/* packet-eigrp.h
 * Routines for EIGRP dissection
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

#ifndef __PACKET_EIGRP_H_
#define __PACKET_EIGRP_H_

extern const value_string eigrp_proto2string[];

#endif
