/* Do not modify this file. Changes will be overwritten.                      */
/* Generated automatically by the ASN.1 to Wireshark dissector compiler       */
/* packet-nbap.c                                                              */
/* asn2wrs.py -L -p nbap -c ./nbap.cnf -s ./packet-nbap-template -D . -O ../.. NBAP-CommonDataTypes.asn NBAP-Constants.asn NBAP-Containers.asn NBAP-IEs.asn NBAP-PDU-Contents.asn NBAP-PDU-Descriptions.asn */

/* packet-nbap-template.c
 * Routines for UMTS Node B Application Part(NBAP) packet dissection
 * Copyright 2005, 2009 Anders Broman <anders.broman@ericsson.com>
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * Ref: 3GPP TS 25.433 version 6.6.0 Release 6
 */

#include "config.h"

#include <epan/packet.h>
#include <epan/sctpppids.h>
#include <epan/asn1.h>
#include <epan/conversation.h>
#include <epan/expert.h>
#include <epan/prefs.h>
#include <epan/proto_data.h>
#include <epan/uat.h>

#include "packet-per.h"
#include "packet-isup.h"
#include "packet-umts_fp.h"
#include "packet-umts_mac.h"
#include "packet-rrc.h"
#include "packet-umts_rlc.h"
#include "packet-nbap.h"

#ifdef _MSC_VER
/* disable: "warning C4146: unary minus operator applied to unsigned type, result still unsigned" */
#pragma warning(disable:4146)
#endif

#define PNAME  "UTRAN Iub interface NBAP signalling"
#define PSNAME "NBAP"
#define PFNAME "nbap"


#define NBAP_IGNORE_PORT 255

/* Debug */
#define DEBUG_NBAP 0
#if DEBUG_NBAP
#include <epan/to_str.h>
#define nbap_debug(...) ws_warning(__VA_ARGS__)
#else
#define nbap_debug(...)
#endif

void proto_register_nbap(void);
void proto_reg_handoff_nbap(void);

/* Protocol Handles */
static dissector_handle_t fp_handle;

#define maxPrivateIEs                  65535
#define maxProtocolExtensions          65535
#define maxProtocolIEs                 65535
#define maxNrOfCodes                   10
#define maxNrOfDLTSs                   15
#define maxNrOfDLTSLCRs                6
#define maxNrOfErrors                  256
#define maxNrOfTFs                     32
#define maxNrOfTFCs                    1024
#define maxNrOfRLs                     16
#define maxNrOfRLs_1                   15
#define maxNrOfRLs_2                   14
#define maxNrOfRLSets                  maxNrOfRLs
#define maxNrOfDPCHs                   240
#define maxNrOfDPCHsPerRL_1            239
#define maxNrOfDPCHLCRs                240
#define maxNrOfDPCHsLCRPerRL_1         95
#define maxNrOfDPCHs768                480
#define maxNrOfDPCHs768PerRL_1         479
#define maxNrOfSCCPCHs                 8
#define maxNrOfSCCPCHsinExt            232
#define maxNrOfSCCPCHs768              480
#define maxNrOfDCHs                    128
#define maxNrOfDSCHs                   32
#define maxNrOfFACHs                   8
#define maxNrOfCCTrCHs                 16
#define maxNrOfPDSCHs                  256
#define maxNrOfHSPDSCHs                16
#define maxNrOfHSPDSCHs768             32
#define maxNrOfPUSCHs                  256
#define maxNrOfPUSCHs_1                255
#define maxNrOfPDSCHSets               256
#define maxNrOfPRACHLCRs               8
#define maxNrOfPUSCHSets               256
#define maxNrOfSCCPCHLCRs              8
#define maxNrOfSCCPCHsLCRinExt         88
#define maxNrOfULTSs                   15
#define maxNrOfULTSLCRs                6
#define maxNrOfUSCHs                   32
#define maxNrOfSlotFormatsPRACH        8
#define maxCellinNodeB                 256
#define maxCCPinNodeB                  256
#define maxCTFC                        16777215
#define maxLocalCellinNodeB            maxCellinNodeB
#define maxFPACHCell                   8
#define maxRACHCell                    maxPRACHCell
#define maxPLCCHCell                   16
#define maxPRACHCell                   16
#define maxSCCPCHCell                  32
#define maxSCCPCHCellinExt             208
#define maxSCCPCHCellinExtLCR          64
#define maxSCCPCHCell768               480
#define maxSCPICHCell                  32
#define maxTTI_count                   4
#define maxIBSEG                       16
#define maxIB                          64
#define maxFACHCell                    256
#define maxRateMatching                256
#define maxHS_PDSCHCodeNrComp_1        15
#define maxHS_SCCHCodeNrComp_1         127
#define maxNrOfCellSyncBursts          10
#define maxNrOfReceptsPerSyncFrame     16
#define maxNrOfMeasNCell               96
#define maxNrOfMeasNCell_1             95
#define maxNrOfSF                      8
#define maxTGPS                        6
#define maxCommunicationContext        1048575
#define maxNrOfLevels                  256
#define maxNoSat                       16
#define maxNoGPSItems                  8
#define maxNrOfHSSCCHs                 32
#define maxNrOfHSSICHs                 4
#define maxNrOfHSSICHs_1               3
#define maxNrOfSyncFramesLCR           512
#define maxNrOfReceptionsperSyncFrameLCR 8
#define maxNrOfSyncDLCodesLCR          32
#define maxNrOfHSSCCHCodes             4
#define maxNrOfMACdFlows               8
#define maxNrOfMACdFlows_1             7
#define maxNrOfMACdPDUIndexes          8
#define maxNrOfMACdPDUIndexes_1        7
#define maxNrOfMACdPDUSize             32
#define maxNrOfNIs                     256
#define maxNrOfPriorityQueues          8
#define maxNrOfPriorityQueues_1        7
#define maxNrOfHARQProcesses           8
#define maxNrOfContextsOnUeList        16
#define maxNrOfCellPortionsPerCell     64
#define maxNrOfCellPortionsPerCell_1   63
#define maxNrOfPriorityClasses         16
#define maxNrOfSatAlmanac_maxNoSat     16
#define maxNrOfE_AGCHs                 32
#define maxNrOfEDCHMACdFlows           8
#define maxNrOfEDCHMACdFlows_1         7
#define maxNrOfE_RGCHs_E_HICHs         32
#define maxNrOfEDCH_HARQ_PO_QUANTSTEPs 6
#define maxNrOfEDCHHARQProcesses2msEDCH 8
#define maxNrOfEDPCCH_PO_QUANTSTEPs    8
#define maxNrOfBits_MACe_PDU_non_scheduled 19982
#define maxNrOfRefETFCIs               8
#define maxNrOfRefETFCI_PO_QUANTSTEPs  29
#define maxNrofSigSeqRGHI_1            39
#define maxNoOfLogicalChannels         16
#define maxNrOfCombEDPDCH              12
#define maxE_RUCCHCell                 16
#define maxNrOfEAGCHCodes              4
#define maxNrOfRefBetas                8
#define maxNrOfE_PUCHSlots             13
#define maxNrOfEAGCHs                  32
#define maxNrOfHS_DSCH_TBSs_HS_SCCHless 4
#define maxNrOfHS_DSCH_TBSs            90
#define maxNrOfEHICHCodes              4
#define maxNrOfE_PUCHSlotsLCR          5
#define maxNrOfEPUCHcodes              16
#define maxNrOfEHICHs                  32
#define maxNrOfCommonMACFlows          8
#define maxNrOfCommonMACFlows_1        7
#define maxNrOfPagingMACFlow           4
#define maxNrOfPagingMACFlow_1         3
#define maxNrOfcommonMACQueues         8
#define maxNrOfpagingMACQueues         8
#define maxNrOfHS_DSCHTBSsE_PCH        2
#define maxGANSSSat                    64
#define maxNoGANSS                     8
#define maxSgnType                     8
#define maxFrequencyinCell             12
#define maxFrequencyinCell_1           11
#define maxHSDPAFrequency              8
#define maxHSDPAFrequency_1            7
#define maxNrOfHSSCCHsinExt            224
#define maxGANSSSatAlmanac             36
#define maxGANSSClockMod               4
#define maxNrOfEDCHRLs                 4
#define maxERNTItoRelease              256
#define maxNrOfCommonEDCH              32
#define maxNrOfCommonMACFlowsLCR       256
#define maxNrOfCommonMACFlowsLCR_1     255
#define maxNrOfHSSCCHsLCR              256
#define maxNrOfEDCHMACdFlowsLCR        256
#define maxNrOfEDCHMACdFlowsLCR_1      255
#define maxNrOfEAGCHsLCR               256
#define maxNrOfEHICHsLCR               256
#define maxnrofERUCCHsLCR              32
#define maxNrOfHSDSCH_1                32
#define maxNrOfHSDSCH                  33
#define maxGANSS_1                     7
#define maxNoOfTBSs_Mapping_HS_DSCH_SPS 4
#define maxNoOfTBSs_Mapping_HS_DSCH_SPS_1 3
#define maxNoOfHS_DSCH_TBSsLCR         64
#define maxNoOfRepetition_Period_LCR   4
#define maxNoOfRepetitionPeriod_SPS_LCR_1 3
#define maxNoOf_HS_SICH_SPS            4
#define maxNoOf_HS_SICH_SPS_1          3
#define maxNoOfNon_HS_SCCH_Assosiated_HS_SICH 4
#define maxNoOfNon_HS_SCCH_Assosiated_HS_SICH_Ext 44
#define maxMBMSServiceSelect           256
#define maxNrOfCellPortionsPerCellLCR  256
#define maxNrOfCellPortionsPerCellLCR_1 255
#define maxNrOfEDCH_1                  32
#define maxNoOfCommonH_RNTI            256
#define maxNrOfCommonMACFlowsLCRExt    248
#define maxofERNTI                     256
#define maxNrOfDCHMeasurementOccasionPatternSequence 6

/* enumerated values for ProcedureCode */
#define NBAP_ID_AUDIT   0
#define NBAP_ID_AUDITREQUIRED   1
#define NBAP_ID_BLOCKRESOURCE   2
#define NBAP_ID_CELLDELETION   3
#define NBAP_ID_CELLRECONFIGURATION   4
#define NBAP_ID_CELLSETUP   5
#define NBAP_ID_CELLSYNCHRONISATIONINITIATION  45
#define NBAP_ID_CELLSYNCHRONISATIONRECONFIGURATION  46
#define NBAP_ID_CELLSYNCHRONISATIONREPORTING  47
#define NBAP_ID_CELLSYNCHRONISATIONTERMINATION  48
#define NBAP_ID_CELLSYNCHRONISATIONFAILURE  49
#define NBAP_ID_COMMONMEASUREMENTFAILURE   6
#define NBAP_ID_COMMONMEASUREMENTINITIATION   7
#define NBAP_ID_COMMONMEASUREMENTREPORT   8
#define NBAP_ID_COMMONMEASUREMENTTERMINATION   9
#define NBAP_ID_COMMONTRANSPORTCHANNELDELETE  10
#define NBAP_ID_COMMONTRANSPORTCHANNELRECONFIGURE  11
#define NBAP_ID_COMMONTRANSPORTCHANNELSETUP  12
#define NBAP_ID_COMPRESSEDMODECOMMAND  14
#define NBAP_ID_DEDICATEDMEASUREMENTFAILURE  16
#define NBAP_ID_DEDICATEDMEASUREMENTINITIATION  17
#define NBAP_ID_DEDICATEDMEASUREMENTREPORT  18
#define NBAP_ID_DEDICATEDMEASUREMENTTERMINATION  19
#define NBAP_ID_DOWNLINKPOWERCONTROL  20
#define NBAP_ID_DOWNLINKPOWERTIMESLOTCONTROL  38
#define NBAP_ID_ERRORINDICATIONFORCOMMON  35
#define NBAP_ID_ERRORINDICATIONFORDEDICATED  21
#define NBAP_ID_INFORMATIONEXCHANGEFAILURE  40
#define NBAP_ID_INFORMATIONEXCHANGEINITIATION  41
#define NBAP_ID_INFORMATIONEXCHANGETERMINATION  42
#define NBAP_ID_INFORMATIONREPORTING  43
#define NBAP_ID_BEARERREARRANGEMENT  50
#define NBAP_ID_MBMSNOTIFICATIONUPDATE  53
#define NBAP_ID_PHYSICALSHAREDCHANNELRECONFIGURATION  37
#define NBAP_ID_PRIVATEMESSAGEFORCOMMON  36
#define NBAP_ID_PRIVATEMESSAGEFORDEDICATED  22
#define NBAP_ID_RADIOLINKADDITION  23
#define NBAP_ID_RADIOLINKDELETION  24
#define NBAP_ID_RADIOLINKFAILURE  25
#define NBAP_ID_RADIOLINKPREEMPTION  39
#define NBAP_ID_RADIOLINKRESTORATION  26
#define NBAP_ID_RADIOLINKSETUP  27
#define NBAP_ID_RESET  13
#define NBAP_ID_RESOURCESTATUSINDICATION  28
#define NBAP_ID_CELLSYNCHRONISATIONADJUSTMENT  44
#define NBAP_ID_SYNCHRONISEDRADIOLINKRECONFIGURATIONCANCELLATION  29
#define NBAP_ID_SYNCHRONISEDRADIOLINKRECONFIGURATIONCOMMIT  30
#define NBAP_ID_SYNCHRONISEDRADIOLINKRECONFIGURATIONPREPARATION  31
#define NBAP_ID_SYSTEMINFORMATIONUPDATE  32
#define NBAP_ID_UNBLOCKRESOURCE  33
#define NBAP_ID_UNSYNCHRONISEDRADIOLINKRECONFIGURATION  34
#define NBAP_ID_RADIOLINKACTIVATION  51
#define NBAP_ID_RADIOLINKPARAMETERUPDATE  52
#define NBAP_ID_UESTATUSUPDATE  54
#define NBAP_ID_SECONDARYULFREQUENCYREPORTING  55
#define NBAP_ID_SECONDARYULFREQUENCYUPDATE  56

typedef enum _DdMode_enum {
  tdd          =   0,
  fdd          =   1,
  common       =   2
} DdMode_enum;

typedef enum _ProtocolIE_ID_enum {
  id_AICH_Information =   0,
  id_AICH_InformationItem_ResourceStatusInd =   1,
  id_BCH_Information =   7,
  id_BCH_InformationItem_ResourceStatusInd =   8,
  id_BCCH_ModificationTime =   9,
  id_BlockingPriorityIndicator =  10,
  id_Cause     =  13,
  id_CCP_InformationItem_AuditRsp =  14,
  id_CCP_InformationList_AuditRsp =  15,
  id_CCP_InformationItem_ResourceStatusInd =  16,
  id_Cell_InformationItem_AuditRsp =  17,
  id_Cell_InformationItem_ResourceStatusInd =  18,
  id_Cell_InformationList_AuditRsp =  19,
  id_CellParameterID =  23,
  id_CFN       =  24,
  id_C_ID      =  25,
  id_Unknown_38 =  38,
  id_CommonMeasurementAccuracy =  39,
  id_CommonMeasurementObjectType_CM_Rprt =  31,
  id_CommonMeasurementObjectType_CM_Rqst =  32,
  id_CommonMeasurementObjectType_CM_Rsp =  33,
  id_CommonMeasurementType =  34,
  id_CommonPhysicalChannelID =  35,
  id_CommonPhysicalChannelType_CTCH_SetupRqstFDD =  36,
  id_CommonPhysicalChannelType_CTCH_SetupRqstTDD =  37,
  id_CommunicationControlPortID =  40,
  id_ConfigurationGenerationID =  43,
  id_CRNC_CommunicationContextID =  44,
  id_CriticalityDiagnostics =  45,
  id_DCHs_to_Add_FDD =  48,
  id_DCH_AddList_RL_ReconfPrepTDD =  49,
  id_DCHs_to_Add_TDD =  50,
  id_DCH_DeleteList_RL_ReconfPrepFDD =  52,
  id_DCH_DeleteList_RL_ReconfPrepTDD =  53,
  id_DCH_DeleteList_RL_ReconfRqstFDD =  54,
  id_DCH_DeleteList_RL_ReconfRqstTDD =  55,
  id_DCH_FDD_Information =  56,
  id_DCH_TDD_Information =  57,
  id_DCH_InformationResponse =  59,
  id_Unknown_60 =  60,
  id_Unknown_61 =  61,
  id_FDD_DCHs_to_Modify =  62,
  id_TDD_DCHs_to_Modify =  63,
  id_DCH_ModifyList_RL_ReconfRqstTDD =  65,
  id_DCH_RearrangeList_Bearer_RearrangeInd = 135,
  id_DedicatedMeasurementObjectType_DM_Rprt =  67,
  id_DedicatedMeasurementObjectType_DM_Rqst =  68,
  id_DedicatedMeasurementObjectType_DM_Rsp =  69,
  id_DedicatedMeasurementType =  70,
  id_Unknown_71 =  71,
  id_DL_CCTrCH_InformationItem_RL_SetupRqstTDD =  72,
  id_DL_CCTrCH_InformationList_RL_AdditionRqstTDD =  73,
  id_DL_CCTrCH_InformationList_RL_SetupRqstTDD =  76,
  id_Unknown_75 =  75,
  id_DL_DPCH_InformationItem_RL_AdditionRqstTDD =  77,
  id_DL_DPCH_InformationList_RL_SetupRqstTDD =  79,
  id_Unknown_80 =  80,
  id_DL_DPCH_Information_RL_ReconfPrepFDD =  81,
  id_DL_DPCH_Information_RL_ReconfRqstFDD =  82,
  id_DL_DPCH_Information_RL_SetupRqstFDD =  83,
  id_DL_DPCH_TimingAdjustment =  21,
  id_DL_ReferencePowerInformationItem_DL_PC_Rqst =  84,
  id_DLReferencePower =  85,
  id_DLReferencePowerList_DL_PC_Rqst =  86,
  id_Unused_ProtocolIE_ID_87 =  87,
  id_Unknown_88 =  88,
  id_Unused_ProtocolIE_ID_89 =  89,
  id_Unused_ProtocolIE_ID_91 =  91,
  id_Unknown_92 =  92,
  id_Unused_ProtocolIE_ID_93 =  93,
  id_Unknown_95 =  95,
  id_DSCHs_to_Add_TDD =  96,
  id_DSCH_Information_DeleteList_RL_ReconfPrepTDD =  98,
  id_DSCH_Information_ModifyList_RL_ReconfPrepTDD = 100,
  id_DSCH_InformationResponse = 105,
  id_Unused_ProtocolIE_ID_106 = 106,
  id_DSCH_TDD_Information = 107,
  id_Unused_ProtocolIE_ID_108 = 108,
  id_Unknown_109 = 109,
  id_Unused_ProtocolIE_ID_112 = 112,
  id_DSCH_RearrangeList_Bearer_RearrangeInd = 136,
  id_End_Of_Audit_Sequence_Indicator = 113,
  id_FACH_Information = 116,
  id_FACH_InformationItem_ResourceStatusInd = 117,
  id_Unknown_118 = 118,
  id_FACH_ParametersList_CTCH_ReconfRqstTDD = 120,
  id_FACH_ParametersListIE_CTCH_SetupRqstFDD = 121,
  id_FACH_ParametersListIE_CTCH_SetupRqstTDD = 122,
  id_IndicationType_ResourceStatusInd = 123,
  id_Local_Cell_ID = 124,
  id_Local_Cell_Group_InformationItem_AuditRsp =   2,
  id_Local_Cell_Group_InformationItem_ResourceStatusInd =   3,
  id_Local_Cell_Group_InformationItem2_ResourceStatusInd =   4,
  id_Local_Cell_Group_InformationList_AuditRsp =   5,
  id_Local_Cell_InformationItem_AuditRsp = 125,
  id_Local_Cell_InformationItem_ResourceStatusInd = 126,
  id_Local_Cell_InformationItem2_ResourceStatusInd = 127,
  id_Local_Cell_InformationList_AuditRsp = 128,
  id_AdjustmentPeriod = 129,
  id_MaxAdjustmentStep = 130,
  id_MaximumTransmissionPower = 131,
  id_MeasurementFilterCoefficient = 132,
  id_MeasurementID = 133,
  id_MessageStructure = 115,
  id_MIB_SB_SIB_InformationList_SystemInfoUpdateRqst = 134,
  id_Unknown_137 = 137,
  id_Unknown_140 = 140,
  id_NodeB_CommunicationContextID = 143,
  id_NeighbouringCellMeasurementInformation = 455,
  id_P_CCPCH_Information = 144,
  id_P_CCPCH_InformationItem_ResourceStatusInd = 145,
  id_P_CPICH_Information = 146,
  id_P_CPICH_InformationItem_ResourceStatusInd = 147,
  id_P_SCH_Information = 148,
  id_Unknown_149 = 149,
  id_PCCPCH_Information_Cell_ReconfRqstTDD = 150,
  id_PCCPCH_Information_Cell_SetupRqstTDD = 151,
  id_Unknown_152 = 152,
  id_Unknown_153 = 153,
  id_PCH_Parameters_CTCH_ReconfRqstTDD = 155,
  id_PCH_ParametersItem_CTCH_SetupRqstFDD = 156,
  id_PCH_ParametersItem_CTCH_SetupRqstTDD = 157,
  id_PCH_Information = 158,
  id_Unknown_159 = 159,
  id_Unknown_160 = 160,
  id_PDSCH_Information_AddListIE_PSCH_ReconfRqst = 161,
  id_PDSCH_Information_ModifyListIE_PSCH_ReconfRqst = 162,
  id_PDSCHSets_AddList_PSCH_ReconfRqst = 163,
  id_PDSCHSets_DeleteList_PSCH_ReconfRqst = 164,
  id_PDSCHSets_ModifyList_PSCH_ReconfRqst = 165,
  id_PICH_Information = 166,
  id_PICH_Parameters_CTCH_ReconfRqstTDD = 168,
  id_PowerAdjustmentType = 169,
  id_PRACH_Information = 170,
  id_Unknown_171 = 171,
  id_Unknown_172 = 172,
  id_Unknown_173 = 173,
  id_PrimaryCCPCH_Information_Cell_ReconfRqstFDD = 175,
  id_PrimaryCCPCH_Information_Cell_SetupRqstFDD = 176,
  id_PrimaryCPICH_Information_Cell_ReconfRqstFDD = 177,
  id_PrimaryCPICH_Information_Cell_SetupRqstFDD = 178,
  id_PrimarySCH_Information_Cell_ReconfRqstFDD = 179,
  id_PrimarySCH_Information_Cell_SetupRqstFDD = 180,
  id_PrimaryScramblingCode = 181,
  id_Unknown_182 = 182,
  id_SCH_Information_Cell_ReconfRqstTDD = 183,
  id_SCH_Information_Cell_SetupRqstTDD = 184,
  id_PUSCH_Information_AddListIE_PSCH_ReconfRqst = 185,
  id_PUSCH_Information_ModifyListIE_PSCH_ReconfRqst = 186,
  id_PUSCHSets_AddList_PSCH_ReconfRqst = 187,
  id_PUSCHSets_DeleteList_PSCH_ReconfRqst = 188,
  id_PUSCHSets_ModifyList_PSCH_ReconfRqst = 189,
  id_RACH_Information = 190,
  id_Unknown_191 = 191,
  id_Unknown_192 = 192,
  id_Unknown_193 = 193,
  id_Unknown_194 = 194,
  id_Unknown_195 = 195,
  id_RACH_ParametersItem_CTCH_SetupRqstFDD = 196,
  id_RACH_ParameterItem_CTCH_SetupRqstTDD = 197,
  id_ReportCharacteristics = 198,
  id_Reporting_Object_RL_FailureInd = 199,
  id_Reporting_Object_RL_RestoreInd = 200,
  id_Unknown_201 = 201,
  id_RL_InformationItem_DM_Rprt = 202,
  id_RL_InformationItem_DM_Rqst = 203,
  id_RL_InformationItem_DM_Rsp = 204,
  id_RL_InformationItem_RL_AdditionRqstFDD = 205,
  id_RL_informationItem_RL_DeletionRqst = 206,
  id_RL_InformationItem_RL_FailureInd = 207,
  id_RL_InformationItem_RL_PreemptRequiredInd = 286,
  id_RL_InformationItem_RL_ReconfPrepFDD = 208,
  id_RL_InformationItem_RL_ReconfRqstFDD = 209,
  id_RL_InformationItem_RL_RestoreInd = 210,
  id_RL_InformationItem_RL_SetupRqstFDD = 211,
  id_RL_InformationList_RL_AdditionRqstFDD = 212,
  id_RL_informationList_RL_DeletionRqst = 213,
  id_RL_InformationList_RL_PreemptRequiredInd = 237,
  id_RL_InformationList_RL_ReconfPrepFDD = 214,
  id_RL_InformationList_RL_ReconfRqstFDD = 215,
  id_RL_InformationList_RL_SetupRqstFDD = 216,
  id_RL_InformationResponseItem_RL_AdditionRspFDD = 217,
  id_RL_InformationResponseItem_RL_ReconfReady = 218,
  id_RL_InformationResponseItem_RL_ReconfRsp = 219,
  id_RL_InformationResponseItem_RL_SetupRspFDD = 220,
  id_RL_InformationResponseList_RL_AdditionRspFDD = 221,
  id_RL_InformationResponseList_RL_ReconfReady = 222,
  id_RL_InformationResponseList_RL_ReconfRsp = 223,
  id_RL_InformationResponseList_RL_SetupRspFDD = 224,
  id_RL_InformationResponse_RL_AdditionRspTDD = 225,
  id_RL_InformationResponse_RL_SetupRspTDD = 226,
  id_RL_Information_RL_AdditionRqstTDD = 227,
  id_RL_Information_RL_ReconfRqstTDD = 228,
  id_RL_Information_RL_ReconfPrepTDD = 229,
  id_RL_Information_RL_SetupRqstTDD = 230,
  id_RL_ReconfigurationFailureItem_RL_ReconfFailure = 236,
  id_RL_Set_InformationItem_DM_Rprt = 238,
  id_Unknown_239 = 239,
  id_RL_Set_InformationItem_DM_Rsp = 240,
  id_RL_Set_InformationItem_RL_FailureInd = 241,
  id_RL_Set_InformationItem_RL_RestoreInd = 242,
  id_Unknown_243 = 243,
  id_Unknown_244 = 244,
  id_Unknown_245 = 245,
  id_Unknown_246 = 246,
  id_S_CCPCH_Information = 247,
  id_Unknown_248 = 248,
  id_S_CPICH_Information = 249,
  id_Unknown_250 = 250,
  id_SCH_Information = 251,
  id_Unknown_252 = 252,
  id_S_SCH_Information = 253,
  id_Unknown_254 = 254,
  id_Unknown_255 = 255,
  id_Unknown_256 = 256,
  id_Secondary_CCPCHListIE_CTCH_ReconfRqstTDD = 257,
  id_Secondary_CCPCH_parameterListIE_CTCH_SetupRqstTDD = 258,
  id_Secondary_CCPCH_Parameters_CTCH_ReconfRqstTDD = 259,
  id_SecondaryCPICH_InformationItem_Cell_ReconfRqstFDD = 260,
  id_SecondaryCPICH_InformationItem_Cell_SetupRqstFDD = 261,
  id_SecondaryCPICH_InformationList_Cell_ReconfRqstFDD = 262,
  id_SecondaryCPICH_InformationList_Cell_SetupRqstFDD = 263,
  id_SecondarySCH_Information_Cell_ReconfRqstFDD = 264,
  id_SecondarySCH_Information_Cell_SetupRqstFDD = 265,
  id_SegmentInformationListIE_SystemInfoUpdate = 266,
  id_Unknown_267 = 267,
  id_SFN       = 268,
  id_SignallingBearerRequestIndicator = 138,
  id_ShutdownTimer = 269,
  id_Start_Of_Audit_Sequence_Indicator = 114,
  id_Successful_RL_InformationRespItem_RL_AdditionFailureFDD = 270,
  id_Successful_RL_InformationRespItem_RL_SetupFailureFDD = 271,
  id_Unknown_272 = 272,
  id_Unknown_273 = 273,
  id_SyncCase  = 274,
  id_SyncCaseIndicatorItem_Cell_SetupRqstTDD_PSCH = 275,
  id_T_Cell    = 276,
  id_TargetCommunicationControlPortID = 139,
  id_TimeSlotConfigurationList_Cell_ReconfRqstTDD = 277,
  id_TimeSlotConfigurationList_Cell_SetupRqstTDD = 278,
  id_TransmissionDiversityApplied = 279,
  id_TypeOfError = 508,
  id_UARFCNforNt = 280,
  id_UARFCNforNd = 281,
  id_UARFCNforNu = 282,
  id_UL_CCTrCH_InformationItem_RL_SetupRqstTDD = 284,
  id_UL_CCTrCH_InformationList_RL_AdditionRqstTDD = 285,
  id_UL_CCTrCH_InformationList_RL_SetupRqstTDD = 288,
  id_UL_DPCH_InformationItem_RL_AdditionRqstTDD = 289,
  id_UL_DPCH_InformationList_RL_SetupRqstTDD = 291,
  id_UL_DPCH_Information_RL_ReconfPrepFDD = 293,
  id_UL_DPCH_Information_RL_ReconfRqstFDD = 294,
  id_UL_DPCH_Information_RL_SetupRqstFDD = 295,
  id_Unsuccessful_RL_InformationRespItem_RL_AdditionFailureFDD = 296,
  id_Unsuccessful_RL_InformationRespItem_RL_SetupFailureFDD = 297,
  id_Unknown_298 = 298,
  id_Unknown_299 = 299,
  id_Unsuccessful_RL_InformationResp_RL_AdditionFailureTDD = 300,
  id_Unsuccessful_RL_InformationResp_RL_SetupFailureTDD = 301,
  id_USCH_Information_Add = 302,
  id_Unknown_303 = 303,
  id_USCH_Information_DeleteList_RL_ReconfPrepTDD = 304,
  id_Unknown_305 = 305,
  id_USCH_Information_ModifyList_RL_ReconfPrepTDD = 306,
  id_Unknown_307 = 307,
  id_Unknown_308 = 308,
  id_USCH_InformationResponse = 309,
  id_USCH_Information = 310,
  id_USCH_RearrangeList_Bearer_RearrangeInd = 141,
  id_Unknown_313 = 313,
  id_Active_Pattern_Sequence_Information = 315,
  id_AICH_ParametersListIE_CTCH_ReconfRqstFDD = 316,
  id_AdjustmentRatio = 317,
  id_Unknown_318 = 318,
  id_Unknown_319 = 319,
  id_Not_Used_320 = 320,
  id_Unknown_321 = 321,
  id_Not_Used_322 = 322,
  id_FACH_ParametersListIE_CTCH_ReconfRqstFDD = 323,
  id_CauseLevel_PSCH_ReconfFailure = 324,
  id_CauseLevel_RL_AdditionFailureFDD = 325,
  id_CauseLevel_RL_AdditionFailureTDD = 326,
  id_CauseLevel_RL_ReconfFailure = 327,
  id_CauseLevel_RL_SetupFailureFDD = 328,
  id_CauseLevel_RL_SetupFailureTDD = 329,
  id_Not_Used_330 = 330,
  id_Unknown_331 = 331,
  id_Not_Used_332 = 332,
  id_Closed_Loop_Timing_Adjustment_Mode = 333,
  id_CommonPhysicalChannelType_CTCH_ReconfRqstFDD = 334,
  id_Compressed_Mode_Deactivation_Flag = 335,
  id_Not_Used_336 = 336,
  id_Unknown_337 = 337,
  id_Unknown_338 = 338,
  id_Unknown_339 = 339,
  id_Unknown_340 = 340,
  id_Unknown_341 = 341,
  id_Not_Used_342 = 342,
  id_Not_Used_343 = 343,
  id_Unknown_344 = 344,
  id_Unknown_345 = 345,
  id_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD = 346,
  id_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD = 347,
  id_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD = 348,
  id_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD = 349,
  id_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD = 350,
  id_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD = 351,
  id_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD = 352,
  id_DL_DPCH_InformationAddListIE_RL_ReconfPrepTDD = 353,
  id_Unknown_354 = 354,
  id_DL_DPCH_InformationModify_AddListIE_RL_ReconfPrepTDD = 355,
  id_DL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD = 356,
  id_DL_DPCH_InformationModify_ModifyListIE_RL_ReconfPrepTDD = 357,
  id_DL_TPC_Pattern01Count = 358,
  id_DPC_Mode  = 450,
  id_DPCHConstant = 359,
  id_Unused_ProtocolIE_ID_94 =  94,
  id_Unused_ProtocolIE_ID_110 = 110,
  id_Unused_ProtocolIE_ID_111 = 111,
  id_Unknown_360 = 360,
  id_Unknown_361 = 361,
  id_FACH_ParametersList_CTCH_SetupRsp = 362,
  id_Unknown_363 = 363,
  id_Unknown_364 = 364,
  id_Unknown_365 = 365,
  id_Unknown_366 = 366,
  id_Unknown_367 = 367,
  id_Unknown_368 = 368,
  id_Limited_power_increase_information_Cell_SetupRqstFDD = 369,
  id_Unknown_370 = 370,
  id_Unknown_371 = 371,
  id_Unknown_372 = 372,
  id_Unknown_373 = 373,
  id_PCH_Parameters_CTCH_SetupRsp = 374,
  id_PCH_ParametersItem_CTCH_ReconfRqstFDD = 375,
  id_Not_Used_376 = 376,
  id_Unknown_377 = 377,
  id_Unknown_378 = 378,
  id_Unknown_379 = 379,
  id_PICH_ParametersItem_CTCH_ReconfRqstFDD = 380,
  id_PRACHConstant = 381,
  id_Unknown_382 = 382,
  id_PRACH_ParametersListIE_CTCH_ReconfRqstFDD = 383,
  id_PUSCHConstant = 384,
  id_RACH_Parameters_CTCH_SetupRsp = 385,
  id_Unknown_386 = 386,
  id_Unknown_387 = 387,
  id_Unknown_388 = 388,
  id_Unknown_389 = 389,
  id_Unknown_390 = 390,
  id_Unknown_391 = 391,
  id_Unknown_392 = 392,
  id_Unused_ProtocolIE_ID_443 = 443,
  id_Synchronisation_Configuration_Cell_ReconfRqst = 393,
  id_Synchronisation_Configuration_Cell_SetupRqst = 394,
  id_Transmission_Gap_Pattern_Sequence_Information = 395,
  id_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD = 396,
  id_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD = 397,
  id_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD = 398,
  id_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD = 399,
  id_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD = 400,
  id_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD = 401,
  id_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD = 402,
  id_UL_DPCH_InformationAddListIE_RL_ReconfPrepTDD = 403,
  id_Unknown_404 = 404,
  id_UL_DPCH_InformationModify_AddListIE_RL_ReconfPrepTDD = 405,
  id_UL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD = 406,
  id_UL_DPCH_InformationModify_ModifyListIE_RL_ReconfPrepTDD = 407,
  id_Unsuccessful_PDSCHSetItem_PSCH_ReconfFailureTDD = 408,
  id_Unsuccessful_PUSCHSetItem_PSCH_ReconfFailureTDD = 409,
  id_Unknown_410 = 410,
  id_Unknown_411 = 411,
  id_CommunicationContextInfoItem_Reset = 412,
  id_Unknown_413 = 413,
  id_CommunicationControlPortInfoItem_Reset = 414,
  id_Unknown_415 = 415,
  id_ResetIndicator = 416,
  id_Unused_ProtocolIE_ID_417 = 417,
  id_Unused_ProtocolIE_ID_418 = 418,
  id_Unused_ProtocolIE_ID_419 = 419,
  id_Unused_ProtocolIE_ID_142 = 142,
  id_TimingAdvanceApplied = 287,
  id_CFNReportingIndicator =   6,
  id_SFNReportingIndicator =  11,
  id_InnerLoopDLPCStatus =  12,
  id_TimeslotISCPInfo = 283,
  id_PICH_ParametersItem_CTCH_SetupRqstTDD = 167,
  id_PRACH_ParametersItem_CTCH_SetupRqstTDD =  20,
  id_CCTrCH_InformationItem_RL_FailureInd =  46,
  id_CCTrCH_InformationItem_RL_RestoreInd =  47,
  id_CauseLevel_SyncAdjustmntFailureTDD = 420,
  id_CellAdjustmentInfo_SyncAdjustmntRqstTDD = 421,
  id_CellAdjustmentInfoItem_SyncAdjustmentRqstTDD = 494,
  id_CellSyncBurstInfoList_CellSyncReconfRqstTDD = 482,
  id_CellSyncBurstTransInit_CellSyncInitiationRqstTDD = 422,
  id_CellSyncBurstMeasureInit_CellSyncInitiationRqstTDD = 423,
  id_CellSyncBurstTransReconfiguration_CellSyncReconfRqstTDD = 424,
  id_CellSyncBurstMeasReconfiguration_CellSyncReconfRqstTDD = 425,
  id_CellSyncBurstTransInfoList_CellSyncReconfRqstTDD = 426,
  id_CellSyncBurstMeasInfoList_CellSyncReconfRqstTDD = 427,
  id_CellSyncBurstTransReconfInfo_CellSyncReconfRqstTDD = 428,
  id_CellSyncInfo_CellSyncReprtTDD = 429,
  id_CSBTransmissionID = 430,
  id_CSBMeasurementID = 431,
  id_IntStdPhCellSyncInfoItem_CellSyncReprtTDD = 432,
  id_NCyclesPerSFNperiod = 433,
  id_NRepetitionsPerCyclePeriod = 434,
  id_SyncFrameNumber = 437,
  id_SynchronisationReportType = 438,
  id_SynchronisationReportCharacteristics = 439,
  id_Unsuccessful_cell_InformationRespItem_SyncAdjustmntFailureTDD = 440,
  id_Unknown_441 = 441,
  id_Unknown_442 = 442,
  id_LateEntranceCellSyncInfoItem_CellSyncReprtTDD = 119,
  id_ReferenceClockAvailability = 435,
  id_ReferenceSFNoffset = 436,
  id_InformationExchangeID = 444,
  id_InformationExchangeObjectType_InfEx_Rqst = 445,
  id_InformationType = 446,
  id_InformationReportCharacteristics = 447,
  id_InformationExchangeObjectType_InfEx_Rsp = 448,
  id_InformationExchangeObjectType_InfEx_Rprt = 449,
  id_IPDLParameter_Information_Cell_ReconfRqstFDD = 451,
  id_IPDLParameter_Information_Cell_SetupRqstFDD = 452,
  id_IPDLParameter_Information_Cell_ReconfRqstTDD = 453,
  id_IPDLParameter_Information_Cell_SetupRqstTDD = 454,
  id_DL_DPCH_LCR_Information_RL_SetupRqstTDD =  74,
  id_DwPCH_LCR_Information =  78,
  id_DwPCH_LCR_InformationList_AuditRsp =  90,
  id_DwPCH_LCR_Information_Cell_SetupRqstTDD =  97,
  id_DwPCH_LCR_Information_Cell_ReconfRqstTDD =  99,
  id_DwPCH_LCR_Information_ResourceStatusInd = 101,
  id_maxFACH_Power_LCR_CTCH_SetupRqstTDD = 154,
  id_maxFACH_Power_LCR_CTCH_ReconfRqstTDD = 174,
  id_FPACH_LCR_Information = 290,
  id_FPACH_LCR_Information_AuditRsp = 292,
  id_FPACH_LCR_InformationList_AuditRsp =  22,
  id_FPACH_LCR_InformationList_ResourceStatusInd = 311,
  id_FPACH_LCR_Parameters_CTCH_SetupRqstTDD = 312,
  id_FPACH_LCR_Parameters_CTCH_ReconfRqstTDD = 314,
  id_PCCPCH_LCR_Information_Cell_SetupRqstTDD = 456,
  id_PCH_Power_LCR_CTCH_SetupRqstTDD = 457,
  id_PCH_Power_LCR_CTCH_ReconfRqstTDD = 458,
  id_PICH_LCR_Parameters_CTCH_SetupRqstTDD = 459,
  id_Unknown_460 = 460,
  id_PRACH_LCR_ParametersList_CTCH_SetupRqstTDD = 461,
  id_Unknown_462 = 462,
  id_RL_InformationResponse_LCR_RL_SetupRspTDD = 463,
  id_Unknown_464 = 464,
  id_Secondary_CCPCH_LCR_parameterList_CTCH_SetupRqstTDD = 465,
  id_TimeSlot  = 495,
  id_TimeSlotConfigurationList_LCR_Cell_ReconfRqstTDD = 466,
  id_TimeSlotConfigurationList_LCR_Cell_SetupRqstTDD = 467,
  id_TimeslotISCP_LCR_InfoList_RL_SetupRqstTDD = 468,
  id_TimeSlotLCR_CM_Rqst = 469,
  id_UL_DPCH_LCR_Information_RL_SetupRqstTDD = 470,
  id_Unknown_471 = 471,
  id_DL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD = 472,
  id_UL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD = 473,
  id_TimeslotISCP_InformationList_LCR_RL_AdditionRqstTDD = 474,
  id_DL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD = 475,
  id_Unknown_476 = 476,
  id_DL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD = 477,
  id_Unknown_478 = 478,
  id_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD = 479,
  id_TimeslotISCPInfoList_LCR_DL_PC_RqstTDD = 480,
  id_UL_DPCH_LCR_InformationAddListIE_RL_ReconfPrepTDD = 481,
  id_UL_DPCH_LCR_InformationModify_AddList = 483,
  id_Unknown_484 = 484,
  id_UL_TimeslotLCR_Information_RL_ReconfPrepTDD = 485,
  id_UL_SIRTarget = 510,
  id_PDSCH_AddInformation_LCR_PSCH_ReconfRqst = 486,
  id_PDSCH_AddInformation_LCR_AddListIE_PSCH_ReconfRqst = 487,
  id_Unused_ProtocolIE_ID_26 =  26,
  id_Unused_ProtocolIE_ID_27 =  27,
  id_PDSCH_ModifyInformation_LCR_PSCH_ReconfRqst = 488,
  id_PDSCH_ModifyInformation_LCR_ModifyListIE_PSCH_ReconfRqst = 489,
  id_PUSCH_AddInformation_LCR_PSCH_ReconfRqst = 490,
  id_PUSCH_AddInformation_LCR_AddListIE_PSCH_ReconfRqst = 491,
  id_PUSCH_ModifyInformation_LCR_PSCH_ReconfRqst = 492,
  id_PUSCH_ModifyInformation_LCR_ModifyListIE_PSCH_ReconfRqst = 493,
  id_timeslotInfo_CellSyncInitiationRqstTDD = 496,
  id_SyncReportType_CellSyncReprtTDD = 497,
  id_Power_Local_Cell_Group_InformationItem_AuditRsp = 498,
  id_Power_Local_Cell_Group_InformationItem_ResourceStatusInd = 499,
  id_Power_Local_Cell_Group_InformationItem2_ResourceStatusInd = 500,
  id_Power_Local_Cell_Group_InformationList_AuditRsp = 501,
  id_Power_Local_Cell_Group_InformationList_ResourceStatusInd = 502,
  id_Power_Local_Cell_Group_InformationList2_ResourceStatusInd = 503,
  id_Power_Local_Cell_Group_ID = 504,
  id_PUSCH_Info_DM_Rqst = 505,
  id_PUSCH_Info_DM_Rsp = 506,
  id_PUSCH_Info_DM_Rprt = 507,
  id_InitDL_Power = 509,
  id_cellSyncBurstRepetitionPeriod = 511,
  id_ReportCharacteristicsType_OnModification = 512,
  id_SFNSFNMeasurementValueInformation = 513,
  id_SFNSFNMeasurementThresholdInformation = 514,
  id_TUTRANGPSMeasurementValueInformation = 515,
  id_TUTRANGPSMeasurementThresholdInformation = 516,
  id_Rx_Timing_Deviation_Value_LCR = 520,
  id_RL_InformationResponse_LCR_RL_AdditionRspTDD =  51,
  id_DL_PowerBalancing_Information =  28,
  id_DL_PowerBalancing_ActivationIndicator =  29,
  id_DL_PowerBalancing_UpdatedIndicator =  30,
  id_CCTrCH_Initial_DL_Power_RL_SetupRqstTDD = 517,
  id_CCTrCH_Initial_DL_Power_RL_AdditionRqstTDD = 518,
  id_CCTrCH_Initial_DL_Power_RL_ReconfPrepTDD = 519,
  id_IPDLParameter_Information_LCR_Cell_SetupRqstTDD =  41,
  id_IPDLParameter_Information_LCR_Cell_ReconfRqstTDD =  42,
  id_HS_PDSCH_HS_SCCH_E_AGCH_E_RGCH_E_HICH_MaxPower_PSCH_ReconfRqst = 522,
  id_HS_PDSCH_HS_SCCH_ScramblingCode_PSCH_ReconfRqst = 523,
  id_HS_PDSCH_FDD_Code_Information_PSCH_ReconfRqst = 524,
  id_HS_SCCH_FDD_Code_Information_PSCH_ReconfRqst = 525,
  id_HS_PDSCH_TDD_Information_PSCH_ReconfRqst = 526,
  id_Add_To_HS_SCCH_Resource_Pool_PSCH_ReconfRqst = 527,
  id_Modify_HS_SCCH_Resource_Pool_PSCH_ReconfRqst = 528,
  id_Delete_From_HS_SCCH_Resource_Pool_PSCH_ReconfRqst = 529,
  id_bindingID = 102,
  id_RL_Specific_DCH_Info = 103,
  id_transportlayeraddress = 104,
  id_DelayedActivation = 231,
  id_DelayedActivationList_RL_ActivationCmdFDD = 232,
  id_DelayedActivationInformation_RL_ActivationCmdFDD = 233,
  id_DelayedActivationList_RL_ActivationCmdTDD = 234,
  id_DelayedActivationInformation_RL_ActivationCmdTDD = 235,
  id_neighbouringTDDCellMeasurementInformationLCR =  58,
  id_SYNCDlCodeId_TransInitLCR_CellSyncInitiationRqstTDD = 543,
  id_SYNCDlCodeId_MeasureInitLCR_CellSyncInitiationRqstTDD = 544,
  id_SYNCDlCodeIdTransReconfInfoLCR_CellSyncReconfRqstTDD = 545,
  id_SYNCDlCodeIdMeasReconfigurationLCR_CellSyncReconfRqstTDD = 546,
  id_SYNCDlCodeIdMeasInfoList_CellSyncReconfRqstTDD = 547,
  id_SyncDLCodeIdsMeasInfoList_CellSyncReprtTDD = 548,
  id_SyncDLCodeIdThreInfoLCR = 549,
  id_NSubCyclesPerCyclePeriod_CellSyncReconfRqstTDD = 550,
  id_DwPCH_Power = 551,
  id_AccumulatedClockupdate_CellSyncReprtTDD = 552,
  id_Angle_Of_Arrival_Value_LCR = 521,
  id_HSDSCH_FDD_Information = 530,
  id_HSDSCH_FDD_Information_Response = 531,
  id_Unknown_532 = 532,
  id_Unknown_533 = 533,
  id_HSDSCH_Information_to_Modify = 534,
  id_HSDSCH_RNTI = 535,
  id_HSDSCH_TDD_Information = 536,
  id_HSDSCH_TDD_Information_Response = 537,
  id_Unknown_538 = 538,
  id_Unknown_539 = 539,
  id_Unknown_540 = 540,
  id_HSPDSCH_RL_ID = 541,
  id_PrimCCPCH_RSCP_DL_PC_RqstTDD = 542,
  id_Unused_ProtocolIE_ID_64 =  64,
  id_PDSCH_RL_ID =  66,
  id_HSDSCH_RearrangeList_Bearer_RearrangeInd = 553,
  id_UL_Synchronisation_Parameters_LCR = 554,
  id_HSDSCH_FDD_Update_Information = 555,
  id_HSDSCH_TDD_Update_Information = 556,
  id_Unknown_557 = 557,
  id_DL_DPCH_TimeSlotFormat_LCR_ModifyItem_RL_ReconfPrepTDD = 558,
  id_UL_DPCH_TimeSlotFormat_LCR_ModifyItem_RL_ReconfPrepTDD = 559,
  id_TDD_TPC_UplinkStepSize_LCR_RL_SetupRqstTDD = 560,
  id_TDD_TPC_UplinkStepSize_LCR_RL_AdditionRqstTDD = 561,
  id_TDD_TPC_DownlinkStepSize_RL_AdditionRqstTDD = 562,
  id_TDD_TPC_UplinkStepSize_InformationAdd_LCR_RL_ReconfPrepTDD = 563,
  id_TDD_TPC_UplinkStepSize_InformationModify_LCR_RL_ReconfPrepTDD = 564,
  id_TDD_TPC_DownlinkStepSize_InformationModify_RL_ReconfPrepTDD = 565,
  id_TDD_TPC_DownlinkStepSize_InformationAdd_RL_ReconfPrepTDD = 566,
  id_CCTrCH_Maximum_DL_Power_RL_SetupRqstTDD = 567,
  id_CCTrCH_Minimum_DL_Power_RL_SetupRqstTDD = 568,
  id_CCTrCH_Maximum_DL_Power_RL_AdditionRqstTDD = 569,
  id_CCTrCH_Minimum_DL_Power_RL_AdditionRqstTDD = 570,
  id_CCTrCH_Maximum_DL_Power_InformationAdd_RL_ReconfPrepTDD = 571,
  id_CCTrCH_Minimum_DL_Power_InformationAdd_RL_ReconfPrepTDD = 572,
  id_CCTrCH_Maximum_DL_Power_InformationModify_RL_ReconfPrepTDD = 573,
  id_CCTrCH_Minimum_DL_Power_InformationModify_RL_ReconfPrepTDD = 574,
  id_Maximum_DL_Power_Modify_LCR_InformationModify_RL_ReconfPrepTDD = 575,
  id_Minimum_DL_Power_Modify_LCR_InformationModify_RL_ReconfPrepTDD = 576,
  id_DL_DPCH_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD = 577,
  id_CCTrCH_Maximum_DL_Power_InformationModify_RL_ReconfRqstTDD = 578,
  id_CCTrCH_Minimum_DL_Power_InformationModify_RL_ReconfRqstTDD = 579,
  id_Initial_DL_Power_TimeslotLCR_InformationItem = 580,
  id_Maximum_DL_Power_TimeslotLCR_InformationItem = 581,
  id_Minimum_DL_Power_TimeslotLCR_InformationItem = 582,
  id_HS_DSCHProvidedBitRateValueInformation = 583,
  id_Unknown_584 = 584,
  id_HS_DSCHRequiredPowerValueInformation = 585,
  id_HS_DSCHRequiredPowerValue = 586,
  id_TransmittedCarrierPowerOfAllCodesNotUsedForHSTransmission = 587,
  id_HS_SICH_Reception_Quality = 588,
  id_HS_SICH_Reception_Quality_Measurement_Value = 589,
  id_HSSICH_Info_DM_Rprt = 590,
  id_HSSICH_Info_DM_Rqst = 591,
  id_HSSICH_Info_DM_Rsp = 592,
  id_Best_Cell_Portions_Value = 593,
  id_Primary_CPICH_Usage_for_Channel_Estimation = 594,
  id_Secondary_CPICH_Information_Change = 595,
  id_NumberOfReportedCellPortions = 596,
  id_CellPortion_InformationItem_Cell_SetupRqstFDD = 597,
  id_CellPortion_InformationList_Cell_SetupRqstFDD = 598,
  id_TimeslotISCP_LCR_InfoList_RL_ReconfPrepTDD = 599,
  id_Secondary_CPICH_Information = 600,
  id_Received_total_wide_band_power_For_CellPortion = 601,
  id_Unidirectional_DCH_Indicator = 602,
  id_TimingAdjustmentValueLCR = 603,
  id_multipleRL_dl_DPCH_InformationList = 604,
  id_multipleRL_dl_DPCH_InformationModifyList = 605,
  id_multipleRL_ul_DPCH_InformationList = 606,
  id_multipleRL_ul_DPCH_InformationModifyList = 607,
  id_RL_ID     = 608,
  id_SAT_Info_Almanac_ExtItem = 609,
  id_HSDPA_Capability = 610,
  id_HSDSCH_Resources_Information_AuditRsp = 611,
  id_HSDSCH_Resources_Information_ResourceStatusInd = 612,
  id_HSDSCH_MACdFlows_to_Add = 613,
  id_HSDSCH_MACdFlows_to_Delete = 614,
  id_HSDSCH_Information_to_Modify_Unsynchronised = 615,
  id_TnlQos    = 616,
  id_Received_total_wide_band_power_For_CellPortion_Value = 617,
  id_Transmitted_Carrier_Power_For_CellPortion = 618,
  id_Transmitted_Carrier_Power_For_CellPortion_Value = 619,
  id_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortion = 620,
  id_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue = 621,
  id_UpPTSInterferenceValue = 622,
  id_PrimaryCCPCH_RSCP_Delta = 623,
  id_MeasurementRecoveryBehavior = 624,
  id_MeasurementRecoveryReportingIndicator = 625,
  id_MeasurementRecoverySupportIndicator = 626,
  id_Tstd_indicator = 627,
  id_multiple_RL_Information_RL_ReconfPrepTDD = 628,
  id_multiple_RL_Information_RL_ReconfRqstTDD = 629,
  id_DL_DPCH_Power_Information_RL_ReconfPrepFDD = 630,
  id_F_DPCH_Information_RL_ReconfPrepFDD = 631,
  id_F_DPCH_Information_RL_SetupRqstFDD = 632,
  id_Additional_S_CCPCH_Parameters_CTCH_ReconfRqstTDD = 633,
  id_Additional_S_CCPCH_Parameters_CTCH_SetupRqstTDD = 634,
  id_Additional_S_CCPCH_LCR_Parameters_CTCH_ReconfRqstTDD = 635,
  id_Additional_S_CCPCH_LCR_Parameters_CTCH_SetupRqstTDD = 636,
  id_MICH_CFN  = 637,
  id_MICH_Information_AuditRsp = 638,
  id_MICH_Information_ResourceStatusInd = 639,
  id_MICH_Parameters_CTCH_ReconfRqstFDD = 640,
  id_MICH_Parameters_CTCH_ReconfRqstTDD = 641,
  id_MICH_Parameters_CTCH_SetupRqstFDD = 642,
  id_MICH_Parameters_CTCH_SetupRqstTDD = 643,
  id_Modification_Period = 644,
  id_NI_Information_NotifUpdateCmd = 645,
  id_S_CCPCH_InformationListExt_AuditRsp = 646,
  id_S_CCPCH_InformationListExt_ResourceStatusInd = 647,
  id_S_CCPCH_LCR_InformationListExt_AuditRsp = 648,
  id_S_CCPCH_LCR_InformationListExt_ResourceStatusInd = 649,
  id_HARQ_Preamble_Mode = 650,
  id_Initial_DL_DPCH_TimingAdjustment = 651,
  id_Initial_DL_DPCH_TimingAdjustment_Allowed = 652,
  id_DLTransmissionBranchLoadValue = 653,
  id_Power_Local_Cell_Group_choice_CM_Rqst = 654,
  id_Power_Local_Cell_Group_choice_CM_Rsp = 655,
  id_Power_Local_Cell_Group_choice_CM_Rprt = 656,
  id_SynchronisationIndicator = 657,
  id_HSDPA_And_EDCH_CellPortion_Information_PSCH_ReconfRqst = 658,
  id_Unused_ProtocolIE_ID_659 = 659,
  id_HS_DSCHRequiredPowerValue_For_Cell_Portion = 660,
  id_HS_DSCHRequiredPowerValueInformation_For_CellPortion = 661,
  id_HS_DSCHProvidedBitRateValueInformation_For_CellPortion = 662,
  id_E_AGCH_And_E_RGCH_E_HICH_FDD_Scrambling_Code = 663,
  id_E_AGCH_FDD_Code_Information = 664,
  id_E_DCH_Capability = 665,
  id_E_DCH_FDD_DL_Control_Channel_Information = 666,
  id_E_DCH_FDD_Information = 667,
  id_E_DCH_FDD_Information_Response = 668,
  id_E_DCH_FDD_Information_to_Modify = 669,
  id_E_DCH_MACdFlows_to_Add = 670,
  id_E_DCH_MACdFlows_to_Delete = 671,
  id_E_DCH_Resources_Information_AuditRsp = 672,
  id_E_DCH_Resources_Information_ResourceStatusInd = 673,
  id_E_DCH_RL_Indication = 674,
  id_E_DCH_RL_Set_ID = 675,
  id_E_DPCH_Information_RL_ReconfPrepFDD = 676,
  id_E_DPCH_Information_RL_SetupRqstFDD = 677,
  id_E_RGCH_E_HICH_FDD_Code_Information = 678,
  id_Serving_E_DCH_RL_ID = 679,
  id_UL_DPDCH_Indicator_For_E_DCH_Operation = 680,
  id_FDD_S_CCPCH_FrameOffset_CTCH_SetupRqstFDD = 681,
  id_E_DPCH_Information_RL_ReconfRqstFDD = 682,
  id_Maximum_Target_ReceivedTotalWideBandPower = 683,
  id_E_DCHProvidedBitRateValueInformation = 684,
  id_HARQ_Preamble_Mode_Activation_Indicator = 685,
  id_RL_Specific_E_DCH_Info = 686,
  id_E_DCH_CapacityConsumptionLaw = 687,
  id_multiple_DedicatedMeasurementValueList_TDD_DM_Rsp = 688,
  id_multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp = 689,
  id_E_DCH_RearrangeList_Bearer_RearrangeInd = 690,
  id_Unused_ProtocolIE_ID_691 = 691,
  id_multipleRL_dl_CCTrCH_InformationModifyList_RL_ReconfRqstTDD = 692,
  id_Target_NonServing_EDCH_To_Total_EDCH_Power_Ratio = 693,
  id_CellPortion_InformationItem_Cell_ReconfRqstFDD = 694,
  id_CellPortion_InformationList_Cell_ReconfRqstFDD = 695,
  id_multiple_PUSCH_InfoList_DM_Rsp = 696,
  id_multiple_PUSCH_InfoList_DM_Rprt = 697,
  id_Reference_ReceivedTotalWideBandPower = 698,
  id_E_DCH_Serving_Cell_Change_Info_Response = 699,
  id_HS_DSCH_Serving_Cell_Change_Info = 700,
  id_HS_DSCH_Serving_Cell_Change_Info_Response = 701,
  id_Serving_Cell_Change_CFN = 702,
  id_E_DCH_HARQ_Combining_Capability = 703,
  id_E_DCH_TTI2ms_Capability = 704,
  id_E_DCH_SF_Capability = 705,
  id_E_DCH_FDD_Update_Information = 706,
  id_F_DPCH_Capability = 707,
  id_E_DCH_Non_serving_Relative_Grant_Down_CommandsValue = 708,
  id_HSSICH_SIRTarget = 709,
  id_multiple_HSSICHMeasurementValueList_TDD_DM_Rsp = 710,
  id_PLCCH_Information_AuditRsp = 711,
  id_PLCCH_Information_ResourceStatusInd = 712,
  id_PLCCH_Information_RL_ReconfPrepTDDLCR = 713,
  id_PLCCH_Information_UL_TimeslotLCR_Info = 714,
  id_PLCCH_InformationList_AuditRsp = 715,
  id_PLCCH_InformationList_ResourceStatusInd = 716,
  id_PLCCH_Parameters_CTCH_ReconfRqstTDD = 717,
  id_S_CCPCH_768_Parameters_CTCH_SetupRqstTDD = 718,
  id_PICH_768_Parameters_CTCH_SetupRqstTDD = 719,
  id_PRACH_768_Parameters_CTCH_SetupRqstTDD = 720,
  id_S_CCPCH_768_Parameters_CTCH_ReconfRqstTDD = 721,
  id_PICH_768_Parameters_CTCH_ReconfRqstTDD = 722,
  id_MICH_768_Parameters_CTCH_ReconfRqstTDD = 723,
  id_CommonPhysicalChannelID768_CommonTrChDeletionReq = 724,
  id_S_CCPCH_768_InformationList_AuditRsp = 725,
  id_S_CCPCH_768_Information_AuditRsp = 726,
  id_neighbouringTDDCellMeasurementInformation768 = 727,
  id_PCCPCH_768_Information_Cell_SetupRqstTDD = 728,
  id_SCH_768_Information_Cell_SetupRqstTDD = 729,
  id_SCH_768_Information_Cell_ReconfRqstTDD = 730,
  id_PCCPCH_768_Information_Cell_ReconfRqstTDD = 731,
  id_P_CCPCH_768_Information_AuditRsp = 732,
  id_PICH_768_Information_AuditRsp = 733,
  id_PRACH_768_InformationList_AuditRsp = 734,
  id_SCH_768_Information_AuditRsp = 735,
  id_MICH_768_Information_AuditRsp = 736,
  id_PRACH_768_Information = 737,
  id_S_CCPCH_768_Information_ResourceStatusInd = 738,
  id_P_CCPCH_768_Information_ResourceStatusInd = 739,
  id_PICH_768_Information_ResourceStatusInd = 740,
  id_PRACH_768_InformationList_ResourceStatusInd = 741,
  id_SCH_768_Information_ResourceStatusInd = 742,
  id_MICH_768_Information_ResourceStatusInd = 743,
  id_S_CCPCH_768_InformationList_ResourceStatusInd = 744,
  id_UL_DPCH_768_Information_RL_SetupRqstTDD = 745,
  id_DL_DPCH_768_Information_RL_SetupRqstTDD = 746,
  id_DL_DPCH_InformationItem_768_RL_AdditionRqstTDD = 747,
  id_UL_DPCH_InformationItem_768_RL_AdditionRqstTDD = 748,
  id_UL_DPCH_768_InformationAddItemIE_RL_ReconfPrepTDD = 749,
  id_UL_DPCH_768_InformationAddListIE_RL_ReconfPrepTDD = 750,
  id_UL_DPCH_768_InformationModify_AddItem = 751,
  id_UL_DPCH_768_InformationModify_AddList = 752,
  id_UL_Timeslot768_Information_RL_ReconfPrepTDD = 753,
  id_DL_DPCH_768_InformationAddItem_RL_ReconfPrepTDD = 754,
  id_DL_DPCH_768_InformationAddList_RL_ReconfPrepTDD = 755,
  id_DL_DPCH_768_InformationModify_AddItem_RL_ReconfPrepTDD = 756,
  id_DL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD = 757,
  id_DL_Timeslot_768_InformationModify_ModifyList_RL_ReconfPrepTDD = 758,
  id_DPCH_ID768_DM_Rqst = 759,
  id_multiple_DedicatedMeasurementValueList_768_TDD_DM_Rsp = 760,
  id_DPCH_ID768_DM_Rsp = 761,
  id_Rx_Timing_Deviation_Value_768 = 762,
  id_DPCH_ID768_DM_Rprt = 763,
  id_PDSCH_AddInformation_768_PSCH_ReconfRqst = 764,
  id_PDSCH_ModifyInformation_768_PSCH_ReconfRqst = 765,
  id_PUSCH_AddInformation_768_PSCH_ReconfRqst = 766,
  id_PUSCH_ModifyInformation_768_PSCH_ReconfRqst = 767,
  id_dL_HS_PDSCH_Timeslot_Information_768_PSCH_ReconfRqst = 768,
  id_hS_SCCH_Information_768_PSCH_ReconfRqst = 769,
  id_hS_SCCH_InformationModify_768_PSCH_ReconfRqst = 770,
  id_hsSCCH_Specific_Information_ResponseTDD768 = 771,
  id_E_DPCH_Information_RL_AdditionReqFDD = 772,
  id_Unknown_773 = 773,
  id_Unknown_774 = 774,
  id_PDSCH_Timeslot_Format_PSCH_ReconfRqst_LCR = 775,
  id_Unknown_776 = 776,
  id_Unknown_777 = 777,
  id_Unknown_778 = 778,
  id_Unknown_779 = 779,
  id_PUSCH_Timeslot_Format_PSCH_ReconfRqst_LCR = 780,
  id_Unknown_781 = 781,
  id_E_DCH_PowerOffset_for_SchedulingInfo = 782,
  id_HSDSCH_Configured_Indicator = 783,
  id_Unknown_784 = 784,
  id_Unknown_785 = 785,
  id_Rx_Timing_Deviation_Value_384_ext = 786,
  id_RTWP_ReportingIndicator = 787,
  id_RTWP_CellPortion_ReportingIndicator = 788,
  id_Received_Scheduled_EDCH_Power_Share_Value = 789,
  id_Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value = 790,
  id_Received_Scheduled_EDCH_Power_Share = 791,
  id_Received_Scheduled_EDCH_Power_Share_For_CellPortion = 792,
  id_tFCI_Presence = 793,
  id_HSSICH_TPC_StepSize = 794,
  id_E_RUCCH_InformationList_AuditRsp = 795,
  id_E_RUCCH_InformationList_ResourceStatusInd = 796,
  id_E_DCH_TDD_CapacityConsumptionLaw = 797,
  id_E_RUCCH_Information = 798,
  id_E_DCH_Information = 799,
  id_E_DCH_Information_Response = 800,
  id_E_DCH_Information_Reconfig = 801,
  id_E_PUCH_Information_PSCH_ReconfRqst = 802,
  id_Add_To_E_AGCH_Resource_Pool_PSCH_ReconfRqst = 803,
  id_Modify_E_AGCH_Resource_Pool_PSCH_ReconfRqst = 804,
  id_Delete_From_E_AGCH_Resource_Pool_PSCH_ReconfRqst = 805,
  id_E_HICH_Information_PSCH_ReconfRqst = 806,
  id_E_HICH_TimeOffset = 807,
  id_Maximum_Generated_ReceivedTotalWideBandPowerInOtherCells = 808,
  id_E_DCH_Serving_RL_ID = 809,
  id_E_RUCCH_768_InformationList_AuditRsp = 810,
  id_E_RUCCH_768_InformationList_ResourceStatusInd = 811,
  id_E_RUCCH_768_Information = 812,
  id_E_DCH_768_Information = 813,
  id_E_DCH_768_Information_Reconfig = 814,
  id_E_PUCH_Information_768_PSCH_ReconfRqst = 815,
  id_Add_To_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst = 816,
  id_Modify_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst = 817,
  id_E_HICH_Information_768_PSCH_ReconfRqst = 818,
  id_ExtendedPropagationDelay = 819,
  id_Extended_Round_Trip_Time_Value = 820,
  id_AlternativeFormatReportingIndicator = 821,
  id_DCH_Indicator_For_E_DCH_HSDPA_Operation = 822,
  id_Reference_ReceivedTotalWideBandPowerReporting = 823,
  id_Reference_ReceivedTotalWideBandPowerSupportIndicator = 824,
  id_ueCapability_Info = 825,
  id_MAChs_ResetIndicator = 826,
  id_Fast_Reconfiguration_Mode = 827,
  id_Fast_Reconfiguration_Permission = 828,
  id_BroadcastReference = 829,
  id_BroadcastCommonTransportBearerIndication = 830,
  id_ContinuousPacketConnectivityDTX_DRX_Capability = 831,
  id_ContinuousPacketConnectivityDTX_DRX_Information = 832,
  id_ContinuousPacketConnectivityHS_SCCH_less_Capability = 833,
  id_ContinuousPacketConnectivityHS_SCCH_less_Information = 834,
  id_ContinuousPacketConnectivityHS_SCCH_less_Information_Response = 835,
  id_CPC_Information = 836,
  id_MIMO_Capability = 837,
  id_MIMO_PilotConfiguration = 838,
  id_Unknown_839 = 839,
  id_Unknown_840 = 840,
  id_MBSFN_Cell_ParameterID_Cell_SetupRqstTDD = 841,
  id_MBSFN_Cell_ParameterID_Cell_ReconfRqstTDD = 842,
  id_S_CCPCH_Modulation = 843,
  id_HS_PDSCH_Code_Change_Grant = 844,
  id_HS_PDSCH_Code_Change_Indicator = 845,
  id_SYNC_UL_Partition_LCR = 846,
  id_E_DCH_LCR_Information = 847,
  id_E_DCH_LCR_Information_Reconfig = 848,
  id_Unknown_849 = 849,
  id_Unknown_850 = 850,
  id_Unknown_851 = 851,
  id_E_PUCH_Information_LCR_PSCH_ReconfRqst = 852,
  id_Add_To_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst = 853,
  id_Modify_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst = 854,
  id_Add_To_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst = 855,
  id_Modify_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst = 856,
  id_Delete_From_E_HICH_Resource_Pool_PSCH_ReconfRqst = 857,
  id_E_HICH_TimeOffsetLCR = 858,
  id_Unknown_859 = 859,
  id_SixtyfourQAM_DL_Capability = 860,
  id_SixteenQAM_UL_Capability = 861,
  id_Unknown_862 = 862,
  id_Unknown_863 = 863,
  id_HSDSCH_MACdPDU_SizeCapability = 864,
  id_HSDSCH_MACdPDUSizeFormat = 865,
  id_MaximumMACdPDU_SizeExtended = 866,
  id_Unknown_867 = 867,
  id_Unknown_868 = 868,
  id_Unknown_869 = 869,
  id_F_DPCH_SlotFormat = 870,
  id_F_DPCH_SlotFormatCapability = 871,
  id_LCRTDD_uplink_Physical_Channel_Capability = 872,
  id_Extended_RNC_ID = 873,
  id_Max_UE_DTX_Cycle = 874,
  id_Unknown_875 = 875,
  id_Secondary_CCPCH_SlotFormat_Extended = 876,
  id_Unknown_877 = 877,
  id_MBSFN_Only_Mode_Indicator_Cell_SetupRqstTDD_LCR = 878,
  id_MBSFN_Only_Mode_Capability = 879,
  id_Time_Slot_Parameter_ID = 880,
  id_Additional_failed_HS_SICH = 881,
  id_Additional_missed_HS_SICH = 882,
  id_Additional_total_HS_SICH = 883,
  id_Additional_HS_SICH_Reception_Quality_Measurement_Value = 884,
  id_Unknown_885 = 885,
  id_Unknown_886 = 886,
  id_GANSS_Common_Data = 887,
  id_GANSS_Information = 888,
  id_GANSS_Generic_Data = 889,
  id_TUTRANGANSSMeasurementThresholdInformation = 890,
  id_TUTRANGANSSMeasurementValueInformation = 891,
  id_ModulationPO_MBSFN = 892,
  id_Unknown_893 = 893,
  id_Unknown_894 = 894,
  id_Enhanced_FACH_Capability = 895,
  id_Enhanced_PCH_Capability = 896,
  id_HSDSCH_Common_System_InformationFDD = 897,
  id_HSDSCH_Common_System_Information_ResponseFDD = 898,
  id_HSDSCH_Paging_System_InformationFDD = 899,
  id_HSDSCH_Paging_System_Information_ResponseFDD = 900,
  id_MBMS_Capability = 901,
  id_Ext_Reference_E_TFCI_PO = 902,
  id_Ext_Max_Bits_MACe_PDU_non_scheduled = 903,
  id_HARQ_MemoryPartitioningInfoExtForMIMO = 904,
  id_MIMO_ActivationIndicator = 905,
  id_MIMO_Mode_Indicator = 906,
  id_MIMO_N_M_Ratio = 907,
  id_IPMulticastIndication = 908,
  id_IPMulticastDataBearerIndication = 909,
  id_TransportBearerNotSetupIndicator = 910,
  id_TransportBearerNotRequestedIndicator = 911,
  id_TimeSlotConfigurationList_LCR_CTCH_SetupRqstTDD = 912,
  id_Cell_Frequency_List_Information_LCR_MulFreq_AuditRsp = 913,
  id_Cell_Frequency_List_InformationItem_LCR_MulFreq_AuditRsp = 914,
  id_Cell_Frequency_List_LCR_MulFreq_Cell_SetupRqstTDD = 915,
  id_UARFCN_Adjustment = 916,
  id_Cell_Frequency_List_Information_LCR_MulFreq_ResourceStatusInd = 917,
  id_Cell_Frequency_List_InformationItem_LCR_MulFreq_ResourceStatusInd = 918,
  id_UPPCHPositionLCR = 919,
  id_UPPCH_LCR_Parameters_CTCH_ReconfRqstTDD = 920,
  id_UPPCH_LCR_InformationList_AuditRsp = 921,
  id_UPPCH_LCR_InformationItem_AuditRsp = 922,
  id_UPPCH_LCR_InformationList_ResourceStatusInd = 923,
  id_UPPCH_LCR_InformationItem_ResourceStatusInd = 924,
  id_multipleFreq_dL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst = 925,
  id_number_Of_Supported_Carriers = 926,
  id_multipleFreq_HSPDSCH_InformationList_ResponseTDDLCR = 927,
  id_Unsuccessful_UARFCNItem_PSCH_ReconfFailureTDD = 928,
  id_multipleFreq_HS_DSCH_Resources_InformationList_AuditRsp = 929,
  id_multipleFreq_HS_DSCH_Resources_InformationList_ResourceStatusInd = 930,
  id_UARFCNSpecificCauseList = 931,
  id_tSN_Length = 932,
  id_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst = 933,
  id_multicarrier_number = 934,
  id_Extended_HS_SCCH_ID = 935,
  id_Extended_HS_SICH_ID = 936,
  id_HSSICH_InfoExt_DM_Rqst = 937,
  id_Delete_From_HS_SCCH_Resource_PoolExt_PSCH_ReconfRqst = 938,
  id_HS_SCCH_InformationExt_LCR_PSCH_ReconfRqst = 939,
  id_HS_SCCH_InformationModifyExt_LCR_PSCH_ReconfRqst = 940,
  id_PowerControlGAP = 941,
  id_MBSFN_SpecialTimeSlot_LCR = 942,
  id_Common_MACFlows_to_DeleteFDD = 943,
  id_Paging_MACFlows_to_DeleteFDD = 944,
  id_E_TFCI_Boost_Information = 945,
  id_SixteenQAM_UL_Operation_Indicator = 946,
  id_SixtyfourQAM_UsageAllowedIndicator = 947,
  id_SixtyfourQAM_DL_UsageIndicator = 948,
  id_Default_Serving_Grant_in_DTX_Cycle2 = 949,
  id_Maximum_Target_ReceivedTotalWideBandPower_LCR = 950,
  id_E_DPDCH_PowerInterpolation = 951,
  id_Extended_E_DCH_LCRTDD_PhysicalLayerCategory = 952,
  id_MultipleFreq_E_DCH_Resources_InformationList_AuditRsp = 953,
  id_MultipleFreq_E_DCH_Resources_InformationList_ResourceStatusInd = 954,
  id_MultipleFreq_E_PUCH_Timeslot_InformationList_LCR_PSCH_ReconfRqst = 955,
  id_MultipleFreq_E_PUCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst = 956,
  id_Extended_E_HICH_ID_TDD = 957,
  id_ContinuousPacketConnectivityHS_SCCH_less_Deactivate_Indicator = 958,
  id_E_DCH_MACdPDU_SizeCapability = 959,
  id_E_DCH_MACdPDUSizeFormat = 960,
  id_MaximumNumber_Of_Retransmission_for_Scheduling_Info_LCRTDD = 961,
  id_E_DCH_RetransmissionTimer_for_SchedulingInfo_LCRTDD = 962,
  id_E_HICH_TimeOffset_Extension = 963,
  id_MultipleFreq_E_HICH_TimeOffsetLCR = 964,
  id_E_PUCH_PowerControlGAP = 965,
  id_HSDSCH_TBSizeTableIndicator = 966,
  id_E_DCH_DL_Control_Channel_Change_Information = 967,
  id_E_DCH_DL_Control_Channel_Grant_Information = 968,
  id_DGANSS_Corrections_Req = 969,
  id_UE_with_enhanced_HS_SCCH_support_indicator = 970,
  id_AdditionalTimeSlotListLCR = 971,
  id_AdditionalMeasurementValueList = 972,
  id_Unknown_973 = 973,
  id_Unknown_974 = 974,
  id_Unknown_975 = 975,
  id_Unknown_976 = 976,
  id_Unknown_977 = 977,
  id_E_AGCH_Table_Choice = 978,
  id_Unknown_979 = 979,
  id_Unknown_980 = 980,
  id_PLCCH_parameters = 981,
  id_E_RUCCH_parameters = 982,
  id_E_RUCCH_768_parameters = 983,
  id_HS_Cause  = 984,
  id_E_Cause   = 985,
  id_Unknown_986 = 986,
  id_Common_EDCH_Capability = 987,
  id_E_AI_Capability = 988,
  id_Common_EDCH_System_InformationFDD = 989,
  id_Common_UL_MACFlows_to_DeleteFDD = 990,
  id_Common_EDCH_MACdFlows_to_DeleteFDD = 991,
  id_Common_EDCH_System_Information_ResponseFDD = 992,
  id_Cell_ERNTI_Status_Information = 993,
  id_Enhanced_UE_DRX_Capability = 994,
  id_Enhanced_UE_DRX_InformationFDD = 995,
  id_TransportBearerRequestIndicator = 996,
  id_SixtyfourQAM_DL_MIMO_Combined_Capability = 997,
  id_E_RNTI    = 998,
  id_MinimumReducedE_DPDCH_GainFactor = 999,
  id_GANSS_Time_ID = 1000,
  id_GANSS_AddIonoModelReq = 1001,
  id_GANSS_EarthOrientParaReq = 1002,
  id_GANSS_AddNavigationModelsReq = 1003,
  id_GANSS_AddUTCModelsReq = 1004,
  id_GANSS_AuxInfoReq = 1005,
  id_GANSS_SBAS_ID = 1006,
  id_GANSS_ID  = 1007,
  id_GANSS_Additional_Ionospheric_Model = 1008,
  id_GANSS_Earth_Orientation_Parameters = 1009,
  id_GANSS_Additional_Time_Models = 1010,
  id_GANSS_Additional_Navigation_Models = 1011,
  id_GANSS_Additional_UTC_Models = 1012,
  id_GANSS_Auxiliary_Information = 1013,
  id_ERACH_CM_Rqst = 1014,
  id_ERACH_CM_Rsp = 1015,
  id_ERACH_CM_Rprt = 1016,
  id_EDCH_RACH_Report_Value = 1017,
  id_EDCH_RACH_Report_IncrDecrThres = 1018,
  id_EDCH_RACH_Report_ThresholdInformation = 1019,
  id_E_DPCCH_Power_Boosting_Capability = 1020,
  id_HSDSCH_Common_System_InformationLCR = 1021,
  id_Unknown_1022 = 1022,
  id_HSDSCH_Common_System_Information_ResponseLCR = 1222,
  id_HSDSCH_Paging_System_InformationLCR = 1023,
  id_HSDSCH_Paging_System_Information_ResponseLCR = 1024,
  id_Common_MACFlows_to_DeleteLCR = 1025,
  id_Paging_MACFlows_to_DeleteLCR = 1026,
  id_Common_EDCH_System_InformationLCR = 1027,
  id_Common_UL_MACFlows_to_DeleteLCR = 1028,
  id_Common_EDCH_MACdFlows_to_DeleteLCR = 1029,
  id_Common_EDCH_System_Information_ResponseLCR = 1030,
  id_Enhanced_UE_DRX_CapabilityLCR = 1031,
  id_Enhanced_UE_DRX_InformationLCR = 1032,
  id_HSDSCH_PreconfigurationSetup = 1033,
  id_HSDSCH_PreconfigurationInfo = 1034,
  id_NoOfTargetCellHS_SCCH_Order = 1035,
  id_EnhancedHSServingCC_Abort = 1036,
  id_Additional_HS_Cell_Information_RL_Setup = 1037,
  id_Additional_HS_Cell_Information_Response = 1038,
  id_Additional_HS_Cell_Information_RL_Addition = 1039,
  id_Additional_HS_Cell_Change_Information_Response = 1040,
  id_Additional_HS_Cell_Information_RL_Reconf_Prep = 1041,
  id_Additional_HS_Cell_Information_RL_Reconf_Req = 1042,
  id_Additional_HS_Cell_Information_RL_Param_Upd = 1043,
  id_Multi_Cell_Capability_Info = 1044,
  id_IMB_Parameters = 1045,
  id_MACes_Maximum_Bitrate_LCR = 1046,
  id_Semi_PersistentScheduling_CapabilityLCR = 1047,
  id_E_DCH_Semi_PersistentScheduling_Information_LCR = 1048,
  id_HS_DSCH_Semi_PersistentScheduling_Information_LCR = 1049,
  id_Add_To_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst = 1050,
  id_Modify_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst = 1051,
  id_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst = 1052,
  id_ContinuousPacketConnectivity_DRX_CapabilityLCR = 1053,
  id_ContinuousPacketConnectivity_DRX_InformationLCR = 1054,
  id_ContinuousPacketConnectivity_DRX_Information_ResponseLCR = 1055,
  id_CPC_InformationLCR = 1056,
  id_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR = 1057,
  id_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR = 1058,
  id_E_AGCH_UE_Inactivity_Monitor_Threshold = 1059,
  id_Unknown_1060 = 1060,
  id_Unknown_1061 = 1061,
  id_Unknown_1062 = 1062,
  id_IdleIntervalInformation = 1063,
  id_GANSS_alm_keplerianNAVAlmanac = 1064,
  id_GANSS_alm_keplerianReducedAlmanac = 1065,
  id_GANSS_alm_keplerianMidiAlmanac = 1066,
  id_GANSS_alm_keplerianGLONASS = 1067,
  id_GANSS_alm_ecefSBASAlmanac = 1068,
  id_Unknown_1069 = 1069,
  id_HSSICH_ReferenceSignal_InformationLCR = 1070,
  id_MIMO_ReferenceSignal_InformationListLCR = 1071,
  id_MIMO_SFMode_For_HSPDSCHDualStream = 1072,
  id_MIMO_SFMode_Supported_For_HSPDSCHDualStream = 1073,
  id_UE_Selected_MBMS_Service_Information = 1074,
  id_Unknown_1075 = 1075,
  id_Unknown_1076 = 1076,
  id_MultiCarrier_HSDSCH_Physical_Layer_Category = 1077,
  id_Common_E_DCH_HSDPCCH_Capability = 1078,
  id_DL_RLC_PDU_Size_Format = 1079,
  id_HSSICH_ReferenceSignal_InformationModifyLCR = 1080,
  id_schedulingPriorityIndicator = 1081,
  id_TimeSlotMeasurementValueListLCR = 1082,
  id_Unknown_1083 = 1083,
  id_Unknown_1084 = 1084,
  id_UE_SupportIndicatorExtension = 1085,
  id_Unknown_1086 = 1086,
  id_Unknown_1087 = 1087,
  id_Single_Stream_MIMO_ActivationIndicator = 1088,
  id_Single_Stream_MIMO_Capability = 1089,
  id_Single_Stream_MIMO_Mode_Indicator = 1090,
  id_Dual_Band_Capability_Info = 1091,
  id_UE_AggregateMaximumBitRate = 1092,
  id_UE_AggregateMaximumBitRate_Enforcement_Indicator = 1093,
  id_Unknown_1094 = 1094,
  id_Unknown_1095 = 1095,
  id_Unknown_1096 = 1096,
  id_Unknown_1097 = 1097,
  id_Unknown_1098 = 1098,
  id_Unknown_1099 = 1099,
  id_Unknown_1100 = 1100,
  id_MIMO_Power_Offset_For_S_CPICH_Capability = 1101,
  id_MIMO_PilotConfigurationExtension = 1102,
  id_TxDiversityOnDLControlChannelsByMIMOUECapability = 1103,
  id_ULTimeslotISCPValue_For_CellPortion = 1104,
  id_UpPTSInterferenceValue_For_CellPortion = 1105,
  id_Best_Cell_Portions_ValueLCR = 1106,
  id_Transmitted_Carrier_Power_For_CellPortion_ValueLCR = 1107,
  id_Received_total_wide_band_power_For_CellPortion_ValueLCR = 1108,
  id_UL_TimeslotISCP_For_CellPortion_Value = 1109,
  id_HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR = 1110,
  id_HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR = 1111,
  id_E_DCHProvidedBitRateValueInformation_For_CellPortion = 1112,
  id_UpPTSInterference_For_CellPortion_Value = 1113,
  id_NumberOfReportedCellPortionsLCR = 1114,
  id_CellPortion_CapabilityLCR = 1115,
  id_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue = 1116,
  id_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortion = 1117,
  id_Unknown_1118 = 1118,
  id_ActivationInformation = 1119,
  id_Additional_EDCH_Cell_Information_RL_Setup_Req = 1120,
  id_Additional_EDCH_Cell_Information_Response = 1121,
  id_Additional_EDCH_Cell_Information_RL_Add_Req = 1122,
  id_Additional_EDCH_Cell_Information_Response_RL_Add = 1123,
  id_Additional_EDCH_Cell_Information_RL_Reconf_Prep = 1124,
  id_Additional_EDCH_Cell_Information_RL_Reconf_Req = 1125,
  id_Additional_EDCH_Cell_Information_Bearer_Rearrangement = 1126,
  id_Additional_EDCH_Cell_Information_RL_Param_Upd = 1127,
  id_Additional_EDCH_Preconfiguration_Information = 1128,
  id_EDCH_Indicator = 1129,
  id_Unknown_1130 = 1130,
  id_HS_DSCH_SPS_Reservation_Indicator = 1131,
  id_E_DCH_SPS_Reservation_Indicator = 1132,
  id_MultipleFreq_HARQ_MemoryPartitioning_InformationList = 1133,
  id_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_Ext = 1134,
  id_RepetitionPeriodIndex = 1135,
  id_MidambleShiftLCR = 1136,
  id_MaxHSDSCH_HSSCCH_Power_per_CELLPORTION = 1137,
  id_DormantModeIndicator = 1138,
  id_DiversityMode = 1139,
  id_TransmitDiversityIndicator = 1140,
  id_NonCellSpecificTxDiversity = 1141,
  id_Cell_Capability_Container = 1142,
  id_E_RNTI_List_Request = 1143,
  id_E_RNTI_List = 1144,
  id_PowerControlGAP_For_CellFACHLCR = 1145,
  id_Unknown_1146 = 1146,
  id_UL_Synchronisation_Parameters_For_FACHLCR = 1147,
  id_HS_DSCH_SPS_Operation_Indicator = 1148,
  id_HSDSCH_RNTI_For_FACH = 1149,
  id_E_RNTI_For_FACH = 1150,
  id_Out_of_Sychronization_Window = 1151,
  id_Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst = 1152,
  id_E_HICH_TimeOffset_ReconfFailureTDD = 1153,
  id_HSSCCH_TPC_StepSize = 1154,
  id_TS0_CapabilityLCR = 1155,
  id_UE_TS0_CapabilityLCR = 1156,
  id_Common_System_Information_ResponseLCR = 1157,
  id_Additional_EDCH_Cell_Information_ResponseRLReconf = 1158,
  id_Multicell_EDCH_InformationItemIEs = 1159,
  id_Multicell_EDCH_RL_Specific_InformationItemIEs = 1160,
  id_Add_To_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext = 1161,
  id_Modify_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext = 1162,
  id_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext = 1163,
  id_Initial_DL_Transmission_Power = 1164,
  id_Maximum_DL_Power = 1165,
  id_Minimum_DL_Power = 1166,
  id_DCH_MeasurementOccasion_Information = 1167,
  id_AssociatedPhsicalChannelID = 1168,
  id_DGNSS_ValidityPeriod = 1169,
  id_PhysicalChannelID_for_CommonERNTI_RequestedIndicator = 1170,
  id_PrecodingWeightSetRestriction = 1171,
  id_Treset_Usage_Indicator = 1172,
  id_Non_Serving_RL_Preconfig_Info = 1173,
  id_Non_Serving_RL_Preconfig_Setup = 1174,
  id_Non_Serving_RL_Preconfig_Removal = 1175,
  id_Additional_E_DCH_Non_Serving_RL_Preconfiguration_Setup = 1176,
  id_Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList = 1177,
  id_Ul_common_E_DCH_MACflow_Specific_InfoListLCR_Ext = 1178,
  id_CommonMACFlow_Specific_InfoList_ResponseLCR_Ext = 1179,
  id_Enabling_Delay_Ext_LCR = 1180,
  id_Unallocated_1181 = 1181,
  id_Unallocated_1182 = 1182,
  id_Unallocated_1183 = 1183,
  id_Unallocated_1184 = 1184,
  id_Unallocated_1185 = 1185,
  id_Unallocated_1186 = 1186,
  id_Unallocated_1187 = 1187,
  id_Unallocated_1188 = 1188,
  id_Unallocated_1189 = 1189,
  id_Unallocated_1190 = 1190,
  id_Unallocated_1191 = 1191,
  id_Unallocated_1192 = 1192,
  id_Unallocated_1193 = 1193,
  id_Unallocated_1194 = 1194,
  id_Unallocated_1195 = 1195,
  id_Unallocated_1196 = 1196,
  id_Unallocated_1197 = 1197,
  id_Unallocated_1198 = 1198,
  id_Unallocated_1199 = 1199,
  id_Unallocated_1200 = 1200,
  id_Unallocated_1201 = 1201,
  id_Unallocated_1202 = 1202,
  id_Unallocated_1203 = 1203,
  id_Unallocated_1204 = 1204,
  id_Unallocated_1205 = 1205,
  id_Unallocated_1206 = 1206,
  id_Unallocated_1207 = 1207,
  id_Unallocated_1208 = 1208,
  id_Unallocated_1209 = 1209,
  id_Unallocated_1210 = 1210,
  id_Unallocated_1211 = 1211,
  id_Unallocated_1212 = 1212,
  id_Unallocated_1213 = 1213,
  id_Unallocated_1214 = 1214,
  id_Unallocated_1215 = 1215,
  id_Unallocated_1216 = 1216,
  id_Unallocated_1217 = 1217,
  id_Unallocated_1218 = 1218,
  id_Unallocated_1219 = 1219,
  id_Unallocated_1220 = 1220,
  id_Unallocated_1221 = 1221
} ProtocolIE_ID_enum;

/* Initialize the protocol and registered fields */
static int proto_nbap = -1;
static int hf_nbap_transportLayerAddress_ipv4 = -1;
static int hf_nbap_transportLayerAddress_ipv6 = -1;
static int hf_nbap_transportLayerAddress_nsap = -1;
static int hf_nbap_reassembled_information_block = -1;

static int hf_nbap_ActivationInformation_PDU = -1;  /* ActivationInformation */
static int hf_nbap_Additional_EDCH_Setup_Info_PDU = -1;  /* Additional_EDCH_Setup_Info */
static int hf_nbap_Additional_EDCH_Cell_Information_Response_List_PDU = -1;  /* Additional_EDCH_Cell_Information_Response_List */
static int hf_nbap_Additional_EDCH_Cell_Information_Response_RLReconf_List_PDU = -1;  /* Additional_EDCH_Cell_Information_Response_RLReconf_List */
static int hf_nbap_Additional_EDCH_Cell_Information_Response_RL_Add_List_PDU = -1;  /* Additional_EDCH_Cell_Information_Response_RL_Add_List */
static int hf_nbap_AdditionalMeasurementValueList_PDU = -1;  /* AdditionalMeasurementValueList */
static int hf_nbap_AdditionalTimeSlotListLCR_PDU = -1;  /* AdditionalTimeSlotListLCR */
static int hf_nbap_Active_Pattern_Sequence_Information_PDU = -1;  /* Active_Pattern_Sequence_Information */
static int hf_nbap_AlternativeFormatReportingIndicator_PDU = -1;  /* AlternativeFormatReportingIndicator */
static int hf_nbap_Angle_Of_Arrival_Value_LCR_PDU = -1;  /* Angle_Of_Arrival_Value_LCR */
static int hf_nbap_BCCH_ModificationTime_PDU = -1;  /* BCCH_ModificationTime */
static int hf_nbap_Best_Cell_Portions_Value_PDU = -1;  /* Best_Cell_Portions_Value */
static int hf_nbap_Best_Cell_Portions_ValueLCR_PDU = -1;  /* Best_Cell_Portions_ValueLCR */
static int hf_nbap_BindingID_PDU = -1;            /* BindingID */
static int hf_nbap_BlockingPriorityIndicator_PDU = -1;  /* BlockingPriorityIndicator */
static int hf_nbap_BroadcastCommonTransportBearerIndication_PDU = -1;  /* BroadcastCommonTransportBearerIndication */
static int hf_nbap_BroadcastReference_PDU = -1;   /* BroadcastReference */
static int hf_nbap_Cause_PDU = -1;                /* Cause */
static int hf_nbap_Cell_Capability_Container_PDU = -1;  /* Cell_Capability_Container */
static int hf_nbap_Cell_ERNTI_Status_Information_PDU = -1;  /* Cell_ERNTI_Status_Information */
static int hf_nbap_CellParameterID_PDU = -1;      /* CellParameterID */
static int hf_nbap_CellPortion_CapabilityLCR_PDU = -1;  /* CellPortion_CapabilityLCR */
static int hf_nbap_CellSyncBurstRepetitionPeriod_PDU = -1;  /* CellSyncBurstRepetitionPeriod */
static int hf_nbap_CFN_PDU = -1;                  /* CFN */
static int hf_nbap_C_ID_PDU = -1;                 /* C_ID */
static int hf_nbap_Closedlooptimingadjustmentmode_PDU = -1;  /* Closedlooptimingadjustmentmode */
static int hf_nbap_Common_EDCH_Capability_PDU = -1;  /* Common_EDCH_Capability */
static int hf_nbap_Common_E_DCH_HSDPCCH_Capability_PDU = -1;  /* Common_E_DCH_HSDPCCH_Capability */
static int hf_nbap_Common_EDCH_System_InformationFDD_PDU = -1;  /* Common_EDCH_System_InformationFDD */
static int hf_nbap_Common_EDCH_System_Information_ResponseFDD_PDU = -1;  /* Common_EDCH_System_Information_ResponseFDD */
static int hf_nbap_E_RNTI_List_PDU = -1;          /* E_RNTI_List */
static int hf_nbap_Common_MACFlows_to_DeleteFDD_PDU = -1;  /* Common_MACFlows_to_DeleteFDD */
static int hf_nbap_CommonMeasurementAccuracy_PDU = -1;  /* CommonMeasurementAccuracy */
static int hf_nbap_CommonMeasurementType_PDU = -1;  /* CommonMeasurementType */
static int hf_nbap_CommonPhysicalChannelID_PDU = -1;  /* CommonPhysicalChannelID */
static int hf_nbap_CommonPhysicalChannelID768_PDU = -1;  /* CommonPhysicalChannelID768 */
static int hf_nbap_Common_PhysicalChannel_Status_Information_PDU = -1;  /* Common_PhysicalChannel_Status_Information */
static int hf_nbap_Common_PhysicalChannel_Status_Information768_PDU = -1;  /* Common_PhysicalChannel_Status_Information768 */
static int hf_nbap_CommonTransportChannel_InformationResponse_PDU = -1;  /* CommonTransportChannel_InformationResponse */
static int hf_nbap_Common_TransportChannel_Status_Information_PDU = -1;  /* Common_TransportChannel_Status_Information */
static int hf_nbap_CommunicationControlPortID_PDU = -1;  /* CommunicationControlPortID */
static int hf_nbap_Compressed_Mode_Deactivation_Flag_PDU = -1;  /* Compressed_Mode_Deactivation_Flag */
static int hf_nbap_ConfigurationGenerationID_PDU = -1;  /* ConfigurationGenerationID */
static int hf_nbap_ConstantValue_PDU = -1;        /* ConstantValue */
static int hf_nbap_ContinuousPacketConnectivityDTX_DRX_Capability_PDU = -1;  /* ContinuousPacketConnectivityDTX_DRX_Capability */
static int hf_nbap_ContinuousPacketConnectivityDTX_DRX_Information_PDU = -1;  /* ContinuousPacketConnectivityDTX_DRX_Information */
static int hf_nbap_ContinuousPacketConnectivityHS_SCCH_less_Capability_PDU = -1;  /* ContinuousPacketConnectivityHS_SCCH_less_Capability */
static int hf_nbap_ContinuousPacketConnectivityHS_SCCH_less_Information_PDU = -1;  /* ContinuousPacketConnectivityHS_SCCH_less_Information */
static int hf_nbap_ContinuousPacketConnectivityHS_SCCH_less_Information_Response_PDU = -1;  /* ContinuousPacketConnectivityHS_SCCH_less_Information_Response */
static int hf_nbap_ControlGAP_PDU = -1;           /* ControlGAP */
static int hf_nbap_CPC_Information_PDU = -1;      /* CPC_Information */
static int hf_nbap_ContinuousPacketConnectivityHS_SCCH_less_Deactivate_Indicator_PDU = -1;  /* ContinuousPacketConnectivityHS_SCCH_less_Deactivate_Indicator */
static int hf_nbap_CriticalityDiagnostics_PDU = -1;  /* CriticalityDiagnostics */
static int hf_nbap_CRNC_CommunicationContextID_PDU = -1;  /* CRNC_CommunicationContextID */
static int hf_nbap_CSBMeasurementID_PDU = -1;     /* CSBMeasurementID */
static int hf_nbap_CSBTransmissionID_PDU = -1;    /* CSBTransmissionID */
static int hf_nbap_Common_EDCH_System_InformationLCR_PDU = -1;  /* Common_EDCH_System_InformationLCR */
static int hf_nbap_Common_EDCH_System_Information_ResponseLCR_PDU = -1;  /* Common_EDCH_System_Information_ResponseLCR */
static int hf_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_Ext_PDU = -1;  /* Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_Ext */
static int hf_nbap_Common_MACFlows_to_DeleteLCR_PDU = -1;  /* Common_MACFlows_to_DeleteLCR */
static int hf_nbap_CPC_InformationLCR_PDU = -1;   /* CPC_InformationLCR */
static int hf_nbap_ContinuousPacketConnectivity_DRX_CapabilityLCR_PDU = -1;  /* ContinuousPacketConnectivity_DRX_CapabilityLCR */
static int hf_nbap_ContinuousPacketConnectivity_DRX_InformationLCR_PDU = -1;  /* ContinuousPacketConnectivity_DRX_InformationLCR */
static int hf_nbap_E_AGCH_UE_Inactivity_Monitor_Threshold_PDU = -1;  /* E_AGCH_UE_Inactivity_Monitor_Threshold */
static int hf_nbap_ContinuousPacketConnectivity_DRX_Information_ResponseLCR_PDU = -1;  /* ContinuousPacketConnectivity_DRX_Information_ResponseLCR */
static int hf_nbap_DCH_FDD_Information_PDU = -1;  /* DCH_FDD_Information */
static int hf_nbap_DCH_Indicator_For_E_DCH_HSDPA_Operation_PDU = -1;  /* DCH_Indicator_For_E_DCH_HSDPA_Operation */
static int hf_nbap_DCH_InformationResponse_PDU = -1;  /* DCH_InformationResponse */
static int hf_nbap_DCH_MeasurementOccasion_Information_PDU = -1;  /* DCH_MeasurementOccasion_Information */
static int hf_nbap_DCH_TDD_Information_PDU = -1;  /* DCH_TDD_Information */
static int hf_nbap_FDD_DCHs_to_Modify_PDU = -1;   /* FDD_DCHs_to_Modify */
static int hf_nbap_TDD_DCHs_to_Modify_PDU = -1;   /* TDD_DCHs_to_Modify */
static int hf_nbap_DedicatedMeasurementType_PDU = -1;  /* DedicatedMeasurementType */
static int hf_nbap_DelayedActivation_PDU = -1;    /* DelayedActivation */
static int hf_nbap_DGANSS_Corrections_Req_PDU = -1;  /* DGANSS_Corrections_Req */
static int hf_nbap_DGNSS_ValidityPeriod_PDU = -1;  /* DGNSS_ValidityPeriod */
static int hf_nbap_DiversityMode_PDU = -1;        /* DiversityMode */
static int hf_nbap_DL_DPCH_TimingAdjustment_PDU = -1;  /* DL_DPCH_TimingAdjustment */
static int hf_nbap_DL_Power_PDU = -1;             /* DL_Power */
static int hf_nbap_DL_PowerBalancing_Information_PDU = -1;  /* DL_PowerBalancing_Information */
static int hf_nbap_DL_PowerBalancing_ActivationIndicator_PDU = -1;  /* DL_PowerBalancing_ActivationIndicator */
static int hf_nbap_DL_PowerBalancing_UpdatedIndicator_PDU = -1;  /* DL_PowerBalancing_UpdatedIndicator */
static int hf_nbap_DL_ScramblingCode_PDU = -1;    /* DL_ScramblingCode */
static int hf_nbap_DL_TimeslotISCPInfo_PDU = -1;  /* DL_TimeslotISCPInfo */
static int hf_nbap_DL_TimeslotISCPInfoLCR_PDU = -1;  /* DL_TimeslotISCPInfoLCR */
static int hf_nbap_DL_TPC_Pattern01Count_PDU = -1;  /* DL_TPC_Pattern01Count */
static int hf_nbap_DLTransmissionBranchLoadValue_PDU = -1;  /* DLTransmissionBranchLoadValue */
static int hf_nbap_MaxHSDSCH_HSSCCH_Power_per_CELLPORTION_PDU = -1;  /* MaxHSDSCH_HSSCCH_Power_per_CELLPORTION */
static int hf_nbap_DPC_Mode_PDU = -1;             /* DPC_Mode */
static int hf_nbap_DPCH_ID768_PDU = -1;           /* DPCH_ID768 */
static int hf_nbap_DSCH_InformationResponse_PDU = -1;  /* DSCH_InformationResponse */
static int hf_nbap_DSCH_TDD_Information_PDU = -1;  /* DSCH_TDD_Information */
static int hf_nbap_Dual_Band_Capability_Info_PDU = -1;  /* Dual_Band_Capability_Info */
static int hf_nbap_DwPCH_Power_PDU = -1;          /* DwPCH_Power */
static int hf_nbap_E_AGCH_Table_Choice_PDU = -1;  /* E_AGCH_Table_Choice */
static int hf_nbap_E_AGCH_FDD_Code_Information_PDU = -1;  /* E_AGCH_FDD_Code_Information */
static int hf_nbap_E_AI_Capability_PDU = -1;      /* E_AI_Capability */
static int hf_nbap_E_DCH_Capability_PDU = -1;     /* E_DCH_Capability */
static int hf_nbap_E_DCHCapacityConsumptionLaw_PDU = -1;  /* E_DCHCapacityConsumptionLaw */
static int hf_nbap_E_DCH_TDD_CapacityConsumptionLaw_PDU = -1;  /* E_DCH_TDD_CapacityConsumptionLaw */
static int hf_nbap_E_DCH_TTI2ms_Capability_PDU = -1;  /* E_DCH_TTI2ms_Capability */
static int hf_nbap_E_DCH_SF_Capability_PDU = -1;  /* E_DCH_SF_Capability */
static int hf_nbap_E_DCH_HARQ_Combining_Capability_PDU = -1;  /* E_DCH_HARQ_Combining_Capability */
static int hf_nbap_E_DCH_FDD_DL_Control_Channel_Information_PDU = -1;  /* E_DCH_FDD_DL_Control_Channel_Information */
static int hf_nbap_E_DCH_FDD_Information_PDU = -1;  /* E_DCH_FDD_Information */
static int hf_nbap_E_DCH_FDD_Information_Response_PDU = -1;  /* E_DCH_FDD_Information_Response */
static int hf_nbap_E_DCH_FDD_Information_to_Modify_PDU = -1;  /* E_DCH_FDD_Information_to_Modify */
static int hf_nbap_E_DCH_FDD_Update_Information_PDU = -1;  /* E_DCH_FDD_Update_Information */
static int hf_nbap_E_DCH_DL_Control_Channel_Change_Information_PDU = -1;  /* E_DCH_DL_Control_Channel_Change_Information */
static int hf_nbap_E_DCH_DL_Control_Channel_Grant_Information_PDU = -1;  /* E_DCH_DL_Control_Channel_Grant_Information */
static int hf_nbap_E_DCH_PowerOffset_for_SchedulingInfo_PDU = -1;  /* E_DCH_PowerOffset_for_SchedulingInfo */
static int hf_nbap_E_DCH_MACdPDU_SizeCapability_PDU = -1;  /* E_DCH_MACdPDU_SizeCapability */
static int hf_nbap_E_DCH_MACdPDUSizeFormat_PDU = -1;  /* E_DCH_MACdPDUSizeFormat */
static int hf_nbap_E_DCH_MACdFlows_Information_PDU = -1;  /* E_DCH_MACdFlows_Information */
static int hf_nbap_E_DCH_MACdFlows_to_Delete_PDU = -1;  /* E_DCH_MACdFlows_to_Delete */
static int hf_nbap_E_DCH_Non_serving_Relative_Grant_Down_Commands_PDU = -1;  /* E_DCH_Non_serving_Relative_Grant_Down_Commands */
static int hf_nbap_Maximum_Target_ReceivedTotalWideBandPower_PDU = -1;  /* Maximum_Target_ReceivedTotalWideBandPower */
static int hf_nbap_Target_NonServing_EDCH_To_Total_EDCH_Power_Ratio_PDU = -1;  /* Target_NonServing_EDCH_To_Total_EDCH_Power_Ratio */
static int hf_nbap_E_DCH_RL_Indication_PDU = -1;  /* E_DCH_RL_Indication */
static int hf_nbap_E_DCH_Serving_Cell_Change_Info_Response_PDU = -1;  /* E_DCH_Serving_Cell_Change_Info_Response */
static int hf_nbap_EDCH_RACH_Report_Value_PDU = -1;  /* EDCH_RACH_Report_Value */
static int hf_nbap_E_DPDCH_PowerInterpolation_PDU = -1;  /* E_DPDCH_PowerInterpolation */
static int hf_nbap_E_DCH_MACdFlows_to_DeleteLCR_PDU = -1;  /* E_DCH_MACdFlows_to_DeleteLCR */
static int hf_nbap_Enhanced_UE_DRX_InformationLCR_PDU = -1;  /* Enhanced_UE_DRX_InformationLCR */
static int hf_nbap_End_Of_Audit_Sequence_Indicator_PDU = -1;  /* End_Of_Audit_Sequence_Indicator */
static int hf_nbap_E_RGCH_E_HICH_FDD_Code_Information_PDU = -1;  /* E_RGCH_E_HICH_FDD_Code_Information */
static int hf_nbap_E_RNTI_PDU = -1;               /* E_RNTI */
static int hf_nbap_E_TFCI_Boost_Information_PDU = -1;  /* E_TFCI_Boost_Information */
static int hf_nbap_E_DCHProvidedBitRate_PDU = -1;  /* E_DCHProvidedBitRate */
static int hf_nbap_E_DCHProvidedBitRateValueInformation_For_CellPortion_PDU = -1;  /* E_DCHProvidedBitRateValueInformation_For_CellPortion */
static int hf_nbap_E_HICH_TimeOffset_PDU = -1;    /* E_HICH_TimeOffset */
static int hf_nbap_E_HICH_TimeOffsetLCR_PDU = -1;  /* E_HICH_TimeOffsetLCR */
static int hf_nbap_E_DCH_Information_PDU = -1;    /* E_DCH_Information */
static int hf_nbap_E_DCH_MACdFlow_Retransmission_Timer_PDU = -1;  /* E_DCH_MACdFlow_Retransmission_Timer */
static int hf_nbap_E_DCH_Information_Response_PDU = -1;  /* E_DCH_Information_Response */
static int hf_nbap_E_DCH_Information_Reconfig_PDU = -1;  /* E_DCH_Information_Reconfig */
static int hf_nbap_Maximum_Generated_ReceivedTotalWideBandPowerInOtherCells_PDU = -1;  /* Maximum_Generated_ReceivedTotalWideBandPowerInOtherCells */
static int hf_nbap_E_DCH_768_Information_PDU = -1;  /* E_DCH_768_Information */
static int hf_nbap_E_DCH_768_Information_Reconfig_PDU = -1;  /* E_DCH_768_Information_Reconfig */
static int hf_nbap_E_DCH_LCR_Information_PDU = -1;  /* E_DCH_LCR_Information */
static int hf_nbap_E_DCH_LCR_Information_Reconfig_PDU = -1;  /* E_DCH_LCR_Information_Reconfig */
static int hf_nbap_DormantModeIndicator_PDU = -1;  /* DormantModeIndicator */
static int hf_nbap_Enhanced_FACH_Capability_PDU = -1;  /* Enhanced_FACH_Capability */
static int hf_nbap_EnhancedHSServingCC_Abort_PDU = -1;  /* EnhancedHSServingCC_Abort */
static int hf_nbap_Enhanced_PCH_Capability_PDU = -1;  /* Enhanced_PCH_Capability */
static int hf_nbap_Enhanced_UE_DRX_Capability_PDU = -1;  /* Enhanced_UE_DRX_Capability */
static int hf_nbap_Enhanced_UE_DRX_InformationFDD_PDU = -1;  /* Enhanced_UE_DRX_InformationFDD */
static int hf_nbap_Extended_E_DCH_LCRTDD_PhysicalLayerCategory_PDU = -1;  /* Extended_E_DCH_LCRTDD_PhysicalLayerCategory */
static int hf_nbap_Ext_Max_Bits_MACe_PDU_non_scheduled_PDU = -1;  /* Ext_Max_Bits_MACe_PDU_non_scheduled */
static int hf_nbap_Ext_Reference_E_TFCI_PO_PDU = -1;  /* Ext_Reference_E_TFCI_PO */
static int hf_nbap_ExtendedPropagationDelay_PDU = -1;  /* ExtendedPropagationDelay */
static int hf_nbap_Extended_RNC_ID_PDU = -1;      /* Extended_RNC_ID */
static int hf_nbap_Extended_Round_Trip_Time_Value_PDU = -1;  /* Extended_Round_Trip_Time_Value */
static int hf_nbap_Extended_HS_SCCH_ID_PDU = -1;  /* Extended_HS_SCCH_ID */
static int hf_nbap_Extended_HS_SICH_ID_PDU = -1;  /* Extended_HS_SICH_ID */
static int hf_nbap_Extended_E_HICH_ID_TDD_PDU = -1;  /* Extended_E_HICH_ID_TDD */
static int hf_nbap_E_DCH_Semi_PersistentScheduling_Information_LCR_PDU = -1;  /* E_DCH_Semi_PersistentScheduling_Information_LCR */
static int hf_nbap_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR_PDU = -1;  /* E_DCH_Semi_PersistentScheduling_Information_ResponseLCR */
static int hf_nbap_Fast_Reconfiguration_Mode_PDU = -1;  /* Fast_Reconfiguration_Mode */
static int hf_nbap_Fast_Reconfiguration_Permission_PDU = -1;  /* Fast_Reconfiguration_Permission */
static int hf_nbap_FDD_S_CCPCH_FrameOffset_PDU = -1;  /* FDD_S_CCPCH_FrameOffset */
static int hf_nbap_F_DPCH_Capability_PDU = -1;    /* F_DPCH_Capability */
static int hf_nbap_F_DPCH_SlotFormat_PDU = -1;    /* F_DPCH_SlotFormat */
static int hf_nbap_F_DPCH_SlotFormatCapability_PDU = -1;  /* F_DPCH_SlotFormatCapability */
static int hf_nbap_FNReportingIndicator_PDU = -1;  /* FNReportingIndicator */
static int hf_nbap_GANSS_AddIonoModelReq_PDU = -1;  /* GANSS_AddIonoModelReq */
static int hf_nbap_GANSS_AddNavigationModelsReq_PDU = -1;  /* GANSS_AddNavigationModelsReq */
static int hf_nbap_GANSS_AddUTCModelsReq_PDU = -1;  /* GANSS_AddUTCModelsReq */
static int hf_nbap_GANSS_Additional_Ionospheric_Model_PDU = -1;  /* GANSS_Additional_Ionospheric_Model */
static int hf_nbap_GANSS_Additional_Navigation_Models_PDU = -1;  /* GANSS_Additional_Navigation_Models */
static int hf_nbap_GANSS_Additional_Time_Models_PDU = -1;  /* GANSS_Additional_Time_Models */
static int hf_nbap_GANSS_Additional_UTC_Models_PDU = -1;  /* GANSS_Additional_UTC_Models */
static int hf_nbap_GANSS_ALM_ECEFsbasAlmanacSet_PDU = -1;  /* GANSS_ALM_ECEFsbasAlmanacSet */
static int hf_nbap_GANSS_ALM_GlonassAlmanacSet_PDU = -1;  /* GANSS_ALM_GlonassAlmanacSet */
static int hf_nbap_GANSS_ALM_MidiAlmanacSet_PDU = -1;  /* GANSS_ALM_MidiAlmanacSet */
static int hf_nbap_GANSS_ALM_NAVKeplerianSet_PDU = -1;  /* GANSS_ALM_NAVKeplerianSet */
static int hf_nbap_GANSS_ALM_ReducedKeplerianSet_PDU = -1;  /* GANSS_ALM_ReducedKeplerianSet */
static int hf_nbap_GANSS_Auxiliary_Information_PDU = -1;  /* GANSS_Auxiliary_Information */
static int hf_nbap_GANSS_AuxInfoReq_PDU = -1;     /* GANSS_AuxInfoReq */
static int hf_nbap_GANSS_Common_Data_PDU = -1;    /* GANSS_Common_Data */
static int hf_nbap_GANSS_Earth_Orientation_Parameters_PDU = -1;  /* GANSS_Earth_Orientation_Parameters */
static int hf_nbap_GANSS_EarthOrientParaReq_PDU = -1;  /* GANSS_EarthOrientParaReq */
static int hf_nbap_GANSS_Generic_Data_PDU = -1;   /* GANSS_Generic_Data */
static int hf_nbap_GANSS_ID_PDU = -1;             /* GANSS_ID */
static int hf_nbap_GANSS_Information_PDU = -1;    /* GANSS_Information */
static int hf_nbap_GANSS_SBAS_ID_PDU = -1;        /* GANSS_SBAS_ID */
static int hf_nbap_GANSS_Time_ID_PDU = -1;        /* GANSS_Time_ID */
static int hf_nbap_HARQ_MemoryPartitioningInfoExtForMIMO_PDU = -1;  /* HARQ_MemoryPartitioningInfoExtForMIMO */
static int hf_nbap_HARQ_Preamble_Mode_PDU = -1;   /* HARQ_Preamble_Mode */
static int hf_nbap_HARQ_Preamble_Mode_Activation_Indicator_PDU = -1;  /* HARQ_Preamble_Mode_Activation_Indicator */
static int hf_nbap_HSDPA_Capability_PDU = -1;     /* HSDPA_Capability */
static int hf_nbap_HS_DSCHProvidedBitRate_PDU = -1;  /* HS_DSCHProvidedBitRate */
static int hf_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortion_PDU = -1;  /* HS_DSCHProvidedBitRateValueInformation_For_CellPortion */
static int hf_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR_PDU = -1;  /* HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR */
static int hf_nbap_HS_DSCHRequiredPower_PDU = -1;  /* HS_DSCHRequiredPower */
static int hf_nbap_HS_DSCHRequiredPowerValue_PDU = -1;  /* HS_DSCHRequiredPowerValue */
static int hf_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortion_PDU = -1;  /* HS_DSCHRequiredPowerValueInformation_For_CellPortion */
static int hf_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR_PDU = -1;  /* HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR */
static int hf_nbap_HSDSCH_Common_System_InformationFDD_PDU = -1;  /* HSDSCH_Common_System_InformationFDD */
static int hf_nbap_HSDSCH_Common_System_Information_ResponseFDD_PDU = -1;  /* HSDSCH_Common_System_Information_ResponseFDD */
static int hf_nbap_HSDSCH_FDD_Information_PDU = -1;  /* HSDSCH_FDD_Information */
static int hf_nbap_HSDSCH_TDD_Information_PDU = -1;  /* HSDSCH_TDD_Information */
static int hf_nbap_HSDSCH_Information_to_Modify_PDU = -1;  /* HSDSCH_Information_to_Modify */
static int hf_nbap_HSDSCH_MACdPDUSizeFormat_PDU = -1;  /* HSDSCH_MACdPDUSizeFormat */
static int hf_nbap_HSDSCH_MACdPDU_SizeCapability_PDU = -1;  /* HSDSCH_MACdPDU_SizeCapability */
static int hf_nbap_HSDSCH_Information_to_Modify_Unsynchronised_PDU = -1;  /* HSDSCH_Information_to_Modify_Unsynchronised */
static int hf_nbap_HSDSCH_FDD_Information_Response_PDU = -1;  /* HSDSCH_FDD_Information_Response */
static int hf_nbap_HSDSCH_Paging_System_InformationFDD_PDU = -1;  /* HSDSCH_Paging_System_InformationFDD */
static int hf_nbap_HSDSCH_Paging_System_Information_ResponseFDD_PDU = -1;  /* HSDSCH_Paging_System_Information_ResponseFDD */
static int hf_nbap_HSDSCH_TDD_Information_Response_PDU = -1;  /* HSDSCH_TDD_Information_Response */
static int hf_nbap_HSDSCH_MACdFlows_Information_PDU = -1;  /* HSDSCH_MACdFlows_Information */
static int hf_nbap_HSDSCH_MACdFlows_to_Delete_PDU = -1;  /* HSDSCH_MACdFlows_to_Delete */
static int hf_nbap_HSDSCH_TBSizeTableIndicator_PDU = -1;  /* HSDSCH_TBSizeTableIndicator */
static int hf_nbap_HSDSCH_PreconfigurationInfo_PDU = -1;  /* HSDSCH_PreconfigurationInfo */
static int hf_nbap_Additional_EDCH_Preconfiguration_Information_PDU = -1;  /* Additional_EDCH_Preconfiguration_Information */
static int hf_nbap_HSDSCH_PreconfigurationSetup_PDU = -1;  /* HSDSCH_PreconfigurationSetup */
static int hf_nbap_HSSCCH_Specific_InformationRespListTDD768_PDU = -1;  /* HSSCCH_Specific_InformationRespListTDD768 */
static int hf_nbap_HS_SICH_Reception_Quality_Value_PDU = -1;  /* HS_SICH_Reception_Quality_Value */
static int hf_nbap_HS_SICH_failed_PDU = -1;       /* HS_SICH_failed */
static int hf_nbap_HS_SICH_missed_PDU = -1;       /* HS_SICH_missed */
static int hf_nbap_HS_SICH_total_PDU = -1;        /* HS_SICH_total */
static int hf_nbap_HS_SICH_Reception_Quality_Measurement_Value_PDU = -1;  /* HS_SICH_Reception_Quality_Measurement_Value */
static int hf_nbap_HSDSCH_RNTI_PDU = -1;          /* HSDSCH_RNTI */
static int hf_nbap_HS_PDSCH_FDD_Code_Information_PDU = -1;  /* HS_PDSCH_FDD_Code_Information */
static int hf_nbap_HS_SICH_ID_PDU = -1;           /* HS_SICH_ID */
static int hf_nbap_HS_SCCH_FDD_Code_Information_PDU = -1;  /* HS_SCCH_FDD_Code_Information */
static int hf_nbap_HS_PDSCH_Code_Change_Indicator_PDU = -1;  /* HS_PDSCH_Code_Change_Indicator */
static int hf_nbap_HS_PDSCH_Code_Change_Grant_PDU = -1;  /* HS_PDSCH_Code_Change_Grant */
static int hf_nbap_HSDSCH_Configured_Indicator_PDU = -1;  /* HSDSCH_Configured_Indicator */
static int hf_nbap_HS_DSCH_Serving_Cell_Change_Info_PDU = -1;  /* HS_DSCH_Serving_Cell_Change_Info */
static int hf_nbap_HS_DSCH_Serving_Cell_Change_Info_Response_PDU = -1;  /* HS_DSCH_Serving_Cell_Change_Info_Response */
static int hf_nbap_HSDSCH_FDD_Update_Information_PDU = -1;  /* HSDSCH_FDD_Update_Information */
static int hf_nbap_HSDSCH_TDD_Update_Information_PDU = -1;  /* HSDSCH_TDD_Update_Information */
static int hf_nbap_HSDSCH_Common_System_InformationLCR_PDU = -1;  /* HSDSCH_Common_System_InformationLCR */
static int hf_nbap_HSDSCH_Common_System_Information_ResponseLCR_PDU = -1;  /* HSDSCH_Common_System_Information_ResponseLCR */
static int hf_nbap_HSDSCH_Paging_System_InformationLCR_PDU = -1;  /* HSDSCH_Paging_System_InformationLCR */
static int hf_nbap_HSDSCH_Paging_System_Information_ResponseLCR_PDU = -1;  /* HSDSCH_Paging_System_Information_ResponseLCR */
static int hf_nbap_HS_DSCH_Semi_PersistentScheduling_Information_LCR_PDU = -1;  /* HS_DSCH_Semi_PersistentScheduling_Information_LCR */
static int hf_nbap_RepetitionPeriodIndex_PDU = -1;  /* RepetitionPeriodIndex */
static int hf_nbap_SPS_Reservation_Indicator_PDU = -1;  /* SPS_Reservation_Indicator */
static int hf_nbap_HS_DSCH_SPS_Operation_Indicator_PDU = -1;  /* HS_DSCH_SPS_Operation_Indicator */
static int hf_nbap_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR_PDU = -1;  /* HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR */
static int hf_nbap_Add_To_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU = -1;  /* Add_To_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst */
static int hf_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext_PDU = -1;  /* Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext */
static int hf_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU = -1;  /* Modify_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst */
static int hf_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext_PDU = -1;  /* Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext */
static int hf_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU = -1;  /* Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst */
static int hf_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext_PDU = -1;  /* Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext */
static int hf_nbap_MIMO_ReferenceSignal_InformationListLCR_PDU = -1;  /* MIMO_ReferenceSignal_InformationListLCR */
static int hf_nbap_HSSICH_ReferenceSignal_InformationLCR_PDU = -1;  /* HSSICH_ReferenceSignal_InformationLCR */
static int hf_nbap_HSSICH_ReferenceSignal_InformationModifyLCR_PDU = -1;  /* HSSICH_ReferenceSignal_InformationModifyLCR */
static int hf_nbap_IMB_Parameters_PDU = -1;       /* IMB_Parameters */
static int hf_nbap_InformationReportCharacteristics_PDU = -1;  /* InformationReportCharacteristics */
static int hf_nbap_InformationExchangeID_PDU = -1;  /* InformationExchangeID */
static int hf_nbap_InformationType_PDU = -1;      /* InformationType */
static int hf_nbap_Initial_DL_DPCH_TimingAdjustment_Allowed_PDU = -1;  /* Initial_DL_DPCH_TimingAdjustment_Allowed */
static int hf_nbap_InnerLoopDLPCStatus_PDU = -1;  /* InnerLoopDLPCStatus */
static int hf_nbap_IPMulticastIndication_PDU = -1;  /* IPMulticastIndication */
static int hf_nbap_IPMulticastDataBearerIndication_PDU = -1;  /* IPMulticastDataBearerIndication */
static int hf_nbap_IdleIntervalInformation_PDU = -1;  /* IdleIntervalInformation */
static int hf_nbap_Local_Cell_ID_PDU = -1;        /* Local_Cell_ID */
static int hf_nbap_LCRTDD_Uplink_Physical_Channel_Capability_PDU = -1;  /* LCRTDD_Uplink_Physical_Channel_Capability */
static int hf_nbap_MAC_PDU_SizeExtended_PDU = -1;  /* MAC_PDU_SizeExtended */
static int hf_nbap_MACes_Maximum_Bitrate_LCR_PDU = -1;  /* MACes_Maximum_Bitrate_LCR */
static int hf_nbap_Maximum_Number_of_Retransmissions_For_E_DCH_PDU = -1;  /* Maximum_Number_of_Retransmissions_For_E_DCH */
static int hf_nbap_Maximum_Target_ReceivedTotalWideBandPower_LCR_PDU = -1;  /* Maximum_Target_ReceivedTotalWideBandPower_LCR */
static int hf_nbap_MaximumTransmissionPower_PDU = -1;  /* MaximumTransmissionPower */
static int hf_nbap_Max_UE_DTX_Cycle_PDU = -1;     /* Max_UE_DTX_Cycle */
static int hf_nbap_MBMS_Capability_PDU = -1;      /* MBMS_Capability */
static int hf_nbap_MeasurementFilterCoefficient_PDU = -1;  /* MeasurementFilterCoefficient */
static int hf_nbap_MeasurementID_PDU = -1;        /* MeasurementID */
static int hf_nbap_MeasurementRecoveryBehavior_PDU = -1;  /* MeasurementRecoveryBehavior */
static int hf_nbap_MeasurementRecoveryReportingIndicator_PDU = -1;  /* MeasurementRecoveryReportingIndicator */
static int hf_nbap_MeasurementRecoverySupportIndicator_PDU = -1;  /* MeasurementRecoverySupportIndicator */
static int hf_nbap_MessageStructure_PDU = -1;     /* MessageStructure */
static int hf_nbap_MICH_CFN_PDU = -1;             /* MICH_CFN */
static int hf_nbap_MidambleShiftLCR_PDU = -1;     /* MidambleShiftLCR */
static int hf_nbap_MIMO_ActivationIndicator_PDU = -1;  /* MIMO_ActivationIndicator */
static int hf_nbap_MIMO_Capability_PDU = -1;      /* MIMO_Capability */
static int hf_nbap_MIMO_Mode_Indicator_PDU = -1;  /* MIMO_Mode_Indicator */
static int hf_nbap_MIMO_N_M_Ratio_PDU = -1;       /* MIMO_N_M_Ratio */
static int hf_nbap_MIMO_PilotConfiguration_PDU = -1;  /* MIMO_PilotConfiguration */
static int hf_nbap_MIMO_PilotConfigurationExtension_PDU = -1;  /* MIMO_PilotConfigurationExtension */
static int hf_nbap_MIMO_PowerOffsetForS_CPICHCapability_PDU = -1;  /* MIMO_PowerOffsetForS_CPICHCapability */
static int hf_nbap_MinimumReducedE_DPDCH_GainFactor_PDU = -1;  /* MinimumReducedE_DPDCH_GainFactor */
static int hf_nbap_Modification_Period_PDU = -1;  /* Modification_Period */
static int hf_nbap_MAChs_ResetIndicator_PDU = -1;  /* MAChs_ResetIndicator */
static int hf_nbap_ModulationMBSFN_PDU = -1;      /* ModulationMBSFN */
static int hf_nbap_ModulationPO_MBSFN_PDU = -1;   /* ModulationPO_MBSFN */
static int hf_nbap_MBSFN_Only_Mode_Indicator_PDU = -1;  /* MBSFN_Only_Mode_Indicator */
static int hf_nbap_MBSFN_Only_Mode_Capability_PDU = -1;  /* MBSFN_Only_Mode_Capability */
static int hf_nbap_Multicarrier_Number_PDU = -1;  /* Multicarrier_Number */
static int hf_nbap_MultipleFreq_HARQ_MemoryPartitioning_InformationList_PDU = -1;  /* MultipleFreq_HARQ_MemoryPartitioning_InformationList */
static int hf_nbap_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR_PDU = -1;  /* MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR */
static int hf_nbap_Multi_Cell_Capability_Info_PDU = -1;  /* Multi_Cell_Capability_Info */
static int hf_nbap_Multicell_EDCH_InformationItemIEs_PDU = -1;  /* Multicell_EDCH_InformationItemIEs */
static int hf_nbap_Multicell_EDCH_RL_Specific_InformationItemIEs_PDU = -1;  /* Multicell_EDCH_RL_Specific_InformationItemIEs */
static int hf_nbap_MIMO_SFMode_For_HSPDSCHDualStream_PDU = -1;  /* MIMO_SFMode_For_HSPDSCHDualStream */
static int hf_nbap_NCyclesPerSFNperiod_PDU = -1;  /* NCyclesPerSFNperiod */
static int hf_nbap_NRepetitionsPerCyclePeriod_PDU = -1;  /* NRepetitionsPerCyclePeriod */
static int hf_nbap_NeighbouringCellMeasurementInformation_PDU = -1;  /* NeighbouringCellMeasurementInformation */
static int hf_nbap_NeighbouringTDDCellMeasurementInformationLCR_PDU = -1;  /* NeighbouringTDDCellMeasurementInformationLCR */
static int hf_nbap_NeighbouringTDDCellMeasurementInformation768_PDU = -1;  /* NeighbouringTDDCellMeasurementInformation768 */
static int hf_nbap_NonCellSpecificTxDiversity_PDU = -1;  /* NonCellSpecificTxDiversity */
static int hf_nbap_NI_Information_PDU = -1;       /* NI_Information */
static int hf_nbap_NodeB_CommunicationContextID_PDU = -1;  /* NodeB_CommunicationContextID */
static int hf_nbap_NumberOfReportedCellPortions_PDU = -1;  /* NumberOfReportedCellPortions */
static int hf_nbap_NumberOfReportedCellPortionsLCR_PDU = -1;  /* NumberOfReportedCellPortionsLCR */
static int hf_nbap_NSubCyclesPerCyclePeriod_PDU = -1;  /* NSubCyclesPerCyclePeriod */
static int hf_nbap_Number_Of_Supported_Carriers_PDU = -1;  /* Number_Of_Supported_Carriers */
static int hf_nbap_NoOfTargetCellHS_SCCH_Order_PDU = -1;  /* NoOfTargetCellHS_SCCH_Order */
static int hf_nbap_Out_of_Sychronization_Window_PDU = -1;  /* Out_of_Sychronization_Window */
static int hf_nbap_Paging_MACFlows_to_DeleteFDD_PDU = -1;  /* Paging_MACFlows_to_DeleteFDD */
static int hf_nbap_PhysicalChannelID_for_CommonERNTI_RequestedIndicator_PDU = -1;  /* PhysicalChannelID_for_CommonERNTI_RequestedIndicator */
static int hf_nbap_PLCCHinformation_PDU = -1;     /* PLCCHinformation */
static int hf_nbap_PowerAdjustmentType_PDU = -1;  /* PowerAdjustmentType */
static int hf_nbap_PrecodingWeightSetRestriction_PDU = -1;  /* PrecodingWeightSetRestriction */
static int hf_nbap_Primary_CPICH_Usage_for_Channel_Estimation_PDU = -1;  /* Primary_CPICH_Usage_for_Channel_Estimation */
static int hf_nbap_PrimaryScramblingCode_PDU = -1;  /* PrimaryScramblingCode */
static int hf_nbap_PrimaryCCPCH_RSCP_PDU = -1;    /* PrimaryCCPCH_RSCP */
static int hf_nbap_PrimaryCCPCH_RSCP_Delta_PDU = -1;  /* PrimaryCCPCH_RSCP_Delta */
static int hf_nbap_UE_Selected_MBMS_Service_Information_PDU = -1;  /* UE_Selected_MBMS_Service_Information */
static int hf_nbap_Paging_MACFlows_to_DeleteLCR_PDU = -1;  /* Paging_MACFlows_to_DeleteLCR */
static int hf_nbap_RL_Specific_DCH_Info_PDU = -1;  /* RL_Specific_DCH_Info */
static int hf_nbap_RL_Specific_E_DCH_Info_PDU = -1;  /* RL_Specific_E_DCH_Info */
static int hf_nbap_Reference_ReceivedTotalWideBandPower_PDU = -1;  /* Reference_ReceivedTotalWideBandPower */
static int hf_nbap_Reference_ReceivedTotalWideBandPowerReporting_PDU = -1;  /* Reference_ReceivedTotalWideBandPowerReporting */
static int hf_nbap_Reference_ReceivedTotalWideBandPowerSupportIndicator_PDU = -1;  /* Reference_ReceivedTotalWideBandPowerSupportIndicator */
static int hf_nbap_ReferenceClockAvailability_PDU = -1;  /* ReferenceClockAvailability */
static int hf_nbap_ReferenceSFNoffset_PDU = -1;   /* ReferenceSFNoffset */
static int hf_nbap_ReportCharacteristics_PDU = -1;  /* ReportCharacteristics */
static int hf_nbap_ReportCharacteristicsType_OnModification_PDU = -1;  /* ReportCharacteristicsType_OnModification */
static int hf_nbap_EDCH_RACH_Report_IncrDecrThres_PDU = -1;  /* EDCH_RACH_Report_IncrDecrThres */
static int hf_nbap_EDCH_RACH_Report_ThresholdInformation_PDU = -1;  /* EDCH_RACH_Report_ThresholdInformation */
static int hf_nbap_RL_ID_PDU = -1;                /* RL_ID */
static int hf_nbap_RL_Set_ID_PDU = -1;            /* RL_Set_ID */
static int hf_nbap_DL_RLC_PDU_Size_Format_PDU = -1;  /* DL_RLC_PDU_Size_Format */
static int hf_nbap_Received_total_wide_band_power_For_CellPortion_Value_PDU = -1;  /* Received_total_wide_band_power_For_CellPortion_Value */
static int hf_nbap_Received_total_wide_band_power_For_CellPortion_ValueLCR_PDU = -1;  /* Received_total_wide_band_power_For_CellPortion_ValueLCR */
static int hf_nbap_Received_total_wide_band_power_Value_IncrDecrThres_PDU = -1;  /* Received_total_wide_band_power_Value_IncrDecrThres */
static int hf_nbap_Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value_PDU = -1;  /* Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value */
static int hf_nbap_Received_Scheduled_EDCH_Power_Share_Value_PDU = -1;  /* Received_Scheduled_EDCH_Power_Share_Value */
static int hf_nbap_RSEPS_Value_IncrDecrThres_PDU = -1;  /* RSEPS_Value_IncrDecrThres */
static int hf_nbap_Rx_Timing_Deviation_Value_LCR_PDU = -1;  /* Rx_Timing_Deviation_Value_LCR */
static int hf_nbap_Rx_Timing_Deviation_Value_768_PDU = -1;  /* Rx_Timing_Deviation_Value_768 */
static int hf_nbap_Rx_Timing_Deviation_Value_384_ext_PDU = -1;  /* Rx_Timing_Deviation_Value_384_ext */
static int hf_nbap_RTWP_ReportingIndicator_PDU = -1;  /* RTWP_ReportingIndicator */
static int hf_nbap_RTWP_CellPortion_ReportingIndicator_PDU = -1;  /* RTWP_CellPortion_ReportingIndicator */
static int hf_nbap_AdjustmentPeriod_PDU = -1;     /* AdjustmentPeriod */
static int hf_nbap_E_DPCCH_Power_Boosting_Capability_PDU = -1;  /* E_DPCCH_Power_Boosting_Capability */
static int hf_nbap_SAT_Info_Almanac_ExtList_PDU = -1;  /* SAT_Info_Almanac_ExtList */
static int hf_nbap_ScaledAdjustmentRatio_PDU = -1;  /* ScaledAdjustmentRatio */
static int hf_nbap_MaxAdjustmentStep_PDU = -1;    /* MaxAdjustmentStep */
static int hf_nbap_SchedulingPriorityIndicator_PDU = -1;  /* SchedulingPriorityIndicator */
static int hf_nbap_Secondary_CPICH_Information_Change_PDU = -1;  /* Secondary_CPICH_Information_Change */
static int hf_nbap_Secondary_CCPCH_SlotFormat_Extended_PDU = -1;  /* Secondary_CCPCH_SlotFormat_Extended */
static int hf_nbap_Serving_E_DCH_RL_ID_PDU = -1;  /* Serving_E_DCH_RL_ID */
static int hf_nbap_SFN_PDU = -1;                  /* SFN */
static int hf_nbap_SFNSFNMeasurementThresholdInformation_PDU = -1;  /* SFNSFNMeasurementThresholdInformation */
static int hf_nbap_SFNSFNMeasurementValueInformation_PDU = -1;  /* SFNSFNMeasurementValueInformation */
static int hf_nbap_ShutdownTimer_PDU = -1;        /* ShutdownTimer */
static int hf_nbap_Single_Stream_MIMO_ActivationIndicator_PDU = -1;  /* Single_Stream_MIMO_ActivationIndicator */
static int hf_nbap_Single_Stream_MIMO_Capability_PDU = -1;  /* Single_Stream_MIMO_Capability */
static int hf_nbap_Single_Stream_MIMO_Mode_Indicator_PDU = -1;  /* Single_Stream_MIMO_Mode_Indicator */
static int hf_nbap_SignallingBearerRequestIndicator_PDU = -1;  /* SignallingBearerRequestIndicator */
static int hf_nbap_SixtyfourQAM_UsageAllowedIndicator_PDU = -1;  /* SixtyfourQAM_UsageAllowedIndicator */
static int hf_nbap_SixtyfourQAM_DL_UsageIndicator_PDU = -1;  /* SixtyfourQAM_DL_UsageIndicator */
static int hf_nbap_SixtyfourQAM_DL_Capability_PDU = -1;  /* SixtyfourQAM_DL_Capability */
static int hf_nbap_SixtyfourQAM_DL_MIMO_Combined_Capability_PDU = -1;  /* SixtyfourQAM_DL_MIMO_Combined_Capability */
static int hf_nbap_SixteenQAM_UL_Capability_PDU = -1;  /* SixteenQAM_UL_Capability */
static int hf_nbap_SixteenQAM_UL_Operation_Indicator_PDU = -1;  /* SixteenQAM_UL_Operation_Indicator */
static int hf_nbap_Start_Of_Audit_Sequence_Indicator_PDU = -1;  /* Start_Of_Audit_Sequence_Indicator */
static int hf_nbap_SyncCase_PDU = -1;             /* SyncCase */
static int hf_nbap_SynchronisationReportCharacteristics_PDU = -1;  /* SynchronisationReportCharacteristics */
static int hf_nbap_SyncDLCodeIdThreInfoLCR_PDU = -1;  /* SyncDLCodeIdThreInfoLCR */
static int hf_nbap_SynchronisationReportType_PDU = -1;  /* SynchronisationReportType */
static int hf_nbap_Semi_PersistentScheduling_CapabilityLCR_PDU = -1;  /* Semi_PersistentScheduling_CapabilityLCR */
static int hf_nbap_T_Cell_PDU = -1;               /* T_Cell */
static int hf_nbap_TDD_DL_DPCH_TimeSlotFormat_LCR_PDU = -1;  /* TDD_DL_DPCH_TimeSlotFormat_LCR */
static int hf_nbap_TDD_TPC_DownlinkStepSize_PDU = -1;  /* TDD_TPC_DownlinkStepSize */
static int hf_nbap_TDD_TPC_UplinkStepSize_LCR_PDU = -1;  /* TDD_TPC_UplinkStepSize_LCR */
static int hf_nbap_TDD_UL_DPCH_TimeSlotFormat_LCR_PDU = -1;  /* TDD_UL_DPCH_TimeSlotFormat_LCR */
static int hf_nbap_TFCI_Presence_PDU = -1;        /* TFCI_Presence */
static int hf_nbap_TimeSlot_PDU = -1;             /* TimeSlot */
static int hf_nbap_TimeSlotLCR_PDU = -1;          /* TimeSlotLCR */
static int hf_nbap_TimeslotLCR_Extension_PDU = -1;  /* TimeslotLCR_Extension */
static int hf_nbap_TimeSlotMeasurementValueListLCR_PDU = -1;  /* TimeSlotMeasurementValueListLCR */
static int hf_nbap_TimingAdjustmentValue_PDU = -1;  /* TimingAdjustmentValue */
static int hf_nbap_TimingAdjustmentValueLCR_PDU = -1;  /* TimingAdjustmentValueLCR */
static int hf_nbap_TimingAdvanceApplied_PDU = -1;  /* TimingAdvanceApplied */
static int hf_nbap_SynchronisationIndicator_PDU = -1;  /* SynchronisationIndicator */
static int hf_nbap_TnlQos_PDU = -1;               /* TnlQos */
static int hf_nbap_Transmission_Gap_Pattern_Sequence_Information_PDU = -1;  /* Transmission_Gap_Pattern_Sequence_Information */
static int hf_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue_PDU = -1;  /* TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue */
static int hf_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue_PDU = -1;  /* TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue */
static int hf_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHSTransmissionValue_PDU = -1;  /* TransmittedCarrierPowerOfAllCodesNotUsedForHSTransmissionValue */
static int hf_nbap_Transmitted_Carrier_Power_For_CellPortion_Value_PDU = -1;  /* Transmitted_Carrier_Power_For_CellPortion_Value */
static int hf_nbap_Transmitted_Carrier_Power_For_CellPortion_ValueLCR_PDU = -1;  /* Transmitted_Carrier_Power_For_CellPortion_ValueLCR */
static int hf_nbap_Transmitted_Carrier_Power_Value_PDU = -1;  /* Transmitted_Carrier_Power_Value */
static int hf_nbap_TransmissionDiversityApplied_PDU = -1;  /* TransmissionDiversityApplied */
static int hf_nbap_TransmitDiversityIndicator_PDU = -1;  /* TransmitDiversityIndicator */
static int hf_nbap_TransportBearerRequestIndicator_PDU = -1;  /* TransportBearerRequestIndicator */
static int hf_nbap_TransportBearerNotRequestedIndicator_PDU = -1;  /* TransportBearerNotRequestedIndicator */
static int hf_nbap_TransportBearerNotSetupIndicator_PDU = -1;  /* TransportBearerNotSetupIndicator */
static int hf_nbap_TransportLayerAddress_PDU = -1;  /* TransportLayerAddress */
static int hf_nbap_TS0_CapabilityLCR_PDU = -1;    /* TS0_CapabilityLCR */
static int hf_nbap_TSTD_Indicator_PDU = -1;       /* TSTD_Indicator */
static int hf_nbap_TSN_Length_PDU = -1;           /* TSN_Length */
static int hf_nbap_TUTRANGANSSMeasurementThresholdInformation_PDU = -1;  /* TUTRANGANSSMeasurementThresholdInformation */
static int hf_nbap_TUTRANGANSSMeasurementValueInformation_PDU = -1;  /* TUTRANGANSSMeasurementValueInformation */
static int hf_nbap_TUTRANGPSMeasurementThresholdInformation_PDU = -1;  /* TUTRANGPSMeasurementThresholdInformation */
static int hf_nbap_TUTRANGPSMeasurementValueInformation_PDU = -1;  /* TUTRANGPSMeasurementValueInformation */
static int hf_nbap_TxDiversityOnDLControlChannelsByMIMOUECapability_PDU = -1;  /* TxDiversityOnDLControlChannelsByMIMOUECapability */
static int hf_nbap_TypeOfError_PDU = -1;          /* TypeOfError */
static int hf_nbap_UARFCN_PDU = -1;               /* UARFCN */
static int hf_nbap_UE_AggregateMaximumBitRate_PDU = -1;  /* UE_AggregateMaximumBitRate */
static int hf_nbap_UE_AggregateMaximumBitRate_Enforcement_Indicator_PDU = -1;  /* UE_AggregateMaximumBitRate_Enforcement_Indicator */
static int hf_nbap_UE_Capability_Information_PDU = -1;  /* UE_Capability_Information */
static int hf_nbap_UE_TS0_CapabilityLCR_PDU = -1;  /* UE_TS0_CapabilityLCR */
static int hf_nbap_UE_SupportIndicatorExtension_PDU = -1;  /* UE_SupportIndicatorExtension */
static int hf_nbap_LCRTDD_HSDSCH_Physical_Layer_Category_PDU = -1;  /* LCRTDD_HSDSCH_Physical_Layer_Category */
static int hf_nbap_UL_DPDCH_Indicator_For_E_DCH_Operation_PDU = -1;  /* UL_DPDCH_Indicator_For_E_DCH_Operation */
static int hf_nbap_UL_SIR_PDU = -1;               /* UL_SIR */
static int hf_nbap_UL_Synchronisation_Parameters_LCR_PDU = -1;  /* UL_Synchronisation_Parameters_LCR */
static int hf_nbap_UPPCHPositionLCR_PDU = -1;     /* UPPCHPositionLCR */
static int hf_nbap_UpPTSInterference_For_CellPortion_Value_PDU = -1;  /* UpPTSInterference_For_CellPortion_Value */
static int hf_nbap_UpPTSInterferenceValue_PDU = -1;  /* UpPTSInterferenceValue */
static int hf_nbap_Unidirectional_DCH_Indicator_PDU = -1;  /* Unidirectional_DCH_Indicator */
static int hf_nbap_USCH_Information_PDU = -1;     /* USCH_Information */
static int hf_nbap_USCH_InformationResponse_PDU = -1;  /* USCH_InformationResponse */
static int hf_nbap_UL_TimeslotISCP_For_CellPortion_Value_PDU = -1;  /* UL_TimeslotISCP_For_CellPortion_Value */
static int hf_nbap_UL_TimeslotISCP_Value_IncrDecrThres_PDU = -1;  /* UL_TimeslotISCP_Value_IncrDecrThres */
static int hf_nbap_CommonTransportChannelSetupRequestFDD_PDU = -1;  /* CommonTransportChannelSetupRequestFDD */
static int hf_nbap_CommonPhysicalChannelType_CTCH_SetupRqstFDD_PDU = -1;  /* CommonPhysicalChannelType_CTCH_SetupRqstFDD */
static int hf_nbap_FACH_ParametersListIE_CTCH_SetupRqstFDD_PDU = -1;  /* FACH_ParametersListIE_CTCH_SetupRqstFDD */
static int hf_nbap_PCH_ParametersItem_CTCH_SetupRqstFDD_PDU = -1;  /* PCH_ParametersItem_CTCH_SetupRqstFDD */
static int hf_nbap_MICH_Parameters_CTCH_SetupRqstFDD_PDU = -1;  /* MICH_Parameters_CTCH_SetupRqstFDD */
static int hf_nbap_RACH_ParametersItem_CTCH_SetupRqstFDD_PDU = -1;  /* RACH_ParametersItem_CTCH_SetupRqstFDD */
static int hf_nbap_CommonTransportChannelSetupRequestTDD_PDU = -1;  /* CommonTransportChannelSetupRequestTDD */
static int hf_nbap_CommonPhysicalChannelType_CTCH_SetupRqstTDD_PDU = -1;  /* CommonPhysicalChannelType_CTCH_SetupRqstTDD */
static int hf_nbap_Secondary_CCPCH_parameterListIE_CTCH_SetupRqstTDD_PDU = -1;  /* Secondary_CCPCH_parameterListIE_CTCH_SetupRqstTDD */
static int hf_nbap_Secondary_CCPCH_LCR_parameterList_CTCH_SetupRqstTDD_PDU = -1;  /* Secondary_CCPCH_LCR_parameterList_CTCH_SetupRqstTDD */
static int hf_nbap_Secondary_CCPCH_768_parameterList_CTCH_SetupRqstTDD_PDU = -1;  /* Secondary_CCPCH_768_parameterList_CTCH_SetupRqstTDD */
static int hf_nbap_FACH_ParametersListIE_CTCH_SetupRqstTDD_PDU = -1;  /* FACH_ParametersListIE_CTCH_SetupRqstTDD */
static int hf_nbap_PCH_ParametersItem_CTCH_SetupRqstTDD_PDU = -1;  /* PCH_ParametersItem_CTCH_SetupRqstTDD */
static int hf_nbap_PICH_ParametersItem_CTCH_SetupRqstTDD_PDU = -1;  /* PICH_ParametersItem_CTCH_SetupRqstTDD */
static int hf_nbap_PICH_LCR_Parameters_CTCH_SetupRqstTDD_PDU = -1;  /* PICH_LCR_Parameters_CTCH_SetupRqstTDD */
static int hf_nbap_PICH_768_ParametersItem_CTCH_SetupRqstTDD_PDU = -1;  /* PICH_768_ParametersItem_CTCH_SetupRqstTDD */
static int hf_nbap_MICH_Parameters_CTCH_SetupRqstTDD_PDU = -1;  /* MICH_Parameters_CTCH_SetupRqstTDD */
static int hf_nbap_TimeSlotConfigurationList_LCR_CTCH_SetupRqstTDD_PDU = -1;  /* TimeSlotConfigurationList_LCR_CTCH_SetupRqstTDD */
static int hf_nbap_Secondary_CCPCH_parameterExtendedList_CTCH_SetupRqstTDD_PDU = -1;  /* Secondary_CCPCH_parameterExtendedList_CTCH_SetupRqstTDD */
static int hf_nbap_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_SetupRqstTDD_PDU = -1;  /* Secondary_CCPCH_LCR_parameterExtendedList_CTCH_SetupRqstTDD */
static int hf_nbap_PRACH_ParametersItem_CTCH_SetupRqstTDD_PDU = -1;  /* PRACH_ParametersItem_CTCH_SetupRqstTDD */
static int hf_nbap_RACH_ParameterItem_CTCH_SetupRqstTDD_PDU = -1;  /* RACH_ParameterItem_CTCH_SetupRqstTDD */
static int hf_nbap_PRACH_LCR_ParametersList_CTCH_SetupRqstTDD_PDU = -1;  /* PRACH_LCR_ParametersList_CTCH_SetupRqstTDD */
static int hf_nbap_PRACH_768_ParametersItem_CTCH_SetupRqstTDD_PDU = -1;  /* PRACH_768_ParametersItem_CTCH_SetupRqstTDD */
static int hf_nbap_FPACH_LCR_Parameters_CTCH_SetupRqstTDD_PDU = -1;  /* FPACH_LCR_Parameters_CTCH_SetupRqstTDD */
static int hf_nbap_PLCCH_parameters_PDU = -1;     /* PLCCH_parameters */
static int hf_nbap_E_RUCCH_parameters_PDU = -1;   /* E_RUCCH_parameters */
static int hf_nbap_E_RUCCH_768_parameters_PDU = -1;  /* E_RUCCH_768_parameters */
static int hf_nbap_CommonTransportChannelSetupResponse_PDU = -1;  /* CommonTransportChannelSetupResponse */
static int hf_nbap_FACH_CommonTransportChannel_InformationResponse_PDU = -1;  /* FACH_CommonTransportChannel_InformationResponse */
static int hf_nbap_CommonTransportChannelSetupFailure_PDU = -1;  /* CommonTransportChannelSetupFailure */
static int hf_nbap_CommonTransportChannelReconfigurationRequestFDD_PDU = -1;  /* CommonTransportChannelReconfigurationRequestFDD */
static int hf_nbap_CommonPhysicalChannelType_CTCH_ReconfRqstFDD_PDU = -1;  /* CommonPhysicalChannelType_CTCH_ReconfRqstFDD */
static int hf_nbap_FACH_ParametersListIE_CTCH_ReconfRqstFDD_PDU = -1;  /* FACH_ParametersListIE_CTCH_ReconfRqstFDD */
static int hf_nbap_PCH_ParametersItem_CTCH_ReconfRqstFDD_PDU = -1;  /* PCH_ParametersItem_CTCH_ReconfRqstFDD */
static int hf_nbap_PICH_ParametersItem_CTCH_ReconfRqstFDD_PDU = -1;  /* PICH_ParametersItem_CTCH_ReconfRqstFDD */
static int hf_nbap_MICH_Parameters_CTCH_ReconfRqstFDD_PDU = -1;  /* MICH_Parameters_CTCH_ReconfRqstFDD */
static int hf_nbap_PRACH_ParametersListIE_CTCH_ReconfRqstFDD_PDU = -1;  /* PRACH_ParametersListIE_CTCH_ReconfRqstFDD */
static int hf_nbap_AICH_ParametersListIE_CTCH_ReconfRqstFDD_PDU = -1;  /* AICH_ParametersListIE_CTCH_ReconfRqstFDD */
static int hf_nbap_CommonTransportChannelReconfigurationRequestTDD_PDU = -1;  /* CommonTransportChannelReconfigurationRequestTDD */
static int hf_nbap_Secondary_CCPCH_Parameters_CTCH_ReconfRqstTDD_PDU = -1;  /* Secondary_CCPCH_Parameters_CTCH_ReconfRqstTDD */
static int hf_nbap_Secondary_CCPCHListIE_CTCH_ReconfRqstTDD_PDU = -1;  /* Secondary_CCPCHListIE_CTCH_ReconfRqstTDD */
static int hf_nbap_Secondary_CCPCH_parameterExtendedList_CTCH_ReconfRqstTDD_PDU = -1;  /* Secondary_CCPCH_parameterExtendedList_CTCH_ReconfRqstTDD */
static int hf_nbap_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_ReconfRqstTDD_PDU = -1;  /* Secondary_CCPCH_LCR_parameterExtendedList_CTCH_ReconfRqstTDD */
static int hf_nbap_PICH_Parameters_CTCH_ReconfRqstTDD_PDU = -1;  /* PICH_Parameters_CTCH_ReconfRqstTDD */
static int hf_nbap_FACH_ParametersList_CTCH_ReconfRqstTDD_PDU = -1;  /* FACH_ParametersList_CTCH_ReconfRqstTDD */
static int hf_nbap_PCH_Parameters_CTCH_ReconfRqstTDD_PDU = -1;  /* PCH_Parameters_CTCH_ReconfRqstTDD */
static int hf_nbap_FPACH_LCR_Parameters_CTCH_ReconfRqstTDD_PDU = -1;  /* FPACH_LCR_Parameters_CTCH_ReconfRqstTDD */
static int hf_nbap_MICH_Parameters_CTCH_ReconfRqstTDD_PDU = -1;  /* MICH_Parameters_CTCH_ReconfRqstTDD */
static int hf_nbap_PLCCH_Parameters_CTCH_ReconfRqstTDD_PDU = -1;  /* PLCCH_Parameters_CTCH_ReconfRqstTDD */
static int hf_nbap_Secondary_CCPCH_768_Parameters_CTCH_ReconfRqstTDD_PDU = -1;  /* Secondary_CCPCH_768_Parameters_CTCH_ReconfRqstTDD */
static int hf_nbap_PICH_768_Parameters_CTCH_ReconfRqstTDD_PDU = -1;  /* PICH_768_Parameters_CTCH_ReconfRqstTDD */
static int hf_nbap_MICH_768_Parameters_CTCH_ReconfRqstTDD_PDU = -1;  /* MICH_768_Parameters_CTCH_ReconfRqstTDD */
static int hf_nbap_UPPCH_LCR_Parameters_CTCH_ReconfRqstTDD_PDU = -1;  /* UPPCH_LCR_Parameters_CTCH_ReconfRqstTDD */
static int hf_nbap_CommonTransportChannelReconfigurationResponse_PDU = -1;  /* CommonTransportChannelReconfigurationResponse */
static int hf_nbap_CommonTransportChannelReconfigurationFailure_PDU = -1;  /* CommonTransportChannelReconfigurationFailure */
static int hf_nbap_CommonTransportChannelDeletionRequest_PDU = -1;  /* CommonTransportChannelDeletionRequest */
static int hf_nbap_CommonTransportChannelDeletionResponse_PDU = -1;  /* CommonTransportChannelDeletionResponse */
static int hf_nbap_BlockResourceRequest_PDU = -1;  /* BlockResourceRequest */
static int hf_nbap_BlockResourceResponse_PDU = -1;  /* BlockResourceResponse */
static int hf_nbap_BlockResourceFailure_PDU = -1;  /* BlockResourceFailure */
static int hf_nbap_UnblockResourceIndication_PDU = -1;  /* UnblockResourceIndication */
static int hf_nbap_AuditRequiredIndication_PDU = -1;  /* AuditRequiredIndication */
static int hf_nbap_AuditRequest_PDU = -1;         /* AuditRequest */
static int hf_nbap_AuditResponse_PDU = -1;        /* AuditResponse */
static int hf_nbap_Cell_InformationList_AuditRsp_PDU = -1;  /* Cell_InformationList_AuditRsp */
static int hf_nbap_Cell_InformationItem_AuditRsp_PDU = -1;  /* Cell_InformationItem_AuditRsp */
static int hf_nbap_FPACH_LCR_InformationList_AuditRsp_PDU = -1;  /* FPACH_LCR_InformationList_AuditRsp */
static int hf_nbap_HS_DSCH_Resources_Information_AuditRsp_PDU = -1;  /* HS_DSCH_Resources_Information_AuditRsp */
static int hf_nbap_S_CCPCH_InformationListExt_AuditRsp_PDU = -1;  /* S_CCPCH_InformationListExt_AuditRsp */
static int hf_nbap_S_CCPCH_LCR_InformationListExt_AuditRsp_PDU = -1;  /* S_CCPCH_LCR_InformationListExt_AuditRsp */
static int hf_nbap_E_DCH_Resources_Information_AuditRsp_PDU = -1;  /* E_DCH_Resources_Information_AuditRsp */
static int hf_nbap_PLCCH_InformationList_AuditRsp_PDU = -1;  /* PLCCH_InformationList_AuditRsp */
static int hf_nbap_S_CCPCH_768_InformationList_AuditRsp_PDU = -1;  /* S_CCPCH_768_InformationList_AuditRsp */
static int hf_nbap_PRACH_768_InformationList_AuditRsp_PDU = -1;  /* PRACH_768_InformationList_AuditRsp */
static int hf_nbap_E_RUCCH_InformationList_AuditRsp_PDU = -1;  /* E_RUCCH_InformationList_AuditRsp */
static int hf_nbap_E_RUCCH_768_InformationList_AuditRsp_PDU = -1;  /* E_RUCCH_768_InformationList_AuditRsp */
static int hf_nbap_Cell_Frequency_List_Information_LCR_MulFreq_AuditRsp_PDU = -1;  /* Cell_Frequency_List_Information_LCR_MulFreq_AuditRsp */
static int hf_nbap_Cell_Frequency_List_InformationItem_LCR_MulFreq_AuditRsp_PDU = -1;  /* Cell_Frequency_List_InformationItem_LCR_MulFreq_AuditRsp */
static int hf_nbap_UPPCH_LCR_InformationList_AuditRsp_PDU = -1;  /* UPPCH_LCR_InformationList_AuditRsp */
static int hf_nbap_UPPCH_LCR_InformationItem_AuditRsp_PDU = -1;  /* UPPCH_LCR_InformationItem_AuditRsp */
static int hf_nbap_MultipleFreq_HS_DSCH_Resources_InformationList_AuditRsp_PDU = -1;  /* MultipleFreq_HS_DSCH_Resources_InformationList_AuditRsp */
static int hf_nbap_MultipleFreq_E_DCH_Resources_InformationList_AuditRsp_PDU = -1;  /* MultipleFreq_E_DCH_Resources_InformationList_AuditRsp */
static int hf_nbap_CCP_InformationList_AuditRsp_PDU = -1;  /* CCP_InformationList_AuditRsp */
static int hf_nbap_CCP_InformationItem_AuditRsp_PDU = -1;  /* CCP_InformationItem_AuditRsp */
static int hf_nbap_Local_Cell_InformationList_AuditRsp_PDU = -1;  /* Local_Cell_InformationList_AuditRsp */
static int hf_nbap_Local_Cell_InformationItem_AuditRsp_PDU = -1;  /* Local_Cell_InformationItem_AuditRsp */
static int hf_nbap_Local_Cell_Group_InformationList_AuditRsp_PDU = -1;  /* Local_Cell_Group_InformationList_AuditRsp */
static int hf_nbap_Local_Cell_Group_InformationItem_AuditRsp_PDU = -1;  /* Local_Cell_Group_InformationItem_AuditRsp */
static int hf_nbap_Power_Local_Cell_Group_InformationList_AuditRsp_PDU = -1;  /* Power_Local_Cell_Group_InformationList_AuditRsp */
static int hf_nbap_Power_Local_Cell_Group_InformationItem_AuditRsp_PDU = -1;  /* Power_Local_Cell_Group_InformationItem_AuditRsp */
static int hf_nbap_AuditFailure_PDU = -1;         /* AuditFailure */
static int hf_nbap_CommonMeasurementInitiationRequest_PDU = -1;  /* CommonMeasurementInitiationRequest */
static int hf_nbap_CommonMeasurementObjectType_CM_Rqst_PDU = -1;  /* CommonMeasurementObjectType_CM_Rqst */
static int hf_nbap_ERACH_CM_Rqst_PDU = -1;        /* ERACH_CM_Rqst */
static int hf_nbap_PowerLocalCellGroup_CM_Rqst_PDU = -1;  /* PowerLocalCellGroup_CM_Rqst */
static int hf_nbap_CommonMeasurementInitiationResponse_PDU = -1;  /* CommonMeasurementInitiationResponse */
static int hf_nbap_CommonMeasurementObjectType_CM_Rsp_PDU = -1;  /* CommonMeasurementObjectType_CM_Rsp */
static int hf_nbap_ERACH_CM_Rsp_PDU = -1;         /* ERACH_CM_Rsp */
static int hf_nbap_PowerLocalCellGroup_CM_Rsp_PDU = -1;  /* PowerLocalCellGroup_CM_Rsp */
static int hf_nbap_CommonMeasurementInitiationFailure_PDU = -1;  /* CommonMeasurementInitiationFailure */
static int hf_nbap_CommonMeasurementReport_PDU = -1;  /* CommonMeasurementReport */
static int hf_nbap_CommonMeasurementObjectType_CM_Rprt_PDU = -1;  /* CommonMeasurementObjectType_CM_Rprt */
static int hf_nbap_ERACH_CM_Rprt_PDU = -1;        /* ERACH_CM_Rprt */
static int hf_nbap_PowerLocalCellGroup_CM_Rprt_PDU = -1;  /* PowerLocalCellGroup_CM_Rprt */
static int hf_nbap_CommonMeasurementTerminationRequest_PDU = -1;  /* CommonMeasurementTerminationRequest */
static int hf_nbap_CommonMeasurementFailureIndication_PDU = -1;  /* CommonMeasurementFailureIndication */
static int hf_nbap_CellSetupRequestFDD_PDU = -1;  /* CellSetupRequestFDD */
static int hf_nbap_Synchronisation_Configuration_Cell_SetupRqst_PDU = -1;  /* Synchronisation_Configuration_Cell_SetupRqst */
static int hf_nbap_PrimarySCH_Information_Cell_SetupRqstFDD_PDU = -1;  /* PrimarySCH_Information_Cell_SetupRqstFDD */
static int hf_nbap_SecondarySCH_Information_Cell_SetupRqstFDD_PDU = -1;  /* SecondarySCH_Information_Cell_SetupRqstFDD */
static int hf_nbap_PrimaryCPICH_Information_Cell_SetupRqstFDD_PDU = -1;  /* PrimaryCPICH_Information_Cell_SetupRqstFDD */
static int hf_nbap_SecondaryCPICH_InformationList_Cell_SetupRqstFDD_PDU = -1;  /* SecondaryCPICH_InformationList_Cell_SetupRqstFDD */
static int hf_nbap_SecondaryCPICH_InformationItem_Cell_SetupRqstFDD_PDU = -1;  /* SecondaryCPICH_InformationItem_Cell_SetupRqstFDD */
static int hf_nbap_PrimaryCCPCH_Information_Cell_SetupRqstFDD_PDU = -1;  /* PrimaryCCPCH_Information_Cell_SetupRqstFDD */
static int hf_nbap_Limited_power_increase_information_Cell_SetupRqstFDD_PDU = -1;  /* Limited_power_increase_information_Cell_SetupRqstFDD */
static int hf_nbap_IPDLParameter_Information_Cell_SetupRqstFDD_PDU = -1;  /* IPDLParameter_Information_Cell_SetupRqstFDD */
static int hf_nbap_CellPortion_InformationList_Cell_SetupRqstFDD_PDU = -1;  /* CellPortion_InformationList_Cell_SetupRqstFDD */
static int hf_nbap_CellPortion_InformationItem_Cell_SetupRqstFDD_PDU = -1;  /* CellPortion_InformationItem_Cell_SetupRqstFDD */
static int hf_nbap_CellSetupRequestTDD_PDU = -1;  /* CellSetupRequestTDD */
static int hf_nbap_SCH_Information_Cell_SetupRqstTDD_PDU = -1;  /* SCH_Information_Cell_SetupRqstTDD */
static int hf_nbap_SyncCaseIndicatorItem_Cell_SetupRqstTDD_PSCH_PDU = -1;  /* SyncCaseIndicatorItem_Cell_SetupRqstTDD_PSCH */
static int hf_nbap_PCCPCH_Information_Cell_SetupRqstTDD_PDU = -1;  /* PCCPCH_Information_Cell_SetupRqstTDD */
static int hf_nbap_TimeSlotConfigurationList_Cell_SetupRqstTDD_PDU = -1;  /* TimeSlotConfigurationList_Cell_SetupRqstTDD */
static int hf_nbap_TimeSlotConfigurationList_LCR_Cell_SetupRqstTDD_PDU = -1;  /* TimeSlotConfigurationList_LCR_Cell_SetupRqstTDD */
static int hf_nbap_PCCPCH_LCR_Information_Cell_SetupRqstTDD_PDU = -1;  /* PCCPCH_LCR_Information_Cell_SetupRqstTDD */
static int hf_nbap_DwPCH_LCR_Information_Cell_SetupRqstTDD_PDU = -1;  /* DwPCH_LCR_Information_Cell_SetupRqstTDD */
static int hf_nbap_IPDLParameter_Information_Cell_SetupRqstTDD_PDU = -1;  /* IPDLParameter_Information_Cell_SetupRqstTDD */
static int hf_nbap_IPDLParameter_Information_LCR_Cell_SetupRqstTDD_PDU = -1;  /* IPDLParameter_Information_LCR_Cell_SetupRqstTDD */
static int hf_nbap_PCCPCH_768_Information_Cell_SetupRqstTDD_PDU = -1;  /* PCCPCH_768_Information_Cell_SetupRqstTDD */
static int hf_nbap_SCH_768_Information_Cell_SetupRqstTDD_PDU = -1;  /* SCH_768_Information_Cell_SetupRqstTDD */
static int hf_nbap_Cell_Frequency_List_LCR_MulFreq_Cell_SetupRqstTDD_PDU = -1;  /* Cell_Frequency_List_LCR_MulFreq_Cell_SetupRqstTDD */
static int hf_nbap_CellSetupResponse_PDU = -1;    /* CellSetupResponse */
static int hf_nbap_CellSetupFailure_PDU = -1;     /* CellSetupFailure */
static int hf_nbap_CellReconfigurationRequestFDD_PDU = -1;  /* CellReconfigurationRequestFDD */
static int hf_nbap_Synchronisation_Configuration_Cell_ReconfRqst_PDU = -1;  /* Synchronisation_Configuration_Cell_ReconfRqst */
static int hf_nbap_PrimarySCH_Information_Cell_ReconfRqstFDD_PDU = -1;  /* PrimarySCH_Information_Cell_ReconfRqstFDD */
static int hf_nbap_SecondarySCH_Information_Cell_ReconfRqstFDD_PDU = -1;  /* SecondarySCH_Information_Cell_ReconfRqstFDD */
static int hf_nbap_PrimaryCPICH_Information_Cell_ReconfRqstFDD_PDU = -1;  /* PrimaryCPICH_Information_Cell_ReconfRqstFDD */
static int hf_nbap_SecondaryCPICH_InformationList_Cell_ReconfRqstFDD_PDU = -1;  /* SecondaryCPICH_InformationList_Cell_ReconfRqstFDD */
static int hf_nbap_SecondaryCPICH_InformationItem_Cell_ReconfRqstFDD_PDU = -1;  /* SecondaryCPICH_InformationItem_Cell_ReconfRqstFDD */
static int hf_nbap_PrimaryCCPCH_Information_Cell_ReconfRqstFDD_PDU = -1;  /* PrimaryCCPCH_Information_Cell_ReconfRqstFDD */
static int hf_nbap_IPDLParameter_Information_Cell_ReconfRqstFDD_PDU = -1;  /* IPDLParameter_Information_Cell_ReconfRqstFDD */
static int hf_nbap_CellPortion_InformationList_Cell_ReconfRqstFDD_PDU = -1;  /* CellPortion_InformationList_Cell_ReconfRqstFDD */
static int hf_nbap_CellPortion_InformationItem_Cell_ReconfRqstFDD_PDU = -1;  /* CellPortion_InformationItem_Cell_ReconfRqstFDD */
static int hf_nbap_CellReconfigurationRequestTDD_PDU = -1;  /* CellReconfigurationRequestTDD */
static int hf_nbap_SCH_Information_Cell_ReconfRqstTDD_PDU = -1;  /* SCH_Information_Cell_ReconfRqstTDD */
static int hf_nbap_PCCPCH_Information_Cell_ReconfRqstTDD_PDU = -1;  /* PCCPCH_Information_Cell_ReconfRqstTDD */
static int hf_nbap_TimeSlotConfigurationList_Cell_ReconfRqstTDD_PDU = -1;  /* TimeSlotConfigurationList_Cell_ReconfRqstTDD */
static int hf_nbap_TimeSlotConfigurationList_LCR_Cell_ReconfRqstTDD_PDU = -1;  /* TimeSlotConfigurationList_LCR_Cell_ReconfRqstTDD */
static int hf_nbap_DwPCH_LCR_Information_Cell_ReconfRqstTDD_PDU = -1;  /* DwPCH_LCR_Information_Cell_ReconfRqstTDD */
static int hf_nbap_IPDLParameter_Information_Cell_ReconfRqstTDD_PDU = -1;  /* IPDLParameter_Information_Cell_ReconfRqstTDD */
static int hf_nbap_IPDLParameter_Information_LCR_Cell_ReconfRqstTDD_PDU = -1;  /* IPDLParameter_Information_LCR_Cell_ReconfRqstTDD */
static int hf_nbap_SCH_768_Information_Cell_ReconfRqstTDD_PDU = -1;  /* SCH_768_Information_Cell_ReconfRqstTDD */
static int hf_nbap_PCCPCH_768_Information_Cell_ReconfRqstTDD_PDU = -1;  /* PCCPCH_768_Information_Cell_ReconfRqstTDD */
static int hf_nbap_UARFCN_Adjustment_PDU = -1;    /* UARFCN_Adjustment */
static int hf_nbap_CellReconfigurationResponse_PDU = -1;  /* CellReconfigurationResponse */
static int hf_nbap_CellReconfigurationFailure_PDU = -1;  /* CellReconfigurationFailure */
static int hf_nbap_CellDeletionRequest_PDU = -1;  /* CellDeletionRequest */
static int hf_nbap_CellDeletionResponse_PDU = -1;  /* CellDeletionResponse */
static int hf_nbap_ResourceStatusIndication_PDU = -1;  /* ResourceStatusIndication */
static int hf_nbap_IndicationType_ResourceStatusInd_PDU = -1;  /* IndicationType_ResourceStatusInd */
static int hf_nbap_Local_Cell_InformationItem_ResourceStatusInd_PDU = -1;  /* Local_Cell_InformationItem_ResourceStatusInd */
static int hf_nbap_Local_Cell_Group_InformationItem_ResourceStatusInd_PDU = -1;  /* Local_Cell_Group_InformationItem_ResourceStatusInd */
static int hf_nbap_Power_Local_Cell_Group_InformationList_ResourceStatusInd_PDU = -1;  /* Power_Local_Cell_Group_InformationList_ResourceStatusInd */
static int hf_nbap_Power_Local_Cell_Group_InformationItem_ResourceStatusInd_PDU = -1;  /* Power_Local_Cell_Group_InformationItem_ResourceStatusInd */
static int hf_nbap_Local_Cell_InformationItem2_ResourceStatusInd_PDU = -1;  /* Local_Cell_InformationItem2_ResourceStatusInd */
static int hf_nbap_Local_Cell_Group_InformationItem2_ResourceStatusInd_PDU = -1;  /* Local_Cell_Group_InformationItem2_ResourceStatusInd */
static int hf_nbap_CCP_InformationItem_ResourceStatusInd_PDU = -1;  /* CCP_InformationItem_ResourceStatusInd */
static int hf_nbap_Cell_InformationItem_ResourceStatusInd_PDU = -1;  /* Cell_InformationItem_ResourceStatusInd */
static int hf_nbap_FPACH_LCR_InformationList_ResourceStatusInd_PDU = -1;  /* FPACH_LCR_InformationList_ResourceStatusInd */
static int hf_nbap_DwPCH_LCR_Information_ResourceStatusInd_PDU = -1;  /* DwPCH_LCR_Information_ResourceStatusInd */
static int hf_nbap_HS_DSCH_Resources_Information_ResourceStatusInd_PDU = -1;  /* HS_DSCH_Resources_Information_ResourceStatusInd */
static int hf_nbap_S_CCPCH_InformationListExt_ResourceStatusInd_PDU = -1;  /* S_CCPCH_InformationListExt_ResourceStatusInd */
static int hf_nbap_S_CCPCH_LCR_InformationListExt_ResourceStatusInd_PDU = -1;  /* S_CCPCH_LCR_InformationListExt_ResourceStatusInd */
static int hf_nbap_E_DCH_Resources_Information_ResourceStatusInd_PDU = -1;  /* E_DCH_Resources_Information_ResourceStatusInd */
static int hf_nbap_PLCCH_InformationList_ResourceStatusInd_PDU = -1;  /* PLCCH_InformationList_ResourceStatusInd */
static int hf_nbap_S_CCPCH_768_InformationList_ResourceStatusInd_PDU = -1;  /* S_CCPCH_768_InformationList_ResourceStatusInd */
static int hf_nbap_PRACH_768_InformationList_ResourceStatusInd_PDU = -1;  /* PRACH_768_InformationList_ResourceStatusInd */
static int hf_nbap_E_RUCCH_InformationList_ResourceStatusInd_PDU = -1;  /* E_RUCCH_InformationList_ResourceStatusInd */
static int hf_nbap_E_RUCCH_768_InformationList_ResourceStatusInd_PDU = -1;  /* E_RUCCH_768_InformationList_ResourceStatusInd */
static int hf_nbap_Cell_Frequency_List_Information_LCR_MulFreq_ResourceStatusInd_PDU = -1;  /* Cell_Frequency_List_Information_LCR_MulFreq_ResourceStatusInd */
static int hf_nbap_Cell_Frequency_List_InformationItem_LCR_MulFreq_ResourceStatusInd_PDU = -1;  /* Cell_Frequency_List_InformationItem_LCR_MulFreq_ResourceStatusInd */
static int hf_nbap_UPPCH_LCR_InformationList_ResourceStatusInd_PDU = -1;  /* UPPCH_LCR_InformationList_ResourceStatusInd */
static int hf_nbap_UPPCH_LCR_InformationItem_ResourceStatusInd_PDU = -1;  /* UPPCH_LCR_InformationItem_ResourceStatusInd */
static int hf_nbap_MultipleFreq_HS_DSCH_Resources_InformationList_ResourceStatusInd_PDU = -1;  /* MultipleFreq_HS_DSCH_Resources_InformationList_ResourceStatusInd */
static int hf_nbap_Power_Local_Cell_Group_InformationList2_ResourceStatusInd_PDU = -1;  /* Power_Local_Cell_Group_InformationList2_ResourceStatusInd */
static int hf_nbap_Power_Local_Cell_Group_InformationItem2_ResourceStatusInd_PDU = -1;  /* Power_Local_Cell_Group_InformationItem2_ResourceStatusInd */
static int hf_nbap_MultipleFreq_E_DCH_Resources_InformationList_ResourceStatusInd_PDU = -1;  /* MultipleFreq_E_DCH_Resources_InformationList_ResourceStatusInd */
static int hf_nbap_SystemInformationUpdateRequest_PDU = -1;  /* SystemInformationUpdateRequest */
static int hf_nbap_MIB_SB_SIB_InformationList_SystemInfoUpdateRqst_PDU = -1;  /* MIB_SB_SIB_InformationList_SystemInfoUpdateRqst */
static int hf_nbap_SegmentInformationListIE_SystemInfoUpdate_PDU = -1;  /* SegmentInformationListIE_SystemInfoUpdate */
static int hf_nbap_SystemInformationUpdateResponse_PDU = -1;  /* SystemInformationUpdateResponse */
static int hf_nbap_SystemInformationUpdateFailure_PDU = -1;  /* SystemInformationUpdateFailure */
static int hf_nbap_RadioLinkSetupRequestFDD_PDU = -1;  /* RadioLinkSetupRequestFDD */
static int hf_nbap_Additional_HS_Cell_Information_RL_Setup_List_PDU = -1;  /* Additional_HS_Cell_Information_RL_Setup_List */
static int hf_nbap_UL_DPCH_Information_RL_SetupRqstFDD_PDU = -1;  /* UL_DPCH_Information_RL_SetupRqstFDD */
static int hf_nbap_DL_DPCH_Information_RL_SetupRqstFDD_PDU = -1;  /* DL_DPCH_Information_RL_SetupRqstFDD */
static int hf_nbap_RL_InformationList_RL_SetupRqstFDD_PDU = -1;  /* RL_InformationList_RL_SetupRqstFDD */
static int hf_nbap_RL_InformationItem_RL_SetupRqstFDD_PDU = -1;  /* RL_InformationItem_RL_SetupRqstFDD */
static int hf_nbap_E_DPCH_Information_RL_SetupRqstFDD_PDU = -1;  /* E_DPCH_Information_RL_SetupRqstFDD */
static int hf_nbap_F_DPCH_Information_RL_SetupRqstFDD_PDU = -1;  /* F_DPCH_Information_RL_SetupRqstFDD */
static int hf_nbap_RadioLinkSetupRequestTDD_PDU = -1;  /* RadioLinkSetupRequestTDD */
static int hf_nbap_UL_CCTrCH_InformationList_RL_SetupRqstTDD_PDU = -1;  /* UL_CCTrCH_InformationList_RL_SetupRqstTDD */
static int hf_nbap_UL_CCTrCH_InformationItem_RL_SetupRqstTDD_PDU = -1;  /* UL_CCTrCH_InformationItem_RL_SetupRqstTDD */
static int hf_nbap_UL_DPCH_InformationItem_RL_SetupRqstTDD_PDU = -1;  /* UL_DPCH_InformationItem_RL_SetupRqstTDD */
static int hf_nbap_UL_DPCH_LCR_Information_RL_SetupRqstTDD_PDU = -1;  /* UL_DPCH_LCR_Information_RL_SetupRqstTDD */
static int hf_nbap_UL_DPCH_768_Information_RL_SetupRqstTDD_PDU = -1;  /* UL_DPCH_768_Information_RL_SetupRqstTDD */
static int hf_nbap_DL_CCTrCH_InformationList_RL_SetupRqstTDD_PDU = -1;  /* DL_CCTrCH_InformationList_RL_SetupRqstTDD */
static int hf_nbap_DL_CCTrCH_InformationItem_RL_SetupRqstTDD_PDU = -1;  /* DL_CCTrCH_InformationItem_RL_SetupRqstTDD */
static int hf_nbap_DL_DPCH_InformationItem_RL_SetupRqstTDD_PDU = -1;  /* DL_DPCH_InformationItem_RL_SetupRqstTDD */
static int hf_nbap_DL_DPCH_LCR_Information_RL_SetupRqstTDD_PDU = -1;  /* DL_DPCH_LCR_Information_RL_SetupRqstTDD */
static int hf_nbap_DL_DPCH_768_Information_RL_SetupRqstTDD_PDU = -1;  /* DL_DPCH_768_Information_RL_SetupRqstTDD */
static int hf_nbap_RL_Information_RL_SetupRqstTDD_PDU = -1;  /* RL_Information_RL_SetupRqstTDD */
static int hf_nbap_RadioLinkSetupResponseFDD_PDU = -1;  /* RadioLinkSetupResponseFDD */
static int hf_nbap_Additional_HS_Cell_Information_Response_List_PDU = -1;  /* Additional_HS_Cell_Information_Response_List */
static int hf_nbap_RL_InformationResponseList_RL_SetupRspFDD_PDU = -1;  /* RL_InformationResponseList_RL_SetupRspFDD */
static int hf_nbap_RL_InformationResponseItem_RL_SetupRspFDD_PDU = -1;  /* RL_InformationResponseItem_RL_SetupRspFDD */
static int hf_nbap_RadioLinkSetupResponseTDD_PDU = -1;  /* RadioLinkSetupResponseTDD */
static int hf_nbap_RL_InformationResponse_RL_SetupRspTDD_PDU = -1;  /* RL_InformationResponse_RL_SetupRspTDD */
static int hf_nbap_RL_InformationResponse_LCR_RL_SetupRspTDD_PDU = -1;  /* RL_InformationResponse_LCR_RL_SetupRspTDD */
static int hf_nbap_RadioLinkSetupFailureFDD_PDU = -1;  /* RadioLinkSetupFailureFDD */
static int hf_nbap_CauseLevel_RL_SetupFailureFDD_PDU = -1;  /* CauseLevel_RL_SetupFailureFDD */
static int hf_nbap_Unsuccessful_RL_InformationRespItem_RL_SetupFailureFDD_PDU = -1;  /* Unsuccessful_RL_InformationRespItem_RL_SetupFailureFDD */
static int hf_nbap_Successful_RL_InformationRespItem_RL_SetupFailureFDD_PDU = -1;  /* Successful_RL_InformationRespItem_RL_SetupFailureFDD */
static int hf_nbap_RadioLinkSetupFailureTDD_PDU = -1;  /* RadioLinkSetupFailureTDD */
static int hf_nbap_CauseLevel_RL_SetupFailureTDD_PDU = -1;  /* CauseLevel_RL_SetupFailureTDD */
static int hf_nbap_Unsuccessful_RL_InformationResp_RL_SetupFailureTDD_PDU = -1;  /* Unsuccessful_RL_InformationResp_RL_SetupFailureTDD */
static int hf_nbap_RadioLinkAdditionRequestFDD_PDU = -1;  /* RadioLinkAdditionRequestFDD */
static int hf_nbap_Additional_HS_Cell_Information_RL_Addition_List_PDU = -1;  /* Additional_HS_Cell_Information_RL_Addition_List */
static int hf_nbap_Additional_EDCH_Cell_Information_RL_Add_Req_PDU = -1;  /* Additional_EDCH_Cell_Information_RL_Add_Req */
static int hf_nbap_RL_InformationList_RL_AdditionRqstFDD_PDU = -1;  /* RL_InformationList_RL_AdditionRqstFDD */
static int hf_nbap_RL_InformationItem_RL_AdditionRqstFDD_PDU = -1;  /* RL_InformationItem_RL_AdditionRqstFDD */
static int hf_nbap_E_DPCH_Information_RL_AdditionReqFDD_PDU = -1;  /* E_DPCH_Information_RL_AdditionReqFDD */
static int hf_nbap_RadioLinkAdditionRequestTDD_PDU = -1;  /* RadioLinkAdditionRequestTDD */
static int hf_nbap_UL_CCTrCH_InformationList_RL_AdditionRqstTDD_PDU = -1;  /* UL_CCTrCH_InformationList_RL_AdditionRqstTDD */
static int hf_nbap_UL_DPCH_InformationItem_RL_AdditionRqstTDD_PDU = -1;  /* UL_DPCH_InformationItem_RL_AdditionRqstTDD */
static int hf_nbap_UL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD_PDU = -1;  /* UL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD */
static int hf_nbap_UL_DPCH_InformationItem_768_RL_AdditionRqstTDD_PDU = -1;  /* UL_DPCH_InformationItem_768_RL_AdditionRqstTDD */
static int hf_nbap_DL_CCTrCH_InformationList_RL_AdditionRqstTDD_PDU = -1;  /* DL_CCTrCH_InformationList_RL_AdditionRqstTDD */
static int hf_nbap_DL_DPCH_InformationItem_RL_AdditionRqstTDD_PDU = -1;  /* DL_DPCH_InformationItem_RL_AdditionRqstTDD */
static int hf_nbap_DL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD_PDU = -1;  /* DL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD */
static int hf_nbap_DL_DPCH_InformationItem_768_RL_AdditionRqstTDD_PDU = -1;  /* DL_DPCH_InformationItem_768_RL_AdditionRqstTDD */
static int hf_nbap_RL_Information_RL_AdditionRqstTDD_PDU = -1;  /* RL_Information_RL_AdditionRqstTDD */
static int hf_nbap_RadioLinkAdditionResponseFDD_PDU = -1;  /* RadioLinkAdditionResponseFDD */
static int hf_nbap_Additional_HS_Cell_Change_Information_Response_List_PDU = -1;  /* Additional_HS_Cell_Change_Information_Response_List */
static int hf_nbap_RL_InformationResponseList_RL_AdditionRspFDD_PDU = -1;  /* RL_InformationResponseList_RL_AdditionRspFDD */
static int hf_nbap_RL_InformationResponseItem_RL_AdditionRspFDD_PDU = -1;  /* RL_InformationResponseItem_RL_AdditionRspFDD */
static int hf_nbap_RadioLinkAdditionResponseTDD_PDU = -1;  /* RadioLinkAdditionResponseTDD */
static int hf_nbap_RL_InformationResponse_RL_AdditionRspTDD_PDU = -1;  /* RL_InformationResponse_RL_AdditionRspTDD */
static int hf_nbap_RL_InformationResponse_LCR_RL_AdditionRspTDD_PDU = -1;  /* RL_InformationResponse_LCR_RL_AdditionRspTDD */
static int hf_nbap_RadioLinkAdditionFailureFDD_PDU = -1;  /* RadioLinkAdditionFailureFDD */
static int hf_nbap_CauseLevel_RL_AdditionFailureFDD_PDU = -1;  /* CauseLevel_RL_AdditionFailureFDD */
static int hf_nbap_Unsuccessful_RL_InformationRespItem_RL_AdditionFailureFDD_PDU = -1;  /* Unsuccessful_RL_InformationRespItem_RL_AdditionFailureFDD */
static int hf_nbap_Successful_RL_InformationRespItem_RL_AdditionFailureFDD_PDU = -1;  /* Successful_RL_InformationRespItem_RL_AdditionFailureFDD */
static int hf_nbap_RadioLinkAdditionFailureTDD_PDU = -1;  /* RadioLinkAdditionFailureTDD */
static int hf_nbap_CauseLevel_RL_AdditionFailureTDD_PDU = -1;  /* CauseLevel_RL_AdditionFailureTDD */
static int hf_nbap_Unsuccessful_RL_InformationResp_RL_AdditionFailureTDD_PDU = -1;  /* Unsuccessful_RL_InformationResp_RL_AdditionFailureTDD */
static int hf_nbap_RadioLinkReconfigurationPrepareFDD_PDU = -1;  /* RadioLinkReconfigurationPrepareFDD */
static int hf_nbap_Additional_HS_Cell_Information_RL_Reconf_Prep_PDU = -1;  /* Additional_HS_Cell_Information_RL_Reconf_Prep */
static int hf_nbap_Additional_EDCH_Cell_Information_RL_Reconf_Prep_PDU = -1;  /* Additional_EDCH_Cell_Information_RL_Reconf_Prep */
static int hf_nbap_UL_DPCH_Information_RL_ReconfPrepFDD_PDU = -1;  /* UL_DPCH_Information_RL_ReconfPrepFDD */
static int hf_nbap_DL_DPCH_Information_RL_ReconfPrepFDD_PDU = -1;  /* DL_DPCH_Information_RL_ReconfPrepFDD */
static int hf_nbap_DL_DPCH_Power_Information_RL_ReconfPrepFDD_PDU = -1;  /* DL_DPCH_Power_Information_RL_ReconfPrepFDD */
static int hf_nbap_DCH_DeleteList_RL_ReconfPrepFDD_PDU = -1;  /* DCH_DeleteList_RL_ReconfPrepFDD */
static int hf_nbap_RL_InformationList_RL_ReconfPrepFDD_PDU = -1;  /* RL_InformationList_RL_ReconfPrepFDD */
static int hf_nbap_RL_InformationItem_RL_ReconfPrepFDD_PDU = -1;  /* RL_InformationItem_RL_ReconfPrepFDD */
static int hf_nbap_E_DPCH_Information_RL_ReconfPrepFDD_PDU = -1;  /* E_DPCH_Information_RL_ReconfPrepFDD */
static int hf_nbap_F_DPCH_Information_RL_ReconfPrepFDD_PDU = -1;  /* F_DPCH_Information_RL_ReconfPrepFDD */
static int hf_nbap_RadioLinkReconfigurationPrepareTDD_PDU = -1;  /* RadioLinkReconfigurationPrepareTDD */
static int hf_nbap_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_PDU = -1;  /* UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD */
static int hf_nbap_UL_DPCH_InformationAddItem_RL_ReconfPrepTDD_PDU = -1;  /* UL_DPCH_InformationAddItem_RL_ReconfPrepTDD */
static int hf_nbap_UL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD_PDU = -1;  /* UL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD */
static int hf_nbap_MultipleRL_UL_DPCH_InformationAddList_RL_ReconfPrepTDD_PDU = -1;  /* MultipleRL_UL_DPCH_InformationAddList_RL_ReconfPrepTDD */
static int hf_nbap_UL_DPCH_768_InformationAddList_RL_ReconfPrepTDD_PDU = -1;  /* UL_DPCH_768_InformationAddList_RL_ReconfPrepTDD */
static int hf_nbap_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_PDU = -1;  /* UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD */
static int hf_nbap_UL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD_PDU = -1;  /* UL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD */
static int hf_nbap_UL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD_PDU = -1;  /* UL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD */
static int hf_nbap_UL_TimeslotLCR_InformationModify_ModifyList_RL_ReconfPrepTDD_PDU = -1;  /* UL_TimeslotLCR_InformationModify_ModifyList_RL_ReconfPrepTDD */
static int hf_nbap_UL_Timeslot768_InformationModify_ModifyList_RL_ReconfPrepTDD_PDU = -1;  /* UL_Timeslot768_InformationModify_ModifyList_RL_ReconfPrepTDD */
static int hf_nbap_UL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD_PDU = -1;  /* UL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD */
static int hf_nbap_UL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD_PDU = -1;  /* UL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD */
static int hf_nbap_MultipleRL_UL_DPCH_InformationModifyList_RL_ReconfPrepTDD_PDU = -1;  /* MultipleRL_UL_DPCH_InformationModifyList_RL_ReconfPrepTDD */
static int hf_nbap_UL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD_PDU = -1;  /* UL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD */
static int hf_nbap_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_PDU = -1;  /* UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD */
static int hf_nbap_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_PDU = -1;  /* DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD */
static int hf_nbap_DL_DPCH_InformationAddItem_RL_ReconfPrepTDD_PDU = -1;  /* DL_DPCH_InformationAddItem_RL_ReconfPrepTDD */
static int hf_nbap_DL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD_PDU = -1;  /* DL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD */
static int hf_nbap_MultipleRL_DL_DPCH_InformationAddList_RL_ReconfPrepTDD_PDU = -1;  /* MultipleRL_DL_DPCH_InformationAddList_RL_ReconfPrepTDD */
static int hf_nbap_DL_DPCH_768_InformationAddList_RL_ReconfPrepTDD_PDU = -1;  /* DL_DPCH_768_InformationAddList_RL_ReconfPrepTDD */
static int hf_nbap_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_PDU = -1;  /* DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD */
static int hf_nbap_DL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD_PDU = -1;  /* DL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD */
static int hf_nbap_DL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD_PDU = -1;  /* DL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD */
static int hf_nbap_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD_PDU = -1;  /* DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD */
static int hf_nbap_DL_Timeslot_768_InformationModify_ModifyList_RL_ReconfPrepTDD_PDU = -1;  /* DL_Timeslot_768_InformationModify_ModifyList_RL_ReconfPrepTDD */
static int hf_nbap_DL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD_PDU = -1;  /* DL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD */
static int hf_nbap_DL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD_PDU = -1;  /* DL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD */
static int hf_nbap_MultipleRL_DL_DPCH_InformationModifyList_RL_ReconfPrepTDD_PDU = -1;  /* MultipleRL_DL_DPCH_InformationModifyList_RL_ReconfPrepTDD */
static int hf_nbap_DL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD_PDU = -1;  /* DL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD */
static int hf_nbap_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_PDU = -1;  /* DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD */
static int hf_nbap_DCH_DeleteList_RL_ReconfPrepTDD_PDU = -1;  /* DCH_DeleteList_RL_ReconfPrepTDD */
static int hf_nbap_DSCH_Information_ModifyList_RL_ReconfPrepTDD_PDU = -1;  /* DSCH_Information_ModifyList_RL_ReconfPrepTDD */
static int hf_nbap_DSCH_Information_DeleteList_RL_ReconfPrepTDD_PDU = -1;  /* DSCH_Information_DeleteList_RL_ReconfPrepTDD */
static int hf_nbap_USCH_Information_ModifyList_RL_ReconfPrepTDD_PDU = -1;  /* USCH_Information_ModifyList_RL_ReconfPrepTDD */
static int hf_nbap_USCH_Information_DeleteList_RL_ReconfPrepTDD_PDU = -1;  /* USCH_Information_DeleteList_RL_ReconfPrepTDD */
static int hf_nbap_MultipleRL_Information_RL_ReconfPrepTDD_PDU = -1;  /* MultipleRL_Information_RL_ReconfPrepTDD */
static int hf_nbap_RL_Information_RL_ReconfPrepTDD_PDU = -1;  /* RL_Information_RL_ReconfPrepTDD */
static int hf_nbap_RadioLinkReconfigurationReady_PDU = -1;  /* RadioLinkReconfigurationReady */
static int hf_nbap_RL_InformationResponseList_RL_ReconfReady_PDU = -1;  /* RL_InformationResponseList_RL_ReconfReady */
static int hf_nbap_RL_InformationResponseItem_RL_ReconfReady_PDU = -1;  /* RL_InformationResponseItem_RL_ReconfReady */
static int hf_nbap_RadioLinkReconfigurationFailure_PDU = -1;  /* RadioLinkReconfigurationFailure */
static int hf_nbap_CauseLevel_RL_ReconfFailure_PDU = -1;  /* CauseLevel_RL_ReconfFailure */
static int hf_nbap_RL_ReconfigurationFailureItem_RL_ReconfFailure_PDU = -1;  /* RL_ReconfigurationFailureItem_RL_ReconfFailure */
static int hf_nbap_RadioLinkReconfigurationCommit_PDU = -1;  /* RadioLinkReconfigurationCommit */
static int hf_nbap_RadioLinkReconfigurationCancel_PDU = -1;  /* RadioLinkReconfigurationCancel */
static int hf_nbap_RadioLinkReconfigurationRequestFDD_PDU = -1;  /* RadioLinkReconfigurationRequestFDD */
static int hf_nbap_Additional_HS_Cell_Information_RL_Reconf_Req_PDU = -1;  /* Additional_HS_Cell_Information_RL_Reconf_Req */
static int hf_nbap_Additional_EDCH_Cell_Information_RL_Reconf_Req_PDU = -1;  /* Additional_EDCH_Cell_Information_RL_Reconf_Req */
static int hf_nbap_UL_DPCH_Information_RL_ReconfRqstFDD_PDU = -1;  /* UL_DPCH_Information_RL_ReconfRqstFDD */
static int hf_nbap_DL_DPCH_Information_RL_ReconfRqstFDD_PDU = -1;  /* DL_DPCH_Information_RL_ReconfRqstFDD */
static int hf_nbap_DCH_DeleteList_RL_ReconfRqstFDD_PDU = -1;  /* DCH_DeleteList_RL_ReconfRqstFDD */
static int hf_nbap_RL_InformationList_RL_ReconfRqstFDD_PDU = -1;  /* RL_InformationList_RL_ReconfRqstFDD */
static int hf_nbap_RL_InformationItem_RL_ReconfRqstFDD_PDU = -1;  /* RL_InformationItem_RL_ReconfRqstFDD */
static int hf_nbap_E_DPCH_Information_RL_ReconfRqstFDD_PDU = -1;  /* E_DPCH_Information_RL_ReconfRqstFDD */
static int hf_nbap_RadioLinkReconfigurationRequestTDD_PDU = -1;  /* RadioLinkReconfigurationRequestTDD */
static int hf_nbap_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU = -1;  /* UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD */
static int hf_nbap_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_PDU = -1;  /* UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD */
static int hf_nbap_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_PDU = -1;  /* UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD */
static int hf_nbap_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_PDU = -1;  /* UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD */
static int hf_nbap_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU = -1;  /* DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD */
static int hf_nbap_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_PDU = -1;  /* DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD */
static int hf_nbap_MultipleRL_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU = -1;  /* MultipleRL_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD */
static int hf_nbap_DL_DPCH_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD_PDU = -1;  /* DL_DPCH_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD */
static int hf_nbap_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_PDU = -1;  /* DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD */
static int hf_nbap_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_PDU = -1;  /* DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD */
static int hf_nbap_DCH_DeleteList_RL_ReconfRqstTDD_PDU = -1;  /* DCH_DeleteList_RL_ReconfRqstTDD */
static int hf_nbap_Multiple_RL_Information_RL_ReconfRqstTDD_PDU = -1;  /* Multiple_RL_Information_RL_ReconfRqstTDD */
static int hf_nbap_RL_Information_RL_ReconfRqstTDD_PDU = -1;  /* RL_Information_RL_ReconfRqstTDD */
static int hf_nbap_RadioLinkReconfigurationResponse_PDU = -1;  /* RadioLinkReconfigurationResponse */
static int hf_nbap_RL_InformationResponseList_RL_ReconfRsp_PDU = -1;  /* RL_InformationResponseList_RL_ReconfRsp */
static int hf_nbap_RL_InformationResponseItem_RL_ReconfRsp_PDU = -1;  /* RL_InformationResponseItem_RL_ReconfRsp */
static int hf_nbap_RadioLinkDeletionRequest_PDU = -1;  /* RadioLinkDeletionRequest */
static int hf_nbap_RL_informationList_RL_DeletionRqst_PDU = -1;  /* RL_informationList_RL_DeletionRqst */
static int hf_nbap_RL_informationItem_RL_DeletionRqst_PDU = -1;  /* RL_informationItem_RL_DeletionRqst */
static int hf_nbap_RadioLinkDeletionResponse_PDU = -1;  /* RadioLinkDeletionResponse */
static int hf_nbap_DL_PowerControlRequest_PDU = -1;  /* DL_PowerControlRequest */
static int hf_nbap_DL_ReferencePowerInformationList_DL_PC_Rqst_PDU = -1;  /* DL_ReferencePowerInformationList_DL_PC_Rqst */
static int hf_nbap_DL_ReferencePowerInformationItem_DL_PC_Rqst_PDU = -1;  /* DL_ReferencePowerInformationItem_DL_PC_Rqst */
static int hf_nbap_DL_PowerTimeslotControlRequest_PDU = -1;  /* DL_PowerTimeslotControlRequest */
static int hf_nbap_DedicatedMeasurementInitiationRequest_PDU = -1;  /* DedicatedMeasurementInitiationRequest */
static int hf_nbap_DedicatedMeasurementObjectType_DM_Rqst_PDU = -1;  /* DedicatedMeasurementObjectType_DM_Rqst */
static int hf_nbap_RL_InformationItem_DM_Rqst_PDU = -1;  /* RL_InformationItem_DM_Rqst */
static int hf_nbap_PUSCH_Info_DM_Rqst_PDU = -1;   /* PUSCH_Info_DM_Rqst */
static int hf_nbap_HSSICH_Info_DM_Rqst_PDU = -1;  /* HSSICH_Info_DM_Rqst */
static int hf_nbap_HSSICH_InfoExt_DM_Rqst_PDU = -1;  /* HSSICH_InfoExt_DM_Rqst */
static int hf_nbap_DedicatedMeasurementInitiationResponse_PDU = -1;  /* DedicatedMeasurementInitiationResponse */
static int hf_nbap_DedicatedMeasurementObjectType_DM_Rsp_PDU = -1;  /* DedicatedMeasurementObjectType_DM_Rsp */
static int hf_nbap_RL_InformationItem_DM_Rsp_PDU = -1;  /* RL_InformationItem_DM_Rsp */
static int hf_nbap_PUSCH_Info_DM_Rsp_PDU = -1;    /* PUSCH_Info_DM_Rsp */
static int hf_nbap_Multiple_PUSCH_InfoList_DM_Rsp_PDU = -1;  /* Multiple_PUSCH_InfoList_DM_Rsp */
static int hf_nbap_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp_PDU = -1;  /* Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp */
static int hf_nbap_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp_PDU = -1;  /* Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp */
static int hf_nbap_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp_PDU = -1;  /* Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp */
static int hf_nbap_Multiple_DedicatedMeasurementValueList_768_TDD_DM_Rsp_PDU = -1;  /* Multiple_DedicatedMeasurementValueList_768_TDD_DM_Rsp */
static int hf_nbap_RL_Set_InformationItem_DM_Rsp_PDU = -1;  /* RL_Set_InformationItem_DM_Rsp */
static int hf_nbap_DedicatedMeasurementInitiationFailure_PDU = -1;  /* DedicatedMeasurementInitiationFailure */
static int hf_nbap_DedicatedMeasurementReport_PDU = -1;  /* DedicatedMeasurementReport */
static int hf_nbap_DedicatedMeasurementObjectType_DM_Rprt_PDU = -1;  /* DedicatedMeasurementObjectType_DM_Rprt */
static int hf_nbap_RL_InformationItem_DM_Rprt_PDU = -1;  /* RL_InformationItem_DM_Rprt */
static int hf_nbap_PUSCH_Info_DM_Rprt_PDU = -1;   /* PUSCH_Info_DM_Rprt */
static int hf_nbap_Multiple_PUSCH_InfoList_DM_Rprt_PDU = -1;  /* Multiple_PUSCH_InfoList_DM_Rprt */
static int hf_nbap_RL_Set_InformationItem_DM_Rprt_PDU = -1;  /* RL_Set_InformationItem_DM_Rprt */
static int hf_nbap_DedicatedMeasurementTerminationRequest_PDU = -1;  /* DedicatedMeasurementTerminationRequest */
static int hf_nbap_DedicatedMeasurementFailureIndication_PDU = -1;  /* DedicatedMeasurementFailureIndication */
static int hf_nbap_RadioLinkFailureIndication_PDU = -1;  /* RadioLinkFailureIndication */
static int hf_nbap_Reporting_Object_RL_FailureInd_PDU = -1;  /* Reporting_Object_RL_FailureInd */
static int hf_nbap_RL_InformationItem_RL_FailureInd_PDU = -1;  /* RL_InformationItem_RL_FailureInd */
static int hf_nbap_RL_Set_InformationItem_RL_FailureInd_PDU = -1;  /* RL_Set_InformationItem_RL_FailureInd */
static int hf_nbap_CCTrCH_InformationItem_RL_FailureInd_PDU = -1;  /* CCTrCH_InformationItem_RL_FailureInd */
static int hf_nbap_RadioLinkPreemptionRequiredIndication_PDU = -1;  /* RadioLinkPreemptionRequiredIndication */
static int hf_nbap_RL_InformationList_RL_PreemptRequiredInd_PDU = -1;  /* RL_InformationList_RL_PreemptRequiredInd */
static int hf_nbap_RL_InformationItem_RL_PreemptRequiredInd_PDU = -1;  /* RL_InformationItem_RL_PreemptRequiredInd */
static int hf_nbap_RadioLinkRestoreIndication_PDU = -1;  /* RadioLinkRestoreIndication */
static int hf_nbap_Reporting_Object_RL_RestoreInd_PDU = -1;  /* Reporting_Object_RL_RestoreInd */
static int hf_nbap_RL_InformationItem_RL_RestoreInd_PDU = -1;  /* RL_InformationItem_RL_RestoreInd */
static int hf_nbap_RL_Set_InformationItem_RL_RestoreInd_PDU = -1;  /* RL_Set_InformationItem_RL_RestoreInd */
static int hf_nbap_CCTrCH_InformationItem_RL_RestoreInd_PDU = -1;  /* CCTrCH_InformationItem_RL_RestoreInd */
static int hf_nbap_CompressedModeCommand_PDU = -1;  /* CompressedModeCommand */
static int hf_nbap_ErrorIndication_PDU = -1;      /* ErrorIndication */
static int hf_nbap_PrivateMessage_PDU = -1;       /* PrivateMessage */
static int hf_nbap_PhysicalSharedChannelReconfigurationRequestFDD_PDU = -1;  /* PhysicalSharedChannelReconfigurationRequestFDD */
static int hf_nbap_HSDPA_And_EDCH_CellPortion_InformationList_PSCH_ReconfRqst_PDU = -1;  /* HSDPA_And_EDCH_CellPortion_InformationList_PSCH_ReconfRqst */
static int hf_nbap_PhysicalSharedChannelReconfigurationRequestTDD_PDU = -1;  /* PhysicalSharedChannelReconfigurationRequestTDD */
static int hf_nbap_PDSCHSets_AddList_PSCH_ReconfRqst_PDU = -1;  /* PDSCHSets_AddList_PSCH_ReconfRqst */
static int hf_nbap_PDSCH_Information_AddItem_PSCH_ReconfRqst_PDU = -1;  /* PDSCH_Information_AddItem_PSCH_ReconfRqst */
static int hf_nbap_PDSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst_PDU = -1;  /* PDSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst */
static int hf_nbap_PDSCH_AddInformation_768_AddItem_PSCH_ReconfRqst_PDU = -1;  /* PDSCH_AddInformation_768_AddItem_PSCH_ReconfRqst */
static int hf_nbap_PDSCHSets_ModifyList_PSCH_ReconfRqst_PDU = -1;  /* PDSCHSets_ModifyList_PSCH_ReconfRqst */
static int hf_nbap_PDSCH_Information_ModifyItem_PSCH_ReconfRqst_PDU = -1;  /* PDSCH_Information_ModifyItem_PSCH_ReconfRqst */
static int hf_nbap_PDSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst_PDU = -1;  /* PDSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst */
static int hf_nbap_PDSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst_PDU = -1;  /* PDSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst */
static int hf_nbap_PDSCHSets_DeleteList_PSCH_ReconfRqst_PDU = -1;  /* PDSCHSets_DeleteList_PSCH_ReconfRqst */
static int hf_nbap_PUSCHSets_AddList_PSCH_ReconfRqst_PDU = -1;  /* PUSCHSets_AddList_PSCH_ReconfRqst */
static int hf_nbap_PUSCH_Information_AddItem_PSCH_ReconfRqst_PDU = -1;  /* PUSCH_Information_AddItem_PSCH_ReconfRqst */
static int hf_nbap_PUSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst_PDU = -1;  /* PUSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst */
static int hf_nbap_PUSCH_AddInformation_768_AddItem_PSCH_ReconfRqst_PDU = -1;  /* PUSCH_AddInformation_768_AddItem_PSCH_ReconfRqst */
static int hf_nbap_PUSCHSets_ModifyList_PSCH_ReconfRqst_PDU = -1;  /* PUSCHSets_ModifyList_PSCH_ReconfRqst */
static int hf_nbap_PUSCH_Information_ModifyItem_PSCH_ReconfRqst_PDU = -1;  /* PUSCH_Information_ModifyItem_PSCH_ReconfRqst */
static int hf_nbap_PUSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst_PDU = -1;  /* PUSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst */
static int hf_nbap_PUSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst_PDU = -1;  /* PUSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst */
static int hf_nbap_PUSCHSets_DeleteList_PSCH_ReconfRqst_PDU = -1;  /* PUSCHSets_DeleteList_PSCH_ReconfRqst */
static int hf_nbap_HS_PDSCH_TDD_Information_PSCH_ReconfRqst_PDU = -1;  /* HS_PDSCH_TDD_Information_PSCH_ReconfRqst */
static int hf_nbap_DL_HS_PDSCH_Timeslot_Information_768_PSCH_ReconfRqst_PDU = -1;  /* DL_HS_PDSCH_Timeslot_Information_768_PSCH_ReconfRqst */
static int hf_nbap_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst_PDU = -1;  /* MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst */
static int hf_nbap_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst_PDU = -1;  /* MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst */
static int hf_nbap_Add_To_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_PDU = -1;  /* Add_To_HS_SCCH_Resource_Pool_PSCH_ReconfRqst */
static int hf_nbap_HS_SCCH_Information_768_PSCH_ReconfRqst_PDU = -1;  /* HS_SCCH_Information_768_PSCH_ReconfRqst */
static int hf_nbap_HS_SCCH_InformationExt_LCR_PSCH_ReconfRqst_PDU = -1;  /* HS_SCCH_InformationExt_LCR_PSCH_ReconfRqst */
static int hf_nbap_Modify_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_PDU = -1;  /* Modify_HS_SCCH_Resource_Pool_PSCH_ReconfRqst */
static int hf_nbap_HS_SCCH_InformationModifyExt_LCR_PSCH_ReconfRqst_PDU = -1;  /* HS_SCCH_InformationModifyExt_LCR_PSCH_ReconfRqst */
static int hf_nbap_HS_SCCH_InformationModify_768_PSCH_ReconfRqst_PDU = -1;  /* HS_SCCH_InformationModify_768_PSCH_ReconfRqst */
static int hf_nbap_Delete_From_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_PDU = -1;  /* Delete_From_HS_SCCH_Resource_Pool_PSCH_ReconfRqst */
static int hf_nbap_E_PUCH_Information_PSCH_ReconfRqst_PDU = -1;  /* E_PUCH_Information_PSCH_ReconfRqst */
static int hf_nbap_Add_To_E_AGCH_Resource_Pool_PSCH_ReconfRqst_PDU = -1;  /* Add_To_E_AGCH_Resource_Pool_PSCH_ReconfRqst */
static int hf_nbap_Modify_E_AGCH_Resource_Pool_PSCH_ReconfRqst_PDU = -1;  /* Modify_E_AGCH_Resource_Pool_PSCH_ReconfRqst */
static int hf_nbap_Delete_From_E_AGCH_Resource_Pool_PSCH_ReconfRqst_PDU = -1;  /* Delete_From_E_AGCH_Resource_Pool_PSCH_ReconfRqst */
static int hf_nbap_E_HICH_Information_PSCH_ReconfRqst_PDU = -1;  /* E_HICH_Information_PSCH_ReconfRqst */
static int hf_nbap_E_PUCH_Information_768_PSCH_ReconfRqst_PDU = -1;  /* E_PUCH_Information_768_PSCH_ReconfRqst */
static int hf_nbap_Add_To_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst_PDU = -1;  /* Add_To_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst */
static int hf_nbap_Modify_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst_PDU = -1;  /* Modify_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst */
static int hf_nbap_E_HICH_Information_768_PSCH_ReconfRqst_PDU = -1;  /* E_HICH_Information_768_PSCH_ReconfRqst */
static int hf_nbap_E_PUCH_Information_LCR_PSCH_ReconfRqst_PDU = -1;  /* E_PUCH_Information_LCR_PSCH_ReconfRqst */
static int hf_nbap_Add_To_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU = -1;  /* Add_To_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst */
static int hf_nbap_Modify_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU = -1;  /* Modify_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst */
static int hf_nbap_Add_To_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU = -1;  /* Add_To_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst */
static int hf_nbap_Modify_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU = -1;  /* Modify_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst */
static int hf_nbap_Delete_From_E_HICH_Resource_Pool_PSCH_ReconfRqst_PDU = -1;  /* Delete_From_E_HICH_Resource_Pool_PSCH_ReconfRqst */
static int hf_nbap_SYNC_UL_Partition_LCR_PDU = -1;  /* SYNC_UL_Partition_LCR */
static int hf_nbap_Delete_From_HS_SCCH_Resource_PoolExt_PSCH_ReconfRqst_PDU = -1;  /* Delete_From_HS_SCCH_Resource_PoolExt_PSCH_ReconfRqst */
static int hf_nbap_MultipleFreq_E_PUCH_Timeslot_InformationList_LCR_PSCH_ReconfRqst_PDU = -1;  /* MultipleFreq_E_PUCH_Timeslot_InformationList_LCR_PSCH_ReconfRqst */
static int hf_nbap_MultipleFreq_E_PUCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst_PDU = -1;  /* MultipleFreq_E_PUCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst */
static int hf_nbap_Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst_PDU = -1;  /* Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst */
static int hf_nbap_PhysicalSharedChannelReconfigurationResponse_PDU = -1;  /* PhysicalSharedChannelReconfigurationResponse */
static int hf_nbap_E_HICH_TimeOffset_ExtensionLCR_PDU = -1;  /* E_HICH_TimeOffset_ExtensionLCR */
static int hf_nbap_MultipleFreq_E_HICH_TimeOffsetLCR_PDU = -1;  /* MultipleFreq_E_HICH_TimeOffsetLCR */
static int hf_nbap_PhysicalSharedChannelReconfigurationFailure_PDU = -1;  /* PhysicalSharedChannelReconfigurationFailure */
static int hf_nbap_CauseLevel_PSCH_ReconfFailure_PDU = -1;  /* CauseLevel_PSCH_ReconfFailure */
static int hf_nbap_Unsuccessful_PDSCHSetItem_PSCH_ReconfFailureTDD_PDU = -1;  /* Unsuccessful_PDSCHSetItem_PSCH_ReconfFailureTDD */
static int hf_nbap_Unsuccessful_PUSCHSetItem_PSCH_ReconfFailureTDD_PDU = -1;  /* Unsuccessful_PUSCHSetItem_PSCH_ReconfFailureTDD */
static int hf_nbap_UARFCNSpecificCauseList_PSCH_ReconfFailureTDD_PDU = -1;  /* UARFCNSpecificCauseList_PSCH_ReconfFailureTDD */
static int hf_nbap_Unsuccessful_UARFCNItem_PSCH_ReconfFailureTDD_PDU = -1;  /* Unsuccessful_UARFCNItem_PSCH_ReconfFailureTDD */
static int hf_nbap_E_HICH_TimeOffset_ReconfFailureTDD_PDU = -1;  /* E_HICH_TimeOffset_ReconfFailureTDD */
static int hf_nbap_Common_System_Information_ResponseLCR_PDU = -1;  /* Common_System_Information_ResponseLCR */
static int hf_nbap_ResetRequest_PDU = -1;         /* ResetRequest */
static int hf_nbap_ResetIndicator_PDU = -1;       /* ResetIndicator */
static int hf_nbap_CommunicationContextInfoItem_Reset_PDU = -1;  /* CommunicationContextInfoItem_Reset */
static int hf_nbap_CommunicationControlPortInfoItem_Reset_PDU = -1;  /* CommunicationControlPortInfoItem_Reset */
static int hf_nbap_ResetResponse_PDU = -1;        /* ResetResponse */
static int hf_nbap_InformationExchangeInitiationRequest_PDU = -1;  /* InformationExchangeInitiationRequest */
static int hf_nbap_InformationExchangeObjectType_InfEx_Rqst_PDU = -1;  /* InformationExchangeObjectType_InfEx_Rqst */
static int hf_nbap_InformationExchangeInitiationResponse_PDU = -1;  /* InformationExchangeInitiationResponse */
static int hf_nbap_InformationExchangeObjectType_InfEx_Rsp_PDU = -1;  /* InformationExchangeObjectType_InfEx_Rsp */
static int hf_nbap_InformationExchangeInitiationFailure_PDU = -1;  /* InformationExchangeInitiationFailure */
static int hf_nbap_InformationReport_PDU = -1;    /* InformationReport */
static int hf_nbap_InformationExchangeObjectType_InfEx_Rprt_PDU = -1;  /* InformationExchangeObjectType_InfEx_Rprt */
static int hf_nbap_InformationExchangeTerminationRequest_PDU = -1;  /* InformationExchangeTerminationRequest */
static int hf_nbap_InformationExchangeFailureIndication_PDU = -1;  /* InformationExchangeFailureIndication */
static int hf_nbap_CellSynchronisationInitiationRequestTDD_PDU = -1;  /* CellSynchronisationInitiationRequestTDD */
static int hf_nbap_TimeslotInfo_CellSyncInitiationRqstTDD_PDU = -1;  /* TimeslotInfo_CellSyncInitiationRqstTDD */
static int hf_nbap_CellSyncBurstTransInit_CellSyncInitiationRqstTDD_PDU = -1;  /* CellSyncBurstTransInit_CellSyncInitiationRqstTDD */
static int hf_nbap_CellSyncBurstMeasureInit_CellSyncInitiationRqstTDD_PDU = -1;  /* CellSyncBurstMeasureInit_CellSyncInitiationRqstTDD */
static int hf_nbap_SYNCDlCodeId_TransInitLCR_CellSyncInitiationRqstTDD_PDU = -1;  /* SYNCDlCodeId_TransInitLCR_CellSyncInitiationRqstTDD */
static int hf_nbap_SYNCDlCodeId_MeasureInitLCR_CellSyncInitiationRqstTDD_PDU = -1;  /* SYNCDlCodeId_MeasureInitLCR_CellSyncInitiationRqstTDD */
static int hf_nbap_CellSynchronisationInitiationResponseTDD_PDU = -1;  /* CellSynchronisationInitiationResponseTDD */
static int hf_nbap_CellSynchronisationInitiationFailureTDD_PDU = -1;  /* CellSynchronisationInitiationFailureTDD */
static int hf_nbap_CellSynchronisationReconfigurationRequestTDD_PDU = -1;  /* CellSynchronisationReconfigurationRequestTDD */
static int hf_nbap_CellSyncBurstTransReconfInfo_CellSyncReconfRqstTDD_PDU = -1;  /* CellSyncBurstTransReconfInfo_CellSyncReconfRqstTDD */
static int hf_nbap_CellSyncBurstMeasInfo_CellSyncReconfRqstTDD_PDU = -1;  /* CellSyncBurstMeasInfo_CellSyncReconfRqstTDD */
static int hf_nbap_CellSyncBurstMeasInfoListIE_CellSyncReconfRqstTDD_PDU = -1;  /* CellSyncBurstMeasInfoListIE_CellSyncReconfRqstTDD */
static int hf_nbap_SYNCDlCodeIdTransReconfInfoLCR_CellSyncReconfRqstTDD_PDU = -1;  /* SYNCDlCodeIdTransReconfInfoLCR_CellSyncReconfRqstTDD */
static int hf_nbap_SYNCDlCodeIdMeasInfoLCR_CellSyncReconfRqstTDD_PDU = -1;  /* SYNCDlCodeIdMeasInfoLCR_CellSyncReconfRqstTDD */
static int hf_nbap_CellSynchronisationReconfigurationResponseTDD_PDU = -1;  /* CellSynchronisationReconfigurationResponseTDD */
static int hf_nbap_CellSynchronisationReconfigurationFailureTDD_PDU = -1;  /* CellSynchronisationReconfigurationFailureTDD */
static int hf_nbap_CellSynchronisationAdjustmentRequestTDD_PDU = -1;  /* CellSynchronisationAdjustmentRequestTDD */
static int hf_nbap_CellAdjustmentInfo_SyncAdjustmentRqstTDD_PDU = -1;  /* CellAdjustmentInfo_SyncAdjustmentRqstTDD */
static int hf_nbap_CellAdjustmentInfoItem_SyncAdjustmentRqstTDD_PDU = -1;  /* CellAdjustmentInfoItem_SyncAdjustmentRqstTDD */
static int hf_nbap_CellSynchronisationAdjustmentResponseTDD_PDU = -1;  /* CellSynchronisationAdjustmentResponseTDD */
static int hf_nbap_CellSynchronisationAdjustmentFailureTDD_PDU = -1;  /* CellSynchronisationAdjustmentFailureTDD */
static int hf_nbap_CauseLevel_SyncAdjustmntFailureTDD_PDU = -1;  /* CauseLevel_SyncAdjustmntFailureTDD */
static int hf_nbap_Unsuccessful_cell_InformationRespItem_SyncAdjustmntFailureTDD_PDU = -1;  /* Unsuccessful_cell_InformationRespItem_SyncAdjustmntFailureTDD */
static int hf_nbap_CellSynchronisationTerminationRequestTDD_PDU = -1;  /* CellSynchronisationTerminationRequestTDD */
static int hf_nbap_CellSynchronisationFailureIndicationTDD_PDU = -1;  /* CellSynchronisationFailureIndicationTDD */
static int hf_nbap_CellSynchronisationReportTDD_PDU = -1;  /* CellSynchronisationReportTDD */
static int hf_nbap_CellSyncInfo_CellSyncReprtTDD_PDU = -1;  /* CellSyncInfo_CellSyncReprtTDD */
static int hf_nbap_SyncReportType_CellSyncReprtTDD_PDU = -1;  /* SyncReportType_CellSyncReprtTDD */
static int hf_nbap_SyncDLCodeIdsMeasInfoList_CellSyncReprtTDD_PDU = -1;  /* SyncDLCodeIdsMeasInfoList_CellSyncReprtTDD */
static int hf_nbap_BearerRearrangementIndication_PDU = -1;  /* BearerRearrangementIndication */
static int hf_nbap_DCH_RearrangeList_Bearer_RearrangeInd_PDU = -1;  /* DCH_RearrangeList_Bearer_RearrangeInd */
static int hf_nbap_DSCH_RearrangeList_Bearer_RearrangeInd_PDU = -1;  /* DSCH_RearrangeList_Bearer_RearrangeInd */
static int hf_nbap_USCH_RearrangeList_Bearer_RearrangeInd_PDU = -1;  /* USCH_RearrangeList_Bearer_RearrangeInd */
static int hf_nbap_HSDSCH_RearrangeList_Bearer_RearrangeInd_PDU = -1;  /* HSDSCH_RearrangeList_Bearer_RearrangeInd */
static int hf_nbap_E_DCH_RearrangeList_Bearer_RearrangeInd_PDU = -1;  /* E_DCH_RearrangeList_Bearer_RearrangeInd */
static int hf_nbap_Additional_EDCH_Cell_Information_Bearer_Rearrangement_List_PDU = -1;  /* Additional_EDCH_Cell_Information_Bearer_Rearrangement_List */
static int hf_nbap_RadioLinkActivationCommandFDD_PDU = -1;  /* RadioLinkActivationCommandFDD */
static int hf_nbap_DelayedActivationInformationList_RL_ActivationCmdFDD_PDU = -1;  /* DelayedActivationInformationList_RL_ActivationCmdFDD */
static int hf_nbap_DelayedActivationInformation_RL_ActivationCmdFDD_PDU = -1;  /* DelayedActivationInformation_RL_ActivationCmdFDD */
static int hf_nbap_RadioLinkActivationCommandTDD_PDU = -1;  /* RadioLinkActivationCommandTDD */
static int hf_nbap_DelayedActivationInformationList_RL_ActivationCmdTDD_PDU = -1;  /* DelayedActivationInformationList_RL_ActivationCmdTDD */
static int hf_nbap_DelayedActivationInformation_RL_ActivationCmdTDD_PDU = -1;  /* DelayedActivationInformation_RL_ActivationCmdTDD */
static int hf_nbap_RadioLinkParameterUpdateIndicationFDD_PDU = -1;  /* RadioLinkParameterUpdateIndicationFDD */
static int hf_nbap_Additional_HS_Cell_Information_RL_Param_Upd_PDU = -1;  /* Additional_HS_Cell_Information_RL_Param_Upd */
static int hf_nbap_Additional_EDCH_Cell_Information_RL_Param_Upd_PDU = -1;  /* Additional_EDCH_Cell_Information_RL_Param_Upd */
static int hf_nbap_RadioLinkParameterUpdateIndicationTDD_PDU = -1;  /* RadioLinkParameterUpdateIndicationTDD */
static int hf_nbap_MBMSNotificationUpdateCommand_PDU = -1;  /* MBMSNotificationUpdateCommand */
static int hf_nbap_UEStatusUpdateCommand_PDU = -1;  /* UEStatusUpdateCommand */
static int hf_nbap_SecondaryULFrequencyReport_PDU = -1;  /* SecondaryULFrequencyReport */
static int hf_nbap_SecondaryULFrequencyUpdateIndication_PDU = -1;  /* SecondaryULFrequencyUpdateIndication */
static int hf_nbap_NBAP_PDU_PDU = -1;             /* NBAP_PDU */
static int hf_nbap_NULL_PDU = -1;                 /* NULL */
static int hf_nbap_local = -1;                    /* INTEGER_0_maxPrivateIEs */
static int hf_nbap_global = -1;                   /* OBJECT_IDENTIFIER */
static int hf_nbap_procedureCode = -1;            /* ProcedureCode */
static int hf_nbap_ddMode = -1;                   /* DdMode */
static int hf_nbap_shortTransActionId = -1;       /* T_shortTransActionId */
static int hf_nbap_longTransActionId = -1;        /* T_longTransActionId */
static int hf_nbap_ProtocolIE_Container_item = -1;  /* ProtocolIE_Field */
static int hf_nbap_id = -1;                       /* ProtocolIE_ID */
static int hf_nbap_criticality = -1;              /* Criticality */
static int hf_nbap_ie_field_value = -1;           /* ProtocolIE_Field_value */
static int hf_nbap_ProtocolExtensionContainer_item = -1;  /* ProtocolExtensionField */
static int hf_nbap_extensionValue = -1;           /* T_extensionValue */
static int hf_nbap_PrivateIE_Container_item = -1;  /* PrivateIE_Field */
static int hf_nbap_id_01 = -1;                    /* PrivateIE_ID */
static int hf_nbap_private_value = -1;            /* PrivateIE_Field_value */
static int hf_nbap_ActivationInformation_item = -1;  /* ActivationInformationItem */
static int hf_nbap_uU_ActivationState = -1;       /* Uu_ActivationState */
static int hf_nbap_iE_Extensions = -1;            /* ProtocolExtensionContainer */
static int hf_nbap_multicell_EDCH_Transport_Bearer_Mode = -1;  /* Multicell_EDCH_Transport_Bearer_Mode */
static int hf_nbap_additional_EDCH_Cell_Information_Setup = -1;  /* Additional_EDCH_Cell_Information_Setup */
static int hf_nbap_Additional_EDCH_Cell_Information_Setup_item = -1;  /* Additional_EDCH_FDD_Setup_Cell_Information */
static int hf_nbap_additional_EDCH_UL_DPCH_Information_Setup = -1;  /* Additional_EDCH_UL_DPCH_Information_Setup */
static int hf_nbap_additional_EDCH_RL_Specific_Information_To_Setup = -1;  /* Additional_EDCH_RL_Specific_Information_To_Setup_List */
static int hf_nbap_additional_EDCH_FDD_Information = -1;  /* Additional_EDCH_FDD_Information */
static int hf_nbap_additional_EDCH_F_DPCH_Information_Setup = -1;  /* Additional_EDCH_F_DPCH_Information */
static int hf_nbap_multicell_EDCH_Information = -1;  /* Multicell_EDCH_Information */
static int hf_nbap_ul_ScramblingCode = -1;        /* UL_ScramblingCode */
static int hf_nbap_ul_SIR_Target = -1;            /* UL_SIR */
static int hf_nbap_fdd_TPC_DownlinkStepSize = -1;  /* FDD_TPC_DownlinkStepSize */
static int hf_nbap_limitedPowerIncrease = -1;     /* LimitedPowerIncrease */
static int hf_nbap_innerLoopDLPCStatus = -1;      /* InnerLoopDLPCStatus */
static int hf_nbap_Additional_EDCH_RL_Specific_Information_To_Setup_List_item = -1;  /* Additional_EDCH_RL_Specific_Information_To_Setup_ItemIEs */
static int hf_nbap_eDCH_Additional_RL_ID = -1;    /* RL_ID */
static int hf_nbap_c_ID = -1;                     /* C_ID */
static int hf_nbap_firstRLS_indicator = -1;       /* FirstRLS_Indicator */
static int hf_nbap_propagationDelay = -1;         /* PropagationDelay */
static int hf_nbap_dl_CodeInformation = -1;       /* FDD_DL_CodeInformation */
static int hf_nbap_initialDL_transmissionPower = -1;  /* DL_Power */
static int hf_nbap_maximumDL_power = -1;          /* DL_Power */
static int hf_nbap_minimumDL_power = -1;          /* DL_Power */
static int hf_nbap_f_DPCH_SlotFormat = -1;        /* F_DPCH_SlotFormat */
static int hf_nbap_e_RNTI = -1;                   /* E_RNTI */
static int hf_nbap_multicell_EDCH_RL_Specific_Information = -1;  /* Multicell_EDCH_RL_Specific_Information */
static int hf_nbap_Additional_EDCH_Cell_Information_To_Add_List_item = -1;  /* Additional_EDCH_Cell_Information_To_Add_ItemIEs */
static int hf_nbap_additional_EDCH_RL_Specific_Information_To_Add_ItemIEs = -1;  /* Additional_EDCH_RL_Specific_Information_To_Add_ItemIEs */
static int hf_nbap_Additional_EDCH_RL_Specific_Information_To_Add_ItemIEs_item = -1;  /* EDCH_Additional_RL_Specific_Information_To_Add_List */
static int hf_nbap_Additional_EDCH_RL_Specific_Information_To_Modify_List_item = -1;  /* Additional_EDCH_RL_Specific_Information_To_Modify_ItemIEs */
static int hf_nbap_additional_EDCH_MAC_d_Flows_Specific_Information = -1;  /* Additional_EDCH_MAC_d_Flows_Specific_Info_List */
static int hf_nbap_hARQ_Process_Allocation_Scheduled_2ms_EDCH = -1;  /* HARQ_Process_Allocation_2ms_EDCH */
static int hf_nbap_e_DCH_Maximum_Bitrate = -1;    /* E_DCH_Maximum_Bitrate */
static int hf_nbap_e_DCH_Processing_Overload_Level = -1;  /* E_DCH_Processing_Overload_Level */
static int hf_nbap_e_DCH_Min_Set_E_TFCI = -1;     /* E_TFCI */
static int hf_nbap_Additional_EDCH_MAC_d_Flows_Specific_Info_List_item = -1;  /* Additional_EDCH_MAC_d_Flows_Specific_Info */
static int hf_nbap_e_DCH_MACdFlow_ID = -1;        /* E_DCH_MACdFlow_ID */
static int hf_nbap_bindingID = -1;                /* BindingID */
static int hf_nbap_transportLayerAddress = -1;    /* TransportLayerAddress */
static int hf_nbap_Additional_EDCH_Cell_Information_Response_List_item = -1;  /* Additional_EDCH_FDD_Information_Response_ItemIEs */
static int hf_nbap_eDCH_Additional_RL_Specific_Information_Response = -1;  /* EDCH_Additional_RL_Specific_Information_Response_List */
static int hf_nbap_additional_EDCH_MAC_d_Flow_Specific_Information_Response = -1;  /* Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List */
static int hf_nbap_EDCH_Additional_RL_Specific_Information_Response_List_item = -1;  /* EDCH_Additional_RL_Specific_Information_Response_ItemIEs */
static int hf_nbap_received_total_wide_band_power = -1;  /* Received_total_wide_band_power_Value */
static int hf_nbap_dL_PowerBalancing_ActivationIndicator = -1;  /* DL_PowerBalancing_ActivationIndicator */
static int hf_nbap_rL_Set_ID = -1;                /* RL_Set_ID */
static int hf_nbap_e_DCH_RL_Set_ID = -1;          /* RL_Set_ID */
static int hf_nbap_e_DCH_FDD_DL_Control_Channel_Information = -1;  /* E_DCH_FDD_DL_Control_Channel_Information */
static int hf_nbap_Additional_EDCH_Cell_Information_Response_RLReconf_List_item = -1;  /* Additional_EDCH_FDD_Information_Response_RLReconf_Items */
static int hf_nbap_additional_EDCH_FDD_Information_Response_ItemIEs = -1;  /* Additional_EDCH_FDD_Information_Response_ItemIEs */
static int hf_nbap_additional_Modififed_EDCH_FDD_Information_Response_ItemIEs = -1;  /* Additional_Modififed_EDCH_FDD_Information_Response_ItemIEs */
static int hf_nbap_eDCH_Additional_Modified_RL_Specific_Information_Response = -1;  /* EDCH_Additional_Modified_RL_Specific_Information_Response_List */
static int hf_nbap_EDCH_Additional_Modified_RL_Specific_Information_Response_List_item = -1;  /* EDCH_Additional_Modified_RL_Specific_Information_Response_List_Items */
static int hf_nbap_dL_PowerBalancing_UpdatedIndicator = -1;  /* DL_PowerBalancing_UpdatedIndicator */
static int hf_nbap_Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List_item = -1;  /* Additional_EDCH_MAC_d_Flows_Specific_Info_Response */
static int hf_nbap_Additional_EDCH_Cell_Information_Response_RL_Add_List_item = -1;  /* Additional_EDCH_Cell_Information_Response_RL_Add_ItemIEs */
static int hf_nbap_additional_EDCH_FDD_Information_Response = -1;  /* Additional_EDCH_FDD_Information_Response_ItemIEs */
static int hf_nbap_additional_EDCH_Serving_Cell_Change_Information_Response = -1;  /* E_DCH_Serving_Cell_Change_Info_Response */
static int hf_nbap_Additional_EDCH_Cell_Information_ConfigurationChange_List_item = -1;  /* Additional_EDCH_ConfigurationChange_Info_ItemIEs */
static int hf_nbap_additional_EDCH_UL_DPCH_Information_Modify = -1;  /* Additional_EDCH_UL_DPCH_Information_Modify */
static int hf_nbap_additional_EDCH_RL_Specific_Information_To_Add = -1;  /* Additional_EDCH_RL_Specific_Information_To_Add_ItemIEs */
static int hf_nbap_additional_EDCH_RL_Specific_Information_To_Modify = -1;  /* Additional_EDCH_RL_Specific_Information_To_Modify_List */
static int hf_nbap_additional_EDCH_FDD_Information_To_Modify = -1;  /* Additional_EDCH_FDD_Information */
static int hf_nbap_additional_EDCH_F_DPCH_Information_Modify = -1;  /* Additional_EDCH_F_DPCH_Information */
static int hf_nbap_Additional_EDCH_Cell_Information_Removal_List_item = -1;  /* Additional_EDCH_Cell_Information_Removal_Info_ItemIEs */
static int hf_nbap_rL_on_Secondary_UL_Frequency = -1;  /* RL_on_Secondary_UL_Frequency */
static int hf_nbap_additional_EDCH_DL_Control_Channel_Change_Information = -1;  /* Additional_EDCH_DL_Control_Channel_Change_Information_List */
static int hf_nbap_Additional_EDCH_DL_Control_Channel_Change_Information_List_item = -1;  /* Additional_EDCH_DL_Control_Channel_Change_Info_ItemIEs */
static int hf_nbap_AdditionalMeasurementValueList_item = -1;  /* AdditionalMeasurementValue */
static int hf_nbap_uARFCN = -1;                   /* UARFCN */
static int hf_nbap_timeSlotMeasurementValueListLCR = -1;  /* TimeSlotMeasurementValueListLCR */
static int hf_nbap_AdditionalTimeSlotListLCR_item = -1;  /* AdditionalTimeSlotLCR */
static int hf_nbap_timeslot_InitiatedListLCR = -1;  /* TimeSlot_InitiatedListLCR */
static int hf_nbap_cMConfigurationChangeCFN = -1;  /* CFN */
static int hf_nbap_transmission_Gap_Pattern_Sequence_Status = -1;  /* Transmission_Gap_Pattern_Sequence_Status_List */
static int hf_nbap_Transmission_Gap_Pattern_Sequence_Status_List_item = -1;  /* Transmission_Gap_Pattern_Sequence_Status_List_item */
static int hf_nbap_tGPSID = -1;                   /* TGPSID */
static int hf_nbap_tGPRC = -1;                    /* TGPRC */
static int hf_nbap_tGCFN = -1;                    /* CFN */
static int hf_nbap_priorityLevel = -1;            /* PriorityLevel */
static int hf_nbap_pre_emptionCapability = -1;    /* Pre_emptionCapability */
static int hf_nbap_pre_emptionVulnerability = -1;  /* Pre_emptionVulnerability */
static int hf_nbap_aOA_LCR = -1;                  /* AOA_LCR */
static int hf_nbap_aOA_LCR_Accuracy_Class = -1;   /* AOA_LCR_Accuracy_Class */
static int hf_nbap_bCCH_Specific_HSDSCH_RNTI = -1;  /* HSDSCH_RNTI */
static int hf_nbap_hSSCCH_Power = -1;             /* DL_Power */
static int hf_nbap_hSPDSCH_Power = -1;            /* DL_Power */
static int hf_nbap_Best_Cell_Portions_Value_item = -1;  /* Best_Cell_Portions_Item */
static int hf_nbap_cellPortionID = -1;            /* CellPortionID */
static int hf_nbap_sIRValue = -1;                 /* SIR_Value */
static int hf_nbap_Best_Cell_Portions_ValueLCR_item = -1;  /* Best_Cell_Portions_ItemLCR */
static int hf_nbap_cellPortionLCRID = -1;         /* CellPortionLCRID */
static int hf_nbap_rSCPValue = -1;                /* RSCP_Value */
static int hf_nbap_commonTransportChannelID = -1;  /* CommonTransportChannelID */
static int hf_nbap_cid = -1;                      /* C_ID */
static int hf_nbap_radioNetwork = -1;             /* CauseRadioNetwork */
static int hf_nbap_transport = -1;                /* CauseTransport */
static int hf_nbap_protocol = -1;                 /* CauseProtocol */
static int hf_nbap_misc = -1;                     /* CauseMisc */
static int hf_nbap_Cell_ERNTI_Status_Information_item = -1;  /* Cell_ERNTI_Status_Information_Item */
static int hf_nbap_vacant_ERNTI = -1;             /* Vacant_ERNTI */
static int hf_nbap_Vacant_ERNTI_item = -1;        /* E_RNTI */
static int hf_nbap_initialPhase_0_1048575 = -1;   /* INTEGER_0_1048575_ */
static int hf_nbap_steadyStatePhase = -1;         /* INTEGER_0_255_ */
static int hf_nbap_initialPhase = -1;             /* INTEGER_0_524287_ */
static int hf_nbap_steadyStatePhase_01 = -1;      /* INTEGER_0_127_ */
static int hf_nbap_CommonChannelsCapacityConsumptionLaw_item = -1;  /* CommonChannelsCapacityConsumptionLaw_item */
static int hf_nbap_dl_Cost = -1;                  /* INTEGER_0_65535 */
static int hf_nbap_ul_Cost = -1;                  /* INTEGER_0_65535 */
static int hf_nbap_common_E_DCH_UL_DPCH_Information = -1;  /* Common_E_DCH_UL_DPCH_InfoItem */
static int hf_nbap_common_E_DCH_EDPCH_Information = -1;  /* Common_E_DCH_EDPCH_InfoItem */
static int hf_nbap_common_E_DCH_Information = -1;  /* Common_E_DCH_InfoItem */
static int hf_nbap_common_E_DCH_HSDPCCH_Information = -1;  /* Common_E_DCH_HSDPCCH_InfoItem */
static int hf_nbap_common_E_DCH_Preamble_Control_Information = -1;  /* Common_E_DCH_Preamble_Control_InfoItem */
static int hf_nbap_common_E_DCH_FDPCH_Information = -1;  /* Common_E_DCH_FDPCH_InfoItem */
static int hf_nbap_common_E_DCH_E_AGCH_ChannelisationCodeNumber = -1;  /* FDD_DL_ChannelisationCodeNumber */
static int hf_nbap_common_E_DCH_Resource_Combination_Information = -1;  /* Common_E_DCH_Resource_Combination_InfoList */
static int hf_nbap_ul_common_E_DCH_MACflow_Specific_Information = -1;  /* Ul_common_E_DCH_MACflow_Specific_InfoList */
static int hf_nbap_uL_SIR_Target = -1;            /* UL_SIR */
static int hf_nbap_dPC_Mode = -1;                 /* DPC_Mode */
static int hf_nbap_maxSet_E_DPDCHs = -1;          /* Max_Set_E_DPDCHs */
static int hf_nbap_ul_PunctureLimit = -1;         /* PunctureLimit */
static int hf_nbap_e_TFCS_Information = -1;       /* E_TFCS_Information */
static int hf_nbap_e_TTI = -1;                    /* E_TTI */
static int hf_nbap_e_DPCCH_PO = -1;               /* E_DPCCH_PO */
static int hf_nbap_e_RGCH_2_IndexStepThreshold = -1;  /* E_RGCH_2_IndexStepThreshold */
static int hf_nbap_e_RGCH_3_IndexStepThreshold = -1;  /* E_RGCH_3_IndexStepThreshold */
static int hf_nbap_hARQ_Info_for_E_DCH = -1;      /* HARQ_Info_for_E_DCH */
static int hf_nbap_e_DCH_Reference_Power_Offset = -1;  /* E_DCH_Reference_Power_Offset */
static int hf_nbap_e_DCH_PowerOffset_for_SchedulingInfo = -1;  /* E_DCH_PowerOffset_for_SchedulingInfo */
static int hf_nbap_max_EDCH_Resource_Allocation_for_CCCH = -1;  /* Max_EDCH_Resource_Allocation_for_CCCH */
static int hf_nbap_max_Period_for_Collistion_Resolution = -1;  /* Max_Period_for_Collistion_Resolution */
static int hf_nbap_max_TB_Sizes = -1;             /* Max_TB_Sizes */
static int hf_nbap_common_E_DCH_ImplicitRelease_Indicator = -1;  /* BOOLEAN */
static int hf_nbap_ackNackRepetitionFactor = -1;  /* AckNack_RepetitionFactor */
static int hf_nbap_ackPowerOffset = -1;           /* Ack_Power_Offset */
static int hf_nbap_nackPowerOffset = -1;          /* Nack_Power_Offset */
static int hf_nbap_common_E_DCH_CQI_Info = -1;    /* Common_E_DCH_CQI_Info */
static int hf_nbap_cqiFeedback_CycleK = -1;       /* CQI_Feedback_Cycle */
static int hf_nbap_cqiRepetitionFactor = -1;      /* CQI_RepetitionFactor */
static int hf_nbap_cqiPowerOffset = -1;           /* CQI_Power_Offset */
static int hf_nbap_measurement_Power_Offset = -1;  /* Measurement_Power_Offset */
static int hf_nbap_commonPhysicalChannelID = -1;  /* CommonPhysicalChannelID */
static int hf_nbap_common_E_DCH_PreambleSignatures = -1;  /* PreambleSignatures */
static int hf_nbap_scramblingCodeNumber = -1;     /* ScramblingCodeNumber */
static int hf_nbap_preambleThreshold = -1;        /* PreambleThreshold */
static int hf_nbap_e_AI_Indicator = -1;           /* E_AI_Indicator */
static int hf_nbap_common_E_DCH_AICH_Information = -1;  /* Common_E_DCH_AICH_Information */
static int hf_nbap_aICH_TransmissionTiming = -1;  /* AICH_TransmissionTiming */
static int hf_nbap_fdd_dl_ChannelisationCodeNumber = -1;  /* FDD_DL_ChannelisationCodeNumber */
static int hf_nbap_aICH_Power = -1;               /* AICH_Power */
static int hf_nbap_sTTD_Indicator = -1;           /* STTD_Indicator */
static int hf_nbap_Common_E_DCH_Resource_Combination_InfoList_item = -1;  /* Common_E_DCH_Resource_Combination_InfoList_Item */
static int hf_nbap_soffset = -1;                  /* Soffset */
static int hf_nbap_f_DPCH_DL_Code_Number = -1;    /* FDD_DL_ChannelisationCodeNumber */
static int hf_nbap_ul_DPCH_ScramblingCode = -1;   /* UL_ScramblingCode */
static int hf_nbap_e_RGCH_E_HICH_Channelisation_Code = -1;  /* FDD_DL_ChannelisationCodeNumber */
static int hf_nbap_e_RGCH_Signature_Sequence = -1;  /* E_RGCH_Signature_Sequence */
static int hf_nbap_e_HICH_Signature_Sequence = -1;  /* E_HICH_Signature_Sequence */
static int hf_nbap_Ul_common_E_DCH_MACflow_Specific_InfoList_item = -1;  /* Ul_common_E_DCH_MACflow_Specific_InfoList_Item */
static int hf_nbap_ul_Common_MACFlowID = -1;      /* Common_MACFlow_ID */
static int hf_nbap_transportBearerRequestIndicator = -1;  /* TransportBearerRequestIndicator */
static int hf_nbap_tnlQos = -1;                   /* TnlQos */
static int hf_nbap_payloadCRC_PresenceIndicator = -1;  /* PayloadCRC_PresenceIndicator */
static int hf_nbap_bundlingModeIndicator = -1;    /* BundlingModeIndicator */
static int hf_nbap_common_E_DCH_MACdFlow_Specific_Information = -1;  /* Common_E_DCH_MACdFlow_Specific_InfoList */
static int hf_nbap_Common_E_DCH_MACdFlow_Specific_InfoList_item = -1;  /* Common_E_DCH_MACdFlow_Specific_InfoList_Item */
static int hf_nbap_common_e_DCH_MACdFlow_ID = -1;  /* E_DCH_MACdFlow_ID */
static int hf_nbap_maximum_Number_of_Retransmissions_For_E_DCH = -1;  /* Maximum_Number_of_Retransmissions_For_E_DCH */
static int hf_nbap_eDCH_HARQ_PO_FDD = -1;         /* E_DCH_HARQ_PO_FDD */
static int hf_nbap_eDCH_MACdFlow_Multiplexing_List = -1;  /* E_DCH_MACdFlow_Multiplexing_List */
static int hf_nbap_common_E_DCHLogicalChannelInformation = -1;  /* Common_E_DCH_LogicalChannel_InfoList */
static int hf_nbap_Common_E_DCH_LogicalChannel_InfoList_item = -1;  /* Common_E_DCH_LogicalChannel_InfoList_Item */
static int hf_nbap_logicalChannelId = -1;         /* LogicalChannelID */
static int hf_nbap_maximumMACcPDU_SizeExtended = -1;  /* MAC_PDU_SizeExtended */
static int hf_nbap_ul_common_E_DCH_MACflow_Specific_InfoResponse = -1;  /* Ul_common_E_DCH_MACflow_Specific_InfoResponseList */
static int hf_nbap_serving_Grant_Value = -1;      /* E_Serving_Grant_Value */
static int hf_nbap_E_RNTI_List_item = -1;         /* E_RNTI */
static int hf_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseList_item = -1;  /* Ul_common_E_DCH_MACflow_Specific_InfoResponseList_Item */
static int hf_nbap_Common_MACFlows_to_DeleteFDD_item = -1;  /* Common_MACFlows_to_DeleteFDD_Item */
static int hf_nbap_common_MACFlow_ID = -1;        /* Common_MACFlow_ID */
static int hf_nbap_CommonMACFlow_Specific_InfoList_item = -1;  /* CommonMACFlow_Specific_InfoItem */
static int hf_nbap_common_MACFlow_Id = -1;        /* Common_MACFlow_ID */
static int hf_nbap_tnl_qos = -1;                  /* TnlQos */
static int hf_nbap_common_MACFlow_PriorityQueue_Information = -1;  /* Common_MACFlow_PriorityQueue_Information */
static int hf_nbap_CommonMACFlow_Specific_InfoList_Response_item = -1;  /* CommonMACFlow_Specific_InfoItem_Response */
static int hf_nbap_commonMACFlow_ID = -1;         /* Common_MACFlow_ID */
static int hf_nbap_hSDSCH_Initial_Capacity_Allocation = -1;  /* HSDSCH_Initial_Capacity_Allocation */
static int hf_nbap_Common_MACFlow_PriorityQueue_Information_item = -1;  /* Common_MACFlow_PriorityQueue_Item */
static int hf_nbap_priority_Queue_Information_for_Enhanced_FACH = -1;  /* Priority_Queue_Information_for_Enhanced_FACH_PCH */
static int hf_nbap_tUTRANGPSMeasurementAccuracyClass = -1;  /* TUTRANGPSAccuracyClass */
static int hf_nbap_tUTRANGANSSMeasurementAccuracyClass = -1;  /* TUTRANGANSSAccuracyClass */
static int hf_nbap_transmitted_carrier_power = -1;  /* Transmitted_Carrier_Power_Value */
static int hf_nbap_acknowledged_prach_preambles = -1;  /* Acknowledged_PRACH_preambles_Value */
static int hf_nbap_uL_TimeslotISCP = -1;          /* UL_TimeslotISCP_Value */
static int hf_nbap_notUsed_1_acknowledged_PCPCH_access_preambles = -1;  /* NULL */
static int hf_nbap_notUsed_2_detected_PCPCH_access_preambles = -1;  /* NULL */
static int hf_nbap_extension_CommonMeasurementValue = -1;  /* Extension_CommonMeasurementValue */
static int hf_nbap_measurementAvailable = -1;     /* CommonMeasurementAvailable */
static int hf_nbap_measurementnotAvailable = -1;  /* CommonMeasurementnotAvailable */
static int hf_nbap_commonmeasurementValue = -1;   /* CommonMeasurementValue */
static int hf_nbap_ie_Extensions = -1;            /* ProtocolExtensionContainer */
static int hf_nbap_resourceOperationalState = -1;  /* ResourceOperationalState */
static int hf_nbap_availabilityStatus = -1;       /* AvailabilityStatus */
static int hf_nbap_commonPhysicalChannelID768 = -1;  /* CommonPhysicalChannelID768 */
static int hf_nbap_uE_DTX_DRX_Offset = -1;        /* UE_DTX_DRX_Offset */
static int hf_nbap_enabling_Delay = -1;           /* Enabling_Delay */
static int hf_nbap_dTX_Information = -1;          /* DTX_Information */
static int hf_nbap_dRX_Information = -1;          /* DRX_Information */
static int hf_nbap_dTX_Information_to_Modify = -1;  /* DTX_Information_to_Modify */
static int hf_nbap_dRX_Information_to_Modify = -1;  /* DRX_Information_to_Modify */
static int hf_nbap_ContinuousPacketConnectivityHS_SCCH_less_Information_item = -1;  /* ContinuousPacketConnectivityHS_SCCH_less_InformationItem */
static int hf_nbap_transport_Block_Size_Index = -1;  /* Transport_Block_Size_Index */
static int hf_nbap_hSPDSCH_Second_Code_Support = -1;  /* HSPDSCH_Second_Code_Support */
static int hf_nbap_hSPDSCH_First_Code_Index = -1;  /* HSPDSCH_First_Code_Index */
static int hf_nbap_hSPDSCH_Second_Code_Index = -1;  /* HSPDSCH_Second_Code_Index */
static int hf_nbap_continuousPacketConnectivityDTX_DRX_Information = -1;  /* ContinuousPacketConnectivityDTX_DRX_Information */
static int hf_nbap_continuousPacketConnectivityDTX_DRX_Information_to_Modify = -1;  /* ContinuousPacketConnectivityDTX_DRX_Information_to_Modify */
static int hf_nbap_continuousPacketConnectivityHS_SCCH_less_Information = -1;  /* ContinuousPacketConnectivityHS_SCCH_less_Information */
static int hf_nbap_procedureID = -1;              /* ProcedureID */
static int hf_nbap_triggeringMessage = -1;        /* TriggeringMessage */
static int hf_nbap_procedureCriticality = -1;     /* Criticality */
static int hf_nbap_transactionID = -1;            /* TransactionID */
static int hf_nbap_iEsCriticalityDiagnostics = -1;  /* CriticalityDiagnostics_IE_List */
static int hf_nbap_CriticalityDiagnostics_IE_List_item = -1;  /* CriticalityDiagnostics_IE_List_item */
static int hf_nbap_iECriticality = -1;            /* Criticality */
static int hf_nbap_iE_ID = -1;                    /* ProtocolIE_ID */
static int hf_nbap_repetitionNumber = -1;         /* RepetitionNumber0 */
static int hf_nbap_ul_common_E_DCH_MACflow_Specific_InformationLCR = -1;  /* Ul_common_E_DCH_MACflow_Specific_InfoListLCR */
static int hf_nbap_common_E_PUCH_InformationLCR = -1;  /* Common_E_PUCH_InformationLCR */
static int hf_nbap_e_TFCS_Information_TDD = -1;   /* E_TFCS_Information_TDD */
static int hf_nbap_maximum_Number_of_Retransmissions_For_SchedulingInfo = -1;  /* Maximum_Number_of_Retransmissions_For_E_DCH */
static int hf_nbap_eDCH_Retransmission_Timer_SchedulingInfo = -1;  /* E_DCH_MACdFlow_Retransmission_Timer */
static int hf_nbap_minCR = -1;                    /* CodeRate */
static int hf_nbap_maxCR = -1;                    /* CodeRate */
static int hf_nbap_harqInfo = -1;                 /* HARQ_Info_for_E_DCH */
static int hf_nbap_pRXdes_base_perURAFCN = -1;    /* PRXdes_base_perURAFCN */
static int hf_nbap_e_PUCH_TPC_StepSize = -1;      /* TDD_TPC_UplinkStepSize_LCR */
static int hf_nbap_e_AGCH_TPC_StepSize = -1;      /* TDD_TPC_DownlinkStepSize */
static int hf_nbap_e_PUCH_PowerControlGAP = -1;   /* ControlGAP */
static int hf_nbap_PRXdes_base_perURAFCN_item = -1;  /* PRXdes_base_Item */
static int hf_nbap_pRXdes_base = -1;              /* PRXdes_base */
static int hf_nbap_Ul_common_E_DCH_MACflow_Specific_InfoListLCR_item = -1;  /* Ul_common_E_DCH_MACflow_Specific_InfoList_ItemLCR */
static int hf_nbap_ul_Common_MACFlowIDLCR = -1;   /* Common_MACFlow_ID_LCR */
static int hf_nbap_common_E_DCH_MACdFlow_Specific_InformationLCR = -1;  /* Common_E_DCH_MACdFlow_Specific_InfoListLCR */
static int hf_nbap_Common_E_DCH_MACdFlow_Specific_InfoListLCR_item = -1;  /* Common_E_DCH_MACdFlow_Specific_InfoList_ItemLCR */
static int hf_nbap_common_e_DCH_MACdFlow_ID_01 = -1;  /* E_DCH_MACdFlow_ID_LCR */
static int hf_nbap_eDCH_HARQ_PO_TDD = -1;         /* E_DCH_HARQ_PO_TDD */
static int hf_nbap_eDCH_MACdFlow_Retransmission_Timer = -1;  /* E_DCH_MACdFlow_Retransmission_Timer */
static int hf_nbap_ul_common_E_DCH_MACflow_Specific_InfoResponseLCR = -1;  /* Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR */
static int hf_nbap_common_E_AGCH_ListLCR = -1;    /* Common_E_AGCH_ListLCR */
static int hf_nbap_common_E_HICH_ListLCR = -1;    /* Common_E_HICH_ListLCR */
static int hf_nbap_common_E_RNTI_Info_LCR = -1;   /* Common_E_RNTI_Info_LCR */
static int hf_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_item = -1;  /* Ul_common_E_DCH_MACflow_Specific_InfoResponseList_ItemLCR */
static int hf_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_Ext_item = -1;  /* Ul_common_E_DCH_MACflow_Specific_InfoResponseList_ItemLCR */
static int hf_nbap_ul_Common_MACFlowID_LCR = -1;  /* Common_MACFlow_ID_LCR */
static int hf_nbap_Common_E_AGCH_ListLCR_item = -1;  /* Common_E_AGCH_ItemLCR */
static int hf_nbap_e_AGCH_ID = -1;                /* E_AGCH_Id */
static int hf_nbap_Common_E_HICH_ListLCR_item = -1;  /* Common_E_HICH_ItemLCR */
static int hf_nbap_eI = -1;                       /* EI */
static int hf_nbap_e_HICH_ID = -1;                /* E_HICH_ID_LCR */
static int hf_nbap_Common_E_RNTI_Info_LCR_item = -1;  /* Common_E_RNTI_Info_ItemLCR */
static int hf_nbap_starting_E_RNTI = -1;          /* E_RNTI */
static int hf_nbap_number_of_Group = -1;          /* INTEGER_1_32 */
static int hf_nbap_number_of_e_E_RNTI_perGroup = -1;  /* INTEGER_1_7 */
static int hf_nbap_Common_MACFlows_to_DeleteLCR_item = -1;  /* Common_MACFlows_to_DeleteLCR_Item */
static int hf_nbap_common_MACFlow_ID_LCR = -1;    /* Common_MACFlow_ID_LCR */
static int hf_nbap_CommonMACFlow_Specific_InfoListLCR_item = -1;  /* CommonMACFlow_Specific_InfoItemLCR */
static int hf_nbap_common_MACFlow_PriorityQueue_InformationLCR = -1;  /* Common_MACFlow_PriorityQueue_Information */
static int hf_nbap_Common_H_RNTI_InformationLCR_item = -1;  /* Common_H_RNTI_InfoItemLCR */
static int hf_nbap_common_H_RNTI = -1;            /* HSDSCH_RNTI */
static int hf_nbap_t_SYNC = -1;                   /* T_SYNC */
static int hf_nbap_t_PROTECT = -1;                /* T_PROTECT */
static int hf_nbap_n_PROTECT = -1;                /* N_PROTECT */
static int hf_nbap_CommonMACFlow_Specific_InfoList_ResponseLCR_item = -1;  /* CommonMACFlow_Specific_InfoItem_ResponseLCR */
static int hf_nbap_continuousPacketConnectivity_DRX_InformationLCR = -1;  /* ContinuousPacketConnectivity_DRX_InformationLCR */
static int hf_nbap_continuousPacketConnectivity_DRX_Information_to_Modify_LCR = -1;  /* ContinuousPacketConnectivity_DRX_Information_to_Modify_LCR */
static int hf_nbap_hS_DSCH_Semi_PersistentScheduling_Information_LCR = -1;  /* HS_DSCH_Semi_PersistentScheduling_Information_LCR */
static int hf_nbap_hS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR = -1;  /* HS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR */
static int hf_nbap_hS_DSCH_SPS_Deactivate_Indicator_LCR = -1;  /* NULL */
static int hf_nbap_e_DCH_Semi_PersistentScheduling_Information_LCR = -1;  /* E_DCH_Semi_PersistentScheduling_Information_LCR */
static int hf_nbap_e_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR = -1;  /* E_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR */
static int hf_nbap_e_DCH_SPS_Deactivate_Indicator_LCR = -1;  /* NULL */
static int hf_nbap_hS_SCCH_DRX_Information_LCR = -1;  /* HS_SCCH_DRX_Information_LCR */
static int hf_nbap_e_AGCH_DRX_Information_LCR = -1;  /* E_AGCH_DRX_Information_LCR */
static int hf_nbap_hS_SCCH_UE_DRX_Cycle_LCR = -1;  /* UE_DRX_Cycle_LCR */
static int hf_nbap_hS_SCCH_Inactivity_Threshold_for_UE_DRX_Cycle_LCR = -1;  /* Inactivity_Threshold_for_UE_DRX_Cycle_LCR */
static int hf_nbap_hS_SCCH_UE_DRX_Offset_LCR = -1;  /* UE_DRX_Offset_LCR */
static int hf_nbap_sameAsHS_SCCH = -1;            /* NULL */
static int hf_nbap_e_AGCH_DRX_Parameters = -1;    /* E_AGCH_DRX_Parameters */
static int hf_nbap_e_AGCH_UE_DRX_Cycle_LCR = -1;  /* UE_DRX_Cycle_LCR */
static int hf_nbap_e_AGCH_UE_Inactivity_Monitor_Threshold = -1;  /* E_AGCH_UE_Inactivity_Monitor_Threshold */
static int hf_nbap_e_AGCH_UE_DRX_Offset_LCR = -1;  /* UE_DRX_Offset_LCR */
static int hf_nbap_dRX_Information_to_Modify_LCR = -1;  /* DRX_Information_to_Modify_LCR */
static int hf_nbap_modify = -1;                   /* DRX_Information_to_Modify_Items_LCR */
static int hf_nbap_deactivate = -1;               /* NULL */
static int hf_nbap_hS_SCCH_DRX_Information_ResponseLCR = -1;  /* HS_SCCH_DRX_Information_ResponseLCR */
static int hf_nbap_e_AGCH_DRX_Information_ResponseLCR = -1;  /* E_AGCH_DRX_Information_ResponseLCR */
static int hf_nbap_e_AGCH_DRX_Parameters_Response = -1;  /* E_AGCH_DRX_Parameters_Response */
static int hf_nbap_DCH_FDD_Information_item = -1;  /* DCH_FDD_InformationItem */
static int hf_nbap_ul_FP_Mode = -1;               /* UL_FP_Mode */
static int hf_nbap_toAWS = -1;                    /* ToAWS */
static int hf_nbap_toAWE = -1;                    /* ToAWE */
static int hf_nbap_dCH_SpecificInformationList = -1;  /* DCH_Specific_FDD_InformationList */
static int hf_nbap_DCH_Specific_FDD_InformationList_item = -1;  /* DCH_Specific_FDD_Item */
static int hf_nbap_dCH_ID = -1;                   /* T_dCH_ID */
static int hf_nbap_ul_TransportFormatSet = -1;    /* T_ul_TransportFormatSet */
static int hf_nbap_dl_TransportFormatSet = -1;    /* T_dl_TransportFormatSet */
static int hf_nbap_allocationRetentionPriority = -1;  /* AllocationRetentionPriority */
static int hf_nbap_frameHandlingPriority = -1;    /* FrameHandlingPriority */
static int hf_nbap_qE_Selector = -1;              /* QE_Selector */
static int hf_nbap_DCH_InformationResponse_item = -1;  /* DCH_InformationResponseItem */
static int hf_nbap_dCH_ID_01 = -1;                /* DCH_ID */
static int hf_nbap_DCH_MeasurementOccasion_Information_item = -1;  /* DchMeasurementOccasionInformation_Item */
static int hf_nbap_pattern_Sequence_Identifier = -1;  /* Pattern_Sequence_Identifier */
static int hf_nbap_status_Flag = -1;              /* Status_Flag */
static int hf_nbap_measurement_Occasion_Pattern_Sequence_parameters = -1;  /* Measurement_Occasion_Pattern_Sequence_parameters */
static int hf_nbap_measurement_Occasion_Pattern_Sequence_parameters_k = -1;  /* INTEGER_1_9 */
static int hf_nbap_measurement_Occasion_Pattern_Sequence_parameters_offset = -1;  /* INTEGER_0_511 */
static int hf_nbap_measurement_Occasion_Pattern_Sequence_parameters_M_Length = -1;  /* INTEGER_1_512 */
static int hf_nbap_measurement_Occasion_Pattern_Sequence_parameters_Timeslot_Bitmap = -1;  /* BIT_STRING_SIZE_7 */
static int hf_nbap_DCH_TDD_Information_item = -1;  /* DCH_TDD_InformationItem */
static int hf_nbap_dCH_SpecificInformationList_01 = -1;  /* DCH_Specific_TDD_InformationList */
static int hf_nbap_DCH_Specific_TDD_InformationList_item = -1;  /* DCH_Specific_TDD_Item */
static int hf_nbap_ul_CCTrCH_ID = -1;             /* CCTrCH_ID */
static int hf_nbap_dl_CCTrCH_ID = -1;             /* CCTrCH_ID */
static int hf_nbap_ul_TransportFormatSet_01 = -1;  /* TransportFormatSet */
static int hf_nbap_dl_TransportFormatSet_01 = -1;  /* TransportFormatSet */
static int hf_nbap_FDD_DCHs_to_Modify_item = -1;  /* FDD_DCHs_to_ModifyItem */
static int hf_nbap_dCH_SpecificInformationList_02 = -1;  /* DCH_ModifySpecificInformation_FDD */
static int hf_nbap_DCH_ModifySpecificInformation_FDD_item = -1;  /* DCH_ModifySpecificItem_FDD */
static int hf_nbap_dCH_ID_02 = -1;                /* T_dCH_ID_01 */
static int hf_nbap_ul_TransportFormatSet_02 = -1;  /* T_ul_TransportFormatSet_01 */
static int hf_nbap_dl_TransportFormatSet_02 = -1;  /* T_dl_TransportFormatSet_01 */
static int hf_nbap_TDD_DCHs_to_Modify_item = -1;  /* DCH_ModifyItem_TDD */
static int hf_nbap_dCH_SpecificInformationList_03 = -1;  /* DCH_ModifySpecificInformation_TDD */
static int hf_nbap_DCH_ModifySpecificInformation_TDD_item = -1;  /* DCH_ModifySpecificItem_TDD */
static int hf_nbap_DedicatedChannelsCapacityConsumptionLaw_item = -1;  /* DedicatedChannelsCapacityConsumptionLaw_item */
static int hf_nbap_dl_Cost_1 = -1;                /* INTEGER_0_65535 */
static int hf_nbap_dl_Cost_2 = -1;                /* INTEGER_0_65535 */
static int hf_nbap_ul_Cost_1 = -1;                /* INTEGER_0_65535 */
static int hf_nbap_ul_Cost_2 = -1;                /* INTEGER_0_65535 */
static int hf_nbap_sIR_Value = -1;                /* SIR_Value */
static int hf_nbap_sIR_ErrorValue = -1;           /* SIR_Error_Value */
static int hf_nbap_transmittedCodePowerValue = -1;  /* Transmitted_Code_Power_Value */
static int hf_nbap_rSCP = -1;                     /* RSCP_Value */
static int hf_nbap_rxTimingDeviationValue = -1;   /* Rx_Timing_Deviation_Value */
static int hf_nbap_roundTripTime = -1;            /* Round_Trip_Time_Value */
static int hf_nbap_extension_DedicatedMeasurementValue = -1;  /* Extension_DedicatedMeasurementValue */
static int hf_nbap_measurementAvailable_01 = -1;  /* DedicatedMeasurementAvailable */
static int hf_nbap_measurementnotAvailable_01 = -1;  /* DedicatedMeasurementnotAvailable */
static int hf_nbap_dedicatedmeasurementValue = -1;  /* DedicatedMeasurementValue */
static int hf_nbap_cFN = -1;                      /* CFN */
static int hf_nbap_cfn = -1;                      /* CFN */
static int hf_nbap_separate_indication = -1;      /* NULL */
static int hf_nbap_activate = -1;                 /* Activate_Info */
static int hf_nbap_deactivate_01 = -1;            /* Deactivate_Info */
static int hf_nbap_activation_type = -1;          /* Execution_Type */
static int hf_nbap_initial_dl_tx_power = -1;      /* DL_Power */
static int hf_nbap_firstRLS_Indicator = -1;       /* FirstRLS_Indicator */
static int hf_nbap_propagation_delay = -1;        /* PropagationDelay */
static int hf_nbap_deactivation_type = -1;        /* Execution_Type */
static int hf_nbap_synchronised = -1;             /* CFN */
static int hf_nbap_unsynchronised = -1;           /* NULL */
static int hf_nbap_dGANSS_ReferenceTime = -1;     /* INTEGER_0_119 */
static int hf_nbap_dGANSS_Information = -1;       /* DGANSS_Information */
static int hf_nbap_dGANSS_Signal_ID = -1;         /* BIT_STRING_SIZE_8 */
static int hf_nbap_DGANSS_Information_item = -1;  /* DGANSS_InformationItem */
static int hf_nbap_gANSS_SignalId = -1;           /* GANSS_Signal_ID */
static int hf_nbap_gANSS_StatusHealth = -1;       /* GANSS_StatusHealth */
static int hf_nbap_dGANSS_SignalInformation = -1;  /* DGANSS_SignalInformation */
static int hf_nbap_DGANSS_SignalInformation_item = -1;  /* DGANSS_SignalInformationItem */
static int hf_nbap_satId = -1;                    /* INTEGER_0_63 */
static int hf_nbap_gANSS_iod = -1;                /* BIT_STRING_SIZE_10 */
static int hf_nbap_udre = -1;                     /* UDRE */
static int hf_nbap_ganss_prc = -1;                /* INTEGER_M2047_2047 */
static int hf_nbap_ganss_rrc = -1;                /* INTEGER_M127_127 */
static int hf_nbap_pRCDeviation = -1;             /* PRCDeviation */
static int hf_nbap_udreGrowthRate = -1;           /* UDREGrowthRate */
static int hf_nbap_udreValidityTime = -1;         /* UDREValidityTime */
static int hf_nbap_gpstow = -1;                   /* GPSTOW */
static int hf_nbap_status_health = -1;            /* GPS_Status_Health */
static int hf_nbap_satelliteinfo = -1;            /* SAT_Info_DGPSCorrections */
static int hf_nbap_prcdeviation = -1;             /* PRCDeviation */
static int hf_nbap_DL_Timeslot_Information_item = -1;  /* DL_Timeslot_InformationItem */
static int hf_nbap_timeSlot = -1;                 /* TimeSlot */
static int hf_nbap_midambleShiftAndBurstType = -1;  /* MidambleShiftAndBurstType */
static int hf_nbap_tFCI_Presence = -1;            /* TFCI_Presence */
static int hf_nbap_dL_Code_Information = -1;      /* TDD_DL_Code_Information */
static int hf_nbap_DL_TimeslotLCR_Information_item = -1;  /* DL_TimeslotLCR_InformationItem */
static int hf_nbap_timeSlotLCR = -1;              /* TimeSlotLCR */
static int hf_nbap_midambleShiftLCR = -1;         /* MidambleShiftLCR */
static int hf_nbap_dL_Code_LCR_Information = -1;  /* TDD_DL_Code_LCR_Information */
static int hf_nbap_DL_Timeslot768_Information_item = -1;  /* DL_Timeslot768_InformationItem */
static int hf_nbap_midambleShiftAndBurstType768 = -1;  /* MidambleShiftAndBurstType768 */
static int hf_nbap_dL_Code_768_Information = -1;  /* TDD_DL_Code_768_Information */
static int hf_nbap_powerAdjustmentType = -1;      /* PowerAdjustmentType */
static int hf_nbap_dLReferencePower = -1;         /* DL_Power */
static int hf_nbap_dLReferencePowerList_DL_PC_Rqst = -1;  /* DL_ReferencePowerInformationList */
static int hf_nbap_maxAdjustmentStep = -1;        /* MaxAdjustmentStep */
static int hf_nbap_adjustmentPeriod = -1;         /* AdjustmentPeriod */
static int hf_nbap_adjustmentRatio = -1;          /* ScaledAdjustmentRatio */
static int hf_nbap_DL_ReferencePowerInformationList_item = -1;  /* DL_ReferencePowerInformationItem */
static int hf_nbap_rL_ID = -1;                    /* RL_ID */
static int hf_nbap_dl_Reference_Power = -1;       /* DL_Power */
static int hf_nbap_DL_TimeslotISCPInfo_item = -1;  /* DL_TimeslotISCPInfoItem */
static int hf_nbap_dL_TimeslotISCP = -1;          /* DL_TimeslotISCP */
static int hf_nbap_DL_TimeslotISCPInfoLCR_item = -1;  /* DL_TimeslotISCPInfoItemLCR */
static int hf_nbap_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst_item = -1;  /* DL_HS_PDSCH_Timeslot_InformationItem_LCR_PSCH_ReconfRqst */
static int hf_nbap_midambleShiftAndBurstType_01 = -1;  /* MidambleShiftLCR */
static int hf_nbap_dl_HS_PDSCH_Codelist_LCR_PSCH_ReconfRqst = -1;  /* DL_HS_PDSCH_Codelist_LCR_PSCH_ReconfRqst */
static int hf_nbap_maxHSDSCH_HSSCCH_Power = -1;   /* MaximumTransmissionPower */
static int hf_nbap_MaxHSDSCH_HSSCCH_Power_per_CELLPORTION_item = -1;  /* MaxHSDSCH_HSSCCH_Power_per_CELLPORTION_Item */
static int hf_nbap_DL_HS_PDSCH_Codelist_LCR_PSCH_ReconfRqst_item = -1;  /* TDD_ChannelisationCode */
static int hf_nbap_uE_DRX_Cycle = -1;             /* UE_DRX_Cycle */
static int hf_nbap_inactivity_Threshold_for_UE_DRX_Cycle = -1;  /* Inactivity_Threshold_for_UE_DRX_Cycle */
static int hf_nbap_inactivity_Threshold_for_UE_Grant_Monitoring = -1;  /* Inactivity_Threshold_for_UE_Grant_Monitoring */
static int hf_nbap_uE_DRX_Grant_Monitoring = -1;  /* UE_DRX_Grant_Monitoring */
static int hf_nbap_modify_01 = -1;                /* DRX_Information_to_Modify_Items */
static int hf_nbap_DSCH_InformationResponse_item = -1;  /* DSCH_InformationResponseItem */
static int hf_nbap_dSCH_ID = -1;                  /* DSCH_ID */
static int hf_nbap_DSCH_TDD_Information_item = -1;  /* DSCH_TDD_InformationItem */
static int hf_nbap_cCTrCH_ID = -1;                /* CCTrCH_ID */
static int hf_nbap_transportFormatSet = -1;       /* TransportFormatSet */
static int hf_nbap_uE_DTX_Cycle1_2ms = -1;        /* UE_DTX_Cycle1_2ms */
static int hf_nbap_uE_DTX_Cycle2_2ms = -1;        /* UE_DTX_Cycle2_2ms */
static int hf_nbap_mAC_DTX_Cycle_2ms = -1;        /* MAC_DTX_Cycle_2ms */
static int hf_nbap_uE_DTX_Cycle1_10ms = -1;       /* UE_DTX_Cycle1_10ms */
static int hf_nbap_uE_DTX_Cycle2_10ms = -1;       /* UE_DTX_Cycle2_10ms */
static int hf_nbap_mAC_DTX_Cycle_10ms = -1;       /* MAC_DTX_Cycle_10ms */
static int hf_nbap_e_DCH_TTI_Length = -1;         /* E_DCH_TTI_Length */
static int hf_nbap_inactivity_Threshold_for_UE_DTX_Cycle2 = -1;  /* Inactivity_Threshold_for_UE_DTX_Cycle2 */
static int hf_nbap_uE_DTX_Long_Preamble = -1;     /* UE_DTX_Long_Preamble */
static int hf_nbap_mAC_Inactivity_Threshold = -1;  /* MAC_Inactivity_Threshold */
static int hf_nbap_cQI_DTX_Timer = -1;            /* CQI_DTX_Timer */
static int hf_nbap_uE_DPCCH_burst1 = -1;          /* UE_DPCCH_burst1 */
static int hf_nbap_uE_DPCCH_burst2 = -1;          /* UE_DPCCH_burst2 */
static int hf_nbap_modify_02 = -1;                /* DTX_Information_to_Modify_Items */
static int hf_nbap_e_DCH_TTI_Length_to_Modify = -1;  /* E_DCH_TTI_Length_to_Modify */
static int hf_nbap_dual_Band_Capability = -1;     /* Dual_Band_Capability */
static int hf_nbap_possible_Secondary_Serving_Cell_List = -1;  /* Possible_Secondary_Serving_Cell_List */
static int hf_nbap_replace = -1;                  /* E_AGCH_FDD_Code_List */
static int hf_nbap_remove = -1;                   /* NULL */
static int hf_nbap_E_AGCH_FDD_Code_List_item = -1;  /* FDD_DL_ChannelisationCodeNumber */
static int hf_nbap_e_DCH_SF_allocation = -1;      /* E_DCH_SF_allocation */
static int hf_nbap_E_DCH_SF_allocation_item = -1;  /* E_DCH_SF_allocation_item */
static int hf_nbap_e_AGCH_And_E_RGCH_E_HICH_FDD_Scrambling_Code = -1;  /* DL_ScramblingCode */
static int hf_nbap_e_AGCH_Channelisation_Code = -1;  /* FDD_DL_ChannelisationCodeNumber */
static int hf_nbap_primary_e_RNTI = -1;           /* E_RNTI */
static int hf_nbap_secondary_e_RNTI = -1;         /* E_RNTI */
static int hf_nbap_primary_Secondary_Grant_Selector = -1;  /* E_Primary_Secondary_Grant_Selector */
static int hf_nbap_e_RGCH_Release_Indicator = -1;  /* E_RGCH_Release_Indicator */
static int hf_nbap_e_DCH_MACdFlows_Information = -1;  /* E_DCH_MACdFlows_Information */
static int hf_nbap_e_DCH_MACdFlow_Specific_InformationResp = -1;  /* E_DCH_MACdFlow_Specific_InformationResp */
static int hf_nbap_e_DCH_MACdFlow_Specific_Info_to_Modify = -1;  /* E_DCH_MACdFlow_Specific_InfoList_to_Modify */
static int hf_nbap_mACeReset_Indicator = -1;      /* MACeReset_Indicator */
static int hf_nbap_e_DCH_MACdFlow_Specific_UpdateInformation = -1;  /* E_DCH_MACdFlow_Specific_UpdateInformation */
static int hf_nbap_E_DCH_MACdFlow_Specific_UpdateInformation_item = -1;  /* E_DCH_MACdFlow_Specific_UpdateInformation_Item */
static int hf_nbap_hARQ_Process_Allocation_NonSched_2ms_EDCH = -1;  /* HARQ_Process_Allocation_2ms_EDCH */
static int hf_nbap_E_DCH_DL_Control_Channel_Change_Information_item = -1;  /* E_DCH_DL_Control_Channel_Change_Information_Item */
static int hf_nbap_e_DCH_RL_ID = -1;              /* RL_ID */
static int hf_nbap_E_DCH_DL_Control_Channel_Grant_Information_item = -1;  /* E_DCH_DL_Control_Channel_Grant_Information_Item */
static int hf_nbap_e_DCH_Non_Scheduled_Transmission_Grant = -1;  /* E_DCH_Non_Scheduled_Transmission_Grant_Items */
static int hf_nbap_e_DCH_Scheduled_Transmission_Grant = -1;  /* NULL */
static int hf_nbap_E_DCH_LogicalChannelInformation_item = -1;  /* E_DCH_LogicalChannelInformationItem */
static int hf_nbap_schedulingPriorityIndicator = -1;  /* SchedulingPriorityIndicator */
static int hf_nbap_schedulingInformation = -1;    /* SchedulingInformation */
static int hf_nbap_mACesGuaranteedBitRate = -1;   /* MACesGuaranteedBitRate */
static int hf_nbap_e_DCH_DDI_Value = -1;          /* E_DCH_DDI_Value */
static int hf_nbap_mACd_PDU_Size_List = -1;       /* E_DCH_MACdPDU_SizeList */
static int hf_nbap_E_DCH_MACdPDU_SizeList_item = -1;  /* E_DCH_MACdPDU_SizeListItem */
static int hf_nbap_mACdPDU_Size = -1;             /* MACdPDU_Size */
static int hf_nbap_E_DCH_LogicalChannelToModify_item = -1;  /* E_DCH_LogicalChannelToModifyItem */
static int hf_nbap_mACd_PDU_Size_List_01 = -1;    /* E_DCH_MACdPDU_SizeToModifyList */
static int hf_nbap_E_DCH_MACdPDU_SizeToModifyList_item = -1;  /* E_DCH_MACdPDU_SizeListItem */
static int hf_nbap_E_DCH_LogicalChannelToDelete_item = -1;  /* E_DCH_LogicalChannelToDeleteItem */
static int hf_nbap_e_DCH_MACdFlow_Specific_Info = -1;  /* E_DCH_MACdFlow_Specific_InfoList */
static int hf_nbap_E_DCH_MACdFlow_Specific_InfoList_item = -1;  /* E_DCH_MACdFlow_Specific_InfoItem */
static int hf_nbap_eDCH_Grant_Type_Information = -1;  /* E_DCH_Grant_Type_Information */
static int hf_nbap_eDCHLogicalChannelInformation = -1;  /* E_DCH_LogicalChannelInformation */
static int hf_nbap_E_DCH_MACdFlow_Specific_InformationResp_item = -1;  /* E_DCH_MACdFlow_Specific_InformationResp_Item */
static int hf_nbap_E_DCH_MACdFlow_Specific_InfoList_to_Modify_item = -1;  /* E_DCH_MACdFlow_Specific_InfoItem_to_Modify */
static int hf_nbap_eDCH_LogicalChannelToAdd = -1;  /* E_DCH_LogicalChannelInformation */
static int hf_nbap_eDCH_LogicalChannelToModify = -1;  /* E_DCH_LogicalChannelToModify */
static int hf_nbap_eDCH_LogicalChannelToDelete = -1;  /* E_DCH_LogicalChannelToDelete */
static int hf_nbap_E_DCH_MACdFlows_to_Delete_item = -1;  /* E_DCH_MACdFlow_to_Delete_Item */
static int hf_nbap_maxBits_MACe_PDU_non_scheduled = -1;  /* Max_Bits_MACe_PDU_non_scheduled */
static int hf_nbap_hARQ_Process_Allocation_NonSched_2ms = -1;  /* HARQ_Process_Allocation_2ms_EDCH */
static int hf_nbap_e_DCH_serving_cell_choice = -1;  /* E_DCH_serving_cell_choice */
static int hf_nbap_e_DCH_serving_cell_change_successful = -1;  /* E_DCH_serving_cell_change_successful */
static int hf_nbap_e_DCH_serving_cell_change_unsuccessful = -1;  /* E_DCH_serving_cell_change_unsuccessful */
static int hf_nbap_e_DCH_RL_InformationList_Rsp = -1;  /* E_DCH_RL_InformationList_Rsp */
static int hf_nbap_E_DCH_RL_InformationList_Rsp_item = -1;  /* E_DCH_RL_InformationList_Rsp_Item */
static int hf_nbap_rl_ID = -1;                    /* RL_ID */
static int hf_nbap_e_DCH_FDD_DL_Control_Channel_Info = -1;  /* E_DCH_FDD_DL_Control_Channel_Information */
static int hf_nbap_cause = -1;                    /* Cause */
static int hf_nbap_EDCH_RACH_Report_Value_item = -1;  /* EDCH_RACH_Report_Value_item */
static int hf_nbap_granted_EDCH_RACH_resources = -1;  /* Granted_EDCH_RACH_Resources_Value */
static int hf_nbap_denied_EDCH_RACH_resources = -1;  /* Denied_EDCH_RACH_Resources_Value */
static int hf_nbap_two_ms = -1;                   /* DTX_Cycle_2ms_Items */
static int hf_nbap_ten_ms = -1;                   /* DTX_Cycle_10ms_Items */
static int hf_nbap_two_ms_01 = -1;                /* DTX_Cycle_2ms_to_Modify_Items */
static int hf_nbap_ten_ms_01 = -1;                /* DTX_Cycle_10ms_to_Modify_Items */
static int hf_nbap_E_DCH_MACdFlows_to_DeleteLCR_item = -1;  /* E_DCH_MACdFlow_to_Delete_ItemLCR */
static int hf_nbap_e_DCH_MACdFlow_ID_LCR = -1;    /* E_DCH_MACdFlow_ID_LCR */
static int hf_nbap_t321 = -1;                     /* T321 */
static int hf_nbap_hS_DSCH_DRX_Cycle_FACH = -1;   /* HS_DSCH_DRX_Cycle_FACH */
static int hf_nbap_hS_DSCH_RX_Burst_FACH = -1;    /* HS_DSCH_RX_Burst_FACH */
static int hf_nbap_replace_01 = -1;               /* E_RGCH_E_HICH_FDD_Code_List */
static int hf_nbap_E_RGCH_E_HICH_FDD_Code_List_item = -1;  /* FDD_DL_ChannelisationCodeNumber */
static int hf_nbap_e_TFCI_BetaEC_Boost = -1;      /* E_TFCI_BetaEC_Boost */
static int hf_nbap_uL_Delta_T2TP = -1;            /* UL_Delta_T2TP */
static int hf_nbap_e_DCH_TFCI_Table_Index = -1;   /* E_DCH_TFCI_Table_Index */
static int hf_nbap_reference_E_TFCI_Information = -1;  /* Reference_E_TFCI_Information */
static int hf_nbap_E_DCHProvidedBitRate_item = -1;  /* E_DCHProvidedBitRate_Item */
static int hf_nbap_e_DCHProvidedBitRateValue = -1;  /* E_DCHProvidedBitRateValue */
static int hf_nbap_E_DCHProvidedBitRateValueInformation_For_CellPortion_item = -1;  /* E_DCHProvidedBitRateValueInformation_For_CellPortion_Item */
static int hf_nbap_e_DCHProvidedBitRateValue_01 = -1;  /* E_DCHProvidedBitRate */
static int hf_nbap_e_PUCH_Information = -1;       /* E_PUCH_Information */
static int hf_nbap_e_DCH_MACdFlows_Information_TDD = -1;  /* E_DCH_MACdFlows_Information_TDD */
static int hf_nbap_e_DCH_Non_Scheduled_Grant_Info = -1;  /* E_DCH_Non_Scheduled_Grant_Info */
static int hf_nbap_e_DCH_TDD_Information = -1;    /* E_DCH_TDD_Information */
static int hf_nbap_n_E_UCCH = -1;                 /* N_E_UCCH */
static int hf_nbap_e_DCH_QPSK_RefBetaInfo = -1;   /* E_DCH_QPSK_RefBetaInfo */
static int hf_nbap_e_DCH_sixteenQAM_RefBetaInfo = -1;  /* E_DCH_sixteenQAM_RefBetaInfo */
static int hf_nbap_E_DCH_QPSK_RefBetaInfo_item = -1;  /* E_DCH_RefBeta_Item */
static int hf_nbap_E_DCH_sixteenQAM_RefBetaInfo_item = -1;  /* E_DCH_RefBeta_Item */
static int hf_nbap_refCodeRate = -1;              /* CodeRate_short */
static int hf_nbap_refBeta = -1;                  /* RefBeta */
static int hf_nbap_E_DCH_MACdFlows_Information_TDD_item = -1;  /* E_DCH_MACdFlow_InfoTDDItem */
static int hf_nbap_eDCH_Grant_TypeTDD = -1;       /* E_DCH_Grant_TypeTDD */
static int hf_nbap_timeslotResource = -1;         /* E_DCH_TimeslotResource */
static int hf_nbap_powerResource = -1;            /* E_DCH_PowerResource */
static int hf_nbap_repetitionPeriod = -1;         /* RepetitionPeriod */
static int hf_nbap_repetitionLength = -1;         /* RepetitionLength */
static int hf_nbap_tddE_PUCH_Offset = -1;         /* TddE_PUCH_Offset */
static int hf_nbap_tdd_ChannelisationCode = -1;   /* TDD_ChannelisationCode */
static int hf_nbap_e_DCH_TDD_Maximum_Bitrate = -1;  /* E_DCH_TDD_Maximum_Bitrate */
static int hf_nbap_e_DCH_TDD_MACdFlow_Specific_InformationResp = -1;  /* E_DCH_TDD_MACdFlow_Specific_InformationResp */
static int hf_nbap_e_AGCH_Specific_Information_ResponseTDD = -1;  /* E_AGCH_Specific_InformationRespListTDD */
static int hf_nbap_scheduled_E_HICH_Specific_InformationResp = -1;  /* Scheduled_E_HICH_Specific_Information_ResponseLCRTDD */
static int hf_nbap_Scheduled_E_HICH_Specific_Information_ResponseLCRTDD_item = -1;  /* Scheduled_E_HICH_Specific_InformationItem_ResponseLCRTDD */
static int hf_nbap_e_HICH_ID_TDD = -1;            /* E_HICH_ID_TDD */
static int hf_nbap_E_DCH_TDD_MACdFlow_Specific_InformationResp_item = -1;  /* E_DCH_TDD_MACdFlow_Specific_InformationResp_Item */
static int hf_nbap_e_DCH_MacdFlow_Id = -1;        /* E_DCH_MACdFlow_ID */
static int hf_nbap_E_AGCH_Specific_InformationRespListTDD_item = -1;  /* E_AGCH_Specific_InformationResp_ItemTDD */
static int hf_nbap_e_AGCH_Id = -1;                /* E_AGCH_Id */
static int hf_nbap_e_DCH_MACdFlows_to_Add = -1;   /* E_DCH_MACdFlows_Information_TDD */
static int hf_nbap_e_DCH_MACdFlows_to_Delete = -1;  /* E_DCH_MACdFlows_to_Delete */
static int hf_nbap_e_DCH_TDD_Information_to_Modify = -1;  /* E_DCH_TDD_Information_to_Modify */
static int hf_nbap_e_DCH_TDD_Information_to_Modify_List = -1;  /* E_DCH_TDD_Information_to_Modify_List */
static int hf_nbap_E_DCH_TDD_Information_to_Modify_List_item = -1;  /* E_DCH_MACdFlow_ModifyTDDItem */
static int hf_nbap_e_DCH_LogicalChannelToAdd = -1;  /* E_DCH_LogicalChannelInformation */
static int hf_nbap_e_DCH_LogicalChannelToModify = -1;  /* E_DCH_LogicalChannelToModify */
static int hf_nbap_e_DCH_LogicalChannelToDelete = -1;  /* E_DCH_LogicalChannelToDelete */
static int hf_nbap_e_DCH_Non_Scheduled_Grant_Info768 = -1;  /* E_DCH_Non_Scheduled_Grant_Info768 */
static int hf_nbap_e_DCH_TDD_Information768 = -1;  /* E_DCH_TDD_Information768 */
static int hf_nbap_tdd_ChannelisationCode768 = -1;  /* TDD_ChannelisationCode768 */
static int hf_nbap_e_DCH_TDD_Maximum_Bitrate768 = -1;  /* E_DCH_TDD_Maximum_Bitrate768 */
static int hf_nbap_e_PUCH_LCR_Information = -1;   /* E_PUCH_LCR_Information */
static int hf_nbap_e_DCH_Non_Scheduled_Grant_LCR_Info = -1;  /* E_DCH_Non_Scheduled_Grant_LCR_Info */
static int hf_nbap_e_DCH_LCRTDD_Information = -1;  /* E_DCH_LCRTDD_Information */
static int hf_nbap_timeslotResourceLCR = -1;      /* E_DCH_TimeslotResourceLCR */
static int hf_nbap_subframeNumber = -1;           /* T_subframeNumber */
static int hf_nbap_n_E_UCCHLCR = -1;              /* N_E_UCCHLCR */
static int hf_nbap_e_HICH_LCR_Information = -1;   /* E_HICH_LCR_Information */
static int hf_nbap_signatureSequenceGroupIndex = -1;  /* SignatureSequenceGroupIndex */
static int hf_nbap_e_DCH_LCRTDD_PhysicalLayerCategory = -1;  /* E_DCH_LCRTDD_PhysicalLayerCategory */
static int hf_nbap_dRX_Interruption_by_HS_DSCH = -1;  /* DRX_Interruption_by_HS_DSCH */
static int hf_nbap_repetition_Period_List_LCR = -1;  /* Repetition_Period_List_LCR */
static int hf_nbap_e_DCH_SPS_Indicator = -1;      /* E_DCH_SPS_Indicator */
static int hf_nbap_sPS_E_DCH_releted_E_HICH_Information = -1;  /* E_HICH_LCR_Information */
static int hf_nbap_timeslot_Resource_Related_Information = -1;  /* E_DCH_TimeslotResourceLCR */
static int hf_nbap_subframeNumber_01 = -1;        /* T_subframeNumber_01 */
static int hf_nbap_FDD_DL_CodeInformation_item = -1;  /* FDD_DL_CodeInformationItem */
static int hf_nbap_dl_ScramblingCode = -1;        /* DL_ScramblingCode */
static int hf_nbap_fdd_DL_ChannelisationCodeNumber = -1;  /* FDD_DL_ChannelisationCodeNumber */
static int hf_nbap_transmissionGapPatternSequenceCodeInformation = -1;  /* TransmissionGapPatternSequenceCodeInformation */
static int hf_nbap_navClockModel = -1;            /* GANSS_NAVclockModel */
static int hf_nbap_cnavClockModel = -1;           /* GANSS_CNAVclockModel */
static int hf_nbap_glonassClockModel = -1;        /* GANSS_GLONASSclockModel */
static int hf_nbap_sbasClockModel = -1;           /* GANSS_SBASclockModel */
static int hf_nbap_navKeplerianSet = -1;          /* GANSS_NavModel_NAVKeplerianSet */
static int hf_nbap_cnavKeplerianSet = -1;         /* GANSS_NavModel_CNAVKeplerianSet */
static int hf_nbap_glonassECEF = -1;              /* GANSS_NavModel_GLONASSecef */
static int hf_nbap_sbasECEF = -1;                 /* GANSS_NavModel_SBASecef */
static int hf_nbap_dataID = -1;                   /* BIT_STRING_SIZE_2 */
static int hf_nbap_alpha_beta_parameters = -1;    /* GPS_Ionospheric_Model */
static int hf_nbap_ganss_Transmission_Time = -1;  /* GANSS_Transmission_Time */
static int hf_nbap_non_broadcastIndication = -1;  /* T_non_broadcastIndication */
static int hf_nbap_ganssSatInfoNavList = -1;      /* Ganss_Sat_Info_AddNavList */
static int hf_nbap_GANSS_Additional_Time_Models_item = -1;  /* GANSS_Time_Model */
static int hf_nbap_utcModel1 = -1;                /* GANSS_UTCmodelSet1 */
static int hf_nbap_utcModel2 = -1;                /* GANSS_UTCmodelSet2 */
static int hf_nbap_utcModel3 = -1;                /* GANSS_UTCmodelSet3 */
static int hf_nbap_ganss_wk_number = -1;          /* INTEGER_0_255 */
static int hf_nbap_gANSS_AlmanacModel = -1;       /* GANSS_AlmanacModel */
static int hf_nbap_gANSS_keplerianParameters = -1;  /* GANSS_KeplerianParametersAlm */
static int hf_nbap_extension_GANSS_AlmanacModel = -1;  /* Extension_GANSS_AlmanacModel */
static int hf_nbap_sat_info_SBASecefList = -1;    /* GANSS_SAT_Info_Almanac_SBASecefList */
static int hf_nbap_sat_info_GLOkpList = -1;       /* GANSS_SAT_Info_Almanac_GLOkpList */
static int hf_nbap_t_oa = -1;                     /* INTEGER_0_255 */
static int hf_nbap_sat_info_MIDIkpList = -1;      /* GANSS_SAT_Info_Almanac_MIDIkpList */
static int hf_nbap_sat_info_NAVkpList = -1;       /* GANSS_SAT_Info_Almanac_NAVkpList */
static int hf_nbap_sat_info_REDkpList = -1;       /* GANSS_SAT_Info_Almanac_REDkpList */
static int hf_nbap_ganssID1 = -1;                 /* GANSS_AuxInfoGANSS_ID1 */
static int hf_nbap_ganssID3 = -1;                 /* GANSS_AuxInfoGANSS_ID3 */
static int hf_nbap_GANSS_AuxInfoGANSS_ID1_item = -1;  /* GANSS_AuxInfoGANSS_ID1_element */
static int hf_nbap_svID = -1;                     /* INTEGER_0_63 */
static int hf_nbap_signalsAvailable = -1;         /* BIT_STRING_SIZE_8 */
static int hf_nbap_GANSS_AuxInfoGANSS_ID3_item = -1;  /* GANSS_AuxInfoGANSS_ID3_element */
static int hf_nbap_channelNumber = -1;            /* INTEGER_M7_13 */
static int hf_nbap_GANSS_Clock_Model_item = -1;   /* GANSS_SatelliteClockModelItem */
static int hf_nbap_cnavToc = -1;                  /* BIT_STRING_SIZE_11 */
static int hf_nbap_cnavTop = -1;                  /* BIT_STRING_SIZE_11 */
static int hf_nbap_cnavURA0 = -1;                 /* BIT_STRING_SIZE_5 */
static int hf_nbap_cnavURA1 = -1;                 /* BIT_STRING_SIZE_3 */
static int hf_nbap_cnavURA2 = -1;                 /* BIT_STRING_SIZE_3 */
static int hf_nbap_cnavAf2 = -1;                  /* BIT_STRING_SIZE_10 */
static int hf_nbap_cnavAf1 = -1;                  /* BIT_STRING_SIZE_20 */
static int hf_nbap_cnavAf0 = -1;                  /* BIT_STRING_SIZE_26 */
static int hf_nbap_cnavTgd = -1;                  /* BIT_STRING_SIZE_13 */
static int hf_nbap_cnavISCl1cp = -1;              /* BIT_STRING_SIZE_13 */
static int hf_nbap_cnavISCl1cd = -1;              /* BIT_STRING_SIZE_13 */
static int hf_nbap_cnavISCl1ca = -1;              /* BIT_STRING_SIZE_13 */
static int hf_nbap_cnavISCl2c = -1;               /* BIT_STRING_SIZE_13 */
static int hf_nbap_cnavISCl5i5 = -1;              /* BIT_STRING_SIZE_13 */
static int hf_nbap_cnavISCl5q5 = -1;              /* BIT_STRING_SIZE_13 */
static int hf_nbap_ganss_Ionospheric_Model = -1;  /* GANSS_Ionospheric_Model */
static int hf_nbap_ganss_Rx_Pos = -1;             /* GANSS_RX_Pos */
static int hf_nbap_ionospheric_Model = -1;        /* BOOLEAN */
static int hf_nbap_ganssTod = -1;                 /* INTEGER_0_59_ */
static int hf_nbap_dataBitAssistancelist = -1;    /* GANSS_DataBitAssistanceList */
static int hf_nbap_GANSS_DataBitAssistanceList_item = -1;  /* GANSS_DataBitAssistanceItem */
static int hf_nbap_dataBitAssistanceSgnList = -1;  /* GANSS_DataBitAssistanceSgnList */
static int hf_nbap_GANSS_DataBitAssistanceSgnList_item = -1;  /* GANSS_DataBitAssistanceSgnItem */
static int hf_nbap_ganss_SignalId = -1;           /* GANSS_Signal_ID */
static int hf_nbap_ganssDataBits = -1;            /* BIT_STRING_SIZE_1_1024 */
static int hf_nbap_ganssTod_01 = -1;              /* INTEGER_0_86399 */
static int hf_nbap_ganss_Data_Bit_Assistance_ReqList = -1;  /* GANSS_Data_Bit_Assistance_ReqList */
static int hf_nbap_ganss_DataBitInterval = -1;    /* INTEGER_0_15 */
static int hf_nbap_ganss_SatelliteInfo = -1;      /* T_ganss_SatelliteInfo */
static int hf_nbap_ganss_SatelliteInfo_item = -1;  /* INTEGER_0_63 */
static int hf_nbap_b1 = -1;                       /* BIT_STRING_SIZE_11 */
static int hf_nbap_b2 = -1;                       /* BIT_STRING_SIZE_10 */
static int hf_nbap_teop = -1;                     /* BIT_STRING_SIZE_16 */
static int hf_nbap_pmX = -1;                      /* BIT_STRING_SIZE_21 */
static int hf_nbap_pmXdot = -1;                   /* BIT_STRING_SIZE_15 */
static int hf_nbap_pmY = -1;                      /* BIT_STRING_SIZE_21 */
static int hf_nbap_pmYdot = -1;                   /* BIT_STRING_SIZE_15 */
static int hf_nbap_deltaUT1 = -1;                 /* BIT_STRING_SIZE_31 */
static int hf_nbap_deltaUT1dot = -1;              /* BIT_STRING_SIZE_19 */
static int hf_nbap_GANSS_GenericDataInfoReqList_item = -1;  /* GANSS_GenericDataInfoReqItem */
static int hf_nbap_ganss_Id = -1;                 /* GANSS_ID */
static int hf_nbap_ganss_Navigation_Model_And_Time_Recovery = -1;  /* BOOLEAN */
static int hf_nbap_ganss_Time_Model_GNSS_GNSS = -1;  /* BIT_STRING_SIZE_9 */
static int hf_nbap_ganss_UTC_Model = -1;          /* BOOLEAN */
static int hf_nbap_ganss_Almanac = -1;            /* BOOLEAN */
static int hf_nbap_gANSS_GenericDataInfoReqItem_ganss_Real_Time_Integrity = -1;  /* BOOLEAN */
static int hf_nbap_ganss_Data_Bit_Assistance_Req = -1;  /* GANSS_Data_Bit_Assistance_ReqItem */
static int hf_nbap_GANSS_Generic_Data_item = -1;  /* GANSS_Generic_DataItem */
static int hf_nbap_dganss_Correction = -1;        /* DGANSSCorrections */
static int hf_nbap_ganss_Navigation_Model_And_Time_Recovery_01 = -1;  /* GANSS_Navigation_Model_And_Time_Recovery */
static int hf_nbap_ganss_Time_Model = -1;         /* GANSS_Time_Model */
static int hf_nbap_ganss_UTC_TIME = -1;           /* GANSS_UTC_Model */
static int hf_nbap_ganss_Almanac_01 = -1;         /* GANSS_Almanac */
static int hf_nbap_ganss_Real_Time_Integrity = -1;  /* GANSS_Real_Time_Integrity */
static int hf_nbap_ganss_Data_Bit_Assistance = -1;  /* GANSS_Data_Bit_Assistance */
static int hf_nbap_gloTau = -1;                   /* BIT_STRING_SIZE_22 */
static int hf_nbap_gloGamma = -1;                 /* BIT_STRING_SIZE_11 */
static int hf_nbap_gloDeltaTau = -1;              /* BIT_STRING_SIZE_5 */
static int hf_nbap_gANSS_CommonDataInfoReq = -1;  /* GANSS_CommonDataInfoReq */
static int hf_nbap_gANSS_GenericDataInfoReqList = -1;  /* GANSS_GenericDataInfoReqList */
static int hf_nbap_alpha_zero_ionos = -1;         /* BIT_STRING_SIZE_12 */
static int hf_nbap_alpha_one_ionos = -1;          /* BIT_STRING_SIZE_12 */
static int hf_nbap_alpha_two_ionos = -1;          /* BIT_STRING_SIZE_12 */
static int hf_nbap_gANSS_IonosphereRegionalStormFlags = -1;  /* GANSS_IonosphereRegionalStormFlags */
static int hf_nbap_storm_flag_one = -1;           /* BOOLEAN */
static int hf_nbap_storm_flag_two = -1;           /* BOOLEAN */
static int hf_nbap_storm_flag_three = -1;         /* BOOLEAN */
static int hf_nbap_storm_flag_four = -1;          /* BOOLEAN */
static int hf_nbap_storm_flag_five = -1;          /* BOOLEAN */
static int hf_nbap_iod_a = -1;                    /* INTEGER_0_3 */
static int hf_nbap_gANSS_SatelliteInformationKP = -1;  /* GANSS_SatelliteInformationKP */
static int hf_nbap_toe_nav = -1;                  /* BIT_STRING_SIZE_14 */
static int hf_nbap_ganss_omega_nav = -1;          /* BIT_STRING_SIZE_32 */
static int hf_nbap_delta_n_nav = -1;              /* BIT_STRING_SIZE_16 */
static int hf_nbap_m_zero_nav = -1;               /* BIT_STRING_SIZE_32 */
static int hf_nbap_omegadot_nav = -1;             /* BIT_STRING_SIZE_24 */
static int hf_nbap_ganss_e_nav = -1;              /* BIT_STRING_SIZE_32 */
static int hf_nbap_idot_nav = -1;                 /* BIT_STRING_SIZE_14 */
static int hf_nbap_a_sqrt_nav = -1;               /* BIT_STRING_SIZE_32 */
static int hf_nbap_i_zero_nav = -1;               /* BIT_STRING_SIZE_32 */
static int hf_nbap_omega_zero_nav = -1;           /* BIT_STRING_SIZE_32 */
static int hf_nbap_c_rs_nav = -1;                 /* BIT_STRING_SIZE_16 */
static int hf_nbap_c_is_nav = -1;                 /* BIT_STRING_SIZE_16 */
static int hf_nbap_c_us_nav = -1;                 /* BIT_STRING_SIZE_16 */
static int hf_nbap_c_rc_nav = -1;                 /* BIT_STRING_SIZE_16 */
static int hf_nbap_c_ic_nav = -1;                 /* BIT_STRING_SIZE_16 */
static int hf_nbap_c_uc_nav = -1;                 /* BIT_STRING_SIZE_16 */
static int hf_nbap_navToc = -1;                   /* BIT_STRING_SIZE_16 */
static int hf_nbap_navaf2 = -1;                   /* BIT_STRING_SIZE_8 */
static int hf_nbap_navaf1 = -1;                   /* BIT_STRING_SIZE_16 */
static int hf_nbap_navaf0 = -1;                   /* BIT_STRING_SIZE_22 */
static int hf_nbap_navTgd = -1;                   /* BIT_STRING_SIZE_8 */
static int hf_nbap_non_broadcastIndication_01 = -1;  /* T_non_broadcastIndication_01 */
static int hf_nbap_ganssSatInfoNav = -1;          /* GANSS_Sat_Info_Nav */
static int hf_nbap_cnavURAindex = -1;             /* BIT_STRING_SIZE_5 */
static int hf_nbap_cnavDeltaA = -1;               /* BIT_STRING_SIZE_26 */
static int hf_nbap_cnavAdot = -1;                 /* BIT_STRING_SIZE_25 */
static int hf_nbap_cnavDeltaNo = -1;              /* BIT_STRING_SIZE_17 */
static int hf_nbap_cnavDeltaNoDot = -1;           /* BIT_STRING_SIZE_23 */
static int hf_nbap_cnavMo = -1;                   /* BIT_STRING_SIZE_33 */
static int hf_nbap_cnavE = -1;                    /* BIT_STRING_SIZE_33 */
static int hf_nbap_cnavOmega = -1;                /* BIT_STRING_SIZE_33 */
static int hf_nbap_cnavOMEGA0 = -1;               /* BIT_STRING_SIZE_33 */
static int hf_nbap_cnavDeltaOmegaDot = -1;        /* BIT_STRING_SIZE_17 */
static int hf_nbap_cnavIo = -1;                   /* BIT_STRING_SIZE_33 */
static int hf_nbap_cnavIoDot = -1;                /* BIT_STRING_SIZE_15 */
static int hf_nbap_cnavCis = -1;                  /* BIT_STRING_SIZE_16 */
static int hf_nbap_cnavCic = -1;                  /* BIT_STRING_SIZE_16 */
static int hf_nbap_cnavCrs = -1;                  /* BIT_STRING_SIZE_24 */
static int hf_nbap_cnavCrc = -1;                  /* BIT_STRING_SIZE_24 */
static int hf_nbap_cnavCus = -1;                  /* BIT_STRING_SIZE_21 */
static int hf_nbap_cnavCuc = -1;                  /* BIT_STRING_SIZE_21 */
static int hf_nbap_gloEn = -1;                    /* BIT_STRING_SIZE_5 */
static int hf_nbap_gloP1 = -1;                    /* BIT_STRING_SIZE_2 */
static int hf_nbap_gloP2 = -1;                    /* BIT_STRING_SIZE_1 */
static int hf_nbap_gloM = -1;                     /* BIT_STRING_SIZE_2 */
static int hf_nbap_gloX = -1;                     /* BIT_STRING_SIZE_27 */
static int hf_nbap_gloXdot = -1;                  /* BIT_STRING_SIZE_24 */
static int hf_nbap_gloXdotdot = -1;               /* BIT_STRING_SIZE_5 */
static int hf_nbap_gloY = -1;                     /* BIT_STRING_SIZE_27 */
static int hf_nbap_gloYdot = -1;                  /* BIT_STRING_SIZE_24 */
static int hf_nbap_gloYdotdot = -1;               /* BIT_STRING_SIZE_5 */
static int hf_nbap_gloZ = -1;                     /* BIT_STRING_SIZE_27 */
static int hf_nbap_gloZdot = -1;                  /* BIT_STRING_SIZE_24 */
static int hf_nbap_gloZdotdot = -1;               /* BIT_STRING_SIZE_5 */
static int hf_nbap_navURA = -1;                   /* BIT_STRING_SIZE_4 */
static int hf_nbap_navFitFlag = -1;               /* BIT_STRING_SIZE_1 */
static int hf_nbap_navToe = -1;                   /* BIT_STRING_SIZE_16 */
static int hf_nbap_navOmega = -1;                 /* BIT_STRING_SIZE_32 */
static int hf_nbap_navDeltaN = -1;                /* BIT_STRING_SIZE_16 */
static int hf_nbap_navM0 = -1;                    /* BIT_STRING_SIZE_32 */
static int hf_nbap_navOmegaADot = -1;             /* BIT_STRING_SIZE_24 */
static int hf_nbap_navE = -1;                     /* BIT_STRING_SIZE_32 */
static int hf_nbap_navIDot = -1;                  /* BIT_STRING_SIZE_14 */
static int hf_nbap_navAPowerHalf = -1;            /* BIT_STRING_SIZE_32 */
static int hf_nbap_navI0 = -1;                    /* BIT_STRING_SIZE_32 */
static int hf_nbap_navOmegaA0 = -1;               /* BIT_STRING_SIZE_32 */
static int hf_nbap_navCrs = -1;                   /* BIT_STRING_SIZE_16 */
static int hf_nbap_navCis = -1;                   /* BIT_STRING_SIZE_16 */
static int hf_nbap_navCus = -1;                   /* BIT_STRING_SIZE_16 */
static int hf_nbap_navCrc = -1;                   /* BIT_STRING_SIZE_16 */
static int hf_nbap_navCic = -1;                   /* BIT_STRING_SIZE_16 */
static int hf_nbap_navCuc = -1;                   /* BIT_STRING_SIZE_16 */
static int hf_nbap_sbasTo = -1;                   /* BIT_STRING_SIZE_13 */
static int hf_nbap_sbasAccuracy = -1;             /* BIT_STRING_SIZE_4 */
static int hf_nbap_sbasXg = -1;                   /* BIT_STRING_SIZE_30 */
static int hf_nbap_sbasYg = -1;                   /* BIT_STRING_SIZE_30 */
static int hf_nbap_sbasZg = -1;                   /* BIT_STRING_SIZE_25 */
static int hf_nbap_sbasXgDot = -1;                /* BIT_STRING_SIZE_17 */
static int hf_nbap_sbasYgDot = -1;                /* BIT_STRING_SIZE_17 */
static int hf_nbap_sbasZgDot = -1;                /* BIT_STRING_SIZE_18 */
static int hf_nbap_sbasXgDotDot = -1;             /* BIT_STRING_SIZE_10 */
static int hf_nbap_sbagYgDotDot = -1;             /* BIT_STRING_SIZE_10 */
static int hf_nbap_sbasZgDotDot = -1;             /* BIT_STRING_SIZE_10 */
static int hf_nbap_gANSS_keplerianParameters_01 = -1;  /* GANSS_KeplerianParametersOrb */
static int hf_nbap_GANSS_Real_Time_Integrity_item = -1;  /* GANSS_RealTimeInformationItem */
static int hf_nbap_bad_ganss_satId = -1;          /* INTEGER_0_63 */
static int hf_nbap_bad_ganss_signalId = -1;       /* BIT_STRING_SIZE_8 */
static int hf_nbap_latitudeSign = -1;             /* T_latitudeSign */
static int hf_nbap_degreesOfLatitude = -1;        /* INTEGER_0_2147483647 */
static int hf_nbap_degreesOfLongitude = -1;       /* INTEGER_M2147483648_2147483647 */
static int hf_nbap_directionOfAltitude = -1;      /* T_directionOfAltitude */
static int hf_nbap_altitude = -1;                 /* INTEGER_0_32767 */
static int hf_nbap_t_oc = -1;                     /* BIT_STRING_SIZE_14 */
static int hf_nbap_a_i2 = -1;                     /* BIT_STRING_SIZE_12 */
static int hf_nbap_a_i1 = -1;                     /* BIT_STRING_SIZE_18 */
static int hf_nbap_a_i0 = -1;                     /* BIT_STRING_SIZE_28 */
static int hf_nbap_t_gd = -1;                     /* BIT_STRING_SIZE_10 */
static int hf_nbap_model_id = -1;                 /* INTEGER_0_1_ */
static int hf_nbap_GANSS_SatelliteInformationKP_item = -1;  /* GANSS_SatelliteInformationKPItem */
static int hf_nbap_ganss_e_alm = -1;              /* BIT_STRING_SIZE_11 */
static int hf_nbap_ganss_delta_I_alm = -1;        /* BIT_STRING_SIZE_11 */
static int hf_nbap_ganss_omegadot_alm = -1;       /* BIT_STRING_SIZE_11 */
static int hf_nbap_ganss_svhealth_alm = -1;       /* BIT_STRING_SIZE_4 */
static int hf_nbap_ganss_delta_a_sqrt_alm = -1;   /* BIT_STRING_SIZE_17 */
static int hf_nbap_ganss_omegazero_alm = -1;      /* BIT_STRING_SIZE_16 */
static int hf_nbap_ganss_m_zero_alm = -1;         /* BIT_STRING_SIZE_16 */
static int hf_nbap_ganss_omega_alm = -1;          /* BIT_STRING_SIZE_16 */
static int hf_nbap_ganss_af_zero_alm = -1;        /* BIT_STRING_SIZE_14 */
static int hf_nbap_ganss_af_one_alm = -1;         /* BIT_STRING_SIZE_11 */
static int hf_nbap_Ganss_Sat_Info_AddNavList_item = -1;  /* Ganss_Sat_Info_AddNavList_item */
static int hf_nbap_svHealth = -1;                 /* BIT_STRING_SIZE_6 */
static int hf_nbap_iod = -1;                      /* BIT_STRING_SIZE_11 */
static int hf_nbap_ganssAddClockModels = -1;      /* GANSS_AddClockModels */
static int hf_nbap_ganssAddOrbitModels = -1;      /* GANSS_AddOrbitModels */
static int hf_nbap_GANSS_SAT_Info_Almanac_GLOkpList_item = -1;  /* GANSS_SAT_Info_Almanac_GLOkp */
static int hf_nbap_gloAlmNA = -1;                 /* BIT_STRING_SIZE_11 */
static int hf_nbap_gloAlmnA = -1;                 /* BIT_STRING_SIZE_5 */
static int hf_nbap_gloAlmHA = -1;                 /* BIT_STRING_SIZE_5 */
static int hf_nbap_gloAlmLambdaA = -1;            /* BIT_STRING_SIZE_21 */
static int hf_nbap_gloAlmTlambdaA = -1;           /* BIT_STRING_SIZE_21 */
static int hf_nbap_gloAlmDeltaIA = -1;            /* BIT_STRING_SIZE_18 */
static int hf_nbap_gloAkmDeltaTA = -1;            /* BIT_STRING_SIZE_22 */
static int hf_nbap_gloAlmDeltaTdotA = -1;         /* BIT_STRING_SIZE_7 */
static int hf_nbap_gloAlmEpsilonA = -1;           /* BIT_STRING_SIZE_15 */
static int hf_nbap_gloAlmOmegaA = -1;             /* BIT_STRING_SIZE_16 */
static int hf_nbap_gloAlmTauA = -1;               /* BIT_STRING_SIZE_10 */
static int hf_nbap_gloAlmCA = -1;                 /* BIT_STRING_SIZE_1 */
static int hf_nbap_gloAlmMA = -1;                 /* BIT_STRING_SIZE_2 */
static int hf_nbap_GANSS_SAT_Info_Almanac_MIDIkpList_item = -1;  /* GANSS_SAT_Info_Almanac_MIDIkp */
static int hf_nbap_midiAlmE = -1;                 /* BIT_STRING_SIZE_11 */
static int hf_nbap_midiAlmDeltaI = -1;            /* BIT_STRING_SIZE_11 */
static int hf_nbap_midiAlmOmegaDot = -1;          /* BIT_STRING_SIZE_11 */
static int hf_nbap_midiAlmSqrtA = -1;             /* BIT_STRING_SIZE_17 */
static int hf_nbap_midiAlmOmega0 = -1;            /* BIT_STRING_SIZE_16 */
static int hf_nbap_midiAlmOmega = -1;             /* BIT_STRING_SIZE_16 */
static int hf_nbap_midiAlmMo = -1;                /* BIT_STRING_SIZE_16 */
static int hf_nbap_midiAlmaf0 = -1;               /* BIT_STRING_SIZE_11 */
static int hf_nbap_midiAlmaf1 = -1;               /* BIT_STRING_SIZE_10 */
static int hf_nbap_midiAlmL1Health = -1;          /* BIT_STRING_SIZE_1 */
static int hf_nbap_midiAlmL2Health = -1;          /* BIT_STRING_SIZE_1 */
static int hf_nbap_midiAlmL5Health = -1;          /* BIT_STRING_SIZE_1 */
static int hf_nbap_GANSS_SAT_Info_Almanac_NAVkpList_item = -1;  /* GANSS_SAT_Info_Almanac_NAVkp */
static int hf_nbap_navAlmE = -1;                  /* BIT_STRING_SIZE_16 */
static int hf_nbap_navAlmDeltaI = -1;             /* BIT_STRING_SIZE_16 */
static int hf_nbap_navAlmOMEGADOT = -1;           /* BIT_STRING_SIZE_16 */
static int hf_nbap_navAlmSVHealth = -1;           /* BIT_STRING_SIZE_8 */
static int hf_nbap_navAlmSqrtA = -1;              /* BIT_STRING_SIZE_24 */
static int hf_nbap_navAlmOMEGAo = -1;             /* BIT_STRING_SIZE_24 */
static int hf_nbap_navAlmOmega = -1;              /* BIT_STRING_SIZE_24 */
static int hf_nbap_navAlmMo = -1;                 /* BIT_STRING_SIZE_24 */
static int hf_nbap_navAlmaf0 = -1;                /* BIT_STRING_SIZE_11 */
static int hf_nbap_navAlmaf1 = -1;                /* BIT_STRING_SIZE_11 */
static int hf_nbap_GANSS_SAT_Info_Almanac_REDkpList_item = -1;  /* GANSS_SAT_Info_Almanac_REDkp */
static int hf_nbap_redAlmDeltaA = -1;             /* BIT_STRING_SIZE_8 */
static int hf_nbap_redAlmOmega0 = -1;             /* BIT_STRING_SIZE_7 */
static int hf_nbap_redAlmPhi0 = -1;               /* BIT_STRING_SIZE_7 */
static int hf_nbap_redAlmL1Health = -1;           /* BIT_STRING_SIZE_1 */
static int hf_nbap_redAlmL2Health = -1;           /* BIT_STRING_SIZE_1 */
static int hf_nbap_redAlmL5Health = -1;           /* BIT_STRING_SIZE_1 */
static int hf_nbap_GANSS_SAT_Info_Almanac_SBASecefList_item = -1;  /* GANSS_SAT_Info_Almanac_SBASecef */
static int hf_nbap_sbasAlmDataID = -1;            /* BIT_STRING_SIZE_2 */
static int hf_nbap_sbasAlmHealth = -1;            /* BIT_STRING_SIZE_8 */
static int hf_nbap_sbasAlmXg = -1;                /* BIT_STRING_SIZE_15 */
static int hf_nbap_sbasAlmYg = -1;                /* BIT_STRING_SIZE_15 */
static int hf_nbap_sbasAlmZg = -1;                /* BIT_STRING_SIZE_9 */
static int hf_nbap_sbasAlmXgdot = -1;             /* BIT_STRING_SIZE_3 */
static int hf_nbap_sbasAlmYgDot = -1;             /* BIT_STRING_SIZE_3 */
static int hf_nbap_sbasAlmZgDot = -1;             /* BIT_STRING_SIZE_4 */
static int hf_nbap_sbasAlmTo = -1;                /* BIT_STRING_SIZE_11 */
static int hf_nbap_GANSS_Sat_Info_Nav_item = -1;  /* GANSS_Sat_Info_Nav_item */
static int hf_nbap_svHealth_01 = -1;              /* BIT_STRING_SIZE_5 */
static int hf_nbap_iod_01 = -1;                   /* BIT_STRING_SIZE_10 */
static int hf_nbap_ganssClockModel = -1;          /* GANSS_Clock_Model */
static int hf_nbap_ganssOrbitModel = -1;          /* GANSS_Orbit_Model */
static int hf_nbap_sbasAgfo = -1;                 /* BIT_STRING_SIZE_12 */
static int hf_nbap_sbasAgf1 = -1;                 /* BIT_STRING_SIZE_8 */
static int hf_nbap_ganss_time_model_Ref_Time = -1;  /* INTEGER_0_37799 */
static int hf_nbap_ganss_t_a0 = -1;               /* INTEGER_M2147483648_2147483647 */
static int hf_nbap_ganss_t_a1 = -1;               /* INTEGER_M8388608_8388607 */
static int hf_nbap_ganss_t_a2 = -1;               /* INTEGER_M64_63 */
static int hf_nbap_gnss_to_id = -1;               /* T_gnss_to_id */
static int hf_nbap_ganss_wk_number_01 = -1;       /* INTEGER_0_8191 */
static int hf_nbap_ganssDay = -1;                 /* INTEGER_0_8191 */
static int hf_nbap_a_one_utc = -1;                /* BIT_STRING_SIZE_24 */
static int hf_nbap_a_zero_utc = -1;               /* BIT_STRING_SIZE_32 */
static int hf_nbap_t_ot_utc = -1;                 /* BIT_STRING_SIZE_8 */
static int hf_nbap_w_n_t_utc = -1;                /* BIT_STRING_SIZE_8 */
static int hf_nbap_delta_t_ls_utc = -1;           /* BIT_STRING_SIZE_8 */
static int hf_nbap_w_n_lsf_utc = -1;              /* BIT_STRING_SIZE_8 */
static int hf_nbap_dn_utc = -1;                   /* BIT_STRING_SIZE_8 */
static int hf_nbap_delta_t_lsf_utc = -1;          /* BIT_STRING_SIZE_8 */
static int hf_nbap_utcA0 = -1;                    /* BIT_STRING_SIZE_16 */
static int hf_nbap_utcA1 = -1;                    /* BIT_STRING_SIZE_13 */
static int hf_nbap_utcA2 = -1;                    /* BIT_STRING_SIZE_7 */
static int hf_nbap_utcDeltaTls = -1;              /* BIT_STRING_SIZE_8 */
static int hf_nbap_utcTot = -1;                   /* BIT_STRING_SIZE_16 */
static int hf_nbap_utcWNot = -1;                  /* BIT_STRING_SIZE_13 */
static int hf_nbap_utcWNlsf = -1;                 /* BIT_STRING_SIZE_8 */
static int hf_nbap_utcDN = -1;                    /* BIT_STRING_SIZE_4 */
static int hf_nbap_utcDeltaTlsf = -1;             /* BIT_STRING_SIZE_8 */
static int hf_nbap_nA = -1;                       /* BIT_STRING_SIZE_11 */
static int hf_nbap_tauC = -1;                     /* BIT_STRING_SIZE_32 */
static int hf_nbap_deltaUT1_01 = -1;              /* GANSS_DeltaUT1 */
static int hf_nbap_kp = -1;                       /* BIT_STRING_SIZE_2 */
static int hf_nbap_utcA1wnt = -1;                 /* BIT_STRING_SIZE_24 */
static int hf_nbap_utcA0wnt = -1;                 /* BIT_STRING_SIZE_32 */
static int hf_nbap_utcTot_01 = -1;                /* BIT_STRING_SIZE_8 */
static int hf_nbap_utcWNt = -1;                   /* BIT_STRING_SIZE_8 */
static int hf_nbap_utcDN_01 = -1;                 /* BIT_STRING_SIZE_8 */
static int hf_nbap_utcStandardID = -1;            /* BIT_STRING_SIZE_3 */
static int hf_nbap_wna_alm = -1;                  /* BIT_STRING_SIZE_8 */
static int hf_nbap_sat_info_almanac = -1;         /* SAT_Info_Almanac */
static int hf_nbap_sVGlobalHealth_alm = -1;       /* BIT_STRING_SIZE_364 */
static int hf_nbap_alpha_zero_ionos_01 = -1;      /* BIT_STRING_SIZE_8 */
static int hf_nbap_alpha_one_ionos_01 = -1;       /* BIT_STRING_SIZE_8 */
static int hf_nbap_alpha_two_ionos_01 = -1;       /* BIT_STRING_SIZE_8 */
static int hf_nbap_alpha_three_ionos = -1;        /* BIT_STRING_SIZE_8 */
static int hf_nbap_beta_zero_ionos = -1;          /* BIT_STRING_SIZE_8 */
static int hf_nbap_beta_one_ionos = -1;           /* BIT_STRING_SIZE_8 */
static int hf_nbap_beta_two_ionos = -1;           /* BIT_STRING_SIZE_8 */
static int hf_nbap_beta_three_ionos = -1;         /* BIT_STRING_SIZE_8 */
static int hf_nbap_GPS_Information_item = -1;     /* GPS_Information_Item */
static int hf_nbap_bad_satellites = -1;           /* GPSBadSat_Info_RealTime_Integrity */
static int hf_nbap_no_bad_satellites = -1;        /* NULL */
static int hf_nbap_sat_info = -1;                 /* SATInfo_RealTime_Integrity */
static int hf_nbap_GPS_NavigationModel_and_TimeRecovery_item = -1;  /* GPS_NavandRecovery_Item */
static int hf_nbap_tx_tow_nav = -1;               /* INTEGER_0_1048575 */
static int hf_nbap_sat_id_nav = -1;               /* SAT_ID */
static int hf_nbap_tlm_message_nav = -1;          /* BIT_STRING_SIZE_14 */
static int hf_nbap_tlm_revd_c_nav = -1;           /* BIT_STRING_SIZE_2 */
static int hf_nbap_ho_word_nav = -1;              /* BIT_STRING_SIZE_22 */
static int hf_nbap_w_n_nav = -1;                  /* BIT_STRING_SIZE_10 */
static int hf_nbap_ca_or_p_on_l2_nav = -1;        /* BIT_STRING_SIZE_2 */
static int hf_nbap_user_range_accuracy_index_nav = -1;  /* BIT_STRING_SIZE_4 */
static int hf_nbap_sv_health_nav = -1;            /* BIT_STRING_SIZE_6 */
static int hf_nbap_iodc_nav = -1;                 /* BIT_STRING_SIZE_10 */
static int hf_nbap_l2_p_dataflag_nav = -1;        /* BIT_STRING_SIZE_1 */
static int hf_nbap_sf1_reserved_nav = -1;         /* BIT_STRING_SIZE_87 */
static int hf_nbap_t_gd_nav = -1;                 /* BIT_STRING_SIZE_8 */
static int hf_nbap_t_oc_nav = -1;                 /* BIT_STRING_SIZE_16 */
static int hf_nbap_a_f_2_nav = -1;                /* BIT_STRING_SIZE_8 */
static int hf_nbap_a_f_1_nav = -1;                /* BIT_STRING_SIZE_16 */
static int hf_nbap_a_f_zero_nav = -1;             /* BIT_STRING_SIZE_22 */
static int hf_nbap_gps_e_nav = -1;                /* BIT_STRING_SIZE_32 */
static int hf_nbap_t_oe_nav = -1;                 /* BIT_STRING_SIZE_16 */
static int hf_nbap_fit_interval_flag_nav = -1;    /* BIT_STRING_SIZE_1 */
static int hf_nbap_aodo_nav = -1;                 /* BIT_STRING_SIZE_5 */
static int hf_nbap_gps_omega_nav = -1;            /* BIT_STRING_SIZE_32 */
static int hf_nbap_spare_zero_fill = -1;          /* BIT_STRING_SIZE_20 */
static int hf_nbap_latitudeSign_01 = -1;          /* T_latitudeSign_01 */
static int hf_nbap_latitude = -1;                 /* INTEGER_0_8388607 */
static int hf_nbap_longitude = -1;                /* INTEGER_M8388608_8388607 */
static int hf_nbap_directionOfAltitude_01 = -1;   /* T_directionOfAltitude_01 */
static int hf_nbap_implicit = -1;                 /* HARQ_MemoryPartitioning_Implicit */
static int hf_nbap_explicit = -1;                 /* HARQ_MemoryPartitioning_Explicit */
static int hf_nbap_number_of_Processes = -1;      /* INTEGER_1_8_ */
static int hf_nbap_hARQ_MemoryPartitioningList = -1;  /* HARQ_MemoryPartitioningList */
static int hf_nbap_HARQ_MemoryPartitioningList_item = -1;  /* HARQ_MemoryPartitioningItem */
static int hf_nbap_HARQ_MemoryPartitioningInfoExtForMIMO_item = -1;  /* HARQ_MemoryPartitioningItem */
static int hf_nbap_process_Memory_Size = -1;      /* T_process_Memory_Size */
static int hf_nbap_HS_DSCHProvidedBitRate_item = -1;  /* HS_DSCHProvidedBitRate_Item */
static int hf_nbap_hS_DSCHProvidedBitRateValue = -1;  /* HS_DSCHProvidedBitRateValue */
static int hf_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortion_item = -1;  /* HS_DSCHProvidedBitRateValueInformation_For_CellPortion_Item */
static int hf_nbap_hS_DSCHProvidedBitRateValue_01 = -1;  /* HS_DSCHProvidedBitRate */
static int hf_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR_item = -1;  /* HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR_Item */
static int hf_nbap_HS_DSCHRequiredPower_item = -1;  /* HS_DSCHRequiredPower_Item */
static int hf_nbap_hS_DSCHRequiredPowerValue = -1;  /* HS_DSCHRequiredPowerValue */
static int hf_nbap_hS_DSCHRequiredPowerPerUEInformation = -1;  /* HS_DSCHRequiredPowerPerUEInformation */
static int hf_nbap_HS_DSCHRequiredPowerPerUEInformation_item = -1;  /* HS_DSCHRequiredPowerPerUEInformation_Item */
static int hf_nbap_cRNC_CommunicationContextID = -1;  /* CRNC_CommunicationContextID */
static int hf_nbap_hS_DSCHRequiredPowerPerUEWeight = -1;  /* HS_DSCHRequiredPowerPerUEWeight */
static int hf_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortion_item = -1;  /* HS_DSCHRequiredPowerValueInformation_For_CellPortion_Item */
static int hf_nbap_hS_DSCHRequiredPowerValue_01 = -1;  /* HS_DSCHRequiredPower */
static int hf_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR_item = -1;  /* HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR_Item */
static int hf_nbap_hsdpa_PICH_Shared_with_PCH = -1;  /* HSDPA_PICH_Shared_with_PCH */
static int hf_nbap_hsdpa_PICH_notShared_with_PCH = -1;  /* HSDPA_PICH_notShared_with_PCH */
static int hf_nbap_hsdpa_PICH_SharedPCH_ID = -1;  /* CommonPhysicalChannelID */
static int hf_nbap_hSDPA_PICH_notShared_ID = -1;  /* CommonPhysicalChannelID */
static int hf_nbap_fdd_DL_Channelisation_CodeNumber = -1;  /* FDD_DL_ChannelisationCodeNumber */
static int hf_nbap_pich_Power = -1;               /* PICH_Power */
static int hf_nbap_pich_Mode = -1;                /* PICH_Mode */
static int hf_nbap_sttd_Indicator = -1;           /* STTD_Indicator */
static int hf_nbap_hsdsch_Common_Information = -1;  /* HSDSCH_Common_Information */
static int hf_nbap_commonMACFlow_Specific_Information = -1;  /* CommonMACFlow_Specific_InfoList */
static int hf_nbap_hsSCCH_Specific_Information_ResponseFDD = -1;  /* HSSCCH_Specific_InformationRespListFDD */
static int hf_nbap_hARQ_MemoryPartitioning = -1;  /* HARQ_MemoryPartitioning */
static int hf_nbap_commonMACFlow_Specific_Info_Response = -1;  /* CommonMACFlow_Specific_InfoList_Response */
static int hf_nbap_cCCH_PriorityQueue_Id = -1;    /* PriorityQueue_Id */
static int hf_nbap_sRB1_PriorityQueue_Id = -1;    /* PriorityQueue_Id */
static int hf_nbap_associatedCommon_MACFlow = -1;  /* Common_MACFlow_ID */
static int hf_nbap_fACH_Measurement_Occasion_Cycle_Length_Coefficient = -1;  /* FACH_Measurement_Occasion_Cycle_Length_Coefficient */
static int hf_nbap_rACH_Measurement_Result = -1;  /* RACH_Measurement_Result */
static int hf_nbap_bCCH_Specific_HSDSCH_RNTI_Information = -1;  /* BCCH_Specific_HSDSCH_RNTI_Information */
static int hf_nbap_hSDSCH_MACdFlows_Information = -1;  /* HSDSCH_MACdFlows_Information */
static int hf_nbap_ueCapability_Info = -1;        /* UE_Capability_Information */
static int hf_nbap_mAChs_Reordering_Buffer_Size_for_RLC_UM = -1;  /* MAChsReorderingBufferSize_for_RLC_UM */
static int hf_nbap_hsscch_PowerOffset = -1;       /* HSSCCH_PowerOffset */
static int hf_nbap_tDD_AckNack_Power_Offset = -1;  /* TDD_AckNack_Power_Offset */
static int hf_nbap_hsDSCH_MACdFlow_Specific_Info_to_Modify = -1;  /* HSDSCH_MACdFlow_Specific_InfoList_to_Modify */
static int hf_nbap_priorityQueueInfotoModify = -1;  /* PriorityQueue_InfoList_to_Modify */
static int hf_nbap_hSSCCHCodeChangeGrant = -1;    /* HSSCCH_Code_Change_Grant */
static int hf_nbap_tDDAckNackPowerOffset = -1;    /* TDD_AckNack_Power_Offset */
static int hf_nbap_HSDSCH_MACdFlow_Specific_InfoList_to_Modify_item = -1;  /* HSDSCH_MACdFlow_Specific_InfoItem_to_Modify */
static int hf_nbap_hsDSCH_MACdFlow_ID = -1;       /* HSDSCH_MACdFlow_ID */
static int hf_nbap_priorityQueueInfotoModifyUnsynchronised = -1;  /* PriorityQueue_InfoList_to_Modify_Unsynchronised */
static int hf_nbap_hsDSCH_MACdFlow_Specific_InformationResp = -1;  /* HSDSCH_MACdFlow_Specific_InformationResp */
static int hf_nbap_sixtyfourQAM_UsageAllowedIndicator = -1;  /* SixtyfourQAM_UsageAllowedIndicator */
static int hf_nbap_hSDSCH_RNTI = -1;              /* HSDSCH_RNTI */
static int hf_nbap_sixtyfourQAM_DL_UsageIndicator = -1;  /* SixtyfourQAM_DL_UsageIndicator */
static int hf_nbap_hSDSCH_TBSizeTableIndicator = -1;  /* HSDSCH_TBSizeTableIndicator */
static int hf_nbap_hSSCCH_CodeChangeGrant = -1;   /* HSSCCH_Code_Change_Grant */
static int hf_nbap_hsSCCHCodeChangeIndicator = -1;  /* HSSCCH_CodeChangeIndicator */
static int hf_nbap_hS_PDSCH_Code_Change_Indicator = -1;  /* HS_PDSCH_Code_Change_Indicator */
static int hf_nbap_hS_DSCH_Secondary_Serving_cell_choice = -1;  /* HS_DSCH_Secondary_Serving_cell_change_choice */
static int hf_nbap_hS_Secondary_Serving_cell_change_successful = -1;  /* HS_Secondary_Serving_cell_change_successful */
static int hf_nbap_hS_Secondary_Serving_cell_change_unsuccessful = -1;  /* HS_Secondary_Serving_cell_change_unsuccessful */
static int hf_nbap_hS_DSCH_FDD_Secondary_Serving_Information_Response = -1;  /* HS_DSCH_FDD_Secondary_Serving_Information_Response */
static int hf_nbap_paging_MACFlow_Specific_Information = -1;  /* Paging_MACFlow_Specific_Information */
static int hf_nbap_number_of_PCCH_transmission = -1;  /* Number_of_PCCH_transmission */
static int hf_nbap_transport_Block_Size_List = -1;  /* Transport_Block_Size_List */
static int hf_nbap_HSDSCH_Paging_System_Information_ResponseFDD_item = -1;  /* HSDSCH_Paging_System_Information_ResponseList */
static int hf_nbap_pagingMACFlow_ID = -1;         /* Paging_MACFlow_ID */
static int hf_nbap_hSPDSCH_Code_Index = -1;       /* HSPDSCH_Code_Index */
static int hf_nbap_hsSCCH_Specific_Information_ResponseTDD = -1;  /* HSSCCH_Specific_InformationRespListTDD */
static int hf_nbap_hsSCCH_Specific_Information_ResponseTDDLCR = -1;  /* HSSCCH_Specific_InformationRespListTDDLCR */
static int hf_nbap_HSDSCH_MACdFlow_Specific_InformationResp_item = -1;  /* HSDSCH_MACdFlow_Specific_InformationResp_Item */
static int hf_nbap_hsDSCHMacdFlow_Id = -1;        /* HSDSCH_MACdFlow_ID */
static int hf_nbap_hSDSCH_MACdFlow_Specific_Info = -1;  /* HSDSCH_MACdFlow_Specific_InfoList */
static int hf_nbap_priorityQueue_Info = -1;       /* PriorityQueue_InfoList */
static int hf_nbap_HSDSCH_MACdFlow_Specific_InfoList_item = -1;  /* HSDSCH_MACdFlow_Specific_InfoItem */
static int hf_nbap_HSDSCH_MACdFlows_to_Delete_item = -1;  /* HSDSCH_MACdFlows_to_Delete_Item */
static int hf_nbap_HSDSCH_Initial_Capacity_Allocation_item = -1;  /* HSDSCH_Initial_Capacity_AllocationItem */
static int hf_nbap_maximum_MACdPDU_Size = -1;     /* MACdPDU_Size */
static int hf_nbap_hSDSCH_InitialWindowSize = -1;  /* HSDSCH_InitialWindowSize */
static int hf_nbap_setsOfHS_SCCH_Codes = -1;      /* SetsOfHS_SCCH_Codes */
static int hf_nbap_hARQ_Preamble_Mode_Activation_Indicator = -1;  /* HARQ_Preamble_Mode_Activation_Indicator */
static int hf_nbap_mIMO_N_M_Ratio = -1;           /* MIMO_N_M_Ratio */
static int hf_nbap_continuousPacketConnectivityHS_SCCH_less_Information_Response = -1;  /* ContinuousPacketConnectivityHS_SCCH_less_Information_Response */
static int hf_nbap_Additional_EDCH_Preconfiguration_Information_item = -1;  /* Additional_EDCH_Preconfiguration_Information_ItemIEs */
static int hf_nbap_mAChsResetScheme = -1;         /* MAChsResetScheme */
static int hf_nbap_hSDSCH_Physical_Layer_Category = -1;  /* INTEGER_1_64_ */
static int hf_nbap_secondaryServingCells = -1;    /* SecondaryServingCells */
static int hf_nbap_numPrimaryHS_SCCH_Codes = -1;  /* NumHS_SCCH_Codes */
static int hf_nbap_hARQ_Preamble_Mode = -1;       /* HARQ_Preamble_Mode */
static int hf_nbap_mIMO_ActivationIndicator = -1;  /* MIMO_ActivationIndicator */
static int hf_nbap_hSDSCH_MACdPDUSizeFormat = -1;  /* HSDSCH_MACdPDUSizeFormat */
static int hf_nbap_uE_with_enhanced_HS_SCCH_support_indicator = -1;  /* NULL */
static int hf_nbap_HS_SCCH_PreconfiguredCodes_item = -1;  /* HS_SCCH_PreconfiguredCodesItem */
static int hf_nbap_hS_SCCH_CodeNumber = -1;       /* HS_SCCH_CodeNumber */
static int hf_nbap_HSSCCH_Specific_InformationRespListFDD_item = -1;  /* HSSCCH_Codes */
static int hf_nbap_codeNumber = -1;               /* INTEGER_0_127 */
static int hf_nbap_HSSCCH_Specific_InformationRespListTDD_item = -1;  /* HSSCCH_Specific_InformationRespItemTDD */
static int hf_nbap_timeslot = -1;                 /* TimeSlot */
static int hf_nbap_tDD_ChannelisationCode = -1;   /* TDD_ChannelisationCode */
static int hf_nbap_hSSICH_Info = -1;              /* HSSICH_Info */
static int hf_nbap_HSSCCH_Specific_InformationRespListTDDLCR_item = -1;  /* HSSCCH_Specific_InformationRespItemTDDLCR */
static int hf_nbap_timeslotLCR = -1;              /* TimeSlotLCR */
static int hf_nbap_first_TDD_ChannelisationCode = -1;  /* TDD_ChannelisationCode */
static int hf_nbap_second_TDD_ChannelisationCode = -1;  /* TDD_ChannelisationCode */
static int hf_nbap_hSSICH_InfoLCR = -1;           /* HSSICH_InfoLCR */
static int hf_nbap_HSSCCH_Specific_InformationRespListTDD768_item = -1;  /* HSSCCH_Specific_InformationRespItemTDD768 */
static int hf_nbap_tDD_ChannelisationCode768 = -1;  /* TDD_ChannelisationCode768 */
static int hf_nbap_hSSICH_Info768 = -1;           /* HSSICH_Info768 */
static int hf_nbap_hsSICH_ID = -1;                /* HS_SICH_ID */
static int hf_nbap_failed_HS_SICH = -1;           /* HS_SICH_failed */
static int hf_nbap_missed_HS_SICH = -1;           /* HS_SICH_missed */
static int hf_nbap_total_HS_SICH = -1;            /* HS_SICH_total */
static int hf_nbap_number_of_HS_PDSCH_codes = -1;  /* INTEGER_0_maxHS_PDSCHCodeNrComp_1 */
static int hf_nbap_hS_PDSCH_Start_code_number = -1;  /* HS_PDSCH_Start_code_number */
static int hf_nbap_replace_02 = -1;               /* HS_SCCH_FDD_Code_List */
static int hf_nbap_HS_SCCH_FDD_Code_List_item = -1;  /* HS_SCCH_FDD_Code_Information_Item */
static int hf_nbap_hspdsch_RL_ID = -1;            /* RL_ID */
static int hf_nbap_hSDSCH_FDD_Information = -1;   /* HSDSCH_FDD_Information */
static int hf_nbap_hsdsch_RNTI = -1;              /* HSDSCH_RNTI */
static int hf_nbap_hS_DSCH_serving_cell_choice = -1;  /* HS_DSCH_serving_cell_choice */
static int hf_nbap_hS_serving_cell_change_successful = -1;  /* HS_serving_cell_change_successful */
static int hf_nbap_hS_serving_cell_change_unsuccessful = -1;  /* HS_serving_cell_change_unsuccessful */
static int hf_nbap_hSDSCH_FDD_Information_Response = -1;  /* HSDSCH_FDD_Information_Response */
static int hf_nbap_hsdpa_PICH_notShared_with_PCHLCR = -1;  /* HSDPA_PICH_notShared_with_PCHLCR */
static int hf_nbap_tdd_ChannelisationCodeLCR = -1;  /* TDD_ChannelisationCodeLCR */
static int hf_nbap_tdd_PhysicalChannelOffset = -1;  /* TDD_PhysicalChannelOffset */
static int hf_nbap_pagingIndicatorLength = -1;    /* PagingIndicatorLength */
static int hf_nbap_pICH_Power = -1;               /* PICH_Power */
static int hf_nbap_second_TDD_ChannelisationCodeLCR = -1;  /* TDD_ChannelisationCodeLCR */
static int hf_nbap_hsdsch_Common_InformationLCR = -1;  /* HSDSCH_Common_InformationLCR */
static int hf_nbap_commonMACFlow_Specific_InformationLCR = -1;  /* CommonMACFlow_Specific_InfoListLCR */
static int hf_nbap_common_H_RNTI_InformationLCR = -1;  /* Common_H_RNTI_InformationLCR */
static int hf_nbap_sync_InformationLCR = -1;      /* Sync_InformationLCR */
static int hf_nbap_hSSICH_SIRTarget = -1;         /* UL_SIR */
static int hf_nbap_hSSICH_TPC_StepSize = -1;      /* TDD_TPC_UplinkStepSize_LCR */
static int hf_nbap_hsSCCH_Specific_Information_ResponseLCR = -1;  /* HSSCCH_Specific_InformationRespListLCR */
static int hf_nbap_commonMACFlow_Specific_Info_ResponseLCR = -1;  /* CommonMACFlow_Specific_InfoList_ResponseLCR */
static int hf_nbap_associatedCommon_MACFlowLCR = -1;  /* Common_MACFlow_ID_LCR */
static int hf_nbap_bCCH_Specific_HSDSCH_RNTI_InformationLCR = -1;  /* BCCH_Specific_HSDSCH_RNTI_InformationLCR */
static int hf_nbap_paging_MACFlow_Specific_InformationLCR = -1;  /* Paging_MACFlow_Specific_InformationLCR */
static int hf_nbap_reception_Window_Size = -1;    /* INTEGER_1_16 */
static int hf_nbap_n_PCH = -1;                    /* INTEGER_1_8 */
static int hf_nbap_paging_Subchannel_Size = -1;   /* INTEGER_1_3 */
static int hf_nbap_HSDSCH_Paging_System_Information_ResponseLCR_item = -1;  /* HSDSCH_Paging_System_Information_ResponseListLCR */
static int hf_nbap_dL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst = -1;  /* DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst */
static int hf_nbap_HSSCCH_Specific_InformationRespListLCR_item = -1;  /* HSSCCH_Specific_InformationRespItemLCR */
static int hf_nbap_hS_SCCH_ID_LCR = -1;           /* HS_SCCH_ID_LCR */
static int hf_nbap_transport_Block_Size_List_01 = -1;  /* Transport_Block_Size_List_LCR */
static int hf_nbap_hS_DSCH_SPS_Reservation_Indicator = -1;  /* SPS_Reservation_Indicator */
static int hf_nbap_hS_DSCH_SPS_Operation_Indicator = -1;  /* HS_DSCH_SPS_Operation_Indicator */
static int hf_nbap_Transport_Block_Size_List_LCR_item = -1;  /* Transport_Block_Size_Item_LCR */
static int hf_nbap_transport_Block_Size_maping_Index_LCR = -1;  /* Transport_Block_Size_maping_Index_LCR */
static int hf_nbap_transport_Block_Size_Index_LCR = -1;  /* Transport_Block_Size_Index_LCR */
static int hf_nbap_Repetition_Period_List_LCR_item = -1;  /* Repetition_Period_Item_LCR */
static int hf_nbap_repetitionPeriodIndex = -1;    /* RepetitionPeriodIndex */
static int hf_nbap_logicalChannellevel = -1;      /* LogicalChannellevel */
static int hf_nbap_priorityQueuelevel = -1;       /* PriorityQueuelevel */
static int hf_nbap_hS_SICH_InformationList_for_HS_DSCH_SPS = -1;  /* HS_SICH_InformationList_for_HS_DSCH_SPS */
static int hf_nbap_initial_HS_PDSCH_SPS_Resource = -1;  /* Initial_HS_PDSCH_SPS_Resource */
static int hf_nbap_buffer_Size_for_HS_DSCH_SPS = -1;  /* Process_Memory_Size */
static int hf_nbap_number_of_Processes_for_HS_DSCH_SPS = -1;  /* Number_of_Processes_for_HS_DSCH_SPS */
static int hf_nbap_HS_SICH_InformationList_for_HS_DSCH_SPS_item = -1;  /* HS_SICH_InformationItem_for_HS_DSCH_SPS */
static int hf_nbap_hS_SICH_Mapping_Index = -1;    /* HS_SICH_Mapping_Index */
static int hf_nbap_hS_SICH_Type = -1;             /* HS_SICH_Type */
static int hf_nbap_hS_SCCH_Associated_HS_SICH = -1;  /* HS_SCCH_Associated_HS_SICH */
static int hf_nbap_non_HS_SCCH_Associated_HS_SICH = -1;  /* Non_HS_SCCH_Associated_HS_SICH */
static int hf_nbap_extended_HS_SICH_ID = -1;      /* Extended_HS_SICH_ID */
static int hf_nbap_non_HS_SCCH_Aassociated_HS_SICH_ID = -1;  /* Non_HS_SCCH_Aassociated_HS_SICH_ID */
static int hf_nbap_hS_PDSCH_Offset = -1;          /* TDD_PhysicalChannelOffset */
static int hf_nbap_timeslot_Resource_Related_Information_01 = -1;  /* HS_DSCH_TimeslotResourceLCR */
static int hf_nbap_startCode = -1;                /* TDD_ChannelisationCode */
static int hf_nbap_endCode = -1;                  /* TDD_ChannelisationCode */
static int hf_nbap_transport_Block_Size_Index_01 = -1;  /* Transport_Block_Size_Index_LCR */
static int hf_nbap_modulationType = -1;           /* ModulationSPS_LCR */
static int hf_nbap_non_HS_SCCH_Associated_HS_SICH_InformationList = -1;  /* Non_HS_SCCH_Associated_HS_SICH_InformationList */
static int hf_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationList_item = -1;  /* Non_HS_SCCH_Associated_HS_SICH_InformationItem */
static int hf_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext_item = -1;  /* Non_HS_SCCH_Associated_HS_SICH_InformationItem */
static int hf_nbap_modify_non_HS_SCCH_Associated_HS_SICH_InformationList = -1;  /* Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList */
static int hf_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList_item = -1;  /* Modify_Non_HS_SCCH_Associated_HS_SICH_InformationItem */
static int hf_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext_item = -1;  /* Modify_Non_HS_SCCH_Associated_HS_SICH_InformationItem */
static int hf_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_item = -1;  /* Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqstItem */
static int hf_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext_item = -1;  /* Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqstItem */
static int hf_nbap_MIMO_ReferenceSignal_InformationListLCR_item = -1;  /* HSSICH_ReferenceSignal_InformationLCR */
static int hf_nbap_midambleConfigurationLCR = -1;  /* MidambleConfigurationLCR */
static int hf_nbap_midambleShift = -1;            /* INTEGER_0_15 */
static int hf_nbap_hSSICH_ReferenceSignal_InformationLCR = -1;  /* HSSICH_ReferenceSignal_InformationLCR */
static int hf_nbap_sub_Frame_Number = -1;         /* Sub_Frame_Number */
static int hf_nbap_onDemand = -1;                 /* NULL */
static int hf_nbap_periodic = -1;                 /* InformationReportCharacteristicsType_ReportPeriodicity */
static int hf_nbap_onModification = -1;           /* InformationReportCharacteristicsType_OnModification */
static int hf_nbap_min = -1;                      /* ReportPeriodicity_Scaledmin */
static int hf_nbap_hours = -1;                    /* ReportPeriodicity_Scaledhour */
static int hf_nbap_information_thresholds = -1;   /* InformationThresholds */
static int hf_nbap_dgps = -1;                     /* DGPSThresholds */
static int hf_nbap_dGANSSThreshold = -1;          /* DGANSSThreshold */
static int hf_nbap_information_Type_Item = -1;    /* Information_Type_Item */
static int hf_nbap_gPSInformation = -1;           /* GPS_Information */
static int hf_nbap_iP_SpacingFDD = -1;            /* T_iP_SpacingFDD */
static int hf_nbap_iP_Length = -1;                /* T_iP_Length */
static int hf_nbap_seed = -1;                     /* INTEGER_0_63 */
static int hf_nbap_burstModeParams = -1;          /* BurstModeParams */
static int hf_nbap_iP_Offset = -1;                /* INTEGER_0_9 */
static int hf_nbap_iP_SpacingTDD = -1;            /* T_iP_SpacingTDD */
static int hf_nbap_iP_Start = -1;                 /* INTEGER_0_4095 */
static int hf_nbap_iP_Slot = -1;                  /* INTEGER_0_14 */
static int hf_nbap_iP_PCCPCH = -1;                /* T_iP_PCCPCH */
static int hf_nbap_iP_SpacingTDD_01 = -1;         /* T_iP_SpacingTDD_01 */
static int hf_nbap_iP_Sub = -1;                   /* T_iP_Sub */
static int hf_nbap_cFNOffset = -1;                /* INTEGER_0_255 */
static int hf_nbap_burstStart = -1;               /* INTEGER_0_15 */
static int hf_nbap_burstLength = -1;              /* INTEGER_10_25 */
static int hf_nbap_burstFreq = -1;                /* INTEGER_1_16 */
static int hf_nbap_idleIntervalInfo_k = -1;       /* T_idleIntervalInfo_k */
static int hf_nbap_idleIntervalInfo_offset = -1;  /* INTEGER_0_7 */
static int hf_nbap_maxTimeslotsPerSubFrame = -1;  /* INTEGER_1_6 */
static int hf_nbap_maxPhysChPerTimeslot = -1;     /* T_maxPhysChPerTimeslot */
static int hf_nbap_MACdPDU_Size_Indexlist_item = -1;  /* MACdPDU_Size_IndexItem */
static int hf_nbap_sID = -1;                      /* SID */
static int hf_nbap_macdPDU_Size = -1;             /* MACdPDU_Size */
static int hf_nbap_MACdPDU_Size_Indexlist_to_Modify_item = -1;  /* MACdPDU_Size_IndexItem_to_Modify */
static int hf_nbap_maximum_TB_Size_cell_edge_users = -1;  /* INTEGER_0_5000_ */
static int hf_nbap_maximum_TB_Size_other_users = -1;  /* INTEGER_0_5000_ */
static int hf_nbap_MessageStructure_item = -1;    /* MessageStructure_item */
static int hf_nbap_repetitionNumber_01 = -1;      /* RepetitionNumber1 */
static int hf_nbap_type1 = -1;                    /* Type1 */
static int hf_nbap_midambleConfigurationBurstType1And3 = -1;  /* MidambleConfigurationBurstType1And3 */
static int hf_nbap_midambleAllocationMode = -1;   /* MidambleAllocationMode1 */
static int hf_nbap_defaultMidamble = -1;          /* NULL */
static int hf_nbap_commonMidamble = -1;           /* NULL */
static int hf_nbap_ueSpecificMidamble = -1;       /* MidambleShiftLong */
static int hf_nbap_type2 = -1;                    /* Type2 */
static int hf_nbap_midambleConfigurationBurstType2 = -1;  /* MidambleConfigurationBurstType2 */
static int hf_nbap_midambleAllocationMode_01 = -1;  /* MidambleAllocationMode2 */
static int hf_nbap_ueSpecificMidamble_01 = -1;    /* MidambleShiftShort */
static int hf_nbap_type3 = -1;                    /* Type3 */
static int hf_nbap_midambleAllocationMode_02 = -1;  /* MidambleAllocationMode3 */
static int hf_nbap_midambleAllocationMode_03 = -1;  /* MidambleAllocationMode */
static int hf_nbap_midambleShift_01 = -1;         /* MidambleShiftLong */
static int hf_nbap_type1_01 = -1;                 /* Type7681 */
static int hf_nbap_midambleAllocationMode_04 = -1;  /* MidambleAllocationMode7681 */
static int hf_nbap_type2_01 = -1;                 /* Type7682 */
static int hf_nbap_midambleConfigurationBurstType2_768 = -1;  /* MidambleConfigurationBurstType2_768 */
static int hf_nbap_midambleAllocationMode_05 = -1;  /* MidambleAllocationMode7682 */
static int hf_nbap_ueSpecificMidamble_02 = -1;    /* MidambleShiftShort768 */
static int hf_nbap_type3_01 = -1;                 /* Type7683 */
static int hf_nbap_midambleAllocationMode_06 = -1;  /* MidambleAllocationMode7683 */
static int hf_nbap_primary_and_secondary_CPICH = -1;  /* CommonPhysicalChannelID */
static int hf_nbap_normal_and_diversity_primary_CPICH = -1;  /* NULL */
static int hf_nbap_primary_and_secondary_CPICH_01 = -1;  /* PrimaryAndSecondaryCPICHContainer */
static int hf_nbap_normal_and_diversity_primary_CPICH_01 = -1;  /* NormalAndDiversityPrimaryCPICHContainer */
static int hf_nbap_addPriorityQueue = -1;         /* PriorityQueue_InfoItem_to_Add */
static int hf_nbap_modifyPriorityQueue = -1;      /* PriorityQueue_InfoItem_to_Modify */
static int hf_nbap_deletePriorityQueue = -1;      /* PriorityQueue_Id */
static int hf_nbap_qPSK = -1;                     /* NULL */
static int hf_nbap_sixteenQAM = -1;               /* MBSFN_CPICH_secondary_CCPCH_power_offset */
static int hf_nbap_MultipleFreq_HARQ_MemoryPartitioning_InformationList_item = -1;  /* MultipleFreq_HARQ_MemoryPartitioning_InformationItem */
static int hf_nbap_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR_item = -1;  /* MultipleFreq_HSPDSCH_InformationItem_ResponseTDDLCR */
static int hf_nbap_multi_Cell_Capability = -1;    /* Multi_Cell_Capability */
static int hf_nbap_dL_PowerBalancing_Information = -1;  /* DL_PowerBalancing_Information */
static int hf_nbap_minimumReducedE_DPDCH_GainFactor = -1;  /* MinimumReducedE_DPDCH_GainFactor */
static int hf_nbap_secondary_UL_Frequency_Activation_State = -1;  /* Secondary_UL_Frequency_Activation_State */
static int hf_nbap_extendedPropagationDelay = -1;  /* ExtendedPropagationDelay */
static int hf_nbap_primary_CPICH_Usage_for_Channel_Estimation = -1;  /* Primary_CPICH_Usage_for_Channel_Estimation */
static int hf_nbap_secondary_CPICH_Information = -1;  /* CommonPhysicalChannelID */
static int hf_nbap_secondary_CPICH_Information_Change = -1;  /* Secondary_CPICH_Information_Change */
static int hf_nbap_e_AGCH_PowerOffset = -1;       /* E_AGCH_PowerOffset */
static int hf_nbap_e_RGCH_PowerOffset = -1;       /* E_RGCH_PowerOffset */
static int hf_nbap_e_HICH_PowerOffset = -1;       /* E_HICH_PowerOffset */
static int hf_nbap_e_DCH_DL_Control_Channel_Grant = -1;  /* NULL */
static int hf_nbap_NeighbouringCellMeasurementInformation_item = -1;  /* NeighbouringCellMeasurementInformation_item */
static int hf_nbap_neighbouringFDDCellMeasurementInformation = -1;  /* NeighbouringFDDCellMeasurementInformation */
static int hf_nbap_neighbouringTDDCellMeasurementInformation = -1;  /* NeighbouringTDDCellMeasurementInformation */
static int hf_nbap_extension_neighbouringCellMeasurementInformation = -1;  /* Extension_neighbouringCellMeasurementInformation */
static int hf_nbap_uC_Id = -1;                    /* UC_Id */
static int hf_nbap_primaryScramblingCode = -1;    /* PrimaryScramblingCode */
static int hf_nbap_cellParameterID = -1;          /* CellParameterID */
static int hf_nbap_NI_Information_item = -1;      /* Notification_Indicator */
static int hf_nbap_Paging_MACFlows_to_DeleteFDD_item = -1;  /* Paging_MACFlows_to_DeleteFDD_Item */
static int hf_nbap_paging_MACFlow_ID = -1;        /* Paging_MACFlow_ID */
static int hf_nbap_Paging_MACFlow_Specific_Information_item = -1;  /* Paging_MAC_Flow_Specific_Information_Item */
static int hf_nbap_paging_MACFlow_Id = -1;        /* Paging_MACFlow_ID */
static int hf_nbap_hSDPA_associated_PICH_Info = -1;  /* HSDPA_Associated_PICH_Information */
static int hf_nbap_paging_MACFlow_PriorityQueue_Information = -1;  /* Paging_MACFlow_PriorityQueue_Information */
static int hf_nbap_Paging_MACFlow_PriorityQueue_Information_item = -1;  /* Paging_MACFlow_PriorityQueue_Item */
static int hf_nbap_priority_Queue_Information_for_Enhanced_PCH = -1;  /* Priority_Queue_Information_for_Enhanced_FACH_PCH */
static int hf_nbap_sequenceNumber = -1;           /* PLCCHsequenceNumber */
static int hf_nbap_Possible_Secondary_Serving_Cell_List_item = -1;  /* Possible_Secondary_Serving_Cell */
static int hf_nbap_local_Cell_ID = -1;            /* Local_Cell_ID */
static int hf_nbap_power_Offset_For_Secondary_CPICH_for_MIMO = -1;  /* PowerOffsetForSecondaryCPICHforMIMO */
static int hf_nbap_priorityQueue_Id = -1;         /* PriorityQueue_Id */
static int hf_nbap_t1 = -1;                       /* T1 */
static int hf_nbap_mAC_ehs_Reset_Timer = -1;      /* MAC_ehs_Reset_Timer */
static int hf_nbap_discardTimer = -1;             /* DiscardTimer */
static int hf_nbap_mAC_hsWindowSize = -1;         /* MAC_hsWindowSize */
static int hf_nbap_maximum_MACcPDU_Size = -1;     /* MAC_PDU_SizeExtended */
static int hf_nbap_PriorityQueue_InfoList_item = -1;  /* PriorityQueue_InfoItem */
static int hf_nbap_priorityQueueId = -1;          /* PriorityQueue_Id */
static int hf_nbap_associatedHSDSCH_MACdFlow = -1;  /* HSDSCH_MACdFlow_ID */
static int hf_nbap_mAChsGuaranteedBitRate = -1;   /* MAChsGuaranteedBitRate */
static int hf_nbap_macdPDU_Size_Index = -1;       /* MACdPDU_Size_Indexlist */
static int hf_nbap_rLC_Mode = -1;                 /* RLC_Mode */
static int hf_nbap_PriorityQueue_InfoList_to_Modify_item = -1;  /* ModifyPriorityQueue */
static int hf_nbap_macdPDU_Size_Index_to_Modify = -1;  /* MACdPDU_Size_Indexlist_to_Modify */
static int hf_nbap_PriorityQueue_InfoList_to_Modify_Unsynchronised_item = -1;  /* PriorityQueue_InfoItem_to_Modify_Unsynchronised */
static int hf_nbap_none = -1;                     /* NULL */
static int hf_nbap_selected_MBMS_Service = -1;    /* Selected_MBMS_Service */
static int hf_nbap_selected_MBMS_Service_List = -1;  /* Selected_MBMS_Service_List */
static int hf_nbap_Selected_MBMS_Service_List_item = -1;  /* Selected_MBMS_Service_Item */
static int hf_nbap_selected_MBMS_Service_TimeSlot_Information_LCR = -1;  /* Selected_MBMS_Service_TimeSlot_Information_LCR */
static int hf_nbap_mBMS_Service_TDM_Information = -1;  /* MBMS_Service_TDM_Information */
static int hf_nbap_Selected_MBMS_Service_TimeSlot_Information_LCR_item = -1;  /* TimeSlotLCR */
static int hf_nbap_transmission_Time_Interval = -1;  /* T_transmission_Time_Interval */
static int hf_nbap_tDM_Rep = -1;                  /* INTEGER_2_9 */
static int hf_nbap_tDM_Offset = -1;               /* INTEGER_0_8 */
static int hf_nbap_tDM_Length = -1;               /* INTEGER_1_8 */
static int hf_nbap_Paging_MACFlow_Specific_InformationLCR_item = -1;  /* Paging_MAC_Flow_Specific_Information_ItemLCR */
static int hf_nbap_hSDPA_associated_PICH_InfoLCR = -1;  /* HSDPA_Associated_PICH_InformationLCR */
static int hf_nbap_paging_MACFlow_PriorityQueue_InformationLCR = -1;  /* Paging_MACFlow_PriorityQueue_Information */
static int hf_nbap_Paging_MACFlows_to_DeleteLCR_item = -1;  /* Paging_MACFlows_to_DeleteLCR_Item */
static int hf_nbap_RL_Specific_DCH_Info_item = -1;  /* RL_Specific_DCH_Info_Item */
static int hf_nbap_dCH_id = -1;                   /* T_dCH_id */
static int hf_nbap_transportlayeraddress = -1;    /* TransportLayerAddress */
static int hf_nbap_rL_Specific_E_DCH_Information = -1;  /* RL_Specific_E_DCH_Information */
static int hf_nbap_RL_Specific_E_DCH_Information_item = -1;  /* RL_Specific_E_DCH_Information_Item */
static int hf_nbap_Reference_E_TFCI_Information_item = -1;  /* Reference_E_TFCI_Information_Item */
static int hf_nbap_reference_E_TFCI = -1;         /* E_TFCI */
static int hf_nbap_reference_E_TFCI_PO = -1;      /* Reference_E_TFCI_PO */
static int hf_nbap_periodic_01 = -1;              /* ReportCharacteristicsType_ReportPeriodicity */
static int hf_nbap_event_a = -1;                  /* ReportCharacteristicsType_EventA */
static int hf_nbap_event_b = -1;                  /* ReportCharacteristicsType_EventB */
static int hf_nbap_event_c = -1;                  /* ReportCharacteristicsType_EventC */
static int hf_nbap_event_d = -1;                  /* ReportCharacteristicsType_EventD */
static int hf_nbap_event_e = -1;                  /* ReportCharacteristicsType_EventE */
static int hf_nbap_event_f = -1;                  /* ReportCharacteristicsType_EventF */
static int hf_nbap_extension_ReportCharacteristics = -1;  /* Extension_ReportCharacteristics */
static int hf_nbap_measurementThreshold = -1;     /* ReportCharacteristicsType_MeasurementThreshold */
static int hf_nbap_measurementHysteresisTime = -1;  /* ReportCharacteristicsType_ScaledMeasurementHysteresisTime */
static int hf_nbap_measurementIncreaseThreshold = -1;  /* ReportCharacteristicsType_MeasurementIncreaseDecreaseThreshold */
static int hf_nbap_measurementChangeTime = -1;    /* ReportCharacteristicsType_ScaledMeasurementChangeTime */
static int hf_nbap_measurementDecreaseThreshold = -1;  /* ReportCharacteristicsType_MeasurementIncreaseDecreaseThreshold */
static int hf_nbap_measurementThreshold1 = -1;    /* ReportCharacteristicsType_MeasurementThreshold */
static int hf_nbap_measurementThreshold2 = -1;    /* ReportCharacteristicsType_MeasurementThreshold */
static int hf_nbap_reportPeriodicity = -1;        /* ReportCharacteristicsType_ReportPeriodicity */
static int hf_nbap_received_total_wide_band_power_01 = -1;  /* Received_total_wide_band_power_Value_IncrDecrThres */
static int hf_nbap_uL_TimeslotISCP_01 = -1;       /* UL_TimeslotISCP_Value_IncrDecrThres */
static int hf_nbap_sir = -1;                      /* SIR_Value_IncrDecrThres */
static int hf_nbap_sir_error = -1;                /* SIR_Error_Value_IncrDecrThres */
static int hf_nbap_transmitted_code_power = -1;   /* Transmitted_Code_Power_Value_IncrDecrThres */
static int hf_nbap_rscp = -1;                     /* RSCP_Value_IncrDecrThres */
static int hf_nbap_round_trip_time = -1;          /* Round_Trip_Time_IncrDecrThres */
static int hf_nbap_extension_ReportCharacteristicsType_MeasurementIncreaseDecreaseThreshold = -1;  /* Extension_ReportCharacteristicsType_MeasurementIncreaseDecreaseThreshold */
static int hf_nbap_sir_01 = -1;                   /* SIR_Value */
static int hf_nbap_sir_error_01 = -1;             /* SIR_Error_Value */
static int hf_nbap_transmitted_code_power_01 = -1;  /* Transmitted_Code_Power_Value */
static int hf_nbap_rscp_01 = -1;                  /* RSCP_Value */
static int hf_nbap_rx_timing_deviation = -1;      /* Rx_Timing_Deviation_Value */
static int hf_nbap_round_trip_time_01 = -1;       /* Round_Trip_Time_Value */
static int hf_nbap_extension_ReportCharacteristicsType_MeasurementThreshold = -1;  /* Extension_ReportCharacteristicsType_MeasurementThreshold */
static int hf_nbap_msec = -1;                     /* MeasurementChangeTime_Scaledmsec */
static int hf_nbap_msec_01 = -1;                  /* MeasurementHysteresisTime_Scaledmsec */
static int hf_nbap_msec_02 = -1;                  /* ReportPeriodicity_Scaledmsec */
static int hf_nbap_Received_total_wide_band_power_For_CellPortion_Value_item = -1;  /* Received_total_wide_band_power_For_CellPortion_Value_Item */
static int hf_nbap_received_total_wide_band_power_value = -1;  /* Received_total_wide_band_power_Value */
static int hf_nbap_Received_total_wide_band_power_For_CellPortion_ValueLCR_item = -1;  /* Received_total_wide_band_power_For_CellPortion_ValueLCR_Item */
static int hf_nbap_Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value_item = -1;  /* Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value_Item */
static int hf_nbap_received_Scheduled_power_share_value = -1;  /* RSEPS_Value */
static int hf_nbap_informationAvailable = -1;     /* InformationAvailable */
static int hf_nbap_informationnotAvailable = -1;  /* InformationnotAvailable */
static int hf_nbap_requesteddataValue = -1;       /* RequestedDataValue */
static int hf_nbap_dgps_corrections = -1;         /* DGPSCorrections */
static int hf_nbap_gps_navandrecovery = -1;       /* GPS_NavigationModel_and_TimeRecovery */
static int hf_nbap_gps_ionos_model = -1;          /* GPS_Ionospheric_Model */
static int hf_nbap_gps_utc_model = -1;            /* GPS_UTC_Model */
static int hf_nbap_gps_almanac = -1;              /* GPS_Almanac */
static int hf_nbap_gps_rt_integrity = -1;         /* GPS_RealTime_Integrity */
static int hf_nbap_gpsrxpos = -1;                 /* GPS_RX_POS */
static int hf_nbap_SAT_Info_Almanac_item = -1;    /* SAT_Info_Almanac_Item */
static int hf_nbap_data_id = -1;                  /* DATA_ID */
static int hf_nbap_sat_id = -1;                   /* SAT_ID */
static int hf_nbap_gps_e_alm = -1;                /* BIT_STRING_SIZE_16 */
static int hf_nbap_gps_toa_alm = -1;              /* BIT_STRING_SIZE_8 */
static int hf_nbap_gps_delta_I_alm = -1;          /* BIT_STRING_SIZE_16 */
static int hf_nbap_omegadot_alm = -1;             /* BIT_STRING_SIZE_16 */
static int hf_nbap_svhealth_alm = -1;             /* BIT_STRING_SIZE_8 */
static int hf_nbap_gps_a_sqrt_alm = -1;           /* BIT_STRING_SIZE_24 */
static int hf_nbap_omegazero_alm = -1;            /* BIT_STRING_SIZE_24 */
static int hf_nbap_m_zero_alm = -1;               /* BIT_STRING_SIZE_24 */
static int hf_nbap_gps_omega_alm = -1;            /* BIT_STRING_SIZE_24 */
static int hf_nbap_gps_af_zero_alm = -1;          /* BIT_STRING_SIZE_11 */
static int hf_nbap_gps_af_one_alm = -1;           /* BIT_STRING_SIZE_11 */
static int hf_nbap_SAT_Info_Almanac_ExtList_item = -1;  /* SAT_Info_Almanac_ExtItem */
static int hf_nbap_SAT_Info_DGPSCorrections_item = -1;  /* SAT_Info_DGPSCorrections_Item */
static int hf_nbap_iode_dgps = -1;                /* BIT_STRING_SIZE_8 */
static int hf_nbap_prc = -1;                      /* PRC */
static int hf_nbap_range_correction_rate = -1;    /* Range_Correction_Rate */
static int hf_nbap_SATInfo_RealTime_Integrity_item = -1;  /* SAT_Info_RealTime_Integrity_Item */
static int hf_nbap_bad_sat_id = -1;               /* SAT_ID */
static int hf_nbap_SecondaryServingCells_item = -1;  /* SecondaryServingCellsItem */
static int hf_nbap_secondaryC_ID = -1;            /* C_ID */
static int hf_nbap_numSecondaryHS_SCCH_Codes = -1;  /* NumHS_SCCH_Codes */
static int hf_nbap_new_secondary_CPICH = -1;      /* CommonPhysicalChannelID */
static int hf_nbap_secondary_CPICH_shall_not_be_used = -1;  /* NULL */
static int hf_nbap_serving_E_DCH_RL_in_this_NodeB = -1;  /* Serving_E_DCH_RL_in_this_NodeB */
static int hf_nbap_serving_E_DCH_RL_not_in_this_NodeB = -1;  /* NULL */
static int hf_nbap_SetsOfHS_SCCH_Codes_item = -1;  /* SetsOfHS_SCCH_CodesItem */
static int hf_nbap_hS_SCCH_PreconfiguredCodes = -1;  /* HS_SCCH_PreconfiguredCodes */
static int hf_nbap_setup = -1;                    /* Additional_EDCH_Setup_Info */
static int hf_nbap_configurationChange = -1;      /* Additional_EDCH_Cell_Information_ConfigurationChange_List */
static int hf_nbap_removal = -1;                  /* Additional_EDCH_Cell_Information_Removal_List */
static int hf_nbap_sFNSFNChangeLimit = -1;        /* SFNSFNChangeLimit */
static int hf_nbap_predictedSFNSFNDeviationLimit = -1;  /* PredictedSFNSFNDeviationLimit */
static int hf_nbap_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation = -1;  /* T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation */
static int hf_nbap_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item = -1;  /* T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item */
static int hf_nbap_sFNSFNValue = -1;              /* SFNSFNValue */
static int hf_nbap_sFNSFNQuality = -1;            /* SFNSFNQuality */
static int hf_nbap_sFNSFNDriftRate = -1;          /* SFNSFNDriftRate */
static int hf_nbap_sFNSFNDriftRateQuality = -1;   /* SFNSFNDriftRateQuality */
static int hf_nbap_sFNSFNTimeStampInformation = -1;  /* SFNSFNTimeStampInformation */
static int hf_nbap_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation = -1;  /* T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation */
static int hf_nbap_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item = -1;  /* T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item */
static int hf_nbap_sFNSFNTimeStamp_FDD = -1;      /* SFN */
static int hf_nbap_sFNSFNTimeStamp_TDD = -1;      /* SFNSFNTimeStamp_TDD */
static int hf_nbap_sFN = -1;                      /* SFN */
static int hf_nbap_sFNSFN_FDD = -1;               /* SFNSFN_FDD */
static int hf_nbap_sFNSFN_TDD = -1;               /* SFNSFN_TDD */
static int hf_nbap_sFNSFN_TDD768 = -1;            /* SFNSFN_TDD768 */
static int hf_nbap_synchronisationReportCharacteristicsType = -1;  /* SynchronisationReportCharacteristicsType */
static int hf_nbap_synchronisationReportCharactThreExc = -1;  /* SynchronisationReportCharactThreExc */
static int hf_nbap_SynchronisationReportCharactThreExc_item = -1;  /* SynchronisationReportCharactThreInfoItem */
static int hf_nbap_syncFrameNumber = -1;          /* SyncFrameNumber */
static int hf_nbap_cellSyncBurstInformation = -1;  /* SEQUENCE_SIZE_1_maxNrOfReceptsPerSyncFrame_OF_SynchronisationReportCharactCellSyncBurstInfoItem */
static int hf_nbap_cellSyncBurstInformation_item = -1;  /* SynchronisationReportCharactCellSyncBurstInfoItem */
static int hf_nbap_cellSyncBurstCode = -1;        /* CellSyncBurstCode */
static int hf_nbap_cellSyncBurstCodeShift = -1;   /* CellSyncBurstCodeShift */
static int hf_nbap_cellSyncBurstTiming = -1;      /* CellSyncBurstTiming */
static int hf_nbap_cellSyncBurstTimingThreshold = -1;  /* CellSyncBurstTimingThreshold */
static int hf_nbap_SyncDLCodeIdThreInfoLCR_item = -1;  /* SyncDLCodeIdThreInfoList */
static int hf_nbap_syncFrameNoToReceive = -1;     /* SyncFrameNumber */
static int hf_nbap_syncDLCodeIdInfoLCR = -1;      /* SyncDLCodeInfoListLCR */
static int hf_nbap_SyncDLCodeInfoListLCR_item = -1;  /* SyncDLCodeInfoItemLCR */
static int hf_nbap_syncDLCodeId = -1;             /* SYNCDlCodeId */
static int hf_nbap_syncDLCodeIdArrivTime = -1;    /* CellSyncBurstTimingLCR */
static int hf_nbap_syncDLCodeIdTimingThre = -1;   /* CellSyncBurstTimingThreshold */
static int hf_nbap_modulation = -1;               /* Modulation */
static int hf_nbap_TDD_DL_Code_Information_item = -1;  /* TDD_DL_Code_InformationItem */
static int hf_nbap_dPCH_ID = -1;                  /* DPCH_ID */
static int hf_nbap_TDD_DL_Code_LCR_Information_item = -1;  /* TDD_DL_Code_LCR_InformationItem */
static int hf_nbap_tdd_DL_DPCH_TimeSlotFormat_LCR = -1;  /* TDD_DL_DPCH_TimeSlotFormat_LCR */
static int hf_nbap_TDD_DL_Code_768_Information_item = -1;  /* TDD_DL_Code_768_InformationItem */
static int hf_nbap_qPSK_01 = -1;                  /* QPSK_DL_DPCH_TimeSlotFormatTDD_LCR */
static int hf_nbap_eightPSK = -1;                 /* EightPSK_DL_DPCH_TimeSlotFormatTDD_LCR */
static int hf_nbap_initialOffset = -1;            /* INTEGER_0_255 */
static int hf_nbap_noinitialOffset = -1;          /* INTEGER_0_63 */
static int hf_nbap_signalledGainFactors = -1;     /* T_signalledGainFactors */
static int hf_nbap_gainFactor = -1;               /* T_gainFactor */
static int hf_nbap_fdd = -1;                      /* T_fdd */
static int hf_nbap_betaC = -1;                    /* BetaCD */
static int hf_nbap_betaD = -1;                    /* BetaCD */
static int hf_nbap_tdd = -1;                      /* BetaCD */
static int hf_nbap_refTFCNumber = -1;             /* RefTFCNumber */
static int hf_nbap_computedGainFactors = -1;      /* RefTFCNumber */
static int hf_nbap_TDD_UL_Code_Information_item = -1;  /* TDD_UL_Code_InformationItem */
static int hf_nbap_TDD_UL_Code_LCR_Information_item = -1;  /* TDD_UL_Code_LCR_InformationItem */
static int hf_nbap_tdd_UL_DPCH_TimeSlotFormat_LCR = -1;  /* TDD_UL_DPCH_TimeSlotFormat_LCR */
static int hf_nbap_TDD_UL_Code_768_Information_item = -1;  /* TDD_UL_Code_768_InformationItem */
static int hf_nbap_qPSK_02 = -1;                  /* QPSK_UL_DPCH_TimeSlotFormatTDD_LCR */
static int hf_nbap_eightPSK_01 = -1;              /* EightPSK_UL_DPCH_TimeSlotFormatTDD_LCR */
static int hf_nbap_tFCI_SignallingOption = -1;    /* TFCI_SignallingMode_TFCI_SignallingOption */
static int hf_nbap_not_Used_splitType = -1;       /* NULL */
static int hf_nbap_not_Used_lengthOfTFCI2 = -1;   /* NULL */
static int hf_nbap_TimeSlot_InitiatedListLCR_item = -1;  /* TimeSlotLCR */
static int hf_nbap_TimeSlotMeasurementValueListLCR_item = -1;  /* TimeSlotMeasurementValueLCR */
static int hf_nbap_commonMeasurementValue = -1;   /* CommonMeasurementValue */
static int hf_nbap_initialPhase_01 = -1;          /* INTEGER_0_1048575_ */
static int hf_nbap_dsField = -1;                  /* DsField */
static int hf_nbap_genericTrafficCategory = -1;   /* GenericTrafficCategory */
static int hf_nbap_Transmission_Gap_Pattern_Sequence_Information_item = -1;  /* Transmission_Gap_Pattern_Sequence_Information_item */
static int hf_nbap_tGSN = -1;                     /* TGSN */
static int hf_nbap_tGL1 = -1;                     /* GapLength */
static int hf_nbap_tGL2 = -1;                     /* GapLength */
static int hf_nbap_tGD = -1;                      /* TGD */
static int hf_nbap_tGPL1 = -1;                    /* GapDuration */
static int hf_nbap_not_to_be_used_1 = -1;         /* GapDuration */
static int hf_nbap_uL_DL_mode = -1;               /* UL_DL_mode */
static int hf_nbap_downlink_Compressed_Mode_Method = -1;  /* Downlink_Compressed_Mode_Method */
static int hf_nbap_uplink_Compressed_Mode_Method = -1;  /* Uplink_Compressed_Mode_Method */
static int hf_nbap_dL_FrameType = -1;             /* DL_FrameType */
static int hf_nbap_delta_SIR1 = -1;               /* DeltaSIR */
static int hf_nbap_delta_SIR_after1 = -1;         /* DeltaSIR */
static int hf_nbap_delta_SIR2 = -1;               /* DeltaSIR */
static int hf_nbap_delta_SIR_after2 = -1;         /* DeltaSIR */
static int hf_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue_item = -1;  /* TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue_Item */
static int hf_nbap_transmittedCarrierPowerOfAllCodesNotUsedForHSTransmissionValue = -1;  /* TransmittedCarrierPowerOfAllCodesNotUsedForHSTransmissionValue */
static int hf_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue_item = -1;  /* TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue_Item */
static int hf_nbap_Transmitted_Carrier_Power_For_CellPortion_Value_item = -1;  /* Transmitted_Carrier_Power_For_CellPortion_Value_Item */
static int hf_nbap_transmitted_Carrier_Power_Value = -1;  /* Transmitted_Carrier_Power_Value */
static int hf_nbap_Transmitted_Carrier_Power_For_CellPortion_ValueLCR_item = -1;  /* Transmitted_Carrier_Power_For_CellPortion_ValueLCR_Item */
static int hf_nbap_tFCSvalues = -1;               /* T_tFCSvalues */
static int hf_nbap_no_Split_in_TFCI = -1;         /* TFCS_TFCSList */
static int hf_nbap_not_Used_split_in_TFCI = -1;   /* NULL */
static int hf_nbap_TFCS_TFCSList_item = -1;       /* TFCS_TFCSList_item */
static int hf_nbap_cTFC = -1;                     /* TFCS_CTFC */
static int hf_nbap_tFC_Beta = -1;                 /* TransportFormatCombination_Beta */
static int hf_nbap_ctfc2bit = -1;                 /* INTEGER_0_3 */
static int hf_nbap_ctfc4bit = -1;                 /* INTEGER_0_15 */
static int hf_nbap_ctfc6bit = -1;                 /* INTEGER_0_63 */
static int hf_nbap_ctfc8bit = -1;                 /* INTEGER_0_255 */
static int hf_nbap_ctfc12bit = -1;                /* INTEGER_0_4095 */
static int hf_nbap_ctfc16bit = -1;                /* INTEGER_0_65535 */
static int hf_nbap_ctfcmaxbit = -1;               /* INTEGER_0_maxCTFC */
static int hf_nbap_Transport_Block_Size_List_item = -1;  /* Transport_Block_Size_List_item */
static int hf_nbap_transport_Block_Size_Index_for_Enhanced_PCH = -1;  /* Transport_Block_Size_Index_for_Enhanced_PCH */
static int hf_nbap_dynamicParts = -1;             /* TransportFormatSet_DynamicPartList */
static int hf_nbap_semi_staticPart = -1;          /* TransportFormatSet_Semi_staticPart */
static int hf_nbap_TransportFormatSet_DynamicPartList_item = -1;  /* TransportFormatSet_DynamicPartList_item */
static int hf_nbap_nrOfTransportBlocks = -1;      /* TransportFormatSet_NrOfTransportBlocks */
static int hf_nbap_transportBlockSize = -1;       /* TransportFormatSet_TransportBlockSize */
static int hf_nbap_mode = -1;                     /* TransportFormatSet_ModeDP */
static int hf_nbap_transmissionTimeIntervalInformation = -1;  /* TransmissionTimeIntervalInformation */
static int hf_nbap_TransmissionTimeIntervalInformation_item = -1;  /* TransmissionTimeIntervalInformation_item */
static int hf_nbap_transmissionTimeInterval = -1;  /* TransportFormatSet_TransmissionTimeIntervalDynamic */
static int hf_nbap_transmissionTimeInterval_01 = -1;  /* TransportFormatSet_TransmissionTimeIntervalSemiStatic */
static int hf_nbap_channelCoding = -1;            /* TransportFormatSet_ChannelCodingType */
static int hf_nbap_codingRate = -1;               /* TransportFormatSet_CodingRate */
static int hf_nbap_rateMatchingAttribute = -1;    /* TransportFormatSet_RateMatchingAttribute */
static int hf_nbap_cRC_Size = -1;                 /* TransportFormatSet_CRC_Size */
static int hf_nbap_mode_01 = -1;                  /* TransportFormatSet_ModeSSP */
static int hf_nbap_tdd_01 = -1;                   /* TDD_TransportFormatSet_ModeDP */
static int hf_nbap_notApplicable = -1;            /* NULL */
static int hf_nbap_tdd_02 = -1;                   /* TransportFormatSet_SecondInterleavingMode */
static int hf_nbap_mS = -1;                       /* INTEGER_0_16383 */
static int hf_nbap_lS = -1;                       /* INTEGER_0_4294967295 */
static int hf_nbap_tUTRANGANSSChangeLimit = -1;   /* INTEGER_1_256 */
static int hf_nbap_predictedTUTRANGANSSDeviationLimit = -1;  /* INTEGER_1_256 */
static int hf_nbap_tUTRANGANSS = -1;              /* TUTRANGANSS */
static int hf_nbap_tUTRANGANSSQuality = -1;       /* INTEGER_0_255 */
static int hf_nbap_tUTRANGANSSDriftRate = -1;     /* INTEGER_M50_50 */
static int hf_nbap_tUTRANGANSSDriftRateQuality = -1;  /* INTEGER_0_50 */
static int hf_nbap_ms_part = -1;                  /* INTEGER_0_16383 */
static int hf_nbap_ls_part = -1;                  /* INTEGER_0_4294967295 */
static int hf_nbap_tUTRANGPSChangeLimit = -1;     /* TUTRANGPSChangeLimit */
static int hf_nbap_predictedTUTRANGPSDeviationLimit = -1;  /* PredictedTUTRANGPSDeviationLimit */
static int hf_nbap_tUTRANGPS = -1;                /* TUTRANGPS */
static int hf_nbap_tUTRANGPSQuality = -1;         /* TUTRANGPSQuality */
static int hf_nbap_tUTRANGPSDriftRate = -1;       /* TUTRANGPSDriftRate */
static int hf_nbap_tUTRANGPSDriftRateQuality = -1;  /* TUTRANGPSDriftRateQuality */
static int hf_nbap_rNC_ID = -1;                   /* RNC_ID */
static int hf_nbap_uE_AggregateMaximumBitRateDownlink = -1;  /* UE_AggregateMaximumBitRateDownlink */
static int hf_nbap_uE_AggregateMaximumBitRateUplink = -1;  /* UE_AggregateMaximumBitRateUplink */
static int hf_nbap_hSDSCH_Physical_Layer_Category_01 = -1;  /* T_hSDSCH_Physical_Layer_Category */
static int hf_nbap_UL_Timeslot_Information_item = -1;  /* UL_Timeslot_InformationItem */
static int hf_nbap_uL_Code_InformationList = -1;  /* TDD_UL_Code_Information */
static int hf_nbap_UL_TimeslotLCR_Information_item = -1;  /* UL_TimeslotLCR_InformationItem */
static int hf_nbap_uL_Code_InformationList_01 = -1;  /* TDD_UL_Code_LCR_Information */
static int hf_nbap_UL_Timeslot768_Information_item = -1;  /* UL_Timeslot768_InformationItem */
static int hf_nbap_uL_Code_InformationList_02 = -1;  /* TDD_UL_Code_768_Information */
static int hf_nbap_uL_ScramblingCodeNumber = -1;  /* UL_ScramblingCodeNumber */
static int hf_nbap_uL_ScramblingCodeLength = -1;  /* UL_ScramblingCodeLength */
static int hf_nbap_uL_Synchronisation_StepSize = -1;  /* UL_Synchronisation_StepSize */
static int hf_nbap_uL_Synchronisation_Frequency = -1;  /* UL_Synchronisation_Frequency */
static int hf_nbap_UL_TimeSlot_ISCP_Info_item = -1;  /* UL_TimeSlot_ISCP_InfoItem */
static int hf_nbap_iSCP = -1;                     /* UL_TimeslotISCP_Value */
static int hf_nbap_UL_TimeSlot_ISCP_LCR_Info_item = -1;  /* UL_TimeSlot_ISCP_LCR_InfoItem */
static int hf_nbap_UpPTSInterference_For_CellPortion_Value_item = -1;  /* UpPTSInterference_For_CellPortion_Value_Item */
static int hf_nbap_upPTSInterferenceValue = -1;   /* UpPTSInterferenceValue */
static int hf_nbap_USCH_Information_item = -1;    /* USCH_InformationItem */
static int hf_nbap_uSCH_ID = -1;                  /* USCH_ID */
static int hf_nbap_USCH_InformationResponse_item = -1;  /* USCH_InformationResponseItem */
static int hf_nbap_UL_TimeslotISCP_For_CellPortion_Value_item = -1;  /* UL_TimeslotISCP_For_CellPortion_Value_Item */
static int hf_nbap_uL_TimeslotISCP_Value = -1;    /* UL_TimeslotISCP_Value */
static int hf_nbap_protocolIEs = -1;              /* ProtocolIE_Container */
static int hf_nbap_protocolExtensions = -1;       /* ProtocolExtensionContainer */
static int hf_nbap_secondary_CCPCH_parameters = -1;  /* Secondary_CCPCH_CTCH_SetupRqstFDD */
static int hf_nbap_pRACH_parameters = -1;         /* PRACH_CTCH_SetupRqstFDD */
static int hf_nbap_notUsed_pCPCHes_parameters = -1;  /* NULL */
static int hf_nbap_fdd_S_CCPCH_Offset = -1;       /* FDD_S_CCPCH_Offset */
static int hf_nbap_tFCS = -1;                     /* TFCS */
static int hf_nbap_secondary_CCPCH_SlotFormat = -1;  /* SecondaryCCPCH_SlotFormat */
static int hf_nbap_multiplexingPosition = -1;     /* MultiplexingPosition */
static int hf_nbap_powerOffsetInformation = -1;   /* PowerOffsetInformation_CTCH_SetupRqstFDD */
static int hf_nbap_fACH_Parameters = -1;          /* FACH_ParametersList_CTCH_SetupRqstFDD */
static int hf_nbap_pCH_Parameters = -1;           /* PCH_Parameters_CTCH_SetupRqstFDD */
static int hf_nbap_pO1_ForTFCI_Bits = -1;         /* PowerOffset */
static int hf_nbap_pO3_ForPilotBits = -1;         /* PowerOffset */
static int hf_nbap_FACH_ParametersListIE_CTCH_SetupRqstFDD_item = -1;  /* FACH_ParametersItem_CTCH_SetupRqstFDD */
static int hf_nbap_maxFACH_Power = -1;            /* DL_Power */
static int hf_nbap_transportFormatSet_01 = -1;    /* T_transportFormatSet */
static int hf_nbap_pCH_Power = -1;                /* DL_Power */
static int hf_nbap_pICH_Parameters = -1;          /* PICH_Parameters_CTCH_SetupRqstFDD */
static int hf_nbap_pICH_Mode = -1;                /* PICH_Mode */
static int hf_nbap_mICH_Power = -1;               /* PICH_Power */
static int hf_nbap_mICH_Mode = -1;                /* MICH_Mode */
static int hf_nbap_preambleSignatures = -1;       /* PreambleSignatures */
static int hf_nbap_allowedSlotFormatInformation = -1;  /* AllowedSlotFormatInformationList_CTCH_SetupRqstFDD */
static int hf_nbap_rACH_SubChannelNumbers = -1;   /* RACH_SubChannelNumbers */
static int hf_nbap_ul_punctureLimit = -1;         /* PunctureLimit */
static int hf_nbap_rACH_Parameters = -1;          /* RACH_Parameters_CTCH_SetupRqstFDD */
static int hf_nbap_aICH_Parameters = -1;          /* AICH_Parameters_CTCH_SetupRqstFDD */
static int hf_nbap_AllowedSlotFormatInformationList_CTCH_SetupRqstFDD_item = -1;  /* AllowedSlotFormatInformationItem_CTCH_SetupRqstFDD */
static int hf_nbap_rACHSlotFormat = -1;           /* RACH_SlotFormat */
static int hf_nbap_secondary_CCPCH_parameters_01 = -1;  /* Secondary_CCPCH_CTCH_SetupRqstTDD */
static int hf_nbap_pRACH_parameters_01 = -1;      /* PRACH_CTCH_SetupRqstTDD */
static int hf_nbap_extension_CommonPhysicalChannelType_CTCH_SetupRqstTDD = -1;  /* Extension_CommonPhysicalChannelType_CTCH_SetupRqstTDD */
static int hf_nbap_sCCPCH_CCTrCH_ID = -1;         /* CCTrCH_ID */
static int hf_nbap_tFCI_Coding = -1;              /* TFCI_Coding */
static int hf_nbap_punctureLimit = -1;            /* PunctureLimit */
static int hf_nbap_secondaryCCPCH_parameterList = -1;  /* Secondary_CCPCH_parameterList_CTCH_SetupRqstTDD */
static int hf_nbap_fACH_ParametersList = -1;      /* FACH_ParametersList_CTCH_SetupRqstTDD */
static int hf_nbap_pCH_Parameters_01 = -1;        /* PCH_Parameters_CTCH_SetupRqstTDD */
static int hf_nbap_Secondary_CCPCH_parameterListIE_CTCH_SetupRqstTDD_item = -1;  /* Secondary_CCPCH_parameterItem_CTCH_SetupRqstTDD */
static int hf_nbap_midambleShiftandBurstType = -1;  /* MidambleShiftAndBurstType */
static int hf_nbap_s_CCPCH_Power = -1;            /* DL_Power */
static int hf_nbap_Secondary_CCPCH_LCR_parameterList_CTCH_SetupRqstTDD_item = -1;  /* Secondary_CCPCH_LCR_parameterItem_CTCH_SetupRqstTDD */
static int hf_nbap_s_CCPCH_TimeSlotFormat_LCR = -1;  /* TDD_DL_DPCH_TimeSlotFormat_LCR */
static int hf_nbap_Secondary_CCPCH_768_parameterList_CTCH_SetupRqstTDD_item = -1;  /* Secondary_CCPCH_768_parameterItem_CTCH_SetupRqstTDD */
static int hf_nbap_tFCI_Presence768 = -1;         /* TFCI_Presence */
static int hf_nbap_midambleShiftandBurstType768 = -1;  /* MidambleShiftAndBurstType768 */
static int hf_nbap_FACH_ParametersListIE_CTCH_SetupRqstTDD_item = -1;  /* FACH_ParametersItem_CTCH_SetupRqstTDD */
static int hf_nbap_fACH_CCTrCH_ID = -1;           /* CCTrCH_ID */
static int hf_nbap_pCH_CCTrCH_ID = -1;            /* CCTrCH_ID */
static int hf_nbap_pICH_Parameters_01 = -1;       /* PICH_Parameters_CTCH_SetupRqstTDD */
static int hf_nbap_midambleshiftAndBurstType = -1;  /* MidambleShiftAndBurstType */
static int hf_nbap_midambleshiftAndBurstType78 = -1;  /* MidambleShiftAndBurstType768 */
static int hf_nbap_notificationIndicatorLength = -1;  /* NotificationIndicatorLength */
static int hf_nbap_mICH_TDDOption_Specific_Parameters = -1;  /* MICH_TDDOption_Specific_Parameters_CTCH_SetupRqstTDD */
static int hf_nbap_hCR_TDD = -1;                  /* MICH_HCR_Parameters_CTCH_SetupRqstTDD */
static int hf_nbap_lCR_TDD = -1;                  /* MICH_LCR_Parameters_CTCH_SetupRqstTDD */
static int hf_nbap_cHipRate768_TDD = -1;          /* MICH_768_Parameters_CTCH_SetupRqstTDD */
static int hf_nbap_tSTD_Indicator = -1;           /* TSTD_Indicator */
static int hf_nbap_midambleshiftAndBurstType768 = -1;  /* MidambleShiftAndBurstType768 */
static int hf_nbap_TimeSlotConfigurationList_LCR_CTCH_SetupRqstTDD_item = -1;  /* TimeSlotConfigurationItem_LCR_CTCH_SetupRqstTDD */
static int hf_nbap_timeslotLCR_Parameter_ID = -1;  /* CellParameterID */
static int hf_nbap_Secondary_CCPCH_parameterExtendedList_CTCH_SetupRqstTDD_item = -1;  /* Secondary_CCPCH_parameterItem_CTCH_SetupRqstTDD */
static int hf_nbap_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_SetupRqstTDD_item = -1;  /* Secondary_CCPCH_LCR_parameterItem_CTCH_SetupRqstTDD */
static int hf_nbap_pRACH_Parameters_CTCH_SetupRqstTDD = -1;  /* PRACH_Parameters_CTCH_SetupRqstTDD */
static int hf_nbap_maxPRACH_MidambleShifts = -1;  /* MaxPRACH_MidambleShifts */
static int hf_nbap_pRACH_Midamble = -1;           /* PRACH_Midamble */
static int hf_nbap_rACH = -1;                     /* RACH_Parameter_CTCH_SetupRqstTDD */
static int hf_nbap_uL_TransportFormatSet = -1;    /* TransportFormatSet */
static int hf_nbap_PRACH_LCR_ParametersList_CTCH_SetupRqstTDD_item = -1;  /* PRACH_LCR_ParametersItem_CTCH_SetupRqstTDD */
static int hf_nbap_fPACH_Power = -1;              /* FPACH_Power */
static int hf_nbap_maxPowerPLCCH = -1;            /* DL_Power */
static int hf_nbap_maxE_RUCCH_MidambleShifts = -1;  /* MaxPRACH_MidambleShifts */
static int hf_nbap_e_RUCCH_Midamble = -1;         /* PRACH_Midamble */
static int hf_nbap_FACH_CommonTransportChannel_InformationResponse_item = -1;  /* CommonTransportChannel_InformationResponse */
static int hf_nbap_secondary_CCPCH_parameters_02 = -1;  /* Secondary_CCPCHList_CTCH_ReconfRqstFDD */
static int hf_nbap_pRACH_parameters_02 = -1;      /* PRACHList_CTCH_ReconfRqstFDD */
static int hf_nbap_notUsed_cPCH_parameters = -1;  /* NULL */
static int hf_nbap_fACH_ParametersList_CTCH_ReconfRqstFDD = -1;  /* FACH_ParametersList_CTCH_ReconfRqstFDD */
static int hf_nbap_pCH_Parameters_CTCH_ReconfRqstFDD = -1;  /* PCH_Parameters_CTCH_ReconfRqstFDD */
static int hf_nbap_pICH_Parameters_CTCH_ReconfRqstFDD = -1;  /* PICH_Parameters_CTCH_ReconfRqstFDD */
static int hf_nbap_FACH_ParametersListIE_CTCH_ReconfRqstFDD_item = -1;  /* FACH_ParametersItem_CTCH_ReconfRqstFDD */
static int hf_nbap_pRACH_ParametersList_CTCH_ReconfRqstFDD = -1;  /* PRACH_ParametersList_CTCH_ReconfRqstFDD */
static int hf_nbap_aICH_ParametersList_CTCH_ReconfRqstFDD = -1;  /* AICH_ParametersList_CTCH_ReconfRqstFDD */
static int hf_nbap_PRACH_ParametersListIE_CTCH_ReconfRqstFDD_item = -1;  /* PRACH_ParametersItem_CTCH_ReconfRqstFDD */
static int hf_nbap_allowedSlotFormatInformation_01 = -1;  /* AllowedSlotFormatInformationList_CTCH_ReconfRqstFDD */
static int hf_nbap_AllowedSlotFormatInformationList_CTCH_ReconfRqstFDD_item = -1;  /* AllowedSlotFormatInformationItem_CTCH_ReconfRqstFDD */
static int hf_nbap_rACH_SlotFormat = -1;          /* RACH_SlotFormat */
static int hf_nbap_AICH_ParametersListIE_CTCH_ReconfRqstFDD_item = -1;  /* AICH_ParametersItem_CTCH_ReconfRqstFDD */
static int hf_nbap_secondaryCCPCHList = -1;       /* Secondary_CCPCHList_CTCH_ReconfRqstTDD */
static int hf_nbap_Secondary_CCPCHListIE_CTCH_ReconfRqstTDD_item = -1;  /* Secondary_CCPCHItem_CTCH_ReconfRqstTDD */
static int hf_nbap_sCCPCH_Power = -1;             /* DL_Power */
static int hf_nbap_Secondary_CCPCH_parameterExtendedList_CTCH_ReconfRqstTDD_item = -1;  /* Secondary_CCPCHItem_CTCH_ReconfRqstTDD */
static int hf_nbap_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_ReconfRqstTDD_item = -1;  /* Secondary_CCPCHItem_CTCH_ReconfRqstTDD */
static int hf_nbap_FACH_ParametersList_CTCH_ReconfRqstTDD_item = -1;  /* FACH_ParametersItem_CTCH_ReconfRqstTDD */
static int hf_nbap_commonPhysicalChannelId = -1;  /* CommonPhysicalChannelID */
static int hf_nbap_fPACHPower = -1;               /* FPACH_Power */
static int hf_nbap_secondaryCCPCH768List = -1;    /* Secondary_CCPCH_768_List_CTCH_ReconfRqstTDD */
static int hf_nbap_Secondary_CCPCH_768_List_CTCH_ReconfRqstTDD_item = -1;  /* Secondary_CCPCH_768_Item_CTCH_ReconfRqstTDD */
static int hf_nbap_uPPCHPositionLCR = -1;         /* UPPCHPositionLCR */
static int hf_nbap_Cell_InformationList_AuditRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_configurationGenerationID = -1;  /* ConfigurationGenerationID */
static int hf_nbap_primary_SCH_Information = -1;  /* P_SCH_Information_AuditRsp */
static int hf_nbap_secondary_SCH_Information = -1;  /* S_SCH_Information_AuditRsp */
static int hf_nbap_primary_CPICH_Information = -1;  /* P_CPICH_Information_AuditRsp */
static int hf_nbap_secondary_CPICH_InformationList = -1;  /* S_CPICH_InformationList_AuditRsp */
static int hf_nbap_primary_CCPCH_Information = -1;  /* P_CCPCH_Information_AuditRsp */
static int hf_nbap_bCH_Information = -1;          /* BCH_Information_AuditRsp */
static int hf_nbap_secondary_CCPCH_InformationList = -1;  /* S_CCPCH_InformationList_AuditRsp */
static int hf_nbap_pCH_Information = -1;          /* PCH_Information_AuditRsp */
static int hf_nbap_pICH_Information = -1;         /* PICH_Information_AuditRsp */
static int hf_nbap_fACH_InformationList = -1;     /* FACH_InformationList_AuditRsp */
static int hf_nbap_pRACH_InformationList = -1;    /* PRACH_InformationList_AuditRsp */
static int hf_nbap_rACH_InformationList = -1;     /* RACH_InformationList_AuditRsp */
static int hf_nbap_aICH_InformationList = -1;     /* AICH_InformationList_AuditRsp */
static int hf_nbap_notUsed_1_pCPCH_InformationList = -1;  /* NULL */
static int hf_nbap_notUsed_2_cPCH_InformationList = -1;  /* NULL */
static int hf_nbap_notUsed_3_aP_AICH_InformationList = -1;  /* NULL */
static int hf_nbap_notUsed_4_cDCA_ICH_InformationList = -1;  /* NULL */
static int hf_nbap_sCH_Information = -1;          /* SCH_Information_AuditRsp */
static int hf_nbap_S_CPICH_InformationList_AuditRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_S_CCPCH_InformationList_AuditRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_FACH_InformationList_AuditRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_PRACH_InformationList_AuditRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_RACH_InformationList_AuditRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_AICH_InformationList_AuditRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_FPACH_LCR_InformationList_AuditRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_S_CCPCH_InformationListExt_AuditRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_S_CCPCH_LCR_InformationListExt_AuditRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_PLCCH_InformationList_AuditRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_S_CCPCH_768_InformationList_AuditRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_PRACH_768_InformationList_AuditRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_E_RUCCH_InformationList_AuditRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_E_RUCCH_768_InformationList_AuditRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_Cell_Frequency_List_Information_LCR_MulFreq_AuditRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_UPPCH_LCR_InformationList_AuditRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_MultipleFreq_HS_DSCH_Resources_InformationList_AuditRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_MultipleFreq_E_DCH_Resources_InformationList_AuditRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_CCP_InformationList_AuditRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_communicationControlPortID = -1;  /* CommunicationControlPortID */
static int hf_nbap_Local_Cell_InformationList_AuditRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_dl_or_global_capacityCredit = -1;  /* DL_or_Global_CapacityCredit */
static int hf_nbap_ul_capacityCredit = -1;        /* UL_CapacityCredit */
static int hf_nbap_commonChannelsCapacityConsumptionLaw = -1;  /* CommonChannelsCapacityConsumptionLaw */
static int hf_nbap_dedicatedChannelsCapacityConsumptionLaw = -1;  /* DedicatedChannelsCapacityConsumptionLaw */
static int hf_nbap_maximumDL_PowerCapability = -1;  /* MaximumDL_PowerCapability */
static int hf_nbap_minSpreadingFactor = -1;       /* MinSpreadingFactor */
static int hf_nbap_minimumDL_PowerCapability = -1;  /* MinimumDL_PowerCapability */
static int hf_nbap_local_Cell_Group_ID = -1;      /* Local_Cell_ID */
static int hf_nbap_Local_Cell_Group_InformationList_AuditRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_Power_Local_Cell_Group_InformationList_AuditRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_power_Local_Cell_Group_ID = -1;  /* Local_Cell_ID */
static int hf_nbap_cell = -1;                     /* Cell_CM_Rqst */
static int hf_nbap_rACH_01 = -1;                  /* RACH_CM_Rqst */
static int hf_nbap_notUsed_cPCH = -1;             /* NULL */
static int hf_nbap_extension_CommonMeasurementObjectType_CM_Rqst = -1;  /* Extension_CommonMeasurementObjectType_CM_Rqst */
static int hf_nbap_powerLocalCellGroupID = -1;    /* Local_Cell_ID */
static int hf_nbap_cell_01 = -1;                  /* Cell_CM_Rsp */
static int hf_nbap_rACH_02 = -1;                  /* RACH_CM_Rsp */
static int hf_nbap_extension_CommonMeasurementObjectType_CM_Rsp = -1;  /* Extension_CommonMeasurementObjectType_CM_Rsp */
static int hf_nbap_cell_02 = -1;                  /* Cell_CM_Rprt */
static int hf_nbap_rACH_03 = -1;                  /* RACH_CM_Rprt */
static int hf_nbap_extension_CommonMeasurementObjectType_CM_Rprt = -1;  /* Extension_CommonMeasurementObjectType_CM_Rprt */
static int hf_nbap_commonMeasurementValueInformation = -1;  /* CommonMeasurementValueInformation */
static int hf_nbap_n_INSYNC_IND = -1;             /* N_INSYNC_IND */
static int hf_nbap_n_OUTSYNC_IND = -1;            /* N_OUTSYNC_IND */
static int hf_nbap_t_RLFAILURE = -1;              /* T_RLFAILURE */
static int hf_nbap_primarySCH_Power = -1;         /* DL_Power */
static int hf_nbap_secondarySCH_Power = -1;       /* DL_Power */
static int hf_nbap_primaryCPICH_Power = -1;       /* PrimaryCPICH_Power */
static int hf_nbap_transmitDiversityIndicator = -1;  /* TransmitDiversityIndicator */
static int hf_nbap_SecondaryCPICH_InformationList_Cell_SetupRqstFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_fDD_DL_ChannelisationCodeNumber = -1;  /* FDD_DL_ChannelisationCodeNumber */
static int hf_nbap_secondaryCPICH_Power = -1;     /* DL_Power */
static int hf_nbap_bCH_information = -1;          /* BCH_Information_Cell_SetupRqstFDD */
static int hf_nbap_bCH_Power = -1;                /* DL_Power */
static int hf_nbap_powerRaiseLimit = -1;          /* PowerRaiseLimit */
static int hf_nbap_dLPowerAveragingWindowSize = -1;  /* DLPowerAveragingWindowSize */
static int hf_nbap_iPDL_FDD_Parameters = -1;      /* IPDL_FDD_Parameters */
static int hf_nbap_iPDL_Indicator = -1;           /* IPDL_Indicator */
static int hf_nbap_CellPortion_InformationList_Cell_SetupRqstFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_associatedSecondaryCPICH = -1;  /* CommonPhysicalChannelID */
static int hf_nbap_maximumTransmissionPowerforCellPortion = -1;  /* MaximumTransmissionPower */
static int hf_nbap_syncCaseIndicator = -1;        /* SyncCaseIndicator_Cell_SetupRqstTDD_PSCH */
static int hf_nbap_sCH_Power = -1;                /* DL_Power */
static int hf_nbap_case1 = -1;                    /* Case1_Cell_SetupRqstTDD */
static int hf_nbap_case2 = -1;                    /* Case2_Cell_SetupRqstTDD */
static int hf_nbap_sCH_TimeSlot = -1;             /* SCH_TimeSlot */
static int hf_nbap_pCCPCH_Power = -1;             /* PCCPCH_Power */
static int hf_nbap_sCTD_Indicator = -1;           /* SCTD_Indicator */
static int hf_nbap_TimeSlotConfigurationList_Cell_SetupRqstTDD_item = -1;  /* TimeSlotConfigurationItem_Cell_SetupRqstTDD */
static int hf_nbap_timeSlotStatus = -1;           /* TimeSlotStatus */
static int hf_nbap_timeSlotDirection = -1;        /* TimeSlotDirection */
static int hf_nbap_TimeSlotConfigurationList_LCR_Cell_SetupRqstTDD_item = -1;  /* TimeSlotConfigurationItem_LCR_Cell_SetupRqstTDD */
static int hf_nbap_dwPCH_Power = -1;              /* DwPCH_Power */
static int hf_nbap_iPDL_TDD_Parameters = -1;      /* IPDL_TDD_Parameters */
static int hf_nbap_iPDL_TDD_Parameters_LCR = -1;  /* IPDL_TDD_Parameters_LCR */
static int hf_nbap_Cell_Frequency_List_LCR_MulFreq_Cell_SetupRqstTDD_item = -1;  /* Cell_Frequency_Item_LCR_MulFreq_Cell_SetupRqstTDD */
static int hf_nbap_timeSlotConfigurationList_LCR_Cell_SetupRqstTDD = -1;  /* TimeSlotConfigurationList_LCR_Cell_SetupRqstTDD */
static int hf_nbap_SecondaryCPICH_InformationList_Cell_ReconfRqstFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_bCH_information_01 = -1;       /* BCH_information_Cell_ReconfRqstFDD */
static int hf_nbap_CellPortion_InformationList_Cell_ReconfRqstFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_TimeSlotConfigurationList_Cell_ReconfRqstTDD_item = -1;  /* TimeSlotConfigurationItem_Cell_ReconfRqstTDD */
static int hf_nbap_TimeSlotConfigurationList_LCR_Cell_ReconfRqstTDD_item = -1;  /* TimeSlotConfigurationItem_LCR_Cell_ReconfRqstTDD */
static int hf_nbap_cell_Frequency_Add_LCR_MulFreq_Cell_ReconfRqstTDD = -1;  /* Cell_Frequency_Add_LCR_MulFreq_Cell_ReconfRqstTDD */
static int hf_nbap_cell_Frequency_ModifyList_LCR_MulFreq_Cell_ReconfRqstTDD = -1;  /* Cell_Frequency_ModifyList_LCR_MulFreq_Cell_ReconfRqstTDD */
static int hf_nbap_cell_Frequency_Delete_LCR_MulFreq_Cell_ReconfRqstTDD = -1;  /* Cell_Frequency_Delete_LCR_MulFreq_Cell_ReconfRqstTDD */
static int hf_nbap_timeSlotConfigurationList_LCR_Cell_ReconfRqstTDD = -1;  /* TimeSlotConfigurationList_LCR_Cell_ReconfRqstTDD */
static int hf_nbap_Cell_Frequency_ModifyList_LCR_MulFreq_Cell_ReconfRqstTDD_item = -1;  /* Cell_Frequency_ModifyItem_LCR_MulFreq_Cell_ReconfRqstTDD */
static int hf_nbap_no_Failure = -1;               /* No_Failure_ResourceStatusInd */
static int hf_nbap_serviceImpacting = -1;         /* ServiceImpacting_ResourceStatusInd */
static int hf_nbap_local_Cell_InformationList = -1;  /* Local_Cell_InformationList_ResourceStatusInd */
static int hf_nbap_local_Cell_Group_InformationList = -1;  /* Local_Cell_Group_InformationList_ResourceStatusInd */
static int hf_nbap_Local_Cell_InformationList_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_local_CellID = -1;             /* Local_Cell_ID */
static int hf_nbap_addorDeleteIndicator = -1;     /* AddorDeleteIndicator */
static int hf_nbap_Local_Cell_Group_InformationList_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_Power_Local_Cell_Group_InformationList_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_local_Cell_InformationList_01 = -1;  /* Local_Cell_InformationList2_ResourceStatusInd */
static int hf_nbap_local_Cell_Group_InformationList_01 = -1;  /* Local_Cell_Group_InformationList2_ResourceStatusInd */
static int hf_nbap_cCP_InformationList = -1;      /* CCP_InformationList_ResourceStatusInd */
static int hf_nbap_cell_InformationList = -1;     /* Cell_InformationList_ResourceStatusInd */
static int hf_nbap_Local_Cell_InformationList2_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_maximum_DL_PowerCapability = -1;  /* MaximumDL_PowerCapability */
static int hf_nbap_Local_Cell_Group_InformationList2_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_CCP_InformationList_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_Cell_InformationList_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_primary_SCH_Information_01 = -1;  /* P_SCH_Information_ResourceStatusInd */
static int hf_nbap_secondary_SCH_Information_01 = -1;  /* S_SCH_Information_ResourceStatusInd */
static int hf_nbap_primary_CPICH_Information_01 = -1;  /* P_CPICH_Information_ResourceStatusInd */
static int hf_nbap_secondary_CPICH_Information_01 = -1;  /* S_CPICH_InformationList_ResourceStatusInd */
static int hf_nbap_primary_CCPCH_Information_01 = -1;  /* P_CCPCH_Information_ResourceStatusInd */
static int hf_nbap_bCH_Information_01 = -1;       /* BCH_Information_ResourceStatusInd */
static int hf_nbap_secondary_CCPCH_InformationList_01 = -1;  /* S_CCPCH_InformationList_ResourceStatusInd */
static int hf_nbap_pCH_Information_01 = -1;       /* PCH_Information_ResourceStatusInd */
static int hf_nbap_pICH_Information_01 = -1;      /* PICH_Information_ResourceStatusInd */
static int hf_nbap_fACH_InformationList_01 = -1;  /* FACH_InformationList_ResourceStatusInd */
static int hf_nbap_pRACH_InformationList_01 = -1;  /* PRACH_InformationList_ResourceStatusInd */
static int hf_nbap_rACH_InformationList_01 = -1;  /* RACH_InformationList_ResourceStatusInd */
static int hf_nbap_aICH_InformationList_01 = -1;  /* AICH_InformationList_ResourceStatusInd */
static int hf_nbap_sCH_Information_01 = -1;       /* SCH_Information_ResourceStatusInd */
static int hf_nbap_S_CPICH_InformationList_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_S_CCPCH_InformationList_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_FACH_InformationList_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_PRACH_InformationList_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_RACH_InformationList_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_AICH_InformationList_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_FPACH_LCR_InformationList_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_S_CCPCH_InformationListExt_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_S_CCPCH_LCR_InformationListExt_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_PLCCH_InformationList_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_S_CCPCH_768_InformationList_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_PRACH_768_InformationList_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_E_RUCCH_InformationList_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_E_RUCCH_768_InformationList_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_Cell_Frequency_List_Information_LCR_MulFreq_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_UPPCH_LCR_InformationList_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_MultipleFreq_HS_DSCH_Resources_InformationList_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_Power_Local_Cell_Group_InformationList2_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_MultipleFreq_E_DCH_Resources_InformationList_ResourceStatusInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_MIB_SB_SIB_InformationList_SystemInfoUpdateRqst_item = -1;  /* MIB_SB_SIB_InformationItem_SystemInfoUpdateRqst */
static int hf_nbap_iB_Type = -1;                  /* IB_Type */
static int hf_nbap_iB_OC_ID = -1;                 /* IB_OC_ID */
static int hf_nbap_deletionIndicator = -1;        /* DeletionIndicator_SystemInfoUpdate */
static int hf_nbap_no_Deletion = -1;              /* No_Deletion_SystemInfoUpdate */
static int hf_nbap_yes_Deletion = -1;             /* NULL */
static int hf_nbap_sIB_Originator = -1;           /* SIB_Originator */
static int hf_nbap_iB_SG_REP = -1;                /* IB_SG_REP */
static int hf_nbap_segmentInformationList = -1;   /* SegmentInformationList_SystemInfoUpdate */
static int hf_nbap_SegmentInformationListIE_SystemInfoUpdate_item = -1;  /* SegmentInformationItem_SystemInfoUpdate */
static int hf_nbap_iB_SG_POS = -1;                /* IB_SG_POS */
static int hf_nbap_segment_Type = -1;             /* Segment_Type */
static int hf_nbap_iB_SG_DATA = -1;               /* IB_SG_DATA */
static int hf_nbap_Additional_HS_Cell_Information_RL_Setup_List_item = -1;  /* Additional_HS_Cell_Information_RL_Setup_ItemIEs */
static int hf_nbap_hSPDSCH_RL_ID = -1;            /* RL_ID */
static int hf_nbap_hS_DSCH_FDD_Secondary_Serving_Information = -1;  /* HS_DSCH_FDD_Secondary_Serving_Information */
static int hf_nbap_minUL_ChannelisationCodeLength = -1;  /* MinUL_ChannelisationCodeLength */
static int hf_nbap_maxNrOfUL_DPDCHs = -1;         /* MaxNrOfUL_DPDCHs */
static int hf_nbap_ul_DPCCH_SlotFormat = -1;      /* UL_DPCCH_SlotFormat */
static int hf_nbap_diversityMode = -1;            /* DiversityMode */
static int hf_nbap_not_Used_sSDT_CellID_Length = -1;  /* NULL */
static int hf_nbap_not_Used_s_FieldLength = -1;   /* NULL */
static int hf_nbap_dl_DPCH_SlotFormat = -1;       /* DL_DPCH_SlotFormat */
static int hf_nbap_tFCI_SignallingMode = -1;      /* TFCI_SignallingMode */
static int hf_nbap_not_Used_pDSCH_RL_ID = -1;     /* NULL */
static int hf_nbap_not_Used_pDSCH_CodeMapping = -1;  /* NULL */
static int hf_nbap_powerOffsetInformation_01 = -1;  /* PowerOffsetInformation_RL_SetupRqstFDD */
static int hf_nbap_pO2_ForTPC_Bits = -1;          /* PowerOffset */
static int hf_nbap_RL_InformationList_RL_SetupRqstFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_frameOffset = -1;              /* FrameOffset */
static int hf_nbap_chipOffset = -1;               /* ChipOffset */
static int hf_nbap_diversityControlField = -1;    /* DiversityControlField */
static int hf_nbap_not_Used_sSDT_Cell_Identity = -1;  /* NULL */
static int hf_nbap_hSDSCH_Configured_Indicator = -1;  /* HSDSCH_Configured_Indicator */
static int hf_nbap_powerOffsetInformation_02 = -1;  /* PowerOffsetInformation_F_DPCH_RL_SetupRqstFDD */
static int hf_nbap_UL_CCTrCH_InformationList_RL_SetupRqstTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_uL_DPCH_Information = -1;      /* UL_DPCH_Information_RL_SetupRqstTDD */
static int hf_nbap_tdd_DPCHOffset = -1;           /* TDD_DPCHOffset */
static int hf_nbap_uL_Timeslot_Information = -1;  /* UL_Timeslot_Information */
static int hf_nbap_uL_TimeslotLCR_Information = -1;  /* UL_TimeslotLCR_Information */
static int hf_nbap_uL_Timeslot768_Information = -1;  /* UL_Timeslot768_Information */
static int hf_nbap_DL_CCTrCH_InformationList_RL_SetupRqstTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_tdd_TPC_DownlinkStepSize = -1;  /* TDD_TPC_DownlinkStepSize */
static int hf_nbap_cCTrCH_TPCList = -1;           /* CCTrCH_TPCList_RL_SetupRqstTDD */
static int hf_nbap_dL_DPCH_Information = -1;      /* DL_DPCH_Information_RL_SetupRqstTDD */
static int hf_nbap_CCTrCH_TPCList_RL_SetupRqstTDD_item = -1;  /* CCTrCH_TPCItem_RL_SetupRqstTDD */
static int hf_nbap_dL_Timeslot_Information = -1;  /* DL_Timeslot_Information */
static int hf_nbap_dL_TimeslotLCR_Information = -1;  /* DL_TimeslotLCR_Information */
static int hf_nbap_tstdIndicator = -1;            /* TSTD_Indicator */
static int hf_nbap_dL_Timeslot768_Information = -1;  /* DL_Timeslot768_Information */
static int hf_nbap_specialBurstScheduling = -1;   /* SpecialBurstScheduling */
static int hf_nbap_dL_TimeSlotISCPInfo = -1;      /* DL_TimeslotISCPInfo */
static int hf_nbap_Additional_HS_Cell_Information_Response_List_item = -1;  /* Additional_HS_Cell_Information_Response_ItemIEs */
static int hf_nbap_RL_InformationResponseList_RL_SetupRspFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_diversityIndication = -1;      /* DiversityIndication_RL_SetupRspFDD */
static int hf_nbap_not_Used_dSCH_InformationResponseList = -1;  /* NULL */
static int hf_nbap_sSDT_SupportIndicator = -1;    /* SSDT_SupportIndicator */
static int hf_nbap_combining = -1;                /* Combining_RL_SetupRspFDD */
static int hf_nbap_nonCombiningOrFirstRL = -1;    /* NonCombiningOrFirstRL_RL_SetupRspFDD */
static int hf_nbap_dCH_InformationResponse = -1;  /* DCH_InformationResponse */
static int hf_nbap_uL_TimeSlot_ISCP_Info = -1;    /* UL_TimeSlot_ISCP_Info */
static int hf_nbap_ul_PhysCH_SF_Variation = -1;   /* UL_PhysCH_SF_Variation */
static int hf_nbap_dCH_InformationResponseList = -1;  /* DCH_InformationResponseList_RL_SetupRspTDD */
static int hf_nbap_dSCH_InformationResponseList = -1;  /* DSCH_InformationResponseList_RL_SetupRspTDD */
static int hf_nbap_uSCH_InformationResponseList = -1;  /* USCH_InformationResponseList_RL_SetupRspTDD */
static int hf_nbap_uL_TimeSlot_ISCP_LCR_Info = -1;  /* UL_TimeSlot_ISCP_LCR_Info */
static int hf_nbap_generalCause = -1;             /* GeneralCauseList_RL_SetupFailureFDD */
static int hf_nbap_rLSpecificCause = -1;          /* RLSpecificCauseList_RL_SetupFailureFDD */
static int hf_nbap_unsuccessful_RL_InformationRespList_RL_SetupFailureFDD = -1;  /* Unsuccessful_RL_InformationRespList_RL_SetupFailureFDD */
static int hf_nbap_successful_RL_InformationRespList_RL_SetupFailureFDD = -1;  /* Successful_RL_InformationRespList_RL_SetupFailureFDD */
static int hf_nbap_Unsuccessful_RL_InformationRespList_RL_SetupFailureFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_Successful_RL_InformationRespList_RL_SetupFailureFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_diversityIndication_01 = -1;   /* DiversityIndication_RL_SetupFailureFDD */
static int hf_nbap_not_Used_tFCI2_BearerInformationResponse = -1;  /* NULL */
static int hf_nbap_combining_01 = -1;             /* Combining_RL_SetupFailureFDD */
static int hf_nbap_nonCombiningOrFirstRL_01 = -1;  /* NonCombiningOrFirstRL_RL_SetupFailureFDD */
static int hf_nbap_generalCause_01 = -1;          /* GeneralCauseList_RL_SetupFailureTDD */
static int hf_nbap_rLSpecificCause_01 = -1;       /* RLSpecificCauseList_RL_SetupFailureTDD */
static int hf_nbap_unsuccessful_RL_InformationRespItem_RL_SetupFailureTDD = -1;  /* Unsuccessful_RL_InformationRespItem_RL_SetupFailureTDD */
static int hf_nbap_Additional_HS_Cell_Information_RL_Addition_List_item = -1;  /* Additional_HS_Cell_Information_RL_Addition_ItemIEs */
static int hf_nbap_setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency = -1;  /* Setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency */
static int hf_nbap_addition = -1;                 /* Additional_EDCH_Cell_Information_To_Add_List */
static int hf_nbap_RL_InformationList_RL_AdditionRqstFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_initialDL_TransmissionPower = -1;  /* DL_Power */
static int hf_nbap_maximumDL_Power = -1;          /* DL_Power */
static int hf_nbap_minimumDL_Power = -1;          /* DL_Power */
static int hf_nbap_not_Used_sSDT_CellIdentity = -1;  /* NULL */
static int hf_nbap_UL_CCTrCH_InformationList_RL_AdditionRqstTDD_item = -1;  /* UL_CCTrCH_InformationItem_RL_AdditionRqstTDD */
static int hf_nbap_uL_DPCH_Information_01 = -1;   /* UL_DPCH_InformationList_RL_AdditionRqstTDD */
static int hf_nbap_DL_CCTrCH_InformationList_RL_AdditionRqstTDD_item = -1;  /* DL_CCTrCH_InformationItem_RL_AdditionRqstTDD */
static int hf_nbap_dL_DPCH_Information_01 = -1;   /* DL_DPCH_InformationList_RL_AdditionRqstTDD */
static int hf_nbap_initial_DL_Transmission_Power = -1;  /* DL_Power */
static int hf_nbap_Additional_HS_Cell_Change_Information_Response_List_item = -1;  /* Additional_HS_Cell_Change_Information_Response_ItemIEs */
static int hf_nbap_hS_DSCH_Secondary_Serving_Cell_Change_Information_Response = -1;  /* HS_DSCH_Secondary_Serving_Cell_Change_Information_Response */
static int hf_nbap_RL_InformationResponseList_RL_AdditionRspFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_diversityIndication_02 = -1;   /* DiversityIndication_RL_AdditionRspFDD */
static int hf_nbap_combining_02 = -1;             /* Combining_RL_AdditionRspFDD */
static int hf_nbap_non_combining = -1;            /* Non_Combining_RL_AdditionRspFDD */
static int hf_nbap_dCH_Information = -1;          /* DCH_Information_RL_AdditionRspTDD */
static int hf_nbap_dSCH_InformationResponseList_01 = -1;  /* DSCH_InformationResponseList_RL_AdditionRspTDD */
static int hf_nbap_uSCH_InformationResponseList_01 = -1;  /* USCH_InformationResponseList_RL_AdditionRspTDD */
static int hf_nbap_diversityIndication_03 = -1;   /* DiversityIndication_RL_AdditionRspTDD */
static int hf_nbap_combining_03 = -1;             /* Combining_RL_AdditionRspTDD */
static int hf_nbap_non_Combining = -1;            /* Non_Combining_RL_AdditionRspTDD */
static int hf_nbap_uL_TimeSlot_ISCP_InfoLCR = -1;  /* UL_TimeSlot_ISCP_LCR_Info */
static int hf_nbap_generalCause_02 = -1;          /* GeneralCauseList_RL_AdditionFailureFDD */
static int hf_nbap_rLSpecificCause_02 = -1;       /* RLSpecificCauseList_RL_AdditionFailureFDD */
static int hf_nbap_unsuccessful_RL_InformationRespList_RL_AdditionFailureFDD = -1;  /* Unsuccessful_RL_InformationRespList_RL_AdditionFailureFDD */
static int hf_nbap_successful_RL_InformationRespList_RL_AdditionFailureFDD = -1;  /* Successful_RL_InformationRespList_RL_AdditionFailureFDD */
static int hf_nbap_Unsuccessful_RL_InformationRespList_RL_AdditionFailureFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_Successful_RL_InformationRespList_RL_AdditionFailureFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_diversityIndication_04 = -1;   /* DiversityIndication_RL_AdditionFailureFDD */
static int hf_nbap_combining_04 = -1;             /* Combining_RL_AdditionFailureFDD */
static int hf_nbap_non_Combining_01 = -1;         /* Non_Combining_RL_AdditionFailureFDD */
static int hf_nbap_generalCause_03 = -1;          /* GeneralCauseList_RL_AdditionFailureTDD */
static int hf_nbap_rLSpecificCause_03 = -1;       /* RLSpecificCauseList_RL_AdditionFailureTDD */
static int hf_nbap_unsuccessful_RL_InformationRespItem_RL_AdditionFailureTDD = -1;  /* Unsuccessful_RL_InformationRespItem_RL_AdditionFailureTDD */
static int hf_nbap_Additional_HS_Cell_Information_RL_Reconf_Prep_item = -1;  /* Additional_HS_Cell_Information_RL_Reconf_Prep_ItemIEs */
static int hf_nbap_hS_DSCH_Secondary_Serving_Information_To_Modify = -1;  /* HS_DSCH_Secondary_Serving_Information_To_Modify */
static int hf_nbap_hS_HS_DSCH_Secondary_Serving_Remove = -1;  /* HS_DSCH_Secondary_Serving_Remove */
static int hf_nbap_setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency = -1;  /* Setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency */
static int hf_nbap_not_Used_sSDT_CellIDLength = -1;  /* NULL */
static int hf_nbap_powerOffsetInformation_03 = -1;  /* PowerOffsetInformation_RL_ReconfPrepFDD */
static int hf_nbap_DCH_DeleteList_RL_ReconfPrepFDD_item = -1;  /* DCH_DeleteItem_RL_ReconfPrepFDD */
static int hf_nbap_RL_InformationList_RL_ReconfPrepFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_maxDL_Power = -1;              /* DL_Power */
static int hf_nbap_minDL_Power = -1;              /* DL_Power */
static int hf_nbap_not_Used_sSDT_Indication = -1;  /* NULL */
static int hf_nbap_powerOffsetInformation_04 = -1;  /* PowerOffsetInformation_F_DPCH_RL_ReconfPrepFDD */
static int hf_nbap_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_item = -1;  /* UL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD */
static int hf_nbap_ul_DPCH_InformationList = -1;  /* UL_DPCH_InformationAddList_RL_ReconfPrepTDD */
static int hf_nbap_uL_Timeslot_InformationLCR = -1;  /* UL_TimeslotLCR_Information */
static int hf_nbap_MultipleRL_UL_DPCH_InformationAddList_RL_ReconfPrepTDD_item = -1;  /* MultipleRL_UL_DPCH_InformationAddListIE_RL_ReconfPrepTDD */
static int hf_nbap_ul_DPCH_InformationListLCR = -1;  /* UL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD */
static int hf_nbap_ul_sir_target = -1;            /* UL_SIR */
static int hf_nbap_tDD_TPC_UplinkStepSize_LCR = -1;  /* TDD_TPC_UplinkStepSize_LCR */
static int hf_nbap_uL_Timeslot_Information768 = -1;  /* UL_Timeslot768_Information */
static int hf_nbap_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_item = -1;  /* UL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD */
static int hf_nbap_ul_DPCH_InformationAddList = -1;  /* UL_DPCH_InformationModify_AddList_RL_ReconfPrepTDD */
static int hf_nbap_ul_DPCH_InformationModifyList = -1;  /* UL_DPCH_InformationModify_ModifyList_RL_ReconfPrepTDD */
static int hf_nbap_ul_DPCH_InformationDeleteList = -1;  /* UL_DPCH_InformationModify_DeleteList_RL_ReconfPrepTDD */
static int hf_nbap_uL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD = -1;  /* UL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD */
static int hf_nbap_UL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD_item = -1;  /* UL_Timeslot_InformationModify_ModifyItem_RL_ReconfPrepTDD */
static int hf_nbap_uL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD = -1;  /* UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD */
static int hf_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD_item = -1;  /* UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD */
static int hf_nbap_UL_TimeslotLCR_InformationModify_ModifyList_RL_ReconfPrepTDD_item = -1;  /* UL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD */
static int hf_nbap_uL_Code_InformationModify_ModifyList_RL_ReconfPrepTDDLCR = -1;  /* UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDDLCR */
static int hf_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDDLCR_item = -1;  /* UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDDLCR */
static int hf_nbap_UL_Timeslot768_InformationModify_ModifyList_RL_ReconfPrepTDD_item = -1;  /* UL_Timeslot_768_InformationModify_ModifyItem_RL_ReconfPrepTDD */
static int hf_nbap_uL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD768 = -1;  /* UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD768 */
static int hf_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD768_item = -1;  /* UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD768 */
static int hf_nbap_UL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD_item = -1;  /* UL_DPCH_InformationModify_DeleteItem_RL_ReconfPrepTDD */
static int hf_nbap_MultipleRL_UL_DPCH_InformationModifyList_RL_ReconfPrepTDD_item = -1;  /* MultipleRL_UL_DPCH_InformationModifyListIE_RL_ReconfPrepTDD */
static int hf_nbap_ul_DPCH_InformationAddListLCR = -1;  /* UL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD */
static int hf_nbap_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_item = -1;  /* UL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD */
static int hf_nbap_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_item = -1;  /* DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD */
static int hf_nbap_cCTrCH_TPCList_01 = -1;        /* CCTrCH_TPCAddList_RL_ReconfPrepTDD */
static int hf_nbap_dl_DPCH_InformationList = -1;  /* DL_DPCH_InformationAddList_RL_ReconfPrepTDD */
static int hf_nbap_CCTrCH_TPCAddList_RL_ReconfPrepTDD_item = -1;  /* CCTrCH_TPCAddItem_RL_ReconfPrepTDD */
static int hf_nbap_dL_Timeslot_InformationLCR = -1;  /* DL_TimeslotLCR_Information */
static int hf_nbap_MultipleRL_DL_DPCH_InformationAddList_RL_ReconfPrepTDD_item = -1;  /* MultipleRL_DL_DPCH_InformationAddListIE_RL_ReconfPrepTDD */
static int hf_nbap_dl_DPCH_InformationListLCR = -1;  /* DL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD */
static int hf_nbap_cCTrCH_Initial_DL_Power = -1;  /* DL_Power */
static int hf_nbap_tDD_TPC_DownlinkStepSize = -1;  /* TDD_TPC_DownlinkStepSize */
static int hf_nbap_cCTrCH_Maximum_DL_Power_InformationAdd_RL_ReconfPrepTDD = -1;  /* DL_Power */
static int hf_nbap_cCTrCH_Minimum_DL_Power_InformationAdd_RL_ReconfPrepTDD = -1;  /* DL_Power */
static int hf_nbap_dL_Timeslot_Information768 = -1;  /* DL_Timeslot768_Information */
static int hf_nbap_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_item = -1;  /* DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD */
static int hf_nbap_cCTrCH_TPCList_02 = -1;        /* CCTrCH_TPCModifyList_RL_ReconfPrepTDD */
static int hf_nbap_dl_DPCH_InformationAddList = -1;  /* DL_DPCH_InformationModify_AddList_RL_ReconfPrepTDD */
static int hf_nbap_dl_DPCH_InformationModifyList = -1;  /* DL_DPCH_InformationModify_ModifyList_RL_ReconfPrepTDD */
static int hf_nbap_dl_DPCH_InformationDeleteList = -1;  /* DL_DPCH_InformationModify_DeleteList_RL_ReconfPrepTDD */
static int hf_nbap_CCTrCH_TPCModifyList_RL_ReconfPrepTDD_item = -1;  /* CCTrCH_TPCModifyItem_RL_ReconfPrepTDD */
static int hf_nbap_dL_Timeslot_InformationAddModify_ModifyList_RL_ReconfPrepTDD = -1;  /* DL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD */
static int hf_nbap_DL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD_item = -1;  /* DL_Timeslot_InformationModify_ModifyItem_RL_ReconfPrepTDD */
static int hf_nbap_dL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD = -1;  /* DL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD */
static int hf_nbap_DL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD_item = -1;  /* DL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD */
static int hf_nbap_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD_item = -1;  /* DL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD */
static int hf_nbap_dL_Code_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD = -1;  /* DL_Code_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD */
static int hf_nbap_DL_Code_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD_item = -1;  /* DL_Code_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD */
static int hf_nbap_DL_Timeslot_768_InformationModify_ModifyList_RL_ReconfPrepTDD_item = -1;  /* DL_Timeslot_768_InformationModify_ModifyItem_RL_ReconfPrepTDD */
static int hf_nbap_dL_Code_768_InformationModify_ModifyList_RL_ReconfPrepTDD = -1;  /* DL_Code_768_InformationModify_ModifyList_RL_ReconfPrepTDD */
static int hf_nbap_DL_Code_768_InformationModify_ModifyList_RL_ReconfPrepTDD_item = -1;  /* DL_Code_768_InformationModify_ModifyItem_RL_ReconfPrepTDD */
static int hf_nbap_dPCH_ID768 = -1;               /* DPCH_ID768 */
static int hf_nbap_DL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD_item = -1;  /* DL_DPCH_InformationModify_DeleteItem_RL_ReconfPrepTDD */
static int hf_nbap_MultipleRL_DL_DPCH_InformationModifyList_RL_ReconfPrepTDD_item = -1;  /* MultipleRL_DL_DPCH_InformationModifyListIE_RL_ReconfPrepTDD */
static int hf_nbap_dl_DPCH_InformationAddListLCR = -1;  /* DL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD */
static int hf_nbap_tDD_TPC_DownlinkStepSize_InformationModify_RL_ReconfPrepTDD = -1;  /* TDD_TPC_DownlinkStepSize */
static int hf_nbap_cCTrCH_Maximum_DL_Power_InformationModify_RL_ReconfPrepTDD = -1;  /* DL_Power */
static int hf_nbap_cCTrCH_Minimum_DL_Power_InformationModify_RL_ReconfPrepTDD = -1;  /* DL_Power */
static int hf_nbap_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_item = -1;  /* DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD */
static int hf_nbap_DCH_DeleteList_RL_ReconfPrepTDD_item = -1;  /* DCH_DeleteItem_RL_ReconfPrepTDD */
static int hf_nbap_DSCH_Information_ModifyList_RL_ReconfPrepTDD_item = -1;  /* DSCH_Information_ModifyItem_RL_ReconfPrepTDD */
static int hf_nbap_DSCH_Information_DeleteList_RL_ReconfPrepTDD_item = -1;  /* DSCH_Information_DeleteItem_RL_ReconfPrepTDD */
static int hf_nbap_USCH_Information_ModifyList_RL_ReconfPrepTDD_item = -1;  /* USCH_Information_ModifyItem_RL_ReconfPrepTDD */
static int hf_nbap_USCH_Information_DeleteList_RL_ReconfPrepTDD_item = -1;  /* USCH_Information_DeleteItem_RL_ReconfPrepTDD */
static int hf_nbap_MultipleRL_Information_RL_ReconfPrepTDD_item = -1;  /* RL_Information_RL_ReconfPrepTDD */
static int hf_nbap_RL_InformationResponseList_RL_ReconfReady_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_dCH_InformationResponseList_RL_ReconfReady = -1;  /* DCH_InformationResponseList_RL_ReconfReady */
static int hf_nbap_dSCH_InformationResponseList_RL_ReconfReady = -1;  /* DSCH_InformationResponseList_RL_ReconfReady */
static int hf_nbap_uSCH_InformationResponseList_RL_ReconfReady = -1;  /* USCH_InformationResponseList_RL_ReconfReady */
static int hf_nbap_generalCause_04 = -1;          /* GeneralCauseList_RL_ReconfFailure */
static int hf_nbap_rLSpecificCause_04 = -1;       /* RLSpecificCauseList_RL_ReconfFailure */
static int hf_nbap_rL_ReconfigurationFailureList_RL_ReconfFailure = -1;  /* RL_ReconfigurationFailureList_RL_ReconfFailure */
static int hf_nbap_RL_ReconfigurationFailureList_RL_ReconfFailure_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_Additional_HS_Cell_Information_RL_Reconf_Req_item = -1;  /* Additional_HS_Cell_Information_RL_Reconf_Req_ItemIEs */
static int hf_nbap_hS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised = -1;  /* HS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised */
static int hf_nbap_hS_DSCH_Secondary_Serving_Remove = -1;  /* HS_DSCH_Secondary_Serving_Remove */
static int hf_nbap_ul_TFCS = -1;                  /* TFCS */
static int hf_nbap_dl_TFCS = -1;                  /* TFCS */
static int hf_nbap_DCH_DeleteList_RL_ReconfRqstFDD_item = -1;  /* DCH_DeleteItem_RL_ReconfRqstFDD */
static int hf_nbap_RL_InformationList_RL_ReconfRqstFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_MultipleRL_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_item = -1;  /* MultipleRL_DL_CCTrCH_InformationModifyListIE_RL_ReconfRqstTDD */
static int hf_nbap_dl_DPCH_LCR_InformationModifyList = -1;  /* DL_DPCH_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD */
static int hf_nbap_cCTrCH_Maximum_DL_Power_InformationModify_RL_ReconfRqstTDD = -1;  /* DL_Power */
static int hf_nbap_cCTrCH_Minimum_DL_Power_InformationModify_RL_ReconfRqstTDD = -1;  /* DL_Power */
static int hf_nbap_dL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD = -1;  /* DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD */
static int hf_nbap_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD_item = -1;  /* DL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfRqstTDD */
static int hf_nbap_maxPowerLCR = -1;              /* DL_Power */
static int hf_nbap_minPowerLCR = -1;              /* DL_Power */
static int hf_nbap_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_DCH_DeleteList_RL_ReconfRqstTDD_item = -1;  /* DCH_DeleteItem_RL_ReconfRqstTDD */
static int hf_nbap_Multiple_RL_Information_RL_ReconfRqstTDD_item = -1;  /* RL_Information_RL_ReconfRqstTDD */
static int hf_nbap_RL_InformationResponseList_RL_ReconfRsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_dCH_InformationResponseList_RL_ReconfRsp = -1;  /* DCH_InformationResponseList_RL_ReconfRsp */
static int hf_nbap_RL_informationList_RL_DeletionRqst_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_DL_ReferencePowerInformationList_DL_PC_Rqst_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_dl_ReferencePower = -1;        /* DL_Power */
static int hf_nbap_rL = -1;                       /* RL_DM_Rqst */
static int hf_nbap_rLS = -1;                      /* RL_Set_DM_Rqst */
static int hf_nbap_all_RL = -1;                   /* AllRL_DM_Rqst */
static int hf_nbap_all_RLS = -1;                  /* AllRL_Set_DM_Rqst */
static int hf_nbap_rL_InformationList = -1;       /* RL_InformationList_DM_Rqst */
static int hf_nbap_RL_InformationList_DM_Rqst_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_PUSCH_Info_DM_Rqst_item = -1;  /* PUSCH_ID */
static int hf_nbap_HSSICH_Info_DM_Rqst_item = -1;  /* HS_SICH_ID */
static int hf_nbap_HSSICH_InfoExt_DM_Rqst_item = -1;  /* Extended_HS_SICH_ID */
static int hf_nbap_rL_Set_InformationList_DM_Rqst = -1;  /* RL_Set_InformationList_DM_Rqst */
static int hf_nbap_RL_Set_InformationList_DM_Rqst_item = -1;  /* RL_Set_InformationItem_DM_Rqst */
static int hf_nbap_rL_01 = -1;                    /* RL_DM_Rsp */
static int hf_nbap_rLS_01 = -1;                   /* RL_Set_DM_Rsp */
static int hf_nbap_all_RL_01 = -1;                /* RL_DM_Rsp */
static int hf_nbap_all_RLS_01 = -1;               /* RL_Set_DM_Rsp */
static int hf_nbap_rL_InformationList_DM_Rsp = -1;  /* RL_InformationList_DM_Rsp */
static int hf_nbap_RL_InformationList_DM_Rsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_dedicatedMeasurementValue = -1;  /* DedicatedMeasurementValue */
static int hf_nbap_PUSCH_Info_DM_Rsp_item = -1;   /* PUSCH_ID */
static int hf_nbap_Multiple_PUSCH_InfoList_DM_Rsp_item = -1;  /* Multiple_PUSCH_InfoListIE_DM_Rsp */
static int hf_nbap_pUSCH_ID = -1;                 /* PUSCH_ID */
static int hf_nbap_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp_item = -1;  /* Multiple_DedicatedMeasurementValueItem_TDD_DM_Rsp */
static int hf_nbap_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp_item = -1;  /* Multiple_DedicatedMeasurementValueItem_LCR_TDD_DM_Rsp */
static int hf_nbap_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp_item = -1;  /* Multiple_HSSICHMeasurementValueItem_TDD_DM_Rsp */
static int hf_nbap_Multiple_DedicatedMeasurementValueList_768_TDD_DM_Rsp_item = -1;  /* Multiple_DedicatedMeasurementValueItem_768_TDD_DM_Rsp */
static int hf_nbap_rL_Set_InformationList_DM_Rsp = -1;  /* RL_Set_InformationList_DM_Rsp */
static int hf_nbap_RL_Set_InformationList_DM_Rsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_rL_02 = -1;                    /* RL_DM_Rprt */
static int hf_nbap_rLS_02 = -1;                   /* RL_Set_DM_Rprt */
static int hf_nbap_all_RL_02 = -1;                /* RL_DM_Rprt */
static int hf_nbap_all_RLS_02 = -1;               /* RL_Set_DM_Rprt */
static int hf_nbap_rL_InformationList_DM_Rprt = -1;  /* RL_InformationList_DM_Rprt */
static int hf_nbap_RL_InformationList_DM_Rprt_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_dedicatedMeasurementValueInformation = -1;  /* DedicatedMeasurementValueInformation */
static int hf_nbap_PUSCH_Info_DM_Rprt_item = -1;  /* PUSCH_ID */
static int hf_nbap_Multiple_PUSCH_InfoList_DM_Rprt_item = -1;  /* Multiple_PUSCH_InfoListIE_DM_Rprt */
static int hf_nbap_rL_Set_InformationList_DM_Rprt = -1;  /* RL_Set_InformationList_DM_Rprt */
static int hf_nbap_RL_Set_InformationList_DM_Rprt_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_rL_03 = -1;                    /* RL_RL_FailureInd */
static int hf_nbap_rL_Set = -1;                   /* RL_Set_RL_FailureInd */
static int hf_nbap_cCTrCH = -1;                   /* CCTrCH_RL_FailureInd */
static int hf_nbap_rL_InformationList_RL_FailureInd = -1;  /* RL_InformationList_RL_FailureInd */
static int hf_nbap_RL_InformationList_RL_FailureInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_rL_Set_InformationList_RL_FailureInd = -1;  /* RL_Set_InformationList_RL_FailureInd */
static int hf_nbap_RL_Set_InformationList_RL_FailureInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_cCTrCH_InformationList_RL_FailureInd = -1;  /* CCTrCH_InformationList_RL_FailureInd */
static int hf_nbap_CCTrCH_InformationList_RL_FailureInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_RL_InformationList_RL_PreemptRequiredInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_rL_04 = -1;                    /* RL_RL_RestoreInd */
static int hf_nbap_rL_Set_01 = -1;                /* RL_Set_RL_RestoreInd */
static int hf_nbap_cCTrCH_01 = -1;                /* CCTrCH_RL_RestoreInd */
static int hf_nbap_rL_InformationList_RL_RestoreInd = -1;  /* RL_InformationList_RL_RestoreInd */
static int hf_nbap_RL_InformationList_RL_RestoreInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_rL_Set_InformationList_RL_RestoreInd = -1;  /* RL_Set_InformationList_RL_RestoreInd */
static int hf_nbap_RL_Set_InformationList_RL_RestoreInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_cCTrCH_InformationList_RL_RestoreInd = -1;  /* CCTrCH_InformationList_RL_RestoreInd */
static int hf_nbap_CCTrCH_InformationList_RL_RestoreInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_privateIEs = -1;               /* PrivateIE_Container */
static int hf_nbap_HSDPA_And_EDCH_CellPortion_InformationList_PSCH_ReconfRqst_item = -1;  /* HSDPA_And_EDCH_CellPortion_InformationItem_PSCH_ReconfRqst */
static int hf_nbap_hS_PDSCH_HS_SCCH_ScramblingCode_PSCH_ReconfRqst = -1;  /* DL_ScramblingCode */
static int hf_nbap_hS_PDSCH_FDD_Code_Information_PSCH_ReconfRqst = -1;  /* HS_PDSCH_FDD_Code_Information */
static int hf_nbap_hS_SCCH_FDD_Code_Information_PSCH_ReconfRqst = -1;  /* HS_SCCH_FDD_Code_Information */
static int hf_nbap_hS_PDSCH_HS_SCCH_E_AGCH_E_RGCH_E_HICH_MaxPower_PSCH_ReconfRqst = -1;  /* MaximumTransmissionPower */
static int hf_nbap_e_AGCH_FDD_Code_Information = -1;  /* E_AGCH_FDD_Code_Information */
static int hf_nbap_e_RGCH_E_HICH_FDD_Code_Information = -1;  /* E_RGCH_E_HICH_FDD_Code_Information */
static int hf_nbap_PDSCHSets_AddList_PSCH_ReconfRqst_item = -1;  /* PDSCHSets_AddItem_PSCH_ReconfRqst */
static int hf_nbap_pDSCHSet_ID = -1;              /* PDSCHSet_ID */
static int hf_nbap_pDSCH_InformationList = -1;    /* PDSCH_Information_AddList_PSCH_ReconfRqst */
static int hf_nbap_dL_Timeslot_InformationAddList_PSCH_ReconfRqst = -1;  /* DL_Timeslot_InformationAddList_PSCH_ReconfRqst */
static int hf_nbap_DL_Timeslot_InformationAddList_PSCH_ReconfRqst_item = -1;  /* DL_Timeslot_InformationAddItem_PSCH_ReconfRqst */
static int hf_nbap_dL_Code_InformationAddList_PSCH_ReconfRqst = -1;  /* DL_Code_InformationAddList_PSCH_ReconfRqst */
static int hf_nbap_DL_Code_InformationAddList_PSCH_ReconfRqst_item = -1;  /* DL_Code_InformationAddItem_PSCH_ReconfRqst */
static int hf_nbap_pDSCH_ID = -1;                 /* PDSCH_ID */
static int hf_nbap_dL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst = -1;  /* DL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst */
static int hf_nbap_DL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst_item = -1;  /* DL_Timeslot_InformationAddItem_LCR_PSCH_ReconfRqst */
static int hf_nbap_dL_Code_InformationAddList_LCR_PSCH_ReconfRqst = -1;  /* DL_Code_InformationAddList_LCR_PSCH_ReconfRqst */
static int hf_nbap_DL_Code_InformationAddList_LCR_PSCH_ReconfRqst_item = -1;  /* DL_Code_InformationAddItem_LCR_PSCH_ReconfRqst */
static int hf_nbap_dL_Timeslot_InformationAddList_768_PSCH_ReconfRqst = -1;  /* DL_Timeslot_InformationAddList_768_PSCH_ReconfRqst */
static int hf_nbap_DL_Timeslot_InformationAddList_768_PSCH_ReconfRqst_item = -1;  /* DL_Timeslot_InformationAddItem_768_PSCH_ReconfRqst */
static int hf_nbap_dL_Code_InformationAddList_768_PSCH_ReconfRqst = -1;  /* DL_Code_InformationAddList_768_PSCH_ReconfRqst */
static int hf_nbap_DL_Code_InformationAddList_768_PSCH_ReconfRqst_item = -1;  /* DL_Code_InformationAddItem_768_PSCH_ReconfRqst */
static int hf_nbap_pDSCH_ID768 = -1;              /* PDSCH_ID768 */
static int hf_nbap_PDSCHSets_ModifyList_PSCH_ReconfRqst_item = -1;  /* PDSCHSets_ModifyItem_PSCH_ReconfRqst */
static int hf_nbap_pDSCH_InformationList_01 = -1;  /* PDSCH_Information_ModifyList_PSCH_ReconfRqst */
static int hf_nbap_dL_Timeslot_InformationModifyList_PSCH_ReconfRqst = -1;  /* DL_Timeslot_InformationModifyList_PSCH_ReconfRqst */
static int hf_nbap_DL_Timeslot_InformationModifyList_PSCH_ReconfRqst_item = -1;  /* DL_Timeslot_InformationModifyItem_PSCH_ReconfRqst */
static int hf_nbap_dL_Code_InformationModifyList_PSCH_ReconfRqst = -1;  /* DL_Code_InformationModifyList_PSCH_ReconfRqst */
static int hf_nbap_DL_Code_InformationModifyList_PSCH_ReconfRqst_item = -1;  /* DL_Code_InformationModifyItem_PSCH_ReconfRqst */
static int hf_nbap_dL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst = -1;  /* DL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst */
static int hf_nbap_DL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst_item = -1;  /* DL_Timeslot_LCR_InformationModifyItem_PSCH_ReconfRqst */
static int hf_nbap_dL_Code_LCR_InformationModifyList_PSCH_ReconfRqst = -1;  /* DL_Code_LCR_InformationModifyList_PSCH_ReconfRqst */
static int hf_nbap_DL_Code_LCR_InformationModifyList_PSCH_ReconfRqst_item = -1;  /* DL_Code_LCR_InformationModifyItem_PSCH_ReconfRqst */
static int hf_nbap_dL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst = -1;  /* DL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst */
static int hf_nbap_DL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst_item = -1;  /* DL_Timeslot_768_InformationModifyItem_PSCH_ReconfRqst */
static int hf_nbap_dL_Code_768_InformationModifyList_PSCH_ReconfRqst = -1;  /* DL_Code_768_InformationModifyList_PSCH_ReconfRqst */
static int hf_nbap_DL_Code_768_InformationModifyList_PSCH_ReconfRqst_item = -1;  /* DL_Code_768_InformationModifyItem_PSCH_ReconfRqst */
static int hf_nbap_PDSCHSets_DeleteList_PSCH_ReconfRqst_item = -1;  /* PDSCHSets_DeleteItem_PSCH_ReconfRqst */
static int hf_nbap_PUSCHSets_AddList_PSCH_ReconfRqst_item = -1;  /* PUSCHSets_AddItem_PSCH_ReconfRqst */
static int hf_nbap_pUSCHSet_ID = -1;              /* PUSCHSet_ID */
static int hf_nbap_pUSCH_InformationList = -1;    /* PUSCH_Information_AddList_PSCH_ReconfRqst */
static int hf_nbap_uL_Timeslot_InformationAddList_PSCH_ReconfRqst = -1;  /* UL_Timeslot_InformationAddList_PSCH_ReconfRqst */
static int hf_nbap_UL_Timeslot_InformationAddList_PSCH_ReconfRqst_item = -1;  /* UL_Timeslot_InformationAddItem_PSCH_ReconfRqst */
static int hf_nbap_uL_Code_InformationAddList_PSCH_ReconfRqst = -1;  /* UL_Code_InformationAddList_PSCH_ReconfRqst */
static int hf_nbap_UL_Code_InformationAddList_PSCH_ReconfRqst_item = -1;  /* UL_Code_InformationAddItem_PSCH_ReconfRqst */
static int hf_nbap_uL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst = -1;  /* UL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst */
static int hf_nbap_UL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst_item = -1;  /* UL_Timeslot_InformationAddItem_LCR_PSCH_ReconfRqst */
static int hf_nbap_uL_Code_InformationAddList_LCR_PSCH_ReconfRqst = -1;  /* UL_Code_InformationAddList_LCR_PSCH_ReconfRqst */
static int hf_nbap_UL_Code_InformationAddList_LCR_PSCH_ReconfRqst_item = -1;  /* UL_Code_InformationAddItem_LCR_PSCH_ReconfRqst */
static int hf_nbap_uL_Timeslot_InformationAddList_768_PSCH_ReconfRqst = -1;  /* UL_Timeslot_InformationAddList_768_PSCH_ReconfRqst */
static int hf_nbap_UL_Timeslot_InformationAddList_768_PSCH_ReconfRqst_item = -1;  /* UL_Timeslot_InformationAddItem_768_PSCH_ReconfRqst */
static int hf_nbap_uL_Code_InformationAddList_768_PSCH_ReconfRqst = -1;  /* UL_Code_InformationAddList_768_PSCH_ReconfRqst */
static int hf_nbap_UL_Code_InformationAddList_768_PSCH_ReconfRqst_item = -1;  /* UL_Code_InformationAddItem_768_PSCH_ReconfRqst */
static int hf_nbap_PUSCHSets_ModifyList_PSCH_ReconfRqst_item = -1;  /* PUSCHSets_ModifyItem_PSCH_ReconfRqst */
static int hf_nbap_pUSCH_InformationList_01 = -1;  /* PUSCH_Information_ModifyList_PSCH_ReconfRqst */
static int hf_nbap_uL_Timeslot_InformationModifyList_PSCH_ReconfRqst = -1;  /* UL_Timeslot_InformationModifyList_PSCH_ReconfRqst */
static int hf_nbap_UL_Timeslot_InformationModifyList_PSCH_ReconfRqst_item = -1;  /* UL_Timeslot_InformationModifyItem_PSCH_ReconfRqst */
static int hf_nbap_uL_Code_InformationModifyList_PSCH_ReconfRqst = -1;  /* UL_Code_InformationModifyList_PSCH_ReconfRqst */
static int hf_nbap_UL_Code_InformationModifyList_PSCH_ReconfRqst_item = -1;  /* UL_Code_InformationModifyItem_PSCH_ReconfRqst */
static int hf_nbap_uL_Timeslot_InformationModifyList_LCR_PSCH_ReconfRqst = -1;  /* UL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst */
static int hf_nbap_UL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst_item = -1;  /* UL_Timeslot_LCR_InformationModifyItem_PSCH_ReconfRqst */
static int hf_nbap_uL_Code_LCR_InformationModifyList_PSCH_ReconfRqst = -1;  /* UL_Code_LCR_InformationModifyList_PSCH_ReconfRqst */
static int hf_nbap_UL_Code_LCR_InformationModifyList_PSCH_ReconfRqst_item = -1;  /* UL_Code_LCR_InformationModifyItem_PSCH_ReconfRqst */
static int hf_nbap_uL_Timeslot_InformationModifyList_768_PSCH_ReconfRqst = -1;  /* UL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst */
static int hf_nbap_UL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst_item = -1;  /* UL_Timeslot_768_InformationModifyItem_PSCH_ReconfRqst */
static int hf_nbap_uL_Code_768_InformationModifyList_PSCH_ReconfRqst = -1;  /* UL_Code_768_InformationModifyList_PSCH_ReconfRqst */
static int hf_nbap_UL_Code_768_InformationModifyList_PSCH_ReconfRqst_item = -1;  /* UL_Code_768_InformationModifyItem_PSCH_ReconfRqst */
static int hf_nbap_PUSCHSets_DeleteList_PSCH_ReconfRqst_item = -1;  /* PUSCHSets_DeleteItem_PSCH_ReconfRqst */
static int hf_nbap_dL_HS_PDSCH_Timeslot_Information_PSCH_ReconfRqst = -1;  /* DL_HS_PDSCH_Timeslot_Information_PSCH_ReconfRqst */
static int hf_nbap_DL_HS_PDSCH_Timeslot_Information_PSCH_ReconfRqst_item = -1;  /* DL_HS_PDSCH_Timeslot_InformationItem_PSCH_ReconfRqst */
static int hf_nbap_dl_HS_PDSCH_Codelist_PSCH_ReconfRqst = -1;  /* DL_HS_PDSCH_Codelist_PSCH_ReconfRqst */
static int hf_nbap_DL_HS_PDSCH_Codelist_PSCH_ReconfRqst_item = -1;  /* TDD_ChannelisationCode */
static int hf_nbap_DL_HS_PDSCH_Timeslot_Information_768_PSCH_ReconfRqst_item = -1;  /* DL_HS_PDSCH_Timeslot_InformationItem_768_PSCH_ReconfRqst */
static int hf_nbap_dl_HS_PDSCH_Codelist_768_PSCH_ReconfRqst = -1;  /* DL_HS_PDSCH_Codelist_768_PSCH_ReconfRqst */
static int hf_nbap_DL_HS_PDSCH_Codelist_768_PSCH_ReconfRqst_item = -1;  /* TDD_ChannelisationCode768 */
static int hf_nbap_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_hS_SCCH_Information_PSCH_ReconfRqst = -1;  /* HS_SCCH_Information_PSCH_ReconfRqst */
static int hf_nbap_hS_SCCH_Information_LCR_PSCH_ReconfRqst = -1;  /* HS_SCCH_Information_LCR_PSCH_ReconfRqst */
static int hf_nbap_HS_SCCH_Information_PSCH_ReconfRqst_item = -1;  /* HS_SCCH_InformationItem_PSCH_ReconfRqst */
static int hf_nbap_hS_SCCH_ID = -1;               /* HS_SCCH_ID */
static int hf_nbap_hS_SCCH_MaxPower = -1;         /* DL_Power */
static int hf_nbap_hS_SICH_Information = -1;      /* HS_SICH_Information_PSCH_ReconfRqst */
static int hf_nbap_HS_SCCH_Information_LCR_PSCH_ReconfRqst_item = -1;  /* HS_SCCH_InformationItem_LCR_PSCH_ReconfRqst */
static int hf_nbap_hS_SICH_Information_LCR = -1;  /* HS_SICH_Information_LCR_PSCH_ReconfRqst */
static int hf_nbap_HS_SCCH_Information_768_PSCH_ReconfRqst_item = -1;  /* HS_SCCH_InformationItem_768_PSCH_ReconfRqst */
static int hf_nbap_hS_SICH_Information_768 = -1;  /* HS_SICH_Information_768_PSCH_ReconfRqst */
static int hf_nbap_HS_SCCH_InformationExt_LCR_PSCH_ReconfRqst_item = -1;  /* HS_SCCH_InformationItem_LCR_PSCH_ReconfRqst */
static int hf_nbap_hS_SCCH_InformationModify_PSCH_ReconfRqst = -1;  /* HS_SCCH_InformationModify_PSCH_ReconfRqst */
static int hf_nbap_hS_SCCH_InformationModify_LCR_PSCH_ReconfRqst = -1;  /* HS_SCCH_InformationModify_LCR_PSCH_ReconfRqst */
static int hf_nbap_hS_SICH_Information_01 = -1;   /* HS_SICH_InformationModify_PSCH_ReconfRqst */
static int hf_nbap_HS_SCCH_InformationModify_LCR_PSCH_ReconfRqst_item = -1;  /* HS_SCCH_InformationModifyItem_LCR_PSCH_ReconfRqst */
static int hf_nbap_hS_SICH_Information_LCR_01 = -1;  /* HS_SICH_InformationModify_LCR_PSCH_ReconfRqst */
static int hf_nbap_HS_SCCH_InformationModifyExt_LCR_PSCH_ReconfRqst_item = -1;  /* HS_SCCH_InformationModifyItem_LCR_PSCH_ReconfRqst */
static int hf_nbap_HS_SCCH_InformationModify_768_PSCH_ReconfRqst_item = -1;  /* HS_SCCH_InformationModifyItem_768_PSCH_ReconfRqst */
static int hf_nbap_hS_SICH_Information_768_01 = -1;  /* HS_SICH_InformationModify_768_PSCH_ReconfRqst */
static int hf_nbap_HS_SCCH_InformationModify_PSCH_ReconfRqst_item = -1;  /* HS_SCCH_InformationModifyItem_PSCH_ReconfRqst */
static int hf_nbap_Delete_From_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_item = -1;  /* Delete_From_HS_SCCH_Resource_PoolItem_PSCH_ReconfRqst */
static int hf_nbap_lTGI_Presence = -1;            /* LTGI_Presence */
static int hf_nbap_sNPL_Reporting_Type = -1;      /* SNPL_Reporting_Type */
static int hf_nbap_e_PUCH_Timeslot_Info = -1;     /* E_PUCH_Timeslot_Info */
static int hf_nbap_E_PUCH_Timeslot_Info_item = -1;  /* TimeSlot */
static int hf_nbap_e_AGCH_Information_PSCH_ReconfRqst = -1;  /* E_AGCH_Information_PSCH_ReconfRqst */
static int hf_nbap_E_AGCH_Information_PSCH_ReconfRqst_item = -1;  /* E_AGCH_InformationItem_PSCH_ReconfRqst */
static int hf_nbap_e_AGCH_MaxPower = -1;          /* DL_Power */
static int hf_nbap_e_AGCH_InformationModify_PSCH_ReconfRqst = -1;  /* E_AGCH_InformationModify_PSCH_ReconfRqst */
static int hf_nbap_E_AGCH_InformationModify_PSCH_ReconfRqst_item = -1;  /* E_AGCH_InformationModifyItem_PSCH_ReconfRqst */
static int hf_nbap_Delete_From_E_AGCH_Resource_Pool_PSCH_ReconfRqst_item = -1;  /* Delete_From_E_AGCH_Resource_PoolItem_PSCH_ReconfRqst */
static int hf_nbap_e_HICH_MaxPower = -1;          /* DL_Power */
static int hf_nbap_e_AGCH_Information_768_PSCH_ReconfRqst = -1;  /* E_AGCH_Information_768_PSCH_ReconfRqst */
static int hf_nbap_E_AGCH_Information_768_PSCH_ReconfRqst_item = -1;  /* E_AGCH_InformationItem_768_PSCH_ReconfRqst */
static int hf_nbap_e_AGCH_InformationModify_768_PSCH_ReconfRqst = -1;  /* E_AGCH_InformationModify_768_PSCH_ReconfRqst */
static int hf_nbap_E_AGCH_InformationModify_768_PSCH_ReconfRqst_item = -1;  /* E_AGCH_InformationModifyItem_768_PSCH_ReconfRqst */
static int hf_nbap_e_PUCH_Timeslot_InfoLCR = -1;  /* E_PUCH_Timeslot_InfoLCR */
static int hf_nbap_E_PUCH_Timeslot_InfoLCR_item = -1;  /* E_PUCH_Timeslot_Item_InfoLCR */
static int hf_nbap_e_PUCH_Codelist_LCR = -1;      /* E_PUCH_Codelist_LCR */
static int hf_nbap_E_PUCH_Codelist_LCR_item = -1;  /* TDD_ChannelisationCode */
static int hf_nbap_e_AGCH_Information_LCR_PSCH_ReconfRqst = -1;  /* E_AGCH_Information_LCR_PSCH_ReconfRqst */
static int hf_nbap_E_AGCH_Information_LCR_PSCH_ReconfRqst_item = -1;  /* E_AGCH_InformationItem_LCR_PSCH_ReconfRqst */
static int hf_nbap_e_AGCH_InformationModify_LCR_PSCH_ReconfRqst = -1;  /* E_AGCH_InformationModify_LCR_PSCH_ReconfRqst */
static int hf_nbap_E_AGCH_InformationModify_LCR_PSCH_ReconfRqst_item = -1;  /* E_AGCH_InformationModifyItem_LCR_PSCH_ReconfRqst */
static int hf_nbap_e_HICH_Information_LCR_PSCH_ReconfRqst = -1;  /* E_HICH_Information_LCR_PSCH_ReconfRqst */
static int hf_nbap_E_HICH_Information_LCR_PSCH_ReconfRqst_item = -1;  /* E_HICH_InformationItem_LCR_PSCH_ReconfRqst */
static int hf_nbap_e_HICH_Type = -1;              /* E_HICH_Type */
static int hf_nbap_e_HICH_InformationModify_LCR_PSCH_ReconfRqst = -1;  /* E_HICH_InformationModify_LCR_PSCH_ReconfRqst */
static int hf_nbap_E_HICH_InformationModify_LCR_PSCH_ReconfRqst_item = -1;  /* E_HICH_InformationModifyItem_LCR_PSCH_ReconfRqst */
static int hf_nbap_Delete_From_E_HICH_Resource_Pool_PSCH_ReconfRqst_item = -1;  /* Delete_From_E_HICH_Resource_PoolItem_PSCH_ReconfRqst */
static int hf_nbap_eRUCCH_SYNC_UL_codes_bitmap = -1;  /* BIT_STRING_SIZE_8 */
static int hf_nbap_Delete_From_HS_SCCH_Resource_PoolExt_PSCH_ReconfRqst_item = -1;  /* Delete_From_HS_SCCH_Resource_PoolItem_PSCH_ReconfRqst */
static int hf_nbap_MultipleFreq_E_PUCH_Timeslot_InformationList_LCR_PSCH_ReconfRqst_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst_item = -1;  /* Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst_Item */
static int hf_nbap_maximum_Target_ReceivedTotalWideBandPower_LCR = -1;  /* Maximum_Target_ReceivedTotalWideBandPower_LCR */
static int hf_nbap_E_HICH_TimeOffset_ExtensionLCR_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_e_HICH_TimeOffsetLCR = -1;     /* E_HICH_TimeOffsetLCR */
static int hf_nbap_generalCause_05 = -1;          /* GeneralCauseList_PSCH_ReconfFailure */
static int hf_nbap_setSpecificCause = -1;         /* SetSpecificCauseList_PSCH_ReconfFailureTDD */
static int hf_nbap_extension_CauseLevel_PSCH_ReconfFailure = -1;  /* Extension_CauseLevel_PSCH_ReconfFailure */
static int hf_nbap_unsuccessful_PDSCHSetList_PSCH_ReconfFailureTDD = -1;  /* Unsuccessful_PDSCHSetList_PSCH_ReconfFailureTDD */
static int hf_nbap_unsuccessful_PUSCHSetList_PSCH_ReconfFailureTDD = -1;  /* Unsuccessful_PUSCHSetList_PSCH_ReconfFailureTDD */
static int hf_nbap_Unsuccessful_PDSCHSetList_PSCH_ReconfFailureTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_Unsuccessful_PUSCHSetList_PSCH_ReconfFailureTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_UARFCNSpecificCauseList_PSCH_ReconfFailureTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_E_HICH_TimeOffset_ReconfFailureTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_hSDSCH_Common_System_Information_ResponseLCR = -1;  /* HSDSCH_Common_System_Information_ResponseLCR */
static int hf_nbap_hSDSCH_Paging_System_Information_ResponseLCR = -1;  /* HSDSCH_Paging_System_Information_ResponseLCR */
static int hf_nbap_common_EDCH_System_Information_ResponseLCR = -1;  /* Common_EDCH_System_Information_ResponseLCR */
static int hf_nbap_communicationContext = -1;     /* CommunicationContextList_Reset */
static int hf_nbap_communicationControlPort = -1;  /* CommunicationControlPortList_Reset */
static int hf_nbap_nodeB = -1;                    /* NULL */
static int hf_nbap_communicationContextInfoList_Reset = -1;  /* CommunicationContextInfoList_Reset */
static int hf_nbap_CommunicationContextInfoList_Reset_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_communicationContextType_Reset = -1;  /* CommunicationContextType_Reset */
static int hf_nbap_nodeB_CommunicationContextID = -1;  /* NodeB_CommunicationContextID */
static int hf_nbap_communicationControlPortInfoList_Reset = -1;  /* CommunicationControlPortInfoList_Reset */
static int hf_nbap_CommunicationControlPortInfoList_Reset_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_cell_03 = -1;                  /* Cell_InfEx_Rqst */
static int hf_nbap_cell_04 = -1;                  /* Cell_InfEx_Rsp */
static int hf_nbap_requestedDataValue = -1;       /* RequestedDataValue */
static int hf_nbap_cell_05 = -1;                  /* Cell_Inf_Rprt */
static int hf_nbap_requestedDataValueInformation = -1;  /* RequestedDataValueInformation */
static int hf_nbap_TimeslotInfo_CellSyncInitiationRqstTDD_item = -1;  /* TimeSlot */
static int hf_nbap_cSBTransmissionID = -1;        /* CSBTransmissionID */
static int hf_nbap_sfn = -1;                      /* SFN */
static int hf_nbap_initialDLTransPower = -1;      /* DL_Power */
static int hf_nbap_cSBMeasurementID = -1;         /* CSBMeasurementID */
static int hf_nbap_synchronisationReportType = -1;  /* SynchronisationReportType */
static int hf_nbap_synchronisationReportCharacteristics = -1;  /* SynchronisationReportCharacteristics */
static int hf_nbap_sYNCDlCodeId = -1;             /* SYNCDlCodeId */
static int hf_nbap_CellSyncBurstTransReconfInfo_CellSyncReconfRqstTDD_item = -1;  /* CellSyncBurstTransInfoItem_CellSyncReconfRqstTDD */
static int hf_nbap_syncFrameNumberToTransmit = -1;  /* SyncFrameNumber */
static int hf_nbap_dlTransPower = -1;             /* DL_Power */
static int hf_nbap_cellSyncBurstMeasInfoList_CellSyncReconfRqstTDD = -1;  /* CellSyncBurstMeasInfoList_CellSyncReconfRqstTDD */
static int hf_nbap_synchronisationReportType_01 = -1;  /* SynchronisationReportTypeIE */
static int hf_nbap_synchronisationReportCharacteristics_01 = -1;  /* SynchronisationReportCharacteristicsIE */
static int hf_nbap_CellSyncBurstMeasInfoListIE_CellSyncReconfRqstTDD_item = -1;  /* CellSyncBurstMeasInfoItem_CellSyncReconfRqstTDD */
static int hf_nbap_syncFrameNrToReceive = -1;     /* SyncFrameNumber */
static int hf_nbap_syncBurstInfo = -1;            /* CellSyncBurstInfoList_CellSyncReconfRqstTDD */
static int hf_nbap_CellSyncBurstInfoList_CellSyncReconfRqstTDD_item = -1;  /* CellSyncBurstInfoItem_CellSyncReconfRqstTDD */
static int hf_nbap_SYNCDlCodeIdTransReconfInfoLCR_CellSyncReconfRqstTDD_item = -1;  /* SYNCDlCodeIdTransReconfItemLCR_CellSyncReconfRqstTDD */
static int hf_nbap_syncFrameNumberforTransmit = -1;  /* SyncFrameNumber */
static int hf_nbap_sYNCDlCodeIdMeasInfoList = -1;  /* SYNCDlCodeIdMeasInfoList_CellSyncReconfRqstTDD */
static int hf_nbap_SYNCDlCodeIdMeasInfoList_CellSyncReconfRqstTDD_item = -1;  /* SYNCDlCodeIdMeasInfoItem_CellSyncReconfRqstTDD */
static int hf_nbap_sYNCDlCodeIdInfoLCR = -1;      /* SYNCDlCodeIdInfoListLCR_CellSyncReconfRqstTDD */
static int hf_nbap_SYNCDlCodeIdInfoListLCR_CellSyncReconfRqstTDD_item = -1;  /* SYNCDlCodeIdInfoItemLCR_CellSyncReconfRqstTDD */
static int hf_nbap_propagationDelayCompensation = -1;  /* TimingAdjustmentValueLCR */
static int hf_nbap_CellAdjustmentInfo_SyncAdjustmentRqstTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_frameAdjustmentValue = -1;     /* FrameAdjustmentValue */
static int hf_nbap_timingAdjustmentValue = -1;    /* TimingAdjustmentValue */
static int hf_nbap_dLTransPower = -1;             /* DL_Power */
static int hf_nbap_generalCause_06 = -1;          /* GeneralCauseList_SyncAdjustmntFailureTDD */
static int hf_nbap_cellSpecificCause = -1;        /* CellSpecificCauseList_SyncAdjustmntFailureTDD */
static int hf_nbap_unsuccessful_cell_InformationRespList_SyncAdjustmntFailureTDD = -1;  /* Unsuccessful_cell_InformationRespList_SyncAdjustmntFailureTDD */
static int hf_nbap_Unsuccessful_cell_InformationRespList_SyncAdjustmntFailureTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_CellSyncInfo_CellSyncReprtTDD_item = -1;  /* CellSyncInfoItemIE_CellSyncReprtTDD */
static int hf_nbap_c_ID_CellSyncReprtTDD = -1;    /* C_ID_IE_CellSyncReprtTDD */
static int hf_nbap_syncReportType_CellSyncReprtTDD = -1;  /* SyncReportTypeIE_CellSyncReprtTDD */
static int hf_nbap_intStdPhSyncInfo_CellSyncReprtTDD = -1;  /* IntStdPhCellSyncInfo_CellSyncReprtTDD */
static int hf_nbap_lateEntrantCell = -1;          /* NULL */
static int hf_nbap_frequencyAcquisition = -1;     /* NULL */
static int hf_nbap_cellSyncBurstMeasuredInfo = -1;  /* CellSyncBurstMeasInfoList_CellSyncReprtTDD */
static int hf_nbap_CellSyncBurstMeasInfoList_CellSyncReprtTDD_item = -1;  /* CellSyncBurstMeasInfoItem_CellSyncReprtTDD */
static int hf_nbap_cellSyncBurstInfo_CellSyncReprtTDD = -1;  /* SEQUENCE_SIZE_1_maxNrOfReceptsPerSyncFrame_OF_CellSyncBurstInfo_CellSyncReprtTDD */
static int hf_nbap_cellSyncBurstInfo_CellSyncReprtTDD_item = -1;  /* CellSyncBurstInfo_CellSyncReprtTDD */
static int hf_nbap_cellSyncBurstAvailable = -1;   /* CellSyncBurstAvailable_CellSyncReprtTDD */
static int hf_nbap_cellSyncBurstNotAvailable = -1;  /* NULL */
static int hf_nbap_cellSyncBurstSIR = -1;         /* CellSyncBurstSIR */
static int hf_nbap_SyncDLCodeIdsMeasInfoList_CellSyncReprtTDD_item = -1;  /* SyncDLCodeIdsMeasInfoItem_CellSyncReprtTDD */
static int hf_nbap_syncDLCodeIdInfo_CellSyncReprtTDD = -1;  /* SyncDLCodeIdInfo_CellSyncReprtTDD */
static int hf_nbap_SyncDLCodeIdInfo_CellSyncReprtTDD_item = -1;  /* SyncDLCodeIdItem_CellSyncReprtTDD */
static int hf_nbap_syncDLCodeIdAvailable = -1;    /* SyncDLCodeIdAvailable_CellSyncReprtTDD */
static int hf_nbap_syncDLCodeIDNotAvailable = -1;  /* NULL */
static int hf_nbap_syncDLCodeIdTiming = -1;       /* CellSyncBurstTimingLCR */
static int hf_nbap_syncDLCodeIdSIR = -1;          /* CellSyncBurstSIR */
static int hf_nbap_DCH_RearrangeList_Bearer_RearrangeInd_item = -1;  /* DCH_RearrangeItem_Bearer_RearrangeInd */
static int hf_nbap_DSCH_RearrangeList_Bearer_RearrangeInd_item = -1;  /* DSCH_RearrangeItem_Bearer_RearrangeInd */
static int hf_nbap_USCH_RearrangeList_Bearer_RearrangeInd_item = -1;  /* USCH_RearrangeItem_Bearer_RearrangeInd */
static int hf_nbap_HSDSCH_RearrangeList_Bearer_RearrangeInd_item = -1;  /* HSDSCH_RearrangeItem_Bearer_RearrangeInd */
static int hf_nbap_E_DCH_RearrangeList_Bearer_RearrangeInd_item = -1;  /* E_DCH_RearrangeItem_Bearer_RearrangeInd */
static int hf_nbap_Additional_EDCH_Cell_Information_Bearer_Rearrangement_List_item = -1;  /* Additional_EDCH_Cell_Information_Bearer_Rearrangement_ItemIEs */
static int hf_nbap_transport_Bearer_Rearrangement_Indicator_for_Additional_EDCH_Separate_Mode = -1;  /* Transport_Bearer_Rearrangement_Indicator_for_Additional_EDCH_Separate_Mode */
static int hf_nbap_DelayedActivationInformationList_RL_ActivationCmdFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_delayed_activation_update = -1;  /* DelayedActivationUpdate */
static int hf_nbap_DelayedActivationInformationList_RL_ActivationCmdTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_nbap_Additional_HS_Cell_Information_RL_Param_Upd_item = -1;  /* Additional_HS_Cell_Information_RL_Param_Upd_ItemIEs */
static int hf_nbap_hS_DSCH_FDD_Secondary_Serving_Update_Information = -1;  /* HS_DSCH_FDD_Secondary_Serving_Update_Information */
static int hf_nbap_Additional_EDCH_Cell_Information_RL_Param_Upd_item = -1;  /* Additional_EDCH_Cell_Information_RL_Param_Upd_ItemIEs */
static int hf_nbap_additional_EDCH_FDD_Update_Information = -1;  /* Additional_EDCH_FDD_Update_Information */
static int hf_nbap_initiatingMessage = -1;        /* InitiatingMessage */
static int hf_nbap_succesfulOutcome = -1;         /* SuccessfulOutcome */
static int hf_nbap_unsuccesfulOutcome = -1;       /* UnsuccessfulOutcome */
static int hf_nbap_outcome = -1;                  /* Outcome */
static int hf_nbap_messageDiscriminator = -1;     /* MessageDiscriminator */
static int hf_nbap_initiatingMessagevalue = -1;   /* InitiatingMessage_value */
static int hf_nbap_successfulOutcome_value = -1;  /* SuccessfulOutcome_value */
static int hf_nbap_unsuccessfulOutcome_value = -1;  /* UnsuccessfulOutcome_value */
static int hf_nbap_outcome_value = -1;            /* Outcome_value */
/* named bits */
static int hf_nbap_PreambleSignatures_signature15 = -1;
static int hf_nbap_PreambleSignatures_signature14 = -1;
static int hf_nbap_PreambleSignatures_signature13 = -1;
static int hf_nbap_PreambleSignatures_signature12 = -1;
static int hf_nbap_PreambleSignatures_signature11 = -1;
static int hf_nbap_PreambleSignatures_signature10 = -1;
static int hf_nbap_PreambleSignatures_signature9 = -1;
static int hf_nbap_PreambleSignatures_signature8 = -1;
static int hf_nbap_PreambleSignatures_signature7 = -1;
static int hf_nbap_PreambleSignatures_signature6 = -1;
static int hf_nbap_PreambleSignatures_signature5 = -1;
static int hf_nbap_PreambleSignatures_signature4 = -1;
static int hf_nbap_PreambleSignatures_signature3 = -1;
static int hf_nbap_PreambleSignatures_signature2 = -1;
static int hf_nbap_PreambleSignatures_signature1 = -1;
static int hf_nbap_PreambleSignatures_signature0 = -1;
static int hf_nbap_RACH_SubChannelNumbers_subCh11 = -1;
static int hf_nbap_RACH_SubChannelNumbers_subCh10 = -1;
static int hf_nbap_RACH_SubChannelNumbers_subCh9 = -1;
static int hf_nbap_RACH_SubChannelNumbers_subCh8 = -1;
static int hf_nbap_RACH_SubChannelNumbers_subCh7 = -1;
static int hf_nbap_RACH_SubChannelNumbers_subCh6 = -1;
static int hf_nbap_RACH_SubChannelNumbers_subCh5 = -1;
static int hf_nbap_RACH_SubChannelNumbers_subCh4 = -1;
static int hf_nbap_RACH_SubChannelNumbers_subCh3 = -1;
static int hf_nbap_RACH_SubChannelNumbers_subCh2 = -1;
static int hf_nbap_RACH_SubChannelNumbers_subCh1 = -1;
static int hf_nbap_RACH_SubChannelNumbers_subCh0 = -1;

/* Initialize the subtree pointers */
static int ett_nbap = -1;
static int ett_nbap_TransportLayerAddress = -1;
static int ett_nbap_TransportLayerAddress_nsap = -1;
static int ett_nbap_ib_sg_data = -1;

static gint ett_nbap_PrivateIE_ID = -1;
static gint ett_nbap_ProcedureID = -1;
static gint ett_nbap_TransactionID = -1;
static gint ett_nbap_ProtocolIE_Container = -1;
static gint ett_nbap_ProtocolIE_Field = -1;
static gint ett_nbap_ProtocolExtensionContainer = -1;
static gint ett_nbap_ProtocolExtensionField = -1;
static gint ett_nbap_PrivateIE_Container = -1;
static gint ett_nbap_PrivateIE_Field = -1;
static gint ett_nbap_ActivationInformation = -1;
static gint ett_nbap_ActivationInformationItem = -1;
static gint ett_nbap_Additional_EDCH_Setup_Info = -1;
static gint ett_nbap_Additional_EDCH_Cell_Information_Setup = -1;
static gint ett_nbap_Additional_EDCH_FDD_Setup_Cell_Information = -1;
static gint ett_nbap_Additional_EDCH_UL_DPCH_Information_Setup = -1;
static gint ett_nbap_Additional_EDCH_F_DPCH_Information = -1;
static gint ett_nbap_Additional_EDCH_RL_Specific_Information_To_Setup_List = -1;
static gint ett_nbap_Additional_EDCH_RL_Specific_Information_To_Setup_ItemIEs = -1;
static gint ett_nbap_Additional_EDCH_Cell_Information_To_Add_List = -1;
static gint ett_nbap_Additional_EDCH_Cell_Information_To_Add_ItemIEs = -1;
static gint ett_nbap_Additional_EDCH_RL_Specific_Information_To_Add_ItemIEs = -1;
static gint ett_nbap_EDCH_Additional_RL_Specific_Information_To_Add_List = -1;
static gint ett_nbap_Additional_EDCH_RL_Specific_Information_To_Modify_List = -1;
static gint ett_nbap_Additional_EDCH_RL_Specific_Information_To_Modify_ItemIEs = -1;
static gint ett_nbap_Additional_EDCH_FDD_Information = -1;
static gint ett_nbap_Additional_EDCH_MAC_d_Flows_Specific_Info_List = -1;
static gint ett_nbap_Additional_EDCH_MAC_d_Flows_Specific_Info = -1;
static gint ett_nbap_Additional_EDCH_Cell_Information_Response_List = -1;
static gint ett_nbap_Additional_EDCH_FDD_Information_Response_ItemIEs = -1;
static gint ett_nbap_EDCH_Additional_RL_Specific_Information_Response_List = -1;
static gint ett_nbap_EDCH_Additional_RL_Specific_Information_Response_ItemIEs = -1;
static gint ett_nbap_Additional_EDCH_Cell_Information_Response_RLReconf_List = -1;
static gint ett_nbap_Additional_EDCH_FDD_Information_Response_RLReconf_Items = -1;
static gint ett_nbap_Additional_Modififed_EDCH_FDD_Information_Response_ItemIEs = -1;
static gint ett_nbap_EDCH_Additional_Modified_RL_Specific_Information_Response_List = -1;
static gint ett_nbap_EDCH_Additional_Modified_RL_Specific_Information_Response_List_Items = -1;
static gint ett_nbap_Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List = -1;
static gint ett_nbap_Additional_EDCH_MAC_d_Flows_Specific_Info_Response = -1;
static gint ett_nbap_Additional_EDCH_Cell_Information_Response_RL_Add_List = -1;
static gint ett_nbap_Additional_EDCH_Cell_Information_Response_RL_Add_ItemIEs = -1;
static gint ett_nbap_Additional_EDCH_Cell_Information_ConfigurationChange_List = -1;
static gint ett_nbap_Additional_EDCH_ConfigurationChange_Info_ItemIEs = -1;
static gint ett_nbap_Additional_EDCH_UL_DPCH_Information_Modify = -1;
static gint ett_nbap_Additional_EDCH_Cell_Information_Removal_List = -1;
static gint ett_nbap_Additional_EDCH_Cell_Information_Removal_Info_ItemIEs = -1;
static gint ett_nbap_Additional_EDCH_FDD_Update_Information = -1;
static gint ett_nbap_Additional_EDCH_DL_Control_Channel_Change_Information_List = -1;
static gint ett_nbap_Additional_EDCH_DL_Control_Channel_Change_Info_ItemIEs = -1;
static gint ett_nbap_AdditionalMeasurementValueList = -1;
static gint ett_nbap_AdditionalMeasurementValue = -1;
static gint ett_nbap_AdditionalTimeSlotListLCR = -1;
static gint ett_nbap_AdditionalTimeSlotLCR = -1;
static gint ett_nbap_Active_Pattern_Sequence_Information = -1;
static gint ett_nbap_Transmission_Gap_Pattern_Sequence_Status_List = -1;
static gint ett_nbap_Transmission_Gap_Pattern_Sequence_Status_List_item = -1;
static gint ett_nbap_AllocationRetentionPriority = -1;
static gint ett_nbap_Angle_Of_Arrival_Value_LCR = -1;
static gint ett_nbap_BCCH_Specific_HSDSCH_RNTI_Information = -1;
static gint ett_nbap_BCCH_Specific_HSDSCH_RNTI_InformationLCR = -1;
static gint ett_nbap_Best_Cell_Portions_Value = -1;
static gint ett_nbap_Best_Cell_Portions_Item = -1;
static gint ett_nbap_Best_Cell_Portions_ValueLCR = -1;
static gint ett_nbap_Best_Cell_Portions_ItemLCR = -1;
static gint ett_nbap_BroadcastCommonTransportBearerIndication = -1;
static gint ett_nbap_Cause = -1;
static gint ett_nbap_Cell_ERNTI_Status_Information = -1;
static gint ett_nbap_Cell_ERNTI_Status_Information_Item = -1;
static gint ett_nbap_Vacant_ERNTI = -1;
static gint ett_nbap_CellSyncBurstTiming = -1;
static gint ett_nbap_CellSyncBurstTimingLCR = -1;
static gint ett_nbap_CommonChannelsCapacityConsumptionLaw = -1;
static gint ett_nbap_CommonChannelsCapacityConsumptionLaw_item = -1;
static gint ett_nbap_Common_EDCH_System_InformationFDD = -1;
static gint ett_nbap_Common_E_DCH_UL_DPCH_InfoItem = -1;
static gint ett_nbap_Common_E_DCH_EDPCH_InfoItem = -1;
static gint ett_nbap_Common_E_DCH_InfoItem = -1;
static gint ett_nbap_Common_E_DCH_HSDPCCH_InfoItem = -1;
static gint ett_nbap_Common_E_DCH_CQI_Info = -1;
static gint ett_nbap_Common_E_DCH_Preamble_Control_InfoItem = -1;
static gint ett_nbap_Common_E_DCH_AICH_Information = -1;
static gint ett_nbap_Common_E_DCH_FDPCH_InfoItem = -1;
static gint ett_nbap_Common_E_DCH_Resource_Combination_InfoList = -1;
static gint ett_nbap_Common_E_DCH_Resource_Combination_InfoList_Item = -1;
static gint ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoList = -1;
static gint ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoList_Item = -1;
static gint ett_nbap_Common_E_DCH_MACdFlow_Specific_InfoList = -1;
static gint ett_nbap_Common_E_DCH_MACdFlow_Specific_InfoList_Item = -1;
static gint ett_nbap_Common_E_DCH_LogicalChannel_InfoList = -1;
static gint ett_nbap_Common_E_DCH_LogicalChannel_InfoList_Item = -1;
static gint ett_nbap_Common_EDCH_System_Information_ResponseFDD = -1;
static gint ett_nbap_E_RNTI_List = -1;
static gint ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseList = -1;
static gint ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseList_Item = -1;
static gint ett_nbap_Common_MACFlows_to_DeleteFDD = -1;
static gint ett_nbap_Common_MACFlows_to_DeleteFDD_Item = -1;
static gint ett_nbap_CommonMACFlow_Specific_InfoList = -1;
static gint ett_nbap_CommonMACFlow_Specific_InfoItem = -1;
static gint ett_nbap_CommonMACFlow_Specific_InfoList_Response = -1;
static gint ett_nbap_CommonMACFlow_Specific_InfoItem_Response = -1;
static gint ett_nbap_Common_MACFlow_PriorityQueue_Information = -1;
static gint ett_nbap_Common_MACFlow_PriorityQueue_Item = -1;
static gint ett_nbap_CommonMeasurementAccuracy = -1;
static gint ett_nbap_CommonMeasurementValue = -1;
static gint ett_nbap_CommonMeasurementValueInformation = -1;
static gint ett_nbap_CommonMeasurementAvailable = -1;
static gint ett_nbap_Common_PhysicalChannel_Status_Information = -1;
static gint ett_nbap_Common_PhysicalChannel_Status_Information768 = -1;
static gint ett_nbap_CommonTransportChannel_InformationResponse = -1;
static gint ett_nbap_Common_TransportChannel_Status_Information = -1;
static gint ett_nbap_ContinuousPacketConnectivityDTX_DRX_Information = -1;
static gint ett_nbap_ContinuousPacketConnectivityDTX_DRX_Information_to_Modify = -1;
static gint ett_nbap_ContinuousPacketConnectivityHS_SCCH_less_Information = -1;
static gint ett_nbap_ContinuousPacketConnectivityHS_SCCH_less_InformationItem = -1;
static gint ett_nbap_ContinuousPacketConnectivityHS_SCCH_less_Information_Response = -1;
static gint ett_nbap_CPC_Information = -1;
static gint ett_nbap_CriticalityDiagnostics = -1;
static gint ett_nbap_CriticalityDiagnostics_IE_List = -1;
static gint ett_nbap_CriticalityDiagnostics_IE_List_item = -1;
static gint ett_nbap_Common_EDCH_System_InformationLCR = -1;
static gint ett_nbap_Common_E_PUCH_InformationLCR = -1;
static gint ett_nbap_PRXdes_base_perURAFCN = -1;
static gint ett_nbap_PRXdes_base_Item = -1;
static gint ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoListLCR = -1;
static gint ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoList_ItemLCR = -1;
static gint ett_nbap_Common_E_DCH_MACdFlow_Specific_InfoListLCR = -1;
static gint ett_nbap_Common_E_DCH_MACdFlow_Specific_InfoList_ItemLCR = -1;
static gint ett_nbap_Common_EDCH_System_Information_ResponseLCR = -1;
static gint ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR = -1;
static gint ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_Ext = -1;
static gint ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseList_ItemLCR = -1;
static gint ett_nbap_Common_E_AGCH_ListLCR = -1;
static gint ett_nbap_Common_E_AGCH_ItemLCR = -1;
static gint ett_nbap_Common_E_HICH_ListLCR = -1;
static gint ett_nbap_Common_E_HICH_ItemLCR = -1;
static gint ett_nbap_Common_E_RNTI_Info_LCR = -1;
static gint ett_nbap_Common_E_RNTI_Info_ItemLCR = -1;
static gint ett_nbap_Common_MACFlows_to_DeleteLCR = -1;
static gint ett_nbap_Common_MACFlows_to_DeleteLCR_Item = -1;
static gint ett_nbap_CommonMACFlow_Specific_InfoListLCR = -1;
static gint ett_nbap_CommonMACFlow_Specific_InfoItemLCR = -1;
static gint ett_nbap_Common_H_RNTI_InformationLCR = -1;
static gint ett_nbap_Common_H_RNTI_InfoItemLCR = -1;
static gint ett_nbap_Sync_InformationLCR = -1;
static gint ett_nbap_CommonMACFlow_Specific_InfoList_ResponseLCR = -1;
static gint ett_nbap_CommonMACFlow_Specific_InfoItem_ResponseLCR = -1;
static gint ett_nbap_CPC_InformationLCR = -1;
static gint ett_nbap_ContinuousPacketConnectivity_DRX_InformationLCR = -1;
static gint ett_nbap_HS_SCCH_DRX_Information_LCR = -1;
static gint ett_nbap_E_AGCH_DRX_Information_LCR = -1;
static gint ett_nbap_E_AGCH_DRX_Parameters = -1;
static gint ett_nbap_ContinuousPacketConnectivity_DRX_Information_to_Modify_LCR = -1;
static gint ett_nbap_DRX_Information_to_Modify_LCR = -1;
static gint ett_nbap_DRX_Information_to_Modify_Items_LCR = -1;
static gint ett_nbap_ContinuousPacketConnectivity_DRX_Information_ResponseLCR = -1;
static gint ett_nbap_HS_SCCH_DRX_Information_ResponseLCR = -1;
static gint ett_nbap_E_AGCH_DRX_Information_ResponseLCR = -1;
static gint ett_nbap_E_AGCH_DRX_Parameters_Response = -1;
static gint ett_nbap_DCH_FDD_Information = -1;
static gint ett_nbap_DCH_FDD_InformationItem = -1;
static gint ett_nbap_DCH_Specific_FDD_InformationList = -1;
static gint ett_nbap_DCH_Specific_FDD_Item = -1;
static gint ett_nbap_DCH_InformationResponse = -1;
static gint ett_nbap_DCH_InformationResponseItem = -1;
static gint ett_nbap_DCH_MeasurementOccasion_Information = -1;
static gint ett_nbap_DchMeasurementOccasionInformation_Item = -1;
static gint ett_nbap_Measurement_Occasion_Pattern_Sequence_parameters = -1;
static gint ett_nbap_DCH_TDD_Information = -1;
static gint ett_nbap_DCH_TDD_InformationItem = -1;
static gint ett_nbap_DCH_Specific_TDD_InformationList = -1;
static gint ett_nbap_DCH_Specific_TDD_Item = -1;
static gint ett_nbap_FDD_DCHs_to_Modify = -1;
static gint ett_nbap_FDD_DCHs_to_ModifyItem = -1;
static gint ett_nbap_DCH_ModifySpecificInformation_FDD = -1;
static gint ett_nbap_DCH_ModifySpecificItem_FDD = -1;
static gint ett_nbap_TDD_DCHs_to_Modify = -1;
static gint ett_nbap_DCH_ModifyItem_TDD = -1;
static gint ett_nbap_DCH_ModifySpecificInformation_TDD = -1;
static gint ett_nbap_DCH_ModifySpecificItem_TDD = -1;
static gint ett_nbap_DedicatedChannelsCapacityConsumptionLaw = -1;
static gint ett_nbap_DedicatedChannelsCapacityConsumptionLaw_item = -1;
static gint ett_nbap_DedicatedMeasurementValue = -1;
static gint ett_nbap_DedicatedMeasurementValueInformation = -1;
static gint ett_nbap_DedicatedMeasurementAvailable = -1;
static gint ett_nbap_DelayedActivation = -1;
static gint ett_nbap_DelayedActivationUpdate = -1;
static gint ett_nbap_Activate_Info = -1;
static gint ett_nbap_Deactivate_Info = -1;
static gint ett_nbap_Execution_Type = -1;
static gint ett_nbap_DGANSSCorrections = -1;
static gint ett_nbap_DGANSS_Corrections_Req = -1;
static gint ett_nbap_DGANSS_Information = -1;
static gint ett_nbap_DGANSS_InformationItem = -1;
static gint ett_nbap_DGANSS_SignalInformation = -1;
static gint ett_nbap_DGANSS_SignalInformationItem = -1;
static gint ett_nbap_DGANSSThreshold = -1;
static gint ett_nbap_DGNSS_ValidityPeriod = -1;
static gint ett_nbap_DGPSCorrections = -1;
static gint ett_nbap_DGPSThresholds = -1;
static gint ett_nbap_DL_Timeslot_Information = -1;
static gint ett_nbap_DL_Timeslot_InformationItem = -1;
static gint ett_nbap_DL_TimeslotLCR_Information = -1;
static gint ett_nbap_DL_TimeslotLCR_InformationItem = -1;
static gint ett_nbap_DL_Timeslot768_Information = -1;
static gint ett_nbap_DL_Timeslot768_InformationItem = -1;
static gint ett_nbap_DL_PowerBalancing_Information = -1;
static gint ett_nbap_DL_ReferencePowerInformationList = -1;
static gint ett_nbap_DL_ReferencePowerInformationItem = -1;
static gint ett_nbap_DL_TimeslotISCPInfo = -1;
static gint ett_nbap_DL_TimeslotISCPInfoItem = -1;
static gint ett_nbap_DL_TimeslotISCPInfoLCR = -1;
static gint ett_nbap_DL_TimeslotISCPInfoItemLCR = -1;
static gint ett_nbap_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_HS_PDSCH_Timeslot_InformationItem_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_MaxHSDSCH_HSSCCH_Power_per_CELLPORTION = -1;
static gint ett_nbap_MaxHSDSCH_HSSCCH_Power_per_CELLPORTION_Item = -1;
static gint ett_nbap_DL_HS_PDSCH_Codelist_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_DRX_Information = -1;
static gint ett_nbap_DRX_Information_to_Modify = -1;
static gint ett_nbap_DRX_Information_to_Modify_Items = -1;
static gint ett_nbap_DSCH_InformationResponse = -1;
static gint ett_nbap_DSCH_InformationResponseItem = -1;
static gint ett_nbap_DSCH_TDD_Information = -1;
static gint ett_nbap_DSCH_TDD_InformationItem = -1;
static gint ett_nbap_DTX_Cycle_2ms_Items = -1;
static gint ett_nbap_DTX_Cycle_2ms_to_Modify_Items = -1;
static gint ett_nbap_DTX_Cycle_10ms_Items = -1;
static gint ett_nbap_DTX_Cycle_10ms_to_Modify_Items = -1;
static gint ett_nbap_DTX_Information = -1;
static gint ett_nbap_DTX_Information_to_Modify = -1;
static gint ett_nbap_DTX_Information_to_Modify_Items = -1;
static gint ett_nbap_Dual_Band_Capability_Info = -1;
static gint ett_nbap_E_AGCH_FDD_Code_Information = -1;
static gint ett_nbap_E_AGCH_FDD_Code_List = -1;
static gint ett_nbap_E_DCHCapacityConsumptionLaw = -1;
static gint ett_nbap_E_DCH_TDD_CapacityConsumptionLaw = -1;
static gint ett_nbap_E_DCH_SF_allocation = -1;
static gint ett_nbap_E_DCH_SF_allocation_item = -1;
static gint ett_nbap_E_DCH_FDD_DL_Control_Channel_Information = -1;
static gint ett_nbap_E_DCH_FDD_Information = -1;
static gint ett_nbap_E_DCH_FDD_Information_Response = -1;
static gint ett_nbap_E_DCH_FDD_Information_to_Modify = -1;
static gint ett_nbap_E_DCH_FDD_Update_Information = -1;
static gint ett_nbap_E_DCH_MACdFlow_Specific_UpdateInformation = -1;
static gint ett_nbap_E_DCH_MACdFlow_Specific_UpdateInformation_Item = -1;
static gint ett_nbap_E_DCH_DL_Control_Channel_Change_Information = -1;
static gint ett_nbap_E_DCH_DL_Control_Channel_Change_Information_Item = -1;
static gint ett_nbap_E_DCH_DL_Control_Channel_Grant_Information = -1;
static gint ett_nbap_E_DCH_DL_Control_Channel_Grant_Information_Item = -1;
static gint ett_nbap_E_DCH_Grant_Type_Information = -1;
static gint ett_nbap_E_DCH_LogicalChannelInformation = -1;
static gint ett_nbap_E_DCH_LogicalChannelInformationItem = -1;
static gint ett_nbap_E_DCH_MACdPDU_SizeList = -1;
static gint ett_nbap_E_DCH_MACdPDU_SizeListItem = -1;
static gint ett_nbap_E_DCH_LogicalChannelToModify = -1;
static gint ett_nbap_E_DCH_LogicalChannelToModifyItem = -1;
static gint ett_nbap_E_DCH_MACdPDU_SizeToModifyList = -1;
static gint ett_nbap_E_DCH_LogicalChannelToDelete = -1;
static gint ett_nbap_E_DCH_LogicalChannelToDeleteItem = -1;
static gint ett_nbap_E_DCH_MACdFlows_Information = -1;
static gint ett_nbap_E_DCH_MACdFlow_Specific_InfoList = -1;
static gint ett_nbap_E_DCH_MACdFlow_Specific_InfoItem = -1;
static gint ett_nbap_E_DCH_MACdFlow_Specific_InformationResp = -1;
static gint ett_nbap_E_DCH_MACdFlow_Specific_InformationResp_Item = -1;
static gint ett_nbap_E_DCH_MACdFlow_Specific_InfoList_to_Modify = -1;
static gint ett_nbap_E_DCH_MACdFlow_Specific_InfoItem_to_Modify = -1;
static gint ett_nbap_E_DCH_MACdFlows_to_Delete = -1;
static gint ett_nbap_E_DCH_MACdFlow_to_Delete_Item = -1;
static gint ett_nbap_E_DCH_Non_Scheduled_Transmission_Grant_Items = -1;
static gint ett_nbap_E_DCH_Serving_Cell_Change_Info_Response = -1;
static gint ett_nbap_E_DCH_serving_cell_choice = -1;
static gint ett_nbap_E_DCH_serving_cell_change_successful = -1;
static gint ett_nbap_E_DCH_RL_InformationList_Rsp = -1;
static gint ett_nbap_E_DCH_RL_InformationList_Rsp_Item = -1;
static gint ett_nbap_E_DCH_serving_cell_change_unsuccessful = -1;
static gint ett_nbap_EDCH_RACH_Report_Value = -1;
static gint ett_nbap_EDCH_RACH_Report_Value_item = -1;
static gint ett_nbap_E_DCH_TTI_Length = -1;
static gint ett_nbap_E_DCH_TTI_Length_to_Modify = -1;
static gint ett_nbap_E_DCH_MACdFlows_to_DeleteLCR = -1;
static gint ett_nbap_E_DCH_MACdFlow_to_Delete_ItemLCR = -1;
static gint ett_nbap_Enhanced_UE_DRX_InformationLCR = -1;
static gint ett_nbap_E_RGCH_E_HICH_FDD_Code_Information = -1;
static gint ett_nbap_E_RGCH_E_HICH_FDD_Code_List = -1;
static gint ett_nbap_E_TFCI_Boost_Information = -1;
static gint ett_nbap_E_TFCS_Information = -1;
static gint ett_nbap_E_DCHProvidedBitRate = -1;
static gint ett_nbap_E_DCHProvidedBitRate_Item = -1;
static gint ett_nbap_E_DCHProvidedBitRateValueInformation_For_CellPortion = -1;
static gint ett_nbap_E_DCHProvidedBitRateValueInformation_For_CellPortion_Item = -1;
static gint ett_nbap_E_DCH_Information = -1;
static gint ett_nbap_E_PUCH_Information = -1;
static gint ett_nbap_E_TFCS_Information_TDD = -1;
static gint ett_nbap_E_DCH_QPSK_RefBetaInfo = -1;
static gint ett_nbap_E_DCH_sixteenQAM_RefBetaInfo = -1;
static gint ett_nbap_E_DCH_RefBeta_Item = -1;
static gint ett_nbap_E_DCH_MACdFlows_Information_TDD = -1;
static gint ett_nbap_E_DCH_MACdFlow_InfoTDDItem = -1;
static gint ett_nbap_E_DCH_Non_Scheduled_Grant_Info = -1;
static gint ett_nbap_E_DCH_TDD_Information = -1;
static gint ett_nbap_E_DCH_Information_Response = -1;
static gint ett_nbap_Scheduled_E_HICH_Specific_Information_ResponseLCRTDD = -1;
static gint ett_nbap_Scheduled_E_HICH_Specific_InformationItem_ResponseLCRTDD = -1;
static gint ett_nbap_E_DCH_TDD_MACdFlow_Specific_InformationResp = -1;
static gint ett_nbap_E_DCH_TDD_MACdFlow_Specific_InformationResp_Item = -1;
static gint ett_nbap_E_AGCH_Specific_InformationRespListTDD = -1;
static gint ett_nbap_E_AGCH_Specific_InformationResp_ItemTDD = -1;
static gint ett_nbap_E_DCH_Information_Reconfig = -1;
static gint ett_nbap_E_DCH_TDD_Information_to_Modify = -1;
static gint ett_nbap_E_DCH_TDD_Information_to_Modify_List = -1;
static gint ett_nbap_E_DCH_MACdFlow_ModifyTDDItem = -1;
static gint ett_nbap_E_DCH_768_Information = -1;
static gint ett_nbap_E_DCH_Non_Scheduled_Grant_Info768 = -1;
static gint ett_nbap_E_DCH_TDD_Information768 = -1;
static gint ett_nbap_E_DCH_768_Information_Reconfig = -1;
static gint ett_nbap_E_DCH_LCR_Information = -1;
static gint ett_nbap_E_PUCH_LCR_Information = -1;
static gint ett_nbap_E_DCH_Non_Scheduled_Grant_LCR_Info = -1;
static gint ett_nbap_E_HICH_LCR_Information = -1;
static gint ett_nbap_E_DCH_LCRTDD_Information = -1;
static gint ett_nbap_E_DCH_LCR_Information_Reconfig = -1;
static gint ett_nbap_Enhanced_UE_DRX_InformationFDD = -1;
static gint ett_nbap_E_DCH_Semi_PersistentScheduling_Information_LCR = -1;
static gint ett_nbap_E_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR = -1;
static gint ett_nbap_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR = -1;
static gint ett_nbap_FDD_DL_CodeInformation = -1;
static gint ett_nbap_FDD_DL_CodeInformationItem = -1;
static gint ett_nbap_GANSS_AddClockModels = -1;
static gint ett_nbap_GANSS_AddOrbitModels = -1;
static gint ett_nbap_GANSS_Additional_Ionospheric_Model = -1;
static gint ett_nbap_GANSS_Additional_Navigation_Models = -1;
static gint ett_nbap_GANSS_Additional_Time_Models = -1;
static gint ett_nbap_GANSS_Additional_UTC_Models = -1;
static gint ett_nbap_GANSS_Almanac = -1;
static gint ett_nbap_GANSS_AlmanacModel = -1;
static gint ett_nbap_GANSS_ALM_ECEFsbasAlmanacSet = -1;
static gint ett_nbap_GANSS_ALM_GlonassAlmanacSet = -1;
static gint ett_nbap_GANSS_ALM_MidiAlmanacSet = -1;
static gint ett_nbap_GANSS_ALM_NAVKeplerianSet = -1;
static gint ett_nbap_GANSS_ALM_ReducedKeplerianSet = -1;
static gint ett_nbap_GANSS_Auxiliary_Information = -1;
static gint ett_nbap_GANSS_AuxInfoGANSS_ID1 = -1;
static gint ett_nbap_GANSS_AuxInfoGANSS_ID1_element = -1;
static gint ett_nbap_GANSS_AuxInfoGANSS_ID3 = -1;
static gint ett_nbap_GANSS_AuxInfoGANSS_ID3_element = -1;
static gint ett_nbap_GANSS_Clock_Model = -1;
static gint ett_nbap_GANSS_CNAVclockModel = -1;
static gint ett_nbap_GANSS_Common_Data = -1;
static gint ett_nbap_GANSS_CommonDataInfoReq = -1;
static gint ett_nbap_GANSS_Data_Bit_Assistance = -1;
static gint ett_nbap_GANSS_DataBitAssistanceList = -1;
static gint ett_nbap_GANSS_DataBitAssistanceItem = -1;
static gint ett_nbap_GANSS_DataBitAssistanceSgnList = -1;
static gint ett_nbap_GANSS_DataBitAssistanceSgnItem = -1;
static gint ett_nbap_GANSS_Data_Bit_Assistance_ReqItem = -1;
static gint ett_nbap_GANSS_Data_Bit_Assistance_ReqList = -1;
static gint ett_nbap_T_ganss_SatelliteInfo = -1;
static gint ett_nbap_GANSS_DeltaUT1 = -1;
static gint ett_nbap_GANSS_Earth_Orientation_Parameters = -1;
static gint ett_nbap_GANSS_GenericDataInfoReqList = -1;
static gint ett_nbap_GANSS_GenericDataInfoReqItem = -1;
static gint ett_nbap_GANSS_Generic_Data = -1;
static gint ett_nbap_GANSS_Generic_DataItem = -1;
static gint ett_nbap_GANSS_GLONASSclockModel = -1;
static gint ett_nbap_GANSS_Information = -1;
static gint ett_nbap_GANSS_Ionospheric_Model = -1;
static gint ett_nbap_GANSS_IonosphereRegionalStormFlags = -1;
static gint ett_nbap_GANSS_KeplerianParametersAlm = -1;
static gint ett_nbap_GANSS_KeplerianParametersOrb = -1;
static gint ett_nbap_GANSS_NAVclockModel = -1;
static gint ett_nbap_GANSS_Navigation_Model_And_Time_Recovery = -1;
static gint ett_nbap_GANSS_NavModel_CNAVKeplerianSet = -1;
static gint ett_nbap_GANSS_NavModel_GLONASSecef = -1;
static gint ett_nbap_GANSS_NavModel_NAVKeplerianSet = -1;
static gint ett_nbap_GANSS_NavModel_SBASecef = -1;
static gint ett_nbap_GANSS_Orbit_Model = -1;
static gint ett_nbap_GANSS_Real_Time_Integrity = -1;
static gint ett_nbap_GANSS_RealTimeInformationItem = -1;
static gint ett_nbap_GANSS_RX_Pos = -1;
static gint ett_nbap_GANSS_SatelliteClockModelItem = -1;
static gint ett_nbap_GANSS_SatelliteInformationKP = -1;
static gint ett_nbap_GANSS_SatelliteInformationKPItem = -1;
static gint ett_nbap_Ganss_Sat_Info_AddNavList = -1;
static gint ett_nbap_Ganss_Sat_Info_AddNavList_item = -1;
static gint ett_nbap_GANSS_SAT_Info_Almanac_GLOkpList = -1;
static gint ett_nbap_GANSS_SAT_Info_Almanac_GLOkp = -1;
static gint ett_nbap_GANSS_SAT_Info_Almanac_MIDIkpList = -1;
static gint ett_nbap_GANSS_SAT_Info_Almanac_MIDIkp = -1;
static gint ett_nbap_GANSS_SAT_Info_Almanac_NAVkpList = -1;
static gint ett_nbap_GANSS_SAT_Info_Almanac_NAVkp = -1;
static gint ett_nbap_GANSS_SAT_Info_Almanac_REDkpList = -1;
static gint ett_nbap_GANSS_SAT_Info_Almanac_REDkp = -1;
static gint ett_nbap_GANSS_SAT_Info_Almanac_SBASecefList = -1;
static gint ett_nbap_GANSS_SAT_Info_Almanac_SBASecef = -1;
static gint ett_nbap_GANSS_Sat_Info_Nav = -1;
static gint ett_nbap_GANSS_Sat_Info_Nav_item = -1;
static gint ett_nbap_GANSS_SBASclockModel = -1;
static gint ett_nbap_GANSS_Time_Model = -1;
static gint ett_nbap_GANSS_Transmission_Time = -1;
static gint ett_nbap_GANSS_UTC_Model = -1;
static gint ett_nbap_GANSS_UTCmodelSet1 = -1;
static gint ett_nbap_GANSS_UTCmodelSet2 = -1;
static gint ett_nbap_GANSS_UTCmodelSet3 = -1;
static gint ett_nbap_GPS_Almanac = -1;
static gint ett_nbap_GPS_Ionospheric_Model = -1;
static gint ett_nbap_GPS_Information = -1;
static gint ett_nbap_GPS_RealTime_Integrity = -1;
static gint ett_nbap_GPSBadSat_Info_RealTime_Integrity = -1;
static gint ett_nbap_GPS_NavigationModel_and_TimeRecovery = -1;
static gint ett_nbap_GPS_NavandRecovery_Item = -1;
static gint ett_nbap_GPS_RX_POS = -1;
static gint ett_nbap_GPS_UTC_Model = -1;
static gint ett_nbap_HARQ_MemoryPartitioning = -1;
static gint ett_nbap_HARQ_MemoryPartitioning_Implicit = -1;
static gint ett_nbap_HARQ_MemoryPartitioning_Explicit = -1;
static gint ett_nbap_HARQ_MemoryPartitioningList = -1;
static gint ett_nbap_HARQ_MemoryPartitioningInfoExtForMIMO = -1;
static gint ett_nbap_HARQ_MemoryPartitioningItem = -1;
static gint ett_nbap_HS_DSCHProvidedBitRate = -1;
static gint ett_nbap_HS_DSCHProvidedBitRate_Item = -1;
static gint ett_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortion = -1;
static gint ett_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortion_Item = -1;
static gint ett_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR = -1;
static gint ett_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR_Item = -1;
static gint ett_nbap_HS_DSCHRequiredPower = -1;
static gint ett_nbap_HS_DSCHRequiredPower_Item = -1;
static gint ett_nbap_HS_DSCHRequiredPowerPerUEInformation = -1;
static gint ett_nbap_HS_DSCHRequiredPowerPerUEInformation_Item = -1;
static gint ett_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortion = -1;
static gint ett_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortion_Item = -1;
static gint ett_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR = -1;
static gint ett_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR_Item = -1;
static gint ett_nbap_HSDPA_Associated_PICH_Information = -1;
static gint ett_nbap_HSDPA_PICH_Shared_with_PCH = -1;
static gint ett_nbap_HSDPA_PICH_notShared_with_PCH = -1;
static gint ett_nbap_HSDSCH_Common_System_InformationFDD = -1;
static gint ett_nbap_HSDSCH_Common_System_Information_ResponseFDD = -1;
static gint ett_nbap_HSDSCH_Common_Information = -1;
static gint ett_nbap_HSDSCH_FDD_Information = -1;
static gint ett_nbap_HSDSCH_TDD_Information = -1;
static gint ett_nbap_HSDSCH_Information_to_Modify = -1;
static gint ett_nbap_HSDSCH_MACdFlow_Specific_InfoList_to_Modify = -1;
static gint ett_nbap_HSDSCH_MACdFlow_Specific_InfoItem_to_Modify = -1;
static gint ett_nbap_HSDSCH_Information_to_Modify_Unsynchronised = -1;
static gint ett_nbap_HSDSCH_FDD_Information_Response = -1;
static gint ett_nbap_HS_DSCH_FDD_Secondary_Serving_Information = -1;
static gint ett_nbap_HS_DSCH_FDD_Secondary_Serving_Information_Response = -1;
static gint ett_nbap_HS_DSCH_Secondary_Serving_Information_To_Modify = -1;
static gint ett_nbap_HS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised = -1;
static gint ett_nbap_HS_DSCH_FDD_Secondary_Serving_Update_Information = -1;
static gint ett_nbap_HS_DSCH_Secondary_Serving_Cell_Change_Information_Response = -1;
static gint ett_nbap_HS_DSCH_Secondary_Serving_cell_change_choice = -1;
static gint ett_nbap_HS_Secondary_Serving_cell_change_successful = -1;
static gint ett_nbap_HS_Secondary_Serving_cell_change_unsuccessful = -1;
static gint ett_nbap_HSDSCH_Paging_System_InformationFDD = -1;
static gint ett_nbap_HSDSCH_Paging_System_Information_ResponseFDD = -1;
static gint ett_nbap_HSDSCH_Paging_System_Information_ResponseList = -1;
static gint ett_nbap_HSDSCH_TDD_Information_Response = -1;
static gint ett_nbap_HSDSCH_MACdFlow_Specific_InformationResp = -1;
static gint ett_nbap_HSDSCH_MACdFlow_Specific_InformationResp_Item = -1;
static gint ett_nbap_HSDSCH_MACdFlows_Information = -1;
static gint ett_nbap_HSDSCH_MACdFlow_Specific_InfoList = -1;
static gint ett_nbap_HSDSCH_MACdFlow_Specific_InfoItem = -1;
static gint ett_nbap_HSDSCH_MACdFlows_to_Delete = -1;
static gint ett_nbap_HSDSCH_MACdFlows_to_Delete_Item = -1;
static gint ett_nbap_HSDSCH_Initial_Capacity_Allocation = -1;
static gint ett_nbap_HSDSCH_Initial_Capacity_AllocationItem = -1;
static gint ett_nbap_HSDSCH_PreconfigurationInfo = -1;
static gint ett_nbap_Additional_EDCH_Preconfiguration_Information = -1;
static gint ett_nbap_Additional_EDCH_Preconfiguration_Information_ItemIEs = -1;
static gint ett_nbap_HSDSCH_PreconfigurationSetup = -1;
static gint ett_nbap_HS_SCCH_PreconfiguredCodes = -1;
static gint ett_nbap_HS_SCCH_PreconfiguredCodesItem = -1;
static gint ett_nbap_HSSCCH_Specific_InformationRespListFDD = -1;
static gint ett_nbap_HSSCCH_Codes = -1;
static gint ett_nbap_HSSCCH_Specific_InformationRespListTDD = -1;
static gint ett_nbap_HSSCCH_Specific_InformationRespItemTDD = -1;
static gint ett_nbap_HSSCCH_Specific_InformationRespListTDDLCR = -1;
static gint ett_nbap_HSSCCH_Specific_InformationRespItemTDDLCR = -1;
static gint ett_nbap_HSSCCH_Specific_InformationRespListTDD768 = -1;
static gint ett_nbap_HSSCCH_Specific_InformationRespItemTDD768 = -1;
static gint ett_nbap_HSSICH_Info = -1;
static gint ett_nbap_HSSICH_InfoLCR = -1;
static gint ett_nbap_HSSICH_Info768 = -1;
static gint ett_nbap_HS_SICH_Reception_Quality_Value = -1;
static gint ett_nbap_HS_PDSCH_FDD_Code_Information = -1;
static gint ett_nbap_HS_SCCH_FDD_Code_Information = -1;
static gint ett_nbap_HS_SCCH_FDD_Code_List = -1;
static gint ett_nbap_HS_DSCH_Serving_Cell_Change_Info = -1;
static gint ett_nbap_HS_DSCH_Serving_Cell_Change_Info_Response = -1;
static gint ett_nbap_HS_DSCH_serving_cell_choice = -1;
static gint ett_nbap_HS_serving_cell_change_successful = -1;
static gint ett_nbap_HS_serving_cell_change_unsuccessful = -1;
static gint ett_nbap_HSDSCH_FDD_Update_Information = -1;
static gint ett_nbap_HSDSCH_TDD_Update_Information = -1;
static gint ett_nbap_HSDPA_Associated_PICH_InformationLCR = -1;
static gint ett_nbap_HSDPA_PICH_notShared_with_PCHLCR = -1;
static gint ett_nbap_HSDSCH_Common_System_InformationLCR = -1;
static gint ett_nbap_HSDSCH_Common_System_Information_ResponseLCR = -1;
static gint ett_nbap_HSDSCH_Common_InformationLCR = -1;
static gint ett_nbap_HSDSCH_Paging_System_InformationLCR = -1;
static gint ett_nbap_HSDSCH_Paging_System_Information_ResponseLCR = -1;
static gint ett_nbap_HSDSCH_Paging_System_Information_ResponseListLCR = -1;
static gint ett_nbap_HSSCCH_Specific_InformationRespListLCR = -1;
static gint ett_nbap_HSSCCH_Specific_InformationRespItemLCR = -1;
static gint ett_nbap_HS_DSCH_Semi_PersistentScheduling_Information_LCR = -1;
static gint ett_nbap_Transport_Block_Size_List_LCR = -1;
static gint ett_nbap_Transport_Block_Size_Item_LCR = -1;
static gint ett_nbap_Repetition_Period_List_LCR = -1;
static gint ett_nbap_Repetition_Period_Item_LCR = -1;
static gint ett_nbap_HS_DSCH_SPS_Operation_Indicator = -1;
static gint ett_nbap_HS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR = -1;
static gint ett_nbap_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR = -1;
static gint ett_nbap_HS_SICH_InformationList_for_HS_DSCH_SPS = -1;
static gint ett_nbap_HS_SICH_InformationItem_for_HS_DSCH_SPS = -1;
static gint ett_nbap_HS_SICH_Type = -1;
static gint ett_nbap_HS_SCCH_Associated_HS_SICH = -1;
static gint ett_nbap_Non_HS_SCCH_Associated_HS_SICH = -1;
static gint ett_nbap_Initial_HS_PDSCH_SPS_Resource = -1;
static gint ett_nbap_Add_To_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationList = -1;
static gint ett_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext = -1;
static gint ett_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationItem = -1;
static gint ett_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList = -1;
static gint ett_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext = -1;
static gint ett_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationItem = -1;
static gint ett_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext = -1;
static gint ett_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqstItem = -1;
static gint ett_nbap_MIMO_ReferenceSignal_InformationListLCR = -1;
static gint ett_nbap_HSSICH_ReferenceSignal_InformationLCR = -1;
static gint ett_nbap_HSSICH_ReferenceSignal_InformationModifyLCR = -1;
static gint ett_nbap_IMB_Parameters = -1;
static gint ett_nbap_InformationReportCharacteristics = -1;
static gint ett_nbap_InformationReportCharacteristicsType_ReportPeriodicity = -1;
static gint ett_nbap_InformationReportCharacteristicsType_OnModification = -1;
static gint ett_nbap_InformationThresholds = -1;
static gint ett_nbap_InformationType = -1;
static gint ett_nbap_IPDL_FDD_Parameters = -1;
static gint ett_nbap_IPDL_TDD_Parameters = -1;
static gint ett_nbap_IPDL_TDD_Parameters_LCR = -1;
static gint ett_nbap_IPMulticastIndication = -1;
static gint ett_nbap_BurstModeParams = -1;
static gint ett_nbap_IdleIntervalInformation = -1;
static gint ett_nbap_LCRTDD_Uplink_Physical_Channel_Capability = -1;
static gint ett_nbap_MACdPDU_Size_Indexlist = -1;
static gint ett_nbap_MACdPDU_Size_IndexItem = -1;
static gint ett_nbap_MACdPDU_Size_Indexlist_to_Modify = -1;
static gint ett_nbap_MACdPDU_Size_IndexItem_to_Modify = -1;
static gint ett_nbap_Max_TB_Sizes = -1;
static gint ett_nbap_MessageStructure = -1;
static gint ett_nbap_MessageStructure_item = -1;
static gint ett_nbap_MidambleShiftAndBurstType = -1;
static gint ett_nbap_Type1 = -1;
static gint ett_nbap_MidambleAllocationMode1 = -1;
static gint ett_nbap_Type2 = -1;
static gint ett_nbap_MidambleAllocationMode2 = -1;
static gint ett_nbap_Type3 = -1;
static gint ett_nbap_MidambleAllocationMode3 = -1;
static gint ett_nbap_MidambleShiftLCR = -1;
static gint ett_nbap_MidambleShiftAndBurstType768 = -1;
static gint ett_nbap_Type7681 = -1;
static gint ett_nbap_MidambleAllocationMode7681 = -1;
static gint ett_nbap_Type7682 = -1;
static gint ett_nbap_MidambleAllocationMode7682 = -1;
static gint ett_nbap_Type7683 = -1;
static gint ett_nbap_MidambleAllocationMode7683 = -1;
static gint ett_nbap_MIMO_PilotConfiguration = -1;
static gint ett_nbap_MIMO_PilotConfigurationExtension = -1;
static gint ett_nbap_ModifyPriorityQueue = -1;
static gint ett_nbap_ModulationPO_MBSFN = -1;
static gint ett_nbap_MultipleFreq_HARQ_MemoryPartitioning_InformationList = -1;
static gint ett_nbap_MultipleFreq_HARQ_MemoryPartitioning_InformationItem = -1;
static gint ett_nbap_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR = -1;
static gint ett_nbap_MultipleFreq_HSPDSCH_InformationItem_ResponseTDDLCR = -1;
static gint ett_nbap_Multi_Cell_Capability_Info = -1;
static gint ett_nbap_Multicell_EDCH_InformationItemIEs = -1;
static gint ett_nbap_Multicell_EDCH_RL_Specific_InformationItemIEs = -1;
static gint ett_nbap_NeighbouringCellMeasurementInformation = -1;
static gint ett_nbap_NeighbouringCellMeasurementInformation_item = -1;
static gint ett_nbap_NeighbouringFDDCellMeasurementInformation = -1;
static gint ett_nbap_NeighbouringTDDCellMeasurementInformation = -1;
static gint ett_nbap_NeighbouringTDDCellMeasurementInformationLCR = -1;
static gint ett_nbap_NeighbouringTDDCellMeasurementInformation768 = -1;
static gint ett_nbap_NI_Information = -1;
static gint ett_nbap_NormalAndDiversityPrimaryCPICHContainer = -1;
static gint ett_nbap_Paging_MACFlows_to_DeleteFDD = -1;
static gint ett_nbap_Paging_MACFlows_to_DeleteFDD_Item = -1;
static gint ett_nbap_Paging_MACFlow_Specific_Information = -1;
static gint ett_nbap_Paging_MAC_Flow_Specific_Information_Item = -1;
static gint ett_nbap_Paging_MACFlow_PriorityQueue_Information = -1;
static gint ett_nbap_Paging_MACFlow_PriorityQueue_Item = -1;
static gint ett_nbap_PLCCHinformation = -1;
static gint ett_nbap_Possible_Secondary_Serving_Cell_List = -1;
static gint ett_nbap_Possible_Secondary_Serving_Cell = -1;
static gint ett_nbap_PreambleSignatures = -1;
static gint ett_nbap_PrimaryAndSecondaryCPICHContainer = -1;
static gint ett_nbap_Priority_Queue_Information_for_Enhanced_FACH_PCH = -1;
static gint ett_nbap_PriorityQueue_InfoList = -1;
static gint ett_nbap_PriorityQueue_InfoItem = -1;
static gint ett_nbap_PriorityQueue_InfoList_to_Modify = -1;
static gint ett_nbap_PriorityQueue_InfoItem_to_Add = -1;
static gint ett_nbap_PriorityQueue_InfoItem_to_Modify = -1;
static gint ett_nbap_PriorityQueue_InfoList_to_Modify_Unsynchronised = -1;
static gint ett_nbap_PriorityQueue_InfoItem_to_Modify_Unsynchronised = -1;
static gint ett_nbap_UE_Selected_MBMS_Service_Information = -1;
static gint ett_nbap_Selected_MBMS_Service = -1;
static gint ett_nbap_Selected_MBMS_Service_List = -1;
static gint ett_nbap_Selected_MBMS_Service_Item = -1;
static gint ett_nbap_Selected_MBMS_Service_TimeSlot_Information_LCR = -1;
static gint ett_nbap_MBMS_Service_TDM_Information = -1;
static gint ett_nbap_Paging_MACFlow_Specific_InformationLCR = -1;
static gint ett_nbap_Paging_MAC_Flow_Specific_Information_ItemLCR = -1;
static gint ett_nbap_Paging_MACFlows_to_DeleteLCR = -1;
static gint ett_nbap_Paging_MACFlows_to_DeleteLCR_Item = -1;
static gint ett_nbap_RACH_SubChannelNumbers = -1;
static gint ett_nbap_RL_Specific_DCH_Info = -1;
static gint ett_nbap_RL_Specific_DCH_Info_Item = -1;
static gint ett_nbap_RL_Specific_E_DCH_Info = -1;
static gint ett_nbap_RL_Specific_E_DCH_Information = -1;
static gint ett_nbap_RL_Specific_E_DCH_Information_Item = -1;
static gint ett_nbap_Reference_E_TFCI_Information = -1;
static gint ett_nbap_Reference_E_TFCI_Information_Item = -1;
static gint ett_nbap_ReportCharacteristics = -1;
static gint ett_nbap_ReportCharacteristicsType_EventA = -1;
static gint ett_nbap_ReportCharacteristicsType_EventB = -1;
static gint ett_nbap_ReportCharacteristicsType_EventC = -1;
static gint ett_nbap_ReportCharacteristicsType_EventD = -1;
static gint ett_nbap_ReportCharacteristicsType_EventE = -1;
static gint ett_nbap_ReportCharacteristicsType_EventF = -1;
static gint ett_nbap_ReportCharacteristicsType_OnModification = -1;
static gint ett_nbap_ReportCharacteristicsType_MeasurementIncreaseDecreaseThreshold = -1;
static gint ett_nbap_EDCH_RACH_Report_IncrDecrThres = -1;
static gint ett_nbap_ReportCharacteristicsType_MeasurementThreshold = -1;
static gint ett_nbap_EDCH_RACH_Report_ThresholdInformation = -1;
static gint ett_nbap_ReportCharacteristicsType_ScaledMeasurementChangeTime = -1;
static gint ett_nbap_ReportCharacteristicsType_ScaledMeasurementHysteresisTime = -1;
static gint ett_nbap_ReportCharacteristicsType_ReportPeriodicity = -1;
static gint ett_nbap_Received_total_wide_band_power_For_CellPortion_Value = -1;
static gint ett_nbap_Received_total_wide_band_power_For_CellPortion_Value_Item = -1;
static gint ett_nbap_Received_total_wide_band_power_For_CellPortion_ValueLCR = -1;
static gint ett_nbap_Received_total_wide_band_power_For_CellPortion_ValueLCR_Item = -1;
static gint ett_nbap_Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value = -1;
static gint ett_nbap_Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value_Item = -1;
static gint ett_nbap_Received_Scheduled_EDCH_Power_Share_Value = -1;
static gint ett_nbap_RequestedDataValueInformation = -1;
static gint ett_nbap_InformationAvailable = -1;
static gint ett_nbap_RequestedDataValue = -1;
static gint ett_nbap_SAT_Info_Almanac = -1;
static gint ett_nbap_SAT_Info_Almanac_Item = -1;
static gint ett_nbap_SAT_Info_Almanac_ExtList = -1;
static gint ett_nbap_SAT_Info_Almanac_ExtItem = -1;
static gint ett_nbap_SAT_Info_DGPSCorrections = -1;
static gint ett_nbap_SAT_Info_DGPSCorrections_Item = -1;
static gint ett_nbap_SATInfo_RealTime_Integrity = -1;
static gint ett_nbap_SAT_Info_RealTime_Integrity_Item = -1;
static gint ett_nbap_SecondaryServingCells = -1;
static gint ett_nbap_SecondaryServingCellsItem = -1;
static gint ett_nbap_Secondary_CPICH_Information_Change = -1;
static gint ett_nbap_Serving_E_DCH_RL_ID = -1;
static gint ett_nbap_Serving_E_DCH_RL_in_this_NodeB = -1;
static gint ett_nbap_SetsOfHS_SCCH_Codes = -1;
static gint ett_nbap_SetsOfHS_SCCH_CodesItem = -1;
static gint ett_nbap_Setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency = -1;
static gint ett_nbap_SFNSFNMeasurementThresholdInformation = -1;
static gint ett_nbap_SFNSFNMeasurementValueInformation = -1;
static gint ett_nbap_T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation = -1;
static gint ett_nbap_T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item = -1;
static gint ett_nbap_T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation = -1;
static gint ett_nbap_T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item = -1;
static gint ett_nbap_SFNSFNTimeStampInformation = -1;
static gint ett_nbap_SFNSFNTimeStamp_TDD = -1;
static gint ett_nbap_SFNSFNValue = -1;
static gint ett_nbap_SynchronisationReportCharacteristics = -1;
static gint ett_nbap_SynchronisationReportCharactThreExc = -1;
static gint ett_nbap_SynchronisationReportCharactThreInfoItem = -1;
static gint ett_nbap_SEQUENCE_SIZE_1_maxNrOfReceptsPerSyncFrame_OF_SynchronisationReportCharactCellSyncBurstInfoItem = -1;
static gint ett_nbap_SynchronisationReportCharactCellSyncBurstInfoItem = -1;
static gint ett_nbap_SyncDLCodeIdThreInfoLCR = -1;
static gint ett_nbap_SyncDLCodeIdThreInfoList = -1;
static gint ett_nbap_SyncDLCodeInfoListLCR = -1;
static gint ett_nbap_SyncDLCodeInfoItemLCR = -1;
static gint ett_nbap_TDD_ChannelisationCodeLCR = -1;
static gint ett_nbap_TDD_DL_Code_Information = -1;
static gint ett_nbap_TDD_DL_Code_InformationItem = -1;
static gint ett_nbap_TDD_DL_Code_LCR_Information = -1;
static gint ett_nbap_TDD_DL_Code_LCR_InformationItem = -1;
static gint ett_nbap_TDD_DL_Code_768_Information = -1;
static gint ett_nbap_TDD_DL_Code_768_InformationItem = -1;
static gint ett_nbap_TDD_DL_DPCH_TimeSlotFormat_LCR = -1;
static gint ett_nbap_TDD_DPCHOffset = -1;
static gint ett_nbap_TransportFormatCombination_Beta = -1;
static gint ett_nbap_T_signalledGainFactors = -1;
static gint ett_nbap_T_gainFactor = -1;
static gint ett_nbap_T_fdd = -1;
static gint ett_nbap_TDD_UL_Code_Information = -1;
static gint ett_nbap_TDD_UL_Code_InformationItem = -1;
static gint ett_nbap_TDD_UL_Code_LCR_Information = -1;
static gint ett_nbap_TDD_UL_Code_LCR_InformationItem = -1;
static gint ett_nbap_TDD_UL_Code_768_Information = -1;
static gint ett_nbap_TDD_UL_Code_768_InformationItem = -1;
static gint ett_nbap_TDD_UL_DPCH_TimeSlotFormat_LCR = -1;
static gint ett_nbap_TFCI_SignallingMode = -1;
static gint ett_nbap_TimeSlot_InitiatedListLCR = -1;
static gint ett_nbap_TimeSlotMeasurementValueListLCR = -1;
static gint ett_nbap_TimeSlotMeasurementValueLCR = -1;
static gint ett_nbap_TimingAdjustmentValue = -1;
static gint ett_nbap_TimingAdjustmentValueLCR = -1;
static gint ett_nbap_TnlQos = -1;
static gint ett_nbap_Transmission_Gap_Pattern_Sequence_Information = -1;
static gint ett_nbap_Transmission_Gap_Pattern_Sequence_Information_item = -1;
static gint ett_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue = -1;
static gint ett_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue_Item = -1;
static gint ett_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue = -1;
static gint ett_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue_Item = -1;
static gint ett_nbap_Transmitted_Carrier_Power_For_CellPortion_Value = -1;
static gint ett_nbap_Transmitted_Carrier_Power_For_CellPortion_Value_Item = -1;
static gint ett_nbap_Transmitted_Carrier_Power_For_CellPortion_ValueLCR = -1;
static gint ett_nbap_Transmitted_Carrier_Power_For_CellPortion_ValueLCR_Item = -1;
static gint ett_nbap_TFCS = -1;
static gint ett_nbap_T_tFCSvalues = -1;
static gint ett_nbap_TFCS_TFCSList = -1;
static gint ett_nbap_TFCS_TFCSList_item = -1;
static gint ett_nbap_TFCS_CTFC = -1;
static gint ett_nbap_Transport_Block_Size_List = -1;
static gint ett_nbap_Transport_Block_Size_List_item = -1;
static gint ett_nbap_TransportFormatSet = -1;
static gint ett_nbap_TransportFormatSet_DynamicPartList = -1;
static gint ett_nbap_TransportFormatSet_DynamicPartList_item = -1;
static gint ett_nbap_TDD_TransportFormatSet_ModeDP = -1;
static gint ett_nbap_TransmissionTimeIntervalInformation = -1;
static gint ett_nbap_TransmissionTimeIntervalInformation_item = -1;
static gint ett_nbap_TransportFormatSet_Semi_staticPart = -1;
static gint ett_nbap_TransportFormatSet_ModeDP = -1;
static gint ett_nbap_TransportFormatSet_ModeSSP = -1;
static gint ett_nbap_TUTRANGANSS = -1;
static gint ett_nbap_TUTRANGANSSMeasurementThresholdInformation = -1;
static gint ett_nbap_TUTRANGANSSMeasurementValueInformation = -1;
static gint ett_nbap_TUTRANGPS = -1;
static gint ett_nbap_TUTRANGPSMeasurementThresholdInformation = -1;
static gint ett_nbap_TUTRANGPSMeasurementValueInformation = -1;
static gint ett_nbap_UC_Id = -1;
static gint ett_nbap_UE_AggregateMaximumBitRate = -1;
static gint ett_nbap_UE_Capability_Information = -1;
static gint ett_nbap_UL_Timeslot_Information = -1;
static gint ett_nbap_UL_Timeslot_InformationItem = -1;
static gint ett_nbap_UL_TimeslotLCR_Information = -1;
static gint ett_nbap_UL_TimeslotLCR_InformationItem = -1;
static gint ett_nbap_UL_Timeslot768_Information = -1;
static gint ett_nbap_UL_Timeslot768_InformationItem = -1;
static gint ett_nbap_UL_ScramblingCode = -1;
static gint ett_nbap_UL_Synchronisation_Parameters_LCR = -1;
static gint ett_nbap_UL_TimeSlot_ISCP_Info = -1;
static gint ett_nbap_UL_TimeSlot_ISCP_InfoItem = -1;
static gint ett_nbap_UL_TimeSlot_ISCP_LCR_Info = -1;
static gint ett_nbap_UL_TimeSlot_ISCP_LCR_InfoItem = -1;
static gint ett_nbap_UpPTSInterference_For_CellPortion_Value = -1;
static gint ett_nbap_UpPTSInterference_For_CellPortion_Value_Item = -1;
static gint ett_nbap_USCH_Information = -1;
static gint ett_nbap_USCH_InformationItem = -1;
static gint ett_nbap_USCH_InformationResponse = -1;
static gint ett_nbap_USCH_InformationResponseItem = -1;
static gint ett_nbap_UL_TimeslotISCP_For_CellPortion_Value = -1;
static gint ett_nbap_UL_TimeslotISCP_For_CellPortion_Value_Item = -1;
static gint ett_nbap_CommonTransportChannelSetupRequestFDD = -1;
static gint ett_nbap_CommonPhysicalChannelType_CTCH_SetupRqstFDD = -1;
static gint ett_nbap_Secondary_CCPCH_CTCH_SetupRqstFDD = -1;
static gint ett_nbap_PowerOffsetInformation_CTCH_SetupRqstFDD = -1;
static gint ett_nbap_FACH_ParametersListIE_CTCH_SetupRqstFDD = -1;
static gint ett_nbap_FACH_ParametersItem_CTCH_SetupRqstFDD = -1;
static gint ett_nbap_PCH_ParametersItem_CTCH_SetupRqstFDD = -1;
static gint ett_nbap_PICH_Parameters_CTCH_SetupRqstFDD = -1;
static gint ett_nbap_MICH_Parameters_CTCH_SetupRqstFDD = -1;
static gint ett_nbap_PRACH_CTCH_SetupRqstFDD = -1;
static gint ett_nbap_AllowedSlotFormatInformationList_CTCH_SetupRqstFDD = -1;
static gint ett_nbap_AllowedSlotFormatInformationItem_CTCH_SetupRqstFDD = -1;
static gint ett_nbap_RACH_ParametersItem_CTCH_SetupRqstFDD = -1;
static gint ett_nbap_AICH_Parameters_CTCH_SetupRqstFDD = -1;
static gint ett_nbap_CommonTransportChannelSetupRequestTDD = -1;
static gint ett_nbap_CommonPhysicalChannelType_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_Secondary_CCPCH_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_Secondary_CCPCH_parameterListIE_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_Secondary_CCPCH_parameterItem_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_Secondary_CCPCH_LCR_parameterList_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_Secondary_CCPCH_LCR_parameterItem_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_Secondary_CCPCH_768_parameterList_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_Secondary_CCPCH_768_parameterItem_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_FACH_ParametersListIE_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_FACH_ParametersItem_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_PCH_ParametersItem_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_PICH_ParametersItem_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_PICH_LCR_Parameters_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_PICH_768_ParametersItem_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_MICH_Parameters_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_MICH_TDDOption_Specific_Parameters_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_MICH_HCR_Parameters_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_MICH_LCR_Parameters_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_MICH_768_Parameters_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_TimeSlotConfigurationList_LCR_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_TimeSlotConfigurationItem_LCR_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_Secondary_CCPCH_parameterExtendedList_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_PRACH_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_PRACH_ParametersItem_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_RACH_ParameterItem_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_PRACH_LCR_ParametersList_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_PRACH_LCR_ParametersItem_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_PRACH_768_ParametersItem_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_FPACH_LCR_Parameters_CTCH_SetupRqstTDD = -1;
static gint ett_nbap_PLCCH_parameters = -1;
static gint ett_nbap_E_RUCCH_parameters = -1;
static gint ett_nbap_E_RUCCH_768_parameters = -1;
static gint ett_nbap_CommonTransportChannelSetupResponse = -1;
static gint ett_nbap_FACH_CommonTransportChannel_InformationResponse = -1;
static gint ett_nbap_CommonTransportChannelSetupFailure = -1;
static gint ett_nbap_CommonTransportChannelReconfigurationRequestFDD = -1;
static gint ett_nbap_CommonPhysicalChannelType_CTCH_ReconfRqstFDD = -1;
static gint ett_nbap_Secondary_CCPCHList_CTCH_ReconfRqstFDD = -1;
static gint ett_nbap_FACH_ParametersListIE_CTCH_ReconfRqstFDD = -1;
static gint ett_nbap_FACH_ParametersItem_CTCH_ReconfRqstFDD = -1;
static gint ett_nbap_PCH_ParametersItem_CTCH_ReconfRqstFDD = -1;
static gint ett_nbap_PICH_ParametersItem_CTCH_ReconfRqstFDD = -1;
static gint ett_nbap_MICH_Parameters_CTCH_ReconfRqstFDD = -1;
static gint ett_nbap_PRACHList_CTCH_ReconfRqstFDD = -1;
static gint ett_nbap_PRACH_ParametersListIE_CTCH_ReconfRqstFDD = -1;
static gint ett_nbap_PRACH_ParametersItem_CTCH_ReconfRqstFDD = -1;
static gint ett_nbap_AllowedSlotFormatInformationList_CTCH_ReconfRqstFDD = -1;
static gint ett_nbap_AllowedSlotFormatInformationItem_CTCH_ReconfRqstFDD = -1;
static gint ett_nbap_AICH_ParametersListIE_CTCH_ReconfRqstFDD = -1;
static gint ett_nbap_AICH_ParametersItem_CTCH_ReconfRqstFDD = -1;
static gint ett_nbap_CommonTransportChannelReconfigurationRequestTDD = -1;
static gint ett_nbap_Secondary_CCPCH_Parameters_CTCH_ReconfRqstTDD = -1;
static gint ett_nbap_Secondary_CCPCHListIE_CTCH_ReconfRqstTDD = -1;
static gint ett_nbap_Secondary_CCPCHItem_CTCH_ReconfRqstTDD = -1;
static gint ett_nbap_Secondary_CCPCH_parameterExtendedList_CTCH_ReconfRqstTDD = -1;
static gint ett_nbap_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_ReconfRqstTDD = -1;
static gint ett_nbap_PICH_Parameters_CTCH_ReconfRqstTDD = -1;
static gint ett_nbap_FACH_ParametersList_CTCH_ReconfRqstTDD = -1;
static gint ett_nbap_FACH_ParametersItem_CTCH_ReconfRqstTDD = -1;
static gint ett_nbap_PCH_Parameters_CTCH_ReconfRqstTDD = -1;
static gint ett_nbap_FPACH_LCR_Parameters_CTCH_ReconfRqstTDD = -1;
static gint ett_nbap_MICH_Parameters_CTCH_ReconfRqstTDD = -1;
static gint ett_nbap_PLCCH_Parameters_CTCH_ReconfRqstTDD = -1;
static gint ett_nbap_Secondary_CCPCH_768_Parameters_CTCH_ReconfRqstTDD = -1;
static gint ett_nbap_Secondary_CCPCH_768_List_CTCH_ReconfRqstTDD = -1;
static gint ett_nbap_Secondary_CCPCH_768_Item_CTCH_ReconfRqstTDD = -1;
static gint ett_nbap_PICH_768_Parameters_CTCH_ReconfRqstTDD = -1;
static gint ett_nbap_MICH_768_Parameters_CTCH_ReconfRqstTDD = -1;
static gint ett_nbap_UPPCH_LCR_Parameters_CTCH_ReconfRqstTDD = -1;
static gint ett_nbap_CommonTransportChannelReconfigurationResponse = -1;
static gint ett_nbap_CommonTransportChannelReconfigurationFailure = -1;
static gint ett_nbap_CommonTransportChannelDeletionRequest = -1;
static gint ett_nbap_CommonTransportChannelDeletionResponse = -1;
static gint ett_nbap_BlockResourceRequest = -1;
static gint ett_nbap_BlockResourceResponse = -1;
static gint ett_nbap_BlockResourceFailure = -1;
static gint ett_nbap_UnblockResourceIndication = -1;
static gint ett_nbap_AuditRequiredIndication = -1;
static gint ett_nbap_AuditRequest = -1;
static gint ett_nbap_AuditResponse = -1;
static gint ett_nbap_Cell_InformationList_AuditRsp = -1;
static gint ett_nbap_Cell_InformationItem_AuditRsp = -1;
static gint ett_nbap_S_CPICH_InformationList_AuditRsp = -1;
static gint ett_nbap_S_CCPCH_InformationList_AuditRsp = -1;
static gint ett_nbap_FACH_InformationList_AuditRsp = -1;
static gint ett_nbap_PRACH_InformationList_AuditRsp = -1;
static gint ett_nbap_RACH_InformationList_AuditRsp = -1;
static gint ett_nbap_AICH_InformationList_AuditRsp = -1;
static gint ett_nbap_FPACH_LCR_InformationList_AuditRsp = -1;
static gint ett_nbap_HS_DSCH_Resources_Information_AuditRsp = -1;
static gint ett_nbap_S_CCPCH_InformationListExt_AuditRsp = -1;
static gint ett_nbap_S_CCPCH_LCR_InformationListExt_AuditRsp = -1;
static gint ett_nbap_E_DCH_Resources_Information_AuditRsp = -1;
static gint ett_nbap_PLCCH_InformationList_AuditRsp = -1;
static gint ett_nbap_S_CCPCH_768_InformationList_AuditRsp = -1;
static gint ett_nbap_PRACH_768_InformationList_AuditRsp = -1;
static gint ett_nbap_E_RUCCH_InformationList_AuditRsp = -1;
static gint ett_nbap_E_RUCCH_768_InformationList_AuditRsp = -1;
static gint ett_nbap_Cell_Frequency_List_Information_LCR_MulFreq_AuditRsp = -1;
static gint ett_nbap_Cell_Frequency_List_InformationItem_LCR_MulFreq_AuditRsp = -1;
static gint ett_nbap_UPPCH_LCR_InformationList_AuditRsp = -1;
static gint ett_nbap_UPPCH_LCR_InformationItem_AuditRsp = -1;
static gint ett_nbap_MultipleFreq_HS_DSCH_Resources_InformationList_AuditRsp = -1;
static gint ett_nbap_MultipleFreq_E_DCH_Resources_InformationList_AuditRsp = -1;
static gint ett_nbap_CCP_InformationList_AuditRsp = -1;
static gint ett_nbap_CCP_InformationItem_AuditRsp = -1;
static gint ett_nbap_Local_Cell_InformationList_AuditRsp = -1;
static gint ett_nbap_Local_Cell_InformationItem_AuditRsp = -1;
static gint ett_nbap_Local_Cell_Group_InformationList_AuditRsp = -1;
static gint ett_nbap_Local_Cell_Group_InformationItem_AuditRsp = -1;
static gint ett_nbap_Power_Local_Cell_Group_InformationList_AuditRsp = -1;
static gint ett_nbap_Power_Local_Cell_Group_InformationItem_AuditRsp = -1;
static gint ett_nbap_AuditFailure = -1;
static gint ett_nbap_CommonMeasurementInitiationRequest = -1;
static gint ett_nbap_CommonMeasurementObjectType_CM_Rqst = -1;
static gint ett_nbap_ERACH_CM_Rqst = -1;
static gint ett_nbap_Cell_CM_Rqst = -1;
static gint ett_nbap_RACH_CM_Rqst = -1;
static gint ett_nbap_PowerLocalCellGroup_CM_Rqst = -1;
static gint ett_nbap_CommonMeasurementInitiationResponse = -1;
static gint ett_nbap_CommonMeasurementObjectType_CM_Rsp = -1;
static gint ett_nbap_ERACH_CM_Rsp = -1;
static gint ett_nbap_Cell_CM_Rsp = -1;
static gint ett_nbap_RACH_CM_Rsp = -1;
static gint ett_nbap_PowerLocalCellGroup_CM_Rsp = -1;
static gint ett_nbap_CommonMeasurementInitiationFailure = -1;
static gint ett_nbap_CommonMeasurementReport = -1;
static gint ett_nbap_CommonMeasurementObjectType_CM_Rprt = -1;
static gint ett_nbap_ERACH_CM_Rprt = -1;
static gint ett_nbap_Cell_CM_Rprt = -1;
static gint ett_nbap_RACH_CM_Rprt = -1;
static gint ett_nbap_PowerLocalCellGroup_CM_Rprt = -1;
static gint ett_nbap_CommonMeasurementTerminationRequest = -1;
static gint ett_nbap_CommonMeasurementFailureIndication = -1;
static gint ett_nbap_CellSetupRequestFDD = -1;
static gint ett_nbap_Synchronisation_Configuration_Cell_SetupRqst = -1;
static gint ett_nbap_PrimarySCH_Information_Cell_SetupRqstFDD = -1;
static gint ett_nbap_SecondarySCH_Information_Cell_SetupRqstFDD = -1;
static gint ett_nbap_PrimaryCPICH_Information_Cell_SetupRqstFDD = -1;
static gint ett_nbap_SecondaryCPICH_InformationList_Cell_SetupRqstFDD = -1;
static gint ett_nbap_SecondaryCPICH_InformationItem_Cell_SetupRqstFDD = -1;
static gint ett_nbap_PrimaryCCPCH_Information_Cell_SetupRqstFDD = -1;
static gint ett_nbap_BCH_Information_Cell_SetupRqstFDD = -1;
static gint ett_nbap_Limited_power_increase_information_Cell_SetupRqstFDD = -1;
static gint ett_nbap_IPDLParameter_Information_Cell_SetupRqstFDD = -1;
static gint ett_nbap_CellPortion_InformationList_Cell_SetupRqstFDD = -1;
static gint ett_nbap_CellPortion_InformationItem_Cell_SetupRqstFDD = -1;
static gint ett_nbap_CellSetupRequestTDD = -1;
static gint ett_nbap_SCH_Information_Cell_SetupRqstTDD = -1;
static gint ett_nbap_SyncCaseIndicatorItem_Cell_SetupRqstTDD_PSCH = -1;
static gint ett_nbap_Case1_Cell_SetupRqstTDD = -1;
static gint ett_nbap_Case2_Cell_SetupRqstTDD = -1;
static gint ett_nbap_PCCPCH_Information_Cell_SetupRqstTDD = -1;
static gint ett_nbap_TimeSlotConfigurationList_Cell_SetupRqstTDD = -1;
static gint ett_nbap_TimeSlotConfigurationItem_Cell_SetupRqstTDD = -1;
static gint ett_nbap_TimeSlotConfigurationList_LCR_Cell_SetupRqstTDD = -1;
static gint ett_nbap_TimeSlotConfigurationItem_LCR_Cell_SetupRqstTDD = -1;
static gint ett_nbap_PCCPCH_LCR_Information_Cell_SetupRqstTDD = -1;
static gint ett_nbap_DwPCH_LCR_Information_Cell_SetupRqstTDD = -1;
static gint ett_nbap_IPDLParameter_Information_Cell_SetupRqstTDD = -1;
static gint ett_nbap_IPDLParameter_Information_LCR_Cell_SetupRqstTDD = -1;
static gint ett_nbap_PCCPCH_768_Information_Cell_SetupRqstTDD = -1;
static gint ett_nbap_SCH_768_Information_Cell_SetupRqstTDD = -1;
static gint ett_nbap_Cell_Frequency_List_LCR_MulFreq_Cell_SetupRqstTDD = -1;
static gint ett_nbap_Cell_Frequency_Item_LCR_MulFreq_Cell_SetupRqstTDD = -1;
static gint ett_nbap_CellSetupResponse = -1;
static gint ett_nbap_CellSetupFailure = -1;
static gint ett_nbap_CellReconfigurationRequestFDD = -1;
static gint ett_nbap_Synchronisation_Configuration_Cell_ReconfRqst = -1;
static gint ett_nbap_PrimarySCH_Information_Cell_ReconfRqstFDD = -1;
static gint ett_nbap_SecondarySCH_Information_Cell_ReconfRqstFDD = -1;
static gint ett_nbap_PrimaryCPICH_Information_Cell_ReconfRqstFDD = -1;
static gint ett_nbap_SecondaryCPICH_InformationList_Cell_ReconfRqstFDD = -1;
static gint ett_nbap_SecondaryCPICH_InformationItem_Cell_ReconfRqstFDD = -1;
static gint ett_nbap_PrimaryCCPCH_Information_Cell_ReconfRqstFDD = -1;
static gint ett_nbap_BCH_information_Cell_ReconfRqstFDD = -1;
static gint ett_nbap_IPDLParameter_Information_Cell_ReconfRqstFDD = -1;
static gint ett_nbap_CellPortion_InformationList_Cell_ReconfRqstFDD = -1;
static gint ett_nbap_CellPortion_InformationItem_Cell_ReconfRqstFDD = -1;
static gint ett_nbap_CellReconfigurationRequestTDD = -1;
static gint ett_nbap_SCH_Information_Cell_ReconfRqstTDD = -1;
static gint ett_nbap_PCCPCH_Information_Cell_ReconfRqstTDD = -1;
static gint ett_nbap_TimeSlotConfigurationList_Cell_ReconfRqstTDD = -1;
static gint ett_nbap_TimeSlotConfigurationItem_Cell_ReconfRqstTDD = -1;
static gint ett_nbap_TimeSlotConfigurationList_LCR_Cell_ReconfRqstTDD = -1;
static gint ett_nbap_TimeSlotConfigurationItem_LCR_Cell_ReconfRqstTDD = -1;
static gint ett_nbap_DwPCH_LCR_Information_Cell_ReconfRqstTDD = -1;
static gint ett_nbap_IPDLParameter_Information_Cell_ReconfRqstTDD = -1;
static gint ett_nbap_IPDLParameter_Information_LCR_Cell_ReconfRqstTDD = -1;
static gint ett_nbap_SCH_768_Information_Cell_ReconfRqstTDD = -1;
static gint ett_nbap_PCCPCH_768_Information_Cell_ReconfRqstTDD = -1;
static gint ett_nbap_UARFCN_Adjustment = -1;
static gint ett_nbap_Cell_Frequency_Add_LCR_MulFreq_Cell_ReconfRqstTDD = -1;
static gint ett_nbap_Cell_Frequency_ModifyList_LCR_MulFreq_Cell_ReconfRqstTDD = -1;
static gint ett_nbap_Cell_Frequency_ModifyItem_LCR_MulFreq_Cell_ReconfRqstTDD = -1;
static gint ett_nbap_Cell_Frequency_Delete_LCR_MulFreq_Cell_ReconfRqstTDD = -1;
static gint ett_nbap_CellReconfigurationResponse = -1;
static gint ett_nbap_CellReconfigurationFailure = -1;
static gint ett_nbap_CellDeletionRequest = -1;
static gint ett_nbap_CellDeletionResponse = -1;
static gint ett_nbap_ResourceStatusIndication = -1;
static gint ett_nbap_IndicationType_ResourceStatusInd = -1;
static gint ett_nbap_No_Failure_ResourceStatusInd = -1;
static gint ett_nbap_Local_Cell_InformationList_ResourceStatusInd = -1;
static gint ett_nbap_Local_Cell_InformationItem_ResourceStatusInd = -1;
static gint ett_nbap_Local_Cell_Group_InformationList_ResourceStatusInd = -1;
static gint ett_nbap_Local_Cell_Group_InformationItem_ResourceStatusInd = -1;
static gint ett_nbap_Power_Local_Cell_Group_InformationList_ResourceStatusInd = -1;
static gint ett_nbap_Power_Local_Cell_Group_InformationItem_ResourceStatusInd = -1;
static gint ett_nbap_ServiceImpacting_ResourceStatusInd = -1;
static gint ett_nbap_Local_Cell_InformationList2_ResourceStatusInd = -1;
static gint ett_nbap_Local_Cell_InformationItem2_ResourceStatusInd = -1;
static gint ett_nbap_Local_Cell_Group_InformationList2_ResourceStatusInd = -1;
static gint ett_nbap_Local_Cell_Group_InformationItem2_ResourceStatusInd = -1;
static gint ett_nbap_CCP_InformationList_ResourceStatusInd = -1;
static gint ett_nbap_CCP_InformationItem_ResourceStatusInd = -1;
static gint ett_nbap_Cell_InformationList_ResourceStatusInd = -1;
static gint ett_nbap_Cell_InformationItem_ResourceStatusInd = -1;
static gint ett_nbap_S_CPICH_InformationList_ResourceStatusInd = -1;
static gint ett_nbap_S_CCPCH_InformationList_ResourceStatusInd = -1;
static gint ett_nbap_FACH_InformationList_ResourceStatusInd = -1;
static gint ett_nbap_PRACH_InformationList_ResourceStatusInd = -1;
static gint ett_nbap_RACH_InformationList_ResourceStatusInd = -1;
static gint ett_nbap_AICH_InformationList_ResourceStatusInd = -1;
static gint ett_nbap_FPACH_LCR_InformationList_ResourceStatusInd = -1;
static gint ett_nbap_HS_DSCH_Resources_Information_ResourceStatusInd = -1;
static gint ett_nbap_S_CCPCH_InformationListExt_ResourceStatusInd = -1;
static gint ett_nbap_S_CCPCH_LCR_InformationListExt_ResourceStatusInd = -1;
static gint ett_nbap_E_DCH_Resources_Information_ResourceStatusInd = -1;
static gint ett_nbap_PLCCH_InformationList_ResourceStatusInd = -1;
static gint ett_nbap_S_CCPCH_768_InformationList_ResourceStatusInd = -1;
static gint ett_nbap_PRACH_768_InformationList_ResourceStatusInd = -1;
static gint ett_nbap_E_RUCCH_InformationList_ResourceStatusInd = -1;
static gint ett_nbap_E_RUCCH_768_InformationList_ResourceStatusInd = -1;
static gint ett_nbap_Cell_Frequency_List_Information_LCR_MulFreq_ResourceStatusInd = -1;
static gint ett_nbap_Cell_Frequency_List_InformationItem_LCR_MulFreq_ResourceStatusInd = -1;
static gint ett_nbap_UPPCH_LCR_InformationList_ResourceStatusInd = -1;
static gint ett_nbap_UPPCH_LCR_InformationItem_ResourceStatusInd = -1;
static gint ett_nbap_MultipleFreq_HS_DSCH_Resources_InformationList_ResourceStatusInd = -1;
static gint ett_nbap_Power_Local_Cell_Group_InformationList2_ResourceStatusInd = -1;
static gint ett_nbap_Power_Local_Cell_Group_InformationItem2_ResourceStatusInd = -1;
static gint ett_nbap_MultipleFreq_E_DCH_Resources_InformationList_ResourceStatusInd = -1;
static gint ett_nbap_SystemInformationUpdateRequest = -1;
static gint ett_nbap_MIB_SB_SIB_InformationList_SystemInfoUpdateRqst = -1;
static gint ett_nbap_MIB_SB_SIB_InformationItem_SystemInfoUpdateRqst = -1;
static gint ett_nbap_DeletionIndicator_SystemInfoUpdate = -1;
static gint ett_nbap_No_Deletion_SystemInfoUpdate = -1;
static gint ett_nbap_SegmentInformationListIE_SystemInfoUpdate = -1;
static gint ett_nbap_SegmentInformationItem_SystemInfoUpdate = -1;
static gint ett_nbap_SystemInformationUpdateResponse = -1;
static gint ett_nbap_SystemInformationUpdateFailure = -1;
static gint ett_nbap_RadioLinkSetupRequestFDD = -1;
static gint ett_nbap_Additional_HS_Cell_Information_RL_Setup_List = -1;
static gint ett_nbap_Additional_HS_Cell_Information_RL_Setup_ItemIEs = -1;
static gint ett_nbap_UL_DPCH_Information_RL_SetupRqstFDD = -1;
static gint ett_nbap_DL_DPCH_Information_RL_SetupRqstFDD = -1;
static gint ett_nbap_PowerOffsetInformation_RL_SetupRqstFDD = -1;
static gint ett_nbap_RL_InformationList_RL_SetupRqstFDD = -1;
static gint ett_nbap_RL_InformationItem_RL_SetupRqstFDD = -1;
static gint ett_nbap_E_DPCH_Information_RL_SetupRqstFDD = -1;
static gint ett_nbap_F_DPCH_Information_RL_SetupRqstFDD = -1;
static gint ett_nbap_PowerOffsetInformation_F_DPCH_RL_SetupRqstFDD = -1;
static gint ett_nbap_RadioLinkSetupRequestTDD = -1;
static gint ett_nbap_UL_CCTrCH_InformationList_RL_SetupRqstTDD = -1;
static gint ett_nbap_UL_CCTrCH_InformationItem_RL_SetupRqstTDD = -1;
static gint ett_nbap_UL_DPCH_InformationItem_RL_SetupRqstTDD = -1;
static gint ett_nbap_UL_DPCH_LCR_Information_RL_SetupRqstTDD = -1;
static gint ett_nbap_UL_DPCH_768_Information_RL_SetupRqstTDD = -1;
static gint ett_nbap_DL_CCTrCH_InformationList_RL_SetupRqstTDD = -1;
static gint ett_nbap_DL_CCTrCH_InformationItem_RL_SetupRqstTDD = -1;
static gint ett_nbap_CCTrCH_TPCList_RL_SetupRqstTDD = -1;
static gint ett_nbap_CCTrCH_TPCItem_RL_SetupRqstTDD = -1;
static gint ett_nbap_DL_DPCH_InformationItem_RL_SetupRqstTDD = -1;
static gint ett_nbap_DL_DPCH_LCR_Information_RL_SetupRqstTDD = -1;
static gint ett_nbap_DL_DPCH_768_Information_RL_SetupRqstTDD = -1;
static gint ett_nbap_RL_Information_RL_SetupRqstTDD = -1;
static gint ett_nbap_RadioLinkSetupResponseFDD = -1;
static gint ett_nbap_Additional_HS_Cell_Information_Response_List = -1;
static gint ett_nbap_Additional_HS_Cell_Information_Response_ItemIEs = -1;
static gint ett_nbap_RL_InformationResponseList_RL_SetupRspFDD = -1;
static gint ett_nbap_RL_InformationResponseItem_RL_SetupRspFDD = -1;
static gint ett_nbap_DiversityIndication_RL_SetupRspFDD = -1;
static gint ett_nbap_Combining_RL_SetupRspFDD = -1;
static gint ett_nbap_NonCombiningOrFirstRL_RL_SetupRspFDD = -1;
static gint ett_nbap_RadioLinkSetupResponseTDD = -1;
static gint ett_nbap_RL_InformationResponse_RL_SetupRspTDD = -1;
static gint ett_nbap_RL_InformationResponse_LCR_RL_SetupRspTDD = -1;
static gint ett_nbap_RadioLinkSetupFailureFDD = -1;
static gint ett_nbap_CauseLevel_RL_SetupFailureFDD = -1;
static gint ett_nbap_GeneralCauseList_RL_SetupFailureFDD = -1;
static gint ett_nbap_RLSpecificCauseList_RL_SetupFailureFDD = -1;
static gint ett_nbap_Unsuccessful_RL_InformationRespList_RL_SetupFailureFDD = -1;
static gint ett_nbap_Unsuccessful_RL_InformationRespItem_RL_SetupFailureFDD = -1;
static gint ett_nbap_Successful_RL_InformationRespList_RL_SetupFailureFDD = -1;
static gint ett_nbap_Successful_RL_InformationRespItem_RL_SetupFailureFDD = -1;
static gint ett_nbap_DiversityIndication_RL_SetupFailureFDD = -1;
static gint ett_nbap_Combining_RL_SetupFailureFDD = -1;
static gint ett_nbap_NonCombiningOrFirstRL_RL_SetupFailureFDD = -1;
static gint ett_nbap_RadioLinkSetupFailureTDD = -1;
static gint ett_nbap_CauseLevel_RL_SetupFailureTDD = -1;
static gint ett_nbap_GeneralCauseList_RL_SetupFailureTDD = -1;
static gint ett_nbap_RLSpecificCauseList_RL_SetupFailureTDD = -1;
static gint ett_nbap_Unsuccessful_RL_InformationResp_RL_SetupFailureTDD = -1;
static gint ett_nbap_RadioLinkAdditionRequestFDD = -1;
static gint ett_nbap_Additional_HS_Cell_Information_RL_Addition_List = -1;
static gint ett_nbap_Additional_EDCH_Cell_Information_RL_Add_Req = -1;
static gint ett_nbap_Setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency = -1;
static gint ett_nbap_Additional_HS_Cell_Information_RL_Addition_ItemIEs = -1;
static gint ett_nbap_RL_InformationList_RL_AdditionRqstFDD = -1;
static gint ett_nbap_RL_InformationItem_RL_AdditionRqstFDD = -1;
static gint ett_nbap_E_DPCH_Information_RL_AdditionReqFDD = -1;
static gint ett_nbap_RadioLinkAdditionRequestTDD = -1;
static gint ett_nbap_UL_CCTrCH_InformationList_RL_AdditionRqstTDD = -1;
static gint ett_nbap_UL_CCTrCH_InformationItem_RL_AdditionRqstTDD = -1;
static gint ett_nbap_UL_DPCH_InformationItem_RL_AdditionRqstTDD = -1;
static gint ett_nbap_UL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD = -1;
static gint ett_nbap_UL_DPCH_InformationItem_768_RL_AdditionRqstTDD = -1;
static gint ett_nbap_DL_CCTrCH_InformationList_RL_AdditionRqstTDD = -1;
static gint ett_nbap_DL_CCTrCH_InformationItem_RL_AdditionRqstTDD = -1;
static gint ett_nbap_DL_DPCH_InformationItem_RL_AdditionRqstTDD = -1;
static gint ett_nbap_DL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD = -1;
static gint ett_nbap_DL_DPCH_InformationItem_768_RL_AdditionRqstTDD = -1;
static gint ett_nbap_RL_Information_RL_AdditionRqstTDD = -1;
static gint ett_nbap_RadioLinkAdditionResponseFDD = -1;
static gint ett_nbap_Additional_HS_Cell_Change_Information_Response_List = -1;
static gint ett_nbap_Additional_HS_Cell_Change_Information_Response_ItemIEs = -1;
static gint ett_nbap_RL_InformationResponseList_RL_AdditionRspFDD = -1;
static gint ett_nbap_RL_InformationResponseItem_RL_AdditionRspFDD = -1;
static gint ett_nbap_DiversityIndication_RL_AdditionRspFDD = -1;
static gint ett_nbap_Combining_RL_AdditionRspFDD = -1;
static gint ett_nbap_Non_Combining_RL_AdditionRspFDD = -1;
static gint ett_nbap_RadioLinkAdditionResponseTDD = -1;
static gint ett_nbap_RL_InformationResponse_RL_AdditionRspTDD = -1;
static gint ett_nbap_DCH_Information_RL_AdditionRspTDD = -1;
static gint ett_nbap_DiversityIndication_RL_AdditionRspTDD = -1;
static gint ett_nbap_Combining_RL_AdditionRspTDD = -1;
static gint ett_nbap_Non_Combining_RL_AdditionRspTDD = -1;
static gint ett_nbap_RL_InformationResponse_LCR_RL_AdditionRspTDD = -1;
static gint ett_nbap_RadioLinkAdditionFailureFDD = -1;
static gint ett_nbap_CauseLevel_RL_AdditionFailureFDD = -1;
static gint ett_nbap_GeneralCauseList_RL_AdditionFailureFDD = -1;
static gint ett_nbap_RLSpecificCauseList_RL_AdditionFailureFDD = -1;
static gint ett_nbap_Unsuccessful_RL_InformationRespList_RL_AdditionFailureFDD = -1;
static gint ett_nbap_Unsuccessful_RL_InformationRespItem_RL_AdditionFailureFDD = -1;
static gint ett_nbap_Successful_RL_InformationRespList_RL_AdditionFailureFDD = -1;
static gint ett_nbap_Successful_RL_InformationRespItem_RL_AdditionFailureFDD = -1;
static gint ett_nbap_DiversityIndication_RL_AdditionFailureFDD = -1;
static gint ett_nbap_Combining_RL_AdditionFailureFDD = -1;
static gint ett_nbap_Non_Combining_RL_AdditionFailureFDD = -1;
static gint ett_nbap_RadioLinkAdditionFailureTDD = -1;
static gint ett_nbap_CauseLevel_RL_AdditionFailureTDD = -1;
static gint ett_nbap_GeneralCauseList_RL_AdditionFailureTDD = -1;
static gint ett_nbap_RLSpecificCauseList_RL_AdditionFailureTDD = -1;
static gint ett_nbap_Unsuccessful_RL_InformationResp_RL_AdditionFailureTDD = -1;
static gint ett_nbap_RadioLinkReconfigurationPrepareFDD = -1;
static gint ett_nbap_Additional_HS_Cell_Information_RL_Reconf_Prep = -1;
static gint ett_nbap_Additional_HS_Cell_Information_RL_Reconf_Prep_ItemIEs = -1;
static gint ett_nbap_Additional_EDCH_Cell_Information_RL_Reconf_Prep = -1;
static gint ett_nbap_UL_DPCH_Information_RL_ReconfPrepFDD = -1;
static gint ett_nbap_DL_DPCH_Information_RL_ReconfPrepFDD = -1;
static gint ett_nbap_DL_DPCH_Power_Information_RL_ReconfPrepFDD = -1;
static gint ett_nbap_PowerOffsetInformation_RL_ReconfPrepFDD = -1;
static gint ett_nbap_DCH_DeleteList_RL_ReconfPrepFDD = -1;
static gint ett_nbap_DCH_DeleteItem_RL_ReconfPrepFDD = -1;
static gint ett_nbap_RL_InformationList_RL_ReconfPrepFDD = -1;
static gint ett_nbap_RL_InformationItem_RL_ReconfPrepFDD = -1;
static gint ett_nbap_E_DPCH_Information_RL_ReconfPrepFDD = -1;
static gint ett_nbap_F_DPCH_Information_RL_ReconfPrepFDD = -1;
static gint ett_nbap_PowerOffsetInformation_F_DPCH_RL_ReconfPrepFDD = -1;
static gint ett_nbap_RadioLinkReconfigurationPrepareTDD = -1;
static gint ett_nbap_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_UL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_UL_DPCH_InformationAddItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_UL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_MultipleRL_UL_DPCH_InformationAddList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_MultipleRL_UL_DPCH_InformationAddListIE_RL_ReconfPrepTDD = -1;
static gint ett_nbap_UL_DPCH_768_InformationAddList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_UL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_UL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_UL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_UL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_UL_Timeslot_InformationModify_ModifyItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_UL_TimeslotLCR_InformationModify_ModifyList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_UL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDDLCR = -1;
static gint ett_nbap_UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDDLCR = -1;
static gint ett_nbap_UL_Timeslot768_InformationModify_ModifyList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_UL_Timeslot_768_InformationModify_ModifyItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD768 = -1;
static gint ett_nbap_UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD768 = -1;
static gint ett_nbap_UL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD = -1;
static gint ett_nbap_UL_DPCH_InformationModify_DeleteItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_UL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_MultipleRL_UL_DPCH_InformationModifyList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_MultipleRL_UL_DPCH_InformationModifyListIE_RL_ReconfPrepTDD = -1;
static gint ett_nbap_UL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_UL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_CCTrCH_TPCAddList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_CCTrCH_TPCAddItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_DPCH_InformationAddItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_MultipleRL_DL_DPCH_InformationAddList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_MultipleRL_DL_DPCH_InformationAddListIE_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_DPCH_768_InformationAddList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_CCTrCH_TPCModifyList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_CCTrCH_TPCModifyItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_Timeslot_InformationModify_ModifyItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_Code_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_Code_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_Timeslot_768_InformationModify_ModifyList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_Timeslot_768_InformationModify_ModifyItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_Code_768_InformationModify_ModifyList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_Code_768_InformationModify_ModifyItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_DPCH_InformationModify_DeleteItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_MultipleRL_DL_DPCH_InformationModifyList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_MultipleRL_DL_DPCH_InformationModifyListIE_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DCH_DeleteList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DCH_DeleteItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DSCH_Information_ModifyList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DSCH_Information_ModifyItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DSCH_Information_DeleteList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_DSCH_Information_DeleteItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_USCH_Information_ModifyList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_USCH_Information_ModifyItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_USCH_Information_DeleteList_RL_ReconfPrepTDD = -1;
static gint ett_nbap_USCH_Information_DeleteItem_RL_ReconfPrepTDD = -1;
static gint ett_nbap_MultipleRL_Information_RL_ReconfPrepTDD = -1;
static gint ett_nbap_RL_Information_RL_ReconfPrepTDD = -1;
static gint ett_nbap_RadioLinkReconfigurationReady = -1;
static gint ett_nbap_RL_InformationResponseList_RL_ReconfReady = -1;
static gint ett_nbap_RL_InformationResponseItem_RL_ReconfReady = -1;
static gint ett_nbap_RadioLinkReconfigurationFailure = -1;
static gint ett_nbap_CauseLevel_RL_ReconfFailure = -1;
static gint ett_nbap_GeneralCauseList_RL_ReconfFailure = -1;
static gint ett_nbap_RLSpecificCauseList_RL_ReconfFailure = -1;
static gint ett_nbap_RL_ReconfigurationFailureList_RL_ReconfFailure = -1;
static gint ett_nbap_RL_ReconfigurationFailureItem_RL_ReconfFailure = -1;
static gint ett_nbap_RadioLinkReconfigurationCommit = -1;
static gint ett_nbap_RadioLinkReconfigurationCancel = -1;
static gint ett_nbap_RadioLinkReconfigurationRequestFDD = -1;
static gint ett_nbap_Additional_HS_Cell_Information_RL_Reconf_Req = -1;
static gint ett_nbap_Additional_HS_Cell_Information_RL_Reconf_Req_ItemIEs = -1;
static gint ett_nbap_Additional_EDCH_Cell_Information_RL_Reconf_Req = -1;
static gint ett_nbap_UL_DPCH_Information_RL_ReconfRqstFDD = -1;
static gint ett_nbap_DL_DPCH_Information_RL_ReconfRqstFDD = -1;
static gint ett_nbap_DCH_DeleteList_RL_ReconfRqstFDD = -1;
static gint ett_nbap_DCH_DeleteItem_RL_ReconfRqstFDD = -1;
static gint ett_nbap_RL_InformationList_RL_ReconfRqstFDD = -1;
static gint ett_nbap_RL_InformationItem_RL_ReconfRqstFDD = -1;
static gint ett_nbap_E_DPCH_Information_RL_ReconfRqstFDD = -1;
static gint ett_nbap_RadioLinkReconfigurationRequestTDD = -1;
static gint ett_nbap_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD = -1;
static gint ett_nbap_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD = -1;
static gint ett_nbap_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD = -1;
static gint ett_nbap_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD = -1;
static gint ett_nbap_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD = -1;
static gint ett_nbap_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD = -1;
static gint ett_nbap_MultipleRL_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD = -1;
static gint ett_nbap_MultipleRL_DL_CCTrCH_InformationModifyListIE_RL_ReconfRqstTDD = -1;
static gint ett_nbap_DL_DPCH_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD = -1;
static gint ett_nbap_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD = -1;
static gint ett_nbap_DL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfRqstTDD = -1;
static gint ett_nbap_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD = -1;
static gint ett_nbap_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD = -1;
static gint ett_nbap_DCH_DeleteList_RL_ReconfRqstTDD = -1;
static gint ett_nbap_DCH_DeleteItem_RL_ReconfRqstTDD = -1;
static gint ett_nbap_Multiple_RL_Information_RL_ReconfRqstTDD = -1;
static gint ett_nbap_RL_Information_RL_ReconfRqstTDD = -1;
static gint ett_nbap_RadioLinkReconfigurationResponse = -1;
static gint ett_nbap_RL_InformationResponseList_RL_ReconfRsp = -1;
static gint ett_nbap_RL_InformationResponseItem_RL_ReconfRsp = -1;
static gint ett_nbap_RadioLinkDeletionRequest = -1;
static gint ett_nbap_RL_informationList_RL_DeletionRqst = -1;
static gint ett_nbap_RL_informationItem_RL_DeletionRqst = -1;
static gint ett_nbap_RadioLinkDeletionResponse = -1;
static gint ett_nbap_DL_PowerControlRequest = -1;
static gint ett_nbap_DL_ReferencePowerInformationList_DL_PC_Rqst = -1;
static gint ett_nbap_DL_ReferencePowerInformationItem_DL_PC_Rqst = -1;
static gint ett_nbap_DL_PowerTimeslotControlRequest = -1;
static gint ett_nbap_DedicatedMeasurementInitiationRequest = -1;
static gint ett_nbap_DedicatedMeasurementObjectType_DM_Rqst = -1;
static gint ett_nbap_RL_DM_Rqst = -1;
static gint ett_nbap_RL_InformationList_DM_Rqst = -1;
static gint ett_nbap_RL_InformationItem_DM_Rqst = -1;
static gint ett_nbap_PUSCH_Info_DM_Rqst = -1;
static gint ett_nbap_HSSICH_Info_DM_Rqst = -1;
static gint ett_nbap_HSSICH_InfoExt_DM_Rqst = -1;
static gint ett_nbap_RL_Set_DM_Rqst = -1;
static gint ett_nbap_RL_Set_InformationList_DM_Rqst = -1;
static gint ett_nbap_RL_Set_InformationItem_DM_Rqst = -1;
static gint ett_nbap_DedicatedMeasurementInitiationResponse = -1;
static gint ett_nbap_DedicatedMeasurementObjectType_DM_Rsp = -1;
static gint ett_nbap_RL_DM_Rsp = -1;
static gint ett_nbap_RL_InformationList_DM_Rsp = -1;
static gint ett_nbap_RL_InformationItem_DM_Rsp = -1;
static gint ett_nbap_PUSCH_Info_DM_Rsp = -1;
static gint ett_nbap_Multiple_PUSCH_InfoList_DM_Rsp = -1;
static gint ett_nbap_Multiple_PUSCH_InfoListIE_DM_Rsp = -1;
static gint ett_nbap_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp = -1;
static gint ett_nbap_Multiple_DedicatedMeasurementValueItem_TDD_DM_Rsp = -1;
static gint ett_nbap_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp = -1;
static gint ett_nbap_Multiple_DedicatedMeasurementValueItem_LCR_TDD_DM_Rsp = -1;
static gint ett_nbap_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp = -1;
static gint ett_nbap_Multiple_HSSICHMeasurementValueItem_TDD_DM_Rsp = -1;
static gint ett_nbap_Multiple_DedicatedMeasurementValueList_768_TDD_DM_Rsp = -1;
static gint ett_nbap_Multiple_DedicatedMeasurementValueItem_768_TDD_DM_Rsp = -1;
static gint ett_nbap_RL_Set_DM_Rsp = -1;
static gint ett_nbap_RL_Set_InformationList_DM_Rsp = -1;
static gint ett_nbap_RL_Set_InformationItem_DM_Rsp = -1;
static gint ett_nbap_DedicatedMeasurementInitiationFailure = -1;
static gint ett_nbap_DedicatedMeasurementReport = -1;
static gint ett_nbap_DedicatedMeasurementObjectType_DM_Rprt = -1;
static gint ett_nbap_RL_DM_Rprt = -1;
static gint ett_nbap_RL_InformationList_DM_Rprt = -1;
static gint ett_nbap_RL_InformationItem_DM_Rprt = -1;
static gint ett_nbap_PUSCH_Info_DM_Rprt = -1;
static gint ett_nbap_Multiple_PUSCH_InfoList_DM_Rprt = -1;
static gint ett_nbap_Multiple_PUSCH_InfoListIE_DM_Rprt = -1;
static gint ett_nbap_RL_Set_DM_Rprt = -1;
static gint ett_nbap_RL_Set_InformationList_DM_Rprt = -1;
static gint ett_nbap_RL_Set_InformationItem_DM_Rprt = -1;
static gint ett_nbap_DedicatedMeasurementTerminationRequest = -1;
static gint ett_nbap_DedicatedMeasurementFailureIndication = -1;
static gint ett_nbap_RadioLinkFailureIndication = -1;
static gint ett_nbap_Reporting_Object_RL_FailureInd = -1;
static gint ett_nbap_RL_RL_FailureInd = -1;
static gint ett_nbap_RL_InformationList_RL_FailureInd = -1;
static gint ett_nbap_RL_InformationItem_RL_FailureInd = -1;
static gint ett_nbap_RL_Set_RL_FailureInd = -1;
static gint ett_nbap_RL_Set_InformationList_RL_FailureInd = -1;
static gint ett_nbap_RL_Set_InformationItem_RL_FailureInd = -1;
static gint ett_nbap_CCTrCH_RL_FailureInd = -1;
static gint ett_nbap_CCTrCH_InformationList_RL_FailureInd = -1;
static gint ett_nbap_CCTrCH_InformationItem_RL_FailureInd = -1;
static gint ett_nbap_RadioLinkPreemptionRequiredIndication = -1;
static gint ett_nbap_RL_InformationList_RL_PreemptRequiredInd = -1;
static gint ett_nbap_RL_InformationItem_RL_PreemptRequiredInd = -1;
static gint ett_nbap_RadioLinkRestoreIndication = -1;
static gint ett_nbap_Reporting_Object_RL_RestoreInd = -1;
static gint ett_nbap_RL_RL_RestoreInd = -1;
static gint ett_nbap_RL_InformationList_RL_RestoreInd = -1;
static gint ett_nbap_RL_InformationItem_RL_RestoreInd = -1;
static gint ett_nbap_RL_Set_RL_RestoreInd = -1;
static gint ett_nbap_RL_Set_InformationList_RL_RestoreInd = -1;
static gint ett_nbap_RL_Set_InformationItem_RL_RestoreInd = -1;
static gint ett_nbap_CCTrCH_RL_RestoreInd = -1;
static gint ett_nbap_CCTrCH_InformationList_RL_RestoreInd = -1;
static gint ett_nbap_CCTrCH_InformationItem_RL_RestoreInd = -1;
static gint ett_nbap_CompressedModeCommand = -1;
static gint ett_nbap_ErrorIndication = -1;
static gint ett_nbap_PrivateMessage = -1;
static gint ett_nbap_PhysicalSharedChannelReconfigurationRequestFDD = -1;
static gint ett_nbap_HSDPA_And_EDCH_CellPortion_InformationList_PSCH_ReconfRqst = -1;
static gint ett_nbap_HSDPA_And_EDCH_CellPortion_InformationItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_PhysicalSharedChannelReconfigurationRequestTDD = -1;
static gint ett_nbap_PDSCHSets_AddList_PSCH_ReconfRqst = -1;
static gint ett_nbap_PDSCHSets_AddItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_PDSCH_Information_AddItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Timeslot_InformationAddList_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Timeslot_InformationAddItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Code_InformationAddList_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Code_InformationAddItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_PDSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Timeslot_InformationAddItem_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Code_InformationAddList_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Code_InformationAddItem_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_PDSCH_AddInformation_768_AddItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Timeslot_InformationAddList_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Timeslot_InformationAddItem_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Code_InformationAddList_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Code_InformationAddItem_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_PDSCHSets_ModifyList_PSCH_ReconfRqst = -1;
static gint ett_nbap_PDSCHSets_ModifyItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_PDSCH_Information_ModifyItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Timeslot_InformationModifyList_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Timeslot_InformationModifyItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Code_InformationModifyList_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Code_InformationModifyItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_PDSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Timeslot_LCR_InformationModifyItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Code_LCR_InformationModifyList_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Code_LCR_InformationModifyItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_PDSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Timeslot_768_InformationModifyItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Code_768_InformationModifyList_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_Code_768_InformationModifyItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_PDSCHSets_DeleteList_PSCH_ReconfRqst = -1;
static gint ett_nbap_PDSCHSets_DeleteItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_PUSCHSets_AddList_PSCH_ReconfRqst = -1;
static gint ett_nbap_PUSCHSets_AddItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_PUSCH_Information_AddItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Timeslot_InformationAddList_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Timeslot_InformationAddItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Code_InformationAddList_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Code_InformationAddItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_PUSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Timeslot_InformationAddItem_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Code_InformationAddList_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Code_InformationAddItem_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_PUSCH_AddInformation_768_AddItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Timeslot_InformationAddList_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Timeslot_InformationAddItem_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Code_InformationAddList_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Code_InformationAddItem_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_PUSCHSets_ModifyList_PSCH_ReconfRqst = -1;
static gint ett_nbap_PUSCHSets_ModifyItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_PUSCH_Information_ModifyItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Timeslot_InformationModifyList_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Timeslot_InformationModifyItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Code_InformationModifyList_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Code_InformationModifyItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_PUSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Timeslot_LCR_InformationModifyItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Code_LCR_InformationModifyList_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Code_LCR_InformationModifyItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_PUSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Timeslot_768_InformationModifyItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Code_768_InformationModifyList_PSCH_ReconfRqst = -1;
static gint ett_nbap_UL_Code_768_InformationModifyItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_PUSCHSets_DeleteList_PSCH_ReconfRqst = -1;
static gint ett_nbap_PUSCHSets_DeleteItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_HS_PDSCH_TDD_Information_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_HS_PDSCH_Timeslot_Information_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_HS_PDSCH_Timeslot_InformationItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_HS_PDSCH_Codelist_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_HS_PDSCH_Timeslot_Information_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_HS_PDSCH_Timeslot_InformationItem_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_DL_HS_PDSCH_Codelist_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_Add_To_HS_SCCH_Resource_Pool_PSCH_ReconfRqst = -1;
static gint ett_nbap_HS_SCCH_Information_PSCH_ReconfRqst = -1;
static gint ett_nbap_HS_SCCH_InformationItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_HS_SICH_Information_PSCH_ReconfRqst = -1;
static gint ett_nbap_HS_SCCH_Information_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_HS_SCCH_InformationItem_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_HS_SICH_Information_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_HS_SCCH_Information_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_HS_SCCH_InformationItem_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_HS_SICH_Information_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_HS_SCCH_InformationExt_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_Modify_HS_SCCH_Resource_Pool_PSCH_ReconfRqst = -1;
static gint ett_nbap_HS_SCCH_InformationModifyItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_HS_SICH_InformationModify_PSCH_ReconfRqst = -1;
static gint ett_nbap_HS_SCCH_InformationModify_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_HS_SCCH_InformationModifyItem_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_HS_SCCH_InformationModifyExt_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_HS_SICH_InformationModify_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_HS_SCCH_InformationModify_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_HS_SCCH_InformationModifyItem_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_HS_SICH_InformationModify_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_HS_SCCH_InformationModify_PSCH_ReconfRqst = -1;
static gint ett_nbap_Delete_From_HS_SCCH_Resource_Pool_PSCH_ReconfRqst = -1;
static gint ett_nbap_Delete_From_HS_SCCH_Resource_PoolItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_E_PUCH_Information_PSCH_ReconfRqst = -1;
static gint ett_nbap_E_PUCH_Timeslot_Info = -1;
static gint ett_nbap_Add_To_E_AGCH_Resource_Pool_PSCH_ReconfRqst = -1;
static gint ett_nbap_E_AGCH_Information_PSCH_ReconfRqst = -1;
static gint ett_nbap_E_AGCH_InformationItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_Modify_E_AGCH_Resource_Pool_PSCH_ReconfRqst = -1;
static gint ett_nbap_E_AGCH_InformationModify_PSCH_ReconfRqst = -1;
static gint ett_nbap_E_AGCH_InformationModifyItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_Delete_From_E_AGCH_Resource_Pool_PSCH_ReconfRqst = -1;
static gint ett_nbap_Delete_From_E_AGCH_Resource_PoolItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_E_HICH_Information_PSCH_ReconfRqst = -1;
static gint ett_nbap_E_PUCH_Information_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_Add_To_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_E_AGCH_Information_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_E_AGCH_InformationItem_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_Modify_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_E_AGCH_InformationModify_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_E_AGCH_InformationModifyItem_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_E_HICH_Information_768_PSCH_ReconfRqst = -1;
static gint ett_nbap_E_PUCH_Information_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_E_PUCH_Timeslot_InfoLCR = -1;
static gint ett_nbap_E_PUCH_Timeslot_Item_InfoLCR = -1;
static gint ett_nbap_E_PUCH_Codelist_LCR = -1;
static gint ett_nbap_Add_To_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_E_AGCH_Information_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_E_AGCH_InformationItem_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_Modify_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_E_AGCH_InformationModify_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_E_AGCH_InformationModifyItem_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_Add_To_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_E_HICH_Information_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_E_HICH_InformationItem_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_Modify_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_E_HICH_InformationModify_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_E_HICH_InformationModifyItem_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_Delete_From_E_HICH_Resource_Pool_PSCH_ReconfRqst = -1;
static gint ett_nbap_Delete_From_E_HICH_Resource_PoolItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_SYNC_UL_Partition_LCR = -1;
static gint ett_nbap_Delete_From_HS_SCCH_Resource_PoolExt_PSCH_ReconfRqst = -1;
static gint ett_nbap_MultipleFreq_E_PUCH_Timeslot_InformationList_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_MultipleFreq_E_PUCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst = -1;
static gint ett_nbap_Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst = -1;
static gint ett_nbap_Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst_Item = -1;
static gint ett_nbap_PhysicalSharedChannelReconfigurationResponse = -1;
static gint ett_nbap_E_HICH_TimeOffset_ExtensionLCR = -1;
static gint ett_nbap_MultipleFreq_E_HICH_TimeOffsetLCR = -1;
static gint ett_nbap_PhysicalSharedChannelReconfigurationFailure = -1;
static gint ett_nbap_CauseLevel_PSCH_ReconfFailure = -1;
static gint ett_nbap_GeneralCauseList_PSCH_ReconfFailure = -1;
static gint ett_nbap_SetSpecificCauseList_PSCH_ReconfFailureTDD = -1;
static gint ett_nbap_Unsuccessful_PDSCHSetList_PSCH_ReconfFailureTDD = -1;
static gint ett_nbap_Unsuccessful_PDSCHSetItem_PSCH_ReconfFailureTDD = -1;
static gint ett_nbap_Unsuccessful_PUSCHSetList_PSCH_ReconfFailureTDD = -1;
static gint ett_nbap_Unsuccessful_PUSCHSetItem_PSCH_ReconfFailureTDD = -1;
static gint ett_nbap_UARFCNSpecificCauseList_PSCH_ReconfFailureTDD = -1;
static gint ett_nbap_Unsuccessful_UARFCNItem_PSCH_ReconfFailureTDD = -1;
static gint ett_nbap_E_HICH_TimeOffset_ReconfFailureTDD = -1;
static gint ett_nbap_Common_System_Information_ResponseLCR = -1;
static gint ett_nbap_ResetRequest = -1;
static gint ett_nbap_ResetIndicator = -1;
static gint ett_nbap_CommunicationContextList_Reset = -1;
static gint ett_nbap_CommunicationContextInfoList_Reset = -1;
static gint ett_nbap_CommunicationContextInfoItem_Reset = -1;
static gint ett_nbap_CommunicationContextType_Reset = -1;
static gint ett_nbap_CommunicationControlPortList_Reset = -1;
static gint ett_nbap_CommunicationControlPortInfoList_Reset = -1;
static gint ett_nbap_CommunicationControlPortInfoItem_Reset = -1;
static gint ett_nbap_ResetResponse = -1;
static gint ett_nbap_InformationExchangeInitiationRequest = -1;
static gint ett_nbap_InformationExchangeObjectType_InfEx_Rqst = -1;
static gint ett_nbap_Cell_InfEx_Rqst = -1;
static gint ett_nbap_InformationExchangeInitiationResponse = -1;
static gint ett_nbap_InformationExchangeObjectType_InfEx_Rsp = -1;
static gint ett_nbap_Cell_InfEx_Rsp = -1;
static gint ett_nbap_InformationExchangeInitiationFailure = -1;
static gint ett_nbap_InformationReport = -1;
static gint ett_nbap_InformationExchangeObjectType_InfEx_Rprt = -1;
static gint ett_nbap_Cell_Inf_Rprt = -1;
static gint ett_nbap_InformationExchangeTerminationRequest = -1;
static gint ett_nbap_InformationExchangeFailureIndication = -1;
static gint ett_nbap_CellSynchronisationInitiationRequestTDD = -1;
static gint ett_nbap_TimeslotInfo_CellSyncInitiationRqstTDD = -1;
static gint ett_nbap_CellSyncBurstTransInit_CellSyncInitiationRqstTDD = -1;
static gint ett_nbap_CellSyncBurstMeasureInit_CellSyncInitiationRqstTDD = -1;
static gint ett_nbap_SYNCDlCodeId_TransInitLCR_CellSyncInitiationRqstTDD = -1;
static gint ett_nbap_SYNCDlCodeId_MeasureInitLCR_CellSyncInitiationRqstTDD = -1;
static gint ett_nbap_CellSynchronisationInitiationResponseTDD = -1;
static gint ett_nbap_CellSynchronisationInitiationFailureTDD = -1;
static gint ett_nbap_CellSynchronisationReconfigurationRequestTDD = -1;
static gint ett_nbap_CellSyncBurstTransReconfInfo_CellSyncReconfRqstTDD = -1;
static gint ett_nbap_CellSyncBurstTransInfoItem_CellSyncReconfRqstTDD = -1;
static gint ett_nbap_CellSyncBurstMeasInfo_CellSyncReconfRqstTDD = -1;
static gint ett_nbap_CellSyncBurstMeasInfoListIE_CellSyncReconfRqstTDD = -1;
static gint ett_nbap_CellSyncBurstMeasInfoItem_CellSyncReconfRqstTDD = -1;
static gint ett_nbap_CellSyncBurstInfoList_CellSyncReconfRqstTDD = -1;
static gint ett_nbap_CellSyncBurstInfoItem_CellSyncReconfRqstTDD = -1;
static gint ett_nbap_SYNCDlCodeIdTransReconfInfoLCR_CellSyncReconfRqstTDD = -1;
static gint ett_nbap_SYNCDlCodeIdTransReconfItemLCR_CellSyncReconfRqstTDD = -1;
static gint ett_nbap_SYNCDlCodeIdMeasInfoLCR_CellSyncReconfRqstTDD = -1;
static gint ett_nbap_SYNCDlCodeIdMeasInfoList_CellSyncReconfRqstTDD = -1;
static gint ett_nbap_SYNCDlCodeIdMeasInfoItem_CellSyncReconfRqstTDD = -1;
static gint ett_nbap_SYNCDlCodeIdInfoListLCR_CellSyncReconfRqstTDD = -1;
static gint ett_nbap_SYNCDlCodeIdInfoItemLCR_CellSyncReconfRqstTDD = -1;
static gint ett_nbap_CellSynchronisationReconfigurationResponseTDD = -1;
static gint ett_nbap_CellSynchronisationReconfigurationFailureTDD = -1;
static gint ett_nbap_CellSynchronisationAdjustmentRequestTDD = -1;
static gint ett_nbap_CellAdjustmentInfo_SyncAdjustmentRqstTDD = -1;
static gint ett_nbap_CellAdjustmentInfoItem_SyncAdjustmentRqstTDD = -1;
static gint ett_nbap_CellSynchronisationAdjustmentResponseTDD = -1;
static gint ett_nbap_CellSynchronisationAdjustmentFailureTDD = -1;
static gint ett_nbap_CauseLevel_SyncAdjustmntFailureTDD = -1;
static gint ett_nbap_GeneralCauseList_SyncAdjustmntFailureTDD = -1;
static gint ett_nbap_CellSpecificCauseList_SyncAdjustmntFailureTDD = -1;
static gint ett_nbap_Unsuccessful_cell_InformationRespList_SyncAdjustmntFailureTDD = -1;
static gint ett_nbap_Unsuccessful_cell_InformationRespItem_SyncAdjustmntFailureTDD = -1;
static gint ett_nbap_CellSynchronisationTerminationRequestTDD = -1;
static gint ett_nbap_CellSynchronisationFailureIndicationTDD = -1;
static gint ett_nbap_CellSynchronisationReportTDD = -1;
static gint ett_nbap_CellSyncInfo_CellSyncReprtTDD = -1;
static gint ett_nbap_CellSyncInfoItemIE_CellSyncReprtTDD = -1;
static gint ett_nbap_SyncReportType_CellSyncReprtTDD = -1;
static gint ett_nbap_IntStdPhCellSyncInfo_CellSyncReprtTDD = -1;
static gint ett_nbap_CellSyncBurstMeasInfoList_CellSyncReprtTDD = -1;
static gint ett_nbap_CellSyncBurstMeasInfoItem_CellSyncReprtTDD = -1;
static gint ett_nbap_SEQUENCE_SIZE_1_maxNrOfReceptsPerSyncFrame_OF_CellSyncBurstInfo_CellSyncReprtTDD = -1;
static gint ett_nbap_CellSyncBurstInfo_CellSyncReprtTDD = -1;
static gint ett_nbap_CellSyncBurstAvailable_CellSyncReprtTDD = -1;
static gint ett_nbap_SyncDLCodeIdsMeasInfoList_CellSyncReprtTDD = -1;
static gint ett_nbap_SyncDLCodeIdsMeasInfoItem_CellSyncReprtTDD = -1;
static gint ett_nbap_SyncDLCodeIdInfo_CellSyncReprtTDD = -1;
static gint ett_nbap_SyncDLCodeIdItem_CellSyncReprtTDD = -1;
static gint ett_nbap_SyncDLCodeIdAvailable_CellSyncReprtTDD = -1;
static gint ett_nbap_BearerRearrangementIndication = -1;
static gint ett_nbap_DCH_RearrangeList_Bearer_RearrangeInd = -1;
static gint ett_nbap_DCH_RearrangeItem_Bearer_RearrangeInd = -1;
static gint ett_nbap_DSCH_RearrangeList_Bearer_RearrangeInd = -1;
static gint ett_nbap_DSCH_RearrangeItem_Bearer_RearrangeInd = -1;
static gint ett_nbap_USCH_RearrangeList_Bearer_RearrangeInd = -1;
static gint ett_nbap_USCH_RearrangeItem_Bearer_RearrangeInd = -1;
static gint ett_nbap_HSDSCH_RearrangeList_Bearer_RearrangeInd = -1;
static gint ett_nbap_HSDSCH_RearrangeItem_Bearer_RearrangeInd = -1;
static gint ett_nbap_E_DCH_RearrangeList_Bearer_RearrangeInd = -1;
static gint ett_nbap_E_DCH_RearrangeItem_Bearer_RearrangeInd = -1;
static gint ett_nbap_Additional_EDCH_Cell_Information_Bearer_Rearrangement_List = -1;
static gint ett_nbap_Additional_EDCH_Cell_Information_Bearer_Rearrangement_ItemIEs = -1;
static gint ett_nbap_RadioLinkActivationCommandFDD = -1;
static gint ett_nbap_DelayedActivationInformationList_RL_ActivationCmdFDD = -1;
static gint ett_nbap_DelayedActivationInformation_RL_ActivationCmdFDD = -1;
static gint ett_nbap_RadioLinkActivationCommandTDD = -1;
static gint ett_nbap_DelayedActivationInformationList_RL_ActivationCmdTDD = -1;
static gint ett_nbap_DelayedActivationInformation_RL_ActivationCmdTDD = -1;
static gint ett_nbap_RadioLinkParameterUpdateIndicationFDD = -1;
static gint ett_nbap_Additional_HS_Cell_Information_RL_Param_Upd = -1;
static gint ett_nbap_Additional_HS_Cell_Information_RL_Param_Upd_ItemIEs = -1;
static gint ett_nbap_Additional_EDCH_Cell_Information_RL_Param_Upd = -1;
static gint ett_nbap_Additional_EDCH_Cell_Information_RL_Param_Upd_ItemIEs = -1;
static gint ett_nbap_RadioLinkParameterUpdateIndicationTDD = -1;
static gint ett_nbap_MBMSNotificationUpdateCommand = -1;
static gint ett_nbap_UEStatusUpdateCommand = -1;
static gint ett_nbap_SecondaryULFrequencyReport = -1;
static gint ett_nbap_SecondaryULFrequencyUpdateIndication = -1;
static gint ett_nbap_NBAP_PDU = -1;
static gint ett_nbap_InitiatingMessage = -1;
static gint ett_nbap_SuccessfulOutcome = -1;
static gint ett_nbap_UnsuccessfulOutcome = -1;
static gint ett_nbap_Outcome = -1;

static expert_field ei_nbap_no_find_port_info = EI_INIT;
static expert_field ei_nbap_no_set_comm_context_id = EI_INIT;
static expert_field ei_nbap_hsdsch_entity_not_specified = EI_INIT;

extern int proto_fp;

static dissector_handle_t nbap_handle;

/*
 * Structure to hold Setup Request/Response message conversation
 * we add all src add/port declared in SetupRequest msg
 * to match it with dst add/port declared in SetupResponse msg
 * so we gonna have conversation with exact match (src and dst addr and port)
 */
typedef struct nbap_setup_conv
{
  guint32 transaction_id;
  guint32 dd_mode;
  guint32 channel_id;
  guint32 request_frame_number;
  address addr;
  guint32 port;
  umts_fp_conversation_info_t *umts_fp_conversation_info;
  conversation_t *conv;
}nbap_setup_conv_t;

/*
 * Hash table to manage Setup Request/Response message conversation
 * we can look in table for proper conversation
 */
static wmem_map_t *nbap_setup_conv_table;

typedef struct
{
  gint num_dch_in_flow;
  gint next_dch;
  gint num_ul_chans;
  gint ul_chan_tf_size[MAX_FP_CHANS];
  gint ul_chan_num_tbs[MAX_FP_CHANS];
  gint num_dl_chans;
  gint dl_chan_tf_size[MAX_FP_CHANS];
  gint dl_chan_num_tbs[MAX_FP_CHANS];
}nbap_dch_channel_info_t;

/* Struct to collect E-DCH data in a packet
 * As the address data comes before the ddi entries
 * we save the address to be able to find the conversation and update the
 * conversation data.
 */
typedef struct
{
  address crnc_address;
  guint16 crnc_port;
  gint no_ddi_entries;
  guint8 edch_ddi[MAX_EDCH_DDIS];
  guint edch_macd_pdu_size[MAX_EDCH_DDIS];
  guint8 edch_type;  /* 1 means T2 */
  guint8 lchId[MAX_EDCH_DDIS]; /*Logical channel ids.*/
} nbap_edch_channel_info_t;


typedef struct
{
  guint32 crnc_address;
  guint16 crnc_port[maxNrOfEDCHMACdFlows];
} nbap_edch_port_info_t;

typedef struct
{
  address crnc_address;
  guint16 crnc_port;
  enum fp_rlc_mode rlc_mode;
  guint32 hsdsch_physical_layer_category;
  guint8 entity;  /* "ns" means type 1 and "ehs" means type 2, type 3 == ?*/
} nbap_hsdsch_channel_info_t;

typedef struct
{
 address crnc_address;
 guint16 crnc_port;
 enum fp_rlc_mode rlc_mode;
} nbap_common_channel_info_t;

/*Stuff for mapping NodeB-Comuncation Context ID to CRNC Communication Context ID*/
typedef struct com_ctxt_{
  /*guint   nodeb_context;*/
  guint crnc_context;
  guint frame_num;
}nbap_com_context_id_t;

enum TransportFormatSet_type_enum
{
  NBAP_DCH_UL,
  NBAP_DCH_DL,
  NBAP_CPCH,
  NBAP_FACH,
  NBAP_PCH
};

#define NBAP_MAX_IB_SEGMENT_LENGTH 222

typedef struct nbap_ib_segment_t {
  guint32 bit_length;
  guint8* data;
} nbap_ib_segment_t;

static nbap_ib_segment_t* nbap_parse_ib_sg_data_var1(packet_info *pinfo, tvbuff_t *tvb,gboolean is_short)
{
  guint8 bit_length;
  guint8* data;
  nbap_ib_segment_t* output;
  if ( tvb_captured_length(tvb) < 2 ) {
    return NULL;
  }
  if (is_short) {
    bit_length = tvb_get_guint8(tvb,0) + 1;
    data = (guint8*)tvb_memdup(pinfo->pool,tvb,1,(bit_length+7)/8);
  }
  else {
    bit_length = NBAP_MAX_IB_SEGMENT_LENGTH;
    data = (guint8*)tvb_memdup(pinfo->pool,tvb,0,(bit_length+7)/8);
  }
  output = wmem_new(pinfo->pool, nbap_ib_segment_t);
  output->bit_length = bit_length;
  output->data = data;
  return output;
}

/*****************************************************************************/
/* Packet private data                                                       */
/* For this dissector, all information passed between different ASN.1 nodes  */
/* should be done only through this API!                                     */
/*****************************************************************************/


typedef struct nbap_private_data_t
{
  guint32 transportLayerAddress_ipv4;
  guint16 binding_id_port;
  enum TransportFormatSet_type_enum transport_format_set_type;
  guint32 procedure_code;
  guint num_items;
  guint32 ul_scrambling_code;
  guint32 com_context_id;
  gint num_dch_in_flow;
  gint hrnti;
  guint32 protocol_ie_id;
  guint32 dd_mode;
  guint32 transaction_id;
  guint32 t_dch_id;
  guint32 dch_id;
  guint32 prev_dch_id;
  guint32 common_physical_channel_id;
  guint32 e_dch_macdflow_id;
  guint32 hsdsch_macdflow_id;
  gboolean max_mac_d_pdu_size_ext_ie_present;
  guint32 e_dch_ddi_value;
  guint32 logical_channel_id;
  guint32 common_macdflow_id;
  guint32 mac_d_pdu_size;
  guint32 common_transport_channel_id;
  gint paging_indications;
  guint32 ib_type;
  guint32 segment_type;
  gboolean crnc_context_present; /* Whether 'com_context_id' is set */
  guint8 dch_crc_present;
  /* Arrays */
  nbap_dch_channel_info_t nbap_dch_chnl_info[256];
  nbap_edch_channel_info_t nbap_edch_channel_info[maxNrOfEDCHMACdFlows];
  gint hsdsch_macdflow_ids[maxNrOfMACdFlows];
  nbap_hsdsch_channel_info_t nbap_hsdsch_channel_info[maxNrOfMACdFlows];
  nbap_common_channel_info_t nbap_common_channel_info[maxNrOfMACdFlows];	/*TODO: Fix this!*/
  wmem_list_t* ib_segments; /* Information block segments */
} nbap_private_data_t;


/* Helper function to get or create a private_data struct */
static nbap_private_data_t* nbap_get_private_data(packet_info *pinfo)
{
  guint8 i;
  /* NOTE: Unlike other ASN.1 dissectors which store information in
  * actx->private_data the NBAP dissector can't do so because some fields
  * are defined as their own 'PDU' (Like BindingID and TransportLayerAddress)
  * in those cases, the generic ASN.1 dissector creates a NEW 'ASN.1 context'
  * (asn1_ctx_t) and hence a new 'private data' field for them so information
  * can't be passes to/from them.
  */
  nbap_private_data_t *private_data = (nbap_private_data_t *)p_get_proto_data(pinfo->pool, pinfo, proto_nbap, 0);
  if(private_data == NULL ) {
    private_data = wmem_new0(pinfo->pool, nbap_private_data_t);
    p_add_proto_data(pinfo->pool, pinfo, proto_nbap, 0, private_data);
    /* Setting  default values */
    private_data->hsdsch_macdflow_id = 3;
    private_data->crnc_context_present = FALSE;
    private_data->procedure_code = 0xFFFF;
    private_data->dd_mode = 0xFFFF;
    private_data->dch_crc_present = 2; /* Unknown */
    for (i = 0; i < maxNrOfMACdFlows; i++) {
        private_data->nbap_hsdsch_channel_info[i].entity = hs;
    }
  }
  return private_data;
}

/* Helper function to reset the private data struct*/
static void nbap_reset_private_data(packet_info *pinfo)
{
  p_remove_proto_data(pinfo->pool, pinfo, proto_nbap, 0);
}

/*****************************************************************************/


/* Global Variables */

/* Variables for sub elements dissection */
static  const gchar *ProcedureID;
/* Trees */
static wmem_tree_t* edch_flow_port_map = NULL;
wmem_tree_t *nbap_scrambling_code_crncc_map = NULL;
wmem_tree_t *nbap_crncc_urnti_map = NULL;
static wmem_tree_t* com_context_map;

/* This table is used externally from FP, MAC and such, TODO: merge this with
 * lch_contents[] */
guint8 lchId_type_table[]= {
  MAC_CONTENT_UNKNOWN,  /* Shouldn't happen*/
  MAC_CONTENT_DCCH,  /* 1 to 4 SRB => DCCH*/
  MAC_CONTENT_DCCH,
  MAC_CONTENT_DCCH,
  MAC_CONTENT_DCCH,
  MAC_CONTENT_CS_DTCH,  /* 5 to 7 Conv CS speech => ?*/
  MAC_CONTENT_CS_DTCH,
  MAC_CONTENT_CS_DTCH,
  MAC_CONTENT_DCCH, /* 8 SRB => DCCH*/
  MAC_CONTENT_PS_DTCH,  /* 9 maps to DTCH*/
  MAC_CONTENT_UNKNOWN,  /* 10 Conv CS unknown*/
  MAC_CONTENT_PS_DTCH,  /* 11 Interactive PS => DTCH*/
  MAC_CONTENT_PS_DTCH,  /* 12 Streaming PS => DTCH*/
  MAC_CONTENT_CS_DTCH,  /* 13 Streaming CS*/
  MAC_CONTENT_PS_DTCH,  /* 14 Interactive PS => DTCH*/
  MAC_CONTENT_CCCH  /* This is CCCH? */
};

/* Mapping logicalchannel id to RLC_MODE */
guint8 lchId_rlc_map[] = {
  0,
  RLC_UM, /* Logical channel id = 1 is SRB1 which uses RLC_UM*/
  RLC_AM,
  RLC_AM,
  RLC_AM,
  RLC_TM, /*5 to 7 Conv CS Speech*/
  RLC_TM,
  RLC_TM, /*...*/
  RLC_AM,
  RLC_AM,
  RLC_AM,
  RLC_AM,
  RLC_AM,
  RLC_AM,
  RLC_AM,
  RLC_AM, /* This is CCCH which is UM?, probably not */
};

/* Preference variables */
/* Array with preference variables for easy looping, TODO: merge this with
 * lchId_type_table[] */
static int lch_contents[16] = {
  MAC_CONTENT_DCCH,
  MAC_CONTENT_DCCH,
  MAC_CONTENT_DCCH,
  MAC_CONTENT_DCCH,
  MAC_CONTENT_CS_DTCH,
  MAC_CONTENT_CS_DTCH,
  MAC_CONTENT_CS_DTCH,
  MAC_CONTENT_DCCH,
  MAC_CONTENT_PS_DTCH,
  MAC_CONTENT_UNKNOWN,
  MAC_CONTENT_PS_DTCH,
  MAC_CONTENT_PS_DTCH,
  MAC_CONTENT_CS_DTCH,
  MAC_CONTENT_PS_DTCH,
  MAC_CONTENT_CCCH,
  MAC_CONTENT_DCCH
};

static const enum_val_t content_types[] = {
  {"MAC_CONTENT_UNKNOWN", "MAC_CONTENT_UNKNOWN", MAC_CONTENT_UNKNOWN},
  {"MAC_CONTENT_DCCH", "MAC_CONTENT_DCCH", MAC_CONTENT_DCCH},
  {"MAC_CONTENT_PS_DTCH", "MAC_CONTENT_PS_DTCH", MAC_CONTENT_PS_DTCH},
  {"MAC_CONTENT_CS_DTCH", "MAC_CONTENT_CS_DTCH", MAC_CONTENT_CS_DTCH},
  {"MAC_CONTENT_CCCH", "MAC_CONTENT_CCCH", MAC_CONTENT_CCCH},
  {NULL, NULL, -1}};

typedef struct {
  const char *name;
  const char *title;
  const char *description;
} preference_strings;

/* This is used when registering preferences, name, title, description */
static const preference_strings ch_strings[] = {
  {"lch1_content", "Logical Channel 1 Content", "foo"},
  {"lch2_content", "Logical Channel 2 Content", "foo"},
  {"lch3_content", "Logical Channel 3 Content", "foo"},
  {"lch4_content", "Logical Channel 4 Content", "foo"},
  {"lch5_content", "Logical Channel 5 Content", "foo"},
  {"lch6_content", "Logical Channel 6 Content", "foo"},
  {"lch7_content", "Logical Channel 7 Content", "foo"},
  {"lch8_content", "Logical Channel 8 Content", "foo"},
  {"lch9_content", "Logical Channel 9 Content", "foo"},
  {"lch10_content", "Logical Channel 10 Content", "foo"},
  {"lch11_content", "Logical Channel 11 Content", "foo"},
  {"lch12_content", "Logical Channel 12 Content", "foo"},
  {"lch13_content", "Logical Channel 13 Content", "foo"},
  {"lch14_content", "Logical Channel 14 Content", "foo"},
  {"lch15_content", "Logical Channel 15 Content", "foo"},
  {"lch16_content", "Logical Channel 16 Content", "foo"}};

enum ib_sg_enc_type {
  IB_SG_DATA_ENC_VAR_1,
  IB_SG_DATA_ENC_VAR_2
};

static const enum_val_t ib_sg_enc_vals[] = {
  {"Encoding Variant 1 (TS 25.433 Annex D.2)",
   "Encoding Variant 1 (TS 25.433 Annex D.2)", IB_SG_DATA_ENC_VAR_1},
  {"Encoding Variant 2 (TS 25.433 Annex D.3)",
   "Encoding Variant 2 (TS 25.433 Annex D.3)", IB_SG_DATA_ENC_VAR_2},
  {NULL, NULL, -1}
};

static gint preferences_ib_sg_data_encoding = IB_SG_DATA_ENC_VAR_1;

/* Dissector tables */
static dissector_table_t nbap_ies_dissector_table;
static dissector_table_t nbap_extension_dissector_table;
static dissector_table_t nbap_proc_imsg_dissector_table;
static dissector_table_t nbap_proc_sout_dissector_table;
static dissector_table_t nbap_proc_uout_dissector_table;

static int dissect_ProtocolIEFieldValue(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *);
static int dissect_ProtocolExtensionFieldExtensionValue(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *);
static int dissect_InitiatingMessageValue(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *);
static int dissect_SuccessfulOutcomeValue(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *);
static int dissect_UnsuccessfulOutcomeValue(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *);

static guint32 calculate_setup_conv_key(const guint32 transaction_id, const guint32 dd_mode, const guint32 channel_id);
static void add_setup_conv(const packet_info *pinfo _U_, const guint32 transaction_id, const guint32 dd_mode, const guint32 channel_id, const guint32 req_frame_number,
           const address *addr, const guint32 port, umts_fp_conversation_info_t * umts_fp_conversation_info, conversation_t *conv);
static nbap_setup_conv_t* find_setup_conv(const packet_info *pinfo _U_, const guint32 transaction_id, const guint32 dd_mode, const guint32 channel_id);
static void delete_setup_conv(nbap_setup_conv_t *conv);

/*Easy way to add hsdhsch binds for corner cases*/
static void add_hsdsch_bind(packet_info * pinfo);


static const value_string nbap_Criticality_vals[] = {
  {   0, "reject" },
  {   1, "ignore" },
  {   2, "notify" },
  { 0, NULL }
};


static int
dissect_nbap_Criticality(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_MessageDiscriminator_vals[] = {
  {   0, "common" },
  {   1, "dedicated" },
  { 0, NULL }
};


static int
dissect_nbap_MessageDiscriminator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_INTEGER_0_maxPrivateIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxPrivateIEs, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_OBJECT_IDENTIFIER(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_object_identifier(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}


static const value_string nbap_PrivateIE_ID_vals[] = {
  {   0, "local" },
  {   1, "global" },
  { 0, NULL }
};

static const per_choice_t PrivateIE_ID_choice[] = {
  {   0, &hf_nbap_local          , ASN1_NO_EXTENSIONS     , dissect_nbap_INTEGER_0_maxPrivateIEs },
  {   1, &hf_nbap_global         , ASN1_NO_EXTENSIONS     , dissect_nbap_OBJECT_IDENTIFIER },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_PrivateIE_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_PrivateIE_ID, PrivateIE_ID_choice,
                                 NULL);

  return offset;
}


static const value_string nbap_ProcedureCode_vals[] = {
  { NBAP_ID_AUDIT, "id-audit" },
  { NBAP_ID_AUDITREQUIRED, "id-auditRequired" },
  { NBAP_ID_BLOCKRESOURCE, "id-blockResource" },
  { NBAP_ID_CELLDELETION, "id-cellDeletion" },
  { NBAP_ID_CELLRECONFIGURATION, "id-cellReconfiguration" },
  { NBAP_ID_CELLSETUP, "id-cellSetup" },
  { NBAP_ID_CELLSYNCHRONISATIONINITIATION, "id-cellSynchronisationInitiation" },
  { NBAP_ID_CELLSYNCHRONISATIONRECONFIGURATION, "id-cellSynchronisationReconfiguration" },
  { NBAP_ID_CELLSYNCHRONISATIONREPORTING, "id-cellSynchronisationReporting" },
  { NBAP_ID_CELLSYNCHRONISATIONTERMINATION, "id-cellSynchronisationTermination" },
  { NBAP_ID_CELLSYNCHRONISATIONFAILURE, "id-cellSynchronisationFailure" },
  { NBAP_ID_COMMONMEASUREMENTFAILURE, "id-commonMeasurementFailure" },
  { NBAP_ID_COMMONMEASUREMENTINITIATION, "id-commonMeasurementInitiation" },
  { NBAP_ID_COMMONMEASUREMENTREPORT, "id-commonMeasurementReport" },
  { NBAP_ID_COMMONMEASUREMENTTERMINATION, "id-commonMeasurementTermination" },
  { NBAP_ID_COMMONTRANSPORTCHANNELDELETE, "id-commonTransportChannelDelete" },
  { NBAP_ID_COMMONTRANSPORTCHANNELRECONFIGURE, "id-commonTransportChannelReconfigure" },
  { NBAP_ID_COMMONTRANSPORTCHANNELSETUP, "id-commonTransportChannelSetup" },
  { NBAP_ID_COMPRESSEDMODECOMMAND, "id-compressedModeCommand" },
  { NBAP_ID_DEDICATEDMEASUREMENTFAILURE, "id-dedicatedMeasurementFailure" },
  { NBAP_ID_DEDICATEDMEASUREMENTINITIATION, "id-dedicatedMeasurementInitiation" },
  { NBAP_ID_DEDICATEDMEASUREMENTREPORT, "id-dedicatedMeasurementReport" },
  { NBAP_ID_DEDICATEDMEASUREMENTTERMINATION, "id-dedicatedMeasurementTermination" },
  { NBAP_ID_DOWNLINKPOWERCONTROL, "id-downlinkPowerControl" },
  { NBAP_ID_DOWNLINKPOWERTIMESLOTCONTROL, "id-downlinkPowerTimeslotControl" },
  { NBAP_ID_ERRORINDICATIONFORCOMMON, "id-errorIndicationForCommon" },
  { NBAP_ID_ERRORINDICATIONFORDEDICATED, "id-errorIndicationForDedicated" },
  { NBAP_ID_INFORMATIONEXCHANGEFAILURE, "id-informationExchangeFailure" },
  { NBAP_ID_INFORMATIONEXCHANGEINITIATION, "id-informationExchangeInitiation" },
  { NBAP_ID_INFORMATIONEXCHANGETERMINATION, "id-informationExchangeTermination" },
  { NBAP_ID_INFORMATIONREPORTING, "id-informationReporting" },
  { NBAP_ID_BEARERREARRANGEMENT, "id-BearerRearrangement" },
  { NBAP_ID_MBMSNOTIFICATIONUPDATE, "id-mBMSNotificationUpdate" },
  { NBAP_ID_PHYSICALSHAREDCHANNELRECONFIGURATION, "id-physicalSharedChannelReconfiguration" },
  { NBAP_ID_PRIVATEMESSAGEFORCOMMON, "id-privateMessageForCommon" },
  { NBAP_ID_PRIVATEMESSAGEFORDEDICATED, "id-privateMessageForDedicated" },
  { NBAP_ID_RADIOLINKADDITION, "id-radioLinkAddition" },
  { NBAP_ID_RADIOLINKDELETION, "id-radioLinkDeletion" },
  { NBAP_ID_RADIOLINKFAILURE, "id-radioLinkFailure" },
  { NBAP_ID_RADIOLINKPREEMPTION, "id-radioLinkPreemption" },
  { NBAP_ID_RADIOLINKRESTORATION, "id-radioLinkRestoration" },
  { NBAP_ID_RADIOLINKSETUP, "id-radioLinkSetup" },
  { NBAP_ID_RESET, "id-reset" },
  { NBAP_ID_RESOURCESTATUSINDICATION, "id-resourceStatusIndication" },
  { NBAP_ID_CELLSYNCHRONISATIONADJUSTMENT, "id-cellSynchronisationAdjustment" },
  { NBAP_ID_SYNCHRONISEDRADIOLINKRECONFIGURATIONCANCELLATION, "id-synchronisedRadioLinkReconfigurationCancellation" },
  { NBAP_ID_SYNCHRONISEDRADIOLINKRECONFIGURATIONCOMMIT, "id-synchronisedRadioLinkReconfigurationCommit" },
  { NBAP_ID_SYNCHRONISEDRADIOLINKRECONFIGURATIONPREPARATION, "id-synchronisedRadioLinkReconfigurationPreparation" },
  { NBAP_ID_SYSTEMINFORMATIONUPDATE, "id-systemInformationUpdate" },
  { NBAP_ID_UNBLOCKRESOURCE, "id-unblockResource" },
  { NBAP_ID_UNSYNCHRONISEDRADIOLINKRECONFIGURATION, "id-unSynchronisedRadioLinkReconfiguration" },
  { NBAP_ID_RADIOLINKACTIVATION, "id-radioLinkActivation" },
  { NBAP_ID_RADIOLINKPARAMETERUPDATE, "id-radioLinkParameterUpdate" },
  { NBAP_ID_UESTATUSUPDATE, "id-uEStatusUpdate" },
  { NBAP_ID_SECONDARYULFREQUENCYREPORTING, "id-secondaryULFrequencyReporting" },
  { NBAP_ID_SECONDARYULFREQUENCYUPDATE, "id-secondaryULFrequencyUpdate" },
  { 0, NULL }
};


static int
dissect_nbap_ProcedureCode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  guint32 procedure_code;
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, &procedure_code, FALSE);

  nbap_get_private_data(actx->pinfo)->procedure_code = procedure_code;
  col_add_fstr(actx->pinfo->cinfo, COL_INFO, "%s ",
                 val_to_str(procedure_code, nbap_ProcedureCode_vals,
                            "unknown message"));


  return offset;
}


static const value_string nbap_DdMode_vals[] = {
  { tdd, "tdd" },
  { fdd, "fdd" },
  { common, "common" },
  { 0, NULL }
};


static int
dissect_nbap_DdMode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, &nbap_get_private_data(actx->pinfo)->dd_mode, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t ProcedureID_sequence[] = {
  { &hf_nbap_procedureCode  , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProcedureCode },
  { &hf_nbap_ddMode         , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DdMode },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ProcedureID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  ProcedureID = NULL;
  nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ProcedureID, ProcedureID_sequence);

  ProcedureID = wmem_strdup_printf(actx->pinfo->pool, "%s/%s",
                                 val_to_str(nbap_private_data->procedure_code, VALS(nbap_ProcedureCode_vals), "unknown(%u)"),
                                 val_to_str(nbap_private_data->dd_mode, VALS(nbap_DdMode_vals), "unknown(%u)"));
  nbap_private_data->crnc_context_present = FALSE; /*Reset CRNC Com context present flag.*/


  return offset;
}


static const value_string nbap_ProtocolIE_ID_vals[] = {
  { id_AICH_Information, "id-AICH-Information" },
  { id_AICH_InformationItem_ResourceStatusInd, "id-AICH-InformationItem-ResourceStatusInd" },
  { id_Local_Cell_Group_InformationItem_AuditRsp, "id-Local-Cell-Group-InformationItem-AuditRsp" },
  { id_Local_Cell_Group_InformationItem_ResourceStatusInd, "id-Local-Cell-Group-InformationItem-ResourceStatusInd" },
  { id_Local_Cell_Group_InformationItem2_ResourceStatusInd, "id-Local-Cell-Group-InformationItem2-ResourceStatusInd" },
  { id_Local_Cell_Group_InformationList_AuditRsp, "id-Local-Cell-Group-InformationList-AuditRsp" },
  { id_CFNReportingIndicator, "id-CFNReportingIndicator" },
  { id_BCH_Information, "id-BCH-Information" },
  { id_BCH_InformationItem_ResourceStatusInd, "id-BCH-InformationItem-ResourceStatusInd" },
  { id_BCCH_ModificationTime, "id-BCCH-ModificationTime" },
  { id_BlockingPriorityIndicator, "id-BlockingPriorityIndicator" },
  { id_SFNReportingIndicator, "id-SFNReportingIndicator" },
  { id_InnerLoopDLPCStatus, "id-InnerLoopDLPCStatus" },
  { id_Cause, "id-Cause" },
  { id_CCP_InformationItem_AuditRsp, "id-CCP-InformationItem-AuditRsp" },
  { id_CCP_InformationList_AuditRsp, "id-CCP-InformationList-AuditRsp" },
  { id_CCP_InformationItem_ResourceStatusInd, "id-CCP-InformationItem-ResourceStatusInd" },
  { id_Cell_InformationItem_AuditRsp, "id-Cell-InformationItem-AuditRsp" },
  { id_Cell_InformationItem_ResourceStatusInd, "id-Cell-InformationItem-ResourceStatusInd" },
  { id_Cell_InformationList_AuditRsp, "id-Cell-InformationList-AuditRsp" },
  { id_PRACH_ParametersItem_CTCH_SetupRqstTDD, "id-PRACH-ParametersItem-CTCH-SetupRqstTDD" },
  { id_DL_DPCH_TimingAdjustment, "id-DL-DPCH-TimingAdjustment" },
  { id_FPACH_LCR_InformationList_AuditRsp, "id-FPACH-LCR-InformationList-AuditRsp" },
  { id_CellParameterID, "id-CellParameterID" },
  { id_CFN, "id-CFN" },
  { id_C_ID, "id-C-ID" },
  { id_Unused_ProtocolIE_ID_26, "id-Unused-ProtocolIE-ID-26" },
  { id_Unused_ProtocolIE_ID_27, "id-Unused-ProtocolIE-ID-27" },
  { id_DL_PowerBalancing_Information, "id-DL-PowerBalancing-Information" },
  { id_DL_PowerBalancing_ActivationIndicator, "id-DL-PowerBalancing-ActivationIndicator" },
  { id_DL_PowerBalancing_UpdatedIndicator, "id-DL-PowerBalancing-UpdatedIndicator" },
  { id_CommonMeasurementObjectType_CM_Rprt, "id-CommonMeasurementObjectType-CM-Rprt" },
  { id_CommonMeasurementObjectType_CM_Rqst, "id-CommonMeasurementObjectType-CM-Rqst" },
  { id_CommonMeasurementObjectType_CM_Rsp, "id-CommonMeasurementObjectType-CM-Rsp" },
  { id_CommonMeasurementType, "id-CommonMeasurementType" },
  { id_CommonPhysicalChannelID, "id-CommonPhysicalChannelID" },
  { id_CommonPhysicalChannelType_CTCH_SetupRqstFDD, "id-CommonPhysicalChannelType-CTCH-SetupRqstFDD" },
  { id_CommonPhysicalChannelType_CTCH_SetupRqstTDD, "id-CommonPhysicalChannelType-CTCH-SetupRqstTDD" },
  { id_Unknown_38, "id-Unknown-38" },
  { id_CommonMeasurementAccuracy, "id-CommonMeasurementAccuracy" },
  { id_CommunicationControlPortID, "id-CommunicationControlPortID" },
  { id_IPDLParameter_Information_LCR_Cell_SetupRqstTDD, "id-IPDLParameter-Information-LCR-Cell-SetupRqstTDD" },
  { id_IPDLParameter_Information_LCR_Cell_ReconfRqstTDD, "id-IPDLParameter-Information-LCR-Cell-ReconfRqstTDD" },
  { id_ConfigurationGenerationID, "id-ConfigurationGenerationID" },
  { id_CRNC_CommunicationContextID, "id-CRNC-CommunicationContextID" },
  { id_CriticalityDiagnostics, "id-CriticalityDiagnostics" },
  { id_CCTrCH_InformationItem_RL_FailureInd, "id-CCTrCH-InformationItem-RL-FailureInd" },
  { id_CCTrCH_InformationItem_RL_RestoreInd, "id-CCTrCH-InformationItem-RL-RestoreInd" },
  { id_DCHs_to_Add_FDD, "id-DCHs-to-Add-FDD" },
  { id_DCH_AddList_RL_ReconfPrepTDD, "id-DCH-AddList-RL-ReconfPrepTDD" },
  { id_DCHs_to_Add_TDD, "id-DCHs-to-Add-TDD" },
  { id_RL_InformationResponse_LCR_RL_AdditionRspTDD, "id-RL-InformationResponse-LCR-RL-AdditionRspTDD" },
  { id_DCH_DeleteList_RL_ReconfPrepFDD, "id-DCH-DeleteList-RL-ReconfPrepFDD" },
  { id_DCH_DeleteList_RL_ReconfPrepTDD, "id-DCH-DeleteList-RL-ReconfPrepTDD" },
  { id_DCH_DeleteList_RL_ReconfRqstFDD, "id-DCH-DeleteList-RL-ReconfRqstFDD" },
  { id_DCH_DeleteList_RL_ReconfRqstTDD, "id-DCH-DeleteList-RL-ReconfRqstTDD" },
  { id_DCH_FDD_Information, "id-DCH-FDD-Information" },
  { id_DCH_TDD_Information, "id-DCH-TDD-Information" },
  { id_neighbouringTDDCellMeasurementInformationLCR, "id-neighbouringTDDCellMeasurementInformationLCR" },
  { id_DCH_InformationResponse, "id-DCH-InformationResponse" },
  { id_Unknown_60, "id-Unknown-60" },
  { id_Unknown_61, "id-Unknown-61" },
  { id_FDD_DCHs_to_Modify, "id-FDD-DCHs-to-Modify" },
  { id_TDD_DCHs_to_Modify, "id-TDD-DCHs-to-Modify" },
  { id_Unused_ProtocolIE_ID_64, "id-Unused-ProtocolIE-ID-64" },
  { id_DCH_ModifyList_RL_ReconfRqstTDD, "id-DCH-ModifyList-RL-ReconfRqstTDD" },
  { id_PDSCH_RL_ID, "id-PDSCH-RL-ID" },
  { id_DedicatedMeasurementObjectType_DM_Rprt, "id-DedicatedMeasurementObjectType-DM-Rprt" },
  { id_DedicatedMeasurementObjectType_DM_Rqst, "id-DedicatedMeasurementObjectType-DM-Rqst" },
  { id_DedicatedMeasurementObjectType_DM_Rsp, "id-DedicatedMeasurementObjectType-DM-Rsp" },
  { id_DedicatedMeasurementType, "id-DedicatedMeasurementType" },
  { id_Unknown_71, "id-Unknown-71" },
  { id_DL_CCTrCH_InformationItem_RL_SetupRqstTDD, "id-DL-CCTrCH-InformationItem-RL-SetupRqstTDD" },
  { id_DL_CCTrCH_InformationList_RL_AdditionRqstTDD, "id-DL-CCTrCH-InformationList-RL-AdditionRqstTDD" },
  { id_DL_DPCH_LCR_Information_RL_SetupRqstTDD, "id-DL-DPCH-LCR-Information-RL-SetupRqstTDD" },
  { id_Unknown_75, "id-Unknown-75" },
  { id_DL_CCTrCH_InformationList_RL_SetupRqstTDD, "id-DL-CCTrCH-InformationList-RL-SetupRqstTDD" },
  { id_DL_DPCH_InformationItem_RL_AdditionRqstTDD, "id-DL-DPCH-InformationItem-RL-AdditionRqstTDD" },
  { id_DwPCH_LCR_Information, "id-DwPCH-LCR-Information" },
  { id_DL_DPCH_InformationList_RL_SetupRqstTDD, "id-DL-DPCH-InformationList-RL-SetupRqstTDD" },
  { id_Unknown_80, "id-Unknown-80" },
  { id_DL_DPCH_Information_RL_ReconfPrepFDD, "id-DL-DPCH-Information-RL-ReconfPrepFDD" },
  { id_DL_DPCH_Information_RL_ReconfRqstFDD, "id-DL-DPCH-Information-RL-ReconfRqstFDD" },
  { id_DL_DPCH_Information_RL_SetupRqstFDD, "id-DL-DPCH-Information-RL-SetupRqstFDD" },
  { id_DL_ReferencePowerInformationItem_DL_PC_Rqst, "id-DL-ReferencePowerInformationItem-DL-PC-Rqst" },
  { id_DLReferencePower, "id-DLReferencePower" },
  { id_DLReferencePowerList_DL_PC_Rqst, "id-DLReferencePowerList-DL-PC-Rqst" },
  { id_Unused_ProtocolIE_ID_87, "id-Unused-ProtocolIE-ID-87" },
  { id_Unknown_88, "id-Unknown-88" },
  { id_Unused_ProtocolIE_ID_89, "id-Unused-ProtocolIE-ID-89" },
  { id_DwPCH_LCR_InformationList_AuditRsp, "id-DwPCH-LCR-InformationList-AuditRsp" },
  { id_Unused_ProtocolIE_ID_91, "id-Unused-ProtocolIE-ID-91" },
  { id_Unknown_92, "id-Unknown-92" },
  { id_Unused_ProtocolIE_ID_93, "id-Unused-ProtocolIE-ID-93" },
  { id_Unused_ProtocolIE_ID_94, "id-Unused-ProtocolIE-ID-94" },
  { id_Unknown_95, "id-Unknown-95" },
  { id_DSCHs_to_Add_TDD, "id-DSCHs-to-Add-TDD" },
  { id_DwPCH_LCR_Information_Cell_SetupRqstTDD, "id-DwPCH-LCR-Information-Cell-SetupRqstTDD" },
  { id_DSCH_Information_DeleteList_RL_ReconfPrepTDD, "id-DSCH-Information-DeleteList-RL-ReconfPrepTDD" },
  { id_DwPCH_LCR_Information_Cell_ReconfRqstTDD, "id-DwPCH-LCR-Information-Cell-ReconfRqstTDD" },
  { id_DSCH_Information_ModifyList_RL_ReconfPrepTDD, "id-DSCH-Information-ModifyList-RL-ReconfPrepTDD" },
  { id_DwPCH_LCR_Information_ResourceStatusInd, "id-DwPCH-LCR-Information-ResourceStatusInd" },
  { id_bindingID, "id-bindingID" },
  { id_RL_Specific_DCH_Info, "id-RL-Specific-DCH-Info" },
  { id_transportlayeraddress, "id-transportlayeraddress" },
  { id_DSCH_InformationResponse, "id-DSCH-InformationResponse" },
  { id_Unused_ProtocolIE_ID_106, "id-Unused-ProtocolIE-ID-106" },
  { id_DSCH_TDD_Information, "id-DSCH-TDD-Information" },
  { id_Unused_ProtocolIE_ID_108, "id-Unused-ProtocolIE-ID-108" },
  { id_Unknown_109, "id-Unknown-109" },
  { id_Unused_ProtocolIE_ID_110, "id-Unused-ProtocolIE-ID-110" },
  { id_Unused_ProtocolIE_ID_111, "id-Unused-ProtocolIE-ID-111" },
  { id_Unused_ProtocolIE_ID_112, "id-Unused-ProtocolIE-ID-112" },
  { id_End_Of_Audit_Sequence_Indicator, "id-End-Of-Audit-Sequence-Indicator" },
  { id_Start_Of_Audit_Sequence_Indicator, "id-Start-Of-Audit-Sequence-Indicator" },
  { id_MessageStructure, "id-MessageStructure" },
  { id_FACH_Information, "id-FACH-Information" },
  { id_FACH_InformationItem_ResourceStatusInd, "id-FACH-InformationItem-ResourceStatusInd" },
  { id_Unknown_118, "id-Unknown-118" },
  { id_LateEntranceCellSyncInfoItem_CellSyncReprtTDD, "id-LateEntranceCellSyncInfoItem-CellSyncReprtTDD" },
  { id_FACH_ParametersList_CTCH_ReconfRqstTDD, "id-FACH-ParametersList-CTCH-ReconfRqstTDD" },
  { id_FACH_ParametersListIE_CTCH_SetupRqstFDD, "id-FACH-ParametersListIE-CTCH-SetupRqstFDD" },
  { id_FACH_ParametersListIE_CTCH_SetupRqstTDD, "id-FACH-ParametersListIE-CTCH-SetupRqstTDD" },
  { id_IndicationType_ResourceStatusInd, "id-IndicationType-ResourceStatusInd" },
  { id_Local_Cell_ID, "id-Local-Cell-ID" },
  { id_Local_Cell_InformationItem_AuditRsp, "id-Local-Cell-InformationItem-AuditRsp" },
  { id_Local_Cell_InformationItem_ResourceStatusInd, "id-Local-Cell-InformationItem-ResourceStatusInd" },
  { id_Local_Cell_InformationItem2_ResourceStatusInd, "id-Local-Cell-InformationItem2-ResourceStatusInd" },
  { id_Local_Cell_InformationList_AuditRsp, "id-Local-Cell-InformationList-AuditRsp" },
  { id_AdjustmentPeriod, "id-AdjustmentPeriod" },
  { id_MaxAdjustmentStep, "id-MaxAdjustmentStep" },
  { id_MaximumTransmissionPower, "id-MaximumTransmissionPower" },
  { id_MeasurementFilterCoefficient, "id-MeasurementFilterCoefficient" },
  { id_MeasurementID, "id-MeasurementID" },
  { id_MIB_SB_SIB_InformationList_SystemInfoUpdateRqst, "id-MIB-SB-SIB-InformationList-SystemInfoUpdateRqst" },
  { id_DCH_RearrangeList_Bearer_RearrangeInd, "id-DCH-RearrangeList-Bearer-RearrangeInd" },
  { id_DSCH_RearrangeList_Bearer_RearrangeInd, "id-DSCH-RearrangeList-Bearer-RearrangeInd" },
  { id_Unknown_137, "id-Unknown-137" },
  { id_SignallingBearerRequestIndicator, "id-SignallingBearerRequestIndicator" },
  { id_TargetCommunicationControlPortID, "id-TargetCommunicationControlPortID" },
  { id_Unknown_140, "id-Unknown-140" },
  { id_USCH_RearrangeList_Bearer_RearrangeInd, "id-USCH-RearrangeList-Bearer-RearrangeInd" },
  { id_Unused_ProtocolIE_ID_142, "id-Unused-ProtocolIE-ID-142" },
  { id_NodeB_CommunicationContextID, "id-NodeB-CommunicationContextID" },
  { id_P_CCPCH_Information, "id-P-CCPCH-Information" },
  { id_P_CCPCH_InformationItem_ResourceStatusInd, "id-P-CCPCH-InformationItem-ResourceStatusInd" },
  { id_P_CPICH_Information, "id-P-CPICH-Information" },
  { id_P_CPICH_InformationItem_ResourceStatusInd, "id-P-CPICH-InformationItem-ResourceStatusInd" },
  { id_P_SCH_Information, "id-P-SCH-Information" },
  { id_Unknown_149, "id-Unknown-149" },
  { id_PCCPCH_Information_Cell_ReconfRqstTDD, "id-PCCPCH-Information-Cell-ReconfRqstTDD" },
  { id_PCCPCH_Information_Cell_SetupRqstTDD, "id-PCCPCH-Information-Cell-SetupRqstTDD" },
  { id_Unknown_152, "id-Unknown-152" },
  { id_Unknown_153, "id-Unknown-153" },
  { id_maxFACH_Power_LCR_CTCH_SetupRqstTDD, "id-maxFACH-Power-LCR-CTCH-SetupRqstTDD" },
  { id_PCH_Parameters_CTCH_ReconfRqstTDD, "id-PCH-Parameters-CTCH-ReconfRqstTDD" },
  { id_PCH_ParametersItem_CTCH_SetupRqstFDD, "id-PCH-ParametersItem-CTCH-SetupRqstFDD" },
  { id_PCH_ParametersItem_CTCH_SetupRqstTDD, "id-PCH-ParametersItem-CTCH-SetupRqstTDD" },
  { id_PCH_Information, "id-PCH-Information" },
  { id_Unknown_159, "id-Unknown-159" },
  { id_Unknown_160, "id-Unknown-160" },
  { id_PDSCH_Information_AddListIE_PSCH_ReconfRqst, "id-PDSCH-Information-AddListIE-PSCH-ReconfRqst" },
  { id_PDSCH_Information_ModifyListIE_PSCH_ReconfRqst, "id-PDSCH-Information-ModifyListIE-PSCH-ReconfRqst" },
  { id_PDSCHSets_AddList_PSCH_ReconfRqst, "id-PDSCHSets-AddList-PSCH-ReconfRqst" },
  { id_PDSCHSets_DeleteList_PSCH_ReconfRqst, "id-PDSCHSets-DeleteList-PSCH-ReconfRqst" },
  { id_PDSCHSets_ModifyList_PSCH_ReconfRqst, "id-PDSCHSets-ModifyList-PSCH-ReconfRqst" },
  { id_PICH_Information, "id-PICH-Information" },
  { id_PICH_ParametersItem_CTCH_SetupRqstTDD, "id-PICH-ParametersItem-CTCH-SetupRqstTDD" },
  { id_PICH_Parameters_CTCH_ReconfRqstTDD, "id-PICH-Parameters-CTCH-ReconfRqstTDD" },
  { id_PowerAdjustmentType, "id-PowerAdjustmentType" },
  { id_PRACH_Information, "id-PRACH-Information" },
  { id_Unknown_171, "id-Unknown-171" },
  { id_Unknown_172, "id-Unknown-172" },
  { id_Unknown_173, "id-Unknown-173" },
  { id_maxFACH_Power_LCR_CTCH_ReconfRqstTDD, "id-maxFACH-Power-LCR-CTCH-ReconfRqstTDD" },
  { id_PrimaryCCPCH_Information_Cell_ReconfRqstFDD, "id-PrimaryCCPCH-Information-Cell-ReconfRqstFDD" },
  { id_PrimaryCCPCH_Information_Cell_SetupRqstFDD, "id-PrimaryCCPCH-Information-Cell-SetupRqstFDD" },
  { id_PrimaryCPICH_Information_Cell_ReconfRqstFDD, "id-PrimaryCPICH-Information-Cell-ReconfRqstFDD" },
  { id_PrimaryCPICH_Information_Cell_SetupRqstFDD, "id-PrimaryCPICH-Information-Cell-SetupRqstFDD" },
  { id_PrimarySCH_Information_Cell_ReconfRqstFDD, "id-PrimarySCH-Information-Cell-ReconfRqstFDD" },
  { id_PrimarySCH_Information_Cell_SetupRqstFDD, "id-PrimarySCH-Information-Cell-SetupRqstFDD" },
  { id_PrimaryScramblingCode, "id-PrimaryScramblingCode" },
  { id_Unknown_182, "id-Unknown-182" },
  { id_SCH_Information_Cell_ReconfRqstTDD, "id-SCH-Information-Cell-ReconfRqstTDD" },
  { id_SCH_Information_Cell_SetupRqstTDD, "id-SCH-Information-Cell-SetupRqstTDD" },
  { id_PUSCH_Information_AddListIE_PSCH_ReconfRqst, "id-PUSCH-Information-AddListIE-PSCH-ReconfRqst" },
  { id_PUSCH_Information_ModifyListIE_PSCH_ReconfRqst, "id-PUSCH-Information-ModifyListIE-PSCH-ReconfRqst" },
  { id_PUSCHSets_AddList_PSCH_ReconfRqst, "id-PUSCHSets-AddList-PSCH-ReconfRqst" },
  { id_PUSCHSets_DeleteList_PSCH_ReconfRqst, "id-PUSCHSets-DeleteList-PSCH-ReconfRqst" },
  { id_PUSCHSets_ModifyList_PSCH_ReconfRqst, "id-PUSCHSets-ModifyList-PSCH-ReconfRqst" },
  { id_RACH_Information, "id-RACH-Information" },
  { id_Unknown_191, "id-Unknown-191" },
  { id_Unknown_192, "id-Unknown-192" },
  { id_Unknown_193, "id-Unknown-193" },
  { id_Unknown_194, "id-Unknown-194" },
  { id_Unknown_195, "id-Unknown-195" },
  { id_RACH_ParametersItem_CTCH_SetupRqstFDD, "id-RACH-ParametersItem-CTCH-SetupRqstFDD" },
  { id_RACH_ParameterItem_CTCH_SetupRqstTDD, "id-RACH-ParameterItem-CTCH-SetupRqstTDD" },
  { id_ReportCharacteristics, "id-ReportCharacteristics" },
  { id_Reporting_Object_RL_FailureInd, "id-Reporting-Object-RL-FailureInd" },
  { id_Reporting_Object_RL_RestoreInd, "id-Reporting-Object-RL-RestoreInd" },
  { id_Unknown_201, "id-Unknown-201" },
  { id_RL_InformationItem_DM_Rprt, "id-RL-InformationItem-DM-Rprt" },
  { id_RL_InformationItem_DM_Rqst, "id-RL-InformationItem-DM-Rqst" },
  { id_RL_InformationItem_DM_Rsp, "id-RL-InformationItem-DM-Rsp" },
  { id_RL_InformationItem_RL_AdditionRqstFDD, "id-RL-InformationItem-RL-AdditionRqstFDD" },
  { id_RL_informationItem_RL_DeletionRqst, "id-RL-informationItem-RL-DeletionRqst" },
  { id_RL_InformationItem_RL_FailureInd, "id-RL-InformationItem-RL-FailureInd" },
  { id_RL_InformationItem_RL_ReconfPrepFDD, "id-RL-InformationItem-RL-ReconfPrepFDD" },
  { id_RL_InformationItem_RL_ReconfRqstFDD, "id-RL-InformationItem-RL-ReconfRqstFDD" },
  { id_RL_InformationItem_RL_RestoreInd, "id-RL-InformationItem-RL-RestoreInd" },
  { id_RL_InformationItem_RL_SetupRqstFDD, "id-RL-InformationItem-RL-SetupRqstFDD" },
  { id_RL_InformationList_RL_AdditionRqstFDD, "id-RL-InformationList-RL-AdditionRqstFDD" },
  { id_RL_informationList_RL_DeletionRqst, "id-RL-informationList-RL-DeletionRqst" },
  { id_RL_InformationList_RL_ReconfPrepFDD, "id-RL-InformationList-RL-ReconfPrepFDD" },
  { id_RL_InformationList_RL_ReconfRqstFDD, "id-RL-InformationList-RL-ReconfRqstFDD" },
  { id_RL_InformationList_RL_SetupRqstFDD, "id-RL-InformationList-RL-SetupRqstFDD" },
  { id_RL_InformationResponseItem_RL_AdditionRspFDD, "id-RL-InformationResponseItem-RL-AdditionRspFDD" },
  { id_RL_InformationResponseItem_RL_ReconfReady, "id-RL-InformationResponseItem-RL-ReconfReady" },
  { id_RL_InformationResponseItem_RL_ReconfRsp, "id-RL-InformationResponseItem-RL-ReconfRsp" },
  { id_RL_InformationResponseItem_RL_SetupRspFDD, "id-RL-InformationResponseItem-RL-SetupRspFDD" },
  { id_RL_InformationResponseList_RL_AdditionRspFDD, "id-RL-InformationResponseList-RL-AdditionRspFDD" },
  { id_RL_InformationResponseList_RL_ReconfReady, "id-RL-InformationResponseList-RL-ReconfReady" },
  { id_RL_InformationResponseList_RL_ReconfRsp, "id-RL-InformationResponseList-RL-ReconfRsp" },
  { id_RL_InformationResponseList_RL_SetupRspFDD, "id-RL-InformationResponseList-RL-SetupRspFDD" },
  { id_RL_InformationResponse_RL_AdditionRspTDD, "id-RL-InformationResponse-RL-AdditionRspTDD" },
  { id_RL_InformationResponse_RL_SetupRspTDD, "id-RL-InformationResponse-RL-SetupRspTDD" },
  { id_RL_Information_RL_AdditionRqstTDD, "id-RL-Information-RL-AdditionRqstTDD" },
  { id_RL_Information_RL_ReconfRqstTDD, "id-RL-Information-RL-ReconfRqstTDD" },
  { id_RL_Information_RL_ReconfPrepTDD, "id-RL-Information-RL-ReconfPrepTDD" },
  { id_RL_Information_RL_SetupRqstTDD, "id-RL-Information-RL-SetupRqstTDD" },
  { id_DelayedActivation, "id-DelayedActivation" },
  { id_DelayedActivationList_RL_ActivationCmdFDD, "id-DelayedActivationList-RL-ActivationCmdFDD" },
  { id_DelayedActivationInformation_RL_ActivationCmdFDD, "id-DelayedActivationInformation-RL-ActivationCmdFDD" },
  { id_DelayedActivationList_RL_ActivationCmdTDD, "id-DelayedActivationList-RL-ActivationCmdTDD" },
  { id_DelayedActivationInformation_RL_ActivationCmdTDD, "id-DelayedActivationInformation-RL-ActivationCmdTDD" },
  { id_RL_ReconfigurationFailureItem_RL_ReconfFailure, "id-RL-ReconfigurationFailureItem-RL-ReconfFailure" },
  { id_RL_InformationList_RL_PreemptRequiredInd, "id-RL-InformationList-RL-PreemptRequiredInd" },
  { id_RL_Set_InformationItem_DM_Rprt, "id-RL-Set-InformationItem-DM-Rprt" },
  { id_Unknown_239, "id-Unknown-239" },
  { id_RL_Set_InformationItem_DM_Rsp, "id-RL-Set-InformationItem-DM-Rsp" },
  { id_RL_Set_InformationItem_RL_FailureInd, "id-RL-Set-InformationItem-RL-FailureInd" },
  { id_RL_Set_InformationItem_RL_RestoreInd, "id-RL-Set-InformationItem-RL-RestoreInd" },
  { id_Unknown_243, "id-Unknown-243" },
  { id_Unknown_244, "id-Unknown-244" },
  { id_Unknown_245, "id-Unknown-245" },
  { id_Unknown_246, "id-Unknown-246" },
  { id_S_CCPCH_Information, "id-S-CCPCH-Information" },
  { id_Unknown_248, "id-Unknown-248" },
  { id_S_CPICH_Information, "id-S-CPICH-Information" },
  { id_Unknown_250, "id-Unknown-250" },
  { id_SCH_Information, "id-SCH-Information" },
  { id_Unknown_252, "id-Unknown-252" },
  { id_S_SCH_Information, "id-S-SCH-Information" },
  { id_Unknown_254, "id-Unknown-254" },
  { id_Unknown_255, "id-Unknown-255" },
  { id_Unknown_256, "id-Unknown-256" },
  { id_Secondary_CCPCHListIE_CTCH_ReconfRqstTDD, "id-Secondary-CCPCHListIE-CTCH-ReconfRqstTDD" },
  { id_Secondary_CCPCH_parameterListIE_CTCH_SetupRqstTDD, "id-Secondary-CCPCH-parameterListIE-CTCH-SetupRqstTDD" },
  { id_Secondary_CCPCH_Parameters_CTCH_ReconfRqstTDD, "id-Secondary-CCPCH-Parameters-CTCH-ReconfRqstTDD" },
  { id_SecondaryCPICH_InformationItem_Cell_ReconfRqstFDD, "id-SecondaryCPICH-InformationItem-Cell-ReconfRqstFDD" },
  { id_SecondaryCPICH_InformationItem_Cell_SetupRqstFDD, "id-SecondaryCPICH-InformationItem-Cell-SetupRqstFDD" },
  { id_SecondaryCPICH_InformationList_Cell_ReconfRqstFDD, "id-SecondaryCPICH-InformationList-Cell-ReconfRqstFDD" },
  { id_SecondaryCPICH_InformationList_Cell_SetupRqstFDD, "id-SecondaryCPICH-InformationList-Cell-SetupRqstFDD" },
  { id_SecondarySCH_Information_Cell_ReconfRqstFDD, "id-SecondarySCH-Information-Cell-ReconfRqstFDD" },
  { id_SecondarySCH_Information_Cell_SetupRqstFDD, "id-SecondarySCH-Information-Cell-SetupRqstFDD" },
  { id_SegmentInformationListIE_SystemInfoUpdate, "id-SegmentInformationListIE-SystemInfoUpdate" },
  { id_Unknown_267, "id-Unknown-267" },
  { id_SFN, "id-SFN" },
  { id_ShutdownTimer, "id-ShutdownTimer" },
  { id_Successful_RL_InformationRespItem_RL_AdditionFailureFDD, "id-Successful-RL-InformationRespItem-RL-AdditionFailureFDD" },
  { id_Successful_RL_InformationRespItem_RL_SetupFailureFDD, "id-Successful-RL-InformationRespItem-RL-SetupFailureFDD" },
  { id_Unknown_272, "id-Unknown-272" },
  { id_Unknown_273, "id-Unknown-273" },
  { id_SyncCase, "id-SyncCase" },
  { id_SyncCaseIndicatorItem_Cell_SetupRqstTDD_PSCH, "id-SyncCaseIndicatorItem-Cell-SetupRqstTDD-PSCH" },
  { id_T_Cell, "id-T-Cell" },
  { id_TimeSlotConfigurationList_Cell_ReconfRqstTDD, "id-TimeSlotConfigurationList-Cell-ReconfRqstTDD" },
  { id_TimeSlotConfigurationList_Cell_SetupRqstTDD, "id-TimeSlotConfigurationList-Cell-SetupRqstTDD" },
  { id_TransmissionDiversityApplied, "id-TransmissionDiversityApplied" },
  { id_UARFCNforNt, "id-UARFCNforNt" },
  { id_UARFCNforNd, "id-UARFCNforNd" },
  { id_UARFCNforNu, "id-UARFCNforNu" },
  { id_TimeslotISCPInfo, "id-TimeslotISCPInfo" },
  { id_UL_CCTrCH_InformationItem_RL_SetupRqstTDD, "id-UL-CCTrCH-InformationItem-RL-SetupRqstTDD" },
  { id_UL_CCTrCH_InformationList_RL_AdditionRqstTDD, "id-UL-CCTrCH-InformationList-RL-AdditionRqstTDD" },
  { id_RL_InformationItem_RL_PreemptRequiredInd, "id-RL-InformationItem-RL-PreemptRequiredInd" },
  { id_TimingAdvanceApplied, "id-TimingAdvanceApplied" },
  { id_UL_CCTrCH_InformationList_RL_SetupRqstTDD, "id-UL-CCTrCH-InformationList-RL-SetupRqstTDD" },
  { id_UL_DPCH_InformationItem_RL_AdditionRqstTDD, "id-UL-DPCH-InformationItem-RL-AdditionRqstTDD" },
  { id_FPACH_LCR_Information, "id-FPACH-LCR-Information" },
  { id_UL_DPCH_InformationList_RL_SetupRqstTDD, "id-UL-DPCH-InformationList-RL-SetupRqstTDD" },
  { id_FPACH_LCR_Information_AuditRsp, "id-FPACH-LCR-Information-AuditRsp" },
  { id_UL_DPCH_Information_RL_ReconfPrepFDD, "id-UL-DPCH-Information-RL-ReconfPrepFDD" },
  { id_UL_DPCH_Information_RL_ReconfRqstFDD, "id-UL-DPCH-Information-RL-ReconfRqstFDD" },
  { id_UL_DPCH_Information_RL_SetupRqstFDD, "id-UL-DPCH-Information-RL-SetupRqstFDD" },
  { id_Unsuccessful_RL_InformationRespItem_RL_AdditionFailureFDD, "id-Unsuccessful-RL-InformationRespItem-RL-AdditionFailureFDD" },
  { id_Unsuccessful_RL_InformationRespItem_RL_SetupFailureFDD, "id-Unsuccessful-RL-InformationRespItem-RL-SetupFailureFDD" },
  { id_Unknown_298, "id-Unknown-298" },
  { id_Unknown_299, "id-Unknown-299" },
  { id_Unsuccessful_RL_InformationResp_RL_AdditionFailureTDD, "id-Unsuccessful-RL-InformationResp-RL-AdditionFailureTDD" },
  { id_Unsuccessful_RL_InformationResp_RL_SetupFailureTDD, "id-Unsuccessful-RL-InformationResp-RL-SetupFailureTDD" },
  { id_USCH_Information_Add, "id-USCH-Information-Add" },
  { id_Unknown_303, "id-Unknown-303" },
  { id_USCH_Information_DeleteList_RL_ReconfPrepTDD, "id-USCH-Information-DeleteList-RL-ReconfPrepTDD" },
  { id_Unknown_305, "id-Unknown-305" },
  { id_USCH_Information_ModifyList_RL_ReconfPrepTDD, "id-USCH-Information-ModifyList-RL-ReconfPrepTDD" },
  { id_Unknown_307, "id-Unknown-307" },
  { id_Unknown_308, "id-Unknown-308" },
  { id_USCH_InformationResponse, "id-USCH-InformationResponse" },
  { id_USCH_Information, "id-USCH-Information" },
  { id_FPACH_LCR_InformationList_ResourceStatusInd, "id-FPACH-LCR-InformationList-ResourceStatusInd" },
  { id_FPACH_LCR_Parameters_CTCH_SetupRqstTDD, "id-FPACH-LCR-Parameters-CTCH-SetupRqstTDD" },
  { id_Unknown_313, "id-Unknown-313" },
  { id_FPACH_LCR_Parameters_CTCH_ReconfRqstTDD, "id-FPACH-LCR-Parameters-CTCH-ReconfRqstTDD" },
  { id_Active_Pattern_Sequence_Information, "id-Active-Pattern-Sequence-Information" },
  { id_AICH_ParametersListIE_CTCH_ReconfRqstFDD, "id-AICH-ParametersListIE-CTCH-ReconfRqstFDD" },
  { id_AdjustmentRatio, "id-AdjustmentRatio" },
  { id_Unknown_318, "id-Unknown-318" },
  { id_Unknown_319, "id-Unknown-319" },
  { id_Not_Used_320, "id-Not-Used-320" },
  { id_Unknown_321, "id-Unknown-321" },
  { id_Not_Used_322, "id-Not-Used-322" },
  { id_FACH_ParametersListIE_CTCH_ReconfRqstFDD, "id-FACH-ParametersListIE-CTCH-ReconfRqstFDD" },
  { id_CauseLevel_PSCH_ReconfFailure, "id-CauseLevel-PSCH-ReconfFailure" },
  { id_CauseLevel_RL_AdditionFailureFDD, "id-CauseLevel-RL-AdditionFailureFDD" },
  { id_CauseLevel_RL_AdditionFailureTDD, "id-CauseLevel-RL-AdditionFailureTDD" },
  { id_CauseLevel_RL_ReconfFailure, "id-CauseLevel-RL-ReconfFailure" },
  { id_CauseLevel_RL_SetupFailureFDD, "id-CauseLevel-RL-SetupFailureFDD" },
  { id_CauseLevel_RL_SetupFailureTDD, "id-CauseLevel-RL-SetupFailureTDD" },
  { id_Not_Used_330, "id-Not-Used-330" },
  { id_Unknown_331, "id-Unknown-331" },
  { id_Not_Used_332, "id-Not-Used-332" },
  { id_Closed_Loop_Timing_Adjustment_Mode, "id-Closed-Loop-Timing-Adjustment-Mode" },
  { id_CommonPhysicalChannelType_CTCH_ReconfRqstFDD, "id-CommonPhysicalChannelType-CTCH-ReconfRqstFDD" },
  { id_Compressed_Mode_Deactivation_Flag, "id-Compressed-Mode-Deactivation-Flag" },
  { id_Not_Used_336, "id-Not-Used-336" },
  { id_Unknown_337, "id-Unknown-337" },
  { id_Unknown_338, "id-Unknown-338" },
  { id_Unknown_339, "id-Unknown-339" },
  { id_Unknown_340, "id-Unknown-340" },
  { id_Unknown_341, "id-Unknown-341" },
  { id_Not_Used_342, "id-Not-Used-342" },
  { id_Not_Used_343, "id-Not-Used-343" },
  { id_Unknown_344, "id-Unknown-344" },
  { id_Unknown_345, "id-Unknown-345" },
  { id_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD, "id-DL-CCTrCH-InformationAddList-RL-ReconfPrepTDD" },
  { id_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD, "id-DL-CCTrCH-InformationDeleteItem-RL-ReconfRqstTDD" },
  { id_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD, "id-DL-CCTrCH-InformationDeleteList-RL-ReconfPrepTDD" },
  { id_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD, "id-DL-CCTrCH-InformationDeleteList-RL-ReconfRqstTDD" },
  { id_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD, "id-DL-CCTrCH-InformationModifyItem-RL-ReconfRqstTDD" },
  { id_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD, "id-DL-CCTrCH-InformationModifyList-RL-ReconfPrepTDD" },
  { id_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD, "id-DL-CCTrCH-InformationModifyList-RL-ReconfRqstTDD" },
  { id_DL_DPCH_InformationAddListIE_RL_ReconfPrepTDD, "id-DL-DPCH-InformationAddListIE-RL-ReconfPrepTDD" },
  { id_Unknown_354, "id-Unknown-354" },
  { id_DL_DPCH_InformationModify_AddListIE_RL_ReconfPrepTDD, "id-DL-DPCH-InformationModify-AddListIE-RL-ReconfPrepTDD" },
  { id_DL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD, "id-DL-DPCH-InformationModify-DeleteListIE-RL-ReconfPrepTDD" },
  { id_DL_DPCH_InformationModify_ModifyListIE_RL_ReconfPrepTDD, "id-DL-DPCH-InformationModify-ModifyListIE-RL-ReconfPrepTDD" },
  { id_DL_TPC_Pattern01Count, "id-DL-TPC-Pattern01Count" },
  { id_DPCHConstant, "id-DPCHConstant" },
  { id_Unknown_360, "id-Unknown-360" },
  { id_Unknown_361, "id-Unknown-361" },
  { id_FACH_ParametersList_CTCH_SetupRsp, "id-FACH-ParametersList-CTCH-SetupRsp" },
  { id_Unknown_363, "id-Unknown-363" },
  { id_Unknown_364, "id-Unknown-364" },
  { id_Unknown_365, "id-Unknown-365" },
  { id_Unknown_366, "id-Unknown-366" },
  { id_Unknown_367, "id-Unknown-367" },
  { id_Unknown_368, "id-Unknown-368" },
  { id_Limited_power_increase_information_Cell_SetupRqstFDD, "id-Limited-power-increase-information-Cell-SetupRqstFDD" },
  { id_Unknown_370, "id-Unknown-370" },
  { id_Unknown_371, "id-Unknown-371" },
  { id_Unknown_372, "id-Unknown-372" },
  { id_Unknown_373, "id-Unknown-373" },
  { id_PCH_Parameters_CTCH_SetupRsp, "id-PCH-Parameters-CTCH-SetupRsp" },
  { id_PCH_ParametersItem_CTCH_ReconfRqstFDD, "id-PCH-ParametersItem-CTCH-ReconfRqstFDD" },
  { id_Not_Used_376, "id-Not-Used-376" },
  { id_Unknown_377, "id-Unknown-377" },
  { id_Unknown_378, "id-Unknown-378" },
  { id_Unknown_379, "id-Unknown-379" },
  { id_PICH_ParametersItem_CTCH_ReconfRqstFDD, "id-PICH-ParametersItem-CTCH-ReconfRqstFDD" },
  { id_PRACHConstant, "id-PRACHConstant" },
  { id_Unknown_382, "id-Unknown-382" },
  { id_PRACH_ParametersListIE_CTCH_ReconfRqstFDD, "id-PRACH-ParametersListIE-CTCH-ReconfRqstFDD" },
  { id_PUSCHConstant, "id-PUSCHConstant" },
  { id_RACH_Parameters_CTCH_SetupRsp, "id-RACH-Parameters-CTCH-SetupRsp" },
  { id_Unknown_386, "id-Unknown-386" },
  { id_Unknown_387, "id-Unknown-387" },
  { id_Unknown_388, "id-Unknown-388" },
  { id_Unknown_389, "id-Unknown-389" },
  { id_Unknown_390, "id-Unknown-390" },
  { id_Unknown_391, "id-Unknown-391" },
  { id_Unknown_392, "id-Unknown-392" },
  { id_Synchronisation_Configuration_Cell_ReconfRqst, "id-Synchronisation-Configuration-Cell-ReconfRqst" },
  { id_Synchronisation_Configuration_Cell_SetupRqst, "id-Synchronisation-Configuration-Cell-SetupRqst" },
  { id_Transmission_Gap_Pattern_Sequence_Information, "id-Transmission-Gap-Pattern-Sequence-Information" },
  { id_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD, "id-UL-CCTrCH-InformationAddList-RL-ReconfPrepTDD" },
  { id_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD, "id-UL-CCTrCH-InformationDeleteItem-RL-ReconfRqstTDD" },
  { id_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD, "id-UL-CCTrCH-InformationDeleteList-RL-ReconfPrepTDD" },
  { id_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD, "id-UL-CCTrCH-InformationDeleteList-RL-ReconfRqstTDD" },
  { id_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD, "id-UL-CCTrCH-InformationModifyItem-RL-ReconfRqstTDD" },
  { id_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD, "id-UL-CCTrCH-InformationModifyList-RL-ReconfPrepTDD" },
  { id_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD, "id-UL-CCTrCH-InformationModifyList-RL-ReconfRqstTDD" },
  { id_UL_DPCH_InformationAddListIE_RL_ReconfPrepTDD, "id-UL-DPCH-InformationAddListIE-RL-ReconfPrepTDD" },
  { id_Unknown_404, "id-Unknown-404" },
  { id_UL_DPCH_InformationModify_AddListIE_RL_ReconfPrepTDD, "id-UL-DPCH-InformationModify-AddListIE-RL-ReconfPrepTDD" },
  { id_UL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD, "id-UL-DPCH-InformationModify-DeleteListIE-RL-ReconfPrepTDD" },
  { id_UL_DPCH_InformationModify_ModifyListIE_RL_ReconfPrepTDD, "id-UL-DPCH-InformationModify-ModifyListIE-RL-ReconfPrepTDD" },
  { id_Unsuccessful_PDSCHSetItem_PSCH_ReconfFailureTDD, "id-Unsuccessful-PDSCHSetItem-PSCH-ReconfFailureTDD" },
  { id_Unsuccessful_PUSCHSetItem_PSCH_ReconfFailureTDD, "id-Unsuccessful-PUSCHSetItem-PSCH-ReconfFailureTDD" },
  { id_Unknown_410, "id-Unknown-410" },
  { id_Unknown_411, "id-Unknown-411" },
  { id_CommunicationContextInfoItem_Reset, "id-CommunicationContextInfoItem-Reset" },
  { id_Unknown_413, "id-Unknown-413" },
  { id_CommunicationControlPortInfoItem_Reset, "id-CommunicationControlPortInfoItem-Reset" },
  { id_Unknown_415, "id-Unknown-415" },
  { id_ResetIndicator, "id-ResetIndicator" },
  { id_Unused_ProtocolIE_ID_417, "id-Unused-ProtocolIE-ID-417" },
  { id_Unused_ProtocolIE_ID_418, "id-Unused-ProtocolIE-ID-418" },
  { id_Unused_ProtocolIE_ID_419, "id-Unused-ProtocolIE-ID-419" },
  { id_CauseLevel_SyncAdjustmntFailureTDD, "id-CauseLevel-SyncAdjustmntFailureTDD" },
  { id_CellAdjustmentInfo_SyncAdjustmntRqstTDD, "id-CellAdjustmentInfo-SyncAdjustmntRqstTDD" },
  { id_CellSyncBurstTransInit_CellSyncInitiationRqstTDD, "id-CellSyncBurstTransInit-CellSyncInitiationRqstTDD" },
  { id_CellSyncBurstMeasureInit_CellSyncInitiationRqstTDD, "id-CellSyncBurstMeasureInit-CellSyncInitiationRqstTDD" },
  { id_CellSyncBurstTransReconfiguration_CellSyncReconfRqstTDD, "id-CellSyncBurstTransReconfiguration-CellSyncReconfRqstTDD" },
  { id_CellSyncBurstMeasReconfiguration_CellSyncReconfRqstTDD, "id-CellSyncBurstMeasReconfiguration-CellSyncReconfRqstTDD" },
  { id_CellSyncBurstTransInfoList_CellSyncReconfRqstTDD, "id-CellSyncBurstTransInfoList-CellSyncReconfRqstTDD" },
  { id_CellSyncBurstMeasInfoList_CellSyncReconfRqstTDD, "id-CellSyncBurstMeasInfoList-CellSyncReconfRqstTDD" },
  { id_CellSyncBurstTransReconfInfo_CellSyncReconfRqstTDD, "id-CellSyncBurstTransReconfInfo-CellSyncReconfRqstTDD" },
  { id_CellSyncInfo_CellSyncReprtTDD, "id-CellSyncInfo-CellSyncReprtTDD" },
  { id_CSBTransmissionID, "id-CSBTransmissionID" },
  { id_CSBMeasurementID, "id-CSBMeasurementID" },
  { id_IntStdPhCellSyncInfoItem_CellSyncReprtTDD, "id-IntStdPhCellSyncInfoItem-CellSyncReprtTDD" },
  { id_NCyclesPerSFNperiod, "id-NCyclesPerSFNperiod" },
  { id_NRepetitionsPerCyclePeriod, "id-NRepetitionsPerCyclePeriod" },
  { id_ReferenceClockAvailability, "id-ReferenceClockAvailability" },
  { id_ReferenceSFNoffset, "id-ReferenceSFNoffset" },
  { id_SyncFrameNumber, "id-SyncFrameNumber" },
  { id_SynchronisationReportType, "id-SynchronisationReportType" },
  { id_SynchronisationReportCharacteristics, "id-SynchronisationReportCharacteristics" },
  { id_Unsuccessful_cell_InformationRespItem_SyncAdjustmntFailureTDD, "id-Unsuccessful-cell-InformationRespItem-SyncAdjustmntFailureTDD" },
  { id_Unknown_441, "id-Unknown-441" },
  { id_Unknown_442, "id-Unknown-442" },
  { id_Unused_ProtocolIE_ID_443, "id-Unused-ProtocolIE-ID-443" },
  { id_InformationExchangeID, "id-InformationExchangeID" },
  { id_InformationExchangeObjectType_InfEx_Rqst, "id-InformationExchangeObjectType-InfEx-Rqst" },
  { id_InformationType, "id-InformationType" },
  { id_InformationReportCharacteristics, "id-InformationReportCharacteristics" },
  { id_InformationExchangeObjectType_InfEx_Rsp, "id-InformationExchangeObjectType-InfEx-Rsp" },
  { id_InformationExchangeObjectType_InfEx_Rprt, "id-InformationExchangeObjectType-InfEx-Rprt" },
  { id_DPC_Mode, "id-DPC-Mode" },
  { id_IPDLParameter_Information_Cell_ReconfRqstFDD, "id-IPDLParameter-Information-Cell-ReconfRqstFDD" },
  { id_IPDLParameter_Information_Cell_SetupRqstFDD, "id-IPDLParameter-Information-Cell-SetupRqstFDD" },
  { id_IPDLParameter_Information_Cell_ReconfRqstTDD, "id-IPDLParameter-Information-Cell-ReconfRqstTDD" },
  { id_IPDLParameter_Information_Cell_SetupRqstTDD, "id-IPDLParameter-Information-Cell-SetupRqstTDD" },
  { id_NeighbouringCellMeasurementInformation, "id-NeighbouringCellMeasurementInformation" },
  { id_PCCPCH_LCR_Information_Cell_SetupRqstTDD, "id-PCCPCH-LCR-Information-Cell-SetupRqstTDD" },
  { id_PCH_Power_LCR_CTCH_SetupRqstTDD, "id-PCH-Power-LCR-CTCH-SetupRqstTDD" },
  { id_PCH_Power_LCR_CTCH_ReconfRqstTDD, "id-PCH-Power-LCR-CTCH-ReconfRqstTDD" },
  { id_PICH_LCR_Parameters_CTCH_SetupRqstTDD, "id-PICH-LCR-Parameters-CTCH-SetupRqstTDD" },
  { id_Unknown_460, "id-Unknown-460" },
  { id_PRACH_LCR_ParametersList_CTCH_SetupRqstTDD, "id-PRACH-LCR-ParametersList-CTCH-SetupRqstTDD" },
  { id_Unknown_462, "id-Unknown-462" },
  { id_RL_InformationResponse_LCR_RL_SetupRspTDD, "id-RL-InformationResponse-LCR-RL-SetupRspTDD" },
  { id_Unknown_464, "id-Unknown-464" },
  { id_Secondary_CCPCH_LCR_parameterList_CTCH_SetupRqstTDD, "id-Secondary-CCPCH-LCR-parameterList-CTCH-SetupRqstTDD" },
  { id_TimeSlotConfigurationList_LCR_Cell_ReconfRqstTDD, "id-TimeSlotConfigurationList-LCR-Cell-ReconfRqstTDD" },
  { id_TimeSlotConfigurationList_LCR_Cell_SetupRqstTDD, "id-TimeSlotConfigurationList-LCR-Cell-SetupRqstTDD" },
  { id_TimeslotISCP_LCR_InfoList_RL_SetupRqstTDD, "id-TimeslotISCP-LCR-InfoList-RL-SetupRqstTDD" },
  { id_TimeSlotLCR_CM_Rqst, "id-TimeSlotLCR-CM-Rqst" },
  { id_UL_DPCH_LCR_Information_RL_SetupRqstTDD, "id-UL-DPCH-LCR-Information-RL-SetupRqstTDD" },
  { id_Unknown_471, "id-Unknown-471" },
  { id_DL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD, "id-DL-DPCH-InformationItem-LCR-RL-AdditionRqstTDD" },
  { id_UL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD, "id-UL-DPCH-InformationItem-LCR-RL-AdditionRqstTDD" },
  { id_TimeslotISCP_InformationList_LCR_RL_AdditionRqstTDD, "id-TimeslotISCP-InformationList-LCR-RL-AdditionRqstTDD" },
  { id_DL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD, "id-DL-DPCH-LCR-InformationAddList-RL-ReconfPrepTDD" },
  { id_Unknown_476, "id-Unknown-476" },
  { id_DL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD, "id-DL-DPCH-LCR-InformationModify-AddList-RL-ReconfPrepTDD" },
  { id_Unknown_478, "id-Unknown-478" },
  { id_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD, "id-DL-Timeslot-LCR-InformationModify-ModifyList-RL-ReconfPrepTDD" },
  { id_TimeslotISCPInfoList_LCR_DL_PC_RqstTDD, "id-TimeslotISCPInfoList-LCR-DL-PC-RqstTDD" },
  { id_UL_DPCH_LCR_InformationAddListIE_RL_ReconfPrepTDD, "id-UL-DPCH-LCR-InformationAddListIE-RL-ReconfPrepTDD" },
  { id_CellSyncBurstInfoList_CellSyncReconfRqstTDD, "id-CellSyncBurstInfoList-CellSyncReconfRqstTDD" },
  { id_UL_DPCH_LCR_InformationModify_AddList, "id-UL-DPCH-LCR-InformationModify-AddList" },
  { id_Unknown_484, "id-Unknown-484" },
  { id_UL_TimeslotLCR_Information_RL_ReconfPrepTDD, "id-UL-TimeslotLCR-Information-RL-ReconfPrepTDD" },
  { id_PDSCH_AddInformation_LCR_PSCH_ReconfRqst, "id-PDSCH-AddInformation-LCR-PSCH-ReconfRqst" },
  { id_PDSCH_AddInformation_LCR_AddListIE_PSCH_ReconfRqst, "id-PDSCH-AddInformation-LCR-AddListIE-PSCH-ReconfRqst" },
  { id_PDSCH_ModifyInformation_LCR_PSCH_ReconfRqst, "id-PDSCH-ModifyInformation-LCR-PSCH-ReconfRqst" },
  { id_PDSCH_ModifyInformation_LCR_ModifyListIE_PSCH_ReconfRqst, "id-PDSCH-ModifyInformation-LCR-ModifyListIE-PSCH-ReconfRqst" },
  { id_PUSCH_AddInformation_LCR_PSCH_ReconfRqst, "id-PUSCH-AddInformation-LCR-PSCH-ReconfRqst" },
  { id_PUSCH_AddInformation_LCR_AddListIE_PSCH_ReconfRqst, "id-PUSCH-AddInformation-LCR-AddListIE-PSCH-ReconfRqst" },
  { id_PUSCH_ModifyInformation_LCR_PSCH_ReconfRqst, "id-PUSCH-ModifyInformation-LCR-PSCH-ReconfRqst" },
  { id_PUSCH_ModifyInformation_LCR_ModifyListIE_PSCH_ReconfRqst, "id-PUSCH-ModifyInformation-LCR-ModifyListIE-PSCH-ReconfRqst" },
  { id_CellAdjustmentInfoItem_SyncAdjustmentRqstTDD, "id-CellAdjustmentInfoItem-SyncAdjustmentRqstTDD" },
  { id_TimeSlot, "id-TimeSlot" },
  { id_timeslotInfo_CellSyncInitiationRqstTDD, "id-timeslotInfo-CellSyncInitiationRqstTDD" },
  { id_SyncReportType_CellSyncReprtTDD, "id-SyncReportType-CellSyncReprtTDD" },
  { id_Power_Local_Cell_Group_InformationItem_AuditRsp, "id-Power-Local-Cell-Group-InformationItem-AuditRsp" },
  { id_Power_Local_Cell_Group_InformationItem_ResourceStatusInd, "id-Power-Local-Cell-Group-InformationItem-ResourceStatusInd" },
  { id_Power_Local_Cell_Group_InformationItem2_ResourceStatusInd, "id-Power-Local-Cell-Group-InformationItem2-ResourceStatusInd" },
  { id_Power_Local_Cell_Group_InformationList_AuditRsp, "id-Power-Local-Cell-Group-InformationList-AuditRsp" },
  { id_Power_Local_Cell_Group_InformationList_ResourceStatusInd, "id-Power-Local-Cell-Group-InformationList-ResourceStatusInd" },
  { id_Power_Local_Cell_Group_InformationList2_ResourceStatusInd, "id-Power-Local-Cell-Group-InformationList2-ResourceStatusInd" },
  { id_Power_Local_Cell_Group_ID, "id-Power-Local-Cell-Group-ID" },
  { id_PUSCH_Info_DM_Rqst, "id-PUSCH-Info-DM-Rqst" },
  { id_PUSCH_Info_DM_Rsp, "id-PUSCH-Info-DM-Rsp" },
  { id_PUSCH_Info_DM_Rprt, "id-PUSCH-Info-DM-Rprt" },
  { id_TypeOfError, "id-TypeOfError" },
  { id_InitDL_Power, "id-InitDL-Power" },
  { id_UL_SIRTarget, "id-UL-SIRTarget" },
  { id_cellSyncBurstRepetitionPeriod, "id-cellSyncBurstRepetitionPeriod" },
  { id_ReportCharacteristicsType_OnModification, "id-ReportCharacteristicsType-OnModification" },
  { id_SFNSFNMeasurementValueInformation, "id-SFNSFNMeasurementValueInformation" },
  { id_SFNSFNMeasurementThresholdInformation, "id-SFNSFNMeasurementThresholdInformation" },
  { id_TUTRANGPSMeasurementValueInformation, "id-TUTRANGPSMeasurementValueInformation" },
  { id_TUTRANGPSMeasurementThresholdInformation, "id-TUTRANGPSMeasurementThresholdInformation" },
  { id_CCTrCH_Initial_DL_Power_RL_SetupRqstTDD, "id-CCTrCH-Initial-DL-Power-RL-SetupRqstTDD" },
  { id_CCTrCH_Initial_DL_Power_RL_AdditionRqstTDD, "id-CCTrCH-Initial-DL-Power-RL-AdditionRqstTDD" },
  { id_CCTrCH_Initial_DL_Power_RL_ReconfPrepTDD, "id-CCTrCH-Initial-DL-Power-RL-ReconfPrepTDD" },
  { id_Rx_Timing_Deviation_Value_LCR, "id-Rx-Timing-Deviation-Value-LCR" },
  { id_Angle_Of_Arrival_Value_LCR, "id-Angle-Of-Arrival-Value-LCR" },
  { id_HS_PDSCH_HS_SCCH_E_AGCH_E_RGCH_E_HICH_MaxPower_PSCH_ReconfRqst, "id-HS-PDSCH-HS-SCCH-E-AGCH-E-RGCH-E-HICH-MaxPower-PSCH-ReconfRqst" },
  { id_HS_PDSCH_HS_SCCH_ScramblingCode_PSCH_ReconfRqst, "id-HS-PDSCH-HS-SCCH-ScramblingCode-PSCH-ReconfRqst" },
  { id_HS_PDSCH_FDD_Code_Information_PSCH_ReconfRqst, "id-HS-PDSCH-FDD-Code-Information-PSCH-ReconfRqst" },
  { id_HS_SCCH_FDD_Code_Information_PSCH_ReconfRqst, "id-HS-SCCH-FDD-Code-Information-PSCH-ReconfRqst" },
  { id_HS_PDSCH_TDD_Information_PSCH_ReconfRqst, "id-HS-PDSCH-TDD-Information-PSCH-ReconfRqst" },
  { id_Add_To_HS_SCCH_Resource_Pool_PSCH_ReconfRqst, "id-Add-To-HS-SCCH-Resource-Pool-PSCH-ReconfRqst" },
  { id_Modify_HS_SCCH_Resource_Pool_PSCH_ReconfRqst, "id-Modify-HS-SCCH-Resource-Pool-PSCH-ReconfRqst" },
  { id_Delete_From_HS_SCCH_Resource_Pool_PSCH_ReconfRqst, "id-Delete-From-HS-SCCH-Resource-Pool-PSCH-ReconfRqst" },
  { id_HSDSCH_FDD_Information, "id-HSDSCH-FDD-Information" },
  { id_HSDSCH_FDD_Information_Response, "id-HSDSCH-FDD-Information-Response" },
  { id_Unknown_532, "id-Unknown-532" },
  { id_Unknown_533, "id-Unknown-533" },
  { id_HSDSCH_Information_to_Modify, "id-HSDSCH-Information-to-Modify" },
  { id_HSDSCH_RNTI, "id-HSDSCH-RNTI" },
  { id_HSDSCH_TDD_Information, "id-HSDSCH-TDD-Information" },
  { id_HSDSCH_TDD_Information_Response, "id-HSDSCH-TDD-Information-Response" },
  { id_Unknown_538, "id-Unknown-538" },
  { id_Unknown_539, "id-Unknown-539" },
  { id_Unknown_540, "id-Unknown-540" },
  { id_HSPDSCH_RL_ID, "id-HSPDSCH-RL-ID" },
  { id_PrimCCPCH_RSCP_DL_PC_RqstTDD, "id-PrimCCPCH-RSCP-DL-PC-RqstTDD" },
  { id_SYNCDlCodeId_TransInitLCR_CellSyncInitiationRqstTDD, "id-SYNCDlCodeId-TransInitLCR-CellSyncInitiationRqstTDD" },
  { id_SYNCDlCodeId_MeasureInitLCR_CellSyncInitiationRqstTDD, "id-SYNCDlCodeId-MeasureInitLCR-CellSyncInitiationRqstTDD" },
  { id_SYNCDlCodeIdTransReconfInfoLCR_CellSyncReconfRqstTDD, "id-SYNCDlCodeIdTransReconfInfoLCR-CellSyncReconfRqstTDD" },
  { id_SYNCDlCodeIdMeasReconfigurationLCR_CellSyncReconfRqstTDD, "id-SYNCDlCodeIdMeasReconfigurationLCR-CellSyncReconfRqstTDD" },
  { id_SYNCDlCodeIdMeasInfoList_CellSyncReconfRqstTDD, "id-SYNCDlCodeIdMeasInfoList-CellSyncReconfRqstTDD" },
  { id_SyncDLCodeIdsMeasInfoList_CellSyncReprtTDD, "id-SyncDLCodeIdsMeasInfoList-CellSyncReprtTDD" },
  { id_SyncDLCodeIdThreInfoLCR, "id-SyncDLCodeIdThreInfoLCR" },
  { id_NSubCyclesPerCyclePeriod_CellSyncReconfRqstTDD, "id-NSubCyclesPerCyclePeriod-CellSyncReconfRqstTDD" },
  { id_DwPCH_Power, "id-DwPCH-Power" },
  { id_AccumulatedClockupdate_CellSyncReprtTDD, "id-AccumulatedClockupdate-CellSyncReprtTDD" },
  { id_HSDSCH_RearrangeList_Bearer_RearrangeInd, "id-HSDSCH-RearrangeList-Bearer-RearrangeInd" },
  { id_UL_Synchronisation_Parameters_LCR, "id-UL-Synchronisation-Parameters-LCR" },
  { id_HSDSCH_FDD_Update_Information, "id-HSDSCH-FDD-Update-Information" },
  { id_HSDSCH_TDD_Update_Information, "id-HSDSCH-TDD-Update-Information" },
  { id_Unknown_557, "id-Unknown-557" },
  { id_DL_DPCH_TimeSlotFormat_LCR_ModifyItem_RL_ReconfPrepTDD, "id-DL-DPCH-TimeSlotFormat-LCR-ModifyItem-RL-ReconfPrepTDD" },
  { id_UL_DPCH_TimeSlotFormat_LCR_ModifyItem_RL_ReconfPrepTDD, "id-UL-DPCH-TimeSlotFormat-LCR-ModifyItem-RL-ReconfPrepTDD" },
  { id_TDD_TPC_UplinkStepSize_LCR_RL_SetupRqstTDD, "id-TDD-TPC-UplinkStepSize-LCR-RL-SetupRqstTDD" },
  { id_TDD_TPC_UplinkStepSize_LCR_RL_AdditionRqstTDD, "id-TDD-TPC-UplinkStepSize-LCR-RL-AdditionRqstTDD" },
  { id_TDD_TPC_DownlinkStepSize_RL_AdditionRqstTDD, "id-TDD-TPC-DownlinkStepSize-RL-AdditionRqstTDD" },
  { id_TDD_TPC_UplinkStepSize_InformationAdd_LCR_RL_ReconfPrepTDD, "id-TDD-TPC-UplinkStepSize-InformationAdd-LCR-RL-ReconfPrepTDD" },
  { id_TDD_TPC_UplinkStepSize_InformationModify_LCR_RL_ReconfPrepTDD, "id-TDD-TPC-UplinkStepSize-InformationModify-LCR-RL-ReconfPrepTDD" },
  { id_TDD_TPC_DownlinkStepSize_InformationModify_RL_ReconfPrepTDD, "id-TDD-TPC-DownlinkStepSize-InformationModify-RL-ReconfPrepTDD" },
  { id_TDD_TPC_DownlinkStepSize_InformationAdd_RL_ReconfPrepTDD, "id-TDD-TPC-DownlinkStepSize-InformationAdd-RL-ReconfPrepTDD" },
  { id_CCTrCH_Maximum_DL_Power_RL_SetupRqstTDD, "id-CCTrCH-Maximum-DL-Power-RL-SetupRqstTDD" },
  { id_CCTrCH_Minimum_DL_Power_RL_SetupRqstTDD, "id-CCTrCH-Minimum-DL-Power-RL-SetupRqstTDD" },
  { id_CCTrCH_Maximum_DL_Power_RL_AdditionRqstTDD, "id-CCTrCH-Maximum-DL-Power-RL-AdditionRqstTDD" },
  { id_CCTrCH_Minimum_DL_Power_RL_AdditionRqstTDD, "id-CCTrCH-Minimum-DL-Power-RL-AdditionRqstTDD" },
  { id_CCTrCH_Maximum_DL_Power_InformationAdd_RL_ReconfPrepTDD, "id-CCTrCH-Maximum-DL-Power-InformationAdd-RL-ReconfPrepTDD" },
  { id_CCTrCH_Minimum_DL_Power_InformationAdd_RL_ReconfPrepTDD, "id-CCTrCH-Minimum-DL-Power-InformationAdd-RL-ReconfPrepTDD" },
  { id_CCTrCH_Maximum_DL_Power_InformationModify_RL_ReconfPrepTDD, "id-CCTrCH-Maximum-DL-Power-InformationModify-RL-ReconfPrepTDD" },
  { id_CCTrCH_Minimum_DL_Power_InformationModify_RL_ReconfPrepTDD, "id-CCTrCH-Minimum-DL-Power-InformationModify-RL-ReconfPrepTDD" },
  { id_Maximum_DL_Power_Modify_LCR_InformationModify_RL_ReconfPrepTDD, "id-Maximum-DL-Power-Modify-LCR-InformationModify-RL-ReconfPrepTDD" },
  { id_Minimum_DL_Power_Modify_LCR_InformationModify_RL_ReconfPrepTDD, "id-Minimum-DL-Power-Modify-LCR-InformationModify-RL-ReconfPrepTDD" },
  { id_DL_DPCH_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD, "id-DL-DPCH-LCR-InformationModify-ModifyList-RL-ReconfRqstTDD" },
  { id_CCTrCH_Maximum_DL_Power_InformationModify_RL_ReconfRqstTDD, "id-CCTrCH-Maximum-DL-Power-InformationModify-RL-ReconfRqstTDD" },
  { id_CCTrCH_Minimum_DL_Power_InformationModify_RL_ReconfRqstTDD, "id-CCTrCH-Minimum-DL-Power-InformationModify-RL-ReconfRqstTDD" },
  { id_Initial_DL_Power_TimeslotLCR_InformationItem, "id-Initial-DL-Power-TimeslotLCR-InformationItem" },
  { id_Maximum_DL_Power_TimeslotLCR_InformationItem, "id-Maximum-DL-Power-TimeslotLCR-InformationItem" },
  { id_Minimum_DL_Power_TimeslotLCR_InformationItem, "id-Minimum-DL-Power-TimeslotLCR-InformationItem" },
  { id_HS_DSCHProvidedBitRateValueInformation, "id-HS-DSCHProvidedBitRateValueInformation" },
  { id_Unknown_584, "id-Unknown-584" },
  { id_HS_DSCHRequiredPowerValueInformation, "id-HS-DSCHRequiredPowerValueInformation" },
  { id_HS_DSCHRequiredPowerValue, "id-HS-DSCHRequiredPowerValue" },
  { id_TransmittedCarrierPowerOfAllCodesNotUsedForHSTransmission, "id-TransmittedCarrierPowerOfAllCodesNotUsedForHSTransmission" },
  { id_HS_SICH_Reception_Quality, "id-HS-SICH-Reception-Quality" },
  { id_HS_SICH_Reception_Quality_Measurement_Value, "id-HS-SICH-Reception-Quality-Measurement-Value" },
  { id_HSSICH_Info_DM_Rprt, "id-HSSICH-Info-DM-Rprt" },
  { id_HSSICH_Info_DM_Rqst, "id-HSSICH-Info-DM-Rqst" },
  { id_HSSICH_Info_DM_Rsp, "id-HSSICH-Info-DM-Rsp" },
  { id_Best_Cell_Portions_Value, "id-Best-Cell-Portions-Value" },
  { id_Primary_CPICH_Usage_for_Channel_Estimation, "id-Primary-CPICH-Usage-for-Channel-Estimation" },
  { id_Secondary_CPICH_Information_Change, "id-Secondary-CPICH-Information-Change" },
  { id_NumberOfReportedCellPortions, "id-NumberOfReportedCellPortions" },
  { id_CellPortion_InformationItem_Cell_SetupRqstFDD, "id-CellPortion-InformationItem-Cell-SetupRqstFDD" },
  { id_CellPortion_InformationList_Cell_SetupRqstFDD, "id-CellPortion-InformationList-Cell-SetupRqstFDD" },
  { id_TimeslotISCP_LCR_InfoList_RL_ReconfPrepTDD, "id-TimeslotISCP-LCR-InfoList-RL-ReconfPrepTDD" },
  { id_Secondary_CPICH_Information, "id-Secondary-CPICH-Information" },
  { id_Received_total_wide_band_power_For_CellPortion, "id-Received-total-wide-band-power-For-CellPortion" },
  { id_Unidirectional_DCH_Indicator, "id-Unidirectional-DCH-Indicator" },
  { id_TimingAdjustmentValueLCR, "id-TimingAdjustmentValueLCR" },
  { id_multipleRL_dl_DPCH_InformationList, "id-multipleRL-dl-DPCH-InformationList" },
  { id_multipleRL_dl_DPCH_InformationModifyList, "id-multipleRL-dl-DPCH-InformationModifyList" },
  { id_multipleRL_ul_DPCH_InformationList, "id-multipleRL-ul-DPCH-InformationList" },
  { id_multipleRL_ul_DPCH_InformationModifyList, "id-multipleRL-ul-DPCH-InformationModifyList" },
  { id_RL_ID, "id-RL-ID" },
  { id_SAT_Info_Almanac_ExtItem, "id-SAT-Info-Almanac-ExtItem" },
  { id_HSDPA_Capability, "id-HSDPA-Capability" },
  { id_HSDSCH_Resources_Information_AuditRsp, "id-HSDSCH-Resources-Information-AuditRsp" },
  { id_HSDSCH_Resources_Information_ResourceStatusInd, "id-HSDSCH-Resources-Information-ResourceStatusInd" },
  { id_HSDSCH_MACdFlows_to_Add, "id-HSDSCH-MACdFlows-to-Add" },
  { id_HSDSCH_MACdFlows_to_Delete, "id-HSDSCH-MACdFlows-to-Delete" },
  { id_HSDSCH_Information_to_Modify_Unsynchronised, "id-HSDSCH-Information-to-Modify-Unsynchronised" },
  { id_TnlQos, "id-TnlQos" },
  { id_Received_total_wide_band_power_For_CellPortion_Value, "id-Received-total-wide-band-power-For-CellPortion-Value" },
  { id_Transmitted_Carrier_Power_For_CellPortion, "id-Transmitted-Carrier-Power-For-CellPortion" },
  { id_Transmitted_Carrier_Power_For_CellPortion_Value, "id-Transmitted-Carrier-Power-For-CellPortion-Value" },
  { id_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortion, "id-TransmittedCarrierPowerOfAllCodesNotUsedForHS-PDSCH-HS-SCCH-E-AGCH-E-RGCHOrE-HICHTransmissionCellPortion" },
  { id_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue, "id-TransmittedCarrierPowerOfAllCodesNotUsedForHS-PDSCH-HS-SCCH-E-AGCH-E-RGCHOrE-HICHTransmissionCellPortionValue" },
  { id_UpPTSInterferenceValue, "id-UpPTSInterferenceValue" },
  { id_PrimaryCCPCH_RSCP_Delta, "id-PrimaryCCPCH-RSCP-Delta" },
  { id_MeasurementRecoveryBehavior, "id-MeasurementRecoveryBehavior" },
  { id_MeasurementRecoveryReportingIndicator, "id-MeasurementRecoveryReportingIndicator" },
  { id_MeasurementRecoverySupportIndicator, "id-MeasurementRecoverySupportIndicator" },
  { id_Tstd_indicator, "id-Tstd-indicator" },
  { id_multiple_RL_Information_RL_ReconfPrepTDD, "id-multiple-RL-Information-RL-ReconfPrepTDD" },
  { id_multiple_RL_Information_RL_ReconfRqstTDD, "id-multiple-RL-Information-RL-ReconfRqstTDD" },
  { id_DL_DPCH_Power_Information_RL_ReconfPrepFDD, "id-DL-DPCH-Power-Information-RL-ReconfPrepFDD" },
  { id_F_DPCH_Information_RL_ReconfPrepFDD, "id-F-DPCH-Information-RL-ReconfPrepFDD" },
  { id_F_DPCH_Information_RL_SetupRqstFDD, "id-F-DPCH-Information-RL-SetupRqstFDD" },
  { id_Additional_S_CCPCH_Parameters_CTCH_ReconfRqstTDD, "id-Additional-S-CCPCH-Parameters-CTCH-ReconfRqstTDD" },
  { id_Additional_S_CCPCH_Parameters_CTCH_SetupRqstTDD, "id-Additional-S-CCPCH-Parameters-CTCH-SetupRqstTDD" },
  { id_Additional_S_CCPCH_LCR_Parameters_CTCH_ReconfRqstTDD, "id-Additional-S-CCPCH-LCR-Parameters-CTCH-ReconfRqstTDD" },
  { id_Additional_S_CCPCH_LCR_Parameters_CTCH_SetupRqstTDD, "id-Additional-S-CCPCH-LCR-Parameters-CTCH-SetupRqstTDD" },
  { id_MICH_CFN, "id-MICH-CFN" },
  { id_MICH_Information_AuditRsp, "id-MICH-Information-AuditRsp" },
  { id_MICH_Information_ResourceStatusInd, "id-MICH-Information-ResourceStatusInd" },
  { id_MICH_Parameters_CTCH_ReconfRqstFDD, "id-MICH-Parameters-CTCH-ReconfRqstFDD" },
  { id_MICH_Parameters_CTCH_ReconfRqstTDD, "id-MICH-Parameters-CTCH-ReconfRqstTDD" },
  { id_MICH_Parameters_CTCH_SetupRqstFDD, "id-MICH-Parameters-CTCH-SetupRqstFDD" },
  { id_MICH_Parameters_CTCH_SetupRqstTDD, "id-MICH-Parameters-CTCH-SetupRqstTDD" },
  { id_Modification_Period, "id-Modification-Period" },
  { id_NI_Information_NotifUpdateCmd, "id-NI-Information-NotifUpdateCmd" },
  { id_S_CCPCH_InformationListExt_AuditRsp, "id-S-CCPCH-InformationListExt-AuditRsp" },
  { id_S_CCPCH_InformationListExt_ResourceStatusInd, "id-S-CCPCH-InformationListExt-ResourceStatusInd" },
  { id_S_CCPCH_LCR_InformationListExt_AuditRsp, "id-S-CCPCH-LCR-InformationListExt-AuditRsp" },
  { id_S_CCPCH_LCR_InformationListExt_ResourceStatusInd, "id-S-CCPCH-LCR-InformationListExt-ResourceStatusInd" },
  { id_HARQ_Preamble_Mode, "id-HARQ-Preamble-Mode" },
  { id_Initial_DL_DPCH_TimingAdjustment, "id-Initial-DL-DPCH-TimingAdjustment" },
  { id_Initial_DL_DPCH_TimingAdjustment_Allowed, "id-Initial-DL-DPCH-TimingAdjustment-Allowed" },
  { id_DLTransmissionBranchLoadValue, "id-DLTransmissionBranchLoadValue" },
  { id_Power_Local_Cell_Group_choice_CM_Rqst, "id-Power-Local-Cell-Group-choice-CM-Rqst" },
  { id_Power_Local_Cell_Group_choice_CM_Rsp, "id-Power-Local-Cell-Group-choice-CM-Rsp" },
  { id_Power_Local_Cell_Group_choice_CM_Rprt, "id-Power-Local-Cell-Group-choice-CM-Rprt" },
  { id_SynchronisationIndicator, "id-SynchronisationIndicator" },
  { id_HSDPA_And_EDCH_CellPortion_Information_PSCH_ReconfRqst, "id-HSDPA-And-EDCH-CellPortion-Information-PSCH-ReconfRqst" },
  { id_Unused_ProtocolIE_ID_659, "id-Unused-ProtocolIE-ID-659" },
  { id_HS_DSCHRequiredPowerValue_For_Cell_Portion, "id-HS-DSCHRequiredPowerValue-For-Cell-Portion" },
  { id_HS_DSCHRequiredPowerValueInformation_For_CellPortion, "id-HS-DSCHRequiredPowerValueInformation-For-CellPortion" },
  { id_HS_DSCHProvidedBitRateValueInformation_For_CellPortion, "id-HS-DSCHProvidedBitRateValueInformation-For-CellPortion" },
  { id_E_AGCH_And_E_RGCH_E_HICH_FDD_Scrambling_Code, "id-E-AGCH-And-E-RGCH-E-HICH-FDD-Scrambling-Code" },
  { id_E_AGCH_FDD_Code_Information, "id-E-AGCH-FDD-Code-Information" },
  { id_E_DCH_Capability, "id-E-DCH-Capability" },
  { id_E_DCH_FDD_DL_Control_Channel_Information, "id-E-DCH-FDD-DL-Control-Channel-Information" },
  { id_E_DCH_FDD_Information, "id-E-DCH-FDD-Information" },
  { id_E_DCH_FDD_Information_Response, "id-E-DCH-FDD-Information-Response" },
  { id_E_DCH_FDD_Information_to_Modify, "id-E-DCH-FDD-Information-to-Modify" },
  { id_E_DCH_MACdFlows_to_Add, "id-E-DCH-MACdFlows-to-Add" },
  { id_E_DCH_MACdFlows_to_Delete, "id-E-DCH-MACdFlows-to-Delete" },
  { id_E_DCH_Resources_Information_AuditRsp, "id-E-DCH-Resources-Information-AuditRsp" },
  { id_E_DCH_Resources_Information_ResourceStatusInd, "id-E-DCH-Resources-Information-ResourceStatusInd" },
  { id_E_DCH_RL_Indication, "id-E-DCH-RL-Indication" },
  { id_E_DCH_RL_Set_ID, "id-E-DCH-RL-Set-ID" },
  { id_E_DPCH_Information_RL_ReconfPrepFDD, "id-E-DPCH-Information-RL-ReconfPrepFDD" },
  { id_E_DPCH_Information_RL_SetupRqstFDD, "id-E-DPCH-Information-RL-SetupRqstFDD" },
  { id_E_RGCH_E_HICH_FDD_Code_Information, "id-E-RGCH-E-HICH-FDD-Code-Information" },
  { id_Serving_E_DCH_RL_ID, "id-Serving-E-DCH-RL-ID" },
  { id_UL_DPDCH_Indicator_For_E_DCH_Operation, "id-UL-DPDCH-Indicator-For-E-DCH-Operation" },
  { id_FDD_S_CCPCH_FrameOffset_CTCH_SetupRqstFDD, "id-FDD-S-CCPCH-FrameOffset-CTCH-SetupRqstFDD" },
  { id_E_DPCH_Information_RL_ReconfRqstFDD, "id-E-DPCH-Information-RL-ReconfRqstFDD" },
  { id_Maximum_Target_ReceivedTotalWideBandPower, "id-Maximum-Target-ReceivedTotalWideBandPower" },
  { id_E_DCHProvidedBitRateValueInformation, "id-E-DCHProvidedBitRateValueInformation" },
  { id_HARQ_Preamble_Mode_Activation_Indicator, "id-HARQ-Preamble-Mode-Activation-Indicator" },
  { id_RL_Specific_E_DCH_Info, "id-RL-Specific-E-DCH-Info" },
  { id_E_DCH_CapacityConsumptionLaw, "id-E-DCH-CapacityConsumptionLaw" },
  { id_multiple_DedicatedMeasurementValueList_TDD_DM_Rsp, "id-multiple-DedicatedMeasurementValueList-TDD-DM-Rsp" },
  { id_multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp, "id-multiple-DedicatedMeasurementValueList-LCR-TDD-DM-Rsp" },
  { id_E_DCH_RearrangeList_Bearer_RearrangeInd, "id-E-DCH-RearrangeList-Bearer-RearrangeInd" },
  { id_Unused_ProtocolIE_ID_691, "id-Unused-ProtocolIE-ID-691" },
  { id_multipleRL_dl_CCTrCH_InformationModifyList_RL_ReconfRqstTDD, "id-multipleRL-dl-CCTrCH-InformationModifyList-RL-ReconfRqstTDD" },
  { id_Target_NonServing_EDCH_To_Total_EDCH_Power_Ratio, "id-Target-NonServing-EDCH-To-Total-EDCH-Power-Ratio" },
  { id_CellPortion_InformationItem_Cell_ReconfRqstFDD, "id-CellPortion-InformationItem-Cell-ReconfRqstFDD" },
  { id_CellPortion_InformationList_Cell_ReconfRqstFDD, "id-CellPortion-InformationList-Cell-ReconfRqstFDD" },
  { id_multiple_PUSCH_InfoList_DM_Rsp, "id-multiple-PUSCH-InfoList-DM-Rsp" },
  { id_multiple_PUSCH_InfoList_DM_Rprt, "id-multiple-PUSCH-InfoList-DM-Rprt" },
  { id_Reference_ReceivedTotalWideBandPower, "id-Reference-ReceivedTotalWideBandPower" },
  { id_E_DCH_Serving_Cell_Change_Info_Response, "id-E-DCH-Serving-Cell-Change-Info-Response" },
  { id_HS_DSCH_Serving_Cell_Change_Info, "id-HS-DSCH-Serving-Cell-Change-Info" },
  { id_HS_DSCH_Serving_Cell_Change_Info_Response, "id-HS-DSCH-Serving-Cell-Change-Info-Response" },
  { id_Serving_Cell_Change_CFN, "id-Serving-Cell-Change-CFN" },
  { id_E_DCH_HARQ_Combining_Capability, "id-E-DCH-HARQ-Combining-Capability" },
  { id_E_DCH_TTI2ms_Capability, "id-E-DCH-TTI2ms-Capability" },
  { id_E_DCH_SF_Capability, "id-E-DCH-SF-Capability" },
  { id_E_DCH_FDD_Update_Information, "id-E-DCH-FDD-Update-Information" },
  { id_F_DPCH_Capability, "id-F-DPCH-Capability" },
  { id_E_DCH_Non_serving_Relative_Grant_Down_CommandsValue, "id-E-DCH-Non-serving-Relative-Grant-Down-CommandsValue" },
  { id_HSSICH_SIRTarget, "id-HSSICH-SIRTarget" },
  { id_multiple_HSSICHMeasurementValueList_TDD_DM_Rsp, "id-multiple-HSSICHMeasurementValueList-TDD-DM-Rsp" },
  { id_PLCCH_Information_AuditRsp, "id-PLCCH-Information-AuditRsp" },
  { id_PLCCH_Information_ResourceStatusInd, "id-PLCCH-Information-ResourceStatusInd" },
  { id_PLCCH_Information_RL_ReconfPrepTDDLCR, "id-PLCCH-Information-RL-ReconfPrepTDDLCR" },
  { id_PLCCH_Information_UL_TimeslotLCR_Info, "id-PLCCH-Information-UL-TimeslotLCR-Info" },
  { id_PLCCH_InformationList_AuditRsp, "id-PLCCH-InformationList-AuditRsp" },
  { id_PLCCH_InformationList_ResourceStatusInd, "id-PLCCH-InformationList-ResourceStatusInd" },
  { id_PLCCH_Parameters_CTCH_ReconfRqstTDD, "id-PLCCH-Parameters-CTCH-ReconfRqstTDD" },
  { id_S_CCPCH_768_Parameters_CTCH_SetupRqstTDD, "id-S-CCPCH-768-Parameters-CTCH-SetupRqstTDD" },
  { id_PICH_768_Parameters_CTCH_SetupRqstTDD, "id-PICH-768-Parameters-CTCH-SetupRqstTDD" },
  { id_PRACH_768_Parameters_CTCH_SetupRqstTDD, "id-PRACH-768-Parameters-CTCH-SetupRqstTDD" },
  { id_S_CCPCH_768_Parameters_CTCH_ReconfRqstTDD, "id-S-CCPCH-768-Parameters-CTCH-ReconfRqstTDD" },
  { id_PICH_768_Parameters_CTCH_ReconfRqstTDD, "id-PICH-768-Parameters-CTCH-ReconfRqstTDD" },
  { id_MICH_768_Parameters_CTCH_ReconfRqstTDD, "id-MICH-768-Parameters-CTCH-ReconfRqstTDD" },
  { id_CommonPhysicalChannelID768_CommonTrChDeletionReq, "id-CommonPhysicalChannelID768-CommonTrChDeletionReq" },
  { id_S_CCPCH_768_InformationList_AuditRsp, "id-S-CCPCH-768-InformationList-AuditRsp" },
  { id_S_CCPCH_768_Information_AuditRsp, "id-S-CCPCH-768-Information-AuditRsp" },
  { id_neighbouringTDDCellMeasurementInformation768, "id-neighbouringTDDCellMeasurementInformation768" },
  { id_PCCPCH_768_Information_Cell_SetupRqstTDD, "id-PCCPCH-768-Information-Cell-SetupRqstTDD" },
  { id_SCH_768_Information_Cell_SetupRqstTDD, "id-SCH-768-Information-Cell-SetupRqstTDD" },
  { id_SCH_768_Information_Cell_ReconfRqstTDD, "id-SCH-768-Information-Cell-ReconfRqstTDD" },
  { id_PCCPCH_768_Information_Cell_ReconfRqstTDD, "id-PCCPCH-768-Information-Cell-ReconfRqstTDD" },
  { id_P_CCPCH_768_Information_AuditRsp, "id-P-CCPCH-768-Information-AuditRsp" },
  { id_PICH_768_Information_AuditRsp, "id-PICH-768-Information-AuditRsp" },
  { id_PRACH_768_InformationList_AuditRsp, "id-PRACH-768-InformationList-AuditRsp" },
  { id_SCH_768_Information_AuditRsp, "id-SCH-768-Information-AuditRsp" },
  { id_MICH_768_Information_AuditRsp, "id-MICH-768-Information-AuditRsp" },
  { id_PRACH_768_Information, "id-PRACH-768-Information" },
  { id_S_CCPCH_768_Information_ResourceStatusInd, "id-S-CCPCH-768-Information-ResourceStatusInd" },
  { id_P_CCPCH_768_Information_ResourceStatusInd, "id-P-CCPCH-768-Information-ResourceStatusInd" },
  { id_PICH_768_Information_ResourceStatusInd, "id-PICH-768-Information-ResourceStatusInd" },
  { id_PRACH_768_InformationList_ResourceStatusInd, "id-PRACH-768-InformationList-ResourceStatusInd" },
  { id_SCH_768_Information_ResourceStatusInd, "id-SCH-768-Information-ResourceStatusInd" },
  { id_MICH_768_Information_ResourceStatusInd, "id-MICH-768-Information-ResourceStatusInd" },
  { id_S_CCPCH_768_InformationList_ResourceStatusInd, "id-S-CCPCH-768-InformationList-ResourceStatusInd" },
  { id_UL_DPCH_768_Information_RL_SetupRqstTDD, "id-UL-DPCH-768-Information-RL-SetupRqstTDD" },
  { id_DL_DPCH_768_Information_RL_SetupRqstTDD, "id-DL-DPCH-768-Information-RL-SetupRqstTDD" },
  { id_DL_DPCH_InformationItem_768_RL_AdditionRqstTDD, "id-DL-DPCH-InformationItem-768-RL-AdditionRqstTDD" },
  { id_UL_DPCH_InformationItem_768_RL_AdditionRqstTDD, "id-UL-DPCH-InformationItem-768-RL-AdditionRqstTDD" },
  { id_UL_DPCH_768_InformationAddItemIE_RL_ReconfPrepTDD, "id-UL-DPCH-768-InformationAddItemIE-RL-ReconfPrepTDD" },
  { id_UL_DPCH_768_InformationAddListIE_RL_ReconfPrepTDD, "id-UL-DPCH-768-InformationAddListIE-RL-ReconfPrepTDD" },
  { id_UL_DPCH_768_InformationModify_AddItem, "id-UL-DPCH-768-InformationModify-AddItem" },
  { id_UL_DPCH_768_InformationModify_AddList, "id-UL-DPCH-768-InformationModify-AddList" },
  { id_UL_Timeslot768_Information_RL_ReconfPrepTDD, "id-UL-Timeslot768-Information-RL-ReconfPrepTDD" },
  { id_DL_DPCH_768_InformationAddItem_RL_ReconfPrepTDD, "id-DL-DPCH-768-InformationAddItem-RL-ReconfPrepTDD" },
  { id_DL_DPCH_768_InformationAddList_RL_ReconfPrepTDD, "id-DL-DPCH-768-InformationAddList-RL-ReconfPrepTDD" },
  { id_DL_DPCH_768_InformationModify_AddItem_RL_ReconfPrepTDD, "id-DL-DPCH-768-InformationModify-AddItem-RL-ReconfPrepTDD" },
  { id_DL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD, "id-DL-DPCH-768-InformationModify-AddList-RL-ReconfPrepTDD" },
  { id_DL_Timeslot_768_InformationModify_ModifyList_RL_ReconfPrepTDD, "id-DL-Timeslot-768-InformationModify-ModifyList-RL-ReconfPrepTDD" },
  { id_DPCH_ID768_DM_Rqst, "id-DPCH-ID768-DM-Rqst" },
  { id_multiple_DedicatedMeasurementValueList_768_TDD_DM_Rsp, "id-multiple-DedicatedMeasurementValueList-768-TDD-DM-Rsp" },
  { id_DPCH_ID768_DM_Rsp, "id-DPCH-ID768-DM-Rsp" },
  { id_Rx_Timing_Deviation_Value_768, "id-Rx-Timing-Deviation-Value-768" },
  { id_DPCH_ID768_DM_Rprt, "id-DPCH-ID768-DM-Rprt" },
  { id_PDSCH_AddInformation_768_PSCH_ReconfRqst, "id-PDSCH-AddInformation-768-PSCH-ReconfRqst" },
  { id_PDSCH_ModifyInformation_768_PSCH_ReconfRqst, "id-PDSCH-ModifyInformation-768-PSCH-ReconfRqst" },
  { id_PUSCH_AddInformation_768_PSCH_ReconfRqst, "id-PUSCH-AddInformation-768-PSCH-ReconfRqst" },
  { id_PUSCH_ModifyInformation_768_PSCH_ReconfRqst, "id-PUSCH-ModifyInformation-768-PSCH-ReconfRqst" },
  { id_dL_HS_PDSCH_Timeslot_Information_768_PSCH_ReconfRqst, "id-dL-HS-PDSCH-Timeslot-Information-768-PSCH-ReconfRqst" },
  { id_hS_SCCH_Information_768_PSCH_ReconfRqst, "id-hS-SCCH-Information-768-PSCH-ReconfRqst" },
  { id_hS_SCCH_InformationModify_768_PSCH_ReconfRqst, "id-hS-SCCH-InformationModify-768-PSCH-ReconfRqst" },
  { id_hsSCCH_Specific_Information_ResponseTDD768, "id-hsSCCH-Specific-Information-ResponseTDD768" },
  { id_E_DPCH_Information_RL_AdditionReqFDD, "id-E-DPCH-Information-RL-AdditionReqFDD" },
  { id_Unknown_773, "id-Unknown-773" },
  { id_Unknown_774, "id-Unknown-774" },
  { id_PDSCH_Timeslot_Format_PSCH_ReconfRqst_LCR, "id-PDSCH-Timeslot-Format-PSCH-ReconfRqst-LCR" },
  { id_Unknown_776, "id-Unknown-776" },
  { id_Unknown_777, "id-Unknown-777" },
  { id_Unknown_778, "id-Unknown-778" },
  { id_Unknown_779, "id-Unknown-779" },
  { id_PUSCH_Timeslot_Format_PSCH_ReconfRqst_LCR, "id-PUSCH-Timeslot-Format-PSCH-ReconfRqst-LCR" },
  { id_Unknown_781, "id-Unknown-781" },
  { id_E_DCH_PowerOffset_for_SchedulingInfo, "id-E-DCH-PowerOffset-for-SchedulingInfo" },
  { id_HSDSCH_Configured_Indicator, "id-HSDSCH-Configured-Indicator" },
  { id_Unknown_784, "id-Unknown-784" },
  { id_Unknown_785, "id-Unknown-785" },
  { id_Rx_Timing_Deviation_Value_384_ext, "id-Rx-Timing-Deviation-Value-384-ext" },
  { id_RTWP_ReportingIndicator, "id-RTWP-ReportingIndicator" },
  { id_RTWP_CellPortion_ReportingIndicator, "id-RTWP-CellPortion-ReportingIndicator" },
  { id_Received_Scheduled_EDCH_Power_Share_Value, "id-Received-Scheduled-EDCH-Power-Share-Value" },
  { id_Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value, "id-Received-Scheduled-EDCH-Power-Share-For-CellPortion-Value" },
  { id_Received_Scheduled_EDCH_Power_Share, "id-Received-Scheduled-EDCH-Power-Share" },
  { id_Received_Scheduled_EDCH_Power_Share_For_CellPortion, "id-Received-Scheduled-EDCH-Power-Share-For-CellPortion" },
  { id_tFCI_Presence, "id-tFCI-Presence" },
  { id_HSSICH_TPC_StepSize, "id-HSSICH-TPC-StepSize" },
  { id_E_RUCCH_InformationList_AuditRsp, "id-E-RUCCH-InformationList-AuditRsp" },
  { id_E_RUCCH_InformationList_ResourceStatusInd, "id-E-RUCCH-InformationList-ResourceStatusInd" },
  { id_E_DCH_TDD_CapacityConsumptionLaw, "id-E-DCH-TDD-CapacityConsumptionLaw" },
  { id_E_RUCCH_Information, "id-E-RUCCH-Information" },
  { id_E_DCH_Information, "id-E-DCH-Information" },
  { id_E_DCH_Information_Response, "id-E-DCH-Information-Response" },
  { id_E_DCH_Information_Reconfig, "id-E-DCH-Information-Reconfig" },
  { id_E_PUCH_Information_PSCH_ReconfRqst, "id-E-PUCH-Information-PSCH-ReconfRqst" },
  { id_Add_To_E_AGCH_Resource_Pool_PSCH_ReconfRqst, "id-Add-To-E-AGCH-Resource-Pool-PSCH-ReconfRqst" },
  { id_Modify_E_AGCH_Resource_Pool_PSCH_ReconfRqst, "id-Modify-E-AGCH-Resource-Pool-PSCH-ReconfRqst" },
  { id_Delete_From_E_AGCH_Resource_Pool_PSCH_ReconfRqst, "id-Delete-From-E-AGCH-Resource-Pool-PSCH-ReconfRqst" },
  { id_E_HICH_Information_PSCH_ReconfRqst, "id-E-HICH-Information-PSCH-ReconfRqst" },
  { id_E_HICH_TimeOffset, "id-E-HICH-TimeOffset" },
  { id_Maximum_Generated_ReceivedTotalWideBandPowerInOtherCells, "id-Maximum-Generated-ReceivedTotalWideBandPowerInOtherCells" },
  { id_E_DCH_Serving_RL_ID, "id-E-DCH-Serving-RL-ID" },
  { id_E_RUCCH_768_InformationList_AuditRsp, "id-E-RUCCH-768-InformationList-AuditRsp" },
  { id_E_RUCCH_768_InformationList_ResourceStatusInd, "id-E-RUCCH-768-InformationList-ResourceStatusInd" },
  { id_E_RUCCH_768_Information, "id-E-RUCCH-768-Information" },
  { id_E_DCH_768_Information, "id-E-DCH-768-Information" },
  { id_E_DCH_768_Information_Reconfig, "id-E-DCH-768-Information-Reconfig" },
  { id_E_PUCH_Information_768_PSCH_ReconfRqst, "id-E-PUCH-Information-768-PSCH-ReconfRqst" },
  { id_Add_To_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst, "id-Add-To-E-AGCH-Resource-Pool-768-PSCH-ReconfRqst" },
  { id_Modify_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst, "id-Modify-E-AGCH-Resource-Pool-768-PSCH-ReconfRqst" },
  { id_E_HICH_Information_768_PSCH_ReconfRqst, "id-E-HICH-Information-768-PSCH-ReconfRqst" },
  { id_ExtendedPropagationDelay, "id-ExtendedPropagationDelay" },
  { id_Extended_Round_Trip_Time_Value, "id-Extended-Round-Trip-Time-Value" },
  { id_AlternativeFormatReportingIndicator, "id-AlternativeFormatReportingIndicator" },
  { id_DCH_Indicator_For_E_DCH_HSDPA_Operation, "id-DCH-Indicator-For-E-DCH-HSDPA-Operation" },
  { id_Reference_ReceivedTotalWideBandPowerReporting, "id-Reference-ReceivedTotalWideBandPowerReporting" },
  { id_Reference_ReceivedTotalWideBandPowerSupportIndicator, "id-Reference-ReceivedTotalWideBandPowerSupportIndicator" },
  { id_ueCapability_Info, "id-ueCapability-Info" },
  { id_MAChs_ResetIndicator, "id-MAChs-ResetIndicator" },
  { id_Fast_Reconfiguration_Mode, "id-Fast-Reconfiguration-Mode" },
  { id_Fast_Reconfiguration_Permission, "id-Fast-Reconfiguration-Permission" },
  { id_BroadcastReference, "id-BroadcastReference" },
  { id_BroadcastCommonTransportBearerIndication, "id-BroadcastCommonTransportBearerIndication" },
  { id_ContinuousPacketConnectivityDTX_DRX_Capability, "id-ContinuousPacketConnectivityDTX-DRX-Capability" },
  { id_ContinuousPacketConnectivityDTX_DRX_Information, "id-ContinuousPacketConnectivityDTX-DRX-Information" },
  { id_ContinuousPacketConnectivityHS_SCCH_less_Capability, "id-ContinuousPacketConnectivityHS-SCCH-less-Capability" },
  { id_ContinuousPacketConnectivityHS_SCCH_less_Information, "id-ContinuousPacketConnectivityHS-SCCH-less-Information" },
  { id_ContinuousPacketConnectivityHS_SCCH_less_Information_Response, "id-ContinuousPacketConnectivityHS-SCCH-less-Information-Response" },
  { id_CPC_Information, "id-CPC-Information" },
  { id_MIMO_Capability, "id-MIMO-Capability" },
  { id_MIMO_PilotConfiguration, "id-MIMO-PilotConfiguration" },
  { id_Unknown_839, "id-Unknown-839" },
  { id_Unknown_840, "id-Unknown-840" },
  { id_MBSFN_Cell_ParameterID_Cell_SetupRqstTDD, "id-MBSFN-Cell-ParameterID-Cell-SetupRqstTDD" },
  { id_MBSFN_Cell_ParameterID_Cell_ReconfRqstTDD, "id-MBSFN-Cell-ParameterID-Cell-ReconfRqstTDD" },
  { id_S_CCPCH_Modulation, "id-S-CCPCH-Modulation" },
  { id_HS_PDSCH_Code_Change_Grant, "id-HS-PDSCH-Code-Change-Grant" },
  { id_HS_PDSCH_Code_Change_Indicator, "id-HS-PDSCH-Code-Change-Indicator" },
  { id_SYNC_UL_Partition_LCR, "id-SYNC-UL-Partition-LCR" },
  { id_E_DCH_LCR_Information, "id-E-DCH-LCR-Information" },
  { id_E_DCH_LCR_Information_Reconfig, "id-E-DCH-LCR-Information-Reconfig" },
  { id_Unknown_849, "id-Unknown-849" },
  { id_Unknown_850, "id-Unknown-850" },
  { id_Unknown_851, "id-Unknown-851" },
  { id_E_PUCH_Information_LCR_PSCH_ReconfRqst, "id-E-PUCH-Information-LCR-PSCH-ReconfRqst" },
  { id_Add_To_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst, "id-Add-To-E-AGCH-Resource-Pool-LCR-PSCH-ReconfRqst" },
  { id_Modify_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst, "id-Modify-E-AGCH-Resource-Pool-LCR-PSCH-ReconfRqst" },
  { id_Add_To_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst, "id-Add-To-E-HICH-Resource-Pool-LCR-PSCH-ReconfRqst" },
  { id_Modify_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst, "id-Modify-E-HICH-Resource-Pool-LCR-PSCH-ReconfRqst" },
  { id_Delete_From_E_HICH_Resource_Pool_PSCH_ReconfRqst, "id-Delete-From-E-HICH-Resource-Pool-PSCH-ReconfRqst" },
  { id_E_HICH_TimeOffsetLCR, "id-E-HICH-TimeOffsetLCR" },
  { id_Unknown_859, "id-Unknown-859" },
  { id_SixtyfourQAM_DL_Capability, "id-SixtyfourQAM-DL-Capability" },
  { id_SixteenQAM_UL_Capability, "id-SixteenQAM-UL-Capability" },
  { id_Unknown_862, "id-Unknown-862" },
  { id_Unknown_863, "id-Unknown-863" },
  { id_HSDSCH_MACdPDU_SizeCapability, "id-HSDSCH-MACdPDU-SizeCapability" },
  { id_HSDSCH_MACdPDUSizeFormat, "id-HSDSCH-MACdPDUSizeFormat" },
  { id_MaximumMACdPDU_SizeExtended, "id-MaximumMACdPDU-SizeExtended" },
  { id_Unknown_867, "id-Unknown-867" },
  { id_Unknown_868, "id-Unknown-868" },
  { id_Unknown_869, "id-Unknown-869" },
  { id_F_DPCH_SlotFormat, "id-F-DPCH-SlotFormat" },
  { id_F_DPCH_SlotFormatCapability, "id-F-DPCH-SlotFormatCapability" },
  { id_LCRTDD_uplink_Physical_Channel_Capability, "id-LCRTDD-uplink-Physical-Channel-Capability" },
  { id_Extended_RNC_ID, "id-Extended-RNC-ID" },
  { id_Max_UE_DTX_Cycle, "id-Max-UE-DTX-Cycle" },
  { id_Unknown_875, "id-Unknown-875" },
  { id_Secondary_CCPCH_SlotFormat_Extended, "id-Secondary-CCPCH-SlotFormat-Extended" },
  { id_Unknown_877, "id-Unknown-877" },
  { id_MBSFN_Only_Mode_Indicator_Cell_SetupRqstTDD_LCR, "id-MBSFN-Only-Mode-Indicator-Cell-SetupRqstTDD-LCR" },
  { id_MBSFN_Only_Mode_Capability, "id-MBSFN-Only-Mode-Capability" },
  { id_Time_Slot_Parameter_ID, "id-Time-Slot-Parameter-ID" },
  { id_Additional_failed_HS_SICH, "id-Additional-failed-HS-SICH" },
  { id_Additional_missed_HS_SICH, "id-Additional-missed-HS-SICH" },
  { id_Additional_total_HS_SICH, "id-Additional-total-HS-SICH" },
  { id_Additional_HS_SICH_Reception_Quality_Measurement_Value, "id-Additional-HS-SICH-Reception-Quality-Measurement-Value" },
  { id_Unknown_885, "id-Unknown-885" },
  { id_Unknown_886, "id-Unknown-886" },
  { id_GANSS_Common_Data, "id-GANSS-Common-Data" },
  { id_GANSS_Information, "id-GANSS-Information" },
  { id_GANSS_Generic_Data, "id-GANSS-Generic-Data" },
  { id_TUTRANGANSSMeasurementThresholdInformation, "id-TUTRANGANSSMeasurementThresholdInformation" },
  { id_TUTRANGANSSMeasurementValueInformation, "id-TUTRANGANSSMeasurementValueInformation" },
  { id_ModulationPO_MBSFN, "id-ModulationPO-MBSFN" },
  { id_Unknown_893, "id-Unknown-893" },
  { id_Unknown_894, "id-Unknown-894" },
  { id_Enhanced_FACH_Capability, "id-Enhanced-FACH-Capability" },
  { id_Enhanced_PCH_Capability, "id-Enhanced-PCH-Capability" },
  { id_HSDSCH_Common_System_InformationFDD, "id-HSDSCH-Common-System-InformationFDD" },
  { id_HSDSCH_Common_System_Information_ResponseFDD, "id-HSDSCH-Common-System-Information-ResponseFDD" },
  { id_HSDSCH_Paging_System_InformationFDD, "id-HSDSCH-Paging-System-InformationFDD" },
  { id_HSDSCH_Paging_System_Information_ResponseFDD, "id-HSDSCH-Paging-System-Information-ResponseFDD" },
  { id_MBMS_Capability, "id-MBMS-Capability" },
  { id_Ext_Reference_E_TFCI_PO, "id-Ext-Reference-E-TFCI-PO" },
  { id_Ext_Max_Bits_MACe_PDU_non_scheduled, "id-Ext-Max-Bits-MACe-PDU-non-scheduled" },
  { id_HARQ_MemoryPartitioningInfoExtForMIMO, "id-HARQ-MemoryPartitioningInfoExtForMIMO" },
  { id_MIMO_ActivationIndicator, "id-MIMO-ActivationIndicator" },
  { id_MIMO_Mode_Indicator, "id-MIMO-Mode-Indicator" },
  { id_MIMO_N_M_Ratio, "id-MIMO-N-M-Ratio" },
  { id_IPMulticastIndication, "id-IPMulticastIndication" },
  { id_IPMulticastDataBearerIndication, "id-IPMulticastDataBearerIndication" },
  { id_TransportBearerNotSetupIndicator, "id-TransportBearerNotSetupIndicator" },
  { id_TransportBearerNotRequestedIndicator, "id-TransportBearerNotRequestedIndicator" },
  { id_TimeSlotConfigurationList_LCR_CTCH_SetupRqstTDD, "id-TimeSlotConfigurationList-LCR-CTCH-SetupRqstTDD" },
  { id_Cell_Frequency_List_Information_LCR_MulFreq_AuditRsp, "id-Cell-Frequency-List-Information-LCR-MulFreq-AuditRsp" },
  { id_Cell_Frequency_List_InformationItem_LCR_MulFreq_AuditRsp, "id-Cell-Frequency-List-InformationItem-LCR-MulFreq-AuditRsp" },
  { id_Cell_Frequency_List_LCR_MulFreq_Cell_SetupRqstTDD, "id-Cell-Frequency-List-LCR-MulFreq-Cell-SetupRqstTDD" },
  { id_UARFCN_Adjustment, "id-UARFCN-Adjustment" },
  { id_Cell_Frequency_List_Information_LCR_MulFreq_ResourceStatusInd, "id-Cell-Frequency-List-Information-LCR-MulFreq-ResourceStatusInd" },
  { id_Cell_Frequency_List_InformationItem_LCR_MulFreq_ResourceStatusInd, "id-Cell-Frequency-List-InformationItem-LCR-MulFreq-ResourceStatusInd" },
  { id_UPPCHPositionLCR, "id-UPPCHPositionLCR" },
  { id_UPPCH_LCR_Parameters_CTCH_ReconfRqstTDD, "id-UPPCH-LCR-Parameters-CTCH-ReconfRqstTDD" },
  { id_UPPCH_LCR_InformationList_AuditRsp, "id-UPPCH-LCR-InformationList-AuditRsp" },
  { id_UPPCH_LCR_InformationItem_AuditRsp, "id-UPPCH-LCR-InformationItem-AuditRsp" },
  { id_UPPCH_LCR_InformationList_ResourceStatusInd, "id-UPPCH-LCR-InformationList-ResourceStatusInd" },
  { id_UPPCH_LCR_InformationItem_ResourceStatusInd, "id-UPPCH-LCR-InformationItem-ResourceStatusInd" },
  { id_multipleFreq_dL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst, "id-multipleFreq-dL-HS-PDSCH-Timeslot-Information-LCR-PSCH-ReconfRqst" },
  { id_number_Of_Supported_Carriers, "id-number-Of-Supported-Carriers" },
  { id_multipleFreq_HSPDSCH_InformationList_ResponseTDDLCR, "id-multipleFreq-HSPDSCH-InformationList-ResponseTDDLCR" },
  { id_Unsuccessful_UARFCNItem_PSCH_ReconfFailureTDD, "id-Unsuccessful-UARFCNItem-PSCH-ReconfFailureTDD" },
  { id_multipleFreq_HS_DSCH_Resources_InformationList_AuditRsp, "id-multipleFreq-HS-DSCH-Resources-InformationList-AuditRsp" },
  { id_multipleFreq_HS_DSCH_Resources_InformationList_ResourceStatusInd, "id-multipleFreq-HS-DSCH-Resources-InformationList-ResourceStatusInd" },
  { id_UARFCNSpecificCauseList, "id-UARFCNSpecificCauseList" },
  { id_tSN_Length, "id-tSN-Length" },
  { id_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst, "id-MultipleFreq-DL-HS-PDSCH-Timeslot-Information-LCRItem-PSCH-ReconfRqst" },
  { id_multicarrier_number, "id-multicarrier-number" },
  { id_Extended_HS_SCCH_ID, "id-Extended-HS-SCCH-ID" },
  { id_Extended_HS_SICH_ID, "id-Extended-HS-SICH-ID" },
  { id_HSSICH_InfoExt_DM_Rqst, "id-HSSICH-InfoExt-DM-Rqst" },
  { id_Delete_From_HS_SCCH_Resource_PoolExt_PSCH_ReconfRqst, "id-Delete-From-HS-SCCH-Resource-PoolExt-PSCH-ReconfRqst" },
  { id_HS_SCCH_InformationExt_LCR_PSCH_ReconfRqst, "id-HS-SCCH-InformationExt-LCR-PSCH-ReconfRqst" },
  { id_HS_SCCH_InformationModifyExt_LCR_PSCH_ReconfRqst, "id-HS-SCCH-InformationModifyExt-LCR-PSCH-ReconfRqst" },
  { id_PowerControlGAP, "id-PowerControlGAP" },
  { id_MBSFN_SpecialTimeSlot_LCR, "id-MBSFN-SpecialTimeSlot-LCR" },
  { id_Common_MACFlows_to_DeleteFDD, "id-Common-MACFlows-to-DeleteFDD" },
  { id_Paging_MACFlows_to_DeleteFDD, "id-Paging-MACFlows-to-DeleteFDD" },
  { id_E_TFCI_Boost_Information, "id-E-TFCI-Boost-Information" },
  { id_SixteenQAM_UL_Operation_Indicator, "id-SixteenQAM-UL-Operation-Indicator" },
  { id_SixtyfourQAM_UsageAllowedIndicator, "id-SixtyfourQAM-UsageAllowedIndicator" },
  { id_SixtyfourQAM_DL_UsageIndicator, "id-SixtyfourQAM-DL-UsageIndicator" },
  { id_Default_Serving_Grant_in_DTX_Cycle2, "id-Default-Serving-Grant-in-DTX-Cycle2" },
  { id_Maximum_Target_ReceivedTotalWideBandPower_LCR, "id-Maximum-Target-ReceivedTotalWideBandPower-LCR" },
  { id_E_DPDCH_PowerInterpolation, "id-E-DPDCH-PowerInterpolation" },
  { id_Extended_E_DCH_LCRTDD_PhysicalLayerCategory, "id-Extended-E-DCH-LCRTDD-PhysicalLayerCategory" },
  { id_MultipleFreq_E_DCH_Resources_InformationList_AuditRsp, "id-MultipleFreq-E-DCH-Resources-InformationList-AuditRsp" },
  { id_MultipleFreq_E_DCH_Resources_InformationList_ResourceStatusInd, "id-MultipleFreq-E-DCH-Resources-InformationList-ResourceStatusInd" },
  { id_MultipleFreq_E_PUCH_Timeslot_InformationList_LCR_PSCH_ReconfRqst, "id-MultipleFreq-E-PUCH-Timeslot-InformationList-LCR-PSCH-ReconfRqst" },
  { id_MultipleFreq_E_PUCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst, "id-MultipleFreq-E-PUCH-Timeslot-Information-LCRItem-PSCH-ReconfRqst" },
  { id_Extended_E_HICH_ID_TDD, "id-Extended-E-HICH-ID-TDD" },
  { id_ContinuousPacketConnectivityHS_SCCH_less_Deactivate_Indicator, "id-ContinuousPacketConnectivityHS-SCCH-less-Deactivate-Indicator" },
  { id_E_DCH_MACdPDU_SizeCapability, "id-E-DCH-MACdPDU-SizeCapability" },
  { id_E_DCH_MACdPDUSizeFormat, "id-E-DCH-MACdPDUSizeFormat" },
  { id_MaximumNumber_Of_Retransmission_for_Scheduling_Info_LCRTDD, "id-MaximumNumber-Of-Retransmission-for-Scheduling-Info-LCRTDD" },
  { id_E_DCH_RetransmissionTimer_for_SchedulingInfo_LCRTDD, "id-E-DCH-RetransmissionTimer-for-SchedulingInfo-LCRTDD" },
  { id_E_HICH_TimeOffset_Extension, "id-E-HICH-TimeOffset-Extension" },
  { id_MultipleFreq_E_HICH_TimeOffsetLCR, "id-MultipleFreq-E-HICH-TimeOffsetLCR" },
  { id_E_PUCH_PowerControlGAP, "id-E-PUCH-PowerControlGAP" },
  { id_HSDSCH_TBSizeTableIndicator, "id-HSDSCH-TBSizeTableIndicator" },
  { id_E_DCH_DL_Control_Channel_Change_Information, "id-E-DCH-DL-Control-Channel-Change-Information" },
  { id_E_DCH_DL_Control_Channel_Grant_Information, "id-E-DCH-DL-Control-Channel-Grant-Information" },
  { id_DGANSS_Corrections_Req, "id-DGANSS-Corrections-Req" },
  { id_UE_with_enhanced_HS_SCCH_support_indicator, "id-UE-with-enhanced-HS-SCCH-support-indicator" },
  { id_AdditionalTimeSlotListLCR, "id-AdditionalTimeSlotListLCR" },
  { id_AdditionalMeasurementValueList, "id-AdditionalMeasurementValueList" },
  { id_Unknown_973, "id-Unknown-973" },
  { id_Unknown_974, "id-Unknown-974" },
  { id_Unknown_975, "id-Unknown-975" },
  { id_Unknown_976, "id-Unknown-976" },
  { id_Unknown_977, "id-Unknown-977" },
  { id_E_AGCH_Table_Choice, "id-E-AGCH-Table-Choice" },
  { id_Unknown_979, "id-Unknown-979" },
  { id_Unknown_980, "id-Unknown-980" },
  { id_PLCCH_parameters, "id-PLCCH-parameters" },
  { id_E_RUCCH_parameters, "id-E-RUCCH-parameters" },
  { id_E_RUCCH_768_parameters, "id-E-RUCCH-768-parameters" },
  { id_HS_Cause, "id-HS-Cause" },
  { id_E_Cause, "id-E-Cause" },
  { id_Unknown_986, "id-Unknown-986" },
  { id_Common_EDCH_Capability, "id-Common-EDCH-Capability" },
  { id_E_AI_Capability, "id-E-AI-Capability" },
  { id_Common_EDCH_System_InformationFDD, "id-Common-EDCH-System-InformationFDD" },
  { id_Common_UL_MACFlows_to_DeleteFDD, "id-Common-UL-MACFlows-to-DeleteFDD" },
  { id_Common_EDCH_MACdFlows_to_DeleteFDD, "id-Common-EDCH-MACdFlows-to-DeleteFDD" },
  { id_Common_EDCH_System_Information_ResponseFDD, "id-Common-EDCH-System-Information-ResponseFDD" },
  { id_Cell_ERNTI_Status_Information, "id-Cell-ERNTI-Status-Information" },
  { id_Enhanced_UE_DRX_Capability, "id-Enhanced-UE-DRX-Capability" },
  { id_Enhanced_UE_DRX_InformationFDD, "id-Enhanced-UE-DRX-InformationFDD" },
  { id_TransportBearerRequestIndicator, "id-TransportBearerRequestIndicator" },
  { id_SixtyfourQAM_DL_MIMO_Combined_Capability, "id-SixtyfourQAM-DL-MIMO-Combined-Capability" },
  { id_E_RNTI, "id-E-RNTI" },
  { id_MinimumReducedE_DPDCH_GainFactor, "id-MinimumReducedE-DPDCH-GainFactor" },
  { id_GANSS_Time_ID, "id-GANSS-Time-ID" },
  { id_GANSS_AddIonoModelReq, "id-GANSS-AddIonoModelReq" },
  { id_GANSS_EarthOrientParaReq, "id-GANSS-EarthOrientParaReq" },
  { id_GANSS_AddNavigationModelsReq, "id-GANSS-AddNavigationModelsReq" },
  { id_GANSS_AddUTCModelsReq, "id-GANSS-AddUTCModelsReq" },
  { id_GANSS_AuxInfoReq, "id-GANSS-AuxInfoReq" },
  { id_GANSS_SBAS_ID, "id-GANSS-SBAS-ID" },
  { id_GANSS_ID, "id-GANSS-ID" },
  { id_GANSS_Additional_Ionospheric_Model, "id-GANSS-Additional-Ionospheric-Model" },
  { id_GANSS_Earth_Orientation_Parameters, "id-GANSS-Earth-Orientation-Parameters" },
  { id_GANSS_Additional_Time_Models, "id-GANSS-Additional-Time-Models" },
  { id_GANSS_Additional_Navigation_Models, "id-GANSS-Additional-Navigation-Models" },
  { id_GANSS_Additional_UTC_Models, "id-GANSS-Additional-UTC-Models" },
  { id_GANSS_Auxiliary_Information, "id-GANSS-Auxiliary-Information" },
  { id_ERACH_CM_Rqst, "id-ERACH-CM-Rqst" },
  { id_ERACH_CM_Rsp, "id-ERACH-CM-Rsp" },
  { id_ERACH_CM_Rprt, "id-ERACH-CM-Rprt" },
  { id_EDCH_RACH_Report_Value, "id-EDCH-RACH-Report-Value" },
  { id_EDCH_RACH_Report_IncrDecrThres, "id-EDCH-RACH-Report-IncrDecrThres" },
  { id_EDCH_RACH_Report_ThresholdInformation, "id-EDCH-RACH-Report-ThresholdInformation" },
  { id_E_DPCCH_Power_Boosting_Capability, "id-E-DPCCH-Power-Boosting-Capability" },
  { id_HSDSCH_Common_System_InformationLCR, "id-HSDSCH-Common-System-InformationLCR" },
  { id_Unknown_1022, "id-Unknown-1022" },
  { id_HSDSCH_Paging_System_InformationLCR, "id-HSDSCH-Paging-System-InformationLCR" },
  { id_HSDSCH_Paging_System_Information_ResponseLCR, "id-HSDSCH-Paging-System-Information-ResponseLCR" },
  { id_Common_MACFlows_to_DeleteLCR, "id-Common-MACFlows-to-DeleteLCR" },
  { id_Paging_MACFlows_to_DeleteLCR, "id-Paging-MACFlows-to-DeleteLCR" },
  { id_Common_EDCH_System_InformationLCR, "id-Common-EDCH-System-InformationLCR" },
  { id_Common_UL_MACFlows_to_DeleteLCR, "id-Common-UL-MACFlows-to-DeleteLCR" },
  { id_Common_EDCH_MACdFlows_to_DeleteLCR, "id-Common-EDCH-MACdFlows-to-DeleteLCR" },
  { id_Common_EDCH_System_Information_ResponseLCR, "id-Common-EDCH-System-Information-ResponseLCR" },
  { id_Enhanced_UE_DRX_CapabilityLCR, "id-Enhanced-UE-DRX-CapabilityLCR" },
  { id_Enhanced_UE_DRX_InformationLCR, "id-Enhanced-UE-DRX-InformationLCR" },
  { id_HSDSCH_PreconfigurationSetup, "id-HSDSCH-PreconfigurationSetup" },
  { id_HSDSCH_PreconfigurationInfo, "id-HSDSCH-PreconfigurationInfo" },
  { id_NoOfTargetCellHS_SCCH_Order, "id-NoOfTargetCellHS-SCCH-Order" },
  { id_EnhancedHSServingCC_Abort, "id-EnhancedHSServingCC-Abort" },
  { id_Additional_HS_Cell_Information_RL_Setup, "id-Additional-HS-Cell-Information-RL-Setup" },
  { id_Additional_HS_Cell_Information_Response, "id-Additional-HS-Cell-Information-Response" },
  { id_Additional_HS_Cell_Information_RL_Addition, "id-Additional-HS-Cell-Information-RL-Addition" },
  { id_Additional_HS_Cell_Change_Information_Response, "id-Additional-HS-Cell-Change-Information-Response" },
  { id_Additional_HS_Cell_Information_RL_Reconf_Prep, "id-Additional-HS-Cell-Information-RL-Reconf-Prep" },
  { id_Additional_HS_Cell_Information_RL_Reconf_Req, "id-Additional-HS-Cell-Information-RL-Reconf-Req" },
  { id_Additional_HS_Cell_Information_RL_Param_Upd, "id-Additional-HS-Cell-Information-RL-Param-Upd" },
  { id_Multi_Cell_Capability_Info, "id-Multi-Cell-Capability-Info" },
  { id_IMB_Parameters, "id-IMB-Parameters" },
  { id_MACes_Maximum_Bitrate_LCR, "id-MACes-Maximum-Bitrate-LCR" },
  { id_Semi_PersistentScheduling_CapabilityLCR, "id-Semi-PersistentScheduling-CapabilityLCR" },
  { id_E_DCH_Semi_PersistentScheduling_Information_LCR, "id-E-DCH-Semi-PersistentScheduling-Information-LCR" },
  { id_HS_DSCH_Semi_PersistentScheduling_Information_LCR, "id-HS-DSCH-Semi-PersistentScheduling-Information-LCR" },
  { id_Add_To_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst, "id-Add-To-Non-HS-SCCH-Associated-HS-SICH-Resource-Pool-LCR-PSCH-ReconfRqst" },
  { id_Modify_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst, "id-Modify-Non-HS-SCCH-Associated-HS-SICH-Resource-Pool-LCR-PSCH-ReconfRqst" },
  { id_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst, "id-Delete-From-Non-HS-SCCH-Associated-HS-SICH-Resource-Pool-LCR-PSCH-ReconfRqst" },
  { id_ContinuousPacketConnectivity_DRX_CapabilityLCR, "id-ContinuousPacketConnectivity-DRX-CapabilityLCR" },
  { id_ContinuousPacketConnectivity_DRX_InformationLCR, "id-ContinuousPacketConnectivity-DRX-InformationLCR" },
  { id_ContinuousPacketConnectivity_DRX_Information_ResponseLCR, "id-ContinuousPacketConnectivity-DRX-Information-ResponseLCR" },
  { id_CPC_InformationLCR, "id-CPC-InformationLCR" },
  { id_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR, "id-HS-DSCH-Semi-PersistentScheduling-Information-ResponseLCR" },
  { id_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR, "id-E-DCH-Semi-PersistentScheduling-Information-ResponseLCR" },
  { id_E_AGCH_UE_Inactivity_Monitor_Threshold, "id-E-AGCH-UE-Inactivity-Monitor-Threshold" },
  { id_Unknown_1060, "id-Unknown-1060" },
  { id_Unknown_1061, "id-Unknown-1061" },
  { id_Unknown_1062, "id-Unknown-1062" },
  { id_IdleIntervalInformation, "id-IdleIntervalInformation" },
  { id_GANSS_alm_keplerianNAVAlmanac, "id-GANSS-alm-keplerianNAVAlmanac" },
  { id_GANSS_alm_keplerianReducedAlmanac, "id-GANSS-alm-keplerianReducedAlmanac" },
  { id_GANSS_alm_keplerianMidiAlmanac, "id-GANSS-alm-keplerianMidiAlmanac" },
  { id_GANSS_alm_keplerianGLONASS, "id-GANSS-alm-keplerianGLONASS" },
  { id_GANSS_alm_ecefSBASAlmanac, "id-GANSS-alm-ecefSBASAlmanac" },
  { id_Unknown_1069, "id-Unknown-1069" },
  { id_HSSICH_ReferenceSignal_InformationLCR, "id-HSSICH-ReferenceSignal-InformationLCR" },
  { id_MIMO_ReferenceSignal_InformationListLCR, "id-MIMO-ReferenceSignal-InformationListLCR" },
  { id_MIMO_SFMode_For_HSPDSCHDualStream, "id-MIMO-SFMode-For-HSPDSCHDualStream" },
  { id_MIMO_SFMode_Supported_For_HSPDSCHDualStream, "id-MIMO-SFMode-Supported-For-HSPDSCHDualStream" },
  { id_UE_Selected_MBMS_Service_Information, "id-UE-Selected-MBMS-Service-Information" },
  { id_Unknown_1075, "id-Unknown-1075" },
  { id_Unknown_1076, "id-Unknown-1076" },
  { id_MultiCarrier_HSDSCH_Physical_Layer_Category, "id-MultiCarrier-HSDSCH-Physical-Layer-Category" },
  { id_Common_E_DCH_HSDPCCH_Capability, "id-Common-E-DCH-HSDPCCH-Capability" },
  { id_DL_RLC_PDU_Size_Format, "id-DL-RLC-PDU-Size-Format" },
  { id_HSSICH_ReferenceSignal_InformationModifyLCR, "id-HSSICH-ReferenceSignal-InformationModifyLCR" },
  { id_schedulingPriorityIndicator, "id-schedulingPriorityIndicator" },
  { id_TimeSlotMeasurementValueListLCR, "id-TimeSlotMeasurementValueListLCR" },
  { id_Unknown_1083, "id-Unknown-1083" },
  { id_Unknown_1084, "id-Unknown-1084" },
  { id_UE_SupportIndicatorExtension, "id-UE-SupportIndicatorExtension" },
  { id_Unknown_1086, "id-Unknown-1086" },
  { id_Unknown_1087, "id-Unknown-1087" },
  { id_Single_Stream_MIMO_ActivationIndicator, "id-Single-Stream-MIMO-ActivationIndicator" },
  { id_Single_Stream_MIMO_Capability, "id-Single-Stream-MIMO-Capability" },
  { id_Single_Stream_MIMO_Mode_Indicator, "id-Single-Stream-MIMO-Mode-Indicator" },
  { id_Dual_Band_Capability_Info, "id-Dual-Band-Capability-Info" },
  { id_UE_AggregateMaximumBitRate, "id-UE-AggregateMaximumBitRate" },
  { id_UE_AggregateMaximumBitRate_Enforcement_Indicator, "id-UE-AggregateMaximumBitRate-Enforcement-Indicator" },
  { id_Unknown_1094, "id-Unknown-1094" },
  { id_Unknown_1095, "id-Unknown-1095" },
  { id_Unknown_1096, "id-Unknown-1096" },
  { id_Unknown_1097, "id-Unknown-1097" },
  { id_Unknown_1098, "id-Unknown-1098" },
  { id_Unknown_1099, "id-Unknown-1099" },
  { id_Unknown_1100, "id-Unknown-1100" },
  { id_MIMO_Power_Offset_For_S_CPICH_Capability, "id-MIMO-Power-Offset-For-S-CPICH-Capability" },
  { id_MIMO_PilotConfigurationExtension, "id-MIMO-PilotConfigurationExtension" },
  { id_TxDiversityOnDLControlChannelsByMIMOUECapability, "id-TxDiversityOnDLControlChannelsByMIMOUECapability" },
  { id_ULTimeslotISCPValue_For_CellPortion, "id-ULTimeslotISCPValue-For-CellPortion" },
  { id_UpPTSInterferenceValue_For_CellPortion, "id-UpPTSInterferenceValue-For-CellPortion" },
  { id_Best_Cell_Portions_ValueLCR, "id-Best-Cell-Portions-ValueLCR" },
  { id_Transmitted_Carrier_Power_For_CellPortion_ValueLCR, "id-Transmitted-Carrier-Power-For-CellPortion-ValueLCR" },
  { id_Received_total_wide_band_power_For_CellPortion_ValueLCR, "id-Received-total-wide-band-power-For-CellPortion-ValueLCR" },
  { id_UL_TimeslotISCP_For_CellPortion_Value, "id-UL-TimeslotISCP-For-CellPortion-Value" },
  { id_HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR, "id-HS-DSCHRequiredPowerValueInformation-For-CellPortionLCR" },
  { id_HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR, "id-HS-DSCHProvidedBitRateValueInformation-For-CellPortionLCR" },
  { id_E_DCHProvidedBitRateValueInformation_For_CellPortion, "id-E-DCHProvidedBitRateValueInformation-For-CellPortion" },
  { id_UpPTSInterference_For_CellPortion_Value, "id-UpPTSInterference-For-CellPortion-Value" },
  { id_NumberOfReportedCellPortionsLCR, "id-NumberOfReportedCellPortionsLCR" },
  { id_CellPortion_CapabilityLCR, "id-CellPortion-CapabilityLCR" },
  { id_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue, "id-TransmittedCarrierPowerOfAllCodesNotUsedForHS-PDSCH-HS-SCCH-E-AGCHOrE-HICHTransmissionCellPortionValue" },
  { id_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortion, "id-TransmittedCarrierPowerOfAllCodesNotUsedForHS-PDSCH-HS-SCCH-E-AGCHOrE-HICHTransmissionCellPortion" },
  { id_Unknown_1118, "id-Unknown-1118" },
  { id_ActivationInformation, "id-ActivationInformation" },
  { id_Additional_EDCH_Cell_Information_RL_Setup_Req, "id-Additional-EDCH-Cell-Information-RL-Setup-Req" },
  { id_Additional_EDCH_Cell_Information_Response, "id-Additional-EDCH-Cell-Information-Response" },
  { id_Additional_EDCH_Cell_Information_RL_Add_Req, "id-Additional-EDCH-Cell-Information-RL-Add-Req" },
  { id_Additional_EDCH_Cell_Information_Response_RL_Add, "id-Additional-EDCH-Cell-Information-Response-RL-Add" },
  { id_Additional_EDCH_Cell_Information_RL_Reconf_Prep, "id-Additional-EDCH-Cell-Information-RL-Reconf-Prep" },
  { id_Additional_EDCH_Cell_Information_RL_Reconf_Req, "id-Additional-EDCH-Cell-Information-RL-Reconf-Req" },
  { id_Additional_EDCH_Cell_Information_Bearer_Rearrangement, "id-Additional-EDCH-Cell-Information-Bearer-Rearrangement" },
  { id_Additional_EDCH_Cell_Information_RL_Param_Upd, "id-Additional-EDCH-Cell-Information-RL-Param-Upd" },
  { id_Additional_EDCH_Preconfiguration_Information, "id-Additional-EDCH-Preconfiguration-Information" },
  { id_EDCH_Indicator, "id-EDCH-Indicator" },
  { id_Unknown_1130, "id-Unknown-1130" },
  { id_HS_DSCH_SPS_Reservation_Indicator, "id-HS-DSCH-SPS-Reservation-Indicator" },
  { id_E_DCH_SPS_Reservation_Indicator, "id-E-DCH-SPS-Reservation-Indicator" },
  { id_MultipleFreq_HARQ_MemoryPartitioning_InformationList, "id-MultipleFreq-HARQ-MemoryPartitioning-InformationList" },
  { id_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_Ext, "id-Ul-common-E-DCH-MACflow-Specific-InfoResponseListLCR-Ext" },
  { id_RepetitionPeriodIndex, "id-RepetitionPeriodIndex" },
  { id_MidambleShiftLCR, "id-MidambleShiftLCR" },
  { id_MaxHSDSCH_HSSCCH_Power_per_CELLPORTION, "id-MaxHSDSCH-HSSCCH-Power-per-CELLPORTION" },
  { id_DormantModeIndicator, "id-DormantModeIndicator" },
  { id_DiversityMode, "id-DiversityMode" },
  { id_TransmitDiversityIndicator, "id-TransmitDiversityIndicator" },
  { id_NonCellSpecificTxDiversity, "id-NonCellSpecificTxDiversity" },
  { id_Cell_Capability_Container, "id-Cell-Capability-Container" },
  { id_E_RNTI_List_Request, "id-E-RNTI-List-Request" },
  { id_E_RNTI_List, "id-E-RNTI-List" },
  { id_PowerControlGAP_For_CellFACHLCR, "id-PowerControlGAP-For-CellFACHLCR" },
  { id_Unknown_1146, "id-Unknown-1146" },
  { id_UL_Synchronisation_Parameters_For_FACHLCR, "id-UL-Synchronisation-Parameters-For-FACHLCR" },
  { id_HS_DSCH_SPS_Operation_Indicator, "id-HS-DSCH-SPS-Operation-Indicator" },
  { id_HSDSCH_RNTI_For_FACH, "id-HSDSCH-RNTI-For-FACH" },
  { id_E_RNTI_For_FACH, "id-E-RNTI-For-FACH" },
  { id_Out_of_Sychronization_Window, "id-Out-of-Sychronization-Window" },
  { id_Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst, "id-Max-RTWP-perUARFCN-Information-LCR-PSCH-ReconfRqst" },
  { id_E_HICH_TimeOffset_ReconfFailureTDD, "id-E-HICH-TimeOffset-ReconfFailureTDD" },
  { id_HSSCCH_TPC_StepSize, "id-HSSCCH-TPC-StepSize" },
  { id_TS0_CapabilityLCR, "id-TS0-CapabilityLCR" },
  { id_UE_TS0_CapabilityLCR, "id-UE-TS0-CapabilityLCR" },
  { id_Common_System_Information_ResponseLCR, "id-Common-System-Information-ResponseLCR" },
  { id_Additional_EDCH_Cell_Information_ResponseRLReconf, "id-Additional-EDCH-Cell-Information-ResponseRLReconf" },
  { id_Multicell_EDCH_InformationItemIEs, "id-Multicell-EDCH-InformationItemIEs" },
  { id_Multicell_EDCH_RL_Specific_InformationItemIEs, "id-Multicell-EDCH-RL-Specific-InformationItemIEs" },
  { id_Add_To_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext, "id-Add-To-Non-HS-SCCH-Associated-HS-SICH-Resource-Pool-LCR-PSCH-ReconfRqst-Ext" },
  { id_Modify_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext, "id-Modify-Non-HS-SCCH-Associated-HS-SICH-Resource-Pool-LCR-PSCH-ReconfRqst-Ext" },
  { id_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext, "id-Delete-From-Non-HS-SCCH-Associated-HS-SICH-Resource-Pool-LCR-PSCH-ReconfRqst-Ext" },
  { id_Initial_DL_Transmission_Power, "id-Initial-DL-Transmission-Power" },
  { id_Maximum_DL_Power, "id-Maximum-DL-Power" },
  { id_Minimum_DL_Power, "id-Minimum-DL-Power" },
  { id_DCH_MeasurementOccasion_Information, "id-DCH-MeasurementOccasion-Information" },
  { id_AssociatedPhsicalChannelID, "id-AssociatedPhsicalChannelID" },
  { id_DGNSS_ValidityPeriod, "id-DGNSS-ValidityPeriod" },
  { id_PhysicalChannelID_for_CommonERNTI_RequestedIndicator, "id-PhysicalChannelID-for-CommonERNTI-RequestedIndicator" },
  { id_PrecodingWeightSetRestriction, "id-PrecodingWeightSetRestriction" },
  { id_Treset_Usage_Indicator, "id-Treset-Usage-Indicator" },
  { id_Non_Serving_RL_Preconfig_Info, "id-Non-Serving-RL-Preconfig-Info" },
  { id_Non_Serving_RL_Preconfig_Setup, "id-Non-Serving-RL-Preconfig-Setup" },
  { id_Non_Serving_RL_Preconfig_Removal, "id-Non-Serving-RL-Preconfig-Removal" },
  { id_Additional_E_DCH_Non_Serving_RL_Preconfiguration_Setup, "id-Additional-E-DCH-Non-Serving-RL-Preconfiguration-Setup" },
  { id_Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList, "id-Additional-E-DCH-New-non-serving-RL-E-DCH-FDD-DL-Control-Channel-InfoList" },
  { id_Ul_common_E_DCH_MACflow_Specific_InfoListLCR_Ext, "id-Ul-common-E-DCH-MACflow-Specific-InfoListLCR-Ext" },
  { id_CommonMACFlow_Specific_InfoList_ResponseLCR_Ext, "id-CommonMACFlow-Specific-InfoList-ResponseLCR-Ext" },
  { id_Enabling_Delay_Ext_LCR, "id-Enabling-Delay-Ext-LCR" },
  { id_Unallocated_1181, "id-Unallocated-1181" },
  { id_Unallocated_1182, "id-Unallocated-1182" },
  { id_Unallocated_1183, "id-Unallocated-1183" },
  { id_Unallocated_1184, "id-Unallocated-1184" },
  { id_Unallocated_1185, "id-Unallocated-1185" },
  { id_Unallocated_1186, "id-Unallocated-1186" },
  { id_Unallocated_1187, "id-Unallocated-1187" },
  { id_Unallocated_1188, "id-Unallocated-1188" },
  { id_Unallocated_1189, "id-Unallocated-1189" },
  { id_Unallocated_1190, "id-Unallocated-1190" },
  { id_Unallocated_1191, "id-Unallocated-1191" },
  { id_Unallocated_1192, "id-Unallocated-1192" },
  { id_Unallocated_1193, "id-Unallocated-1193" },
  { id_Unallocated_1194, "id-Unallocated-1194" },
  { id_Unallocated_1195, "id-Unallocated-1195" },
  { id_Unallocated_1196, "id-Unallocated-1196" },
  { id_Unallocated_1197, "id-Unallocated-1197" },
  { id_Unallocated_1198, "id-Unallocated-1198" },
  { id_Unallocated_1199, "id-Unallocated-1199" },
  { id_Unallocated_1200, "id-Unallocated-1200" },
  { id_Unallocated_1201, "id-Unallocated-1201" },
  { id_Unallocated_1202, "id-Unallocated-1202" },
  { id_Unallocated_1203, "id-Unallocated-1203" },
  { id_Unallocated_1204, "id-Unallocated-1204" },
  { id_Unallocated_1205, "id-Unallocated-1205" },
  { id_Unallocated_1206, "id-Unallocated-1206" },
  { id_Unallocated_1207, "id-Unallocated-1207" },
  { id_Unallocated_1208, "id-Unallocated-1208" },
  { id_Unallocated_1209, "id-Unallocated-1209" },
  { id_Unallocated_1210, "id-Unallocated-1210" },
  { id_Unallocated_1211, "id-Unallocated-1211" },
  { id_Unallocated_1212, "id-Unallocated-1212" },
  { id_Unallocated_1213, "id-Unallocated-1213" },
  { id_Unallocated_1214, "id-Unallocated-1214" },
  { id_Unallocated_1215, "id-Unallocated-1215" },
  { id_Unallocated_1216, "id-Unallocated-1216" },
  { id_Unallocated_1217, "id-Unallocated-1217" },
  { id_Unallocated_1218, "id-Unallocated-1218" },
  { id_Unallocated_1219, "id-Unallocated-1219" },
  { id_Unallocated_1220, "id-Unallocated-1220" },
  { id_Unallocated_1221, "id-Unallocated-1221" },
  { id_HSDSCH_Common_System_Information_ResponseLCR, "id-HSDSCH-Common-System-Information-ResponseLCR" },
  { 0, NULL }
};

static value_string_ext nbap_ProtocolIE_ID_vals_ext = VALUE_STRING_EXT_INIT(nbap_ProtocolIE_ID_vals);


static int
dissect_nbap_ProtocolIE_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  guint32 protocol_ie_id;
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxProtocolIEs, &protocol_ie_id, FALSE);

  nbap_get_private_data(actx->pinfo)->protocol_ie_id = protocol_ie_id; /* To carry around the packet */
  if (tree) {
    proto_item_append_text(proto_item_get_parent_nth(actx->created_item, 2), ": %s", val_to_str_ext(protocol_ie_id, &nbap_ProtocolIE_ID_vals_ext, "unknown (%d)"));
  }

  return offset;
}



static int
dissect_nbap_T_shortTransActionId(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, &nbap_get_private_data(actx->pinfo)->transaction_id, FALSE);

  return offset;
}



static int
dissect_nbap_T_longTransActionId(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 32767U, &nbap_get_private_data(actx->pinfo)->transaction_id, FALSE);

  return offset;
}


static const value_string nbap_TransactionID_vals[] = {
  {   0, "shortTransActionId" },
  {   1, "longTransActionId" },
  { 0, NULL }
};

static const per_choice_t TransactionID_choice[] = {
  {   0, &hf_nbap_shortTransActionId, ASN1_NO_EXTENSIONS     , dissect_nbap_T_shortTransActionId },
  {   1, &hf_nbap_longTransActionId, ASN1_NO_EXTENSIONS     , dissect_nbap_T_longTransActionId },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_TransactionID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_TransactionID, TransactionID_choice,
                                 NULL);

  return offset;
}


static const value_string nbap_TriggeringMessage_vals[] = {
  {   0, "initiating-message" },
  {   1, "successful-outcome" },
  {   2, "unsuccessfull-outcome" },
  {   3, "outcome" },
  { 0, NULL }
};


static int
dissect_nbap_TriggeringMessage(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_ProtocolIE_Field_value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_open_type_pdu_new(tvb, offset, actx, tree, hf_index, dissect_ProtocolIEFieldValue);

  return offset;
}


static const per_sequence_t ProtocolIE_Field_sequence[] = {
  { &hf_nbap_id             , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_ID },
  { &hf_nbap_criticality    , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Criticality },
  { &hf_nbap_ie_field_value , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Field_value },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ProtocolIE_Field(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ProtocolIE_Field, ProtocolIE_Field_sequence);

  return offset;
}


static const per_sequence_t ProtocolIE_Container_sequence_of[1] = {
  { &hf_nbap_ProtocolIE_Container_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Field },
};

static int
dissect_nbap_ProtocolIE_Container(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_ProtocolIE_Container, ProtocolIE_Container_sequence_of,
                                                  0, maxProtocolIEs, FALSE);

  return offset;
}



static int
dissect_nbap_ProtocolIE_Single_Container(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Field(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_T_extensionValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_open_type_pdu_new(tvb, offset, actx, tree, hf_index, dissect_ProtocolExtensionFieldExtensionValue);

  return offset;
}


static const per_sequence_t ProtocolExtensionField_sequence[] = {
  { &hf_nbap_id             , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_ID },
  { &hf_nbap_criticality    , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Criticality },
  { &hf_nbap_extensionValue , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_T_extensionValue },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ProtocolExtensionField(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ProtocolExtensionField, ProtocolExtensionField_sequence);

  return offset;
}


static const per_sequence_t ProtocolExtensionContainer_sequence_of[1] = {
  { &hf_nbap_ProtocolExtensionContainer_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolExtensionField },
};

static int
dissect_nbap_ProtocolExtensionContainer(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_ProtocolExtensionContainer, ProtocolExtensionContainer_sequence_of,
                                                  1, maxProtocolExtensions, FALSE);

  return offset;
}



static int
dissect_nbap_PrivateIE_Field_value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_open_type(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}


static const per_sequence_t PrivateIE_Field_sequence[] = {
  { &hf_nbap_id_01          , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_PrivateIE_ID },
  { &hf_nbap_criticality    , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Criticality },
  { &hf_nbap_private_value  , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_PrivateIE_Field_value },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PrivateIE_Field(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PrivateIE_Field, PrivateIE_Field_sequence);

  return offset;
}


static const per_sequence_t PrivateIE_Container_sequence_of[1] = {
  { &hf_nbap_PrivateIE_Container_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_PrivateIE_Field },
};

static int
dissect_nbap_PrivateIE_Container(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_PrivateIE_Container, PrivateIE_Container_sequence_of,
                                                  1, maxPrivateIEs, FALSE);

  return offset;
}



static int
dissect_nbap_AckNack_RepetitionFactor(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 4U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_Ack_Power_Offset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 8U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_Acknowledged_PRACH_preambles_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 240U, NULL, TRUE);

  return offset;
}


static const value_string nbap_Uu_ActivationState_vals[] = {
  {   0, "activated" },
  {   1, "de-activated" },
  { 0, NULL }
};


static int
dissect_nbap_Uu_ActivationState(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t ActivationInformationItem_sequence[] = {
  { &hf_nbap_uU_ActivationState, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Uu_ActivationState },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ActivationInformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ActivationInformationItem, ActivationInformationItem_sequence);

  return offset;
}


static const per_sequence_t ActivationInformation_sequence_of[1] = {
  { &hf_nbap_ActivationInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ActivationInformationItem },
};

static int
dissect_nbap_ActivationInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_ActivationInformation, ActivationInformation_sequence_of,
                                                  1, maxNrOfEDCH_1, FALSE);

  return offset;
}


static const value_string nbap_Multicell_EDCH_Transport_Bearer_Mode_vals[] = {
  {   0, "separate-Iub-Transport-Bearer-Mode" },
  {   1, "uL-Flow-Multiplexing-Mode" },
  { 0, NULL }
};


static int
dissect_nbap_Multicell_EDCH_Transport_Bearer_Mode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_UL_ScramblingCodeNumber(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  guint32 ul_scrambling_code;
  guint32 com_context_id;
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 16777215U, &ul_scrambling_code, FALSE);

  nbap_get_private_data(actx->pinfo)->ul_scrambling_code = ul_scrambling_code;
  com_context_id = nbap_get_private_data(actx->pinfo)->com_context_id;
  if(ul_scrambling_code != 0 && com_context_id != 0) {
    wmem_tree_insert32(nbap_scrambling_code_crncc_map,ul_scrambling_code,GUINT_TO_POINTER(com_context_id));
  }


  return offset;
}


static const value_string nbap_UL_ScramblingCodeLength_vals[] = {
  {   0, "short" },
  {   1, "long" },
  { 0, NULL }
};


static int
dissect_nbap_UL_ScramblingCodeLength(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t UL_ScramblingCode_sequence[] = {
  { &hf_nbap_uL_ScramblingCodeNumber, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_ScramblingCodeNumber },
  { &hf_nbap_uL_ScramblingCodeLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_ScramblingCodeLength },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_ScramblingCode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_ScramblingCode, UL_ScramblingCode_sequence);

  return offset;
}



static int
dissect_nbap_UL_SIR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -82, 173U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_UL_DPCH_Information_Setup_sequence[] = {
  { &hf_nbap_ul_ScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_ScramblingCode },
  { &hf_nbap_ul_SIR_Target  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_SIR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_UL_DPCH_Information_Setup(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_UL_DPCH_Information_Setup, Additional_EDCH_UL_DPCH_Information_Setup_sequence);

  return offset;
}



static int
dissect_nbap_RL_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 31U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_C_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 65535U, NULL, FALSE);

  return offset;
}


static const value_string nbap_FirstRLS_Indicator_vals[] = {
  {   0, "first-RLS" },
  {   1, "not-first-RLS" },
  { 0, NULL }
};


static int
dissect_nbap_FirstRLS_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_PropagationDelay(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_DL_ScramblingCode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_FDD_DL_ChannelisationCodeNumber(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 511U, NULL, FALSE);

  return offset;
}


static const value_string nbap_TransmissionGapPatternSequenceCodeInformation_vals[] = {
  {   0, "code-change" },
  {   1, "nocode-change" },
  { 0, NULL }
};


static int
dissect_nbap_TransmissionGapPatternSequenceCodeInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t FDD_DL_CodeInformationItem_sequence[] = {
  { &hf_nbap_dl_ScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_ScramblingCode },
  { &hf_nbap_fdd_DL_ChannelisationCodeNumber, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_DL_ChannelisationCodeNumber },
  { &hf_nbap_transmissionGapPatternSequenceCodeInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransmissionGapPatternSequenceCodeInformation },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_FDD_DL_CodeInformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_FDD_DL_CodeInformationItem, FDD_DL_CodeInformationItem_sequence);

  return offset;
}


static const per_sequence_t FDD_DL_CodeInformation_sequence_of[1] = {
  { &hf_nbap_FDD_DL_CodeInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_DL_CodeInformationItem },
};

static int
dissect_nbap_FDD_DL_CodeInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_FDD_DL_CodeInformation, FDD_DL_CodeInformation_sequence_of,
                                                  1, maxNrOfCodes, FALSE);

  return offset;
}



static int
dissect_nbap_DL_Power(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -350, 150U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_F_DPCH_SlotFormat(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 9U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_E_RNTI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 65535U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_Multicell_EDCH_RL_Specific_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t Additional_EDCH_RL_Specific_Information_To_Setup_ItemIEs_sequence[] = {
  { &hf_nbap_eDCH_Additional_RL_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_c_ID           , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_C_ID },
  { &hf_nbap_firstRLS_indicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FirstRLS_Indicator },
  { &hf_nbap_propagationDelay, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PropagationDelay },
  { &hf_nbap_dl_CodeInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_DL_CodeInformation },
  { &hf_nbap_initialDL_transmissionPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_maximumDL_power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_minimumDL_power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_f_DPCH_SlotFormat, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_F_DPCH_SlotFormat },
  { &hf_nbap_e_RNTI         , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_RNTI },
  { &hf_nbap_multicell_EDCH_RL_Specific_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Multicell_EDCH_RL_Specific_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_RL_Specific_Information_To_Setup_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_RL_Specific_Information_To_Setup_ItemIEs, Additional_EDCH_RL_Specific_Information_To_Setup_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_RL_Specific_Information_To_Setup_List_sequence_of[1] = {
  { &hf_nbap_Additional_EDCH_RL_Specific_Information_To_Setup_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_EDCH_RL_Specific_Information_To_Setup_ItemIEs },
};

static int
dissect_nbap_Additional_EDCH_RL_Specific_Information_To_Setup_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Additional_EDCH_RL_Specific_Information_To_Setup_List, Additional_EDCH_RL_Specific_Information_To_Setup_List_sequence_of,
                                                  1, maxNrOfEDCHRLs, FALSE);

  return offset;
}



static int
dissect_nbap_E_DCH_MACdFlow_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfEDCHMACdFlows_1, &nbap_get_private_data(actx->pinfo)->e_dch_macdflow_id, FALSE);

  return offset;
}



static int
dissect_nbap_BindingID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  tvbuff_t *parameter_tvb=NULL;
  guint16 binding_id_port;

  offset = dissect_per_octet_string(tvb, offset, actx, tree, hf_index,
                                       1, 4, TRUE, &parameter_tvb);


  if (!parameter_tvb)
    return offset;

  if(tvb_reported_length(parameter_tvb)>=2){
    binding_id_port = tvb_get_ntohs(parameter_tvb,0);
    nbap_get_private_data(actx->pinfo)->binding_id_port = binding_id_port;
    proto_item_append_text(actx->created_item, " (%u)",binding_id_port);
  }


  return offset;
}



static int
dissect_nbap_TransportLayerAddress(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  tvbuff_t *parameter_tvb=NULL;
  proto_item *item;
  proto_tree *subtree, *nsap_tree;
  guint8 *padded_nsap_bytes;
  tvbuff_t *nsap_tvb;
  gint tvb_len;
  nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);

  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     1, 160, TRUE, NULL, 0, &parameter_tvb, NULL);

  if (!parameter_tvb)
    return offset;
  /* Get the length */
  tvb_len = tvb_reported_length(parameter_tvb);
  subtree = proto_item_add_subtree(actx->created_item, ett_nbap_TransportLayerAddress);
  if (tvb_len==4){
    /* IPv4 */
    proto_tree_add_item(subtree, hf_nbap_transportLayerAddress_ipv4, parameter_tvb, 0, tvb_len, ENC_BIG_ENDIAN);
    nbap_private_data->transportLayerAddress_ipv4 = tvb_get_ipv4(parameter_tvb, 0);
  }
  if (tvb_len==16){
    /* IPv6 */
    proto_tree_add_item(subtree, hf_nbap_transportLayerAddress_ipv6, parameter_tvb, 0, tvb_len, ENC_NA);
  }
  if (tvb_len == 20 || tvb_len == 7){
    /* NSAP */
    if (tvb_len == 7){
      /* Unpadded IPv4 NSAP */
      /* Creating a new TVB with padding */
      padded_nsap_bytes = (guint8*) wmem_alloc0(actx->pinfo->pool, 20);
      tvb_memcpy(parameter_tvb, padded_nsap_bytes, 0, tvb_len);
      nsap_tvb = tvb_new_child_real_data(tvb, padded_nsap_bytes, 20, 20);
      add_new_data_source(actx->pinfo, nsap_tvb, "Padded NSAP Data");
    }else{
      /* Padded NSAP*/
      nsap_tvb = parameter_tvb;
    }
    item = proto_tree_add_item(subtree, hf_nbap_transportLayerAddress_nsap, parameter_tvb, 0, tvb_len, ENC_NA);
    nsap_tree = proto_item_add_subtree(item, ett_nbap_TransportLayerAddress_nsap);
    if(tvb_get_ntoh24(parameter_tvb,0) == 0x350001){
      /* IPv4 */
      nbap_private_data->transportLayerAddress_ipv4 = tvb_get_ipv4(parameter_tvb, 3);
    }
    dissect_nsap(nsap_tvb, 0, 20, nsap_tree);
  }


  return offset;
}


static const per_sequence_t Additional_EDCH_MAC_d_Flows_Specific_Info_sequence[] = {
  { &hf_nbap_e_DCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_ID },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_MAC_d_Flows_Specific_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_MAC_d_Flows_Specific_Info, Additional_EDCH_MAC_d_Flows_Specific_Info_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_MAC_d_Flows_Specific_Info_List_sequence_of[1] = {
  { &hf_nbap_Additional_EDCH_MAC_d_Flows_Specific_Info_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_EDCH_MAC_d_Flows_Specific_Info },
};

static int
dissect_nbap_Additional_EDCH_MAC_d_Flows_Specific_Info_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Additional_EDCH_MAC_d_Flows_Specific_Info_List, Additional_EDCH_MAC_d_Flows_Specific_Info_List_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}



static int
dissect_nbap_HARQ_Process_Allocation_2ms_EDCH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     maxNrOfEDCHHARQProcesses2msEDCH, maxNrOfEDCHHARQProcesses2msEDCH, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_E_DCH_Maximum_Bitrate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 5742U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_E_DCH_Processing_Overload_Level(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 10U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_E_TFCI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_FDD_Information_sequence[] = {
  { &hf_nbap_additional_EDCH_MAC_d_Flows_Specific_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Additional_EDCH_MAC_d_Flows_Specific_Info_List },
  { &hf_nbap_hARQ_Process_Allocation_Scheduled_2ms_EDCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HARQ_Process_Allocation_2ms_EDCH },
  { &hf_nbap_e_DCH_Maximum_Bitrate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_Maximum_Bitrate },
  { &hf_nbap_e_DCH_Processing_Overload_Level, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_Processing_Overload_Level },
  { &hf_nbap_e_DCH_Min_Set_E_TFCI, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_TFCI },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_FDD_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_FDD_Information, Additional_EDCH_FDD_Information_sequence);

  return offset;
}


static const value_string nbap_FDD_TPC_DownlinkStepSize_vals[] = {
  {   0, "step-size0-5" },
  {   1, "step-size1" },
  {   2, "step-size1-5" },
  {   3, "step-size2" },
  { 0, NULL }
};


static int
dissect_nbap_FDD_TPC_DownlinkStepSize(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_LimitedPowerIncrease_vals[] = {
  {   0, "used" },
  {   1, "not-used" },
  { 0, NULL }
};


static int
dissect_nbap_LimitedPowerIncrease(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_InnerLoopDLPCStatus_vals[] = {
  {   0, "active" },
  {   1, "inactive" },
  { 0, NULL }
};


static int
dissect_nbap_InnerLoopDLPCStatus(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t Additional_EDCH_F_DPCH_Information_sequence[] = {
  { &hf_nbap_fdd_TPC_DownlinkStepSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_TPC_DownlinkStepSize },
  { &hf_nbap_limitedPowerIncrease, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_LimitedPowerIncrease },
  { &hf_nbap_innerLoopDLPCStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_InnerLoopDLPCStatus },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_F_DPCH_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_F_DPCH_Information, Additional_EDCH_F_DPCH_Information_sequence);

  return offset;
}



static int
dissect_nbap_Multicell_EDCH_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t Additional_EDCH_FDD_Setup_Cell_Information_sequence[] = {
  { &hf_nbap_additional_EDCH_UL_DPCH_Information_Setup, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_EDCH_UL_DPCH_Information_Setup },
  { &hf_nbap_additional_EDCH_RL_Specific_Information_To_Setup, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_EDCH_RL_Specific_Information_To_Setup_List },
  { &hf_nbap_additional_EDCH_FDD_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Additional_EDCH_FDD_Information },
  { &hf_nbap_additional_EDCH_F_DPCH_Information_Setup, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_EDCH_F_DPCH_Information },
  { &hf_nbap_multicell_EDCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Multicell_EDCH_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_FDD_Setup_Cell_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_FDD_Setup_Cell_Information, Additional_EDCH_FDD_Setup_Cell_Information_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_Setup_sequence_of[1] = {
  { &hf_nbap_Additional_EDCH_Cell_Information_Setup_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_EDCH_FDD_Setup_Cell_Information },
};

static int
dissect_nbap_Additional_EDCH_Cell_Information_Setup(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Additional_EDCH_Cell_Information_Setup, Additional_EDCH_Cell_Information_Setup_sequence_of,
                                                  1, maxNrOfEDCH_1, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_Setup_Info_sequence[] = {
  { &hf_nbap_multicell_EDCH_Transport_Bearer_Mode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Multicell_EDCH_Transport_Bearer_Mode },
  { &hf_nbap_additional_EDCH_Cell_Information_Setup, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_EDCH_Cell_Information_Setup },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_Setup_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_Setup_Info, Additional_EDCH_Setup_Info_sequence);

  return offset;
}


static const per_sequence_t EDCH_Additional_RL_Specific_Information_To_Add_List_sequence[] = {
  { &hf_nbap_eDCH_Additional_RL_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_c_ID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_C_ID },
  { &hf_nbap_dl_CodeInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_DL_CodeInformation },
  { &hf_nbap_initialDL_transmissionPower, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_maximumDL_power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_minimumDL_power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_f_DPCH_SlotFormat, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_F_DPCH_SlotFormat },
  { &hf_nbap_multicell_EDCH_RL_Specific_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Multicell_EDCH_RL_Specific_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_EDCH_Additional_RL_Specific_Information_To_Add_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_EDCH_Additional_RL_Specific_Information_To_Add_List, EDCH_Additional_RL_Specific_Information_To_Add_List_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_RL_Specific_Information_To_Add_ItemIEs_sequence_of[1] = {
  { &hf_nbap_Additional_EDCH_RL_Specific_Information_To_Add_ItemIEs_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_EDCH_Additional_RL_Specific_Information_To_Add_List },
};

static int
dissect_nbap_Additional_EDCH_RL_Specific_Information_To_Add_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Additional_EDCH_RL_Specific_Information_To_Add_ItemIEs, Additional_EDCH_RL_Specific_Information_To_Add_ItemIEs_sequence_of,
                                                  1, maxNrOfEDCHRLs, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_To_Add_ItemIEs_sequence[] = {
  { &hf_nbap_additional_EDCH_RL_Specific_Information_To_Add_ItemIEs, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_EDCH_RL_Specific_Information_To_Add_ItemIEs },
  { &hf_nbap_additional_EDCH_FDD_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Additional_EDCH_FDD_Information },
  { &hf_nbap_multicell_EDCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Multicell_EDCH_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_Cell_Information_To_Add_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_Cell_Information_To_Add_ItemIEs, Additional_EDCH_Cell_Information_To_Add_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_To_Add_List_sequence_of[1] = {
  { &hf_nbap_Additional_EDCH_Cell_Information_To_Add_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_EDCH_Cell_Information_To_Add_ItemIEs },
};

static int
dissect_nbap_Additional_EDCH_Cell_Information_To_Add_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Additional_EDCH_Cell_Information_To_Add_List, Additional_EDCH_Cell_Information_To_Add_List_sequence_of,
                                                  1, maxNrOfEDCH_1, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_RL_Specific_Information_To_Modify_ItemIEs_sequence[] = {
  { &hf_nbap_eDCH_Additional_RL_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_dl_CodeInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_FDD_DL_CodeInformation },
  { &hf_nbap_maximumDL_power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_minimumDL_power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_f_DPCH_SlotFormat, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_F_DPCH_SlotFormat },
  { &hf_nbap_multicell_EDCH_RL_Specific_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Multicell_EDCH_RL_Specific_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_RL_Specific_Information_To_Modify_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_RL_Specific_Information_To_Modify_ItemIEs, Additional_EDCH_RL_Specific_Information_To_Modify_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_RL_Specific_Information_To_Modify_List_sequence_of[1] = {
  { &hf_nbap_Additional_EDCH_RL_Specific_Information_To_Modify_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_EDCH_RL_Specific_Information_To_Modify_ItemIEs },
};

static int
dissect_nbap_Additional_EDCH_RL_Specific_Information_To_Modify_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Additional_EDCH_RL_Specific_Information_To_Modify_List, Additional_EDCH_RL_Specific_Information_To_Modify_List_sequence_of,
                                                  1, maxNrOfEDCHRLs, FALSE);

  return offset;
}



static int
dissect_nbap_Received_total_wide_band_power_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 621U, NULL, FALSE);

  return offset;
}


static const value_string nbap_DL_PowerBalancing_ActivationIndicator_vals[] = {
  {   0, "dL-PowerBalancing-Activated" },
  { 0, NULL }
};


static int
dissect_nbap_DL_PowerBalancing_ActivationIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_RL_Set_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 31U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_E_RGCH_Signature_Sequence(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrofSigSeqRGHI_1, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_E_HICH_Signature_Sequence(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrofSigSeqRGHI_1, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_E_Serving_Grant_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 38U, NULL, FALSE);

  return offset;
}


static const value_string nbap_E_Primary_Secondary_Grant_Selector_vals[] = {
  {   0, "primary" },
  {   1, "secondary" },
  { 0, NULL }
};


static int
dissect_nbap_E_Primary_Secondary_Grant_Selector(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_E_RGCH_Release_Indicator_vals[] = {
  {   0, "e-RGCHreleased" },
  { 0, NULL }
};


static int
dissect_nbap_E_RGCH_Release_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t E_DCH_FDD_DL_Control_Channel_Information_sequence[] = {
  { &hf_nbap_e_AGCH_And_E_RGCH_E_HICH_FDD_Scrambling_Code, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_ScramblingCode },
  { &hf_nbap_e_AGCH_Channelisation_Code, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_FDD_DL_ChannelisationCodeNumber },
  { &hf_nbap_primary_e_RNTI , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_RNTI },
  { &hf_nbap_secondary_e_RNTI, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_RNTI },
  { &hf_nbap_e_RGCH_E_HICH_Channelisation_Code, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_FDD_DL_ChannelisationCodeNumber },
  { &hf_nbap_e_RGCH_Signature_Sequence, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_RGCH_Signature_Sequence },
  { &hf_nbap_e_HICH_Signature_Sequence, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_HICH_Signature_Sequence },
  { &hf_nbap_serving_Grant_Value, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_Serving_Grant_Value },
  { &hf_nbap_primary_Secondary_Grant_Selector, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_Primary_Secondary_Grant_Selector },
  { &hf_nbap_e_RGCH_Release_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_RGCH_Release_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_FDD_DL_Control_Channel_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_FDD_DL_Control_Channel_Information, E_DCH_FDD_DL_Control_Channel_Information_sequence);

  return offset;
}


static const per_sequence_t EDCH_Additional_RL_Specific_Information_Response_ItemIEs_sequence[] = {
  { &hf_nbap_eDCH_Additional_RL_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_received_total_wide_band_power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Received_total_wide_band_power_Value },
  { &hf_nbap_dL_PowerBalancing_ActivationIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_PowerBalancing_ActivationIndicator },
  { &hf_nbap_rL_Set_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_Set_ID },
  { &hf_nbap_e_DCH_RL_Set_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_Set_ID },
  { &hf_nbap_e_DCH_FDD_DL_Control_Channel_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_FDD_DL_Control_Channel_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_EDCH_Additional_RL_Specific_Information_Response_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_EDCH_Additional_RL_Specific_Information_Response_ItemIEs, EDCH_Additional_RL_Specific_Information_Response_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t EDCH_Additional_RL_Specific_Information_Response_List_sequence_of[1] = {
  { &hf_nbap_EDCH_Additional_RL_Specific_Information_Response_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_EDCH_Additional_RL_Specific_Information_Response_ItemIEs },
};

static int
dissect_nbap_EDCH_Additional_RL_Specific_Information_Response_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_EDCH_Additional_RL_Specific_Information_Response_List, EDCH_Additional_RL_Specific_Information_Response_List_sequence_of,
                                                  1, maxNrOfEDCHRLs, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_MAC_d_Flows_Specific_Info_Response_sequence[] = {
  { &hf_nbap_e_DCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_ID },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_MAC_d_Flows_Specific_Info_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_MAC_d_Flows_Specific_Info_Response, Additional_EDCH_MAC_d_Flows_Specific_Info_Response_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List_sequence_of[1] = {
  { &hf_nbap_Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_EDCH_MAC_d_Flows_Specific_Info_Response },
};

static int
dissect_nbap_Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List, Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_FDD_Information_Response_ItemIEs_sequence[] = {
  { &hf_nbap_eDCH_Additional_RL_Specific_Information_Response, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_EDCH_Additional_RL_Specific_Information_Response_List },
  { &hf_nbap_additional_EDCH_MAC_d_Flow_Specific_Information_Response, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List },
  { &hf_nbap_hARQ_Process_Allocation_Scheduled_2ms_EDCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HARQ_Process_Allocation_2ms_EDCH },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_FDD_Information_Response_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_FDD_Information_Response_ItemIEs, Additional_EDCH_FDD_Information_Response_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_Response_List_sequence_of[1] = {
  { &hf_nbap_Additional_EDCH_Cell_Information_Response_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_EDCH_FDD_Information_Response_ItemIEs },
};

static int
dissect_nbap_Additional_EDCH_Cell_Information_Response_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Additional_EDCH_Cell_Information_Response_List, Additional_EDCH_Cell_Information_Response_List_sequence_of,
                                                  1, maxNrOfEDCH_1, FALSE);

  return offset;
}


static const value_string nbap_DL_PowerBalancing_UpdatedIndicator_vals[] = {
  {   0, "dL-PowerBalancing-Updated" },
  { 0, NULL }
};


static int
dissect_nbap_DL_PowerBalancing_UpdatedIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t EDCH_Additional_Modified_RL_Specific_Information_Response_List_Items_sequence[] = {
  { &hf_nbap_eDCH_Additional_RL_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_dL_PowerBalancing_UpdatedIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_PowerBalancing_UpdatedIndicator },
  { &hf_nbap_e_DCH_FDD_DL_Control_Channel_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_FDD_DL_Control_Channel_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_EDCH_Additional_Modified_RL_Specific_Information_Response_List_Items(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_EDCH_Additional_Modified_RL_Specific_Information_Response_List_Items, EDCH_Additional_Modified_RL_Specific_Information_Response_List_Items_sequence);

  return offset;
}


static const per_sequence_t EDCH_Additional_Modified_RL_Specific_Information_Response_List_sequence_of[1] = {
  { &hf_nbap_EDCH_Additional_Modified_RL_Specific_Information_Response_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_EDCH_Additional_Modified_RL_Specific_Information_Response_List_Items },
};

static int
dissect_nbap_EDCH_Additional_Modified_RL_Specific_Information_Response_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_EDCH_Additional_Modified_RL_Specific_Information_Response_List, EDCH_Additional_Modified_RL_Specific_Information_Response_List_sequence_of,
                                                  1, maxNrOfEDCHRLs, FALSE);

  return offset;
}


static const per_sequence_t Additional_Modififed_EDCH_FDD_Information_Response_ItemIEs_sequence[] = {
  { &hf_nbap_eDCH_Additional_Modified_RL_Specific_Information_Response, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_EDCH_Additional_Modified_RL_Specific_Information_Response_List },
  { &hf_nbap_additional_EDCH_MAC_d_Flow_Specific_Information_Response, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List },
  { &hf_nbap_hARQ_Process_Allocation_Scheduled_2ms_EDCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HARQ_Process_Allocation_2ms_EDCH },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_Modififed_EDCH_FDD_Information_Response_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_Modififed_EDCH_FDD_Information_Response_ItemIEs, Additional_Modififed_EDCH_FDD_Information_Response_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_FDD_Information_Response_RLReconf_Items_sequence[] = {
  { &hf_nbap_additional_EDCH_FDD_Information_Response_ItemIEs, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Additional_EDCH_FDD_Information_Response_ItemIEs },
  { &hf_nbap_additional_Modififed_EDCH_FDD_Information_Response_ItemIEs, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Additional_Modififed_EDCH_FDD_Information_Response_ItemIEs },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_FDD_Information_Response_RLReconf_Items(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_FDD_Information_Response_RLReconf_Items, Additional_EDCH_FDD_Information_Response_RLReconf_Items_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_Response_RLReconf_List_sequence_of[1] = {
  { &hf_nbap_Additional_EDCH_Cell_Information_Response_RLReconf_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_EDCH_FDD_Information_Response_RLReconf_Items },
};

static int
dissect_nbap_Additional_EDCH_Cell_Information_Response_RLReconf_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Additional_EDCH_Cell_Information_Response_RLReconf_List, Additional_EDCH_Cell_Information_Response_RLReconf_List_sequence_of,
                                                  1, maxNrOfEDCH_1, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_RL_InformationList_Rsp_Item_sequence[] = {
  { &hf_nbap_rl_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_e_DCH_FDD_DL_Control_Channel_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_FDD_DL_Control_Channel_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_RL_InformationList_Rsp_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_RL_InformationList_Rsp_Item, E_DCH_RL_InformationList_Rsp_Item_sequence);

  return offset;
}


static const per_sequence_t E_DCH_RL_InformationList_Rsp_sequence_of[1] = {
  { &hf_nbap_E_DCH_RL_InformationList_Rsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_RL_InformationList_Rsp_Item },
};

static int
dissect_nbap_E_DCH_RL_InformationList_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_DCH_RL_InformationList_Rsp, E_DCH_RL_InformationList_Rsp_sequence_of,
                                                  0, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_serving_cell_change_successful_sequence[] = {
  { &hf_nbap_e_DCH_RL_InformationList_Rsp, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_RL_InformationList_Rsp },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_serving_cell_change_successful(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_serving_cell_change_successful, E_DCH_serving_cell_change_successful_sequence);

  return offset;
}


static const value_string nbap_CauseRadioNetwork_vals[] = {
  {   0, "unknown-C-ID" },
  {   1, "cell-not-available" },
  {   2, "power-level-not-supported" },
  {   3, "dl-radio-resources-not-available" },
  {   4, "ul-radio-resources-not-available" },
  {   5, "rl-already-ActivatedOrAllocated" },
  {   6, "nodeB-Resources-unavailable" },
  {   7, "measurement-not-supported-for-the-object" },
  {   8, "combining-resources-not-available" },
  {   9, "requested-configuration-not-supported" },
  {  10, "synchronisation-failure" },
  {  11, "priority-transport-channel-established" },
  {  12, "sIB-Origination-in-Node-B-not-Supported" },
  {  13, "requested-tx-diversity-mode-not-supported" },
  {  14, "unspecified" },
  {  15, "bCCH-scheduling-error" },
  {  16, "measurement-temporarily-not-available" },
  {  17, "invalid-CM-settings" },
  {  18, "reconfiguration-CFN-not-elapsed" },
  {  19, "number-of-DL-codes-not-supported" },
  {  20, "s-cpich-not-supported" },
  {  21, "combining-not-supported" },
  {  22, "ul-sf-not-supported" },
  {  23, "dl-SF-not-supported" },
  {  24, "common-transport-channel-type-not-supported" },
  {  25, "dedicated-transport-channel-type-not-supported" },
  {  26, "downlink-shared-channel-type-not-supported" },
  {  27, "uplink-shared-channel-type-not-supported" },
  {  28, "cm-not-supported" },
  {  29, "tx-diversity-no-longer-supported" },
  {  30, "unknown-Local-Cell-ID" },
  {  31, "number-of-UL-codes-not-supported" },
  {  32, "information-temporarily-not-available" },
  {  33, "information-provision-not-supported-for-the-object" },
  {  34, "cell-synchronisation-not-supported" },
  {  35, "cell-synchronisation-adjustment-not-supported" },
  {  36, "dpc-mode-change-not-supported" },
  {  37, "iPDL-already-activated" },
  {  38, "iPDL-not-supported" },
  {  39, "iPDL-parameters-not-available" },
  {  40, "frequency-acquisition-not-supported" },
  {  41, "power-balancing-status-not-compatible" },
  {  42, "requested-typeofbearer-re-arrangement-not-supported" },
  {  43, "signalling-Bearer-Re-arrangement-not-supported" },
  {  44, "bearer-Re-arrangement-needed" },
  {  45, "delayed-activation-not-supported" },
  {  46, "rl-timing-adjustment-not-supported" },
  {  47, "mich-not-supported" },
  {  48, "f-DPCH-not-supported" },
  {  49, "modification-period-not-available" },
  {  50, "pLCCH-not-supported" },
  {  51, "continuous-packet-connectivity-DTX-DRX-operation-not-available" },
  {  52, "continuous-packet-connectivity-UE-DTX-Cycle-not-available" },
  {  53, "mIMO-not-available" },
  {  54, "e-DCH-MACdPDU-SizeFormat-not-available" },
  {  55, "multi-Cell-operation-not-available" },
  {  56, "semi-Persistent-scheduling-not-supported" },
  {  57, "continuous-Packet-Connectivity-DRX-not-supported" },
  {  58, "continuous-Packet-Connectivity-DRX-not-available" },
  {  59, "sixtyfourQAM-DL-and-MIMO-Combined-not-available" },
  {  60, "s-cpich-power-offset-not-available" },
  {  61, "tx-diversity-for-mimo-on-DL-control-channels-not-available" },
  {  62, "single-Stream-MIMO-not-available" },
  {  63, "multi-Cell-operation-with-MIMO-not-available" },
  {  64, "multi-Cell-operation-with-Single-Stream-MIMO-not-available" },
  {  65, "cellSpecificTxDiversityHandlingForMultiCellOperationNotAvailable" },
  {  66, "multi-Cell-EDCH-operation-not-available" },
  { 0, NULL }
};


static int
dissect_nbap_CauseRadioNetwork(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     31, NULL, TRUE, 36, NULL);

  return offset;
}


static const value_string nbap_CauseTransport_vals[] = {
  {   0, "transport-resource-unavailable" },
  {   1, "unspecified" },
  { 0, NULL }
};


static int
dissect_nbap_CauseTransport(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_CauseProtocol_vals[] = {
  {   0, "transfer-syntax-error" },
  {   1, "abstract-syntax-error-reject" },
  {   2, "abstract-syntax-error-ignore-and-notify" },
  {   3, "message-not-compatible-with-receiver-state" },
  {   4, "semantic-error" },
  {   5, "unspecified" },
  {   6, "abstract-syntax-error-falsely-constructed-message" },
  { 0, NULL }
};


static int
dissect_nbap_CauseProtocol(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     7, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_CauseMisc_vals[] = {
  {   0, "control-processing-overload" },
  {   1, "hardware-failure" },
  {   2, "oam-intervention" },
  {   3, "not-enough-user-plane-processing-resources" },
  {   4, "unspecified" },
  { 0, NULL }
};


static int
dissect_nbap_CauseMisc(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     5, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_Cause_vals[] = {
  {   0, "radioNetwork" },
  {   1, "transport" },
  {   2, "protocol" },
  {   3, "misc" },
  { 0, NULL }
};

static const per_choice_t Cause_choice[] = {
  {   0, &hf_nbap_radioNetwork   , ASN1_EXTENSION_ROOT    , dissect_nbap_CauseRadioNetwork },
  {   1, &hf_nbap_transport      , ASN1_EXTENSION_ROOT    , dissect_nbap_CauseTransport },
  {   2, &hf_nbap_protocol       , ASN1_EXTENSION_ROOT    , dissect_nbap_CauseProtocol },
  {   3, &hf_nbap_misc           , ASN1_EXTENSION_ROOT    , dissect_nbap_CauseMisc },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_Cause(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_Cause, Cause_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t E_DCH_serving_cell_change_unsuccessful_sequence[] = {
  { &hf_nbap_cause          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Cause },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_serving_cell_change_unsuccessful(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_serving_cell_change_unsuccessful, E_DCH_serving_cell_change_unsuccessful_sequence);

  return offset;
}


static const value_string nbap_E_DCH_serving_cell_choice_vals[] = {
  {   0, "e-DCH-serving-cell-change-successful" },
  {   1, "e-DCH-serving-cell-change-unsuccessful" },
  { 0, NULL }
};

static const per_choice_t E_DCH_serving_cell_choice_choice[] = {
  {   0, &hf_nbap_e_DCH_serving_cell_change_successful, ASN1_EXTENSION_ROOT    , dissect_nbap_E_DCH_serving_cell_change_successful },
  {   1, &hf_nbap_e_DCH_serving_cell_change_unsuccessful, ASN1_EXTENSION_ROOT    , dissect_nbap_E_DCH_serving_cell_change_unsuccessful },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_E_DCH_serving_cell_choice(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_E_DCH_serving_cell_choice, E_DCH_serving_cell_choice_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t E_DCH_Serving_Cell_Change_Info_Response_sequence[] = {
  { &hf_nbap_e_DCH_serving_cell_choice, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_serving_cell_choice },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_Serving_Cell_Change_Info_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_Serving_Cell_Change_Info_Response, E_DCH_Serving_Cell_Change_Info_Response_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_Response_RL_Add_ItemIEs_sequence[] = {
  { &hf_nbap_additional_EDCH_FDD_Information_Response, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Additional_EDCH_FDD_Information_Response_ItemIEs },
  { &hf_nbap_additional_EDCH_Serving_Cell_Change_Information_Response, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_Serving_Cell_Change_Info_Response },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_Cell_Information_Response_RL_Add_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_Cell_Information_Response_RL_Add_ItemIEs, Additional_EDCH_Cell_Information_Response_RL_Add_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_Response_RL_Add_List_sequence_of[1] = {
  { &hf_nbap_Additional_EDCH_Cell_Information_Response_RL_Add_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_EDCH_Cell_Information_Response_RL_Add_ItemIEs },
};

static int
dissect_nbap_Additional_EDCH_Cell_Information_Response_RL_Add_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Additional_EDCH_Cell_Information_Response_RL_Add_List, Additional_EDCH_Cell_Information_Response_RL_Add_List_sequence_of,
                                                  1, maxNrOfEDCH_1, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_UL_DPCH_Information_Modify_sequence[] = {
  { &hf_nbap_ul_ScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_ScramblingCode },
  { &hf_nbap_ul_SIR_Target  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_SIR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_UL_DPCH_Information_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_UL_DPCH_Information_Modify, Additional_EDCH_UL_DPCH_Information_Modify_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_ConfigurationChange_Info_ItemIEs_sequence[] = {
  { &hf_nbap_additional_EDCH_UL_DPCH_Information_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Additional_EDCH_UL_DPCH_Information_Modify },
  { &hf_nbap_additional_EDCH_RL_Specific_Information_To_Add, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Additional_EDCH_RL_Specific_Information_To_Add_ItemIEs },
  { &hf_nbap_additional_EDCH_RL_Specific_Information_To_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Additional_EDCH_RL_Specific_Information_To_Modify_List },
  { &hf_nbap_additional_EDCH_FDD_Information_To_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Additional_EDCH_FDD_Information },
  { &hf_nbap_additional_EDCH_F_DPCH_Information_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Additional_EDCH_F_DPCH_Information },
  { &hf_nbap_multicell_EDCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Multicell_EDCH_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_ConfigurationChange_Info_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_ConfigurationChange_Info_ItemIEs, Additional_EDCH_ConfigurationChange_Info_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_ConfigurationChange_List_sequence_of[1] = {
  { &hf_nbap_Additional_EDCH_Cell_Information_ConfigurationChange_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_EDCH_ConfigurationChange_Info_ItemIEs },
};

static int
dissect_nbap_Additional_EDCH_Cell_Information_ConfigurationChange_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Additional_EDCH_Cell_Information_ConfigurationChange_List, Additional_EDCH_Cell_Information_ConfigurationChange_List_sequence_of,
                                                  1, maxNrOfEDCH_1, FALSE);

  return offset;
}


static const value_string nbap_RL_on_Secondary_UL_Frequency_vals[] = {
  {   0, "remove" },
  { 0, NULL }
};


static int
dissect_nbap_RL_on_Secondary_UL_Frequency(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_Removal_Info_ItemIEs_sequence[] = {
  { &hf_nbap_rL_on_Secondary_UL_Frequency, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_on_Secondary_UL_Frequency },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_Cell_Information_Removal_Info_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_Cell_Information_Removal_Info_ItemIEs, Additional_EDCH_Cell_Information_Removal_Info_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_Removal_List_sequence_of[1] = {
  { &hf_nbap_Additional_EDCH_Cell_Information_Removal_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_EDCH_Cell_Information_Removal_Info_ItemIEs },
};

static int
dissect_nbap_Additional_EDCH_Cell_Information_Removal_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Additional_EDCH_Cell_Information_Removal_List, Additional_EDCH_Cell_Information_Removal_List_sequence_of,
                                                  1, maxNrOfEDCH_1, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_DL_Control_Channel_Change_Info_ItemIEs_sequence[] = {
  { &hf_nbap_eDCH_Additional_RL_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_DL_Control_Channel_Change_Info_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_DL_Control_Channel_Change_Info_ItemIEs, Additional_EDCH_DL_Control_Channel_Change_Info_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_DL_Control_Channel_Change_Information_List_sequence_of[1] = {
  { &hf_nbap_Additional_EDCH_DL_Control_Channel_Change_Information_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_EDCH_DL_Control_Channel_Change_Info_ItemIEs },
};

static int
dissect_nbap_Additional_EDCH_DL_Control_Channel_Change_Information_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Additional_EDCH_DL_Control_Channel_Change_Information_List, Additional_EDCH_DL_Control_Channel_Change_Information_List_sequence_of,
                                                  1, maxNrOfEDCHRLs, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_FDD_Update_Information_sequence[] = {
  { &hf_nbap_hARQ_Process_Allocation_Scheduled_2ms_EDCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HARQ_Process_Allocation_2ms_EDCH },
  { &hf_nbap_additional_EDCH_DL_Control_Channel_Change_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Additional_EDCH_DL_Control_Channel_Change_Information_List },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_FDD_Update_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_FDD_Update_Information, Additional_EDCH_FDD_Update_Information_sequence);

  return offset;
}



static int
dissect_nbap_UARFCN(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 16383U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_TimeSlotLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 6U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_Transmitted_Carrier_Power_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 100U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_UL_TimeslotISCP_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_NULL(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_Extension_CommonMeasurementValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string nbap_CommonMeasurementValue_vals[] = {
  {   0, "transmitted-carrier-power" },
  {   1, "received-total-wide-band-power" },
  {   2, "acknowledged-prach-preambles" },
  {   3, "uL-TimeslotISCP" },
  {   4, "notUsed-1-acknowledged-PCPCH-access-preambles" },
  {   5, "notUsed-2-detected-PCPCH-access-preambles" },
  {   6, "extension-CommonMeasurementValue" },
  { 0, NULL }
};

static const per_choice_t CommonMeasurementValue_choice[] = {
  {   0, &hf_nbap_transmitted_carrier_power, ASN1_EXTENSION_ROOT    , dissect_nbap_Transmitted_Carrier_Power_Value },
  {   1, &hf_nbap_received_total_wide_band_power, ASN1_EXTENSION_ROOT    , dissect_nbap_Received_total_wide_band_power_Value },
  {   2, &hf_nbap_acknowledged_prach_preambles, ASN1_EXTENSION_ROOT    , dissect_nbap_Acknowledged_PRACH_preambles_Value },
  {   3, &hf_nbap_uL_TimeslotISCP, ASN1_EXTENSION_ROOT    , dissect_nbap_UL_TimeslotISCP_Value },
  {   4, &hf_nbap_notUsed_1_acknowledged_PCPCH_access_preambles, ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {   5, &hf_nbap_notUsed_2_detected_PCPCH_access_preambles, ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {   6, &hf_nbap_extension_CommonMeasurementValue, ASN1_NOT_EXTENSION_ROOT, dissect_nbap_Extension_CommonMeasurementValue },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_CommonMeasurementValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_CommonMeasurementValue, CommonMeasurementValue_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t TimeSlotMeasurementValueLCR_sequence[] = {
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_commonMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonMeasurementValue },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TimeSlotMeasurementValueLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TimeSlotMeasurementValueLCR, TimeSlotMeasurementValueLCR_sequence);

  return offset;
}


static const per_sequence_t TimeSlotMeasurementValueListLCR_sequence_of[1] = {
  { &hf_nbap_TimeSlotMeasurementValueListLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotMeasurementValueLCR },
};

static int
dissect_nbap_TimeSlotMeasurementValueListLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_TimeSlotMeasurementValueListLCR, TimeSlotMeasurementValueListLCR_sequence_of,
                                                  1, 6, FALSE);

  return offset;
}


static const per_sequence_t AdditionalMeasurementValue_sequence[] = {
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UARFCN },
  { &hf_nbap_timeSlotMeasurementValueListLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotMeasurementValueListLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_AdditionalMeasurementValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_AdditionalMeasurementValue, AdditionalMeasurementValue_sequence);

  return offset;
}


static const per_sequence_t AdditionalMeasurementValueList_sequence_of[1] = {
  { &hf_nbap_AdditionalMeasurementValueList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_AdditionalMeasurementValue },
};

static int
dissect_nbap_AdditionalMeasurementValueList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_AdditionalMeasurementValueList, AdditionalMeasurementValueList_sequence_of,
                                                  1, maxFrequencyinCell_1, FALSE);

  return offset;
}


static const per_sequence_t TimeSlot_InitiatedListLCR_sequence_of[1] = {
  { &hf_nbap_TimeSlot_InitiatedListLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
};

static int
dissect_nbap_TimeSlot_InitiatedListLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_TimeSlot_InitiatedListLCR, TimeSlot_InitiatedListLCR_sequence_of,
                                                  0, 6, FALSE);

  return offset;
}


static const per_sequence_t AdditionalTimeSlotLCR_sequence[] = {
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UARFCN },
  { &hf_nbap_timeslot_InitiatedListLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TimeSlot_InitiatedListLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_AdditionalTimeSlotLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_AdditionalTimeSlotLCR, AdditionalTimeSlotLCR_sequence);

  return offset;
}


static const per_sequence_t AdditionalTimeSlotListLCR_sequence_of[1] = {
  { &hf_nbap_AdditionalTimeSlotListLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_AdditionalTimeSlotLCR },
};

static int
dissect_nbap_AdditionalTimeSlotListLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_AdditionalTimeSlotListLCR, AdditionalTimeSlotListLCR_sequence_of,
                                                  0, maxFrequencyinCell_1, FALSE);

  return offset;
}


static const value_string nbap_AddorDeleteIndicator_vals[] = {
  {   0, "add" },
  {   1, "delete" },
  { 0, NULL }
};


static int
dissect_nbap_AddorDeleteIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_CFN(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_TGPSID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxTGPS, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_TGPRC(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 511U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Transmission_Gap_Pattern_Sequence_Status_List_item_sequence[] = {
  { &hf_nbap_tGPSID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TGPSID },
  { &hf_nbap_tGPRC          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TGPRC },
  { &hf_nbap_tGCFN          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CFN },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Transmission_Gap_Pattern_Sequence_Status_List_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Transmission_Gap_Pattern_Sequence_Status_List_item, Transmission_Gap_Pattern_Sequence_Status_List_item_sequence);

  return offset;
}


static const per_sequence_t Transmission_Gap_Pattern_Sequence_Status_List_sequence_of[1] = {
  { &hf_nbap_Transmission_Gap_Pattern_Sequence_Status_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Transmission_Gap_Pattern_Sequence_Status_List_item },
};

static int
dissect_nbap_Transmission_Gap_Pattern_Sequence_Status_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Transmission_Gap_Pattern_Sequence_Status_List, Transmission_Gap_Pattern_Sequence_Status_List_sequence_of,
                                                  0, maxTGPS, FALSE);

  return offset;
}


static const per_sequence_t Active_Pattern_Sequence_Information_sequence[] = {
  { &hf_nbap_cMConfigurationChangeCFN, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CFN },
  { &hf_nbap_transmission_Gap_Pattern_Sequence_Status, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Transmission_Gap_Pattern_Sequence_Status_List },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Active_Pattern_Sequence_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Active_Pattern_Sequence_Information, Active_Pattern_Sequence_Information_sequence);

  return offset;
}



static int
dissect_nbap_AICH_Power(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -22, 5U, NULL, FALSE);

  return offset;
}


static const value_string nbap_AICH_TransmissionTiming_vals[] = {
  {   0, "v0" },
  {   1, "v1" },
  { 0, NULL }
};


static int
dissect_nbap_AICH_TransmissionTiming(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_PriorityLevel(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}


static const value_string nbap_Pre_emptionCapability_vals[] = {
  {   0, "shall-not-trigger-pre-emption" },
  {   1, "may-trigger-pre-emption" },
  { 0, NULL }
};


static int
dissect_nbap_Pre_emptionCapability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_Pre_emptionVulnerability_vals[] = {
  {   0, "not-pre-emptable" },
  {   1, "pre-emptable" },
  { 0, NULL }
};


static int
dissect_nbap_Pre_emptionVulnerability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t AllocationRetentionPriority_sequence[] = {
  { &hf_nbap_priorityLevel  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PriorityLevel },
  { &hf_nbap_pre_emptionCapability, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Pre_emptionCapability },
  { &hf_nbap_pre_emptionVulnerability, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Pre_emptionVulnerability },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_AllocationRetentionPriority(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_AllocationRetentionPriority, AllocationRetentionPriority_sequence);

  return offset;
}


static const value_string nbap_AlternativeFormatReportingIndicator_vals[] = {
  {   0, "alternativeFormatAllowed" },
  { 0, NULL }
};


static int
dissect_nbap_AlternativeFormatReportingIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_AOA_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 719U, NULL, FALSE);

  return offset;
}


static const value_string nbap_AOA_LCR_Accuracy_Class_vals[] = {
  {   0, "a" },
  {   1, "b" },
  {   2, "c" },
  {   3, "d" },
  {   4, "e" },
  {   5, "f" },
  {   6, "g" },
  {   7, "h" },
  { 0, NULL }
};


static int
dissect_nbap_AOA_LCR_Accuracy_Class(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t Angle_Of_Arrival_Value_LCR_sequence[] = {
  { &hf_nbap_aOA_LCR        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AOA_LCR },
  { &hf_nbap_aOA_LCR_Accuracy_Class, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AOA_LCR_Accuracy_Class },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Angle_Of_Arrival_Value_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Angle_Of_Arrival_Value_LCR, Angle_Of_Arrival_Value_LCR_sequence);

  return offset;
}


static const value_string nbap_AvailabilityStatus_vals[] = {
  {   0, "empty" },
  {   1, "in-test" },
  {   2, "failed" },
  {   3, "power-off" },
  {   4, "off-line" },
  {   5, "off-duty" },
  {   6, "dependency" },
  {   7, "degraded" },
  {   8, "not-installed" },
  {   9, "log-full" },
  { 0, NULL }
};


static int
dissect_nbap_AvailabilityStatus(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     10, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_HSDSCH_RNTI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  gint hrnti;
umts_fp_conversation_info_t *umts_fp_conversation_info = NULL;
fp_hsdsch_channel_info_t* fp_hsdsch_channel_info = NULL;
address     null_addr;
conversation_t   *conversation = NULL;
int i;
nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
nbap_hsdsch_channel_info_t* nbap_hsdsch_channel_info;
nbap_hsdsch_channel_info = nbap_private_data->nbap_hsdsch_channel_info;

  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 65535U, &hrnti, FALSE);

  nbap_private_data->hrnti = hrnti;

    if (PINFO_FD_VISITED(actx->pinfo)){
        return offset;
    }

    /*Find the conversations assoicated with the HS-DSCH flows in this packet and set proper H-RNTI*/
    clear_address(&null_addr);
    for (i = 0; i < maxNrOfMACdFlows; i++) {
        if (nbap_hsdsch_channel_info[i].crnc_port != 0){
            conversation = find_conversation(actx->pinfo->num, &(nbap_hsdsch_channel_info[i].crnc_address), &null_addr,
                               CONVERSATION_UDP, nbap_hsdsch_channel_info[i].crnc_port, 0, NO_ADDR_B);
            if(conversation != NULL){
                umts_fp_conversation_info = (umts_fp_conversation_info_t *)conversation_get_proto_data(conversation, proto_fp);
                DISSECTOR_ASSERT(umts_fp_conversation_info != NULL);
                fp_hsdsch_channel_info = (fp_hsdsch_channel_info_t*)umts_fp_conversation_info->channel_specific_info;
                DISSECTOR_ASSERT(fp_hsdsch_channel_info != NULL);
                fp_hsdsch_channel_info->hrnti = nbap_private_data->hrnti;
            }
         }
     }


  return offset;
}


static const per_sequence_t BCCH_Specific_HSDSCH_RNTI_Information_sequence[] = {
  { &hf_nbap_bCCH_Specific_HSDSCH_RNTI, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_RNTI },
  { &hf_nbap_hSSCCH_Power   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_hSPDSCH_Power  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_BCCH_Specific_HSDSCH_RNTI_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_BCCH_Specific_HSDSCH_RNTI_Information, BCCH_Specific_HSDSCH_RNTI_Information_sequence);

  return offset;
}


static const per_sequence_t BCCH_Specific_HSDSCH_RNTI_InformationLCR_sequence[] = {
  { &hf_nbap_bCCH_Specific_HSDSCH_RNTI, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_RNTI },
  { &hf_nbap_hSSCCH_Power   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_hSPDSCH_Power  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_BCCH_Specific_HSDSCH_RNTI_InformationLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_BCCH_Specific_HSDSCH_RNTI_InformationLCR, BCCH_Specific_HSDSCH_RNTI_InformationLCR_sequence);

  return offset;
}



static int
dissect_nbap_BCCH_ModificationTime(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 511U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_CellPortionID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfCellPortionsPerCell_1, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_SIR_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 63U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Best_Cell_Portions_Item_sequence[] = {
  { &hf_nbap_cellPortionID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellPortionID },
  { &hf_nbap_sIRValue       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SIR_Value },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Best_Cell_Portions_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Best_Cell_Portions_Item, Best_Cell_Portions_Item_sequence);

  return offset;
}


static const per_sequence_t Best_Cell_Portions_Value_sequence_of[1] = {
  { &hf_nbap_Best_Cell_Portions_Value_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Best_Cell_Portions_Item },
};

static int
dissect_nbap_Best_Cell_Portions_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Best_Cell_Portions_Value, Best_Cell_Portions_Value_sequence_of,
                                                  1, maxNrOfCellPortionsPerCell, FALSE);

  return offset;
}



static int
dissect_nbap_CellPortionLCRID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfCellPortionsPerCellLCR_1, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_RSCP_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Best_Cell_Portions_ItemLCR_sequence[] = {
  { &hf_nbap_cellPortionLCRID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellPortionLCRID },
  { &hf_nbap_rSCPValue      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RSCP_Value },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Best_Cell_Portions_ItemLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Best_Cell_Portions_ItemLCR, Best_Cell_Portions_ItemLCR_sequence);

  return offset;
}


static const per_sequence_t Best_Cell_Portions_ValueLCR_sequence_of[1] = {
  { &hf_nbap_Best_Cell_Portions_ValueLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Best_Cell_Portions_ItemLCR },
};

static int
dissect_nbap_Best_Cell_Portions_ValueLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Best_Cell_Portions_ValueLCR, Best_Cell_Portions_ValueLCR_sequence_of,
                                                  1, maxNrOfCellPortionsPerCellLCR, FALSE);

  return offset;
}



static int
dissect_nbap_BetaCD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}


static const value_string nbap_BlockingPriorityIndicator_vals[] = {
  {   0, "high" },
  {   1, "normal" },
  {   2, "low" },
  { 0, NULL }
};


static int
dissect_nbap_BlockingPriorityIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_SCTD_Indicator_vals[] = {
  {   0, "active" },
  {   1, "inactive" },
  { 0, NULL }
};


static int
dissect_nbap_SCTD_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_BundlingModeIndicator_vals[] = {
  {   0, "bundling" },
  {   1, "no-bundling" },
  { 0, NULL }
};


static int
dissect_nbap_BundlingModeIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_CommonTransportChannelID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  guint32 commontransportchannelid;
  nbap_dch_channel_info_t* nbap_dch_chnl_info;
  nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
  nbap_dch_chnl_info = nbap_private_data->nbap_dch_chnl_info;
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, &commontransportchannelid, FALSE);

  nbap_private_data->common_transport_channel_id = commontransportchannelid;
  if(commontransportchannelid<maxNrOfDCHs)
    nbap_dch_chnl_info[commontransportchannelid].next_dch = 0;


  return offset;
}


static const per_sequence_t BroadcastCommonTransportBearerIndication_sequence[] = {
  { &hf_nbap_commonTransportChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonTransportChannelID },
  { &hf_nbap_cid            , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_C_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_BroadcastCommonTransportBearerIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_BroadcastCommonTransportBearerIndication, BroadcastCommonTransportBearerIndication_sequence);

  return offset;
}



static int
dissect_nbap_BroadcastReference(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     24, 24, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_CCTrCH_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_Cell_Capability_Container(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     128, 128, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t Vacant_ERNTI_sequence_of[1] = {
  { &hf_nbap_Vacant_ERNTI_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_RNTI },
};

static int
dissect_nbap_Vacant_ERNTI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Vacant_ERNTI, Vacant_ERNTI_sequence_of,
                                                  1, maxERNTItoRelease, FALSE);

  return offset;
}


static const per_sequence_t Cell_ERNTI_Status_Information_Item_sequence[] = {
  { &hf_nbap_c_ID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_C_ID },
  { &hf_nbap_vacant_ERNTI   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Vacant_ERNTI },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Cell_ERNTI_Status_Information_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Cell_ERNTI_Status_Information_Item, Cell_ERNTI_Status_Information_Item_sequence);

  return offset;
}


static const per_sequence_t Cell_ERNTI_Status_Information_sequence_of[1] = {
  { &hf_nbap_Cell_ERNTI_Status_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Cell_ERNTI_Status_Information_Item },
};

static int
dissect_nbap_Cell_ERNTI_Status_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Cell_ERNTI_Status_Information, Cell_ERNTI_Status_Information_sequence_of,
                                                  1, maxCellinNodeB, FALSE);

  return offset;
}



static int
dissect_nbap_CellParameterID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, NULL, TRUE);

  return offset;
}


static const value_string nbap_CellPortion_CapabilityLCR_vals[] = {
  {   0, "cell-portion-capable" },
  {   1, "cell-portion-non-capable" },
  { 0, NULL }
};


static int
dissect_nbap_CellPortion_CapabilityLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_CellSyncBurstCode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 7U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_CellSyncBurstCodeShift(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 7U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_CellSyncBurstRepetitionPeriod(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 4095U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_CellSyncBurstSIR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 31U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_INTEGER_0_1048575_(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 1048575U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_INTEGER_0_255_(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, TRUE);

  return offset;
}


static const value_string nbap_CellSyncBurstTiming_vals[] = {
  {   0, "initialPhase" },
  {   1, "steadyStatePhase" },
  { 0, NULL }
};

static const per_choice_t CellSyncBurstTiming_choice[] = {
  {   0, &hf_nbap_initialPhase_0_1048575, ASN1_NO_EXTENSIONS     , dissect_nbap_INTEGER_0_1048575_ },
  {   1, &hf_nbap_steadyStatePhase, ASN1_NO_EXTENSIONS     , dissect_nbap_INTEGER_0_255_ },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_CellSyncBurstTiming(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_CellSyncBurstTiming, CellSyncBurstTiming_choice,
                                 NULL);

  return offset;
}



static int
dissect_nbap_INTEGER_0_524287_(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 524287U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_INTEGER_0_127_(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, NULL, TRUE);

  return offset;
}


static const value_string nbap_CellSyncBurstTimingLCR_vals[] = {
  {   0, "initialPhase" },
  {   1, "steadyStatePhase" },
  { 0, NULL }
};

static const per_choice_t CellSyncBurstTimingLCR_choice[] = {
  {   0, &hf_nbap_initialPhase   , ASN1_NO_EXTENSIONS     , dissect_nbap_INTEGER_0_524287_ },
  {   1, &hf_nbap_steadyStatePhase_01, ASN1_NO_EXTENSIONS     , dissect_nbap_INTEGER_0_127_ },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_CellSyncBurstTimingLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_CellSyncBurstTimingLCR, CellSyncBurstTimingLCR_choice,
                                 NULL);

  return offset;
}



static int
dissect_nbap_CellSyncBurstTimingThreshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 254U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_ChipOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 38399U, NULL, FALSE);

  return offset;
}


static const value_string nbap_Closedlooptimingadjustmentmode_vals[] = {
  {   0, "adj-1-slot" },
  {   1, "adj-2-slot" },
  { 0, NULL }
};


static int
dissect_nbap_Closedlooptimingadjustmentmode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_CodeRate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 63U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_CodeRate_short(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 10U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_INTEGER_0_65535(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 65535U, NULL, FALSE);

  return offset;
}


static const per_sequence_t CommonChannelsCapacityConsumptionLaw_item_sequence[] = {
  { &hf_nbap_dl_Cost        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_65535 },
  { &hf_nbap_ul_Cost        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_65535 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommonChannelsCapacityConsumptionLaw_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommonChannelsCapacityConsumptionLaw_item, CommonChannelsCapacityConsumptionLaw_item_sequence);

  return offset;
}


static const per_sequence_t CommonChannelsCapacityConsumptionLaw_sequence_of[1] = {
  { &hf_nbap_CommonChannelsCapacityConsumptionLaw_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_CommonChannelsCapacityConsumptionLaw_item },
};

static int
dissect_nbap_CommonChannelsCapacityConsumptionLaw(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_CommonChannelsCapacityConsumptionLaw, CommonChannelsCapacityConsumptionLaw_sequence_of,
                                                  1, maxNrOfSF, FALSE);

  return offset;
}


static const value_string nbap_Common_EDCH_Capability_vals[] = {
  {   0, "common-EDCH-capable" },
  {   1, "common-EDCH-non-capable" },
  { 0, NULL }
};


static int
dissect_nbap_Common_EDCH_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_Common_E_DCH_HSDPCCH_Capability_vals[] = {
  {   0, "hSDPCCH-non-capable" },
  {   1, "aCK-NACK-capable" },
  {   2, "aCK-NACK-CQI-capable" },
  { 0, NULL }
};


static int
dissect_nbap_Common_E_DCH_HSDPCCH_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_DPC_Mode_vals[] = {
  {   0, "mode0" },
  {   1, "mode1" },
  { 0, NULL }
};


static int
dissect_nbap_DPC_Mode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t Common_E_DCH_UL_DPCH_InfoItem_sequence[] = {
  { &hf_nbap_uL_SIR_Target  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_SIR },
  { &hf_nbap_dPC_Mode       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DPC_Mode },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_E_DCH_UL_DPCH_InfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_E_DCH_UL_DPCH_InfoItem, Common_E_DCH_UL_DPCH_InfoItem_sequence);

  return offset;
}


static const value_string nbap_Max_Set_E_DPDCHs_vals[] = {
  {   0, "vN256" },
  {   1, "vN128" },
  {   2, "vN64" },
  {   3, "vN32" },
  {   4, "vN16" },
  {   5, "vN8" },
  {   6, "vN4" },
  {   7, "v2xN4" },
  {   8, "v2xN2" },
  {   9, "v2xN2plus2xN4" },
  {  10, "v2xM2plus2xM4" },
  { 0, NULL }
};


static int
dissect_nbap_Max_Set_E_DPDCHs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     10, NULL, TRUE, 1, NULL);

  return offset;
}



static int
dissect_nbap_PunctureLimit(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_E_DCH_TFCI_Table_Index(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 1U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_Reference_E_TFCI_PO(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfRefETFCI_PO_QUANTSTEPs, NULL, FALSE);

  return offset;
}


static const per_sequence_t Reference_E_TFCI_Information_Item_sequence[] = {
  { &hf_nbap_reference_E_TFCI, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_TFCI },
  { &hf_nbap_reference_E_TFCI_PO, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Reference_E_TFCI_PO },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Reference_E_TFCI_Information_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Reference_E_TFCI_Information_Item, Reference_E_TFCI_Information_Item_sequence);

  return offset;
}


static const per_sequence_t Reference_E_TFCI_Information_sequence_of[1] = {
  { &hf_nbap_Reference_E_TFCI_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Reference_E_TFCI_Information_Item },
};

static int
dissect_nbap_Reference_E_TFCI_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Reference_E_TFCI_Information, Reference_E_TFCI_Information_sequence_of,
                                                  1, maxNrOfRefETFCIs, FALSE);

  return offset;
}


static const per_sequence_t E_TFCS_Information_sequence[] = {
  { &hf_nbap_e_DCH_TFCI_Table_Index, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_TFCI_Table_Index },
  { &hf_nbap_e_DCH_Min_Set_E_TFCI, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_TFCI },
  { &hf_nbap_reference_E_TFCI_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Reference_E_TFCI_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_TFCS_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_TFCS_Information, E_TFCS_Information_sequence);

  return offset;
}


static const value_string nbap_E_TTI_vals[] = {
  {   0, "e-TTI-2ms" },
  {   1, "e-TTI-10ms" },
  { 0, NULL }
};


static int
dissect_nbap_E_TTI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_E_DPCCH_PO(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfEDPCCH_PO_QUANTSTEPs, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_E_RGCH_2_IndexStepThreshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 37U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_E_RGCH_3_IndexStepThreshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 37U, NULL, FALSE);

  return offset;
}


static const value_string nbap_HARQ_Info_for_E_DCH_vals[] = {
  {   0, "rv0" },
  {   1, "rvtable" },
  { 0, NULL }
};


static int
dissect_nbap_HARQ_Info_for_E_DCH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t Common_E_DCH_EDPCH_InfoItem_sequence[] = {
  { &hf_nbap_maxSet_E_DPDCHs, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Max_Set_E_DPDCHs },
  { &hf_nbap_ul_PunctureLimit, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PunctureLimit },
  { &hf_nbap_e_TFCS_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_TFCS_Information },
  { &hf_nbap_e_TTI          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_TTI },
  { &hf_nbap_e_DPCCH_PO     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DPCCH_PO },
  { &hf_nbap_e_RGCH_2_IndexStepThreshold, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_RGCH_2_IndexStepThreshold },
  { &hf_nbap_e_RGCH_3_IndexStepThreshold, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_RGCH_3_IndexStepThreshold },
  { &hf_nbap_hARQ_Info_for_E_DCH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HARQ_Info_for_E_DCH },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_E_DCH_EDPCH_InfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_E_DCH_EDPCH_InfoItem, Common_E_DCH_EDPCH_InfoItem_sequence);

  return offset;
}



static int
dissect_nbap_E_DCH_Reference_Power_Offset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfEDCH_HARQ_PO_QUANTSTEPs, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_E_DCH_PowerOffset_for_SchedulingInfo(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfEDCH_HARQ_PO_QUANTSTEPs, NULL, FALSE);

  return offset;
}


static const value_string nbap_Max_EDCH_Resource_Allocation_for_CCCH_vals[] = {
  {   0, "v8" },
  {   1, "v12" },
  {   2, "v16" },
  {   3, "v24" },
  {   4, "v32" },
  {   5, "v40" },
  {   6, "v80" },
  {   7, "v120" },
  { 0, NULL }
};


static int
dissect_nbap_Max_EDCH_Resource_Allocation_for_CCCH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_Max_Period_for_Collistion_Resolution(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            8U, 24U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_INTEGER_0_5000_(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 5000U, NULL, TRUE);

  return offset;
}


static const per_sequence_t Max_TB_Sizes_sequence[] = {
  { &hf_nbap_maximum_TB_Size_cell_edge_users, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_5000_ },
  { &hf_nbap_maximum_TB_Size_other_users, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_5000_ },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Max_TB_Sizes(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Max_TB_Sizes, Max_TB_Sizes_sequence);

  return offset;
}



static int
dissect_nbap_BOOLEAN(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_boolean(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}


static const per_sequence_t Common_E_DCH_InfoItem_sequence[] = {
  { &hf_nbap_e_DCH_Reference_Power_Offset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_Reference_Power_Offset },
  { &hf_nbap_e_DCH_PowerOffset_for_SchedulingInfo, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_PowerOffset_for_SchedulingInfo },
  { &hf_nbap_max_EDCH_Resource_Allocation_for_CCCH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Max_EDCH_Resource_Allocation_for_CCCH },
  { &hf_nbap_max_Period_for_Collistion_Resolution, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Max_Period_for_Collistion_Resolution },
  { &hf_nbap_max_TB_Sizes   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Max_TB_Sizes },
  { &hf_nbap_common_E_DCH_ImplicitRelease_Indicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BOOLEAN },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_E_DCH_InfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_E_DCH_InfoItem, Common_E_DCH_InfoItem_sequence);

  return offset;
}



static int
dissect_nbap_Nack_Power_Offset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 8U, NULL, TRUE);

  return offset;
}


static const value_string nbap_CQI_Feedback_Cycle_vals[] = {
  {   0, "v0" },
  {   1, "v2" },
  {   2, "v4" },
  {   3, "v8" },
  {   4, "v10" },
  {   5, "v20" },
  {   6, "v40" },
  {   7, "v80" },
  {   8, "v160" },
  {   9, "v16" },
  {  10, "v32" },
  {  11, "v64" },
  { 0, NULL }
};


static int
dissect_nbap_CQI_Feedback_Cycle(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     9, NULL, TRUE, 3, NULL);

  return offset;
}



static int
dissect_nbap_CQI_RepetitionFactor(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 4U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_CQI_Power_Offset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 8U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_Measurement_Power_Offset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -12, 26U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Common_E_DCH_CQI_Info_sequence[] = {
  { &hf_nbap_cqiFeedback_CycleK, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CQI_Feedback_Cycle },
  { &hf_nbap_cqiRepetitionFactor, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CQI_RepetitionFactor },
  { &hf_nbap_cqiPowerOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CQI_Power_Offset },
  { &hf_nbap_measurement_Power_Offset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Measurement_Power_Offset },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_E_DCH_CQI_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_E_DCH_CQI_Info, Common_E_DCH_CQI_Info_sequence);

  return offset;
}


static const per_sequence_t Common_E_DCH_HSDPCCH_InfoItem_sequence[] = {
  { &hf_nbap_ackNackRepetitionFactor, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AckNack_RepetitionFactor },
  { &hf_nbap_ackPowerOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Ack_Power_Offset },
  { &hf_nbap_nackPowerOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Nack_Power_Offset },
  { &hf_nbap_common_E_DCH_CQI_Info, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Common_E_DCH_CQI_Info },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_E_DCH_HSDPCCH_InfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_E_DCH_HSDPCCH_InfoItem, Common_E_DCH_HSDPCCH_InfoItem_sequence);

  return offset;
}



static int
dissect_nbap_CommonPhysicalChannelID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  guint32 commonphysicalchannelid;
  nbap_dch_channel_info_t* nbap_dch_chnl_info;
  nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
  nbap_dch_chnl_info = nbap_private_data->nbap_dch_chnl_info;
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, &commonphysicalchannelid, FALSE);

  nbap_private_data->common_physical_channel_id = commonphysicalchannelid;
  if(commonphysicalchannelid<maxNrOfDCHs)
    nbap_dch_chnl_info[commonphysicalchannelid].next_dch = 0;


  return offset;
}


static int * const PreambleSignatures_bits[] = {
  &hf_nbap_PreambleSignatures_signature15,
  &hf_nbap_PreambleSignatures_signature14,
  &hf_nbap_PreambleSignatures_signature13,
  &hf_nbap_PreambleSignatures_signature12,
  &hf_nbap_PreambleSignatures_signature11,
  &hf_nbap_PreambleSignatures_signature10,
  &hf_nbap_PreambleSignatures_signature9,
  &hf_nbap_PreambleSignatures_signature8,
  &hf_nbap_PreambleSignatures_signature7,
  &hf_nbap_PreambleSignatures_signature6,
  &hf_nbap_PreambleSignatures_signature5,
  &hf_nbap_PreambleSignatures_signature4,
  &hf_nbap_PreambleSignatures_signature3,
  &hf_nbap_PreambleSignatures_signature2,
  &hf_nbap_PreambleSignatures_signature1,
  &hf_nbap_PreambleSignatures_signature0,
  NULL
};

static int
dissect_nbap_PreambleSignatures(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     16, 16, FALSE, PreambleSignatures_bits, 16, NULL, NULL);

  return offset;
}



static int
dissect_nbap_ScramblingCodeNumber(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_PreambleThreshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 72U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_E_AI_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_boolean(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}


static const value_string nbap_STTD_Indicator_vals[] = {
  {   0, "active" },
  {   1, "inactive" },
  { 0, NULL }
};


static int
dissect_nbap_STTD_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t Common_E_DCH_AICH_Information_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_aICH_TransmissionTiming, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AICH_TransmissionTiming },
  { &hf_nbap_fdd_dl_ChannelisationCodeNumber, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_DL_ChannelisationCodeNumber },
  { &hf_nbap_aICH_Power     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AICH_Power },
  { &hf_nbap_sTTD_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_STTD_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_E_DCH_AICH_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_E_DCH_AICH_Information, Common_E_DCH_AICH_Information_sequence);

  return offset;
}


static const per_sequence_t Common_E_DCH_Preamble_Control_InfoItem_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_common_E_DCH_PreambleSignatures, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PreambleSignatures },
  { &hf_nbap_scramblingCodeNumber, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ScramblingCodeNumber },
  { &hf_nbap_preambleThreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PreambleThreshold },
  { &hf_nbap_e_AI_Indicator , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_AI_Indicator },
  { &hf_nbap_common_E_DCH_AICH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Common_E_DCH_AICH_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_E_DCH_Preamble_Control_InfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_E_DCH_Preamble_Control_InfoItem, Common_E_DCH_Preamble_Control_InfoItem_sequence);

  return offset;
}


static const per_sequence_t Common_E_DCH_FDPCH_InfoItem_sequence[] = {
  { &hf_nbap_f_DPCH_SlotFormat, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_F_DPCH_SlotFormat },
  { &hf_nbap_fdd_TPC_DownlinkStepSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_TPC_DownlinkStepSize },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_E_DCH_FDPCH_InfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_E_DCH_FDPCH_InfoItem, Common_E_DCH_FDPCH_InfoItem_sequence);

  return offset;
}



static int
dissect_nbap_Soffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 9U, NULL, TRUE);

  return offset;
}


static const per_sequence_t Common_E_DCH_Resource_Combination_InfoList_Item_sequence[] = {
  { &hf_nbap_soffset        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Soffset },
  { &hf_nbap_f_DPCH_DL_Code_Number, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_DL_ChannelisationCodeNumber },
  { &hf_nbap_ul_DPCH_ScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_ScramblingCode },
  { &hf_nbap_e_RGCH_E_HICH_Channelisation_Code, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_DL_ChannelisationCodeNumber },
  { &hf_nbap_e_RGCH_Signature_Sequence, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_RGCH_Signature_Sequence },
  { &hf_nbap_e_HICH_Signature_Sequence, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_HICH_Signature_Sequence },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_E_DCH_Resource_Combination_InfoList_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_E_DCH_Resource_Combination_InfoList_Item, Common_E_DCH_Resource_Combination_InfoList_Item_sequence);

  return offset;
}


static const per_sequence_t Common_E_DCH_Resource_Combination_InfoList_sequence_of[1] = {
  { &hf_nbap_Common_E_DCH_Resource_Combination_InfoList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Common_E_DCH_Resource_Combination_InfoList_Item },
};

static int
dissect_nbap_Common_E_DCH_Resource_Combination_InfoList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Common_E_DCH_Resource_Combination_InfoList, Common_E_DCH_Resource_Combination_InfoList_sequence_of,
                                                  1, maxNrOfCommonEDCH, FALSE);

  return offset;
}



static int
dissect_nbap_Common_MACFlow_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfCommonMACFlows_1, &nbap_get_private_data(actx->pinfo)->common_macdflow_id, FALSE);

  return offset;
}


static const value_string nbap_TransportBearerRequestIndicator_vals[] = {
  {   0, "bearerRequested" },
  {   1, "bearerNotRequested" },
  { 0, NULL }
};


static int
dissect_nbap_TransportBearerRequestIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_DsField(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     8, 8, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_GenericTrafficCategory(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     8, 8, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const value_string nbap_TnlQos_vals[] = {
  {   0, "dsField" },
  {   1, "genericTrafficCategory" },
  { 0, NULL }
};

static const per_choice_t TnlQos_choice[] = {
  {   0, &hf_nbap_dsField        , ASN1_EXTENSION_ROOT    , dissect_nbap_DsField },
  {   1, &hf_nbap_genericTrafficCategory, ASN1_EXTENSION_ROOT    , dissect_nbap_GenericTrafficCategory },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_TnlQos(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_TnlQos, TnlQos_choice,
                                 NULL);

  return offset;
}


static const value_string nbap_PayloadCRC_PresenceIndicator_vals[] = {
  {   0, "cRC-Included" },
  {   1, "cRC-NotIncluded" },
  { 0, NULL }
};


static int
dissect_nbap_PayloadCRC_PresenceIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
guint32 payload_crc_value;

  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, &payload_crc_value, TRUE, 0, NULL);


    if(payload_crc_value == 0){
        nbap_get_private_data(actx->pinfo)->dch_crc_present = TRUE;
    }else{
        nbap_get_private_data(actx->pinfo)->dch_crc_present = FALSE;
    }


  return offset;
}



static int
dissect_nbap_Maximum_Number_of_Retransmissions_For_E_DCH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_E_DCH_HARQ_PO_FDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfEDCH_HARQ_PO_QUANTSTEPs, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_E_DCH_MACdFlow_Multiplexing_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     maxNrOfEDCHMACdFlows, maxNrOfEDCHMACdFlows, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_LogicalChannelID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  guint32 logical_channel_id;
  guint num_items;
  nbap_edch_channel_info_t* nbap_edch_channel_info;
  nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
  nbap_edch_channel_info = nbap_private_data->nbap_edch_channel_info;
  /* Set logical channel id for this entry*/
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 15U, &logical_channel_id, FALSE);


    nbap_private_data->logical_channel_id = logical_channel_id;
    num_items = nbap_private_data->num_items;
    if(num_items > 0 && num_items < MAX_EDCH_DDIS + 1)
        nbap_edch_channel_info[nbap_private_data->e_dch_macdflow_id].lchId[num_items-1] = nbap_private_data->logical_channel_id;


  return offset;
}



static int
dissect_nbap_MAC_PDU_SizeExtended(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
guint32 hsdsch_macdflow_id;
nbap_hsdsch_channel_info_t* nbap_hsdsch_channel_info;
nbap_hsdsch_channel_info = nbap_get_private_data(actx->pinfo)->nbap_hsdsch_channel_info;
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 1504U, NULL, TRUE);

    nbap_get_private_data(actx->pinfo)->max_mac_d_pdu_size_ext_ie_present = TRUE;
    hsdsch_macdflow_id = nbap_get_private_data(actx->pinfo)->hsdsch_macdflow_id;
    if(nbap_hsdsch_channel_info[hsdsch_macdflow_id].crnc_port != 0){
            nbap_hsdsch_channel_info[hsdsch_macdflow_id].entity = ehs;
    }


  return offset;
}


static const per_sequence_t Common_E_DCH_LogicalChannel_InfoList_Item_sequence[] = {
  { &hf_nbap_logicalChannelId, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_LogicalChannelID },
  { &hf_nbap_maximumMACcPDU_SizeExtended, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MAC_PDU_SizeExtended },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_E_DCH_LogicalChannel_InfoList_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_E_DCH_LogicalChannel_InfoList_Item, Common_E_DCH_LogicalChannel_InfoList_Item_sequence);

  return offset;
}


static const per_sequence_t Common_E_DCH_LogicalChannel_InfoList_sequence_of[1] = {
  { &hf_nbap_Common_E_DCH_LogicalChannel_InfoList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Common_E_DCH_LogicalChannel_InfoList_Item },
};

static int
dissect_nbap_Common_E_DCH_LogicalChannel_InfoList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Common_E_DCH_LogicalChannel_InfoList, Common_E_DCH_LogicalChannel_InfoList_sequence_of,
                                                  1, maxNoOfLogicalChannels, FALSE);

  return offset;
}


static const per_sequence_t Common_E_DCH_MACdFlow_Specific_InfoList_Item_sequence[] = {
  { &hf_nbap_common_e_DCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_ID },
  { &hf_nbap_maximum_Number_of_Retransmissions_For_E_DCH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Maximum_Number_of_Retransmissions_For_E_DCH },
  { &hf_nbap_eDCH_HARQ_PO_FDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_HARQ_PO_FDD },
  { &hf_nbap_eDCH_MACdFlow_Multiplexing_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_MACdFlow_Multiplexing_List },
  { &hf_nbap_common_E_DCHLogicalChannelInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Common_E_DCH_LogicalChannel_InfoList },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_E_DCH_MACdFlow_Specific_InfoList_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_E_DCH_MACdFlow_Specific_InfoList_Item, Common_E_DCH_MACdFlow_Specific_InfoList_Item_sequence);

  return offset;
}


static const per_sequence_t Common_E_DCH_MACdFlow_Specific_InfoList_sequence_of[1] = {
  { &hf_nbap_Common_E_DCH_MACdFlow_Specific_InfoList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Common_E_DCH_MACdFlow_Specific_InfoList_Item },
};

static int
dissect_nbap_Common_E_DCH_MACdFlow_Specific_InfoList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Common_E_DCH_MACdFlow_Specific_InfoList, Common_E_DCH_MACdFlow_Specific_InfoList_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}


static const per_sequence_t Ul_common_E_DCH_MACflow_Specific_InfoList_Item_sequence[] = {
  { &hf_nbap_ul_Common_MACFlowID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Common_MACFlow_ID },
  { &hf_nbap_transportBearerRequestIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportBearerRequestIndicator },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_tnlQos         , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TnlQos },
  { &hf_nbap_payloadCRC_PresenceIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PayloadCRC_PresenceIndicator },
  { &hf_nbap_bundlingModeIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BundlingModeIndicator },
  { &hf_nbap_common_E_DCH_MACdFlow_Specific_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Common_E_DCH_MACdFlow_Specific_InfoList },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Ul_common_E_DCH_MACflow_Specific_InfoList_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoList_Item, Ul_common_E_DCH_MACflow_Specific_InfoList_Item_sequence);

  return offset;
}


static const per_sequence_t Ul_common_E_DCH_MACflow_Specific_InfoList_sequence_of[1] = {
  { &hf_nbap_Ul_common_E_DCH_MACflow_Specific_InfoList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Ul_common_E_DCH_MACflow_Specific_InfoList_Item },
};

static int
dissect_nbap_Ul_common_E_DCH_MACflow_Specific_InfoList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoList, Ul_common_E_DCH_MACflow_Specific_InfoList_sequence_of,
                                                  1, maxNrOfCommonMACFlows, FALSE);

  return offset;
}


static const per_sequence_t Common_EDCH_System_InformationFDD_sequence[] = {
  { &hf_nbap_common_E_DCH_UL_DPCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Common_E_DCH_UL_DPCH_InfoItem },
  { &hf_nbap_common_E_DCH_EDPCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Common_E_DCH_EDPCH_InfoItem },
  { &hf_nbap_common_E_DCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Common_E_DCH_InfoItem },
  { &hf_nbap_common_E_DCH_HSDPCCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Common_E_DCH_HSDPCCH_InfoItem },
  { &hf_nbap_common_E_DCH_Preamble_Control_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Common_E_DCH_Preamble_Control_InfoItem },
  { &hf_nbap_common_E_DCH_FDPCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Common_E_DCH_FDPCH_InfoItem },
  { &hf_nbap_common_E_DCH_E_AGCH_ChannelisationCodeNumber, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_FDD_DL_ChannelisationCodeNumber },
  { &hf_nbap_common_E_DCH_Resource_Combination_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Common_E_DCH_Resource_Combination_InfoList },
  { &hf_nbap_ul_common_E_DCH_MACflow_Specific_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Ul_common_E_DCH_MACflow_Specific_InfoList },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_EDCH_System_InformationFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_EDCH_System_InformationFDD, Common_EDCH_System_InformationFDD_sequence);

  return offset;
}


static const per_sequence_t Ul_common_E_DCH_MACflow_Specific_InfoResponseList_Item_sequence[] = {
  { &hf_nbap_ul_Common_MACFlowID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Common_MACFlow_ID },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseList_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseList_Item, Ul_common_E_DCH_MACflow_Specific_InfoResponseList_Item_sequence);

  return offset;
}


static const per_sequence_t Ul_common_E_DCH_MACflow_Specific_InfoResponseList_sequence_of[1] = {
  { &hf_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseList_Item },
};

static int
dissect_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseList, Ul_common_E_DCH_MACflow_Specific_InfoResponseList_sequence_of,
                                                  1, maxNrOfCommonMACFlows, FALSE);

  return offset;
}


static const per_sequence_t Common_EDCH_System_Information_ResponseFDD_sequence[] = {
  { &hf_nbap_ul_common_E_DCH_MACflow_Specific_InfoResponse, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseList },
  { &hf_nbap_serving_Grant_Value, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_Serving_Grant_Value },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_EDCH_System_Information_ResponseFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_EDCH_System_Information_ResponseFDD, Common_EDCH_System_Information_ResponseFDD_sequence);

  return offset;
}


static const per_sequence_t E_RNTI_List_sequence_of[1] = {
  { &hf_nbap_E_RNTI_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_RNTI },
};

static int
dissect_nbap_E_RNTI_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_RNTI_List, E_RNTI_List_sequence_of,
                                                  1, maxofERNTI, FALSE);

  return offset;
}


static const per_sequence_t Common_MACFlows_to_DeleteFDD_Item_sequence[] = {
  { &hf_nbap_common_MACFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Common_MACFlow_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_MACFlows_to_DeleteFDD_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_MACFlows_to_DeleteFDD_Item, Common_MACFlows_to_DeleteFDD_Item_sequence);

  return offset;
}


static const per_sequence_t Common_MACFlows_to_DeleteFDD_sequence_of[1] = {
  { &hf_nbap_Common_MACFlows_to_DeleteFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Common_MACFlows_to_DeleteFDD_Item },
};

static int
dissect_nbap_Common_MACFlows_to_DeleteFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Common_MACFlows_to_DeleteFDD, Common_MACFlows_to_DeleteFDD_sequence_of,
                                                  1, maxNrOfCommonMACFlows, FALSE);

  return offset;
}



static int
dissect_nbap_PriorityQueue_Id(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfPriorityQueues_1, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_SchedulingPriorityIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}


static const value_string nbap_T1_vals[] = {
  {   0, "v10" },
  {   1, "v20" },
  {   2, "v30" },
  {   3, "v40" },
  {   4, "v50" },
  {   5, "v60" },
  {   6, "v70" },
  {   7, "v80" },
  {   8, "v90" },
  {   9, "v100" },
  {  10, "v120" },
  {  11, "v140" },
  {  12, "v160" },
  {  13, "v200" },
  {  14, "v300" },
  {  15, "v400" },
  { 0, NULL }
};


static int
dissect_nbap_T1(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     16, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_MAC_ehs_Reset_Timer_vals[] = {
  {   0, "v1" },
  {   1, "v2" },
  {   2, "v3" },
  {   3, "v4" },
  { 0, NULL }
};


static int
dissect_nbap_MAC_ehs_Reset_Timer(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_DiscardTimer_vals[] = {
  {   0, "v20" },
  {   1, "v40" },
  {   2, "v60" },
  {   3, "v80" },
  {   4, "v100" },
  {   5, "v120" },
  {   6, "v140" },
  {   7, "v160" },
  {   8, "v180" },
  {   9, "v200" },
  {  10, "v250" },
  {  11, "v300" },
  {  12, "v400" },
  {  13, "v500" },
  {  14, "v750" },
  {  15, "v1000" },
  {  16, "v1250" },
  {  17, "v1500" },
  {  18, "v1750" },
  {  19, "v2000" },
  {  20, "v2500" },
  {  21, "v3000" },
  {  22, "v3500" },
  {  23, "v4000" },
  {  24, "v4500" },
  {  25, "v5000" },
  {  26, "v7500" },
  { 0, NULL }
};

static value_string_ext nbap_DiscardTimer_vals_ext = VALUE_STRING_EXT_INIT(nbap_DiscardTimer_vals);


static int
dissect_nbap_DiscardTimer(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     27, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_MAC_hsWindowSize_vals[] = {
  {   0, "v4" },
  {   1, "v6" },
  {   2, "v8" },
  {   3, "v12" },
  {   4, "v16" },
  {   5, "v24" },
  {   6, "v32" },
  {   7, "v64" },
  {   8, "v128" },
  { 0, NULL }
};


static int
dissect_nbap_MAC_hsWindowSize(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     7, NULL, TRUE, 2, NULL);

  return offset;
}


static const per_sequence_t Priority_Queue_Information_for_Enhanced_FACH_PCH_sequence[] = {
  { &hf_nbap_priorityQueue_Id, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PriorityQueue_Id },
  { &hf_nbap_schedulingPriorityIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SchedulingPriorityIndicator },
  { &hf_nbap_t1             , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T1 },
  { &hf_nbap_mAC_ehs_Reset_Timer, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MAC_ehs_Reset_Timer },
  { &hf_nbap_discardTimer   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DiscardTimer },
  { &hf_nbap_mAC_hsWindowSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MAC_hsWindowSize },
  { &hf_nbap_maximum_MACcPDU_Size, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MAC_PDU_SizeExtended },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Priority_Queue_Information_for_Enhanced_FACH_PCH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Priority_Queue_Information_for_Enhanced_FACH_PCH, Priority_Queue_Information_for_Enhanced_FACH_PCH_sequence);

  return offset;
}


static const per_sequence_t Common_MACFlow_PriorityQueue_Item_sequence[] = {
  { &hf_nbap_priority_Queue_Information_for_Enhanced_FACH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Priority_Queue_Information_for_Enhanced_FACH_PCH },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_MACFlow_PriorityQueue_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_MACFlow_PriorityQueue_Item, Common_MACFlow_PriorityQueue_Item_sequence);

  return offset;
}


static const per_sequence_t Common_MACFlow_PriorityQueue_Information_sequence_of[1] = {
  { &hf_nbap_Common_MACFlow_PriorityQueue_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Common_MACFlow_PriorityQueue_Item },
};

static int
dissect_nbap_Common_MACFlow_PriorityQueue_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Common_MACFlow_PriorityQueue_Information, Common_MACFlow_PriorityQueue_Information_sequence_of,
                                                  1, maxNrOfcommonMACQueues, FALSE);

  return offset;
}


static const per_sequence_t CommonMACFlow_Specific_InfoItem_sequence[] = {
  { &hf_nbap_common_MACFlow_Id, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Common_MACFlow_ID },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_tnl_qos        , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TnlQos },
  { &hf_nbap_common_MACFlow_PriorityQueue_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Common_MACFlow_PriorityQueue_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommonMACFlow_Specific_InfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    address     dst_addr;
    guint32 transportLayerAddress_ipv4;
    guint16 bindingID;
    guint32 common_macdflow_id;
    nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
    nbap_private_data->transportLayerAddress_ipv4 = 0;
    nbap_private_data->binding_id_port = 0;
    nbap_private_data->num_items++;
    nbap_common_channel_info_t* nbap_common_channel_info;
    nbap_common_channel_info = nbap_private_data->nbap_common_channel_info;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommonMACFlow_Specific_InfoItem, CommonMACFlow_Specific_InfoItem_sequence);


    transportLayerAddress_ipv4 = nbap_private_data->transportLayerAddress_ipv4;
    bindingID = nbap_private_data->binding_id_port;
    if (PINFO_FD_VISITED(actx->pinfo) || transportLayerAddress_ipv4 == 0 || bindingID == 0){
        return offset;
    }

    set_address(&dst_addr, AT_IPv4, 4, &transportLayerAddress_ipv4);

    /* Set address for collection of common entries */
    common_macdflow_id = nbap_private_data->common_macdflow_id;
    copy_address_wmem(actx->pinfo->pool,&(nbap_common_channel_info[common_macdflow_id].crnc_address),&dst_addr);
    nbap_common_channel_info[common_macdflow_id].crnc_port = nbap_private_data->binding_id_port;


  return offset;
}


static const per_sequence_t CommonMACFlow_Specific_InfoList_sequence_of[1] = {
  { &hf_nbap_CommonMACFlow_Specific_InfoList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_CommonMACFlow_Specific_InfoItem },
};

static int
dissect_nbap_CommonMACFlow_Specific_InfoList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_CommonMACFlow_Specific_InfoList, CommonMACFlow_Specific_InfoList_sequence_of,
                                                  1, maxNrOfCommonMACFlows, FALSE);

  return offset;
}



static int
dissect_nbap_MACdPDU_Size(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
guint32 mac_d_pdu_size;
guint num_items;
nbap_edch_channel_info_t* nbap_edch_channel_info;
nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
nbap_edch_channel_info = nbap_private_data->nbap_edch_channel_info;
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 5000U, &mac_d_pdu_size, TRUE);


    nbap_private_data->mac_d_pdu_size = mac_d_pdu_size;
    if (PINFO_FD_VISITED(actx->pinfo))
    {
        return offset;
    }
    num_items = nbap_private_data->num_items;
    if(num_items > 0 && num_items < MAX_EDCH_DDIS + 1)
        nbap_edch_channel_info[nbap_private_data->e_dch_macdflow_id].edch_macd_pdu_size[num_items-1] = nbap_private_data->mac_d_pdu_size;


  return offset;
}



static int
dissect_nbap_HSDSCH_InitialWindowSize(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t HSDSCH_Initial_Capacity_AllocationItem_sequence[] = {
  { &hf_nbap_schedulingPriorityIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SchedulingPriorityIndicator },
  { &hf_nbap_maximum_MACdPDU_Size, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MACdPDU_Size },
  { &hf_nbap_hSDSCH_InitialWindowSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_InitialWindowSize },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_Initial_Capacity_AllocationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_Initial_Capacity_AllocationItem, HSDSCH_Initial_Capacity_AllocationItem_sequence);

  return offset;
}


static const per_sequence_t HSDSCH_Initial_Capacity_Allocation_sequence_of[1] = {
  { &hf_nbap_HSDSCH_Initial_Capacity_Allocation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_Initial_Capacity_AllocationItem },
};

static int
dissect_nbap_HSDSCH_Initial_Capacity_Allocation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HSDSCH_Initial_Capacity_Allocation, HSDSCH_Initial_Capacity_Allocation_sequence_of,
                                                  1, maxNrOfPriorityQueues, FALSE);

  return offset;
}


static const per_sequence_t CommonMACFlow_Specific_InfoItem_Response_sequence[] = {
  { &hf_nbap_commonMACFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Common_MACFlow_ID },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_hSDSCH_Initial_Capacity_Allocation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSDSCH_Initial_Capacity_Allocation },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommonMACFlow_Specific_InfoItem_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommonMACFlow_Specific_InfoItem_Response, CommonMACFlow_Specific_InfoItem_Response_sequence);

  return offset;
}


static const per_sequence_t CommonMACFlow_Specific_InfoList_Response_sequence_of[1] = {
  { &hf_nbap_CommonMACFlow_Specific_InfoList_Response_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_CommonMACFlow_Specific_InfoItem_Response },
};

static int
dissect_nbap_CommonMACFlow_Specific_InfoList_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_CommonMACFlow_Specific_InfoList_Response, CommonMACFlow_Specific_InfoList_Response_sequence_of,
                                                  1, maxNrOfCommonMACFlows, FALSE);

  return offset;
}


static const value_string nbap_TUTRANGPSAccuracyClass_vals[] = {
  {   0, "accuracy-class-A" },
  {   1, "accuracy-class-B" },
  {   2, "accuracy-class-C" },
  { 0, NULL }
};


static int
dissect_nbap_TUTRANGPSAccuracyClass(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_TUTRANGANSSAccuracyClass_vals[] = {
  {   0, "ganssAccuracy-class-A" },
  {   1, "ganssAccuracy-class-B" },
  {   2, "ganssAccuracy-class-C" },
  { 0, NULL }
};


static int
dissect_nbap_TUTRANGANSSAccuracyClass(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_CommonMeasurementAccuracy_vals[] = {
  {   0, "tUTRANGPSMeasurementAccuracyClass" },
  {   1, "tUTRANGANSSMeasurementAccuracyClass" },
  { 0, NULL }
};

static const per_choice_t CommonMeasurementAccuracy_choice[] = {
  {   0, &hf_nbap_tUTRANGPSMeasurementAccuracyClass, ASN1_EXTENSION_ROOT    , dissect_nbap_TUTRANGPSAccuracyClass },
  {   1, &hf_nbap_tUTRANGANSSMeasurementAccuracyClass, ASN1_NOT_EXTENSION_ROOT, dissect_nbap_TUTRANGANSSAccuracyClass },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_CommonMeasurementAccuracy(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_CommonMeasurementAccuracy, CommonMeasurementAccuracy_choice,
                                 NULL);

  return offset;
}


static const value_string nbap_CommonMeasurementType_vals[] = {
  {   0, "received-total-wide-band-power" },
  {   1, "transmitted-carrier-power" },
  {   2, "acknowledged-prach-preambles" },
  {   3, "ul-timeslot-iscp" },
  {   4, "notUsed-1-acknowledged-PCPCH-access-preambles" },
  {   5, "notUsed-2-detected-PCPCH-access-preambles" },
  {   6, "uTRAN-GPS-Timing-of-Cell-Frames-for-UE-Positioning" },
  {   7, "sFN-SFN-Observed-Time-Difference" },
  {   8, "transmittedCarrierPowerOfAllCodesNotUsedForHSTransmission" },
  {   9, "hS-DSCH-Required-Power" },
  {  10, "hS-DSCH-Provided-Bit-Rate" },
  {  11, "received-total-wide-band-power-for-cellPortion" },
  {  12, "transmitted-carrier-power-for-cellPortion" },
  {  13, "transmittedCarrierPowerOfAllCodesNotUsedForHS-PDSCH-HS-SCCH-E-AGCH-E-RGCHOrE-HICHTransmission-for-cellPortion" },
  {  14, "upPTS-Interference" },
  {  15, "dLTransmissionBranchLoad" },
  {  16, "hS-DSCH-Required-Power-for-cell-portion" },
  {  17, "hS-DSCH-Provided-Bit-Rate-for-cell-portion" },
  {  18, "e-DCH-Provided-Bit-Rate" },
  {  19, "e-DCH-Non-serving-Relative-Grant-Down-Commands" },
  {  20, "received-Scheduled-EDCH-Power-Share" },
  {  21, "received-Scheduled-EDCH-Power-Share-for-cellPortion" },
  {  22, "uTRAN-GANSS-timing-of-cell-frames-for-UE-Positioning" },
  {  23, "eDCH-RACH-report" },
  {  24, "transmittedCarrierPowerOfAllCodesNotUsedForHS-PDSCH-HS-SCCH-E-AGCHOrE-HICHTransmission-for-cellPortion" },
  {  25, "ul-timeslot-iscp-for-cellPortion" },
  {  26, "upPTS-Interference-for-cellPortion" },
  {  27, "e-DCH-Provided-Bit-Rate-for-cellPortion" },
  { 0, NULL }
};

static value_string_ext nbap_CommonMeasurementType_vals_ext = VALUE_STRING_EXT_INIT(nbap_CommonMeasurementType_vals);


static int
dissect_nbap_CommonMeasurementType(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     6, NULL, TRUE, 22, NULL);

  return offset;
}


static const per_sequence_t CommonMeasurementAvailable_sequence[] = {
  { &hf_nbap_commonmeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonMeasurementValue },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommonMeasurementAvailable(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommonMeasurementAvailable, CommonMeasurementAvailable_sequence);

  return offset;
}



static int
dissect_nbap_CommonMeasurementnotAvailable(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string nbap_CommonMeasurementValueInformation_vals[] = {
  {   0, "measurementAvailable" },
  {   1, "measurementnotAvailable" },
  { 0, NULL }
};

static const per_choice_t CommonMeasurementValueInformation_choice[] = {
  {   0, &hf_nbap_measurementAvailable, ASN1_NO_EXTENSIONS     , dissect_nbap_CommonMeasurementAvailable },
  {   1, &hf_nbap_measurementnotAvailable, ASN1_NO_EXTENSIONS     , dissect_nbap_CommonMeasurementnotAvailable },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_CommonMeasurementValueInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_CommonMeasurementValueInformation, CommonMeasurementValueInformation_choice,
                                 NULL);

  return offset;
}



static int
dissect_nbap_CommonPhysicalChannelID768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 511U, NULL, FALSE);

  return offset;
}


static const value_string nbap_ResourceOperationalState_vals[] = {
  {   0, "enabled" },
  {   1, "disabled" },
  { 0, NULL }
};


static int
dissect_nbap_ResourceOperationalState(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t Common_PhysicalChannel_Status_Information_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_resourceOperationalState, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ResourceOperationalState },
  { &hf_nbap_availabilityStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AvailabilityStatus },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_PhysicalChannel_Status_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_PhysicalChannel_Status_Information, Common_PhysicalChannel_Status_Information_sequence);

  return offset;
}


static const per_sequence_t Common_PhysicalChannel_Status_Information768_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID768 },
  { &hf_nbap_resourceOperationalState, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ResourceOperationalState },
  { &hf_nbap_availabilityStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AvailabilityStatus },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_PhysicalChannel_Status_Information768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_PhysicalChannel_Status_Information768, Common_PhysicalChannel_Status_Information768_sequence);

  return offset;
}


static const per_sequence_t CommonTransportChannel_InformationResponse_sequence[] = {
  { &hf_nbap_commonTransportChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonTransportChannelID },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommonTransportChannel_InformationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
address dst_addr;
nbap_setup_conv_t *request_conv;
conversation_t *conv;
guint32 transportLayerAddress_ipv4;
guint16 bindingID;
nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);

nbap_debug("Frame %u CommonTransportChannel-InformationResponse Start", actx->pinfo->num);

nbap_private_data->transportLayerAddress_ipv4 = 0;
nbap_private_data->binding_id_port = 0;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommonTransportChannel_InformationResponse, CommonTransportChannel_InformationResponse_sequence);


  transportLayerAddress_ipv4 = nbap_private_data->transportLayerAddress_ipv4;
  bindingID = nbap_private_data->binding_id_port;
  if (PINFO_FD_VISITED(actx->pinfo) || transportLayerAddress_ipv4 == 0 || bindingID == 0){
    return offset;
  }

  set_address(&dst_addr, AT_IPv4, 4, &transportLayerAddress_ipv4);

  request_conv = find_setup_conv(actx->pinfo, nbap_private_data->transaction_id,nbap_private_data->dd_mode,nbap_private_data->common_transport_channel_id);

  if(request_conv == NULL){
    return offset;
  }

  conv = request_conv->conv;
  conversation_set_addr2(conv, &dst_addr);
  conversation_set_port2(conv, bindingID);

  delete_setup_conv(request_conv);
  nbap_debug("    Frame %u conversation setup frame: %u %s:%u -> %s:%u", actx->pinfo->num, conv->setup_frame,
            address_to_str(actx->pinfo->pool, &conv->key_ptr->addr1), conv->key_ptr->port1,
            address_to_str(actx->pinfo->pool, &conv->key_ptr->addr2), conv->key_ptr->port2);

  nbap_debug("Frame %u CommonTransportChannel-InformationResponse End", actx->pinfo->num);



  return offset;
}


static const per_sequence_t Common_TransportChannel_Status_Information_sequence[] = {
  { &hf_nbap_commonTransportChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonTransportChannelID },
  { &hf_nbap_resourceOperationalState, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ResourceOperationalState },
  { &hf_nbap_availabilityStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AvailabilityStatus },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_TransportChannel_Status_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_TransportChannel_Status_Information, Common_TransportChannel_Status_Information_sequence);

  return offset;
}



static int
dissect_nbap_CommunicationControlPortID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 65535U, NULL, FALSE);

  return offset;
}


static const value_string nbap_Compressed_Mode_Deactivation_Flag_vals[] = {
  {   0, "deactivate" },
  {   1, "maintain-Active" },
  { 0, NULL }
};


static int
dissect_nbap_Compressed_Mode_Deactivation_Flag(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_ConfigurationGenerationID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_ConstantValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -10, 10U, NULL, TRUE);

  return offset;
}


static const value_string nbap_ContinuousPacketConnectivityDTX_DRX_Capability_vals[] = {
  {   0, "continuous-Packet-Connectivity-DTX-DRX-capable" },
  {   1, "continuous-Packet-Connectivity-DTX-DRX-non-capable" },
  { 0, NULL }
};


static int
dissect_nbap_ContinuousPacketConnectivityDTX_DRX_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_UE_DTX_DRX_Offset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 159U, NULL, FALSE);

  return offset;
}


static const value_string nbap_Enabling_Delay_vals[] = {
  {   0, "v0" },
  {   1, "v1" },
  {   2, "v2" },
  {   3, "v4" },
  {   4, "v8" },
  {   5, "v16" },
  {   6, "v32" },
  {   7, "v64" },
  {   8, "v128" },
  { 0, NULL }
};


static int
dissect_nbap_Enabling_Delay(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     9, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_UE_DTX_Cycle1_2ms_vals[] = {
  {   0, "v1" },
  {   1, "v4" },
  {   2, "v5" },
  {   3, "v8" },
  {   4, "v10" },
  {   5, "v16" },
  {   6, "v20" },
  { 0, NULL }
};


static int
dissect_nbap_UE_DTX_Cycle1_2ms(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     7, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_UE_DTX_Cycle2_2ms_vals[] = {
  {   0, "v4" },
  {   1, "v5" },
  {   2, "v8" },
  {   3, "v10" },
  {   4, "v16" },
  {   5, "v20" },
  {   6, "v32" },
  {   7, "v40" },
  {   8, "v64" },
  {   9, "v80" },
  {  10, "v128" },
  {  11, "v160" },
  { 0, NULL }
};


static int
dissect_nbap_UE_DTX_Cycle2_2ms(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     12, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_MAC_DTX_Cycle_2ms_vals[] = {
  {   0, "v1" },
  {   1, "v4" },
  {   2, "v5" },
  {   3, "v8" },
  {   4, "v10" },
  {   5, "v16" },
  {   6, "v20" },
  { 0, NULL }
};


static int
dissect_nbap_MAC_DTX_Cycle_2ms(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     7, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t DTX_Cycle_2ms_Items_sequence[] = {
  { &hf_nbap_uE_DTX_Cycle1_2ms, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UE_DTX_Cycle1_2ms },
  { &hf_nbap_uE_DTX_Cycle2_2ms, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UE_DTX_Cycle2_2ms },
  { &hf_nbap_mAC_DTX_Cycle_2ms, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MAC_DTX_Cycle_2ms },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DTX_Cycle_2ms_Items(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DTX_Cycle_2ms_Items, DTX_Cycle_2ms_Items_sequence);

  return offset;
}


static const value_string nbap_UE_DTX_Cycle1_10ms_vals[] = {
  {   0, "v1" },
  {   1, "v5" },
  {   2, "v10" },
  {   3, "v20" },
  { 0, NULL }
};


static int
dissect_nbap_UE_DTX_Cycle1_10ms(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_UE_DTX_Cycle2_10ms_vals[] = {
  {   0, "v5" },
  {   1, "v10" },
  {   2, "v20" },
  {   3, "v40" },
  {   4, "v80" },
  {   5, "v160" },
  { 0, NULL }
};


static int
dissect_nbap_UE_DTX_Cycle2_10ms(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     6, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_MAC_DTX_Cycle_10ms_vals[] = {
  {   0, "v5" },
  {   1, "v10" },
  {   2, "v20" },
  { 0, NULL }
};


static int
dissect_nbap_MAC_DTX_Cycle_10ms(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t DTX_Cycle_10ms_Items_sequence[] = {
  { &hf_nbap_uE_DTX_Cycle1_10ms, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UE_DTX_Cycle1_10ms },
  { &hf_nbap_uE_DTX_Cycle2_10ms, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UE_DTX_Cycle2_10ms },
  { &hf_nbap_mAC_DTX_Cycle_10ms, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MAC_DTX_Cycle_10ms },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DTX_Cycle_10ms_Items(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DTX_Cycle_10ms_Items, DTX_Cycle_10ms_Items_sequence);

  return offset;
}


static const value_string nbap_E_DCH_TTI_Length_vals[] = {
  {   0, "two-ms" },
  {   1, "ten-ms" },
  { 0, NULL }
};

static const per_choice_t E_DCH_TTI_Length_choice[] = {
  {   0, &hf_nbap_two_ms         , ASN1_EXTENSION_ROOT    , dissect_nbap_DTX_Cycle_2ms_Items },
  {   1, &hf_nbap_ten_ms         , ASN1_EXTENSION_ROOT    , dissect_nbap_DTX_Cycle_10ms_Items },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_E_DCH_TTI_Length(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_E_DCH_TTI_Length, E_DCH_TTI_Length_choice,
                                 NULL);

  return offset;
}


static const value_string nbap_Inactivity_Threshold_for_UE_DTX_Cycle2_vals[] = {
  {   0, "v1" },
  {   1, "v4" },
  {   2, "v8" },
  {   3, "v16" },
  {   4, "v32" },
  {   5, "v64" },
  {   6, "v128" },
  {   7, "v256" },
  { 0, NULL }
};


static int
dissect_nbap_Inactivity_Threshold_for_UE_DTX_Cycle2(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_UE_DTX_Long_Preamble_vals[] = {
  {   0, "v2" },
  {   1, "v4" },
  {   2, "v15" },
  { 0, NULL }
};


static int
dissect_nbap_UE_DTX_Long_Preamble(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_MAC_Inactivity_Threshold_vals[] = {
  {   0, "v1" },
  {   1, "v2" },
  {   2, "v4" },
  {   3, "v8" },
  {   4, "v16" },
  {   5, "v32" },
  {   6, "v64" },
  {   7, "v128" },
  {   8, "v256" },
  {   9, "v512" },
  {  10, "infinity" },
  { 0, NULL }
};


static int
dissect_nbap_MAC_Inactivity_Threshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     11, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_CQI_DTX_Timer_vals[] = {
  {   0, "v0" },
  {   1, "v1" },
  {   2, "v2" },
  {   3, "v4" },
  {   4, "v8" },
  {   5, "v16" },
  {   6, "v32" },
  {   7, "v64" },
  {   8, "v128" },
  {   9, "v256" },
  {  10, "v512" },
  {  11, "infinity" },
  { 0, NULL }
};


static int
dissect_nbap_CQI_DTX_Timer(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     12, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_UE_DPCCH_burst1_vals[] = {
  {   0, "v1" },
  {   1, "v2" },
  {   2, "v5" },
  { 0, NULL }
};


static int
dissect_nbap_UE_DPCCH_burst1(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_UE_DPCCH_burst2_vals[] = {
  {   0, "v1" },
  {   1, "v2" },
  {   2, "v5" },
  { 0, NULL }
};


static int
dissect_nbap_UE_DPCCH_burst2(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t DTX_Information_sequence[] = {
  { &hf_nbap_e_DCH_TTI_Length, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_TTI_Length },
  { &hf_nbap_inactivity_Threshold_for_UE_DTX_Cycle2, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Inactivity_Threshold_for_UE_DTX_Cycle2 },
  { &hf_nbap_uE_DTX_Long_Preamble, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UE_DTX_Long_Preamble },
  { &hf_nbap_mAC_Inactivity_Threshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MAC_Inactivity_Threshold },
  { &hf_nbap_cQI_DTX_Timer  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CQI_DTX_Timer },
  { &hf_nbap_uE_DPCCH_burst1, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UE_DPCCH_burst1 },
  { &hf_nbap_uE_DPCCH_burst2, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UE_DPCCH_burst2 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DTX_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DTX_Information, DTX_Information_sequence);

  return offset;
}


static const value_string nbap_UE_DRX_Cycle_vals[] = {
  {   0, "v4" },
  {   1, "v5" },
  {   2, "v8" },
  {   3, "v10" },
  {   4, "v16" },
  {   5, "v20" },
  { 0, NULL }
};


static int
dissect_nbap_UE_DRX_Cycle(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     6, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_Inactivity_Threshold_for_UE_DRX_Cycle_vals[] = {
  {   0, "v0" },
  {   1, "v1" },
  {   2, "v2" },
  {   3, "v4" },
  {   4, "v8" },
  {   5, "v16" },
  {   6, "v32" },
  {   7, "v64" },
  {   8, "v128" },
  {   9, "v256" },
  {  10, "v512" },
  { 0, NULL }
};


static int
dissect_nbap_Inactivity_Threshold_for_UE_DRX_Cycle(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     11, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_Inactivity_Threshold_for_UE_Grant_Monitoring_vals[] = {
  {   0, "v0" },
  {   1, "v1" },
  {   2, "v2" },
  {   3, "v4" },
  {   4, "v8" },
  {   5, "v16" },
  {   6, "v32" },
  {   7, "v64" },
  {   8, "v128" },
  {   9, "v256" },
  { 0, NULL }
};


static int
dissect_nbap_Inactivity_Threshold_for_UE_Grant_Monitoring(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     10, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_UE_DRX_Grant_Monitoring(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_boolean(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}


static const per_sequence_t DRX_Information_sequence[] = {
  { &hf_nbap_uE_DRX_Cycle   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UE_DRX_Cycle },
  { &hf_nbap_inactivity_Threshold_for_UE_DRX_Cycle, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Inactivity_Threshold_for_UE_DRX_Cycle },
  { &hf_nbap_inactivity_Threshold_for_UE_Grant_Monitoring, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Inactivity_Threshold_for_UE_Grant_Monitoring },
  { &hf_nbap_uE_DRX_Grant_Monitoring, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UE_DRX_Grant_Monitoring },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DRX_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DRX_Information, DRX_Information_sequence);

  return offset;
}


static const per_sequence_t ContinuousPacketConnectivityDTX_DRX_Information_sequence[] = {
  { &hf_nbap_uE_DTX_DRX_Offset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UE_DTX_DRX_Offset },
  { &hf_nbap_enabling_Delay , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Enabling_Delay },
  { &hf_nbap_dTX_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DTX_Information },
  { &hf_nbap_dRX_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DRX_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ContinuousPacketConnectivityDTX_DRX_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ContinuousPacketConnectivityDTX_DRX_Information, ContinuousPacketConnectivityDTX_DRX_Information_sequence);

  return offset;
}


static const per_sequence_t DTX_Cycle_2ms_to_Modify_Items_sequence[] = {
  { &hf_nbap_uE_DTX_Cycle1_2ms, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UE_DTX_Cycle1_2ms },
  { &hf_nbap_uE_DTX_Cycle2_2ms, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UE_DTX_Cycle2_2ms },
  { &hf_nbap_mAC_DTX_Cycle_2ms, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MAC_DTX_Cycle_2ms },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DTX_Cycle_2ms_to_Modify_Items(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DTX_Cycle_2ms_to_Modify_Items, DTX_Cycle_2ms_to_Modify_Items_sequence);

  return offset;
}


static const per_sequence_t DTX_Cycle_10ms_to_Modify_Items_sequence[] = {
  { &hf_nbap_uE_DTX_Cycle1_10ms, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UE_DTX_Cycle1_10ms },
  { &hf_nbap_uE_DTX_Cycle2_10ms, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UE_DTX_Cycle2_10ms },
  { &hf_nbap_mAC_DTX_Cycle_10ms, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MAC_DTX_Cycle_10ms },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DTX_Cycle_10ms_to_Modify_Items(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DTX_Cycle_10ms_to_Modify_Items, DTX_Cycle_10ms_to_Modify_Items_sequence);

  return offset;
}


static const value_string nbap_E_DCH_TTI_Length_to_Modify_vals[] = {
  {   0, "two-ms" },
  {   1, "ten-ms" },
  { 0, NULL }
};

static const per_choice_t E_DCH_TTI_Length_to_Modify_choice[] = {
  {   0, &hf_nbap_two_ms_01      , ASN1_EXTENSION_ROOT    , dissect_nbap_DTX_Cycle_2ms_to_Modify_Items },
  {   1, &hf_nbap_ten_ms_01      , ASN1_EXTENSION_ROOT    , dissect_nbap_DTX_Cycle_10ms_to_Modify_Items },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_E_DCH_TTI_Length_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_E_DCH_TTI_Length_to_Modify, E_DCH_TTI_Length_to_Modify_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t DTX_Information_to_Modify_Items_sequence[] = {
  { &hf_nbap_e_DCH_TTI_Length_to_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_TTI_Length_to_Modify },
  { &hf_nbap_inactivity_Threshold_for_UE_DTX_Cycle2, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Inactivity_Threshold_for_UE_DTX_Cycle2 },
  { &hf_nbap_uE_DTX_Long_Preamble, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UE_DTX_Long_Preamble },
  { &hf_nbap_mAC_Inactivity_Threshold, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MAC_Inactivity_Threshold },
  { &hf_nbap_cQI_DTX_Timer  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CQI_DTX_Timer },
  { &hf_nbap_uE_DPCCH_burst1, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UE_DPCCH_burst1 },
  { &hf_nbap_uE_DPCCH_burst2, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UE_DPCCH_burst2 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DTX_Information_to_Modify_Items(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DTX_Information_to_Modify_Items, DTX_Information_to_Modify_Items_sequence);

  return offset;
}


static const value_string nbap_DTX_Information_to_Modify_vals[] = {
  {   0, "modify" },
  {   1, "deactivate" },
  { 0, NULL }
};

static const per_choice_t DTX_Information_to_Modify_choice[] = {
  {   0, &hf_nbap_modify_02      , ASN1_EXTENSION_ROOT    , dissect_nbap_DTX_Information_to_Modify_Items },
  {   1, &hf_nbap_deactivate     , ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_DTX_Information_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_DTX_Information_to_Modify, DTX_Information_to_Modify_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t DRX_Information_to_Modify_Items_sequence[] = {
  { &hf_nbap_uE_DRX_Cycle   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UE_DRX_Cycle },
  { &hf_nbap_inactivity_Threshold_for_UE_DRX_Cycle, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Inactivity_Threshold_for_UE_DRX_Cycle },
  { &hf_nbap_inactivity_Threshold_for_UE_Grant_Monitoring, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Inactivity_Threshold_for_UE_Grant_Monitoring },
  { &hf_nbap_uE_DRX_Grant_Monitoring, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UE_DRX_Grant_Monitoring },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DRX_Information_to_Modify_Items(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DRX_Information_to_Modify_Items, DRX_Information_to_Modify_Items_sequence);

  return offset;
}


static const value_string nbap_DRX_Information_to_Modify_vals[] = {
  {   0, "modify" },
  {   1, "deactivate" },
  { 0, NULL }
};

static const per_choice_t DRX_Information_to_Modify_choice[] = {
  {   0, &hf_nbap_modify_01      , ASN1_EXTENSION_ROOT    , dissect_nbap_DRX_Information_to_Modify_Items },
  {   1, &hf_nbap_deactivate     , ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_DRX_Information_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_DRX_Information_to_Modify, DRX_Information_to_Modify_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t ContinuousPacketConnectivityDTX_DRX_Information_to_Modify_sequence[] = {
  { &hf_nbap_uE_DTX_DRX_Offset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UE_DTX_DRX_Offset },
  { &hf_nbap_enabling_Delay , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Enabling_Delay },
  { &hf_nbap_dTX_Information_to_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DTX_Information_to_Modify },
  { &hf_nbap_dRX_Information_to_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DRX_Information_to_Modify },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ContinuousPacketConnectivityDTX_DRX_Information_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ContinuousPacketConnectivityDTX_DRX_Information_to_Modify, ContinuousPacketConnectivityDTX_DRX_Information_to_Modify_sequence);

  return offset;
}


static const value_string nbap_ContinuousPacketConnectivityHS_SCCH_less_Capability_vals[] = {
  {   0, "continuous-Packet-Connectivity-HS-SCCH-less-capable" },
  {   1, "continuous-Packet-Connectivity-HS-SCCH-less-capable-non-capable" },
  { 0, NULL }
};


static int
dissect_nbap_ContinuousPacketConnectivityHS_SCCH_less_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_Transport_Block_Size_Index(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxNrOfHS_DSCH_TBSs, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_HSPDSCH_Second_Code_Support(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_boolean(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}


static const per_sequence_t ContinuousPacketConnectivityHS_SCCH_less_InformationItem_sequence[] = {
  { &hf_nbap_transport_Block_Size_Index, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Transport_Block_Size_Index },
  { &hf_nbap_hSPDSCH_Second_Code_Support, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSPDSCH_Second_Code_Support },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ContinuousPacketConnectivityHS_SCCH_less_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ContinuousPacketConnectivityHS_SCCH_less_InformationItem, ContinuousPacketConnectivityHS_SCCH_less_InformationItem_sequence);

  return offset;
}


static const per_sequence_t ContinuousPacketConnectivityHS_SCCH_less_Information_sequence_of[1] = {
  { &hf_nbap_ContinuousPacketConnectivityHS_SCCH_less_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ContinuousPacketConnectivityHS_SCCH_less_InformationItem },
};

static int
dissect_nbap_ContinuousPacketConnectivityHS_SCCH_less_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_ContinuousPacketConnectivityHS_SCCH_less_Information, ContinuousPacketConnectivityHS_SCCH_less_Information_sequence_of,
                                                  1, maxNrOfHS_DSCH_TBSs_HS_SCCHless, FALSE);

  return offset;
}



static int
dissect_nbap_HSPDSCH_First_Code_Index(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxHS_PDSCHCodeNrComp_1, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_HSPDSCH_Second_Code_Index(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxHS_PDSCHCodeNrComp_1, NULL, FALSE);

  return offset;
}


static const per_sequence_t ContinuousPacketConnectivityHS_SCCH_less_Information_Response_sequence[] = {
  { &hf_nbap_hSPDSCH_First_Code_Index, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSPDSCH_First_Code_Index },
  { &hf_nbap_hSPDSCH_Second_Code_Index, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSPDSCH_Second_Code_Index },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ContinuousPacketConnectivityHS_SCCH_less_Information_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ContinuousPacketConnectivityHS_SCCH_less_Information_Response, ContinuousPacketConnectivityHS_SCCH_less_Information_Response_sequence);

  return offset;
}



static int
dissect_nbap_ControlGAP(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t CPC_Information_sequence[] = {
  { &hf_nbap_continuousPacketConnectivityDTX_DRX_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ContinuousPacketConnectivityDTX_DRX_Information },
  { &hf_nbap_continuousPacketConnectivityDTX_DRX_Information_to_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ContinuousPacketConnectivityDTX_DRX_Information_to_Modify },
  { &hf_nbap_continuousPacketConnectivityHS_SCCH_less_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ContinuousPacketConnectivityHS_SCCH_less_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CPC_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CPC_Information, CPC_Information_sequence);

  return offset;
}



static int
dissect_nbap_ContinuousPacketConnectivityHS_SCCH_less_Deactivate_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_RepetitionNumber0(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t CriticalityDiagnostics_IE_List_item_sequence[] = {
  { &hf_nbap_iECriticality  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Criticality },
  { &hf_nbap_iE_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_ID },
  { &hf_nbap_repetitionNumber, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RepetitionNumber0 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CriticalityDiagnostics_IE_List_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CriticalityDiagnostics_IE_List_item, CriticalityDiagnostics_IE_List_item_sequence);

  return offset;
}


static const per_sequence_t CriticalityDiagnostics_IE_List_sequence_of[1] = {
  { &hf_nbap_CriticalityDiagnostics_IE_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_CriticalityDiagnostics_IE_List_item },
};

static int
dissect_nbap_CriticalityDiagnostics_IE_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_CriticalityDiagnostics_IE_List, CriticalityDiagnostics_IE_List_sequence_of,
                                                  1, maxNrOfErrors, FALSE);

  return offset;
}


static const per_sequence_t CriticalityDiagnostics_sequence[] = {
  { &hf_nbap_procedureID    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProcedureID },
  { &hf_nbap_triggeringMessage, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TriggeringMessage },
  { &hf_nbap_procedureCriticality, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Criticality },
  { &hf_nbap_transactionID  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransactionID },
  { &hf_nbap_iEsCriticalityDiagnostics, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CriticalityDiagnostics_IE_List },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CriticalityDiagnostics(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CriticalityDiagnostics, CriticalityDiagnostics_sequence);

  return offset;
}



static int
dissect_nbap_CRNC_CommunicationContextID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
guint32 com_context_id;
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 1048575U, &com_context_id, FALSE);

nbap_get_private_data(actx->pinfo)->com_context_id = com_context_id;
nbap_get_private_data(actx->pinfo)->crnc_context_present = TRUE;


  return offset;
}



static int
dissect_nbap_CSBMeasurementID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 65535U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_CSBTransmissionID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 65535U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_Common_MACFlow_ID_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfCommonMACFlowsLCR_1, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_E_DCH_MACdFlow_ID_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfEDCHMACdFlowsLCR_1, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_E_DCH_HARQ_PO_TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 6U, NULL, FALSE);

  return offset;
}


static const value_string nbap_E_DCH_MACdFlow_Retransmission_Timer_vals[] = {
  {   0, "ms10" },
  {   1, "ms15" },
  {   2, "ms20" },
  {   3, "ms25" },
  {   4, "ms30" },
  {   5, "ms35" },
  {   6, "ms40" },
  {   7, "ms45" },
  {   8, "ms50" },
  {   9, "ms55" },
  {  10, "ms60" },
  {  11, "ms65" },
  {  12, "ms70" },
  {  13, "ms75" },
  {  14, "ms80" },
  {  15, "ms85" },
  {  16, "ms90" },
  {  17, "ms95" },
  {  18, "ms100" },
  {  19, "ms110" },
  {  20, "ms120" },
  {  21, "ms140" },
  {  22, "ms160" },
  {  23, "ms200" },
  {  24, "ms240" },
  {  25, "ms280" },
  {  26, "ms320" },
  {  27, "ms400" },
  {  28, "ms480" },
  {  29, "ms560" },
  { 0, NULL }
};

static value_string_ext nbap_E_DCH_MACdFlow_Retransmission_Timer_vals_ext = VALUE_STRING_EXT_INIT(nbap_E_DCH_MACdFlow_Retransmission_Timer_vals);


static int
dissect_nbap_E_DCH_MACdFlow_Retransmission_Timer(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     30, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t Common_E_DCH_MACdFlow_Specific_InfoList_ItemLCR_sequence[] = {
  { &hf_nbap_common_e_DCH_MACdFlow_ID_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_ID_LCR },
  { &hf_nbap_maximum_Number_of_Retransmissions_For_E_DCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Maximum_Number_of_Retransmissions_For_E_DCH },
  { &hf_nbap_eDCH_MACdFlow_Multiplexing_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_MACdFlow_Multiplexing_List },
  { &hf_nbap_common_E_DCHLogicalChannelInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Common_E_DCH_LogicalChannel_InfoList },
  { &hf_nbap_eDCH_HARQ_PO_TDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_HARQ_PO_TDD },
  { &hf_nbap_eDCH_MACdFlow_Retransmission_Timer, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_MACdFlow_Retransmission_Timer },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_E_DCH_MACdFlow_Specific_InfoList_ItemLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_E_DCH_MACdFlow_Specific_InfoList_ItemLCR, Common_E_DCH_MACdFlow_Specific_InfoList_ItemLCR_sequence);

  return offset;
}


static const per_sequence_t Common_E_DCH_MACdFlow_Specific_InfoListLCR_sequence_of[1] = {
  { &hf_nbap_Common_E_DCH_MACdFlow_Specific_InfoListLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Common_E_DCH_MACdFlow_Specific_InfoList_ItemLCR },
};

static int
dissect_nbap_Common_E_DCH_MACdFlow_Specific_InfoListLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Common_E_DCH_MACdFlow_Specific_InfoListLCR, Common_E_DCH_MACdFlow_Specific_InfoListLCR_sequence_of,
                                                  1, maxNrOfEDCHMACdFlowsLCR, FALSE);

  return offset;
}


static const per_sequence_t Ul_common_E_DCH_MACflow_Specific_InfoList_ItemLCR_sequence[] = {
  { &hf_nbap_ul_Common_MACFlowIDLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Common_MACFlow_ID_LCR },
  { &hf_nbap_transportBearerRequestIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportBearerRequestIndicator },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_tnlQos         , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TnlQos },
  { &hf_nbap_payloadCRC_PresenceIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PayloadCRC_PresenceIndicator },
  { &hf_nbap_common_E_DCH_MACdFlow_Specific_InformationLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Common_E_DCH_MACdFlow_Specific_InfoListLCR },
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UARFCN },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Ul_common_E_DCH_MACflow_Specific_InfoList_ItemLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoList_ItemLCR, Ul_common_E_DCH_MACflow_Specific_InfoList_ItemLCR_sequence);

  return offset;
}


static const per_sequence_t Ul_common_E_DCH_MACflow_Specific_InfoListLCR_sequence_of[1] = {
  { &hf_nbap_Ul_common_E_DCH_MACflow_Specific_InfoListLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Ul_common_E_DCH_MACflow_Specific_InfoList_ItemLCR },
};

static int
dissect_nbap_Ul_common_E_DCH_MACflow_Specific_InfoListLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoListLCR, Ul_common_E_DCH_MACflow_Specific_InfoListLCR_sequence_of,
                                                  1, maxNrOfCommonMACFlows, FALSE);

  return offset;
}



static int
dissect_nbap_PRXdes_base(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -112, -50, NULL, FALSE);

  return offset;
}


static const per_sequence_t PRXdes_base_Item_sequence[] = {
  { &hf_nbap_pRXdes_base    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PRXdes_base },
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UARFCN },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PRXdes_base_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PRXdes_base_Item, PRXdes_base_Item_sequence);

  return offset;
}


static const per_sequence_t PRXdes_base_perURAFCN_sequence_of[1] = {
  { &hf_nbap_PRXdes_base_perURAFCN_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_PRXdes_base_Item },
};

static int
dissect_nbap_PRXdes_base_perURAFCN(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_PRXdes_base_perURAFCN, PRXdes_base_perURAFCN_sequence_of,
                                                  1, maxFrequencyinCell, FALSE);

  return offset;
}


static const value_string nbap_TDD_TPC_UplinkStepSize_LCR_vals[] = {
  {   0, "step-size1" },
  {   1, "step-size2" },
  {   2, "step-size3" },
  { 0, NULL }
};


static int
dissect_nbap_TDD_TPC_UplinkStepSize_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_TDD_TPC_DownlinkStepSize_vals[] = {
  {   0, "step-size1" },
  {   1, "step-size2" },
  {   2, "step-size3" },
  { 0, NULL }
};


static int
dissect_nbap_TDD_TPC_DownlinkStepSize(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t Common_E_PUCH_InformationLCR_sequence[] = {
  { &hf_nbap_minCR          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CodeRate },
  { &hf_nbap_maxCR          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CodeRate },
  { &hf_nbap_harqInfo       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HARQ_Info_for_E_DCH },
  { &hf_nbap_pRXdes_base_perURAFCN, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PRXdes_base_perURAFCN },
  { &hf_nbap_e_PUCH_TPC_StepSize, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_TPC_UplinkStepSize_LCR },
  { &hf_nbap_e_AGCH_TPC_StepSize, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_TPC_DownlinkStepSize },
  { &hf_nbap_e_PUCH_PowerControlGAP, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ControlGAP },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_E_PUCH_InformationLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_E_PUCH_InformationLCR, Common_E_PUCH_InformationLCR_sequence);

  return offset;
}



static int
dissect_nbap_RefBeta(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -15, 16U, NULL, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_RefBeta_Item_sequence[] = {
  { &hf_nbap_refCodeRate    , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_CodeRate_short },
  { &hf_nbap_refBeta        , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_RefBeta },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_RefBeta_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_RefBeta_Item, E_DCH_RefBeta_Item_sequence);

  return offset;
}


static const per_sequence_t E_DCH_QPSK_RefBetaInfo_sequence_of[1] = {
  { &hf_nbap_E_DCH_QPSK_RefBetaInfo_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_RefBeta_Item },
};

static int
dissect_nbap_E_DCH_QPSK_RefBetaInfo(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_DCH_QPSK_RefBetaInfo, E_DCH_QPSK_RefBetaInfo_sequence_of,
                                                  1, maxNrOfRefBetas, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_sixteenQAM_RefBetaInfo_sequence_of[1] = {
  { &hf_nbap_E_DCH_sixteenQAM_RefBetaInfo_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_RefBeta_Item },
};

static int
dissect_nbap_E_DCH_sixteenQAM_RefBetaInfo(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_DCH_sixteenQAM_RefBetaInfo, E_DCH_sixteenQAM_RefBetaInfo_sequence_of,
                                                  1, maxNrOfRefBetas, FALSE);

  return offset;
}


static const per_sequence_t E_TFCS_Information_TDD_sequence[] = {
  { &hf_nbap_e_DCH_QPSK_RefBetaInfo, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_QPSK_RefBetaInfo },
  { &hf_nbap_e_DCH_sixteenQAM_RefBetaInfo, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_sixteenQAM_RefBetaInfo },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_TFCS_Information_TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_TFCS_Information_TDD, E_TFCS_Information_TDD_sequence);

  return offset;
}


static const per_sequence_t Common_EDCH_System_InformationLCR_sequence[] = {
  { &hf_nbap_ul_common_E_DCH_MACflow_Specific_InformationLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Ul_common_E_DCH_MACflow_Specific_InfoListLCR },
  { &hf_nbap_common_E_PUCH_InformationLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Common_E_PUCH_InformationLCR },
  { &hf_nbap_e_TFCS_Information_TDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_TFCS_Information_TDD },
  { &hf_nbap_maximum_Number_of_Retransmissions_For_SchedulingInfo, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Maximum_Number_of_Retransmissions_For_E_DCH },
  { &hf_nbap_eDCH_Retransmission_Timer_SchedulingInfo, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_MACdFlow_Retransmission_Timer },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_EDCH_System_InformationLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_EDCH_System_InformationLCR, Common_EDCH_System_InformationLCR_sequence);

  return offset;
}


static const per_sequence_t Ul_common_E_DCH_MACflow_Specific_InfoResponseList_ItemLCR_sequence[] = {
  { &hf_nbap_ul_Common_MACFlowID_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Common_MACFlow_ID_LCR },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UARFCN },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseList_ItemLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseList_ItemLCR, Ul_common_E_DCH_MACflow_Specific_InfoResponseList_ItemLCR_sequence);

  return offset;
}


static const per_sequence_t Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_sequence_of[1] = {
  { &hf_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseList_ItemLCR },
};

static int
dissect_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR, Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_sequence_of,
                                                  1, maxNrOfCommonMACFlows, FALSE);

  return offset;
}



static int
dissect_nbap_E_AGCH_Id(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 31U, NULL, TRUE);

  return offset;
}


static const per_sequence_t Common_E_AGCH_ItemLCR_sequence[] = {
  { &hf_nbap_e_AGCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_AGCH_Id },
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UARFCN },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_E_AGCH_ItemLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_E_AGCH_ItemLCR, Common_E_AGCH_ItemLCR_sequence);

  return offset;
}


static const per_sequence_t Common_E_AGCH_ListLCR_sequence_of[1] = {
  { &hf_nbap_Common_E_AGCH_ListLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Common_E_AGCH_ItemLCR },
};

static int
dissect_nbap_Common_E_AGCH_ListLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Common_E_AGCH_ListLCR, Common_E_AGCH_ListLCR_sequence_of,
                                                  1, maxNrOfEAGCHsLCR, FALSE);

  return offset;
}



static int
dissect_nbap_EI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 3U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_E_HICH_ID_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Common_E_HICH_ItemLCR_sequence[] = {
  { &hf_nbap_eI             , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_EI },
  { &hf_nbap_e_HICH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_HICH_ID_LCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_E_HICH_ItemLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_E_HICH_ItemLCR, Common_E_HICH_ItemLCR_sequence);

  return offset;
}


static const per_sequence_t Common_E_HICH_ListLCR_sequence_of[1] = {
  { &hf_nbap_Common_E_HICH_ListLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Common_E_HICH_ItemLCR },
};

static int
dissect_nbap_Common_E_HICH_ListLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Common_E_HICH_ListLCR, Common_E_HICH_ListLCR_sequence_of,
                                                  1, maxNrOfEHICHsLCR, FALSE);

  return offset;
}



static int
dissect_nbap_INTEGER_1_32(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 32U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_INTEGER_1_7(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 7U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Common_E_RNTI_Info_ItemLCR_sequence[] = {
  { &hf_nbap_starting_E_RNTI, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_RNTI },
  { &hf_nbap_number_of_Group, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_1_32 },
  { &hf_nbap_number_of_e_E_RNTI_perGroup, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_1_7 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_E_RNTI_Info_ItemLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_E_RNTI_Info_ItemLCR, Common_E_RNTI_Info_ItemLCR_sequence);

  return offset;
}


static const per_sequence_t Common_E_RNTI_Info_LCR_sequence_of[1] = {
  { &hf_nbap_Common_E_RNTI_Info_LCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Common_E_RNTI_Info_ItemLCR },
};

static int
dissect_nbap_Common_E_RNTI_Info_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Common_E_RNTI_Info_LCR, Common_E_RNTI_Info_LCR_sequence_of,
                                                  1, maxnrofERUCCHsLCR, FALSE);

  return offset;
}


static const per_sequence_t Common_EDCH_System_Information_ResponseLCR_sequence[] = {
  { &hf_nbap_ul_common_E_DCH_MACflow_Specific_InfoResponseLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR },
  { &hf_nbap_common_E_AGCH_ListLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Common_E_AGCH_ListLCR },
  { &hf_nbap_common_E_HICH_ListLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Common_E_HICH_ListLCR },
  { &hf_nbap_common_E_RNTI_Info_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Common_E_RNTI_Info_LCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_EDCH_System_Information_ResponseLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_EDCH_System_Information_ResponseLCR, Common_EDCH_System_Information_ResponseLCR_sequence);

  return offset;
}


static const per_sequence_t Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_Ext_sequence_of[1] = {
  { &hf_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_Ext_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseList_ItemLCR },
};

static int
dissect_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_Ext(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_Ext, Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_Ext_sequence_of,
                                                  1, maxNrOfCommonMACFlowsLCRExt, FALSE);

  return offset;
}


static const per_sequence_t Common_MACFlows_to_DeleteLCR_Item_sequence[] = {
  { &hf_nbap_common_MACFlow_ID_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Common_MACFlow_ID_LCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_MACFlows_to_DeleteLCR_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_MACFlows_to_DeleteLCR_Item, Common_MACFlows_to_DeleteLCR_Item_sequence);

  return offset;
}


static const per_sequence_t Common_MACFlows_to_DeleteLCR_sequence_of[1] = {
  { &hf_nbap_Common_MACFlows_to_DeleteLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Common_MACFlows_to_DeleteLCR_Item },
};

static int
dissect_nbap_Common_MACFlows_to_DeleteLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Common_MACFlows_to_DeleteLCR, Common_MACFlows_to_DeleteLCR_sequence_of,
                                                  1, maxNrOfCommonMACFlowsLCR, FALSE);

  return offset;
}


static const per_sequence_t CommonMACFlow_Specific_InfoItemLCR_sequence[] = {
  { &hf_nbap_common_MACFlow_ID_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Common_MACFlow_ID_LCR },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_tnl_qos        , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TnlQos },
  { &hf_nbap_common_MACFlow_PriorityQueue_InformationLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Common_MACFlow_PriorityQueue_Information },
  { &hf_nbap_transportBearerRequestIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportBearerRequestIndicator },
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UARFCN },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommonMACFlow_Specific_InfoItemLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommonMACFlow_Specific_InfoItemLCR, CommonMACFlow_Specific_InfoItemLCR_sequence);

  return offset;
}


static const per_sequence_t CommonMACFlow_Specific_InfoListLCR_sequence_of[1] = {
  { &hf_nbap_CommonMACFlow_Specific_InfoListLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_CommonMACFlow_Specific_InfoItemLCR },
};

static int
dissect_nbap_CommonMACFlow_Specific_InfoListLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_CommonMACFlow_Specific_InfoListLCR, CommonMACFlow_Specific_InfoListLCR_sequence_of,
                                                  1, maxNrOfCommonMACFlowsLCR, FALSE);

  return offset;
}


static const per_sequence_t Common_H_RNTI_InfoItemLCR_sequence[] = {
  { &hf_nbap_common_H_RNTI  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_RNTI },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_H_RNTI_InfoItemLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_H_RNTI_InfoItemLCR, Common_H_RNTI_InfoItemLCR_sequence);

  return offset;
}


static const per_sequence_t Common_H_RNTI_InformationLCR_sequence_of[1] = {
  { &hf_nbap_Common_H_RNTI_InformationLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Common_H_RNTI_InfoItemLCR },
};

static int
dissect_nbap_Common_H_RNTI_InformationLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Common_H_RNTI_InformationLCR, Common_H_RNTI_InformationLCR_sequence_of,
                                                  1, maxNoOfCommonH_RNTI, FALSE);

  return offset;
}


static const value_string nbap_T_SYNC_vals[] = {
  {   0, "v40" },
  {   1, "v80" },
  {   2, "v120" },
  {   3, "v160" },
  {   4, "v200" },
  {   5, "v300" },
  {   6, "v400" },
  {   7, "v500" },
  { 0, NULL }
};


static int
dissect_nbap_T_SYNC(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_T_PROTECT_vals[] = {
  {   0, "v40" },
  {   1, "v60" },
  {   2, "v80" },
  {   3, "v100" },
  {   4, "v120" },
  {   5, "v200" },
  {   6, "v400" },
  { 0, NULL }
};


static int
dissect_nbap_T_PROTECT(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     7, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_N_PROTECT(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 7U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Sync_InformationLCR_sequence[] = {
  { &hf_nbap_t_SYNC         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_SYNC },
  { &hf_nbap_t_PROTECT      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_PROTECT },
  { &hf_nbap_n_PROTECT      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_N_PROTECT },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Sync_InformationLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Sync_InformationLCR, Sync_InformationLCR_sequence);

  return offset;
}


static const per_sequence_t CommonMACFlow_Specific_InfoItem_ResponseLCR_sequence[] = {
  { &hf_nbap_common_MACFlow_ID_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Common_MACFlow_ID_LCR },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_hSDSCH_Initial_Capacity_Allocation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSDSCH_Initial_Capacity_Allocation },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommonMACFlow_Specific_InfoItem_ResponseLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommonMACFlow_Specific_InfoItem_ResponseLCR, CommonMACFlow_Specific_InfoItem_ResponseLCR_sequence);

  return offset;
}


static const per_sequence_t CommonMACFlow_Specific_InfoList_ResponseLCR_sequence_of[1] = {
  { &hf_nbap_CommonMACFlow_Specific_InfoList_ResponseLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_CommonMACFlow_Specific_InfoItem_ResponseLCR },
};

static int
dissect_nbap_CommonMACFlow_Specific_InfoList_ResponseLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_CommonMACFlow_Specific_InfoList_ResponseLCR, CommonMACFlow_Specific_InfoList_ResponseLCR_sequence_of,
                                                  1, maxNrOfCommonMACFlows, FALSE);

  return offset;
}


static const value_string nbap_UE_DRX_Cycle_LCR_vals[] = {
  {   0, "v1" },
  {   1, "v2" },
  {   2, "v4" },
  {   3, "v8" },
  {   4, "v16" },
  {   5, "v32" },
  {   6, "v64" },
  { 0, NULL }
};


static int
dissect_nbap_UE_DRX_Cycle_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     7, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_Inactivity_Threshold_for_UE_DRX_Cycle_LCR_vals[] = {
  {   0, "v1" },
  {   1, "v2" },
  {   2, "v4" },
  {   3, "v8" },
  {   4, "v16" },
  {   5, "v32" },
  {   6, "v64" },
  { 0, NULL }
};


static int
dissect_nbap_Inactivity_Threshold_for_UE_DRX_Cycle_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     7, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_UE_DRX_Offset_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 63U, NULL, FALSE);

  return offset;
}


static const per_sequence_t HS_SCCH_DRX_Information_LCR_sequence[] = {
  { &hf_nbap_hS_SCCH_UE_DRX_Cycle_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UE_DRX_Cycle_LCR },
  { &hf_nbap_hS_SCCH_Inactivity_Threshold_for_UE_DRX_Cycle_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Inactivity_Threshold_for_UE_DRX_Cycle_LCR },
  { &hf_nbap_hS_SCCH_UE_DRX_Offset_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UE_DRX_Offset_LCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_SCCH_DRX_Information_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_SCCH_DRX_Information_LCR, HS_SCCH_DRX_Information_LCR_sequence);

  return offset;
}


static const value_string nbap_E_AGCH_UE_Inactivity_Monitor_Threshold_vals[] = {
  {   0, "v0" },
  {   1, "v1" },
  {   2, "v2" },
  {   3, "v4" },
  {   4, "v8" },
  {   5, "v16" },
  {   6, "v32" },
  {   7, "v64" },
  {   8, "v128" },
  {   9, "v256" },
  {  10, "v512" },
  {  11, "infinity" },
  { 0, NULL }
};

static value_string_ext nbap_E_AGCH_UE_Inactivity_Monitor_Threshold_vals_ext = VALUE_STRING_EXT_INIT(nbap_E_AGCH_UE_Inactivity_Monitor_Threshold_vals);


static int
dissect_nbap_E_AGCH_UE_Inactivity_Monitor_Threshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     12, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t E_AGCH_DRX_Parameters_sequence[] = {
  { &hf_nbap_e_AGCH_UE_DRX_Cycle_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UE_DRX_Cycle_LCR },
  { &hf_nbap_e_AGCH_UE_Inactivity_Monitor_Threshold, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_AGCH_UE_Inactivity_Monitor_Threshold },
  { &hf_nbap_e_AGCH_UE_DRX_Offset_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UE_DRX_Offset_LCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_AGCH_DRX_Parameters(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_AGCH_DRX_Parameters, E_AGCH_DRX_Parameters_sequence);

  return offset;
}


static const value_string nbap_E_AGCH_DRX_Information_LCR_vals[] = {
  {   0, "sameAsHS-SCCH" },
  {   1, "e-AGCH-DRX-Parameters" },
  { 0, NULL }
};

static const per_choice_t E_AGCH_DRX_Information_LCR_choice[] = {
  {   0, &hf_nbap_sameAsHS_SCCH  , ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {   1, &hf_nbap_e_AGCH_DRX_Parameters, ASN1_EXTENSION_ROOT    , dissect_nbap_E_AGCH_DRX_Parameters },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_E_AGCH_DRX_Information_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_E_AGCH_DRX_Information_LCR, E_AGCH_DRX_Information_LCR_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t ContinuousPacketConnectivity_DRX_InformationLCR_sequence[] = {
  { &hf_nbap_enabling_Delay , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Enabling_Delay },
  { &hf_nbap_hS_SCCH_DRX_Information_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SCCH_DRX_Information_LCR },
  { &hf_nbap_e_AGCH_DRX_Information_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_AGCH_DRX_Information_LCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ContinuousPacketConnectivity_DRX_InformationLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ContinuousPacketConnectivity_DRX_InformationLCR, ContinuousPacketConnectivity_DRX_InformationLCR_sequence);

  return offset;
}


static const per_sequence_t DRX_Information_to_Modify_Items_LCR_sequence[] = {
  { &hf_nbap_hS_SCCH_DRX_Information_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_SCCH_DRX_Information_LCR },
  { &hf_nbap_e_AGCH_DRX_Information_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_AGCH_DRX_Information_LCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DRX_Information_to_Modify_Items_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DRX_Information_to_Modify_Items_LCR, DRX_Information_to_Modify_Items_LCR_sequence);

  return offset;
}


static const value_string nbap_DRX_Information_to_Modify_LCR_vals[] = {
  {   0, "modify" },
  {   1, "deactivate" },
  { 0, NULL }
};

static const per_choice_t DRX_Information_to_Modify_LCR_choice[] = {
  {   0, &hf_nbap_modify         , ASN1_EXTENSION_ROOT    , dissect_nbap_DRX_Information_to_Modify_Items_LCR },
  {   1, &hf_nbap_deactivate     , ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_DRX_Information_to_Modify_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_DRX_Information_to_Modify_LCR, DRX_Information_to_Modify_LCR_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t ContinuousPacketConnectivity_DRX_Information_to_Modify_LCR_sequence[] = {
  { &hf_nbap_enabling_Delay , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Enabling_Delay },
  { &hf_nbap_dRX_Information_to_Modify_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DRX_Information_to_Modify_LCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ContinuousPacketConnectivity_DRX_Information_to_Modify_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ContinuousPacketConnectivity_DRX_Information_to_Modify_LCR, ContinuousPacketConnectivity_DRX_Information_to_Modify_LCR_sequence);

  return offset;
}



static int
dissect_nbap_Transport_Block_Size_maping_Index_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNoOfTBSs_Mapping_HS_DSCH_SPS_1, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_Transport_Block_Size_Index_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxNoOfHS_DSCH_TBSsLCR, NULL, FALSE);

  return offset;
}


static const per_sequence_t Transport_Block_Size_Item_LCR_sequence[] = {
  { &hf_nbap_transport_Block_Size_maping_Index_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Transport_Block_Size_maping_Index_LCR },
  { &hf_nbap_transport_Block_Size_Index_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Transport_Block_Size_Index_LCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Transport_Block_Size_Item_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Transport_Block_Size_Item_LCR, Transport_Block_Size_Item_LCR_sequence);

  return offset;
}


static const per_sequence_t Transport_Block_Size_List_LCR_sequence_of[1] = {
  { &hf_nbap_Transport_Block_Size_List_LCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Transport_Block_Size_Item_LCR },
};

static int
dissect_nbap_Transport_Block_Size_List_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Transport_Block_Size_List_LCR, Transport_Block_Size_List_LCR_sequence_of,
                                                  1, maxNoOfTBSs_Mapping_HS_DSCH_SPS, FALSE);

  return offset;
}



static int
dissect_nbap_RepetitionPeriodIndex(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNoOfRepetitionPeriod_SPS_LCR_1, NULL, FALSE);

  return offset;
}


static const value_string nbap_RepetitionPeriod_vals[] = {
  {   0, "v1" },
  {   1, "v2" },
  {   2, "v4" },
  {   3, "v8" },
  {   4, "v16" },
  {   5, "v32" },
  {   6, "v64" },
  { 0, NULL }
};


static int
dissect_nbap_RepetitionPeriod(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     7, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_RepetitionLength(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 63U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Repetition_Period_Item_LCR_sequence[] = {
  { &hf_nbap_repetitionPeriodIndex, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriodIndex },
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RepetitionLength },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Repetition_Period_Item_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Repetition_Period_Item_LCR, Repetition_Period_Item_LCR_sequence);

  return offset;
}


static const per_sequence_t Repetition_Period_List_LCR_sequence_of[1] = {
  { &hf_nbap_Repetition_Period_List_LCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Repetition_Period_Item_LCR },
};

static int
dissect_nbap_Repetition_Period_List_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Repetition_Period_List_LCR, Repetition_Period_List_LCR_sequence_of,
                                                  1, maxNoOfRepetition_Period_LCR, FALSE);

  return offset;
}


static const value_string nbap_SPS_Reservation_Indicator_vals[] = {
  {   0, "reserve" },
  { 0, NULL }
};


static int
dissect_nbap_SPS_Reservation_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_LogicalChannellevel(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     16, 16, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_PriorityQueuelevel(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     8, 8, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const value_string nbap_HS_DSCH_SPS_Operation_Indicator_vals[] = {
  {   0, "logicalChannellevel" },
  {   1, "priorityQueuelevel" },
  { 0, NULL }
};

static const per_choice_t HS_DSCH_SPS_Operation_Indicator_choice[] = {
  {   0, &hf_nbap_logicalChannellevel, ASN1_EXTENSION_ROOT    , dissect_nbap_LogicalChannellevel },
  {   1, &hf_nbap_priorityQueuelevel, ASN1_EXTENSION_ROOT    , dissect_nbap_PriorityQueuelevel },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_HS_DSCH_SPS_Operation_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_HS_DSCH_SPS_Operation_Indicator, HS_DSCH_SPS_Operation_Indicator_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t HS_DSCH_Semi_PersistentScheduling_Information_LCR_sequence[] = {
  { &hf_nbap_transport_Block_Size_List_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Transport_Block_Size_List_LCR },
  { &hf_nbap_repetition_Period_List_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Repetition_Period_List_LCR },
  { &hf_nbap_hS_DSCH_SPS_Reservation_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SPS_Reservation_Indicator },
  { &hf_nbap_hS_DSCH_SPS_Operation_Indicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCH_SPS_Operation_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_DSCH_Semi_PersistentScheduling_Information_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_DSCH_Semi_PersistentScheduling_Information_LCR, HS_DSCH_Semi_PersistentScheduling_Information_LCR_sequence);

  return offset;
}


static const per_sequence_t HS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR_sequence[] = {
  { &hf_nbap_transport_Block_Size_List_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Transport_Block_Size_List_LCR },
  { &hf_nbap_repetition_Period_List_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Repetition_Period_List_LCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR, HS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR_sequence);

  return offset;
}



static int
dissect_nbap_E_DCH_SPS_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     16, 16, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_E_HICH_ID_TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 31U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_SignatureSequenceGroupIndex(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 19U, NULL, FALSE);

  return offset;
}


static const per_sequence_t E_HICH_LCR_Information_sequence[] = {
  { &hf_nbap_e_HICH_ID_TDD  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_HICH_ID_TDD },
  { &hf_nbap_signatureSequenceGroupIndex, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SignatureSequenceGroupIndex },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_HICH_LCR_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_HICH_LCR_Information, E_HICH_LCR_Information_sequence);

  return offset;
}


static const per_sequence_t E_DCH_Semi_PersistentScheduling_Information_LCR_sequence[] = {
  { &hf_nbap_repetition_Period_List_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Repetition_Period_List_LCR },
  { &hf_nbap_e_DCH_SPS_Indicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_SPS_Indicator },
  { &hf_nbap_sPS_E_DCH_releted_E_HICH_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_HICH_LCR_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_Semi_PersistentScheduling_Information_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_Semi_PersistentScheduling_Information_LCR, E_DCH_Semi_PersistentScheduling_Information_LCR_sequence);

  return offset;
}


static const per_sequence_t E_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR_sequence[] = {
  { &hf_nbap_repetition_Period_List_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Repetition_Period_List_LCR },
  { &hf_nbap_e_DCH_SPS_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_SPS_Indicator },
  { &hf_nbap_sPS_E_DCH_releted_E_HICH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_HICH_LCR_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR, E_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR_sequence);

  return offset;
}


static const per_sequence_t CPC_InformationLCR_sequence[] = {
  { &hf_nbap_continuousPacketConnectivity_DRX_InformationLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ContinuousPacketConnectivity_DRX_InformationLCR },
  { &hf_nbap_continuousPacketConnectivity_DRX_Information_to_Modify_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ContinuousPacketConnectivity_DRX_Information_to_Modify_LCR },
  { &hf_nbap_hS_DSCH_Semi_PersistentScheduling_Information_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_DSCH_Semi_PersistentScheduling_Information_LCR },
  { &hf_nbap_hS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR },
  { &hf_nbap_hS_DSCH_SPS_Deactivate_Indicator_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_e_DCH_Semi_PersistentScheduling_Information_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_Semi_PersistentScheduling_Information_LCR },
  { &hf_nbap_e_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR },
  { &hf_nbap_e_DCH_SPS_Deactivate_Indicator_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CPC_InformationLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CPC_InformationLCR, CPC_InformationLCR_sequence);

  return offset;
}


static const value_string nbap_ContinuousPacketConnectivity_DRX_CapabilityLCR_vals[] = {
  {   0, "continuous-Packet-Connectivity-DRX-Capable" },
  {   1, "continuous-Packet-Connectivity-DRX-Non-Capable" },
  { 0, NULL }
};


static int
dissect_nbap_ContinuousPacketConnectivity_DRX_CapabilityLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t HS_SCCH_DRX_Information_ResponseLCR_sequence[] = {
  { &hf_nbap_hS_SCCH_UE_DRX_Cycle_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UE_DRX_Cycle_LCR },
  { &hf_nbap_hS_SCCH_Inactivity_Threshold_for_UE_DRX_Cycle_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Inactivity_Threshold_for_UE_DRX_Cycle_LCR },
  { &hf_nbap_hS_SCCH_UE_DRX_Offset_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UE_DRX_Offset_LCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_SCCH_DRX_Information_ResponseLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_SCCH_DRX_Information_ResponseLCR, HS_SCCH_DRX_Information_ResponseLCR_sequence);

  return offset;
}


static const per_sequence_t E_AGCH_DRX_Parameters_Response_sequence[] = {
  { &hf_nbap_e_AGCH_UE_DRX_Cycle_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UE_DRX_Cycle_LCR },
  { &hf_nbap_e_AGCH_UE_Inactivity_Monitor_Threshold, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_AGCH_UE_Inactivity_Monitor_Threshold },
  { &hf_nbap_e_AGCH_UE_DRX_Offset_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UE_DRX_Offset_LCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_AGCH_DRX_Parameters_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_AGCH_DRX_Parameters_Response, E_AGCH_DRX_Parameters_Response_sequence);

  return offset;
}


static const value_string nbap_E_AGCH_DRX_Information_ResponseLCR_vals[] = {
  {   0, "sameAsHS-SCCH" },
  {   1, "e-AGCH-DRX-Parameters-Response" },
  { 0, NULL }
};

static const per_choice_t E_AGCH_DRX_Information_ResponseLCR_choice[] = {
  {   0, &hf_nbap_sameAsHS_SCCH  , ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {   1, &hf_nbap_e_AGCH_DRX_Parameters_Response, ASN1_EXTENSION_ROOT    , dissect_nbap_E_AGCH_DRX_Parameters_Response },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_E_AGCH_DRX_Information_ResponseLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_E_AGCH_DRX_Information_ResponseLCR, E_AGCH_DRX_Information_ResponseLCR_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t ContinuousPacketConnectivity_DRX_Information_ResponseLCR_sequence[] = {
  { &hf_nbap_enabling_Delay , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Enabling_Delay },
  { &hf_nbap_hS_SCCH_DRX_Information_ResponseLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_SCCH_DRX_Information_ResponseLCR },
  { &hf_nbap_e_AGCH_DRX_Information_ResponseLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_AGCH_DRX_Information_ResponseLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ContinuousPacketConnectivity_DRX_Information_ResponseLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ContinuousPacketConnectivity_DRX_Information_ResponseLCR, ContinuousPacketConnectivity_DRX_Information_ResponseLCR_sequence);

  return offset;
}



static int
dissect_nbap_DATA_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 3U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_DCH_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, &nbap_get_private_data(actx->pinfo)->t_dch_id, FALSE);

  return offset;
}


static const value_string nbap_UL_FP_Mode_vals[] = {
  {   0, "normal" },
  {   1, "silent" },
  { 0, NULL }
};


static int
dissect_nbap_UL_FP_Mode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_ToAWS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 1279U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_ToAWE(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 2559U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_T_dCH_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  guint32 dch_id;
  gint num_dch_in_flow;
  nbap_dch_channel_info_t* nbap_dch_chnl_info;
  nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
  nbap_dch_chnl_info = nbap_private_data->nbap_dch_chnl_info;
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, &dch_id, FALSE);

  num_dch_in_flow = nbap_private_data->num_dch_in_flow;
  nbap_private_data->dch_id = dch_id;
  if(num_dch_in_flow>0){
      guint32 prev_dch_id = nbap_private_data->prev_dch_id;
      nbap_dch_chnl_info[dch_id].next_dch = 0;
      if(prev_dch_id != 0 && prev_dch_id != 0xffffffff && prev_dch_id != dch_id){
          nbap_dch_chnl_info[prev_dch_id].next_dch = dch_id;
      }
  }


  return offset;
}



static int
dissect_nbap_TransportFormatSet_NrOfTransportBlocks(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
guint32 NrOfTransportBlocks;
guint num_items;
guint32 dch_id;
guint32 common_physical_channel_id;
guint32 common_transport_channel_id;
nbap_dch_channel_info_t* nbap_dch_chnl_info;
nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
nbap_dch_chnl_info = nbap_private_data->nbap_dch_chnl_info;

  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 512U, &NrOfTransportBlocks, FALSE);


    if(nbap_private_data->num_items>0){
        num_items = nbap_private_data->num_items;
        dch_id = nbap_private_data->dch_id;
        if (num_items > 0 && num_items < MAX_FP_CHANS + 1 && dch_id != 0xffffffff) {
            common_physical_channel_id = nbap_private_data->common_physical_channel_id;
            common_transport_channel_id = nbap_private_data->common_transport_channel_id;
            switch(nbap_private_data->transport_format_set_type){
                case NBAP_DCH_UL:
                    nbap_dch_chnl_info[dch_id].num_ul_chans++;
                    nbap_dch_chnl_info[dch_id].ul_chan_num_tbs[num_items-1] = NrOfTransportBlocks;
                    break;
                case NBAP_DCH_DL:
                    nbap_dch_chnl_info[dch_id].num_dl_chans++;
                    nbap_dch_chnl_info[dch_id].dl_chan_num_tbs[num_items-1] = NrOfTransportBlocks;
                    break;
                case NBAP_CPCH:
                    nbap_dch_chnl_info[common_physical_channel_id].num_ul_chans++;
                    nbap_dch_chnl_info[common_physical_channel_id].ul_chan_num_tbs[num_items-1] = NrOfTransportBlocks;
                    nbap_dch_chnl_info[common_physical_channel_id].num_dl_chans++;
                    nbap_dch_chnl_info[common_physical_channel_id].dl_chan_num_tbs[num_items-1] = NrOfTransportBlocks;
                    break;
                case NBAP_PCH:
                    nbap_dch_chnl_info[common_transport_channel_id].num_ul_chans++;
                    nbap_dch_chnl_info[common_transport_channel_id].ul_chan_num_tbs[num_items-1] = NrOfTransportBlocks;
                    nbap_dch_chnl_info[common_transport_channel_id].num_dl_chans++;
                    nbap_dch_chnl_info[common_transport_channel_id].dl_chan_num_tbs[num_items-1] = NrOfTransportBlocks;
                    break;
                default:
                    break;
            }
        }
    }


  return offset;
}



static int
dissect_nbap_TransportFormatSet_TransportBlockSize(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
guint32 TransportBlockSize;
guint num_items;
guint32 dch_id;
guint32 common_physical_channel_id;
guint32 common_transport_channel_id;
nbap_dch_channel_info_t* nbap_dch_chnl_info;
nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
nbap_dch_chnl_info = nbap_private_data->nbap_dch_chnl_info;

  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 5000U, &TransportBlockSize, FALSE);


    if(nbap_private_data->num_items>0){
        num_items = nbap_private_data->num_items;
        dch_id = nbap_private_data->dch_id;
        if (num_items > 0 && num_items < MAX_FP_CHANS + 1 && dch_id != 0xffffffff) {
            common_physical_channel_id = nbap_private_data->common_physical_channel_id;
            common_transport_channel_id = nbap_private_data->common_transport_channel_id;
            switch(nbap_private_data->transport_format_set_type){
                case NBAP_DCH_UL:
                    nbap_dch_chnl_info[dch_id].ul_chan_tf_size[num_items-1] = TransportBlockSize;
                    break;
                case NBAP_DCH_DL:
                    nbap_dch_chnl_info[dch_id].dl_chan_tf_size[num_items-1] = TransportBlockSize;
                    break;
                case NBAP_CPCH:
                    nbap_dch_chnl_info[common_physical_channel_id].ul_chan_tf_size[num_items-1] = TransportBlockSize;
                    nbap_dch_chnl_info[common_physical_channel_id].dl_chan_tf_size[num_items-1] = TransportBlockSize;
                    break;
                case NBAP_PCH:
                    nbap_dch_chnl_info[common_transport_channel_id].ul_chan_tf_size[num_items-1] = TransportBlockSize;
                    nbap_dch_chnl_info[common_transport_channel_id].dl_chan_tf_size[num_items-1] = TransportBlockSize;
                    break;
                default:
                    break;
            }
        }
    }


  return offset;
}


static const value_string nbap_TransportFormatSet_TransmissionTimeIntervalDynamic_vals[] = {
  {   0, "msec-10" },
  {   1, "msec-20" },
  {   2, "msec-40" },
  {   3, "msec-80" },
  { 0, NULL }
};


static int
dissect_nbap_TransportFormatSet_TransmissionTimeIntervalDynamic(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t TransmissionTimeIntervalInformation_item_sequence[] = {
  { &hf_nbap_transmissionTimeInterval, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportFormatSet_TransmissionTimeIntervalDynamic },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TransmissionTimeIntervalInformation_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TransmissionTimeIntervalInformation_item, TransmissionTimeIntervalInformation_item_sequence);

  return offset;
}


static const per_sequence_t TransmissionTimeIntervalInformation_sequence_of[1] = {
  { &hf_nbap_TransmissionTimeIntervalInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TransmissionTimeIntervalInformation_item },
};

static int
dissect_nbap_TransmissionTimeIntervalInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_TransmissionTimeIntervalInformation, TransmissionTimeIntervalInformation_sequence_of,
                                                  1, maxTTI_count, FALSE);

  return offset;
}


static const per_sequence_t TDD_TransportFormatSet_ModeDP_sequence[] = {
  { &hf_nbap_transmissionTimeIntervalInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransmissionTimeIntervalInformation },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TDD_TransportFormatSet_ModeDP(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TDD_TransportFormatSet_ModeDP, TDD_TransportFormatSet_ModeDP_sequence);

  return offset;
}


static const value_string nbap_TransportFormatSet_ModeDP_vals[] = {
  {   0, "tdd" },
  {   1, "notApplicable" },
  { 0, NULL }
};

static const per_choice_t TransportFormatSet_ModeDP_choice[] = {
  {   0, &hf_nbap_tdd_01         , ASN1_EXTENSION_ROOT    , dissect_nbap_TDD_TransportFormatSet_ModeDP },
  {   1, &hf_nbap_notApplicable  , ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_TransportFormatSet_ModeDP(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_TransportFormatSet_ModeDP, TransportFormatSet_ModeDP_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t TransportFormatSet_DynamicPartList_item_sequence[] = {
  { &hf_nbap_nrOfTransportBlocks, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportFormatSet_NrOfTransportBlocks },
  { &hf_nbap_transportBlockSize, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportFormatSet_TransportBlockSize },
  { &hf_nbap_mode           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportFormatSet_ModeDP },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TransportFormatSet_DynamicPartList_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    nbap_get_private_data(actx->pinfo)->num_items++;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TransportFormatSet_DynamicPartList_item, TransportFormatSet_DynamicPartList_item_sequence);

  return offset;
}


static const per_sequence_t TransportFormatSet_DynamicPartList_sequence_of[1] = {
  { &hf_nbap_TransportFormatSet_DynamicPartList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TransportFormatSet_DynamicPartList_item },
};

static int
dissect_nbap_TransportFormatSet_DynamicPartList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    nbap_get_private_data(actx->pinfo)->num_items = 0;

  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_TransportFormatSet_DynamicPartList, TransportFormatSet_DynamicPartList_sequence_of,
                                                  1, maxNrOfTFs, FALSE);

  return offset;
}


static const value_string nbap_TransportFormatSet_TransmissionTimeIntervalSemiStatic_vals[] = {
  {   0, "msec-10" },
  {   1, "msec-20" },
  {   2, "msec-40" },
  {   3, "msec-80" },
  {   4, "dynamic" },
  {   5, "msec-5" },
  { 0, NULL }
};


static int
dissect_nbap_TransportFormatSet_TransmissionTimeIntervalSemiStatic(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     5, NULL, TRUE, 1, NULL);

  return offset;
}


static const value_string nbap_TransportFormatSet_ChannelCodingType_vals[] = {
  {   0, "no-codingTDD" },
  {   1, "convolutional-coding" },
  {   2, "turbo-coding" },
  { 0, NULL }
};


static int
dissect_nbap_TransportFormatSet_ChannelCodingType(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_TransportFormatSet_CodingRate_vals[] = {
  {   0, "half" },
  {   1, "third" },
  { 0, NULL }
};


static int
dissect_nbap_TransportFormatSet_CodingRate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_TransportFormatSet_RateMatchingAttribute(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxRateMatching, NULL, FALSE);

  return offset;
}


static const value_string nbap_TransportFormatSet_CRC_Size_vals[] = {
  {   0, "v0" },
  {   1, "v8" },
  {   2, "v12" },
  {   3, "v16" },
  {   4, "v24" },
  { 0, NULL }
};


static int
dissect_nbap_TransportFormatSet_CRC_Size(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     5, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_TransportFormatSet_SecondInterleavingMode_vals[] = {
  {   0, "frame-related" },
  {   1, "timeSlot-related" },
  { 0, NULL }
};


static int
dissect_nbap_TransportFormatSet_SecondInterleavingMode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_TransportFormatSet_ModeSSP_vals[] = {
  {   0, "tdd" },
  {   1, "notApplicable" },
  { 0, NULL }
};

static const per_choice_t TransportFormatSet_ModeSSP_choice[] = {
  {   0, &hf_nbap_tdd_02         , ASN1_EXTENSION_ROOT    , dissect_nbap_TransportFormatSet_SecondInterleavingMode },
  {   1, &hf_nbap_notApplicable  , ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_TransportFormatSet_ModeSSP(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_TransportFormatSet_ModeSSP, TransportFormatSet_ModeSSP_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t TransportFormatSet_Semi_staticPart_sequence[] = {
  { &hf_nbap_transmissionTimeInterval_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportFormatSet_TransmissionTimeIntervalSemiStatic },
  { &hf_nbap_channelCoding  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportFormatSet_ChannelCodingType },
  { &hf_nbap_codingRate     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportFormatSet_CodingRate },
  { &hf_nbap_rateMatchingAttribute, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportFormatSet_RateMatchingAttribute },
  { &hf_nbap_cRC_Size       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportFormatSet_CRC_Size },
  { &hf_nbap_mode_01        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportFormatSet_ModeSSP },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TransportFormatSet_Semi_staticPart(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TransportFormatSet_Semi_staticPart, TransportFormatSet_Semi_staticPart_sequence);

  return offset;
}


static const per_sequence_t TransportFormatSet_sequence[] = {
  { &hf_nbap_dynamicParts   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportFormatSet_DynamicPartList },
  { &hf_nbap_semi_staticPart, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportFormatSet_Semi_staticPart },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TransportFormatSet(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TransportFormatSet, TransportFormatSet_sequence);

  return offset;
}



static int
dissect_nbap_T_ul_TransportFormatSet(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
    guint32 dch_id = nbap_private_data->dch_id;
    nbap_dch_channel_info_t* nbap_dch_chnl_info;
    nbap_dch_chnl_info = nbap_private_data->nbap_dch_chnl_info;
    nbap_private_data->transport_format_set_type = NBAP_DCH_UL;
    if (dch_id != 0xffffffff) {
        nbap_dch_chnl_info[dch_id].num_ul_chans = 0;
    }

  offset = dissect_nbap_TransportFormatSet(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_T_dl_TransportFormatSet(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
    guint32 dch_id = nbap_private_data->dch_id;
    nbap_dch_channel_info_t* nbap_dch_chnl_info;
    nbap_dch_chnl_info = nbap_private_data->nbap_dch_chnl_info;
    nbap_private_data->transport_format_set_type = NBAP_DCH_DL;
    if (dch_id != 0xffffffff) {
        nbap_dch_chnl_info[dch_id].num_dl_chans = 0;
    }

  offset = dissect_nbap_TransportFormatSet(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_FrameHandlingPriority(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}


static const value_string nbap_QE_Selector_vals[] = {
  {   0, "selected" },
  {   1, "non-selected" },
  { 0, NULL }
};


static int
dissect_nbap_QE_Selector(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t DCH_Specific_FDD_Item_sequence[] = {
  { &hf_nbap_dCH_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_dCH_ID },
  { &hf_nbap_ul_TransportFormatSet, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_ul_TransportFormatSet },
  { &hf_nbap_dl_TransportFormatSet, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_dl_TransportFormatSet },
  { &hf_nbap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AllocationRetentionPriority },
  { &hf_nbap_frameHandlingPriority, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FrameHandlingPriority },
  { &hf_nbap_qE_Selector    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_QE_Selector },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DCH_Specific_FDD_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    nbap_get_private_data(actx->pinfo)->num_dch_in_flow++;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DCH_Specific_FDD_Item, DCH_Specific_FDD_Item_sequence);

    nbap_get_private_data(actx->pinfo)->prev_dch_id = nbap_get_private_data(actx->pinfo)->dch_id;

  return offset;
}


static const per_sequence_t DCH_Specific_FDD_InformationList_sequence_of[1] = {
  { &hf_nbap_DCH_Specific_FDD_InformationList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_Specific_FDD_Item },
};

static int
dissect_nbap_DCH_Specific_FDD_InformationList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    nbap_get_private_data(actx->pinfo)->num_dch_in_flow = 0;
    nbap_get_private_data(actx->pinfo)->prev_dch_id = 0;

  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DCH_Specific_FDD_InformationList, DCH_Specific_FDD_InformationList_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t DCH_FDD_InformationItem_sequence[] = {
  { &hf_nbap_payloadCRC_PresenceIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PayloadCRC_PresenceIndicator },
  { &hf_nbap_ul_FP_Mode     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_FP_Mode },
  { &hf_nbap_toAWS          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ToAWS },
  { &hf_nbap_toAWE          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ToAWE },
  { &hf_nbap_dCH_SpecificInformationList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_Specific_FDD_InformationList },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DCH_FDD_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DCH_FDD_InformationItem, DCH_FDD_InformationItem_sequence);

  return offset;
}


static const per_sequence_t DCH_FDD_Information_sequence_of[1] = {
  { &hf_nbap_DCH_FDD_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_FDD_InformationItem },
};

static int
dissect_nbap_DCH_FDD_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DCH_FDD_Information, DCH_FDD_Information_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}


static const value_string nbap_DCH_Indicator_For_E_DCH_HSDPA_Operation_vals[] = {
  {   0, "dch-not-present" },
  { 0, NULL }
};


static int
dissect_nbap_DCH_Indicator_For_E_DCH_HSDPA_Operation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t DCH_InformationResponseItem_sequence[] = {
  { &hf_nbap_dCH_ID_01      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_ID },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DCH_InformationResponseItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DCH_InformationResponseItem, DCH_InformationResponseItem_sequence);

  return offset;
}


static const per_sequence_t DCH_InformationResponse_sequence_of[1] = {
  { &hf_nbap_DCH_InformationResponse_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_InformationResponseItem },
};

static int
dissect_nbap_DCH_InformationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DCH_InformationResponse, DCH_InformationResponse_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}



static int
dissect_nbap_Pattern_Sequence_Identifier(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxNrOfDCHMeasurementOccasionPatternSequence, NULL, FALSE);

  return offset;
}


static const value_string nbap_Status_Flag_vals[] = {
  {   0, "activate" },
  {   1, "deactivate" },
  { 0, NULL }
};


static int
dissect_nbap_Status_Flag(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_INTEGER_1_9(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 9U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_INTEGER_0_511(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 511U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_INTEGER_1_512(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 512U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_7(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     7, 7, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t Measurement_Occasion_Pattern_Sequence_parameters_sequence[] = {
  { &hf_nbap_measurement_Occasion_Pattern_Sequence_parameters_k, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_1_9 },
  { &hf_nbap_measurement_Occasion_Pattern_Sequence_parameters_offset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_511 },
  { &hf_nbap_measurement_Occasion_Pattern_Sequence_parameters_M_Length, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_1_512 },
  { &hf_nbap_measurement_Occasion_Pattern_Sequence_parameters_Timeslot_Bitmap, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_7 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Measurement_Occasion_Pattern_Sequence_parameters(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Measurement_Occasion_Pattern_Sequence_parameters, Measurement_Occasion_Pattern_Sequence_parameters_sequence);

  return offset;
}


static const per_sequence_t DchMeasurementOccasionInformation_Item_sequence[] = {
  { &hf_nbap_pattern_Sequence_Identifier, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Pattern_Sequence_Identifier },
  { &hf_nbap_status_Flag    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Status_Flag },
  { &hf_nbap_measurement_Occasion_Pattern_Sequence_parameters, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Measurement_Occasion_Pattern_Sequence_parameters },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DchMeasurementOccasionInformation_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DchMeasurementOccasionInformation_Item, DchMeasurementOccasionInformation_Item_sequence);

  return offset;
}


static const per_sequence_t DCH_MeasurementOccasion_Information_sequence_of[1] = {
  { &hf_nbap_DCH_MeasurementOccasion_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DchMeasurementOccasionInformation_Item },
};

static int
dissect_nbap_DCH_MeasurementOccasion_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DCH_MeasurementOccasion_Information, DCH_MeasurementOccasion_Information_sequence_of,
                                                  1, maxNrOfDCHMeasurementOccasionPatternSequence, FALSE);

  return offset;
}


static const per_sequence_t DCH_Specific_TDD_Item_sequence[] = {
  { &hf_nbap_dCH_ID_01      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_ID },
  { &hf_nbap_ul_CCTrCH_ID   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_dl_CCTrCH_ID   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_ul_TransportFormatSet_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportFormatSet },
  { &hf_nbap_dl_TransportFormatSet_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportFormatSet },
  { &hf_nbap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AllocationRetentionPriority },
  { &hf_nbap_frameHandlingPriority, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FrameHandlingPriority },
  { &hf_nbap_qE_Selector    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_QE_Selector },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DCH_Specific_TDD_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DCH_Specific_TDD_Item, DCH_Specific_TDD_Item_sequence);

  return offset;
}


static const per_sequence_t DCH_Specific_TDD_InformationList_sequence_of[1] = {
  { &hf_nbap_DCH_Specific_TDD_InformationList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_Specific_TDD_Item },
};

static int
dissect_nbap_DCH_Specific_TDD_InformationList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DCH_Specific_TDD_InformationList, DCH_Specific_TDD_InformationList_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t DCH_TDD_InformationItem_sequence[] = {
  { &hf_nbap_payloadCRC_PresenceIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PayloadCRC_PresenceIndicator },
  { &hf_nbap_ul_FP_Mode     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_FP_Mode },
  { &hf_nbap_toAWS          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ToAWS },
  { &hf_nbap_toAWE          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ToAWE },
  { &hf_nbap_dCH_SpecificInformationList_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_Specific_TDD_InformationList },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DCH_TDD_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DCH_TDD_InformationItem, DCH_TDD_InformationItem_sequence);

  return offset;
}


static const per_sequence_t DCH_TDD_Information_sequence_of[1] = {
  { &hf_nbap_DCH_TDD_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_TDD_InformationItem },
};

static int
dissect_nbap_DCH_TDD_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DCH_TDD_Information, DCH_TDD_Information_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}



static int
dissect_nbap_T_dCH_ID_01(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_DCH_ID(tvb, offset, actx, tree, hf_index);

    nbap_get_private_data(actx->pinfo)->dch_id = nbap_get_private_data(actx->pinfo)->t_dch_id;

  return offset;
}



static int
dissect_nbap_T_ul_TransportFormatSet_01(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
    guint32 dch_id = nbap_private_data->dch_id;
    nbap_dch_channel_info_t* nbap_dch_chnl_info;
    nbap_dch_chnl_info = nbap_private_data->nbap_dch_chnl_info;
    nbap_private_data->transport_format_set_type = NBAP_DCH_UL;
    if (dch_id != 0xffffffff) {
        nbap_dch_chnl_info[dch_id].num_ul_chans = 0;
    }

  offset = dissect_nbap_TransportFormatSet(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_T_dl_TransportFormatSet_01(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
    guint32 dch_id = nbap_private_data->dch_id;
    nbap_dch_channel_info_t* nbap_dch_chnl_info;
    nbap_dch_chnl_info = nbap_private_data->nbap_dch_chnl_info;
    nbap_private_data->transport_format_set_type = NBAP_DCH_DL;
    if (dch_id != 0xffffffff) {
        nbap_dch_chnl_info[dch_id].num_dl_chans = 0;
    }

  offset = dissect_nbap_TransportFormatSet(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t DCH_ModifySpecificItem_FDD_sequence[] = {
  { &hf_nbap_dCH_ID_02      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_dCH_ID_01 },
  { &hf_nbap_ul_TransportFormatSet_02, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_T_ul_TransportFormatSet_01 },
  { &hf_nbap_dl_TransportFormatSet_02, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_T_dl_TransportFormatSet_01 },
  { &hf_nbap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_AllocationRetentionPriority },
  { &hf_nbap_frameHandlingPriority, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_FrameHandlingPriority },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DCH_ModifySpecificItem_FDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DCH_ModifySpecificItem_FDD, DCH_ModifySpecificItem_FDD_sequence);

  return offset;
}


static const per_sequence_t DCH_ModifySpecificInformation_FDD_sequence_of[1] = {
  { &hf_nbap_DCH_ModifySpecificInformation_FDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_ModifySpecificItem_FDD },
};

static int
dissect_nbap_DCH_ModifySpecificInformation_FDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DCH_ModifySpecificInformation_FDD, DCH_ModifySpecificInformation_FDD_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t FDD_DCHs_to_ModifyItem_sequence[] = {
  { &hf_nbap_ul_FP_Mode     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_FP_Mode },
  { &hf_nbap_toAWS          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ToAWS },
  { &hf_nbap_toAWE          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ToAWE },
  { &hf_nbap_transportBearerRequestIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportBearerRequestIndicator },
  { &hf_nbap_dCH_SpecificInformationList_02, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_ModifySpecificInformation_FDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_FDD_DCHs_to_ModifyItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_FDD_DCHs_to_ModifyItem, FDD_DCHs_to_ModifyItem_sequence);

  return offset;
}


static const per_sequence_t FDD_DCHs_to_Modify_sequence_of[1] = {
  { &hf_nbap_FDD_DCHs_to_Modify_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_DCHs_to_ModifyItem },
};

static int
dissect_nbap_FDD_DCHs_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_FDD_DCHs_to_Modify, FDD_DCHs_to_Modify_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t DCH_ModifySpecificItem_TDD_sequence[] = {
  { &hf_nbap_dCH_ID_01      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_ID },
  { &hf_nbap_ul_CCTrCH_ID   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CCTrCH_ID },
  { &hf_nbap_dl_CCTrCH_ID   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CCTrCH_ID },
  { &hf_nbap_ul_TransportFormatSet_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportFormatSet },
  { &hf_nbap_dl_TransportFormatSet_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportFormatSet },
  { &hf_nbap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_AllocationRetentionPriority },
  { &hf_nbap_frameHandlingPriority, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_FrameHandlingPriority },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DCH_ModifySpecificItem_TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DCH_ModifySpecificItem_TDD, DCH_ModifySpecificItem_TDD_sequence);

  return offset;
}


static const per_sequence_t DCH_ModifySpecificInformation_TDD_sequence_of[1] = {
  { &hf_nbap_DCH_ModifySpecificInformation_TDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_ModifySpecificItem_TDD },
};

static int
dissect_nbap_DCH_ModifySpecificInformation_TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DCH_ModifySpecificInformation_TDD, DCH_ModifySpecificInformation_TDD_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t DCH_ModifyItem_TDD_sequence[] = {
  { &hf_nbap_ul_FP_Mode     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_FP_Mode },
  { &hf_nbap_toAWS          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ToAWS },
  { &hf_nbap_toAWE          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ToAWE },
  { &hf_nbap_transportBearerRequestIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportBearerRequestIndicator },
  { &hf_nbap_dCH_SpecificInformationList_03, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_ModifySpecificInformation_TDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DCH_ModifyItem_TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DCH_ModifyItem_TDD, DCH_ModifyItem_TDD_sequence);

  return offset;
}


static const per_sequence_t TDD_DCHs_to_Modify_sequence_of[1] = {
  { &hf_nbap_TDD_DCHs_to_Modify_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_ModifyItem_TDD },
};

static int
dissect_nbap_TDD_DCHs_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_TDD_DCHs_to_Modify, TDD_DCHs_to_Modify_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t DedicatedChannelsCapacityConsumptionLaw_item_sequence[] = {
  { &hf_nbap_dl_Cost_1      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_65535 },
  { &hf_nbap_dl_Cost_2      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_65535 },
  { &hf_nbap_ul_Cost_1      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_65535 },
  { &hf_nbap_ul_Cost_2      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_65535 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DedicatedChannelsCapacityConsumptionLaw_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DedicatedChannelsCapacityConsumptionLaw_item, DedicatedChannelsCapacityConsumptionLaw_item_sequence);

  return offset;
}


static const per_sequence_t DedicatedChannelsCapacityConsumptionLaw_sequence_of[1] = {
  { &hf_nbap_DedicatedChannelsCapacityConsumptionLaw_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DedicatedChannelsCapacityConsumptionLaw_item },
};

static int
dissect_nbap_DedicatedChannelsCapacityConsumptionLaw(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DedicatedChannelsCapacityConsumptionLaw, DedicatedChannelsCapacityConsumptionLaw_sequence_of,
                                                  1, maxNrOfSF, FALSE);

  return offset;
}


static const value_string nbap_DedicatedMeasurementType_vals[] = {
  {   0, "sir" },
  {   1, "sir-error" },
  {   2, "transmitted-code-power" },
  {   3, "rscp" },
  {   4, "rx-timing-deviation" },
  {   5, "round-trip-time" },
  {   6, "rx-timing-deviation-LCR" },
  {   7, "angle-Of-Arrival-LCR" },
  {   8, "hs-sich-quality" },
  {   9, "best-Cell-Portions" },
  {  10, "rx-timing-deviation-768" },
  {  11, "rx-timing-deviation-384-extended" },
  {  12, "best-Cell-PortionsLCR" },
  { 0, NULL }
};

static value_string_ext nbap_DedicatedMeasurementType_vals_ext = VALUE_STRING_EXT_INIT(nbap_DedicatedMeasurementType_vals);


static int
dissect_nbap_DedicatedMeasurementType(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     6, NULL, TRUE, 7, NULL);

  return offset;
}



static int
dissect_nbap_SIR_Error_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 125U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_Transmitted_Code_Power_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_Rx_Timing_Deviation_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 8191U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_Round_Trip_Time_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 32767U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_Extension_DedicatedMeasurementValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string nbap_DedicatedMeasurementValue_vals[] = {
  {   0, "sIR-Value" },
  {   1, "sIR-ErrorValue" },
  {   2, "transmittedCodePowerValue" },
  {   3, "rSCP" },
  {   4, "rxTimingDeviationValue" },
  {   5, "roundTripTime" },
  {   6, "extension-DedicatedMeasurementValue" },
  { 0, NULL }
};

static const per_choice_t DedicatedMeasurementValue_choice[] = {
  {   0, &hf_nbap_sIR_Value      , ASN1_EXTENSION_ROOT    , dissect_nbap_SIR_Value },
  {   1, &hf_nbap_sIR_ErrorValue , ASN1_EXTENSION_ROOT    , dissect_nbap_SIR_Error_Value },
  {   2, &hf_nbap_transmittedCodePowerValue, ASN1_EXTENSION_ROOT    , dissect_nbap_Transmitted_Code_Power_Value },
  {   3, &hf_nbap_rSCP           , ASN1_EXTENSION_ROOT    , dissect_nbap_RSCP_Value },
  {   4, &hf_nbap_rxTimingDeviationValue, ASN1_EXTENSION_ROOT    , dissect_nbap_Rx_Timing_Deviation_Value },
  {   5, &hf_nbap_roundTripTime  , ASN1_EXTENSION_ROOT    , dissect_nbap_Round_Trip_Time_Value },
  {   6, &hf_nbap_extension_DedicatedMeasurementValue, ASN1_NOT_EXTENSION_ROOT, dissect_nbap_Extension_DedicatedMeasurementValue },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_DedicatedMeasurementValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_DedicatedMeasurementValue, DedicatedMeasurementValue_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t DedicatedMeasurementAvailable_sequence[] = {
  { &hf_nbap_dedicatedmeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DedicatedMeasurementValue },
  { &hf_nbap_cFN            , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CFN },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DedicatedMeasurementAvailable(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DedicatedMeasurementAvailable, DedicatedMeasurementAvailable_sequence);

  return offset;
}



static int
dissect_nbap_DedicatedMeasurementnotAvailable(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string nbap_DedicatedMeasurementValueInformation_vals[] = {
  {   0, "measurementAvailable" },
  {   1, "measurementnotAvailable" },
  { 0, NULL }
};

static const per_choice_t DedicatedMeasurementValueInformation_choice[] = {
  {   0, &hf_nbap_measurementAvailable_01, ASN1_NO_EXTENSIONS     , dissect_nbap_DedicatedMeasurementAvailable },
  {   1, &hf_nbap_measurementnotAvailable_01, ASN1_NO_EXTENSIONS     , dissect_nbap_DedicatedMeasurementnotAvailable },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_DedicatedMeasurementValueInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_DedicatedMeasurementValueInformation, DedicatedMeasurementValueInformation_choice,
                                 NULL);

  return offset;
}


static const value_string nbap_DelayedActivation_vals[] = {
  {   0, "cfn" },
  {   1, "separate-indication" },
  { 0, NULL }
};

static const per_choice_t DelayedActivation_choice[] = {
  {   0, &hf_nbap_cfn            , ASN1_NO_EXTENSIONS     , dissect_nbap_CFN },
  {   1, &hf_nbap_separate_indication, ASN1_NO_EXTENSIONS     , dissect_nbap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_DelayedActivation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_DelayedActivation, DelayedActivation_choice,
                                 NULL);

  return offset;
}


static const value_string nbap_Execution_Type_vals[] = {
  {   0, "synchronised" },
  {   1, "unsynchronised" },
  { 0, NULL }
};

static const per_choice_t Execution_Type_choice[] = {
  {   0, &hf_nbap_synchronised   , ASN1_NO_EXTENSIONS     , dissect_nbap_CFN },
  {   1, &hf_nbap_unsynchronised , ASN1_NO_EXTENSIONS     , dissect_nbap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_Execution_Type(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_Execution_Type, Execution_Type_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Activate_Info_sequence[] = {
  { &hf_nbap_activation_type, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Execution_Type },
  { &hf_nbap_initial_dl_tx_power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_firstRLS_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_FirstRLS_Indicator },
  { &hf_nbap_propagation_delay, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PropagationDelay },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Activate_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Activate_Info, Activate_Info_sequence);

  return offset;
}


static const per_sequence_t Deactivate_Info_sequence[] = {
  { &hf_nbap_deactivation_type, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Execution_Type },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Deactivate_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Deactivate_Info, Deactivate_Info_sequence);

  return offset;
}


static const value_string nbap_DelayedActivationUpdate_vals[] = {
  {   0, "activate" },
  {   1, "deactivate" },
  { 0, NULL }
};

static const per_choice_t DelayedActivationUpdate_choice[] = {
  {   0, &hf_nbap_activate       , ASN1_NO_EXTENSIONS     , dissect_nbap_Activate_Info },
  {   1, &hf_nbap_deactivate_01  , ASN1_NO_EXTENSIONS     , dissect_nbap_Deactivate_Info },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_DelayedActivationUpdate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_DelayedActivationUpdate, DelayedActivationUpdate_choice,
                                 NULL);

  return offset;
}



static int
dissect_nbap_DeltaSIR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 30U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_INTEGER_0_119(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 119U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_GANSS_Signal_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 7U, NULL, TRUE);

  return offset;
}


static const value_string nbap_GANSS_StatusHealth_vals[] = {
  {   0, "udre-scale-1dot0" },
  {   1, "udre-scale-0dot75" },
  {   2, "udre-scale-0dot5" },
  {   3, "udre-scale-0dot3" },
  {   4, "udre-scale-0dot2" },
  {   5, "udre-scale-0dot1" },
  {   6, "no-data" },
  {   7, "invalid-data" },
  { 0, NULL }
};


static int
dissect_nbap_GANSS_StatusHealth(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_INTEGER_0_63(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 63U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_10(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     10, 10, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const value_string nbap_UDRE_vals[] = {
  {   0, "udre-minusequal-one-m" },
  {   1, "udre-betweenoneandfour-m" },
  {   2, "udre-betweenfourandeight-m" },
  {   3, "udre-greaterequaleight-m" },
  { 0, NULL }
};


static int
dissect_nbap_UDRE(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_INTEGER_M2047_2047(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -2047, 2047U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_INTEGER_M127_127(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -127, 127U, NULL, FALSE);

  return offset;
}


static const per_sequence_t DGANSS_SignalInformationItem_sequence[] = {
  { &hf_nbap_satId          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_63 },
  { &hf_nbap_gANSS_iod      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_10 },
  { &hf_nbap_udre           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UDRE },
  { &hf_nbap_ganss_prc      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_M2047_2047 },
  { &hf_nbap_ganss_rrc      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_M127_127 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DGANSS_SignalInformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DGANSS_SignalInformationItem, DGANSS_SignalInformationItem_sequence);

  return offset;
}


static const per_sequence_t DGANSS_SignalInformation_sequence_of[1] = {
  { &hf_nbap_DGANSS_SignalInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DGANSS_SignalInformationItem },
};

static int
dissect_nbap_DGANSS_SignalInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DGANSS_SignalInformation, DGANSS_SignalInformation_sequence_of,
                                                  1, maxGANSSSat, FALSE);

  return offset;
}


static const per_sequence_t DGANSS_InformationItem_sequence[] = {
  { &hf_nbap_gANSS_SignalId , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GANSS_Signal_ID },
  { &hf_nbap_gANSS_StatusHealth, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_StatusHealth },
  { &hf_nbap_dGANSS_SignalInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DGANSS_SignalInformation },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DGANSS_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DGANSS_InformationItem, DGANSS_InformationItem_sequence);

  return offset;
}


static const per_sequence_t DGANSS_Information_sequence_of[1] = {
  { &hf_nbap_DGANSS_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DGANSS_InformationItem },
};

static int
dissect_nbap_DGANSS_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DGANSS_Information, DGANSS_Information_sequence_of,
                                                  1, maxSgnType, FALSE);

  return offset;
}


static const per_sequence_t DGANSSCorrections_sequence[] = {
  { &hf_nbap_dGANSS_ReferenceTime, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_119 },
  { &hf_nbap_dGANSS_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DGANSS_Information },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DGANSSCorrections(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DGANSSCorrections, DGANSSCorrections_sequence);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_8(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     8, 8, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t DGANSS_Corrections_Req_sequence[] = {
  { &hf_nbap_dGANSS_Signal_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DGANSS_Corrections_Req(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DGANSS_Corrections_Req, DGANSS_Corrections_Req_sequence);

  return offset;
}


static const value_string nbap_PRCDeviation_vals[] = {
  {   0, "one" },
  {   1, "two" },
  {   2, "five" },
  {   3, "ten" },
  { 0, NULL }
};


static int
dissect_nbap_PRCDeviation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t DGANSSThreshold_sequence[] = {
  { &hf_nbap_pRCDeviation   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PRCDeviation },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DGANSSThreshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DGANSSThreshold, DGANSSThreshold_sequence);

  return offset;
}


static const value_string nbap_UDREGrowthRate_vals[] = {
  {   0, "growth-1-point-5" },
  {   1, "growth-2" },
  {   2, "growth-4" },
  {   3, "growth-6" },
  {   4, "growth-8" },
  {   5, "growth-10" },
  {   6, "growth-12" },
  {   7, "growth-16" },
  { 0, NULL }
};


static int
dissect_nbap_UDREGrowthRate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_UDREValidityTime_vals[] = {
  {   0, "val-20sec" },
  {   1, "val-40sec" },
  {   2, "val-80sec" },
  {   3, "val-160sec" },
  {   4, "val-320sec" },
  {   5, "val-640sec" },
  {   6, "val-1280sec" },
  {   7, "val-2560sec" },
  { 0, NULL }
};


static int
dissect_nbap_UDREValidityTime(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t DGNSS_ValidityPeriod_sequence[] = {
  { &hf_nbap_udreGrowthRate , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UDREGrowthRate },
  { &hf_nbap_udreValidityTime, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UDREValidityTime },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DGNSS_ValidityPeriod(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DGNSS_ValidityPeriod, DGNSS_ValidityPeriod_sequence);

  return offset;
}



static int
dissect_nbap_GPSTOW(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 604799U, NULL, FALSE);

  return offset;
}


static const value_string nbap_GPS_Status_Health_vals[] = {
  {   0, "udre-scale-1dot0" },
  {   1, "udre-scale-0dot75" },
  {   2, "udre-scale-0dot5" },
  {   3, "udre-scale-0dot3" },
  {   4, "udre-scale-0dot1" },
  {   5, "no-data" },
  {   6, "invalid-data" },
  { 0, NULL }
};


static int
dissect_nbap_GPS_Status_Health(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     7, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_SAT_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 63U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_PRC(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -2047, 2047U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_Range_Correction_Rate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -127, 127U, NULL, FALSE);

  return offset;
}


static const per_sequence_t SAT_Info_DGPSCorrections_Item_sequence[] = {
  { &hf_nbap_sat_id         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SAT_ID },
  { &hf_nbap_iode_dgps      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_udre           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UDRE },
  { &hf_nbap_prc            , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PRC },
  { &hf_nbap_range_correction_rate, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Range_Correction_Rate },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SAT_Info_DGPSCorrections_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SAT_Info_DGPSCorrections_Item, SAT_Info_DGPSCorrections_Item_sequence);

  return offset;
}


static const per_sequence_t SAT_Info_DGPSCorrections_sequence_of[1] = {
  { &hf_nbap_SAT_Info_DGPSCorrections_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_SAT_Info_DGPSCorrections_Item },
};

static int
dissect_nbap_SAT_Info_DGPSCorrections(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_SAT_Info_DGPSCorrections, SAT_Info_DGPSCorrections_sequence_of,
                                                  1, maxNoSat, FALSE);

  return offset;
}


static const per_sequence_t DGPSCorrections_sequence[] = {
  { &hf_nbap_gpstow         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GPSTOW },
  { &hf_nbap_status_health  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GPS_Status_Health },
  { &hf_nbap_satelliteinfo  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SAT_Info_DGPSCorrections },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DGPSCorrections(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DGPSCorrections, DGPSCorrections_sequence);

  return offset;
}


static const per_sequence_t DGPSThresholds_sequence[] = {
  { &hf_nbap_prcdeviation   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PRCDeviation },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DGPSThresholds(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DGPSThresholds, DGPSThresholds_sequence);

  return offset;
}


static const value_string nbap_DiversityControlField_vals[] = {
  {   0, "may" },
  {   1, "must" },
  {   2, "must-not" },
  { 0, NULL }
};


static int
dissect_nbap_DiversityControlField(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_DiversityMode_vals[] = {
  {   0, "none" },
  {   1, "sTTD" },
  {   2, "closed-loop-mode1" },
  {   3, "not-used-closed-loop-mode2" },
  { 0, NULL }
};


static int
dissect_nbap_DiversityMode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_DL_DPCH_SlotFormat(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 16U, NULL, TRUE);

  return offset;
}


static const value_string nbap_DL_DPCH_TimingAdjustment_vals[] = {
  {   0, "timing-advance" },
  {   1, "timing-delay" },
  { 0, NULL }
};


static int
dissect_nbap_DL_DPCH_TimingAdjustment(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_TimeSlot(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 14U, NULL, FALSE);

  return offset;
}


static const value_string nbap_MidambleConfigurationBurstType1And3_vals[] = {
  {   0, "v4" },
  {   1, "v8" },
  {   2, "v16" },
  { 0, NULL }
};


static int
dissect_nbap_MidambleConfigurationBurstType1And3(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_MidambleShiftLong(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}


static const value_string nbap_MidambleAllocationMode1_vals[] = {
  {   0, "defaultMidamble" },
  {   1, "commonMidamble" },
  {   2, "ueSpecificMidamble" },
  { 0, NULL }
};

static const per_choice_t MidambleAllocationMode1_choice[] = {
  {   0, &hf_nbap_defaultMidamble, ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {   1, &hf_nbap_commonMidamble , ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {   2, &hf_nbap_ueSpecificMidamble, ASN1_EXTENSION_ROOT    , dissect_nbap_MidambleShiftLong },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_MidambleAllocationMode1(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_MidambleAllocationMode1, MidambleAllocationMode1_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Type1_sequence[] = {
  { &hf_nbap_midambleConfigurationBurstType1And3, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleConfigurationBurstType1And3 },
  { &hf_nbap_midambleAllocationMode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleAllocationMode1 },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Type1(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Type1, Type1_sequence);

  return offset;
}


static const value_string nbap_MidambleConfigurationBurstType2_vals[] = {
  {   0, "v3" },
  {   1, "v6" },
  { 0, NULL }
};


static int
dissect_nbap_MidambleConfigurationBurstType2(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_MidambleShiftShort(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 5U, NULL, FALSE);

  return offset;
}


static const value_string nbap_MidambleAllocationMode2_vals[] = {
  {   0, "defaultMidamble" },
  {   1, "commonMidamble" },
  {   2, "ueSpecificMidamble" },
  { 0, NULL }
};

static const per_choice_t MidambleAllocationMode2_choice[] = {
  {   0, &hf_nbap_defaultMidamble, ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {   1, &hf_nbap_commonMidamble , ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {   2, &hf_nbap_ueSpecificMidamble_01, ASN1_EXTENSION_ROOT    , dissect_nbap_MidambleShiftShort },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_MidambleAllocationMode2(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_MidambleAllocationMode2, MidambleAllocationMode2_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Type2_sequence[] = {
  { &hf_nbap_midambleConfigurationBurstType2, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleConfigurationBurstType2 },
  { &hf_nbap_midambleAllocationMode_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleAllocationMode2 },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Type2(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Type2, Type2_sequence);

  return offset;
}


static const value_string nbap_MidambleAllocationMode3_vals[] = {
  {   0, "defaultMidamble" },
  {   1, "ueSpecificMidamble" },
  { 0, NULL }
};

static const per_choice_t MidambleAllocationMode3_choice[] = {
  {   0, &hf_nbap_defaultMidamble, ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {   1, &hf_nbap_ueSpecificMidamble, ASN1_EXTENSION_ROOT    , dissect_nbap_MidambleShiftLong },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_MidambleAllocationMode3(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_MidambleAllocationMode3, MidambleAllocationMode3_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Type3_sequence[] = {
  { &hf_nbap_midambleConfigurationBurstType1And3, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleConfigurationBurstType1And3 },
  { &hf_nbap_midambleAllocationMode_02, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleAllocationMode3 },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Type3(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Type3, Type3_sequence);

  return offset;
}


static const value_string nbap_MidambleShiftAndBurstType_vals[] = {
  {   0, "type1" },
  {   1, "type2" },
  {   2, "type3" },
  { 0, NULL }
};

static const per_choice_t MidambleShiftAndBurstType_choice[] = {
  {   0, &hf_nbap_type1          , ASN1_EXTENSION_ROOT    , dissect_nbap_Type1 },
  {   1, &hf_nbap_type2          , ASN1_EXTENSION_ROOT    , dissect_nbap_Type2 },
  {   2, &hf_nbap_type3          , ASN1_EXTENSION_ROOT    , dissect_nbap_Type3 },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_MidambleShiftAndBurstType(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_MidambleShiftAndBurstType, MidambleShiftAndBurstType_choice,
                                 NULL);

  return offset;
}


static const value_string nbap_TFCI_Presence_vals[] = {
  {   0, "present" },
  {   1, "not-present" },
  { 0, NULL }
};


static int
dissect_nbap_TFCI_Presence(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_DPCH_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 239U, NULL, FALSE);

  return offset;
}


static const value_string nbap_TDD_ChannelisationCode_vals[] = {
  {   0, "chCode1div1" },
  {   1, "chCode2div1" },
  {   2, "chCode2div2" },
  {   3, "chCode4div1" },
  {   4, "chCode4div2" },
  {   5, "chCode4div3" },
  {   6, "chCode4div4" },
  {   7, "chCode8div1" },
  {   8, "chCode8div2" },
  {   9, "chCode8div3" },
  {  10, "chCode8div4" },
  {  11, "chCode8div5" },
  {  12, "chCode8div6" },
  {  13, "chCode8div7" },
  {  14, "chCode8div8" },
  {  15, "chCode16div1" },
  {  16, "chCode16div2" },
  {  17, "chCode16div3" },
  {  18, "chCode16div4" },
  {  19, "chCode16div5" },
  {  20, "chCode16div6" },
  {  21, "chCode16div7" },
  {  22, "chCode16div8" },
  {  23, "chCode16div9" },
  {  24, "chCode16div10" },
  {  25, "chCode16div11" },
  {  26, "chCode16div12" },
  {  27, "chCode16div13" },
  {  28, "chCode16div14" },
  {  29, "chCode16div15" },
  {  30, "chCode16div16" },
  { 0, NULL }
};

static value_string_ext nbap_TDD_ChannelisationCode_vals_ext = VALUE_STRING_EXT_INIT(nbap_TDD_ChannelisationCode_vals);


static int
dissect_nbap_TDD_ChannelisationCode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     31, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t TDD_DL_Code_InformationItem_sequence[] = {
  { &hf_nbap_dPCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DPCH_ID },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TDD_DL_Code_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TDD_DL_Code_InformationItem, TDD_DL_Code_InformationItem_sequence);

  return offset;
}


static const per_sequence_t TDD_DL_Code_Information_sequence_of[1] = {
  { &hf_nbap_TDD_DL_Code_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DL_Code_InformationItem },
};

static int
dissect_nbap_TDD_DL_Code_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_TDD_DL_Code_Information, TDD_DL_Code_Information_sequence_of,
                                                  1, maxNrOfDPCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_Timeslot_InformationItem_sequence[] = {
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCI_Presence },
  { &hf_nbap_dL_Code_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DL_Code_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_Timeslot_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_Timeslot_InformationItem, DL_Timeslot_InformationItem_sequence);

  return offset;
}


static const per_sequence_t DL_Timeslot_Information_sequence_of[1] = {
  { &hf_nbap_DL_Timeslot_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Timeslot_InformationItem },
};

static int
dissect_nbap_DL_Timeslot_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_Timeslot_Information, DL_Timeslot_Information_sequence_of,
                                                  1, maxNrOfDLTSs, FALSE);

  return offset;
}


static const value_string nbap_MidambleAllocationMode_vals[] = {
  {   0, "defaultMidamble" },
  {   1, "commonMidamble" },
  {   2, "uESpecificMidamble" },
  { 0, NULL }
};


static int
dissect_nbap_MidambleAllocationMode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_MidambleConfigurationLCR_vals[] = {
  {   0, "v2" },
  {   1, "v4" },
  {   2, "v6" },
  {   3, "v8" },
  {   4, "v10" },
  {   5, "v12" },
  {   6, "v14" },
  {   7, "v16" },
  { 0, NULL }
};


static int
dissect_nbap_MidambleConfigurationLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t MidambleShiftLCR_sequence[] = {
  { &hf_nbap_midambleAllocationMode_03, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleAllocationMode },
  { &hf_nbap_midambleShift_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftLong },
  { &hf_nbap_midambleConfigurationLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleConfigurationLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MidambleShiftLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MidambleShiftLCR, MidambleShiftLCR_sequence);

  return offset;
}


static const value_string nbap_Modulation_vals[] = {
  {   0, "qPSK" },
  {   1, "eightPSK" },
  { 0, NULL }
};


static int
dissect_nbap_Modulation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t TDD_ChannelisationCodeLCR_sequence[] = {
  { &hf_nbap_tDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_modulation     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Modulation },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TDD_ChannelisationCodeLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TDD_ChannelisationCodeLCR, TDD_ChannelisationCodeLCR_sequence);

  return offset;
}



static int
dissect_nbap_QPSK_DL_DPCH_TimeSlotFormatTDD_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 24U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_EightPSK_DL_DPCH_TimeSlotFormatTDD_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 24U, NULL, TRUE);

  return offset;
}


static const value_string nbap_TDD_DL_DPCH_TimeSlotFormat_LCR_vals[] = {
  {   0, "qPSK" },
  {   1, "eightPSK" },
  { 0, NULL }
};

static const per_choice_t TDD_DL_DPCH_TimeSlotFormat_LCR_choice[] = {
  {   0, &hf_nbap_qPSK_01        , ASN1_EXTENSION_ROOT    , dissect_nbap_QPSK_DL_DPCH_TimeSlotFormatTDD_LCR },
  {   1, &hf_nbap_eightPSK       , ASN1_EXTENSION_ROOT    , dissect_nbap_EightPSK_DL_DPCH_TimeSlotFormatTDD_LCR },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_TDD_DL_DPCH_TimeSlotFormat_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_TDD_DL_DPCH_TimeSlotFormat_LCR, TDD_DL_DPCH_TimeSlotFormat_LCR_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t TDD_DL_Code_LCR_InformationItem_sequence[] = {
  { &hf_nbap_dPCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DPCH_ID },
  { &hf_nbap_tdd_ChannelisationCodeLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCodeLCR },
  { &hf_nbap_tdd_DL_DPCH_TimeSlotFormat_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DL_DPCH_TimeSlotFormat_LCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TDD_DL_Code_LCR_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TDD_DL_Code_LCR_InformationItem, TDD_DL_Code_LCR_InformationItem_sequence);

  return offset;
}


static const per_sequence_t TDD_DL_Code_LCR_Information_sequence_of[1] = {
  { &hf_nbap_TDD_DL_Code_LCR_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DL_Code_LCR_InformationItem },
};

static int
dissect_nbap_TDD_DL_Code_LCR_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_TDD_DL_Code_LCR_Information, TDD_DL_Code_LCR_Information_sequence_of,
                                                  1, maxNrOfDPCHLCRs, FALSE);

  return offset;
}


static const per_sequence_t DL_TimeslotLCR_InformationItem_sequence[] = {
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCI_Presence },
  { &hf_nbap_dL_Code_LCR_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DL_Code_LCR_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_TimeslotLCR_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_TimeslotLCR_InformationItem, DL_TimeslotLCR_InformationItem_sequence);

  return offset;
}


static const per_sequence_t DL_TimeslotLCR_Information_sequence_of[1] = {
  { &hf_nbap_DL_TimeslotLCR_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_TimeslotLCR_InformationItem },
};

static int
dissect_nbap_DL_TimeslotLCR_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_TimeslotLCR_Information, DL_TimeslotLCR_Information_sequence_of,
                                                  1, maxNrOfDLTSLCRs, FALSE);

  return offset;
}


static const value_string nbap_MidambleAllocationMode7681_vals[] = {
  {   0, "defaultMidamble" },
  {   1, "commonMidamble" },
  {   2, "ueSpecificMidamble" },
  { 0, NULL }
};

static const per_choice_t MidambleAllocationMode7681_choice[] = {
  {   0, &hf_nbap_defaultMidamble, ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {   1, &hf_nbap_commonMidamble , ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {   2, &hf_nbap_ueSpecificMidamble, ASN1_EXTENSION_ROOT    , dissect_nbap_MidambleShiftLong },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_MidambleAllocationMode7681(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_MidambleAllocationMode7681, MidambleAllocationMode7681_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Type7681_sequence[] = {
  { &hf_nbap_midambleConfigurationBurstType1And3, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleConfigurationBurstType1And3 },
  { &hf_nbap_midambleAllocationMode_04, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleAllocationMode7681 },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Type7681(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Type7681, Type7681_sequence);

  return offset;
}


static const value_string nbap_MidambleConfigurationBurstType2_768_vals[] = {
  {   0, "v4" },
  {   1, "v8" },
  { 0, NULL }
};


static int
dissect_nbap_MidambleConfigurationBurstType2_768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_MidambleShiftShort768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 7U, NULL, FALSE);

  return offset;
}


static const value_string nbap_MidambleAllocationMode7682_vals[] = {
  {   0, "defaultMidamble" },
  {   1, "commonMidamble" },
  {   2, "ueSpecificMidamble" },
  { 0, NULL }
};

static const per_choice_t MidambleAllocationMode7682_choice[] = {
  {   0, &hf_nbap_defaultMidamble, ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {   1, &hf_nbap_commonMidamble , ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {   2, &hf_nbap_ueSpecificMidamble_02, ASN1_EXTENSION_ROOT    , dissect_nbap_MidambleShiftShort768 },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_MidambleAllocationMode7682(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_MidambleAllocationMode7682, MidambleAllocationMode7682_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Type7682_sequence[] = {
  { &hf_nbap_midambleConfigurationBurstType2_768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleConfigurationBurstType2_768 },
  { &hf_nbap_midambleAllocationMode_05, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleAllocationMode7682 },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Type7682(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Type7682, Type7682_sequence);

  return offset;
}


static const value_string nbap_MidambleAllocationMode7683_vals[] = {
  {   0, "defaultMidamble" },
  {   1, "ueSpecificMidamble" },
  { 0, NULL }
};

static const per_choice_t MidambleAllocationMode7683_choice[] = {
  {   0, &hf_nbap_defaultMidamble, ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {   1, &hf_nbap_ueSpecificMidamble, ASN1_EXTENSION_ROOT    , dissect_nbap_MidambleShiftLong },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_MidambleAllocationMode7683(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_MidambleAllocationMode7683, MidambleAllocationMode7683_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Type7683_sequence[] = {
  { &hf_nbap_midambleConfigurationBurstType1And3, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleConfigurationBurstType1And3 },
  { &hf_nbap_midambleAllocationMode_06, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleAllocationMode7683 },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Type7683(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Type7683, Type7683_sequence);

  return offset;
}


static const value_string nbap_MidambleShiftAndBurstType768_vals[] = {
  {   0, "type1" },
  {   1, "type2" },
  {   2, "type3" },
  { 0, NULL }
};

static const per_choice_t MidambleShiftAndBurstType768_choice[] = {
  {   0, &hf_nbap_type1_01       , ASN1_EXTENSION_ROOT    , dissect_nbap_Type7681 },
  {   1, &hf_nbap_type2_01       , ASN1_EXTENSION_ROOT    , dissect_nbap_Type7682 },
  {   2, &hf_nbap_type3_01       , ASN1_EXTENSION_ROOT    , dissect_nbap_Type7683 },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_MidambleShiftAndBurstType768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_MidambleShiftAndBurstType768, MidambleShiftAndBurstType768_choice,
                                 NULL);

  return offset;
}


static const value_string nbap_TDD_ChannelisationCode768_vals[] = {
  {   0, "chCode1div1" },
  {   1, "chCode2div1" },
  {   2, "chCode2div2" },
  {   3, "chCode4div1" },
  {   4, "chCode4div2" },
  {   5, "chCode4div3" },
  {   6, "chCode4div4" },
  {   7, "chCode8div1" },
  {   8, "chCode8div2" },
  {   9, "chCode8div3" },
  {  10, "chCode8div4" },
  {  11, "chCode8div5" },
  {  12, "chCode8div6" },
  {  13, "chCode8div7" },
  {  14, "chCode8div8" },
  {  15, "chCode16div1" },
  {  16, "chCode16div2" },
  {  17, "chCode16div3" },
  {  18, "chCode16div4" },
  {  19, "chCode16div5" },
  {  20, "chCode16div6" },
  {  21, "chCode16div7" },
  {  22, "chCode16div8" },
  {  23, "chCode16div9" },
  {  24, "chCode16div10" },
  {  25, "chCode16div11" },
  {  26, "chCode16div12" },
  {  27, "chCode16div13" },
  {  28, "chCode16div14" },
  {  29, "chCode16div15" },
  {  30, "chCode16div16" },
  {  31, "chCode32div1" },
  {  32, "chCode32div2" },
  {  33, "chCode32div3" },
  {  34, "chCode32div4" },
  {  35, "chCode32div5" },
  {  36, "chCode32div6" },
  {  37, "chCode32div7" },
  {  38, "chCode32div8" },
  {  39, "chCode32div9" },
  {  40, "chCode32div10" },
  {  41, "chCode32div11" },
  {  42, "chCode32div12" },
  {  43, "chCode32div13" },
  {  44, "chCode32div14" },
  {  45, "chCode32div15" },
  {  46, "chCode32div16" },
  {  47, "chCode32div17" },
  {  48, "chCode32div18" },
  {  49, "chCode32div19" },
  {  50, "chCode32div20" },
  {  51, "chCode32div21" },
  {  52, "chCode32div22" },
  {  53, "chCode32div23" },
  {  54, "chCode32div24" },
  {  55, "chCode32div25" },
  {  56, "chCode32div26" },
  {  57, "chCode32div27" },
  {  58, "chCode32div28" },
  {  59, "chCode32div29" },
  {  60, "chCode32div30" },
  {  61, "chCode32div31" },
  {  62, "chCode32div32" },
  { 0, NULL }
};

static value_string_ext nbap_TDD_ChannelisationCode768_vals_ext = VALUE_STRING_EXT_INIT(nbap_TDD_ChannelisationCode768_vals);


static int
dissect_nbap_TDD_ChannelisationCode768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     63, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t TDD_DL_Code_768_InformationItem_sequence[] = {
  { &hf_nbap_dPCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DPCH_ID },
  { &hf_nbap_tdd_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode768 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TDD_DL_Code_768_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TDD_DL_Code_768_InformationItem, TDD_DL_Code_768_InformationItem_sequence);

  return offset;
}


static const per_sequence_t TDD_DL_Code_768_Information_sequence_of[1] = {
  { &hf_nbap_TDD_DL_Code_768_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DL_Code_768_InformationItem },
};

static int
dissect_nbap_TDD_DL_Code_768_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_TDD_DL_Code_768_Information, TDD_DL_Code_768_Information_sequence_of,
                                                  1, maxNrOfDPCHs768, FALSE);

  return offset;
}


static const per_sequence_t DL_Timeslot768_InformationItem_sequence[] = {
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType768 },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCI_Presence },
  { &hf_nbap_dL_Code_768_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DL_Code_768_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_Timeslot768_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_Timeslot768_InformationItem, DL_Timeslot768_InformationItem_sequence);

  return offset;
}


static const per_sequence_t DL_Timeslot768_Information_sequence_of[1] = {
  { &hf_nbap_DL_Timeslot768_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Timeslot768_InformationItem },
};

static int
dissect_nbap_DL_Timeslot768_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_Timeslot768_Information, DL_Timeslot768_Information_sequence_of,
                                                  1, maxNrOfDLTSs, FALSE);

  return offset;
}


static const value_string nbap_DL_FrameType_vals[] = {
  {   0, "typeA" },
  {   1, "typeB" },
  { 0, NULL }
};


static int
dissect_nbap_DL_FrameType(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_DL_or_Global_CapacityCredit(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 65535U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_DLPowerAveragingWindowSize(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 60U, NULL, FALSE);

  return offset;
}


static const value_string nbap_PowerAdjustmentType_vals[] = {
  {   0, "none" },
  {   1, "common" },
  {   2, "individual" },
  { 0, NULL }
};


static int
dissect_nbap_PowerAdjustmentType(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t DL_ReferencePowerInformationItem_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_dl_Reference_Power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_ReferencePowerInformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_ReferencePowerInformationItem, DL_ReferencePowerInformationItem_sequence);

  return offset;
}


static const per_sequence_t DL_ReferencePowerInformationList_sequence_of[1] = {
  { &hf_nbap_DL_ReferencePowerInformationList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_ReferencePowerInformationItem },
};

static int
dissect_nbap_DL_ReferencePowerInformationList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_ReferencePowerInformationList, DL_ReferencePowerInformationList_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}



static int
dissect_nbap_MaxAdjustmentStep(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 10U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_AdjustmentPeriod(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 256U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_ScaledAdjustmentRatio(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 100U, NULL, FALSE);

  return offset;
}


static const per_sequence_t DL_PowerBalancing_Information_sequence[] = {
  { &hf_nbap_powerAdjustmentType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PowerAdjustmentType },
  { &hf_nbap_dLReferencePower, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_dLReferencePowerList_DL_PC_Rqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_ReferencePowerInformationList },
  { &hf_nbap_maxAdjustmentStep, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MaxAdjustmentStep },
  { &hf_nbap_adjustmentPeriod, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_AdjustmentPeriod },
  { &hf_nbap_adjustmentRatio, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ScaledAdjustmentRatio },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_PowerBalancing_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_PowerBalancing_Information, DL_PowerBalancing_Information_sequence);

  return offset;
}



static int
dissect_nbap_DL_TimeslotISCP(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 91U, NULL, FALSE);

  return offset;
}


static const per_sequence_t DL_TimeslotISCPInfoItem_sequence[] = {
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_dL_TimeslotISCP, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_TimeslotISCP },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_TimeslotISCPInfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_TimeslotISCPInfoItem, DL_TimeslotISCPInfoItem_sequence);

  return offset;
}


static const per_sequence_t DL_TimeslotISCPInfo_sequence_of[1] = {
  { &hf_nbap_DL_TimeslotISCPInfo_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_TimeslotISCPInfoItem },
};

static int
dissect_nbap_DL_TimeslotISCPInfo(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_TimeslotISCPInfo, DL_TimeslotISCPInfo_sequence_of,
                                                  1, maxNrOfDLTSs, FALSE);

  return offset;
}


static const per_sequence_t DL_TimeslotISCPInfoItemLCR_sequence[] = {
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_dL_TimeslotISCP, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_TimeslotISCP },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_TimeslotISCPInfoItemLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_TimeslotISCPInfoItemLCR, DL_TimeslotISCPInfoItemLCR_sequence);

  return offset;
}


static const per_sequence_t DL_TimeslotISCPInfoLCR_sequence_of[1] = {
  { &hf_nbap_DL_TimeslotISCPInfoLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_TimeslotISCPInfoItemLCR },
};

static int
dissect_nbap_DL_TimeslotISCPInfoLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_TimeslotISCPInfoLCR, DL_TimeslotISCPInfoLCR_sequence_of,
                                                  1, maxNrOfDLTSLCRs, FALSE);

  return offset;
}



static int
dissect_nbap_DL_TPC_Pattern01Count(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 30U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_DLTransmissionBranchLoadValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 101U, NULL, TRUE);

  return offset;
}


static const value_string nbap_Downlink_Compressed_Mode_Method_vals[] = {
  {   0, "not-Used-puncturing" },
  {   1, "sFdiv2" },
  {   2, "higher-layer-scheduling" },
  { 0, NULL }
};


static int
dissect_nbap_Downlink_Compressed_Mode_Method(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t DL_HS_PDSCH_Codelist_LCR_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_DL_HS_PDSCH_Codelist_LCR_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
};

static int
dissect_nbap_DL_HS_PDSCH_Codelist_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_HS_PDSCH_Codelist_LCR_PSCH_ReconfRqst, DL_HS_PDSCH_Codelist_LCR_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfHSPDSCHs, FALSE);

  return offset;
}



static int
dissect_nbap_MaximumTransmissionPower(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 500U, NULL, FALSE);

  return offset;
}


static const per_sequence_t DL_HS_PDSCH_Timeslot_InformationItem_LCR_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftAndBurstType_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_dl_HS_PDSCH_Codelist_LCR_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_HS_PDSCH_Codelist_LCR_PSCH_ReconfRqst },
  { &hf_nbap_maxHSDSCH_HSSCCH_Power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MaximumTransmissionPower },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_HS_PDSCH_Timeslot_InformationItem_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_HS_PDSCH_Timeslot_InformationItem_LCR_PSCH_ReconfRqst, DL_HS_PDSCH_Timeslot_InformationItem_LCR_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_HS_PDSCH_Timeslot_InformationItem_LCR_PSCH_ReconfRqst },
};

static int
dissect_nbap_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst, DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfDLTSLCRs, FALSE);

  return offset;
}


static const per_sequence_t MaxHSDSCH_HSSCCH_Power_per_CELLPORTION_Item_sequence[] = {
  { &hf_nbap_cellPortionLCRID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellPortionLCRID },
  { &hf_nbap_maxHSDSCH_HSSCCH_Power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MaximumTransmissionPower },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MaxHSDSCH_HSSCCH_Power_per_CELLPORTION_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MaxHSDSCH_HSSCCH_Power_per_CELLPORTION_Item, MaxHSDSCH_HSSCCH_Power_per_CELLPORTION_Item_sequence);

  return offset;
}


static const per_sequence_t MaxHSDSCH_HSSCCH_Power_per_CELLPORTION_sequence_of[1] = {
  { &hf_nbap_MaxHSDSCH_HSSCCH_Power_per_CELLPORTION_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_MaxHSDSCH_HSSCCH_Power_per_CELLPORTION_Item },
};

static int
dissect_nbap_MaxHSDSCH_HSSCCH_Power_per_CELLPORTION(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_MaxHSDSCH_HSSCCH_Power_per_CELLPORTION, MaxHSDSCH_HSSCCH_Power_per_CELLPORTION_sequence_of,
                                                  1, maxNrOfCellPortionsPerCellLCR, FALSE);

  return offset;
}



static int
dissect_nbap_DPCH_ID768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 479U, NULL, FALSE);

  return offset;
}


static const value_string nbap_DRX_Interruption_by_HS_DSCH_vals[] = {
  {   0, "drx-Interruption-Configured" },
  {   1, "drx-Interruption-Not-Configured" },
  { 0, NULL }
};


static int
dissect_nbap_DRX_Interruption_by_HS_DSCH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_DSCH_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t DSCH_InformationResponseItem_sequence[] = {
  { &hf_nbap_dSCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DSCH_ID },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DSCH_InformationResponseItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DSCH_InformationResponseItem, DSCH_InformationResponseItem_sequence);

  return offset;
}


static const per_sequence_t DSCH_InformationResponse_sequence_of[1] = {
  { &hf_nbap_DSCH_InformationResponse_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DSCH_InformationResponseItem },
};

static int
dissect_nbap_DSCH_InformationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DSCH_InformationResponse, DSCH_InformationResponse_sequence_of,
                                                  1, maxNrOfDSCHs, FALSE);

  return offset;
}


static const per_sequence_t DSCH_TDD_InformationItem_sequence[] = {
  { &hf_nbap_dSCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DSCH_ID },
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_transportFormatSet, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportFormatSet },
  { &hf_nbap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AllocationRetentionPriority },
  { &hf_nbap_frameHandlingPriority, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FrameHandlingPriority },
  { &hf_nbap_toAWS          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ToAWS },
  { &hf_nbap_toAWE          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ToAWE },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DSCH_TDD_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DSCH_TDD_InformationItem, DSCH_TDD_InformationItem_sequence);

  return offset;
}


static const per_sequence_t DSCH_TDD_Information_sequence_of[1] = {
  { &hf_nbap_DSCH_TDD_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DSCH_TDD_InformationItem },
};

static int
dissect_nbap_DSCH_TDD_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DSCH_TDD_Information, DSCH_TDD_Information_sequence_of,
                                                  1, maxNrOfDSCHs, FALSE);

  return offset;
}


static const value_string nbap_Dual_Band_Capability_vals[] = {
  {   0, "dual-Band-Capable" },
  {   1, "dual-Band-non-Capable" },
  { 0, NULL }
};


static int
dissect_nbap_Dual_Band_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_Local_Cell_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 268435455U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Possible_Secondary_Serving_Cell_sequence[] = {
  { &hf_nbap_local_Cell_ID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Local_Cell_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Possible_Secondary_Serving_Cell(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Possible_Secondary_Serving_Cell, Possible_Secondary_Serving_Cell_sequence);

  return offset;
}


static const per_sequence_t Possible_Secondary_Serving_Cell_List_sequence_of[1] = {
  { &hf_nbap_Possible_Secondary_Serving_Cell_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Possible_Secondary_Serving_Cell },
};

static int
dissect_nbap_Possible_Secondary_Serving_Cell_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Possible_Secondary_Serving_Cell_List, Possible_Secondary_Serving_Cell_List_sequence_of,
                                                  1, maxNrOfHSDSCH_1, FALSE);

  return offset;
}


static const per_sequence_t Dual_Band_Capability_Info_sequence[] = {
  { &hf_nbap_dual_Band_Capability, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Dual_Band_Capability },
  { &hf_nbap_possible_Secondary_Serving_Cell_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Possible_Secondary_Serving_Cell_List },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Dual_Band_Capability_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Dual_Band_Capability_Info, Dual_Band_Capability_Info_sequence);

  return offset;
}



static int
dissect_nbap_DwPCH_Power(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -150, 400U, NULL, TRUE);

  return offset;
}


static const value_string nbap_E_AGCH_Table_Choice_vals[] = {
  {   0, "table16B" },
  {   1, "table16B-1" },
  { 0, NULL }
};


static int
dissect_nbap_E_AGCH_Table_Choice(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t E_AGCH_FDD_Code_List_sequence_of[1] = {
  { &hf_nbap_E_AGCH_FDD_Code_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_DL_ChannelisationCodeNumber },
};

static int
dissect_nbap_E_AGCH_FDD_Code_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_AGCH_FDD_Code_List, E_AGCH_FDD_Code_List_sequence_of,
                                                  1, maxNrOfE_AGCHs, FALSE);

  return offset;
}


static const value_string nbap_E_AGCH_FDD_Code_Information_vals[] = {
  {   0, "replace" },
  {   1, "remove" },
  { 0, NULL }
};

static const per_choice_t E_AGCH_FDD_Code_Information_choice[] = {
  {   0, &hf_nbap_replace        , ASN1_EXTENSION_ROOT    , dissect_nbap_E_AGCH_FDD_Code_List },
  {   1, &hf_nbap_remove         , ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_E_AGCH_FDD_Code_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_E_AGCH_FDD_Code_Information, E_AGCH_FDD_Code_Information_choice,
                                 NULL);

  return offset;
}


static const value_string nbap_E_AI_Capability_vals[] = {
  {   0, "e-AI-capable" },
  {   1, "e-AI-non-capable" },
  { 0, NULL }
};


static int
dissect_nbap_E_AI_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_E_DCH_Capability_vals[] = {
  {   0, "e-DCH-capable" },
  {   1, "e-DCH-non-capable" },
  { 0, NULL }
};


static int
dissect_nbap_E_DCH_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t E_DCH_SF_allocation_item_sequence[] = {
  { &hf_nbap_ul_Cost_1      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_65535 },
  { &hf_nbap_ul_Cost_2      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_65535 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_SF_allocation_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_SF_allocation_item, E_DCH_SF_allocation_item_sequence);

  return offset;
}


static const per_sequence_t E_DCH_SF_allocation_sequence_of[1] = {
  { &hf_nbap_E_DCH_SF_allocation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_SF_allocation_item },
};

static int
dissect_nbap_E_DCH_SF_allocation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_DCH_SF_allocation, E_DCH_SF_allocation_sequence_of,
                                                  1, maxNrOfCombEDPDCH, FALSE);

  return offset;
}


static const per_sequence_t E_DCHCapacityConsumptionLaw_sequence[] = {
  { &hf_nbap_e_DCH_SF_allocation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_SF_allocation },
  { &hf_nbap_dl_Cost_1      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_INTEGER_0_65535 },
  { &hf_nbap_dl_Cost_2      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_INTEGER_0_65535 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCHCapacityConsumptionLaw(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCHCapacityConsumptionLaw, E_DCHCapacityConsumptionLaw_sequence);

  return offset;
}


static const per_sequence_t E_DCH_TDD_CapacityConsumptionLaw_sequence[] = {
  { &hf_nbap_ul_Cost        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_65535 },
  { &hf_nbap_dl_Cost        , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_INTEGER_0_65535 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_TDD_CapacityConsumptionLaw(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_TDD_CapacityConsumptionLaw, E_DCH_TDD_CapacityConsumptionLaw_sequence);

  return offset;
}



static int
dissect_nbap_E_DCH_TTI2ms_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_boolean(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}


static const value_string nbap_E_DCH_SF_Capability_vals[] = {
  {   0, "sf64" },
  {   1, "sf32" },
  {   2, "sf16" },
  {   3, "sf8" },
  {   4, "sf4" },
  {   5, "sf4x2" },
  {   6, "sf2x2" },
  {   7, "sf4x2-and-sf2x2" },
  { 0, NULL }
};


static int
dissect_nbap_E_DCH_SF_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_E_DCH_HARQ_Combining_Capability_vals[] = {
  {   0, "iR-Combining-capable" },
  {   1, "chase-Combining-capable" },
  {   2, "iR-and-Chase-Combining-capable" },
  { 0, NULL }
};


static int
dissect_nbap_E_DCH_HARQ_Combining_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_E_DCH_DDI_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  guint32 e_dch_ddi_value;
  guint num_items;
  nbap_edch_channel_info_t* nbap_edch_channel_info;
  nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
  nbap_edch_channel_info = nbap_private_data->nbap_edch_channel_info;
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 62U, &e_dch_ddi_value, FALSE);


  nbap_private_data->e_dch_ddi_value = e_dch_ddi_value;
  if (PINFO_FD_VISITED(actx->pinfo))
  {
      return offset;
  }
  num_items = nbap_private_data->num_items;
  if(num_items > 0 && num_items < MAX_EDCH_DDIS + 1)
      nbap_edch_channel_info[nbap_private_data->e_dch_macdflow_id].edch_ddi[num_items-1] = nbap_private_data->e_dch_ddi_value;


  return offset;
}



static int
dissect_nbap_Max_Bits_MACe_PDU_non_scheduled(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxNrOfBits_MACe_PDU_non_scheduled, NULL, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_Non_Scheduled_Transmission_Grant_Items_sequence[] = {
  { &hf_nbap_maxBits_MACe_PDU_non_scheduled, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Max_Bits_MACe_PDU_non_scheduled },
  { &hf_nbap_hARQ_Process_Allocation_NonSched_2ms, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HARQ_Process_Allocation_2ms_EDCH },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_Non_Scheduled_Transmission_Grant_Items(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_Non_Scheduled_Transmission_Grant_Items, E_DCH_Non_Scheduled_Transmission_Grant_Items_sequence);

  return offset;
}


static const value_string nbap_E_DCH_Grant_Type_Information_vals[] = {
  {   0, "e-DCH-Non-Scheduled-Transmission-Grant" },
  {   1, "e-DCH-Scheduled-Transmission-Grant" },
  { 0, NULL }
};

static const per_choice_t E_DCH_Grant_Type_Information_choice[] = {
  {   0, &hf_nbap_e_DCH_Non_Scheduled_Transmission_Grant, ASN1_EXTENSION_ROOT    , dissect_nbap_E_DCH_Non_Scheduled_Transmission_Grant_Items },
  {   1, &hf_nbap_e_DCH_Scheduled_Transmission_Grant, ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_E_DCH_Grant_Type_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_E_DCH_Grant_Type_Information, E_DCH_Grant_Type_Information_choice,
                                 NULL);

  return offset;
}


static const value_string nbap_SchedulingInformation_vals[] = {
  {   0, "included" },
  {   1, "not-included" },
  { 0, NULL }
};


static int
dissect_nbap_SchedulingInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_MACesGuaranteedBitRate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 16777215U, NULL, TRUE);

  return offset;
}


static const per_sequence_t E_DCH_MACdPDU_SizeListItem_sequence[] = {
  { &hf_nbap_mACdPDU_Size   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MACdPDU_Size },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_MACdPDU_SizeListItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_MACdPDU_SizeListItem, E_DCH_MACdPDU_SizeListItem_sequence);

  return offset;
}


static const per_sequence_t E_DCH_MACdPDU_SizeList_sequence_of[1] = {
  { &hf_nbap_E_DCH_MACdPDU_SizeList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdPDU_SizeListItem },
};

static int
dissect_nbap_E_DCH_MACdPDU_SizeList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_DCH_MACdPDU_SizeList, E_DCH_MACdPDU_SizeList_sequence_of,
                                                  1, maxNrOfMACdPDUSize, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_LogicalChannelInformationItem_sequence[] = {
  { &hf_nbap_logicalChannelId, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_LogicalChannelID },
  { &hf_nbap_schedulingPriorityIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SchedulingPriorityIndicator },
  { &hf_nbap_schedulingInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SchedulingInformation },
  { &hf_nbap_mACesGuaranteedBitRate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MACesGuaranteedBitRate },
  { &hf_nbap_e_DCH_DDI_Value, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_DDI_Value },
  { &hf_nbap_mACd_PDU_Size_List, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdPDU_SizeList },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_LogicalChannelInformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    nbap_get_private_data(actx->pinfo)->num_items++;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_LogicalChannelInformationItem, E_DCH_LogicalChannelInformationItem_sequence);

  return offset;
}


static const per_sequence_t E_DCH_LogicalChannelInformation_sequence_of[1] = {
  { &hf_nbap_E_DCH_LogicalChannelInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_LogicalChannelInformationItem },
};

static int
dissect_nbap_E_DCH_LogicalChannelInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
    nbap_private_data->num_items = 0;
    nbap_edch_channel_info_t* nbap_edch_channel_info;
    nbap_edch_channel_info = nbap_private_data->nbap_edch_channel_info;

  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_DCH_LogicalChannelInformation, E_DCH_LogicalChannelInformation_sequence_of,
                                                  1, maxNoOfLogicalChannels, FALSE);


nbap_edch_channel_info[nbap_private_data->e_dch_macdflow_id].no_ddi_entries = nbap_private_data->num_items;


  return offset;
}


static const per_sequence_t E_DCH_MACdFlow_Specific_InfoItem_sequence[] = {
  { &hf_nbap_e_DCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_ID },
  { &hf_nbap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AllocationRetentionPriority },
  { &hf_nbap_tnlQos         , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TnlQos },
  { &hf_nbap_payloadCRC_PresenceIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PayloadCRC_PresenceIndicator },
  { &hf_nbap_maximum_Number_of_Retransmissions_For_E_DCH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Maximum_Number_of_Retransmissions_For_E_DCH },
  { &hf_nbap_eDCH_HARQ_PO_FDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_HARQ_PO_FDD },
  { &hf_nbap_eDCH_MACdFlow_Multiplexing_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_MACdFlow_Multiplexing_List },
  { &hf_nbap_eDCH_Grant_Type_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_Grant_Type_Information },
  { &hf_nbap_bundlingModeIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BundlingModeIndicator },
  { &hf_nbap_eDCHLogicalChannelInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_LogicalChannelInformation },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_MACdFlow_Specific_InfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
umts_fp_conversation_info_t *p_conv_data = NULL;
fp_edch_channel_info_t* fp_edch_channel_info = NULL;
address     null_addr;
conversation_t   *p_conv;
guint32 no_ddi_entries, i;
guint32 e_dch_macdflow_id;
nbap_edch_channel_info_t* nbap_edch_channel_info;
nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
nbap_edch_channel_info = nbap_private_data->nbap_edch_channel_info;

    /* Resetting entity recognition flag to check if it's set in this InfoItem */
    nbap_private_data->max_mac_d_pdu_size_ext_ie_present = FALSE;
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_MACdFlow_Specific_InfoItem, E_DCH_MACdFlow_Specific_InfoItem_sequence);

    if (PINFO_FD_VISITED(actx->pinfo)) {
        return offset;
    }

    /* Check if we have conversation info */
    e_dch_macdflow_id = nbap_private_data->e_dch_macdflow_id;
    clear_address(&null_addr);
    p_conv = find_conversation(actx->pinfo->num, &nbap_edch_channel_info[e_dch_macdflow_id].crnc_address, &null_addr,
                               CONVERSATION_UDP, nbap_edch_channel_info[e_dch_macdflow_id].crnc_port, 0, NO_ADDR_B);
    if(!p_conv)
        return offset;

    p_conv_data = (umts_fp_conversation_info_t *)conversation_get_proto_data(p_conv, proto_fp);

    if(!p_conv_data)
        return offset;

    fp_edch_channel_info = (fp_edch_channel_info_t*)p_conv_data->channel_specific_info;
    if(p_conv_data->channel != CHANNEL_EDCH  || !fp_edch_channel_info)
        return offset;

    no_ddi_entries = fp_edch_channel_info->no_ddi_entries = nbap_edch_channel_info[e_dch_macdflow_id].no_ddi_entries;
    for (i = 0; i < no_ddi_entries; i++) {
        fp_edch_channel_info->edch_ddi[i]            = nbap_edch_channel_info[e_dch_macdflow_id].edch_ddi[i];
        fp_edch_channel_info->edch_macd_pdu_size[i]  = nbap_edch_channel_info[e_dch_macdflow_id].edch_macd_pdu_size[i];

        fp_edch_channel_info->edch_lchId[i]          =  nbap_edch_channel_info[e_dch_macdflow_id].lchId[i];
    }

    p_conv_data->dch_crc_present = nbap_private_data->dch_crc_present;

    /* Figure out MAC entity: MAC-e/es or MAC-i/is
     * Then derive the type of E-DCH frame:
     * MAC-e/es => Type 1
     * MAC-i/is => Type 2
     * The specifications isn't very clear about the indicator for what entity
     * should be used. For now, it seems like the presence of the "Maximum MAC-d PDU Size Extended IE"
     * indicates MAC-i/is and it's absense means MAC-e/es
     */
    if(nbap_private_data->max_mac_d_pdu_size_ext_ie_present){
        fp_edch_channel_info->edch_type = 1; /* 1 means Type 2 */
    }else{
        fp_edch_channel_info->edch_type = 0;  /* 0 means Type 1 */
    }
    /* use to display e_dch_macdflow_id */
    p_conv_data->num_dch_in_flow = 1;
    p_conv_data->dch_ids_in_flow_list[0] = nbap_private_data->e_dch_macdflow_id;


  return offset;
}


static const per_sequence_t E_DCH_MACdFlow_Specific_InfoList_sequence_of[1] = {
  { &hf_nbap_E_DCH_MACdFlow_Specific_InfoList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_Specific_InfoItem },
};

static int
dissect_nbap_E_DCH_MACdFlow_Specific_InfoList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_DCH_MACdFlow_Specific_InfoList, E_DCH_MACdFlow_Specific_InfoList_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_MACdFlows_Information_sequence[] = {
  { &hf_nbap_e_DCH_MACdFlow_Specific_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_Specific_InfoList },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_MACdFlows_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_MACdFlows_Information, E_DCH_MACdFlows_Information_sequence);

  return offset;
}


static const per_sequence_t E_DCH_FDD_Information_sequence[] = {
  { &hf_nbap_e_DCH_MACdFlows_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlows_Information },
  { &hf_nbap_hARQ_Process_Allocation_Scheduled_2ms_EDCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HARQ_Process_Allocation_2ms_EDCH },
  { &hf_nbap_e_DCH_Maximum_Bitrate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_Maximum_Bitrate },
  { &hf_nbap_e_DCH_Processing_Overload_Level, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_Processing_Overload_Level },
  { &hf_nbap_e_DCH_Reference_Power_Offset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_Reference_Power_Offset },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_FDD_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_FDD_Information, E_DCH_FDD_Information_sequence);

  return offset;
}


static const per_sequence_t E_DCH_MACdFlow_Specific_InformationResp_Item_sequence[] = {
  { &hf_nbap_e_DCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_ID },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_hARQ_Process_Allocation_NonSched_2ms_EDCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HARQ_Process_Allocation_2ms_EDCH },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_MACdFlow_Specific_InformationResp_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_MACdFlow_Specific_InformationResp_Item, E_DCH_MACdFlow_Specific_InformationResp_Item_sequence);

  return offset;
}


static const per_sequence_t E_DCH_MACdFlow_Specific_InformationResp_sequence_of[1] = {
  { &hf_nbap_E_DCH_MACdFlow_Specific_InformationResp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_Specific_InformationResp_Item },
};

static int
dissect_nbap_E_DCH_MACdFlow_Specific_InformationResp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_DCH_MACdFlow_Specific_InformationResp, E_DCH_MACdFlow_Specific_InformationResp_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_FDD_Information_Response_sequence[] = {
  { &hf_nbap_e_DCH_MACdFlow_Specific_InformationResp, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_MACdFlow_Specific_InformationResp },
  { &hf_nbap_hARQ_Process_Allocation_Scheduled_2ms_EDCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HARQ_Process_Allocation_2ms_EDCH },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_FDD_Information_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_FDD_Information_Response, E_DCH_FDD_Information_Response_sequence);

  return offset;
}


static const per_sequence_t E_DCH_MACdPDU_SizeToModifyList_sequence_of[1] = {
  { &hf_nbap_E_DCH_MACdPDU_SizeToModifyList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdPDU_SizeListItem },
};

static int
dissect_nbap_E_DCH_MACdPDU_SizeToModifyList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_DCH_MACdPDU_SizeToModifyList, E_DCH_MACdPDU_SizeToModifyList_sequence_of,
                                                  0, maxNrOfMACdPDUSize, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_LogicalChannelToModifyItem_sequence[] = {
  { &hf_nbap_logicalChannelId, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_LogicalChannelID },
  { &hf_nbap_schedulingPriorityIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SchedulingPriorityIndicator },
  { &hf_nbap_schedulingInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SchedulingInformation },
  { &hf_nbap_mACesGuaranteedBitRate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MACesGuaranteedBitRate },
  { &hf_nbap_e_DCH_DDI_Value, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_DDI_Value },
  { &hf_nbap_mACd_PDU_Size_List_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdPDU_SizeToModifyList },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_LogicalChannelToModifyItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_LogicalChannelToModifyItem, E_DCH_LogicalChannelToModifyItem_sequence);

    nbap_get_private_data(actx->pinfo)->num_items++;

  return offset;
}


static const per_sequence_t E_DCH_LogicalChannelToModify_sequence_of[1] = {
  { &hf_nbap_E_DCH_LogicalChannelToModify_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_LogicalChannelToModifyItem },
};

static int
dissect_nbap_E_DCH_LogicalChannelToModify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_DCH_LogicalChannelToModify, E_DCH_LogicalChannelToModify_sequence_of,
                                                  1, maxNoOfLogicalChannels, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_LogicalChannelToDeleteItem_sequence[] = {
  { &hf_nbap_logicalChannelId, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_LogicalChannelID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_LogicalChannelToDeleteItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_LogicalChannelToDeleteItem, E_DCH_LogicalChannelToDeleteItem_sequence);

  return offset;
}


static const per_sequence_t E_DCH_LogicalChannelToDelete_sequence_of[1] = {
  { &hf_nbap_E_DCH_LogicalChannelToDelete_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_LogicalChannelToDeleteItem },
};

static int
dissect_nbap_E_DCH_LogicalChannelToDelete(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_DCH_LogicalChannelToDelete, E_DCH_LogicalChannelToDelete_sequence_of,
                                                  1, maxNoOfLogicalChannels, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_MACdFlow_Specific_InfoItem_to_Modify_sequence[] = {
  { &hf_nbap_e_DCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_ID },
  { &hf_nbap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_AllocationRetentionPriority },
  { &hf_nbap_transportBearerRequestIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportBearerRequestIndicator },
  { &hf_nbap_tnlQos         , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TnlQos },
  { &hf_nbap_maximum_Number_of_Retransmissions_For_E_DCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Maximum_Number_of_Retransmissions_For_E_DCH },
  { &hf_nbap_eDCH_HARQ_PO_FDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_HARQ_PO_FDD },
  { &hf_nbap_eDCH_MACdFlow_Multiplexing_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_MACdFlow_Multiplexing_List },
  { &hf_nbap_eDCH_Grant_Type_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_Grant_Type_Information },
  { &hf_nbap_bundlingModeIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BundlingModeIndicator },
  { &hf_nbap_eDCH_LogicalChannelToAdd, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_LogicalChannelInformation },
  { &hf_nbap_eDCH_LogicalChannelToModify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_LogicalChannelToModify },
  { &hf_nbap_eDCH_LogicalChannelToDelete, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_LogicalChannelToDelete },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_MACdFlow_Specific_InfoItem_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
guint32 no_ddi_entries, i;
address     null_addr;
nbap_edch_port_info_t *old_info;
umts_fp_conversation_info_t *p_conv_data = NULL;
fp_edch_channel_info_t* fp_edch_channel_info;
conversation_t   *p_conv;
guint32 e_dch_macdflow_id;
nbap_edch_channel_info_t* nbap_edch_channel_info;
nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
nbap_edch_channel_info = nbap_private_data->nbap_edch_channel_info;

nbap_private_data->num_items = 1;
   offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_MACdFlow_Specific_InfoItem_to_Modify, E_DCH_MACdFlow_Specific_InfoItem_to_Modify_sequence);


    if (PINFO_FD_VISITED(actx->pinfo))
    {
        return offset;
    }

    nbap_debug("Frame %u E-DCH-MACdFlow-Specific-InfoItem-to-Modify",
        actx->pinfo->num);

    /****** Look up old port and ip information since this is not included in this message ******/
    /*Find proper communication context ID*/
    if(nbap_private_data->crnc_context_present){
        /*umts_fp_conversation_info->com_context_id = nbap_private_data->com_context_id;*/
    }else{
        expert_add_info(actx->pinfo, NULL, &ei_nbap_no_set_comm_context_id);
    }

    /*This should not happen*/
    if(( old_info = (nbap_edch_port_info_t *)wmem_tree_lookup32(edch_flow_port_map,nbap_private_data->com_context_id)) == NULL ){
        expert_add_info(actx->pinfo, NULL, &ei_nbap_no_find_port_info);
        return offset;
    }
    nbap_debug("    Found com_context_id %u",nbap_private_data->com_context_id);

    /*Set the appropriate port, cheat and use same variable.*/
    e_dch_macdflow_id = nbap_private_data->e_dch_macdflow_id;
    nbap_private_data->binding_id_port = old_info->crnc_port[e_dch_macdflow_id];

    nbap_debug(" Port %u loaded from old_info->crnc_port[e_dch_macdflow_id %u]",nbap_private_data->binding_id_port, e_dch_macdflow_id);

    /*TODO: Fix this for ipv6 as well!*/
    nbap_private_data->transportLayerAddress_ipv4 = old_info->crnc_address;

    /*Do the configurations*/
    /* Check if we have conversation info */
    clear_address(&null_addr);
    p_conv = find_conversation(actx->pinfo->num, &nbap_edch_channel_info[e_dch_macdflow_id].crnc_address, &null_addr,
                               CONVERSATION_UDP, nbap_edch_channel_info[e_dch_macdflow_id].crnc_port, 0, NO_ADDR_B);
    if(!p_conv)
        return offset;

    p_conv_data = (umts_fp_conversation_info_t *)conversation_get_proto_data(p_conv, proto_fp);

    if(!p_conv_data)
        return offset;

    fp_edch_channel_info = (fp_edch_channel_info_t*)p_conv_data->channel_specific_info;
    if(p_conv_data->channel != CHANNEL_EDCH  || !fp_edch_channel_info)
        return offset;

    no_ddi_entries = fp_edch_channel_info->no_ddi_entries = nbap_edch_channel_info[e_dch_macdflow_id].no_ddi_entries;
    for (i = 0; i < no_ddi_entries; i++) {
        fp_edch_channel_info->edch_ddi[i]            = nbap_edch_channel_info[e_dch_macdflow_id].edch_ddi[i];
        fp_edch_channel_info->edch_macd_pdu_size[i]  = nbap_edch_channel_info[e_dch_macdflow_id].edch_macd_pdu_size[i];

        fp_edch_channel_info->edch_lchId[i]          =  nbap_edch_channel_info[e_dch_macdflow_id].lchId[i];
    }

    p_conv_data->dch_crc_present = nbap_private_data->dch_crc_present;

    /* Figure out MAC entity: MAC-e/es or MAC-i/is
     * Then derive the type of E-DCH frame:
     * MAC-e/es => Type 1
     * MAC-i/is => Type 2
     * The specifications isn't very clear about the indicator for what entity
     * should be used. For now, it seems like the presence of the "Maximum MAC-d PDU Size Extended IE"
     * indicates MAC-i/is and it's absense means MAC-e/es
     */
    if(nbap_private_data->max_mac_d_pdu_size_ext_ie_present){
        fp_edch_channel_info->edch_type = 1; /* 1 means Type 2 */
    }else{
        fp_edch_channel_info->edch_type = 0;  /* 0 means Type 1 */
    }
    /* use to display e_dch_macdflow_id */
    p_conv_data->num_dch_in_flow = 1;
    p_conv_data->dch_ids_in_flow_list[0] = e_dch_macdflow_id;


  return offset;
}


static const per_sequence_t E_DCH_MACdFlow_Specific_InfoList_to_Modify_sequence_of[1] = {
  { &hf_nbap_E_DCH_MACdFlow_Specific_InfoList_to_Modify_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_Specific_InfoItem_to_Modify },
};

static int
dissect_nbap_E_DCH_MACdFlow_Specific_InfoList_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_DCH_MACdFlow_Specific_InfoList_to_Modify, E_DCH_MACdFlow_Specific_InfoList_to_Modify_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}


static const value_string nbap_MACeReset_Indicator_vals[] = {
  {   0, "mACeReset" },
  { 0, NULL }
};


static int
dissect_nbap_MACeReset_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t E_DCH_FDD_Information_to_Modify_sequence[] = {
  { &hf_nbap_e_DCH_MACdFlow_Specific_Info_to_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_MACdFlow_Specific_InfoList_to_Modify },
  { &hf_nbap_hARQ_Process_Allocation_Scheduled_2ms_EDCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HARQ_Process_Allocation_2ms_EDCH },
  { &hf_nbap_e_DCH_Maximum_Bitrate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_Maximum_Bitrate },
  { &hf_nbap_e_DCH_Processing_Overload_Level, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_Processing_Overload_Level },
  { &hf_nbap_e_DCH_Reference_Power_Offset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_Reference_Power_Offset },
  { &hf_nbap_mACeReset_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MACeReset_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_FDD_Information_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {

address     dst_addr, null_addr;
conversation_t *conversation,*old_conversation = NULL;
umts_fp_conversation_info_t *umts_fp_conversation_info _U_;
fp_edch_channel_info_t* fp_edch_channel_info;
void *conv_proto_data = NULL;
guint32 transportLayerAddress_ipv4;
guint16 bindingID;
guint32 e_dch_macdflow_id;
nbap_edch_channel_info_t* nbap_edch_channel_info;
nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
nbap_edch_channel_info = nbap_private_data->nbap_edch_channel_info;

nbap_private_data->transportLayerAddress_ipv4 = 0;
nbap_private_data->binding_id_port = 0;

   offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_FDD_Information_to_Modify, E_DCH_FDD_Information_to_Modify_sequence);


        transportLayerAddress_ipv4 = nbap_private_data->transportLayerAddress_ipv4;
        bindingID = nbap_private_data->binding_id_port;
        if (PINFO_FD_VISITED(actx->pinfo) || transportLayerAddress_ipv4 == 0 || bindingID == 0){
            return offset;
        }
        /*
         * Basically the idea here is that we create a new converation (Which is ok? maybe?)
         * And then hijack the old conversation and let lower tree items configure that hijacked data.
         * */
        clear_address(&null_addr);

        set_address(&dst_addr, AT_IPv4, 4, &transportLayerAddress_ipv4);

        old_conversation = find_conversation(actx->pinfo->num, &dst_addr,
            &null_addr, CONVERSATION_UDP, bindingID,
            0, NO_ADDR_B|NO_PORT_B);

        if(old_conversation){
            nbap_debug("Frame %u E-DCH-FDD-Information-to-Modify: found old conv on IP %s Port %u",
                actx->pinfo->num,
                address_to_str(actx->pinfo->pool, &dst_addr), bindingID);
        }else{
            nbap_debug("Frame %u E-DCH-FDD-Information-to-Modify: Did not find old conv on IP %s Port %u",
                actx->pinfo->num,
                address_to_str(actx->pinfo->pool, &dst_addr), bindingID);
        }

        /* It's not part of any conversation - create a new one. */
        conversation = conversation_new(actx->pinfo->num, &dst_addr,
                &null_addr, CONVERSATION_UDP, bindingID,
                0, NO_ADDR2|NO_PORT2);

            /* Set dissector */
            conversation_set_dissector(conversation, fp_handle);
            if(actx->pinfo->link_dir==P2P_DIR_DL){
                umts_fp_conversation_info = wmem_new0(wmem_file_scope(), umts_fp_conversation_info_t);
                /* Steal the old information */
                if (old_conversation) {
                    conv_proto_data = conversation_get_proto_data(old_conversation, proto_fp);
                    if (conv_proto_data)
                        memcpy(umts_fp_conversation_info,conv_proto_data,sizeof(umts_fp_conversation_info_t));
                }
                /* Overwrite the data */
                umts_fp_conversation_info->iface_type        = IuB_Interface;
                umts_fp_conversation_info->division          = Division_FDD;
                umts_fp_conversation_info->channel           = CHANNEL_EDCH;
                umts_fp_conversation_info->dl_frame_number   = 0;
                umts_fp_conversation_info->ul_frame_number   = actx->pinfo->num;
                copy_address_wmem(wmem_file_scope(), &(umts_fp_conversation_info->crnc_address), &dst_addr);
                umts_fp_conversation_info->crnc_port         = bindingID;
                umts_fp_conversation_info->rlc_mode          = FP_RLC_MODE_UNKNOWN;

                fp_edch_channel_info = wmem_new0(wmem_file_scope(), fp_edch_channel_info_t);
                umts_fp_conversation_info->channel_specific_info = (void*)fp_edch_channel_info;

                if(nbap_private_data->crnc_context_present){
                    umts_fp_conversation_info->com_context_id = nbap_private_data->com_context_id;
                }else{
                    expert_add_info(actx->pinfo, NULL, &ei_nbap_no_set_comm_context_id);
                }

                /* Set address for collection of DDI entries */
                e_dch_macdflow_id = nbap_private_data->e_dch_macdflow_id;
                copy_address_wmem(actx->pinfo->pool,&(nbap_edch_channel_info[e_dch_macdflow_id].crnc_address),&dst_addr);
                nbap_edch_channel_info[e_dch_macdflow_id].crnc_port = bindingID;

                /*Indicate that the frag table has to be reset*/
                umts_fp_conversation_info->reset_frag = TRUE;

                set_umts_fp_conv_data(conversation, umts_fp_conversation_info);
            }


  return offset;
}


static const per_sequence_t E_DCH_MACdFlow_Specific_UpdateInformation_Item_sequence[] = {
  { &hf_nbap_e_DCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_ID },
  { &hf_nbap_hARQ_Process_Allocation_NonSched_2ms_EDCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HARQ_Process_Allocation_2ms_EDCH },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_MACdFlow_Specific_UpdateInformation_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_MACdFlow_Specific_UpdateInformation_Item, E_DCH_MACdFlow_Specific_UpdateInformation_Item_sequence);

  return offset;
}


static const per_sequence_t E_DCH_MACdFlow_Specific_UpdateInformation_sequence_of[1] = {
  { &hf_nbap_E_DCH_MACdFlow_Specific_UpdateInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_Specific_UpdateInformation_Item },
};

static int
dissect_nbap_E_DCH_MACdFlow_Specific_UpdateInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_DCH_MACdFlow_Specific_UpdateInformation, E_DCH_MACdFlow_Specific_UpdateInformation_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_FDD_Update_Information_sequence[] = {
  { &hf_nbap_e_DCH_MACdFlow_Specific_UpdateInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_MACdFlow_Specific_UpdateInformation },
  { &hf_nbap_hARQ_Process_Allocation_Scheduled_2ms_EDCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HARQ_Process_Allocation_2ms_EDCH },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_FDD_Update_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_FDD_Update_Information, E_DCH_FDD_Update_Information_sequence);

  return offset;
}


static const per_sequence_t E_DCH_DL_Control_Channel_Change_Information_Item_sequence[] = {
  { &hf_nbap_e_DCH_RL_ID    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_DL_Control_Channel_Change_Information_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_DL_Control_Channel_Change_Information_Item, E_DCH_DL_Control_Channel_Change_Information_Item_sequence);

  return offset;
}


static const per_sequence_t E_DCH_DL_Control_Channel_Change_Information_sequence_of[1] = {
  { &hf_nbap_E_DCH_DL_Control_Channel_Change_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_DL_Control_Channel_Change_Information_Item },
};

static int
dissect_nbap_E_DCH_DL_Control_Channel_Change_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_DCH_DL_Control_Channel_Change_Information, E_DCH_DL_Control_Channel_Change_Information_sequence_of,
                                                  1, maxNrOfEDCHRLs, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_DL_Control_Channel_Grant_Information_Item_sequence[] = {
  { &hf_nbap_e_DCH_RL_ID    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_DL_Control_Channel_Grant_Information_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_DL_Control_Channel_Grant_Information_Item, E_DCH_DL_Control_Channel_Grant_Information_Item_sequence);

  return offset;
}


static const per_sequence_t E_DCH_DL_Control_Channel_Grant_Information_sequence_of[1] = {
  { &hf_nbap_E_DCH_DL_Control_Channel_Grant_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_DL_Control_Channel_Grant_Information_Item },
};

static int
dissect_nbap_E_DCH_DL_Control_Channel_Grant_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_DCH_DL_Control_Channel_Grant_Information, E_DCH_DL_Control_Channel_Grant_Information_sequence_of,
                                                  1, maxNrOfEDCHRLs, FALSE);

  return offset;
}


static const value_string nbap_E_DCH_MACdPDU_SizeCapability_vals[] = {
  {   0, "fixedSizeCapable" },
  {   1, "flexibleSizeCapable" },
  { 0, NULL }
};


static int
dissect_nbap_E_DCH_MACdPDU_SizeCapability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_E_DCH_MACdPDUSizeFormat_vals[] = {
  {   0, "fixedMACdPDU-Size" },
  {   1, "flexibleMACdPDU-Size" },
  { 0, NULL }
};


static int
dissect_nbap_E_DCH_MACdPDUSizeFormat(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t E_DCH_MACdFlow_to_Delete_Item_sequence[] = {
  { &hf_nbap_e_DCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_MACdFlow_to_Delete_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_MACdFlow_to_Delete_Item, E_DCH_MACdFlow_to_Delete_Item_sequence);

  return offset;
}


static const per_sequence_t E_DCH_MACdFlows_to_Delete_sequence_of[1] = {
  { &hf_nbap_E_DCH_MACdFlows_to_Delete_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_to_Delete_Item },
};

static int
dissect_nbap_E_DCH_MACdFlows_to_Delete(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_DCH_MACdFlows_to_Delete, E_DCH_MACdFlows_to_Delete_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}



static int
dissect_nbap_E_DCH_Non_serving_Relative_Grant_Down_Commands(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 100U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_E_DCHProvidedBitRateValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 16777215U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_Maximum_Target_ReceivedTotalWideBandPower(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 621U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_Target_NonServing_EDCH_To_Total_EDCH_Power_Ratio(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 100U, NULL, FALSE);

  return offset;
}


static const value_string nbap_E_DCH_RL_Indication_vals[] = {
  {   0, "e-DCH" },
  {   1, "non-e-DCH" },
  { 0, NULL }
};


static int
dissect_nbap_E_DCH_RL_Indication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_Granted_EDCH_RACH_Resources_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 240U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_Denied_EDCH_RACH_Resources_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 240U, NULL, TRUE);

  return offset;
}


static const per_sequence_t EDCH_RACH_Report_Value_item_sequence[] = {
  { &hf_nbap_granted_EDCH_RACH_resources, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Granted_EDCH_RACH_Resources_Value },
  { &hf_nbap_denied_EDCH_RACH_resources, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Denied_EDCH_RACH_Resources_Value },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_EDCH_RACH_Report_Value_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_EDCH_RACH_Report_Value_item, EDCH_RACH_Report_Value_item_sequence);

  return offset;
}


static const per_sequence_t EDCH_RACH_Report_Value_sequence_of[1] = {
  { &hf_nbap_EDCH_RACH_Report_Value_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_EDCH_RACH_Report_Value_item },
};

static int
dissect_nbap_EDCH_RACH_Report_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_EDCH_RACH_Report_Value, EDCH_RACH_Report_Value_sequence_of,
                                                  1, maxNrOfCommonEDCH, FALSE);

  return offset;
}



static int
dissect_nbap_E_DPDCH_PowerInterpolation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_boolean(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}


static const per_sequence_t E_DCH_MACdFlow_to_Delete_ItemLCR_sequence[] = {
  { &hf_nbap_e_DCH_MACdFlow_ID_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_ID_LCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_MACdFlow_to_Delete_ItemLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_MACdFlow_to_Delete_ItemLCR, E_DCH_MACdFlow_to_Delete_ItemLCR_sequence);

  return offset;
}


static const per_sequence_t E_DCH_MACdFlows_to_DeleteLCR_sequence_of[1] = {
  { &hf_nbap_E_DCH_MACdFlows_to_DeleteLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_to_Delete_ItemLCR },
};

static int
dissect_nbap_E_DCH_MACdFlows_to_DeleteLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_DCH_MACdFlows_to_DeleteLCR, E_DCH_MACdFlows_to_DeleteLCR_sequence_of,
                                                  1, maxNrOfEDCHMACdFlowsLCR, FALSE);

  return offset;
}


static const value_string nbap_T321_vals[] = {
  {   0, "v100" },
  {   1, "v200" },
  {   2, "v400" },
  {   3, "v800" },
  { 0, NULL }
};


static int
dissect_nbap_T321(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_HS_DSCH_DRX_Cycle_FACH_vals[] = {
  {   0, "v4" },
  {   1, "v8" },
  {   2, "v16" },
  {   3, "v32" },
  { 0, NULL }
};


static int
dissect_nbap_HS_DSCH_DRX_Cycle_FACH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_HS_DSCH_RX_Burst_FACH_vals[] = {
  {   0, "v1" },
  {   1, "v2" },
  {   2, "v4" },
  {   3, "v8" },
  {   4, "v16" },
  { 0, NULL }
};


static int
dissect_nbap_HS_DSCH_RX_Burst_FACH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     5, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t Enhanced_UE_DRX_InformationLCR_sequence[] = {
  { &hf_nbap_t321           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T321 },
  { &hf_nbap_hS_DSCH_DRX_Cycle_FACH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCH_DRX_Cycle_FACH },
  { &hf_nbap_hS_DSCH_RX_Burst_FACH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCH_RX_Burst_FACH },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Enhanced_UE_DRX_InformationLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Enhanced_UE_DRX_InformationLCR, Enhanced_UE_DRX_InformationLCR_sequence);

  return offset;
}


static const value_string nbap_End_Of_Audit_Sequence_Indicator_vals[] = {
  {   0, "end-of-audit-sequence" },
  {   1, "not-end-of-audit-sequence" },
  { 0, NULL }
};


static int
dissect_nbap_End_Of_Audit_Sequence_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t E_RGCH_E_HICH_FDD_Code_List_sequence_of[1] = {
  { &hf_nbap_E_RGCH_E_HICH_FDD_Code_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_DL_ChannelisationCodeNumber },
};

static int
dissect_nbap_E_RGCH_E_HICH_FDD_Code_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_RGCH_E_HICH_FDD_Code_List, E_RGCH_E_HICH_FDD_Code_List_sequence_of,
                                                  1, maxNrOfE_RGCHs_E_HICHs, FALSE);

  return offset;
}


static const value_string nbap_E_RGCH_E_HICH_FDD_Code_Information_vals[] = {
  {   0, "replace" },
  {   1, "remove" },
  { 0, NULL }
};

static const per_choice_t E_RGCH_E_HICH_FDD_Code_Information_choice[] = {
  {   0, &hf_nbap_replace_01     , ASN1_EXTENSION_ROOT    , dissect_nbap_E_RGCH_E_HICH_FDD_Code_List },
  {   1, &hf_nbap_remove         , ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_E_RGCH_E_HICH_FDD_Code_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_E_RGCH_E_HICH_FDD_Code_Information, E_RGCH_E_HICH_FDD_Code_Information_choice,
                                 NULL);

  return offset;
}



static int
dissect_nbap_E_TFCI_BetaEC_Boost(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_UL_Delta_T2TP(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 6U, NULL, TRUE);

  return offset;
}


static const per_sequence_t E_TFCI_Boost_Information_sequence[] = {
  { &hf_nbap_e_TFCI_BetaEC_Boost, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_TFCI_BetaEC_Boost },
  { &hf_nbap_uL_Delta_T2TP  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_Delta_T2TP },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_TFCI_Boost_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_TFCI_Boost_Information, E_TFCI_Boost_Information_sequence);

  return offset;
}


static const per_sequence_t E_DCHProvidedBitRate_Item_sequence[] = {
  { &hf_nbap_schedulingPriorityIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SchedulingPriorityIndicator },
  { &hf_nbap_e_DCHProvidedBitRateValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCHProvidedBitRateValue },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCHProvidedBitRate_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCHProvidedBitRate_Item, E_DCHProvidedBitRate_Item_sequence);

  return offset;
}


static const per_sequence_t E_DCHProvidedBitRate_sequence_of[1] = {
  { &hf_nbap_E_DCHProvidedBitRate_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCHProvidedBitRate_Item },
};

static int
dissect_nbap_E_DCHProvidedBitRate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_DCHProvidedBitRate, E_DCHProvidedBitRate_sequence_of,
                                                  1, maxNrOfPriorityClasses, FALSE);

  return offset;
}


static const per_sequence_t E_DCHProvidedBitRateValueInformation_For_CellPortion_Item_sequence[] = {
  { &hf_nbap_cellPortionLCRID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellPortionLCRID },
  { &hf_nbap_e_DCHProvidedBitRateValue_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCHProvidedBitRate },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCHProvidedBitRateValueInformation_For_CellPortion_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCHProvidedBitRateValueInformation_For_CellPortion_Item, E_DCHProvidedBitRateValueInformation_For_CellPortion_Item_sequence);

  return offset;
}


static const per_sequence_t E_DCHProvidedBitRateValueInformation_For_CellPortion_sequence_of[1] = {
  { &hf_nbap_E_DCHProvidedBitRateValueInformation_For_CellPortion_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCHProvidedBitRateValueInformation_For_CellPortion_Item },
};

static int
dissect_nbap_E_DCHProvidedBitRateValueInformation_For_CellPortion(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_DCHProvidedBitRateValueInformation_For_CellPortion, E_DCHProvidedBitRateValueInformation_For_CellPortion_sequence_of,
                                                  1, maxNrOfCellPortionsPerCellLCR, FALSE);

  return offset;
}



static int
dissect_nbap_E_AGCH_PowerOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_E_RGCH_PowerOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_E_HICH_PowerOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_E_HICH_TimeOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            4U, 44U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_E_HICH_TimeOffsetLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            4U, 15U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_N_E_UCCH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 12U, NULL, FALSE);

  return offset;
}


static const per_sequence_t E_PUCH_Information_sequence[] = {
  { &hf_nbap_minCR          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CodeRate },
  { &hf_nbap_maxCR          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CodeRate },
  { &hf_nbap_harqInfo       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HARQ_Info_for_E_DCH },
  { &hf_nbap_n_E_UCCH       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_N_E_UCCH },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_PUCH_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_PUCH_Information, E_PUCH_Information_sequence);

  return offset;
}


static const value_string nbap_E_DCH_Grant_TypeTDD_vals[] = {
  {   0, "scheduled" },
  {   1, "non-scheduled" },
  { 0, NULL }
};


static int
dissect_nbap_E_DCH_Grant_TypeTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t E_DCH_MACdFlow_InfoTDDItem_sequence[] = {
  { &hf_nbap_e_DCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_ID },
  { &hf_nbap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AllocationRetentionPriority },
  { &hf_nbap_tnlQos         , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TnlQos },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_payloadCRC_PresenceIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PayloadCRC_PresenceIndicator },
  { &hf_nbap_maximum_Number_of_Retransmissions_For_E_DCH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Maximum_Number_of_Retransmissions_For_E_DCH },
  { &hf_nbap_eDCH_HARQ_PO_TDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_HARQ_PO_TDD },
  { &hf_nbap_eDCH_MACdFlow_Multiplexing_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_MACdFlow_Multiplexing_List },
  { &hf_nbap_eDCH_Grant_TypeTDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_Grant_TypeTDD },
  { &hf_nbap_eDCHLogicalChannelInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_LogicalChannelInformation },
  { &hf_nbap_eDCH_MACdFlow_Retransmission_Timer, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_MACdFlow_Retransmission_Timer },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_MACdFlow_InfoTDDItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_MACdFlow_InfoTDDItem, E_DCH_MACdFlow_InfoTDDItem_sequence);

  return offset;
}


static const per_sequence_t E_DCH_MACdFlows_Information_TDD_sequence_of[1] = {
  { &hf_nbap_E_DCH_MACdFlows_Information_TDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_InfoTDDItem },
};

static int
dissect_nbap_E_DCH_MACdFlows_Information_TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_DCH_MACdFlows_Information_TDD, E_DCH_MACdFlows_Information_TDD_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}



static int
dissect_nbap_E_DCH_TimeslotResource(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     13, 13, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_E_DCH_PowerResource(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 32U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_TddE_PUCH_Offset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_Non_Scheduled_Grant_Info_sequence[] = {
  { &hf_nbap_timeslotResource, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_TimeslotResource },
  { &hf_nbap_powerResource  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_PowerResource },
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tddE_PUCH_Offset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TddE_PUCH_Offset },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_Non_Scheduled_Grant_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_Non_Scheduled_Grant_Info, E_DCH_Non_Scheduled_Grant_Info_sequence);

  return offset;
}



static int
dissect_nbap_E_DCH_TDD_Maximum_Bitrate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 9201U, NULL, TRUE);

  return offset;
}


static const per_sequence_t E_DCH_TDD_Information_sequence[] = {
  { &hf_nbap_e_DCH_TDD_Maximum_Bitrate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_TDD_Maximum_Bitrate },
  { &hf_nbap_e_DCH_Processing_Overload_Level, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_Processing_Overload_Level },
  { &hf_nbap_e_DCH_PowerOffset_for_SchedulingInfo, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_PowerOffset_for_SchedulingInfo },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_TDD_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_TDD_Information, E_DCH_TDD_Information_sequence);

  return offset;
}


static const per_sequence_t E_DCH_Information_sequence[] = {
  { &hf_nbap_e_PUCH_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_PUCH_Information },
  { &hf_nbap_e_TFCS_Information_TDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_TFCS_Information_TDD },
  { &hf_nbap_e_DCH_MACdFlows_Information_TDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlows_Information_TDD },
  { &hf_nbap_e_DCH_Non_Scheduled_Grant_Info, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_Non_Scheduled_Grant_Info },
  { &hf_nbap_e_DCH_TDD_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_TDD_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_Information, E_DCH_Information_sequence);

  return offset;
}



static int
dissect_nbap_E_DCH_TimeslotResourceLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     5, 5, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t E_DCH_TDD_MACdFlow_Specific_InformationResp_Item_sequence[] = {
  { &hf_nbap_e_DCH_MacdFlow_Id, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_ID },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_TDD_MACdFlow_Specific_InformationResp_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_TDD_MACdFlow_Specific_InformationResp_Item, E_DCH_TDD_MACdFlow_Specific_InformationResp_Item_sequence);

  return offset;
}


static const per_sequence_t E_DCH_TDD_MACdFlow_Specific_InformationResp_sequence_of[1] = {
  { &hf_nbap_E_DCH_TDD_MACdFlow_Specific_InformationResp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_TDD_MACdFlow_Specific_InformationResp_Item },
};

static int
dissect_nbap_E_DCH_TDD_MACdFlow_Specific_InformationResp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_DCH_TDD_MACdFlow_Specific_InformationResp, E_DCH_TDD_MACdFlow_Specific_InformationResp_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}


static const per_sequence_t E_AGCH_Specific_InformationResp_ItemTDD_sequence[] = {
  { &hf_nbap_e_AGCH_Id      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_AGCH_Id },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_AGCH_Specific_InformationResp_ItemTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_AGCH_Specific_InformationResp_ItemTDD, E_AGCH_Specific_InformationResp_ItemTDD_sequence);

  return offset;
}


static const per_sequence_t E_AGCH_Specific_InformationRespListTDD_sequence_of[1] = {
  { &hf_nbap_E_AGCH_Specific_InformationRespListTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_AGCH_Specific_InformationResp_ItemTDD },
};

static int
dissect_nbap_E_AGCH_Specific_InformationRespListTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_AGCH_Specific_InformationRespListTDD, E_AGCH_Specific_InformationRespListTDD_sequence_of,
                                                  1, maxNrOfEAGCHCodes, FALSE);

  return offset;
}


static const per_sequence_t Scheduled_E_HICH_Specific_InformationItem_ResponseLCRTDD_sequence[] = {
  { &hf_nbap_eI             , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_EI },
  { &hf_nbap_e_HICH_ID_TDD  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_HICH_ID_TDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Scheduled_E_HICH_Specific_InformationItem_ResponseLCRTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Scheduled_E_HICH_Specific_InformationItem_ResponseLCRTDD, Scheduled_E_HICH_Specific_InformationItem_ResponseLCRTDD_sequence);

  return offset;
}


static const per_sequence_t Scheduled_E_HICH_Specific_Information_ResponseLCRTDD_sequence_of[1] = {
  { &hf_nbap_Scheduled_E_HICH_Specific_Information_ResponseLCRTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Scheduled_E_HICH_Specific_InformationItem_ResponseLCRTDD },
};

static int
dissect_nbap_Scheduled_E_HICH_Specific_Information_ResponseLCRTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Scheduled_E_HICH_Specific_Information_ResponseLCRTDD, Scheduled_E_HICH_Specific_Information_ResponseLCRTDD_sequence_of,
                                                  1, maxNrOfEHICHCodes, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_Information_Response_sequence[] = {
  { &hf_nbap_e_DCH_TDD_MACdFlow_Specific_InformationResp, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_TDD_MACdFlow_Specific_InformationResp },
  { &hf_nbap_e_AGCH_Specific_Information_ResponseTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_AGCH_Specific_InformationRespListTDD },
  { &hf_nbap_e_RNTI         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_RNTI },
  { &hf_nbap_scheduled_E_HICH_Specific_InformationResp, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Scheduled_E_HICH_Specific_Information_ResponseLCRTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_Information_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_Information_Response, E_DCH_Information_Response_sequence);

  return offset;
}


static const value_string nbap_E_HICH_Type_vals[] = {
  {   0, "scheduled" },
  {   1, "non-scheduled" },
  { 0, NULL }
};


static int
dissect_nbap_E_HICH_Type(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t E_DCH_MACdFlow_ModifyTDDItem_sequence[] = {
  { &hf_nbap_e_DCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_ID },
  { &hf_nbap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_AllocationRetentionPriority },
  { &hf_nbap_transportBearerRequestIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportBearerRequestIndicator },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_tnlQos         , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TnlQos },
  { &hf_nbap_maximum_Number_of_Retransmissions_For_E_DCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Maximum_Number_of_Retransmissions_For_E_DCH },
  { &hf_nbap_eDCH_HARQ_PO_TDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_HARQ_PO_TDD },
  { &hf_nbap_eDCH_MACdFlow_Multiplexing_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_MACdFlow_Multiplexing_List },
  { &hf_nbap_eDCH_Grant_TypeTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_Grant_TypeTDD },
  { &hf_nbap_e_DCH_LogicalChannelToAdd, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_LogicalChannelInformation },
  { &hf_nbap_e_DCH_LogicalChannelToModify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_LogicalChannelToModify },
  { &hf_nbap_e_DCH_LogicalChannelToDelete, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_LogicalChannelToDelete },
  { &hf_nbap_eDCH_MACdFlow_Retransmission_Timer, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_MACdFlow_Retransmission_Timer },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_MACdFlow_ModifyTDDItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_MACdFlow_ModifyTDDItem, E_DCH_MACdFlow_ModifyTDDItem_sequence);

  return offset;
}


static const per_sequence_t E_DCH_TDD_Information_to_Modify_List_sequence_of[1] = {
  { &hf_nbap_E_DCH_TDD_Information_to_Modify_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_ModifyTDDItem },
};

static int
dissect_nbap_E_DCH_TDD_Information_to_Modify_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_DCH_TDD_Information_to_Modify_List, E_DCH_TDD_Information_to_Modify_List_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_TDD_Information_to_Modify_sequence[] = {
  { &hf_nbap_e_DCH_TDD_Information_to_Modify_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_TDD_Information_to_Modify_List },
  { &hf_nbap_mACeReset_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MACeReset_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_TDD_Information_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_TDD_Information_to_Modify, E_DCH_TDD_Information_to_Modify_sequence);

  return offset;
}


static const per_sequence_t E_DCH_Information_Reconfig_sequence[] = {
  { &hf_nbap_e_PUCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_PUCH_Information },
  { &hf_nbap_e_TFCS_Information_TDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_TFCS_Information_TDD },
  { &hf_nbap_e_DCH_MACdFlows_to_Add, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_MACdFlows_Information_TDD },
  { &hf_nbap_e_DCH_MACdFlows_to_Delete, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_MACdFlows_to_Delete },
  { &hf_nbap_e_DCH_Non_Scheduled_Grant_Info, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_Non_Scheduled_Grant_Info },
  { &hf_nbap_e_DCH_TDD_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_TDD_Information },
  { &hf_nbap_e_DCH_TDD_Information_to_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_TDD_Information_to_Modify },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_Information_Reconfig(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_Information_Reconfig, E_DCH_Information_Reconfig_sequence);

  return offset;
}



static int
dissect_nbap_Maximum_Generated_ReceivedTotalWideBandPowerInOtherCells(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 621U, NULL, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_Non_Scheduled_Grant_Info768_sequence[] = {
  { &hf_nbap_timeslotResource, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_TimeslotResource },
  { &hf_nbap_powerResource  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_PowerResource },
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tddE_PUCH_Offset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TddE_PUCH_Offset },
  { &hf_nbap_tdd_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode768 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_Non_Scheduled_Grant_Info768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_Non_Scheduled_Grant_Info768, E_DCH_Non_Scheduled_Grant_Info768_sequence);

  return offset;
}



static int
dissect_nbap_E_DCH_TDD_Maximum_Bitrate768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 17713U, NULL, TRUE);

  return offset;
}


static const per_sequence_t E_DCH_TDD_Information768_sequence[] = {
  { &hf_nbap_e_DCH_TDD_Maximum_Bitrate768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_TDD_Maximum_Bitrate768 },
  { &hf_nbap_e_DCH_Processing_Overload_Level, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_Processing_Overload_Level },
  { &hf_nbap_e_DCH_PowerOffset_for_SchedulingInfo, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_PowerOffset_for_SchedulingInfo },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_TDD_Information768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_TDD_Information768, E_DCH_TDD_Information768_sequence);

  return offset;
}


static const per_sequence_t E_DCH_768_Information_sequence[] = {
  { &hf_nbap_e_PUCH_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_PUCH_Information },
  { &hf_nbap_e_TFCS_Information_TDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_TFCS_Information_TDD },
  { &hf_nbap_e_DCH_MACdFlows_Information_TDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlows_Information_TDD },
  { &hf_nbap_e_DCH_Non_Scheduled_Grant_Info768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_Non_Scheduled_Grant_Info768 },
  { &hf_nbap_e_DCH_TDD_Information768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_TDD_Information768 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_768_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_768_Information, E_DCH_768_Information_sequence);

  return offset;
}


static const per_sequence_t E_DCH_768_Information_Reconfig_sequence[] = {
  { &hf_nbap_e_PUCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_PUCH_Information },
  { &hf_nbap_e_TFCS_Information_TDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_TFCS_Information_TDD },
  { &hf_nbap_e_DCH_MACdFlows_to_Add, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_MACdFlows_Information_TDD },
  { &hf_nbap_e_DCH_MACdFlows_to_Delete, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_MACdFlows_to_Delete },
  { &hf_nbap_e_DCH_Non_Scheduled_Grant_Info768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_Non_Scheduled_Grant_Info768 },
  { &hf_nbap_e_DCH_TDD_Information768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_TDD_Information768 },
  { &hf_nbap_e_DCH_TDD_Information_to_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_TDD_Information_to_Modify },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_768_Information_Reconfig(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_768_Information_Reconfig, E_DCH_768_Information_Reconfig_sequence);

  return offset;
}


static const per_sequence_t E_PUCH_LCR_Information_sequence[] = {
  { &hf_nbap_minCR          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CodeRate },
  { &hf_nbap_maxCR          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CodeRate },
  { &hf_nbap_harqInfo       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HARQ_Info_for_E_DCH },
  { &hf_nbap_pRXdes_base    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PRXdes_base },
  { &hf_nbap_e_PUCH_TPC_StepSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_TPC_UplinkStepSize_LCR },
  { &hf_nbap_e_AGCH_TPC_StepSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_TPC_DownlinkStepSize },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_PUCH_LCR_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_PUCH_LCR_Information, E_PUCH_LCR_Information_sequence);

  return offset;
}


static const value_string nbap_T_subframeNumber_vals[] = {
  {   0, "v0" },
  {   1, "v1" },
  { 0, NULL }
};


static int
dissect_nbap_T_subframeNumber(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_N_E_UCCHLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 8U, NULL, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_Non_Scheduled_Grant_LCR_Info_sequence[] = {
  { &hf_nbap_timeslotResourceLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_TimeslotResourceLCR },
  { &hf_nbap_powerResource  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_PowerResource },
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_subframeNumber , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_subframeNumber },
  { &hf_nbap_tddE_PUCH_Offset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TddE_PUCH_Offset },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_n_E_UCCHLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_N_E_UCCHLCR },
  { &hf_nbap_e_HICH_LCR_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_HICH_LCR_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_Non_Scheduled_Grant_LCR_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_Non_Scheduled_Grant_LCR_Info, E_DCH_Non_Scheduled_Grant_LCR_Info_sequence);

  return offset;
}



static int
dissect_nbap_E_DCH_LCRTDD_PhysicalLayerCategory(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 5U, NULL, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_LCRTDD_Information_sequence[] = {
  { &hf_nbap_e_DCH_LCRTDD_PhysicalLayerCategory, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_LCRTDD_PhysicalLayerCategory },
  { &hf_nbap_e_DCH_Processing_Overload_Level, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_Processing_Overload_Level },
  { &hf_nbap_e_DCH_PowerOffset_for_SchedulingInfo, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_PowerOffset_for_SchedulingInfo },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_LCRTDD_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_LCRTDD_Information, E_DCH_LCRTDD_Information_sequence);

  return offset;
}


static const per_sequence_t E_DCH_LCR_Information_sequence[] = {
  { &hf_nbap_e_PUCH_LCR_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_PUCH_LCR_Information },
  { &hf_nbap_e_TFCS_Information_TDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_TFCS_Information_TDD },
  { &hf_nbap_e_DCH_MACdFlows_Information_TDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlows_Information_TDD },
  { &hf_nbap_e_DCH_Non_Scheduled_Grant_LCR_Info, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_Non_Scheduled_Grant_LCR_Info },
  { &hf_nbap_e_DCH_LCRTDD_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_LCRTDD_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_LCR_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_LCR_Information, E_DCH_LCR_Information_sequence);

  return offset;
}


static const per_sequence_t E_DCH_LCR_Information_Reconfig_sequence[] = {
  { &hf_nbap_e_PUCH_LCR_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_PUCH_LCR_Information },
  { &hf_nbap_e_TFCS_Information_TDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_TFCS_Information_TDD },
  { &hf_nbap_e_DCH_MACdFlows_to_Add, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_MACdFlows_Information_TDD },
  { &hf_nbap_e_DCH_MACdFlows_to_Delete, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_MACdFlows_to_Delete },
  { &hf_nbap_e_DCH_Non_Scheduled_Grant_LCR_Info, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_Non_Scheduled_Grant_LCR_Info },
  { &hf_nbap_e_DCH_LCRTDD_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_LCRTDD_Information },
  { &hf_nbap_e_DCH_TDD_Information_to_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_TDD_Information_to_Modify },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_LCR_Information_Reconfig(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_LCR_Information_Reconfig, E_DCH_LCR_Information_Reconfig_sequence);

  return offset;
}


static const value_string nbap_DormantModeIndicator_vals[] = {
  {   0, "enterDormantMode" },
  {   1, "leaveDormantMode" },
  { 0, NULL }
};


static int
dissect_nbap_DormantModeIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_Enhanced_FACH_Capability_vals[] = {
  {   0, "enhanced-FACH-capable" },
  {   1, "enhanced-FACH-non-capable" },
  { 0, NULL }
};


static int
dissect_nbap_Enhanced_FACH_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_EnhancedHSServingCC_Abort_vals[] = {
  {   0, "abortEnhancedHSServingCC" },
  { 0, NULL }
};


static int
dissect_nbap_EnhancedHSServingCC_Abort(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_Enhanced_PCH_Capability_vals[] = {
  {   0, "enhanced-PCH-capable" },
  {   1, "enhanced-PCH-non-capable" },
  { 0, NULL }
};


static int
dissect_nbap_Enhanced_PCH_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_Enhanced_UE_DRX_Capability_vals[] = {
  {   0, "enhanced-UE-DRX-capable" },
  {   1, "enhanced-UE-DRX-non-capable" },
  { 0, NULL }
};


static int
dissect_nbap_Enhanced_UE_DRX_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t Enhanced_UE_DRX_InformationFDD_sequence[] = {
  { &hf_nbap_t321           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T321 },
  { &hf_nbap_hS_DSCH_DRX_Cycle_FACH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCH_DRX_Cycle_FACH },
  { &hf_nbap_hS_DSCH_RX_Burst_FACH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCH_RX_Burst_FACH },
  { &hf_nbap_dRX_Interruption_by_HS_DSCH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DRX_Interruption_by_HS_DSCH },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Enhanced_UE_DRX_InformationFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Enhanced_UE_DRX_InformationFDD, Enhanced_UE_DRX_InformationFDD_sequence);

  return offset;
}



static int
dissect_nbap_Extended_E_DCH_LCRTDD_PhysicalLayerCategory(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            6U, 6U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_Ext_Max_Bits_MACe_PDU_non_scheduled(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            19983U, 22978U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_Ext_Reference_E_TFCI_PO(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            30U, 31U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_ExtendedPropagationDelay(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            255U, 1023U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_Extended_RNC_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            4096U, 65535U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_Extended_Round_Trip_Time_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            32767U, 103041U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_Extended_HS_SCCH_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            32U, 255U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_Extended_HS_SICH_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            32U, 255U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_Extended_E_HICH_ID_TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            32U, 255U, NULL, FALSE);

  return offset;
}


static const value_string nbap_T_subframeNumber_01_vals[] = {
  {   0, "v0" },
  {   1, "v1" },
  { 0, NULL }
};


static int
dissect_nbap_T_subframeNumber_01(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t E_DCH_Semi_PersistentScheduling_Information_ResponseLCR_sequence[] = {
  { &hf_nbap_timeslot_Resource_Related_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_TimeslotResourceLCR },
  { &hf_nbap_powerResource  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_PowerResource },
  { &hf_nbap_repetition_Period_List_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Repetition_Period_List_LCR },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_subframeNumber_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_subframeNumber_01 },
  { &hf_nbap_tddE_PUCH_Offset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TddE_PUCH_Offset },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_n_E_UCCHLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_N_E_UCCHLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR, E_DCH_Semi_PersistentScheduling_Information_ResponseLCR_sequence);

  return offset;
}



static int
dissect_nbap_FACH_Measurement_Occasion_Cycle_Length_Coefficient(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 12U, NULL, FALSE);

  return offset;
}


static const value_string nbap_Fast_Reconfiguration_Mode_vals[] = {
  {   0, "fast" },
  { 0, NULL }
};


static int
dissect_nbap_Fast_Reconfiguration_Mode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_Fast_Reconfiguration_Permission_vals[] = {
  {   0, "allowed" },
  { 0, NULL }
};


static int
dissect_nbap_Fast_Reconfiguration_Permission(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_FDD_S_CCPCH_FrameOffset_vals[] = {
  {   0, "v1" },
  {   1, "v2" },
  {   2, "v4" },
  { 0, NULL }
};


static int
dissect_nbap_FDD_S_CCPCH_FrameOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_FDD_S_CCPCH_Offset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 149U, NULL, FALSE);

  return offset;
}


static const value_string nbap_F_DPCH_Capability_vals[] = {
  {   0, "f-DPCH-capable" },
  {   1, "f-DPCH-non-capable" },
  { 0, NULL }
};


static int
dissect_nbap_F_DPCH_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_F_DPCH_SlotFormatCapability_vals[] = {
  {   0, "f-DPCH-slot-format-capable" },
  {   1, "f-DPCH-slot-format-non-capable" },
  { 0, NULL }
};


static int
dissect_nbap_F_DPCH_SlotFormatCapability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_FNReportingIndicator_vals[] = {
  {   0, "fN-reporting-required" },
  {   1, "fN-reporting-not-required" },
  { 0, NULL }
};


static int
dissect_nbap_FNReportingIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_FrameAdjustmentValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 4095U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_FrameOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_FPACH_Power(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -150, 400U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_16(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     16, 16, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_22(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     22, 22, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t GANSS_NAVclockModel_sequence[] = {
  { &hf_nbap_navToc         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_navaf2         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_navaf1         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_navaf0         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_22 },
  { &hf_nbap_navTgd         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_NAVclockModel(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_NAVclockModel, GANSS_NAVclockModel_sequence);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_11(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     11, 11, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_5(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     5, 5, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_3(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     3, 3, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_20(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     20, 20, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_26(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     26, 26, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_13(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     13, 13, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t GANSS_CNAVclockModel_sequence[] = {
  { &hf_nbap_cnavToc        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_11 },
  { &hf_nbap_cnavTop        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_11 },
  { &hf_nbap_cnavURA0       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_5 },
  { &hf_nbap_cnavURA1       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_3 },
  { &hf_nbap_cnavURA2       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_3 },
  { &hf_nbap_cnavAf2        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_10 },
  { &hf_nbap_cnavAf1        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_20 },
  { &hf_nbap_cnavAf0        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_26 },
  { &hf_nbap_cnavTgd        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_13 },
  { &hf_nbap_cnavISCl1cp    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BIT_STRING_SIZE_13 },
  { &hf_nbap_cnavISCl1cd    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BIT_STRING_SIZE_13 },
  { &hf_nbap_cnavISCl1ca    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BIT_STRING_SIZE_13 },
  { &hf_nbap_cnavISCl2c     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BIT_STRING_SIZE_13 },
  { &hf_nbap_cnavISCl5i5    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BIT_STRING_SIZE_13 },
  { &hf_nbap_cnavISCl5q5    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BIT_STRING_SIZE_13 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_CNAVclockModel(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_CNAVclockModel, GANSS_CNAVclockModel_sequence);

  return offset;
}


static const per_sequence_t GANSS_GLONASSclockModel_sequence[] = {
  { &hf_nbap_gloTau         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_22 },
  { &hf_nbap_gloGamma       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_11 },
  { &hf_nbap_gloDeltaTau    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BIT_STRING_SIZE_5 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_GLONASSclockModel(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_GLONASSclockModel, GANSS_GLONASSclockModel_sequence);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_12(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     12, 12, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t GANSS_SBASclockModel_sequence[] = {
  { &hf_nbap_sbasTo         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_13 },
  { &hf_nbap_sbasAgfo       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_12 },
  { &hf_nbap_sbasAgf1       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_SBASclockModel(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_SBASclockModel, GANSS_SBASclockModel_sequence);

  return offset;
}


static const value_string nbap_GANSS_AddClockModels_vals[] = {
  {   0, "navClockModel" },
  {   1, "cnavClockModel" },
  {   2, "glonassClockModel" },
  {   3, "sbasClockModel" },
  { 0, NULL }
};

static const per_choice_t GANSS_AddClockModels_choice[] = {
  {   0, &hf_nbap_navClockModel  , ASN1_EXTENSION_ROOT    , dissect_nbap_GANSS_NAVclockModel },
  {   1, &hf_nbap_cnavClockModel , ASN1_EXTENSION_ROOT    , dissect_nbap_GANSS_CNAVclockModel },
  {   2, &hf_nbap_glonassClockModel, ASN1_EXTENSION_ROOT    , dissect_nbap_GANSS_GLONASSclockModel },
  {   3, &hf_nbap_sbasClockModel , ASN1_EXTENSION_ROOT    , dissect_nbap_GANSS_SBASclockModel },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_GANSS_AddClockModels(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_GANSS_AddClockModels, GANSS_AddClockModels_choice,
                                 NULL);

  return offset;
}



static int
dissect_nbap_GANSS_AddIonoModelReq(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     2, 2, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_GANSS_AddNavigationModelsReq(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_boolean(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_4(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     4, 4, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_1(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     1, 1, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_32(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     32, 32, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_24(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     24, 24, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_14(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     14, 14, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t GANSS_NavModel_NAVKeplerianSet_sequence[] = {
  { &hf_nbap_navURA         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_4 },
  { &hf_nbap_navFitFlag     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_1 },
  { &hf_nbap_navToe         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_navOmega       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_32 },
  { &hf_nbap_navDeltaN      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_navM0          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_32 },
  { &hf_nbap_navOmegaADot   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_24 },
  { &hf_nbap_navE           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_32 },
  { &hf_nbap_navIDot        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_14 },
  { &hf_nbap_navAPowerHalf  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_32 },
  { &hf_nbap_navI0          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_32 },
  { &hf_nbap_navOmegaA0     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_32 },
  { &hf_nbap_navCrs         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_navCis         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_navCus         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_navCrc         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_navCic         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_navCuc         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_NavModel_NAVKeplerianSet(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_NavModel_NAVKeplerianSet, GANSS_NavModel_NAVKeplerianSet_sequence);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_25(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     25, 25, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_17(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     17, 17, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_23(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     23, 23, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_33(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     33, 33, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_15(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     15, 15, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_21(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     21, 21, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t GANSS_NavModel_CNAVKeplerianSet_sequence[] = {
  { &hf_nbap_cnavTop        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_11 },
  { &hf_nbap_cnavURAindex   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_5 },
  { &hf_nbap_cnavDeltaA     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_26 },
  { &hf_nbap_cnavAdot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_25 },
  { &hf_nbap_cnavDeltaNo    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_17 },
  { &hf_nbap_cnavDeltaNoDot , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_23 },
  { &hf_nbap_cnavMo         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_33 },
  { &hf_nbap_cnavE          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_33 },
  { &hf_nbap_cnavOmega      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_33 },
  { &hf_nbap_cnavOMEGA0     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_33 },
  { &hf_nbap_cnavDeltaOmegaDot, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_17 },
  { &hf_nbap_cnavIo         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_33 },
  { &hf_nbap_cnavIoDot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_15 },
  { &hf_nbap_cnavCis        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_cnavCic        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_cnavCrs        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_24 },
  { &hf_nbap_cnavCrc        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_24 },
  { &hf_nbap_cnavCus        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_21 },
  { &hf_nbap_cnavCuc        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_21 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_NavModel_CNAVKeplerianSet(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_NavModel_CNAVKeplerianSet, GANSS_NavModel_CNAVKeplerianSet_sequence);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_2(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     2, 2, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_27(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     27, 27, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t GANSS_NavModel_GLONASSecef_sequence[] = {
  { &hf_nbap_gloEn          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_5 },
  { &hf_nbap_gloP1          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_2 },
  { &hf_nbap_gloP2          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_1 },
  { &hf_nbap_gloM           , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BIT_STRING_SIZE_2 },
  { &hf_nbap_gloX           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_27 },
  { &hf_nbap_gloXdot        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_24 },
  { &hf_nbap_gloXdotdot     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_5 },
  { &hf_nbap_gloY           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_27 },
  { &hf_nbap_gloYdot        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_24 },
  { &hf_nbap_gloYdotdot     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_5 },
  { &hf_nbap_gloZ           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_27 },
  { &hf_nbap_gloZdot        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_24 },
  { &hf_nbap_gloZdotdot     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_5 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_NavModel_GLONASSecef(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_NavModel_GLONASSecef, GANSS_NavModel_GLONASSecef_sequence);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_30(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     30, 30, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_18(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     18, 18, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t GANSS_NavModel_SBASecef_sequence[] = {
  { &hf_nbap_sbasTo         , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BIT_STRING_SIZE_13 },
  { &hf_nbap_sbasAccuracy   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_4 },
  { &hf_nbap_sbasXg         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_30 },
  { &hf_nbap_sbasYg         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_30 },
  { &hf_nbap_sbasZg         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_25 },
  { &hf_nbap_sbasXgDot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_17 },
  { &hf_nbap_sbasYgDot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_17 },
  { &hf_nbap_sbasZgDot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_18 },
  { &hf_nbap_sbasXgDotDot   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_10 },
  { &hf_nbap_sbagYgDotDot   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_10 },
  { &hf_nbap_sbasZgDotDot   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_10 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_NavModel_SBASecef(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_NavModel_SBASecef, GANSS_NavModel_SBASecef_sequence);

  return offset;
}


static const value_string nbap_GANSS_AddOrbitModels_vals[] = {
  {   0, "navKeplerianSet" },
  {   1, "cnavKeplerianSet" },
  {   2, "glonassECEF" },
  {   3, "sbasECEF" },
  { 0, NULL }
};

static const per_choice_t GANSS_AddOrbitModels_choice[] = {
  {   0, &hf_nbap_navKeplerianSet, ASN1_EXTENSION_ROOT    , dissect_nbap_GANSS_NavModel_NAVKeplerianSet },
  {   1, &hf_nbap_cnavKeplerianSet, ASN1_EXTENSION_ROOT    , dissect_nbap_GANSS_NavModel_CNAVKeplerianSet },
  {   2, &hf_nbap_glonassECEF    , ASN1_EXTENSION_ROOT    , dissect_nbap_GANSS_NavModel_GLONASSecef },
  {   3, &hf_nbap_sbasECEF       , ASN1_EXTENSION_ROOT    , dissect_nbap_GANSS_NavModel_SBASecef },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_GANSS_AddOrbitModels(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_GANSS_AddOrbitModels, GANSS_AddOrbitModels_choice,
                                 NULL);

  return offset;
}



static int
dissect_nbap_GANSS_AddUTCModelsReq(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_boolean(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}


static const per_sequence_t GPS_Ionospheric_Model_sequence[] = {
  { &hf_nbap_alpha_zero_ionos_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_alpha_one_ionos_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_alpha_two_ionos_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_alpha_three_ionos, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_beta_zero_ionos, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_beta_one_ionos , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_beta_two_ionos , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_beta_three_ionos, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GPS_Ionospheric_Model(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GPS_Ionospheric_Model, GPS_Ionospheric_Model_sequence);

  return offset;
}


static const per_sequence_t GANSS_Additional_Ionospheric_Model_sequence[] = {
  { &hf_nbap_dataID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_2 },
  { &hf_nbap_alpha_beta_parameters, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GPS_Ionospheric_Model },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_Additional_Ionospheric_Model(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_Additional_Ionospheric_Model, GANSS_Additional_Ionospheric_Model_sequence);

  return offset;
}



static int
dissect_nbap_INTEGER_0_8191(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 8191U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_INTEGER_0_86399(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 86399U, NULL, FALSE);

  return offset;
}


static const per_sequence_t GANSS_Transmission_Time_sequence[] = {
  { &hf_nbap_ganssDay       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_INTEGER_0_8191 },
  { &hf_nbap_ganssTod_01    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_86399 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_Transmission_Time(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_Transmission_Time, GANSS_Transmission_Time_sequence);

  return offset;
}


static const value_string nbap_T_non_broadcastIndication_vals[] = {
  {   0, "true" },
  { 0, NULL }
};


static int
dissect_nbap_T_non_broadcastIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_6(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     6, 6, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t Ganss_Sat_Info_AddNavList_item_sequence[] = {
  { &hf_nbap_satId          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_63 },
  { &hf_nbap_svHealth       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_6 },
  { &hf_nbap_iod            , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_11 },
  { &hf_nbap_ganssAddClockModels, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_AddClockModels },
  { &hf_nbap_ganssAddOrbitModels, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_AddOrbitModels },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Ganss_Sat_Info_AddNavList_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Ganss_Sat_Info_AddNavList_item, Ganss_Sat_Info_AddNavList_item_sequence);

  return offset;
}


static const per_sequence_t Ganss_Sat_Info_AddNavList_sequence_of[1] = {
  { &hf_nbap_Ganss_Sat_Info_AddNavList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Ganss_Sat_Info_AddNavList_item },
};

static int
dissect_nbap_Ganss_Sat_Info_AddNavList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Ganss_Sat_Info_AddNavList, Ganss_Sat_Info_AddNavList_sequence_of,
                                                  1, maxGANSSSat, FALSE);

  return offset;
}


static const per_sequence_t GANSS_Additional_Navigation_Models_sequence[] = {
  { &hf_nbap_ganss_Transmission_Time, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_Transmission_Time },
  { &hf_nbap_non_broadcastIndication, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_T_non_broadcastIndication },
  { &hf_nbap_ganssSatInfoNavList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Ganss_Sat_Info_AddNavList },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_Additional_Navigation_Models(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_Additional_Navigation_Models, GANSS_Additional_Navigation_Models_sequence);

  return offset;
}



static int
dissect_nbap_INTEGER_0_37799(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 37799U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_INTEGER_M2147483648_2147483647(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            G_MININT32, 2147483647U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_INTEGER_M8388608_8388607(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -8388608, 8388607U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_INTEGER_M64_63(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -64, 63U, NULL, FALSE);

  return offset;
}


static const value_string nbap_T_gnss_to_id_vals[] = {
  {   0, "gps" },
  {   1, "galileo" },
  {   2, "qzss" },
  {   3, "glonass" },
  { 0, NULL }
};


static int
dissect_nbap_T_gnss_to_id(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 3, NULL);

  return offset;
}


static const per_sequence_t GANSS_Time_Model_sequence[] = {
  { &hf_nbap_ganss_time_model_Ref_Time, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_37799 },
  { &hf_nbap_ganss_t_a0     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_M2147483648_2147483647 },
  { &hf_nbap_ganss_t_a1     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_INTEGER_M8388608_8388607 },
  { &hf_nbap_ganss_t_a2     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_INTEGER_M64_63 },
  { &hf_nbap_gnss_to_id     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_gnss_to_id },
  { &hf_nbap_ganss_wk_number_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_INTEGER_0_8191 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_Time_Model(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_Time_Model, GANSS_Time_Model_sequence);

  return offset;
}


static const per_sequence_t GANSS_Additional_Time_Models_sequence_of[1] = {
  { &hf_nbap_GANSS_Additional_Time_Models_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_Time_Model },
};

static int
dissect_nbap_GANSS_Additional_Time_Models(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_GANSS_Additional_Time_Models, GANSS_Additional_Time_Models_sequence_of,
                                                  1, maxGANSS_1, FALSE);

  return offset;
}


static const per_sequence_t GANSS_UTCmodelSet1_sequence[] = {
  { &hf_nbap_utcA0          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_utcA1          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_13 },
  { &hf_nbap_utcA2          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_7 },
  { &hf_nbap_utcDeltaTls    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_utcTot         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_utcWNot        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_13 },
  { &hf_nbap_utcWNlsf       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_utcDN          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_4 },
  { &hf_nbap_utcDeltaTlsf   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_UTCmodelSet1(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_UTCmodelSet1, GANSS_UTCmodelSet1_sequence);

  return offset;
}


static const per_sequence_t GANSS_DeltaUT1_sequence[] = {
  { &hf_nbap_b1             , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_11 },
  { &hf_nbap_b2             , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_10 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_DeltaUT1(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_DeltaUT1, GANSS_DeltaUT1_sequence);

  return offset;
}


static const per_sequence_t GANSS_UTCmodelSet2_sequence[] = {
  { &hf_nbap_nA             , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_11 },
  { &hf_nbap_tauC           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_32 },
  { &hf_nbap_deltaUT1_01    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GANSS_DeltaUT1 },
  { &hf_nbap_kp             , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BIT_STRING_SIZE_2 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_UTCmodelSet2(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_UTCmodelSet2, GANSS_UTCmodelSet2_sequence);

  return offset;
}


static const per_sequence_t GANSS_UTCmodelSet3_sequence[] = {
  { &hf_nbap_utcA1wnt       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_24 },
  { &hf_nbap_utcA0wnt       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_32 },
  { &hf_nbap_utcTot_01      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_utcWNt         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_utcDeltaTls    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_utcWNlsf       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_utcDN_01       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_utcDeltaTlsf   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_utcStandardID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_3 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_UTCmodelSet3(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_UTCmodelSet3, GANSS_UTCmodelSet3_sequence);

  return offset;
}


static const value_string nbap_GANSS_Additional_UTC_Models_vals[] = {
  {   0, "utcModel1" },
  {   1, "utcModel2" },
  {   2, "utcModel3" },
  { 0, NULL }
};

static const per_choice_t GANSS_Additional_UTC_Models_choice[] = {
  {   0, &hf_nbap_utcModel1      , ASN1_EXTENSION_ROOT    , dissect_nbap_GANSS_UTCmodelSet1 },
  {   1, &hf_nbap_utcModel2      , ASN1_EXTENSION_ROOT    , dissect_nbap_GANSS_UTCmodelSet2 },
  {   2, &hf_nbap_utcModel3      , ASN1_EXTENSION_ROOT    , dissect_nbap_GANSS_UTCmodelSet3 },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_GANSS_Additional_UTC_Models(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_GANSS_Additional_UTC_Models, GANSS_Additional_UTC_Models_choice,
                                 NULL);

  return offset;
}



static int
dissect_nbap_INTEGER_0_255(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_INTEGER_0_3(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 3U, NULL, FALSE);

  return offset;
}


static const per_sequence_t GANSS_SatelliteInformationKPItem_sequence[] = {
  { &hf_nbap_satId          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_63 },
  { &hf_nbap_ganss_e_alm    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_11 },
  { &hf_nbap_ganss_delta_I_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_11 },
  { &hf_nbap_ganss_omegadot_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_11 },
  { &hf_nbap_ganss_svhealth_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_4 },
  { &hf_nbap_ganss_delta_a_sqrt_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_17 },
  { &hf_nbap_ganss_omegazero_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_ganss_m_zero_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_ganss_omega_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_ganss_af_zero_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_14 },
  { &hf_nbap_ganss_af_one_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_11 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_SatelliteInformationKPItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_SatelliteInformationKPItem, GANSS_SatelliteInformationKPItem_sequence);

  return offset;
}


static const per_sequence_t GANSS_SatelliteInformationKP_sequence_of[1] = {
  { &hf_nbap_GANSS_SatelliteInformationKP_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_SatelliteInformationKPItem },
};

static int
dissect_nbap_GANSS_SatelliteInformationKP(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_GANSS_SatelliteInformationKP, GANSS_SatelliteInformationKP_sequence_of,
                                                  1, maxGANSSSatAlmanac, FALSE);

  return offset;
}


static const per_sequence_t GANSS_KeplerianParametersAlm_sequence[] = {
  { &hf_nbap_t_oa           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_255 },
  { &hf_nbap_iod_a          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_3 },
  { &hf_nbap_gANSS_SatelliteInformationKP, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_SatelliteInformationKP },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_KeplerianParametersAlm(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_KeplerianParametersAlm, GANSS_KeplerianParametersAlm_sequence);

  return offset;
}



static int
dissect_nbap_Extension_GANSS_AlmanacModel(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string nbap_GANSS_AlmanacModel_vals[] = {
  {   0, "gANSS-keplerianParameters" },
  {   1, "extension-GANSS-AlmanacModel" },
  { 0, NULL }
};

static const per_choice_t GANSS_AlmanacModel_choice[] = {
  {   0, &hf_nbap_gANSS_keplerianParameters, ASN1_EXTENSION_ROOT    , dissect_nbap_GANSS_KeplerianParametersAlm },
  {   1, &hf_nbap_extension_GANSS_AlmanacModel, ASN1_NOT_EXTENSION_ROOT, dissect_nbap_Extension_GANSS_AlmanacModel },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_GANSS_AlmanacModel(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_GANSS_AlmanacModel, GANSS_AlmanacModel_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t GANSS_Almanac_sequence[] = {
  { &hf_nbap_ganss_wk_number, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_255 },
  { &hf_nbap_gANSS_AlmanacModel, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_AlmanacModel },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_Almanac(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_Almanac, GANSS_Almanac_sequence);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_9(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     9, 9, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t GANSS_SAT_Info_Almanac_SBASecef_sequence[] = {
  { &hf_nbap_sbasAlmDataID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_2 },
  { &hf_nbap_svID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_63 },
  { &hf_nbap_sbasAlmHealth  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_sbasAlmXg      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_15 },
  { &hf_nbap_sbasAlmYg      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_15 },
  { &hf_nbap_sbasAlmZg      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_9 },
  { &hf_nbap_sbasAlmXgdot   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_3 },
  { &hf_nbap_sbasAlmYgDot   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_3 },
  { &hf_nbap_sbasAlmZgDot   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_4 },
  { &hf_nbap_sbasAlmTo      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_11 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_SAT_Info_Almanac_SBASecef(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_SAT_Info_Almanac_SBASecef, GANSS_SAT_Info_Almanac_SBASecef_sequence);

  return offset;
}


static const per_sequence_t GANSS_SAT_Info_Almanac_SBASecefList_sequence_of[1] = {
  { &hf_nbap_GANSS_SAT_Info_Almanac_SBASecefList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_SAT_Info_Almanac_SBASecef },
};

static int
dissect_nbap_GANSS_SAT_Info_Almanac_SBASecefList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_GANSS_SAT_Info_Almanac_SBASecefList, GANSS_SAT_Info_Almanac_SBASecefList_sequence_of,
                                                  1, maxGANSSSatAlmanac, FALSE);

  return offset;
}


static const per_sequence_t GANSS_ALM_ECEFsbasAlmanacSet_sequence[] = {
  { &hf_nbap_sat_info_SBASecefList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_SAT_Info_Almanac_SBASecefList },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_ALM_ECEFsbasAlmanacSet(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_ALM_ECEFsbasAlmanacSet, GANSS_ALM_ECEFsbasAlmanacSet_sequence);

  return offset;
}


static const per_sequence_t GANSS_SAT_Info_Almanac_GLOkp_sequence[] = {
  { &hf_nbap_gloAlmNA       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_11 },
  { &hf_nbap_gloAlmnA       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_5 },
  { &hf_nbap_gloAlmHA       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_5 },
  { &hf_nbap_gloAlmLambdaA  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_21 },
  { &hf_nbap_gloAlmTlambdaA , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_21 },
  { &hf_nbap_gloAlmDeltaIA  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_18 },
  { &hf_nbap_gloAkmDeltaTA  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_22 },
  { &hf_nbap_gloAlmDeltaTdotA, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_7 },
  { &hf_nbap_gloAlmEpsilonA , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_15 },
  { &hf_nbap_gloAlmOmegaA   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_gloAlmTauA     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_10 },
  { &hf_nbap_gloAlmCA       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_1 },
  { &hf_nbap_gloAlmMA       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BIT_STRING_SIZE_2 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_SAT_Info_Almanac_GLOkp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_SAT_Info_Almanac_GLOkp, GANSS_SAT_Info_Almanac_GLOkp_sequence);

  return offset;
}


static const per_sequence_t GANSS_SAT_Info_Almanac_GLOkpList_sequence_of[1] = {
  { &hf_nbap_GANSS_SAT_Info_Almanac_GLOkpList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_SAT_Info_Almanac_GLOkp },
};

static int
dissect_nbap_GANSS_SAT_Info_Almanac_GLOkpList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_GANSS_SAT_Info_Almanac_GLOkpList, GANSS_SAT_Info_Almanac_GLOkpList_sequence_of,
                                                  1, maxGANSSSatAlmanac, FALSE);

  return offset;
}


static const per_sequence_t GANSS_ALM_GlonassAlmanacSet_sequence[] = {
  { &hf_nbap_sat_info_GLOkpList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_SAT_Info_Almanac_GLOkpList },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_ALM_GlonassAlmanacSet(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_ALM_GlonassAlmanacSet, GANSS_ALM_GlonassAlmanacSet_sequence);

  return offset;
}


static const per_sequence_t GANSS_SAT_Info_Almanac_MIDIkp_sequence[] = {
  { &hf_nbap_svID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_63 },
  { &hf_nbap_midiAlmE       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_11 },
  { &hf_nbap_midiAlmDeltaI  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_11 },
  { &hf_nbap_midiAlmOmegaDot, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_11 },
  { &hf_nbap_midiAlmSqrtA   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_17 },
  { &hf_nbap_midiAlmOmega0  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_midiAlmOmega   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_midiAlmMo      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_midiAlmaf0     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_11 },
  { &hf_nbap_midiAlmaf1     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_10 },
  { &hf_nbap_midiAlmL1Health, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_1 },
  { &hf_nbap_midiAlmL2Health, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_1 },
  { &hf_nbap_midiAlmL5Health, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_1 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_SAT_Info_Almanac_MIDIkp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_SAT_Info_Almanac_MIDIkp, GANSS_SAT_Info_Almanac_MIDIkp_sequence);

  return offset;
}


static const per_sequence_t GANSS_SAT_Info_Almanac_MIDIkpList_sequence_of[1] = {
  { &hf_nbap_GANSS_SAT_Info_Almanac_MIDIkpList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_SAT_Info_Almanac_MIDIkp },
};

static int
dissect_nbap_GANSS_SAT_Info_Almanac_MIDIkpList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_GANSS_SAT_Info_Almanac_MIDIkpList, GANSS_SAT_Info_Almanac_MIDIkpList_sequence_of,
                                                  1, maxGANSSSatAlmanac, FALSE);

  return offset;
}


static const per_sequence_t GANSS_ALM_MidiAlmanacSet_sequence[] = {
  { &hf_nbap_t_oa           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_255 },
  { &hf_nbap_sat_info_MIDIkpList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_SAT_Info_Almanac_MIDIkpList },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_ALM_MidiAlmanacSet(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_ALM_MidiAlmanacSet, GANSS_ALM_MidiAlmanacSet_sequence);

  return offset;
}


static const per_sequence_t GANSS_SAT_Info_Almanac_NAVkp_sequence[] = {
  { &hf_nbap_svID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_63 },
  { &hf_nbap_navAlmE        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_navAlmDeltaI   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_navAlmOMEGADOT , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_navAlmSVHealth , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_navAlmSqrtA    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_24 },
  { &hf_nbap_navAlmOMEGAo   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_24 },
  { &hf_nbap_navAlmOmega    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_24 },
  { &hf_nbap_navAlmMo       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_24 },
  { &hf_nbap_navAlmaf0      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_11 },
  { &hf_nbap_navAlmaf1      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_11 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_SAT_Info_Almanac_NAVkp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_SAT_Info_Almanac_NAVkp, GANSS_SAT_Info_Almanac_NAVkp_sequence);

  return offset;
}


static const per_sequence_t GANSS_SAT_Info_Almanac_NAVkpList_sequence_of[1] = {
  { &hf_nbap_GANSS_SAT_Info_Almanac_NAVkpList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_SAT_Info_Almanac_NAVkp },
};

static int
dissect_nbap_GANSS_SAT_Info_Almanac_NAVkpList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_GANSS_SAT_Info_Almanac_NAVkpList, GANSS_SAT_Info_Almanac_NAVkpList_sequence_of,
                                                  1, maxGANSSSatAlmanac, FALSE);

  return offset;
}


static const per_sequence_t GANSS_ALM_NAVKeplerianSet_sequence[] = {
  { &hf_nbap_t_oa           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_255 },
  { &hf_nbap_sat_info_NAVkpList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_SAT_Info_Almanac_NAVkpList },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_ALM_NAVKeplerianSet(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_ALM_NAVKeplerianSet, GANSS_ALM_NAVKeplerianSet_sequence);

  return offset;
}


static const per_sequence_t GANSS_SAT_Info_Almanac_REDkp_sequence[] = {
  { &hf_nbap_svID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_63 },
  { &hf_nbap_redAlmDeltaA   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_redAlmOmega0   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_7 },
  { &hf_nbap_redAlmPhi0     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_7 },
  { &hf_nbap_redAlmL1Health , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_1 },
  { &hf_nbap_redAlmL2Health , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_1 },
  { &hf_nbap_redAlmL5Health , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_1 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_SAT_Info_Almanac_REDkp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_SAT_Info_Almanac_REDkp, GANSS_SAT_Info_Almanac_REDkp_sequence);

  return offset;
}


static const per_sequence_t GANSS_SAT_Info_Almanac_REDkpList_sequence_of[1] = {
  { &hf_nbap_GANSS_SAT_Info_Almanac_REDkpList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_SAT_Info_Almanac_REDkp },
};

static int
dissect_nbap_GANSS_SAT_Info_Almanac_REDkpList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_GANSS_SAT_Info_Almanac_REDkpList, GANSS_SAT_Info_Almanac_REDkpList_sequence_of,
                                                  1, maxGANSSSatAlmanac, FALSE);

  return offset;
}


static const per_sequence_t GANSS_ALM_ReducedKeplerianSet_sequence[] = {
  { &hf_nbap_t_oa           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_255 },
  { &hf_nbap_sat_info_REDkpList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_SAT_Info_Almanac_REDkpList },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_ALM_ReducedKeplerianSet(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_ALM_ReducedKeplerianSet, GANSS_ALM_ReducedKeplerianSet_sequence);

  return offset;
}


static const per_sequence_t GANSS_AuxInfoGANSS_ID1_element_sequence[] = {
  { &hf_nbap_svID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_63 },
  { &hf_nbap_signalsAvailable, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_AuxInfoGANSS_ID1_element(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_AuxInfoGANSS_ID1_element, GANSS_AuxInfoGANSS_ID1_element_sequence);

  return offset;
}


static const per_sequence_t GANSS_AuxInfoGANSS_ID1_sequence_of[1] = {
  { &hf_nbap_GANSS_AuxInfoGANSS_ID1_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_AuxInfoGANSS_ID1_element },
};

static int
dissect_nbap_GANSS_AuxInfoGANSS_ID1(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_GANSS_AuxInfoGANSS_ID1, GANSS_AuxInfoGANSS_ID1_sequence_of,
                                                  1, maxGANSSSat, FALSE);

  return offset;
}



static int
dissect_nbap_INTEGER_M7_13(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -7, 13U, NULL, FALSE);

  return offset;
}


static const per_sequence_t GANSS_AuxInfoGANSS_ID3_element_sequence[] = {
  { &hf_nbap_svID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_63 },
  { &hf_nbap_signalsAvailable, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_channelNumber  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_M7_13 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_AuxInfoGANSS_ID3_element(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_AuxInfoGANSS_ID3_element, GANSS_AuxInfoGANSS_ID3_element_sequence);

  return offset;
}


static const per_sequence_t GANSS_AuxInfoGANSS_ID3_sequence_of[1] = {
  { &hf_nbap_GANSS_AuxInfoGANSS_ID3_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_AuxInfoGANSS_ID3_element },
};

static int
dissect_nbap_GANSS_AuxInfoGANSS_ID3(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_GANSS_AuxInfoGANSS_ID3, GANSS_AuxInfoGANSS_ID3_sequence_of,
                                                  1, maxGANSSSat, FALSE);

  return offset;
}


static const value_string nbap_GANSS_Auxiliary_Information_vals[] = {
  {   0, "ganssID1" },
  {   1, "ganssID3" },
  { 0, NULL }
};

static const per_choice_t GANSS_Auxiliary_Information_choice[] = {
  {   0, &hf_nbap_ganssID1       , ASN1_EXTENSION_ROOT    , dissect_nbap_GANSS_AuxInfoGANSS_ID1 },
  {   1, &hf_nbap_ganssID3       , ASN1_EXTENSION_ROOT    , dissect_nbap_GANSS_AuxInfoGANSS_ID3 },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_GANSS_Auxiliary_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_GANSS_Auxiliary_Information, GANSS_Auxiliary_Information_choice,
                                 NULL);

  return offset;
}



static int
dissect_nbap_GANSS_AuxInfoReq(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_boolean(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_28(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     28, 28, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_INTEGER_0_1_(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 1U, NULL, TRUE);

  return offset;
}


static const per_sequence_t GANSS_SatelliteClockModelItem_sequence[] = {
  { &hf_nbap_t_oc           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_14 },
  { &hf_nbap_a_i2           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_12 },
  { &hf_nbap_a_i1           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_18 },
  { &hf_nbap_a_i0           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_28 },
  { &hf_nbap_t_gd           , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BIT_STRING_SIZE_10 },
  { &hf_nbap_model_id       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_INTEGER_0_1_ },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_SatelliteClockModelItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_SatelliteClockModelItem, GANSS_SatelliteClockModelItem_sequence);

  return offset;
}


static const per_sequence_t GANSS_Clock_Model_sequence_of[1] = {
  { &hf_nbap_GANSS_Clock_Model_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_SatelliteClockModelItem },
};

static int
dissect_nbap_GANSS_Clock_Model(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_GANSS_Clock_Model, GANSS_Clock_Model_sequence_of,
                                                  1, maxGANSSClockMod, FALSE);

  return offset;
}


static const per_sequence_t GANSS_IonosphereRegionalStormFlags_sequence[] = {
  { &hf_nbap_storm_flag_one , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BOOLEAN },
  { &hf_nbap_storm_flag_two , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BOOLEAN },
  { &hf_nbap_storm_flag_three, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BOOLEAN },
  { &hf_nbap_storm_flag_four, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BOOLEAN },
  { &hf_nbap_storm_flag_five, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BOOLEAN },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_IonosphereRegionalStormFlags(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_IonosphereRegionalStormFlags, GANSS_IonosphereRegionalStormFlags_sequence);

  return offset;
}


static const per_sequence_t GANSS_Ionospheric_Model_sequence[] = {
  { &hf_nbap_alpha_zero_ionos, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_12 },
  { &hf_nbap_alpha_one_ionos, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_12 },
  { &hf_nbap_alpha_two_ionos, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_12 },
  { &hf_nbap_gANSS_IonosphereRegionalStormFlags, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GANSS_IonosphereRegionalStormFlags },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_Ionospheric_Model(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_Ionospheric_Model, GANSS_Ionospheric_Model_sequence);

  return offset;
}


static const value_string nbap_T_latitudeSign_vals[] = {
  {   0, "north" },
  {   1, "south" },
  { 0, NULL }
};


static int
dissect_nbap_T_latitudeSign(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_INTEGER_0_2147483647(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 2147483647U, NULL, FALSE);

  return offset;
}


static const value_string nbap_T_directionOfAltitude_vals[] = {
  {   0, "height" },
  {   1, "depth" },
  { 0, NULL }
};


static int
dissect_nbap_T_directionOfAltitude(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_INTEGER_0_32767(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 32767U, NULL, FALSE);

  return offset;
}


static const per_sequence_t GANSS_RX_Pos_sequence[] = {
  { &hf_nbap_latitudeSign   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_latitudeSign },
  { &hf_nbap_degreesOfLatitude, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_2147483647 },
  { &hf_nbap_degreesOfLongitude, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_M2147483648_2147483647 },
  { &hf_nbap_directionOfAltitude, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_directionOfAltitude },
  { &hf_nbap_altitude       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_32767 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_RX_Pos(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_RX_Pos, GANSS_RX_Pos_sequence);

  return offset;
}


static const per_sequence_t GANSS_Common_Data_sequence[] = {
  { &hf_nbap_ganss_Ionospheric_Model, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GANSS_Ionospheric_Model },
  { &hf_nbap_ganss_Rx_Pos   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GANSS_RX_Pos },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_Common_Data(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_Common_Data, GANSS_Common_Data_sequence);

  return offset;
}


static const per_sequence_t GANSS_CommonDataInfoReq_sequence[] = {
  { &hf_nbap_ionospheric_Model, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BOOLEAN },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_CommonDataInfoReq(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_CommonDataInfoReq, GANSS_CommonDataInfoReq_sequence);

  return offset;
}



static int
dissect_nbap_INTEGER_0_59_(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 59U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_1_1024(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     1, 1024, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t GANSS_DataBitAssistanceSgnItem_sequence[] = {
  { &hf_nbap_ganss_SignalId , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_Signal_ID },
  { &hf_nbap_ganssDataBits  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_1_1024 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_DataBitAssistanceSgnItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_DataBitAssistanceSgnItem, GANSS_DataBitAssistanceSgnItem_sequence);

  return offset;
}


static const per_sequence_t GANSS_DataBitAssistanceSgnList_sequence_of[1] = {
  { &hf_nbap_GANSS_DataBitAssistanceSgnList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_DataBitAssistanceSgnItem },
};

static int
dissect_nbap_GANSS_DataBitAssistanceSgnList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_GANSS_DataBitAssistanceSgnList, GANSS_DataBitAssistanceSgnList_sequence_of,
                                                  1, maxSgnType, FALSE);

  return offset;
}


static const per_sequence_t GANSS_DataBitAssistanceItem_sequence[] = {
  { &hf_nbap_satId          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_63 },
  { &hf_nbap_dataBitAssistanceSgnList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_DataBitAssistanceSgnList },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_DataBitAssistanceItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_DataBitAssistanceItem, GANSS_DataBitAssistanceItem_sequence);

  return offset;
}


static const per_sequence_t GANSS_DataBitAssistanceList_sequence_of[1] = {
  { &hf_nbap_GANSS_DataBitAssistanceList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_DataBitAssistanceItem },
};

static int
dissect_nbap_GANSS_DataBitAssistanceList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_GANSS_DataBitAssistanceList, GANSS_DataBitAssistanceList_sequence_of,
                                                  1, maxGANSSSat, FALSE);

  return offset;
}


static const per_sequence_t GANSS_Data_Bit_Assistance_sequence[] = {
  { &hf_nbap_ganssTod       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_59_ },
  { &hf_nbap_dataBitAssistancelist, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_DataBitAssistanceList },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_Data_Bit_Assistance(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_Data_Bit_Assistance, GANSS_Data_Bit_Assistance_sequence);

  return offset;
}



static int
dissect_nbap_INTEGER_0_15(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}


static const per_sequence_t T_ganss_SatelliteInfo_sequence_of[1] = {
  { &hf_nbap_ganss_SatelliteInfo_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_63 },
};

static int
dissect_nbap_T_ganss_SatelliteInfo(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_T_ganss_SatelliteInfo, T_ganss_SatelliteInfo_sequence_of,
                                                  1, maxGANSSSat, FALSE);

  return offset;
}


static const per_sequence_t GANSS_Data_Bit_Assistance_ReqList_sequence[] = {
  { &hf_nbap_dGANSS_Signal_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_ganss_DataBitInterval, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_15 },
  { &hf_nbap_ganss_SatelliteInfo, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_T_ganss_SatelliteInfo },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_Data_Bit_Assistance_ReqList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_Data_Bit_Assistance_ReqList, GANSS_Data_Bit_Assistance_ReqList_sequence);

  return offset;
}


static const per_sequence_t GANSS_Data_Bit_Assistance_ReqItem_sequence[] = {
  { &hf_nbap_ganssTod_01    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_86399 },
  { &hf_nbap_ganss_Data_Bit_Assistance_ReqList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_Data_Bit_Assistance_ReqList },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_Data_Bit_Assistance_ReqItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_Data_Bit_Assistance_ReqItem, GANSS_Data_Bit_Assistance_ReqItem_sequence);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_31(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     31, 31, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_19(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     19, 19, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t GANSS_Earth_Orientation_Parameters_sequence[] = {
  { &hf_nbap_teop           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_pmX            , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_21 },
  { &hf_nbap_pmXdot         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_15 },
  { &hf_nbap_pmY            , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_21 },
  { &hf_nbap_pmYdot         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_15 },
  { &hf_nbap_deltaUT1       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_31 },
  { &hf_nbap_deltaUT1dot    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_19 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_Earth_Orientation_Parameters(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_Earth_Orientation_Parameters, GANSS_Earth_Orientation_Parameters_sequence);

  return offset;
}



static int
dissect_nbap_GANSS_EarthOrientParaReq(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_boolean(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}



static int
dissect_nbap_GANSS_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 7U, NULL, TRUE);

  return offset;
}


static const per_sequence_t GANSS_GenericDataInfoReqItem_sequence[] = {
  { &hf_nbap_ganss_Id       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GANSS_ID },
  { &hf_nbap_ganss_Navigation_Model_And_Time_Recovery, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BOOLEAN },
  { &hf_nbap_ganss_Time_Model_GNSS_GNSS, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BIT_STRING_SIZE_9 },
  { &hf_nbap_ganss_UTC_Model, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BOOLEAN },
  { &hf_nbap_ganss_Almanac  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BOOLEAN },
  { &hf_nbap_gANSS_GenericDataInfoReqItem_ganss_Real_Time_Integrity, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BOOLEAN },
  { &hf_nbap_ganss_Data_Bit_Assistance_Req, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GANSS_Data_Bit_Assistance_ReqItem },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_GenericDataInfoReqItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_GenericDataInfoReqItem, GANSS_GenericDataInfoReqItem_sequence);

  return offset;
}


static const per_sequence_t GANSS_GenericDataInfoReqList_sequence_of[1] = {
  { &hf_nbap_GANSS_GenericDataInfoReqList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_GenericDataInfoReqItem },
};

static int
dissect_nbap_GANSS_GenericDataInfoReqList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_GANSS_GenericDataInfoReqList, GANSS_GenericDataInfoReqList_sequence_of,
                                                  1, maxNoGANSS, FALSE);

  return offset;
}


static const value_string nbap_T_non_broadcastIndication_01_vals[] = {
  {   0, "true" },
  { 0, NULL }
};


static int
dissect_nbap_T_non_broadcastIndication_01(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t GANSS_KeplerianParametersOrb_sequence[] = {
  { &hf_nbap_toe_nav        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_14 },
  { &hf_nbap_ganss_omega_nav, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_32 },
  { &hf_nbap_delta_n_nav    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_m_zero_nav     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_32 },
  { &hf_nbap_omegadot_nav   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_24 },
  { &hf_nbap_ganss_e_nav    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_32 },
  { &hf_nbap_idot_nav       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_14 },
  { &hf_nbap_a_sqrt_nav     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_32 },
  { &hf_nbap_i_zero_nav     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_32 },
  { &hf_nbap_omega_zero_nav , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_32 },
  { &hf_nbap_c_rs_nav       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_c_is_nav       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_c_us_nav       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_c_rc_nav       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_c_ic_nav       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_c_uc_nav       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_KeplerianParametersOrb(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_KeplerianParametersOrb, GANSS_KeplerianParametersOrb_sequence);

  return offset;
}


static const value_string nbap_GANSS_Orbit_Model_vals[] = {
  {   0, "gANSS-keplerianParameters" },
  { 0, NULL }
};

static const per_choice_t GANSS_Orbit_Model_choice[] = {
  {   0, &hf_nbap_gANSS_keplerianParameters_01, ASN1_EXTENSION_ROOT    , dissect_nbap_GANSS_KeplerianParametersOrb },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_GANSS_Orbit_Model(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_GANSS_Orbit_Model, GANSS_Orbit_Model_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t GANSS_Sat_Info_Nav_item_sequence[] = {
  { &hf_nbap_satId          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_63 },
  { &hf_nbap_svHealth_01    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_5 },
  { &hf_nbap_iod_01         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_10 },
  { &hf_nbap_ganssClockModel, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_Clock_Model },
  { &hf_nbap_ganssOrbitModel, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_Orbit_Model },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_Sat_Info_Nav_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_Sat_Info_Nav_item, GANSS_Sat_Info_Nav_item_sequence);

  return offset;
}


static const per_sequence_t GANSS_Sat_Info_Nav_sequence_of[1] = {
  { &hf_nbap_GANSS_Sat_Info_Nav_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_Sat_Info_Nav_item },
};

static int
dissect_nbap_GANSS_Sat_Info_Nav(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_GANSS_Sat_Info_Nav, GANSS_Sat_Info_Nav_sequence_of,
                                                  1, maxGANSSSat, FALSE);

  return offset;
}


static const per_sequence_t GANSS_Navigation_Model_And_Time_Recovery_sequence[] = {
  { &hf_nbap_ganss_Transmission_Time, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_Transmission_Time },
  { &hf_nbap_non_broadcastIndication_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_T_non_broadcastIndication_01 },
  { &hf_nbap_ganssSatInfoNav, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_Sat_Info_Nav },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_Navigation_Model_And_Time_Recovery(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_Navigation_Model_And_Time_Recovery, GANSS_Navigation_Model_And_Time_Recovery_sequence);

  return offset;
}


static const per_sequence_t GANSS_UTC_Model_sequence[] = {
  { &hf_nbap_a_one_utc      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_24 },
  { &hf_nbap_a_zero_utc     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_32 },
  { &hf_nbap_t_ot_utc       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_w_n_t_utc      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_delta_t_ls_utc , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_w_n_lsf_utc    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_dn_utc         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_delta_t_lsf_utc, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_UTC_Model(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_UTC_Model, GANSS_UTC_Model_sequence);

  return offset;
}


static const per_sequence_t GANSS_RealTimeInformationItem_sequence[] = {
  { &hf_nbap_bad_ganss_satId, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_63 },
  { &hf_nbap_bad_ganss_signalId, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_RealTimeInformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_RealTimeInformationItem, GANSS_RealTimeInformationItem_sequence);

  return offset;
}


static const per_sequence_t GANSS_Real_Time_Integrity_sequence_of[1] = {
  { &hf_nbap_GANSS_Real_Time_Integrity_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_RealTimeInformationItem },
};

static int
dissect_nbap_GANSS_Real_Time_Integrity(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_GANSS_Real_Time_Integrity, GANSS_Real_Time_Integrity_sequence_of,
                                                  1, maxGANSSSat, FALSE);

  return offset;
}


static const per_sequence_t GANSS_Generic_DataItem_sequence[] = {
  { &hf_nbap_ganss_Id       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GANSS_ID },
  { &hf_nbap_dganss_Correction, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DGANSSCorrections },
  { &hf_nbap_ganss_Navigation_Model_And_Time_Recovery_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GANSS_Navigation_Model_And_Time_Recovery },
  { &hf_nbap_ganss_Time_Model, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GANSS_Time_Model },
  { &hf_nbap_ganss_UTC_TIME , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GANSS_UTC_Model },
  { &hf_nbap_ganss_Almanac_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GANSS_Almanac },
  { &hf_nbap_ganss_Real_Time_Integrity, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GANSS_Real_Time_Integrity },
  { &hf_nbap_ganss_Data_Bit_Assistance, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GANSS_Data_Bit_Assistance },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_Generic_DataItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_Generic_DataItem, GANSS_Generic_DataItem_sequence);

  return offset;
}


static const per_sequence_t GANSS_Generic_Data_sequence_of[1] = {
  { &hf_nbap_GANSS_Generic_Data_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_GANSS_Generic_DataItem },
};

static int
dissect_nbap_GANSS_Generic_Data(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_GANSS_Generic_Data, GANSS_Generic_Data_sequence_of,
                                                  1, maxNoGANSS, FALSE);

  return offset;
}


static const per_sequence_t GANSS_Information_sequence[] = {
  { &hf_nbap_gANSS_CommonDataInfoReq, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GANSS_CommonDataInfoReq },
  { &hf_nbap_gANSS_GenericDataInfoReqList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GANSS_GenericDataInfoReqList },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GANSS_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GANSS_Information, GANSS_Information_sequence);

  return offset;
}


static const value_string nbap_GANSS_SBAS_ID_vals[] = {
  {   0, "waas" },
  {   1, "egnos" },
  {   2, "msas" },
  {   3, "gagan" },
  { 0, NULL }
};


static int
dissect_nbap_GANSS_SBAS_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_GANSS_Time_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 7U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_GapLength(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 14U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_GapDuration(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 144U, NULL, TRUE);

  return offset;
}


static const per_sequence_t SAT_Info_Almanac_Item_sequence[] = {
  { &hf_nbap_data_id        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DATA_ID },
  { &hf_nbap_sat_id         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SAT_ID },
  { &hf_nbap_gps_e_alm      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_gps_toa_alm    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_gps_delta_I_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_omegadot_alm   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_svhealth_alm   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_gps_a_sqrt_alm , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_24 },
  { &hf_nbap_omegazero_alm  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_24 },
  { &hf_nbap_m_zero_alm     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_24 },
  { &hf_nbap_gps_omega_alm  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_24 },
  { &hf_nbap_gps_af_zero_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_11 },
  { &hf_nbap_gps_af_one_alm , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_11 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SAT_Info_Almanac_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SAT_Info_Almanac_Item, SAT_Info_Almanac_Item_sequence);

  return offset;
}


static const per_sequence_t SAT_Info_Almanac_sequence_of[1] = {
  { &hf_nbap_SAT_Info_Almanac_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_SAT_Info_Almanac_Item },
};

static int
dissect_nbap_SAT_Info_Almanac(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_SAT_Info_Almanac, SAT_Info_Almanac_sequence_of,
                                                  1, maxNoSat, FALSE);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_364(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     364, 364, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t GPS_Almanac_sequence[] = {
  { &hf_nbap_wna_alm        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_sat_info_almanac, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SAT_Info_Almanac },
  { &hf_nbap_sVGlobalHealth_alm, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BIT_STRING_SIZE_364 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GPS_Almanac(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GPS_Almanac, GPS_Almanac_sequence);

  return offset;
}


static const value_string nbap_GPS_Information_Item_vals[] = {
  {   0, "gps-navigation-model-and-time-recovery" },
  {   1, "gps-ionospheric-model" },
  {   2, "gps-utc-model" },
  {   3, "gps-almanac" },
  {   4, "gps-rt-integrity" },
  { 0, NULL }
};


static int
dissect_nbap_GPS_Information_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     5, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t GPS_Information_sequence_of[1] = {
  { &hf_nbap_GPS_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_GPS_Information_Item },
};

static int
dissect_nbap_GPS_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_GPS_Information, GPS_Information_sequence_of,
                                                  0, maxNoGPSItems, FALSE);

  return offset;
}


static const per_sequence_t SAT_Info_RealTime_Integrity_Item_sequence[] = {
  { &hf_nbap_bad_sat_id     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SAT_ID },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SAT_Info_RealTime_Integrity_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SAT_Info_RealTime_Integrity_Item, SAT_Info_RealTime_Integrity_Item_sequence);

  return offset;
}


static const per_sequence_t SATInfo_RealTime_Integrity_sequence_of[1] = {
  { &hf_nbap_SATInfo_RealTime_Integrity_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_SAT_Info_RealTime_Integrity_Item },
};

static int
dissect_nbap_SATInfo_RealTime_Integrity(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_SATInfo_RealTime_Integrity, SATInfo_RealTime_Integrity_sequence_of,
                                                  1, maxNoSat, FALSE);

  return offset;
}


static const per_sequence_t GPSBadSat_Info_RealTime_Integrity_sequence[] = {
  { &hf_nbap_sat_info       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SATInfo_RealTime_Integrity },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GPSBadSat_Info_RealTime_Integrity(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GPSBadSat_Info_RealTime_Integrity, GPSBadSat_Info_RealTime_Integrity_sequence);

  return offset;
}


static const value_string nbap_GPS_RealTime_Integrity_vals[] = {
  {   0, "bad-satellites" },
  {   1, "no-bad-satellites" },
  { 0, NULL }
};

static const per_choice_t GPS_RealTime_Integrity_choice[] = {
  {   0, &hf_nbap_bad_satellites , ASN1_NO_EXTENSIONS     , dissect_nbap_GPSBadSat_Info_RealTime_Integrity },
  {   1, &hf_nbap_no_bad_satellites, ASN1_NO_EXTENSIONS     , dissect_nbap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_GPS_RealTime_Integrity(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_GPS_RealTime_Integrity, GPS_RealTime_Integrity_choice,
                                 NULL);

  return offset;
}



static int
dissect_nbap_INTEGER_0_1048575(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 1048575U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_BIT_STRING_SIZE_87(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     87, 87, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t GPS_NavandRecovery_Item_sequence[] = {
  { &hf_nbap_tx_tow_nav     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_1048575 },
  { &hf_nbap_sat_id_nav     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SAT_ID },
  { &hf_nbap_tlm_message_nav, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_14 },
  { &hf_nbap_tlm_revd_c_nav , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_2 },
  { &hf_nbap_ho_word_nav    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_22 },
  { &hf_nbap_w_n_nav        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_10 },
  { &hf_nbap_ca_or_p_on_l2_nav, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_2 },
  { &hf_nbap_user_range_accuracy_index_nav, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_4 },
  { &hf_nbap_sv_health_nav  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_6 },
  { &hf_nbap_iodc_nav       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_10 },
  { &hf_nbap_l2_p_dataflag_nav, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_1 },
  { &hf_nbap_sf1_reserved_nav, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_87 },
  { &hf_nbap_t_gd_nav       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_t_oc_nav       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_a_f_2_nav      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_a_f_1_nav      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_a_f_zero_nav   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_22 },
  { &hf_nbap_c_rs_nav       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_delta_n_nav    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_m_zero_nav     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_32 },
  { &hf_nbap_c_uc_nav       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_gps_e_nav      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_32 },
  { &hf_nbap_c_us_nav       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_a_sqrt_nav     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_32 },
  { &hf_nbap_t_oe_nav       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_fit_interval_flag_nav, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_1 },
  { &hf_nbap_aodo_nav       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_5 },
  { &hf_nbap_c_ic_nav       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_omega_zero_nav , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_32 },
  { &hf_nbap_c_is_nav       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_i_zero_nav     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_32 },
  { &hf_nbap_c_rc_nav       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_gps_omega_nav  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_32 },
  { &hf_nbap_omegadot_nav   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_24 },
  { &hf_nbap_idot_nav       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_14 },
  { &hf_nbap_spare_zero_fill, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_20 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GPS_NavandRecovery_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GPS_NavandRecovery_Item, GPS_NavandRecovery_Item_sequence);

  return offset;
}


static const per_sequence_t GPS_NavigationModel_and_TimeRecovery_sequence_of[1] = {
  { &hf_nbap_GPS_NavigationModel_and_TimeRecovery_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_GPS_NavandRecovery_Item },
};

static int
dissect_nbap_GPS_NavigationModel_and_TimeRecovery(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_GPS_NavigationModel_and_TimeRecovery, GPS_NavigationModel_and_TimeRecovery_sequence_of,
                                                  1, maxNoSat, FALSE);

  return offset;
}


static const value_string nbap_T_latitudeSign_01_vals[] = {
  {   0, "north" },
  {   1, "south" },
  { 0, NULL }
};


static int
dissect_nbap_T_latitudeSign_01(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_INTEGER_0_8388607(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 8388607U, NULL, FALSE);

  return offset;
}


static const value_string nbap_T_directionOfAltitude_01_vals[] = {
  {   0, "height" },
  {   1, "depth" },
  { 0, NULL }
};


static int
dissect_nbap_T_directionOfAltitude_01(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t GPS_RX_POS_sequence[] = {
  { &hf_nbap_latitudeSign_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_latitudeSign_01 },
  { &hf_nbap_latitude       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_8388607 },
  { &hf_nbap_longitude      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_M8388608_8388607 },
  { &hf_nbap_directionOfAltitude_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_directionOfAltitude_01 },
  { &hf_nbap_altitude       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_32767 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GPS_RX_POS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GPS_RX_POS, GPS_RX_POS_sequence);

  return offset;
}


static const per_sequence_t GPS_UTC_Model_sequence[] = {
  { &hf_nbap_a_one_utc      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_24 },
  { &hf_nbap_a_zero_utc     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_32 },
  { &hf_nbap_t_ot_utc       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_delta_t_ls_utc , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_w_n_t_utc      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_w_n_lsf_utc    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_dn_utc         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_delta_t_lsf_utc, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GPS_UTC_Model(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GPS_UTC_Model, GPS_UTC_Model_sequence);

  return offset;
}



static int
dissect_nbap_INTEGER_1_8_(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 8U, NULL, TRUE);

  return offset;
}


static const per_sequence_t HARQ_MemoryPartitioning_Implicit_sequence[] = {
  { &hf_nbap_number_of_Processes, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_1_8_ },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HARQ_MemoryPartitioning_Implicit(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HARQ_MemoryPartitioning_Implicit, HARQ_MemoryPartitioning_Implicit_sequence);

  return offset;
}


static const value_string nbap_T_process_Memory_Size_vals[] = {
  {   0, "hms800" },
  {   1, "hms1600" },
  {   2, "hms2400" },
  {   3, "hms3200" },
  {   4, "hms4000" },
  {   5, "hms4800" },
  {   6, "hms5600" },
  {   7, "hms6400" },
  {   8, "hms7200" },
  {   9, "hms8000" },
  {  10, "hms8800" },
  {  11, "hms9600" },
  {  12, "hms10400" },
  {  13, "hms11200" },
  {  14, "hms12000" },
  {  15, "hms12800" },
  {  16, "hms13600" },
  {  17, "hms14400" },
  {  18, "hms15200" },
  {  19, "hms16000" },
  {  20, "hms17600" },
  {  21, "hms19200" },
  {  22, "hms20800" },
  {  23, "hms22400" },
  {  24, "hms24000" },
  {  25, "hms25600" },
  {  26, "hms27200" },
  {  27, "hms28800" },
  {  28, "hms30400" },
  {  29, "hms32000" },
  {  30, "hms36000" },
  {  31, "hms40000" },
  {  32, "hms44000" },
  {  33, "hms48000" },
  {  34, "hms52000" },
  {  35, "hms56000" },
  {  36, "hms60000" },
  {  37, "hms64000" },
  {  38, "hms68000" },
  {  39, "hms72000" },
  {  40, "hms76000" },
  {  41, "hms80000" },
  {  42, "hms88000" },
  {  43, "hms96000" },
  {  44, "hms104000" },
  {  45, "hms112000" },
  {  46, "hms120000" },
  {  47, "hms128000" },
  {  48, "hms136000" },
  {  49, "hms144000" },
  {  50, "hms152000" },
  {  51, "hms160000" },
  {  52, "hms176000" },
  {  53, "hms192000" },
  {  54, "hms208000" },
  {  55, "hms224000" },
  {  56, "hms240000" },
  {  57, "hms256000" },
  {  58, "hms272000" },
  {  59, "hms288000" },
  {  60, "hms304000" },
  { 0, NULL }
};


static int
dissect_nbap_T_process_Memory_Size(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     61, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t HARQ_MemoryPartitioningItem_sequence[] = {
  { &hf_nbap_process_Memory_Size, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_process_Memory_Size },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HARQ_MemoryPartitioningItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HARQ_MemoryPartitioningItem, HARQ_MemoryPartitioningItem_sequence);

  return offset;
}


static const per_sequence_t HARQ_MemoryPartitioningList_sequence_of[1] = {
  { &hf_nbap_HARQ_MemoryPartitioningList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HARQ_MemoryPartitioningItem },
};

static int
dissect_nbap_HARQ_MemoryPartitioningList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HARQ_MemoryPartitioningList, HARQ_MemoryPartitioningList_sequence_of,
                                                  1, maxNrOfHARQProcesses, FALSE);

  return offset;
}


static const per_sequence_t HARQ_MemoryPartitioning_Explicit_sequence[] = {
  { &hf_nbap_hARQ_MemoryPartitioningList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HARQ_MemoryPartitioningList },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HARQ_MemoryPartitioning_Explicit(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HARQ_MemoryPartitioning_Explicit, HARQ_MemoryPartitioning_Explicit_sequence);

  return offset;
}


static const value_string nbap_HARQ_MemoryPartitioning_vals[] = {
  {   0, "implicit" },
  {   1, "explicit" },
  { 0, NULL }
};

static const per_choice_t HARQ_MemoryPartitioning_choice[] = {
  {   0, &hf_nbap_implicit       , ASN1_EXTENSION_ROOT    , dissect_nbap_HARQ_MemoryPartitioning_Implicit },
  {   1, &hf_nbap_explicit       , ASN1_EXTENSION_ROOT    , dissect_nbap_HARQ_MemoryPartitioning_Explicit },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_HARQ_MemoryPartitioning(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_HARQ_MemoryPartitioning, HARQ_MemoryPartitioning_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t HARQ_MemoryPartitioningInfoExtForMIMO_sequence_of[1] = {
  { &hf_nbap_HARQ_MemoryPartitioningInfoExtForMIMO_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HARQ_MemoryPartitioningItem },
};

static int
dissect_nbap_HARQ_MemoryPartitioningInfoExtForMIMO(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HARQ_MemoryPartitioningInfoExtForMIMO, HARQ_MemoryPartitioningInfoExtForMIMO_sequence_of,
                                                  4, 8, FALSE);

  return offset;
}


static const value_string nbap_HARQ_Preamble_Mode_vals[] = {
  {   0, "mode0" },
  {   1, "mode1" },
  { 0, NULL }
};


static int
dissect_nbap_HARQ_Preamble_Mode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_HARQ_Preamble_Mode_Activation_Indicator_vals[] = {
  {   0, "harqPreambleModeActivated" },
  { 0, NULL }
};


static int
dissect_nbap_HARQ_Preamble_Mode_Activation_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_HSDPA_Capability_vals[] = {
  {   0, "hsdpa-capable" },
  {   1, "hsdpa-non-capable" },
  { 0, NULL }
};


static int
dissect_nbap_HSDPA_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_HS_DSCHProvidedBitRateValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 16777215U, NULL, TRUE);

  return offset;
}


static const per_sequence_t HS_DSCHProvidedBitRate_Item_sequence[] = {
  { &hf_nbap_schedulingPriorityIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SchedulingPriorityIndicator },
  { &hf_nbap_hS_DSCHProvidedBitRateValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCHProvidedBitRateValue },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_DSCHProvidedBitRate_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_DSCHProvidedBitRate_Item, HS_DSCHProvidedBitRate_Item_sequence);

  return offset;
}


static const per_sequence_t HS_DSCHProvidedBitRate_sequence_of[1] = {
  { &hf_nbap_HS_DSCHProvidedBitRate_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCHProvidedBitRate_Item },
};

static int
dissect_nbap_HS_DSCHProvidedBitRate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HS_DSCHProvidedBitRate, HS_DSCHProvidedBitRate_sequence_of,
                                                  1, maxNrOfPriorityClasses, FALSE);

  return offset;
}


static const per_sequence_t HS_DSCHProvidedBitRateValueInformation_For_CellPortion_Item_sequence[] = {
  { &hf_nbap_cellPortionID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellPortionID },
  { &hf_nbap_hS_DSCHProvidedBitRateValue_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCHProvidedBitRate },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortion_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortion_Item, HS_DSCHProvidedBitRateValueInformation_For_CellPortion_Item_sequence);

  return offset;
}


static const per_sequence_t HS_DSCHProvidedBitRateValueInformation_For_CellPortion_sequence_of[1] = {
  { &hf_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortion_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortion_Item },
};

static int
dissect_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortion(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortion, HS_DSCHProvidedBitRateValueInformation_For_CellPortion_sequence_of,
                                                  1, maxNrOfCellPortionsPerCell, FALSE);

  return offset;
}


static const per_sequence_t HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR_Item_sequence[] = {
  { &hf_nbap_cellPortionLCRID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellPortionLCRID },
  { &hf_nbap_hS_DSCHProvidedBitRateValue_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCHProvidedBitRate },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR_Item, HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR_Item_sequence);

  return offset;
}


static const per_sequence_t HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR_sequence_of[1] = {
  { &hf_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR_Item },
};

static int
dissect_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR, HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR_sequence_of,
                                                  1, maxNrOfCellPortionsPerCellLCR, FALSE);

  return offset;
}



static int
dissect_nbap_HS_DSCHRequiredPowerValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 1000U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_HS_DSCHRequiredPowerPerUEWeight(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 100U, NULL, FALSE);

  return offset;
}


static const per_sequence_t HS_DSCHRequiredPowerPerUEInformation_Item_sequence[] = {
  { &hf_nbap_cRNC_CommunicationContextID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CRNC_CommunicationContextID },
  { &hf_nbap_hS_DSCHRequiredPowerPerUEWeight, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_DSCHRequiredPowerPerUEWeight },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_DSCHRequiredPowerPerUEInformation_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_DSCHRequiredPowerPerUEInformation_Item, HS_DSCHRequiredPowerPerUEInformation_Item_sequence);

  return offset;
}


static const per_sequence_t HS_DSCHRequiredPowerPerUEInformation_sequence_of[1] = {
  { &hf_nbap_HS_DSCHRequiredPowerPerUEInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCHRequiredPowerPerUEInformation_Item },
};

static int
dissect_nbap_HS_DSCHRequiredPowerPerUEInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HS_DSCHRequiredPowerPerUEInformation, HS_DSCHRequiredPowerPerUEInformation_sequence_of,
                                                  1, maxNrOfContextsOnUeList, FALSE);

  return offset;
}


static const per_sequence_t HS_DSCHRequiredPower_Item_sequence[] = {
  { &hf_nbap_schedulingPriorityIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SchedulingPriorityIndicator },
  { &hf_nbap_hS_DSCHRequiredPowerValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCHRequiredPowerValue },
  { &hf_nbap_hS_DSCHRequiredPowerPerUEInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_DSCHRequiredPowerPerUEInformation },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_DSCHRequiredPower_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_DSCHRequiredPower_Item, HS_DSCHRequiredPower_Item_sequence);

  return offset;
}


static const per_sequence_t HS_DSCHRequiredPower_sequence_of[1] = {
  { &hf_nbap_HS_DSCHRequiredPower_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCHRequiredPower_Item },
};

static int
dissect_nbap_HS_DSCHRequiredPower(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HS_DSCHRequiredPower, HS_DSCHRequiredPower_sequence_of,
                                                  1, maxNrOfPriorityClasses, FALSE);

  return offset;
}


static const per_sequence_t HS_DSCHRequiredPowerValueInformation_For_CellPortion_Item_sequence[] = {
  { &hf_nbap_cellPortionID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellPortionID },
  { &hf_nbap_hS_DSCHRequiredPowerValue_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCHRequiredPower },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortion_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortion_Item, HS_DSCHRequiredPowerValueInformation_For_CellPortion_Item_sequence);

  return offset;
}


static const per_sequence_t HS_DSCHRequiredPowerValueInformation_For_CellPortion_sequence_of[1] = {
  { &hf_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortion_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortion_Item },
};

static int
dissect_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortion(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortion, HS_DSCHRequiredPowerValueInformation_For_CellPortion_sequence_of,
                                                  1, maxNrOfCellPortionsPerCell, FALSE);

  return offset;
}


static const per_sequence_t HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR_Item_sequence[] = {
  { &hf_nbap_cellPortionLCRID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellPortionLCRID },
  { &hf_nbap_hS_DSCHRequiredPowerValue_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCHRequiredPower },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR_Item, HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR_Item_sequence);

  return offset;
}


static const per_sequence_t HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR_sequence_of[1] = {
  { &hf_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR_Item },
};

static int
dissect_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR, HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR_sequence_of,
                                                  1, maxNrOfCellPortionsPerCellLCR, FALSE);

  return offset;
}


static const per_sequence_t HSDPA_PICH_Shared_with_PCH_sequence[] = {
  { &hf_nbap_hsdpa_PICH_SharedPCH_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDPA_PICH_Shared_with_PCH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDPA_PICH_Shared_with_PCH, HSDPA_PICH_Shared_with_PCH_sequence);

  return offset;
}



static int
dissect_nbap_PICH_Power(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -10, 5U, NULL, FALSE);

  return offset;
}


static const value_string nbap_PICH_Mode_vals[] = {
  {   0, "v18" },
  {   1, "v36" },
  {   2, "v72" },
  {   3, "v144" },
  { 0, NULL }
};


static int
dissect_nbap_PICH_Mode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
guint32 PICH_Mode = 0;
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, &PICH_Mode, TRUE, 0, NULL);


    switch(PICH_Mode){
        case 0: /*  v18 */
            nbap_get_private_data(actx->pinfo)->paging_indications = 18;
            break;
        case 1: /*  v36 */
            nbap_get_private_data(actx->pinfo)->paging_indications = 36;
            break;
        case 2: /*  v72 */
            nbap_get_private_data(actx->pinfo)->paging_indications = 72;
            break;
        case 3: /*  v144 */
            nbap_get_private_data(actx->pinfo)->paging_indications = 144;
            break;
    }


  return offset;
}


static const per_sequence_t HSDPA_PICH_notShared_with_PCH_sequence[] = {
  { &hf_nbap_hSDPA_PICH_notShared_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_fdd_DL_Channelisation_CodeNumber, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_DL_ChannelisationCodeNumber },
  { &hf_nbap_pich_Power     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PICH_Power },
  { &hf_nbap_pich_Mode      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PICH_Mode },
  { &hf_nbap_sttd_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_STTD_Indicator },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDPA_PICH_notShared_with_PCH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDPA_PICH_notShared_with_PCH, HSDPA_PICH_notShared_with_PCH_sequence);

  return offset;
}


static const value_string nbap_HSDPA_Associated_PICH_Information_vals[] = {
  {   0, "hsdpa-PICH-Shared-with-PCH" },
  {   1, "hsdpa-PICH-notShared-with-PCH" },
  { 0, NULL }
};

static const per_choice_t HSDPA_Associated_PICH_Information_choice[] = {
  {   0, &hf_nbap_hsdpa_PICH_Shared_with_PCH, ASN1_EXTENSION_ROOT    , dissect_nbap_HSDPA_PICH_Shared_with_PCH },
  {   1, &hf_nbap_hsdpa_PICH_notShared_with_PCH, ASN1_EXTENSION_ROOT    , dissect_nbap_HSDPA_PICH_notShared_with_PCH },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_HSDPA_Associated_PICH_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_HSDPA_Associated_PICH_Information, HSDPA_Associated_PICH_Information_choice,
                                 NULL);

  return offset;
}


static const value_string nbap_RACH_Measurement_Result_vals[] = {
  {   0, "cpich-EcNo" },
  {   1, "cpich-RSCP" },
  {   2, "pathloss" },
  { 0, NULL }
};


static int
dissect_nbap_RACH_Measurement_Result(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t HSDSCH_Common_Information_sequence[] = {
  { &hf_nbap_cCCH_PriorityQueue_Id, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PriorityQueue_Id },
  { &hf_nbap_sRB1_PriorityQueue_Id, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PriorityQueue_Id },
  { &hf_nbap_associatedCommon_MACFlow, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Common_MACFlow_ID },
  { &hf_nbap_fACH_Measurement_Occasion_Cycle_Length_Coefficient, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_FACH_Measurement_Occasion_Cycle_Length_Coefficient },
  { &hf_nbap_rACH_Measurement_Result, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RACH_Measurement_Result },
  { &hf_nbap_bCCH_Specific_HSDSCH_RNTI_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BCCH_Specific_HSDSCH_RNTI_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_Common_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_Common_Information, HSDSCH_Common_Information_sequence);

  return offset;
}


static const per_sequence_t HSDSCH_Common_System_InformationFDD_sequence[] = {
  { &hf_nbap_hsdsch_Common_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSDSCH_Common_Information },
  { &hf_nbap_commonMACFlow_Specific_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CommonMACFlow_Specific_InfoList },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_Common_System_InformationFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
/*
 * 5.1.6 High Speed Downlink Shared Channels
 * The Data Transfer procedure is used to transfer a HS-DSCH DATA FRAME (TYPE 1, TYPE 2 [FDD and 1.28Mcps
 * TDD - or TYPE3]) from the CRNC to a Node B. HS-DSCH DATA FRAME TYPE 2 is selected if the IE HS-DSCH
 * MAC-d PDU Size Format in NBAP (TS 25.433 [6]) is present and set to "Flexible MAC-d PDU Size" [FDD and
 * 1.28Mcps TDD - or if the IE HS-DSCH Common System Information is present and the UE is in Cell_FACH state. HS-
 * DSCH DATA FRAME TYPE 3 is selected if the IE HS-DSCH Paging System Information in NBAP (TS 25.433 [6]) is
 * present and the UE is in Cell_PCH state or URA_PCH state]. HS-DSCH DATA FRAME TYPE 1 is selected in any
 * other case.
 */
umts_fp_conversation_info_t *umts_fp_conversation_info = NULL;
fp_hsdsch_channel_info_t* fp_hsdsch_channel_info = NULL;
address     null_addr;
conversation_t   *conversation = NULL;
nbap_common_channel_info_t* nbap_common_channel_info;
nbap_common_channel_info = nbap_get_private_data(actx->pinfo)->nbap_common_channel_info;

int i;

    if (!PINFO_FD_VISITED(actx->pinfo)){
        /* Set port to zero use that as an indication of whether we have data or not */
        for (i = 0; i < maxNrOfCommonMACFlows; i++) {
            nbap_common_channel_info[i].crnc_port = 0;
            nbap_common_channel_info[i].rlc_mode = FP_RLC_MODE_UNKNOWN;

        }

    }
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_Common_System_InformationFDD, HSDSCH_Common_System_InformationFDD_sequence);


    if (PINFO_FD_VISITED(actx->pinfo)){
        return offset;
    }
    /* Set port to zero use that as an indication of whether we have data or not */
    clear_address(&null_addr);
    for (i = 0; i < maxNrOfCommonMACFlows; i++) {
        if (nbap_common_channel_info[i].crnc_port != 0){

            conversation = find_conversation(actx->pinfo->num, &(nbap_common_channel_info[i].crnc_address), &null_addr,
                               CONVERSATION_UDP, nbap_common_channel_info[i].crnc_port, 0, NO_ADDR_B);

            if (conversation == NULL) {
                conversation = conversation_new(actx->pinfo->num, &(nbap_common_channel_info[i].crnc_address),
                    &null_addr, CONVERSATION_UDP, nbap_common_channel_info[i].crnc_port,
                    0, NO_ADDR2|NO_PORT2);

                /* Set dissector */
                conversation_set_dissector(conversation, fp_handle);
                /*Set NBAP configuration to lower layers*/
                if(actx->pinfo->link_dir==P2P_DIR_DL){

                    umts_fp_conversation_info = wmem_new0(wmem_file_scope(), umts_fp_conversation_info_t);
                    /*Select frame type = 3 according to paragraph 5.1.6 in 3GPP TS 25.435*/
                    umts_fp_conversation_info->channel = CHANNEL_HSDSCH_COMMON;
                    umts_fp_conversation_info->division          = Division_FDD;

                    umts_fp_conversation_info->dl_frame_number   = 0;
                    umts_fp_conversation_info->ul_frame_number   = actx->pinfo->num;
                    copy_address_wmem(wmem_file_scope(), &(umts_fp_conversation_info->crnc_address), &nbap_common_channel_info[i].crnc_address);
                    umts_fp_conversation_info->crnc_port         = nbap_common_channel_info[i].crnc_port;

                    fp_hsdsch_channel_info = wmem_new0(wmem_file_scope(), fp_hsdsch_channel_info_t);
                    umts_fp_conversation_info->channel_specific_info = (void*)fp_hsdsch_channel_info;
                    fp_hsdsch_channel_info->common_macdflow_id = nbap_get_private_data(actx->pinfo)->common_macdflow_id;
                    fp_hsdsch_channel_info->hsdsch_entity = ehs;

                    umts_fp_conversation_info->num_dch_in_flow++;
                    umts_fp_conversation_info->dch_ids_in_flow_list[umts_fp_conversation_info->num_dch_in_flow -1] = i;

                    set_umts_fp_conv_data(conversation, umts_fp_conversation_info);
                }
            }
        }
    }

  return offset;
}



static int
dissect_nbap_INTEGER_0_127(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, NULL, FALSE);

  return offset;
}


static const per_sequence_t HSSCCH_Codes_sequence[] = {
  { &hf_nbap_codeNumber     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_127 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSSCCH_Codes(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSSCCH_Codes, HSSCCH_Codes_sequence);

  return offset;
}


static const per_sequence_t HSSCCH_Specific_InformationRespListFDD_sequence_of[1] = {
  { &hf_nbap_HSSCCH_Specific_InformationRespListFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HSSCCH_Codes },
};

static int
dissect_nbap_HSSCCH_Specific_InformationRespListFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HSSCCH_Specific_InformationRespListFDD, HSSCCH_Specific_InformationRespListFDD_sequence_of,
                                                  1, maxNrOfHSSCCHCodes, FALSE);

  return offset;
}


static const per_sequence_t HSDSCH_Common_System_Information_ResponseFDD_sequence[] = {
  { &hf_nbap_hsSCCH_Specific_Information_ResponseFDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSSCCH_Specific_InformationRespListFDD },
  { &hf_nbap_hARQ_MemoryPartitioning, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HARQ_MemoryPartitioning },
  { &hf_nbap_commonMACFlow_Specific_Info_Response, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CommonMACFlow_Specific_InfoList_Response },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_Common_System_Information_ResponseFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_Common_System_Information_ResponseFDD, HSDSCH_Common_System_Information_ResponseFDD_sequence);

  return offset;
}



static int
dissect_nbap_HSDSCH_MACdFlow_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  guint32 hsdsch_macdflow_id;
  guint num_items;
  gint* hsdsch_macdflow_ids;
  nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfMACdFlows_1, &hsdsch_macdflow_id, FALSE);

  nbap_private_data->hsdsch_macdflow_id = hsdsch_macdflow_id;

  num_items = nbap_private_data->num_items;
DISSECTOR_ASSERT(num_items < maxNrOfMACdFlows+1);
DISSECTOR_ASSERT(num_items > 0);

hsdsch_macdflow_ids = nbap_private_data->hsdsch_macdflow_ids;
hsdsch_macdflow_ids[num_items-1] = hsdsch_macdflow_id;


  return offset;
}


static const per_sequence_t HSDSCH_MACdFlow_Specific_InfoItem_sequence[] = {
  { &hf_nbap_hsDSCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_MACdFlow_ID },
  { &hf_nbap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AllocationRetentionPriority },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_MACdFlow_Specific_InfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    address     dst_addr;
    guint32 transportLayerAddress_ipv4;
    guint16 bindingID;
    guint32 hsdsch_macdflow_id;
    nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
    nbap_hsdsch_channel_info_t* nbap_hsdsch_channel_info;
    nbap_hsdsch_channel_info = nbap_private_data->nbap_hsdsch_channel_info;

    nbap_private_data->transportLayerAddress_ipv4 = 0;
    nbap_private_data->binding_id_port = 0;
    nbap_private_data->num_items++;
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_MACdFlow_Specific_InfoItem, HSDSCH_MACdFlow_Specific_InfoItem_sequence);


    transportLayerAddress_ipv4 = nbap_private_data->transportLayerAddress_ipv4;
    bindingID = nbap_private_data->binding_id_port;
    if (PINFO_FD_VISITED(actx->pinfo) || transportLayerAddress_ipv4 == 0 || bindingID == 0){
        return offset;
    }

    set_address(&dst_addr, AT_IPv4, 4, &transportLayerAddress_ipv4);

    /* Set address for collection of HSDSCH entries */
    hsdsch_macdflow_id = nbap_private_data->hsdsch_macdflow_id;
    copy_address_wmem(actx->pinfo->pool,&(nbap_hsdsch_channel_info[hsdsch_macdflow_id].crnc_address),&dst_addr);
    nbap_hsdsch_channel_info[hsdsch_macdflow_id].crnc_port = bindingID;


  return offset;
}


static const per_sequence_t HSDSCH_MACdFlow_Specific_InfoList_sequence_of[1] = {
  { &hf_nbap_HSDSCH_MACdFlow_Specific_InfoList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_MACdFlow_Specific_InfoItem },
};

static int
dissect_nbap_HSDSCH_MACdFlow_Specific_InfoList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HSDSCH_MACdFlow_Specific_InfoList, HSDSCH_MACdFlow_Specific_InfoList_sequence_of,
                                                  1, maxNrOfMACdFlows, FALSE);

  return offset;
}



static int
dissect_nbap_MAChsGuaranteedBitRate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 16777215U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_SID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfMACdPDUIndexes_1, NULL, FALSE);

  return offset;
}


static const per_sequence_t MACdPDU_Size_IndexItem_sequence[] = {
  { &hf_nbap_sID            , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SID },
  { &hf_nbap_macdPDU_Size   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MACdPDU_Size },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MACdPDU_Size_IndexItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MACdPDU_Size_IndexItem, MACdPDU_Size_IndexItem_sequence);

  return offset;
}


static const per_sequence_t MACdPDU_Size_Indexlist_sequence_of[1] = {
  { &hf_nbap_MACdPDU_Size_Indexlist_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_MACdPDU_Size_IndexItem },
};

static int
dissect_nbap_MACdPDU_Size_Indexlist(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_MACdPDU_Size_Indexlist, MACdPDU_Size_Indexlist_sequence_of,
                                                  1, maxNrOfMACdPDUIndexes, FALSE);

  return offset;
}


static const value_string nbap_RLC_Mode_vals[] = {
  {   0, "rLC-AM" },
  {   1, "rLC-UM" },
  { 0, NULL }
};


static int
dissect_nbap_RLC_Mode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    guint32 rlc_mode;
    nbap_hsdsch_channel_info_t* nbap_hsdsch_channel_info;
    nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
    nbap_hsdsch_channel_info = nbap_private_data->nbap_hsdsch_channel_info;
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, &rlc_mode, TRUE, 0, NULL);

    switch(rlc_mode){
        case 0:
            /* rLC-AM */
            nbap_hsdsch_channel_info[nbap_private_data->hsdsch_macdflow_id].rlc_mode = FP_RLC_AM;
            break;
        case 1:
            /* rLC-UM */
            nbap_hsdsch_channel_info[nbap_private_data->hsdsch_macdflow_id].rlc_mode = FP_RLC_UM;
            break;
        default:
            break;
    }


  return offset;
}


static const per_sequence_t PriorityQueue_InfoItem_sequence[] = {
  { &hf_nbap_priorityQueueId, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PriorityQueue_Id },
  { &hf_nbap_associatedHSDSCH_MACdFlow, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_MACdFlow_ID },
  { &hf_nbap_schedulingPriorityIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SchedulingPriorityIndicator },
  { &hf_nbap_t1             , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T1 },
  { &hf_nbap_discardTimer   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DiscardTimer },
  { &hf_nbap_mAC_hsWindowSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MAC_hsWindowSize },
  { &hf_nbap_mAChsGuaranteedBitRate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MAChsGuaranteedBitRate },
  { &hf_nbap_macdPDU_Size_Index, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MACdPDU_Size_Indexlist },
  { &hf_nbap_rLC_Mode       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RLC_Mode },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PriorityQueue_InfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    nbap_get_private_data(actx->pinfo)->num_items++;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PriorityQueue_InfoItem, PriorityQueue_InfoItem_sequence);

  return offset;
}


static const per_sequence_t PriorityQueue_InfoList_sequence_of[1] = {
  { &hf_nbap_PriorityQueue_InfoList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_PriorityQueue_InfoItem },
};

static int
dissect_nbap_PriorityQueue_InfoList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_PriorityQueue_InfoList, PriorityQueue_InfoList_sequence_of,
                                                  1, maxNrOfPriorityQueues, FALSE);

  return offset;
}


static const per_sequence_t HSDSCH_MACdFlows_Information_sequence[] = {
  { &hf_nbap_hSDSCH_MACdFlow_Specific_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_MACdFlow_Specific_InfoList },
  { &hf_nbap_priorityQueue_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PriorityQueue_InfoList },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_MACdFlows_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    int protocol_ie_id;
    guint32 i;
    nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
    nbap_private_data->num_items = 0;
    protocol_ie_id = nbap_private_data->protocol_ie_id;
    nbap_hsdsch_channel_info_t* nbap_hsdsch_channel_info;
    nbap_hsdsch_channel_info = nbap_private_data->nbap_hsdsch_channel_info;

    /*Handle special cases, when the tree is weird, ie. no useful message appears in the tree, like HSDHSCH-Information-FDD */
    switch(protocol_ie_id){
            /*This flow must also be added*/
            case id_HSDSCH_MACdFlows_to_Add:
                if (!PINFO_FD_VISITED(actx->pinfo)){
                /* Set port to zero use that as an indication of whether we have data or not */
                    for (i = 0; i < maxNrOfMACdFlows; i++) {
                        nbap_hsdsch_channel_info[i].crnc_port = 0;
                        nbap_hsdsch_channel_info[i].rlc_mode = FP_RLC_MODE_UNKNOWN;
                        /*XXX: Added 29 jun*/
                        nbap_hsdsch_channel_info[i].entity = entity_not_specified;  /* Maybe this should default to entity = hs*/
                    }
                }
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_MACdFlows_Information, HSDSCH_MACdFlows_Information_sequence);

                add_hsdsch_bind(actx->pinfo);

            break;
            default:
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_MACdFlows_Information, HSDSCH_MACdFlows_Information_sequence);

            break;
        }


  return offset;
}



static int
dissect_nbap_T_hSDSCH_Physical_Layer_Category(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    guint32 hsdsch_physical_layer_category;
    nbap_hsdsch_channel_info_t* nbap_hsdsch_channel_info;
    nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
    nbap_hsdsch_channel_info = nbap_private_data->nbap_hsdsch_channel_info;

  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 64U, &hsdsch_physical_layer_category, TRUE);


    nbap_hsdsch_channel_info[nbap_private_data->hsdsch_macdflow_id].hsdsch_physical_layer_category = hsdsch_physical_layer_category;


  return offset;
}


static const per_sequence_t UE_Capability_Information_sequence[] = {
  { &hf_nbap_hSDSCH_Physical_Layer_Category_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_hSDSCH_Physical_Layer_Category },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UE_Capability_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UE_Capability_Information, UE_Capability_Information_sequence);

  return offset;
}



static int
dissect_nbap_MAChsReorderingBufferSize_for_RLC_UM(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 300U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_HSSCCH_PowerOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t HSDSCH_FDD_Information_sequence[] = {
  { &hf_nbap_hSDSCH_MACdFlows_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_MACdFlows_Information },
  { &hf_nbap_ueCapability_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UE_Capability_Information },
  { &hf_nbap_mAChs_Reordering_Buffer_Size_for_RLC_UM, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MAChsReorderingBufferSize_for_RLC_UM },
  { &hf_nbap_cqiFeedback_CycleK, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CQI_Feedback_Cycle },
  { &hf_nbap_cqiRepetitionFactor, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CQI_RepetitionFactor },
  { &hf_nbap_ackNackRepetitionFactor, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AckNack_RepetitionFactor },
  { &hf_nbap_cqiPowerOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CQI_Power_Offset },
  { &hf_nbap_ackPowerOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Ack_Power_Offset },
  { &hf_nbap_nackPowerOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Nack_Power_Offset },
  { &hf_nbap_hsscch_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSSCCH_PowerOffset },
  { &hf_nbap_measurement_Power_Offset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Measurement_Power_Offset },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_FDD_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
/*
 * Collect the information about the HSDSCH MACdFlows set up conversation(s) and set the conversation data.
 */
    address     null_addr;
    conversation_t *conversation = NULL;
    umts_fp_conversation_info_t *umts_fp_conversation_info;
    fp_hsdsch_channel_info_t* fp_hsdsch_channel_info = NULL;
    guint32 i;
    nbap_hsdsch_channel_info_t* nbap_hsdsch_channel_info;
    nbap_hsdsch_channel_info = nbap_get_private_data(actx->pinfo)->nbap_hsdsch_channel_info;

    if (!PINFO_FD_VISITED(actx->pinfo)){
        /* Set port to zero use that as an indication of whether we have data or not */
        for (i = 0; i < maxNrOfMACdFlows; i++) {
            nbap_hsdsch_channel_info[i].crnc_port = 0;
            nbap_hsdsch_channel_info[i].rlc_mode = FP_RLC_MODE_UNKNOWN;
            /*XXX: Added 29 jun*/
            nbap_hsdsch_channel_info[i].entity = entity_not_specified;  /* Maybe this should default to entity = hs*/
        }
    }

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_FDD_Information, HSDSCH_FDD_Information_sequence);


    if (PINFO_FD_VISITED(actx->pinfo)){
        return offset;
    }

    /* Set port to zero use that as an indication of whether we have data or not */
    clear_address(&null_addr);
    for (i = 0; i < maxNrOfMACdFlows; i++) {
        if (nbap_hsdsch_channel_info[i].crnc_port != 0){
            nbap_debug("Frame %u HSDSCH-MACdFlows-Information:hsdsch_macdflow_id %u Look for conv on IP %s Port %u",
                        actx->pinfo->num,
                        i,
                        address_to_str (actx->pinfo->pool, &(nbap_hsdsch_channel_info[i].crnc_address)),
                        nbap_hsdsch_channel_info[i].crnc_port);
            conversation = find_conversation(actx->pinfo->num, &(nbap_hsdsch_channel_info[i].crnc_address), &null_addr,
                               CONVERSATION_UDP, nbap_hsdsch_channel_info[i].crnc_port, 0, NO_ADDR_B);

            if (conversation == NULL) {
                /* It's not part of any conversation - create a new one. */
                nbap_debug("Frame %u HSDSCH-MACdFlows-Information: Set up conv on Port %u", actx->pinfo->num, nbap_hsdsch_channel_info[i].crnc_port);
                conversation = conversation_new(actx->pinfo->num, &(nbap_hsdsch_channel_info[i].crnc_address),
                    &null_addr, CONVERSATION_UDP, nbap_hsdsch_channel_info[i].crnc_port,
                    0, NO_ADDR2|NO_PORT2);

                /* Set dissector */
                conversation_set_dissector(conversation, fp_handle);

                if(actx->pinfo->link_dir==P2P_DIR_DL){
                    umts_fp_conversation_info = wmem_new0(wmem_file_scope(), umts_fp_conversation_info_t);
                    /* Fill in the HSDSCH relevant data */

                    umts_fp_conversation_info->iface_type        = IuB_Interface;
                    umts_fp_conversation_info->division          = Division_FDD;
                    umts_fp_conversation_info->channel           = CHANNEL_HSDSCH;
                    umts_fp_conversation_info->dl_frame_number   = 0;
                    umts_fp_conversation_info->ul_frame_number   = actx->pinfo->num;
                    copy_address_wmem(wmem_file_scope(), &(umts_fp_conversation_info->crnc_address), &nbap_hsdsch_channel_info[i].crnc_address);
                    umts_fp_conversation_info->crnc_port         = nbap_hsdsch_channel_info[i].crnc_port;

                    fp_hsdsch_channel_info = wmem_new0(wmem_file_scope(), fp_hsdsch_channel_info_t);
                    umts_fp_conversation_info->channel_specific_info = (void*)fp_hsdsch_channel_info;
                    /*Added june 3, normally just the iterator variable*/
                    fp_hsdsch_channel_info->hsdsch_macdflow_id = i ; /*hsdsch_macdflow_ids[i];*/ /* hsdsch_macdflow_id;*/

                    /*Added july 2012*/
                    umts_fp_conversation_info->com_context_id = nbap_get_private_data(actx->pinfo)->com_context_id;

                    /* Cheat and use the DCH entries */
                    umts_fp_conversation_info->num_dch_in_flow++;
                    umts_fp_conversation_info->dch_ids_in_flow_list[umts_fp_conversation_info->num_dch_in_flow -1] = i;

                    /* The information collected from the 'hsdsch_physical_layer_category' field
                     * is used here to tell if the UE supports MAC-ehs or MAC-hs.
                     * This logic is based on this line in TS 25.306 / Section 5.1 :
                     * "... A UE that supports categories greater or equal to category 13, also supports MAC-ehs."
                     */
                    if(nbap_hsdsch_channel_info[i].entity == entity_not_specified ){

                        if(nbap_hsdsch_channel_info[i].hsdsch_physical_layer_category > 12){
                            fp_hsdsch_channel_info->hsdsch_entity = ehs;
                        }else{
                            fp_hsdsch_channel_info->hsdsch_entity = hs;
                        }
                    }else{
                        fp_hsdsch_channel_info->hsdsch_entity = (enum fp_hsdsch_entity)nbap_hsdsch_channel_info[i].entity;
                    }
                    umts_fp_conversation_info->rlc_mode = nbap_hsdsch_channel_info[i].rlc_mode;
                    set_umts_fp_conv_data(conversation, umts_fp_conversation_info);
                }
            }
        }
    }


  return offset;
}



static int
dissect_nbap_TDD_AckNack_Power_Offset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -7, 8U, NULL, TRUE);

  return offset;
}


static const per_sequence_t HSDSCH_TDD_Information_sequence[] = {
  { &hf_nbap_hSDSCH_MACdFlows_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_MACdFlows_Information },
  { &hf_nbap_ueCapability_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UE_Capability_Information },
  { &hf_nbap_mAChs_Reordering_Buffer_Size_for_RLC_UM, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MAChsReorderingBufferSize_for_RLC_UM },
  { &hf_nbap_tDD_AckNack_Power_Offset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_AckNack_Power_Offset },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_TDD_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_TDD_Information, HSDSCH_TDD_Information_sequence);

  return offset;
}


static const per_sequence_t HSDSCH_MACdFlow_Specific_InfoItem_to_Modify_sequence[] = {
  { &hf_nbap_hsDSCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_MACdFlow_ID },
  { &hf_nbap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_AllocationRetentionPriority },
  { &hf_nbap_transportBearerRequestIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportBearerRequestIndicator },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_MACdFlow_Specific_InfoItem_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    address     dst_addr;
    guint32 transportLayerAddress_ipv4;
    guint16 bindingID;
    nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
    nbap_private_data->transportLayerAddress_ipv4 = 0;
    nbap_private_data->binding_id_port = 0;
    nbap_private_data->num_items++;
    nbap_hsdsch_channel_info_t* nbap_hsdsch_channel_info;
    nbap_hsdsch_channel_info = nbap_private_data->nbap_hsdsch_channel_info;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_MACdFlow_Specific_InfoItem_to_Modify, HSDSCH_MACdFlow_Specific_InfoItem_to_Modify_sequence);


    transportLayerAddress_ipv4 = nbap_private_data->transportLayerAddress_ipv4;
    bindingID = nbap_private_data->binding_id_port;
    if (PINFO_FD_VISITED(actx->pinfo) || transportLayerAddress_ipv4 == 0 || bindingID == 0){
        return offset;
    }

    set_address(&dst_addr, AT_IPv4, 4, &transportLayerAddress_ipv4);

    /* Set address for collection of HSDSCH entries */
    copy_address_wmem(actx->pinfo->pool,&(nbap_hsdsch_channel_info[nbap_private_data->hsdsch_macdflow_id].crnc_address),&dst_addr);
    nbap_hsdsch_channel_info[nbap_private_data->hsdsch_macdflow_id].crnc_port = bindingID;


  return offset;
}


static const per_sequence_t HSDSCH_MACdFlow_Specific_InfoList_to_Modify_sequence_of[1] = {
  { &hf_nbap_HSDSCH_MACdFlow_Specific_InfoList_to_Modify_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_MACdFlow_Specific_InfoItem_to_Modify },
};

static int
dissect_nbap_HSDSCH_MACdFlow_Specific_InfoList_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HSDSCH_MACdFlow_Specific_InfoList_to_Modify, HSDSCH_MACdFlow_Specific_InfoList_to_Modify_sequence_of,
                                                  1, maxNrOfMACdFlows, FALSE);

  return offset;
}


static const per_sequence_t PriorityQueue_InfoItem_to_Add_sequence[] = {
  { &hf_nbap_priorityQueueId, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PriorityQueue_Id },
  { &hf_nbap_associatedHSDSCH_MACdFlow, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_MACdFlow_ID },
  { &hf_nbap_schedulingPriorityIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SchedulingPriorityIndicator },
  { &hf_nbap_t1             , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T1 },
  { &hf_nbap_discardTimer   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DiscardTimer },
  { &hf_nbap_mAC_hsWindowSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MAC_hsWindowSize },
  { &hf_nbap_mAChsGuaranteedBitRate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MAChsGuaranteedBitRate },
  { &hf_nbap_macdPDU_Size_Index, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MACdPDU_Size_Indexlist },
  { &hf_nbap_rLC_Mode       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RLC_Mode },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PriorityQueue_InfoItem_to_Add(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    nbap_get_private_data(actx->pinfo)->num_items = 1;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PriorityQueue_InfoItem_to_Add, PriorityQueue_InfoItem_to_Add_sequence);

  return offset;
}


static const per_sequence_t MACdPDU_Size_IndexItem_to_Modify_sequence[] = {
  { &hf_nbap_sID            , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SID },
  { &hf_nbap_macdPDU_Size   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MACdPDU_Size },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MACdPDU_Size_IndexItem_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MACdPDU_Size_IndexItem_to_Modify, MACdPDU_Size_IndexItem_to_Modify_sequence);

  return offset;
}


static const per_sequence_t MACdPDU_Size_Indexlist_to_Modify_sequence_of[1] = {
  { &hf_nbap_MACdPDU_Size_Indexlist_to_Modify_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_MACdPDU_Size_IndexItem_to_Modify },
};

static int
dissect_nbap_MACdPDU_Size_Indexlist_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_MACdPDU_Size_Indexlist_to_Modify, MACdPDU_Size_Indexlist_to_Modify_sequence_of,
                                                  1, maxNrOfMACdPDUIndexes, FALSE);

  return offset;
}


static const per_sequence_t PriorityQueue_InfoItem_to_Modify_sequence[] = {
  { &hf_nbap_priorityQueueId, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PriorityQueue_Id },
  { &hf_nbap_schedulingPriorityIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SchedulingPriorityIndicator },
  { &hf_nbap_t1             , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_T1 },
  { &hf_nbap_discardTimer   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DiscardTimer },
  { &hf_nbap_mAC_hsWindowSize, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MAC_hsWindowSize },
  { &hf_nbap_mAChsGuaranteedBitRate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MAChsGuaranteedBitRate },
  { &hf_nbap_macdPDU_Size_Index_to_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MACdPDU_Size_Indexlist_to_Modify },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PriorityQueue_InfoItem_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PriorityQueue_InfoItem_to_Modify, PriorityQueue_InfoItem_to_Modify_sequence);

  return offset;
}


static const value_string nbap_ModifyPriorityQueue_vals[] = {
  {   0, "addPriorityQueue" },
  {   1, "modifyPriorityQueue" },
  {   2, "deletePriorityQueue" },
  { 0, NULL }
};

static const per_choice_t ModifyPriorityQueue_choice[] = {
  {   0, &hf_nbap_addPriorityQueue, ASN1_EXTENSION_ROOT    , dissect_nbap_PriorityQueue_InfoItem_to_Add },
  {   1, &hf_nbap_modifyPriorityQueue, ASN1_EXTENSION_ROOT    , dissect_nbap_PriorityQueue_InfoItem_to_Modify },
  {   2, &hf_nbap_deletePriorityQueue, ASN1_EXTENSION_ROOT    , dissect_nbap_PriorityQueue_Id },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_ModifyPriorityQueue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_ModifyPriorityQueue, ModifyPriorityQueue_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t PriorityQueue_InfoList_to_Modify_sequence_of[1] = {
  { &hf_nbap_PriorityQueue_InfoList_to_Modify_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ModifyPriorityQueue },
};

static int
dissect_nbap_PriorityQueue_InfoList_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_PriorityQueue_InfoList_to_Modify, PriorityQueue_InfoList_to_Modify_sequence_of,
                                                  1, maxNrOfPriorityQueues, FALSE);

  return offset;
}


static const value_string nbap_HSSCCH_Code_Change_Grant_vals[] = {
  {   0, "changeGranted" },
  { 0, NULL }
};


static int
dissect_nbap_HSSCCH_Code_Change_Grant(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t HSDSCH_Information_to_Modify_sequence[] = {
  { &hf_nbap_hsDSCH_MACdFlow_Specific_Info_to_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSDSCH_MACdFlow_Specific_InfoList_to_Modify },
  { &hf_nbap_priorityQueueInfotoModify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PriorityQueue_InfoList_to_Modify },
  { &hf_nbap_mAChs_Reordering_Buffer_Size_for_RLC_UM, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MAChsReorderingBufferSize_for_RLC_UM },
  { &hf_nbap_cqiFeedback_CycleK, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CQI_Feedback_Cycle },
  { &hf_nbap_cqiRepetitionFactor, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CQI_RepetitionFactor },
  { &hf_nbap_ackNackRepetitionFactor, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_AckNack_RepetitionFactor },
  { &hf_nbap_cqiPowerOffset , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CQI_Power_Offset },
  { &hf_nbap_ackPowerOffset , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Ack_Power_Offset },
  { &hf_nbap_nackPowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Nack_Power_Offset },
  { &hf_nbap_hsscch_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSSCCH_PowerOffset },
  { &hf_nbap_measurement_Power_Offset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Measurement_Power_Offset },
  { &hf_nbap_hSSCCHCodeChangeGrant, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSSCCH_Code_Change_Grant },
  { &hf_nbap_tDDAckNackPowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_AckNack_Power_Offset },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_Information_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
/*
 * This is pretty much the same like if we setup a previous flow
 * Collect the information about the HSDSCH MACdFlows set up conversation(s) and set the conversation data.
 */
    address     null_addr;
    conversation_t *conversation = NULL;
    umts_fp_conversation_info_t *umts_fp_conversation_info;
    fp_hsdsch_channel_info_t* fp_hsdsch_channel_info = NULL;
    guint32 i;
    nbap_hsdsch_channel_info_t* nbap_hsdsch_channel_info;
    nbap_hsdsch_channel_info = nbap_get_private_data(actx->pinfo)->nbap_hsdsch_channel_info;

    if (!PINFO_FD_VISITED(actx->pinfo)){
        /* Set port to zero use that as an indication of whether we have data or not */
        for (i = 0; i < maxNrOfMACdFlows; i++) {
            nbap_hsdsch_channel_info[i].crnc_port = 0;
            nbap_hsdsch_channel_info[i].rlc_mode = FP_RLC_MODE_UNKNOWN;
            /*XXX: Added 29 jun*/
            nbap_hsdsch_channel_info[i].entity = entity_not_specified;  /* Maybe this should default to entity = hs*/
        }
    }

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_Information_to_Modify, HSDSCH_Information_to_Modify_sequence);


    if (PINFO_FD_VISITED(actx->pinfo)){
        return offset;
    }

    /* Set port to zero use that as an indication of whether we have data or not */
    clear_address(&null_addr);

    nbap_debug("Frame %u HSDSCH-MACdFlows-Information Start",
        actx->pinfo->num);

    for (i = 0; i < maxNrOfMACdFlows; i++) {
        if (nbap_hsdsch_channel_info[i].crnc_port != 0){
            nbap_debug("    hsdsch_macdflow_id %u Look for conv on IP %s Port %u",
                        i,
                        address_to_str (actx->pinfo->pool, &(nbap_hsdsch_channel_info[i].crnc_address)),
                        nbap_hsdsch_channel_info[i].crnc_port);
            conversation = find_conversation(actx->pinfo->num, &(nbap_hsdsch_channel_info[i].crnc_address), &null_addr,
                               CONVERSATION_UDP, nbap_hsdsch_channel_info[i].crnc_port, 0, NO_ADDR_B);

            if (conversation == NULL) {
                /* It's not part of any conversation - create a new one. */
                nbap_debug("    Set up conv on Port %u", nbap_hsdsch_channel_info[i].crnc_port);

                conversation = conversation_new(actx->pinfo->num, &(nbap_hsdsch_channel_info[i].crnc_address),
                    &null_addr, CONVERSATION_UDP, nbap_hsdsch_channel_info[i].crnc_port,
                    0, NO_ADDR2|NO_PORT2);

                /* Set dissector */
                conversation_set_dissector(conversation, fp_handle);

                if(actx->pinfo->link_dir==P2P_DIR_DL){
                    umts_fp_conversation_info = wmem_new0(wmem_file_scope(), umts_fp_conversation_info_t);
                    /* Fill in the HSDSCH relevant data */

                    umts_fp_conversation_info->iface_type        = IuB_Interface;
                    umts_fp_conversation_info->division          = Division_FDD;
                    umts_fp_conversation_info->channel           = CHANNEL_HSDSCH;
                    umts_fp_conversation_info->dl_frame_number   = 0;
                    umts_fp_conversation_info->ul_frame_number   = actx->pinfo->num;
                    copy_address_wmem(wmem_file_scope(), &(umts_fp_conversation_info->crnc_address), &nbap_hsdsch_channel_info[i].crnc_address);
                    umts_fp_conversation_info->crnc_port         = nbap_hsdsch_channel_info[i].crnc_port;

                    fp_hsdsch_channel_info = wmem_new0(wmem_file_scope(), fp_hsdsch_channel_info_t);
                    umts_fp_conversation_info->channel_specific_info = (void*)fp_hsdsch_channel_info;
                    /*Added june 3, normally just the iterator variable*/
                    fp_hsdsch_channel_info->hsdsch_macdflow_id = i ; /*hsdsch_macdflow_ids[i];*/ /* hsdsch_macdflow_id;*/

                    /*Added july 2012*/
                    umts_fp_conversation_info->com_context_id = nbap_get_private_data(actx->pinfo)->com_context_id;

                    /* Cheat and use the DCH entries */
                    umts_fp_conversation_info->num_dch_in_flow++;
                    umts_fp_conversation_info->dch_ids_in_flow_list[umts_fp_conversation_info->num_dch_in_flow -1] = i;

                    /* The information collected from the 'hsdsch_physical_layer_category' field
                     * is used here to tell if the UE supports MAC-ehs or MAC-hs.
                     * This logic is based on this line in TS 25.306 / Section 5.1 :
                     * "... A UE that supports categories greater or equal to category 13, also supports MAC-ehs."
                     */
                    if(nbap_hsdsch_channel_info[i].entity == entity_not_specified ){

                        if(nbap_hsdsch_channel_info[i].hsdsch_physical_layer_category > 12){
                            fp_hsdsch_channel_info->hsdsch_entity = ehs;
                        }else{
                            fp_hsdsch_channel_info->hsdsch_entity = hs;
                        }
                    }else{
                        fp_hsdsch_channel_info->hsdsch_entity = (enum fp_hsdsch_entity)nbap_hsdsch_channel_info[i].entity;
                    }

                    umts_fp_conversation_info->rlc_mode = nbap_hsdsch_channel_info[i].rlc_mode;

                    /*Indicate that the frag table has to be reset*/
                    umts_fp_conversation_info->reset_frag = TRUE;

                    set_umts_fp_conv_data(conversation, umts_fp_conversation_info);
                }
            }
        }
        nbap_debug("Frame %u HSDSCH-MACdFlows-Information End",
            actx->pinfo->num);

    }


  return offset;
}


static const value_string nbap_HSDSCH_MACdPDUSizeFormat_vals[] = {
  {   0, "indexedMACdPDU-Size" },
  {   1, "flexibleMACdPDU-Size" },
  { 0, NULL }
};


static int
dissect_nbap_HSDSCH_MACdPDUSizeFormat(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_HSDSCH_MACdPDU_SizeCapability_vals[] = {
  {   0, "indexedSizeCapable" },
  {   1, "flexibleSizeCapable" },
  { 0, NULL }
};


static int
dissect_nbap_HSDSCH_MACdPDU_SizeCapability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t PriorityQueue_InfoItem_to_Modify_Unsynchronised_sequence[] = {
  { &hf_nbap_priorityQueueId, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PriorityQueue_Id },
  { &hf_nbap_schedulingPriorityIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SchedulingPriorityIndicator },
  { &hf_nbap_discardTimer   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DiscardTimer },
  { &hf_nbap_mAChsGuaranteedBitRate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MAChsGuaranteedBitRate },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PriorityQueue_InfoItem_to_Modify_Unsynchronised(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PriorityQueue_InfoItem_to_Modify_Unsynchronised, PriorityQueue_InfoItem_to_Modify_Unsynchronised_sequence);

  return offset;
}


static const per_sequence_t PriorityQueue_InfoList_to_Modify_Unsynchronised_sequence_of[1] = {
  { &hf_nbap_PriorityQueue_InfoList_to_Modify_Unsynchronised_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_PriorityQueue_InfoItem_to_Modify_Unsynchronised },
};

static int
dissect_nbap_PriorityQueue_InfoList_to_Modify_Unsynchronised(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_PriorityQueue_InfoList_to_Modify_Unsynchronised, PriorityQueue_InfoList_to_Modify_Unsynchronised_sequence_of,
                                                  1, maxNrOfPriorityQueues, FALSE);

  return offset;
}


static const per_sequence_t HSDSCH_Information_to_Modify_Unsynchronised_sequence[] = {
  { &hf_nbap_hsDSCH_MACdFlow_Specific_Info_to_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSDSCH_MACdFlow_Specific_InfoList_to_Modify },
  { &hf_nbap_priorityQueueInfotoModifyUnsynchronised, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PriorityQueue_InfoList_to_Modify_Unsynchronised },
  { &hf_nbap_cqiPowerOffset , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CQI_Power_Offset },
  { &hf_nbap_ackPowerOffset , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Ack_Power_Offset },
  { &hf_nbap_nackPowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Nack_Power_Offset },
  { &hf_nbap_hsscch_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSSCCH_PowerOffset },
  { &hf_nbap_tDDAckNackPowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_AckNack_Power_Offset },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_Information_to_Modify_Unsynchronised(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_Information_to_Modify_Unsynchronised, HSDSCH_Information_to_Modify_Unsynchronised_sequence);

  return offset;
}


static const per_sequence_t HSDSCH_MACdFlow_Specific_InformationResp_Item_sequence[] = {
  { &hf_nbap_hsDSCHMacdFlow_Id, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_MACdFlow_ID },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_hSDSCH_Initial_Capacity_Allocation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSDSCH_Initial_Capacity_Allocation },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_MACdFlow_Specific_InformationResp_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    nbap_get_private_data(actx->pinfo)->num_items++;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_MACdFlow_Specific_InformationResp_Item, HSDSCH_MACdFlow_Specific_InformationResp_Item_sequence);

  return offset;
}


static const per_sequence_t HSDSCH_MACdFlow_Specific_InformationResp_sequence_of[1] = {
  { &hf_nbap_HSDSCH_MACdFlow_Specific_InformationResp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_MACdFlow_Specific_InformationResp_Item },
};

static int
dissect_nbap_HSDSCH_MACdFlow_Specific_InformationResp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HSDSCH_MACdFlow_Specific_InformationResp, HSDSCH_MACdFlow_Specific_InformationResp_sequence_of,
                                                  1, maxNrOfMACdFlows, FALSE);

  return offset;
}


static const per_sequence_t HSDSCH_FDD_Information_Response_sequence[] = {
  { &hf_nbap_hsDSCH_MACdFlow_Specific_InformationResp, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSDSCH_MACdFlow_Specific_InformationResp },
  { &hf_nbap_hsSCCH_Specific_Information_ResponseFDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSSCCH_Specific_InformationRespListFDD },
  { &hf_nbap_hARQ_MemoryPartitioning, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HARQ_MemoryPartitioning },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_FDD_Information_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_FDD_Information_Response, HSDSCH_FDD_Information_Response_sequence);

  return offset;
}


static const value_string nbap_SixtyfourQAM_UsageAllowedIndicator_vals[] = {
  {   0, "allowed" },
  {   1, "not-allowed" },
  { 0, NULL }
};


static int
dissect_nbap_SixtyfourQAM_UsageAllowedIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t HS_DSCH_FDD_Secondary_Serving_Information_sequence[] = {
  { &hf_nbap_hsscch_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSSCCH_PowerOffset },
  { &hf_nbap_measurement_Power_Offset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Measurement_Power_Offset },
  { &hf_nbap_sixtyfourQAM_UsageAllowedIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SixtyfourQAM_UsageAllowedIndicator },
  { &hf_nbap_hSDSCH_RNTI    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_RNTI },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_DSCH_FDD_Secondary_Serving_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_DSCH_FDD_Secondary_Serving_Information, HS_DSCH_FDD_Secondary_Serving_Information_sequence);

  return offset;
}


static const value_string nbap_SixtyfourQAM_DL_UsageIndicator_vals[] = {
  {   0, "sixtyfourQAM-DL-used" },
  {   1, "sixtyfourQAM-DL-not-used" },
  { 0, NULL }
};


static int
dissect_nbap_SixtyfourQAM_DL_UsageIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_HSDSCH_TBSizeTableIndicator_vals[] = {
  {   0, "octet-aligned" },
  { 0, NULL }
};


static int
dissect_nbap_HSDSCH_TBSizeTableIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t HS_DSCH_FDD_Secondary_Serving_Information_Response_sequence[] = {
  { &hf_nbap_hsSCCH_Specific_Information_ResponseFDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSSCCH_Specific_InformationRespListFDD },
  { &hf_nbap_sixtyfourQAM_DL_UsageIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SixtyfourQAM_DL_UsageIndicator },
  { &hf_nbap_hSDSCH_TBSizeTableIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSDSCH_TBSizeTableIndicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_DSCH_FDD_Secondary_Serving_Information_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_DSCH_FDD_Secondary_Serving_Information_Response, HS_DSCH_FDD_Secondary_Serving_Information_Response_sequence);

  return offset;
}


static const per_sequence_t HS_DSCH_Secondary_Serving_Information_To_Modify_sequence[] = {
  { &hf_nbap_hsscch_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSSCCH_PowerOffset },
  { &hf_nbap_measurement_Power_Offset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Measurement_Power_Offset },
  { &hf_nbap_hSSCCH_CodeChangeGrant, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSSCCH_Code_Change_Grant },
  { &hf_nbap_sixtyfourQAM_UsageAllowedIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SixtyfourQAM_UsageAllowedIndicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_DSCH_Secondary_Serving_Information_To_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_DSCH_Secondary_Serving_Information_To_Modify, HS_DSCH_Secondary_Serving_Information_To_Modify_sequence);

  return offset;
}


static const per_sequence_t HS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised_sequence[] = {
  { &hf_nbap_hsscch_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSSCCH_PowerOffset },
  { &hf_nbap_sixtyfourQAM_UsageAllowedIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SixtyfourQAM_UsageAllowedIndicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised, HS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised_sequence);

  return offset;
}


static const value_string nbap_HSSCCH_CodeChangeIndicator_vals[] = {
  {   0, "hsSCCHCodeChangeNeeded" },
  { 0, NULL }
};


static int
dissect_nbap_HSSCCH_CodeChangeIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_HS_PDSCH_Code_Change_Indicator_vals[] = {
  {   0, "hsPDSCHCodeChangeNeeded" },
  { 0, NULL }
};


static int
dissect_nbap_HS_PDSCH_Code_Change_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t HS_DSCH_FDD_Secondary_Serving_Update_Information_sequence[] = {
  { &hf_nbap_hsSCCHCodeChangeIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSSCCH_CodeChangeIndicator },
  { &hf_nbap_hS_PDSCH_Code_Change_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_PDSCH_Code_Change_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_DSCH_FDD_Secondary_Serving_Update_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_DSCH_FDD_Secondary_Serving_Update_Information, HS_DSCH_FDD_Secondary_Serving_Update_Information_sequence);

  return offset;
}


static const per_sequence_t HS_Secondary_Serving_cell_change_successful_sequence[] = {
  { &hf_nbap_hS_DSCH_FDD_Secondary_Serving_Information_Response, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCH_FDD_Secondary_Serving_Information_Response },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_Secondary_Serving_cell_change_successful(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_Secondary_Serving_cell_change_successful, HS_Secondary_Serving_cell_change_successful_sequence);

  return offset;
}


static const per_sequence_t HS_Secondary_Serving_cell_change_unsuccessful_sequence[] = {
  { &hf_nbap_cause          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Cause },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_Secondary_Serving_cell_change_unsuccessful(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_Secondary_Serving_cell_change_unsuccessful, HS_Secondary_Serving_cell_change_unsuccessful_sequence);

  return offset;
}


static const value_string nbap_HS_DSCH_Secondary_Serving_cell_change_choice_vals[] = {
  {   0, "hS-Secondary-Serving-cell-change-successful" },
  {   1, "hS-Secondary-Serving-cell-change-unsuccessful" },
  { 0, NULL }
};

static const per_choice_t HS_DSCH_Secondary_Serving_cell_change_choice_choice[] = {
  {   0, &hf_nbap_hS_Secondary_Serving_cell_change_successful, ASN1_EXTENSION_ROOT    , dissect_nbap_HS_Secondary_Serving_cell_change_successful },
  {   1, &hf_nbap_hS_Secondary_Serving_cell_change_unsuccessful, ASN1_EXTENSION_ROOT    , dissect_nbap_HS_Secondary_Serving_cell_change_unsuccessful },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_HS_DSCH_Secondary_Serving_cell_change_choice(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_HS_DSCH_Secondary_Serving_cell_change_choice, HS_DSCH_Secondary_Serving_cell_change_choice_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t HS_DSCH_Secondary_Serving_Cell_Change_Information_Response_sequence[] = {
  { &hf_nbap_hS_DSCH_Secondary_Serving_cell_choice, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCH_Secondary_Serving_cell_change_choice },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_DSCH_Secondary_Serving_Cell_Change_Information_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_DSCH_Secondary_Serving_Cell_Change_Information_Response, HS_DSCH_Secondary_Serving_Cell_Change_Information_Response_sequence);

  return offset;
}



static int
dissect_nbap_HS_DSCH_Secondary_Serving_Remove(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_Paging_MACFlow_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfPagingMACFlow_1, NULL, FALSE);

  return offset;
}


static const per_sequence_t Paging_MACFlow_PriorityQueue_Item_sequence[] = {
  { &hf_nbap_priority_Queue_Information_for_Enhanced_PCH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Priority_Queue_Information_for_Enhanced_FACH_PCH },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Paging_MACFlow_PriorityQueue_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Paging_MACFlow_PriorityQueue_Item, Paging_MACFlow_PriorityQueue_Item_sequence);

  return offset;
}


static const per_sequence_t Paging_MACFlow_PriorityQueue_Information_sequence_of[1] = {
  { &hf_nbap_Paging_MACFlow_PriorityQueue_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Paging_MACFlow_PriorityQueue_Item },
};

static int
dissect_nbap_Paging_MACFlow_PriorityQueue_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Paging_MACFlow_PriorityQueue_Information, Paging_MACFlow_PriorityQueue_Information_sequence_of,
                                                  1, maxNrOfpagingMACQueues, FALSE);

  return offset;
}


static const per_sequence_t Paging_MAC_Flow_Specific_Information_Item_sequence[] = {
  { &hf_nbap_paging_MACFlow_Id, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Paging_MACFlow_ID },
  { &hf_nbap_hSDPA_associated_PICH_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSDPA_Associated_PICH_Information },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_tnl_qos        , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TnlQos },
  { &hf_nbap_toAWS          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ToAWS },
  { &hf_nbap_toAWE          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ToAWE },
  { &hf_nbap_paging_MACFlow_PriorityQueue_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Paging_MACFlow_PriorityQueue_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Paging_MAC_Flow_Specific_Information_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Paging_MAC_Flow_Specific_Information_Item, Paging_MAC_Flow_Specific_Information_Item_sequence);

  return offset;
}


static const per_sequence_t Paging_MACFlow_Specific_Information_sequence_of[1] = {
  { &hf_nbap_Paging_MACFlow_Specific_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Paging_MAC_Flow_Specific_Information_Item },
};

static int
dissect_nbap_Paging_MACFlow_Specific_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Paging_MACFlow_Specific_Information, Paging_MACFlow_Specific_Information_sequence_of,
                                                  1, maxNrOfPagingMACFlow, FALSE);

  return offset;
}



static int
dissect_nbap_Number_of_PCCH_transmission(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 5U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_Transport_Block_Size_Index_for_Enhanced_PCH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 32U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Transport_Block_Size_List_item_sequence[] = {
  { &hf_nbap_transport_Block_Size_Index_for_Enhanced_PCH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Transport_Block_Size_Index_for_Enhanced_PCH },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Transport_Block_Size_List_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Transport_Block_Size_List_item, Transport_Block_Size_List_item_sequence);

  return offset;
}


static const per_sequence_t Transport_Block_Size_List_sequence_of[1] = {
  { &hf_nbap_Transport_Block_Size_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Transport_Block_Size_List_item },
};

static int
dissect_nbap_Transport_Block_Size_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Transport_Block_Size_List, Transport_Block_Size_List_sequence_of,
                                                  1, maxNrOfHS_DSCHTBSsE_PCH, FALSE);

  return offset;
}


static const per_sequence_t HSDSCH_Paging_System_InformationFDD_sequence[] = {
  { &hf_nbap_paging_MACFlow_Specific_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Paging_MACFlow_Specific_Information },
  { &hf_nbap_hSSCCH_Power   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_hSPDSCH_Power  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_number_of_PCCH_transmission, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Number_of_PCCH_transmission },
  { &hf_nbap_transport_Block_Size_List, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Transport_Block_Size_List },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_Paging_System_InformationFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  /*
  ws_warning("HS-DSCH Type 3 NOT Implemented!");
  */
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_Paging_System_InformationFDD, HSDSCH_Paging_System_InformationFDD_sequence);




  return offset;
}



static int
dissect_nbap_HSPDSCH_Code_Index(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxHS_PDSCHCodeNrComp_1, NULL, FALSE);

  return offset;
}


static const per_sequence_t HSDSCH_Paging_System_Information_ResponseList_sequence[] = {
  { &hf_nbap_pagingMACFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Paging_MACFlow_ID },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_hSPDSCH_Code_Index, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSPDSCH_Code_Index },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_Paging_System_Information_ResponseList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_Paging_System_Information_ResponseList, HSDSCH_Paging_System_Information_ResponseList_sequence);

  return offset;
}


static const per_sequence_t HSDSCH_Paging_System_Information_ResponseFDD_sequence_of[1] = {
  { &hf_nbap_HSDSCH_Paging_System_Information_ResponseFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_Paging_System_Information_ResponseList },
};

static int
dissect_nbap_HSDSCH_Paging_System_Information_ResponseFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HSDSCH_Paging_System_Information_ResponseFDD, HSDSCH_Paging_System_Information_ResponseFDD_sequence_of,
                                                  1, maxNrOfPagingMACFlow, FALSE);

  return offset;
}



static int
dissect_nbap_HS_SICH_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 31U, NULL, FALSE);

  return offset;
}


static const per_sequence_t HSSICH_Info_sequence[] = {
  { &hf_nbap_hsSICH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SICH_ID },
  { &hf_nbap_timeslot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_tDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSSICH_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSSICH_Info, HSSICH_Info_sequence);

  return offset;
}


static const per_sequence_t HSSCCH_Specific_InformationRespItemTDD_sequence[] = {
  { &hf_nbap_timeslot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_tDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_hSSICH_Info    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSSICH_Info },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSSCCH_Specific_InformationRespItemTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSSCCH_Specific_InformationRespItemTDD, HSSCCH_Specific_InformationRespItemTDD_sequence);

  return offset;
}


static const per_sequence_t HSSCCH_Specific_InformationRespListTDD_sequence_of[1] = {
  { &hf_nbap_HSSCCH_Specific_InformationRespListTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HSSCCH_Specific_InformationRespItemTDD },
};

static int
dissect_nbap_HSSCCH_Specific_InformationRespListTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HSSCCH_Specific_InformationRespListTDD, HSSCCH_Specific_InformationRespListTDD_sequence_of,
                                                  1, maxNrOfHSSCCHCodes, FALSE);

  return offset;
}


static const per_sequence_t HSSICH_InfoLCR_sequence[] = {
  { &hf_nbap_hsSICH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SICH_ID },
  { &hf_nbap_timeslotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_tDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSSICH_InfoLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSSICH_InfoLCR, HSSICH_InfoLCR_sequence);

  return offset;
}


static const per_sequence_t HSSCCH_Specific_InformationRespItemTDDLCR_sequence[] = {
  { &hf_nbap_timeslotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_first_TDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_second_TDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_hSSICH_InfoLCR , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSSICH_InfoLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSSCCH_Specific_InformationRespItemTDDLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSSCCH_Specific_InformationRespItemTDDLCR, HSSCCH_Specific_InformationRespItemTDDLCR_sequence);

  return offset;
}


static const per_sequence_t HSSCCH_Specific_InformationRespListTDDLCR_sequence_of[1] = {
  { &hf_nbap_HSSCCH_Specific_InformationRespListTDDLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HSSCCH_Specific_InformationRespItemTDDLCR },
};

static int
dissect_nbap_HSSCCH_Specific_InformationRespListTDDLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HSSCCH_Specific_InformationRespListTDDLCR, HSSCCH_Specific_InformationRespListTDDLCR_sequence_of,
                                                  1, maxNrOfHSSCCHCodes, FALSE);

  return offset;
}


static const per_sequence_t HSDSCH_TDD_Information_Response_sequence[] = {
  { &hf_nbap_hsDSCH_MACdFlow_Specific_InformationResp, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSDSCH_MACdFlow_Specific_InformationResp },
  { &hf_nbap_hsSCCH_Specific_Information_ResponseTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSSCCH_Specific_InformationRespListTDD },
  { &hf_nbap_hsSCCH_Specific_Information_ResponseTDDLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSSCCH_Specific_InformationRespListTDDLCR },
  { &hf_nbap_hARQ_MemoryPartitioning, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HARQ_MemoryPartitioning },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_TDD_Information_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_TDD_Information_Response, HSDSCH_TDD_Information_Response_sequence);

  return offset;
}


static const per_sequence_t HSDSCH_MACdFlows_to_Delete_Item_sequence[] = {
  { &hf_nbap_hsDSCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_MACdFlow_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_MACdFlows_to_Delete_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    nbap_get_private_data(actx->pinfo)->num_items++;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_MACdFlows_to_Delete_Item, HSDSCH_MACdFlows_to_Delete_Item_sequence);

  return offset;
}


static const per_sequence_t HSDSCH_MACdFlows_to_Delete_sequence_of[1] = {
  { &hf_nbap_HSDSCH_MACdFlows_to_Delete_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_MACdFlows_to_Delete_Item },
};

static int
dissect_nbap_HSDSCH_MACdFlows_to_Delete(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    nbap_get_private_data(actx->pinfo)->num_items = 0;

  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HSDSCH_MACdFlows_to_Delete, HSDSCH_MACdFlows_to_Delete_sequence_of,
                                                  1, maxNrOfMACdFlows, FALSE);

  return offset;
}



static int
dissect_nbap_HS_SCCH_CodeNumber(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, NULL, FALSE);

  return offset;
}


static const per_sequence_t HS_SCCH_PreconfiguredCodesItem_sequence[] = {
  { &hf_nbap_hS_SCCH_CodeNumber, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SCCH_CodeNumber },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_SCCH_PreconfiguredCodesItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_SCCH_PreconfiguredCodesItem, HS_SCCH_PreconfiguredCodesItem_sequence);

  return offset;
}


static const per_sequence_t HS_SCCH_PreconfiguredCodes_sequence_of[1] = {
  { &hf_nbap_HS_SCCH_PreconfiguredCodes_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SCCH_PreconfiguredCodesItem },
};

static int
dissect_nbap_HS_SCCH_PreconfiguredCodes(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HS_SCCH_PreconfiguredCodes, HS_SCCH_PreconfiguredCodes_sequence_of,
                                                  1, maxNrOfHSSCCHCodes, FALSE);

  return offset;
}


static const per_sequence_t SetsOfHS_SCCH_CodesItem_sequence[] = {
  { &hf_nbap_hS_SCCH_PreconfiguredCodes, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SCCH_PreconfiguredCodes },
  { &hf_nbap_sixtyfourQAM_DL_UsageIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SixtyfourQAM_DL_UsageIndicator },
  { &hf_nbap_hSDSCH_TBSizeTableIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSDSCH_TBSizeTableIndicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SetsOfHS_SCCH_CodesItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SetsOfHS_SCCH_CodesItem, SetsOfHS_SCCH_CodesItem_sequence);

  return offset;
}


static const per_sequence_t SetsOfHS_SCCH_Codes_sequence_of[1] = {
  { &hf_nbap_SetsOfHS_SCCH_Codes_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_SetsOfHS_SCCH_CodesItem },
};

static int
dissect_nbap_SetsOfHS_SCCH_Codes(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_SetsOfHS_SCCH_Codes, SetsOfHS_SCCH_Codes_sequence_of,
                                                  1, maxNrOfHSDSCH, FALSE);

  return offset;
}


static const value_string nbap_MIMO_N_M_Ratio_vals[] = {
  {   0, "v1-2" },
  {   1, "v2-3" },
  {   2, "v3-4" },
  {   3, "v4-5" },
  {   4, "v5-6" },
  {   5, "v6-7" },
  {   6, "v7-8" },
  {   7, "v8-9" },
  {   8, "v9-10" },
  {   9, "v1-1" },
  { 0, NULL }
};


static int
dissect_nbap_MIMO_N_M_Ratio(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     10, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t HSDSCH_PreconfigurationInfo_sequence[] = {
  { &hf_nbap_setsOfHS_SCCH_Codes, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SetsOfHS_SCCH_Codes },
  { &hf_nbap_hARQ_MemoryPartitioning, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HARQ_MemoryPartitioning },
  { &hf_nbap_e_DCH_FDD_DL_Control_Channel_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DCH_FDD_DL_Control_Channel_Information },
  { &hf_nbap_hARQ_Preamble_Mode_Activation_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HARQ_Preamble_Mode_Activation_Indicator },
  { &hf_nbap_mIMO_N_M_Ratio , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MIMO_N_M_Ratio },
  { &hf_nbap_continuousPacketConnectivityHS_SCCH_less_Information_Response, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ContinuousPacketConnectivityHS_SCCH_less_Information_Response },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_PreconfigurationInfo(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_PreconfigurationInfo, HSDSCH_PreconfigurationInfo_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_Preconfiguration_Information_ItemIEs_sequence[] = {
  { &hf_nbap_e_DCH_FDD_DL_Control_Channel_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_FDD_DL_Control_Channel_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_Preconfiguration_Information_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_Preconfiguration_Information_ItemIEs, Additional_EDCH_Preconfiguration_Information_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_Preconfiguration_Information_sequence_of[1] = {
  { &hf_nbap_Additional_EDCH_Preconfiguration_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_EDCH_Preconfiguration_Information_ItemIEs },
};

static int
dissect_nbap_Additional_EDCH_Preconfiguration_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Additional_EDCH_Preconfiguration_Information, Additional_EDCH_Preconfiguration_Information_sequence_of,
                                                  1, maxNrOfEDCH_1, FALSE);

  return offset;
}


static const value_string nbap_MAChsResetScheme_vals[] = {
  {   0, "always" },
  {   1, "interNodeB-change" },
  { 0, NULL }
};


static int
dissect_nbap_MAChsResetScheme(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_INTEGER_1_64_(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 64U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_NumHS_SCCH_Codes(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxNrOfHSSCCHCodes, NULL, FALSE);

  return offset;
}


static const per_sequence_t SecondaryServingCellsItem_sequence[] = {
  { &hf_nbap_secondaryC_ID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_C_ID },
  { &hf_nbap_numSecondaryHS_SCCH_Codes, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NumHS_SCCH_Codes },
  { &hf_nbap_sixtyfourQAM_UsageAllowedIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SixtyfourQAM_UsageAllowedIndicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SecondaryServingCellsItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SecondaryServingCellsItem, SecondaryServingCellsItem_sequence);

  return offset;
}


static const per_sequence_t SecondaryServingCells_sequence_of[1] = {
  { &hf_nbap_SecondaryServingCells_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_SecondaryServingCellsItem },
};

static int
dissect_nbap_SecondaryServingCells(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_SecondaryServingCells, SecondaryServingCells_sequence_of,
                                                  1, maxNrOfHSDSCH_1, FALSE);

  return offset;
}



static int
dissect_nbap_MIMO_ActivationIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t HSDSCH_PreconfigurationSetup_sequence[] = {
  { &hf_nbap_mAChsResetScheme, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MAChsResetScheme },
  { &hf_nbap_hSDSCH_Physical_Layer_Category, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_1_64_ },
  { &hf_nbap_mAChs_Reordering_Buffer_Size_for_RLC_UM, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MAChsReorderingBufferSize_for_RLC_UM },
  { &hf_nbap_secondaryServingCells, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SecondaryServingCells },
  { &hf_nbap_numPrimaryHS_SCCH_Codes, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NumHS_SCCH_Codes },
  { &hf_nbap_hARQ_Preamble_Mode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HARQ_Preamble_Mode },
  { &hf_nbap_mIMO_ActivationIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MIMO_ActivationIndicator },
  { &hf_nbap_hSDSCH_MACdPDUSizeFormat, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSDSCH_MACdPDUSizeFormat },
  { &hf_nbap_sixtyfourQAM_UsageAllowedIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SixtyfourQAM_UsageAllowedIndicator },
  { &hf_nbap_uE_with_enhanced_HS_SCCH_support_indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_continuousPacketConnectivityHS_SCCH_less_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ContinuousPacketConnectivityHS_SCCH_less_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_PreconfigurationSetup(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_PreconfigurationSetup, HSDSCH_PreconfigurationSetup_sequence);

  return offset;
}


static const per_sequence_t HSSICH_Info768_sequence[] = {
  { &hf_nbap_hsSICH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SICH_ID },
  { &hf_nbap_timeslot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType768 },
  { &hf_nbap_tDD_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode768 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSSICH_Info768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSSICH_Info768, HSSICH_Info768_sequence);

  return offset;
}


static const per_sequence_t HSSCCH_Specific_InformationRespItemTDD768_sequence[] = {
  { &hf_nbap_timeslot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType768 },
  { &hf_nbap_tDD_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode768 },
  { &hf_nbap_hSSICH_Info768 , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSSICH_Info768 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSSCCH_Specific_InformationRespItemTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSSCCH_Specific_InformationRespItemTDD768, HSSCCH_Specific_InformationRespItemTDD768_sequence);

  return offset;
}


static const per_sequence_t HSSCCH_Specific_InformationRespListTDD768_sequence_of[1] = {
  { &hf_nbap_HSSCCH_Specific_InformationRespListTDD768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HSSCCH_Specific_InformationRespItemTDD768 },
};

static int
dissect_nbap_HSSCCH_Specific_InformationRespListTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HSSCCH_Specific_InformationRespListTDD768, HSSCCH_Specific_InformationRespListTDD768_sequence_of,
                                                  1, maxNrOfHSSCCHCodes, FALSE);

  return offset;
}



static int
dissect_nbap_HS_SICH_failed(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 20U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_HS_SICH_missed(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 20U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_HS_SICH_total(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 20U, NULL, FALSE);

  return offset;
}


static const per_sequence_t HS_SICH_Reception_Quality_Value_sequence[] = {
  { &hf_nbap_failed_HS_SICH , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SICH_failed },
  { &hf_nbap_missed_HS_SICH , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SICH_missed },
  { &hf_nbap_total_HS_SICH  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SICH_total },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_SICH_Reception_Quality_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_SICH_Reception_Quality_Value, HS_SICH_Reception_Quality_Value_sequence);

  return offset;
}



static int
dissect_nbap_HS_SICH_Reception_Quality_Measurement_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 20U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_INTEGER_0_maxHS_PDSCHCodeNrComp_1(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxHS_PDSCHCodeNrComp_1, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_HS_PDSCH_Start_code_number(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxHS_PDSCHCodeNrComp_1, NULL, FALSE);

  return offset;
}


static const per_sequence_t HS_PDSCH_FDD_Code_Information_sequence[] = {
  { &hf_nbap_number_of_HS_PDSCH_codes, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_maxHS_PDSCHCodeNrComp_1 },
  { &hf_nbap_hS_PDSCH_Start_code_number, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_PDSCH_Start_code_number },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_PDSCH_FDD_Code_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_PDSCH_FDD_Code_Information, HS_PDSCH_FDD_Code_Information_sequence);

  return offset;
}



static int
dissect_nbap_HS_SCCH_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 31U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_HS_SCCH_FDD_Code_Information_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxHS_SCCHCodeNrComp_1, NULL, FALSE);

  return offset;
}


static const per_sequence_t HS_SCCH_FDD_Code_List_sequence_of[1] = {
  { &hf_nbap_HS_SCCH_FDD_Code_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SCCH_FDD_Code_Information_Item },
};

static int
dissect_nbap_HS_SCCH_FDD_Code_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HS_SCCH_FDD_Code_List, HS_SCCH_FDD_Code_List_sequence_of,
                                                  1, maxNrOfHSSCCHs, FALSE);

  return offset;
}


static const value_string nbap_HS_SCCH_FDD_Code_Information_vals[] = {
  {   0, "replace" },
  {   1, "remove" },
  { 0, NULL }
};

static const per_choice_t HS_SCCH_FDD_Code_Information_choice[] = {
  {   0, &hf_nbap_replace_02     , ASN1_EXTENSION_ROOT    , dissect_nbap_HS_SCCH_FDD_Code_List },
  {   1, &hf_nbap_remove         , ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_HS_SCCH_FDD_Code_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_HS_SCCH_FDD_Code_Information, HS_SCCH_FDD_Code_Information_choice,
                                 NULL);

  return offset;
}


static const value_string nbap_HS_PDSCH_Code_Change_Grant_vals[] = {
  {   0, "changeGranted" },
  { 0, NULL }
};


static int
dissect_nbap_HS_PDSCH_Code_Change_Grant(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_HSDSCH_Configured_Indicator_vals[] = {
  {   0, "configured-HS-DSCH" },
  {   1, "no-configured-HS-DSCH" },
  { 0, NULL }
};


static int
dissect_nbap_HSDSCH_Configured_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t HS_DSCH_Serving_Cell_Change_Info_sequence[] = {
  { &hf_nbap_hspdsch_RL_ID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_hSDSCH_FDD_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSDSCH_FDD_Information },
  { &hf_nbap_hsdsch_RNTI    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_RNTI },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_DSCH_Serving_Cell_Change_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_DSCH_Serving_Cell_Change_Info, HS_DSCH_Serving_Cell_Change_Info_sequence);

  return offset;
}


static const per_sequence_t HS_serving_cell_change_successful_sequence[] = {
  { &hf_nbap_hSDSCH_FDD_Information_Response, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_FDD_Information_Response },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_serving_cell_change_successful(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_serving_cell_change_successful, HS_serving_cell_change_successful_sequence);

  return offset;
}


static const per_sequence_t HS_serving_cell_change_unsuccessful_sequence[] = {
  { &hf_nbap_cause          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Cause },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_serving_cell_change_unsuccessful(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_serving_cell_change_unsuccessful, HS_serving_cell_change_unsuccessful_sequence);

  return offset;
}


static const value_string nbap_HS_DSCH_serving_cell_choice_vals[] = {
  {   0, "hS-serving-cell-change-successful" },
  {   1, "hS-serving-cell-change-unsuccessful" },
  { 0, NULL }
};

static const per_choice_t HS_DSCH_serving_cell_choice_choice[] = {
  {   0, &hf_nbap_hS_serving_cell_change_successful, ASN1_EXTENSION_ROOT    , dissect_nbap_HS_serving_cell_change_successful },
  {   1, &hf_nbap_hS_serving_cell_change_unsuccessful, ASN1_EXTENSION_ROOT    , dissect_nbap_HS_serving_cell_change_unsuccessful },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_HS_DSCH_serving_cell_choice(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_HS_DSCH_serving_cell_choice, HS_DSCH_serving_cell_choice_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t HS_DSCH_Serving_Cell_Change_Info_Response_sequence[] = {
  { &hf_nbap_hS_DSCH_serving_cell_choice, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCH_serving_cell_choice },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_DSCH_Serving_Cell_Change_Info_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_DSCH_Serving_Cell_Change_Info_Response, HS_DSCH_Serving_Cell_Change_Info_Response_sequence);

  return offset;
}


static const per_sequence_t HSDSCH_FDD_Update_Information_sequence[] = {
  { &hf_nbap_hsSCCHCodeChangeIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSSCCH_CodeChangeIndicator },
  { &hf_nbap_cqiFeedback_CycleK, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CQI_Feedback_Cycle },
  { &hf_nbap_cqiRepetitionFactor, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CQI_RepetitionFactor },
  { &hf_nbap_ackNackRepetitionFactor, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_AckNack_RepetitionFactor },
  { &hf_nbap_cqiPowerOffset , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CQI_Power_Offset },
  { &hf_nbap_ackPowerOffset , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Ack_Power_Offset },
  { &hf_nbap_nackPowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Nack_Power_Offset },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_FDD_Update_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_FDD_Update_Information, HSDSCH_FDD_Update_Information_sequence);

  return offset;
}


static const per_sequence_t HSDSCH_TDD_Update_Information_sequence[] = {
  { &hf_nbap_hsSCCHCodeChangeIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSSCCH_CodeChangeIndicator },
  { &hf_nbap_tDDAckNackPowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_AckNack_Power_Offset },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_TDD_Update_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_TDD_Update_Information, HSDSCH_TDD_Update_Information_sequence);

  return offset;
}



static int
dissect_nbap_TDD_PhysicalChannelOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 63U, NULL, FALSE);

  return offset;
}


static const value_string nbap_PagingIndicatorLength_vals[] = {
  {   0, "v2" },
  {   1, "v4" },
  {   2, "v8" },
  { 0, NULL }
};


static int
dissect_nbap_PagingIndicatorLength(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t HSDPA_PICH_notShared_with_PCHLCR_sequence[] = {
  { &hf_nbap_hSDPA_PICH_notShared_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_tdd_ChannelisationCodeLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCodeLCR },
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_tdd_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_pagingIndicatorLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PagingIndicatorLength },
  { &hf_nbap_pICH_Power     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PICH_Power },
  { &hf_nbap_second_TDD_ChannelisationCodeLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCodeLCR },
  { &hf_nbap_sttd_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_STTD_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDPA_PICH_notShared_with_PCHLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDPA_PICH_notShared_with_PCHLCR, HSDPA_PICH_notShared_with_PCHLCR_sequence);

  return offset;
}


static const value_string nbap_HSDPA_Associated_PICH_InformationLCR_vals[] = {
  {   0, "hsdpa-PICH-Shared-with-PCH" },
  {   1, "hsdpa-PICH-notShared-with-PCHLCR" },
  { 0, NULL }
};

static const per_choice_t HSDPA_Associated_PICH_InformationLCR_choice[] = {
  {   0, &hf_nbap_hsdpa_PICH_Shared_with_PCH, ASN1_EXTENSION_ROOT    , dissect_nbap_HSDPA_PICH_Shared_with_PCH },
  {   1, &hf_nbap_hsdpa_PICH_notShared_with_PCHLCR, ASN1_EXTENSION_ROOT    , dissect_nbap_HSDPA_PICH_notShared_with_PCHLCR },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_HSDPA_Associated_PICH_InformationLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_HSDPA_Associated_PICH_InformationLCR, HSDPA_Associated_PICH_InformationLCR_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t HSDSCH_Common_InformationLCR_sequence[] = {
  { &hf_nbap_cCCH_PriorityQueue_Id, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PriorityQueue_Id },
  { &hf_nbap_sRB1_PriorityQueue_Id, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PriorityQueue_Id },
  { &hf_nbap_associatedCommon_MACFlowLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Common_MACFlow_ID_LCR },
  { &hf_nbap_fACH_Measurement_Occasion_Cycle_Length_Coefficient, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_FACH_Measurement_Occasion_Cycle_Length_Coefficient },
  { &hf_nbap_bCCH_Specific_HSDSCH_RNTI_InformationLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BCCH_Specific_HSDSCH_RNTI_InformationLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_Common_InformationLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_Common_InformationLCR, HSDSCH_Common_InformationLCR_sequence);

  return offset;
}


static const per_sequence_t HSDSCH_Common_System_InformationLCR_sequence[] = {
  { &hf_nbap_hsdsch_Common_InformationLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSDSCH_Common_InformationLCR },
  { &hf_nbap_commonMACFlow_Specific_InformationLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CommonMACFlow_Specific_InfoListLCR },
  { &hf_nbap_common_H_RNTI_InformationLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Common_H_RNTI_InformationLCR },
  { &hf_nbap_sync_InformationLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Sync_InformationLCR },
  { &hf_nbap_tDD_AckNack_Power_Offset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_AckNack_Power_Offset },
  { &hf_nbap_hSSICH_SIRTarget, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_SIR },
  { &hf_nbap_hSSICH_TPC_StepSize, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_TPC_UplinkStepSize_LCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_Common_System_InformationLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_Common_System_InformationLCR, HSDSCH_Common_System_InformationLCR_sequence);

  return offset;
}



static int
dissect_nbap_HS_SCCH_ID_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t HSSCCH_Specific_InformationRespItemLCR_sequence[] = {
  { &hf_nbap_hS_SCCH_ID_LCR , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SCCH_ID_LCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSSCCH_Specific_InformationRespItemLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSSCCH_Specific_InformationRespItemLCR, HSSCCH_Specific_InformationRespItemLCR_sequence);

  return offset;
}


static const per_sequence_t HSSCCH_Specific_InformationRespListLCR_sequence_of[1] = {
  { &hf_nbap_HSSCCH_Specific_InformationRespListLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HSSCCH_Specific_InformationRespItemLCR },
};

static int
dissect_nbap_HSSCCH_Specific_InformationRespListLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HSSCCH_Specific_InformationRespListLCR, HSSCCH_Specific_InformationRespListLCR_sequence_of,
                                                  1, maxNrOfHSSCCHsLCR, FALSE);

  return offset;
}


static const per_sequence_t HSDSCH_Common_System_Information_ResponseLCR_sequence[] = {
  { &hf_nbap_hsSCCH_Specific_Information_ResponseLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSSCCH_Specific_InformationRespListLCR },
  { &hf_nbap_hARQ_MemoryPartitioning, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HARQ_MemoryPartitioning },
  { &hf_nbap_commonMACFlow_Specific_Info_ResponseLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CommonMACFlow_Specific_InfoList_ResponseLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_Common_System_Information_ResponseLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_Common_System_Information_ResponseLCR, HSDSCH_Common_System_Information_ResponseLCR_sequence);

  return offset;
}


static const per_sequence_t Paging_MAC_Flow_Specific_Information_ItemLCR_sequence[] = {
  { &hf_nbap_paging_MACFlow_Id, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Paging_MACFlow_ID },
  { &hf_nbap_hSDPA_associated_PICH_InfoLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSDPA_Associated_PICH_InformationLCR },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_tnl_qos        , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TnlQos },
  { &hf_nbap_toAWS          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ToAWS },
  { &hf_nbap_toAWE          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ToAWE },
  { &hf_nbap_paging_MACFlow_PriorityQueue_InformationLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Paging_MACFlow_PriorityQueue_Information },
  { &hf_nbap_transportBearerRequestIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportBearerRequestIndicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Paging_MAC_Flow_Specific_Information_ItemLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Paging_MAC_Flow_Specific_Information_ItemLCR, Paging_MAC_Flow_Specific_Information_ItemLCR_sequence);

  return offset;
}


static const per_sequence_t Paging_MACFlow_Specific_InformationLCR_sequence_of[1] = {
  { &hf_nbap_Paging_MACFlow_Specific_InformationLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Paging_MAC_Flow_Specific_Information_ItemLCR },
};

static int
dissect_nbap_Paging_MACFlow_Specific_InformationLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Paging_MACFlow_Specific_InformationLCR, Paging_MACFlow_Specific_InformationLCR_sequence_of,
                                                  1, maxNrOfPagingMACFlow, FALSE);

  return offset;
}



static int
dissect_nbap_INTEGER_1_16(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 16U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_INTEGER_1_8(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 8U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_INTEGER_1_3(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 3U, NULL, FALSE);

  return offset;
}


static const per_sequence_t HSDSCH_Paging_System_InformationLCR_sequence[] = {
  { &hf_nbap_paging_MACFlow_Specific_InformationLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Paging_MACFlow_Specific_InformationLCR },
  { &hf_nbap_hSSCCH_Power   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_hSPDSCH_Power  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_reception_Window_Size, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_INTEGER_1_16 },
  { &hf_nbap_n_PCH          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_INTEGER_1_8 },
  { &hf_nbap_paging_Subchannel_Size, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_INTEGER_1_3 },
  { &hf_nbap_transport_Block_Size_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Transport_Block_Size_List },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_Paging_System_InformationLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_Paging_System_InformationLCR, HSDSCH_Paging_System_InformationLCR_sequence);

  return offset;
}


static const per_sequence_t HSDSCH_Paging_System_Information_ResponseListLCR_sequence[] = {
  { &hf_nbap_pagingMACFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Paging_MACFlow_ID },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_dL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_Paging_System_Information_ResponseListLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_Paging_System_Information_ResponseListLCR, HSDSCH_Paging_System_Information_ResponseListLCR_sequence);

  return offset;
}


static const per_sequence_t HSDSCH_Paging_System_Information_ResponseLCR_sequence_of[1] = {
  { &hf_nbap_HSDSCH_Paging_System_Information_ResponseLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_Paging_System_Information_ResponseListLCR },
};

static int
dissect_nbap_HSDSCH_Paging_System_Information_ResponseLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HSDSCH_Paging_System_Information_ResponseLCR, HSDSCH_Paging_System_Information_ResponseLCR_sequence_of,
                                                  1, maxNrOfPagingMACFlow, FALSE);

  return offset;
}



static int
dissect_nbap_HS_SICH_Mapping_Index(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNoOf_HS_SICH_SPS_1, NULL, FALSE);

  return offset;
}


static const per_sequence_t HS_SCCH_Associated_HS_SICH_sequence[] = {
  { &hf_nbap_hsSICH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SICH_ID },
  { &hf_nbap_extended_HS_SICH_ID, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Extended_HS_SICH_ID },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_SCCH_Associated_HS_SICH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_SCCH_Associated_HS_SICH, HS_SCCH_Associated_HS_SICH_sequence);

  return offset;
}



static int
dissect_nbap_Non_HS_SCCH_Aassociated_HS_SICH_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Non_HS_SCCH_Associated_HS_SICH_sequence[] = {
  { &hf_nbap_non_HS_SCCH_Aassociated_HS_SICH_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Non_HS_SCCH_Aassociated_HS_SICH_ID },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Non_HS_SCCH_Associated_HS_SICH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Non_HS_SCCH_Associated_HS_SICH, Non_HS_SCCH_Associated_HS_SICH_sequence);

  return offset;
}


static const value_string nbap_HS_SICH_Type_vals[] = {
  {   0, "hS-SCCH-Associated-HS-SICH" },
  {   1, "non-HS-SCCH-Associated-HS-SICH" },
  { 0, NULL }
};

static const per_choice_t HS_SICH_Type_choice[] = {
  {   0, &hf_nbap_hS_SCCH_Associated_HS_SICH, ASN1_EXTENSION_ROOT    , dissect_nbap_HS_SCCH_Associated_HS_SICH },
  {   1, &hf_nbap_non_HS_SCCH_Associated_HS_SICH, ASN1_EXTENSION_ROOT    , dissect_nbap_Non_HS_SCCH_Associated_HS_SICH },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_HS_SICH_Type(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_HS_SICH_Type, HS_SICH_Type_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t HS_SICH_InformationItem_for_HS_DSCH_SPS_sequence[] = {
  { &hf_nbap_hS_SICH_Mapping_Index, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_SICH_Mapping_Index },
  { &hf_nbap_hS_SICH_Type   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SICH_Type },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_SICH_InformationItem_for_HS_DSCH_SPS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_SICH_InformationItem_for_HS_DSCH_SPS, HS_SICH_InformationItem_for_HS_DSCH_SPS_sequence);

  return offset;
}


static const per_sequence_t HS_SICH_InformationList_for_HS_DSCH_SPS_sequence_of[1] = {
  { &hf_nbap_HS_SICH_InformationList_for_HS_DSCH_SPS_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SICH_InformationItem_for_HS_DSCH_SPS },
};

static int
dissect_nbap_HS_SICH_InformationList_for_HS_DSCH_SPS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HS_SICH_InformationList_for_HS_DSCH_SPS, HS_SICH_InformationList_for_HS_DSCH_SPS_sequence_of,
                                                  1, maxNoOf_HS_SICH_SPS, FALSE);

  return offset;
}



static int
dissect_nbap_HS_DSCH_TimeslotResourceLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     5, 5, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const value_string nbap_ModulationSPS_LCR_vals[] = {
  {   0, "qPSK" },
  {   1, "sixteenQAM" },
  { 0, NULL }
};


static int
dissect_nbap_ModulationSPS_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t Initial_HS_PDSCH_SPS_Resource_sequence[] = {
  { &hf_nbap_repetitionPeriodIndex, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriodIndex },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RepetitionLength },
  { &hf_nbap_hS_PDSCH_Offset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_timeslot_Resource_Related_Information_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCH_TimeslotResourceLCR },
  { &hf_nbap_startCode      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_endCode        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_transport_Block_Size_Index_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Transport_Block_Size_Index_LCR },
  { &hf_nbap_modulationType , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ModulationSPS_LCR },
  { &hf_nbap_hS_SICH_Mapping_Index, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SICH_Mapping_Index },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Initial_HS_PDSCH_SPS_Resource(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Initial_HS_PDSCH_SPS_Resource, Initial_HS_PDSCH_SPS_Resource_sequence);

  return offset;
}


static const value_string nbap_Process_Memory_Size_vals[] = {
  {   0, "hms800" },
  {   1, "hms1600" },
  {   2, "hms2400" },
  {   3, "hms3200" },
  {   4, "hms4000" },
  {   5, "hms4800" },
  {   6, "hms5600" },
  {   7, "hms6400" },
  {   8, "hms7200" },
  {   9, "hms8000" },
  {  10, "hms8800" },
  {  11, "hms9600" },
  {  12, "hms10400" },
  {  13, "hms11200" },
  {  14, "hms12000" },
  {  15, "hms12800" },
  {  16, "hms13600" },
  {  17, "hms14400" },
  {  18, "hms15200" },
  {  19, "hms16000" },
  {  20, "hms17600" },
  {  21, "hms19200" },
  {  22, "hms20800" },
  {  23, "hms22400" },
  {  24, "hms24000" },
  {  25, "hms25600" },
  {  26, "hms27200" },
  {  27, "hms28800" },
  {  28, "hms30400" },
  {  29, "hms32000" },
  {  30, "hms36000" },
  {  31, "hms40000" },
  {  32, "hms44000" },
  {  33, "hms48000" },
  {  34, "hms52000" },
  {  35, "hms56000" },
  {  36, "hms60000" },
  {  37, "hms64000" },
  {  38, "hms68000" },
  {  39, "hms72000" },
  {  40, "hms76000" },
  {  41, "hms80000" },
  {  42, "hms88000" },
  {  43, "hms96000" },
  {  44, "hms104000" },
  {  45, "hms112000" },
  {  46, "hms120000" },
  {  47, "hms128000" },
  {  48, "hms136000" },
  {  49, "hms144000" },
  {  50, "hms152000" },
  {  51, "hms160000" },
  {  52, "hms176000" },
  {  53, "hms192000" },
  {  54, "hms208000" },
  {  55, "hms224000" },
  {  56, "hms240000" },
  {  57, "hms256000" },
  {  58, "hms272000" },
  {  59, "hms288000" },
  {  60, "hms304000" },
  { 0, NULL }
};

static value_string_ext nbap_Process_Memory_Size_vals_ext = VALUE_STRING_EXT_INIT(nbap_Process_Memory_Size_vals);


static int
dissect_nbap_Process_Memory_Size(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     61, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_Number_of_Processes_for_HS_DSCH_SPS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 16U, NULL, FALSE);

  return offset;
}


static const per_sequence_t HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR_sequence[] = {
  { &hf_nbap_hS_SICH_InformationList_for_HS_DSCH_SPS, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SICH_InformationList_for_HS_DSCH_SPS },
  { &hf_nbap_initial_HS_PDSCH_SPS_Resource, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Initial_HS_PDSCH_SPS_Resource },
  { &hf_nbap_buffer_Size_for_HS_DSCH_SPS, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Process_Memory_Size },
  { &hf_nbap_number_of_Processes_for_HS_DSCH_SPS, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Number_of_Processes_for_HS_DSCH_SPS },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR, HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR_sequence);

  return offset;
}


static const per_sequence_t Non_HS_SCCH_Associated_HS_SICH_InformationItem_sequence[] = {
  { &hf_nbap_non_HS_SCCH_Aassociated_HS_SICH_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Non_HS_SCCH_Aassociated_HS_SICH_ID },
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UARFCN },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationItem, Non_HS_SCCH_Associated_HS_SICH_InformationItem_sequence);

  return offset;
}


static const per_sequence_t Non_HS_SCCH_Associated_HS_SICH_InformationList_sequence_of[1] = {
  { &hf_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationItem },
};

static int
dissect_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationList, Non_HS_SCCH_Associated_HS_SICH_InformationList_sequence_of,
                                                  0, maxNoOfNon_HS_SCCH_Assosiated_HS_SICH, FALSE);

  return offset;
}


static const per_sequence_t Add_To_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_non_HS_SCCH_Associated_HS_SICH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationList },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Add_To_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Add_To_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst, Add_To_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext_sequence_of[1] = {
  { &hf_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationItem },
};

static int
dissect_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext, Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext_sequence_of,
                                                  0, maxNoOfNon_HS_SCCH_Assosiated_HS_SICH_Ext, FALSE);

  return offset;
}


static const per_sequence_t Modify_Non_HS_SCCH_Associated_HS_SICH_InformationItem_sequence[] = {
  { &hf_nbap_non_HS_SCCH_Aassociated_HS_SICH_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Non_HS_SCCH_Aassociated_HS_SICH_ID },
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UARFCN },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationItem, Modify_Non_HS_SCCH_Associated_HS_SICH_InformationItem_sequence);

  return offset;
}


static const per_sequence_t Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList_sequence_of[1] = {
  { &hf_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationItem },
};

static int
dissect_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList, Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList_sequence_of,
                                                  0, maxNoOfNon_HS_SCCH_Assosiated_HS_SICH, FALSE);

  return offset;
}


static const per_sequence_t Modify_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_modify_non_HS_SCCH_Associated_HS_SICH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst, Modify_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext_sequence_of[1] = {
  { &hf_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationItem },
};

static int
dissect_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext, Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext_sequence_of,
                                                  0, maxNoOfNon_HS_SCCH_Assosiated_HS_SICH_Ext, FALSE);

  return offset;
}


static const per_sequence_t Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqstItem_sequence[] = {
  { &hf_nbap_non_HS_SCCH_Aassociated_HS_SICH_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Non_HS_SCCH_Aassociated_HS_SICH_ID },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqstItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqstItem, Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqstItem_sequence);

  return offset;
}


static const per_sequence_t Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqstItem },
};

static int
dissect_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst, Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_sequence_of,
                                                  0, maxNoOfNon_HS_SCCH_Assosiated_HS_SICH, FALSE);

  return offset;
}


static const per_sequence_t Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext_sequence_of[1] = {
  { &hf_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqstItem },
};

static int
dissect_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext, Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext_sequence_of,
                                                  0, maxNoOfNon_HS_SCCH_Assosiated_HS_SICH_Ext, FALSE);

  return offset;
}


static const per_sequence_t HSSICH_ReferenceSignal_InformationLCR_sequence[] = {
  { &hf_nbap_midambleConfigurationLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleConfigurationLCR },
  { &hf_nbap_midambleShift  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_15 },
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSSICH_ReferenceSignal_InformationLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSSICH_ReferenceSignal_InformationLCR, HSSICH_ReferenceSignal_InformationLCR_sequence);

  return offset;
}


static const per_sequence_t MIMO_ReferenceSignal_InformationListLCR_sequence_of[1] = {
  { &hf_nbap_MIMO_ReferenceSignal_InformationListLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HSSICH_ReferenceSignal_InformationLCR },
};

static int
dissect_nbap_MIMO_ReferenceSignal_InformationListLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_MIMO_ReferenceSignal_InformationListLCR, MIMO_ReferenceSignal_InformationListLCR_sequence_of,
                                                  1, maxNrOfHSSCCHCodes, FALSE);

  return offset;
}


static const per_sequence_t HSSICH_ReferenceSignal_InformationModifyLCR_sequence[] = {
  { &hf_nbap_hSSICH_ReferenceSignal_InformationLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSSICH_ReferenceSignal_InformationLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSSICH_ReferenceSignal_InformationModifyLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSSICH_ReferenceSignal_InformationModifyLCR, HSSICH_ReferenceSignal_InformationModifyLCR_sequence);

  return offset;
}



static int
dissect_nbap_IB_OC_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 16U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_IB_SG_DATA(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  tvbuff_t *parameter_tvb=NULL;
  tvbuff_t *final_tvb=NULL;
  proto_item *ti;
  proto_tree *subtree;
  guint32 segment_type;
  /* reassembly variables */
  gboolean is_short = FALSE;
  guint32 total_bit_size = 0;
  guint32 total_byte_size = 0;
  nbap_ib_segment_t* nbap_ib_segment;
  wmem_list_t *list = NULL;
  wmem_list_frame_t *curr_frame;
  guint8 *final_arr;
  guint8 final_byte_off = 0;
  guint8 final_bit_off = 0x80;
  guint8 *source;
  guint32 bit_length;
  guint32 byte_off = 0;
  guint32 bit_off = 0x80;
  guint32 sources_count;
  guint8* data;
  guint32 per_length;
  guint32 first_off;
  nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);

  /* compute aligned PER length*/
  first_off = offset; /* Saving initial offset for the default body */
  offset = ((offset+7)/8)*8; /* Round to nearest byte */
  per_length = tvb_get_bits8(tvb, offset, 8);
  offset += 8;
  if ((per_length & 0x80) == 0x80) {
    if ((per_length & 0xc0) == 0x80) {
      per_length &= 0x3f;
      per_length <<= 8;
      per_length += tvb_get_bits8(tvb, offset, 8);
      offset += 8;
    } else {
      per_length = 0;
    }
  }
  offset = first_off;

  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     NO_BOUND, NO_BOUND, FALSE, NULL, 0, &parameter_tvb, NULL);


  if(!parameter_tvb)
    return offset;
  segment_type = nbap_private_data->segment_type;
  switch(segment_type) {
    case 5: /*complete-SIB */
      final_tvb = tvb_new_subset_length(parameter_tvb,0,tvb_captured_length(parameter_tvb));
      break;
    case 6: /*complete-SIB-short */
      if(preferences_ib_sg_data_encoding == IB_SG_DATA_ENC_VAR_1) {
        /* Simply skipping the first byte (containing the length) */
        final_tvb = tvb_new_subset_length(parameter_tvb, 1, tvb_captured_length(parameter_tvb)-1);
      }
      else { /* This is IB_SG_DATA_ENC_VAR_2 */
        /* No length in tvb, just take everything as is*/
        final_tvb = tvb_new_subset_length(parameter_tvb, 0, tvb_captured_length(parameter_tvb));
      }
      break;
    default: /* First, subsequent or last */
      if(preferences_ib_sg_data_encoding == IB_SG_DATA_ENC_VAR_1) {
          is_short = ( segment_type == 1 || segment_type == 4 ); /* first-short or last-short */
          nbap_ib_segment = nbap_parse_ib_sg_data_var1(actx->pinfo, parameter_tvb, is_short);
          if (nbap_ib_segment == NULL ) { /* failed to parse */
            return offset;
          }
      }
      else { /* This is IB_SG_DATA_ENC_VAR_2 */
        /* Using the per encoded length */
        data = (guint8*)tvb_memdup(actx->pinfo->pool, parameter_tvb, 0, (per_length + 7)/8);
        nbap_ib_segment = wmem_new(actx->pinfo->pool, nbap_ib_segment_t);
        nbap_ib_segment->bit_length = per_length;
        nbap_ib_segment->data = data;
      }

      list = nbap_private_data->ib_segments;
      if (!list) {
        if ( segment_type == 0 || segment_type == 1 ) { /* first or first-short */
          list = wmem_list_new(actx->pinfo->pool);
          nbap_private_data->ib_segments = list;
        }
        else {
          return offset;
        }
      }
      wmem_list_append(list,(void*)nbap_ib_segment);
      if ( segment_type <= 2 ) { /* first, first-short or subsequent */
        return offset;
      }
      break;
  }

  if ( segment_type == 3 || segment_type == 4 ) { /* last or last-short */
    /* Sum all length of all segments */
    sources_count = wmem_list_count(list);
    curr_frame = wmem_list_head(list);
    for (guint32 src_indx = 0; src_indx < sources_count; src_indx++) {
      nbap_ib_segment = (nbap_ib_segment_t*)(wmem_list_frame_data(curr_frame));
      total_bit_size += nbap_ib_segment->bit_length;
      curr_frame = wmem_list_frame_next(curr_frame);
    }

    /* Create an array large enough for all segments */
    total_byte_size = (total_bit_size+7)/8;
    final_arr = wmem_alloc0_array(actx->pinfo->pool,guint8,total_byte_size);

    /* Reassemble all segment into the final array */
    curr_frame = wmem_list_head(list);
    for (guint32 src_indx = 0; src_indx < sources_count; src_indx++) {
      nbap_ib_segment = (nbap_ib_segment_t*)(wmem_list_frame_data(curr_frame));
      source = nbap_ib_segment->data;
      bit_length = nbap_ib_segment->bit_length;
      byte_off = 0;
      bit_off = 0x80;
      for (guint32 i=0;i<bit_length;i++) {
        if (((*(source+byte_off)) & bit_off) == bit_off) {
          final_arr[final_byte_off] |= final_bit_off;
        }
        bit_off >>= 1;
        if ( bit_off == 0x00 ) {
          byte_off += 1;
          bit_off = 0x80;
        }
        final_bit_off >>= 1;
        if ( final_bit_off == 0x00 ) {
          final_byte_off += 1;
          final_bit_off = 0x80;
        }
      }
      curr_frame = wmem_list_frame_next(curr_frame);
    }

    /* Creating TVB from the reassembled data */
    final_tvb = tvb_new_child_real_data(tvb,final_arr,total_byte_size,total_byte_size);
    add_new_data_source(actx->pinfo, final_tvb, "Reassembled Information Block");

    /* Reset segments list */
    nbap_private_data->ib_segments = NULL;

    /* Add 'reassembled' item to tree */
    ti = proto_tree_add_item(tree, hf_nbap_reassembled_information_block, final_tvb, 0, total_byte_size, ENC_NA);
  }
  else { /* Complete SIB */
    ti = actx->created_item;
  }

  subtree = proto_item_add_subtree(ti, ett_nbap_ib_sg_data);
  col_set_fence(actx->pinfo->cinfo, COL_INFO);
  col_append_str(actx->pinfo->cinfo, COL_INFO," (");
  switch(nbap_private_data->ib_type){
    case 0:
      /* mIB */
      dissect_rrc_MasterInformationBlock_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 1:
      /* iB-Type: sB1 (1) */
      dissect_rrc_SysInfoTypeSB1_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 2:
      /* iB-Type: sB2 (2) */
      dissect_rrc_SysInfoTypeSB2_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 3:
      /* iB-Type: sIB1 (3) */
      dissect_rrc_SysInfoType1_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 4:
      /* iB-Type: sIB2 (4) */
      dissect_rrc_SysInfoType2_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 5:
      /* iB-Type: sIB3 (5) */
      dissect_rrc_SysInfoType3_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 6:
      /* iB-Type: sIB4 (6) */
      dissect_rrc_SysInfoType4_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 7:
      /* iB-Type: sIB5 (7) */
      dissect_rrc_SysInfoType5_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 8:
      /* iB-Type: sIB6 (8) */
      dissect_rrc_SysInfoType6_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 9:
      /* iB-Type: sIB7 (9) */
      dissect_rrc_SysInfoType7_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 13:
      /* iB-Type: sIB11 (13) */
      dissect_rrc_SysInfoType11_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 14:
      /* iB-Type: sIB12 (14) */
      dissect_rrc_SysInfoType12_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 15:
      /* iB-Type: sIB13 (15) */
      dissect_rrc_SysInfoType13_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 16:
      /* iB-Type: sIB13.1 (16) */
      dissect_rrc_SysInfoType13_1_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 17:
      /* iB-Type: sIB13.2 (17) */
      dissect_rrc_SysInfoType13_2_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 18:
      /* iB-Type: sIB13.3 (18) */
      dissect_rrc_SysInfoType13_3_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 19:
      /* iB-Type: sIB13.4 (19) */
      dissect_rrc_SysInfoType13_4_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 20:
      /* iB-Type: sIB14 (20) */
      dissect_rrc_SysInfoType14_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 21:
      /* iB-Type: sIB15 (21) */
      dissect_rrc_SysInfoType15_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 22:
      /* iB-Type: sIB15.1 (22) */
      dissect_rrc_SysInfoType15_1_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 23:
      /* iB-Type: sIB15.2 (23) */
      dissect_rrc_SysInfoType15_2_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 24:
      /* iB-Type: sIB15.3 (24) */
      dissect_rrc_SysInfoType15_3_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 25:
      /* iB-Type: sIB16 (25) */
      dissect_rrc_SysInfoType16_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 26:
      /* iB-Type: sIB17 (26) */
      dissect_rrc_SysInfoType17_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 27:
      /* iB-Type: sIB15.4 (27) */
      dissect_rrc_SysInfoType15_4_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 28:
      /* iB-Type: sIB18 (28) */
      dissect_rrc_SysInfoType18_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    case 40:
      /* iB-Type: sIB19 (40) */
      dissect_rrc_SysInfoType19_PDU(final_tvb, actx->pinfo, subtree, NULL);
      break;
    default:
      break;
    }
  col_append_str(actx->pinfo->cinfo, COL_INFO,")");
  col_set_fence(actx->pinfo->cinfo, COL_INFO);


  return offset;
}



static int
dissect_nbap_IB_SG_POS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 4094U, NULL, FALSE);

  return offset;
}


static const value_string nbap_IB_SG_REP_vals[] = {
  {   0, "rep4" },
  {   1, "rep8" },
  {   2, "rep16" },
  {   3, "rep32" },
  {   4, "rep64" },
  {   5, "rep128" },
  {   6, "rep256" },
  {   7, "rep512" },
  {   8, "rep1024" },
  {   9, "rep2048" },
  {  10, "rep4096" },
  { 0, NULL }
};


static int
dissect_nbap_IB_SG_REP(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     11, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_IB_Type_vals[] = {
  {   0, "mIB" },
  {   1, "sB1" },
  {   2, "sB2" },
  {   3, "sIB1" },
  {   4, "sIB2" },
  {   5, "sIB3" },
  {   6, "sIB4" },
  {   7, "sIB5" },
  {   8, "sIB6" },
  {   9, "sIB7" },
  {  10, "not-Used-sIB8" },
  {  11, "not-Used-sIB9" },
  {  12, "not-Used-sIB10" },
  {  13, "sIB11" },
  {  14, "sIB12" },
  {  15, "sIB13" },
  {  16, "sIB13dot1" },
  {  17, "sIB13dot2" },
  {  18, "sIB13dot3" },
  {  19, "sIB13dot4" },
  {  20, "sIB14" },
  {  21, "sIB15" },
  {  22, "sIB15dot1" },
  {  23, "sIB15dot2" },
  {  24, "sIB15dot3" },
  {  25, "sIB16" },
  {  26, "sIB17" },
  {  27, "sIB15dot4" },
  {  28, "sIB18" },
  {  29, "sIB15dot5" },
  {  30, "sIB5bis" },
  {  31, "sIB11bis" },
  {  32, "sIB15bis" },
  {  33, "sIB15dot1bis" },
  {  34, "sIB15dot2bis" },
  {  35, "sIB15dot3bis" },
  {  36, "sIB15dot6" },
  {  37, "sIB15dot7" },
  {  38, "sIB15dot8" },
  {  39, "sIB15dot2ter" },
  {  40, "sIB19" },
  { 0, NULL }
};


static int
dissect_nbap_IB_Type(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     26, &nbap_get_private_data(actx->pinfo)->ib_type, TRUE, 15, NULL);

  return offset;
}



static int
dissect_nbap_Sub_Frame_Number(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 4U, NULL, TRUE);

  return offset;
}


static const per_sequence_t IMB_Parameters_sequence[] = {
  { &hf_nbap_sub_Frame_Number, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Sub_Frame_Number },
  { &hf_nbap_fdd_dl_ChannelisationCodeNumber, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_FDD_DL_ChannelisationCodeNumber },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_IMB_Parameters(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_IMB_Parameters, IMB_Parameters_sequence);

  return offset;
}



static int
dissect_nbap_ReportPeriodicity_Scaledmin(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 60U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_ReportPeriodicity_Scaledhour(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 24U, NULL, TRUE);

  return offset;
}


static const value_string nbap_InformationReportCharacteristicsType_ReportPeriodicity_vals[] = {
  {   0, "min" },
  {   1, "hours" },
  { 0, NULL }
};

static const per_choice_t InformationReportCharacteristicsType_ReportPeriodicity_choice[] = {
  {   0, &hf_nbap_min            , ASN1_EXTENSION_ROOT    , dissect_nbap_ReportPeriodicity_Scaledmin },
  {   1, &hf_nbap_hours          , ASN1_EXTENSION_ROOT    , dissect_nbap_ReportPeriodicity_Scaledhour },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_InformationReportCharacteristicsType_ReportPeriodicity(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_InformationReportCharacteristicsType_ReportPeriodicity, InformationReportCharacteristicsType_ReportPeriodicity_choice,
                                 NULL);

  return offset;
}


static const value_string nbap_InformationThresholds_vals[] = {
  {   0, "dgps" },
  {   1, "dGANSSThreshold" },
  { 0, NULL }
};

static const per_choice_t InformationThresholds_choice[] = {
  {   0, &hf_nbap_dgps           , ASN1_EXTENSION_ROOT    , dissect_nbap_DGPSThresholds },
  {   1, &hf_nbap_dGANSSThreshold, ASN1_NOT_EXTENSION_ROOT, dissect_nbap_DGANSSThreshold },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_InformationThresholds(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_InformationThresholds, InformationThresholds_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t InformationReportCharacteristicsType_OnModification_sequence[] = {
  { &hf_nbap_information_thresholds, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_InformationThresholds },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_InformationReportCharacteristicsType_OnModification(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_InformationReportCharacteristicsType_OnModification, InformationReportCharacteristicsType_OnModification_sequence);

  return offset;
}


static const value_string nbap_InformationReportCharacteristics_vals[] = {
  {   0, "onDemand" },
  {   1, "periodic" },
  {   2, "onModification" },
  { 0, NULL }
};

static const per_choice_t InformationReportCharacteristics_choice[] = {
  {   0, &hf_nbap_onDemand       , ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {   1, &hf_nbap_periodic       , ASN1_EXTENSION_ROOT    , dissect_nbap_InformationReportCharacteristicsType_ReportPeriodicity },
  {   2, &hf_nbap_onModification , ASN1_EXTENSION_ROOT    , dissect_nbap_InformationReportCharacteristicsType_OnModification },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_InformationReportCharacteristics(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_InformationReportCharacteristics, InformationReportCharacteristics_choice,
                                 NULL);

  return offset;
}



static int
dissect_nbap_InformationExchangeID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 1048575U, NULL, FALSE);

  return offset;
}


static const value_string nbap_Information_Type_Item_vals[] = {
  {   0, "gpsinformation" },
  {   1, "dgpscorrections" },
  {   2, "gpsrxpos" },
  {   3, "gANSSInformation" },
  {   4, "dGANSSCorrections" },
  {   5, "gANSS-RX-Pos" },
  { 0, NULL }
};


static int
dissect_nbap_Information_Type_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 3, NULL);

  return offset;
}


static const per_sequence_t InformationType_sequence[] = {
  { &hf_nbap_information_Type_Item, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Information_Type_Item },
  { &hf_nbap_gPSInformation , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GPS_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_InformationType(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_InformationType, InformationType_sequence);

  return offset;
}


static const value_string nbap_Initial_DL_DPCH_TimingAdjustment_Allowed_vals[] = {
  {   0, "initial-DL-DPCH-TimingAdjustment-Allowed" },
  { 0, NULL }
};


static int
dissect_nbap_Initial_DL_DPCH_TimingAdjustment_Allowed(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_IPDL_Indicator_vals[] = {
  {   0, "active" },
  {   1, "inactive" },
  { 0, NULL }
};


static int
dissect_nbap_IPDL_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_T_iP_SpacingFDD_vals[] = {
  {   0, "sp5" },
  {   1, "sp7" },
  {   2, "sp10" },
  {   3, "sp15" },
  {   4, "sp20" },
  {   5, "sp30" },
  {   6, "sp40" },
  {   7, "sp50" },
  { 0, NULL }
};


static int
dissect_nbap_T_iP_SpacingFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_T_iP_Length_vals[] = {
  {   0, "len5" },
  {   1, "len10" },
  { 0, NULL }
};


static int
dissect_nbap_T_iP_Length(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_INTEGER_10_25(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            10U, 25U, NULL, FALSE);

  return offset;
}


static const per_sequence_t BurstModeParams_sequence[] = {
  { &hf_nbap_burstStart     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_15 },
  { &hf_nbap_burstLength    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_10_25 },
  { &hf_nbap_burstFreq      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_1_16 },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_BurstModeParams(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_BurstModeParams, BurstModeParams_sequence);

  return offset;
}



static int
dissect_nbap_INTEGER_0_9(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 9U, NULL, FALSE);

  return offset;
}


static const per_sequence_t IPDL_FDD_Parameters_sequence[] = {
  { &hf_nbap_iP_SpacingFDD  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_iP_SpacingFDD },
  { &hf_nbap_iP_Length      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_iP_Length },
  { &hf_nbap_seed           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_63 },
  { &hf_nbap_burstModeParams, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BurstModeParams },
  { &hf_nbap_iP_Offset      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_9 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_IPDL_FDD_Parameters(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_IPDL_FDD_Parameters, IPDL_FDD_Parameters_sequence);

  return offset;
}


static const value_string nbap_T_iP_SpacingTDD_vals[] = {
  {   0, "sp30" },
  {   1, "sp40" },
  {   2, "sp50" },
  {   3, "sp70" },
  {   4, "sp100" },
  { 0, NULL }
};


static int
dissect_nbap_T_iP_SpacingTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     5, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_INTEGER_0_4095(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 4095U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_INTEGER_0_14(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 14U, NULL, FALSE);

  return offset;
}


static const value_string nbap_T_iP_PCCPCH_vals[] = {
  {   0, "switchOff-1-Frame" },
  {   1, "switchOff-2-Frames" },
  { 0, NULL }
};


static int
dissect_nbap_T_iP_PCCPCH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t IPDL_TDD_Parameters_sequence[] = {
  { &hf_nbap_iP_SpacingTDD  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_iP_SpacingTDD },
  { &hf_nbap_iP_Start       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_4095 },
  { &hf_nbap_iP_Slot        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_14 },
  { &hf_nbap_iP_PCCPCH      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_iP_PCCPCH },
  { &hf_nbap_burstModeParams, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BurstModeParams },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_IPDL_TDD_Parameters(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_IPDL_TDD_Parameters, IPDL_TDD_Parameters_sequence);

  return offset;
}


static const value_string nbap_T_iP_SpacingTDD_01_vals[] = {
  {   0, "sp30" },
  {   1, "sp40" },
  {   2, "sp50" },
  {   3, "sp70" },
  {   4, "sp100" },
  { 0, NULL }
};


static int
dissect_nbap_T_iP_SpacingTDD_01(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     5, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_T_iP_Sub_vals[] = {
  {   0, "first" },
  {   1, "second" },
  {   2, "both" },
  { 0, NULL }
};


static int
dissect_nbap_T_iP_Sub(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t IPDL_TDD_Parameters_LCR_sequence[] = {
  { &hf_nbap_iP_SpacingTDD_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_iP_SpacingTDD_01 },
  { &hf_nbap_iP_Start       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_4095 },
  { &hf_nbap_iP_Sub         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_iP_Sub },
  { &hf_nbap_burstModeParams, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BurstModeParams },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_IPDL_TDD_Parameters_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_IPDL_TDD_Parameters_LCR, IPDL_TDD_Parameters_LCR_sequence);

  return offset;
}


static const per_sequence_t IPMulticastIndication_sequence[] = {
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportLayerAddress },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BindingID },
  { &hf_nbap_cFNOffset      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_255 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_IPMulticastIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_IPMulticastIndication, IPMulticastIndication_sequence);

  return offset;
}



static int
dissect_nbap_IPMulticastDataBearerIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_boolean(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}


static const value_string nbap_T_idleIntervalInfo_k_vals[] = {
  {   0, "none" },
  {   2, "two" },
  {   3, "three" },
  { 0, NULL }
};


static int
dissect_nbap_T_idleIntervalInfo_k(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 3U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_INTEGER_0_7(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 7U, NULL, FALSE);

  return offset;
}


static const per_sequence_t IdleIntervalInformation_sequence[] = {
  { &hf_nbap_idleIntervalInfo_k, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_idleIntervalInfo_k },
  { &hf_nbap_idleIntervalInfo_offset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_7 },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_IdleIntervalInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_IdleIntervalInformation, IdleIntervalInformation_sequence);

  return offset;
}



static int
dissect_nbap_LTGI_Presence(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_boolean(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}



static int
dissect_nbap_INTEGER_1_6(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 6U, NULL, FALSE);

  return offset;
}


static const value_string nbap_T_maxPhysChPerTimeslot_vals[] = {
  {   0, "one" },
  {   1, "two" },
  {   2, "three" },
  {   3, "four" },
  { 0, NULL }
};


static int
dissect_nbap_T_maxPhysChPerTimeslot(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 2, NULL);

  return offset;
}


static const per_sequence_t LCRTDD_Uplink_Physical_Channel_Capability_sequence[] = {
  { &hf_nbap_maxTimeslotsPerSubFrame, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_1_6 },
  { &hf_nbap_maxPhysChPerTimeslot, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_maxPhysChPerTimeslot },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_LCRTDD_Uplink_Physical_Channel_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_LCRTDD_Uplink_Physical_Channel_Capability, LCRTDD_Uplink_Physical_Channel_Capability_sequence);

  return offset;
}



static int
dissect_nbap_MACes_Maximum_Bitrate_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 256000000U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_MaximumDL_PowerCapability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 500U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_Maximum_Target_ReceivedTotalWideBandPower_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 621U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_MaxNrOfUL_DPDCHs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 6U, NULL, FALSE);

  return offset;
}


static const value_string nbap_MaxPRACH_MidambleShifts_vals[] = {
  {   0, "shift4" },
  {   1, "shift8" },
  {   2, "shift16" },
  { 0, NULL }
};


static int
dissect_nbap_MaxPRACH_MidambleShifts(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 1, NULL);

  return offset;
}


static const value_string nbap_Max_UE_DTX_Cycle_vals[] = {
  {   0, "v5" },
  {   1, "v10" },
  {   2, "v20" },
  {   3, "v40" },
  {   4, "v64" },
  {   5, "v80" },
  {   6, "v128" },
  {   7, "v160" },
  { 0, NULL }
};


static int
dissect_nbap_Max_UE_DTX_Cycle(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_MBMS_Capability_vals[] = {
  {   0, "mbms-capable" },
  {   1, "mbms-non-capable" },
  { 0, NULL }
};


static int
dissect_nbap_MBMS_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_MeasurementFilterCoefficient_vals[] = {
  {   0, "k0" },
  {   1, "k1" },
  {   2, "k2" },
  {   3, "k3" },
  {   4, "k4" },
  {   5, "k5" },
  {   6, "k6" },
  {   7, "k7" },
  {   8, "k8" },
  {   9, "k9" },
  {  10, "k11" },
  {  11, "k13" },
  {  12, "k15" },
  {  13, "k17" },
  {  14, "k19" },
  { 0, NULL }
};

static value_string_ext nbap_MeasurementFilterCoefficient_vals_ext = VALUE_STRING_EXT_INIT(nbap_MeasurementFilterCoefficient_vals);


static int
dissect_nbap_MeasurementFilterCoefficient(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     15, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_MeasurementID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 1048575U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_MeasurementRecoveryBehavior(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_MeasurementRecoveryReportingIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_MeasurementRecoverySupportIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_RepetitionNumber1(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 256U, NULL, FALSE);

  return offset;
}


static const per_sequence_t MessageStructure_item_sequence[] = {
  { &hf_nbap_iE_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_ID },
  { &hf_nbap_repetitionNumber_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RepetitionNumber1 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MessageStructure_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MessageStructure_item, MessageStructure_item_sequence);

  return offset;
}


static const per_sequence_t MessageStructure_sequence_of[1] = {
  { &hf_nbap_MessageStructure_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_MessageStructure_item },
};

static int
dissect_nbap_MessageStructure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_MessageStructure, MessageStructure_sequence_of,
                                                  1, maxNrOfLevels, FALSE);

  return offset;
}



static int
dissect_nbap_MICH_CFN(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 4095U, NULL, FALSE);

  return offset;
}


static const value_string nbap_MICH_Mode_vals[] = {
  {   0, "v18" },
  {   1, "v36" },
  {   2, "v72" },
  {   3, "v144" },
  {   4, "v16" },
  {   5, "v32" },
  {   6, "v64" },
  {   7, "v128" },
  { 0, NULL }
};


static int
dissect_nbap_MICH_Mode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 4, NULL);

  return offset;
}


static const value_string nbap_MIMO_Capability_vals[] = {
  {   0, "mimo-capable" },
  {   1, "mimo-non-capable" },
  { 0, NULL }
};


static int
dissect_nbap_MIMO_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_MIMO_Mode_Indicator_vals[] = {
  {   0, "activate" },
  {   1, "deactivate" },
  { 0, NULL }
};


static int
dissect_nbap_MIMO_Mode_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_MIMO_PilotConfiguration_vals[] = {
  {   0, "primary-and-secondary-CPICH" },
  {   1, "normal-and-diversity-primary-CPICH" },
  { 0, NULL }
};

static const per_choice_t MIMO_PilotConfiguration_choice[] = {
  {   0, &hf_nbap_primary_and_secondary_CPICH, ASN1_EXTENSION_ROOT    , dissect_nbap_CommonPhysicalChannelID },
  {   1, &hf_nbap_normal_and_diversity_primary_CPICH, ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_MIMO_PilotConfiguration(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_MIMO_PilotConfiguration, MIMO_PilotConfiguration_choice,
                                 NULL);

  return offset;
}



static int
dissect_nbap_PowerOffsetForSecondaryCPICHforMIMO(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -6, 0U, NULL, FALSE);

  return offset;
}


static const per_sequence_t PrimaryAndSecondaryCPICHContainer_sequence[] = {
  { &hf_nbap_power_Offset_For_Secondary_CPICH_for_MIMO, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PowerOffsetForSecondaryCPICHforMIMO },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PrimaryAndSecondaryCPICHContainer(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PrimaryAndSecondaryCPICHContainer, PrimaryAndSecondaryCPICHContainer_sequence);

  return offset;
}


static const per_sequence_t NormalAndDiversityPrimaryCPICHContainer_sequence[] = {
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_NormalAndDiversityPrimaryCPICHContainer(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_NormalAndDiversityPrimaryCPICHContainer, NormalAndDiversityPrimaryCPICHContainer_sequence);

  return offset;
}


static const value_string nbap_MIMO_PilotConfigurationExtension_vals[] = {
  {   0, "primary-and-secondary-CPICH" },
  {   1, "normal-and-diversity-primary-CPICH" },
  { 0, NULL }
};

static const per_choice_t MIMO_PilotConfigurationExtension_choice[] = {
  {   0, &hf_nbap_primary_and_secondary_CPICH_01, ASN1_EXTENSION_ROOT    , dissect_nbap_PrimaryAndSecondaryCPICHContainer },
  {   1, &hf_nbap_normal_and_diversity_primary_CPICH_01, ASN1_EXTENSION_ROOT    , dissect_nbap_NormalAndDiversityPrimaryCPICHContainer },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_MIMO_PilotConfigurationExtension(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_MIMO_PilotConfigurationExtension, MIMO_PilotConfigurationExtension_choice,
                                 NULL);

  return offset;
}


static const value_string nbap_MIMO_PowerOffsetForS_CPICHCapability_vals[] = {
  {   0, "s-CPICH-Power-Offset-Capable" },
  {   1, "s-CPICH-Power-Offset-Not-Capable" },
  { 0, NULL }
};


static int
dissect_nbap_MIMO_PowerOffsetForS_CPICHCapability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_MinimumDL_PowerCapability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 800U, NULL, FALSE);

  return offset;
}


static const value_string nbap_MinimumReducedE_DPDCH_GainFactor_vals[] = {
  {   0, "m8-15" },
  {   1, "m11-15" },
  {   2, "m15-15" },
  {   3, "m21-15" },
  {   4, "m30-15" },
  {   5, "m42-15" },
  {   6, "m60-15" },
  {   7, "m84-15" },
  { 0, NULL }
};


static int
dissect_nbap_MinimumReducedE_DPDCH_GainFactor(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_MinSpreadingFactor_vals[] = {
  {   0, "v4" },
  {   1, "v8" },
  {   2, "v16" },
  {   3, "v32" },
  {   4, "v64" },
  {   5, "v128" },
  {   6, "v256" },
  {   7, "v512" },
  { 0, NULL }
};


static int
dissect_nbap_MinSpreadingFactor(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_Modification_Period_vals[] = {
  {   0, "v1280" },
  {   1, "v2560" },
  {   2, "v5120" },
  {   3, "v10240" },
  { 0, NULL }
};


static int
dissect_nbap_Modification_Period(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_MinUL_ChannelisationCodeLength_vals[] = {
  {   0, "v4" },
  {   1, "v8" },
  {   2, "v16" },
  {   3, "v32" },
  {   4, "v64" },
  {   5, "v128" },
  {   6, "v256" },
  { 0, NULL }
};


static int
dissect_nbap_MinUL_ChannelisationCodeLength(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     7, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_MultiplexingPosition_vals[] = {
  {   0, "fixed" },
  {   1, "flexible" },
  { 0, NULL }
};


static int
dissect_nbap_MultiplexingPosition(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_MAChs_ResetIndicator_vals[] = {
  {   0, "mAChs-NotReset" },
  { 0, NULL }
};


static int
dissect_nbap_MAChs_ResetIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_ModulationMBSFN_vals[] = {
  {   0, "qPSK" },
  {   1, "sixteenQAM" },
  { 0, NULL }
};


static int
dissect_nbap_ModulationMBSFN(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_MBSFN_CPICH_secondary_CCPCH_power_offset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -11, 4U, NULL, TRUE);

  return offset;
}


static const value_string nbap_ModulationPO_MBSFN_vals[] = {
  {   0, "qPSK" },
  {   1, "sixteenQAM" },
  { 0, NULL }
};

static const per_choice_t ModulationPO_MBSFN_choice[] = {
  {   0, &hf_nbap_qPSK           , ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {   1, &hf_nbap_sixteenQAM     , ASN1_EXTENSION_ROOT    , dissect_nbap_MBSFN_CPICH_secondary_CCPCH_power_offset },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_ModulationPO_MBSFN(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_ModulationPO_MBSFN, ModulationPO_MBSFN_choice,
                                 NULL);

  return offset;
}


static const value_string nbap_MBSFN_Only_Mode_Indicator_vals[] = {
  {   0, "mBSFN-Only-Mode" },
  { 0, NULL }
};


static int
dissect_nbap_MBSFN_Only_Mode_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_MBSFN_Only_Mode_Capability_vals[] = {
  {   0, "mBSFN-Only-Mode-capable" },
  {   1, "mBSFN-Only-Mode-non-capable" },
  { 0, NULL }
};


static int
dissect_nbap_MBSFN_Only_Mode_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_Multicarrier_Number(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxHSDPAFrequency, NULL, FALSE);

  return offset;
}


static const per_sequence_t MultipleFreq_HARQ_MemoryPartitioning_InformationItem_sequence[] = {
  { &hf_nbap_hARQ_MemoryPartitioning, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HARQ_MemoryPartitioning },
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UARFCN },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MultipleFreq_HARQ_MemoryPartitioning_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MultipleFreq_HARQ_MemoryPartitioning_InformationItem, MultipleFreq_HARQ_MemoryPartitioning_InformationItem_sequence);

  return offset;
}


static const per_sequence_t MultipleFreq_HARQ_MemoryPartitioning_InformationList_sequence_of[1] = {
  { &hf_nbap_MultipleFreq_HARQ_MemoryPartitioning_InformationList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_MultipleFreq_HARQ_MemoryPartitioning_InformationItem },
};

static int
dissect_nbap_MultipleFreq_HARQ_MemoryPartitioning_InformationList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_MultipleFreq_HARQ_MemoryPartitioning_InformationList, MultipleFreq_HARQ_MemoryPartitioning_InformationList_sequence_of,
                                                  1, maxFrequencyinCell_1, FALSE);

  return offset;
}


static const per_sequence_t MultipleFreq_HSPDSCH_InformationItem_ResponseTDDLCR_sequence[] = {
  { &hf_nbap_hsSCCH_Specific_Information_ResponseTDDLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSSCCH_Specific_InformationRespListTDDLCR },
  { &hf_nbap_hARQ_MemoryPartitioning, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HARQ_MemoryPartitioning },
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UARFCN },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MultipleFreq_HSPDSCH_InformationItem_ResponseTDDLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MultipleFreq_HSPDSCH_InformationItem_ResponseTDDLCR, MultipleFreq_HSPDSCH_InformationItem_ResponseTDDLCR_sequence);

  return offset;
}


static const per_sequence_t MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR_sequence_of[1] = {
  { &hf_nbap_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_MultipleFreq_HSPDSCH_InformationItem_ResponseTDDLCR },
};

static int
dissect_nbap_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR, MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR_sequence_of,
                                                  1, maxHSDPAFrequency_1, FALSE);

  return offset;
}


static const value_string nbap_Multi_Cell_Capability_vals[] = {
  {   0, "multi-Cell-Capable" },
  {   1, "multi-Cell-non-Capable" },
  { 0, NULL }
};


static int
dissect_nbap_Multi_Cell_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t Multi_Cell_Capability_Info_sequence[] = {
  { &hf_nbap_multi_Cell_Capability, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Multi_Cell_Capability },
  { &hf_nbap_possible_Secondary_Serving_Cell_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Possible_Secondary_Serving_Cell_List },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Multi_Cell_Capability_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Multi_Cell_Capability_Info, Multi_Cell_Capability_Info_sequence);

  return offset;
}


static const value_string nbap_Secondary_UL_Frequency_Activation_State_vals[] = {
  {   0, "activated" },
  {   1, "deactivated" },
  { 0, NULL }
};


static int
dissect_nbap_Secondary_UL_Frequency_Activation_State(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t Multicell_EDCH_InformationItemIEs_sequence[] = {
  { &hf_nbap_dL_PowerBalancing_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_PowerBalancing_Information },
  { &hf_nbap_minimumReducedE_DPDCH_GainFactor, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MinimumReducedE_DPDCH_GainFactor },
  { &hf_nbap_secondary_UL_Frequency_Activation_State, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Secondary_UL_Frequency_Activation_State },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Multicell_EDCH_InformationItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Multicell_EDCH_InformationItemIEs, Multicell_EDCH_InformationItemIEs_sequence);

  return offset;
}


static const value_string nbap_Primary_CPICH_Usage_for_Channel_Estimation_vals[] = {
  {   0, "primary-CPICH-may-be-used" },
  {   1, "primary-CPICH-shall-not-be-used" },
  { 0, NULL }
};


static int
dissect_nbap_Primary_CPICH_Usage_for_Channel_Estimation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_Secondary_CPICH_Information_Change_vals[] = {
  {   0, "new-secondary-CPICH" },
  {   1, "secondary-CPICH-shall-not-be-used" },
  { 0, NULL }
};

static const per_choice_t Secondary_CPICH_Information_Change_choice[] = {
  {   0, &hf_nbap_new_secondary_CPICH, ASN1_EXTENSION_ROOT    , dissect_nbap_CommonPhysicalChannelID },
  {   1, &hf_nbap_secondary_CPICH_shall_not_be_used, ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_Secondary_CPICH_Information_Change(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_Secondary_CPICH_Information_Change, Secondary_CPICH_Information_Change_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Multicell_EDCH_RL_Specific_InformationItemIEs_sequence[] = {
  { &hf_nbap_extendedPropagationDelay, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ExtendedPropagationDelay },
  { &hf_nbap_primary_CPICH_Usage_for_Channel_Estimation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Primary_CPICH_Usage_for_Channel_Estimation },
  { &hf_nbap_secondary_CPICH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_secondary_CPICH_Information_Change, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Secondary_CPICH_Information_Change },
  { &hf_nbap_e_AGCH_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_AGCH_PowerOffset },
  { &hf_nbap_e_RGCH_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_RGCH_PowerOffset },
  { &hf_nbap_e_HICH_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_HICH_PowerOffset },
  { &hf_nbap_dLReferencePower, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_e_DCH_DL_Control_Channel_Grant, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Multicell_EDCH_RL_Specific_InformationItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Multicell_EDCH_RL_Specific_InformationItemIEs, Multicell_EDCH_RL_Specific_InformationItemIEs_sequence);

  return offset;
}


static const value_string nbap_MIMO_SFMode_For_HSPDSCHDualStream_vals[] = {
  {   0, "sF1" },
  {   1, "sF1SF16" },
  { 0, NULL }
};


static int
dissect_nbap_MIMO_SFMode_For_HSPDSCHDualStream(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_NCyclesPerSFNperiod_vals[] = {
  {   0, "v1" },
  {   1, "v2" },
  {   2, "v4" },
  {   3, "v8" },
  {   4, "v16" },
  {   5, "v32" },
  {   6, "v64" },
  { 0, NULL }
};


static int
dissect_nbap_NCyclesPerSFNperiod(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 3, NULL);

  return offset;
}



static int
dissect_nbap_NRepetitionsPerCyclePeriod(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            2U, 10U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_N_INSYNC_IND(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 256U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_N_OUTSYNC_IND(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 256U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_RNC_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 4095U, NULL, FALSE);

  return offset;
}


static const per_sequence_t UC_Id_sequence[] = {
  { &hf_nbap_rNC_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RNC_ID },
  { &hf_nbap_c_ID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_C_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UC_Id(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UC_Id, UC_Id_sequence);

  return offset;
}



static int
dissect_nbap_PrimaryScramblingCode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 511U, NULL, FALSE);

  return offset;
}


static const per_sequence_t NeighbouringFDDCellMeasurementInformation_sequence[] = {
  { &hf_nbap_uC_Id          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UC_Id },
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UARFCN },
  { &hf_nbap_primaryScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PrimaryScramblingCode },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_NeighbouringFDDCellMeasurementInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_NeighbouringFDDCellMeasurementInformation, NeighbouringFDDCellMeasurementInformation_sequence);

  return offset;
}


static const per_sequence_t NeighbouringTDDCellMeasurementInformation_sequence[] = {
  { &hf_nbap_uC_Id          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UC_Id },
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UARFCN },
  { &hf_nbap_cellParameterID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellParameterID },
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_NeighbouringTDDCellMeasurementInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_NeighbouringTDDCellMeasurementInformation, NeighbouringTDDCellMeasurementInformation_sequence);

  return offset;
}



static int
dissect_nbap_Extension_neighbouringCellMeasurementInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string nbap_NeighbouringCellMeasurementInformation_item_vals[] = {
  {   0, "neighbouringFDDCellMeasurementInformation" },
  {   1, "neighbouringTDDCellMeasurementInformation" },
  {   2, "extension-neighbouringCellMeasurementInformation" },
  { 0, NULL }
};

static const per_choice_t NeighbouringCellMeasurementInformation_item_choice[] = {
  {   0, &hf_nbap_neighbouringFDDCellMeasurementInformation, ASN1_EXTENSION_ROOT    , dissect_nbap_NeighbouringFDDCellMeasurementInformation },
  {   1, &hf_nbap_neighbouringTDDCellMeasurementInformation, ASN1_EXTENSION_ROOT    , dissect_nbap_NeighbouringTDDCellMeasurementInformation },
  {   2, &hf_nbap_extension_neighbouringCellMeasurementInformation, ASN1_NOT_EXTENSION_ROOT, dissect_nbap_Extension_neighbouringCellMeasurementInformation },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_NeighbouringCellMeasurementInformation_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_NeighbouringCellMeasurementInformation_item, NeighbouringCellMeasurementInformation_item_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t NeighbouringCellMeasurementInformation_sequence_of[1] = {
  { &hf_nbap_NeighbouringCellMeasurementInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_NeighbouringCellMeasurementInformation_item },
};

static int
dissect_nbap_NeighbouringCellMeasurementInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_NeighbouringCellMeasurementInformation, NeighbouringCellMeasurementInformation_sequence_of,
                                                  1, maxNrOfMeasNCell, FALSE);

  return offset;
}


static const per_sequence_t NeighbouringTDDCellMeasurementInformationLCR_sequence[] = {
  { &hf_nbap_uC_Id          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UC_Id },
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UARFCN },
  { &hf_nbap_cellParameterID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellParameterID },
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_NeighbouringTDDCellMeasurementInformationLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_NeighbouringTDDCellMeasurementInformationLCR, NeighbouringTDDCellMeasurementInformationLCR_sequence);

  return offset;
}


static const per_sequence_t NeighbouringTDDCellMeasurementInformation768_sequence[] = {
  { &hf_nbap_uC_Id          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UC_Id },
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UARFCN },
  { &hf_nbap_cellParameterID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellParameterID },
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftAndBurstType768 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_NeighbouringTDDCellMeasurementInformation768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_NeighbouringTDDCellMeasurementInformation768, NeighbouringTDDCellMeasurementInformation768_sequence);

  return offset;
}


static const value_string nbap_NonCellSpecificTxDiversity_vals[] = {
  {   0, "txDiversity" },
  { 0, NULL }
};


static int
dissect_nbap_NonCellSpecificTxDiversity(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_Notification_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 65535U, NULL, FALSE);

  return offset;
}


static const per_sequence_t NI_Information_sequence_of[1] = {
  { &hf_nbap_NI_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Notification_Indicator },
};

static int
dissect_nbap_NI_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_NI_Information, NI_Information_sequence_of,
                                                  1, maxNrOfNIs, FALSE);

  return offset;
}



static int
dissect_nbap_NodeB_CommunicationContextID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  gboolean crnc_context_present;
  guint node_b_com_context_id;
  nbap_com_context_id_t *cur_val;
  nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);

  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 1048575U, &node_b_com_context_id, FALSE);


  /* Checking if CRNC context is present in this frame */
  crnc_context_present = nbap_private_data->crnc_context_present;
  if(crnc_context_present) {
    /* This message contains both context fields. Updaaing the contexts map if needed. */
    if (PINFO_FD_VISITED(actx->pinfo)){
        return offset;
    }

    /* Making sure this Node B context isn't already mapped to a CRNC context */
    if(wmem_tree_lookup32(com_context_map, node_b_com_context_id) == NULL) {
       /* Creating new mapping and adding to map */
       cur_val = wmem_new(wmem_file_scope(), nbap_com_context_id_t);
       cur_val->crnc_context = nbap_private_data->com_context_id;
       cur_val->frame_num = actx->pinfo->num;
       wmem_tree_insert32(com_context_map, node_b_com_context_id, cur_val);
    }
  }
  else {
    /* No CRNC context field in this message, check if Node B context is already mapped to CRNC context. */
    cur_val = (nbap_com_context_id_t *)wmem_tree_lookup32(com_context_map,node_b_com_context_id);
    if(cur_val != NULL){
      /* A mapping was found. Adding to prvivate data. */
      nbap_private_data->com_context_id = cur_val->crnc_context;
      nbap_private_data->crnc_context_present = TRUE;
    }
  }


  return offset;
}


static const value_string nbap_NotificationIndicatorLength_vals[] = {
  {   0, "v2" },
  {   1, "v4" },
  {   2, "v8" },
  { 0, NULL }
};


static int
dissect_nbap_NotificationIndicatorLength(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_NumberOfReportedCellPortions(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxNrOfCellPortionsPerCell, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_NumberOfReportedCellPortionsLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxNrOfCellPortionsPerCellLCR, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_NSubCyclesPerCyclePeriod(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 16U, NULL, TRUE);

  return offset;
}


static const value_string nbap_Number_Of_Supported_Carriers_vals[] = {
  {   0, "one-one-carrier" },
  {   1, "one-three-carrier" },
  {   2, "three-three-carrier" },
  {   3, "one-six-carrier" },
  {   4, "three-six-carrier" },
  {   5, "six-six-carrier" },
  { 0, NULL }
};


static int
dissect_nbap_Number_Of_Supported_Carriers(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     6, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_NoOfTargetCellHS_SCCH_Order(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 30U, NULL, FALSE);

  return offset;
}


static const value_string nbap_Out_of_Sychronization_Window_vals[] = {
  {   0, "ms40" },
  {   1, "ms80" },
  {   2, "ms160" },
  {   3, "ms320" },
  {   4, "ms640" },
  { 0, NULL }
};


static int
dissect_nbap_Out_of_Sychronization_Window(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     5, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_PCCPCH_Power(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -150, 400U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_PDSCH_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_PDSCH_ID768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 511U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_PDSCHSet_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Paging_MACFlows_to_DeleteFDD_Item_sequence[] = {
  { &hf_nbap_paging_MACFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Paging_MACFlow_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Paging_MACFlows_to_DeleteFDD_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Paging_MACFlows_to_DeleteFDD_Item, Paging_MACFlows_to_DeleteFDD_Item_sequence);

  return offset;
}


static const per_sequence_t Paging_MACFlows_to_DeleteFDD_sequence_of[1] = {
  { &hf_nbap_Paging_MACFlows_to_DeleteFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Paging_MACFlows_to_DeleteFDD_Item },
};

static int
dissect_nbap_Paging_MACFlows_to_DeleteFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Paging_MACFlows_to_DeleteFDD, Paging_MACFlows_to_DeleteFDD_sequence_of,
                                                  1, maxNrOfPagingMACFlow, FALSE);

  return offset;
}


static const value_string nbap_PhysicalChannelID_for_CommonERNTI_RequestedIndicator_vals[] = {
  {   0, "requested" },
  { 0, NULL }
};


static int
dissect_nbap_PhysicalChannelID_for_CommonERNTI_RequestedIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_PLCCHsequenceNumber(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 14U, NULL, FALSE);

  return offset;
}


static const per_sequence_t PLCCHinformation_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_sequenceNumber , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PLCCHsequenceNumber },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PLCCHinformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PLCCHinformation, PLCCHinformation_sequence);

  return offset;
}



static int
dissect_nbap_PowerOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 24U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_PowerRaiseLimit(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 10U, NULL, FALSE);

  return offset;
}


static const value_string nbap_PRACH_Midamble_vals[] = {
  {   0, "inverted" },
  {   1, "direct" },
  { 0, NULL }
};


static int
dissect_nbap_PRACH_Midamble(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_PrecodingWeightSetRestriction_vals[] = {
  {   0, "preferred" },
  {   1, "not-preferred" },
  { 0, NULL }
};


static int
dissect_nbap_PrecodingWeightSetRestriction(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_PredictedSFNSFNDeviationLimit(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 256U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_PredictedTUTRANGPSDeviationLimit(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 256U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_PrimaryCPICH_Power(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -100, 500U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_PrimaryCCPCH_RSCP(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 91U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_PrimaryCCPCH_RSCP_Delta(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -5, -1, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_SCH_TimeSlot(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 6U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_PUSCH_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Selected_MBMS_Service_TimeSlot_Information_LCR_sequence_of[1] = {
  { &hf_nbap_Selected_MBMS_Service_TimeSlot_Information_LCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
};

static int
dissect_nbap_Selected_MBMS_Service_TimeSlot_Information_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Selected_MBMS_Service_TimeSlot_Information_LCR, Selected_MBMS_Service_TimeSlot_Information_LCR_sequence_of,
                                                  1, 7, FALSE);

  return offset;
}


static const value_string nbap_T_transmission_Time_Interval_vals[] = {
  {   0, "v10" },
  {   1, "v20" },
  {   2, "v40" },
  {   3, "v80" },
  { 0, NULL }
};


static int
dissect_nbap_T_transmission_Time_Interval(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_INTEGER_2_9(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            2U, 9U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_INTEGER_0_8(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 8U, NULL, FALSE);

  return offset;
}


static const per_sequence_t MBMS_Service_TDM_Information_sequence[] = {
  { &hf_nbap_transmission_Time_Interval, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_transmission_Time_Interval },
  { &hf_nbap_tDM_Rep        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_2_9 },
  { &hf_nbap_tDM_Offset     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_8 },
  { &hf_nbap_tDM_Length     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_1_8 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MBMS_Service_TDM_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MBMS_Service_TDM_Information, MBMS_Service_TDM_Information_sequence);

  return offset;
}


static const per_sequence_t Selected_MBMS_Service_Item_sequence[] = {
  { &hf_nbap_selected_MBMS_Service_TimeSlot_Information_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Selected_MBMS_Service_TimeSlot_Information_LCR },
  { &hf_nbap_mBMS_Service_TDM_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MBMS_Service_TDM_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Selected_MBMS_Service_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Selected_MBMS_Service_Item, Selected_MBMS_Service_Item_sequence);

  return offset;
}


static const per_sequence_t Selected_MBMS_Service_List_sequence_of[1] = {
  { &hf_nbap_Selected_MBMS_Service_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Selected_MBMS_Service_Item },
};

static int
dissect_nbap_Selected_MBMS_Service_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Selected_MBMS_Service_List, Selected_MBMS_Service_List_sequence_of,
                                                  1, maxMBMSServiceSelect, FALSE);

  return offset;
}


static const per_sequence_t Selected_MBMS_Service_sequence[] = {
  { &hf_nbap_selected_MBMS_Service_List, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Selected_MBMS_Service_List },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Selected_MBMS_Service(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Selected_MBMS_Service, Selected_MBMS_Service_sequence);

  return offset;
}


static const value_string nbap_UE_Selected_MBMS_Service_Information_vals[] = {
  {   0, "none" },
  {   1, "selected-MBMS-Service" },
  { 0, NULL }
};

static const per_choice_t UE_Selected_MBMS_Service_Information_choice[] = {
  {   0, &hf_nbap_none           , ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {   1, &hf_nbap_selected_MBMS_Service, ASN1_EXTENSION_ROOT    , dissect_nbap_Selected_MBMS_Service },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_UE_Selected_MBMS_Service_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_UE_Selected_MBMS_Service_Information, UE_Selected_MBMS_Service_Information_choice,
                                 NULL);

  return offset;
}



static int
dissect_nbap_PUSCHSet_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Paging_MACFlows_to_DeleteLCR_Item_sequence[] = {
  { &hf_nbap_paging_MACFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Paging_MACFlow_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Paging_MACFlows_to_DeleteLCR_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Paging_MACFlows_to_DeleteLCR_Item, Paging_MACFlows_to_DeleteLCR_Item_sequence);

  return offset;
}


static const per_sequence_t Paging_MACFlows_to_DeleteLCR_sequence_of[1] = {
  { &hf_nbap_Paging_MACFlows_to_DeleteLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Paging_MACFlows_to_DeleteLCR_Item },
};

static int
dissect_nbap_Paging_MACFlows_to_DeleteLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Paging_MACFlows_to_DeleteLCR, Paging_MACFlows_to_DeleteLCR_sequence_of,
                                                  1, maxNrOfPagingMACFlow, FALSE);

  return offset;
}


static const value_string nbap_RACH_SlotFormat_vals[] = {
  {   0, "v0" },
  {   1, "v1" },
  {   2, "v2" },
  {   3, "v3" },
  { 0, NULL }
};


static int
dissect_nbap_RACH_SlotFormat(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}


static int * const RACH_SubChannelNumbers_bits[] = {
  &hf_nbap_RACH_SubChannelNumbers_subCh11,
  &hf_nbap_RACH_SubChannelNumbers_subCh10,
  &hf_nbap_RACH_SubChannelNumbers_subCh9,
  &hf_nbap_RACH_SubChannelNumbers_subCh8,
  &hf_nbap_RACH_SubChannelNumbers_subCh7,
  &hf_nbap_RACH_SubChannelNumbers_subCh6,
  &hf_nbap_RACH_SubChannelNumbers_subCh5,
  &hf_nbap_RACH_SubChannelNumbers_subCh4,
  &hf_nbap_RACH_SubChannelNumbers_subCh3,
  &hf_nbap_RACH_SubChannelNumbers_subCh2,
  &hf_nbap_RACH_SubChannelNumbers_subCh1,
  &hf_nbap_RACH_SubChannelNumbers_subCh0,
  NULL
};

static int
dissect_nbap_RACH_SubChannelNumbers(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     12, 12, FALSE, RACH_SubChannelNumbers_bits, 12, NULL, NULL);

  return offset;
}



static int
dissect_nbap_T_dCH_id(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_DCH_ID(tvb, offset, actx, tree, hf_index);

    nbap_get_private_data(actx->pinfo)->dch_id = nbap_get_private_data(actx->pinfo)->t_dch_id;

  return offset;
}


static const per_sequence_t RL_Specific_DCH_Info_Item_sequence[] = {
  { &hf_nbap_dCH_id         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_dCH_id },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportlayeraddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_Specific_DCH_Info_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
address     dst_addr, null_addr;
conversation_t *conversation = NULL;
umts_fp_conversation_info_t *umts_fp_conversation_info;
int i, j, num_tf;
guint32 transportLayerAddress_ipv4;
guint16 bindingID;
guint32 dch_id;
nbap_dch_channel_info_t* nbap_dch_chnl_info;
nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
nbap_dch_chnl_info = nbap_private_data->nbap_dch_chnl_info;

nbap_private_data->transportLayerAddress_ipv4 = 0;
nbap_private_data->binding_id_port = 0;
nbap_private_data->dch_id = 0xFFFFFFFF;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_Specific_DCH_Info_Item, RL_Specific_DCH_Info_Item_sequence);


        transportLayerAddress_ipv4 = nbap_private_data->transportLayerAddress_ipv4;
        bindingID = nbap_private_data->binding_id_port;
        if (PINFO_FD_VISITED(actx->pinfo) || transportLayerAddress_ipv4 == 0 || bindingID == 0){
            return offset;
        }

        /*RBS might sometimes send a nonsens bind, to indicate that no DCH is present*/
        if(bindingID == NBAP_IGNORE_PORT){
                return offset;
        }

        clear_address(&null_addr);

        set_address(&dst_addr, AT_IPv4, 4, &transportLayerAddress_ipv4);

        conversation = find_conversation(actx->pinfo->num, &dst_addr,
            &null_addr, CONVERSATION_UDP, nbap_private_data->binding_id_port,
            0, NO_ADDR_B|NO_PORT_B);

        if (conversation == NULL) {
            /* It's not part of any conversation - create a new one. */
            conversation = conversation_new(actx->pinfo->num, &dst_addr,
                &null_addr, CONVERSATION_UDP, nbap_private_data->binding_id_port,
                0, NO_ADDR2|NO_PORT2);

            /* Set dissector */
            conversation_set_dissector(conversation, fp_handle);
            if(actx->pinfo->link_dir==P2P_DIR_DL){
                umts_fp_conversation_info = wmem_new0(wmem_file_scope(), umts_fp_conversation_info_t);

                /* Fill in the data */
                umts_fp_conversation_info->iface_type        = IuB_Interface;
                umts_fp_conversation_info->division          = Division_FDD;
                umts_fp_conversation_info->channel           = CHANNEL_DCH;
                umts_fp_conversation_info->dl_frame_number   = 0;
                umts_fp_conversation_info->ul_frame_number   = actx->pinfo->num;
                copy_address_wmem(wmem_file_scope(), &(umts_fp_conversation_info->crnc_address), &dst_addr);
                umts_fp_conversation_info->crnc_port         = bindingID;
                umts_fp_conversation_info->scrambling_code   = nbap_private_data->ul_scrambling_code;
                umts_fp_conversation_info->rlc_mode          = FP_RLC_MODE_UNKNOWN;

                /* DCH's in this flow */
                umts_fp_conversation_info->dch_crc_present = nbap_private_data->dch_crc_present;

                /*Save unique UE-identifier */
                umts_fp_conversation_info->com_context_id = nbap_private_data->com_context_id;

                /*UPLINK*/
                /* Set data for First or single channel */
                dch_id = nbap_private_data->dch_id;
                if (dch_id != 0xffffffff) {
                    umts_fp_conversation_info->fp_dch_channel_info[0].num_ul_chans = num_tf = nbap_dch_chnl_info[dch_id].num_ul_chans;
                    for (j = 0; j < num_tf; j++) {
                        umts_fp_conversation_info->fp_dch_channel_info[0].ul_chan_tf_size[j] = nbap_dch_chnl_info[dch_id].ul_chan_tf_size[j];
                        umts_fp_conversation_info->fp_dch_channel_info[0].ul_chan_num_tbs[j] = nbap_dch_chnl_info[dch_id].ul_chan_num_tbs[j];
                    }

                    /* Traffic flows per DCH(DL) */
                    umts_fp_conversation_info->fp_dch_channel_info[0].num_dl_chans = num_tf = nbap_dch_chnl_info[dch_id].num_dl_chans;
                    for (j = 0; j < num_tf; j++) {
                        umts_fp_conversation_info->fp_dch_channel_info[0].dl_chan_tf_size[j] = nbap_dch_chnl_info[dch_id].dl_chan_tf_size[j];
                        umts_fp_conversation_info->fp_dch_channel_info[0].dl_chan_num_tbs[j] = nbap_dch_chnl_info[dch_id].dl_chan_num_tbs[j];
                    }

                    /* Set data for associated DCH's if we have any */
                    i = dch_id;
                    umts_fp_conversation_info->dch_ids_in_flow_list[0] = dch_id;
                    while(nbap_dch_chnl_info[i].next_dch != 0 && umts_fp_conversation_info->num_dch_in_flow < FP_maxNrOfDCHs){
                        i = nbap_dch_chnl_info[i].next_dch;
                        umts_fp_conversation_info->num_dch_in_flow++;
                        umts_fp_conversation_info->dch_ids_in_flow_list[umts_fp_conversation_info->num_dch_in_flow] = i;   /*Set transport channel id*/

                        /*Setting Logical Channel's for this DCH*/
                        /* Traffic flows per DCH(UL) */
                        umts_fp_conversation_info->fp_dch_channel_info[umts_fp_conversation_info->num_dch_in_flow].num_ul_chans = num_tf = nbap_dch_chnl_info[i].num_ul_chans;
                        for (j = 0; j < num_tf; j++) {
                                umts_fp_conversation_info->fp_dch_channel_info[umts_fp_conversation_info->num_dch_in_flow].ul_chan_tf_size[j] = nbap_dch_chnl_info[i].ul_chan_tf_size[j];
                                umts_fp_conversation_info->fp_dch_channel_info[umts_fp_conversation_info->num_dch_in_flow].ul_chan_num_tbs[j] = nbap_dch_chnl_info[i].ul_chan_num_tbs[j];
                        }

                        /* Traffic flows per DCH(DL) */
                        umts_fp_conversation_info->fp_dch_channel_info[umts_fp_conversation_info->num_dch_in_flow].num_dl_chans = num_tf = nbap_dch_chnl_info[i].num_dl_chans;
                        for (j = 0; j < num_tf; j++) {
                                umts_fp_conversation_info->fp_dch_channel_info[umts_fp_conversation_info->num_dch_in_flow].dl_chan_tf_size[j] = nbap_dch_chnl_info[i].dl_chan_tf_size[j];
                                umts_fp_conversation_info->fp_dch_channel_info[umts_fp_conversation_info->num_dch_in_flow].dl_chan_num_tbs[j] = nbap_dch_chnl_info[i].dl_chan_num_tbs[j];
                        }
                    }
                    umts_fp_conversation_info->num_dch_in_flow++;
                    umts_fp_conversation_info->dch_ids_in_flow_list[umts_fp_conversation_info->num_dch_in_flow] = i;
                    set_umts_fp_conv_data(conversation, umts_fp_conversation_info);
                }
            }
            nbap_debug("Frame %u RL-Specific-DCH-Info-Item Start", actx->pinfo->num);
            nbap_debug("    Total no of ch in flow will be: %d", umts_fp_conversation_info->num_dch_in_flow);
            nbap_debug("Frame %u RL-Specific-DCH-Info-Item End", actx->pinfo->num);
        }


  return offset;
}


static const per_sequence_t RL_Specific_DCH_Info_sequence_of[1] = {
  { &hf_nbap_RL_Specific_DCH_Info_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_RL_Specific_DCH_Info_Item },
};

static int
dissect_nbap_RL_Specific_DCH_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RL_Specific_DCH_Info, RL_Specific_DCH_Info_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t RL_Specific_E_DCH_Information_Item_sequence[] = {
  { &hf_nbap_e_DCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_ID },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportlayeraddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_Specific_E_DCH_Information_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
address     dst_addr, null_addr;
conversation_t *conversation;
umts_fp_conversation_info_t *umts_fp_conversation_info = NULL;
fp_edch_channel_info_t* fp_edch_channel_info;
nbap_edch_port_info_t *old_info = NULL;
guint32 transportLayerAddress_ipv4;
guint16 bindingID;
guint32 e_dch_macdflow_id;
nbap_edch_channel_info_t* nbap_edch_channel_info;
nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
nbap_edch_channel_info = nbap_private_data->nbap_edch_channel_info;

nbap_private_data->transportLayerAddress_ipv4 = 0;
nbap_private_data->binding_id_port = 0;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_Specific_E_DCH_Information_Item, RL_Specific_E_DCH_Information_Item_sequence);


        transportLayerAddress_ipv4 = nbap_private_data->transportLayerAddress_ipv4;
        bindingID = nbap_private_data->binding_id_port;
        if (PINFO_FD_VISITED(actx->pinfo) || transportLayerAddress_ipv4 == 0 || bindingID == 0){
            return offset;
        }
        clear_address(&null_addr);

        set_address(&dst_addr, AT_IPv4, 4, &transportLayerAddress_ipv4);

        conversation = find_conversation(actx->pinfo->num, &dst_addr,
            &null_addr, CONVERSATION_UDP, bindingID,
            0, NO_ADDR_B|NO_PORT_B);
        if (conversation) {
            umts_fp_conversation_info = (umts_fp_conversation_info_t*)conversation_get_proto_data(conversation, proto_fp);
        }

        /* We must also check if this port is about to be overriden, if that's the case we
         * might already have a DCH entry on this port which should be overwritten
         */
        if ((conversation == NULL) || (umts_fp_conversation_info && umts_fp_conversation_info->channel == CHANNEL_DCH)) {

            /* It's not part of any conversation - create a new one. */
            conversation = conversation_new(actx->pinfo->num, &dst_addr,
                &null_addr, CONVERSATION_UDP, bindingID,
                0, NO_ADDR2|NO_PORT2);

            /* Set dissector */
            conversation_set_dissector(conversation, fp_handle);
            if(actx->pinfo->link_dir==P2P_DIR_DL){
                umts_fp_conversation_info = wmem_new0(wmem_file_scope(), umts_fp_conversation_info_t);
                /* Fill in the data */
                umts_fp_conversation_info->iface_type        = IuB_Interface;
                umts_fp_conversation_info->division          = Division_FDD;
                umts_fp_conversation_info->channel           = CHANNEL_EDCH;
                umts_fp_conversation_info->dl_frame_number   = 0;
                umts_fp_conversation_info->ul_frame_number   = actx->pinfo->num;
                copy_address_wmem(wmem_file_scope(), &(umts_fp_conversation_info->crnc_address), &dst_addr);
                umts_fp_conversation_info->crnc_port         = bindingID;
                umts_fp_conversation_info->rlc_mode          = FP_RLC_MODE_UNKNOWN;

                fp_edch_channel_info = wmem_new0(wmem_file_scope(), fp_edch_channel_info_t);
                umts_fp_conversation_info->channel_specific_info = (void*)fp_edch_channel_info;

                if(nbap_private_data->crnc_context_present){
                    umts_fp_conversation_info->com_context_id = nbap_private_data->com_context_id;
                }else{
                    expert_add_info(actx->pinfo, NULL, &ei_nbap_no_set_comm_context_id);
                }

                /* Check if we allready have this context */
                e_dch_macdflow_id = nbap_private_data->e_dch_macdflow_id;
                if( (old_info = (nbap_edch_port_info_t *)wmem_tree_lookup32(edch_flow_port_map,nbap_private_data->com_context_id)) == NULL ){
                    nbap_edch_port_info_t * nbap_edch_port_info;

                    nbap_edch_port_info = wmem_new0(wmem_file_scope(), nbap_edch_port_info_t);

                    /*Saving port/flow map based on context id for future reconfigurations*/
                    nbap_edch_port_info->crnc_port[e_dch_macdflow_id] = bindingID;
                    /*Ip address might be useful as well*/
                    nbap_edch_port_info->crnc_address = nbap_private_data->transportLayerAddress_ipv4;

                    nbap_debug("Frame %u RL-Specific-E-DCH-Information-Item Start",
                        actx->pinfo->num);

                    nbap_debug("    wmem_tree_insert32(edch_flow_port_map) com_context_id %u e_dch_macdflow_id %u IP %s Port %u",
                        umts_fp_conversation_info->com_context_id,e_dch_macdflow_id,
                        address_to_str(actx->pinfo->pool, &dst_addr),bindingID);

                    wmem_tree_insert32(edch_flow_port_map, umts_fp_conversation_info->com_context_id, nbap_edch_port_info);
                }else{
                    nbap_debug("    Insert in existing edch_flow_port_map com_context_id %u e_dch_macdflow_id %u IP %s Port %u",
                        umts_fp_conversation_info->com_context_id,e_dch_macdflow_id,
                        address_to_str(actx->pinfo->pool, &dst_addr), bindingID);

                    /* Must be same ADDRESS */
                    old_info->crnc_port[e_dch_macdflow_id] = bindingID;
                }

                /* Set address for collection of DDI entries */
                copy_address_wmem(actx->pinfo->pool,&(nbap_edch_channel_info[e_dch_macdflow_id].crnc_address),&dst_addr);
                nbap_edch_channel_info[e_dch_macdflow_id].crnc_port = bindingID;

                set_umts_fp_conv_data(conversation, umts_fp_conversation_info);

                nbap_debug("Frame %u RL-Specific-E-DCH-Information-Item End", actx->pinfo->num);
            }
    }


  return offset;
}


static const per_sequence_t RL_Specific_E_DCH_Information_sequence_of[1] = {
  { &hf_nbap_RL_Specific_E_DCH_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_RL_Specific_E_DCH_Information_Item },
};

static int
dissect_nbap_RL_Specific_E_DCH_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RL_Specific_E_DCH_Information, RL_Specific_E_DCH_Information_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}


static const per_sequence_t RL_Specific_E_DCH_Info_sequence[] = {
  { &hf_nbap_rL_Specific_E_DCH_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_Specific_E_DCH_Information },
  { &hf_nbap_e_AGCH_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_AGCH_PowerOffset },
  { &hf_nbap_e_RGCH_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_RGCH_PowerOffset },
  { &hf_nbap_e_HICH_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_HICH_PowerOffset },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_Specific_E_DCH_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_Specific_E_DCH_Info, RL_Specific_E_DCH_Info_sequence);

  return offset;
}



static int
dissect_nbap_Reference_ReceivedTotalWideBandPower(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 621U, NULL, FALSE);

  return offset;
}


static const value_string nbap_Reference_ReceivedTotalWideBandPowerReporting_vals[] = {
  {   0, "reference-ReceivedTotalWideBandPower-Requested" },
  { 0, NULL }
};


static int
dissect_nbap_Reference_ReceivedTotalWideBandPowerReporting(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_Reference_ReceivedTotalWideBandPowerSupportIndicator_vals[] = {
  {   0, "indication-of-Reference-ReceivedTotalWideBandPower-supported" },
  { 0, NULL }
};


static int
dissect_nbap_Reference_ReceivedTotalWideBandPowerSupportIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_ReferenceClockAvailability_vals[] = {
  {   0, "available" },
  {   1, "notAvailable" },
  { 0, NULL }
};


static int
dissect_nbap_ReferenceClockAvailability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_ReferenceSFNoffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_RefTFCNumber(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 3U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_ReportPeriodicity_Scaledmsec(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 6000U, NULL, TRUE);

  return offset;
}


static const value_string nbap_ReportCharacteristicsType_ReportPeriodicity_vals[] = {
  {   0, "msec" },
  {   1, "min" },
  { 0, NULL }
};

static const per_choice_t ReportCharacteristicsType_ReportPeriodicity_choice[] = {
  {   0, &hf_nbap_msec_02        , ASN1_EXTENSION_ROOT    , dissect_nbap_ReportPeriodicity_Scaledmsec },
  {   1, &hf_nbap_min            , ASN1_EXTENSION_ROOT    , dissect_nbap_ReportPeriodicity_Scaledmin },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_ReportCharacteristicsType_ReportPeriodicity(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_ReportCharacteristicsType_ReportPeriodicity, ReportCharacteristicsType_ReportPeriodicity_choice,
                                 NULL);

  return offset;
}



static int
dissect_nbap_Extension_ReportCharacteristicsType_MeasurementThreshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string nbap_ReportCharacteristicsType_MeasurementThreshold_vals[] = {
  {   0, "received-total-wide-band-power" },
  {   1, "transmitted-carrier-power" },
  {   2, "acknowledged-prach-preambles" },
  {   3, "uL-TimeslotISCP" },
  {   4, "sir" },
  {   5, "sir-error" },
  {   6, "transmitted-code-power" },
  {   7, "rscp" },
  {   8, "rx-timing-deviation" },
  {   9, "round-trip-time" },
  {  10, "notUsed-1-acknowledged-PCPCH-access-preambles" },
  {  11, "notUsed-2-detected-PCPCH-access-preambles" },
  {  12, "extension-ReportCharacteristicsType-MeasurementThreshold" },
  { 0, NULL }
};

static const per_choice_t ReportCharacteristicsType_MeasurementThreshold_choice[] = {
  {   0, &hf_nbap_received_total_wide_band_power, ASN1_EXTENSION_ROOT    , dissect_nbap_Received_total_wide_band_power_Value },
  {   1, &hf_nbap_transmitted_carrier_power, ASN1_EXTENSION_ROOT    , dissect_nbap_Transmitted_Carrier_Power_Value },
  {   2, &hf_nbap_acknowledged_prach_preambles, ASN1_EXTENSION_ROOT    , dissect_nbap_Acknowledged_PRACH_preambles_Value },
  {   3, &hf_nbap_uL_TimeslotISCP, ASN1_EXTENSION_ROOT    , dissect_nbap_UL_TimeslotISCP_Value },
  {   4, &hf_nbap_sir_01         , ASN1_EXTENSION_ROOT    , dissect_nbap_SIR_Value },
  {   5, &hf_nbap_sir_error_01   , ASN1_EXTENSION_ROOT    , dissect_nbap_SIR_Error_Value },
  {   6, &hf_nbap_transmitted_code_power_01, ASN1_EXTENSION_ROOT    , dissect_nbap_Transmitted_Code_Power_Value },
  {   7, &hf_nbap_rscp_01        , ASN1_EXTENSION_ROOT    , dissect_nbap_RSCP_Value },
  {   8, &hf_nbap_rx_timing_deviation, ASN1_EXTENSION_ROOT    , dissect_nbap_Rx_Timing_Deviation_Value },
  {   9, &hf_nbap_round_trip_time_01, ASN1_EXTENSION_ROOT    , dissect_nbap_Round_Trip_Time_Value },
  {  10, &hf_nbap_notUsed_1_acknowledged_PCPCH_access_preambles, ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {  11, &hf_nbap_notUsed_2_detected_PCPCH_access_preambles, ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {  12, &hf_nbap_extension_ReportCharacteristicsType_MeasurementThreshold, ASN1_NOT_EXTENSION_ROOT, dissect_nbap_Extension_ReportCharacteristicsType_MeasurementThreshold },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_ReportCharacteristicsType_MeasurementThreshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_ReportCharacteristicsType_MeasurementThreshold, ReportCharacteristicsType_MeasurementThreshold_choice,
                                 NULL);

  return offset;
}



static int
dissect_nbap_MeasurementHysteresisTime_Scaledmsec(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 6000U, NULL, TRUE);

  return offset;
}


static const value_string nbap_ReportCharacteristicsType_ScaledMeasurementHysteresisTime_vals[] = {
  {   0, "msec" },
  { 0, NULL }
};

static const per_choice_t ReportCharacteristicsType_ScaledMeasurementHysteresisTime_choice[] = {
  {   0, &hf_nbap_msec_01        , ASN1_EXTENSION_ROOT    , dissect_nbap_MeasurementHysteresisTime_Scaledmsec },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_ReportCharacteristicsType_ScaledMeasurementHysteresisTime(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_ReportCharacteristicsType_ScaledMeasurementHysteresisTime, ReportCharacteristicsType_ScaledMeasurementHysteresisTime_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t ReportCharacteristicsType_EventA_sequence[] = {
  { &hf_nbap_measurementThreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ReportCharacteristicsType_MeasurementThreshold },
  { &hf_nbap_measurementHysteresisTime, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ReportCharacteristicsType_ScaledMeasurementHysteresisTime },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ReportCharacteristicsType_EventA(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ReportCharacteristicsType_EventA, ReportCharacteristicsType_EventA_sequence);

  return offset;
}


static const per_sequence_t ReportCharacteristicsType_EventB_sequence[] = {
  { &hf_nbap_measurementThreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ReportCharacteristicsType_MeasurementThreshold },
  { &hf_nbap_measurementHysteresisTime, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ReportCharacteristicsType_ScaledMeasurementHysteresisTime },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ReportCharacteristicsType_EventB(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ReportCharacteristicsType_EventB, ReportCharacteristicsType_EventB_sequence);

  return offset;
}



static int
dissect_nbap_Received_total_wide_band_power_Value_IncrDecrThres(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 620U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_UL_TimeslotISCP_Value_IncrDecrThres(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 126U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_SIR_Value_IncrDecrThres(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 62U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_SIR_Error_Value_IncrDecrThres(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 124U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_Transmitted_Code_Power_Value_IncrDecrThres(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 112U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_RSCP_Value_IncrDecrThres(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 126U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_Round_Trip_Time_IncrDecrThres(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 32766U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_Extension_ReportCharacteristicsType_MeasurementIncreaseDecreaseThreshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string nbap_ReportCharacteristicsType_MeasurementIncreaseDecreaseThreshold_vals[] = {
  {   0, "received-total-wide-band-power" },
  {   1, "transmitted-carrier-power" },
  {   2, "acknowledged-prach-preambles" },
  {   3, "uL-TimeslotISCP" },
  {   4, "sir" },
  {   5, "sir-error" },
  {   6, "transmitted-code-power" },
  {   7, "rscp" },
  {   8, "round-trip-time" },
  {   9, "notUsed-1-acknowledged-PCPCH-access-preambles" },
  {  10, "notUsed-2-detected-PCPCH-access-preambles" },
  {  11, "extension-ReportCharacteristicsType-MeasurementIncreaseDecreaseThreshold" },
  { 0, NULL }
};

static const per_choice_t ReportCharacteristicsType_MeasurementIncreaseDecreaseThreshold_choice[] = {
  {   0, &hf_nbap_received_total_wide_band_power_01, ASN1_EXTENSION_ROOT    , dissect_nbap_Received_total_wide_band_power_Value_IncrDecrThres },
  {   1, &hf_nbap_transmitted_carrier_power, ASN1_EXTENSION_ROOT    , dissect_nbap_Transmitted_Carrier_Power_Value },
  {   2, &hf_nbap_acknowledged_prach_preambles, ASN1_EXTENSION_ROOT    , dissect_nbap_Acknowledged_PRACH_preambles_Value },
  {   3, &hf_nbap_uL_TimeslotISCP_01, ASN1_EXTENSION_ROOT    , dissect_nbap_UL_TimeslotISCP_Value_IncrDecrThres },
  {   4, &hf_nbap_sir            , ASN1_EXTENSION_ROOT    , dissect_nbap_SIR_Value_IncrDecrThres },
  {   5, &hf_nbap_sir_error      , ASN1_EXTENSION_ROOT    , dissect_nbap_SIR_Error_Value_IncrDecrThres },
  {   6, &hf_nbap_transmitted_code_power, ASN1_EXTENSION_ROOT    , dissect_nbap_Transmitted_Code_Power_Value_IncrDecrThres },
  {   7, &hf_nbap_rscp           , ASN1_EXTENSION_ROOT    , dissect_nbap_RSCP_Value_IncrDecrThres },
  {   8, &hf_nbap_round_trip_time, ASN1_EXTENSION_ROOT    , dissect_nbap_Round_Trip_Time_IncrDecrThres },
  {   9, &hf_nbap_notUsed_1_acknowledged_PCPCH_access_preambles, ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {  10, &hf_nbap_notUsed_2_detected_PCPCH_access_preambles, ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {  11, &hf_nbap_extension_ReportCharacteristicsType_MeasurementIncreaseDecreaseThreshold, ASN1_NOT_EXTENSION_ROOT, dissect_nbap_Extension_ReportCharacteristicsType_MeasurementIncreaseDecreaseThreshold },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_ReportCharacteristicsType_MeasurementIncreaseDecreaseThreshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_ReportCharacteristicsType_MeasurementIncreaseDecreaseThreshold, ReportCharacteristicsType_MeasurementIncreaseDecreaseThreshold_choice,
                                 NULL);

  return offset;
}



static int
dissect_nbap_MeasurementChangeTime_Scaledmsec(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 6000U, NULL, TRUE);

  return offset;
}


static const value_string nbap_ReportCharacteristicsType_ScaledMeasurementChangeTime_vals[] = {
  {   0, "msec" },
  { 0, NULL }
};

static const per_choice_t ReportCharacteristicsType_ScaledMeasurementChangeTime_choice[] = {
  {   0, &hf_nbap_msec           , ASN1_EXTENSION_ROOT    , dissect_nbap_MeasurementChangeTime_Scaledmsec },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_ReportCharacteristicsType_ScaledMeasurementChangeTime(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_ReportCharacteristicsType_ScaledMeasurementChangeTime, ReportCharacteristicsType_ScaledMeasurementChangeTime_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t ReportCharacteristicsType_EventC_sequence[] = {
  { &hf_nbap_measurementIncreaseThreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ReportCharacteristicsType_MeasurementIncreaseDecreaseThreshold },
  { &hf_nbap_measurementChangeTime, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ReportCharacteristicsType_ScaledMeasurementChangeTime },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ReportCharacteristicsType_EventC(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ReportCharacteristicsType_EventC, ReportCharacteristicsType_EventC_sequence);

  return offset;
}


static const per_sequence_t ReportCharacteristicsType_EventD_sequence[] = {
  { &hf_nbap_measurementDecreaseThreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ReportCharacteristicsType_MeasurementIncreaseDecreaseThreshold },
  { &hf_nbap_measurementChangeTime, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ReportCharacteristicsType_ScaledMeasurementChangeTime },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ReportCharacteristicsType_EventD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ReportCharacteristicsType_EventD, ReportCharacteristicsType_EventD_sequence);

  return offset;
}


static const per_sequence_t ReportCharacteristicsType_EventE_sequence[] = {
  { &hf_nbap_measurementThreshold1, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ReportCharacteristicsType_MeasurementThreshold },
  { &hf_nbap_measurementThreshold2, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ReportCharacteristicsType_MeasurementThreshold },
  { &hf_nbap_measurementHysteresisTime, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ReportCharacteristicsType_ScaledMeasurementHysteresisTime },
  { &hf_nbap_reportPeriodicity, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ReportCharacteristicsType_ReportPeriodicity },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ReportCharacteristicsType_EventE(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ReportCharacteristicsType_EventE, ReportCharacteristicsType_EventE_sequence);

  return offset;
}


static const per_sequence_t ReportCharacteristicsType_EventF_sequence[] = {
  { &hf_nbap_measurementThreshold1, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ReportCharacteristicsType_MeasurementThreshold },
  { &hf_nbap_measurementThreshold2, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ReportCharacteristicsType_MeasurementThreshold },
  { &hf_nbap_measurementHysteresisTime, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ReportCharacteristicsType_ScaledMeasurementHysteresisTime },
  { &hf_nbap_reportPeriodicity, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ReportCharacteristicsType_ReportPeriodicity },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ReportCharacteristicsType_EventF(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ReportCharacteristicsType_EventF, ReportCharacteristicsType_EventF_sequence);

  return offset;
}



static int
dissect_nbap_Extension_ReportCharacteristics(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string nbap_ReportCharacteristics_vals[] = {
  {   0, "onDemand" },
  {   1, "periodic" },
  {   2, "event-a" },
  {   3, "event-b" },
  {   4, "event-c" },
  {   5, "event-d" },
  {   6, "event-e" },
  {   7, "event-f" },
  {   8, "extension-ReportCharacteristics" },
  { 0, NULL }
};

static const per_choice_t ReportCharacteristics_choice[] = {
  {   0, &hf_nbap_onDemand       , ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {   1, &hf_nbap_periodic_01    , ASN1_EXTENSION_ROOT    , dissect_nbap_ReportCharacteristicsType_ReportPeriodicity },
  {   2, &hf_nbap_event_a        , ASN1_EXTENSION_ROOT    , dissect_nbap_ReportCharacteristicsType_EventA },
  {   3, &hf_nbap_event_b        , ASN1_EXTENSION_ROOT    , dissect_nbap_ReportCharacteristicsType_EventB },
  {   4, &hf_nbap_event_c        , ASN1_EXTENSION_ROOT    , dissect_nbap_ReportCharacteristicsType_EventC },
  {   5, &hf_nbap_event_d        , ASN1_EXTENSION_ROOT    , dissect_nbap_ReportCharacteristicsType_EventD },
  {   6, &hf_nbap_event_e        , ASN1_EXTENSION_ROOT    , dissect_nbap_ReportCharacteristicsType_EventE },
  {   7, &hf_nbap_event_f        , ASN1_EXTENSION_ROOT    , dissect_nbap_ReportCharacteristicsType_EventF },
  {   8, &hf_nbap_extension_ReportCharacteristics, ASN1_NOT_EXTENSION_ROOT, dissect_nbap_Extension_ReportCharacteristics },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_ReportCharacteristics(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_ReportCharacteristics, ReportCharacteristics_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t ReportCharacteristicsType_OnModification_sequence[] = {
  { &hf_nbap_measurementThreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ReportCharacteristicsType_MeasurementThreshold },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ReportCharacteristicsType_OnModification(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ReportCharacteristicsType_OnModification, ReportCharacteristicsType_OnModification_sequence);

  return offset;
}


static const per_sequence_t EDCH_RACH_Report_IncrDecrThres_sequence[] = {
  { &hf_nbap_denied_EDCH_RACH_resources, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Denied_EDCH_RACH_Resources_Value },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_EDCH_RACH_Report_IncrDecrThres(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_EDCH_RACH_Report_IncrDecrThres, EDCH_RACH_Report_IncrDecrThres_sequence);

  return offset;
}


static const per_sequence_t EDCH_RACH_Report_ThresholdInformation_sequence[] = {
  { &hf_nbap_denied_EDCH_RACH_resources, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Denied_EDCH_RACH_Resources_Value },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_EDCH_RACH_Report_ThresholdInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_EDCH_RACH_Report_ThresholdInformation, EDCH_RACH_Report_ThresholdInformation_sequence);

  return offset;
}


static const value_string nbap_DL_RLC_PDU_Size_Format_vals[] = {
  {   0, "fixed-RLC-PDU-Size" },
  {   1, "flexible-RLC-PDU-Size" },
  { 0, NULL }
};


static int
dissect_nbap_DL_RLC_PDU_Size_Format(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t Received_total_wide_band_power_For_CellPortion_Value_Item_sequence[] = {
  { &hf_nbap_cellPortionID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellPortionID },
  { &hf_nbap_received_total_wide_band_power_value, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Received_total_wide_band_power_Value },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Received_total_wide_band_power_For_CellPortion_Value_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Received_total_wide_band_power_For_CellPortion_Value_Item, Received_total_wide_band_power_For_CellPortion_Value_Item_sequence);

  return offset;
}


static const per_sequence_t Received_total_wide_band_power_For_CellPortion_Value_sequence_of[1] = {
  { &hf_nbap_Received_total_wide_band_power_For_CellPortion_Value_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Received_total_wide_band_power_For_CellPortion_Value_Item },
};

static int
dissect_nbap_Received_total_wide_band_power_For_CellPortion_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Received_total_wide_band_power_For_CellPortion_Value, Received_total_wide_band_power_For_CellPortion_Value_sequence_of,
                                                  1, maxNrOfCellPortionsPerCell, FALSE);

  return offset;
}


static const per_sequence_t Received_total_wide_band_power_For_CellPortion_ValueLCR_Item_sequence[] = {
  { &hf_nbap_cellPortionLCRID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellPortionLCRID },
  { &hf_nbap_received_total_wide_band_power_value, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Received_total_wide_band_power_Value },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Received_total_wide_band_power_For_CellPortion_ValueLCR_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Received_total_wide_band_power_For_CellPortion_ValueLCR_Item, Received_total_wide_band_power_For_CellPortion_ValueLCR_Item_sequence);

  return offset;
}


static const per_sequence_t Received_total_wide_band_power_For_CellPortion_ValueLCR_sequence_of[1] = {
  { &hf_nbap_Received_total_wide_band_power_For_CellPortion_ValueLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Received_total_wide_band_power_For_CellPortion_ValueLCR_Item },
};

static int
dissect_nbap_Received_total_wide_band_power_For_CellPortion_ValueLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Received_total_wide_band_power_For_CellPortion_ValueLCR, Received_total_wide_band_power_For_CellPortion_ValueLCR_sequence_of,
                                                  1, maxNrOfCellPortionsPerCellLCR, FALSE);

  return offset;
}



static int
dissect_nbap_RSEPS_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 151U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value_Item_sequence[] = {
  { &hf_nbap_cellPortionID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellPortionID },
  { &hf_nbap_received_Scheduled_power_share_value, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RSEPS_Value },
  { &hf_nbap_received_total_wide_band_power_value, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Received_total_wide_band_power_Value },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value_Item, Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value_Item_sequence);

  return offset;
}


static const per_sequence_t Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value_sequence_of[1] = {
  { &hf_nbap_Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value_Item },
};

static int
dissect_nbap_Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value, Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value_sequence_of,
                                                  1, maxNrOfCellPortionsPerCell, FALSE);

  return offset;
}


static const per_sequence_t Received_Scheduled_EDCH_Power_Share_Value_sequence[] = {
  { &hf_nbap_received_Scheduled_power_share_value, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RSEPS_Value },
  { &hf_nbap_received_total_wide_band_power_value, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Received_total_wide_band_power_Value },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Received_Scheduled_EDCH_Power_Share_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Received_Scheduled_EDCH_Power_Share_Value, Received_Scheduled_EDCH_Power_Share_Value_sequence);

  return offset;
}



static int
dissect_nbap_RSEPS_Value_IncrDecrThres(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 151U, NULL, FALSE);

  return offset;
}


static const per_sequence_t RequestedDataValue_sequence[] = {
  { &hf_nbap_dgps_corrections, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DGPSCorrections },
  { &hf_nbap_gps_navandrecovery, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GPS_NavigationModel_and_TimeRecovery },
  { &hf_nbap_gps_ionos_model, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GPS_Ionospheric_Model },
  { &hf_nbap_gps_utc_model  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GPS_UTC_Model },
  { &hf_nbap_gps_almanac    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GPS_Almanac },
  { &hf_nbap_gps_rt_integrity, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GPS_RealTime_Integrity },
  { &hf_nbap_gpsrxpos       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GPS_RX_POS },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RequestedDataValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RequestedDataValue, RequestedDataValue_sequence);

  return offset;
}


static const per_sequence_t InformationAvailable_sequence[] = {
  { &hf_nbap_requesteddataValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RequestedDataValue },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_InformationAvailable(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_InformationAvailable, InformationAvailable_sequence);

  return offset;
}



static int
dissect_nbap_InformationnotAvailable(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string nbap_RequestedDataValueInformation_vals[] = {
  {   0, "informationAvailable" },
  {   1, "informationnotAvailable" },
  { 0, NULL }
};

static const per_choice_t RequestedDataValueInformation_choice[] = {
  {   0, &hf_nbap_informationAvailable, ASN1_NO_EXTENSIONS     , dissect_nbap_InformationAvailable },
  {   1, &hf_nbap_informationnotAvailable, ASN1_NO_EXTENSIONS     , dissect_nbap_InformationnotAvailable },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_RequestedDataValueInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_RequestedDataValueInformation, RequestedDataValueInformation_choice,
                                 NULL);

  return offset;
}



static int
dissect_nbap_Rx_Timing_Deviation_Value_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 511U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_Rx_Timing_Deviation_Value_768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 65535U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_Rx_Timing_Deviation_Value_384_ext(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 32767U, NULL, FALSE);

  return offset;
}


static const value_string nbap_RTWP_ReportingIndicator_vals[] = {
  {   0, "rTWP-reporting-required" },
  { 0, NULL }
};


static int
dissect_nbap_RTWP_ReportingIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_RTWP_CellPortion_ReportingIndicator_vals[] = {
  {   0, "rTWP-CellPortion-reporting-required" },
  { 0, NULL }
};


static int
dissect_nbap_RTWP_CellPortion_ReportingIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_E_DPCCH_Power_Boosting_Capability_vals[] = {
  {   0, "e-DPCCH-Power-Boosting-capable" },
  {   1, "e-DPCCH-Power-Boosting-non-capable" },
  { 0, NULL }
};


static int
dissect_nbap_E_DPCCH_Power_Boosting_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t SAT_Info_Almanac_ExtItem_sequence[] = {
  { &hf_nbap_data_id        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DATA_ID },
  { &hf_nbap_sat_id         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SAT_ID },
  { &hf_nbap_gps_e_alm      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_gps_toa_alm    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_gps_delta_I_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_omegadot_alm   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_16 },
  { &hf_nbap_svhealth_alm   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_gps_a_sqrt_alm , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_24 },
  { &hf_nbap_omegazero_alm  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_24 },
  { &hf_nbap_m_zero_alm     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_24 },
  { &hf_nbap_gps_omega_alm  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_24 },
  { &hf_nbap_gps_af_zero_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_11 },
  { &hf_nbap_gps_af_one_alm , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_11 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SAT_Info_Almanac_ExtItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SAT_Info_Almanac_ExtItem, SAT_Info_Almanac_ExtItem_sequence);

  return offset;
}


static const per_sequence_t SAT_Info_Almanac_ExtList_sequence_of[1] = {
  { &hf_nbap_SAT_Info_Almanac_ExtList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_SAT_Info_Almanac_ExtItem },
};

static int
dissect_nbap_SAT_Info_Almanac_ExtList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_SAT_Info_Almanac_ExtList, SAT_Info_Almanac_ExtList_sequence_of,
                                                  1, maxNrOfSatAlmanac_maxNoSat, FALSE);

  return offset;
}



static int
dissect_nbap_SecondaryCCPCH_SlotFormat(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 17U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_Secondary_CCPCH_SlotFormat_Extended(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            18U, 23U, NULL, TRUE);

  return offset;
}


static const value_string nbap_Segment_Type_vals[] = {
  {   0, "first-segment" },
  {   1, "first-segment-short" },
  {   2, "subsequent-segment" },
  {   3, "last-segment" },
  {   4, "last-segment-short" },
  {   5, "complete-SIB" },
  {   6, "complete-SIB-short" },
  { 0, NULL }
};


static int
dissect_nbap_Segment_Type(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     7, &nbap_get_private_data(actx->pinfo)->segment_type, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t Serving_E_DCH_RL_in_this_NodeB_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Serving_E_DCH_RL_in_this_NodeB(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Serving_E_DCH_RL_in_this_NodeB, Serving_E_DCH_RL_in_this_NodeB_sequence);

  return offset;
}


static const value_string nbap_Serving_E_DCH_RL_ID_vals[] = {
  {   0, "serving-E-DCH-RL-in-this-NodeB" },
  {   1, "serving-E-DCH-RL-not-in-this-NodeB" },
  { 0, NULL }
};

static const per_choice_t Serving_E_DCH_RL_ID_choice[] = {
  {   0, &hf_nbap_serving_E_DCH_RL_in_this_NodeB, ASN1_EXTENSION_ROOT    , dissect_nbap_Serving_E_DCH_RL_in_this_NodeB },
  {   1, &hf_nbap_serving_E_DCH_RL_not_in_this_NodeB, ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_Serving_E_DCH_RL_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_Serving_E_DCH_RL_ID, Serving_E_DCH_RL_ID_choice,
                                 NULL);

  return offset;
}


static const value_string nbap_Setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency_vals[] = {
  {   0, "setup" },
  {   1, "configurationChange" },
  {   2, "removal" },
  { 0, NULL }
};

static const per_choice_t Setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency_choice[] = {
  {   0, &hf_nbap_setup          , ASN1_EXTENSION_ROOT    , dissect_nbap_Additional_EDCH_Setup_Info },
  {   1, &hf_nbap_configurationChange, ASN1_EXTENSION_ROOT    , dissect_nbap_Additional_EDCH_Cell_Information_ConfigurationChange_List },
  {   2, &hf_nbap_removal        , ASN1_EXTENSION_ROOT    , dissect_nbap_Additional_EDCH_Cell_Information_Removal_List },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_Setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_Setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency, Setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency_choice,
                                 NULL);

  return offset;
}



static int
dissect_nbap_SFN(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 4095U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_SFNSFN_FDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 614399U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_SFNSFN_TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 40961U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_SFNSFN_TDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 81923U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_SFNSFNChangeLimit(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 256U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_SFNSFNDriftRate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -100, 100U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_SFNSFNDriftRateQuality(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 100U, NULL, FALSE);

  return offset;
}


static const per_sequence_t SFNSFNMeasurementThresholdInformation_sequence[] = {
  { &hf_nbap_sFNSFNChangeLimit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SFNSFNChangeLimit },
  { &hf_nbap_predictedSFNSFNDeviationLimit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PredictedSFNSFNDeviationLimit },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SFNSFNMeasurementThresholdInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SFNSFNMeasurementThresholdInformation, SFNSFNMeasurementThresholdInformation_sequence);

  return offset;
}


static const value_string nbap_SFNSFNValue_vals[] = {
  {   0, "sFNSFN-FDD" },
  {   1, "sFNSFN-TDD" },
  {   2, "sFNSFN-TDD768" },
  { 0, NULL }
};

static const per_choice_t SFNSFNValue_choice[] = {
  {   0, &hf_nbap_sFNSFN_FDD     , ASN1_EXTENSION_ROOT    , dissect_nbap_SFNSFN_FDD },
  {   1, &hf_nbap_sFNSFN_TDD     , ASN1_EXTENSION_ROOT    , dissect_nbap_SFNSFN_TDD },
  {   2, &hf_nbap_sFNSFN_TDD768  , ASN1_NOT_EXTENSION_ROOT, dissect_nbap_SFNSFN_TDD768 },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_SFNSFNValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_SFNSFNValue, SFNSFNValue_choice,
                                 NULL);

  return offset;
}



static int
dissect_nbap_SFNSFNQuality(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t SFNSFNTimeStamp_TDD_sequence[] = {
  { &hf_nbap_sFN            , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SFN },
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SFNSFNTimeStamp_TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SFNSFNTimeStamp_TDD, SFNSFNTimeStamp_TDD_sequence);

  return offset;
}


static const value_string nbap_SFNSFNTimeStampInformation_vals[] = {
  {   0, "sFNSFNTimeStamp-FDD" },
  {   1, "sFNSFNTimeStamp-TDD" },
  { 0, NULL }
};

static const per_choice_t SFNSFNTimeStampInformation_choice[] = {
  {   0, &hf_nbap_sFNSFNTimeStamp_FDD, ASN1_EXTENSION_ROOT    , dissect_nbap_SFN },
  {   1, &hf_nbap_sFNSFNTimeStamp_TDD, ASN1_EXTENSION_ROOT    , dissect_nbap_SFNSFNTimeStamp_TDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_SFNSFNTimeStampInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_SFNSFNTimeStampInformation, SFNSFNTimeStampInformation_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item_sequence[] = {
  { &hf_nbap_uC_Id          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UC_Id },
  { &hf_nbap_sFNSFNValue    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SFNSFNValue },
  { &hf_nbap_sFNSFNQuality  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SFNSFNQuality },
  { &hf_nbap_sFNSFNDriftRate, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SFNSFNDriftRate },
  { &hf_nbap_sFNSFNDriftRateQuality, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SFNSFNDriftRateQuality },
  { &hf_nbap_sFNSFNTimeStampInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SFNSFNTimeStampInformation },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item, T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item_sequence);

  return offset;
}


static const per_sequence_t T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_sequence_of[1] = {
  { &hf_nbap_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item },
};

static int
dissect_nbap_T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation, T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_sequence_of,
                                                  1, maxNrOfMeasNCell, FALSE);

  return offset;
}


static const per_sequence_t T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item_sequence[] = {
  { &hf_nbap_uC_Id          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UC_Id },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item, T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item_sequence);

  return offset;
}


static const per_sequence_t T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_sequence_of[1] = {
  { &hf_nbap_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item },
};

static int
dissect_nbap_T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation, T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_sequence_of,
                                                  0, maxNrOfMeasNCell_1, FALSE);

  return offset;
}


static const per_sequence_t SFNSFNMeasurementValueInformation_sequence[] = {
  { &hf_nbap_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation },
  { &hf_nbap_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SFNSFNMeasurementValueInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SFNSFNMeasurementValueInformation, SFNSFNMeasurementValueInformation_sequence);

  return offset;
}



static int
dissect_nbap_ShutdownTimer(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 3600U, NULL, FALSE);

  return offset;
}


static const value_string nbap_SIB_Originator_vals[] = {
  {   0, "nodeB" },
  {   1, "cRNC" },
  { 0, NULL }
};


static int
dissect_nbap_SIB_Originator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_Single_Stream_MIMO_ActivationIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string nbap_Single_Stream_MIMO_Capability_vals[] = {
  {   0, "single-stream-mimo-capable" },
  {   1, "single-stream-mimo-non-capable" },
  { 0, NULL }
};


static int
dissect_nbap_Single_Stream_MIMO_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_Single_Stream_MIMO_Mode_Indicator_vals[] = {
  {   0, "activate" },
  {   1, "deactivate" },
  { 0, NULL }
};


static int
dissect_nbap_Single_Stream_MIMO_Mode_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_SignallingBearerRequestIndicator_vals[] = {
  {   0, "bearerRequested" },
  { 0, NULL }
};


static int
dissect_nbap_SignallingBearerRequestIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_SixtyfourQAM_DL_Capability_vals[] = {
  {   0, "sixtyfourQAM-DL-supported" },
  {   1, "sixtyfourQAM-DL-not-supported" },
  { 0, NULL }
};


static int
dissect_nbap_SixtyfourQAM_DL_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_SixtyfourQAM_DL_MIMO_Combined_Capability_vals[] = {
  {   0, "sixtyfourQAM-DL-MIMO-Combined-capable" },
  {   1, "sixtyfourQAM-DL-MIMO-Combined-non-capable" },
  { 0, NULL }
};


static int
dissect_nbap_SixtyfourQAM_DL_MIMO_Combined_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_SixteenQAM_UL_Capability_vals[] = {
  {   0, "sixteenQAM-UL-capable" },
  {   1, "sixteenQAM-UL-non-capable" },
  { 0, NULL }
};


static int
dissect_nbap_SixteenQAM_UL_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_SixteenQAM_UL_Operation_Indicator_vals[] = {
  {   0, "activate" },
  {   1, "deactivate" },
  { 0, NULL }
};


static int
dissect_nbap_SixteenQAM_UL_Operation_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_SNPL_Reporting_Type_vals[] = {
  {   0, "type1" },
  {   1, "type2" },
  { 0, NULL }
};


static int
dissect_nbap_SNPL_Reporting_Type(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_SpecialBurstScheduling(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 256U, NULL, FALSE);

  return offset;
}


static const value_string nbap_Start_Of_Audit_Sequence_Indicator_vals[] = {
  {   0, "start-of-audit-sequence" },
  {   1, "not-start-of-audit-sequence" },
  { 0, NULL }
};


static int
dissect_nbap_Start_Of_Audit_Sequence_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_SSDT_SupportIndicator_vals[] = {
  {   0, "not-Used-sSDT-Supported" },
  {   1, "sSDT-not-supported" },
  { 0, NULL }
};


static int
dissect_nbap_SSDT_SupportIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_SyncCase(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 2U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_SYNCDlCodeId(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 32U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_SyncFrameNumber(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 10U, NULL, FALSE);

  return offset;
}


static const value_string nbap_SynchronisationReportCharacteristicsType_vals[] = {
  {   0, "frameRelated" },
  {   1, "sFNperiodRelated" },
  {   2, "cycleLengthRelated" },
  {   3, "thresholdExceeding" },
  {   4, "frequencyAcquisitionCompleted" },
  { 0, NULL }
};


static int
dissect_nbap_SynchronisationReportCharacteristicsType(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     5, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t SynchronisationReportCharactCellSyncBurstInfoItem_sequence[] = {
  { &hf_nbap_cellSyncBurstCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellSyncBurstCode },
  { &hf_nbap_cellSyncBurstCodeShift, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellSyncBurstCodeShift },
  { &hf_nbap_cellSyncBurstTiming, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CellSyncBurstTiming },
  { &hf_nbap_cellSyncBurstTimingThreshold, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CellSyncBurstTimingThreshold },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SynchronisationReportCharactCellSyncBurstInfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SynchronisationReportCharactCellSyncBurstInfoItem, SynchronisationReportCharactCellSyncBurstInfoItem_sequence);

  return offset;
}


static const per_sequence_t SEQUENCE_SIZE_1_maxNrOfReceptsPerSyncFrame_OF_SynchronisationReportCharactCellSyncBurstInfoItem_sequence_of[1] = {
  { &hf_nbap_cellSyncBurstInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_SynchronisationReportCharactCellSyncBurstInfoItem },
};

static int
dissect_nbap_SEQUENCE_SIZE_1_maxNrOfReceptsPerSyncFrame_OF_SynchronisationReportCharactCellSyncBurstInfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_SEQUENCE_SIZE_1_maxNrOfReceptsPerSyncFrame_OF_SynchronisationReportCharactCellSyncBurstInfoItem, SEQUENCE_SIZE_1_maxNrOfReceptsPerSyncFrame_OF_SynchronisationReportCharactCellSyncBurstInfoItem_sequence_of,
                                                  1, maxNrOfReceptsPerSyncFrame, FALSE);

  return offset;
}


static const per_sequence_t SynchronisationReportCharactThreInfoItem_sequence[] = {
  { &hf_nbap_syncFrameNumber, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SyncFrameNumber },
  { &hf_nbap_cellSyncBurstInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SEQUENCE_SIZE_1_maxNrOfReceptsPerSyncFrame_OF_SynchronisationReportCharactCellSyncBurstInfoItem },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SynchronisationReportCharactThreInfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SynchronisationReportCharactThreInfoItem, SynchronisationReportCharactThreInfoItem_sequence);

  return offset;
}


static const per_sequence_t SynchronisationReportCharactThreExc_sequence_of[1] = {
  { &hf_nbap_SynchronisationReportCharactThreExc_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_SynchronisationReportCharactThreInfoItem },
};

static int
dissect_nbap_SynchronisationReportCharactThreExc(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_SynchronisationReportCharactThreExc, SynchronisationReportCharactThreExc_sequence_of,
                                                  1, maxNrOfCellSyncBursts, FALSE);

  return offset;
}


static const per_sequence_t SynchronisationReportCharacteristics_sequence[] = {
  { &hf_nbap_synchronisationReportCharacteristicsType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SynchronisationReportCharacteristicsType },
  { &hf_nbap_synchronisationReportCharactThreExc, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SynchronisationReportCharactThreExc },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SynchronisationReportCharacteristics(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SynchronisationReportCharacteristics, SynchronisationReportCharacteristics_sequence);

  return offset;
}


static const per_sequence_t SyncDLCodeInfoItemLCR_sequence[] = {
  { &hf_nbap_syncDLCodeId   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SYNCDlCodeId },
  { &hf_nbap_syncDLCodeIdArrivTime, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CellSyncBurstTimingLCR },
  { &hf_nbap_syncDLCodeIdTimingThre, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CellSyncBurstTimingThreshold },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SyncDLCodeInfoItemLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SyncDLCodeInfoItemLCR, SyncDLCodeInfoItemLCR_sequence);

  return offset;
}


static const per_sequence_t SyncDLCodeInfoListLCR_sequence_of[1] = {
  { &hf_nbap_SyncDLCodeInfoListLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_SyncDLCodeInfoItemLCR },
};

static int
dissect_nbap_SyncDLCodeInfoListLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_SyncDLCodeInfoListLCR, SyncDLCodeInfoListLCR_sequence_of,
                                                  1, maxNrOfSyncDLCodesLCR, FALSE);

  return offset;
}


static const per_sequence_t SyncDLCodeIdThreInfoList_sequence[] = {
  { &hf_nbap_syncFrameNoToReceive, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SyncFrameNumber },
  { &hf_nbap_syncDLCodeIdInfoLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SyncDLCodeInfoListLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SyncDLCodeIdThreInfoList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SyncDLCodeIdThreInfoList, SyncDLCodeIdThreInfoList_sequence);

  return offset;
}


static const per_sequence_t SyncDLCodeIdThreInfoLCR_sequence_of[1] = {
  { &hf_nbap_SyncDLCodeIdThreInfoLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_SyncDLCodeIdThreInfoList },
};

static int
dissect_nbap_SyncDLCodeIdThreInfoLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_SyncDLCodeIdThreInfoLCR, SyncDLCodeIdThreInfoLCR_sequence_of,
                                                  0, maxNrOfSyncFramesLCR, FALSE);

  return offset;
}


static const value_string nbap_SynchronisationReportType_vals[] = {
  {   0, "initialPhase" },
  {   1, "steadyStatePhase" },
  {   2, "lateEntrantCell" },
  {   3, "frequencyAcquisition" },
  { 0, NULL }
};


static int
dissect_nbap_SynchronisationReportType(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_Semi_PersistentScheduling_CapabilityLCR_vals[] = {
  {   0, "semi-Persistent-scheduling-Capable" },
  {   1, "semi-Persistent-scheduling-Non-Capable" },
  { 0, NULL }
};


static int
dissect_nbap_Semi_PersistentScheduling_CapabilityLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_T_Cell_vals[] = {
  {   0, "v0" },
  {   1, "v1" },
  {   2, "v2" },
  {   3, "v3" },
  {   4, "v4" },
  {   5, "v5" },
  {   6, "v6" },
  {   7, "v7" },
  {   8, "v8" },
  {   9, "v9" },
  { 0, NULL }
};


static int
dissect_nbap_T_Cell(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     10, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_T_RLFAILURE(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}


static const value_string nbap_TDD_DPCHOffset_vals[] = {
  {   0, "initialOffset" },
  {   1, "noinitialOffset" },
  { 0, NULL }
};

static const per_choice_t TDD_DPCHOffset_choice[] = {
  {   0, &hf_nbap_initialOffset  , ASN1_NO_EXTENSIONS     , dissect_nbap_INTEGER_0_255 },
  {   1, &hf_nbap_noinitialOffset, ASN1_NO_EXTENSIONS     , dissect_nbap_INTEGER_0_63 },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_TDD_DPCHOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_TDD_DPCHOffset, TDD_DPCHOffset_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t T_fdd_sequence[] = {
  { &hf_nbap_betaC          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BetaCD },
  { &hf_nbap_betaD          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BetaCD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_T_fdd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_T_fdd, T_fdd_sequence);

  return offset;
}


static const value_string nbap_T_gainFactor_vals[] = {
  {   0, "fdd" },
  {   1, "tdd" },
  { 0, NULL }
};

static const per_choice_t T_gainFactor_choice[] = {
  {   0, &hf_nbap_fdd            , ASN1_EXTENSION_ROOT    , dissect_nbap_T_fdd },
  {   1, &hf_nbap_tdd            , ASN1_EXTENSION_ROOT    , dissect_nbap_BetaCD },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_T_gainFactor(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_T_gainFactor, T_gainFactor_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t T_signalledGainFactors_sequence[] = {
  { &hf_nbap_gainFactor     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_gainFactor },
  { &hf_nbap_refTFCNumber   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RefTFCNumber },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_T_signalledGainFactors(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_T_signalledGainFactors, T_signalledGainFactors_sequence);

  return offset;
}


static const value_string nbap_TransportFormatCombination_Beta_vals[] = {
  {   0, "signalledGainFactors" },
  {   1, "computedGainFactors" },
  { 0, NULL }
};

static const per_choice_t TransportFormatCombination_Beta_choice[] = {
  {   0, &hf_nbap_signalledGainFactors, ASN1_EXTENSION_ROOT    , dissect_nbap_T_signalledGainFactors },
  {   1, &hf_nbap_computedGainFactors, ASN1_EXTENSION_ROOT    , dissect_nbap_RefTFCNumber },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_TransportFormatCombination_Beta(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_TransportFormatCombination_Beta, TransportFormatCombination_Beta_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t TDD_UL_Code_InformationItem_sequence[] = {
  { &hf_nbap_dPCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DPCH_ID },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TDD_UL_Code_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TDD_UL_Code_InformationItem, TDD_UL_Code_InformationItem_sequence);

  return offset;
}


static const per_sequence_t TDD_UL_Code_Information_sequence_of[1] = {
  { &hf_nbap_TDD_UL_Code_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_UL_Code_InformationItem },
};

static int
dissect_nbap_TDD_UL_Code_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_TDD_UL_Code_Information, TDD_UL_Code_Information_sequence_of,
                                                  1, maxNrOfDPCHs, FALSE);

  return offset;
}



static int
dissect_nbap_QPSK_UL_DPCH_TimeSlotFormatTDD_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 69U, NULL, TRUE);

  return offset;
}



static int
dissect_nbap_EightPSK_UL_DPCH_TimeSlotFormatTDD_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 24U, NULL, TRUE);

  return offset;
}


static const value_string nbap_TDD_UL_DPCH_TimeSlotFormat_LCR_vals[] = {
  {   0, "qPSK" },
  {   1, "eightPSK" },
  { 0, NULL }
};

static const per_choice_t TDD_UL_DPCH_TimeSlotFormat_LCR_choice[] = {
  {   0, &hf_nbap_qPSK_02        , ASN1_EXTENSION_ROOT    , dissect_nbap_QPSK_UL_DPCH_TimeSlotFormatTDD_LCR },
  {   1, &hf_nbap_eightPSK_01    , ASN1_EXTENSION_ROOT    , dissect_nbap_EightPSK_UL_DPCH_TimeSlotFormatTDD_LCR },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_TDD_UL_DPCH_TimeSlotFormat_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_TDD_UL_DPCH_TimeSlotFormat_LCR, TDD_UL_DPCH_TimeSlotFormat_LCR_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t TDD_UL_Code_LCR_InformationItem_sequence[] = {
  { &hf_nbap_dPCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DPCH_ID },
  { &hf_nbap_tdd_ChannelisationCodeLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCodeLCR },
  { &hf_nbap_tdd_UL_DPCH_TimeSlotFormat_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_UL_DPCH_TimeSlotFormat_LCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TDD_UL_Code_LCR_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TDD_UL_Code_LCR_InformationItem, TDD_UL_Code_LCR_InformationItem_sequence);

  return offset;
}


static const per_sequence_t TDD_UL_Code_LCR_Information_sequence_of[1] = {
  { &hf_nbap_TDD_UL_Code_LCR_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_UL_Code_LCR_InformationItem },
};

static int
dissect_nbap_TDD_UL_Code_LCR_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_TDD_UL_Code_LCR_Information, TDD_UL_Code_LCR_Information_sequence_of,
                                                  1, maxNrOfDPCHLCRs, FALSE);

  return offset;
}


static const per_sequence_t TDD_UL_Code_768_InformationItem_sequence[] = {
  { &hf_nbap_dPCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DPCH_ID },
  { &hf_nbap_tdd_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode768 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TDD_UL_Code_768_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TDD_UL_Code_768_InformationItem, TDD_UL_Code_768_InformationItem_sequence);

  return offset;
}


static const per_sequence_t TDD_UL_Code_768_Information_sequence_of[1] = {
  { &hf_nbap_TDD_UL_Code_768_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_UL_Code_768_InformationItem },
};

static int
dissect_nbap_TDD_UL_Code_768_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_TDD_UL_Code_768_Information, TDD_UL_Code_768_Information_sequence_of,
                                                  1, maxNrOfDPCHs, FALSE);

  return offset;
}


static const value_string nbap_TFCI_Coding_vals[] = {
  {   0, "v4" },
  {   1, "v8" },
  {   2, "v16" },
  {   3, "v32" },
  { 0, NULL }
};


static int
dissect_nbap_TFCI_Coding(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_TFCI_SignallingMode_TFCI_SignallingOption_vals[] = {
  {   0, "normal" },
  {   1, "not-Used-split" },
  { 0, NULL }
};


static int
dissect_nbap_TFCI_SignallingMode_TFCI_SignallingOption(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t TFCI_SignallingMode_sequence[] = {
  { &hf_nbap_tFCI_SignallingOption, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCI_SignallingMode_TFCI_SignallingOption },
  { &hf_nbap_not_Used_splitType, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_not_Used_lengthOfTFCI2, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TFCI_SignallingMode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TFCI_SignallingMode, TFCI_SignallingMode_sequence);

  return offset;
}



static int
dissect_nbap_TGD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 269U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_TGSN(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 14U, NULL, FALSE);

  return offset;
}


static const value_string nbap_TimeSlotDirection_vals[] = {
  {   0, "ul" },
  {   1, "dl" },
  { 0, NULL }
};


static int
dissect_nbap_TimeSlotDirection(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_TimeslotLCR_Extension_vals[] = {
  {   0, "ts7" },
  { 0, NULL }
};


static int
dissect_nbap_TimeslotLCR_Extension(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_TimeSlotStatus_vals[] = {
  {   0, "active" },
  {   1, "not-active" },
  { 0, NULL }
};


static int
dissect_nbap_TimeSlotStatus(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_TimingAdjustmentValue_vals[] = {
  {   0, "initialPhase" },
  {   1, "steadyStatePhase" },
  { 0, NULL }
};

static const per_choice_t TimingAdjustmentValue_choice[] = {
  {   0, &hf_nbap_initialPhase_01, ASN1_NO_EXTENSIONS     , dissect_nbap_INTEGER_0_1048575_ },
  {   1, &hf_nbap_steadyStatePhase, ASN1_NO_EXTENSIONS     , dissect_nbap_INTEGER_0_255_ },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_TimingAdjustmentValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_TimingAdjustmentValue, TimingAdjustmentValue_choice,
                                 NULL);

  return offset;
}


static const value_string nbap_TimingAdjustmentValueLCR_vals[] = {
  {   0, "initialPhase" },
  {   1, "steadyStatePhase" },
  { 0, NULL }
};

static const per_choice_t TimingAdjustmentValueLCR_choice[] = {
  {   0, &hf_nbap_initialPhase   , ASN1_NO_EXTENSIONS     , dissect_nbap_INTEGER_0_524287_ },
  {   1, &hf_nbap_steadyStatePhase_01, ASN1_NO_EXTENSIONS     , dissect_nbap_INTEGER_0_127_ },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_TimingAdjustmentValueLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_TimingAdjustmentValueLCR, TimingAdjustmentValueLCR_choice,
                                 NULL);

  return offset;
}


static const value_string nbap_TimingAdvanceApplied_vals[] = {
  {   0, "yes" },
  {   1, "no" },
  { 0, NULL }
};


static int
dissect_nbap_TimingAdvanceApplied(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_SynchronisationIndicator_vals[] = {
  {   0, "timingMaintainedSynchronisation" },
  { 0, NULL }
};


static int
dissect_nbap_SynchronisationIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string nbap_UL_DL_mode_vals[] = {
  {   0, "ul-only" },
  {   1, "dl-only" },
  {   2, "both-ul-and-dl" },
  { 0, NULL }
};


static int
dissect_nbap_UL_DL_mode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_Uplink_Compressed_Mode_Method_vals[] = {
  {   0, "sFdiv2" },
  {   1, "higher-layer-scheduling" },
  { 0, NULL }
};


static int
dissect_nbap_Uplink_Compressed_Mode_Method(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t Transmission_Gap_Pattern_Sequence_Information_item_sequence[] = {
  { &hf_nbap_tGPSID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TGPSID },
  { &hf_nbap_tGSN           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TGSN },
  { &hf_nbap_tGL1           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GapLength },
  { &hf_nbap_tGL2           , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GapLength },
  { &hf_nbap_tGD            , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TGD },
  { &hf_nbap_tGPL1          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_GapDuration },
  { &hf_nbap_not_to_be_used_1, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_GapDuration },
  { &hf_nbap_uL_DL_mode     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_DL_mode },
  { &hf_nbap_downlink_Compressed_Mode_Method, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Downlink_Compressed_Mode_Method },
  { &hf_nbap_uplink_Compressed_Mode_Method, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Uplink_Compressed_Mode_Method },
  { &hf_nbap_dL_FrameType   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_FrameType },
  { &hf_nbap_delta_SIR1     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DeltaSIR },
  { &hf_nbap_delta_SIR_after1, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DeltaSIR },
  { &hf_nbap_delta_SIR2     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DeltaSIR },
  { &hf_nbap_delta_SIR_after2, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DeltaSIR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Transmission_Gap_Pattern_Sequence_Information_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Transmission_Gap_Pattern_Sequence_Information_item, Transmission_Gap_Pattern_Sequence_Information_item_sequence);

  return offset;
}


static const per_sequence_t Transmission_Gap_Pattern_Sequence_Information_sequence_of[1] = {
  { &hf_nbap_Transmission_Gap_Pattern_Sequence_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Transmission_Gap_Pattern_Sequence_Information_item },
};

static int
dissect_nbap_Transmission_Gap_Pattern_Sequence_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Transmission_Gap_Pattern_Sequence_Information, Transmission_Gap_Pattern_Sequence_Information_sequence_of,
                                                  1, maxTGPS, FALSE);

  return offset;
}



static int
dissect_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHSTransmissionValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 100U, NULL, FALSE);

  return offset;
}


static const per_sequence_t TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue_Item_sequence[] = {
  { &hf_nbap_cellPortionID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellPortionID },
  { &hf_nbap_transmittedCarrierPowerOfAllCodesNotUsedForHSTransmissionValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHSTransmissionValue },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue_Item, TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue_Item_sequence);

  return offset;
}


static const per_sequence_t TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue_sequence_of[1] = {
  { &hf_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue_Item },
};

static int
dissect_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue, TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue_sequence_of,
                                                  1, maxNrOfCellPortionsPerCell, FALSE);

  return offset;
}


static const per_sequence_t TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue_Item_sequence[] = {
  { &hf_nbap_cellPortionLCRID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellPortionLCRID },
  { &hf_nbap_transmittedCarrierPowerOfAllCodesNotUsedForHSTransmissionValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHSTransmissionValue },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue_Item, TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue_Item_sequence);

  return offset;
}


static const per_sequence_t TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue_sequence_of[1] = {
  { &hf_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue_Item },
};

static int
dissect_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue, TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue_sequence_of,
                                                  1, maxNrOfCellPortionsPerCellLCR, FALSE);

  return offset;
}


static const per_sequence_t Transmitted_Carrier_Power_For_CellPortion_Value_Item_sequence[] = {
  { &hf_nbap_cellPortionID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellPortionID },
  { &hf_nbap_transmitted_Carrier_Power_Value, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Transmitted_Carrier_Power_Value },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Transmitted_Carrier_Power_For_CellPortion_Value_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Transmitted_Carrier_Power_For_CellPortion_Value_Item, Transmitted_Carrier_Power_For_CellPortion_Value_Item_sequence);

  return offset;
}


static const per_sequence_t Transmitted_Carrier_Power_For_CellPortion_Value_sequence_of[1] = {
  { &hf_nbap_Transmitted_Carrier_Power_For_CellPortion_Value_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Transmitted_Carrier_Power_For_CellPortion_Value_Item },
};

static int
dissect_nbap_Transmitted_Carrier_Power_For_CellPortion_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Transmitted_Carrier_Power_For_CellPortion_Value, Transmitted_Carrier_Power_For_CellPortion_Value_sequence_of,
                                                  1, maxNrOfCellPortionsPerCell, FALSE);

  return offset;
}


static const per_sequence_t Transmitted_Carrier_Power_For_CellPortion_ValueLCR_Item_sequence[] = {
  { &hf_nbap_cellPortionLCRID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellPortionLCRID },
  { &hf_nbap_transmitted_Carrier_Power_Value, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Transmitted_Carrier_Power_Value },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Transmitted_Carrier_Power_For_CellPortion_ValueLCR_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Transmitted_Carrier_Power_For_CellPortion_ValueLCR_Item, Transmitted_Carrier_Power_For_CellPortion_ValueLCR_Item_sequence);

  return offset;
}


static const per_sequence_t Transmitted_Carrier_Power_For_CellPortion_ValueLCR_sequence_of[1] = {
  { &hf_nbap_Transmitted_Carrier_Power_For_CellPortion_ValueLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Transmitted_Carrier_Power_For_CellPortion_ValueLCR_Item },
};

static int
dissect_nbap_Transmitted_Carrier_Power_For_CellPortion_ValueLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Transmitted_Carrier_Power_For_CellPortion_ValueLCR, Transmitted_Carrier_Power_For_CellPortion_ValueLCR_sequence_of,
                                                  1, maxNrOfCellPortionsPerCellLCR, FALSE);

  return offset;
}



static int
dissect_nbap_TransmissionDiversityApplied(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_boolean(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}


static const value_string nbap_TransmitDiversityIndicator_vals[] = {
  {   0, "active" },
  {   1, "inactive" },
  { 0, NULL }
};


static int
dissect_nbap_TransmitDiversityIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_INTEGER_0_maxCTFC(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxCTFC, NULL, FALSE);

  return offset;
}


static const value_string nbap_TFCS_CTFC_vals[] = {
  {   0, "ctfc2bit" },
  {   1, "ctfc4bit" },
  {   2, "ctfc6bit" },
  {   3, "ctfc8bit" },
  {   4, "ctfc12bit" },
  {   5, "ctfc16bit" },
  {   6, "ctfcmaxbit" },
  { 0, NULL }
};

static const per_choice_t TFCS_CTFC_choice[] = {
  {   0, &hf_nbap_ctfc2bit       , ASN1_NO_EXTENSIONS     , dissect_nbap_INTEGER_0_3 },
  {   1, &hf_nbap_ctfc4bit       , ASN1_NO_EXTENSIONS     , dissect_nbap_INTEGER_0_15 },
  {   2, &hf_nbap_ctfc6bit       , ASN1_NO_EXTENSIONS     , dissect_nbap_INTEGER_0_63 },
  {   3, &hf_nbap_ctfc8bit       , ASN1_NO_EXTENSIONS     , dissect_nbap_INTEGER_0_255 },
  {   4, &hf_nbap_ctfc12bit      , ASN1_NO_EXTENSIONS     , dissect_nbap_INTEGER_0_4095 },
  {   5, &hf_nbap_ctfc16bit      , ASN1_NO_EXTENSIONS     , dissect_nbap_INTEGER_0_65535 },
  {   6, &hf_nbap_ctfcmaxbit     , ASN1_NO_EXTENSIONS     , dissect_nbap_INTEGER_0_maxCTFC },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_TFCS_CTFC(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_TFCS_CTFC, TFCS_CTFC_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t TFCS_TFCSList_item_sequence[] = {
  { &hf_nbap_cTFC           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCS_CTFC },
  { &hf_nbap_tFC_Beta       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportFormatCombination_Beta },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TFCS_TFCSList_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TFCS_TFCSList_item, TFCS_TFCSList_item_sequence);

  return offset;
}


static const per_sequence_t TFCS_TFCSList_sequence_of[1] = {
  { &hf_nbap_TFCS_TFCSList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TFCS_TFCSList_item },
};

static int
dissect_nbap_TFCS_TFCSList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_TFCS_TFCSList, TFCS_TFCSList_sequence_of,
                                                  1, maxNrOfTFCs, FALSE);

  return offset;
}


static const value_string nbap_T_tFCSvalues_vals[] = {
  {   0, "no-Split-in-TFCI" },
  {   1, "not-Used-split-in-TFCI" },
  { 0, NULL }
};

static const per_choice_t T_tFCSvalues_choice[] = {
  {   0, &hf_nbap_no_Split_in_TFCI, ASN1_EXTENSION_ROOT    , dissect_nbap_TFCS_TFCSList },
  {   1, &hf_nbap_not_Used_split_in_TFCI, ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_T_tFCSvalues(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_T_tFCSvalues, T_tFCSvalues_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t TFCS_sequence[] = {
  { &hf_nbap_tFCSvalues     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_tFCSvalues },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TFCS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TFCS, TFCS_sequence);

  return offset;
}


static const value_string nbap_TransportBearerNotRequestedIndicator_vals[] = {
  {   0, "transport-bearer-shall-not-be-established" },
  {   1, "transport-bearer-may-not-be-established" },
  { 0, NULL }
};


static int
dissect_nbap_TransportBearerNotRequestedIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_TransportBearerNotSetupIndicator_vals[] = {
  {   0, "transport-bearer-not-setup" },
  { 0, NULL }
};


static int
dissect_nbap_TransportBearerNotSetupIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_TS0_CapabilityLCR_vals[] = {
  {   0, "tS0-Capable" },
  {   1, "tS0-Not-Capable" },
  { 0, NULL }
};


static int
dissect_nbap_TS0_CapabilityLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_TSTD_Indicator_vals[] = {
  {   0, "active" },
  {   1, "inactive" },
  { 0, NULL }
};


static int
dissect_nbap_TSTD_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_TSN_Length_vals[] = {
  {   0, "tsn-6bits" },
  {   1, "tsn-9bits" },
  { 0, NULL }
};


static int
dissect_nbap_TSN_Length(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_INTEGER_0_16383(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 16383U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_INTEGER_0_4294967295(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 4294967295U, NULL, FALSE);

  return offset;
}


static const per_sequence_t TUTRANGANSS_sequence[] = {
  { &hf_nbap_mS             , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_16383 },
  { &hf_nbap_lS             , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_4294967295 },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TUTRANGANSS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TUTRANGANSS, TUTRANGANSS_sequence);

  return offset;
}



static int
dissect_nbap_INTEGER_1_256(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 256U, NULL, FALSE);

  return offset;
}


static const per_sequence_t TUTRANGANSSMeasurementThresholdInformation_sequence[] = {
  { &hf_nbap_tUTRANGANSSChangeLimit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_INTEGER_1_256 },
  { &hf_nbap_predictedTUTRANGANSSDeviationLimit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_INTEGER_1_256 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TUTRANGANSSMeasurementThresholdInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TUTRANGANSSMeasurementThresholdInformation, TUTRANGANSSMeasurementThresholdInformation_sequence);

  return offset;
}



static int
dissect_nbap_INTEGER_M50_50(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -50, 50U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_INTEGER_0_50(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 50U, NULL, FALSE);

  return offset;
}


static const per_sequence_t TUTRANGANSSMeasurementValueInformation_sequence[] = {
  { &hf_nbap_tUTRANGANSS    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TUTRANGANSS },
  { &hf_nbap_tUTRANGANSSQuality, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_INTEGER_0_255 },
  { &hf_nbap_tUTRANGANSSDriftRate, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_M50_50 },
  { &hf_nbap_tUTRANGANSSDriftRateQuality, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_INTEGER_0_50 },
  { &hf_nbap_ie_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TUTRANGANSSMeasurementValueInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TUTRANGANSSMeasurementValueInformation, TUTRANGANSSMeasurementValueInformation_sequence);

  return offset;
}


static const per_sequence_t TUTRANGPS_sequence[] = {
  { &hf_nbap_ms_part        , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_16383 },
  { &hf_nbap_ls_part        , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_INTEGER_0_4294967295 },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TUTRANGPS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TUTRANGPS, TUTRANGPS_sequence);

  return offset;
}



static int
dissect_nbap_TUTRANGPSChangeLimit(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 256U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_TUTRANGPSDriftRate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -50, 50U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_TUTRANGPSDriftRateQuality(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 50U, NULL, FALSE);

  return offset;
}


static const per_sequence_t TUTRANGPSMeasurementThresholdInformation_sequence[] = {
  { &hf_nbap_tUTRANGPSChangeLimit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TUTRANGPSChangeLimit },
  { &hf_nbap_predictedTUTRANGPSDeviationLimit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PredictedTUTRANGPSDeviationLimit },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TUTRANGPSMeasurementThresholdInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TUTRANGPSMeasurementThresholdInformation, TUTRANGPSMeasurementThresholdInformation_sequence);

  return offset;
}



static int
dissect_nbap_TUTRANGPSQuality(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t TUTRANGPSMeasurementValueInformation_sequence[] = {
  { &hf_nbap_tUTRANGPS      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TUTRANGPS },
  { &hf_nbap_tUTRANGPSQuality, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TUTRANGPSQuality },
  { &hf_nbap_tUTRANGPSDriftRate, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TUTRANGPSDriftRate },
  { &hf_nbap_tUTRANGPSDriftRateQuality, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TUTRANGPSDriftRateQuality },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TUTRANGPSMeasurementValueInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TUTRANGPSMeasurementValueInformation, TUTRANGPSMeasurementValueInformation_sequence);

  return offset;
}


static const value_string nbap_TxDiversityOnDLControlChannelsByMIMOUECapability_vals[] = {
  {   0, "dL-Control-Channel-Tx-Diversity-for-MIMO-UE-with-non-diverse-P-CPICH-Capable" },
  {   1, "dL-Control-Channel-Tx-Diversity-for-MIMO-UE-with-non-diverse-P-CPICH-Not-Capable" },
  { 0, NULL }
};


static int
dissect_nbap_TxDiversityOnDLControlChannelsByMIMOUECapability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string nbap_TypeOfError_vals[] = {
  {   0, "not-understood" },
  {   1, "missing" },
  { 0, NULL }
};


static int
dissect_nbap_TypeOfError(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_nbap_UE_AggregateMaximumBitRateDownlink(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 1000000000U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_UE_AggregateMaximumBitRateUplink(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 1000000000U, NULL, FALSE);

  return offset;
}


static const per_sequence_t UE_AggregateMaximumBitRate_sequence[] = {
  { &hf_nbap_uE_AggregateMaximumBitRateDownlink, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UE_AggregateMaximumBitRateDownlink },
  { &hf_nbap_uE_AggregateMaximumBitRateUplink, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UE_AggregateMaximumBitRateUplink },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UE_AggregateMaximumBitRate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UE_AggregateMaximumBitRate, UE_AggregateMaximumBitRate_sequence);

  return offset;
}



static int
dissect_nbap_UE_AggregateMaximumBitRate_Enforcement_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string nbap_UE_TS0_CapabilityLCR_vals[] = {
  {   0, "uE-TS0-Capable" },
  {   1, "uE-TS0-Not-Capable" },
  { 0, NULL }
};


static int
dissect_nbap_UE_TS0_CapabilityLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_UE_SupportIndicatorExtension(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     32, 32, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_nbap_LCRTDD_HSDSCH_Physical_Layer_Category(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 64U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_UL_CapacityCredit(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 65535U, NULL, FALSE);

  return offset;
}


static const value_string nbap_UL_DPDCH_Indicator_For_E_DCH_Operation_vals[] = {
  {   0, "ul-DPDCH-present" },
  {   1, "ul-DPDCH-not-present" },
  { 0, NULL }
};


static int
dissect_nbap_UL_DPDCH_Indicator_For_E_DCH_Operation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t UL_Timeslot_InformationItem_sequence[] = {
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCI_Presence },
  { &hf_nbap_uL_Code_InformationList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_UL_Code_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_Timeslot_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_Timeslot_InformationItem, UL_Timeslot_InformationItem_sequence);

  return offset;
}


static const per_sequence_t UL_Timeslot_Information_sequence_of[1] = {
  { &hf_nbap_UL_Timeslot_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Timeslot_InformationItem },
};

static int
dissect_nbap_UL_Timeslot_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_Timeslot_Information, UL_Timeslot_Information_sequence_of,
                                                  1, maxNrOfULTSs, FALSE);

  return offset;
}


static const per_sequence_t UL_TimeslotLCR_InformationItem_sequence[] = {
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCI_Presence },
  { &hf_nbap_uL_Code_InformationList_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_UL_Code_LCR_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_TimeslotLCR_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_TimeslotLCR_InformationItem, UL_TimeslotLCR_InformationItem_sequence);

  return offset;
}


static const per_sequence_t UL_TimeslotLCR_Information_sequence_of[1] = {
  { &hf_nbap_UL_TimeslotLCR_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_TimeslotLCR_InformationItem },
};

static int
dissect_nbap_UL_TimeslotLCR_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_TimeslotLCR_Information, UL_TimeslotLCR_Information_sequence_of,
                                                  1, maxNrOfULTSLCRs, FALSE);

  return offset;
}


static const per_sequence_t UL_Timeslot768_InformationItem_sequence[] = {
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType768 },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCI_Presence },
  { &hf_nbap_uL_Code_InformationList_02, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_UL_Code_768_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_Timeslot768_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_Timeslot768_InformationItem, UL_Timeslot768_InformationItem_sequence);

  return offset;
}


static const per_sequence_t UL_Timeslot768_Information_sequence_of[1] = {
  { &hf_nbap_UL_Timeslot768_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Timeslot768_InformationItem },
};

static int
dissect_nbap_UL_Timeslot768_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_Timeslot768_Information, UL_Timeslot768_Information_sequence_of,
                                                  1, maxNrOfULTSs, FALSE);

  return offset;
}



static int
dissect_nbap_UL_DPCCH_SlotFormat(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 5U, NULL, TRUE);

  return offset;
}


static const value_string nbap_UL_PhysCH_SF_Variation_vals[] = {
  {   0, "sf-variation-supported" },
  {   1, "sf-variation-not-supported" },
  { 0, NULL }
};


static int
dissect_nbap_UL_PhysCH_SF_Variation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_UL_Synchronisation_StepSize(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 8U, NULL, FALSE);

  return offset;
}



static int
dissect_nbap_UL_Synchronisation_Frequency(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 8U, NULL, FALSE);

  return offset;
}


static const per_sequence_t UL_Synchronisation_Parameters_LCR_sequence[] = {
  { &hf_nbap_uL_Synchronisation_StepSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Synchronisation_StepSize },
  { &hf_nbap_uL_Synchronisation_Frequency, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Synchronisation_Frequency },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_Synchronisation_Parameters_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_Synchronisation_Parameters_LCR, UL_Synchronisation_Parameters_LCR_sequence);

  return offset;
}



static int
dissect_nbap_UPPCHPositionLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, NULL, FALSE);

  return offset;
}


static const per_sequence_t UL_TimeSlot_ISCP_InfoItem_sequence[] = {
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_iSCP           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_TimeslotISCP_Value },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_TimeSlot_ISCP_InfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_TimeSlot_ISCP_InfoItem, UL_TimeSlot_ISCP_InfoItem_sequence);

  return offset;
}


static const per_sequence_t UL_TimeSlot_ISCP_Info_sequence_of[1] = {
  { &hf_nbap_UL_TimeSlot_ISCP_Info_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_TimeSlot_ISCP_InfoItem },
};

static int
dissect_nbap_UL_TimeSlot_ISCP_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_TimeSlot_ISCP_Info, UL_TimeSlot_ISCP_Info_sequence_of,
                                                  1, maxNrOfULTSs, FALSE);

  return offset;
}


static const per_sequence_t UL_TimeSlot_ISCP_LCR_InfoItem_sequence[] = {
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_iSCP           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_TimeslotISCP_Value },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_TimeSlot_ISCP_LCR_InfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_TimeSlot_ISCP_LCR_InfoItem, UL_TimeSlot_ISCP_LCR_InfoItem_sequence);

  return offset;
}


static const per_sequence_t UL_TimeSlot_ISCP_LCR_Info_sequence_of[1] = {
  { &hf_nbap_UL_TimeSlot_ISCP_LCR_Info_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_TimeSlot_ISCP_LCR_InfoItem },
};

static int
dissect_nbap_UL_TimeSlot_ISCP_LCR_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_TimeSlot_ISCP_LCR_Info, UL_TimeSlot_ISCP_LCR_Info_sequence_of,
                                                  1, maxNrOfULTSLCRs, FALSE);

  return offset;
}



static int
dissect_nbap_UpPTSInterferenceValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, NULL, TRUE);

  return offset;
}


static const per_sequence_t UpPTSInterference_For_CellPortion_Value_Item_sequence[] = {
  { &hf_nbap_cellPortionLCRID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellPortionLCRID },
  { &hf_nbap_upPTSInterferenceValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UpPTSInterferenceValue },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UpPTSInterference_For_CellPortion_Value_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UpPTSInterference_For_CellPortion_Value_Item, UpPTSInterference_For_CellPortion_Value_Item_sequence);

  return offset;
}


static const per_sequence_t UpPTSInterference_For_CellPortion_Value_sequence_of[1] = {
  { &hf_nbap_UpPTSInterference_For_CellPortion_Value_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UpPTSInterference_For_CellPortion_Value_Item },
};

static int
dissect_nbap_UpPTSInterference_For_CellPortion_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UpPTSInterference_For_CellPortion_Value, UpPTSInterference_For_CellPortion_Value_sequence_of,
                                                  1, maxNrOfCellPortionsPerCellLCR, FALSE);

  return offset;
}


static const value_string nbap_Unidirectional_DCH_Indicator_vals[] = {
  {   0, "downlink-DCH-only" },
  {   1, "uplink-DCH-only" },
  { 0, NULL }
};


static int
dissect_nbap_Unidirectional_DCH_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_nbap_USCH_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t USCH_InformationItem_sequence[] = {
  { &hf_nbap_uSCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_USCH_ID },
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_transportFormatSet, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportFormatSet },
  { &hf_nbap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AllocationRetentionPriority },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_USCH_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_USCH_InformationItem, USCH_InformationItem_sequence);

  return offset;
}


static const per_sequence_t USCH_Information_sequence_of[1] = {
  { &hf_nbap_USCH_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_USCH_InformationItem },
};

static int
dissect_nbap_USCH_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_USCH_Information, USCH_Information_sequence_of,
                                                  1, maxNrOfUSCHs, FALSE);

  return offset;
}


static const per_sequence_t USCH_InformationResponseItem_sequence[] = {
  { &hf_nbap_uSCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_USCH_ID },
  { &hf_nbap_bindingID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BindingID },
  { &hf_nbap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportLayerAddress },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_USCH_InformationResponseItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_USCH_InformationResponseItem, USCH_InformationResponseItem_sequence);

  return offset;
}


static const per_sequence_t USCH_InformationResponse_sequence_of[1] = {
  { &hf_nbap_USCH_InformationResponse_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_USCH_InformationResponseItem },
};

static int
dissect_nbap_USCH_InformationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_USCH_InformationResponse, USCH_InformationResponse_sequence_of,
                                                  1, maxNrOfUSCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_TimeslotISCP_For_CellPortion_Value_Item_sequence[] = {
  { &hf_nbap_cellPortionLCRID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellPortionLCRID },
  { &hf_nbap_uL_TimeslotISCP_Value, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_TimeslotISCP_Value },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_TimeslotISCP_For_CellPortion_Value_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_TimeslotISCP_For_CellPortion_Value_Item, UL_TimeslotISCP_For_CellPortion_Value_Item_sequence);

  return offset;
}


static const per_sequence_t UL_TimeslotISCP_For_CellPortion_Value_sequence_of[1] = {
  { &hf_nbap_UL_TimeslotISCP_For_CellPortion_Value_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_TimeslotISCP_For_CellPortion_Value_Item },
};

static int
dissect_nbap_UL_TimeslotISCP_For_CellPortion_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_TimeslotISCP_For_CellPortion_Value, UL_TimeslotISCP_For_CellPortion_Value_sequence_of,
                                                  1, maxNrOfCellPortionsPerCellLCR, FALSE);

  return offset;
}


static const per_sequence_t CommonTransportChannelSetupRequestFDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommonTransportChannelSetupRequestFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"CommonTransportChannelSetupRequest(FDD) ");
/* CRNC -> Node B */
actx->pinfo->link_dir=P2P_DIR_DL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommonTransportChannelSetupRequestFDD, CommonTransportChannelSetupRequestFDD_sequence);

  return offset;
}


static const per_sequence_t PowerOffsetInformation_CTCH_SetupRqstFDD_sequence[] = {
  { &hf_nbap_pO1_ForTFCI_Bits, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PowerOffset },
  { &hf_nbap_pO3_ForPilotBits, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PowerOffset },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PowerOffsetInformation_CTCH_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PowerOffsetInformation_CTCH_SetupRqstFDD, PowerOffsetInformation_CTCH_SetupRqstFDD_sequence);

  return offset;
}



static int
dissect_nbap_FACH_ParametersList_CTCH_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_PCH_Parameters_CTCH_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_CTCH_SetupRqstFDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_fdd_S_CCPCH_Offset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_S_CCPCH_Offset },
  { &hf_nbap_dl_ScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_ScramblingCode },
  { &hf_nbap_fdd_DL_ChannelisationCodeNumber, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_DL_ChannelisationCodeNumber },
  { &hf_nbap_tFCS           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCS },
  { &hf_nbap_secondary_CCPCH_SlotFormat, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SecondaryCCPCH_SlotFormat },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCI_Presence },
  { &hf_nbap_multiplexingPosition, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MultiplexingPosition },
  { &hf_nbap_powerOffsetInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PowerOffsetInformation_CTCH_SetupRqstFDD },
  { &hf_nbap_sTTD_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_STTD_Indicator },
  { &hf_nbap_fACH_Parameters, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_FACH_ParametersList_CTCH_SetupRqstFDD },
  { &hf_nbap_pCH_Parameters , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PCH_Parameters_CTCH_SetupRqstFDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Secondary_CCPCH_CTCH_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Secondary_CCPCH_CTCH_SetupRqstFDD, Secondary_CCPCH_CTCH_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t AllowedSlotFormatInformationItem_CTCH_SetupRqstFDD_sequence[] = {
  { &hf_nbap_rACHSlotFormat , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RACH_SlotFormat },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_AllowedSlotFormatInformationItem_CTCH_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_AllowedSlotFormatInformationItem_CTCH_SetupRqstFDD, AllowedSlotFormatInformationItem_CTCH_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t AllowedSlotFormatInformationList_CTCH_SetupRqstFDD_sequence_of[1] = {
  { &hf_nbap_AllowedSlotFormatInformationList_CTCH_SetupRqstFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_AllowedSlotFormatInformationItem_CTCH_SetupRqstFDD },
};

static int
dissect_nbap_AllowedSlotFormatInformationList_CTCH_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_AllowedSlotFormatInformationList_CTCH_SetupRqstFDD, AllowedSlotFormatInformationList_CTCH_SetupRqstFDD_sequence_of,
                                                  1, maxNrOfSlotFormatsPRACH, FALSE);

  return offset;
}



static int
dissect_nbap_RACH_Parameters_CTCH_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t AICH_Parameters_CTCH_SetupRqstFDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_aICH_TransmissionTiming, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AICH_TransmissionTiming },
  { &hf_nbap_fdd_dl_ChannelisationCodeNumber, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_DL_ChannelisationCodeNumber },
  { &hf_nbap_aICH_Power     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AICH_Power },
  { &hf_nbap_sTTD_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_STTD_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_AICH_Parameters_CTCH_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_AICH_Parameters_CTCH_SetupRqstFDD, AICH_Parameters_CTCH_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t PRACH_CTCH_SetupRqstFDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_scramblingCodeNumber, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ScramblingCodeNumber },
  { &hf_nbap_tFCS           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCS },
  { &hf_nbap_preambleSignatures, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PreambleSignatures },
  { &hf_nbap_allowedSlotFormatInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AllowedSlotFormatInformationList_CTCH_SetupRqstFDD },
  { &hf_nbap_rACH_SubChannelNumbers, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RACH_SubChannelNumbers },
  { &hf_nbap_ul_punctureLimit, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PunctureLimit },
  { &hf_nbap_preambleThreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PreambleThreshold },
  { &hf_nbap_rACH_Parameters, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RACH_Parameters_CTCH_SetupRqstFDD },
  { &hf_nbap_aICH_Parameters, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AICH_Parameters_CTCH_SetupRqstFDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PRACH_CTCH_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PRACH_CTCH_SetupRqstFDD, PRACH_CTCH_SetupRqstFDD_sequence);

  return offset;
}


static const value_string nbap_CommonPhysicalChannelType_CTCH_SetupRqstFDD_vals[] = {
  {   0, "secondary-CCPCH-parameters" },
  {   1, "pRACH-parameters" },
  {   2, "notUsed-pCPCHes-parameters" },
  { 0, NULL }
};

static const per_choice_t CommonPhysicalChannelType_CTCH_SetupRqstFDD_choice[] = {
  {   0, &hf_nbap_secondary_CCPCH_parameters, ASN1_EXTENSION_ROOT    , dissect_nbap_Secondary_CCPCH_CTCH_SetupRqstFDD },
  {   1, &hf_nbap_pRACH_parameters, ASN1_EXTENSION_ROOT    , dissect_nbap_PRACH_CTCH_SetupRqstFDD },
  {   2, &hf_nbap_notUsed_pCPCHes_parameters, ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_CommonPhysicalChannelType_CTCH_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_CommonPhysicalChannelType_CTCH_SetupRqstFDD, CommonPhysicalChannelType_CTCH_SetupRqstFDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t FACH_ParametersItem_CTCH_SetupRqstFDD_sequence[] = {
  { &hf_nbap_commonTransportChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonTransportChannelID },
  { &hf_nbap_transportFormatSet, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportFormatSet },
  { &hf_nbap_toAWS          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ToAWS },
  { &hf_nbap_toAWE          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ToAWE },
  { &hf_nbap_maxFACH_Power  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_FACH_ParametersItem_CTCH_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
address dst_addr, null_addr;
conversation_t *conversation;

fp_fach_channel_info_t* fp_fach_channel_info;
umts_fp_conversation_info_t *umts_fp_conversation_info;
int i, j, num_tf;
guint32 transportLayerAddress_ipv4;
guint16 bindingID;
guint32 common_physical_channel_id;
nbap_dch_channel_info_t* nbap_dch_chnl_info;
nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
nbap_dch_chnl_info = nbap_private_data->nbap_dch_chnl_info;

nbap_private_data->transportLayerAddress_ipv4 = 0;
nbap_private_data->binding_id_port = 0;

nbap_private_data->transport_format_set_type = NBAP_CPCH;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_FACH_ParametersItem_CTCH_SetupRqstFDD, FACH_ParametersItem_CTCH_SetupRqstFDD_sequence);


  transportLayerAddress_ipv4 = nbap_private_data->transportLayerAddress_ipv4;
  bindingID = nbap_private_data->binding_id_port;
  if (PINFO_FD_VISITED(actx->pinfo) || transportLayerAddress_ipv4 == 0 || bindingID == 0){
    return offset;
  }
  clear_address(&null_addr);

  set_address(&dst_addr, AT_IPv4, 4, &transportLayerAddress_ipv4);

  conversation = conversation_new(actx->pinfo->num, &dst_addr, &null_addr, CONVERSATION_UDP, bindingID, 0, NO_ADDR2|NO_PORT2);

  /* Set dissector */
  conversation_set_dissector(conversation, fp_handle);

  if(actx->pinfo->link_dir==P2P_DIR_DL){
    umts_fp_conversation_info = wmem_new0(wmem_file_scope(), umts_fp_conversation_info_t);
    /* Fill in the data */
    umts_fp_conversation_info->iface_type        = IuB_Interface;
    umts_fp_conversation_info->division          = Division_FDD;
    umts_fp_conversation_info->channel           = CHANNEL_FACH_FDD;
    umts_fp_conversation_info->dl_frame_number   = 0;
    umts_fp_conversation_info->ul_frame_number   = actx->pinfo->num;
    copy_address_wmem(wmem_file_scope(), &(umts_fp_conversation_info->crnc_address), &dst_addr);
    umts_fp_conversation_info->crnc_port         = bindingID;
    umts_fp_conversation_info->rlc_mode          = FP_RLC_MODE_UNKNOWN;
    /*Save unique UE-identifier */
    umts_fp_conversation_info->com_context_id = nbap_private_data->crnc_context_present ? nbap_private_data->com_context_id : 1;

    /* Adding the 'channel specific info' for FACH */
    fp_fach_channel_info = wmem_new0(wmem_file_scope(), fp_fach_channel_info_t);
    fp_fach_channel_info->crnti_to_urnti_map = wmem_tree_new_autoreset(wmem_epan_scope(), wmem_file_scope());
    umts_fp_conversation_info->channel_specific_info = (void*)fp_fach_channel_info;

    /* DCH's in this flow */
    umts_fp_conversation_info->dch_crc_present = nbap_private_data->dch_crc_present;
    /* Set data for First or single channel */
    common_physical_channel_id = nbap_private_data->common_physical_channel_id;
    umts_fp_conversation_info->fp_dch_channel_info[0].num_ul_chans = num_tf = nbap_dch_chnl_info[common_physical_channel_id].num_ul_chans;
    for (j = 0; j < num_tf; j++) {
      umts_fp_conversation_info->fp_dch_channel_info[0].ul_chan_tf_size[j] = nbap_dch_chnl_info[common_physical_channel_id].ul_chan_tf_size[j];
      umts_fp_conversation_info->fp_dch_channel_info[0].ul_chan_num_tbs[j] = nbap_dch_chnl_info[common_physical_channel_id].ul_chan_num_tbs[j];
    }

    /* Traffic flows per DCH(DL) */
    umts_fp_conversation_info->fp_dch_channel_info[0].num_dl_chans = num_tf = nbap_dch_chnl_info[common_physical_channel_id].num_dl_chans;
    for (j = 0; j < num_tf; j++) {
      umts_fp_conversation_info->fp_dch_channel_info[0].dl_chan_tf_size[j] = nbap_dch_chnl_info[common_physical_channel_id].dl_chan_tf_size[j];
      umts_fp_conversation_info->fp_dch_channel_info[0].dl_chan_num_tbs[j] = nbap_dch_chnl_info[common_physical_channel_id].dl_chan_num_tbs[j];
    }

    /* Set data for associated DCH's if we have any */
    i = common_physical_channel_id;
    umts_fp_conversation_info->dch_ids_in_flow_list[0] = common_physical_channel_id;
    while(nbap_dch_chnl_info[i].next_dch != 0 && umts_fp_conversation_info->num_dch_in_flow < FP_maxNrOfDCHs){
      i = nbap_dch_chnl_info[i].next_dch;
      umts_fp_conversation_info->num_dch_in_flow++;
      umts_fp_conversation_info->dch_ids_in_flow_list[umts_fp_conversation_info->num_dch_in_flow] = i;
      /* Traffic flows per DCH(UL) */
      umts_fp_conversation_info->fp_dch_channel_info[umts_fp_conversation_info->num_dch_in_flow].num_ul_chans = num_tf = nbap_dch_chnl_info[i].num_ul_chans;
      for (j = 0; j < num_tf; j++) {
        umts_fp_conversation_info->fp_dch_channel_info[umts_fp_conversation_info->num_dch_in_flow].ul_chan_tf_size[j] = nbap_dch_chnl_info[i].ul_chan_tf_size[j];
        umts_fp_conversation_info->fp_dch_channel_info[umts_fp_conversation_info->num_dch_in_flow].ul_chan_num_tbs[j] = nbap_dch_chnl_info[i].ul_chan_num_tbs[j];
      }

      /* Traffic flows per DCH(DL) */
      umts_fp_conversation_info->fp_dch_channel_info[umts_fp_conversation_info->num_dch_in_flow].num_dl_chans = num_tf = nbap_dch_chnl_info[i].num_dl_chans;
      for (j = 0; j < num_tf; j++) {
        umts_fp_conversation_info->fp_dch_channel_info[umts_fp_conversation_info->num_dch_in_flow].dl_chan_tf_size[j] = nbap_dch_chnl_info[i].dl_chan_tf_size[j];
        umts_fp_conversation_info->fp_dch_channel_info[umts_fp_conversation_info->num_dch_in_flow].dl_chan_num_tbs[j] = nbap_dch_chnl_info[i].dl_chan_num_tbs[j];
      }
    }
    umts_fp_conversation_info->num_dch_in_flow++;
    umts_fp_conversation_info->dch_ids_in_flow_list[umts_fp_conversation_info->num_dch_in_flow] = i;
    set_umts_fp_conv_data(conversation, umts_fp_conversation_info);

    /* Add Setup Conversation to list, we need it in response msg */
    add_setup_conv(actx->pinfo, nbap_private_data->transaction_id,nbap_private_data->dd_mode,nbap_private_data->common_transport_channel_id, actx->pinfo->num, &dst_addr, bindingID, umts_fp_conversation_info, conversation);
  }


  return offset;
}


static const per_sequence_t FACH_ParametersListIE_CTCH_SetupRqstFDD_sequence_of[1] = {
  { &hf_nbap_FACH_ParametersListIE_CTCH_SetupRqstFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_FACH_ParametersItem_CTCH_SetupRqstFDD },
};

static int
dissect_nbap_FACH_ParametersListIE_CTCH_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_FACH_ParametersListIE_CTCH_SetupRqstFDD, FACH_ParametersListIE_CTCH_SetupRqstFDD_sequence_of,
                                                  1, maxNrOfFACHs, FALSE);

  return offset;
}



static int
dissect_nbap_T_transportFormatSet(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
    nbap_dch_channel_info_t* nbap_dch_chnl_info;
    nbap_dch_chnl_info = nbap_private_data->nbap_dch_chnl_info;
    nbap_private_data->transport_format_set_type = NBAP_PCH;
    nbap_dch_chnl_info[nbap_private_data->common_transport_channel_id].num_dl_chans = 0;
    nbap_dch_chnl_info[nbap_private_data->common_transport_channel_id].num_ul_chans = 0;

  offset = dissect_nbap_TransportFormatSet(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t PICH_Parameters_CTCH_SetupRqstFDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_fdd_dl_ChannelisationCodeNumber, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_DL_ChannelisationCodeNumber },
  { &hf_nbap_pICH_Power     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PICH_Power },
  { &hf_nbap_pICH_Mode      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PICH_Mode },
  { &hf_nbap_sTTD_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_STTD_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PICH_Parameters_CTCH_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PICH_Parameters_CTCH_SetupRqstFDD, PICH_Parameters_CTCH_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t PCH_ParametersItem_CTCH_SetupRqstFDD_sequence[] = {
  { &hf_nbap_commonTransportChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonTransportChannelID },
  { &hf_nbap_transportFormatSet_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_transportFormatSet },
  { &hf_nbap_toAWS          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ToAWS },
  { &hf_nbap_toAWE          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ToAWE },
  { &hf_nbap_pCH_Power      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_pICH_Parameters, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PICH_Parameters_CTCH_SetupRqstFDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PCH_ParametersItem_CTCH_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {

address dst_addr, null_addr;
conversation_t *conversation;

fp_pch_channel_info_t *fp_pch_channel_info;
umts_fp_conversation_info_t *umts_fp_conversation_info;
int i, j, num_tf;
guint32 transportLayerAddress_ipv4;
guint16 bindingID;
guint32 common_transport_channel_id;
nbap_dch_channel_info_t* nbap_dch_chnl_info;
nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
nbap_dch_chnl_info = nbap_private_data->nbap_dch_chnl_info;

nbap_private_data->transportLayerAddress_ipv4 = 0;
nbap_private_data->binding_id_port = 0;

/* There can only be one item, set num_items here to collect the TransportFormatSet data */
nbap_private_data->num_items = 1;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PCH_ParametersItem_CTCH_SetupRqstFDD, PCH_ParametersItem_CTCH_SetupRqstFDD_sequence);


  transportLayerAddress_ipv4 = nbap_private_data->transportLayerAddress_ipv4;
  bindingID = nbap_private_data->binding_id_port;
  if (PINFO_FD_VISITED(actx->pinfo) || transportLayerAddress_ipv4 == 0 || bindingID == 0){
    return offset;
  }
  clear_address(&null_addr);

  set_address(&dst_addr, AT_IPv4, 4, &transportLayerAddress_ipv4);

  conversation = conversation_new(actx->pinfo->num, &dst_addr, &null_addr, CONVERSATION_UDP, bindingID, 0, NO_ADDR2|NO_PORT2);

  /* Set dissector */
  conversation_set_dissector(conversation, fp_handle);

  if(actx->pinfo->link_dir==P2P_DIR_DL){
    umts_fp_conversation_info = wmem_new0(wmem_file_scope(), umts_fp_conversation_info_t);
    /* Fill in the data */
    umts_fp_conversation_info->iface_type         = IuB_Interface;
    umts_fp_conversation_info->division           = Division_FDD;
    umts_fp_conversation_info->channel            = CHANNEL_PCH;
    umts_fp_conversation_info->dl_frame_number    = 0;
    umts_fp_conversation_info->ul_frame_number    = actx->pinfo->num;
    copy_address_wmem(wmem_file_scope(), &(umts_fp_conversation_info->crnc_address), &dst_addr);
    umts_fp_conversation_info->crnc_port          = bindingID;
    umts_fp_conversation_info->rlc_mode           = FP_RLC_MODE_UNKNOWN;

    fp_pch_channel_info = wmem_new0(wmem_file_scope(), fp_pch_channel_info_t);
    fp_pch_channel_info->paging_indications = nbap_private_data->paging_indications;
    umts_fp_conversation_info->channel_specific_info = (void*)fp_pch_channel_info;

    /* DCH's in this flow */
    umts_fp_conversation_info->dch_crc_present = nbap_private_data->dch_crc_present;
    /* Set data for First or single channel */
    common_transport_channel_id = nbap_private_data->common_transport_channel_id;
    umts_fp_conversation_info->fp_dch_channel_info[0].num_ul_chans = num_tf = nbap_dch_chnl_info[common_transport_channel_id].num_ul_chans;

    nbap_debug("Frame %u PCH-ParametersItem-CTCH-SetupRqstFDD Start: num_tf %u", actx->pinfo->num, num_tf);

    for (j = 0; j < num_tf; j++) {
      umts_fp_conversation_info->fp_dch_channel_info[0].ul_chan_tf_size[j] = nbap_dch_chnl_info[common_transport_channel_id].ul_chan_tf_size[j];
      umts_fp_conversation_info->fp_dch_channel_info[0].ul_chan_num_tbs[j] = nbap_dch_chnl_info[common_transport_channel_id].ul_chan_num_tbs[j];
      nbap_debug("    UL tf %u ul_chan_tf_size %u",j, nbap_dch_chnl_info[common_transport_channel_id].ul_chan_tf_size[j]);
    }

    /* Traffic flows per DCH(DL) */
    umts_fp_conversation_info->fp_dch_channel_info[0].num_dl_chans = num_tf = nbap_dch_chnl_info[common_transport_channel_id].num_dl_chans;
    for (j = 0; j < num_tf; j++) {
      umts_fp_conversation_info->fp_dch_channel_info[0].dl_chan_tf_size[j] = nbap_dch_chnl_info[common_transport_channel_id].dl_chan_tf_size[j];
      umts_fp_conversation_info->fp_dch_channel_info[0].dl_chan_num_tbs[j] = nbap_dch_chnl_info[common_transport_channel_id].dl_chan_num_tbs[j];
      nbap_debug("    DL tf %u ul_chan_tf_size %u",j, nbap_dch_chnl_info[common_transport_channel_id].dl_chan_tf_size[j]);
    }

    /* Set data for associated DCH's if we have any */
    i = common_transport_channel_id;

    nbap_debug("    commontransportchannelid %u next ch %u",common_transport_channel_id, nbap_dch_chnl_info[i].next_dch);

    umts_fp_conversation_info->dch_ids_in_flow_list[0] = common_transport_channel_id;
    while(nbap_dch_chnl_info[i].next_dch != 0 && umts_fp_conversation_info->num_dch_in_flow < FP_maxNrOfDCHs){
      i = nbap_dch_chnl_info[i].next_dch;
      umts_fp_conversation_info->num_dch_in_flow++;
      umts_fp_conversation_info->dch_ids_in_flow_list[umts_fp_conversation_info->num_dch_in_flow] = i;
      /* Traffic flows per DCH(UL) */
      umts_fp_conversation_info->fp_dch_channel_info[umts_fp_conversation_info->num_dch_in_flow].num_ul_chans = num_tf = nbap_dch_chnl_info[i].num_ul_chans;
      for (j = 0; j < num_tf; j++) {
        umts_fp_conversation_info->fp_dch_channel_info[umts_fp_conversation_info->num_dch_in_flow].ul_chan_tf_size[j] = nbap_dch_chnl_info[i].ul_chan_tf_size[j];
        umts_fp_conversation_info->fp_dch_channel_info[umts_fp_conversation_info->num_dch_in_flow].ul_chan_num_tbs[j] = nbap_dch_chnl_info[i].ul_chan_num_tbs[j];
      }

      /* Traffic flows per DCH(DL) */
      umts_fp_conversation_info->fp_dch_channel_info[umts_fp_conversation_info->num_dch_in_flow].num_dl_chans = num_tf = nbap_dch_chnl_info[i].num_dl_chans;
      for (j = 0; j < num_tf; j++) {
        umts_fp_conversation_info->fp_dch_channel_info[umts_fp_conversation_info->num_dch_in_flow].dl_chan_tf_size[j] = nbap_dch_chnl_info[i].dl_chan_tf_size[j];
        umts_fp_conversation_info->fp_dch_channel_info[umts_fp_conversation_info->num_dch_in_flow].dl_chan_num_tbs[j] = nbap_dch_chnl_info[i].dl_chan_num_tbs[j];
      }
    }
    umts_fp_conversation_info->num_dch_in_flow++;

    nbap_debug("    num_dch_in_flow %u", umts_fp_conversation_info->num_dch_in_flow);

    umts_fp_conversation_info->dch_ids_in_flow_list[umts_fp_conversation_info->num_dch_in_flow] = i;
    set_umts_fp_conv_data(conversation, umts_fp_conversation_info);

    /* Add Setup Conversation to list, we need it in response msg */
    add_setup_conv(actx->pinfo, nbap_private_data->transaction_id, nbap_private_data->dd_mode, common_transport_channel_id, actx->pinfo->num, &dst_addr, bindingID, umts_fp_conversation_info, conversation);

    nbap_debug("Frame %u PCH-ParametersItem-CTCH-SetupRqstFDD End", actx->pinfo->num);
  }


  return offset;
}


static const per_sequence_t MICH_Parameters_CTCH_SetupRqstFDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_fdd_dl_ChannelisationCodeNumber, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_DL_ChannelisationCodeNumber },
  { &hf_nbap_mICH_Power     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PICH_Power },
  { &hf_nbap_mICH_Mode      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MICH_Mode },
  { &hf_nbap_sTTD_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_STTD_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MICH_Parameters_CTCH_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MICH_Parameters_CTCH_SetupRqstFDD, MICH_Parameters_CTCH_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t RACH_ParametersItem_CTCH_SetupRqstFDD_sequence[] = {
  { &hf_nbap_commonTransportChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonTransportChannelID },
  { &hf_nbap_transportFormatSet, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportFormatSet },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RACH_ParametersItem_CTCH_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
address dst_addr, null_addr;
conversation_t *conversation;
fp_rach_channel_info_t* fp_rach_channel_info;
umts_fp_conversation_info_t *umts_fp_conversation_info;
int j, num_tf;
guint32 transportLayerAddress_ipv4;
guint16 bindingID;
guint32 common_physical_channel_id;
nbap_dch_channel_info_t* nbap_dch_chnl_info;
nbap_private_data_t* nbap_private_data = nbap_get_private_data(actx->pinfo);
nbap_dch_chnl_info = nbap_private_data->nbap_dch_chnl_info;

nbap_private_data->transportLayerAddress_ipv4 = 0;
nbap_private_data->binding_id_port = 0;
nbap_private_data->transport_format_set_type = NBAP_CPCH;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RACH_ParametersItem_CTCH_SetupRqstFDD, RACH_ParametersItem_CTCH_SetupRqstFDD_sequence);


  transportLayerAddress_ipv4 = nbap_private_data->transportLayerAddress_ipv4;
  bindingID = nbap_private_data->binding_id_port;
  if (PINFO_FD_VISITED(actx->pinfo) || transportLayerAddress_ipv4 == 0 || bindingID == 0){
    return offset;
  }
  clear_address(&null_addr);

  set_address(&dst_addr, AT_IPv4, 4, &transportLayerAddress_ipv4);

  conversation = conversation_new(actx->pinfo->num, &dst_addr, &null_addr, CONVERSATION_UDP, bindingID, 0, NO_ADDR2|NO_PORT2);

  conversation_set_dissector(conversation, fp_handle);

  if(actx->pinfo->link_dir==P2P_DIR_DL){
    umts_fp_conversation_info = wmem_new0(wmem_file_scope(), umts_fp_conversation_info_t);
    /* Fill in the data */
    umts_fp_conversation_info->iface_type        = IuB_Interface;
    umts_fp_conversation_info->division          = Division_FDD;
    umts_fp_conversation_info->channel           = CHANNEL_RACH_FDD;
    umts_fp_conversation_info->dl_frame_number   = 0;
    umts_fp_conversation_info->ul_frame_number   = actx->pinfo->num;
    copy_address_wmem(wmem_file_scope(), &(umts_fp_conversation_info->crnc_address), &dst_addr);
    umts_fp_conversation_info->crnc_port         = bindingID;
    umts_fp_conversation_info->rlc_mode          = FP_RLC_MODE_UNKNOWN;

    /* Adding the 'channel specific info' for RACH */
    fp_rach_channel_info = wmem_new0(wmem_file_scope(), fp_rach_channel_info_t);
    fp_rach_channel_info->crnti_to_urnti_map = wmem_tree_new_autoreset(wmem_epan_scope(), wmem_file_scope());
    umts_fp_conversation_info->channel_specific_info = (void*)fp_rach_channel_info;

    /*Save unique UE-identifier */
    umts_fp_conversation_info->com_context_id = nbap_private_data->crnc_context_present ? nbap_private_data->com_context_id : 1;

    /* DCH's in this flow */
    umts_fp_conversation_info->dch_crc_present = nbap_private_data->dch_crc_present;
    /* Set data for First or single channel */
    common_physical_channel_id = nbap_private_data->common_physical_channel_id;
    umts_fp_conversation_info->fp_dch_channel_info[0].num_ul_chans = num_tf = nbap_dch_chnl_info[common_physical_channel_id].num_ul_chans;
    for (j = 0; j < num_tf; j++) {
      umts_fp_conversation_info->fp_dch_channel_info[0].ul_chan_tf_size[j] = nbap_dch_chnl_info[common_physical_channel_id].ul_chan_tf_size[j];
      umts_fp_conversation_info->fp_dch_channel_info[0].ul_chan_num_tbs[j] = nbap_dch_chnl_info[common_physical_channel_id].ul_chan_num_tbs[j];
    }

    /* Traffic flows per DCH(DL) */
    umts_fp_conversation_info->fp_dch_channel_info[0].num_dl_chans = num_tf = nbap_dch_chnl_info[common_physical_channel_id].num_dl_chans;
    for (j = 0; j < num_tf; j++) {
      umts_fp_conversation_info->fp_dch_channel_info[0].dl_chan_tf_size[j] = nbap_dch_chnl_info[common_physical_channel_id].dl_chan_tf_size[j];
      umts_fp_conversation_info->fp_dch_channel_info[0].dl_chan_num_tbs[j] = nbap_dch_chnl_info[common_physical_channel_id].dl_chan_num_tbs[j];
    }

    umts_fp_conversation_info->dch_ids_in_flow_list[0] = common_physical_channel_id;
    umts_fp_conversation_info->num_dch_in_flow=1;
    set_umts_fp_conv_data(conversation, umts_fp_conversation_info);

    /* Add Setup Conversation to list, we need it in response msg */
    add_setup_conv(actx->pinfo, nbap_private_data->transaction_id,nbap_private_data->dd_mode,nbap_private_data->common_transport_channel_id, actx->pinfo->num, &dst_addr, bindingID, umts_fp_conversation_info, conversation);
  }


  return offset;
}


static const per_sequence_t CommonTransportChannelSetupRequestTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommonTransportChannelSetupRequestTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommonTransportChannelSetupRequestTDD, CommonTransportChannelSetupRequestTDD_sequence);

  return offset;
}



static int
dissect_nbap_Secondary_CCPCH_parameterList_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_FACH_ParametersList_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_PCH_Parameters_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_CTCH_SetupRqstTDD_sequence[] = {
  { &hf_nbap_sCCPCH_CCTrCH_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_tFCS           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCS },
  { &hf_nbap_tFCI_Coding    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCI_Coding },
  { &hf_nbap_punctureLimit  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PunctureLimit },
  { &hf_nbap_secondaryCCPCH_parameterList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Secondary_CCPCH_parameterList_CTCH_SetupRqstTDD },
  { &hf_nbap_fACH_ParametersList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_FACH_ParametersList_CTCH_SetupRqstTDD },
  { &hf_nbap_pCH_Parameters_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PCH_Parameters_CTCH_SetupRqstTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Secondary_CCPCH_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Secondary_CCPCH_CTCH_SetupRqstTDD, Secondary_CCPCH_CTCH_SetupRqstTDD_sequence);

  return offset;
}



static int
dissect_nbap_PRACH_Parameters_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t PRACH_CTCH_SetupRqstTDD_sequence[] = {
  { &hf_nbap_pRACH_Parameters_CTCH_SetupRqstTDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PRACH_Parameters_CTCH_SetupRqstTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PRACH_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PRACH_CTCH_SetupRqstTDD, PRACH_CTCH_SetupRqstTDD_sequence);

  return offset;
}



static int
dissect_nbap_Extension_CommonPhysicalChannelType_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string nbap_CommonPhysicalChannelType_CTCH_SetupRqstTDD_vals[] = {
  {   0, "secondary-CCPCH-parameters" },
  {   1, "pRACH-parameters" },
  {   2, "extension-CommonPhysicalChannelType-CTCH-SetupRqstTDD" },
  { 0, NULL }
};

static const per_choice_t CommonPhysicalChannelType_CTCH_SetupRqstTDD_choice[] = {
  {   0, &hf_nbap_secondary_CCPCH_parameters_01, ASN1_EXTENSION_ROOT    , dissect_nbap_Secondary_CCPCH_CTCH_SetupRqstTDD },
  {   1, &hf_nbap_pRACH_parameters_01, ASN1_EXTENSION_ROOT    , dissect_nbap_PRACH_CTCH_SetupRqstTDD },
  {   2, &hf_nbap_extension_CommonPhysicalChannelType_CTCH_SetupRqstTDD, ASN1_NOT_EXTENSION_ROOT, dissect_nbap_Extension_CommonPhysicalChannelType_CTCH_SetupRqstTDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_CommonPhysicalChannelType_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_CommonPhysicalChannelType_CTCH_SetupRqstTDD, CommonPhysicalChannelType_CTCH_SetupRqstTDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_parameterItem_CTCH_SetupRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_timeslot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftandBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_tdd_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_s_CCPCH_Power  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Secondary_CCPCH_parameterItem_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Secondary_CCPCH_parameterItem_CTCH_SetupRqstTDD, Secondary_CCPCH_parameterItem_CTCH_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_parameterListIE_CTCH_SetupRqstTDD_sequence_of[1] = {
  { &hf_nbap_Secondary_CCPCH_parameterListIE_CTCH_SetupRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Secondary_CCPCH_parameterItem_CTCH_SetupRqstTDD },
};

static int
dissect_nbap_Secondary_CCPCH_parameterListIE_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Secondary_CCPCH_parameterListIE_CTCH_SetupRqstTDD, Secondary_CCPCH_parameterListIE_CTCH_SetupRqstTDD_sequence_of,
                                                  1, maxNrOfSCCPCHs, FALSE);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_LCR_parameterItem_CTCH_SetupRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_tdd_ChannelisationCodeLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCodeLCR },
  { &hf_nbap_timeslotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_tdd_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_s_CCPCH_Power  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_s_CCPCH_TimeSlotFormat_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DL_DPCH_TimeSlotFormat_LCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Secondary_CCPCH_LCR_parameterItem_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Secondary_CCPCH_LCR_parameterItem_CTCH_SetupRqstTDD, Secondary_CCPCH_LCR_parameterItem_CTCH_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_LCR_parameterList_CTCH_SetupRqstTDD_sequence_of[1] = {
  { &hf_nbap_Secondary_CCPCH_LCR_parameterList_CTCH_SetupRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Secondary_CCPCH_LCR_parameterItem_CTCH_SetupRqstTDD },
};

static int
dissect_nbap_Secondary_CCPCH_LCR_parameterList_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Secondary_CCPCH_LCR_parameterList_CTCH_SetupRqstTDD, Secondary_CCPCH_LCR_parameterList_CTCH_SetupRqstTDD_sequence_of,
                                                  1, maxNrOfSCCPCHLCRs, FALSE);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_768_parameterItem_CTCH_SetupRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID768 },
  { &hf_nbap_tdd_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode768 },
  { &hf_nbap_timeslot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_tFCI_Presence768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCI_Presence },
  { &hf_nbap_midambleShiftandBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType768 },
  { &hf_nbap_tdd_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_s_CCPCH_Power  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Secondary_CCPCH_768_parameterItem_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Secondary_CCPCH_768_parameterItem_CTCH_SetupRqstTDD, Secondary_CCPCH_768_parameterItem_CTCH_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_768_parameterList_CTCH_SetupRqstTDD_sequence_of[1] = {
  { &hf_nbap_Secondary_CCPCH_768_parameterList_CTCH_SetupRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Secondary_CCPCH_768_parameterItem_CTCH_SetupRqstTDD },
};

static int
dissect_nbap_Secondary_CCPCH_768_parameterList_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Secondary_CCPCH_768_parameterList_CTCH_SetupRqstTDD, Secondary_CCPCH_768_parameterList_CTCH_SetupRqstTDD_sequence_of,
                                                  1, maxNrOfSCCPCHs768, FALSE);

  return offset;
}


static const per_sequence_t FACH_ParametersItem_CTCH_SetupRqstTDD_sequence[] = {
  { &hf_nbap_commonTransportChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonTransportChannelID },
  { &hf_nbap_fACH_CCTrCH_ID , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_dl_TransportFormatSet_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportFormatSet },
  { &hf_nbap_toAWS          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ToAWS },
  { &hf_nbap_toAWE          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ToAWE },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_FACH_ParametersItem_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_FACH_ParametersItem_CTCH_SetupRqstTDD, FACH_ParametersItem_CTCH_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t FACH_ParametersListIE_CTCH_SetupRqstTDD_sequence_of[1] = {
  { &hf_nbap_FACH_ParametersListIE_CTCH_SetupRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_FACH_ParametersItem_CTCH_SetupRqstTDD },
};

static int
dissect_nbap_FACH_ParametersListIE_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_FACH_ParametersListIE_CTCH_SetupRqstTDD, FACH_ParametersListIE_CTCH_SetupRqstTDD_sequence_of,
                                                  1, maxNrOfFACHs, FALSE);

  return offset;
}



static int
dissect_nbap_PICH_Parameters_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t PCH_ParametersItem_CTCH_SetupRqstTDD_sequence[] = {
  { &hf_nbap_commonTransportChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonTransportChannelID },
  { &hf_nbap_pCH_CCTrCH_ID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_dl_TransportFormatSet_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportFormatSet },
  { &hf_nbap_toAWS          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ToAWS },
  { &hf_nbap_toAWE          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ToAWE },
  { &hf_nbap_pICH_Parameters_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PICH_Parameters_CTCH_SetupRqstTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PCH_ParametersItem_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PCH_ParametersItem_CTCH_SetupRqstTDD, PCH_ParametersItem_CTCH_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t PICH_ParametersItem_CTCH_SetupRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleshiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_tdd_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_pagingIndicatorLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PagingIndicatorLength },
  { &hf_nbap_pICH_Power     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PICH_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PICH_ParametersItem_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PICH_ParametersItem_CTCH_SetupRqstTDD, PICH_ParametersItem_CTCH_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t PICH_LCR_Parameters_CTCH_SetupRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_tdd_ChannelisationCodeLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCodeLCR },
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_tdd_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_pagingIndicatorLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PagingIndicatorLength },
  { &hf_nbap_pICH_Power     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PICH_Power },
  { &hf_nbap_second_TDD_ChannelisationCodeLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCodeLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PICH_LCR_Parameters_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PICH_LCR_Parameters_CTCH_SetupRqstTDD, PICH_LCR_Parameters_CTCH_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t PICH_768_ParametersItem_CTCH_SetupRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID768 },
  { &hf_nbap_tdd_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode768 },
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleshiftAndBurstType78, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType768 },
  { &hf_nbap_tdd_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_pagingIndicatorLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PagingIndicatorLength },
  { &hf_nbap_pICH_Power     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PICH_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PICH_768_ParametersItem_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PICH_768_ParametersItem_CTCH_SetupRqstTDD, PICH_768_ParametersItem_CTCH_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t MICH_HCR_Parameters_CTCH_SetupRqstTDD_sequence[] = {
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleshiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MICH_HCR_Parameters_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MICH_HCR_Parameters_CTCH_SetupRqstTDD, MICH_HCR_Parameters_CTCH_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t MICH_LCR_Parameters_CTCH_SetupRqstTDD_sequence[] = {
  { &hf_nbap_tdd_ChannelisationCodeLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCodeLCR },
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_second_TDD_ChannelisationCodeLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCodeLCR },
  { &hf_nbap_tSTD_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TSTD_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MICH_LCR_Parameters_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MICH_LCR_Parameters_CTCH_SetupRqstTDD, MICH_LCR_Parameters_CTCH_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t MICH_768_Parameters_CTCH_SetupRqstTDD_sequence[] = {
  { &hf_nbap_tdd_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode768 },
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleshiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType768 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MICH_768_Parameters_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MICH_768_Parameters_CTCH_SetupRqstTDD, MICH_768_Parameters_CTCH_SetupRqstTDD_sequence);

  return offset;
}


static const value_string nbap_MICH_TDDOption_Specific_Parameters_CTCH_SetupRqstTDD_vals[] = {
  {   0, "hCR-TDD" },
  {   1, "lCR-TDD" },
  {   2, "cHipRate768-TDD" },
  { 0, NULL }
};

static const per_choice_t MICH_TDDOption_Specific_Parameters_CTCH_SetupRqstTDD_choice[] = {
  {   0, &hf_nbap_hCR_TDD        , ASN1_EXTENSION_ROOT    , dissect_nbap_MICH_HCR_Parameters_CTCH_SetupRqstTDD },
  {   1, &hf_nbap_lCR_TDD        , ASN1_EXTENSION_ROOT    , dissect_nbap_MICH_LCR_Parameters_CTCH_SetupRqstTDD },
  {   2, &hf_nbap_cHipRate768_TDD, ASN1_NOT_EXTENSION_ROOT, dissect_nbap_MICH_768_Parameters_CTCH_SetupRqstTDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_MICH_TDDOption_Specific_Parameters_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_MICH_TDDOption_Specific_Parameters_CTCH_SetupRqstTDD, MICH_TDDOption_Specific_Parameters_CTCH_SetupRqstTDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t MICH_Parameters_CTCH_SetupRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_tdd_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_notificationIndicatorLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_NotificationIndicatorLength },
  { &hf_nbap_mICH_Power     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PICH_Power },
  { &hf_nbap_mICH_TDDOption_Specific_Parameters, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MICH_TDDOption_Specific_Parameters_CTCH_SetupRqstTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MICH_Parameters_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MICH_Parameters_CTCH_SetupRqstTDD, MICH_Parameters_CTCH_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t TimeSlotConfigurationItem_LCR_CTCH_SetupRqstTDD_sequence[] = {
  { &hf_nbap_timeslotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_timeslotLCR_Parameter_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellParameterID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TimeSlotConfigurationItem_LCR_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TimeSlotConfigurationItem_LCR_CTCH_SetupRqstTDD, TimeSlotConfigurationItem_LCR_CTCH_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t TimeSlotConfigurationList_LCR_CTCH_SetupRqstTDD_sequence_of[1] = {
  { &hf_nbap_TimeSlotConfigurationList_LCR_CTCH_SetupRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotConfigurationItem_LCR_CTCH_SetupRqstTDD },
};

static int
dissect_nbap_TimeSlotConfigurationList_LCR_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_TimeSlotConfigurationList_LCR_CTCH_SetupRqstTDD, TimeSlotConfigurationList_LCR_CTCH_SetupRqstTDD_sequence_of,
                                                  1, 7, FALSE);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_parameterExtendedList_CTCH_SetupRqstTDD_sequence_of[1] = {
  { &hf_nbap_Secondary_CCPCH_parameterExtendedList_CTCH_SetupRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Secondary_CCPCH_parameterItem_CTCH_SetupRqstTDD },
};

static int
dissect_nbap_Secondary_CCPCH_parameterExtendedList_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Secondary_CCPCH_parameterExtendedList_CTCH_SetupRqstTDD, Secondary_CCPCH_parameterExtendedList_CTCH_SetupRqstTDD_sequence_of,
                                                  1, maxNrOfSCCPCHsinExt, FALSE);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_LCR_parameterExtendedList_CTCH_SetupRqstTDD_sequence_of[1] = {
  { &hf_nbap_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_SetupRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Secondary_CCPCH_LCR_parameterItem_CTCH_SetupRqstTDD },
};

static int
dissect_nbap_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_SetupRqstTDD, Secondary_CCPCH_LCR_parameterExtendedList_CTCH_SetupRqstTDD_sequence_of,
                                                  1, maxNrOfSCCPCHsLCRinExt, FALSE);

  return offset;
}



static int
dissect_nbap_RACH_Parameter_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t PRACH_ParametersItem_CTCH_SetupRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_tFCS           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCS },
  { &hf_nbap_timeslot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_maxPRACH_MidambleShifts, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MaxPRACH_MidambleShifts },
  { &hf_nbap_pRACH_Midamble , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PRACH_Midamble },
  { &hf_nbap_rACH           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RACH_Parameter_CTCH_SetupRqstTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PRACH_ParametersItem_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PRACH_ParametersItem_CTCH_SetupRqstTDD, PRACH_ParametersItem_CTCH_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t RACH_ParameterItem_CTCH_SetupRqstTDD_sequence[] = {
  { &hf_nbap_commonTransportChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonTransportChannelID },
  { &hf_nbap_uL_TransportFormatSet, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportFormatSet },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RACH_ParameterItem_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RACH_ParameterItem_CTCH_SetupRqstTDD, RACH_ParameterItem_CTCH_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t PRACH_LCR_ParametersItem_CTCH_SetupRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_tFCS           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCS },
  { &hf_nbap_timeslotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_tdd_ChannelisationCodeLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCodeLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_rACH           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RACH_Parameter_CTCH_SetupRqstTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PRACH_LCR_ParametersItem_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PRACH_LCR_ParametersItem_CTCH_SetupRqstTDD, PRACH_LCR_ParametersItem_CTCH_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t PRACH_LCR_ParametersList_CTCH_SetupRqstTDD_sequence_of[1] = {
  { &hf_nbap_PRACH_LCR_ParametersList_CTCH_SetupRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_PRACH_LCR_ParametersItem_CTCH_SetupRqstTDD },
};

static int
dissect_nbap_PRACH_LCR_ParametersList_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_PRACH_LCR_ParametersList_CTCH_SetupRqstTDD, PRACH_LCR_ParametersList_CTCH_SetupRqstTDD_sequence_of,
                                                  1, maxNrOfPRACHLCRs, FALSE);

  return offset;
}


static const per_sequence_t PRACH_768_ParametersItem_CTCH_SetupRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID768 },
  { &hf_nbap_tFCS           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCS },
  { &hf_nbap_timeslot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_tdd_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode768 },
  { &hf_nbap_maxPRACH_MidambleShifts, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MaxPRACH_MidambleShifts },
  { &hf_nbap_pRACH_Midamble , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PRACH_Midamble },
  { &hf_nbap_rACH           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RACH_Parameter_CTCH_SetupRqstTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PRACH_768_ParametersItem_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PRACH_768_ParametersItem_CTCH_SetupRqstTDD, PRACH_768_ParametersItem_CTCH_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t FPACH_LCR_Parameters_CTCH_SetupRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_tdd_ChannelisationCodeLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCodeLCR },
  { &hf_nbap_timeslotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_fPACH_Power    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FPACH_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_FPACH_LCR_Parameters_CTCH_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_FPACH_LCR_Parameters_CTCH_SetupRqstTDD, FPACH_LCR_Parameters_CTCH_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t PLCCH_parameters_sequence[] = {
  { &hf_nbap_maxPowerPLCCH  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_timeslotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PLCCH_parameters(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PLCCH_parameters, PLCCH_parameters_sequence);

  return offset;
}


static const per_sequence_t E_RUCCH_parameters_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_timeslot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_maxE_RUCCH_MidambleShifts, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MaxPRACH_MidambleShifts },
  { &hf_nbap_e_RUCCH_Midamble, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PRACH_Midamble },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_RUCCH_parameters(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_RUCCH_parameters, E_RUCCH_parameters_sequence);

  return offset;
}


static const per_sequence_t E_RUCCH_768_parameters_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID768 },
  { &hf_nbap_timeslot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_tdd_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode768 },
  { &hf_nbap_maxE_RUCCH_MidambleShifts, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MaxPRACH_MidambleShifts },
  { &hf_nbap_e_RUCCH_Midamble, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PRACH_Midamble },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_RUCCH_768_parameters(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_RUCCH_768_parameters, E_RUCCH_768_parameters_sequence);

  return offset;
}


static const per_sequence_t CommonTransportChannelSetupResponse_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommonTransportChannelSetupResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"CommonTransportChannelSetupResponse ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommonTransportChannelSetupResponse, CommonTransportChannelSetupResponse_sequence);

  return offset;
}


static const per_sequence_t FACH_CommonTransportChannel_InformationResponse_sequence_of[1] = {
  { &hf_nbap_FACH_CommonTransportChannel_InformationResponse_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_CommonTransportChannel_InformationResponse },
};

static int
dissect_nbap_FACH_CommonTransportChannel_InformationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_FACH_CommonTransportChannel_InformationResponse, FACH_CommonTransportChannel_InformationResponse_sequence_of,
                                                  1, maxNrOfFACHs, FALSE);

  return offset;
}


static const per_sequence_t CommonTransportChannelSetupFailure_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommonTransportChannelSetupFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"CommonTransportChannelSetupFailure ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommonTransportChannelSetupFailure, CommonTransportChannelSetupFailure_sequence);

  return offset;
}


static const per_sequence_t CommonTransportChannelReconfigurationRequestFDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommonTransportChannelReconfigurationRequestFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommonTransportChannelReconfigurationRequestFDD, CommonTransportChannelReconfigurationRequestFDD_sequence);

  return offset;
}



static int
dissect_nbap_FACH_ParametersList_CTCH_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_PCH_Parameters_CTCH_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_PICH_Parameters_CTCH_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t Secondary_CCPCHList_CTCH_ReconfRqstFDD_sequence[] = {
  { &hf_nbap_fACH_ParametersList_CTCH_ReconfRqstFDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_FACH_ParametersList_CTCH_ReconfRqstFDD },
  { &hf_nbap_pCH_Parameters_CTCH_ReconfRqstFDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PCH_Parameters_CTCH_ReconfRqstFDD },
  { &hf_nbap_pICH_Parameters_CTCH_ReconfRqstFDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PICH_Parameters_CTCH_ReconfRqstFDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Secondary_CCPCHList_CTCH_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Secondary_CCPCHList_CTCH_ReconfRqstFDD, Secondary_CCPCHList_CTCH_ReconfRqstFDD_sequence);

  return offset;
}



static int
dissect_nbap_PRACH_ParametersList_CTCH_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_AICH_ParametersList_CTCH_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t PRACHList_CTCH_ReconfRqstFDD_sequence[] = {
  { &hf_nbap_pRACH_ParametersList_CTCH_ReconfRqstFDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PRACH_ParametersList_CTCH_ReconfRqstFDD },
  { &hf_nbap_aICH_ParametersList_CTCH_ReconfRqstFDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_AICH_ParametersList_CTCH_ReconfRqstFDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PRACHList_CTCH_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PRACHList_CTCH_ReconfRqstFDD, PRACHList_CTCH_ReconfRqstFDD_sequence);

  return offset;
}


static const value_string nbap_CommonPhysicalChannelType_CTCH_ReconfRqstFDD_vals[] = {
  {   0, "secondary-CCPCH-parameters" },
  {   1, "pRACH-parameters" },
  {   2, "notUsed-cPCH-parameters" },
  { 0, NULL }
};

static const per_choice_t CommonPhysicalChannelType_CTCH_ReconfRqstFDD_choice[] = {
  {   0, &hf_nbap_secondary_CCPCH_parameters_02, ASN1_EXTENSION_ROOT    , dissect_nbap_Secondary_CCPCHList_CTCH_ReconfRqstFDD },
  {   1, &hf_nbap_pRACH_parameters_02, ASN1_EXTENSION_ROOT    , dissect_nbap_PRACHList_CTCH_ReconfRqstFDD },
  {   2, &hf_nbap_notUsed_cPCH_parameters, ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_CommonPhysicalChannelType_CTCH_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_CommonPhysicalChannelType_CTCH_ReconfRqstFDD, CommonPhysicalChannelType_CTCH_ReconfRqstFDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t FACH_ParametersItem_CTCH_ReconfRqstFDD_sequence[] = {
  { &hf_nbap_commonTransportChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonTransportChannelID },
  { &hf_nbap_maxFACH_Power  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_toAWS          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ToAWS },
  { &hf_nbap_toAWE          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ToAWE },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_FACH_ParametersItem_CTCH_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_FACH_ParametersItem_CTCH_ReconfRqstFDD, FACH_ParametersItem_CTCH_ReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t FACH_ParametersListIE_CTCH_ReconfRqstFDD_sequence_of[1] = {
  { &hf_nbap_FACH_ParametersListIE_CTCH_ReconfRqstFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_FACH_ParametersItem_CTCH_ReconfRqstFDD },
};

static int
dissect_nbap_FACH_ParametersListIE_CTCH_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_FACH_ParametersListIE_CTCH_ReconfRqstFDD, FACH_ParametersListIE_CTCH_ReconfRqstFDD_sequence_of,
                                                  1, maxFACHCell, FALSE);

  return offset;
}


static const per_sequence_t PCH_ParametersItem_CTCH_ReconfRqstFDD_sequence[] = {
  { &hf_nbap_commonTransportChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonTransportChannelID },
  { &hf_nbap_pCH_Power      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_toAWS          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ToAWS },
  { &hf_nbap_toAWE          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ToAWE },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PCH_ParametersItem_CTCH_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PCH_ParametersItem_CTCH_ReconfRqstFDD, PCH_ParametersItem_CTCH_ReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t PICH_ParametersItem_CTCH_ReconfRqstFDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_pICH_Power     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PICH_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PICH_ParametersItem_CTCH_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PICH_ParametersItem_CTCH_ReconfRqstFDD, PICH_ParametersItem_CTCH_ReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t MICH_Parameters_CTCH_ReconfRqstFDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_mICH_Power     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PICH_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MICH_Parameters_CTCH_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MICH_Parameters_CTCH_ReconfRqstFDD, MICH_Parameters_CTCH_ReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t AllowedSlotFormatInformationItem_CTCH_ReconfRqstFDD_sequence[] = {
  { &hf_nbap_rACH_SlotFormat, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RACH_SlotFormat },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_AllowedSlotFormatInformationItem_CTCH_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_AllowedSlotFormatInformationItem_CTCH_ReconfRqstFDD, AllowedSlotFormatInformationItem_CTCH_ReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t AllowedSlotFormatInformationList_CTCH_ReconfRqstFDD_sequence_of[1] = {
  { &hf_nbap_AllowedSlotFormatInformationList_CTCH_ReconfRqstFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_AllowedSlotFormatInformationItem_CTCH_ReconfRqstFDD },
};

static int
dissect_nbap_AllowedSlotFormatInformationList_CTCH_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_AllowedSlotFormatInformationList_CTCH_ReconfRqstFDD, AllowedSlotFormatInformationList_CTCH_ReconfRqstFDD_sequence_of,
                                                  1, maxNrOfSlotFormatsPRACH, FALSE);

  return offset;
}


static const per_sequence_t PRACH_ParametersItem_CTCH_ReconfRqstFDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_preambleSignatures, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PreambleSignatures },
  { &hf_nbap_allowedSlotFormatInformation_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_AllowedSlotFormatInformationList_CTCH_ReconfRqstFDD },
  { &hf_nbap_rACH_SubChannelNumbers, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RACH_SubChannelNumbers },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PRACH_ParametersItem_CTCH_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PRACH_ParametersItem_CTCH_ReconfRqstFDD, PRACH_ParametersItem_CTCH_ReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t PRACH_ParametersListIE_CTCH_ReconfRqstFDD_sequence_of[1] = {
  { &hf_nbap_PRACH_ParametersListIE_CTCH_ReconfRqstFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_PRACH_ParametersItem_CTCH_ReconfRqstFDD },
};

static int
dissect_nbap_PRACH_ParametersListIE_CTCH_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_PRACH_ParametersListIE_CTCH_ReconfRqstFDD, PRACH_ParametersListIE_CTCH_ReconfRqstFDD_sequence_of,
                                                  1, maxPRACHCell, FALSE);

  return offset;
}


static const per_sequence_t AICH_ParametersItem_CTCH_ReconfRqstFDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_aICH_Power     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_AICH_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_AICH_ParametersItem_CTCH_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_AICH_ParametersItem_CTCH_ReconfRqstFDD, AICH_ParametersItem_CTCH_ReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t AICH_ParametersListIE_CTCH_ReconfRqstFDD_sequence_of[1] = {
  { &hf_nbap_AICH_ParametersListIE_CTCH_ReconfRqstFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_AICH_ParametersItem_CTCH_ReconfRqstFDD },
};

static int
dissect_nbap_AICH_ParametersListIE_CTCH_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_AICH_ParametersListIE_CTCH_ReconfRqstFDD, AICH_ParametersListIE_CTCH_ReconfRqstFDD_sequence_of,
                                                  1, maxPRACHCell, FALSE);

  return offset;
}


static const per_sequence_t CommonTransportChannelReconfigurationRequestTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommonTransportChannelReconfigurationRequestTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommonTransportChannelReconfigurationRequestTDD, CommonTransportChannelReconfigurationRequestTDD_sequence);

  return offset;
}



static int
dissect_nbap_Secondary_CCPCHList_CTCH_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_Parameters_CTCH_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_secondaryCCPCHList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Secondary_CCPCHList_CTCH_ReconfRqstTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Secondary_CCPCH_Parameters_CTCH_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Secondary_CCPCH_Parameters_CTCH_ReconfRqstTDD, Secondary_CCPCH_Parameters_CTCH_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t Secondary_CCPCHItem_CTCH_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_sCCPCH_Power   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Secondary_CCPCHItem_CTCH_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Secondary_CCPCHItem_CTCH_ReconfRqstTDD, Secondary_CCPCHItem_CTCH_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t Secondary_CCPCHListIE_CTCH_ReconfRqstTDD_sequence_of[1] = {
  { &hf_nbap_Secondary_CCPCHListIE_CTCH_ReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Secondary_CCPCHItem_CTCH_ReconfRqstTDD },
};

static int
dissect_nbap_Secondary_CCPCHListIE_CTCH_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Secondary_CCPCHListIE_CTCH_ReconfRqstTDD, Secondary_CCPCHListIE_CTCH_ReconfRqstTDD_sequence_of,
                                                  1, maxNrOfSCCPCHs, FALSE);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_parameterExtendedList_CTCH_ReconfRqstTDD_sequence_of[1] = {
  { &hf_nbap_Secondary_CCPCH_parameterExtendedList_CTCH_ReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Secondary_CCPCHItem_CTCH_ReconfRqstTDD },
};

static int
dissect_nbap_Secondary_CCPCH_parameterExtendedList_CTCH_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Secondary_CCPCH_parameterExtendedList_CTCH_ReconfRqstTDD, Secondary_CCPCH_parameterExtendedList_CTCH_ReconfRqstTDD_sequence_of,
                                                  1, maxNrOfSCCPCHsinExt, FALSE);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_LCR_parameterExtendedList_CTCH_ReconfRqstTDD_sequence_of[1] = {
  { &hf_nbap_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_ReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Secondary_CCPCHItem_CTCH_ReconfRqstTDD },
};

static int
dissect_nbap_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_ReconfRqstTDD, Secondary_CCPCH_LCR_parameterExtendedList_CTCH_ReconfRqstTDD_sequence_of,
                                                  1, maxNrOfSCCPCHsLCRinExt, FALSE);

  return offset;
}


static const per_sequence_t PICH_Parameters_CTCH_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_pICH_Power     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PICH_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PICH_Parameters_CTCH_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PICH_Parameters_CTCH_ReconfRqstTDD, PICH_Parameters_CTCH_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t FACH_ParametersItem_CTCH_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_commonTransportChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonTransportChannelID },
  { &hf_nbap_toAWS          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ToAWS },
  { &hf_nbap_toAWE          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ToAWE },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_FACH_ParametersItem_CTCH_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_FACH_ParametersItem_CTCH_ReconfRqstTDD, FACH_ParametersItem_CTCH_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t FACH_ParametersList_CTCH_ReconfRqstTDD_sequence_of[1] = {
  { &hf_nbap_FACH_ParametersList_CTCH_ReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_FACH_ParametersItem_CTCH_ReconfRqstTDD },
};

static int
dissect_nbap_FACH_ParametersList_CTCH_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_FACH_ParametersList_CTCH_ReconfRqstTDD, FACH_ParametersList_CTCH_ReconfRqstTDD_sequence_of,
                                                  0, maxNrOfFACHs, FALSE);

  return offset;
}


static const per_sequence_t PCH_Parameters_CTCH_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_commonTransportChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonTransportChannelID },
  { &hf_nbap_toAWS          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ToAWS },
  { &hf_nbap_toAWE          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ToAWE },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PCH_Parameters_CTCH_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PCH_Parameters_CTCH_ReconfRqstTDD, PCH_Parameters_CTCH_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t FPACH_LCR_Parameters_CTCH_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelId, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_fPACHPower     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_FPACH_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_FPACH_LCR_Parameters_CTCH_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_FPACH_LCR_Parameters_CTCH_ReconfRqstTDD, FPACH_LCR_Parameters_CTCH_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t MICH_Parameters_CTCH_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_mICH_Power     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PICH_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MICH_Parameters_CTCH_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MICH_Parameters_CTCH_ReconfRqstTDD, MICH_Parameters_CTCH_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t PLCCH_Parameters_CTCH_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_maxPowerPLCCH  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PLCCH_Parameters_CTCH_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PLCCH_Parameters_CTCH_ReconfRqstTDD, PLCCH_Parameters_CTCH_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_768_Item_CTCH_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID768 },
  { &hf_nbap_sCCPCH_Power   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Secondary_CCPCH_768_Item_CTCH_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Secondary_CCPCH_768_Item_CTCH_ReconfRqstTDD, Secondary_CCPCH_768_Item_CTCH_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_768_List_CTCH_ReconfRqstTDD_sequence_of[1] = {
  { &hf_nbap_Secondary_CCPCH_768_List_CTCH_ReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Secondary_CCPCH_768_Item_CTCH_ReconfRqstTDD },
};

static int
dissect_nbap_Secondary_CCPCH_768_List_CTCH_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Secondary_CCPCH_768_List_CTCH_ReconfRqstTDD, Secondary_CCPCH_768_List_CTCH_ReconfRqstTDD_sequence_of,
                                                  1, maxNrOfSCCPCHs768, FALSE);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_768_Parameters_CTCH_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_secondaryCCPCH768List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Secondary_CCPCH_768_List_CTCH_ReconfRqstTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Secondary_CCPCH_768_Parameters_CTCH_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Secondary_CCPCH_768_Parameters_CTCH_ReconfRqstTDD, Secondary_CCPCH_768_Parameters_CTCH_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t PICH_768_Parameters_CTCH_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID768 },
  { &hf_nbap_pICH_Power     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PICH_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PICH_768_Parameters_CTCH_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PICH_768_Parameters_CTCH_ReconfRqstTDD, PICH_768_Parameters_CTCH_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t MICH_768_Parameters_CTCH_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID768 },
  { &hf_nbap_mICH_Power     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PICH_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MICH_768_Parameters_CTCH_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MICH_768_Parameters_CTCH_ReconfRqstTDD, MICH_768_Parameters_CTCH_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t UPPCH_LCR_Parameters_CTCH_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_uPPCHPositionLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UPPCHPositionLCR },
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UARFCN },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UPPCH_LCR_Parameters_CTCH_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UPPCH_LCR_Parameters_CTCH_ReconfRqstTDD, UPPCH_LCR_Parameters_CTCH_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t CommonTransportChannelReconfigurationResponse_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommonTransportChannelReconfigurationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommonTransportChannelReconfigurationResponse, CommonTransportChannelReconfigurationResponse_sequence);

  return offset;
}


static const per_sequence_t CommonTransportChannelReconfigurationFailure_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommonTransportChannelReconfigurationFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommonTransportChannelReconfigurationFailure, CommonTransportChannelReconfigurationFailure_sequence);

  return offset;
}


static const per_sequence_t CommonTransportChannelDeletionRequest_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommonTransportChannelDeletionRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommonTransportChannelDeletionRequest, CommonTransportChannelDeletionRequest_sequence);

  return offset;
}


static const per_sequence_t CommonTransportChannelDeletionResponse_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommonTransportChannelDeletionResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommonTransportChannelDeletionResponse, CommonTransportChannelDeletionResponse_sequence);

  return offset;
}


static const per_sequence_t BlockResourceRequest_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_BlockResourceRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_BlockResourceRequest, BlockResourceRequest_sequence);

  return offset;
}


static const per_sequence_t BlockResourceResponse_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_BlockResourceResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_BlockResourceResponse, BlockResourceResponse_sequence);

  return offset;
}


static const per_sequence_t BlockResourceFailure_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_BlockResourceFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_BlockResourceFailure, BlockResourceFailure_sequence);

  return offset;
}


static const per_sequence_t UnblockResourceIndication_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UnblockResourceIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"UnblockResourceIndication ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UnblockResourceIndication, UnblockResourceIndication_sequence);

  return offset;
}


static const per_sequence_t AuditRequiredIndication_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_AuditRequiredIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"AuditRequiredIndication ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_AuditRequiredIndication, AuditRequiredIndication_sequence);

  return offset;
}


static const per_sequence_t AuditRequest_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_AuditRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"AuditRequest ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_AuditRequest, AuditRequest_sequence);

  return offset;
}


static const per_sequence_t AuditResponse_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_AuditResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"AuditResponse ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_AuditResponse, AuditResponse_sequence);

  return offset;
}


static const per_sequence_t Cell_InformationList_AuditRsp_sequence_of[1] = {
  { &hf_nbap_Cell_InformationList_AuditRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_Cell_InformationList_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Cell_InformationList_AuditRsp, Cell_InformationList_AuditRsp_sequence_of,
                                                  1, maxCellinNodeB, FALSE);

  return offset;
}



static int
dissect_nbap_P_SCH_Information_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_S_SCH_Information_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_P_CPICH_Information_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t S_CPICH_InformationList_AuditRsp_sequence_of[1] = {
  { &hf_nbap_S_CPICH_InformationList_AuditRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_S_CPICH_InformationList_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_S_CPICH_InformationList_AuditRsp, S_CPICH_InformationList_AuditRsp_sequence_of,
                                                  1, maxSCPICHCell, FALSE);

  return offset;
}



static int
dissect_nbap_P_CCPCH_Information_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_BCH_Information_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t S_CCPCH_InformationList_AuditRsp_sequence_of[1] = {
  { &hf_nbap_S_CCPCH_InformationList_AuditRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_S_CCPCH_InformationList_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_S_CCPCH_InformationList_AuditRsp, S_CCPCH_InformationList_AuditRsp_sequence_of,
                                                  1, maxSCCPCHCell, FALSE);

  return offset;
}



static int
dissect_nbap_PCH_Information_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_PICH_Information_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t FACH_InformationList_AuditRsp_sequence_of[1] = {
  { &hf_nbap_FACH_InformationList_AuditRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_FACH_InformationList_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_FACH_InformationList_AuditRsp, FACH_InformationList_AuditRsp_sequence_of,
                                                  1, maxFACHCell, FALSE);

  return offset;
}


static const per_sequence_t PRACH_InformationList_AuditRsp_sequence_of[1] = {
  { &hf_nbap_PRACH_InformationList_AuditRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_PRACH_InformationList_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_PRACH_InformationList_AuditRsp, PRACH_InformationList_AuditRsp_sequence_of,
                                                  1, maxPRACHCell, FALSE);

  return offset;
}


static const per_sequence_t RACH_InformationList_AuditRsp_sequence_of[1] = {
  { &hf_nbap_RACH_InformationList_AuditRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_RACH_InformationList_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RACH_InformationList_AuditRsp, RACH_InformationList_AuditRsp_sequence_of,
                                                  1, maxRACHCell, FALSE);

  return offset;
}


static const per_sequence_t AICH_InformationList_AuditRsp_sequence_of[1] = {
  { &hf_nbap_AICH_InformationList_AuditRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_AICH_InformationList_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_AICH_InformationList_AuditRsp, AICH_InformationList_AuditRsp_sequence_of,
                                                  1, maxPRACHCell, FALSE);

  return offset;
}



static int
dissect_nbap_SCH_Information_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t Cell_InformationItem_AuditRsp_sequence[] = {
  { &hf_nbap_c_ID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_C_ID },
  { &hf_nbap_configurationGenerationID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ConfigurationGenerationID },
  { &hf_nbap_resourceOperationalState, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ResourceOperationalState },
  { &hf_nbap_availabilityStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AvailabilityStatus },
  { &hf_nbap_local_Cell_ID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Local_Cell_ID },
  { &hf_nbap_primary_SCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_P_SCH_Information_AuditRsp },
  { &hf_nbap_secondary_SCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_S_SCH_Information_AuditRsp },
  { &hf_nbap_primary_CPICH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_P_CPICH_Information_AuditRsp },
  { &hf_nbap_secondary_CPICH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_S_CPICH_InformationList_AuditRsp },
  { &hf_nbap_primary_CCPCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_P_CCPCH_Information_AuditRsp },
  { &hf_nbap_bCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BCH_Information_AuditRsp },
  { &hf_nbap_secondary_CCPCH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_S_CCPCH_InformationList_AuditRsp },
  { &hf_nbap_pCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PCH_Information_AuditRsp },
  { &hf_nbap_pICH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PICH_Information_AuditRsp },
  { &hf_nbap_fACH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_FACH_InformationList_AuditRsp },
  { &hf_nbap_pRACH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PRACH_InformationList_AuditRsp },
  { &hf_nbap_rACH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RACH_InformationList_AuditRsp },
  { &hf_nbap_aICH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_AICH_InformationList_AuditRsp },
  { &hf_nbap_notUsed_1_pCPCH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_notUsed_2_cPCH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_notUsed_3_aP_AICH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_notUsed_4_cDCA_ICH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_sCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SCH_Information_AuditRsp },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Cell_InformationItem_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Cell_InformationItem_AuditRsp, Cell_InformationItem_AuditRsp_sequence);

  return offset;
}


static const per_sequence_t FPACH_LCR_InformationList_AuditRsp_sequence_of[1] = {
  { &hf_nbap_FPACH_LCR_InformationList_AuditRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_FPACH_LCR_InformationList_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_FPACH_LCR_InformationList_AuditRsp, FPACH_LCR_InformationList_AuditRsp_sequence_of,
                                                  1, maxFPACHCell, FALSE);

  return offset;
}


static const per_sequence_t HS_DSCH_Resources_Information_AuditRsp_sequence[] = {
  { &hf_nbap_resourceOperationalState, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ResourceOperationalState },
  { &hf_nbap_availabilityStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AvailabilityStatus },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_DSCH_Resources_Information_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_DSCH_Resources_Information_AuditRsp, HS_DSCH_Resources_Information_AuditRsp_sequence);

  return offset;
}


static const per_sequence_t S_CCPCH_InformationListExt_AuditRsp_sequence_of[1] = {
  { &hf_nbap_S_CCPCH_InformationListExt_AuditRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_S_CCPCH_InformationListExt_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_S_CCPCH_InformationListExt_AuditRsp, S_CCPCH_InformationListExt_AuditRsp_sequence_of,
                                                  1, maxSCCPCHCellinExt, FALSE);

  return offset;
}


static const per_sequence_t S_CCPCH_LCR_InformationListExt_AuditRsp_sequence_of[1] = {
  { &hf_nbap_S_CCPCH_LCR_InformationListExt_AuditRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_S_CCPCH_LCR_InformationListExt_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_S_CCPCH_LCR_InformationListExt_AuditRsp, S_CCPCH_LCR_InformationListExt_AuditRsp_sequence_of,
                                                  1, maxSCCPCHCellinExtLCR, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_Resources_Information_AuditRsp_sequence[] = {
  { &hf_nbap_resourceOperationalState, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ResourceOperationalState },
  { &hf_nbap_availabilityStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AvailabilityStatus },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_Resources_Information_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_Resources_Information_AuditRsp, E_DCH_Resources_Information_AuditRsp_sequence);

  return offset;
}


static const per_sequence_t PLCCH_InformationList_AuditRsp_sequence_of[1] = {
  { &hf_nbap_PLCCH_InformationList_AuditRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_PLCCH_InformationList_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_PLCCH_InformationList_AuditRsp, PLCCH_InformationList_AuditRsp_sequence_of,
                                                  1, maxPLCCHCell, FALSE);

  return offset;
}


static const per_sequence_t S_CCPCH_768_InformationList_AuditRsp_sequence_of[1] = {
  { &hf_nbap_S_CCPCH_768_InformationList_AuditRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_S_CCPCH_768_InformationList_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_S_CCPCH_768_InformationList_AuditRsp, S_CCPCH_768_InformationList_AuditRsp_sequence_of,
                                                  1, maxSCCPCHCell768, FALSE);

  return offset;
}


static const per_sequence_t PRACH_768_InformationList_AuditRsp_sequence_of[1] = {
  { &hf_nbap_PRACH_768_InformationList_AuditRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_PRACH_768_InformationList_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_PRACH_768_InformationList_AuditRsp, PRACH_768_InformationList_AuditRsp_sequence_of,
                                                  1, maxPRACHCell, FALSE);

  return offset;
}


static const per_sequence_t E_RUCCH_InformationList_AuditRsp_sequence_of[1] = {
  { &hf_nbap_E_RUCCH_InformationList_AuditRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_E_RUCCH_InformationList_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_RUCCH_InformationList_AuditRsp, E_RUCCH_InformationList_AuditRsp_sequence_of,
                                                  1, maxE_RUCCHCell, FALSE);

  return offset;
}


static const per_sequence_t E_RUCCH_768_InformationList_AuditRsp_sequence_of[1] = {
  { &hf_nbap_E_RUCCH_768_InformationList_AuditRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_E_RUCCH_768_InformationList_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_RUCCH_768_InformationList_AuditRsp, E_RUCCH_768_InformationList_AuditRsp_sequence_of,
                                                  1, maxE_RUCCHCell, FALSE);

  return offset;
}


static const per_sequence_t Cell_Frequency_List_Information_LCR_MulFreq_AuditRsp_sequence_of[1] = {
  { &hf_nbap_Cell_Frequency_List_Information_LCR_MulFreq_AuditRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_Cell_Frequency_List_Information_LCR_MulFreq_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Cell_Frequency_List_Information_LCR_MulFreq_AuditRsp, Cell_Frequency_List_Information_LCR_MulFreq_AuditRsp_sequence_of,
                                                  1, maxFrequencyinCell, FALSE);

  return offset;
}


static const per_sequence_t Cell_Frequency_List_InformationItem_LCR_MulFreq_AuditRsp_sequence[] = {
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UARFCN },
  { &hf_nbap_resourceOperationalState, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ResourceOperationalState },
  { &hf_nbap_availabilityStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AvailabilityStatus },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Cell_Frequency_List_InformationItem_LCR_MulFreq_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Cell_Frequency_List_InformationItem_LCR_MulFreq_AuditRsp, Cell_Frequency_List_InformationItem_LCR_MulFreq_AuditRsp_sequence);

  return offset;
}


static const per_sequence_t UPPCH_LCR_InformationList_AuditRsp_sequence_of[1] = {
  { &hf_nbap_UPPCH_LCR_InformationList_AuditRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_UPPCH_LCR_InformationList_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UPPCH_LCR_InformationList_AuditRsp, UPPCH_LCR_InformationList_AuditRsp_sequence_of,
                                                  1, maxFrequencyinCell, FALSE);

  return offset;
}


static const per_sequence_t UPPCH_LCR_InformationItem_AuditRsp_sequence[] = {
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UARFCN },
  { &hf_nbap_uPPCHPositionLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UPPCHPositionLCR },
  { &hf_nbap_resourceOperationalState, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ResourceOperationalState },
  { &hf_nbap_availabilityStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AvailabilityStatus },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UPPCH_LCR_InformationItem_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UPPCH_LCR_InformationItem_AuditRsp, UPPCH_LCR_InformationItem_AuditRsp_sequence);

  return offset;
}


static const per_sequence_t MultipleFreq_HS_DSCH_Resources_InformationList_AuditRsp_sequence_of[1] = {
  { &hf_nbap_MultipleFreq_HS_DSCH_Resources_InformationList_AuditRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_MultipleFreq_HS_DSCH_Resources_InformationList_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_MultipleFreq_HS_DSCH_Resources_InformationList_AuditRsp, MultipleFreq_HS_DSCH_Resources_InformationList_AuditRsp_sequence_of,
                                                  1, maxFrequencyinCell_1, FALSE);

  return offset;
}


static const per_sequence_t MultipleFreq_E_DCH_Resources_InformationList_AuditRsp_sequence_of[1] = {
  { &hf_nbap_MultipleFreq_E_DCH_Resources_InformationList_AuditRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_MultipleFreq_E_DCH_Resources_InformationList_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_MultipleFreq_E_DCH_Resources_InformationList_AuditRsp, MultipleFreq_E_DCH_Resources_InformationList_AuditRsp_sequence_of,
                                                  1, maxFrequencyinCell_1, FALSE);

  return offset;
}


static const per_sequence_t CCP_InformationList_AuditRsp_sequence_of[1] = {
  { &hf_nbap_CCP_InformationList_AuditRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_CCP_InformationList_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_CCP_InformationList_AuditRsp, CCP_InformationList_AuditRsp_sequence_of,
                                                  1, maxCCPinNodeB, FALSE);

  return offset;
}


static const per_sequence_t CCP_InformationItem_AuditRsp_sequence[] = {
  { &hf_nbap_communicationControlPortID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommunicationControlPortID },
  { &hf_nbap_resourceOperationalState, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ResourceOperationalState },
  { &hf_nbap_availabilityStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AvailabilityStatus },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CCP_InformationItem_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CCP_InformationItem_AuditRsp, CCP_InformationItem_AuditRsp_sequence);

  return offset;
}


static const per_sequence_t Local_Cell_InformationList_AuditRsp_sequence_of[1] = {
  { &hf_nbap_Local_Cell_InformationList_AuditRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_Local_Cell_InformationList_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Local_Cell_InformationList_AuditRsp, Local_Cell_InformationList_AuditRsp_sequence_of,
                                                  1, maxLocalCellinNodeB, FALSE);

  return offset;
}


static const per_sequence_t Local_Cell_InformationItem_AuditRsp_sequence[] = {
  { &hf_nbap_local_Cell_ID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Local_Cell_ID },
  { &hf_nbap_dl_or_global_capacityCredit, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_or_Global_CapacityCredit },
  { &hf_nbap_ul_capacityCredit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_CapacityCredit },
  { &hf_nbap_commonChannelsCapacityConsumptionLaw, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonChannelsCapacityConsumptionLaw },
  { &hf_nbap_dedicatedChannelsCapacityConsumptionLaw, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DedicatedChannelsCapacityConsumptionLaw },
  { &hf_nbap_maximumDL_PowerCapability, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MaximumDL_PowerCapability },
  { &hf_nbap_minSpreadingFactor, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MinSpreadingFactor },
  { &hf_nbap_minimumDL_PowerCapability, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MinimumDL_PowerCapability },
  { &hf_nbap_local_Cell_Group_ID, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Local_Cell_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Local_Cell_InformationItem_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Local_Cell_InformationItem_AuditRsp, Local_Cell_InformationItem_AuditRsp_sequence);

  return offset;
}


static const per_sequence_t Local_Cell_Group_InformationList_AuditRsp_sequence_of[1] = {
  { &hf_nbap_Local_Cell_Group_InformationList_AuditRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_Local_Cell_Group_InformationList_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Local_Cell_Group_InformationList_AuditRsp, Local_Cell_Group_InformationList_AuditRsp_sequence_of,
                                                  1, maxLocalCellinNodeB, FALSE);

  return offset;
}


static const per_sequence_t Local_Cell_Group_InformationItem_AuditRsp_sequence[] = {
  { &hf_nbap_local_Cell_Group_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Local_Cell_ID },
  { &hf_nbap_dl_or_global_capacityCredit, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_or_Global_CapacityCredit },
  { &hf_nbap_ul_capacityCredit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_CapacityCredit },
  { &hf_nbap_commonChannelsCapacityConsumptionLaw, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonChannelsCapacityConsumptionLaw },
  { &hf_nbap_dedicatedChannelsCapacityConsumptionLaw, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DedicatedChannelsCapacityConsumptionLaw },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Local_Cell_Group_InformationItem_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Local_Cell_Group_InformationItem_AuditRsp, Local_Cell_Group_InformationItem_AuditRsp_sequence);

  return offset;
}


static const per_sequence_t Power_Local_Cell_Group_InformationList_AuditRsp_sequence_of[1] = {
  { &hf_nbap_Power_Local_Cell_Group_InformationList_AuditRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_Power_Local_Cell_Group_InformationList_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Power_Local_Cell_Group_InformationList_AuditRsp, Power_Local_Cell_Group_InformationList_AuditRsp_sequence_of,
                                                  1, maxLocalCellinNodeB, FALSE);

  return offset;
}


static const per_sequence_t Power_Local_Cell_Group_InformationItem_AuditRsp_sequence[] = {
  { &hf_nbap_power_Local_Cell_Group_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Local_Cell_ID },
  { &hf_nbap_maximumDL_PowerCapability, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MaximumDL_PowerCapability },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Power_Local_Cell_Group_InformationItem_AuditRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Power_Local_Cell_Group_InformationItem_AuditRsp, Power_Local_Cell_Group_InformationItem_AuditRsp_sequence);

  return offset;
}


static const per_sequence_t AuditFailure_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_AuditFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"AuditFailure ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;


  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_AuditFailure, AuditFailure_sequence);

  return offset;
}


static const per_sequence_t CommonMeasurementInitiationRequest_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommonMeasurementInitiationRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"CommonMeasurementInitiationRequest ");
/* CRNC -> Node B */
actx->pinfo->link_dir=P2P_DIR_DL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommonMeasurementInitiationRequest, CommonMeasurementInitiationRequest_sequence);

  return offset;
}


static const per_sequence_t Cell_CM_Rqst_sequence[] = {
  { &hf_nbap_c_ID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_C_ID },
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TimeSlot },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Cell_CM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Cell_CM_Rqst, Cell_CM_Rqst_sequence);

  return offset;
}


static const per_sequence_t RACH_CM_Rqst_sequence[] = {
  { &hf_nbap_c_ID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_C_ID },
  { &hf_nbap_commonTransportChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonTransportChannelID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RACH_CM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RACH_CM_Rqst, RACH_CM_Rqst_sequence);

  return offset;
}



static int
dissect_nbap_Extension_CommonMeasurementObjectType_CM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string nbap_CommonMeasurementObjectType_CM_Rqst_vals[] = {
  {   0, "cell" },
  {   1, "rACH" },
  {   2, "notUsed-cPCH" },
  {   3, "extension-CommonMeasurementObjectType-CM-Rqst" },
  { 0, NULL }
};

static const per_choice_t CommonMeasurementObjectType_CM_Rqst_choice[] = {
  {   0, &hf_nbap_cell           , ASN1_EXTENSION_ROOT    , dissect_nbap_Cell_CM_Rqst },
  {   1, &hf_nbap_rACH_01        , ASN1_EXTENSION_ROOT    , dissect_nbap_RACH_CM_Rqst },
  {   2, &hf_nbap_notUsed_cPCH   , ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {   3, &hf_nbap_extension_CommonMeasurementObjectType_CM_Rqst, ASN1_NOT_EXTENSION_ROOT, dissect_nbap_Extension_CommonMeasurementObjectType_CM_Rqst },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_CommonMeasurementObjectType_CM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_CommonMeasurementObjectType_CM_Rqst, CommonMeasurementObjectType_CM_Rqst_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t ERACH_CM_Rqst_sequence[] = {
  { &hf_nbap_c_ID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_C_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ERACH_CM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ERACH_CM_Rqst, ERACH_CM_Rqst_sequence);

  return offset;
}


static const per_sequence_t PowerLocalCellGroup_CM_Rqst_sequence[] = {
  { &hf_nbap_powerLocalCellGroupID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Local_Cell_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PowerLocalCellGroup_CM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PowerLocalCellGroup_CM_Rqst, PowerLocalCellGroup_CM_Rqst_sequence);

  return offset;
}


static const per_sequence_t CommonMeasurementInitiationResponse_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommonMeasurementInitiationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"CommonMeasurementInitiationResponse ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommonMeasurementInitiationResponse, CommonMeasurementInitiationResponse_sequence);

  return offset;
}


static const per_sequence_t Cell_CM_Rsp_sequence[] = {
  { &hf_nbap_commonMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonMeasurementValue },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Cell_CM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Cell_CM_Rsp, Cell_CM_Rsp_sequence);

  return offset;
}


static const per_sequence_t RACH_CM_Rsp_sequence[] = {
  { &hf_nbap_commonMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonMeasurementValue },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RACH_CM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RACH_CM_Rsp, RACH_CM_Rsp_sequence);

  return offset;
}



static int
dissect_nbap_Extension_CommonMeasurementObjectType_CM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string nbap_CommonMeasurementObjectType_CM_Rsp_vals[] = {
  {   0, "cell" },
  {   1, "rACH" },
  {   2, "notUsed-cPCH" },
  {   3, "extension-CommonMeasurementObjectType-CM-Rsp" },
  { 0, NULL }
};

static const per_choice_t CommonMeasurementObjectType_CM_Rsp_choice[] = {
  {   0, &hf_nbap_cell_01        , ASN1_EXTENSION_ROOT    , dissect_nbap_Cell_CM_Rsp },
  {   1, &hf_nbap_rACH_02        , ASN1_EXTENSION_ROOT    , dissect_nbap_RACH_CM_Rsp },
  {   2, &hf_nbap_notUsed_cPCH   , ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {   3, &hf_nbap_extension_CommonMeasurementObjectType_CM_Rsp, ASN1_NOT_EXTENSION_ROOT, dissect_nbap_Extension_CommonMeasurementObjectType_CM_Rsp },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_CommonMeasurementObjectType_CM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_CommonMeasurementObjectType_CM_Rsp, CommonMeasurementObjectType_CM_Rsp_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t ERACH_CM_Rsp_sequence[] = {
  { &hf_nbap_commonMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonMeasurementValue },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ERACH_CM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ERACH_CM_Rsp, ERACH_CM_Rsp_sequence);

  return offset;
}


static const per_sequence_t PowerLocalCellGroup_CM_Rsp_sequence[] = {
  { &hf_nbap_commonMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonMeasurementValue },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PowerLocalCellGroup_CM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PowerLocalCellGroup_CM_Rsp, PowerLocalCellGroup_CM_Rsp_sequence);

  return offset;
}


static const per_sequence_t CommonMeasurementInitiationFailure_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommonMeasurementInitiationFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"CommonMeasurementInitiationFailure ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommonMeasurementInitiationFailure, CommonMeasurementInitiationFailure_sequence);

  return offset;
}


static const per_sequence_t CommonMeasurementReport_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommonMeasurementReport(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"CommonMeasurementReport ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommonMeasurementReport, CommonMeasurementReport_sequence);

  return offset;
}


static const per_sequence_t Cell_CM_Rprt_sequence[] = {
  { &hf_nbap_commonMeasurementValueInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonMeasurementValueInformation },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Cell_CM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Cell_CM_Rprt, Cell_CM_Rprt_sequence);

  return offset;
}


static const per_sequence_t RACH_CM_Rprt_sequence[] = {
  { &hf_nbap_commonMeasurementValueInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonMeasurementValueInformation },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RACH_CM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RACH_CM_Rprt, RACH_CM_Rprt_sequence);

  return offset;
}



static int
dissect_nbap_Extension_CommonMeasurementObjectType_CM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string nbap_CommonMeasurementObjectType_CM_Rprt_vals[] = {
  {   0, "cell" },
  {   1, "rACH" },
  {   2, "notUsed-cPCH" },
  {   3, "extension-CommonMeasurementObjectType-CM-Rprt" },
  { 0, NULL }
};

static const per_choice_t CommonMeasurementObjectType_CM_Rprt_choice[] = {
  {   0, &hf_nbap_cell_02        , ASN1_EXTENSION_ROOT    , dissect_nbap_Cell_CM_Rprt },
  {   1, &hf_nbap_rACH_03        , ASN1_EXTENSION_ROOT    , dissect_nbap_RACH_CM_Rprt },
  {   2, &hf_nbap_notUsed_cPCH   , ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {   3, &hf_nbap_extension_CommonMeasurementObjectType_CM_Rprt, ASN1_NOT_EXTENSION_ROOT, dissect_nbap_Extension_CommonMeasurementObjectType_CM_Rprt },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_CommonMeasurementObjectType_CM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_CommonMeasurementObjectType_CM_Rprt, CommonMeasurementObjectType_CM_Rprt_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t ERACH_CM_Rprt_sequence[] = {
  { &hf_nbap_commonMeasurementValueInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonMeasurementValueInformation },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ERACH_CM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ERACH_CM_Rprt, ERACH_CM_Rprt_sequence);

  return offset;
}


static const per_sequence_t PowerLocalCellGroup_CM_Rprt_sequence[] = {
  { &hf_nbap_commonMeasurementValueInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonMeasurementValueInformation },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PowerLocalCellGroup_CM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PowerLocalCellGroup_CM_Rprt, PowerLocalCellGroup_CM_Rprt_sequence);

  return offset;
}


static const per_sequence_t CommonMeasurementTerminationRequest_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommonMeasurementTerminationRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"CommonMeasurementTerminationRequest ");
/* CRNC -> Node B */
actx->pinfo->link_dir=P2P_DIR_DL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommonMeasurementTerminationRequest, CommonMeasurementTerminationRequest_sequence);

  return offset;
}


static const per_sequence_t CommonMeasurementFailureIndication_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommonMeasurementFailureIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"CommonMeasurementFailureIndication ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommonMeasurementFailureIndication, CommonMeasurementFailureIndication_sequence);

  return offset;
}


static const per_sequence_t CellSetupRequestFDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSetupRequestFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"CellSetupRequest(FDD) ");
/* CRNC -> Node B */
actx->pinfo->link_dir=P2P_DIR_DL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSetupRequestFDD, CellSetupRequestFDD_sequence);

  return offset;
}


static const per_sequence_t Synchronisation_Configuration_Cell_SetupRqst_sequence[] = {
  { &hf_nbap_n_INSYNC_IND   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_N_INSYNC_IND },
  { &hf_nbap_n_OUTSYNC_IND  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_N_OUTSYNC_IND },
  { &hf_nbap_t_RLFAILURE    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_RLFAILURE },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Synchronisation_Configuration_Cell_SetupRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Synchronisation_Configuration_Cell_SetupRqst, Synchronisation_Configuration_Cell_SetupRqst_sequence);

  return offset;
}


static const per_sequence_t PrimarySCH_Information_Cell_SetupRqstFDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_primarySCH_Power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_tSTD_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TSTD_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PrimarySCH_Information_Cell_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PrimarySCH_Information_Cell_SetupRqstFDD, PrimarySCH_Information_Cell_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t SecondarySCH_Information_Cell_SetupRqstFDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_secondarySCH_Power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_tSTD_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TSTD_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SecondarySCH_Information_Cell_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SecondarySCH_Information_Cell_SetupRqstFDD, SecondarySCH_Information_Cell_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t PrimaryCPICH_Information_Cell_SetupRqstFDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_primaryCPICH_Power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PrimaryCPICH_Power },
  { &hf_nbap_transmitDiversityIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransmitDiversityIndicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PrimaryCPICH_Information_Cell_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PrimaryCPICH_Information_Cell_SetupRqstFDD, PrimaryCPICH_Information_Cell_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t SecondaryCPICH_InformationList_Cell_SetupRqstFDD_sequence_of[1] = {
  { &hf_nbap_SecondaryCPICH_InformationList_Cell_SetupRqstFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_SecondaryCPICH_InformationList_Cell_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_SecondaryCPICH_InformationList_Cell_SetupRqstFDD, SecondaryCPICH_InformationList_Cell_SetupRqstFDD_sequence_of,
                                                  1, maxSCPICHCell, FALSE);

  return offset;
}


static const per_sequence_t SecondaryCPICH_InformationItem_Cell_SetupRqstFDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_dl_ScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_ScramblingCode },
  { &hf_nbap_fDD_DL_ChannelisationCodeNumber, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_DL_ChannelisationCodeNumber },
  { &hf_nbap_secondaryCPICH_Power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_transmitDiversityIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransmitDiversityIndicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SecondaryCPICH_InformationItem_Cell_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SecondaryCPICH_InformationItem_Cell_SetupRqstFDD, SecondaryCPICH_InformationItem_Cell_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t BCH_Information_Cell_SetupRqstFDD_sequence[] = {
  { &hf_nbap_commonTransportChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonTransportChannelID },
  { &hf_nbap_bCH_Power      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_BCH_Information_Cell_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_BCH_Information_Cell_SetupRqstFDD, BCH_Information_Cell_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t PrimaryCCPCH_Information_Cell_SetupRqstFDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_bCH_information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BCH_Information_Cell_SetupRqstFDD },
  { &hf_nbap_sTTD_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_STTD_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PrimaryCCPCH_Information_Cell_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PrimaryCCPCH_Information_Cell_SetupRqstFDD, PrimaryCCPCH_Information_Cell_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t Limited_power_increase_information_Cell_SetupRqstFDD_sequence[] = {
  { &hf_nbap_powerRaiseLimit, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PowerRaiseLimit },
  { &hf_nbap_dLPowerAveragingWindowSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DLPowerAveragingWindowSize },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Limited_power_increase_information_Cell_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Limited_power_increase_information_Cell_SetupRqstFDD, Limited_power_increase_information_Cell_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t IPDLParameter_Information_Cell_SetupRqstFDD_sequence[] = {
  { &hf_nbap_iPDL_FDD_Parameters, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_IPDL_FDD_Parameters },
  { &hf_nbap_iPDL_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_IPDL_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_IPDLParameter_Information_Cell_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_IPDLParameter_Information_Cell_SetupRqstFDD, IPDLParameter_Information_Cell_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t CellPortion_InformationList_Cell_SetupRqstFDD_sequence_of[1] = {
  { &hf_nbap_CellPortion_InformationList_Cell_SetupRqstFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_CellPortion_InformationList_Cell_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_CellPortion_InformationList_Cell_SetupRqstFDD, CellPortion_InformationList_Cell_SetupRqstFDD_sequence_of,
                                                  1, maxNrOfCellPortionsPerCell, FALSE);

  return offset;
}


static const per_sequence_t CellPortion_InformationItem_Cell_SetupRqstFDD_sequence[] = {
  { &hf_nbap_cellPortionID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellPortionID },
  { &hf_nbap_associatedSecondaryCPICH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_maximumTransmissionPowerforCellPortion, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MaximumTransmissionPower },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellPortion_InformationItem_Cell_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellPortion_InformationItem_Cell_SetupRqstFDD, CellPortion_InformationItem_Cell_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t CellSetupRequestTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSetupRequestTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSetupRequestTDD, CellSetupRequestTDD_sequence);

  return offset;
}



static int
dissect_nbap_SyncCaseIndicator_Cell_SetupRqstTDD_PSCH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t SCH_Information_Cell_SetupRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_syncCaseIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SyncCaseIndicator_Cell_SetupRqstTDD_PSCH },
  { &hf_nbap_sCH_Power      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_tSTD_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TSTD_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SCH_Information_Cell_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SCH_Information_Cell_SetupRqstTDD, SCH_Information_Cell_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t Case1_Cell_SetupRqstTDD_sequence[] = {
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Case1_Cell_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Case1_Cell_SetupRqstTDD, Case1_Cell_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t Case2_Cell_SetupRqstTDD_sequence[] = {
  { &hf_nbap_sCH_TimeSlot   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SCH_TimeSlot },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Case2_Cell_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Case2_Cell_SetupRqstTDD, Case2_Cell_SetupRqstTDD_sequence);

  return offset;
}


static const value_string nbap_SyncCaseIndicatorItem_Cell_SetupRqstTDD_PSCH_vals[] = {
  {   0, "case1" },
  {   1, "case2" },
  { 0, NULL }
};

static const per_choice_t SyncCaseIndicatorItem_Cell_SetupRqstTDD_PSCH_choice[] = {
  {   0, &hf_nbap_case1          , ASN1_EXTENSION_ROOT    , dissect_nbap_Case1_Cell_SetupRqstTDD },
  {   1, &hf_nbap_case2          , ASN1_EXTENSION_ROOT    , dissect_nbap_Case2_Cell_SetupRqstTDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_SyncCaseIndicatorItem_Cell_SetupRqstTDD_PSCH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_SyncCaseIndicatorItem_Cell_SetupRqstTDD_PSCH, SyncCaseIndicatorItem_Cell_SetupRqstTDD_PSCH_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t PCCPCH_Information_Cell_SetupRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_tdd_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_pCCPCH_Power   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PCCPCH_Power },
  { &hf_nbap_sCTD_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SCTD_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PCCPCH_Information_Cell_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PCCPCH_Information_Cell_SetupRqstTDD, PCCPCH_Information_Cell_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t TimeSlotConfigurationItem_Cell_SetupRqstTDD_sequence[] = {
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_timeSlotStatus , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotStatus },
  { &hf_nbap_timeSlotDirection, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotDirection },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TimeSlotConfigurationItem_Cell_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TimeSlotConfigurationItem_Cell_SetupRqstTDD, TimeSlotConfigurationItem_Cell_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t TimeSlotConfigurationList_Cell_SetupRqstTDD_sequence_of[1] = {
  { &hf_nbap_TimeSlotConfigurationList_Cell_SetupRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotConfigurationItem_Cell_SetupRqstTDD },
};

static int
dissect_nbap_TimeSlotConfigurationList_Cell_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_TimeSlotConfigurationList_Cell_SetupRqstTDD, TimeSlotConfigurationList_Cell_SetupRqstTDD_sequence_of,
                                                  1, 15, FALSE);

  return offset;
}


static const per_sequence_t TimeSlotConfigurationItem_LCR_Cell_SetupRqstTDD_sequence[] = {
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_timeSlotStatus , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotStatus },
  { &hf_nbap_timeSlotDirection, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotDirection },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TimeSlotConfigurationItem_LCR_Cell_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TimeSlotConfigurationItem_LCR_Cell_SetupRqstTDD, TimeSlotConfigurationItem_LCR_Cell_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t TimeSlotConfigurationList_LCR_Cell_SetupRqstTDD_sequence_of[1] = {
  { &hf_nbap_TimeSlotConfigurationList_LCR_Cell_SetupRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotConfigurationItem_LCR_Cell_SetupRqstTDD },
};

static int
dissect_nbap_TimeSlotConfigurationList_LCR_Cell_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_TimeSlotConfigurationList_LCR_Cell_SetupRqstTDD, TimeSlotConfigurationList_LCR_Cell_SetupRqstTDD_sequence_of,
                                                  1, 7, FALSE);

  return offset;
}


static const per_sequence_t PCCPCH_LCR_Information_Cell_SetupRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_tdd_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_pCCPCH_Power   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PCCPCH_Power },
  { &hf_nbap_sCTD_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SCTD_Indicator },
  { &hf_nbap_tSTD_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TSTD_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PCCPCH_LCR_Information_Cell_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PCCPCH_LCR_Information_Cell_SetupRqstTDD, PCCPCH_LCR_Information_Cell_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DwPCH_LCR_Information_Cell_SetupRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelId, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_tSTD_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TSTD_Indicator },
  { &hf_nbap_dwPCH_Power    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DwPCH_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DwPCH_LCR_Information_Cell_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DwPCH_LCR_Information_Cell_SetupRqstTDD, DwPCH_LCR_Information_Cell_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t IPDLParameter_Information_Cell_SetupRqstTDD_sequence[] = {
  { &hf_nbap_iPDL_TDD_Parameters, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_IPDL_TDD_Parameters },
  { &hf_nbap_iPDL_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_IPDL_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_IPDLParameter_Information_Cell_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_IPDLParameter_Information_Cell_SetupRqstTDD, IPDLParameter_Information_Cell_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t IPDLParameter_Information_LCR_Cell_SetupRqstTDD_sequence[] = {
  { &hf_nbap_iPDL_TDD_Parameters_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_IPDL_TDD_Parameters_LCR },
  { &hf_nbap_iPDL_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_IPDL_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_IPDLParameter_Information_LCR_Cell_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_IPDLParameter_Information_LCR_Cell_SetupRqstTDD, IPDLParameter_Information_LCR_Cell_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t PCCPCH_768_Information_Cell_SetupRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID768 },
  { &hf_nbap_tdd_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_pCCPCH_Power   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PCCPCH_Power },
  { &hf_nbap_sCTD_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SCTD_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PCCPCH_768_Information_Cell_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PCCPCH_768_Information_Cell_SetupRqstTDD, PCCPCH_768_Information_Cell_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t SCH_768_Information_Cell_SetupRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID768 },
  { &hf_nbap_syncCaseIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SyncCaseIndicator_Cell_SetupRqstTDD_PSCH },
  { &hf_nbap_sCH_Power      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_tSTD_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TSTD_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SCH_768_Information_Cell_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SCH_768_Information_Cell_SetupRqstTDD, SCH_768_Information_Cell_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t Cell_Frequency_Item_LCR_MulFreq_Cell_SetupRqstTDD_sequence[] = {
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UARFCN },
  { &hf_nbap_timeSlotConfigurationList_LCR_Cell_SetupRqstTDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotConfigurationList_LCR_Cell_SetupRqstTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Cell_Frequency_Item_LCR_MulFreq_Cell_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Cell_Frequency_Item_LCR_MulFreq_Cell_SetupRqstTDD, Cell_Frequency_Item_LCR_MulFreq_Cell_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t Cell_Frequency_List_LCR_MulFreq_Cell_SetupRqstTDD_sequence_of[1] = {
  { &hf_nbap_Cell_Frequency_List_LCR_MulFreq_Cell_SetupRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Cell_Frequency_Item_LCR_MulFreq_Cell_SetupRqstTDD },
};

static int
dissect_nbap_Cell_Frequency_List_LCR_MulFreq_Cell_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Cell_Frequency_List_LCR_MulFreq_Cell_SetupRqstTDD, Cell_Frequency_List_LCR_MulFreq_Cell_SetupRqstTDD_sequence_of,
                                                  1, maxFrequencyinCell_1, FALSE);

  return offset;
}


static const per_sequence_t CellSetupResponse_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSetupResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"CellSetupResponse ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSetupResponse, CellSetupResponse_sequence);

  return offset;
}


static const per_sequence_t CellSetupFailure_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSetupFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"CellSetupFailure ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;


  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSetupFailure, CellSetupFailure_sequence);

  return offset;
}


static const per_sequence_t CellReconfigurationRequestFDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellReconfigurationRequestFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellReconfigurationRequestFDD, CellReconfigurationRequestFDD_sequence);

  return offset;
}


static const per_sequence_t Synchronisation_Configuration_Cell_ReconfRqst_sequence[] = {
  { &hf_nbap_n_INSYNC_IND   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_N_INSYNC_IND },
  { &hf_nbap_n_OUTSYNC_IND  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_N_OUTSYNC_IND },
  { &hf_nbap_t_RLFAILURE    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_T_RLFAILURE },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Synchronisation_Configuration_Cell_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Synchronisation_Configuration_Cell_ReconfRqst, Synchronisation_Configuration_Cell_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t PrimarySCH_Information_Cell_ReconfRqstFDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_primarySCH_Power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PrimarySCH_Information_Cell_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PrimarySCH_Information_Cell_ReconfRqstFDD, PrimarySCH_Information_Cell_ReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t SecondarySCH_Information_Cell_ReconfRqstFDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_secondarySCH_Power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SecondarySCH_Information_Cell_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SecondarySCH_Information_Cell_ReconfRqstFDD, SecondarySCH_Information_Cell_ReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t PrimaryCPICH_Information_Cell_ReconfRqstFDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_primaryCPICH_Power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PrimaryCPICH_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PrimaryCPICH_Information_Cell_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PrimaryCPICH_Information_Cell_ReconfRqstFDD, PrimaryCPICH_Information_Cell_ReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t SecondaryCPICH_InformationList_Cell_ReconfRqstFDD_sequence_of[1] = {
  { &hf_nbap_SecondaryCPICH_InformationList_Cell_ReconfRqstFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_SecondaryCPICH_InformationList_Cell_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_SecondaryCPICH_InformationList_Cell_ReconfRqstFDD, SecondaryCPICH_InformationList_Cell_ReconfRqstFDD_sequence_of,
                                                  1, maxSCPICHCell, FALSE);

  return offset;
}


static const per_sequence_t SecondaryCPICH_InformationItem_Cell_ReconfRqstFDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_secondaryCPICH_Power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SecondaryCPICH_InformationItem_Cell_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SecondaryCPICH_InformationItem_Cell_ReconfRqstFDD, SecondaryCPICH_InformationItem_Cell_ReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t BCH_information_Cell_ReconfRqstFDD_sequence[] = {
  { &hf_nbap_commonTransportChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonTransportChannelID },
  { &hf_nbap_bCH_Power      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_BCH_information_Cell_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_BCH_information_Cell_ReconfRqstFDD, BCH_information_Cell_ReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t PrimaryCCPCH_Information_Cell_ReconfRqstFDD_sequence[] = {
  { &hf_nbap_bCH_information_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BCH_information_Cell_ReconfRqstFDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PrimaryCCPCH_Information_Cell_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PrimaryCCPCH_Information_Cell_ReconfRqstFDD, PrimaryCCPCH_Information_Cell_ReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t IPDLParameter_Information_Cell_ReconfRqstFDD_sequence[] = {
  { &hf_nbap_iPDL_FDD_Parameters, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_IPDL_FDD_Parameters },
  { &hf_nbap_iPDL_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_IPDL_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_IPDLParameter_Information_Cell_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_IPDLParameter_Information_Cell_ReconfRqstFDD, IPDLParameter_Information_Cell_ReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t CellPortion_InformationList_Cell_ReconfRqstFDD_sequence_of[1] = {
  { &hf_nbap_CellPortion_InformationList_Cell_ReconfRqstFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_CellPortion_InformationList_Cell_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_CellPortion_InformationList_Cell_ReconfRqstFDD, CellPortion_InformationList_Cell_ReconfRqstFDD_sequence_of,
                                                  1, maxNrOfCellPortionsPerCell, FALSE);

  return offset;
}


static const per_sequence_t CellPortion_InformationItem_Cell_ReconfRqstFDD_sequence[] = {
  { &hf_nbap_cellPortionID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellPortionID },
  { &hf_nbap_maximumTransmissionPowerforCellPortion, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MaximumTransmissionPower },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellPortion_InformationItem_Cell_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellPortion_InformationItem_Cell_ReconfRqstFDD, CellPortion_InformationItem_Cell_ReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t CellReconfigurationRequestTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellReconfigurationRequestTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellReconfigurationRequestTDD, CellReconfigurationRequestTDD_sequence);

  return offset;
}


static const per_sequence_t SCH_Information_Cell_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_sCH_Power      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SCH_Information_Cell_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SCH_Information_Cell_ReconfRqstTDD, SCH_Information_Cell_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t PCCPCH_Information_Cell_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_pCCPCH_Power   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PCCPCH_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PCCPCH_Information_Cell_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PCCPCH_Information_Cell_ReconfRqstTDD, PCCPCH_Information_Cell_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t TimeSlotConfigurationItem_Cell_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_timeSlotStatus , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotStatus },
  { &hf_nbap_timeSlotDirection, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotDirection },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TimeSlotConfigurationItem_Cell_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TimeSlotConfigurationItem_Cell_ReconfRqstTDD, TimeSlotConfigurationItem_Cell_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t TimeSlotConfigurationList_Cell_ReconfRqstTDD_sequence_of[1] = {
  { &hf_nbap_TimeSlotConfigurationList_Cell_ReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotConfigurationItem_Cell_ReconfRqstTDD },
};

static int
dissect_nbap_TimeSlotConfigurationList_Cell_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_TimeSlotConfigurationList_Cell_ReconfRqstTDD, TimeSlotConfigurationList_Cell_ReconfRqstTDD_sequence_of,
                                                  1, 15, FALSE);

  return offset;
}


static const per_sequence_t TimeSlotConfigurationItem_LCR_Cell_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_timeSlotStatus , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotStatus },
  { &hf_nbap_timeSlotDirection, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotDirection },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_TimeSlotConfigurationItem_LCR_Cell_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_TimeSlotConfigurationItem_LCR_Cell_ReconfRqstTDD, TimeSlotConfigurationItem_LCR_Cell_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t TimeSlotConfigurationList_LCR_Cell_ReconfRqstTDD_sequence_of[1] = {
  { &hf_nbap_TimeSlotConfigurationList_LCR_Cell_ReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotConfigurationItem_LCR_Cell_ReconfRqstTDD },
};

static int
dissect_nbap_TimeSlotConfigurationList_LCR_Cell_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_TimeSlotConfigurationList_LCR_Cell_ReconfRqstTDD, TimeSlotConfigurationList_LCR_Cell_ReconfRqstTDD_sequence_of,
                                                  1, 7, FALSE);

  return offset;
}


static const per_sequence_t DwPCH_LCR_Information_Cell_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelId, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID },
  { &hf_nbap_dwPCH_Power    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DwPCH_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DwPCH_LCR_Information_Cell_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DwPCH_LCR_Information_Cell_ReconfRqstTDD, DwPCH_LCR_Information_Cell_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t IPDLParameter_Information_Cell_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_iPDL_TDD_Parameters, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_IPDL_TDD_Parameters },
  { &hf_nbap_iPDL_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_IPDL_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_IPDLParameter_Information_Cell_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_IPDLParameter_Information_Cell_ReconfRqstTDD, IPDLParameter_Information_Cell_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t IPDLParameter_Information_LCR_Cell_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_iPDL_TDD_Parameters_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_IPDL_TDD_Parameters_LCR },
  { &hf_nbap_iPDL_Indicator , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_IPDL_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_IPDLParameter_Information_LCR_Cell_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_IPDLParameter_Information_LCR_Cell_ReconfRqstTDD, IPDLParameter_Information_LCR_Cell_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t SCH_768_Information_Cell_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID768 },
  { &hf_nbap_sCH_Power      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SCH_768_Information_Cell_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SCH_768_Information_Cell_ReconfRqstTDD, SCH_768_Information_Cell_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t PCCPCH_768_Information_Cell_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_commonPhysicalChannelID768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonPhysicalChannelID768 },
  { &hf_nbap_pCCPCH_Power   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PCCPCH_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PCCPCH_768_Information_Cell_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PCCPCH_768_Information_Cell_ReconfRqstTDD, PCCPCH_768_Information_Cell_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t Cell_Frequency_Add_LCR_MulFreq_Cell_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UARFCN },
  { &hf_nbap_timeSlotConfigurationList_LCR_Cell_ReconfRqstTDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotConfigurationList_LCR_Cell_ReconfRqstTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Cell_Frequency_Add_LCR_MulFreq_Cell_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Cell_Frequency_Add_LCR_MulFreq_Cell_ReconfRqstTDD, Cell_Frequency_Add_LCR_MulFreq_Cell_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t Cell_Frequency_ModifyItem_LCR_MulFreq_Cell_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UARFCN },
  { &hf_nbap_timeSlotConfigurationList_LCR_Cell_ReconfRqstTDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotConfigurationList_LCR_Cell_ReconfRqstTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Cell_Frequency_ModifyItem_LCR_MulFreq_Cell_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Cell_Frequency_ModifyItem_LCR_MulFreq_Cell_ReconfRqstTDD, Cell_Frequency_ModifyItem_LCR_MulFreq_Cell_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t Cell_Frequency_ModifyList_LCR_MulFreq_Cell_ReconfRqstTDD_sequence_of[1] = {
  { &hf_nbap_Cell_Frequency_ModifyList_LCR_MulFreq_Cell_ReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Cell_Frequency_ModifyItem_LCR_MulFreq_Cell_ReconfRqstTDD },
};

static int
dissect_nbap_Cell_Frequency_ModifyList_LCR_MulFreq_Cell_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Cell_Frequency_ModifyList_LCR_MulFreq_Cell_ReconfRqstTDD, Cell_Frequency_ModifyList_LCR_MulFreq_Cell_ReconfRqstTDD_sequence_of,
                                                  1, maxFrequencyinCell_1, FALSE);

  return offset;
}


static const per_sequence_t Cell_Frequency_Delete_LCR_MulFreq_Cell_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UARFCN },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Cell_Frequency_Delete_LCR_MulFreq_Cell_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Cell_Frequency_Delete_LCR_MulFreq_Cell_ReconfRqstTDD, Cell_Frequency_Delete_LCR_MulFreq_Cell_ReconfRqstTDD_sequence);

  return offset;
}


static const value_string nbap_UARFCN_Adjustment_vals[] = {
  {   0, "cell-Frequency-Add-LCR-MulFreq-Cell-ReconfRqstTDD" },
  {   1, "cell-Frequency-ModifyList-LCR-MulFreq-Cell-ReconfRqstTDD" },
  {   2, "cell-Frequency-Delete-LCR-MulFreq-Cell-ReconfRqstTDD" },
  { 0, NULL }
};

static const per_choice_t UARFCN_Adjustment_choice[] = {
  {   0, &hf_nbap_cell_Frequency_Add_LCR_MulFreq_Cell_ReconfRqstTDD, ASN1_EXTENSION_ROOT    , dissect_nbap_Cell_Frequency_Add_LCR_MulFreq_Cell_ReconfRqstTDD },
  {   1, &hf_nbap_cell_Frequency_ModifyList_LCR_MulFreq_Cell_ReconfRqstTDD, ASN1_EXTENSION_ROOT    , dissect_nbap_Cell_Frequency_ModifyList_LCR_MulFreq_Cell_ReconfRqstTDD },
  {   2, &hf_nbap_cell_Frequency_Delete_LCR_MulFreq_Cell_ReconfRqstTDD, ASN1_EXTENSION_ROOT    , dissect_nbap_Cell_Frequency_Delete_LCR_MulFreq_Cell_ReconfRqstTDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_UARFCN_Adjustment(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_UARFCN_Adjustment, UARFCN_Adjustment_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t CellReconfigurationResponse_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellReconfigurationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellReconfigurationResponse, CellReconfigurationResponse_sequence);

  return offset;
}


static const per_sequence_t CellReconfigurationFailure_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellReconfigurationFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellReconfigurationFailure, CellReconfigurationFailure_sequence);

  return offset;
}


static const per_sequence_t CellDeletionRequest_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellDeletionRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellDeletionRequest, CellDeletionRequest_sequence);

  return offset;
}


static const per_sequence_t CellDeletionResponse_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellDeletionResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellDeletionResponse, CellDeletionResponse_sequence);

  return offset;
}


static const per_sequence_t ResourceStatusIndication_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ResourceStatusIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"ResourceStatusIndication ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ResourceStatusIndication, ResourceStatusIndication_sequence);

  return offset;
}


static const per_sequence_t Local_Cell_InformationList_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_Local_Cell_InformationList_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_Local_Cell_InformationList_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Local_Cell_InformationList_ResourceStatusInd, Local_Cell_InformationList_ResourceStatusInd_sequence_of,
                                                  1, maxLocalCellinNodeB, FALSE);

  return offset;
}


static const per_sequence_t Local_Cell_Group_InformationList_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_Local_Cell_Group_InformationList_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_Local_Cell_Group_InformationList_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Local_Cell_Group_InformationList_ResourceStatusInd, Local_Cell_Group_InformationList_ResourceStatusInd_sequence_of,
                                                  1, maxLocalCellinNodeB, FALSE);

  return offset;
}


static const per_sequence_t No_Failure_ResourceStatusInd_sequence[] = {
  { &hf_nbap_local_Cell_InformationList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Local_Cell_InformationList_ResourceStatusInd },
  { &hf_nbap_local_Cell_Group_InformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Local_Cell_Group_InformationList_ResourceStatusInd },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_No_Failure_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_No_Failure_ResourceStatusInd, No_Failure_ResourceStatusInd_sequence);

  return offset;
}


static const per_sequence_t Local_Cell_InformationList2_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_Local_Cell_InformationList2_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_Local_Cell_InformationList2_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Local_Cell_InformationList2_ResourceStatusInd, Local_Cell_InformationList2_ResourceStatusInd_sequence_of,
                                                  1, maxLocalCellinNodeB, FALSE);

  return offset;
}


static const per_sequence_t Local_Cell_Group_InformationList2_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_Local_Cell_Group_InformationList2_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_Local_Cell_Group_InformationList2_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Local_Cell_Group_InformationList2_ResourceStatusInd, Local_Cell_Group_InformationList2_ResourceStatusInd_sequence_of,
                                                  1, maxLocalCellinNodeB, FALSE);

  return offset;
}


static const per_sequence_t CCP_InformationList_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_CCP_InformationList_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_CCP_InformationList_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_CCP_InformationList_ResourceStatusInd, CCP_InformationList_ResourceStatusInd_sequence_of,
                                                  1, maxCCPinNodeB, FALSE);

  return offset;
}


static const per_sequence_t Cell_InformationList_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_Cell_InformationList_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_Cell_InformationList_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Cell_InformationList_ResourceStatusInd, Cell_InformationList_ResourceStatusInd_sequence_of,
                                                  1, maxCellinNodeB, FALSE);

  return offset;
}


static const per_sequence_t ServiceImpacting_ResourceStatusInd_sequence[] = {
  { &hf_nbap_local_Cell_InformationList_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Local_Cell_InformationList2_ResourceStatusInd },
  { &hf_nbap_local_Cell_Group_InformationList_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Local_Cell_Group_InformationList2_ResourceStatusInd },
  { &hf_nbap_cCP_InformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CCP_InformationList_ResourceStatusInd },
  { &hf_nbap_cell_InformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Cell_InformationList_ResourceStatusInd },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ServiceImpacting_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ServiceImpacting_ResourceStatusInd, ServiceImpacting_ResourceStatusInd_sequence);

  return offset;
}


static const value_string nbap_IndicationType_ResourceStatusInd_vals[] = {
  {   0, "no-Failure" },
  {   1, "serviceImpacting" },
  { 0, NULL }
};

static const per_choice_t IndicationType_ResourceStatusInd_choice[] = {
  {   0, &hf_nbap_no_Failure     , ASN1_EXTENSION_ROOT    , dissect_nbap_No_Failure_ResourceStatusInd },
  {   1, &hf_nbap_serviceImpacting, ASN1_EXTENSION_ROOT    , dissect_nbap_ServiceImpacting_ResourceStatusInd },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_IndicationType_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_IndicationType_ResourceStatusInd, IndicationType_ResourceStatusInd_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Local_Cell_InformationItem_ResourceStatusInd_sequence[] = {
  { &hf_nbap_local_CellID   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Local_Cell_ID },
  { &hf_nbap_addorDeleteIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AddorDeleteIndicator },
  { &hf_nbap_dl_or_global_capacityCredit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_or_Global_CapacityCredit },
  { &hf_nbap_ul_capacityCredit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_CapacityCredit },
  { &hf_nbap_commonChannelsCapacityConsumptionLaw, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CommonChannelsCapacityConsumptionLaw },
  { &hf_nbap_dedicatedChannelsCapacityConsumptionLaw, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DedicatedChannelsCapacityConsumptionLaw },
  { &hf_nbap_maximumDL_PowerCapability, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MaximumDL_PowerCapability },
  { &hf_nbap_minSpreadingFactor, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MinSpreadingFactor },
  { &hf_nbap_minimumDL_PowerCapability, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MinimumDL_PowerCapability },
  { &hf_nbap_local_Cell_Group_ID, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Local_Cell_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Local_Cell_InformationItem_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Local_Cell_InformationItem_ResourceStatusInd, Local_Cell_InformationItem_ResourceStatusInd_sequence);

  return offset;
}


static const per_sequence_t Local_Cell_Group_InformationItem_ResourceStatusInd_sequence[] = {
  { &hf_nbap_local_Cell_Group_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Local_Cell_ID },
  { &hf_nbap_dl_or_global_capacityCredit, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_or_Global_CapacityCredit },
  { &hf_nbap_ul_capacityCredit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_CapacityCredit },
  { &hf_nbap_commonChannelsCapacityConsumptionLaw, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommonChannelsCapacityConsumptionLaw },
  { &hf_nbap_dedicatedChannelsCapacityConsumptionLaw, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DedicatedChannelsCapacityConsumptionLaw },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Local_Cell_Group_InformationItem_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Local_Cell_Group_InformationItem_ResourceStatusInd, Local_Cell_Group_InformationItem_ResourceStatusInd_sequence);

  return offset;
}


static const per_sequence_t Power_Local_Cell_Group_InformationList_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_Power_Local_Cell_Group_InformationList_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_Power_Local_Cell_Group_InformationList_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Power_Local_Cell_Group_InformationList_ResourceStatusInd, Power_Local_Cell_Group_InformationList_ResourceStatusInd_sequence_of,
                                                  1, maxLocalCellinNodeB, FALSE);

  return offset;
}


static const per_sequence_t Power_Local_Cell_Group_InformationItem_ResourceStatusInd_sequence[] = {
  { &hf_nbap_power_Local_Cell_Group_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Local_Cell_ID },
  { &hf_nbap_maximumDL_PowerCapability, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MaximumDL_PowerCapability },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Power_Local_Cell_Group_InformationItem_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Power_Local_Cell_Group_InformationItem_ResourceStatusInd, Power_Local_Cell_Group_InformationItem_ResourceStatusInd_sequence);

  return offset;
}


static const per_sequence_t Local_Cell_InformationItem2_ResourceStatusInd_sequence[] = {
  { &hf_nbap_local_Cell_ID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Local_Cell_ID },
  { &hf_nbap_dl_or_global_capacityCredit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_or_Global_CapacityCredit },
  { &hf_nbap_ul_capacityCredit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_CapacityCredit },
  { &hf_nbap_commonChannelsCapacityConsumptionLaw, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CommonChannelsCapacityConsumptionLaw },
  { &hf_nbap_dedicatedChannelsCapacityConsumptionLaw, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DedicatedChannelsCapacityConsumptionLaw },
  { &hf_nbap_maximum_DL_PowerCapability, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MaximumDL_PowerCapability },
  { &hf_nbap_minSpreadingFactor, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MinSpreadingFactor },
  { &hf_nbap_minimumDL_PowerCapability, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MinimumDL_PowerCapability },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Local_Cell_InformationItem2_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Local_Cell_InformationItem2_ResourceStatusInd, Local_Cell_InformationItem2_ResourceStatusInd_sequence);

  return offset;
}


static const per_sequence_t Local_Cell_Group_InformationItem2_ResourceStatusInd_sequence[] = {
  { &hf_nbap_local_Cell_Group_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Local_Cell_ID },
  { &hf_nbap_dl_or_global_capacityCredit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_or_Global_CapacityCredit },
  { &hf_nbap_ul_capacityCredit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_CapacityCredit },
  { &hf_nbap_commonChannelsCapacityConsumptionLaw, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CommonChannelsCapacityConsumptionLaw },
  { &hf_nbap_dedicatedChannelsCapacityConsumptionLaw, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DedicatedChannelsCapacityConsumptionLaw },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Local_Cell_Group_InformationItem2_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Local_Cell_Group_InformationItem2_ResourceStatusInd, Local_Cell_Group_InformationItem2_ResourceStatusInd_sequence);

  return offset;
}


static const per_sequence_t CCP_InformationItem_ResourceStatusInd_sequence[] = {
  { &hf_nbap_communicationControlPortID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommunicationControlPortID },
  { &hf_nbap_resourceOperationalState, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ResourceOperationalState },
  { &hf_nbap_availabilityStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AvailabilityStatus },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CCP_InformationItem_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CCP_InformationItem_ResourceStatusInd, CCP_InformationItem_ResourceStatusInd_sequence);

  return offset;
}



static int
dissect_nbap_P_SCH_Information_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_S_SCH_Information_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_P_CPICH_Information_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t S_CPICH_InformationList_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_S_CPICH_InformationList_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_S_CPICH_InformationList_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_S_CPICH_InformationList_ResourceStatusInd, S_CPICH_InformationList_ResourceStatusInd_sequence_of,
                                                  1, maxSCPICHCell, FALSE);

  return offset;
}



static int
dissect_nbap_P_CCPCH_Information_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_BCH_Information_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t S_CCPCH_InformationList_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_S_CCPCH_InformationList_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_S_CCPCH_InformationList_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_S_CCPCH_InformationList_ResourceStatusInd, S_CCPCH_InformationList_ResourceStatusInd_sequence_of,
                                                  1, maxSCCPCHCell, FALSE);

  return offset;
}



static int
dissect_nbap_PCH_Information_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_PICH_Information_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t FACH_InformationList_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_FACH_InformationList_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_FACH_InformationList_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_FACH_InformationList_ResourceStatusInd, FACH_InformationList_ResourceStatusInd_sequence_of,
                                                  1, maxFACHCell, FALSE);

  return offset;
}


static const per_sequence_t PRACH_InformationList_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_PRACH_InformationList_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_PRACH_InformationList_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_PRACH_InformationList_ResourceStatusInd, PRACH_InformationList_ResourceStatusInd_sequence_of,
                                                  1, maxPRACHCell, FALSE);

  return offset;
}


static const per_sequence_t RACH_InformationList_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_RACH_InformationList_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_RACH_InformationList_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RACH_InformationList_ResourceStatusInd, RACH_InformationList_ResourceStatusInd_sequence_of,
                                                  1, maxPRACHCell, FALSE);

  return offset;
}


static const per_sequence_t AICH_InformationList_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_AICH_InformationList_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_AICH_InformationList_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_AICH_InformationList_ResourceStatusInd, AICH_InformationList_ResourceStatusInd_sequence_of,
                                                  1, maxPRACHCell, FALSE);

  return offset;
}



static int
dissect_nbap_SCH_Information_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t Cell_InformationItem_ResourceStatusInd_sequence[] = {
  { &hf_nbap_c_ID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_C_ID },
  { &hf_nbap_resourceOperationalState, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ResourceOperationalState },
  { &hf_nbap_availabilityStatus, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_AvailabilityStatus },
  { &hf_nbap_primary_SCH_Information_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_P_SCH_Information_ResourceStatusInd },
  { &hf_nbap_secondary_SCH_Information_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_S_SCH_Information_ResourceStatusInd },
  { &hf_nbap_primary_CPICH_Information_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_P_CPICH_Information_ResourceStatusInd },
  { &hf_nbap_secondary_CPICH_Information_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_S_CPICH_InformationList_ResourceStatusInd },
  { &hf_nbap_primary_CCPCH_Information_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_P_CCPCH_Information_ResourceStatusInd },
  { &hf_nbap_bCH_Information_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_BCH_Information_ResourceStatusInd },
  { &hf_nbap_secondary_CCPCH_InformationList_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_S_CCPCH_InformationList_ResourceStatusInd },
  { &hf_nbap_pCH_Information_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PCH_Information_ResourceStatusInd },
  { &hf_nbap_pICH_Information_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PICH_Information_ResourceStatusInd },
  { &hf_nbap_fACH_InformationList_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_FACH_InformationList_ResourceStatusInd },
  { &hf_nbap_pRACH_InformationList_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PRACH_InformationList_ResourceStatusInd },
  { &hf_nbap_rACH_InformationList_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RACH_InformationList_ResourceStatusInd },
  { &hf_nbap_aICH_InformationList_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_AICH_InformationList_ResourceStatusInd },
  { &hf_nbap_notUsed_1_pCPCH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_notUsed_2_cPCH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_notUsed_3_aP_AICH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_notUsed_4_cDCA_ICH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_sCH_Information_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SCH_Information_ResourceStatusInd },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Cell_InformationItem_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Cell_InformationItem_ResourceStatusInd, Cell_InformationItem_ResourceStatusInd_sequence);

  return offset;
}


static const per_sequence_t FPACH_LCR_InformationList_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_FPACH_LCR_InformationList_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_FPACH_LCR_InformationList_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_FPACH_LCR_InformationList_ResourceStatusInd, FPACH_LCR_InformationList_ResourceStatusInd_sequence_of,
                                                  1, maxFPACHCell, FALSE);

  return offset;
}



static int
dissect_nbap_DwPCH_LCR_Information_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t HS_DSCH_Resources_Information_ResourceStatusInd_sequence[] = {
  { &hf_nbap_resourceOperationalState, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ResourceOperationalState },
  { &hf_nbap_availabilityStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AvailabilityStatus },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_DSCH_Resources_Information_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_DSCH_Resources_Information_ResourceStatusInd, HS_DSCH_Resources_Information_ResourceStatusInd_sequence);

  return offset;
}


static const per_sequence_t S_CCPCH_InformationListExt_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_S_CCPCH_InformationListExt_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_S_CCPCH_InformationListExt_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_S_CCPCH_InformationListExt_ResourceStatusInd, S_CCPCH_InformationListExt_ResourceStatusInd_sequence_of,
                                                  1, maxSCCPCHCellinExt, FALSE);

  return offset;
}


static const per_sequence_t S_CCPCH_LCR_InformationListExt_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_S_CCPCH_LCR_InformationListExt_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_S_CCPCH_LCR_InformationListExt_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_S_CCPCH_LCR_InformationListExt_ResourceStatusInd, S_CCPCH_LCR_InformationListExt_ResourceStatusInd_sequence_of,
                                                  1, maxSCCPCHCellinExtLCR, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_Resources_Information_ResourceStatusInd_sequence[] = {
  { &hf_nbap_resourceOperationalState, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ResourceOperationalState },
  { &hf_nbap_availabilityStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AvailabilityStatus },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_Resources_Information_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_Resources_Information_ResourceStatusInd, E_DCH_Resources_Information_ResourceStatusInd_sequence);

  return offset;
}


static const per_sequence_t PLCCH_InformationList_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_PLCCH_InformationList_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_PLCCH_InformationList_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_PLCCH_InformationList_ResourceStatusInd, PLCCH_InformationList_ResourceStatusInd_sequence_of,
                                                  1, maxPLCCHCell, FALSE);

  return offset;
}


static const per_sequence_t S_CCPCH_768_InformationList_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_S_CCPCH_768_InformationList_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_S_CCPCH_768_InformationList_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_S_CCPCH_768_InformationList_ResourceStatusInd, S_CCPCH_768_InformationList_ResourceStatusInd_sequence_of,
                                                  1, maxSCCPCHCell768, FALSE);

  return offset;
}


static const per_sequence_t PRACH_768_InformationList_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_PRACH_768_InformationList_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_PRACH_768_InformationList_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_PRACH_768_InformationList_ResourceStatusInd, PRACH_768_InformationList_ResourceStatusInd_sequence_of,
                                                  1, maxPRACHCell, FALSE);

  return offset;
}


static const per_sequence_t E_RUCCH_InformationList_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_E_RUCCH_InformationList_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_E_RUCCH_InformationList_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_RUCCH_InformationList_ResourceStatusInd, E_RUCCH_InformationList_ResourceStatusInd_sequence_of,
                                                  1, maxE_RUCCHCell, FALSE);

  return offset;
}


static const per_sequence_t E_RUCCH_768_InformationList_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_E_RUCCH_768_InformationList_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_E_RUCCH_768_InformationList_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_RUCCH_768_InformationList_ResourceStatusInd, E_RUCCH_768_InformationList_ResourceStatusInd_sequence_of,
                                                  1, maxE_RUCCHCell, FALSE);

  return offset;
}


static const per_sequence_t Cell_Frequency_List_Information_LCR_MulFreq_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_Cell_Frequency_List_Information_LCR_MulFreq_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_Cell_Frequency_List_Information_LCR_MulFreq_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Cell_Frequency_List_Information_LCR_MulFreq_ResourceStatusInd, Cell_Frequency_List_Information_LCR_MulFreq_ResourceStatusInd_sequence_of,
                                                  1, maxFrequencyinCell, FALSE);

  return offset;
}


static const per_sequence_t Cell_Frequency_List_InformationItem_LCR_MulFreq_ResourceStatusInd_sequence[] = {
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UARFCN },
  { &hf_nbap_resourceOperationalState, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ResourceOperationalState },
  { &hf_nbap_availabilityStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AvailabilityStatus },
  { &hf_nbap_cause          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Cause },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Cell_Frequency_List_InformationItem_LCR_MulFreq_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Cell_Frequency_List_InformationItem_LCR_MulFreq_ResourceStatusInd, Cell_Frequency_List_InformationItem_LCR_MulFreq_ResourceStatusInd_sequence);

  return offset;
}


static const per_sequence_t UPPCH_LCR_InformationList_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_UPPCH_LCR_InformationList_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_UPPCH_LCR_InformationList_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UPPCH_LCR_InformationList_ResourceStatusInd, UPPCH_LCR_InformationList_ResourceStatusInd_sequence_of,
                                                  1, maxFrequencyinCell, FALSE);

  return offset;
}


static const per_sequence_t UPPCH_LCR_InformationItem_ResourceStatusInd_sequence[] = {
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UARFCN },
  { &hf_nbap_uPPCHPositionLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UPPCHPositionLCR },
  { &hf_nbap_resourceOperationalState, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ResourceOperationalState },
  { &hf_nbap_availabilityStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_AvailabilityStatus },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UPPCH_LCR_InformationItem_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UPPCH_LCR_InformationItem_ResourceStatusInd, UPPCH_LCR_InformationItem_ResourceStatusInd_sequence);

  return offset;
}


static const per_sequence_t MultipleFreq_HS_DSCH_Resources_InformationList_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_MultipleFreq_HS_DSCH_Resources_InformationList_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_MultipleFreq_HS_DSCH_Resources_InformationList_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_MultipleFreq_HS_DSCH_Resources_InformationList_ResourceStatusInd, MultipleFreq_HS_DSCH_Resources_InformationList_ResourceStatusInd_sequence_of,
                                                  1, maxFrequencyinCell_1, FALSE);

  return offset;
}


static const per_sequence_t Power_Local_Cell_Group_InformationList2_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_Power_Local_Cell_Group_InformationList2_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_Power_Local_Cell_Group_InformationList2_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Power_Local_Cell_Group_InformationList2_ResourceStatusInd, Power_Local_Cell_Group_InformationList2_ResourceStatusInd_sequence_of,
                                                  1, maxLocalCellinNodeB, FALSE);

  return offset;
}


static const per_sequence_t Power_Local_Cell_Group_InformationItem2_ResourceStatusInd_sequence[] = {
  { &hf_nbap_power_Local_Cell_Group_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Local_Cell_ID },
  { &hf_nbap_maximumDL_PowerCapability, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MaximumDL_PowerCapability },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Power_Local_Cell_Group_InformationItem2_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Power_Local_Cell_Group_InformationItem2_ResourceStatusInd, Power_Local_Cell_Group_InformationItem2_ResourceStatusInd_sequence);

  return offset;
}


static const per_sequence_t MultipleFreq_E_DCH_Resources_InformationList_ResourceStatusInd_sequence_of[1] = {
  { &hf_nbap_MultipleFreq_E_DCH_Resources_InformationList_ResourceStatusInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_MultipleFreq_E_DCH_Resources_InformationList_ResourceStatusInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_MultipleFreq_E_DCH_Resources_InformationList_ResourceStatusInd, MultipleFreq_E_DCH_Resources_InformationList_ResourceStatusInd_sequence_of,
                                                  1, maxFrequencyinCell_1, FALSE);

  return offset;
}


static const per_sequence_t SystemInformationUpdateRequest_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SystemInformationUpdateRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"SystemInformationUpdateRequest ");
/* CRNC -> Node B */
actx->pinfo->link_dir=P2P_DIR_DL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SystemInformationUpdateRequest, SystemInformationUpdateRequest_sequence);

  return offset;
}



static int
dissect_nbap_SegmentInformationList_SystemInfoUpdate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t No_Deletion_SystemInfoUpdate_sequence[] = {
  { &hf_nbap_sIB_Originator , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SIB_Originator },
  { &hf_nbap_iB_SG_REP      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_IB_SG_REP },
  { &hf_nbap_segmentInformationList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SegmentInformationList_SystemInfoUpdate },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_No_Deletion_SystemInfoUpdate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_No_Deletion_SystemInfoUpdate, No_Deletion_SystemInfoUpdate_sequence);

  return offset;
}


static const value_string nbap_DeletionIndicator_SystemInfoUpdate_vals[] = {
  {   0, "no-Deletion" },
  {   1, "yes-Deletion" },
  { 0, NULL }
};

static const per_choice_t DeletionIndicator_SystemInfoUpdate_choice[] = {
  {   0, &hf_nbap_no_Deletion    , ASN1_NO_EXTENSIONS     , dissect_nbap_No_Deletion_SystemInfoUpdate },
  {   1, &hf_nbap_yes_Deletion   , ASN1_NO_EXTENSIONS     , dissect_nbap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_DeletionIndicator_SystemInfoUpdate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_DeletionIndicator_SystemInfoUpdate, DeletionIndicator_SystemInfoUpdate_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t MIB_SB_SIB_InformationItem_SystemInfoUpdateRqst_sequence[] = {
  { &hf_nbap_iB_Type        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_IB_Type },
  { &hf_nbap_iB_OC_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_IB_OC_ID },
  { &hf_nbap_deletionIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DeletionIndicator_SystemInfoUpdate },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MIB_SB_SIB_InformationItem_SystemInfoUpdateRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    nbap_get_private_data(actx->pinfo)->ib_type = 10; /* not-Used-sIB8 */
    nbap_get_private_data(actx->pinfo)->segment_type = 0;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MIB_SB_SIB_InformationItem_SystemInfoUpdateRqst, MIB_SB_SIB_InformationItem_SystemInfoUpdateRqst_sequence);

  return offset;
}


static const per_sequence_t MIB_SB_SIB_InformationList_SystemInfoUpdateRqst_sequence_of[1] = {
  { &hf_nbap_MIB_SB_SIB_InformationList_SystemInfoUpdateRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_MIB_SB_SIB_InformationItem_SystemInfoUpdateRqst },
};

static int
dissect_nbap_MIB_SB_SIB_InformationList_SystemInfoUpdateRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_MIB_SB_SIB_InformationList_SystemInfoUpdateRqst, MIB_SB_SIB_InformationList_SystemInfoUpdateRqst_sequence_of,
                                                  1, maxIB, FALSE);

  return offset;
}


static const per_sequence_t SegmentInformationItem_SystemInfoUpdate_sequence[] = {
  { &hf_nbap_iB_SG_POS      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_IB_SG_POS },
  { &hf_nbap_segment_Type   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Segment_Type },
  { &hf_nbap_iB_SG_DATA     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_IB_SG_DATA },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SegmentInformationItem_SystemInfoUpdate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SegmentInformationItem_SystemInfoUpdate, SegmentInformationItem_SystemInfoUpdate_sequence);

  return offset;
}


static const per_sequence_t SegmentInformationListIE_SystemInfoUpdate_sequence_of[1] = {
  { &hf_nbap_SegmentInformationListIE_SystemInfoUpdate_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_SegmentInformationItem_SystemInfoUpdate },
};

static int
dissect_nbap_SegmentInformationListIE_SystemInfoUpdate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_SegmentInformationListIE_SystemInfoUpdate, SegmentInformationListIE_SystemInfoUpdate_sequence_of,
                                                  1, maxIBSEG, FALSE);

  return offset;
}


static const per_sequence_t SystemInformationUpdateResponse_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SystemInformationUpdateResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"SystemInformationUpdateResponse ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SystemInformationUpdateResponse, SystemInformationUpdateResponse_sequence);

  return offset;
}


static const per_sequence_t SystemInformationUpdateFailure_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SystemInformationUpdateFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"SystemInformationUpdateFailure ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;


  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SystemInformationUpdateFailure, SystemInformationUpdateFailure_sequence);

  return offset;
}


static const per_sequence_t RadioLinkSetupRequestFDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkSetupRequestFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {

    col_set_str(actx->pinfo->cinfo, COL_INFO,"RadioLinkSetupRequest(FDD) ");
    /* CRNC -> Node B */
    actx->pinfo->link_dir=P2P_DIR_DL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkSetupRequestFDD, RadioLinkSetupRequestFDD_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_RL_Setup_ItemIEs_sequence[] = {
  { &hf_nbap_hSPDSCH_RL_ID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_c_ID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_C_ID },
  { &hf_nbap_hS_DSCH_FDD_Secondary_Serving_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCH_FDD_Secondary_Serving_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_HS_Cell_Information_RL_Setup_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_HS_Cell_Information_RL_Setup_ItemIEs, Additional_HS_Cell_Information_RL_Setup_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_RL_Setup_List_sequence_of[1] = {
  { &hf_nbap_Additional_HS_Cell_Information_RL_Setup_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_HS_Cell_Information_RL_Setup_ItemIEs },
};

static int
dissect_nbap_Additional_HS_Cell_Information_RL_Setup_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Additional_HS_Cell_Information_RL_Setup_List, Additional_HS_Cell_Information_RL_Setup_List_sequence_of,
                                                  1, maxNrOfHSDSCH_1, FALSE);

  return offset;
}


static const per_sequence_t UL_DPCH_Information_RL_SetupRqstFDD_sequence[] = {
  { &hf_nbap_ul_ScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_ScramblingCode },
  { &hf_nbap_minUL_ChannelisationCodeLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MinUL_ChannelisationCodeLength },
  { &hf_nbap_maxNrOfUL_DPDCHs, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MaxNrOfUL_DPDCHs },
  { &hf_nbap_ul_PunctureLimit, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PunctureLimit },
  { &hf_nbap_tFCS           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCS },
  { &hf_nbap_ul_DPCCH_SlotFormat, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_DPCCH_SlotFormat },
  { &hf_nbap_ul_SIR_Target  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_SIR },
  { &hf_nbap_diversityMode  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DiversityMode },
  { &hf_nbap_not_Used_sSDT_CellID_Length, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_not_Used_s_FieldLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_DPCH_Information_RL_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_DPCH_Information_RL_SetupRqstFDD, UL_DPCH_Information_RL_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t PowerOffsetInformation_RL_SetupRqstFDD_sequence[] = {
  { &hf_nbap_pO1_ForTFCI_Bits, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PowerOffset },
  { &hf_nbap_pO2_ForTPC_Bits, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PowerOffset },
  { &hf_nbap_pO3_ForPilotBits, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PowerOffset },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PowerOffsetInformation_RL_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PowerOffsetInformation_RL_SetupRqstFDD, PowerOffsetInformation_RL_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t DL_DPCH_Information_RL_SetupRqstFDD_sequence[] = {
  { &hf_nbap_tFCS           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCS },
  { &hf_nbap_dl_DPCH_SlotFormat, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_DPCH_SlotFormat },
  { &hf_nbap_tFCI_SignallingMode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCI_SignallingMode },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCI_Presence },
  { &hf_nbap_multiplexingPosition, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MultiplexingPosition },
  { &hf_nbap_not_Used_pDSCH_RL_ID, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_not_Used_pDSCH_CodeMapping, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_powerOffsetInformation_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PowerOffsetInformation_RL_SetupRqstFDD },
  { &hf_nbap_fdd_TPC_DownlinkStepSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_TPC_DownlinkStepSize },
  { &hf_nbap_limitedPowerIncrease, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_LimitedPowerIncrease },
  { &hf_nbap_innerLoopDLPCStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_InnerLoopDLPCStatus },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_DPCH_Information_RL_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_DPCH_Information_RL_SetupRqstFDD, DL_DPCH_Information_RL_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t RL_InformationList_RL_SetupRqstFDD_sequence_of[1] = {
  { &hf_nbap_RL_InformationList_RL_SetupRqstFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_RL_InformationList_RL_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RL_InformationList_RL_SetupRqstFDD, RL_InformationList_RL_SetupRqstFDD_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RL_InformationItem_RL_SetupRqstFDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_c_ID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_C_ID },
  { &hf_nbap_firstRLS_indicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FirstRLS_Indicator },
  { &hf_nbap_frameOffset    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FrameOffset },
  { &hf_nbap_chipOffset     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ChipOffset },
  { &hf_nbap_propagationDelay, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PropagationDelay },
  { &hf_nbap_diversityControlField, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DiversityControlField },
  { &hf_nbap_dl_CodeInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_DL_CodeInformation },
  { &hf_nbap_initialDL_transmissionPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_maximumDL_power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_minimumDL_power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_not_Used_sSDT_Cell_Identity, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_transmitDiversityIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransmitDiversityIndicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_InformationItem_RL_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_InformationItem_RL_SetupRqstFDD, RL_InformationItem_RL_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t E_DPCH_Information_RL_SetupRqstFDD_sequence[] = {
  { &hf_nbap_maxSet_E_DPDCHs, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Max_Set_E_DPDCHs },
  { &hf_nbap_ul_PunctureLimit, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PunctureLimit },
  { &hf_nbap_e_TFCS_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_TFCS_Information },
  { &hf_nbap_e_TTI          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_TTI },
  { &hf_nbap_e_DPCCH_PO     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DPCCH_PO },
  { &hf_nbap_e_RGCH_2_IndexStepThreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_RGCH_2_IndexStepThreshold },
  { &hf_nbap_e_RGCH_3_IndexStepThreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_RGCH_3_IndexStepThreshold },
  { &hf_nbap_hARQ_Info_for_E_DCH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HARQ_Info_for_E_DCH },
  { &hf_nbap_hSDSCH_Configured_Indicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_Configured_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DPCH_Information_RL_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DPCH_Information_RL_SetupRqstFDD, E_DPCH_Information_RL_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t PowerOffsetInformation_F_DPCH_RL_SetupRqstFDD_sequence[] = {
  { &hf_nbap_pO2_ForTPC_Bits, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PowerOffset },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PowerOffsetInformation_F_DPCH_RL_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PowerOffsetInformation_F_DPCH_RL_SetupRqstFDD, PowerOffsetInformation_F_DPCH_RL_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t F_DPCH_Information_RL_SetupRqstFDD_sequence[] = {
  { &hf_nbap_powerOffsetInformation_02, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PowerOffsetInformation_F_DPCH_RL_SetupRqstFDD },
  { &hf_nbap_fdd_TPC_DownlinkStepSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_TPC_DownlinkStepSize },
  { &hf_nbap_limitedPowerIncrease, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_LimitedPowerIncrease },
  { &hf_nbap_innerLoopDLPCStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_InnerLoopDLPCStatus },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_F_DPCH_Information_RL_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_F_DPCH_Information_RL_SetupRqstFDD, F_DPCH_Information_RL_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkSetupRequestTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkSetupRequestTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkSetupRequestTDD, RadioLinkSetupRequestTDD_sequence);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationList_RL_SetupRqstTDD_sequence_of[1] = {
  { &hf_nbap_UL_CCTrCH_InformationList_RL_SetupRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_UL_CCTrCH_InformationList_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_CCTrCH_InformationList_RL_SetupRqstTDD, UL_CCTrCH_InformationList_RL_SetupRqstTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}



static int
dissect_nbap_UL_DPCH_Information_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationItem_RL_SetupRqstTDD_sequence[] = {
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_tFCS           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCS },
  { &hf_nbap_tFCI_Coding    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCI_Coding },
  { &hf_nbap_punctureLimit  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PunctureLimit },
  { &hf_nbap_uL_DPCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_DPCH_Information_RL_SetupRqstTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_CCTrCH_InformationItem_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_CCTrCH_InformationItem_RL_SetupRqstTDD, UL_CCTrCH_InformationItem_RL_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t UL_DPCH_InformationItem_RL_SetupRqstTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_uL_Timeslot_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Timeslot_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_DPCH_InformationItem_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_DPCH_InformationItem_RL_SetupRqstTDD, UL_DPCH_InformationItem_RL_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t UL_DPCH_LCR_Information_RL_SetupRqstTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_uL_TimeslotLCR_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_TimeslotLCR_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_DPCH_LCR_Information_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_DPCH_LCR_Information_RL_SetupRqstTDD, UL_DPCH_LCR_Information_RL_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t UL_DPCH_768_Information_RL_SetupRqstTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_uL_Timeslot768_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Timeslot768_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_DPCH_768_Information_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_DPCH_768_Information_RL_SetupRqstTDD, UL_DPCH_768_Information_RL_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationList_RL_SetupRqstTDD_sequence_of[1] = {
  { &hf_nbap_DL_CCTrCH_InformationList_RL_SetupRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_DL_CCTrCH_InformationList_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_CCTrCH_InformationList_RL_SetupRqstTDD, DL_CCTrCH_InformationList_RL_SetupRqstTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t CCTrCH_TPCItem_RL_SetupRqstTDD_sequence[] = {
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CCTrCH_TPCItem_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CCTrCH_TPCItem_RL_SetupRqstTDD, CCTrCH_TPCItem_RL_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t CCTrCH_TPCList_RL_SetupRqstTDD_sequence_of[1] = {
  { &hf_nbap_CCTrCH_TPCList_RL_SetupRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_TPCItem_RL_SetupRqstTDD },
};

static int
dissect_nbap_CCTrCH_TPCList_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_CCTrCH_TPCList_RL_SetupRqstTDD, CCTrCH_TPCList_RL_SetupRqstTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}



static int
dissect_nbap_DL_DPCH_Information_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationItem_RL_SetupRqstTDD_sequence[] = {
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_tFCS           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCS },
  { &hf_nbap_tFCI_Coding    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCI_Coding },
  { &hf_nbap_punctureLimit  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PunctureLimit },
  { &hf_nbap_tdd_TPC_DownlinkStepSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_TPC_DownlinkStepSize },
  { &hf_nbap_cCTrCH_TPCList , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CCTrCH_TPCList_RL_SetupRqstTDD },
  { &hf_nbap_dL_DPCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_DPCH_Information_RL_SetupRqstTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_CCTrCH_InformationItem_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_CCTrCH_InformationItem_RL_SetupRqstTDD, DL_CCTrCH_InformationItem_RL_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DL_DPCH_InformationItem_RL_SetupRqstTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_dL_Timeslot_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Timeslot_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_DPCH_InformationItem_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_DPCH_InformationItem_RL_SetupRqstTDD, DL_DPCH_InformationItem_RL_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DL_DPCH_LCR_Information_RL_SetupRqstTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_dL_TimeslotLCR_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_TimeslotLCR_Information },
  { &hf_nbap_tstdIndicator  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TSTD_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_DPCH_LCR_Information_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_DPCH_LCR_Information_RL_SetupRqstTDD, DL_DPCH_LCR_Information_RL_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DL_DPCH_768_Information_RL_SetupRqstTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_dL_Timeslot768_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Timeslot768_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_DPCH_768_Information_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_DPCH_768_Information_RL_SetupRqstTDD, DL_DPCH_768_Information_RL_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t RL_Information_RL_SetupRqstTDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_c_ID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_C_ID },
  { &hf_nbap_frameOffset    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FrameOffset },
  { &hf_nbap_specialBurstScheduling, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SpecialBurstScheduling },
  { &hf_nbap_initialDL_transmissionPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_maximumDL_power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_minimumDL_power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_dL_TimeSlotISCPInfo, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_TimeslotISCPInfo },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_Information_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_Information_RL_SetupRqstTDD, RL_Information_RL_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkSetupResponseFDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkSetupResponseFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {

    col_set_str(actx->pinfo->cinfo, COL_INFO,"RadioLinkSetupResponse(FDD) ");
    /* Node B -> CRNC */
    actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkSetupResponseFDD, RadioLinkSetupResponseFDD_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_Response_ItemIEs_sequence[] = {
  { &hf_nbap_hSPDSCH_RL_ID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_hS_DSCH_FDD_Secondary_Serving_Information_Response, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCH_FDD_Secondary_Serving_Information_Response },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_HS_Cell_Information_Response_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_HS_Cell_Information_Response_ItemIEs, Additional_HS_Cell_Information_Response_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_Response_List_sequence_of[1] = {
  { &hf_nbap_Additional_HS_Cell_Information_Response_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_HS_Cell_Information_Response_ItemIEs },
};

static int
dissect_nbap_Additional_HS_Cell_Information_Response_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Additional_HS_Cell_Information_Response_List, Additional_HS_Cell_Information_Response_List_sequence_of,
                                                  1, maxNrOfHSDSCH_1, FALSE);

  return offset;
}


static const per_sequence_t RL_InformationResponseList_RL_SetupRspFDD_sequence_of[1] = {
  { &hf_nbap_RL_InformationResponseList_RL_SetupRspFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_RL_InformationResponseList_RL_SetupRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RL_InformationResponseList_RL_SetupRspFDD, RL_InformationResponseList_RL_SetupRspFDD_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t Combining_RL_SetupRspFDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Combining_RL_SetupRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Combining_RL_SetupRspFDD, Combining_RL_SetupRspFDD_sequence);

  return offset;
}


static const per_sequence_t NonCombiningOrFirstRL_RL_SetupRspFDD_sequence[] = {
  { &hf_nbap_dCH_InformationResponse, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_InformationResponse },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_NonCombiningOrFirstRL_RL_SetupRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_NonCombiningOrFirstRL_RL_SetupRspFDD, NonCombiningOrFirstRL_RL_SetupRspFDD_sequence);

  return offset;
}


static const value_string nbap_DiversityIndication_RL_SetupRspFDD_vals[] = {
  {   0, "combining" },
  {   1, "nonCombiningOrFirstRL" },
  { 0, NULL }
};

static const per_choice_t DiversityIndication_RL_SetupRspFDD_choice[] = {
  {   0, &hf_nbap_combining      , ASN1_NO_EXTENSIONS     , dissect_nbap_Combining_RL_SetupRspFDD },
  {   1, &hf_nbap_nonCombiningOrFirstRL, ASN1_NO_EXTENSIONS     , dissect_nbap_NonCombiningOrFirstRL_RL_SetupRspFDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_DiversityIndication_RL_SetupRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_DiversityIndication_RL_SetupRspFDD, DiversityIndication_RL_SetupRspFDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t RL_InformationResponseItem_RL_SetupRspFDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_rL_Set_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_Set_ID },
  { &hf_nbap_received_total_wide_band_power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Received_total_wide_band_power_Value },
  { &hf_nbap_diversityIndication, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DiversityIndication_RL_SetupRspFDD },
  { &hf_nbap_not_Used_dSCH_InformationResponseList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_sSDT_SupportIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SSDT_SupportIndicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_InformationResponseItem_RL_SetupRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_InformationResponseItem_RL_SetupRspFDD, RL_InformationResponseItem_RL_SetupRspFDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkSetupResponseTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkSetupResponseTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkSetupResponseTDD, RadioLinkSetupResponseTDD_sequence);

  return offset;
}



static int
dissect_nbap_DCH_InformationResponseList_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_DSCH_InformationResponseList_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_USCH_InformationResponseList_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t RL_InformationResponse_RL_SetupRspTDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_uL_TimeSlot_ISCP_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_TimeSlot_ISCP_Info },
  { &hf_nbap_ul_PhysCH_SF_Variation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_PhysCH_SF_Variation },
  { &hf_nbap_dCH_InformationResponseList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DCH_InformationResponseList_RL_SetupRspTDD },
  { &hf_nbap_dSCH_InformationResponseList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DSCH_InformationResponseList_RL_SetupRspTDD },
  { &hf_nbap_uSCH_InformationResponseList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_USCH_InformationResponseList_RL_SetupRspTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_InformationResponse_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_InformationResponse_RL_SetupRspTDD, RL_InformationResponse_RL_SetupRspTDD_sequence);

  return offset;
}


static const per_sequence_t RL_InformationResponse_LCR_RL_SetupRspTDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_uL_TimeSlot_ISCP_LCR_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_TimeSlot_ISCP_LCR_Info },
  { &hf_nbap_ul_PhysCH_SF_Variation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_PhysCH_SF_Variation },
  { &hf_nbap_dCH_InformationResponseList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DCH_InformationResponseList_RL_SetupRspTDD },
  { &hf_nbap_dSCH_InformationResponseList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DSCH_InformationResponseList_RL_SetupRspTDD },
  { &hf_nbap_uSCH_InformationResponseList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_USCH_InformationResponseList_RL_SetupRspTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_InformationResponse_LCR_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_InformationResponse_LCR_RL_SetupRspTDD, RL_InformationResponse_LCR_RL_SetupRspTDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkSetupFailureFDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkSetupFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"RadioLinkSetupFailure(FDD) ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;


  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkSetupFailureFDD, RadioLinkSetupFailureFDD_sequence);

  return offset;
}


static const per_sequence_t GeneralCauseList_RL_SetupFailureFDD_sequence[] = {
  { &hf_nbap_cause          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Cause },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GeneralCauseList_RL_SetupFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GeneralCauseList_RL_SetupFailureFDD, GeneralCauseList_RL_SetupFailureFDD_sequence);

  return offset;
}


static const per_sequence_t Unsuccessful_RL_InformationRespList_RL_SetupFailureFDD_sequence_of[1] = {
  { &hf_nbap_Unsuccessful_RL_InformationRespList_RL_SetupFailureFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_Unsuccessful_RL_InformationRespList_RL_SetupFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Unsuccessful_RL_InformationRespList_RL_SetupFailureFDD, Unsuccessful_RL_InformationRespList_RL_SetupFailureFDD_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t Successful_RL_InformationRespList_RL_SetupFailureFDD_sequence_of[1] = {
  { &hf_nbap_Successful_RL_InformationRespList_RL_SetupFailureFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_Successful_RL_InformationRespList_RL_SetupFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Successful_RL_InformationRespList_RL_SetupFailureFDD, Successful_RL_InformationRespList_RL_SetupFailureFDD_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RLSpecificCauseList_RL_SetupFailureFDD_sequence[] = {
  { &hf_nbap_unsuccessful_RL_InformationRespList_RL_SetupFailureFDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Unsuccessful_RL_InformationRespList_RL_SetupFailureFDD },
  { &hf_nbap_successful_RL_InformationRespList_RL_SetupFailureFDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Successful_RL_InformationRespList_RL_SetupFailureFDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RLSpecificCauseList_RL_SetupFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RLSpecificCauseList_RL_SetupFailureFDD, RLSpecificCauseList_RL_SetupFailureFDD_sequence);

  return offset;
}


static const value_string nbap_CauseLevel_RL_SetupFailureFDD_vals[] = {
  {   0, "generalCause" },
  {   1, "rLSpecificCause" },
  { 0, NULL }
};

static const per_choice_t CauseLevel_RL_SetupFailureFDD_choice[] = {
  {   0, &hf_nbap_generalCause   , ASN1_EXTENSION_ROOT    , dissect_nbap_GeneralCauseList_RL_SetupFailureFDD },
  {   1, &hf_nbap_rLSpecificCause, ASN1_EXTENSION_ROOT    , dissect_nbap_RLSpecificCauseList_RL_SetupFailureFDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_CauseLevel_RL_SetupFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_CauseLevel_RL_SetupFailureFDD, CauseLevel_RL_SetupFailureFDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Unsuccessful_RL_InformationRespItem_RL_SetupFailureFDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_cause          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Cause },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Unsuccessful_RL_InformationRespItem_RL_SetupFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Unsuccessful_RL_InformationRespItem_RL_SetupFailureFDD, Unsuccessful_RL_InformationRespItem_RL_SetupFailureFDD_sequence);

  return offset;
}


static const per_sequence_t Combining_RL_SetupFailureFDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Combining_RL_SetupFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Combining_RL_SetupFailureFDD, Combining_RL_SetupFailureFDD_sequence);

  return offset;
}


static const per_sequence_t NonCombiningOrFirstRL_RL_SetupFailureFDD_sequence[] = {
  { &hf_nbap_dCH_InformationResponse, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_InformationResponse },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_NonCombiningOrFirstRL_RL_SetupFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_NonCombiningOrFirstRL_RL_SetupFailureFDD, NonCombiningOrFirstRL_RL_SetupFailureFDD_sequence);

  return offset;
}


static const value_string nbap_DiversityIndication_RL_SetupFailureFDD_vals[] = {
  {   0, "combining" },
  {   1, "nonCombiningOrFirstRL" },
  { 0, NULL }
};

static const per_choice_t DiversityIndication_RL_SetupFailureFDD_choice[] = {
  {   0, &hf_nbap_combining_01   , ASN1_NO_EXTENSIONS     , dissect_nbap_Combining_RL_SetupFailureFDD },
  {   1, &hf_nbap_nonCombiningOrFirstRL_01, ASN1_NO_EXTENSIONS     , dissect_nbap_NonCombiningOrFirstRL_RL_SetupFailureFDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_DiversityIndication_RL_SetupFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_DiversityIndication_RL_SetupFailureFDD, DiversityIndication_RL_SetupFailureFDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Successful_RL_InformationRespItem_RL_SetupFailureFDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_rL_Set_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_Set_ID },
  { &hf_nbap_received_total_wide_band_power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Received_total_wide_band_power_Value },
  { &hf_nbap_diversityIndication_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DiversityIndication_RL_SetupFailureFDD },
  { &hf_nbap_not_Used_dSCH_InformationResponseList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_not_Used_tFCI2_BearerInformationResponse, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_sSDT_SupportIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SSDT_SupportIndicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Successful_RL_InformationRespItem_RL_SetupFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Successful_RL_InformationRespItem_RL_SetupFailureFDD, Successful_RL_InformationRespItem_RL_SetupFailureFDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkSetupFailureTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkSetupFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkSetupFailureTDD, RadioLinkSetupFailureTDD_sequence);

  return offset;
}


static const per_sequence_t GeneralCauseList_RL_SetupFailureTDD_sequence[] = {
  { &hf_nbap_cause          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Cause },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GeneralCauseList_RL_SetupFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GeneralCauseList_RL_SetupFailureTDD, GeneralCauseList_RL_SetupFailureTDD_sequence);

  return offset;
}



static int
dissect_nbap_Unsuccessful_RL_InformationRespItem_RL_SetupFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t RLSpecificCauseList_RL_SetupFailureTDD_sequence[] = {
  { &hf_nbap_unsuccessful_RL_InformationRespItem_RL_SetupFailureTDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Unsuccessful_RL_InformationRespItem_RL_SetupFailureTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RLSpecificCauseList_RL_SetupFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RLSpecificCauseList_RL_SetupFailureTDD, RLSpecificCauseList_RL_SetupFailureTDD_sequence);

  return offset;
}


static const value_string nbap_CauseLevel_RL_SetupFailureTDD_vals[] = {
  {   0, "generalCause" },
  {   1, "rLSpecificCause" },
  { 0, NULL }
};

static const per_choice_t CauseLevel_RL_SetupFailureTDD_choice[] = {
  {   0, &hf_nbap_generalCause_01, ASN1_EXTENSION_ROOT    , dissect_nbap_GeneralCauseList_RL_SetupFailureTDD },
  {   1, &hf_nbap_rLSpecificCause_01, ASN1_EXTENSION_ROOT    , dissect_nbap_RLSpecificCauseList_RL_SetupFailureTDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_CauseLevel_RL_SetupFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_CauseLevel_RL_SetupFailureTDD, CauseLevel_RL_SetupFailureTDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Unsuccessful_RL_InformationResp_RL_SetupFailureTDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_cause          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Cause },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Unsuccessful_RL_InformationResp_RL_SetupFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Unsuccessful_RL_InformationResp_RL_SetupFailureTDD, Unsuccessful_RL_InformationResp_RL_SetupFailureTDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkAdditionRequestFDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkAdditionRequestFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"RadioLinkAdditionRequest(FDD) ");


  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkAdditionRequestFDD, RadioLinkAdditionRequestFDD_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_RL_Addition_ItemIEs_sequence[] = {
  { &hf_nbap_hSPDSCH_RL_ID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_c_ID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_C_ID },
  { &hf_nbap_hS_DSCH_FDD_Secondary_Serving_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCH_FDD_Secondary_Serving_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_HS_Cell_Information_RL_Addition_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_HS_Cell_Information_RL_Addition_ItemIEs, Additional_HS_Cell_Information_RL_Addition_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_RL_Addition_List_sequence_of[1] = {
  { &hf_nbap_Additional_HS_Cell_Information_RL_Addition_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_HS_Cell_Information_RL_Addition_ItemIEs },
};

static int
dissect_nbap_Additional_HS_Cell_Information_RL_Addition_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Additional_HS_Cell_Information_RL_Addition_List, Additional_HS_Cell_Information_RL_Addition_List_sequence_of,
                                                  1, maxNrOfHSDSCH_1, FALSE);

  return offset;
}


static const value_string nbap_Setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency_vals[] = {
  {   0, "setup" },
  {   1, "addition" },
  { 0, NULL }
};

static const per_choice_t Setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency_choice[] = {
  {   0, &hf_nbap_setup          , ASN1_EXTENSION_ROOT    , dissect_nbap_Additional_EDCH_Setup_Info },
  {   1, &hf_nbap_addition       , ASN1_EXTENSION_ROOT    , dissect_nbap_Additional_EDCH_Cell_Information_To_Add_List },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_Setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_Setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency, Setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_RL_Add_Req_sequence[] = {
  { &hf_nbap_setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_Cell_Information_RL_Add_Req(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_Cell_Information_RL_Add_Req, Additional_EDCH_Cell_Information_RL_Add_Req_sequence);

  return offset;
}


static const per_sequence_t RL_InformationList_RL_AdditionRqstFDD_sequence_of[1] = {
  { &hf_nbap_RL_InformationList_RL_AdditionRqstFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_RL_InformationList_RL_AdditionRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RL_InformationList_RL_AdditionRqstFDD, RL_InformationList_RL_AdditionRqstFDD_sequence_of,
                                                  1, maxNrOfRLs_1, FALSE);

  return offset;
}


static const per_sequence_t RL_InformationItem_RL_AdditionRqstFDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_c_ID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_C_ID },
  { &hf_nbap_frameOffset    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FrameOffset },
  { &hf_nbap_chipOffset     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ChipOffset },
  { &hf_nbap_diversityControlField, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DiversityControlField },
  { &hf_nbap_dl_CodeInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_DL_CodeInformation },
  { &hf_nbap_initialDL_TransmissionPower, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_maximumDL_Power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_minimumDL_Power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_not_Used_sSDT_CellIdentity, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_transmitDiversityIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransmitDiversityIndicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_InformationItem_RL_AdditionRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_InformationItem_RL_AdditionRqstFDD, RL_InformationItem_RL_AdditionRqstFDD_sequence);

  return offset;
}


static const per_sequence_t E_DPCH_Information_RL_AdditionReqFDD_sequence[] = {
  { &hf_nbap_maxSet_E_DPDCHs, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Max_Set_E_DPDCHs },
  { &hf_nbap_ul_PunctureLimit, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PunctureLimit },
  { &hf_nbap_e_TFCS_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_TFCS_Information },
  { &hf_nbap_e_TTI          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_TTI },
  { &hf_nbap_e_DPCCH_PO     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DPCCH_PO },
  { &hf_nbap_e_RGCH_2_IndexStepThreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_RGCH_2_IndexStepThreshold },
  { &hf_nbap_e_RGCH_3_IndexStepThreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_RGCH_3_IndexStepThreshold },
  { &hf_nbap_hARQ_Info_for_E_DCH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HARQ_Info_for_E_DCH },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DPCH_Information_RL_AdditionReqFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DPCH_Information_RL_AdditionReqFDD, E_DPCH_Information_RL_AdditionReqFDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkAdditionRequestTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkAdditionRequestTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkAdditionRequestTDD, RadioLinkAdditionRequestTDD_sequence);

  return offset;
}



static int
dissect_nbap_UL_DPCH_InformationList_RL_AdditionRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationItem_RL_AdditionRqstTDD_sequence[] = {
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_uL_DPCH_Information_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_DPCH_InformationList_RL_AdditionRqstTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_CCTrCH_InformationItem_RL_AdditionRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_CCTrCH_InformationItem_RL_AdditionRqstTDD, UL_CCTrCH_InformationItem_RL_AdditionRqstTDD_sequence);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationList_RL_AdditionRqstTDD_sequence_of[1] = {
  { &hf_nbap_UL_CCTrCH_InformationList_RL_AdditionRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_CCTrCH_InformationItem_RL_AdditionRqstTDD },
};

static int
dissect_nbap_UL_CCTrCH_InformationList_RL_AdditionRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_CCTrCH_InformationList_RL_AdditionRqstTDD, UL_CCTrCH_InformationList_RL_AdditionRqstTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_DPCH_InformationItem_RL_AdditionRqstTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_uL_Timeslot_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Timeslot_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_DPCH_InformationItem_RL_AdditionRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_DPCH_InformationItem_RL_AdditionRqstTDD, UL_DPCH_InformationItem_RL_AdditionRqstTDD_sequence);

  return offset;
}


static const per_sequence_t UL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_uL_TimeslotLCR_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_TimeslotLCR_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD, UL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD_sequence);

  return offset;
}


static const per_sequence_t UL_DPCH_InformationItem_768_RL_AdditionRqstTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_uL_Timeslot768_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Timeslot768_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_DPCH_InformationItem_768_RL_AdditionRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_DPCH_InformationItem_768_RL_AdditionRqstTDD, UL_DPCH_InformationItem_768_RL_AdditionRqstTDD_sequence);

  return offset;
}



static int
dissect_nbap_DL_DPCH_InformationList_RL_AdditionRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationItem_RL_AdditionRqstTDD_sequence[] = {
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_dL_DPCH_Information_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_DPCH_InformationList_RL_AdditionRqstTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_CCTrCH_InformationItem_RL_AdditionRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_CCTrCH_InformationItem_RL_AdditionRqstTDD, DL_CCTrCH_InformationItem_RL_AdditionRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationList_RL_AdditionRqstTDD_sequence_of[1] = {
  { &hf_nbap_DL_CCTrCH_InformationList_RL_AdditionRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_CCTrCH_InformationItem_RL_AdditionRqstTDD },
};

static int
dissect_nbap_DL_CCTrCH_InformationList_RL_AdditionRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_CCTrCH_InformationList_RL_AdditionRqstTDD, DL_CCTrCH_InformationList_RL_AdditionRqstTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_DPCH_InformationItem_RL_AdditionRqstTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_dL_Timeslot_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Timeslot_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_DPCH_InformationItem_RL_AdditionRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_DPCH_InformationItem_RL_AdditionRqstTDD, DL_DPCH_InformationItem_RL_AdditionRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_dL_TimeslotLCR_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_TimeslotLCR_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD, DL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DL_DPCH_InformationItem_768_RL_AdditionRqstTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_dL_Timeslot768_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Timeslot768_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_DPCH_InformationItem_768_RL_AdditionRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_DPCH_InformationItem_768_RL_AdditionRqstTDD, DL_DPCH_InformationItem_768_RL_AdditionRqstTDD_sequence);

  return offset;
}


static const per_sequence_t RL_Information_RL_AdditionRqstTDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_c_ID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_C_ID },
  { &hf_nbap_frameOffset    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FrameOffset },
  { &hf_nbap_diversityControlField, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DiversityControlField },
  { &hf_nbap_initial_DL_Transmission_Power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_maximumDL_Power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_minimumDL_Power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_dL_TimeSlotISCPInfo, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_TimeslotISCPInfo },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_Information_RL_AdditionRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_Information_RL_AdditionRqstTDD, RL_Information_RL_AdditionRqstTDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkAdditionResponseFDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkAdditionResponseFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"RadioLinkAdditionResponse(FDD) ");


  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkAdditionResponseFDD, RadioLinkAdditionResponseFDD_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Change_Information_Response_ItemIEs_sequence[] = {
  { &hf_nbap_hSPDSCH_RL_ID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_hS_DSCH_Secondary_Serving_Cell_Change_Information_Response, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCH_Secondary_Serving_Cell_Change_Information_Response },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_HS_Cell_Change_Information_Response_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_HS_Cell_Change_Information_Response_ItemIEs, Additional_HS_Cell_Change_Information_Response_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Change_Information_Response_List_sequence_of[1] = {
  { &hf_nbap_Additional_HS_Cell_Change_Information_Response_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_HS_Cell_Change_Information_Response_ItemIEs },
};

static int
dissect_nbap_Additional_HS_Cell_Change_Information_Response_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Additional_HS_Cell_Change_Information_Response_List, Additional_HS_Cell_Change_Information_Response_List_sequence_of,
                                                  1, maxNrOfHSDSCH_1, FALSE);

  return offset;
}


static const per_sequence_t RL_InformationResponseList_RL_AdditionRspFDD_sequence_of[1] = {
  { &hf_nbap_RL_InformationResponseList_RL_AdditionRspFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_RL_InformationResponseList_RL_AdditionRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RL_InformationResponseList_RL_AdditionRspFDD, RL_InformationResponseList_RL_AdditionRspFDD_sequence_of,
                                                  1, maxNrOfRLs_1, FALSE);

  return offset;
}


static const per_sequence_t Combining_RL_AdditionRspFDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Combining_RL_AdditionRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Combining_RL_AdditionRspFDD, Combining_RL_AdditionRspFDD_sequence);

  return offset;
}


static const per_sequence_t Non_Combining_RL_AdditionRspFDD_sequence[] = {
  { &hf_nbap_dCH_InformationResponse, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_InformationResponse },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Non_Combining_RL_AdditionRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Non_Combining_RL_AdditionRspFDD, Non_Combining_RL_AdditionRspFDD_sequence);

  return offset;
}


static const value_string nbap_DiversityIndication_RL_AdditionRspFDD_vals[] = {
  {   0, "combining" },
  {   1, "non-combining" },
  { 0, NULL }
};

static const per_choice_t DiversityIndication_RL_AdditionRspFDD_choice[] = {
  {   0, &hf_nbap_combining_02   , ASN1_NO_EXTENSIONS     , dissect_nbap_Combining_RL_AdditionRspFDD },
  {   1, &hf_nbap_non_combining  , ASN1_NO_EXTENSIONS     , dissect_nbap_Non_Combining_RL_AdditionRspFDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_DiversityIndication_RL_AdditionRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_DiversityIndication_RL_AdditionRspFDD, DiversityIndication_RL_AdditionRspFDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t RL_InformationResponseItem_RL_AdditionRspFDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_rL_Set_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_Set_ID },
  { &hf_nbap_received_total_wide_band_power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Received_total_wide_band_power_Value },
  { &hf_nbap_diversityIndication_02, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DiversityIndication_RL_AdditionRspFDD },
  { &hf_nbap_sSDT_SupportIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SSDT_SupportIndicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_InformationResponseItem_RL_AdditionRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_InformationResponseItem_RL_AdditionRspFDD, RL_InformationResponseItem_RL_AdditionRspFDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkAdditionResponseTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkAdditionResponseTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkAdditionResponseTDD, RadioLinkAdditionResponseTDD_sequence);

  return offset;
}


static const per_sequence_t Combining_RL_AdditionRspTDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Combining_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Combining_RL_AdditionRspTDD, Combining_RL_AdditionRspTDD_sequence);

  return offset;
}


static const per_sequence_t Non_Combining_RL_AdditionRspTDD_sequence[] = {
  { &hf_nbap_dCH_InformationResponse, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_InformationResponse },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Non_Combining_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Non_Combining_RL_AdditionRspTDD, Non_Combining_RL_AdditionRspTDD_sequence);

  return offset;
}


static const value_string nbap_DiversityIndication_RL_AdditionRspTDD_vals[] = {
  {   0, "combining" },
  {   1, "non-Combining" },
  { 0, NULL }
};

static const per_choice_t DiversityIndication_RL_AdditionRspTDD_choice[] = {
  {   0, &hf_nbap_combining_03   , ASN1_NO_EXTENSIONS     , dissect_nbap_Combining_RL_AdditionRspTDD },
  {   1, &hf_nbap_non_Combining  , ASN1_NO_EXTENSIONS     , dissect_nbap_Non_Combining_RL_AdditionRspTDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_DiversityIndication_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_DiversityIndication_RL_AdditionRspTDD, DiversityIndication_RL_AdditionRspTDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t DCH_Information_RL_AdditionRspTDD_sequence[] = {
  { &hf_nbap_diversityIndication_03, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DiversityIndication_RL_AdditionRspTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DCH_Information_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DCH_Information_RL_AdditionRspTDD, DCH_Information_RL_AdditionRspTDD_sequence);

  return offset;
}



static int
dissect_nbap_DSCH_InformationResponseList_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_USCH_InformationResponseList_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t RL_InformationResponse_RL_AdditionRspTDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_uL_TimeSlot_ISCP_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_TimeSlot_ISCP_Info },
  { &hf_nbap_ul_PhysCH_SF_Variation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_PhysCH_SF_Variation },
  { &hf_nbap_dCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DCH_Information_RL_AdditionRspTDD },
  { &hf_nbap_dSCH_InformationResponseList_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DSCH_InformationResponseList_RL_AdditionRspTDD },
  { &hf_nbap_uSCH_InformationResponseList_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_USCH_InformationResponseList_RL_AdditionRspTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_InformationResponse_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_InformationResponse_RL_AdditionRspTDD, RL_InformationResponse_RL_AdditionRspTDD_sequence);

  return offset;
}


static const per_sequence_t RL_InformationResponse_LCR_RL_AdditionRspTDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_uL_TimeSlot_ISCP_InfoLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_TimeSlot_ISCP_LCR_Info },
  { &hf_nbap_ul_PhysCH_SF_Variation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_PhysCH_SF_Variation },
  { &hf_nbap_dCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DCH_Information_RL_AdditionRspTDD },
  { &hf_nbap_dSCH_InformationResponseList_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DSCH_InformationResponseList_RL_AdditionRspTDD },
  { &hf_nbap_uSCH_InformationResponseList_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_USCH_InformationResponseList_RL_AdditionRspTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_InformationResponse_LCR_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_InformationResponse_LCR_RL_AdditionRspTDD, RL_InformationResponse_LCR_RL_AdditionRspTDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkAdditionFailureFDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkAdditionFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"RadioLinkAdditionRequest(FDD) ");


  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkAdditionFailureFDD, RadioLinkAdditionFailureFDD_sequence);

  return offset;
}


static const per_sequence_t GeneralCauseList_RL_AdditionFailureFDD_sequence[] = {
  { &hf_nbap_cause          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Cause },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GeneralCauseList_RL_AdditionFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GeneralCauseList_RL_AdditionFailureFDD, GeneralCauseList_RL_AdditionFailureFDD_sequence);

  return offset;
}


static const per_sequence_t Unsuccessful_RL_InformationRespList_RL_AdditionFailureFDD_sequence_of[1] = {
  { &hf_nbap_Unsuccessful_RL_InformationRespList_RL_AdditionFailureFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_Unsuccessful_RL_InformationRespList_RL_AdditionFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Unsuccessful_RL_InformationRespList_RL_AdditionFailureFDD, Unsuccessful_RL_InformationRespList_RL_AdditionFailureFDD_sequence_of,
                                                  1, maxNrOfRLs_1, FALSE);

  return offset;
}


static const per_sequence_t Successful_RL_InformationRespList_RL_AdditionFailureFDD_sequence_of[1] = {
  { &hf_nbap_Successful_RL_InformationRespList_RL_AdditionFailureFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_Successful_RL_InformationRespList_RL_AdditionFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Successful_RL_InformationRespList_RL_AdditionFailureFDD, Successful_RL_InformationRespList_RL_AdditionFailureFDD_sequence_of,
                                                  1, maxNrOfRLs_2, FALSE);

  return offset;
}


static const per_sequence_t RLSpecificCauseList_RL_AdditionFailureFDD_sequence[] = {
  { &hf_nbap_unsuccessful_RL_InformationRespList_RL_AdditionFailureFDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Unsuccessful_RL_InformationRespList_RL_AdditionFailureFDD },
  { &hf_nbap_successful_RL_InformationRespList_RL_AdditionFailureFDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Successful_RL_InformationRespList_RL_AdditionFailureFDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RLSpecificCauseList_RL_AdditionFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RLSpecificCauseList_RL_AdditionFailureFDD, RLSpecificCauseList_RL_AdditionFailureFDD_sequence);

  return offset;
}


static const value_string nbap_CauseLevel_RL_AdditionFailureFDD_vals[] = {
  {   0, "generalCause" },
  {   1, "rLSpecificCause" },
  { 0, NULL }
};

static const per_choice_t CauseLevel_RL_AdditionFailureFDD_choice[] = {
  {   0, &hf_nbap_generalCause_02, ASN1_EXTENSION_ROOT    , dissect_nbap_GeneralCauseList_RL_AdditionFailureFDD },
  {   1, &hf_nbap_rLSpecificCause_02, ASN1_EXTENSION_ROOT    , dissect_nbap_RLSpecificCauseList_RL_AdditionFailureFDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_CauseLevel_RL_AdditionFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_CauseLevel_RL_AdditionFailureFDD, CauseLevel_RL_AdditionFailureFDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Unsuccessful_RL_InformationRespItem_RL_AdditionFailureFDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_cause          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Cause },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Unsuccessful_RL_InformationRespItem_RL_AdditionFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Unsuccessful_RL_InformationRespItem_RL_AdditionFailureFDD, Unsuccessful_RL_InformationRespItem_RL_AdditionFailureFDD_sequence);

  return offset;
}


static const per_sequence_t Combining_RL_AdditionFailureFDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Combining_RL_AdditionFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Combining_RL_AdditionFailureFDD, Combining_RL_AdditionFailureFDD_sequence);

  return offset;
}


static const per_sequence_t Non_Combining_RL_AdditionFailureFDD_sequence[] = {
  { &hf_nbap_dCH_InformationResponse, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_InformationResponse },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Non_Combining_RL_AdditionFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Non_Combining_RL_AdditionFailureFDD, Non_Combining_RL_AdditionFailureFDD_sequence);

  return offset;
}


static const value_string nbap_DiversityIndication_RL_AdditionFailureFDD_vals[] = {
  {   0, "combining" },
  {   1, "non-Combining" },
  { 0, NULL }
};

static const per_choice_t DiversityIndication_RL_AdditionFailureFDD_choice[] = {
  {   0, &hf_nbap_combining_04   , ASN1_NO_EXTENSIONS     , dissect_nbap_Combining_RL_AdditionFailureFDD },
  {   1, &hf_nbap_non_Combining_01, ASN1_NO_EXTENSIONS     , dissect_nbap_Non_Combining_RL_AdditionFailureFDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_DiversityIndication_RL_AdditionFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_DiversityIndication_RL_AdditionFailureFDD, DiversityIndication_RL_AdditionFailureFDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Successful_RL_InformationRespItem_RL_AdditionFailureFDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_rL_Set_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_Set_ID },
  { &hf_nbap_received_total_wide_band_power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Received_total_wide_band_power_Value },
  { &hf_nbap_diversityIndication_04, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DiversityIndication_RL_AdditionFailureFDD },
  { &hf_nbap_sSDT_SupportIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SSDT_SupportIndicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Successful_RL_InformationRespItem_RL_AdditionFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Successful_RL_InformationRespItem_RL_AdditionFailureFDD, Successful_RL_InformationRespItem_RL_AdditionFailureFDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkAdditionFailureTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkAdditionFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkAdditionFailureTDD, RadioLinkAdditionFailureTDD_sequence);

  return offset;
}


static const per_sequence_t GeneralCauseList_RL_AdditionFailureTDD_sequence[] = {
  { &hf_nbap_cause          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Cause },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GeneralCauseList_RL_AdditionFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GeneralCauseList_RL_AdditionFailureTDD, GeneralCauseList_RL_AdditionFailureTDD_sequence);

  return offset;
}



static int
dissect_nbap_Unsuccessful_RL_InformationRespItem_RL_AdditionFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t RLSpecificCauseList_RL_AdditionFailureTDD_sequence[] = {
  { &hf_nbap_unsuccessful_RL_InformationRespItem_RL_AdditionFailureTDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Unsuccessful_RL_InformationRespItem_RL_AdditionFailureTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RLSpecificCauseList_RL_AdditionFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RLSpecificCauseList_RL_AdditionFailureTDD, RLSpecificCauseList_RL_AdditionFailureTDD_sequence);

  return offset;
}


static const value_string nbap_CauseLevel_RL_AdditionFailureTDD_vals[] = {
  {   0, "generalCause" },
  {   1, "rLSpecificCause" },
  { 0, NULL }
};

static const per_choice_t CauseLevel_RL_AdditionFailureTDD_choice[] = {
  {   0, &hf_nbap_generalCause_03, ASN1_EXTENSION_ROOT    , dissect_nbap_GeneralCauseList_RL_AdditionFailureTDD },
  {   1, &hf_nbap_rLSpecificCause_03, ASN1_EXTENSION_ROOT    , dissect_nbap_RLSpecificCauseList_RL_AdditionFailureTDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_CauseLevel_RL_AdditionFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_CauseLevel_RL_AdditionFailureTDD, CauseLevel_RL_AdditionFailureTDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Unsuccessful_RL_InformationResp_RL_AdditionFailureTDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_cause          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Cause },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Unsuccessful_RL_InformationResp_RL_AdditionFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Unsuccessful_RL_InformationResp_RL_AdditionFailureTDD, Unsuccessful_RL_InformationResp_RL_AdditionFailureTDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkReconfigurationPrepareFDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkReconfigurationPrepareFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
    col_set_str(actx->pinfo->cinfo, COL_INFO,"RadioLinkReconfigurationPrepare(FDD) ");
    /* CRNC -> Node B */
    actx->pinfo->link_dir=P2P_DIR_DL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkReconfigurationPrepareFDD, RadioLinkReconfigurationPrepareFDD_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_RL_Reconf_Prep_ItemIEs_sequence[] = {
  { &hf_nbap_hSPDSCH_RL_ID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_c_ID           , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_C_ID },
  { &hf_nbap_hS_DSCH_FDD_Secondary_Serving_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_DSCH_FDD_Secondary_Serving_Information },
  { &hf_nbap_hS_DSCH_Secondary_Serving_Information_To_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_DSCH_Secondary_Serving_Information_To_Modify },
  { &hf_nbap_hS_HS_DSCH_Secondary_Serving_Remove, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_DSCH_Secondary_Serving_Remove },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_HS_Cell_Information_RL_Reconf_Prep_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_HS_Cell_Information_RL_Reconf_Prep_ItemIEs, Additional_HS_Cell_Information_RL_Reconf_Prep_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_RL_Reconf_Prep_sequence_of[1] = {
  { &hf_nbap_Additional_HS_Cell_Information_RL_Reconf_Prep_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_HS_Cell_Information_RL_Reconf_Prep_ItemIEs },
};

static int
dissect_nbap_Additional_HS_Cell_Information_RL_Reconf_Prep(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Additional_HS_Cell_Information_RL_Reconf_Prep, Additional_HS_Cell_Information_RL_Reconf_Prep_sequence_of,
                                                  1, maxNrOfHSDSCH_1, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_RL_Reconf_Prep_sequence[] = {
  { &hf_nbap_setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_Cell_Information_RL_Reconf_Prep(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_Cell_Information_RL_Reconf_Prep, Additional_EDCH_Cell_Information_RL_Reconf_Prep_sequence);

  return offset;
}


static const per_sequence_t UL_DPCH_Information_RL_ReconfPrepFDD_sequence[] = {
  { &hf_nbap_ul_ScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_ScramblingCode },
  { &hf_nbap_ul_SIR_Target  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_SIR },
  { &hf_nbap_minUL_ChannelisationCodeLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MinUL_ChannelisationCodeLength },
  { &hf_nbap_maxNrOfUL_DPDCHs, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MaxNrOfUL_DPDCHs },
  { &hf_nbap_ul_PunctureLimit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PunctureLimit },
  { &hf_nbap_tFCS           , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCS },
  { &hf_nbap_ul_DPCCH_SlotFormat, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_DPCCH_SlotFormat },
  { &hf_nbap_diversityMode  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DiversityMode },
  { &hf_nbap_not_Used_sSDT_CellIDLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_not_Used_s_FieldLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_DPCH_Information_RL_ReconfPrepFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_DPCH_Information_RL_ReconfPrepFDD, UL_DPCH_Information_RL_ReconfPrepFDD_sequence);

  return offset;
}


static const per_sequence_t DL_DPCH_Information_RL_ReconfPrepFDD_sequence[] = {
  { &hf_nbap_tFCS           , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCS },
  { &hf_nbap_dl_DPCH_SlotFormat, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_DPCH_SlotFormat },
  { &hf_nbap_tFCI_SignallingMode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCI_SignallingMode },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCI_Presence },
  { &hf_nbap_multiplexingPosition, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MultiplexingPosition },
  { &hf_nbap_not_Used_pDSCH_CodeMapping, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_not_Used_pDSCH_RL_ID, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_limitedPowerIncrease, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_LimitedPowerIncrease },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_DPCH_Information_RL_ReconfPrepFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_DPCH_Information_RL_ReconfPrepFDD, DL_DPCH_Information_RL_ReconfPrepFDD_sequence);

  return offset;
}


static const per_sequence_t PowerOffsetInformation_RL_ReconfPrepFDD_sequence[] = {
  { &hf_nbap_pO1_ForTFCI_Bits, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PowerOffset },
  { &hf_nbap_pO2_ForTPC_Bits, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PowerOffset },
  { &hf_nbap_pO3_ForPilotBits, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PowerOffset },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PowerOffsetInformation_RL_ReconfPrepFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PowerOffsetInformation_RL_ReconfPrepFDD, PowerOffsetInformation_RL_ReconfPrepFDD_sequence);

  return offset;
}


static const per_sequence_t DL_DPCH_Power_Information_RL_ReconfPrepFDD_sequence[] = {
  { &hf_nbap_powerOffsetInformation_03, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PowerOffsetInformation_RL_ReconfPrepFDD },
  { &hf_nbap_fdd_TPC_DownlinkStepSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_TPC_DownlinkStepSize },
  { &hf_nbap_innerLoopDLPCStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_InnerLoopDLPCStatus },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_DPCH_Power_Information_RL_ReconfPrepFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_DPCH_Power_Information_RL_ReconfPrepFDD, DL_DPCH_Power_Information_RL_ReconfPrepFDD_sequence);

  return offset;
}


static const per_sequence_t DCH_DeleteItem_RL_ReconfPrepFDD_sequence[] = {
  { &hf_nbap_dCH_ID_01      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DCH_DeleteItem_RL_ReconfPrepFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DCH_DeleteItem_RL_ReconfPrepFDD, DCH_DeleteItem_RL_ReconfPrepFDD_sequence);

  return offset;
}


static const per_sequence_t DCH_DeleteList_RL_ReconfPrepFDD_sequence_of[1] = {
  { &hf_nbap_DCH_DeleteList_RL_ReconfPrepFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_DeleteItem_RL_ReconfPrepFDD },
};

static int
dissect_nbap_DCH_DeleteList_RL_ReconfPrepFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DCH_DeleteList_RL_ReconfPrepFDD, DCH_DeleteList_RL_ReconfPrepFDD_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t RL_InformationList_RL_ReconfPrepFDD_sequence_of[1] = {
  { &hf_nbap_RL_InformationList_RL_ReconfPrepFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_RL_InformationList_RL_ReconfPrepFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RL_InformationList_RL_ReconfPrepFDD, RL_InformationList_RL_ReconfPrepFDD_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RL_InformationItem_RL_ReconfPrepFDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_dl_CodeInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_FDD_DL_CodeInformation },
  { &hf_nbap_maxDL_Power    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_minDL_Power    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_not_Used_sSDT_Indication, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_not_Used_sSDT_Cell_Identity, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_transmitDiversityIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransmitDiversityIndicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_InformationItem_RL_ReconfPrepFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_InformationItem_RL_ReconfPrepFDD, RL_InformationItem_RL_ReconfPrepFDD_sequence);

  return offset;
}


static const per_sequence_t E_DPCH_Information_RL_ReconfPrepFDD_sequence[] = {
  { &hf_nbap_maxSet_E_DPDCHs, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Max_Set_E_DPDCHs },
  { &hf_nbap_ul_PunctureLimit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PunctureLimit },
  { &hf_nbap_e_TFCS_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_TFCS_Information },
  { &hf_nbap_e_TTI          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_TTI },
  { &hf_nbap_e_DPCCH_PO     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DPCCH_PO },
  { &hf_nbap_e_RGCH_2_IndexStepThreshold, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_RGCH_2_IndexStepThreshold },
  { &hf_nbap_e_RGCH_3_IndexStepThreshold, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_RGCH_3_IndexStepThreshold },
  { &hf_nbap_hARQ_Info_for_E_DCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HARQ_Info_for_E_DCH },
  { &hf_nbap_hSDSCH_Configured_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSDSCH_Configured_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DPCH_Information_RL_ReconfPrepFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DPCH_Information_RL_ReconfPrepFDD, E_DPCH_Information_RL_ReconfPrepFDD_sequence);

  return offset;
}


static const per_sequence_t PowerOffsetInformation_F_DPCH_RL_ReconfPrepFDD_sequence[] = {
  { &hf_nbap_pO2_ForTPC_Bits, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PowerOffset },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PowerOffsetInformation_F_DPCH_RL_ReconfPrepFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PowerOffsetInformation_F_DPCH_RL_ReconfPrepFDD, PowerOffsetInformation_F_DPCH_RL_ReconfPrepFDD_sequence);

  return offset;
}


static const per_sequence_t F_DPCH_Information_RL_ReconfPrepFDD_sequence[] = {
  { &hf_nbap_powerOffsetInformation_04, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PowerOffsetInformation_F_DPCH_RL_ReconfPrepFDD },
  { &hf_nbap_fdd_TPC_DownlinkStepSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_FDD_TPC_DownlinkStepSize },
  { &hf_nbap_limitedPowerIncrease, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_LimitedPowerIncrease },
  { &hf_nbap_innerLoopDLPCStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_InnerLoopDLPCStatus },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_F_DPCH_Information_RL_ReconfPrepFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_F_DPCH_Information_RL_ReconfPrepFDD, F_DPCH_Information_RL_ReconfPrepFDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkReconfigurationPrepareTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkReconfigurationPrepareTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkReconfigurationPrepareTDD, RadioLinkReconfigurationPrepareTDD_sequence);

  return offset;
}



static int
dissect_nbap_UL_DPCH_InformationAddList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_tFCS           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCS },
  { &hf_nbap_tFCI_Coding    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCI_Coding },
  { &hf_nbap_punctureLimit  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PunctureLimit },
  { &hf_nbap_ul_DPCH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_DPCH_InformationAddList_RL_ReconfPrepTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD, UL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD, UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_DPCH_InformationAddItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_uL_Timeslot_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Timeslot_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_DPCH_InformationAddItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_DPCH_InformationAddItem_RL_ReconfPrepTDD, UL_DPCH_InformationAddItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t UL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_uL_Timeslot_InformationLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_TimeslotLCR_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD, UL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t MultipleRL_UL_DPCH_InformationAddListIE_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_ul_DPCH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_DPCH_InformationAddList_RL_ReconfPrepTDD },
  { &hf_nbap_ul_DPCH_InformationListLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD },
  { &hf_nbap_ul_sir_target  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_SIR },
  { &hf_nbap_tDD_TPC_UplinkStepSize_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_TPC_UplinkStepSize_LCR },
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RL_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MultipleRL_UL_DPCH_InformationAddListIE_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MultipleRL_UL_DPCH_InformationAddListIE_RL_ReconfPrepTDD, MultipleRL_UL_DPCH_InformationAddListIE_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t MultipleRL_UL_DPCH_InformationAddList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_MultipleRL_UL_DPCH_InformationAddList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_MultipleRL_UL_DPCH_InformationAddListIE_RL_ReconfPrepTDD },
};

static int
dissect_nbap_MultipleRL_UL_DPCH_InformationAddList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_MultipleRL_UL_DPCH_InformationAddList_RL_ReconfPrepTDD, MultipleRL_UL_DPCH_InformationAddList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfRLs_1, FALSE);

  return offset;
}


static const per_sequence_t UL_DPCH_768_InformationAddList_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_uL_Timeslot_Information768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Timeslot768_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_DPCH_768_InformationAddList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_DPCH_768_InformationAddList_RL_ReconfPrepTDD, UL_DPCH_768_InformationAddList_RL_ReconfPrepTDD_sequence);

  return offset;
}



static int
dissect_nbap_UL_DPCH_InformationModify_AddList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_UL_DPCH_InformationModify_ModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_UL_DPCH_InformationModify_DeleteList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_tFCS           , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCS },
  { &hf_nbap_tFCI_Coding    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCI_Coding },
  { &hf_nbap_punctureLimit  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PunctureLimit },
  { &hf_nbap_ul_DPCH_InformationAddList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_DPCH_InformationModify_AddList_RL_ReconfPrepTDD },
  { &hf_nbap_ul_DPCH_InformationModifyList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_DPCH_InformationModify_ModifyList_RL_ReconfPrepTDD },
  { &hf_nbap_ul_DPCH_InformationDeleteList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_DPCH_InformationModify_DeleteList_RL_ReconfPrepTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD, UL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD, UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_uL_Timeslot_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Timeslot_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD, UL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_dPCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DPCH_ID },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD, UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD, UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfDPCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_Timeslot_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCI_Presence },
  { &hf_nbap_uL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_Timeslot_InformationModify_ModifyItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_Timeslot_InformationModify_ModifyItem_RL_ReconfPrepTDD, UL_Timeslot_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t UL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_UL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Timeslot_InformationModify_ModifyItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_UL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD, UL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfULTSs, FALSE);

  return offset;
}


static const per_sequence_t UL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_uL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD, UL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDDLCR_sequence[] = {
  { &hf_nbap_dPCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DPCH_ID },
  { &hf_nbap_tdd_ChannelisationCodeLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_ChannelisationCodeLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDDLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDDLCR, UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDDLCR_sequence);

  return offset;
}


static const per_sequence_t UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDDLCR_sequence_of[1] = {
  { &hf_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDDLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDDLCR },
};

static int
dissect_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDDLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDDLCR, UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDDLCR_sequence_of,
                                                  1, maxNrOfDPCHLCRs, FALSE);

  return offset;
}


static const per_sequence_t UL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCI_Presence },
  { &hf_nbap_uL_Code_InformationModify_ModifyList_RL_ReconfPrepTDDLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDDLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD, UL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t UL_TimeslotLCR_InformationModify_ModifyList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_UL_TimeslotLCR_InformationModify_ModifyList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_UL_TimeslotLCR_InformationModify_ModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_TimeslotLCR_InformationModify_ModifyList_RL_ReconfPrepTDD, UL_TimeslotLCR_InformationModify_ModifyList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfULTSLCRs, FALSE);

  return offset;
}


static const per_sequence_t UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD768_sequence[] = {
  { &hf_nbap_dPCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DPCH_ID },
  { &hf_nbap_tdd_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_ChannelisationCode768 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD768, UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD768_sequence);

  return offset;
}


static const per_sequence_t UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD768_sequence_of[1] = {
  { &hf_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD768 },
};

static int
dissect_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD768, UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD768_sequence_of,
                                                  1, maxNrOfDPCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_Timeslot_768_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftAndBurstType768 },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCI_Presence },
  { &hf_nbap_uL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD768 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_Timeslot_768_InformationModify_ModifyItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_Timeslot_768_InformationModify_ModifyItem_RL_ReconfPrepTDD, UL_Timeslot_768_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t UL_Timeslot768_InformationModify_ModifyList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_UL_Timeslot768_InformationModify_ModifyList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Timeslot_768_InformationModify_ModifyItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_UL_Timeslot768_InformationModify_ModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_Timeslot768_InformationModify_ModifyList_RL_ReconfPrepTDD, UL_Timeslot768_InformationModify_ModifyList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfULTSs, FALSE);

  return offset;
}


static const per_sequence_t UL_DPCH_InformationModify_DeleteItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_dPCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DPCH_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_DPCH_InformationModify_DeleteItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_DPCH_InformationModify_DeleteItem_RL_ReconfPrepTDD, UL_DPCH_InformationModify_DeleteItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t UL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_UL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_DPCH_InformationModify_DeleteItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_UL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD, UL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfDPCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_uL_Timeslot_InformationLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_TimeslotLCR_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD, UL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t MultipleRL_UL_DPCH_InformationModifyListIE_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_ul_DPCH_InformationAddList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_DPCH_InformationModify_AddList_RL_ReconfPrepTDD },
  { &hf_nbap_ul_DPCH_InformationModifyList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_DPCH_InformationModify_ModifyList_RL_ReconfPrepTDD },
  { &hf_nbap_ul_DPCH_InformationDeleteList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_DPCH_InformationModify_DeleteList_RL_ReconfPrepTDD },
  { &hf_nbap_ul_DPCH_InformationAddListLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD },
  { &hf_nbap_ul_sir_target  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_SIR },
  { &hf_nbap_tDD_TPC_UplinkStepSize_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_TPC_UplinkStepSize_LCR },
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RL_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MultipleRL_UL_DPCH_InformationModifyListIE_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MultipleRL_UL_DPCH_InformationModifyListIE_RL_ReconfPrepTDD, MultipleRL_UL_DPCH_InformationModifyListIE_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t MultipleRL_UL_DPCH_InformationModifyList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_MultipleRL_UL_DPCH_InformationModifyList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_MultipleRL_UL_DPCH_InformationModifyListIE_RL_ReconfPrepTDD },
};

static int
dissect_nbap_MultipleRL_UL_DPCH_InformationModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_MultipleRL_UL_DPCH_InformationModifyList_RL_ReconfPrepTDD, MultipleRL_UL_DPCH_InformationModifyList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfRLs_1, FALSE);

  return offset;
}


static const per_sequence_t UL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_uL_Timeslot_Information768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Timeslot768_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD, UL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD, UL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD, UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t CCTrCH_TPCAddItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CCTrCH_TPCAddItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CCTrCH_TPCAddItem_RL_ReconfPrepTDD, CCTrCH_TPCAddItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t CCTrCH_TPCAddList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_CCTrCH_TPCAddList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_TPCAddItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_CCTrCH_TPCAddList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_CCTrCH_TPCAddList_RL_ReconfPrepTDD, CCTrCH_TPCAddList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}



static int
dissect_nbap_DL_DPCH_InformationAddList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_tFCS           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCS },
  { &hf_nbap_tFCI_Coding    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCI_Coding },
  { &hf_nbap_punctureLimit  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PunctureLimit },
  { &hf_nbap_cCTrCH_TPCList_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CCTrCH_TPCAddList_RL_ReconfPrepTDD },
  { &hf_nbap_dl_DPCH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_DPCH_InformationAddList_RL_ReconfPrepTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD, DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD, DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_DPCH_InformationAddItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_dL_Timeslot_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Timeslot_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_DPCH_InformationAddItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_DPCH_InformationAddItem_RL_ReconfPrepTDD, DL_DPCH_InformationAddItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_dL_Timeslot_InformationLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_TimeslotLCR_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD, DL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t MultipleRL_DL_DPCH_InformationAddListIE_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_dl_DPCH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_DPCH_InformationAddList_RL_ReconfPrepTDD },
  { &hf_nbap_dl_DPCH_InformationListLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD },
  { &hf_nbap_cCTrCH_Initial_DL_Power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_tDD_TPC_DownlinkStepSize, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_TPC_DownlinkStepSize },
  { &hf_nbap_cCTrCH_Maximum_DL_Power_InformationAdd_RL_ReconfPrepTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_cCTrCH_Minimum_DL_Power_InformationAdd_RL_ReconfPrepTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RL_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MultipleRL_DL_DPCH_InformationAddListIE_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MultipleRL_DL_DPCH_InformationAddListIE_RL_ReconfPrepTDD, MultipleRL_DL_DPCH_InformationAddListIE_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t MultipleRL_DL_DPCH_InformationAddList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_MultipleRL_DL_DPCH_InformationAddList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_MultipleRL_DL_DPCH_InformationAddListIE_RL_ReconfPrepTDD },
};

static int
dissect_nbap_MultipleRL_DL_DPCH_InformationAddList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_MultipleRL_DL_DPCH_InformationAddList_RL_ReconfPrepTDD, MultipleRL_DL_DPCH_InformationAddList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfRLs_1, FALSE);

  return offset;
}


static const per_sequence_t DL_DPCH_768_InformationAddList_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_dL_Timeslot_Information768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Timeslot768_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_DPCH_768_InformationAddList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_DPCH_768_InformationAddList_RL_ReconfPrepTDD, DL_DPCH_768_InformationAddList_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t CCTrCH_TPCModifyItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CCTrCH_TPCModifyItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CCTrCH_TPCModifyItem_RL_ReconfPrepTDD, CCTrCH_TPCModifyItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t CCTrCH_TPCModifyList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_CCTrCH_TPCModifyList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_TPCModifyItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_CCTrCH_TPCModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_CCTrCH_TPCModifyList_RL_ReconfPrepTDD, CCTrCH_TPCModifyList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}



static int
dissect_nbap_DL_DPCH_InformationModify_AddList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_DL_DPCH_InformationModify_ModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_DL_DPCH_InformationModify_DeleteList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_tFCS           , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCS },
  { &hf_nbap_tFCI_Coding    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCI_Coding },
  { &hf_nbap_punctureLimit  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PunctureLimit },
  { &hf_nbap_cCTrCH_TPCList_02, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CCTrCH_TPCModifyList_RL_ReconfPrepTDD },
  { &hf_nbap_dl_DPCH_InformationAddList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_DPCH_InformationModify_AddList_RL_ReconfPrepTDD },
  { &hf_nbap_dl_DPCH_InformationModifyList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_DPCH_InformationModify_ModifyList_RL_ReconfPrepTDD },
  { &hf_nbap_dl_DPCH_InformationDeleteList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_DPCH_InformationModify_DeleteList_RL_ReconfPrepTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD, DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD, DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_dL_Timeslot_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Timeslot_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD, DL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_dPCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DPCH_ID },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD, DL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_DL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_DL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD, DL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD_sequence_of,
                                                  0, maxNrOfDPCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_Timeslot_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCI_Presence },
  { &hf_nbap_dL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_Timeslot_InformationModify_ModifyItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_Timeslot_InformationModify_ModifyItem_RL_ReconfPrepTDD, DL_Timeslot_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_DL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Timeslot_InformationModify_ModifyItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_DL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD, DL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfDLTSs, FALSE);

  return offset;
}


static const per_sequence_t DL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_dL_Timeslot_InformationAddModify_ModifyList_RL_ReconfPrepTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD, DL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DL_Code_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_dPCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DPCH_ID },
  { &hf_nbap_tdd_ChannelisationCodeLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_ChannelisationCodeLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_Code_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_Code_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD, DL_Code_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DL_Code_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_DL_Code_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Code_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_DL_Code_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_Code_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD, DL_Code_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfDPCHLCRs, FALSE);

  return offset;
}


static const per_sequence_t DL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCI_Presence },
  { &hf_nbap_dL_Code_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Code_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD, DL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD, DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfDLTSLCRs, FALSE);

  return offset;
}


static const per_sequence_t DL_Code_768_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_dPCH_ID768     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DPCH_ID768 },
  { &hf_nbap_tdd_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_ChannelisationCode768 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_Code_768_InformationModify_ModifyItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_Code_768_InformationModify_ModifyItem_RL_ReconfPrepTDD, DL_Code_768_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DL_Code_768_InformationModify_ModifyList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_DL_Code_768_InformationModify_ModifyList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Code_768_InformationModify_ModifyItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_DL_Code_768_InformationModify_ModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_Code_768_InformationModify_ModifyList_RL_ReconfPrepTDD, DL_Code_768_InformationModify_ModifyList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfDPCHs768, FALSE);

  return offset;
}


static const per_sequence_t DL_Timeslot_768_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCI_Presence },
  { &hf_nbap_dL_Code_768_InformationModify_ModifyList_RL_ReconfPrepTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Code_768_InformationModify_ModifyList_RL_ReconfPrepTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_Timeslot_768_InformationModify_ModifyItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_Timeslot_768_InformationModify_ModifyItem_RL_ReconfPrepTDD, DL_Timeslot_768_InformationModify_ModifyItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DL_Timeslot_768_InformationModify_ModifyList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_DL_Timeslot_768_InformationModify_ModifyList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Timeslot_768_InformationModify_ModifyItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_DL_Timeslot_768_InformationModify_ModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_Timeslot_768_InformationModify_ModifyList_RL_ReconfPrepTDD, DL_Timeslot_768_InformationModify_ModifyList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfDLTSs, FALSE);

  return offset;
}


static const per_sequence_t DL_DPCH_InformationModify_DeleteItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_dPCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DPCH_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_DPCH_InformationModify_DeleteItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_DPCH_InformationModify_DeleteItem_RL_ReconfPrepTDD, DL_DPCH_InformationModify_DeleteItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_DL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_DPCH_InformationModify_DeleteItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_DL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD, DL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfDPCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_dL_Timeslot_InformationLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_TimeslotLCR_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD, DL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t MultipleRL_DL_DPCH_InformationModifyListIE_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_dl_DPCH_InformationAddList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_DPCH_InformationModify_AddList_RL_ReconfPrepTDD },
  { &hf_nbap_dl_DPCH_InformationModifyList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_DPCH_InformationModify_ModifyList_RL_ReconfPrepTDD },
  { &hf_nbap_dl_DPCH_InformationDeleteList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_DPCH_InformationModify_DeleteList_RL_ReconfPrepTDD },
  { &hf_nbap_dl_DPCH_InformationAddListLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD },
  { &hf_nbap_tDD_TPC_DownlinkStepSize_InformationModify_RL_ReconfPrepTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_TPC_DownlinkStepSize },
  { &hf_nbap_cCTrCH_Maximum_DL_Power_InformationModify_RL_ReconfPrepTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_cCTrCH_Minimum_DL_Power_InformationModify_RL_ReconfPrepTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RL_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MultipleRL_DL_DPCH_InformationModifyListIE_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MultipleRL_DL_DPCH_InformationModifyListIE_RL_ReconfPrepTDD, MultipleRL_DL_DPCH_InformationModifyListIE_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t MultipleRL_DL_DPCH_InformationModifyList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_MultipleRL_DL_DPCH_InformationModifyList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_MultipleRL_DL_DPCH_InformationModifyListIE_RL_ReconfPrepTDD },
};

static int
dissect_nbap_MultipleRL_DL_DPCH_InformationModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_MultipleRL_DL_DPCH_InformationModifyList_RL_ReconfPrepTDD, MultipleRL_DL_DPCH_InformationModifyList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfRLs_1, FALSE);

  return offset;
}


static const per_sequence_t DL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_DPCHOffset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_DPCHOffset },
  { &hf_nbap_dL_Timeslot_Information768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Timeslot768_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD, DL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD, DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD, DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t DCH_DeleteItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_dCH_ID_01      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DCH_DeleteItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DCH_DeleteItem_RL_ReconfPrepTDD, DCH_DeleteItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DCH_DeleteList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_DCH_DeleteList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_DeleteItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_DCH_DeleteList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DCH_DeleteList_RL_ReconfPrepTDD, DCH_DeleteList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t DSCH_Information_ModifyItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_dSCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DSCH_ID },
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CCTrCH_ID },
  { &hf_nbap_transportFormatSet, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportFormatSet },
  { &hf_nbap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_AllocationRetentionPriority },
  { &hf_nbap_frameHandlingPriority, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_FrameHandlingPriority },
  { &hf_nbap_toAWS          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ToAWS },
  { &hf_nbap_toAWE          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ToAWE },
  { &hf_nbap_transportBearerRequestIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportBearerRequestIndicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DSCH_Information_ModifyItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DSCH_Information_ModifyItem_RL_ReconfPrepTDD, DSCH_Information_ModifyItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DSCH_Information_ModifyList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_DSCH_Information_ModifyList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DSCH_Information_ModifyItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_DSCH_Information_ModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DSCH_Information_ModifyList_RL_ReconfPrepTDD, DSCH_Information_ModifyList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfDSCHs, FALSE);

  return offset;
}


static const per_sequence_t DSCH_Information_DeleteItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_dSCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DSCH_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DSCH_Information_DeleteItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DSCH_Information_DeleteItem_RL_ReconfPrepTDD, DSCH_Information_DeleteItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DSCH_Information_DeleteList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_DSCH_Information_DeleteList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DSCH_Information_DeleteItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_DSCH_Information_DeleteList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DSCH_Information_DeleteList_RL_ReconfPrepTDD, DSCH_Information_DeleteList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfDSCHs, FALSE);

  return offset;
}


static const per_sequence_t USCH_Information_ModifyItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_uSCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_USCH_ID },
  { &hf_nbap_transportFormatSet, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TransportFormatSet },
  { &hf_nbap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_AllocationRetentionPriority },
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CCTrCH_ID },
  { &hf_nbap_transportBearerRequestIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TransportBearerRequestIndicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_USCH_Information_ModifyItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_USCH_Information_ModifyItem_RL_ReconfPrepTDD, USCH_Information_ModifyItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t USCH_Information_ModifyList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_USCH_Information_ModifyList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_USCH_Information_ModifyItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_USCH_Information_ModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_USCH_Information_ModifyList_RL_ReconfPrepTDD, USCH_Information_ModifyList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfUSCHs, FALSE);

  return offset;
}


static const per_sequence_t USCH_Information_DeleteItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_uSCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_USCH_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_USCH_Information_DeleteItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_USCH_Information_DeleteItem_RL_ReconfPrepTDD, USCH_Information_DeleteItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t USCH_Information_DeleteList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_USCH_Information_DeleteList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_USCH_Information_DeleteItem_RL_ReconfPrepTDD },
};

static int
dissect_nbap_USCH_Information_DeleteList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_USCH_Information_DeleteList_RL_ReconfPrepTDD, USCH_Information_DeleteList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfUSCHs, FALSE);

  return offset;
}


static const per_sequence_t RL_Information_RL_ReconfPrepTDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_maxDL_Power    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_minDL_Power    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_Information_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_Information_RL_ReconfPrepTDD, RL_Information_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t MultipleRL_Information_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_nbap_MultipleRL_Information_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_RL_Information_RL_ReconfPrepTDD },
};

static int
dissect_nbap_MultipleRL_Information_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_MultipleRL_Information_RL_ReconfPrepTDD, MultipleRL_Information_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfRLs_1, FALSE);

  return offset;
}


static const per_sequence_t RadioLinkReconfigurationReady_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkReconfigurationReady(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"RadioLinkReconfigurationReady ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkReconfigurationReady, RadioLinkReconfigurationReady_sequence);

  return offset;
}


static const per_sequence_t RL_InformationResponseList_RL_ReconfReady_sequence_of[1] = {
  { &hf_nbap_RL_InformationResponseList_RL_ReconfReady_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_RL_InformationResponseList_RL_ReconfReady(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RL_InformationResponseList_RL_ReconfReady, RL_InformationResponseList_RL_ReconfReady_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}



static int
dissect_nbap_DCH_InformationResponseList_RL_ReconfReady(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_DSCH_InformationResponseList_RL_ReconfReady(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_USCH_InformationResponseList_RL_ReconfReady(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t RL_InformationResponseItem_RL_ReconfReady_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_dCH_InformationResponseList_RL_ReconfReady, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DCH_InformationResponseList_RL_ReconfReady },
  { &hf_nbap_dSCH_InformationResponseList_RL_ReconfReady, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DSCH_InformationResponseList_RL_ReconfReady },
  { &hf_nbap_uSCH_InformationResponseList_RL_ReconfReady, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_USCH_InformationResponseList_RL_ReconfReady },
  { &hf_nbap_not_Used_tFCI2_BearerInformationResponse, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_NULL },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_InformationResponseItem_RL_ReconfReady(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_InformationResponseItem_RL_ReconfReady, RL_InformationResponseItem_RL_ReconfReady_sequence);

  return offset;
}


static const per_sequence_t RadioLinkReconfigurationFailure_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkReconfigurationFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"RadioLinkReconfigurationFailure ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;


col_set_str(actx->pinfo->cinfo, COL_INFO,"RadioLinkReconfigurationFailure ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkReconfigurationFailure, RadioLinkReconfigurationFailure_sequence);

  return offset;
}


static const per_sequence_t GeneralCauseList_RL_ReconfFailure_sequence[] = {
  { &hf_nbap_cause          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Cause },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GeneralCauseList_RL_ReconfFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GeneralCauseList_RL_ReconfFailure, GeneralCauseList_RL_ReconfFailure_sequence);

  return offset;
}


static const per_sequence_t RL_ReconfigurationFailureList_RL_ReconfFailure_sequence_of[1] = {
  { &hf_nbap_RL_ReconfigurationFailureList_RL_ReconfFailure_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_RL_ReconfigurationFailureList_RL_ReconfFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RL_ReconfigurationFailureList_RL_ReconfFailure, RL_ReconfigurationFailureList_RL_ReconfFailure_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RLSpecificCauseList_RL_ReconfFailure_sequence[] = {
  { &hf_nbap_rL_ReconfigurationFailureList_RL_ReconfFailure, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RL_ReconfigurationFailureList_RL_ReconfFailure },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RLSpecificCauseList_RL_ReconfFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RLSpecificCauseList_RL_ReconfFailure, RLSpecificCauseList_RL_ReconfFailure_sequence);

  return offset;
}


static const value_string nbap_CauseLevel_RL_ReconfFailure_vals[] = {
  {   0, "generalCause" },
  {   1, "rLSpecificCause" },
  { 0, NULL }
};

static const per_choice_t CauseLevel_RL_ReconfFailure_choice[] = {
  {   0, &hf_nbap_generalCause_04, ASN1_EXTENSION_ROOT    , dissect_nbap_GeneralCauseList_RL_ReconfFailure },
  {   1, &hf_nbap_rLSpecificCause_04, ASN1_EXTENSION_ROOT    , dissect_nbap_RLSpecificCauseList_RL_ReconfFailure },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_CauseLevel_RL_ReconfFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_CauseLevel_RL_ReconfFailure, CauseLevel_RL_ReconfFailure_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t RL_ReconfigurationFailureItem_RL_ReconfFailure_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_cause          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Cause },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_ReconfigurationFailureItem_RL_ReconfFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_ReconfigurationFailureItem_RL_ReconfFailure, RL_ReconfigurationFailureItem_RL_ReconfFailure_sequence);

  return offset;
}


static const per_sequence_t RadioLinkReconfigurationCommit_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkReconfigurationCommit(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"RadioLinkReconfigurationCommit ");
/* CRNC -> Node B */
actx->pinfo->link_dir=P2P_DIR_DL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkReconfigurationCommit, RadioLinkReconfigurationCommit_sequence);

  return offset;
}


static const per_sequence_t RadioLinkReconfigurationCancel_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkReconfigurationCancel(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"RadioLinkReconfigurationCancel ");
/* CRNC -> Node B */
actx->pinfo->link_dir=P2P_DIR_DL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkReconfigurationCancel, RadioLinkReconfigurationCancel_sequence);

  return offset;
}


static const per_sequence_t RadioLinkReconfigurationRequestFDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkReconfigurationRequestFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"RadioLinkReconfigurationRequestFDD(FDD) ");
/* CRNC -> Node B */
actx->pinfo->link_dir=P2P_DIR_DL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkReconfigurationRequestFDD, RadioLinkReconfigurationRequestFDD_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_RL_Reconf_Req_ItemIEs_sequence[] = {
  { &hf_nbap_hSPDSCH_RL_ID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_c_ID           , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_C_ID },
  { &hf_nbap_hS_DSCH_FDD_Secondary_Serving_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_DSCH_FDD_Secondary_Serving_Information },
  { &hf_nbap_hS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised },
  { &hf_nbap_hS_DSCH_Secondary_Serving_Remove, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_DSCH_Secondary_Serving_Remove },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_HS_Cell_Information_RL_Reconf_Req_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_HS_Cell_Information_RL_Reconf_Req_ItemIEs, Additional_HS_Cell_Information_RL_Reconf_Req_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_RL_Reconf_Req_sequence_of[1] = {
  { &hf_nbap_Additional_HS_Cell_Information_RL_Reconf_Req_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_HS_Cell_Information_RL_Reconf_Req_ItemIEs },
};

static int
dissect_nbap_Additional_HS_Cell_Information_RL_Reconf_Req(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Additional_HS_Cell_Information_RL_Reconf_Req, Additional_HS_Cell_Information_RL_Reconf_Req_sequence_of,
                                                  1, maxNrOfHSDSCH_1, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_RL_Reconf_Req_sequence[] = {
  { &hf_nbap_setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_Cell_Information_RL_Reconf_Req(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_Cell_Information_RL_Reconf_Req, Additional_EDCH_Cell_Information_RL_Reconf_Req_sequence);

  return offset;
}


static const per_sequence_t UL_DPCH_Information_RL_ReconfRqstFDD_sequence[] = {
  { &hf_nbap_ul_TFCS        , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCS },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_DPCH_Information_RL_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_DPCH_Information_RL_ReconfRqstFDD, UL_DPCH_Information_RL_ReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t DL_DPCH_Information_RL_ReconfRqstFDD_sequence[] = {
  { &hf_nbap_dl_TFCS        , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCS },
  { &hf_nbap_tFCI_SignallingMode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCI_SignallingMode },
  { &hf_nbap_limitedPowerIncrease, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_LimitedPowerIncrease },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_DPCH_Information_RL_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_DPCH_Information_RL_ReconfRqstFDD, DL_DPCH_Information_RL_ReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t DCH_DeleteItem_RL_ReconfRqstFDD_sequence[] = {
  { &hf_nbap_dCH_ID_01      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DCH_DeleteItem_RL_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DCH_DeleteItem_RL_ReconfRqstFDD, DCH_DeleteItem_RL_ReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t DCH_DeleteList_RL_ReconfRqstFDD_sequence_of[1] = {
  { &hf_nbap_DCH_DeleteList_RL_ReconfRqstFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_DeleteItem_RL_ReconfRqstFDD },
};

static int
dissect_nbap_DCH_DeleteList_RL_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DCH_DeleteList_RL_ReconfRqstFDD, DCH_DeleteList_RL_ReconfRqstFDD_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t RL_InformationList_RL_ReconfRqstFDD_sequence_of[1] = {
  { &hf_nbap_RL_InformationList_RL_ReconfRqstFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_RL_InformationList_RL_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RL_InformationList_RL_ReconfRqstFDD, RL_InformationList_RL_ReconfRqstFDD_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RL_InformationItem_RL_ReconfRqstFDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_maxDL_Power    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_minDL_Power    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_dl_CodeInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_FDD_DL_CodeInformation },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_InformationItem_RL_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_InformationItem_RL_ReconfRqstFDD, RL_InformationItem_RL_ReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t E_DPCH_Information_RL_ReconfRqstFDD_sequence[] = {
  { &hf_nbap_maxSet_E_DPDCHs, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Max_Set_E_DPDCHs },
  { &hf_nbap_ul_PunctureLimit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PunctureLimit },
  { &hf_nbap_e_TFCS_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_TFCS_Information },
  { &hf_nbap_e_TTI          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_TTI },
  { &hf_nbap_e_DPCCH_PO     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_DPCCH_PO },
  { &hf_nbap_e_RGCH_2_IndexStepThreshold, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_RGCH_2_IndexStepThreshold },
  { &hf_nbap_e_RGCH_3_IndexStepThreshold, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_RGCH_3_IndexStepThreshold },
  { &hf_nbap_hARQ_Info_for_E_DCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HARQ_Info_for_E_DCH },
  { &hf_nbap_hSDSCH_Configured_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSDSCH_Configured_Indicator },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DPCH_Information_RL_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DPCH_Information_RL_ReconfRqstFDD, E_DPCH_Information_RL_ReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkReconfigurationRequestTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkReconfigurationRequestTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkReconfigurationRequestTDD, RadioLinkReconfigurationRequestTDD_sequence);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_sequence_of[1] = {
  { &hf_nbap_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD, UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_tFCS           , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCS },
  { &hf_nbap_punctureLimit  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PunctureLimit },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD, UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_sequence_of[1] = {
  { &hf_nbap_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD, UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD, UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_sequence_of[1] = {
  { &hf_nbap_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD, DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_tFCS           , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCS },
  { &hf_nbap_punctureLimit  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PunctureLimit },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD, DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_maxPowerLCR    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_minPowerLCR    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfRqstTDD, DL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD_sequence_of[1] = {
  { &hf_nbap_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfRqstTDD },
};

static int
dissect_nbap_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD, DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD_sequence_of,
                                                  1, maxNrOfDLTSLCRs, FALSE);

  return offset;
}


static const per_sequence_t DL_DPCH_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_dL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_DPCH_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_DPCH_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD, DL_DPCH_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t MultipleRL_DL_CCTrCH_InformationModifyListIE_RL_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_dl_DPCH_LCR_InformationModifyList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_DPCH_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD },
  { &hf_nbap_cCTrCH_Maximum_DL_Power_InformationModify_RL_ReconfRqstTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_cCTrCH_Minimum_DL_Power_InformationModify_RL_ReconfRqstTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RL_ID },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MultipleRL_DL_CCTrCH_InformationModifyListIE_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MultipleRL_DL_CCTrCH_InformationModifyListIE_RL_ReconfRqstTDD, MultipleRL_DL_CCTrCH_InformationModifyListIE_RL_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t MultipleRL_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_sequence_of[1] = {
  { &hf_nbap_MultipleRL_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_MultipleRL_DL_CCTrCH_InformationModifyListIE_RL_ReconfRqstTDD },
};

static int
dissect_nbap_MultipleRL_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_MultipleRL_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD, MultipleRL_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_sequence_of,
                                                  1, maxNrOfRLs_1, FALSE);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_sequence_of[1] = {
  { &hf_nbap_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD, DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD, DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DCH_DeleteItem_RL_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_dCH_ID_01      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DCH_DeleteItem_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DCH_DeleteItem_RL_ReconfRqstTDD, DCH_DeleteItem_RL_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DCH_DeleteList_RL_ReconfRqstTDD_sequence_of[1] = {
  { &hf_nbap_DCH_DeleteList_RL_ReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_DeleteItem_RL_ReconfRqstTDD },
};

static int
dissect_nbap_DCH_DeleteList_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DCH_DeleteList_RL_ReconfRqstTDD, DCH_DeleteList_RL_ReconfRqstTDD_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t RL_Information_RL_ReconfRqstTDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_maxDL_Power    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_minDL_Power    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_Information_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_Information_RL_ReconfRqstTDD, RL_Information_RL_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t Multiple_RL_Information_RL_ReconfRqstTDD_sequence_of[1] = {
  { &hf_nbap_Multiple_RL_Information_RL_ReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_RL_Information_RL_ReconfRqstTDD },
};

static int
dissect_nbap_Multiple_RL_Information_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Multiple_RL_Information_RL_ReconfRqstTDD, Multiple_RL_Information_RL_ReconfRqstTDD_sequence_of,
                                                  1, maxNrOfRLs_1, FALSE);

  return offset;
}


static const per_sequence_t RadioLinkReconfigurationResponse_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkReconfigurationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"RadioLinkReconfigurationResponse ");


  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkReconfigurationResponse, RadioLinkReconfigurationResponse_sequence);

  return offset;
}


static const per_sequence_t RL_InformationResponseList_RL_ReconfRsp_sequence_of[1] = {
  { &hf_nbap_RL_InformationResponseList_RL_ReconfRsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_RL_InformationResponseList_RL_ReconfRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RL_InformationResponseList_RL_ReconfRsp, RL_InformationResponseList_RL_ReconfRsp_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}



static int
dissect_nbap_DCH_InformationResponseList_RL_ReconfRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t RL_InformationResponseItem_RL_ReconfRsp_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_dCH_InformationResponseList_RL_ReconfRsp, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DCH_InformationResponseList_RL_ReconfRsp },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_InformationResponseItem_RL_ReconfRsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_InformationResponseItem_RL_ReconfRsp, RL_InformationResponseItem_RL_ReconfRsp_sequence);

  return offset;
}


static const per_sequence_t RadioLinkDeletionRequest_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkDeletionRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"RadioLinkDeletionRequest ");
/* CRNC -> Node B */
actx->pinfo->link_dir=P2P_DIR_DL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkDeletionRequest, RadioLinkDeletionRequest_sequence);

  return offset;
}


static const per_sequence_t RL_informationList_RL_DeletionRqst_sequence_of[1] = {
  { &hf_nbap_RL_informationList_RL_DeletionRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_RL_informationList_RL_DeletionRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RL_informationList_RL_DeletionRqst, RL_informationList_RL_DeletionRqst_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RL_informationItem_RL_DeletionRqst_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_informationItem_RL_DeletionRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_informationItem_RL_DeletionRqst, RL_informationItem_RL_DeletionRqst_sequence);

  return offset;
}


static const per_sequence_t RadioLinkDeletionResponse_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkDeletionResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"RadioLinkDeletionResponse ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkDeletionResponse, RadioLinkDeletionResponse_sequence);

  return offset;
}


static const per_sequence_t DL_PowerControlRequest_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_PowerControlRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"DL-PowerControlRequest ");
/* CRNC -> Node B */
actx->pinfo->link_dir=P2P_DIR_DL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_PowerControlRequest, DL_PowerControlRequest_sequence);

  return offset;
}


static const per_sequence_t DL_ReferencePowerInformationList_DL_PC_Rqst_sequence_of[1] = {
  { &hf_nbap_DL_ReferencePowerInformationList_DL_PC_Rqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_DL_ReferencePowerInformationList_DL_PC_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_ReferencePowerInformationList_DL_PC_Rqst, DL_ReferencePowerInformationList_DL_PC_Rqst_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t DL_ReferencePowerInformationItem_DL_PC_Rqst_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_dl_ReferencePower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_ReferencePowerInformationItem_DL_PC_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_ReferencePowerInformationItem_DL_PC_Rqst, DL_ReferencePowerInformationItem_DL_PC_Rqst_sequence);

  return offset;
}


static const per_sequence_t DL_PowerTimeslotControlRequest_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_PowerTimeslotControlRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"DL-PowerTimeslotControlRequest ");
/* CRNC -> Node B */
actx->pinfo->link_dir=P2P_DIR_DL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_PowerTimeslotControlRequest, DL_PowerTimeslotControlRequest_sequence);

  return offset;
}


static const per_sequence_t DedicatedMeasurementInitiationRequest_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DedicatedMeasurementInitiationRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"DedicatedMeasurementInitiationRequest ");
/* CRNC -> Node B */
actx->pinfo->link_dir=P2P_DIR_DL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DedicatedMeasurementInitiationRequest, DedicatedMeasurementInitiationRequest_sequence);

  return offset;
}


static const per_sequence_t RL_InformationList_DM_Rqst_sequence_of[1] = {
  { &hf_nbap_RL_InformationList_DM_Rqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_RL_InformationList_DM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RL_InformationList_DM_Rqst, RL_InformationList_DM_Rqst_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RL_DM_Rqst_sequence[] = {
  { &hf_nbap_rL_InformationList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_InformationList_DM_Rqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_DM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_DM_Rqst, RL_DM_Rqst_sequence);

  return offset;
}


static const per_sequence_t RL_Set_InformationItem_DM_Rqst_sequence[] = {
  { &hf_nbap_rL_Set_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_Set_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_Set_InformationItem_DM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_Set_InformationItem_DM_Rqst, RL_Set_InformationItem_DM_Rqst_sequence);

  return offset;
}


static const per_sequence_t RL_Set_InformationList_DM_Rqst_sequence_of[1] = {
  { &hf_nbap_RL_Set_InformationList_DM_Rqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_RL_Set_InformationItem_DM_Rqst },
};

static int
dissect_nbap_RL_Set_InformationList_DM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RL_Set_InformationList_DM_Rqst, RL_Set_InformationList_DM_Rqst_sequence_of,
                                                  1, maxNrOfRLSets, FALSE);

  return offset;
}


static const per_sequence_t RL_Set_DM_Rqst_sequence[] = {
  { &hf_nbap_rL_Set_InformationList_DM_Rqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_Set_InformationList_DM_Rqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_Set_DM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_Set_DM_Rqst, RL_Set_DM_Rqst_sequence);

  return offset;
}



static int
dissect_nbap_AllRL_DM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_AllRL_Set_DM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string nbap_DedicatedMeasurementObjectType_DM_Rqst_vals[] = {
  {   0, "rL" },
  {   1, "rLS" },
  {   2, "all-RL" },
  {   3, "all-RLS" },
  { 0, NULL }
};

static const per_choice_t DedicatedMeasurementObjectType_DM_Rqst_choice[] = {
  {   0, &hf_nbap_rL             , ASN1_EXTENSION_ROOT    , dissect_nbap_RL_DM_Rqst },
  {   1, &hf_nbap_rLS            , ASN1_EXTENSION_ROOT    , dissect_nbap_RL_Set_DM_Rqst },
  {   2, &hf_nbap_all_RL         , ASN1_EXTENSION_ROOT    , dissect_nbap_AllRL_DM_Rqst },
  {   3, &hf_nbap_all_RLS        , ASN1_EXTENSION_ROOT    , dissect_nbap_AllRL_Set_DM_Rqst },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_DedicatedMeasurementObjectType_DM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_DedicatedMeasurementObjectType_DM_Rqst, DedicatedMeasurementObjectType_DM_Rqst_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t RL_InformationItem_DM_Rqst_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_dPCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DPCH_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_InformationItem_DM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_InformationItem_DM_Rqst, RL_InformationItem_DM_Rqst_sequence);

  return offset;
}


static const per_sequence_t PUSCH_Info_DM_Rqst_sequence_of[1] = {
  { &hf_nbap_PUSCH_Info_DM_Rqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_PUSCH_ID },
};

static int
dissect_nbap_PUSCH_Info_DM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_PUSCH_Info_DM_Rqst, PUSCH_Info_DM_Rqst_sequence_of,
                                                  1, maxNrOfPUSCHs, FALSE);

  return offset;
}


static const per_sequence_t HSSICH_Info_DM_Rqst_sequence_of[1] = {
  { &hf_nbap_HSSICH_Info_DM_Rqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SICH_ID },
};

static int
dissect_nbap_HSSICH_Info_DM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HSSICH_Info_DM_Rqst, HSSICH_Info_DM_Rqst_sequence_of,
                                                  1, maxNrOfHSSICHs, FALSE);

  return offset;
}


static const per_sequence_t HSSICH_InfoExt_DM_Rqst_sequence_of[1] = {
  { &hf_nbap_HSSICH_InfoExt_DM_Rqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Extended_HS_SICH_ID },
};

static int
dissect_nbap_HSSICH_InfoExt_DM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HSSICH_InfoExt_DM_Rqst, HSSICH_InfoExt_DM_Rqst_sequence_of,
                                                  1, maxNrOfHSSICHs, FALSE);

  return offset;
}


static const per_sequence_t DedicatedMeasurementInitiationResponse_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DedicatedMeasurementInitiationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"DedicatedMeasurementInitiationResponse ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DedicatedMeasurementInitiationResponse, DedicatedMeasurementInitiationResponse_sequence);

  return offset;
}


static const per_sequence_t RL_InformationList_DM_Rsp_sequence_of[1] = {
  { &hf_nbap_RL_InformationList_DM_Rsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_RL_InformationList_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RL_InformationList_DM_Rsp, RL_InformationList_DM_Rsp_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RL_DM_Rsp_sequence[] = {
  { &hf_nbap_rL_InformationList_DM_Rsp, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_InformationList_DM_Rsp },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_DM_Rsp, RL_DM_Rsp_sequence);

  return offset;
}


static const per_sequence_t RL_Set_InformationList_DM_Rsp_sequence_of[1] = {
  { &hf_nbap_RL_Set_InformationList_DM_Rsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_RL_Set_InformationList_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RL_Set_InformationList_DM_Rsp, RL_Set_InformationList_DM_Rsp_sequence_of,
                                                  1, maxNrOfRLSets, FALSE);

  return offset;
}


static const per_sequence_t RL_Set_DM_Rsp_sequence[] = {
  { &hf_nbap_rL_Set_InformationList_DM_Rsp, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_Set_InformationList_DM_Rsp },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_Set_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_Set_DM_Rsp, RL_Set_DM_Rsp_sequence);

  return offset;
}


static const value_string nbap_DedicatedMeasurementObjectType_DM_Rsp_vals[] = {
  {   0, "rL" },
  {   1, "rLS" },
  {   2, "all-RL" },
  {   3, "all-RLS" },
  { 0, NULL }
};

static const per_choice_t DedicatedMeasurementObjectType_DM_Rsp_choice[] = {
  {   0, &hf_nbap_rL_01          , ASN1_EXTENSION_ROOT    , dissect_nbap_RL_DM_Rsp },
  {   1, &hf_nbap_rLS_01         , ASN1_EXTENSION_ROOT    , dissect_nbap_RL_Set_DM_Rsp },
  {   2, &hf_nbap_all_RL_01      , ASN1_EXTENSION_ROOT    , dissect_nbap_RL_DM_Rsp },
  {   3, &hf_nbap_all_RLS_01     , ASN1_EXTENSION_ROOT    , dissect_nbap_RL_Set_DM_Rsp },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_DedicatedMeasurementObjectType_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_DedicatedMeasurementObjectType_DM_Rsp, DedicatedMeasurementObjectType_DM_Rsp_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t RL_InformationItem_DM_Rsp_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_dPCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DPCH_ID },
  { &hf_nbap_dedicatedMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DedicatedMeasurementValue },
  { &hf_nbap_cFN            , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CFN },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_InformationItem_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_InformationItem_DM_Rsp, RL_InformationItem_DM_Rsp_sequence);

  return offset;
}


static const per_sequence_t PUSCH_Info_DM_Rsp_sequence_of[1] = {
  { &hf_nbap_PUSCH_Info_DM_Rsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_PUSCH_ID },
};

static int
dissect_nbap_PUSCH_Info_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_PUSCH_Info_DM_Rsp, PUSCH_Info_DM_Rsp_sequence_of,
                                                  1, maxNrOfPUSCHs, FALSE);

  return offset;
}


static const per_sequence_t Multiple_PUSCH_InfoListIE_DM_Rsp_sequence[] = {
  { &hf_nbap_pUSCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PUSCH_ID },
  { &hf_nbap_dedicatedMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DedicatedMeasurementValue },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Multiple_PUSCH_InfoListIE_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Multiple_PUSCH_InfoListIE_DM_Rsp, Multiple_PUSCH_InfoListIE_DM_Rsp_sequence);

  return offset;
}


static const per_sequence_t Multiple_PUSCH_InfoList_DM_Rsp_sequence_of[1] = {
  { &hf_nbap_Multiple_PUSCH_InfoList_DM_Rsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Multiple_PUSCH_InfoListIE_DM_Rsp },
};

static int
dissect_nbap_Multiple_PUSCH_InfoList_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Multiple_PUSCH_InfoList_DM_Rsp, Multiple_PUSCH_InfoList_DM_Rsp_sequence_of,
                                                  1, maxNrOfPUSCHs_1, FALSE);

  return offset;
}


static const per_sequence_t Multiple_DedicatedMeasurementValueItem_TDD_DM_Rsp_sequence[] = {
  { &hf_nbap_dPCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DPCH_ID },
  { &hf_nbap_dedicatedMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DedicatedMeasurementValue },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Multiple_DedicatedMeasurementValueItem_TDD_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Multiple_DedicatedMeasurementValueItem_TDD_DM_Rsp, Multiple_DedicatedMeasurementValueItem_TDD_DM_Rsp_sequence);

  return offset;
}


static const per_sequence_t Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp_sequence_of[1] = {
  { &hf_nbap_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Multiple_DedicatedMeasurementValueItem_TDD_DM_Rsp },
};

static int
dissect_nbap_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp, Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp_sequence_of,
                                                  1, maxNrOfDPCHsPerRL_1, FALSE);

  return offset;
}


static const per_sequence_t Multiple_DedicatedMeasurementValueItem_LCR_TDD_DM_Rsp_sequence[] = {
  { &hf_nbap_dPCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DPCH_ID },
  { &hf_nbap_dedicatedMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DedicatedMeasurementValue },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Multiple_DedicatedMeasurementValueItem_LCR_TDD_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Multiple_DedicatedMeasurementValueItem_LCR_TDD_DM_Rsp, Multiple_DedicatedMeasurementValueItem_LCR_TDD_DM_Rsp_sequence);

  return offset;
}


static const per_sequence_t Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp_sequence_of[1] = {
  { &hf_nbap_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Multiple_DedicatedMeasurementValueItem_LCR_TDD_DM_Rsp },
};

static int
dissect_nbap_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp, Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp_sequence_of,
                                                  1, maxNrOfDPCHsLCRPerRL_1, FALSE);

  return offset;
}


static const per_sequence_t Multiple_HSSICHMeasurementValueItem_TDD_DM_Rsp_sequence[] = {
  { &hf_nbap_hsSICH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SICH_ID },
  { &hf_nbap_dedicatedMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DedicatedMeasurementValue },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Multiple_HSSICHMeasurementValueItem_TDD_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Multiple_HSSICHMeasurementValueItem_TDD_DM_Rsp, Multiple_HSSICHMeasurementValueItem_TDD_DM_Rsp_sequence);

  return offset;
}


static const per_sequence_t Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp_sequence_of[1] = {
  { &hf_nbap_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Multiple_HSSICHMeasurementValueItem_TDD_DM_Rsp },
};

static int
dissect_nbap_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp, Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp_sequence_of,
                                                  1, maxNrOfHSSICHs_1, FALSE);

  return offset;
}


static const per_sequence_t Multiple_DedicatedMeasurementValueItem_768_TDD_DM_Rsp_sequence[] = {
  { &hf_nbap_dPCH_ID768     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DPCH_ID768 },
  { &hf_nbap_dedicatedMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DedicatedMeasurementValue },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Multiple_DedicatedMeasurementValueItem_768_TDD_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Multiple_DedicatedMeasurementValueItem_768_TDD_DM_Rsp, Multiple_DedicatedMeasurementValueItem_768_TDD_DM_Rsp_sequence);

  return offset;
}


static const per_sequence_t Multiple_DedicatedMeasurementValueList_768_TDD_DM_Rsp_sequence_of[1] = {
  { &hf_nbap_Multiple_DedicatedMeasurementValueList_768_TDD_DM_Rsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Multiple_DedicatedMeasurementValueItem_768_TDD_DM_Rsp },
};

static int
dissect_nbap_Multiple_DedicatedMeasurementValueList_768_TDD_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Multiple_DedicatedMeasurementValueList_768_TDD_DM_Rsp, Multiple_DedicatedMeasurementValueList_768_TDD_DM_Rsp_sequence_of,
                                                  1, maxNrOfDPCHs768PerRL_1, FALSE);

  return offset;
}


static const per_sequence_t RL_Set_InformationItem_DM_Rsp_sequence[] = {
  { &hf_nbap_rL_Set_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_Set_ID },
  { &hf_nbap_dedicatedMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DedicatedMeasurementValue },
  { &hf_nbap_cFN            , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CFN },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_Set_InformationItem_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_Set_InformationItem_DM_Rsp, RL_Set_InformationItem_DM_Rsp_sequence);

  return offset;
}


static const per_sequence_t DedicatedMeasurementInitiationFailure_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DedicatedMeasurementInitiationFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"DedicatedMeasurementInitiationFailure ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DedicatedMeasurementInitiationFailure, DedicatedMeasurementInitiationFailure_sequence);

  return offset;
}


static const per_sequence_t DedicatedMeasurementReport_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DedicatedMeasurementReport(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"DedicatedMeasurementReport ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DedicatedMeasurementReport, DedicatedMeasurementReport_sequence);

  return offset;
}


static const per_sequence_t RL_InformationList_DM_Rprt_sequence_of[1] = {
  { &hf_nbap_RL_InformationList_DM_Rprt_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_RL_InformationList_DM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RL_InformationList_DM_Rprt, RL_InformationList_DM_Rprt_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RL_DM_Rprt_sequence[] = {
  { &hf_nbap_rL_InformationList_DM_Rprt, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_InformationList_DM_Rprt },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_DM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_DM_Rprt, RL_DM_Rprt_sequence);

  return offset;
}


static const per_sequence_t RL_Set_InformationList_DM_Rprt_sequence_of[1] = {
  { &hf_nbap_RL_Set_InformationList_DM_Rprt_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_RL_Set_InformationList_DM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RL_Set_InformationList_DM_Rprt, RL_Set_InformationList_DM_Rprt_sequence_of,
                                                  1, maxNrOfRLSets, FALSE);

  return offset;
}


static const per_sequence_t RL_Set_DM_Rprt_sequence[] = {
  { &hf_nbap_rL_Set_InformationList_DM_Rprt, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_Set_InformationList_DM_Rprt },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_Set_DM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_Set_DM_Rprt, RL_Set_DM_Rprt_sequence);

  return offset;
}


static const value_string nbap_DedicatedMeasurementObjectType_DM_Rprt_vals[] = {
  {   0, "rL" },
  {   1, "rLS" },
  {   2, "all-RL" },
  {   3, "all-RLS" },
  { 0, NULL }
};

static const per_choice_t DedicatedMeasurementObjectType_DM_Rprt_choice[] = {
  {   0, &hf_nbap_rL_02          , ASN1_EXTENSION_ROOT    , dissect_nbap_RL_DM_Rprt },
  {   1, &hf_nbap_rLS_02         , ASN1_EXTENSION_ROOT    , dissect_nbap_RL_Set_DM_Rprt },
  {   2, &hf_nbap_all_RL_02      , ASN1_EXTENSION_ROOT    , dissect_nbap_RL_DM_Rprt },
  {   3, &hf_nbap_all_RLS_02     , ASN1_EXTENSION_ROOT    , dissect_nbap_RL_Set_DM_Rprt },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_DedicatedMeasurementObjectType_DM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_DedicatedMeasurementObjectType_DM_Rprt, DedicatedMeasurementObjectType_DM_Rprt_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t RL_InformationItem_DM_Rprt_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_dPCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DPCH_ID },
  { &hf_nbap_dedicatedMeasurementValueInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DedicatedMeasurementValueInformation },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_InformationItem_DM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_InformationItem_DM_Rprt, RL_InformationItem_DM_Rprt_sequence);

  return offset;
}


static const per_sequence_t PUSCH_Info_DM_Rprt_sequence_of[1] = {
  { &hf_nbap_PUSCH_Info_DM_Rprt_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_PUSCH_ID },
};

static int
dissect_nbap_PUSCH_Info_DM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_PUSCH_Info_DM_Rprt, PUSCH_Info_DM_Rprt_sequence_of,
                                                  0, maxNrOfPUSCHs, FALSE);

  return offset;
}


static const per_sequence_t Multiple_PUSCH_InfoListIE_DM_Rprt_sequence[] = {
  { &hf_nbap_pUSCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PUSCH_ID },
  { &hf_nbap_dedicatedMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DedicatedMeasurementValue },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Multiple_PUSCH_InfoListIE_DM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Multiple_PUSCH_InfoListIE_DM_Rprt, Multiple_PUSCH_InfoListIE_DM_Rprt_sequence);

  return offset;
}


static const per_sequence_t Multiple_PUSCH_InfoList_DM_Rprt_sequence_of[1] = {
  { &hf_nbap_Multiple_PUSCH_InfoList_DM_Rprt_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Multiple_PUSCH_InfoListIE_DM_Rprt },
};

static int
dissect_nbap_Multiple_PUSCH_InfoList_DM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Multiple_PUSCH_InfoList_DM_Rprt, Multiple_PUSCH_InfoList_DM_Rprt_sequence_of,
                                                  1, maxNrOfPUSCHs_1, FALSE);

  return offset;
}


static const per_sequence_t RL_Set_InformationItem_DM_Rprt_sequence[] = {
  { &hf_nbap_rL_Set_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_Set_ID },
  { &hf_nbap_dedicatedMeasurementValueInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DedicatedMeasurementValueInformation },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_Set_InformationItem_DM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_Set_InformationItem_DM_Rprt, RL_Set_InformationItem_DM_Rprt_sequence);

  return offset;
}


static const per_sequence_t DedicatedMeasurementTerminationRequest_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DedicatedMeasurementTerminationRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"DedicatedMeasurementTerminationRequest ");
/* CRNC -> Node B */
actx->pinfo->link_dir=P2P_DIR_DL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DedicatedMeasurementTerminationRequest, DedicatedMeasurementTerminationRequest_sequence);

  return offset;
}


static const per_sequence_t DedicatedMeasurementFailureIndication_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DedicatedMeasurementFailureIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"DedicatedMeasurementFailureIndication ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DedicatedMeasurementFailureIndication, DedicatedMeasurementFailureIndication_sequence);

  return offset;
}


static const per_sequence_t RadioLinkFailureIndication_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkFailureIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"RadioLinkFailureIndication ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkFailureIndication, RadioLinkFailureIndication_sequence);

  return offset;
}


static const per_sequence_t RL_InformationList_RL_FailureInd_sequence_of[1] = {
  { &hf_nbap_RL_InformationList_RL_FailureInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_RL_InformationList_RL_FailureInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RL_InformationList_RL_FailureInd, RL_InformationList_RL_FailureInd_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RL_RL_FailureInd_sequence[] = {
  { &hf_nbap_rL_InformationList_RL_FailureInd, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_InformationList_RL_FailureInd },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_RL_FailureInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_RL_FailureInd, RL_RL_FailureInd_sequence);

  return offset;
}


static const per_sequence_t RL_Set_InformationList_RL_FailureInd_sequence_of[1] = {
  { &hf_nbap_RL_Set_InformationList_RL_FailureInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_RL_Set_InformationList_RL_FailureInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RL_Set_InformationList_RL_FailureInd, RL_Set_InformationList_RL_FailureInd_sequence_of,
                                                  1, maxNrOfRLSets, FALSE);

  return offset;
}


static const per_sequence_t RL_Set_RL_FailureInd_sequence[] = {
  { &hf_nbap_rL_Set_InformationList_RL_FailureInd, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_Set_InformationList_RL_FailureInd },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_Set_RL_FailureInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_Set_RL_FailureInd, RL_Set_RL_FailureInd_sequence);

  return offset;
}


static const per_sequence_t CCTrCH_InformationList_RL_FailureInd_sequence_of[1] = {
  { &hf_nbap_CCTrCH_InformationList_RL_FailureInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_CCTrCH_InformationList_RL_FailureInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_CCTrCH_InformationList_RL_FailureInd, CCTrCH_InformationList_RL_FailureInd_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t CCTrCH_RL_FailureInd_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_cCTrCH_InformationList_RL_FailureInd, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_InformationList_RL_FailureInd },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CCTrCH_RL_FailureInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CCTrCH_RL_FailureInd, CCTrCH_RL_FailureInd_sequence);

  return offset;
}


static const value_string nbap_Reporting_Object_RL_FailureInd_vals[] = {
  {   0, "rL" },
  {   1, "rL-Set" },
  {   2, "cCTrCH" },
  { 0, NULL }
};

static const per_choice_t Reporting_Object_RL_FailureInd_choice[] = {
  {   0, &hf_nbap_rL_03          , ASN1_EXTENSION_ROOT    , dissect_nbap_RL_RL_FailureInd },
  {   1, &hf_nbap_rL_Set         , ASN1_EXTENSION_ROOT    , dissect_nbap_RL_Set_RL_FailureInd },
  {   2, &hf_nbap_cCTrCH         , ASN1_NOT_EXTENSION_ROOT, dissect_nbap_CCTrCH_RL_FailureInd },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_Reporting_Object_RL_FailureInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_Reporting_Object_RL_FailureInd, Reporting_Object_RL_FailureInd_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t RL_InformationItem_RL_FailureInd_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_cause          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Cause },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_InformationItem_RL_FailureInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_InformationItem_RL_FailureInd, RL_InformationItem_RL_FailureInd_sequence);

  return offset;
}


static const per_sequence_t RL_Set_InformationItem_RL_FailureInd_sequence[] = {
  { &hf_nbap_rL_Set_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_Set_ID },
  { &hf_nbap_cause          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Cause },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_Set_InformationItem_RL_FailureInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_Set_InformationItem_RL_FailureInd, RL_Set_InformationItem_RL_FailureInd_sequence);

  return offset;
}


static const per_sequence_t CCTrCH_InformationItem_RL_FailureInd_sequence[] = {
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_cause          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Cause },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CCTrCH_InformationItem_RL_FailureInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CCTrCH_InformationItem_RL_FailureInd, CCTrCH_InformationItem_RL_FailureInd_sequence);

  return offset;
}


static const per_sequence_t RadioLinkPreemptionRequiredIndication_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkPreemptionRequiredIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"RadioLinkPreemptionRequiredIndication ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkPreemptionRequiredIndication, RadioLinkPreemptionRequiredIndication_sequence);

  return offset;
}


static const per_sequence_t RL_InformationList_RL_PreemptRequiredInd_sequence_of[1] = {
  { &hf_nbap_RL_InformationList_RL_PreemptRequiredInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_RL_InformationList_RL_PreemptRequiredInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RL_InformationList_RL_PreemptRequiredInd, RL_InformationList_RL_PreemptRequiredInd_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RL_InformationItem_RL_PreemptRequiredInd_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_InformationItem_RL_PreemptRequiredInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_InformationItem_RL_PreemptRequiredInd, RL_InformationItem_RL_PreemptRequiredInd_sequence);

  return offset;
}


static const per_sequence_t RadioLinkRestoreIndication_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkRestoreIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"RadioLinkRestoreIndication ");
/* CRNC -> Node B */
actx->pinfo->link_dir=P2P_DIR_DL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkRestoreIndication, RadioLinkRestoreIndication_sequence);

  return offset;
}


static const per_sequence_t RL_InformationList_RL_RestoreInd_sequence_of[1] = {
  { &hf_nbap_RL_InformationList_RL_RestoreInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_RL_InformationList_RL_RestoreInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RL_InformationList_RL_RestoreInd, RL_InformationList_RL_RestoreInd_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RL_RL_RestoreInd_sequence[] = {
  { &hf_nbap_rL_InformationList_RL_RestoreInd, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_InformationList_RL_RestoreInd },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_RL_RestoreInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_RL_RestoreInd, RL_RL_RestoreInd_sequence);

  return offset;
}


static const per_sequence_t RL_Set_InformationList_RL_RestoreInd_sequence_of[1] = {
  { &hf_nbap_RL_Set_InformationList_RL_RestoreInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_RL_Set_InformationList_RL_RestoreInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_RL_Set_InformationList_RL_RestoreInd, RL_Set_InformationList_RL_RestoreInd_sequence_of,
                                                  1, maxNrOfRLSets, FALSE);

  return offset;
}


static const per_sequence_t RL_Set_RL_RestoreInd_sequence[] = {
  { &hf_nbap_rL_Set_InformationList_RL_RestoreInd, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_Set_InformationList_RL_RestoreInd },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_Set_RL_RestoreInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_Set_RL_RestoreInd, RL_Set_RL_RestoreInd_sequence);

  return offset;
}


static const per_sequence_t CCTrCH_InformationList_RL_RestoreInd_sequence_of[1] = {
  { &hf_nbap_CCTrCH_InformationList_RL_RestoreInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_CCTrCH_InformationList_RL_RestoreInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_CCTrCH_InformationList_RL_RestoreInd, CCTrCH_InformationList_RL_RestoreInd_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t CCTrCH_RL_RestoreInd_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_cCTrCH_InformationList_RL_RestoreInd, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_InformationList_RL_RestoreInd },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CCTrCH_RL_RestoreInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CCTrCH_RL_RestoreInd, CCTrCH_RL_RestoreInd_sequence);

  return offset;
}


static const value_string nbap_Reporting_Object_RL_RestoreInd_vals[] = {
  {   0, "rL" },
  {   1, "rL-Set" },
  {   2, "cCTrCH" },
  { 0, NULL }
};

static const per_choice_t Reporting_Object_RL_RestoreInd_choice[] = {
  {   0, &hf_nbap_rL_04          , ASN1_EXTENSION_ROOT    , dissect_nbap_RL_RL_RestoreInd },
  {   1, &hf_nbap_rL_Set_01      , ASN1_EXTENSION_ROOT    , dissect_nbap_RL_Set_RL_RestoreInd },
  {   2, &hf_nbap_cCTrCH_01      , ASN1_NOT_EXTENSION_ROOT, dissect_nbap_CCTrCH_RL_RestoreInd },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_Reporting_Object_RL_RestoreInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_Reporting_Object_RL_RestoreInd, Reporting_Object_RL_RestoreInd_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t RL_InformationItem_RL_RestoreInd_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_InformationItem_RL_RestoreInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_InformationItem_RL_RestoreInd, RL_InformationItem_RL_RestoreInd_sequence);

  return offset;
}


static const per_sequence_t RL_Set_InformationItem_RL_RestoreInd_sequence[] = {
  { &hf_nbap_rL_Set_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_Set_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RL_Set_InformationItem_RL_RestoreInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RL_Set_InformationItem_RL_RestoreInd, RL_Set_InformationItem_RL_RestoreInd_sequence);

  return offset;
}


static const per_sequence_t CCTrCH_InformationItem_RL_RestoreInd_sequence[] = {
  { &hf_nbap_cCTrCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CCTrCH_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CCTrCH_InformationItem_RL_RestoreInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CCTrCH_InformationItem_RL_RestoreInd, CCTrCH_InformationItem_RL_RestoreInd_sequence);

  return offset;
}


static const per_sequence_t CompressedModeCommand_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CompressedModeCommand(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"CompressedModeCommand ");
/* CRNC -> Node B */
actx->pinfo->link_dir=P2P_DIR_DL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CompressedModeCommand, CompressedModeCommand_sequence);

  return offset;
}


static const per_sequence_t ErrorIndication_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ErrorIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"ErrorIndication ");


  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ErrorIndication, ErrorIndication_sequence);

  return offset;
}


static const per_sequence_t PrivateMessage_sequence[] = {
  { &hf_nbap_privateIEs     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PrivateIE_Container },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PrivateMessage(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"PrivateMessage ");

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PrivateMessage, PrivateMessage_sequence);

  return offset;
}


static const per_sequence_t PhysicalSharedChannelReconfigurationRequestFDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PhysicalSharedChannelReconfigurationRequestFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"PhysicalSharedChannelReconfigurationRequest(FDD) ");
/* CRNC -> Node B */
actx->pinfo->link_dir=P2P_DIR_DL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PhysicalSharedChannelReconfigurationRequestFDD, PhysicalSharedChannelReconfigurationRequestFDD_sequence);

  return offset;
}


static const per_sequence_t HSDPA_And_EDCH_CellPortion_InformationItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_cellPortionID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellPortionID },
  { &hf_nbap_hS_PDSCH_HS_SCCH_ScramblingCode_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_ScramblingCode },
  { &hf_nbap_hS_PDSCH_FDD_Code_Information_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_PDSCH_FDD_Code_Information },
  { &hf_nbap_hS_SCCH_FDD_Code_Information_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_SCCH_FDD_Code_Information },
  { &hf_nbap_hS_PDSCH_HS_SCCH_E_AGCH_E_RGCH_E_HICH_MaxPower_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MaximumTransmissionPower },
  { &hf_nbap_e_AGCH_And_E_RGCH_E_HICH_FDD_Scrambling_Code, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_ScramblingCode },
  { &hf_nbap_e_AGCH_FDD_Code_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_AGCH_FDD_Code_Information },
  { &hf_nbap_e_RGCH_E_HICH_FDD_Code_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_RGCH_E_HICH_FDD_Code_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDPA_And_EDCH_CellPortion_InformationItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDPA_And_EDCH_CellPortion_InformationItem_PSCH_ReconfRqst, HSDPA_And_EDCH_CellPortion_InformationItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t HSDPA_And_EDCH_CellPortion_InformationList_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_HSDPA_And_EDCH_CellPortion_InformationList_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HSDPA_And_EDCH_CellPortion_InformationItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_HSDPA_And_EDCH_CellPortion_InformationList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HSDPA_And_EDCH_CellPortion_InformationList_PSCH_ReconfRqst, HSDPA_And_EDCH_CellPortion_InformationList_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfCellPortionsPerCell, FALSE);

  return offset;
}


static const per_sequence_t PhysicalSharedChannelReconfigurationRequestTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PhysicalSharedChannelReconfigurationRequestTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PhysicalSharedChannelReconfigurationRequestTDD, PhysicalSharedChannelReconfigurationRequestTDD_sequence);

  return offset;
}



static int
dissect_nbap_PDSCH_Information_AddList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t PDSCHSets_AddItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_pDSCHSet_ID    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PDSCHSet_ID },
  { &hf_nbap_pDSCH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PDSCH_Information_AddList_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PDSCHSets_AddItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PDSCHSets_AddItem_PSCH_ReconfRqst, PDSCHSets_AddItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t PDSCHSets_AddList_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_PDSCHSets_AddList_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_PDSCHSets_AddItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_PDSCHSets_AddList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_PDSCHSets_AddList_PSCH_ReconfRqst, PDSCHSets_AddList_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfPDSCHSets, FALSE);

  return offset;
}


static const per_sequence_t DL_Code_InformationAddItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_pDSCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PDSCH_ID },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_Code_InformationAddItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_Code_InformationAddItem_PSCH_ReconfRqst, DL_Code_InformationAddItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t DL_Code_InformationAddList_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_DL_Code_InformationAddList_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Code_InformationAddItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_DL_Code_InformationAddList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_Code_InformationAddList_PSCH_ReconfRqst, DL_Code_InformationAddList_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfPDSCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_Timeslot_InformationAddItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCI_Presence },
  { &hf_nbap_dL_Code_InformationAddList_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Code_InformationAddList_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_Timeslot_InformationAddItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_Timeslot_InformationAddItem_PSCH_ReconfRqst, DL_Timeslot_InformationAddItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t DL_Timeslot_InformationAddList_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_DL_Timeslot_InformationAddList_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Timeslot_InformationAddItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_DL_Timeslot_InformationAddList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_Timeslot_InformationAddList_PSCH_ReconfRqst, DL_Timeslot_InformationAddList_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfDLTSs, FALSE);

  return offset;
}


static const per_sequence_t PDSCH_Information_AddItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_dL_Timeslot_InformationAddList_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Timeslot_InformationAddList_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PDSCH_Information_AddItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PDSCH_Information_AddItem_PSCH_ReconfRqst, PDSCH_Information_AddItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t DL_Code_InformationAddItem_LCR_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_pDSCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PDSCH_ID },
  { &hf_nbap_tdd_ChannelisationCodeLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCodeLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_Code_InformationAddItem_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_Code_InformationAddItem_LCR_PSCH_ReconfRqst, DL_Code_InformationAddItem_LCR_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t DL_Code_InformationAddList_LCR_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_DL_Code_InformationAddList_LCR_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Code_InformationAddItem_LCR_PSCH_ReconfRqst },
};

static int
dissect_nbap_DL_Code_InformationAddList_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_Code_InformationAddList_LCR_PSCH_ReconfRqst, DL_Code_InformationAddList_LCR_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfPDSCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_Timeslot_InformationAddItem_LCR_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCI_Presence },
  { &hf_nbap_dL_Code_InformationAddList_LCR_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Code_InformationAddList_LCR_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_Timeslot_InformationAddItem_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_Timeslot_InformationAddItem_LCR_PSCH_ReconfRqst, DL_Timeslot_InformationAddItem_LCR_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t DL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_DL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Timeslot_InformationAddItem_LCR_PSCH_ReconfRqst },
};

static int
dissect_nbap_DL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst, DL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfDLTSLCRs, FALSE);

  return offset;
}


static const per_sequence_t PDSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_dL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PDSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PDSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst, PDSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t DL_Code_InformationAddItem_768_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_pDSCH_ID768    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PDSCH_ID768 },
  { &hf_nbap_tdd_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode768 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_Code_InformationAddItem_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_Code_InformationAddItem_768_PSCH_ReconfRqst, DL_Code_InformationAddItem_768_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t DL_Code_InformationAddList_768_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_DL_Code_InformationAddList_768_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Code_InformationAddItem_768_PSCH_ReconfRqst },
};

static int
dissect_nbap_DL_Code_InformationAddList_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_Code_InformationAddList_768_PSCH_ReconfRqst, DL_Code_InformationAddList_768_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfPDSCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_Timeslot_InformationAddItem_768_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType768 },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCI_Presence },
  { &hf_nbap_dL_Code_InformationAddList_768_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Code_InformationAddList_768_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_Timeslot_InformationAddItem_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_Timeslot_InformationAddItem_768_PSCH_ReconfRqst, DL_Timeslot_InformationAddItem_768_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t DL_Timeslot_InformationAddList_768_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_DL_Timeslot_InformationAddList_768_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Timeslot_InformationAddItem_768_PSCH_ReconfRqst },
};

static int
dissect_nbap_DL_Timeslot_InformationAddList_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_Timeslot_InformationAddList_768_PSCH_ReconfRqst, DL_Timeslot_InformationAddList_768_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfDLTSs, FALSE);

  return offset;
}


static const per_sequence_t PDSCH_AddInformation_768_AddItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_dL_Timeslot_InformationAddList_768_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Timeslot_InformationAddList_768_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PDSCH_AddInformation_768_AddItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PDSCH_AddInformation_768_AddItem_PSCH_ReconfRqst, PDSCH_AddInformation_768_AddItem_PSCH_ReconfRqst_sequence);

  return offset;
}



static int
dissect_nbap_PDSCH_Information_ModifyList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t PDSCHSets_ModifyItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_pDSCHSet_ID    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PDSCHSet_ID },
  { &hf_nbap_pDSCH_InformationList_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PDSCH_Information_ModifyList_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PDSCHSets_ModifyItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PDSCHSets_ModifyItem_PSCH_ReconfRqst, PDSCHSets_ModifyItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t PDSCHSets_ModifyList_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_PDSCHSets_ModifyList_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_PDSCHSets_ModifyItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_PDSCHSets_ModifyList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_PDSCHSets_ModifyList_PSCH_ReconfRqst, PDSCHSets_ModifyList_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfPDSCHSets, FALSE);

  return offset;
}


static const per_sequence_t DL_Code_InformationModifyItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_pDSCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PDSCH_ID },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_Code_InformationModifyItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_Code_InformationModifyItem_PSCH_ReconfRqst, DL_Code_InformationModifyItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t DL_Code_InformationModifyList_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_DL_Code_InformationModifyList_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Code_InformationModifyItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_DL_Code_InformationModifyList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_Code_InformationModifyList_PSCH_ReconfRqst, DL_Code_InformationModifyList_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfPDSCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_Timeslot_InformationModifyItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCI_Presence },
  { &hf_nbap_dL_Code_InformationModifyList_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Code_InformationModifyList_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_Timeslot_InformationModifyItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_Timeslot_InformationModifyItem_PSCH_ReconfRqst, DL_Timeslot_InformationModifyItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t DL_Timeslot_InformationModifyList_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_DL_Timeslot_InformationModifyList_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Timeslot_InformationModifyItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_DL_Timeslot_InformationModifyList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_Timeslot_InformationModifyList_PSCH_ReconfRqst, DL_Timeslot_InformationModifyList_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfDLTSs, FALSE);

  return offset;
}


static const per_sequence_t PDSCH_Information_ModifyItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_dL_Timeslot_InformationModifyList_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Timeslot_InformationModifyList_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PDSCH_Information_ModifyItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PDSCH_Information_ModifyItem_PSCH_ReconfRqst, PDSCH_Information_ModifyItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t DL_Code_LCR_InformationModifyItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_pDSCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PDSCH_ID },
  { &hf_nbap_tdd_ChannelisationCodeLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCodeLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_Code_LCR_InformationModifyItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_Code_LCR_InformationModifyItem_PSCH_ReconfRqst, DL_Code_LCR_InformationModifyItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t DL_Code_LCR_InformationModifyList_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_DL_Code_LCR_InformationModifyList_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Code_LCR_InformationModifyItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_DL_Code_LCR_InformationModifyList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_Code_LCR_InformationModifyList_PSCH_ReconfRqst, DL_Code_LCR_InformationModifyList_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfPDSCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_Timeslot_LCR_InformationModifyItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCI_Presence },
  { &hf_nbap_dL_Code_LCR_InformationModifyList_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Code_LCR_InformationModifyList_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_Timeslot_LCR_InformationModifyItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_Timeslot_LCR_InformationModifyItem_PSCH_ReconfRqst, DL_Timeslot_LCR_InformationModifyItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t DL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_DL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Timeslot_LCR_InformationModifyItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_DL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst, DL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfDLTSLCRs, FALSE);

  return offset;
}


static const per_sequence_t PDSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_dL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PDSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PDSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst, PDSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t DL_Code_768_InformationModifyItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_pDSCH_ID768    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PDSCH_ID768 },
  { &hf_nbap_tdd_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode768 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_Code_768_InformationModifyItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_Code_768_InformationModifyItem_PSCH_ReconfRqst, DL_Code_768_InformationModifyItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t DL_Code_768_InformationModifyList_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_DL_Code_768_InformationModifyList_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Code_768_InformationModifyItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_DL_Code_768_InformationModifyList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_Code_768_InformationModifyList_PSCH_ReconfRqst, DL_Code_768_InformationModifyList_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfPDSCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_Timeslot_768_InformationModifyItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftAndBurstType768 },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCI_Presence },
  { &hf_nbap_dL_Code_768_InformationModifyList_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Code_768_InformationModifyList_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_Timeslot_768_InformationModifyItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_Timeslot_768_InformationModifyItem_PSCH_ReconfRqst, DL_Timeslot_768_InformationModifyItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t DL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_DL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Timeslot_768_InformationModifyItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_DL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst, DL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfDLTSs, FALSE);

  return offset;
}


static const per_sequence_t PDSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_dL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PDSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PDSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst, PDSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t PDSCHSets_DeleteItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_pDSCHSet_ID    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PDSCHSet_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PDSCHSets_DeleteItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PDSCHSets_DeleteItem_PSCH_ReconfRqst, PDSCHSets_DeleteItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t PDSCHSets_DeleteList_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_PDSCHSets_DeleteList_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_PDSCHSets_DeleteItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_PDSCHSets_DeleteList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_PDSCHSets_DeleteList_PSCH_ReconfRqst, PDSCHSets_DeleteList_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfPDSCHSets, FALSE);

  return offset;
}



static int
dissect_nbap_PUSCH_Information_AddList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t PUSCHSets_AddItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_pUSCHSet_ID    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PUSCHSet_ID },
  { &hf_nbap_pUSCH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_PUSCH_Information_AddList_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PUSCHSets_AddItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PUSCHSets_AddItem_PSCH_ReconfRqst, PUSCHSets_AddItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t PUSCHSets_AddList_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_PUSCHSets_AddList_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_PUSCHSets_AddItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_PUSCHSets_AddList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_PUSCHSets_AddList_PSCH_ReconfRqst, PUSCHSets_AddList_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfPUSCHSets, FALSE);

  return offset;
}


static const per_sequence_t UL_Code_InformationAddItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_pUSCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PUSCH_ID },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_Code_InformationAddItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_Code_InformationAddItem_PSCH_ReconfRqst, UL_Code_InformationAddItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t UL_Code_InformationAddList_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_UL_Code_InformationAddList_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Code_InformationAddItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_UL_Code_InformationAddList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_Code_InformationAddList_PSCH_ReconfRqst, UL_Code_InformationAddList_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfPUSCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_Timeslot_InformationAddItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCI_Presence },
  { &hf_nbap_uL_Code_InformationAddList_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Code_InformationAddList_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_Timeslot_InformationAddItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_Timeslot_InformationAddItem_PSCH_ReconfRqst, UL_Timeslot_InformationAddItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t UL_Timeslot_InformationAddList_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_UL_Timeslot_InformationAddList_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Timeslot_InformationAddItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_UL_Timeslot_InformationAddList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_Timeslot_InformationAddList_PSCH_ReconfRqst, UL_Timeslot_InformationAddList_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfULTSs, FALSE);

  return offset;
}


static const per_sequence_t PUSCH_Information_AddItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_uL_Timeslot_InformationAddList_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Timeslot_InformationAddList_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PUSCH_Information_AddItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PUSCH_Information_AddItem_PSCH_ReconfRqst, PUSCH_Information_AddItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t UL_Code_InformationAddItem_LCR_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_pUSCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PUSCH_ID },
  { &hf_nbap_tdd_ChannelisationCodeLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCodeLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_Code_InformationAddItem_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_Code_InformationAddItem_LCR_PSCH_ReconfRqst, UL_Code_InformationAddItem_LCR_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t UL_Code_InformationAddList_LCR_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_UL_Code_InformationAddList_LCR_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Code_InformationAddItem_LCR_PSCH_ReconfRqst },
};

static int
dissect_nbap_UL_Code_InformationAddList_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_Code_InformationAddList_LCR_PSCH_ReconfRqst, UL_Code_InformationAddList_LCR_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfPUSCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_Timeslot_InformationAddItem_LCR_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCI_Presence },
  { &hf_nbap_uL_Code_InformationAddList_LCR_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Code_InformationAddList_LCR_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_Timeslot_InformationAddItem_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_Timeslot_InformationAddItem_LCR_PSCH_ReconfRqst, UL_Timeslot_InformationAddItem_LCR_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t UL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_UL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Timeslot_InformationAddItem_LCR_PSCH_ReconfRqst },
};

static int
dissect_nbap_UL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst, UL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfULTSLCRs, FALSE);

  return offset;
}


static const per_sequence_t PUSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_uL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PUSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PUSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst, PUSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t UL_Code_InformationAddItem_768_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_pUSCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PUSCH_ID },
  { &hf_nbap_tdd_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode768 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_Code_InformationAddItem_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_Code_InformationAddItem_768_PSCH_ReconfRqst, UL_Code_InformationAddItem_768_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t UL_Code_InformationAddList_768_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_UL_Code_InformationAddList_768_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Code_InformationAddItem_768_PSCH_ReconfRqst },
};

static int
dissect_nbap_UL_Code_InformationAddList_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_Code_InformationAddList_768_PSCH_ReconfRqst, UL_Code_InformationAddList_768_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfPUSCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_Timeslot_InformationAddItem_768_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType768 },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TFCI_Presence },
  { &hf_nbap_uL_Code_InformationAddList_768_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Code_InformationAddList_768_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_Timeslot_InformationAddItem_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_Timeslot_InformationAddItem_768_PSCH_ReconfRqst, UL_Timeslot_InformationAddItem_768_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t UL_Timeslot_InformationAddList_768_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_UL_Timeslot_InformationAddList_768_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Timeslot_InformationAddItem_768_PSCH_ReconfRqst },
};

static int
dissect_nbap_UL_Timeslot_InformationAddList_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_Timeslot_InformationAddList_768_PSCH_ReconfRqst, UL_Timeslot_InformationAddList_768_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfULTSs, FALSE);

  return offset;
}


static const per_sequence_t PUSCH_AddInformation_768_AddItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_uL_Timeslot_InformationAddList_768_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Timeslot_InformationAddList_768_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PUSCH_AddInformation_768_AddItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PUSCH_AddInformation_768_AddItem_PSCH_ReconfRqst, PUSCH_AddInformation_768_AddItem_PSCH_ReconfRqst_sequence);

  return offset;
}



static int
dissect_nbap_PUSCH_Information_ModifyList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t PUSCHSets_ModifyItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_pUSCHSet_ID    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PUSCHSet_ID },
  { &hf_nbap_pUSCH_InformationList_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PUSCH_Information_ModifyList_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PUSCHSets_ModifyItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PUSCHSets_ModifyItem_PSCH_ReconfRqst, PUSCHSets_ModifyItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t PUSCHSets_ModifyList_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_PUSCHSets_ModifyList_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_PUSCHSets_ModifyItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_PUSCHSets_ModifyList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_PUSCHSets_ModifyList_PSCH_ReconfRqst, PUSCHSets_ModifyList_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfPUSCHSets, FALSE);

  return offset;
}


static const per_sequence_t UL_Code_InformationModifyItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_pUSCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PUSCH_ID },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_Code_InformationModifyItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_Code_InformationModifyItem_PSCH_ReconfRqst, UL_Code_InformationModifyItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t UL_Code_InformationModifyList_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_UL_Code_InformationModifyList_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Code_InformationModifyItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_UL_Code_InformationModifyList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_Code_InformationModifyList_PSCH_ReconfRqst, UL_Code_InformationModifyList_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfPUSCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_Timeslot_InformationModifyItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCI_Presence },
  { &hf_nbap_uL_Code_InformationModifyList_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_Code_InformationModifyList_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_Timeslot_InformationModifyItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_Timeslot_InformationModifyItem_PSCH_ReconfRqst, UL_Timeslot_InformationModifyItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t UL_Timeslot_InformationModifyList_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_UL_Timeslot_InformationModifyList_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Timeslot_InformationModifyItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_UL_Timeslot_InformationModifyList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_Timeslot_InformationModifyList_PSCH_ReconfRqst, UL_Timeslot_InformationModifyList_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfULTSs, FALSE);

  return offset;
}


static const per_sequence_t PUSCH_Information_ModifyItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_uL_Timeslot_InformationModifyList_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_Timeslot_InformationModifyList_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PUSCH_Information_ModifyItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PUSCH_Information_ModifyItem_PSCH_ReconfRqst, PUSCH_Information_ModifyItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t UL_Code_LCR_InformationModifyItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_pUSCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PUSCH_ID },
  { &hf_nbap_tdd_ChannelisationCodeLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCodeLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_Code_LCR_InformationModifyItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_Code_LCR_InformationModifyItem_PSCH_ReconfRqst, UL_Code_LCR_InformationModifyItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t UL_Code_LCR_InformationModifyList_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_UL_Code_LCR_InformationModifyList_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Code_LCR_InformationModifyItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_UL_Code_LCR_InformationModifyList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_Code_LCR_InformationModifyList_PSCH_ReconfRqst, UL_Code_LCR_InformationModifyList_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfPUSCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_Timeslot_LCR_InformationModifyItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCI_Presence },
  { &hf_nbap_uL_Code_LCR_InformationModifyList_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_Code_LCR_InformationModifyList_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_Timeslot_LCR_InformationModifyItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_Timeslot_LCR_InformationModifyItem_PSCH_ReconfRqst, UL_Timeslot_LCR_InformationModifyItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t UL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_UL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Timeslot_LCR_InformationModifyItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_UL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst, UL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfULTSLCRs, FALSE);

  return offset;
}


static const per_sequence_t PUSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_uL_Timeslot_InformationModifyList_LCR_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PUSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PUSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst, PUSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t UL_Code_768_InformationModifyItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_pUSCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PUSCH_ID },
  { &hf_nbap_tdd_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode768 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_Code_768_InformationModifyItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_Code_768_InformationModifyItem_PSCH_ReconfRqst, UL_Code_768_InformationModifyItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t UL_Code_768_InformationModifyList_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_UL_Code_768_InformationModifyList_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Code_768_InformationModifyItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_UL_Code_768_InformationModifyList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_Code_768_InformationModifyList_PSCH_ReconfRqst, UL_Code_768_InformationModifyList_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfPUSCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_Timeslot_768_InformationModifyItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftAndBurstType768 },
  { &hf_nbap_tFCI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TFCI_Presence },
  { &hf_nbap_uL_Code_768_InformationModifyList_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_Code_768_InformationModifyList_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UL_Timeslot_768_InformationModifyItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UL_Timeslot_768_InformationModifyItem_PSCH_ReconfRqst, UL_Timeslot_768_InformationModifyItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t UL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_UL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UL_Timeslot_768_InformationModifyItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_UL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst, UL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfULTSs, FALSE);

  return offset;
}


static const per_sequence_t PUSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RepetitionPeriod },
  { &hf_nbap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_RepetitionLength },
  { &hf_nbap_tdd_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_PhysicalChannelOffset },
  { &hf_nbap_uL_Timeslot_InformationModifyList_768_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_UL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PUSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PUSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst, PUSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t PUSCHSets_DeleteItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_pUSCHSet_ID    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PUSCHSet_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PUSCHSets_DeleteItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PUSCHSets_DeleteItem_PSCH_ReconfRqst, PUSCHSets_DeleteItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t PUSCHSets_DeleteList_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_PUSCHSets_DeleteList_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_PUSCHSets_DeleteItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_PUSCHSets_DeleteList_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_PUSCHSets_DeleteList_PSCH_ReconfRqst, PUSCHSets_DeleteList_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfPUSCHSets, FALSE);

  return offset;
}


static const per_sequence_t DL_HS_PDSCH_Codelist_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_DL_HS_PDSCH_Codelist_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
};

static int
dissect_nbap_DL_HS_PDSCH_Codelist_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_HS_PDSCH_Codelist_PSCH_ReconfRqst, DL_HS_PDSCH_Codelist_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfHSPDSCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_HS_PDSCH_Timeslot_InformationItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_dl_HS_PDSCH_Codelist_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_HS_PDSCH_Codelist_PSCH_ReconfRqst },
  { &hf_nbap_maxHSDSCH_HSSCCH_Power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MaximumTransmissionPower },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_HS_PDSCH_Timeslot_InformationItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_HS_PDSCH_Timeslot_InformationItem_PSCH_ReconfRqst, DL_HS_PDSCH_Timeslot_InformationItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t DL_HS_PDSCH_Timeslot_Information_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_DL_HS_PDSCH_Timeslot_Information_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_HS_PDSCH_Timeslot_InformationItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_DL_HS_PDSCH_Timeslot_Information_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_HS_PDSCH_Timeslot_Information_PSCH_ReconfRqst, DL_HS_PDSCH_Timeslot_Information_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfDLTSs, FALSE);

  return offset;
}


static const per_sequence_t HS_PDSCH_TDD_Information_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_dL_HS_PDSCH_Timeslot_Information_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_HS_PDSCH_Timeslot_Information_PSCH_ReconfRqst },
  { &hf_nbap_dL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_PDSCH_TDD_Information_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_PDSCH_TDD_Information_PSCH_ReconfRqst, HS_PDSCH_TDD_Information_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t DL_HS_PDSCH_Codelist_768_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_DL_HS_PDSCH_Codelist_768_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode768 },
};

static int
dissect_nbap_DL_HS_PDSCH_Codelist_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_HS_PDSCH_Codelist_768_PSCH_ReconfRqst, DL_HS_PDSCH_Codelist_768_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfHSPDSCHs768, FALSE);

  return offset;
}


static const per_sequence_t DL_HS_PDSCH_Timeslot_InformationItem_768_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType768 },
  { &hf_nbap_dl_HS_PDSCH_Codelist_768_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_HS_PDSCH_Codelist_768_PSCH_ReconfRqst },
  { &hf_nbap_maxHSDSCH_HSSCCH_Power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MaximumTransmissionPower },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DL_HS_PDSCH_Timeslot_InformationItem_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DL_HS_PDSCH_Timeslot_InformationItem_768_PSCH_ReconfRqst, DL_HS_PDSCH_Timeslot_InformationItem_768_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t DL_HS_PDSCH_Timeslot_Information_768_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_DL_HS_PDSCH_Timeslot_Information_768_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DL_HS_PDSCH_Timeslot_InformationItem_768_PSCH_ReconfRqst },
};

static int
dissect_nbap_DL_HS_PDSCH_Timeslot_Information_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DL_HS_PDSCH_Timeslot_Information_768_PSCH_ReconfRqst, DL_HS_PDSCH_Timeslot_Information_768_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfDLTSs, FALSE);

  return offset;
}


static const per_sequence_t MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst, MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst_sequence_of,
                                                  1, maxFrequencyinCell_1, FALSE);

  return offset;
}


static const per_sequence_t MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_dL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst },
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UARFCN },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst, MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t HS_SICH_Information_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_hsSICH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SICH_ID },
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_SICH_Information_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_SICH_Information_PSCH_ReconfRqst, HS_SICH_Information_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t HS_SCCH_InformationItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_hS_SCCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SCCH_ID },
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_hS_SCCH_MaxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_hS_SICH_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SICH_Information_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_SCCH_InformationItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_SCCH_InformationItem_PSCH_ReconfRqst, HS_SCCH_InformationItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t HS_SCCH_Information_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_HS_SCCH_Information_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SCCH_InformationItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_HS_SCCH_Information_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HS_SCCH_Information_PSCH_ReconfRqst, HS_SCCH_Information_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfHSSCCHs, FALSE);

  return offset;
}


static const per_sequence_t HS_SICH_Information_LCR_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_hsSICH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SICH_ID },
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_SICH_Information_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_SICH_Information_LCR_PSCH_ReconfRqst, HS_SICH_Information_LCR_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t HS_SCCH_InformationItem_LCR_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_hS_SCCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SCCH_ID },
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_first_TDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_second_TDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_hS_SCCH_MaxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_hS_SICH_Information_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SICH_Information_LCR_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_SCCH_InformationItem_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_SCCH_InformationItem_LCR_PSCH_ReconfRqst, HS_SCCH_InformationItem_LCR_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t HS_SCCH_Information_LCR_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_HS_SCCH_Information_LCR_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SCCH_InformationItem_LCR_PSCH_ReconfRqst },
};

static int
dissect_nbap_HS_SCCH_Information_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HS_SCCH_Information_LCR_PSCH_ReconfRqst, HS_SCCH_Information_LCR_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfHSSCCHs, FALSE);

  return offset;
}


static const per_sequence_t Add_To_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_hS_SCCH_Information_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_SCCH_Information_PSCH_ReconfRqst },
  { &hf_nbap_hS_SCCH_Information_LCR_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_SCCH_Information_LCR_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Add_To_HS_SCCH_Resource_Pool_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Add_To_HS_SCCH_Resource_Pool_PSCH_ReconfRqst, Add_To_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t HS_SICH_Information_768_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_hsSICH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SICH_ID },
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType768 },
  { &hf_nbap_tdd_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode768 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_SICH_Information_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_SICH_Information_768_PSCH_ReconfRqst, HS_SICH_Information_768_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t HS_SCCH_InformationItem_768_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_hS_SCCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SCCH_ID },
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType768 },
  { &hf_nbap_tdd_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode768 },
  { &hf_nbap_hS_SCCH_MaxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_hS_SICH_Information_768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SICH_Information_768_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_SCCH_InformationItem_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_SCCH_InformationItem_768_PSCH_ReconfRqst, HS_SCCH_InformationItem_768_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t HS_SCCH_Information_768_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_HS_SCCH_Information_768_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SCCH_InformationItem_768_PSCH_ReconfRqst },
};

static int
dissect_nbap_HS_SCCH_Information_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HS_SCCH_Information_768_PSCH_ReconfRqst, HS_SCCH_Information_768_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfHSSCCHs, FALSE);

  return offset;
}


static const per_sequence_t HS_SCCH_InformationExt_LCR_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_HS_SCCH_InformationExt_LCR_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SCCH_InformationItem_LCR_PSCH_ReconfRqst },
};

static int
dissect_nbap_HS_SCCH_InformationExt_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HS_SCCH_InformationExt_LCR_PSCH_ReconfRqst, HS_SCCH_InformationExt_LCR_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfHSSCCHsinExt, FALSE);

  return offset;
}


static const per_sequence_t HS_SICH_InformationModify_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_hsSICH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SICH_ID },
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_SICH_InformationModify_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_SICH_InformationModify_PSCH_ReconfRqst, HS_SICH_InformationModify_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t HS_SCCH_InformationModifyItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_hS_SCCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SCCH_ID },
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_hS_SCCH_MaxPower, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_hS_SICH_Information_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_SICH_InformationModify_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_SCCH_InformationModifyItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_SCCH_InformationModifyItem_PSCH_ReconfRqst, HS_SCCH_InformationModifyItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t HS_SCCH_InformationModify_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_HS_SCCH_InformationModify_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SCCH_InformationModifyItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_HS_SCCH_InformationModify_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HS_SCCH_InformationModify_PSCH_ReconfRqst, HS_SCCH_InformationModify_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfHSSCCHs, FALSE);

  return offset;
}


static const per_sequence_t HS_SICH_InformationModify_LCR_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_hsSICH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SICH_ID },
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_SICH_InformationModify_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_SICH_InformationModify_LCR_PSCH_ReconfRqst, HS_SICH_InformationModify_LCR_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t HS_SCCH_InformationModifyItem_LCR_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_hS_SCCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SCCH_ID },
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_first_TDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_second_TDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_hS_SCCH_MaxPower, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_hS_SICH_Information_LCR_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_SICH_InformationModify_LCR_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_SCCH_InformationModifyItem_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_SCCH_InformationModifyItem_LCR_PSCH_ReconfRqst, HS_SCCH_InformationModifyItem_LCR_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t HS_SCCH_InformationModify_LCR_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_HS_SCCH_InformationModify_LCR_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SCCH_InformationModifyItem_LCR_PSCH_ReconfRqst },
};

static int
dissect_nbap_HS_SCCH_InformationModify_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HS_SCCH_InformationModify_LCR_PSCH_ReconfRqst, HS_SCCH_InformationModify_LCR_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfHSSCCHs, FALSE);

  return offset;
}


static const per_sequence_t Modify_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_hS_SCCH_InformationModify_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_SCCH_InformationModify_PSCH_ReconfRqst },
  { &hf_nbap_hS_SCCH_InformationModify_LCR_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_SCCH_InformationModify_LCR_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Modify_HS_SCCH_Resource_Pool_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Modify_HS_SCCH_Resource_Pool_PSCH_ReconfRqst, Modify_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t HS_SCCH_InformationModifyExt_LCR_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_HS_SCCH_InformationModifyExt_LCR_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SCCH_InformationModifyItem_LCR_PSCH_ReconfRqst },
};

static int
dissect_nbap_HS_SCCH_InformationModifyExt_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HS_SCCH_InformationModifyExt_LCR_PSCH_ReconfRqst, HS_SCCH_InformationModifyExt_LCR_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfHSSCCHsinExt, FALSE);

  return offset;
}


static const per_sequence_t HS_SICH_InformationModify_768_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_hsSICH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SICH_ID },
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType768 },
  { &hf_nbap_tdd_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode768 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_SICH_InformationModify_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_SICH_InformationModify_768_PSCH_ReconfRqst, HS_SICH_InformationModify_768_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t HS_SCCH_InformationModifyItem_768_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_hS_SCCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SCCH_ID },
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType768 },
  { &hf_nbap_tdd_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode768 },
  { &hf_nbap_hS_SCCH_MaxPower, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_hS_SICH_Information_768_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HS_SICH_InformationModify_768_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HS_SCCH_InformationModifyItem_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HS_SCCH_InformationModifyItem_768_PSCH_ReconfRqst, HS_SCCH_InformationModifyItem_768_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t HS_SCCH_InformationModify_768_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_HS_SCCH_InformationModify_768_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SCCH_InformationModifyItem_768_PSCH_ReconfRqst },
};

static int
dissect_nbap_HS_SCCH_InformationModify_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HS_SCCH_InformationModify_768_PSCH_ReconfRqst, HS_SCCH_InformationModify_768_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfHSSCCHs, FALSE);

  return offset;
}


static const per_sequence_t Delete_From_HS_SCCH_Resource_PoolItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_hS_SCCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_SCCH_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Delete_From_HS_SCCH_Resource_PoolItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Delete_From_HS_SCCH_Resource_PoolItem_PSCH_ReconfRqst, Delete_From_HS_SCCH_Resource_PoolItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t Delete_From_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_Delete_From_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Delete_From_HS_SCCH_Resource_PoolItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_Delete_From_HS_SCCH_Resource_Pool_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Delete_From_HS_SCCH_Resource_Pool_PSCH_ReconfRqst, Delete_From_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfHSSCCHs, FALSE);

  return offset;
}


static const per_sequence_t E_PUCH_Timeslot_Info_sequence_of[1] = {
  { &hf_nbap_E_PUCH_Timeslot_Info_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
};

static int
dissect_nbap_E_PUCH_Timeslot_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_PUCH_Timeslot_Info, E_PUCH_Timeslot_Info_sequence_of,
                                                  1, maxNrOfE_PUCHSlots, FALSE);

  return offset;
}


static const per_sequence_t E_PUCH_Information_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_lTGI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_LTGI_Presence },
  { &hf_nbap_sNPL_Reporting_Type, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SNPL_Reporting_Type },
  { &hf_nbap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_e_PUCH_Timeslot_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_PUCH_Timeslot_Info },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_PUCH_Information_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_PUCH_Information_PSCH_ReconfRqst, E_PUCH_Information_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t E_AGCH_InformationItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_e_AGCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_AGCH_Id },
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_e_AGCH_MaxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_AGCH_InformationItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_AGCH_InformationItem_PSCH_ReconfRqst, E_AGCH_InformationItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t E_AGCH_Information_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_E_AGCH_Information_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_AGCH_InformationItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_E_AGCH_Information_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_AGCH_Information_PSCH_ReconfRqst, E_AGCH_Information_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfEAGCHs, FALSE);

  return offset;
}


static const per_sequence_t Add_To_E_AGCH_Resource_Pool_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_e_AGCH_Information_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_AGCH_Information_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Add_To_E_AGCH_Resource_Pool_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Add_To_E_AGCH_Resource_Pool_PSCH_ReconfRqst, Add_To_E_AGCH_Resource_Pool_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t E_AGCH_InformationModifyItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_e_AGCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_AGCH_Id },
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_e_AGCH_MaxPower, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_AGCH_InformationModifyItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_AGCH_InformationModifyItem_PSCH_ReconfRqst, E_AGCH_InformationModifyItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t E_AGCH_InformationModify_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_E_AGCH_InformationModify_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_AGCH_InformationModifyItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_E_AGCH_InformationModify_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_AGCH_InformationModify_PSCH_ReconfRqst, E_AGCH_InformationModify_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfEAGCHs, FALSE);

  return offset;
}


static const per_sequence_t Modify_E_AGCH_Resource_Pool_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_e_AGCH_InformationModify_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_AGCH_InformationModify_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Modify_E_AGCH_Resource_Pool_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Modify_E_AGCH_Resource_Pool_PSCH_ReconfRqst, Modify_E_AGCH_Resource_Pool_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t Delete_From_E_AGCH_Resource_PoolItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_e_AGCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_AGCH_Id },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Delete_From_E_AGCH_Resource_PoolItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Delete_From_E_AGCH_Resource_PoolItem_PSCH_ReconfRqst, Delete_From_E_AGCH_Resource_PoolItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t Delete_From_E_AGCH_Resource_Pool_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_Delete_From_E_AGCH_Resource_Pool_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Delete_From_E_AGCH_Resource_PoolItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_Delete_From_E_AGCH_Resource_Pool_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Delete_From_E_AGCH_Resource_Pool_PSCH_ReconfRqst, Delete_From_E_AGCH_Resource_Pool_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfEAGCHs, FALSE);

  return offset;
}


static const per_sequence_t E_HICH_Information_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_e_HICH_MaxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_HICH_Information_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_HICH_Information_PSCH_ReconfRqst, E_HICH_Information_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t E_PUCH_Information_768_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_lTGI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_LTGI_Presence },
  { &hf_nbap_sNPL_Reporting_Type, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SNPL_Reporting_Type },
  { &hf_nbap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType768 },
  { &hf_nbap_e_PUCH_Timeslot_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_PUCH_Timeslot_Info },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_PUCH_Information_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_PUCH_Information_768_PSCH_ReconfRqst, E_PUCH_Information_768_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t E_AGCH_InformationItem_768_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_e_AGCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_AGCH_Id },
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType768 },
  { &hf_nbap_tdd_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode768 },
  { &hf_nbap_e_AGCH_MaxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_AGCH_InformationItem_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_AGCH_InformationItem_768_PSCH_ReconfRqst, E_AGCH_InformationItem_768_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t E_AGCH_Information_768_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_E_AGCH_Information_768_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_AGCH_InformationItem_768_PSCH_ReconfRqst },
};

static int
dissect_nbap_E_AGCH_Information_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_AGCH_Information_768_PSCH_ReconfRqst, E_AGCH_Information_768_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfEAGCHs, FALSE);

  return offset;
}


static const per_sequence_t Add_To_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_e_AGCH_Information_768_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_AGCH_Information_768_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Add_To_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Add_To_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst, Add_To_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t E_AGCH_InformationModifyItem_768_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_e_AGCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_AGCH_Id },
  { &hf_nbap_timeSlot       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TimeSlot },
  { &hf_nbap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftAndBurstType768 },
  { &hf_nbap_tdd_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_ChannelisationCode768 },
  { &hf_nbap_e_AGCH_MaxPower, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_AGCH_InformationModifyItem_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_AGCH_InformationModifyItem_768_PSCH_ReconfRqst, E_AGCH_InformationModifyItem_768_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t E_AGCH_InformationModify_768_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_E_AGCH_InformationModify_768_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_AGCH_InformationModifyItem_768_PSCH_ReconfRqst },
};

static int
dissect_nbap_E_AGCH_InformationModify_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_AGCH_InformationModify_768_PSCH_ReconfRqst, E_AGCH_InformationModify_768_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfEAGCHs, FALSE);

  return offset;
}


static const per_sequence_t Modify_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_e_AGCH_InformationModify_768_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_AGCH_InformationModify_768_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Modify_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Modify_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst, Modify_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t E_HICH_Information_768_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftAndBurstType768 },
  { &hf_nbap_tdd_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode768 },
  { &hf_nbap_e_HICH_MaxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_HICH_Information_768_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_HICH_Information_768_PSCH_ReconfRqst, E_HICH_Information_768_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t E_PUCH_Codelist_LCR_sequence_of[1] = {
  { &hf_nbap_E_PUCH_Codelist_LCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
};

static int
dissect_nbap_E_PUCH_Codelist_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_PUCH_Codelist_LCR, E_PUCH_Codelist_LCR_sequence_of,
                                                  1, maxNrOfEPUCHcodes, FALSE);

  return offset;
}


static const per_sequence_t E_PUCH_Timeslot_Item_InfoLCR_sequence[] = {
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftAndBurstType_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_e_PUCH_Codelist_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_PUCH_Codelist_LCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_PUCH_Timeslot_Item_InfoLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_PUCH_Timeslot_Item_InfoLCR, E_PUCH_Timeslot_Item_InfoLCR_sequence);

  return offset;
}


static const per_sequence_t E_PUCH_Timeslot_InfoLCR_sequence_of[1] = {
  { &hf_nbap_E_PUCH_Timeslot_InfoLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_PUCH_Timeslot_Item_InfoLCR },
};

static int
dissect_nbap_E_PUCH_Timeslot_InfoLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_PUCH_Timeslot_InfoLCR, E_PUCH_Timeslot_InfoLCR_sequence_of,
                                                  1, maxNrOfE_PUCHSlotsLCR, FALSE);

  return offset;
}


static const per_sequence_t E_PUCH_Information_LCR_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_lTGI_Presence  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_LTGI_Presence },
  { &hf_nbap_sNPL_Reporting_Type, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SNPL_Reporting_Type },
  { &hf_nbap_e_PUCH_Timeslot_InfoLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_PUCH_Timeslot_InfoLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_PUCH_Information_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_PUCH_Information_LCR_PSCH_ReconfRqst, E_PUCH_Information_LCR_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t E_AGCH_InformationItem_LCR_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_e_AGCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_AGCH_Id },
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_first_TDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_second_TDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_e_AGCH_MaxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_AGCH_InformationItem_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_AGCH_InformationItem_LCR_PSCH_ReconfRqst, E_AGCH_InformationItem_LCR_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t E_AGCH_Information_LCR_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_E_AGCH_Information_LCR_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_AGCH_InformationItem_LCR_PSCH_ReconfRqst },
};

static int
dissect_nbap_E_AGCH_Information_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_AGCH_Information_LCR_PSCH_ReconfRqst, E_AGCH_Information_LCR_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfEAGCHs, FALSE);

  return offset;
}


static const per_sequence_t Add_To_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_e_AGCH_Information_LCR_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_AGCH_Information_LCR_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Add_To_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Add_To_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst, Add_To_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t E_AGCH_InformationModifyItem_LCR_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_e_AGCH_ID      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_AGCH_Id },
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_first_TDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_second_TDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_e_AGCH_MaxPower, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_AGCH_InformationModifyItem_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_AGCH_InformationModifyItem_LCR_PSCH_ReconfRqst, E_AGCH_InformationModifyItem_LCR_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t E_AGCH_InformationModify_LCR_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_E_AGCH_InformationModify_LCR_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_AGCH_InformationModifyItem_LCR_PSCH_ReconfRqst },
};

static int
dissect_nbap_E_AGCH_InformationModify_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_AGCH_InformationModify_LCR_PSCH_ReconfRqst, E_AGCH_InformationModify_LCR_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfEAGCHs, FALSE);

  return offset;
}


static const per_sequence_t Modify_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_e_AGCH_InformationModify_LCR_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_AGCH_InformationModify_LCR_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Modify_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Modify_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst, Modify_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t E_HICH_InformationItem_LCR_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_e_HICH_ID_TDD  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_HICH_ID_TDD },
  { &hf_nbap_e_HICH_Type    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_HICH_Type },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_e_HICH_MaxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_HICH_InformationItem_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_HICH_InformationItem_LCR_PSCH_ReconfRqst, E_HICH_InformationItem_LCR_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t E_HICH_Information_LCR_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_E_HICH_Information_LCR_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_HICH_InformationItem_LCR_PSCH_ReconfRqst },
};

static int
dissect_nbap_E_HICH_Information_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_HICH_Information_LCR_PSCH_ReconfRqst, E_HICH_Information_LCR_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfEHICHs, FALSE);

  return offset;
}


static const per_sequence_t Add_To_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_e_HICH_Information_LCR_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_HICH_Information_LCR_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Add_To_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Add_To_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst, Add_To_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t E_HICH_InformationModifyItem_LCR_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_e_HICH_ID_TDD  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_HICH_ID_TDD },
  { &hf_nbap_e_HICH_Type    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_HICH_Type },
  { &hf_nbap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TDD_ChannelisationCode },
  { &hf_nbap_timeSlotLCR    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TimeSlotLCR },
  { &hf_nbap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_MidambleShiftLCR },
  { &hf_nbap_e_HICH_MaxPower, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_HICH_InformationModifyItem_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_HICH_InformationModifyItem_LCR_PSCH_ReconfRqst, E_HICH_InformationModifyItem_LCR_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t E_HICH_InformationModify_LCR_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_E_HICH_InformationModify_LCR_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_HICH_InformationModifyItem_LCR_PSCH_ReconfRqst },
};

static int
dissect_nbap_E_HICH_InformationModify_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_HICH_InformationModify_LCR_PSCH_ReconfRqst, E_HICH_InformationModify_LCR_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfEHICHs, FALSE);

  return offset;
}


static const per_sequence_t Modify_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_e_HICH_InformationModify_LCR_PSCH_ReconfRqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_HICH_InformationModify_LCR_PSCH_ReconfRqst },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Modify_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Modify_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst, Modify_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t Delete_From_E_HICH_Resource_PoolItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_e_HICH_ID_TDD  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_HICH_ID_TDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Delete_From_E_HICH_Resource_PoolItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Delete_From_E_HICH_Resource_PoolItem_PSCH_ReconfRqst, Delete_From_E_HICH_Resource_PoolItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t Delete_From_E_HICH_Resource_Pool_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_Delete_From_E_HICH_Resource_Pool_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Delete_From_E_HICH_Resource_PoolItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_Delete_From_E_HICH_Resource_Pool_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Delete_From_E_HICH_Resource_Pool_PSCH_ReconfRqst, Delete_From_E_HICH_Resource_Pool_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfEHICHs, FALSE);

  return offset;
}


static const per_sequence_t SYNC_UL_Partition_LCR_sequence[] = {
  { &hf_nbap_eRUCCH_SYNC_UL_codes_bitmap, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_BIT_STRING_SIZE_8 },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SYNC_UL_Partition_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SYNC_UL_Partition_LCR, SYNC_UL_Partition_LCR_sequence);

  return offset;
}


static const per_sequence_t Delete_From_HS_SCCH_Resource_PoolExt_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_Delete_From_HS_SCCH_Resource_PoolExt_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Delete_From_HS_SCCH_Resource_PoolItem_PSCH_ReconfRqst },
};

static int
dissect_nbap_Delete_From_HS_SCCH_Resource_PoolExt_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Delete_From_HS_SCCH_Resource_PoolExt_PSCH_ReconfRqst, Delete_From_HS_SCCH_Resource_PoolExt_PSCH_ReconfRqst_sequence_of,
                                                  1, maxNrOfHSSCCHsinExt, FALSE);

  return offset;
}


static const per_sequence_t MultipleFreq_E_PUCH_Timeslot_InformationList_LCR_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_MultipleFreq_E_PUCH_Timeslot_InformationList_LCR_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_MultipleFreq_E_PUCH_Timeslot_InformationList_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_MultipleFreq_E_PUCH_Timeslot_InformationList_LCR_PSCH_ReconfRqst, MultipleFreq_E_PUCH_Timeslot_InformationList_LCR_PSCH_ReconfRqst_sequence_of,
                                                  1, maxFrequencyinCell_1, FALSE);

  return offset;
}


static const per_sequence_t MultipleFreq_E_PUCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst_sequence[] = {
  { &hf_nbap_e_PUCH_Timeslot_InfoLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_E_PUCH_Timeslot_InfoLCR },
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UARFCN },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MultipleFreq_E_PUCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MultipleFreq_E_PUCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst, MultipleFreq_E_PUCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst_sequence);

  return offset;
}


static const per_sequence_t Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst_Item_sequence[] = {
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UARFCN },
  { &hf_nbap_maximum_Target_ReceivedTotalWideBandPower_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Maximum_Target_ReceivedTotalWideBandPower_LCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst_Item, Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst_Item_sequence);

  return offset;
}


static const per_sequence_t Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst_sequence_of[1] = {
  { &hf_nbap_Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst_Item },
};

static int
dissect_nbap_Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst, Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst_sequence_of,
                                                  1, maxFrequencyinCell, FALSE);

  return offset;
}


static const per_sequence_t PhysicalSharedChannelReconfigurationResponse_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PhysicalSharedChannelReconfigurationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"PhysicalSharedChannelReconfigurationResponse ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;

  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PhysicalSharedChannelReconfigurationResponse, PhysicalSharedChannelReconfigurationResponse_sequence);

  return offset;
}


static const per_sequence_t E_HICH_TimeOffset_ExtensionLCR_sequence_of[1] = {
  { &hf_nbap_E_HICH_TimeOffset_ExtensionLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_E_HICH_TimeOffset_ExtensionLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_HICH_TimeOffset_ExtensionLCR, E_HICH_TimeOffset_ExtensionLCR_sequence_of,
                                                  1, maxFrequencyinCell_1, FALSE);

  return offset;
}


static const per_sequence_t MultipleFreq_E_HICH_TimeOffsetLCR_sequence[] = {
  { &hf_nbap_e_HICH_TimeOffsetLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_HICH_TimeOffsetLCR },
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UARFCN },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MultipleFreq_E_HICH_TimeOffsetLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MultipleFreq_E_HICH_TimeOffsetLCR, MultipleFreq_E_HICH_TimeOffsetLCR_sequence);

  return offset;
}


static const per_sequence_t PhysicalSharedChannelReconfigurationFailure_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_PhysicalSharedChannelReconfigurationFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"PhysicalSharedChannelReconfigurationFailure ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;


  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_PhysicalSharedChannelReconfigurationFailure, PhysicalSharedChannelReconfigurationFailure_sequence);

  return offset;
}


static const per_sequence_t GeneralCauseList_PSCH_ReconfFailure_sequence[] = {
  { &hf_nbap_cause          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Cause },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GeneralCauseList_PSCH_ReconfFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GeneralCauseList_PSCH_ReconfFailure, GeneralCauseList_PSCH_ReconfFailure_sequence);

  return offset;
}


static const per_sequence_t Unsuccessful_PDSCHSetList_PSCH_ReconfFailureTDD_sequence_of[1] = {
  { &hf_nbap_Unsuccessful_PDSCHSetList_PSCH_ReconfFailureTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_Unsuccessful_PDSCHSetList_PSCH_ReconfFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Unsuccessful_PDSCHSetList_PSCH_ReconfFailureTDD, Unsuccessful_PDSCHSetList_PSCH_ReconfFailureTDD_sequence_of,
                                                  0, maxNrOfPDSCHSets, FALSE);

  return offset;
}


static const per_sequence_t Unsuccessful_PUSCHSetList_PSCH_ReconfFailureTDD_sequence_of[1] = {
  { &hf_nbap_Unsuccessful_PUSCHSetList_PSCH_ReconfFailureTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_Unsuccessful_PUSCHSetList_PSCH_ReconfFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Unsuccessful_PUSCHSetList_PSCH_ReconfFailureTDD, Unsuccessful_PUSCHSetList_PSCH_ReconfFailureTDD_sequence_of,
                                                  0, maxNrOfPUSCHSets, FALSE);

  return offset;
}


static const per_sequence_t SetSpecificCauseList_PSCH_ReconfFailureTDD_sequence[] = {
  { &hf_nbap_unsuccessful_PDSCHSetList_PSCH_ReconfFailureTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Unsuccessful_PDSCHSetList_PSCH_ReconfFailureTDD },
  { &hf_nbap_unsuccessful_PUSCHSetList_PSCH_ReconfFailureTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_Unsuccessful_PUSCHSetList_PSCH_ReconfFailureTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SetSpecificCauseList_PSCH_ReconfFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SetSpecificCauseList_PSCH_ReconfFailureTDD, SetSpecificCauseList_PSCH_ReconfFailureTDD_sequence);

  return offset;
}



static int
dissect_nbap_Extension_CauseLevel_PSCH_ReconfFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string nbap_CauseLevel_PSCH_ReconfFailure_vals[] = {
  {   0, "generalCause" },
  {   1, "setSpecificCause" },
  {   2, "extension-CauseLevel-PSCH-ReconfFailure" },
  { 0, NULL }
};

static const per_choice_t CauseLevel_PSCH_ReconfFailure_choice[] = {
  {   0, &hf_nbap_generalCause_05, ASN1_EXTENSION_ROOT    , dissect_nbap_GeneralCauseList_PSCH_ReconfFailure },
  {   1, &hf_nbap_setSpecificCause, ASN1_EXTENSION_ROOT    , dissect_nbap_SetSpecificCauseList_PSCH_ReconfFailureTDD },
  {   2, &hf_nbap_extension_CauseLevel_PSCH_ReconfFailure, ASN1_NOT_EXTENSION_ROOT, dissect_nbap_Extension_CauseLevel_PSCH_ReconfFailure },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_CauseLevel_PSCH_ReconfFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_CauseLevel_PSCH_ReconfFailure, CauseLevel_PSCH_ReconfFailure_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Unsuccessful_PDSCHSetItem_PSCH_ReconfFailureTDD_sequence[] = {
  { &hf_nbap_pDSCHSet_ID    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PDSCHSet_ID },
  { &hf_nbap_cause          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Cause },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Unsuccessful_PDSCHSetItem_PSCH_ReconfFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Unsuccessful_PDSCHSetItem_PSCH_ReconfFailureTDD, Unsuccessful_PDSCHSetItem_PSCH_ReconfFailureTDD_sequence);

  return offset;
}


static const per_sequence_t Unsuccessful_PUSCHSetItem_PSCH_ReconfFailureTDD_sequence[] = {
  { &hf_nbap_pUSCHSet_ID    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_PUSCHSet_ID },
  { &hf_nbap_cause          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Cause },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Unsuccessful_PUSCHSetItem_PSCH_ReconfFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Unsuccessful_PUSCHSetItem_PSCH_ReconfFailureTDD, Unsuccessful_PUSCHSetItem_PSCH_ReconfFailureTDD_sequence);

  return offset;
}


static const per_sequence_t UARFCNSpecificCauseList_PSCH_ReconfFailureTDD_sequence_of[1] = {
  { &hf_nbap_UARFCNSpecificCauseList_PSCH_ReconfFailureTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_UARFCNSpecificCauseList_PSCH_ReconfFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_UARFCNSpecificCauseList_PSCH_ReconfFailureTDD, UARFCNSpecificCauseList_PSCH_ReconfFailureTDD_sequence_of,
                                                  0, maxFrequencyinCell, FALSE);

  return offset;
}


static const per_sequence_t Unsuccessful_UARFCNItem_PSCH_ReconfFailureTDD_sequence[] = {
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UARFCN },
  { &hf_nbap_cause          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Cause },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Unsuccessful_UARFCNItem_PSCH_ReconfFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Unsuccessful_UARFCNItem_PSCH_ReconfFailureTDD, Unsuccessful_UARFCNItem_PSCH_ReconfFailureTDD_sequence);

  return offset;
}


static const per_sequence_t E_HICH_TimeOffset_ReconfFailureTDD_sequence_of[1] = {
  { &hf_nbap_E_HICH_TimeOffset_ReconfFailureTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_E_HICH_TimeOffset_ReconfFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_HICH_TimeOffset_ReconfFailureTDD, E_HICH_TimeOffset_ReconfFailureTDD_sequence_of,
                                                  1, maxFrequencyinCell, FALSE);

  return offset;
}


static const per_sequence_t Common_System_Information_ResponseLCR_sequence[] = {
  { &hf_nbap_hSDSCH_Common_System_Information_ResponseLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_Common_System_Information_ResponseLCR },
  { &hf_nbap_hSDSCH_Paging_System_Information_ResponseLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_HSDSCH_Paging_System_Information_ResponseLCR },
  { &hf_nbap_common_EDCH_System_Information_ResponseLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Common_EDCH_System_Information_ResponseLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Common_System_Information_ResponseLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Common_System_Information_ResponseLCR, Common_System_Information_ResponseLCR_sequence);

  return offset;
}


static const per_sequence_t ResetRequest_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ResetRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ResetRequest, ResetRequest_sequence);

  return offset;
}


static const per_sequence_t CommunicationContextInfoList_Reset_sequence_of[1] = {
  { &hf_nbap_CommunicationContextInfoList_Reset_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_CommunicationContextInfoList_Reset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_CommunicationContextInfoList_Reset, CommunicationContextInfoList_Reset_sequence_of,
                                                  1, maxCommunicationContext, FALSE);

  return offset;
}


static const per_sequence_t CommunicationContextList_Reset_sequence[] = {
  { &hf_nbap_communicationContextInfoList_Reset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommunicationContextInfoList_Reset },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommunicationContextList_Reset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommunicationContextList_Reset, CommunicationContextList_Reset_sequence);

  return offset;
}


static const per_sequence_t CommunicationControlPortInfoList_Reset_sequence_of[1] = {
  { &hf_nbap_CommunicationControlPortInfoList_Reset_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_CommunicationControlPortInfoList_Reset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_CommunicationControlPortInfoList_Reset, CommunicationControlPortInfoList_Reset_sequence_of,
                                                  1, maxCCPinNodeB, FALSE);

  return offset;
}


static const per_sequence_t CommunicationControlPortList_Reset_sequence[] = {
  { &hf_nbap_communicationControlPortInfoList_Reset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommunicationControlPortInfoList_Reset },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommunicationControlPortList_Reset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommunicationControlPortList_Reset, CommunicationControlPortList_Reset_sequence);

  return offset;
}


static const value_string nbap_ResetIndicator_vals[] = {
  {   0, "communicationContext" },
  {   1, "communicationControlPort" },
  {   2, "nodeB" },
  { 0, NULL }
};

static const per_choice_t ResetIndicator_choice[] = {
  {   0, &hf_nbap_communicationContext, ASN1_EXTENSION_ROOT    , dissect_nbap_CommunicationContextList_Reset },
  {   1, &hf_nbap_communicationControlPort, ASN1_EXTENSION_ROOT    , dissect_nbap_CommunicationControlPortList_Reset },
  {   2, &hf_nbap_nodeB          , ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_ResetIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_ResetIndicator, ResetIndicator_choice,
                                 NULL);

  return offset;
}


static const value_string nbap_CommunicationContextType_Reset_vals[] = {
  {   0, "cRNC-CommunicationContextID" },
  {   1, "nodeB-CommunicationContextID" },
  { 0, NULL }
};

static const per_choice_t CommunicationContextType_Reset_choice[] = {
  {   0, &hf_nbap_cRNC_CommunicationContextID, ASN1_EXTENSION_ROOT    , dissect_nbap_CRNC_CommunicationContextID },
  {   1, &hf_nbap_nodeB_CommunicationContextID, ASN1_EXTENSION_ROOT    , dissect_nbap_NodeB_CommunicationContextID },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_CommunicationContextType_Reset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_CommunicationContextType_Reset, CommunicationContextType_Reset_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t CommunicationContextInfoItem_Reset_sequence[] = {
  { &hf_nbap_communicationContextType_Reset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommunicationContextType_Reset },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommunicationContextInfoItem_Reset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommunicationContextInfoItem_Reset, CommunicationContextInfoItem_Reset_sequence);

  return offset;
}


static const per_sequence_t CommunicationControlPortInfoItem_Reset_sequence[] = {
  { &hf_nbap_communicationControlPortID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CommunicationControlPortID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CommunicationControlPortInfoItem_Reset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CommunicationControlPortInfoItem_Reset, CommunicationControlPortInfoItem_Reset_sequence);

  return offset;
}


static const per_sequence_t ResetResponse_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_ResetResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_ResetResponse, ResetResponse_sequence);

  return offset;
}


static const per_sequence_t InformationExchangeInitiationRequest_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_InformationExchangeInitiationRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_InformationExchangeInitiationRequest, InformationExchangeInitiationRequest_sequence);

  return offset;
}


static const per_sequence_t Cell_InfEx_Rqst_sequence[] = {
  { &hf_nbap_c_ID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_C_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Cell_InfEx_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Cell_InfEx_Rqst, Cell_InfEx_Rqst_sequence);

  return offset;
}


static const value_string nbap_InformationExchangeObjectType_InfEx_Rqst_vals[] = {
  {   0, "cell" },
  { 0, NULL }
};

static const per_choice_t InformationExchangeObjectType_InfEx_Rqst_choice[] = {
  {   0, &hf_nbap_cell_03        , ASN1_EXTENSION_ROOT    , dissect_nbap_Cell_InfEx_Rqst },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_InformationExchangeObjectType_InfEx_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_InformationExchangeObjectType_InfEx_Rqst, InformationExchangeObjectType_InfEx_Rqst_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t InformationExchangeInitiationResponse_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_InformationExchangeInitiationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_InformationExchangeInitiationResponse, InformationExchangeInitiationResponse_sequence);

  return offset;
}


static const per_sequence_t Cell_InfEx_Rsp_sequence[] = {
  { &hf_nbap_requestedDataValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RequestedDataValue },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Cell_InfEx_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Cell_InfEx_Rsp, Cell_InfEx_Rsp_sequence);

  return offset;
}


static const value_string nbap_InformationExchangeObjectType_InfEx_Rsp_vals[] = {
  {   0, "cell" },
  { 0, NULL }
};

static const per_choice_t InformationExchangeObjectType_InfEx_Rsp_choice[] = {
  {   0, &hf_nbap_cell_04        , ASN1_EXTENSION_ROOT    , dissect_nbap_Cell_InfEx_Rsp },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_InformationExchangeObjectType_InfEx_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_InformationExchangeObjectType_InfEx_Rsp, InformationExchangeObjectType_InfEx_Rsp_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t InformationExchangeInitiationFailure_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_InformationExchangeInitiationFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_InformationExchangeInitiationFailure, InformationExchangeInitiationFailure_sequence);

  return offset;
}


static const per_sequence_t InformationReport_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_InformationReport(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_InformationReport, InformationReport_sequence);

  return offset;
}


static const per_sequence_t Cell_Inf_Rprt_sequence[] = {
  { &hf_nbap_requestedDataValueInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RequestedDataValueInformation },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Cell_Inf_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Cell_Inf_Rprt, Cell_Inf_Rprt_sequence);

  return offset;
}


static const value_string nbap_InformationExchangeObjectType_InfEx_Rprt_vals[] = {
  {   0, "cell" },
  { 0, NULL }
};

static const per_choice_t InformationExchangeObjectType_InfEx_Rprt_choice[] = {
  {   0, &hf_nbap_cell_05        , ASN1_EXTENSION_ROOT    , dissect_nbap_Cell_Inf_Rprt },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_InformationExchangeObjectType_InfEx_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_InformationExchangeObjectType_InfEx_Rprt, InformationExchangeObjectType_InfEx_Rprt_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t InformationExchangeTerminationRequest_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_InformationExchangeTerminationRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_InformationExchangeTerminationRequest, InformationExchangeTerminationRequest_sequence);

  return offset;
}


static const per_sequence_t InformationExchangeFailureIndication_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_InformationExchangeFailureIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_InformationExchangeFailureIndication, InformationExchangeFailureIndication_sequence);

  return offset;
}


static const per_sequence_t CellSynchronisationInitiationRequestTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSynchronisationInitiationRequestTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSynchronisationInitiationRequestTDD, CellSynchronisationInitiationRequestTDD_sequence);

  return offset;
}


static const per_sequence_t TimeslotInfo_CellSyncInitiationRqstTDD_sequence_of[1] = {
  { &hf_nbap_TimeslotInfo_CellSyncInitiationRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TimeSlot },
};

static int
dissect_nbap_TimeslotInfo_CellSyncInitiationRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_TimeslotInfo_CellSyncInitiationRqstTDD, TimeslotInfo_CellSyncInitiationRqstTDD_sequence_of,
                                                  1, 15, FALSE);

  return offset;
}


static const per_sequence_t CellSyncBurstTransInit_CellSyncInitiationRqstTDD_sequence[] = {
  { &hf_nbap_cSBTransmissionID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CSBTransmissionID },
  { &hf_nbap_sfn            , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SFN },
  { &hf_nbap_cellSyncBurstCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellSyncBurstCode },
  { &hf_nbap_cellSyncBurstCodeShift, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellSyncBurstCodeShift },
  { &hf_nbap_initialDLTransPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSyncBurstTransInit_CellSyncInitiationRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSyncBurstTransInit_CellSyncInitiationRqstTDD, CellSyncBurstTransInit_CellSyncInitiationRqstTDD_sequence);

  return offset;
}


static const per_sequence_t CellSyncBurstMeasureInit_CellSyncInitiationRqstTDD_sequence[] = {
  { &hf_nbap_cSBMeasurementID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CSBMeasurementID },
  { &hf_nbap_cellSyncBurstCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellSyncBurstCode },
  { &hf_nbap_cellSyncBurstCodeShift, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellSyncBurstCodeShift },
  { &hf_nbap_synchronisationReportType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SynchronisationReportType },
  { &hf_nbap_sfn            , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SFN },
  { &hf_nbap_synchronisationReportCharacteristics, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SynchronisationReportCharacteristics },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSyncBurstMeasureInit_CellSyncInitiationRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSyncBurstMeasureInit_CellSyncInitiationRqstTDD, CellSyncBurstMeasureInit_CellSyncInitiationRqstTDD_sequence);

  return offset;
}


static const per_sequence_t SYNCDlCodeId_TransInitLCR_CellSyncInitiationRqstTDD_sequence[] = {
  { &hf_nbap_cSBTransmissionID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CSBTransmissionID },
  { &hf_nbap_sfn            , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SFN },
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UARFCN },
  { &hf_nbap_sYNCDlCodeId   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SYNCDlCodeId },
  { &hf_nbap_dwPCH_Power    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DwPCH_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SYNCDlCodeId_TransInitLCR_CellSyncInitiationRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SYNCDlCodeId_TransInitLCR_CellSyncInitiationRqstTDD, SYNCDlCodeId_TransInitLCR_CellSyncInitiationRqstTDD_sequence);

  return offset;
}


static const per_sequence_t SYNCDlCodeId_MeasureInitLCR_CellSyncInitiationRqstTDD_sequence[] = {
  { &hf_nbap_cSBMeasurementID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CSBMeasurementID },
  { &hf_nbap_sfn            , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SFN },
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UARFCN },
  { &hf_nbap_sYNCDlCodeId   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SYNCDlCodeId },
  { &hf_nbap_synchronisationReportType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SynchronisationReportType },
  { &hf_nbap_synchronisationReportCharacteristics, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SynchronisationReportCharacteristics },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SYNCDlCodeId_MeasureInitLCR_CellSyncInitiationRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SYNCDlCodeId_MeasureInitLCR_CellSyncInitiationRqstTDD, SYNCDlCodeId_MeasureInitLCR_CellSyncInitiationRqstTDD_sequence);

  return offset;
}


static const per_sequence_t CellSynchronisationInitiationResponseTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSynchronisationInitiationResponseTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSynchronisationInitiationResponseTDD, CellSynchronisationInitiationResponseTDD_sequence);

  return offset;
}


static const per_sequence_t CellSynchronisationInitiationFailureTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSynchronisationInitiationFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSynchronisationInitiationFailureTDD, CellSynchronisationInitiationFailureTDD_sequence);

  return offset;
}


static const per_sequence_t CellSynchronisationReconfigurationRequestTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSynchronisationReconfigurationRequestTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSynchronisationReconfigurationRequestTDD, CellSynchronisationReconfigurationRequestTDD_sequence);

  return offset;
}


static const per_sequence_t CellSyncBurstTransInfoItem_CellSyncReconfRqstTDD_sequence[] = {
  { &hf_nbap_cSBTransmissionID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CSBTransmissionID },
  { &hf_nbap_syncFrameNumberToTransmit, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SyncFrameNumber },
  { &hf_nbap_cellSyncBurstCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CellSyncBurstCode },
  { &hf_nbap_cellSyncBurstCodeShift, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_CellSyncBurstCodeShift },
  { &hf_nbap_dlTransPower   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSyncBurstTransInfoItem_CellSyncReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSyncBurstTransInfoItem_CellSyncReconfRqstTDD, CellSyncBurstTransInfoItem_CellSyncReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t CellSyncBurstTransReconfInfo_CellSyncReconfRqstTDD_sequence_of[1] = {
  { &hf_nbap_CellSyncBurstTransReconfInfo_CellSyncReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_CellSyncBurstTransInfoItem_CellSyncReconfRqstTDD },
};

static int
dissect_nbap_CellSyncBurstTransReconfInfo_CellSyncReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_CellSyncBurstTransReconfInfo_CellSyncReconfRqstTDD, CellSyncBurstTransReconfInfo_CellSyncReconfRqstTDD_sequence_of,
                                                  1, maxNrOfCellSyncBursts, FALSE);

  return offset;
}



static int
dissect_nbap_CellSyncBurstMeasInfoList_CellSyncReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_SynchronisationReportTypeIE(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_SynchronisationReportCharacteristicsIE(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t CellSyncBurstMeasInfo_CellSyncReconfRqstTDD_sequence[] = {
  { &hf_nbap_cellSyncBurstMeasInfoList_CellSyncReconfRqstTDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellSyncBurstMeasInfoList_CellSyncReconfRqstTDD },
  { &hf_nbap_synchronisationReportType_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SynchronisationReportTypeIE },
  { &hf_nbap_synchronisationReportCharacteristics_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SynchronisationReportCharacteristicsIE },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSyncBurstMeasInfo_CellSyncReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSyncBurstMeasInfo_CellSyncReconfRqstTDD, CellSyncBurstMeasInfo_CellSyncReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t CellSyncBurstInfoItem_CellSyncReconfRqstTDD_sequence[] = {
  { &hf_nbap_cSBMeasurementID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CSBMeasurementID },
  { &hf_nbap_cellSyncBurstCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellSyncBurstCode },
  { &hf_nbap_cellSyncBurstCodeShift, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellSyncBurstCodeShift },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSyncBurstInfoItem_CellSyncReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSyncBurstInfoItem_CellSyncReconfRqstTDD, CellSyncBurstInfoItem_CellSyncReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t CellSyncBurstInfoList_CellSyncReconfRqstTDD_sequence_of[1] = {
  { &hf_nbap_CellSyncBurstInfoList_CellSyncReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_CellSyncBurstInfoItem_CellSyncReconfRqstTDD },
};

static int
dissect_nbap_CellSyncBurstInfoList_CellSyncReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_CellSyncBurstInfoList_CellSyncReconfRqstTDD, CellSyncBurstInfoList_CellSyncReconfRqstTDD_sequence_of,
                                                  1, maxNrOfReceptsPerSyncFrame, FALSE);

  return offset;
}


static const per_sequence_t CellSyncBurstMeasInfoItem_CellSyncReconfRqstTDD_sequence[] = {
  { &hf_nbap_syncFrameNrToReceive, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SyncFrameNumber },
  { &hf_nbap_syncBurstInfo  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellSyncBurstInfoList_CellSyncReconfRqstTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSyncBurstMeasInfoItem_CellSyncReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSyncBurstMeasInfoItem_CellSyncReconfRqstTDD, CellSyncBurstMeasInfoItem_CellSyncReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t CellSyncBurstMeasInfoListIE_CellSyncReconfRqstTDD_sequence_of[1] = {
  { &hf_nbap_CellSyncBurstMeasInfoListIE_CellSyncReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_CellSyncBurstMeasInfoItem_CellSyncReconfRqstTDD },
};

static int
dissect_nbap_CellSyncBurstMeasInfoListIE_CellSyncReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_CellSyncBurstMeasInfoListIE_CellSyncReconfRqstTDD, CellSyncBurstMeasInfoListIE_CellSyncReconfRqstTDD_sequence_of,
                                                  1, maxNrOfCellSyncBursts, FALSE);

  return offset;
}


static const per_sequence_t SYNCDlCodeIdTransReconfItemLCR_CellSyncReconfRqstTDD_sequence[] = {
  { &hf_nbap_cSBTransmissionID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CSBTransmissionID },
  { &hf_nbap_syncFrameNumberforTransmit, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SyncFrameNumber },
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UARFCN },
  { &hf_nbap_sYNCDlCodeId   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SYNCDlCodeId },
  { &hf_nbap_dwPCH_Power    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DwPCH_Power },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SYNCDlCodeIdTransReconfItemLCR_CellSyncReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SYNCDlCodeIdTransReconfItemLCR_CellSyncReconfRqstTDD, SYNCDlCodeIdTransReconfItemLCR_CellSyncReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t SYNCDlCodeIdTransReconfInfoLCR_CellSyncReconfRqstTDD_sequence_of[1] = {
  { &hf_nbap_SYNCDlCodeIdTransReconfInfoLCR_CellSyncReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_SYNCDlCodeIdTransReconfItemLCR_CellSyncReconfRqstTDD },
};

static int
dissect_nbap_SYNCDlCodeIdTransReconfInfoLCR_CellSyncReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_SYNCDlCodeIdTransReconfInfoLCR_CellSyncReconfRqstTDD, SYNCDlCodeIdTransReconfInfoLCR_CellSyncReconfRqstTDD_sequence_of,
                                                  1, maxNrOfSyncFramesLCR, FALSE);

  return offset;
}


static const per_sequence_t SYNCDlCodeIdInfoItemLCR_CellSyncReconfRqstTDD_sequence[] = {
  { &hf_nbap_cSBMeasurementID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CSBMeasurementID },
  { &hf_nbap_sYNCDlCodeId   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SYNCDlCodeId },
  { &hf_nbap_uARFCN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_UARFCN },
  { &hf_nbap_propagationDelayCompensation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TimingAdjustmentValueLCR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SYNCDlCodeIdInfoItemLCR_CellSyncReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SYNCDlCodeIdInfoItemLCR_CellSyncReconfRqstTDD, SYNCDlCodeIdInfoItemLCR_CellSyncReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t SYNCDlCodeIdInfoListLCR_CellSyncReconfRqstTDD_sequence_of[1] = {
  { &hf_nbap_SYNCDlCodeIdInfoListLCR_CellSyncReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_SYNCDlCodeIdInfoItemLCR_CellSyncReconfRqstTDD },
};

static int
dissect_nbap_SYNCDlCodeIdInfoListLCR_CellSyncReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_SYNCDlCodeIdInfoListLCR_CellSyncReconfRqstTDD, SYNCDlCodeIdInfoListLCR_CellSyncReconfRqstTDD_sequence_of,
                                                  1, maxNrOfReceptionsperSyncFrameLCR, FALSE);

  return offset;
}


static const per_sequence_t SYNCDlCodeIdMeasInfoItem_CellSyncReconfRqstTDD_sequence[] = {
  { &hf_nbap_syncFrameNrToReceive, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SyncFrameNumber },
  { &hf_nbap_sYNCDlCodeIdInfoLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SYNCDlCodeIdInfoListLCR_CellSyncReconfRqstTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SYNCDlCodeIdMeasInfoItem_CellSyncReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SYNCDlCodeIdMeasInfoItem_CellSyncReconfRqstTDD, SYNCDlCodeIdMeasInfoItem_CellSyncReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t SYNCDlCodeIdMeasInfoList_CellSyncReconfRqstTDD_sequence_of[1] = {
  { &hf_nbap_SYNCDlCodeIdMeasInfoList_CellSyncReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_SYNCDlCodeIdMeasInfoItem_CellSyncReconfRqstTDD },
};

static int
dissect_nbap_SYNCDlCodeIdMeasInfoList_CellSyncReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_SYNCDlCodeIdMeasInfoList_CellSyncReconfRqstTDD, SYNCDlCodeIdMeasInfoList_CellSyncReconfRqstTDD_sequence_of,
                                                  1, maxNrOfSyncDLCodesLCR, FALSE);

  return offset;
}


static const per_sequence_t SYNCDlCodeIdMeasInfoLCR_CellSyncReconfRqstTDD_sequence[] = {
  { &hf_nbap_sYNCDlCodeIdMeasInfoList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SYNCDlCodeIdMeasInfoList_CellSyncReconfRqstTDD },
  { &hf_nbap_synchronisationReportType, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SynchronisationReportType },
  { &hf_nbap_synchronisationReportCharacteristics, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SynchronisationReportCharacteristics },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SYNCDlCodeIdMeasInfoLCR_CellSyncReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SYNCDlCodeIdMeasInfoLCR_CellSyncReconfRqstTDD, SYNCDlCodeIdMeasInfoLCR_CellSyncReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t CellSynchronisationReconfigurationResponseTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSynchronisationReconfigurationResponseTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSynchronisationReconfigurationResponseTDD, CellSynchronisationReconfigurationResponseTDD_sequence);

  return offset;
}


static const per_sequence_t CellSynchronisationReconfigurationFailureTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSynchronisationReconfigurationFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSynchronisationReconfigurationFailureTDD, CellSynchronisationReconfigurationFailureTDD_sequence);

  return offset;
}


static const per_sequence_t CellSynchronisationAdjustmentRequestTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSynchronisationAdjustmentRequestTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSynchronisationAdjustmentRequestTDD, CellSynchronisationAdjustmentRequestTDD_sequence);

  return offset;
}


static const per_sequence_t CellAdjustmentInfo_SyncAdjustmentRqstTDD_sequence_of[1] = {
  { &hf_nbap_CellAdjustmentInfo_SyncAdjustmentRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_CellAdjustmentInfo_SyncAdjustmentRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_CellAdjustmentInfo_SyncAdjustmentRqstTDD, CellAdjustmentInfo_SyncAdjustmentRqstTDD_sequence_of,
                                                  1, maxCellinNodeB, FALSE);

  return offset;
}


static const per_sequence_t CellAdjustmentInfoItem_SyncAdjustmentRqstTDD_sequence[] = {
  { &hf_nbap_c_ID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_C_ID },
  { &hf_nbap_frameAdjustmentValue, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_FrameAdjustmentValue },
  { &hf_nbap_timingAdjustmentValue, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_TimingAdjustmentValue },
  { &hf_nbap_dLTransPower   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_DL_Power },
  { &hf_nbap_sfn            , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SFN },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellAdjustmentInfoItem_SyncAdjustmentRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellAdjustmentInfoItem_SyncAdjustmentRqstTDD, CellAdjustmentInfoItem_SyncAdjustmentRqstTDD_sequence);

  return offset;
}


static const per_sequence_t CellSynchronisationAdjustmentResponseTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSynchronisationAdjustmentResponseTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSynchronisationAdjustmentResponseTDD, CellSynchronisationAdjustmentResponseTDD_sequence);

  return offset;
}


static const per_sequence_t CellSynchronisationAdjustmentFailureTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSynchronisationAdjustmentFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSynchronisationAdjustmentFailureTDD, CellSynchronisationAdjustmentFailureTDD_sequence);

  return offset;
}


static const per_sequence_t GeneralCauseList_SyncAdjustmntFailureTDD_sequence[] = {
  { &hf_nbap_cause          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Cause },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_GeneralCauseList_SyncAdjustmntFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_GeneralCauseList_SyncAdjustmntFailureTDD, GeneralCauseList_SyncAdjustmntFailureTDD_sequence);

  return offset;
}


static const per_sequence_t Unsuccessful_cell_InformationRespList_SyncAdjustmntFailureTDD_sequence_of[1] = {
  { &hf_nbap_Unsuccessful_cell_InformationRespList_SyncAdjustmntFailureTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_Unsuccessful_cell_InformationRespList_SyncAdjustmntFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Unsuccessful_cell_InformationRespList_SyncAdjustmntFailureTDD, Unsuccessful_cell_InformationRespList_SyncAdjustmntFailureTDD_sequence_of,
                                                  1, maxCellinNodeB, FALSE);

  return offset;
}


static const per_sequence_t CellSpecificCauseList_SyncAdjustmntFailureTDD_sequence[] = {
  { &hf_nbap_unsuccessful_cell_InformationRespList_SyncAdjustmntFailureTDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Unsuccessful_cell_InformationRespList_SyncAdjustmntFailureTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSpecificCauseList_SyncAdjustmntFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSpecificCauseList_SyncAdjustmntFailureTDD, CellSpecificCauseList_SyncAdjustmntFailureTDD_sequence);

  return offset;
}


static const value_string nbap_CauseLevel_SyncAdjustmntFailureTDD_vals[] = {
  {   0, "generalCause" },
  {   1, "cellSpecificCause" },
  { 0, NULL }
};

static const per_choice_t CauseLevel_SyncAdjustmntFailureTDD_choice[] = {
  {   0, &hf_nbap_generalCause_06, ASN1_EXTENSION_ROOT    , dissect_nbap_GeneralCauseList_SyncAdjustmntFailureTDD },
  {   1, &hf_nbap_cellSpecificCause, ASN1_EXTENSION_ROOT    , dissect_nbap_CellSpecificCauseList_SyncAdjustmntFailureTDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_CauseLevel_SyncAdjustmntFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_CauseLevel_SyncAdjustmntFailureTDD, CauseLevel_SyncAdjustmntFailureTDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Unsuccessful_cell_InformationRespItem_SyncAdjustmntFailureTDD_sequence[] = {
  { &hf_nbap_c_ID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_C_ID },
  { &hf_nbap_cause          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Cause },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Unsuccessful_cell_InformationRespItem_SyncAdjustmntFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Unsuccessful_cell_InformationRespItem_SyncAdjustmntFailureTDD, Unsuccessful_cell_InformationRespItem_SyncAdjustmntFailureTDD_sequence);

  return offset;
}


static const per_sequence_t CellSynchronisationTerminationRequestTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSynchronisationTerminationRequestTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSynchronisationTerminationRequestTDD, CellSynchronisationTerminationRequestTDD_sequence);

  return offset;
}


static const per_sequence_t CellSynchronisationFailureIndicationTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSynchronisationFailureIndicationTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSynchronisationFailureIndicationTDD, CellSynchronisationFailureIndicationTDD_sequence);

  return offset;
}


static const per_sequence_t CellSynchronisationReportTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSynchronisationReportTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSynchronisationReportTDD, CellSynchronisationReportTDD_sequence);

  return offset;
}



static int
dissect_nbap_C_ID_IE_CellSyncReprtTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_nbap_SyncReportTypeIE_CellSyncReprtTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_nbap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t CellSyncInfoItemIE_CellSyncReprtTDD_sequence[] = {
  { &hf_nbap_c_ID_CellSyncReprtTDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_C_ID_IE_CellSyncReprtTDD },
  { &hf_nbap_syncReportType_CellSyncReprtTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_SyncReportTypeIE_CellSyncReprtTDD },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSyncInfoItemIE_CellSyncReprtTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSyncInfoItemIE_CellSyncReprtTDD, CellSyncInfoItemIE_CellSyncReprtTDD_sequence);

  return offset;
}


static const per_sequence_t CellSyncInfo_CellSyncReprtTDD_sequence_of[1] = {
  { &hf_nbap_CellSyncInfo_CellSyncReprtTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_CellSyncInfoItemIE_CellSyncReprtTDD },
};

static int
dissect_nbap_CellSyncInfo_CellSyncReprtTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_CellSyncInfo_CellSyncReprtTDD, CellSyncInfo_CellSyncReprtTDD_sequence_of,
                                                  1, maxCellinNodeB, FALSE);

  return offset;
}


static const per_sequence_t CellSyncBurstAvailable_CellSyncReprtTDD_sequence[] = {
  { &hf_nbap_cellSyncBurstTiming, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellSyncBurstTiming },
  { &hf_nbap_cellSyncBurstSIR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellSyncBurstSIR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSyncBurstAvailable_CellSyncReprtTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSyncBurstAvailable_CellSyncReprtTDD, CellSyncBurstAvailable_CellSyncReprtTDD_sequence);

  return offset;
}


static const value_string nbap_CellSyncBurstInfo_CellSyncReprtTDD_vals[] = {
  {   0, "cellSyncBurstAvailable" },
  {   1, "cellSyncBurstNotAvailable" },
  { 0, NULL }
};

static const per_choice_t CellSyncBurstInfo_CellSyncReprtTDD_choice[] = {
  {   0, &hf_nbap_cellSyncBurstAvailable, ASN1_EXTENSION_ROOT    , dissect_nbap_CellSyncBurstAvailable_CellSyncReprtTDD },
  {   1, &hf_nbap_cellSyncBurstNotAvailable, ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_CellSyncBurstInfo_CellSyncReprtTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_CellSyncBurstInfo_CellSyncReprtTDD, CellSyncBurstInfo_CellSyncReprtTDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t SEQUENCE_SIZE_1_maxNrOfReceptsPerSyncFrame_OF_CellSyncBurstInfo_CellSyncReprtTDD_sequence_of[1] = {
  { &hf_nbap_cellSyncBurstInfo_CellSyncReprtTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_CellSyncBurstInfo_CellSyncReprtTDD },
};

static int
dissect_nbap_SEQUENCE_SIZE_1_maxNrOfReceptsPerSyncFrame_OF_CellSyncBurstInfo_CellSyncReprtTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_SEQUENCE_SIZE_1_maxNrOfReceptsPerSyncFrame_OF_CellSyncBurstInfo_CellSyncReprtTDD, SEQUENCE_SIZE_1_maxNrOfReceptsPerSyncFrame_OF_CellSyncBurstInfo_CellSyncReprtTDD_sequence_of,
                                                  1, maxNrOfReceptsPerSyncFrame, FALSE);

  return offset;
}


static const per_sequence_t CellSyncBurstMeasInfoItem_CellSyncReprtTDD_sequence[] = {
  { &hf_nbap_sFN            , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SFN },
  { &hf_nbap_cellSyncBurstInfo_CellSyncReprtTDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SEQUENCE_SIZE_1_maxNrOfReceptsPerSyncFrame_OF_CellSyncBurstInfo_CellSyncReprtTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_CellSyncBurstMeasInfoItem_CellSyncReprtTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_CellSyncBurstMeasInfoItem_CellSyncReprtTDD, CellSyncBurstMeasInfoItem_CellSyncReprtTDD_sequence);

  return offset;
}


static const per_sequence_t CellSyncBurstMeasInfoList_CellSyncReprtTDD_sequence_of[1] = {
  { &hf_nbap_CellSyncBurstMeasInfoList_CellSyncReprtTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_CellSyncBurstMeasInfoItem_CellSyncReprtTDD },
};

static int
dissect_nbap_CellSyncBurstMeasInfoList_CellSyncReprtTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_CellSyncBurstMeasInfoList_CellSyncReprtTDD, CellSyncBurstMeasInfoList_CellSyncReprtTDD_sequence_of,
                                                  0, maxNrOfCellSyncBursts, FALSE);

  return offset;
}


static const per_sequence_t IntStdPhCellSyncInfo_CellSyncReprtTDD_sequence[] = {
  { &hf_nbap_cellSyncBurstMeasuredInfo, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellSyncBurstMeasInfoList_CellSyncReprtTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_IntStdPhCellSyncInfo_CellSyncReprtTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_IntStdPhCellSyncInfo_CellSyncReprtTDD, IntStdPhCellSyncInfo_CellSyncReprtTDD_sequence);

  return offset;
}


static const value_string nbap_SyncReportType_CellSyncReprtTDD_vals[] = {
  {   0, "intStdPhSyncInfo-CellSyncReprtTDD" },
  {   1, "lateEntrantCell" },
  {   2, "frequencyAcquisition" },
  { 0, NULL }
};

static const per_choice_t SyncReportType_CellSyncReprtTDD_choice[] = {
  {   0, &hf_nbap_intStdPhSyncInfo_CellSyncReprtTDD, ASN1_EXTENSION_ROOT    , dissect_nbap_IntStdPhCellSyncInfo_CellSyncReprtTDD },
  {   1, &hf_nbap_lateEntrantCell, ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  {   2, &hf_nbap_frequencyAcquisition, ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_SyncReportType_CellSyncReprtTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_SyncReportType_CellSyncReprtTDD, SyncReportType_CellSyncReprtTDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t SyncDLCodeIdAvailable_CellSyncReprtTDD_sequence[] = {
  { &hf_nbap_syncDLCodeIdTiming, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellSyncBurstTimingLCR },
  { &hf_nbap_syncDLCodeIdSIR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_CellSyncBurstSIR },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SyncDLCodeIdAvailable_CellSyncReprtTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SyncDLCodeIdAvailable_CellSyncReprtTDD, SyncDLCodeIdAvailable_CellSyncReprtTDD_sequence);

  return offset;
}


static const value_string nbap_SyncDLCodeIdItem_CellSyncReprtTDD_vals[] = {
  {   0, "syncDLCodeIdAvailable" },
  {   1, "syncDLCodeIDNotAvailable" },
  { 0, NULL }
};

static const per_choice_t SyncDLCodeIdItem_CellSyncReprtTDD_choice[] = {
  {   0, &hf_nbap_syncDLCodeIdAvailable, ASN1_EXTENSION_ROOT    , dissect_nbap_SyncDLCodeIdAvailable_CellSyncReprtTDD },
  {   1, &hf_nbap_syncDLCodeIDNotAvailable, ASN1_EXTENSION_ROOT    , dissect_nbap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_SyncDLCodeIdItem_CellSyncReprtTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_SyncDLCodeIdItem_CellSyncReprtTDD, SyncDLCodeIdItem_CellSyncReprtTDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t SyncDLCodeIdInfo_CellSyncReprtTDD_sequence_of[1] = {
  { &hf_nbap_SyncDLCodeIdInfo_CellSyncReprtTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_SyncDLCodeIdItem_CellSyncReprtTDD },
};

static int
dissect_nbap_SyncDLCodeIdInfo_CellSyncReprtTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_SyncDLCodeIdInfo_CellSyncReprtTDD, SyncDLCodeIdInfo_CellSyncReprtTDD_sequence_of,
                                                  1, maxNrOfReceptionsperSyncFrameLCR, FALSE);

  return offset;
}


static const per_sequence_t SyncDLCodeIdsMeasInfoItem_CellSyncReprtTDD_sequence[] = {
  { &hf_nbap_sFN            , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SFN },
  { &hf_nbap_syncDLCodeIdInfo_CellSyncReprtTDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_SyncDLCodeIdInfo_CellSyncReprtTDD },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SyncDLCodeIdsMeasInfoItem_CellSyncReprtTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SyncDLCodeIdsMeasInfoItem_CellSyncReprtTDD, SyncDLCodeIdsMeasInfoItem_CellSyncReprtTDD_sequence);

  return offset;
}


static const per_sequence_t SyncDLCodeIdsMeasInfoList_CellSyncReprtTDD_sequence_of[1] = {
  { &hf_nbap_SyncDLCodeIdsMeasInfoList_CellSyncReprtTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_SyncDLCodeIdsMeasInfoItem_CellSyncReprtTDD },
};

static int
dissect_nbap_SyncDLCodeIdsMeasInfoList_CellSyncReprtTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_SyncDLCodeIdsMeasInfoList_CellSyncReprtTDD, SyncDLCodeIdsMeasInfoList_CellSyncReprtTDD_sequence_of,
                                                  0, maxNrOfSyncFramesLCR, FALSE);

  return offset;
}


static const per_sequence_t BearerRearrangementIndication_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_BearerRearrangementIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"BearerRearrangementIndication ");


  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_BearerRearrangementIndication, BearerRearrangementIndication_sequence);

  return offset;
}


static const per_sequence_t DCH_RearrangeItem_Bearer_RearrangeInd_sequence[] = {
  { &hf_nbap_dCH_ID_01      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DCH_RearrangeItem_Bearer_RearrangeInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DCH_RearrangeItem_Bearer_RearrangeInd, DCH_RearrangeItem_Bearer_RearrangeInd_sequence);

  return offset;
}


static const per_sequence_t DCH_RearrangeList_Bearer_RearrangeInd_sequence_of[1] = {
  { &hf_nbap_DCH_RearrangeList_Bearer_RearrangeInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DCH_RearrangeItem_Bearer_RearrangeInd },
};

static int
dissect_nbap_DCH_RearrangeList_Bearer_RearrangeInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DCH_RearrangeList_Bearer_RearrangeInd, DCH_RearrangeList_Bearer_RearrangeInd_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t DSCH_RearrangeItem_Bearer_RearrangeInd_sequence[] = {
  { &hf_nbap_dSCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DSCH_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DSCH_RearrangeItem_Bearer_RearrangeInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DSCH_RearrangeItem_Bearer_RearrangeInd, DSCH_RearrangeItem_Bearer_RearrangeInd_sequence);

  return offset;
}


static const per_sequence_t DSCH_RearrangeList_Bearer_RearrangeInd_sequence_of[1] = {
  { &hf_nbap_DSCH_RearrangeList_Bearer_RearrangeInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_DSCH_RearrangeItem_Bearer_RearrangeInd },
};

static int
dissect_nbap_DSCH_RearrangeList_Bearer_RearrangeInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DSCH_RearrangeList_Bearer_RearrangeInd, DSCH_RearrangeList_Bearer_RearrangeInd_sequence_of,
                                                  1, maxNrOfDSCHs, FALSE);

  return offset;
}


static const per_sequence_t USCH_RearrangeItem_Bearer_RearrangeInd_sequence[] = {
  { &hf_nbap_uSCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_USCH_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_USCH_RearrangeItem_Bearer_RearrangeInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_USCH_RearrangeItem_Bearer_RearrangeInd, USCH_RearrangeItem_Bearer_RearrangeInd_sequence);

  return offset;
}


static const per_sequence_t USCH_RearrangeList_Bearer_RearrangeInd_sequence_of[1] = {
  { &hf_nbap_USCH_RearrangeList_Bearer_RearrangeInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_USCH_RearrangeItem_Bearer_RearrangeInd },
};

static int
dissect_nbap_USCH_RearrangeList_Bearer_RearrangeInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_USCH_RearrangeList_Bearer_RearrangeInd, USCH_RearrangeList_Bearer_RearrangeInd_sequence_of,
                                                  1, maxNrOfUSCHs, FALSE);

  return offset;
}


static const per_sequence_t HSDSCH_RearrangeItem_Bearer_RearrangeInd_sequence[] = {
  { &hf_nbap_hsDSCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_MACdFlow_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_HSDSCH_RearrangeItem_Bearer_RearrangeInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_HSDSCH_RearrangeItem_Bearer_RearrangeInd, HSDSCH_RearrangeItem_Bearer_RearrangeInd_sequence);

  return offset;
}


static const per_sequence_t HSDSCH_RearrangeList_Bearer_RearrangeInd_sequence_of[1] = {
  { &hf_nbap_HSDSCH_RearrangeList_Bearer_RearrangeInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_HSDSCH_RearrangeItem_Bearer_RearrangeInd },
};

static int
dissect_nbap_HSDSCH_RearrangeList_Bearer_RearrangeInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_HSDSCH_RearrangeList_Bearer_RearrangeInd, HSDSCH_RearrangeList_Bearer_RearrangeInd_sequence_of,
                                                  1, maxNrOfMACdFlows, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_RearrangeItem_Bearer_RearrangeInd_sequence[] = {
  { &hf_nbap_e_DCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_MACdFlow_ID },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_E_DCH_RearrangeItem_Bearer_RearrangeInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_E_DCH_RearrangeItem_Bearer_RearrangeInd, E_DCH_RearrangeItem_Bearer_RearrangeInd_sequence);

  return offset;
}


static const per_sequence_t E_DCH_RearrangeList_Bearer_RearrangeInd_sequence_of[1] = {
  { &hf_nbap_E_DCH_RearrangeList_Bearer_RearrangeInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_E_DCH_RearrangeItem_Bearer_RearrangeInd },
};

static int
dissect_nbap_E_DCH_RearrangeList_Bearer_RearrangeInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_E_DCH_RearrangeList_Bearer_RearrangeInd, E_DCH_RearrangeList_Bearer_RearrangeInd_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}


static const value_string nbap_Transport_Bearer_Rearrangement_Indicator_for_Additional_EDCH_Separate_Mode_vals[] = {
  {   0, "bearer-for-primary-carrier" },
  {   1, "bearer-for-secondary-carrier" },
  {   2, "bearers-for-both-primary-and-secondary-carriers" },
  { 0, NULL }
};


static int
dissect_nbap_Transport_Bearer_Rearrangement_Indicator_for_Additional_EDCH_Separate_Mode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_Bearer_Rearrangement_ItemIEs_sequence[] = {
  { &hf_nbap_transport_Bearer_Rearrangement_Indicator_for_Additional_EDCH_Separate_Mode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Transport_Bearer_Rearrangement_Indicator_for_Additional_EDCH_Separate_Mode },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_Cell_Information_Bearer_Rearrangement_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_Cell_Information_Bearer_Rearrangement_ItemIEs, Additional_EDCH_Cell_Information_Bearer_Rearrangement_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_Bearer_Rearrangement_List_sequence_of[1] = {
  { &hf_nbap_Additional_EDCH_Cell_Information_Bearer_Rearrangement_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_EDCH_Cell_Information_Bearer_Rearrangement_ItemIEs },
};

static int
dissect_nbap_Additional_EDCH_Cell_Information_Bearer_Rearrangement_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Additional_EDCH_Cell_Information_Bearer_Rearrangement_List, Additional_EDCH_Cell_Information_Bearer_Rearrangement_List_sequence_of,
                                                  1, maxNrOfEDCH_1, FALSE);

  return offset;
}


static const per_sequence_t RadioLinkActivationCommandFDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkActivationCommandFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkActivationCommandFDD, RadioLinkActivationCommandFDD_sequence);

  return offset;
}


static const per_sequence_t DelayedActivationInformationList_RL_ActivationCmdFDD_sequence_of[1] = {
  { &hf_nbap_DelayedActivationInformationList_RL_ActivationCmdFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_DelayedActivationInformationList_RL_ActivationCmdFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DelayedActivationInformationList_RL_ActivationCmdFDD, DelayedActivationInformationList_RL_ActivationCmdFDD_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t DelayedActivationInformation_RL_ActivationCmdFDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_delayed_activation_update, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DelayedActivationUpdate },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DelayedActivationInformation_RL_ActivationCmdFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DelayedActivationInformation_RL_ActivationCmdFDD, DelayedActivationInformation_RL_ActivationCmdFDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkActivationCommandTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkActivationCommandTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkActivationCommandTDD, RadioLinkActivationCommandTDD_sequence);

  return offset;
}


static const per_sequence_t DelayedActivationInformationList_RL_ActivationCmdTDD_sequence_of[1] = {
  { &hf_nbap_DelayedActivationInformationList_RL_ActivationCmdTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Single_Container },
};

static int
dissect_nbap_DelayedActivationInformationList_RL_ActivationCmdTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_DelayedActivationInformationList_RL_ActivationCmdTDD, DelayedActivationInformationList_RL_ActivationCmdTDD_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t DelayedActivationInformation_RL_ActivationCmdTDD_sequence[] = {
  { &hf_nbap_rL_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_delayed_activation_update, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_DelayedActivationUpdate },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_DelayedActivationInformation_RL_ActivationCmdTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_DelayedActivationInformation_RL_ActivationCmdTDD, DelayedActivationInformation_RL_ActivationCmdTDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkParameterUpdateIndicationFDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkParameterUpdateIndicationFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
col_set_str(actx->pinfo->cinfo, COL_INFO,"RadioLinkParameterUpdateIndication(FDD) ");
/* Node B -> CRNC */
actx->pinfo->link_dir=P2P_DIR_UL;


  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkParameterUpdateIndicationFDD, RadioLinkParameterUpdateIndicationFDD_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_RL_Param_Upd_ItemIEs_sequence[] = {
  { &hf_nbap_hSPDSCH_RL_ID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_RL_ID },
  { &hf_nbap_hS_DSCH_FDD_Secondary_Serving_Update_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_HS_DSCH_FDD_Secondary_Serving_Update_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_HS_Cell_Information_RL_Param_Upd_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_HS_Cell_Information_RL_Param_Upd_ItemIEs, Additional_HS_Cell_Information_RL_Param_Upd_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_RL_Param_Upd_sequence_of[1] = {
  { &hf_nbap_Additional_HS_Cell_Information_RL_Param_Upd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_HS_Cell_Information_RL_Param_Upd_ItemIEs },
};

static int
dissect_nbap_Additional_HS_Cell_Information_RL_Param_Upd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Additional_HS_Cell_Information_RL_Param_Upd, Additional_HS_Cell_Information_RL_Param_Upd_sequence_of,
                                                  1, maxNrOfHSDSCH_1, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_RL_Param_Upd_ItemIEs_sequence[] = {
  { &hf_nbap_additional_EDCH_FDD_Update_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_EDCH_FDD_Update_Information },
  { &hf_nbap_iE_Extensions  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Additional_EDCH_Cell_Information_RL_Param_Upd_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Additional_EDCH_Cell_Information_RL_Param_Upd_ItemIEs, Additional_EDCH_Cell_Information_RL_Param_Upd_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_RL_Param_Upd_sequence_of[1] = {
  { &hf_nbap_Additional_EDCH_Cell_Information_RL_Param_Upd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Additional_EDCH_Cell_Information_RL_Param_Upd_ItemIEs },
};

static int
dissect_nbap_Additional_EDCH_Cell_Information_RL_Param_Upd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_nbap_Additional_EDCH_Cell_Information_RL_Param_Upd, Additional_EDCH_Cell_Information_RL_Param_Upd_sequence_of,
                                                  1, maxNrOfEDCH_1, FALSE);

  return offset;
}


static const per_sequence_t RadioLinkParameterUpdateIndicationTDD_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_RadioLinkParameterUpdateIndicationTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_RadioLinkParameterUpdateIndicationTDD, RadioLinkParameterUpdateIndicationTDD_sequence);

  return offset;
}


static const per_sequence_t MBMSNotificationUpdateCommand_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_MBMSNotificationUpdateCommand(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_MBMSNotificationUpdateCommand, MBMSNotificationUpdateCommand_sequence);

  return offset;
}


static const per_sequence_t UEStatusUpdateCommand_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UEStatusUpdateCommand(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UEStatusUpdateCommand, UEStatusUpdateCommand_sequence);

  return offset;
}


static const per_sequence_t SecondaryULFrequencyReport_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SecondaryULFrequencyReport(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SecondaryULFrequencyReport, SecondaryULFrequencyReport_sequence);

  return offset;
}


static const per_sequence_t SecondaryULFrequencyUpdateIndication_sequence[] = {
  { &hf_nbap_protocolIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_nbap_ProtocolIE_Container },
  { &hf_nbap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_nbap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SecondaryULFrequencyUpdateIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SecondaryULFrequencyUpdateIndication, SecondaryULFrequencyUpdateIndication_sequence);

  return offset;
}



static int
dissect_nbap_InitiatingMessage_value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_open_type_pdu_new(tvb, offset, actx, tree, hf_index, dissect_InitiatingMessageValue);

  return offset;
}


static const per_sequence_t InitiatingMessage_sequence[] = {
  { &hf_nbap_procedureID    , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProcedureID },
  { &hf_nbap_criticality    , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Criticality },
  { &hf_nbap_messageDiscriminator, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_MessageDiscriminator },
  { &hf_nbap_transactionID  , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TransactionID },
  { &hf_nbap_initiatingMessagevalue, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_InitiatingMessage_value },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_InitiatingMessage(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_InitiatingMessage, InitiatingMessage_sequence);

  return offset;
}



static int
dissect_nbap_SuccessfulOutcome_value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_open_type_pdu_new(tvb, offset, actx, tree, hf_index, dissect_SuccessfulOutcomeValue);

  return offset;
}


static const per_sequence_t SuccessfulOutcome_sequence[] = {
  { &hf_nbap_procedureID    , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProcedureID },
  { &hf_nbap_criticality    , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Criticality },
  { &hf_nbap_messageDiscriminator, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_MessageDiscriminator },
  { &hf_nbap_transactionID  , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TransactionID },
  { &hf_nbap_successfulOutcome_value, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_SuccessfulOutcome_value },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_SuccessfulOutcome(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_SuccessfulOutcome, SuccessfulOutcome_sequence);

  return offset;
}



static int
dissect_nbap_UnsuccessfulOutcome_value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_open_type_pdu_new(tvb, offset, actx, tree, hf_index, dissect_UnsuccessfulOutcomeValue);

  return offset;
}


static const per_sequence_t UnsuccessfulOutcome_sequence[] = {
  { &hf_nbap_procedureID    , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProcedureID },
  { &hf_nbap_criticality    , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Criticality },
  { &hf_nbap_messageDiscriminator, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_MessageDiscriminator },
  { &hf_nbap_transactionID  , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TransactionID },
  { &hf_nbap_unsuccessfulOutcome_value, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_UnsuccessfulOutcome_value },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_UnsuccessfulOutcome(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_UnsuccessfulOutcome, UnsuccessfulOutcome_sequence);

  return offset;
}



static int
dissect_nbap_Outcome_value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_open_type(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}


static const per_sequence_t Outcome_sequence[] = {
  { &hf_nbap_procedureID    , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_ProcedureID },
  { &hf_nbap_criticality    , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Criticality },
  { &hf_nbap_messageDiscriminator, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_MessageDiscriminator },
  { &hf_nbap_transactionID  , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_TransactionID },
  { &hf_nbap_outcome_value  , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_nbap_Outcome_value },
  { NULL, 0, 0, NULL }
};

static int
dissect_nbap_Outcome(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_nbap_Outcome, Outcome_sequence);

  return offset;
}


static const value_string nbap_NBAP_PDU_vals[] = {
  {   0, "initiatingMessage" },
  {   1, "succesfulOutcome" },
  {   2, "unsuccesfulOutcome" },
  {   3, "outcome" },
  { 0, NULL }
};

static const per_choice_t NBAP_PDU_choice[] = {
  {   0, &hf_nbap_initiatingMessage, ASN1_EXTENSION_ROOT    , dissect_nbap_InitiatingMessage },
  {   1, &hf_nbap_succesfulOutcome, ASN1_EXTENSION_ROOT    , dissect_nbap_SuccessfulOutcome },
  {   2, &hf_nbap_unsuccesfulOutcome, ASN1_EXTENSION_ROOT    , dissect_nbap_UnsuccessfulOutcome },
  {   3, &hf_nbap_outcome        , ASN1_EXTENSION_ROOT    , dissect_nbap_Outcome },
  { 0, NULL, 0, NULL }
};

static int
dissect_nbap_NBAP_PDU(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_nbap_NBAP_PDU, NBAP_PDU_choice,
                                 NULL);

  return offset;
}

/*--- PDUs ---*/

static int dissect_ActivationInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ActivationInformation(tvb, offset, &asn1_ctx, tree, hf_nbap_ActivationInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_EDCH_Setup_Info_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Additional_EDCH_Setup_Info(tvb, offset, &asn1_ctx, tree, hf_nbap_Additional_EDCH_Setup_Info_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_EDCH_Cell_Information_Response_List_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Additional_EDCH_Cell_Information_Response_List(tvb, offset, &asn1_ctx, tree, hf_nbap_Additional_EDCH_Cell_Information_Response_List_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_EDCH_Cell_Information_Response_RLReconf_List_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Additional_EDCH_Cell_Information_Response_RLReconf_List(tvb, offset, &asn1_ctx, tree, hf_nbap_Additional_EDCH_Cell_Information_Response_RLReconf_List_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_EDCH_Cell_Information_Response_RL_Add_List_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Additional_EDCH_Cell_Information_Response_RL_Add_List(tvb, offset, &asn1_ctx, tree, hf_nbap_Additional_EDCH_Cell_Information_Response_RL_Add_List_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_AdditionalMeasurementValueList_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_AdditionalMeasurementValueList(tvb, offset, &asn1_ctx, tree, hf_nbap_AdditionalMeasurementValueList_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_AdditionalTimeSlotListLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_AdditionalTimeSlotListLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_AdditionalTimeSlotListLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Active_Pattern_Sequence_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Active_Pattern_Sequence_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_Active_Pattern_Sequence_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_AlternativeFormatReportingIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_AlternativeFormatReportingIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_AlternativeFormatReportingIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Angle_Of_Arrival_Value_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Angle_Of_Arrival_Value_LCR(tvb, offset, &asn1_ctx, tree, hf_nbap_Angle_Of_Arrival_Value_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_BCCH_ModificationTime_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_BCCH_ModificationTime(tvb, offset, &asn1_ctx, tree, hf_nbap_BCCH_ModificationTime_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Best_Cell_Portions_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Best_Cell_Portions_Value(tvb, offset, &asn1_ctx, tree, hf_nbap_Best_Cell_Portions_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Best_Cell_Portions_ValueLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Best_Cell_Portions_ValueLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_Best_Cell_Portions_ValueLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_BindingID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_BindingID(tvb, offset, &asn1_ctx, tree, hf_nbap_BindingID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_BlockingPriorityIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_BlockingPriorityIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_BlockingPriorityIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_BroadcastCommonTransportBearerIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_BroadcastCommonTransportBearerIndication(tvb, offset, &asn1_ctx, tree, hf_nbap_BroadcastCommonTransportBearerIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_BroadcastReference_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_BroadcastReference(tvb, offset, &asn1_ctx, tree, hf_nbap_BroadcastReference_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Cause_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Cause(tvb, offset, &asn1_ctx, tree, hf_nbap_Cause_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Cell_Capability_Container_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Cell_Capability_Container(tvb, offset, &asn1_ctx, tree, hf_nbap_Cell_Capability_Container_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Cell_ERNTI_Status_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Cell_ERNTI_Status_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_Cell_ERNTI_Status_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellParameterID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellParameterID(tvb, offset, &asn1_ctx, tree, hf_nbap_CellParameterID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellPortion_CapabilityLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellPortion_CapabilityLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_CellPortion_CapabilityLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellSyncBurstRepetitionPeriod_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellSyncBurstRepetitionPeriod(tvb, offset, &asn1_ctx, tree, hf_nbap_CellSyncBurstRepetitionPeriod_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CFN_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CFN(tvb, offset, &asn1_ctx, tree, hf_nbap_CFN_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_C_ID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_C_ID(tvb, offset, &asn1_ctx, tree, hf_nbap_C_ID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Closedlooptimingadjustmentmode_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Closedlooptimingadjustmentmode(tvb, offset, &asn1_ctx, tree, hf_nbap_Closedlooptimingadjustmentmode_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Common_EDCH_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Common_EDCH_Capability(tvb, offset, &asn1_ctx, tree, hf_nbap_Common_EDCH_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Common_E_DCH_HSDPCCH_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Common_E_DCH_HSDPCCH_Capability(tvb, offset, &asn1_ctx, tree, hf_nbap_Common_E_DCH_HSDPCCH_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Common_EDCH_System_InformationFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Common_EDCH_System_InformationFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Common_EDCH_System_InformationFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Common_EDCH_System_Information_ResponseFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Common_EDCH_System_Information_ResponseFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Common_EDCH_System_Information_ResponseFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_RNTI_List_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_RNTI_List(tvb, offset, &asn1_ctx, tree, hf_nbap_E_RNTI_List_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Common_MACFlows_to_DeleteFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Common_MACFlows_to_DeleteFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Common_MACFlows_to_DeleteFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonMeasurementAccuracy_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonMeasurementAccuracy(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonMeasurementAccuracy_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonMeasurementType_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonMeasurementType(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonMeasurementType_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonPhysicalChannelID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonPhysicalChannelID(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonPhysicalChannelID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonPhysicalChannelID768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonPhysicalChannelID768(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonPhysicalChannelID768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Common_PhysicalChannel_Status_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Common_PhysicalChannel_Status_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_Common_PhysicalChannel_Status_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Common_PhysicalChannel_Status_Information768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Common_PhysicalChannel_Status_Information768(tvb, offset, &asn1_ctx, tree, hf_nbap_Common_PhysicalChannel_Status_Information768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonTransportChannel_InformationResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonTransportChannel_InformationResponse(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonTransportChannel_InformationResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Common_TransportChannel_Status_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Common_TransportChannel_Status_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_Common_TransportChannel_Status_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommunicationControlPortID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommunicationControlPortID(tvb, offset, &asn1_ctx, tree, hf_nbap_CommunicationControlPortID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Compressed_Mode_Deactivation_Flag_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Compressed_Mode_Deactivation_Flag(tvb, offset, &asn1_ctx, tree, hf_nbap_Compressed_Mode_Deactivation_Flag_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ConfigurationGenerationID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ConfigurationGenerationID(tvb, offset, &asn1_ctx, tree, hf_nbap_ConfigurationGenerationID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ConstantValue_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ConstantValue(tvb, offset, &asn1_ctx, tree, hf_nbap_ConstantValue_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ContinuousPacketConnectivityDTX_DRX_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ContinuousPacketConnectivityDTX_DRX_Capability(tvb, offset, &asn1_ctx, tree, hf_nbap_ContinuousPacketConnectivityDTX_DRX_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ContinuousPacketConnectivityDTX_DRX_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ContinuousPacketConnectivityDTX_DRX_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_ContinuousPacketConnectivityDTX_DRX_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ContinuousPacketConnectivityHS_SCCH_less_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ContinuousPacketConnectivityHS_SCCH_less_Capability(tvb, offset, &asn1_ctx, tree, hf_nbap_ContinuousPacketConnectivityHS_SCCH_less_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ContinuousPacketConnectivityHS_SCCH_less_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ContinuousPacketConnectivityHS_SCCH_less_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_ContinuousPacketConnectivityHS_SCCH_less_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ContinuousPacketConnectivityHS_SCCH_less_Information_Response_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ContinuousPacketConnectivityHS_SCCH_less_Information_Response(tvb, offset, &asn1_ctx, tree, hf_nbap_ContinuousPacketConnectivityHS_SCCH_less_Information_Response_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ControlGAP_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ControlGAP(tvb, offset, &asn1_ctx, tree, hf_nbap_ControlGAP_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CPC_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CPC_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_CPC_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ContinuousPacketConnectivityHS_SCCH_less_Deactivate_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ContinuousPacketConnectivityHS_SCCH_less_Deactivate_Indicator(tvb, offset, &asn1_ctx, tree, hf_nbap_ContinuousPacketConnectivityHS_SCCH_less_Deactivate_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CriticalityDiagnostics_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CriticalityDiagnostics(tvb, offset, &asn1_ctx, tree, hf_nbap_CriticalityDiagnostics_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CRNC_CommunicationContextID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CRNC_CommunicationContextID(tvb, offset, &asn1_ctx, tree, hf_nbap_CRNC_CommunicationContextID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CSBMeasurementID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CSBMeasurementID(tvb, offset, &asn1_ctx, tree, hf_nbap_CSBMeasurementID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CSBTransmissionID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CSBTransmissionID(tvb, offset, &asn1_ctx, tree, hf_nbap_CSBTransmissionID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Common_EDCH_System_InformationLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Common_EDCH_System_InformationLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_Common_EDCH_System_InformationLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Common_EDCH_System_Information_ResponseLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Common_EDCH_System_Information_ResponseLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_Common_EDCH_System_Information_ResponseLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_Ext_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_Ext(tvb, offset, &asn1_ctx, tree, hf_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_Ext_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Common_MACFlows_to_DeleteLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Common_MACFlows_to_DeleteLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_Common_MACFlows_to_DeleteLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CPC_InformationLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CPC_InformationLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_CPC_InformationLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ContinuousPacketConnectivity_DRX_CapabilityLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ContinuousPacketConnectivity_DRX_CapabilityLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_ContinuousPacketConnectivity_DRX_CapabilityLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ContinuousPacketConnectivity_DRX_InformationLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ContinuousPacketConnectivity_DRX_InformationLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_ContinuousPacketConnectivity_DRX_InformationLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_AGCH_UE_Inactivity_Monitor_Threshold_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_AGCH_UE_Inactivity_Monitor_Threshold(tvb, offset, &asn1_ctx, tree, hf_nbap_E_AGCH_UE_Inactivity_Monitor_Threshold_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ContinuousPacketConnectivity_DRX_Information_ResponseLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ContinuousPacketConnectivity_DRX_Information_ResponseLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_ContinuousPacketConnectivity_DRX_Information_ResponseLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DCH_FDD_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DCH_FDD_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_DCH_FDD_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DCH_Indicator_For_E_DCH_HSDPA_Operation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DCH_Indicator_For_E_DCH_HSDPA_Operation(tvb, offset, &asn1_ctx, tree, hf_nbap_DCH_Indicator_For_E_DCH_HSDPA_Operation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DCH_InformationResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DCH_InformationResponse(tvb, offset, &asn1_ctx, tree, hf_nbap_DCH_InformationResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DCH_MeasurementOccasion_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DCH_MeasurementOccasion_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_DCH_MeasurementOccasion_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DCH_TDD_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DCH_TDD_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_DCH_TDD_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FDD_DCHs_to_Modify_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_FDD_DCHs_to_Modify(tvb, offset, &asn1_ctx, tree, hf_nbap_FDD_DCHs_to_Modify_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TDD_DCHs_to_Modify_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TDD_DCHs_to_Modify(tvb, offset, &asn1_ctx, tree, hf_nbap_TDD_DCHs_to_Modify_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DedicatedMeasurementType_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DedicatedMeasurementType(tvb, offset, &asn1_ctx, tree, hf_nbap_DedicatedMeasurementType_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DelayedActivation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DelayedActivation(tvb, offset, &asn1_ctx, tree, hf_nbap_DelayedActivation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DGANSS_Corrections_Req_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DGANSS_Corrections_Req(tvb, offset, &asn1_ctx, tree, hf_nbap_DGANSS_Corrections_Req_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DGNSS_ValidityPeriod_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DGNSS_ValidityPeriod(tvb, offset, &asn1_ctx, tree, hf_nbap_DGNSS_ValidityPeriod_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DiversityMode_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DiversityMode(tvb, offset, &asn1_ctx, tree, hf_nbap_DiversityMode_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_TimingAdjustment_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_DPCH_TimingAdjustment(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_DPCH_TimingAdjustment_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_Power_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_Power(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_Power_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_PowerBalancing_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_PowerBalancing_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_PowerBalancing_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_PowerBalancing_ActivationIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_PowerBalancing_ActivationIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_PowerBalancing_ActivationIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_PowerBalancing_UpdatedIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_PowerBalancing_UpdatedIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_PowerBalancing_UpdatedIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_ScramblingCode_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_ScramblingCode(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_ScramblingCode_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_TimeslotISCPInfo_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_TimeslotISCPInfo(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_TimeslotISCPInfo_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_TimeslotISCPInfoLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_TimeslotISCPInfoLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_TimeslotISCPInfoLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_TPC_Pattern01Count_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_TPC_Pattern01Count(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_TPC_Pattern01Count_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DLTransmissionBranchLoadValue_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DLTransmissionBranchLoadValue(tvb, offset, &asn1_ctx, tree, hf_nbap_DLTransmissionBranchLoadValue_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MaxHSDSCH_HSSCCH_Power_per_CELLPORTION_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MaxHSDSCH_HSSCCH_Power_per_CELLPORTION(tvb, offset, &asn1_ctx, tree, hf_nbap_MaxHSDSCH_HSSCCH_Power_per_CELLPORTION_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DPC_Mode_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DPC_Mode(tvb, offset, &asn1_ctx, tree, hf_nbap_DPC_Mode_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DPCH_ID768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DPCH_ID768(tvb, offset, &asn1_ctx, tree, hf_nbap_DPCH_ID768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DSCH_InformationResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DSCH_InformationResponse(tvb, offset, &asn1_ctx, tree, hf_nbap_DSCH_InformationResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DSCH_TDD_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DSCH_TDD_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_DSCH_TDD_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Dual_Band_Capability_Info_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Dual_Band_Capability_Info(tvb, offset, &asn1_ctx, tree, hf_nbap_Dual_Band_Capability_Info_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DwPCH_Power_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DwPCH_Power(tvb, offset, &asn1_ctx, tree, hf_nbap_DwPCH_Power_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_AGCH_Table_Choice_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_AGCH_Table_Choice(tvb, offset, &asn1_ctx, tree, hf_nbap_E_AGCH_Table_Choice_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_AGCH_FDD_Code_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_AGCH_FDD_Code_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_E_AGCH_FDD_Code_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_AI_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_AI_Capability(tvb, offset, &asn1_ctx, tree, hf_nbap_E_AI_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_Capability(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCHCapacityConsumptionLaw_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCHCapacityConsumptionLaw(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCHCapacityConsumptionLaw_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_TDD_CapacityConsumptionLaw_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_TDD_CapacityConsumptionLaw(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_TDD_CapacityConsumptionLaw_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_TTI2ms_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_TTI2ms_Capability(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_TTI2ms_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_SF_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_SF_Capability(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_SF_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_HARQ_Combining_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_HARQ_Combining_Capability(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_HARQ_Combining_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_FDD_DL_Control_Channel_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_FDD_DL_Control_Channel_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_FDD_DL_Control_Channel_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_FDD_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_FDD_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_FDD_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_FDD_Information_Response_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_FDD_Information_Response(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_FDD_Information_Response_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_FDD_Information_to_Modify_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_FDD_Information_to_Modify(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_FDD_Information_to_Modify_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_FDD_Update_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_FDD_Update_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_FDD_Update_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_DL_Control_Channel_Change_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_DL_Control_Channel_Change_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_DL_Control_Channel_Change_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_DL_Control_Channel_Grant_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_DL_Control_Channel_Grant_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_DL_Control_Channel_Grant_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_PowerOffset_for_SchedulingInfo_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_PowerOffset_for_SchedulingInfo(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_PowerOffset_for_SchedulingInfo_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_MACdPDU_SizeCapability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_MACdPDU_SizeCapability(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_MACdPDU_SizeCapability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_MACdPDUSizeFormat_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_MACdPDUSizeFormat(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_MACdPDUSizeFormat_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_MACdFlows_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_MACdFlows_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_MACdFlows_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_MACdFlows_to_Delete_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_MACdFlows_to_Delete(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_MACdFlows_to_Delete_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_Non_serving_Relative_Grant_Down_Commands_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_Non_serving_Relative_Grant_Down_Commands(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_Non_serving_Relative_Grant_Down_Commands_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Maximum_Target_ReceivedTotalWideBandPower_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Maximum_Target_ReceivedTotalWideBandPower(tvb, offset, &asn1_ctx, tree, hf_nbap_Maximum_Target_ReceivedTotalWideBandPower_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Target_NonServing_EDCH_To_Total_EDCH_Power_Ratio_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Target_NonServing_EDCH_To_Total_EDCH_Power_Ratio(tvb, offset, &asn1_ctx, tree, hf_nbap_Target_NonServing_EDCH_To_Total_EDCH_Power_Ratio_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_RL_Indication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_RL_Indication(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_RL_Indication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_Serving_Cell_Change_Info_Response_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_Serving_Cell_Change_Info_Response(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_Serving_Cell_Change_Info_Response_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EDCH_RACH_Report_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_EDCH_RACH_Report_Value(tvb, offset, &asn1_ctx, tree, hf_nbap_EDCH_RACH_Report_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DPDCH_PowerInterpolation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DPDCH_PowerInterpolation(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DPDCH_PowerInterpolation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_MACdFlows_to_DeleteLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_MACdFlows_to_DeleteLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_MACdFlows_to_DeleteLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Enhanced_UE_DRX_InformationLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Enhanced_UE_DRX_InformationLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_Enhanced_UE_DRX_InformationLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_End_Of_Audit_Sequence_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_End_Of_Audit_Sequence_Indicator(tvb, offset, &asn1_ctx, tree, hf_nbap_End_Of_Audit_Sequence_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_RGCH_E_HICH_FDD_Code_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_RGCH_E_HICH_FDD_Code_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_E_RGCH_E_HICH_FDD_Code_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_RNTI_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_RNTI(tvb, offset, &asn1_ctx, tree, hf_nbap_E_RNTI_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_TFCI_Boost_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_TFCI_Boost_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_E_TFCI_Boost_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCHProvidedBitRate_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCHProvidedBitRate(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCHProvidedBitRate_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCHProvidedBitRateValueInformation_For_CellPortion_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCHProvidedBitRateValueInformation_For_CellPortion(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCHProvidedBitRateValueInformation_For_CellPortion_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_HICH_TimeOffset_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_HICH_TimeOffset(tvb, offset, &asn1_ctx, tree, hf_nbap_E_HICH_TimeOffset_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_HICH_TimeOffsetLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_HICH_TimeOffsetLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_E_HICH_TimeOffsetLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_MACdFlow_Retransmission_Timer_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_MACdFlow_Retransmission_Timer(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_MACdFlow_Retransmission_Timer_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_Information_Response_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_Information_Response(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_Information_Response_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_Information_Reconfig_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_Information_Reconfig(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_Information_Reconfig_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Maximum_Generated_ReceivedTotalWideBandPowerInOtherCells_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Maximum_Generated_ReceivedTotalWideBandPowerInOtherCells(tvb, offset, &asn1_ctx, tree, hf_nbap_Maximum_Generated_ReceivedTotalWideBandPowerInOtherCells_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_768_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_768_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_768_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_768_Information_Reconfig_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_768_Information_Reconfig(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_768_Information_Reconfig_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_LCR_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_LCR_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_LCR_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_LCR_Information_Reconfig_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_LCR_Information_Reconfig(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_LCR_Information_Reconfig_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DormantModeIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DormantModeIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_DormantModeIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Enhanced_FACH_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Enhanced_FACH_Capability(tvb, offset, &asn1_ctx, tree, hf_nbap_Enhanced_FACH_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EnhancedHSServingCC_Abort_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_EnhancedHSServingCC_Abort(tvb, offset, &asn1_ctx, tree, hf_nbap_EnhancedHSServingCC_Abort_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Enhanced_PCH_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Enhanced_PCH_Capability(tvb, offset, &asn1_ctx, tree, hf_nbap_Enhanced_PCH_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Enhanced_UE_DRX_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Enhanced_UE_DRX_Capability(tvb, offset, &asn1_ctx, tree, hf_nbap_Enhanced_UE_DRX_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Enhanced_UE_DRX_InformationFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Enhanced_UE_DRX_InformationFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Enhanced_UE_DRX_InformationFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Extended_E_DCH_LCRTDD_PhysicalLayerCategory_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Extended_E_DCH_LCRTDD_PhysicalLayerCategory(tvb, offset, &asn1_ctx, tree, hf_nbap_Extended_E_DCH_LCRTDD_PhysicalLayerCategory_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Ext_Max_Bits_MACe_PDU_non_scheduled_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Ext_Max_Bits_MACe_PDU_non_scheduled(tvb, offset, &asn1_ctx, tree, hf_nbap_Ext_Max_Bits_MACe_PDU_non_scheduled_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Ext_Reference_E_TFCI_PO_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Ext_Reference_E_TFCI_PO(tvb, offset, &asn1_ctx, tree, hf_nbap_Ext_Reference_E_TFCI_PO_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ExtendedPropagationDelay_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ExtendedPropagationDelay(tvb, offset, &asn1_ctx, tree, hf_nbap_ExtendedPropagationDelay_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Extended_RNC_ID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Extended_RNC_ID(tvb, offset, &asn1_ctx, tree, hf_nbap_Extended_RNC_ID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Extended_Round_Trip_Time_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Extended_Round_Trip_Time_Value(tvb, offset, &asn1_ctx, tree, hf_nbap_Extended_Round_Trip_Time_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Extended_HS_SCCH_ID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Extended_HS_SCCH_ID(tvb, offset, &asn1_ctx, tree, hf_nbap_Extended_HS_SCCH_ID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Extended_HS_SICH_ID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Extended_HS_SICH_ID(tvb, offset, &asn1_ctx, tree, hf_nbap_Extended_HS_SICH_ID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Extended_E_HICH_ID_TDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Extended_E_HICH_ID_TDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Extended_E_HICH_ID_TDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_Semi_PersistentScheduling_Information_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_Semi_PersistentScheduling_Information_LCR(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_Semi_PersistentScheduling_Information_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Fast_Reconfiguration_Mode_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Fast_Reconfiguration_Mode(tvb, offset, &asn1_ctx, tree, hf_nbap_Fast_Reconfiguration_Mode_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Fast_Reconfiguration_Permission_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Fast_Reconfiguration_Permission(tvb, offset, &asn1_ctx, tree, hf_nbap_Fast_Reconfiguration_Permission_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FDD_S_CCPCH_FrameOffset_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_FDD_S_CCPCH_FrameOffset(tvb, offset, &asn1_ctx, tree, hf_nbap_FDD_S_CCPCH_FrameOffset_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_F_DPCH_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_F_DPCH_Capability(tvb, offset, &asn1_ctx, tree, hf_nbap_F_DPCH_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_F_DPCH_SlotFormat_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_F_DPCH_SlotFormat(tvb, offset, &asn1_ctx, tree, hf_nbap_F_DPCH_SlotFormat_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_F_DPCH_SlotFormatCapability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_F_DPCH_SlotFormatCapability(tvb, offset, &asn1_ctx, tree, hf_nbap_F_DPCH_SlotFormatCapability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FNReportingIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_FNReportingIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_FNReportingIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_AddIonoModelReq_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_GANSS_AddIonoModelReq(tvb, offset, &asn1_ctx, tree, hf_nbap_GANSS_AddIonoModelReq_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_AddNavigationModelsReq_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_GANSS_AddNavigationModelsReq(tvb, offset, &asn1_ctx, tree, hf_nbap_GANSS_AddNavigationModelsReq_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_AddUTCModelsReq_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_GANSS_AddUTCModelsReq(tvb, offset, &asn1_ctx, tree, hf_nbap_GANSS_AddUTCModelsReq_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_Additional_Ionospheric_Model_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_GANSS_Additional_Ionospheric_Model(tvb, offset, &asn1_ctx, tree, hf_nbap_GANSS_Additional_Ionospheric_Model_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_Additional_Navigation_Models_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_GANSS_Additional_Navigation_Models(tvb, offset, &asn1_ctx, tree, hf_nbap_GANSS_Additional_Navigation_Models_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_Additional_Time_Models_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_GANSS_Additional_Time_Models(tvb, offset, &asn1_ctx, tree, hf_nbap_GANSS_Additional_Time_Models_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_Additional_UTC_Models_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_GANSS_Additional_UTC_Models(tvb, offset, &asn1_ctx, tree, hf_nbap_GANSS_Additional_UTC_Models_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_ALM_ECEFsbasAlmanacSet_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_GANSS_ALM_ECEFsbasAlmanacSet(tvb, offset, &asn1_ctx, tree, hf_nbap_GANSS_ALM_ECEFsbasAlmanacSet_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_ALM_GlonassAlmanacSet_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_GANSS_ALM_GlonassAlmanacSet(tvb, offset, &asn1_ctx, tree, hf_nbap_GANSS_ALM_GlonassAlmanacSet_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_ALM_MidiAlmanacSet_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_GANSS_ALM_MidiAlmanacSet(tvb, offset, &asn1_ctx, tree, hf_nbap_GANSS_ALM_MidiAlmanacSet_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_ALM_NAVKeplerianSet_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_GANSS_ALM_NAVKeplerianSet(tvb, offset, &asn1_ctx, tree, hf_nbap_GANSS_ALM_NAVKeplerianSet_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_ALM_ReducedKeplerianSet_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_GANSS_ALM_ReducedKeplerianSet(tvb, offset, &asn1_ctx, tree, hf_nbap_GANSS_ALM_ReducedKeplerianSet_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_Auxiliary_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_GANSS_Auxiliary_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_GANSS_Auxiliary_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_AuxInfoReq_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_GANSS_AuxInfoReq(tvb, offset, &asn1_ctx, tree, hf_nbap_GANSS_AuxInfoReq_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_Common_Data_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_GANSS_Common_Data(tvb, offset, &asn1_ctx, tree, hf_nbap_GANSS_Common_Data_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_Earth_Orientation_Parameters_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_GANSS_Earth_Orientation_Parameters(tvb, offset, &asn1_ctx, tree, hf_nbap_GANSS_Earth_Orientation_Parameters_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_EarthOrientParaReq_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_GANSS_EarthOrientParaReq(tvb, offset, &asn1_ctx, tree, hf_nbap_GANSS_EarthOrientParaReq_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_Generic_Data_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_GANSS_Generic_Data(tvb, offset, &asn1_ctx, tree, hf_nbap_GANSS_Generic_Data_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_ID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_GANSS_ID(tvb, offset, &asn1_ctx, tree, hf_nbap_GANSS_ID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_GANSS_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_GANSS_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_SBAS_ID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_GANSS_SBAS_ID(tvb, offset, &asn1_ctx, tree, hf_nbap_GANSS_SBAS_ID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_Time_ID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_GANSS_Time_ID(tvb, offset, &asn1_ctx, tree, hf_nbap_GANSS_Time_ID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HARQ_MemoryPartitioningInfoExtForMIMO_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HARQ_MemoryPartitioningInfoExtForMIMO(tvb, offset, &asn1_ctx, tree, hf_nbap_HARQ_MemoryPartitioningInfoExtForMIMO_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HARQ_Preamble_Mode_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HARQ_Preamble_Mode(tvb, offset, &asn1_ctx, tree, hf_nbap_HARQ_Preamble_Mode_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HARQ_Preamble_Mode_Activation_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HARQ_Preamble_Mode_Activation_Indicator(tvb, offset, &asn1_ctx, tree, hf_nbap_HARQ_Preamble_Mode_Activation_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDPA_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDPA_Capability(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDPA_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_DSCHProvidedBitRate_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_DSCHProvidedBitRate(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_DSCHProvidedBitRate_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_DSCHProvidedBitRateValueInformation_For_CellPortion_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortion(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortion_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_DSCHRequiredPower_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_DSCHRequiredPower(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_DSCHRequiredPower_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_DSCHRequiredPowerValue_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_DSCHRequiredPowerValue(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_DSCHRequiredPowerValue_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_DSCHRequiredPowerValueInformation_For_CellPortion_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortion(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortion_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_Common_System_InformationFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_Common_System_InformationFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_Common_System_InformationFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_Common_System_Information_ResponseFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_Common_System_Information_ResponseFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_Common_System_Information_ResponseFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_FDD_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_FDD_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_FDD_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_TDD_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_TDD_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_TDD_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_Information_to_Modify_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_Information_to_Modify(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_Information_to_Modify_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_MACdPDUSizeFormat_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_MACdPDUSizeFormat(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_MACdPDUSizeFormat_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_MACdPDU_SizeCapability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_MACdPDU_SizeCapability(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_MACdPDU_SizeCapability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_Information_to_Modify_Unsynchronised_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_Information_to_Modify_Unsynchronised(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_Information_to_Modify_Unsynchronised_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_FDD_Information_Response_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_FDD_Information_Response(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_FDD_Information_Response_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_Paging_System_InformationFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_Paging_System_InformationFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_Paging_System_InformationFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_Paging_System_Information_ResponseFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_Paging_System_Information_ResponseFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_Paging_System_Information_ResponseFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_TDD_Information_Response_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_TDD_Information_Response(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_TDD_Information_Response_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_MACdFlows_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_MACdFlows_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_MACdFlows_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_MACdFlows_to_Delete_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_MACdFlows_to_Delete(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_MACdFlows_to_Delete_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_TBSizeTableIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_TBSizeTableIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_TBSizeTableIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_PreconfigurationInfo_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_PreconfigurationInfo(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_PreconfigurationInfo_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_EDCH_Preconfiguration_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Additional_EDCH_Preconfiguration_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_Additional_EDCH_Preconfiguration_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_PreconfigurationSetup_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_PreconfigurationSetup(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_PreconfigurationSetup_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSSCCH_Specific_InformationRespListTDD768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSSCCH_Specific_InformationRespListTDD768(tvb, offset, &asn1_ctx, tree, hf_nbap_HSSCCH_Specific_InformationRespListTDD768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_SICH_Reception_Quality_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_SICH_Reception_Quality_Value(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_SICH_Reception_Quality_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_SICH_failed_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_SICH_failed(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_SICH_failed_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_SICH_missed_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_SICH_missed(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_SICH_missed_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_SICH_total_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_SICH_total(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_SICH_total_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_SICH_Reception_Quality_Measurement_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_SICH_Reception_Quality_Measurement_Value(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_SICH_Reception_Quality_Measurement_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_RNTI_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_RNTI(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_RNTI_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_PDSCH_FDD_Code_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_PDSCH_FDD_Code_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_PDSCH_FDD_Code_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_SICH_ID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_SICH_ID(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_SICH_ID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_SCCH_FDD_Code_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_SCCH_FDD_Code_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_SCCH_FDD_Code_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_PDSCH_Code_Change_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_PDSCH_Code_Change_Indicator(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_PDSCH_Code_Change_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_PDSCH_Code_Change_Grant_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_PDSCH_Code_Change_Grant(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_PDSCH_Code_Change_Grant_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_Configured_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_Configured_Indicator(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_Configured_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_DSCH_Serving_Cell_Change_Info_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_DSCH_Serving_Cell_Change_Info(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_DSCH_Serving_Cell_Change_Info_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_DSCH_Serving_Cell_Change_Info_Response_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_DSCH_Serving_Cell_Change_Info_Response(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_DSCH_Serving_Cell_Change_Info_Response_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_FDD_Update_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_FDD_Update_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_FDD_Update_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_TDD_Update_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_TDD_Update_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_TDD_Update_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_Common_System_InformationLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_Common_System_InformationLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_Common_System_InformationLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_Common_System_Information_ResponseLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_Common_System_Information_ResponseLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_Common_System_Information_ResponseLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_Paging_System_InformationLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_Paging_System_InformationLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_Paging_System_InformationLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_Paging_System_Information_ResponseLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_Paging_System_Information_ResponseLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_Paging_System_Information_ResponseLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_DSCH_Semi_PersistentScheduling_Information_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_DSCH_Semi_PersistentScheduling_Information_LCR(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_DSCH_Semi_PersistentScheduling_Information_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RepetitionPeriodIndex_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RepetitionPeriodIndex(tvb, offset, &asn1_ctx, tree, hf_nbap_RepetitionPeriodIndex_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SPS_Reservation_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SPS_Reservation_Indicator(tvb, offset, &asn1_ctx, tree, hf_nbap_SPS_Reservation_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_DSCH_SPS_Operation_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_DSCH_SPS_Operation_Indicator(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_DSCH_SPS_Operation_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Add_To_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Add_To_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_Add_To_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext(tvb, offset, &asn1_ctx, tree, hf_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Modify_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext(tvb, offset, &asn1_ctx, tree, hf_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext(tvb, offset, &asn1_ctx, tree, hf_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MIMO_ReferenceSignal_InformationListLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MIMO_ReferenceSignal_InformationListLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_MIMO_ReferenceSignal_InformationListLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSSICH_ReferenceSignal_InformationLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSSICH_ReferenceSignal_InformationLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_HSSICH_ReferenceSignal_InformationLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSSICH_ReferenceSignal_InformationModifyLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSSICH_ReferenceSignal_InformationModifyLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_HSSICH_ReferenceSignal_InformationModifyLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_IMB_Parameters_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_IMB_Parameters(tvb, offset, &asn1_ctx, tree, hf_nbap_IMB_Parameters_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationReportCharacteristics_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_InformationReportCharacteristics(tvb, offset, &asn1_ctx, tree, hf_nbap_InformationReportCharacteristics_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationExchangeID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_InformationExchangeID(tvb, offset, &asn1_ctx, tree, hf_nbap_InformationExchangeID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationType_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_InformationType(tvb, offset, &asn1_ctx, tree, hf_nbap_InformationType_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Initial_DL_DPCH_TimingAdjustment_Allowed_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Initial_DL_DPCH_TimingAdjustment_Allowed(tvb, offset, &asn1_ctx, tree, hf_nbap_Initial_DL_DPCH_TimingAdjustment_Allowed_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InnerLoopDLPCStatus_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_InnerLoopDLPCStatus(tvb, offset, &asn1_ctx, tree, hf_nbap_InnerLoopDLPCStatus_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_IPMulticastIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_IPMulticastIndication(tvb, offset, &asn1_ctx, tree, hf_nbap_IPMulticastIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_IPMulticastDataBearerIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_IPMulticastDataBearerIndication(tvb, offset, &asn1_ctx, tree, hf_nbap_IPMulticastDataBearerIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_IdleIntervalInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_IdleIntervalInformation(tvb, offset, &asn1_ctx, tree, hf_nbap_IdleIntervalInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Local_Cell_ID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Local_Cell_ID(tvb, offset, &asn1_ctx, tree, hf_nbap_Local_Cell_ID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_LCRTDD_Uplink_Physical_Channel_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_LCRTDD_Uplink_Physical_Channel_Capability(tvb, offset, &asn1_ctx, tree, hf_nbap_LCRTDD_Uplink_Physical_Channel_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MAC_PDU_SizeExtended_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MAC_PDU_SizeExtended(tvb, offset, &asn1_ctx, tree, hf_nbap_MAC_PDU_SizeExtended_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MACes_Maximum_Bitrate_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MACes_Maximum_Bitrate_LCR(tvb, offset, &asn1_ctx, tree, hf_nbap_MACes_Maximum_Bitrate_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Maximum_Number_of_Retransmissions_For_E_DCH_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Maximum_Number_of_Retransmissions_For_E_DCH(tvb, offset, &asn1_ctx, tree, hf_nbap_Maximum_Number_of_Retransmissions_For_E_DCH_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Maximum_Target_ReceivedTotalWideBandPower_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Maximum_Target_ReceivedTotalWideBandPower_LCR(tvb, offset, &asn1_ctx, tree, hf_nbap_Maximum_Target_ReceivedTotalWideBandPower_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MaximumTransmissionPower_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MaximumTransmissionPower(tvb, offset, &asn1_ctx, tree, hf_nbap_MaximumTransmissionPower_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Max_UE_DTX_Cycle_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Max_UE_DTX_Cycle(tvb, offset, &asn1_ctx, tree, hf_nbap_Max_UE_DTX_Cycle_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MBMS_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MBMS_Capability(tvb, offset, &asn1_ctx, tree, hf_nbap_MBMS_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MeasurementFilterCoefficient_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MeasurementFilterCoefficient(tvb, offset, &asn1_ctx, tree, hf_nbap_MeasurementFilterCoefficient_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MeasurementID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MeasurementID(tvb, offset, &asn1_ctx, tree, hf_nbap_MeasurementID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MeasurementRecoveryBehavior_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MeasurementRecoveryBehavior(tvb, offset, &asn1_ctx, tree, hf_nbap_MeasurementRecoveryBehavior_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MeasurementRecoveryReportingIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MeasurementRecoveryReportingIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_MeasurementRecoveryReportingIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MeasurementRecoverySupportIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MeasurementRecoverySupportIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_MeasurementRecoverySupportIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MessageStructure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MessageStructure(tvb, offset, &asn1_ctx, tree, hf_nbap_MessageStructure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MICH_CFN_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MICH_CFN(tvb, offset, &asn1_ctx, tree, hf_nbap_MICH_CFN_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MidambleShiftLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MidambleShiftLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_MidambleShiftLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MIMO_ActivationIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MIMO_ActivationIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_MIMO_ActivationIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MIMO_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MIMO_Capability(tvb, offset, &asn1_ctx, tree, hf_nbap_MIMO_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MIMO_Mode_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MIMO_Mode_Indicator(tvb, offset, &asn1_ctx, tree, hf_nbap_MIMO_Mode_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MIMO_N_M_Ratio_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MIMO_N_M_Ratio(tvb, offset, &asn1_ctx, tree, hf_nbap_MIMO_N_M_Ratio_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MIMO_PilotConfiguration_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MIMO_PilotConfiguration(tvb, offset, &asn1_ctx, tree, hf_nbap_MIMO_PilotConfiguration_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MIMO_PilotConfigurationExtension_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MIMO_PilotConfigurationExtension(tvb, offset, &asn1_ctx, tree, hf_nbap_MIMO_PilotConfigurationExtension_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MIMO_PowerOffsetForS_CPICHCapability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MIMO_PowerOffsetForS_CPICHCapability(tvb, offset, &asn1_ctx, tree, hf_nbap_MIMO_PowerOffsetForS_CPICHCapability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MinimumReducedE_DPDCH_GainFactor_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MinimumReducedE_DPDCH_GainFactor(tvb, offset, &asn1_ctx, tree, hf_nbap_MinimumReducedE_DPDCH_GainFactor_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Modification_Period_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Modification_Period(tvb, offset, &asn1_ctx, tree, hf_nbap_Modification_Period_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MAChs_ResetIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MAChs_ResetIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_MAChs_ResetIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ModulationMBSFN_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ModulationMBSFN(tvb, offset, &asn1_ctx, tree, hf_nbap_ModulationMBSFN_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ModulationPO_MBSFN_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ModulationPO_MBSFN(tvb, offset, &asn1_ctx, tree, hf_nbap_ModulationPO_MBSFN_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MBSFN_Only_Mode_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MBSFN_Only_Mode_Indicator(tvb, offset, &asn1_ctx, tree, hf_nbap_MBSFN_Only_Mode_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MBSFN_Only_Mode_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MBSFN_Only_Mode_Capability(tvb, offset, &asn1_ctx, tree, hf_nbap_MBSFN_Only_Mode_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multicarrier_Number_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Multicarrier_Number(tvb, offset, &asn1_ctx, tree, hf_nbap_Multicarrier_Number_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MultipleFreq_HARQ_MemoryPartitioning_InformationList_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MultipleFreq_HARQ_MemoryPartitioning_InformationList(tvb, offset, &asn1_ctx, tree, hf_nbap_MultipleFreq_HARQ_MemoryPartitioning_InformationList_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multi_Cell_Capability_Info_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Multi_Cell_Capability_Info(tvb, offset, &asn1_ctx, tree, hf_nbap_Multi_Cell_Capability_Info_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multicell_EDCH_InformationItemIEs_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Multicell_EDCH_InformationItemIEs(tvb, offset, &asn1_ctx, tree, hf_nbap_Multicell_EDCH_InformationItemIEs_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multicell_EDCH_RL_Specific_InformationItemIEs_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Multicell_EDCH_RL_Specific_InformationItemIEs(tvb, offset, &asn1_ctx, tree, hf_nbap_Multicell_EDCH_RL_Specific_InformationItemIEs_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MIMO_SFMode_For_HSPDSCHDualStream_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MIMO_SFMode_For_HSPDSCHDualStream(tvb, offset, &asn1_ctx, tree, hf_nbap_MIMO_SFMode_For_HSPDSCHDualStream_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NCyclesPerSFNperiod_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_NCyclesPerSFNperiod(tvb, offset, &asn1_ctx, tree, hf_nbap_NCyclesPerSFNperiod_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NRepetitionsPerCyclePeriod_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_NRepetitionsPerCyclePeriod(tvb, offset, &asn1_ctx, tree, hf_nbap_NRepetitionsPerCyclePeriod_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NeighbouringCellMeasurementInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_NeighbouringCellMeasurementInformation(tvb, offset, &asn1_ctx, tree, hf_nbap_NeighbouringCellMeasurementInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NeighbouringTDDCellMeasurementInformationLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_NeighbouringTDDCellMeasurementInformationLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_NeighbouringTDDCellMeasurementInformationLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NeighbouringTDDCellMeasurementInformation768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_NeighbouringTDDCellMeasurementInformation768(tvb, offset, &asn1_ctx, tree, hf_nbap_NeighbouringTDDCellMeasurementInformation768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NonCellSpecificTxDiversity_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_NonCellSpecificTxDiversity(tvb, offset, &asn1_ctx, tree, hf_nbap_NonCellSpecificTxDiversity_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NI_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_NI_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_NI_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NodeB_CommunicationContextID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_NodeB_CommunicationContextID(tvb, offset, &asn1_ctx, tree, hf_nbap_NodeB_CommunicationContextID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NumberOfReportedCellPortions_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_NumberOfReportedCellPortions(tvb, offset, &asn1_ctx, tree, hf_nbap_NumberOfReportedCellPortions_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NumberOfReportedCellPortionsLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_NumberOfReportedCellPortionsLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_NumberOfReportedCellPortionsLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NSubCyclesPerCyclePeriod_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_NSubCyclesPerCyclePeriod(tvb, offset, &asn1_ctx, tree, hf_nbap_NSubCyclesPerCyclePeriod_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Number_Of_Supported_Carriers_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Number_Of_Supported_Carriers(tvb, offset, &asn1_ctx, tree, hf_nbap_Number_Of_Supported_Carriers_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NoOfTargetCellHS_SCCH_Order_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_NoOfTargetCellHS_SCCH_Order(tvb, offset, &asn1_ctx, tree, hf_nbap_NoOfTargetCellHS_SCCH_Order_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Out_of_Sychronization_Window_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Out_of_Sychronization_Window(tvb, offset, &asn1_ctx, tree, hf_nbap_Out_of_Sychronization_Window_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Paging_MACFlows_to_DeleteFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Paging_MACFlows_to_DeleteFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Paging_MACFlows_to_DeleteFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PhysicalChannelID_for_CommonERNTI_RequestedIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PhysicalChannelID_for_CommonERNTI_RequestedIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_PhysicalChannelID_for_CommonERNTI_RequestedIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PLCCHinformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PLCCHinformation(tvb, offset, &asn1_ctx, tree, hf_nbap_PLCCHinformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PowerAdjustmentType_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PowerAdjustmentType(tvb, offset, &asn1_ctx, tree, hf_nbap_PowerAdjustmentType_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PrecodingWeightSetRestriction_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PrecodingWeightSetRestriction(tvb, offset, &asn1_ctx, tree, hf_nbap_PrecodingWeightSetRestriction_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Primary_CPICH_Usage_for_Channel_Estimation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Primary_CPICH_Usage_for_Channel_Estimation(tvb, offset, &asn1_ctx, tree, hf_nbap_Primary_CPICH_Usage_for_Channel_Estimation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PrimaryScramblingCode_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PrimaryScramblingCode(tvb, offset, &asn1_ctx, tree, hf_nbap_PrimaryScramblingCode_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PrimaryCCPCH_RSCP_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PrimaryCCPCH_RSCP(tvb, offset, &asn1_ctx, tree, hf_nbap_PrimaryCCPCH_RSCP_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PrimaryCCPCH_RSCP_Delta_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PrimaryCCPCH_RSCP_Delta(tvb, offset, &asn1_ctx, tree, hf_nbap_PrimaryCCPCH_RSCP_Delta_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UE_Selected_MBMS_Service_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UE_Selected_MBMS_Service_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_UE_Selected_MBMS_Service_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Paging_MACFlows_to_DeleteLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Paging_MACFlows_to_DeleteLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_Paging_MACFlows_to_DeleteLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Specific_DCH_Info_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_Specific_DCH_Info(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_Specific_DCH_Info_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Specific_E_DCH_Info_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_Specific_E_DCH_Info(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_Specific_E_DCH_Info_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Reference_ReceivedTotalWideBandPower_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Reference_ReceivedTotalWideBandPower(tvb, offset, &asn1_ctx, tree, hf_nbap_Reference_ReceivedTotalWideBandPower_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Reference_ReceivedTotalWideBandPowerReporting_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Reference_ReceivedTotalWideBandPowerReporting(tvb, offset, &asn1_ctx, tree, hf_nbap_Reference_ReceivedTotalWideBandPowerReporting_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Reference_ReceivedTotalWideBandPowerSupportIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Reference_ReceivedTotalWideBandPowerSupportIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_Reference_ReceivedTotalWideBandPowerSupportIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ReferenceClockAvailability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ReferenceClockAvailability(tvb, offset, &asn1_ctx, tree, hf_nbap_ReferenceClockAvailability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ReferenceSFNoffset_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ReferenceSFNoffset(tvb, offset, &asn1_ctx, tree, hf_nbap_ReferenceSFNoffset_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ReportCharacteristics_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ReportCharacteristics(tvb, offset, &asn1_ctx, tree, hf_nbap_ReportCharacteristics_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ReportCharacteristicsType_OnModification_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ReportCharacteristicsType_OnModification(tvb, offset, &asn1_ctx, tree, hf_nbap_ReportCharacteristicsType_OnModification_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EDCH_RACH_Report_IncrDecrThres_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_EDCH_RACH_Report_IncrDecrThres(tvb, offset, &asn1_ctx, tree, hf_nbap_EDCH_RACH_Report_IncrDecrThres_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EDCH_RACH_Report_ThresholdInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_EDCH_RACH_Report_ThresholdInformation(tvb, offset, &asn1_ctx, tree, hf_nbap_EDCH_RACH_Report_ThresholdInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_ID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_ID(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_ID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Set_ID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_Set_ID(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_Set_ID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_RLC_PDU_Size_Format_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_RLC_PDU_Size_Format(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_RLC_PDU_Size_Format_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Received_total_wide_band_power_For_CellPortion_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Received_total_wide_band_power_For_CellPortion_Value(tvb, offset, &asn1_ctx, tree, hf_nbap_Received_total_wide_band_power_For_CellPortion_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Received_total_wide_band_power_For_CellPortion_ValueLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Received_total_wide_band_power_For_CellPortion_ValueLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_Received_total_wide_band_power_For_CellPortion_ValueLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Received_total_wide_band_power_Value_IncrDecrThres_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Received_total_wide_band_power_Value_IncrDecrThres(tvb, offset, &asn1_ctx, tree, hf_nbap_Received_total_wide_band_power_Value_IncrDecrThres_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value(tvb, offset, &asn1_ctx, tree, hf_nbap_Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Received_Scheduled_EDCH_Power_Share_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Received_Scheduled_EDCH_Power_Share_Value(tvb, offset, &asn1_ctx, tree, hf_nbap_Received_Scheduled_EDCH_Power_Share_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RSEPS_Value_IncrDecrThres_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RSEPS_Value_IncrDecrThres(tvb, offset, &asn1_ctx, tree, hf_nbap_RSEPS_Value_IncrDecrThres_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Rx_Timing_Deviation_Value_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Rx_Timing_Deviation_Value_LCR(tvb, offset, &asn1_ctx, tree, hf_nbap_Rx_Timing_Deviation_Value_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Rx_Timing_Deviation_Value_768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Rx_Timing_Deviation_Value_768(tvb, offset, &asn1_ctx, tree, hf_nbap_Rx_Timing_Deviation_Value_768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Rx_Timing_Deviation_Value_384_ext_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Rx_Timing_Deviation_Value_384_ext(tvb, offset, &asn1_ctx, tree, hf_nbap_Rx_Timing_Deviation_Value_384_ext_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RTWP_ReportingIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RTWP_ReportingIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_RTWP_ReportingIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RTWP_CellPortion_ReportingIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RTWP_CellPortion_ReportingIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_RTWP_CellPortion_ReportingIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_AdjustmentPeriod_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_AdjustmentPeriod(tvb, offset, &asn1_ctx, tree, hf_nbap_AdjustmentPeriod_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DPCCH_Power_Boosting_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DPCCH_Power_Boosting_Capability(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DPCCH_Power_Boosting_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SAT_Info_Almanac_ExtList_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SAT_Info_Almanac_ExtList(tvb, offset, &asn1_ctx, tree, hf_nbap_SAT_Info_Almanac_ExtList_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ScaledAdjustmentRatio_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ScaledAdjustmentRatio(tvb, offset, &asn1_ctx, tree, hf_nbap_ScaledAdjustmentRatio_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MaxAdjustmentStep_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MaxAdjustmentStep(tvb, offset, &asn1_ctx, tree, hf_nbap_MaxAdjustmentStep_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SchedulingPriorityIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SchedulingPriorityIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_SchedulingPriorityIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Secondary_CPICH_Information_Change_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Secondary_CPICH_Information_Change(tvb, offset, &asn1_ctx, tree, hf_nbap_Secondary_CPICH_Information_Change_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Secondary_CCPCH_SlotFormat_Extended_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Secondary_CCPCH_SlotFormat_Extended(tvb, offset, &asn1_ctx, tree, hf_nbap_Secondary_CCPCH_SlotFormat_Extended_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Serving_E_DCH_RL_ID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Serving_E_DCH_RL_ID(tvb, offset, &asn1_ctx, tree, hf_nbap_Serving_E_DCH_RL_ID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SFN_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SFN(tvb, offset, &asn1_ctx, tree, hf_nbap_SFN_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SFNSFNMeasurementThresholdInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SFNSFNMeasurementThresholdInformation(tvb, offset, &asn1_ctx, tree, hf_nbap_SFNSFNMeasurementThresholdInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SFNSFNMeasurementValueInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SFNSFNMeasurementValueInformation(tvb, offset, &asn1_ctx, tree, hf_nbap_SFNSFNMeasurementValueInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ShutdownTimer_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ShutdownTimer(tvb, offset, &asn1_ctx, tree, hf_nbap_ShutdownTimer_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Single_Stream_MIMO_ActivationIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Single_Stream_MIMO_ActivationIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_Single_Stream_MIMO_ActivationIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Single_Stream_MIMO_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Single_Stream_MIMO_Capability(tvb, offset, &asn1_ctx, tree, hf_nbap_Single_Stream_MIMO_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Single_Stream_MIMO_Mode_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Single_Stream_MIMO_Mode_Indicator(tvb, offset, &asn1_ctx, tree, hf_nbap_Single_Stream_MIMO_Mode_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SignallingBearerRequestIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SignallingBearerRequestIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_SignallingBearerRequestIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SixtyfourQAM_UsageAllowedIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SixtyfourQAM_UsageAllowedIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_SixtyfourQAM_UsageAllowedIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SixtyfourQAM_DL_UsageIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SixtyfourQAM_DL_UsageIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_SixtyfourQAM_DL_UsageIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SixtyfourQAM_DL_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SixtyfourQAM_DL_Capability(tvb, offset, &asn1_ctx, tree, hf_nbap_SixtyfourQAM_DL_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SixtyfourQAM_DL_MIMO_Combined_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SixtyfourQAM_DL_MIMO_Combined_Capability(tvb, offset, &asn1_ctx, tree, hf_nbap_SixtyfourQAM_DL_MIMO_Combined_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SixteenQAM_UL_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SixteenQAM_UL_Capability(tvb, offset, &asn1_ctx, tree, hf_nbap_SixteenQAM_UL_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SixteenQAM_UL_Operation_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SixteenQAM_UL_Operation_Indicator(tvb, offset, &asn1_ctx, tree, hf_nbap_SixteenQAM_UL_Operation_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Start_Of_Audit_Sequence_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Start_Of_Audit_Sequence_Indicator(tvb, offset, &asn1_ctx, tree, hf_nbap_Start_Of_Audit_Sequence_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SyncCase_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SyncCase(tvb, offset, &asn1_ctx, tree, hf_nbap_SyncCase_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SynchronisationReportCharacteristics_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SynchronisationReportCharacteristics(tvb, offset, &asn1_ctx, tree, hf_nbap_SynchronisationReportCharacteristics_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SyncDLCodeIdThreInfoLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SyncDLCodeIdThreInfoLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_SyncDLCodeIdThreInfoLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SynchronisationReportType_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SynchronisationReportType(tvb, offset, &asn1_ctx, tree, hf_nbap_SynchronisationReportType_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Semi_PersistentScheduling_CapabilityLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Semi_PersistentScheduling_CapabilityLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_Semi_PersistentScheduling_CapabilityLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_T_Cell_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_T_Cell(tvb, offset, &asn1_ctx, tree, hf_nbap_T_Cell_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TDD_DL_DPCH_TimeSlotFormat_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TDD_DL_DPCH_TimeSlotFormat_LCR(tvb, offset, &asn1_ctx, tree, hf_nbap_TDD_DL_DPCH_TimeSlotFormat_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TDD_TPC_DownlinkStepSize_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TDD_TPC_DownlinkStepSize(tvb, offset, &asn1_ctx, tree, hf_nbap_TDD_TPC_DownlinkStepSize_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TDD_TPC_UplinkStepSize_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TDD_TPC_UplinkStepSize_LCR(tvb, offset, &asn1_ctx, tree, hf_nbap_TDD_TPC_UplinkStepSize_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TDD_UL_DPCH_TimeSlotFormat_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TDD_UL_DPCH_TimeSlotFormat_LCR(tvb, offset, &asn1_ctx, tree, hf_nbap_TDD_UL_DPCH_TimeSlotFormat_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TFCI_Presence_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TFCI_Presence(tvb, offset, &asn1_ctx, tree, hf_nbap_TFCI_Presence_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TimeSlot_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TimeSlot(tvb, offset, &asn1_ctx, tree, hf_nbap_TimeSlot_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TimeSlotLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TimeSlotLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_TimeSlotLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TimeslotLCR_Extension_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TimeslotLCR_Extension(tvb, offset, &asn1_ctx, tree, hf_nbap_TimeslotLCR_Extension_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TimeSlotMeasurementValueListLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TimeSlotMeasurementValueListLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_TimeSlotMeasurementValueListLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TimingAdjustmentValue_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TimingAdjustmentValue(tvb, offset, &asn1_ctx, tree, hf_nbap_TimingAdjustmentValue_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TimingAdjustmentValueLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TimingAdjustmentValueLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_TimingAdjustmentValueLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TimingAdvanceApplied_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TimingAdvanceApplied(tvb, offset, &asn1_ctx, tree, hf_nbap_TimingAdvanceApplied_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SynchronisationIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SynchronisationIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_SynchronisationIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TnlQos_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TnlQos(tvb, offset, &asn1_ctx, tree, hf_nbap_TnlQos_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Transmission_Gap_Pattern_Sequence_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Transmission_Gap_Pattern_Sequence_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_Transmission_Gap_Pattern_Sequence_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue(tvb, offset, &asn1_ctx, tree, hf_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue(tvb, offset, &asn1_ctx, tree, hf_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TransmittedCarrierPowerOfAllCodesNotUsedForHSTransmissionValue_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHSTransmissionValue(tvb, offset, &asn1_ctx, tree, hf_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHSTransmissionValue_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Transmitted_Carrier_Power_For_CellPortion_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Transmitted_Carrier_Power_For_CellPortion_Value(tvb, offset, &asn1_ctx, tree, hf_nbap_Transmitted_Carrier_Power_For_CellPortion_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Transmitted_Carrier_Power_For_CellPortion_ValueLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Transmitted_Carrier_Power_For_CellPortion_ValueLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_Transmitted_Carrier_Power_For_CellPortion_ValueLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Transmitted_Carrier_Power_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Transmitted_Carrier_Power_Value(tvb, offset, &asn1_ctx, tree, hf_nbap_Transmitted_Carrier_Power_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TransmissionDiversityApplied_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TransmissionDiversityApplied(tvb, offset, &asn1_ctx, tree, hf_nbap_TransmissionDiversityApplied_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TransmitDiversityIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TransmitDiversityIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_TransmitDiversityIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TransportBearerRequestIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TransportBearerRequestIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_TransportBearerRequestIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TransportBearerNotRequestedIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TransportBearerNotRequestedIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_TransportBearerNotRequestedIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TransportBearerNotSetupIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TransportBearerNotSetupIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_TransportBearerNotSetupIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TransportLayerAddress_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TransportLayerAddress(tvb, offset, &asn1_ctx, tree, hf_nbap_TransportLayerAddress_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TS0_CapabilityLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TS0_CapabilityLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_TS0_CapabilityLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TSTD_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TSTD_Indicator(tvb, offset, &asn1_ctx, tree, hf_nbap_TSTD_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TSN_Length_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TSN_Length(tvb, offset, &asn1_ctx, tree, hf_nbap_TSN_Length_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TUTRANGANSSMeasurementThresholdInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TUTRANGANSSMeasurementThresholdInformation(tvb, offset, &asn1_ctx, tree, hf_nbap_TUTRANGANSSMeasurementThresholdInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TUTRANGANSSMeasurementValueInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TUTRANGANSSMeasurementValueInformation(tvb, offset, &asn1_ctx, tree, hf_nbap_TUTRANGANSSMeasurementValueInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TUTRANGPSMeasurementThresholdInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TUTRANGPSMeasurementThresholdInformation(tvb, offset, &asn1_ctx, tree, hf_nbap_TUTRANGPSMeasurementThresholdInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TUTRANGPSMeasurementValueInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TUTRANGPSMeasurementValueInformation(tvb, offset, &asn1_ctx, tree, hf_nbap_TUTRANGPSMeasurementValueInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TxDiversityOnDLControlChannelsByMIMOUECapability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TxDiversityOnDLControlChannelsByMIMOUECapability(tvb, offset, &asn1_ctx, tree, hf_nbap_TxDiversityOnDLControlChannelsByMIMOUECapability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TypeOfError_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TypeOfError(tvb, offset, &asn1_ctx, tree, hf_nbap_TypeOfError_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UARFCN_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UARFCN(tvb, offset, &asn1_ctx, tree, hf_nbap_UARFCN_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UE_AggregateMaximumBitRate_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UE_AggregateMaximumBitRate(tvb, offset, &asn1_ctx, tree, hf_nbap_UE_AggregateMaximumBitRate_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UE_AggregateMaximumBitRate_Enforcement_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UE_AggregateMaximumBitRate_Enforcement_Indicator(tvb, offset, &asn1_ctx, tree, hf_nbap_UE_AggregateMaximumBitRate_Enforcement_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UE_Capability_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UE_Capability_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_UE_Capability_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UE_TS0_CapabilityLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UE_TS0_CapabilityLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_UE_TS0_CapabilityLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UE_SupportIndicatorExtension_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UE_SupportIndicatorExtension(tvb, offset, &asn1_ctx, tree, hf_nbap_UE_SupportIndicatorExtension_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_LCRTDD_HSDSCH_Physical_Layer_Category_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_LCRTDD_HSDSCH_Physical_Layer_Category(tvb, offset, &asn1_ctx, tree, hf_nbap_LCRTDD_HSDSCH_Physical_Layer_Category_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPDCH_Indicator_For_E_DCH_Operation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_DPDCH_Indicator_For_E_DCH_Operation(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_DPDCH_Indicator_For_E_DCH_Operation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_SIR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_SIR(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_SIR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_Synchronisation_Parameters_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_Synchronisation_Parameters_LCR(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_Synchronisation_Parameters_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UPPCHPositionLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UPPCHPositionLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_UPPCHPositionLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UpPTSInterference_For_CellPortion_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UpPTSInterference_For_CellPortion_Value(tvb, offset, &asn1_ctx, tree, hf_nbap_UpPTSInterference_For_CellPortion_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UpPTSInterferenceValue_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UpPTSInterferenceValue(tvb, offset, &asn1_ctx, tree, hf_nbap_UpPTSInterferenceValue_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Unidirectional_DCH_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Unidirectional_DCH_Indicator(tvb, offset, &asn1_ctx, tree, hf_nbap_Unidirectional_DCH_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_USCH_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_USCH_Information(tvb, offset, &asn1_ctx, tree, hf_nbap_USCH_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_USCH_InformationResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_USCH_InformationResponse(tvb, offset, &asn1_ctx, tree, hf_nbap_USCH_InformationResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_TimeslotISCP_For_CellPortion_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_TimeslotISCP_For_CellPortion_Value(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_TimeslotISCP_For_CellPortion_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_TimeslotISCP_Value_IncrDecrThres_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_TimeslotISCP_Value_IncrDecrThres(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_TimeslotISCP_Value_IncrDecrThres_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonTransportChannelSetupRequestFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonTransportChannelSetupRequestFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonTransportChannelSetupRequestFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonPhysicalChannelType_CTCH_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonPhysicalChannelType_CTCH_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonPhysicalChannelType_CTCH_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FACH_ParametersListIE_CTCH_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_FACH_ParametersListIE_CTCH_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_FACH_ParametersListIE_CTCH_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PCH_ParametersItem_CTCH_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PCH_ParametersItem_CTCH_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PCH_ParametersItem_CTCH_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MICH_Parameters_CTCH_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MICH_Parameters_CTCH_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_MICH_Parameters_CTCH_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RACH_ParametersItem_CTCH_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RACH_ParametersItem_CTCH_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RACH_ParametersItem_CTCH_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonTransportChannelSetupRequestTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonTransportChannelSetupRequestTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonTransportChannelSetupRequestTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonPhysicalChannelType_CTCH_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonPhysicalChannelType_CTCH_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonPhysicalChannelType_CTCH_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Secondary_CCPCH_parameterListIE_CTCH_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Secondary_CCPCH_parameterListIE_CTCH_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Secondary_CCPCH_parameterListIE_CTCH_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Secondary_CCPCH_LCR_parameterList_CTCH_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Secondary_CCPCH_LCR_parameterList_CTCH_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Secondary_CCPCH_LCR_parameterList_CTCH_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Secondary_CCPCH_768_parameterList_CTCH_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Secondary_CCPCH_768_parameterList_CTCH_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Secondary_CCPCH_768_parameterList_CTCH_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FACH_ParametersListIE_CTCH_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_FACH_ParametersListIE_CTCH_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_FACH_ParametersListIE_CTCH_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PCH_ParametersItem_CTCH_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PCH_ParametersItem_CTCH_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PCH_ParametersItem_CTCH_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PICH_ParametersItem_CTCH_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PICH_ParametersItem_CTCH_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PICH_ParametersItem_CTCH_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PICH_LCR_Parameters_CTCH_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PICH_LCR_Parameters_CTCH_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PICH_LCR_Parameters_CTCH_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PICH_768_ParametersItem_CTCH_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PICH_768_ParametersItem_CTCH_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PICH_768_ParametersItem_CTCH_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MICH_Parameters_CTCH_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MICH_Parameters_CTCH_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_MICH_Parameters_CTCH_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TimeSlotConfigurationList_LCR_CTCH_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TimeSlotConfigurationList_LCR_CTCH_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_TimeSlotConfigurationList_LCR_CTCH_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Secondary_CCPCH_parameterExtendedList_CTCH_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Secondary_CCPCH_parameterExtendedList_CTCH_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Secondary_CCPCH_parameterExtendedList_CTCH_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PRACH_ParametersItem_CTCH_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PRACH_ParametersItem_CTCH_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PRACH_ParametersItem_CTCH_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RACH_ParameterItem_CTCH_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RACH_ParameterItem_CTCH_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RACH_ParameterItem_CTCH_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PRACH_LCR_ParametersList_CTCH_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PRACH_LCR_ParametersList_CTCH_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PRACH_LCR_ParametersList_CTCH_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PRACH_768_ParametersItem_CTCH_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PRACH_768_ParametersItem_CTCH_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PRACH_768_ParametersItem_CTCH_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FPACH_LCR_Parameters_CTCH_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_FPACH_LCR_Parameters_CTCH_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_FPACH_LCR_Parameters_CTCH_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PLCCH_parameters_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PLCCH_parameters(tvb, offset, &asn1_ctx, tree, hf_nbap_PLCCH_parameters_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_RUCCH_parameters_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_RUCCH_parameters(tvb, offset, &asn1_ctx, tree, hf_nbap_E_RUCCH_parameters_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_RUCCH_768_parameters_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_RUCCH_768_parameters(tvb, offset, &asn1_ctx, tree, hf_nbap_E_RUCCH_768_parameters_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonTransportChannelSetupResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonTransportChannelSetupResponse(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonTransportChannelSetupResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FACH_CommonTransportChannel_InformationResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_FACH_CommonTransportChannel_InformationResponse(tvb, offset, &asn1_ctx, tree, hf_nbap_FACH_CommonTransportChannel_InformationResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonTransportChannelSetupFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonTransportChannelSetupFailure(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonTransportChannelSetupFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonTransportChannelReconfigurationRequestFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonTransportChannelReconfigurationRequestFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonTransportChannelReconfigurationRequestFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonPhysicalChannelType_CTCH_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonPhysicalChannelType_CTCH_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonPhysicalChannelType_CTCH_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FACH_ParametersListIE_CTCH_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_FACH_ParametersListIE_CTCH_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_FACH_ParametersListIE_CTCH_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PCH_ParametersItem_CTCH_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PCH_ParametersItem_CTCH_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PCH_ParametersItem_CTCH_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PICH_ParametersItem_CTCH_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PICH_ParametersItem_CTCH_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PICH_ParametersItem_CTCH_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MICH_Parameters_CTCH_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MICH_Parameters_CTCH_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_MICH_Parameters_CTCH_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PRACH_ParametersListIE_CTCH_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PRACH_ParametersListIE_CTCH_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PRACH_ParametersListIE_CTCH_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_AICH_ParametersListIE_CTCH_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_AICH_ParametersListIE_CTCH_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_AICH_ParametersListIE_CTCH_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonTransportChannelReconfigurationRequestTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonTransportChannelReconfigurationRequestTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonTransportChannelReconfigurationRequestTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Secondary_CCPCH_Parameters_CTCH_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Secondary_CCPCH_Parameters_CTCH_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Secondary_CCPCH_Parameters_CTCH_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Secondary_CCPCHListIE_CTCH_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Secondary_CCPCHListIE_CTCH_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Secondary_CCPCHListIE_CTCH_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Secondary_CCPCH_parameterExtendedList_CTCH_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Secondary_CCPCH_parameterExtendedList_CTCH_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Secondary_CCPCH_parameterExtendedList_CTCH_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PICH_Parameters_CTCH_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PICH_Parameters_CTCH_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PICH_Parameters_CTCH_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FACH_ParametersList_CTCH_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_FACH_ParametersList_CTCH_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_FACH_ParametersList_CTCH_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PCH_Parameters_CTCH_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PCH_Parameters_CTCH_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PCH_Parameters_CTCH_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FPACH_LCR_Parameters_CTCH_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_FPACH_LCR_Parameters_CTCH_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_FPACH_LCR_Parameters_CTCH_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MICH_Parameters_CTCH_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MICH_Parameters_CTCH_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_MICH_Parameters_CTCH_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PLCCH_Parameters_CTCH_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PLCCH_Parameters_CTCH_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PLCCH_Parameters_CTCH_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Secondary_CCPCH_768_Parameters_CTCH_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Secondary_CCPCH_768_Parameters_CTCH_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Secondary_CCPCH_768_Parameters_CTCH_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PICH_768_Parameters_CTCH_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PICH_768_Parameters_CTCH_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PICH_768_Parameters_CTCH_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MICH_768_Parameters_CTCH_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MICH_768_Parameters_CTCH_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_MICH_768_Parameters_CTCH_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UPPCH_LCR_Parameters_CTCH_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UPPCH_LCR_Parameters_CTCH_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UPPCH_LCR_Parameters_CTCH_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonTransportChannelReconfigurationResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonTransportChannelReconfigurationResponse(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonTransportChannelReconfigurationResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonTransportChannelReconfigurationFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonTransportChannelReconfigurationFailure(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonTransportChannelReconfigurationFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonTransportChannelDeletionRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonTransportChannelDeletionRequest(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonTransportChannelDeletionRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonTransportChannelDeletionResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonTransportChannelDeletionResponse(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonTransportChannelDeletionResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_BlockResourceRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_BlockResourceRequest(tvb, offset, &asn1_ctx, tree, hf_nbap_BlockResourceRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_BlockResourceResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_BlockResourceResponse(tvb, offset, &asn1_ctx, tree, hf_nbap_BlockResourceResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_BlockResourceFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_BlockResourceFailure(tvb, offset, &asn1_ctx, tree, hf_nbap_BlockResourceFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UnblockResourceIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UnblockResourceIndication(tvb, offset, &asn1_ctx, tree, hf_nbap_UnblockResourceIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_AuditRequiredIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_AuditRequiredIndication(tvb, offset, &asn1_ctx, tree, hf_nbap_AuditRequiredIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_AuditRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_AuditRequest(tvb, offset, &asn1_ctx, tree, hf_nbap_AuditRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_AuditResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_AuditResponse(tvb, offset, &asn1_ctx, tree, hf_nbap_AuditResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Cell_InformationList_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Cell_InformationList_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_Cell_InformationList_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Cell_InformationItem_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Cell_InformationItem_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_Cell_InformationItem_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FPACH_LCR_InformationList_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_FPACH_LCR_InformationList_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_FPACH_LCR_InformationList_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_DSCH_Resources_Information_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_DSCH_Resources_Information_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_DSCH_Resources_Information_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_S_CCPCH_InformationListExt_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_S_CCPCH_InformationListExt_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_S_CCPCH_InformationListExt_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_S_CCPCH_LCR_InformationListExt_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_S_CCPCH_LCR_InformationListExt_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_S_CCPCH_LCR_InformationListExt_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_Resources_Information_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_Resources_Information_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_Resources_Information_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PLCCH_InformationList_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PLCCH_InformationList_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_PLCCH_InformationList_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_S_CCPCH_768_InformationList_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_S_CCPCH_768_InformationList_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_S_CCPCH_768_InformationList_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PRACH_768_InformationList_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PRACH_768_InformationList_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_PRACH_768_InformationList_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_RUCCH_InformationList_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_RUCCH_InformationList_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_E_RUCCH_InformationList_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_RUCCH_768_InformationList_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_RUCCH_768_InformationList_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_E_RUCCH_768_InformationList_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Cell_Frequency_List_Information_LCR_MulFreq_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Cell_Frequency_List_Information_LCR_MulFreq_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_Cell_Frequency_List_Information_LCR_MulFreq_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Cell_Frequency_List_InformationItem_LCR_MulFreq_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Cell_Frequency_List_InformationItem_LCR_MulFreq_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_Cell_Frequency_List_InformationItem_LCR_MulFreq_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UPPCH_LCR_InformationList_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UPPCH_LCR_InformationList_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_UPPCH_LCR_InformationList_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UPPCH_LCR_InformationItem_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UPPCH_LCR_InformationItem_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_UPPCH_LCR_InformationItem_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MultipleFreq_HS_DSCH_Resources_InformationList_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MultipleFreq_HS_DSCH_Resources_InformationList_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_MultipleFreq_HS_DSCH_Resources_InformationList_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MultipleFreq_E_DCH_Resources_InformationList_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MultipleFreq_E_DCH_Resources_InformationList_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_MultipleFreq_E_DCH_Resources_InformationList_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CCP_InformationList_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CCP_InformationList_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_CCP_InformationList_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CCP_InformationItem_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CCP_InformationItem_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_CCP_InformationItem_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Local_Cell_InformationList_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Local_Cell_InformationList_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_Local_Cell_InformationList_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Local_Cell_InformationItem_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Local_Cell_InformationItem_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_Local_Cell_InformationItem_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Local_Cell_Group_InformationList_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Local_Cell_Group_InformationList_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_Local_Cell_Group_InformationList_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Local_Cell_Group_InformationItem_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Local_Cell_Group_InformationItem_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_Local_Cell_Group_InformationItem_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Power_Local_Cell_Group_InformationList_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Power_Local_Cell_Group_InformationList_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_Power_Local_Cell_Group_InformationList_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Power_Local_Cell_Group_InformationItem_AuditRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Power_Local_Cell_Group_InformationItem_AuditRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_Power_Local_Cell_Group_InformationItem_AuditRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_AuditFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_AuditFailure(tvb, offset, &asn1_ctx, tree, hf_nbap_AuditFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonMeasurementInitiationRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonMeasurementInitiationRequest(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonMeasurementInitiationRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonMeasurementObjectType_CM_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonMeasurementObjectType_CM_Rqst(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonMeasurementObjectType_CM_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ERACH_CM_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ERACH_CM_Rqst(tvb, offset, &asn1_ctx, tree, hf_nbap_ERACH_CM_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PowerLocalCellGroup_CM_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PowerLocalCellGroup_CM_Rqst(tvb, offset, &asn1_ctx, tree, hf_nbap_PowerLocalCellGroup_CM_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonMeasurementInitiationResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonMeasurementInitiationResponse(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonMeasurementInitiationResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonMeasurementObjectType_CM_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonMeasurementObjectType_CM_Rsp(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonMeasurementObjectType_CM_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ERACH_CM_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ERACH_CM_Rsp(tvb, offset, &asn1_ctx, tree, hf_nbap_ERACH_CM_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PowerLocalCellGroup_CM_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PowerLocalCellGroup_CM_Rsp(tvb, offset, &asn1_ctx, tree, hf_nbap_PowerLocalCellGroup_CM_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonMeasurementInitiationFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonMeasurementInitiationFailure(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonMeasurementInitiationFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonMeasurementReport_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonMeasurementReport(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonMeasurementReport_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonMeasurementObjectType_CM_Rprt_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonMeasurementObjectType_CM_Rprt(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonMeasurementObjectType_CM_Rprt_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ERACH_CM_Rprt_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ERACH_CM_Rprt(tvb, offset, &asn1_ctx, tree, hf_nbap_ERACH_CM_Rprt_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PowerLocalCellGroup_CM_Rprt_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PowerLocalCellGroup_CM_Rprt(tvb, offset, &asn1_ctx, tree, hf_nbap_PowerLocalCellGroup_CM_Rprt_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonMeasurementTerminationRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonMeasurementTerminationRequest(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonMeasurementTerminationRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonMeasurementFailureIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommonMeasurementFailureIndication(tvb, offset, &asn1_ctx, tree, hf_nbap_CommonMeasurementFailureIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellSetupRequestFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellSetupRequestFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellSetupRequestFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Synchronisation_Configuration_Cell_SetupRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Synchronisation_Configuration_Cell_SetupRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_Synchronisation_Configuration_Cell_SetupRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PrimarySCH_Information_Cell_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PrimarySCH_Information_Cell_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PrimarySCH_Information_Cell_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SecondarySCH_Information_Cell_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SecondarySCH_Information_Cell_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_SecondarySCH_Information_Cell_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PrimaryCPICH_Information_Cell_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PrimaryCPICH_Information_Cell_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PrimaryCPICH_Information_Cell_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SecondaryCPICH_InformationList_Cell_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SecondaryCPICH_InformationList_Cell_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_SecondaryCPICH_InformationList_Cell_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SecondaryCPICH_InformationItem_Cell_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SecondaryCPICH_InformationItem_Cell_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_SecondaryCPICH_InformationItem_Cell_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PrimaryCCPCH_Information_Cell_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PrimaryCCPCH_Information_Cell_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PrimaryCCPCH_Information_Cell_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Limited_power_increase_information_Cell_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Limited_power_increase_information_Cell_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Limited_power_increase_information_Cell_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_IPDLParameter_Information_Cell_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_IPDLParameter_Information_Cell_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_IPDLParameter_Information_Cell_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellPortion_InformationList_Cell_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellPortion_InformationList_Cell_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellPortion_InformationList_Cell_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellPortion_InformationItem_Cell_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellPortion_InformationItem_Cell_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellPortion_InformationItem_Cell_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellSetupRequestTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellSetupRequestTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellSetupRequestTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SCH_Information_Cell_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SCH_Information_Cell_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_SCH_Information_Cell_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SyncCaseIndicatorItem_Cell_SetupRqstTDD_PSCH_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SyncCaseIndicatorItem_Cell_SetupRqstTDD_PSCH(tvb, offset, &asn1_ctx, tree, hf_nbap_SyncCaseIndicatorItem_Cell_SetupRqstTDD_PSCH_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PCCPCH_Information_Cell_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PCCPCH_Information_Cell_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PCCPCH_Information_Cell_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TimeSlotConfigurationList_Cell_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TimeSlotConfigurationList_Cell_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_TimeSlotConfigurationList_Cell_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TimeSlotConfigurationList_LCR_Cell_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TimeSlotConfigurationList_LCR_Cell_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_TimeSlotConfigurationList_LCR_Cell_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PCCPCH_LCR_Information_Cell_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PCCPCH_LCR_Information_Cell_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PCCPCH_LCR_Information_Cell_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DwPCH_LCR_Information_Cell_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DwPCH_LCR_Information_Cell_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DwPCH_LCR_Information_Cell_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_IPDLParameter_Information_Cell_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_IPDLParameter_Information_Cell_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_IPDLParameter_Information_Cell_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_IPDLParameter_Information_LCR_Cell_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_IPDLParameter_Information_LCR_Cell_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_IPDLParameter_Information_LCR_Cell_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PCCPCH_768_Information_Cell_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PCCPCH_768_Information_Cell_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PCCPCH_768_Information_Cell_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SCH_768_Information_Cell_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SCH_768_Information_Cell_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_SCH_768_Information_Cell_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Cell_Frequency_List_LCR_MulFreq_Cell_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Cell_Frequency_List_LCR_MulFreq_Cell_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Cell_Frequency_List_LCR_MulFreq_Cell_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellSetupResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellSetupResponse(tvb, offset, &asn1_ctx, tree, hf_nbap_CellSetupResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellSetupFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellSetupFailure(tvb, offset, &asn1_ctx, tree, hf_nbap_CellSetupFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellReconfigurationRequestFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellReconfigurationRequestFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellReconfigurationRequestFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Synchronisation_Configuration_Cell_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Synchronisation_Configuration_Cell_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_Synchronisation_Configuration_Cell_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PrimarySCH_Information_Cell_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PrimarySCH_Information_Cell_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PrimarySCH_Information_Cell_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SecondarySCH_Information_Cell_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SecondarySCH_Information_Cell_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_SecondarySCH_Information_Cell_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PrimaryCPICH_Information_Cell_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PrimaryCPICH_Information_Cell_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PrimaryCPICH_Information_Cell_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SecondaryCPICH_InformationList_Cell_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SecondaryCPICH_InformationList_Cell_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_SecondaryCPICH_InformationList_Cell_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SecondaryCPICH_InformationItem_Cell_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SecondaryCPICH_InformationItem_Cell_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_SecondaryCPICH_InformationItem_Cell_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PrimaryCCPCH_Information_Cell_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PrimaryCCPCH_Information_Cell_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PrimaryCCPCH_Information_Cell_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_IPDLParameter_Information_Cell_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_IPDLParameter_Information_Cell_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_IPDLParameter_Information_Cell_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellPortion_InformationList_Cell_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellPortion_InformationList_Cell_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellPortion_InformationList_Cell_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellPortion_InformationItem_Cell_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellPortion_InformationItem_Cell_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellPortion_InformationItem_Cell_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellReconfigurationRequestTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellReconfigurationRequestTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellReconfigurationRequestTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SCH_Information_Cell_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SCH_Information_Cell_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_SCH_Information_Cell_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PCCPCH_Information_Cell_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PCCPCH_Information_Cell_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PCCPCH_Information_Cell_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TimeSlotConfigurationList_Cell_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TimeSlotConfigurationList_Cell_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_TimeSlotConfigurationList_Cell_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TimeSlotConfigurationList_LCR_Cell_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TimeSlotConfigurationList_LCR_Cell_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_TimeSlotConfigurationList_LCR_Cell_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DwPCH_LCR_Information_Cell_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DwPCH_LCR_Information_Cell_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DwPCH_LCR_Information_Cell_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_IPDLParameter_Information_Cell_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_IPDLParameter_Information_Cell_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_IPDLParameter_Information_Cell_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_IPDLParameter_Information_LCR_Cell_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_IPDLParameter_Information_LCR_Cell_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_IPDLParameter_Information_LCR_Cell_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SCH_768_Information_Cell_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SCH_768_Information_Cell_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_SCH_768_Information_Cell_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PCCPCH_768_Information_Cell_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PCCPCH_768_Information_Cell_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PCCPCH_768_Information_Cell_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UARFCN_Adjustment_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UARFCN_Adjustment(tvb, offset, &asn1_ctx, tree, hf_nbap_UARFCN_Adjustment_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellReconfigurationResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellReconfigurationResponse(tvb, offset, &asn1_ctx, tree, hf_nbap_CellReconfigurationResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellReconfigurationFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellReconfigurationFailure(tvb, offset, &asn1_ctx, tree, hf_nbap_CellReconfigurationFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellDeletionRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellDeletionRequest(tvb, offset, &asn1_ctx, tree, hf_nbap_CellDeletionRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellDeletionResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellDeletionResponse(tvb, offset, &asn1_ctx, tree, hf_nbap_CellDeletionResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ResourceStatusIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ResourceStatusIndication(tvb, offset, &asn1_ctx, tree, hf_nbap_ResourceStatusIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_IndicationType_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_IndicationType_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_IndicationType_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Local_Cell_InformationItem_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Local_Cell_InformationItem_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_Local_Cell_InformationItem_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Local_Cell_Group_InformationItem_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Local_Cell_Group_InformationItem_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_Local_Cell_Group_InformationItem_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Power_Local_Cell_Group_InformationList_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Power_Local_Cell_Group_InformationList_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_Power_Local_Cell_Group_InformationList_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Power_Local_Cell_Group_InformationItem_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Power_Local_Cell_Group_InformationItem_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_Power_Local_Cell_Group_InformationItem_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Local_Cell_InformationItem2_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Local_Cell_InformationItem2_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_Local_Cell_InformationItem2_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Local_Cell_Group_InformationItem2_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Local_Cell_Group_InformationItem2_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_Local_Cell_Group_InformationItem2_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CCP_InformationItem_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CCP_InformationItem_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_CCP_InformationItem_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Cell_InformationItem_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Cell_InformationItem_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_Cell_InformationItem_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FPACH_LCR_InformationList_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_FPACH_LCR_InformationList_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_FPACH_LCR_InformationList_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DwPCH_LCR_Information_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DwPCH_LCR_Information_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_DwPCH_LCR_Information_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_DSCH_Resources_Information_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_DSCH_Resources_Information_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_DSCH_Resources_Information_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_S_CCPCH_InformationListExt_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_S_CCPCH_InformationListExt_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_S_CCPCH_InformationListExt_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_S_CCPCH_LCR_InformationListExt_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_S_CCPCH_LCR_InformationListExt_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_S_CCPCH_LCR_InformationListExt_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_Resources_Information_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_Resources_Information_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_Resources_Information_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PLCCH_InformationList_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PLCCH_InformationList_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_PLCCH_InformationList_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_S_CCPCH_768_InformationList_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_S_CCPCH_768_InformationList_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_S_CCPCH_768_InformationList_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PRACH_768_InformationList_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PRACH_768_InformationList_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_PRACH_768_InformationList_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_RUCCH_InformationList_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_RUCCH_InformationList_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_E_RUCCH_InformationList_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_RUCCH_768_InformationList_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_RUCCH_768_InformationList_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_E_RUCCH_768_InformationList_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Cell_Frequency_List_Information_LCR_MulFreq_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Cell_Frequency_List_Information_LCR_MulFreq_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_Cell_Frequency_List_Information_LCR_MulFreq_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Cell_Frequency_List_InformationItem_LCR_MulFreq_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Cell_Frequency_List_InformationItem_LCR_MulFreq_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_Cell_Frequency_List_InformationItem_LCR_MulFreq_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UPPCH_LCR_InformationList_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UPPCH_LCR_InformationList_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_UPPCH_LCR_InformationList_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UPPCH_LCR_InformationItem_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UPPCH_LCR_InformationItem_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_UPPCH_LCR_InformationItem_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MultipleFreq_HS_DSCH_Resources_InformationList_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MultipleFreq_HS_DSCH_Resources_InformationList_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_MultipleFreq_HS_DSCH_Resources_InformationList_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Power_Local_Cell_Group_InformationList2_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Power_Local_Cell_Group_InformationList2_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_Power_Local_Cell_Group_InformationList2_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Power_Local_Cell_Group_InformationItem2_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Power_Local_Cell_Group_InformationItem2_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_Power_Local_Cell_Group_InformationItem2_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MultipleFreq_E_DCH_Resources_InformationList_ResourceStatusInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MultipleFreq_E_DCH_Resources_InformationList_ResourceStatusInd(tvb, offset, &asn1_ctx, tree, hf_nbap_MultipleFreq_E_DCH_Resources_InformationList_ResourceStatusInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SystemInformationUpdateRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SystemInformationUpdateRequest(tvb, offset, &asn1_ctx, tree, hf_nbap_SystemInformationUpdateRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MIB_SB_SIB_InformationList_SystemInfoUpdateRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MIB_SB_SIB_InformationList_SystemInfoUpdateRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_MIB_SB_SIB_InformationList_SystemInfoUpdateRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SegmentInformationListIE_SystemInfoUpdate_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SegmentInformationListIE_SystemInfoUpdate(tvb, offset, &asn1_ctx, tree, hf_nbap_SegmentInformationListIE_SystemInfoUpdate_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SystemInformationUpdateResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SystemInformationUpdateResponse(tvb, offset, &asn1_ctx, tree, hf_nbap_SystemInformationUpdateResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SystemInformationUpdateFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SystemInformationUpdateFailure(tvb, offset, &asn1_ctx, tree, hf_nbap_SystemInformationUpdateFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkSetupRequestFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkSetupRequestFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkSetupRequestFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_HS_Cell_Information_RL_Setup_List_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Additional_HS_Cell_Information_RL_Setup_List(tvb, offset, &asn1_ctx, tree, hf_nbap_Additional_HS_Cell_Information_RL_Setup_List_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_Information_RL_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_DPCH_Information_RL_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_DPCH_Information_RL_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_Information_RL_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_DPCH_Information_RL_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_DPCH_Information_RL_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationList_RL_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationList_RL_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationList_RL_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationItem_RL_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationItem_RL_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationItem_RL_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DPCH_Information_RL_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DPCH_Information_RL_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DPCH_Information_RL_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_F_DPCH_Information_RL_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_F_DPCH_Information_RL_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_F_DPCH_Information_RL_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkSetupRequestTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkSetupRequestTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkSetupRequestTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_InformationList_RL_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_CCTrCH_InformationList_RL_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_CCTrCH_InformationList_RL_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_InformationItem_RL_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_CCTrCH_InformationItem_RL_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_CCTrCH_InformationItem_RL_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_InformationItem_RL_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_DPCH_InformationItem_RL_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_DPCH_InformationItem_RL_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_LCR_Information_RL_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_DPCH_LCR_Information_RL_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_DPCH_LCR_Information_RL_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_768_Information_RL_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_DPCH_768_Information_RL_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_DPCH_768_Information_RL_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationList_RL_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_CCTrCH_InformationList_RL_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_CCTrCH_InformationList_RL_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationItem_RL_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_CCTrCH_InformationItem_RL_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_CCTrCH_InformationItem_RL_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_InformationItem_RL_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_DPCH_InformationItem_RL_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_DPCH_InformationItem_RL_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_LCR_Information_RL_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_DPCH_LCR_Information_RL_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_DPCH_LCR_Information_RL_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_768_Information_RL_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_DPCH_768_Information_RL_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_DPCH_768_Information_RL_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Information_RL_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_Information_RL_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_Information_RL_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkSetupResponseFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkSetupResponseFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkSetupResponseFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_HS_Cell_Information_Response_List_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Additional_HS_Cell_Information_Response_List(tvb, offset, &asn1_ctx, tree, hf_nbap_Additional_HS_Cell_Information_Response_List_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponseList_RL_SetupRspFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationResponseList_RL_SetupRspFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationResponseList_RL_SetupRspFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponseItem_RL_SetupRspFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationResponseItem_RL_SetupRspFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationResponseItem_RL_SetupRspFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkSetupResponseTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkSetupResponseTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkSetupResponseTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponse_RL_SetupRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationResponse_RL_SetupRspTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationResponse_RL_SetupRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponse_LCR_RL_SetupRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationResponse_LCR_RL_SetupRspTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationResponse_LCR_RL_SetupRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkSetupFailureFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkSetupFailureFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkSetupFailureFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CauseLevel_RL_SetupFailureFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CauseLevel_RL_SetupFailureFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CauseLevel_RL_SetupFailureFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Unsuccessful_RL_InformationRespItem_RL_SetupFailureFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Unsuccessful_RL_InformationRespItem_RL_SetupFailureFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Unsuccessful_RL_InformationRespItem_RL_SetupFailureFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Successful_RL_InformationRespItem_RL_SetupFailureFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Successful_RL_InformationRespItem_RL_SetupFailureFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Successful_RL_InformationRespItem_RL_SetupFailureFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkSetupFailureTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkSetupFailureTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkSetupFailureTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CauseLevel_RL_SetupFailureTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CauseLevel_RL_SetupFailureTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CauseLevel_RL_SetupFailureTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Unsuccessful_RL_InformationResp_RL_SetupFailureTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Unsuccessful_RL_InformationResp_RL_SetupFailureTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Unsuccessful_RL_InformationResp_RL_SetupFailureTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkAdditionRequestFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkAdditionRequestFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkAdditionRequestFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_HS_Cell_Information_RL_Addition_List_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Additional_HS_Cell_Information_RL_Addition_List(tvb, offset, &asn1_ctx, tree, hf_nbap_Additional_HS_Cell_Information_RL_Addition_List_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_EDCH_Cell_Information_RL_Add_Req_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Additional_EDCH_Cell_Information_RL_Add_Req(tvb, offset, &asn1_ctx, tree, hf_nbap_Additional_EDCH_Cell_Information_RL_Add_Req_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationList_RL_AdditionRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationList_RL_AdditionRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationList_RL_AdditionRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationItem_RL_AdditionRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationItem_RL_AdditionRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationItem_RL_AdditionRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DPCH_Information_RL_AdditionReqFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DPCH_Information_RL_AdditionReqFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DPCH_Information_RL_AdditionReqFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkAdditionRequestTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkAdditionRequestTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkAdditionRequestTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_InformationList_RL_AdditionRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_CCTrCH_InformationList_RL_AdditionRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_CCTrCH_InformationList_RL_AdditionRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_InformationItem_RL_AdditionRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_DPCH_InformationItem_RL_AdditionRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_DPCH_InformationItem_RL_AdditionRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_InformationItem_768_RL_AdditionRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_DPCH_InformationItem_768_RL_AdditionRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_DPCH_InformationItem_768_RL_AdditionRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationList_RL_AdditionRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_CCTrCH_InformationList_RL_AdditionRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_CCTrCH_InformationList_RL_AdditionRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_InformationItem_RL_AdditionRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_DPCH_InformationItem_RL_AdditionRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_DPCH_InformationItem_RL_AdditionRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_InformationItem_768_RL_AdditionRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_DPCH_InformationItem_768_RL_AdditionRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_DPCH_InformationItem_768_RL_AdditionRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Information_RL_AdditionRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_Information_RL_AdditionRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_Information_RL_AdditionRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkAdditionResponseFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkAdditionResponseFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkAdditionResponseFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_HS_Cell_Change_Information_Response_List_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Additional_HS_Cell_Change_Information_Response_List(tvb, offset, &asn1_ctx, tree, hf_nbap_Additional_HS_Cell_Change_Information_Response_List_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponseList_RL_AdditionRspFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationResponseList_RL_AdditionRspFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationResponseList_RL_AdditionRspFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponseItem_RL_AdditionRspFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationResponseItem_RL_AdditionRspFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationResponseItem_RL_AdditionRspFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkAdditionResponseTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkAdditionResponseTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkAdditionResponseTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponse_RL_AdditionRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationResponse_RL_AdditionRspTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationResponse_RL_AdditionRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponse_LCR_RL_AdditionRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationResponse_LCR_RL_AdditionRspTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationResponse_LCR_RL_AdditionRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkAdditionFailureFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkAdditionFailureFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkAdditionFailureFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CauseLevel_RL_AdditionFailureFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CauseLevel_RL_AdditionFailureFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CauseLevel_RL_AdditionFailureFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Unsuccessful_RL_InformationRespItem_RL_AdditionFailureFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Unsuccessful_RL_InformationRespItem_RL_AdditionFailureFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Unsuccessful_RL_InformationRespItem_RL_AdditionFailureFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Successful_RL_InformationRespItem_RL_AdditionFailureFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Successful_RL_InformationRespItem_RL_AdditionFailureFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Successful_RL_InformationRespItem_RL_AdditionFailureFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkAdditionFailureTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkAdditionFailureTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkAdditionFailureTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CauseLevel_RL_AdditionFailureTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CauseLevel_RL_AdditionFailureTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CauseLevel_RL_AdditionFailureTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Unsuccessful_RL_InformationResp_RL_AdditionFailureTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Unsuccessful_RL_InformationResp_RL_AdditionFailureTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Unsuccessful_RL_InformationResp_RL_AdditionFailureTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkReconfigurationPrepareFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkReconfigurationPrepareFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkReconfigurationPrepareFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_HS_Cell_Information_RL_Reconf_Prep_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Additional_HS_Cell_Information_RL_Reconf_Prep(tvb, offset, &asn1_ctx, tree, hf_nbap_Additional_HS_Cell_Information_RL_Reconf_Prep_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_EDCH_Cell_Information_RL_Reconf_Prep_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Additional_EDCH_Cell_Information_RL_Reconf_Prep(tvb, offset, &asn1_ctx, tree, hf_nbap_Additional_EDCH_Cell_Information_RL_Reconf_Prep_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_Information_RL_ReconfPrepFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_DPCH_Information_RL_ReconfPrepFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_DPCH_Information_RL_ReconfPrepFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_Information_RL_ReconfPrepFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_DPCH_Information_RL_ReconfPrepFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_DPCH_Information_RL_ReconfPrepFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_Power_Information_RL_ReconfPrepFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_DPCH_Power_Information_RL_ReconfPrepFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_DPCH_Power_Information_RL_ReconfPrepFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DCH_DeleteList_RL_ReconfPrepFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DCH_DeleteList_RL_ReconfPrepFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DCH_DeleteList_RL_ReconfPrepFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationList_RL_ReconfPrepFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationList_RL_ReconfPrepFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationList_RL_ReconfPrepFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationItem_RL_ReconfPrepFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationItem_RL_ReconfPrepFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationItem_RL_ReconfPrepFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DPCH_Information_RL_ReconfPrepFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DPCH_Information_RL_ReconfPrepFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DPCH_Information_RL_ReconfPrepFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_F_DPCH_Information_RL_ReconfPrepFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_F_DPCH_Information_RL_ReconfPrepFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_F_DPCH_Information_RL_ReconfPrepFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkReconfigurationPrepareTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkReconfigurationPrepareTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkReconfigurationPrepareTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_InformationAddItem_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_DPCH_InformationAddItem_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_DPCH_InformationAddItem_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MultipleRL_UL_DPCH_InformationAddList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MultipleRL_UL_DPCH_InformationAddList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_MultipleRL_UL_DPCH_InformationAddList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_768_InformationAddList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_DPCH_768_InformationAddList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_DPCH_768_InformationAddList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_TimeslotLCR_InformationModify_ModifyList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_TimeslotLCR_InformationModify_ModifyList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_TimeslotLCR_InformationModify_ModifyList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_Timeslot768_InformationModify_ModifyList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_Timeslot768_InformationModify_ModifyList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_Timeslot768_InformationModify_ModifyList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MultipleRL_UL_DPCH_InformationModifyList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MultipleRL_UL_DPCH_InformationModifyList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_MultipleRL_UL_DPCH_InformationModifyList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_InformationAddItem_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_DPCH_InformationAddItem_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_DPCH_InformationAddItem_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MultipleRL_DL_DPCH_InformationAddList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MultipleRL_DL_DPCH_InformationAddList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_MultipleRL_DL_DPCH_InformationAddList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_768_InformationAddList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_DPCH_768_InformationAddList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_DPCH_768_InformationAddList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_Timeslot_768_InformationModify_ModifyList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_Timeslot_768_InformationModify_ModifyList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_Timeslot_768_InformationModify_ModifyList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MultipleRL_DL_DPCH_InformationModifyList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MultipleRL_DL_DPCH_InformationModifyList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_MultipleRL_DL_DPCH_InformationModifyList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DCH_DeleteList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DCH_DeleteList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DCH_DeleteList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DSCH_Information_ModifyList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DSCH_Information_ModifyList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DSCH_Information_ModifyList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DSCH_Information_DeleteList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DSCH_Information_DeleteList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DSCH_Information_DeleteList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_USCH_Information_ModifyList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_USCH_Information_ModifyList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_USCH_Information_ModifyList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_USCH_Information_DeleteList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_USCH_Information_DeleteList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_USCH_Information_DeleteList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MultipleRL_Information_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MultipleRL_Information_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_MultipleRL_Information_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Information_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_Information_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_Information_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkReconfigurationReady_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkReconfigurationReady(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkReconfigurationReady_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponseList_RL_ReconfReady_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationResponseList_RL_ReconfReady(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationResponseList_RL_ReconfReady_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponseItem_RL_ReconfReady_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationResponseItem_RL_ReconfReady(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationResponseItem_RL_ReconfReady_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkReconfigurationFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkReconfigurationFailure(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkReconfigurationFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CauseLevel_RL_ReconfFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CauseLevel_RL_ReconfFailure(tvb, offset, &asn1_ctx, tree, hf_nbap_CauseLevel_RL_ReconfFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_ReconfigurationFailureItem_RL_ReconfFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_ReconfigurationFailureItem_RL_ReconfFailure(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_ReconfigurationFailureItem_RL_ReconfFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkReconfigurationCommit_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkReconfigurationCommit(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkReconfigurationCommit_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkReconfigurationCancel_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkReconfigurationCancel(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkReconfigurationCancel_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkReconfigurationRequestFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkReconfigurationRequestFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkReconfigurationRequestFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_HS_Cell_Information_RL_Reconf_Req_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Additional_HS_Cell_Information_RL_Reconf_Req(tvb, offset, &asn1_ctx, tree, hf_nbap_Additional_HS_Cell_Information_RL_Reconf_Req_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_EDCH_Cell_Information_RL_Reconf_Req_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Additional_EDCH_Cell_Information_RL_Reconf_Req(tvb, offset, &asn1_ctx, tree, hf_nbap_Additional_EDCH_Cell_Information_RL_Reconf_Req_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_Information_RL_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_DPCH_Information_RL_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_DPCH_Information_RL_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_Information_RL_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_DPCH_Information_RL_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_DPCH_Information_RL_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DCH_DeleteList_RL_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DCH_DeleteList_RL_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DCH_DeleteList_RL_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationList_RL_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationList_RL_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationList_RL_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationItem_RL_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationItem_RL_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationItem_RL_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DPCH_Information_RL_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DPCH_Information_RL_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DPCH_Information_RL_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkReconfigurationRequestTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkReconfigurationRequestTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkReconfigurationRequestTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MultipleRL_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MultipleRL_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_MultipleRL_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_DPCH_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_DPCH_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DCH_DeleteList_RL_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DCH_DeleteList_RL_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DCH_DeleteList_RL_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multiple_RL_Information_RL_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Multiple_RL_Information_RL_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Multiple_RL_Information_RL_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Information_RL_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_Information_RL_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_Information_RL_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkReconfigurationResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkReconfigurationResponse(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkReconfigurationResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponseList_RL_ReconfRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationResponseList_RL_ReconfRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationResponseList_RL_ReconfRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponseItem_RL_ReconfRsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationResponseItem_RL_ReconfRsp(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationResponseItem_RL_ReconfRsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkDeletionRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkDeletionRequest(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkDeletionRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_informationList_RL_DeletionRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_informationList_RL_DeletionRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_informationList_RL_DeletionRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_informationItem_RL_DeletionRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_informationItem_RL_DeletionRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_informationItem_RL_DeletionRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkDeletionResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkDeletionResponse(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkDeletionResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_PowerControlRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_PowerControlRequest(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_PowerControlRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_ReferencePowerInformationList_DL_PC_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_ReferencePowerInformationList_DL_PC_Rqst(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_ReferencePowerInformationList_DL_PC_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_ReferencePowerInformationItem_DL_PC_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_ReferencePowerInformationItem_DL_PC_Rqst(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_ReferencePowerInformationItem_DL_PC_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_PowerTimeslotControlRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_PowerTimeslotControlRequest(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_PowerTimeslotControlRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DedicatedMeasurementInitiationRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DedicatedMeasurementInitiationRequest(tvb, offset, &asn1_ctx, tree, hf_nbap_DedicatedMeasurementInitiationRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DedicatedMeasurementObjectType_DM_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DedicatedMeasurementObjectType_DM_Rqst(tvb, offset, &asn1_ctx, tree, hf_nbap_DedicatedMeasurementObjectType_DM_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationItem_DM_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationItem_DM_Rqst(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationItem_DM_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PUSCH_Info_DM_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PUSCH_Info_DM_Rqst(tvb, offset, &asn1_ctx, tree, hf_nbap_PUSCH_Info_DM_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSSICH_Info_DM_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSSICH_Info_DM_Rqst(tvb, offset, &asn1_ctx, tree, hf_nbap_HSSICH_Info_DM_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSSICH_InfoExt_DM_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSSICH_InfoExt_DM_Rqst(tvb, offset, &asn1_ctx, tree, hf_nbap_HSSICH_InfoExt_DM_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DedicatedMeasurementInitiationResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DedicatedMeasurementInitiationResponse(tvb, offset, &asn1_ctx, tree, hf_nbap_DedicatedMeasurementInitiationResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DedicatedMeasurementObjectType_DM_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DedicatedMeasurementObjectType_DM_Rsp(tvb, offset, &asn1_ctx, tree, hf_nbap_DedicatedMeasurementObjectType_DM_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationItem_DM_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationItem_DM_Rsp(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationItem_DM_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PUSCH_Info_DM_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PUSCH_Info_DM_Rsp(tvb, offset, &asn1_ctx, tree, hf_nbap_PUSCH_Info_DM_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multiple_PUSCH_InfoList_DM_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Multiple_PUSCH_InfoList_DM_Rsp(tvb, offset, &asn1_ctx, tree, hf_nbap_Multiple_PUSCH_InfoList_DM_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp(tvb, offset, &asn1_ctx, tree, hf_nbap_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp(tvb, offset, &asn1_ctx, tree, hf_nbap_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp(tvb, offset, &asn1_ctx, tree, hf_nbap_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multiple_DedicatedMeasurementValueList_768_TDD_DM_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Multiple_DedicatedMeasurementValueList_768_TDD_DM_Rsp(tvb, offset, &asn1_ctx, tree, hf_nbap_Multiple_DedicatedMeasurementValueList_768_TDD_DM_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Set_InformationItem_DM_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_Set_InformationItem_DM_Rsp(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_Set_InformationItem_DM_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DedicatedMeasurementInitiationFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DedicatedMeasurementInitiationFailure(tvb, offset, &asn1_ctx, tree, hf_nbap_DedicatedMeasurementInitiationFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DedicatedMeasurementReport_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DedicatedMeasurementReport(tvb, offset, &asn1_ctx, tree, hf_nbap_DedicatedMeasurementReport_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DedicatedMeasurementObjectType_DM_Rprt_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DedicatedMeasurementObjectType_DM_Rprt(tvb, offset, &asn1_ctx, tree, hf_nbap_DedicatedMeasurementObjectType_DM_Rprt_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationItem_DM_Rprt_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationItem_DM_Rprt(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationItem_DM_Rprt_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PUSCH_Info_DM_Rprt_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PUSCH_Info_DM_Rprt(tvb, offset, &asn1_ctx, tree, hf_nbap_PUSCH_Info_DM_Rprt_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multiple_PUSCH_InfoList_DM_Rprt_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Multiple_PUSCH_InfoList_DM_Rprt(tvb, offset, &asn1_ctx, tree, hf_nbap_Multiple_PUSCH_InfoList_DM_Rprt_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Set_InformationItem_DM_Rprt_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_Set_InformationItem_DM_Rprt(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_Set_InformationItem_DM_Rprt_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DedicatedMeasurementTerminationRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DedicatedMeasurementTerminationRequest(tvb, offset, &asn1_ctx, tree, hf_nbap_DedicatedMeasurementTerminationRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DedicatedMeasurementFailureIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DedicatedMeasurementFailureIndication(tvb, offset, &asn1_ctx, tree, hf_nbap_DedicatedMeasurementFailureIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkFailureIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkFailureIndication(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkFailureIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Reporting_Object_RL_FailureInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Reporting_Object_RL_FailureInd(tvb, offset, &asn1_ctx, tree, hf_nbap_Reporting_Object_RL_FailureInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationItem_RL_FailureInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationItem_RL_FailureInd(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationItem_RL_FailureInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Set_InformationItem_RL_FailureInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_Set_InformationItem_RL_FailureInd(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_Set_InformationItem_RL_FailureInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CCTrCH_InformationItem_RL_FailureInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CCTrCH_InformationItem_RL_FailureInd(tvb, offset, &asn1_ctx, tree, hf_nbap_CCTrCH_InformationItem_RL_FailureInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkPreemptionRequiredIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkPreemptionRequiredIndication(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkPreemptionRequiredIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationList_RL_PreemptRequiredInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationList_RL_PreemptRequiredInd(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationList_RL_PreemptRequiredInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationItem_RL_PreemptRequiredInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationItem_RL_PreemptRequiredInd(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationItem_RL_PreemptRequiredInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkRestoreIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkRestoreIndication(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkRestoreIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Reporting_Object_RL_RestoreInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Reporting_Object_RL_RestoreInd(tvb, offset, &asn1_ctx, tree, hf_nbap_Reporting_Object_RL_RestoreInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationItem_RL_RestoreInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_InformationItem_RL_RestoreInd(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_InformationItem_RL_RestoreInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Set_InformationItem_RL_RestoreInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RL_Set_InformationItem_RL_RestoreInd(tvb, offset, &asn1_ctx, tree, hf_nbap_RL_Set_InformationItem_RL_RestoreInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CCTrCH_InformationItem_RL_RestoreInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CCTrCH_InformationItem_RL_RestoreInd(tvb, offset, &asn1_ctx, tree, hf_nbap_CCTrCH_InformationItem_RL_RestoreInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CompressedModeCommand_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CompressedModeCommand(tvb, offset, &asn1_ctx, tree, hf_nbap_CompressedModeCommand_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ErrorIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ErrorIndication(tvb, offset, &asn1_ctx, tree, hf_nbap_ErrorIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PrivateMessage_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PrivateMessage(tvb, offset, &asn1_ctx, tree, hf_nbap_PrivateMessage_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PhysicalSharedChannelReconfigurationRequestFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PhysicalSharedChannelReconfigurationRequestFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PhysicalSharedChannelReconfigurationRequestFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDPA_And_EDCH_CellPortion_InformationList_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDPA_And_EDCH_CellPortion_InformationList_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDPA_And_EDCH_CellPortion_InformationList_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PhysicalSharedChannelReconfigurationRequestTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PhysicalSharedChannelReconfigurationRequestTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_PhysicalSharedChannelReconfigurationRequestTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PDSCHSets_AddList_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PDSCHSets_AddList_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_PDSCHSets_AddList_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PDSCH_Information_AddItem_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PDSCH_Information_AddItem_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_PDSCH_Information_AddItem_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PDSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PDSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_PDSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PDSCH_AddInformation_768_AddItem_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PDSCH_AddInformation_768_AddItem_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_PDSCH_AddInformation_768_AddItem_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PDSCHSets_ModifyList_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PDSCHSets_ModifyList_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_PDSCHSets_ModifyList_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PDSCH_Information_ModifyItem_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PDSCH_Information_ModifyItem_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_PDSCH_Information_ModifyItem_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PDSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PDSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_PDSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PDSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PDSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_PDSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PDSCHSets_DeleteList_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PDSCHSets_DeleteList_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_PDSCHSets_DeleteList_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PUSCHSets_AddList_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PUSCHSets_AddList_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_PUSCHSets_AddList_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PUSCH_Information_AddItem_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PUSCH_Information_AddItem_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_PUSCH_Information_AddItem_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PUSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PUSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_PUSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PUSCH_AddInformation_768_AddItem_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PUSCH_AddInformation_768_AddItem_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_PUSCH_AddInformation_768_AddItem_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PUSCHSets_ModifyList_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PUSCHSets_ModifyList_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_PUSCHSets_ModifyList_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PUSCH_Information_ModifyItem_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PUSCH_Information_ModifyItem_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_PUSCH_Information_ModifyItem_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PUSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PUSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_PUSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PUSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PUSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_PUSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PUSCHSets_DeleteList_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PUSCHSets_DeleteList_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_PUSCHSets_DeleteList_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_PDSCH_TDD_Information_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_PDSCH_TDD_Information_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_PDSCH_TDD_Information_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_HS_PDSCH_Timeslot_Information_768_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DL_HS_PDSCH_Timeslot_Information_768_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_DL_HS_PDSCH_Timeslot_Information_768_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Add_To_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Add_To_HS_SCCH_Resource_Pool_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_Add_To_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_SCCH_Information_768_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_SCCH_Information_768_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_SCCH_Information_768_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_SCCH_InformationExt_LCR_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_SCCH_InformationExt_LCR_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_SCCH_InformationExt_LCR_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Modify_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Modify_HS_SCCH_Resource_Pool_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_Modify_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_SCCH_InformationModifyExt_LCR_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_SCCH_InformationModifyExt_LCR_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_SCCH_InformationModifyExt_LCR_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_SCCH_InformationModify_768_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HS_SCCH_InformationModify_768_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_HS_SCCH_InformationModify_768_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Delete_From_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Delete_From_HS_SCCH_Resource_Pool_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_Delete_From_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_PUCH_Information_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_PUCH_Information_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_E_PUCH_Information_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Add_To_E_AGCH_Resource_Pool_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Add_To_E_AGCH_Resource_Pool_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_Add_To_E_AGCH_Resource_Pool_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Modify_E_AGCH_Resource_Pool_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Modify_E_AGCH_Resource_Pool_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_Modify_E_AGCH_Resource_Pool_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Delete_From_E_AGCH_Resource_Pool_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Delete_From_E_AGCH_Resource_Pool_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_Delete_From_E_AGCH_Resource_Pool_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_HICH_Information_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_HICH_Information_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_E_HICH_Information_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_PUCH_Information_768_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_PUCH_Information_768_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_E_PUCH_Information_768_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Add_To_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Add_To_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_Add_To_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Modify_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Modify_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_Modify_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_HICH_Information_768_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_HICH_Information_768_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_E_HICH_Information_768_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_PUCH_Information_LCR_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_PUCH_Information_LCR_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_E_PUCH_Information_LCR_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Add_To_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Add_To_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_Add_To_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Modify_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Modify_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_Modify_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Add_To_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Add_To_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_Add_To_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Modify_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Modify_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_Modify_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Delete_From_E_HICH_Resource_Pool_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Delete_From_E_HICH_Resource_Pool_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_Delete_From_E_HICH_Resource_Pool_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SYNC_UL_Partition_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SYNC_UL_Partition_LCR(tvb, offset, &asn1_ctx, tree, hf_nbap_SYNC_UL_Partition_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Delete_From_HS_SCCH_Resource_PoolExt_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Delete_From_HS_SCCH_Resource_PoolExt_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_Delete_From_HS_SCCH_Resource_PoolExt_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MultipleFreq_E_PUCH_Timeslot_InformationList_LCR_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MultipleFreq_E_PUCH_Timeslot_InformationList_LCR_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_MultipleFreq_E_PUCH_Timeslot_InformationList_LCR_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MultipleFreq_E_PUCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MultipleFreq_E_PUCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_MultipleFreq_E_PUCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst(tvb, offset, &asn1_ctx, tree, hf_nbap_Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PhysicalSharedChannelReconfigurationResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PhysicalSharedChannelReconfigurationResponse(tvb, offset, &asn1_ctx, tree, hf_nbap_PhysicalSharedChannelReconfigurationResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_HICH_TimeOffset_ExtensionLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_HICH_TimeOffset_ExtensionLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_E_HICH_TimeOffset_ExtensionLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MultipleFreq_E_HICH_TimeOffsetLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MultipleFreq_E_HICH_TimeOffsetLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_MultipleFreq_E_HICH_TimeOffsetLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PhysicalSharedChannelReconfigurationFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_PhysicalSharedChannelReconfigurationFailure(tvb, offset, &asn1_ctx, tree, hf_nbap_PhysicalSharedChannelReconfigurationFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CauseLevel_PSCH_ReconfFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CauseLevel_PSCH_ReconfFailure(tvb, offset, &asn1_ctx, tree, hf_nbap_CauseLevel_PSCH_ReconfFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Unsuccessful_PDSCHSetItem_PSCH_ReconfFailureTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Unsuccessful_PDSCHSetItem_PSCH_ReconfFailureTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Unsuccessful_PDSCHSetItem_PSCH_ReconfFailureTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Unsuccessful_PUSCHSetItem_PSCH_ReconfFailureTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Unsuccessful_PUSCHSetItem_PSCH_ReconfFailureTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Unsuccessful_PUSCHSetItem_PSCH_ReconfFailureTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UARFCNSpecificCauseList_PSCH_ReconfFailureTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UARFCNSpecificCauseList_PSCH_ReconfFailureTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_UARFCNSpecificCauseList_PSCH_ReconfFailureTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Unsuccessful_UARFCNItem_PSCH_ReconfFailureTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Unsuccessful_UARFCNItem_PSCH_ReconfFailureTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Unsuccessful_UARFCNItem_PSCH_ReconfFailureTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_HICH_TimeOffset_ReconfFailureTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_HICH_TimeOffset_ReconfFailureTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_E_HICH_TimeOffset_ReconfFailureTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Common_System_Information_ResponseLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Common_System_Information_ResponseLCR(tvb, offset, &asn1_ctx, tree, hf_nbap_Common_System_Information_ResponseLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ResetRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ResetRequest(tvb, offset, &asn1_ctx, tree, hf_nbap_ResetRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ResetIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ResetIndicator(tvb, offset, &asn1_ctx, tree, hf_nbap_ResetIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommunicationContextInfoItem_Reset_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommunicationContextInfoItem_Reset(tvb, offset, &asn1_ctx, tree, hf_nbap_CommunicationContextInfoItem_Reset_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommunicationControlPortInfoItem_Reset_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CommunicationControlPortInfoItem_Reset(tvb, offset, &asn1_ctx, tree, hf_nbap_CommunicationControlPortInfoItem_Reset_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ResetResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_ResetResponse(tvb, offset, &asn1_ctx, tree, hf_nbap_ResetResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationExchangeInitiationRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_InformationExchangeInitiationRequest(tvb, offset, &asn1_ctx, tree, hf_nbap_InformationExchangeInitiationRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationExchangeObjectType_InfEx_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_InformationExchangeObjectType_InfEx_Rqst(tvb, offset, &asn1_ctx, tree, hf_nbap_InformationExchangeObjectType_InfEx_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationExchangeInitiationResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_InformationExchangeInitiationResponse(tvb, offset, &asn1_ctx, tree, hf_nbap_InformationExchangeInitiationResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationExchangeObjectType_InfEx_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_InformationExchangeObjectType_InfEx_Rsp(tvb, offset, &asn1_ctx, tree, hf_nbap_InformationExchangeObjectType_InfEx_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationExchangeInitiationFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_InformationExchangeInitiationFailure(tvb, offset, &asn1_ctx, tree, hf_nbap_InformationExchangeInitiationFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationReport_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_InformationReport(tvb, offset, &asn1_ctx, tree, hf_nbap_InformationReport_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationExchangeObjectType_InfEx_Rprt_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_InformationExchangeObjectType_InfEx_Rprt(tvb, offset, &asn1_ctx, tree, hf_nbap_InformationExchangeObjectType_InfEx_Rprt_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationExchangeTerminationRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_InformationExchangeTerminationRequest(tvb, offset, &asn1_ctx, tree, hf_nbap_InformationExchangeTerminationRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationExchangeFailureIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_InformationExchangeFailureIndication(tvb, offset, &asn1_ctx, tree, hf_nbap_InformationExchangeFailureIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellSynchronisationInitiationRequestTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellSynchronisationInitiationRequestTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellSynchronisationInitiationRequestTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TimeslotInfo_CellSyncInitiationRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_TimeslotInfo_CellSyncInitiationRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_TimeslotInfo_CellSyncInitiationRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellSyncBurstTransInit_CellSyncInitiationRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellSyncBurstTransInit_CellSyncInitiationRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellSyncBurstTransInit_CellSyncInitiationRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellSyncBurstMeasureInit_CellSyncInitiationRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellSyncBurstMeasureInit_CellSyncInitiationRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellSyncBurstMeasureInit_CellSyncInitiationRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SYNCDlCodeId_TransInitLCR_CellSyncInitiationRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SYNCDlCodeId_TransInitLCR_CellSyncInitiationRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_SYNCDlCodeId_TransInitLCR_CellSyncInitiationRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SYNCDlCodeId_MeasureInitLCR_CellSyncInitiationRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SYNCDlCodeId_MeasureInitLCR_CellSyncInitiationRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_SYNCDlCodeId_MeasureInitLCR_CellSyncInitiationRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellSynchronisationInitiationResponseTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellSynchronisationInitiationResponseTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellSynchronisationInitiationResponseTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellSynchronisationInitiationFailureTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellSynchronisationInitiationFailureTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellSynchronisationInitiationFailureTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellSynchronisationReconfigurationRequestTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellSynchronisationReconfigurationRequestTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellSynchronisationReconfigurationRequestTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellSyncBurstTransReconfInfo_CellSyncReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellSyncBurstTransReconfInfo_CellSyncReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellSyncBurstTransReconfInfo_CellSyncReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellSyncBurstMeasInfo_CellSyncReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellSyncBurstMeasInfo_CellSyncReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellSyncBurstMeasInfo_CellSyncReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellSyncBurstMeasInfoListIE_CellSyncReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellSyncBurstMeasInfoListIE_CellSyncReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellSyncBurstMeasInfoListIE_CellSyncReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SYNCDlCodeIdTransReconfInfoLCR_CellSyncReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SYNCDlCodeIdTransReconfInfoLCR_CellSyncReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_SYNCDlCodeIdTransReconfInfoLCR_CellSyncReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SYNCDlCodeIdMeasInfoLCR_CellSyncReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SYNCDlCodeIdMeasInfoLCR_CellSyncReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_SYNCDlCodeIdMeasInfoLCR_CellSyncReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellSynchronisationReconfigurationResponseTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellSynchronisationReconfigurationResponseTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellSynchronisationReconfigurationResponseTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellSynchronisationReconfigurationFailureTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellSynchronisationReconfigurationFailureTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellSynchronisationReconfigurationFailureTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellSynchronisationAdjustmentRequestTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellSynchronisationAdjustmentRequestTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellSynchronisationAdjustmentRequestTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellAdjustmentInfo_SyncAdjustmentRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellAdjustmentInfo_SyncAdjustmentRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellAdjustmentInfo_SyncAdjustmentRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellAdjustmentInfoItem_SyncAdjustmentRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellAdjustmentInfoItem_SyncAdjustmentRqstTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellAdjustmentInfoItem_SyncAdjustmentRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellSynchronisationAdjustmentResponseTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellSynchronisationAdjustmentResponseTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellSynchronisationAdjustmentResponseTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellSynchronisationAdjustmentFailureTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellSynchronisationAdjustmentFailureTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellSynchronisationAdjustmentFailureTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CauseLevel_SyncAdjustmntFailureTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CauseLevel_SyncAdjustmntFailureTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CauseLevel_SyncAdjustmntFailureTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Unsuccessful_cell_InformationRespItem_SyncAdjustmntFailureTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Unsuccessful_cell_InformationRespItem_SyncAdjustmntFailureTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_Unsuccessful_cell_InformationRespItem_SyncAdjustmntFailureTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellSynchronisationTerminationRequestTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellSynchronisationTerminationRequestTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellSynchronisationTerminationRequestTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellSynchronisationFailureIndicationTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellSynchronisationFailureIndicationTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellSynchronisationFailureIndicationTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellSynchronisationReportTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellSynchronisationReportTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellSynchronisationReportTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellSyncInfo_CellSyncReprtTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_CellSyncInfo_CellSyncReprtTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_CellSyncInfo_CellSyncReprtTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SyncReportType_CellSyncReprtTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SyncReportType_CellSyncReprtTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_SyncReportType_CellSyncReprtTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SyncDLCodeIdsMeasInfoList_CellSyncReprtTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SyncDLCodeIdsMeasInfoList_CellSyncReprtTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_SyncDLCodeIdsMeasInfoList_CellSyncReprtTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_BearerRearrangementIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_BearerRearrangementIndication(tvb, offset, &asn1_ctx, tree, hf_nbap_BearerRearrangementIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DCH_RearrangeList_Bearer_RearrangeInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DCH_RearrangeList_Bearer_RearrangeInd(tvb, offset, &asn1_ctx, tree, hf_nbap_DCH_RearrangeList_Bearer_RearrangeInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DSCH_RearrangeList_Bearer_RearrangeInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DSCH_RearrangeList_Bearer_RearrangeInd(tvb, offset, &asn1_ctx, tree, hf_nbap_DSCH_RearrangeList_Bearer_RearrangeInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_USCH_RearrangeList_Bearer_RearrangeInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_USCH_RearrangeList_Bearer_RearrangeInd(tvb, offset, &asn1_ctx, tree, hf_nbap_USCH_RearrangeList_Bearer_RearrangeInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_RearrangeList_Bearer_RearrangeInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_HSDSCH_RearrangeList_Bearer_RearrangeInd(tvb, offset, &asn1_ctx, tree, hf_nbap_HSDSCH_RearrangeList_Bearer_RearrangeInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_RearrangeList_Bearer_RearrangeInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_E_DCH_RearrangeList_Bearer_RearrangeInd(tvb, offset, &asn1_ctx, tree, hf_nbap_E_DCH_RearrangeList_Bearer_RearrangeInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_EDCH_Cell_Information_Bearer_Rearrangement_List_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Additional_EDCH_Cell_Information_Bearer_Rearrangement_List(tvb, offset, &asn1_ctx, tree, hf_nbap_Additional_EDCH_Cell_Information_Bearer_Rearrangement_List_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkActivationCommandFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkActivationCommandFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkActivationCommandFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DelayedActivationInformationList_RL_ActivationCmdFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DelayedActivationInformationList_RL_ActivationCmdFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DelayedActivationInformationList_RL_ActivationCmdFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DelayedActivationInformation_RL_ActivationCmdFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DelayedActivationInformation_RL_ActivationCmdFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DelayedActivationInformation_RL_ActivationCmdFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkActivationCommandTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkActivationCommandTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkActivationCommandTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DelayedActivationInformationList_RL_ActivationCmdTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DelayedActivationInformationList_RL_ActivationCmdTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DelayedActivationInformationList_RL_ActivationCmdTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DelayedActivationInformation_RL_ActivationCmdTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_DelayedActivationInformation_RL_ActivationCmdTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_DelayedActivationInformation_RL_ActivationCmdTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkParameterUpdateIndicationFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkParameterUpdateIndicationFDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkParameterUpdateIndicationFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_HS_Cell_Information_RL_Param_Upd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Additional_HS_Cell_Information_RL_Param_Upd(tvb, offset, &asn1_ctx, tree, hf_nbap_Additional_HS_Cell_Information_RL_Param_Upd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_EDCH_Cell_Information_RL_Param_Upd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_Additional_EDCH_Cell_Information_RL_Param_Upd(tvb, offset, &asn1_ctx, tree, hf_nbap_Additional_EDCH_Cell_Information_RL_Param_Upd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkParameterUpdateIndicationTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_RadioLinkParameterUpdateIndicationTDD(tvb, offset, &asn1_ctx, tree, hf_nbap_RadioLinkParameterUpdateIndicationTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MBMSNotificationUpdateCommand_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_MBMSNotificationUpdateCommand(tvb, offset, &asn1_ctx, tree, hf_nbap_MBMSNotificationUpdateCommand_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UEStatusUpdateCommand_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_UEStatusUpdateCommand(tvb, offset, &asn1_ctx, tree, hf_nbap_UEStatusUpdateCommand_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SecondaryULFrequencyReport_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SecondaryULFrequencyReport(tvb, offset, &asn1_ctx, tree, hf_nbap_SecondaryULFrequencyReport_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SecondaryULFrequencyUpdateIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_SecondaryULFrequencyUpdateIndication(tvb, offset, &asn1_ctx, tree, hf_nbap_SecondaryULFrequencyUpdateIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NBAP_PDU_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_NBAP_PDU(tvb, offset, &asn1_ctx, tree, hf_nbap_NBAP_PDU_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NULL_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_nbap_NULL(tvb, offset, &asn1_ctx, tree, hf_nbap_NULL_PDU);
  offset += 7; offset >>= 3;
  return offset;
}


static int dissect_ProtocolIEFieldValue(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  guint32 protocol_ie_id;
  protocol_ie_id = nbap_get_private_data(pinfo)->protocol_ie_id;
  return (dissector_try_uint_new(nbap_ies_dissector_table, protocol_ie_id, tvb, pinfo, tree, FALSE, NULL)) ? tvb_captured_length(tvb) : 0;
}

static int dissect_ProtocolExtensionFieldExtensionValue(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  guint32 protocol_ie_id;
  protocol_ie_id = nbap_get_private_data(pinfo)->protocol_ie_id;
  return (dissector_try_uint_new(nbap_extension_dissector_table, protocol_ie_id, tvb, pinfo, tree, FALSE, NULL)) ? tvb_captured_length(tvb) : 0;
}

static int dissect_InitiatingMessageValue(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  if (!ProcedureID) return 0;
  return (dissector_try_string(nbap_proc_imsg_dissector_table, ProcedureID, tvb, pinfo, tree, NULL)) ? tvb_captured_length(tvb) : 0;
}

static int dissect_SuccessfulOutcomeValue(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  if (!ProcedureID) return 0;
  return (dissector_try_string(nbap_proc_sout_dissector_table, ProcedureID, tvb, pinfo, tree, NULL)) ? tvb_captured_length(tvb) : 0;
}

static int dissect_UnsuccessfulOutcomeValue(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  if (!ProcedureID) return 0;
  return (dissector_try_string(nbap_proc_uout_dissector_table, ProcedureID, tvb, pinfo, tree, NULL)) ? tvb_captured_length(tvb) : 0;
}
static void add_hsdsch_bind(packet_info *pinfo){
  address null_addr;
  conversation_t *conversation = NULL;
  umts_fp_conversation_info_t *umts_fp_conversation_info;
  fp_hsdsch_channel_info_t* fp_hsdsch_channel_info = NULL;
  guint32 i;
  nbap_private_data_t* nbap_private_data;
  nbap_hsdsch_channel_info_t* nbap_hsdsch_channel_info;

  if (PINFO_FD_VISITED(pinfo)){
    return;
  }

  nbap_private_data = nbap_get_private_data(pinfo);
  nbap_hsdsch_channel_info = nbap_private_data->nbap_hsdsch_channel_info;
  /* Set port to zero use that as an indication of whether we have data or not */
  clear_address(&null_addr);
  for (i = 0; i < maxNrOfMACdFlows; i++) {
    if (nbap_hsdsch_channel_info[i].crnc_port != 0){
      conversation = find_conversation(pinfo->num, &(nbap_hsdsch_channel_info[i].crnc_address), &null_addr, CONVERSATION_UDP,
                                      nbap_hsdsch_channel_info[i].crnc_port, 0, NO_ADDR_B);

      if (conversation == NULL) {
        /* It's not part of any conversation - create a new one. */
        conversation = conversation_new(pinfo->num, &(nbap_hsdsch_channel_info[i].crnc_address), &null_addr, CONVERSATION_UDP,
                                       nbap_hsdsch_channel_info[i].crnc_port, 0, NO_ADDR2|NO_PORT2);

        /* Set dissector */
        conversation_set_dissector(conversation, fp_handle);

        if(pinfo->link_dir==P2P_DIR_DL){
          umts_fp_conversation_info = wmem_new0(wmem_file_scope(), umts_fp_conversation_info_t);
          /* Fill in the HSDSCH relevant data */
          umts_fp_conversation_info->iface_type        = IuB_Interface;
          umts_fp_conversation_info->division          = Division_FDD;
          umts_fp_conversation_info->channel           = CHANNEL_HSDSCH;
          umts_fp_conversation_info->dl_frame_number   = 0;
          umts_fp_conversation_info->ul_frame_number   = pinfo->num;
          copy_address_wmem(wmem_file_scope(), &(umts_fp_conversation_info->crnc_address), &nbap_hsdsch_channel_info[i].crnc_address);
          umts_fp_conversation_info->crnc_port         = nbap_hsdsch_channel_info[i].crnc_port;

          fp_hsdsch_channel_info = wmem_new0(wmem_file_scope(), fp_hsdsch_channel_info_t);
          umts_fp_conversation_info->channel_specific_info = (void*)fp_hsdsch_channel_info;
          /*Added june 3, normally just the iterator variable*/
          fp_hsdsch_channel_info->hsdsch_macdflow_id = i ; /*hsdsch_macdflow_ids[i];*/ /* hsdsch_macdflow_id;*/

          if (nbap_private_data->crnc_context_present) {
            umts_fp_conversation_info->com_context_id = nbap_private_data->com_context_id;
          } else {
            /* XXX: This expert info doesn't get added in subsequent passes,
             * but probably should.
             */
            expert_add_info(pinfo, NULL, &ei_nbap_no_set_comm_context_id);
          }

          /* Cheat and use the DCH entries */
          umts_fp_conversation_info->num_dch_in_flow++;
          umts_fp_conversation_info->dch_ids_in_flow_list[umts_fp_conversation_info->num_dch_in_flow -1] = i;

          if(nbap_hsdsch_channel_info[i].entity == entity_not_specified ){
            /*Error*/
            expert_add_info(pinfo, NULL, &ei_nbap_hsdsch_entity_not_specified);
          }else{
            fp_hsdsch_channel_info->hsdsch_entity = (enum fp_hsdsch_entity)nbap_hsdsch_channel_info[i].entity;
          }
          umts_fp_conversation_info->rlc_mode = nbap_hsdsch_channel_info[i].rlc_mode;
          set_umts_fp_conv_data(conversation, umts_fp_conversation_info);
        }
      }
    }
  }
}

/*
 * Function used to manage conversation declared in Setup Request/Response message
 */
static guint32 calculate_setup_conv_key(const guint32 transaction_id, const guint32 dd_mode, const guint32 channel_id)
{
  /* We need to pack 3 values on 32 bits:
   * 31-16 transaction_id
   * 15-14 dd_mode
   * 13-0  channel_id
   */
  guint32 key;
  key = transaction_id << 16;
  key |= (dd_mode & 0x03) << 14;
  key |= (channel_id & 0x3fff);
  nbap_debug("\tCalculating key 0x%04x", key);
  return key;
}

static void add_setup_conv(const packet_info *pinfo _U_, const guint32 transaction_id, const guint32 dd_mode, const guint32 channel_id, const guint32 req_frame_number,
              const address *addr, const guint32 port, umts_fp_conversation_info_t * umts_fp_conversation_info, conversation_t *conv)
{
  nbap_setup_conv_t *new_conv = NULL;
  guint32 key;

  nbap_debug("Creating new setup conv\t TransactionID: %u\tddMode: %u\tChannelID: %u\t %s:%u",
  transaction_id, dd_mode, channel_id, address_to_str(pinfo->pool, addr), port);

  new_conv = wmem_new0(wmem_file_scope(), nbap_setup_conv_t);

  /* fill with data */
  new_conv->transaction_id = transaction_id;
  new_conv->dd_mode = dd_mode;
  new_conv->channel_id = channel_id;
  new_conv->request_frame_number = req_frame_number;
  copy_address_wmem(wmem_file_scope(), &new_conv->addr, addr);
  new_conv->port = port;
  new_conv->umts_fp_conversation_info = umts_fp_conversation_info;
  new_conv->conv = conv;

  key = calculate_setup_conv_key(new_conv->transaction_id, new_conv->dd_mode, new_conv->channel_id);

  wmem_map_insert(nbap_setup_conv_table, GUINT_TO_POINTER(key), new_conv);
}

static nbap_setup_conv_t* find_setup_conv(const packet_info *pinfo _U_, const guint32 transaction_id, const guint32 dd_mode, const guint32 channel_id)
{
  nbap_setup_conv_t *conv;
  guint32 key;
  nbap_debug("Looking for Setup Conversation match\t TransactionID: %u\t ddMode: %u\t ChannelID: %u", transaction_id, dd_mode, channel_id);

  key = calculate_setup_conv_key(transaction_id, dd_mode, channel_id);

  conv = (nbap_setup_conv_t*) wmem_map_lookup(nbap_setup_conv_table, GUINT_TO_POINTER(key));

  if(conv == NULL){
    nbap_debug("\tDidn't find Setup Conversation match");
  }else{
    nbap_debug("\tFOUND Setup Conversation match\t TransactionID: %u\t ddMode: %u\t ChannelID: %u\t %s:%u",
         conv->transaction_id, conv->dd_mode, conv->channel_id, address_to_str(pinfo->pool, &(conv->addr)), conv->port);
  }

  return conv;
}

static void delete_setup_conv(nbap_setup_conv_t *conv)
{
  guint32 key;

  /* check if conversation exist */
  if(conv == NULL){
    nbap_debug("Trying delete Setup Conversation that does not exist (ptr == NULL)\t");
    return;
  }
  key = calculate_setup_conv_key(conv->transaction_id, conv->dd_mode, conv->channel_id);
  wmem_map_remove(nbap_setup_conv_table, GUINT_TO_POINTER(key));
}

static void nbap_init(void){
  guint8 i;
  /*Initialize*/
  com_context_map = wmem_tree_new_autoreset(wmem_epan_scope(), wmem_file_scope());

  /*Initialize structure for muxed flow indication*/
  edch_flow_port_map = wmem_tree_new_autoreset(wmem_epan_scope(), wmem_file_scope());

  /*Initialize Setup Conversation hash table*/
  nbap_setup_conv_table = wmem_map_new(wmem_file_scope(), g_direct_hash, g_direct_equal);
  /*Initializing Scrambling Code to C-RNC Context & C-RNC Context to U-RNTI maps*/
  nbap_scrambling_code_crncc_map = wmem_tree_new_autoreset(wmem_epan_scope(), wmem_file_scope());
  nbap_crncc_urnti_map = wmem_tree_new_autoreset(wmem_epan_scope(), wmem_file_scope());

  for (i = 0; i < 15; i++) {
    lchId_type_table[i+1] = lch_contents[i];
  }
}

static int
dissect_nbap(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data)
{
  proto_item *nbap_item = NULL;
  proto_tree *nbap_tree = NULL;
  /* make entry in the Protocol column on summary display */
  col_set_str(pinfo->cinfo, COL_PROTOCOL, "NBAP");

  /* create the nbap protocol tree */
  nbap_item = proto_tree_add_item(tree, proto_nbap, tvb, 0, -1, ENC_NA);
  nbap_tree = proto_item_add_subtree(nbap_item, ett_nbap);

  /* Clearing any old 'private data' stored */
  nbap_reset_private_data(pinfo);

  return dissect_NBAP_PDU_PDU(tvb, pinfo, nbap_tree, data);
}

/* Highest ProcedureCode value, used in heuristics */
#define NBAP_MAX_PC 56 /* id-secondaryULFrequencyUpdate = 56*/
#define NBAP_MSG_MIN_LENGTH 7
static gboolean
dissect_nbap_heur(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data)
{
  guint8 pdu_type;
  guint8 procedure_id;
  guint8 dd_mode;
  guint8 criticality;
  guint8 transaction_id_type;
  guint length;
  int length_field_offset;

  #define PDU_TYPE_OFFSET 0
  #define PROC_CODE_OFFSET 1
  #define DD_CRIT_OFFSET 2
  if (tvb_captured_length(tvb) < NBAP_MSG_MIN_LENGTH) {
    return FALSE;
  }

  pdu_type = tvb_get_guint8(tvb, PDU_TYPE_OFFSET);
  if (pdu_type & 0x1f) {
    /* pdu_type is not 0x00 (initiatingMessage), 0x20 (succesfulOutcome),
       0x40 (unsuccesfulOutcome) or 0x60 (outcome), ignore extension bit (0x80) */
    return FALSE;
  }

  procedure_id = tvb_get_guint8(tvb, PROC_CODE_OFFSET);
  if (procedure_id > NBAP_MAX_PC) {
      return FALSE;
  }

  dd_mode = tvb_get_guint8(tvb, DD_CRIT_OFFSET) >> 5;
  if (dd_mode >= 0x03) {
    /* dd_mode is not 0x00 (tdd), 0x01 (fdd) or 0x02 (common) */
    return FALSE;
  }

  criticality = (tvb_get_guint8(tvb, DD_CRIT_OFFSET) & 0x18) >> 3;
  if (criticality == 0x03) {
    /* criticality is not 0x00 (reject), 0x01 (ignore) or 0x02 (notify) */
    return FALSE;
  }

  /* Finding the offset for the length field - depends on wether the transaction id is long or short */
  transaction_id_type = (tvb_get_guint8(tvb, DD_CRIT_OFFSET) & 0x02) >> 1;
  if(transaction_id_type == 0x00) { /* Short transaction id - 1 byte*/
    length_field_offset = 4;
  }
  else { /* Long transaction id - 2 bytes*/
    length_field_offset = 5;
  }

  /* compute aligned PER length determinant without calling dissect_per_length_determinant()
     to avoid exceptions and info added to tree, info column and expert info */
  length = tvb_get_guint8(tvb, length_field_offset);
  length_field_offset += 1;
  if (length & 0x80) {
    if ((length & 0xc0) == 0x80) {
      length &= 0x3f;
      length <<= 8;
      length += tvb_get_guint8(tvb, length_field_offset);
      length_field_offset += 1;
    } else {
      length = 0;
    }
  }
  if (length!= (tvb_reported_length(tvb) - length_field_offset)){
    return FALSE;
  }

  dissect_nbap(tvb, pinfo, tree, data);

  return TRUE;
}

/*--- proto_register_nbap -------------------------------------------*/
void proto_register_nbap(void)
{
  module_t *nbap_module;
  guint8 i;

  /* List of fields */
  static hf_register_info hf[] = {
  { &hf_nbap_transportLayerAddress_ipv4,
    { "transportLayerAddress IPv4", "nbap.transportLayerAddress_ipv4",
    FT_IPv4, BASE_NONE, NULL, 0,
  NULL, HFILL }},
  { &hf_nbap_transportLayerAddress_ipv6,
    { "transportLayerAddress IPv6", "nbap.transportLayerAddress_ipv6",
    FT_IPv6, BASE_NONE, NULL, 0,
    NULL, HFILL }},
  { &hf_nbap_transportLayerAddress_nsap,
    { "transportLayerAddress NSAP", "nbap.transportLayerAddress_NSAP",
    FT_BYTES, BASE_NONE, NULL, 0,
    NULL, HFILL }},
  { &hf_nbap_reassembled_information_block,
    { "Reassembled Information Block", "nbap.reassembled_information_block",
    FT_BYTES, BASE_NONE, NULL, 0,
    NULL, HFILL }},
    { &hf_nbap_ActivationInformation_PDU,
      { "ActivationInformation", "nbap.ActivationInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_EDCH_Setup_Info_PDU,
      { "Additional-EDCH-Setup-Info", "nbap.Additional_EDCH_Setup_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_EDCH_Cell_Information_Response_List_PDU,
      { "Additional-EDCH-Cell-Information-Response-List", "nbap.Additional_EDCH_Cell_Information_Response_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_EDCH_Cell_Information_Response_RLReconf_List_PDU,
      { "Additional-EDCH-Cell-Information-Response-RLReconf-List", "nbap.Additional_EDCH_Cell_Information_Response_RLReconf_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_EDCH_Cell_Information_Response_RL_Add_List_PDU,
      { "Additional-EDCH-Cell-Information-Response-RL-Add-List", "nbap.Additional_EDCH_Cell_Information_Response_RL_Add_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_AdditionalMeasurementValueList_PDU,
      { "AdditionalMeasurementValueList", "nbap.AdditionalMeasurementValueList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_AdditionalTimeSlotListLCR_PDU,
      { "AdditionalTimeSlotListLCR", "nbap.AdditionalTimeSlotListLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Active_Pattern_Sequence_Information_PDU,
      { "Active-Pattern-Sequence-Information", "nbap.Active_Pattern_Sequence_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_AlternativeFormatReportingIndicator_PDU,
      { "AlternativeFormatReportingIndicator", "nbap.AlternativeFormatReportingIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_AlternativeFormatReportingIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Angle_Of_Arrival_Value_LCR_PDU,
      { "Angle-Of-Arrival-Value-LCR", "nbap.Angle_Of_Arrival_Value_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_BCCH_ModificationTime_PDU,
      { "BCCH-ModificationTime", "nbap.BCCH_ModificationTime",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Best_Cell_Portions_Value_PDU,
      { "Best-Cell-Portions-Value", "nbap.Best_Cell_Portions_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Best_Cell_Portions_ValueLCR_PDU,
      { "Best-Cell-Portions-ValueLCR", "nbap.Best_Cell_Portions_ValueLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_BindingID_PDU,
      { "BindingID", "nbap.BindingID",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_BlockingPriorityIndicator_PDU,
      { "BlockingPriorityIndicator", "nbap.BlockingPriorityIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_BlockingPriorityIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_BroadcastCommonTransportBearerIndication_PDU,
      { "BroadcastCommonTransportBearerIndication", "nbap.BroadcastCommonTransportBearerIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_BroadcastReference_PDU,
      { "BroadcastReference", "nbap.BroadcastReference",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Cause_PDU,
      { "Cause", "nbap.Cause",
        FT_UINT32, BASE_DEC, VALS(nbap_Cause_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Cell_Capability_Container_PDU,
      { "Cell-Capability-Container", "nbap.Cell_Capability_Container",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Cell_ERNTI_Status_Information_PDU,
      { "Cell-ERNTI-Status-Information", "nbap.Cell_ERNTI_Status_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellParameterID_PDU,
      { "CellParameterID", "nbap.CellParameterID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellPortion_CapabilityLCR_PDU,
      { "CellPortion-CapabilityLCR", "nbap.CellPortion_CapabilityLCR",
        FT_UINT32, BASE_DEC, VALS(nbap_CellPortion_CapabilityLCR_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_CellSyncBurstRepetitionPeriod_PDU,
      { "CellSyncBurstRepetitionPeriod", "nbap.CellSyncBurstRepetitionPeriod",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CFN_PDU,
      { "CFN", "nbap.CFN",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_C_ID_PDU,
      { "C-ID", "nbap.C_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Closedlooptimingadjustmentmode_PDU,
      { "Closedlooptimingadjustmentmode", "nbap.Closedlooptimingadjustmentmode",
        FT_UINT32, BASE_DEC, VALS(nbap_Closedlooptimingadjustmentmode_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Common_EDCH_Capability_PDU,
      { "Common-EDCH-Capability", "nbap.Common_EDCH_Capability",
        FT_UINT32, BASE_DEC, VALS(nbap_Common_EDCH_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Common_E_DCH_HSDPCCH_Capability_PDU,
      { "Common-E-DCH-HSDPCCH-Capability", "nbap.Common_E_DCH_HSDPCCH_Capability",
        FT_UINT32, BASE_DEC, VALS(nbap_Common_E_DCH_HSDPCCH_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Common_EDCH_System_InformationFDD_PDU,
      { "Common-EDCH-System-InformationFDD", "nbap.Common_EDCH_System_InformationFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Common_EDCH_System_Information_ResponseFDD_PDU,
      { "Common-EDCH-System-Information-ResponseFDD", "nbap.Common_EDCH_System_Information_ResponseFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_RNTI_List_PDU,
      { "E-RNTI-List", "nbap.E_RNTI_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Common_MACFlows_to_DeleteFDD_PDU,
      { "Common-MACFlows-to-DeleteFDD", "nbap.Common_MACFlows_to_DeleteFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonMeasurementAccuracy_PDU,
      { "CommonMeasurementAccuracy", "nbap.CommonMeasurementAccuracy",
        FT_UINT32, BASE_DEC, VALS(nbap_CommonMeasurementAccuracy_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_CommonMeasurementType_PDU,
      { "CommonMeasurementType", "nbap.CommonMeasurementType",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_CommonMeasurementType_vals_ext, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonPhysicalChannelID_PDU,
      { "CommonPhysicalChannelID", "nbap.CommonPhysicalChannelID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonPhysicalChannelID768_PDU,
      { "CommonPhysicalChannelID768", "nbap.CommonPhysicalChannelID768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Common_PhysicalChannel_Status_Information_PDU,
      { "Common-PhysicalChannel-Status-Information", "nbap.Common_PhysicalChannel_Status_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Common_PhysicalChannel_Status_Information768_PDU,
      { "Common-PhysicalChannel-Status-Information768", "nbap.Common_PhysicalChannel_Status_Information768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonTransportChannel_InformationResponse_PDU,
      { "CommonTransportChannel-InformationResponse", "nbap.CommonTransportChannel_InformationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Common_TransportChannel_Status_Information_PDU,
      { "Common-TransportChannel-Status-Information", "nbap.Common_TransportChannel_Status_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommunicationControlPortID_PDU,
      { "CommunicationControlPortID", "nbap.CommunicationControlPortID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Compressed_Mode_Deactivation_Flag_PDU,
      { "Compressed-Mode-Deactivation-Flag", "nbap.Compressed_Mode_Deactivation_Flag",
        FT_UINT32, BASE_DEC, VALS(nbap_Compressed_Mode_Deactivation_Flag_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_ConfigurationGenerationID_PDU,
      { "ConfigurationGenerationID", "nbap.ConfigurationGenerationID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ConstantValue_PDU,
      { "ConstantValue", "nbap.ConstantValue",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ContinuousPacketConnectivityDTX_DRX_Capability_PDU,
      { "ContinuousPacketConnectivityDTX-DRX-Capability", "nbap.ContinuousPacketConnectivityDTX_DRX_Capability",
        FT_UINT32, BASE_DEC, VALS(nbap_ContinuousPacketConnectivityDTX_DRX_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_ContinuousPacketConnectivityDTX_DRX_Information_PDU,
      { "ContinuousPacketConnectivityDTX-DRX-Information", "nbap.ContinuousPacketConnectivityDTX_DRX_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ContinuousPacketConnectivityHS_SCCH_less_Capability_PDU,
      { "ContinuousPacketConnectivityHS-SCCH-less-Capability", "nbap.ContinuousPacketConnectivityHS_SCCH_less_Capability",
        FT_UINT32, BASE_DEC, VALS(nbap_ContinuousPacketConnectivityHS_SCCH_less_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_ContinuousPacketConnectivityHS_SCCH_less_Information_PDU,
      { "ContinuousPacketConnectivityHS-SCCH-less-Information", "nbap.ContinuousPacketConnectivityHS_SCCH_less_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ContinuousPacketConnectivityHS_SCCH_less_Information_Response_PDU,
      { "ContinuousPacketConnectivityHS-SCCH-less-Information-Response", "nbap.ContinuousPacketConnectivityHS_SCCH_less_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ControlGAP_PDU,
      { "ControlGAP", "nbap.ControlGAP",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CPC_Information_PDU,
      { "CPC-Information", "nbap.CPC_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ContinuousPacketConnectivityHS_SCCH_less_Deactivate_Indicator_PDU,
      { "ContinuousPacketConnectivityHS-SCCH-less-Deactivate-Indicator", "nbap.ContinuousPacketConnectivityHS_SCCH_less_Deactivate_Indicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CriticalityDiagnostics_PDU,
      { "CriticalityDiagnostics", "nbap.CriticalityDiagnostics_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CRNC_CommunicationContextID_PDU,
      { "CRNC-CommunicationContextID", "nbap.CRNC_CommunicationContextID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CSBMeasurementID_PDU,
      { "CSBMeasurementID", "nbap.CSBMeasurementID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CSBTransmissionID_PDU,
      { "CSBTransmissionID", "nbap.CSBTransmissionID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Common_EDCH_System_InformationLCR_PDU,
      { "Common-EDCH-System-InformationLCR", "nbap.Common_EDCH_System_InformationLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Common_EDCH_System_Information_ResponseLCR_PDU,
      { "Common-EDCH-System-Information-ResponseLCR", "nbap.Common_EDCH_System_Information_ResponseLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_Ext_PDU,
      { "Ul-common-E-DCH-MACflow-Specific-InfoResponseListLCR-Ext", "nbap.Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_Ext",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Common_MACFlows_to_DeleteLCR_PDU,
      { "Common-MACFlows-to-DeleteLCR", "nbap.Common_MACFlows_to_DeleteLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CPC_InformationLCR_PDU,
      { "CPC-InformationLCR", "nbap.CPC_InformationLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ContinuousPacketConnectivity_DRX_CapabilityLCR_PDU,
      { "ContinuousPacketConnectivity-DRX-CapabilityLCR", "nbap.ContinuousPacketConnectivity_DRX_CapabilityLCR",
        FT_UINT32, BASE_DEC, VALS(nbap_ContinuousPacketConnectivity_DRX_CapabilityLCR_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_ContinuousPacketConnectivity_DRX_InformationLCR_PDU,
      { "ContinuousPacketConnectivity-DRX-InformationLCR", "nbap.ContinuousPacketConnectivity_DRX_InformationLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_AGCH_UE_Inactivity_Monitor_Threshold_PDU,
      { "E-AGCH-UE-Inactivity-Monitor-Threshold", "nbap.E_AGCH_UE_Inactivity_Monitor_Threshold",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_E_AGCH_UE_Inactivity_Monitor_Threshold_vals_ext, 0,
        NULL, HFILL }},
    { &hf_nbap_ContinuousPacketConnectivity_DRX_Information_ResponseLCR_PDU,
      { "ContinuousPacketConnectivity-DRX-Information-ResponseLCR", "nbap.ContinuousPacketConnectivity_DRX_Information_ResponseLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DCH_FDD_Information_PDU,
      { "DCH-FDD-Information", "nbap.DCH_FDD_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DCH_Indicator_For_E_DCH_HSDPA_Operation_PDU,
      { "DCH-Indicator-For-E-DCH-HSDPA-Operation", "nbap.DCH_Indicator_For_E_DCH_HSDPA_Operation",
        FT_UINT32, BASE_DEC, VALS(nbap_DCH_Indicator_For_E_DCH_HSDPA_Operation_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_DCH_InformationResponse_PDU,
      { "DCH-InformationResponse", "nbap.DCH_InformationResponse",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DCH_MeasurementOccasion_Information_PDU,
      { "DCH-MeasurementOccasion-Information", "nbap.DCH_MeasurementOccasion_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DCH_TDD_Information_PDU,
      { "DCH-TDD-Information", "nbap.DCH_TDD_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_FDD_DCHs_to_Modify_PDU,
      { "FDD-DCHs-to-Modify", "nbap.FDD_DCHs_to_Modify",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TDD_DCHs_to_Modify_PDU,
      { "TDD-DCHs-to-Modify", "nbap.TDD_DCHs_to_Modify",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DedicatedMeasurementType_PDU,
      { "DedicatedMeasurementType", "nbap.DedicatedMeasurementType",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_DedicatedMeasurementType_vals_ext, 0,
        NULL, HFILL }},
    { &hf_nbap_DelayedActivation_PDU,
      { "DelayedActivation", "nbap.DelayedActivation",
        FT_UINT32, BASE_DEC, VALS(nbap_DelayedActivation_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_DGANSS_Corrections_Req_PDU,
      { "DGANSS-Corrections-Req", "nbap.DGANSS_Corrections_Req_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DGNSS_ValidityPeriod_PDU,
      { "DGNSS-ValidityPeriod", "nbap.DGNSS_ValidityPeriod_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DiversityMode_PDU,
      { "DiversityMode", "nbap.DiversityMode",
        FT_UINT32, BASE_DEC, VALS(nbap_DiversityMode_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_DL_DPCH_TimingAdjustment_PDU,
      { "DL-DPCH-TimingAdjustment", "nbap.DL_DPCH_TimingAdjustment",
        FT_UINT32, BASE_DEC, VALS(nbap_DL_DPCH_TimingAdjustment_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_DL_Power_PDU,
      { "DL-Power", "nbap.DL_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_PowerBalancing_Information_PDU,
      { "DL-PowerBalancing-Information", "nbap.DL_PowerBalancing_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_PowerBalancing_ActivationIndicator_PDU,
      { "DL-PowerBalancing-ActivationIndicator", "nbap.DL_PowerBalancing_ActivationIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_DL_PowerBalancing_ActivationIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_DL_PowerBalancing_UpdatedIndicator_PDU,
      { "DL-PowerBalancing-UpdatedIndicator", "nbap.DL_PowerBalancing_UpdatedIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_DL_PowerBalancing_UpdatedIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_DL_ScramblingCode_PDU,
      { "DL-ScramblingCode", "nbap.DL_ScramblingCode",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_TimeslotISCPInfo_PDU,
      { "DL-TimeslotISCPInfo", "nbap.DL_TimeslotISCPInfo",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_TimeslotISCPInfoLCR_PDU,
      { "DL-TimeslotISCPInfoLCR", "nbap.DL_TimeslotISCPInfoLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_TPC_Pattern01Count_PDU,
      { "DL-TPC-Pattern01Count", "nbap.DL_TPC_Pattern01Count",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DLTransmissionBranchLoadValue_PDU,
      { "DLTransmissionBranchLoadValue", "nbap.DLTransmissionBranchLoadValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MaxHSDSCH_HSSCCH_Power_per_CELLPORTION_PDU,
      { "MaxHSDSCH-HSSCCH-Power-per-CELLPORTION", "nbap.MaxHSDSCH_HSSCCH_Power_per_CELLPORTION",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DPC_Mode_PDU,
      { "DPC-Mode", "nbap.DPC_Mode",
        FT_UINT32, BASE_DEC, VALS(nbap_DPC_Mode_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_DPCH_ID768_PDU,
      { "DPCH-ID768", "nbap.DPCH_ID768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DSCH_InformationResponse_PDU,
      { "DSCH-InformationResponse", "nbap.DSCH_InformationResponse",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DSCH_TDD_Information_PDU,
      { "DSCH-TDD-Information", "nbap.DSCH_TDD_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Dual_Band_Capability_Info_PDU,
      { "Dual-Band-Capability-Info", "nbap.Dual_Band_Capability_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DwPCH_Power_PDU,
      { "DwPCH-Power", "nbap.DwPCH_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_AGCH_Table_Choice_PDU,
      { "E-AGCH-Table-Choice", "nbap.E_AGCH_Table_Choice",
        FT_UINT32, BASE_DEC, VALS(nbap_E_AGCH_Table_Choice_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_E_AGCH_FDD_Code_Information_PDU,
      { "E-AGCH-FDD-Code-Information", "nbap.E_AGCH_FDD_Code_Information",
        FT_UINT32, BASE_DEC, VALS(nbap_E_AGCH_FDD_Code_Information_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_E_AI_Capability_PDU,
      { "E-AI-Capability", "nbap.E_AI_Capability",
        FT_UINT32, BASE_DEC, VALS(nbap_E_AI_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_Capability_PDU,
      { "E-DCH-Capability", "nbap.E_DCH_Capability",
        FT_UINT32, BASE_DEC, VALS(nbap_E_DCH_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCHCapacityConsumptionLaw_PDU,
      { "E-DCHCapacityConsumptionLaw", "nbap.E_DCHCapacityConsumptionLaw_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_TDD_CapacityConsumptionLaw_PDU,
      { "E-DCH-TDD-CapacityConsumptionLaw", "nbap.E_DCH_TDD_CapacityConsumptionLaw_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_TTI2ms_Capability_PDU,
      { "E-DCH-TTI2ms-Capability", "nbap.E_DCH_TTI2ms_Capability",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_SF_Capability_PDU,
      { "E-DCH-SF-Capability", "nbap.E_DCH_SF_Capability",
        FT_UINT32, BASE_DEC, VALS(nbap_E_DCH_SF_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_HARQ_Combining_Capability_PDU,
      { "E-DCH-HARQ-Combining-Capability", "nbap.E_DCH_HARQ_Combining_Capability",
        FT_UINT32, BASE_DEC, VALS(nbap_E_DCH_HARQ_Combining_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_FDD_DL_Control_Channel_Information_PDU,
      { "E-DCH-FDD-DL-Control-Channel-Information", "nbap.E_DCH_FDD_DL_Control_Channel_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_FDD_Information_PDU,
      { "E-DCH-FDD-Information", "nbap.E_DCH_FDD_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_FDD_Information_Response_PDU,
      { "E-DCH-FDD-Information-Response", "nbap.E_DCH_FDD_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_FDD_Information_to_Modify_PDU,
      { "E-DCH-FDD-Information-to-Modify", "nbap.E_DCH_FDD_Information_to_Modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_FDD_Update_Information_PDU,
      { "E-DCH-FDD-Update-Information", "nbap.E_DCH_FDD_Update_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_DL_Control_Channel_Change_Information_PDU,
      { "E-DCH-DL-Control-Channel-Change-Information", "nbap.E_DCH_DL_Control_Channel_Change_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_DL_Control_Channel_Grant_Information_PDU,
      { "E-DCH-DL-Control-Channel-Grant-Information", "nbap.E_DCH_DL_Control_Channel_Grant_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_PowerOffset_for_SchedulingInfo_PDU,
      { "E-DCH-PowerOffset-for-SchedulingInfo", "nbap.E_DCH_PowerOffset_for_SchedulingInfo",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_MACdPDU_SizeCapability_PDU,
      { "E-DCH-MACdPDU-SizeCapability", "nbap.E_DCH_MACdPDU_SizeCapability",
        FT_UINT32, BASE_DEC, VALS(nbap_E_DCH_MACdPDU_SizeCapability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_MACdPDUSizeFormat_PDU,
      { "E-DCH-MACdPDUSizeFormat", "nbap.E_DCH_MACdPDUSizeFormat",
        FT_UINT32, BASE_DEC, VALS(nbap_E_DCH_MACdPDUSizeFormat_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_MACdFlows_Information_PDU,
      { "E-DCH-MACdFlows-Information", "nbap.E_DCH_MACdFlows_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_MACdFlows_to_Delete_PDU,
      { "E-DCH-MACdFlows-to-Delete", "nbap.E_DCH_MACdFlows_to_Delete",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_Non_serving_Relative_Grant_Down_Commands_PDU,
      { "E-DCH-Non-serving-Relative-Grant-Down-Commands", "nbap.E_DCH_Non_serving_Relative_Grant_Down_Commands",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Maximum_Target_ReceivedTotalWideBandPower_PDU,
      { "Maximum-Target-ReceivedTotalWideBandPower", "nbap.Maximum_Target_ReceivedTotalWideBandPower",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Target_NonServing_EDCH_To_Total_EDCH_Power_Ratio_PDU,
      { "Target-NonServing-EDCH-To-Total-EDCH-Power-Ratio", "nbap.Target_NonServing_EDCH_To_Total_EDCH_Power_Ratio",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_RL_Indication_PDU,
      { "E-DCH-RL-Indication", "nbap.E_DCH_RL_Indication",
        FT_UINT32, BASE_DEC, VALS(nbap_E_DCH_RL_Indication_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_Serving_Cell_Change_Info_Response_PDU,
      { "E-DCH-Serving-Cell-Change-Info-Response", "nbap.E_DCH_Serving_Cell_Change_Info_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_EDCH_RACH_Report_Value_PDU,
      { "EDCH-RACH-Report-Value", "nbap.EDCH_RACH_Report_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DPDCH_PowerInterpolation_PDU,
      { "E-DPDCH-PowerInterpolation", "nbap.E_DPDCH_PowerInterpolation",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_MACdFlows_to_DeleteLCR_PDU,
      { "E-DCH-MACdFlows-to-DeleteLCR", "nbap.E_DCH_MACdFlows_to_DeleteLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Enhanced_UE_DRX_InformationLCR_PDU,
      { "Enhanced-UE-DRX-InformationLCR", "nbap.Enhanced_UE_DRX_InformationLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_End_Of_Audit_Sequence_Indicator_PDU,
      { "End-Of-Audit-Sequence-Indicator", "nbap.End_Of_Audit_Sequence_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_End_Of_Audit_Sequence_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_E_RGCH_E_HICH_FDD_Code_Information_PDU,
      { "E-RGCH-E-HICH-FDD-Code-Information", "nbap.E_RGCH_E_HICH_FDD_Code_Information",
        FT_UINT32, BASE_DEC, VALS(nbap_E_RGCH_E_HICH_FDD_Code_Information_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_E_RNTI_PDU,
      { "E-RNTI", "nbap.E_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_TFCI_Boost_Information_PDU,
      { "E-TFCI-Boost-Information", "nbap.E_TFCI_Boost_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCHProvidedBitRate_PDU,
      { "E-DCHProvidedBitRate", "nbap.E_DCHProvidedBitRate",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCHProvidedBitRateValueInformation_For_CellPortion_PDU,
      { "E-DCHProvidedBitRateValueInformation-For-CellPortion", "nbap.E_DCHProvidedBitRateValueInformation_For_CellPortion",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_HICH_TimeOffset_PDU,
      { "E-HICH-TimeOffset", "nbap.E_HICH_TimeOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_HICH_TimeOffsetLCR_PDU,
      { "E-HICH-TimeOffsetLCR", "nbap.E_HICH_TimeOffsetLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_Information_PDU,
      { "E-DCH-Information", "nbap.E_DCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_MACdFlow_Retransmission_Timer_PDU,
      { "E-DCH-MACdFlow-Retransmission-Timer", "nbap.E_DCH_MACdFlow_Retransmission_Timer",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_E_DCH_MACdFlow_Retransmission_Timer_vals_ext, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_Information_Response_PDU,
      { "E-DCH-Information-Response", "nbap.E_DCH_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_Information_Reconfig_PDU,
      { "E-DCH-Information-Reconfig", "nbap.E_DCH_Information_Reconfig_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Maximum_Generated_ReceivedTotalWideBandPowerInOtherCells_PDU,
      { "Maximum-Generated-ReceivedTotalWideBandPowerInOtherCells", "nbap.Maximum_Generated_ReceivedTotalWideBandPowerInOtherCells",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_768_Information_PDU,
      { "E-DCH-768-Information", "nbap.E_DCH_768_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_768_Information_Reconfig_PDU,
      { "E-DCH-768-Information-Reconfig", "nbap.E_DCH_768_Information_Reconfig_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_LCR_Information_PDU,
      { "E-DCH-LCR-Information", "nbap.E_DCH_LCR_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_LCR_Information_Reconfig_PDU,
      { "E-DCH-LCR-Information-Reconfig", "nbap.E_DCH_LCR_Information_Reconfig_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DormantModeIndicator_PDU,
      { "DormantModeIndicator", "nbap.DormantModeIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_DormantModeIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Enhanced_FACH_Capability_PDU,
      { "Enhanced-FACH-Capability", "nbap.Enhanced_FACH_Capability",
        FT_UINT32, BASE_DEC, VALS(nbap_Enhanced_FACH_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_EnhancedHSServingCC_Abort_PDU,
      { "EnhancedHSServingCC-Abort", "nbap.EnhancedHSServingCC_Abort",
        FT_UINT32, BASE_DEC, VALS(nbap_EnhancedHSServingCC_Abort_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Enhanced_PCH_Capability_PDU,
      { "Enhanced-PCH-Capability", "nbap.Enhanced_PCH_Capability",
        FT_UINT32, BASE_DEC, VALS(nbap_Enhanced_PCH_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Enhanced_UE_DRX_Capability_PDU,
      { "Enhanced-UE-DRX-Capability", "nbap.Enhanced_UE_DRX_Capability",
        FT_UINT32, BASE_DEC, VALS(nbap_Enhanced_UE_DRX_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Enhanced_UE_DRX_InformationFDD_PDU,
      { "Enhanced-UE-DRX-InformationFDD", "nbap.Enhanced_UE_DRX_InformationFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Extended_E_DCH_LCRTDD_PhysicalLayerCategory_PDU,
      { "Extended-E-DCH-LCRTDD-PhysicalLayerCategory", "nbap.Extended_E_DCH_LCRTDD_PhysicalLayerCategory",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Ext_Max_Bits_MACe_PDU_non_scheduled_PDU,
      { "Ext-Max-Bits-MACe-PDU-non-scheduled", "nbap.Ext_Max_Bits_MACe_PDU_non_scheduled",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Ext_Reference_E_TFCI_PO_PDU,
      { "Ext-Reference-E-TFCI-PO", "nbap.Ext_Reference_E_TFCI_PO",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ExtendedPropagationDelay_PDU,
      { "ExtendedPropagationDelay", "nbap.ExtendedPropagationDelay",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Extended_RNC_ID_PDU,
      { "Extended-RNC-ID", "nbap.Extended_RNC_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Extended_Round_Trip_Time_Value_PDU,
      { "Extended-Round-Trip-Time-Value", "nbap.Extended_Round_Trip_Time_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Extended_HS_SCCH_ID_PDU,
      { "Extended-HS-SCCH-ID", "nbap.Extended_HS_SCCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Extended_HS_SICH_ID_PDU,
      { "Extended-HS-SICH-ID", "nbap.Extended_HS_SICH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Extended_E_HICH_ID_TDD_PDU,
      { "Extended-E-HICH-ID-TDD", "nbap.Extended_E_HICH_ID_TDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_Semi_PersistentScheduling_Information_LCR_PDU,
      { "E-DCH-Semi-PersistentScheduling-Information-LCR", "nbap.E_DCH_Semi_PersistentScheduling_Information_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR_PDU,
      { "E-DCH-Semi-PersistentScheduling-Information-ResponseLCR", "nbap.E_DCH_Semi_PersistentScheduling_Information_ResponseLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Fast_Reconfiguration_Mode_PDU,
      { "Fast-Reconfiguration-Mode", "nbap.Fast_Reconfiguration_Mode",
        FT_UINT32, BASE_DEC, VALS(nbap_Fast_Reconfiguration_Mode_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Fast_Reconfiguration_Permission_PDU,
      { "Fast-Reconfiguration-Permission", "nbap.Fast_Reconfiguration_Permission",
        FT_UINT32, BASE_DEC, VALS(nbap_Fast_Reconfiguration_Permission_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_FDD_S_CCPCH_FrameOffset_PDU,
      { "FDD-S-CCPCH-FrameOffset", "nbap.FDD_S_CCPCH_FrameOffset",
        FT_UINT32, BASE_DEC, VALS(nbap_FDD_S_CCPCH_FrameOffset_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_F_DPCH_Capability_PDU,
      { "F-DPCH-Capability", "nbap.F_DPCH_Capability",
        FT_UINT32, BASE_DEC, VALS(nbap_F_DPCH_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_F_DPCH_SlotFormat_PDU,
      { "F-DPCH-SlotFormat", "nbap.F_DPCH_SlotFormat",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_F_DPCH_SlotFormatCapability_PDU,
      { "F-DPCH-SlotFormatCapability", "nbap.F_DPCH_SlotFormatCapability",
        FT_UINT32, BASE_DEC, VALS(nbap_F_DPCH_SlotFormatCapability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_FNReportingIndicator_PDU,
      { "FNReportingIndicator", "nbap.FNReportingIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_FNReportingIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_GANSS_AddIonoModelReq_PDU,
      { "GANSS-AddIonoModelReq", "nbap.GANSS_AddIonoModelReq",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_GANSS_AddNavigationModelsReq_PDU,
      { "GANSS-AddNavigationModelsReq", "nbap.GANSS_AddNavigationModelsReq",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_GANSS_AddUTCModelsReq_PDU,
      { "GANSS-AddUTCModelsReq", "nbap.GANSS_AddUTCModelsReq",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_GANSS_Additional_Ionospheric_Model_PDU,
      { "GANSS-Additional-Ionospheric-Model", "nbap.GANSS_Additional_Ionospheric_Model_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_GANSS_Additional_Navigation_Models_PDU,
      { "GANSS-Additional-Navigation-Models", "nbap.GANSS_Additional_Navigation_Models_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_GANSS_Additional_Time_Models_PDU,
      { "GANSS-Additional-Time-Models", "nbap.GANSS_Additional_Time_Models",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_GANSS_Additional_UTC_Models_PDU,
      { "GANSS-Additional-UTC-Models", "nbap.GANSS_Additional_UTC_Models",
        FT_UINT32, BASE_DEC, VALS(nbap_GANSS_Additional_UTC_Models_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_GANSS_ALM_ECEFsbasAlmanacSet_PDU,
      { "GANSS-ALM-ECEFsbasAlmanacSet", "nbap.GANSS_ALM_ECEFsbasAlmanacSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_GANSS_ALM_GlonassAlmanacSet_PDU,
      { "GANSS-ALM-GlonassAlmanacSet", "nbap.GANSS_ALM_GlonassAlmanacSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_GANSS_ALM_MidiAlmanacSet_PDU,
      { "GANSS-ALM-MidiAlmanacSet", "nbap.GANSS_ALM_MidiAlmanacSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_GANSS_ALM_NAVKeplerianSet_PDU,
      { "GANSS-ALM-NAVKeplerianSet", "nbap.GANSS_ALM_NAVKeplerianSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_GANSS_ALM_ReducedKeplerianSet_PDU,
      { "GANSS-ALM-ReducedKeplerianSet", "nbap.GANSS_ALM_ReducedKeplerianSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_GANSS_Auxiliary_Information_PDU,
      { "GANSS-Auxiliary-Information", "nbap.GANSS_Auxiliary_Information",
        FT_UINT32, BASE_DEC, VALS(nbap_GANSS_Auxiliary_Information_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_GANSS_AuxInfoReq_PDU,
      { "GANSS-AuxInfoReq", "nbap.GANSS_AuxInfoReq",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_GANSS_Common_Data_PDU,
      { "GANSS-Common-Data", "nbap.GANSS_Common_Data_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_GANSS_Earth_Orientation_Parameters_PDU,
      { "GANSS-Earth-Orientation-Parameters", "nbap.GANSS_Earth_Orientation_Parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_GANSS_EarthOrientParaReq_PDU,
      { "GANSS-EarthOrientParaReq", "nbap.GANSS_EarthOrientParaReq",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_GANSS_Generic_Data_PDU,
      { "GANSS-Generic-Data", "nbap.GANSS_Generic_Data",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_GANSS_ID_PDU,
      { "GANSS-ID", "nbap.GANSS_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_GANSS_Information_PDU,
      { "GANSS-Information", "nbap.GANSS_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_GANSS_SBAS_ID_PDU,
      { "GANSS-SBAS-ID", "nbap.GANSS_SBAS_ID",
        FT_UINT32, BASE_DEC, VALS(nbap_GANSS_SBAS_ID_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_GANSS_Time_ID_PDU,
      { "GANSS-Time-ID", "nbap.GANSS_Time_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HARQ_MemoryPartitioningInfoExtForMIMO_PDU,
      { "HARQ-MemoryPartitioningInfoExtForMIMO", "nbap.HARQ_MemoryPartitioningInfoExtForMIMO",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HARQ_Preamble_Mode_PDU,
      { "HARQ-Preamble-Mode", "nbap.HARQ_Preamble_Mode",
        FT_UINT32, BASE_DEC, VALS(nbap_HARQ_Preamble_Mode_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_HARQ_Preamble_Mode_Activation_Indicator_PDU,
      { "HARQ-Preamble-Mode-Activation-Indicator", "nbap.HARQ_Preamble_Mode_Activation_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_HARQ_Preamble_Mode_Activation_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_HSDPA_Capability_PDU,
      { "HSDPA-Capability", "nbap.HSDPA_Capability",
        FT_UINT32, BASE_DEC, VALS(nbap_HSDPA_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_HS_DSCHProvidedBitRate_PDU,
      { "HS-DSCHProvidedBitRate", "nbap.HS_DSCHProvidedBitRate",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortion_PDU,
      { "HS-DSCHProvidedBitRateValueInformation-For-CellPortion", "nbap.HS_DSCHProvidedBitRateValueInformation_For_CellPortion",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR_PDU,
      { "HS-DSCHProvidedBitRateValueInformation-For-CellPortionLCR", "nbap.HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_DSCHRequiredPower_PDU,
      { "HS-DSCHRequiredPower", "nbap.HS_DSCHRequiredPower",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_DSCHRequiredPowerValue_PDU,
      { "HS-DSCHRequiredPowerValue", "nbap.HS_DSCHRequiredPowerValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortion_PDU,
      { "HS-DSCHRequiredPowerValueInformation-For-CellPortion", "nbap.HS_DSCHRequiredPowerValueInformation_For_CellPortion",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR_PDU,
      { "HS-DSCHRequiredPowerValueInformation-For-CellPortionLCR", "nbap.HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_Common_System_InformationFDD_PDU,
      { "HSDSCH-Common-System-InformationFDD", "nbap.HSDSCH_Common_System_InformationFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_Common_System_Information_ResponseFDD_PDU,
      { "HSDSCH-Common-System-Information-ResponseFDD", "nbap.HSDSCH_Common_System_Information_ResponseFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_FDD_Information_PDU,
      { "HSDSCH-FDD-Information", "nbap.HSDSCH_FDD_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_TDD_Information_PDU,
      { "HSDSCH-TDD-Information", "nbap.HSDSCH_TDD_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_Information_to_Modify_PDU,
      { "HSDSCH-Information-to-Modify", "nbap.HSDSCH_Information_to_Modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_MACdPDUSizeFormat_PDU,
      { "HSDSCH-MACdPDUSizeFormat", "nbap.HSDSCH_MACdPDUSizeFormat",
        FT_UINT32, BASE_DEC, VALS(nbap_HSDSCH_MACdPDUSizeFormat_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_MACdPDU_SizeCapability_PDU,
      { "HSDSCH-MACdPDU-SizeCapability", "nbap.HSDSCH_MACdPDU_SizeCapability",
        FT_UINT32, BASE_DEC, VALS(nbap_HSDSCH_MACdPDU_SizeCapability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_Information_to_Modify_Unsynchronised_PDU,
      { "HSDSCH-Information-to-Modify-Unsynchronised", "nbap.HSDSCH_Information_to_Modify_Unsynchronised_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_FDD_Information_Response_PDU,
      { "HSDSCH-FDD-Information-Response", "nbap.HSDSCH_FDD_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_Paging_System_InformationFDD_PDU,
      { "HSDSCH-Paging-System-InformationFDD", "nbap.HSDSCH_Paging_System_InformationFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_Paging_System_Information_ResponseFDD_PDU,
      { "HSDSCH-Paging-System-Information-ResponseFDD", "nbap.HSDSCH_Paging_System_Information_ResponseFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_TDD_Information_Response_PDU,
      { "HSDSCH-TDD-Information-Response", "nbap.HSDSCH_TDD_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_MACdFlows_Information_PDU,
      { "HSDSCH-MACdFlows-Information", "nbap.HSDSCH_MACdFlows_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_MACdFlows_to_Delete_PDU,
      { "HSDSCH-MACdFlows-to-Delete", "nbap.HSDSCH_MACdFlows_to_Delete",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_TBSizeTableIndicator_PDU,
      { "HSDSCH-TBSizeTableIndicator", "nbap.HSDSCH_TBSizeTableIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_HSDSCH_TBSizeTableIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_PreconfigurationInfo_PDU,
      { "HSDSCH-PreconfigurationInfo", "nbap.HSDSCH_PreconfigurationInfo_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_EDCH_Preconfiguration_Information_PDU,
      { "Additional-EDCH-Preconfiguration-Information", "nbap.Additional_EDCH_Preconfiguration_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_PreconfigurationSetup_PDU,
      { "HSDSCH-PreconfigurationSetup", "nbap.HSDSCH_PreconfigurationSetup_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSSCCH_Specific_InformationRespListTDD768_PDU,
      { "HSSCCH-Specific-InformationRespListTDD768", "nbap.HSSCCH_Specific_InformationRespListTDD768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_SICH_Reception_Quality_Value_PDU,
      { "HS-SICH-Reception-Quality-Value", "nbap.HS_SICH_Reception_Quality_Value_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_SICH_failed_PDU,
      { "HS-SICH-failed", "nbap.HS_SICH_failed",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_SICH_missed_PDU,
      { "HS-SICH-missed", "nbap.HS_SICH_missed",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_SICH_total_PDU,
      { "HS-SICH-total", "nbap.HS_SICH_total",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_SICH_Reception_Quality_Measurement_Value_PDU,
      { "HS-SICH-Reception-Quality-Measurement-Value", "nbap.HS_SICH_Reception_Quality_Measurement_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_RNTI_PDU,
      { "HSDSCH-RNTI", "nbap.HSDSCH_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_PDSCH_FDD_Code_Information_PDU,
      { "HS-PDSCH-FDD-Code-Information", "nbap.HS_PDSCH_FDD_Code_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_SICH_ID_PDU,
      { "HS-SICH-ID", "nbap.HS_SICH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_SCCH_FDD_Code_Information_PDU,
      { "HS-SCCH-FDD-Code-Information", "nbap.HS_SCCH_FDD_Code_Information",
        FT_UINT32, BASE_DEC, VALS(nbap_HS_SCCH_FDD_Code_Information_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_HS_PDSCH_Code_Change_Indicator_PDU,
      { "HS-PDSCH-Code-Change-Indicator", "nbap.HS_PDSCH_Code_Change_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_HS_PDSCH_Code_Change_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_HS_PDSCH_Code_Change_Grant_PDU,
      { "HS-PDSCH-Code-Change-Grant", "nbap.HS_PDSCH_Code_Change_Grant",
        FT_UINT32, BASE_DEC, VALS(nbap_HS_PDSCH_Code_Change_Grant_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_Configured_Indicator_PDU,
      { "HSDSCH-Configured-Indicator", "nbap.HSDSCH_Configured_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_HSDSCH_Configured_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_HS_DSCH_Serving_Cell_Change_Info_PDU,
      { "HS-DSCH-Serving-Cell-Change-Info", "nbap.HS_DSCH_Serving_Cell_Change_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_DSCH_Serving_Cell_Change_Info_Response_PDU,
      { "HS-DSCH-Serving-Cell-Change-Info-Response", "nbap.HS_DSCH_Serving_Cell_Change_Info_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_FDD_Update_Information_PDU,
      { "HSDSCH-FDD-Update-Information", "nbap.HSDSCH_FDD_Update_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_TDD_Update_Information_PDU,
      { "HSDSCH-TDD-Update-Information", "nbap.HSDSCH_TDD_Update_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_Common_System_InformationLCR_PDU,
      { "HSDSCH-Common-System-InformationLCR", "nbap.HSDSCH_Common_System_InformationLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_Common_System_Information_ResponseLCR_PDU,
      { "HSDSCH-Common-System-Information-ResponseLCR", "nbap.HSDSCH_Common_System_Information_ResponseLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_Paging_System_InformationLCR_PDU,
      { "HSDSCH-Paging-System-InformationLCR", "nbap.HSDSCH_Paging_System_InformationLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_Paging_System_Information_ResponseLCR_PDU,
      { "HSDSCH-Paging-System-Information-ResponseLCR", "nbap.HSDSCH_Paging_System_Information_ResponseLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_DSCH_Semi_PersistentScheduling_Information_LCR_PDU,
      { "HS-DSCH-Semi-PersistentScheduling-Information-LCR", "nbap.HS_DSCH_Semi_PersistentScheduling_Information_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RepetitionPeriodIndex_PDU,
      { "RepetitionPeriodIndex", "nbap.RepetitionPeriodIndex",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SPS_Reservation_Indicator_PDU,
      { "SPS-Reservation-Indicator", "nbap.SPS_Reservation_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_SPS_Reservation_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_HS_DSCH_SPS_Operation_Indicator_PDU,
      { "HS-DSCH-SPS-Operation-Indicator", "nbap.HS_DSCH_SPS_Operation_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_HS_DSCH_SPS_Operation_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR_PDU,
      { "HS-DSCH-Semi-PersistentScheduling-Information-ResponseLCR", "nbap.HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Add_To_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU,
      { "Add-To-Non-HS-SCCH-Associated-HS-SICH-Resource-Pool-LCR-PSCH-ReconfRqst", "nbap.Add_To_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext_PDU,
      { "Non-HS-SCCH-Associated-HS-SICH-InformationList-Ext", "nbap.Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU,
      { "Modify-Non-HS-SCCH-Associated-HS-SICH-Resource-Pool-LCR-PSCH-ReconfRqst", "nbap.Modify_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext_PDU,
      { "Modify-Non-HS-SCCH-Associated-HS-SICH-InformationList-Ext", "nbap.Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU,
      { "Delete-From-Non-HS-SCCH-Associated-HS-SICH-Resource-Pool-LCR-PSCH-ReconfRqst", "nbap.Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext_PDU,
      { "Delete-From-Non-HS-SCCH-Associated-HS-SICH-Resource-Pool-LCR-PSCH-ReconfRqst-Ext", "nbap.Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MIMO_ReferenceSignal_InformationListLCR_PDU,
      { "MIMO-ReferenceSignal-InformationListLCR", "nbap.MIMO_ReferenceSignal_InformationListLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSSICH_ReferenceSignal_InformationLCR_PDU,
      { "HSSICH-ReferenceSignal-InformationLCR", "nbap.HSSICH_ReferenceSignal_InformationLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSSICH_ReferenceSignal_InformationModifyLCR_PDU,
      { "HSSICH-ReferenceSignal-InformationModifyLCR", "nbap.HSSICH_ReferenceSignal_InformationModifyLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_IMB_Parameters_PDU,
      { "IMB-Parameters", "nbap.IMB_Parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_InformationReportCharacteristics_PDU,
      { "InformationReportCharacteristics", "nbap.InformationReportCharacteristics",
        FT_UINT32, BASE_DEC, VALS(nbap_InformationReportCharacteristics_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_InformationExchangeID_PDU,
      { "InformationExchangeID", "nbap.InformationExchangeID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_InformationType_PDU,
      { "InformationType", "nbap.InformationType_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Initial_DL_DPCH_TimingAdjustment_Allowed_PDU,
      { "Initial-DL-DPCH-TimingAdjustment-Allowed", "nbap.Initial_DL_DPCH_TimingAdjustment_Allowed",
        FT_UINT32, BASE_DEC, VALS(nbap_Initial_DL_DPCH_TimingAdjustment_Allowed_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_InnerLoopDLPCStatus_PDU,
      { "InnerLoopDLPCStatus", "nbap.InnerLoopDLPCStatus",
        FT_UINT32, BASE_DEC, VALS(nbap_InnerLoopDLPCStatus_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_IPMulticastIndication_PDU,
      { "IPMulticastIndication", "nbap.IPMulticastIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_IPMulticastDataBearerIndication_PDU,
      { "IPMulticastDataBearerIndication", "nbap.IPMulticastDataBearerIndication",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_IdleIntervalInformation_PDU,
      { "IdleIntervalInformation", "nbap.IdleIntervalInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Local_Cell_ID_PDU,
      { "Local-Cell-ID", "nbap.Local_Cell_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_LCRTDD_Uplink_Physical_Channel_Capability_PDU,
      { "LCRTDD-Uplink-Physical-Channel-Capability", "nbap.LCRTDD_Uplink_Physical_Channel_Capability_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MAC_PDU_SizeExtended_PDU,
      { "MAC-PDU-SizeExtended", "nbap.MAC_PDU_SizeExtended",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MACes_Maximum_Bitrate_LCR_PDU,
      { "MACes-Maximum-Bitrate-LCR", "nbap.MACes_Maximum_Bitrate_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Maximum_Number_of_Retransmissions_For_E_DCH_PDU,
      { "Maximum-Number-of-Retransmissions-For-E-DCH", "nbap.Maximum_Number_of_Retransmissions_For_E_DCH",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Maximum_Target_ReceivedTotalWideBandPower_LCR_PDU,
      { "Maximum-Target-ReceivedTotalWideBandPower-LCR", "nbap.Maximum_Target_ReceivedTotalWideBandPower_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MaximumTransmissionPower_PDU,
      { "MaximumTransmissionPower", "nbap.MaximumTransmissionPower",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Max_UE_DTX_Cycle_PDU,
      { "Max-UE-DTX-Cycle", "nbap.Max_UE_DTX_Cycle",
        FT_UINT32, BASE_DEC, VALS(nbap_Max_UE_DTX_Cycle_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_MBMS_Capability_PDU,
      { "MBMS-Capability", "nbap.MBMS_Capability",
        FT_UINT32, BASE_DEC, VALS(nbap_MBMS_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_MeasurementFilterCoefficient_PDU,
      { "MeasurementFilterCoefficient", "nbap.MeasurementFilterCoefficient",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_MeasurementFilterCoefficient_vals_ext, 0,
        NULL, HFILL }},
    { &hf_nbap_MeasurementID_PDU,
      { "MeasurementID", "nbap.MeasurementID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MeasurementRecoveryBehavior_PDU,
      { "MeasurementRecoveryBehavior", "nbap.MeasurementRecoveryBehavior_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MeasurementRecoveryReportingIndicator_PDU,
      { "MeasurementRecoveryReportingIndicator", "nbap.MeasurementRecoveryReportingIndicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MeasurementRecoverySupportIndicator_PDU,
      { "MeasurementRecoverySupportIndicator", "nbap.MeasurementRecoverySupportIndicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MessageStructure_PDU,
      { "MessageStructure", "nbap.MessageStructure",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MICH_CFN_PDU,
      { "MICH-CFN", "nbap.MICH_CFN",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MidambleShiftLCR_PDU,
      { "MidambleShiftLCR", "nbap.MidambleShiftLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MIMO_ActivationIndicator_PDU,
      { "MIMO-ActivationIndicator", "nbap.MIMO_ActivationIndicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MIMO_Capability_PDU,
      { "MIMO-Capability", "nbap.MIMO_Capability",
        FT_UINT32, BASE_DEC, VALS(nbap_MIMO_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_MIMO_Mode_Indicator_PDU,
      { "MIMO-Mode-Indicator", "nbap.MIMO_Mode_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_MIMO_Mode_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_MIMO_N_M_Ratio_PDU,
      { "MIMO-N-M-Ratio", "nbap.MIMO_N_M_Ratio",
        FT_UINT32, BASE_DEC, VALS(nbap_MIMO_N_M_Ratio_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_MIMO_PilotConfiguration_PDU,
      { "MIMO-PilotConfiguration", "nbap.MIMO_PilotConfiguration",
        FT_UINT32, BASE_DEC, VALS(nbap_MIMO_PilotConfiguration_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_MIMO_PilotConfigurationExtension_PDU,
      { "MIMO-PilotConfigurationExtension", "nbap.MIMO_PilotConfigurationExtension",
        FT_UINT32, BASE_DEC, VALS(nbap_MIMO_PilotConfigurationExtension_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_MIMO_PowerOffsetForS_CPICHCapability_PDU,
      { "MIMO-PowerOffsetForS-CPICHCapability", "nbap.MIMO_PowerOffsetForS_CPICHCapability",
        FT_UINT32, BASE_DEC, VALS(nbap_MIMO_PowerOffsetForS_CPICHCapability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_MinimumReducedE_DPDCH_GainFactor_PDU,
      { "MinimumReducedE-DPDCH-GainFactor", "nbap.MinimumReducedE_DPDCH_GainFactor",
        FT_UINT32, BASE_DEC, VALS(nbap_MinimumReducedE_DPDCH_GainFactor_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Modification_Period_PDU,
      { "Modification-Period", "nbap.Modification_Period",
        FT_UINT32, BASE_DEC, VALS(nbap_Modification_Period_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_MAChs_ResetIndicator_PDU,
      { "MAChs-ResetIndicator", "nbap.MAChs_ResetIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_MAChs_ResetIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_ModulationMBSFN_PDU,
      { "ModulationMBSFN", "nbap.ModulationMBSFN",
        FT_UINT32, BASE_DEC, VALS(nbap_ModulationMBSFN_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_ModulationPO_MBSFN_PDU,
      { "ModulationPO-MBSFN", "nbap.ModulationPO_MBSFN",
        FT_UINT32, BASE_DEC, VALS(nbap_ModulationPO_MBSFN_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_MBSFN_Only_Mode_Indicator_PDU,
      { "MBSFN-Only-Mode-Indicator", "nbap.MBSFN_Only_Mode_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_MBSFN_Only_Mode_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_MBSFN_Only_Mode_Capability_PDU,
      { "MBSFN-Only-Mode-Capability", "nbap.MBSFN_Only_Mode_Capability",
        FT_UINT32, BASE_DEC, VALS(nbap_MBSFN_Only_Mode_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Multicarrier_Number_PDU,
      { "Multicarrier-Number", "nbap.Multicarrier_Number",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleFreq_HARQ_MemoryPartitioning_InformationList_PDU,
      { "MultipleFreq-HARQ-MemoryPartitioning-InformationList", "nbap.MultipleFreq_HARQ_MemoryPartitioning_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR_PDU,
      { "MultipleFreq-HSPDSCH-InformationList-ResponseTDDLCR", "nbap.MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Multi_Cell_Capability_Info_PDU,
      { "Multi-Cell-Capability-Info", "nbap.Multi_Cell_Capability_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Multicell_EDCH_InformationItemIEs_PDU,
      { "Multicell-EDCH-InformationItemIEs", "nbap.Multicell_EDCH_InformationItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Multicell_EDCH_RL_Specific_InformationItemIEs_PDU,
      { "Multicell-EDCH-RL-Specific-InformationItemIEs", "nbap.Multicell_EDCH_RL_Specific_InformationItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MIMO_SFMode_For_HSPDSCHDualStream_PDU,
      { "MIMO-SFMode-For-HSPDSCHDualStream", "nbap.MIMO_SFMode_For_HSPDSCHDualStream",
        FT_UINT32, BASE_DEC, VALS(nbap_MIMO_SFMode_For_HSPDSCHDualStream_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_NCyclesPerSFNperiod_PDU,
      { "NCyclesPerSFNperiod", "nbap.NCyclesPerSFNperiod",
        FT_UINT32, BASE_DEC, VALS(nbap_NCyclesPerSFNperiod_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_NRepetitionsPerCyclePeriod_PDU,
      { "NRepetitionsPerCyclePeriod", "nbap.NRepetitionsPerCyclePeriod",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_NeighbouringCellMeasurementInformation_PDU,
      { "NeighbouringCellMeasurementInformation", "nbap.NeighbouringCellMeasurementInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_NeighbouringTDDCellMeasurementInformationLCR_PDU,
      { "NeighbouringTDDCellMeasurementInformationLCR", "nbap.NeighbouringTDDCellMeasurementInformationLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_NeighbouringTDDCellMeasurementInformation768_PDU,
      { "NeighbouringTDDCellMeasurementInformation768", "nbap.NeighbouringTDDCellMeasurementInformation768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_NonCellSpecificTxDiversity_PDU,
      { "NonCellSpecificTxDiversity", "nbap.NonCellSpecificTxDiversity",
        FT_UINT32, BASE_DEC, VALS(nbap_NonCellSpecificTxDiversity_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_NI_Information_PDU,
      { "NI-Information", "nbap.NI_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_NodeB_CommunicationContextID_PDU,
      { "NodeB-CommunicationContextID", "nbap.NodeB_CommunicationContextID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_NumberOfReportedCellPortions_PDU,
      { "NumberOfReportedCellPortions", "nbap.NumberOfReportedCellPortions",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_NumberOfReportedCellPortionsLCR_PDU,
      { "NumberOfReportedCellPortionsLCR", "nbap.NumberOfReportedCellPortionsLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_NSubCyclesPerCyclePeriod_PDU,
      { "NSubCyclesPerCyclePeriod", "nbap.NSubCyclesPerCyclePeriod",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Number_Of_Supported_Carriers_PDU,
      { "Number-Of-Supported-Carriers", "nbap.Number_Of_Supported_Carriers",
        FT_UINT32, BASE_DEC, VALS(nbap_Number_Of_Supported_Carriers_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_NoOfTargetCellHS_SCCH_Order_PDU,
      { "NoOfTargetCellHS-SCCH-Order", "nbap.NoOfTargetCellHS_SCCH_Order",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Out_of_Sychronization_Window_PDU,
      { "Out-of-Sychronization-Window", "nbap.Out_of_Sychronization_Window",
        FT_UINT32, BASE_DEC, VALS(nbap_Out_of_Sychronization_Window_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Paging_MACFlows_to_DeleteFDD_PDU,
      { "Paging-MACFlows-to-DeleteFDD", "nbap.Paging_MACFlows_to_DeleteFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PhysicalChannelID_for_CommonERNTI_RequestedIndicator_PDU,
      { "PhysicalChannelID-for-CommonERNTI-RequestedIndicator", "nbap.PhysicalChannelID_for_CommonERNTI_RequestedIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_PhysicalChannelID_for_CommonERNTI_RequestedIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_PLCCHinformation_PDU,
      { "PLCCHinformation", "nbap.PLCCHinformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PowerAdjustmentType_PDU,
      { "PowerAdjustmentType", "nbap.PowerAdjustmentType",
        FT_UINT32, BASE_DEC, VALS(nbap_PowerAdjustmentType_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_PrecodingWeightSetRestriction_PDU,
      { "PrecodingWeightSetRestriction", "nbap.PrecodingWeightSetRestriction",
        FT_UINT32, BASE_DEC, VALS(nbap_PrecodingWeightSetRestriction_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Primary_CPICH_Usage_for_Channel_Estimation_PDU,
      { "Primary-CPICH-Usage-for-Channel-Estimation", "nbap.Primary_CPICH_Usage_for_Channel_Estimation",
        FT_UINT32, BASE_DEC, VALS(nbap_Primary_CPICH_Usage_for_Channel_Estimation_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_PrimaryScramblingCode_PDU,
      { "PrimaryScramblingCode", "nbap.PrimaryScramblingCode",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PrimaryCCPCH_RSCP_PDU,
      { "PrimaryCCPCH-RSCP", "nbap.PrimaryCCPCH_RSCP",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PrimaryCCPCH_RSCP_Delta_PDU,
      { "PrimaryCCPCH-RSCP-Delta", "nbap.PrimaryCCPCH_RSCP_Delta",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UE_Selected_MBMS_Service_Information_PDU,
      { "UE-Selected-MBMS-Service-Information", "nbap.UE_Selected_MBMS_Service_Information",
        FT_UINT32, BASE_DEC, VALS(nbap_UE_Selected_MBMS_Service_Information_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Paging_MACFlows_to_DeleteLCR_PDU,
      { "Paging-MACFlows-to-DeleteLCR", "nbap.Paging_MACFlows_to_DeleteLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_Specific_DCH_Info_PDU,
      { "RL-Specific-DCH-Info", "nbap.RL_Specific_DCH_Info",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_Specific_E_DCH_Info_PDU,
      { "RL-Specific-E-DCH-Info", "nbap.RL_Specific_E_DCH_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Reference_ReceivedTotalWideBandPower_PDU,
      { "Reference-ReceivedTotalWideBandPower", "nbap.Reference_ReceivedTotalWideBandPower",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Reference_ReceivedTotalWideBandPowerReporting_PDU,
      { "Reference-ReceivedTotalWideBandPowerReporting", "nbap.Reference_ReceivedTotalWideBandPowerReporting",
        FT_UINT32, BASE_DEC, VALS(nbap_Reference_ReceivedTotalWideBandPowerReporting_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Reference_ReceivedTotalWideBandPowerSupportIndicator_PDU,
      { "Reference-ReceivedTotalWideBandPowerSupportIndicator", "nbap.Reference_ReceivedTotalWideBandPowerSupportIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_Reference_ReceivedTotalWideBandPowerSupportIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_ReferenceClockAvailability_PDU,
      { "ReferenceClockAvailability", "nbap.ReferenceClockAvailability",
        FT_UINT32, BASE_DEC, VALS(nbap_ReferenceClockAvailability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_ReferenceSFNoffset_PDU,
      { "ReferenceSFNoffset", "nbap.ReferenceSFNoffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ReportCharacteristics_PDU,
      { "ReportCharacteristics", "nbap.ReportCharacteristics",
        FT_UINT32, BASE_DEC, VALS(nbap_ReportCharacteristics_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_ReportCharacteristicsType_OnModification_PDU,
      { "ReportCharacteristicsType-OnModification", "nbap.ReportCharacteristicsType_OnModification_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_EDCH_RACH_Report_IncrDecrThres_PDU,
      { "EDCH-RACH-Report-IncrDecrThres", "nbap.EDCH_RACH_Report_IncrDecrThres_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_EDCH_RACH_Report_ThresholdInformation_PDU,
      { "EDCH-RACH-Report-ThresholdInformation", "nbap.EDCH_RACH_Report_ThresholdInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_ID_PDU,
      { "RL-ID", "nbap.RL_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_Set_ID_PDU,
      { "RL-Set-ID", "nbap.RL_Set_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_RLC_PDU_Size_Format_PDU,
      { "DL-RLC-PDU-Size-Format", "nbap.DL_RLC_PDU_Size_Format",
        FT_UINT32, BASE_DEC, VALS(nbap_DL_RLC_PDU_Size_Format_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Received_total_wide_band_power_For_CellPortion_Value_PDU,
      { "Received-total-wide-band-power-For-CellPortion-Value", "nbap.Received_total_wide_band_power_For_CellPortion_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Received_total_wide_band_power_For_CellPortion_ValueLCR_PDU,
      { "Received-total-wide-band-power-For-CellPortion-ValueLCR", "nbap.Received_total_wide_band_power_For_CellPortion_ValueLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Received_total_wide_band_power_Value_IncrDecrThres_PDU,
      { "Received-total-wide-band-power-Value-IncrDecrThres", "nbap.Received_total_wide_band_power_Value_IncrDecrThres",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value_PDU,
      { "Received-Scheduled-EDCH-Power-Share-For-CellPortion-Value", "nbap.Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Received_Scheduled_EDCH_Power_Share_Value_PDU,
      { "Received-Scheduled-EDCH-Power-Share-Value", "nbap.Received_Scheduled_EDCH_Power_Share_Value_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RSEPS_Value_IncrDecrThres_PDU,
      { "RSEPS-Value-IncrDecrThres", "nbap.RSEPS_Value_IncrDecrThres",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Rx_Timing_Deviation_Value_LCR_PDU,
      { "Rx-Timing-Deviation-Value-LCR", "nbap.Rx_Timing_Deviation_Value_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Rx_Timing_Deviation_Value_768_PDU,
      { "Rx-Timing-Deviation-Value-768", "nbap.Rx_Timing_Deviation_Value_768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Rx_Timing_Deviation_Value_384_ext_PDU,
      { "Rx-Timing-Deviation-Value-384-ext", "nbap.Rx_Timing_Deviation_Value_384_ext",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RTWP_ReportingIndicator_PDU,
      { "RTWP-ReportingIndicator", "nbap.RTWP_ReportingIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_RTWP_ReportingIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_RTWP_CellPortion_ReportingIndicator_PDU,
      { "RTWP-CellPortion-ReportingIndicator", "nbap.RTWP_CellPortion_ReportingIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_RTWP_CellPortion_ReportingIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_AdjustmentPeriod_PDU,
      { "AdjustmentPeriod", "nbap.AdjustmentPeriod",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DPCCH_Power_Boosting_Capability_PDU,
      { "E-DPCCH-Power-Boosting-Capability", "nbap.E_DPCCH_Power_Boosting_Capability",
        FT_UINT32, BASE_DEC, VALS(nbap_E_DPCCH_Power_Boosting_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_SAT_Info_Almanac_ExtList_PDU,
      { "SAT-Info-Almanac-ExtList", "nbap.SAT_Info_Almanac_ExtList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ScaledAdjustmentRatio_PDU,
      { "ScaledAdjustmentRatio", "nbap.ScaledAdjustmentRatio",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MaxAdjustmentStep_PDU,
      { "MaxAdjustmentStep", "nbap.MaxAdjustmentStep",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SchedulingPriorityIndicator_PDU,
      { "SchedulingPriorityIndicator", "nbap.SchedulingPriorityIndicator",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Secondary_CPICH_Information_Change_PDU,
      { "Secondary-CPICH-Information-Change", "nbap.Secondary_CPICH_Information_Change",
        FT_UINT32, BASE_DEC, VALS(nbap_Secondary_CPICH_Information_Change_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Secondary_CCPCH_SlotFormat_Extended_PDU,
      { "Secondary-CCPCH-SlotFormat-Extended", "nbap.Secondary_CCPCH_SlotFormat_Extended",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Serving_E_DCH_RL_ID_PDU,
      { "Serving-E-DCH-RL-ID", "nbap.Serving_E_DCH_RL_ID",
        FT_UINT32, BASE_DEC, VALS(nbap_Serving_E_DCH_RL_ID_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_SFN_PDU,
      { "SFN", "nbap.SFN",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SFNSFNMeasurementThresholdInformation_PDU,
      { "SFNSFNMeasurementThresholdInformation", "nbap.SFNSFNMeasurementThresholdInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SFNSFNMeasurementValueInformation_PDU,
      { "SFNSFNMeasurementValueInformation", "nbap.SFNSFNMeasurementValueInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ShutdownTimer_PDU,
      { "ShutdownTimer", "nbap.ShutdownTimer",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Single_Stream_MIMO_ActivationIndicator_PDU,
      { "Single-Stream-MIMO-ActivationIndicator", "nbap.Single_Stream_MIMO_ActivationIndicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Single_Stream_MIMO_Capability_PDU,
      { "Single-Stream-MIMO-Capability", "nbap.Single_Stream_MIMO_Capability",
        FT_UINT32, BASE_DEC, VALS(nbap_Single_Stream_MIMO_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Single_Stream_MIMO_Mode_Indicator_PDU,
      { "Single-Stream-MIMO-Mode-Indicator", "nbap.Single_Stream_MIMO_Mode_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_Single_Stream_MIMO_Mode_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_SignallingBearerRequestIndicator_PDU,
      { "SignallingBearerRequestIndicator", "nbap.SignallingBearerRequestIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_SignallingBearerRequestIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_SixtyfourQAM_UsageAllowedIndicator_PDU,
      { "SixtyfourQAM-UsageAllowedIndicator", "nbap.SixtyfourQAM_UsageAllowedIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_SixtyfourQAM_UsageAllowedIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_SixtyfourQAM_DL_UsageIndicator_PDU,
      { "SixtyfourQAM-DL-UsageIndicator", "nbap.SixtyfourQAM_DL_UsageIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_SixtyfourQAM_DL_UsageIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_SixtyfourQAM_DL_Capability_PDU,
      { "SixtyfourQAM-DL-Capability", "nbap.SixtyfourQAM_DL_Capability",
        FT_UINT32, BASE_DEC, VALS(nbap_SixtyfourQAM_DL_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_SixtyfourQAM_DL_MIMO_Combined_Capability_PDU,
      { "SixtyfourQAM-DL-MIMO-Combined-Capability", "nbap.SixtyfourQAM_DL_MIMO_Combined_Capability",
        FT_UINT32, BASE_DEC, VALS(nbap_SixtyfourQAM_DL_MIMO_Combined_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_SixteenQAM_UL_Capability_PDU,
      { "SixteenQAM-UL-Capability", "nbap.SixteenQAM_UL_Capability",
        FT_UINT32, BASE_DEC, VALS(nbap_SixteenQAM_UL_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_SixteenQAM_UL_Operation_Indicator_PDU,
      { "SixteenQAM-UL-Operation-Indicator", "nbap.SixteenQAM_UL_Operation_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_SixteenQAM_UL_Operation_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Start_Of_Audit_Sequence_Indicator_PDU,
      { "Start-Of-Audit-Sequence-Indicator", "nbap.Start_Of_Audit_Sequence_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_Start_Of_Audit_Sequence_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_SyncCase_PDU,
      { "SyncCase", "nbap.SyncCase",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SynchronisationReportCharacteristics_PDU,
      { "SynchronisationReportCharacteristics", "nbap.SynchronisationReportCharacteristics_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SyncDLCodeIdThreInfoLCR_PDU,
      { "SyncDLCodeIdThreInfoLCR", "nbap.SyncDLCodeIdThreInfoLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SynchronisationReportType_PDU,
      { "SynchronisationReportType", "nbap.SynchronisationReportType",
        FT_UINT32, BASE_DEC, VALS(nbap_SynchronisationReportType_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Semi_PersistentScheduling_CapabilityLCR_PDU,
      { "Semi-PersistentScheduling-CapabilityLCR", "nbap.Semi_PersistentScheduling_CapabilityLCR",
        FT_UINT32, BASE_DEC, VALS(nbap_Semi_PersistentScheduling_CapabilityLCR_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_T_Cell_PDU,
      { "T-Cell", "nbap.T_Cell",
        FT_UINT32, BASE_DEC, VALS(nbap_T_Cell_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TDD_DL_DPCH_TimeSlotFormat_LCR_PDU,
      { "TDD-DL-DPCH-TimeSlotFormat-LCR", "nbap.TDD_DL_DPCH_TimeSlotFormat_LCR",
        FT_UINT32, BASE_DEC, VALS(nbap_TDD_DL_DPCH_TimeSlotFormat_LCR_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TDD_TPC_DownlinkStepSize_PDU,
      { "TDD-TPC-DownlinkStepSize", "nbap.TDD_TPC_DownlinkStepSize",
        FT_UINT32, BASE_DEC, VALS(nbap_TDD_TPC_DownlinkStepSize_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TDD_TPC_UplinkStepSize_LCR_PDU,
      { "TDD-TPC-UplinkStepSize-LCR", "nbap.TDD_TPC_UplinkStepSize_LCR",
        FT_UINT32, BASE_DEC, VALS(nbap_TDD_TPC_UplinkStepSize_LCR_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TDD_UL_DPCH_TimeSlotFormat_LCR_PDU,
      { "TDD-UL-DPCH-TimeSlotFormat-LCR", "nbap.TDD_UL_DPCH_TimeSlotFormat_LCR",
        FT_UINT32, BASE_DEC, VALS(nbap_TDD_UL_DPCH_TimeSlotFormat_LCR_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TFCI_Presence_PDU,
      { "TFCI-Presence", "nbap.TFCI_Presence",
        FT_UINT32, BASE_DEC, VALS(nbap_TFCI_Presence_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TimeSlot_PDU,
      { "TimeSlot", "nbap.TimeSlot",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TimeSlotLCR_PDU,
      { "TimeSlotLCR", "nbap.TimeSlotLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TimeslotLCR_Extension_PDU,
      { "TimeslotLCR-Extension", "nbap.TimeslotLCR_Extension",
        FT_UINT32, BASE_DEC, VALS(nbap_TimeslotLCR_Extension_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TimeSlotMeasurementValueListLCR_PDU,
      { "TimeSlotMeasurementValueListLCR", "nbap.TimeSlotMeasurementValueListLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TimingAdjustmentValue_PDU,
      { "TimingAdjustmentValue", "nbap.TimingAdjustmentValue",
        FT_UINT32, BASE_DEC, VALS(nbap_TimingAdjustmentValue_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TimingAdjustmentValueLCR_PDU,
      { "TimingAdjustmentValueLCR", "nbap.TimingAdjustmentValueLCR",
        FT_UINT32, BASE_DEC, VALS(nbap_TimingAdjustmentValueLCR_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TimingAdvanceApplied_PDU,
      { "TimingAdvanceApplied", "nbap.TimingAdvanceApplied",
        FT_UINT32, BASE_DEC, VALS(nbap_TimingAdvanceApplied_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_SynchronisationIndicator_PDU,
      { "SynchronisationIndicator", "nbap.SynchronisationIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_SynchronisationIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TnlQos_PDU,
      { "TnlQos", "nbap.TnlQos",
        FT_UINT32, BASE_DEC, VALS(nbap_TnlQos_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Transmission_Gap_Pattern_Sequence_Information_PDU,
      { "Transmission-Gap-Pattern-Sequence-Information", "nbap.Transmission_Gap_Pattern_Sequence_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue_PDU,
      { "TransmittedCarrierPowerOfAllCodesNotUsedForHS-PDSCH-HS-SCCH-E-AGCH-E-RGCHOrE-HICHTransmissionCellPortionValue", "nbap.TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue_PDU,
      { "TransmittedCarrierPowerOfAllCodesNotUsedForHS-PDSCH-HS-SCCH-E-AGCHOrE-HICHTransmissionCellPortionValue", "nbap.TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHSTransmissionValue_PDU,
      { "TransmittedCarrierPowerOfAllCodesNotUsedForHSTransmissionValue", "nbap.TransmittedCarrierPowerOfAllCodesNotUsedForHSTransmissionValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Transmitted_Carrier_Power_For_CellPortion_Value_PDU,
      { "Transmitted-Carrier-Power-For-CellPortion-Value", "nbap.Transmitted_Carrier_Power_For_CellPortion_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Transmitted_Carrier_Power_For_CellPortion_ValueLCR_PDU,
      { "Transmitted-Carrier-Power-For-CellPortion-ValueLCR", "nbap.Transmitted_Carrier_Power_For_CellPortion_ValueLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Transmitted_Carrier_Power_Value_PDU,
      { "Transmitted-Carrier-Power-Value", "nbap.Transmitted_Carrier_Power_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TransmissionDiversityApplied_PDU,
      { "TransmissionDiversityApplied", "nbap.TransmissionDiversityApplied",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TransmitDiversityIndicator_PDU,
      { "TransmitDiversityIndicator", "nbap.TransmitDiversityIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_TransmitDiversityIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TransportBearerRequestIndicator_PDU,
      { "TransportBearerRequestIndicator", "nbap.TransportBearerRequestIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_TransportBearerRequestIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TransportBearerNotRequestedIndicator_PDU,
      { "TransportBearerNotRequestedIndicator", "nbap.TransportBearerNotRequestedIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_TransportBearerNotRequestedIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TransportBearerNotSetupIndicator_PDU,
      { "TransportBearerNotSetupIndicator", "nbap.TransportBearerNotSetupIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_TransportBearerNotSetupIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TransportLayerAddress_PDU,
      { "TransportLayerAddress", "nbap.TransportLayerAddress",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TS0_CapabilityLCR_PDU,
      { "TS0-CapabilityLCR", "nbap.TS0_CapabilityLCR",
        FT_UINT32, BASE_DEC, VALS(nbap_TS0_CapabilityLCR_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TSTD_Indicator_PDU,
      { "TSTD-Indicator", "nbap.TSTD_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_TSTD_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TSN_Length_PDU,
      { "TSN-Length", "nbap.TSN_Length",
        FT_UINT32, BASE_DEC, VALS(nbap_TSN_Length_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TUTRANGANSSMeasurementThresholdInformation_PDU,
      { "TUTRANGANSSMeasurementThresholdInformation", "nbap.TUTRANGANSSMeasurementThresholdInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TUTRANGANSSMeasurementValueInformation_PDU,
      { "TUTRANGANSSMeasurementValueInformation", "nbap.TUTRANGANSSMeasurementValueInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TUTRANGPSMeasurementThresholdInformation_PDU,
      { "TUTRANGPSMeasurementThresholdInformation", "nbap.TUTRANGPSMeasurementThresholdInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TUTRANGPSMeasurementValueInformation_PDU,
      { "TUTRANGPSMeasurementValueInformation", "nbap.TUTRANGPSMeasurementValueInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TxDiversityOnDLControlChannelsByMIMOUECapability_PDU,
      { "TxDiversityOnDLControlChannelsByMIMOUECapability", "nbap.TxDiversityOnDLControlChannelsByMIMOUECapability",
        FT_UINT32, BASE_DEC, VALS(nbap_TxDiversityOnDLControlChannelsByMIMOUECapability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TypeOfError_PDU,
      { "TypeOfError", "nbap.TypeOfError",
        FT_UINT32, BASE_DEC, VALS(nbap_TypeOfError_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_UARFCN_PDU,
      { "UARFCN", "nbap.UARFCN",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UE_AggregateMaximumBitRate_PDU,
      { "UE-AggregateMaximumBitRate", "nbap.UE_AggregateMaximumBitRate_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UE_AggregateMaximumBitRate_Enforcement_Indicator_PDU,
      { "UE-AggregateMaximumBitRate-Enforcement-Indicator", "nbap.UE_AggregateMaximumBitRate_Enforcement_Indicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UE_Capability_Information_PDU,
      { "UE-Capability-Information", "nbap.UE_Capability_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UE_TS0_CapabilityLCR_PDU,
      { "UE-TS0-CapabilityLCR", "nbap.UE_TS0_CapabilityLCR",
        FT_UINT32, BASE_DEC, VALS(nbap_UE_TS0_CapabilityLCR_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_UE_SupportIndicatorExtension_PDU,
      { "UE-SupportIndicatorExtension", "nbap.UE_SupportIndicatorExtension",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_LCRTDD_HSDSCH_Physical_Layer_Category_PDU,
      { "LCRTDD-HSDSCH-Physical-Layer-Category", "nbap.LCRTDD_HSDSCH_Physical_Layer_Category",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_DPDCH_Indicator_For_E_DCH_Operation_PDU,
      { "UL-DPDCH-Indicator-For-E-DCH-Operation", "nbap.UL_DPDCH_Indicator_For_E_DCH_Operation",
        FT_UINT32, BASE_DEC, VALS(nbap_UL_DPDCH_Indicator_For_E_DCH_Operation_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_UL_SIR_PDU,
      { "UL-SIR", "nbap.UL_SIR",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_Synchronisation_Parameters_LCR_PDU,
      { "UL-Synchronisation-Parameters-LCR", "nbap.UL_Synchronisation_Parameters_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UPPCHPositionLCR_PDU,
      { "UPPCHPositionLCR", "nbap.UPPCHPositionLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UpPTSInterference_For_CellPortion_Value_PDU,
      { "UpPTSInterference-For-CellPortion-Value", "nbap.UpPTSInterference_For_CellPortion_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UpPTSInterferenceValue_PDU,
      { "UpPTSInterferenceValue", "nbap.UpPTSInterferenceValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Unidirectional_DCH_Indicator_PDU,
      { "Unidirectional-DCH-Indicator", "nbap.Unidirectional_DCH_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_Unidirectional_DCH_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_USCH_Information_PDU,
      { "USCH-Information", "nbap.USCH_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_USCH_InformationResponse_PDU,
      { "USCH-InformationResponse", "nbap.USCH_InformationResponse",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_TimeslotISCP_For_CellPortion_Value_PDU,
      { "UL-TimeslotISCP-For-CellPortion-Value", "nbap.UL_TimeslotISCP_For_CellPortion_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_TimeslotISCP_Value_IncrDecrThres_PDU,
      { "UL-TimeslotISCP-Value-IncrDecrThres", "nbap.UL_TimeslotISCP_Value_IncrDecrThres",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonTransportChannelSetupRequestFDD_PDU,
      { "CommonTransportChannelSetupRequestFDD", "nbap.CommonTransportChannelSetupRequestFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonPhysicalChannelType_CTCH_SetupRqstFDD_PDU,
      { "CommonPhysicalChannelType-CTCH-SetupRqstFDD", "nbap.CommonPhysicalChannelType_CTCH_SetupRqstFDD",
        FT_UINT32, BASE_DEC, VALS(nbap_CommonPhysicalChannelType_CTCH_SetupRqstFDD_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_FACH_ParametersListIE_CTCH_SetupRqstFDD_PDU,
      { "FACH-ParametersListIE-CTCH-SetupRqstFDD", "nbap.FACH_ParametersListIE_CTCH_SetupRqstFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PCH_ParametersItem_CTCH_SetupRqstFDD_PDU,
      { "PCH-ParametersItem-CTCH-SetupRqstFDD", "nbap.PCH_ParametersItem_CTCH_SetupRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MICH_Parameters_CTCH_SetupRqstFDD_PDU,
      { "MICH-Parameters-CTCH-SetupRqstFDD", "nbap.MICH_Parameters_CTCH_SetupRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RACH_ParametersItem_CTCH_SetupRqstFDD_PDU,
      { "RACH-ParametersItem-CTCH-SetupRqstFDD", "nbap.RACH_ParametersItem_CTCH_SetupRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonTransportChannelSetupRequestTDD_PDU,
      { "CommonTransportChannelSetupRequestTDD", "nbap.CommonTransportChannelSetupRequestTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonPhysicalChannelType_CTCH_SetupRqstTDD_PDU,
      { "CommonPhysicalChannelType-CTCH-SetupRqstTDD", "nbap.CommonPhysicalChannelType_CTCH_SetupRqstTDD",
        FT_UINT32, BASE_DEC, VALS(nbap_CommonPhysicalChannelType_CTCH_SetupRqstTDD_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Secondary_CCPCH_parameterListIE_CTCH_SetupRqstTDD_PDU,
      { "Secondary-CCPCH-parameterListIE-CTCH-SetupRqstTDD", "nbap.Secondary_CCPCH_parameterListIE_CTCH_SetupRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Secondary_CCPCH_LCR_parameterList_CTCH_SetupRqstTDD_PDU,
      { "Secondary-CCPCH-LCR-parameterList-CTCH-SetupRqstTDD", "nbap.Secondary_CCPCH_LCR_parameterList_CTCH_SetupRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Secondary_CCPCH_768_parameterList_CTCH_SetupRqstTDD_PDU,
      { "Secondary-CCPCH-768-parameterList-CTCH-SetupRqstTDD", "nbap.Secondary_CCPCH_768_parameterList_CTCH_SetupRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_FACH_ParametersListIE_CTCH_SetupRqstTDD_PDU,
      { "FACH-ParametersListIE-CTCH-SetupRqstTDD", "nbap.FACH_ParametersListIE_CTCH_SetupRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PCH_ParametersItem_CTCH_SetupRqstTDD_PDU,
      { "PCH-ParametersItem-CTCH-SetupRqstTDD", "nbap.PCH_ParametersItem_CTCH_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PICH_ParametersItem_CTCH_SetupRqstTDD_PDU,
      { "PICH-ParametersItem-CTCH-SetupRqstTDD", "nbap.PICH_ParametersItem_CTCH_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PICH_LCR_Parameters_CTCH_SetupRqstTDD_PDU,
      { "PICH-LCR-Parameters-CTCH-SetupRqstTDD", "nbap.PICH_LCR_Parameters_CTCH_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PICH_768_ParametersItem_CTCH_SetupRqstTDD_PDU,
      { "PICH-768-ParametersItem-CTCH-SetupRqstTDD", "nbap.PICH_768_ParametersItem_CTCH_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MICH_Parameters_CTCH_SetupRqstTDD_PDU,
      { "MICH-Parameters-CTCH-SetupRqstTDD", "nbap.MICH_Parameters_CTCH_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TimeSlotConfigurationList_LCR_CTCH_SetupRqstTDD_PDU,
      { "TimeSlotConfigurationList-LCR-CTCH-SetupRqstTDD", "nbap.TimeSlotConfigurationList_LCR_CTCH_SetupRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Secondary_CCPCH_parameterExtendedList_CTCH_SetupRqstTDD_PDU,
      { "Secondary-CCPCH-parameterExtendedList-CTCH-SetupRqstTDD", "nbap.Secondary_CCPCH_parameterExtendedList_CTCH_SetupRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_SetupRqstTDD_PDU,
      { "Secondary-CCPCH-LCR-parameterExtendedList-CTCH-SetupRqstTDD", "nbap.Secondary_CCPCH_LCR_parameterExtendedList_CTCH_SetupRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PRACH_ParametersItem_CTCH_SetupRqstTDD_PDU,
      { "PRACH-ParametersItem-CTCH-SetupRqstTDD", "nbap.PRACH_ParametersItem_CTCH_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RACH_ParameterItem_CTCH_SetupRqstTDD_PDU,
      { "RACH-ParameterItem-CTCH-SetupRqstTDD", "nbap.RACH_ParameterItem_CTCH_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PRACH_LCR_ParametersList_CTCH_SetupRqstTDD_PDU,
      { "PRACH-LCR-ParametersList-CTCH-SetupRqstTDD", "nbap.PRACH_LCR_ParametersList_CTCH_SetupRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PRACH_768_ParametersItem_CTCH_SetupRqstTDD_PDU,
      { "PRACH-768-ParametersItem-CTCH-SetupRqstTDD", "nbap.PRACH_768_ParametersItem_CTCH_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_FPACH_LCR_Parameters_CTCH_SetupRqstTDD_PDU,
      { "FPACH-LCR-Parameters-CTCH-SetupRqstTDD", "nbap.FPACH_LCR_Parameters_CTCH_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PLCCH_parameters_PDU,
      { "PLCCH-parameters", "nbap.PLCCH_parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_RUCCH_parameters_PDU,
      { "E-RUCCH-parameters", "nbap.E_RUCCH_parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_RUCCH_768_parameters_PDU,
      { "E-RUCCH-768-parameters", "nbap.E_RUCCH_768_parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonTransportChannelSetupResponse_PDU,
      { "CommonTransportChannelSetupResponse", "nbap.CommonTransportChannelSetupResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_FACH_CommonTransportChannel_InformationResponse_PDU,
      { "FACH-CommonTransportChannel-InformationResponse", "nbap.FACH_CommonTransportChannel_InformationResponse",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonTransportChannelSetupFailure_PDU,
      { "CommonTransportChannelSetupFailure", "nbap.CommonTransportChannelSetupFailure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonTransportChannelReconfigurationRequestFDD_PDU,
      { "CommonTransportChannelReconfigurationRequestFDD", "nbap.CommonTransportChannelReconfigurationRequestFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonPhysicalChannelType_CTCH_ReconfRqstFDD_PDU,
      { "CommonPhysicalChannelType-CTCH-ReconfRqstFDD", "nbap.CommonPhysicalChannelType_CTCH_ReconfRqstFDD",
        FT_UINT32, BASE_DEC, VALS(nbap_CommonPhysicalChannelType_CTCH_ReconfRqstFDD_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_FACH_ParametersListIE_CTCH_ReconfRqstFDD_PDU,
      { "FACH-ParametersListIE-CTCH-ReconfRqstFDD", "nbap.FACH_ParametersListIE_CTCH_ReconfRqstFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PCH_ParametersItem_CTCH_ReconfRqstFDD_PDU,
      { "PCH-ParametersItem-CTCH-ReconfRqstFDD", "nbap.PCH_ParametersItem_CTCH_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PICH_ParametersItem_CTCH_ReconfRqstFDD_PDU,
      { "PICH-ParametersItem-CTCH-ReconfRqstFDD", "nbap.PICH_ParametersItem_CTCH_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MICH_Parameters_CTCH_ReconfRqstFDD_PDU,
      { "MICH-Parameters-CTCH-ReconfRqstFDD", "nbap.MICH_Parameters_CTCH_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PRACH_ParametersListIE_CTCH_ReconfRqstFDD_PDU,
      { "PRACH-ParametersListIE-CTCH-ReconfRqstFDD", "nbap.PRACH_ParametersListIE_CTCH_ReconfRqstFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_AICH_ParametersListIE_CTCH_ReconfRqstFDD_PDU,
      { "AICH-ParametersListIE-CTCH-ReconfRqstFDD", "nbap.AICH_ParametersListIE_CTCH_ReconfRqstFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonTransportChannelReconfigurationRequestTDD_PDU,
      { "CommonTransportChannelReconfigurationRequestTDD", "nbap.CommonTransportChannelReconfigurationRequestTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Secondary_CCPCH_Parameters_CTCH_ReconfRqstTDD_PDU,
      { "Secondary-CCPCH-Parameters-CTCH-ReconfRqstTDD", "nbap.Secondary_CCPCH_Parameters_CTCH_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Secondary_CCPCHListIE_CTCH_ReconfRqstTDD_PDU,
      { "Secondary-CCPCHListIE-CTCH-ReconfRqstTDD", "nbap.Secondary_CCPCHListIE_CTCH_ReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Secondary_CCPCH_parameterExtendedList_CTCH_ReconfRqstTDD_PDU,
      { "Secondary-CCPCH-parameterExtendedList-CTCH-ReconfRqstTDD", "nbap.Secondary_CCPCH_parameterExtendedList_CTCH_ReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_ReconfRqstTDD_PDU,
      { "Secondary-CCPCH-LCR-parameterExtendedList-CTCH-ReconfRqstTDD", "nbap.Secondary_CCPCH_LCR_parameterExtendedList_CTCH_ReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PICH_Parameters_CTCH_ReconfRqstTDD_PDU,
      { "PICH-Parameters-CTCH-ReconfRqstTDD", "nbap.PICH_Parameters_CTCH_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_FACH_ParametersList_CTCH_ReconfRqstTDD_PDU,
      { "FACH-ParametersList-CTCH-ReconfRqstTDD", "nbap.FACH_ParametersList_CTCH_ReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PCH_Parameters_CTCH_ReconfRqstTDD_PDU,
      { "PCH-Parameters-CTCH-ReconfRqstTDD", "nbap.PCH_Parameters_CTCH_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_FPACH_LCR_Parameters_CTCH_ReconfRqstTDD_PDU,
      { "FPACH-LCR-Parameters-CTCH-ReconfRqstTDD", "nbap.FPACH_LCR_Parameters_CTCH_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MICH_Parameters_CTCH_ReconfRqstTDD_PDU,
      { "MICH-Parameters-CTCH-ReconfRqstTDD", "nbap.MICH_Parameters_CTCH_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PLCCH_Parameters_CTCH_ReconfRqstTDD_PDU,
      { "PLCCH-Parameters-CTCH-ReconfRqstTDD", "nbap.PLCCH_Parameters_CTCH_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Secondary_CCPCH_768_Parameters_CTCH_ReconfRqstTDD_PDU,
      { "Secondary-CCPCH-768-Parameters-CTCH-ReconfRqstTDD", "nbap.Secondary_CCPCH_768_Parameters_CTCH_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PICH_768_Parameters_CTCH_ReconfRqstTDD_PDU,
      { "PICH-768-Parameters-CTCH-ReconfRqstTDD", "nbap.PICH_768_Parameters_CTCH_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MICH_768_Parameters_CTCH_ReconfRqstTDD_PDU,
      { "MICH-768-Parameters-CTCH-ReconfRqstTDD", "nbap.MICH_768_Parameters_CTCH_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UPPCH_LCR_Parameters_CTCH_ReconfRqstTDD_PDU,
      { "UPPCH-LCR-Parameters-CTCH-ReconfRqstTDD", "nbap.UPPCH_LCR_Parameters_CTCH_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonTransportChannelReconfigurationResponse_PDU,
      { "CommonTransportChannelReconfigurationResponse", "nbap.CommonTransportChannelReconfigurationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonTransportChannelReconfigurationFailure_PDU,
      { "CommonTransportChannelReconfigurationFailure", "nbap.CommonTransportChannelReconfigurationFailure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonTransportChannelDeletionRequest_PDU,
      { "CommonTransportChannelDeletionRequest", "nbap.CommonTransportChannelDeletionRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonTransportChannelDeletionResponse_PDU,
      { "CommonTransportChannelDeletionResponse", "nbap.CommonTransportChannelDeletionResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_BlockResourceRequest_PDU,
      { "BlockResourceRequest", "nbap.BlockResourceRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_BlockResourceResponse_PDU,
      { "BlockResourceResponse", "nbap.BlockResourceResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_BlockResourceFailure_PDU,
      { "BlockResourceFailure", "nbap.BlockResourceFailure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UnblockResourceIndication_PDU,
      { "UnblockResourceIndication", "nbap.UnblockResourceIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_AuditRequiredIndication_PDU,
      { "AuditRequiredIndication", "nbap.AuditRequiredIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_AuditRequest_PDU,
      { "AuditRequest", "nbap.AuditRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_AuditResponse_PDU,
      { "AuditResponse", "nbap.AuditResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Cell_InformationList_AuditRsp_PDU,
      { "Cell-InformationList-AuditRsp", "nbap.Cell_InformationList_AuditRsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Cell_InformationItem_AuditRsp_PDU,
      { "Cell-InformationItem-AuditRsp", "nbap.Cell_InformationItem_AuditRsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_FPACH_LCR_InformationList_AuditRsp_PDU,
      { "FPACH-LCR-InformationList-AuditRsp", "nbap.FPACH_LCR_InformationList_AuditRsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_DSCH_Resources_Information_AuditRsp_PDU,
      { "HS-DSCH-Resources-Information-AuditRsp", "nbap.HS_DSCH_Resources_Information_AuditRsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_S_CCPCH_InformationListExt_AuditRsp_PDU,
      { "S-CCPCH-InformationListExt-AuditRsp", "nbap.S_CCPCH_InformationListExt_AuditRsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_S_CCPCH_LCR_InformationListExt_AuditRsp_PDU,
      { "S-CCPCH-LCR-InformationListExt-AuditRsp", "nbap.S_CCPCH_LCR_InformationListExt_AuditRsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_Resources_Information_AuditRsp_PDU,
      { "E-DCH-Resources-Information-AuditRsp", "nbap.E_DCH_Resources_Information_AuditRsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PLCCH_InformationList_AuditRsp_PDU,
      { "PLCCH-InformationList-AuditRsp", "nbap.PLCCH_InformationList_AuditRsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_S_CCPCH_768_InformationList_AuditRsp_PDU,
      { "S-CCPCH-768-InformationList-AuditRsp", "nbap.S_CCPCH_768_InformationList_AuditRsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PRACH_768_InformationList_AuditRsp_PDU,
      { "PRACH-768-InformationList-AuditRsp", "nbap.PRACH_768_InformationList_AuditRsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_RUCCH_InformationList_AuditRsp_PDU,
      { "E-RUCCH-InformationList-AuditRsp", "nbap.E_RUCCH_InformationList_AuditRsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_RUCCH_768_InformationList_AuditRsp_PDU,
      { "E-RUCCH-768-InformationList-AuditRsp", "nbap.E_RUCCH_768_InformationList_AuditRsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Cell_Frequency_List_Information_LCR_MulFreq_AuditRsp_PDU,
      { "Cell-Frequency-List-Information-LCR-MulFreq-AuditRsp", "nbap.Cell_Frequency_List_Information_LCR_MulFreq_AuditRsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Cell_Frequency_List_InformationItem_LCR_MulFreq_AuditRsp_PDU,
      { "Cell-Frequency-List-InformationItem-LCR-MulFreq-AuditRsp", "nbap.Cell_Frequency_List_InformationItem_LCR_MulFreq_AuditRsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UPPCH_LCR_InformationList_AuditRsp_PDU,
      { "UPPCH-LCR-InformationList-AuditRsp", "nbap.UPPCH_LCR_InformationList_AuditRsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UPPCH_LCR_InformationItem_AuditRsp_PDU,
      { "UPPCH-LCR-InformationItem-AuditRsp", "nbap.UPPCH_LCR_InformationItem_AuditRsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleFreq_HS_DSCH_Resources_InformationList_AuditRsp_PDU,
      { "MultipleFreq-HS-DSCH-Resources-InformationList-AuditRsp", "nbap.MultipleFreq_HS_DSCH_Resources_InformationList_AuditRsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleFreq_E_DCH_Resources_InformationList_AuditRsp_PDU,
      { "MultipleFreq-E-DCH-Resources-InformationList-AuditRsp", "nbap.MultipleFreq_E_DCH_Resources_InformationList_AuditRsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CCP_InformationList_AuditRsp_PDU,
      { "CCP-InformationList-AuditRsp", "nbap.CCP_InformationList_AuditRsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CCP_InformationItem_AuditRsp_PDU,
      { "CCP-InformationItem-AuditRsp", "nbap.CCP_InformationItem_AuditRsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Local_Cell_InformationList_AuditRsp_PDU,
      { "Local-Cell-InformationList-AuditRsp", "nbap.Local_Cell_InformationList_AuditRsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Local_Cell_InformationItem_AuditRsp_PDU,
      { "Local-Cell-InformationItem-AuditRsp", "nbap.Local_Cell_InformationItem_AuditRsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Local_Cell_Group_InformationList_AuditRsp_PDU,
      { "Local-Cell-Group-InformationList-AuditRsp", "nbap.Local_Cell_Group_InformationList_AuditRsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Local_Cell_Group_InformationItem_AuditRsp_PDU,
      { "Local-Cell-Group-InformationItem-AuditRsp", "nbap.Local_Cell_Group_InformationItem_AuditRsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Power_Local_Cell_Group_InformationList_AuditRsp_PDU,
      { "Power-Local-Cell-Group-InformationList-AuditRsp", "nbap.Power_Local_Cell_Group_InformationList_AuditRsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Power_Local_Cell_Group_InformationItem_AuditRsp_PDU,
      { "Power-Local-Cell-Group-InformationItem-AuditRsp", "nbap.Power_Local_Cell_Group_InformationItem_AuditRsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_AuditFailure_PDU,
      { "AuditFailure", "nbap.AuditFailure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonMeasurementInitiationRequest_PDU,
      { "CommonMeasurementInitiationRequest", "nbap.CommonMeasurementInitiationRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonMeasurementObjectType_CM_Rqst_PDU,
      { "CommonMeasurementObjectType-CM-Rqst", "nbap.CommonMeasurementObjectType_CM_Rqst",
        FT_UINT32, BASE_DEC, VALS(nbap_CommonMeasurementObjectType_CM_Rqst_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_ERACH_CM_Rqst_PDU,
      { "ERACH-CM-Rqst", "nbap.ERACH_CM_Rqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PowerLocalCellGroup_CM_Rqst_PDU,
      { "PowerLocalCellGroup-CM-Rqst", "nbap.PowerLocalCellGroup_CM_Rqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonMeasurementInitiationResponse_PDU,
      { "CommonMeasurementInitiationResponse", "nbap.CommonMeasurementInitiationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonMeasurementObjectType_CM_Rsp_PDU,
      { "CommonMeasurementObjectType-CM-Rsp", "nbap.CommonMeasurementObjectType_CM_Rsp",
        FT_UINT32, BASE_DEC, VALS(nbap_CommonMeasurementObjectType_CM_Rsp_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_ERACH_CM_Rsp_PDU,
      { "ERACH-CM-Rsp", "nbap.ERACH_CM_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PowerLocalCellGroup_CM_Rsp_PDU,
      { "PowerLocalCellGroup-CM-Rsp", "nbap.PowerLocalCellGroup_CM_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonMeasurementInitiationFailure_PDU,
      { "CommonMeasurementInitiationFailure", "nbap.CommonMeasurementInitiationFailure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonMeasurementReport_PDU,
      { "CommonMeasurementReport", "nbap.CommonMeasurementReport_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonMeasurementObjectType_CM_Rprt_PDU,
      { "CommonMeasurementObjectType-CM-Rprt", "nbap.CommonMeasurementObjectType_CM_Rprt",
        FT_UINT32, BASE_DEC, VALS(nbap_CommonMeasurementObjectType_CM_Rprt_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_ERACH_CM_Rprt_PDU,
      { "ERACH-CM-Rprt", "nbap.ERACH_CM_Rprt_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PowerLocalCellGroup_CM_Rprt_PDU,
      { "PowerLocalCellGroup-CM-Rprt", "nbap.PowerLocalCellGroup_CM_Rprt_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonMeasurementTerminationRequest_PDU,
      { "CommonMeasurementTerminationRequest", "nbap.CommonMeasurementTerminationRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonMeasurementFailureIndication_PDU,
      { "CommonMeasurementFailureIndication", "nbap.CommonMeasurementFailureIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSetupRequestFDD_PDU,
      { "CellSetupRequestFDD", "nbap.CellSetupRequestFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Synchronisation_Configuration_Cell_SetupRqst_PDU,
      { "Synchronisation-Configuration-Cell-SetupRqst", "nbap.Synchronisation_Configuration_Cell_SetupRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PrimarySCH_Information_Cell_SetupRqstFDD_PDU,
      { "PrimarySCH-Information-Cell-SetupRqstFDD", "nbap.PrimarySCH_Information_Cell_SetupRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SecondarySCH_Information_Cell_SetupRqstFDD_PDU,
      { "SecondarySCH-Information-Cell-SetupRqstFDD", "nbap.SecondarySCH_Information_Cell_SetupRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PrimaryCPICH_Information_Cell_SetupRqstFDD_PDU,
      { "PrimaryCPICH-Information-Cell-SetupRqstFDD", "nbap.PrimaryCPICH_Information_Cell_SetupRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SecondaryCPICH_InformationList_Cell_SetupRqstFDD_PDU,
      { "SecondaryCPICH-InformationList-Cell-SetupRqstFDD", "nbap.SecondaryCPICH_InformationList_Cell_SetupRqstFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SecondaryCPICH_InformationItem_Cell_SetupRqstFDD_PDU,
      { "SecondaryCPICH-InformationItem-Cell-SetupRqstFDD", "nbap.SecondaryCPICH_InformationItem_Cell_SetupRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PrimaryCCPCH_Information_Cell_SetupRqstFDD_PDU,
      { "PrimaryCCPCH-Information-Cell-SetupRqstFDD", "nbap.PrimaryCCPCH_Information_Cell_SetupRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Limited_power_increase_information_Cell_SetupRqstFDD_PDU,
      { "Limited-power-increase-information-Cell-SetupRqstFDD", "nbap.Limited_power_increase_information_Cell_SetupRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_IPDLParameter_Information_Cell_SetupRqstFDD_PDU,
      { "IPDLParameter-Information-Cell-SetupRqstFDD", "nbap.IPDLParameter_Information_Cell_SetupRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellPortion_InformationList_Cell_SetupRqstFDD_PDU,
      { "CellPortion-InformationList-Cell-SetupRqstFDD", "nbap.CellPortion_InformationList_Cell_SetupRqstFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellPortion_InformationItem_Cell_SetupRqstFDD_PDU,
      { "CellPortion-InformationItem-Cell-SetupRqstFDD", "nbap.CellPortion_InformationItem_Cell_SetupRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSetupRequestTDD_PDU,
      { "CellSetupRequestTDD", "nbap.CellSetupRequestTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SCH_Information_Cell_SetupRqstTDD_PDU,
      { "SCH-Information-Cell-SetupRqstTDD", "nbap.SCH_Information_Cell_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SyncCaseIndicatorItem_Cell_SetupRqstTDD_PSCH_PDU,
      { "SyncCaseIndicatorItem-Cell-SetupRqstTDD-PSCH", "nbap.SyncCaseIndicatorItem_Cell_SetupRqstTDD_PSCH",
        FT_UINT32, BASE_DEC, VALS(nbap_SyncCaseIndicatorItem_Cell_SetupRqstTDD_PSCH_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_PCCPCH_Information_Cell_SetupRqstTDD_PDU,
      { "PCCPCH-Information-Cell-SetupRqstTDD", "nbap.PCCPCH_Information_Cell_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TimeSlotConfigurationList_Cell_SetupRqstTDD_PDU,
      { "TimeSlotConfigurationList-Cell-SetupRqstTDD", "nbap.TimeSlotConfigurationList_Cell_SetupRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TimeSlotConfigurationList_LCR_Cell_SetupRqstTDD_PDU,
      { "TimeSlotConfigurationList-LCR-Cell-SetupRqstTDD", "nbap.TimeSlotConfigurationList_LCR_Cell_SetupRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PCCPCH_LCR_Information_Cell_SetupRqstTDD_PDU,
      { "PCCPCH-LCR-Information-Cell-SetupRqstTDD", "nbap.PCCPCH_LCR_Information_Cell_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DwPCH_LCR_Information_Cell_SetupRqstTDD_PDU,
      { "DwPCH-LCR-Information-Cell-SetupRqstTDD", "nbap.DwPCH_LCR_Information_Cell_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_IPDLParameter_Information_Cell_SetupRqstTDD_PDU,
      { "IPDLParameter-Information-Cell-SetupRqstTDD", "nbap.IPDLParameter_Information_Cell_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_IPDLParameter_Information_LCR_Cell_SetupRqstTDD_PDU,
      { "IPDLParameter-Information-LCR-Cell-SetupRqstTDD", "nbap.IPDLParameter_Information_LCR_Cell_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PCCPCH_768_Information_Cell_SetupRqstTDD_PDU,
      { "PCCPCH-768-Information-Cell-SetupRqstTDD", "nbap.PCCPCH_768_Information_Cell_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SCH_768_Information_Cell_SetupRqstTDD_PDU,
      { "SCH-768-Information-Cell-SetupRqstTDD", "nbap.SCH_768_Information_Cell_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Cell_Frequency_List_LCR_MulFreq_Cell_SetupRqstTDD_PDU,
      { "Cell-Frequency-List-LCR-MulFreq-Cell-SetupRqstTDD", "nbap.Cell_Frequency_List_LCR_MulFreq_Cell_SetupRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSetupResponse_PDU,
      { "CellSetupResponse", "nbap.CellSetupResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSetupFailure_PDU,
      { "CellSetupFailure", "nbap.CellSetupFailure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellReconfigurationRequestFDD_PDU,
      { "CellReconfigurationRequestFDD", "nbap.CellReconfigurationRequestFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Synchronisation_Configuration_Cell_ReconfRqst_PDU,
      { "Synchronisation-Configuration-Cell-ReconfRqst", "nbap.Synchronisation_Configuration_Cell_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PrimarySCH_Information_Cell_ReconfRqstFDD_PDU,
      { "PrimarySCH-Information-Cell-ReconfRqstFDD", "nbap.PrimarySCH_Information_Cell_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SecondarySCH_Information_Cell_ReconfRqstFDD_PDU,
      { "SecondarySCH-Information-Cell-ReconfRqstFDD", "nbap.SecondarySCH_Information_Cell_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PrimaryCPICH_Information_Cell_ReconfRqstFDD_PDU,
      { "PrimaryCPICH-Information-Cell-ReconfRqstFDD", "nbap.PrimaryCPICH_Information_Cell_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SecondaryCPICH_InformationList_Cell_ReconfRqstFDD_PDU,
      { "SecondaryCPICH-InformationList-Cell-ReconfRqstFDD", "nbap.SecondaryCPICH_InformationList_Cell_ReconfRqstFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SecondaryCPICH_InformationItem_Cell_ReconfRqstFDD_PDU,
      { "SecondaryCPICH-InformationItem-Cell-ReconfRqstFDD", "nbap.SecondaryCPICH_InformationItem_Cell_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PrimaryCCPCH_Information_Cell_ReconfRqstFDD_PDU,
      { "PrimaryCCPCH-Information-Cell-ReconfRqstFDD", "nbap.PrimaryCCPCH_Information_Cell_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_IPDLParameter_Information_Cell_ReconfRqstFDD_PDU,
      { "IPDLParameter-Information-Cell-ReconfRqstFDD", "nbap.IPDLParameter_Information_Cell_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellPortion_InformationList_Cell_ReconfRqstFDD_PDU,
      { "CellPortion-InformationList-Cell-ReconfRqstFDD", "nbap.CellPortion_InformationList_Cell_ReconfRqstFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellPortion_InformationItem_Cell_ReconfRqstFDD_PDU,
      { "CellPortion-InformationItem-Cell-ReconfRqstFDD", "nbap.CellPortion_InformationItem_Cell_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellReconfigurationRequestTDD_PDU,
      { "CellReconfigurationRequestTDD", "nbap.CellReconfigurationRequestTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SCH_Information_Cell_ReconfRqstTDD_PDU,
      { "SCH-Information-Cell-ReconfRqstTDD", "nbap.SCH_Information_Cell_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PCCPCH_Information_Cell_ReconfRqstTDD_PDU,
      { "PCCPCH-Information-Cell-ReconfRqstTDD", "nbap.PCCPCH_Information_Cell_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TimeSlotConfigurationList_Cell_ReconfRqstTDD_PDU,
      { "TimeSlotConfigurationList-Cell-ReconfRqstTDD", "nbap.TimeSlotConfigurationList_Cell_ReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TimeSlotConfigurationList_LCR_Cell_ReconfRqstTDD_PDU,
      { "TimeSlotConfigurationList-LCR-Cell-ReconfRqstTDD", "nbap.TimeSlotConfigurationList_LCR_Cell_ReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DwPCH_LCR_Information_Cell_ReconfRqstTDD_PDU,
      { "DwPCH-LCR-Information-Cell-ReconfRqstTDD", "nbap.DwPCH_LCR_Information_Cell_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_IPDLParameter_Information_Cell_ReconfRqstTDD_PDU,
      { "IPDLParameter-Information-Cell-ReconfRqstTDD", "nbap.IPDLParameter_Information_Cell_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_IPDLParameter_Information_LCR_Cell_ReconfRqstTDD_PDU,
      { "IPDLParameter-Information-LCR-Cell-ReconfRqstTDD", "nbap.IPDLParameter_Information_LCR_Cell_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SCH_768_Information_Cell_ReconfRqstTDD_PDU,
      { "SCH-768-Information-Cell-ReconfRqstTDD", "nbap.SCH_768_Information_Cell_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PCCPCH_768_Information_Cell_ReconfRqstTDD_PDU,
      { "PCCPCH-768-Information-Cell-ReconfRqstTDD", "nbap.PCCPCH_768_Information_Cell_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UARFCN_Adjustment_PDU,
      { "UARFCN-Adjustment", "nbap.UARFCN_Adjustment",
        FT_UINT32, BASE_DEC, VALS(nbap_UARFCN_Adjustment_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_CellReconfigurationResponse_PDU,
      { "CellReconfigurationResponse", "nbap.CellReconfigurationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellReconfigurationFailure_PDU,
      { "CellReconfigurationFailure", "nbap.CellReconfigurationFailure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellDeletionRequest_PDU,
      { "CellDeletionRequest", "nbap.CellDeletionRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellDeletionResponse_PDU,
      { "CellDeletionResponse", "nbap.CellDeletionResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ResourceStatusIndication_PDU,
      { "ResourceStatusIndication", "nbap.ResourceStatusIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_IndicationType_ResourceStatusInd_PDU,
      { "IndicationType-ResourceStatusInd", "nbap.IndicationType_ResourceStatusInd",
        FT_UINT32, BASE_DEC, VALS(nbap_IndicationType_ResourceStatusInd_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Local_Cell_InformationItem_ResourceStatusInd_PDU,
      { "Local-Cell-InformationItem-ResourceStatusInd", "nbap.Local_Cell_InformationItem_ResourceStatusInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Local_Cell_Group_InformationItem_ResourceStatusInd_PDU,
      { "Local-Cell-Group-InformationItem-ResourceStatusInd", "nbap.Local_Cell_Group_InformationItem_ResourceStatusInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Power_Local_Cell_Group_InformationList_ResourceStatusInd_PDU,
      { "Power-Local-Cell-Group-InformationList-ResourceStatusInd", "nbap.Power_Local_Cell_Group_InformationList_ResourceStatusInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Power_Local_Cell_Group_InformationItem_ResourceStatusInd_PDU,
      { "Power-Local-Cell-Group-InformationItem-ResourceStatusInd", "nbap.Power_Local_Cell_Group_InformationItem_ResourceStatusInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Local_Cell_InformationItem2_ResourceStatusInd_PDU,
      { "Local-Cell-InformationItem2-ResourceStatusInd", "nbap.Local_Cell_InformationItem2_ResourceStatusInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Local_Cell_Group_InformationItem2_ResourceStatusInd_PDU,
      { "Local-Cell-Group-InformationItem2-ResourceStatusInd", "nbap.Local_Cell_Group_InformationItem2_ResourceStatusInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CCP_InformationItem_ResourceStatusInd_PDU,
      { "CCP-InformationItem-ResourceStatusInd", "nbap.CCP_InformationItem_ResourceStatusInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Cell_InformationItem_ResourceStatusInd_PDU,
      { "Cell-InformationItem-ResourceStatusInd", "nbap.Cell_InformationItem_ResourceStatusInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_FPACH_LCR_InformationList_ResourceStatusInd_PDU,
      { "FPACH-LCR-InformationList-ResourceStatusInd", "nbap.FPACH_LCR_InformationList_ResourceStatusInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DwPCH_LCR_Information_ResourceStatusInd_PDU,
      { "DwPCH-LCR-Information-ResourceStatusInd", "nbap.DwPCH_LCR_Information_ResourceStatusInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_DSCH_Resources_Information_ResourceStatusInd_PDU,
      { "HS-DSCH-Resources-Information-ResourceStatusInd", "nbap.HS_DSCH_Resources_Information_ResourceStatusInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_S_CCPCH_InformationListExt_ResourceStatusInd_PDU,
      { "S-CCPCH-InformationListExt-ResourceStatusInd", "nbap.S_CCPCH_InformationListExt_ResourceStatusInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_S_CCPCH_LCR_InformationListExt_ResourceStatusInd_PDU,
      { "S-CCPCH-LCR-InformationListExt-ResourceStatusInd", "nbap.S_CCPCH_LCR_InformationListExt_ResourceStatusInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_Resources_Information_ResourceStatusInd_PDU,
      { "E-DCH-Resources-Information-ResourceStatusInd", "nbap.E_DCH_Resources_Information_ResourceStatusInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PLCCH_InformationList_ResourceStatusInd_PDU,
      { "PLCCH-InformationList-ResourceStatusInd", "nbap.PLCCH_InformationList_ResourceStatusInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_S_CCPCH_768_InformationList_ResourceStatusInd_PDU,
      { "S-CCPCH-768-InformationList-ResourceStatusInd", "nbap.S_CCPCH_768_InformationList_ResourceStatusInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PRACH_768_InformationList_ResourceStatusInd_PDU,
      { "PRACH-768-InformationList-ResourceStatusInd", "nbap.PRACH_768_InformationList_ResourceStatusInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_RUCCH_InformationList_ResourceStatusInd_PDU,
      { "E-RUCCH-InformationList-ResourceStatusInd", "nbap.E_RUCCH_InformationList_ResourceStatusInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_RUCCH_768_InformationList_ResourceStatusInd_PDU,
      { "E-RUCCH-768-InformationList-ResourceStatusInd", "nbap.E_RUCCH_768_InformationList_ResourceStatusInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Cell_Frequency_List_Information_LCR_MulFreq_ResourceStatusInd_PDU,
      { "Cell-Frequency-List-Information-LCR-MulFreq-ResourceStatusInd", "nbap.Cell_Frequency_List_Information_LCR_MulFreq_ResourceStatusInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Cell_Frequency_List_InformationItem_LCR_MulFreq_ResourceStatusInd_PDU,
      { "Cell-Frequency-List-InformationItem-LCR-MulFreq-ResourceStatusInd", "nbap.Cell_Frequency_List_InformationItem_LCR_MulFreq_ResourceStatusInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UPPCH_LCR_InformationList_ResourceStatusInd_PDU,
      { "UPPCH-LCR-InformationList-ResourceStatusInd", "nbap.UPPCH_LCR_InformationList_ResourceStatusInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UPPCH_LCR_InformationItem_ResourceStatusInd_PDU,
      { "UPPCH-LCR-InformationItem-ResourceStatusInd", "nbap.UPPCH_LCR_InformationItem_ResourceStatusInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleFreq_HS_DSCH_Resources_InformationList_ResourceStatusInd_PDU,
      { "MultipleFreq-HS-DSCH-Resources-InformationList-ResourceStatusInd", "nbap.MultipleFreq_HS_DSCH_Resources_InformationList_ResourceStatusInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Power_Local_Cell_Group_InformationList2_ResourceStatusInd_PDU,
      { "Power-Local-Cell-Group-InformationList2-ResourceStatusInd", "nbap.Power_Local_Cell_Group_InformationList2_ResourceStatusInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Power_Local_Cell_Group_InformationItem2_ResourceStatusInd_PDU,
      { "Power-Local-Cell-Group-InformationItem2-ResourceStatusInd", "nbap.Power_Local_Cell_Group_InformationItem2_ResourceStatusInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleFreq_E_DCH_Resources_InformationList_ResourceStatusInd_PDU,
      { "MultipleFreq-E-DCH-Resources-InformationList-ResourceStatusInd", "nbap.MultipleFreq_E_DCH_Resources_InformationList_ResourceStatusInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SystemInformationUpdateRequest_PDU,
      { "SystemInformationUpdateRequest", "nbap.SystemInformationUpdateRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MIB_SB_SIB_InformationList_SystemInfoUpdateRqst_PDU,
      { "MIB-SB-SIB-InformationList-SystemInfoUpdateRqst", "nbap.MIB_SB_SIB_InformationList_SystemInfoUpdateRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SegmentInformationListIE_SystemInfoUpdate_PDU,
      { "SegmentInformationListIE-SystemInfoUpdate", "nbap.SegmentInformationListIE_SystemInfoUpdate",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SystemInformationUpdateResponse_PDU,
      { "SystemInformationUpdateResponse", "nbap.SystemInformationUpdateResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SystemInformationUpdateFailure_PDU,
      { "SystemInformationUpdateFailure", "nbap.SystemInformationUpdateFailure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkSetupRequestFDD_PDU,
      { "RadioLinkSetupRequestFDD", "nbap.RadioLinkSetupRequestFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_HS_Cell_Information_RL_Setup_List_PDU,
      { "Additional-HS-Cell-Information-RL-Setup-List", "nbap.Additional_HS_Cell_Information_RL_Setup_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_DPCH_Information_RL_SetupRqstFDD_PDU,
      { "UL-DPCH-Information-RL-SetupRqstFDD", "nbap.UL_DPCH_Information_RL_SetupRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_DPCH_Information_RL_SetupRqstFDD_PDU,
      { "DL-DPCH-Information-RL-SetupRqstFDD", "nbap.DL_DPCH_Information_RL_SetupRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationList_RL_SetupRqstFDD_PDU,
      { "RL-InformationList-RL-SetupRqstFDD", "nbap.RL_InformationList_RL_SetupRqstFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationItem_RL_SetupRqstFDD_PDU,
      { "RL-InformationItem-RL-SetupRqstFDD", "nbap.RL_InformationItem_RL_SetupRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DPCH_Information_RL_SetupRqstFDD_PDU,
      { "E-DPCH-Information-RL-SetupRqstFDD", "nbap.E_DPCH_Information_RL_SetupRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_F_DPCH_Information_RL_SetupRqstFDD_PDU,
      { "F-DPCH-Information-RL-SetupRqstFDD", "nbap.F_DPCH_Information_RL_SetupRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkSetupRequestTDD_PDU,
      { "RadioLinkSetupRequestTDD", "nbap.RadioLinkSetupRequestTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_CCTrCH_InformationList_RL_SetupRqstTDD_PDU,
      { "UL-CCTrCH-InformationList-RL-SetupRqstTDD", "nbap.UL_CCTrCH_InformationList_RL_SetupRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_CCTrCH_InformationItem_RL_SetupRqstTDD_PDU,
      { "UL-CCTrCH-InformationItem-RL-SetupRqstTDD", "nbap.UL_CCTrCH_InformationItem_RL_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_DPCH_InformationItem_RL_SetupRqstTDD_PDU,
      { "UL-DPCH-InformationItem-RL-SetupRqstTDD", "nbap.UL_DPCH_InformationItem_RL_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_DPCH_LCR_Information_RL_SetupRqstTDD_PDU,
      { "UL-DPCH-LCR-Information-RL-SetupRqstTDD", "nbap.UL_DPCH_LCR_Information_RL_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_DPCH_768_Information_RL_SetupRqstTDD_PDU,
      { "UL-DPCH-768-Information-RL-SetupRqstTDD", "nbap.UL_DPCH_768_Information_RL_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_CCTrCH_InformationList_RL_SetupRqstTDD_PDU,
      { "DL-CCTrCH-InformationList-RL-SetupRqstTDD", "nbap.DL_CCTrCH_InformationList_RL_SetupRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_CCTrCH_InformationItem_RL_SetupRqstTDD_PDU,
      { "DL-CCTrCH-InformationItem-RL-SetupRqstTDD", "nbap.DL_CCTrCH_InformationItem_RL_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_DPCH_InformationItem_RL_SetupRqstTDD_PDU,
      { "DL-DPCH-InformationItem-RL-SetupRqstTDD", "nbap.DL_DPCH_InformationItem_RL_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_DPCH_LCR_Information_RL_SetupRqstTDD_PDU,
      { "DL-DPCH-LCR-Information-RL-SetupRqstTDD", "nbap.DL_DPCH_LCR_Information_RL_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_DPCH_768_Information_RL_SetupRqstTDD_PDU,
      { "DL-DPCH-768-Information-RL-SetupRqstTDD", "nbap.DL_DPCH_768_Information_RL_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_Information_RL_SetupRqstTDD_PDU,
      { "RL-Information-RL-SetupRqstTDD", "nbap.RL_Information_RL_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkSetupResponseFDD_PDU,
      { "RadioLinkSetupResponseFDD", "nbap.RadioLinkSetupResponseFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_HS_Cell_Information_Response_List_PDU,
      { "Additional-HS-Cell-Information-Response-List", "nbap.Additional_HS_Cell_Information_Response_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationResponseList_RL_SetupRspFDD_PDU,
      { "RL-InformationResponseList-RL-SetupRspFDD", "nbap.RL_InformationResponseList_RL_SetupRspFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationResponseItem_RL_SetupRspFDD_PDU,
      { "RL-InformationResponseItem-RL-SetupRspFDD", "nbap.RL_InformationResponseItem_RL_SetupRspFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkSetupResponseTDD_PDU,
      { "RadioLinkSetupResponseTDD", "nbap.RadioLinkSetupResponseTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationResponse_RL_SetupRspTDD_PDU,
      { "RL-InformationResponse-RL-SetupRspTDD", "nbap.RL_InformationResponse_RL_SetupRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationResponse_LCR_RL_SetupRspTDD_PDU,
      { "RL-InformationResponse-LCR-RL-SetupRspTDD", "nbap.RL_InformationResponse_LCR_RL_SetupRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkSetupFailureFDD_PDU,
      { "RadioLinkSetupFailureFDD", "nbap.RadioLinkSetupFailureFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CauseLevel_RL_SetupFailureFDD_PDU,
      { "CauseLevel-RL-SetupFailureFDD", "nbap.CauseLevel_RL_SetupFailureFDD",
        FT_UINT32, BASE_DEC, VALS(nbap_CauseLevel_RL_SetupFailureFDD_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Unsuccessful_RL_InformationRespItem_RL_SetupFailureFDD_PDU,
      { "Unsuccessful-RL-InformationRespItem-RL-SetupFailureFDD", "nbap.Unsuccessful_RL_InformationRespItem_RL_SetupFailureFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Successful_RL_InformationRespItem_RL_SetupFailureFDD_PDU,
      { "Successful-RL-InformationRespItem-RL-SetupFailureFDD", "nbap.Successful_RL_InformationRespItem_RL_SetupFailureFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkSetupFailureTDD_PDU,
      { "RadioLinkSetupFailureTDD", "nbap.RadioLinkSetupFailureTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CauseLevel_RL_SetupFailureTDD_PDU,
      { "CauseLevel-RL-SetupFailureTDD", "nbap.CauseLevel_RL_SetupFailureTDD",
        FT_UINT32, BASE_DEC, VALS(nbap_CauseLevel_RL_SetupFailureTDD_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Unsuccessful_RL_InformationResp_RL_SetupFailureTDD_PDU,
      { "Unsuccessful-RL-InformationResp-RL-SetupFailureTDD", "nbap.Unsuccessful_RL_InformationResp_RL_SetupFailureTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkAdditionRequestFDD_PDU,
      { "RadioLinkAdditionRequestFDD", "nbap.RadioLinkAdditionRequestFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_HS_Cell_Information_RL_Addition_List_PDU,
      { "Additional-HS-Cell-Information-RL-Addition-List", "nbap.Additional_HS_Cell_Information_RL_Addition_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_EDCH_Cell_Information_RL_Add_Req_PDU,
      { "Additional-EDCH-Cell-Information-RL-Add-Req", "nbap.Additional_EDCH_Cell_Information_RL_Add_Req_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationList_RL_AdditionRqstFDD_PDU,
      { "RL-InformationList-RL-AdditionRqstFDD", "nbap.RL_InformationList_RL_AdditionRqstFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationItem_RL_AdditionRqstFDD_PDU,
      { "RL-InformationItem-RL-AdditionRqstFDD", "nbap.RL_InformationItem_RL_AdditionRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DPCH_Information_RL_AdditionReqFDD_PDU,
      { "E-DPCH-Information-RL-AdditionReqFDD", "nbap.E_DPCH_Information_RL_AdditionReqFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkAdditionRequestTDD_PDU,
      { "RadioLinkAdditionRequestTDD", "nbap.RadioLinkAdditionRequestTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_CCTrCH_InformationList_RL_AdditionRqstTDD_PDU,
      { "UL-CCTrCH-InformationList-RL-AdditionRqstTDD", "nbap.UL_CCTrCH_InformationList_RL_AdditionRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_DPCH_InformationItem_RL_AdditionRqstTDD_PDU,
      { "UL-DPCH-InformationItem-RL-AdditionRqstTDD", "nbap.UL_DPCH_InformationItem_RL_AdditionRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD_PDU,
      { "UL-DPCH-InformationItem-LCR-RL-AdditionRqstTDD", "nbap.UL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_DPCH_InformationItem_768_RL_AdditionRqstTDD_PDU,
      { "UL-DPCH-InformationItem-768-RL-AdditionRqstTDD", "nbap.UL_DPCH_InformationItem_768_RL_AdditionRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_CCTrCH_InformationList_RL_AdditionRqstTDD_PDU,
      { "DL-CCTrCH-InformationList-RL-AdditionRqstTDD", "nbap.DL_CCTrCH_InformationList_RL_AdditionRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_DPCH_InformationItem_RL_AdditionRqstTDD_PDU,
      { "DL-DPCH-InformationItem-RL-AdditionRqstTDD", "nbap.DL_DPCH_InformationItem_RL_AdditionRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD_PDU,
      { "DL-DPCH-InformationItem-LCR-RL-AdditionRqstTDD", "nbap.DL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_DPCH_InformationItem_768_RL_AdditionRqstTDD_PDU,
      { "DL-DPCH-InformationItem-768-RL-AdditionRqstTDD", "nbap.DL_DPCH_InformationItem_768_RL_AdditionRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_Information_RL_AdditionRqstTDD_PDU,
      { "RL-Information-RL-AdditionRqstTDD", "nbap.RL_Information_RL_AdditionRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkAdditionResponseFDD_PDU,
      { "RadioLinkAdditionResponseFDD", "nbap.RadioLinkAdditionResponseFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_HS_Cell_Change_Information_Response_List_PDU,
      { "Additional-HS-Cell-Change-Information-Response-List", "nbap.Additional_HS_Cell_Change_Information_Response_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationResponseList_RL_AdditionRspFDD_PDU,
      { "RL-InformationResponseList-RL-AdditionRspFDD", "nbap.RL_InformationResponseList_RL_AdditionRspFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationResponseItem_RL_AdditionRspFDD_PDU,
      { "RL-InformationResponseItem-RL-AdditionRspFDD", "nbap.RL_InformationResponseItem_RL_AdditionRspFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkAdditionResponseTDD_PDU,
      { "RadioLinkAdditionResponseTDD", "nbap.RadioLinkAdditionResponseTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationResponse_RL_AdditionRspTDD_PDU,
      { "RL-InformationResponse-RL-AdditionRspTDD", "nbap.RL_InformationResponse_RL_AdditionRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationResponse_LCR_RL_AdditionRspTDD_PDU,
      { "RL-InformationResponse-LCR-RL-AdditionRspTDD", "nbap.RL_InformationResponse_LCR_RL_AdditionRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkAdditionFailureFDD_PDU,
      { "RadioLinkAdditionFailureFDD", "nbap.RadioLinkAdditionFailureFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CauseLevel_RL_AdditionFailureFDD_PDU,
      { "CauseLevel-RL-AdditionFailureFDD", "nbap.CauseLevel_RL_AdditionFailureFDD",
        FT_UINT32, BASE_DEC, VALS(nbap_CauseLevel_RL_AdditionFailureFDD_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Unsuccessful_RL_InformationRespItem_RL_AdditionFailureFDD_PDU,
      { "Unsuccessful-RL-InformationRespItem-RL-AdditionFailureFDD", "nbap.Unsuccessful_RL_InformationRespItem_RL_AdditionFailureFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Successful_RL_InformationRespItem_RL_AdditionFailureFDD_PDU,
      { "Successful-RL-InformationRespItem-RL-AdditionFailureFDD", "nbap.Successful_RL_InformationRespItem_RL_AdditionFailureFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkAdditionFailureTDD_PDU,
      { "RadioLinkAdditionFailureTDD", "nbap.RadioLinkAdditionFailureTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CauseLevel_RL_AdditionFailureTDD_PDU,
      { "CauseLevel-RL-AdditionFailureTDD", "nbap.CauseLevel_RL_AdditionFailureTDD",
        FT_UINT32, BASE_DEC, VALS(nbap_CauseLevel_RL_AdditionFailureTDD_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Unsuccessful_RL_InformationResp_RL_AdditionFailureTDD_PDU,
      { "Unsuccessful-RL-InformationResp-RL-AdditionFailureTDD", "nbap.Unsuccessful_RL_InformationResp_RL_AdditionFailureTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkReconfigurationPrepareFDD_PDU,
      { "RadioLinkReconfigurationPrepareFDD", "nbap.RadioLinkReconfigurationPrepareFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_HS_Cell_Information_RL_Reconf_Prep_PDU,
      { "Additional-HS-Cell-Information-RL-Reconf-Prep", "nbap.Additional_HS_Cell_Information_RL_Reconf_Prep",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_EDCH_Cell_Information_RL_Reconf_Prep_PDU,
      { "Additional-EDCH-Cell-Information-RL-Reconf-Prep", "nbap.Additional_EDCH_Cell_Information_RL_Reconf_Prep_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_DPCH_Information_RL_ReconfPrepFDD_PDU,
      { "UL-DPCH-Information-RL-ReconfPrepFDD", "nbap.UL_DPCH_Information_RL_ReconfPrepFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_DPCH_Information_RL_ReconfPrepFDD_PDU,
      { "DL-DPCH-Information-RL-ReconfPrepFDD", "nbap.DL_DPCH_Information_RL_ReconfPrepFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_DPCH_Power_Information_RL_ReconfPrepFDD_PDU,
      { "DL-DPCH-Power-Information-RL-ReconfPrepFDD", "nbap.DL_DPCH_Power_Information_RL_ReconfPrepFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DCH_DeleteList_RL_ReconfPrepFDD_PDU,
      { "DCH-DeleteList-RL-ReconfPrepFDD", "nbap.DCH_DeleteList_RL_ReconfPrepFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationList_RL_ReconfPrepFDD_PDU,
      { "RL-InformationList-RL-ReconfPrepFDD", "nbap.RL_InformationList_RL_ReconfPrepFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationItem_RL_ReconfPrepFDD_PDU,
      { "RL-InformationItem-RL-ReconfPrepFDD", "nbap.RL_InformationItem_RL_ReconfPrepFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DPCH_Information_RL_ReconfPrepFDD_PDU,
      { "E-DPCH-Information-RL-ReconfPrepFDD", "nbap.E_DPCH_Information_RL_ReconfPrepFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_F_DPCH_Information_RL_ReconfPrepFDD_PDU,
      { "F-DPCH-Information-RL-ReconfPrepFDD", "nbap.F_DPCH_Information_RL_ReconfPrepFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkReconfigurationPrepareTDD_PDU,
      { "RadioLinkReconfigurationPrepareTDD", "nbap.RadioLinkReconfigurationPrepareTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_PDU,
      { "UL-CCTrCH-InformationAddList-RL-ReconfPrepTDD", "nbap.UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_DPCH_InformationAddItem_RL_ReconfPrepTDD_PDU,
      { "UL-DPCH-InformationAddItem-RL-ReconfPrepTDD", "nbap.UL_DPCH_InformationAddItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD_PDU,
      { "UL-DPCH-LCR-InformationAddList-RL-ReconfPrepTDD", "nbap.UL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleRL_UL_DPCH_InformationAddList_RL_ReconfPrepTDD_PDU,
      { "MultipleRL-UL-DPCH-InformationAddList-RL-ReconfPrepTDD", "nbap.MultipleRL_UL_DPCH_InformationAddList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_DPCH_768_InformationAddList_RL_ReconfPrepTDD_PDU,
      { "UL-DPCH-768-InformationAddList-RL-ReconfPrepTDD", "nbap.UL_DPCH_768_InformationAddList_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_PDU,
      { "UL-CCTrCH-InformationModifyList-RL-ReconfPrepTDD", "nbap.UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD_PDU,
      { "UL-DPCH-InformationModify-AddItem-RL-ReconfPrepTDD", "nbap.UL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD_PDU,
      { "UL-DPCH-InformationModify-ModifyItem-RL-ReconfPrepTDD", "nbap.UL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_TimeslotLCR_InformationModify_ModifyList_RL_ReconfPrepTDD_PDU,
      { "UL-TimeslotLCR-InformationModify-ModifyList-RL-ReconfPrepTDD", "nbap.UL_TimeslotLCR_InformationModify_ModifyList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_Timeslot768_InformationModify_ModifyList_RL_ReconfPrepTDD_PDU,
      { "UL-Timeslot768-InformationModify-ModifyList-RL-ReconfPrepTDD", "nbap.UL_Timeslot768_InformationModify_ModifyList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD_PDU,
      { "UL-DPCH-InformationModify-DeleteListIE-RL-ReconfPrepTDD", "nbap.UL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD_PDU,
      { "UL-DPCH-LCR-InformationModify-AddList-RL-ReconfPrepTDD", "nbap.UL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleRL_UL_DPCH_InformationModifyList_RL_ReconfPrepTDD_PDU,
      { "MultipleRL-UL-DPCH-InformationModifyList-RL-ReconfPrepTDD", "nbap.MultipleRL_UL_DPCH_InformationModifyList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD_PDU,
      { "UL-DPCH-768-InformationModify-AddList-RL-ReconfPrepTDD", "nbap.UL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_PDU,
      { "UL-CCTrCH-InformationDeleteList-RL-ReconfPrepTDD", "nbap.UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_PDU,
      { "DL-CCTrCH-InformationAddList-RL-ReconfPrepTDD", "nbap.DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_DPCH_InformationAddItem_RL_ReconfPrepTDD_PDU,
      { "DL-DPCH-InformationAddItem-RL-ReconfPrepTDD", "nbap.DL_DPCH_InformationAddItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD_PDU,
      { "DL-DPCH-LCR-InformationAddList-RL-ReconfPrepTDD", "nbap.DL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleRL_DL_DPCH_InformationAddList_RL_ReconfPrepTDD_PDU,
      { "MultipleRL-DL-DPCH-InformationAddList-RL-ReconfPrepTDD", "nbap.MultipleRL_DL_DPCH_InformationAddList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_DPCH_768_InformationAddList_RL_ReconfPrepTDD_PDU,
      { "DL-DPCH-768-InformationAddList-RL-ReconfPrepTDD", "nbap.DL_DPCH_768_InformationAddList_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_PDU,
      { "DL-CCTrCH-InformationModifyList-RL-ReconfPrepTDD", "nbap.DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD_PDU,
      { "DL-DPCH-InformationModify-AddItem-RL-ReconfPrepTDD", "nbap.DL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD_PDU,
      { "DL-DPCH-InformationModify-ModifyItem-RL-ReconfPrepTDD", "nbap.DL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD_PDU,
      { "DL-Timeslot-LCR-InformationModify-ModifyList-RL-ReconfPrepTDD", "nbap.DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_Timeslot_768_InformationModify_ModifyList_RL_ReconfPrepTDD_PDU,
      { "DL-Timeslot-768-InformationModify-ModifyList-RL-ReconfPrepTDD", "nbap.DL_Timeslot_768_InformationModify_ModifyList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD_PDU,
      { "DL-DPCH-InformationModify-DeleteListIE-RL-ReconfPrepTDD", "nbap.DL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD_PDU,
      { "DL-DPCH-LCR-InformationModify-AddList-RL-ReconfPrepTDD", "nbap.DL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleRL_DL_DPCH_InformationModifyList_RL_ReconfPrepTDD_PDU,
      { "MultipleRL-DL-DPCH-InformationModifyList-RL-ReconfPrepTDD", "nbap.MultipleRL_DL_DPCH_InformationModifyList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD_PDU,
      { "DL-DPCH-768-InformationModify-AddList-RL-ReconfPrepTDD", "nbap.DL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_PDU,
      { "DL-CCTrCH-InformationDeleteList-RL-ReconfPrepTDD", "nbap.DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DCH_DeleteList_RL_ReconfPrepTDD_PDU,
      { "DCH-DeleteList-RL-ReconfPrepTDD", "nbap.DCH_DeleteList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DSCH_Information_ModifyList_RL_ReconfPrepTDD_PDU,
      { "DSCH-Information-ModifyList-RL-ReconfPrepTDD", "nbap.DSCH_Information_ModifyList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DSCH_Information_DeleteList_RL_ReconfPrepTDD_PDU,
      { "DSCH-Information-DeleteList-RL-ReconfPrepTDD", "nbap.DSCH_Information_DeleteList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_USCH_Information_ModifyList_RL_ReconfPrepTDD_PDU,
      { "USCH-Information-ModifyList-RL-ReconfPrepTDD", "nbap.USCH_Information_ModifyList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_USCH_Information_DeleteList_RL_ReconfPrepTDD_PDU,
      { "USCH-Information-DeleteList-RL-ReconfPrepTDD", "nbap.USCH_Information_DeleteList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleRL_Information_RL_ReconfPrepTDD_PDU,
      { "MultipleRL-Information-RL-ReconfPrepTDD", "nbap.MultipleRL_Information_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_Information_RL_ReconfPrepTDD_PDU,
      { "RL-Information-RL-ReconfPrepTDD", "nbap.RL_Information_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkReconfigurationReady_PDU,
      { "RadioLinkReconfigurationReady", "nbap.RadioLinkReconfigurationReady_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationResponseList_RL_ReconfReady_PDU,
      { "RL-InformationResponseList-RL-ReconfReady", "nbap.RL_InformationResponseList_RL_ReconfReady",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationResponseItem_RL_ReconfReady_PDU,
      { "RL-InformationResponseItem-RL-ReconfReady", "nbap.RL_InformationResponseItem_RL_ReconfReady_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkReconfigurationFailure_PDU,
      { "RadioLinkReconfigurationFailure", "nbap.RadioLinkReconfigurationFailure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CauseLevel_RL_ReconfFailure_PDU,
      { "CauseLevel-RL-ReconfFailure", "nbap.CauseLevel_RL_ReconfFailure",
        FT_UINT32, BASE_DEC, VALS(nbap_CauseLevel_RL_ReconfFailure_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_RL_ReconfigurationFailureItem_RL_ReconfFailure_PDU,
      { "RL-ReconfigurationFailureItem-RL-ReconfFailure", "nbap.RL_ReconfigurationFailureItem_RL_ReconfFailure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkReconfigurationCommit_PDU,
      { "RadioLinkReconfigurationCommit", "nbap.RadioLinkReconfigurationCommit_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkReconfigurationCancel_PDU,
      { "RadioLinkReconfigurationCancel", "nbap.RadioLinkReconfigurationCancel_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkReconfigurationRequestFDD_PDU,
      { "RadioLinkReconfigurationRequestFDD", "nbap.RadioLinkReconfigurationRequestFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_HS_Cell_Information_RL_Reconf_Req_PDU,
      { "Additional-HS-Cell-Information-RL-Reconf-Req", "nbap.Additional_HS_Cell_Information_RL_Reconf_Req",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_EDCH_Cell_Information_RL_Reconf_Req_PDU,
      { "Additional-EDCH-Cell-Information-RL-Reconf-Req", "nbap.Additional_EDCH_Cell_Information_RL_Reconf_Req_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_DPCH_Information_RL_ReconfRqstFDD_PDU,
      { "UL-DPCH-Information-RL-ReconfRqstFDD", "nbap.UL_DPCH_Information_RL_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_DPCH_Information_RL_ReconfRqstFDD_PDU,
      { "DL-DPCH-Information-RL-ReconfRqstFDD", "nbap.DL_DPCH_Information_RL_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DCH_DeleteList_RL_ReconfRqstFDD_PDU,
      { "DCH-DeleteList-RL-ReconfRqstFDD", "nbap.DCH_DeleteList_RL_ReconfRqstFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationList_RL_ReconfRqstFDD_PDU,
      { "RL-InformationList-RL-ReconfRqstFDD", "nbap.RL_InformationList_RL_ReconfRqstFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationItem_RL_ReconfRqstFDD_PDU,
      { "RL-InformationItem-RL-ReconfRqstFDD", "nbap.RL_InformationItem_RL_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DPCH_Information_RL_ReconfRqstFDD_PDU,
      { "E-DPCH-Information-RL-ReconfRqstFDD", "nbap.E_DPCH_Information_RL_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkReconfigurationRequestTDD_PDU,
      { "RadioLinkReconfigurationRequestTDD", "nbap.RadioLinkReconfigurationRequestTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU,
      { "UL-CCTrCH-InformationModifyList-RL-ReconfRqstTDD", "nbap.UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_PDU,
      { "UL-CCTrCH-InformationModifyItem-RL-ReconfRqstTDD", "nbap.UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_PDU,
      { "UL-CCTrCH-InformationDeleteList-RL-ReconfRqstTDD", "nbap.UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_PDU,
      { "UL-CCTrCH-InformationDeleteItem-RL-ReconfRqstTDD", "nbap.UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU,
      { "DL-CCTrCH-InformationModifyList-RL-ReconfRqstTDD", "nbap.DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_PDU,
      { "DL-CCTrCH-InformationModifyItem-RL-ReconfRqstTDD", "nbap.DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleRL_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU,
      { "MultipleRL-DL-CCTrCH-InformationModifyList-RL-ReconfRqstTDD", "nbap.MultipleRL_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_DPCH_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD_PDU,
      { "DL-DPCH-LCR-InformationModify-ModifyList-RL-ReconfRqstTDD", "nbap.DL_DPCH_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_PDU,
      { "DL-CCTrCH-InformationDeleteList-RL-ReconfRqstTDD", "nbap.DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_PDU,
      { "DL-CCTrCH-InformationDeleteItem-RL-ReconfRqstTDD", "nbap.DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DCH_DeleteList_RL_ReconfRqstTDD_PDU,
      { "DCH-DeleteList-RL-ReconfRqstTDD", "nbap.DCH_DeleteList_RL_ReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Multiple_RL_Information_RL_ReconfRqstTDD_PDU,
      { "Multiple-RL-Information-RL-ReconfRqstTDD", "nbap.Multiple_RL_Information_RL_ReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_Information_RL_ReconfRqstTDD_PDU,
      { "RL-Information-RL-ReconfRqstTDD", "nbap.RL_Information_RL_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkReconfigurationResponse_PDU,
      { "RadioLinkReconfigurationResponse", "nbap.RadioLinkReconfigurationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationResponseList_RL_ReconfRsp_PDU,
      { "RL-InformationResponseList-RL-ReconfRsp", "nbap.RL_InformationResponseList_RL_ReconfRsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationResponseItem_RL_ReconfRsp_PDU,
      { "RL-InformationResponseItem-RL-ReconfRsp", "nbap.RL_InformationResponseItem_RL_ReconfRsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkDeletionRequest_PDU,
      { "RadioLinkDeletionRequest", "nbap.RadioLinkDeletionRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_informationList_RL_DeletionRqst_PDU,
      { "RL-informationList-RL-DeletionRqst", "nbap.RL_informationList_RL_DeletionRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_informationItem_RL_DeletionRqst_PDU,
      { "RL-informationItem-RL-DeletionRqst", "nbap.RL_informationItem_RL_DeletionRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkDeletionResponse_PDU,
      { "RadioLinkDeletionResponse", "nbap.RadioLinkDeletionResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_PowerControlRequest_PDU,
      { "DL-PowerControlRequest", "nbap.DL_PowerControlRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_ReferencePowerInformationList_DL_PC_Rqst_PDU,
      { "DL-ReferencePowerInformationList-DL-PC-Rqst", "nbap.DL_ReferencePowerInformationList_DL_PC_Rqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_ReferencePowerInformationItem_DL_PC_Rqst_PDU,
      { "DL-ReferencePowerInformationItem-DL-PC-Rqst", "nbap.DL_ReferencePowerInformationItem_DL_PC_Rqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_PowerTimeslotControlRequest_PDU,
      { "DL-PowerTimeslotControlRequest", "nbap.DL_PowerTimeslotControlRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DedicatedMeasurementInitiationRequest_PDU,
      { "DedicatedMeasurementInitiationRequest", "nbap.DedicatedMeasurementInitiationRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DedicatedMeasurementObjectType_DM_Rqst_PDU,
      { "DedicatedMeasurementObjectType-DM-Rqst", "nbap.DedicatedMeasurementObjectType_DM_Rqst",
        FT_UINT32, BASE_DEC, VALS(nbap_DedicatedMeasurementObjectType_DM_Rqst_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationItem_DM_Rqst_PDU,
      { "RL-InformationItem-DM-Rqst", "nbap.RL_InformationItem_DM_Rqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PUSCH_Info_DM_Rqst_PDU,
      { "PUSCH-Info-DM-Rqst", "nbap.PUSCH_Info_DM_Rqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSSICH_Info_DM_Rqst_PDU,
      { "HSSICH-Info-DM-Rqst", "nbap.HSSICH_Info_DM_Rqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSSICH_InfoExt_DM_Rqst_PDU,
      { "HSSICH-InfoExt-DM-Rqst", "nbap.HSSICH_InfoExt_DM_Rqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DedicatedMeasurementInitiationResponse_PDU,
      { "DedicatedMeasurementInitiationResponse", "nbap.DedicatedMeasurementInitiationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DedicatedMeasurementObjectType_DM_Rsp_PDU,
      { "DedicatedMeasurementObjectType-DM-Rsp", "nbap.DedicatedMeasurementObjectType_DM_Rsp",
        FT_UINT32, BASE_DEC, VALS(nbap_DedicatedMeasurementObjectType_DM_Rsp_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationItem_DM_Rsp_PDU,
      { "RL-InformationItem-DM-Rsp", "nbap.RL_InformationItem_DM_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PUSCH_Info_DM_Rsp_PDU,
      { "PUSCH-Info-DM-Rsp", "nbap.PUSCH_Info_DM_Rsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Multiple_PUSCH_InfoList_DM_Rsp_PDU,
      { "Multiple-PUSCH-InfoList-DM-Rsp", "nbap.Multiple_PUSCH_InfoList_DM_Rsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp_PDU,
      { "Multiple-DedicatedMeasurementValueList-TDD-DM-Rsp", "nbap.Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp_PDU,
      { "Multiple-DedicatedMeasurementValueList-LCR-TDD-DM-Rsp", "nbap.Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp_PDU,
      { "Multiple-HSSICHMeasurementValueList-TDD-DM-Rsp", "nbap.Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Multiple_DedicatedMeasurementValueList_768_TDD_DM_Rsp_PDU,
      { "Multiple-DedicatedMeasurementValueList-768-TDD-DM-Rsp", "nbap.Multiple_DedicatedMeasurementValueList_768_TDD_DM_Rsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_Set_InformationItem_DM_Rsp_PDU,
      { "RL-Set-InformationItem-DM-Rsp", "nbap.RL_Set_InformationItem_DM_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DedicatedMeasurementInitiationFailure_PDU,
      { "DedicatedMeasurementInitiationFailure", "nbap.DedicatedMeasurementInitiationFailure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DedicatedMeasurementReport_PDU,
      { "DedicatedMeasurementReport", "nbap.DedicatedMeasurementReport_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DedicatedMeasurementObjectType_DM_Rprt_PDU,
      { "DedicatedMeasurementObjectType-DM-Rprt", "nbap.DedicatedMeasurementObjectType_DM_Rprt",
        FT_UINT32, BASE_DEC, VALS(nbap_DedicatedMeasurementObjectType_DM_Rprt_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationItem_DM_Rprt_PDU,
      { "RL-InformationItem-DM-Rprt", "nbap.RL_InformationItem_DM_Rprt_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PUSCH_Info_DM_Rprt_PDU,
      { "PUSCH-Info-DM-Rprt", "nbap.PUSCH_Info_DM_Rprt",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Multiple_PUSCH_InfoList_DM_Rprt_PDU,
      { "Multiple-PUSCH-InfoList-DM-Rprt", "nbap.Multiple_PUSCH_InfoList_DM_Rprt",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_Set_InformationItem_DM_Rprt_PDU,
      { "RL-Set-InformationItem-DM-Rprt", "nbap.RL_Set_InformationItem_DM_Rprt_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DedicatedMeasurementTerminationRequest_PDU,
      { "DedicatedMeasurementTerminationRequest", "nbap.DedicatedMeasurementTerminationRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DedicatedMeasurementFailureIndication_PDU,
      { "DedicatedMeasurementFailureIndication", "nbap.DedicatedMeasurementFailureIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkFailureIndication_PDU,
      { "RadioLinkFailureIndication", "nbap.RadioLinkFailureIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Reporting_Object_RL_FailureInd_PDU,
      { "Reporting-Object-RL-FailureInd", "nbap.Reporting_Object_RL_FailureInd",
        FT_UINT32, BASE_DEC, VALS(nbap_Reporting_Object_RL_FailureInd_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationItem_RL_FailureInd_PDU,
      { "RL-InformationItem-RL-FailureInd", "nbap.RL_InformationItem_RL_FailureInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_Set_InformationItem_RL_FailureInd_PDU,
      { "RL-Set-InformationItem-RL-FailureInd", "nbap.RL_Set_InformationItem_RL_FailureInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CCTrCH_InformationItem_RL_FailureInd_PDU,
      { "CCTrCH-InformationItem-RL-FailureInd", "nbap.CCTrCH_InformationItem_RL_FailureInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkPreemptionRequiredIndication_PDU,
      { "RadioLinkPreemptionRequiredIndication", "nbap.RadioLinkPreemptionRequiredIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationList_RL_PreemptRequiredInd_PDU,
      { "RL-InformationList-RL-PreemptRequiredInd", "nbap.RL_InformationList_RL_PreemptRequiredInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationItem_RL_PreemptRequiredInd_PDU,
      { "RL-InformationItem-RL-PreemptRequiredInd", "nbap.RL_InformationItem_RL_PreemptRequiredInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkRestoreIndication_PDU,
      { "RadioLinkRestoreIndication", "nbap.RadioLinkRestoreIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Reporting_Object_RL_RestoreInd_PDU,
      { "Reporting-Object-RL-RestoreInd", "nbap.Reporting_Object_RL_RestoreInd",
        FT_UINT32, BASE_DEC, VALS(nbap_Reporting_Object_RL_RestoreInd_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationItem_RL_RestoreInd_PDU,
      { "RL-InformationItem-RL-RestoreInd", "nbap.RL_InformationItem_RL_RestoreInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_Set_InformationItem_RL_RestoreInd_PDU,
      { "RL-Set-InformationItem-RL-RestoreInd", "nbap.RL_Set_InformationItem_RL_RestoreInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CCTrCH_InformationItem_RL_RestoreInd_PDU,
      { "CCTrCH-InformationItem-RL-RestoreInd", "nbap.CCTrCH_InformationItem_RL_RestoreInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CompressedModeCommand_PDU,
      { "CompressedModeCommand", "nbap.CompressedModeCommand_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ErrorIndication_PDU,
      { "ErrorIndication", "nbap.ErrorIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PrivateMessage_PDU,
      { "PrivateMessage", "nbap.PrivateMessage_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PhysicalSharedChannelReconfigurationRequestFDD_PDU,
      { "PhysicalSharedChannelReconfigurationRequestFDD", "nbap.PhysicalSharedChannelReconfigurationRequestFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDPA_And_EDCH_CellPortion_InformationList_PSCH_ReconfRqst_PDU,
      { "HSDPA-And-EDCH-CellPortion-InformationList-PSCH-ReconfRqst", "nbap.HSDPA_And_EDCH_CellPortion_InformationList_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PhysicalSharedChannelReconfigurationRequestTDD_PDU,
      { "PhysicalSharedChannelReconfigurationRequestTDD", "nbap.PhysicalSharedChannelReconfigurationRequestTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PDSCHSets_AddList_PSCH_ReconfRqst_PDU,
      { "PDSCHSets-AddList-PSCH-ReconfRqst", "nbap.PDSCHSets_AddList_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PDSCH_Information_AddItem_PSCH_ReconfRqst_PDU,
      { "PDSCH-Information-AddItem-PSCH-ReconfRqst", "nbap.PDSCH_Information_AddItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PDSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst_PDU,
      { "PDSCH-AddInformation-LCR-AddItem-PSCH-ReconfRqst", "nbap.PDSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PDSCH_AddInformation_768_AddItem_PSCH_ReconfRqst_PDU,
      { "PDSCH-AddInformation-768-AddItem-PSCH-ReconfRqst", "nbap.PDSCH_AddInformation_768_AddItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PDSCHSets_ModifyList_PSCH_ReconfRqst_PDU,
      { "PDSCHSets-ModifyList-PSCH-ReconfRqst", "nbap.PDSCHSets_ModifyList_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PDSCH_Information_ModifyItem_PSCH_ReconfRqst_PDU,
      { "PDSCH-Information-ModifyItem-PSCH-ReconfRqst", "nbap.PDSCH_Information_ModifyItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PDSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst_PDU,
      { "PDSCH-ModifyInformation-LCR-ModifyItem-PSCH-ReconfRqst", "nbap.PDSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PDSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst_PDU,
      { "PDSCH-ModifyInformation-768-ModifyItem-PSCH-ReconfRqst", "nbap.PDSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PDSCHSets_DeleteList_PSCH_ReconfRqst_PDU,
      { "PDSCHSets-DeleteList-PSCH-ReconfRqst", "nbap.PDSCHSets_DeleteList_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PUSCHSets_AddList_PSCH_ReconfRqst_PDU,
      { "PUSCHSets-AddList-PSCH-ReconfRqst", "nbap.PUSCHSets_AddList_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PUSCH_Information_AddItem_PSCH_ReconfRqst_PDU,
      { "PUSCH-Information-AddItem-PSCH-ReconfRqst", "nbap.PUSCH_Information_AddItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PUSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst_PDU,
      { "PUSCH-AddInformation-LCR-AddItem-PSCH-ReconfRqst", "nbap.PUSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PUSCH_AddInformation_768_AddItem_PSCH_ReconfRqst_PDU,
      { "PUSCH-AddInformation-768-AddItem-PSCH-ReconfRqst", "nbap.PUSCH_AddInformation_768_AddItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PUSCHSets_ModifyList_PSCH_ReconfRqst_PDU,
      { "PUSCHSets-ModifyList-PSCH-ReconfRqst", "nbap.PUSCHSets_ModifyList_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PUSCH_Information_ModifyItem_PSCH_ReconfRqst_PDU,
      { "PUSCH-Information-ModifyItem-PSCH-ReconfRqst", "nbap.PUSCH_Information_ModifyItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PUSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst_PDU,
      { "PUSCH-ModifyInformation-LCR-ModifyItem-PSCH-ReconfRqst", "nbap.PUSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PUSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst_PDU,
      { "PUSCH-ModifyInformation-768-ModifyItem-PSCH-ReconfRqst", "nbap.PUSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PUSCHSets_DeleteList_PSCH_ReconfRqst_PDU,
      { "PUSCHSets-DeleteList-PSCH-ReconfRqst", "nbap.PUSCHSets_DeleteList_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_PDSCH_TDD_Information_PSCH_ReconfRqst_PDU,
      { "HS-PDSCH-TDD-Information-PSCH-ReconfRqst", "nbap.HS_PDSCH_TDD_Information_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_HS_PDSCH_Timeslot_Information_768_PSCH_ReconfRqst_PDU,
      { "DL-HS-PDSCH-Timeslot-Information-768-PSCH-ReconfRqst", "nbap.DL_HS_PDSCH_Timeslot_Information_768_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst_PDU,
      { "MultipleFreq-DL-HS-PDSCH-Timeslot-Information-LCR-PSCH-ReconfRqst", "nbap.MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst_PDU,
      { "MultipleFreq-DL-HS-PDSCH-Timeslot-Information-LCRItem-PSCH-ReconfRqst", "nbap.MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Add_To_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_PDU,
      { "Add-To-HS-SCCH-Resource-Pool-PSCH-ReconfRqst", "nbap.Add_To_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_SCCH_Information_768_PSCH_ReconfRqst_PDU,
      { "HS-SCCH-Information-768-PSCH-ReconfRqst", "nbap.HS_SCCH_Information_768_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_SCCH_InformationExt_LCR_PSCH_ReconfRqst_PDU,
      { "HS-SCCH-InformationExt-LCR-PSCH-ReconfRqst", "nbap.HS_SCCH_InformationExt_LCR_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Modify_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_PDU,
      { "Modify-HS-SCCH-Resource-Pool-PSCH-ReconfRqst", "nbap.Modify_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_SCCH_InformationModifyExt_LCR_PSCH_ReconfRqst_PDU,
      { "HS-SCCH-InformationModifyExt-LCR-PSCH-ReconfRqst", "nbap.HS_SCCH_InformationModifyExt_LCR_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_SCCH_InformationModify_768_PSCH_ReconfRqst_PDU,
      { "HS-SCCH-InformationModify-768-PSCH-ReconfRqst", "nbap.HS_SCCH_InformationModify_768_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Delete_From_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_PDU,
      { "Delete-From-HS-SCCH-Resource-Pool-PSCH-ReconfRqst", "nbap.Delete_From_HS_SCCH_Resource_Pool_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_PUCH_Information_PSCH_ReconfRqst_PDU,
      { "E-PUCH-Information-PSCH-ReconfRqst", "nbap.E_PUCH_Information_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Add_To_E_AGCH_Resource_Pool_PSCH_ReconfRqst_PDU,
      { "Add-To-E-AGCH-Resource-Pool-PSCH-ReconfRqst", "nbap.Add_To_E_AGCH_Resource_Pool_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Modify_E_AGCH_Resource_Pool_PSCH_ReconfRqst_PDU,
      { "Modify-E-AGCH-Resource-Pool-PSCH-ReconfRqst", "nbap.Modify_E_AGCH_Resource_Pool_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Delete_From_E_AGCH_Resource_Pool_PSCH_ReconfRqst_PDU,
      { "Delete-From-E-AGCH-Resource-Pool-PSCH-ReconfRqst", "nbap.Delete_From_E_AGCH_Resource_Pool_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_HICH_Information_PSCH_ReconfRqst_PDU,
      { "E-HICH-Information-PSCH-ReconfRqst", "nbap.E_HICH_Information_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_PUCH_Information_768_PSCH_ReconfRqst_PDU,
      { "E-PUCH-Information-768-PSCH-ReconfRqst", "nbap.E_PUCH_Information_768_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Add_To_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst_PDU,
      { "Add-To-E-AGCH-Resource-Pool-768-PSCH-ReconfRqst", "nbap.Add_To_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Modify_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst_PDU,
      { "Modify-E-AGCH-Resource-Pool-768-PSCH-ReconfRqst", "nbap.Modify_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_HICH_Information_768_PSCH_ReconfRqst_PDU,
      { "E-HICH-Information-768-PSCH-ReconfRqst", "nbap.E_HICH_Information_768_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_PUCH_Information_LCR_PSCH_ReconfRqst_PDU,
      { "E-PUCH-Information-LCR-PSCH-ReconfRqst", "nbap.E_PUCH_Information_LCR_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Add_To_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU,
      { "Add-To-E-AGCH-Resource-Pool-LCR-PSCH-ReconfRqst", "nbap.Add_To_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Modify_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU,
      { "Modify-E-AGCH-Resource-Pool-LCR-PSCH-ReconfRqst", "nbap.Modify_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Add_To_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU,
      { "Add-To-E-HICH-Resource-Pool-LCR-PSCH-ReconfRqst", "nbap.Add_To_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Modify_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU,
      { "Modify-E-HICH-Resource-Pool-LCR-PSCH-ReconfRqst", "nbap.Modify_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Delete_From_E_HICH_Resource_Pool_PSCH_ReconfRqst_PDU,
      { "Delete-From-E-HICH-Resource-Pool-PSCH-ReconfRqst", "nbap.Delete_From_E_HICH_Resource_Pool_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SYNC_UL_Partition_LCR_PDU,
      { "SYNC-UL-Partition-LCR", "nbap.SYNC_UL_Partition_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Delete_From_HS_SCCH_Resource_PoolExt_PSCH_ReconfRqst_PDU,
      { "Delete-From-HS-SCCH-Resource-PoolExt-PSCH-ReconfRqst", "nbap.Delete_From_HS_SCCH_Resource_PoolExt_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleFreq_E_PUCH_Timeslot_InformationList_LCR_PSCH_ReconfRqst_PDU,
      { "MultipleFreq-E-PUCH-Timeslot-InformationList-LCR-PSCH-ReconfRqst", "nbap.MultipleFreq_E_PUCH_Timeslot_InformationList_LCR_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleFreq_E_PUCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst_PDU,
      { "MultipleFreq-E-PUCH-Timeslot-Information-LCRItem-PSCH-ReconfRqst", "nbap.MultipleFreq_E_PUCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst_PDU,
      { "Max-RTWP-perUARFCN-Information-LCR-PSCH-ReconfRqst", "nbap.Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PhysicalSharedChannelReconfigurationResponse_PDU,
      { "PhysicalSharedChannelReconfigurationResponse", "nbap.PhysicalSharedChannelReconfigurationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_HICH_TimeOffset_ExtensionLCR_PDU,
      { "E-HICH-TimeOffset-ExtensionLCR", "nbap.E_HICH_TimeOffset_ExtensionLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleFreq_E_HICH_TimeOffsetLCR_PDU,
      { "MultipleFreq-E-HICH-TimeOffsetLCR", "nbap.MultipleFreq_E_HICH_TimeOffsetLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PhysicalSharedChannelReconfigurationFailure_PDU,
      { "PhysicalSharedChannelReconfigurationFailure", "nbap.PhysicalSharedChannelReconfigurationFailure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CauseLevel_PSCH_ReconfFailure_PDU,
      { "CauseLevel-PSCH-ReconfFailure", "nbap.CauseLevel_PSCH_ReconfFailure",
        FT_UINT32, BASE_DEC, VALS(nbap_CauseLevel_PSCH_ReconfFailure_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Unsuccessful_PDSCHSetItem_PSCH_ReconfFailureTDD_PDU,
      { "Unsuccessful-PDSCHSetItem-PSCH-ReconfFailureTDD", "nbap.Unsuccessful_PDSCHSetItem_PSCH_ReconfFailureTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Unsuccessful_PUSCHSetItem_PSCH_ReconfFailureTDD_PDU,
      { "Unsuccessful-PUSCHSetItem-PSCH-ReconfFailureTDD", "nbap.Unsuccessful_PUSCHSetItem_PSCH_ReconfFailureTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UARFCNSpecificCauseList_PSCH_ReconfFailureTDD_PDU,
      { "UARFCNSpecificCauseList-PSCH-ReconfFailureTDD", "nbap.UARFCNSpecificCauseList_PSCH_ReconfFailureTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Unsuccessful_UARFCNItem_PSCH_ReconfFailureTDD_PDU,
      { "Unsuccessful-UARFCNItem-PSCH-ReconfFailureTDD", "nbap.Unsuccessful_UARFCNItem_PSCH_ReconfFailureTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_HICH_TimeOffset_ReconfFailureTDD_PDU,
      { "E-HICH-TimeOffset-ReconfFailureTDD", "nbap.E_HICH_TimeOffset_ReconfFailureTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Common_System_Information_ResponseLCR_PDU,
      { "Common-System-Information-ResponseLCR", "nbap.Common_System_Information_ResponseLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ResetRequest_PDU,
      { "ResetRequest", "nbap.ResetRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ResetIndicator_PDU,
      { "ResetIndicator", "nbap.ResetIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_ResetIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_CommunicationContextInfoItem_Reset_PDU,
      { "CommunicationContextInfoItem-Reset", "nbap.CommunicationContextInfoItem_Reset_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommunicationControlPortInfoItem_Reset_PDU,
      { "CommunicationControlPortInfoItem-Reset", "nbap.CommunicationControlPortInfoItem_Reset_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ResetResponse_PDU,
      { "ResetResponse", "nbap.ResetResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_InformationExchangeInitiationRequest_PDU,
      { "InformationExchangeInitiationRequest", "nbap.InformationExchangeInitiationRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_InformationExchangeObjectType_InfEx_Rqst_PDU,
      { "InformationExchangeObjectType-InfEx-Rqst", "nbap.InformationExchangeObjectType_InfEx_Rqst",
        FT_UINT32, BASE_DEC, VALS(nbap_InformationExchangeObjectType_InfEx_Rqst_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_InformationExchangeInitiationResponse_PDU,
      { "InformationExchangeInitiationResponse", "nbap.InformationExchangeInitiationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_InformationExchangeObjectType_InfEx_Rsp_PDU,
      { "InformationExchangeObjectType-InfEx-Rsp", "nbap.InformationExchangeObjectType_InfEx_Rsp",
        FT_UINT32, BASE_DEC, VALS(nbap_InformationExchangeObjectType_InfEx_Rsp_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_InformationExchangeInitiationFailure_PDU,
      { "InformationExchangeInitiationFailure", "nbap.InformationExchangeInitiationFailure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_InformationReport_PDU,
      { "InformationReport", "nbap.InformationReport_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_InformationExchangeObjectType_InfEx_Rprt_PDU,
      { "InformationExchangeObjectType-InfEx-Rprt", "nbap.InformationExchangeObjectType_InfEx_Rprt",
        FT_UINT32, BASE_DEC, VALS(nbap_InformationExchangeObjectType_InfEx_Rprt_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_InformationExchangeTerminationRequest_PDU,
      { "InformationExchangeTerminationRequest", "nbap.InformationExchangeTerminationRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_InformationExchangeFailureIndication_PDU,
      { "InformationExchangeFailureIndication", "nbap.InformationExchangeFailureIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSynchronisationInitiationRequestTDD_PDU,
      { "CellSynchronisationInitiationRequestTDD", "nbap.CellSynchronisationInitiationRequestTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TimeslotInfo_CellSyncInitiationRqstTDD_PDU,
      { "TimeslotInfo-CellSyncInitiationRqstTDD", "nbap.TimeslotInfo_CellSyncInitiationRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSyncBurstTransInit_CellSyncInitiationRqstTDD_PDU,
      { "CellSyncBurstTransInit-CellSyncInitiationRqstTDD", "nbap.CellSyncBurstTransInit_CellSyncInitiationRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSyncBurstMeasureInit_CellSyncInitiationRqstTDD_PDU,
      { "CellSyncBurstMeasureInit-CellSyncInitiationRqstTDD", "nbap.CellSyncBurstMeasureInit_CellSyncInitiationRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SYNCDlCodeId_TransInitLCR_CellSyncInitiationRqstTDD_PDU,
      { "SYNCDlCodeId-TransInitLCR-CellSyncInitiationRqstTDD", "nbap.SYNCDlCodeId_TransInitLCR_CellSyncInitiationRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SYNCDlCodeId_MeasureInitLCR_CellSyncInitiationRqstTDD_PDU,
      { "SYNCDlCodeId-MeasureInitLCR-CellSyncInitiationRqstTDD", "nbap.SYNCDlCodeId_MeasureInitLCR_CellSyncInitiationRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSynchronisationInitiationResponseTDD_PDU,
      { "CellSynchronisationInitiationResponseTDD", "nbap.CellSynchronisationInitiationResponseTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSynchronisationInitiationFailureTDD_PDU,
      { "CellSynchronisationInitiationFailureTDD", "nbap.CellSynchronisationInitiationFailureTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSynchronisationReconfigurationRequestTDD_PDU,
      { "CellSynchronisationReconfigurationRequestTDD", "nbap.CellSynchronisationReconfigurationRequestTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSyncBurstTransReconfInfo_CellSyncReconfRqstTDD_PDU,
      { "CellSyncBurstTransReconfInfo-CellSyncReconfRqstTDD", "nbap.CellSyncBurstTransReconfInfo_CellSyncReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSyncBurstMeasInfo_CellSyncReconfRqstTDD_PDU,
      { "CellSyncBurstMeasInfo-CellSyncReconfRqstTDD", "nbap.CellSyncBurstMeasInfo_CellSyncReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSyncBurstMeasInfoListIE_CellSyncReconfRqstTDD_PDU,
      { "CellSyncBurstMeasInfoListIE-CellSyncReconfRqstTDD", "nbap.CellSyncBurstMeasInfoListIE_CellSyncReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SYNCDlCodeIdTransReconfInfoLCR_CellSyncReconfRqstTDD_PDU,
      { "SYNCDlCodeIdTransReconfInfoLCR-CellSyncReconfRqstTDD", "nbap.SYNCDlCodeIdTransReconfInfoLCR_CellSyncReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SYNCDlCodeIdMeasInfoLCR_CellSyncReconfRqstTDD_PDU,
      { "SYNCDlCodeIdMeasInfoLCR-CellSyncReconfRqstTDD", "nbap.SYNCDlCodeIdMeasInfoLCR_CellSyncReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSynchronisationReconfigurationResponseTDD_PDU,
      { "CellSynchronisationReconfigurationResponseTDD", "nbap.CellSynchronisationReconfigurationResponseTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSynchronisationReconfigurationFailureTDD_PDU,
      { "CellSynchronisationReconfigurationFailureTDD", "nbap.CellSynchronisationReconfigurationFailureTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSynchronisationAdjustmentRequestTDD_PDU,
      { "CellSynchronisationAdjustmentRequestTDD", "nbap.CellSynchronisationAdjustmentRequestTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellAdjustmentInfo_SyncAdjustmentRqstTDD_PDU,
      { "CellAdjustmentInfo-SyncAdjustmentRqstTDD", "nbap.CellAdjustmentInfo_SyncAdjustmentRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellAdjustmentInfoItem_SyncAdjustmentRqstTDD_PDU,
      { "CellAdjustmentInfoItem-SyncAdjustmentRqstTDD", "nbap.CellAdjustmentInfoItem_SyncAdjustmentRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSynchronisationAdjustmentResponseTDD_PDU,
      { "CellSynchronisationAdjustmentResponseTDD", "nbap.CellSynchronisationAdjustmentResponseTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSynchronisationAdjustmentFailureTDD_PDU,
      { "CellSynchronisationAdjustmentFailureTDD", "nbap.CellSynchronisationAdjustmentFailureTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CauseLevel_SyncAdjustmntFailureTDD_PDU,
      { "CauseLevel-SyncAdjustmntFailureTDD", "nbap.CauseLevel_SyncAdjustmntFailureTDD",
        FT_UINT32, BASE_DEC, VALS(nbap_CauseLevel_SyncAdjustmntFailureTDD_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Unsuccessful_cell_InformationRespItem_SyncAdjustmntFailureTDD_PDU,
      { "Unsuccessful-cell-InformationRespItem-SyncAdjustmntFailureTDD", "nbap.Unsuccessful_cell_InformationRespItem_SyncAdjustmntFailureTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSynchronisationTerminationRequestTDD_PDU,
      { "CellSynchronisationTerminationRequestTDD", "nbap.CellSynchronisationTerminationRequestTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSynchronisationFailureIndicationTDD_PDU,
      { "CellSynchronisationFailureIndicationTDD", "nbap.CellSynchronisationFailureIndicationTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSynchronisationReportTDD_PDU,
      { "CellSynchronisationReportTDD", "nbap.CellSynchronisationReportTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSyncInfo_CellSyncReprtTDD_PDU,
      { "CellSyncInfo-CellSyncReprtTDD", "nbap.CellSyncInfo_CellSyncReprtTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SyncReportType_CellSyncReprtTDD_PDU,
      { "SyncReportType-CellSyncReprtTDD", "nbap.SyncReportType_CellSyncReprtTDD",
        FT_UINT32, BASE_DEC, VALS(nbap_SyncReportType_CellSyncReprtTDD_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_SyncDLCodeIdsMeasInfoList_CellSyncReprtTDD_PDU,
      { "SyncDLCodeIdsMeasInfoList-CellSyncReprtTDD", "nbap.SyncDLCodeIdsMeasInfoList_CellSyncReprtTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_BearerRearrangementIndication_PDU,
      { "BearerRearrangementIndication", "nbap.BearerRearrangementIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DCH_RearrangeList_Bearer_RearrangeInd_PDU,
      { "DCH-RearrangeList-Bearer-RearrangeInd", "nbap.DCH_RearrangeList_Bearer_RearrangeInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DSCH_RearrangeList_Bearer_RearrangeInd_PDU,
      { "DSCH-RearrangeList-Bearer-RearrangeInd", "nbap.DSCH_RearrangeList_Bearer_RearrangeInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_USCH_RearrangeList_Bearer_RearrangeInd_PDU,
      { "USCH-RearrangeList-Bearer-RearrangeInd", "nbap.USCH_RearrangeList_Bearer_RearrangeInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_RearrangeList_Bearer_RearrangeInd_PDU,
      { "HSDSCH-RearrangeList-Bearer-RearrangeInd", "nbap.HSDSCH_RearrangeList_Bearer_RearrangeInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_RearrangeList_Bearer_RearrangeInd_PDU,
      { "E-DCH-RearrangeList-Bearer-RearrangeInd", "nbap.E_DCH_RearrangeList_Bearer_RearrangeInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_EDCH_Cell_Information_Bearer_Rearrangement_List_PDU,
      { "Additional-EDCH-Cell-Information-Bearer-Rearrangement-List", "nbap.Additional_EDCH_Cell_Information_Bearer_Rearrangement_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkActivationCommandFDD_PDU,
      { "RadioLinkActivationCommandFDD", "nbap.RadioLinkActivationCommandFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DelayedActivationInformationList_RL_ActivationCmdFDD_PDU,
      { "DelayedActivationInformationList-RL-ActivationCmdFDD", "nbap.DelayedActivationInformationList_RL_ActivationCmdFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DelayedActivationInformation_RL_ActivationCmdFDD_PDU,
      { "DelayedActivationInformation-RL-ActivationCmdFDD", "nbap.DelayedActivationInformation_RL_ActivationCmdFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkActivationCommandTDD_PDU,
      { "RadioLinkActivationCommandTDD", "nbap.RadioLinkActivationCommandTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DelayedActivationInformationList_RL_ActivationCmdTDD_PDU,
      { "DelayedActivationInformationList-RL-ActivationCmdTDD", "nbap.DelayedActivationInformationList_RL_ActivationCmdTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DelayedActivationInformation_RL_ActivationCmdTDD_PDU,
      { "DelayedActivationInformation-RL-ActivationCmdTDD", "nbap.DelayedActivationInformation_RL_ActivationCmdTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkParameterUpdateIndicationFDD_PDU,
      { "RadioLinkParameterUpdateIndicationFDD", "nbap.RadioLinkParameterUpdateIndicationFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_HS_Cell_Information_RL_Param_Upd_PDU,
      { "Additional-HS-Cell-Information-RL-Param-Upd", "nbap.Additional_HS_Cell_Information_RL_Param_Upd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_EDCH_Cell_Information_RL_Param_Upd_PDU,
      { "Additional-EDCH-Cell-Information-RL-Param-Upd", "nbap.Additional_EDCH_Cell_Information_RL_Param_Upd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RadioLinkParameterUpdateIndicationTDD_PDU,
      { "RadioLinkParameterUpdateIndicationTDD", "nbap.RadioLinkParameterUpdateIndicationTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MBMSNotificationUpdateCommand_PDU,
      { "MBMSNotificationUpdateCommand", "nbap.MBMSNotificationUpdateCommand_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UEStatusUpdateCommand_PDU,
      { "UEStatusUpdateCommand", "nbap.UEStatusUpdateCommand_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SecondaryULFrequencyReport_PDU,
      { "SecondaryULFrequencyReport", "nbap.SecondaryULFrequencyReport_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SecondaryULFrequencyUpdateIndication_PDU,
      { "SecondaryULFrequencyUpdateIndication", "nbap.SecondaryULFrequencyUpdateIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_NBAP_PDU_PDU,
      { "NBAP-PDU", "nbap.NBAP_PDU",
        FT_UINT32, BASE_DEC, VALS(nbap_NBAP_PDU_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_NULL_PDU,
      { "NULL", "nbap.NULL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_local,
      { "local", "nbap.local",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_maxPrivateIEs", HFILL }},
    { &hf_nbap_global,
      { "global", "nbap.global",
        FT_OID, BASE_NONE, NULL, 0,
        "OBJECT_IDENTIFIER", HFILL }},
    { &hf_nbap_procedureCode,
      { "procedureCode", "nbap.procedureCode",
        FT_UINT32, BASE_DEC, VALS(nbap_ProcedureCode_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_ddMode,
      { "ddMode", "nbap.ddMode",
        FT_UINT32, BASE_DEC, VALS(nbap_DdMode_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_shortTransActionId,
      { "shortTransActionId", "nbap.shortTransActionId",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_longTransActionId,
      { "longTransActionId", "nbap.longTransActionId",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ProtocolIE_Container_item,
      { "ProtocolIE-Field", "nbap.ProtocolIE_Field_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_id,
      { "id", "nbap.id",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_ProtocolIE_ID_vals_ext, 0,
        "ProtocolIE_ID", HFILL }},
    { &hf_nbap_criticality,
      { "criticality", "nbap.criticality",
        FT_UINT32, BASE_DEC, VALS(nbap_Criticality_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_ie_field_value,
      { "value", "nbap.value_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "ProtocolIE_Field_value", HFILL }},
    { &hf_nbap_ProtocolExtensionContainer_item,
      { "ProtocolExtensionField", "nbap.ProtocolExtensionField_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_extensionValue,
      { "extensionValue", "nbap.extensionValue_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PrivateIE_Container_item,
      { "PrivateIE-Field", "nbap.PrivateIE_Field_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_id_01,
      { "id", "nbap.id",
        FT_UINT32, BASE_DEC, VALS(nbap_PrivateIE_ID_vals), 0,
        "PrivateIE_ID", HFILL }},
    { &hf_nbap_private_value,
      { "value", "nbap.value_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PrivateIE_Field_value", HFILL }},
    { &hf_nbap_ActivationInformation_item,
      { "ActivationInformationItem", "nbap.ActivationInformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uU_ActivationState,
      { "uU-ActivationState", "nbap.uU_ActivationState",
        FT_UINT32, BASE_DEC, VALS(nbap_Uu_ActivationState_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_iE_Extensions,
      { "iE-Extensions", "nbap.iE_Extensions",
        FT_UINT32, BASE_DEC, NULL, 0,
        "ProtocolExtensionContainer", HFILL }},
    { &hf_nbap_multicell_EDCH_Transport_Bearer_Mode,
      { "multicell-EDCH-Transport-Bearer-Mode", "nbap.multicell_EDCH_Transport_Bearer_Mode",
        FT_UINT32, BASE_DEC, VALS(nbap_Multicell_EDCH_Transport_Bearer_Mode_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_additional_EDCH_Cell_Information_Setup,
      { "additional-EDCH-Cell-Information-Setup", "nbap.additional_EDCH_Cell_Information_Setup",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_EDCH_Cell_Information_Setup_item,
      { "Additional-EDCH-FDD-Setup-Cell-Information", "nbap.Additional_EDCH_FDD_Setup_Cell_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_additional_EDCH_UL_DPCH_Information_Setup,
      { "additional-EDCH-UL-DPCH-Information-Setup", "nbap.additional_EDCH_UL_DPCH_Information_Setup_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_additional_EDCH_RL_Specific_Information_To_Setup,
      { "additional-EDCH-RL-Specific-Information-To-Setup", "nbap.additional_EDCH_RL_Specific_Information_To_Setup",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Additional_EDCH_RL_Specific_Information_To_Setup_List", HFILL }},
    { &hf_nbap_additional_EDCH_FDD_Information,
      { "additional-EDCH-FDD-Information", "nbap.additional_EDCH_FDD_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_additional_EDCH_F_DPCH_Information_Setup,
      { "additional-EDCH-F-DPCH-Information-Setup", "nbap.additional_EDCH_F_DPCH_Information_Setup_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Additional_EDCH_F_DPCH_Information", HFILL }},
    { &hf_nbap_multicell_EDCH_Information,
      { "multicell-EDCH-Information", "nbap.multicell_EDCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ul_ScramblingCode,
      { "ul-ScramblingCode", "nbap.ul_ScramblingCode_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ul_SIR_Target,
      { "ul-SIR-Target", "nbap.ul_SIR_Target",
        FT_INT32, BASE_DEC, NULL, 0,
        "UL_SIR", HFILL }},
    { &hf_nbap_fdd_TPC_DownlinkStepSize,
      { "fdd-TPC-DownlinkStepSize", "nbap.fdd_TPC_DownlinkStepSize",
        FT_UINT32, BASE_DEC, VALS(nbap_FDD_TPC_DownlinkStepSize_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_limitedPowerIncrease,
      { "limitedPowerIncrease", "nbap.limitedPowerIncrease",
        FT_UINT32, BASE_DEC, VALS(nbap_LimitedPowerIncrease_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_innerLoopDLPCStatus,
      { "innerLoopDLPCStatus", "nbap.innerLoopDLPCStatus",
        FT_UINT32, BASE_DEC, VALS(nbap_InnerLoopDLPCStatus_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_EDCH_RL_Specific_Information_To_Setup_List_item,
      { "Additional-EDCH-RL-Specific-Information-To-Setup-ItemIEs", "nbap.Additional_EDCH_RL_Specific_Information_To_Setup_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_eDCH_Additional_RL_ID,
      { "eDCH-Additional-RL-ID", "nbap.eDCH_Additional_RL_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RL_ID", HFILL }},
    { &hf_nbap_c_ID,
      { "c-ID", "nbap.c_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_firstRLS_indicator,
      { "firstRLS-indicator", "nbap.firstRLS_indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_FirstRLS_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_propagationDelay,
      { "propagationDelay", "nbap.propagationDelay",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dl_CodeInformation,
      { "dl-CodeInformation", "nbap.dl_CodeInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        "FDD_DL_CodeInformation", HFILL }},
    { &hf_nbap_initialDL_transmissionPower,
      { "initialDL-transmissionPower", "nbap.initialDL_transmissionPower",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_maximumDL_power,
      { "maximumDL-power", "nbap.maximumDL_power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_minimumDL_power,
      { "minimumDL-power", "nbap.minimumDL_power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_f_DPCH_SlotFormat,
      { "f-DPCH-SlotFormat", "nbap.f_DPCH_SlotFormat",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_RNTI,
      { "e-RNTI", "nbap.e_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_multicell_EDCH_RL_Specific_Information,
      { "multicell-EDCH-RL-Specific-Information", "nbap.multicell_EDCH_RL_Specific_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_EDCH_Cell_Information_To_Add_List_item,
      { "Additional-EDCH-Cell-Information-To-Add-ItemIEs", "nbap.Additional_EDCH_Cell_Information_To_Add_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_additional_EDCH_RL_Specific_Information_To_Add_ItemIEs,
      { "additional-EDCH-RL-Specific-Information-To-Add-ItemIEs", "nbap.additional_EDCH_RL_Specific_Information_To_Add_ItemIEs",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_EDCH_RL_Specific_Information_To_Add_ItemIEs_item,
      { "EDCH-Additional-RL-Specific-Information-To-Add-List", "nbap.EDCH_Additional_RL_Specific_Information_To_Add_List_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_EDCH_RL_Specific_Information_To_Modify_List_item,
      { "Additional-EDCH-RL-Specific-Information-To-Modify-ItemIEs", "nbap.Additional_EDCH_RL_Specific_Information_To_Modify_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_additional_EDCH_MAC_d_Flows_Specific_Information,
      { "additional-EDCH-MAC-d-Flows-Specific-Information", "nbap.additional_EDCH_MAC_d_Flows_Specific_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Additional_EDCH_MAC_d_Flows_Specific_Info_List", HFILL }},
    { &hf_nbap_hARQ_Process_Allocation_Scheduled_2ms_EDCH,
      { "hARQ-Process-Allocation-Scheduled-2ms-EDCH", "nbap.hARQ_Process_Allocation_Scheduled_2ms_EDCH",
        FT_BYTES, BASE_NONE, NULL, 0,
        "HARQ_Process_Allocation_2ms_EDCH", HFILL }},
    { &hf_nbap_e_DCH_Maximum_Bitrate,
      { "e-DCH-Maximum-Bitrate", "nbap.e_DCH_Maximum_Bitrate",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_Processing_Overload_Level,
      { "e-DCH-Processing-Overload-Level", "nbap.e_DCH_Processing_Overload_Level",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_Min_Set_E_TFCI,
      { "e-DCH-Min-Set-E-TFCI", "nbap.e_DCH_Min_Set_E_TFCI",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_TFCI", HFILL }},
    { &hf_nbap_Additional_EDCH_MAC_d_Flows_Specific_Info_List_item,
      { "Additional-EDCH-MAC-d-Flows-Specific-Info", "nbap.Additional_EDCH_MAC_d_Flows_Specific_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_MACdFlow_ID,
      { "e-DCH-MACdFlow-ID", "nbap.e_DCH_MACdFlow_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_bindingID,
      { "bindingID", "nbap.bindingID",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_transportLayerAddress,
      { "transportLayerAddress", "nbap.transportLayerAddress",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_EDCH_Cell_Information_Response_List_item,
      { "Additional-EDCH-FDD-Information-Response-ItemIEs", "nbap.Additional_EDCH_FDD_Information_Response_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_eDCH_Additional_RL_Specific_Information_Response,
      { "eDCH-Additional-RL-Specific-Information-Response", "nbap.eDCH_Additional_RL_Specific_Information_Response",
        FT_UINT32, BASE_DEC, NULL, 0,
        "EDCH_Additional_RL_Specific_Information_Response_List", HFILL }},
    { &hf_nbap_additional_EDCH_MAC_d_Flow_Specific_Information_Response,
      { "additional-EDCH-MAC-d-Flow-Specific-Information-Response", "nbap.additional_EDCH_MAC_d_Flow_Specific_Information_Response",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List", HFILL }},
    { &hf_nbap_EDCH_Additional_RL_Specific_Information_Response_List_item,
      { "EDCH-Additional-RL-Specific-Information-Response-ItemIEs", "nbap.EDCH_Additional_RL_Specific_Information_Response_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_received_total_wide_band_power,
      { "received-total-wide-band-power", "nbap.received_total_wide_band_power",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Received_total_wide_band_power_Value", HFILL }},
    { &hf_nbap_dL_PowerBalancing_ActivationIndicator,
      { "dL-PowerBalancing-ActivationIndicator", "nbap.dL_PowerBalancing_ActivationIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_DL_PowerBalancing_ActivationIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_rL_Set_ID,
      { "rL-Set-ID", "nbap.rL_Set_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_RL_Set_ID,
      { "e-DCH-RL-Set-ID", "nbap.e_DCH_RL_Set_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RL_Set_ID", HFILL }},
    { &hf_nbap_e_DCH_FDD_DL_Control_Channel_Information,
      { "e-DCH-FDD-DL-Control-Channel-Information", "nbap.e_DCH_FDD_DL_Control_Channel_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_EDCH_Cell_Information_Response_RLReconf_List_item,
      { "Additional-EDCH-FDD-Information-Response-RLReconf-Items", "nbap.Additional_EDCH_FDD_Information_Response_RLReconf_Items_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_additional_EDCH_FDD_Information_Response_ItemIEs,
      { "additional-EDCH-FDD-Information-Response-ItemIEs", "nbap.additional_EDCH_FDD_Information_Response_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_additional_Modififed_EDCH_FDD_Information_Response_ItemIEs,
      { "additional-Modififed-EDCH-FDD-Information-Response-ItemIEs", "nbap.additional_Modififed_EDCH_FDD_Information_Response_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_eDCH_Additional_Modified_RL_Specific_Information_Response,
      { "eDCH-Additional-Modified-RL-Specific-Information-Response", "nbap.eDCH_Additional_Modified_RL_Specific_Information_Response",
        FT_UINT32, BASE_DEC, NULL, 0,
        "EDCH_Additional_Modified_RL_Specific_Information_Response_List", HFILL }},
    { &hf_nbap_EDCH_Additional_Modified_RL_Specific_Information_Response_List_item,
      { "EDCH-Additional-Modified-RL-Specific-Information-Response-List-Items", "nbap.EDCH_Additional_Modified_RL_Specific_Information_Response_List_Items_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_PowerBalancing_UpdatedIndicator,
      { "dL-PowerBalancing-UpdatedIndicator", "nbap.dL_PowerBalancing_UpdatedIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_DL_PowerBalancing_UpdatedIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List_item,
      { "Additional-EDCH-MAC-d-Flows-Specific-Info-Response", "nbap.Additional_EDCH_MAC_d_Flows_Specific_Info_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_EDCH_Cell_Information_Response_RL_Add_List_item,
      { "Additional-EDCH-Cell-Information-Response-RL-Add-ItemIEs", "nbap.Additional_EDCH_Cell_Information_Response_RL_Add_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_additional_EDCH_FDD_Information_Response,
      { "additional-EDCH-FDD-Information-Response", "nbap.additional_EDCH_FDD_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Additional_EDCH_FDD_Information_Response_ItemIEs", HFILL }},
    { &hf_nbap_additional_EDCH_Serving_Cell_Change_Information_Response,
      { "additional-EDCH-Serving-Cell-Change-Information-Response", "nbap.additional_EDCH_Serving_Cell_Change_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "E_DCH_Serving_Cell_Change_Info_Response", HFILL }},
    { &hf_nbap_Additional_EDCH_Cell_Information_ConfigurationChange_List_item,
      { "Additional-EDCH-ConfigurationChange-Info-ItemIEs", "nbap.Additional_EDCH_ConfigurationChange_Info_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_additional_EDCH_UL_DPCH_Information_Modify,
      { "additional-EDCH-UL-DPCH-Information-Modify", "nbap.additional_EDCH_UL_DPCH_Information_Modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_additional_EDCH_RL_Specific_Information_To_Add,
      { "additional-EDCH-RL-Specific-Information-To-Add", "nbap.additional_EDCH_RL_Specific_Information_To_Add",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Additional_EDCH_RL_Specific_Information_To_Add_ItemIEs", HFILL }},
    { &hf_nbap_additional_EDCH_RL_Specific_Information_To_Modify,
      { "additional-EDCH-RL-Specific-Information-To-Modify", "nbap.additional_EDCH_RL_Specific_Information_To_Modify",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Additional_EDCH_RL_Specific_Information_To_Modify_List", HFILL }},
    { &hf_nbap_additional_EDCH_FDD_Information_To_Modify,
      { "additional-EDCH-FDD-Information-To-Modify", "nbap.additional_EDCH_FDD_Information_To_Modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Additional_EDCH_FDD_Information", HFILL }},
    { &hf_nbap_additional_EDCH_F_DPCH_Information_Modify,
      { "additional-EDCH-F-DPCH-Information-Modify", "nbap.additional_EDCH_F_DPCH_Information_Modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Additional_EDCH_F_DPCH_Information", HFILL }},
    { &hf_nbap_Additional_EDCH_Cell_Information_Removal_List_item,
      { "Additional-EDCH-Cell-Information-Removal-Info-ItemIEs", "nbap.Additional_EDCH_Cell_Information_Removal_Info_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_rL_on_Secondary_UL_Frequency,
      { "rL-on-Secondary-UL-Frequency", "nbap.rL_on_Secondary_UL_Frequency",
        FT_UINT32, BASE_DEC, VALS(nbap_RL_on_Secondary_UL_Frequency_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_additional_EDCH_DL_Control_Channel_Change_Information,
      { "additional-EDCH-DL-Control-Channel-Change-Information", "nbap.additional_EDCH_DL_Control_Channel_Change_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Additional_EDCH_DL_Control_Channel_Change_Information_List", HFILL }},
    { &hf_nbap_Additional_EDCH_DL_Control_Channel_Change_Information_List_item,
      { "Additional-EDCH-DL-Control-Channel-Change-Info-ItemIEs", "nbap.Additional_EDCH_DL_Control_Channel_Change_Info_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_AdditionalMeasurementValueList_item,
      { "AdditionalMeasurementValue", "nbap.AdditionalMeasurementValue_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uARFCN,
      { "uARFCN", "nbap.uARFCN",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_timeSlotMeasurementValueListLCR,
      { "timeSlotMeasurementValueListLCR", "nbap.timeSlotMeasurementValueListLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_AdditionalTimeSlotListLCR_item,
      { "AdditionalTimeSlotLCR", "nbap.AdditionalTimeSlotLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_timeslot_InitiatedListLCR,
      { "timeslot-InitiatedListLCR", "nbap.timeslot_InitiatedListLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cMConfigurationChangeCFN,
      { "cMConfigurationChangeCFN", "nbap.cMConfigurationChangeCFN",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CFN", HFILL }},
    { &hf_nbap_transmission_Gap_Pattern_Sequence_Status,
      { "transmission-Gap-Pattern-Sequence-Status", "nbap.transmission_Gap_Pattern_Sequence_Status",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Transmission_Gap_Pattern_Sequence_Status_List", HFILL }},
    { &hf_nbap_Transmission_Gap_Pattern_Sequence_Status_List_item,
      { "Transmission-Gap-Pattern-Sequence-Status-List item", "nbap.Transmission_Gap_Pattern_Sequence_Status_List_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tGPSID,
      { "tGPSID", "nbap.tGPSID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tGPRC,
      { "tGPRC", "nbap.tGPRC",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tGCFN,
      { "tGCFN", "nbap.tGCFN",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CFN", HFILL }},
    { &hf_nbap_priorityLevel,
      { "priorityLevel", "nbap.priorityLevel",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_pre_emptionCapability,
      { "pre-emptionCapability", "nbap.pre_emptionCapability",
        FT_UINT32, BASE_DEC, VALS(nbap_Pre_emptionCapability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_pre_emptionVulnerability,
      { "pre-emptionVulnerability", "nbap.pre_emptionVulnerability",
        FT_UINT32, BASE_DEC, VALS(nbap_Pre_emptionVulnerability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_aOA_LCR,
      { "aOA-LCR", "nbap.aOA_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_aOA_LCR_Accuracy_Class,
      { "aOA-LCR-Accuracy-Class", "nbap.aOA_LCR_Accuracy_Class",
        FT_UINT32, BASE_DEC, VALS(nbap_AOA_LCR_Accuracy_Class_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_bCCH_Specific_HSDSCH_RNTI,
      { "bCCH-Specific-HSDSCH-RNTI", "nbap.bCCH_Specific_HSDSCH_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HSDSCH_RNTI", HFILL }},
    { &hf_nbap_hSSCCH_Power,
      { "hSSCCH-Power", "nbap.hSSCCH_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_hSPDSCH_Power,
      { "hSPDSCH-Power", "nbap.hSPDSCH_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_Best_Cell_Portions_Value_item,
      { "Best-Cell-Portions-Item", "nbap.Best_Cell_Portions_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cellPortionID,
      { "cellPortionID", "nbap.cellPortionID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sIRValue,
      { "sIRValue", "nbap.sIRValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SIR_Value", HFILL }},
    { &hf_nbap_Best_Cell_Portions_ValueLCR_item,
      { "Best-Cell-Portions-ItemLCR", "nbap.Best_Cell_Portions_ItemLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cellPortionLCRID,
      { "cellPortionLCRID", "nbap.cellPortionLCRID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_rSCPValue,
      { "rSCPValue", "nbap.rSCPValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RSCP_Value", HFILL }},
    { &hf_nbap_commonTransportChannelID,
      { "commonTransportChannelID", "nbap.commonTransportChannelID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cid,
      { "cid", "nbap.cid",
        FT_UINT32, BASE_DEC, NULL, 0,
        "C_ID", HFILL }},
    { &hf_nbap_radioNetwork,
      { "radioNetwork", "nbap.radioNetwork",
        FT_UINT32, BASE_DEC, VALS(nbap_CauseRadioNetwork_vals), 0,
        "CauseRadioNetwork", HFILL }},
    { &hf_nbap_transport,
      { "transport", "nbap.transport",
        FT_UINT32, BASE_DEC, VALS(nbap_CauseTransport_vals), 0,
        "CauseTransport", HFILL }},
    { &hf_nbap_protocol,
      { "protocol", "nbap.protocol",
        FT_UINT32, BASE_DEC, VALS(nbap_CauseProtocol_vals), 0,
        "CauseProtocol", HFILL }},
    { &hf_nbap_misc,
      { "misc", "nbap.misc",
        FT_UINT32, BASE_DEC, VALS(nbap_CauseMisc_vals), 0,
        "CauseMisc", HFILL }},
    { &hf_nbap_Cell_ERNTI_Status_Information_item,
      { "Cell-ERNTI-Status-Information-Item", "nbap.Cell_ERNTI_Status_Information_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_vacant_ERNTI,
      { "vacant-ERNTI", "nbap.vacant_ERNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Vacant_ERNTI_item,
      { "E-RNTI", "nbap.E_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_initialPhase_0_1048575,
      { "initialPhase", "nbap.initialPhase",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_1048575_", HFILL }},
    { &hf_nbap_steadyStatePhase,
      { "steadyStatePhase", "nbap.steadyStatePhase",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_255_", HFILL }},
    { &hf_nbap_initialPhase,
      { "initialPhase", "nbap.initialPhase",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_524287_", HFILL }},
    { &hf_nbap_steadyStatePhase_01,
      { "steadyStatePhase", "nbap.steadyStatePhase",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_127_", HFILL }},
    { &hf_nbap_CommonChannelsCapacityConsumptionLaw_item,
      { "CommonChannelsCapacityConsumptionLaw item", "nbap.CommonChannelsCapacityConsumptionLaw_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dl_Cost,
      { "dl-Cost", "nbap.dl_Cost",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_65535", HFILL }},
    { &hf_nbap_ul_Cost,
      { "ul-Cost", "nbap.ul_Cost",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_65535", HFILL }},
    { &hf_nbap_common_E_DCH_UL_DPCH_Information,
      { "common-E-DCH-UL-DPCH-Information", "nbap.common_E_DCH_UL_DPCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Common_E_DCH_UL_DPCH_InfoItem", HFILL }},
    { &hf_nbap_common_E_DCH_EDPCH_Information,
      { "common-E-DCH-EDPCH-Information", "nbap.common_E_DCH_EDPCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Common_E_DCH_EDPCH_InfoItem", HFILL }},
    { &hf_nbap_common_E_DCH_Information,
      { "common-E-DCH-Information", "nbap.common_E_DCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Common_E_DCH_InfoItem", HFILL }},
    { &hf_nbap_common_E_DCH_HSDPCCH_Information,
      { "common-E-DCH-HSDPCCH-Information", "nbap.common_E_DCH_HSDPCCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Common_E_DCH_HSDPCCH_InfoItem", HFILL }},
    { &hf_nbap_common_E_DCH_Preamble_Control_Information,
      { "common-E-DCH-Preamble-Control-Information", "nbap.common_E_DCH_Preamble_Control_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Common_E_DCH_Preamble_Control_InfoItem", HFILL }},
    { &hf_nbap_common_E_DCH_FDPCH_Information,
      { "common-E-DCH-FDPCH-Information", "nbap.common_E_DCH_FDPCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Common_E_DCH_FDPCH_InfoItem", HFILL }},
    { &hf_nbap_common_E_DCH_E_AGCH_ChannelisationCodeNumber,
      { "common-E-DCH-E-AGCH-ChannelisationCodeNumber", "nbap.common_E_DCH_E_AGCH_ChannelisationCodeNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        "FDD_DL_ChannelisationCodeNumber", HFILL }},
    { &hf_nbap_common_E_DCH_Resource_Combination_Information,
      { "common-E-DCH-Resource-Combination-Information", "nbap.common_E_DCH_Resource_Combination_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Common_E_DCH_Resource_Combination_InfoList", HFILL }},
    { &hf_nbap_ul_common_E_DCH_MACflow_Specific_Information,
      { "ul-common-E-DCH-MACflow-Specific-Information", "nbap.ul_common_E_DCH_MACflow_Specific_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Ul_common_E_DCH_MACflow_Specific_InfoList", HFILL }},
    { &hf_nbap_uL_SIR_Target,
      { "uL-SIR-Target", "nbap.uL_SIR_Target",
        FT_INT32, BASE_DEC, NULL, 0,
        "UL_SIR", HFILL }},
    { &hf_nbap_dPC_Mode,
      { "dPC-Mode", "nbap.dPC_Mode",
        FT_UINT32, BASE_DEC, VALS(nbap_DPC_Mode_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_maxSet_E_DPDCHs,
      { "maxSet-E-DPDCHs", "nbap.maxSet_E_DPDCHs",
        FT_UINT32, BASE_DEC, VALS(nbap_Max_Set_E_DPDCHs_vals), 0,
        "Max_Set_E_DPDCHs", HFILL }},
    { &hf_nbap_ul_PunctureLimit,
      { "ul-PunctureLimit", "nbap.ul_PunctureLimit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PunctureLimit", HFILL }},
    { &hf_nbap_e_TFCS_Information,
      { "e-TFCS-Information", "nbap.e_TFCS_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_TTI,
      { "e-TTI", "nbap.e_TTI",
        FT_UINT32, BASE_DEC, VALS(nbap_E_TTI_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_e_DPCCH_PO,
      { "e-DPCCH-PO", "nbap.e_DPCCH_PO",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_RGCH_2_IndexStepThreshold,
      { "e-RGCH-2-IndexStepThreshold", "nbap.e_RGCH_2_IndexStepThreshold",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_RGCH_3_IndexStepThreshold,
      { "e-RGCH-3-IndexStepThreshold", "nbap.e_RGCH_3_IndexStepThreshold",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hARQ_Info_for_E_DCH,
      { "hARQ-Info-for-E-DCH", "nbap.hARQ_Info_for_E_DCH",
        FT_UINT32, BASE_DEC, VALS(nbap_HARQ_Info_for_E_DCH_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_Reference_Power_Offset,
      { "e-DCH-Reference-Power-Offset", "nbap.e_DCH_Reference_Power_Offset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_PowerOffset_for_SchedulingInfo,
      { "e-DCH-PowerOffset-for-SchedulingInfo", "nbap.e_DCH_PowerOffset_for_SchedulingInfo",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_max_EDCH_Resource_Allocation_for_CCCH,
      { "max-EDCH-Resource-Allocation-for-CCCH", "nbap.max_EDCH_Resource_Allocation_for_CCCH",
        FT_UINT32, BASE_DEC, VALS(nbap_Max_EDCH_Resource_Allocation_for_CCCH_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_max_Period_for_Collistion_Resolution,
      { "max-Period-for-Collistion-Resolution", "nbap.max_Period_for_Collistion_Resolution",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_max_TB_Sizes,
      { "max-TB-Sizes", "nbap.max_TB_Sizes_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_common_E_DCH_ImplicitRelease_Indicator,
      { "common-E-DCH-ImplicitRelease-Indicator", "nbap.common_E_DCH_ImplicitRelease_Indicator",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        "BOOLEAN", HFILL }},
    { &hf_nbap_ackNackRepetitionFactor,
      { "ackNackRepetitionFactor", "nbap.ackNackRepetitionFactor",
        FT_UINT32, BASE_DEC, NULL, 0,
        "AckNack_RepetitionFactor", HFILL }},
    { &hf_nbap_ackPowerOffset,
      { "ackPowerOffset", "nbap.ackPowerOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Ack_Power_Offset", HFILL }},
    { &hf_nbap_nackPowerOffset,
      { "nackPowerOffset", "nbap.nackPowerOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Nack_Power_Offset", HFILL }},
    { &hf_nbap_common_E_DCH_CQI_Info,
      { "common-E-DCH-CQI-Info", "nbap.common_E_DCH_CQI_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cqiFeedback_CycleK,
      { "cqiFeedback-CycleK", "nbap.cqiFeedback_CycleK",
        FT_UINT32, BASE_DEC, VALS(nbap_CQI_Feedback_Cycle_vals), 0,
        "CQI_Feedback_Cycle", HFILL }},
    { &hf_nbap_cqiRepetitionFactor,
      { "cqiRepetitionFactor", "nbap.cqiRepetitionFactor",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CQI_RepetitionFactor", HFILL }},
    { &hf_nbap_cqiPowerOffset,
      { "cqiPowerOffset", "nbap.cqiPowerOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CQI_Power_Offset", HFILL }},
    { &hf_nbap_measurement_Power_Offset,
      { "measurement-Power-Offset", "nbap.measurement_Power_Offset",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_commonPhysicalChannelID,
      { "commonPhysicalChannelID", "nbap.commonPhysicalChannelID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_common_E_DCH_PreambleSignatures,
      { "common-E-DCH-PreambleSignatures", "nbap.common_E_DCH_PreambleSignatures",
        FT_BYTES, BASE_NONE, NULL, 0,
        "PreambleSignatures", HFILL }},
    { &hf_nbap_scramblingCodeNumber,
      { "scramblingCodeNumber", "nbap.scramblingCodeNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_preambleThreshold,
      { "preambleThreshold", "nbap.preambleThreshold",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_AI_Indicator,
      { "e-AI-Indicator", "nbap.e_AI_Indicator",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_common_E_DCH_AICH_Information,
      { "common-E-DCH-AICH-Information", "nbap.common_E_DCH_AICH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_aICH_TransmissionTiming,
      { "aICH-TransmissionTiming", "nbap.aICH_TransmissionTiming",
        FT_UINT32, BASE_DEC, VALS(nbap_AICH_TransmissionTiming_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_fdd_dl_ChannelisationCodeNumber,
      { "fdd-dl-ChannelisationCodeNumber", "nbap.fdd_dl_ChannelisationCodeNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_aICH_Power,
      { "aICH-Power", "nbap.aICH_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sTTD_Indicator,
      { "sTTD-Indicator", "nbap.sTTD_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_STTD_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Common_E_DCH_Resource_Combination_InfoList_item,
      { "Common-E-DCH-Resource-Combination-InfoList-Item", "nbap.Common_E_DCH_Resource_Combination_InfoList_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_soffset,
      { "soffset", "nbap.soffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_f_DPCH_DL_Code_Number,
      { "f-DPCH-DL-Code-Number", "nbap.f_DPCH_DL_Code_Number",
        FT_UINT32, BASE_DEC, NULL, 0,
        "FDD_DL_ChannelisationCodeNumber", HFILL }},
    { &hf_nbap_ul_DPCH_ScramblingCode,
      { "ul-DPCH-ScramblingCode", "nbap.ul_DPCH_ScramblingCode_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_ScramblingCode", HFILL }},
    { &hf_nbap_e_RGCH_E_HICH_Channelisation_Code,
      { "e-RGCH-E-HICH-Channelisation-Code", "nbap.e_RGCH_E_HICH_Channelisation_Code",
        FT_UINT32, BASE_DEC, NULL, 0,
        "FDD_DL_ChannelisationCodeNumber", HFILL }},
    { &hf_nbap_e_RGCH_Signature_Sequence,
      { "e-RGCH-Signature-Sequence", "nbap.e_RGCH_Signature_Sequence",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_HICH_Signature_Sequence,
      { "e-HICH-Signature-Sequence", "nbap.e_HICH_Signature_Sequence",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Ul_common_E_DCH_MACflow_Specific_InfoList_item,
      { "Ul-common-E-DCH-MACflow-Specific-InfoList-Item", "nbap.Ul_common_E_DCH_MACflow_Specific_InfoList_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ul_Common_MACFlowID,
      { "ul-Common-MACFlowID", "nbap.ul_Common_MACFlowID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Common_MACFlow_ID", HFILL }},
    { &hf_nbap_transportBearerRequestIndicator,
      { "transportBearerRequestIndicator", "nbap.transportBearerRequestIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_TransportBearerRequestIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_tnlQos,
      { "tnlQos", "nbap.tnlQos",
        FT_UINT32, BASE_DEC, VALS(nbap_TnlQos_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_payloadCRC_PresenceIndicator,
      { "payloadCRC-PresenceIndicator", "nbap.payloadCRC_PresenceIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_PayloadCRC_PresenceIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_bundlingModeIndicator,
      { "bundlingModeIndicator", "nbap.bundlingModeIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_BundlingModeIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_common_E_DCH_MACdFlow_Specific_Information,
      { "common-E-DCH-MACdFlow-Specific-Information", "nbap.common_E_DCH_MACdFlow_Specific_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Common_E_DCH_MACdFlow_Specific_InfoList", HFILL }},
    { &hf_nbap_Common_E_DCH_MACdFlow_Specific_InfoList_item,
      { "Common-E-DCH-MACdFlow-Specific-InfoList-Item", "nbap.Common_E_DCH_MACdFlow_Specific_InfoList_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_common_e_DCH_MACdFlow_ID,
      { "common-e-DCH-MACdFlow-ID", "nbap.common_e_DCH_MACdFlow_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_MACdFlow_ID", HFILL }},
    { &hf_nbap_maximum_Number_of_Retransmissions_For_E_DCH,
      { "maximum-Number-of-Retransmissions-For-E-DCH", "nbap.maximum_Number_of_Retransmissions_For_E_DCH",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_eDCH_HARQ_PO_FDD,
      { "eDCH-HARQ-PO-FDD", "nbap.eDCH_HARQ_PO_FDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_HARQ_PO_FDD", HFILL }},
    { &hf_nbap_eDCH_MACdFlow_Multiplexing_List,
      { "eDCH-MACdFlow-Multiplexing-List", "nbap.eDCH_MACdFlow_Multiplexing_List",
        FT_BYTES, BASE_NONE, NULL, 0,
        "E_DCH_MACdFlow_Multiplexing_List", HFILL }},
    { &hf_nbap_common_E_DCHLogicalChannelInformation,
      { "common-E-DCHLogicalChannelInformation", "nbap.common_E_DCHLogicalChannelInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Common_E_DCH_LogicalChannel_InfoList", HFILL }},
    { &hf_nbap_Common_E_DCH_LogicalChannel_InfoList_item,
      { "Common-E-DCH-LogicalChannel-InfoList-Item", "nbap.Common_E_DCH_LogicalChannel_InfoList_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_logicalChannelId,
      { "logicalChannelId", "nbap.logicalChannelId",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_maximumMACcPDU_SizeExtended,
      { "maximumMACcPDU-SizeExtended", "nbap.maximumMACcPDU_SizeExtended",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MAC_PDU_SizeExtended", HFILL }},
    { &hf_nbap_ul_common_E_DCH_MACflow_Specific_InfoResponse,
      { "ul-common-E-DCH-MACflow-Specific-InfoResponse", "nbap.ul_common_E_DCH_MACflow_Specific_InfoResponse",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Ul_common_E_DCH_MACflow_Specific_InfoResponseList", HFILL }},
    { &hf_nbap_serving_Grant_Value,
      { "serving-Grant-Value", "nbap.serving_Grant_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_Serving_Grant_Value", HFILL }},
    { &hf_nbap_E_RNTI_List_item,
      { "E-RNTI", "nbap.E_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseList_item,
      { "Ul-common-E-DCH-MACflow-Specific-InfoResponseList-Item", "nbap.Ul_common_E_DCH_MACflow_Specific_InfoResponseList_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Common_MACFlows_to_DeleteFDD_item,
      { "Common-MACFlows-to-DeleteFDD-Item", "nbap.Common_MACFlows_to_DeleteFDD_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_common_MACFlow_ID,
      { "common-MACFlow-ID", "nbap.common_MACFlow_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonMACFlow_Specific_InfoList_item,
      { "CommonMACFlow-Specific-InfoItem", "nbap.CommonMACFlow_Specific_InfoItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_common_MACFlow_Id,
      { "common-MACFlow-Id", "nbap.common_MACFlow_Id",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tnl_qos,
      { "tnl-qos", "nbap.tnl_qos",
        FT_UINT32, BASE_DEC, VALS(nbap_TnlQos_vals), 0,
        "TnlQos", HFILL }},
    { &hf_nbap_common_MACFlow_PriorityQueue_Information,
      { "common-MACFlow-PriorityQueue-Information", "nbap.common_MACFlow_PriorityQueue_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonMACFlow_Specific_InfoList_Response_item,
      { "CommonMACFlow-Specific-InfoItem-Response", "nbap.CommonMACFlow_Specific_InfoItem_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_commonMACFlow_ID,
      { "commonMACFlow-ID", "nbap.commonMACFlow_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Common_MACFlow_ID", HFILL }},
    { &hf_nbap_hSDSCH_Initial_Capacity_Allocation,
      { "hSDSCH-Initial-Capacity-Allocation", "nbap.hSDSCH_Initial_Capacity_Allocation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Common_MACFlow_PriorityQueue_Information_item,
      { "Common-MACFlow-PriorityQueue-Item", "nbap.Common_MACFlow_PriorityQueue_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_priority_Queue_Information_for_Enhanced_FACH,
      { "priority-Queue-Information-for-Enhanced-FACH", "nbap.priority_Queue_Information_for_Enhanced_FACH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Priority_Queue_Information_for_Enhanced_FACH_PCH", HFILL }},
    { &hf_nbap_tUTRANGPSMeasurementAccuracyClass,
      { "tUTRANGPSMeasurementAccuracyClass", "nbap.tUTRANGPSMeasurementAccuracyClass",
        FT_UINT32, BASE_DEC, VALS(nbap_TUTRANGPSAccuracyClass_vals), 0,
        "TUTRANGPSAccuracyClass", HFILL }},
    { &hf_nbap_tUTRANGANSSMeasurementAccuracyClass,
      { "tUTRANGANSSMeasurementAccuracyClass", "nbap.tUTRANGANSSMeasurementAccuracyClass",
        FT_UINT32, BASE_DEC, VALS(nbap_TUTRANGANSSAccuracyClass_vals), 0,
        "TUTRANGANSSAccuracyClass", HFILL }},
    { &hf_nbap_transmitted_carrier_power,
      { "transmitted-carrier-power", "nbap.transmitted_carrier_power",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Transmitted_Carrier_Power_Value", HFILL }},
    { &hf_nbap_acknowledged_prach_preambles,
      { "acknowledged-prach-preambles", "nbap.acknowledged_prach_preambles",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Acknowledged_PRACH_preambles_Value", HFILL }},
    { &hf_nbap_uL_TimeslotISCP,
      { "uL-TimeslotISCP", "nbap.uL_TimeslotISCP",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UL_TimeslotISCP_Value", HFILL }},
    { &hf_nbap_notUsed_1_acknowledged_PCPCH_access_preambles,
      { "notUsed-1-acknowledged-PCPCH-access-preambles", "nbap.notUsed_1_acknowledged_PCPCH_access_preambles_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_notUsed_2_detected_PCPCH_access_preambles,
      { "notUsed-2-detected-PCPCH-access-preambles", "nbap.notUsed_2_detected_PCPCH_access_preambles_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_extension_CommonMeasurementValue,
      { "extension-CommonMeasurementValue", "nbap.extension_CommonMeasurementValue_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_measurementAvailable,
      { "measurementAvailable", "nbap.measurementAvailable_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "CommonMeasurementAvailable", HFILL }},
    { &hf_nbap_measurementnotAvailable,
      { "measurementnotAvailable", "nbap.measurementnotAvailable_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "CommonMeasurementnotAvailable", HFILL }},
    { &hf_nbap_commonmeasurementValue,
      { "commonmeasurementValue", "nbap.commonmeasurementValue",
        FT_UINT32, BASE_DEC, VALS(nbap_CommonMeasurementValue_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_ie_Extensions,
      { "ie-Extensions", "nbap.ie_Extensions",
        FT_UINT32, BASE_DEC, NULL, 0,
        "ProtocolExtensionContainer", HFILL }},
    { &hf_nbap_resourceOperationalState,
      { "resourceOperationalState", "nbap.resourceOperationalState",
        FT_UINT32, BASE_DEC, VALS(nbap_ResourceOperationalState_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_availabilityStatus,
      { "availabilityStatus", "nbap.availabilityStatus",
        FT_UINT32, BASE_DEC, VALS(nbap_AvailabilityStatus_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_commonPhysicalChannelID768,
      { "commonPhysicalChannelID768", "nbap.commonPhysicalChannelID768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uE_DTX_DRX_Offset,
      { "uE-DTX-DRX-Offset", "nbap.uE_DTX_DRX_Offset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_enabling_Delay,
      { "enabling-Delay", "nbap.enabling_Delay",
        FT_UINT32, BASE_DEC, VALS(nbap_Enabling_Delay_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_dTX_Information,
      { "dTX-Information", "nbap.dTX_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dRX_Information,
      { "dRX-Information", "nbap.dRX_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dTX_Information_to_Modify,
      { "dTX-Information-to-Modify", "nbap.dTX_Information_to_Modify",
        FT_UINT32, BASE_DEC, VALS(nbap_DTX_Information_to_Modify_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_dRX_Information_to_Modify,
      { "dRX-Information-to-Modify", "nbap.dRX_Information_to_Modify",
        FT_UINT32, BASE_DEC, VALS(nbap_DRX_Information_to_Modify_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_ContinuousPacketConnectivityHS_SCCH_less_Information_item,
      { "ContinuousPacketConnectivityHS-SCCH-less-InformationItem", "nbap.ContinuousPacketConnectivityHS_SCCH_less_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_transport_Block_Size_Index,
      { "transport-Block-Size-Index", "nbap.transport_Block_Size_Index",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hSPDSCH_Second_Code_Support,
      { "hSPDSCH-Second-Code-Support", "nbap.hSPDSCH_Second_Code_Support",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hSPDSCH_First_Code_Index,
      { "hSPDSCH-First-Code-Index", "nbap.hSPDSCH_First_Code_Index",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hSPDSCH_Second_Code_Index,
      { "hSPDSCH-Second-Code-Index", "nbap.hSPDSCH_Second_Code_Index",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_continuousPacketConnectivityDTX_DRX_Information,
      { "continuousPacketConnectivityDTX-DRX-Information", "nbap.continuousPacketConnectivityDTX_DRX_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_continuousPacketConnectivityDTX_DRX_Information_to_Modify,
      { "continuousPacketConnectivityDTX-DRX-Information-to-Modify", "nbap.continuousPacketConnectivityDTX_DRX_Information_to_Modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_continuousPacketConnectivityHS_SCCH_less_Information,
      { "continuousPacketConnectivityHS-SCCH-less-Information", "nbap.continuousPacketConnectivityHS_SCCH_less_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_procedureID,
      { "procedureID", "nbap.procedureID_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_triggeringMessage,
      { "triggeringMessage", "nbap.triggeringMessage",
        FT_UINT32, BASE_DEC, VALS(nbap_TriggeringMessage_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_procedureCriticality,
      { "procedureCriticality", "nbap.procedureCriticality",
        FT_UINT32, BASE_DEC, VALS(nbap_Criticality_vals), 0,
        "Criticality", HFILL }},
    { &hf_nbap_transactionID,
      { "transactionID", "nbap.transactionID",
        FT_UINT32, BASE_DEC, VALS(nbap_TransactionID_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_iEsCriticalityDiagnostics,
      { "iEsCriticalityDiagnostics", "nbap.iEsCriticalityDiagnostics",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CriticalityDiagnostics_IE_List", HFILL }},
    { &hf_nbap_CriticalityDiagnostics_IE_List_item,
      { "CriticalityDiagnostics-IE-List item", "nbap.CriticalityDiagnostics_IE_List_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_iECriticality,
      { "iECriticality", "nbap.iECriticality",
        FT_UINT32, BASE_DEC, VALS(nbap_Criticality_vals), 0,
        "Criticality", HFILL }},
    { &hf_nbap_iE_ID,
      { "iE-ID", "nbap.iE_ID",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_ProtocolIE_ID_vals_ext, 0,
        "ProtocolIE_ID", HFILL }},
    { &hf_nbap_repetitionNumber,
      { "repetitionNumber", "nbap.repetitionNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RepetitionNumber0", HFILL }},
    { &hf_nbap_ul_common_E_DCH_MACflow_Specific_InformationLCR,
      { "ul-common-E-DCH-MACflow-Specific-InformationLCR", "nbap.ul_common_E_DCH_MACflow_Specific_InformationLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Ul_common_E_DCH_MACflow_Specific_InfoListLCR", HFILL }},
    { &hf_nbap_common_E_PUCH_InformationLCR,
      { "common-E-PUCH-InformationLCR", "nbap.common_E_PUCH_InformationLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_TFCS_Information_TDD,
      { "e-TFCS-Information-TDD", "nbap.e_TFCS_Information_TDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_maximum_Number_of_Retransmissions_For_SchedulingInfo,
      { "maximum-Number-of-Retransmissions-For-SchedulingInfo", "nbap.maximum_Number_of_Retransmissions_For_SchedulingInfo",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Maximum_Number_of_Retransmissions_For_E_DCH", HFILL }},
    { &hf_nbap_eDCH_Retransmission_Timer_SchedulingInfo,
      { "eDCH-Retransmission-Timer-SchedulingInfo", "nbap.eDCH_Retransmission_Timer_SchedulingInfo",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_E_DCH_MACdFlow_Retransmission_Timer_vals_ext, 0,
        "E_DCH_MACdFlow_Retransmission_Timer", HFILL }},
    { &hf_nbap_minCR,
      { "minCR", "nbap.minCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CodeRate", HFILL }},
    { &hf_nbap_maxCR,
      { "maxCR", "nbap.maxCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CodeRate", HFILL }},
    { &hf_nbap_harqInfo,
      { "harqInfo", "nbap.harqInfo",
        FT_UINT32, BASE_DEC, VALS(nbap_HARQ_Info_for_E_DCH_vals), 0,
        "HARQ_Info_for_E_DCH", HFILL }},
    { &hf_nbap_pRXdes_base_perURAFCN,
      { "pRXdes-base-perURAFCN", "nbap.pRXdes_base_perURAFCN",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_PUCH_TPC_StepSize,
      { "e-PUCH-TPC-StepSize", "nbap.e_PUCH_TPC_StepSize",
        FT_UINT32, BASE_DEC, VALS(nbap_TDD_TPC_UplinkStepSize_LCR_vals), 0,
        "TDD_TPC_UplinkStepSize_LCR", HFILL }},
    { &hf_nbap_e_AGCH_TPC_StepSize,
      { "e-AGCH-TPC-StepSize", "nbap.e_AGCH_TPC_StepSize",
        FT_UINT32, BASE_DEC, VALS(nbap_TDD_TPC_DownlinkStepSize_vals), 0,
        "TDD_TPC_DownlinkStepSize", HFILL }},
    { &hf_nbap_e_PUCH_PowerControlGAP,
      { "e-PUCH-PowerControlGAP", "nbap.e_PUCH_PowerControlGAP",
        FT_UINT32, BASE_DEC, NULL, 0,
        "ControlGAP", HFILL }},
    { &hf_nbap_PRXdes_base_perURAFCN_item,
      { "PRXdes-base-Item", "nbap.PRXdes_base_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_pRXdes_base,
      { "pRXdes-base", "nbap.pRXdes_base",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Ul_common_E_DCH_MACflow_Specific_InfoListLCR_item,
      { "Ul-common-E-DCH-MACflow-Specific-InfoList-ItemLCR", "nbap.Ul_common_E_DCH_MACflow_Specific_InfoList_ItemLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ul_Common_MACFlowIDLCR,
      { "ul-Common-MACFlowIDLCR", "nbap.ul_Common_MACFlowIDLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Common_MACFlow_ID_LCR", HFILL }},
    { &hf_nbap_common_E_DCH_MACdFlow_Specific_InformationLCR,
      { "common-E-DCH-MACdFlow-Specific-InformationLCR", "nbap.common_E_DCH_MACdFlow_Specific_InformationLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Common_E_DCH_MACdFlow_Specific_InfoListLCR", HFILL }},
    { &hf_nbap_Common_E_DCH_MACdFlow_Specific_InfoListLCR_item,
      { "Common-E-DCH-MACdFlow-Specific-InfoList-ItemLCR", "nbap.Common_E_DCH_MACdFlow_Specific_InfoList_ItemLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_common_e_DCH_MACdFlow_ID_01,
      { "common-e-DCH-MACdFlow-ID", "nbap.common_e_DCH_MACdFlow_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_MACdFlow_ID_LCR", HFILL }},
    { &hf_nbap_eDCH_HARQ_PO_TDD,
      { "eDCH-HARQ-PO-TDD", "nbap.eDCH_HARQ_PO_TDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_HARQ_PO_TDD", HFILL }},
    { &hf_nbap_eDCH_MACdFlow_Retransmission_Timer,
      { "eDCH-MACdFlow-Retransmission-Timer", "nbap.eDCH_MACdFlow_Retransmission_Timer",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_E_DCH_MACdFlow_Retransmission_Timer_vals_ext, 0,
        "E_DCH_MACdFlow_Retransmission_Timer", HFILL }},
    { &hf_nbap_ul_common_E_DCH_MACflow_Specific_InfoResponseLCR,
      { "ul-common-E-DCH-MACflow-Specific-InfoResponseLCR", "nbap.ul_common_E_DCH_MACflow_Specific_InfoResponseLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR", HFILL }},
    { &hf_nbap_common_E_AGCH_ListLCR,
      { "common-E-AGCH-ListLCR", "nbap.common_E_AGCH_ListLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_common_E_HICH_ListLCR,
      { "common-E-HICH-ListLCR", "nbap.common_E_HICH_ListLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_common_E_RNTI_Info_LCR,
      { "common-E-RNTI-Info-LCR", "nbap.common_E_RNTI_Info_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_item,
      { "Ul-common-E-DCH-MACflow-Specific-InfoResponseList-ItemLCR", "nbap.Ul_common_E_DCH_MACflow_Specific_InfoResponseList_ItemLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_Ext_item,
      { "Ul-common-E-DCH-MACflow-Specific-InfoResponseList-ItemLCR", "nbap.Ul_common_E_DCH_MACflow_Specific_InfoResponseList_ItemLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ul_Common_MACFlowID_LCR,
      { "ul-Common-MACFlowID-LCR", "nbap.ul_Common_MACFlowID_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Common_MACFlow_ID_LCR", HFILL }},
    { &hf_nbap_Common_E_AGCH_ListLCR_item,
      { "Common-E-AGCH-ItemLCR", "nbap.Common_E_AGCH_ItemLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_AGCH_ID,
      { "e-AGCH-ID", "nbap.e_AGCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Common_E_HICH_ListLCR_item,
      { "Common-E-HICH-ItemLCR", "nbap.Common_E_HICH_ItemLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_eI,
      { "eI", "nbap.eI",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_HICH_ID,
      { "e-HICH-ID", "nbap.e_HICH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_HICH_ID_LCR", HFILL }},
    { &hf_nbap_Common_E_RNTI_Info_LCR_item,
      { "Common-E-RNTI-Info-ItemLCR", "nbap.Common_E_RNTI_Info_ItemLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_starting_E_RNTI,
      { "starting-E-RNTI", "nbap.starting_E_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_RNTI", HFILL }},
    { &hf_nbap_number_of_Group,
      { "number-of-Group", "nbap.number_of_Group",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_32", HFILL }},
    { &hf_nbap_number_of_e_E_RNTI_perGroup,
      { "number-of-e-E-RNTI-perGroup", "nbap.number_of_e_E_RNTI_perGroup",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_7", HFILL }},
    { &hf_nbap_Common_MACFlows_to_DeleteLCR_item,
      { "Common-MACFlows-to-DeleteLCR-Item", "nbap.Common_MACFlows_to_DeleteLCR_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_common_MACFlow_ID_LCR,
      { "common-MACFlow-ID-LCR", "nbap.common_MACFlow_ID_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonMACFlow_Specific_InfoListLCR_item,
      { "CommonMACFlow-Specific-InfoItemLCR", "nbap.CommonMACFlow_Specific_InfoItemLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_common_MACFlow_PriorityQueue_InformationLCR,
      { "common-MACFlow-PriorityQueue-InformationLCR", "nbap.common_MACFlow_PriorityQueue_InformationLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Common_MACFlow_PriorityQueue_Information", HFILL }},
    { &hf_nbap_Common_H_RNTI_InformationLCR_item,
      { "Common-H-RNTI-InfoItemLCR", "nbap.Common_H_RNTI_InfoItemLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_common_H_RNTI,
      { "common-H-RNTI", "nbap.common_H_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HSDSCH_RNTI", HFILL }},
    { &hf_nbap_t_SYNC,
      { "t-SYNC", "nbap.t_SYNC",
        FT_UINT32, BASE_DEC, VALS(nbap_T_SYNC_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_t_PROTECT,
      { "t-PROTECT", "nbap.t_PROTECT",
        FT_UINT32, BASE_DEC, VALS(nbap_T_PROTECT_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_n_PROTECT,
      { "n-PROTECT", "nbap.n_PROTECT",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommonMACFlow_Specific_InfoList_ResponseLCR_item,
      { "CommonMACFlow-Specific-InfoItem-ResponseLCR", "nbap.CommonMACFlow_Specific_InfoItem_ResponseLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_continuousPacketConnectivity_DRX_InformationLCR,
      { "continuousPacketConnectivity-DRX-InformationLCR", "nbap.continuousPacketConnectivity_DRX_InformationLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_continuousPacketConnectivity_DRX_Information_to_Modify_LCR,
      { "continuousPacketConnectivity-DRX-Information-to-Modify-LCR", "nbap.continuousPacketConnectivity_DRX_Information_to_Modify_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_DSCH_Semi_PersistentScheduling_Information_LCR,
      { "hS-DSCH-Semi-PersistentScheduling-Information-LCR", "nbap.hS_DSCH_Semi_PersistentScheduling_Information_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR,
      { "hS-DSCH-Semi-PersistentScheduling-Information-to-Modify-LCR", "nbap.hS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_DSCH_SPS_Deactivate_Indicator_LCR,
      { "hS-DSCH-SPS-Deactivate-Indicator-LCR", "nbap.hS_DSCH_SPS_Deactivate_Indicator_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_Semi_PersistentScheduling_Information_LCR,
      { "e-DCH-Semi-PersistentScheduling-Information-LCR", "nbap.e_DCH_Semi_PersistentScheduling_Information_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR,
      { "e-DCH-Semi-PersistentScheduling-Information-to-Modify-LCR", "nbap.e_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_SPS_Deactivate_Indicator_LCR,
      { "e-DCH-SPS-Deactivate-Indicator-LCR", "nbap.e_DCH_SPS_Deactivate_Indicator_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_SCCH_DRX_Information_LCR,
      { "hS-SCCH-DRX-Information-LCR", "nbap.hS_SCCH_DRX_Information_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_AGCH_DRX_Information_LCR,
      { "e-AGCH-DRX-Information-LCR", "nbap.e_AGCH_DRX_Information_LCR",
        FT_UINT32, BASE_DEC, VALS(nbap_E_AGCH_DRX_Information_LCR_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_hS_SCCH_UE_DRX_Cycle_LCR,
      { "hS-SCCH-UE-DRX-Cycle-LCR", "nbap.hS_SCCH_UE_DRX_Cycle_LCR",
        FT_UINT32, BASE_DEC, VALS(nbap_UE_DRX_Cycle_LCR_vals), 0,
        "UE_DRX_Cycle_LCR", HFILL }},
    { &hf_nbap_hS_SCCH_Inactivity_Threshold_for_UE_DRX_Cycle_LCR,
      { "hS-SCCH-Inactivity-Threshold-for-UE-DRX-Cycle-LCR", "nbap.hS_SCCH_Inactivity_Threshold_for_UE_DRX_Cycle_LCR",
        FT_UINT32, BASE_DEC, VALS(nbap_Inactivity_Threshold_for_UE_DRX_Cycle_LCR_vals), 0,
        "Inactivity_Threshold_for_UE_DRX_Cycle_LCR", HFILL }},
    { &hf_nbap_hS_SCCH_UE_DRX_Offset_LCR,
      { "hS-SCCH-UE-DRX-Offset-LCR", "nbap.hS_SCCH_UE_DRX_Offset_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UE_DRX_Offset_LCR", HFILL }},
    { &hf_nbap_sameAsHS_SCCH,
      { "sameAsHS-SCCH", "nbap.sameAsHS_SCCH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_AGCH_DRX_Parameters,
      { "e-AGCH-DRX-Parameters", "nbap.e_AGCH_DRX_Parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_AGCH_UE_DRX_Cycle_LCR,
      { "e-AGCH-UE-DRX-Cycle-LCR", "nbap.e_AGCH_UE_DRX_Cycle_LCR",
        FT_UINT32, BASE_DEC, VALS(nbap_UE_DRX_Cycle_LCR_vals), 0,
        "UE_DRX_Cycle_LCR", HFILL }},
    { &hf_nbap_e_AGCH_UE_Inactivity_Monitor_Threshold,
      { "e-AGCH-UE-Inactivity-Monitor-Threshold", "nbap.e_AGCH_UE_Inactivity_Monitor_Threshold",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_E_AGCH_UE_Inactivity_Monitor_Threshold_vals_ext, 0,
        NULL, HFILL }},
    { &hf_nbap_e_AGCH_UE_DRX_Offset_LCR,
      { "e-AGCH-UE-DRX-Offset-LCR", "nbap.e_AGCH_UE_DRX_Offset_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UE_DRX_Offset_LCR", HFILL }},
    { &hf_nbap_dRX_Information_to_Modify_LCR,
      { "dRX-Information-to-Modify-LCR", "nbap.dRX_Information_to_Modify_LCR",
        FT_UINT32, BASE_DEC, VALS(nbap_DRX_Information_to_Modify_LCR_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_modify,
      { "modify", "nbap.modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DRX_Information_to_Modify_Items_LCR", HFILL }},
    { &hf_nbap_deactivate,
      { "deactivate", "nbap.deactivate_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_SCCH_DRX_Information_ResponseLCR,
      { "hS-SCCH-DRX-Information-ResponseLCR", "nbap.hS_SCCH_DRX_Information_ResponseLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_AGCH_DRX_Information_ResponseLCR,
      { "e-AGCH-DRX-Information-ResponseLCR", "nbap.e_AGCH_DRX_Information_ResponseLCR",
        FT_UINT32, BASE_DEC, VALS(nbap_E_AGCH_DRX_Information_ResponseLCR_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_e_AGCH_DRX_Parameters_Response,
      { "e-AGCH-DRX-Parameters-Response", "nbap.e_AGCH_DRX_Parameters_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DCH_FDD_Information_item,
      { "DCH-FDD-InformationItem", "nbap.DCH_FDD_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ul_FP_Mode,
      { "ul-FP-Mode", "nbap.ul_FP_Mode",
        FT_UINT32, BASE_DEC, VALS(nbap_UL_FP_Mode_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_toAWS,
      { "toAWS", "nbap.toAWS",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_toAWE,
      { "toAWE", "nbap.toAWE",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dCH_SpecificInformationList,
      { "dCH-SpecificInformationList", "nbap.dCH_SpecificInformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DCH_Specific_FDD_InformationList", HFILL }},
    { &hf_nbap_DCH_Specific_FDD_InformationList_item,
      { "DCH-Specific-FDD-Item", "nbap.DCH_Specific_FDD_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dCH_ID,
      { "dCH-ID", "nbap.dCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ul_TransportFormatSet,
      { "ul-TransportFormatSet", "nbap.ul_TransportFormatSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dl_TransportFormatSet,
      { "dl-TransportFormatSet", "nbap.dl_TransportFormatSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_allocationRetentionPriority,
      { "allocationRetentionPriority", "nbap.allocationRetentionPriority_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_frameHandlingPriority,
      { "frameHandlingPriority", "nbap.frameHandlingPriority",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_qE_Selector,
      { "qE-Selector", "nbap.qE_Selector",
        FT_UINT32, BASE_DEC, VALS(nbap_QE_Selector_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_DCH_InformationResponse_item,
      { "DCH-InformationResponseItem", "nbap.DCH_InformationResponseItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dCH_ID_01,
      { "dCH-ID", "nbap.dCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DCH_MeasurementOccasion_Information_item,
      { "DchMeasurementOccasionInformation-Item", "nbap.DchMeasurementOccasionInformation_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_pattern_Sequence_Identifier,
      { "pattern-Sequence-Identifier", "nbap.pattern_Sequence_Identifier",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_status_Flag,
      { "status-Flag", "nbap.status_Flag",
        FT_UINT32, BASE_DEC, VALS(nbap_Status_Flag_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_measurement_Occasion_Pattern_Sequence_parameters,
      { "measurement-Occasion-Pattern-Sequence-parameters", "nbap.measurement_Occasion_Pattern_Sequence_parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_measurement_Occasion_Pattern_Sequence_parameters_k,
      { "measurement-Occasion-Pattern-Sequence-parameters-k", "nbap.measurement_Occasion_Pattern_Sequence_parameters_k",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_9", HFILL }},
    { &hf_nbap_measurement_Occasion_Pattern_Sequence_parameters_offset,
      { "measurement-Occasion-Pattern-Sequence-parameters-offset", "nbap.measurement_Occasion_Pattern_Sequence_parameters_offset",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_511", HFILL }},
    { &hf_nbap_measurement_Occasion_Pattern_Sequence_parameters_M_Length,
      { "measurement-Occasion-Pattern-Sequence-parameters-M-Length", "nbap.measurement_Occasion_Pattern_Sequence_parameters_M_Length",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_512", HFILL }},
    { &hf_nbap_measurement_Occasion_Pattern_Sequence_parameters_Timeslot_Bitmap,
      { "measurement-Occasion-Pattern-Sequence-parameters-Timeslot-Bitmap", "nbap.measurement_Occasion_Pattern_Sequence_parameters_Timeslot_Bitmap",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_7", HFILL }},
    { &hf_nbap_DCH_TDD_Information_item,
      { "DCH-TDD-InformationItem", "nbap.DCH_TDD_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dCH_SpecificInformationList_01,
      { "dCH-SpecificInformationList", "nbap.dCH_SpecificInformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DCH_Specific_TDD_InformationList", HFILL }},
    { &hf_nbap_DCH_Specific_TDD_InformationList_item,
      { "DCH-Specific-TDD-Item", "nbap.DCH_Specific_TDD_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ul_CCTrCH_ID,
      { "ul-CCTrCH-ID", "nbap.ul_CCTrCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CCTrCH_ID", HFILL }},
    { &hf_nbap_dl_CCTrCH_ID,
      { "dl-CCTrCH-ID", "nbap.dl_CCTrCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CCTrCH_ID", HFILL }},
    { &hf_nbap_ul_TransportFormatSet_01,
      { "ul-TransportFormatSet", "nbap.ul_TransportFormatSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "TransportFormatSet", HFILL }},
    { &hf_nbap_dl_TransportFormatSet_01,
      { "dl-TransportFormatSet", "nbap.dl_TransportFormatSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "TransportFormatSet", HFILL }},
    { &hf_nbap_FDD_DCHs_to_Modify_item,
      { "FDD-DCHs-to-ModifyItem", "nbap.FDD_DCHs_to_ModifyItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dCH_SpecificInformationList_02,
      { "dCH-SpecificInformationList", "nbap.dCH_SpecificInformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DCH_ModifySpecificInformation_FDD", HFILL }},
    { &hf_nbap_DCH_ModifySpecificInformation_FDD_item,
      { "DCH-ModifySpecificItem-FDD", "nbap.DCH_ModifySpecificItem_FDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dCH_ID_02,
      { "dCH-ID", "nbap.dCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "T_dCH_ID_01", HFILL }},
    { &hf_nbap_ul_TransportFormatSet_02,
      { "ul-TransportFormatSet", "nbap.ul_TransportFormatSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "T_ul_TransportFormatSet_01", HFILL }},
    { &hf_nbap_dl_TransportFormatSet_02,
      { "dl-TransportFormatSet", "nbap.dl_TransportFormatSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "T_dl_TransportFormatSet_01", HFILL }},
    { &hf_nbap_TDD_DCHs_to_Modify_item,
      { "DCH-ModifyItem-TDD", "nbap.DCH_ModifyItem_TDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dCH_SpecificInformationList_03,
      { "dCH-SpecificInformationList", "nbap.dCH_SpecificInformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DCH_ModifySpecificInformation_TDD", HFILL }},
    { &hf_nbap_DCH_ModifySpecificInformation_TDD_item,
      { "DCH-ModifySpecificItem-TDD", "nbap.DCH_ModifySpecificItem_TDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DedicatedChannelsCapacityConsumptionLaw_item,
      { "DedicatedChannelsCapacityConsumptionLaw item", "nbap.DedicatedChannelsCapacityConsumptionLaw_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dl_Cost_1,
      { "dl-Cost-1", "nbap.dl_Cost_1",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_65535", HFILL }},
    { &hf_nbap_dl_Cost_2,
      { "dl-Cost-2", "nbap.dl_Cost_2",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_65535", HFILL }},
    { &hf_nbap_ul_Cost_1,
      { "ul-Cost-1", "nbap.ul_Cost_1",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_65535", HFILL }},
    { &hf_nbap_ul_Cost_2,
      { "ul-Cost-2", "nbap.ul_Cost_2",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_65535", HFILL }},
    { &hf_nbap_sIR_Value,
      { "sIR-Value", "nbap.sIR_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sIR_ErrorValue,
      { "sIR-ErrorValue", "nbap.sIR_ErrorValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SIR_Error_Value", HFILL }},
    { &hf_nbap_transmittedCodePowerValue,
      { "transmittedCodePowerValue", "nbap.transmittedCodePowerValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Transmitted_Code_Power_Value", HFILL }},
    { &hf_nbap_rSCP,
      { "rSCP", "nbap.rSCP",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RSCP_Value", HFILL }},
    { &hf_nbap_rxTimingDeviationValue,
      { "rxTimingDeviationValue", "nbap.rxTimingDeviationValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Rx_Timing_Deviation_Value", HFILL }},
    { &hf_nbap_roundTripTime,
      { "roundTripTime", "nbap.roundTripTime",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Round_Trip_Time_Value", HFILL }},
    { &hf_nbap_extension_DedicatedMeasurementValue,
      { "extension-DedicatedMeasurementValue", "nbap.extension_DedicatedMeasurementValue_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_measurementAvailable_01,
      { "measurementAvailable", "nbap.measurementAvailable_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DedicatedMeasurementAvailable", HFILL }},
    { &hf_nbap_measurementnotAvailable_01,
      { "measurementnotAvailable", "nbap.measurementnotAvailable_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DedicatedMeasurementnotAvailable", HFILL }},
    { &hf_nbap_dedicatedmeasurementValue,
      { "dedicatedmeasurementValue", "nbap.dedicatedmeasurementValue",
        FT_UINT32, BASE_DEC, VALS(nbap_DedicatedMeasurementValue_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_cFN,
      { "cFN", "nbap.cFN",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cfn,
      { "cfn", "nbap.cfn",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_separate_indication,
      { "separate-indication", "nbap.separate_indication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_activate,
      { "activate", "nbap.activate_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Activate_Info", HFILL }},
    { &hf_nbap_deactivate_01,
      { "deactivate", "nbap.deactivate_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Deactivate_Info", HFILL }},
    { &hf_nbap_activation_type,
      { "activation-type", "nbap.activation_type",
        FT_UINT32, BASE_DEC, VALS(nbap_Execution_Type_vals), 0,
        "Execution_Type", HFILL }},
    { &hf_nbap_initial_dl_tx_power,
      { "initial-dl-tx-power", "nbap.initial_dl_tx_power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_firstRLS_Indicator,
      { "firstRLS-Indicator", "nbap.firstRLS_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_FirstRLS_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_propagation_delay,
      { "propagation-delay", "nbap.propagation_delay",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PropagationDelay", HFILL }},
    { &hf_nbap_deactivation_type,
      { "deactivation-type", "nbap.deactivation_type",
        FT_UINT32, BASE_DEC, VALS(nbap_Execution_Type_vals), 0,
        "Execution_Type", HFILL }},
    { &hf_nbap_synchronised,
      { "synchronised", "nbap.synchronised",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CFN", HFILL }},
    { &hf_nbap_unsynchronised,
      { "unsynchronised", "nbap.unsynchronised_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dGANSS_ReferenceTime,
      { "dGANSS-ReferenceTime", "nbap.dGANSS_ReferenceTime",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_119", HFILL }},
    { &hf_nbap_dGANSS_Information,
      { "dGANSS-Information", "nbap.dGANSS_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dGANSS_Signal_ID,
      { "dGANSS-Signal-ID", "nbap.dGANSS_Signal_ID",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_DGANSS_Information_item,
      { "DGANSS-InformationItem", "nbap.DGANSS_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_gANSS_SignalId,
      { "gANSS-SignalId", "nbap.gANSS_SignalId",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_Signal_ID", HFILL }},
    { &hf_nbap_gANSS_StatusHealth,
      { "gANSS-StatusHealth", "nbap.gANSS_StatusHealth",
        FT_UINT32, BASE_DEC, VALS(nbap_GANSS_StatusHealth_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_dGANSS_SignalInformation,
      { "dGANSS-SignalInformation", "nbap.dGANSS_SignalInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DGANSS_SignalInformation_item,
      { "DGANSS-SignalInformationItem", "nbap.DGANSS_SignalInformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_satId,
      { "satId", "nbap.satId",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_63", HFILL }},
    { &hf_nbap_gANSS_iod,
      { "gANSS-iod", "nbap.gANSS_iod",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_nbap_udre,
      { "udre", "nbap.udre",
        FT_UINT32, BASE_DEC, VALS(nbap_UDRE_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_ganss_prc,
      { "ganss-prc", "nbap.ganss_prc",
        FT_INT32, BASE_DEC, NULL, 0,
        "INTEGER_M2047_2047", HFILL }},
    { &hf_nbap_ganss_rrc,
      { "ganss-rrc", "nbap.ganss_rrc",
        FT_INT32, BASE_DEC, NULL, 0,
        "INTEGER_M127_127", HFILL }},
    { &hf_nbap_pRCDeviation,
      { "pRCDeviation", "nbap.pRCDeviation",
        FT_UINT32, BASE_DEC, VALS(nbap_PRCDeviation_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_udreGrowthRate,
      { "udreGrowthRate", "nbap.udreGrowthRate",
        FT_UINT32, BASE_DEC, VALS(nbap_UDREGrowthRate_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_udreValidityTime,
      { "udreValidityTime", "nbap.udreValidityTime",
        FT_UINT32, BASE_DEC, VALS(nbap_UDREValidityTime_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_gpstow,
      { "gpstow", "nbap.gpstow",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_status_health,
      { "status-health", "nbap.status_health",
        FT_UINT32, BASE_DEC, VALS(nbap_GPS_Status_Health_vals), 0,
        "GPS_Status_Health", HFILL }},
    { &hf_nbap_satelliteinfo,
      { "satelliteinfo", "nbap.satelliteinfo",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SAT_Info_DGPSCorrections", HFILL }},
    { &hf_nbap_prcdeviation,
      { "prcdeviation", "nbap.prcdeviation",
        FT_UINT32, BASE_DEC, VALS(nbap_PRCDeviation_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_DL_Timeslot_Information_item,
      { "DL-Timeslot-InformationItem", "nbap.DL_Timeslot_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_timeSlot,
      { "timeSlot", "nbap.timeSlot",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_midambleShiftAndBurstType,
      { "midambleShiftAndBurstType", "nbap.midambleShiftAndBurstType",
        FT_UINT32, BASE_DEC, VALS(nbap_MidambleShiftAndBurstType_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_tFCI_Presence,
      { "tFCI-Presence", "nbap.tFCI_Presence",
        FT_UINT32, BASE_DEC, VALS(nbap_TFCI_Presence_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_dL_Code_Information,
      { "dL-Code-Information", "nbap.dL_Code_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TDD_DL_Code_Information", HFILL }},
    { &hf_nbap_DL_TimeslotLCR_Information_item,
      { "DL-TimeslotLCR-InformationItem", "nbap.DL_TimeslotLCR_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_timeSlotLCR,
      { "timeSlotLCR", "nbap.timeSlotLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_midambleShiftLCR,
      { "midambleShiftLCR", "nbap.midambleShiftLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_Code_LCR_Information,
      { "dL-Code-LCR-Information", "nbap.dL_Code_LCR_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TDD_DL_Code_LCR_Information", HFILL }},
    { &hf_nbap_DL_Timeslot768_Information_item,
      { "DL-Timeslot768-InformationItem", "nbap.DL_Timeslot768_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_midambleShiftAndBurstType768,
      { "midambleShiftAndBurstType768", "nbap.midambleShiftAndBurstType768",
        FT_UINT32, BASE_DEC, VALS(nbap_MidambleShiftAndBurstType768_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_dL_Code_768_Information,
      { "dL-Code-768-Information", "nbap.dL_Code_768_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TDD_DL_Code_768_Information", HFILL }},
    { &hf_nbap_powerAdjustmentType,
      { "powerAdjustmentType", "nbap.powerAdjustmentType",
        FT_UINT32, BASE_DEC, VALS(nbap_PowerAdjustmentType_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_dLReferencePower,
      { "dLReferencePower", "nbap.dLReferencePower",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_dLReferencePowerList_DL_PC_Rqst,
      { "dLReferencePowerList-DL-PC-Rqst", "nbap.dLReferencePowerList_DL_PC_Rqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DL_ReferencePowerInformationList", HFILL }},
    { &hf_nbap_maxAdjustmentStep,
      { "maxAdjustmentStep", "nbap.maxAdjustmentStep",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_adjustmentPeriod,
      { "adjustmentPeriod", "nbap.adjustmentPeriod",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_adjustmentRatio,
      { "adjustmentRatio", "nbap.adjustmentRatio",
        FT_UINT32, BASE_DEC, NULL, 0,
        "ScaledAdjustmentRatio", HFILL }},
    { &hf_nbap_DL_ReferencePowerInformationList_item,
      { "DL-ReferencePowerInformationItem", "nbap.DL_ReferencePowerInformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_rL_ID,
      { "rL-ID", "nbap.rL_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dl_Reference_Power,
      { "dl-Reference-Power", "nbap.dl_Reference_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_DL_TimeslotISCPInfo_item,
      { "DL-TimeslotISCPInfoItem", "nbap.DL_TimeslotISCPInfoItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_TimeslotISCP,
      { "dL-TimeslotISCP", "nbap.dL_TimeslotISCP",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_TimeslotISCPInfoLCR_item,
      { "DL-TimeslotISCPInfoItemLCR", "nbap.DL_TimeslotISCPInfoItemLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst_item,
      { "DL-HS-PDSCH-Timeslot-InformationItem-LCR-PSCH-ReconfRqst", "nbap.DL_HS_PDSCH_Timeslot_InformationItem_LCR_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_midambleShiftAndBurstType_01,
      { "midambleShiftAndBurstType", "nbap.midambleShiftAndBurstType_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "MidambleShiftLCR", HFILL }},
    { &hf_nbap_dl_HS_PDSCH_Codelist_LCR_PSCH_ReconfRqst,
      { "dl-HS-PDSCH-Codelist-LCR-PSCH-ReconfRqst", "nbap.dl_HS_PDSCH_Codelist_LCR_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_maxHSDSCH_HSSCCH_Power,
      { "maxHSDSCH-HSSCCH-Power", "nbap.maxHSDSCH_HSSCCH_Power",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MaximumTransmissionPower", HFILL }},
    { &hf_nbap_MaxHSDSCH_HSSCCH_Power_per_CELLPORTION_item,
      { "MaxHSDSCH-HSSCCH-Power-per-CELLPORTION-Item", "nbap.MaxHSDSCH_HSSCCH_Power_per_CELLPORTION_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_HS_PDSCH_Codelist_LCR_PSCH_ReconfRqst_item,
      { "TDD-ChannelisationCode", "nbap.TDD_ChannelisationCode",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_TDD_ChannelisationCode_vals_ext, 0,
        NULL, HFILL }},
    { &hf_nbap_uE_DRX_Cycle,
      { "uE-DRX-Cycle", "nbap.uE_DRX_Cycle",
        FT_UINT32, BASE_DEC, VALS(nbap_UE_DRX_Cycle_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_inactivity_Threshold_for_UE_DRX_Cycle,
      { "inactivity-Threshold-for-UE-DRX-Cycle", "nbap.inactivity_Threshold_for_UE_DRX_Cycle",
        FT_UINT32, BASE_DEC, VALS(nbap_Inactivity_Threshold_for_UE_DRX_Cycle_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_inactivity_Threshold_for_UE_Grant_Monitoring,
      { "inactivity-Threshold-for-UE-Grant-Monitoring", "nbap.inactivity_Threshold_for_UE_Grant_Monitoring",
        FT_UINT32, BASE_DEC, VALS(nbap_Inactivity_Threshold_for_UE_Grant_Monitoring_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_uE_DRX_Grant_Monitoring,
      { "uE-DRX-Grant-Monitoring", "nbap.uE_DRX_Grant_Monitoring",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_modify_01,
      { "modify", "nbap.modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DRX_Information_to_Modify_Items", HFILL }},
    { &hf_nbap_DSCH_InformationResponse_item,
      { "DSCH-InformationResponseItem", "nbap.DSCH_InformationResponseItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dSCH_ID,
      { "dSCH-ID", "nbap.dSCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DSCH_TDD_Information_item,
      { "DSCH-TDD-InformationItem", "nbap.DSCH_TDD_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cCTrCH_ID,
      { "cCTrCH-ID", "nbap.cCTrCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_transportFormatSet,
      { "transportFormatSet", "nbap.transportFormatSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uE_DTX_Cycle1_2ms,
      { "uE-DTX-Cycle1-2ms", "nbap.uE_DTX_Cycle1_2ms",
        FT_UINT32, BASE_DEC, VALS(nbap_UE_DTX_Cycle1_2ms_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_uE_DTX_Cycle2_2ms,
      { "uE-DTX-Cycle2-2ms", "nbap.uE_DTX_Cycle2_2ms",
        FT_UINT32, BASE_DEC, VALS(nbap_UE_DTX_Cycle2_2ms_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_mAC_DTX_Cycle_2ms,
      { "mAC-DTX-Cycle-2ms", "nbap.mAC_DTX_Cycle_2ms",
        FT_UINT32, BASE_DEC, VALS(nbap_MAC_DTX_Cycle_2ms_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_uE_DTX_Cycle1_10ms,
      { "uE-DTX-Cycle1-10ms", "nbap.uE_DTX_Cycle1_10ms",
        FT_UINT32, BASE_DEC, VALS(nbap_UE_DTX_Cycle1_10ms_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_uE_DTX_Cycle2_10ms,
      { "uE-DTX-Cycle2-10ms", "nbap.uE_DTX_Cycle2_10ms",
        FT_UINT32, BASE_DEC, VALS(nbap_UE_DTX_Cycle2_10ms_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_mAC_DTX_Cycle_10ms,
      { "mAC-DTX-Cycle-10ms", "nbap.mAC_DTX_Cycle_10ms",
        FT_UINT32, BASE_DEC, VALS(nbap_MAC_DTX_Cycle_10ms_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_TTI_Length,
      { "e-DCH-TTI-Length", "nbap.e_DCH_TTI_Length",
        FT_UINT32, BASE_DEC, VALS(nbap_E_DCH_TTI_Length_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_inactivity_Threshold_for_UE_DTX_Cycle2,
      { "inactivity-Threshold-for-UE-DTX-Cycle2", "nbap.inactivity_Threshold_for_UE_DTX_Cycle2",
        FT_UINT32, BASE_DEC, VALS(nbap_Inactivity_Threshold_for_UE_DTX_Cycle2_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_uE_DTX_Long_Preamble,
      { "uE-DTX-Long-Preamble", "nbap.uE_DTX_Long_Preamble",
        FT_UINT32, BASE_DEC, VALS(nbap_UE_DTX_Long_Preamble_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_mAC_Inactivity_Threshold,
      { "mAC-Inactivity-Threshold", "nbap.mAC_Inactivity_Threshold",
        FT_UINT32, BASE_DEC, VALS(nbap_MAC_Inactivity_Threshold_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_cQI_DTX_Timer,
      { "cQI-DTX-Timer", "nbap.cQI_DTX_Timer",
        FT_UINT32, BASE_DEC, VALS(nbap_CQI_DTX_Timer_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_uE_DPCCH_burst1,
      { "uE-DPCCH-burst1", "nbap.uE_DPCCH_burst1",
        FT_UINT32, BASE_DEC, VALS(nbap_UE_DPCCH_burst1_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_uE_DPCCH_burst2,
      { "uE-DPCCH-burst2", "nbap.uE_DPCCH_burst2",
        FT_UINT32, BASE_DEC, VALS(nbap_UE_DPCCH_burst2_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_modify_02,
      { "modify", "nbap.modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DTX_Information_to_Modify_Items", HFILL }},
    { &hf_nbap_e_DCH_TTI_Length_to_Modify,
      { "e-DCH-TTI-Length-to-Modify", "nbap.e_DCH_TTI_Length_to_Modify",
        FT_UINT32, BASE_DEC, VALS(nbap_E_DCH_TTI_Length_to_Modify_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_dual_Band_Capability,
      { "dual-Band-Capability", "nbap.dual_Band_Capability",
        FT_UINT32, BASE_DEC, VALS(nbap_Dual_Band_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_possible_Secondary_Serving_Cell_List,
      { "possible-Secondary-Serving-Cell-List", "nbap.possible_Secondary_Serving_Cell_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_replace,
      { "replace", "nbap.replace",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_AGCH_FDD_Code_List", HFILL }},
    { &hf_nbap_remove,
      { "remove", "nbap.remove_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_AGCH_FDD_Code_List_item,
      { "FDD-DL-ChannelisationCodeNumber", "nbap.FDD_DL_ChannelisationCodeNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_SF_allocation,
      { "e-DCH-SF-allocation", "nbap.e_DCH_SF_allocation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_SF_allocation_item,
      { "E-DCH-SF-allocation item", "nbap.E_DCH_SF_allocation_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_AGCH_And_E_RGCH_E_HICH_FDD_Scrambling_Code,
      { "e-AGCH-And-E-RGCH-E-HICH-FDD-Scrambling-Code", "nbap.e_AGCH_And_E_RGCH_E_HICH_FDD_Scrambling_Code",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DL_ScramblingCode", HFILL }},
    { &hf_nbap_e_AGCH_Channelisation_Code,
      { "e-AGCH-Channelisation-Code", "nbap.e_AGCH_Channelisation_Code",
        FT_UINT32, BASE_DEC, NULL, 0,
        "FDD_DL_ChannelisationCodeNumber", HFILL }},
    { &hf_nbap_primary_e_RNTI,
      { "primary-e-RNTI", "nbap.primary_e_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_RNTI", HFILL }},
    { &hf_nbap_secondary_e_RNTI,
      { "secondary-e-RNTI", "nbap.secondary_e_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_RNTI", HFILL }},
    { &hf_nbap_primary_Secondary_Grant_Selector,
      { "primary-Secondary-Grant-Selector", "nbap.primary_Secondary_Grant_Selector",
        FT_UINT32, BASE_DEC, VALS(nbap_E_Primary_Secondary_Grant_Selector_vals), 0,
        "E_Primary_Secondary_Grant_Selector", HFILL }},
    { &hf_nbap_e_RGCH_Release_Indicator,
      { "e-RGCH-Release-Indicator", "nbap.e_RGCH_Release_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_E_RGCH_Release_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_MACdFlows_Information,
      { "e-DCH-MACdFlows-Information", "nbap.e_DCH_MACdFlows_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_MACdFlow_Specific_InformationResp,
      { "e-DCH-MACdFlow-Specific-InformationResp", "nbap.e_DCH_MACdFlow_Specific_InformationResp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_MACdFlow_Specific_Info_to_Modify,
      { "e-DCH-MACdFlow-Specific-Info-to-Modify", "nbap.e_DCH_MACdFlow_Specific_Info_to_Modify",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_MACdFlow_Specific_InfoList_to_Modify", HFILL }},
    { &hf_nbap_mACeReset_Indicator,
      { "mACeReset-Indicator", "nbap.mACeReset_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_MACeReset_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_MACdFlow_Specific_UpdateInformation,
      { "e-DCH-MACdFlow-Specific-UpdateInformation", "nbap.e_DCH_MACdFlow_Specific_UpdateInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_MACdFlow_Specific_UpdateInformation_item,
      { "E-DCH-MACdFlow-Specific-UpdateInformation-Item", "nbap.E_DCH_MACdFlow_Specific_UpdateInformation_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hARQ_Process_Allocation_NonSched_2ms_EDCH,
      { "hARQ-Process-Allocation-NonSched-2ms-EDCH", "nbap.hARQ_Process_Allocation_NonSched_2ms_EDCH",
        FT_BYTES, BASE_NONE, NULL, 0,
        "HARQ_Process_Allocation_2ms_EDCH", HFILL }},
    { &hf_nbap_E_DCH_DL_Control_Channel_Change_Information_item,
      { "E-DCH-DL-Control-Channel-Change-Information-Item", "nbap.E_DCH_DL_Control_Channel_Change_Information_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_RL_ID,
      { "e-DCH-RL-ID", "nbap.e_DCH_RL_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RL_ID", HFILL }},
    { &hf_nbap_E_DCH_DL_Control_Channel_Grant_Information_item,
      { "E-DCH-DL-Control-Channel-Grant-Information-Item", "nbap.E_DCH_DL_Control_Channel_Grant_Information_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_Non_Scheduled_Transmission_Grant,
      { "e-DCH-Non-Scheduled-Transmission-Grant", "nbap.e_DCH_Non_Scheduled_Transmission_Grant_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "E_DCH_Non_Scheduled_Transmission_Grant_Items", HFILL }},
    { &hf_nbap_e_DCH_Scheduled_Transmission_Grant,
      { "e-DCH-Scheduled-Transmission-Grant", "nbap.e_DCH_Scheduled_Transmission_Grant_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_LogicalChannelInformation_item,
      { "E-DCH-LogicalChannelInformationItem", "nbap.E_DCH_LogicalChannelInformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_schedulingPriorityIndicator,
      { "schedulingPriorityIndicator", "nbap.schedulingPriorityIndicator",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_schedulingInformation,
      { "schedulingInformation", "nbap.schedulingInformation",
        FT_UINT32, BASE_DEC, VALS(nbap_SchedulingInformation_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_mACesGuaranteedBitRate,
      { "mACesGuaranteedBitRate", "nbap.mACesGuaranteedBitRate",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_DDI_Value,
      { "e-DCH-DDI-Value", "nbap.e_DCH_DDI_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_mACd_PDU_Size_List,
      { "mACd-PDU-Size-List", "nbap.mACd_PDU_Size_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_MACdPDU_SizeList", HFILL }},
    { &hf_nbap_E_DCH_MACdPDU_SizeList_item,
      { "E-DCH-MACdPDU-SizeListItem", "nbap.E_DCH_MACdPDU_SizeListItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_mACdPDU_Size,
      { "mACdPDU-Size", "nbap.mACdPDU_Size",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_LogicalChannelToModify_item,
      { "E-DCH-LogicalChannelToModifyItem", "nbap.E_DCH_LogicalChannelToModifyItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_mACd_PDU_Size_List_01,
      { "mACd-PDU-Size-List", "nbap.mACd_PDU_Size_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_MACdPDU_SizeToModifyList", HFILL }},
    { &hf_nbap_E_DCH_MACdPDU_SizeToModifyList_item,
      { "E-DCH-MACdPDU-SizeListItem", "nbap.E_DCH_MACdPDU_SizeListItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_LogicalChannelToDelete_item,
      { "E-DCH-LogicalChannelToDeleteItem", "nbap.E_DCH_LogicalChannelToDeleteItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_MACdFlow_Specific_Info,
      { "e-DCH-MACdFlow-Specific-Info", "nbap.e_DCH_MACdFlow_Specific_Info",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_MACdFlow_Specific_InfoList", HFILL }},
    { &hf_nbap_E_DCH_MACdFlow_Specific_InfoList_item,
      { "E-DCH-MACdFlow-Specific-InfoItem", "nbap.E_DCH_MACdFlow_Specific_InfoItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_eDCH_Grant_Type_Information,
      { "eDCH-Grant-Type-Information", "nbap.eDCH_Grant_Type_Information",
        FT_UINT32, BASE_DEC, VALS(nbap_E_DCH_Grant_Type_Information_vals), 0,
        "E_DCH_Grant_Type_Information", HFILL }},
    { &hf_nbap_eDCHLogicalChannelInformation,
      { "eDCHLogicalChannelInformation", "nbap.eDCHLogicalChannelInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_LogicalChannelInformation", HFILL }},
    { &hf_nbap_E_DCH_MACdFlow_Specific_InformationResp_item,
      { "E-DCH-MACdFlow-Specific-InformationResp-Item", "nbap.E_DCH_MACdFlow_Specific_InformationResp_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_MACdFlow_Specific_InfoList_to_Modify_item,
      { "E-DCH-MACdFlow-Specific-InfoItem-to-Modify", "nbap.E_DCH_MACdFlow_Specific_InfoItem_to_Modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_eDCH_LogicalChannelToAdd,
      { "eDCH-LogicalChannelToAdd", "nbap.eDCH_LogicalChannelToAdd",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_LogicalChannelInformation", HFILL }},
    { &hf_nbap_eDCH_LogicalChannelToModify,
      { "eDCH-LogicalChannelToModify", "nbap.eDCH_LogicalChannelToModify",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_LogicalChannelToModify", HFILL }},
    { &hf_nbap_eDCH_LogicalChannelToDelete,
      { "eDCH-LogicalChannelToDelete", "nbap.eDCH_LogicalChannelToDelete",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_LogicalChannelToDelete", HFILL }},
    { &hf_nbap_E_DCH_MACdFlows_to_Delete_item,
      { "E-DCH-MACdFlow-to-Delete-Item", "nbap.E_DCH_MACdFlow_to_Delete_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_maxBits_MACe_PDU_non_scheduled,
      { "maxBits-MACe-PDU-non-scheduled", "nbap.maxBits_MACe_PDU_non_scheduled",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Max_Bits_MACe_PDU_non_scheduled", HFILL }},
    { &hf_nbap_hARQ_Process_Allocation_NonSched_2ms,
      { "hARQ-Process-Allocation-NonSched-2ms", "nbap.hARQ_Process_Allocation_NonSched_2ms",
        FT_BYTES, BASE_NONE, NULL, 0,
        "HARQ_Process_Allocation_2ms_EDCH", HFILL }},
    { &hf_nbap_e_DCH_serving_cell_choice,
      { "e-DCH-serving-cell-choice", "nbap.e_DCH_serving_cell_choice",
        FT_UINT32, BASE_DEC, VALS(nbap_E_DCH_serving_cell_choice_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_serving_cell_change_successful,
      { "e-DCH-serving-cell-change-successful", "nbap.e_DCH_serving_cell_change_successful_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_serving_cell_change_unsuccessful,
      { "e-DCH-serving-cell-change-unsuccessful", "nbap.e_DCH_serving_cell_change_unsuccessful_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_RL_InformationList_Rsp,
      { "e-DCH-RL-InformationList-Rsp", "nbap.e_DCH_RL_InformationList_Rsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_RL_InformationList_Rsp_item,
      { "E-DCH-RL-InformationList-Rsp-Item", "nbap.E_DCH_RL_InformationList_Rsp_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_rl_ID,
      { "rl-ID", "nbap.rl_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_FDD_DL_Control_Channel_Info,
      { "e-DCH-FDD-DL-Control-Channel-Info", "nbap.e_DCH_FDD_DL_Control_Channel_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "E_DCH_FDD_DL_Control_Channel_Information", HFILL }},
    { &hf_nbap_cause,
      { "cause", "nbap.cause",
        FT_UINT32, BASE_DEC, VALS(nbap_Cause_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_EDCH_RACH_Report_Value_item,
      { "EDCH-RACH-Report-Value item", "nbap.EDCH_RACH_Report_Value_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_granted_EDCH_RACH_resources,
      { "granted-EDCH-RACH-resources", "nbap.granted_EDCH_RACH_resources",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Granted_EDCH_RACH_Resources_Value", HFILL }},
    { &hf_nbap_denied_EDCH_RACH_resources,
      { "denied-EDCH-RACH-resources", "nbap.denied_EDCH_RACH_resources",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Denied_EDCH_RACH_Resources_Value", HFILL }},
    { &hf_nbap_two_ms,
      { "two-ms", "nbap.two_ms_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DTX_Cycle_2ms_Items", HFILL }},
    { &hf_nbap_ten_ms,
      { "ten-ms", "nbap.ten_ms_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DTX_Cycle_10ms_Items", HFILL }},
    { &hf_nbap_two_ms_01,
      { "two-ms", "nbap.two_ms_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DTX_Cycle_2ms_to_Modify_Items", HFILL }},
    { &hf_nbap_ten_ms_01,
      { "ten-ms", "nbap.ten_ms_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DTX_Cycle_10ms_to_Modify_Items", HFILL }},
    { &hf_nbap_E_DCH_MACdFlows_to_DeleteLCR_item,
      { "E-DCH-MACdFlow-to-Delete-ItemLCR", "nbap.E_DCH_MACdFlow_to_Delete_ItemLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_MACdFlow_ID_LCR,
      { "e-DCH-MACdFlow-ID-LCR", "nbap.e_DCH_MACdFlow_ID_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_t321,
      { "t321", "nbap.t321",
        FT_UINT32, BASE_DEC, VALS(nbap_T321_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_hS_DSCH_DRX_Cycle_FACH,
      { "hS-DSCH-DRX-Cycle-FACH", "nbap.hS_DSCH_DRX_Cycle_FACH",
        FT_UINT32, BASE_DEC, VALS(nbap_HS_DSCH_DRX_Cycle_FACH_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_hS_DSCH_RX_Burst_FACH,
      { "hS-DSCH-RX-Burst-FACH", "nbap.hS_DSCH_RX_Burst_FACH",
        FT_UINT32, BASE_DEC, VALS(nbap_HS_DSCH_RX_Burst_FACH_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_replace_01,
      { "replace", "nbap.replace",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_RGCH_E_HICH_FDD_Code_List", HFILL }},
    { &hf_nbap_E_RGCH_E_HICH_FDD_Code_List_item,
      { "FDD-DL-ChannelisationCodeNumber", "nbap.FDD_DL_ChannelisationCodeNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_TFCI_BetaEC_Boost,
      { "e-TFCI-BetaEC-Boost", "nbap.e_TFCI_BetaEC_Boost",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_Delta_T2TP,
      { "uL-Delta-T2TP", "nbap.uL_Delta_T2TP",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_TFCI_Table_Index,
      { "e-DCH-TFCI-Table-Index", "nbap.e_DCH_TFCI_Table_Index",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_reference_E_TFCI_Information,
      { "reference-E-TFCI-Information", "nbap.reference_E_TFCI_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCHProvidedBitRate_item,
      { "E-DCHProvidedBitRate-Item", "nbap.E_DCHProvidedBitRate_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCHProvidedBitRateValue,
      { "e-DCHProvidedBitRateValue", "nbap.e_DCHProvidedBitRateValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCHProvidedBitRateValueInformation_For_CellPortion_item,
      { "E-DCHProvidedBitRateValueInformation-For-CellPortion-Item", "nbap.E_DCHProvidedBitRateValueInformation_For_CellPortion_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCHProvidedBitRateValue_01,
      { "e-DCHProvidedBitRateValue", "nbap.e_DCHProvidedBitRateValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCHProvidedBitRate", HFILL }},
    { &hf_nbap_e_PUCH_Information,
      { "e-PUCH-Information", "nbap.e_PUCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_MACdFlows_Information_TDD,
      { "e-DCH-MACdFlows-Information-TDD", "nbap.e_DCH_MACdFlows_Information_TDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_Non_Scheduled_Grant_Info,
      { "e-DCH-Non-Scheduled-Grant-Info", "nbap.e_DCH_Non_Scheduled_Grant_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_TDD_Information,
      { "e-DCH-TDD-Information", "nbap.e_DCH_TDD_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_n_E_UCCH,
      { "n-E-UCCH", "nbap.n_E_UCCH",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_QPSK_RefBetaInfo,
      { "e-DCH-QPSK-RefBetaInfo", "nbap.e_DCH_QPSK_RefBetaInfo",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_sixteenQAM_RefBetaInfo,
      { "e-DCH-sixteenQAM-RefBetaInfo", "nbap.e_DCH_sixteenQAM_RefBetaInfo",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_QPSK_RefBetaInfo_item,
      { "E-DCH-RefBeta-Item", "nbap.E_DCH_RefBeta_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_sixteenQAM_RefBetaInfo_item,
      { "E-DCH-RefBeta-Item", "nbap.E_DCH_RefBeta_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_refCodeRate,
      { "refCodeRate", "nbap.refCodeRate",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CodeRate_short", HFILL }},
    { &hf_nbap_refBeta,
      { "refBeta", "nbap.refBeta",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_MACdFlows_Information_TDD_item,
      { "E-DCH-MACdFlow-InfoTDDItem", "nbap.E_DCH_MACdFlow_InfoTDDItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_eDCH_Grant_TypeTDD,
      { "eDCH-Grant-TypeTDD", "nbap.eDCH_Grant_TypeTDD",
        FT_UINT32, BASE_DEC, VALS(nbap_E_DCH_Grant_TypeTDD_vals), 0,
        "E_DCH_Grant_TypeTDD", HFILL }},
    { &hf_nbap_timeslotResource,
      { "timeslotResource", "nbap.timeslotResource",
        FT_BYTES, BASE_NONE, NULL, 0,
        "E_DCH_TimeslotResource", HFILL }},
    { &hf_nbap_powerResource,
      { "powerResource", "nbap.powerResource",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_PowerResource", HFILL }},
    { &hf_nbap_repetitionPeriod,
      { "repetitionPeriod", "nbap.repetitionPeriod",
        FT_UINT32, BASE_DEC, VALS(nbap_RepetitionPeriod_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_repetitionLength,
      { "repetitionLength", "nbap.repetitionLength",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tddE_PUCH_Offset,
      { "tddE-PUCH-Offset", "nbap.tddE_PUCH_Offset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tdd_ChannelisationCode,
      { "tdd-ChannelisationCode", "nbap.tdd_ChannelisationCode",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_TDD_ChannelisationCode_vals_ext, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_TDD_Maximum_Bitrate,
      { "e-DCH-TDD-Maximum-Bitrate", "nbap.e_DCH_TDD_Maximum_Bitrate",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_TDD_MACdFlow_Specific_InformationResp,
      { "e-DCH-TDD-MACdFlow-Specific-InformationResp", "nbap.e_DCH_TDD_MACdFlow_Specific_InformationResp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_AGCH_Specific_Information_ResponseTDD,
      { "e-AGCH-Specific-Information-ResponseTDD", "nbap.e_AGCH_Specific_Information_ResponseTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_AGCH_Specific_InformationRespListTDD", HFILL }},
    { &hf_nbap_scheduled_E_HICH_Specific_InformationResp,
      { "scheduled-E-HICH-Specific-InformationResp", "nbap.scheduled_E_HICH_Specific_InformationResp",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Scheduled_E_HICH_Specific_Information_ResponseLCRTDD", HFILL }},
    { &hf_nbap_Scheduled_E_HICH_Specific_Information_ResponseLCRTDD_item,
      { "Scheduled-E-HICH-Specific-InformationItem-ResponseLCRTDD", "nbap.Scheduled_E_HICH_Specific_InformationItem_ResponseLCRTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_HICH_ID_TDD,
      { "e-HICH-ID-TDD", "nbap.e_HICH_ID_TDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_TDD_MACdFlow_Specific_InformationResp_item,
      { "E-DCH-TDD-MACdFlow-Specific-InformationResp-Item", "nbap.E_DCH_TDD_MACdFlow_Specific_InformationResp_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_MacdFlow_Id,
      { "e-DCH-MacdFlow-Id", "nbap.e_DCH_MacdFlow_Id",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_AGCH_Specific_InformationRespListTDD_item,
      { "E-AGCH-Specific-InformationResp-ItemTDD", "nbap.E_AGCH_Specific_InformationResp_ItemTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_AGCH_Id,
      { "e-AGCH-Id", "nbap.e_AGCH_Id",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_MACdFlows_to_Add,
      { "e-DCH-MACdFlows-to-Add", "nbap.e_DCH_MACdFlows_to_Add",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_MACdFlows_Information_TDD", HFILL }},
    { &hf_nbap_e_DCH_MACdFlows_to_Delete,
      { "e-DCH-MACdFlows-to-Delete", "nbap.e_DCH_MACdFlows_to_Delete",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_TDD_Information_to_Modify,
      { "e-DCH-TDD-Information-to-Modify", "nbap.e_DCH_TDD_Information_to_Modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_TDD_Information_to_Modify_List,
      { "e-DCH-TDD-Information-to-Modify-List", "nbap.e_DCH_TDD_Information_to_Modify_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_TDD_Information_to_Modify_List_item,
      { "E-DCH-MACdFlow-ModifyTDDItem", "nbap.E_DCH_MACdFlow_ModifyTDDItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_LogicalChannelToAdd,
      { "e-DCH-LogicalChannelToAdd", "nbap.e_DCH_LogicalChannelToAdd",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_LogicalChannelInformation", HFILL }},
    { &hf_nbap_e_DCH_LogicalChannelToModify,
      { "e-DCH-LogicalChannelToModify", "nbap.e_DCH_LogicalChannelToModify",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_LogicalChannelToDelete,
      { "e-DCH-LogicalChannelToDelete", "nbap.e_DCH_LogicalChannelToDelete",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_Non_Scheduled_Grant_Info768,
      { "e-DCH-Non-Scheduled-Grant-Info768", "nbap.e_DCH_Non_Scheduled_Grant_Info768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_TDD_Information768,
      { "e-DCH-TDD-Information768", "nbap.e_DCH_TDD_Information768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tdd_ChannelisationCode768,
      { "tdd-ChannelisationCode768", "nbap.tdd_ChannelisationCode768",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_TDD_ChannelisationCode768_vals_ext, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_TDD_Maximum_Bitrate768,
      { "e-DCH-TDD-Maximum-Bitrate768", "nbap.e_DCH_TDD_Maximum_Bitrate768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_PUCH_LCR_Information,
      { "e-PUCH-LCR-Information", "nbap.e_PUCH_LCR_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_Non_Scheduled_Grant_LCR_Info,
      { "e-DCH-Non-Scheduled-Grant-LCR-Info", "nbap.e_DCH_Non_Scheduled_Grant_LCR_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_LCRTDD_Information,
      { "e-DCH-LCRTDD-Information", "nbap.e_DCH_LCRTDD_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_timeslotResourceLCR,
      { "timeslotResourceLCR", "nbap.timeslotResourceLCR",
        FT_BYTES, BASE_NONE, NULL, 0,
        "E_DCH_TimeslotResourceLCR", HFILL }},
    { &hf_nbap_subframeNumber,
      { "subframeNumber", "nbap.subframeNumber",
        FT_UINT32, BASE_DEC, VALS(nbap_T_subframeNumber_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_n_E_UCCHLCR,
      { "n-E-UCCHLCR", "nbap.n_E_UCCHLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_HICH_LCR_Information,
      { "e-HICH-LCR-Information", "nbap.e_HICH_LCR_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_signatureSequenceGroupIndex,
      { "signatureSequenceGroupIndex", "nbap.signatureSequenceGroupIndex",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_LCRTDD_PhysicalLayerCategory,
      { "e-DCH-LCRTDD-PhysicalLayerCategory", "nbap.e_DCH_LCRTDD_PhysicalLayerCategory",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dRX_Interruption_by_HS_DSCH,
      { "dRX-Interruption-by-HS-DSCH", "nbap.dRX_Interruption_by_HS_DSCH",
        FT_UINT32, BASE_DEC, VALS(nbap_DRX_Interruption_by_HS_DSCH_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_repetition_Period_List_LCR,
      { "repetition-Period-List-LCR", "nbap.repetition_Period_List_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_SPS_Indicator,
      { "e-DCH-SPS-Indicator", "nbap.e_DCH_SPS_Indicator",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sPS_E_DCH_releted_E_HICH_Information,
      { "sPS-E-DCH-releted-E-HICH-Information", "nbap.sPS_E_DCH_releted_E_HICH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "E_HICH_LCR_Information", HFILL }},
    { &hf_nbap_timeslot_Resource_Related_Information,
      { "timeslot-Resource-Related-Information", "nbap.timeslot_Resource_Related_Information",
        FT_BYTES, BASE_NONE, NULL, 0,
        "E_DCH_TimeslotResourceLCR", HFILL }},
    { &hf_nbap_subframeNumber_01,
      { "subframeNumber", "nbap.subframeNumber",
        FT_UINT32, BASE_DEC, VALS(nbap_T_subframeNumber_01_vals), 0,
        "T_subframeNumber_01", HFILL }},
    { &hf_nbap_FDD_DL_CodeInformation_item,
      { "FDD-DL-CodeInformationItem", "nbap.FDD_DL_CodeInformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dl_ScramblingCode,
      { "dl-ScramblingCode", "nbap.dl_ScramblingCode",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_fdd_DL_ChannelisationCodeNumber,
      { "fdd-DL-ChannelisationCodeNumber", "nbap.fdd_DL_ChannelisationCodeNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_transmissionGapPatternSequenceCodeInformation,
      { "transmissionGapPatternSequenceCodeInformation", "nbap.transmissionGapPatternSequenceCodeInformation",
        FT_UINT32, BASE_DEC, VALS(nbap_TransmissionGapPatternSequenceCodeInformation_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_navClockModel,
      { "navClockModel", "nbap.navClockModel_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_NAVclockModel", HFILL }},
    { &hf_nbap_cnavClockModel,
      { "cnavClockModel", "nbap.cnavClockModel_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_CNAVclockModel", HFILL }},
    { &hf_nbap_glonassClockModel,
      { "glonassClockModel", "nbap.glonassClockModel_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_GLONASSclockModel", HFILL }},
    { &hf_nbap_sbasClockModel,
      { "sbasClockModel", "nbap.sbasClockModel_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_SBASclockModel", HFILL }},
    { &hf_nbap_navKeplerianSet,
      { "navKeplerianSet", "nbap.navKeplerianSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_NavModel_NAVKeplerianSet", HFILL }},
    { &hf_nbap_cnavKeplerianSet,
      { "cnavKeplerianSet", "nbap.cnavKeplerianSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_NavModel_CNAVKeplerianSet", HFILL }},
    { &hf_nbap_glonassECEF,
      { "glonassECEF", "nbap.glonassECEF_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_NavModel_GLONASSecef", HFILL }},
    { &hf_nbap_sbasECEF,
      { "sbasECEF", "nbap.sbasECEF_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_NavModel_SBASecef", HFILL }},
    { &hf_nbap_dataID,
      { "dataID", "nbap.dataID",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_2", HFILL }},
    { &hf_nbap_alpha_beta_parameters,
      { "alpha-beta-parameters", "nbap.alpha_beta_parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GPS_Ionospheric_Model", HFILL }},
    { &hf_nbap_ganss_Transmission_Time,
      { "ganss-Transmission-Time", "nbap.ganss_Transmission_Time_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_non_broadcastIndication,
      { "non-broadcastIndication", "nbap.non_broadcastIndication",
        FT_UINT32, BASE_DEC, VALS(nbap_T_non_broadcastIndication_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_ganssSatInfoNavList,
      { "ganssSatInfoNavList", "nbap.ganssSatInfoNavList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Ganss_Sat_Info_AddNavList", HFILL }},
    { &hf_nbap_GANSS_Additional_Time_Models_item,
      { "GANSS-Time-Model", "nbap.GANSS_Time_Model_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_utcModel1,
      { "utcModel1", "nbap.utcModel1_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_UTCmodelSet1", HFILL }},
    { &hf_nbap_utcModel2,
      { "utcModel2", "nbap.utcModel2_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_UTCmodelSet2", HFILL }},
    { &hf_nbap_utcModel3,
      { "utcModel3", "nbap.utcModel3_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_UTCmodelSet3", HFILL }},
    { &hf_nbap_ganss_wk_number,
      { "ganss-wk-number", "nbap.ganss_wk_number",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_255", HFILL }},
    { &hf_nbap_gANSS_AlmanacModel,
      { "gANSS-AlmanacModel", "nbap.gANSS_AlmanacModel",
        FT_UINT32, BASE_DEC, VALS(nbap_GANSS_AlmanacModel_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_gANSS_keplerianParameters,
      { "gANSS-keplerianParameters", "nbap.gANSS_keplerianParameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_KeplerianParametersAlm", HFILL }},
    { &hf_nbap_extension_GANSS_AlmanacModel,
      { "extension-GANSS-AlmanacModel", "nbap.extension_GANSS_AlmanacModel_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sat_info_SBASecefList,
      { "sat-info-SBASecefList", "nbap.sat_info_SBASecefList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_SAT_Info_Almanac_SBASecefList", HFILL }},
    { &hf_nbap_sat_info_GLOkpList,
      { "sat-info-GLOkpList", "nbap.sat_info_GLOkpList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_SAT_Info_Almanac_GLOkpList", HFILL }},
    { &hf_nbap_t_oa,
      { "t-oa", "nbap.t_oa",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_255", HFILL }},
    { &hf_nbap_sat_info_MIDIkpList,
      { "sat-info-MIDIkpList", "nbap.sat_info_MIDIkpList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_SAT_Info_Almanac_MIDIkpList", HFILL }},
    { &hf_nbap_sat_info_NAVkpList,
      { "sat-info-NAVkpList", "nbap.sat_info_NAVkpList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_SAT_Info_Almanac_NAVkpList", HFILL }},
    { &hf_nbap_sat_info_REDkpList,
      { "sat-info-REDkpList", "nbap.sat_info_REDkpList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_SAT_Info_Almanac_REDkpList", HFILL }},
    { &hf_nbap_ganssID1,
      { "ganssID1", "nbap.ganssID1",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_AuxInfoGANSS_ID1", HFILL }},
    { &hf_nbap_ganssID3,
      { "ganssID3", "nbap.ganssID3",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_AuxInfoGANSS_ID3", HFILL }},
    { &hf_nbap_GANSS_AuxInfoGANSS_ID1_item,
      { "GANSS-AuxInfoGANSS-ID1-element", "nbap.GANSS_AuxInfoGANSS_ID1_element_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_svID,
      { "svID", "nbap.svID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_63", HFILL }},
    { &hf_nbap_signalsAvailable,
      { "signalsAvailable", "nbap.signalsAvailable",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_GANSS_AuxInfoGANSS_ID3_item,
      { "GANSS-AuxInfoGANSS-ID3-element", "nbap.GANSS_AuxInfoGANSS_ID3_element_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_channelNumber,
      { "channelNumber", "nbap.channelNumber",
        FT_INT32, BASE_DEC, NULL, 0,
        "INTEGER_M7_13", HFILL }},
    { &hf_nbap_GANSS_Clock_Model_item,
      { "GANSS-SatelliteClockModelItem", "nbap.GANSS_SatelliteClockModelItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cnavToc,
      { "cnavToc", "nbap.cnavToc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_nbap_cnavTop,
      { "cnavTop", "nbap.cnavTop",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_nbap_cnavURA0,
      { "cnavURA0", "nbap.cnavURA0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_nbap_cnavURA1,
      { "cnavURA1", "nbap.cnavURA1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_3", HFILL }},
    { &hf_nbap_cnavURA2,
      { "cnavURA2", "nbap.cnavURA2",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_3", HFILL }},
    { &hf_nbap_cnavAf2,
      { "cnavAf2", "nbap.cnavAf2",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_nbap_cnavAf1,
      { "cnavAf1", "nbap.cnavAf1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_20", HFILL }},
    { &hf_nbap_cnavAf0,
      { "cnavAf0", "nbap.cnavAf0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_26", HFILL }},
    { &hf_nbap_cnavTgd,
      { "cnavTgd", "nbap.cnavTgd",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_13", HFILL }},
    { &hf_nbap_cnavISCl1cp,
      { "cnavISCl1cp", "nbap.cnavISCl1cp",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_13", HFILL }},
    { &hf_nbap_cnavISCl1cd,
      { "cnavISCl1cd", "nbap.cnavISCl1cd",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_13", HFILL }},
    { &hf_nbap_cnavISCl1ca,
      { "cnavISCl1ca", "nbap.cnavISCl1ca",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_13", HFILL }},
    { &hf_nbap_cnavISCl2c,
      { "cnavISCl2c", "nbap.cnavISCl2c",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_13", HFILL }},
    { &hf_nbap_cnavISCl5i5,
      { "cnavISCl5i5", "nbap.cnavISCl5i5",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_13", HFILL }},
    { &hf_nbap_cnavISCl5q5,
      { "cnavISCl5q5", "nbap.cnavISCl5q5",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_13", HFILL }},
    { &hf_nbap_ganss_Ionospheric_Model,
      { "ganss-Ionospheric-Model", "nbap.ganss_Ionospheric_Model_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ganss_Rx_Pos,
      { "ganss-Rx-Pos", "nbap.ganss_Rx_Pos_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ionospheric_Model,
      { "ionospheric-Model", "nbap.ionospheric_Model",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        "BOOLEAN", HFILL }},
    { &hf_nbap_ganssTod,
      { "ganssTod", "nbap.ganssTod",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_59_", HFILL }},
    { &hf_nbap_dataBitAssistancelist,
      { "dataBitAssistancelist", "nbap.dataBitAssistancelist",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_DataBitAssistanceList", HFILL }},
    { &hf_nbap_GANSS_DataBitAssistanceList_item,
      { "GANSS-DataBitAssistanceItem", "nbap.GANSS_DataBitAssistanceItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dataBitAssistanceSgnList,
      { "dataBitAssistanceSgnList", "nbap.dataBitAssistanceSgnList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_DataBitAssistanceSgnList", HFILL }},
    { &hf_nbap_GANSS_DataBitAssistanceSgnList_item,
      { "GANSS-DataBitAssistanceSgnItem", "nbap.GANSS_DataBitAssistanceSgnItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ganss_SignalId,
      { "ganss-SignalId", "nbap.ganss_SignalId",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_Signal_ID", HFILL }},
    { &hf_nbap_ganssDataBits,
      { "ganssDataBits", "nbap.ganssDataBits",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1_1024", HFILL }},
    { &hf_nbap_ganssTod_01,
      { "ganssTod", "nbap.ganssTod",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_86399", HFILL }},
    { &hf_nbap_ganss_Data_Bit_Assistance_ReqList,
      { "ganss-Data-Bit-Assistance-ReqList", "nbap.ganss_Data_Bit_Assistance_ReqList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ganss_DataBitInterval,
      { "ganss-DataBitInterval", "nbap.ganss_DataBitInterval",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_15", HFILL }},
    { &hf_nbap_ganss_SatelliteInfo,
      { "ganss-SatelliteInfo", "nbap.ganss_SatelliteInfo",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ganss_SatelliteInfo_item,
      { "ganss-SatelliteInfo item", "nbap.ganss_SatelliteInfo_item",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_63", HFILL }},
    { &hf_nbap_b1,
      { "b1", "nbap.b1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_nbap_b2,
      { "b2", "nbap.b2",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_nbap_teop,
      { "teop", "nbap.teop",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_pmX,
      { "pmX", "nbap.pmX",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_21", HFILL }},
    { &hf_nbap_pmXdot,
      { "pmXdot", "nbap.pmXdot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_15", HFILL }},
    { &hf_nbap_pmY,
      { "pmY", "nbap.pmY",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_21", HFILL }},
    { &hf_nbap_pmYdot,
      { "pmYdot", "nbap.pmYdot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_15", HFILL }},
    { &hf_nbap_deltaUT1,
      { "deltaUT1", "nbap.deltaUT1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_31", HFILL }},
    { &hf_nbap_deltaUT1dot,
      { "deltaUT1dot", "nbap.deltaUT1dot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_19", HFILL }},
    { &hf_nbap_GANSS_GenericDataInfoReqList_item,
      { "GANSS-GenericDataInfoReqItem", "nbap.GANSS_GenericDataInfoReqItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ganss_Id,
      { "ganss-Id", "nbap.ganss_Id",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ganss_Navigation_Model_And_Time_Recovery,
      { "ganss-Navigation-Model-And-Time-Recovery", "nbap.ganss_Navigation_Model_And_Time_Recovery",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        "BOOLEAN", HFILL }},
    { &hf_nbap_ganss_Time_Model_GNSS_GNSS,
      { "ganss-Time-Model-GNSS-GNSS", "nbap.ganss_Time_Model_GNSS_GNSS",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_9", HFILL }},
    { &hf_nbap_ganss_UTC_Model,
      { "ganss-UTC-Model", "nbap.ganss_UTC_Model",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        "BOOLEAN", HFILL }},
    { &hf_nbap_ganss_Almanac,
      { "ganss-Almanac", "nbap.ganss_Almanac",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        "BOOLEAN", HFILL }},
    { &hf_nbap_gANSS_GenericDataInfoReqItem_ganss_Real_Time_Integrity,
      { "ganss-Real-Time-Integrity", "nbap.gANSS-GenericDataInfoReqItem.ganss-Real-Time-Integrity",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        "BOOLEAN", HFILL }},
    { &hf_nbap_ganss_Data_Bit_Assistance_Req,
      { "ganss-Data-Bit-Assistance-Req", "nbap.ganss_Data_Bit_Assistance_Req_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_Data_Bit_Assistance_ReqItem", HFILL }},
    { &hf_nbap_GANSS_Generic_Data_item,
      { "GANSS-Generic-DataItem", "nbap.GANSS_Generic_DataItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dganss_Correction,
      { "dganss-Correction", "nbap.dganss_Correction_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DGANSSCorrections", HFILL }},
    { &hf_nbap_ganss_Navigation_Model_And_Time_Recovery_01,
      { "ganss-Navigation-Model-And-Time-Recovery", "nbap.ganss_Navigation_Model_And_Time_Recovery_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ganss_Time_Model,
      { "ganss-Time-Model", "nbap.ganss_Time_Model_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ganss_UTC_TIME,
      { "ganss-UTC-TIME", "nbap.ganss_UTC_TIME_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_UTC_Model", HFILL }},
    { &hf_nbap_ganss_Almanac_01,
      { "ganss-Almanac", "nbap.ganss_Almanac_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ganss_Real_Time_Integrity,
      { "ganss-Real-Time-Integrity", "nbap.ganss_Real_Time_Integrity",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ganss_Data_Bit_Assistance,
      { "ganss-Data-Bit-Assistance", "nbap.ganss_Data_Bit_Assistance_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_gloTau,
      { "gloTau", "nbap.gloTau",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_22", HFILL }},
    { &hf_nbap_gloGamma,
      { "gloGamma", "nbap.gloGamma",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_nbap_gloDeltaTau,
      { "gloDeltaTau", "nbap.gloDeltaTau",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_nbap_gANSS_CommonDataInfoReq,
      { "gANSS-CommonDataInfoReq", "nbap.gANSS_CommonDataInfoReq_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_gANSS_GenericDataInfoReqList,
      { "gANSS-GenericDataInfoReqList", "nbap.gANSS_GenericDataInfoReqList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_alpha_zero_ionos,
      { "alpha-zero-ionos", "nbap.alpha_zero_ionos",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_12", HFILL }},
    { &hf_nbap_alpha_one_ionos,
      { "alpha-one-ionos", "nbap.alpha_one_ionos",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_12", HFILL }},
    { &hf_nbap_alpha_two_ionos,
      { "alpha-two-ionos", "nbap.alpha_two_ionos",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_12", HFILL }},
    { &hf_nbap_gANSS_IonosphereRegionalStormFlags,
      { "gANSS-IonosphereRegionalStormFlags", "nbap.gANSS_IonosphereRegionalStormFlags_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_storm_flag_one,
      { "storm-flag-one", "nbap.storm_flag_one",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        "BOOLEAN", HFILL }},
    { &hf_nbap_storm_flag_two,
      { "storm-flag-two", "nbap.storm_flag_two",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        "BOOLEAN", HFILL }},
    { &hf_nbap_storm_flag_three,
      { "storm-flag-three", "nbap.storm_flag_three",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        "BOOLEAN", HFILL }},
    { &hf_nbap_storm_flag_four,
      { "storm-flag-four", "nbap.storm_flag_four",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        "BOOLEAN", HFILL }},
    { &hf_nbap_storm_flag_five,
      { "storm-flag-five", "nbap.storm_flag_five",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        "BOOLEAN", HFILL }},
    { &hf_nbap_iod_a,
      { "iod-a", "nbap.iod_a",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_3", HFILL }},
    { &hf_nbap_gANSS_SatelliteInformationKP,
      { "gANSS-SatelliteInformationKP", "nbap.gANSS_SatelliteInformationKP",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_toe_nav,
      { "toe-nav", "nbap.toe_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_14", HFILL }},
    { &hf_nbap_ganss_omega_nav,
      { "ganss-omega-nav", "nbap.ganss_omega_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_nbap_delta_n_nav,
      { "delta-n-nav", "nbap.delta_n_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_m_zero_nav,
      { "m-zero-nav", "nbap.m_zero_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_nbap_omegadot_nav,
      { "omegadot-nav", "nbap.omegadot_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_nbap_ganss_e_nav,
      { "ganss-e-nav", "nbap.ganss_e_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_nbap_idot_nav,
      { "idot-nav", "nbap.idot_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_14", HFILL }},
    { &hf_nbap_a_sqrt_nav,
      { "a-sqrt-nav", "nbap.a_sqrt_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_nbap_i_zero_nav,
      { "i-zero-nav", "nbap.i_zero_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_nbap_omega_zero_nav,
      { "omega-zero-nav", "nbap.omega_zero_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_nbap_c_rs_nav,
      { "c-rs-nav", "nbap.c_rs_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_c_is_nav,
      { "c-is-nav", "nbap.c_is_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_c_us_nav,
      { "c-us-nav", "nbap.c_us_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_c_rc_nav,
      { "c-rc-nav", "nbap.c_rc_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_c_ic_nav,
      { "c-ic-nav", "nbap.c_ic_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_c_uc_nav,
      { "c-uc-nav", "nbap.c_uc_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_navToc,
      { "navToc", "nbap.navToc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_navaf2,
      { "navaf2", "nbap.navaf2",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_navaf1,
      { "navaf1", "nbap.navaf1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_navaf0,
      { "navaf0", "nbap.navaf0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_22", HFILL }},
    { &hf_nbap_navTgd,
      { "navTgd", "nbap.navTgd",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_non_broadcastIndication_01,
      { "non-broadcastIndication", "nbap.non_broadcastIndication",
        FT_UINT32, BASE_DEC, VALS(nbap_T_non_broadcastIndication_01_vals), 0,
        "T_non_broadcastIndication_01", HFILL }},
    { &hf_nbap_ganssSatInfoNav,
      { "ganssSatInfoNav", "nbap.ganssSatInfoNav",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_Sat_Info_Nav", HFILL }},
    { &hf_nbap_cnavURAindex,
      { "cnavURAindex", "nbap.cnavURAindex",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_nbap_cnavDeltaA,
      { "cnavDeltaA", "nbap.cnavDeltaA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_26", HFILL }},
    { &hf_nbap_cnavAdot,
      { "cnavAdot", "nbap.cnavAdot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_25", HFILL }},
    { &hf_nbap_cnavDeltaNo,
      { "cnavDeltaNo", "nbap.cnavDeltaNo",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_17", HFILL }},
    { &hf_nbap_cnavDeltaNoDot,
      { "cnavDeltaNoDot", "nbap.cnavDeltaNoDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_23", HFILL }},
    { &hf_nbap_cnavMo,
      { "cnavMo", "nbap.cnavMo",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_33", HFILL }},
    { &hf_nbap_cnavE,
      { "cnavE", "nbap.cnavE",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_33", HFILL }},
    { &hf_nbap_cnavOmega,
      { "cnavOmega", "nbap.cnavOmega",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_33", HFILL }},
    { &hf_nbap_cnavOMEGA0,
      { "cnavOMEGA0", "nbap.cnavOMEGA0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_33", HFILL }},
    { &hf_nbap_cnavDeltaOmegaDot,
      { "cnavDeltaOmegaDot", "nbap.cnavDeltaOmegaDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_17", HFILL }},
    { &hf_nbap_cnavIo,
      { "cnavIo", "nbap.cnavIo",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_33", HFILL }},
    { &hf_nbap_cnavIoDot,
      { "cnavIoDot", "nbap.cnavIoDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_15", HFILL }},
    { &hf_nbap_cnavCis,
      { "cnavCis", "nbap.cnavCis",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_cnavCic,
      { "cnavCic", "nbap.cnavCic",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_cnavCrs,
      { "cnavCrs", "nbap.cnavCrs",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_nbap_cnavCrc,
      { "cnavCrc", "nbap.cnavCrc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_nbap_cnavCus,
      { "cnavCus", "nbap.cnavCus",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_21", HFILL }},
    { &hf_nbap_cnavCuc,
      { "cnavCuc", "nbap.cnavCuc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_21", HFILL }},
    { &hf_nbap_gloEn,
      { "gloEn", "nbap.gloEn",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_nbap_gloP1,
      { "gloP1", "nbap.gloP1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_2", HFILL }},
    { &hf_nbap_gloP2,
      { "gloP2", "nbap.gloP2",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1", HFILL }},
    { &hf_nbap_gloM,
      { "gloM", "nbap.gloM",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_2", HFILL }},
    { &hf_nbap_gloX,
      { "gloX", "nbap.gloX",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_27", HFILL }},
    { &hf_nbap_gloXdot,
      { "gloXdot", "nbap.gloXdot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_nbap_gloXdotdot,
      { "gloXdotdot", "nbap.gloXdotdot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_nbap_gloY,
      { "gloY", "nbap.gloY",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_27", HFILL }},
    { &hf_nbap_gloYdot,
      { "gloYdot", "nbap.gloYdot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_nbap_gloYdotdot,
      { "gloYdotdot", "nbap.gloYdotdot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_nbap_gloZ,
      { "gloZ", "nbap.gloZ",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_27", HFILL }},
    { &hf_nbap_gloZdot,
      { "gloZdot", "nbap.gloZdot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_nbap_gloZdotdot,
      { "gloZdotdot", "nbap.gloZdotdot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_nbap_navURA,
      { "navURA", "nbap.navURA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_4", HFILL }},
    { &hf_nbap_navFitFlag,
      { "navFitFlag", "nbap.navFitFlag",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1", HFILL }},
    { &hf_nbap_navToe,
      { "navToe", "nbap.navToe",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_navOmega,
      { "navOmega", "nbap.navOmega",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_nbap_navDeltaN,
      { "navDeltaN", "nbap.navDeltaN",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_navM0,
      { "navM0", "nbap.navM0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_nbap_navOmegaADot,
      { "navOmegaADot", "nbap.navOmegaADot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_nbap_navE,
      { "navE", "nbap.navE",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_nbap_navIDot,
      { "navIDot", "nbap.navIDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_14", HFILL }},
    { &hf_nbap_navAPowerHalf,
      { "navAPowerHalf", "nbap.navAPowerHalf",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_nbap_navI0,
      { "navI0", "nbap.navI0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_nbap_navOmegaA0,
      { "navOmegaA0", "nbap.navOmegaA0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_nbap_navCrs,
      { "navCrs", "nbap.navCrs",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_navCis,
      { "navCis", "nbap.navCis",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_navCus,
      { "navCus", "nbap.navCus",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_navCrc,
      { "navCrc", "nbap.navCrc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_navCic,
      { "navCic", "nbap.navCic",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_navCuc,
      { "navCuc", "nbap.navCuc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_sbasTo,
      { "sbasTo", "nbap.sbasTo",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_13", HFILL }},
    { &hf_nbap_sbasAccuracy,
      { "sbasAccuracy", "nbap.sbasAccuracy",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_4", HFILL }},
    { &hf_nbap_sbasXg,
      { "sbasXg", "nbap.sbasXg",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_30", HFILL }},
    { &hf_nbap_sbasYg,
      { "sbasYg", "nbap.sbasYg",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_30", HFILL }},
    { &hf_nbap_sbasZg,
      { "sbasZg", "nbap.sbasZg",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_25", HFILL }},
    { &hf_nbap_sbasXgDot,
      { "sbasXgDot", "nbap.sbasXgDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_17", HFILL }},
    { &hf_nbap_sbasYgDot,
      { "sbasYgDot", "nbap.sbasYgDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_17", HFILL }},
    { &hf_nbap_sbasZgDot,
      { "sbasZgDot", "nbap.sbasZgDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_18", HFILL }},
    { &hf_nbap_sbasXgDotDot,
      { "sbasXgDotDot", "nbap.sbasXgDotDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_nbap_sbagYgDotDot,
      { "sbagYgDotDot", "nbap.sbagYgDotDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_nbap_sbasZgDotDot,
      { "sbasZgDotDot", "nbap.sbasZgDotDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_nbap_gANSS_keplerianParameters_01,
      { "gANSS-keplerianParameters", "nbap.gANSS_keplerianParameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_KeplerianParametersOrb", HFILL }},
    { &hf_nbap_GANSS_Real_Time_Integrity_item,
      { "GANSS-RealTimeInformationItem", "nbap.GANSS_RealTimeInformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_bad_ganss_satId,
      { "bad-ganss-satId", "nbap.bad_ganss_satId",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_63", HFILL }},
    { &hf_nbap_bad_ganss_signalId,
      { "bad-ganss-signalId", "nbap.bad_ganss_signalId",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_latitudeSign,
      { "latitudeSign", "nbap.latitudeSign",
        FT_UINT32, BASE_DEC, VALS(nbap_T_latitudeSign_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_degreesOfLatitude,
      { "degreesOfLatitude", "nbap.degreesOfLatitude",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_2147483647", HFILL }},
    { &hf_nbap_degreesOfLongitude,
      { "degreesOfLongitude", "nbap.degreesOfLongitude",
        FT_INT32, BASE_DEC, NULL, 0,
        "INTEGER_M2147483648_2147483647", HFILL }},
    { &hf_nbap_directionOfAltitude,
      { "directionOfAltitude", "nbap.directionOfAltitude",
        FT_UINT32, BASE_DEC, VALS(nbap_T_directionOfAltitude_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_altitude,
      { "altitude", "nbap.altitude",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_32767", HFILL }},
    { &hf_nbap_t_oc,
      { "t-oc", "nbap.t_oc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_14", HFILL }},
    { &hf_nbap_a_i2,
      { "a-i2", "nbap.a_i2",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_12", HFILL }},
    { &hf_nbap_a_i1,
      { "a-i1", "nbap.a_i1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_18", HFILL }},
    { &hf_nbap_a_i0,
      { "a-i0", "nbap.a_i0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_28", HFILL }},
    { &hf_nbap_t_gd,
      { "t-gd", "nbap.t_gd",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_nbap_model_id,
      { "model-id", "nbap.model_id",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_1_", HFILL }},
    { &hf_nbap_GANSS_SatelliteInformationKP_item,
      { "GANSS-SatelliteInformationKPItem", "nbap.GANSS_SatelliteInformationKPItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ganss_e_alm,
      { "ganss-e-alm", "nbap.ganss_e_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_nbap_ganss_delta_I_alm,
      { "ganss-delta-I-alm", "nbap.ganss_delta_I_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_nbap_ganss_omegadot_alm,
      { "ganss-omegadot-alm", "nbap.ganss_omegadot_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_nbap_ganss_svhealth_alm,
      { "ganss-svhealth-alm", "nbap.ganss_svhealth_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_4", HFILL }},
    { &hf_nbap_ganss_delta_a_sqrt_alm,
      { "ganss-delta-a-sqrt-alm", "nbap.ganss_delta_a_sqrt_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_17", HFILL }},
    { &hf_nbap_ganss_omegazero_alm,
      { "ganss-omegazero-alm", "nbap.ganss_omegazero_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_ganss_m_zero_alm,
      { "ganss-m-zero-alm", "nbap.ganss_m_zero_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_ganss_omega_alm,
      { "ganss-omega-alm", "nbap.ganss_omega_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_ganss_af_zero_alm,
      { "ganss-af-zero-alm", "nbap.ganss_af_zero_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_14", HFILL }},
    { &hf_nbap_ganss_af_one_alm,
      { "ganss-af-one-alm", "nbap.ganss_af_one_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_nbap_Ganss_Sat_Info_AddNavList_item,
      { "Ganss-Sat-Info-AddNavList item", "nbap.Ganss_Sat_Info_AddNavList_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_svHealth,
      { "svHealth", "nbap.svHealth",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_6", HFILL }},
    { &hf_nbap_iod,
      { "iod", "nbap.iod",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_nbap_ganssAddClockModels,
      { "ganssAddClockModels", "nbap.ganssAddClockModels",
        FT_UINT32, BASE_DEC, VALS(nbap_GANSS_AddClockModels_vals), 0,
        "GANSS_AddClockModels", HFILL }},
    { &hf_nbap_ganssAddOrbitModels,
      { "ganssAddOrbitModels", "nbap.ganssAddOrbitModels",
        FT_UINT32, BASE_DEC, VALS(nbap_GANSS_AddOrbitModels_vals), 0,
        "GANSS_AddOrbitModels", HFILL }},
    { &hf_nbap_GANSS_SAT_Info_Almanac_GLOkpList_item,
      { "GANSS-SAT-Info-Almanac-GLOkp", "nbap.GANSS_SAT_Info_Almanac_GLOkp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_gloAlmNA,
      { "gloAlmNA", "nbap.gloAlmNA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_nbap_gloAlmnA,
      { "gloAlmnA", "nbap.gloAlmnA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_nbap_gloAlmHA,
      { "gloAlmHA", "nbap.gloAlmHA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_nbap_gloAlmLambdaA,
      { "gloAlmLambdaA", "nbap.gloAlmLambdaA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_21", HFILL }},
    { &hf_nbap_gloAlmTlambdaA,
      { "gloAlmTlambdaA", "nbap.gloAlmTlambdaA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_21", HFILL }},
    { &hf_nbap_gloAlmDeltaIA,
      { "gloAlmDeltaIA", "nbap.gloAlmDeltaIA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_18", HFILL }},
    { &hf_nbap_gloAkmDeltaTA,
      { "gloAkmDeltaTA", "nbap.gloAkmDeltaTA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_22", HFILL }},
    { &hf_nbap_gloAlmDeltaTdotA,
      { "gloAlmDeltaTdotA", "nbap.gloAlmDeltaTdotA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_7", HFILL }},
    { &hf_nbap_gloAlmEpsilonA,
      { "gloAlmEpsilonA", "nbap.gloAlmEpsilonA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_15", HFILL }},
    { &hf_nbap_gloAlmOmegaA,
      { "gloAlmOmegaA", "nbap.gloAlmOmegaA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_gloAlmTauA,
      { "gloAlmTauA", "nbap.gloAlmTauA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_nbap_gloAlmCA,
      { "gloAlmCA", "nbap.gloAlmCA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1", HFILL }},
    { &hf_nbap_gloAlmMA,
      { "gloAlmMA", "nbap.gloAlmMA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_2", HFILL }},
    { &hf_nbap_GANSS_SAT_Info_Almanac_MIDIkpList_item,
      { "GANSS-SAT-Info-Almanac-MIDIkp", "nbap.GANSS_SAT_Info_Almanac_MIDIkp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_midiAlmE,
      { "midiAlmE", "nbap.midiAlmE",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_nbap_midiAlmDeltaI,
      { "midiAlmDeltaI", "nbap.midiAlmDeltaI",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_nbap_midiAlmOmegaDot,
      { "midiAlmOmegaDot", "nbap.midiAlmOmegaDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_nbap_midiAlmSqrtA,
      { "midiAlmSqrtA", "nbap.midiAlmSqrtA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_17", HFILL }},
    { &hf_nbap_midiAlmOmega0,
      { "midiAlmOmega0", "nbap.midiAlmOmega0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_midiAlmOmega,
      { "midiAlmOmega", "nbap.midiAlmOmega",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_midiAlmMo,
      { "midiAlmMo", "nbap.midiAlmMo",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_midiAlmaf0,
      { "midiAlmaf0", "nbap.midiAlmaf0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_nbap_midiAlmaf1,
      { "midiAlmaf1", "nbap.midiAlmaf1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_nbap_midiAlmL1Health,
      { "midiAlmL1Health", "nbap.midiAlmL1Health",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1", HFILL }},
    { &hf_nbap_midiAlmL2Health,
      { "midiAlmL2Health", "nbap.midiAlmL2Health",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1", HFILL }},
    { &hf_nbap_midiAlmL5Health,
      { "midiAlmL5Health", "nbap.midiAlmL5Health",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1", HFILL }},
    { &hf_nbap_GANSS_SAT_Info_Almanac_NAVkpList_item,
      { "GANSS-SAT-Info-Almanac-NAVkp", "nbap.GANSS_SAT_Info_Almanac_NAVkp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_navAlmE,
      { "navAlmE", "nbap.navAlmE",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_navAlmDeltaI,
      { "navAlmDeltaI", "nbap.navAlmDeltaI",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_navAlmOMEGADOT,
      { "navAlmOMEGADOT", "nbap.navAlmOMEGADOT",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_navAlmSVHealth,
      { "navAlmSVHealth", "nbap.navAlmSVHealth",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_navAlmSqrtA,
      { "navAlmSqrtA", "nbap.navAlmSqrtA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_nbap_navAlmOMEGAo,
      { "navAlmOMEGAo", "nbap.navAlmOMEGAo",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_nbap_navAlmOmega,
      { "navAlmOmega", "nbap.navAlmOmega",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_nbap_navAlmMo,
      { "navAlmMo", "nbap.navAlmMo",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_nbap_navAlmaf0,
      { "navAlmaf0", "nbap.navAlmaf0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_nbap_navAlmaf1,
      { "navAlmaf1", "nbap.navAlmaf1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_nbap_GANSS_SAT_Info_Almanac_REDkpList_item,
      { "GANSS-SAT-Info-Almanac-REDkp", "nbap.GANSS_SAT_Info_Almanac_REDkp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_redAlmDeltaA,
      { "redAlmDeltaA", "nbap.redAlmDeltaA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_redAlmOmega0,
      { "redAlmOmega0", "nbap.redAlmOmega0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_7", HFILL }},
    { &hf_nbap_redAlmPhi0,
      { "redAlmPhi0", "nbap.redAlmPhi0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_7", HFILL }},
    { &hf_nbap_redAlmL1Health,
      { "redAlmL1Health", "nbap.redAlmL1Health",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1", HFILL }},
    { &hf_nbap_redAlmL2Health,
      { "redAlmL2Health", "nbap.redAlmL2Health",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1", HFILL }},
    { &hf_nbap_redAlmL5Health,
      { "redAlmL5Health", "nbap.redAlmL5Health",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1", HFILL }},
    { &hf_nbap_GANSS_SAT_Info_Almanac_SBASecefList_item,
      { "GANSS-SAT-Info-Almanac-SBASecef", "nbap.GANSS_SAT_Info_Almanac_SBASecef_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sbasAlmDataID,
      { "sbasAlmDataID", "nbap.sbasAlmDataID",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_2", HFILL }},
    { &hf_nbap_sbasAlmHealth,
      { "sbasAlmHealth", "nbap.sbasAlmHealth",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_sbasAlmXg,
      { "sbasAlmXg", "nbap.sbasAlmXg",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_15", HFILL }},
    { &hf_nbap_sbasAlmYg,
      { "sbasAlmYg", "nbap.sbasAlmYg",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_15", HFILL }},
    { &hf_nbap_sbasAlmZg,
      { "sbasAlmZg", "nbap.sbasAlmZg",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_9", HFILL }},
    { &hf_nbap_sbasAlmXgdot,
      { "sbasAlmXgdot", "nbap.sbasAlmXgdot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_3", HFILL }},
    { &hf_nbap_sbasAlmYgDot,
      { "sbasAlmYgDot", "nbap.sbasAlmYgDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_3", HFILL }},
    { &hf_nbap_sbasAlmZgDot,
      { "sbasAlmZgDot", "nbap.sbasAlmZgDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_4", HFILL }},
    { &hf_nbap_sbasAlmTo,
      { "sbasAlmTo", "nbap.sbasAlmTo",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_nbap_GANSS_Sat_Info_Nav_item,
      { "GANSS-Sat-Info-Nav item", "nbap.GANSS_Sat_Info_Nav_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_svHealth_01,
      { "svHealth", "nbap.svHealth",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_nbap_iod_01,
      { "iod", "nbap.iod",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_nbap_ganssClockModel,
      { "ganssClockModel", "nbap.ganssClockModel",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_Clock_Model", HFILL }},
    { &hf_nbap_ganssOrbitModel,
      { "ganssOrbitModel", "nbap.ganssOrbitModel",
        FT_UINT32, BASE_DEC, VALS(nbap_GANSS_Orbit_Model_vals), 0,
        "GANSS_Orbit_Model", HFILL }},
    { &hf_nbap_sbasAgfo,
      { "sbasAgfo", "nbap.sbasAgfo",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_12", HFILL }},
    { &hf_nbap_sbasAgf1,
      { "sbasAgf1", "nbap.sbasAgf1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_ganss_time_model_Ref_Time,
      { "ganss-time-model-Ref-Time", "nbap.ganss_time_model_Ref_Time",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_37799", HFILL }},
    { &hf_nbap_ganss_t_a0,
      { "ganss-t-a0", "nbap.ganss_t_a0",
        FT_INT32, BASE_DEC, NULL, 0,
        "INTEGER_M2147483648_2147483647", HFILL }},
    { &hf_nbap_ganss_t_a1,
      { "ganss-t-a1", "nbap.ganss_t_a1",
        FT_INT32, BASE_DEC, NULL, 0,
        "INTEGER_M8388608_8388607", HFILL }},
    { &hf_nbap_ganss_t_a2,
      { "ganss-t-a2", "nbap.ganss_t_a2",
        FT_INT32, BASE_DEC, NULL, 0,
        "INTEGER_M64_63", HFILL }},
    { &hf_nbap_gnss_to_id,
      { "gnss-to-id", "nbap.gnss_to_id",
        FT_UINT32, BASE_DEC, VALS(nbap_T_gnss_to_id_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_ganss_wk_number_01,
      { "ganss-wk-number", "nbap.ganss_wk_number",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_8191", HFILL }},
    { &hf_nbap_ganssDay,
      { "ganssDay", "nbap.ganssDay",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_8191", HFILL }},
    { &hf_nbap_a_one_utc,
      { "a-one-utc", "nbap.a_one_utc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_nbap_a_zero_utc,
      { "a-zero-utc", "nbap.a_zero_utc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_nbap_t_ot_utc,
      { "t-ot-utc", "nbap.t_ot_utc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_w_n_t_utc,
      { "w-n-t-utc", "nbap.w_n_t_utc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_delta_t_ls_utc,
      { "delta-t-ls-utc", "nbap.delta_t_ls_utc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_w_n_lsf_utc,
      { "w-n-lsf-utc", "nbap.w_n_lsf_utc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_dn_utc,
      { "dn-utc", "nbap.dn_utc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_delta_t_lsf_utc,
      { "delta-t-lsf-utc", "nbap.delta_t_lsf_utc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_utcA0,
      { "utcA0", "nbap.utcA0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_utcA1,
      { "utcA1", "nbap.utcA1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_13", HFILL }},
    { &hf_nbap_utcA2,
      { "utcA2", "nbap.utcA2",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_7", HFILL }},
    { &hf_nbap_utcDeltaTls,
      { "utcDeltaTls", "nbap.utcDeltaTls",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_utcTot,
      { "utcTot", "nbap.utcTot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_utcWNot,
      { "utcWNot", "nbap.utcWNot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_13", HFILL }},
    { &hf_nbap_utcWNlsf,
      { "utcWNlsf", "nbap.utcWNlsf",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_utcDN,
      { "utcDN", "nbap.utcDN",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_4", HFILL }},
    { &hf_nbap_utcDeltaTlsf,
      { "utcDeltaTlsf", "nbap.utcDeltaTlsf",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_nA,
      { "nA", "nbap.nA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_nbap_tauC,
      { "tauC", "nbap.tauC",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_nbap_deltaUT1_01,
      { "deltaUT1", "nbap.deltaUT1_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_DeltaUT1", HFILL }},
    { &hf_nbap_kp,
      { "kp", "nbap.kp",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_2", HFILL }},
    { &hf_nbap_utcA1wnt,
      { "utcA1wnt", "nbap.utcA1wnt",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_nbap_utcA0wnt,
      { "utcA0wnt", "nbap.utcA0wnt",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_nbap_utcTot_01,
      { "utcTot", "nbap.utcTot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_utcWNt,
      { "utcWNt", "nbap.utcWNt",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_utcDN_01,
      { "utcDN", "nbap.utcDN",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_utcStandardID,
      { "utcStandardID", "nbap.utcStandardID",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_3", HFILL }},
    { &hf_nbap_wna_alm,
      { "wna-alm", "nbap.wna_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_sat_info_almanac,
      { "sat-info-almanac", "nbap.sat_info_almanac",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sVGlobalHealth_alm,
      { "sVGlobalHealth-alm", "nbap.sVGlobalHealth_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_364", HFILL }},
    { &hf_nbap_alpha_zero_ionos_01,
      { "alpha-zero-ionos", "nbap.alpha_zero_ionos",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_alpha_one_ionos_01,
      { "alpha-one-ionos", "nbap.alpha_one_ionos",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_alpha_two_ionos_01,
      { "alpha-two-ionos", "nbap.alpha_two_ionos",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_alpha_three_ionos,
      { "alpha-three-ionos", "nbap.alpha_three_ionos",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_beta_zero_ionos,
      { "beta-zero-ionos", "nbap.beta_zero_ionos",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_beta_one_ionos,
      { "beta-one-ionos", "nbap.beta_one_ionos",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_beta_two_ionos,
      { "beta-two-ionos", "nbap.beta_two_ionos",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_beta_three_ionos,
      { "beta-three-ionos", "nbap.beta_three_ionos",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_GPS_Information_item,
      { "GPS-Information-Item", "nbap.GPS_Information_Item",
        FT_UINT32, BASE_DEC, VALS(nbap_GPS_Information_Item_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_bad_satellites,
      { "bad-satellites", "nbap.bad_satellites_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GPSBadSat_Info_RealTime_Integrity", HFILL }},
    { &hf_nbap_no_bad_satellites,
      { "no-bad-satellites", "nbap.no_bad_satellites_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sat_info,
      { "sat-info", "nbap.sat_info",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SATInfo_RealTime_Integrity", HFILL }},
    { &hf_nbap_GPS_NavigationModel_and_TimeRecovery_item,
      { "GPS-NavandRecovery-Item", "nbap.GPS_NavandRecovery_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tx_tow_nav,
      { "tx-tow-nav", "nbap.tx_tow_nav",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_1048575", HFILL }},
    { &hf_nbap_sat_id_nav,
      { "sat-id-nav", "nbap.sat_id_nav",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SAT_ID", HFILL }},
    { &hf_nbap_tlm_message_nav,
      { "tlm-message-nav", "nbap.tlm_message_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_14", HFILL }},
    { &hf_nbap_tlm_revd_c_nav,
      { "tlm-revd-c-nav", "nbap.tlm_revd_c_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_2", HFILL }},
    { &hf_nbap_ho_word_nav,
      { "ho-word-nav", "nbap.ho_word_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_22", HFILL }},
    { &hf_nbap_w_n_nav,
      { "w-n-nav", "nbap.w_n_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_nbap_ca_or_p_on_l2_nav,
      { "ca-or-p-on-l2-nav", "nbap.ca_or_p_on_l2_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_2", HFILL }},
    { &hf_nbap_user_range_accuracy_index_nav,
      { "user-range-accuracy-index-nav", "nbap.user_range_accuracy_index_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_4", HFILL }},
    { &hf_nbap_sv_health_nav,
      { "sv-health-nav", "nbap.sv_health_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_6", HFILL }},
    { &hf_nbap_iodc_nav,
      { "iodc-nav", "nbap.iodc_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_nbap_l2_p_dataflag_nav,
      { "l2-p-dataflag-nav", "nbap.l2_p_dataflag_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1", HFILL }},
    { &hf_nbap_sf1_reserved_nav,
      { "sf1-reserved-nav", "nbap.sf1_reserved_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_87", HFILL }},
    { &hf_nbap_t_gd_nav,
      { "t-gd-nav", "nbap.t_gd_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_t_oc_nav,
      { "t-oc-nav", "nbap.t_oc_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_a_f_2_nav,
      { "a-f-2-nav", "nbap.a_f_2_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_a_f_1_nav,
      { "a-f-1-nav", "nbap.a_f_1_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_a_f_zero_nav,
      { "a-f-zero-nav", "nbap.a_f_zero_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_22", HFILL }},
    { &hf_nbap_gps_e_nav,
      { "gps-e-nav", "nbap.gps_e_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_nbap_t_oe_nav,
      { "t-oe-nav", "nbap.t_oe_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_fit_interval_flag_nav,
      { "fit-interval-flag-nav", "nbap.fit_interval_flag_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1", HFILL }},
    { &hf_nbap_aodo_nav,
      { "aodo-nav", "nbap.aodo_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_nbap_gps_omega_nav,
      { "gps-omega-nav", "nbap.gps_omega_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_nbap_spare_zero_fill,
      { "spare-zero-fill", "nbap.spare_zero_fill",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_20", HFILL }},
    { &hf_nbap_latitudeSign_01,
      { "latitudeSign", "nbap.latitudeSign",
        FT_UINT32, BASE_DEC, VALS(nbap_T_latitudeSign_01_vals), 0,
        "T_latitudeSign_01", HFILL }},
    { &hf_nbap_latitude,
      { "latitude", "nbap.latitude",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_8388607", HFILL }},
    { &hf_nbap_longitude,
      { "longitude", "nbap.longitude",
        FT_INT32, BASE_DEC, NULL, 0,
        "INTEGER_M8388608_8388607", HFILL }},
    { &hf_nbap_directionOfAltitude_01,
      { "directionOfAltitude", "nbap.directionOfAltitude",
        FT_UINT32, BASE_DEC, VALS(nbap_T_directionOfAltitude_01_vals), 0,
        "T_directionOfAltitude_01", HFILL }},
    { &hf_nbap_implicit,
      { "implicit", "nbap.implicit_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "HARQ_MemoryPartitioning_Implicit", HFILL }},
    { &hf_nbap_explicit,
      { "explicit", "nbap.explicit_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "HARQ_MemoryPartitioning_Explicit", HFILL }},
    { &hf_nbap_number_of_Processes,
      { "number-of-Processes", "nbap.number_of_Processes",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_8_", HFILL }},
    { &hf_nbap_hARQ_MemoryPartitioningList,
      { "hARQ-MemoryPartitioningList", "nbap.hARQ_MemoryPartitioningList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HARQ_MemoryPartitioningList_item,
      { "HARQ-MemoryPartitioningItem", "nbap.HARQ_MemoryPartitioningItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HARQ_MemoryPartitioningInfoExtForMIMO_item,
      { "HARQ-MemoryPartitioningItem", "nbap.HARQ_MemoryPartitioningItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_process_Memory_Size,
      { "process-Memory-Size", "nbap.process_Memory_Size",
        FT_UINT32, BASE_DEC, VALS(nbap_T_process_Memory_Size_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_HS_DSCHProvidedBitRate_item,
      { "HS-DSCHProvidedBitRate-Item", "nbap.HS_DSCHProvidedBitRate_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_DSCHProvidedBitRateValue,
      { "hS-DSCHProvidedBitRateValue", "nbap.hS_DSCHProvidedBitRateValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortion_item,
      { "HS-DSCHProvidedBitRateValueInformation-For-CellPortion-Item", "nbap.HS_DSCHProvidedBitRateValueInformation_For_CellPortion_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_DSCHProvidedBitRateValue_01,
      { "hS-DSCHProvidedBitRateValue", "nbap.hS_DSCHProvidedBitRateValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HS_DSCHProvidedBitRate", HFILL }},
    { &hf_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR_item,
      { "HS-DSCHProvidedBitRateValueInformation-For-CellPortionLCR-Item", "nbap.HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_DSCHRequiredPower_item,
      { "HS-DSCHRequiredPower-Item", "nbap.HS_DSCHRequiredPower_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_DSCHRequiredPowerValue,
      { "hS-DSCHRequiredPowerValue", "nbap.hS_DSCHRequiredPowerValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_DSCHRequiredPowerPerUEInformation,
      { "hS-DSCHRequiredPowerPerUEInformation", "nbap.hS_DSCHRequiredPowerPerUEInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_DSCHRequiredPowerPerUEInformation_item,
      { "HS-DSCHRequiredPowerPerUEInformation-Item", "nbap.HS_DSCHRequiredPowerPerUEInformation_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cRNC_CommunicationContextID,
      { "cRNC-CommunicationContextID", "nbap.cRNC_CommunicationContextID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_DSCHRequiredPowerPerUEWeight,
      { "hS-DSCHRequiredPowerPerUEWeight", "nbap.hS_DSCHRequiredPowerPerUEWeight",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortion_item,
      { "HS-DSCHRequiredPowerValueInformation-For-CellPortion-Item", "nbap.HS_DSCHRequiredPowerValueInformation_For_CellPortion_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_DSCHRequiredPowerValue_01,
      { "hS-DSCHRequiredPowerValue", "nbap.hS_DSCHRequiredPowerValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HS_DSCHRequiredPower", HFILL }},
    { &hf_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR_item,
      { "HS-DSCHRequiredPowerValueInformation-For-CellPortionLCR-Item", "nbap.HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hsdpa_PICH_Shared_with_PCH,
      { "hsdpa-PICH-Shared-with-PCH", "nbap.hsdpa_PICH_Shared_with_PCH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hsdpa_PICH_notShared_with_PCH,
      { "hsdpa-PICH-notShared-with-PCH", "nbap.hsdpa_PICH_notShared_with_PCH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hsdpa_PICH_SharedPCH_ID,
      { "hsdpa-PICH-SharedPCH-ID", "nbap.hsdpa_PICH_SharedPCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CommonPhysicalChannelID", HFILL }},
    { &hf_nbap_hSDPA_PICH_notShared_ID,
      { "hSDPA-PICH-notShared-ID", "nbap.hSDPA_PICH_notShared_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CommonPhysicalChannelID", HFILL }},
    { &hf_nbap_fdd_DL_Channelisation_CodeNumber,
      { "fdd-DL-Channelisation-CodeNumber", "nbap.fdd_DL_Channelisation_CodeNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        "FDD_DL_ChannelisationCodeNumber", HFILL }},
    { &hf_nbap_pich_Power,
      { "pich-Power", "nbap.pich_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_pich_Mode,
      { "pich-Mode", "nbap.pich_Mode",
        FT_UINT32, BASE_DEC, VALS(nbap_PICH_Mode_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_sttd_Indicator,
      { "sttd-Indicator", "nbap.sttd_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_STTD_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_hsdsch_Common_Information,
      { "hsdsch-Common-Information", "nbap.hsdsch_Common_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_commonMACFlow_Specific_Information,
      { "commonMACFlow-Specific-Information", "nbap.commonMACFlow_Specific_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CommonMACFlow_Specific_InfoList", HFILL }},
    { &hf_nbap_hsSCCH_Specific_Information_ResponseFDD,
      { "hsSCCH-Specific-Information-ResponseFDD", "nbap.hsSCCH_Specific_Information_ResponseFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HSSCCH_Specific_InformationRespListFDD", HFILL }},
    { &hf_nbap_hARQ_MemoryPartitioning,
      { "hARQ-MemoryPartitioning", "nbap.hARQ_MemoryPartitioning",
        FT_UINT32, BASE_DEC, VALS(nbap_HARQ_MemoryPartitioning_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_commonMACFlow_Specific_Info_Response,
      { "commonMACFlow-Specific-Info-Response", "nbap.commonMACFlow_Specific_Info_Response",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CommonMACFlow_Specific_InfoList_Response", HFILL }},
    { &hf_nbap_cCCH_PriorityQueue_Id,
      { "cCCH-PriorityQueue-Id", "nbap.cCCH_PriorityQueue_Id",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PriorityQueue_Id", HFILL }},
    { &hf_nbap_sRB1_PriorityQueue_Id,
      { "sRB1-PriorityQueue-Id", "nbap.sRB1_PriorityQueue_Id",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PriorityQueue_Id", HFILL }},
    { &hf_nbap_associatedCommon_MACFlow,
      { "associatedCommon-MACFlow", "nbap.associatedCommon_MACFlow",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Common_MACFlow_ID", HFILL }},
    { &hf_nbap_fACH_Measurement_Occasion_Cycle_Length_Coefficient,
      { "fACH-Measurement-Occasion-Cycle-Length-Coefficient", "nbap.fACH_Measurement_Occasion_Cycle_Length_Coefficient",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_rACH_Measurement_Result,
      { "rACH-Measurement-Result", "nbap.rACH_Measurement_Result",
        FT_UINT32, BASE_DEC, VALS(nbap_RACH_Measurement_Result_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_bCCH_Specific_HSDSCH_RNTI_Information,
      { "bCCH-Specific-HSDSCH-RNTI-Information", "nbap.bCCH_Specific_HSDSCH_RNTI_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hSDSCH_MACdFlows_Information,
      { "hSDSCH-MACdFlows-Information", "nbap.hSDSCH_MACdFlows_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ueCapability_Info,
      { "ueCapability-Info", "nbap.ueCapability_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UE_Capability_Information", HFILL }},
    { &hf_nbap_mAChs_Reordering_Buffer_Size_for_RLC_UM,
      { "mAChs-Reordering-Buffer-Size-for-RLC-UM", "nbap.mAChs_Reordering_Buffer_Size_for_RLC_UM",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MAChsReorderingBufferSize_for_RLC_UM", HFILL }},
    { &hf_nbap_hsscch_PowerOffset,
      { "hsscch-PowerOffset", "nbap.hsscch_PowerOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tDD_AckNack_Power_Offset,
      { "tDD-AckNack-Power-Offset", "nbap.tDD_AckNack_Power_Offset",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hsDSCH_MACdFlow_Specific_Info_to_Modify,
      { "hsDSCH-MACdFlow-Specific-Info-to-Modify", "nbap.hsDSCH_MACdFlow_Specific_Info_to_Modify",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HSDSCH_MACdFlow_Specific_InfoList_to_Modify", HFILL }},
    { &hf_nbap_priorityQueueInfotoModify,
      { "priorityQueueInfotoModify", "nbap.priorityQueueInfotoModify",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PriorityQueue_InfoList_to_Modify", HFILL }},
    { &hf_nbap_hSSCCHCodeChangeGrant,
      { "hSSCCHCodeChangeGrant", "nbap.hSSCCHCodeChangeGrant",
        FT_UINT32, BASE_DEC, VALS(nbap_HSSCCH_Code_Change_Grant_vals), 0,
        "HSSCCH_Code_Change_Grant", HFILL }},
    { &hf_nbap_tDDAckNackPowerOffset,
      { "tDDAckNackPowerOffset", "nbap.tDDAckNackPowerOffset",
        FT_INT32, BASE_DEC, NULL, 0,
        "TDD_AckNack_Power_Offset", HFILL }},
    { &hf_nbap_HSDSCH_MACdFlow_Specific_InfoList_to_Modify_item,
      { "HSDSCH-MACdFlow-Specific-InfoItem-to-Modify", "nbap.HSDSCH_MACdFlow_Specific_InfoItem_to_Modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hsDSCH_MACdFlow_ID,
      { "hsDSCH-MACdFlow-ID", "nbap.hsDSCH_MACdFlow_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_priorityQueueInfotoModifyUnsynchronised,
      { "priorityQueueInfotoModifyUnsynchronised", "nbap.priorityQueueInfotoModifyUnsynchronised",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PriorityQueue_InfoList_to_Modify_Unsynchronised", HFILL }},
    { &hf_nbap_hsDSCH_MACdFlow_Specific_InformationResp,
      { "hsDSCH-MACdFlow-Specific-InformationResp", "nbap.hsDSCH_MACdFlow_Specific_InformationResp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sixtyfourQAM_UsageAllowedIndicator,
      { "sixtyfourQAM-UsageAllowedIndicator", "nbap.sixtyfourQAM_UsageAllowedIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_SixtyfourQAM_UsageAllowedIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_hSDSCH_RNTI,
      { "hSDSCH-RNTI", "nbap.hSDSCH_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sixtyfourQAM_DL_UsageIndicator,
      { "sixtyfourQAM-DL-UsageIndicator", "nbap.sixtyfourQAM_DL_UsageIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_SixtyfourQAM_DL_UsageIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_hSDSCH_TBSizeTableIndicator,
      { "hSDSCH-TBSizeTableIndicator", "nbap.hSDSCH_TBSizeTableIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_HSDSCH_TBSizeTableIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_hSSCCH_CodeChangeGrant,
      { "hSSCCH-CodeChangeGrant", "nbap.hSSCCH_CodeChangeGrant",
        FT_UINT32, BASE_DEC, VALS(nbap_HSSCCH_Code_Change_Grant_vals), 0,
        "HSSCCH_Code_Change_Grant", HFILL }},
    { &hf_nbap_hsSCCHCodeChangeIndicator,
      { "hsSCCHCodeChangeIndicator", "nbap.hsSCCHCodeChangeIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_HSSCCH_CodeChangeIndicator_vals), 0,
        "HSSCCH_CodeChangeIndicator", HFILL }},
    { &hf_nbap_hS_PDSCH_Code_Change_Indicator,
      { "hS-PDSCH-Code-Change-Indicator", "nbap.hS_PDSCH_Code_Change_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_HS_PDSCH_Code_Change_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_hS_DSCH_Secondary_Serving_cell_choice,
      { "hS-DSCH-Secondary-Serving-cell-choice", "nbap.hS_DSCH_Secondary_Serving_cell_choice",
        FT_UINT32, BASE_DEC, VALS(nbap_HS_DSCH_Secondary_Serving_cell_change_choice_vals), 0,
        "HS_DSCH_Secondary_Serving_cell_change_choice", HFILL }},
    { &hf_nbap_hS_Secondary_Serving_cell_change_successful,
      { "hS-Secondary-Serving-cell-change-successful", "nbap.hS_Secondary_Serving_cell_change_successful_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_Secondary_Serving_cell_change_unsuccessful,
      { "hS-Secondary-Serving-cell-change-unsuccessful", "nbap.hS_Secondary_Serving_cell_change_unsuccessful_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_DSCH_FDD_Secondary_Serving_Information_Response,
      { "hS-DSCH-FDD-Secondary-Serving-Information-Response", "nbap.hS_DSCH_FDD_Secondary_Serving_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_paging_MACFlow_Specific_Information,
      { "paging-MACFlow-Specific-Information", "nbap.paging_MACFlow_Specific_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_number_of_PCCH_transmission,
      { "number-of-PCCH-transmission", "nbap.number_of_PCCH_transmission",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_transport_Block_Size_List,
      { "transport-Block-Size-List", "nbap.transport_Block_Size_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_Paging_System_Information_ResponseFDD_item,
      { "HSDSCH-Paging-System-Information-ResponseList", "nbap.HSDSCH_Paging_System_Information_ResponseList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_pagingMACFlow_ID,
      { "pagingMACFlow-ID", "nbap.pagingMACFlow_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Paging_MACFlow_ID", HFILL }},
    { &hf_nbap_hSPDSCH_Code_Index,
      { "hSPDSCH-Code-Index", "nbap.hSPDSCH_Code_Index",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hsSCCH_Specific_Information_ResponseTDD,
      { "hsSCCH-Specific-Information-ResponseTDD", "nbap.hsSCCH_Specific_Information_ResponseTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HSSCCH_Specific_InformationRespListTDD", HFILL }},
    { &hf_nbap_hsSCCH_Specific_Information_ResponseTDDLCR,
      { "hsSCCH-Specific-Information-ResponseTDDLCR", "nbap.hsSCCH_Specific_Information_ResponseTDDLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HSSCCH_Specific_InformationRespListTDDLCR", HFILL }},
    { &hf_nbap_HSDSCH_MACdFlow_Specific_InformationResp_item,
      { "HSDSCH-MACdFlow-Specific-InformationResp-Item", "nbap.HSDSCH_MACdFlow_Specific_InformationResp_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hsDSCHMacdFlow_Id,
      { "hsDSCHMacdFlow-Id", "nbap.hsDSCHMacdFlow_Id",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HSDSCH_MACdFlow_ID", HFILL }},
    { &hf_nbap_hSDSCH_MACdFlow_Specific_Info,
      { "hSDSCH-MACdFlow-Specific-Info", "nbap.hSDSCH_MACdFlow_Specific_Info",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HSDSCH_MACdFlow_Specific_InfoList", HFILL }},
    { &hf_nbap_priorityQueue_Info,
      { "priorityQueue-Info", "nbap.priorityQueue_Info",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PriorityQueue_InfoList", HFILL }},
    { &hf_nbap_HSDSCH_MACdFlow_Specific_InfoList_item,
      { "HSDSCH-MACdFlow-Specific-InfoItem", "nbap.HSDSCH_MACdFlow_Specific_InfoItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_MACdFlows_to_Delete_item,
      { "HSDSCH-MACdFlows-to-Delete-Item", "nbap.HSDSCH_MACdFlows_to_Delete_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_Initial_Capacity_Allocation_item,
      { "HSDSCH-Initial-Capacity-AllocationItem", "nbap.HSDSCH_Initial_Capacity_AllocationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_maximum_MACdPDU_Size,
      { "maximum-MACdPDU-Size", "nbap.maximum_MACdPDU_Size",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MACdPDU_Size", HFILL }},
    { &hf_nbap_hSDSCH_InitialWindowSize,
      { "hSDSCH-InitialWindowSize", "nbap.hSDSCH_InitialWindowSize",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_setsOfHS_SCCH_Codes,
      { "setsOfHS-SCCH-Codes", "nbap.setsOfHS_SCCH_Codes",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hARQ_Preamble_Mode_Activation_Indicator,
      { "hARQ-Preamble-Mode-Activation-Indicator", "nbap.hARQ_Preamble_Mode_Activation_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_HARQ_Preamble_Mode_Activation_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_mIMO_N_M_Ratio,
      { "mIMO-N-M-Ratio", "nbap.mIMO_N_M_Ratio",
        FT_UINT32, BASE_DEC, VALS(nbap_MIMO_N_M_Ratio_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_continuousPacketConnectivityHS_SCCH_less_Information_Response,
      { "continuousPacketConnectivityHS-SCCH-less-Information-Response", "nbap.continuousPacketConnectivityHS_SCCH_less_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_EDCH_Preconfiguration_Information_item,
      { "Additional-EDCH-Preconfiguration-Information-ItemIEs", "nbap.Additional_EDCH_Preconfiguration_Information_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_mAChsResetScheme,
      { "mAChsResetScheme", "nbap.mAChsResetScheme",
        FT_UINT32, BASE_DEC, VALS(nbap_MAChsResetScheme_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_hSDSCH_Physical_Layer_Category,
      { "hSDSCH-Physical-Layer-Category", "nbap.hSDSCH_Physical_Layer_Category",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_64_", HFILL }},
    { &hf_nbap_secondaryServingCells,
      { "secondaryServingCells", "nbap.secondaryServingCells",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_numPrimaryHS_SCCH_Codes,
      { "numPrimaryHS-SCCH-Codes", "nbap.numPrimaryHS_SCCH_Codes",
        FT_UINT32, BASE_DEC, NULL, 0,
        "NumHS_SCCH_Codes", HFILL }},
    { &hf_nbap_hARQ_Preamble_Mode,
      { "hARQ-Preamble-Mode", "nbap.hARQ_Preamble_Mode",
        FT_UINT32, BASE_DEC, VALS(nbap_HARQ_Preamble_Mode_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_mIMO_ActivationIndicator,
      { "mIMO-ActivationIndicator", "nbap.mIMO_ActivationIndicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hSDSCH_MACdPDUSizeFormat,
      { "hSDSCH-MACdPDUSizeFormat", "nbap.hSDSCH_MACdPDUSizeFormat",
        FT_UINT32, BASE_DEC, VALS(nbap_HSDSCH_MACdPDUSizeFormat_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_uE_with_enhanced_HS_SCCH_support_indicator,
      { "uE-with-enhanced-HS-SCCH-support-indicator", "nbap.uE_with_enhanced_HS_SCCH_support_indicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_SCCH_PreconfiguredCodes_item,
      { "HS-SCCH-PreconfiguredCodesItem", "nbap.HS_SCCH_PreconfiguredCodesItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_SCCH_CodeNumber,
      { "hS-SCCH-CodeNumber", "nbap.hS_SCCH_CodeNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSSCCH_Specific_InformationRespListFDD_item,
      { "HSSCCH-Codes", "nbap.HSSCCH_Codes_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_codeNumber,
      { "codeNumber", "nbap.codeNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_127", HFILL }},
    { &hf_nbap_HSSCCH_Specific_InformationRespListTDD_item,
      { "HSSCCH-Specific-InformationRespItemTDD", "nbap.HSSCCH_Specific_InformationRespItemTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_timeslot,
      { "timeslot", "nbap.timeslot",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tDD_ChannelisationCode,
      { "tDD-ChannelisationCode", "nbap.tDD_ChannelisationCode",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_TDD_ChannelisationCode_vals_ext, 0,
        NULL, HFILL }},
    { &hf_nbap_hSSICH_Info,
      { "hSSICH-Info", "nbap.hSSICH_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSSCCH_Specific_InformationRespListTDDLCR_item,
      { "HSSCCH-Specific-InformationRespItemTDDLCR", "nbap.HSSCCH_Specific_InformationRespItemTDDLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_timeslotLCR,
      { "timeslotLCR", "nbap.timeslotLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_first_TDD_ChannelisationCode,
      { "first-TDD-ChannelisationCode", "nbap.first_TDD_ChannelisationCode",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_TDD_ChannelisationCode_vals_ext, 0,
        "TDD_ChannelisationCode", HFILL }},
    { &hf_nbap_second_TDD_ChannelisationCode,
      { "second-TDD-ChannelisationCode", "nbap.second_TDD_ChannelisationCode",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_TDD_ChannelisationCode_vals_ext, 0,
        "TDD_ChannelisationCode", HFILL }},
    { &hf_nbap_hSSICH_InfoLCR,
      { "hSSICH-InfoLCR", "nbap.hSSICH_InfoLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSSCCH_Specific_InformationRespListTDD768_item,
      { "HSSCCH-Specific-InformationRespItemTDD768", "nbap.HSSCCH_Specific_InformationRespItemTDD768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tDD_ChannelisationCode768,
      { "tDD-ChannelisationCode768", "nbap.tDD_ChannelisationCode768",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_TDD_ChannelisationCode768_vals_ext, 0,
        NULL, HFILL }},
    { &hf_nbap_hSSICH_Info768,
      { "hSSICH-Info768", "nbap.hSSICH_Info768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hsSICH_ID,
      { "hsSICH-ID", "nbap.hsSICH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HS_SICH_ID", HFILL }},
    { &hf_nbap_failed_HS_SICH,
      { "failed-HS-SICH", "nbap.failed_HS_SICH",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HS_SICH_failed", HFILL }},
    { &hf_nbap_missed_HS_SICH,
      { "missed-HS-SICH", "nbap.missed_HS_SICH",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HS_SICH_missed", HFILL }},
    { &hf_nbap_total_HS_SICH,
      { "total-HS-SICH", "nbap.total_HS_SICH",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HS_SICH_total", HFILL }},
    { &hf_nbap_number_of_HS_PDSCH_codes,
      { "number-of-HS-PDSCH-codes", "nbap.number_of_HS_PDSCH_codes",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_maxHS_PDSCHCodeNrComp_1", HFILL }},
    { &hf_nbap_hS_PDSCH_Start_code_number,
      { "hS-PDSCH-Start-code-number", "nbap.hS_PDSCH_Start_code_number",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_replace_02,
      { "replace", "nbap.replace",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HS_SCCH_FDD_Code_List", HFILL }},
    { &hf_nbap_HS_SCCH_FDD_Code_List_item,
      { "HS-SCCH-FDD-Code-Information-Item", "nbap.HS_SCCH_FDD_Code_Information_Item",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hspdsch_RL_ID,
      { "hspdsch-RL-ID", "nbap.hspdsch_RL_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RL_ID", HFILL }},
    { &hf_nbap_hSDSCH_FDD_Information,
      { "hSDSCH-FDD-Information", "nbap.hSDSCH_FDD_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hsdsch_RNTI,
      { "hsdsch-RNTI", "nbap.hsdsch_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_DSCH_serving_cell_choice,
      { "hS-DSCH-serving-cell-choice", "nbap.hS_DSCH_serving_cell_choice",
        FT_UINT32, BASE_DEC, VALS(nbap_HS_DSCH_serving_cell_choice_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_hS_serving_cell_change_successful,
      { "hS-serving-cell-change-successful", "nbap.hS_serving_cell_change_successful_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_serving_cell_change_unsuccessful,
      { "hS-serving-cell-change-unsuccessful", "nbap.hS_serving_cell_change_unsuccessful_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hSDSCH_FDD_Information_Response,
      { "hSDSCH-FDD-Information-Response", "nbap.hSDSCH_FDD_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hsdpa_PICH_notShared_with_PCHLCR,
      { "hsdpa-PICH-notShared-with-PCHLCR", "nbap.hsdpa_PICH_notShared_with_PCHLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tdd_ChannelisationCodeLCR,
      { "tdd-ChannelisationCodeLCR", "nbap.tdd_ChannelisationCodeLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tdd_PhysicalChannelOffset,
      { "tdd-PhysicalChannelOffset", "nbap.tdd_PhysicalChannelOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_pagingIndicatorLength,
      { "pagingIndicatorLength", "nbap.pagingIndicatorLength",
        FT_UINT32, BASE_DEC, VALS(nbap_PagingIndicatorLength_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_pICH_Power,
      { "pICH-Power", "nbap.pICH_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_second_TDD_ChannelisationCodeLCR,
      { "second-TDD-ChannelisationCodeLCR", "nbap.second_TDD_ChannelisationCodeLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "TDD_ChannelisationCodeLCR", HFILL }},
    { &hf_nbap_hsdsch_Common_InformationLCR,
      { "hsdsch-Common-InformationLCR", "nbap.hsdsch_Common_InformationLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_commonMACFlow_Specific_InformationLCR,
      { "commonMACFlow-Specific-InformationLCR", "nbap.commonMACFlow_Specific_InformationLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CommonMACFlow_Specific_InfoListLCR", HFILL }},
    { &hf_nbap_common_H_RNTI_InformationLCR,
      { "common-H-RNTI-InformationLCR", "nbap.common_H_RNTI_InformationLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sync_InformationLCR,
      { "sync-InformationLCR", "nbap.sync_InformationLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hSSICH_SIRTarget,
      { "hSSICH-SIRTarget", "nbap.hSSICH_SIRTarget",
        FT_INT32, BASE_DEC, NULL, 0,
        "UL_SIR", HFILL }},
    { &hf_nbap_hSSICH_TPC_StepSize,
      { "hSSICH-TPC-StepSize", "nbap.hSSICH_TPC_StepSize",
        FT_UINT32, BASE_DEC, VALS(nbap_TDD_TPC_UplinkStepSize_LCR_vals), 0,
        "TDD_TPC_UplinkStepSize_LCR", HFILL }},
    { &hf_nbap_hsSCCH_Specific_Information_ResponseLCR,
      { "hsSCCH-Specific-Information-ResponseLCR", "nbap.hsSCCH_Specific_Information_ResponseLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HSSCCH_Specific_InformationRespListLCR", HFILL }},
    { &hf_nbap_commonMACFlow_Specific_Info_ResponseLCR,
      { "commonMACFlow-Specific-Info-ResponseLCR", "nbap.commonMACFlow_Specific_Info_ResponseLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CommonMACFlow_Specific_InfoList_ResponseLCR", HFILL }},
    { &hf_nbap_associatedCommon_MACFlowLCR,
      { "associatedCommon-MACFlowLCR", "nbap.associatedCommon_MACFlowLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Common_MACFlow_ID_LCR", HFILL }},
    { &hf_nbap_bCCH_Specific_HSDSCH_RNTI_InformationLCR,
      { "bCCH-Specific-HSDSCH-RNTI-InformationLCR", "nbap.bCCH_Specific_HSDSCH_RNTI_InformationLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_paging_MACFlow_Specific_InformationLCR,
      { "paging-MACFlow-Specific-InformationLCR", "nbap.paging_MACFlow_Specific_InformationLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_reception_Window_Size,
      { "reception-Window-Size", "nbap.reception_Window_Size",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_16", HFILL }},
    { &hf_nbap_n_PCH,
      { "n-PCH", "nbap.n_PCH",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_8", HFILL }},
    { &hf_nbap_paging_Subchannel_Size,
      { "paging-Subchannel-Size", "nbap.paging_Subchannel_Size",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_3", HFILL }},
    { &hf_nbap_HSDSCH_Paging_System_Information_ResponseLCR_item,
      { "HSDSCH-Paging-System-Information-ResponseListLCR", "nbap.HSDSCH_Paging_System_Information_ResponseListLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst,
      { "dL-HS-PDSCH-Timeslot-Information-LCR-PSCH-ReconfRqst", "nbap.dL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSSCCH_Specific_InformationRespListLCR_item,
      { "HSSCCH-Specific-InformationRespItemLCR", "nbap.HSSCCH_Specific_InformationRespItemLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_SCCH_ID_LCR,
      { "hS-SCCH-ID-LCR", "nbap.hS_SCCH_ID_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_transport_Block_Size_List_01,
      { "transport-Block-Size-List", "nbap.transport_Block_Size_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Transport_Block_Size_List_LCR", HFILL }},
    { &hf_nbap_hS_DSCH_SPS_Reservation_Indicator,
      { "hS-DSCH-SPS-Reservation-Indicator", "nbap.hS_DSCH_SPS_Reservation_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_SPS_Reservation_Indicator_vals), 0,
        "SPS_Reservation_Indicator", HFILL }},
    { &hf_nbap_hS_DSCH_SPS_Operation_Indicator,
      { "hS-DSCH-SPS-Operation-Indicator", "nbap.hS_DSCH_SPS_Operation_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_HS_DSCH_SPS_Operation_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Transport_Block_Size_List_LCR_item,
      { "Transport-Block-Size-Item-LCR", "nbap.Transport_Block_Size_Item_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_transport_Block_Size_maping_Index_LCR,
      { "transport-Block-Size-maping-Index-LCR", "nbap.transport_Block_Size_maping_Index_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_transport_Block_Size_Index_LCR,
      { "transport-Block-Size-Index-LCR", "nbap.transport_Block_Size_Index_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Repetition_Period_List_LCR_item,
      { "Repetition-Period-Item-LCR", "nbap.Repetition_Period_Item_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_repetitionPeriodIndex,
      { "repetitionPeriodIndex", "nbap.repetitionPeriodIndex",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_logicalChannellevel,
      { "logicalChannellevel", "nbap.logicalChannellevel",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_priorityQueuelevel,
      { "priorityQueuelevel", "nbap.priorityQueuelevel",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_SICH_InformationList_for_HS_DSCH_SPS,
      { "hS-SICH-InformationList-for-HS-DSCH-SPS", "nbap.hS_SICH_InformationList_for_HS_DSCH_SPS",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_initial_HS_PDSCH_SPS_Resource,
      { "initial-HS-PDSCH-SPS-Resource", "nbap.initial_HS_PDSCH_SPS_Resource_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_buffer_Size_for_HS_DSCH_SPS,
      { "buffer-Size-for-HS-DSCH-SPS", "nbap.buffer_Size_for_HS_DSCH_SPS",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_Process_Memory_Size_vals_ext, 0,
        "Process_Memory_Size", HFILL }},
    { &hf_nbap_number_of_Processes_for_HS_DSCH_SPS,
      { "number-of-Processes-for-HS-DSCH-SPS", "nbap.number_of_Processes_for_HS_DSCH_SPS",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_SICH_InformationList_for_HS_DSCH_SPS_item,
      { "HS-SICH-InformationItem-for-HS-DSCH-SPS", "nbap.HS_SICH_InformationItem_for_HS_DSCH_SPS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_SICH_Mapping_Index,
      { "hS-SICH-Mapping-Index", "nbap.hS_SICH_Mapping_Index",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_SICH_Type,
      { "hS-SICH-Type", "nbap.hS_SICH_Type",
        FT_UINT32, BASE_DEC, VALS(nbap_HS_SICH_Type_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_hS_SCCH_Associated_HS_SICH,
      { "hS-SCCH-Associated-HS-SICH", "nbap.hS_SCCH_Associated_HS_SICH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_non_HS_SCCH_Associated_HS_SICH,
      { "non-HS-SCCH-Associated-HS-SICH", "nbap.non_HS_SCCH_Associated_HS_SICH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_extended_HS_SICH_ID,
      { "extended-HS-SICH-ID", "nbap.extended_HS_SICH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_non_HS_SCCH_Aassociated_HS_SICH_ID,
      { "non-HS-SCCH-Aassociated-HS-SICH-ID", "nbap.non_HS_SCCH_Aassociated_HS_SICH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_PDSCH_Offset,
      { "hS-PDSCH-Offset", "nbap.hS_PDSCH_Offset",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TDD_PhysicalChannelOffset", HFILL }},
    { &hf_nbap_timeslot_Resource_Related_Information_01,
      { "timeslot-Resource-Related-Information", "nbap.timeslot_Resource_Related_Information",
        FT_BYTES, BASE_NONE, NULL, 0,
        "HS_DSCH_TimeslotResourceLCR", HFILL }},
    { &hf_nbap_startCode,
      { "startCode", "nbap.startCode",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_TDD_ChannelisationCode_vals_ext, 0,
        "TDD_ChannelisationCode", HFILL }},
    { &hf_nbap_endCode,
      { "endCode", "nbap.endCode",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_TDD_ChannelisationCode_vals_ext, 0,
        "TDD_ChannelisationCode", HFILL }},
    { &hf_nbap_transport_Block_Size_Index_01,
      { "transport-Block-Size-Index", "nbap.transport_Block_Size_Index",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Transport_Block_Size_Index_LCR", HFILL }},
    { &hf_nbap_modulationType,
      { "modulationType", "nbap.modulationType",
        FT_UINT32, BASE_DEC, VALS(nbap_ModulationSPS_LCR_vals), 0,
        "ModulationSPS_LCR", HFILL }},
    { &hf_nbap_non_HS_SCCH_Associated_HS_SICH_InformationList,
      { "non-HS-SCCH-Associated-HS-SICH-InformationList", "nbap.non_HS_SCCH_Associated_HS_SICH_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationList_item,
      { "Non-HS-SCCH-Associated-HS-SICH-InformationItem", "nbap.Non_HS_SCCH_Associated_HS_SICH_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext_item,
      { "Non-HS-SCCH-Associated-HS-SICH-InformationItem", "nbap.Non_HS_SCCH_Associated_HS_SICH_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_modify_non_HS_SCCH_Associated_HS_SICH_InformationList,
      { "modify-non-HS-SCCH-Associated-HS-SICH-InformationList", "nbap.modify_non_HS_SCCH_Associated_HS_SICH_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList_item,
      { "Modify-Non-HS-SCCH-Associated-HS-SICH-InformationItem", "nbap.Modify_Non_HS_SCCH_Associated_HS_SICH_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext_item,
      { "Modify-Non-HS-SCCH-Associated-HS-SICH-InformationItem", "nbap.Modify_Non_HS_SCCH_Associated_HS_SICH_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_item,
      { "Delete-From-Non-HS-SCCH-Associated-HS-SICH-Resource-Pool-LCR-PSCH-ReconfRqstItem", "nbap.Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqstItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext_item,
      { "Delete-From-Non-HS-SCCH-Associated-HS-SICH-Resource-Pool-LCR-PSCH-ReconfRqstItem", "nbap.Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqstItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MIMO_ReferenceSignal_InformationListLCR_item,
      { "HSSICH-ReferenceSignal-InformationLCR", "nbap.HSSICH_ReferenceSignal_InformationLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_midambleConfigurationLCR,
      { "midambleConfigurationLCR", "nbap.midambleConfigurationLCR",
        FT_UINT32, BASE_DEC, VALS(nbap_MidambleConfigurationLCR_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_midambleShift,
      { "midambleShift", "nbap.midambleShift",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_15", HFILL }},
    { &hf_nbap_hSSICH_ReferenceSignal_InformationLCR,
      { "hSSICH-ReferenceSignal-InformationLCR", "nbap.hSSICH_ReferenceSignal_InformationLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sub_Frame_Number,
      { "sub-Frame-Number", "nbap.sub_Frame_Number",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_onDemand,
      { "onDemand", "nbap.onDemand_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_periodic,
      { "periodic", "nbap.periodic",
        FT_UINT32, BASE_DEC, VALS(nbap_InformationReportCharacteristicsType_ReportPeriodicity_vals), 0,
        "InformationReportCharacteristicsType_ReportPeriodicity", HFILL }},
    { &hf_nbap_onModification,
      { "onModification", "nbap.onModification_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "InformationReportCharacteristicsType_OnModification", HFILL }},
    { &hf_nbap_min,
      { "min", "nbap.min",
        FT_UINT32, BASE_DEC, NULL, 0,
        "ReportPeriodicity_Scaledmin", HFILL }},
    { &hf_nbap_hours,
      { "hours", "nbap.hours",
        FT_UINT32, BASE_DEC, NULL, 0,
        "ReportPeriodicity_Scaledhour", HFILL }},
    { &hf_nbap_information_thresholds,
      { "information-thresholds", "nbap.information_thresholds",
        FT_UINT32, BASE_DEC, VALS(nbap_InformationThresholds_vals), 0,
        "InformationThresholds", HFILL }},
    { &hf_nbap_dgps,
      { "dgps", "nbap.dgps_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DGPSThresholds", HFILL }},
    { &hf_nbap_dGANSSThreshold,
      { "dGANSSThreshold", "nbap.dGANSSThreshold_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_information_Type_Item,
      { "information-Type-Item", "nbap.information_Type_Item",
        FT_UINT32, BASE_DEC, VALS(nbap_Information_Type_Item_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_gPSInformation,
      { "gPSInformation", "nbap.gPSInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GPS_Information", HFILL }},
    { &hf_nbap_iP_SpacingFDD,
      { "iP-SpacingFDD", "nbap.iP_SpacingFDD",
        FT_UINT32, BASE_DEC, VALS(nbap_T_iP_SpacingFDD_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_iP_Length,
      { "iP-Length", "nbap.iP_Length",
        FT_UINT32, BASE_DEC, VALS(nbap_T_iP_Length_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_seed,
      { "seed", "nbap.seed",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_63", HFILL }},
    { &hf_nbap_burstModeParams,
      { "burstModeParams", "nbap.burstModeParams_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_iP_Offset,
      { "iP-Offset", "nbap.iP_Offset",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_9", HFILL }},
    { &hf_nbap_iP_SpacingTDD,
      { "iP-SpacingTDD", "nbap.iP_SpacingTDD",
        FT_UINT32, BASE_DEC, VALS(nbap_T_iP_SpacingTDD_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_iP_Start,
      { "iP-Start", "nbap.iP_Start",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_4095", HFILL }},
    { &hf_nbap_iP_Slot,
      { "iP-Slot", "nbap.iP_Slot",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_14", HFILL }},
    { &hf_nbap_iP_PCCPCH,
      { "iP-PCCPCH", "nbap.iP_PCCPCH",
        FT_UINT32, BASE_DEC, VALS(nbap_T_iP_PCCPCH_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_iP_SpacingTDD_01,
      { "iP-SpacingTDD", "nbap.iP_SpacingTDD",
        FT_UINT32, BASE_DEC, VALS(nbap_T_iP_SpacingTDD_01_vals), 0,
        "T_iP_SpacingTDD_01", HFILL }},
    { &hf_nbap_iP_Sub,
      { "iP-Sub", "nbap.iP_Sub",
        FT_UINT32, BASE_DEC, VALS(nbap_T_iP_Sub_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_cFNOffset,
      { "cFNOffset", "nbap.cFNOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_255", HFILL }},
    { &hf_nbap_burstStart,
      { "burstStart", "nbap.burstStart",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_15", HFILL }},
    { &hf_nbap_burstLength,
      { "burstLength", "nbap.burstLength",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_10_25", HFILL }},
    { &hf_nbap_burstFreq,
      { "burstFreq", "nbap.burstFreq",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_16", HFILL }},
    { &hf_nbap_idleIntervalInfo_k,
      { "idleIntervalInfo-k", "nbap.idleIntervalInfo_k",
        FT_UINT32, BASE_DEC, VALS(nbap_T_idleIntervalInfo_k_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_idleIntervalInfo_offset,
      { "idleIntervalInfo-offset", "nbap.idleIntervalInfo_offset",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_7", HFILL }},
    { &hf_nbap_maxTimeslotsPerSubFrame,
      { "maxTimeslotsPerSubFrame", "nbap.maxTimeslotsPerSubFrame",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_6", HFILL }},
    { &hf_nbap_maxPhysChPerTimeslot,
      { "maxPhysChPerTimeslot", "nbap.maxPhysChPerTimeslot",
        FT_UINT32, BASE_DEC, VALS(nbap_T_maxPhysChPerTimeslot_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_MACdPDU_Size_Indexlist_item,
      { "MACdPDU-Size-IndexItem", "nbap.MACdPDU_Size_IndexItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sID,
      { "sID", "nbap.sID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_macdPDU_Size,
      { "macdPDU-Size", "nbap.macdPDU_Size",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MACdPDU_Size_Indexlist_to_Modify_item,
      { "MACdPDU-Size-IndexItem-to-Modify", "nbap.MACdPDU_Size_IndexItem_to_Modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_maximum_TB_Size_cell_edge_users,
      { "maximum-TB-Size-cell-edge-users", "nbap.maximum_TB_Size_cell_edge_users",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_5000_", HFILL }},
    { &hf_nbap_maximum_TB_Size_other_users,
      { "maximum-TB-Size-other-users", "nbap.maximum_TB_Size_other_users",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_5000_", HFILL }},
    { &hf_nbap_MessageStructure_item,
      { "MessageStructure item", "nbap.MessageStructure_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_repetitionNumber_01,
      { "repetitionNumber", "nbap.repetitionNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RepetitionNumber1", HFILL }},
    { &hf_nbap_type1,
      { "type1", "nbap.type1_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_midambleConfigurationBurstType1And3,
      { "midambleConfigurationBurstType1And3", "nbap.midambleConfigurationBurstType1And3",
        FT_UINT32, BASE_DEC, VALS(nbap_MidambleConfigurationBurstType1And3_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_midambleAllocationMode,
      { "midambleAllocationMode", "nbap.midambleAllocationMode",
        FT_UINT32, BASE_DEC, VALS(nbap_MidambleAllocationMode1_vals), 0,
        "MidambleAllocationMode1", HFILL }},
    { &hf_nbap_defaultMidamble,
      { "defaultMidamble", "nbap.defaultMidamble_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_commonMidamble,
      { "commonMidamble", "nbap.commonMidamble_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ueSpecificMidamble,
      { "ueSpecificMidamble", "nbap.ueSpecificMidamble",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MidambleShiftLong", HFILL }},
    { &hf_nbap_type2,
      { "type2", "nbap.type2_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_midambleConfigurationBurstType2,
      { "midambleConfigurationBurstType2", "nbap.midambleConfigurationBurstType2",
        FT_UINT32, BASE_DEC, VALS(nbap_MidambleConfigurationBurstType2_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_midambleAllocationMode_01,
      { "midambleAllocationMode", "nbap.midambleAllocationMode",
        FT_UINT32, BASE_DEC, VALS(nbap_MidambleAllocationMode2_vals), 0,
        "MidambleAllocationMode2", HFILL }},
    { &hf_nbap_ueSpecificMidamble_01,
      { "ueSpecificMidamble", "nbap.ueSpecificMidamble",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MidambleShiftShort", HFILL }},
    { &hf_nbap_type3,
      { "type3", "nbap.type3_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_midambleAllocationMode_02,
      { "midambleAllocationMode", "nbap.midambleAllocationMode",
        FT_UINT32, BASE_DEC, VALS(nbap_MidambleAllocationMode3_vals), 0,
        "MidambleAllocationMode3", HFILL }},
    { &hf_nbap_midambleAllocationMode_03,
      { "midambleAllocationMode", "nbap.midambleAllocationMode",
        FT_UINT32, BASE_DEC, VALS(nbap_MidambleAllocationMode_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_midambleShift_01,
      { "midambleShift", "nbap.midambleShift",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MidambleShiftLong", HFILL }},
    { &hf_nbap_type1_01,
      { "type1", "nbap.type1_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Type7681", HFILL }},
    { &hf_nbap_midambleAllocationMode_04,
      { "midambleAllocationMode", "nbap.midambleAllocationMode",
        FT_UINT32, BASE_DEC, VALS(nbap_MidambleAllocationMode7681_vals), 0,
        "MidambleAllocationMode7681", HFILL }},
    { &hf_nbap_type2_01,
      { "type2", "nbap.type2_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Type7682", HFILL }},
    { &hf_nbap_midambleConfigurationBurstType2_768,
      { "midambleConfigurationBurstType2-768", "nbap.midambleConfigurationBurstType2_768",
        FT_UINT32, BASE_DEC, VALS(nbap_MidambleConfigurationBurstType2_768_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_midambleAllocationMode_05,
      { "midambleAllocationMode", "nbap.midambleAllocationMode",
        FT_UINT32, BASE_DEC, VALS(nbap_MidambleAllocationMode7682_vals), 0,
        "MidambleAllocationMode7682", HFILL }},
    { &hf_nbap_ueSpecificMidamble_02,
      { "ueSpecificMidamble", "nbap.ueSpecificMidamble",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MidambleShiftShort768", HFILL }},
    { &hf_nbap_type3_01,
      { "type3", "nbap.type3_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Type7683", HFILL }},
    { &hf_nbap_midambleAllocationMode_06,
      { "midambleAllocationMode", "nbap.midambleAllocationMode",
        FT_UINT32, BASE_DEC, VALS(nbap_MidambleAllocationMode7683_vals), 0,
        "MidambleAllocationMode7683", HFILL }},
    { &hf_nbap_primary_and_secondary_CPICH,
      { "primary-and-secondary-CPICH", "nbap.primary_and_secondary_CPICH",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CommonPhysicalChannelID", HFILL }},
    { &hf_nbap_normal_and_diversity_primary_CPICH,
      { "normal-and-diversity-primary-CPICH", "nbap.normal_and_diversity_primary_CPICH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_primary_and_secondary_CPICH_01,
      { "primary-and-secondary-CPICH", "nbap.primary_and_secondary_CPICH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PrimaryAndSecondaryCPICHContainer", HFILL }},
    { &hf_nbap_normal_and_diversity_primary_CPICH_01,
      { "normal-and-diversity-primary-CPICH", "nbap.normal_and_diversity_primary_CPICH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "NormalAndDiversityPrimaryCPICHContainer", HFILL }},
    { &hf_nbap_addPriorityQueue,
      { "addPriorityQueue", "nbap.addPriorityQueue_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PriorityQueue_InfoItem_to_Add", HFILL }},
    { &hf_nbap_modifyPriorityQueue,
      { "modifyPriorityQueue", "nbap.modifyPriorityQueue_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PriorityQueue_InfoItem_to_Modify", HFILL }},
    { &hf_nbap_deletePriorityQueue,
      { "deletePriorityQueue", "nbap.deletePriorityQueue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PriorityQueue_Id", HFILL }},
    { &hf_nbap_qPSK,
      { "qPSK", "nbap.qPSK_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sixteenQAM,
      { "sixteenQAM", "nbap.sixteenQAM",
        FT_INT32, BASE_DEC, NULL, 0,
        "MBSFN_CPICH_secondary_CCPCH_power_offset", HFILL }},
    { &hf_nbap_MultipleFreq_HARQ_MemoryPartitioning_InformationList_item,
      { "MultipleFreq-HARQ-MemoryPartitioning-InformationItem", "nbap.MultipleFreq_HARQ_MemoryPartitioning_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR_item,
      { "MultipleFreq-HSPDSCH-InformationItem-ResponseTDDLCR", "nbap.MultipleFreq_HSPDSCH_InformationItem_ResponseTDDLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_multi_Cell_Capability,
      { "multi-Cell-Capability", "nbap.multi_Cell_Capability",
        FT_UINT32, BASE_DEC, VALS(nbap_Multi_Cell_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_dL_PowerBalancing_Information,
      { "dL-PowerBalancing-Information", "nbap.dL_PowerBalancing_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_minimumReducedE_DPDCH_GainFactor,
      { "minimumReducedE-DPDCH-GainFactor", "nbap.minimumReducedE_DPDCH_GainFactor",
        FT_UINT32, BASE_DEC, VALS(nbap_MinimumReducedE_DPDCH_GainFactor_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_secondary_UL_Frequency_Activation_State,
      { "secondary-UL-Frequency-Activation-State", "nbap.secondary_UL_Frequency_Activation_State",
        FT_UINT32, BASE_DEC, VALS(nbap_Secondary_UL_Frequency_Activation_State_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_extendedPropagationDelay,
      { "extendedPropagationDelay", "nbap.extendedPropagationDelay",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_primary_CPICH_Usage_for_Channel_Estimation,
      { "primary-CPICH-Usage-for-Channel-Estimation", "nbap.primary_CPICH_Usage_for_Channel_Estimation",
        FT_UINT32, BASE_DEC, VALS(nbap_Primary_CPICH_Usage_for_Channel_Estimation_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_secondary_CPICH_Information,
      { "secondary-CPICH-Information", "nbap.secondary_CPICH_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CommonPhysicalChannelID", HFILL }},
    { &hf_nbap_secondary_CPICH_Information_Change,
      { "secondary-CPICH-Information-Change", "nbap.secondary_CPICH_Information_Change",
        FT_UINT32, BASE_DEC, VALS(nbap_Secondary_CPICH_Information_Change_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_e_AGCH_PowerOffset,
      { "e-AGCH-PowerOffset", "nbap.e_AGCH_PowerOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_RGCH_PowerOffset,
      { "e-RGCH-PowerOffset", "nbap.e_RGCH_PowerOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_HICH_PowerOffset,
      { "e-HICH-PowerOffset", "nbap.e_HICH_PowerOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_DCH_DL_Control_Channel_Grant,
      { "e-DCH-DL-Control-Channel-Grant", "nbap.e_DCH_DL_Control_Channel_Grant_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_NeighbouringCellMeasurementInformation_item,
      { "NeighbouringCellMeasurementInformation item", "nbap.NeighbouringCellMeasurementInformation_item",
        FT_UINT32, BASE_DEC, VALS(nbap_NeighbouringCellMeasurementInformation_item_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_neighbouringFDDCellMeasurementInformation,
      { "neighbouringFDDCellMeasurementInformation", "nbap.neighbouringFDDCellMeasurementInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_neighbouringTDDCellMeasurementInformation,
      { "neighbouringTDDCellMeasurementInformation", "nbap.neighbouringTDDCellMeasurementInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_extension_neighbouringCellMeasurementInformation,
      { "extension-neighbouringCellMeasurementInformation", "nbap.extension_neighbouringCellMeasurementInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uC_Id,
      { "uC-Id", "nbap.uC_Id_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_primaryScramblingCode,
      { "primaryScramblingCode", "nbap.primaryScramblingCode",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cellParameterID,
      { "cellParameterID", "nbap.cellParameterID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_NI_Information_item,
      { "Notification-Indicator", "nbap.Notification_Indicator",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Paging_MACFlows_to_DeleteFDD_item,
      { "Paging-MACFlows-to-DeleteFDD-Item", "nbap.Paging_MACFlows_to_DeleteFDD_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_paging_MACFlow_ID,
      { "paging-MACFlow-ID", "nbap.paging_MACFlow_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Paging_MACFlow_Specific_Information_item,
      { "Paging-MAC-Flow-Specific-Information-Item", "nbap.Paging_MAC_Flow_Specific_Information_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_paging_MACFlow_Id,
      { "paging-MACFlow-Id", "nbap.paging_MACFlow_Id",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hSDPA_associated_PICH_Info,
      { "hSDPA-associated-PICH-Info", "nbap.hSDPA_associated_PICH_Info",
        FT_UINT32, BASE_DEC, VALS(nbap_HSDPA_Associated_PICH_Information_vals), 0,
        "HSDPA_Associated_PICH_Information", HFILL }},
    { &hf_nbap_paging_MACFlow_PriorityQueue_Information,
      { "paging-MACFlow-PriorityQueue-Information", "nbap.paging_MACFlow_PriorityQueue_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Paging_MACFlow_PriorityQueue_Information_item,
      { "Paging-MACFlow-PriorityQueue-Item", "nbap.Paging_MACFlow_PriorityQueue_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_priority_Queue_Information_for_Enhanced_PCH,
      { "priority-Queue-Information-for-Enhanced-PCH", "nbap.priority_Queue_Information_for_Enhanced_PCH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Priority_Queue_Information_for_Enhanced_FACH_PCH", HFILL }},
    { &hf_nbap_sequenceNumber,
      { "sequenceNumber", "nbap.sequenceNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PLCCHsequenceNumber", HFILL }},
    { &hf_nbap_Possible_Secondary_Serving_Cell_List_item,
      { "Possible-Secondary-Serving-Cell", "nbap.Possible_Secondary_Serving_Cell_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_local_Cell_ID,
      { "local-Cell-ID", "nbap.local_Cell_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_power_Offset_For_Secondary_CPICH_for_MIMO,
      { "power-Offset-For-Secondary-CPICH-for-MIMO", "nbap.power_Offset_For_Secondary_CPICH_for_MIMO",
        FT_INT32, BASE_DEC, NULL, 0,
        "PowerOffsetForSecondaryCPICHforMIMO", HFILL }},
    { &hf_nbap_priorityQueue_Id,
      { "priorityQueue-Id", "nbap.priorityQueue_Id",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_t1,
      { "t1", "nbap.t1",
        FT_UINT32, BASE_DEC, VALS(nbap_T1_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_mAC_ehs_Reset_Timer,
      { "mAC-ehs-Reset-Timer", "nbap.mAC_ehs_Reset_Timer",
        FT_UINT32, BASE_DEC, VALS(nbap_MAC_ehs_Reset_Timer_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_discardTimer,
      { "discardTimer", "nbap.discardTimer",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_DiscardTimer_vals_ext, 0,
        NULL, HFILL }},
    { &hf_nbap_mAC_hsWindowSize,
      { "mAC-hsWindowSize", "nbap.mAC_hsWindowSize",
        FT_UINT32, BASE_DEC, VALS(nbap_MAC_hsWindowSize_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_maximum_MACcPDU_Size,
      { "maximum-MACcPDU-Size", "nbap.maximum_MACcPDU_Size",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MAC_PDU_SizeExtended", HFILL }},
    { &hf_nbap_PriorityQueue_InfoList_item,
      { "PriorityQueue-InfoItem", "nbap.PriorityQueue_InfoItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_priorityQueueId,
      { "priorityQueueId", "nbap.priorityQueueId",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PriorityQueue_Id", HFILL }},
    { &hf_nbap_associatedHSDSCH_MACdFlow,
      { "associatedHSDSCH-MACdFlow", "nbap.associatedHSDSCH_MACdFlow",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HSDSCH_MACdFlow_ID", HFILL }},
    { &hf_nbap_mAChsGuaranteedBitRate,
      { "mAChsGuaranteedBitRate", "nbap.mAChsGuaranteedBitRate",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_macdPDU_Size_Index,
      { "macdPDU-Size-Index", "nbap.macdPDU_Size_Index",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MACdPDU_Size_Indexlist", HFILL }},
    { &hf_nbap_rLC_Mode,
      { "rLC-Mode", "nbap.rLC_Mode",
        FT_UINT32, BASE_DEC, VALS(nbap_RLC_Mode_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_PriorityQueue_InfoList_to_Modify_item,
      { "ModifyPriorityQueue", "nbap.ModifyPriorityQueue",
        FT_UINT32, BASE_DEC, VALS(nbap_ModifyPriorityQueue_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_macdPDU_Size_Index_to_Modify,
      { "macdPDU-Size-Index-to-Modify", "nbap.macdPDU_Size_Index_to_Modify",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MACdPDU_Size_Indexlist_to_Modify", HFILL }},
    { &hf_nbap_PriorityQueue_InfoList_to_Modify_Unsynchronised_item,
      { "PriorityQueue-InfoItem-to-Modify-Unsynchronised", "nbap.PriorityQueue_InfoItem_to_Modify_Unsynchronised_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_none,
      { "none", "nbap.none_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_selected_MBMS_Service,
      { "selected-MBMS-Service", "nbap.selected_MBMS_Service_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_selected_MBMS_Service_List,
      { "selected-MBMS-Service-List", "nbap.selected_MBMS_Service_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Selected_MBMS_Service_List_item,
      { "Selected-MBMS-Service-Item", "nbap.Selected_MBMS_Service_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_selected_MBMS_Service_TimeSlot_Information_LCR,
      { "selected-MBMS-Service-TimeSlot-Information-LCR", "nbap.selected_MBMS_Service_TimeSlot_Information_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_mBMS_Service_TDM_Information,
      { "mBMS-Service-TDM-Information", "nbap.mBMS_Service_TDM_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Selected_MBMS_Service_TimeSlot_Information_LCR_item,
      { "TimeSlotLCR", "nbap.TimeSlotLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_transmission_Time_Interval,
      { "transmission-Time-Interval", "nbap.transmission_Time_Interval",
        FT_UINT32, BASE_DEC, VALS(nbap_T_transmission_Time_Interval_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_tDM_Rep,
      { "tDM-Rep", "nbap.tDM_Rep",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_2_9", HFILL }},
    { &hf_nbap_tDM_Offset,
      { "tDM-Offset", "nbap.tDM_Offset",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_8", HFILL }},
    { &hf_nbap_tDM_Length,
      { "tDM-Length", "nbap.tDM_Length",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_8", HFILL }},
    { &hf_nbap_Paging_MACFlow_Specific_InformationLCR_item,
      { "Paging-MAC-Flow-Specific-Information-ItemLCR", "nbap.Paging_MAC_Flow_Specific_Information_ItemLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hSDPA_associated_PICH_InfoLCR,
      { "hSDPA-associated-PICH-InfoLCR", "nbap.hSDPA_associated_PICH_InfoLCR",
        FT_UINT32, BASE_DEC, VALS(nbap_HSDPA_Associated_PICH_InformationLCR_vals), 0,
        "HSDPA_Associated_PICH_InformationLCR", HFILL }},
    { &hf_nbap_paging_MACFlow_PriorityQueue_InformationLCR,
      { "paging-MACFlow-PriorityQueue-InformationLCR", "nbap.paging_MACFlow_PriorityQueue_InformationLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Paging_MACFlow_PriorityQueue_Information", HFILL }},
    { &hf_nbap_Paging_MACFlows_to_DeleteLCR_item,
      { "Paging-MACFlows-to-DeleteLCR-Item", "nbap.Paging_MACFlows_to_DeleteLCR_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_Specific_DCH_Info_item,
      { "RL-Specific-DCH-Info-Item", "nbap.RL_Specific_DCH_Info_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dCH_id,
      { "dCH-id", "nbap.dCH_id",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_transportlayeraddress,
      { "transportlayeraddress", "nbap.transportlayeraddress",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_rL_Specific_E_DCH_Information,
      { "rL-Specific-E-DCH-Information", "nbap.rL_Specific_E_DCH_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_Specific_E_DCH_Information_item,
      { "RL-Specific-E-DCH-Information-Item", "nbap.RL_Specific_E_DCH_Information_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Reference_E_TFCI_Information_item,
      { "Reference-E-TFCI-Information-Item", "nbap.Reference_E_TFCI_Information_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_reference_E_TFCI,
      { "reference-E-TFCI", "nbap.reference_E_TFCI",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_TFCI", HFILL }},
    { &hf_nbap_reference_E_TFCI_PO,
      { "reference-E-TFCI-PO", "nbap.reference_E_TFCI_PO",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_periodic_01,
      { "periodic", "nbap.periodic",
        FT_UINT32, BASE_DEC, VALS(nbap_ReportCharacteristicsType_ReportPeriodicity_vals), 0,
        "ReportCharacteristicsType_ReportPeriodicity", HFILL }},
    { &hf_nbap_event_a,
      { "event-a", "nbap.event_a_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "ReportCharacteristicsType_EventA", HFILL }},
    { &hf_nbap_event_b,
      { "event-b", "nbap.event_b_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "ReportCharacteristicsType_EventB", HFILL }},
    { &hf_nbap_event_c,
      { "event-c", "nbap.event_c_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "ReportCharacteristicsType_EventC", HFILL }},
    { &hf_nbap_event_d,
      { "event-d", "nbap.event_d_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "ReportCharacteristicsType_EventD", HFILL }},
    { &hf_nbap_event_e,
      { "event-e", "nbap.event_e_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "ReportCharacteristicsType_EventE", HFILL }},
    { &hf_nbap_event_f,
      { "event-f", "nbap.event_f_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "ReportCharacteristicsType_EventF", HFILL }},
    { &hf_nbap_extension_ReportCharacteristics,
      { "extension-ReportCharacteristics", "nbap.extension_ReportCharacteristics_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_measurementThreshold,
      { "measurementThreshold", "nbap.measurementThreshold",
        FT_UINT32, BASE_DEC, VALS(nbap_ReportCharacteristicsType_MeasurementThreshold_vals), 0,
        "ReportCharacteristicsType_MeasurementThreshold", HFILL }},
    { &hf_nbap_measurementHysteresisTime,
      { "measurementHysteresisTime", "nbap.measurementHysteresisTime",
        FT_UINT32, BASE_DEC, VALS(nbap_ReportCharacteristicsType_ScaledMeasurementHysteresisTime_vals), 0,
        "ReportCharacteristicsType_ScaledMeasurementHysteresisTime", HFILL }},
    { &hf_nbap_measurementIncreaseThreshold,
      { "measurementIncreaseThreshold", "nbap.measurementIncreaseThreshold",
        FT_UINT32, BASE_DEC, VALS(nbap_ReportCharacteristicsType_MeasurementIncreaseDecreaseThreshold_vals), 0,
        "ReportCharacteristicsType_MeasurementIncreaseDecreaseThreshold", HFILL }},
    { &hf_nbap_measurementChangeTime,
      { "measurementChangeTime", "nbap.measurementChangeTime",
        FT_UINT32, BASE_DEC, VALS(nbap_ReportCharacteristicsType_ScaledMeasurementChangeTime_vals), 0,
        "ReportCharacteristicsType_ScaledMeasurementChangeTime", HFILL }},
    { &hf_nbap_measurementDecreaseThreshold,
      { "measurementDecreaseThreshold", "nbap.measurementDecreaseThreshold",
        FT_UINT32, BASE_DEC, VALS(nbap_ReportCharacteristicsType_MeasurementIncreaseDecreaseThreshold_vals), 0,
        "ReportCharacteristicsType_MeasurementIncreaseDecreaseThreshold", HFILL }},
    { &hf_nbap_measurementThreshold1,
      { "measurementThreshold1", "nbap.measurementThreshold1",
        FT_UINT32, BASE_DEC, VALS(nbap_ReportCharacteristicsType_MeasurementThreshold_vals), 0,
        "ReportCharacteristicsType_MeasurementThreshold", HFILL }},
    { &hf_nbap_measurementThreshold2,
      { "measurementThreshold2", "nbap.measurementThreshold2",
        FT_UINT32, BASE_DEC, VALS(nbap_ReportCharacteristicsType_MeasurementThreshold_vals), 0,
        "ReportCharacteristicsType_MeasurementThreshold", HFILL }},
    { &hf_nbap_reportPeriodicity,
      { "reportPeriodicity", "nbap.reportPeriodicity",
        FT_UINT32, BASE_DEC, VALS(nbap_ReportCharacteristicsType_ReportPeriodicity_vals), 0,
        "ReportCharacteristicsType_ReportPeriodicity", HFILL }},
    { &hf_nbap_received_total_wide_band_power_01,
      { "received-total-wide-band-power", "nbap.received_total_wide_band_power",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Received_total_wide_band_power_Value_IncrDecrThres", HFILL }},
    { &hf_nbap_uL_TimeslotISCP_01,
      { "uL-TimeslotISCP", "nbap.uL_TimeslotISCP",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UL_TimeslotISCP_Value_IncrDecrThres", HFILL }},
    { &hf_nbap_sir,
      { "sir", "nbap.sir",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SIR_Value_IncrDecrThres", HFILL }},
    { &hf_nbap_sir_error,
      { "sir-error", "nbap.sir_error",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SIR_Error_Value_IncrDecrThres", HFILL }},
    { &hf_nbap_transmitted_code_power,
      { "transmitted-code-power", "nbap.transmitted_code_power",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Transmitted_Code_Power_Value_IncrDecrThres", HFILL }},
    { &hf_nbap_rscp,
      { "rscp", "nbap.rscp",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RSCP_Value_IncrDecrThres", HFILL }},
    { &hf_nbap_round_trip_time,
      { "round-trip-time", "nbap.round_trip_time",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Round_Trip_Time_IncrDecrThres", HFILL }},
    { &hf_nbap_extension_ReportCharacteristicsType_MeasurementIncreaseDecreaseThreshold,
      { "extension-ReportCharacteristicsType-MeasurementIncreaseDecreaseThreshold", "nbap.extension_ReportCharacteristicsType_MeasurementIncreaseDecreaseThreshold_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sir_01,
      { "sir", "nbap.sir",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SIR_Value", HFILL }},
    { &hf_nbap_sir_error_01,
      { "sir-error", "nbap.sir_error",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SIR_Error_Value", HFILL }},
    { &hf_nbap_transmitted_code_power_01,
      { "transmitted-code-power", "nbap.transmitted_code_power",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Transmitted_Code_Power_Value", HFILL }},
    { &hf_nbap_rscp_01,
      { "rscp", "nbap.rscp",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RSCP_Value", HFILL }},
    { &hf_nbap_rx_timing_deviation,
      { "rx-timing-deviation", "nbap.rx_timing_deviation",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Rx_Timing_Deviation_Value", HFILL }},
    { &hf_nbap_round_trip_time_01,
      { "round-trip-time", "nbap.round_trip_time",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Round_Trip_Time_Value", HFILL }},
    { &hf_nbap_extension_ReportCharacteristicsType_MeasurementThreshold,
      { "extension-ReportCharacteristicsType-MeasurementThreshold", "nbap.extension_ReportCharacteristicsType_MeasurementThreshold_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_msec,
      { "msec", "nbap.msec",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MeasurementChangeTime_Scaledmsec", HFILL }},
    { &hf_nbap_msec_01,
      { "msec", "nbap.msec",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MeasurementHysteresisTime_Scaledmsec", HFILL }},
    { &hf_nbap_msec_02,
      { "msec", "nbap.msec",
        FT_UINT32, BASE_DEC, NULL, 0,
        "ReportPeriodicity_Scaledmsec", HFILL }},
    { &hf_nbap_Received_total_wide_band_power_For_CellPortion_Value_item,
      { "Received-total-wide-band-power-For-CellPortion-Value-Item", "nbap.Received_total_wide_band_power_For_CellPortion_Value_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_received_total_wide_band_power_value,
      { "received-total-wide-band-power-value", "nbap.received_total_wide_band_power_value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Received_total_wide_band_power_For_CellPortion_ValueLCR_item,
      { "Received-total-wide-band-power-For-CellPortion-ValueLCR-Item", "nbap.Received_total_wide_band_power_For_CellPortion_ValueLCR_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value_item,
      { "Received-Scheduled-EDCH-Power-Share-For-CellPortion-Value-Item", "nbap.Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_received_Scheduled_power_share_value,
      { "received-Scheduled-power-share-value", "nbap.received_Scheduled_power_share_value",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RSEPS_Value", HFILL }},
    { &hf_nbap_informationAvailable,
      { "informationAvailable", "nbap.informationAvailable_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_informationnotAvailable,
      { "informationnotAvailable", "nbap.informationnotAvailable_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_requesteddataValue,
      { "requesteddataValue", "nbap.requesteddataValue_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dgps_corrections,
      { "dgps-corrections", "nbap.dgps_corrections_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DGPSCorrections", HFILL }},
    { &hf_nbap_gps_navandrecovery,
      { "gps-navandrecovery", "nbap.gps_navandrecovery",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GPS_NavigationModel_and_TimeRecovery", HFILL }},
    { &hf_nbap_gps_ionos_model,
      { "gps-ionos-model", "nbap.gps_ionos_model_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GPS_Ionospheric_Model", HFILL }},
    { &hf_nbap_gps_utc_model,
      { "gps-utc-model", "nbap.gps_utc_model_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_gps_almanac,
      { "gps-almanac", "nbap.gps_almanac_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_gps_rt_integrity,
      { "gps-rt-integrity", "nbap.gps_rt_integrity",
        FT_UINT32, BASE_DEC, VALS(nbap_GPS_RealTime_Integrity_vals), 0,
        "GPS_RealTime_Integrity", HFILL }},
    { &hf_nbap_gpsrxpos,
      { "gpsrxpos", "nbap.gpsrxpos_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GPS_RX_POS", HFILL }},
    { &hf_nbap_SAT_Info_Almanac_item,
      { "SAT-Info-Almanac-Item", "nbap.SAT_Info_Almanac_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_data_id,
      { "data-id", "nbap.data_id",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sat_id,
      { "sat-id", "nbap.sat_id",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_gps_e_alm,
      { "gps-e-alm", "nbap.gps_e_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_gps_toa_alm,
      { "gps-toa-alm", "nbap.gps_toa_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_gps_delta_I_alm,
      { "gps-delta-I-alm", "nbap.gps_delta_I_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_omegadot_alm,
      { "omegadot-alm", "nbap.omegadot_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_nbap_svhealth_alm,
      { "svhealth-alm", "nbap.svhealth_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_gps_a_sqrt_alm,
      { "gps-a-sqrt-alm", "nbap.gps_a_sqrt_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_nbap_omegazero_alm,
      { "omegazero-alm", "nbap.omegazero_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_nbap_m_zero_alm,
      { "m-zero-alm", "nbap.m_zero_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_nbap_gps_omega_alm,
      { "gps-omega-alm", "nbap.gps_omega_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_nbap_gps_af_zero_alm,
      { "gps-af-zero-alm", "nbap.gps_af_zero_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_nbap_gps_af_one_alm,
      { "gps-af-one-alm", "nbap.gps_af_one_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_nbap_SAT_Info_Almanac_ExtList_item,
      { "SAT-Info-Almanac-ExtItem", "nbap.SAT_Info_Almanac_ExtItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SAT_Info_DGPSCorrections_item,
      { "SAT-Info-DGPSCorrections-Item", "nbap.SAT_Info_DGPSCorrections_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_iode_dgps,
      { "iode-dgps", "nbap.iode_dgps",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_prc,
      { "prc", "nbap.prc",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_range_correction_rate,
      { "range-correction-rate", "nbap.range_correction_rate",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SATInfo_RealTime_Integrity_item,
      { "SAT-Info-RealTime-Integrity-Item", "nbap.SAT_Info_RealTime_Integrity_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_bad_sat_id,
      { "bad-sat-id", "nbap.bad_sat_id",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SAT_ID", HFILL }},
    { &hf_nbap_SecondaryServingCells_item,
      { "SecondaryServingCellsItem", "nbap.SecondaryServingCellsItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_secondaryC_ID,
      { "secondaryC-ID", "nbap.secondaryC_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "C_ID", HFILL }},
    { &hf_nbap_numSecondaryHS_SCCH_Codes,
      { "numSecondaryHS-SCCH-Codes", "nbap.numSecondaryHS_SCCH_Codes",
        FT_UINT32, BASE_DEC, NULL, 0,
        "NumHS_SCCH_Codes", HFILL }},
    { &hf_nbap_new_secondary_CPICH,
      { "new-secondary-CPICH", "nbap.new_secondary_CPICH",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CommonPhysicalChannelID", HFILL }},
    { &hf_nbap_secondary_CPICH_shall_not_be_used,
      { "secondary-CPICH-shall-not-be-used", "nbap.secondary_CPICH_shall_not_be_used_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_serving_E_DCH_RL_in_this_NodeB,
      { "serving-E-DCH-RL-in-this-NodeB", "nbap.serving_E_DCH_RL_in_this_NodeB_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_serving_E_DCH_RL_not_in_this_NodeB,
      { "serving-E-DCH-RL-not-in-this-NodeB", "nbap.serving_E_DCH_RL_not_in_this_NodeB_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SetsOfHS_SCCH_Codes_item,
      { "SetsOfHS-SCCH-CodesItem", "nbap.SetsOfHS_SCCH_CodesItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_SCCH_PreconfiguredCodes,
      { "hS-SCCH-PreconfiguredCodes", "nbap.hS_SCCH_PreconfiguredCodes",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_setup,
      { "setup", "nbap.setup_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Additional_EDCH_Setup_Info", HFILL }},
    { &hf_nbap_configurationChange,
      { "configurationChange", "nbap.configurationChange",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Additional_EDCH_Cell_Information_ConfigurationChange_List", HFILL }},
    { &hf_nbap_removal,
      { "removal", "nbap.removal",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Additional_EDCH_Cell_Information_Removal_List", HFILL }},
    { &hf_nbap_sFNSFNChangeLimit,
      { "sFNSFNChangeLimit", "nbap.sFNSFNChangeLimit",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_predictedSFNSFNDeviationLimit,
      { "predictedSFNSFNDeviationLimit", "nbap.predictedSFNSFNDeviationLimit",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation,
      { "successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation", "nbap.successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item,
      { "successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation item", "nbap.successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sFNSFNValue,
      { "sFNSFNValue", "nbap.sFNSFNValue",
        FT_UINT32, BASE_DEC, VALS(nbap_SFNSFNValue_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_sFNSFNQuality,
      { "sFNSFNQuality", "nbap.sFNSFNQuality",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sFNSFNDriftRate,
      { "sFNSFNDriftRate", "nbap.sFNSFNDriftRate",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sFNSFNDriftRateQuality,
      { "sFNSFNDriftRateQuality", "nbap.sFNSFNDriftRateQuality",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sFNSFNTimeStampInformation,
      { "sFNSFNTimeStampInformation", "nbap.sFNSFNTimeStampInformation",
        FT_UINT32, BASE_DEC, VALS(nbap_SFNSFNTimeStampInformation_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation,
      { "unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation", "nbap.unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item,
      { "unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation item", "nbap.unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sFNSFNTimeStamp_FDD,
      { "sFNSFNTimeStamp-FDD", "nbap.sFNSFNTimeStamp_FDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SFN", HFILL }},
    { &hf_nbap_sFNSFNTimeStamp_TDD,
      { "sFNSFNTimeStamp-TDD", "nbap.sFNSFNTimeStamp_TDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sFN,
      { "sFN", "nbap.sFN",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sFNSFN_FDD,
      { "sFNSFN-FDD", "nbap.sFNSFN_FDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sFNSFN_TDD,
      { "sFNSFN-TDD", "nbap.sFNSFN_TDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sFNSFN_TDD768,
      { "sFNSFN-TDD768", "nbap.sFNSFN_TDD768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_synchronisationReportCharacteristicsType,
      { "synchronisationReportCharacteristicsType", "nbap.synchronisationReportCharacteristicsType",
        FT_UINT32, BASE_DEC, VALS(nbap_SynchronisationReportCharacteristicsType_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_synchronisationReportCharactThreExc,
      { "synchronisationReportCharactThreExc", "nbap.synchronisationReportCharactThreExc",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SynchronisationReportCharactThreExc_item,
      { "SynchronisationReportCharactThreInfoItem", "nbap.SynchronisationReportCharactThreInfoItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_syncFrameNumber,
      { "syncFrameNumber", "nbap.syncFrameNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cellSyncBurstInformation,
      { "cellSyncBurstInformation", "nbap.cellSyncBurstInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SEQUENCE_SIZE_1_maxNrOfReceptsPerSyncFrame_OF_SynchronisationReportCharactCellSyncBurstInfoItem", HFILL }},
    { &hf_nbap_cellSyncBurstInformation_item,
      { "SynchronisationReportCharactCellSyncBurstInfoItem", "nbap.SynchronisationReportCharactCellSyncBurstInfoItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cellSyncBurstCode,
      { "cellSyncBurstCode", "nbap.cellSyncBurstCode",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cellSyncBurstCodeShift,
      { "cellSyncBurstCodeShift", "nbap.cellSyncBurstCodeShift",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cellSyncBurstTiming,
      { "cellSyncBurstTiming", "nbap.cellSyncBurstTiming",
        FT_UINT32, BASE_DEC, VALS(nbap_CellSyncBurstTiming_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_cellSyncBurstTimingThreshold,
      { "cellSyncBurstTimingThreshold", "nbap.cellSyncBurstTimingThreshold",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SyncDLCodeIdThreInfoLCR_item,
      { "SyncDLCodeIdThreInfoList", "nbap.SyncDLCodeIdThreInfoList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_syncFrameNoToReceive,
      { "syncFrameNoToReceive", "nbap.syncFrameNoToReceive",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SyncFrameNumber", HFILL }},
    { &hf_nbap_syncDLCodeIdInfoLCR,
      { "syncDLCodeIdInfoLCR", "nbap.syncDLCodeIdInfoLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SyncDLCodeInfoListLCR", HFILL }},
    { &hf_nbap_SyncDLCodeInfoListLCR_item,
      { "SyncDLCodeInfoItemLCR", "nbap.SyncDLCodeInfoItemLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_syncDLCodeId,
      { "syncDLCodeId", "nbap.syncDLCodeId",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_syncDLCodeIdArrivTime,
      { "syncDLCodeIdArrivTime", "nbap.syncDLCodeIdArrivTime",
        FT_UINT32, BASE_DEC, VALS(nbap_CellSyncBurstTimingLCR_vals), 0,
        "CellSyncBurstTimingLCR", HFILL }},
    { &hf_nbap_syncDLCodeIdTimingThre,
      { "syncDLCodeIdTimingThre", "nbap.syncDLCodeIdTimingThre",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CellSyncBurstTimingThreshold", HFILL }},
    { &hf_nbap_modulation,
      { "modulation", "nbap.modulation",
        FT_UINT32, BASE_DEC, VALS(nbap_Modulation_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TDD_DL_Code_Information_item,
      { "TDD-DL-Code-InformationItem", "nbap.TDD_DL_Code_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dPCH_ID,
      { "dPCH-ID", "nbap.dPCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TDD_DL_Code_LCR_Information_item,
      { "TDD-DL-Code-LCR-InformationItem", "nbap.TDD_DL_Code_LCR_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tdd_DL_DPCH_TimeSlotFormat_LCR,
      { "tdd-DL-DPCH-TimeSlotFormat-LCR", "nbap.tdd_DL_DPCH_TimeSlotFormat_LCR",
        FT_UINT32, BASE_DEC, VALS(nbap_TDD_DL_DPCH_TimeSlotFormat_LCR_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TDD_DL_Code_768_Information_item,
      { "TDD-DL-Code-768-InformationItem", "nbap.TDD_DL_Code_768_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_qPSK_01,
      { "qPSK", "nbap.qPSK",
        FT_UINT32, BASE_DEC, NULL, 0,
        "QPSK_DL_DPCH_TimeSlotFormatTDD_LCR", HFILL }},
    { &hf_nbap_eightPSK,
      { "eightPSK", "nbap.eightPSK",
        FT_UINT32, BASE_DEC, NULL, 0,
        "EightPSK_DL_DPCH_TimeSlotFormatTDD_LCR", HFILL }},
    { &hf_nbap_initialOffset,
      { "initialOffset", "nbap.initialOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_255", HFILL }},
    { &hf_nbap_noinitialOffset,
      { "noinitialOffset", "nbap.noinitialOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_63", HFILL }},
    { &hf_nbap_signalledGainFactors,
      { "signalledGainFactors", "nbap.signalledGainFactors_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_gainFactor,
      { "gainFactor", "nbap.gainFactor",
        FT_UINT32, BASE_DEC, VALS(nbap_T_gainFactor_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_fdd,
      { "fdd", "nbap.fdd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_betaC,
      { "betaC", "nbap.betaC",
        FT_UINT32, BASE_DEC, NULL, 0,
        "BetaCD", HFILL }},
    { &hf_nbap_betaD,
      { "betaD", "nbap.betaD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "BetaCD", HFILL }},
    { &hf_nbap_tdd,
      { "tdd", "nbap.tdd",
        FT_UINT32, BASE_DEC, NULL, 0,
        "BetaCD", HFILL }},
    { &hf_nbap_refTFCNumber,
      { "refTFCNumber", "nbap.refTFCNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_computedGainFactors,
      { "computedGainFactors", "nbap.computedGainFactors",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RefTFCNumber", HFILL }},
    { &hf_nbap_TDD_UL_Code_Information_item,
      { "TDD-UL-Code-InformationItem", "nbap.TDD_UL_Code_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TDD_UL_Code_LCR_Information_item,
      { "TDD-UL-Code-LCR-InformationItem", "nbap.TDD_UL_Code_LCR_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tdd_UL_DPCH_TimeSlotFormat_LCR,
      { "tdd-UL-DPCH-TimeSlotFormat-LCR", "nbap.tdd_UL_DPCH_TimeSlotFormat_LCR",
        FT_UINT32, BASE_DEC, VALS(nbap_TDD_UL_DPCH_TimeSlotFormat_LCR_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TDD_UL_Code_768_Information_item,
      { "TDD-UL-Code-768-InformationItem", "nbap.TDD_UL_Code_768_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_qPSK_02,
      { "qPSK", "nbap.qPSK",
        FT_UINT32, BASE_DEC, NULL, 0,
        "QPSK_UL_DPCH_TimeSlotFormatTDD_LCR", HFILL }},
    { &hf_nbap_eightPSK_01,
      { "eightPSK", "nbap.eightPSK",
        FT_UINT32, BASE_DEC, NULL, 0,
        "EightPSK_UL_DPCH_TimeSlotFormatTDD_LCR", HFILL }},
    { &hf_nbap_tFCI_SignallingOption,
      { "tFCI-SignallingOption", "nbap.tFCI_SignallingOption",
        FT_UINT32, BASE_DEC, VALS(nbap_TFCI_SignallingMode_TFCI_SignallingOption_vals), 0,
        "TFCI_SignallingMode_TFCI_SignallingOption", HFILL }},
    { &hf_nbap_not_Used_splitType,
      { "not-Used-splitType", "nbap.not_Used_splitType_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_not_Used_lengthOfTFCI2,
      { "not-Used-lengthOfTFCI2", "nbap.not_Used_lengthOfTFCI2_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TimeSlot_InitiatedListLCR_item,
      { "TimeSlotLCR", "nbap.TimeSlotLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TimeSlotMeasurementValueListLCR_item,
      { "TimeSlotMeasurementValueLCR", "nbap.TimeSlotMeasurementValueLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_commonMeasurementValue,
      { "commonMeasurementValue", "nbap.commonMeasurementValue",
        FT_UINT32, BASE_DEC, VALS(nbap_CommonMeasurementValue_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_initialPhase_01,
      { "initialPhase", "nbap.initialPhase",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_1048575_", HFILL }},
    { &hf_nbap_dsField,
      { "dsField", "nbap.dsField",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_genericTrafficCategory,
      { "genericTrafficCategory", "nbap.genericTrafficCategory",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Transmission_Gap_Pattern_Sequence_Information_item,
      { "Transmission-Gap-Pattern-Sequence-Information item", "nbap.Transmission_Gap_Pattern_Sequence_Information_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tGSN,
      { "tGSN", "nbap.tGSN",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tGL1,
      { "tGL1", "nbap.tGL1",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GapLength", HFILL }},
    { &hf_nbap_tGL2,
      { "tGL2", "nbap.tGL2",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GapLength", HFILL }},
    { &hf_nbap_tGD,
      { "tGD", "nbap.tGD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tGPL1,
      { "tGPL1", "nbap.tGPL1",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GapDuration", HFILL }},
    { &hf_nbap_not_to_be_used_1,
      { "not-to-be-used-1", "nbap.not_to_be_used_1",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GapDuration", HFILL }},
    { &hf_nbap_uL_DL_mode,
      { "uL-DL-mode", "nbap.uL_DL_mode",
        FT_UINT32, BASE_DEC, VALS(nbap_UL_DL_mode_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_downlink_Compressed_Mode_Method,
      { "downlink-Compressed-Mode-Method", "nbap.downlink_Compressed_Mode_Method",
        FT_UINT32, BASE_DEC, VALS(nbap_Downlink_Compressed_Mode_Method_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_uplink_Compressed_Mode_Method,
      { "uplink-Compressed-Mode-Method", "nbap.uplink_Compressed_Mode_Method",
        FT_UINT32, BASE_DEC, VALS(nbap_Uplink_Compressed_Mode_Method_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_dL_FrameType,
      { "dL-FrameType", "nbap.dL_FrameType",
        FT_UINT32, BASE_DEC, VALS(nbap_DL_FrameType_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_delta_SIR1,
      { "delta-SIR1", "nbap.delta_SIR1",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DeltaSIR", HFILL }},
    { &hf_nbap_delta_SIR_after1,
      { "delta-SIR-after1", "nbap.delta_SIR_after1",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DeltaSIR", HFILL }},
    { &hf_nbap_delta_SIR2,
      { "delta-SIR2", "nbap.delta_SIR2",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DeltaSIR", HFILL }},
    { &hf_nbap_delta_SIR_after2,
      { "delta-SIR-after2", "nbap.delta_SIR_after2",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DeltaSIR", HFILL }},
    { &hf_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue_item,
      { "TransmittedCarrierPowerOfAllCodesNotUsedForHS-PDSCH-HS-SCCH-E-AGCH-E-RGCHOrE-HICHTransmissionCellPortionValue-Item", "nbap.TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_transmittedCarrierPowerOfAllCodesNotUsedForHSTransmissionValue,
      { "transmittedCarrierPowerOfAllCodesNotUsedForHSTransmissionValue", "nbap.transmittedCarrierPowerOfAllCodesNotUsedForHSTransmissionValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue_item,
      { "TransmittedCarrierPowerOfAllCodesNotUsedForHS-PDSCH-HS-SCCH-E-AGCHOrE-HICHTransmissionCellPortionValue-Item", "nbap.TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Transmitted_Carrier_Power_For_CellPortion_Value_item,
      { "Transmitted-Carrier-Power-For-CellPortion-Value-Item", "nbap.Transmitted_Carrier_Power_For_CellPortion_Value_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_transmitted_Carrier_Power_Value,
      { "transmitted-Carrier-Power-Value", "nbap.transmitted_Carrier_Power_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Transmitted_Carrier_Power_For_CellPortion_ValueLCR_item,
      { "Transmitted-Carrier-Power-For-CellPortion-ValueLCR-Item", "nbap.Transmitted_Carrier_Power_For_CellPortion_ValueLCR_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tFCSvalues,
      { "tFCSvalues", "nbap.tFCSvalues",
        FT_UINT32, BASE_DEC, VALS(nbap_T_tFCSvalues_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_no_Split_in_TFCI,
      { "no-Split-in-TFCI", "nbap.no_Split_in_TFCI",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TFCS_TFCSList", HFILL }},
    { &hf_nbap_not_Used_split_in_TFCI,
      { "not-Used-split-in-TFCI", "nbap.not_Used_split_in_TFCI_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TFCS_TFCSList_item,
      { "TFCS-TFCSList item", "nbap.TFCS_TFCSList_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cTFC,
      { "cTFC", "nbap.cTFC",
        FT_UINT32, BASE_DEC, VALS(nbap_TFCS_CTFC_vals), 0,
        "TFCS_CTFC", HFILL }},
    { &hf_nbap_tFC_Beta,
      { "tFC-Beta", "nbap.tFC_Beta",
        FT_UINT32, BASE_DEC, VALS(nbap_TransportFormatCombination_Beta_vals), 0,
        "TransportFormatCombination_Beta", HFILL }},
    { &hf_nbap_ctfc2bit,
      { "ctfc2bit", "nbap.ctfc2bit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_3", HFILL }},
    { &hf_nbap_ctfc4bit,
      { "ctfc4bit", "nbap.ctfc4bit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_15", HFILL }},
    { &hf_nbap_ctfc6bit,
      { "ctfc6bit", "nbap.ctfc6bit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_63", HFILL }},
    { &hf_nbap_ctfc8bit,
      { "ctfc8bit", "nbap.ctfc8bit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_255", HFILL }},
    { &hf_nbap_ctfc12bit,
      { "ctfc12bit", "nbap.ctfc12bit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_4095", HFILL }},
    { &hf_nbap_ctfc16bit,
      { "ctfc16bit", "nbap.ctfc16bit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_65535", HFILL }},
    { &hf_nbap_ctfcmaxbit,
      { "ctfcmaxbit", "nbap.ctfcmaxbit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_maxCTFC", HFILL }},
    { &hf_nbap_Transport_Block_Size_List_item,
      { "Transport-Block-Size-List item", "nbap.Transport_Block_Size_List_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_transport_Block_Size_Index_for_Enhanced_PCH,
      { "transport-Block-Size-Index-for-Enhanced-PCH", "nbap.transport_Block_Size_Index_for_Enhanced_PCH",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dynamicParts,
      { "dynamicParts", "nbap.dynamicParts",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TransportFormatSet_DynamicPartList", HFILL }},
    { &hf_nbap_semi_staticPart,
      { "semi-staticPart", "nbap.semi_staticPart_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "TransportFormatSet_Semi_staticPart", HFILL }},
    { &hf_nbap_TransportFormatSet_DynamicPartList_item,
      { "TransportFormatSet-DynamicPartList item", "nbap.TransportFormatSet_DynamicPartList_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_nrOfTransportBlocks,
      { "nrOfTransportBlocks", "nbap.nrOfTransportBlocks",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TransportFormatSet_NrOfTransportBlocks", HFILL }},
    { &hf_nbap_transportBlockSize,
      { "transportBlockSize", "nbap.transportBlockSize",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TransportFormatSet_TransportBlockSize", HFILL }},
    { &hf_nbap_mode,
      { "mode", "nbap.mode",
        FT_UINT32, BASE_DEC, VALS(nbap_TransportFormatSet_ModeDP_vals), 0,
        "TransportFormatSet_ModeDP", HFILL }},
    { &hf_nbap_transmissionTimeIntervalInformation,
      { "transmissionTimeIntervalInformation", "nbap.transmissionTimeIntervalInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TransmissionTimeIntervalInformation_item,
      { "TransmissionTimeIntervalInformation item", "nbap.TransmissionTimeIntervalInformation_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_transmissionTimeInterval,
      { "transmissionTimeInterval", "nbap.transmissionTimeInterval",
        FT_UINT32, BASE_DEC, VALS(nbap_TransportFormatSet_TransmissionTimeIntervalDynamic_vals), 0,
        "TransportFormatSet_TransmissionTimeIntervalDynamic", HFILL }},
    { &hf_nbap_transmissionTimeInterval_01,
      { "transmissionTimeInterval", "nbap.transmissionTimeInterval",
        FT_UINT32, BASE_DEC, VALS(nbap_TransportFormatSet_TransmissionTimeIntervalSemiStatic_vals), 0,
        "TransportFormatSet_TransmissionTimeIntervalSemiStatic", HFILL }},
    { &hf_nbap_channelCoding,
      { "channelCoding", "nbap.channelCoding",
        FT_UINT32, BASE_DEC, VALS(nbap_TransportFormatSet_ChannelCodingType_vals), 0,
        "TransportFormatSet_ChannelCodingType", HFILL }},
    { &hf_nbap_codingRate,
      { "codingRate", "nbap.codingRate",
        FT_UINT32, BASE_DEC, VALS(nbap_TransportFormatSet_CodingRate_vals), 0,
        "TransportFormatSet_CodingRate", HFILL }},
    { &hf_nbap_rateMatchingAttribute,
      { "rateMatchingAttribute", "nbap.rateMatchingAttribute",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TransportFormatSet_RateMatchingAttribute", HFILL }},
    { &hf_nbap_cRC_Size,
      { "cRC-Size", "nbap.cRC_Size",
        FT_UINT32, BASE_DEC, VALS(nbap_TransportFormatSet_CRC_Size_vals), 0,
        "TransportFormatSet_CRC_Size", HFILL }},
    { &hf_nbap_mode_01,
      { "mode", "nbap.mode",
        FT_UINT32, BASE_DEC, VALS(nbap_TransportFormatSet_ModeSSP_vals), 0,
        "TransportFormatSet_ModeSSP", HFILL }},
    { &hf_nbap_tdd_01,
      { "tdd", "nbap.tdd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "TDD_TransportFormatSet_ModeDP", HFILL }},
    { &hf_nbap_notApplicable,
      { "notApplicable", "nbap.notApplicable_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tdd_02,
      { "tdd", "nbap.tdd",
        FT_UINT32, BASE_DEC, VALS(nbap_TransportFormatSet_SecondInterleavingMode_vals), 0,
        "TransportFormatSet_SecondInterleavingMode", HFILL }},
    { &hf_nbap_mS,
      { "mS", "nbap.mS",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_16383", HFILL }},
    { &hf_nbap_lS,
      { "lS", "nbap.lS",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_4294967295", HFILL }},
    { &hf_nbap_tUTRANGANSSChangeLimit,
      { "tUTRANGANSSChangeLimit", "nbap.tUTRANGANSSChangeLimit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_256", HFILL }},
    { &hf_nbap_predictedTUTRANGANSSDeviationLimit,
      { "predictedTUTRANGANSSDeviationLimit", "nbap.predictedTUTRANGANSSDeviationLimit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_256", HFILL }},
    { &hf_nbap_tUTRANGANSS,
      { "tUTRANGANSS", "nbap.tUTRANGANSS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tUTRANGANSSQuality,
      { "tUTRANGANSSQuality", "nbap.tUTRANGANSSQuality",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_255", HFILL }},
    { &hf_nbap_tUTRANGANSSDriftRate,
      { "tUTRANGANSSDriftRate", "nbap.tUTRANGANSSDriftRate",
        FT_INT32, BASE_DEC, NULL, 0,
        "INTEGER_M50_50", HFILL }},
    { &hf_nbap_tUTRANGANSSDriftRateQuality,
      { "tUTRANGANSSDriftRateQuality", "nbap.tUTRANGANSSDriftRateQuality",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_50", HFILL }},
    { &hf_nbap_ms_part,
      { "ms-part", "nbap.ms_part",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_16383", HFILL }},
    { &hf_nbap_ls_part,
      { "ls-part", "nbap.ls_part",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_4294967295", HFILL }},
    { &hf_nbap_tUTRANGPSChangeLimit,
      { "tUTRANGPSChangeLimit", "nbap.tUTRANGPSChangeLimit",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_predictedTUTRANGPSDeviationLimit,
      { "predictedTUTRANGPSDeviationLimit", "nbap.predictedTUTRANGPSDeviationLimit",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tUTRANGPS,
      { "tUTRANGPS", "nbap.tUTRANGPS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tUTRANGPSQuality,
      { "tUTRANGPSQuality", "nbap.tUTRANGPSQuality",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tUTRANGPSDriftRate,
      { "tUTRANGPSDriftRate", "nbap.tUTRANGPSDriftRate",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tUTRANGPSDriftRateQuality,
      { "tUTRANGPSDriftRateQuality", "nbap.tUTRANGPSDriftRateQuality",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_rNC_ID,
      { "rNC-ID", "nbap.rNC_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uE_AggregateMaximumBitRateDownlink,
      { "uE-AggregateMaximumBitRateDownlink", "nbap.uE_AggregateMaximumBitRateDownlink",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uE_AggregateMaximumBitRateUplink,
      { "uE-AggregateMaximumBitRateUplink", "nbap.uE_AggregateMaximumBitRateUplink",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hSDSCH_Physical_Layer_Category_01,
      { "hSDSCH-Physical-Layer-Category", "nbap.hSDSCH_Physical_Layer_Category",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_Timeslot_Information_item,
      { "UL-Timeslot-InformationItem", "nbap.UL_Timeslot_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_Code_InformationList,
      { "uL-Code-InformationList", "nbap.uL_Code_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TDD_UL_Code_Information", HFILL }},
    { &hf_nbap_UL_TimeslotLCR_Information_item,
      { "UL-TimeslotLCR-InformationItem", "nbap.UL_TimeslotLCR_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_Code_InformationList_01,
      { "uL-Code-InformationList", "nbap.uL_Code_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TDD_UL_Code_LCR_Information", HFILL }},
    { &hf_nbap_UL_Timeslot768_Information_item,
      { "UL-Timeslot768-InformationItem", "nbap.UL_Timeslot768_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_Code_InformationList_02,
      { "uL-Code-InformationList", "nbap.uL_Code_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TDD_UL_Code_768_Information", HFILL }},
    { &hf_nbap_uL_ScramblingCodeNumber,
      { "uL-ScramblingCodeNumber", "nbap.uL_ScramblingCodeNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_ScramblingCodeLength,
      { "uL-ScramblingCodeLength", "nbap.uL_ScramblingCodeLength",
        FT_UINT32, BASE_DEC, VALS(nbap_UL_ScramblingCodeLength_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_uL_Synchronisation_StepSize,
      { "uL-Synchronisation-StepSize", "nbap.uL_Synchronisation_StepSize",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_Synchronisation_Frequency,
      { "uL-Synchronisation-Frequency", "nbap.uL_Synchronisation_Frequency",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_TimeSlot_ISCP_Info_item,
      { "UL-TimeSlot-ISCP-InfoItem", "nbap.UL_TimeSlot_ISCP_InfoItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_iSCP,
      { "iSCP", "nbap.iSCP",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UL_TimeslotISCP_Value", HFILL }},
    { &hf_nbap_UL_TimeSlot_ISCP_LCR_Info_item,
      { "UL-TimeSlot-ISCP-LCR-InfoItem", "nbap.UL_TimeSlot_ISCP_LCR_InfoItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UpPTSInterference_For_CellPortion_Value_item,
      { "UpPTSInterference-For-CellPortion-Value-Item", "nbap.UpPTSInterference_For_CellPortion_Value_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_upPTSInterferenceValue,
      { "upPTSInterferenceValue", "nbap.upPTSInterferenceValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_USCH_Information_item,
      { "USCH-InformationItem", "nbap.USCH_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uSCH_ID,
      { "uSCH-ID", "nbap.uSCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_USCH_InformationResponse_item,
      { "USCH-InformationResponseItem", "nbap.USCH_InformationResponseItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_TimeslotISCP_For_CellPortion_Value_item,
      { "UL-TimeslotISCP-For-CellPortion-Value-Item", "nbap.UL_TimeslotISCP_For_CellPortion_Value_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_TimeslotISCP_Value,
      { "uL-TimeslotISCP-Value", "nbap.uL_TimeslotISCP_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_protocolIEs,
      { "protocolIEs", "nbap.protocolIEs",
        FT_UINT32, BASE_DEC, NULL, 0,
        "ProtocolIE_Container", HFILL }},
    { &hf_nbap_protocolExtensions,
      { "protocolExtensions", "nbap.protocolExtensions",
        FT_UINT32, BASE_DEC, NULL, 0,
        "ProtocolExtensionContainer", HFILL }},
    { &hf_nbap_secondary_CCPCH_parameters,
      { "secondary-CCPCH-parameters", "nbap.secondary_CCPCH_parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Secondary_CCPCH_CTCH_SetupRqstFDD", HFILL }},
    { &hf_nbap_pRACH_parameters,
      { "pRACH-parameters", "nbap.pRACH_parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PRACH_CTCH_SetupRqstFDD", HFILL }},
    { &hf_nbap_notUsed_pCPCHes_parameters,
      { "notUsed-pCPCHes-parameters", "nbap.notUsed_pCPCHes_parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_fdd_S_CCPCH_Offset,
      { "fdd-S-CCPCH-Offset", "nbap.fdd_S_CCPCH_Offset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tFCS,
      { "tFCS", "nbap.tFCS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_secondary_CCPCH_SlotFormat,
      { "secondary-CCPCH-SlotFormat", "nbap.secondary_CCPCH_SlotFormat",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SecondaryCCPCH_SlotFormat", HFILL }},
    { &hf_nbap_multiplexingPosition,
      { "multiplexingPosition", "nbap.multiplexingPosition",
        FT_UINT32, BASE_DEC, VALS(nbap_MultiplexingPosition_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_powerOffsetInformation,
      { "powerOffsetInformation", "nbap.powerOffsetInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PowerOffsetInformation_CTCH_SetupRqstFDD", HFILL }},
    { &hf_nbap_fACH_Parameters,
      { "fACH-Parameters", "nbap.fACH_Parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "FACH_ParametersList_CTCH_SetupRqstFDD", HFILL }},
    { &hf_nbap_pCH_Parameters,
      { "pCH-Parameters", "nbap.pCH_Parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PCH_Parameters_CTCH_SetupRqstFDD", HFILL }},
    { &hf_nbap_pO1_ForTFCI_Bits,
      { "pO1-ForTFCI-Bits", "nbap.pO1_ForTFCI_Bits",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PowerOffset", HFILL }},
    { &hf_nbap_pO3_ForPilotBits,
      { "pO3-ForPilotBits", "nbap.pO3_ForPilotBits",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PowerOffset", HFILL }},
    { &hf_nbap_FACH_ParametersListIE_CTCH_SetupRqstFDD_item,
      { "FACH-ParametersItem-CTCH-SetupRqstFDD", "nbap.FACH_ParametersItem_CTCH_SetupRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_maxFACH_Power,
      { "maxFACH-Power", "nbap.maxFACH_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_transportFormatSet_01,
      { "transportFormatSet", "nbap.transportFormatSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_pCH_Power,
      { "pCH-Power", "nbap.pCH_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_pICH_Parameters,
      { "pICH-Parameters", "nbap.pICH_Parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PICH_Parameters_CTCH_SetupRqstFDD", HFILL }},
    { &hf_nbap_pICH_Mode,
      { "pICH-Mode", "nbap.pICH_Mode",
        FT_UINT32, BASE_DEC, VALS(nbap_PICH_Mode_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_mICH_Power,
      { "mICH-Power", "nbap.mICH_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        "PICH_Power", HFILL }},
    { &hf_nbap_mICH_Mode,
      { "mICH-Mode", "nbap.mICH_Mode",
        FT_UINT32, BASE_DEC, VALS(nbap_MICH_Mode_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_preambleSignatures,
      { "preambleSignatures", "nbap.preambleSignatures",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_allowedSlotFormatInformation,
      { "allowedSlotFormatInformation", "nbap.allowedSlotFormatInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        "AllowedSlotFormatInformationList_CTCH_SetupRqstFDD", HFILL }},
    { &hf_nbap_rACH_SubChannelNumbers,
      { "rACH-SubChannelNumbers", "nbap.rACH_SubChannelNumbers",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ul_punctureLimit,
      { "ul-punctureLimit", "nbap.ul_punctureLimit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PunctureLimit", HFILL }},
    { &hf_nbap_rACH_Parameters,
      { "rACH-Parameters", "nbap.rACH_Parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RACH_Parameters_CTCH_SetupRqstFDD", HFILL }},
    { &hf_nbap_aICH_Parameters,
      { "aICH-Parameters", "nbap.aICH_Parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "AICH_Parameters_CTCH_SetupRqstFDD", HFILL }},
    { &hf_nbap_AllowedSlotFormatInformationList_CTCH_SetupRqstFDD_item,
      { "AllowedSlotFormatInformationItem-CTCH-SetupRqstFDD", "nbap.AllowedSlotFormatInformationItem_CTCH_SetupRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_rACHSlotFormat,
      { "rACHSlotFormat", "nbap.rACHSlotFormat",
        FT_UINT32, BASE_DEC, VALS(nbap_RACH_SlotFormat_vals), 0,
        "RACH_SlotFormat", HFILL }},
    { &hf_nbap_secondary_CCPCH_parameters_01,
      { "secondary-CCPCH-parameters", "nbap.secondary_CCPCH_parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Secondary_CCPCH_CTCH_SetupRqstTDD", HFILL }},
    { &hf_nbap_pRACH_parameters_01,
      { "pRACH-parameters", "nbap.pRACH_parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PRACH_CTCH_SetupRqstTDD", HFILL }},
    { &hf_nbap_extension_CommonPhysicalChannelType_CTCH_SetupRqstTDD,
      { "extension-CommonPhysicalChannelType-CTCH-SetupRqstTDD", "nbap.extension_CommonPhysicalChannelType_CTCH_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sCCPCH_CCTrCH_ID,
      { "sCCPCH-CCTrCH-ID", "nbap.sCCPCH_CCTrCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CCTrCH_ID", HFILL }},
    { &hf_nbap_tFCI_Coding,
      { "tFCI-Coding", "nbap.tFCI_Coding",
        FT_UINT32, BASE_DEC, VALS(nbap_TFCI_Coding_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_punctureLimit,
      { "punctureLimit", "nbap.punctureLimit",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_secondaryCCPCH_parameterList,
      { "secondaryCCPCH-parameterList", "nbap.secondaryCCPCH_parameterList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Secondary_CCPCH_parameterList_CTCH_SetupRqstTDD", HFILL }},
    { &hf_nbap_fACH_ParametersList,
      { "fACH-ParametersList", "nbap.fACH_ParametersList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "FACH_ParametersList_CTCH_SetupRqstTDD", HFILL }},
    { &hf_nbap_pCH_Parameters_01,
      { "pCH-Parameters", "nbap.pCH_Parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PCH_Parameters_CTCH_SetupRqstTDD", HFILL }},
    { &hf_nbap_Secondary_CCPCH_parameterListIE_CTCH_SetupRqstTDD_item,
      { "Secondary-CCPCH-parameterItem-CTCH-SetupRqstTDD", "nbap.Secondary_CCPCH_parameterItem_CTCH_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_midambleShiftandBurstType,
      { "midambleShiftandBurstType", "nbap.midambleShiftandBurstType",
        FT_UINT32, BASE_DEC, VALS(nbap_MidambleShiftAndBurstType_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_s_CCPCH_Power,
      { "s-CCPCH-Power", "nbap.s_CCPCH_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_Secondary_CCPCH_LCR_parameterList_CTCH_SetupRqstTDD_item,
      { "Secondary-CCPCH-LCR-parameterItem-CTCH-SetupRqstTDD", "nbap.Secondary_CCPCH_LCR_parameterItem_CTCH_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_s_CCPCH_TimeSlotFormat_LCR,
      { "s-CCPCH-TimeSlotFormat-LCR", "nbap.s_CCPCH_TimeSlotFormat_LCR",
        FT_UINT32, BASE_DEC, VALS(nbap_TDD_DL_DPCH_TimeSlotFormat_LCR_vals), 0,
        "TDD_DL_DPCH_TimeSlotFormat_LCR", HFILL }},
    { &hf_nbap_Secondary_CCPCH_768_parameterList_CTCH_SetupRqstTDD_item,
      { "Secondary-CCPCH-768-parameterItem-CTCH-SetupRqstTDD", "nbap.Secondary_CCPCH_768_parameterItem_CTCH_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tFCI_Presence768,
      { "tFCI-Presence768", "nbap.tFCI_Presence768",
        FT_UINT32, BASE_DEC, VALS(nbap_TFCI_Presence_vals), 0,
        "TFCI_Presence", HFILL }},
    { &hf_nbap_midambleShiftandBurstType768,
      { "midambleShiftandBurstType768", "nbap.midambleShiftandBurstType768",
        FT_UINT32, BASE_DEC, VALS(nbap_MidambleShiftAndBurstType768_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_FACH_ParametersListIE_CTCH_SetupRqstTDD_item,
      { "FACH-ParametersItem-CTCH-SetupRqstTDD", "nbap.FACH_ParametersItem_CTCH_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_fACH_CCTrCH_ID,
      { "fACH-CCTrCH-ID", "nbap.fACH_CCTrCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CCTrCH_ID", HFILL }},
    { &hf_nbap_pCH_CCTrCH_ID,
      { "pCH-CCTrCH-ID", "nbap.pCH_CCTrCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CCTrCH_ID", HFILL }},
    { &hf_nbap_pICH_Parameters_01,
      { "pICH-Parameters", "nbap.pICH_Parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PICH_Parameters_CTCH_SetupRqstTDD", HFILL }},
    { &hf_nbap_midambleshiftAndBurstType,
      { "midambleshiftAndBurstType", "nbap.midambleshiftAndBurstType",
        FT_UINT32, BASE_DEC, VALS(nbap_MidambleShiftAndBurstType_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_midambleshiftAndBurstType78,
      { "midambleshiftAndBurstType78", "nbap.midambleshiftAndBurstType78",
        FT_UINT32, BASE_DEC, VALS(nbap_MidambleShiftAndBurstType768_vals), 0,
        "MidambleShiftAndBurstType768", HFILL }},
    { &hf_nbap_notificationIndicatorLength,
      { "notificationIndicatorLength", "nbap.notificationIndicatorLength",
        FT_UINT32, BASE_DEC, VALS(nbap_NotificationIndicatorLength_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_mICH_TDDOption_Specific_Parameters,
      { "mICH-TDDOption-Specific-Parameters", "nbap.mICH_TDDOption_Specific_Parameters",
        FT_UINT32, BASE_DEC, VALS(nbap_MICH_TDDOption_Specific_Parameters_CTCH_SetupRqstTDD_vals), 0,
        "MICH_TDDOption_Specific_Parameters_CTCH_SetupRqstTDD", HFILL }},
    { &hf_nbap_hCR_TDD,
      { "hCR-TDD", "nbap.hCR_TDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "MICH_HCR_Parameters_CTCH_SetupRqstTDD", HFILL }},
    { &hf_nbap_lCR_TDD,
      { "lCR-TDD", "nbap.lCR_TDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "MICH_LCR_Parameters_CTCH_SetupRqstTDD", HFILL }},
    { &hf_nbap_cHipRate768_TDD,
      { "cHipRate768-TDD", "nbap.cHipRate768_TDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "MICH_768_Parameters_CTCH_SetupRqstTDD", HFILL }},
    { &hf_nbap_tSTD_Indicator,
      { "tSTD-Indicator", "nbap.tSTD_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_TSTD_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_midambleshiftAndBurstType768,
      { "midambleshiftAndBurstType768", "nbap.midambleshiftAndBurstType768",
        FT_UINT32, BASE_DEC, VALS(nbap_MidambleShiftAndBurstType768_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TimeSlotConfigurationList_LCR_CTCH_SetupRqstTDD_item,
      { "TimeSlotConfigurationItem-LCR-CTCH-SetupRqstTDD", "nbap.TimeSlotConfigurationItem_LCR_CTCH_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_timeslotLCR_Parameter_ID,
      { "timeslotLCR-Parameter-ID", "nbap.timeslotLCR_Parameter_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CellParameterID", HFILL }},
    { &hf_nbap_Secondary_CCPCH_parameterExtendedList_CTCH_SetupRqstTDD_item,
      { "Secondary-CCPCH-parameterItem-CTCH-SetupRqstTDD", "nbap.Secondary_CCPCH_parameterItem_CTCH_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_SetupRqstTDD_item,
      { "Secondary-CCPCH-LCR-parameterItem-CTCH-SetupRqstTDD", "nbap.Secondary_CCPCH_LCR_parameterItem_CTCH_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_pRACH_Parameters_CTCH_SetupRqstTDD,
      { "pRACH-Parameters-CTCH-SetupRqstTDD", "nbap.pRACH_Parameters_CTCH_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_maxPRACH_MidambleShifts,
      { "maxPRACH-MidambleShifts", "nbap.maxPRACH_MidambleShifts",
        FT_UINT32, BASE_DEC, VALS(nbap_MaxPRACH_MidambleShifts_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_pRACH_Midamble,
      { "pRACH-Midamble", "nbap.pRACH_Midamble",
        FT_UINT32, BASE_DEC, VALS(nbap_PRACH_Midamble_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_rACH,
      { "rACH", "nbap.rACH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RACH_Parameter_CTCH_SetupRqstTDD", HFILL }},
    { &hf_nbap_uL_TransportFormatSet,
      { "uL-TransportFormatSet", "nbap.uL_TransportFormatSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "TransportFormatSet", HFILL }},
    { &hf_nbap_PRACH_LCR_ParametersList_CTCH_SetupRqstTDD_item,
      { "PRACH-LCR-ParametersItem-CTCH-SetupRqstTDD", "nbap.PRACH_LCR_ParametersItem_CTCH_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_fPACH_Power,
      { "fPACH-Power", "nbap.fPACH_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_maxPowerPLCCH,
      { "maxPowerPLCCH", "nbap.maxPowerPLCCH",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_maxE_RUCCH_MidambleShifts,
      { "maxE-RUCCH-MidambleShifts", "nbap.maxE_RUCCH_MidambleShifts",
        FT_UINT32, BASE_DEC, VALS(nbap_MaxPRACH_MidambleShifts_vals), 0,
        "MaxPRACH_MidambleShifts", HFILL }},
    { &hf_nbap_e_RUCCH_Midamble,
      { "e-RUCCH-Midamble", "nbap.e_RUCCH_Midamble",
        FT_UINT32, BASE_DEC, VALS(nbap_PRACH_Midamble_vals), 0,
        "PRACH_Midamble", HFILL }},
    { &hf_nbap_FACH_CommonTransportChannel_InformationResponse_item,
      { "CommonTransportChannel-InformationResponse", "nbap.CommonTransportChannel_InformationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_secondary_CCPCH_parameters_02,
      { "secondary-CCPCH-parameters", "nbap.secondary_CCPCH_parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Secondary_CCPCHList_CTCH_ReconfRqstFDD", HFILL }},
    { &hf_nbap_pRACH_parameters_02,
      { "pRACH-parameters", "nbap.pRACH_parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PRACHList_CTCH_ReconfRqstFDD", HFILL }},
    { &hf_nbap_notUsed_cPCH_parameters,
      { "notUsed-cPCH-parameters", "nbap.notUsed_cPCH_parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_fACH_ParametersList_CTCH_ReconfRqstFDD,
      { "fACH-ParametersList-CTCH-ReconfRqstFDD", "nbap.fACH_ParametersList_CTCH_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_pCH_Parameters_CTCH_ReconfRqstFDD,
      { "pCH-Parameters-CTCH-ReconfRqstFDD", "nbap.pCH_Parameters_CTCH_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_pICH_Parameters_CTCH_ReconfRqstFDD,
      { "pICH-Parameters-CTCH-ReconfRqstFDD", "nbap.pICH_Parameters_CTCH_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_FACH_ParametersListIE_CTCH_ReconfRqstFDD_item,
      { "FACH-ParametersItem-CTCH-ReconfRqstFDD", "nbap.FACH_ParametersItem_CTCH_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_pRACH_ParametersList_CTCH_ReconfRqstFDD,
      { "pRACH-ParametersList-CTCH-ReconfRqstFDD", "nbap.pRACH_ParametersList_CTCH_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_aICH_ParametersList_CTCH_ReconfRqstFDD,
      { "aICH-ParametersList-CTCH-ReconfRqstFDD", "nbap.aICH_ParametersList_CTCH_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PRACH_ParametersListIE_CTCH_ReconfRqstFDD_item,
      { "PRACH-ParametersItem-CTCH-ReconfRqstFDD", "nbap.PRACH_ParametersItem_CTCH_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_allowedSlotFormatInformation_01,
      { "allowedSlotFormatInformation", "nbap.allowedSlotFormatInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        "AllowedSlotFormatInformationList_CTCH_ReconfRqstFDD", HFILL }},
    { &hf_nbap_AllowedSlotFormatInformationList_CTCH_ReconfRqstFDD_item,
      { "AllowedSlotFormatInformationItem-CTCH-ReconfRqstFDD", "nbap.AllowedSlotFormatInformationItem_CTCH_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_rACH_SlotFormat,
      { "rACH-SlotFormat", "nbap.rACH_SlotFormat",
        FT_UINT32, BASE_DEC, VALS(nbap_RACH_SlotFormat_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_AICH_ParametersListIE_CTCH_ReconfRqstFDD_item,
      { "AICH-ParametersItem-CTCH-ReconfRqstFDD", "nbap.AICH_ParametersItem_CTCH_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_secondaryCCPCHList,
      { "secondaryCCPCHList", "nbap.secondaryCCPCHList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Secondary_CCPCHList_CTCH_ReconfRqstTDD", HFILL }},
    { &hf_nbap_Secondary_CCPCHListIE_CTCH_ReconfRqstTDD_item,
      { "Secondary-CCPCHItem-CTCH-ReconfRqstTDD", "nbap.Secondary_CCPCHItem_CTCH_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sCCPCH_Power,
      { "sCCPCH-Power", "nbap.sCCPCH_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_Secondary_CCPCH_parameterExtendedList_CTCH_ReconfRqstTDD_item,
      { "Secondary-CCPCHItem-CTCH-ReconfRqstTDD", "nbap.Secondary_CCPCHItem_CTCH_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_ReconfRqstTDD_item,
      { "Secondary-CCPCHItem-CTCH-ReconfRqstTDD", "nbap.Secondary_CCPCHItem_CTCH_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_FACH_ParametersList_CTCH_ReconfRqstTDD_item,
      { "FACH-ParametersItem-CTCH-ReconfRqstTDD", "nbap.FACH_ParametersItem_CTCH_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_commonPhysicalChannelId,
      { "commonPhysicalChannelId", "nbap.commonPhysicalChannelId",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_fPACHPower,
      { "fPACHPower", "nbap.fPACHPower",
        FT_INT32, BASE_DEC, NULL, 0,
        "FPACH_Power", HFILL }},
    { &hf_nbap_secondaryCCPCH768List,
      { "secondaryCCPCH768List", "nbap.secondaryCCPCH768List",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Secondary_CCPCH_768_List_CTCH_ReconfRqstTDD", HFILL }},
    { &hf_nbap_Secondary_CCPCH_768_List_CTCH_ReconfRqstTDD_item,
      { "Secondary-CCPCH-768-Item-CTCH-ReconfRqstTDD", "nbap.Secondary_CCPCH_768_Item_CTCH_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uPPCHPositionLCR,
      { "uPPCHPositionLCR", "nbap.uPPCHPositionLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Cell_InformationList_AuditRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_configurationGenerationID,
      { "configurationGenerationID", "nbap.configurationGenerationID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_primary_SCH_Information,
      { "primary-SCH-Information", "nbap.primary_SCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "P_SCH_Information_AuditRsp", HFILL }},
    { &hf_nbap_secondary_SCH_Information,
      { "secondary-SCH-Information", "nbap.secondary_SCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "S_SCH_Information_AuditRsp", HFILL }},
    { &hf_nbap_primary_CPICH_Information,
      { "primary-CPICH-Information", "nbap.primary_CPICH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "P_CPICH_Information_AuditRsp", HFILL }},
    { &hf_nbap_secondary_CPICH_InformationList,
      { "secondary-CPICH-InformationList", "nbap.secondary_CPICH_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "S_CPICH_InformationList_AuditRsp", HFILL }},
    { &hf_nbap_primary_CCPCH_Information,
      { "primary-CCPCH-Information", "nbap.primary_CCPCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "P_CCPCH_Information_AuditRsp", HFILL }},
    { &hf_nbap_bCH_Information,
      { "bCH-Information", "nbap.bCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "BCH_Information_AuditRsp", HFILL }},
    { &hf_nbap_secondary_CCPCH_InformationList,
      { "secondary-CCPCH-InformationList", "nbap.secondary_CCPCH_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "S_CCPCH_InformationList_AuditRsp", HFILL }},
    { &hf_nbap_pCH_Information,
      { "pCH-Information", "nbap.pCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PCH_Information_AuditRsp", HFILL }},
    { &hf_nbap_pICH_Information,
      { "pICH-Information", "nbap.pICH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PICH_Information_AuditRsp", HFILL }},
    { &hf_nbap_fACH_InformationList,
      { "fACH-InformationList", "nbap.fACH_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "FACH_InformationList_AuditRsp", HFILL }},
    { &hf_nbap_pRACH_InformationList,
      { "pRACH-InformationList", "nbap.pRACH_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PRACH_InformationList_AuditRsp", HFILL }},
    { &hf_nbap_rACH_InformationList,
      { "rACH-InformationList", "nbap.rACH_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RACH_InformationList_AuditRsp", HFILL }},
    { &hf_nbap_aICH_InformationList,
      { "aICH-InformationList", "nbap.aICH_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "AICH_InformationList_AuditRsp", HFILL }},
    { &hf_nbap_notUsed_1_pCPCH_InformationList,
      { "notUsed-1-pCPCH-InformationList", "nbap.notUsed_1_pCPCH_InformationList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_notUsed_2_cPCH_InformationList,
      { "notUsed-2-cPCH-InformationList", "nbap.notUsed_2_cPCH_InformationList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_notUsed_3_aP_AICH_InformationList,
      { "notUsed-3-aP-AICH-InformationList", "nbap.notUsed_3_aP_AICH_InformationList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_notUsed_4_cDCA_ICH_InformationList,
      { "notUsed-4-cDCA-ICH-InformationList", "nbap.notUsed_4_cDCA_ICH_InformationList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sCH_Information,
      { "sCH-Information", "nbap.sCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "SCH_Information_AuditRsp", HFILL }},
    { &hf_nbap_S_CPICH_InformationList_AuditRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_S_CCPCH_InformationList_AuditRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_FACH_InformationList_AuditRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PRACH_InformationList_AuditRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RACH_InformationList_AuditRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_AICH_InformationList_AuditRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_FPACH_LCR_InformationList_AuditRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_S_CCPCH_InformationListExt_AuditRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_S_CCPCH_LCR_InformationListExt_AuditRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PLCCH_InformationList_AuditRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_S_CCPCH_768_InformationList_AuditRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PRACH_768_InformationList_AuditRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_RUCCH_InformationList_AuditRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_RUCCH_768_InformationList_AuditRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Cell_Frequency_List_Information_LCR_MulFreq_AuditRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UPPCH_LCR_InformationList_AuditRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleFreq_HS_DSCH_Resources_InformationList_AuditRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleFreq_E_DCH_Resources_InformationList_AuditRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CCP_InformationList_AuditRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_communicationControlPortID,
      { "communicationControlPortID", "nbap.communicationControlPortID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Local_Cell_InformationList_AuditRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dl_or_global_capacityCredit,
      { "dl-or-global-capacityCredit", "nbap.dl_or_global_capacityCredit",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ul_capacityCredit,
      { "ul-capacityCredit", "nbap.ul_capacityCredit",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_commonChannelsCapacityConsumptionLaw,
      { "commonChannelsCapacityConsumptionLaw", "nbap.commonChannelsCapacityConsumptionLaw",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dedicatedChannelsCapacityConsumptionLaw,
      { "dedicatedChannelsCapacityConsumptionLaw", "nbap.dedicatedChannelsCapacityConsumptionLaw",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_maximumDL_PowerCapability,
      { "maximumDL-PowerCapability", "nbap.maximumDL_PowerCapability",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_minSpreadingFactor,
      { "minSpreadingFactor", "nbap.minSpreadingFactor",
        FT_UINT32, BASE_DEC, VALS(nbap_MinSpreadingFactor_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_minimumDL_PowerCapability,
      { "minimumDL-PowerCapability", "nbap.minimumDL_PowerCapability",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_local_Cell_Group_ID,
      { "local-Cell-Group-ID", "nbap.local_Cell_Group_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Local_Cell_ID", HFILL }},
    { &hf_nbap_Local_Cell_Group_InformationList_AuditRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Power_Local_Cell_Group_InformationList_AuditRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_power_Local_Cell_Group_ID,
      { "power-Local-Cell-Group-ID", "nbap.power_Local_Cell_Group_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Local_Cell_ID", HFILL }},
    { &hf_nbap_cell,
      { "cell", "nbap.cell_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Cell_CM_Rqst", HFILL }},
    { &hf_nbap_rACH_01,
      { "rACH", "nbap.rACH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RACH_CM_Rqst", HFILL }},
    { &hf_nbap_notUsed_cPCH,
      { "notUsed-cPCH", "nbap.notUsed_cPCH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_extension_CommonMeasurementObjectType_CM_Rqst,
      { "extension-CommonMeasurementObjectType-CM-Rqst", "nbap.extension_CommonMeasurementObjectType_CM_Rqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_powerLocalCellGroupID,
      { "powerLocalCellGroupID", "nbap.powerLocalCellGroupID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Local_Cell_ID", HFILL }},
    { &hf_nbap_cell_01,
      { "cell", "nbap.cell_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Cell_CM_Rsp", HFILL }},
    { &hf_nbap_rACH_02,
      { "rACH", "nbap.rACH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RACH_CM_Rsp", HFILL }},
    { &hf_nbap_extension_CommonMeasurementObjectType_CM_Rsp,
      { "extension-CommonMeasurementObjectType-CM-Rsp", "nbap.extension_CommonMeasurementObjectType_CM_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cell_02,
      { "cell", "nbap.cell_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Cell_CM_Rprt", HFILL }},
    { &hf_nbap_rACH_03,
      { "rACH", "nbap.rACH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RACH_CM_Rprt", HFILL }},
    { &hf_nbap_extension_CommonMeasurementObjectType_CM_Rprt,
      { "extension-CommonMeasurementObjectType-CM-Rprt", "nbap.extension_CommonMeasurementObjectType_CM_Rprt_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_commonMeasurementValueInformation,
      { "commonMeasurementValueInformation", "nbap.commonMeasurementValueInformation",
        FT_UINT32, BASE_DEC, VALS(nbap_CommonMeasurementValueInformation_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_n_INSYNC_IND,
      { "n-INSYNC-IND", "nbap.n_INSYNC_IND",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_n_OUTSYNC_IND,
      { "n-OUTSYNC-IND", "nbap.n_OUTSYNC_IND",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_t_RLFAILURE,
      { "t-RLFAILURE", "nbap.t_RLFAILURE",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_primarySCH_Power,
      { "primarySCH-Power", "nbap.primarySCH_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_secondarySCH_Power,
      { "secondarySCH-Power", "nbap.secondarySCH_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_primaryCPICH_Power,
      { "primaryCPICH-Power", "nbap.primaryCPICH_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_transmitDiversityIndicator,
      { "transmitDiversityIndicator", "nbap.transmitDiversityIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_TransmitDiversityIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_SecondaryCPICH_InformationList_Cell_SetupRqstFDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_fDD_DL_ChannelisationCodeNumber,
      { "fDD-DL-ChannelisationCodeNumber", "nbap.fDD_DL_ChannelisationCodeNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_secondaryCPICH_Power,
      { "secondaryCPICH-Power", "nbap.secondaryCPICH_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_bCH_information,
      { "bCH-information", "nbap.bCH_information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "BCH_Information_Cell_SetupRqstFDD", HFILL }},
    { &hf_nbap_bCH_Power,
      { "bCH-Power", "nbap.bCH_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_powerRaiseLimit,
      { "powerRaiseLimit", "nbap.powerRaiseLimit",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dLPowerAveragingWindowSize,
      { "dLPowerAveragingWindowSize", "nbap.dLPowerAveragingWindowSize",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_iPDL_FDD_Parameters,
      { "iPDL-FDD-Parameters", "nbap.iPDL_FDD_Parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_iPDL_Indicator,
      { "iPDL-Indicator", "nbap.iPDL_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_IPDL_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_CellPortion_InformationList_Cell_SetupRqstFDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_associatedSecondaryCPICH,
      { "associatedSecondaryCPICH", "nbap.associatedSecondaryCPICH",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CommonPhysicalChannelID", HFILL }},
    { &hf_nbap_maximumTransmissionPowerforCellPortion,
      { "maximumTransmissionPowerforCellPortion", "nbap.maximumTransmissionPowerforCellPortion",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MaximumTransmissionPower", HFILL }},
    { &hf_nbap_syncCaseIndicator,
      { "syncCaseIndicator", "nbap.syncCaseIndicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "SyncCaseIndicator_Cell_SetupRqstTDD_PSCH", HFILL }},
    { &hf_nbap_sCH_Power,
      { "sCH-Power", "nbap.sCH_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_case1,
      { "case1", "nbap.case1_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Case1_Cell_SetupRqstTDD", HFILL }},
    { &hf_nbap_case2,
      { "case2", "nbap.case2_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Case2_Cell_SetupRqstTDD", HFILL }},
    { &hf_nbap_sCH_TimeSlot,
      { "sCH-TimeSlot", "nbap.sCH_TimeSlot",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_pCCPCH_Power,
      { "pCCPCH-Power", "nbap.pCCPCH_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sCTD_Indicator,
      { "sCTD-Indicator", "nbap.sCTD_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_SCTD_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TimeSlotConfigurationList_Cell_SetupRqstTDD_item,
      { "TimeSlotConfigurationItem-Cell-SetupRqstTDD", "nbap.TimeSlotConfigurationItem_Cell_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_timeSlotStatus,
      { "timeSlotStatus", "nbap.timeSlotStatus",
        FT_UINT32, BASE_DEC, VALS(nbap_TimeSlotStatus_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_timeSlotDirection,
      { "timeSlotDirection", "nbap.timeSlotDirection",
        FT_UINT32, BASE_DEC, VALS(nbap_TimeSlotDirection_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TimeSlotConfigurationList_LCR_Cell_SetupRqstTDD_item,
      { "TimeSlotConfigurationItem-LCR-Cell-SetupRqstTDD", "nbap.TimeSlotConfigurationItem_LCR_Cell_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dwPCH_Power,
      { "dwPCH-Power", "nbap.dwPCH_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_iPDL_TDD_Parameters,
      { "iPDL-TDD-Parameters", "nbap.iPDL_TDD_Parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_iPDL_TDD_Parameters_LCR,
      { "iPDL-TDD-Parameters-LCR", "nbap.iPDL_TDD_Parameters_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Cell_Frequency_List_LCR_MulFreq_Cell_SetupRqstTDD_item,
      { "Cell-Frequency-Item-LCR-MulFreq-Cell-SetupRqstTDD", "nbap.Cell_Frequency_Item_LCR_MulFreq_Cell_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_timeSlotConfigurationList_LCR_Cell_SetupRqstTDD,
      { "timeSlotConfigurationList-LCR-Cell-SetupRqstTDD", "nbap.timeSlotConfigurationList_LCR_Cell_SetupRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SecondaryCPICH_InformationList_Cell_ReconfRqstFDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_bCH_information_01,
      { "bCH-information", "nbap.bCH_information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "BCH_information_Cell_ReconfRqstFDD", HFILL }},
    { &hf_nbap_CellPortion_InformationList_Cell_ReconfRqstFDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TimeSlotConfigurationList_Cell_ReconfRqstTDD_item,
      { "TimeSlotConfigurationItem-Cell-ReconfRqstTDD", "nbap.TimeSlotConfigurationItem_Cell_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_TimeSlotConfigurationList_LCR_Cell_ReconfRqstTDD_item,
      { "TimeSlotConfigurationItem-LCR-Cell-ReconfRqstTDD", "nbap.TimeSlotConfigurationItem_LCR_Cell_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cell_Frequency_Add_LCR_MulFreq_Cell_ReconfRqstTDD,
      { "cell-Frequency-Add-LCR-MulFreq-Cell-ReconfRqstTDD", "nbap.cell_Frequency_Add_LCR_MulFreq_Cell_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cell_Frequency_ModifyList_LCR_MulFreq_Cell_ReconfRqstTDD,
      { "cell-Frequency-ModifyList-LCR-MulFreq-Cell-ReconfRqstTDD", "nbap.cell_Frequency_ModifyList_LCR_MulFreq_Cell_ReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cell_Frequency_Delete_LCR_MulFreq_Cell_ReconfRqstTDD,
      { "cell-Frequency-Delete-LCR-MulFreq-Cell-ReconfRqstTDD", "nbap.cell_Frequency_Delete_LCR_MulFreq_Cell_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_timeSlotConfigurationList_LCR_Cell_ReconfRqstTDD,
      { "timeSlotConfigurationList-LCR-Cell-ReconfRqstTDD", "nbap.timeSlotConfigurationList_LCR_Cell_ReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Cell_Frequency_ModifyList_LCR_MulFreq_Cell_ReconfRqstTDD_item,
      { "Cell-Frequency-ModifyItem-LCR-MulFreq-Cell-ReconfRqstTDD", "nbap.Cell_Frequency_ModifyItem_LCR_MulFreq_Cell_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_no_Failure,
      { "no-Failure", "nbap.no_Failure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "No_Failure_ResourceStatusInd", HFILL }},
    { &hf_nbap_serviceImpacting,
      { "serviceImpacting", "nbap.serviceImpacting_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "ServiceImpacting_ResourceStatusInd", HFILL }},
    { &hf_nbap_local_Cell_InformationList,
      { "local-Cell-InformationList", "nbap.local_Cell_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Local_Cell_InformationList_ResourceStatusInd", HFILL }},
    { &hf_nbap_local_Cell_Group_InformationList,
      { "local-Cell-Group-InformationList", "nbap.local_Cell_Group_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Local_Cell_Group_InformationList_ResourceStatusInd", HFILL }},
    { &hf_nbap_Local_Cell_InformationList_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_local_CellID,
      { "local-CellID", "nbap.local_CellID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Local_Cell_ID", HFILL }},
    { &hf_nbap_addorDeleteIndicator,
      { "addorDeleteIndicator", "nbap.addorDeleteIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_AddorDeleteIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_Local_Cell_Group_InformationList_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Power_Local_Cell_Group_InformationList_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_local_Cell_InformationList_01,
      { "local-Cell-InformationList", "nbap.local_Cell_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Local_Cell_InformationList2_ResourceStatusInd", HFILL }},
    { &hf_nbap_local_Cell_Group_InformationList_01,
      { "local-Cell-Group-InformationList", "nbap.local_Cell_Group_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Local_Cell_Group_InformationList2_ResourceStatusInd", HFILL }},
    { &hf_nbap_cCP_InformationList,
      { "cCP-InformationList", "nbap.cCP_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CCP_InformationList_ResourceStatusInd", HFILL }},
    { &hf_nbap_cell_InformationList,
      { "cell-InformationList", "nbap.cell_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Cell_InformationList_ResourceStatusInd", HFILL }},
    { &hf_nbap_Local_Cell_InformationList2_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_maximum_DL_PowerCapability,
      { "maximum-DL-PowerCapability", "nbap.maximum_DL_PowerCapability",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MaximumDL_PowerCapability", HFILL }},
    { &hf_nbap_Local_Cell_Group_InformationList2_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CCP_InformationList_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Cell_InformationList_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_primary_SCH_Information_01,
      { "primary-SCH-Information", "nbap.primary_SCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "P_SCH_Information_ResourceStatusInd", HFILL }},
    { &hf_nbap_secondary_SCH_Information_01,
      { "secondary-SCH-Information", "nbap.secondary_SCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "S_SCH_Information_ResourceStatusInd", HFILL }},
    { &hf_nbap_primary_CPICH_Information_01,
      { "primary-CPICH-Information", "nbap.primary_CPICH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "P_CPICH_Information_ResourceStatusInd", HFILL }},
    { &hf_nbap_secondary_CPICH_Information_01,
      { "secondary-CPICH-Information", "nbap.secondary_CPICH_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "S_CPICH_InformationList_ResourceStatusInd", HFILL }},
    { &hf_nbap_primary_CCPCH_Information_01,
      { "primary-CCPCH-Information", "nbap.primary_CCPCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "P_CCPCH_Information_ResourceStatusInd", HFILL }},
    { &hf_nbap_bCH_Information_01,
      { "bCH-Information", "nbap.bCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "BCH_Information_ResourceStatusInd", HFILL }},
    { &hf_nbap_secondary_CCPCH_InformationList_01,
      { "secondary-CCPCH-InformationList", "nbap.secondary_CCPCH_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "S_CCPCH_InformationList_ResourceStatusInd", HFILL }},
    { &hf_nbap_pCH_Information_01,
      { "pCH-Information", "nbap.pCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PCH_Information_ResourceStatusInd", HFILL }},
    { &hf_nbap_pICH_Information_01,
      { "pICH-Information", "nbap.pICH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PICH_Information_ResourceStatusInd", HFILL }},
    { &hf_nbap_fACH_InformationList_01,
      { "fACH-InformationList", "nbap.fACH_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "FACH_InformationList_ResourceStatusInd", HFILL }},
    { &hf_nbap_pRACH_InformationList_01,
      { "pRACH-InformationList", "nbap.pRACH_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PRACH_InformationList_ResourceStatusInd", HFILL }},
    { &hf_nbap_rACH_InformationList_01,
      { "rACH-InformationList", "nbap.rACH_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RACH_InformationList_ResourceStatusInd", HFILL }},
    { &hf_nbap_aICH_InformationList_01,
      { "aICH-InformationList", "nbap.aICH_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "AICH_InformationList_ResourceStatusInd", HFILL }},
    { &hf_nbap_sCH_Information_01,
      { "sCH-Information", "nbap.sCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "SCH_Information_ResourceStatusInd", HFILL }},
    { &hf_nbap_S_CPICH_InformationList_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_S_CCPCH_InformationList_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_FACH_InformationList_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PRACH_InformationList_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RACH_InformationList_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_AICH_InformationList_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_FPACH_LCR_InformationList_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_S_CCPCH_InformationListExt_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_S_CCPCH_LCR_InformationListExt_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PLCCH_InformationList_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_S_CCPCH_768_InformationList_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PRACH_768_InformationList_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_RUCCH_InformationList_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_RUCCH_768_InformationList_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Cell_Frequency_List_Information_LCR_MulFreq_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UPPCH_LCR_InformationList_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleFreq_HS_DSCH_Resources_InformationList_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Power_Local_Cell_Group_InformationList2_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleFreq_E_DCH_Resources_InformationList_ResourceStatusInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MIB_SB_SIB_InformationList_SystemInfoUpdateRqst_item,
      { "MIB-SB-SIB-InformationItem-SystemInfoUpdateRqst", "nbap.MIB_SB_SIB_InformationItem_SystemInfoUpdateRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_iB_Type,
      { "iB-Type", "nbap.iB_Type",
        FT_UINT32, BASE_DEC, VALS(nbap_IB_Type_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_iB_OC_ID,
      { "iB-OC-ID", "nbap.iB_OC_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_deletionIndicator,
      { "deletionIndicator", "nbap.deletionIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_DeletionIndicator_SystemInfoUpdate_vals), 0,
        "DeletionIndicator_SystemInfoUpdate", HFILL }},
    { &hf_nbap_no_Deletion,
      { "no-Deletion", "nbap.no_Deletion_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "No_Deletion_SystemInfoUpdate", HFILL }},
    { &hf_nbap_yes_Deletion,
      { "yes-Deletion", "nbap.yes_Deletion_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sIB_Originator,
      { "sIB-Originator", "nbap.sIB_Originator",
        FT_UINT32, BASE_DEC, VALS(nbap_SIB_Originator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_iB_SG_REP,
      { "iB-SG-REP", "nbap.iB_SG_REP",
        FT_UINT32, BASE_DEC, VALS(nbap_IB_SG_REP_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_segmentInformationList,
      { "segmentInformationList", "nbap.segmentInformationList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "SegmentInformationList_SystemInfoUpdate", HFILL }},
    { &hf_nbap_SegmentInformationListIE_SystemInfoUpdate_item,
      { "SegmentInformationItem-SystemInfoUpdate", "nbap.SegmentInformationItem_SystemInfoUpdate_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_iB_SG_POS,
      { "iB-SG-POS", "nbap.iB_SG_POS",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_segment_Type,
      { "segment-Type", "nbap.segment_Type",
        FT_UINT32, BASE_DEC, VALS(nbap_Segment_Type_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_iB_SG_DATA,
      { "iB-SG-DATA", "nbap.iB_SG_DATA",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_HS_Cell_Information_RL_Setup_List_item,
      { "Additional-HS-Cell-Information-RL-Setup-ItemIEs", "nbap.Additional_HS_Cell_Information_RL_Setup_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hSPDSCH_RL_ID,
      { "hSPDSCH-RL-ID", "nbap.hSPDSCH_RL_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RL_ID", HFILL }},
    { &hf_nbap_hS_DSCH_FDD_Secondary_Serving_Information,
      { "hS-DSCH-FDD-Secondary-Serving-Information", "nbap.hS_DSCH_FDD_Secondary_Serving_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_minUL_ChannelisationCodeLength,
      { "minUL-ChannelisationCodeLength", "nbap.minUL_ChannelisationCodeLength",
        FT_UINT32, BASE_DEC, VALS(nbap_MinUL_ChannelisationCodeLength_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_maxNrOfUL_DPDCHs,
      { "maxNrOfUL-DPDCHs", "nbap.maxNrOfUL_DPDCHs",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ul_DPCCH_SlotFormat,
      { "ul-DPCCH-SlotFormat", "nbap.ul_DPCCH_SlotFormat",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_diversityMode,
      { "diversityMode", "nbap.diversityMode",
        FT_UINT32, BASE_DEC, VALS(nbap_DiversityMode_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_not_Used_sSDT_CellID_Length,
      { "not-Used-sSDT-CellID-Length", "nbap.not_Used_sSDT_CellID_Length_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_not_Used_s_FieldLength,
      { "not-Used-s-FieldLength", "nbap.not_Used_s_FieldLength_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dl_DPCH_SlotFormat,
      { "dl-DPCH-SlotFormat", "nbap.dl_DPCH_SlotFormat",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tFCI_SignallingMode,
      { "tFCI-SignallingMode", "nbap.tFCI_SignallingMode_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_not_Used_pDSCH_RL_ID,
      { "not-Used-pDSCH-RL-ID", "nbap.not_Used_pDSCH_RL_ID_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_not_Used_pDSCH_CodeMapping,
      { "not-Used-pDSCH-CodeMapping", "nbap.not_Used_pDSCH_CodeMapping_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_powerOffsetInformation_01,
      { "powerOffsetInformation", "nbap.powerOffsetInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PowerOffsetInformation_RL_SetupRqstFDD", HFILL }},
    { &hf_nbap_pO2_ForTPC_Bits,
      { "pO2-ForTPC-Bits", "nbap.pO2_ForTPC_Bits",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PowerOffset", HFILL }},
    { &hf_nbap_RL_InformationList_RL_SetupRqstFDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_frameOffset,
      { "frameOffset", "nbap.frameOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_chipOffset,
      { "chipOffset", "nbap.chipOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_diversityControlField,
      { "diversityControlField", "nbap.diversityControlField",
        FT_UINT32, BASE_DEC, VALS(nbap_DiversityControlField_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_not_Used_sSDT_Cell_Identity,
      { "not-Used-sSDT-Cell-Identity", "nbap.not_Used_sSDT_Cell_Identity_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hSDSCH_Configured_Indicator,
      { "hSDSCH-Configured-Indicator", "nbap.hSDSCH_Configured_Indicator",
        FT_UINT32, BASE_DEC, VALS(nbap_HSDSCH_Configured_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_powerOffsetInformation_02,
      { "powerOffsetInformation", "nbap.powerOffsetInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PowerOffsetInformation_F_DPCH_RL_SetupRqstFDD", HFILL }},
    { &hf_nbap_UL_CCTrCH_InformationList_RL_SetupRqstTDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_DPCH_Information,
      { "uL-DPCH-Information", "nbap.uL_DPCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_DPCH_Information_RL_SetupRqstTDD", HFILL }},
    { &hf_nbap_tdd_DPCHOffset,
      { "tdd-DPCHOffset", "nbap.tdd_DPCHOffset",
        FT_UINT32, BASE_DEC, VALS(nbap_TDD_DPCHOffset_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_uL_Timeslot_Information,
      { "uL-Timeslot-Information", "nbap.uL_Timeslot_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_TimeslotLCR_Information,
      { "uL-TimeslotLCR-Information", "nbap.uL_TimeslotLCR_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_Timeslot768_Information,
      { "uL-Timeslot768-Information", "nbap.uL_Timeslot768_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_CCTrCH_InformationList_RL_SetupRqstTDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tdd_TPC_DownlinkStepSize,
      { "tdd-TPC-DownlinkStepSize", "nbap.tdd_TPC_DownlinkStepSize",
        FT_UINT32, BASE_DEC, VALS(nbap_TDD_TPC_DownlinkStepSize_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_cCTrCH_TPCList,
      { "cCTrCH-TPCList", "nbap.cCTrCH_TPCList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CCTrCH_TPCList_RL_SetupRqstTDD", HFILL }},
    { &hf_nbap_dL_DPCH_Information,
      { "dL-DPCH-Information", "nbap.dL_DPCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_DPCH_Information_RL_SetupRqstTDD", HFILL }},
    { &hf_nbap_CCTrCH_TPCList_RL_SetupRqstTDD_item,
      { "CCTrCH-TPCItem-RL-SetupRqstTDD", "nbap.CCTrCH_TPCItem_RL_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_Timeslot_Information,
      { "dL-Timeslot-Information", "nbap.dL_Timeslot_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_TimeslotLCR_Information,
      { "dL-TimeslotLCR-Information", "nbap.dL_TimeslotLCR_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_tstdIndicator,
      { "tstdIndicator", "nbap.tstdIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_TSTD_Indicator_vals), 0,
        "TSTD_Indicator", HFILL }},
    { &hf_nbap_dL_Timeslot768_Information,
      { "dL-Timeslot768-Information", "nbap.dL_Timeslot768_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_specialBurstScheduling,
      { "specialBurstScheduling", "nbap.specialBurstScheduling",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_TimeSlotISCPInfo,
      { "dL-TimeSlotISCPInfo", "nbap.dL_TimeSlotISCPInfo",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_HS_Cell_Information_Response_List_item,
      { "Additional-HS-Cell-Information-Response-ItemIEs", "nbap.Additional_HS_Cell_Information_Response_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationResponseList_RL_SetupRspFDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_diversityIndication,
      { "diversityIndication", "nbap.diversityIndication",
        FT_UINT32, BASE_DEC, VALS(nbap_DiversityIndication_RL_SetupRspFDD_vals), 0,
        "DiversityIndication_RL_SetupRspFDD", HFILL }},
    { &hf_nbap_not_Used_dSCH_InformationResponseList,
      { "not-Used-dSCH-InformationResponseList", "nbap.not_Used_dSCH_InformationResponseList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sSDT_SupportIndicator,
      { "sSDT-SupportIndicator", "nbap.sSDT_SupportIndicator",
        FT_UINT32, BASE_DEC, VALS(nbap_SSDT_SupportIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_combining,
      { "combining", "nbap.combining_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Combining_RL_SetupRspFDD", HFILL }},
    { &hf_nbap_nonCombiningOrFirstRL,
      { "nonCombiningOrFirstRL", "nbap.nonCombiningOrFirstRL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "NonCombiningOrFirstRL_RL_SetupRspFDD", HFILL }},
    { &hf_nbap_dCH_InformationResponse,
      { "dCH-InformationResponse", "nbap.dCH_InformationResponse",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_TimeSlot_ISCP_Info,
      { "uL-TimeSlot-ISCP-Info", "nbap.uL_TimeSlot_ISCP_Info",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ul_PhysCH_SF_Variation,
      { "ul-PhysCH-SF-Variation", "nbap.ul_PhysCH_SF_Variation",
        FT_UINT32, BASE_DEC, VALS(nbap_UL_PhysCH_SF_Variation_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_dCH_InformationResponseList,
      { "dCH-InformationResponseList", "nbap.dCH_InformationResponseList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DCH_InformationResponseList_RL_SetupRspTDD", HFILL }},
    { &hf_nbap_dSCH_InformationResponseList,
      { "dSCH-InformationResponseList", "nbap.dSCH_InformationResponseList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DSCH_InformationResponseList_RL_SetupRspTDD", HFILL }},
    { &hf_nbap_uSCH_InformationResponseList,
      { "uSCH-InformationResponseList", "nbap.uSCH_InformationResponseList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "USCH_InformationResponseList_RL_SetupRspTDD", HFILL }},
    { &hf_nbap_uL_TimeSlot_ISCP_LCR_Info,
      { "uL-TimeSlot-ISCP-LCR-Info", "nbap.uL_TimeSlot_ISCP_LCR_Info",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_generalCause,
      { "generalCause", "nbap.generalCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GeneralCauseList_RL_SetupFailureFDD", HFILL }},
    { &hf_nbap_rLSpecificCause,
      { "rLSpecificCause", "nbap.rLSpecificCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RLSpecificCauseList_RL_SetupFailureFDD", HFILL }},
    { &hf_nbap_unsuccessful_RL_InformationRespList_RL_SetupFailureFDD,
      { "unsuccessful-RL-InformationRespList-RL-SetupFailureFDD", "nbap.unsuccessful_RL_InformationRespList_RL_SetupFailureFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_successful_RL_InformationRespList_RL_SetupFailureFDD,
      { "successful-RL-InformationRespList-RL-SetupFailureFDD", "nbap.successful_RL_InformationRespList_RL_SetupFailureFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Unsuccessful_RL_InformationRespList_RL_SetupFailureFDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Successful_RL_InformationRespList_RL_SetupFailureFDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_diversityIndication_01,
      { "diversityIndication", "nbap.diversityIndication",
        FT_UINT32, BASE_DEC, VALS(nbap_DiversityIndication_RL_SetupFailureFDD_vals), 0,
        "DiversityIndication_RL_SetupFailureFDD", HFILL }},
    { &hf_nbap_not_Used_tFCI2_BearerInformationResponse,
      { "not-Used-tFCI2-BearerInformationResponse", "nbap.not_Used_tFCI2_BearerInformationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_combining_01,
      { "combining", "nbap.combining_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Combining_RL_SetupFailureFDD", HFILL }},
    { &hf_nbap_nonCombiningOrFirstRL_01,
      { "nonCombiningOrFirstRL", "nbap.nonCombiningOrFirstRL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "NonCombiningOrFirstRL_RL_SetupFailureFDD", HFILL }},
    { &hf_nbap_generalCause_01,
      { "generalCause", "nbap.generalCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GeneralCauseList_RL_SetupFailureTDD", HFILL }},
    { &hf_nbap_rLSpecificCause_01,
      { "rLSpecificCause", "nbap.rLSpecificCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RLSpecificCauseList_RL_SetupFailureTDD", HFILL }},
    { &hf_nbap_unsuccessful_RL_InformationRespItem_RL_SetupFailureTDD,
      { "unsuccessful-RL-InformationRespItem-RL-SetupFailureTDD", "nbap.unsuccessful_RL_InformationRespItem_RL_SetupFailureTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_HS_Cell_Information_RL_Addition_List_item,
      { "Additional-HS-Cell-Information-RL-Addition-ItemIEs", "nbap.Additional_HS_Cell_Information_RL_Addition_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency,
      { "setup-Or-Addition-Of-EDCH-On-secondary-UL-Frequency", "nbap.setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency",
        FT_UINT32, BASE_DEC, VALS(nbap_Setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_addition,
      { "addition", "nbap.addition",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Additional_EDCH_Cell_Information_To_Add_List", HFILL }},
    { &hf_nbap_RL_InformationList_RL_AdditionRqstFDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_initialDL_TransmissionPower,
      { "initialDL-TransmissionPower", "nbap.initialDL_TransmissionPower",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_maximumDL_Power,
      { "maximumDL-Power", "nbap.maximumDL_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_minimumDL_Power,
      { "minimumDL-Power", "nbap.minimumDL_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_not_Used_sSDT_CellIdentity,
      { "not-Used-sSDT-CellIdentity", "nbap.not_Used_sSDT_CellIdentity_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_CCTrCH_InformationList_RL_AdditionRqstTDD_item,
      { "UL-CCTrCH-InformationItem-RL-AdditionRqstTDD", "nbap.UL_CCTrCH_InformationItem_RL_AdditionRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_DPCH_Information_01,
      { "uL-DPCH-Information", "nbap.uL_DPCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_DPCH_InformationList_RL_AdditionRqstTDD", HFILL }},
    { &hf_nbap_DL_CCTrCH_InformationList_RL_AdditionRqstTDD_item,
      { "DL-CCTrCH-InformationItem-RL-AdditionRqstTDD", "nbap.DL_CCTrCH_InformationItem_RL_AdditionRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_DPCH_Information_01,
      { "dL-DPCH-Information", "nbap.dL_DPCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_DPCH_InformationList_RL_AdditionRqstTDD", HFILL }},
    { &hf_nbap_initial_DL_Transmission_Power,
      { "initial-DL-Transmission-Power", "nbap.initial_DL_Transmission_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_Additional_HS_Cell_Change_Information_Response_List_item,
      { "Additional-HS-Cell-Change-Information-Response-ItemIEs", "nbap.Additional_HS_Cell_Change_Information_Response_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_DSCH_Secondary_Serving_Cell_Change_Information_Response,
      { "hS-DSCH-Secondary-Serving-Cell-Change-Information-Response", "nbap.hS_DSCH_Secondary_Serving_Cell_Change_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationResponseList_RL_AdditionRspFDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_diversityIndication_02,
      { "diversityIndication", "nbap.diversityIndication",
        FT_UINT32, BASE_DEC, VALS(nbap_DiversityIndication_RL_AdditionRspFDD_vals), 0,
        "DiversityIndication_RL_AdditionRspFDD", HFILL }},
    { &hf_nbap_combining_02,
      { "combining", "nbap.combining_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Combining_RL_AdditionRspFDD", HFILL }},
    { &hf_nbap_non_combining,
      { "non-combining", "nbap.non_combining_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Non_Combining_RL_AdditionRspFDD", HFILL }},
    { &hf_nbap_dCH_Information,
      { "dCH-Information", "nbap.dCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DCH_Information_RL_AdditionRspTDD", HFILL }},
    { &hf_nbap_dSCH_InformationResponseList_01,
      { "dSCH-InformationResponseList", "nbap.dSCH_InformationResponseList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DSCH_InformationResponseList_RL_AdditionRspTDD", HFILL }},
    { &hf_nbap_uSCH_InformationResponseList_01,
      { "uSCH-InformationResponseList", "nbap.uSCH_InformationResponseList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "USCH_InformationResponseList_RL_AdditionRspTDD", HFILL }},
    { &hf_nbap_diversityIndication_03,
      { "diversityIndication", "nbap.diversityIndication",
        FT_UINT32, BASE_DEC, VALS(nbap_DiversityIndication_RL_AdditionRspTDD_vals), 0,
        "DiversityIndication_RL_AdditionRspTDD", HFILL }},
    { &hf_nbap_combining_03,
      { "combining", "nbap.combining_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Combining_RL_AdditionRspTDD", HFILL }},
    { &hf_nbap_non_Combining,
      { "non-Combining", "nbap.non_Combining_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Non_Combining_RL_AdditionRspTDD", HFILL }},
    { &hf_nbap_uL_TimeSlot_ISCP_InfoLCR,
      { "uL-TimeSlot-ISCP-InfoLCR", "nbap.uL_TimeSlot_ISCP_InfoLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UL_TimeSlot_ISCP_LCR_Info", HFILL }},
    { &hf_nbap_generalCause_02,
      { "generalCause", "nbap.generalCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GeneralCauseList_RL_AdditionFailureFDD", HFILL }},
    { &hf_nbap_rLSpecificCause_02,
      { "rLSpecificCause", "nbap.rLSpecificCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RLSpecificCauseList_RL_AdditionFailureFDD", HFILL }},
    { &hf_nbap_unsuccessful_RL_InformationRespList_RL_AdditionFailureFDD,
      { "unsuccessful-RL-InformationRespList-RL-AdditionFailureFDD", "nbap.unsuccessful_RL_InformationRespList_RL_AdditionFailureFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_successful_RL_InformationRespList_RL_AdditionFailureFDD,
      { "successful-RL-InformationRespList-RL-AdditionFailureFDD", "nbap.successful_RL_InformationRespList_RL_AdditionFailureFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Unsuccessful_RL_InformationRespList_RL_AdditionFailureFDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Successful_RL_InformationRespList_RL_AdditionFailureFDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_diversityIndication_04,
      { "diversityIndication", "nbap.diversityIndication",
        FT_UINT32, BASE_DEC, VALS(nbap_DiversityIndication_RL_AdditionFailureFDD_vals), 0,
        "DiversityIndication_RL_AdditionFailureFDD", HFILL }},
    { &hf_nbap_combining_04,
      { "combining", "nbap.combining_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Combining_RL_AdditionFailureFDD", HFILL }},
    { &hf_nbap_non_Combining_01,
      { "non-Combining", "nbap.non_Combining_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Non_Combining_RL_AdditionFailureFDD", HFILL }},
    { &hf_nbap_generalCause_03,
      { "generalCause", "nbap.generalCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GeneralCauseList_RL_AdditionFailureTDD", HFILL }},
    { &hf_nbap_rLSpecificCause_03,
      { "rLSpecificCause", "nbap.rLSpecificCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RLSpecificCauseList_RL_AdditionFailureTDD", HFILL }},
    { &hf_nbap_unsuccessful_RL_InformationRespItem_RL_AdditionFailureTDD,
      { "unsuccessful-RL-InformationRespItem-RL-AdditionFailureTDD", "nbap.unsuccessful_RL_InformationRespItem_RL_AdditionFailureTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_HS_Cell_Information_RL_Reconf_Prep_item,
      { "Additional-HS-Cell-Information-RL-Reconf-Prep-ItemIEs", "nbap.Additional_HS_Cell_Information_RL_Reconf_Prep_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_DSCH_Secondary_Serving_Information_To_Modify,
      { "hS-DSCH-Secondary-Serving-Information-To-Modify", "nbap.hS_DSCH_Secondary_Serving_Information_To_Modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_HS_DSCH_Secondary_Serving_Remove,
      { "hS-HS-DSCH-Secondary-Serving-Remove", "nbap.hS_HS_DSCH_Secondary_Serving_Remove_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "HS_DSCH_Secondary_Serving_Remove", HFILL }},
    { &hf_nbap_setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency,
      { "setup-Or-ConfigurationChange-Or-Removal-Of-EDCH-On-secondary-UL-Frequency", "nbap.setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency",
        FT_UINT32, BASE_DEC, VALS(nbap_Setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_not_Used_sSDT_CellIDLength,
      { "not-Used-sSDT-CellIDLength", "nbap.not_Used_sSDT_CellIDLength_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_powerOffsetInformation_03,
      { "powerOffsetInformation", "nbap.powerOffsetInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PowerOffsetInformation_RL_ReconfPrepFDD", HFILL }},
    { &hf_nbap_DCH_DeleteList_RL_ReconfPrepFDD_item,
      { "DCH-DeleteItem-RL-ReconfPrepFDD", "nbap.DCH_DeleteItem_RL_ReconfPrepFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationList_RL_ReconfPrepFDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_maxDL_Power,
      { "maxDL-Power", "nbap.maxDL_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_minDL_Power,
      { "minDL-Power", "nbap.minDL_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_not_Used_sSDT_Indication,
      { "not-Used-sSDT-Indication", "nbap.not_Used_sSDT_Indication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_powerOffsetInformation_04,
      { "powerOffsetInformation", "nbap.powerOffsetInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PowerOffsetInformation_F_DPCH_RL_ReconfPrepFDD", HFILL }},
    { &hf_nbap_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_item,
      { "UL-CCTrCH-InformationAddItem-RL-ReconfPrepTDD", "nbap.UL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ul_DPCH_InformationList,
      { "ul-DPCH-InformationList", "nbap.ul_DPCH_InformationList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_DPCH_InformationAddList_RL_ReconfPrepTDD", HFILL }},
    { &hf_nbap_uL_Timeslot_InformationLCR,
      { "uL-Timeslot-InformationLCR", "nbap.uL_Timeslot_InformationLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UL_TimeslotLCR_Information", HFILL }},
    { &hf_nbap_MultipleRL_UL_DPCH_InformationAddList_RL_ReconfPrepTDD_item,
      { "MultipleRL-UL-DPCH-InformationAddListIE-RL-ReconfPrepTDD", "nbap.MultipleRL_UL_DPCH_InformationAddListIE_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ul_DPCH_InformationListLCR,
      { "ul-DPCH-InformationListLCR", "nbap.ul_DPCH_InformationListLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD", HFILL }},
    { &hf_nbap_ul_sir_target,
      { "ul-sir-target", "nbap.ul_sir_target",
        FT_INT32, BASE_DEC, NULL, 0,
        "UL_SIR", HFILL }},
    { &hf_nbap_tDD_TPC_UplinkStepSize_LCR,
      { "tDD-TPC-UplinkStepSize-LCR", "nbap.tDD_TPC_UplinkStepSize_LCR",
        FT_UINT32, BASE_DEC, VALS(nbap_TDD_TPC_UplinkStepSize_LCR_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_uL_Timeslot_Information768,
      { "uL-Timeslot-Information768", "nbap.uL_Timeslot_Information768",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UL_Timeslot768_Information", HFILL }},
    { &hf_nbap_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_item,
      { "UL-CCTrCH-InformationModifyItem-RL-ReconfPrepTDD", "nbap.UL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ul_DPCH_InformationAddList,
      { "ul-DPCH-InformationAddList", "nbap.ul_DPCH_InformationAddList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_DPCH_InformationModify_AddList_RL_ReconfPrepTDD", HFILL }},
    { &hf_nbap_ul_DPCH_InformationModifyList,
      { "ul-DPCH-InformationModifyList", "nbap.ul_DPCH_InformationModifyList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_DPCH_InformationModify_ModifyList_RL_ReconfPrepTDD", HFILL }},
    { &hf_nbap_ul_DPCH_InformationDeleteList,
      { "ul-DPCH-InformationDeleteList", "nbap.ul_DPCH_InformationDeleteList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_DPCH_InformationModify_DeleteList_RL_ReconfPrepTDD", HFILL }},
    { &hf_nbap_uL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD,
      { "uL-Timeslot-InformationModify-ModifyList-RL-ReconfPrepTDD", "nbap.uL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD_item,
      { "UL-Timeslot-InformationModify-ModifyItem-RL-ReconfPrepTDD", "nbap.UL_Timeslot_InformationModify_ModifyItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD,
      { "uL-Code-InformationModify-ModifyList-RL-ReconfPrepTDD", "nbap.uL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD_item,
      { "UL-Code-InformationModify-ModifyItem-RL-ReconfPrepTDD", "nbap.UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_TimeslotLCR_InformationModify_ModifyList_RL_ReconfPrepTDD_item,
      { "UL-Timeslot-LCR-InformationModify-ModifyItem-RL-ReconfPrepTDD", "nbap.UL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_Code_InformationModify_ModifyList_RL_ReconfPrepTDDLCR,
      { "uL-Code-InformationModify-ModifyList-RL-ReconfPrepTDDLCR", "nbap.uL_Code_InformationModify_ModifyList_RL_ReconfPrepTDDLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDDLCR_item,
      { "UL-Code-InformationModify-ModifyItem-RL-ReconfPrepTDDLCR", "nbap.UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDDLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_Timeslot768_InformationModify_ModifyList_RL_ReconfPrepTDD_item,
      { "UL-Timeslot-768-InformationModify-ModifyItem-RL-ReconfPrepTDD", "nbap.UL_Timeslot_768_InformationModify_ModifyItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD768,
      { "uL-Code-InformationModify-ModifyList-RL-ReconfPrepTDD768", "nbap.uL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD768_item,
      { "UL-Code-InformationModify-ModifyItem-RL-ReconfPrepTDD768", "nbap.UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD_item,
      { "UL-DPCH-InformationModify-DeleteItem-RL-ReconfPrepTDD", "nbap.UL_DPCH_InformationModify_DeleteItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleRL_UL_DPCH_InformationModifyList_RL_ReconfPrepTDD_item,
      { "MultipleRL-UL-DPCH-InformationModifyListIE-RL-ReconfPrepTDD", "nbap.MultipleRL_UL_DPCH_InformationModifyListIE_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ul_DPCH_InformationAddListLCR,
      { "ul-DPCH-InformationAddListLCR", "nbap.ul_DPCH_InformationAddListLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD", HFILL }},
    { &hf_nbap_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_item,
      { "UL-CCTrCH-InformationDeleteItem-RL-ReconfPrepTDD", "nbap.UL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_item,
      { "DL-CCTrCH-InformationAddItem-RL-ReconfPrepTDD", "nbap.DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cCTrCH_TPCList_01,
      { "cCTrCH-TPCList", "nbap.cCTrCH_TPCList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CCTrCH_TPCAddList_RL_ReconfPrepTDD", HFILL }},
    { &hf_nbap_dl_DPCH_InformationList,
      { "dl-DPCH-InformationList", "nbap.dl_DPCH_InformationList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_DPCH_InformationAddList_RL_ReconfPrepTDD", HFILL }},
    { &hf_nbap_CCTrCH_TPCAddList_RL_ReconfPrepTDD_item,
      { "CCTrCH-TPCAddItem-RL-ReconfPrepTDD", "nbap.CCTrCH_TPCAddItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_Timeslot_InformationLCR,
      { "dL-Timeslot-InformationLCR", "nbap.dL_Timeslot_InformationLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DL_TimeslotLCR_Information", HFILL }},
    { &hf_nbap_MultipleRL_DL_DPCH_InformationAddList_RL_ReconfPrepTDD_item,
      { "MultipleRL-DL-DPCH-InformationAddListIE-RL-ReconfPrepTDD", "nbap.MultipleRL_DL_DPCH_InformationAddListIE_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dl_DPCH_InformationListLCR,
      { "dl-DPCH-InformationListLCR", "nbap.dl_DPCH_InformationListLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD", HFILL }},
    { &hf_nbap_cCTrCH_Initial_DL_Power,
      { "cCTrCH-Initial-DL-Power", "nbap.cCTrCH_Initial_DL_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_tDD_TPC_DownlinkStepSize,
      { "tDD-TPC-DownlinkStepSize", "nbap.tDD_TPC_DownlinkStepSize",
        FT_UINT32, BASE_DEC, VALS(nbap_TDD_TPC_DownlinkStepSize_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_cCTrCH_Maximum_DL_Power_InformationAdd_RL_ReconfPrepTDD,
      { "cCTrCH-Maximum-DL-Power-InformationAdd-RL-ReconfPrepTDD", "nbap.cCTrCH_Maximum_DL_Power_InformationAdd_RL_ReconfPrepTDD",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_cCTrCH_Minimum_DL_Power_InformationAdd_RL_ReconfPrepTDD,
      { "cCTrCH-Minimum-DL-Power-InformationAdd-RL-ReconfPrepTDD", "nbap.cCTrCH_Minimum_DL_Power_InformationAdd_RL_ReconfPrepTDD",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_dL_Timeslot_Information768,
      { "dL-Timeslot-Information768", "nbap.dL_Timeslot_Information768",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DL_Timeslot768_Information", HFILL }},
    { &hf_nbap_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_item,
      { "DL-CCTrCH-InformationModifyItem-RL-ReconfPrepTDD", "nbap.DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cCTrCH_TPCList_02,
      { "cCTrCH-TPCList", "nbap.cCTrCH_TPCList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CCTrCH_TPCModifyList_RL_ReconfPrepTDD", HFILL }},
    { &hf_nbap_dl_DPCH_InformationAddList,
      { "dl-DPCH-InformationAddList", "nbap.dl_DPCH_InformationAddList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_DPCH_InformationModify_AddList_RL_ReconfPrepTDD", HFILL }},
    { &hf_nbap_dl_DPCH_InformationModifyList,
      { "dl-DPCH-InformationModifyList", "nbap.dl_DPCH_InformationModifyList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_DPCH_InformationModify_ModifyList_RL_ReconfPrepTDD", HFILL }},
    { &hf_nbap_dl_DPCH_InformationDeleteList,
      { "dl-DPCH-InformationDeleteList", "nbap.dl_DPCH_InformationDeleteList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_DPCH_InformationModify_DeleteList_RL_ReconfPrepTDD", HFILL }},
    { &hf_nbap_CCTrCH_TPCModifyList_RL_ReconfPrepTDD_item,
      { "CCTrCH-TPCModifyItem-RL-ReconfPrepTDD", "nbap.CCTrCH_TPCModifyItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_Timeslot_InformationAddModify_ModifyList_RL_ReconfPrepTDD,
      { "dL-Timeslot-InformationAddModify-ModifyList-RL-ReconfPrepTDD", "nbap.dL_Timeslot_InformationAddModify_ModifyList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD", HFILL }},
    { &hf_nbap_DL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD_item,
      { "DL-Timeslot-InformationModify-ModifyItem-RL-ReconfPrepTDD", "nbap.DL_Timeslot_InformationModify_ModifyItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD,
      { "dL-Code-InformationModify-ModifyList-RL-ReconfPrepTDD", "nbap.dL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD_item,
      { "DL-Code-InformationModify-ModifyItem-RL-ReconfPrepTDD", "nbap.DL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD_item,
      { "DL-Timeslot-LCR-InformationModify-ModifyItem-RL-ReconfPrepTDD", "nbap.DL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_Code_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD,
      { "dL-Code-LCR-InformationModify-ModifyList-RL-ReconfPrepTDD", "nbap.dL_Code_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_Code_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD_item,
      { "DL-Code-LCR-InformationModify-ModifyItem-RL-ReconfPrepTDD", "nbap.DL_Code_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_Timeslot_768_InformationModify_ModifyList_RL_ReconfPrepTDD_item,
      { "DL-Timeslot-768-InformationModify-ModifyItem-RL-ReconfPrepTDD", "nbap.DL_Timeslot_768_InformationModify_ModifyItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_Code_768_InformationModify_ModifyList_RL_ReconfPrepTDD,
      { "dL-Code-768-InformationModify-ModifyList-RL-ReconfPrepTDD", "nbap.dL_Code_768_InformationModify_ModifyList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_Code_768_InformationModify_ModifyList_RL_ReconfPrepTDD_item,
      { "DL-Code-768-InformationModify-ModifyItem-RL-ReconfPrepTDD", "nbap.DL_Code_768_InformationModify_ModifyItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dPCH_ID768,
      { "dPCH-ID768", "nbap.dPCH_ID768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD_item,
      { "DL-DPCH-InformationModify-DeleteItem-RL-ReconfPrepTDD", "nbap.DL_DPCH_InformationModify_DeleteItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleRL_DL_DPCH_InformationModifyList_RL_ReconfPrepTDD_item,
      { "MultipleRL-DL-DPCH-InformationModifyListIE-RL-ReconfPrepTDD", "nbap.MultipleRL_DL_DPCH_InformationModifyListIE_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dl_DPCH_InformationAddListLCR,
      { "dl-DPCH-InformationAddListLCR", "nbap.dl_DPCH_InformationAddListLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD", HFILL }},
    { &hf_nbap_tDD_TPC_DownlinkStepSize_InformationModify_RL_ReconfPrepTDD,
      { "tDD-TPC-DownlinkStepSize-InformationModify-RL-ReconfPrepTDD", "nbap.tDD_TPC_DownlinkStepSize_InformationModify_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, VALS(nbap_TDD_TPC_DownlinkStepSize_vals), 0,
        "TDD_TPC_DownlinkStepSize", HFILL }},
    { &hf_nbap_cCTrCH_Maximum_DL_Power_InformationModify_RL_ReconfPrepTDD,
      { "cCTrCH-Maximum-DL-Power-InformationModify-RL-ReconfPrepTDD", "nbap.cCTrCH_Maximum_DL_Power_InformationModify_RL_ReconfPrepTDD",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_cCTrCH_Minimum_DL_Power_InformationModify_RL_ReconfPrepTDD,
      { "cCTrCH-Minimum-DL-Power-InformationModify-RL-ReconfPrepTDD", "nbap.cCTrCH_Minimum_DL_Power_InformationModify_RL_ReconfPrepTDD",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_item,
      { "DL-CCTrCH-InformationDeleteItem-RL-ReconfPrepTDD", "nbap.DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DCH_DeleteList_RL_ReconfPrepTDD_item,
      { "DCH-DeleteItem-RL-ReconfPrepTDD", "nbap.DCH_DeleteItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DSCH_Information_ModifyList_RL_ReconfPrepTDD_item,
      { "DSCH-Information-ModifyItem-RL-ReconfPrepTDD", "nbap.DSCH_Information_ModifyItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DSCH_Information_DeleteList_RL_ReconfPrepTDD_item,
      { "DSCH-Information-DeleteItem-RL-ReconfPrepTDD", "nbap.DSCH_Information_DeleteItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_USCH_Information_ModifyList_RL_ReconfPrepTDD_item,
      { "USCH-Information-ModifyItem-RL-ReconfPrepTDD", "nbap.USCH_Information_ModifyItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_USCH_Information_DeleteList_RL_ReconfPrepTDD_item,
      { "USCH-Information-DeleteItem-RL-ReconfPrepTDD", "nbap.USCH_Information_DeleteItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleRL_Information_RL_ReconfPrepTDD_item,
      { "RL-Information-RL-ReconfPrepTDD", "nbap.RL_Information_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationResponseList_RL_ReconfReady_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dCH_InformationResponseList_RL_ReconfReady,
      { "dCH-InformationResponseList-RL-ReconfReady", "nbap.dCH_InformationResponseList_RL_ReconfReady_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dSCH_InformationResponseList_RL_ReconfReady,
      { "dSCH-InformationResponseList-RL-ReconfReady", "nbap.dSCH_InformationResponseList_RL_ReconfReady_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uSCH_InformationResponseList_RL_ReconfReady,
      { "uSCH-InformationResponseList-RL-ReconfReady", "nbap.uSCH_InformationResponseList_RL_ReconfReady_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_generalCause_04,
      { "generalCause", "nbap.generalCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GeneralCauseList_RL_ReconfFailure", HFILL }},
    { &hf_nbap_rLSpecificCause_04,
      { "rLSpecificCause", "nbap.rLSpecificCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RLSpecificCauseList_RL_ReconfFailure", HFILL }},
    { &hf_nbap_rL_ReconfigurationFailureList_RL_ReconfFailure,
      { "rL-ReconfigurationFailureList-RL-ReconfFailure", "nbap.rL_ReconfigurationFailureList_RL_ReconfFailure",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_ReconfigurationFailureList_RL_ReconfFailure_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_HS_Cell_Information_RL_Reconf_Req_item,
      { "Additional-HS-Cell-Information-RL-Reconf-Req-ItemIEs", "nbap.Additional_HS_Cell_Information_RL_Reconf_Req_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised,
      { "hS-DSCH-FDD-Secondary-Serving-Information-To-Modify-Unsynchronised", "nbap.hS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_DSCH_Secondary_Serving_Remove,
      { "hS-DSCH-Secondary-Serving-Remove", "nbap.hS_DSCH_Secondary_Serving_Remove_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_ul_TFCS,
      { "ul-TFCS", "nbap.ul_TFCS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "TFCS", HFILL }},
    { &hf_nbap_dl_TFCS,
      { "dl-TFCS", "nbap.dl_TFCS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "TFCS", HFILL }},
    { &hf_nbap_DCH_DeleteList_RL_ReconfRqstFDD_item,
      { "DCH-DeleteItem-RL-ReconfRqstFDD", "nbap.DCH_DeleteItem_RL_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationList_RL_ReconfRqstFDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleRL_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_item,
      { "MultipleRL-DL-CCTrCH-InformationModifyListIE-RL-ReconfRqstTDD", "nbap.MultipleRL_DL_CCTrCH_InformationModifyListIE_RL_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dl_DPCH_LCR_InformationModifyList,
      { "dl-DPCH-LCR-InformationModifyList", "nbap.dl_DPCH_LCR_InformationModifyList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_DPCH_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD", HFILL }},
    { &hf_nbap_cCTrCH_Maximum_DL_Power_InformationModify_RL_ReconfRqstTDD,
      { "cCTrCH-Maximum-DL-Power-InformationModify-RL-ReconfRqstTDD", "nbap.cCTrCH_Maximum_DL_Power_InformationModify_RL_ReconfRqstTDD",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_cCTrCH_Minimum_DL_Power_InformationModify_RL_ReconfRqstTDD,
      { "cCTrCH-Minimum-DL-Power-InformationModify-RL-ReconfRqstTDD", "nbap.cCTrCH_Minimum_DL_Power_InformationModify_RL_ReconfRqstTDD",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_dL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD,
      { "dL-Timeslot-LCR-InformationModify-ModifyList-RL-ReconfRqstTDD", "nbap.dL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD_item,
      { "DL-Timeslot-LCR-InformationModify-ModifyItem-RL-ReconfRqstTDD", "nbap.DL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_maxPowerLCR,
      { "maxPowerLCR", "nbap.maxPowerLCR",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_minPowerLCR,
      { "minPowerLCR", "nbap.minPowerLCR",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DCH_DeleteList_RL_ReconfRqstTDD_item,
      { "DCH-DeleteItem-RL-ReconfRqstTDD", "nbap.DCH_DeleteItem_RL_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Multiple_RL_Information_RL_ReconfRqstTDD_item,
      { "RL-Information-RL-ReconfRqstTDD", "nbap.RL_Information_RL_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationResponseList_RL_ReconfRsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dCH_InformationResponseList_RL_ReconfRsp,
      { "dCH-InformationResponseList-RL-ReconfRsp", "nbap.dCH_InformationResponseList_RL_ReconfRsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_informationList_RL_DeletionRqst_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_ReferencePowerInformationList_DL_PC_Rqst_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dl_ReferencePower,
      { "dl-ReferencePower", "nbap.dl_ReferencePower",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_rL,
      { "rL", "nbap.rL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_DM_Rqst", HFILL }},
    { &hf_nbap_rLS,
      { "rLS", "nbap.rLS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_Set_DM_Rqst", HFILL }},
    { &hf_nbap_all_RL,
      { "all-RL", "nbap.all_RL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "AllRL_DM_Rqst", HFILL }},
    { &hf_nbap_all_RLS,
      { "all-RLS", "nbap.all_RLS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "AllRL_Set_DM_Rqst", HFILL }},
    { &hf_nbap_rL_InformationList,
      { "rL-InformationList", "nbap.rL_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RL_InformationList_DM_Rqst", HFILL }},
    { &hf_nbap_RL_InformationList_DM_Rqst_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PUSCH_Info_DM_Rqst_item,
      { "PUSCH-ID", "nbap.PUSCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSSICH_Info_DM_Rqst_item,
      { "HS-SICH-ID", "nbap.HS_SICH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSSICH_InfoExt_DM_Rqst_item,
      { "Extended-HS-SICH-ID", "nbap.Extended_HS_SICH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_rL_Set_InformationList_DM_Rqst,
      { "rL-Set-InformationList-DM-Rqst", "nbap.rL_Set_InformationList_DM_Rqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_Set_InformationList_DM_Rqst_item,
      { "RL-Set-InformationItem-DM-Rqst", "nbap.RL_Set_InformationItem_DM_Rqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_rL_01,
      { "rL", "nbap.rL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_DM_Rsp", HFILL }},
    { &hf_nbap_rLS_01,
      { "rLS", "nbap.rLS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_Set_DM_Rsp", HFILL }},
    { &hf_nbap_all_RL_01,
      { "all-RL", "nbap.all_RL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_DM_Rsp", HFILL }},
    { &hf_nbap_all_RLS_01,
      { "all-RLS", "nbap.all_RLS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_Set_DM_Rsp", HFILL }},
    { &hf_nbap_rL_InformationList_DM_Rsp,
      { "rL-InformationList-DM-Rsp", "nbap.rL_InformationList_DM_Rsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationList_DM_Rsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dedicatedMeasurementValue,
      { "dedicatedMeasurementValue", "nbap.dedicatedMeasurementValue",
        FT_UINT32, BASE_DEC, VALS(nbap_DedicatedMeasurementValue_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_PUSCH_Info_DM_Rsp_item,
      { "PUSCH-ID", "nbap.PUSCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Multiple_PUSCH_InfoList_DM_Rsp_item,
      { "Multiple-PUSCH-InfoListIE-DM-Rsp", "nbap.Multiple_PUSCH_InfoListIE_DM_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_pUSCH_ID,
      { "pUSCH-ID", "nbap.pUSCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp_item,
      { "Multiple-DedicatedMeasurementValueItem-TDD-DM-Rsp", "nbap.Multiple_DedicatedMeasurementValueItem_TDD_DM_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp_item,
      { "Multiple-DedicatedMeasurementValueItem-LCR-TDD-DM-Rsp", "nbap.Multiple_DedicatedMeasurementValueItem_LCR_TDD_DM_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp_item,
      { "Multiple-HSSICHMeasurementValueItem-TDD-DM-Rsp", "nbap.Multiple_HSSICHMeasurementValueItem_TDD_DM_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Multiple_DedicatedMeasurementValueList_768_TDD_DM_Rsp_item,
      { "Multiple-DedicatedMeasurementValueItem-768-TDD-DM-Rsp", "nbap.Multiple_DedicatedMeasurementValueItem_768_TDD_DM_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_rL_Set_InformationList_DM_Rsp,
      { "rL-Set-InformationList-DM-Rsp", "nbap.rL_Set_InformationList_DM_Rsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_Set_InformationList_DM_Rsp_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_rL_02,
      { "rL", "nbap.rL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_DM_Rprt", HFILL }},
    { &hf_nbap_rLS_02,
      { "rLS", "nbap.rLS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_Set_DM_Rprt", HFILL }},
    { &hf_nbap_all_RL_02,
      { "all-RL", "nbap.all_RL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_DM_Rprt", HFILL }},
    { &hf_nbap_all_RLS_02,
      { "all-RLS", "nbap.all_RLS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_Set_DM_Rprt", HFILL }},
    { &hf_nbap_rL_InformationList_DM_Rprt,
      { "rL-InformationList-DM-Rprt", "nbap.rL_InformationList_DM_Rprt",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationList_DM_Rprt_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dedicatedMeasurementValueInformation,
      { "dedicatedMeasurementValueInformation", "nbap.dedicatedMeasurementValueInformation",
        FT_UINT32, BASE_DEC, VALS(nbap_DedicatedMeasurementValueInformation_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_PUSCH_Info_DM_Rprt_item,
      { "PUSCH-ID", "nbap.PUSCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Multiple_PUSCH_InfoList_DM_Rprt_item,
      { "Multiple-PUSCH-InfoListIE-DM-Rprt", "nbap.Multiple_PUSCH_InfoListIE_DM_Rprt_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_rL_Set_InformationList_DM_Rprt,
      { "rL-Set-InformationList-DM-Rprt", "nbap.rL_Set_InformationList_DM_Rprt",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_Set_InformationList_DM_Rprt_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_rL_03,
      { "rL", "nbap.rL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_RL_FailureInd", HFILL }},
    { &hf_nbap_rL_Set,
      { "rL-Set", "nbap.rL_Set_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_Set_RL_FailureInd", HFILL }},
    { &hf_nbap_cCTrCH,
      { "cCTrCH", "nbap.cCTrCH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "CCTrCH_RL_FailureInd", HFILL }},
    { &hf_nbap_rL_InformationList_RL_FailureInd,
      { "rL-InformationList-RL-FailureInd", "nbap.rL_InformationList_RL_FailureInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationList_RL_FailureInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_rL_Set_InformationList_RL_FailureInd,
      { "rL-Set-InformationList-RL-FailureInd", "nbap.rL_Set_InformationList_RL_FailureInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_Set_InformationList_RL_FailureInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cCTrCH_InformationList_RL_FailureInd,
      { "cCTrCH-InformationList-RL-FailureInd", "nbap.cCTrCH_InformationList_RL_FailureInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CCTrCH_InformationList_RL_FailureInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationList_RL_PreemptRequiredInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_rL_04,
      { "rL", "nbap.rL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_RL_RestoreInd", HFILL }},
    { &hf_nbap_rL_Set_01,
      { "rL-Set", "nbap.rL_Set_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_Set_RL_RestoreInd", HFILL }},
    { &hf_nbap_cCTrCH_01,
      { "cCTrCH", "nbap.cCTrCH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "CCTrCH_RL_RestoreInd", HFILL }},
    { &hf_nbap_rL_InformationList_RL_RestoreInd,
      { "rL-InformationList-RL-RestoreInd", "nbap.rL_InformationList_RL_RestoreInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_InformationList_RL_RestoreInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_rL_Set_InformationList_RL_RestoreInd,
      { "rL-Set-InformationList-RL-RestoreInd", "nbap.rL_Set_InformationList_RL_RestoreInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_RL_Set_InformationList_RL_RestoreInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cCTrCH_InformationList_RL_RestoreInd,
      { "cCTrCH-InformationList-RL-RestoreInd", "nbap.cCTrCH_InformationList_RL_RestoreInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CCTrCH_InformationList_RL_RestoreInd_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_privateIEs,
      { "privateIEs", "nbap.privateIEs",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PrivateIE_Container", HFILL }},
    { &hf_nbap_HSDPA_And_EDCH_CellPortion_InformationList_PSCH_ReconfRqst_item,
      { "HSDPA-And-EDCH-CellPortion-InformationItem-PSCH-ReconfRqst", "nbap.HSDPA_And_EDCH_CellPortion_InformationItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_PDSCH_HS_SCCH_ScramblingCode_PSCH_ReconfRqst,
      { "hS-PDSCH-HS-SCCH-ScramblingCode-PSCH-ReconfRqst", "nbap.hS_PDSCH_HS_SCCH_ScramblingCode_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DL_ScramblingCode", HFILL }},
    { &hf_nbap_hS_PDSCH_FDD_Code_Information_PSCH_ReconfRqst,
      { "hS-PDSCH-FDD-Code-Information-PSCH-ReconfRqst", "nbap.hS_PDSCH_FDD_Code_Information_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "HS_PDSCH_FDD_Code_Information", HFILL }},
    { &hf_nbap_hS_SCCH_FDD_Code_Information_PSCH_ReconfRqst,
      { "hS-SCCH-FDD-Code-Information-PSCH-ReconfRqst", "nbap.hS_SCCH_FDD_Code_Information_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, VALS(nbap_HS_SCCH_FDD_Code_Information_vals), 0,
        "HS_SCCH_FDD_Code_Information", HFILL }},
    { &hf_nbap_hS_PDSCH_HS_SCCH_E_AGCH_E_RGCH_E_HICH_MaxPower_PSCH_ReconfRqst,
      { "hS-PDSCH-HS-SCCH-E-AGCH-E-RGCH-E-HICH-MaxPower-PSCH-ReconfRqst", "nbap.hS_PDSCH_HS_SCCH_E_AGCH_E_RGCH_E_HICH_MaxPower_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MaximumTransmissionPower", HFILL }},
    { &hf_nbap_e_AGCH_FDD_Code_Information,
      { "e-AGCH-FDD-Code-Information", "nbap.e_AGCH_FDD_Code_Information",
        FT_UINT32, BASE_DEC, VALS(nbap_E_AGCH_FDD_Code_Information_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_e_RGCH_E_HICH_FDD_Code_Information,
      { "e-RGCH-E-HICH-FDD-Code-Information", "nbap.e_RGCH_E_HICH_FDD_Code_Information",
        FT_UINT32, BASE_DEC, VALS(nbap_E_RGCH_E_HICH_FDD_Code_Information_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_PDSCHSets_AddList_PSCH_ReconfRqst_item,
      { "PDSCHSets-AddItem-PSCH-ReconfRqst", "nbap.PDSCHSets_AddItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_pDSCHSet_ID,
      { "pDSCHSet-ID", "nbap.pDSCHSet_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_pDSCH_InformationList,
      { "pDSCH-InformationList", "nbap.pDSCH_InformationList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PDSCH_Information_AddList_PSCH_ReconfRqst", HFILL }},
    { &hf_nbap_dL_Timeslot_InformationAddList_PSCH_ReconfRqst,
      { "dL-Timeslot-InformationAddList-PSCH-ReconfRqst", "nbap.dL_Timeslot_InformationAddList_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_Timeslot_InformationAddList_PSCH_ReconfRqst_item,
      { "DL-Timeslot-InformationAddItem-PSCH-ReconfRqst", "nbap.DL_Timeslot_InformationAddItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_Code_InformationAddList_PSCH_ReconfRqst,
      { "dL-Code-InformationAddList-PSCH-ReconfRqst", "nbap.dL_Code_InformationAddList_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_Code_InformationAddList_PSCH_ReconfRqst_item,
      { "DL-Code-InformationAddItem-PSCH-ReconfRqst", "nbap.DL_Code_InformationAddItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_pDSCH_ID,
      { "pDSCH-ID", "nbap.pDSCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst,
      { "dL-Timeslot-InformationAddList-LCR-PSCH-ReconfRqst", "nbap.dL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst_item,
      { "DL-Timeslot-InformationAddItem-LCR-PSCH-ReconfRqst", "nbap.DL_Timeslot_InformationAddItem_LCR_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_Code_InformationAddList_LCR_PSCH_ReconfRqst,
      { "dL-Code-InformationAddList-LCR-PSCH-ReconfRqst", "nbap.dL_Code_InformationAddList_LCR_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_Code_InformationAddList_LCR_PSCH_ReconfRqst_item,
      { "DL-Code-InformationAddItem-LCR-PSCH-ReconfRqst", "nbap.DL_Code_InformationAddItem_LCR_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_Timeslot_InformationAddList_768_PSCH_ReconfRqst,
      { "dL-Timeslot-InformationAddList-768-PSCH-ReconfRqst", "nbap.dL_Timeslot_InformationAddList_768_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_Timeslot_InformationAddList_768_PSCH_ReconfRqst_item,
      { "DL-Timeslot-InformationAddItem-768-PSCH-ReconfRqst", "nbap.DL_Timeslot_InformationAddItem_768_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_Code_InformationAddList_768_PSCH_ReconfRqst,
      { "dL-Code-InformationAddList-768-PSCH-ReconfRqst", "nbap.dL_Code_InformationAddList_768_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_Code_InformationAddList_768_PSCH_ReconfRqst_item,
      { "DL-Code-InformationAddItem-768-PSCH-ReconfRqst", "nbap.DL_Code_InformationAddItem_768_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_pDSCH_ID768,
      { "pDSCH-ID768", "nbap.pDSCH_ID768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PDSCHSets_ModifyList_PSCH_ReconfRqst_item,
      { "PDSCHSets-ModifyItem-PSCH-ReconfRqst", "nbap.PDSCHSets_ModifyItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_pDSCH_InformationList_01,
      { "pDSCH-InformationList", "nbap.pDSCH_InformationList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PDSCH_Information_ModifyList_PSCH_ReconfRqst", HFILL }},
    { &hf_nbap_dL_Timeslot_InformationModifyList_PSCH_ReconfRqst,
      { "dL-Timeslot-InformationModifyList-PSCH-ReconfRqst", "nbap.dL_Timeslot_InformationModifyList_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_Timeslot_InformationModifyList_PSCH_ReconfRqst_item,
      { "DL-Timeslot-InformationModifyItem-PSCH-ReconfRqst", "nbap.DL_Timeslot_InformationModifyItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_Code_InformationModifyList_PSCH_ReconfRqst,
      { "dL-Code-InformationModifyList-PSCH-ReconfRqst", "nbap.dL_Code_InformationModifyList_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_Code_InformationModifyList_PSCH_ReconfRqst_item,
      { "DL-Code-InformationModifyItem-PSCH-ReconfRqst", "nbap.DL_Code_InformationModifyItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst,
      { "dL-Timeslot-LCR-InformationModifyList-PSCH-ReconfRqst", "nbap.dL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst_item,
      { "DL-Timeslot-LCR-InformationModifyItem-PSCH-ReconfRqst", "nbap.DL_Timeslot_LCR_InformationModifyItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_Code_LCR_InformationModifyList_PSCH_ReconfRqst,
      { "dL-Code-LCR-InformationModifyList-PSCH-ReconfRqst", "nbap.dL_Code_LCR_InformationModifyList_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_Code_LCR_InformationModifyList_PSCH_ReconfRqst_item,
      { "DL-Code-LCR-InformationModifyItem-PSCH-ReconfRqst", "nbap.DL_Code_LCR_InformationModifyItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst,
      { "dL-Timeslot-768-InformationModifyList-PSCH-ReconfRqst", "nbap.dL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst_item,
      { "DL-Timeslot-768-InformationModifyItem-PSCH-ReconfRqst", "nbap.DL_Timeslot_768_InformationModifyItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_Code_768_InformationModifyList_PSCH_ReconfRqst,
      { "dL-Code-768-InformationModifyList-PSCH-ReconfRqst", "nbap.dL_Code_768_InformationModifyList_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_Code_768_InformationModifyList_PSCH_ReconfRqst_item,
      { "DL-Code-768-InformationModifyItem-PSCH-ReconfRqst", "nbap.DL_Code_768_InformationModifyItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PDSCHSets_DeleteList_PSCH_ReconfRqst_item,
      { "PDSCHSets-DeleteItem-PSCH-ReconfRqst", "nbap.PDSCHSets_DeleteItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PUSCHSets_AddList_PSCH_ReconfRqst_item,
      { "PUSCHSets-AddItem-PSCH-ReconfRqst", "nbap.PUSCHSets_AddItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_pUSCHSet_ID,
      { "pUSCHSet-ID", "nbap.pUSCHSet_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_pUSCH_InformationList,
      { "pUSCH-InformationList", "nbap.pUSCH_InformationList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PUSCH_Information_AddList_PSCH_ReconfRqst", HFILL }},
    { &hf_nbap_uL_Timeslot_InformationAddList_PSCH_ReconfRqst,
      { "uL-Timeslot-InformationAddList-PSCH-ReconfRqst", "nbap.uL_Timeslot_InformationAddList_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_Timeslot_InformationAddList_PSCH_ReconfRqst_item,
      { "UL-Timeslot-InformationAddItem-PSCH-ReconfRqst", "nbap.UL_Timeslot_InformationAddItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_Code_InformationAddList_PSCH_ReconfRqst,
      { "uL-Code-InformationAddList-PSCH-ReconfRqst", "nbap.uL_Code_InformationAddList_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_Code_InformationAddList_PSCH_ReconfRqst_item,
      { "UL-Code-InformationAddItem-PSCH-ReconfRqst", "nbap.UL_Code_InformationAddItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst,
      { "uL-Timeslot-InformationAddList-LCR-PSCH-ReconfRqst", "nbap.uL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst_item,
      { "UL-Timeslot-InformationAddItem-LCR-PSCH-ReconfRqst", "nbap.UL_Timeslot_InformationAddItem_LCR_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_Code_InformationAddList_LCR_PSCH_ReconfRqst,
      { "uL-Code-InformationAddList-LCR-PSCH-ReconfRqst", "nbap.uL_Code_InformationAddList_LCR_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_Code_InformationAddList_LCR_PSCH_ReconfRqst_item,
      { "UL-Code-InformationAddItem-LCR-PSCH-ReconfRqst", "nbap.UL_Code_InformationAddItem_LCR_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_Timeslot_InformationAddList_768_PSCH_ReconfRqst,
      { "uL-Timeslot-InformationAddList-768-PSCH-ReconfRqst", "nbap.uL_Timeslot_InformationAddList_768_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_Timeslot_InformationAddList_768_PSCH_ReconfRqst_item,
      { "UL-Timeslot-InformationAddItem-768-PSCH-ReconfRqst", "nbap.UL_Timeslot_InformationAddItem_768_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_Code_InformationAddList_768_PSCH_ReconfRqst,
      { "uL-Code-InformationAddList-768-PSCH-ReconfRqst", "nbap.uL_Code_InformationAddList_768_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_Code_InformationAddList_768_PSCH_ReconfRqst_item,
      { "UL-Code-InformationAddItem-768-PSCH-ReconfRqst", "nbap.UL_Code_InformationAddItem_768_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PUSCHSets_ModifyList_PSCH_ReconfRqst_item,
      { "PUSCHSets-ModifyItem-PSCH-ReconfRqst", "nbap.PUSCHSets_ModifyItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_pUSCH_InformationList_01,
      { "pUSCH-InformationList", "nbap.pUSCH_InformationList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PUSCH_Information_ModifyList_PSCH_ReconfRqst", HFILL }},
    { &hf_nbap_uL_Timeslot_InformationModifyList_PSCH_ReconfRqst,
      { "uL-Timeslot-InformationModifyList-PSCH-ReconfRqst", "nbap.uL_Timeslot_InformationModifyList_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_Timeslot_InformationModifyList_PSCH_ReconfRqst_item,
      { "UL-Timeslot-InformationModifyItem-PSCH-ReconfRqst", "nbap.UL_Timeslot_InformationModifyItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_Code_InformationModifyList_PSCH_ReconfRqst,
      { "uL-Code-InformationModifyList-PSCH-ReconfRqst", "nbap.uL_Code_InformationModifyList_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_Code_InformationModifyList_PSCH_ReconfRqst_item,
      { "UL-Code-InformationModifyItem-PSCH-ReconfRqst", "nbap.UL_Code_InformationModifyItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_Timeslot_InformationModifyList_LCR_PSCH_ReconfRqst,
      { "uL-Timeslot-InformationModifyList-LCR-PSCH-ReconfRqst", "nbap.uL_Timeslot_InformationModifyList_LCR_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst", HFILL }},
    { &hf_nbap_UL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst_item,
      { "UL-Timeslot-LCR-InformationModifyItem-PSCH-ReconfRqst", "nbap.UL_Timeslot_LCR_InformationModifyItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_Code_LCR_InformationModifyList_PSCH_ReconfRqst,
      { "uL-Code-LCR-InformationModifyList-PSCH-ReconfRqst", "nbap.uL_Code_LCR_InformationModifyList_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_Code_LCR_InformationModifyList_PSCH_ReconfRqst_item,
      { "UL-Code-LCR-InformationModifyItem-PSCH-ReconfRqst", "nbap.UL_Code_LCR_InformationModifyItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_Timeslot_InformationModifyList_768_PSCH_ReconfRqst,
      { "uL-Timeslot-InformationModifyList-768-PSCH-ReconfRqst", "nbap.uL_Timeslot_InformationModifyList_768_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst", HFILL }},
    { &hf_nbap_UL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst_item,
      { "UL-Timeslot-768-InformationModifyItem-PSCH-ReconfRqst", "nbap.UL_Timeslot_768_InformationModifyItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_uL_Code_768_InformationModifyList_PSCH_ReconfRqst,
      { "uL-Code-768-InformationModifyList-PSCH-ReconfRqst", "nbap.uL_Code_768_InformationModifyList_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UL_Code_768_InformationModifyList_PSCH_ReconfRqst_item,
      { "UL-Code-768-InformationModifyItem-PSCH-ReconfRqst", "nbap.UL_Code_768_InformationModifyItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_PUSCHSets_DeleteList_PSCH_ReconfRqst_item,
      { "PUSCHSets-DeleteItem-PSCH-ReconfRqst", "nbap.PUSCHSets_DeleteItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dL_HS_PDSCH_Timeslot_Information_PSCH_ReconfRqst,
      { "dL-HS-PDSCH-Timeslot-Information-PSCH-ReconfRqst", "nbap.dL_HS_PDSCH_Timeslot_Information_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_HS_PDSCH_Timeslot_Information_PSCH_ReconfRqst_item,
      { "DL-HS-PDSCH-Timeslot-InformationItem-PSCH-ReconfRqst", "nbap.DL_HS_PDSCH_Timeslot_InformationItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dl_HS_PDSCH_Codelist_PSCH_ReconfRqst,
      { "dl-HS-PDSCH-Codelist-PSCH-ReconfRqst", "nbap.dl_HS_PDSCH_Codelist_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_HS_PDSCH_Codelist_PSCH_ReconfRqst_item,
      { "TDD-ChannelisationCode", "nbap.TDD_ChannelisationCode",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_TDD_ChannelisationCode_vals_ext, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_HS_PDSCH_Timeslot_Information_768_PSCH_ReconfRqst_item,
      { "DL-HS-PDSCH-Timeslot-InformationItem-768-PSCH-ReconfRqst", "nbap.DL_HS_PDSCH_Timeslot_InformationItem_768_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_dl_HS_PDSCH_Codelist_768_PSCH_ReconfRqst,
      { "dl-HS-PDSCH-Codelist-768-PSCH-ReconfRqst", "nbap.dl_HS_PDSCH_Codelist_768_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DL_HS_PDSCH_Codelist_768_PSCH_ReconfRqst_item,
      { "TDD-ChannelisationCode768", "nbap.TDD_ChannelisationCode768",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_TDD_ChannelisationCode768_vals_ext, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_SCCH_Information_PSCH_ReconfRqst,
      { "hS-SCCH-Information-PSCH-ReconfRqst", "nbap.hS_SCCH_Information_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_SCCH_Information_LCR_PSCH_ReconfRqst,
      { "hS-SCCH-Information-LCR-PSCH-ReconfRqst", "nbap.hS_SCCH_Information_LCR_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_SCCH_Information_PSCH_ReconfRqst_item,
      { "HS-SCCH-InformationItem-PSCH-ReconfRqst", "nbap.HS_SCCH_InformationItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_SCCH_ID,
      { "hS-SCCH-ID", "nbap.hS_SCCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_SCCH_MaxPower,
      { "hS-SCCH-MaxPower", "nbap.hS_SCCH_MaxPower",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_hS_SICH_Information,
      { "hS-SICH-Information", "nbap.hS_SICH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "HS_SICH_Information_PSCH_ReconfRqst", HFILL }},
    { &hf_nbap_HS_SCCH_Information_LCR_PSCH_ReconfRqst_item,
      { "HS-SCCH-InformationItem-LCR-PSCH-ReconfRqst", "nbap.HS_SCCH_InformationItem_LCR_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_SICH_Information_LCR,
      { "hS-SICH-Information-LCR", "nbap.hS_SICH_Information_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "HS_SICH_Information_LCR_PSCH_ReconfRqst", HFILL }},
    { &hf_nbap_HS_SCCH_Information_768_PSCH_ReconfRqst_item,
      { "HS-SCCH-InformationItem-768-PSCH-ReconfRqst", "nbap.HS_SCCH_InformationItem_768_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_SICH_Information_768,
      { "hS-SICH-Information-768", "nbap.hS_SICH_Information_768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "HS_SICH_Information_768_PSCH_ReconfRqst", HFILL }},
    { &hf_nbap_HS_SCCH_InformationExt_LCR_PSCH_ReconfRqst_item,
      { "HS-SCCH-InformationItem-LCR-PSCH-ReconfRqst", "nbap.HS_SCCH_InformationItem_LCR_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_SCCH_InformationModify_PSCH_ReconfRqst,
      { "hS-SCCH-InformationModify-PSCH-ReconfRqst", "nbap.hS_SCCH_InformationModify_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_SCCH_InformationModify_LCR_PSCH_ReconfRqst,
      { "hS-SCCH-InformationModify-LCR-PSCH-ReconfRqst", "nbap.hS_SCCH_InformationModify_LCR_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_SICH_Information_01,
      { "hS-SICH-Information", "nbap.hS_SICH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "HS_SICH_InformationModify_PSCH_ReconfRqst", HFILL }},
    { &hf_nbap_HS_SCCH_InformationModify_LCR_PSCH_ReconfRqst_item,
      { "HS-SCCH-InformationModifyItem-LCR-PSCH-ReconfRqst", "nbap.HS_SCCH_InformationModifyItem_LCR_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_SICH_Information_LCR_01,
      { "hS-SICH-Information-LCR", "nbap.hS_SICH_Information_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "HS_SICH_InformationModify_LCR_PSCH_ReconfRqst", HFILL }},
    { &hf_nbap_HS_SCCH_InformationModifyExt_LCR_PSCH_ReconfRqst_item,
      { "HS-SCCH-InformationModifyItem-LCR-PSCH-ReconfRqst", "nbap.HS_SCCH_InformationModifyItem_LCR_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HS_SCCH_InformationModify_768_PSCH_ReconfRqst_item,
      { "HS-SCCH-InformationModifyItem-768-PSCH-ReconfRqst", "nbap.HS_SCCH_InformationModifyItem_768_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_SICH_Information_768_01,
      { "hS-SICH-Information-768", "nbap.hS_SICH_Information_768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "HS_SICH_InformationModify_768_PSCH_ReconfRqst", HFILL }},
    { &hf_nbap_HS_SCCH_InformationModify_PSCH_ReconfRqst_item,
      { "HS-SCCH-InformationModifyItem-PSCH-ReconfRqst", "nbap.HS_SCCH_InformationModifyItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Delete_From_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_item,
      { "Delete-From-HS-SCCH-Resource-PoolItem-PSCH-ReconfRqst", "nbap.Delete_From_HS_SCCH_Resource_PoolItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_lTGI_Presence,
      { "lTGI-Presence", "nbap.lTGI_Presence",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sNPL_Reporting_Type,
      { "sNPL-Reporting-Type", "nbap.sNPL_Reporting_Type",
        FT_UINT32, BASE_DEC, VALS(nbap_SNPL_Reporting_Type_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_e_PUCH_Timeslot_Info,
      { "e-PUCH-Timeslot-Info", "nbap.e_PUCH_Timeslot_Info",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_PUCH_Timeslot_Info_item,
      { "TimeSlot", "nbap.TimeSlot",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_AGCH_Information_PSCH_ReconfRqst,
      { "e-AGCH-Information-PSCH-ReconfRqst", "nbap.e_AGCH_Information_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_AGCH_Information_PSCH_ReconfRqst_item,
      { "E-AGCH-InformationItem-PSCH-ReconfRqst", "nbap.E_AGCH_InformationItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_AGCH_MaxPower,
      { "e-AGCH-MaxPower", "nbap.e_AGCH_MaxPower",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_e_AGCH_InformationModify_PSCH_ReconfRqst,
      { "e-AGCH-InformationModify-PSCH-ReconfRqst", "nbap.e_AGCH_InformationModify_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_AGCH_InformationModify_PSCH_ReconfRqst_item,
      { "E-AGCH-InformationModifyItem-PSCH-ReconfRqst", "nbap.E_AGCH_InformationModifyItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Delete_From_E_AGCH_Resource_Pool_PSCH_ReconfRqst_item,
      { "Delete-From-E-AGCH-Resource-PoolItem-PSCH-ReconfRqst", "nbap.Delete_From_E_AGCH_Resource_PoolItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_HICH_MaxPower,
      { "e-HICH-MaxPower", "nbap.e_HICH_MaxPower",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_e_AGCH_Information_768_PSCH_ReconfRqst,
      { "e-AGCH-Information-768-PSCH-ReconfRqst", "nbap.e_AGCH_Information_768_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_AGCH_Information_768_PSCH_ReconfRqst_item,
      { "E-AGCH-InformationItem-768-PSCH-ReconfRqst", "nbap.E_AGCH_InformationItem_768_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_AGCH_InformationModify_768_PSCH_ReconfRqst,
      { "e-AGCH-InformationModify-768-PSCH-ReconfRqst", "nbap.e_AGCH_InformationModify_768_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_AGCH_InformationModify_768_PSCH_ReconfRqst_item,
      { "E-AGCH-InformationModifyItem-768-PSCH-ReconfRqst", "nbap.E_AGCH_InformationModifyItem_768_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_PUCH_Timeslot_InfoLCR,
      { "e-PUCH-Timeslot-InfoLCR", "nbap.e_PUCH_Timeslot_InfoLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_PUCH_Timeslot_InfoLCR_item,
      { "E-PUCH-Timeslot-Item-InfoLCR", "nbap.E_PUCH_Timeslot_Item_InfoLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_PUCH_Codelist_LCR,
      { "e-PUCH-Codelist-LCR", "nbap.e_PUCH_Codelist_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_PUCH_Codelist_LCR_item,
      { "TDD-ChannelisationCode", "nbap.TDD_ChannelisationCode",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &nbap_TDD_ChannelisationCode_vals_ext, 0,
        NULL, HFILL }},
    { &hf_nbap_e_AGCH_Information_LCR_PSCH_ReconfRqst,
      { "e-AGCH-Information-LCR-PSCH-ReconfRqst", "nbap.e_AGCH_Information_LCR_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_AGCH_Information_LCR_PSCH_ReconfRqst_item,
      { "E-AGCH-InformationItem-LCR-PSCH-ReconfRqst", "nbap.E_AGCH_InformationItem_LCR_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_AGCH_InformationModify_LCR_PSCH_ReconfRqst,
      { "e-AGCH-InformationModify-LCR-PSCH-ReconfRqst", "nbap.e_AGCH_InformationModify_LCR_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_AGCH_InformationModify_LCR_PSCH_ReconfRqst_item,
      { "E-AGCH-InformationModifyItem-LCR-PSCH-ReconfRqst", "nbap.E_AGCH_InformationModifyItem_LCR_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_HICH_Information_LCR_PSCH_ReconfRqst,
      { "e-HICH-Information-LCR-PSCH-ReconfRqst", "nbap.e_HICH_Information_LCR_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_HICH_Information_LCR_PSCH_ReconfRqst_item,
      { "E-HICH-InformationItem-LCR-PSCH-ReconfRqst", "nbap.E_HICH_InformationItem_LCR_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_HICH_Type,
      { "e-HICH-Type", "nbap.e_HICH_Type",
        FT_UINT32, BASE_DEC, VALS(nbap_E_HICH_Type_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_e_HICH_InformationModify_LCR_PSCH_ReconfRqst,
      { "e-HICH-InformationModify-LCR-PSCH-ReconfRqst", "nbap.e_HICH_InformationModify_LCR_PSCH_ReconfRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_HICH_InformationModify_LCR_PSCH_ReconfRqst_item,
      { "E-HICH-InformationModifyItem-LCR-PSCH-ReconfRqst", "nbap.E_HICH_InformationModifyItem_LCR_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Delete_From_E_HICH_Resource_Pool_PSCH_ReconfRqst_item,
      { "Delete-From-E-HICH-Resource-PoolItem-PSCH-ReconfRqst", "nbap.Delete_From_E_HICH_Resource_PoolItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_eRUCCH_SYNC_UL_codes_bitmap,
      { "eRUCCH-SYNC-UL-codes-bitmap", "nbap.eRUCCH_SYNC_UL_codes_bitmap",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_nbap_Delete_From_HS_SCCH_Resource_PoolExt_PSCH_ReconfRqst_item,
      { "Delete-From-HS-SCCH-Resource-PoolItem-PSCH-ReconfRqst", "nbap.Delete_From_HS_SCCH_Resource_PoolItem_PSCH_ReconfRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_MultipleFreq_E_PUCH_Timeslot_InformationList_LCR_PSCH_ReconfRqst_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst_item,
      { "Max-RTWP-perUARFCN-Information-LCR-PSCH-ReconfRqst-Item", "nbap.Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_maximum_Target_ReceivedTotalWideBandPower_LCR,
      { "maximum-Target-ReceivedTotalWideBandPower-LCR", "nbap.maximum_Target_ReceivedTotalWideBandPower_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_HICH_TimeOffset_ExtensionLCR_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_e_HICH_TimeOffsetLCR,
      { "e-HICH-TimeOffsetLCR", "nbap.e_HICH_TimeOffsetLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_generalCause_05,
      { "generalCause", "nbap.generalCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GeneralCauseList_PSCH_ReconfFailure", HFILL }},
    { &hf_nbap_setSpecificCause,
      { "setSpecificCause", "nbap.setSpecificCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "SetSpecificCauseList_PSCH_ReconfFailureTDD", HFILL }},
    { &hf_nbap_extension_CauseLevel_PSCH_ReconfFailure,
      { "extension-CauseLevel-PSCH-ReconfFailure", "nbap.extension_CauseLevel_PSCH_ReconfFailure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_unsuccessful_PDSCHSetList_PSCH_ReconfFailureTDD,
      { "unsuccessful-PDSCHSetList-PSCH-ReconfFailureTDD", "nbap.unsuccessful_PDSCHSetList_PSCH_ReconfFailureTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_unsuccessful_PUSCHSetList_PSCH_ReconfFailureTDD,
      { "unsuccessful-PUSCHSetList-PSCH-ReconfFailureTDD", "nbap.unsuccessful_PUSCHSetList_PSCH_ReconfFailureTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Unsuccessful_PDSCHSetList_PSCH_ReconfFailureTDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Unsuccessful_PUSCHSetList_PSCH_ReconfFailureTDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_UARFCNSpecificCauseList_PSCH_ReconfFailureTDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_HICH_TimeOffset_ReconfFailureTDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hSDSCH_Common_System_Information_ResponseLCR,
      { "hSDSCH-Common-System-Information-ResponseLCR", "nbap.hSDSCH_Common_System_Information_ResponseLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hSDSCH_Paging_System_Information_ResponseLCR,
      { "hSDSCH-Paging-System-Information-ResponseLCR", "nbap.hSDSCH_Paging_System_Information_ResponseLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_common_EDCH_System_Information_ResponseLCR,
      { "common-EDCH-System-Information-ResponseLCR", "nbap.common_EDCH_System_Information_ResponseLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_communicationContext,
      { "communicationContext", "nbap.communicationContext_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "CommunicationContextList_Reset", HFILL }},
    { &hf_nbap_communicationControlPort,
      { "communicationControlPort", "nbap.communicationControlPort_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "CommunicationControlPortList_Reset", HFILL }},
    { &hf_nbap_nodeB,
      { "nodeB", "nbap.nodeB_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_communicationContextInfoList_Reset,
      { "communicationContextInfoList-Reset", "nbap.communicationContextInfoList_Reset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommunicationContextInfoList_Reset_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_communicationContextType_Reset,
      { "communicationContextType-Reset", "nbap.communicationContextType_Reset",
        FT_UINT32, BASE_DEC, VALS(nbap_CommunicationContextType_Reset_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_nodeB_CommunicationContextID,
      { "nodeB-CommunicationContextID", "nbap.nodeB_CommunicationContextID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_communicationControlPortInfoList_Reset,
      { "communicationControlPortInfoList-Reset", "nbap.communicationControlPortInfoList_Reset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CommunicationControlPortInfoList_Reset_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cell_03,
      { "cell", "nbap.cell_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Cell_InfEx_Rqst", HFILL }},
    { &hf_nbap_cell_04,
      { "cell", "nbap.cell_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Cell_InfEx_Rsp", HFILL }},
    { &hf_nbap_requestedDataValue,
      { "requestedDataValue", "nbap.requestedDataValue_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cell_05,
      { "cell", "nbap.cell_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Cell_Inf_Rprt", HFILL }},
    { &hf_nbap_requestedDataValueInformation,
      { "requestedDataValueInformation", "nbap.requestedDataValueInformation",
        FT_UINT32, BASE_DEC, VALS(nbap_RequestedDataValueInformation_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_TimeslotInfo_CellSyncInitiationRqstTDD_item,
      { "TimeSlot", "nbap.TimeSlot",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cSBTransmissionID,
      { "cSBTransmissionID", "nbap.cSBTransmissionID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sfn,
      { "sfn", "nbap.sfn",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_initialDLTransPower,
      { "initialDLTransPower", "nbap.initialDLTransPower",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_cSBMeasurementID,
      { "cSBMeasurementID", "nbap.cSBMeasurementID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_synchronisationReportType,
      { "synchronisationReportType", "nbap.synchronisationReportType",
        FT_UINT32, BASE_DEC, VALS(nbap_SynchronisationReportType_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_synchronisationReportCharacteristics,
      { "synchronisationReportCharacteristics", "nbap.synchronisationReportCharacteristics_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sYNCDlCodeId,
      { "sYNCDlCodeId", "nbap.sYNCDlCodeId",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSyncBurstTransReconfInfo_CellSyncReconfRqstTDD_item,
      { "CellSyncBurstTransInfoItem-CellSyncReconfRqstTDD", "nbap.CellSyncBurstTransInfoItem_CellSyncReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_syncFrameNumberToTransmit,
      { "syncFrameNumberToTransmit", "nbap.syncFrameNumberToTransmit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SyncFrameNumber", HFILL }},
    { &hf_nbap_dlTransPower,
      { "dlTransPower", "nbap.dlTransPower",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_cellSyncBurstMeasInfoList_CellSyncReconfRqstTDD,
      { "cellSyncBurstMeasInfoList-CellSyncReconfRqstTDD", "nbap.cellSyncBurstMeasInfoList_CellSyncReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_synchronisationReportType_01,
      { "synchronisationReportType", "nbap.synchronisationReportType_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "SynchronisationReportTypeIE", HFILL }},
    { &hf_nbap_synchronisationReportCharacteristics_01,
      { "synchronisationReportCharacteristics", "nbap.synchronisationReportCharacteristics_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "SynchronisationReportCharacteristicsIE", HFILL }},
    { &hf_nbap_CellSyncBurstMeasInfoListIE_CellSyncReconfRqstTDD_item,
      { "CellSyncBurstMeasInfoItem-CellSyncReconfRqstTDD", "nbap.CellSyncBurstMeasInfoItem_CellSyncReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_syncFrameNrToReceive,
      { "syncFrameNrToReceive", "nbap.syncFrameNrToReceive",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SyncFrameNumber", HFILL }},
    { &hf_nbap_syncBurstInfo,
      { "syncBurstInfo", "nbap.syncBurstInfo",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CellSyncBurstInfoList_CellSyncReconfRqstTDD", HFILL }},
    { &hf_nbap_CellSyncBurstInfoList_CellSyncReconfRqstTDD_item,
      { "CellSyncBurstInfoItem-CellSyncReconfRqstTDD", "nbap.CellSyncBurstInfoItem_CellSyncReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SYNCDlCodeIdTransReconfInfoLCR_CellSyncReconfRqstTDD_item,
      { "SYNCDlCodeIdTransReconfItemLCR-CellSyncReconfRqstTDD", "nbap.SYNCDlCodeIdTransReconfItemLCR_CellSyncReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_syncFrameNumberforTransmit,
      { "syncFrameNumberforTransmit", "nbap.syncFrameNumberforTransmit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SyncFrameNumber", HFILL }},
    { &hf_nbap_sYNCDlCodeIdMeasInfoList,
      { "sYNCDlCodeIdMeasInfoList", "nbap.sYNCDlCodeIdMeasInfoList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SYNCDlCodeIdMeasInfoList_CellSyncReconfRqstTDD", HFILL }},
    { &hf_nbap_SYNCDlCodeIdMeasInfoList_CellSyncReconfRqstTDD_item,
      { "SYNCDlCodeIdMeasInfoItem-CellSyncReconfRqstTDD", "nbap.SYNCDlCodeIdMeasInfoItem_CellSyncReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_sYNCDlCodeIdInfoLCR,
      { "sYNCDlCodeIdInfoLCR", "nbap.sYNCDlCodeIdInfoLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SYNCDlCodeIdInfoListLCR_CellSyncReconfRqstTDD", HFILL }},
    { &hf_nbap_SYNCDlCodeIdInfoListLCR_CellSyncReconfRqstTDD_item,
      { "SYNCDlCodeIdInfoItemLCR-CellSyncReconfRqstTDD", "nbap.SYNCDlCodeIdInfoItemLCR_CellSyncReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_propagationDelayCompensation,
      { "propagationDelayCompensation", "nbap.propagationDelayCompensation",
        FT_UINT32, BASE_DEC, VALS(nbap_TimingAdjustmentValueLCR_vals), 0,
        "TimingAdjustmentValueLCR", HFILL }},
    { &hf_nbap_CellAdjustmentInfo_SyncAdjustmentRqstTDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_frameAdjustmentValue,
      { "frameAdjustmentValue", "nbap.frameAdjustmentValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_timingAdjustmentValue,
      { "timingAdjustmentValue", "nbap.timingAdjustmentValue",
        FT_UINT32, BASE_DEC, VALS(nbap_TimingAdjustmentValue_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_dLTransPower,
      { "dLTransPower", "nbap.dLTransPower",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_nbap_generalCause_06,
      { "generalCause", "nbap.generalCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GeneralCauseList_SyncAdjustmntFailureTDD", HFILL }},
    { &hf_nbap_cellSpecificCause,
      { "cellSpecificCause", "nbap.cellSpecificCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "CellSpecificCauseList_SyncAdjustmntFailureTDD", HFILL }},
    { &hf_nbap_unsuccessful_cell_InformationRespList_SyncAdjustmntFailureTDD,
      { "unsuccessful-cell-InformationRespList-SyncAdjustmntFailureTDD", "nbap.unsuccessful_cell_InformationRespList_SyncAdjustmntFailureTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Unsuccessful_cell_InformationRespList_SyncAdjustmntFailureTDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_CellSyncInfo_CellSyncReprtTDD_item,
      { "CellSyncInfoItemIE-CellSyncReprtTDD", "nbap.CellSyncInfoItemIE_CellSyncReprtTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_c_ID_CellSyncReprtTDD,
      { "c-ID-CellSyncReprtTDD", "nbap.c_ID_CellSyncReprtTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "C_ID_IE_CellSyncReprtTDD", HFILL }},
    { &hf_nbap_syncReportType_CellSyncReprtTDD,
      { "syncReportType-CellSyncReprtTDD", "nbap.syncReportType_CellSyncReprtTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "SyncReportTypeIE_CellSyncReprtTDD", HFILL }},
    { &hf_nbap_intStdPhSyncInfo_CellSyncReprtTDD,
      { "intStdPhSyncInfo-CellSyncReprtTDD", "nbap.intStdPhSyncInfo_CellSyncReprtTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "IntStdPhCellSyncInfo_CellSyncReprtTDD", HFILL }},
    { &hf_nbap_lateEntrantCell,
      { "lateEntrantCell", "nbap.lateEntrantCell_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_frequencyAcquisition,
      { "frequencyAcquisition", "nbap.frequencyAcquisition_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cellSyncBurstMeasuredInfo,
      { "cellSyncBurstMeasuredInfo", "nbap.cellSyncBurstMeasuredInfo",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CellSyncBurstMeasInfoList_CellSyncReprtTDD", HFILL }},
    { &hf_nbap_CellSyncBurstMeasInfoList_CellSyncReprtTDD_item,
      { "CellSyncBurstMeasInfoItem-CellSyncReprtTDD", "nbap.CellSyncBurstMeasInfoItem_CellSyncReprtTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cellSyncBurstInfo_CellSyncReprtTDD,
      { "cellSyncBurstInfo-CellSyncReprtTDD", "nbap.cellSyncBurstInfo_CellSyncReprtTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SEQUENCE_SIZE_1_maxNrOfReceptsPerSyncFrame_OF_CellSyncBurstInfo_CellSyncReprtTDD", HFILL }},
    { &hf_nbap_cellSyncBurstInfo_CellSyncReprtTDD_item,
      { "CellSyncBurstInfo-CellSyncReprtTDD", "nbap.CellSyncBurstInfo_CellSyncReprtTDD",
        FT_UINT32, BASE_DEC, VALS(nbap_CellSyncBurstInfo_CellSyncReprtTDD_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_cellSyncBurstAvailable,
      { "cellSyncBurstAvailable", "nbap.cellSyncBurstAvailable_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "CellSyncBurstAvailable_CellSyncReprtTDD", HFILL }},
    { &hf_nbap_cellSyncBurstNotAvailable,
      { "cellSyncBurstNotAvailable", "nbap.cellSyncBurstNotAvailable_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_cellSyncBurstSIR,
      { "cellSyncBurstSIR", "nbap.cellSyncBurstSIR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SyncDLCodeIdsMeasInfoList_CellSyncReprtTDD_item,
      { "SyncDLCodeIdsMeasInfoItem-CellSyncReprtTDD", "nbap.SyncDLCodeIdsMeasInfoItem_CellSyncReprtTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_syncDLCodeIdInfo_CellSyncReprtTDD,
      { "syncDLCodeIdInfo-CellSyncReprtTDD", "nbap.syncDLCodeIdInfo_CellSyncReprtTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_SyncDLCodeIdInfo_CellSyncReprtTDD_item,
      { "SyncDLCodeIdItem-CellSyncReprtTDD", "nbap.SyncDLCodeIdItem_CellSyncReprtTDD",
        FT_UINT32, BASE_DEC, VALS(nbap_SyncDLCodeIdItem_CellSyncReprtTDD_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_syncDLCodeIdAvailable,
      { "syncDLCodeIdAvailable", "nbap.syncDLCodeIdAvailable_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "SyncDLCodeIdAvailable_CellSyncReprtTDD", HFILL }},
    { &hf_nbap_syncDLCodeIDNotAvailable,
      { "syncDLCodeIDNotAvailable", "nbap.syncDLCodeIDNotAvailable_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_syncDLCodeIdTiming,
      { "syncDLCodeIdTiming", "nbap.syncDLCodeIdTiming",
        FT_UINT32, BASE_DEC, VALS(nbap_CellSyncBurstTimingLCR_vals), 0,
        "CellSyncBurstTimingLCR", HFILL }},
    { &hf_nbap_syncDLCodeIdSIR,
      { "syncDLCodeIdSIR", "nbap.syncDLCodeIdSIR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CellSyncBurstSIR", HFILL }},
    { &hf_nbap_DCH_RearrangeList_Bearer_RearrangeInd_item,
      { "DCH-RearrangeItem-Bearer-RearrangeInd", "nbap.DCH_RearrangeItem_Bearer_RearrangeInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_DSCH_RearrangeList_Bearer_RearrangeInd_item,
      { "DSCH-RearrangeItem-Bearer-RearrangeInd", "nbap.DSCH_RearrangeItem_Bearer_RearrangeInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_USCH_RearrangeList_Bearer_RearrangeInd_item,
      { "USCH-RearrangeItem-Bearer-RearrangeInd", "nbap.USCH_RearrangeItem_Bearer_RearrangeInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_HSDSCH_RearrangeList_Bearer_RearrangeInd_item,
      { "HSDSCH-RearrangeItem-Bearer-RearrangeInd", "nbap.HSDSCH_RearrangeItem_Bearer_RearrangeInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_E_DCH_RearrangeList_Bearer_RearrangeInd_item,
      { "E-DCH-RearrangeItem-Bearer-RearrangeInd", "nbap.E_DCH_RearrangeItem_Bearer_RearrangeInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_EDCH_Cell_Information_Bearer_Rearrangement_List_item,
      { "Additional-EDCH-Cell-Information-Bearer-Rearrangement-ItemIEs", "nbap.Additional_EDCH_Cell_Information_Bearer_Rearrangement_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_transport_Bearer_Rearrangement_Indicator_for_Additional_EDCH_Separate_Mode,
      { "transport-Bearer-Rearrangement-Indicator-for-Additional-EDCH-Separate-Mode", "nbap.transport_Bearer_Rearrangement_Indicator_for_Additional_EDCH_Separate_Mode",
        FT_UINT32, BASE_DEC, VALS(nbap_Transport_Bearer_Rearrangement_Indicator_for_Additional_EDCH_Separate_Mode_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_DelayedActivationInformationList_RL_ActivationCmdFDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_delayed_activation_update,
      { "delayed-activation-update", "nbap.delayed_activation_update",
        FT_UINT32, BASE_DEC, VALS(nbap_DelayedActivationUpdate_vals), 0,
        "DelayedActivationUpdate", HFILL }},
    { &hf_nbap_DelayedActivationInformationList_RL_ActivationCmdTDD_item,
      { "ProtocolIE-Single-Container", "nbap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_HS_Cell_Information_RL_Param_Upd_item,
      { "Additional-HS-Cell-Information-RL-Param-Upd-ItemIEs", "nbap.Additional_HS_Cell_Information_RL_Param_Upd_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_hS_DSCH_FDD_Secondary_Serving_Update_Information,
      { "hS-DSCH-FDD-Secondary-Serving-Update-Information", "nbap.hS_DSCH_FDD_Secondary_Serving_Update_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_Additional_EDCH_Cell_Information_RL_Param_Upd_item,
      { "Additional-EDCH-Cell-Information-RL-Param-Upd-ItemIEs", "nbap.Additional_EDCH_Cell_Information_RL_Param_Upd_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_additional_EDCH_FDD_Update_Information,
      { "additional-EDCH-FDD-Update-Information", "nbap.additional_EDCH_FDD_Update_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_initiatingMessage,
      { "initiatingMessage", "nbap.initiatingMessage_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_succesfulOutcome,
      { "succesfulOutcome", "nbap.succesfulOutcome_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "SuccessfulOutcome", HFILL }},
    { &hf_nbap_unsuccesfulOutcome,
      { "unsuccesfulOutcome", "nbap.unsuccesfulOutcome_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UnsuccessfulOutcome", HFILL }},
    { &hf_nbap_outcome,
      { "outcome", "nbap.outcome_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_nbap_messageDiscriminator,
      { "messageDiscriminator", "nbap.messageDiscriminator",
        FT_UINT32, BASE_DEC, VALS(nbap_MessageDiscriminator_vals), 0,
        NULL, HFILL }},
    { &hf_nbap_initiatingMessagevalue,
      { "value", "nbap.value_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "InitiatingMessage_value", HFILL }},
    { &hf_nbap_successfulOutcome_value,
      { "value", "nbap.value_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "SuccessfulOutcome_value", HFILL }},
    { &hf_nbap_unsuccessfulOutcome_value,
      { "value", "nbap.value_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UnsuccessfulOutcome_value", HFILL }},
    { &hf_nbap_outcome_value,
      { "value", "nbap.value_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Outcome_value", HFILL }},
    { &hf_nbap_PreambleSignatures_signature15,
      { "signature15", "nbap.PreambleSignatures.signature15",
        FT_BOOLEAN, 8, NULL, 0x80,
        NULL, HFILL }},
    { &hf_nbap_PreambleSignatures_signature14,
      { "signature14", "nbap.PreambleSignatures.signature14",
        FT_BOOLEAN, 8, NULL, 0x40,
        NULL, HFILL }},
    { &hf_nbap_PreambleSignatures_signature13,
      { "signature13", "nbap.PreambleSignatures.signature13",
        FT_BOOLEAN, 8, NULL, 0x20,
        NULL, HFILL }},
    { &hf_nbap_PreambleSignatures_signature12,
      { "signature12", "nbap.PreambleSignatures.signature12",
        FT_BOOLEAN, 8, NULL, 0x10,
        NULL, HFILL }},
    { &hf_nbap_PreambleSignatures_signature11,
      { "signature11", "nbap.PreambleSignatures.signature11",
        FT_BOOLEAN, 8, NULL, 0x08,
        NULL, HFILL }},
    { &hf_nbap_PreambleSignatures_signature10,
      { "signature10", "nbap.PreambleSignatures.signature10",
        FT_BOOLEAN, 8, NULL, 0x04,
        NULL, HFILL }},
    { &hf_nbap_PreambleSignatures_signature9,
      { "signature9", "nbap.PreambleSignatures.signature9",
        FT_BOOLEAN, 8, NULL, 0x02,
        NULL, HFILL }},
    { &hf_nbap_PreambleSignatures_signature8,
      { "signature8", "nbap.PreambleSignatures.signature8",
        FT_BOOLEAN, 8, NULL, 0x01,
        NULL, HFILL }},
    { &hf_nbap_PreambleSignatures_signature7,
      { "signature7", "nbap.PreambleSignatures.signature7",
        FT_BOOLEAN, 8, NULL, 0x80,
        NULL, HFILL }},
    { &hf_nbap_PreambleSignatures_signature6,
      { "signature6", "nbap.PreambleSignatures.signature6",
        FT_BOOLEAN, 8, NULL, 0x40,
        NULL, HFILL }},
    { &hf_nbap_PreambleSignatures_signature5,
      { "signature5", "nbap.PreambleSignatures.signature5",
        FT_BOOLEAN, 8, NULL, 0x20,
        NULL, HFILL }},
    { &hf_nbap_PreambleSignatures_signature4,
      { "signature4", "nbap.PreambleSignatures.signature4",
        FT_BOOLEAN, 8, NULL, 0x10,
        NULL, HFILL }},
    { &hf_nbap_PreambleSignatures_signature3,
      { "signature3", "nbap.PreambleSignatures.signature3",
        FT_BOOLEAN, 8, NULL, 0x08,
        NULL, HFILL }},
    { &hf_nbap_PreambleSignatures_signature2,
      { "signature2", "nbap.PreambleSignatures.signature2",
        FT_BOOLEAN, 8, NULL, 0x04,
        NULL, HFILL }},
    { &hf_nbap_PreambleSignatures_signature1,
      { "signature1", "nbap.PreambleSignatures.signature1",
        FT_BOOLEAN, 8, NULL, 0x02,
        NULL, HFILL }},
    { &hf_nbap_PreambleSignatures_signature0,
      { "signature0", "nbap.PreambleSignatures.signature0",
        FT_BOOLEAN, 8, NULL, 0x01,
        NULL, HFILL }},
    { &hf_nbap_RACH_SubChannelNumbers_subCh11,
      { "subCh11", "nbap.RACH.SubChannelNumbers.subCh11",
        FT_BOOLEAN, 8, NULL, 0x80,
        NULL, HFILL }},
    { &hf_nbap_RACH_SubChannelNumbers_subCh10,
      { "subCh10", "nbap.RACH.SubChannelNumbers.subCh10",
        FT_BOOLEAN, 8, NULL, 0x40,
        NULL, HFILL }},
    { &hf_nbap_RACH_SubChannelNumbers_subCh9,
      { "subCh9", "nbap.RACH.SubChannelNumbers.subCh9",
        FT_BOOLEAN, 8, NULL, 0x20,
        NULL, HFILL }},
    { &hf_nbap_RACH_SubChannelNumbers_subCh8,
      { "subCh8", "nbap.RACH.SubChannelNumbers.subCh8",
        FT_BOOLEAN, 8, NULL, 0x10,
        NULL, HFILL }},
    { &hf_nbap_RACH_SubChannelNumbers_subCh7,
      { "subCh7", "nbap.RACH.SubChannelNumbers.subCh7",
        FT_BOOLEAN, 8, NULL, 0x08,
        NULL, HFILL }},
    { &hf_nbap_RACH_SubChannelNumbers_subCh6,
      { "subCh6", "nbap.RACH.SubChannelNumbers.subCh6",
        FT_BOOLEAN, 8, NULL, 0x04,
        NULL, HFILL }},
    { &hf_nbap_RACH_SubChannelNumbers_subCh5,
      { "subCh5", "nbap.RACH.SubChannelNumbers.subCh5",
        FT_BOOLEAN, 8, NULL, 0x02,
        NULL, HFILL }},
    { &hf_nbap_RACH_SubChannelNumbers_subCh4,
      { "subCh4", "nbap.RACH.SubChannelNumbers.subCh4",
        FT_BOOLEAN, 8, NULL, 0x01,
        NULL, HFILL }},
    { &hf_nbap_RACH_SubChannelNumbers_subCh3,
      { "subCh3", "nbap.RACH.SubChannelNumbers.subCh3",
        FT_BOOLEAN, 8, NULL, 0x80,
        NULL, HFILL }},
    { &hf_nbap_RACH_SubChannelNumbers_subCh2,
      { "subCh2", "nbap.RACH.SubChannelNumbers.subCh2",
        FT_BOOLEAN, 8, NULL, 0x40,
        NULL, HFILL }},
    { &hf_nbap_RACH_SubChannelNumbers_subCh1,
      { "subCh1", "nbap.RACH.SubChannelNumbers.subCh1",
        FT_BOOLEAN, 8, NULL, 0x20,
        NULL, HFILL }},
    { &hf_nbap_RACH_SubChannelNumbers_subCh0,
      { "subCh0", "nbap.RACH.SubChannelNumbers.subCh0",
        FT_BOOLEAN, 8, NULL, 0x10,
        NULL, HFILL }},
  };

  /* List of subtrees */
  static gint *ett[] = {
    &ett_nbap,
    &ett_nbap_TransportLayerAddress,
    &ett_nbap_TransportLayerAddress_nsap,
    &ett_nbap_ib_sg_data,
    &ett_nbap_PrivateIE_ID,
    &ett_nbap_ProcedureID,
    &ett_nbap_TransactionID,
    &ett_nbap_ProtocolIE_Container,
    &ett_nbap_ProtocolIE_Field,
    &ett_nbap_ProtocolExtensionContainer,
    &ett_nbap_ProtocolExtensionField,
    &ett_nbap_PrivateIE_Container,
    &ett_nbap_PrivateIE_Field,
    &ett_nbap_ActivationInformation,
    &ett_nbap_ActivationInformationItem,
    &ett_nbap_Additional_EDCH_Setup_Info,
    &ett_nbap_Additional_EDCH_Cell_Information_Setup,
    &ett_nbap_Additional_EDCH_FDD_Setup_Cell_Information,
    &ett_nbap_Additional_EDCH_UL_DPCH_Information_Setup,
    &ett_nbap_Additional_EDCH_F_DPCH_Information,
    &ett_nbap_Additional_EDCH_RL_Specific_Information_To_Setup_List,
    &ett_nbap_Additional_EDCH_RL_Specific_Information_To_Setup_ItemIEs,
    &ett_nbap_Additional_EDCH_Cell_Information_To_Add_List,
    &ett_nbap_Additional_EDCH_Cell_Information_To_Add_ItemIEs,
    &ett_nbap_Additional_EDCH_RL_Specific_Information_To_Add_ItemIEs,
    &ett_nbap_EDCH_Additional_RL_Specific_Information_To_Add_List,
    &ett_nbap_Additional_EDCH_RL_Specific_Information_To_Modify_List,
    &ett_nbap_Additional_EDCH_RL_Specific_Information_To_Modify_ItemIEs,
    &ett_nbap_Additional_EDCH_FDD_Information,
    &ett_nbap_Additional_EDCH_MAC_d_Flows_Specific_Info_List,
    &ett_nbap_Additional_EDCH_MAC_d_Flows_Specific_Info,
    &ett_nbap_Additional_EDCH_Cell_Information_Response_List,
    &ett_nbap_Additional_EDCH_FDD_Information_Response_ItemIEs,
    &ett_nbap_EDCH_Additional_RL_Specific_Information_Response_List,
    &ett_nbap_EDCH_Additional_RL_Specific_Information_Response_ItemIEs,
    &ett_nbap_Additional_EDCH_Cell_Information_Response_RLReconf_List,
    &ett_nbap_Additional_EDCH_FDD_Information_Response_RLReconf_Items,
    &ett_nbap_Additional_Modififed_EDCH_FDD_Information_Response_ItemIEs,
    &ett_nbap_EDCH_Additional_Modified_RL_Specific_Information_Response_List,
    &ett_nbap_EDCH_Additional_Modified_RL_Specific_Information_Response_List_Items,
    &ett_nbap_Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List,
    &ett_nbap_Additional_EDCH_MAC_d_Flows_Specific_Info_Response,
    &ett_nbap_Additional_EDCH_Cell_Information_Response_RL_Add_List,
    &ett_nbap_Additional_EDCH_Cell_Information_Response_RL_Add_ItemIEs,
    &ett_nbap_Additional_EDCH_Cell_Information_ConfigurationChange_List,
    &ett_nbap_Additional_EDCH_ConfigurationChange_Info_ItemIEs,
    &ett_nbap_Additional_EDCH_UL_DPCH_Information_Modify,
    &ett_nbap_Additional_EDCH_Cell_Information_Removal_List,
    &ett_nbap_Additional_EDCH_Cell_Information_Removal_Info_ItemIEs,
    &ett_nbap_Additional_EDCH_FDD_Update_Information,
    &ett_nbap_Additional_EDCH_DL_Control_Channel_Change_Information_List,
    &ett_nbap_Additional_EDCH_DL_Control_Channel_Change_Info_ItemIEs,
    &ett_nbap_AdditionalMeasurementValueList,
    &ett_nbap_AdditionalMeasurementValue,
    &ett_nbap_AdditionalTimeSlotListLCR,
    &ett_nbap_AdditionalTimeSlotLCR,
    &ett_nbap_Active_Pattern_Sequence_Information,
    &ett_nbap_Transmission_Gap_Pattern_Sequence_Status_List,
    &ett_nbap_Transmission_Gap_Pattern_Sequence_Status_List_item,
    &ett_nbap_AllocationRetentionPriority,
    &ett_nbap_Angle_Of_Arrival_Value_LCR,
    &ett_nbap_BCCH_Specific_HSDSCH_RNTI_Information,
    &ett_nbap_BCCH_Specific_HSDSCH_RNTI_InformationLCR,
    &ett_nbap_Best_Cell_Portions_Value,
    &ett_nbap_Best_Cell_Portions_Item,
    &ett_nbap_Best_Cell_Portions_ValueLCR,
    &ett_nbap_Best_Cell_Portions_ItemLCR,
    &ett_nbap_BroadcastCommonTransportBearerIndication,
    &ett_nbap_Cause,
    &ett_nbap_Cell_ERNTI_Status_Information,
    &ett_nbap_Cell_ERNTI_Status_Information_Item,
    &ett_nbap_Vacant_ERNTI,
    &ett_nbap_CellSyncBurstTiming,
    &ett_nbap_CellSyncBurstTimingLCR,
    &ett_nbap_CommonChannelsCapacityConsumptionLaw,
    &ett_nbap_CommonChannelsCapacityConsumptionLaw_item,
    &ett_nbap_Common_EDCH_System_InformationFDD,
    &ett_nbap_Common_E_DCH_UL_DPCH_InfoItem,
    &ett_nbap_Common_E_DCH_EDPCH_InfoItem,
    &ett_nbap_Common_E_DCH_InfoItem,
    &ett_nbap_Common_E_DCH_HSDPCCH_InfoItem,
    &ett_nbap_Common_E_DCH_CQI_Info,
    &ett_nbap_Common_E_DCH_Preamble_Control_InfoItem,
    &ett_nbap_Common_E_DCH_AICH_Information,
    &ett_nbap_Common_E_DCH_FDPCH_InfoItem,
    &ett_nbap_Common_E_DCH_Resource_Combination_InfoList,
    &ett_nbap_Common_E_DCH_Resource_Combination_InfoList_Item,
    &ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoList,
    &ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoList_Item,
    &ett_nbap_Common_E_DCH_MACdFlow_Specific_InfoList,
    &ett_nbap_Common_E_DCH_MACdFlow_Specific_InfoList_Item,
    &ett_nbap_Common_E_DCH_LogicalChannel_InfoList,
    &ett_nbap_Common_E_DCH_LogicalChannel_InfoList_Item,
    &ett_nbap_Common_EDCH_System_Information_ResponseFDD,
    &ett_nbap_E_RNTI_List,
    &ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseList,
    &ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseList_Item,
    &ett_nbap_Common_MACFlows_to_DeleteFDD,
    &ett_nbap_Common_MACFlows_to_DeleteFDD_Item,
    &ett_nbap_CommonMACFlow_Specific_InfoList,
    &ett_nbap_CommonMACFlow_Specific_InfoItem,
    &ett_nbap_CommonMACFlow_Specific_InfoList_Response,
    &ett_nbap_CommonMACFlow_Specific_InfoItem_Response,
    &ett_nbap_Common_MACFlow_PriorityQueue_Information,
    &ett_nbap_Common_MACFlow_PriorityQueue_Item,
    &ett_nbap_CommonMeasurementAccuracy,
    &ett_nbap_CommonMeasurementValue,
    &ett_nbap_CommonMeasurementValueInformation,
    &ett_nbap_CommonMeasurementAvailable,
    &ett_nbap_Common_PhysicalChannel_Status_Information,
    &ett_nbap_Common_PhysicalChannel_Status_Information768,
    &ett_nbap_CommonTransportChannel_InformationResponse,
    &ett_nbap_Common_TransportChannel_Status_Information,
    &ett_nbap_ContinuousPacketConnectivityDTX_DRX_Information,
    &ett_nbap_ContinuousPacketConnectivityDTX_DRX_Information_to_Modify,
    &ett_nbap_ContinuousPacketConnectivityHS_SCCH_less_Information,
    &ett_nbap_ContinuousPacketConnectivityHS_SCCH_less_InformationItem,
    &ett_nbap_ContinuousPacketConnectivityHS_SCCH_less_Information_Response,
    &ett_nbap_CPC_Information,
    &ett_nbap_CriticalityDiagnostics,
    &ett_nbap_CriticalityDiagnostics_IE_List,
    &ett_nbap_CriticalityDiagnostics_IE_List_item,
    &ett_nbap_Common_EDCH_System_InformationLCR,
    &ett_nbap_Common_E_PUCH_InformationLCR,
    &ett_nbap_PRXdes_base_perURAFCN,
    &ett_nbap_PRXdes_base_Item,
    &ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoListLCR,
    &ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoList_ItemLCR,
    &ett_nbap_Common_E_DCH_MACdFlow_Specific_InfoListLCR,
    &ett_nbap_Common_E_DCH_MACdFlow_Specific_InfoList_ItemLCR,
    &ett_nbap_Common_EDCH_System_Information_ResponseLCR,
    &ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR,
    &ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_Ext,
    &ett_nbap_Ul_common_E_DCH_MACflow_Specific_InfoResponseList_ItemLCR,
    &ett_nbap_Common_E_AGCH_ListLCR,
    &ett_nbap_Common_E_AGCH_ItemLCR,
    &ett_nbap_Common_E_HICH_ListLCR,
    &ett_nbap_Common_E_HICH_ItemLCR,
    &ett_nbap_Common_E_RNTI_Info_LCR,
    &ett_nbap_Common_E_RNTI_Info_ItemLCR,
    &ett_nbap_Common_MACFlows_to_DeleteLCR,
    &ett_nbap_Common_MACFlows_to_DeleteLCR_Item,
    &ett_nbap_CommonMACFlow_Specific_InfoListLCR,
    &ett_nbap_CommonMACFlow_Specific_InfoItemLCR,
    &ett_nbap_Common_H_RNTI_InformationLCR,
    &ett_nbap_Common_H_RNTI_InfoItemLCR,
    &ett_nbap_Sync_InformationLCR,
    &ett_nbap_CommonMACFlow_Specific_InfoList_ResponseLCR,
    &ett_nbap_CommonMACFlow_Specific_InfoItem_ResponseLCR,
    &ett_nbap_CPC_InformationLCR,
    &ett_nbap_ContinuousPacketConnectivity_DRX_InformationLCR,
    &ett_nbap_HS_SCCH_DRX_Information_LCR,
    &ett_nbap_E_AGCH_DRX_Information_LCR,
    &ett_nbap_E_AGCH_DRX_Parameters,
    &ett_nbap_ContinuousPacketConnectivity_DRX_Information_to_Modify_LCR,
    &ett_nbap_DRX_Information_to_Modify_LCR,
    &ett_nbap_DRX_Information_to_Modify_Items_LCR,
    &ett_nbap_ContinuousPacketConnectivity_DRX_Information_ResponseLCR,
    &ett_nbap_HS_SCCH_DRX_Information_ResponseLCR,
    &ett_nbap_E_AGCH_DRX_Information_ResponseLCR,
    &ett_nbap_E_AGCH_DRX_Parameters_Response,
    &ett_nbap_DCH_FDD_Information,
    &ett_nbap_DCH_FDD_InformationItem,
    &ett_nbap_DCH_Specific_FDD_InformationList,
    &ett_nbap_DCH_Specific_FDD_Item,
    &ett_nbap_DCH_InformationResponse,
    &ett_nbap_DCH_InformationResponseItem,
    &ett_nbap_DCH_MeasurementOccasion_Information,
    &ett_nbap_DchMeasurementOccasionInformation_Item,
    &ett_nbap_Measurement_Occasion_Pattern_Sequence_parameters,
    &ett_nbap_DCH_TDD_Information,
    &ett_nbap_DCH_TDD_InformationItem,
    &ett_nbap_DCH_Specific_TDD_InformationList,
    &ett_nbap_DCH_Specific_TDD_Item,
    &ett_nbap_FDD_DCHs_to_Modify,
    &ett_nbap_FDD_DCHs_to_ModifyItem,
    &ett_nbap_DCH_ModifySpecificInformation_FDD,
    &ett_nbap_DCH_ModifySpecificItem_FDD,
    &ett_nbap_TDD_DCHs_to_Modify,
    &ett_nbap_DCH_ModifyItem_TDD,
    &ett_nbap_DCH_ModifySpecificInformation_TDD,
    &ett_nbap_DCH_ModifySpecificItem_TDD,
    &ett_nbap_DedicatedChannelsCapacityConsumptionLaw,
    &ett_nbap_DedicatedChannelsCapacityConsumptionLaw_item,
    &ett_nbap_DedicatedMeasurementValue,
    &ett_nbap_DedicatedMeasurementValueInformation,
    &ett_nbap_DedicatedMeasurementAvailable,
    &ett_nbap_DelayedActivation,
    &ett_nbap_DelayedActivationUpdate,
    &ett_nbap_Activate_Info,
    &ett_nbap_Deactivate_Info,
    &ett_nbap_Execution_Type,
    &ett_nbap_DGANSSCorrections,
    &ett_nbap_DGANSS_Corrections_Req,
    &ett_nbap_DGANSS_Information,
    &ett_nbap_DGANSS_InformationItem,
    &ett_nbap_DGANSS_SignalInformation,
    &ett_nbap_DGANSS_SignalInformationItem,
    &ett_nbap_DGANSSThreshold,
    &ett_nbap_DGNSS_ValidityPeriod,
    &ett_nbap_DGPSCorrections,
    &ett_nbap_DGPSThresholds,
    &ett_nbap_DL_Timeslot_Information,
    &ett_nbap_DL_Timeslot_InformationItem,
    &ett_nbap_DL_TimeslotLCR_Information,
    &ett_nbap_DL_TimeslotLCR_InformationItem,
    &ett_nbap_DL_Timeslot768_Information,
    &ett_nbap_DL_Timeslot768_InformationItem,
    &ett_nbap_DL_PowerBalancing_Information,
    &ett_nbap_DL_ReferencePowerInformationList,
    &ett_nbap_DL_ReferencePowerInformationItem,
    &ett_nbap_DL_TimeslotISCPInfo,
    &ett_nbap_DL_TimeslotISCPInfoItem,
    &ett_nbap_DL_TimeslotISCPInfoLCR,
    &ett_nbap_DL_TimeslotISCPInfoItemLCR,
    &ett_nbap_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst,
    &ett_nbap_DL_HS_PDSCH_Timeslot_InformationItem_LCR_PSCH_ReconfRqst,
    &ett_nbap_MaxHSDSCH_HSSCCH_Power_per_CELLPORTION,
    &ett_nbap_MaxHSDSCH_HSSCCH_Power_per_CELLPORTION_Item,
    &ett_nbap_DL_HS_PDSCH_Codelist_LCR_PSCH_ReconfRqst,
    &ett_nbap_DRX_Information,
    &ett_nbap_DRX_Information_to_Modify,
    &ett_nbap_DRX_Information_to_Modify_Items,
    &ett_nbap_DSCH_InformationResponse,
    &ett_nbap_DSCH_InformationResponseItem,
    &ett_nbap_DSCH_TDD_Information,
    &ett_nbap_DSCH_TDD_InformationItem,
    &ett_nbap_DTX_Cycle_2ms_Items,
    &ett_nbap_DTX_Cycle_2ms_to_Modify_Items,
    &ett_nbap_DTX_Cycle_10ms_Items,
    &ett_nbap_DTX_Cycle_10ms_to_Modify_Items,
    &ett_nbap_DTX_Information,
    &ett_nbap_DTX_Information_to_Modify,
    &ett_nbap_DTX_Information_to_Modify_Items,
    &ett_nbap_Dual_Band_Capability_Info,
    &ett_nbap_E_AGCH_FDD_Code_Information,
    &ett_nbap_E_AGCH_FDD_Code_List,
    &ett_nbap_E_DCHCapacityConsumptionLaw,
    &ett_nbap_E_DCH_TDD_CapacityConsumptionLaw,
    &ett_nbap_E_DCH_SF_allocation,
    &ett_nbap_E_DCH_SF_allocation_item,
    &ett_nbap_E_DCH_FDD_DL_Control_Channel_Information,
    &ett_nbap_E_DCH_FDD_Information,
    &ett_nbap_E_DCH_FDD_Information_Response,
    &ett_nbap_E_DCH_FDD_Information_to_Modify,
    &ett_nbap_E_DCH_FDD_Update_Information,
    &ett_nbap_E_DCH_MACdFlow_Specific_UpdateInformation,
    &ett_nbap_E_DCH_MACdFlow_Specific_UpdateInformation_Item,
    &ett_nbap_E_DCH_DL_Control_Channel_Change_Information,
    &ett_nbap_E_DCH_DL_Control_Channel_Change_Information_Item,
    &ett_nbap_E_DCH_DL_Control_Channel_Grant_Information,
    &ett_nbap_E_DCH_DL_Control_Channel_Grant_Information_Item,
    &ett_nbap_E_DCH_Grant_Type_Information,
    &ett_nbap_E_DCH_LogicalChannelInformation,
    &ett_nbap_E_DCH_LogicalChannelInformationItem,
    &ett_nbap_E_DCH_MACdPDU_SizeList,
    &ett_nbap_E_DCH_MACdPDU_SizeListItem,
    &ett_nbap_E_DCH_LogicalChannelToModify,
    &ett_nbap_E_DCH_LogicalChannelToModifyItem,
    &ett_nbap_E_DCH_MACdPDU_SizeToModifyList,
    &ett_nbap_E_DCH_LogicalChannelToDelete,
    &ett_nbap_E_DCH_LogicalChannelToDeleteItem,
    &ett_nbap_E_DCH_MACdFlows_Information,
    &ett_nbap_E_DCH_MACdFlow_Specific_InfoList,
    &ett_nbap_E_DCH_MACdFlow_Specific_InfoItem,
    &ett_nbap_E_DCH_MACdFlow_Specific_InformationResp,
    &ett_nbap_E_DCH_MACdFlow_Specific_InformationResp_Item,
    &ett_nbap_E_DCH_MACdFlow_Specific_InfoList_to_Modify,
    &ett_nbap_E_DCH_MACdFlow_Specific_InfoItem_to_Modify,
    &ett_nbap_E_DCH_MACdFlows_to_Delete,
    &ett_nbap_E_DCH_MACdFlow_to_Delete_Item,
    &ett_nbap_E_DCH_Non_Scheduled_Transmission_Grant_Items,
    &ett_nbap_E_DCH_Serving_Cell_Change_Info_Response,
    &ett_nbap_E_DCH_serving_cell_choice,
    &ett_nbap_E_DCH_serving_cell_change_successful,
    &ett_nbap_E_DCH_RL_InformationList_Rsp,
    &ett_nbap_E_DCH_RL_InformationList_Rsp_Item,
    &ett_nbap_E_DCH_serving_cell_change_unsuccessful,
    &ett_nbap_EDCH_RACH_Report_Value,
    &ett_nbap_EDCH_RACH_Report_Value_item,
    &ett_nbap_E_DCH_TTI_Length,
    &ett_nbap_E_DCH_TTI_Length_to_Modify,
    &ett_nbap_E_DCH_MACdFlows_to_DeleteLCR,
    &ett_nbap_E_DCH_MACdFlow_to_Delete_ItemLCR,
    &ett_nbap_Enhanced_UE_DRX_InformationLCR,
    &ett_nbap_E_RGCH_E_HICH_FDD_Code_Information,
    &ett_nbap_E_RGCH_E_HICH_FDD_Code_List,
    &ett_nbap_E_TFCI_Boost_Information,
    &ett_nbap_E_TFCS_Information,
    &ett_nbap_E_DCHProvidedBitRate,
    &ett_nbap_E_DCHProvidedBitRate_Item,
    &ett_nbap_E_DCHProvidedBitRateValueInformation_For_CellPortion,
    &ett_nbap_E_DCHProvidedBitRateValueInformation_For_CellPortion_Item,
    &ett_nbap_E_DCH_Information,
    &ett_nbap_E_PUCH_Information,
    &ett_nbap_E_TFCS_Information_TDD,
    &ett_nbap_E_DCH_QPSK_RefBetaInfo,
    &ett_nbap_E_DCH_sixteenQAM_RefBetaInfo,
    &ett_nbap_E_DCH_RefBeta_Item,
    &ett_nbap_E_DCH_MACdFlows_Information_TDD,
    &ett_nbap_E_DCH_MACdFlow_InfoTDDItem,
    &ett_nbap_E_DCH_Non_Scheduled_Grant_Info,
    &ett_nbap_E_DCH_TDD_Information,
    &ett_nbap_E_DCH_Information_Response,
    &ett_nbap_Scheduled_E_HICH_Specific_Information_ResponseLCRTDD,
    &ett_nbap_Scheduled_E_HICH_Specific_InformationItem_ResponseLCRTDD,
    &ett_nbap_E_DCH_TDD_MACdFlow_Specific_InformationResp,
    &ett_nbap_E_DCH_TDD_MACdFlow_Specific_InformationResp_Item,
    &ett_nbap_E_AGCH_Specific_InformationRespListTDD,
    &ett_nbap_E_AGCH_Specific_InformationResp_ItemTDD,
    &ett_nbap_E_DCH_Information_Reconfig,
    &ett_nbap_E_DCH_TDD_Information_to_Modify,
    &ett_nbap_E_DCH_TDD_Information_to_Modify_List,
    &ett_nbap_E_DCH_MACdFlow_ModifyTDDItem,
    &ett_nbap_E_DCH_768_Information,
    &ett_nbap_E_DCH_Non_Scheduled_Grant_Info768,
    &ett_nbap_E_DCH_TDD_Information768,
    &ett_nbap_E_DCH_768_Information_Reconfig,
    &ett_nbap_E_DCH_LCR_Information,
    &ett_nbap_E_PUCH_LCR_Information,
    &ett_nbap_E_DCH_Non_Scheduled_Grant_LCR_Info,
    &ett_nbap_E_HICH_LCR_Information,
    &ett_nbap_E_DCH_LCRTDD_Information,
    &ett_nbap_E_DCH_LCR_Information_Reconfig,
    &ett_nbap_Enhanced_UE_DRX_InformationFDD,
    &ett_nbap_E_DCH_Semi_PersistentScheduling_Information_LCR,
    &ett_nbap_E_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR,
    &ett_nbap_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR,
    &ett_nbap_FDD_DL_CodeInformation,
    &ett_nbap_FDD_DL_CodeInformationItem,
    &ett_nbap_GANSS_AddClockModels,
    &ett_nbap_GANSS_AddOrbitModels,
    &ett_nbap_GANSS_Additional_Ionospheric_Model,
    &ett_nbap_GANSS_Additional_Navigation_Models,
    &ett_nbap_GANSS_Additional_Time_Models,
    &ett_nbap_GANSS_Additional_UTC_Models,
    &ett_nbap_GANSS_Almanac,
    &ett_nbap_GANSS_AlmanacModel,
    &ett_nbap_GANSS_ALM_ECEFsbasAlmanacSet,
    &ett_nbap_GANSS_ALM_GlonassAlmanacSet,
    &ett_nbap_GANSS_ALM_MidiAlmanacSet,
    &ett_nbap_GANSS_ALM_NAVKeplerianSet,
    &ett_nbap_GANSS_ALM_ReducedKeplerianSet,
    &ett_nbap_GANSS_Auxiliary_Information,
    &ett_nbap_GANSS_AuxInfoGANSS_ID1,
    &ett_nbap_GANSS_AuxInfoGANSS_ID1_element,
    &ett_nbap_GANSS_AuxInfoGANSS_ID3,
    &ett_nbap_GANSS_AuxInfoGANSS_ID3_element,
    &ett_nbap_GANSS_Clock_Model,
    &ett_nbap_GANSS_CNAVclockModel,
    &ett_nbap_GANSS_Common_Data,
    &ett_nbap_GANSS_CommonDataInfoReq,
    &ett_nbap_GANSS_Data_Bit_Assistance,
    &ett_nbap_GANSS_DataBitAssistanceList,
    &ett_nbap_GANSS_DataBitAssistanceItem,
    &ett_nbap_GANSS_DataBitAssistanceSgnList,
    &ett_nbap_GANSS_DataBitAssistanceSgnItem,
    &ett_nbap_GANSS_Data_Bit_Assistance_ReqItem,
    &ett_nbap_GANSS_Data_Bit_Assistance_ReqList,
    &ett_nbap_T_ganss_SatelliteInfo,
    &ett_nbap_GANSS_DeltaUT1,
    &ett_nbap_GANSS_Earth_Orientation_Parameters,
    &ett_nbap_GANSS_GenericDataInfoReqList,
    &ett_nbap_GANSS_GenericDataInfoReqItem,
    &ett_nbap_GANSS_Generic_Data,
    &ett_nbap_GANSS_Generic_DataItem,
    &ett_nbap_GANSS_GLONASSclockModel,
    &ett_nbap_GANSS_Information,
    &ett_nbap_GANSS_Ionospheric_Model,
    &ett_nbap_GANSS_IonosphereRegionalStormFlags,
    &ett_nbap_GANSS_KeplerianParametersAlm,
    &ett_nbap_GANSS_KeplerianParametersOrb,
    &ett_nbap_GANSS_NAVclockModel,
    &ett_nbap_GANSS_Navigation_Model_And_Time_Recovery,
    &ett_nbap_GANSS_NavModel_CNAVKeplerianSet,
    &ett_nbap_GANSS_NavModel_GLONASSecef,
    &ett_nbap_GANSS_NavModel_NAVKeplerianSet,
    &ett_nbap_GANSS_NavModel_SBASecef,
    &ett_nbap_GANSS_Orbit_Model,
    &ett_nbap_GANSS_Real_Time_Integrity,
    &ett_nbap_GANSS_RealTimeInformationItem,
    &ett_nbap_GANSS_RX_Pos,
    &ett_nbap_GANSS_SatelliteClockModelItem,
    &ett_nbap_GANSS_SatelliteInformationKP,
    &ett_nbap_GANSS_SatelliteInformationKPItem,
    &ett_nbap_Ganss_Sat_Info_AddNavList,
    &ett_nbap_Ganss_Sat_Info_AddNavList_item,
    &ett_nbap_GANSS_SAT_Info_Almanac_GLOkpList,
    &ett_nbap_GANSS_SAT_Info_Almanac_GLOkp,
    &ett_nbap_GANSS_SAT_Info_Almanac_MIDIkpList,
    &ett_nbap_GANSS_SAT_Info_Almanac_MIDIkp,
    &ett_nbap_GANSS_SAT_Info_Almanac_NAVkpList,
    &ett_nbap_GANSS_SAT_Info_Almanac_NAVkp,
    &ett_nbap_GANSS_SAT_Info_Almanac_REDkpList,
    &ett_nbap_GANSS_SAT_Info_Almanac_REDkp,
    &ett_nbap_GANSS_SAT_Info_Almanac_SBASecefList,
    &ett_nbap_GANSS_SAT_Info_Almanac_SBASecef,
    &ett_nbap_GANSS_Sat_Info_Nav,
    &ett_nbap_GANSS_Sat_Info_Nav_item,
    &ett_nbap_GANSS_SBASclockModel,
    &ett_nbap_GANSS_Time_Model,
    &ett_nbap_GANSS_Transmission_Time,
    &ett_nbap_GANSS_UTC_Model,
    &ett_nbap_GANSS_UTCmodelSet1,
    &ett_nbap_GANSS_UTCmodelSet2,
    &ett_nbap_GANSS_UTCmodelSet3,
    &ett_nbap_GPS_Almanac,
    &ett_nbap_GPS_Ionospheric_Model,
    &ett_nbap_GPS_Information,
    &ett_nbap_GPS_RealTime_Integrity,
    &ett_nbap_GPSBadSat_Info_RealTime_Integrity,
    &ett_nbap_GPS_NavigationModel_and_TimeRecovery,
    &ett_nbap_GPS_NavandRecovery_Item,
    &ett_nbap_GPS_RX_POS,
    &ett_nbap_GPS_UTC_Model,
    &ett_nbap_HARQ_MemoryPartitioning,
    &ett_nbap_HARQ_MemoryPartitioning_Implicit,
    &ett_nbap_HARQ_MemoryPartitioning_Explicit,
    &ett_nbap_HARQ_MemoryPartitioningList,
    &ett_nbap_HARQ_MemoryPartitioningInfoExtForMIMO,
    &ett_nbap_HARQ_MemoryPartitioningItem,
    &ett_nbap_HS_DSCHProvidedBitRate,
    &ett_nbap_HS_DSCHProvidedBitRate_Item,
    &ett_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortion,
    &ett_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortion_Item,
    &ett_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR,
    &ett_nbap_HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR_Item,
    &ett_nbap_HS_DSCHRequiredPower,
    &ett_nbap_HS_DSCHRequiredPower_Item,
    &ett_nbap_HS_DSCHRequiredPowerPerUEInformation,
    &ett_nbap_HS_DSCHRequiredPowerPerUEInformation_Item,
    &ett_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortion,
    &ett_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortion_Item,
    &ett_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR,
    &ett_nbap_HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR_Item,
    &ett_nbap_HSDPA_Associated_PICH_Information,
    &ett_nbap_HSDPA_PICH_Shared_with_PCH,
    &ett_nbap_HSDPA_PICH_notShared_with_PCH,
    &ett_nbap_HSDSCH_Common_System_InformationFDD,
    &ett_nbap_HSDSCH_Common_System_Information_ResponseFDD,
    &ett_nbap_HSDSCH_Common_Information,
    &ett_nbap_HSDSCH_FDD_Information,
    &ett_nbap_HSDSCH_TDD_Information,
    &ett_nbap_HSDSCH_Information_to_Modify,
    &ett_nbap_HSDSCH_MACdFlow_Specific_InfoList_to_Modify,
    &ett_nbap_HSDSCH_MACdFlow_Specific_InfoItem_to_Modify,
    &ett_nbap_HSDSCH_Information_to_Modify_Unsynchronised,
    &ett_nbap_HSDSCH_FDD_Information_Response,
    &ett_nbap_HS_DSCH_FDD_Secondary_Serving_Information,
    &ett_nbap_HS_DSCH_FDD_Secondary_Serving_Information_Response,
    &ett_nbap_HS_DSCH_Secondary_Serving_Information_To_Modify,
    &ett_nbap_HS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised,
    &ett_nbap_HS_DSCH_FDD_Secondary_Serving_Update_Information,
    &ett_nbap_HS_DSCH_Secondary_Serving_Cell_Change_Information_Response,
    &ett_nbap_HS_DSCH_Secondary_Serving_cell_change_choice,
    &ett_nbap_HS_Secondary_Serving_cell_change_successful,
    &ett_nbap_HS_Secondary_Serving_cell_change_unsuccessful,
    &ett_nbap_HSDSCH_Paging_System_InformationFDD,
    &ett_nbap_HSDSCH_Paging_System_Information_ResponseFDD,
    &ett_nbap_HSDSCH_Paging_System_Information_ResponseList,
    &ett_nbap_HSDSCH_TDD_Information_Response,
    &ett_nbap_HSDSCH_MACdFlow_Specific_InformationResp,
    &ett_nbap_HSDSCH_MACdFlow_Specific_InformationResp_Item,
    &ett_nbap_HSDSCH_MACdFlows_Information,
    &ett_nbap_HSDSCH_MACdFlow_Specific_InfoList,
    &ett_nbap_HSDSCH_MACdFlow_Specific_InfoItem,
    &ett_nbap_HSDSCH_MACdFlows_to_Delete,
    &ett_nbap_HSDSCH_MACdFlows_to_Delete_Item,
    &ett_nbap_HSDSCH_Initial_Capacity_Allocation,
    &ett_nbap_HSDSCH_Initial_Capacity_AllocationItem,
    &ett_nbap_HSDSCH_PreconfigurationInfo,
    &ett_nbap_Additional_EDCH_Preconfiguration_Information,
    &ett_nbap_Additional_EDCH_Preconfiguration_Information_ItemIEs,
    &ett_nbap_HSDSCH_PreconfigurationSetup,
    &ett_nbap_HS_SCCH_PreconfiguredCodes,
    &ett_nbap_HS_SCCH_PreconfiguredCodesItem,
    &ett_nbap_HSSCCH_Specific_InformationRespListFDD,
    &ett_nbap_HSSCCH_Codes,
    &ett_nbap_HSSCCH_Specific_InformationRespListTDD,
    &ett_nbap_HSSCCH_Specific_InformationRespItemTDD,
    &ett_nbap_HSSCCH_Specific_InformationRespListTDDLCR,
    &ett_nbap_HSSCCH_Specific_InformationRespItemTDDLCR,
    &ett_nbap_HSSCCH_Specific_InformationRespListTDD768,
    &ett_nbap_HSSCCH_Specific_InformationRespItemTDD768,
    &ett_nbap_HSSICH_Info,
    &ett_nbap_HSSICH_InfoLCR,
    &ett_nbap_HSSICH_Info768,
    &ett_nbap_HS_SICH_Reception_Quality_Value,
    &ett_nbap_HS_PDSCH_FDD_Code_Information,
    &ett_nbap_HS_SCCH_FDD_Code_Information,
    &ett_nbap_HS_SCCH_FDD_Code_List,
    &ett_nbap_HS_DSCH_Serving_Cell_Change_Info,
    &ett_nbap_HS_DSCH_Serving_Cell_Change_Info_Response,
    &ett_nbap_HS_DSCH_serving_cell_choice,
    &ett_nbap_HS_serving_cell_change_successful,
    &ett_nbap_HS_serving_cell_change_unsuccessful,
    &ett_nbap_HSDSCH_FDD_Update_Information,
    &ett_nbap_HSDSCH_TDD_Update_Information,
    &ett_nbap_HSDPA_Associated_PICH_InformationLCR,
    &ett_nbap_HSDPA_PICH_notShared_with_PCHLCR,
    &ett_nbap_HSDSCH_Common_System_InformationLCR,
    &ett_nbap_HSDSCH_Common_System_Information_ResponseLCR,
    &ett_nbap_HSDSCH_Common_InformationLCR,
    &ett_nbap_HSDSCH_Paging_System_InformationLCR,
    &ett_nbap_HSDSCH_Paging_System_Information_ResponseLCR,
    &ett_nbap_HSDSCH_Paging_System_Information_ResponseListLCR,
    &ett_nbap_HSSCCH_Specific_InformationRespListLCR,
    &ett_nbap_HSSCCH_Specific_InformationRespItemLCR,
    &ett_nbap_HS_DSCH_Semi_PersistentScheduling_Information_LCR,
    &ett_nbap_Transport_Block_Size_List_LCR,
    &ett_nbap_Transport_Block_Size_Item_LCR,
    &ett_nbap_Repetition_Period_List_LCR,
    &ett_nbap_Repetition_Period_Item_LCR,
    &ett_nbap_HS_DSCH_SPS_Operation_Indicator,
    &ett_nbap_HS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR,
    &ett_nbap_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR,
    &ett_nbap_HS_SICH_InformationList_for_HS_DSCH_SPS,
    &ett_nbap_HS_SICH_InformationItem_for_HS_DSCH_SPS,
    &ett_nbap_HS_SICH_Type,
    &ett_nbap_HS_SCCH_Associated_HS_SICH,
    &ett_nbap_Non_HS_SCCH_Associated_HS_SICH,
    &ett_nbap_Initial_HS_PDSCH_SPS_Resource,
    &ett_nbap_Add_To_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst,
    &ett_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationList,
    &ett_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext,
    &ett_nbap_Non_HS_SCCH_Associated_HS_SICH_InformationItem,
    &ett_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst,
    &ett_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList,
    &ett_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext,
    &ett_nbap_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationItem,
    &ett_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst,
    &ett_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext,
    &ett_nbap_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqstItem,
    &ett_nbap_MIMO_ReferenceSignal_InformationListLCR,
    &ett_nbap_HSSICH_ReferenceSignal_InformationLCR,
    &ett_nbap_HSSICH_ReferenceSignal_InformationModifyLCR,
    &ett_nbap_IMB_Parameters,
    &ett_nbap_InformationReportCharacteristics,
    &ett_nbap_InformationReportCharacteristicsType_ReportPeriodicity,
    &ett_nbap_InformationReportCharacteristicsType_OnModification,
    &ett_nbap_InformationThresholds,
    &ett_nbap_InformationType,
    &ett_nbap_IPDL_FDD_Parameters,
    &ett_nbap_IPDL_TDD_Parameters,
    &ett_nbap_IPDL_TDD_Parameters_LCR,
    &ett_nbap_IPMulticastIndication,
    &ett_nbap_BurstModeParams,
    &ett_nbap_IdleIntervalInformation,
    &ett_nbap_LCRTDD_Uplink_Physical_Channel_Capability,
    &ett_nbap_MACdPDU_Size_Indexlist,
    &ett_nbap_MACdPDU_Size_IndexItem,
    &ett_nbap_MACdPDU_Size_Indexlist_to_Modify,
    &ett_nbap_MACdPDU_Size_IndexItem_to_Modify,
    &ett_nbap_Max_TB_Sizes,
    &ett_nbap_MessageStructure,
    &ett_nbap_MessageStructure_item,
    &ett_nbap_MidambleShiftAndBurstType,
    &ett_nbap_Type1,
    &ett_nbap_MidambleAllocationMode1,
    &ett_nbap_Type2,
    &ett_nbap_MidambleAllocationMode2,
    &ett_nbap_Type3,
    &ett_nbap_MidambleAllocationMode3,
    &ett_nbap_MidambleShiftLCR,
    &ett_nbap_MidambleShiftAndBurstType768,
    &ett_nbap_Type7681,
    &ett_nbap_MidambleAllocationMode7681,
    &ett_nbap_Type7682,
    &ett_nbap_MidambleAllocationMode7682,
    &ett_nbap_Type7683,
    &ett_nbap_MidambleAllocationMode7683,
    &ett_nbap_MIMO_PilotConfiguration,
    &ett_nbap_MIMO_PilotConfigurationExtension,
    &ett_nbap_ModifyPriorityQueue,
    &ett_nbap_ModulationPO_MBSFN,
    &ett_nbap_MultipleFreq_HARQ_MemoryPartitioning_InformationList,
    &ett_nbap_MultipleFreq_HARQ_MemoryPartitioning_InformationItem,
    &ett_nbap_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR,
    &ett_nbap_MultipleFreq_HSPDSCH_InformationItem_ResponseTDDLCR,
    &ett_nbap_Multi_Cell_Capability_Info,
    &ett_nbap_Multicell_EDCH_InformationItemIEs,
    &ett_nbap_Multicell_EDCH_RL_Specific_InformationItemIEs,
    &ett_nbap_NeighbouringCellMeasurementInformation,
    &ett_nbap_NeighbouringCellMeasurementInformation_item,
    &ett_nbap_NeighbouringFDDCellMeasurementInformation,
    &ett_nbap_NeighbouringTDDCellMeasurementInformation,
    &ett_nbap_NeighbouringTDDCellMeasurementInformationLCR,
    &ett_nbap_NeighbouringTDDCellMeasurementInformation768,
    &ett_nbap_NI_Information,
    &ett_nbap_NormalAndDiversityPrimaryCPICHContainer,
    &ett_nbap_Paging_MACFlows_to_DeleteFDD,
    &ett_nbap_Paging_MACFlows_to_DeleteFDD_Item,
    &ett_nbap_Paging_MACFlow_Specific_Information,
    &ett_nbap_Paging_MAC_Flow_Specific_Information_Item,
    &ett_nbap_Paging_MACFlow_PriorityQueue_Information,
    &ett_nbap_Paging_MACFlow_PriorityQueue_Item,
    &ett_nbap_PLCCHinformation,
    &ett_nbap_Possible_Secondary_Serving_Cell_List,
    &ett_nbap_Possible_Secondary_Serving_Cell,
    &ett_nbap_PreambleSignatures,
    &ett_nbap_PrimaryAndSecondaryCPICHContainer,
    &ett_nbap_Priority_Queue_Information_for_Enhanced_FACH_PCH,
    &ett_nbap_PriorityQueue_InfoList,
    &ett_nbap_PriorityQueue_InfoItem,
    &ett_nbap_PriorityQueue_InfoList_to_Modify,
    &ett_nbap_PriorityQueue_InfoItem_to_Add,
    &ett_nbap_PriorityQueue_InfoItem_to_Modify,
    &ett_nbap_PriorityQueue_InfoList_to_Modify_Unsynchronised,
    &ett_nbap_PriorityQueue_InfoItem_to_Modify_Unsynchronised,
    &ett_nbap_UE_Selected_MBMS_Service_Information,
    &ett_nbap_Selected_MBMS_Service,
    &ett_nbap_Selected_MBMS_Service_List,
    &ett_nbap_Selected_MBMS_Service_Item,
    &ett_nbap_Selected_MBMS_Service_TimeSlot_Information_LCR,
    &ett_nbap_MBMS_Service_TDM_Information,
    &ett_nbap_Paging_MACFlow_Specific_InformationLCR,
    &ett_nbap_Paging_MAC_Flow_Specific_Information_ItemLCR,
    &ett_nbap_Paging_MACFlows_to_DeleteLCR,
    &ett_nbap_Paging_MACFlows_to_DeleteLCR_Item,
    &ett_nbap_RACH_SubChannelNumbers,
    &ett_nbap_RL_Specific_DCH_Info,
    &ett_nbap_RL_Specific_DCH_Info_Item,
    &ett_nbap_RL_Specific_E_DCH_Info,
    &ett_nbap_RL_Specific_E_DCH_Information,
    &ett_nbap_RL_Specific_E_DCH_Information_Item,
    &ett_nbap_Reference_E_TFCI_Information,
    &ett_nbap_Reference_E_TFCI_Information_Item,
    &ett_nbap_ReportCharacteristics,
    &ett_nbap_ReportCharacteristicsType_EventA,
    &ett_nbap_ReportCharacteristicsType_EventB,
    &ett_nbap_ReportCharacteristicsType_EventC,
    &ett_nbap_ReportCharacteristicsType_EventD,
    &ett_nbap_ReportCharacteristicsType_EventE,
    &ett_nbap_ReportCharacteristicsType_EventF,
    &ett_nbap_ReportCharacteristicsType_OnModification,
    &ett_nbap_ReportCharacteristicsType_MeasurementIncreaseDecreaseThreshold,
    &ett_nbap_EDCH_RACH_Report_IncrDecrThres,
    &ett_nbap_ReportCharacteristicsType_MeasurementThreshold,
    &ett_nbap_EDCH_RACH_Report_ThresholdInformation,
    &ett_nbap_ReportCharacteristicsType_ScaledMeasurementChangeTime,
    &ett_nbap_ReportCharacteristicsType_ScaledMeasurementHysteresisTime,
    &ett_nbap_ReportCharacteristicsType_ReportPeriodicity,
    &ett_nbap_Received_total_wide_band_power_For_CellPortion_Value,
    &ett_nbap_Received_total_wide_band_power_For_CellPortion_Value_Item,
    &ett_nbap_Received_total_wide_band_power_For_CellPortion_ValueLCR,
    &ett_nbap_Received_total_wide_band_power_For_CellPortion_ValueLCR_Item,
    &ett_nbap_Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value,
    &ett_nbap_Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value_Item,
    &ett_nbap_Received_Scheduled_EDCH_Power_Share_Value,
    &ett_nbap_RequestedDataValueInformation,
    &ett_nbap_InformationAvailable,
    &ett_nbap_RequestedDataValue,
    &ett_nbap_SAT_Info_Almanac,
    &ett_nbap_SAT_Info_Almanac_Item,
    &ett_nbap_SAT_Info_Almanac_ExtList,
    &ett_nbap_SAT_Info_Almanac_ExtItem,
    &ett_nbap_SAT_Info_DGPSCorrections,
    &ett_nbap_SAT_Info_DGPSCorrections_Item,
    &ett_nbap_SATInfo_RealTime_Integrity,
    &ett_nbap_SAT_Info_RealTime_Integrity_Item,
    &ett_nbap_SecondaryServingCells,
    &ett_nbap_SecondaryServingCellsItem,
    &ett_nbap_Secondary_CPICH_Information_Change,
    &ett_nbap_Serving_E_DCH_RL_ID,
    &ett_nbap_Serving_E_DCH_RL_in_this_NodeB,
    &ett_nbap_SetsOfHS_SCCH_Codes,
    &ett_nbap_SetsOfHS_SCCH_CodesItem,
    &ett_nbap_Setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency,
    &ett_nbap_SFNSFNMeasurementThresholdInformation,
    &ett_nbap_SFNSFNMeasurementValueInformation,
    &ett_nbap_T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation,
    &ett_nbap_T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item,
    &ett_nbap_T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation,
    &ett_nbap_T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item,
    &ett_nbap_SFNSFNTimeStampInformation,
    &ett_nbap_SFNSFNTimeStamp_TDD,
    &ett_nbap_SFNSFNValue,
    &ett_nbap_SynchronisationReportCharacteristics,
    &ett_nbap_SynchronisationReportCharactThreExc,
    &ett_nbap_SynchronisationReportCharactThreInfoItem,
    &ett_nbap_SEQUENCE_SIZE_1_maxNrOfReceptsPerSyncFrame_OF_SynchronisationReportCharactCellSyncBurstInfoItem,
    &ett_nbap_SynchronisationReportCharactCellSyncBurstInfoItem,
    &ett_nbap_SyncDLCodeIdThreInfoLCR,
    &ett_nbap_SyncDLCodeIdThreInfoList,
    &ett_nbap_SyncDLCodeInfoListLCR,
    &ett_nbap_SyncDLCodeInfoItemLCR,
    &ett_nbap_TDD_ChannelisationCodeLCR,
    &ett_nbap_TDD_DL_Code_Information,
    &ett_nbap_TDD_DL_Code_InformationItem,
    &ett_nbap_TDD_DL_Code_LCR_Information,
    &ett_nbap_TDD_DL_Code_LCR_InformationItem,
    &ett_nbap_TDD_DL_Code_768_Information,
    &ett_nbap_TDD_DL_Code_768_InformationItem,
    &ett_nbap_TDD_DL_DPCH_TimeSlotFormat_LCR,
    &ett_nbap_TDD_DPCHOffset,
    &ett_nbap_TransportFormatCombination_Beta,
    &ett_nbap_T_signalledGainFactors,
    &ett_nbap_T_gainFactor,
    &ett_nbap_T_fdd,
    &ett_nbap_TDD_UL_Code_Information,
    &ett_nbap_TDD_UL_Code_InformationItem,
    &ett_nbap_TDD_UL_Code_LCR_Information,
    &ett_nbap_TDD_UL_Code_LCR_InformationItem,
    &ett_nbap_TDD_UL_Code_768_Information,
    &ett_nbap_TDD_UL_Code_768_InformationItem,
    &ett_nbap_TDD_UL_DPCH_TimeSlotFormat_LCR,
    &ett_nbap_TFCI_SignallingMode,
    &ett_nbap_TimeSlot_InitiatedListLCR,
    &ett_nbap_TimeSlotMeasurementValueListLCR,
    &ett_nbap_TimeSlotMeasurementValueLCR,
    &ett_nbap_TimingAdjustmentValue,
    &ett_nbap_TimingAdjustmentValueLCR,
    &ett_nbap_TnlQos,
    &ett_nbap_Transmission_Gap_Pattern_Sequence_Information,
    &ett_nbap_Transmission_Gap_Pattern_Sequence_Information_item,
    &ett_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue,
    &ett_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue_Item,
    &ett_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue,
    &ett_nbap_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue_Item,
    &ett_nbap_Transmitted_Carrier_Power_For_CellPortion_Value,
    &ett_nbap_Transmitted_Carrier_Power_For_CellPortion_Value_Item,
    &ett_nbap_Transmitted_Carrier_Power_For_CellPortion_ValueLCR,
    &ett_nbap_Transmitted_Carrier_Power_For_CellPortion_ValueLCR_Item,
    &ett_nbap_TFCS,
    &ett_nbap_T_tFCSvalues,
    &ett_nbap_TFCS_TFCSList,
    &ett_nbap_TFCS_TFCSList_item,
    &ett_nbap_TFCS_CTFC,
    &ett_nbap_Transport_Block_Size_List,
    &ett_nbap_Transport_Block_Size_List_item,
    &ett_nbap_TransportFormatSet,
    &ett_nbap_TransportFormatSet_DynamicPartList,
    &ett_nbap_TransportFormatSet_DynamicPartList_item,
    &ett_nbap_TDD_TransportFormatSet_ModeDP,
    &ett_nbap_TransmissionTimeIntervalInformation,
    &ett_nbap_TransmissionTimeIntervalInformation_item,
    &ett_nbap_TransportFormatSet_Semi_staticPart,
    &ett_nbap_TransportFormatSet_ModeDP,
    &ett_nbap_TransportFormatSet_ModeSSP,
    &ett_nbap_TUTRANGANSS,
    &ett_nbap_TUTRANGANSSMeasurementThresholdInformation,
    &ett_nbap_TUTRANGANSSMeasurementValueInformation,
    &ett_nbap_TUTRANGPS,
    &ett_nbap_TUTRANGPSMeasurementThresholdInformation,
    &ett_nbap_TUTRANGPSMeasurementValueInformation,
    &ett_nbap_UC_Id,
    &ett_nbap_UE_AggregateMaximumBitRate,
    &ett_nbap_UE_Capability_Information,
    &ett_nbap_UL_Timeslot_Information,
    &ett_nbap_UL_Timeslot_InformationItem,
    &ett_nbap_UL_TimeslotLCR_Information,
    &ett_nbap_UL_TimeslotLCR_InformationItem,
    &ett_nbap_UL_Timeslot768_Information,
    &ett_nbap_UL_Timeslot768_InformationItem,
    &ett_nbap_UL_ScramblingCode,
    &ett_nbap_UL_Synchronisation_Parameters_LCR,
    &ett_nbap_UL_TimeSlot_ISCP_Info,
    &ett_nbap_UL_TimeSlot_ISCP_InfoItem,
    &ett_nbap_UL_TimeSlot_ISCP_LCR_Info,
    &ett_nbap_UL_TimeSlot_ISCP_LCR_InfoItem,
    &ett_nbap_UpPTSInterference_For_CellPortion_Value,
    &ett_nbap_UpPTSInterference_For_CellPortion_Value_Item,
    &ett_nbap_USCH_Information,
    &ett_nbap_USCH_InformationItem,
    &ett_nbap_USCH_InformationResponse,
    &ett_nbap_USCH_InformationResponseItem,
    &ett_nbap_UL_TimeslotISCP_For_CellPortion_Value,
    &ett_nbap_UL_TimeslotISCP_For_CellPortion_Value_Item,
    &ett_nbap_CommonTransportChannelSetupRequestFDD,
    &ett_nbap_CommonPhysicalChannelType_CTCH_SetupRqstFDD,
    &ett_nbap_Secondary_CCPCH_CTCH_SetupRqstFDD,
    &ett_nbap_PowerOffsetInformation_CTCH_SetupRqstFDD,
    &ett_nbap_FACH_ParametersListIE_CTCH_SetupRqstFDD,
    &ett_nbap_FACH_ParametersItem_CTCH_SetupRqstFDD,
    &ett_nbap_PCH_ParametersItem_CTCH_SetupRqstFDD,
    &ett_nbap_PICH_Parameters_CTCH_SetupRqstFDD,
    &ett_nbap_MICH_Parameters_CTCH_SetupRqstFDD,
    &ett_nbap_PRACH_CTCH_SetupRqstFDD,
    &ett_nbap_AllowedSlotFormatInformationList_CTCH_SetupRqstFDD,
    &ett_nbap_AllowedSlotFormatInformationItem_CTCH_SetupRqstFDD,
    &ett_nbap_RACH_ParametersItem_CTCH_SetupRqstFDD,
    &ett_nbap_AICH_Parameters_CTCH_SetupRqstFDD,
    &ett_nbap_CommonTransportChannelSetupRequestTDD,
    &ett_nbap_CommonPhysicalChannelType_CTCH_SetupRqstTDD,
    &ett_nbap_Secondary_CCPCH_CTCH_SetupRqstTDD,
    &ett_nbap_Secondary_CCPCH_parameterListIE_CTCH_SetupRqstTDD,
    &ett_nbap_Secondary_CCPCH_parameterItem_CTCH_SetupRqstTDD,
    &ett_nbap_Secondary_CCPCH_LCR_parameterList_CTCH_SetupRqstTDD,
    &ett_nbap_Secondary_CCPCH_LCR_parameterItem_CTCH_SetupRqstTDD,
    &ett_nbap_Secondary_CCPCH_768_parameterList_CTCH_SetupRqstTDD,
    &ett_nbap_Secondary_CCPCH_768_parameterItem_CTCH_SetupRqstTDD,
    &ett_nbap_FACH_ParametersListIE_CTCH_SetupRqstTDD,
    &ett_nbap_FACH_ParametersItem_CTCH_SetupRqstTDD,
    &ett_nbap_PCH_ParametersItem_CTCH_SetupRqstTDD,
    &ett_nbap_PICH_ParametersItem_CTCH_SetupRqstTDD,
    &ett_nbap_PICH_LCR_Parameters_CTCH_SetupRqstTDD,
    &ett_nbap_PICH_768_ParametersItem_CTCH_SetupRqstTDD,
    &ett_nbap_MICH_Parameters_CTCH_SetupRqstTDD,
    &ett_nbap_MICH_TDDOption_Specific_Parameters_CTCH_SetupRqstTDD,
    &ett_nbap_MICH_HCR_Parameters_CTCH_SetupRqstTDD,
    &ett_nbap_MICH_LCR_Parameters_CTCH_SetupRqstTDD,
    &ett_nbap_MICH_768_Parameters_CTCH_SetupRqstTDD,
    &ett_nbap_TimeSlotConfigurationList_LCR_CTCH_SetupRqstTDD,
    &ett_nbap_TimeSlotConfigurationItem_LCR_CTCH_SetupRqstTDD,
    &ett_nbap_Secondary_CCPCH_parameterExtendedList_CTCH_SetupRqstTDD,
    &ett_nbap_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_SetupRqstTDD,
    &ett_nbap_PRACH_CTCH_SetupRqstTDD,
    &ett_nbap_PRACH_ParametersItem_CTCH_SetupRqstTDD,
    &ett_nbap_RACH_ParameterItem_CTCH_SetupRqstTDD,
    &ett_nbap_PRACH_LCR_ParametersList_CTCH_SetupRqstTDD,
    &ett_nbap_PRACH_LCR_ParametersItem_CTCH_SetupRqstTDD,
    &ett_nbap_PRACH_768_ParametersItem_CTCH_SetupRqstTDD,
    &ett_nbap_FPACH_LCR_Parameters_CTCH_SetupRqstTDD,
    &ett_nbap_PLCCH_parameters,
    &ett_nbap_E_RUCCH_parameters,
    &ett_nbap_E_RUCCH_768_parameters,
    &ett_nbap_CommonTransportChannelSetupResponse,
    &ett_nbap_FACH_CommonTransportChannel_InformationResponse,
    &ett_nbap_CommonTransportChannelSetupFailure,
    &ett_nbap_CommonTransportChannelReconfigurationRequestFDD,
    &ett_nbap_CommonPhysicalChannelType_CTCH_ReconfRqstFDD,
    &ett_nbap_Secondary_CCPCHList_CTCH_ReconfRqstFDD,
    &ett_nbap_FACH_ParametersListIE_CTCH_ReconfRqstFDD,
    &ett_nbap_FACH_ParametersItem_CTCH_ReconfRqstFDD,
    &ett_nbap_PCH_ParametersItem_CTCH_ReconfRqstFDD,
    &ett_nbap_PICH_ParametersItem_CTCH_ReconfRqstFDD,
    &ett_nbap_MICH_Parameters_CTCH_ReconfRqstFDD,
    &ett_nbap_PRACHList_CTCH_ReconfRqstFDD,
    &ett_nbap_PRACH_ParametersListIE_CTCH_ReconfRqstFDD,
    &ett_nbap_PRACH_ParametersItem_CTCH_ReconfRqstFDD,
    &ett_nbap_AllowedSlotFormatInformationList_CTCH_ReconfRqstFDD,
    &ett_nbap_AllowedSlotFormatInformationItem_CTCH_ReconfRqstFDD,
    &ett_nbap_AICH_ParametersListIE_CTCH_ReconfRqstFDD,
    &ett_nbap_AICH_ParametersItem_CTCH_ReconfRqstFDD,
    &ett_nbap_CommonTransportChannelReconfigurationRequestTDD,
    &ett_nbap_Secondary_CCPCH_Parameters_CTCH_ReconfRqstTDD,
    &ett_nbap_Secondary_CCPCHListIE_CTCH_ReconfRqstTDD,
    &ett_nbap_Secondary_CCPCHItem_CTCH_ReconfRqstTDD,
    &ett_nbap_Secondary_CCPCH_parameterExtendedList_CTCH_ReconfRqstTDD,
    &ett_nbap_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_ReconfRqstTDD,
    &ett_nbap_PICH_Parameters_CTCH_ReconfRqstTDD,
    &ett_nbap_FACH_ParametersList_CTCH_ReconfRqstTDD,
    &ett_nbap_FACH_ParametersItem_CTCH_ReconfRqstTDD,
    &ett_nbap_PCH_Parameters_CTCH_ReconfRqstTDD,
    &ett_nbap_FPACH_LCR_Parameters_CTCH_ReconfRqstTDD,
    &ett_nbap_MICH_Parameters_CTCH_ReconfRqstTDD,
    &ett_nbap_PLCCH_Parameters_CTCH_ReconfRqstTDD,
    &ett_nbap_Secondary_CCPCH_768_Parameters_CTCH_ReconfRqstTDD,
    &ett_nbap_Secondary_CCPCH_768_List_CTCH_ReconfRqstTDD,
    &ett_nbap_Secondary_CCPCH_768_Item_CTCH_ReconfRqstTDD,
    &ett_nbap_PICH_768_Parameters_CTCH_ReconfRqstTDD,
    &ett_nbap_MICH_768_Parameters_CTCH_ReconfRqstTDD,
    &ett_nbap_UPPCH_LCR_Parameters_CTCH_ReconfRqstTDD,
    &ett_nbap_CommonTransportChannelReconfigurationResponse,
    &ett_nbap_CommonTransportChannelReconfigurationFailure,
    &ett_nbap_CommonTransportChannelDeletionRequest,
    &ett_nbap_CommonTransportChannelDeletionResponse,
    &ett_nbap_BlockResourceRequest,
    &ett_nbap_BlockResourceResponse,
    &ett_nbap_BlockResourceFailure,
    &ett_nbap_UnblockResourceIndication,
    &ett_nbap_AuditRequiredIndication,
    &ett_nbap_AuditRequest,
    &ett_nbap_AuditResponse,
    &ett_nbap_Cell_InformationList_AuditRsp,
    &ett_nbap_Cell_InformationItem_AuditRsp,
    &ett_nbap_S_CPICH_InformationList_AuditRsp,
    &ett_nbap_S_CCPCH_InformationList_AuditRsp,
    &ett_nbap_FACH_InformationList_AuditRsp,
    &ett_nbap_PRACH_InformationList_AuditRsp,
    &ett_nbap_RACH_InformationList_AuditRsp,
    &ett_nbap_AICH_InformationList_AuditRsp,
    &ett_nbap_FPACH_LCR_InformationList_AuditRsp,
    &ett_nbap_HS_DSCH_Resources_Information_AuditRsp,
    &ett_nbap_S_CCPCH_InformationListExt_AuditRsp,
    &ett_nbap_S_CCPCH_LCR_InformationListExt_AuditRsp,
    &ett_nbap_E_DCH_Resources_Information_AuditRsp,
    &ett_nbap_PLCCH_InformationList_AuditRsp,
    &ett_nbap_S_CCPCH_768_InformationList_AuditRsp,
    &ett_nbap_PRACH_768_InformationList_AuditRsp,
    &ett_nbap_E_RUCCH_InformationList_AuditRsp,
    &ett_nbap_E_RUCCH_768_InformationList_AuditRsp,
    &ett_nbap_Cell_Frequency_List_Information_LCR_MulFreq_AuditRsp,
    &ett_nbap_Cell_Frequency_List_InformationItem_LCR_MulFreq_AuditRsp,
    &ett_nbap_UPPCH_LCR_InformationList_AuditRsp,
    &ett_nbap_UPPCH_LCR_InformationItem_AuditRsp,
    &ett_nbap_MultipleFreq_HS_DSCH_Resources_InformationList_AuditRsp,
    &ett_nbap_MultipleFreq_E_DCH_Resources_InformationList_AuditRsp,
    &ett_nbap_CCP_InformationList_AuditRsp,
    &ett_nbap_CCP_InformationItem_AuditRsp,
    &ett_nbap_Local_Cell_InformationList_AuditRsp,
    &ett_nbap_Local_Cell_InformationItem_AuditRsp,
    &ett_nbap_Local_Cell_Group_InformationList_AuditRsp,
    &ett_nbap_Local_Cell_Group_InformationItem_AuditRsp,
    &ett_nbap_Power_Local_Cell_Group_InformationList_AuditRsp,
    &ett_nbap_Power_Local_Cell_Group_InformationItem_AuditRsp,
    &ett_nbap_AuditFailure,
    &ett_nbap_CommonMeasurementInitiationRequest,
    &ett_nbap_CommonMeasurementObjectType_CM_Rqst,
    &ett_nbap_ERACH_CM_Rqst,
    &ett_nbap_Cell_CM_Rqst,
    &ett_nbap_RACH_CM_Rqst,
    &ett_nbap_PowerLocalCellGroup_CM_Rqst,
    &ett_nbap_CommonMeasurementInitiationResponse,
    &ett_nbap_CommonMeasurementObjectType_CM_Rsp,
    &ett_nbap_ERACH_CM_Rsp,
    &ett_nbap_Cell_CM_Rsp,
    &ett_nbap_RACH_CM_Rsp,
    &ett_nbap_PowerLocalCellGroup_CM_Rsp,
    &ett_nbap_CommonMeasurementInitiationFailure,
    &ett_nbap_CommonMeasurementReport,
    &ett_nbap_CommonMeasurementObjectType_CM_Rprt,
    &ett_nbap_ERACH_CM_Rprt,
    &ett_nbap_Cell_CM_Rprt,
    &ett_nbap_RACH_CM_Rprt,
    &ett_nbap_PowerLocalCellGroup_CM_Rprt,
    &ett_nbap_CommonMeasurementTerminationRequest,
    &ett_nbap_CommonMeasurementFailureIndication,
    &ett_nbap_CellSetupRequestFDD,
    &ett_nbap_Synchronisation_Configuration_Cell_SetupRqst,
    &ett_nbap_PrimarySCH_Information_Cell_SetupRqstFDD,
    &ett_nbap_SecondarySCH_Information_Cell_SetupRqstFDD,
    &ett_nbap_PrimaryCPICH_Information_Cell_SetupRqstFDD,
    &ett_nbap_SecondaryCPICH_InformationList_Cell_SetupRqstFDD,
    &ett_nbap_SecondaryCPICH_InformationItem_Cell_SetupRqstFDD,
    &ett_nbap_PrimaryCCPCH_Information_Cell_SetupRqstFDD,
    &ett_nbap_BCH_Information_Cell_SetupRqstFDD,
    &ett_nbap_Limited_power_increase_information_Cell_SetupRqstFDD,
    &ett_nbap_IPDLParameter_Information_Cell_SetupRqstFDD,
    &ett_nbap_CellPortion_InformationList_Cell_SetupRqstFDD,
    &ett_nbap_CellPortion_InformationItem_Cell_SetupRqstFDD,
    &ett_nbap_CellSetupRequestTDD,
    &ett_nbap_SCH_Information_Cell_SetupRqstTDD,
    &ett_nbap_SyncCaseIndicatorItem_Cell_SetupRqstTDD_PSCH,
    &ett_nbap_Case1_Cell_SetupRqstTDD,
    &ett_nbap_Case2_Cell_SetupRqstTDD,
    &ett_nbap_PCCPCH_Information_Cell_SetupRqstTDD,
    &ett_nbap_TimeSlotConfigurationList_Cell_SetupRqstTDD,
    &ett_nbap_TimeSlotConfigurationItem_Cell_SetupRqstTDD,
    &ett_nbap_TimeSlotConfigurationList_LCR_Cell_SetupRqstTDD,
    &ett_nbap_TimeSlotConfigurationItem_LCR_Cell_SetupRqstTDD,
    &ett_nbap_PCCPCH_LCR_Information_Cell_SetupRqstTDD,
    &ett_nbap_DwPCH_LCR_Information_Cell_SetupRqstTDD,
    &ett_nbap_IPDLParameter_Information_Cell_SetupRqstTDD,
    &ett_nbap_IPDLParameter_Information_LCR_Cell_SetupRqstTDD,
    &ett_nbap_PCCPCH_768_Information_Cell_SetupRqstTDD,
    &ett_nbap_SCH_768_Information_Cell_SetupRqstTDD,
    &ett_nbap_Cell_Frequency_List_LCR_MulFreq_Cell_SetupRqstTDD,
    &ett_nbap_Cell_Frequency_Item_LCR_MulFreq_Cell_SetupRqstTDD,
    &ett_nbap_CellSetupResponse,
    &ett_nbap_CellSetupFailure,
    &ett_nbap_CellReconfigurationRequestFDD,
    &ett_nbap_Synchronisation_Configuration_Cell_ReconfRqst,
    &ett_nbap_PrimarySCH_Information_Cell_ReconfRqstFDD,
    &ett_nbap_SecondarySCH_Information_Cell_ReconfRqstFDD,
    &ett_nbap_PrimaryCPICH_Information_Cell_ReconfRqstFDD,
    &ett_nbap_SecondaryCPICH_InformationList_Cell_ReconfRqstFDD,
    &ett_nbap_SecondaryCPICH_InformationItem_Cell_ReconfRqstFDD,
    &ett_nbap_PrimaryCCPCH_Information_Cell_ReconfRqstFDD,
    &ett_nbap_BCH_information_Cell_ReconfRqstFDD,
    &ett_nbap_IPDLParameter_Information_Cell_ReconfRqstFDD,
    &ett_nbap_CellPortion_InformationList_Cell_ReconfRqstFDD,
    &ett_nbap_CellPortion_InformationItem_Cell_ReconfRqstFDD,
    &ett_nbap_CellReconfigurationRequestTDD,
    &ett_nbap_SCH_Information_Cell_ReconfRqstTDD,
    &ett_nbap_PCCPCH_Information_Cell_ReconfRqstTDD,
    &ett_nbap_TimeSlotConfigurationList_Cell_ReconfRqstTDD,
    &ett_nbap_TimeSlotConfigurationItem_Cell_ReconfRqstTDD,
    &ett_nbap_TimeSlotConfigurationList_LCR_Cell_ReconfRqstTDD,
    &ett_nbap_TimeSlotConfigurationItem_LCR_Cell_ReconfRqstTDD,
    &ett_nbap_DwPCH_LCR_Information_Cell_ReconfRqstTDD,
    &ett_nbap_IPDLParameter_Information_Cell_ReconfRqstTDD,
    &ett_nbap_IPDLParameter_Information_LCR_Cell_ReconfRqstTDD,
    &ett_nbap_SCH_768_Information_Cell_ReconfRqstTDD,
    &ett_nbap_PCCPCH_768_Information_Cell_ReconfRqstTDD,
    &ett_nbap_UARFCN_Adjustment,
    &ett_nbap_Cell_Frequency_Add_LCR_MulFreq_Cell_ReconfRqstTDD,
    &ett_nbap_Cell_Frequency_ModifyList_LCR_MulFreq_Cell_ReconfRqstTDD,
    &ett_nbap_Cell_Frequency_ModifyItem_LCR_MulFreq_Cell_ReconfRqstTDD,
    &ett_nbap_Cell_Frequency_Delete_LCR_MulFreq_Cell_ReconfRqstTDD,
    &ett_nbap_CellReconfigurationResponse,
    &ett_nbap_CellReconfigurationFailure,
    &ett_nbap_CellDeletionRequest,
    &ett_nbap_CellDeletionResponse,
    &ett_nbap_ResourceStatusIndication,
    &ett_nbap_IndicationType_ResourceStatusInd,
    &ett_nbap_No_Failure_ResourceStatusInd,
    &ett_nbap_Local_Cell_InformationList_ResourceStatusInd,
    &ett_nbap_Local_Cell_InformationItem_ResourceStatusInd,
    &ett_nbap_Local_Cell_Group_InformationList_ResourceStatusInd,
    &ett_nbap_Local_Cell_Group_InformationItem_ResourceStatusInd,
    &ett_nbap_Power_Local_Cell_Group_InformationList_ResourceStatusInd,
    &ett_nbap_Power_Local_Cell_Group_InformationItem_ResourceStatusInd,
    &ett_nbap_ServiceImpacting_ResourceStatusInd,
    &ett_nbap_Local_Cell_InformationList2_ResourceStatusInd,
    &ett_nbap_Local_Cell_InformationItem2_ResourceStatusInd,
    &ett_nbap_Local_Cell_Group_InformationList2_ResourceStatusInd,
    &ett_nbap_Local_Cell_Group_InformationItem2_ResourceStatusInd,
    &ett_nbap_CCP_InformationList_ResourceStatusInd,
    &ett_nbap_CCP_InformationItem_ResourceStatusInd,
    &ett_nbap_Cell_InformationList_ResourceStatusInd,
    &ett_nbap_Cell_InformationItem_ResourceStatusInd,
    &ett_nbap_S_CPICH_InformationList_ResourceStatusInd,
    &ett_nbap_S_CCPCH_InformationList_ResourceStatusInd,
    &ett_nbap_FACH_InformationList_ResourceStatusInd,
    &ett_nbap_PRACH_InformationList_ResourceStatusInd,
    &ett_nbap_RACH_InformationList_ResourceStatusInd,
    &ett_nbap_AICH_InformationList_ResourceStatusInd,
    &ett_nbap_FPACH_LCR_InformationList_ResourceStatusInd,
    &ett_nbap_HS_DSCH_Resources_Information_ResourceStatusInd,
    &ett_nbap_S_CCPCH_InformationListExt_ResourceStatusInd,
    &ett_nbap_S_CCPCH_LCR_InformationListExt_ResourceStatusInd,
    &ett_nbap_E_DCH_Resources_Information_ResourceStatusInd,
    &ett_nbap_PLCCH_InformationList_ResourceStatusInd,
    &ett_nbap_S_CCPCH_768_InformationList_ResourceStatusInd,
    &ett_nbap_PRACH_768_InformationList_ResourceStatusInd,
    &ett_nbap_E_RUCCH_InformationList_ResourceStatusInd,
    &ett_nbap_E_RUCCH_768_InformationList_ResourceStatusInd,
    &ett_nbap_Cell_Frequency_List_Information_LCR_MulFreq_ResourceStatusInd,
    &ett_nbap_Cell_Frequency_List_InformationItem_LCR_MulFreq_ResourceStatusInd,
    &ett_nbap_UPPCH_LCR_InformationList_ResourceStatusInd,
    &ett_nbap_UPPCH_LCR_InformationItem_ResourceStatusInd,
    &ett_nbap_MultipleFreq_HS_DSCH_Resources_InformationList_ResourceStatusInd,
    &ett_nbap_Power_Local_Cell_Group_InformationList2_ResourceStatusInd,
    &ett_nbap_Power_Local_Cell_Group_InformationItem2_ResourceStatusInd,
    &ett_nbap_MultipleFreq_E_DCH_Resources_InformationList_ResourceStatusInd,
    &ett_nbap_SystemInformationUpdateRequest,
    &ett_nbap_MIB_SB_SIB_InformationList_SystemInfoUpdateRqst,
    &ett_nbap_MIB_SB_SIB_InformationItem_SystemInfoUpdateRqst,
    &ett_nbap_DeletionIndicator_SystemInfoUpdate,
    &ett_nbap_No_Deletion_SystemInfoUpdate,
    &ett_nbap_SegmentInformationListIE_SystemInfoUpdate,
    &ett_nbap_SegmentInformationItem_SystemInfoUpdate,
    &ett_nbap_SystemInformationUpdateResponse,
    &ett_nbap_SystemInformationUpdateFailure,
    &ett_nbap_RadioLinkSetupRequestFDD,
    &ett_nbap_Additional_HS_Cell_Information_RL_Setup_List,
    &ett_nbap_Additional_HS_Cell_Information_RL_Setup_ItemIEs,
    &ett_nbap_UL_DPCH_Information_RL_SetupRqstFDD,
    &ett_nbap_DL_DPCH_Information_RL_SetupRqstFDD,
    &ett_nbap_PowerOffsetInformation_RL_SetupRqstFDD,
    &ett_nbap_RL_InformationList_RL_SetupRqstFDD,
    &ett_nbap_RL_InformationItem_RL_SetupRqstFDD,
    &ett_nbap_E_DPCH_Information_RL_SetupRqstFDD,
    &ett_nbap_F_DPCH_Information_RL_SetupRqstFDD,
    &ett_nbap_PowerOffsetInformation_F_DPCH_RL_SetupRqstFDD,
    &ett_nbap_RadioLinkSetupRequestTDD,
    &ett_nbap_UL_CCTrCH_InformationList_RL_SetupRqstTDD,
    &ett_nbap_UL_CCTrCH_InformationItem_RL_SetupRqstTDD,
    &ett_nbap_UL_DPCH_InformationItem_RL_SetupRqstTDD,
    &ett_nbap_UL_DPCH_LCR_Information_RL_SetupRqstTDD,
    &ett_nbap_UL_DPCH_768_Information_RL_SetupRqstTDD,
    &ett_nbap_DL_CCTrCH_InformationList_RL_SetupRqstTDD,
    &ett_nbap_DL_CCTrCH_InformationItem_RL_SetupRqstTDD,
    &ett_nbap_CCTrCH_TPCList_RL_SetupRqstTDD,
    &ett_nbap_CCTrCH_TPCItem_RL_SetupRqstTDD,
    &ett_nbap_DL_DPCH_InformationItem_RL_SetupRqstTDD,
    &ett_nbap_DL_DPCH_LCR_Information_RL_SetupRqstTDD,
    &ett_nbap_DL_DPCH_768_Information_RL_SetupRqstTDD,
    &ett_nbap_RL_Information_RL_SetupRqstTDD,
    &ett_nbap_RadioLinkSetupResponseFDD,
    &ett_nbap_Additional_HS_Cell_Information_Response_List,
    &ett_nbap_Additional_HS_Cell_Information_Response_ItemIEs,
    &ett_nbap_RL_InformationResponseList_RL_SetupRspFDD,
    &ett_nbap_RL_InformationResponseItem_RL_SetupRspFDD,
    &ett_nbap_DiversityIndication_RL_SetupRspFDD,
    &ett_nbap_Combining_RL_SetupRspFDD,
    &ett_nbap_NonCombiningOrFirstRL_RL_SetupRspFDD,
    &ett_nbap_RadioLinkSetupResponseTDD,
    &ett_nbap_RL_InformationResponse_RL_SetupRspTDD,
    &ett_nbap_RL_InformationResponse_LCR_RL_SetupRspTDD,
    &ett_nbap_RadioLinkSetupFailureFDD,
    &ett_nbap_CauseLevel_RL_SetupFailureFDD,
    &ett_nbap_GeneralCauseList_RL_SetupFailureFDD,
    &ett_nbap_RLSpecificCauseList_RL_SetupFailureFDD,
    &ett_nbap_Unsuccessful_RL_InformationRespList_RL_SetupFailureFDD,
    &ett_nbap_Unsuccessful_RL_InformationRespItem_RL_SetupFailureFDD,
    &ett_nbap_Successful_RL_InformationRespList_RL_SetupFailureFDD,
    &ett_nbap_Successful_RL_InformationRespItem_RL_SetupFailureFDD,
    &ett_nbap_DiversityIndication_RL_SetupFailureFDD,
    &ett_nbap_Combining_RL_SetupFailureFDD,
    &ett_nbap_NonCombiningOrFirstRL_RL_SetupFailureFDD,
    &ett_nbap_RadioLinkSetupFailureTDD,
    &ett_nbap_CauseLevel_RL_SetupFailureTDD,
    &ett_nbap_GeneralCauseList_RL_SetupFailureTDD,
    &ett_nbap_RLSpecificCauseList_RL_SetupFailureTDD,
    &ett_nbap_Unsuccessful_RL_InformationResp_RL_SetupFailureTDD,
    &ett_nbap_RadioLinkAdditionRequestFDD,
    &ett_nbap_Additional_HS_Cell_Information_RL_Addition_List,
    &ett_nbap_Additional_EDCH_Cell_Information_RL_Add_Req,
    &ett_nbap_Setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency,
    &ett_nbap_Additional_HS_Cell_Information_RL_Addition_ItemIEs,
    &ett_nbap_RL_InformationList_RL_AdditionRqstFDD,
    &ett_nbap_RL_InformationItem_RL_AdditionRqstFDD,
    &ett_nbap_E_DPCH_Information_RL_AdditionReqFDD,
    &ett_nbap_RadioLinkAdditionRequestTDD,
    &ett_nbap_UL_CCTrCH_InformationList_RL_AdditionRqstTDD,
    &ett_nbap_UL_CCTrCH_InformationItem_RL_AdditionRqstTDD,
    &ett_nbap_UL_DPCH_InformationItem_RL_AdditionRqstTDD,
    &ett_nbap_UL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD,
    &ett_nbap_UL_DPCH_InformationItem_768_RL_AdditionRqstTDD,
    &ett_nbap_DL_CCTrCH_InformationList_RL_AdditionRqstTDD,
    &ett_nbap_DL_CCTrCH_InformationItem_RL_AdditionRqstTDD,
    &ett_nbap_DL_DPCH_InformationItem_RL_AdditionRqstTDD,
    &ett_nbap_DL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD,
    &ett_nbap_DL_DPCH_InformationItem_768_RL_AdditionRqstTDD,
    &ett_nbap_RL_Information_RL_AdditionRqstTDD,
    &ett_nbap_RadioLinkAdditionResponseFDD,
    &ett_nbap_Additional_HS_Cell_Change_Information_Response_List,
    &ett_nbap_Additional_HS_Cell_Change_Information_Response_ItemIEs,
    &ett_nbap_RL_InformationResponseList_RL_AdditionRspFDD,
    &ett_nbap_RL_InformationResponseItem_RL_AdditionRspFDD,
    &ett_nbap_DiversityIndication_RL_AdditionRspFDD,
    &ett_nbap_Combining_RL_AdditionRspFDD,
    &ett_nbap_Non_Combining_RL_AdditionRspFDD,
    &ett_nbap_RadioLinkAdditionResponseTDD,
    &ett_nbap_RL_InformationResponse_RL_AdditionRspTDD,
    &ett_nbap_DCH_Information_RL_AdditionRspTDD,
    &ett_nbap_DiversityIndication_RL_AdditionRspTDD,
    &ett_nbap_Combining_RL_AdditionRspTDD,
    &ett_nbap_Non_Combining_RL_AdditionRspTDD,
    &ett_nbap_RL_InformationResponse_LCR_RL_AdditionRspTDD,
    &ett_nbap_RadioLinkAdditionFailureFDD,
    &ett_nbap_CauseLevel_RL_AdditionFailureFDD,
    &ett_nbap_GeneralCauseList_RL_AdditionFailureFDD,
    &ett_nbap_RLSpecificCauseList_RL_AdditionFailureFDD,
    &ett_nbap_Unsuccessful_RL_InformationRespList_RL_AdditionFailureFDD,
    &ett_nbap_Unsuccessful_RL_InformationRespItem_RL_AdditionFailureFDD,
    &ett_nbap_Successful_RL_InformationRespList_RL_AdditionFailureFDD,
    &ett_nbap_Successful_RL_InformationRespItem_RL_AdditionFailureFDD,
    &ett_nbap_DiversityIndication_RL_AdditionFailureFDD,
    &ett_nbap_Combining_RL_AdditionFailureFDD,
    &ett_nbap_Non_Combining_RL_AdditionFailureFDD,
    &ett_nbap_RadioLinkAdditionFailureTDD,
    &ett_nbap_CauseLevel_RL_AdditionFailureTDD,
    &ett_nbap_GeneralCauseList_RL_AdditionFailureTDD,
    &ett_nbap_RLSpecificCauseList_RL_AdditionFailureTDD,
    &ett_nbap_Unsuccessful_RL_InformationResp_RL_AdditionFailureTDD,
    &ett_nbap_RadioLinkReconfigurationPrepareFDD,
    &ett_nbap_Additional_HS_Cell_Information_RL_Reconf_Prep,
    &ett_nbap_Additional_HS_Cell_Information_RL_Reconf_Prep_ItemIEs,
    &ett_nbap_Additional_EDCH_Cell_Information_RL_Reconf_Prep,
    &ett_nbap_UL_DPCH_Information_RL_ReconfPrepFDD,
    &ett_nbap_DL_DPCH_Information_RL_ReconfPrepFDD,
    &ett_nbap_DL_DPCH_Power_Information_RL_ReconfPrepFDD,
    &ett_nbap_PowerOffsetInformation_RL_ReconfPrepFDD,
    &ett_nbap_DCH_DeleteList_RL_ReconfPrepFDD,
    &ett_nbap_DCH_DeleteItem_RL_ReconfPrepFDD,
    &ett_nbap_RL_InformationList_RL_ReconfPrepFDD,
    &ett_nbap_RL_InformationItem_RL_ReconfPrepFDD,
    &ett_nbap_E_DPCH_Information_RL_ReconfPrepFDD,
    &ett_nbap_F_DPCH_Information_RL_ReconfPrepFDD,
    &ett_nbap_PowerOffsetInformation_F_DPCH_RL_ReconfPrepFDD,
    &ett_nbap_RadioLinkReconfigurationPrepareTDD,
    &ett_nbap_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD,
    &ett_nbap_UL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD,
    &ett_nbap_UL_DPCH_InformationAddItem_RL_ReconfPrepTDD,
    &ett_nbap_UL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD,
    &ett_nbap_MultipleRL_UL_DPCH_InformationAddList_RL_ReconfPrepTDD,
    &ett_nbap_MultipleRL_UL_DPCH_InformationAddListIE_RL_ReconfPrepTDD,
    &ett_nbap_UL_DPCH_768_InformationAddList_RL_ReconfPrepTDD,
    &ett_nbap_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD,
    &ett_nbap_UL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD,
    &ett_nbap_UL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD,
    &ett_nbap_UL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD,
    &ett_nbap_UL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD,
    &ett_nbap_UL_Timeslot_InformationModify_ModifyItem_RL_ReconfPrepTDD,
    &ett_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD,
    &ett_nbap_UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD,
    &ett_nbap_UL_TimeslotLCR_InformationModify_ModifyList_RL_ReconfPrepTDD,
    &ett_nbap_UL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD,
    &ett_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDDLCR,
    &ett_nbap_UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDDLCR,
    &ett_nbap_UL_Timeslot768_InformationModify_ModifyList_RL_ReconfPrepTDD,
    &ett_nbap_UL_Timeslot_768_InformationModify_ModifyItem_RL_ReconfPrepTDD,
    &ett_nbap_UL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD768,
    &ett_nbap_UL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD768,
    &ett_nbap_UL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD,
    &ett_nbap_UL_DPCH_InformationModify_DeleteItem_RL_ReconfPrepTDD,
    &ett_nbap_UL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD,
    &ett_nbap_MultipleRL_UL_DPCH_InformationModifyList_RL_ReconfPrepTDD,
    &ett_nbap_MultipleRL_UL_DPCH_InformationModifyListIE_RL_ReconfPrepTDD,
    &ett_nbap_UL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD,
    &ett_nbap_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD,
    &ett_nbap_UL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD,
    &ett_nbap_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD,
    &ett_nbap_DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD,
    &ett_nbap_CCTrCH_TPCAddList_RL_ReconfPrepTDD,
    &ett_nbap_CCTrCH_TPCAddItem_RL_ReconfPrepTDD,
    &ett_nbap_DL_DPCH_InformationAddItem_RL_ReconfPrepTDD,
    &ett_nbap_DL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD,
    &ett_nbap_MultipleRL_DL_DPCH_InformationAddList_RL_ReconfPrepTDD,
    &ett_nbap_MultipleRL_DL_DPCH_InformationAddListIE_RL_ReconfPrepTDD,
    &ett_nbap_DL_DPCH_768_InformationAddList_RL_ReconfPrepTDD,
    &ett_nbap_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD,
    &ett_nbap_DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD,
    &ett_nbap_CCTrCH_TPCModifyList_RL_ReconfPrepTDD,
    &ett_nbap_CCTrCH_TPCModifyItem_RL_ReconfPrepTDD,
    &ett_nbap_DL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD,
    &ett_nbap_DL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD,
    &ett_nbap_DL_Timeslot_InformationModify_ModifyList_RL_ReconfPrepTDD,
    &ett_nbap_DL_Timeslot_InformationModify_ModifyItem_RL_ReconfPrepTDD,
    &ett_nbap_DL_Code_InformationModify_ModifyList_RL_ReconfPrepTDD,
    &ett_nbap_DL_Code_InformationModify_ModifyItem_RL_ReconfPrepTDD,
    &ett_nbap_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD,
    &ett_nbap_DL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD,
    &ett_nbap_DL_Code_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD,
    &ett_nbap_DL_Code_LCR_InformationModify_ModifyItem_RL_ReconfPrepTDD,
    &ett_nbap_DL_Timeslot_768_InformationModify_ModifyList_RL_ReconfPrepTDD,
    &ett_nbap_DL_Timeslot_768_InformationModify_ModifyItem_RL_ReconfPrepTDD,
    &ett_nbap_DL_Code_768_InformationModify_ModifyList_RL_ReconfPrepTDD,
    &ett_nbap_DL_Code_768_InformationModify_ModifyItem_RL_ReconfPrepTDD,
    &ett_nbap_DL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD,
    &ett_nbap_DL_DPCH_InformationModify_DeleteItem_RL_ReconfPrepTDD,
    &ett_nbap_DL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD,
    &ett_nbap_MultipleRL_DL_DPCH_InformationModifyList_RL_ReconfPrepTDD,
    &ett_nbap_MultipleRL_DL_DPCH_InformationModifyListIE_RL_ReconfPrepTDD,
    &ett_nbap_DL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD,
    &ett_nbap_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD,
    &ett_nbap_DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD,
    &ett_nbap_DCH_DeleteList_RL_ReconfPrepTDD,
    &ett_nbap_DCH_DeleteItem_RL_ReconfPrepTDD,
    &ett_nbap_DSCH_Information_ModifyList_RL_ReconfPrepTDD,
    &ett_nbap_DSCH_Information_ModifyItem_RL_ReconfPrepTDD,
    &ett_nbap_DSCH_Information_DeleteList_RL_ReconfPrepTDD,
    &ett_nbap_DSCH_Information_DeleteItem_RL_ReconfPrepTDD,
    &ett_nbap_USCH_Information_ModifyList_RL_ReconfPrepTDD,
    &ett_nbap_USCH_Information_ModifyItem_RL_ReconfPrepTDD,
    &ett_nbap_USCH_Information_DeleteList_RL_ReconfPrepTDD,
    &ett_nbap_USCH_Information_DeleteItem_RL_ReconfPrepTDD,
    &ett_nbap_MultipleRL_Information_RL_ReconfPrepTDD,
    &ett_nbap_RL_Information_RL_ReconfPrepTDD,
    &ett_nbap_RadioLinkReconfigurationReady,
    &ett_nbap_RL_InformationResponseList_RL_ReconfReady,
    &ett_nbap_RL_InformationResponseItem_RL_ReconfReady,
    &ett_nbap_RadioLinkReconfigurationFailure,
    &ett_nbap_CauseLevel_RL_ReconfFailure,
    &ett_nbap_GeneralCauseList_RL_ReconfFailure,
    &ett_nbap_RLSpecificCauseList_RL_ReconfFailure,
    &ett_nbap_RL_ReconfigurationFailureList_RL_ReconfFailure,
    &ett_nbap_RL_ReconfigurationFailureItem_RL_ReconfFailure,
    &ett_nbap_RadioLinkReconfigurationCommit,
    &ett_nbap_RadioLinkReconfigurationCancel,
    &ett_nbap_RadioLinkReconfigurationRequestFDD,
    &ett_nbap_Additional_HS_Cell_Information_RL_Reconf_Req,
    &ett_nbap_Additional_HS_Cell_Information_RL_Reconf_Req_ItemIEs,
    &ett_nbap_Additional_EDCH_Cell_Information_RL_Reconf_Req,
    &ett_nbap_UL_DPCH_Information_RL_ReconfRqstFDD,
    &ett_nbap_DL_DPCH_Information_RL_ReconfRqstFDD,
    &ett_nbap_DCH_DeleteList_RL_ReconfRqstFDD,
    &ett_nbap_DCH_DeleteItem_RL_ReconfRqstFDD,
    &ett_nbap_RL_InformationList_RL_ReconfRqstFDD,
    &ett_nbap_RL_InformationItem_RL_ReconfRqstFDD,
    &ett_nbap_E_DPCH_Information_RL_ReconfRqstFDD,
    &ett_nbap_RadioLinkReconfigurationRequestTDD,
    &ett_nbap_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD,
    &ett_nbap_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD,
    &ett_nbap_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD,
    &ett_nbap_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD,
    &ett_nbap_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD,
    &ett_nbap_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD,
    &ett_nbap_MultipleRL_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD,
    &ett_nbap_MultipleRL_DL_CCTrCH_InformationModifyListIE_RL_ReconfRqstTDD,
    &ett_nbap_DL_DPCH_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD,
    &ett_nbap_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD,
    &ett_nbap_DL_Timeslot_LCR_InformationModify_ModifyItem_RL_ReconfRqstTDD,
    &ett_nbap_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD,
    &ett_nbap_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD,
    &ett_nbap_DCH_DeleteList_RL_ReconfRqstTDD,
    &ett_nbap_DCH_DeleteItem_RL_ReconfRqstTDD,
    &ett_nbap_Multiple_RL_Information_RL_ReconfRqstTDD,
    &ett_nbap_RL_Information_RL_ReconfRqstTDD,
    &ett_nbap_RadioLinkReconfigurationResponse,
    &ett_nbap_RL_InformationResponseList_RL_ReconfRsp,
    &ett_nbap_RL_InformationResponseItem_RL_ReconfRsp,
    &ett_nbap_RadioLinkDeletionRequest,
    &ett_nbap_RL_informationList_RL_DeletionRqst,
    &ett_nbap_RL_informationItem_RL_DeletionRqst,
    &ett_nbap_RadioLinkDeletionResponse,
    &ett_nbap_DL_PowerControlRequest,
    &ett_nbap_DL_ReferencePowerInformationList_DL_PC_Rqst,
    &ett_nbap_DL_ReferencePowerInformationItem_DL_PC_Rqst,
    &ett_nbap_DL_PowerTimeslotControlRequest,
    &ett_nbap_DedicatedMeasurementInitiationRequest,
    &ett_nbap_DedicatedMeasurementObjectType_DM_Rqst,
    &ett_nbap_RL_DM_Rqst,
    &ett_nbap_RL_InformationList_DM_Rqst,
    &ett_nbap_RL_InformationItem_DM_Rqst,
    &ett_nbap_PUSCH_Info_DM_Rqst,
    &ett_nbap_HSSICH_Info_DM_Rqst,
    &ett_nbap_HSSICH_InfoExt_DM_Rqst,
    &ett_nbap_RL_Set_DM_Rqst,
    &ett_nbap_RL_Set_InformationList_DM_Rqst,
    &ett_nbap_RL_Set_InformationItem_DM_Rqst,
    &ett_nbap_DedicatedMeasurementInitiationResponse,
    &ett_nbap_DedicatedMeasurementObjectType_DM_Rsp,
    &ett_nbap_RL_DM_Rsp,
    &ett_nbap_RL_InformationList_DM_Rsp,
    &ett_nbap_RL_InformationItem_DM_Rsp,
    &ett_nbap_PUSCH_Info_DM_Rsp,
    &ett_nbap_Multiple_PUSCH_InfoList_DM_Rsp,
    &ett_nbap_Multiple_PUSCH_InfoListIE_DM_Rsp,
    &ett_nbap_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp,
    &ett_nbap_Multiple_DedicatedMeasurementValueItem_TDD_DM_Rsp,
    &ett_nbap_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp,
    &ett_nbap_Multiple_DedicatedMeasurementValueItem_LCR_TDD_DM_Rsp,
    &ett_nbap_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp,
    &ett_nbap_Multiple_HSSICHMeasurementValueItem_TDD_DM_Rsp,
    &ett_nbap_Multiple_DedicatedMeasurementValueList_768_TDD_DM_Rsp,
    &ett_nbap_Multiple_DedicatedMeasurementValueItem_768_TDD_DM_Rsp,
    &ett_nbap_RL_Set_DM_Rsp,
    &ett_nbap_RL_Set_InformationList_DM_Rsp,
    &ett_nbap_RL_Set_InformationItem_DM_Rsp,
    &ett_nbap_DedicatedMeasurementInitiationFailure,
    &ett_nbap_DedicatedMeasurementReport,
    &ett_nbap_DedicatedMeasurementObjectType_DM_Rprt,
    &ett_nbap_RL_DM_Rprt,
    &ett_nbap_RL_InformationList_DM_Rprt,
    &ett_nbap_RL_InformationItem_DM_Rprt,
    &ett_nbap_PUSCH_Info_DM_Rprt,
    &ett_nbap_Multiple_PUSCH_InfoList_DM_Rprt,
    &ett_nbap_Multiple_PUSCH_InfoListIE_DM_Rprt,
    &ett_nbap_RL_Set_DM_Rprt,
    &ett_nbap_RL_Set_InformationList_DM_Rprt,
    &ett_nbap_RL_Set_InformationItem_DM_Rprt,
    &ett_nbap_DedicatedMeasurementTerminationRequest,
    &ett_nbap_DedicatedMeasurementFailureIndication,
    &ett_nbap_RadioLinkFailureIndication,
    &ett_nbap_Reporting_Object_RL_FailureInd,
    &ett_nbap_RL_RL_FailureInd,
    &ett_nbap_RL_InformationList_RL_FailureInd,
    &ett_nbap_RL_InformationItem_RL_FailureInd,
    &ett_nbap_RL_Set_RL_FailureInd,
    &ett_nbap_RL_Set_InformationList_RL_FailureInd,
    &ett_nbap_RL_Set_InformationItem_RL_FailureInd,
    &ett_nbap_CCTrCH_RL_FailureInd,
    &ett_nbap_CCTrCH_InformationList_RL_FailureInd,
    &ett_nbap_CCTrCH_InformationItem_RL_FailureInd,
    &ett_nbap_RadioLinkPreemptionRequiredIndication,
    &ett_nbap_RL_InformationList_RL_PreemptRequiredInd,
    &ett_nbap_RL_InformationItem_RL_PreemptRequiredInd,
    &ett_nbap_RadioLinkRestoreIndication,
    &ett_nbap_Reporting_Object_RL_RestoreInd,
    &ett_nbap_RL_RL_RestoreInd,
    &ett_nbap_RL_InformationList_RL_RestoreInd,
    &ett_nbap_RL_InformationItem_RL_RestoreInd,
    &ett_nbap_RL_Set_RL_RestoreInd,
    &ett_nbap_RL_Set_InformationList_RL_RestoreInd,
    &ett_nbap_RL_Set_InformationItem_RL_RestoreInd,
    &ett_nbap_CCTrCH_RL_RestoreInd,
    &ett_nbap_CCTrCH_InformationList_RL_RestoreInd,
    &ett_nbap_CCTrCH_InformationItem_RL_RestoreInd,
    &ett_nbap_CompressedModeCommand,
    &ett_nbap_ErrorIndication,
    &ett_nbap_PrivateMessage,
    &ett_nbap_PhysicalSharedChannelReconfigurationRequestFDD,
    &ett_nbap_HSDPA_And_EDCH_CellPortion_InformationList_PSCH_ReconfRqst,
    &ett_nbap_HSDPA_And_EDCH_CellPortion_InformationItem_PSCH_ReconfRqst,
    &ett_nbap_PhysicalSharedChannelReconfigurationRequestTDD,
    &ett_nbap_PDSCHSets_AddList_PSCH_ReconfRqst,
    &ett_nbap_PDSCHSets_AddItem_PSCH_ReconfRqst,
    &ett_nbap_PDSCH_Information_AddItem_PSCH_ReconfRqst,
    &ett_nbap_DL_Timeslot_InformationAddList_PSCH_ReconfRqst,
    &ett_nbap_DL_Timeslot_InformationAddItem_PSCH_ReconfRqst,
    &ett_nbap_DL_Code_InformationAddList_PSCH_ReconfRqst,
    &ett_nbap_DL_Code_InformationAddItem_PSCH_ReconfRqst,
    &ett_nbap_PDSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst,
    &ett_nbap_DL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst,
    &ett_nbap_DL_Timeslot_InformationAddItem_LCR_PSCH_ReconfRqst,
    &ett_nbap_DL_Code_InformationAddList_LCR_PSCH_ReconfRqst,
    &ett_nbap_DL_Code_InformationAddItem_LCR_PSCH_ReconfRqst,
    &ett_nbap_PDSCH_AddInformation_768_AddItem_PSCH_ReconfRqst,
    &ett_nbap_DL_Timeslot_InformationAddList_768_PSCH_ReconfRqst,
    &ett_nbap_DL_Timeslot_InformationAddItem_768_PSCH_ReconfRqst,
    &ett_nbap_DL_Code_InformationAddList_768_PSCH_ReconfRqst,
    &ett_nbap_DL_Code_InformationAddItem_768_PSCH_ReconfRqst,
    &ett_nbap_PDSCHSets_ModifyList_PSCH_ReconfRqst,
    &ett_nbap_PDSCHSets_ModifyItem_PSCH_ReconfRqst,
    &ett_nbap_PDSCH_Information_ModifyItem_PSCH_ReconfRqst,
    &ett_nbap_DL_Timeslot_InformationModifyList_PSCH_ReconfRqst,
    &ett_nbap_DL_Timeslot_InformationModifyItem_PSCH_ReconfRqst,
    &ett_nbap_DL_Code_InformationModifyList_PSCH_ReconfRqst,
    &ett_nbap_DL_Code_InformationModifyItem_PSCH_ReconfRqst,
    &ett_nbap_PDSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst,
    &ett_nbap_DL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst,
    &ett_nbap_DL_Timeslot_LCR_InformationModifyItem_PSCH_ReconfRqst,
    &ett_nbap_DL_Code_LCR_InformationModifyList_PSCH_ReconfRqst,
    &ett_nbap_DL_Code_LCR_InformationModifyItem_PSCH_ReconfRqst,
    &ett_nbap_PDSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst,
    &ett_nbap_DL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst,
    &ett_nbap_DL_Timeslot_768_InformationModifyItem_PSCH_ReconfRqst,
    &ett_nbap_DL_Code_768_InformationModifyList_PSCH_ReconfRqst,
    &ett_nbap_DL_Code_768_InformationModifyItem_PSCH_ReconfRqst,
    &ett_nbap_PDSCHSets_DeleteList_PSCH_ReconfRqst,
    &ett_nbap_PDSCHSets_DeleteItem_PSCH_ReconfRqst,
    &ett_nbap_PUSCHSets_AddList_PSCH_ReconfRqst,
    &ett_nbap_PUSCHSets_AddItem_PSCH_ReconfRqst,
    &ett_nbap_PUSCH_Information_AddItem_PSCH_ReconfRqst,
    &ett_nbap_UL_Timeslot_InformationAddList_PSCH_ReconfRqst,
    &ett_nbap_UL_Timeslot_InformationAddItem_PSCH_ReconfRqst,
    &ett_nbap_UL_Code_InformationAddList_PSCH_ReconfRqst,
    &ett_nbap_UL_Code_InformationAddItem_PSCH_ReconfRqst,
    &ett_nbap_PUSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst,
    &ett_nbap_UL_Timeslot_InformationAddList_LCR_PSCH_ReconfRqst,
    &ett_nbap_UL_Timeslot_InformationAddItem_LCR_PSCH_ReconfRqst,
    &ett_nbap_UL_Code_InformationAddList_LCR_PSCH_ReconfRqst,
    &ett_nbap_UL_Code_InformationAddItem_LCR_PSCH_ReconfRqst,
    &ett_nbap_PUSCH_AddInformation_768_AddItem_PSCH_ReconfRqst,
    &ett_nbap_UL_Timeslot_InformationAddList_768_PSCH_ReconfRqst,
    &ett_nbap_UL_Timeslot_InformationAddItem_768_PSCH_ReconfRqst,
    &ett_nbap_UL_Code_InformationAddList_768_PSCH_ReconfRqst,
    &ett_nbap_UL_Code_InformationAddItem_768_PSCH_ReconfRqst,
    &ett_nbap_PUSCHSets_ModifyList_PSCH_ReconfRqst,
    &ett_nbap_PUSCHSets_ModifyItem_PSCH_ReconfRqst,
    &ett_nbap_PUSCH_Information_ModifyItem_PSCH_ReconfRqst,
    &ett_nbap_UL_Timeslot_InformationModifyList_PSCH_ReconfRqst,
    &ett_nbap_UL_Timeslot_InformationModifyItem_PSCH_ReconfRqst,
    &ett_nbap_UL_Code_InformationModifyList_PSCH_ReconfRqst,
    &ett_nbap_UL_Code_InformationModifyItem_PSCH_ReconfRqst,
    &ett_nbap_PUSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst,
    &ett_nbap_UL_Timeslot_LCR_InformationModifyList_PSCH_ReconfRqst,
    &ett_nbap_UL_Timeslot_LCR_InformationModifyItem_PSCH_ReconfRqst,
    &ett_nbap_UL_Code_LCR_InformationModifyList_PSCH_ReconfRqst,
    &ett_nbap_UL_Code_LCR_InformationModifyItem_PSCH_ReconfRqst,
    &ett_nbap_PUSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst,
    &ett_nbap_UL_Timeslot_768_InformationModifyList_PSCH_ReconfRqst,
    &ett_nbap_UL_Timeslot_768_InformationModifyItem_PSCH_ReconfRqst,
    &ett_nbap_UL_Code_768_InformationModifyList_PSCH_ReconfRqst,
    &ett_nbap_UL_Code_768_InformationModifyItem_PSCH_ReconfRqst,
    &ett_nbap_PUSCHSets_DeleteList_PSCH_ReconfRqst,
    &ett_nbap_PUSCHSets_DeleteItem_PSCH_ReconfRqst,
    &ett_nbap_HS_PDSCH_TDD_Information_PSCH_ReconfRqst,
    &ett_nbap_DL_HS_PDSCH_Timeslot_Information_PSCH_ReconfRqst,
    &ett_nbap_DL_HS_PDSCH_Timeslot_InformationItem_PSCH_ReconfRqst,
    &ett_nbap_DL_HS_PDSCH_Codelist_PSCH_ReconfRqst,
    &ett_nbap_DL_HS_PDSCH_Timeslot_Information_768_PSCH_ReconfRqst,
    &ett_nbap_DL_HS_PDSCH_Timeslot_InformationItem_768_PSCH_ReconfRqst,
    &ett_nbap_DL_HS_PDSCH_Codelist_768_PSCH_ReconfRqst,
    &ett_nbap_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst,
    &ett_nbap_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst,
    &ett_nbap_Add_To_HS_SCCH_Resource_Pool_PSCH_ReconfRqst,
    &ett_nbap_HS_SCCH_Information_PSCH_ReconfRqst,
    &ett_nbap_HS_SCCH_InformationItem_PSCH_ReconfRqst,
    &ett_nbap_HS_SICH_Information_PSCH_ReconfRqst,
    &ett_nbap_HS_SCCH_Information_LCR_PSCH_ReconfRqst,
    &ett_nbap_HS_SCCH_InformationItem_LCR_PSCH_ReconfRqst,
    &ett_nbap_HS_SICH_Information_LCR_PSCH_ReconfRqst,
    &ett_nbap_HS_SCCH_Information_768_PSCH_ReconfRqst,
    &ett_nbap_HS_SCCH_InformationItem_768_PSCH_ReconfRqst,
    &ett_nbap_HS_SICH_Information_768_PSCH_ReconfRqst,
    &ett_nbap_HS_SCCH_InformationExt_LCR_PSCH_ReconfRqst,
    &ett_nbap_Modify_HS_SCCH_Resource_Pool_PSCH_ReconfRqst,
    &ett_nbap_HS_SCCH_InformationModifyItem_PSCH_ReconfRqst,
    &ett_nbap_HS_SICH_InformationModify_PSCH_ReconfRqst,
    &ett_nbap_HS_SCCH_InformationModify_LCR_PSCH_ReconfRqst,
    &ett_nbap_HS_SCCH_InformationModifyItem_LCR_PSCH_ReconfRqst,
    &ett_nbap_HS_SCCH_InformationModifyExt_LCR_PSCH_ReconfRqst,
    &ett_nbap_HS_SICH_InformationModify_LCR_PSCH_ReconfRqst,
    &ett_nbap_HS_SCCH_InformationModify_768_PSCH_ReconfRqst,
    &ett_nbap_HS_SCCH_InformationModifyItem_768_PSCH_ReconfRqst,
    &ett_nbap_HS_SICH_InformationModify_768_PSCH_ReconfRqst,
    &ett_nbap_HS_SCCH_InformationModify_PSCH_ReconfRqst,
    &ett_nbap_Delete_From_HS_SCCH_Resource_Pool_PSCH_ReconfRqst,
    &ett_nbap_Delete_From_HS_SCCH_Resource_PoolItem_PSCH_ReconfRqst,
    &ett_nbap_E_PUCH_Information_PSCH_ReconfRqst,
    &ett_nbap_E_PUCH_Timeslot_Info,
    &ett_nbap_Add_To_E_AGCH_Resource_Pool_PSCH_ReconfRqst,
    &ett_nbap_E_AGCH_Information_PSCH_ReconfRqst,
    &ett_nbap_E_AGCH_InformationItem_PSCH_ReconfRqst,
    &ett_nbap_Modify_E_AGCH_Resource_Pool_PSCH_ReconfRqst,
    &ett_nbap_E_AGCH_InformationModify_PSCH_ReconfRqst,
    &ett_nbap_E_AGCH_InformationModifyItem_PSCH_ReconfRqst,
    &ett_nbap_Delete_From_E_AGCH_Resource_Pool_PSCH_ReconfRqst,
    &ett_nbap_Delete_From_E_AGCH_Resource_PoolItem_PSCH_ReconfRqst,
    &ett_nbap_E_HICH_Information_PSCH_ReconfRqst,
    &ett_nbap_E_PUCH_Information_768_PSCH_ReconfRqst,
    &ett_nbap_Add_To_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst,
    &ett_nbap_E_AGCH_Information_768_PSCH_ReconfRqst,
    &ett_nbap_E_AGCH_InformationItem_768_PSCH_ReconfRqst,
    &ett_nbap_Modify_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst,
    &ett_nbap_E_AGCH_InformationModify_768_PSCH_ReconfRqst,
    &ett_nbap_E_AGCH_InformationModifyItem_768_PSCH_ReconfRqst,
    &ett_nbap_E_HICH_Information_768_PSCH_ReconfRqst,
    &ett_nbap_E_PUCH_Information_LCR_PSCH_ReconfRqst,
    &ett_nbap_E_PUCH_Timeslot_InfoLCR,
    &ett_nbap_E_PUCH_Timeslot_Item_InfoLCR,
    &ett_nbap_E_PUCH_Codelist_LCR,
    &ett_nbap_Add_To_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst,
    &ett_nbap_E_AGCH_Information_LCR_PSCH_ReconfRqst,
    &ett_nbap_E_AGCH_InformationItem_LCR_PSCH_ReconfRqst,
    &ett_nbap_Modify_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst,
    &ett_nbap_E_AGCH_InformationModify_LCR_PSCH_ReconfRqst,
    &ett_nbap_E_AGCH_InformationModifyItem_LCR_PSCH_ReconfRqst,
    &ett_nbap_Add_To_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst,
    &ett_nbap_E_HICH_Information_LCR_PSCH_ReconfRqst,
    &ett_nbap_E_HICH_InformationItem_LCR_PSCH_ReconfRqst,
    &ett_nbap_Modify_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst,
    &ett_nbap_E_HICH_InformationModify_LCR_PSCH_ReconfRqst,
    &ett_nbap_E_HICH_InformationModifyItem_LCR_PSCH_ReconfRqst,
    &ett_nbap_Delete_From_E_HICH_Resource_Pool_PSCH_ReconfRqst,
    &ett_nbap_Delete_From_E_HICH_Resource_PoolItem_PSCH_ReconfRqst,
    &ett_nbap_SYNC_UL_Partition_LCR,
    &ett_nbap_Delete_From_HS_SCCH_Resource_PoolExt_PSCH_ReconfRqst,
    &ett_nbap_MultipleFreq_E_PUCH_Timeslot_InformationList_LCR_PSCH_ReconfRqst,
    &ett_nbap_MultipleFreq_E_PUCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst,
    &ett_nbap_Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst,
    &ett_nbap_Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst_Item,
    &ett_nbap_PhysicalSharedChannelReconfigurationResponse,
    &ett_nbap_E_HICH_TimeOffset_ExtensionLCR,
    &ett_nbap_MultipleFreq_E_HICH_TimeOffsetLCR,
    &ett_nbap_PhysicalSharedChannelReconfigurationFailure,
    &ett_nbap_CauseLevel_PSCH_ReconfFailure,
    &ett_nbap_GeneralCauseList_PSCH_ReconfFailure,
    &ett_nbap_SetSpecificCauseList_PSCH_ReconfFailureTDD,
    &ett_nbap_Unsuccessful_PDSCHSetList_PSCH_ReconfFailureTDD,
    &ett_nbap_Unsuccessful_PDSCHSetItem_PSCH_ReconfFailureTDD,
    &ett_nbap_Unsuccessful_PUSCHSetList_PSCH_ReconfFailureTDD,
    &ett_nbap_Unsuccessful_PUSCHSetItem_PSCH_ReconfFailureTDD,
    &ett_nbap_UARFCNSpecificCauseList_PSCH_ReconfFailureTDD,
    &ett_nbap_Unsuccessful_UARFCNItem_PSCH_ReconfFailureTDD,
    &ett_nbap_E_HICH_TimeOffset_ReconfFailureTDD,
    &ett_nbap_Common_System_Information_ResponseLCR,
    &ett_nbap_ResetRequest,
    &ett_nbap_ResetIndicator,
    &ett_nbap_CommunicationContextList_Reset,
    &ett_nbap_CommunicationContextInfoList_Reset,
    &ett_nbap_CommunicationContextInfoItem_Reset,
    &ett_nbap_CommunicationContextType_Reset,
    &ett_nbap_CommunicationControlPortList_Reset,
    &ett_nbap_CommunicationControlPortInfoList_Reset,
    &ett_nbap_CommunicationControlPortInfoItem_Reset,
    &ett_nbap_ResetResponse,
    &ett_nbap_InformationExchangeInitiationRequest,
    &ett_nbap_InformationExchangeObjectType_InfEx_Rqst,
    &ett_nbap_Cell_InfEx_Rqst,
    &ett_nbap_InformationExchangeInitiationResponse,
    &ett_nbap_InformationExchangeObjectType_InfEx_Rsp,
    &ett_nbap_Cell_InfEx_Rsp,
    &ett_nbap_InformationExchangeInitiationFailure,
    &ett_nbap_InformationReport,
    &ett_nbap_InformationExchangeObjectType_InfEx_Rprt,
    &ett_nbap_Cell_Inf_Rprt,
    &ett_nbap_InformationExchangeTerminationRequest,
    &ett_nbap_InformationExchangeFailureIndication,
    &ett_nbap_CellSynchronisationInitiationRequestTDD,
    &ett_nbap_TimeslotInfo_CellSyncInitiationRqstTDD,
    &ett_nbap_CellSyncBurstTransInit_CellSyncInitiationRqstTDD,
    &ett_nbap_CellSyncBurstMeasureInit_CellSyncInitiationRqstTDD,
    &ett_nbap_SYNCDlCodeId_TransInitLCR_CellSyncInitiationRqstTDD,
    &ett_nbap_SYNCDlCodeId_MeasureInitLCR_CellSyncInitiationRqstTDD,
    &ett_nbap_CellSynchronisationInitiationResponseTDD,
    &ett_nbap_CellSynchronisationInitiationFailureTDD,
    &ett_nbap_CellSynchronisationReconfigurationRequestTDD,
    &ett_nbap_CellSyncBurstTransReconfInfo_CellSyncReconfRqstTDD,
    &ett_nbap_CellSyncBurstTransInfoItem_CellSyncReconfRqstTDD,
    &ett_nbap_CellSyncBurstMeasInfo_CellSyncReconfRqstTDD,
    &ett_nbap_CellSyncBurstMeasInfoListIE_CellSyncReconfRqstTDD,
    &ett_nbap_CellSyncBurstMeasInfoItem_CellSyncReconfRqstTDD,
    &ett_nbap_CellSyncBurstInfoList_CellSyncReconfRqstTDD,
    &ett_nbap_CellSyncBurstInfoItem_CellSyncReconfRqstTDD,
    &ett_nbap_SYNCDlCodeIdTransReconfInfoLCR_CellSyncReconfRqstTDD,
    &ett_nbap_SYNCDlCodeIdTransReconfItemLCR_CellSyncReconfRqstTDD,
    &ett_nbap_SYNCDlCodeIdMeasInfoLCR_CellSyncReconfRqstTDD,
    &ett_nbap_SYNCDlCodeIdMeasInfoList_CellSyncReconfRqstTDD,
    &ett_nbap_SYNCDlCodeIdMeasInfoItem_CellSyncReconfRqstTDD,
    &ett_nbap_SYNCDlCodeIdInfoListLCR_CellSyncReconfRqstTDD,
    &ett_nbap_SYNCDlCodeIdInfoItemLCR_CellSyncReconfRqstTDD,
    &ett_nbap_CellSynchronisationReconfigurationResponseTDD,
    &ett_nbap_CellSynchronisationReconfigurationFailureTDD,
    &ett_nbap_CellSynchronisationAdjustmentRequestTDD,
    &ett_nbap_CellAdjustmentInfo_SyncAdjustmentRqstTDD,
    &ett_nbap_CellAdjustmentInfoItem_SyncAdjustmentRqstTDD,
    &ett_nbap_CellSynchronisationAdjustmentResponseTDD,
    &ett_nbap_CellSynchronisationAdjustmentFailureTDD,
    &ett_nbap_CauseLevel_SyncAdjustmntFailureTDD,
    &ett_nbap_GeneralCauseList_SyncAdjustmntFailureTDD,
    &ett_nbap_CellSpecificCauseList_SyncAdjustmntFailureTDD,
    &ett_nbap_Unsuccessful_cell_InformationRespList_SyncAdjustmntFailureTDD,
    &ett_nbap_Unsuccessful_cell_InformationRespItem_SyncAdjustmntFailureTDD,
    &ett_nbap_CellSynchronisationTerminationRequestTDD,
    &ett_nbap_CellSynchronisationFailureIndicationTDD,
    &ett_nbap_CellSynchronisationReportTDD,
    &ett_nbap_CellSyncInfo_CellSyncReprtTDD,
    &ett_nbap_CellSyncInfoItemIE_CellSyncReprtTDD,
    &ett_nbap_SyncReportType_CellSyncReprtTDD,
    &ett_nbap_IntStdPhCellSyncInfo_CellSyncReprtTDD,
    &ett_nbap_CellSyncBurstMeasInfoList_CellSyncReprtTDD,
    &ett_nbap_CellSyncBurstMeasInfoItem_CellSyncReprtTDD,
    &ett_nbap_SEQUENCE_SIZE_1_maxNrOfReceptsPerSyncFrame_OF_CellSyncBurstInfo_CellSyncReprtTDD,
    &ett_nbap_CellSyncBurstInfo_CellSyncReprtTDD,
    &ett_nbap_CellSyncBurstAvailable_CellSyncReprtTDD,
    &ett_nbap_SyncDLCodeIdsMeasInfoList_CellSyncReprtTDD,
    &ett_nbap_SyncDLCodeIdsMeasInfoItem_CellSyncReprtTDD,
    &ett_nbap_SyncDLCodeIdInfo_CellSyncReprtTDD,
    &ett_nbap_SyncDLCodeIdItem_CellSyncReprtTDD,
    &ett_nbap_SyncDLCodeIdAvailable_CellSyncReprtTDD,
    &ett_nbap_BearerRearrangementIndication,
    &ett_nbap_DCH_RearrangeList_Bearer_RearrangeInd,
    &ett_nbap_DCH_RearrangeItem_Bearer_RearrangeInd,
    &ett_nbap_DSCH_RearrangeList_Bearer_RearrangeInd,
    &ett_nbap_DSCH_RearrangeItem_Bearer_RearrangeInd,
    &ett_nbap_USCH_RearrangeList_Bearer_RearrangeInd,
    &ett_nbap_USCH_RearrangeItem_Bearer_RearrangeInd,
    &ett_nbap_HSDSCH_RearrangeList_Bearer_RearrangeInd,
    &ett_nbap_HSDSCH_RearrangeItem_Bearer_RearrangeInd,
    &ett_nbap_E_DCH_RearrangeList_Bearer_RearrangeInd,
    &ett_nbap_E_DCH_RearrangeItem_Bearer_RearrangeInd,
    &ett_nbap_Additional_EDCH_Cell_Information_Bearer_Rearrangement_List,
    &ett_nbap_Additional_EDCH_Cell_Information_Bearer_Rearrangement_ItemIEs,
    &ett_nbap_RadioLinkActivationCommandFDD,
    &ett_nbap_DelayedActivationInformationList_RL_ActivationCmdFDD,
    &ett_nbap_DelayedActivationInformation_RL_ActivationCmdFDD,
    &ett_nbap_RadioLinkActivationCommandTDD,
    &ett_nbap_DelayedActivationInformationList_RL_ActivationCmdTDD,
    &ett_nbap_DelayedActivationInformation_RL_ActivationCmdTDD,
    &ett_nbap_RadioLinkParameterUpdateIndicationFDD,
    &ett_nbap_Additional_HS_Cell_Information_RL_Param_Upd,
    &ett_nbap_Additional_HS_Cell_Information_RL_Param_Upd_ItemIEs,
    &ett_nbap_Additional_EDCH_Cell_Information_RL_Param_Upd,
    &ett_nbap_Additional_EDCH_Cell_Information_RL_Param_Upd_ItemIEs,
    &ett_nbap_RadioLinkParameterUpdateIndicationTDD,
    &ett_nbap_MBMSNotificationUpdateCommand,
    &ett_nbap_UEStatusUpdateCommand,
    &ett_nbap_SecondaryULFrequencyReport,
    &ett_nbap_SecondaryULFrequencyUpdateIndication,
    &ett_nbap_NBAP_PDU,
    &ett_nbap_InitiatingMessage,
    &ett_nbap_SuccessfulOutcome,
    &ett_nbap_UnsuccessfulOutcome,
    &ett_nbap_Outcome,
  };

  static ei_register_info ei[] = {
    { &ei_nbap_no_set_comm_context_id, { "nbap.no_set_comm_context_id", PI_MALFORMED, PI_WARN, "Couldn't not set Communication Context-ID, fragments over reconfigured channels might fail", EXPFILL }},
    { &ei_nbap_no_find_port_info, { "nbap.no_find_port_info", PI_MALFORMED, PI_WARN, "Couldn't not find port information for reconfigured E-DCH flow, unable to reconfigure", EXPFILL }},
    { &ei_nbap_hsdsch_entity_not_specified, { "nbap.hsdsch_entity_not_specified", PI_MALFORMED,PI_ERROR, "HSDSCH Entity not specified!", EXPFILL }},
  };

  expert_module_t* expert_nbap;

  /* Register protocol */
  proto_nbap = proto_register_protocol(PNAME, PSNAME, PFNAME);
  /* Register fields and subtrees */
  proto_register_field_array(proto_nbap, hf, array_length(hf));
  proto_register_subtree_array(ett, array_length(ett));
  expert_nbap = expert_register_protocol(proto_nbap);
  expert_register_field_array(expert_nbap, ei, array_length(ei));

  /* Register dissector */
  nbap_handle = register_dissector("nbap", dissect_nbap, proto_nbap);

  nbap_module = prefs_register_protocol(proto_nbap, NULL);

  /* Register preferences for mapping logical channel IDs to MAC content types. */
  for (i = 0; i < 16; i++) {
    prefs_register_enum_preference(nbap_module, ch_strings[i].name, ch_strings[i].title, ch_strings[i].description, &lch_contents[i], content_types, FALSE);
  }
  prefs_register_enum_preference(nbap_module, "ib_sg_data_encoding",
    "IB_SG_DATA encoding",
    "Encoding used for the IB-SG-DATA element carrying segments of information blocks",
    &preferences_ib_sg_data_encoding, ib_sg_enc_vals, FALSE);

  /* Register dissector tables */
  nbap_ies_dissector_table = register_dissector_table("nbap.ies", "NBAP-PROTOCOL-IES", proto_nbap, FT_UINT32, BASE_DEC);
  nbap_extension_dissector_table = register_dissector_table("nbap.extension", "NBAP-PROTOCOL-EXTENSION", proto_nbap, FT_UINT32, BASE_DEC);
  nbap_proc_imsg_dissector_table = register_dissector_table("nbap.proc.imsg", "NBAP-ELEMENTARY-PROCEDURE InitiatingMessage", proto_nbap, FT_STRING, STRING_CASE_SENSITIVE);
  nbap_proc_sout_dissector_table = register_dissector_table("nbap.proc.sout", "NBAP-ELEMENTARY-PROCEDURE SuccessfulOutcome", proto_nbap, FT_STRING, STRING_CASE_SENSITIVE);
  nbap_proc_uout_dissector_table = register_dissector_table("nbap.proc.uout", "NBAP-ELEMENTARY-PROCEDURE UnsuccessfulOutcome", proto_nbap, FT_STRING, STRING_CASE_SENSITIVE);

  register_init_routine(nbap_init);
}

/*
 * #define	EXTRA_PPI 1
 */
/*--- proto_reg_handoff_nbap ---------------------------------------*/
void
proto_reg_handoff_nbap(void)
{
  fp_handle = find_dissector("fp");
  dissector_add_uint("sctp.ppi", NBAP_PAYLOAD_PROTOCOL_ID, nbap_handle);
#ifdef EXTRA_PPI
  dissector_add_uint("sctp.ppi", 17, nbap_handle);
#endif
  dissector_add_for_decode_as("sctp.port", nbap_handle);
  heur_dissector_add("sctp", dissect_nbap_heur, "NBAP over SCTP", "nbap_sctp", proto_nbap, HEURISTIC_ENABLE);

  dissector_add_uint("nbap.ies", id_TUTRANGPSMeasurementValueInformation, create_dissector_handle(dissect_TUTRANGPSMeasurementValueInformation_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_SFNSFNMeasurementValueInformation, create_dissector_handle(dissect_SFNSFNMeasurementValueInformation_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_TransmittedCarrierPowerOfAllCodesNotUsedForHSTransmission, create_dissector_handle(dissect_TransmittedCarrierPowerOfAllCodesNotUsedForHSTransmissionValue_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_HS_DSCHRequiredPowerValueInformation, create_dissector_handle(dissect_HS_DSCHRequiredPower_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_HS_DSCHProvidedBitRateValueInformation, create_dissector_handle(dissect_HS_DSCHProvidedBitRate_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Transmitted_Carrier_Power_For_CellPortion_Value, create_dissector_handle(dissect_Transmitted_Carrier_Power_For_CellPortion_Value_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Received_total_wide_band_power_For_CellPortion_Value, create_dissector_handle(dissect_Received_total_wide_band_power_For_CellPortion_Value_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue, create_dissector_handle(dissect_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortionValue_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UpPTSInterferenceValue, create_dissector_handle(dissect_UpPTSInterferenceValue_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DLTransmissionBranchLoadValue, create_dissector_handle(dissect_DLTransmissionBranchLoadValue_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_HS_DSCHRequiredPowerValueInformation_For_CellPortion, create_dissector_handle(dissect_HS_DSCHRequiredPowerValueInformation_For_CellPortion_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_HS_DSCHProvidedBitRateValueInformation_For_CellPortion, create_dissector_handle(dissect_HS_DSCHProvidedBitRateValueInformation_For_CellPortion_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_E_DCHProvidedBitRateValueInformation, create_dissector_handle(dissect_E_DCHProvidedBitRate_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_E_DCH_Non_serving_Relative_Grant_Down_CommandsValue, create_dissector_handle(dissect_E_DCH_Non_serving_Relative_Grant_Down_Commands_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Received_Scheduled_EDCH_Power_Share_Value, create_dissector_handle(dissect_Received_Scheduled_EDCH_Power_Share_Value_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value, create_dissector_handle(dissect_Received_Scheduled_EDCH_Power_Share_For_CellPortion_Value_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_TUTRANGANSSMeasurementValueInformation, create_dissector_handle(dissect_TUTRANGANSSMeasurementValueInformation_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Rx_Timing_Deviation_Value_LCR, create_dissector_handle(dissect_Rx_Timing_Deviation_Value_LCR_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Angle_Of_Arrival_Value_LCR, create_dissector_handle(dissect_Angle_Of_Arrival_Value_LCR_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_HS_SICH_Reception_Quality, create_dissector_handle(dissect_HS_SICH_Reception_Quality_Value_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Best_Cell_Portions_Value, create_dissector_handle(dissect_Best_Cell_Portions_Value_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Rx_Timing_Deviation_Value_768, create_dissector_handle(dissect_Rx_Timing_Deviation_Value_768_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Rx_Timing_Deviation_Value_384_ext, create_dissector_handle(dissect_Rx_Timing_Deviation_Value_384_ext_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Extended_Round_Trip_Time_Value, create_dissector_handle(dissect_Extended_Round_Trip_Time_Value_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_neighbouringTDDCellMeasurementInformationLCR, create_dissector_handle(dissect_NeighbouringTDDCellMeasurementInformationLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_neighbouringTDDCellMeasurementInformation768, create_dissector_handle(dissect_NeighbouringTDDCellMeasurementInformation768_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_ReportCharacteristicsType_OnModification, create_dissector_handle(dissect_ReportCharacteristicsType_OnModification_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Transmitted_Carrier_Power_For_CellPortion, create_dissector_handle(dissect_Transmitted_Carrier_Power_Value_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Received_total_wide_band_power_For_CellPortion, create_dissector_handle(dissect_Received_total_wide_band_power_Value_IncrDecrThres_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCH_E_RGCHOrE_HICHTransmissionCellPortion, create_dissector_handle(dissect_TransmittedCarrierPowerOfAllCodesNotUsedForHSTransmissionValue_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Received_Scheduled_EDCH_Power_Share, create_dissector_handle(dissect_RSEPS_Value_IncrDecrThres_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_TUTRANGPSMeasurementThresholdInformation, create_dissector_handle(dissect_TUTRANGPSMeasurementThresholdInformation_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_SFNSFNMeasurementThresholdInformation, create_dissector_handle(dissect_SFNSFNMeasurementThresholdInformation_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_HS_SICH_Reception_Quality_Measurement_Value, create_dissector_handle(dissect_HS_SICH_Reception_Quality_Measurement_Value_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_HS_DSCHRequiredPowerValue, create_dissector_handle(dissect_HS_DSCHRequiredPowerValue_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_HS_DSCHRequiredPowerValue_For_Cell_Portion, create_dissector_handle(dissect_HS_DSCHRequiredPowerValue_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Received_Scheduled_EDCH_Power_Share_For_CellPortion, create_dissector_handle(dissect_RSEPS_Value_IncrDecrThres_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Additional_HS_SICH_Reception_Quality_Measurement_Value, create_dissector_handle(dissect_HS_SICH_Reception_Quality_Measurement_Value_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_TUTRANGANSSMeasurementThresholdInformation, create_dissector_handle(dissect_TUTRANGANSSMeasurementThresholdInformation_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_C_ID, create_dissector_handle(dissect_C_ID_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_ConfigurationGenerationID, create_dissector_handle(dissect_ConfigurationGenerationID_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CommonPhysicalChannelType_CTCH_SetupRqstFDD, create_dissector_handle(dissect_CommonPhysicalChannelType_CTCH_SetupRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_FACH_ParametersListIE_CTCH_SetupRqstFDD, create_dissector_handle(dissect_FACH_ParametersListIE_CTCH_SetupRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PCH_ParametersItem_CTCH_SetupRqstFDD, create_dissector_handle(dissect_PCH_ParametersItem_CTCH_SetupRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RACH_ParametersItem_CTCH_SetupRqstFDD, create_dissector_handle(dissect_RACH_ParametersItem_CTCH_SetupRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CommonPhysicalChannelType_CTCH_SetupRqstTDD, create_dissector_handle(dissect_CommonPhysicalChannelType_CTCH_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Secondary_CCPCH_parameterListIE_CTCH_SetupRqstTDD, create_dissector_handle(dissect_Secondary_CCPCH_parameterListIE_CTCH_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Secondary_CCPCH_LCR_parameterList_CTCH_SetupRqstTDD, create_dissector_handle(dissect_Secondary_CCPCH_LCR_parameterList_CTCH_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_FACH_ParametersListIE_CTCH_SetupRqstTDD, create_dissector_handle(dissect_FACH_ParametersListIE_CTCH_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PCH_ParametersItem_CTCH_SetupRqstTDD, create_dissector_handle(dissect_PCH_ParametersItem_CTCH_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PICH_ParametersItem_CTCH_SetupRqstTDD, create_dissector_handle(dissect_PICH_ParametersItem_CTCH_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PICH_LCR_Parameters_CTCH_SetupRqstTDD, create_dissector_handle(dissect_PICH_LCR_Parameters_CTCH_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PRACH_ParametersItem_CTCH_SetupRqstTDD, create_dissector_handle(dissect_PRACH_ParametersItem_CTCH_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PRACH_LCR_ParametersList_CTCH_SetupRqstTDD, create_dissector_handle(dissect_PRACH_LCR_ParametersList_CTCH_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RACH_ParameterItem_CTCH_SetupRqstTDD, create_dissector_handle(dissect_RACH_ParameterItem_CTCH_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_FACH_ParametersList_CTCH_SetupRsp, create_dissector_handle(dissect_FACH_CommonTransportChannel_InformationResponse_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PCH_Parameters_CTCH_SetupRsp, create_dissector_handle(dissect_CommonTransportChannel_InformationResponse_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RACH_Parameters_CTCH_SetupRsp, create_dissector_handle(dissect_CommonTransportChannel_InformationResponse_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CriticalityDiagnostics, create_dissector_handle(dissect_CriticalityDiagnostics_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Cause, create_dissector_handle(dissect_Cause_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CommonPhysicalChannelType_CTCH_ReconfRqstFDD, create_dissector_handle(dissect_CommonPhysicalChannelType_CTCH_ReconfRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_FACH_ParametersListIE_CTCH_ReconfRqstFDD, create_dissector_handle(dissect_FACH_ParametersListIE_CTCH_ReconfRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PCH_ParametersItem_CTCH_ReconfRqstFDD, create_dissector_handle(dissect_PCH_ParametersItem_CTCH_ReconfRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PICH_ParametersItem_CTCH_ReconfRqstFDD, create_dissector_handle(dissect_PICH_ParametersItem_CTCH_ReconfRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PRACH_ParametersListIE_CTCH_ReconfRqstFDD, create_dissector_handle(dissect_PRACH_ParametersListIE_CTCH_ReconfRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_AICH_ParametersListIE_CTCH_ReconfRqstFDD, create_dissector_handle(dissect_AICH_ParametersListIE_CTCH_ReconfRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Secondary_CCPCH_Parameters_CTCH_ReconfRqstTDD, create_dissector_handle(dissect_Secondary_CCPCH_Parameters_CTCH_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PICH_Parameters_CTCH_ReconfRqstTDD, create_dissector_handle(dissect_PICH_Parameters_CTCH_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_FACH_ParametersList_CTCH_ReconfRqstTDD, create_dissector_handle(dissect_FACH_ParametersList_CTCH_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PCH_Parameters_CTCH_ReconfRqstTDD, create_dissector_handle(dissect_PCH_Parameters_CTCH_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Secondary_CCPCHListIE_CTCH_ReconfRqstTDD, create_dissector_handle(dissect_Secondary_CCPCHListIE_CTCH_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CommonPhysicalChannelID, create_dissector_handle(dissect_CommonPhysicalChannelID_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_BlockingPriorityIndicator, create_dissector_handle(dissect_BlockingPriorityIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_ShutdownTimer, create_dissector_handle(dissect_ShutdownTimer_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Start_Of_Audit_Sequence_Indicator, create_dissector_handle(dissect_Start_Of_Audit_Sequence_Indicator_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_End_Of_Audit_Sequence_Indicator, create_dissector_handle(dissect_End_Of_Audit_Sequence_Indicator_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Cell_InformationList_AuditRsp, create_dissector_handle(dissect_Cell_InformationList_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CCP_InformationList_AuditRsp, create_dissector_handle(dissect_CCP_InformationList_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Local_Cell_InformationList_AuditRsp, create_dissector_handle(dissect_Local_Cell_InformationList_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Local_Cell_Group_InformationList_AuditRsp, create_dissector_handle(dissect_Local_Cell_Group_InformationList_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Cell_InformationItem_AuditRsp, create_dissector_handle(dissect_Cell_InformationItem_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_P_SCH_Information, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_S_SCH_Information, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_P_CPICH_Information, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_S_CPICH_Information, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_P_CCPCH_Information, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_BCH_Information, create_dissector_handle(dissect_Common_TransportChannel_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_S_CCPCH_Information, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PCH_Information, create_dissector_handle(dissect_Common_TransportChannel_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PICH_Information, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_FACH_Information, create_dissector_handle(dissect_Common_TransportChannel_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PRACH_Information, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RACH_Information, create_dissector_handle(dissect_Common_TransportChannel_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_AICH_Information, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_SCH_Information, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CCP_InformationItem_AuditRsp, create_dissector_handle(dissect_CCP_InformationItem_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_FPACH_LCR_Information_AuditRsp, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_S_CCPCH_768_Information_AuditRsp, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information768_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PRACH_768_Information, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information768_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Local_Cell_InformationItem_AuditRsp, create_dissector_handle(dissect_Local_Cell_InformationItem_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Local_Cell_Group_InformationItem_AuditRsp, create_dissector_handle(dissect_Local_Cell_Group_InformationItem_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Power_Local_Cell_Group_InformationItem_AuditRsp, create_dissector_handle(dissect_Power_Local_Cell_Group_InformationItem_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PLCCH_Information_AuditRsp, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_E_RUCCH_Information, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_E_RUCCH_768_Information, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information768_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_MeasurementID, create_dissector_handle(dissect_MeasurementID_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CommonMeasurementObjectType_CM_Rqst, create_dissector_handle(dissect_CommonMeasurementObjectType_CM_Rqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CommonMeasurementType, create_dissector_handle(dissect_CommonMeasurementType_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_MeasurementFilterCoefficient, create_dissector_handle(dissect_MeasurementFilterCoefficient_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_ReportCharacteristics, create_dissector_handle(dissect_ReportCharacteristics_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_SFNReportingIndicator, create_dissector_handle(dissect_FNReportingIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_SFN, create_dissector_handle(dissect_SFN_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Power_Local_Cell_Group_choice_CM_Rqst, create_dissector_handle(dissect_PowerLocalCellGroup_CM_Rqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CommonMeasurementObjectType_CM_Rsp, create_dissector_handle(dissect_CommonMeasurementObjectType_CM_Rsp_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Power_Local_Cell_Group_choice_CM_Rsp, create_dissector_handle(dissect_PowerLocalCellGroup_CM_Rsp_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CommonMeasurementObjectType_CM_Rprt, create_dissector_handle(dissect_CommonMeasurementObjectType_CM_Rprt_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Power_Local_Cell_Group_choice_CM_Rprt, create_dissector_handle(dissect_PowerLocalCellGroup_CM_Rprt_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Local_Cell_ID, create_dissector_handle(dissect_Local_Cell_ID_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_T_Cell, create_dissector_handle(dissect_T_Cell_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UARFCNforNu, create_dissector_handle(dissect_UARFCN_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UARFCNforNd, create_dissector_handle(dissect_UARFCN_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_MaximumTransmissionPower, create_dissector_handle(dissect_MaximumTransmissionPower_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Closed_Loop_Timing_Adjustment_Mode, create_dissector_handle(dissect_Closedlooptimingadjustmentmode_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PrimaryScramblingCode, create_dissector_handle(dissect_PrimaryScramblingCode_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Synchronisation_Configuration_Cell_SetupRqst, create_dissector_handle(dissect_Synchronisation_Configuration_Cell_SetupRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DL_TPC_Pattern01Count, create_dissector_handle(dissect_DL_TPC_Pattern01Count_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PrimarySCH_Information_Cell_SetupRqstFDD, create_dissector_handle(dissect_PrimarySCH_Information_Cell_SetupRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_SecondarySCH_Information_Cell_SetupRqstFDD, create_dissector_handle(dissect_SecondarySCH_Information_Cell_SetupRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PrimaryCPICH_Information_Cell_SetupRqstFDD, create_dissector_handle(dissect_PrimaryCPICH_Information_Cell_SetupRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_SecondaryCPICH_InformationList_Cell_SetupRqstFDD, create_dissector_handle(dissect_SecondaryCPICH_InformationList_Cell_SetupRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PrimaryCCPCH_Information_Cell_SetupRqstFDD, create_dissector_handle(dissect_PrimaryCCPCH_Information_Cell_SetupRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Limited_power_increase_information_Cell_SetupRqstFDD, create_dissector_handle(dissect_Limited_power_increase_information_Cell_SetupRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_SecondaryCPICH_InformationItem_Cell_SetupRqstFDD, create_dissector_handle(dissect_SecondaryCPICH_InformationItem_Cell_SetupRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CellPortion_InformationItem_Cell_SetupRqstFDD, create_dissector_handle(dissect_CellPortion_InformationItem_Cell_SetupRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UARFCNforNt, create_dissector_handle(dissect_UARFCN_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CellParameterID, create_dissector_handle(dissect_CellParameterID_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_TransmissionDiversityApplied, create_dissector_handle(dissect_TransmissionDiversityApplied_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_SyncCase, create_dissector_handle(dissect_SyncCase_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DPCHConstant, create_dissector_handle(dissect_ConstantValue_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PUSCHConstant, create_dissector_handle(dissect_ConstantValue_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PRACHConstant, create_dissector_handle(dissect_ConstantValue_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_TimingAdvanceApplied, create_dissector_handle(dissect_TimingAdvanceApplied_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_SCH_Information_Cell_SetupRqstTDD, create_dissector_handle(dissect_SCH_Information_Cell_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PCCPCH_Information_Cell_SetupRqstTDD, create_dissector_handle(dissect_PCCPCH_Information_Cell_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_TimeSlotConfigurationList_Cell_SetupRqstTDD, create_dissector_handle(dissect_TimeSlotConfigurationList_Cell_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_SyncCaseIndicatorItem_Cell_SetupRqstTDD_PSCH, create_dissector_handle(dissect_SyncCaseIndicatorItem_Cell_SetupRqstTDD_PSCH_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Synchronisation_Configuration_Cell_ReconfRqst, create_dissector_handle(dissect_Synchronisation_Configuration_Cell_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PrimarySCH_Information_Cell_ReconfRqstFDD, create_dissector_handle(dissect_PrimarySCH_Information_Cell_ReconfRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_SecondarySCH_Information_Cell_ReconfRqstFDD, create_dissector_handle(dissect_SecondarySCH_Information_Cell_ReconfRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PrimaryCPICH_Information_Cell_ReconfRqstFDD, create_dissector_handle(dissect_PrimaryCPICH_Information_Cell_ReconfRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_SecondaryCPICH_InformationList_Cell_ReconfRqstFDD, create_dissector_handle(dissect_SecondaryCPICH_InformationList_Cell_ReconfRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PrimaryCCPCH_Information_Cell_ReconfRqstFDD, create_dissector_handle(dissect_PrimaryCCPCH_Information_Cell_ReconfRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_SecondaryCPICH_InformationItem_Cell_ReconfRqstFDD, create_dissector_handle(dissect_SecondaryCPICH_InformationItem_Cell_ReconfRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CellPortion_InformationItem_Cell_ReconfRqstFDD, create_dissector_handle(dissect_CellPortion_InformationItem_Cell_ReconfRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_SCH_Information_Cell_ReconfRqstTDD, create_dissector_handle(dissect_SCH_Information_Cell_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PCCPCH_Information_Cell_ReconfRqstTDD, create_dissector_handle(dissect_PCCPCH_Information_Cell_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_TimeSlotConfigurationList_Cell_ReconfRqstTDD, create_dissector_handle(dissect_TimeSlotConfigurationList_Cell_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_IndicationType_ResourceStatusInd, create_dissector_handle(dissect_IndicationType_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Local_Cell_InformationItem_ResourceStatusInd, create_dissector_handle(dissect_Local_Cell_InformationItem_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Local_Cell_Group_InformationItem_ResourceStatusInd, create_dissector_handle(dissect_Local_Cell_Group_InformationItem_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Power_Local_Cell_Group_InformationItem_ResourceStatusInd, create_dissector_handle(dissect_Power_Local_Cell_Group_InformationItem_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Local_Cell_InformationItem2_ResourceStatusInd, create_dissector_handle(dissect_Local_Cell_InformationItem2_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Local_Cell_Group_InformationItem2_ResourceStatusInd, create_dissector_handle(dissect_Local_Cell_Group_InformationItem2_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Power_Local_Cell_Group_InformationItem2_ResourceStatusInd, create_dissector_handle(dissect_Power_Local_Cell_Group_InformationItem2_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CCP_InformationItem_ResourceStatusInd, create_dissector_handle(dissect_CCP_InformationItem_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Cell_InformationItem_ResourceStatusInd, create_dissector_handle(dissect_Cell_InformationItem_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_FPACH_LCR_Information, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DwPCH_LCR_Information, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PLCCH_Information_ResourceStatusInd, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_S_CCPCH_768_Information_ResourceStatusInd, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information768_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_BCCH_ModificationTime, create_dissector_handle(dissect_BCCH_ModificationTime_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_MIB_SB_SIB_InformationList_SystemInfoUpdateRqst, create_dissector_handle(dissect_MIB_SB_SIB_InformationList_SystemInfoUpdateRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_SegmentInformationListIE_SystemInfoUpdate, create_dissector_handle(dissect_SegmentInformationListIE_SystemInfoUpdate_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CRNC_CommunicationContextID, create_dissector_handle(dissect_CRNC_CommunicationContextID_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UL_DPCH_Information_RL_SetupRqstFDD, create_dissector_handle(dissect_UL_DPCH_Information_RL_SetupRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DL_DPCH_Information_RL_SetupRqstFDD, create_dissector_handle(dissect_DL_DPCH_Information_RL_SetupRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DCH_FDD_Information, create_dissector_handle(dissect_DCH_FDD_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationList_RL_SetupRqstFDD, create_dissector_handle(dissect_RL_InformationList_RL_SetupRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Transmission_Gap_Pattern_Sequence_Information, create_dissector_handle(dissect_Transmission_Gap_Pattern_Sequence_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Active_Pattern_Sequence_Information, create_dissector_handle(dissect_Active_Pattern_Sequence_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationItem_RL_SetupRqstFDD, create_dissector_handle(dissect_RL_InformationItem_RL_SetupRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UL_CCTrCH_InformationList_RL_SetupRqstTDD, create_dissector_handle(dissect_UL_CCTrCH_InformationList_RL_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DL_CCTrCH_InformationList_RL_SetupRqstTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationList_RL_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DCH_TDD_Information, create_dissector_handle(dissect_DCH_TDD_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DSCH_TDD_Information, create_dissector_handle(dissect_DSCH_TDD_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_USCH_Information, create_dissector_handle(dissect_USCH_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_Information_RL_SetupRqstTDD, create_dissector_handle(dissect_RL_Information_RL_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UL_CCTrCH_InformationItem_RL_SetupRqstTDD, create_dissector_handle(dissect_UL_CCTrCH_InformationItem_RL_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UL_DPCH_InformationList_RL_SetupRqstTDD, create_dissector_handle(dissect_UL_DPCH_InformationItem_RL_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DL_CCTrCH_InformationItem_RL_SetupRqstTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationItem_RL_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DL_DPCH_InformationList_RL_SetupRqstTDD, create_dissector_handle(dissect_DL_DPCH_InformationItem_RL_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_NodeB_CommunicationContextID, create_dissector_handle(dissect_NodeB_CommunicationContextID_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CommunicationControlPortID, create_dissector_handle(dissect_CommunicationControlPortID_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationResponseList_RL_SetupRspFDD, create_dissector_handle(dissect_RL_InformationResponseList_RL_SetupRspFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationResponseItem_RL_SetupRspFDD, create_dissector_handle(dissect_RL_InformationResponseItem_RL_SetupRspFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationResponse_RL_SetupRspTDD, create_dissector_handle(dissect_RL_InformationResponse_RL_SetupRspTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DCH_InformationResponse, create_dissector_handle(dissect_DCH_InformationResponse_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DSCH_InformationResponse, create_dissector_handle(dissect_DSCH_InformationResponse_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_USCH_InformationResponse, create_dissector_handle(dissect_USCH_InformationResponse_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CauseLevel_RL_SetupFailureFDD, create_dissector_handle(dissect_CauseLevel_RL_SetupFailureFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Unsuccessful_RL_InformationRespItem_RL_SetupFailureFDD, create_dissector_handle(dissect_Unsuccessful_RL_InformationRespItem_RL_SetupFailureFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Successful_RL_InformationRespItem_RL_SetupFailureFDD, create_dissector_handle(dissect_Successful_RL_InformationRespItem_RL_SetupFailureFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CauseLevel_RL_SetupFailureTDD, create_dissector_handle(dissect_CauseLevel_RL_SetupFailureTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Unsuccessful_RL_InformationResp_RL_SetupFailureTDD, create_dissector_handle(dissect_Unsuccessful_RL_InformationResp_RL_SetupFailureTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Compressed_Mode_Deactivation_Flag, create_dissector_handle(dissect_Compressed_Mode_Deactivation_Flag_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationList_RL_AdditionRqstFDD, create_dissector_handle(dissect_RL_InformationList_RL_AdditionRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationItem_RL_AdditionRqstFDD, create_dissector_handle(dissect_RL_InformationItem_RL_AdditionRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UL_CCTrCH_InformationList_RL_AdditionRqstTDD, create_dissector_handle(dissect_UL_CCTrCH_InformationList_RL_AdditionRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DL_CCTrCH_InformationList_RL_AdditionRqstTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationList_RL_AdditionRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_Information_RL_AdditionRqstTDD, create_dissector_handle(dissect_RL_Information_RL_AdditionRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UL_DPCH_InformationItem_RL_AdditionRqstTDD, create_dissector_handle(dissect_UL_DPCH_InformationItem_RL_AdditionRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DL_DPCH_InformationItem_RL_AdditionRqstTDD, create_dissector_handle(dissect_DL_DPCH_InformationItem_RL_AdditionRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationResponseList_RL_AdditionRspFDD, create_dissector_handle(dissect_RL_InformationResponseList_RL_AdditionRspFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationResponseItem_RL_AdditionRspFDD, create_dissector_handle(dissect_RL_InformationResponseItem_RL_AdditionRspFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationResponse_RL_AdditionRspTDD, create_dissector_handle(dissect_RL_InformationResponse_RL_AdditionRspTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CauseLevel_RL_AdditionFailureFDD, create_dissector_handle(dissect_CauseLevel_RL_AdditionFailureFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Unsuccessful_RL_InformationRespItem_RL_AdditionFailureFDD, create_dissector_handle(dissect_Unsuccessful_RL_InformationRespItem_RL_AdditionFailureFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Successful_RL_InformationRespItem_RL_AdditionFailureFDD, create_dissector_handle(dissect_Successful_RL_InformationRespItem_RL_AdditionFailureFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CauseLevel_RL_AdditionFailureTDD, create_dissector_handle(dissect_CauseLevel_RL_AdditionFailureTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Unsuccessful_RL_InformationResp_RL_AdditionFailureTDD, create_dissector_handle(dissect_Unsuccessful_RL_InformationResp_RL_AdditionFailureTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UL_DPCH_Information_RL_ReconfPrepFDD, create_dissector_handle(dissect_UL_DPCH_Information_RL_ReconfPrepFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DL_DPCH_Information_RL_ReconfPrepFDD, create_dissector_handle(dissect_DL_DPCH_Information_RL_ReconfPrepFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_FDD_DCHs_to_Modify, create_dissector_handle(dissect_FDD_DCHs_to_Modify_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DCHs_to_Add_FDD, create_dissector_handle(dissect_DCH_FDD_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DCH_DeleteList_RL_ReconfPrepFDD, create_dissector_handle(dissect_DCH_DeleteList_RL_ReconfPrepFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationList_RL_ReconfPrepFDD, create_dissector_handle(dissect_RL_InformationList_RL_ReconfPrepFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationItem_RL_ReconfPrepFDD, create_dissector_handle(dissect_RL_InformationItem_RL_ReconfPrepFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD, create_dissector_handle(dissect_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD, create_dissector_handle(dissect_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD, create_dissector_handle(dissect_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_TDD_DCHs_to_Modify, create_dissector_handle(dissect_TDD_DCHs_to_Modify_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DCHs_to_Add_TDD, create_dissector_handle(dissect_DCH_TDD_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DCH_DeleteList_RL_ReconfPrepTDD, create_dissector_handle(dissect_DCH_DeleteList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DSCH_Information_ModifyList_RL_ReconfPrepTDD, create_dissector_handle(dissect_DSCH_Information_ModifyList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DSCHs_to_Add_TDD, create_dissector_handle(dissect_DSCH_TDD_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DSCH_Information_DeleteList_RL_ReconfPrepTDD, create_dissector_handle(dissect_DSCH_Information_DeleteList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_USCH_Information_ModifyList_RL_ReconfPrepTDD, create_dissector_handle(dissect_USCH_Information_ModifyList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_USCH_Information_Add, create_dissector_handle(dissect_USCH_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_USCH_Information_DeleteList_RL_ReconfPrepTDD, create_dissector_handle(dissect_USCH_Information_DeleteList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_Information_RL_ReconfPrepTDD, create_dissector_handle(dissect_RL_Information_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UL_DPCH_InformationAddListIE_RL_ReconfPrepTDD, create_dissector_handle(dissect_UL_DPCH_InformationAddItem_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UL_DPCH_InformationModify_AddListIE_RL_ReconfPrepTDD, create_dissector_handle(dissect_UL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UL_DPCH_InformationModify_ModifyListIE_RL_ReconfPrepTDD, create_dissector_handle(dissect_UL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD, create_dissector_handle(dissect_UL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DL_DPCH_InformationAddListIE_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_DPCH_InformationAddItem_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DL_DPCH_InformationModify_AddListIE_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_DPCH_InformationModify_AddItem_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DL_DPCH_InformationModify_ModifyListIE_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_DPCH_InformationModify_ModifyItem_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_DPCH_InformationModify_DeleteListIE_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationResponseList_RL_ReconfReady, create_dissector_handle(dissect_RL_InformationResponseList_RL_ReconfReady_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationResponseItem_RL_ReconfReady, create_dissector_handle(dissect_RL_InformationResponseItem_RL_ReconfReady_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CauseLevel_RL_ReconfFailure, create_dissector_handle(dissect_CauseLevel_RL_ReconfFailure_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_ReconfigurationFailureItem_RL_ReconfFailure, create_dissector_handle(dissect_RL_ReconfigurationFailureItem_RL_ReconfFailure_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CFN, create_dissector_handle(dissect_CFN_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UL_DPCH_Information_RL_ReconfRqstFDD, create_dissector_handle(dissect_UL_DPCH_Information_RL_ReconfRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DL_DPCH_Information_RL_ReconfRqstFDD, create_dissector_handle(dissect_DL_DPCH_Information_RL_ReconfRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DCH_DeleteList_RL_ReconfRqstFDD, create_dissector_handle(dissect_DCH_DeleteList_RL_ReconfRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationList_RL_ReconfRqstFDD, create_dissector_handle(dissect_RL_InformationList_RL_ReconfRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationItem_RL_ReconfRqstFDD, create_dissector_handle(dissect_RL_InformationItem_RL_ReconfRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD, create_dissector_handle(dissect_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD, create_dissector_handle(dissect_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DCH_DeleteList_RL_ReconfRqstTDD, create_dissector_handle(dissect_DCH_DeleteList_RL_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_Information_RL_ReconfRqstTDD, create_dissector_handle(dissect_RL_Information_RL_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD, create_dissector_handle(dissect_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD, create_dissector_handle(dissect_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationResponseList_RL_ReconfRsp, create_dissector_handle(dissect_RL_InformationResponseList_RL_ReconfRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationResponseItem_RL_ReconfRsp, create_dissector_handle(dissect_RL_InformationResponseItem_RL_ReconfRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_informationList_RL_DeletionRqst, create_dissector_handle(dissect_RL_informationList_RL_DeletionRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_informationItem_RL_DeletionRqst, create_dissector_handle(dissect_RL_informationItem_RL_DeletionRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PowerAdjustmentType, create_dissector_handle(dissect_PowerAdjustmentType_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DLReferencePower, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_InnerLoopDLPCStatus, create_dissector_handle(dissect_InnerLoopDLPCStatus_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DLReferencePowerList_DL_PC_Rqst, create_dissector_handle(dissect_DL_ReferencePowerInformationList_DL_PC_Rqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_MaxAdjustmentStep, create_dissector_handle(dissect_MaxAdjustmentStep_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_AdjustmentPeriod, create_dissector_handle(dissect_AdjustmentPeriod_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_AdjustmentRatio, create_dissector_handle(dissect_ScaledAdjustmentRatio_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DL_ReferencePowerInformationItem_DL_PC_Rqst, create_dissector_handle(dissect_DL_ReferencePowerInformationItem_DL_PC_Rqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_TimeslotISCPInfo, create_dissector_handle(dissect_DL_TimeslotISCPInfo_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DedicatedMeasurementObjectType_DM_Rqst, create_dissector_handle(dissect_DedicatedMeasurementObjectType_DM_Rqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DedicatedMeasurementType, create_dissector_handle(dissect_DedicatedMeasurementType_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CFNReportingIndicator, create_dissector_handle(dissect_FNReportingIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationItem_DM_Rqst, create_dissector_handle(dissect_RL_InformationItem_DM_Rqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DedicatedMeasurementObjectType_DM_Rsp, create_dissector_handle(dissect_DedicatedMeasurementObjectType_DM_Rsp_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationItem_DM_Rsp, create_dissector_handle(dissect_RL_InformationItem_DM_Rsp_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_Set_InformationItem_DM_Rsp, create_dissector_handle(dissect_RL_Set_InformationItem_DM_Rsp_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DedicatedMeasurementObjectType_DM_Rprt, create_dissector_handle(dissect_DedicatedMeasurementObjectType_DM_Rprt_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationItem_DM_Rprt, create_dissector_handle(dissect_RL_InformationItem_DM_Rprt_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_Set_InformationItem_DM_Rprt, create_dissector_handle(dissect_RL_Set_InformationItem_DM_Rprt_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Reporting_Object_RL_FailureInd, create_dissector_handle(dissect_Reporting_Object_RL_FailureInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationItem_RL_FailureInd, create_dissector_handle(dissect_RL_InformationItem_RL_FailureInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_Set_InformationItem_RL_FailureInd, create_dissector_handle(dissect_RL_Set_InformationItem_RL_FailureInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CCTrCH_InformationItem_RL_FailureInd, create_dissector_handle(dissect_CCTrCH_InformationItem_RL_FailureInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationList_RL_PreemptRequiredInd, create_dissector_handle(dissect_RL_InformationList_RL_PreemptRequiredInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationItem_RL_PreemptRequiredInd, create_dissector_handle(dissect_RL_InformationItem_RL_PreemptRequiredInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Reporting_Object_RL_RestoreInd, create_dissector_handle(dissect_Reporting_Object_RL_RestoreInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_InformationItem_RL_RestoreInd, create_dissector_handle(dissect_RL_InformationItem_RL_RestoreInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_RL_Set_InformationItem_RL_RestoreInd, create_dissector_handle(dissect_RL_Set_InformationItem_RL_RestoreInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CCTrCH_InformationItem_RL_RestoreInd, create_dissector_handle(dissect_CCTrCH_InformationItem_RL_RestoreInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_HS_PDSCH_HS_SCCH_E_AGCH_E_RGCH_E_HICH_MaxPower_PSCH_ReconfRqst, create_dissector_handle(dissect_MaximumTransmissionPower_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_HS_PDSCH_HS_SCCH_ScramblingCode_PSCH_ReconfRqst, create_dissector_handle(dissect_DL_ScramblingCode_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_HS_PDSCH_FDD_Code_Information_PSCH_ReconfRqst, create_dissector_handle(dissect_HS_PDSCH_FDD_Code_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_HS_SCCH_FDD_Code_Information_PSCH_ReconfRqst, create_dissector_handle(dissect_HS_SCCH_FDD_Code_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PDSCHSets_AddList_PSCH_ReconfRqst, create_dissector_handle(dissect_PDSCHSets_AddList_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PDSCHSets_ModifyList_PSCH_ReconfRqst, create_dissector_handle(dissect_PDSCHSets_ModifyList_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PDSCHSets_DeleteList_PSCH_ReconfRqst, create_dissector_handle(dissect_PDSCHSets_DeleteList_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PUSCHSets_AddList_PSCH_ReconfRqst, create_dissector_handle(dissect_PUSCHSets_AddList_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PUSCHSets_ModifyList_PSCH_ReconfRqst, create_dissector_handle(dissect_PUSCHSets_ModifyList_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PUSCHSets_DeleteList_PSCH_ReconfRqst, create_dissector_handle(dissect_PUSCHSets_DeleteList_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PDSCH_Information_AddListIE_PSCH_ReconfRqst, create_dissector_handle(dissect_PDSCH_Information_AddItem_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PDSCH_Information_ModifyListIE_PSCH_ReconfRqst, create_dissector_handle(dissect_PDSCH_Information_ModifyItem_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PDSCH_ModifyInformation_LCR_PSCH_ReconfRqst, create_dissector_handle(dissect_PDSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PUSCH_Information_AddListIE_PSCH_ReconfRqst, create_dissector_handle(dissect_PUSCH_Information_AddItem_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PUSCH_Information_ModifyListIE_PSCH_ReconfRqst, create_dissector_handle(dissect_PUSCH_Information_ModifyItem_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PUSCH_ModifyInformation_LCR_PSCH_ReconfRqst, create_dissector_handle(dissect_PUSCH_ModifyInformation_LCR_ModifyItem_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CauseLevel_PSCH_ReconfFailure, create_dissector_handle(dissect_CauseLevel_PSCH_ReconfFailure_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Unsuccessful_PDSCHSetItem_PSCH_ReconfFailureTDD, create_dissector_handle(dissect_Unsuccessful_PDSCHSetItem_PSCH_ReconfFailureTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Unsuccessful_PUSCHSetItem_PSCH_ReconfFailureTDD, create_dissector_handle(dissect_Unsuccessful_PUSCHSetItem_PSCH_ReconfFailureTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_ResetIndicator, create_dissector_handle(dissect_ResetIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CommunicationContextInfoItem_Reset, create_dissector_handle(dissect_CommunicationContextInfoItem_Reset_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CommunicationControlPortInfoItem_Reset, create_dissector_handle(dissect_CommunicationControlPortInfoItem_Reset_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_InformationExchangeID, create_dissector_handle(dissect_InformationExchangeID_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_InformationExchangeObjectType_InfEx_Rqst, create_dissector_handle(dissect_InformationExchangeObjectType_InfEx_Rqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_InformationType, create_dissector_handle(dissect_InformationType_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_InformationReportCharacteristics, create_dissector_handle(dissect_InformationReportCharacteristics_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_InformationExchangeObjectType_InfEx_Rsp, create_dissector_handle(dissect_InformationExchangeObjectType_InfEx_Rsp_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_InformationExchangeObjectType_InfEx_Rprt, create_dissector_handle(dissect_InformationExchangeObjectType_InfEx_Rprt_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_cellSyncBurstRepetitionPeriod, create_dissector_handle(dissect_CellSyncBurstRepetitionPeriod_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_timeslotInfo_CellSyncInitiationRqstTDD, create_dissector_handle(dissect_TimeslotInfo_CellSyncInitiationRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CellSyncBurstTransInit_CellSyncInitiationRqstTDD, create_dissector_handle(dissect_CellSyncBurstTransInit_CellSyncInitiationRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CellSyncBurstMeasureInit_CellSyncInitiationRqstTDD, create_dissector_handle(dissect_CellSyncBurstMeasureInit_CellSyncInitiationRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_TimeSlot, create_dissector_handle(dissect_TimeSlot_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_NCyclesPerSFNperiod, create_dissector_handle(dissect_NCyclesPerSFNperiod_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_NRepetitionsPerCyclePeriod, create_dissector_handle(dissect_NRepetitionsPerCyclePeriod_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CellSyncBurstTransReconfInfo_CellSyncReconfRqstTDD, create_dissector_handle(dissect_CellSyncBurstTransReconfInfo_CellSyncReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CellSyncBurstMeasReconfiguration_CellSyncReconfRqstTDD, create_dissector_handle(dissect_CellSyncBurstMeasInfo_CellSyncReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CellSyncBurstMeasInfoList_CellSyncReconfRqstTDD, create_dissector_handle(dissect_CellSyncBurstMeasInfoListIE_CellSyncReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_SynchronisationReportType, create_dissector_handle(dissect_SynchronisationReportType_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_SynchronisationReportCharacteristics, create_dissector_handle(dissect_SynchronisationReportCharacteristics_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CellAdjustmentInfo_SyncAdjustmntRqstTDD, create_dissector_handle(dissect_CellAdjustmentInfo_SyncAdjustmentRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CellAdjustmentInfoItem_SyncAdjustmentRqstTDD, create_dissector_handle(dissect_CellAdjustmentInfoItem_SyncAdjustmentRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CauseLevel_SyncAdjustmntFailureTDD, create_dissector_handle(dissect_CauseLevel_SyncAdjustmntFailureTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Unsuccessful_cell_InformationRespItem_SyncAdjustmntFailureTDD, create_dissector_handle(dissect_Unsuccessful_cell_InformationRespItem_SyncAdjustmntFailureTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CSBTransmissionID, create_dissector_handle(dissect_CSBTransmissionID_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CSBMeasurementID, create_dissector_handle(dissect_CSBMeasurementID_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_CellSyncInfo_CellSyncReprtTDD, create_dissector_handle(dissect_CellSyncInfo_CellSyncReprtTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_SyncReportType_CellSyncReprtTDD, create_dissector_handle(dissect_SyncReportType_CellSyncReprtTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_SignallingBearerRequestIndicator, create_dissector_handle(dissect_SignallingBearerRequestIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DCH_RearrangeList_Bearer_RearrangeInd, create_dissector_handle(dissect_DCH_RearrangeList_Bearer_RearrangeInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DSCH_RearrangeList_Bearer_RearrangeInd, create_dissector_handle(dissect_DSCH_RearrangeList_Bearer_RearrangeInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_USCH_RearrangeList_Bearer_RearrangeInd, create_dissector_handle(dissect_USCH_RearrangeList_Bearer_RearrangeInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_HSDSCH_RearrangeList_Bearer_RearrangeInd, create_dissector_handle(dissect_HSDSCH_RearrangeList_Bearer_RearrangeInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DelayedActivationList_RL_ActivationCmdFDD, create_dissector_handle(dissect_DelayedActivationInformationList_RL_ActivationCmdFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DelayedActivationInformation_RL_ActivationCmdFDD, create_dissector_handle(dissect_DelayedActivationInformation_RL_ActivationCmdFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DelayedActivationList_RL_ActivationCmdTDD, create_dissector_handle(dissect_DelayedActivationInformationList_RL_ActivationCmdTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_DelayedActivationInformation_RL_ActivationCmdTDD, create_dissector_handle(dissect_DelayedActivationInformation_RL_ActivationCmdTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_HSDSCH_FDD_Update_Information, create_dissector_handle(dissect_HSDSCH_FDD_Update_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_HSDSCH_TDD_Update_Information, create_dissector_handle(dissect_HSDSCH_TDD_Update_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Modification_Period, create_dissector_handle(dissect_Modification_Period_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_MICH_CFN, create_dissector_handle(dissect_MICH_CFN_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_NI_Information_NotifUpdateCmd, create_dissector_handle(dissect_NI_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Cell_Frequency_List_InformationItem_LCR_MulFreq_AuditRsp, create_dissector_handle(dissect_Cell_Frequency_List_InformationItem_LCR_MulFreq_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Cell_Frequency_List_InformationItem_LCR_MulFreq_ResourceStatusInd, create_dissector_handle(dissect_Cell_Frequency_List_InformationItem_LCR_MulFreq_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UPPCH_LCR_InformationItem_AuditRsp, create_dissector_handle(dissect_UPPCH_LCR_InformationItem_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UPPCH_LCR_InformationItem_ResourceStatusInd, create_dissector_handle(dissect_UPPCH_LCR_InformationItem_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Unsuccessful_UARFCNItem_PSCH_ReconfFailureTDD, create_dissector_handle(dissect_Unsuccessful_UARFCNItem_PSCH_ReconfFailureTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UARFCNSpecificCauseList, create_dissector_handle(dissect_UARFCNSpecificCauseList_PSCH_ReconfFailureTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst, create_dissector_handle(dissect_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_MultipleFreq_E_PUCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst, create_dissector_handle(dissect_MultipleFreq_E_PUCH_Timeslot_Information_LCRItem_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_MultipleFreq_E_HICH_TimeOffsetLCR, create_dissector_handle(dissect_MultipleFreq_E_HICH_TimeOffsetLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_PLCCH_parameters, create_dissector_handle(dissect_PLCCH_parameters_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_E_RUCCH_parameters, create_dissector_handle(dissect_E_RUCCH_parameters_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_E_RUCCH_768_parameters, create_dissector_handle(dissect_E_RUCCH_768_parameters_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Cell_ERNTI_Status_Information, create_dissector_handle(dissect_Cell_ERNTI_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_ERACH_CM_Rqst, create_dissector_handle(dissect_ERACH_CM_Rqst_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_ERACH_CM_Rsp, create_dissector_handle(dissect_ERACH_CM_Rsp_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_ERACH_CM_Rprt, create_dissector_handle(dissect_ERACH_CM_Rprt_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_EDCH_RACH_Report_Value, create_dissector_handle(dissect_EDCH_RACH_Report_Value_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_EDCH_RACH_Report_IncrDecrThres, create_dissector_handle(dissect_EDCH_RACH_Report_IncrDecrThres_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_EDCH_RACH_Report_ThresholdInformation, create_dissector_handle(dissect_EDCH_RACH_Report_ThresholdInformation_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_GANSS_alm_keplerianNAVAlmanac, create_dissector_handle(dissect_GANSS_ALM_NAVKeplerianSet_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_GANSS_alm_keplerianReducedAlmanac, create_dissector_handle(dissect_GANSS_ALM_ReducedKeplerianSet_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_GANSS_alm_keplerianMidiAlmanac, create_dissector_handle(dissect_GANSS_ALM_MidiAlmanacSet_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_GANSS_alm_keplerianGLONASS, create_dissector_handle(dissect_GANSS_ALM_GlonassAlmanacSet_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_GANSS_alm_ecefSBASAlmanac, create_dissector_handle(dissect_GANSS_ALM_ECEFsbasAlmanacSet_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_ULTimeslotISCPValue_For_CellPortion, create_dissector_handle(dissect_UL_TimeslotISCP_Value_IncrDecrThres_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UpPTSInterferenceValue_For_CellPortion, create_dissector_handle(dissect_UpPTSInterferenceValue_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Best_Cell_Portions_ValueLCR, create_dissector_handle(dissect_Best_Cell_Portions_ValueLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Transmitted_Carrier_Power_For_CellPortion_ValueLCR, create_dissector_handle(dissect_Transmitted_Carrier_Power_For_CellPortion_ValueLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Received_total_wide_band_power_For_CellPortion_ValueLCR, create_dissector_handle(dissect_Received_total_wide_band_power_For_CellPortion_ValueLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UL_TimeslotISCP_For_CellPortion_Value, create_dissector_handle(dissect_UL_TimeslotISCP_For_CellPortion_Value_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR, create_dissector_handle(dissect_HS_DSCHRequiredPowerValueInformation_For_CellPortionLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR, create_dissector_handle(dissect_HS_DSCHProvidedBitRateValueInformation_For_CellPortionLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_E_DCHProvidedBitRateValueInformation_For_CellPortion, create_dissector_handle(dissect_E_DCHProvidedBitRateValueInformation_For_CellPortion_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_UpPTSInterference_For_CellPortion_Value, create_dissector_handle(dissect_UpPTSInterference_For_CellPortion_Value_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue, create_dissector_handle(dissect_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortionValue_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_TransmittedCarrierPowerOfAllCodesNotUsedForHS_PDSCH_HS_SCCH_E_AGCHOrE_HICHTransmissionCellPortion, create_dissector_handle(dissect_TransmittedCarrierPowerOfAllCodesNotUsedForHSTransmissionValue_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_ActivationInformation, create_dissector_handle(dissect_ActivationInformation_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_BroadcastCommonTransportBearerIndication, create_dissector_handle(dissect_BroadcastCommonTransportBearerIndication_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MessageStructure, create_dissector_handle(dissect_MessageStructure_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TypeOfError, create_dissector_handle(dissect_TypeOfError_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TnlQos, create_dissector_handle(dissect_TnlQos_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Unidirectional_DCH_Indicator, create_dissector_handle(dissect_Unidirectional_DCH_Indicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_ExtendedPropagationDelay, create_dissector_handle(dissect_ExtendedPropagationDelay_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Initial_DL_Power_TimeslotLCR_InformationItem, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Maximum_DL_Power_TimeslotLCR_InformationItem, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Minimum_DL_Power_TimeslotLCR_InformationItem, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_bindingID, create_dissector_handle(dissect_BindingID_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_transportlayeraddress, create_dissector_handle(dissect_TransportLayerAddress_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_PowerOffset_for_SchedulingInfo, create_dissector_handle(dissect_E_DCH_PowerOffset_for_SchedulingInfo_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_SAT_Info_Almanac_ExtItem, create_dissector_handle(dissect_SAT_Info_Almanac_ExtList_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HARQ_Preamble_Mode, create_dissector_handle(dissect_HARQ_Preamble_Mode_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_MACdPDUSizeFormat, create_dissector_handle(dissect_HSDSCH_MACdPDUSizeFormat_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSSICH_SIRTarget, create_dissector_handle(dissect_UL_SIR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSSICH_TPC_StepSize, create_dissector_handle(dissect_TDD_TPC_UplinkStepSize_LCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_ueCapability_Info, create_dissector_handle(dissect_UE_Capability_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HS_PDSCH_Code_Change_Grant, create_dissector_handle(dissect_HS_PDSCH_Code_Change_Grant_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HARQ_Preamble_Mode_Activation_Indicator, create_dissector_handle(dissect_HARQ_Preamble_Mode_Activation_Indicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_hsSCCH_Specific_Information_ResponseTDD768, create_dissector_handle(dissect_HSSCCH_Specific_InformationRespListTDD768_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MaximumMACdPDU_SizeExtended, create_dissector_handle(dissect_MAC_PDU_SizeExtended_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_failed_HS_SICH, create_dissector_handle(dissect_HS_SICH_failed_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_missed_HS_SICH, create_dissector_handle(dissect_HS_SICH_missed_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_total_HS_SICH, create_dissector_handle(dissect_HS_SICH_total_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_ContinuousPacketConnectivityHS_SCCH_less_Information, create_dissector_handle(dissect_ContinuousPacketConnectivityHS_SCCH_less_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_ContinuousPacketConnectivityHS_SCCH_less_Information_Response, create_dissector_handle(dissect_ContinuousPacketConnectivityHS_SCCH_less_Information_Response_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HS_PDSCH_Code_Change_Indicator, create_dissector_handle(dissect_HS_PDSCH_Code_Change_Indicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_GANSS_Information, create_dissector_handle(dissect_GANSS_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_GANSS_Common_Data, create_dissector_handle(dissect_GANSS_Common_Data_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_GANSS_Generic_Data, create_dissector_handle(dissect_GANSS_Generic_Data_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_SyncDLCodeIdThreInfoLCR, create_dissector_handle(dissect_SyncDLCodeIdThreInfoLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Extended_RNC_ID, create_dissector_handle(dissect_Extended_RNC_ID_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_LCRTDD_uplink_Physical_Channel_Capability, create_dissector_handle(dissect_LCRTDD_Uplink_Physical_Channel_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PLCCH_Information_UL_TimeslotLCR_Info, create_dissector_handle(dissect_PLCCHinformation_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MICH_Parameters_CTCH_SetupRqstFDD, create_dissector_handle(dissect_MICH_Parameters_CTCH_SetupRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_FDD_S_CCPCH_FrameOffset_CTCH_SetupRqstFDD, create_dissector_handle(dissect_FDD_S_CCPCH_FrameOffset_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_ModulationPO_MBSFN, create_dissector_handle(dissect_ModulationPO_MBSFN_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Secondary_CCPCH_SlotFormat_Extended, create_dissector_handle(dissect_Secondary_CCPCH_SlotFormat_Extended_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_BroadcastReference, create_dissector_handle(dissect_BroadcastReference_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Tstd_indicator, create_dissector_handle(dissect_TSTD_Indicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MICH_Parameters_CTCH_SetupRqstTDD, create_dissector_handle(dissect_MICH_Parameters_CTCH_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_S_CCPCH_Parameters_CTCH_SetupRqstTDD, create_dissector_handle(dissect_Secondary_CCPCH_parameterExtendedList_CTCH_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_S_CCPCH_LCR_Parameters_CTCH_SetupRqstTDD, create_dissector_handle(dissect_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_S_CCPCH_768_Parameters_CTCH_SetupRqstTDD, create_dissector_handle(dissect_Secondary_CCPCH_768_parameterList_CTCH_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_S_CCPCH_Modulation, create_dissector_handle(dissect_ModulationMBSFN_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_tFCI_Presence, create_dissector_handle(dissect_TFCI_Presence_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_maxFACH_Power_LCR_CTCH_SetupRqstTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PCH_Power_LCR_CTCH_SetupRqstTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PICH_768_Parameters_CTCH_SetupRqstTDD, create_dissector_handle(dissect_PICH_768_ParametersItem_CTCH_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_FPACH_LCR_Parameters_CTCH_SetupRqstTDD, create_dissector_handle(dissect_FPACH_LCR_Parameters_CTCH_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PRACH_768_Parameters_CTCH_SetupRqstTDD, create_dissector_handle(dissect_PRACH_768_ParametersItem_CTCH_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MICH_Parameters_CTCH_ReconfRqstFDD, create_dissector_handle(dissect_MICH_Parameters_CTCH_ReconfRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_FPACH_LCR_Parameters_CTCH_ReconfRqstTDD, create_dissector_handle(dissect_FPACH_LCR_Parameters_CTCH_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MICH_Parameters_CTCH_ReconfRqstTDD, create_dissector_handle(dissect_MICH_Parameters_CTCH_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PLCCH_Parameters_CTCH_ReconfRqstTDD, create_dissector_handle(dissect_PLCCH_Parameters_CTCH_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_S_CCPCH_768_Parameters_CTCH_ReconfRqstTDD, create_dissector_handle(dissect_Secondary_CCPCH_768_Parameters_CTCH_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PICH_768_Parameters_CTCH_ReconfRqstTDD, create_dissector_handle(dissect_PICH_768_Parameters_CTCH_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MICH_768_Parameters_CTCH_ReconfRqstTDD, create_dissector_handle(dissect_MICH_768_Parameters_CTCH_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_S_CCPCH_Parameters_CTCH_ReconfRqstTDD, create_dissector_handle(dissect_Secondary_CCPCH_parameterExtendedList_CTCH_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_S_CCPCH_LCR_Parameters_CTCH_ReconfRqstTDD, create_dissector_handle(dissect_Secondary_CCPCH_LCR_parameterExtendedList_CTCH_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_maxFACH_Power_LCR_CTCH_ReconfRqstTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PCH_Power_LCR_CTCH_ReconfRqstTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_CommonPhysicalChannelID768_CommonTrChDeletionReq, create_dissector_handle(dissect_CommonPhysicalChannelID768_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Power_Local_Cell_Group_InformationList_AuditRsp, create_dissector_handle(dissect_Power_Local_Cell_Group_InformationList_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_FPACH_LCR_InformationList_AuditRsp, create_dissector_handle(dissect_FPACH_LCR_InformationList_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DwPCH_LCR_InformationList_AuditRsp, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_Resources_Information_AuditRsp, create_dissector_handle(dissect_HS_DSCH_Resources_Information_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MICH_Information_AuditRsp, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_S_CCPCH_InformationListExt_AuditRsp, create_dissector_handle(dissect_S_CCPCH_InformationListExt_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_S_CCPCH_LCR_InformationListExt_AuditRsp, create_dissector_handle(dissect_S_CCPCH_LCR_InformationListExt_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_Resources_Information_AuditRsp, create_dissector_handle(dissect_E_DCH_Resources_Information_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PLCCH_InformationList_AuditRsp, create_dissector_handle(dissect_PLCCH_InformationList_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_P_CCPCH_768_Information_AuditRsp, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information768_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_S_CCPCH_768_InformationList_AuditRsp, create_dissector_handle(dissect_S_CCPCH_768_InformationList_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PICH_768_Information_AuditRsp, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information768_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PRACH_768_InformationList_AuditRsp, create_dissector_handle(dissect_PRACH_768_InformationList_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_SCH_768_Information_AuditRsp, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information768_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MICH_768_Information_AuditRsp, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information768_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_RUCCH_InformationList_AuditRsp, create_dissector_handle(dissect_E_RUCCH_InformationList_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_RUCCH_768_InformationList_AuditRsp, create_dissector_handle(dissect_E_RUCCH_768_InformationList_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_ReferenceClockAvailability, create_dissector_handle(dissect_ReferenceClockAvailability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Power_Local_Cell_Group_ID, create_dissector_handle(dissect_Local_Cell_ID_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDPA_Capability, create_dissector_handle(dissect_HSDPA_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_Capability, create_dissector_handle(dissect_E_DCH_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_TTI2ms_Capability, create_dissector_handle(dissect_E_DCH_TTI2ms_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_SF_Capability, create_dissector_handle(dissect_E_DCH_SF_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_HARQ_Combining_Capability, create_dissector_handle(dissect_E_DCH_HARQ_Combining_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_CapacityConsumptionLaw, create_dissector_handle(dissect_E_DCHCapacityConsumptionLaw_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_F_DPCH_Capability, create_dissector_handle(dissect_F_DPCH_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_TDD_CapacityConsumptionLaw, create_dissector_handle(dissect_E_DCH_TDD_CapacityConsumptionLaw_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_ContinuousPacketConnectivityDTX_DRX_Capability, create_dissector_handle(dissect_ContinuousPacketConnectivityDTX_DRX_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Max_UE_DTX_Cycle, create_dissector_handle(dissect_Max_UE_DTX_Cycle_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_ContinuousPacketConnectivityHS_SCCH_less_Capability, create_dissector_handle(dissect_ContinuousPacketConnectivityHS_SCCH_less_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MIMO_Capability, create_dissector_handle(dissect_MIMO_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_SixtyfourQAM_DL_Capability, create_dissector_handle(dissect_SixtyfourQAM_DL_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MBMS_Capability, create_dissector_handle(dissect_MBMS_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Enhanced_FACH_Capability, create_dissector_handle(dissect_Enhanced_FACH_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Enhanced_PCH_Capability, create_dissector_handle(dissect_Enhanced_PCH_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_SixteenQAM_UL_Capability, create_dissector_handle(dissect_SixteenQAM_UL_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_MACdPDU_SizeCapability, create_dissector_handle(dissect_HSDSCH_MACdPDU_SizeCapability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_F_DPCH_SlotFormatCapability, create_dissector_handle(dissect_F_DPCH_SlotFormatCapability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_CommonMeasurementAccuracy, create_dissector_handle(dissect_CommonMeasurementAccuracy_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MeasurementRecoveryBehavior, create_dissector_handle(dissect_MeasurementRecoveryBehavior_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_RTWP_ReportingIndicator, create_dissector_handle(dissect_RTWP_ReportingIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_RTWP_CellPortion_ReportingIndicator, create_dissector_handle(dissect_RTWP_CellPortion_ReportingIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Reference_ReceivedTotalWideBandPowerReporting, create_dissector_handle(dissect_Reference_ReceivedTotalWideBandPowerReporting_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TimeSlotLCR_CM_Rqst, create_dissector_handle(dissect_TimeSlotLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_NeighbouringCellMeasurementInformation, create_dissector_handle(dissect_NeighbouringCellMeasurementInformation_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MeasurementRecoverySupportIndicator, create_dissector_handle(dissect_MeasurementRecoverySupportIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Reference_ReceivedTotalWideBandPowerSupportIndicator, create_dissector_handle(dissect_Reference_ReceivedTotalWideBandPowerSupportIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Reference_ReceivedTotalWideBandPower, create_dissector_handle(dissect_Reference_ReceivedTotalWideBandPower_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MeasurementRecoveryReportingIndicator, create_dissector_handle(dissect_MeasurementRecoveryReportingIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_IPDLParameter_Information_Cell_SetupRqstFDD, create_dissector_handle(dissect_IPDLParameter_Information_Cell_SetupRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_CellPortion_InformationList_Cell_SetupRqstFDD, create_dissector_handle(dissect_CellPortion_InformationList_Cell_SetupRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MIMO_PilotConfiguration, create_dissector_handle(dissect_MIMO_PilotConfiguration_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TimeSlotConfigurationList_LCR_Cell_SetupRqstTDD, create_dissector_handle(dissect_TimeSlotConfigurationList_LCR_Cell_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PCCPCH_LCR_Information_Cell_SetupRqstTDD, create_dissector_handle(dissect_PCCPCH_LCR_Information_Cell_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DwPCH_LCR_Information_Cell_SetupRqstTDD, create_dissector_handle(dissect_DwPCH_LCR_Information_Cell_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_ReferenceSFNoffset, create_dissector_handle(dissect_ReferenceSFNoffset_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_IPDLParameter_Information_Cell_SetupRqstTDD, create_dissector_handle(dissect_IPDLParameter_Information_Cell_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_IPDLParameter_Information_LCR_Cell_SetupRqstTDD, create_dissector_handle(dissect_IPDLParameter_Information_LCR_Cell_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PCCPCH_768_Information_Cell_SetupRqstTDD, create_dissector_handle(dissect_PCCPCH_768_Information_Cell_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_SCH_768_Information_Cell_SetupRqstTDD, create_dissector_handle(dissect_SCH_768_Information_Cell_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MBSFN_Only_Mode_Indicator_Cell_SetupRqstTDD_LCR, create_dissector_handle(dissect_MBSFN_Only_Mode_Indicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MBSFN_Cell_ParameterID_Cell_SetupRqstTDD, create_dissector_handle(dissect_CellParameterID_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Time_Slot_Parameter_ID, create_dissector_handle(dissect_CellParameterID_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_IPDLParameter_Information_Cell_ReconfRqstFDD, create_dissector_handle(dissect_IPDLParameter_Information_Cell_ReconfRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_CellPortion_InformationList_Cell_ReconfRqstFDD, create_dissector_handle(dissect_CellPortion_InformationList_Cell_ReconfRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TimeSlotConfigurationList_LCR_Cell_ReconfRqstTDD, create_dissector_handle(dissect_TimeSlotConfigurationList_LCR_Cell_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DwPCH_LCR_Information_Cell_ReconfRqstTDD, create_dissector_handle(dissect_DwPCH_LCR_Information_Cell_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_IPDLParameter_Information_Cell_ReconfRqstTDD, create_dissector_handle(dissect_IPDLParameter_Information_Cell_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_IPDLParameter_Information_LCR_Cell_ReconfRqstTDD, create_dissector_handle(dissect_IPDLParameter_Information_LCR_Cell_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_SCH_768_Information_Cell_ReconfRqstTDD, create_dissector_handle(dissect_SCH_768_Information_Cell_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PCCPCH_768_Information_Cell_ReconfRqstTDD, create_dissector_handle(dissect_PCCPCH_768_Information_Cell_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MBSFN_Cell_ParameterID_Cell_ReconfRqstTDD, create_dissector_handle(dissect_CellParameterID_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Power_Local_Cell_Group_InformationList_ResourceStatusInd, create_dissector_handle(dissect_Power_Local_Cell_Group_InformationList_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MBSFN_Only_Mode_Capability, create_dissector_handle(dissect_MBSFN_Only_Mode_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Power_Local_Cell_Group_InformationList2_ResourceStatusInd, create_dissector_handle(dissect_Power_Local_Cell_Group_InformationList2_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_FPACH_LCR_InformationList_ResourceStatusInd, create_dissector_handle(dissect_FPACH_LCR_InformationList_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DwPCH_LCR_Information_ResourceStatusInd, create_dissector_handle(dissect_DwPCH_LCR_Information_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_Resources_Information_ResourceStatusInd, create_dissector_handle(dissect_HS_DSCH_Resources_Information_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MICH_Information_ResourceStatusInd, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_S_CCPCH_InformationListExt_ResourceStatusInd, create_dissector_handle(dissect_S_CCPCH_InformationListExt_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_S_CCPCH_LCR_InformationListExt_ResourceStatusInd, create_dissector_handle(dissect_S_CCPCH_LCR_InformationListExt_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_Resources_Information_ResourceStatusInd, create_dissector_handle(dissect_E_DCH_Resources_Information_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PLCCH_InformationList_ResourceStatusInd, create_dissector_handle(dissect_PLCCH_InformationList_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_P_CCPCH_768_Information_ResourceStatusInd, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information768_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_S_CCPCH_768_InformationList_ResourceStatusInd, create_dissector_handle(dissect_S_CCPCH_768_InformationList_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PICH_768_Information_ResourceStatusInd, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information768_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PRACH_768_InformationList_ResourceStatusInd, create_dissector_handle(dissect_PRACH_768_InformationList_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_SCH_768_Information_ResourceStatusInd, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information768_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MICH_768_Information_ResourceStatusInd, create_dissector_handle(dissect_Common_PhysicalChannel_Status_Information768_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_RUCCH_InformationList_ResourceStatusInd, create_dissector_handle(dissect_E_RUCCH_InformationList_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_RUCCH_768_InformationList_ResourceStatusInd, create_dissector_handle(dissect_E_RUCCH_768_InformationList_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DL_PowerBalancing_Information, create_dissector_handle(dissect_DL_PowerBalancing_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_FDD_Information, create_dissector_handle(dissect_HSDSCH_FDD_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_RNTI, create_dissector_handle(dissect_HSDSCH_RNTI_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSPDSCH_RL_ID, create_dissector_handle(dissect_RL_ID_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DPCH_Information_RL_SetupRqstFDD, create_dissector_handle(dissect_E_DPCH_Information_RL_SetupRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_FDD_Information, create_dissector_handle(dissect_E_DCH_FDD_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Serving_E_DCH_RL_ID, create_dissector_handle(dissect_Serving_E_DCH_RL_ID_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_F_DPCH_Information_RL_SetupRqstFDD, create_dissector_handle(dissect_F_DPCH_Information_RL_SetupRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Initial_DL_DPCH_TimingAdjustment_Allowed, create_dissector_handle(dissect_Initial_DL_DPCH_TimingAdjustment_Allowed_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DCH_Indicator_For_E_DCH_HSDPA_Operation, create_dissector_handle(dissect_DCH_Indicator_For_E_DCH_HSDPA_Operation_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Serving_Cell_Change_CFN, create_dissector_handle(dissect_CFN_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_ContinuousPacketConnectivityDTX_DRX_Information, create_dissector_handle(dissect_ContinuousPacketConnectivityDTX_DRX_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DPC_Mode, create_dissector_handle(dissect_DPC_Mode_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UL_DPDCH_Indicator_For_E_DCH_Operation, create_dissector_handle(dissect_UL_DPDCH_Indicator_For_E_DCH_Operation_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_RL_Specific_DCH_Info, create_dissector_handle(dissect_RL_Specific_DCH_Info_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DelayedActivation, create_dissector_handle(dissect_DelayedActivation_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Primary_CPICH_Usage_for_Channel_Estimation, create_dissector_handle(dissect_Primary_CPICH_Usage_for_Channel_Estimation_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Secondary_CPICH_Information, create_dissector_handle(dissect_CommonPhysicalChannelID_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_RL_Indication, create_dissector_handle(dissect_E_DCH_RL_Indication_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_RL_Specific_E_DCH_Info, create_dissector_handle(dissect_RL_Specific_E_DCH_Info_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_SynchronisationIndicator, create_dissector_handle(dissect_SynchronisationIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_F_DPCH_SlotFormat, create_dissector_handle(dissect_F_DPCH_SlotFormat_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_TDD_Information, create_dissector_handle(dissect_HSDSCH_TDD_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PDSCH_RL_ID, create_dissector_handle(dissect_RL_ID_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_Information, create_dissector_handle(dissect_E_DCH_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_Serving_RL_ID, create_dissector_handle(dissect_RL_ID_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_768_Information, create_dissector_handle(dissect_E_DCH_768_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_LCR_Information, create_dissector_handle(dissect_E_DCH_LCR_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UL_DPCH_LCR_Information_RL_SetupRqstTDD, create_dissector_handle(dissect_UL_DPCH_LCR_Information_RL_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UL_SIRTarget, create_dissector_handle(dissect_UL_SIR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TDD_TPC_UplinkStepSize_LCR_RL_SetupRqstTDD, create_dissector_handle(dissect_TDD_TPC_UplinkStepSize_LCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UL_DPCH_768_Information_RL_SetupRqstTDD, create_dissector_handle(dissect_UL_DPCH_768_Information_RL_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DL_DPCH_LCR_Information_RL_SetupRqstTDD, create_dissector_handle(dissect_DL_DPCH_LCR_Information_RL_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_CCTrCH_Initial_DL_Power_RL_SetupRqstTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_CCTrCH_Maximum_DL_Power_RL_SetupRqstTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_CCTrCH_Minimum_DL_Power_RL_SetupRqstTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DL_DPCH_768_Information_RL_SetupRqstTDD, create_dissector_handle(dissect_DL_DPCH_768_Information_RL_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TimeslotISCP_LCR_InfoList_RL_SetupRqstTDD, create_dissector_handle(dissect_DL_TimeslotISCPInfoLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UL_Synchronisation_Parameters_LCR, create_dissector_handle(dissect_UL_Synchronisation_Parameters_LCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_FDD_Information_Response, create_dissector_handle(dissect_HSDSCH_FDD_Information_Response_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DL_PowerBalancing_ActivationIndicator, create_dissector_handle(dissect_DL_PowerBalancing_ActivationIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_RL_Set_ID, create_dissector_handle(dissect_RL_Set_ID_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_FDD_DL_Control_Channel_Information, create_dissector_handle(dissect_E_DCH_FDD_DL_Control_Channel_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Initial_DL_DPCH_TimingAdjustment, create_dissector_handle(dissect_DL_DPCH_TimingAdjustment_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_FDD_Information_Response, create_dissector_handle(dissect_E_DCH_FDD_Information_Response_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_RL_InformationResponse_LCR_RL_SetupRspTDD, create_dissector_handle(dissect_RL_InformationResponse_LCR_RL_SetupRspTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_TDD_Information_Response, create_dissector_handle(dissect_HSDSCH_TDD_Information_Response_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_Information_Response, create_dissector_handle(dissect_E_DCH_Information_Response_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HS_DSCH_Serving_Cell_Change_Info, create_dissector_handle(dissect_HS_DSCH_Serving_Cell_Change_Info_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DPCH_Information_RL_AdditionReqFDD, create_dissector_handle(dissect_E_DPCH_Information_RL_AdditionReqFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DLReferencePower, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_Configured_Indicator, create_dissector_handle(dissect_HSDSCH_Configured_Indicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD, create_dissector_handle(dissect_UL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TDD_TPC_UplinkStepSize_LCR_RL_AdditionRqstTDD, create_dissector_handle(dissect_TDD_TPC_UplinkStepSize_LCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UL_DPCH_InformationItem_768_RL_AdditionRqstTDD, create_dissector_handle(dissect_UL_DPCH_InformationItem_768_RL_AdditionRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD, create_dissector_handle(dissect_DL_DPCH_InformationItem_LCR_RL_AdditionRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_CCTrCH_Initial_DL_Power_RL_AdditionRqstTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TDD_TPC_DownlinkStepSize_RL_AdditionRqstTDD, create_dissector_handle(dissect_TDD_TPC_DownlinkStepSize_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_CCTrCH_Maximum_DL_Power_RL_AdditionRqstTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_CCTrCH_Minimum_DL_Power_RL_AdditionRqstTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DL_DPCH_InformationItem_768_RL_AdditionRqstTDD, create_dissector_handle(dissect_DL_DPCH_InformationItem_768_RL_AdditionRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TimeslotISCP_InformationList_LCR_RL_AdditionRqstTDD, create_dissector_handle(dissect_DL_TimeslotISCPInfoLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HS_DSCH_Serving_Cell_Change_Info_Response, create_dissector_handle(dissect_HS_DSCH_Serving_Cell_Change_Info_Response_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_Serving_Cell_Change_Info_Response, create_dissector_handle(dissect_E_DCH_Serving_Cell_Change_Info_Response_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MAChs_ResetIndicator, create_dissector_handle(dissect_MAChs_ResetIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_RL_InformationResponse_LCR_RL_AdditionRspTDD, create_dissector_handle(dissect_RL_InformationResponse_LCR_RL_AdditionRspTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_SignallingBearerRequestIndicator, create_dissector_handle(dissect_SignallingBearerRequestIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_Information_to_Modify, create_dissector_handle(dissect_HSDSCH_Information_to_Modify_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_MACdFlows_to_Add, create_dissector_handle(dissect_HSDSCH_MACdFlows_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_MACdFlows_to_Delete, create_dissector_handle(dissect_HSDSCH_MACdFlows_to_Delete_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DPCH_Information_RL_ReconfPrepFDD, create_dissector_handle(dissect_E_DPCH_Information_RL_ReconfPrepFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_FDD_Information_to_Modify, create_dissector_handle(dissect_E_DCH_FDD_Information_to_Modify_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_MACdFlows_to_Add, create_dissector_handle(dissect_E_DCH_MACdFlows_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_MACdFlows_to_Delete, create_dissector_handle(dissect_E_DCH_MACdFlows_to_Delete_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_F_DPCH_Information_RL_ReconfPrepFDD, create_dissector_handle(dissect_F_DPCH_Information_RL_ReconfPrepFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Fast_Reconfiguration_Mode, create_dissector_handle(dissect_Fast_Reconfiguration_Mode_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_CPC_Information, create_dissector_handle(dissect_CPC_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DL_DPCH_Power_Information_RL_ReconfPrepFDD, create_dissector_handle(dissect_DL_DPCH_Power_Information_RL_ReconfPrepFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DL_DPCH_TimingAdjustment, create_dissector_handle(dissect_DL_DPCH_TimingAdjustment_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Secondary_CPICH_Information_Change, create_dissector_handle(dissect_Secondary_CPICH_Information_Change_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_multiple_RL_Information_RL_ReconfPrepTDD, create_dissector_handle(dissect_MultipleRL_Information_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_Information_Reconfig, create_dissector_handle(dissect_E_DCH_Information_Reconfig_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_768_Information_Reconfig, create_dissector_handle(dissect_E_DCH_768_Information_Reconfig_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_LCR_Information_Reconfig, create_dissector_handle(dissect_E_DCH_LCR_Information_Reconfig_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UL_DPCH_LCR_InformationAddListIE_RL_ReconfPrepTDD, create_dissector_handle(dissect_UL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TDD_TPC_UplinkStepSize_InformationAdd_LCR_RL_ReconfPrepTDD, create_dissector_handle(dissect_TDD_TPC_UplinkStepSize_LCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_RL_ID, create_dissector_handle(dissect_RL_ID_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_multipleRL_ul_DPCH_InformationList, create_dissector_handle(dissect_MultipleRL_UL_DPCH_InformationAddList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UL_DPCH_768_InformationAddItemIE_RL_ReconfPrepTDD, create_dissector_handle(dissect_UL_DPCH_768_InformationAddList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UL_DPCH_768_InformationAddListIE_RL_ReconfPrepTDD, create_dissector_handle(dissect_UL_DPCH_768_InformationAddList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UL_DPCH_LCR_InformationModify_AddList, create_dissector_handle(dissect_UL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TDD_TPC_UplinkStepSize_InformationModify_LCR_RL_ReconfPrepTDD, create_dissector_handle(dissect_TDD_TPC_UplinkStepSize_LCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_multipleRL_ul_DPCH_InformationModifyList, create_dissector_handle(dissect_MultipleRL_UL_DPCH_InformationModifyList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UL_DPCH_768_InformationModify_AddItem, create_dissector_handle(dissect_UL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UL_DPCH_768_InformationModify_AddList, create_dissector_handle(dissect_UL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UL_TimeslotLCR_Information_RL_ReconfPrepTDD, create_dissector_handle(dissect_UL_TimeslotLCR_InformationModify_ModifyList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UL_Timeslot768_Information_RL_ReconfPrepTDD, create_dissector_handle(dissect_UL_Timeslot768_InformationModify_ModifyList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PLCCH_Information_RL_ReconfPrepTDDLCR, create_dissector_handle(dissect_PLCCHinformation_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UL_DPCH_TimeSlotFormat_LCR_ModifyItem_RL_ReconfPrepTDD, create_dissector_handle(dissect_TDD_UL_DPCH_TimeSlotFormat_LCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_DPCH_LCR_InformationAddList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_CCTrCH_Initial_DL_Power_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TDD_TPC_DownlinkStepSize_InformationAdd_RL_ReconfPrepTDD, create_dissector_handle(dissect_TDD_TPC_DownlinkStepSize_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_CCTrCH_Maximum_DL_Power_InformationAdd_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_CCTrCH_Minimum_DL_Power_InformationAdd_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_multipleRL_dl_DPCH_InformationList, create_dissector_handle(dissect_MultipleRL_DL_DPCH_InformationAddList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DL_DPCH_768_InformationAddItem_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_DPCH_768_InformationAddList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DL_DPCH_768_InformationAddList_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_DPCH_768_InformationAddList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_DPCH_LCR_InformationModify_AddList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TDD_TPC_DownlinkStepSize_InformationModify_RL_ReconfPrepTDD, create_dissector_handle(dissect_TDD_TPC_DownlinkStepSize_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_CCTrCH_Maximum_DL_Power_InformationModify_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_CCTrCH_Minimum_DL_Power_InformationModify_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_multipleRL_dl_DPCH_InformationModifyList, create_dissector_handle(dissect_MultipleRL_DL_DPCH_InformationModifyList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DL_DPCH_768_InformationModify_AddItem_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_DPCH_768_InformationModify_AddList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_Timeslot_LCR_InformationModify_ModifyList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DL_Timeslot_768_InformationModify_ModifyList_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_Timeslot_768_InformationModify_ModifyList_RL_ReconfPrepTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Maximum_DL_Power_Modify_LCR_InformationModify_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Minimum_DL_Power_Modify_LCR_InformationModify_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DL_DPCH_TimeSlotFormat_LCR_ModifyItem_RL_ReconfPrepTDD, create_dissector_handle(dissect_TDD_DL_DPCH_TimeSlotFormat_LCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_InitDL_Power, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TimeslotISCP_LCR_InfoList_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_TimeslotISCPInfoLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TargetCommunicationControlPortID, create_dissector_handle(dissect_CommunicationControlPortID_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Fast_Reconfiguration_Permission, create_dissector_handle(dissect_Fast_Reconfiguration_Permission_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DL_PowerBalancing_UpdatedIndicator, create_dissector_handle(dissect_DL_PowerBalancing_UpdatedIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_Information_to_Modify_Unsynchronised, create_dissector_handle(dissect_HSDSCH_Information_to_Modify_Unsynchronised_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DPCH_Information_RL_ReconfRqstFDD, create_dissector_handle(dissect_E_DPCH_Information_RL_ReconfRqstFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_multiple_RL_Information_RL_ReconfRqstTDD, create_dissector_handle(dissect_Multiple_RL_Information_RL_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DL_DPCH_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD, create_dissector_handle(dissect_DL_DPCH_LCR_InformationModify_ModifyList_RL_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_CCTrCH_Maximum_DL_Power_InformationModify_RL_ReconfRqstTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_CCTrCH_Minimum_DL_Power_InformationModify_RL_ReconfRqstTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_multipleRL_dl_CCTrCH_InformationModifyList_RL_ReconfRqstTDD, create_dissector_handle(dissect_MultipleRL_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TimeslotISCPInfoList_LCR_DL_PC_RqstTDD, create_dissector_handle(dissect_DL_TimeslotISCPInfoLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PrimCCPCH_RSCP_DL_PC_RqstTDD, create_dissector_handle(dissect_PrimaryCCPCH_RSCP_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PrimaryCCPCH_RSCP_Delta, create_dissector_handle(dissect_PrimaryCCPCH_RSCP_Delta_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_NumberOfReportedCellPortions, create_dissector_handle(dissect_NumberOfReportedCellPortions_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_AlternativeFormatReportingIndicator, create_dissector_handle(dissect_AlternativeFormatReportingIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PUSCH_Info_DM_Rqst, create_dissector_handle(dissect_PUSCH_Info_DM_Rqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSSICH_Info_DM_Rqst, create_dissector_handle(dissect_HSSICH_Info_DM_Rqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DPCH_ID768_DM_Rqst, create_dissector_handle(dissect_DPCH_ID768_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PUSCH_Info_DM_Rsp, create_dissector_handle(dissect_PUSCH_Info_DM_Rsp_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSSICH_Info_DM_Rsp, create_dissector_handle(dissect_HS_SICH_ID_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_multiple_DedicatedMeasurementValueList_TDD_DM_Rsp, create_dissector_handle(dissect_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp, create_dissector_handle(dissect_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_multiple_PUSCH_InfoList_DM_Rsp, create_dissector_handle(dissect_Multiple_PUSCH_InfoList_DM_Rsp_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_multiple_HSSICHMeasurementValueList_TDD_DM_Rsp, create_dissector_handle(dissect_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DPCH_ID768_DM_Rsp, create_dissector_handle(dissect_DPCH_ID768_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_multiple_DedicatedMeasurementValueList_768_TDD_DM_Rsp, create_dissector_handle(dissect_Multiple_DedicatedMeasurementValueList_768_TDD_DM_Rsp_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PUSCH_Info_DM_Rprt, create_dissector_handle(dissect_PUSCH_Info_DM_Rprt_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSSICH_Info_DM_Rprt, create_dissector_handle(dissect_HS_SICH_ID_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_multiple_PUSCH_InfoList_DM_Rprt, create_dissector_handle(dissect_Multiple_PUSCH_InfoList_DM_Rprt_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DPCH_ID768_DM_Rprt, create_dissector_handle(dissect_DPCH_ID768_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_AGCH_And_E_RGCH_E_HICH_FDD_Scrambling_Code, create_dissector_handle(dissect_DL_ScramblingCode_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_AGCH_FDD_Code_Information, create_dissector_handle(dissect_E_AGCH_FDD_Code_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_RGCH_E_HICH_FDD_Code_Information, create_dissector_handle(dissect_E_RGCH_E_HICH_FDD_Code_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDPA_And_EDCH_CellPortion_Information_PSCH_ReconfRqst, create_dissector_handle(dissect_HSDPA_And_EDCH_CellPortion_InformationList_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Maximum_Target_ReceivedTotalWideBandPower, create_dissector_handle(dissect_Maximum_Target_ReceivedTotalWideBandPower_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Target_NonServing_EDCH_To_Total_EDCH_Power_Ratio, create_dissector_handle(dissect_Target_NonServing_EDCH_To_Total_EDCH_Power_Ratio_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_Common_System_InformationFDD, create_dissector_handle(dissect_HSDSCH_Common_System_InformationFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_Paging_System_InformationFDD, create_dissector_handle(dissect_HSDSCH_Paging_System_InformationFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HS_PDSCH_TDD_Information_PSCH_ReconfRqst, create_dissector_handle(dissect_HS_PDSCH_TDD_Information_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Add_To_HS_SCCH_Resource_Pool_PSCH_ReconfRqst, create_dissector_handle(dissect_Add_To_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Modify_HS_SCCH_Resource_Pool_PSCH_ReconfRqst, create_dissector_handle(dissect_Modify_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Delete_From_HS_SCCH_Resource_Pool_PSCH_ReconfRqst, create_dissector_handle(dissect_Delete_From_HS_SCCH_Resource_Pool_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_ConfigurationGenerationID, create_dissector_handle(dissect_ConfigurationGenerationID_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_PUCH_Information_PSCH_ReconfRqst, create_dissector_handle(dissect_E_PUCH_Information_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Add_To_E_AGCH_Resource_Pool_PSCH_ReconfRqst, create_dissector_handle(dissect_Add_To_E_AGCH_Resource_Pool_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Modify_E_AGCH_Resource_Pool_PSCH_ReconfRqst, create_dissector_handle(dissect_Modify_E_AGCH_Resource_Pool_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Delete_From_E_AGCH_Resource_Pool_PSCH_ReconfRqst, create_dissector_handle(dissect_Delete_From_E_AGCH_Resource_Pool_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_HICH_Information_PSCH_ReconfRqst, create_dissector_handle(dissect_E_HICH_Information_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Maximum_Generated_ReceivedTotalWideBandPowerInOtherCells, create_dissector_handle(dissect_Maximum_Generated_ReceivedTotalWideBandPowerInOtherCells_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_PUCH_Information_768_PSCH_ReconfRqst, create_dissector_handle(dissect_E_PUCH_Information_768_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Add_To_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst, create_dissector_handle(dissect_Add_To_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Modify_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst, create_dissector_handle(dissect_Modify_E_AGCH_Resource_Pool_768_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_HICH_Information_768_PSCH_ReconfRqst, create_dissector_handle(dissect_E_HICH_Information_768_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_PUCH_Information_LCR_PSCH_ReconfRqst, create_dissector_handle(dissect_E_PUCH_Information_LCR_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Add_To_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst, create_dissector_handle(dissect_Add_To_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Modify_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst, create_dissector_handle(dissect_Modify_E_AGCH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Add_To_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst, create_dissector_handle(dissect_Add_To_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Modify_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst, create_dissector_handle(dissect_Modify_E_HICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Delete_From_E_HICH_Resource_Pool_PSCH_ReconfRqst, create_dissector_handle(dissect_Delete_From_E_HICH_Resource_Pool_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_SYNC_UL_Partition_LCR, create_dissector_handle(dissect_SYNC_UL_Partition_LCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PDSCH_AddInformation_LCR_PSCH_ReconfRqst, create_dissector_handle(dissect_PDSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PDSCH_AddInformation_768_PSCH_ReconfRqst, create_dissector_handle(dissect_PDSCH_AddInformation_768_AddItem_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PDSCH_Timeslot_Format_PSCH_ReconfRqst_LCR, create_dissector_handle(dissect_TDD_DL_DPCH_TimeSlotFormat_LCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PDSCH_ModifyInformation_768_PSCH_ReconfRqst, create_dissector_handle(dissect_PDSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PUSCH_AddInformation_LCR_PSCH_ReconfRqst, create_dissector_handle(dissect_PUSCH_AddInformation_LCR_AddItem_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PUSCH_AddInformation_768_PSCH_ReconfRqst, create_dissector_handle(dissect_PUSCH_AddInformation_768_AddItem_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PUSCH_Timeslot_Format_PSCH_ReconfRqst_LCR, create_dissector_handle(dissect_TDD_UL_DPCH_TimeSlotFormat_LCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PUSCH_ModifyInformation_768_PSCH_ReconfRqst, create_dissector_handle(dissect_PUSCH_ModifyInformation_768_ModifyItem_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_dL_HS_PDSCH_Timeslot_Information_768_PSCH_ReconfRqst, create_dissector_handle(dissect_DL_HS_PDSCH_Timeslot_Information_768_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_hS_SCCH_Information_768_PSCH_ReconfRqst, create_dissector_handle(dissect_HS_SCCH_Information_768_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_hS_SCCH_InformationModify_768_PSCH_ReconfRqst, create_dissector_handle(dissect_HS_SCCH_InformationModify_768_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_HICH_TimeOffset, create_dissector_handle(dissect_E_HICH_TimeOffset_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_HICH_TimeOffsetLCR, create_dissector_handle(dissect_E_HICH_TimeOffsetLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_Common_System_Information_ResponseFDD, create_dissector_handle(dissect_HSDSCH_Common_System_Information_ResponseFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_Paging_System_Information_ResponseFDD, create_dissector_handle(dissect_HSDSCH_Paging_System_Information_ResponseFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_SYNCDlCodeId_TransInitLCR_CellSyncInitiationRqstTDD, create_dissector_handle(dissect_SYNCDlCodeId_TransInitLCR_CellSyncInitiationRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_SYNCDlCodeId_MeasureInitLCR_CellSyncInitiationRqstTDD, create_dissector_handle(dissect_SYNCDlCodeId_MeasureInitLCR_CellSyncInitiationRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_NSubCyclesPerCyclePeriod_CellSyncReconfRqstTDD, create_dissector_handle(dissect_NSubCyclesPerCyclePeriod_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_SYNCDlCodeIdTransReconfInfoLCR_CellSyncReconfRqstTDD, create_dissector_handle(dissect_SYNCDlCodeIdTransReconfInfoLCR_CellSyncReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_SYNCDlCodeIdMeasReconfigurationLCR_CellSyncReconfRqstTDD, create_dissector_handle(dissect_SYNCDlCodeIdMeasInfoLCR_CellSyncReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DwPCH_Power, create_dissector_handle(dissect_DwPCH_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TimingAdjustmentValueLCR, create_dissector_handle(dissect_TimingAdjustmentValueLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_AccumulatedClockupdate_CellSyncReprtTDD, create_dissector_handle(dissect_TimingAdjustmentValue_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_SyncDLCodeIdsMeasInfoList_CellSyncReprtTDD, create_dissector_handle(dissect_SyncDLCodeIdsMeasInfoList_CellSyncReprtTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_RearrangeList_Bearer_RearrangeInd, create_dissector_handle(dissect_E_DCH_RearrangeList_Bearer_RearrangeInd_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_FDD_Update_Information, create_dissector_handle(dissect_E_DCH_FDD_Update_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_IPMulticastIndication, create_dissector_handle(dissect_IPMulticastIndication_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TimeSlotConfigurationList_LCR_CTCH_SetupRqstTDD, create_dissector_handle(dissect_TimeSlotConfigurationList_LCR_CTCH_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Cell_Frequency_List_Information_LCR_MulFreq_AuditRsp, create_dissector_handle(dissect_Cell_Frequency_List_Information_LCR_MulFreq_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Cell_Frequency_List_LCR_MulFreq_Cell_SetupRqstTDD, create_dissector_handle(dissect_Cell_Frequency_List_LCR_MulFreq_Cell_SetupRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UARFCN_Adjustment, create_dissector_handle(dissect_UARFCN_Adjustment_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Cell_Frequency_List_Information_LCR_MulFreq_ResourceStatusInd, create_dissector_handle(dissect_Cell_Frequency_List_Information_LCR_MulFreq_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UPPCHPositionLCR, create_dissector_handle(dissect_UPPCHPositionLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UPPCH_LCR_Parameters_CTCH_ReconfRqstTDD, create_dissector_handle(dissect_UPPCH_LCR_Parameters_CTCH_ReconfRqstTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UPPCH_LCR_InformationList_AuditRsp, create_dissector_handle(dissect_UPPCH_LCR_InformationList_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UPPCH_LCR_InformationList_ResourceStatusInd, create_dissector_handle(dissect_UPPCH_LCR_InformationList_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_multipleFreq_dL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst, create_dissector_handle(dissect_MultipleFreq_DL_HS_PDSCH_Timeslot_Information_LCR_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_multipleFreq_HS_DSCH_Resources_InformationList_AuditRsp, create_dissector_handle(dissect_MultipleFreq_HS_DSCH_Resources_InformationList_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_multipleFreq_HS_DSCH_Resources_InformationList_ResourceStatusInd, create_dissector_handle(dissect_MultipleFreq_HS_DSCH_Resources_InformationList_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Extended_HS_SCCH_ID, create_dissector_handle(dissect_Extended_HS_SCCH_ID_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Extended_HS_SICH_ID, create_dissector_handle(dissect_Extended_HS_SICH_ID_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSSICH_InfoExt_DM_Rqst, create_dissector_handle(dissect_HSSICH_InfoExt_DM_Rqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Delete_From_HS_SCCH_Resource_PoolExt_PSCH_ReconfRqst, create_dissector_handle(dissect_Delete_From_HS_SCCH_Resource_PoolExt_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HS_SCCH_InformationExt_LCR_PSCH_ReconfRqst, create_dissector_handle(dissect_HS_SCCH_InformationExt_LCR_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HS_SCCH_InformationModifyExt_LCR_PSCH_ReconfRqst, create_dissector_handle(dissect_HS_SCCH_InformationModifyExt_LCR_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PowerControlGAP, create_dissector_handle(dissect_ControlGAP_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MBSFN_SpecialTimeSlot_LCR, create_dissector_handle(dissect_TimeslotLCR_Extension_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Common_MACFlows_to_DeleteFDD, create_dissector_handle(dissect_Common_MACFlows_to_DeleteFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Paging_MACFlows_to_DeleteFDD, create_dissector_handle(dissect_Paging_MACFlows_to_DeleteFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Maximum_Target_ReceivedTotalWideBandPower_LCR, create_dissector_handle(dissect_Maximum_Target_ReceivedTotalWideBandPower_LCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DPDCH_PowerInterpolation, create_dissector_handle(dissect_E_DPDCH_PowerInterpolation_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_TFCI_Boost_Information, create_dissector_handle(dissect_E_TFCI_Boost_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Ext_Max_Bits_MACe_PDU_non_scheduled, create_dissector_handle(dissect_Ext_Max_Bits_MACe_PDU_non_scheduled_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Ext_Reference_E_TFCI_PO, create_dissector_handle(dissect_Ext_Reference_E_TFCI_PO_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HARQ_MemoryPartitioningInfoExtForMIMO, create_dissector_handle(dissect_HARQ_MemoryPartitioningInfoExtForMIMO_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_IPMulticastDataBearerIndication, create_dissector_handle(dissect_IPMulticastDataBearerIndication_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MIMO_ActivationIndicator, create_dissector_handle(dissect_MIMO_ActivationIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MIMO_Mode_Indicator, create_dissector_handle(dissect_MIMO_Mode_Indicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MIMO_N_M_Ratio, create_dissector_handle(dissect_MIMO_N_M_Ratio_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_multicarrier_number, create_dissector_handle(dissect_Multicarrier_Number_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_number_Of_Supported_Carriers, create_dissector_handle(dissect_Number_Of_Supported_Carriers_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_multipleFreq_HSPDSCH_InformationList_ResponseTDDLCR, create_dissector_handle(dissect_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_SixtyfourQAM_UsageAllowedIndicator, create_dissector_handle(dissect_SixtyfourQAM_UsageAllowedIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_SixtyfourQAM_DL_UsageIndicator, create_dissector_handle(dissect_SixtyfourQAM_DL_UsageIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_SixteenQAM_UL_Operation_Indicator, create_dissector_handle(dissect_SixteenQAM_UL_Operation_Indicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TransportBearerNotRequestedIndicator, create_dissector_handle(dissect_TransportBearerNotRequestedIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TransportBearerNotSetupIndicator, create_dissector_handle(dissect_TransportBearerNotSetupIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_tSN_Length, create_dissector_handle(dissect_TSN_Length_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Extended_E_DCH_LCRTDD_PhysicalLayerCategory, create_dissector_handle(dissect_Extended_E_DCH_LCRTDD_PhysicalLayerCategory_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MultipleFreq_E_DCH_Resources_InformationList_AuditRsp, create_dissector_handle(dissect_MultipleFreq_E_DCH_Resources_InformationList_AuditRsp_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MultipleFreq_E_DCH_Resources_InformationList_ResourceStatusInd, create_dissector_handle(dissect_MultipleFreq_E_DCH_Resources_InformationList_ResourceStatusInd_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MultipleFreq_E_PUCH_Timeslot_InformationList_LCR_PSCH_ReconfRqst, create_dissector_handle(dissect_MultipleFreq_E_PUCH_Timeslot_InformationList_LCR_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Extended_E_HICH_ID_TDD, create_dissector_handle(dissect_Extended_E_HICH_ID_TDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_ContinuousPacketConnectivityHS_SCCH_less_Deactivate_Indicator, create_dissector_handle(dissect_ContinuousPacketConnectivityHS_SCCH_less_Deactivate_Indicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_MACdPDU_SizeCapability, create_dissector_handle(dissect_E_DCH_MACdPDU_SizeCapability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_MACdPDUSizeFormat, create_dissector_handle(dissect_E_DCH_MACdPDUSizeFormat_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MaximumNumber_Of_Retransmission_for_Scheduling_Info_LCRTDD, create_dissector_handle(dissect_Maximum_Number_of_Retransmissions_For_E_DCH_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_RetransmissionTimer_for_SchedulingInfo_LCRTDD, create_dissector_handle(dissect_E_DCH_MACdFlow_Retransmission_Timer_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_HICH_TimeOffset_Extension, create_dissector_handle(dissect_E_HICH_TimeOffset_ExtensionLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_PUCH_PowerControlGAP, create_dissector_handle(dissect_ControlGAP_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_TBSizeTableIndicator, create_dissector_handle(dissect_HSDSCH_TBSizeTableIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_DL_Control_Channel_Change_Information, create_dissector_handle(dissect_E_DCH_DL_Control_Channel_Change_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_DL_Control_Channel_Grant_Information, create_dissector_handle(dissect_E_DCH_DL_Control_Channel_Grant_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DGANSS_Corrections_Req, create_dissector_handle(dissect_DGANSS_Corrections_Req_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_AdditionalTimeSlotListLCR, create_dissector_handle(dissect_AdditionalTimeSlotListLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_AdditionalMeasurementValueList, create_dissector_handle(dissect_AdditionalMeasurementValueList_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_AGCH_Table_Choice, create_dissector_handle(dissect_E_AGCH_Table_Choice_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HS_Cause, create_dissector_handle(dissect_Cause_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_Cause, create_dissector_handle(dissect_Cause_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Common_EDCH_Capability, create_dissector_handle(dissect_Common_EDCH_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_AI_Capability, create_dissector_handle(dissect_E_AI_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Common_EDCH_System_InformationFDD, create_dissector_handle(dissect_Common_EDCH_System_InformationFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Common_UL_MACFlows_to_DeleteFDD, create_dissector_handle(dissect_Common_MACFlows_to_DeleteFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Common_EDCH_MACdFlows_to_DeleteFDD, create_dissector_handle(dissect_E_DCH_MACdFlows_to_Delete_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Common_EDCH_System_Information_ResponseFDD, create_dissector_handle(dissect_Common_EDCH_System_Information_ResponseFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Enhanced_UE_DRX_Capability, create_dissector_handle(dissect_Enhanced_UE_DRX_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Enhanced_UE_DRX_InformationFDD, create_dissector_handle(dissect_Enhanced_UE_DRX_InformationFDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TransportBearerRequestIndicator, create_dissector_handle(dissect_TransportBearerRequestIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_SixtyfourQAM_DL_MIMO_Combined_Capability, create_dissector_handle(dissect_SixtyfourQAM_DL_MIMO_Combined_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_RNTI, create_dissector_handle(dissect_E_RNTI_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MinimumReducedE_DPDCH_GainFactor, create_dissector_handle(dissect_MinimumReducedE_DPDCH_GainFactor_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_GANSS_Time_ID, create_dissector_handle(dissect_GANSS_Time_ID_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_GANSS_AddIonoModelReq, create_dissector_handle(dissect_GANSS_AddIonoModelReq_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_GANSS_EarthOrientParaReq, create_dissector_handle(dissect_GANSS_EarthOrientParaReq_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_GANSS_AddNavigationModelsReq, create_dissector_handle(dissect_GANSS_AddNavigationModelsReq_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_GANSS_AddUTCModelsReq, create_dissector_handle(dissect_GANSS_AddUTCModelsReq_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_GANSS_AuxInfoReq, create_dissector_handle(dissect_GANSS_AuxInfoReq_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_GANSS_SBAS_ID, create_dissector_handle(dissect_GANSS_SBAS_ID_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_GANSS_ID, create_dissector_handle(dissect_GANSS_ID_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_GANSS_Additional_Ionospheric_Model, create_dissector_handle(dissect_GANSS_Additional_Ionospheric_Model_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_GANSS_Earth_Orientation_Parameters, create_dissector_handle(dissect_GANSS_Earth_Orientation_Parameters_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_GANSS_Additional_Time_Models, create_dissector_handle(dissect_GANSS_Additional_Time_Models_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_GANSS_Additional_Navigation_Models, create_dissector_handle(dissect_GANSS_Additional_Navigation_Models_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_GANSS_Additional_UTC_Models, create_dissector_handle(dissect_GANSS_Additional_UTC_Models_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_GANSS_Auxiliary_Information, create_dissector_handle(dissect_GANSS_Auxiliary_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DPCCH_Power_Boosting_Capability, create_dissector_handle(dissect_E_DPCCH_Power_Boosting_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_Common_System_InformationLCR, create_dissector_handle(dissect_HSDSCH_Common_System_InformationLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_Common_System_Information_ResponseLCR, create_dissector_handle(dissect_HSDSCH_Common_System_Information_ResponseLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_Paging_System_InformationLCR, create_dissector_handle(dissect_HSDSCH_Paging_System_InformationLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_Paging_System_Information_ResponseLCR, create_dissector_handle(dissect_HSDSCH_Paging_System_Information_ResponseLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Common_MACFlows_to_DeleteLCR, create_dissector_handle(dissect_Common_MACFlows_to_DeleteLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Paging_MACFlows_to_DeleteLCR, create_dissector_handle(dissect_Paging_MACFlows_to_DeleteLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Common_EDCH_System_InformationLCR, create_dissector_handle(dissect_Common_EDCH_System_InformationLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Common_UL_MACFlows_to_DeleteLCR, create_dissector_handle(dissect_Common_MACFlows_to_DeleteLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Common_EDCH_MACdFlows_to_DeleteLCR, create_dissector_handle(dissect_E_DCH_MACdFlows_to_DeleteLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Common_EDCH_System_Information_ResponseLCR, create_dissector_handle(dissect_Common_EDCH_System_Information_ResponseLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Enhanced_UE_DRX_CapabilityLCR, create_dissector_handle(dissect_Enhanced_UE_DRX_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Enhanced_UE_DRX_InformationLCR, create_dissector_handle(dissect_Enhanced_UE_DRX_InformationLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_PreconfigurationSetup, create_dissector_handle(dissect_HSDSCH_PreconfigurationSetup_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSDSCH_PreconfigurationInfo, create_dissector_handle(dissect_HSDSCH_PreconfigurationInfo_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_NoOfTargetCellHS_SCCH_Order, create_dissector_handle(dissect_NoOfTargetCellHS_SCCH_Order_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_EnhancedHSServingCC_Abort, create_dissector_handle(dissect_EnhancedHSServingCC_Abort_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_HS_Cell_Information_RL_Setup, create_dissector_handle(dissect_Additional_HS_Cell_Information_RL_Setup_List_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_HS_Cell_Information_Response, create_dissector_handle(dissect_Additional_HS_Cell_Information_Response_List_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_HS_Cell_Information_RL_Addition, create_dissector_handle(dissect_Additional_HS_Cell_Information_RL_Addition_List_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_HS_Cell_Change_Information_Response, create_dissector_handle(dissect_Additional_HS_Cell_Change_Information_Response_List_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_HS_Cell_Information_RL_Reconf_Prep, create_dissector_handle(dissect_Additional_HS_Cell_Information_RL_Reconf_Prep_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_HS_Cell_Information_RL_Reconf_Req, create_dissector_handle(dissect_Additional_HS_Cell_Information_RL_Reconf_Req_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_HS_Cell_Information_RL_Param_Upd, create_dissector_handle(dissect_Additional_HS_Cell_Information_RL_Param_Upd_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Multi_Cell_Capability_Info, create_dissector_handle(dissect_Multi_Cell_Capability_Info_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_IMB_Parameters, create_dissector_handle(dissect_IMB_Parameters_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MACes_Maximum_Bitrate_LCR, create_dissector_handle(dissect_MACes_Maximum_Bitrate_LCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Semi_PersistentScheduling_CapabilityLCR, create_dissector_handle(dissect_Semi_PersistentScheduling_CapabilityLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_Semi_PersistentScheduling_Information_LCR, create_dissector_handle(dissect_E_DCH_Semi_PersistentScheduling_Information_LCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HS_DSCH_Semi_PersistentScheduling_Information_LCR, create_dissector_handle(dissect_HS_DSCH_Semi_PersistentScheduling_Information_LCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Add_To_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst, create_dissector_handle(dissect_Add_To_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Modify_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst, create_dissector_handle(dissect_Modify_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst, create_dissector_handle(dissect_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_ContinuousPacketConnectivity_DRX_CapabilityLCR, create_dissector_handle(dissect_ContinuousPacketConnectivity_DRX_CapabilityLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_ContinuousPacketConnectivity_DRX_InformationLCR, create_dissector_handle(dissect_ContinuousPacketConnectivity_DRX_InformationLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_ContinuousPacketConnectivity_DRX_Information_ResponseLCR, create_dissector_handle(dissect_ContinuousPacketConnectivity_DRX_Information_ResponseLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_CPC_InformationLCR, create_dissector_handle(dissect_CPC_InformationLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR, create_dissector_handle(dissect_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR, create_dissector_handle(dissect_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_AGCH_UE_Inactivity_Monitor_Threshold, create_dissector_handle(dissect_E_AGCH_UE_Inactivity_Monitor_Threshold_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_IdleIntervalInformation, create_dissector_handle(dissect_IdleIntervalInformation_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSSICH_ReferenceSignal_InformationLCR, create_dissector_handle(dissect_HSSICH_ReferenceSignal_InformationLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MIMO_ReferenceSignal_InformationListLCR, create_dissector_handle(dissect_MIMO_ReferenceSignal_InformationListLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MIMO_SFMode_For_HSPDSCHDualStream, create_dissector_handle(dissect_MIMO_SFMode_For_HSPDSCHDualStream_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MIMO_SFMode_Supported_For_HSPDSCHDualStream, create_dissector_handle(dissect_MIMO_SFMode_For_HSPDSCHDualStream_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UE_Selected_MBMS_Service_Information, create_dissector_handle(dissect_UE_Selected_MBMS_Service_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MultiCarrier_HSDSCH_Physical_Layer_Category, create_dissector_handle(dissect_LCRTDD_HSDSCH_Physical_Layer_Category_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Common_E_DCH_HSDPCCH_Capability, create_dissector_handle(dissect_Common_E_DCH_HSDPCCH_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DL_RLC_PDU_Size_Format, create_dissector_handle(dissect_DL_RLC_PDU_Size_Format_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSSICH_ReferenceSignal_InformationModifyLCR, create_dissector_handle(dissect_HSSICH_ReferenceSignal_InformationModifyLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_schedulingPriorityIndicator, create_dissector_handle(dissect_SchedulingPriorityIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TimeSlotMeasurementValueListLCR, create_dissector_handle(dissect_TimeSlotMeasurementValueListLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UE_SupportIndicatorExtension, create_dissector_handle(dissect_UE_SupportIndicatorExtension_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Single_Stream_MIMO_ActivationIndicator, create_dissector_handle(dissect_Single_Stream_MIMO_ActivationIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Single_Stream_MIMO_Capability, create_dissector_handle(dissect_Single_Stream_MIMO_Capability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Single_Stream_MIMO_Mode_Indicator, create_dissector_handle(dissect_Single_Stream_MIMO_Mode_Indicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Dual_Band_Capability_Info, create_dissector_handle(dissect_Dual_Band_Capability_Info_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UE_AggregateMaximumBitRate, create_dissector_handle(dissect_UE_AggregateMaximumBitRate_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UE_AggregateMaximumBitRate_Enforcement_Indicator, create_dissector_handle(dissect_UE_AggregateMaximumBitRate_Enforcement_Indicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MIMO_Power_Offset_For_S_CPICH_Capability, create_dissector_handle(dissect_MIMO_PowerOffsetForS_CPICHCapability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MIMO_PilotConfigurationExtension, create_dissector_handle(dissect_MIMO_PilotConfigurationExtension_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TxDiversityOnDLControlChannelsByMIMOUECapability, create_dissector_handle(dissect_TxDiversityOnDLControlChannelsByMIMOUECapability_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_NumberOfReportedCellPortionsLCR, create_dissector_handle(dissect_NumberOfReportedCellPortionsLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_CellPortion_CapabilityLCR, create_dissector_handle(dissect_CellPortion_CapabilityLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_EDCH_Cell_Information_RL_Setup_Req, create_dissector_handle(dissect_Additional_EDCH_Setup_Info_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_EDCH_Cell_Information_Response, create_dissector_handle(dissect_Additional_EDCH_Cell_Information_Response_List_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_EDCH_Cell_Information_RL_Add_Req, create_dissector_handle(dissect_Additional_EDCH_Cell_Information_RL_Add_Req_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_EDCH_Cell_Information_Response_RL_Add, create_dissector_handle(dissect_Additional_EDCH_Cell_Information_Response_RL_Add_List_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_EDCH_Cell_Information_RL_Reconf_Prep, create_dissector_handle(dissect_Additional_EDCH_Cell_Information_RL_Reconf_Prep_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_EDCH_Cell_Information_RL_Reconf_Req, create_dissector_handle(dissect_Additional_EDCH_Cell_Information_RL_Reconf_Req_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_EDCH_Cell_Information_Bearer_Rearrangement, create_dissector_handle(dissect_Additional_EDCH_Cell_Information_Bearer_Rearrangement_List_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_EDCH_Cell_Information_RL_Param_Upd, create_dissector_handle(dissect_Additional_EDCH_Cell_Information_RL_Param_Upd_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_EDCH_Preconfiguration_Information, create_dissector_handle(dissect_Additional_EDCH_Preconfiguration_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_EDCH_Indicator, create_dissector_handle(dissect_NULL_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HS_DSCH_SPS_Reservation_Indicator, create_dissector_handle(dissect_SPS_Reservation_Indicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_DCH_SPS_Reservation_Indicator, create_dissector_handle(dissect_SPS_Reservation_Indicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MultipleFreq_HARQ_MemoryPartitioning_InformationList, create_dissector_handle(dissect_MultipleFreq_HARQ_MemoryPartitioning_InformationList_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_Ext, create_dissector_handle(dissect_Ul_common_E_DCH_MACflow_Specific_InfoResponseListLCR_Ext_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_RepetitionPeriodIndex, create_dissector_handle(dissect_RepetitionPeriodIndex_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MidambleShiftLCR, create_dissector_handle(dissect_MidambleShiftLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_MaxHSDSCH_HSSCCH_Power_per_CELLPORTION, create_dissector_handle(dissect_MaxHSDSCH_HSSCCH_Power_per_CELLPORTION_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DormantModeIndicator, create_dissector_handle(dissect_DormantModeIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DiversityMode, create_dissector_handle(dissect_DiversityMode_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TransmitDiversityIndicator, create_dissector_handle(dissect_TransmitDiversityIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_NonCellSpecificTxDiversity, create_dissector_handle(dissect_NonCellSpecificTxDiversity_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Cell_Capability_Container, create_dissector_handle(dissect_Cell_Capability_Container_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_RNTI_List_Request, create_dissector_handle(dissect_NULL_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_RNTI_List, create_dissector_handle(dissect_E_RNTI_List_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PowerControlGAP_For_CellFACHLCR, create_dissector_handle(dissect_ControlGAP_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UL_Synchronisation_Parameters_For_FACHLCR, create_dissector_handle(dissect_UL_Synchronisation_Parameters_LCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HS_DSCH_SPS_Operation_Indicator, create_dissector_handle(dissect_HS_DSCH_SPS_Operation_Indicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Out_of_Sychronization_Window, create_dissector_handle(dissect_Out_of_Sychronization_Window_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst, create_dissector_handle(dissect_Max_RTWP_perUARFCN_Information_LCR_PSCH_ReconfRqst_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_E_HICH_TimeOffset_ReconfFailureTDD, create_dissector_handle(dissect_E_HICH_TimeOffset_ReconfFailureTDD_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_HSSCCH_TPC_StepSize, create_dissector_handle(dissect_TDD_TPC_DownlinkStepSize_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_TS0_CapabilityLCR, create_dissector_handle(dissect_TS0_CapabilityLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_UE_TS0_CapabilityLCR, create_dissector_handle(dissect_UE_TS0_CapabilityLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Common_System_Information_ResponseLCR, create_dissector_handle(dissect_Common_System_Information_ResponseLCR_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Additional_EDCH_Cell_Information_ResponseRLReconf, create_dissector_handle(dissect_Additional_EDCH_Cell_Information_Response_RLReconf_List_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Multicell_EDCH_InformationItemIEs, create_dissector_handle(dissect_Multicell_EDCH_InformationItemIEs_PDU, proto_nbap));
  dissector_add_uint("nbap.ies", id_Multicell_EDCH_RL_Specific_InformationItemIEs, create_dissector_handle(dissect_Multicell_EDCH_RL_Specific_InformationItemIEs_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Add_To_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext, create_dissector_handle(dissect_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Modify_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext, create_dissector_handle(dissect_Modify_Non_HS_SCCH_Associated_HS_SICH_InformationList_Ext_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext, create_dissector_handle(dissect_Delete_From_Non_HS_SCCH_Associated_HS_SICH_Resource_Pool_LCR_PSCH_ReconfRqst_Ext_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Initial_DL_Transmission_Power, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Maximum_DL_Power, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_Minimum_DL_Power, create_dissector_handle(dissect_DL_Power_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DCH_MeasurementOccasion_Information, create_dissector_handle(dissect_DCH_MeasurementOccasion_Information_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_AssociatedPhsicalChannelID, create_dissector_handle(dissect_CommonPhysicalChannelID_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_DGNSS_ValidityPeriod, create_dissector_handle(dissect_DGNSS_ValidityPeriod_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PhysicalChannelID_for_CommonERNTI_RequestedIndicator, create_dissector_handle(dissect_PhysicalChannelID_for_CommonERNTI_RequestedIndicator_PDU, proto_nbap));
  dissector_add_uint("nbap.extension", id_PrecodingWeightSetRestriction, create_dissector_handle(dissect_PrecodingWeightSetRestriction_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-cellSetup/fdd", create_dissector_handle(dissect_CellSetupRequestFDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-cellSetup/fdd", create_dissector_handle(dissect_CellSetupResponse_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-cellSetup/fdd", create_dissector_handle(dissect_CellSetupFailure_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-cellSetup/tdd", create_dissector_handle(dissect_CellSetupRequestTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-cellSetup/tdd", create_dissector_handle(dissect_CellSetupResponse_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-cellSetup/tdd", create_dissector_handle(dissect_CellSetupFailure_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-cellReconfiguration/fdd", create_dissector_handle(dissect_CellReconfigurationRequestFDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-cellReconfiguration/fdd", create_dissector_handle(dissect_CellReconfigurationResponse_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-cellReconfiguration/fdd", create_dissector_handle(dissect_CellReconfigurationFailure_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-cellReconfiguration/tdd", create_dissector_handle(dissect_CellReconfigurationRequestTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-cellReconfiguration/tdd", create_dissector_handle(dissect_CellReconfigurationResponse_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-cellReconfiguration/tdd", create_dissector_handle(dissect_CellReconfigurationFailure_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-cellDeletion/common", create_dissector_handle(dissect_CellDeletionRequest_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-cellDeletion/common", create_dissector_handle(dissect_CellDeletionResponse_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-commonTransportChannelSetup/fdd", create_dissector_handle(dissect_CommonTransportChannelSetupRequestFDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-commonTransportChannelSetup/fdd", create_dissector_handle(dissect_CommonTransportChannelSetupResponse_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-commonTransportChannelSetup/fdd", create_dissector_handle(dissect_CommonTransportChannelSetupFailure_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-commonTransportChannelSetup/tdd", create_dissector_handle(dissect_CommonTransportChannelSetupRequestTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-commonTransportChannelSetup/tdd", create_dissector_handle(dissect_CommonTransportChannelSetupResponse_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-commonTransportChannelSetup/tdd", create_dissector_handle(dissect_CommonTransportChannelSetupFailure_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-commonTransportChannelReconfigure/fdd", create_dissector_handle(dissect_CommonTransportChannelReconfigurationRequestFDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-commonTransportChannelReconfigure/fdd", create_dissector_handle(dissect_CommonTransportChannelReconfigurationResponse_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-commonTransportChannelReconfigure/fdd", create_dissector_handle(dissect_CommonTransportChannelReconfigurationFailure_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-commonTransportChannelReconfigure/tdd", create_dissector_handle(dissect_CommonTransportChannelReconfigurationRequestTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-commonTransportChannelReconfigure/tdd", create_dissector_handle(dissect_CommonTransportChannelReconfigurationResponse_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-commonTransportChannelReconfigure/tdd", create_dissector_handle(dissect_CommonTransportChannelReconfigurationFailure_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-commonTransportChannelDelete/common", create_dissector_handle(dissect_CommonTransportChannelDeletionRequest_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-commonTransportChannelDelete/common", create_dissector_handle(dissect_CommonTransportChannelDeletionResponse_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-audit/common", create_dissector_handle(dissect_AuditRequest_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-audit/common", create_dissector_handle(dissect_AuditResponse_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-audit/common", create_dissector_handle(dissect_AuditFailure_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-blockResource/common", create_dissector_handle(dissect_BlockResourceRequest_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-blockResource/common", create_dissector_handle(dissect_BlockResourceResponse_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-blockResource/common", create_dissector_handle(dissect_BlockResourceFailure_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-radioLinkSetup/fdd", create_dissector_handle(dissect_RadioLinkSetupRequestFDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-radioLinkSetup/fdd", create_dissector_handle(dissect_RadioLinkSetupResponseFDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-radioLinkSetup/fdd", create_dissector_handle(dissect_RadioLinkSetupFailureFDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-radioLinkSetup/tdd", create_dissector_handle(dissect_RadioLinkSetupRequestTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-radioLinkSetup/tdd", create_dissector_handle(dissect_RadioLinkSetupResponseTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-radioLinkSetup/tdd", create_dissector_handle(dissect_RadioLinkSetupFailureTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-systemInformationUpdate/common", create_dissector_handle(dissect_SystemInformationUpdateRequest_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-systemInformationUpdate/common", create_dissector_handle(dissect_SystemInformationUpdateResponse_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-systemInformationUpdate/common", create_dissector_handle(dissect_SystemInformationUpdateFailure_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-reset/common", create_dissector_handle(dissect_ResetRequest_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-reset/common", create_dissector_handle(dissect_ResetResponse_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-commonMeasurementInitiation/common", create_dissector_handle(dissect_CommonMeasurementInitiationRequest_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-commonMeasurementInitiation/common", create_dissector_handle(dissect_CommonMeasurementInitiationResponse_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-commonMeasurementInitiation/common", create_dissector_handle(dissect_CommonMeasurementInitiationFailure_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-radioLinkAddition/fdd", create_dissector_handle(dissect_RadioLinkAdditionRequestFDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-radioLinkAddition/fdd", create_dissector_handle(dissect_RadioLinkAdditionResponseFDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-radioLinkAddition/fdd", create_dissector_handle(dissect_RadioLinkAdditionFailureFDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-radioLinkAddition/tdd", create_dissector_handle(dissect_RadioLinkAdditionRequestTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-radioLinkAddition/tdd", create_dissector_handle(dissect_RadioLinkAdditionResponseTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-radioLinkAddition/tdd", create_dissector_handle(dissect_RadioLinkAdditionFailureTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-radioLinkDeletion/common", create_dissector_handle(dissect_RadioLinkDeletionRequest_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-radioLinkDeletion/common", create_dissector_handle(dissect_RadioLinkDeletionResponse_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-synchronisedRadioLinkReconfigurationPreparation/fdd", create_dissector_handle(dissect_RadioLinkReconfigurationPrepareFDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-synchronisedRadioLinkReconfigurationPreparation/fdd", create_dissector_handle(dissect_RadioLinkReconfigurationReady_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-synchronisedRadioLinkReconfigurationPreparation/fdd", create_dissector_handle(dissect_RadioLinkReconfigurationFailure_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-synchronisedRadioLinkReconfigurationPreparation/tdd", create_dissector_handle(dissect_RadioLinkReconfigurationPrepareTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-synchronisedRadioLinkReconfigurationPreparation/tdd", create_dissector_handle(dissect_RadioLinkReconfigurationReady_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-synchronisedRadioLinkReconfigurationPreparation/tdd", create_dissector_handle(dissect_RadioLinkReconfigurationFailure_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-unSynchronisedRadioLinkReconfiguration/fdd", create_dissector_handle(dissect_RadioLinkReconfigurationRequestFDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-unSynchronisedRadioLinkReconfiguration/fdd", create_dissector_handle(dissect_RadioLinkReconfigurationResponse_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-unSynchronisedRadioLinkReconfiguration/fdd", create_dissector_handle(dissect_RadioLinkReconfigurationFailure_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-unSynchronisedRadioLinkReconfiguration/tdd", create_dissector_handle(dissect_RadioLinkReconfigurationRequestTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-unSynchronisedRadioLinkReconfiguration/tdd", create_dissector_handle(dissect_RadioLinkReconfigurationResponse_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-unSynchronisedRadioLinkReconfiguration/tdd", create_dissector_handle(dissect_RadioLinkReconfigurationFailure_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-dedicatedMeasurementInitiation/common", create_dissector_handle(dissect_DedicatedMeasurementInitiationRequest_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-dedicatedMeasurementInitiation/common", create_dissector_handle(dissect_DedicatedMeasurementInitiationResponse_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-dedicatedMeasurementInitiation/common", create_dissector_handle(dissect_DedicatedMeasurementInitiationFailure_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-physicalSharedChannelReconfiguration/fdd", create_dissector_handle(dissect_PhysicalSharedChannelReconfigurationRequestFDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-physicalSharedChannelReconfiguration/fdd", create_dissector_handle(dissect_PhysicalSharedChannelReconfigurationResponse_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-physicalSharedChannelReconfiguration/fdd", create_dissector_handle(dissect_PhysicalSharedChannelReconfigurationFailure_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-physicalSharedChannelReconfiguration/tdd", create_dissector_handle(dissect_PhysicalSharedChannelReconfigurationRequestTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-physicalSharedChannelReconfiguration/tdd", create_dissector_handle(dissect_PhysicalSharedChannelReconfigurationResponse_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-physicalSharedChannelReconfiguration/tdd", create_dissector_handle(dissect_PhysicalSharedChannelReconfigurationFailure_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-informationExchangeInitiation/common", create_dissector_handle(dissect_InformationExchangeInitiationRequest_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-informationExchangeInitiation/common", create_dissector_handle(dissect_InformationExchangeInitiationResponse_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-informationExchangeInitiation/common", create_dissector_handle(dissect_InformationExchangeInitiationFailure_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-cellSynchronisationInitiation/tdd", create_dissector_handle(dissect_CellSynchronisationInitiationRequestTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-cellSynchronisationInitiation/tdd", create_dissector_handle(dissect_CellSynchronisationInitiationResponseTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-cellSynchronisationInitiation/tdd", create_dissector_handle(dissect_CellSynchronisationInitiationFailureTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-cellSynchronisationReconfiguration/tdd", create_dissector_handle(dissect_CellSynchronisationReconfigurationRequestTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-cellSynchronisationReconfiguration/tdd", create_dissector_handle(dissect_CellSynchronisationReconfigurationResponseTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-cellSynchronisationReconfiguration/tdd", create_dissector_handle(dissect_CellSynchronisationReconfigurationFailureTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-cellSynchronisationAdjustment/tdd", create_dissector_handle(dissect_CellSynchronisationAdjustmentRequestTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.sout", "id-cellSynchronisationAdjustment/tdd", create_dissector_handle(dissect_CellSynchronisationAdjustmentResponseTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.uout", "id-cellSynchronisationAdjustment/tdd", create_dissector_handle(dissect_CellSynchronisationAdjustmentFailureTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-resourceStatusIndication/common", create_dissector_handle(dissect_ResourceStatusIndication_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-auditRequired/common", create_dissector_handle(dissect_AuditRequiredIndication_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-commonMeasurementReport/common", create_dissector_handle(dissect_CommonMeasurementReport_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-commonMeasurementTermination/common", create_dissector_handle(dissect_CommonMeasurementTerminationRequest_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-commonMeasurementFailure/common", create_dissector_handle(dissect_CommonMeasurementFailureIndication_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-synchronisedRadioLinkReconfigurationCommit/common", create_dissector_handle(dissect_RadioLinkReconfigurationCommit_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-synchronisedRadioLinkReconfigurationCancellation/common", create_dissector_handle(dissect_RadioLinkReconfigurationCancel_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-radioLinkFailure/common", create_dissector_handle(dissect_RadioLinkFailureIndication_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-radioLinkPreemption/common", create_dissector_handle(dissect_RadioLinkPreemptionRequiredIndication_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-radioLinkRestoration/common", create_dissector_handle(dissect_RadioLinkRestoreIndication_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-dedicatedMeasurementReport/common", create_dissector_handle(dissect_DedicatedMeasurementReport_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-dedicatedMeasurementTermination/common", create_dissector_handle(dissect_DedicatedMeasurementTerminationRequest_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-dedicatedMeasurementFailure/common", create_dissector_handle(dissect_DedicatedMeasurementFailureIndication_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-downlinkPowerControl/fdd", create_dissector_handle(dissect_DL_PowerControlRequest_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-downlinkPowerTimeslotControl/tdd", create_dissector_handle(dissect_DL_PowerTimeslotControlRequest_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-compressedModeCommand/fdd", create_dissector_handle(dissect_CompressedModeCommand_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-unblockResource/common", create_dissector_handle(dissect_UnblockResourceIndication_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-errorIndicationForDedicated/common", create_dissector_handle(dissect_ErrorIndication_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-errorIndicationForCommon/common", create_dissector_handle(dissect_ErrorIndication_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-cellSynchronisationReporting/tdd", create_dissector_handle(dissect_CellSynchronisationReportTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-cellSynchronisationTermination/tdd", create_dissector_handle(dissect_CellSynchronisationTerminationRequestTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-cellSynchronisationFailure/tdd", create_dissector_handle(dissect_CellSynchronisationFailureIndicationTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-privateMessageForDedicated/common", create_dissector_handle(dissect_PrivateMessage_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-privateMessageForCommon/common", create_dissector_handle(dissect_PrivateMessage_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-informationReporting/common", create_dissector_handle(dissect_InformationReport_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-informationExchangeTermination/common", create_dissector_handle(dissect_InformationExchangeTerminationRequest_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-informationExchangeFailure/common", create_dissector_handle(dissect_InformationExchangeFailureIndication_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-BearerRearrangement/common", create_dissector_handle(dissect_BearerRearrangementIndication_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-radioLinkActivation/fdd", create_dissector_handle(dissect_RadioLinkActivationCommandFDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-radioLinkActivation/tdd", create_dissector_handle(dissect_RadioLinkActivationCommandTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-radioLinkParameterUpdate/fdd", create_dissector_handle(dissect_RadioLinkParameterUpdateIndicationFDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-radioLinkParameterUpdate/tdd", create_dissector_handle(dissect_RadioLinkParameterUpdateIndicationTDD_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-mBMSNotificationUpdate/common", create_dissector_handle(dissect_MBMSNotificationUpdateCommand_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-uEStatusUpdate/common", create_dissector_handle(dissect_UEStatusUpdateCommand_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-secondaryULFrequencyReporting/fdd", create_dissector_handle(dissect_SecondaryULFrequencyReport_PDU, proto_nbap));
  dissector_add_string("nbap.proc.imsg", "id-secondaryULFrequencyUpdate/fdd", create_dissector_handle(dissect_SecondaryULFrequencyUpdateIndication_PDU, proto_nbap));

}

