/* packet-parlay.c
 *
 * Routines for IDL dissection
 *
 * Autogenerated from idl2wrs
 * Copyright 2001 Frank Singleton <frank.singleton@ericsson.com>
 */


/*
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 */

/*
 * SPDX-License-Identifier: GPL-2.0-or-later
 */


#include "config.h"

#include <string.h>
#include <epan/packet.h>
#include <epan/proto.h>
#include <epan/proto_data.h>
#include "packet-giop.h"
#include <epan/expert.h>

#include "ws_diag_control.h"
#include "ws_compiler_tests.h"

#ifdef _MSC_VER
/* disable warning: "unreference local variable" */
#pragma warning(disable:4101)
#endif

/* XXX this should be autogenerated, or the warnings fixed in the generator */
DIAG_OFF(unused-function)
DIAG_OFF(unused-variable)
#if WS_IS_AT_LEAST_GNUC_VERSION(6,0)
DIAG_OFF(unused-const-variable)
#endif

void proto_register_giop_parlay(void);
void proto_reg_handoff_giop_parlay(void);

/* Initialise the protocol and subtree pointers */
static int proto_parlay = -1;
static gint ett_parlay = -1;


/* Initialise the initial Alignment */
static guint32  boundary = GIOP_HEADER_SIZE;  /* initial value */

static int hf_operationrequest = -1;/* Request_Operation field */

        /* Operation filters */
static int hf_org_csapi_IpService_setCallbackWithSessionID_sessionID = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpInitial_initiateAuthentication_authType = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpInitial_initiateAuthenticationWithVersion_authType = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpInitial_initiateAuthenticationWithVersion_frameworkVersion = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAuthentication_requestAccess_accessType = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpClientAccess_terminateAccess_terminationText = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpClientAccess_terminateAccess_signingAlgorithm = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpClientAccess_terminateAccess_digitalSignature_loop = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpClientAccess_terminateAccess_digitalSignature = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_obtainInterface_interfaceName = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_obtainInterfaceWithCallback_interfaceName = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_endAccess_endAccessProperties_loop = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_listInterfaces_return_loop = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_listInterfaces_return = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_releaseInterface_interfaceName = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_selectSigningAlgorithm_return = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_selectSigningAlgorithm_signingAlgorithmCaps = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_terminateAccess_terminationText = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_terminateAccess_digitalSignature_loop = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_terminateAccess_digitalSignature = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_relinquishInterface_interfaceName = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_relinquishInterface_terminationText = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_relinquishInterface_digitalSignature_loop = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_relinquishInterface_digitalSignature = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_return_loop = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_return = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_challenge_loop = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_challenge = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_return_loop = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_return = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_challenge_loop = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_challenge = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_selectEncryptionMethod_return = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_selectEncryptionMethod_encryptionCaps = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_return_loop = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_return = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_challenge_loop = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_challenge = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_selectAuthenticationMechanism_return = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_selectAuthenticationMechanism_authMechanismList = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_return_loop = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_return = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_challenge_loop = -1;
static int hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_challenge = -1;
static int hf_org_csapi_fw_fw_application_notification_IpAppEventNotification_reportNotification_assignmentID = -1;
static int hf_org_csapi_fw_fw_application_notification_IpEventNotification_createNotification_return = -1;
static int hf_org_csapi_fw_fw_application_notification_IpEventNotification_destroyNotification_assignmentID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_activityTestRes_activityTestID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_activityTestRes_activityTestResult = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_appActivityTestReq_activityTestID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_fwFaultReportInd_fault = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_fwFaultRecoveryInd_fault = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_svcUnavailableInd_serviceID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_svcUnavailableInd_reason = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordRes_serviceIDs_loop = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordRes_serviceIDs = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_fwUnavailableInd_reason = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_activityTestErr_activityTestID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordErr_faultStatisticsError = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordErr_serviceIDs_loop = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordErr_serviceIDs = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_appUnavailableInd_serviceID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_svcAvailStatusInd_serviceID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_svcAvailStatusInd_reason = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordRes_faultStatsReqID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordRes_serviceIDs_loop = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordRes_serviceIDs = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_faultStatsReqID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_faultStatistics_loop = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_faultStatistics = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_serviceIDs_loop = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_serviceIDs = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordReq_faultStatsReqID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_fwAvailStatusInd_reason = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadRes_loadStatistics_loop = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadErr_loadStatisticsError = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_loadLevelNotification_loadStatistics_loop = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryAppLoadStatsReq_loadStatsReqID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsRes_loadStatsReqID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsRes_loadStatistics_loop = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsErr_loadStatsReqID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsErr_loadStatisticsError = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpLoadManager_reportLoad_loadLevel = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadReq_serviceIDs_loop = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadReq_serviceIDs = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadRes_loadStatistics_loop = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadErr_loadStatisticsError = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpLoadManager_createLoadLevelNotification_serviceIDs_loop = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpLoadManager_createLoadLevelNotification_serviceIDs = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpLoadManager_destroyLoadLevelNotification_serviceIDs_loop = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpLoadManager_destroyLoadLevelNotification_serviceIDs = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpLoadManager_resumeNotification_serviceIDs_loop = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpLoadManager_resumeNotification_serviceIDs = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpLoadManager_suspendNotification_serviceIDs_loop = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpLoadManager_suspendNotification_serviceIDs = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadStatsReq_loadStatsReqID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadStatsReq_serviceIDs_loop = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadStatsReq_serviceIDs = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsRes_loadStatsReqID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsRes_loadStatistics_loop = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsErr_loadStatsReqID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsErr_loadStatisticsError = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppOAM_systemDateTimeQuery_return = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppOAM_systemDateTimeQuery_systemDateAndTime = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpOAM_systemDateTimeQuery_return = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpOAM_systemDateTimeQuery_clientDateAndTime = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpFaultManager_activityTestReq_activityTestID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpFaultManager_activityTestReq_svcID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpFaultManager_appActivityTestRes_activityTestID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpFaultManager_appActivityTestRes_activityTestResult = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpFaultManager_svcUnavailableInd_serviceID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpFaultManager_genFaultStatsRecordReq_serviceIDs_loop = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpFaultManager_genFaultStatsRecordReq_serviceIDs = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpFaultManager_appActivityTestErr_activityTestID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpFaultManager_appUnavailableInd_serviceID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpFaultManager_genFaultStatsRecordErr_faultStatisticsError = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpFaultManager_appAvailStatusInd_reason = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordReq_faultStatsReqID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordReq_serviceIDs_loop = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordReq_serviceIDs = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordRes_faultStatsReqID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordErr_faultStatsReqID = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordErr_faultStatisticsError = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpHeartBeatMgmt_enableHeartBeat_interval = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpHeartBeatMgmt_changeInterval_interval = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppHeartBeatMgmt_enableAppHeartBeat_interval = -1;
static int hf_org_csapi_fw_fw_application_integrity_IpAppHeartBeatMgmt_changeInterval_interval = -1;
static int hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listServiceTypes_return_loop = -1;
static int hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listServiceTypes_return = -1;
static int hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_describeServiceType_name = -1;
static int hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_return_loop = -1;
static int hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_serviceTypeName = -1;
static int hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_desiredPropertyList_loop = -1;
static int hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_max = -1;
static int hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listSubscribedServices_return_loop = -1;
static int hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement_return_loop = -1;
static int hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement_return = -1;
static int hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement_serviceToken = -1;
static int hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement_agreementText = -1;
static int hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement_signingAlgorithm = -1;
static int hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_terminateServiceAgreement_serviceToken = -1;
static int hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_terminateServiceAgreement_terminationText = -1;
static int hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_terminateServiceAgreement_digitalSignature_loop = -1;
static int hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_terminateServiceAgreement_digitalSignature = -1;
static int hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_signServiceAgreement_serviceToken = -1;
static int hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_signServiceAgreement_agreementText = -1;
static int hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_signServiceAgreement_signingAlgorithm = -1;
static int hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_terminateServiceAgreement_serviceToken = -1;
static int hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_terminateServiceAgreement_terminationText = -1;
static int hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_terminateServiceAgreement_digitalSignature_loop = -1;
static int hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_terminateServiceAgreement_digitalSignature = -1;
static int hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_selectService_return = -1;
static int hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_selectService_serviceID = -1;
static int hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_initiateSignServiceAgreement_serviceToken = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_createServiceProfile_return = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_deleteServiceProfile_serviceProfileID = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_assign_sagID = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_assign_serviceProfileID = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_deassign_sagID = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_deassign_serviceProfileID = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_requestConflictInfo_return_loop = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listServiceProfiles_return_loop = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listServiceProfiles_return = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_describeServiceProfile_serviceProfileID = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listAssignedMembers_return_loop = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listAssignedMembers_return = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listAssignedMembers_serviceProfileID = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractManagement_createServiceContract_return = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractManagement_deleteServiceContract_serviceContractID = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_describeServiceContract_serviceContractID = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceContracts_return_loop = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceContracts_return = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceProfiles_return_loop = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceProfiles_return = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceProfiles_serviceContractID = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpEntOpAccountManagement_modifyEntOpAccount_enterpriseOperatorProperties_loop = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_deleteClientApp_clientAppID = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_createSAG_clientAppIDs_loop = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_createSAG_clientAppIDs = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_deleteSAG_sagID = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_addSAGMembers_sagID = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_addSAGMembers_clientAppIDs_loop = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_addSAGMembers_clientAppIDs = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_removeSAGMembers_sagID = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_removeSAGMembers_clientAppIDList_loop = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_removeSAGMembers_clientAppIDList = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_requestConflictInfo_return_loop = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_describeClientApp_clientAppID = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientApps_return_loop = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientApps_return = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_describeSAG_return = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_describeSAG_sagID = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGs_return_loop = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGs_return = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGMembers_return_loop = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGMembers_return = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGMembers_sagID = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientAppMembership_return_loop = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientAppMembership_return = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientAppMembership_clientAppID = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_notification_IpClientEventNotification_reportNotification_assignmentID = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_notification_IpEventNotification_createNotification_return = -1;
static int hf_org_csapi_fw_fw_enterprise_operator_notification_IpEventNotification_destroyNotification_assignmentID = -1;
static int hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listServiceTypes_return_loop = -1;
static int hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listServiceTypes_return = -1;
static int hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_describeServiceType_name = -1;
static int hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_return_loop = -1;
static int hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_serviceTypeName = -1;
static int hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_desiredPropertyList_loop = -1;
static int hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_max = -1;
static int hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listRegisteredServices_return_loop = -1;
static int hf_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_createServiceManager_application = -1;
static int hf_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_createServiceManager_serviceProperties_loop = -1;
static int hf_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_createServiceManager_serviceInstanceID = -1;
static int hf_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_destroyServiceManager_serviceInstance = -1;
static int hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerService_return = -1;
static int hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerService_serviceTypeName = -1;
static int hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerService_servicePropertyList_loop = -1;
static int hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_announceServiceAvailability_serviceID = -1;
static int hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_unregisterService_serviceID = -1;
static int hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_describeService_serviceID = -1;
static int hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_unannounceService_serviceID = -1;
static int hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_return = -1;
static int hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_serviceTypeName = -1;
static int hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_servicePropertyList_loop = -1;
static int hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_extendedServicePropertyList_loop = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadRes_loadStatistics_loop = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadErr_loadStatisticsError = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_loadLevelNotification_loadStatistics_loop = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_querySvcLoadStatsReq_loadStatsReqID = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsRes_loadStatsReqID = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsRes_loadStatistics_loop = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsErr_loadStatsReqID = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsErr_loadStatisticsError = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_reportLoad_loadLevel = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_queryLoadReq_querySubject = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadRes_loadStatistics_loop = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadErr_loadStatisticError = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_createLoadLevelNotification_notificationSubject = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_destroyLoadLevelNotification_notificationSubject = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_suspendNotification_notificationSubject = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_resumeNotification_notificationSubject = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_queryLoadStatsReq_loadStatsReqID = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_queryLoadStatsReq_querySubject = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsRes_loadStatsReqID = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsRes_loadStatistics_loop = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsErr_loadStatsReqID = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsErr_loadStatisticError = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_activityTestRes_activityTestID = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_activityTestRes_activityTestResult = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_svcActivityTestReq_activityTestID = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_fwFaultReportInd_fault = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_fwFaultRecoveryInd_fault = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_fwUnavailableInd_reason = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordRes_recordSubject = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_activityTestErr_activityTestID = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordErr_faultStatisticsError = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordErr_recordSubject = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordReq_serviceIDs_loop = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordReq_serviceIDs = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_appAvailStatusInd_reason = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordRes_faultStatsReqID = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordRes_recordSubject = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordErr_faultStatsReqID = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordErr_faultStatisticsError = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordErr_recordSubject = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordReq_faultStatsReqID = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_fwAvailStatusInd_reason = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_activityTestReq_activityTestID = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_activityTestReq_testSubject = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_svcActivityTestRes_activityTestID = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_svcActivityTestRes_activityTestResult = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordReq_recordSubject = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_svcUnavailableInd_reason = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_svcActivityTestErr_activityTestID = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordRes_serviceIDs_loop = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordRes_serviceIDs = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordErr_faultStatisticsError = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordErr_serviceIDs_loop = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordErr_serviceIDs = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatsRecordErr_faultStatisticsError = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_svcAvailStatusInd_reason = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatisticsRecordReq_faultStatsReqID = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatisticsRecordReq_recordSubject = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatisticsRecordRes_faultStatsReqID = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatisticsRecordErr_faultStatsReqID = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatisticsRecordErr_faultStatisticsError = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcOAM_systemDateTimeQuery_return = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcOAM_systemDateTimeQuery_systemDateAndTime = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwOAM_systemDateTimeQuery_return = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwOAM_systemDateTimeQuery_clientDateAndTime = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwHeartBeatMgmt_enableHeartBeat_interval = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpFwHeartBeatMgmt_changeInterval_interval = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcHeartBeatMgmt_enableSvcHeartBeat_interval = -1;
static int hf_org_csapi_fw_fw_service_integrity_IpSvcHeartBeatMgmt_changeInterval_interval = -1;
static int hf_org_csapi_fw_fw_service_notification_IpFwEventNotification_createNotification_return = -1;
static int hf_org_csapi_fw_fw_service_notification_IpFwEventNotification_destroyNotification_assignmentID = -1;
static int hf_org_csapi_fw_fw_service_notification_IpSvcEventNotification_reportNotification_assignmentID = -1;
static int hf_org_csapi_cc_gccs_IpAppCall_routeRes_callSessionID = -1;
static int hf_org_csapi_cc_gccs_IpAppCall_routeRes_callLegSessionID = -1;
static int hf_org_csapi_cc_gccs_IpAppCall_routeErr_callSessionID = -1;
static int hf_org_csapi_cc_gccs_IpAppCall_routeErr_callLegSessionID = -1;
static int hf_org_csapi_cc_gccs_IpAppCall_getCallInfoRes_callSessionID = -1;
static int hf_org_csapi_cc_gccs_IpAppCall_getCallInfoErr_callSessionID = -1;
static int hf_org_csapi_cc_gccs_IpAppCall_superviseCallRes_callSessionID = -1;
static int hf_org_csapi_cc_gccs_IpAppCall_superviseCallRes_report = -1;
static int hf_org_csapi_cc_gccs_IpAppCall_superviseCallRes_usedTime = -1;
static int hf_org_csapi_cc_gccs_IpAppCall_superviseCallErr_callSessionID = -1;
static int hf_org_csapi_cc_gccs_IpAppCall_callFaultDetected_callSessionID = -1;
static int hf_org_csapi_cc_gccs_IpAppCall_callFaultDetected_fault = -1;
static int hf_org_csapi_cc_gccs_IpAppCall_getMoreDialledDigitsRes_callSessionID = -1;
static int hf_org_csapi_cc_gccs_IpAppCall_getMoreDialledDigitsRes_digits = -1;
static int hf_org_csapi_cc_gccs_IpAppCall_getMoreDialledDigitsErr_callSessionID = -1;
static int hf_org_csapi_cc_gccs_IpAppCall_callEnded_callSessionID = -1;
static int hf_org_csapi_cc_gccs_IpCall_routeReq_return = -1;
static int hf_org_csapi_cc_gccs_IpCall_routeReq_callSessionID = -1;
static int hf_org_csapi_cc_gccs_IpCall_routeReq_responseRequested_loop = -1;
static int hf_org_csapi_cc_gccs_IpCall_routeReq_appInfo_loop = -1;
static int hf_org_csapi_cc_gccs_IpCall_release_callSessionID = -1;
static int hf_org_csapi_cc_gccs_IpCall_deassignCall_callSessionID = -1;
static int hf_org_csapi_cc_gccs_IpCall_getCallInfoReq_callSessionID = -1;
static int hf_org_csapi_cc_gccs_IpCall_getCallInfoReq_callInfoRequested = -1;
static int hf_org_csapi_cc_gccs_IpCall_setCallChargePlan_callSessionID = -1;
static int hf_org_csapi_cc_gccs_IpCall_setAdviceOfCharge_callSessionID = -1;
static int hf_org_csapi_cc_gccs_IpCall_setAdviceOfCharge_tariffSwitch = -1;
static int hf_org_csapi_cc_gccs_IpCall_getMoreDialledDigitsReq_callSessionID = -1;
static int hf_org_csapi_cc_gccs_IpCall_getMoreDialledDigitsReq_length = -1;
static int hf_org_csapi_cc_gccs_IpCall_superviseCallReq_callSessionID = -1;
static int hf_org_csapi_cc_gccs_IpCall_superviseCallReq_time = -1;
static int hf_org_csapi_cc_gccs_IpCall_superviseCallReq_treatment = -1;
static int hf_org_csapi_cc_gccs_IpCall_continueProcessing_callSessionID = -1;
static int hf_org_csapi_cc_gccs_IpAppCallControlManager_callAborted_callReference = -1;
static int hf_org_csapi_cc_gccs_IpAppCallControlManager_callEventNotify_assignmentID = -1;
static int hf_org_csapi_cc_gccs_IpAppCallControlManager_callOverloadEncountered_assignmentID = -1;
static int hf_org_csapi_cc_gccs_IpAppCallControlManager_callOverloadCeased_assignmentID = -1;
static int hf_org_csapi_cc_gccs_IpAppCallControlManager_abortMultipleCalls_callReferenceSet_loop = -1;
static int hf_org_csapi_cc_gccs_IpAppCallControlManager_abortMultipleCalls_callReferenceSet = -1;
static int hf_org_csapi_cc_gccs_IpCallControlManager_enableCallNotification_return = -1;
static int hf_org_csapi_cc_gccs_IpCallControlManager_disableCallNotification_assignmentID = -1;
static int hf_org_csapi_cc_gccs_IpCallControlManager_setCallLoadControl_return = -1;
static int hf_org_csapi_cc_gccs_IpCallControlManager_setCallLoadControl_duration = -1;
static int hf_org_csapi_cc_gccs_IpCallControlManager_changeCallNotification_assignmentID = -1;
static int hf_org_csapi_cc_gccs_IpCallControlManager_getCriteria_return_loop = -1;
static int hf_org_csapi_cc_mpccs_IpAppCallLeg_eventReportRes_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpAppCallLeg_eventReportErr_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpAppCallLeg_attachMediaRes_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpAppCallLeg_attachMediaErr_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpAppCallLeg_detachMediaRes_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpAppCallLeg_detachMediaErr_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpAppCallLeg_getInfoRes_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpAppCallLeg_getInfoErr_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpAppCallLeg_routeErr_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpAppCallLeg_superviseRes_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpAppCallLeg_superviseRes_report = -1;
static int hf_org_csapi_cc_mpccs_IpAppCallLeg_superviseRes_usedTime = -1;
static int hf_org_csapi_cc_mpccs_IpAppCallLeg_superviseErr_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpAppCallLeg_callLegEnded_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpAppCallLeg_callLegEnded_cause = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_createNotification_return = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_destroyNotification_assignmentID = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_changeNotification_assignmentID = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_getNotification_return_loop = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_setCallLoadControl_return = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_setCallLoadControl_duration = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_enableNotifications_return = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_getNextNotification_reset = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_routeReq_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_routeReq_appInfo_loop = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_eventReportReq_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_eventReportReq_eventsRequested_loop = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_release_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_release_cause = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_getInfoReq_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_getInfoReq_callLegInfoRequested = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_getCall_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_attachMediaReq_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_detachMediaReq_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_getCurrentDestinationAddress_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_continueProcessing_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_setChargePlan_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_setAdviceOfCharge_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_setAdviceOfCharge_tariffSwitch = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_superviseReq_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_superviseReq_time = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_superviseReq_treatment = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_deassign_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_getProperties_return_loop = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_getProperties_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_getProperties_propertyNames_loop = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_getProperties_propertyNames = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_setProperties_callLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpCallLeg_setProperties_properties_loop = -1;
static int hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_getInfoRes_callSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_getInfoErr_callSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_superviseRes_callSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_superviseRes_report = -1;
static int hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_superviseRes_usedTime = -1;
static int hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_superviseErr_callSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_callEnded_callSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_createAndRouteCallLegErr_callSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCall_getCallLegs_return_loop = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCall_getCallLegs_callSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCall_createCallLeg_callSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_callSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_eventsRequested_loop = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_appInfo_loop = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCall_release_callSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCall_release_cause = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCall_deassignCall_callSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCall_getInfoReq_callSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCall_getInfoReq_callInfoRequested = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCall_setChargePlan_callSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCall_setAdviceOfCharge_callSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCall_setAdviceOfCharge_tariffSwitch = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCall_superviseReq_callSessionID = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCall_superviseReq_time = -1;
static int hf_org_csapi_cc_mpccs_IpMultiPartyCall_superviseReq_treatment = -1;
static int hf_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_reportNotification_callLegReferenceSet_loop = -1;
static int hf_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_reportNotification_assignmentID = -1;
static int hf_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_callAborted_callReference = -1;
static int hf_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_callOverloadEncountered_assignmentID = -1;
static int hf_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_callOverloadCeased_assignmentID = -1;
static int hf_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_abortMultipleCalls_callReferenceSet_loop = -1;
static int hf_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_abortMultipleCalls_callReferenceSet = -1;
static int hf_org_csapi_cc_mmccs_IpAppMultiMediaCall_superviseVolumeRes_callSessionID = -1;
static int hf_org_csapi_cc_mmccs_IpAppMultiMediaCall_superviseVolumeRes_report = -1;
static int hf_org_csapi_cc_mmccs_IpAppMultiMediaCall_superviseVolumeRes_qualityOfService = -1;
static int hf_org_csapi_cc_mmccs_IpAppMultiMediaCall_superviseVolumeErr_callSessionID = -1;
static int hf_org_csapi_cc_mmccs_IpAppMultiMediaCallLeg_mediaStreamMonitorRes_callLegSessionID = -1;
static int hf_org_csapi_cc_mmccs_IpAppMultiMediaCallLeg_mediaStreamMonitorRes_streams_loop = -1;
static int hf_org_csapi_cc_mmccs_IpAppMultiMediaCallLeg_mediaStreamMonitorRes_type = -1;
static int hf_org_csapi_cc_mmccs_IpMultiMediaStream_subtract_mediaStreamSessionID = -1;
static int hf_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamAllow_callLegSessionID = -1;
static int hf_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamAllow_mediaStreamList_loop = -1;
static int hf_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamAllow_mediaStreamList = -1;
static int hf_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamMonitorReq_callLegSessionID = -1;
static int hf_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamMonitorReq_mediaStreamEventCriteria_loop = -1;
static int hf_org_csapi_cc_mmccs_IpMultiMediaCallLeg_getMediaStreams_return_loop = -1;
static int hf_org_csapi_cc_mmccs_IpMultiMediaCallLeg_getMediaStreams_callLegSessionID = -1;
static int hf_org_csapi_cc_mmccs_IpMultiMediaCall_superviseVolumeReq_callSessionID = -1;
static int hf_org_csapi_cc_mmccs_IpMultiMediaCall_superviseVolumeReq_treatment = -1;
static int hf_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_callLegReferenceSet_loop = -1;
static int hf_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_mediaStreams_loop = -1;
static int hf_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_type = -1;
static int hf_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_qualityOfService = -1;
static int hf_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_assignmentID = -1;
static int hf_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_createMediaNotification_return = -1;
static int hf_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_destroyMediaNotification_assignmentID = -1;
static int hf_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_changeMediaNotification_assignmentID = -1;
static int hf_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_getMediaNotification_return_loop = -1;
static int hf_org_csapi_cc_cccs_IpAppSubConfCall_chairSelection_subConferenceSessionID = -1;
static int hf_org_csapi_cc_cccs_IpAppSubConfCall_chairSelection_callLegSessionID = -1;
static int hf_org_csapi_cc_cccs_IpAppSubConfCall_floorRequest_subConferenceSessionID = -1;
static int hf_org_csapi_cc_cccs_IpAppSubConfCall_floorRequest_callLegSessionID = -1;
static int hf_org_csapi_cc_cccs_IpAppConfCall_partyJoined_conferenceSessionID = -1;
static int hf_org_csapi_cc_cccs_IpAppConfCall_leaveMonitorRes_conferenceSessionID = -1;
static int hf_org_csapi_cc_cccs_IpAppConfCall_leaveMonitorRes_callLeg = -1;
static int hf_org_csapi_cc_cccs_IpConfCallControlManager_createConference_numberOfSubConferences = -1;
static int hf_org_csapi_cc_cccs_IpConfCallControlManager_createConference_numberOfParticipants = -1;
static int hf_org_csapi_cc_cccs_IpConfCallControlManager_createConference_duration = -1;
static int hf_org_csapi_cc_cccs_IpConfCallControlManager_reserveResources_startTime = -1;
static int hf_org_csapi_cc_cccs_IpConfCallControlManager_reserveResources_numberOfParticipants = -1;
static int hf_org_csapi_cc_cccs_IpConfCallControlManager_reserveResources_duration = -1;
static int hf_org_csapi_cc_cccs_IpSubConfCall_splitSubConference_subConferenceSessionID = -1;
static int hf_org_csapi_cc_cccs_IpSubConfCall_splitSubConference_callLegList_loop = -1;
static int hf_org_csapi_cc_cccs_IpSubConfCall_splitSubConference_callLegList = -1;
static int hf_org_csapi_cc_cccs_IpSubConfCall_mergeSubConference_subConferenceCallSessionID = -1;
static int hf_org_csapi_cc_cccs_IpSubConfCall_mergeSubConference_targetSubConferenceCall = -1;
static int hf_org_csapi_cc_cccs_IpSubConfCall_moveCallLeg_subConferenceCallSessionID = -1;
static int hf_org_csapi_cc_cccs_IpSubConfCall_moveCallLeg_targetSubConferenceCall = -1;
static int hf_org_csapi_cc_cccs_IpSubConfCall_moveCallLeg_callLeg = -1;
static int hf_org_csapi_cc_cccs_IpSubConfCall_inspectVideo_subConferenceSessionID = -1;
static int hf_org_csapi_cc_cccs_IpSubConfCall_inspectVideo_inspectedCallLeg = -1;
static int hf_org_csapi_cc_cccs_IpSubConfCall_inspectVideoCancel_subConferenceSessionID = -1;
static int hf_org_csapi_cc_cccs_IpSubConfCall_appointSpeaker_subConferenceSessionID = -1;
static int hf_org_csapi_cc_cccs_IpSubConfCall_appointSpeaker_speakerCallLeg = -1;
static int hf_org_csapi_cc_cccs_IpSubConfCall_chairSelection_subConferenceSessionID = -1;
static int hf_org_csapi_cc_cccs_IpSubConfCall_chairSelection_chairCallLeg = -1;
static int hf_org_csapi_cc_cccs_IpSubConfCall_changeConferencePolicy_subConferenceSessionID = -1;
static int hf_org_csapi_cc_cccs_IpConfCall_getSubConferences_return_loop = -1;
static int hf_org_csapi_cc_cccs_IpConfCall_getSubConferences_conferenceSessionID = -1;
static int hf_org_csapi_cc_cccs_IpConfCall_createSubConference_conferenceSessionID = -1;
static int hf_org_csapi_cc_cccs_IpConfCall_leaveMonitorReq_conferenceSessionID = -1;
static int hf_org_csapi_cc_cccs_IpConfCall_getConferenceAddress_conferenceSessionID = -1;
static int hf_org_csapi_ui_IpAppUI_sendInfoRes_userInteractionSessionID = -1;
static int hf_org_csapi_ui_IpAppUI_sendInfoRes_assignmentID = -1;
static int hf_org_csapi_ui_IpAppUI_sendInfoRes_response = -1;
static int hf_org_csapi_ui_IpAppUI_sendInfoErr_userInteractionSessionID = -1;
static int hf_org_csapi_ui_IpAppUI_sendInfoErr_assignmentID = -1;
static int hf_org_csapi_ui_IpAppUI_sendInfoErr_error = -1;
static int hf_org_csapi_ui_IpAppUI_sendInfoAndCollectRes_userInteractionSessionID = -1;
static int hf_org_csapi_ui_IpAppUI_sendInfoAndCollectRes_assignmentID = -1;
static int hf_org_csapi_ui_IpAppUI_sendInfoAndCollectRes_response = -1;
static int hf_org_csapi_ui_IpAppUI_sendInfoAndCollectRes_collectedInfo = -1;
static int hf_org_csapi_ui_IpAppUI_sendInfoAndCollectErr_userInteractionSessionID = -1;
static int hf_org_csapi_ui_IpAppUI_sendInfoAndCollectErr_assignmentID = -1;
static int hf_org_csapi_ui_IpAppUI_sendInfoAndCollectErr_error = -1;
static int hf_org_csapi_ui_IpAppUI_userInteractionFaultDetected_userInteractionSessionID = -1;
static int hf_org_csapi_ui_IpAppUI_userInteractionFaultDetected_fault = -1;
static int hf_org_csapi_ui_IpAppUICall_recordMessageRes_userInteractionSessionID = -1;
static int hf_org_csapi_ui_IpAppUICall_recordMessageRes_assignmentID = -1;
static int hf_org_csapi_ui_IpAppUICall_recordMessageRes_response = -1;
static int hf_org_csapi_ui_IpAppUICall_recordMessageRes_messageID = -1;
static int hf_org_csapi_ui_IpAppUICall_recordMessageErr_userInteractionSessionID = -1;
static int hf_org_csapi_ui_IpAppUICall_recordMessageErr_assignmentID = -1;
static int hf_org_csapi_ui_IpAppUICall_recordMessageErr_error = -1;
static int hf_org_csapi_ui_IpAppUICall_deleteMessageRes_usrInteractionSessionID = -1;
static int hf_org_csapi_ui_IpAppUICall_deleteMessageRes_response = -1;
static int hf_org_csapi_ui_IpAppUICall_deleteMessageRes_assignmentID = -1;
static int hf_org_csapi_ui_IpAppUICall_deleteMessageErr_usrInteractionSessionID = -1;
static int hf_org_csapi_ui_IpAppUICall_deleteMessageErr_error = -1;
static int hf_org_csapi_ui_IpAppUICall_deleteMessageErr_assignmentID = -1;
static int hf_org_csapi_ui_IpAppUICall_abortActionRes_userInteractionSessionID = -1;
static int hf_org_csapi_ui_IpAppUICall_abortActionRes_assignmentID = -1;
static int hf_org_csapi_ui_IpAppUICall_abortActionErr_userInteractionSessionID = -1;
static int hf_org_csapi_ui_IpAppUICall_abortActionErr_assignmentID = -1;
static int hf_org_csapi_ui_IpAppUICall_abortActionErr_error = -1;
static int hf_org_csapi_ui_IpAppUICall_getMessageRes_userInteractionSessionID = -1;
static int hf_org_csapi_ui_IpAppUICall_getMessageRes_assignmentID = -1;
static int hf_org_csapi_ui_IpAppUICall_getMessageErr_userInteractionSessionID = -1;
static int hf_org_csapi_ui_IpAppUICall_getMessageErr_assignmentID = -1;
static int hf_org_csapi_ui_IpAppUICall_getMessageErr_error = -1;
static int hf_org_csapi_ui_IpAppUIManager_reportEventNotification_assignmentID = -1;
static int hf_org_csapi_ui_IpAppUIManager_abortMultipleUserInteractions_userInteractionSet_loop = -1;
static int hf_org_csapi_ui_IpUI_sendInfoReq_return = -1;
static int hf_org_csapi_ui_IpUI_sendInfoReq_userInteractionSessionID = -1;
static int hf_org_csapi_ui_IpUI_sendInfoReq_language = -1;
static int hf_org_csapi_ui_IpUI_sendInfoReq_variableInfo_loop = -1;
static int hf_org_csapi_ui_IpUI_sendInfoReq_repeatIndicator = -1;
static int hf_org_csapi_ui_IpUI_sendInfoReq_responseRequested = -1;
static int hf_org_csapi_ui_IpUI_sendInfoAndCollectReq_return = -1;
static int hf_org_csapi_ui_IpUI_sendInfoAndCollectReq_userInteractionSessionID = -1;
static int hf_org_csapi_ui_IpUI_sendInfoAndCollectReq_language = -1;
static int hf_org_csapi_ui_IpUI_sendInfoAndCollectReq_variableInfo_loop = -1;
static int hf_org_csapi_ui_IpUI_sendInfoAndCollectReq_responseRequested = -1;
static int hf_org_csapi_ui_IpUI_release_userInteractionSessionID = -1;
static int hf_org_csapi_ui_IpUI_setOriginatingAddress_userInteractionSessionID = -1;
static int hf_org_csapi_ui_IpUI_setOriginatingAddress_origin = -1;
static int hf_org_csapi_ui_IpUI_getOriginatingAddress_return = -1;
static int hf_org_csapi_ui_IpUI_getOriginatingAddress_userInteractionSessionID = -1;
static int hf_org_csapi_ui_IpUICall_recordMessageReq_return = -1;
static int hf_org_csapi_ui_IpUICall_recordMessageReq_userInteractionSessionID = -1;
static int hf_org_csapi_ui_IpUICall_deleteMessageReq_return = -1;
static int hf_org_csapi_ui_IpUICall_deleteMessageReq_usrInteractionSessionID = -1;
static int hf_org_csapi_ui_IpUICall_deleteMessageReq_messageID = -1;
static int hf_org_csapi_ui_IpUICall_abortActionReq_userInteractionSessionID = -1;
static int hf_org_csapi_ui_IpUICall_abortActionReq_assignmentID = -1;
static int hf_org_csapi_ui_IpUICall_getMessageReq_return = -1;
static int hf_org_csapi_ui_IpUICall_getMessageReq_userInteractionSessionID = -1;
static int hf_org_csapi_ui_IpUICall_getMessageReq_messageID = -1;
static int hf_org_csapi_ui_IpUIManager_createNotification_return = -1;
static int hf_org_csapi_ui_IpUIManager_destroyNotification_assignmentID = -1;
static int hf_org_csapi_ui_IpUIManager_changeNotification_assignmentID = -1;
static int hf_org_csapi_ui_IpUIManager_getNotification_return_loop = -1;
static int hf_org_csapi_ui_IpUIManager_enableNotifications_return = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_getMessageRes_usrInteractionSessionID = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_getMessageRes_assignmentID = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_getMessageErr_usrInteractionSessionID = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_getMessageErr_assignmentID = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_getMessageErr_error = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_deleteMessageRes_usrInteractionSessionID = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_deleteMessageRes_response = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_deleteMessageRes_assignmentID = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_deleteMessageErr_usrInteractionSessionID = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_deleteMessageErr_error = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_deleteMessageErr_assignmentID = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_putMessageRes_usrInteractionSessionID = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_putMessageRes_assignmentID = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_putMessageRes_messageID = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_putMessageErr_usrInteractionSessionID = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_putMessageErr_assignmentID = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_putMessageErr_error = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_getMessageListRes_usrInteractionSessionID = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_getMessageListRes_assignmentID = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_getMessageListRes_messageIDList_loop = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_getMessageListRes_messageIDList = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_getMessageListRes_final = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_getMessageListErr_usrInteractionSessionID = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_getMessageListErr_assignmentID = -1;
static int hf_org_csapi_ui_IpAppUIAdminManager_getMessageListErr_error = -1;
static int hf_org_csapi_ui_IpUIAdminManager_getMessageReq_return = -1;
static int hf_org_csapi_ui_IpUIAdminManager_getMessageReq_usrInteractionSessionID = -1;
static int hf_org_csapi_ui_IpUIAdminManager_getMessageReq_messageID = -1;
static int hf_org_csapi_ui_IpUIAdminManager_putMessageReq_return = -1;
static int hf_org_csapi_ui_IpUIAdminManager_putMessageReq_usrInteractionSessionID = -1;
static int hf_org_csapi_ui_IpUIAdminManager_deleteMessageReq_return = -1;
static int hf_org_csapi_ui_IpUIAdminManager_deleteMessageReq_usrInteractionSessionID = -1;
static int hf_org_csapi_ui_IpUIAdminManager_deleteMessageReq_messageID = -1;
static int hf_org_csapi_ui_IpUIAdminManager_getMessageListReq_return = -1;
static int hf_org_csapi_ui_IpUIAdminManager_getMessageListReq_usrInteractionSessionID = -1;
static int hf_org_csapi_ui_IpUIAdminManager_getMessageListReq_reset = -1;
static int hf_org_csapi_mm_ul_IpAppUserLocation_locationReportRes_assignmentId = -1;
static int hf_org_csapi_mm_ul_IpAppUserLocation_locationReportRes_locations_loop = -1;
static int hf_org_csapi_mm_ul_IpAppUserLocation_locationReportErr_assignmentId = -1;
static int hf_org_csapi_mm_ul_IpAppUserLocation_locationReportErr_cause = -1;
static int hf_org_csapi_mm_ul_IpAppUserLocation_locationReportErr_diagnostic = -1;
static int hf_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportRes_assignmentId = -1;
static int hf_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportRes_locations_loop = -1;
static int hf_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportErr_assignmentId = -1;
static int hf_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportErr_cause = -1;
static int hf_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportErr_diagnostic = -1;
static int hf_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReport_assignmentId = -1;
static int hf_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReport_locations_loop = -1;
static int hf_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReportErr_assignmentId = -1;
static int hf_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReportErr_cause = -1;
static int hf_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReportErr_diagnostic = -1;
static int hf_org_csapi_mm_ul_IpUserLocation_locationReportReq_return = -1;
static int hf_org_csapi_mm_ul_IpUserLocation_locationReportReq_users_loop = -1;
static int hf_org_csapi_mm_ul_IpUserLocation_extendedLocationReportReq_return = -1;
static int hf_org_csapi_mm_ul_IpUserLocation_extendedLocationReportReq_users_loop = -1;
static int hf_org_csapi_mm_ul_IpUserLocation_periodicLocationReportingStartReq_return = -1;
static int hf_org_csapi_mm_ul_IpUserLocation_periodicLocationReportingStartReq_users_loop = -1;
static int hf_org_csapi_mm_ul_IpUserLocation_periodicLocationReportingStartReq_reportingInterval = -1;
static int hf_org_csapi_mm_ul_IpUserLocation_getNextPeriodicLocationRequest_reset = -1;
static int hf_org_csapi_mm_ul_IpAppTriggeredUserLocation_triggeredLocationReport_assignmentId = -1;
static int hf_org_csapi_mm_ul_IpAppTriggeredUserLocation_triggeredLocationReport_criterion = -1;
static int hf_org_csapi_mm_ul_IpAppTriggeredUserLocation_triggeredLocationReportErr_assignmentId = -1;
static int hf_org_csapi_mm_ul_IpAppTriggeredUserLocation_triggeredLocationReportErr_cause = -1;
static int hf_org_csapi_mm_ul_IpAppTriggeredUserLocation_triggeredLocationReportErr_diagnostic = -1;
static int hf_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_return = -1;
static int hf_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_users_loop = -1;
static int hf_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_triggers_loop = -1;
static int hf_org_csapi_mm_ul_IpTriggeredUserLocation_getNextTriggeredLocationRequest_reset = -1;
static int hf_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportRes_assignmentId = -1;
static int hf_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportRes_locations_loop = -1;
static int hf_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportErr_assignmentId = -1;
static int hf_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportErr_cause = -1;
static int hf_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportErr_diagnostic = -1;
static int hf_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReport_assignmentId = -1;
static int hf_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReport_locations_loop = -1;
static int hf_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReportErr_assignmentId = -1;
static int hf_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReportErr_cause = -1;
static int hf_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReportErr_diagnostic = -1;
static int hf_org_csapi_mm_ulc_IpAppUserLocationCamel_triggeredLocationReport_assignmentId = -1;
static int hf_org_csapi_mm_ulc_IpAppUserLocationCamel_triggeredLocationReportErr_assignmentId = -1;
static int hf_org_csapi_mm_ulc_IpAppUserLocationCamel_triggeredLocationReportErr_cause = -1;
static int hf_org_csapi_mm_ulc_IpAppUserLocationCamel_triggeredLocationReportErr_diagnostic = -1;
static int hf_org_csapi_mm_ulc_IpUserLocationCamel_locationReportReq_return = -1;
static int hf_org_csapi_mm_ulc_IpUserLocationCamel_locationReportReq_users_loop = -1;
static int hf_org_csapi_mm_ulc_IpUserLocationCamel_periodicLocationReportingStartReq_return = -1;
static int hf_org_csapi_mm_ulc_IpUserLocationCamel_periodicLocationReportingStartReq_users_loop = -1;
static int hf_org_csapi_mm_ulc_IpUserLocationCamel_periodicLocationReportingStartReq_reportingInterval = -1;
static int hf_org_csapi_mm_ulc_IpUserLocationCamel_triggeredLocationReportingStartReq_return = -1;
static int hf_org_csapi_mm_ulc_IpUserLocationCamel_triggeredLocationReportingStartReq_users_loop = -1;
static int hf_org_csapi_mm_ulc_IpUserLocationCamel_getNextPeriodicLocationRequest_reset = -1;
static int hf_org_csapi_mm_ulc_IpUserLocationCamel_getNextTriggeredLocationRequest_reset = -1;
static int hf_org_csapi_mm_ule_IpAppUserLocationEmergency_emergencyLocationReport_assignmentId = -1;
static int hf_org_csapi_mm_ule_IpAppUserLocationEmergency_emergencyLocationReportErr_assignmentId = -1;
static int hf_org_csapi_mm_ule_IpAppUserLocationEmergency_emergencyLocationReportErr_cause = -1;
static int hf_org_csapi_mm_ule_IpAppUserLocationEmergency_emergencyLocationReportErr_diagnostic = -1;
static int hf_org_csapi_mm_ule_IpUserLocationEmergency_emergencyLocationReportReq_return = -1;
static int hf_org_csapi_mm_ule_IpUserLocationEmergency_subscribeEmergencyLocationReports_return = -1;
static int hf_org_csapi_mm_ule_IpUserLocationEmergency_unSubscribeEmergencyLocationReports_assignmentId = -1;
static int hf_org_csapi_mm_us_IpAppUserStatus_statusReportRes_assignmentId = -1;
static int hf_org_csapi_mm_us_IpAppUserStatus_statusReportRes_status_loop = -1;
static int hf_org_csapi_mm_us_IpAppUserStatus_statusReportErr_assignmentId = -1;
static int hf_org_csapi_mm_us_IpAppUserStatus_statusReportErr_cause = -1;
static int hf_org_csapi_mm_us_IpAppUserStatus_statusReportErr_diagnostic = -1;
static int hf_org_csapi_mm_us_IpAppUserStatus_triggeredStatusReport_assignmentId = -1;
static int hf_org_csapi_mm_us_IpAppUserStatus_triggeredStatusReportErr_assignmentId = -1;
static int hf_org_csapi_mm_us_IpAppUserStatus_triggeredStatusReportErr_cause = -1;
static int hf_org_csapi_mm_us_IpAppUserStatus_triggeredStatusReportErr_diagnostic = -1;
static int hf_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportRes_assignmentId = -1;
static int hf_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportRes_status_loop = -1;
static int hf_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportErr_assignmentId = -1;
static int hf_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportErr_cause = -1;
static int hf_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportErr_diagnostic = -1;
static int hf_org_csapi_mm_us_IpAppUserStatus_extTriggeredStatusReport_assignmentId = -1;
static int hf_org_csapi_mm_us_IpAppUserStatus_extTriggeredStatusReportErr_assignmentId = -1;
static int hf_org_csapi_mm_us_IpAppUserStatus_extTriggeredStatusReportErr_cause = -1;
static int hf_org_csapi_mm_us_IpAppUserStatus_extTriggeredStatusReportErr_diagnostic = -1;
static int hf_org_csapi_mm_us_IpUserStatus_statusReportReq_return = -1;
static int hf_org_csapi_mm_us_IpUserStatus_statusReportReq_users_loop = -1;
static int hf_org_csapi_mm_us_IpUserStatus_triggeredStatusReportingStartReq_return = -1;
static int hf_org_csapi_mm_us_IpUserStatus_triggeredStatusReportingStartReq_users_loop = -1;
static int hf_org_csapi_mm_us_IpUserStatus_getNextTriggeredStatusRequest_reset = -1;
static int hf_org_csapi_mm_us_IpUserStatus_extendedStatusReportReq_return = -1;
static int hf_org_csapi_mm_us_IpUserStatus_extendedStatusReportReq_users_loop = -1;
static int hf_org_csapi_mm_us_IpUserStatus_extTriggeredStatusReportingStartReq_return = -1;
static int hf_org_csapi_mm_us_IpUserStatus_extTriggeredStatusReportingStartReq_users_loop = -1;
static int hf_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_return_loop = -1;
static int hf_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_assignmentId = -1;
static int hf_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_contactInformation_loop = -1;
static int hf_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotificationStartErr_assignmentId = -1;
static int hf_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotificationStartErr_cause = -1;
static int hf_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotificationStartErr_diagnostic = -1;
static int hf_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_return = -1;
static int hf_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_users_loop = -1;
static int hf_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_criteria_loop = -1;
static int hf_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStop_assignmentId = -1;
static int hf_org_csapi_termcap_IpTerminalCapabilities_getTerminalCapabilities_terminalIdentity = -1;
static int hf_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReport_assignmentID = -1;
static int hf_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReport_terminals_loop = -1;
static int hf_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReport_criteria = -1;
static int hf_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReportErr_assignmentId = -1;
static int hf_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReportErr_terminals_loop = -1;
static int hf_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReportErr_cause = -1;
static int hf_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStartReq_return = -1;
static int hf_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStartReq_terminals_loop = -1;
static int hf_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStartReq_criteria = -1;
static int hf_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStop_assignmentID = -1;
static int hf_org_csapi_dsc_IpAppDataSession_connectRes_dataSessionID = -1;
static int hf_org_csapi_dsc_IpAppDataSession_connectRes_assignmentID = -1;
static int hf_org_csapi_dsc_IpAppDataSession_connectErr_dataSessionID = -1;
static int hf_org_csapi_dsc_IpAppDataSession_connectErr_assignmentID = -1;
static int hf_org_csapi_dsc_IpAppDataSession_superviseDataSessionRes_dataSessionID = -1;
static int hf_org_csapi_dsc_IpAppDataSession_superviseDataSessionRes_report = -1;
static int hf_org_csapi_dsc_IpAppDataSession_superviseDataSessionRes_qualityOfService = -1;
static int hf_org_csapi_dsc_IpAppDataSession_superviseDataSessionErr_dataSessionID = -1;
static int hf_org_csapi_dsc_IpAppDataSession_dataSessionFaultDetected_dataSessionID = -1;
static int hf_org_csapi_dsc_IpAppDataSession_dataSessionFaultDetected_fault = -1;
static int hf_org_csapi_dsc_IpAppDataSessionControlManager_dataSessionAborted_dataSession = -1;
static int hf_org_csapi_dsc_IpAppDataSessionControlManager_reportNotification_assignmentID = -1;
static int hf_org_csapi_dsc_IpAppDataSessionControlManager_abortMultipleDataSessions_dataSessionSet_loop = -1;
static int hf_org_csapi_dsc_IpAppDataSessionControlManager_abortMultipleDataSessions_dataSessionSet = -1;
static int hf_org_csapi_dsc_IpDataSession_connectReq_return = -1;
static int hf_org_csapi_dsc_IpDataSession_connectReq_dataSessionID = -1;
static int hf_org_csapi_dsc_IpDataSession_connectReq_responseRequested_loop = -1;
static int hf_org_csapi_dsc_IpDataSession_release_dataSessionID = -1;
static int hf_org_csapi_dsc_IpDataSession_superviseDataSessionReq_dataSessionID = -1;
static int hf_org_csapi_dsc_IpDataSession_superviseDataSessionReq_treatment = -1;
static int hf_org_csapi_dsc_IpDataSession_setDataSessionChargePlan_dataSessionID = -1;
static int hf_org_csapi_dsc_IpDataSession_setAdviceOfCharge_dataSessionID = -1;
static int hf_org_csapi_dsc_IpDataSession_setAdviceOfCharge_tariffSwitch = -1;
static int hf_org_csapi_dsc_IpDataSession_deassignDataSession_dataSessionID = -1;
static int hf_org_csapi_dsc_IpDataSession_continueProcessing_dataSessionID = -1;
static int hf_org_csapi_dsc_IpDataSessionControlManager_createNotification_return = -1;
static int hf_org_csapi_dsc_IpDataSessionControlManager_destroyNotification_assignmentID = -1;
static int hf_org_csapi_dsc_IpDataSessionControlManager_changeNotification_assignmentID = -1;
static int hf_org_csapi_dsc_IpDataSessionControlManager_enableNotifications_return = -1;
static int hf_org_csapi_dsc_IpDataSessionControlManager_getNotifications_return_loop = -1;
static int hf_org_csapi_dsc_IpDataSessionControlManager_createNotifications_return = -1;
static int hf_org_csapi_gms_IpMessagingManager_openMailbox_authenticationInfo = -1;
static int hf_org_csapi_gms_IpMessagingManager_enableMessagingNotification_return = -1;
static int hf_org_csapi_gms_IpMessagingManager_disableMessagingNotification_assignmentID = -1;
static int hf_org_csapi_gms_IpAppMessagingManager_mailboxTerminated_mailboxSessionID = -1;
static int hf_org_csapi_gms_IpAppMessagingManager_mailboxFaultDetected_mailboxSessionID = -1;
static int hf_org_csapi_gms_IpAppMessagingManager_mailboxFaultDetected_fault = -1;
static int hf_org_csapi_gms_IpAppMessagingManager_messagingEventNotify_assignmentID = -1;
static int hf_org_csapi_gms_IpMailbox_close_mailboxSessionID = -1;
static int hf_org_csapi_gms_IpMailbox_lock_mailboxSessionID = -1;
static int hf_org_csapi_gms_IpMailbox_unlock_mailboxSessionID = -1;
static int hf_org_csapi_gms_IpMailbox_getInfoAmount_return = -1;
static int hf_org_csapi_gms_IpMailbox_getInfoAmount_mailboxSessionID = -1;
static int hf_org_csapi_gms_IpMailbox_getInfoProperties_return_loop = -1;
static int hf_org_csapi_gms_IpMailbox_getInfoProperties_mailboxSessionID = -1;
static int hf_org_csapi_gms_IpMailbox_getInfoProperties_firstProperty = -1;
static int hf_org_csapi_gms_IpMailbox_getInfoProperties_numberOfProperties = -1;
static int hf_org_csapi_gms_IpMailbox_setInfoProperties_mailboxSessionID = -1;
static int hf_org_csapi_gms_IpMailbox_setInfoProperties_firstProperty = -1;
static int hf_org_csapi_gms_IpMailbox_setInfoProperties_mailboxInfoProperties_loop = -1;
static int hf_org_csapi_gms_IpMailbox_openFolder_mailboxSessionID = -1;
static int hf_org_csapi_gms_IpMailbox_openFolder_folderID = -1;
static int hf_org_csapi_gms_IpMailbox_createFolder_mailboxSessionID = -1;
static int hf_org_csapi_gms_IpMailbox_createFolder_folderID = -1;
static int hf_org_csapi_gms_IpMailbox_remove_authenticationInfo = -1;
static int hf_org_csapi_gms_IpMessage_getInfoAmount_return = -1;
static int hf_org_csapi_gms_IpMessage_getInfoAmount_folderSessionID = -1;
static int hf_org_csapi_gms_IpMessage_getInfoAmount_messageID = -1;
static int hf_org_csapi_gms_IpMessage_getInfoProperties_return_loop = -1;
static int hf_org_csapi_gms_IpMessage_getInfoProperties_folderSessionID = -1;
static int hf_org_csapi_gms_IpMessage_getInfoProperties_messageID = -1;
static int hf_org_csapi_gms_IpMessage_getInfoProperties_firstProperty = -1;
static int hf_org_csapi_gms_IpMessage_getInfoProperties_numberOfProperties = -1;
static int hf_org_csapi_gms_IpMessage_setInfoProperties_folderSessionID = -1;
static int hf_org_csapi_gms_IpMessage_setInfoProperties_messageID = -1;
static int hf_org_csapi_gms_IpMessage_setInfoProperties_firstProperty = -1;
static int hf_org_csapi_gms_IpMessage_setInfoProperties_messageInfoProperties_loop = -1;
static int hf_org_csapi_gms_IpMessage_remove_folderSessionID = -1;
static int hf_org_csapi_gms_IpMessage_remove_messageID = -1;
static int hf_org_csapi_gms_IpMessage_getContent_return = -1;
static int hf_org_csapi_gms_IpMessage_getContent_folderSessionID = -1;
static int hf_org_csapi_gms_IpMessage_getContent_messageID = -1;
static int hf_org_csapi_gms_IpMailboxFolder_getInfoAmount_return = -1;
static int hf_org_csapi_gms_IpMailboxFolder_getInfoAmount_folderSessionID = -1;
static int hf_org_csapi_gms_IpMailboxFolder_getInfoProperties_return_loop = -1;
static int hf_org_csapi_gms_IpMailboxFolder_getInfoProperties_folderSessionID = -1;
static int hf_org_csapi_gms_IpMailboxFolder_getInfoProperties_firstProperty = -1;
static int hf_org_csapi_gms_IpMailboxFolder_getInfoProperties_numberOfProperties = -1;
static int hf_org_csapi_gms_IpMailboxFolder_setInfoProperties_folderSessionID = -1;
static int hf_org_csapi_gms_IpMailboxFolder_setInfoProperties_firstProperty = -1;
static int hf_org_csapi_gms_IpMailboxFolder_setInfoProperties_folderInfoProperties_loop = -1;
static int hf_org_csapi_gms_IpMailboxFolder_putMessage_folderSessionID = -1;
static int hf_org_csapi_gms_IpMailboxFolder_putMessage_message = -1;
static int hf_org_csapi_gms_IpMailboxFolder_putMessage_messageInfoProperties_loop = -1;
static int hf_org_csapi_gms_IpMailboxFolder_getMessage_folderSessionID = -1;
static int hf_org_csapi_gms_IpMailboxFolder_getMessage_messageID = -1;
static int hf_org_csapi_gms_IpMailboxFolder_close_mailboxSessionID = -1;
static int hf_org_csapi_gms_IpMailboxFolder_close_folderSessionID = -1;
static int hf_org_csapi_gms_IpMailboxFolder_remove_mailboxSessionID = -1;
static int hf_org_csapi_gms_IpMailboxFolder_remove_folderID = -1;
static int hf_org_csapi_cm_IpVPrP_getVPrPID_return = -1;
static int hf_org_csapi_cm_IpVPrP_getSlaID_return = -1;
static int hf_org_csapi_cm_IpVPrP_getStatus_return = -1;
static int hf_org_csapi_cm_IpVPrN_getVPrPList_return_loop = -1;
static int hf_org_csapi_cm_IpVPrN_getVPrPList_return = -1;
static int hf_org_csapi_cm_IpVPrN_getVPrP_vPrPID = -1;
static int hf_org_csapi_cm_IpVPrN_deleteVPrP_vPrPID = -1;
static int hf_org_csapi_cm_IpQoSMenu_getTemplate_templateType = -1;
static int hf_org_csapi_cm_IpQoSMenu_getTemplateList_return_loop = -1;
static int hf_org_csapi_cm_IpQoSMenu_getTemplateList_return = -1;
static int hf_org_csapi_cm_IpQoSTemplate_getTemplateType_return = -1;
static int hf_org_csapi_cm_IpQoSTemplate_getDescription_return = -1;
static int hf_org_csapi_cm_IpQoSTemplate_setSlaID_slaID = -1;
static int hf_org_csapi_cm_IpEnterpriseNetwork_getSiteList_return_loop = -1;
static int hf_org_csapi_cm_IpEnterpriseNetwork_getSiteList_return = -1;
static int hf_org_csapi_cm_IpEnterpriseNetwork_getSite_siteID = -1;
static int hf_org_csapi_cm_IpEnterpriseNetworkSite_getSAPList_return_loop = -1;
static int hf_org_csapi_cm_IpEnterpriseNetworkSite_getSAPList_return = -1;
static int hf_org_csapi_cm_IpEnterpriseNetworkSite_getSiteID_return = -1;
static int hf_org_csapi_cm_IpEnterpriseNetworkSite_getSiteLocation_return = -1;
static int hf_org_csapi_cm_IpEnterpriseNetworkSite_getSiteDescription_return = -1;
static int hf_org_csapi_cm_IpEnterpriseNetworkSite_getSAPIPSubnet_sapID = -1;
static int hf_org_csapi_am_IpAppAccountManager_reportNotification_assignmentId = -1;
static int hf_org_csapi_am_IpAppAccountManager_queryBalanceRes_queryId = -1;
static int hf_org_csapi_am_IpAppAccountManager_queryBalanceRes_balances_loop = -1;
static int hf_org_csapi_am_IpAppAccountManager_queryBalanceErr_queryId = -1;
static int hf_org_csapi_am_IpAppAccountManager_queryBalanceErr_cause = -1;
static int hf_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryRes_retrievalID = -1;
static int hf_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryRes_transactionHistory_loop = -1;
static int hf_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryErr_retrievalID = -1;
static int hf_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryErr_transactionHistoryError = -1;
static int hf_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateRes_queryId = -1;
static int hf_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateRes_dates_loop = -1;
static int hf_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateErr_queryId = -1;
static int hf_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateErr_cause = -1;
static int hf_org_csapi_am_IpAppAccountManager_updateBalanceRes_requestId = -1;
static int hf_org_csapi_am_IpAppAccountManager_updateBalanceErr_requestId = -1;
static int hf_org_csapi_am_IpAppAccountManager_updateBalanceErr_cause = -1;
static int hf_org_csapi_am_IpAppAccountManager_createVoucherRes_requestId = -1;
static int hf_org_csapi_am_IpAppAccountManager_createVoucherRes_voucherId = -1;
static int hf_org_csapi_am_IpAppAccountManager_createVoucherErr_requestId = -1;
static int hf_org_csapi_am_IpAppAccountManager_createVoucherErr_cause = -1;
static int hf_org_csapi_am_IpAppAccountManager_destroyVoucherRes_requestId = -1;
static int hf_org_csapi_am_IpAppAccountManager_destroyVoucherRes_voucherId = -1;
static int hf_org_csapi_am_IpAppAccountManager_destroyVoucherErr_requestId = -1;
static int hf_org_csapi_am_IpAppAccountManager_destroyVoucherErr_voucherId = -1;
static int hf_org_csapi_am_IpAppAccountManager_destroyVoucherErr_cause = -1;
static int hf_org_csapi_am_IpAppAccountManager_queryVoucherRes_queryId = -1;
static int hf_org_csapi_am_IpAppAccountManager_queryVoucherErr_queryId = -1;
static int hf_org_csapi_am_IpAppAccountManager_queryVoucherErr_voucherId = -1;
static int hf_org_csapi_am_IpAppAccountManager_queryVoucherErr_cause = -1;
static int hf_org_csapi_am_IpAppAccountManager_queryUserVouchersRes_queryId = -1;
static int hf_org_csapi_am_IpAppAccountManager_queryUserVouchersRes_vouchers_loop = -1;
static int hf_org_csapi_am_IpAppAccountManager_queryUserVouchersErr_queryId = -1;
static int hf_org_csapi_am_IpAppAccountManager_queryUserVouchersErr_cause = -1;
static int hf_org_csapi_am_IpAccountManager_createNotification_return = -1;
static int hf_org_csapi_am_IpAccountManager_destroyNotification_assignmentId = -1;
static int hf_org_csapi_am_IpAccountManager_queryBalanceReq_return = -1;
static int hf_org_csapi_am_IpAccountManager_queryBalanceReq_users_loop = -1;
static int hf_org_csapi_am_IpAccountManager_changeNotification_assignmentID = -1;
static int hf_org_csapi_am_IpAccountManager_getNotification_return_loop = -1;
static int hf_org_csapi_am_IpAccountManager_retrieveTransactionHistoryReq_return = -1;
static int hf_org_csapi_am_IpAccountManager_enableNotifications_return = -1;
static int hf_org_csapi_am_IpAccountManager_queryBalanceExpiryDateReq_return = -1;
static int hf_org_csapi_am_IpAccountManager_queryBalanceExpiryDateReq_users_loop = -1;
static int hf_org_csapi_am_IpAccountManager_updateBalanceReq_return = -1;
static int hf_org_csapi_am_IpAccountManager_updateBalanceReq_debit = -1;
static int hf_org_csapi_am_IpAccountManager_updateBalanceReq_period = -1;
static int hf_org_csapi_am_IpAccountManager_createVoucherReq_return = -1;
static int hf_org_csapi_am_IpAccountManager_destroyVoucherReq_return = -1;
static int hf_org_csapi_am_IpAccountManager_destroyVoucherReq_voucherId = -1;
static int hf_org_csapi_am_IpAccountManager_queryVoucherReq_return = -1;
static int hf_org_csapi_am_IpAccountManager_queryVoucherReq_voucherId = -1;
static int hf_org_csapi_am_IpAccountManager_queryUserVouchersReq_return = -1;
static int hf_org_csapi_cs_IpAppChargingSession_creditAmountErr_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_creditAmountErr_requestNumber = -1;
static int hf_org_csapi_cs_IpAppChargingSession_creditAmountErr_error = -1;
static int hf_org_csapi_cs_IpAppChargingSession_creditAmountErr_requestNumberNextRequest = -1;
static int hf_org_csapi_cs_IpAppChargingSession_creditAmountRes_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_creditAmountRes_requestNumber = -1;
static int hf_org_csapi_cs_IpAppChargingSession_creditAmountRes_requestNumberNextRequest = -1;
static int hf_org_csapi_cs_IpAppChargingSession_creditUnitErr_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_creditUnitErr_requestNumber = -1;
static int hf_org_csapi_cs_IpAppChargingSession_creditUnitErr_error = -1;
static int hf_org_csapi_cs_IpAppChargingSession_creditUnitErr_requestNumberNextRequest = -1;
static int hf_org_csapi_cs_IpAppChargingSession_creditUnitRes_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_creditUnitRes_requestNumber = -1;
static int hf_org_csapi_cs_IpAppChargingSession_creditUnitRes_creditedVolumes_loop = -1;
static int hf_org_csapi_cs_IpAppChargingSession_creditUnitRes_reservedUnitsLeft_loop = -1;
static int hf_org_csapi_cs_IpAppChargingSession_creditUnitRes_requestNumberNextRequest = -1;
static int hf_org_csapi_cs_IpAppChargingSession_debitAmountErr_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_debitAmountErr_requestNumber = -1;
static int hf_org_csapi_cs_IpAppChargingSession_debitAmountErr_error = -1;
static int hf_org_csapi_cs_IpAppChargingSession_debitAmountErr_requestNumberNextRequest = -1;
static int hf_org_csapi_cs_IpAppChargingSession_debitAmountRes_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_debitAmountRes_requestNumber = -1;
static int hf_org_csapi_cs_IpAppChargingSession_debitAmountRes_requestNumberNextRequest = -1;
static int hf_org_csapi_cs_IpAppChargingSession_debitUnitErr_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_debitUnitErr_requestNumber = -1;
static int hf_org_csapi_cs_IpAppChargingSession_debitUnitErr_error = -1;
static int hf_org_csapi_cs_IpAppChargingSession_debitUnitErr_requestNumberNextRequest = -1;
static int hf_org_csapi_cs_IpAppChargingSession_debitUnitRes_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_debitUnitRes_requestNumber = -1;
static int hf_org_csapi_cs_IpAppChargingSession_debitUnitRes_debitedVolumes_loop = -1;
static int hf_org_csapi_cs_IpAppChargingSession_debitUnitRes_reservedUnitsLeft_loop = -1;
static int hf_org_csapi_cs_IpAppChargingSession_debitUnitRes_requestNumberNextRequest = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directCreditAmountErr_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directCreditAmountErr_requestNumber = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directCreditAmountErr_error = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directCreditAmountErr_requestNumberNextRequest = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directCreditAmountRes_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directCreditAmountRes_requestNumber = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directCreditAmountRes_requestNumberNextRequest = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directCreditUnitErr_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directCreditUnitErr_requestNumber = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directCreditUnitErr_error = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directCreditUnitErr_requestNumberNextRequest = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directCreditUnitRes_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directCreditUnitRes_requestNumber = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directCreditUnitRes_creditedVolumes_loop = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directCreditUnitRes_requestNumberNextRequest = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directDebitAmountErr_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directDebitAmountErr_requestNumber = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directDebitAmountErr_error = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directDebitAmountErr_requestNumberNextRequest = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directDebitAmountRes_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directDebitAmountRes_requestNumber = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directDebitAmountRes_requestNumberNextRequest = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directDebitUnitErr_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directDebitUnitErr_requestNumber = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directDebitUnitErr_error = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directDebitUnitErr_requestNumberNextRequest = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directDebitUnitRes_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directDebitUnitRes_requestNumber = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directDebitUnitRes_debitedVolumes_loop = -1;
static int hf_org_csapi_cs_IpAppChargingSession_directDebitUnitRes_requestNumberNextRequest = -1;
static int hf_org_csapi_cs_IpAppChargingSession_extendLifeTimeErr_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_extendLifeTimeErr_error = -1;
static int hf_org_csapi_cs_IpAppChargingSession_extendLifeTimeRes_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_extendLifeTimeRes_sessionTimeLeft = -1;
static int hf_org_csapi_cs_IpAppChargingSession_rateErr_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_rateErr_error = -1;
static int hf_org_csapi_cs_IpAppChargingSession_rateRes_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_rateRes_rates_loop = -1;
static int hf_org_csapi_cs_IpAppChargingSession_rateRes_validityTimeLeft = -1;
static int hf_org_csapi_cs_IpAppChargingSession_reserveAmountErr_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_reserveAmountErr_requestNumber = -1;
static int hf_org_csapi_cs_IpAppChargingSession_reserveAmountErr_error = -1;
static int hf_org_csapi_cs_IpAppChargingSession_reserveAmountErr_requestNumberNextRequest = -1;
static int hf_org_csapi_cs_IpAppChargingSession_reserveAmountRes_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_reserveAmountRes_requestNumber = -1;
static int hf_org_csapi_cs_IpAppChargingSession_reserveAmountRes_sessionTimeLeft = -1;
static int hf_org_csapi_cs_IpAppChargingSession_reserveAmountRes_requestNumberNextRequest = -1;
static int hf_org_csapi_cs_IpAppChargingSession_reserveUnitErr_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_reserveUnitErr_requestNumber = -1;
static int hf_org_csapi_cs_IpAppChargingSession_reserveUnitErr_error = -1;
static int hf_org_csapi_cs_IpAppChargingSession_reserveUnitErr_requestNumberNextRequest = -1;
static int hf_org_csapi_cs_IpAppChargingSession_reserveUnitRes_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_reserveUnitRes_requestNumber = -1;
static int hf_org_csapi_cs_IpAppChargingSession_reserveUnitRes_reservedUnits_loop = -1;
static int hf_org_csapi_cs_IpAppChargingSession_reserveUnitRes_sessionTimeLeft = -1;
static int hf_org_csapi_cs_IpAppChargingSession_reserveUnitRes_requestNumberNextRequest = -1;
static int hf_org_csapi_cs_IpAppChargingSession_sessionEnded_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingSession_sessionEnded_report = -1;
static int hf_org_csapi_cs_IpChargingSession_creditAmountReq_sessionID = -1;
static int hf_org_csapi_cs_IpChargingSession_creditAmountReq_closeReservation = -1;
static int hf_org_csapi_cs_IpChargingSession_creditAmountReq_requestNumber = -1;
static int hf_org_csapi_cs_IpChargingSession_creditUnitReq_sessionID = -1;
static int hf_org_csapi_cs_IpChargingSession_creditUnitReq_volumes_loop = -1;
static int hf_org_csapi_cs_IpChargingSession_creditUnitReq_closeReservation = -1;
static int hf_org_csapi_cs_IpChargingSession_creditUnitReq_requestNumber = -1;
static int hf_org_csapi_cs_IpChargingSession_debitAmountReq_sessionID = -1;
static int hf_org_csapi_cs_IpChargingSession_debitAmountReq_closeReservation = -1;
static int hf_org_csapi_cs_IpChargingSession_debitAmountReq_requestNumber = -1;
static int hf_org_csapi_cs_IpChargingSession_debitUnitReq_sessionID = -1;
static int hf_org_csapi_cs_IpChargingSession_debitUnitReq_volumes_loop = -1;
static int hf_org_csapi_cs_IpChargingSession_debitUnitReq_closeReservation = -1;
static int hf_org_csapi_cs_IpChargingSession_debitUnitReq_requestNumber = -1;
static int hf_org_csapi_cs_IpChargingSession_directCreditAmountReq_sessionID = -1;
static int hf_org_csapi_cs_IpChargingSession_directCreditAmountReq_chargingParameters_loop = -1;
static int hf_org_csapi_cs_IpChargingSession_directCreditAmountReq_requestNumber = -1;
static int hf_org_csapi_cs_IpChargingSession_directCreditUnitReq_sessionID = -1;
static int hf_org_csapi_cs_IpChargingSession_directCreditUnitReq_chargingParameters_loop = -1;
static int hf_org_csapi_cs_IpChargingSession_directCreditUnitReq_volumes_loop = -1;
static int hf_org_csapi_cs_IpChargingSession_directCreditUnitReq_requestNumber = -1;
static int hf_org_csapi_cs_IpChargingSession_directDebitAmountReq_sessionID = -1;
static int hf_org_csapi_cs_IpChargingSession_directDebitAmountReq_chargingParameters_loop = -1;
static int hf_org_csapi_cs_IpChargingSession_directDebitAmountReq_requestNumber = -1;
static int hf_org_csapi_cs_IpChargingSession_directDebitUnitReq_sessionID = -1;
static int hf_org_csapi_cs_IpChargingSession_directDebitUnitReq_chargingParameters_loop = -1;
static int hf_org_csapi_cs_IpChargingSession_directDebitUnitReq_volumes_loop = -1;
static int hf_org_csapi_cs_IpChargingSession_directDebitUnitReq_requestNumber = -1;
static int hf_org_csapi_cs_IpChargingSession_extendLifeTimeReq_sessionID = -1;
static int hf_org_csapi_cs_IpChargingSession_getAmountLeft_sessionID = -1;
static int hf_org_csapi_cs_IpChargingSession_getLifeTimeLeft_return = -1;
static int hf_org_csapi_cs_IpChargingSession_getLifeTimeLeft_sessionID = -1;
static int hf_org_csapi_cs_IpChargingSession_getUnitLeft_return_loop = -1;
static int hf_org_csapi_cs_IpChargingSession_getUnitLeft_sessionID = -1;
static int hf_org_csapi_cs_IpChargingSession_rateReq_sessionID = -1;
static int hf_org_csapi_cs_IpChargingSession_rateReq_chargingParameters_loop = -1;
static int hf_org_csapi_cs_IpChargingSession_release_sessionID = -1;
static int hf_org_csapi_cs_IpChargingSession_release_requestNumber = -1;
static int hf_org_csapi_cs_IpChargingSession_reserveAmountReq_sessionID = -1;
static int hf_org_csapi_cs_IpChargingSession_reserveAmountReq_chargingParameters_loop = -1;
static int hf_org_csapi_cs_IpChargingSession_reserveAmountReq_requestNumber = -1;
static int hf_org_csapi_cs_IpChargingSession_reserveUnitReq_sessionID = -1;
static int hf_org_csapi_cs_IpChargingSession_reserveUnitReq_chargingParameters_loop = -1;
static int hf_org_csapi_cs_IpChargingSession_reserveUnitReq_volumes_loop = -1;
static int hf_org_csapi_cs_IpChargingSession_reserveUnitReq_requestNumber = -1;
static int hf_org_csapi_cs_IpAppChargingManager_sessionAborted_sessionID = -1;
static int hf_org_csapi_cs_IpAppChargingManager_abortMultipleChargingSessions_chargingSessionSet_loop = -1;
static int hf_org_csapi_cs_IpAppChargingManager_abortMultipleChargingSessions_chargingSessionSet = -1;
static int hf_org_csapi_cs_IpChargingManager_createChargingSession_sessionDescription = -1;
static int hf_org_csapi_cs_IpChargingManager_createSplitChargingSession_sessionDescription = -1;
static int hf_org_csapi_cs_IpChargingManager_createSplitChargingSession_users_loop = -1;
static int hf_org_csapi_policy_evaluation_IpAppPolicyDomain_reportNotification_assignmentID = -1;
static int hf_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyRes_assignmentID = -1;
static int hf_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyRes_outputVariables_loop = -1;
static int hf_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyErr_assignmentID = -1;
static int hf_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyErr_error = -1;
static int hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_return_loop = -1;
static int hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_domainName = -1;
static int hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_signatureName = -1;
static int hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_inputVariables_loop = -1;
static int hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicyReq_return = -1;
static int hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicyReq_domainName = -1;
static int hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicyReq_signatureName = -1;
static int hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicyReq_inputVariables_loop = -1;
static int hf_org_csapi_policy_evaluation_IpPolicyEvalManager_abortEvalPolicyReq_domainName = -1;
static int hf_org_csapi_policy_evaluation_IpPolicyEvalManager_abortEvalPolicyReq_assignmentID = -1;
static int hf_org_csapi_policy_evaluation_IpPolicyEvalManager_generateEvent_domainName = -1;
static int hf_org_csapi_policy_evaluation_IpPolicyEvalManager_generateEvent_eventDefinitionName = -1;
static int hf_org_csapi_policy_evaluation_IpPolicyEvalManager_generateEvent_attributes_loop = -1;
static int hf_org_csapi_policy_evaluation_IpPolicyEvalManager_createNotification_return = -1;
static int hf_org_csapi_policy_evaluation_IpPolicyEvalManager_createNotification_domainName = -1;
static int hf_org_csapi_policy_evaluation_IpPolicyEvalManager_createNotification_events_loop = -1;
static int hf_org_csapi_policy_evaluation_IpPolicyEvalManager_createNotification_events = -1;
static int hf_org_csapi_policy_evaluation_IpPolicyEvalManager_destroyNotification_assignmentID = -1;
static int hf_org_csapi_policy_evaluation_IpPolicyEvalManager_destroyNotification_events_loop = -1;
static int hf_org_csapi_policy_evaluation_IpPolicyEvalManager_destroyNotification_events = -1;
static int hf_org_csapi_policy_provisioning_IpPolicy_getAttribute_attributeName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicy_getAttributes_return_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicy_getAttributes_attributeNames_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicy_getAttributes_attributeNames = -1;
static int hf_org_csapi_policy_provisioning_IpPolicy_setAttributes_targetAttributes_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyIterator_getList_return_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyIterator_getList_return = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyIterator_getList_startIndex = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyIterator_getList_numberRequested = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRepository_createRepository_repositoryName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRepository_getRepository_repositoryName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRepository_removeRepository_repositoryName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRepository_getRepositoryCount_return = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRepository_createCondition_conditionName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRepository_createCondition_conditionType = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRepository_createCondition_conditionAttributes_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRepository_getCondition_conditionName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRepository_removeCondition_conditionName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRepository_getConditionCount_return = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRepository_createAction_actionName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRepository_createAction_actionType = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRepository_createAction_actionAttributes_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRepository_getAction_actionName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRepository_removeAction_actionName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRepository_getActionCount_return = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRule_createCondition_conditionName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRule_createCondition_conditionType = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRule_createCondition_conditionAttributes_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRule_getCondition_conditionName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRule_removeCondition_conditionName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRule_getConditionCount_return = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRule_createAction_actionName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRule_createAction_actionType = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRule_createAction_actionAttributes_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRule_getAction_actionName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRule_removeAction_actionName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRule_getActionCount_return = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRule_setValidityPeriodConditionByName_conditionName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRule_setConditionList_conditionList_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRule_getConditionList_return_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRule_setActionList_actionList_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyRule_getActionList_return_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyManager_createDomain_domainName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyManager_getDomain_domainName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyManager_removeDomain_domainName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyManager_getDomainCount_return = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_return_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_return = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_matchingAttributes_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyManager_createRepository_repositoryName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyManager_getRepository_repositoryName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyManager_removeRepository_repositoryName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyManager_getRepositoryCount_return = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyManager_commitTransaction_return = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyGroup_createGroup_groupName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyGroup_getGroup_groupName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyGroup_removeGroup_groupName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyGroup_getGroupCount_return = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyGroup_createRule_ruleName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyGroup_getRule_ruleName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyGroup_removeRule_ruleName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyGroup_getRuleCount_return = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyEventDefinition_setRequiredAttributes_requiredAttributes_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyEventDefinition_setOptionalAttributes_optionalAttributes_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyEventDefinition_getRequiredAttributes_return_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyEventDefinition_getOptionalAttributes_return_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicySignature_setInputVariables_inputVariables_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicySignature_setInputVariables_inputVariables = -1;
static int hf_org_csapi_policy_provisioning_IpPolicySignature_setOutputVariables_outputVariables_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicySignature_setOutputVariables_outputVariables = -1;
static int hf_org_csapi_policy_provisioning_IpPolicySignature_getInputVariables_return_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicySignature_getInputVariables_return = -1;
static int hf_org_csapi_policy_provisioning_IpPolicySignature_getOutputVariables_return_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicySignature_getOutputVariables_return = -1;
static int hf_org_csapi_policy_provisioning_IpPolicySignature_setGroupNames_groupNames_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicySignature_setGroupNames_groupNames = -1;
static int hf_org_csapi_policy_provisioning_IpPolicySignature_setPolicyRoles_roleNames_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicySignature_setPolicyRoles_roleNames = -1;
static int hf_org_csapi_policy_provisioning_IpPolicySignature_getGroupNames_return_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicySignature_getGroupNames_return = -1;
static int hf_org_csapi_policy_provisioning_IpPolicySignature_getPolicyRoles_return_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicySignature_getPolicyRoles_return = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_createDomain_domainName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_getDomain_domainName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_removeDomain_domainName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_getDomainCount_return = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_createGroup_groupName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_getGroup_groupName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_removeGroup_groupName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_getGroupCount_return = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_createRule_ruleName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_getRule_ruleName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_removeRule_ruleName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_getRuleCount_return = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_eventDefinitionName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_requiredAttributes_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_requiredAttributes = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_optionalAttributes_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_optionalAttributes = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_getEventDefinition_eventDefinitionName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_removeEventDefinition_eventDefinitionName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_getEventDefinitionCount_return = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_createVariableSet_variableSetName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariableSet_return_loop = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariableSet_variableSetName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_removeVariableSet_variableSetName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariableSetCount_return = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_createVariable_variableSetName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_createVariable_variableName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_setVariableValue_variableSetName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_setVariableValue_variableName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariableType_variableSetName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariableType_variableName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariableValue_variableSetName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariableValue_variableName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariable_variableSetName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariable_variableName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_removeVariable_variablSetName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_removeVariable_variableName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_createSignature_signatureName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_getSignature_signatureName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_removeSignature_signatureName = -1;
static int hf_org_csapi_policy_provisioning_IpPolicyDomain_getSignatureCount_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_identityTypes_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_identityTypes = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteIdentity_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteIdentity_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteIdentity_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isIdentity_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isIdentity_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isIdentity_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isIdentity_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_identityTypes_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_identityTypes = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteGroupIdentity_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteGroupIdentity_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteGroupIdentity_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addToGroup_group = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addToGroup_member = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addToGroup_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addToGroup_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeFromGroup_group = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeFromGroup_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeFromGroup_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeFromGroup_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isGroupIdentity_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isGroupIdentity_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isGroupIdentity_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isGroupIdentity_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addAlias_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addAlias_alias = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addAlias_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addAlias_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeAliases_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeAliases_alias = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeAliases_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeAliases_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_lookupByAlias_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_lookupByAlias_alias = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_lookupByAlias_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_lookupByAlias_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_identityTypes_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_identityTypes = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_identityTypes_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_identityTypes = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType_typeName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_identityType = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_attributeNames_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_attributeNames = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_identityType = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_attributes_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_agentName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_agentTypes_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_agentTypes = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_deleteAgent_agentName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_deleteAgent_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_deleteAgent_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isAgent_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isAgent_agentName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isAgent_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isAgent_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_agentName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_capabilities_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_capabilities = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_agentName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_capabilities_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_capabilities = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_agentName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_agentName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf_agentName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf_capability = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_agentName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_agentTypes_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_agentTypes = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_agentName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_agentTypes_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_agentTypes = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_agentName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType_agentName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType_typeName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_agentName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_agentType = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_attributeNames_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_attributeNames = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_agentName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_agentType = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_attributes_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_assignAgent_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_assignAgent_agentName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_assignAgent_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_assignAgent_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_unassignAgent_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_unassignAgent_agentName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_unassignAgent_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_unassignAgent_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_agentName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_capability = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf_capability = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityAttribute_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityAttribute_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityAttribute_attributeName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityAttribute_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityAttribute_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_getIdentityAttributeDefinition_attributeName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_getIdentityAttributeDefinition_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_getIdentityAttributeDefinition_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_typeName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_attributeNames_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_attributeNames = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityType_typeName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityType_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityType_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_typeName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_attributeNames_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_attributeNames = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_typeName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_attributeNames_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_attributeNames = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_typeName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentAttribute_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentAttribute_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentAttribute_attributeName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentAttribute_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentAttribute_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_getAgentAttributeDefinition_attributeName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_getAgentAttributeDefinition_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_getAgentAttributeDefinition_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_typeName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_attributeNames_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_attributeNames = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentType_typeName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentType_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentType_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_typeName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_attributeNames_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_attributeNames = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_typeName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_attributeNames_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_attributeNames = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_typeName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapabilityAttribute_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapabilityAttribute_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapabilityAttribute_attributeName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapabilityAttribute_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapabilityAttribute_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_getCapabilityAttributeDefinition_attributeName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_getCapabilityAttributeDefinition_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_getCapabilityAttributeDefinition_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_capabilityName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_attributeNames_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_attributeNames = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapability_capabilityName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapability_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapability_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_capabilityName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_attributeNames_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_attributeNames = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_capabilityName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_attributeNames_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_attributeNames = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_capabilityName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_agentType = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_capabilities_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_capabilities = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_agentType = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_capabilities_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_capabilities = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_agentType = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_return_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_return = -1;
static int hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_askerData_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_obtainInterface_interfaceName = -1;
static int hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAccessControl_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAccessControl_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAccessControl_authToken = -1;
static int hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_setAccessControl_identity = -1;
static int hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_setAccessControl_operation = -1;
static int hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_setAccessControl_authToken_loop = -1;
static int hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_setAccessControl_authToken = -1;
static int hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_identity = -1;
static int hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_identityType = -1;
static int hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_attributes_loop = -1;
static int hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_authToken_loop = -1;
static int hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_authToken = -1;
static int hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_identity = -1;
static int hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_identityType = -1;
static int hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_attributeNames_loop = -1;
static int hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_attributeNames = -1;
static int hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_expiresIn = -1;
static int hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_authToken_loop = -1;
static int hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_authToken = -1;
static int hf_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_return_loop = -1;
static int hf_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_identity = -1;
static int hf_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_identityType = -1;
static int hf_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_attributeNames_loop = -1;
static int hf_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_attributeNames = -1;
static int hf_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_authToken_loop = -1;
static int hf_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_authToken = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_agent = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_agentType = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_attributes_loop = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_authToken_loop = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_authToken = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_agent = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_capability = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_attributes_loop = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_authToken_loop = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_authToken = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_agent = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_agentType = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_attributeNames_loop = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_attributeNames = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_expiresIn = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_authToken_loop = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_authToken = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_agent = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_capability = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_attributeNames_loop = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_attributeNames = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_expiresIn = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_authToken_loop = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_authToken = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_return_loop = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_agent = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_agentType = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_attributeNames_loop = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_attributeNames = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_authToken_loop = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_authToken = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_return_loop = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_agent = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_capability = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_attributeNames_loop = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_attributeNames = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_authToken_loop = -1;
static int hf_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_authToken = -1;
static int hf_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_return_loop = -1;
static int hf_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_identity = -1;
static int hf_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_attributeNames_loop = -1;
static int hf_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_attributeNames = -1;
static int hf_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_authToken_loop = -1;
static int hf_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_authToken = -1;
static int hf_org_csapi_pam_access_IpPAMAvailability_getAvailability_return_loop = -1;
static int hf_org_csapi_pam_access_IpPAMAvailability_getAvailability_identity = -1;
static int hf_org_csapi_pam_access_IpPAMAvailability_getAvailability_attributeNames_loop = -1;
static int hf_org_csapi_pam_access_IpPAMAvailability_getAvailability_attributeNames = -1;
static int hf_org_csapi_pam_access_IpPAMAvailability_getAvailability_authToken_loop = -1;
static int hf_org_csapi_pam_access_IpPAMAvailability_getAvailability_authToken = -1;
static int hf_org_csapi_pam_access_IpPAMAvailability_getPreference_identity = -1;
static int hf_org_csapi_pam_access_IpPAMAvailability_getPreference_authToken_loop = -1;
static int hf_org_csapi_pam_access_IpPAMAvailability_getPreference_authToken = -1;
static int hf_org_csapi_pam_access_IpPAMAvailability_setPreference_identity = -1;
static int hf_org_csapi_pam_access_IpPAMAvailability_setPreference_operation = -1;
static int hf_org_csapi_pam_access_IpPAMAvailability_setPreference_authToken_loop = -1;
static int hf_org_csapi_pam_access_IpPAMAvailability_setPreference_authToken = -1;
static int hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_return_loop = -1;
static int hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_return = -1;
static int hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_askerData_loop = -1;
static int hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_obtainInterface_interfaceName = -1;
static int hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAccessControl_identity = -1;
static int hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAccessControl_authToken_loop = -1;
static int hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAccessControl_authToken = -1;
static int hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_setAccessControl_identity = -1;
static int hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_setAccessControl_operation = -1;
static int hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_setAccessControl_authToken_loop = -1;
static int hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_setAccessControl_authToken = -1;
static int hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_activateService_identity = -1;
static int hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_activateService_authToken_loop = -1;
static int hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_activateService_authToken = -1;
static int hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_deactivateService_identity = -1;
static int hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_deactivateService_authToken_loop = -1;
static int hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_deactivateService_authToken = -1;
static int hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_isActiveIdentity_return = -1;
static int hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_isActiveIdentity_identity = -1;
static int hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_isActiveIdentity_authToken_loop = -1;
static int hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_isActiveIdentity_authToken = -1;
static int hf_org_csapi_pam_event_IpAppPAMEventHandler_eventNotify_eventID = -1;
static int hf_org_csapi_pam_event_IpAppPAMEventHandler_eventNotify_eventInfo_loop = -1;
static int hf_org_csapi_pam_event_IpAppPAMEventHandler_eventNotifyErr_eventID = -1;
static int hf_org_csapi_pam_event_IpPAMEventHandler_isRegistered_return = -1;
static int hf_org_csapi_pam_event_IpPAMEventHandler_isRegistered_clientID = -1;
static int hf_org_csapi_pam_event_IpPAMEventHandler_isRegistered_authToken_loop = -1;
static int hf_org_csapi_pam_event_IpPAMEventHandler_isRegistered_authToken = -1;
static int hf_org_csapi_pam_event_IpPAMEventHandler_registerAppInterface_return = -1;
static int hf_org_csapi_pam_event_IpPAMEventHandler_registerAppInterface_authToken_loop = -1;
static int hf_org_csapi_pam_event_IpPAMEventHandler_registerAppInterface_authToken = -1;
static int hf_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_return = -1;
static int hf_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_clientID = -1;
static int hf_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_eventList_loop = -1;
static int hf_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_validFor = -1;
static int hf_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_authToken_loop = -1;
static int hf_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_authToken = -1;
static int hf_org_csapi_pam_event_IpPAMEventHandler_deregisterAppInterface_clientID = -1;
static int hf_org_csapi_pam_event_IpPAMEventHandler_deregisterAppInterface_authToken_loop = -1;
static int hf_org_csapi_pam_event_IpPAMEventHandler_deregisterAppInterface_authToken = -1;
static int hf_org_csapi_pam_event_IpPAMEventHandler_deregisterFromEvent_eventID = -1;
static int hf_org_csapi_pam_event_IpPAMEventHandler_deregisterFromEvent_authToken_loop = -1;
static int hf_org_csapi_pam_event_IpPAMEventHandler_deregisterFromEvent_authToken = -1;
static int hf_org_csapi_pam_event_IpPAMEventManager_getAuthToken_return_loop = -1;
static int hf_org_csapi_pam_event_IpPAMEventManager_getAuthToken_return = -1;
static int hf_org_csapi_pam_event_IpPAMEventManager_getAuthToken_askerData_loop = -1;
static int hf_org_csapi_pam_event_IpPAMEventManager_obtainInterface_interfaceName = -1;
static int hf_org_csapi_pam_event_IpPAMEventManager_getAccessControl_identity = -1;
static int hf_org_csapi_pam_event_IpPAMEventManager_getAccessControl_authToken_loop = -1;
static int hf_org_csapi_pam_event_IpPAMEventManager_getAccessControl_authToken = -1;
static int hf_org_csapi_pam_event_IpPAMEventManager_setAccessControl_identity = -1;
static int hf_org_csapi_pam_event_IpPAMEventManager_setAccessControl_operation = -1;
static int hf_org_csapi_pam_event_IpPAMEventManager_setAccessControl_authToken_loop = -1;
static int hf_org_csapi_pam_event_IpPAMEventManager_setAccessControl_authToken = -1;
static int hf_org_csapi_pam_event_IpPAMEventManager_activateService_identity = -1;
static int hf_org_csapi_pam_event_IpPAMEventManager_activateService_authToken_loop = -1;
static int hf_org_csapi_pam_event_IpPAMEventManager_activateService_authToken = -1;
static int hf_org_csapi_pam_event_IpPAMEventManager_deactivateService_identity = -1;
static int hf_org_csapi_pam_event_IpPAMEventManager_deactivateService_authToken_loop = -1;
static int hf_org_csapi_pam_event_IpPAMEventManager_deactivateService_authToken = -1;
static int hf_org_csapi_pam_event_IpPAMEventManager_isActiveIdentity_return = -1;
static int hf_org_csapi_pam_event_IpPAMEventManager_isActiveIdentity_identity = -1;
static int hf_org_csapi_pam_event_IpPAMEventManager_isActiveIdentity_authToken_loop = -1;
static int hf_org_csapi_pam_event_IpPAMEventManager_isActiveIdentity_authToken = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessagingManager_reportNotification_assignmentID = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessagingManager_reportNotification_eventInfo_loop = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMailboxes_mailboxSet_loop = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMultiMediaMessagingSessions_multiMediaMessagingSet_loop = -1;
static int hf_org_csapi_mmm_IpMultiMediaMessagingManager_openMailbox_mailboxID = -1;
static int hf_org_csapi_mmm_IpMultiMediaMessagingManager_openMailbox_authenticationInfo = -1;
static int hf_org_csapi_mmm_IpMultiMediaMessagingManager_createNotification_return = -1;
static int hf_org_csapi_mmm_IpMultiMediaMessagingManager_createNotification_eventCriteria_loop = -1;
static int hf_org_csapi_mmm_IpMultiMediaMessagingManager_destroyNotification_assignmentID = -1;
static int hf_org_csapi_mmm_IpMultiMediaMessagingManager_changeNotification_assignmentID = -1;
static int hf_org_csapi_mmm_IpMultiMediaMessagingManager_changeNotification_eventCriteria_loop = -1;
static int hf_org_csapi_mmm_IpMultiMediaMessagingManager_getNextNotification_reset = -1;
static int hf_org_csapi_mmm_IpMultiMediaMessagingManager_enableNotifications_return = -1;
static int hf_org_csapi_mmm_IpAppMailbox_createFolderRes_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_createFolderRes_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_createFolderRes_folderID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_createFolderErr_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_createFolderErr_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_createFolderErr_error = -1;
static int hf_org_csapi_mmm_IpAppMailbox_createFolderErr_errorDetails = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFoldersRes_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFoldersRes_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFoldersRes_folderID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFoldersRes_folderNames_loop = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFoldersRes_folderNames = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFoldersErr_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFoldersErr_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFoldersErr_folderID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFoldersErr_error = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFoldersErr_errorDetails = -1;
static int hf_org_csapi_mmm_IpAppMailbox_deleteFolderRes_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_deleteFolderRes_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_deleteFolderErr_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_deleteFolderErr_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_deleteFolderErr_error = -1;
static int hf_org_csapi_mmm_IpAppMailbox_deleteFolderErr_errorDetails = -1;
static int hf_org_csapi_mmm_IpAppMailbox_copyFolderRes_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_copyFolderRes_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_copyFolderErr_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_copyFolderErr_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_copyFolderErr_error = -1;
static int hf_org_csapi_mmm_IpAppMailbox_copyFolderErr_errorDetails = -1;
static int hf_org_csapi_mmm_IpAppMailbox_moveFolderRes_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_moveFolderRes_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_moveFolderErr_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_moveFolderErr_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_moveFolderErr_error = -1;
static int hf_org_csapi_mmm_IpAppMailbox_moveFolderErr_errorDetails = -1;
static int hf_org_csapi_mmm_IpAppMailbox_putMessageRes_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_putMessageRes_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_putMessageRes_messageID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_putMessageErr_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_putMessageErr_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_putMessageErr_error = -1;
static int hf_org_csapi_mmm_IpAppMailbox_putMessageErr_errorDetails = -1;
static int hf_org_csapi_mmm_IpAppMailbox_copyMessageRes_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_copyMessageRes_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_copyMessageErr_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_copyMessageErr_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_copyMessageErr_error = -1;
static int hf_org_csapi_mmm_IpAppMailbox_copyMessageErr_errorDetails = -1;
static int hf_org_csapi_mmm_IpAppMailbox_moveMessageRes_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_moveMessageRes_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_moveMessageErr_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_moveMessageErr_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_moveMessageErr_error = -1;
static int hf_org_csapi_mmm_IpAppMailbox_moveMessageErr_errorDetails = -1;
static int hf_org_csapi_mmm_IpAppMailbox_deleteMessageRes_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_deleteMessageRes_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_deleteMessageErr_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_deleteMessageErr_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_deleteMessageErr_error = -1;
static int hf_org_csapi_mmm_IpAppMailbox_deleteMessageErr_errorDetails = -1;
static int hf_org_csapi_mmm_IpAppMailbox_listMessagesRes_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_listMessagesRes_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_listMessagesRes_messageList_loop = -1;
static int hf_org_csapi_mmm_IpAppMailbox_listMessagesRes_final = -1;
static int hf_org_csapi_mmm_IpAppMailbox_listMessagesErr_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_listMessagesErr_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_listMessagesErr_error = -1;
static int hf_org_csapi_mmm_IpAppMailbox_listMessagesErr_errorDetails = -1;
static int hf_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsRes_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsRes_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsRes_partsList_loop = -1;
static int hf_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsErr_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsErr_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsErr_error = -1;
static int hf_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsErr_errorDetails = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsRes_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsRes_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsRes_bodyParts_loop = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsErr_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsErr_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsErr_error = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsErr_errorDetails = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageHeadersRes_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageHeadersRes_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageHeadersRes_headers_loop = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageHeadersErr_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageHeadersErr_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageHeadersErr_error = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageHeadersErr_errorDetails = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageContentRes_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageContentRes_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageContentRes_contentType = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageContentRes_contentTransferEncoding = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageContentRes_content_loop = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageContentRes_content = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageContentErr_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageContentErr_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageContentErr_error = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageContentErr_errorDetails = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFullMessageRes_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFullMessageRes_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFullMessageRes_message_loop = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFullMessageRes_message = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFullMessageErr_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFullMessageErr_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFullMessageErr_error = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFullMessageErr_errorDetails = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesRes_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesRes_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesRes_returnedProperties_loop = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesRes_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesRes_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesRes_folderID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesRes_returnedProperties_loop = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesRes_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesRes_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesRes_messageID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesRes_returnedProperties_loop = -1;
static int hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesRes_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesRes_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesRes_messageID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesRes_propertiesUpdated_loop = -1;
static int hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesErr_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesErr_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesErr_messageID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesErr_propertiesNotUpdated_loop = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesErr_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesErr_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesErr_error = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesErr_errorDetails = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesErr_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesErr_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesErr_error = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesErr_errorDetails = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesErr_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesErr_requestID = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesErr_error = -1;
static int hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesErr_errorDetails = -1;
static int hf_org_csapi_mmm_IpMailbox_close_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpMailbox_createFolderReq_return = -1;
static int hf_org_csapi_mmm_IpMailbox_createFolderReq_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpMailbox_createFolderReq_folderID = -1;
static int hf_org_csapi_mmm_IpMailbox_getFoldersReq_return = -1;
static int hf_org_csapi_mmm_IpMailbox_getFoldersReq_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpMailbox_getFoldersReq_folderID = -1;
static int hf_org_csapi_mmm_IpMailbox_deleteFolderReq_return = -1;
static int hf_org_csapi_mmm_IpMailbox_deleteFolderReq_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpMailbox_deleteFolderReq_folderID = -1;
static int hf_org_csapi_mmm_IpMailbox_copyFolderReq_return = -1;
static int hf_org_csapi_mmm_IpMailbox_copyFolderReq_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpMailbox_copyFolderReq_sourceFolderID = -1;
static int hf_org_csapi_mmm_IpMailbox_copyFolderReq_destinationFolderID = -1;
static int hf_org_csapi_mmm_IpMailbox_moveFolderReq_return = -1;
static int hf_org_csapi_mmm_IpMailbox_moveFolderReq_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpMailbox_moveFolderReq_sourceFolderID = -1;
static int hf_org_csapi_mmm_IpMailbox_moveFolderReq_destinationFolderID = -1;
static int hf_org_csapi_mmm_IpMailbox_putMessageReq_return = -1;
static int hf_org_csapi_mmm_IpMailbox_putMessageReq_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpMailbox_putMessageReq_folderID = -1;
static int hf_org_csapi_mmm_IpMailbox_putMessageReq_message_loop = -1;
static int hf_org_csapi_mmm_IpMailbox_putMessageReq_message = -1;
static int hf_org_csapi_mmm_IpMailbox_copyMessageReq_return = -1;
static int hf_org_csapi_mmm_IpMailbox_copyMessageReq_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpMailbox_copyMessageReq_fromFolderID = -1;
static int hf_org_csapi_mmm_IpMailbox_copyMessageReq_toFolderID = -1;
static int hf_org_csapi_mmm_IpMailbox_copyMessageReq_messageID = -1;
static int hf_org_csapi_mmm_IpMailbox_moveMessageReq_return = -1;
static int hf_org_csapi_mmm_IpMailbox_moveMessageReq_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpMailbox_moveMessageReq_fromFolderID = -1;
static int hf_org_csapi_mmm_IpMailbox_moveMessageReq_toFolderID = -1;
static int hf_org_csapi_mmm_IpMailbox_moveMessageReq_messageID = -1;
static int hf_org_csapi_mmm_IpMailbox_deleteMessageReq_return = -1;
static int hf_org_csapi_mmm_IpMailbox_deleteMessageReq_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpMailbox_deleteMessageReq_fromFolderID = -1;
static int hf_org_csapi_mmm_IpMailbox_deleteMessageReq_messageID = -1;
static int hf_org_csapi_mmm_IpMailbox_listMessagesReq_return = -1;
static int hf_org_csapi_mmm_IpMailbox_listMessagesReq_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpMailbox_listMessagesReq_folderID = -1;
static int hf_org_csapi_mmm_IpMailbox_listMessagesReq_reset = -1;
static int hf_org_csapi_mmm_IpMailbox_listMessageBodyPartsReq_return = -1;
static int hf_org_csapi_mmm_IpMailbox_listMessageBodyPartsReq_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpMailbox_listMessageBodyPartsReq_folderID = -1;
static int hf_org_csapi_mmm_IpMailbox_listMessageBodyPartsReq_messageID = -1;
static int hf_org_csapi_mmm_IpMailbox_listMessageBodyPartsReq_maxNestingLevel = -1;
static int hf_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_return = -1;
static int hf_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_folderID = -1;
static int hf_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_messageID = -1;
static int hf_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_partIDs_loop = -1;
static int hf_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_partIDs = -1;
static int hf_org_csapi_mmm_IpMailbox_getMessageHeadersReq_return = -1;
static int hf_org_csapi_mmm_IpMailbox_getMessageHeadersReq_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpMailbox_getMessageHeadersReq_folderID = -1;
static int hf_org_csapi_mmm_IpMailbox_getMessageHeadersReq_messageID = -1;
static int hf_org_csapi_mmm_IpMailbox_getMessageContentReq_return = -1;
static int hf_org_csapi_mmm_IpMailbox_getMessageContentReq_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpMailbox_getMessageContentReq_folderID = -1;
static int hf_org_csapi_mmm_IpMailbox_getMessageContentReq_messageID = -1;
static int hf_org_csapi_mmm_IpMailbox_getFullMessageReq_return = -1;
static int hf_org_csapi_mmm_IpMailbox_getFullMessageReq_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpMailbox_getFullMessageReq_folderID = -1;
static int hf_org_csapi_mmm_IpMailbox_getFullMessageReq_messageID = -1;
static int hf_org_csapi_mmm_IpMailbox_getMailboxInfoPropertiesReq_return = -1;
static int hf_org_csapi_mmm_IpMailbox_getMailboxInfoPropertiesReq_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpMailbox_getFolderInfoPropertiesReq_return = -1;
static int hf_org_csapi_mmm_IpMailbox_getFolderInfoPropertiesReq_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpMailbox_getFolderInfoPropertiesReq_folderID = -1;
static int hf_org_csapi_mmm_IpMailbox_getMessageInfoPropertiesReq_return = -1;
static int hf_org_csapi_mmm_IpMailbox_getMessageInfoPropertiesReq_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpMailbox_getMessageInfoPropertiesReq_messageID = -1;
static int hf_org_csapi_mmm_IpMailbox_setMessageInfoPropertiesReq_return = -1;
static int hf_org_csapi_mmm_IpMailbox_setMessageInfoPropertiesReq_mailboxSessionID = -1;
static int hf_org_csapi_mmm_IpMailbox_setMessageInfoPropertiesReq_messageID = -1;
static int hf_org_csapi_mmm_IpMailbox_setMessageInfoPropertiesReq_properties_loop = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_sendMessageRes_sessionID = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_sendMessageRes_assignmentID = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_sendMessageErr_sessionID = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_sendMessageErr_assignmentID = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_sendMessageErr_error = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_sendMessageErr_errorDetails = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_cancelMessageRes_sessionID = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_cancelMessageRes_assignmentID = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_cancelMessageErr_sessionID = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_cancelMessageErr_assignmentID = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_cancelMessageErr_error = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_cancelMessageErr_errorDetails = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusRes_sessionID = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusRes_assignmentID = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusRes_result_loop = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusErr_sessionID = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusErr_assignmentID = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusErr_error = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusErr_errorDetails = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageStatusReport_sessionID = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageStatusReport_assignmentID = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageStatusReport_deliveryReportType = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageStatusReport_deliveryReportInfo = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_sessionID = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_message_loop = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_message = -1;
static int hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_headers_loop = -1;
static int hf_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_return = -1;
static int hf_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_sessionID = -1;
static int hf_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_deliveryType = -1;
static int hf_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_messageTreatment_loop = -1;
static int hf_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_message_loop = -1;
static int hf_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_message = -1;
static int hf_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_additionalHeaders_loop = -1;
static int hf_org_csapi_mmm_IpMultiMediaMessaging_cancelMessageReq_sessionID = -1;
static int hf_org_csapi_mmm_IpMultiMediaMessaging_cancelMessageReq_assignmentID = -1;
static int hf_org_csapi_mmm_IpMultiMediaMessaging_queryStatusReq_sessionID = -1;
static int hf_org_csapi_mmm_IpMultiMediaMessaging_queryStatusReq_assignmentID = -1;
static int hf_org_csapi_mmm_IpMultiMediaMessaging_close_sessionID = -1;
        /* Struct filters */
static int hf_org_csapi_TpAddress_Plan = -1;
static int hf_org_csapi_TpAddress_AddrString = -1;
static int hf_org_csapi_TpAddress_Name = -1;
static int hf_org_csapi_TpAddress_Presentation = -1;
static int hf_org_csapi_TpAddress_Screening = -1;
static int hf_org_csapi_TpAddress_SubAddressString = -1;
static int hf_org_csapi_TpAddressRange_Plan = -1;
static int hf_org_csapi_TpAddressRange_AddrString = -1;
static int hf_org_csapi_TpAddressRange_Name = -1;
static int hf_org_csapi_TpAddressRange_SubAddressString = -1;
static int hf_org_csapi_TpCAIElements_UnitsPerInterval = -1;
static int hf_org_csapi_TpCAIElements_SecondsPerTimeInterval = -1;
static int hf_org_csapi_TpCAIElements_ScalingFactor = -1;
static int hf_org_csapi_TpCAIElements_UnitIncrement = -1;
static int hf_org_csapi_TpCAIElements_UnitsPerDataInterval = -1;
static int hf_org_csapi_TpCAIElements_SegmentsPerDataInterval = -1;
static int hf_org_csapi_TpCAIElements_InitialSecsPerTimeInterval = -1;
static int hf_org_csapi_TpChargePerTime_InitialCharge = -1;
static int hf_org_csapi_TpChargePerTime_CurrentChargePerMinute = -1;
static int hf_org_csapi_TpChargePerTime_NextChargePerMinute = -1;
static int hf_org_csapi_TpAoCInfo_Currency = -1;
static int hf_org_csapi_TpTimeInterval_StartTime = -1;
static int hf_org_csapi_TpTimeInterval_StopTime = -1;
static int hf_org_csapi_TpStructuredAttributeValue_Type = -1;
static int hf_org_csapi_TpAttribute_AttributeName = -1;
static int hf_org_csapi_fw_TpFaultStats_Fault = -1;
static int hf_org_csapi_fw_TpFaultStats_Occurrences = -1;
static int hf_org_csapi_fw_TpFaultStats_MaxDuration = -1;
static int hf_org_csapi_fw_TpFaultStats_TotalDuration = -1;
static int hf_org_csapi_fw_TpFaultStats_NumberOfClientsAffected = -1;
static int hf_org_csapi_fw_TpFaultStatsRecord_FaultStatsSet_loop = -1;
static int hf_org_csapi_fw_TpLoadPolicy_LoadPolicy = -1;
static int hf_org_csapi_fw_TpLoadStatisticData_LoadValue = -1;
static int hf_org_csapi_fw_TpLoadStatisticData_LoadLevel = -1;
static int hf_org_csapi_fw_TpLoadThreshold_LoadThreshold = -1;
static int hf_org_csapi_fw_TpLoadInitVal_LoadLevel = -1;
static int hf_org_csapi_fw_TpProperty_PropertyName = -1;
static int hf_org_csapi_fw_TpProperty_PropertyValue = -1;
static int hf_org_csapi_fw_TpClientAppDescription_ClientAppID = -1;
static int hf_org_csapi_fw_TpClientAppDescription_ClientAppProperties_loop = -1;
static int hf_org_csapi_fw_TpClientAppDescription_HasAccessSession = -1;
static int hf_org_csapi_fw_TpClientAppDescription_HasServiceInstances = -1;
static int hf_org_csapi_fw_TpEntOp_EntOpID = -1;
static int hf_org_csapi_fw_TpEntOp_EntOpProperties_loop = -1;
static int hf_org_csapi_fw_TpSag_SagID = -1;
static int hf_org_csapi_fw_TpSag_SagDescription = -1;
static int hf_org_csapi_fw_TpServiceProperty_ServicePropertyName = -1;
static int hf_org_csapi_fw_TpServiceProperty_ServicePropertyValueList_loop = -1;
static int hf_org_csapi_fw_TpServiceProperty_ServicePropertyValueList = -1;
static int hf_org_csapi_fw_TpServiceDescription_ServiceTypeName = -1;
static int hf_org_csapi_fw_TpServiceDescription_ServicePropertyList_loop = -1;
static int hf_org_csapi_fw_TpService_ServiceID = -1;
static int hf_org_csapi_fw_TpServiceProfileDescription_ServiceContractID = -1;
static int hf_org_csapi_fw_TpServiceProfileDescription_ServiceStartDate = -1;
static int hf_org_csapi_fw_TpServiceProfileDescription_ServiceEndDate = -1;
static int hf_org_csapi_fw_TpServiceProfileDescription_ServiceTypeName = -1;
static int hf_org_csapi_fw_TpServiceProfileDescription_ServiceSubscriptionProperties_loop = -1;
static int hf_org_csapi_fw_TpServiceProfileDescription_InUse = -1;
static int hf_org_csapi_fw_TpServiceProfileDescription_ServiceID = -1;
static int hf_org_csapi_fw_TpServiceTypeProperty_ServicePropertyName = -1;
static int hf_org_csapi_fw_TpServiceTypeProperty_ServiceTypePropertyMode = -1;
static int hf_org_csapi_fw_TpServiceTypeProperty_ServicePropertyTypeName = -1;
static int hf_org_csapi_fw_TpServiceTypeDescription_ServiceTypePropertyList_loop = -1;
static int hf_org_csapi_fw_TpServiceTypeDescription_ServiceTypeNameList_loop = -1;
static int hf_org_csapi_fw_TpServiceTypeDescription_ServiceTypeNameList = -1;
static int hf_org_csapi_fw_TpServiceTypeDescription_AvailableOrUnavailable = -1;
static int hf_org_csapi_fw_TpSignatureAndServiceMgr_DigitalSignature_loop = -1;
static int hf_org_csapi_fw_TpSignatureAndServiceMgr_DigitalSignature = -1;
static int hf_org_csapi_fw_TpPerson_PersonName = -1;
static int hf_org_csapi_fw_TpPerson_PostalAddress = -1;
static int hf_org_csapi_fw_TpPerson_TelephoneNumber = -1;
static int hf_org_csapi_fw_TpPerson_Email = -1;
static int hf_org_csapi_fw_TpPerson_HomePage = -1;
static int hf_org_csapi_fw_TpPerson_PersonProperties_loop = -1;
static int hf_org_csapi_fw_TpServiceContractDescription_ServiceStartDate = -1;
static int hf_org_csapi_fw_TpServiceContractDescription_ServiceEndDate = -1;
static int hf_org_csapi_fw_TpServiceContractDescription_ServiceTypeName = -1;
static int hf_org_csapi_fw_TpServiceContractDescription_ServiceID = -1;
static int hf_org_csapi_fw_TpServiceContractDescription_ServiceSubscriptionProperties_loop = -1;
static int hf_org_csapi_fw_TpServiceContractDescription_InUse = -1;
static int hf_org_csapi_fw_TpLoadStatistic_TimeStamp = -1;
static int hf_org_csapi_fw_TpServiceContract_ServiceContractID = -1;
static int hf_org_csapi_fw_TpServiceProfile_ServiceProfileID = -1;
static int hf_org_csapi_fw_TpSagProfilePair_Sag = -1;
static int hf_org_csapi_fw_TpSagProfilePair_ServiceProfile = -1;
static int hf_org_csapi_fw_TpAddSagMembersConflict_ClientApplication = -1;
static int hf_org_csapi_fw_TpAddSagMembersConflict_Service = -1;
static int hf_org_csapi_fw_TpAssignSagToServiceProfileConflict_ClientApplication = -1;
static int hf_org_csapi_fw_TpAssignSagToServiceProfileConflict_Service = -1;
static int hf_org_csapi_fw_TpServiceTypePropertyValue_ServicePropertyName = -1;
static int hf_org_csapi_fw_TpServiceTypePropertyValue_ServiceTypePropertyMode = -1;
static int hf_org_csapi_fw_TpServiceTypePropertyValue_ServicePropertyTypeName = -1;
static int hf_org_csapi_fw_TpServiceTypePropertyValue_ServicePropertyValueList_loop = -1;
static int hf_org_csapi_fw_TpServiceTypePropertyValue_ServicePropertyValueList = -1;
static int hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_ServiceType = -1;
static int hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_ServiceID = -1;
static int hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_CompatibleServiceID = -1;
static int hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_BackwardCompatibilityLevel = -1;
static int hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_MigrationRequired = -1;
static int hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_DataMigrated = -1;
static int hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_MigrationDataAndTime = -1;
static int hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_MigrationAdditionalInfo_loop = -1;
static int hf_org_csapi_fw_TpFwAgreementInfo_ClientApplicationID = -1;
static int hf_org_csapi_fw_TpFwAgreementInfo_ServiceID = -1;
static int hf_org_csapi_fw_TpFwAgreementInfo_ServiceContractID = -1;
static int hf_org_csapi_fw_TpFwAgreementInfo_ServiceProfileID = -1;
static int hf_org_csapi_ui_TpUIMessageCriteria_EndSequence = -1;
static int hf_org_csapi_ui_TpUIMessageCriteria_MaxMessageTime = -1;
static int hf_org_csapi_ui_TpUIMessageCriteria_MaxMessageSize = -1;
static int hf_org_csapi_ui_TpUIEventInfo_ServiceCode = -1;
static int hf_org_csapi_ui_TpUIEventInfo_DataTypeIndication = -1;
static int hf_org_csapi_ui_TpUIEventInfo_DataString = -1;
static int hf_org_csapi_ui_TpUIEventCriteria_ServiceCode = -1;
static int hf_org_csapi_ui_TpUIEventCriteriaResult_AssignmentID = -1;
static int hf_org_csapi_ui_TpUIEventNotificationInfo_ServiceCode = -1;
static int hf_org_csapi_ui_TpUIEventNotificationInfo_DataTypeIndication = -1;
static int hf_org_csapi_ui_TpUIEventNotificationInfo_UIEventData_loop = -1;
static int hf_org_csapi_ui_TpUIEventNotificationInfo_UIEventData = -1;
static int hf_org_csapi_ui_TpUIRecognitionProperty_PropertyName = -1;
static int hf_org_csapi_ui_TpUIRecognitionProperty_PropertyValue = -1;
static int hf_org_csapi_ui_TpUIRecognitionCriteria_SpeakerID = -1;
static int hf_org_csapi_ui_TpUIRecognitionCriteria_Properties_loop = -1;
static int hf_org_csapi_ui_TpUIRecognitionCriteria_Grammar = -1;
static int hf_org_csapi_ui_TpUICollectCriteria_MinLength = -1;
static int hf_org_csapi_ui_TpUICollectCriteria_MaxLength = -1;
static int hf_org_csapi_ui_TpUICollectCriteria_EndSequence = -1;
static int hf_org_csapi_ui_TpUICollectCriteria_StartTimeout = -1;
static int hf_org_csapi_ui_TpUICollectCriteria_InterCharTimeout = -1;
static int hf_org_csapi_ui_TpUICollectCriteria_CollectMode = -1;
static int hf_org_csapi_ui_TpUIWordOverride_Spelling = -1;
static int hf_org_csapi_ui_TpUIWordOverride_PronounceType = -1;
static int hf_org_csapi_ui_TpUIWordOverride_PronounceAs = -1;
static int hf_org_csapi_ui_TpUISynthesisInfoData_SpeakerGender = -1;
static int hf_org_csapi_ui_TpUISynthesisInfoData_SpeakerAge = -1;
static int hf_org_csapi_ui_TpUISynthesisInfoData_SpeakerRate = -1;
static int hf_org_csapi_ui_TpUISynthesisInfoData_SpeakerRange = -1;
static int hf_org_csapi_ui_TpUISynthesisInfoData_TextData = -1;
static int hf_org_csapi_ui_TpUISynthesisInfoData_WordOverrideSet_loop = -1;
static int hf_org_csapi_cc_TpCallError_ErrorTime = -1;
static int hf_org_csapi_cc_TpCallError_ErrorType = -1;
static int hf_org_csapi_cc_TpCallEndedReport_CallLegSessionID = -1;
static int hf_org_csapi_cc_TpCallEndedReport_Cause = -1;
static int hf_org_csapi_cc_TpCallInfoReport_CallInfoType = -1;
static int hf_org_csapi_cc_TpCallInfoReport_CallInitiationStartTime = -1;
static int hf_org_csapi_cc_TpCallInfoReport_CallConnectedToResourceTime = -1;
static int hf_org_csapi_cc_TpCallInfoReport_CallConnectedToDestinationTime = -1;
static int hf_org_csapi_cc_TpCallInfoReport_CallEndTime = -1;
static int hf_org_csapi_cc_TpCallInfoReport_Cause = -1;
static int hf_org_csapi_cc_TpCallServiceCode_CallServiceCodeType = -1;
static int hf_org_csapi_cc_TpCallServiceCode_ServiceCodeValue = -1;
static int hf_org_csapi_cc_TpCallTreatment_CallTreatmentType = -1;
static int hf_org_csapi_cc_TpCallTreatment_ReleaseCause = -1;
static int hf_org_csapi_cc_TpCallChargePlan_ChargeOrderType = -1;
static int hf_org_csapi_cc_TpCallChargePlan_TransparentCharge_loop = -1;
static int hf_org_csapi_cc_TpCallChargePlan_TransparentCharge = -1;
static int hf_org_csapi_cc_TpCallChargePlan_ChargePlan = -1;
static int hf_org_csapi_cc_TpCallChargePlan_AdditionalInfo_loop = -1;
static int hf_org_csapi_cc_TpCallChargePlan_AdditionalInfo = -1;
static int hf_org_csapi_cc_TpCallChargePlan_PartyToCharge = -1;
static int hf_org_csapi_cc_gccs_TpCallEventCriteria_CallEventName = -1;
static int hf_org_csapi_cc_gccs_TpCallEventCriteria_CallNotificationType = -1;
static int hf_org_csapi_cc_gccs_TpCallEventCriteria_MonitorMode = -1;
static int hf_org_csapi_cc_gccs_TpCallEventCriteriaResult_AssignmentID = -1;
static int hf_org_csapi_cc_gccs_TpCallReleaseCause_Value = -1;
static int hf_org_csapi_cc_gccs_TpCallReleaseCause_Location = -1;
static int hf_org_csapi_cc_gccs_TpCallReport_MonitorMode = -1;
static int hf_org_csapi_cc_gccs_TpCallReport_CallEventTime = -1;
static int hf_org_csapi_cc_gccs_TpCallReport_CallReportType = -1;
static int hf_org_csapi_cc_gccs_TpCallReportRequest_MonitorMode = -1;
static int hf_org_csapi_cc_gccs_TpCallReportRequest_CallReportType = -1;
static int hf_org_csapi_cc_gccs_TpCallEventInfo_CallAppInfo_loop = -1;
static int hf_org_csapi_cc_gccs_TpCallEventInfo_CallEventName = -1;
static int hf_org_csapi_cc_gccs_TpCallEventInfo_CallNotificationType = -1;
static int hf_org_csapi_cc_gccs_TpCallEventInfo_MonitorMode = -1;
static int hf_org_csapi_cc_gccs_TpCallEndedReport_CallLegSessionID = -1;
static int hf_org_csapi_cc_gccs_TpCallInfoReport_CallInfoType = -1;
static int hf_org_csapi_cc_gccs_TpCallInfoReport_CallInitiationStartTime = -1;
static int hf_org_csapi_cc_gccs_TpCallInfoReport_CallConnectedToResourceTime = -1;
static int hf_org_csapi_cc_gccs_TpCallInfoReport_CallConnectedToDestinationTime = -1;
static int hf_org_csapi_cc_gccs_TpCallInfoReport_CallEndTime = -1;
static int hf_org_csapi_cc_gccs_TpCallTreatment_CallTreatmentType = -1;
static int hf_org_csapi_cc_gccs_TpCallIdentifier_CallSessionID = -1;
static int hf_org_csapi_cc_TpCallEventRequest_CallEventType = -1;
static int hf_org_csapi_cc_TpCallEventRequest_CallMonitorMode = -1;
static int hf_org_csapi_cc_TpCallNotificationRequest_CallEventsRequested_loop = -1;
static int hf_org_csapi_cc_TpNotificationRequested_AssignmentID = -1;
static int hf_org_csapi_cc_TpCallLegConnectionProperties_AttachMechanism = -1;
static int hf_org_csapi_cc_TpCallEventInfo_CallEventType = -1;
static int hf_org_csapi_cc_TpCallEventInfo_CallMonitorMode = -1;
static int hf_org_csapi_cc_TpCallEventInfo_CallEventTime = -1;
static int hf_org_csapi_cc_TpNotificationRequestedSetEntry_NotificationRequestSet_loop = -1;
static int hf_org_csapi_cc_TpNotificationRequestedSetEntry_Final = -1;
static int hf_org_csapi_cc_TpCarrier_CarrierID_loop = -1;
static int hf_org_csapi_cc_TpCarrier_CarrierID = -1;
static int hf_org_csapi_cc_TpCarrier_CarrierSelectionField = -1;
static int hf_org_csapi_cc_TpCallNotificationInfo_CallAppInfo_loop = -1;
static int hf_org_csapi_cc_TpCallLegInfoReport_CallLegInfoType = -1;
static int hf_org_csapi_cc_TpCallLegInfoReport_CallLegStartTime = -1;
static int hf_org_csapi_cc_TpCallLegInfoReport_CallLegConnectedToResourceTime = -1;
static int hf_org_csapi_cc_TpCallLegInfoReport_CallLegConnectedToAddressTime = -1;
static int hf_org_csapi_cc_TpCallLegInfoReport_CallLegEndTime = -1;
static int hf_org_csapi_cc_TpCallLegInfoReport_CallLegReleaseCause = -1;
static int hf_org_csapi_cc_TpCallLegInfoReport_CallAppInfo_loop = -1;
static int hf_org_csapi_cc_TpCallLegProperty_CallLegPropertyName = -1;
static int hf_org_csapi_cc_TpCallLegProperty_CallLegPropertyValue = -1;
static int hf_org_csapi_cc_mpccs_TpMultiPartyCallIdentifier_CallSessionID = -1;
static int hf_org_csapi_cc_mpccs_TpCallLegIdentifier_CallLegSessionID = -1;
static int hf_org_csapi_cc_mpccs_TpAppCallLegCallBack_AppCallLegSet_loop = -1;
static int hf_org_csapi_cc_mmccs_TpCallSuperviseVolume_VolumeQuantity = -1;
static int hf_org_csapi_cc_mmccs_TpCallSuperviseVolume_VolumeUnit = -1;
static int hf_org_csapi_cc_mmccs_TpMediaStreamRequest_Direction = -1;
static int hf_org_csapi_cc_mmccs_TpMediaStreamRequest_MediaMonitorMode = -1;
static int hf_org_csapi_cc_mmccs_TpMediaStreamRequest_EventType = -1;
static int hf_org_csapi_cc_mmccs_TpMediaStream_Direction = -1;
static int hf_org_csapi_cc_mmccs_TpMediaStream_ChannelSessionID = -1;
static int hf_org_csapi_cc_mmccs_TpNotificationMediaRequest_MediaStreamsRequested_loop = -1;
static int hf_org_csapi_cc_mmccs_TpMediaNotificationRequested_AssignmentID = -1;
static int hf_org_csapi_cc_mmccs_TpMultiMediaCallIdentifier_MMCallSessionID = -1;
static int hf_org_csapi_cc_mmccs_TpMultiMediaCallLegIdentifier_MMCallLegSessionID = -1;
static int hf_org_csapi_cc_mmccs_TpAppMultiMediaCallLegCallBack_AppCallLegSet_loop = -1;
static int hf_org_csapi_cc_cccs_TpMonoMediaConfPolicy_JoinAllowed = -1;
static int hf_org_csapi_cc_cccs_TpJoinEventInfo_CallAppInfo_loop = -1;
static int hf_org_csapi_cc_cccs_TpConfSearchCriteria_StartSearch = -1;
static int hf_org_csapi_cc_cccs_TpConfSearchCriteria_StopSearch = -1;
static int hf_org_csapi_cc_cccs_TpConfSearchCriteria_RequestedResources = -1;
static int hf_org_csapi_cc_cccs_TpConfSearchCriteria_RequestedDuration = -1;
static int hf_org_csapi_cc_cccs_TpConfSearchResult_MatchFound = -1;
static int hf_org_csapi_cc_cccs_TpConfSearchResult_ActualStartTime = -1;
static int hf_org_csapi_cc_cccs_TpConfSearchResult_ActualResources = -1;
static int hf_org_csapi_cc_cccs_TpConfSearchResult_ActualDuration = -1;
static int hf_org_csapi_cc_cccs_TpMultiMediaConfPolicy_JoinAllowed = -1;
static int hf_org_csapi_cc_cccs_TpMultiMediaConfPolicy_MediaAllowed = -1;
static int hf_org_csapi_cc_cccs_TpMultiMediaConfPolicy_Chaired = -1;
static int hf_org_csapi_cc_cccs_TpMultiMediaConfPolicy_VideoHandling = -1;
static int hf_org_csapi_cc_cccs_TpResourceReservation_ReservationID = -1;
static int hf_org_csapi_cc_cccs_TpConfCallIdentifier_ConfCallSessionID = -1;
static int hf_org_csapi_cc_cccs_TpSubConfCallIdentifier_SubConfCallSessionID = -1;
static int hf_org_csapi_ui_TpUIIdentifier_UserInteractionSessionID = -1;
static int hf_org_csapi_ui_TpUICallIdentifier_UserInteractionSessionID = -1;
static int hf_org_csapi_mm_TpLocationResponseTime_ResponseTime = -1;
static int hf_org_csapi_mm_TpLocationResponseTime_TimerValue = -1;
static int hf_org_csapi_mm_TpLocationTrigger_Longitude = -1;
static int hf_org_csapi_mm_TpLocationTrigger_Latitude = -1;
static int hf_org_csapi_mm_TpLocationTrigger_AreaSemiMajor = -1;
static int hf_org_csapi_mm_TpLocationTrigger_AreaSemiMinor = -1;
static int hf_org_csapi_mm_TpLocationTrigger_AngleOfSemiMajor = -1;
static int hf_org_csapi_mm_TpLocationTrigger_Criterion = -1;
static int hf_org_csapi_mm_TpLocationTrigger_ReportingInterval = -1;
static int hf_org_csapi_mm_TpLocationRequest_RequestedAccuracy = -1;
static int hf_org_csapi_mm_TpLocationRequest_AltitudeRequested = -1;
static int hf_org_csapi_mm_TpLocationRequest_Type = -1;
static int hf_org_csapi_mm_TpLocationRequest_Priority = -1;
static int hf_org_csapi_mm_TpLocationRequest_RequestedLocationMethod = -1;
static int hf_org_csapi_mm_TpGeographicalPosition_Longitude = -1;
static int hf_org_csapi_mm_TpGeographicalPosition_Latitude = -1;
static int hf_org_csapi_mm_TpGeographicalPosition_TypeOfUncertaintyShape = -1;
static int hf_org_csapi_mm_TpGeographicalPosition_UncertaintyInnerSemiMajor = -1;
static int hf_org_csapi_mm_TpGeographicalPosition_UncertaintyOuterSemiMajor = -1;
static int hf_org_csapi_mm_TpGeographicalPosition_UncertaintyInnerSemiMinor = -1;
static int hf_org_csapi_mm_TpGeographicalPosition_UncertaintyOuterSemiMinor = -1;
static int hf_org_csapi_mm_TpGeographicalPosition_AngleOfSemiMajor = -1;
static int hf_org_csapi_mm_TpGeographicalPosition_SegmentStartAngle = -1;
static int hf_org_csapi_mm_TpGeographicalPosition_SegmentEndAngle = -1;
static int hf_org_csapi_mm_TpMobilityStopAssignmentData_AssignmentId = -1;
static int hf_org_csapi_mm_TpMobilityStopAssignmentData_StopScope = -1;
static int hf_org_csapi_mm_TpMobilityStopAssignmentData_Users_loop = -1;
static int hf_org_csapi_mm_TpUlExtendedData_TerminalType = -1;
static int hf_org_csapi_mm_TpUlExtendedData_AltitudePresent = -1;
static int hf_org_csapi_mm_TpUlExtendedData_Altitude = -1;
static int hf_org_csapi_mm_TpUlExtendedData_UncertaintyAltitude = -1;
static int hf_org_csapi_mm_TpUlExtendedData_TimestampPresent = -1;
static int hf_org_csapi_mm_TpUlExtendedData_Timestamp = -1;
static int hf_org_csapi_mm_TpUlExtendedData_UsedLocationMethod = -1;
static int hf_org_csapi_mm_TpUserLocation_StatusCode = -1;
static int hf_org_csapi_mm_TpUserLocationExtended_StatusCode = -1;
static int hf_org_csapi_mm_TpUserLocationExtended_Locations_loop = -1;
static int hf_org_csapi_mm_TpLocationTriggerCamel_UpdateInsideVlr = -1;
static int hf_org_csapi_mm_TpLocationTriggerCamel_UpdateOutsideVlr = -1;
static int hf_org_csapi_mm_TpUserLocationCamel_StatusCode = -1;
static int hf_org_csapi_mm_TpUserLocationCamel_GeographicalPositionPresent = -1;
static int hf_org_csapi_mm_TpUserLocationCamel_TimestampPresent = -1;
static int hf_org_csapi_mm_TpUserLocationCamel_Timestamp = -1;
static int hf_org_csapi_mm_TpUserLocationCamel_VlrNumberPresent = -1;
static int hf_org_csapi_mm_TpUserLocationCamel_LocationNumberPresent = -1;
static int hf_org_csapi_mm_TpUserLocationCamel_CellIdOrLaiPresent = -1;
static int hf_org_csapi_mm_TpUserLocationCamel_CellIdOrLai = -1;
static int hf_org_csapi_mm_TpUserLocationEmergencyRequest_UserAddressPresent = -1;
static int hf_org_csapi_mm_TpUserLocationEmergencyRequest_NaEsrdPresent = -1;
static int hf_org_csapi_mm_TpUserLocationEmergencyRequest_NaEsrd = -1;
static int hf_org_csapi_mm_TpUserLocationEmergencyRequest_NaEsrkPresent = -1;
static int hf_org_csapi_mm_TpUserLocationEmergencyRequest_NaEsrk = -1;
static int hf_org_csapi_mm_TpUserLocationEmergencyRequest_ImeiPresent = -1;
static int hf_org_csapi_mm_TpUserLocationEmergencyRequest_Imei = -1;
static int hf_org_csapi_mm_TpUserLocationEmergency_StatusCode = -1;
static int hf_org_csapi_mm_TpUserLocationEmergency_UserIdPresent = -1;
static int hf_org_csapi_mm_TpUserLocationEmergency_NaEsrdPresent = -1;
static int hf_org_csapi_mm_TpUserLocationEmergency_NaEsrd = -1;
static int hf_org_csapi_mm_TpUserLocationEmergency_NaEsrkPresent = -1;
static int hf_org_csapi_mm_TpUserLocationEmergency_NaEsrk = -1;
static int hf_org_csapi_mm_TpUserLocationEmergency_ImeiPresent = -1;
static int hf_org_csapi_mm_TpUserLocationEmergency_Imei = -1;
static int hf_org_csapi_mm_TpUserLocationEmergency_TriggeringEvent = -1;
static int hf_org_csapi_mm_TpUserLocationEmergency_GeographicalPositionPresent = -1;
static int hf_org_csapi_mm_TpUserLocationEmergency_AltitudePresent = -1;
static int hf_org_csapi_mm_TpUserLocationEmergency_Altitude = -1;
static int hf_org_csapi_mm_TpUserLocationEmergency_UncertaintyAltitude = -1;
static int hf_org_csapi_mm_TpUserLocationEmergency_TimestampPresent = -1;
static int hf_org_csapi_mm_TpUserLocationEmergency_Timestamp = -1;
static int hf_org_csapi_mm_TpUserLocationEmergency_UsedLocationMethod = -1;
static int hf_org_csapi_mm_TpUserStatus_StatusCode = -1;
static int hf_org_csapi_mm_TpUserStatus_Status = -1;
static int hf_org_csapi_mm_TpUserStatus_TerminalType = -1;
static int hf_org_csapi_mm_TpTriggeredStatusRequest_Users_loop = -1;
static int hf_org_csapi_mm_TpTriggeredStatusRequest_AssignmentID = -1;
static int hf_org_csapi_mm_TpTriggeredStatusRequestSetEntry_Requests_loop = -1;
static int hf_org_csapi_mm_TpTriggeredStatusRequestSetEntry_Final = -1;
static int hf_org_csapi_mm_TpPeriodicLocationRequest_Users_loop = -1;
static int hf_org_csapi_mm_TpPeriodicLocationRequest_ReportingInterval = -1;
static int hf_org_csapi_mm_TpPeriodicLocationRequest_AssignmentID = -1;
static int hf_org_csapi_mm_TpPeriodicLocationRequestSetEntry_Requests_loop = -1;
static int hf_org_csapi_mm_TpPeriodicLocationRequestSetEntry_Final = -1;
static int hf_org_csapi_mm_TpTriggeredLocationRequest_Users_loop = -1;
static int hf_org_csapi_mm_TpTriggeredLocationRequest_AssignmentID = -1;
static int hf_org_csapi_mm_TpTriggeredLocationRequestSetEntry_Requests_loop = -1;
static int hf_org_csapi_mm_TpTriggeredLocationRequestSetEntry_Final = -1;
static int hf_org_csapi_mm_TpUserStatusIndicatorExtended_UserStatusIndicator = -1;
static int hf_org_csapi_mm_TpUserStatusIndicatorExtended_PDPContextActive = -1;
static int hf_org_csapi_mm_TpUserInfo_UserName = -1;
static int hf_org_csapi_mm_TpUserInfo_Password = -1;
static int hf_org_csapi_mm_TpNetworkStatusIndicator_CountryCode = -1;
static int hf_org_csapi_mm_TpNetworkStatusIndicator_MobileNetworkCode = -1;
static int hf_org_csapi_mm_TpNetworkStatusIndicator_AccessTechnology = -1;
static int hf_org_csapi_mm_TpNetworkStatusIndicator_RoamingStatus = -1;
static int hf_org_csapi_mm_TpUserStatusExtended_UserID = -1;
static int hf_org_csapi_mm_TpUserStatusExtended_StatusCode = -1;
static int hf_org_csapi_mm_TpUserStatusExtended_TerminalType = -1;
static int hf_org_csapi_mm_TpUserStatusExtended_AuthenticationStatus = -1;
static int hf_org_csapi_mm_TpUserStatusExtended_UserIPAddress = -1;
static int hf_org_csapi_mm_TpUserStatusExtended_UserMSISDN = -1;
static int hf_org_csapi_mm_TpUserStatusExtended_UserConnectionID = -1;
static int hf_org_csapi_mm_TpUserStatusExtended_AccessPointName = -1;
static int hf_org_csapi_mm_TpBindingNotificationCriteria_NotificationCriteriaType = -1;
static int hf_org_csapi_mm_TpBindingNotificationCriteria_ContactAddressSet_loop = -1;
static int hf_org_csapi_termcap_TpTerminalCapabilities_TerminalCapabilities = -1;
static int hf_org_csapi_termcap_TpTerminalCapabilities_StatusCode = -1;
static int hf_org_csapi_termcap_TpTerminalCapabilityScope_ScopeType = -1;
static int hf_org_csapi_termcap_TpTerminalCapabilityScope_Scope = -1;
static int hf_org_csapi_dsc_TpChargePerVolume_InitialCharge = -1;
static int hf_org_csapi_dsc_TpChargePerVolume_CurrentChargePerKilobyte = -1;
static int hf_org_csapi_dsc_TpChargePerVolume_NextChargePerKilobyte = -1;
static int hf_org_csapi_dsc_TpDataSessionChargePlan_Currency = -1;
static int hf_org_csapi_dsc_TpDataSessionChargePlan_AdditionalInfo = -1;
static int hf_org_csapi_dsc_TpDataSessionError_ErrorTime = -1;
static int hf_org_csapi_dsc_TpDataSessionError_ErrorType = -1;
static int hf_org_csapi_dsc_TpDataSessionEventCriteria_DataSessionEventName = -1;
static int hf_org_csapi_dsc_TpDataSessionEventCriteria_MonitorMode = -1;
static int hf_org_csapi_dsc_TpDataSessionEventInfo_DataSessionEventName = -1;
static int hf_org_csapi_dsc_TpDataSessionEventInfo_MonitorMode = -1;
static int hf_org_csapi_dsc_TpDataSessionEventInfo_QoSClass = -1;
static int hf_org_csapi_dsc_TpDataSessionReleaseCause_Value = -1;
static int hf_org_csapi_dsc_TpDataSessionReleaseCause_Location = -1;
static int hf_org_csapi_dsc_TpDataSessionReport_MonitorMode = -1;
static int hf_org_csapi_dsc_TpDataSessionReport_DataSessionEventTime = -1;
static int hf_org_csapi_dsc_TpDataSessionReport_DataSessionReportType = -1;
static int hf_org_csapi_dsc_TpDataSessionReportRequest_MonitorMode = -1;
static int hf_org_csapi_dsc_TpDataSessionReportRequest_DataSessionReportType = -1;
static int hf_org_csapi_dsc_TpDataSessionSuperviseVolume_VolumeQuantity = -1;
static int hf_org_csapi_dsc_TpDataSessionSuperviseVolume_VolumeUnit = -1;
static int hf_org_csapi_dsc_TpDataSessionEventCriteriaResult_AssignmentID = -1;
static int hf_org_csapi_dsc_TpDataSessionIdentifier_DataSessionID = -1;
static int hf_org_csapi_gms_TpGMSNewMessageArrivedInfo_FolderID = -1;
static int hf_org_csapi_gms_TpGMSNewMessageArrivedInfo_MessageID = -1;
static int hf_org_csapi_gms_TpGMSNewMessageArrivedInfo_NumberOfProperties = -1;
static int hf_org_csapi_gms_TpGMSNewMessageArrivedCriteria_AuthenticationInfo = -1;
static int hf_org_csapi_gms_TpMailboxIdentifier_SessionID = -1;
static int hf_org_csapi_gms_TpMailboxFolderIdentifier_SessionID = -1;
static int hf_org_csapi_cm_TpNameDescrpTagTimePeriod_duration = -1;
static int hf_org_csapi_cm_TpNameDescrpTagTimeOfDay_name = -1;
static int hf_org_csapi_cm_TpNameDescrpTagTimeOfDay_description = -1;
static int hf_org_csapi_cm_TpNameDescrpTagTimeOfDay_tag = -1;
static int hf_org_csapi_cm_TpNameDescrpTagTimeOfDay_value = -1;
static int hf_org_csapi_cm_TpNameDescrpTagString_name = -1;
static int hf_org_csapi_cm_TpNameDescrpTagString_description = -1;
static int hf_org_csapi_cm_TpNameDescrpTagString_tag = -1;
static int hf_org_csapi_cm_TpNameDescrpTagString_value = -1;
static int hf_org_csapi_cm_TpNameDescrpTagMonth_name = -1;
static int hf_org_csapi_cm_TpNameDescrpTagMonth_description = -1;
static int hf_org_csapi_cm_TpNameDescrpTagMonth_tag = -1;
static int hf_org_csapi_cm_TpNameDescrpTagMonth_value = -1;
static int hf_org_csapi_cm_TpNameDescrpTagInt_name = -1;
static int hf_org_csapi_cm_TpNameDescrpTagInt_description = -1;
static int hf_org_csapi_cm_TpNameDescrpTagInt_tag = -1;
static int hf_org_csapi_cm_TpNameDescrpTagInt_value = -1;
static int hf_org_csapi_cm_TpNameDescrpTagDir_name = -1;
static int hf_org_csapi_cm_TpNameDescrpTagDir_description = -1;
static int hf_org_csapi_cm_TpNameDescrpTagDir_tag = -1;
static int hf_org_csapi_cm_TpNameDescrpTagDir_value = -1;
static int hf_org_csapi_cm_TpNameDescrpTagDayOfWeek_name = -1;
static int hf_org_csapi_cm_TpNameDescrpTagDayOfWeek_description = -1;
static int hf_org_csapi_cm_TpNameDescrpTagDayOfWeek_tag = -1;
static int hf_org_csapi_cm_TpNameDescrpTagDayOfWeek_value = -1;
static int hf_org_csapi_cm_TpNameDescrpTagDateTime_name = -1;
static int hf_org_csapi_cm_TpNameDescrpTagDateTime_description = -1;
static int hf_org_csapi_cm_TpNameDescrpTagDateTime_tag = -1;
static int hf_org_csapi_cm_TpNameDescrpTagDateTime_value = -1;
static int hf_org_csapi_cm_TpEndpoint_type = -1;
static int hf_org_csapi_cm_TpEndpoint_id = -1;
static int hf_org_csapi_cm_TpNameDescrpTagExcessLoadAction_name = -1;
static int hf_org_csapi_cm_TpNameDescrpTagExcessLoadAction_description = -1;
static int hf_org_csapi_cm_TpNameDescrpTagExcessLoadAction_tag = -1;
static int hf_org_csapi_cm_TpNameDescrpTagExcessLoadAction_value = -1;
static int hf_org_csapi_cm_TpDsCodepoint_match = -1;
static int hf_org_csapi_cm_TpDsCodepoint_mask = -1;
static int hf_org_csapi_cm_TpIPSubnet_subnetNumber = -1;
static int hf_org_csapi_cm_TpIPSubnet_subnetMask = -1;
static int hf_org_csapi_cm_TpIPSubnet_addressType = -1;
static int hf_org_csapi_cm_TpIPSubnet_IPVersionSupport = -1;
static int hf_org_csapi_am_TpBalanceInfo_Currency = -1;
static int hf_org_csapi_am_TpBalanceInfo_ValuePartA = -1;
static int hf_org_csapi_am_TpBalanceInfo_ValuePartB = -1;
static int hf_org_csapi_am_TpBalanceInfo_Exponent = -1;
static int hf_org_csapi_am_TpBalanceInfo_AdditionalInfo = -1;
static int hf_org_csapi_am_TpChargingEventInfo_ChargingEventName = -1;
static int hf_org_csapi_am_TpChargingEventInfo_ChargingEventTime = -1;
static int hf_org_csapi_am_TpBalance_StatusCode = -1;
static int hf_org_csapi_am_TpTransactionHistory_TransactionID = -1;
static int hf_org_csapi_am_TpTransactionHistory_TimeStamp = -1;
static int hf_org_csapi_am_TpTransactionHistory_AdditionalInfo = -1;
static int hf_org_csapi_am_TpChargingEventCriteria_ChargingEvents_loop = -1;
static int hf_org_csapi_am_TpChargingEventCriteria_ChargingEvents = -1;
static int hf_org_csapi_am_TpChargingEventCriteria_Users_loop = -1;
static int hf_org_csapi_am_TpChargingEventCriteriaResult_AssignmentID = -1;
static int hf_org_csapi_am_TpBalanceExpiryDate_StatusCode = -1;
static int hf_org_csapi_am_TpBalanceExpiryDate_ExpiryDate = -1;
static int hf_org_csapi_am_TpVoucher_VoucherID = -1;
static int hf_org_csapi_cs_TpMerchantAccountID_MerchantID = -1;
static int hf_org_csapi_cs_TpMerchantAccountID_AccountID = -1;
static int hf_org_csapi_cs_TpCorrelationID_CorrelationID = -1;
static int hf_org_csapi_cs_TpCorrelationID_CorrelationType = -1;
static int hf_org_csapi_cs_TpAmount_Number = -1;
static int hf_org_csapi_cs_TpAmount_Exponent = -1;
static int hf_org_csapi_cs_TpChargingPrice_Currency = -1;
static int hf_org_csapi_cs_TpVolume_Unit = -1;
static int hf_org_csapi_cs_TpChargingSessionID_ChargingSessionID = -1;
static int hf_org_csapi_cs_TpChargingSessionID_RequestNumberFirstRequest = -1;
static int hf_org_csapi_cs_TpApplicationDescription_Text = -1;
static int hf_org_csapi_cs_TpApplicationDescription_AppInformation_loop = -1;
static int hf_org_csapi_cs_TpChargingParameter_ParameterID = -1;
static int hf_org_csapi_policy_TpPolicyEvent_EventID = -1;
static int hf_org_csapi_policy_TpPolicyEvent_TimeGenerated = -1;
static int hf_org_csapi_policy_TpPolicyEvent_Attributes_loop = -1;
static int hf_org_csapi_policy_TpPolicyEvent_EventDefinitionName = -1;
static int hf_org_csapi_policy_TpPolicyEvent_EventDomainName = -1;
static int hf_org_csapi_policy_TpPolicyNameValue_Name = -1;
static int hf_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Names_loop = -1;
static int hf_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Names = -1;
static int hf_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Types_loop = -1;
static int hf_org_csapi_policy_TpPolicyType_TpPolicyListType_ElementType_loop = -1;
static int hf_org_csapi_policy_TpPolicyVar_VarName = -1;
static int hf_org_csapi_policy_TpPolicyConditionListElement_GroupNumber = -1;
static int hf_org_csapi_policy_TpPolicyConditionListElement_Negated = -1;
static int hf_org_csapi_policy_TpPolicyActionListElement_SequenceNumber = -1;
static int hf_org_csapi_pam_TpPAMAttributeDef_Name = -1;
static int hf_org_csapi_pam_TpPAMAttributeDef_Type = -1;
static int hf_org_csapi_pam_TpPAMAttributeDef_IsStatic = -1;
static int hf_org_csapi_pam_TpPAMAttributeDef_IsRevertOnExpiration = -1;
static int hf_org_csapi_pam_TpPAMAttribute_AttributeName = -1;
static int hf_org_csapi_pam_TpPAMAttribute_ExpiresIn = -1;
static int hf_org_csapi_pam_TpPAMPresenceData_Name = -1;
static int hf_org_csapi_pam_TpPAMPresenceData_subscriberStatus = -1;
static int hf_org_csapi_pam_TpPAMPresenceData_networkStatus = -1;
static int hf_org_csapi_pam_TpPAMPresenceData_communicationMeans = -1;
static int hf_org_csapi_pam_TpPAMPresenceData_subscriberProvidedLocation = -1;
static int hf_org_csapi_pam_TpPAMPresenceData_networkProvidedLocation = -1;
static int hf_org_csapi_pam_TpPAMPresenceData_Priority = -1;
static int hf_org_csapi_pam_TpPAMPresenceData_otherInfo = -1;
static int hf_org_csapi_pam_TpPAMAvailabilityProfile_PrivacyCode = -1;
static int hf_org_csapi_pam_TpPAMCommunicationContext_CommunicationCapability = -1;
static int hf_org_csapi_pam_TpPAMContext_AskerData_loop = -1;
static int hf_org_csapi_pam_TpPAMAccessControlData_DefaultPolicy = -1;
static int hf_org_csapi_pam_TpPAMAccessControlData_AllowList_loop = -1;
static int hf_org_csapi_pam_TpPAMAccessControlData_AllowList = -1;
static int hf_org_csapi_pam_TpPAMAccessControlData_DenyList_loop = -1;
static int hf_org_csapi_pam_TpPAMAccessControlData_DenyList = -1;
static int hf_org_csapi_pam_TpPAMICEventData_IdentityType_loop = -1;
static int hf_org_csapi_pam_TpPAMICEventData_IdentityType = -1;
static int hf_org_csapi_pam_TpPAMICNotificationData_Identities_loop = -1;
static int hf_org_csapi_pam_TpPAMICNotificationData_Identities = -1;
static int hf_org_csapi_pam_TpPAMIDEventData_IdentityName_loop = -1;
static int hf_org_csapi_pam_TpPAMIDEventData_IdentityName = -1;
static int hf_org_csapi_pam_TpPAMIDEventData_IdentityType_loop = -1;
static int hf_org_csapi_pam_TpPAMIDEventData_IdentityType = -1;
static int hf_org_csapi_pam_TpPAMIDNotificationData_Identities_loop = -1;
static int hf_org_csapi_pam_TpPAMIDNotificationData_Identities = -1;
static int hf_org_csapi_pam_TpPAMGMCEventData_GroupName_loop = -1;
static int hf_org_csapi_pam_TpPAMGMCEventData_GroupName = -1;
static int hf_org_csapi_pam_TpPAMGMCEventData_GroupType_loop = -1;
static int hf_org_csapi_pam_TpPAMGMCEventData_GroupType = -1;
static int hf_org_csapi_pam_TpPAMGMCNotificationData_Groups_loop = -1;
static int hf_org_csapi_pam_TpPAMGMCNotificationData_Groups = -1;
static int hf_org_csapi_pam_TpPAMACEventData_AgentType_loop = -1;
static int hf_org_csapi_pam_TpPAMACEventData_AgentType = -1;
static int hf_org_csapi_pam_TpPAMACNotificationData_Agents_loop = -1;
static int hf_org_csapi_pam_TpPAMACNotificationData_Agents = -1;
static int hf_org_csapi_pam_TpPAMADEventData_AgentName_loop = -1;
static int hf_org_csapi_pam_TpPAMADEventData_AgentName = -1;
static int hf_org_csapi_pam_TpPAMADEventData_AgentType_loop = -1;
static int hf_org_csapi_pam_TpPAMADEventData_AgentType = -1;
static int hf_org_csapi_pam_TpPAMADNotificationData_Agents_loop = -1;
static int hf_org_csapi_pam_TpPAMADNotificationData_Agents = -1;
static int hf_org_csapi_pam_TpPAMAAEventData_IdentityName_loop = -1;
static int hf_org_csapi_pam_TpPAMAAEventData_IdentityName = -1;
static int hf_org_csapi_pam_TpPAMAAEventData_IdentityType_loop = -1;
static int hf_org_csapi_pam_TpPAMAAEventData_IdentityType = -1;
static int hf_org_csapi_pam_TpPAMAAEventData_AgentName_loop = -1;
static int hf_org_csapi_pam_TpPAMAAEventData_AgentName = -1;
static int hf_org_csapi_pam_TpPAMAAEventData_AgentType_loop = -1;
static int hf_org_csapi_pam_TpPAMAAEventData_AgentType = -1;
static int hf_org_csapi_pam_TpPAMAANotificationData_Identity = -1;
static int hf_org_csapi_pam_TpPAMAANotificationData_Agent = -1;
static int hf_org_csapi_pam_TpPAMAUEventData_IdentityName_loop = -1;
static int hf_org_csapi_pam_TpPAMAUEventData_IdentityName = -1;
static int hf_org_csapi_pam_TpPAMAUEventData_IdentityType_loop = -1;
static int hf_org_csapi_pam_TpPAMAUEventData_IdentityType = -1;
static int hf_org_csapi_pam_TpPAMAUEventData_AgentName_loop = -1;
static int hf_org_csapi_pam_TpPAMAUEventData_AgentName = -1;
static int hf_org_csapi_pam_TpPAMAUEventData_AgentType_loop = -1;
static int hf_org_csapi_pam_TpPAMAUEventData_AgentType = -1;
static int hf_org_csapi_pam_TpPAMAUNotificationData_Identity = -1;
static int hf_org_csapi_pam_TpPAMAUNotificationData_Agent = -1;
static int hf_org_csapi_pam_TpPAMCCEventData_IdentityName_loop = -1;
static int hf_org_csapi_pam_TpPAMCCEventData_IdentityName = -1;
static int hf_org_csapi_pam_TpPAMCCEventData_IdentityType_loop = -1;
static int hf_org_csapi_pam_TpPAMCCEventData_IdentityType = -1;
static int hf_org_csapi_pam_TpPAMCCEventData_Capabilities_loop = -1;
static int hf_org_csapi_pam_TpPAMCCEventData_Capabilities = -1;
static int hf_org_csapi_pam_TpPAMCCNotificationData_Identity = -1;
static int hf_org_csapi_pam_TpPAMCCNotificationData_Capabilities_loop = -1;
static int hf_org_csapi_pam_TpPAMCCNotificationData_Capabilities = -1;
static int hf_org_csapi_pam_TpPAMACPSEventData_AgentName_loop = -1;
static int hf_org_csapi_pam_TpPAMACPSEventData_AgentName = -1;
static int hf_org_csapi_pam_TpPAMACPSEventData_AgentType_loop = -1;
static int hf_org_csapi_pam_TpPAMACPSEventData_AgentType = -1;
static int hf_org_csapi_pam_TpPAMACPSEventData_Capabilities_loop = -1;
static int hf_org_csapi_pam_TpPAMACPSEventData_Capabilities = -1;
static int hf_org_csapi_pam_TpPAMACPSEventData_AttributeNames_loop = -1;
static int hf_org_csapi_pam_TpPAMACPSEventData_AttributeNames = -1;
static int hf_org_csapi_pam_TpPAMACPSEventData_ReportingPeriod = -1;
static int hf_org_csapi_pam_TpPAMACPSNotificationData_Agent = -1;
static int hf_org_csapi_pam_TpPAMACPSNotificationData_Capability = -1;
static int hf_org_csapi_pam_TpPAMACPSNotificationData_AttributeNames_loop = -1;
static int hf_org_csapi_pam_TpPAMACPSNotificationData_AttributeNames = -1;
static int hf_org_csapi_pam_TpPAMAPSEventData_AgentName_loop = -1;
static int hf_org_csapi_pam_TpPAMAPSEventData_AgentName = -1;
static int hf_org_csapi_pam_TpPAMAPSEventData_AgentType_loop = -1;
static int hf_org_csapi_pam_TpPAMAPSEventData_AgentType = -1;
static int hf_org_csapi_pam_TpPAMAPSEventData_AttributeNames_loop = -1;
static int hf_org_csapi_pam_TpPAMAPSEventData_AttributeNames = -1;
static int hf_org_csapi_pam_TpPAMAPSEventData_ReportingPeriod = -1;
static int hf_org_csapi_pam_TpPAMAPSNotificationData_Agent = -1;
static int hf_org_csapi_pam_TpPAMAPSNotificationData_AttributeNames_loop = -1;
static int hf_org_csapi_pam_TpPAMAPSNotificationData_AttributeNames = -1;
static int hf_org_csapi_pam_TpPAMIPSEventData_IdentityName_loop = -1;
static int hf_org_csapi_pam_TpPAMIPSEventData_IdentityName = -1;
static int hf_org_csapi_pam_TpPAMIPSEventData_IdentityType_loop = -1;
static int hf_org_csapi_pam_TpPAMIPSEventData_IdentityType = -1;
static int hf_org_csapi_pam_TpPAMIPSEventData_AttributeNames_loop = -1;
static int hf_org_csapi_pam_TpPAMIPSEventData_AttributeNames = -1;
static int hf_org_csapi_pam_TpPAMIPSEventData_ReportingPeriod = -1;
static int hf_org_csapi_pam_TpPAMIPSNotificationData_Identity = -1;
static int hf_org_csapi_pam_TpPAMIPSNotificationData_Attributes_loop = -1;
static int hf_org_csapi_pam_TpPAMAVCEventData_IdentityName_loop = -1;
static int hf_org_csapi_pam_TpPAMAVCEventData_IdentityName = -1;
static int hf_org_csapi_pam_TpPAMAVCEventData_IdentityType_loop = -1;
static int hf_org_csapi_pam_TpPAMAVCEventData_IdentityType = -1;
static int hf_org_csapi_pam_TpPAMAVCEventData_PAMContext_loop = -1;
static int hf_org_csapi_pam_TpPAMAVCEventData_AttributeNames_loop = -1;
static int hf_org_csapi_pam_TpPAMAVCEventData_AttributeNames = -1;
static int hf_org_csapi_pam_TpPAMAVCEventData_ReportingPeriod = -1;
static int hf_org_csapi_pam_TpPAMAVCNotificationData_Identity = -1;
static int hf_org_csapi_pam_TpPAMAVCNotificationData_Availability_loop = -1;
static int hf_org_csapi_pam_TpPAMWCEventData_Events_loop = -1;
static int hf_org_csapi_pam_TpPAMWCEventData_Events = -1;
static int hf_org_csapi_pam_TpPAMWCEventData_IdentityName_loop = -1;
static int hf_org_csapi_pam_TpPAMWCEventData_IdentityName = -1;
static int hf_org_csapi_pam_TpPAMWCEventData_IdentityType_loop = -1;
static int hf_org_csapi_pam_TpPAMWCEventData_IdentityType = -1;
static int hf_org_csapi_pam_TpPAMWCEventData_ReportingPeriod = -1;
static int hf_org_csapi_pam_TpPAMWCNotificationData_Event = -1;
static int hf_org_csapi_pam_TpPAMWCNotificationData_ChangeType = -1;
static int hf_org_csapi_pam_TpPAMWCNotificationData_Identity = -1;
static int hf_org_csapi_pam_TpPAMWCNotificationData_Watchers_loop = -1;
static int hf_org_csapi_pam_TpPAMWCNotificationData_Watchers = -1;
static int hf_org_csapi_pam_TpPAMErrorInfo_Cause = -1;
static int hf_org_csapi_mmm_TpMailboxIdentifier_SessionID = -1;
static int hf_org_csapi_mmm_TpMultiMediaMessagingIdentifier_SessionID = -1;
static int hf_org_csapi_mmm_TpListMessagesCriteria_OnlyUnreadMessages = -1;
static int hf_org_csapi_mmm_TpMailboxFolderStatusInformation_TotalMessageCount = -1;
static int hf_org_csapi_mmm_TpMessageDescription_MessageID = -1;
static int hf_org_csapi_mmm_TpMessageDescription_To_loop = -1;
static int hf_org_csapi_mmm_TpMessageDescription_Subject = -1;
static int hf_org_csapi_mmm_TpMessageDescription_ReceivedDate = -1;
static int hf_org_csapi_mmm_TpMessageDescription_Size = -1;
static int hf_org_csapi_mmm_TpBodyPartDescription_ContentDescription = -1;
static int hf_org_csapi_mmm_TpBodyPartDescription_ContentSize = -1;
static int hf_org_csapi_mmm_TpBodyPartDescription_ContentType = -1;
static int hf_org_csapi_mmm_TpBodyPartDescription_ContentTransferEncoding = -1;
static int hf_org_csapi_mmm_TpBodyPartDescription_ContentID = -1;
static int hf_org_csapi_mmm_TpBodyPartDescription_ContentDisposition = -1;
static int hf_org_csapi_mmm_TpBodyPartDescription_PartID = -1;
static int hf_org_csapi_mmm_TpBodyPartDescription_NestingLevel = -1;
static int hf_org_csapi_mmm_TpBodyPart_BodyPartContent_loop = -1;
static int hf_org_csapi_mmm_TpBodyPart_BodyPartContent = -1;
static int hf_org_csapi_mmm_TpGenericHeaderField_FieldName = -1;
static int hf_org_csapi_mmm_TpGenericHeaderField_FieldValue = -1;
static int hf_org_csapi_mmm_TpMessageInfoPropertyError_MessagePropertyName = -1;
static int hf_org_csapi_mmm_TpMessageInfoPropertyError_Error = -1;
static int hf_org_csapi_mmm_TpNewMailboxMessageArrivedCriteria_MailboxID = -1;
static int hf_org_csapi_mmm_TpNewMailboxMessageArrivedCriteria_AuthenticationInfo = -1;
static int hf_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_MailboxID = -1;
static int hf_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_FolderID = -1;
static int hf_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_MessageDescription_loop = -1;
static int hf_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_ExtendedHeaderInformation_loop = -1;
static int hf_org_csapi_mmm_TpQueryStatusReport_ReportedStatus = -1;
static int hf_org_csapi_mmm_TpTerminatingAddressList_ToAddressList_loop = -1;
static int hf_org_csapi_mmm_TpTerminatingAddressList_CcAddressList_loop = -1;
static int hf_org_csapi_mmm_TpTerminatingAddressList_BccAddressList_loop = -1;
static int hf_org_csapi_mmm_TpNewMessageArrivedCriteria_CreateMultiMediaMessagingSession = -1;
static int hf_org_csapi_mmm_TpMessagingNotificationRequested_EventCriteria_loop = -1;
static int hf_org_csapi_mmm_TpMessagingNotificationRequested_AssignmentID = -1;
static int hf_org_csapi_mmm_TpMessagingNotificationRequestedSetEntry_MessagingNotificationRequestedSet_loop = -1;
static int hf_org_csapi_mmm_TpMessagingNotificationRequestedSetEntry_Final = -1;
static int hf_org_csapi_mmm_TpNewMessageArrivedInfo_DestinationAddressSet_loop = -1;
static int hf_org_csapi_mmm_TpNewMessageArrivedInfo_Message_loop = -1;
static int hf_org_csapi_mmm_TpNewMessageArrivedInfo_Message = -1;
static int hf_org_csapi_mmm_TpNewMessageArrivedInfo_Headers_loop = -1;
        /* User exception filters */
static int hf_org_csapi_TpCommonExceptions_ExceptionType = -1;
static int hf_org_csapi_TpCommonExceptions_ExtraInformation = -1;
static int hf_org_csapi_P_INVALID_INTERFACE_TYPE_ExtraInformation = -1;
static int hf_org_csapi_P_INVALID_SESSION_ID_ExtraInformation = -1;
static int hf_org_csapi_fw_P_INVALID_DOMAIN_ID_ExtraInformation = -1;
static int hf_org_csapi_fw_P_INVALID_AUTH_TYPE_ExtraInformation = -1;
static int hf_org_csapi_P_INVALID_VERSION_ExtraInformation = -1;
static int hf_org_csapi_fw_P_ACCESS_DENIED_ExtraInformation = -1;
static int hf_org_csapi_fw_P_INVALID_ACCESS_TYPE_ExtraInformation = -1;
static int hf_org_csapi_fw_P_INVALID_SIGNING_ALGORITHM_ExtraInformation = -1;
static int hf_org_csapi_fw_P_INVALID_SIGNATURE_ExtraInformation = -1;
static int hf_org_csapi_P_INVALID_INTERFACE_NAME_ExtraInformation = -1;
static int hf_org_csapi_fw_P_INVALID_PROPERTY_ExtraInformation = -1;
static int hf_org_csapi_fw_P_NO_ACCEPTABLE_SIGNING_ALGORITHM_ExtraInformation = -1;
static int hf_org_csapi_fw_P_NO_ACCEPTABLE_ENCRYPTION_CAPABILITY_ExtraInformation = -1;
static int hf_org_csapi_fw_P_NO_ACCEPTABLE_AUTHENTICATION_MECHANISM_ExtraInformation = -1;
static int hf_org_csapi_P_INVALID_CRITERIA_ExtraInformation = -1;
static int hf_org_csapi_P_INVALID_EVENT_TYPE_ExtraInformation = -1;
static int hf_org_csapi_P_INVALID_ASSIGNMENT_ID_ExtraInformation = -1;
static int hf_org_csapi_fw_P_INVALID_SERVICE_ID_ExtraInformation = -1;
static int hf_org_csapi_fw_P_SERVICE_NOT_ENABLED_ExtraInformation = -1;
static int hf_org_csapi_P_UNAUTHORISED_PARAMETER_VALUE_ExtraInformation = -1;
static int hf_org_csapi_P_INVALID_TIME_AND_DATE_FORMAT_ExtraInformation = -1;
static int hf_org_csapi_fw_P_INVALID_ACTIVITY_TEST_ID_ExtraInformation = -1;
static int hf_org_csapi_fw_P_ILLEGAL_SERVICE_TYPE_ExtraInformation = -1;
static int hf_org_csapi_fw_P_UNKNOWN_SERVICE_TYPE_ExtraInformation = -1;
static int hf_org_csapi_fw_P_INVALID_AGREEMENT_TEXT_ExtraInformation = -1;
static int hf_org_csapi_fw_P_INVALID_SERVICE_TOKEN_ExtraInformation = -1;
static int hf_org_csapi_fw_P_SERVICE_ACCESS_DENIED_ExtraInformation = -1;
static int hf_org_csapi_fw_P_INVALID_SERVICE_PROFILE_ID_ExtraInformation = -1;
static int hf_org_csapi_fw_P_INVALID_SAG_ID_ExtraInformation = -1;
static int hf_org_csapi_fw_P_INVALID_SAG_TO_SERVICE_PROFILE_ASSIGNMENT_ExtraInformation = -1;
static int hf_org_csapi_fw_P_INVALID_SERVICE_CONTRACT_ID_ExtraInformation = -1;
static int hf_org_csapi_fw_P_INVALID_CLIENT_APP_ID_ExtraInformation = -1;
static int hf_org_csapi_fw_P_INVALID_ADDITION_TO_SAG_ExtraInformation = -1;
static int hf_org_csapi_fw_P_PROPERTY_TYPE_MISMATCH_ExtraInformation = -1;
static int hf_org_csapi_fw_P_DUPLICATE_PROPERTY_NAME_ExtraInformation = -1;
static int hf_org_csapi_fw_P_MISSING_MANDATORY_PROPERTY_ExtraInformation = -1;
static int hf_org_csapi_fw_P_SERVICE_TYPE_UNAVAILABLE_ExtraInformation = -1;
static int hf_org_csapi_fw_P_ILLEGAL_SERVICE_ID_ExtraInformation = -1;
static int hf_org_csapi_fw_P_UNKNOWN_SERVICE_ID_ExtraInformation = -1;
static int hf_org_csapi_P_INVALID_ADDRESS_ExtraInformation = -1;
static int hf_org_csapi_P_UNSUPPORTED_ADDRESS_PLAN_ExtraInformation = -1;
static int hf_org_csapi_P_INVALID_NETWORK_STATE_ExtraInformation = -1;
static int hf_org_csapi_P_INVALID_CURRENCY_ExtraInformation = -1;
static int hf_org_csapi_P_INVALID_AMOUNT_ExtraInformation = -1;
static int hf_org_csapi_P_INFORMATION_NOT_AVAILABLE_ExtraInformation = -1;
static int hf_org_csapi_ui_P_ILLEGAL_ID_ExtraInformation = -1;
static int hf_org_csapi_ui_P_ID_NOT_FOUND_ExtraInformation = -1;
static int hf_org_csapi_ui_P_ILLEGAL_RANGE_ExtraInformation = -1;
static int hf_org_csapi_ui_P_INVALID_COLLECTION_CRITERIA_ExtraInformation = -1;
static int hf_org_csapi_P_APPLICATION_NOT_ACTIVATED_ExtraInformation = -1;
static int hf_org_csapi_mm_P_REQUESTED_ACCURACY_CANNOT_BE_DELIVERED_ExtraInformation = -1;
static int hf_org_csapi_mm_P_REQUESTED_RESPONSE_TIME_CANNOT_BE_DELIVERED_ExtraInformation = -1;
static int hf_org_csapi_mm_P_INVALID_REPORTING_INTERVAL_ExtraInformation = -1;
static int hf_org_csapi_mm_P_TRIGGER_CONDITIONS_NOT_SUBSCRIBED_ExtraInformation = -1;
static int hf_org_csapi_P_UNKNOWN_SUBSCRIBER_ExtraInformation = -1;
static int hf_org_csapi_termcap_P_INVALID_TERMINAL_ID_ExtraInformation = -1;
static int hf_org_csapi_gms_P_GMS_INVALID_MAILBOX_ExtraInformation = -1;
static int hf_org_csapi_gms_P_GMS_INVALID_AUTHENTICATION_INFORMATION_ExtraInformation = -1;
static int hf_org_csapi_gms_P_GMS_LOCKING_LOCKED_MAILBOX_ExtraInformation = -1;
static int hf_org_csapi_gms_P_GMS_UNLOCKING_UNLOCKED_MAILBOX_ExtraInformation = -1;
static int hf_org_csapi_gms_P_GMS_CANNOT_UNLOCK_MAILBOX_ExtraInformation = -1;
static int hf_org_csapi_gms_P_GMS_NUMBER_NOT_POSITIVE_ExtraInformation = -1;
static int hf_org_csapi_gms_P_GMS_PROPERTY_NOT_SET_ExtraInformation = -1;
static int hf_org_csapi_gms_P_GMS_MAILBOX_LOCKED_ExtraInformation = -1;
static int hf_org_csapi_gms_P_GMS_FOLDER_IS_OPEN_ExtraInformation = -1;
static int hf_org_csapi_gms_P_GMS_INVALID_FOLDER_ID_ExtraInformation = -1;
static int hf_org_csapi_gms_P_GMS_INSUFFICIENT_PRIVILEGE_ExtraInformation = -1;
static int hf_org_csapi_gms_P_GMS_MAILBOX_OPEN_ExtraInformation = -1;
static int hf_org_csapi_gms_P_GMS_INVALID_MESSAGE_ID_ExtraInformation = -1;
static int hf_org_csapi_gms_P_GMS_MESSAGE_NOT_REMOVED_ExtraInformation = -1;
static int hf_org_csapi_cm_P_UNKNOWN_VPRP_ID_ExtraInformation = -1;
static int hf_org_csapi_cm_P_UNKNOWN_SLA_ID_ExtraInformation = -1;
static int hf_org_csapi_cm_P_UNKNOWN_STATUS_ExtraInformation = -1;
static int hf_org_csapi_cm_P_UNKNOWN_QOS_INFO_ExtraInformation = -1;
static int hf_org_csapi_cm_P_UNKNOWN_VALIDITY_INFO_ExtraInformation = -1;
static int hf_org_csapi_cm_P_UNKNOWN_PIPEQOSINFO_ExtraInformation = -1;
static int hf_org_csapi_cm_P_UNKNOWN_DSCODEPOINT_ExtraInformation = -1;
static int hf_org_csapi_cm_P_UNKNOWN_VPRP_ExtraInformation = -1;
static int hf_org_csapi_cm_P_ILLEGAL_VPRPID_ExtraInformation = -1;
static int hf_org_csapi_cm_P_ILLEGAL_REF_VALUE_ExtraInformation = -1;
static int hf_org_csapi_cm_P_UNKNOWN_INTERFACE_ExtraInformation = -1;
static int hf_org_csapi_cm_P_CANT_DELETE_VPRP_ExtraInformation = -1;
static int hf_org_csapi_cm_P_UNKNOWN_TEMPLATES_ExtraInformation = -1;
static int hf_org_csapi_cm_P_UNKNOWN_MENU_ExtraInformation = -1;
static int hf_org_csapi_cm_P_UNKNOWN_ENTERPRISE_NETWORK_ExtraInformation = -1;
static int hf_org_csapi_cm_P_UNKNOWN_TEMPLATE_TYPE_ExtraInformation = -1;
static int hf_org_csapi_cm_P_UNKNOWN_DESCRIPTION_ExtraInformation = -1;
static int hf_org_csapi_cm_P_ILLEGAL_SLA_ID_ExtraInformation = -1;
static int hf_org_csapi_cm_P_ILLEGAL_TAG_ExtraInformation = -1;
static int hf_org_csapi_cm_P_ILLEGAL_VALUE_ExtraInformation = -1;
static int hf_org_csapi_cm_P_ILLEGAL_COMBINATION_ExtraInformation = -1;
static int hf_org_csapi_cm_P_UNKNOWN_SITES_ExtraInformation = -1;
static int hf_org_csapi_cm_P_UNKNOWN_VPRN_ExtraInformation = -1;
static int hf_org_csapi_cm_P_ILLEGAL_SITE_ID_ExtraInformation = -1;
static int hf_org_csapi_cm_P_UNKNOWN_SITE_ID_ExtraInformation = -1;
static int hf_org_csapi_cm_P_UNKNOWN_SAPS_ExtraInformation = -1;
static int hf_org_csapi_cm_P_UNKNOWN_SITE_LOCATION_ExtraInformation = -1;
static int hf_org_csapi_cm_P_UNKNOWN_SITE_DESCRIPTION_ExtraInformation = -1;
static int hf_org_csapi_cm_P_UNKNOWN_IPSUBNET_ExtraInformation = -1;
static int hf_org_csapi_cm_P_UNKNOWN_SAP_ExtraInformation = -1;
static int hf_org_csapi_am_P_UNAUTHORIZED_APPLICATION_ExtraInformation = -1;
static int hf_org_csapi_cs_P_INVALID_REQUEST_NUMBER_ExtraInformation = -1;
static int hf_org_csapi_cs_P_INVALID_VOLUME_ExtraInformation = -1;
static int hf_org_csapi_cs_P_INVALID_USER_ExtraInformation = -1;
static int hf_org_csapi_cs_P_INVALID_ACCOUNT_ExtraInformation = -1;
static int hf_org_csapi_policy_P_ACCESS_VIOLATION_ExtraInformation = -1;
static int hf_org_csapi_policy_P_SYNTAX_ERROR_ExtraInformation = -1;
static int hf_org_csapi_policy_P_NAME_SPACE_ERROR_ExtraInformation = -1;
static int hf_org_csapi_policy_P_NO_TRANSACTION_IN_PROCESS_ExtraInformation = -1;
static int hf_org_csapi_policy_P_TRANSACTION_IN_PROCESS_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_IDENTITY_EXISTS_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_UNKNOWN_TYPE_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_INVALID_CREDENTIAL_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_UNKNOWN_IDENTITY_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_MEMBER_EXISTS_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_UNKNOWN_GROUP_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_UNKNOWN_MEMBER_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_IS_CYCLIC_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_NOT_MEMBER_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_ALIAS_EXISTS_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_ALIAS_NOT_UNIQUE_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_UNASSIGNED_ALIAS_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_UNKNOWN_ALIAS_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_TYPE_ASSOCIATED_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_DISASSOCIATED_TYPE_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_UNKNOWN_ATTRIBUTE_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_UNKNOWN_ATTRIBUTES_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_AGENT_EXISTS_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_UNKNOWN_AGENT_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_NO_CAPABILITY_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_UNKNOWN_ASSIGNMENT_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_ATTRIBUTE_EXISTS_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_TYPE_EXISTS_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_UNKNOWN_CAPABILITY_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_UNAVAILABLE_INTERFACE_ExtraInformation = -1;
static int hf_org_csapi_pam_P_PAM_NOT_REGISTERED_ExtraInformation = -1;
static int hf_org_csapi_mmm_P_MMM_INVALID_MAILBOX_ExtraInformation = -1;
static int hf_org_csapi_mmm_P_MMM_INVALID_AUTHENTICATION_INFORMATION_ExtraInformation = -1;
static int hf_org_csapi_mmm_P_MMM_INVALID_FOLDER_ID_ExtraInformation = -1;
static int hf_org_csapi_mmm_P_MMM_MAX_MESSAGE_SIZE_EXCEEDED_ExtraInformation = -1;
static int hf_org_csapi_mmm_P_MMM_INVALID_MESSAGE_ID_ExtraInformation = -1;
static int hf_org_csapi_mmm_P_MMM_INVALID_PART_ID_ExtraInformation = -1;
static int hf_org_csapi_mmm_P_MMM_INVALID_PROPERTY_ExtraInformation = -1;
static int hf_org_csapi_mmm_P_MMM_INVALID_DELIVERY_TYPE_ExtraInformation = -1;
static int hf_org_csapi_mmm_P_MMM_DELIVERY_TYPE_ADDRESS_TYPE_MISMATCH_ExtraInformation = -1;
static int hf_org_csapi_mmm_P_MMM_DELIVERY_TYPE_MESSAGE_TYPE_MISMATCH_ExtraInformation = -1;
static int hf_org_csapi_mmm_P_MMM_INVALID_DELIVERY_TIME_ExtraInformation = -1;
static int hf_org_csapi_mmm_P_MMM_INVALID_VALIDITY_TIME_ExtraInformation = -1;
static int hf_org_csapi_mmm_P_MMM_MAX_SUBJECT_SIZE_EXCEEDED_ExtraInformation = -1;
static int hf_org_csapi_mmm_P_MMM_INVALID_HEADER_ExtraInformation = -1;
static int hf_org_csapi_mmm_P_MMM_CANNOT_CANCEL_ExtraInformation = -1;
static int hf_org_csapi_mmm_P_MMM_INFORMATION_NOT_AVAILABLE_ExtraInformation = -1;
        /* Union filters */
static int hf_org_csapi_TpAoCOrder_TpAoCOrder = -1;
static int hf_org_csapi_TpAoCOrder_NetworkCharge = -1;
static int hf_org_csapi_TpSimpleAttributeValue_TpSimpleAttributeValue = -1;
static int hf_org_csapi_TpSimpleAttributeValue_BooleanValue = -1;
static int hf_org_csapi_TpSimpleAttributeValue_OctetValue = -1;
static int hf_org_csapi_TpSimpleAttributeValue_CharValue = -1;
static int hf_org_csapi_TpSimpleAttributeValue_WCharValue_len = -1;
static int hf_org_csapi_TpSimpleAttributeValue_WCharValue = -1;
static int hf_org_csapi_TpSimpleAttributeValue_StringValue = -1;
static int hf_org_csapi_TpSimpleAttributeValue_WStringValue = -1;
static int hf_org_csapi_TpSimpleAttributeValue_Int16Value = -1;
static int hf_org_csapi_TpSimpleAttributeValue_UnsignedInt16Value = -1;
static int hf_org_csapi_TpSimpleAttributeValue_Int32Value = -1;
static int hf_org_csapi_TpSimpleAttributeValue_UnsignedInt32Value = -1;
static int hf_org_csapi_TpSimpleAttributeValue_Int64Value = -1;
static int hf_org_csapi_TpSimpleAttributeValue_UnsignedInt64Value = -1;
static int hf_org_csapi_TpSimpleAttributeValue_FloatValue = -1;
static int hf_org_csapi_TpSimpleAttributeValue_DoubleValue = -1;
static int hf_org_csapi_TpAttributeValue_TpAttributeValue = -1;
static int hf_org_csapi_TpAttributeValue_XMLValue = -1;
static int hf_org_csapi_fw_TpFwEventCriteria_TpFwEventCriteria = -1;
static int hf_org_csapi_fw_TpFwEventCriteria_EventNameUndefined = -1;
static int hf_org_csapi_fw_TpFwEventCriteria_ServiceTypeNameList_loop = -1;
static int hf_org_csapi_fw_TpFwEventCriteria_ServiceTypeNameList = -1;
static int hf_org_csapi_fw_TpFwEventCriteria_UnavailableServiceTypeNameList_loop = -1;
static int hf_org_csapi_fw_TpFwEventCriteria_UnavailableServiceTypeNameList = -1;
static int hf_org_csapi_fw_TpFwEventCriteria_CompatibleServiceTypeNameList_loop = -1;
static int hf_org_csapi_fw_TpFwEventCriteria_CompatibleServiceTypeNameList = -1;
static int hf_org_csapi_fw_TpFwEventCriteria_SessionCreatedList_loop = -1;
static int hf_org_csapi_fw_TpFwEventCriteria_SessionCreatedList = -1;
static int hf_org_csapi_fw_TpFwEventCriteria_SessionTerminatedList_loop = -1;
static int hf_org_csapi_fw_TpFwEventCriteria_SessionTerminatedList = -1;
static int hf_org_csapi_fw_TpFwEventCriteria_AgreementSignedList_loop = -1;
static int hf_org_csapi_fw_TpFwEventCriteria_AgreementSignedList = -1;
static int hf_org_csapi_fw_TpFwEventCriteria_AgreementEndedList_loop = -1;
static int hf_org_csapi_fw_TpFwEventCriteria_AgreementEndedList = -1;
static int hf_org_csapi_fw_TpLoadStatisticEntityID_TpLoadStatisticEntityID = -1;
static int hf_org_csapi_fw_TpLoadStatisticEntityID_FrameworkID = -1;
static int hf_org_csapi_fw_TpLoadStatisticEntityID_ServiceID = -1;
static int hf_org_csapi_fw_TpLoadStatisticEntityID_ClientAppID = -1;
static int hf_org_csapi_fw_TpLoadStatisticInfo_TpLoadStatisticInfo = -1;
static int hf_org_csapi_fw_TpLoadStatisticInfo_LoadStatisticError = -1;
static int hf_org_csapi_fw_TpDomainID_TpDomainID = -1;
static int hf_org_csapi_fw_TpDomainID_FwID = -1;
static int hf_org_csapi_fw_TpDomainID_ClientAppID = -1;
static int hf_org_csapi_fw_TpDomainID_EntOpID = -1;
static int hf_org_csapi_fw_TpDomainID_ServiceID = -1;
static int hf_org_csapi_fw_TpDomainID_ServiceSupplierID = -1;
static int hf_org_csapi_fw_TpMigrationAdditionalInfo_TpMigrationAdditionalInfo = -1;
static int hf_org_csapi_fw_TpMigrationAdditionalInfo_Dummy = -1;
static int hf_org_csapi_fw_TpFwEventInfo_TpFwEventInfo = -1;
static int hf_org_csapi_fw_TpFwEventInfo_EventNameUndefined = -1;
static int hf_org_csapi_fw_TpFwEventInfo_ServiceIDList_loop = -1;
static int hf_org_csapi_fw_TpFwEventInfo_ServiceIDList = -1;
static int hf_org_csapi_fw_TpFwEventInfo_UnavailableServiceIDList_loop = -1;
static int hf_org_csapi_fw_TpFwEventInfo_UnavailableServiceIDList = -1;
static int hf_org_csapi_fw_TpFwEventInfo_AppSessionCreated = -1;
static int hf_org_csapi_fw_TpFwEventInfo_AppSessionTerminated = -1;
static int hf_org_csapi_ui_TpUIVariableInfo_TpUIVariableInfo = -1;
static int hf_org_csapi_ui_TpUIVariableInfo_VariablePartInteger = -1;
static int hf_org_csapi_ui_TpUIVariableInfo_VariablePartAddress = -1;
static int hf_org_csapi_ui_TpUIVariableInfo_VariablePartTime = -1;
static int hf_org_csapi_ui_TpUIVariableInfo_VariablePartDate = -1;
static int hf_org_csapi_ui_TpUIVariableInfo_VariablePartPrice = -1;
static int hf_org_csapi_ui_TpUIInfo_TpUIInfo = -1;
static int hf_org_csapi_ui_TpUIInfo_InfoID = -1;
static int hf_org_csapi_ui_TpUIInfo_InfoData = -1;
static int hf_org_csapi_ui_TpUIInfo_InfoAddress = -1;
static int hf_org_csapi_ui_TpUIInfo_InfoBinData_loop = -1;
static int hf_org_csapi_ui_TpUIInfo_InfoBinData = -1;
static int hf_org_csapi_ui_TpUIInfo_InfoUUEncData = -1;
static int hf_org_csapi_ui_TpUIInfo_InfoMimeData_loop = -1;
static int hf_org_csapi_ui_TpUIInfo_InfoMimeData = -1;
static int hf_org_csapi_ui_TpUIInfo_InfoWaveData_loop = -1;
static int hf_org_csapi_ui_TpUIInfo_InfoWaveData = -1;
static int hf_org_csapi_ui_TpUIInfo_InfoAuData_loop = -1;
static int hf_org_csapi_ui_TpUIInfo_InfoAuData = -1;
static int hf_org_csapi_ui_TpUIInfo_InfoVXMLData = -1;
static int hf_org_csapi_cc_TpCallAdditionalErrorInfo_TpCallAdditionalErrorInfo = -1;
static int hf_org_csapi_cc_TpCallAdditionalErrorInfo_CallErrorInvalidAddress = -1;
static int hf_org_csapi_cc_TpCallAdditionalErrorInfo_Dummy = -1;
static int hf_org_csapi_cc_TpCallLoadControlMechanism_TpCallLoadControlMechanism = -1;
static int hf_org_csapi_cc_TpCallLoadControlMechanism_CallLoadControlPerInterval = -1;
static int hf_org_csapi_cc_TpCallAdditionalTreatmentInfo_TpCallAdditionalTreatmentInfo = -1;
static int hf_org_csapi_cc_TpCallAdditionalTreatmentInfo_Dummy = -1;
static int hf_org_csapi_cc_TpCallPartyToChargeAdditionalInfo_TpCallPartyToChargeAdditionalInfo = -1;
static int hf_org_csapi_cc_TpCallPartyToChargeAdditionalInfo_Dummy = -1;
static int hf_org_csapi_cc_gccs_TpCallAppInfo_TpCallAppInfo = -1;
static int hf_org_csapi_cc_gccs_TpCallAppInfo_CallAppAlertingMechanism = -1;
static int hf_org_csapi_cc_gccs_TpCallAppInfo_CallAppNetworkAccessType = -1;
static int hf_org_csapi_cc_gccs_TpCallAppInfo_CallAppTeleService = -1;
static int hf_org_csapi_cc_gccs_TpCallAppInfo_CallAppBearerService = -1;
static int hf_org_csapi_cc_gccs_TpCallAppInfo_CallAppPartyCategory = -1;
static int hf_org_csapi_cc_gccs_TpCallAppInfo_CallAppGenericInfo = -1;
static int hf_org_csapi_cc_gccs_TpCallAppInfo_Dummy = -1;
static int hf_org_csapi_cc_gccs_TpCallAdditionalReportInfo_TpCallAdditionalReportInfo = -1;
static int hf_org_csapi_cc_gccs_TpCallAdditionalReportInfo_QueueStatus = -1;
static int hf_org_csapi_cc_gccs_TpCallAdditionalReportInfo_Dummy = -1;
static int hf_org_csapi_cc_gccs_TpCallAdditionalReportCriteria_TpCallAdditionalReportCriteria = -1;
static int hf_org_csapi_cc_gccs_TpCallAdditionalReportCriteria_NoAnswerDuration = -1;
static int hf_org_csapi_cc_gccs_TpCallAdditionalReportCriteria_Dummy = -1;
static int hf_org_csapi_cc_TpCallAdditionalEventInfo_TpCallAdditionalEventInfo = -1;
static int hf_org_csapi_cc_TpCallAdditionalEventInfo_OriginatingReleaseCause = -1;
static int hf_org_csapi_cc_TpCallAdditionalEventInfo_TerminatingReleaseCause = -1;
static int hf_org_csapi_cc_TpCallAdditionalEventInfo_Dummy = -1;
static int hf_org_csapi_cc_TpAdditionalCallEventCriteria_TpAdditionalCallEventCriteria = -1;
static int hf_org_csapi_cc_TpAdditionalCallEventCriteria_MinAddressLength = -1;
static int hf_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingServiceCode_loop = -1;
static int hf_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingReleaseCauseSet_loop = -1;
static int hf_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingReleaseCauseSet = -1;
static int hf_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingReleaseCauseSet_loop = -1;
static int hf_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingReleaseCauseSet = -1;
static int hf_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingServiceCode_loop = -1;
static int hf_org_csapi_cc_TpAdditionalCallEventCriteria_Dummy = -1;
static int hf_org_csapi_cc_TpCallAppInfo_TpCallAppInfo = -1;
static int hf_org_csapi_cc_TpCallAppInfo_CallAppAlertingMechanism = -1;
static int hf_org_csapi_cc_TpCallAppInfo_CallAppNetworkAccessType = -1;
static int hf_org_csapi_cc_TpCallAppInfo_CallAppTeleService = -1;
static int hf_org_csapi_cc_TpCallAppInfo_CallAppBearerService = -1;
static int hf_org_csapi_cc_TpCallAppInfo_CallAppPartyCategory = -1;
static int hf_org_csapi_cc_TpCallAppInfo_CallAppGenericInfo = -1;
static int hf_org_csapi_cc_TpCallAppInfo_CallHighProbabilityCompletion = -1;
static int hf_org_csapi_cc_TpCallAppInfo_CallAppCarrier_loop = -1;
static int hf_org_csapi_cc_TpCallAppInfo_Dummy = -1;
static int hf_org_csapi_cc_mpccs_TpAppMultiPartyCallBack_TpAppMultiPartyCallBack = -1;
static int hf_org_csapi_cc_mpccs_TpAppMultiPartyCallBack_Dummy = -1;
static int hf_org_csapi_cc_mmccs_TpMediaStreamDataTypeRequest_TpMediaStreamDataTypeRequest = -1;
static int hf_org_csapi_cc_mmccs_TpMediaStreamDataTypeRequest_Audio = -1;
static int hf_org_csapi_cc_mmccs_TpMediaStreamDataTypeRequest_Video = -1;
static int hf_org_csapi_cc_mmccs_TpMediaStreamDataTypeRequest_Data = -1;
static int hf_org_csapi_cc_mmccs_TpAppMultiMediaCallBack_TpAppMultiMediaCallBack = -1;
static int hf_org_csapi_cc_mmccs_TpAppMultiMediaCallBack_Dummy = -1;
static int hf_org_csapi_cc_cccs_TpConfPolicy_TpConfPolicy = -1;
static int hf_org_csapi_cc_cccs_TpConfPolicy_Dummy = -1;
static int hf_org_csapi_ui_TpUITargetObject_TpUITargetObject = -1;
static int hf_org_csapi_mm_TpBindingEntry_TpBindingEntry = -1;
static int hf_org_csapi_mm_TpBindingEntry_UBExpires = -1;
static int hf_org_csapi_mm_TpBindingEntry_UBPreference = -1;
static int hf_org_csapi_mm_TpBindingEntry_Dummy = -1;
static int hf_org_csapi_dsc_TpDataSessionChargeOrder_TpDataSessionChargeOrder = -1;
static int hf_org_csapi_dsc_TpDataSessionChargeOrder_NetworkCharge = -1;
static int hf_org_csapi_dsc_TpDataSessionAdditionalErrorInfo_TpDataSessionAdditionalErrorInfo = -1;
static int hf_org_csapi_dsc_TpDataSessionAdditionalErrorInfo_DataSessionErrorInvalidAddress = -1;
static int hf_org_csapi_dsc_TpDataSessionAdditionalErrorInfo_Dummy = -1;
static int hf_org_csapi_dsc_TpDataSessionAdditionalReportInfo_TpDataSessionAdditionalReportInfo = -1;
static int hf_org_csapi_dsc_TpDataSessionAdditionalReportInfo_Dummy = -1;
static int hf_org_csapi_gms_TpMessageInfoProperty_TpMessageInfoProperty = -1;
static int hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageID = -1;
static int hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageSubject = -1;
static int hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageDateSent = -1;
static int hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageDateReceived = -1;
static int hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageDateChanged = -1;
static int hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageSize = -1;
static int hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessagePriority = -1;
static int hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageFormat = -1;
static int hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageFolder = -1;
static int hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageStatus = -1;
static int hf_org_csapi_gms_TpMessageInfoProperty_Dummy = -1;
static int hf_org_csapi_gms_TpMailboxInfoProperty_TpMailboxInfoProperty = -1;
static int hf_org_csapi_gms_TpMailboxInfoProperty_MessagingMailboxOwner = -1;
static int hf_org_csapi_gms_TpMailboxInfoProperty_MessagingMailboxFolder = -1;
static int hf_org_csapi_gms_TpMailboxInfoProperty_MessagingMailboxDateCreated = -1;
static int hf_org_csapi_gms_TpMailboxInfoProperty_MessagingMailboxDateChanged = -1;
static int hf_org_csapi_gms_TpMailboxInfoProperty_Dummy = -1;
static int hf_org_csapi_gms_TpMessagingEventInfo_TpMessagingEventInfo = -1;
static int hf_org_csapi_gms_TpMessagingEventInfo_EventNameUndefined = -1;
static int hf_org_csapi_gms_TpMessagingEventCriteria_TpMessagingEventCriteria = -1;
static int hf_org_csapi_gms_TpMessagingEventCriteria_Dummy = -1;
static int hf_org_csapi_gms_TpFolderInfoProperty_TpFolderInfoProperty = -1;
static int hf_org_csapi_gms_TpFolderInfoProperty_MessagingFolderID = -1;
static int hf_org_csapi_gms_TpFolderInfoProperty_MessagingFolderMessage = -1;
static int hf_org_csapi_gms_TpFolderInfoProperty_MessagingFolderSubfolder = -1;
static int hf_org_csapi_gms_TpFolderInfoProperty_MessagingFolderDateCreated = -1;
static int hf_org_csapi_gms_TpFolderInfoProperty_MessagingFolderDateChanged = -1;
static int hf_org_csapi_gms_TpFolderInfoProperty_Dummy = -1;
static int hf_org_csapi_cs_TpChargingParameterValue_TpChargingParameterValue = -1;
static int hf_org_csapi_cs_TpChargingParameterValue_IntValue = -1;
static int hf_org_csapi_cs_TpChargingParameterValue_FloatValue = -1;
static int hf_org_csapi_cs_TpChargingParameterValue_StringValue = -1;
static int hf_org_csapi_cs_TpChargingParameterValue_BooleanValue = -1;
static int hf_org_csapi_cs_TpChargingParameterValue_OctetValue_loop = -1;
static int hf_org_csapi_cs_TpChargingParameterValue_OctetValue = -1;
static int hf_org_csapi_cs_TpAppInformation_TpAppInformation = -1;
static int hf_org_csapi_cs_TpAppInformation_Timestamp = -1;
static int hf_org_csapi_policy_TpPolicyType_TpPolicyType = -1;
static int hf_org_csapi_policy_TpPolicyType_SimpleType = -1;
static int hf_org_csapi_policy_TpPolicyType_StructuredType = -1;
static int hf_org_csapi_policy_TpPolicyType_XMLString = -1;
static int hf_org_csapi_pam_TpPAMContextData_TpPAMContextData = -1;
static int hf_org_csapi_pam_TpPAMContextData_Dummy = -1;
static int hf_org_csapi_pam_TpPAMEventInfo_TpPAMEventInfo = -1;
static int hf_org_csapi_pam_TpPAMNotificationInfo_TpPAMNotificationInfo = -1;
static int hf_org_csapi_pam_TpPAMPreferenceData_TpPAMPreferenceData = -1;
static int hf_org_csapi_pam_TpPAMPreferenceData_Dummy = -1;
static int hf_org_csapi_mmm_TpFolderInfoProperty_TpFolderInfoProperty = -1;
static int hf_org_csapi_mmm_TpFolderInfoProperty_FolderDateCreated = -1;
static int hf_org_csapi_mmm_TpFolderInfoProperty_FolderDateChanged = -1;
static int hf_org_csapi_mmm_TpFolderInfoProperty_FolderSize = -1;
static int hf_org_csapi_mmm_TpFolderInfoProperty_FolderNumberOfMessages = -1;
static int hf_org_csapi_mmm_TpFolderInfoProperty_Dummy = -1;
static int hf_org_csapi_mmm_TpMailboxInfoProperty_TpMailboxInfoProperty = -1;
static int hf_org_csapi_mmm_TpMailboxInfoProperty_MailboxOwner = -1;
static int hf_org_csapi_mmm_TpMailboxInfoProperty_MailboxDateCreated = -1;
static int hf_org_csapi_mmm_TpMailboxInfoProperty_MailboxDateChanged = -1;
static int hf_org_csapi_mmm_TpMailboxInfoProperty_MailboxSize = -1;
static int hf_org_csapi_mmm_TpMailboxInfoProperty_Dummy = -1;
static int hf_org_csapi_mmm_TpMessageInfoProperty_TpMessageInfoProperty = -1;
static int hf_org_csapi_mmm_TpMessageInfoProperty_MessageDateCreated = -1;
static int hf_org_csapi_mmm_TpMessageInfoProperty_MessageDateReceived = -1;
static int hf_org_csapi_mmm_TpMessageInfoProperty_MessageDateChanged = -1;
static int hf_org_csapi_mmm_TpMessageInfoProperty_MessageSize = -1;
static int hf_org_csapi_mmm_TpMessageInfoProperty_MessageStatus = -1;
static int hf_org_csapi_mmm_TpMessageInfoProperty_Dummy = -1;
static int hf_org_csapi_mmm_TpMessageHeaderField_TpMessageHeaderField = -1;
static int hf_org_csapi_mmm_TpMessageHeaderField_DateSent = -1;
static int hf_org_csapi_mmm_TpMessageHeaderField_From_loop = -1;
static int hf_org_csapi_mmm_TpMessageHeaderField_ReplyTo_loop = -1;
static int hf_org_csapi_mmm_TpMessageHeaderField_To_loop = -1;
static int hf_org_csapi_mmm_TpMessageHeaderField_Cc_loop = -1;
static int hf_org_csapi_mmm_TpMessageHeaderField_Bcc_loop = -1;
static int hf_org_csapi_mmm_TpMessageHeaderField_RFC822MessageID = -1;
static int hf_org_csapi_mmm_TpMessageHeaderField_InReplyTo_loop = -1;
static int hf_org_csapi_mmm_TpMessageHeaderField_InReplyTo = -1;
static int hf_org_csapi_mmm_TpMessageHeaderField_References_loop = -1;
static int hf_org_csapi_mmm_TpMessageHeaderField_References = -1;
static int hf_org_csapi_mmm_TpMessageHeaderField_Subject = -1;
static int hf_org_csapi_mmm_TpMessageHeaderField_Comments = -1;
static int hf_org_csapi_mmm_TpMessageHeaderField_Keywords_loop = -1;
static int hf_org_csapi_mmm_TpMessageHeaderField_Keywords = -1;
static int hf_org_csapi_mmm_TpMessageHeaderField_MimeVersion = -1;
static int hf_org_csapi_mmm_TpMessageHeaderField_MimeContent = -1;
static int hf_org_csapi_mmm_TpMessageHeaderField_MimeEncoding = -1;
static int hf_org_csapi_mmm_TpMessageHeaderField_MimeID = -1;
static int hf_org_csapi_mmm_TpMessageHeaderField_MimeDescription = -1;
static int hf_org_csapi_mmm_TpMessageHeaderField_MimeDisposition = -1;
static int hf_org_csapi_mmm_TpMessageHeaderField_Priority = -1;
static int hf_org_csapi_mmm_TpDeliveryTime_TpDeliveryTime = -1;
static int hf_org_csapi_mmm_TpDeliveryTime_DeliveryTime = -1;
static int hf_org_csapi_mmm_TpDeliveryTime_Dummy = -1;
static int hf_org_csapi_mmm_TpMessageTreatment_TpMessageTreatment = -1;
static int hf_org_csapi_mmm_TpMessageTreatment_DeliveryReport = -1;
static int hf_org_csapi_mmm_TpMessageTreatment_BillingID = -1;
static int hf_org_csapi_mmm_TpMessageTreatment_ValidityTime = -1;
static int hf_org_csapi_mmm_TpMessagingEventCriteria_TpMessagingEventCriteria = -1;
static int hf_org_csapi_mmm_TpMessagingEventCriteria_Dummy = -1;
static int hf_org_csapi_mmm_TpMessagingEventInfo_TpMessagingEventInfo = -1;
static int hf_org_csapi_mmm_TpMessagingEventInfo_EventNameUndefined = -1;
        /* Expert info filters */
static expert_field ei_parlay_unknown_giop_msg = EI_INIT;
static expert_field ei_parlay_unknown_exception = EI_INIT;
static expert_field ei_parlay_unknown_reply_status = EI_INIT;

#define MAX_RECURSION_DEPTH 50 // Arbitrarily chosen.

static proto_tree *start_dissecting(tvbuff_t *tvb, packet_info *pinfo, proto_tree *ptree, int *offset);


/* Struct prototype declaration Start */

/* Struct = IDL:org/csapi/TpAddress:1.0 */
static void decode_org_csapi_TpAddress_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/TpAddressRange:1.0 */
static void decode_org_csapi_TpAddressRange_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/TpCAIElements:1.0 */
static void decode_org_csapi_TpCAIElements_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/TpChargeAdviceInfo:1.0 */
static void decode_org_csapi_TpChargeAdviceInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/TpChargePerTime:1.0 */
static void decode_org_csapi_TpChargePerTime_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/TpAoCInfo:1.0 */
static void decode_org_csapi_TpAoCInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/TpTimeInterval:1.0 */
static void decode_org_csapi_TpTimeInterval_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/TpStructuredAttributeValue:1.0 */
static void decode_org_csapi_TpStructuredAttributeValue_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/TpAttribute:1.0 */
static void decode_org_csapi_TpAttribute_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpFaultStats:1.0 */
static void decode_org_csapi_fw_TpFaultStats_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpFaultStatsRecord:1.0 */
static void decode_org_csapi_fw_TpFaultStatsRecord_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpLoadPolicy:1.0 */
static void decode_org_csapi_fw_TpLoadPolicy_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpLoadStatisticData:1.0 */
static void decode_org_csapi_fw_TpLoadStatisticData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpLoadThreshold:1.0 */
static void decode_org_csapi_fw_TpLoadThreshold_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpLoadInitVal:1.0 */
static void decode_org_csapi_fw_TpLoadInitVal_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpProperty:1.0 */
static void decode_org_csapi_fw_TpProperty_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpClientAppDescription:1.0 */
static void decode_org_csapi_fw_TpClientAppDescription_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpEntOp:1.0 */
static void decode_org_csapi_fw_TpEntOp_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpSag:1.0 */
static void decode_org_csapi_fw_TpSag_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpServiceProperty:1.0 */
static void decode_org_csapi_fw_TpServiceProperty_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpServiceDescription:1.0 */
static void decode_org_csapi_fw_TpServiceDescription_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpService:1.0 */
static void decode_org_csapi_fw_TpService_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpServiceProfileDescription:1.0 */
static void decode_org_csapi_fw_TpServiceProfileDescription_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpServiceTypeProperty:1.0 */
static void decode_org_csapi_fw_TpServiceTypeProperty_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpServiceTypeDescription:1.0 */
static void decode_org_csapi_fw_TpServiceTypeDescription_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpSignatureAndServiceMgr:1.0 */
static void decode_org_csapi_fw_TpSignatureAndServiceMgr_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpPerson:1.0 */
static void decode_org_csapi_fw_TpPerson_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpServiceContractDescription:1.0 */
static void decode_org_csapi_fw_TpServiceContractDescription_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpLoadStatistic:1.0 */
static void decode_org_csapi_fw_TpLoadStatistic_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpServiceContract:1.0 */
static void decode_org_csapi_fw_TpServiceContract_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpServiceProfile:1.0 */
static void decode_org_csapi_fw_TpServiceProfile_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpAuthDomain:1.0 */
static void decode_org_csapi_fw_TpAuthDomain_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpSagProfilePair:1.0 */
static void decode_org_csapi_fw_TpSagProfilePair_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpAddSagMembersConflict:1.0 */
static void decode_org_csapi_fw_TpAddSagMembersConflict_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpAssignSagToServiceProfileConflict:1.0 */
static void decode_org_csapi_fw_TpAssignSagToServiceProfileConflict_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpServiceTypePropertyValue:1.0 */
static void decode_org_csapi_fw_TpServiceTypePropertyValue_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpFwMigrationServiceAvailableInfo:1.0 */
static void decode_org_csapi_fw_TpFwMigrationServiceAvailableInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/fw/TpFwAgreementInfo:1.0 */
static void decode_org_csapi_fw_TpFwAgreementInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/ui/TpUIMessageCriteria:1.0 */
static void decode_org_csapi_ui_TpUIMessageCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/ui/TpUIEventInfo:1.0 */
static void decode_org_csapi_ui_TpUIEventInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/ui/TpUIEventCriteria:1.0 */
static void decode_org_csapi_ui_TpUIEventCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/ui/TpUIEventCriteriaResult:1.0 */
static void decode_org_csapi_ui_TpUIEventCriteriaResult_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/ui/TpUIEventNotificationInfo:1.0 */
static void decode_org_csapi_ui_TpUIEventNotificationInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/ui/TpUIRecognitionProperty:1.0 */
static void decode_org_csapi_ui_TpUIRecognitionProperty_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/ui/TpUIRecognitionCriteria:1.0 */
static void decode_org_csapi_ui_TpUIRecognitionCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/ui/TpUICollectCriteria:1.0 */
static void decode_org_csapi_ui_TpUICollectCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/ui/TpUIWordOverride:1.0 */
static void decode_org_csapi_ui_TpUIWordOverride_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/ui/TpUISynthesisInfoData:1.0 */
static void decode_org_csapi_ui_TpUISynthesisInfoData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/TpCallError:1.0 */
static void decode_org_csapi_cc_TpCallError_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/TpCallEndedReport:1.0 */
static void decode_org_csapi_cc_TpCallEndedReport_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/TpCallInfoReport:1.0 */
static void decode_org_csapi_cc_TpCallInfoReport_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/TpCallServiceCode:1.0 */
static void decode_org_csapi_cc_TpCallServiceCode_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/TpCallTreatment:1.0 */
static void decode_org_csapi_cc_TpCallTreatment_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/TpCallChargePlan:1.0 */
static void decode_org_csapi_cc_TpCallChargePlan_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/gccs/TpCallEventCriteria:1.0 */
static void decode_org_csapi_cc_gccs_TpCallEventCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/gccs/TpCallEventCriteriaResult:1.0 */
static void decode_org_csapi_cc_gccs_TpCallEventCriteriaResult_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/gccs/TpCallReleaseCause:1.0 */
static void decode_org_csapi_cc_gccs_TpCallReleaseCause_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/gccs/TpCallReport:1.0 */
static void decode_org_csapi_cc_gccs_TpCallReport_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/gccs/TpCallReportRequest:1.0 */
static void decode_org_csapi_cc_gccs_TpCallReportRequest_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/gccs/TpCallEventInfo:1.0 */
static void decode_org_csapi_cc_gccs_TpCallEventInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/gccs/TpCallEndedReport:1.0 */
static void decode_org_csapi_cc_gccs_TpCallEndedReport_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/gccs/TpCallInfoReport:1.0 */
static void decode_org_csapi_cc_gccs_TpCallInfoReport_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/gccs/TpCallTreatment:1.0 */
static void decode_org_csapi_cc_gccs_TpCallTreatment_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/gccs/TpCallIdentifier:1.0 */
static void decode_org_csapi_cc_gccs_TpCallIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/TpCallNotificationScope:1.0 */
static void decode_org_csapi_cc_TpCallNotificationScope_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/TpCallNotificationReportScope:1.0 */
static void decode_org_csapi_cc_TpCallNotificationReportScope_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/TpCallEventRequest:1.0 */
static void decode_org_csapi_cc_TpCallEventRequest_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/TpCallNotificationRequest:1.0 */
static void decode_org_csapi_cc_TpCallNotificationRequest_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/TpNotificationRequested:1.0 */
static void decode_org_csapi_cc_TpNotificationRequested_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/TpCallLegConnectionProperties:1.0 */
static void decode_org_csapi_cc_TpCallLegConnectionProperties_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/TpCallEventInfo:1.0 */
static void decode_org_csapi_cc_TpCallEventInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/TpNotificationRequestedSetEntry:1.0 */
static void decode_org_csapi_cc_TpNotificationRequestedSetEntry_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/TpCarrier:1.0 */
static void decode_org_csapi_cc_TpCarrier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/TpCallNotificationInfo:1.0 */
static void decode_org_csapi_cc_TpCallNotificationInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/TpCallLegInfoReport:1.0 */
static void decode_org_csapi_cc_TpCallLegInfoReport_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/TpCallLegProperty:1.0 */
static void decode_org_csapi_cc_TpCallLegProperty_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/mpccs/TpMultiPartyCallIdentifier:1.0 */
static void decode_org_csapi_cc_mpccs_TpMultiPartyCallIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/mpccs/TpCallLegIdentifier:1.0 */
static void decode_org_csapi_cc_mpccs_TpCallLegIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/mpccs/TpAppCallLegCallBack:1.0 */
static void decode_org_csapi_cc_mpccs_TpAppCallLegCallBack_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/mmccs/TpCallSuperviseVolume:1.0 */
static void decode_org_csapi_cc_mmccs_TpCallSuperviseVolume_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/mmccs/TpMediaStreamRequest:1.0 */
static void decode_org_csapi_cc_mmccs_TpMediaStreamRequest_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/mmccs/TpMediaStream:1.0 */
static void decode_org_csapi_cc_mmccs_TpMediaStream_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/mmccs/TpNotificationMediaRequest:1.0 */
static void decode_org_csapi_cc_mmccs_TpNotificationMediaRequest_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/mmccs/TpMediaNotificationRequested:1.0 */
static void decode_org_csapi_cc_mmccs_TpMediaNotificationRequested_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/mmccs/TpMultiMediaCallIdentifier:1.0 */
static void decode_org_csapi_cc_mmccs_TpMultiMediaCallIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/mmccs/TpMultiMediaCallLegIdentifier:1.0 */
static void decode_org_csapi_cc_mmccs_TpMultiMediaCallLegIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/mmccs/TpAppMultiMediaCallLegCallBack:1.0 */
static void decode_org_csapi_cc_mmccs_TpAppMultiMediaCallLegCallBack_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/cccs/TpMonoMediaConfPolicy:1.0 */
static void decode_org_csapi_cc_cccs_TpMonoMediaConfPolicy_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/cccs/TpJoinEventInfo:1.0 */
static void decode_org_csapi_cc_cccs_TpJoinEventInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/cccs/TpConfSearchCriteria:1.0 */
static void decode_org_csapi_cc_cccs_TpConfSearchCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/cccs/TpConfSearchResult:1.0 */
static void decode_org_csapi_cc_cccs_TpConfSearchResult_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/cccs/TpMultiMediaConfPolicy:1.0 */
static void decode_org_csapi_cc_cccs_TpMultiMediaConfPolicy_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/cccs/TpResourceReservation:1.0 */
static void decode_org_csapi_cc_cccs_TpResourceReservation_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/cccs/TpConfCallIdentifier:1.0 */
static void decode_org_csapi_cc_cccs_TpConfCallIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cc/cccs/TpSubConfCallIdentifier:1.0 */
static void decode_org_csapi_cc_cccs_TpSubConfCallIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/ui/TpUIIdentifier:1.0 */
static void decode_org_csapi_ui_TpUIIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/ui/TpUICallIdentifier:1.0 */
static void decode_org_csapi_ui_TpUICallIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpLocationResponseTime:1.0 */
static void decode_org_csapi_mm_TpLocationResponseTime_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpLocationTrigger:1.0 */
static void decode_org_csapi_mm_TpLocationTrigger_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpLocationRequest:1.0 */
static void decode_org_csapi_mm_TpLocationRequest_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpGeographicalPosition:1.0 */
static void decode_org_csapi_mm_TpGeographicalPosition_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpMobilityStopAssignmentData:1.0 */
static void decode_org_csapi_mm_TpMobilityStopAssignmentData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpUlExtendedData:1.0 */
static void decode_org_csapi_mm_TpUlExtendedData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpUserLocation:1.0 */
static void decode_org_csapi_mm_TpUserLocation_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpUserLocationExtended:1.0 */
static void decode_org_csapi_mm_TpUserLocationExtended_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpLocationTriggerCamel:1.0 */
static void decode_org_csapi_mm_TpLocationTriggerCamel_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpUserLocationCamel:1.0 */
static void decode_org_csapi_mm_TpUserLocationCamel_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpUserLocationEmergencyRequest:1.0 */
static void decode_org_csapi_mm_TpUserLocationEmergencyRequest_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpUserLocationEmergency:1.0 */
static void decode_org_csapi_mm_TpUserLocationEmergency_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpUserStatus:1.0 */
static void decode_org_csapi_mm_TpUserStatus_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpTriggeredStatusRequest:1.0 */
static void decode_org_csapi_mm_TpTriggeredStatusRequest_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpTriggeredStatusRequestSetEntry:1.0 */
static void decode_org_csapi_mm_TpTriggeredStatusRequestSetEntry_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpPeriodicLocationRequest:1.0 */
static void decode_org_csapi_mm_TpPeriodicLocationRequest_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpPeriodicLocationRequestSetEntry:1.0 */
static void decode_org_csapi_mm_TpPeriodicLocationRequestSetEntry_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpTriggeredLocationRequest:1.0 */
static void decode_org_csapi_mm_TpTriggeredLocationRequest_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpTriggeredLocationRequestSetEntry:1.0 */
static void decode_org_csapi_mm_TpTriggeredLocationRequestSetEntry_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpUserStatusIndicatorExtended:1.0 */
static void decode_org_csapi_mm_TpUserStatusIndicatorExtended_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpUserInfo:1.0 */
static void decode_org_csapi_mm_TpUserInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpNetworkStatusIndicator:1.0 */
static void decode_org_csapi_mm_TpNetworkStatusIndicator_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpUserStatusExtended:1.0 */
static void decode_org_csapi_mm_TpUserStatusExtended_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mm/TpBindingNotificationCriteria:1.0 */
static void decode_org_csapi_mm_TpBindingNotificationCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/termcap/TpTerminalCapabilities:1.0 */
static void decode_org_csapi_termcap_TpTerminalCapabilities_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/termcap/TpTerminalCapabilityScope:1.0 */
static void decode_org_csapi_termcap_TpTerminalCapabilityScope_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/dsc/TpChargePerVolume:1.0 */
static void decode_org_csapi_dsc_TpChargePerVolume_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/dsc/TpDataSessionChargePlan:1.0 */
static void decode_org_csapi_dsc_TpDataSessionChargePlan_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/dsc/TpDataSessionError:1.0 */
static void decode_org_csapi_dsc_TpDataSessionError_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/dsc/TpDataSessionEventCriteria:1.0 */
static void decode_org_csapi_dsc_TpDataSessionEventCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/dsc/TpDataSessionEventInfo:1.0 */
static void decode_org_csapi_dsc_TpDataSessionEventInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/dsc/TpDataSessionReleaseCause:1.0 */
static void decode_org_csapi_dsc_TpDataSessionReleaseCause_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/dsc/TpDataSessionReport:1.0 */
static void decode_org_csapi_dsc_TpDataSessionReport_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/dsc/TpDataSessionReportRequest:1.0 */
static void decode_org_csapi_dsc_TpDataSessionReportRequest_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/dsc/TpDataSessionSuperviseVolume:1.0 */
static void decode_org_csapi_dsc_TpDataSessionSuperviseVolume_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/dsc/TpDataSessionEventCriteriaResult:1.0 */
static void decode_org_csapi_dsc_TpDataSessionEventCriteriaResult_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/dsc/TpDataSessionIdentifier:1.0 */
static void decode_org_csapi_dsc_TpDataSessionIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/gms/TpGMSNewMessageArrivedInfo:1.0 */
static void decode_org_csapi_gms_TpGMSNewMessageArrivedInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/gms/TpGMSNewMessageArrivedCriteria:1.0 */
static void decode_org_csapi_gms_TpGMSNewMessageArrivedCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/gms/TpMailboxIdentifier:1.0 */
static void decode_org_csapi_gms_TpMailboxIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/gms/TpMailboxFolderIdentifier:1.0 */
static void decode_org_csapi_gms_TpMailboxFolderIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cm/TpNameDescrpTagTimePeriod:1.0 */
static void decode_org_csapi_cm_TpNameDescrpTagTimePeriod_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cm/TpNameDescrpTagTimeOfDay:1.0 */
static void decode_org_csapi_cm_TpNameDescrpTagTimeOfDay_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cm/TpNameDescrpTagString:1.0 */
static void decode_org_csapi_cm_TpNameDescrpTagString_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cm/TpNameDescrpTagMonth:1.0 */
static void decode_org_csapi_cm_TpNameDescrpTagMonth_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cm/TpNameDescrpTagInt:1.0 */
static void decode_org_csapi_cm_TpNameDescrpTagInt_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cm/TpNameDescrpTagDir:1.0 */
static void decode_org_csapi_cm_TpNameDescrpTagDir_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cm/TpNameDescrpTagDayOfWeek:1.0 */
static void decode_org_csapi_cm_TpNameDescrpTagDayOfWeek_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cm/TpNameDescrpTagDateTime:1.0 */
static void decode_org_csapi_cm_TpNameDescrpTagDateTime_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cm/TpLossDescriptor:1.0 */
static void decode_org_csapi_cm_TpLossDescriptor_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cm/TpLoadDescriptor:1.0 */
static void decode_org_csapi_cm_TpLoadDescriptor_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cm/TpJitterDescriptor:1.0 */
static void decode_org_csapi_cm_TpJitterDescriptor_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cm/TpEndpoint:1.0 */
static void decode_org_csapi_cm_TpEndpoint_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cm/TpDelayDescriptor:1.0 */
static void decode_org_csapi_cm_TpDelayDescriptor_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cm/TpNameDescrpTagExcessLoadAction:1.0 */
static void decode_org_csapi_cm_TpNameDescrpTagExcessLoadAction_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cm/TpValidityInfo:1.0 */
static void decode_org_csapi_cm_TpValidityInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cm/TpProvisionedQoSInfo:1.0 */
static void decode_org_csapi_cm_TpProvisionedQoSInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cm/TpPipeQoSInfo:1.0 */
static void decode_org_csapi_cm_TpPipeQoSInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cm/TpDsCodepoint:1.0 */
static void decode_org_csapi_cm_TpDsCodepoint_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cm/TpIPSubnet:1.0 */
static void decode_org_csapi_cm_TpIPSubnet_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/am/TpBalanceInfo:1.0 */
static void decode_org_csapi_am_TpBalanceInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/am/TpChargingEventInfo:1.0 */
static void decode_org_csapi_am_TpChargingEventInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/am/TpBalance:1.0 */
static void decode_org_csapi_am_TpBalance_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/am/TpTransactionHistory:1.0 */
static void decode_org_csapi_am_TpTransactionHistory_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/am/TpChargingEventCriteria:1.0 */
static void decode_org_csapi_am_TpChargingEventCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/am/TpChargingEventCriteriaResult:1.0 */
static void decode_org_csapi_am_TpChargingEventCriteriaResult_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/am/TpBalanceExpiryDate:1.0 */
static void decode_org_csapi_am_TpBalanceExpiryDate_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/am/TpVoucher:1.0 */
static void decode_org_csapi_am_TpVoucher_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cs/TpMerchantAccountID:1.0 */
static void decode_org_csapi_cs_TpMerchantAccountID_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cs/TpCorrelationID:1.0 */
static void decode_org_csapi_cs_TpCorrelationID_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cs/TpAmount:1.0 */
static void decode_org_csapi_cs_TpAmount_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cs/TpChargingPrice:1.0 */
static void decode_org_csapi_cs_TpChargingPrice_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cs/TpVolume:1.0 */
static void decode_org_csapi_cs_TpVolume_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cs/TpChargingSessionID:1.0 */
static void decode_org_csapi_cs_TpChargingSessionID_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cs/TpPriceVolume:1.0 */
static void decode_org_csapi_cs_TpPriceVolume_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cs/TpApplicationDescription:1.0 */
static void decode_org_csapi_cs_TpApplicationDescription_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/cs/TpChargingParameter:1.0 */
static void decode_org_csapi_cs_TpChargingParameter_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/policy/TpPolicyEvent:1.0 */
static void decode_org_csapi_policy_TpPolicyEvent_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/policy/TpPolicyNameValue:1.0 */
static void decode_org_csapi_policy_TpPolicyNameValue_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/policy/TpPolicyType/TpPolicyRecordType:1.0 */
static void decode_org_csapi_policy_TpPolicyType_TpPolicyRecordType_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/policy/TpPolicyType/TpPolicyListType:1.0 */
static void decode_org_csapi_policy_TpPolicyType_TpPolicyListType_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/policy/TpPolicyVar:1.0 */
static void decode_org_csapi_policy_TpPolicyVar_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/policy/TpPolicyConditionListElement:1.0 */
static void decode_org_csapi_policy_TpPolicyConditionListElement_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/policy/TpPolicyActionListElement:1.0 */
static void decode_org_csapi_policy_TpPolicyActionListElement_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMAttributeDef:1.0 */
static void decode_org_csapi_pam_TpPAMAttributeDef_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMAttribute:1.0 */
static void decode_org_csapi_pam_TpPAMAttribute_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMPresenceData:1.0 */
static void decode_org_csapi_pam_TpPAMPresenceData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMAvailabilityProfile:1.0 */
static void decode_org_csapi_pam_TpPAMAvailabilityProfile_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMCommunicationContext:1.0 */
static void decode_org_csapi_pam_TpPAMCommunicationContext_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMContext:1.0 */
static void decode_org_csapi_pam_TpPAMContext_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMAccessControlData:1.0 */
static void decode_org_csapi_pam_TpPAMAccessControlData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMICEventData:1.0 */
static void decode_org_csapi_pam_TpPAMICEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMICNotificationData:1.0 */
static void decode_org_csapi_pam_TpPAMICNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMIDEventData:1.0 */
static void decode_org_csapi_pam_TpPAMIDEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMIDNotificationData:1.0 */
static void decode_org_csapi_pam_TpPAMIDNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMGMCEventData:1.0 */
static void decode_org_csapi_pam_TpPAMGMCEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMGMCNotificationData:1.0 */
static void decode_org_csapi_pam_TpPAMGMCNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMACEventData:1.0 */
static void decode_org_csapi_pam_TpPAMACEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMACNotificationData:1.0 */
static void decode_org_csapi_pam_TpPAMACNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMADEventData:1.0 */
static void decode_org_csapi_pam_TpPAMADEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMADNotificationData:1.0 */
static void decode_org_csapi_pam_TpPAMADNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMAAEventData:1.0 */
static void decode_org_csapi_pam_TpPAMAAEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMAANotificationData:1.0 */
static void decode_org_csapi_pam_TpPAMAANotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMAUEventData:1.0 */
static void decode_org_csapi_pam_TpPAMAUEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMAUNotificationData:1.0 */
static void decode_org_csapi_pam_TpPAMAUNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMCCEventData:1.0 */
static void decode_org_csapi_pam_TpPAMCCEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMCCNotificationData:1.0 */
static void decode_org_csapi_pam_TpPAMCCNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMACPSEventData:1.0 */
static void decode_org_csapi_pam_TpPAMACPSEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMACPSNotificationData:1.0 */
static void decode_org_csapi_pam_TpPAMACPSNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMAPSEventData:1.0 */
static void decode_org_csapi_pam_TpPAMAPSEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMAPSNotificationData:1.0 */
static void decode_org_csapi_pam_TpPAMAPSNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMIPSEventData:1.0 */
static void decode_org_csapi_pam_TpPAMIPSEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMIPSNotificationData:1.0 */
static void decode_org_csapi_pam_TpPAMIPSNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMAVCEventData:1.0 */
static void decode_org_csapi_pam_TpPAMAVCEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMAVCNotificationData:1.0 */
static void decode_org_csapi_pam_TpPAMAVCNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMWCEventData:1.0 */
static void decode_org_csapi_pam_TpPAMWCEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMWCNotificationData:1.0 */
static void decode_org_csapi_pam_TpPAMWCNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/pam/TpPAMErrorInfo:1.0 */
static void decode_org_csapi_pam_TpPAMErrorInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mmm/TpMailboxIdentifier:1.0 */
static void decode_org_csapi_mmm_TpMailboxIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mmm/TpMultiMediaMessagingIdentifier:1.0 */
static void decode_org_csapi_mmm_TpMultiMediaMessagingIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mmm/TpListMessagesCriteria:1.0 */
static void decode_org_csapi_mmm_TpListMessagesCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mmm/TpMailboxFolderStatusInformation:1.0 */
static void decode_org_csapi_mmm_TpMailboxFolderStatusInformation_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mmm/TpMessageDescription:1.0 */
static void decode_org_csapi_mmm_TpMessageDescription_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mmm/TpBodyPartDescription:1.0 */
static void decode_org_csapi_mmm_TpBodyPartDescription_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mmm/TpBodyPart:1.0 */
static void decode_org_csapi_mmm_TpBodyPart_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mmm/TpGenericHeaderField:1.0 */
static void decode_org_csapi_mmm_TpGenericHeaderField_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mmm/TpMessageInfoPropertyError:1.0 */
static void decode_org_csapi_mmm_TpMessageInfoPropertyError_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mmm/TpNewMailboxMessageArrivedCriteria:1.0 */
static void decode_org_csapi_mmm_TpNewMailboxMessageArrivedCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mmm/TpNewMailboxMessageArrivedInfo:1.0 */
static void decode_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mmm/TpQueryStatusReport:1.0 */
static void decode_org_csapi_mmm_TpQueryStatusReport_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mmm/TpTerminatingAddressList:1.0 */
static void decode_org_csapi_mmm_TpTerminatingAddressList_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mmm/TpNewMessageArrivedCriteria:1.0 */
static void decode_org_csapi_mmm_TpNewMessageArrivedCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mmm/TpMessagingNotificationRequested:1.0 */
static void decode_org_csapi_mmm_TpMessagingNotificationRequested_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mmm/TpMessagingNotificationRequestedSetEntry:1.0 */
static void decode_org_csapi_mmm_TpMessagingNotificationRequestedSetEntry_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct = IDL:org/csapi/mmm/TpNewMessageArrivedInfo:1.0 */
static void decode_org_csapi_mmm_TpNewMessageArrivedInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Struct prototype declaration End */

/* Union prototype declaration Start */

/* Union = IDL:org/csapi/TpAoCOrder:1.0 */
static void decode_org_csapi_TpAoCOrder_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/TpSimpleAttributeValue:1.0 */
static void decode_org_csapi_TpSimpleAttributeValue_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/TpAttributeValue:1.0 */
static void decode_org_csapi_TpAttributeValue_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/fw/TpFwEventCriteria:1.0 */
static void decode_org_csapi_fw_TpFwEventCriteria_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/fw/TpLoadStatisticEntityID:1.0 */
static void decode_org_csapi_fw_TpLoadStatisticEntityID_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/fw/TpLoadStatisticInfo:1.0 */
static void decode_org_csapi_fw_TpLoadStatisticInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/fw/TpDomainID:1.0 */
static void decode_org_csapi_fw_TpDomainID_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/fw/TpMigrationAdditionalInfo:1.0 */
static void decode_org_csapi_fw_TpMigrationAdditionalInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/fw/TpFwEventInfo:1.0 */
static void decode_org_csapi_fw_TpFwEventInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/ui/TpUIVariableInfo:1.0 */
static void decode_org_csapi_ui_TpUIVariableInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/ui/TpUIInfo:1.0 */
static void decode_org_csapi_ui_TpUIInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/cc/TpCallAdditionalErrorInfo:1.0 */
static void decode_org_csapi_cc_TpCallAdditionalErrorInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/cc/TpCallLoadControlMechanism:1.0 */
static void decode_org_csapi_cc_TpCallLoadControlMechanism_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/cc/TpCallAdditionalTreatmentInfo:1.0 */
static void decode_org_csapi_cc_TpCallAdditionalTreatmentInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/cc/TpCallPartyToChargeAdditionalInfo:1.0 */
static void decode_org_csapi_cc_TpCallPartyToChargeAdditionalInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/cc/gccs/TpCallAppInfo:1.0 */
static void decode_org_csapi_cc_gccs_TpCallAppInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/cc/gccs/TpCallAdditionalReportInfo:1.0 */
static void decode_org_csapi_cc_gccs_TpCallAdditionalReportInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/cc/gccs/TpCallAdditionalReportCriteria:1.0 */
static void decode_org_csapi_cc_gccs_TpCallAdditionalReportCriteria_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/cc/TpCallAdditionalEventInfo:1.0 */
static void decode_org_csapi_cc_TpCallAdditionalEventInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/cc/TpAdditionalCallEventCriteria:1.0 */
static void decode_org_csapi_cc_TpAdditionalCallEventCriteria_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/cc/TpCallAppInfo:1.0 */
static void decode_org_csapi_cc_TpCallAppInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/cc/mpccs/TpAppMultiPartyCallBack:1.0 */
static void decode_org_csapi_cc_mpccs_TpAppMultiPartyCallBack_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/cc/mmccs/TpMediaStreamDataTypeRequest:1.0 */
static void decode_org_csapi_cc_mmccs_TpMediaStreamDataTypeRequest_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/cc/mmccs/TpAppMultiMediaCallBack:1.0 */
static void decode_org_csapi_cc_mmccs_TpAppMultiMediaCallBack_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/cc/cccs/TpConfPolicy:1.0 */
static void decode_org_csapi_cc_cccs_TpConfPolicy_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/ui/TpUITargetObject:1.0 */
static void decode_org_csapi_ui_TpUITargetObject_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/mm/TpBindingEntry:1.0 */
static void decode_org_csapi_mm_TpBindingEntry_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/dsc/TpDataSessionChargeOrder:1.0 */
static void decode_org_csapi_dsc_TpDataSessionChargeOrder_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/dsc/TpDataSessionAdditionalErrorInfo:1.0 */
static void decode_org_csapi_dsc_TpDataSessionAdditionalErrorInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/dsc/TpDataSessionAdditionalReportInfo:1.0 */
static void decode_org_csapi_dsc_TpDataSessionAdditionalReportInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/gms/TpMessageInfoProperty:1.0 */
static void decode_org_csapi_gms_TpMessageInfoProperty_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/gms/TpMailboxInfoProperty:1.0 */
static void decode_org_csapi_gms_TpMailboxInfoProperty_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/gms/TpMessagingEventInfo:1.0 */
static void decode_org_csapi_gms_TpMessagingEventInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/gms/TpMessagingEventCriteria:1.0 */
static void decode_org_csapi_gms_TpMessagingEventCriteria_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/gms/TpFolderInfoProperty:1.0 */
static void decode_org_csapi_gms_TpFolderInfoProperty_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/cs/TpChargingParameterValue:1.0 */
static void decode_org_csapi_cs_TpChargingParameterValue_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/cs/TpAppInformation:1.0 */
static void decode_org_csapi_cs_TpAppInformation_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/policy/TpPolicyType:1.0 */
static void decode_org_csapi_policy_TpPolicyType_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/pam/TpPAMContextData:1.0 */
static void decode_org_csapi_pam_TpPAMContextData_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/pam/TpPAMEventInfo:1.0 */
static void decode_org_csapi_pam_TpPAMEventInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/pam/TpPAMNotificationInfo:1.0 */
static void decode_org_csapi_pam_TpPAMNotificationInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/pam/TpPAMPreferenceData:1.0 */
static void decode_org_csapi_pam_TpPAMPreferenceData_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/mmm/TpFolderInfoProperty:1.0 */
static void decode_org_csapi_mmm_TpFolderInfoProperty_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/mmm/TpMailboxInfoProperty:1.0 */
static void decode_org_csapi_mmm_TpMailboxInfoProperty_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/mmm/TpMessageInfoProperty:1.0 */
static void decode_org_csapi_mmm_TpMessageInfoProperty_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/mmm/TpMessageHeaderField:1.0 */
static void decode_org_csapi_mmm_TpMessageHeaderField_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/mmm/TpDeliveryTime:1.0 */
static void decode_org_csapi_mmm_TpDeliveryTime_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/mmm/TpMessageTreatment:1.0 */
static void decode_org_csapi_mmm_TpMessageTreatment_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/mmm/TpMessagingEventCriteria:1.0 */
static void decode_org_csapi_mmm_TpMessagingEventCriteria_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);


/* Union = IDL:org/csapi/mmm/TpMessagingEventInfo:1.0 */
static void decode_org_csapi_mmm_TpMessagingEventInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_);

/* Union prototype declaration End */
/*
 * IDL Enums Start
 */

/*
 * Enum = IDL:org/csapi/TpAddressError:1.0
 */
static const value_string org_csapi_TpAddressError[] = {

    { 0, "P_ADDRESS_INVALID_UNDEFINED" },
    { 1, "P_ADDRESS_INVALID_MISSING" },
    { 2, "P_ADDRESS_INVALID_MISSING_ELEMENT" },
    { 3, "P_ADDRESS_INVALID_OUT_OF_RANGE" },
    { 4, "P_ADDRESS_INVALID_INCOMPLETE" },
    { 5, "P_ADDRESS_INVALID_CANNOT_DECODE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/TpAddressPlan:1.0
 */
static const value_string org_csapi_TpAddressPlan[] = {

    { 0, "P_ADDRESS_PLAN_NOT_PRESENT" },
    { 1, "P_ADDRESS_PLAN_UNDEFINED" },
    { 2, "P_ADDRESS_PLAN_IP" },
    { 3, "P_ADDRESS_PLAN_MULTICAST" },
    { 4, "P_ADDRESS_PLAN_UNICAST" },
    { 5, "P_ADDRESS_PLAN_E164" },
    { 6, "P_ADDRESS_PLAN_AESA" },
    { 7, "P_ADDRESS_PLAN_URL" },
    { 8, "P_ADDRESS_PLAN_NSAP" },
    { 9, "P_ADDRESS_PLAN_SMTP" },
    { 10, "P_ADDRESS_PLAN_MSMAIL" },
    { 11, "P_ADDRESS_PLAN_X400" },
    { 12, "P_ADDRESS_PLAN_SIP" },
    { 13, "P_ADDRESS_PLAN_ANY" },
    { 14, "P_ADDRESS_PLAN_NATIONAL" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/TpAddressPresentation:1.0
 */
static const value_string org_csapi_TpAddressPresentation[] = {

    { 0, "P_ADDRESS_PRESENTATION_UNDEFINED" },
    { 1, "P_ADDRESS_PRESENTATION_ALLOWED" },
    { 2, "P_ADDRESS_PRESENTATION_RESTRICTED" },
    { 3, "P_ADDRESS_PRESENTATION_ADDRESS_NOT_AVAILABLE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/TpAddressScreening:1.0
 */
static const value_string org_csapi_TpAddressScreening[] = {

    { 0, "P_ADDRESS_SCREENING_UNDEFINED" },
    { 1, "P_ADDRESS_SCREENING_USER_VERIFIED_PASSED" },
    { 2, "P_ADDRESS_SCREENING_USER_NOT_VERIFIED" },
    { 3, "P_ADDRESS_SCREENING_USER_VERIFIED_FAILED" },
    { 4, "P_ADDRESS_SCREENING_NETWORK" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/TpCallAoCOrderCategory:1.0
 */
static const value_string org_csapi_TpCallAoCOrderCategory[] = {

    { 0, "P_CHARGE_ADVICE_INFO" },
    { 1, "P_CHARGE_PER_TIME" },
    { 2, "P_CHARGE_NETWORK" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/TpDataSessionQosClass:1.0
 */
static const value_string org_csapi_TpDataSessionQosClass[] = {

    { 0, "P_DATA_SESSION_QOS_CLASS_CONVERSATIONAL" },
    { 1, "P_DATA_SESSION_QOS_CLASS_STREAMING" },
    { 2, "P_DATA_SESSION_QOS_CLASS_INTERACTIVE" },
    { 3, "P_DATA_SESSION_QOS_CLASS_BACKGROUND" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/TpSimpleAttributeTypeInfo:1.0
 */
static const value_string org_csapi_TpSimpleAttributeTypeInfo[] = {

    { 0, "P_BOOLEAN" },
    { 1, "P_OCTET" },
    { 2, "P_CHAR" },
    { 3, "P_WCHAR" },
    { 4, "P_STRING" },
    { 5, "P_WSTRING" },
    { 6, "P_INT16" },
    { 7, "P_UNSIGNED_INT16" },
    { 8, "P_INT32" },
    { 9, "P_UNSIGNED_INT32" },
    { 10, "P_INT64" },
    { 11, "P_UNSIGNED_INT64" },
    { 12, "P_FLOAT" },
    { 13, "P_DOUBLE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/TpAttributeTagInfo:1.0
 */
static const value_string org_csapi_TpAttributeTagInfo[] = {

    { 0, "P_SIMPLE_TYPE" },
    { 1, "P_STRUCTURED_TYPE" },
    { 2, "P_XML_TYPE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/fw/TpAPIUnavailReason:1.0
 */
static const value_string org_csapi_fw_TpAPIUnavailReason[] = {

    { 0, "API_UNAVAILABLE_UNDEFINED" },
    { 1, "API_UNAVAILABLE_LOCAL_FAILURE" },
    { 2, "API_UNAVAILABLE_GATEWAY_FAILURE" },
    { 3, "API_UNAVAILABLE_OVERLOADED" },
    { 4, "API_UNAVAILABLE_CLOSED" },
    { 5, "API_UNAVAILABLE_PROTOCOL_FAILURE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/fw/TpDomainIDType:1.0
 */
static const value_string org_csapi_fw_TpDomainIDType[] = {

    { 0, "P_FW" },
    { 1, "P_CLIENT_APPLICATION" },
    { 2, "P_ENT_OP" },
    { 3, "P_SERVICE_INSTANCE" },
    { 4, "P_SERVICE_SUPPLIER" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/fw/TpFwEventName:1.0
 */
static const value_string org_csapi_fw_TpFwEventName[] = {

    { 0, "P_EVENT_FW_NAME_UNDEFINED" },
    { 1, "P_EVENT_FW_SERVICE_AVAILABLE" },
    { 2, "P_EVENT_FW_SERVICE_UNAVAILABLE" },
    { 3, "P_EVENT_FW_MIGRATION_SERVICE_AVAILABLE" },
    { 4, "P_EVENT_FW_APP_SESSION_CREATED" },
    { 5, "P_EVENT_FW_APP_SESSION_TERMINATED" },
    { 6, "P_EVENT_FW_APP_AGREEMENT_SIGNED" },
    { 7, "P_EVENT_FW_APP_AGREEMENT_ENDED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/fw/TpFWExceptionType:1.0
 */
static const value_string org_csapi_fw_TpFWExceptionType[] = {

    { 0, "P_FW_DUMMY" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/fw/TpFwUnavailReason:1.0
 */
static const value_string org_csapi_fw_TpFwUnavailReason[] = {

    { 0, "FW_UNAVAILABLE_UNDEFINED" },
    { 1, "FW_UNAVAILABLE_LOCAL_FAILURE" },
    { 2, "FW_UNAVAILABLE_GATEWAY_FAILURE" },
    { 3, "FW_UNAVAILABLE_OVERLOADED" },
    { 4, "FW_UNAVAILABLE_CLOSED" },
    { 5, "FW_UNAVAILABLE_PROTOCOL_FAILURE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/fw/TpInterfaceFault:1.0
 */
static const value_string org_csapi_fw_TpInterfaceFault[] = {

    { 0, "INTERFACE_FAULT_UNDEFINED" },
    { 1, "INTERFACE_FAULT_LOCAL_FAILURE" },
    { 2, "INTERFACE_FAULT_GATEWAY_FAILURE" },
    { 3, "INTERFACE_FAULT_PROTOCOL_ERROR" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/fw/TpLoadLevel:1.0
 */
static const value_string org_csapi_fw_TpLoadLevel[] = {

    { 0, "LOAD_LEVEL_NORMAL" },
    { 1, "LOAD_LEVEL_OVERLOAD" },
    { 2, "LOAD_LEVEL_SEVERE_OVERLOAD" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/fw/TpLoadStatisticEntityType:1.0
 */
static const value_string org_csapi_fw_TpLoadStatisticEntityType[] = {

    { 0, "P_LOAD_STATISTICS_FW_TYPE" },
    { 1, "P_LOAD_STATISTICS_SVC_TYPE" },
    { 2, "P_LOAD_STATISTICS_APP_TYPE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/fw/TpLoadStatisticInfoType:1.0
 */
static const value_string org_csapi_fw_TpLoadStatisticInfoType[] = {

    { 0, "P_LOAD_STATISTICS_VALID" },
    { 1, "P_LOAD_STATISTICS_INVALID" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/fw/TpLoadStatusError:1.0
 */
static const value_string org_csapi_fw_TpLoadStatusError[] = {

    { 0, "LOAD_STATUS_ERROR_UNDEFINED" },
    { 1, "LOAD_STATUS_ERROR_UNAVAILABLE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/fw/TpServiceTypePropertyMode:1.0
 */
static const value_string org_csapi_fw_TpServiceTypePropertyMode[] = {

    { 0, "NORMAL" },
    { 1, "MANDATORY" },
    { 2, "READONLY" },
    { 3, "MANDATORY_READONLY" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/fw/TpSvcUnavailReason:1.0
 */
static const value_string org_csapi_fw_TpSvcUnavailReason[] = {

    { 0, "SERVICE_UNAVAILABLE_UNDEFINED" },
    { 1, "SERVICE_UNAVAILABLE_LOCAL_FAILURE" },
    { 2, "SERVICE_UNAVAILABLE_GATEWAY_FAILURE" },
    { 3, "SERVICE_UNAVAILABLE_OVERLOADED" },
    { 4, "SERVICE_UNAVAILABLE_CLOSED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/fw/TpLoadStatisticError:1.0
 */
static const value_string org_csapi_fw_TpLoadStatisticError[] = {

    { 0, "P_LOAD_INFO_ERROR_UNDEFINED" },
    { 1, "P_LOAD_INFO_UNAVAILABLE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/fw/TpSubjectType:1.0
 */
static const value_string org_csapi_fw_TpSubjectType[] = {

    { 0, "P_SUBJECT_UNDEFINED" },
    { 1, "P_SUBJECT_CLIENT_APP" },
    { 2, "P_SUBJECT_FW" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/fw/TpFaultStatisticsError:1.0
 */
static const value_string org_csapi_fw_TpFaultStatisticsError[] = {

    { 0, "P_FAULT_INFO_ERROR_UNDEFINED" },
    { 1, "P_FAULT_INFO_UNAVAILABLE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/fw/TpSvcAvailStatusReason:1.0
 */
static const value_string org_csapi_fw_TpSvcAvailStatusReason[] = {

    { 0, "SVC_UNAVAILABLE_UNDEFINED" },
    { 1, "SVC_UNAVAILABLE_LOCAL_FAILURE" },
    { 2, "SVC_UNAVAILABLE_GATEWAY_FAILURE" },
    { 3, "SVC_UNAVAILABLE_OVERLOADED" },
    { 4, "SVC_UNAVAILABLE_CLOSED" },
    { 5, "SVC_UNAVAILABLE_NO_RESPONSE" },
    { 6, "SVC_UNAVAILABLE_SW_UPGRADE" },
    { 7, "SVC_AVAILABLE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/fw/TpAppAvailStatusReason:1.0
 */
static const value_string org_csapi_fw_TpAppAvailStatusReason[] = {

    { 0, "APP_UNAVAILABLE_UNDEFINED" },
    { 1, "APP_UNAVAILABLE_LOCAL_FAILURE" },
    { 2, "APP_UNAVAILABLE_REMOTE_FAILURE" },
    { 3, "APP_UNAVAILABLE_OVERLOADED" },
    { 4, "APP_UNAVAILABLE_CLOSED" },
    { 5, "APP_UNAVAILABLE_NO_RESPONSE" },
    { 6, "APP_UNAVAILABLE_SW_UPGRADE" },
    { 7, "APP_AVAILABLE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/fw/TpMigrationAdditionalInfoType:1.0
 */
static const value_string org_csapi_fw_TpMigrationAdditionalInfoType[] = {

    { 0, "P_MIGRATION_INFO_UNDEFINED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/fw/TpFwAvailStatusReason:1.0
 */
static const value_string org_csapi_fw_TpFwAvailStatusReason[] = {

    { 0, "FRAMEWORK_UNAVAILABLE_UNDEFINED" },
    { 1, "FRAMEWORK_UNAVAILABLE_LOCAL_FAILURE" },
    { 2, "FRAMEWORK_UNAVAILABLE_REMOTE_FAILURE" },
    { 3, "FRAMEWORK_UNAVAILABLE_OVERLOADED" },
    { 4, "FRAMEWORK_UNAVAILABLE_CLOSED" },
    { 5, "FRAMEWORK_UNAVAILABLE_PROTOCOL_FAILURE" },
    { 6, "FRAMEWORK_UNAVAILABLE_SW_UPGRADE" },
    { 7, "FRAMEWORK_AVAILABLE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/ui/TpUIVariablePartType:1.0
 */
static const value_string org_csapi_ui_TpUIVariablePartType[] = {

    { 0, "P_UI_VARIABLE_PART_INT" },
    { 1, "P_UI_VARIABLE_PART_ADDRESS" },
    { 2, "P_UI_VARIABLE_PART_TIME" },
    { 3, "P_UI_VARIABLE_PART_DATE" },
    { 4, "P_UI_VARIABLE_PART_PRICE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/ui/TpUIReport:1.0
 */
static const value_string org_csapi_ui_TpUIReport[] = {

    { 0, "P_UI_REPORT_UNDEFINED" },
    { 1, "P_UI_REPORT_INFO_SENT" },
    { 2, "P_UI_REPORT_INFO_COLLECTED" },
    { 3, "P_UI_REPORT_NO_INPUT" },
    { 4, "P_UI_REPORT_TIMEOUT" },
    { 5, "P_UI_REPORT_MESSAGE_STORED" },
    { 6, "P_UI_REPORT_MESSAGE_NOT_STORED" },
    { 7, "P_UI_REPORT_MESSAGE_DELETED" },
    { 8, "P_UI_REPORT_MESSAGE_NOT_DELETED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/ui/TpUIInfoType:1.0
 */
static const value_string org_csapi_ui_TpUIInfoType[] = {

    { 0, "P_UI_INFO_ID" },
    { 1, "P_UI_INFO_DATA" },
    { 2, "P_UI_INFO_ADDRESS" },
    { 3, "P_UI_INFO_BIN_DATA" },
    { 4, "P_UI_INFO_UUENCODED" },
    { 5, "P_UI_INFO_MIME" },
    { 6, "P_UI_INFO_WAVE" },
    { 7, "P_UI_INFO_AU" },
    { 8, "P_UI_INFO_VXML" },
    { 9, "P_UI_INFO_SYNTHESIS" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/ui/TpUIFault:1.0
 */
static const value_string org_csapi_ui_TpUIFault[] = {

    { 0, "P_UI_FAULT_UNDEFINED" },
    { 1, "P_UI_CALL_ENDED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/ui/TpUIEventInfoDataType:1.0
 */
static const value_string org_csapi_ui_TpUIEventInfoDataType[] = {

    { 0, "P_UI_EVENT_DATA_TYPE_UNDEFINED" },
    { 1, "P_UI_EVENT_DATA_TYPE_UNSPECIFIED" },
    { 2, "P_UI_EVENT_DATA_TYPE_TEXT" },
    { 3, "P_UI_EVENT_DATA_TYPE_USSD_DATA" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/ui/TpUIError:1.0
 */
static const value_string org_csapi_ui_TpUIError[] = {

    { 0, "P_UI_ERROR_UNDEFINED" },
    { 1, "P_UI_ERROR_ILLEGAL_INFO" },
    { 2, "P_UI_ERROR_ID_NOT_FOUND" },
    { 3, "P_UI_ERROR_RESOURCE_UNAVAILABLE" },
    { 4, "P_UI_ERROR_ILLEGAL_RANGE" },
    { 5, "P_UI_ERROR_IMPROPER_USER_RESPONSE" },
    { 6, "P_UI_ERROR_ABANDON" },
    { 7, "P_UI_ERROR_NO_OPERATION_ACTIVE" },
    { 8, "P_UI_ERROR_NO_SPACE_AVAILABLE" },
    { 9, "P_UI_ERROR_RESOURCE_TIMEOUT" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/ui/TpUICollectMode:1.0
 */
static const value_string org_csapi_ui_TpUICollectMode[] = {

    { 0, "P_UI_COLLECT_MODE_DTMF" },
    { 1, "P_UI_COLLECT_MODE_VOICE" },
    { 2, "P_UI_COLLECT_MODE_DTMFANDVOICE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/ui/TpUIPronounceType:1.0
 */
static const value_string org_csapi_ui_TpUIPronounceType[] = {

    { 0, "P_UI_PRONOUNCE_IPA" },
    { 1, "P_UI_PRONOUNCE_SOUNDSLIKE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/ui/TpUISynthesisRange:1.0
 */
static const value_string org_csapi_ui_TpUISynthesisRange[] = {

    { 0, "P_UI_RANGE_CALMER" },
    { 1, "P_UI_RANGE_CALM" },
    { 2, "P_UI_RANGE_AVERAGE" },
    { 3, "P_UI_RANGE_EXCITED" },
    { 4, "P_UI_RANGE_MORE_EXCITED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/ui/TpUISynthesisRate:1.0
 */
static const value_string org_csapi_ui_TpUISynthesisRate[] = {

    { 0, "P_UI_RATE_SLOW" },
    { 1, "P_UI_RATE_AVERAGE" },
    { 2, "P_UI_RATE_FAST" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/ui/TpUISynthesisAge:1.0
 */
static const value_string org_csapi_ui_TpUISynthesisAge[] = {

    { 0, "P_UI_AGE_CHILD" },
    { 1, "P_UI_AGE_YOUNG_ADULT" },
    { 2, "P_UI_AGE_ADULT" },
    { 3, "P_UI_AGE_OLDER_ADULT" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/ui/TpUISynthesisGender:1.0
 */
static const value_string org_csapi_ui_TpUISynthesisGender[] = {

    { 0, "P_UI_GENDER_MALE" },
    { 1, "P_UI_GENDER_FEMALE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/TpReleaseCause:1.0
 */
static const value_string org_csapi_cc_TpReleaseCause[] = {

    { 0, "P_UNDEFINED" },
    { 1, "P_USER_NOT_AVAILABLE" },
    { 2, "P_BUSY" },
    { 3, "P_NO_ANSWER" },
    { 4, "P_NOT_REACHABLE" },
    { 5, "P_ROUTING_FAILURE" },
    { 6, "P_PREMATURE_DISCONNECT" },
    { 7, "P_DISCONNECTED" },
    { 8, "P_CALL_RESTRICTED" },
    { 9, "P_UNAVAILABLE_RESOURCE" },
    { 10, "P_GENERAL_FAILURE" },
    { 11, "P_TIMER_EXPIRY" },
    { 12, "P_UNSUPPORTED_MEDIA" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/TpCallMonitorMode:1.0
 */
static const value_string org_csapi_cc_TpCallMonitorMode[] = {

    { 0, "P_CALL_MONITOR_MODE_INTERRUPT" },
    { 1, "P_CALL_MONITOR_MODE_NOTIFY" },
    { 2, "P_CALL_MONITOR_MODE_DO_NOT_MONITOR" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/TpCallBearerService:1.0
 */
static const value_string org_csapi_cc_TpCallBearerService[] = {

    { 0, "P_CALL_BEARER_SERVICE_UNKNOWN" },
    { 1, "P_CALL_BEARER_SERVICE_SPEECH" },
    { 2, "P_CALL_BEARER_SERVICE_DIGITALUNRESTRICTED" },
    { 3, "P_CALL_BEARER_SERVICE_DIGITALRESTRICTED" },
    { 4, "P_CALL_BEARER_SERVICE_AUDIO" },
    { 5, "P_CALL_BEARER_SERVICE_DIGITALUNRESTRICTEDTONES" },
    { 6, "P_CALL_BEARER_SERVICE_VIDEO" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/TpCallChargeOrderCategory:1.0
 */
static const value_string org_csapi_cc_TpCallChargeOrderCategory[] = {

    { 0, "P_CALL_CHARGE_TRANSPARENT" },
    { 1, "P_CALL_CHARGE_PREDEFINED_SET" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/TpCallErrorType:1.0
 */
static const value_string org_csapi_cc_TpCallErrorType[] = {

    { 0, "P_CALL_ERROR_UNDEFINED" },
    { 1, "P_CALL_ERROR_INVALID_ADDRESS" },
    { 2, "P_CALL_ERROR_INVALID_STATE" },
    { 3, "P_CALL_ERROR_RESOURCE_UNAVAILABLE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/TpCallLoadControlMechanismType:1.0
 */
static const value_string org_csapi_cc_TpCallLoadControlMechanismType[] = {

    { 0, "P_CALL_LOAD_CONTROL_PER_INTERVAL" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/TpCallNetworkAccessType:1.0
 */
static const value_string org_csapi_cc_TpCallNetworkAccessType[] = {

    { 0, "P_CALL_NETWORK_ACCESS_TYPE_UNKNOWN" },
    { 1, "P_CALL_NETWORK_ACCESS_TYPE_POT" },
    { 2, "P_CALL_NETWORK_ACCESS_TYPE_ISDN" },
    { 3, "P_CALL_NETWORK_ACCESS_TYPE_DIALUPINTERNET" },
    { 4, "P_CALL_NETWORK_ACCESS_TYPE_XDSL" },
    { 5, "P_CALL_NETWORK_ACCESS_TYPE_WIRELESS" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/TpCallPartyCategory:1.0
 */
static const value_string org_csapi_cc_TpCallPartyCategory[] = {

    { 0, "P_CALL_PARTY_CATEGORY_UNKNOWN" },
    { 1, "P_CALL_PARTY_CATEGORY_OPERATOR_F" },
    { 2, "P_CALL_PARTY_CATEGORY_OPERATOR_E" },
    { 3, "P_CALL_PARTY_CATEGORY_OPERATOR_G" },
    { 4, "P_CALL_PARTY_CATEGORY_OPERATOR_R" },
    { 5, "P_CALL_PARTY_CATEGORY_OPERATOR_S" },
    { 6, "P_CALL_PARTY_CATEGORY_ORDINARY_SUB" },
    { 7, "P_CALL_PARTY_CATEGORY_PRIORITY_SUB" },
    { 8, "P_CALL_PARTY_CATEGORY_DATA_CALL" },
    { 9, "P_CALL_PARTY_CATEGORY_TEST_CALL" },
    { 10, "P_CALL_PARTY_CATEGORY_PAYPHONE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/TpCallServiceCodeType:1.0
 */
static const value_string org_csapi_cc_TpCallServiceCodeType[] = {

    { 0, "P_CALL_SERVICE_CODE_UNDEFINED" },
    { 1, "P_CALL_SERVICE_CODE_DIGITS" },
    { 2, "P_CALL_SERVICE_CODE_FACILITY" },
    { 3, "P_CALL_SERVICE_CODE_U2U" },
    { 4, "P_CALL_SERVICE_CODE_HOOKFLASH" },
    { 5, "P_CALL_SERVICE_CODE_RECALL" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/TpCallTeleService:1.0
 */
static const value_string org_csapi_cc_TpCallTeleService[] = {

    { 0, "P_CALL_TELE_SERVICE_UNKNOWN" },
    { 1, "P_CALL_TELE_SERVICE_TELEPHONY" },
    { 2, "P_CALL_TELE_SERVICE_FAX_2_3" },
    { 3, "P_CALL_TELE_SERVICE_FAX_4_I" },
    { 4, "P_CALL_TELE_SERVICE_FAX_4_II_III" },
    { 5, "P_CALL_TELE_SERVICE_VIDEOTEX_SYN" },
    { 6, "P_CALL_TELE_SERVICE_VIDEOTEX_INT" },
    { 7, "P_CALL_TELE_SERVICE_TELEX" },
    { 8, "P_CALL_TELE_SERVICE_MHS" },
    { 9, "P_CALL_TELE_SERVICE_OSI" },
    { 10, "P_CALL_TELE_SERVICE_FTAM" },
    { 11, "P_CALL_TELE_SERVICE_VIDEO" },
    { 12, "P_CALL_TELE_SERVICE_VIDEO_CONF" },
    { 13, "P_CALL_TELE_SERVICE_AUDIOGRAPH_CONF" },
    { 14, "P_CALL_TELE_SERVICE_MULTIMEDIA" },
    { 15, "P_CALL_TELE_SERVICE_CS_INI_H221" },
    { 16, "P_CALL_TELE_SERVICE_CS_SUB_H221" },
    { 17, "P_CALL_TELE_SERVICE_CS_INI_CALL" },
    { 18, "P_CALL_TELE_SERVICE_DATATRAFFIC" },
    { 19, "P_CALL_TELE_SERVICE_EMERGENCY_CALLS" },
    { 20, "P_CALL_TELE_SERVICE_SMS_MT_PP" },
    { 21, "P_CALL_TELE_SERVICE_SMS_MO_PP" },
    { 22, "P_CALL_TELE_SERVICE_CELL_BROADCAST" },
    { 23, "P_CALL_TELE_SERVICE_ALT_SPEECH_FAX_3" },
    { 24, "P_CALL_TELE_SERVICE_AUTOMATIC_FAX_3" },
    { 25, "P_CALL_TELE_SERVICE_VOICE_GROUP_CALL" },
    { 26, "P_CALL_TELE_SERVICE_VOICE_BROADCAST" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/TpCallTreatmentType:1.0
 */
static const value_string org_csapi_cc_TpCallTreatmentType[] = {

    { 0, "P_CALL_TREATMENT_DEFAULT" },
    { 1, "P_CALL_TREATMENT_RELEASE" },
    { 2, "P_CALL_TREATMENT_SIAR" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/TpCallPartyToChargeType:1.0
 */
static const value_string org_csapi_cc_TpCallPartyToChargeType[] = {

    { 0, "P_CALL_PARTY_ORIGINATING" },
    { 1, "P_CALL_PARTY_DESTINATION" },
    { 2, "P_CALL_PARTY_SPECIAL" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/gccs/TpCallNotificationType:1.0
 */
static const value_string org_csapi_cc_gccs_TpCallNotificationType[] = {

    { 0, "P_ORIGINATING" },
    { 1, "P_TERMINATING" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/gccs/TpCallAppInfoType:1.0
 */
static const value_string org_csapi_cc_gccs_TpCallAppInfoType[] = {

    { 0, "P_CALL_APP_UNDEFINED" },
    { 1, "P_CALL_APP_ALERTING_MECHANISM" },
    { 2, "P_CALL_APP_NETWORK_ACCESS_TYPE" },
    { 3, "P_CALL_APP_TELE_SERVICE" },
    { 4, "P_CALL_APP_BEARER_SERVICE" },
    { 5, "P_CALL_APP_PARTY_CATEGORY" },
    { 6, "P_CALL_APP_PRESENTATION_ADDRESS" },
    { 7, "P_CALL_APP_GENERIC_INFO" },
    { 8, "P_CALL_APP_ADDITIONAL_ADDRESS" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/gccs/TpCallReportType:1.0
 */
static const value_string org_csapi_cc_gccs_TpCallReportType[] = {

    { 0, "P_CALL_REPORT_UNDEFINED" },
    { 1, "P_CALL_REPORT_PROGRESS" },
    { 2, "P_CALL_REPORT_ALERTING" },
    { 3, "P_CALL_REPORT_ANSWER" },
    { 4, "P_CALL_REPORT_BUSY" },
    { 5, "P_CALL_REPORT_NO_ANSWER" },
    { 6, "P_CALL_REPORT_DISCONNECT" },
    { 7, "P_CALL_REPORT_REDIRECTED" },
    { 8, "P_CALL_REPORT_SERVICE_CODE" },
    { 9, "P_CALL_REPORT_ROUTING_FAILURE" },
    { 10, "P_CALL_REPORT_QUEUED" },
    { 11, "P_CALL_REPORT_NOT_REACHABLE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/gccs/TpCallFault:1.0
 */
static const value_string org_csapi_cc_gccs_TpCallFault[] = {

    { 0, "P_CALL_FAULT_UNDEFINED" },
    { 1, "P_CALL_TIMEOUT_ON_RELEASE" },
    { 2, "P_CALL_TIMEOUT_ON_INTERRUPT" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/TpCallAppInfoType:1.0
 */
static const value_string org_csapi_cc_TpCallAppInfoType[] = {

    { 0, "P_CALL_APP_UNDEFINED" },
    { 1, "P_CALL_APP_ALERTING_MECHANISM" },
    { 2, "P_CALL_APP_NETWORK_ACCESS_TYPE" },
    { 3, "P_CALL_APP_TELE_SERVICE" },
    { 4, "P_CALL_APP_BEARER_SERVICE" },
    { 5, "P_CALL_APP_PARTY_CATEGORY" },
    { 6, "P_CALL_APP_PRESENTATION_ADDRESS" },
    { 7, "P_CALL_APP_GENERIC_INFO" },
    { 8, "P_CALL_APP_ADDITIONAL_ADDRESS" },
    { 9, "P_CALL_APP_ORIGINAL_DESTINATION_ADDRESS" },
    { 10, "P_CALL_APP_REDIRECTING_ADDRESS" },
    { 11, "P_CALL_APP_HIGH_PROBABILITY_COMPLETION" },
    { 12, "P_CALL_APP_CARRIER" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/TpCallEventType:1.0
 */
static const value_string org_csapi_cc_TpCallEventType[] = {

    { 0, "P_CALL_EVENT_UNDEFINED" },
    { 1, "P_CALL_EVENT_ORIGINATING_CALL_ATTEMPT" },
    { 2, "P_CALL_EVENT_ORIGINATING_CALL_ATTEMPT_AUTHORISED" },
    { 3, "P_CALL_EVENT_ADDRESS_COLLECTED" },
    { 4, "P_CALL_EVENT_ADDRESS_ANALYSED" },
    { 5, "P_CALL_EVENT_ORIGINATING_SERVICE_CODE" },
    { 6, "P_CALL_EVENT_ORIGINATING_RELEASE" },
    { 7, "P_CALL_EVENT_TERMINATING_CALL_ATTEMPT" },
    { 8, "P_CALL_EVENT_TERMINATING_CALL_ATTEMPT_AUTHORISED" },
    { 9, "P_CALL_EVENT_ALERTING" },
    { 10, "P_CALL_EVENT_ANSWER" },
    { 11, "P_CALL_EVENT_TERMINATING_RELEASE" },
    { 12, "P_CALL_EVENT_REDIRECTED" },
    { 13, "P_CALL_EVENT_TERMINATING_SERVICE_CODE" },
    { 14, "P_CALL_EVENT_QUEUED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/TpCallLegAttachMechanism:1.0
 */
static const value_string org_csapi_cc_TpCallLegAttachMechanism[] = {

    { 0, "P_CALLLEG_ATTACH_IMPLICITLY" },
    { 1, "P_CALLLEG_ATTACH_EXPLICITLY" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/TpCarrierSelectionField:1.0
 */
static const value_string org_csapi_cc_TpCarrierSelectionField[] = {

    { 0, "P_CIC_UNDEFINED" },
    { 1, "P_CIC_NO_INPUT" },
    { 2, "P_CIC_INPUT" },
    { 3, "P_CIC_UNDETERMINED" },
    { 4, "P_CIC_NOT_PRESCRIBED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/mpccs/TpAppMultiPartyCallBackRefType:1.0
 */
static const value_string org_csapi_cc_mpccs_TpAppMultiPartyCallBackRefType[] = {

    { 0, "P_APP_CALLBACK_UNDEFINED" },
    { 1, "P_APP_MULTIPARTY_CALL_CALLBACK" },
    { 2, "P_APP_CALL_LEG_CALLBACK" },
    { 3, "P_APP_CALL_AND_CALL_LEG_CALLBACK" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/mmccs/TpMediaStreamDirection:1.0
 */
static const value_string org_csapi_cc_mmccs_TpMediaStreamDirection[] = {

    { 0, "P_SEND_ONLY" },
    { 1, "P_RECEIVE_ONLY" },
    { 2, "P_SEND_RECEIVE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/mmccs/TpMediaStreamEventType:1.0
 */
static const value_string org_csapi_cc_mmccs_TpMediaStreamEventType[] = {

    { 0, "P_MEDIA_STREAM_ADDED" },
    { 1, "P_MEDIA_STREAM_SUBTRACTED" },
    { 2, "P_MEDIA_STREAM_QOS_CLASS_CHANGED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/mmccs/TpMediaStreamDataTypeRequestType:1.0
 */
static const value_string org_csapi_cc_mmccs_TpMediaStreamDataTypeRequestType[] = {

    { 0, "P_AUDIO_CAPABILITIES" },
    { 1, "P_VIDEO_CAPABILITIES" },
    { 2, "P_DATA_CAPABILITIES" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/mmccs/TpAppMultiMediaCallBackRefType:1.0
 */
static const value_string org_csapi_cc_mmccs_TpAppMultiMediaCallBackRefType[] = {

    { 0, "P_APP_CALLBACK_UNDEFINED" },
    { 1, "P_APP_MULTIMEDIA_CALL_CALLBACK" },
    { 2, "P_APP_CALL_LEG_CALLBACK" },
    { 3, "P_APP_CALL_AND_CALL_LEG_CALLBACK" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/cccs/TpConfPolicyType:1.0
 */
static const value_string org_csapi_cc_cccs_TpConfPolicyType[] = {

    { 0, "P_CONFERENCE_POLICY_UNDEFINED" },
    { 1, "P_CONFERENCE_POLICY_MONOMEDIA" },
    { 2, "P_CONFERENCE_POLICY_MULTIMEDIA" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cc/cccs/TpVideoHandlingType:1.0
 */
static const value_string org_csapi_cc_cccs_TpVideoHandlingType[] = {

    { 0, "P_MIXED_VIDEO" },
    { 1, "P_SWITCHED_VIDEO_CHAIR_CONTROLLED" },
    { 2, "P_SWITCHED_VIDEO_VOICE_CONTROLLED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/ui/TpUITargetObjectType:1.0
 */
static const value_string org_csapi_ui_TpUITargetObjectType[] = {

    { 0, "P_UI_TARGET_OBJECT_CALL" },
    { 1, "P_UI_TARGET_OBJECT_MULTI_PARTY_CALL" },
    { 2, "P_UI_TARGET_OBJECT_CALL_LEG" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mm/TpLocationPriority:1.0
 */
static const value_string org_csapi_mm_TpLocationPriority[] = {

    { 0, "P_M_NORMAL" },
    { 1, "P_M_HIGH" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mm/TpLocationResponseIndicator:1.0
 */
static const value_string org_csapi_mm_TpLocationResponseIndicator[] = {

    { 0, "P_M_NO_DELAY" },
    { 1, "P_M_LOW_DELAY" },
    { 2, "P_M_DELAY_TOLERANT" },
    { 3, "P_M_USE_TIMER_VALUE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mm/TpLocationTriggerCriteria:1.0
 */
static const value_string org_csapi_mm_TpLocationTriggerCriteria[] = {

    { 0, "P_UL_ENTERING_AREA" },
    { 1, "P_UL_LEAVING_AREA" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mm/TpLocationType:1.0
 */
static const value_string org_csapi_mm_TpLocationType[] = {

    { 0, "P_M_CURRENT" },
    { 1, "P_M_CURRENT_OR_LAST_KNOWN" },
    { 2, "P_M_INITIAL" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mm/TpLocationUncertaintyShape:1.0
 */
static const value_string org_csapi_mm_TpLocationUncertaintyShape[] = {

    { 0, "P_M_SHAPE_NONE" },
    { 1, "P_M_SHAPE_CIRCLE" },
    { 2, "P_M_SHAPE_CIRCLE_SECTOR" },
    { 3, "P_M_SHAPE_CIRCLE_ARC_STRIPE" },
    { 4, "P_M_SHAPE_ELLIPSE" },
    { 5, "P_M_SHAPE_ELLIPSE_SECTOR" },
    { 6, "P_M_SHAPE_ELLIPSE_ARC_STRIPE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mm/TpMobilityDiagnostic:1.0
 */
static const value_string org_csapi_mm_TpMobilityDiagnostic[] = {

    { 0, "P_M_NO_INFORMATION" },
    { 1, "P_M_APPL_NOT_IN_PRIV_EXCEPT_LST" },
    { 2, "P_M_CALL_TO_USER_NOT_SETUP" },
    { 3, "P_M_PRIVACY_OVERRIDE_NOT_APPLIC" },
    { 4, "P_M_DISALL_BY_LOCAL_REGULAT_REQ" },
    { 5, "P_M_CONGESTION" },
    { 6, "P_M_INSUFFICIENT_RESOURCES" },
    { 7, "P_M_INSUFFICIENT_MEAS_DATA" },
    { 8, "P_M_INCONSISTENT_MEAS_DATA" },
    { 9, "P_M_LOC_PROC_NOT_COMPLETED" },
    { 10, "P_M_LOC_PROC_NOT_SUPP_BY_USER" },
    { 11, "P_M_QOS_NOT_ATTAINABLE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mm/TpMobilityError:1.0
 */
static const value_string org_csapi_mm_TpMobilityError[] = {

    { 0, "P_M_OK" },
    { 1, "P_M_SYSTEM_FAILURE" },
    { 2, "P_M_UNAUTHORIZED_NETWORK" },
    { 3, "P_M_UNAUTHORIZED_APPLICATION" },
    { 4, "P_M_UNKNOWN_SUBSCRIBER" },
    { 5, "P_M_ABSENT_SUBSCRIBER" },
    { 6, "P_M_POSITION_METHOD_FAILURE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mm/TpMobilityStopScope:1.0
 */
static const value_string org_csapi_mm_TpMobilityStopScope[] = {

    { 0, "P_M_ALL_IN_ASSIGNMENT" },
    { 1, "P_M_SPECIFIED_USERS" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mm/TpTerminalType:1.0
 */
static const value_string org_csapi_mm_TpTerminalType[] = {

    { 0, "P_M_FIXED" },
    { 1, "P_M_MOBILE" },
    { 2, "P_M_IP" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mm/TpUserLocationEmergencyTrigger:1.0
 */
static const value_string org_csapi_mm_TpUserLocationEmergencyTrigger[] = {

    { 0, "P_ULE_CALL_ORIGINATION" },
    { 1, "P_ULE_CALL_RELEASE" },
    { 2, "P_ULE_LOCATION_REQUEST" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mm/TpUserStatusIndicator:1.0
 */
static const value_string org_csapi_mm_TpUserStatusIndicator[] = {

    { 0, "P_US_REACHABLE" },
    { 1, "P_US_NOT_REACHABLE" },
    { 2, "P_US_BUSY" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mm/TpAuthStatusIndicator:1.0
 */
static const value_string org_csapi_mm_TpAuthStatusIndicator[] = {

    { 0, "P_AUTHENTICATED_FOR_NETWORK_ONLY" },
    { 1, "P_AUTHENTICATED_FOR_NETWORK_AND_IP_SERVICES" },
    { 2, "P_NOT_AUTHENTICATED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mm/TpAccessTechnology:1.0
 */
static const value_string org_csapi_mm_TpAccessTechnology[] = {

    { 0, "P_MOBILE_ACCESS_PS" },
    { 1, "P_MOBILE_ACCESS_CS" },
    { 2, "P_FIXED_ACCESS" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mm/TpRoamingStatus:1.0
 */
static const value_string org_csapi_mm_TpRoamingStatus[] = {

    { 0, "P_UNKNOWN_ROAMING_STATUS" },
    { 1, "P_HOME_USER" },
    { 2, "P_NATIONAL_ROAMING" },
    { 3, "P_INTERNATIONAL_ROAMING" },
    { 4, "P_OPERATOR_ROAMING" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mm/TpBindingEntryType:1.0
 */
static const value_string org_csapi_mm_TpBindingEntryType[] = {

    { 0, "P_UB_UNDEFINED" },
    { 1, "P_UB_EXPIRES" },
    { 2, "P_UB_CONTACT_ADDRESS" },
    { 3, "P_UB_PREFERENCE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/termcap/TpTerminalCapabilitiesError:1.0
 */
static const value_string org_csapi_termcap_TpTerminalCapabilitiesError[] = {

    { 0, "P_TERMCAP_ERROR_UNDEFINED" },
    { 1, "P_TERMCAP_INVALID_TERMINALID" },
    { 2, "P_TERMCAP_SYSTEM_FAILURE" },
    { 3, "P_TERMCAP_INFO_UNAVAILABLE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/termcap/TpTerminalCapabilityScopeType:1.0
 */
static const value_string org_csapi_termcap_TpTerminalCapabilityScopeType[] = {

    { 0, "P_TERMINAL_CAPABILITY_SCOPE_TYPE_UNDEFINED" },
    { 1, "P_TERMINAL_CAPABILITY_SCOPE_TYPE_CCPP" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/dsc/TpDataSessionChargeOrderCategory:1.0
 */
static const value_string org_csapi_dsc_TpDataSessionChargeOrderCategory[] = {

    { 0, "P_DATA_SESSION_CHARGE_PER_VOLUME" },
    { 1, "P_DATA_SESSION_CHARGE_NETWORK" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/dsc/TpDataSessionErrorType:1.0
 */
static const value_string org_csapi_dsc_TpDataSessionErrorType[] = {

    { 0, "P_DATA_SESSION_ERROR_UNDEFINED" },
    { 1, "P_DATA_SESSION_ERROR_INVALID_ADDRESS" },
    { 2, "P_DATA_SESSION_ERROR_INVALID_STATE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/dsc/TpDataSessionFault:1.0
 */
static const value_string org_csapi_dsc_TpDataSessionFault[] = {

    { 0, "P_DATA_SESSION_FAULT_UNDEFINED" },
    { 1, "P_DATA_SESSION_FAULT_USER_ABORTED" },
    { 2, "P_DATA_SESSION_TIMEOUT_ON_RELEASE" },
    { 3, "P_DATA_SESSION_TIMEOUT_ON_INTERRUPT" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/dsc/TpDataSessionMonitorMode:1.0
 */
static const value_string org_csapi_dsc_TpDataSessionMonitorMode[] = {

    { 0, "P_DATA_SESSION_MONITOR_MODE_INTERRUPT" },
    { 1, "P_DATA_SESSION_MONITOR_MODE_NOTIFY" },
    { 2, "P_DATA_SESSION_MONITOR_MODE_DO_NOT_MONITOR" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/dsc/TpDataSessionReportType:1.0
 */
static const value_string org_csapi_dsc_TpDataSessionReportType[] = {

    { 0, "P_DATA_SESSION_REPORT_UNDEFINED" },
    { 1, "P_DATA_SESSION_REPORT_CONNECTED" },
    { 2, "P_DATA_SESSION_REPORT_DISCONNECT" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/gms/TpMessagingFault:1.0
 */
static const value_string org_csapi_gms_TpMessagingFault[] = {

    { 0, "P_MESSAGING_FAULT_UNDEFINED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/gms/TpMessagingEventName:1.0
 */
static const value_string org_csapi_gms_TpMessagingEventName[] = {

    { 0, "P_EVENT_GMS_NAME_UNDEFINED" },
    { 1, "P_EVENT_GMS_NEW_MESSAGE_ARRIVED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/gms/TpMessageStatus:1.0
 */
static const value_string org_csapi_gms_TpMessageStatus[] = {

    { 0, "P_MESSAGING_MESSAGE_STATUS_READ_MESSAGE" },
    { 1, "P_MESSAGING_MESSAGE_STATUS_UNREAD_MESSAGE" },
    { 2, "P_MESSAGING_MESSAGE_STATUS_FORWARDED_MESSAGE" },
    { 3, "P_MESSAGING_MESSAGE_STATUS_REPLIED_TO_MESSAGE" },
    { 4, "P_MESSAGING_MESSAGE_STATUS_SAVED_OR_UNSENT_MESSAGE" },
    { 5, "P_MESSAGING_MESSAGE_STATUS_NOTIFICATION_THAT_A_MESSAGE_WAS_DELIVERED" },
    { 6, "P_MESSAGING_MESSAGE_STATUS_NOTIFICATION_THAT_A_MESSAGE_WAS_READ" },
    { 7, "P_MESSAGING_MESSAGE_STATUS_NOTIFICATION_THAT_A_MESSAGE_WAS_NOT_DELIVERED" },
    { 8, "P_MESSAGING_MESSAGE_STATUS_NOTIFICATION_THAT_A_MESSAGE_WAS_NOT_READ" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/gms/TpMessagePriority:1.0
 */
static const value_string org_csapi_gms_TpMessagePriority[] = {

    { 0, "P_MESSAGING_MESSAGE_PRIORITY_UNDEFINED" },
    { 1, "P_MESSAGING_MESSAGE_PRIORITY_HIGH" },
    { 2, "P_MESSAGING_MESSAGE_PRIORITY_LOW" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/gms/TpMessageInfoPropertyName:1.0
 */
static const value_string org_csapi_gms_TpMessageInfoPropertyName[] = {

    { 0, "P_MESSAGING_MESSAGE_UNDEFINED" },
    { 1, "P_MESSAGING_MESSAGE_ID" },
    { 2, "P_MESSAGING_MESSAGE_SUBJECT" },
    { 3, "P_MESSAGING_MESSAGE_DATE_SENT" },
    { 4, "P_MESSAGING_MESSAGE_DATE_RECEIVED" },
    { 5, "P_MESSAGING_MESSAGE_DATE_CHANGED" },
    { 6, "P_MESSAGING_MESSAGE_SENT_FROM" },
    { 7, "P_MESSAGING_MESSAGE_SENT_TO" },
    { 8, "P_MESSAGING_MESSAGE_CC_TO" },
    { 9, "P_MESSAGING_MESSAGE_BCC_TO" },
    { 10, "P_MESSAGING_MESSAGE_SIZE" },
    { 11, "P_MESSAGING_MESSAGE_PRIORITY" },
    { 12, "P_MESSAGING_MESSAGE_FORMAT" },
    { 13, "P_MESSAGING_MESSAGE_FOLDER" },
    { 14, "P_MESSAGING_MESSAGE_STATUS" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/gms/TpMessageFormat:1.0
 */
static const value_string org_csapi_gms_TpMessageFormat[] = {

    { 0, "P_MESSAGING_MESSAGE_FORMAT_UNDEFINED" },
    { 1, "P_MESSAGING_MESSAGE_FORMAT_TEXT" },
    { 2, "P_MESSAGING_MESSAGE_FORMAT_BINARY" },
    { 3, "P_MESSAGING_MESSAGE_FORMAT_UUENCODED" },
    { 4, "P_MESSAGING_MESSAGE_FORMAT_MIME" },
    { 5, "P_MESSAGING_MESSAGE_FORMAT_WAVE" },
    { 6, "P_MESSAGING_MESSAGE_FORMAT_AU" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/gms/TpMailboxInfoPropertyName:1.0
 */
static const value_string org_csapi_gms_TpMailboxInfoPropertyName[] = {

    { 0, "P_MESSAGING_MAILBOX_UNDEFINED" },
    { 1, "P_MESSAGING_MAILBOX_ID" },
    { 2, "P_MESSAGING_MAILBOX_OWNER" },
    { 3, "P_MESSAGING_MAILBOX_FOLDER" },
    { 4, "P_MESSAGING_MAILBOX_DATE_CREATED" },
    { 5, "P_MESSAGING_MAILBOX_DATE_CHANGED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/gms/TpFolderInfoPropertyName:1.0
 */
static const value_string org_csapi_gms_TpFolderInfoPropertyName[] = {

    { 0, "P_MESSAGING_FOLDER_UNDEFINED" },
    { 1, "P_MESSAGING_FOLDER_ID" },
    { 2, "P_MESSAGING_FOLDER_MESSAGE" },
    { 3, "P_MESSAGING_FOLDER_SUBFOLDER" },
    { 4, "P_MESSAGING_FOLDER_DATE_CREATED" },
    { 5, "P_MESSAGING_FOLDER_DATE_CHANGED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cm/TpTrafficDirection:1.0
 */
static const value_string org_csapi_cm_TpTrafficDirection[] = {

    { 0, "UNIDIRECTIONAL" },
    { 1, "BIDIRECTIONAL" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cm/TpTagValue:1.0
 */
static const value_string org_csapi_cm_TpTagValue[] = {

    { 0, "PROVIDER_SPECIFIED" },
    { 1, "OPERATOR_SPECIFIED" },
    { 2, "UNSPECIFIED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cm/TpSiteOrSap:1.0
 */
static const value_string org_csapi_cm_TpSiteOrSap[] = {

    { 0, "SITE" },
    { 1, "SAP" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cm/TpIPVersion:1.0
 */
static const value_string org_csapi_cm_TpIPVersion[] = {

    { 0, "VERSION_UNKNOWN" },
    { 1, "VERSION_IPV4" },
    { 2, "VERSION_IPV6" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cm/TpIPv4AddType:1.0
 */
static const value_string org_csapi_cm_TpIPv4AddType[] = {

    { 0, "IPV4_ADD_CLASS_A" },
    { 1, "IPV4_ADD_CLASS_B" },
    { 2, "IPV4_ADD_CLASS_C" },
    { 3, "IPV4_ADD_CLASS_D" },
    { 4, "IPV4_ADD_CLASS_E" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cm/TpAction:1.0
 */
static const value_string org_csapi_cm_TpAction[] = {

    { 0, "DROP" },
    { 1, "TRANSMIT" },
    { 2, "RESHAPE" },
    { 3, "REMARK" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cm/TpVprpStatus:1.0
 */
static const value_string org_csapi_cm_TpVprpStatus[] = {

    { 0, "ACTIVE" },
    { 1, "PENDING" },
    { 2, "DISALLOWED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/am/TpChargingEventName:1.0
 */
static const value_string org_csapi_am_TpChargingEventName[] = {

    { 0, "P_AM_CHARGING" },
    { 1, "P_AM_RECHARGING" },
    { 2, "P_AM_ACCOUNT_LOW" },
    { 3, "P_AM_ACCOUNT_ZERO" },
    { 4, "P_AM_ACCOUNT_DISABLED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/am/TpBalanceQueryError:1.0
 */
static const value_string org_csapi_am_TpBalanceQueryError[] = {

    { 0, "P_BALANCE_QUERY_OK" },
    { 1, "P_BALANCE_QUERY_ERROR_UNDEFINED" },
    { 2, "P_BALANCE_QUERY_UNKNOWN_SUBSCRIBER" },
    { 3, "P_BALANCE_QUERY_UNAUTHORIZED_APPLICATION" },
    { 4, "P_BALANCE_QUERY_SYSTEM_FAILURE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/am/TpTransactionHistoryStatus:1.0
 */
static const value_string org_csapi_am_TpTransactionHistoryStatus[] = {

    { 0, "P_AM_TRANSACTION_ERROR_UNSPECIFIED" },
    { 1, "P_AM_TRANSACTION_INVALID_INTERVAL" },
    { 2, "P_AM_TRANSACTION_UNKNOWN_ACCOUNT" },
    { 3, "P_AM_TRANSACTION_UNAUTHORIZED_APPLICATION" },
    { 4, "P_AM_TRANSACTION_PROCESSING_ERROR" },
    { 5, "P_AM_TRANSACTION_SYSTEM_FAILURE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/am/TpVoucherError:1.0
 */
static const value_string org_csapi_am_TpVoucherError[] = {

    { 0, "P_VOUCHER_OK" },
    { 1, "P_VOUCHER_UNDEFINED" },
    { 2, "P_VOUCHER_UNKNOWN_SUBSCRIBER" },
    { 3, "P_VOUCHER_UNAUTHORIZED_APPLICATION" },
    { 4, "P_VOUCHER_SYSTEM_FAILURE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cs/TpChargingParameterValueType:1.0
 */
static const value_string org_csapi_cs_TpChargingParameterValueType[] = {

    { 0, "P_CHS_PARAMETER_INT32" },
    { 1, "P_CHS_PARAMETER_FLOAT" },
    { 2, "P_CHS_PARAMETER_STRING" },
    { 3, "P_CHS_PARAMETER_BOOLEAN" },
    { 4, "P_CHS_PARAMETER_OCTETSET" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cs/TpChargingError:1.0
 */
static const value_string org_csapi_cs_TpChargingError[] = {

    { 0, "P_CHS_ERR_UNDEFINED" },
    { 1, "P_CHS_ERR_ACCOUNT" },
    { 2, "P_CHS_ERR_USER" },
    { 3, "P_CHS_ERR_PARAMETER" },
    { 4, "P_CHS_ERR_NO_DEBIT" },
    { 5, "P_CHS_ERR_NO_CREDIT" },
    { 6, "P_CHS_ERR_VOLUMES" },
    { 7, "P_CHS_ERR_CURRENCY" },
    { 8, "P_CHS_ERR_NO_EXTEND" },
    { 9, "P_CHS_ERR_RESERVATION_LIMIT" },
    { 10, "P_CHS_ERR_CONFIRMATION_REQUIRED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cs/TpSessionEndedCause:1.0
 */
static const value_string org_csapi_cs_TpSessionEndedCause[] = {

    { 0, "P_CHS_CAUSE_UNDEFINED" },
    { 1, "P_CHS_CAUSE_TIMER_EXPIRED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/cs/TpAppInformationType:1.0
 */
static const value_string org_csapi_cs_TpAppInformationType[] = {

    { 0, "P_APP_INF_TIMESTAMP" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/policy/TpPolicyConditionListType:1.0
 */
static const value_string org_csapi_policy_TpPolicyConditionListType[] = {

    { 0, "P_PM_DNF" },
    { 1, "P_PM_CNF" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/policy/TpPolicyConditionType:1.0
 */
static const value_string org_csapi_policy_TpPolicyConditionType[] = {

    { 0, "P_PM_TIME_PERIOD_CONDITION" },
    { 1, "P_PM_EVENT_CONDITION" },
    { 2, "P_PM_EXPRESSION_CONDITION" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/policy/TpPolicyActionType:1.0
 */
static const value_string org_csapi_policy_TpPolicyActionType[] = {

    { 0, "P_PM_EVENT_ACTION" },
    { 1, "P_PM_EXPRESSION_ACTION" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/policy/TpPolicyTypeInfo:1.0
 */
static const value_string org_csapi_policy_TpPolicyTypeInfo[] = {

    { 0, "P_PM_SIMPLE_TYPE" },
    { 1, "P_PM_TYPE_RECORD" },
    { 2, "P_PM_TYPE_LIST" },
    { 3, "P_PM_STRUCTURED_TYPE" },
    { 4, "P_PM_XML_TYPE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/pam/TpPAMContextName:1.0
 */
static const value_string org_csapi_pam_TpPAMContextName[] = {

    { 0, "PAM_CONTEXT_ANY" },
    { 1, "PAM_CONTEXT_COMMUNICATION" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/pam/TpPAMACLDefault:1.0
 */
static const value_string org_csapi_pam_TpPAMACLDefault[] = {

    { 0, "PAM_ACCESS_ALLOW" },
    { 1, "PAM_ACCESS_DENY" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/pam/TpPAMPreferenceType:1.0
 */
static const value_string org_csapi_pam_TpPAMPreferenceType[] = {

    { 0, "PAM_ACCESS_LIST" },
    { 1, "PAM_EXTERNAL_CONTROL" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/pam/TpPAMEventName:1.0
 */
static const value_string org_csapi_pam_TpPAMEventName[] = {

    { 0, "PAM_CE_IDENTITY_PRESENCE_SET" },
    { 1, "PAM_CE_AVAILABILITY_CHANGED" },
    { 2, "PAM_CE_WATCHERS_CHANGED" },
    { 3, "PAM_CE_IDENTITY_CREATED" },
    { 4, "PAM_CE_IDENTITY_DELETED" },
    { 5, "PAM_CE_GROUP_MEMBERSHIP_CHANGED" },
    { 6, "PAM_CE_AGENT_CREATED" },
    { 7, "PAM_CE_AGENT_DELETED" },
    { 8, "PAM_CE_AGENT_ASSIGNED" },
    { 9, "PAM_CE_AGENT_UNASSIGNED" },
    { 10, "PAM_CE_CAPABILITY_CHANGED" },
    { 11, "PAM_CE_AGENT_CAPABILITY_PRESENCE_SET" },
    { 12, "PAM_CE_AGENT_PRESENCE_SET" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/pam/TpPAMwatcherChangeType:1.0
 */
static const value_string org_csapi_pam_TpPAMwatcherChangeType[] = {

    { 0, "PAM_WATCHERS_PERIODIC" },
    { 1, "PAM_WATCHERS_ADDED" },
    { 2, "PAM_WATCHERS_DELETED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/pam/TpPAMErrorCause:1.0
 */
static const value_string org_csapi_pam_TpPAMErrorCause[] = {

    { 0, "P_PAM_CAUSE_UNDEFINED" },
    { 1, "P_PAM_CAUSE_INVALID_ADDRESS" },
    { 2, "P_PAM_CAUSE_SYSTEM_FAILURE" },
    { 3, "P_PAM_CAUSE_INFO_UNAVAILABLE" },
    { 4, "P_PAM_CAUSE_EVENT_REGISTRATION_CANCELLED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mmm/TpFolderInfoPropertyName:1.0
 */
static const value_string org_csapi_mmm_TpFolderInfoPropertyName[] = {

    { 0, "P_MMM_FOLDER_UNDEFINED" },
    { 1, "P_MMM_FOLDER_DATE_CREATED" },
    { 2, "P_MMM_FOLDER_DATE_CHANGED" },
    { 3, "P_MMM_FOLDER_SIZE" },
    { 4, "P_MMM_FOLDER_NUMBER_OF_MESSAGES" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mmm/TpMailboxInfoPropertyName:1.0
 */
static const value_string org_csapi_mmm_TpMailboxInfoPropertyName[] = {

    { 0, "P_MMM_MAILBOX_UNDEFINED" },
    { 1, "P_MMM_MAILBOX_OWNER" },
    { 2, "P_MMM_MAILBOX_DATE_CREATED" },
    { 3, "P_MMM_MAILBOX_DATE_CHANGED" },
    { 4, "P_MMM_MAILBOX_SIZE" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mmm/TpMessageInfoPropertyName:1.0
 */
static const value_string org_csapi_mmm_TpMessageInfoPropertyName[] = {

    { 0, "P_MMM_MESSAGE_UNDEFINED" },
    { 1, "P_MMM_MESSAGE_DATE_CREATED" },
    { 2, "P_MMM_MESSAGE_DATE_RECEIVED" },
    { 3, "P_MMM_MESSAGE_DATE_CHANGED" },
    { 4, "P_MMM_MESSAGE_SIZE" },
    { 5, "P_MMM_MESSAGE_STATUS" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mmm/TpMessagePriority:1.0
 */
static const value_string org_csapi_mmm_TpMessagePriority[] = {

    { 0, "P_MMM_MESSAGE_PRIORITY_UNDEFINED" },
    { 1, "P_MMM_MESSAGE_PRIORITY_HIGH" },
    { 2, "P_MMM_MESSAGE_PRIORITY_LOW" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mmm/TpMessageHeaderFieldType:1.0
 */
static const value_string org_csapi_mmm_TpMessageHeaderFieldType[] = {

    { 0, "P_MESSAGE_DATE_SENT" },
    { 1, "P_MESSAGE_SENT_FROM" },
    { 2, "P_MESSAGE_SENDER" },
    { 3, "P_MESSAGE_REPLY_TO" },
    { 4, "P_MESSAGE_SENT_TO" },
    { 5, "P_MESSAGE_CC_TO" },
    { 6, "P_MESSAGE_BCC_TO" },
    { 7, "P_MESSAGE_RFC822_MESSAGE_ID" },
    { 8, "P_MESSAGE_IN_REPLY_TO" },
    { 9, "P_MESSAGE_REFERENCES" },
    { 10, "P_MESSAGE_SUBJECT" },
    { 11, "P_MESSAGE_COMMENTS" },
    { 12, "P_MESSAGE_KEYWORDS" },
    { 13, "P_MESSAGE_TRACE_FIELD" },
    { 14, "P_MESSAGE_RESENT_FIELD" },
    { 15, "P_MESSAGE_MIME_VERSION" },
    { 16, "P_MESSAGE_MIME_CONTENT" },
    { 17, "P_MESSAGE_MIME_ENCODING" },
    { 18, "P_MESSAGE_MIME_ID" },
    { 19, "P_MESSAGE_MIME_DESCRIPTION" },
    { 20, "P_MESSAGE_MIME_DISPOSITION" },
    { 21, "P_MESSAGE_MIME_EXTENSION_FIELD" },
    { 22, "P_MESSAGE_EXTENSION_FIELD" },
    { 23, "P_MESSAGE_PRIORITY" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mmm/TpMailboxMessageStatus:1.0
 */
static const value_string org_csapi_mmm_TpMailboxMessageStatus[] = {

    { 0, "P_MMM_RECEIVED_MSG_STATUS_READ" },
    { 1, "P_MMM_RECEIVED_MSG_STATUS_UNREAD" },
    { 2, "P_MMM_RECEIVED_MSG_STATUS_FORWARDED" },
    { 3, "P_MMM_RECEIVED_MSG_STATUS_REPLIED_TO" },
    { 4, "P_MMM_DRAFT_MSG_STATUS_SAVED_OR_UNSENT" },
    { 5, "P_MMM_SENT_MSG_STATUS_SENT" },
    { 6, "P_MMM_SENT_MSG_STATUS_DELIVERED" },
    { 7, "P_MMM_SENT_MSG_STATUS_READ" },
    { 8, "P_MMM_SENT_MSG_STATUS_DELETED_UNREAD" },
    { 9, "P_MMM_SENT_MSG_STATUS_NOT_DELIVERABLE" },
    { 10, "P_MMM_SENT_MSG_STATUS_EXPIRED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mmm/TpSetPropertyError:1.0
 */
static const value_string org_csapi_mmm_TpSetPropertyError[] = {

    { 0, "P_MMM_PROPERTY_NOT_SET" },
    { 1, "P_MMM_PROPERTY_READONLY" },
    { 2, "P_MMM_PROPERTY_INSUFFICIENT_PRIVILEGE" },
    { 3, "P_MMM_PROPERTY_NAME_UNKNOWN" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mmm/TpMessagingEventName:1.0
 */
static const value_string org_csapi_mmm_TpMessagingEventName[] = {

    { 0, "P_EVENT_MSG_NAME_UNDEFINED" },
    { 1, "P_EVENT_MSG_NEW_MAILBOX_MESSAGE_ARRIVED" },
    { 2, "P_EVENT_MSG_NEW_MESSAGE_ARRIVED" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mmm/TpDeliveryTimeType:1.0
 */
static const value_string org_csapi_mmm_TpDeliveryTimeType[] = {

    { 0, "P_MMM_SEND_IMMEDIATE" },
    { 1, "P_MMM_DELIVERY_TIME" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mmm/TpMessageTreatmentType:1.0
 */
static const value_string org_csapi_mmm_TpMessageTreatmentType[] = {

    { 0, "P_MMM_TREATMENT_UNDEFINED" },
    { 1, "P_MMM_TREATMENT_REPORT_REQUESTED" },
    { 2, "P_MMM_TREATMENT_BILLING_ID" },
    { 3, "P_MMM_TREATMENT_DELIVERY_TIME" },
    { 4, "P_MMM_TREATMENT_VALIDITY_TIME" },
    { 0,       NULL },
};

/*
 * Enum = IDL:org/csapi/mmm/TpMessagingError:1.0
 */
static const value_string org_csapi_mmm_TpMessagingError[] = {

    { 0, "P_MMM_ERROR_UNDEFINED" },
    { 1, "P_MMM_ERROR_INVALID_AUTHENTICATION_INFORMATION" },
    { 2, "P_MMM_ERROR_INVALID_MAILBOX" },
    { 3, "P_MMM_ERROR_INVALID_DELIVERY_TYPE" },
    { 4, "P_MMM_ERROR_MAX_MESSAGE_SIZE_EXCEEDED" },
    { 5, "P_MMM_ERROR_INVALID_FOLDER_ID" },
    { 6, "P_MMM_ERROR_INVALID_MESSAGE_ID" },
    { 7, "P_MMM_ERROR_INVALID_PART_ID" },
    { 8, "P_MMM_ERROR_DELIVERY_TYPE_ADDRESS_TYPE_MISMATCH" },
    { 9, "P_MMM_ERROR_DELIVERY_TYPE_MESSAGE_TYPE_MISMATCH" },
    { 10, "P_MMM_ERROR_INVALID_DELIVERY_TIME" },
    { 11, "P_MMM_ERROR_INVALID_VALIDITY_TIME" },
    { 12, "P_MMM_ERROR_MAX_SUBJECT_SIZE_EXCEEDED" },
    { 13, "P_MMM_ERROR_INVALID_ID" },
    { 14, "P_MMM_ERROR_INVALID_NESTING_LEVEL" },
    { 15, "P_MMM_ERROR_INVALID_CRITERIA" },
    { 16, "P_MMM_ERROR_INFORMATION_NOT_AVAILABLE" },
    { 17, "P_MMM_ERROR_CANNOT_CANCEL" },
    { 18, "P_MMM_ERROR_INVALID_HEADER" },
    { 19, "P_MMM_INVALID_NETWORK_STATE" },
    { 20, "P_MMM_ERROR_RESOURCE_UNAVAILABLE" },
    { 21, "P_MMM_ERROR_RESOURCE_TIMEOUT" },
    { 0,       NULL },
};

/*
 * IDL Enums End
 */

/*  Begin Exception Helper Functions  */


/* Exception = IDL:org/csapi/TpCommonExceptions:1.0 */
static void
decode_ex_org_csapi_TpCommonExceptions(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    proto_tree_add_int(tree, hf_org_csapi_TpCommonExceptions_ExceptionType, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_TpCommonExceptions_ExtraInformation);

}

/* Exception = IDL:org/csapi/P_INVALID_INTERFACE_TYPE:1.0 */
static void
decode_ex_org_csapi_P_INVALID_INTERFACE_TYPE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_P_INVALID_INTERFACE_TYPE_ExtraInformation);

}

/* Exception = IDL:org/csapi/P_INVALID_SESSION_ID:1.0 */
static void
decode_ex_org_csapi_P_INVALID_SESSION_ID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_P_INVALID_SESSION_ID_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_INVALID_DOMAIN_ID:1.0 */
static void
decode_ex_org_csapi_fw_P_INVALID_DOMAIN_ID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_INVALID_DOMAIN_ID_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_INVALID_AUTH_TYPE:1.0 */
static void
decode_ex_org_csapi_fw_P_INVALID_AUTH_TYPE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_INVALID_AUTH_TYPE_ExtraInformation);

}

/* Exception = IDL:org/csapi/P_INVALID_VERSION:1.0 */
static void
decode_ex_org_csapi_P_INVALID_VERSION(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_P_INVALID_VERSION_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_ACCESS_DENIED:1.0 */
static void
decode_ex_org_csapi_fw_P_ACCESS_DENIED(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_ACCESS_DENIED_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_INVALID_ACCESS_TYPE:1.0 */
static void
decode_ex_org_csapi_fw_P_INVALID_ACCESS_TYPE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_INVALID_ACCESS_TYPE_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_INVALID_SIGNING_ALGORITHM:1.0 */
static void
decode_ex_org_csapi_fw_P_INVALID_SIGNING_ALGORITHM(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_INVALID_SIGNING_ALGORITHM_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_INVALID_SIGNATURE:1.0 */
static void
decode_ex_org_csapi_fw_P_INVALID_SIGNATURE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_INVALID_SIGNATURE_ExtraInformation);

}

/* Exception = IDL:org/csapi/P_INVALID_INTERFACE_NAME:1.0 */
static void
decode_ex_org_csapi_P_INVALID_INTERFACE_NAME(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_P_INVALID_INTERFACE_NAME_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_INVALID_PROPERTY:1.0 */
static void
decode_ex_org_csapi_fw_P_INVALID_PROPERTY(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_INVALID_PROPERTY_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_NO_ACCEPTABLE_SIGNING_ALGORITHM:1.0 */
static void
decode_ex_org_csapi_fw_P_NO_ACCEPTABLE_SIGNING_ALGORITHM(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_NO_ACCEPTABLE_SIGNING_ALGORITHM_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_NO_ACCEPTABLE_ENCRYPTION_CAPABILITY:1.0 */
static void
decode_ex_org_csapi_fw_P_NO_ACCEPTABLE_ENCRYPTION_CAPABILITY(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_NO_ACCEPTABLE_ENCRYPTION_CAPABILITY_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_NO_ACCEPTABLE_AUTHENTICATION_MECHANISM:1.0 */
static void
decode_ex_org_csapi_fw_P_NO_ACCEPTABLE_AUTHENTICATION_MECHANISM(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_NO_ACCEPTABLE_AUTHENTICATION_MECHANISM_ExtraInformation);

}

/* Exception = IDL:org/csapi/P_INVALID_CRITERIA:1.0 */
static void
decode_ex_org_csapi_P_INVALID_CRITERIA(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_P_INVALID_CRITERIA_ExtraInformation);

}

/* Exception = IDL:org/csapi/P_INVALID_EVENT_TYPE:1.0 */
static void
decode_ex_org_csapi_P_INVALID_EVENT_TYPE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_P_INVALID_EVENT_TYPE_ExtraInformation);

}

/* Exception = IDL:org/csapi/P_INVALID_ASSIGNMENT_ID:1.0 */
static void
decode_ex_org_csapi_P_INVALID_ASSIGNMENT_ID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_P_INVALID_ASSIGNMENT_ID_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_INVALID_SERVICE_ID:1.0 */
static void
decode_ex_org_csapi_fw_P_INVALID_SERVICE_ID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_INVALID_SERVICE_ID_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_SERVICE_NOT_ENABLED:1.0 */
static void
decode_ex_org_csapi_fw_P_SERVICE_NOT_ENABLED(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_SERVICE_NOT_ENABLED_ExtraInformation);

}

/* Exception = IDL:org/csapi/P_UNAUTHORISED_PARAMETER_VALUE:1.0 */
static void
decode_ex_org_csapi_P_UNAUTHORISED_PARAMETER_VALUE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_P_UNAUTHORISED_PARAMETER_VALUE_ExtraInformation);

}

/* Exception = IDL:org/csapi/P_INVALID_TIME_AND_DATE_FORMAT:1.0 */
static void
decode_ex_org_csapi_P_INVALID_TIME_AND_DATE_FORMAT(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_P_INVALID_TIME_AND_DATE_FORMAT_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_INVALID_ACTIVITY_TEST_ID:1.0 */
static void
decode_ex_org_csapi_fw_P_INVALID_ACTIVITY_TEST_ID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_INVALID_ACTIVITY_TEST_ID_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_ILLEGAL_SERVICE_TYPE:1.0 */
static void
decode_ex_org_csapi_fw_P_ILLEGAL_SERVICE_TYPE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_ILLEGAL_SERVICE_TYPE_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_UNKNOWN_SERVICE_TYPE:1.0 */
static void
decode_ex_org_csapi_fw_P_UNKNOWN_SERVICE_TYPE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_UNKNOWN_SERVICE_TYPE_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_INVALID_AGREEMENT_TEXT:1.0 */
static void
decode_ex_org_csapi_fw_P_INVALID_AGREEMENT_TEXT(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_INVALID_AGREEMENT_TEXT_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_INVALID_SERVICE_TOKEN:1.0 */
static void
decode_ex_org_csapi_fw_P_INVALID_SERVICE_TOKEN(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_INVALID_SERVICE_TOKEN_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_SERVICE_ACCESS_DENIED:1.0 */
static void
decode_ex_org_csapi_fw_P_SERVICE_ACCESS_DENIED(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_SERVICE_ACCESS_DENIED_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_INVALID_SERVICE_PROFILE_ID:1.0 */
static void
decode_ex_org_csapi_fw_P_INVALID_SERVICE_PROFILE_ID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_INVALID_SERVICE_PROFILE_ID_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_INVALID_SAG_ID:1.0 */
static void
decode_ex_org_csapi_fw_P_INVALID_SAG_ID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_INVALID_SAG_ID_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_INVALID_SAG_TO_SERVICE_PROFILE_ASSIGNMENT:1.0 */
static void
decode_ex_org_csapi_fw_P_INVALID_SAG_TO_SERVICE_PROFILE_ASSIGNMENT(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_INVALID_SAG_TO_SERVICE_PROFILE_ASSIGNMENT_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_INVALID_SERVICE_CONTRACT_ID:1.0 */
static void
decode_ex_org_csapi_fw_P_INVALID_SERVICE_CONTRACT_ID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_INVALID_SERVICE_CONTRACT_ID_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_INVALID_CLIENT_APP_ID:1.0 */
static void
decode_ex_org_csapi_fw_P_INVALID_CLIENT_APP_ID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_INVALID_CLIENT_APP_ID_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_INVALID_ADDITION_TO_SAG:1.0 */
static void
decode_ex_org_csapi_fw_P_INVALID_ADDITION_TO_SAG(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_INVALID_ADDITION_TO_SAG_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_PROPERTY_TYPE_MISMATCH:1.0 */
static void
decode_ex_org_csapi_fw_P_PROPERTY_TYPE_MISMATCH(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_PROPERTY_TYPE_MISMATCH_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_DUPLICATE_PROPERTY_NAME:1.0 */
static void
decode_ex_org_csapi_fw_P_DUPLICATE_PROPERTY_NAME(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_DUPLICATE_PROPERTY_NAME_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_MISSING_MANDATORY_PROPERTY:1.0 */
static void
decode_ex_org_csapi_fw_P_MISSING_MANDATORY_PROPERTY(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_MISSING_MANDATORY_PROPERTY_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_SERVICE_TYPE_UNAVAILABLE:1.0 */
static void
decode_ex_org_csapi_fw_P_SERVICE_TYPE_UNAVAILABLE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_SERVICE_TYPE_UNAVAILABLE_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_ILLEGAL_SERVICE_ID:1.0 */
static void
decode_ex_org_csapi_fw_P_ILLEGAL_SERVICE_ID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_ILLEGAL_SERVICE_ID_ExtraInformation);

}

/* Exception = IDL:org/csapi/fw/P_UNKNOWN_SERVICE_ID:1.0 */
static void
decode_ex_org_csapi_fw_P_UNKNOWN_SERVICE_ID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_P_UNKNOWN_SERVICE_ID_ExtraInformation);

}

/* Exception = IDL:org/csapi/P_INVALID_ADDRESS:1.0 */
static void
decode_ex_org_csapi_P_INVALID_ADDRESS(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_P_INVALID_ADDRESS_ExtraInformation);

}

/* Exception = IDL:org/csapi/P_UNSUPPORTED_ADDRESS_PLAN:1.0 */
static void
decode_ex_org_csapi_P_UNSUPPORTED_ADDRESS_PLAN(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_P_UNSUPPORTED_ADDRESS_PLAN_ExtraInformation);

}

/* Exception = IDL:org/csapi/P_INVALID_NETWORK_STATE:1.0 */
static void
decode_ex_org_csapi_P_INVALID_NETWORK_STATE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_P_INVALID_NETWORK_STATE_ExtraInformation);

}

/* Exception = IDL:org/csapi/P_INVALID_CURRENCY:1.0 */
static void
decode_ex_org_csapi_P_INVALID_CURRENCY(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_P_INVALID_CURRENCY_ExtraInformation);

}

/* Exception = IDL:org/csapi/P_INVALID_AMOUNT:1.0 */
static void
decode_ex_org_csapi_P_INVALID_AMOUNT(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_P_INVALID_AMOUNT_ExtraInformation);

}

/* Exception = IDL:org/csapi/P_INFORMATION_NOT_AVAILABLE:1.0 */
static void
decode_ex_org_csapi_P_INFORMATION_NOT_AVAILABLE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_P_INFORMATION_NOT_AVAILABLE_ExtraInformation);

}

/* Exception = IDL:org/csapi/ui/P_ILLEGAL_ID:1.0 */
static void
decode_ex_org_csapi_ui_P_ILLEGAL_ID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_P_ILLEGAL_ID_ExtraInformation);

}

/* Exception = IDL:org/csapi/ui/P_ID_NOT_FOUND:1.0 */
static void
decode_ex_org_csapi_ui_P_ID_NOT_FOUND(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_P_ID_NOT_FOUND_ExtraInformation);

}

/* Exception = IDL:org/csapi/ui/P_ILLEGAL_RANGE:1.0 */
static void
decode_ex_org_csapi_ui_P_ILLEGAL_RANGE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_P_ILLEGAL_RANGE_ExtraInformation);

}

/* Exception = IDL:org/csapi/ui/P_INVALID_COLLECTION_CRITERIA:1.0 */
static void
decode_ex_org_csapi_ui_P_INVALID_COLLECTION_CRITERIA(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_P_INVALID_COLLECTION_CRITERIA_ExtraInformation);

}

/* Exception = IDL:org/csapi/P_APPLICATION_NOT_ACTIVATED:1.0 */
static void
decode_ex_org_csapi_P_APPLICATION_NOT_ACTIVATED(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_P_APPLICATION_NOT_ACTIVATED_ExtraInformation);

}

/* Exception = IDL:org/csapi/mm/P_REQUESTED_ACCURACY_CANNOT_BE_DELIVERED:1.0 */
static void
decode_ex_org_csapi_mm_P_REQUESTED_ACCURACY_CANNOT_BE_DELIVERED(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_P_REQUESTED_ACCURACY_CANNOT_BE_DELIVERED_ExtraInformation);

}

/* Exception = IDL:org/csapi/mm/P_REQUESTED_RESPONSE_TIME_CANNOT_BE_DELIVERED:1.0 */
static void
decode_ex_org_csapi_mm_P_REQUESTED_RESPONSE_TIME_CANNOT_BE_DELIVERED(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_P_REQUESTED_RESPONSE_TIME_CANNOT_BE_DELIVERED_ExtraInformation);

}

/* Exception = IDL:org/csapi/mm/P_INVALID_REPORTING_INTERVAL:1.0 */
static void
decode_ex_org_csapi_mm_P_INVALID_REPORTING_INTERVAL(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_P_INVALID_REPORTING_INTERVAL_ExtraInformation);

}

/* Exception = IDL:org/csapi/mm/P_TRIGGER_CONDITIONS_NOT_SUBSCRIBED:1.0 */
static void
decode_ex_org_csapi_mm_P_TRIGGER_CONDITIONS_NOT_SUBSCRIBED(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_P_TRIGGER_CONDITIONS_NOT_SUBSCRIBED_ExtraInformation);

}

/* Exception = IDL:org/csapi/P_UNKNOWN_SUBSCRIBER:1.0 */
static void
decode_ex_org_csapi_P_UNKNOWN_SUBSCRIBER(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_P_UNKNOWN_SUBSCRIBER_ExtraInformation);

}

/* Exception = IDL:org/csapi/termcap/P_INVALID_TERMINAL_ID:1.0 */
static void
decode_ex_org_csapi_termcap_P_INVALID_TERMINAL_ID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_termcap_P_INVALID_TERMINAL_ID_ExtraInformation);

}

/* Exception = IDL:org/csapi/gms/P_GMS_INVALID_MAILBOX:1.0 */
static void
decode_ex_org_csapi_gms_P_GMS_INVALID_MAILBOX(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_P_GMS_INVALID_MAILBOX_ExtraInformation);

}

/* Exception = IDL:org/csapi/gms/P_GMS_INVALID_AUTHENTICATION_INFORMATION:1.0 */
static void
decode_ex_org_csapi_gms_P_GMS_INVALID_AUTHENTICATION_INFORMATION(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_P_GMS_INVALID_AUTHENTICATION_INFORMATION_ExtraInformation);

}

/* Exception = IDL:org/csapi/gms/P_GMS_LOCKING_LOCKED_MAILBOX:1.0 */
static void
decode_ex_org_csapi_gms_P_GMS_LOCKING_LOCKED_MAILBOX(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_P_GMS_LOCKING_LOCKED_MAILBOX_ExtraInformation);

}

/* Exception = IDL:org/csapi/gms/P_GMS_UNLOCKING_UNLOCKED_MAILBOX:1.0 */
static void
decode_ex_org_csapi_gms_P_GMS_UNLOCKING_UNLOCKED_MAILBOX(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_P_GMS_UNLOCKING_UNLOCKED_MAILBOX_ExtraInformation);

}

/* Exception = IDL:org/csapi/gms/P_GMS_CANNOT_UNLOCK_MAILBOX:1.0 */
static void
decode_ex_org_csapi_gms_P_GMS_CANNOT_UNLOCK_MAILBOX(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_P_GMS_CANNOT_UNLOCK_MAILBOX_ExtraInformation);

}

/* Exception = IDL:org/csapi/gms/P_GMS_NUMBER_NOT_POSITIVE:1.0 */
static void
decode_ex_org_csapi_gms_P_GMS_NUMBER_NOT_POSITIVE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_P_GMS_NUMBER_NOT_POSITIVE_ExtraInformation);

}

/* Exception = IDL:org/csapi/gms/P_GMS_PROPERTY_NOT_SET:1.0 */
static void
decode_ex_org_csapi_gms_P_GMS_PROPERTY_NOT_SET(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_P_GMS_PROPERTY_NOT_SET_ExtraInformation);

}

/* Exception = IDL:org/csapi/gms/P_GMS_MAILBOX_LOCKED:1.0 */
static void
decode_ex_org_csapi_gms_P_GMS_MAILBOX_LOCKED(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_P_GMS_MAILBOX_LOCKED_ExtraInformation);

}

/* Exception = IDL:org/csapi/gms/P_GMS_FOLDER_IS_OPEN:1.0 */
static void
decode_ex_org_csapi_gms_P_GMS_FOLDER_IS_OPEN(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_P_GMS_FOLDER_IS_OPEN_ExtraInformation);

}

/* Exception = IDL:org/csapi/gms/P_GMS_INVALID_FOLDER_ID:1.0 */
static void
decode_ex_org_csapi_gms_P_GMS_INVALID_FOLDER_ID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_P_GMS_INVALID_FOLDER_ID_ExtraInformation);

}

/* Exception = IDL:org/csapi/gms/P_GMS_INSUFFICIENT_PRIVILEGE:1.0 */
static void
decode_ex_org_csapi_gms_P_GMS_INSUFFICIENT_PRIVILEGE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_P_GMS_INSUFFICIENT_PRIVILEGE_ExtraInformation);

}

/* Exception = IDL:org/csapi/gms/P_GMS_MAILBOX_OPEN:1.0 */
static void
decode_ex_org_csapi_gms_P_GMS_MAILBOX_OPEN(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_P_GMS_MAILBOX_OPEN_ExtraInformation);

}

/* Exception = IDL:org/csapi/gms/P_GMS_INVALID_MESSAGE_ID:1.0 */
static void
decode_ex_org_csapi_gms_P_GMS_INVALID_MESSAGE_ID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_P_GMS_INVALID_MESSAGE_ID_ExtraInformation);

}

/* Exception = IDL:org/csapi/gms/P_GMS_MESSAGE_NOT_REMOVED:1.0 */
static void
decode_ex_org_csapi_gms_P_GMS_MESSAGE_NOT_REMOVED(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_P_GMS_MESSAGE_NOT_REMOVED_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_UNKNOWN_VPRP_ID:1.0 */
static void
decode_ex_org_csapi_cm_P_UNKNOWN_VPRP_ID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_UNKNOWN_VPRP_ID_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_UNKNOWN_SLA_ID:1.0 */
static void
decode_ex_org_csapi_cm_P_UNKNOWN_SLA_ID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_UNKNOWN_SLA_ID_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_UNKNOWN_STATUS:1.0 */
static void
decode_ex_org_csapi_cm_P_UNKNOWN_STATUS(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_UNKNOWN_STATUS_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_UNKNOWN_QOS_INFO:1.0 */
static void
decode_ex_org_csapi_cm_P_UNKNOWN_QOS_INFO(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_UNKNOWN_QOS_INFO_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_UNKNOWN_VALIDITY_INFO:1.0 */
static void
decode_ex_org_csapi_cm_P_UNKNOWN_VALIDITY_INFO(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_UNKNOWN_VALIDITY_INFO_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_UNKNOWN_PIPEQOSINFO:1.0 */
static void
decode_ex_org_csapi_cm_P_UNKNOWN_PIPEQOSINFO(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_UNKNOWN_PIPEQOSINFO_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_UNKNOWN_DSCODEPOINT:1.0 */
static void
decode_ex_org_csapi_cm_P_UNKNOWN_DSCODEPOINT(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_UNKNOWN_DSCODEPOINT_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_UNKNOWN_VPRP:1.0 */
static void
decode_ex_org_csapi_cm_P_UNKNOWN_VPRP(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_UNKNOWN_VPRP_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_ILLEGAL_VPRPID:1.0 */
static void
decode_ex_org_csapi_cm_P_ILLEGAL_VPRPID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_ILLEGAL_VPRPID_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_ILLEGAL_REF_VALUE:1.0 */
static void
decode_ex_org_csapi_cm_P_ILLEGAL_REF_VALUE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_ILLEGAL_REF_VALUE_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_UNKNOWN_INTERFACE:1.0 */
static void
decode_ex_org_csapi_cm_P_UNKNOWN_INTERFACE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_UNKNOWN_INTERFACE_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_CANT_DELETE_VPRP:1.0 */
static void
decode_ex_org_csapi_cm_P_CANT_DELETE_VPRP(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_CANT_DELETE_VPRP_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_UNKNOWN_TEMPLATES:1.0 */
static void
decode_ex_org_csapi_cm_P_UNKNOWN_TEMPLATES(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_UNKNOWN_TEMPLATES_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_UNKNOWN_MENU:1.0 */
static void
decode_ex_org_csapi_cm_P_UNKNOWN_MENU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_UNKNOWN_MENU_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_UNKNOWN_ENTERPRISE_NETWORK:1.0 */
static void
decode_ex_org_csapi_cm_P_UNKNOWN_ENTERPRISE_NETWORK(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_UNKNOWN_ENTERPRISE_NETWORK_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_UNKNOWN_TEMPLATE_TYPE:1.0 */
static void
decode_ex_org_csapi_cm_P_UNKNOWN_TEMPLATE_TYPE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_UNKNOWN_TEMPLATE_TYPE_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_UNKNOWN_DESCRIPTION:1.0 */
static void
decode_ex_org_csapi_cm_P_UNKNOWN_DESCRIPTION(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_UNKNOWN_DESCRIPTION_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_ILLEGAL_SLA_ID:1.0 */
static void
decode_ex_org_csapi_cm_P_ILLEGAL_SLA_ID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_ILLEGAL_SLA_ID_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_ILLEGAL_TAG:1.0 */
static void
decode_ex_org_csapi_cm_P_ILLEGAL_TAG(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_ILLEGAL_TAG_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_ILLEGAL_VALUE:1.0 */
static void
decode_ex_org_csapi_cm_P_ILLEGAL_VALUE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_ILLEGAL_VALUE_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_ILLEGAL_COMBINATION:1.0 */
static void
decode_ex_org_csapi_cm_P_ILLEGAL_COMBINATION(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_ILLEGAL_COMBINATION_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_UNKNOWN_SITES:1.0 */
static void
decode_ex_org_csapi_cm_P_UNKNOWN_SITES(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_UNKNOWN_SITES_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_UNKNOWN_VPRN:1.0 */
static void
decode_ex_org_csapi_cm_P_UNKNOWN_VPRN(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_UNKNOWN_VPRN_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_ILLEGAL_SITE_ID:1.0 */
static void
decode_ex_org_csapi_cm_P_ILLEGAL_SITE_ID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_ILLEGAL_SITE_ID_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_UNKNOWN_SITE_ID:1.0 */
static void
decode_ex_org_csapi_cm_P_UNKNOWN_SITE_ID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_UNKNOWN_SITE_ID_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_UNKNOWN_SAPS:1.0 */
static void
decode_ex_org_csapi_cm_P_UNKNOWN_SAPS(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_UNKNOWN_SAPS_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_UNKNOWN_SITE_LOCATION:1.0 */
static void
decode_ex_org_csapi_cm_P_UNKNOWN_SITE_LOCATION(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_UNKNOWN_SITE_LOCATION_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_UNKNOWN_SITE_DESCRIPTION:1.0 */
static void
decode_ex_org_csapi_cm_P_UNKNOWN_SITE_DESCRIPTION(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_UNKNOWN_SITE_DESCRIPTION_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_UNKNOWN_IPSUBNET:1.0 */
static void
decode_ex_org_csapi_cm_P_UNKNOWN_IPSUBNET(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_UNKNOWN_IPSUBNET_ExtraInformation);

}

/* Exception = IDL:org/csapi/cm/P_UNKNOWN_SAP:1.0 */
static void
decode_ex_org_csapi_cm_P_UNKNOWN_SAP(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_P_UNKNOWN_SAP_ExtraInformation);

}

/* Exception = IDL:org/csapi/am/P_UNAUTHORIZED_APPLICATION:1.0 */
static void
decode_ex_org_csapi_am_P_UNAUTHORIZED_APPLICATION(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_am_P_UNAUTHORIZED_APPLICATION_ExtraInformation);

}

/* Exception = IDL:org/csapi/cs/P_INVALID_REQUEST_NUMBER:1.0 */
static void
decode_ex_org_csapi_cs_P_INVALID_REQUEST_NUMBER(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cs_P_INVALID_REQUEST_NUMBER_ExtraInformation);

}

/* Exception = IDL:org/csapi/cs/P_INVALID_VOLUME:1.0 */
static void
decode_ex_org_csapi_cs_P_INVALID_VOLUME(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cs_P_INVALID_VOLUME_ExtraInformation);

}

/* Exception = IDL:org/csapi/cs/P_INVALID_USER:1.0 */
static void
decode_ex_org_csapi_cs_P_INVALID_USER(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cs_P_INVALID_USER_ExtraInformation);

}

/* Exception = IDL:org/csapi/cs/P_INVALID_ACCOUNT:1.0 */
static void
decode_ex_org_csapi_cs_P_INVALID_ACCOUNT(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cs_P_INVALID_ACCOUNT_ExtraInformation);

}

/* Exception = IDL:org/csapi/policy/P_ACCESS_VIOLATION:1.0 */
static void
decode_ex_org_csapi_policy_P_ACCESS_VIOLATION(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_P_ACCESS_VIOLATION_ExtraInformation);

}

/* Exception = IDL:org/csapi/policy/P_SYNTAX_ERROR:1.0 */
static void
decode_ex_org_csapi_policy_P_SYNTAX_ERROR(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_P_SYNTAX_ERROR_ExtraInformation);

}

/* Exception = IDL:org/csapi/policy/P_NAME_SPACE_ERROR:1.0 */
static void
decode_ex_org_csapi_policy_P_NAME_SPACE_ERROR(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_P_NAME_SPACE_ERROR_ExtraInformation);

}

/* Exception = IDL:org/csapi/policy/P_NO_TRANSACTION_IN_PROCESS:1.0 */
static void
decode_ex_org_csapi_policy_P_NO_TRANSACTION_IN_PROCESS(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_P_NO_TRANSACTION_IN_PROCESS_ExtraInformation);

}

/* Exception = IDL:org/csapi/policy/P_TRANSACTION_IN_PROCESS:1.0 */
static void
decode_ex_org_csapi_policy_P_TRANSACTION_IN_PROCESS(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_P_TRANSACTION_IN_PROCESS_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_IDENTITY_EXISTS:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_IDENTITY_EXISTS(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_IDENTITY_EXISTS_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_UNKNOWN_TYPE:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_UNKNOWN_TYPE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_UNKNOWN_TYPE_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_INVALID_CREDENTIAL:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_INVALID_CREDENTIAL(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_INVALID_CREDENTIAL_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_UNKNOWN_IDENTITY:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_UNKNOWN_IDENTITY(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_UNKNOWN_IDENTITY_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_MEMBER_EXISTS:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_MEMBER_EXISTS(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_MEMBER_EXISTS_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_UNKNOWN_GROUP:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_UNKNOWN_GROUP(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_UNKNOWN_GROUP_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_UNKNOWN_MEMBER:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_UNKNOWN_MEMBER(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_UNKNOWN_MEMBER_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_IS_CYCLIC:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_IS_CYCLIC(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_IS_CYCLIC_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_NOT_MEMBER:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_NOT_MEMBER(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_NOT_MEMBER_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_ALIAS_EXISTS:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_ALIAS_EXISTS(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_ALIAS_EXISTS_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_ALIAS_NOT_UNIQUE:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_ALIAS_NOT_UNIQUE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_ALIAS_NOT_UNIQUE_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_UNASSIGNED_ALIAS:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_UNASSIGNED_ALIAS(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_UNASSIGNED_ALIAS_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_UNKNOWN_ALIAS:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_UNKNOWN_ALIAS(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_UNKNOWN_ALIAS_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_TYPE_ASSOCIATED:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_TYPE_ASSOCIATED(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_TYPE_ASSOCIATED_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_DISASSOCIATED_TYPE:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_DISASSOCIATED_TYPE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_DISASSOCIATED_TYPE_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_UNKNOWN_ATTRIBUTE:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_UNKNOWN_ATTRIBUTE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_UNKNOWN_ATTRIBUTE_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_UNKNOWN_ATTRIBUTES:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_UNKNOWN_ATTRIBUTES(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_UNKNOWN_ATTRIBUTES_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_AGENT_EXISTS:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_AGENT_EXISTS(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_AGENT_EXISTS_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_UNKNOWN_AGENT:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_UNKNOWN_AGENT(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_UNKNOWN_AGENT_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_NO_CAPABILITY:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_NO_CAPABILITY(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_NO_CAPABILITY_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_UNKNOWN_ASSIGNMENT:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_UNKNOWN_ASSIGNMENT(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_UNKNOWN_ASSIGNMENT_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_ATTRIBUTE_EXISTS:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_ATTRIBUTE_EXISTS(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_ATTRIBUTE_EXISTS_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_TYPE_EXISTS:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_TYPE_EXISTS(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_TYPE_EXISTS_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_UNKNOWN_CAPABILITY:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_UNKNOWN_CAPABILITY(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_UNKNOWN_CAPABILITY_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_UNAVAILABLE_INTERFACE:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_UNAVAILABLE_INTERFACE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_UNAVAILABLE_INTERFACE_ExtraInformation);

}

/* Exception = IDL:org/csapi/pam/P_PAM_NOT_REGISTERED:1.0 */
static void
decode_ex_org_csapi_pam_P_PAM_NOT_REGISTERED(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_P_PAM_NOT_REGISTERED_ExtraInformation);

}

/* Exception = IDL:org/csapi/mmm/P_MMM_INVALID_MAILBOX:1.0 */
static void
decode_ex_org_csapi_mmm_P_MMM_INVALID_MAILBOX(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_P_MMM_INVALID_MAILBOX_ExtraInformation);

}

/* Exception = IDL:org/csapi/mmm/P_MMM_INVALID_AUTHENTICATION_INFORMATION:1.0 */
static void
decode_ex_org_csapi_mmm_P_MMM_INVALID_AUTHENTICATION_INFORMATION(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_P_MMM_INVALID_AUTHENTICATION_INFORMATION_ExtraInformation);

}

/* Exception = IDL:org/csapi/mmm/P_MMM_INVALID_FOLDER_ID:1.0 */
static void
decode_ex_org_csapi_mmm_P_MMM_INVALID_FOLDER_ID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_P_MMM_INVALID_FOLDER_ID_ExtraInformation);

}

/* Exception = IDL:org/csapi/mmm/P_MMM_MAX_MESSAGE_SIZE_EXCEEDED:1.0 */
static void
decode_ex_org_csapi_mmm_P_MMM_MAX_MESSAGE_SIZE_EXCEEDED(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_P_MMM_MAX_MESSAGE_SIZE_EXCEEDED_ExtraInformation);

}

/* Exception = IDL:org/csapi/mmm/P_MMM_INVALID_MESSAGE_ID:1.0 */
static void
decode_ex_org_csapi_mmm_P_MMM_INVALID_MESSAGE_ID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_P_MMM_INVALID_MESSAGE_ID_ExtraInformation);

}

/* Exception = IDL:org/csapi/mmm/P_MMM_INVALID_PART_ID:1.0 */
static void
decode_ex_org_csapi_mmm_P_MMM_INVALID_PART_ID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_P_MMM_INVALID_PART_ID_ExtraInformation);

}

/* Exception = IDL:org/csapi/mmm/P_MMM_INVALID_PROPERTY:1.0 */
static void
decode_ex_org_csapi_mmm_P_MMM_INVALID_PROPERTY(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_P_MMM_INVALID_PROPERTY_ExtraInformation);

}

/* Exception = IDL:org/csapi/mmm/P_MMM_INVALID_DELIVERY_TYPE:1.0 */
static void
decode_ex_org_csapi_mmm_P_MMM_INVALID_DELIVERY_TYPE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_P_MMM_INVALID_DELIVERY_TYPE_ExtraInformation);

}

/* Exception = IDL:org/csapi/mmm/P_MMM_DELIVERY_TYPE_ADDRESS_TYPE_MISMATCH:1.0 */
static void
decode_ex_org_csapi_mmm_P_MMM_DELIVERY_TYPE_ADDRESS_TYPE_MISMATCH(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_P_MMM_DELIVERY_TYPE_ADDRESS_TYPE_MISMATCH_ExtraInformation);

}

/* Exception = IDL:org/csapi/mmm/P_MMM_DELIVERY_TYPE_MESSAGE_TYPE_MISMATCH:1.0 */
static void
decode_ex_org_csapi_mmm_P_MMM_DELIVERY_TYPE_MESSAGE_TYPE_MISMATCH(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_P_MMM_DELIVERY_TYPE_MESSAGE_TYPE_MISMATCH_ExtraInformation);

}

/* Exception = IDL:org/csapi/mmm/P_MMM_INVALID_DELIVERY_TIME:1.0 */
static void
decode_ex_org_csapi_mmm_P_MMM_INVALID_DELIVERY_TIME(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_P_MMM_INVALID_DELIVERY_TIME_ExtraInformation);

}

/* Exception = IDL:org/csapi/mmm/P_MMM_INVALID_VALIDITY_TIME:1.0 */
static void
decode_ex_org_csapi_mmm_P_MMM_INVALID_VALIDITY_TIME(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_P_MMM_INVALID_VALIDITY_TIME_ExtraInformation);

}

/* Exception = IDL:org/csapi/mmm/P_MMM_MAX_SUBJECT_SIZE_EXCEEDED:1.0 */
static void
decode_ex_org_csapi_mmm_P_MMM_MAX_SUBJECT_SIZE_EXCEEDED(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_P_MMM_MAX_SUBJECT_SIZE_EXCEEDED_ExtraInformation);

}

/* Exception = IDL:org/csapi/mmm/P_MMM_INVALID_HEADER:1.0 */
static void
decode_ex_org_csapi_mmm_P_MMM_INVALID_HEADER(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_P_MMM_INVALID_HEADER_ExtraInformation);

}

/* Exception = IDL:org/csapi/mmm/P_MMM_CANNOT_CANCEL:1.0 */
static void
decode_ex_org_csapi_mmm_P_MMM_CANNOT_CANCEL(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_P_MMM_CANNOT_CANCEL_ExtraInformation);

}

/* Exception = IDL:org/csapi/mmm/P_MMM_INFORMATION_NOT_AVAILABLE:1.0 */
static void
decode_ex_org_csapi_mmm_P_MMM_INFORMATION_NOT_AVAILABLE(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item *item _U_;

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_P_MMM_INFORMATION_NOT_AVAILABLE_ExtraInformation);

}


/*  End Exception Helper Functions  */

/*
 * Main delegator for exception handling
 *
 */
static gboolean
decode_user_exception(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *ptree _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_tree *tree _U_;

    if (!header->exception_id)
        return FALSE;

    if (strcmp(header->exception_id, "IDL:org/csapi/TpCommonExceptions:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_TpCommonExceptions(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/TpCommonExceptions:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/P_INVALID_INTERFACE_TYPE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_P_INVALID_INTERFACE_TYPE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/P_INVALID_INTERFACE_TYPE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/P_INVALID_SESSION_ID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_P_INVALID_SESSION_ID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/P_INVALID_SESSION_ID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_INVALID_DOMAIN_ID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_INVALID_DOMAIN_ID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_INVALID_DOMAIN_ID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_INVALID_AUTH_TYPE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_INVALID_AUTH_TYPE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_INVALID_AUTH_TYPE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/P_INVALID_VERSION:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_P_INVALID_VERSION(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/P_INVALID_VERSION:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_ACCESS_DENIED:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_ACCESS_DENIED(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_ACCESS_DENIED:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_INVALID_ACCESS_TYPE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_INVALID_ACCESS_TYPE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_INVALID_ACCESS_TYPE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_INVALID_SIGNING_ALGORITHM:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_INVALID_SIGNING_ALGORITHM(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_INVALID_SIGNING_ALGORITHM:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_INVALID_SIGNATURE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_INVALID_SIGNATURE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_INVALID_SIGNATURE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/P_INVALID_INTERFACE_NAME:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_P_INVALID_INTERFACE_NAME(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/P_INVALID_INTERFACE_NAME:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_INVALID_PROPERTY:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_INVALID_PROPERTY(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_INVALID_PROPERTY:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_NO_ACCEPTABLE_SIGNING_ALGORITHM:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_NO_ACCEPTABLE_SIGNING_ALGORITHM(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_NO_ACCEPTABLE_SIGNING_ALGORITHM:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_NO_ACCEPTABLE_ENCRYPTION_CAPABILITY:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_NO_ACCEPTABLE_ENCRYPTION_CAPABILITY(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_NO_ACCEPTABLE_ENCRYPTION_CAPABILITY:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_NO_ACCEPTABLE_AUTHENTICATION_MECHANISM:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_NO_ACCEPTABLE_AUTHENTICATION_MECHANISM(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_NO_ACCEPTABLE_AUTHENTICATION_MECHANISM:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/P_INVALID_CRITERIA:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_P_INVALID_CRITERIA(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/P_INVALID_CRITERIA:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/P_INVALID_EVENT_TYPE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_P_INVALID_EVENT_TYPE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/P_INVALID_EVENT_TYPE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/P_INVALID_ASSIGNMENT_ID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_P_INVALID_ASSIGNMENT_ID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/P_INVALID_ASSIGNMENT_ID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_INVALID_SERVICE_ID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_INVALID_SERVICE_ID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_INVALID_SERVICE_ID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_SERVICE_NOT_ENABLED:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_SERVICE_NOT_ENABLED(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_SERVICE_NOT_ENABLED:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/P_UNAUTHORISED_PARAMETER_VALUE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_P_UNAUTHORISED_PARAMETER_VALUE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/P_UNAUTHORISED_PARAMETER_VALUE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/P_INVALID_TIME_AND_DATE_FORMAT:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_P_INVALID_TIME_AND_DATE_FORMAT(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/P_INVALID_TIME_AND_DATE_FORMAT:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_INVALID_ACTIVITY_TEST_ID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_INVALID_ACTIVITY_TEST_ID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_INVALID_ACTIVITY_TEST_ID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_ILLEGAL_SERVICE_TYPE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_ILLEGAL_SERVICE_TYPE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_ILLEGAL_SERVICE_TYPE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_UNKNOWN_SERVICE_TYPE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_UNKNOWN_SERVICE_TYPE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_UNKNOWN_SERVICE_TYPE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_INVALID_AGREEMENT_TEXT:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_INVALID_AGREEMENT_TEXT(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_INVALID_AGREEMENT_TEXT:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_INVALID_SERVICE_TOKEN:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_INVALID_SERVICE_TOKEN(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_INVALID_SERVICE_TOKEN:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_SERVICE_ACCESS_DENIED:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_SERVICE_ACCESS_DENIED(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_SERVICE_ACCESS_DENIED:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_INVALID_SERVICE_PROFILE_ID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_INVALID_SERVICE_PROFILE_ID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_INVALID_SERVICE_PROFILE_ID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_INVALID_SAG_ID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_INVALID_SAG_ID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_INVALID_SAG_ID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_INVALID_SAG_TO_SERVICE_PROFILE_ASSIGNMENT:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_INVALID_SAG_TO_SERVICE_PROFILE_ASSIGNMENT(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_INVALID_SAG_TO_SERVICE_PROFILE_ASSIGNMENT:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_INVALID_SERVICE_CONTRACT_ID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_INVALID_SERVICE_CONTRACT_ID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_INVALID_SERVICE_CONTRACT_ID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_INVALID_CLIENT_APP_ID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_INVALID_CLIENT_APP_ID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_INVALID_CLIENT_APP_ID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_INVALID_ADDITION_TO_SAG:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_INVALID_ADDITION_TO_SAG(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_INVALID_ADDITION_TO_SAG:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_PROPERTY_TYPE_MISMATCH:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_PROPERTY_TYPE_MISMATCH(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_PROPERTY_TYPE_MISMATCH:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_DUPLICATE_PROPERTY_NAME:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_DUPLICATE_PROPERTY_NAME(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_DUPLICATE_PROPERTY_NAME:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_MISSING_MANDATORY_PROPERTY:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_MISSING_MANDATORY_PROPERTY(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_MISSING_MANDATORY_PROPERTY:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_SERVICE_TYPE_UNAVAILABLE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_SERVICE_TYPE_UNAVAILABLE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_SERVICE_TYPE_UNAVAILABLE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_ILLEGAL_SERVICE_ID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_ILLEGAL_SERVICE_ID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_ILLEGAL_SERVICE_ID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/fw/P_UNKNOWN_SERVICE_ID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_fw_P_UNKNOWN_SERVICE_ID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/fw/P_UNKNOWN_SERVICE_ID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/P_INVALID_ADDRESS:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_P_INVALID_ADDRESS(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/P_INVALID_ADDRESS:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/P_UNSUPPORTED_ADDRESS_PLAN:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_P_UNSUPPORTED_ADDRESS_PLAN(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/P_UNSUPPORTED_ADDRESS_PLAN:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/P_INVALID_NETWORK_STATE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_P_INVALID_NETWORK_STATE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/P_INVALID_NETWORK_STATE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/P_INVALID_CURRENCY:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_P_INVALID_CURRENCY(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/P_INVALID_CURRENCY:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/P_INVALID_AMOUNT:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_P_INVALID_AMOUNT(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/P_INVALID_AMOUNT:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/P_INFORMATION_NOT_AVAILABLE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_P_INFORMATION_NOT_AVAILABLE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/P_INFORMATION_NOT_AVAILABLE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/ui/P_ILLEGAL_ID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_ui_P_ILLEGAL_ID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/ui/P_ILLEGAL_ID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/ui/P_ID_NOT_FOUND:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_ui_P_ID_NOT_FOUND(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/ui/P_ID_NOT_FOUND:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/ui/P_ILLEGAL_RANGE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_ui_P_ILLEGAL_RANGE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/ui/P_ILLEGAL_RANGE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/ui/P_INVALID_COLLECTION_CRITERIA:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_ui_P_INVALID_COLLECTION_CRITERIA(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/ui/P_INVALID_COLLECTION_CRITERIA:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/P_APPLICATION_NOT_ACTIVATED:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_P_APPLICATION_NOT_ACTIVATED(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/P_APPLICATION_NOT_ACTIVATED:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/mm/P_REQUESTED_ACCURACY_CANNOT_BE_DELIVERED:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_mm_P_REQUESTED_ACCURACY_CANNOT_BE_DELIVERED(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/mm/P_REQUESTED_ACCURACY_CANNOT_BE_DELIVERED:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/mm/P_REQUESTED_RESPONSE_TIME_CANNOT_BE_DELIVERED:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_mm_P_REQUESTED_RESPONSE_TIME_CANNOT_BE_DELIVERED(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/mm/P_REQUESTED_RESPONSE_TIME_CANNOT_BE_DELIVERED:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/mm/P_INVALID_REPORTING_INTERVAL:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_mm_P_INVALID_REPORTING_INTERVAL(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/mm/P_INVALID_REPORTING_INTERVAL:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/mm/P_TRIGGER_CONDITIONS_NOT_SUBSCRIBED:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_mm_P_TRIGGER_CONDITIONS_NOT_SUBSCRIBED(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/mm/P_TRIGGER_CONDITIONS_NOT_SUBSCRIBED:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/P_UNKNOWN_SUBSCRIBER:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_P_UNKNOWN_SUBSCRIBER(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/P_UNKNOWN_SUBSCRIBER:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/termcap/P_INVALID_TERMINAL_ID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_termcap_P_INVALID_TERMINAL_ID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/termcap/P_INVALID_TERMINAL_ID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/gms/P_GMS_INVALID_MAILBOX:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_gms_P_GMS_INVALID_MAILBOX(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/gms/P_GMS_INVALID_MAILBOX:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/gms/P_GMS_INVALID_AUTHENTICATION_INFORMATION:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_gms_P_GMS_INVALID_AUTHENTICATION_INFORMATION(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/gms/P_GMS_INVALID_AUTHENTICATION_INFORMATION:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/gms/P_GMS_LOCKING_LOCKED_MAILBOX:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_gms_P_GMS_LOCKING_LOCKED_MAILBOX(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/gms/P_GMS_LOCKING_LOCKED_MAILBOX:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/gms/P_GMS_UNLOCKING_UNLOCKED_MAILBOX:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_gms_P_GMS_UNLOCKING_UNLOCKED_MAILBOX(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/gms/P_GMS_UNLOCKING_UNLOCKED_MAILBOX:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/gms/P_GMS_CANNOT_UNLOCK_MAILBOX:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_gms_P_GMS_CANNOT_UNLOCK_MAILBOX(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/gms/P_GMS_CANNOT_UNLOCK_MAILBOX:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/gms/P_GMS_NUMBER_NOT_POSITIVE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_gms_P_GMS_NUMBER_NOT_POSITIVE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/gms/P_GMS_NUMBER_NOT_POSITIVE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/gms/P_GMS_PROPERTY_NOT_SET:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_gms_P_GMS_PROPERTY_NOT_SET(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/gms/P_GMS_PROPERTY_NOT_SET:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/gms/P_GMS_MAILBOX_LOCKED:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_gms_P_GMS_MAILBOX_LOCKED(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/gms/P_GMS_MAILBOX_LOCKED:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/gms/P_GMS_FOLDER_IS_OPEN:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_gms_P_GMS_FOLDER_IS_OPEN(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/gms/P_GMS_FOLDER_IS_OPEN:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/gms/P_GMS_INVALID_FOLDER_ID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_gms_P_GMS_INVALID_FOLDER_ID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/gms/P_GMS_INVALID_FOLDER_ID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/gms/P_GMS_INSUFFICIENT_PRIVILEGE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_gms_P_GMS_INSUFFICIENT_PRIVILEGE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/gms/P_GMS_INSUFFICIENT_PRIVILEGE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/gms/P_GMS_MAILBOX_OPEN:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_gms_P_GMS_MAILBOX_OPEN(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/gms/P_GMS_MAILBOX_OPEN:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/gms/P_GMS_INVALID_MESSAGE_ID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_gms_P_GMS_INVALID_MESSAGE_ID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/gms/P_GMS_INVALID_MESSAGE_ID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/gms/P_GMS_MESSAGE_NOT_REMOVED:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_gms_P_GMS_MESSAGE_NOT_REMOVED(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/gms/P_GMS_MESSAGE_NOT_REMOVED:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_UNKNOWN_VPRP_ID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_UNKNOWN_VPRP_ID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_UNKNOWN_VPRP_ID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_UNKNOWN_SLA_ID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_UNKNOWN_SLA_ID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_UNKNOWN_SLA_ID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_UNKNOWN_STATUS:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_UNKNOWN_STATUS(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_UNKNOWN_STATUS:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_UNKNOWN_QOS_INFO:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_UNKNOWN_QOS_INFO(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_UNKNOWN_QOS_INFO:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_UNKNOWN_VALIDITY_INFO:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_UNKNOWN_VALIDITY_INFO(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_UNKNOWN_VALIDITY_INFO:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_UNKNOWN_PIPEQOSINFO:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_UNKNOWN_PIPEQOSINFO(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_UNKNOWN_PIPEQOSINFO:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_UNKNOWN_DSCODEPOINT:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_UNKNOWN_DSCODEPOINT(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_UNKNOWN_DSCODEPOINT:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_UNKNOWN_VPRP:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_UNKNOWN_VPRP(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_UNKNOWN_VPRP:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_ILLEGAL_VPRPID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_ILLEGAL_VPRPID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_ILLEGAL_VPRPID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_ILLEGAL_REF_VALUE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_ILLEGAL_REF_VALUE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_ILLEGAL_REF_VALUE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_UNKNOWN_INTERFACE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_UNKNOWN_INTERFACE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_UNKNOWN_INTERFACE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_CANT_DELETE_VPRP:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_CANT_DELETE_VPRP(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_CANT_DELETE_VPRP:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_UNKNOWN_TEMPLATES:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_UNKNOWN_TEMPLATES(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_UNKNOWN_TEMPLATES:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_UNKNOWN_MENU:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_UNKNOWN_MENU(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_UNKNOWN_MENU:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_UNKNOWN_ENTERPRISE_NETWORK:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_UNKNOWN_ENTERPRISE_NETWORK(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_UNKNOWN_ENTERPRISE_NETWORK:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_UNKNOWN_TEMPLATE_TYPE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_UNKNOWN_TEMPLATE_TYPE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_UNKNOWN_TEMPLATE_TYPE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_UNKNOWN_DESCRIPTION:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_UNKNOWN_DESCRIPTION(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_UNKNOWN_DESCRIPTION:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_ILLEGAL_SLA_ID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_ILLEGAL_SLA_ID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_ILLEGAL_SLA_ID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_ILLEGAL_TAG:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_ILLEGAL_TAG(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_ILLEGAL_TAG:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_ILLEGAL_VALUE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_ILLEGAL_VALUE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_ILLEGAL_VALUE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_ILLEGAL_COMBINATION:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_ILLEGAL_COMBINATION(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_ILLEGAL_COMBINATION:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_UNKNOWN_SITES:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_UNKNOWN_SITES(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_UNKNOWN_SITES:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_UNKNOWN_VPRN:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_UNKNOWN_VPRN(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_UNKNOWN_VPRN:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_ILLEGAL_SITE_ID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_ILLEGAL_SITE_ID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_ILLEGAL_SITE_ID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_UNKNOWN_SITE_ID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_UNKNOWN_SITE_ID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_UNKNOWN_SITE_ID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_UNKNOWN_SAPS:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_UNKNOWN_SAPS(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_UNKNOWN_SAPS:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_UNKNOWN_SITE_LOCATION:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_UNKNOWN_SITE_LOCATION(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_UNKNOWN_SITE_LOCATION:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_UNKNOWN_SITE_DESCRIPTION:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_UNKNOWN_SITE_DESCRIPTION(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_UNKNOWN_SITE_DESCRIPTION:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_UNKNOWN_IPSUBNET:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_UNKNOWN_IPSUBNET(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_UNKNOWN_IPSUBNET:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cm/P_UNKNOWN_SAP:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cm_P_UNKNOWN_SAP(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cm/P_UNKNOWN_SAP:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/am/P_UNAUTHORIZED_APPLICATION:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_am_P_UNAUTHORIZED_APPLICATION(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/am/P_UNAUTHORIZED_APPLICATION:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cs/P_INVALID_REQUEST_NUMBER:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cs_P_INVALID_REQUEST_NUMBER(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cs/P_INVALID_REQUEST_NUMBER:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cs/P_INVALID_VOLUME:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cs_P_INVALID_VOLUME(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cs/P_INVALID_VOLUME:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cs/P_INVALID_USER:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cs_P_INVALID_USER(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cs/P_INVALID_USER:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/cs/P_INVALID_ACCOUNT:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_cs_P_INVALID_ACCOUNT(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/cs/P_INVALID_ACCOUNT:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/policy/P_ACCESS_VIOLATION:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_policy_P_ACCESS_VIOLATION(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/policy/P_ACCESS_VIOLATION:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/policy/P_SYNTAX_ERROR:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_policy_P_SYNTAX_ERROR(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/policy/P_SYNTAX_ERROR:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/policy/P_NAME_SPACE_ERROR:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_policy_P_NAME_SPACE_ERROR(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/policy/P_NAME_SPACE_ERROR:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/policy/P_NO_TRANSACTION_IN_PROCESS:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_policy_P_NO_TRANSACTION_IN_PROCESS(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/policy/P_NO_TRANSACTION_IN_PROCESS:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/policy/P_TRANSACTION_IN_PROCESS:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_policy_P_TRANSACTION_IN_PROCESS(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/policy/P_TRANSACTION_IN_PROCESS:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_IDENTITY_EXISTS:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_IDENTITY_EXISTS(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_IDENTITY_EXISTS:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_UNKNOWN_TYPE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_UNKNOWN_TYPE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_UNKNOWN_TYPE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_INVALID_CREDENTIAL:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_INVALID_CREDENTIAL(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_INVALID_CREDENTIAL:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_UNKNOWN_IDENTITY:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_UNKNOWN_IDENTITY(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_UNKNOWN_IDENTITY:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_MEMBER_EXISTS:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_MEMBER_EXISTS(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_MEMBER_EXISTS:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_UNKNOWN_GROUP:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_UNKNOWN_GROUP(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_UNKNOWN_GROUP:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_UNKNOWN_MEMBER:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_UNKNOWN_MEMBER(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_UNKNOWN_MEMBER:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_IS_CYCLIC:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_IS_CYCLIC(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_IS_CYCLIC:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_NOT_MEMBER:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_NOT_MEMBER(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_NOT_MEMBER:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_ALIAS_EXISTS:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_ALIAS_EXISTS(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_ALIAS_EXISTS:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_ALIAS_NOT_UNIQUE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_ALIAS_NOT_UNIQUE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_ALIAS_NOT_UNIQUE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_UNASSIGNED_ALIAS:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_UNASSIGNED_ALIAS(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_UNASSIGNED_ALIAS:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_UNKNOWN_ALIAS:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_UNKNOWN_ALIAS(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_UNKNOWN_ALIAS:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_TYPE_ASSOCIATED:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_TYPE_ASSOCIATED(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_TYPE_ASSOCIATED:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_DISASSOCIATED_TYPE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_DISASSOCIATED_TYPE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_DISASSOCIATED_TYPE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_UNKNOWN_ATTRIBUTE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_UNKNOWN_ATTRIBUTE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_UNKNOWN_ATTRIBUTE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_UNKNOWN_ATTRIBUTES:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_UNKNOWN_ATTRIBUTES(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_UNKNOWN_ATTRIBUTES:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_AGENT_EXISTS:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_AGENT_EXISTS(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_AGENT_EXISTS:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_UNKNOWN_AGENT:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_UNKNOWN_AGENT(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_UNKNOWN_AGENT:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_NO_CAPABILITY:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_NO_CAPABILITY(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_NO_CAPABILITY:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_UNKNOWN_ASSIGNMENT:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_UNKNOWN_ASSIGNMENT(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_UNKNOWN_ASSIGNMENT:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_ATTRIBUTE_EXISTS:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_ATTRIBUTE_EXISTS(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_ATTRIBUTE_EXISTS:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_TYPE_EXISTS:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_TYPE_EXISTS(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_TYPE_EXISTS:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_UNKNOWN_CAPABILITY:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_UNKNOWN_CAPABILITY(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_UNKNOWN_CAPABILITY:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_UNAVAILABLE_INTERFACE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_UNAVAILABLE_INTERFACE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_UNAVAILABLE_INTERFACE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/pam/P_PAM_NOT_REGISTERED:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_pam_P_PAM_NOT_REGISTERED(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/pam/P_PAM_NOT_REGISTERED:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/mmm/P_MMM_INVALID_MAILBOX:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_mmm_P_MMM_INVALID_MAILBOX(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/mmm/P_MMM_INVALID_MAILBOX:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/mmm/P_MMM_INVALID_AUTHENTICATION_INFORMATION:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_mmm_P_MMM_INVALID_AUTHENTICATION_INFORMATION(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/mmm/P_MMM_INVALID_AUTHENTICATION_INFORMATION:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/mmm/P_MMM_INVALID_FOLDER_ID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_mmm_P_MMM_INVALID_FOLDER_ID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/mmm/P_MMM_INVALID_FOLDER_ID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/mmm/P_MMM_MAX_MESSAGE_SIZE_EXCEEDED:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_mmm_P_MMM_MAX_MESSAGE_SIZE_EXCEEDED(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/mmm/P_MMM_MAX_MESSAGE_SIZE_EXCEEDED:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/mmm/P_MMM_INVALID_MESSAGE_ID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_mmm_P_MMM_INVALID_MESSAGE_ID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/mmm/P_MMM_INVALID_MESSAGE_ID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/mmm/P_MMM_INVALID_PART_ID:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_mmm_P_MMM_INVALID_PART_ID(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/mmm/P_MMM_INVALID_PART_ID:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/mmm/P_MMM_INVALID_PROPERTY:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_mmm_P_MMM_INVALID_PROPERTY(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/mmm/P_MMM_INVALID_PROPERTY:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/mmm/P_MMM_INVALID_DELIVERY_TYPE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_mmm_P_MMM_INVALID_DELIVERY_TYPE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/mmm/P_MMM_INVALID_DELIVERY_TYPE:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/mmm/P_MMM_DELIVERY_TYPE_ADDRESS_TYPE_MISMATCH:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_mmm_P_MMM_DELIVERY_TYPE_ADDRESS_TYPE_MISMATCH(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/mmm/P_MMM_DELIVERY_TYPE_ADDRESS_TYPE_MISMATCH:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/mmm/P_MMM_DELIVERY_TYPE_MESSAGE_TYPE_MISMATCH:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_mmm_P_MMM_DELIVERY_TYPE_MESSAGE_TYPE_MISMATCH(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/mmm/P_MMM_DELIVERY_TYPE_MESSAGE_TYPE_MISMATCH:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/mmm/P_MMM_INVALID_DELIVERY_TIME:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_mmm_P_MMM_INVALID_DELIVERY_TIME(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/mmm/P_MMM_INVALID_DELIVERY_TIME:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/mmm/P_MMM_INVALID_VALIDITY_TIME:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_mmm_P_MMM_INVALID_VALIDITY_TIME(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/mmm/P_MMM_INVALID_VALIDITY_TIME:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/mmm/P_MMM_MAX_SUBJECT_SIZE_EXCEEDED:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_mmm_P_MMM_MAX_SUBJECT_SIZE_EXCEEDED(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/mmm/P_MMM_MAX_SUBJECT_SIZE_EXCEEDED:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/mmm/P_MMM_INVALID_HEADER:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_mmm_P_MMM_INVALID_HEADER(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/mmm/P_MMM_INVALID_HEADER:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/mmm/P_MMM_CANNOT_CANCEL:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_mmm_P_MMM_CANNOT_CANCEL(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/mmm/P_MMM_CANNOT_CANCEL:1.0  */
        return TRUE;
    }

    if (strcmp(header->exception_id, "IDL:org/csapi/mmm/P_MMM_INFORMATION_NOT_AVAILABLE:1.0") == 0) {
        tree = start_dissecting(tvb, pinfo, ptree, offset);
        decode_ex_org_csapi_mmm_P_MMM_INFORMATION_NOT_AVAILABLE(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);   /*  IDL:org/csapi/mmm/P_MMM_INFORMATION_NOT_AVAILABLE:1.0  */
        return TRUE;
    }


    return FALSE;    /* user exception not found */
}

/*
 * IDL:org/csapi/IpService/setCallback:1.0
 */
static void
decode_org_csapi_IpService_setCallback(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/IpService/setCallbackWithSessionID:1.0
 */
static void
decode_org_csapi_IpService_setCallbackWithSessionID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        proto_tree_add_int(tree, hf_org_csapi_IpService_setCallbackWithSessionID_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_access/trust_and_security/IpInitial/initiateAuthentication:1.0
 */
static void
decode_org_csapi_fw_fw_access_trust_and_security_IpInitial_initiateAuthentication(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_fw_TpAuthDomain"  */
        decode_org_csapi_fw_TpAuthDomain_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpAuthDomain"  */
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_access_trust_and_security_IpInitial_initiateAuthentication_authType);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_fw_TpAuthDomain"  */
            decode_org_csapi_fw_TpAuthDomain_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpAuthDomain"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_access/trust_and_security/IpInitial/initiateAuthenticationWithVersion:1.0
 */
static void
decode_org_csapi_fw_fw_access_trust_and_security_IpInitial_initiateAuthenticationWithVersion(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_fw_TpAuthDomain"  */
        decode_org_csapi_fw_TpAuthDomain_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpAuthDomain"  */
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_access_trust_and_security_IpInitial_initiateAuthenticationWithVersion_authType);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_access_trust_and_security_IpInitial_initiateAuthenticationWithVersion_frameworkVersion);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_fw_TpAuthDomain"  */
            decode_org_csapi_fw_TpAuthDomain_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpAuthDomain"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_access/trust_and_security/IpAuthentication/requestAccess:1.0
 */
static void
decode_org_csapi_fw_fw_access_trust_and_security_IpAuthentication_requestAccess(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_access_trust_and_security_IpAuthentication_requestAccess_accessType);

        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_access/trust_and_security/IpClientAccess/terminateAccess:1.0
 */
static void
decode_org_csapi_fw_fw_access_trust_and_security_IpClientAccess_terminateAccess(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpClientAccess_terminateAccess_digitalSignature;
    guint32   i_org_csapi_fw_fw_access_trust_and_security_IpClientAccess_terminateAccess_digitalSignature;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_access_trust_and_security_IpClientAccess_terminateAccess_terminationText);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_access_trust_and_security_IpClientAccess_terminateAccess_signingAlgorithm);

        u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpClientAccess_terminateAccess_digitalSignature = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpClientAccess_terminateAccess_digitalSignature_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpClientAccess_terminateAccess_digitalSignature);

        for (i_org_csapi_fw_fw_access_trust_and_security_IpClientAccess_terminateAccess_digitalSignature=0; i_org_csapi_fw_fw_access_trust_and_security_IpClientAccess_terminateAccess_digitalSignature < u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpClientAccess_terminateAccess_digitalSignature; i_org_csapi_fw_fw_access_trust_and_security_IpClientAccess_terminateAccess_digitalSignature++) {

            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpClientAccess_terminateAccess_digitalSignature, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_access/trust_and_security/IpAccess/obtainInterface:1.0
 */
static void
decode_org_csapi_fw_fw_access_trust_and_security_IpAccess_obtainInterface(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_obtainInterface_interfaceName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_access/trust_and_security/IpAccess/obtainInterfaceWithCallback:1.0
 */
static void
decode_org_csapi_fw_fw_access_trust_and_security_IpAccess_obtainInterfaceWithCallback(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_obtainInterfaceWithCallback_interfaceName);

        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_access/trust_and_security/IpAccess/endAccess:1.0
 */
static void
decode_org_csapi_fw_fw_access_trust_and_security_IpAccess_endAccess(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAccess_endAccess_endAccessProperties;
    guint32   i_org_csapi_fw_fw_access_trust_and_security_IpAccess_endAccess_endAccessProperties;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAccess_endAccess_endAccessProperties = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_endAccess_endAccessProperties_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAccess_endAccess_endAccessProperties);

        for (i_org_csapi_fw_fw_access_trust_and_security_IpAccess_endAccess_endAccessProperties=0; i_org_csapi_fw_fw_access_trust_and_security_IpAccess_endAccess_endAccessProperties < u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAccess_endAccess_endAccessProperties; i_org_csapi_fw_fw_access_trust_and_security_IpAccess_endAccess_endAccessProperties++) {

            /*  Begin struct "org_csapi_fw_TpProperty"  */
            decode_org_csapi_fw_TpProperty_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpProperty"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_access/trust_and_security/IpAccess/listInterfaces:1.0
 */
static void
decode_org_csapi_fw_fw_access_trust_and_security_IpAccess_listInterfaces(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAccess_listInterfaces_return;
    guint32   i_org_csapi_fw_fw_access_trust_and_security_IpAccess_listInterfaces_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAccess_listInterfaces_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_listInterfaces_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAccess_listInterfaces_return);

            for (i_org_csapi_fw_fw_access_trust_and_security_IpAccess_listInterfaces_return=0; i_org_csapi_fw_fw_access_trust_and_security_IpAccess_listInterfaces_return < u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAccess_listInterfaces_return; i_org_csapi_fw_fw_access_trust_and_security_IpAccess_listInterfaces_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_listInterfaces_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_access/trust_and_security/IpAccess/releaseInterface:1.0
 */
static void
decode_org_csapi_fw_fw_access_trust_and_security_IpAccess_releaseInterface(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_releaseInterface_interfaceName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_access/trust_and_security/IpAccess/selectSigningAlgorithm:1.0
 */
static void
decode_org_csapi_fw_fw_access_trust_and_security_IpAccess_selectSigningAlgorithm(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_selectSigningAlgorithm_signingAlgorithmCaps);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_selectSigningAlgorithm_return);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_access/trust_and_security/IpAccess/terminateAccess:1.0
 */
static void
decode_org_csapi_fw_fw_access_trust_and_security_IpAccess_terminateAccess(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAccess_terminateAccess_digitalSignature;
    guint32   i_org_csapi_fw_fw_access_trust_and_security_IpAccess_terminateAccess_digitalSignature;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_terminateAccess_terminationText);

        u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAccess_terminateAccess_digitalSignature = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_terminateAccess_digitalSignature_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAccess_terminateAccess_digitalSignature);

        for (i_org_csapi_fw_fw_access_trust_and_security_IpAccess_terminateAccess_digitalSignature=0; i_org_csapi_fw_fw_access_trust_and_security_IpAccess_terminateAccess_digitalSignature < u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAccess_terminateAccess_digitalSignature; i_org_csapi_fw_fw_access_trust_and_security_IpAccess_terminateAccess_digitalSignature++) {

            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_terminateAccess_digitalSignature, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_access/trust_and_security/IpAccess/relinquishInterface:1.0
 */
static void
decode_org_csapi_fw_fw_access_trust_and_security_IpAccess_relinquishInterface(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAccess_relinquishInterface_digitalSignature;
    guint32   i_org_csapi_fw_fw_access_trust_and_security_IpAccess_relinquishInterface_digitalSignature;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_relinquishInterface_interfaceName);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_relinquishInterface_terminationText);

        u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAccess_relinquishInterface_digitalSignature = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_relinquishInterface_digitalSignature_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAccess_relinquishInterface_digitalSignature);

        for (i_org_csapi_fw_fw_access_trust_and_security_IpAccess_relinquishInterface_digitalSignature=0; i_org_csapi_fw_fw_access_trust_and_security_IpAccess_relinquishInterface_digitalSignature < u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAccess_relinquishInterface_digitalSignature; i_org_csapi_fw_fw_access_trust_and_security_IpAccess_relinquishInterface_digitalSignature++) {

            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_relinquishInterface_digitalSignature, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_access/trust_and_security/IpClientAPILevelAuthentication/authenticate:1.0
 */
static void
decode_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_challenge;
    guint32   i_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_challenge;
    guint32   u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_return;
    guint32   i_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_challenge = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_challenge_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_challenge);

        for (i_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_challenge=0; i_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_challenge < u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_challenge; i_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_challenge++) {

            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_challenge, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_return);

            for (i_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_return=0; i_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_return < u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_return; i_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_return++) {

                proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_return, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_access/trust_and_security/IpClientAPILevelAuthentication/abortAuthentication:1.0
 */
static void
decode_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_abortAuthentication(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_access/trust_and_security/IpClientAPILevelAuthentication/authenticationSucceeded:1.0
 */
static void
decode_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticationSucceeded(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_access/trust_and_security/IpClientAPILevelAuthentication/challenge:1.0
 */
static void
decode_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_challenge;
    guint32   i_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_challenge;
    guint32   u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_return;
    guint32   i_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_challenge = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_challenge_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_challenge);

        for (i_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_challenge=0; i_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_challenge < u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_challenge; i_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_challenge++) {

            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_challenge, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_return);

            for (i_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_return=0; i_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_return < u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_return; i_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_return++) {

                proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_return, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_access/trust_and_security/IpAPILevelAuthentication/selectEncryptionMethod:1.0
 */
static void
decode_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_selectEncryptionMethod(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_selectEncryptionMethod_encryptionCaps);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_selectEncryptionMethod_return);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_access/trust_and_security/IpAPILevelAuthentication/authenticate:1.0
 */
static void
decode_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_challenge;
    guint32   i_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_challenge;
    guint32   u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_return;
    guint32   i_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_challenge = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_challenge_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_challenge);

        for (i_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_challenge=0; i_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_challenge < u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_challenge; i_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_challenge++) {

            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_challenge, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_return);

            for (i_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_return=0; i_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_return < u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_return; i_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_return++) {

                proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_return, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_access/trust_and_security/IpAPILevelAuthentication/abortAuthentication:1.0
 */
static void
decode_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_abortAuthentication(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_access/trust_and_security/IpAPILevelAuthentication/authenticationSucceeded:1.0
 */
static void
decode_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticationSucceeded(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_access/trust_and_security/IpAPILevelAuthentication/selectAuthenticationMechanism:1.0
 */
static void
decode_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_selectAuthenticationMechanism(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_selectAuthenticationMechanism_authMechanismList);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_selectAuthenticationMechanism_return);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_access/trust_and_security/IpAPILevelAuthentication/challenge:1.0
 */
static void
decode_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_challenge;
    guint32   i_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_challenge;
    guint32   u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_return;
    guint32   i_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_challenge = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_challenge_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_challenge);

        for (i_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_challenge=0; i_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_challenge < u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_challenge; i_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_challenge++) {

            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_challenge, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_return);

            for (i_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_return=0; i_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_return < u_octet4_loop_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_return; i_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_return++) {

                proto_tree_add_uint(tree, hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_return, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/notification/IpAppEventNotification/reportNotification:1.0
 */
static void
decode_org_csapi_fw_fw_application_notification_IpAppEventNotification_reportNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin union "org_csapi_fw_TpFwEventInfo"  */
        decode_org_csapi_fw_TpFwEventInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_fw_TpFwEventInfo"  */
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_notification_IpAppEventNotification_reportNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/notification/IpAppEventNotification/notificationTerminated:1.0
 */
static void
decode_org_csapi_fw_fw_application_notification_IpAppEventNotification_notificationTerminated(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/notification/IpEventNotification/createNotification:1.0
 */
static void
decode_org_csapi_fw_fw_application_notification_IpEventNotification_createNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin union "org_csapi_fw_TpFwEventCriteria"  */
        decode_org_csapi_fw_TpFwEventCriteria_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_fw_TpFwEventCriteria"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_notification_IpEventNotification_createNotification_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/notification/IpEventNotification/destroyNotification:1.0
 */
static void
decode_org_csapi_fw_fw_application_notification_IpEventNotification_destroyNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_notification_IpEventNotification_destroyNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppFaultManager/activityTestRes:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_activityTestRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_activityTestRes_activityTestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_activityTestRes_activityTestResult);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppFaultManager/appActivityTestReq:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_appActivityTestReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_appActivityTestReq_activityTestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppFaultManager/fwFaultReportInd:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_fwFaultReportInd(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_fwFaultReportInd_fault, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppFaultManager/fwFaultRecoveryInd:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_fwFaultRecoveryInd(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_fwFaultRecoveryInd_fault, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppFaultManager/svcUnavailableInd:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_svcUnavailableInd(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_svcUnavailableInd_serviceID);

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_svcUnavailableInd_reason, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppFaultManager/genFaultStatsRecordRes:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordRes_serviceIDs;
    guint32   i_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordRes_serviceIDs;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_fw_TpFaultStatsRecord"  */
        decode_org_csapi_fw_TpFaultStatsRecord_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpFaultStatsRecord"  */
        u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordRes_serviceIDs = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordRes_serviceIDs_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordRes_serviceIDs);

        for (i_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordRes_serviceIDs=0; i_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordRes_serviceIDs < u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordRes_serviceIDs; i_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordRes_serviceIDs++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordRes_serviceIDs);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppFaultManager/fwUnavailableInd:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_fwUnavailableInd(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_fwUnavailableInd_reason, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppFaultManager/activityTestErr:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_activityTestErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_activityTestErr_activityTestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppFaultManager/genFaultStatsRecordErr:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    guint32   u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordErr_serviceIDs;
    guint32   i_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordErr_serviceIDs;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordErr_faultStatisticsError, tvb, *offset-4, 4, u_octet4);

        u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordErr_serviceIDs = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordErr_serviceIDs_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordErr_serviceIDs);

        for (i_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordErr_serviceIDs=0; i_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordErr_serviceIDs < u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordErr_serviceIDs; i_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordErr_serviceIDs++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordErr_serviceIDs);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppFaultManager/appUnavailableInd:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_appUnavailableInd(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_appUnavailableInd_serviceID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppFaultManager/genFaultStatsRecordReq:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_TpTimeInterval"  */
        decode_org_csapi_TpTimeInterval_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpTimeInterval"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppFaultManager/svcAvailStatusInd:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_svcAvailStatusInd(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_svcAvailStatusInd_serviceID);

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_svcAvailStatusInd_reason, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppFaultManager/generateFaultStatisticsRecordRes:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordRes_serviceIDs;
    guint32   i_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordRes_serviceIDs;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordRes_faultStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_fw_TpFaultStatsRecord"  */
        decode_org_csapi_fw_TpFaultStatsRecord_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpFaultStatsRecord"  */
        u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordRes_serviceIDs = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordRes_serviceIDs_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordRes_serviceIDs);

        for (i_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordRes_serviceIDs=0; i_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordRes_serviceIDs < u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordRes_serviceIDs; i_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordRes_serviceIDs++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordRes_serviceIDs);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppFaultManager/generateFaultStatisticsRecordErr:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_faultStatistics;
    guint32   i_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_faultStatistics;
    guint32   u_octet4;
    guint32   u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_serviceIDs;
    guint32   i_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_serviceIDs;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_faultStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_faultStatistics = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_faultStatistics_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_faultStatistics);

        for (i_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_faultStatistics=0; i_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_faultStatistics < u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_faultStatistics; i_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_faultStatistics++) {

            u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_faultStatistics, tvb, *offset-4, 4, u_octet4);

        }

        u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_serviceIDs = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_serviceIDs_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_serviceIDs);

        for (i_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_serviceIDs=0; i_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_serviceIDs < u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_serviceIDs; i_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_serviceIDs++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_serviceIDs);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppFaultManager/generateFaultStatisticsRecordReq:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordReq_faultStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_TpTimeInterval"  */
        decode_org_csapi_TpTimeInterval_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpTimeInterval"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppFaultManager/fwAvailStatusInd:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_fwAvailStatusInd(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_fwAvailStatusInd_reason, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppLoadManager/queryAppLoadReq:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryAppLoadReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_TpTimeInterval"  */
        decode_org_csapi_TpTimeInterval_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpTimeInterval"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppLoadManager/queryLoadRes:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadRes_loadStatistics;
    guint32   i_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadRes_loadStatistics;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadRes_loadStatistics = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadRes_loadStatistics_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadRes_loadStatistics);

        for (i_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadRes_loadStatistics=0; i_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadRes_loadStatistics < u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadRes_loadStatistics; i_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadRes_loadStatistics++) {

            /*  Begin struct "org_csapi_fw_TpLoadStatistic"  */
            decode_org_csapi_fw_TpLoadStatistic_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpLoadStatistic"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppLoadManager/queryLoadErr:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadErr_loadStatisticsError, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppLoadManager/loadLevelNotification:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppLoadManager_loadLevelNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppLoadManager_loadLevelNotification_loadStatistics;
    guint32   i_org_csapi_fw_fw_application_integrity_IpAppLoadManager_loadLevelNotification_loadStatistics;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppLoadManager_loadLevelNotification_loadStatistics = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_loadLevelNotification_loadStatistics_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppLoadManager_loadLevelNotification_loadStatistics);

        for (i_org_csapi_fw_fw_application_integrity_IpAppLoadManager_loadLevelNotification_loadStatistics=0; i_org_csapi_fw_fw_application_integrity_IpAppLoadManager_loadLevelNotification_loadStatistics < u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppLoadManager_loadLevelNotification_loadStatistics; i_org_csapi_fw_fw_application_integrity_IpAppLoadManager_loadLevelNotification_loadStatistics++) {

            /*  Begin struct "org_csapi_fw_TpLoadStatistic"  */
            decode_org_csapi_fw_TpLoadStatistic_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpLoadStatistic"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppLoadManager/resumeNotification:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppLoadManager_resumeNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppLoadManager/suspendNotification:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppLoadManager_suspendNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppLoadManager/createLoadLevelNotification:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppLoadManager_createLoadLevelNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppLoadManager/destroyLoadLevelNotification:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppLoadManager_destroyLoadLevelNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppLoadManager/queryAppLoadStatsReq:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryAppLoadStatsReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryAppLoadStatsReq_loadStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_TpTimeInterval"  */
        decode_org_csapi_TpTimeInterval_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpTimeInterval"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppLoadManager/queryLoadStatsRes:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsRes_loadStatistics;
    guint32   i_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsRes_loadStatistics;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsRes_loadStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsRes_loadStatistics = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsRes_loadStatistics_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsRes_loadStatistics);

        for (i_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsRes_loadStatistics=0; i_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsRes_loadStatistics < u_octet4_loop_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsRes_loadStatistics; i_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsRes_loadStatistics++) {

            /*  Begin struct "org_csapi_fw_TpLoadStatistic"  */
            decode_org_csapi_fw_TpLoadStatistic_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpLoadStatistic"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppLoadManager/queryLoadStatsErr:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsErr_loadStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsErr_loadStatisticsError, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpLoadManager/reportLoad:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpLoadManager_reportLoad(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpLoadManager_reportLoad_loadLevel, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpLoadManager/queryLoadReq:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadReq_serviceIDs;
    guint32   i_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadReq_serviceIDs;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadReq_serviceIDs = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadReq_serviceIDs_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadReq_serviceIDs);

        for (i_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadReq_serviceIDs=0; i_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadReq_serviceIDs < u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadReq_serviceIDs; i_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadReq_serviceIDs++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadReq_serviceIDs);

        }

        /*  Begin struct "org_csapi_TpTimeInterval"  */
        decode_org_csapi_TpTimeInterval_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpTimeInterval"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpLoadManager/queryAppLoadRes:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadRes_loadStatistics;
    guint32   i_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadRes_loadStatistics;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadRes_loadStatistics = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadRes_loadStatistics_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadRes_loadStatistics);

        for (i_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadRes_loadStatistics=0; i_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadRes_loadStatistics < u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadRes_loadStatistics; i_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadRes_loadStatistics++) {

            /*  Begin struct "org_csapi_fw_TpLoadStatistic"  */
            decode_org_csapi_fw_TpLoadStatistic_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpLoadStatistic"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpLoadManager/queryAppLoadErr:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadErr_loadStatisticsError, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpLoadManager/createLoadLevelNotification:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpLoadManager_createLoadLevelNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_createLoadLevelNotification_serviceIDs;
    guint32   i_org_csapi_fw_fw_application_integrity_IpLoadManager_createLoadLevelNotification_serviceIDs;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_createLoadLevelNotification_serviceIDs = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpLoadManager_createLoadLevelNotification_serviceIDs_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_createLoadLevelNotification_serviceIDs);

        for (i_org_csapi_fw_fw_application_integrity_IpLoadManager_createLoadLevelNotification_serviceIDs=0; i_org_csapi_fw_fw_application_integrity_IpLoadManager_createLoadLevelNotification_serviceIDs < u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_createLoadLevelNotification_serviceIDs; i_org_csapi_fw_fw_application_integrity_IpLoadManager_createLoadLevelNotification_serviceIDs++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpLoadManager_createLoadLevelNotification_serviceIDs);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpLoadManager/destroyLoadLevelNotification:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpLoadManager_destroyLoadLevelNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_destroyLoadLevelNotification_serviceIDs;
    guint32   i_org_csapi_fw_fw_application_integrity_IpLoadManager_destroyLoadLevelNotification_serviceIDs;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_destroyLoadLevelNotification_serviceIDs = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpLoadManager_destroyLoadLevelNotification_serviceIDs_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_destroyLoadLevelNotification_serviceIDs);

        for (i_org_csapi_fw_fw_application_integrity_IpLoadManager_destroyLoadLevelNotification_serviceIDs=0; i_org_csapi_fw_fw_application_integrity_IpLoadManager_destroyLoadLevelNotification_serviceIDs < u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_destroyLoadLevelNotification_serviceIDs; i_org_csapi_fw_fw_application_integrity_IpLoadManager_destroyLoadLevelNotification_serviceIDs++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpLoadManager_destroyLoadLevelNotification_serviceIDs);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpLoadManager/resumeNotification:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpLoadManager_resumeNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_resumeNotification_serviceIDs;
    guint32   i_org_csapi_fw_fw_application_integrity_IpLoadManager_resumeNotification_serviceIDs;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_resumeNotification_serviceIDs = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpLoadManager_resumeNotification_serviceIDs_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_resumeNotification_serviceIDs);

        for (i_org_csapi_fw_fw_application_integrity_IpLoadManager_resumeNotification_serviceIDs=0; i_org_csapi_fw_fw_application_integrity_IpLoadManager_resumeNotification_serviceIDs < u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_resumeNotification_serviceIDs; i_org_csapi_fw_fw_application_integrity_IpLoadManager_resumeNotification_serviceIDs++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpLoadManager_resumeNotification_serviceIDs);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpLoadManager/suspendNotification:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpLoadManager_suspendNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_suspendNotification_serviceIDs;
    guint32   i_org_csapi_fw_fw_application_integrity_IpLoadManager_suspendNotification_serviceIDs;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_suspendNotification_serviceIDs = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpLoadManager_suspendNotification_serviceIDs_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_suspendNotification_serviceIDs);

        for (i_org_csapi_fw_fw_application_integrity_IpLoadManager_suspendNotification_serviceIDs=0; i_org_csapi_fw_fw_application_integrity_IpLoadManager_suspendNotification_serviceIDs < u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_suspendNotification_serviceIDs; i_org_csapi_fw_fw_application_integrity_IpLoadManager_suspendNotification_serviceIDs++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpLoadManager_suspendNotification_serviceIDs);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpLoadManager/queryLoadStatsReq:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadStatsReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadStatsReq_serviceIDs;
    guint32   i_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadStatsReq_serviceIDs;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadStatsReq_loadStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadStatsReq_serviceIDs = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadStatsReq_serviceIDs_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadStatsReq_serviceIDs);

        for (i_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadStatsReq_serviceIDs=0; i_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadStatsReq_serviceIDs < u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadStatsReq_serviceIDs; i_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadStatsReq_serviceIDs++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadStatsReq_serviceIDs);

        }

        /*  Begin struct "org_csapi_TpTimeInterval"  */
        decode_org_csapi_TpTimeInterval_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpTimeInterval"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpLoadManager/queryAppLoadStatsRes:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsRes_loadStatistics;
    guint32   i_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsRes_loadStatistics;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsRes_loadStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsRes_loadStatistics = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsRes_loadStatistics_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsRes_loadStatistics);

        for (i_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsRes_loadStatistics=0; i_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsRes_loadStatistics < u_octet4_loop_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsRes_loadStatistics; i_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsRes_loadStatistics++) {

            /*  Begin struct "org_csapi_fw_TpLoadStatistic"  */
            decode_org_csapi_fw_TpLoadStatistic_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpLoadStatistic"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpLoadManager/queryAppLoadStatsErr:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsErr_loadStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsErr_loadStatisticsError, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppOAM/systemDateTimeQuery:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppOAM_systemDateTimeQuery(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpAppOAM_systemDateTimeQuery_systemDateAndTime);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpAppOAM_systemDateTimeQuery_return);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpOAM/systemDateTimeQuery:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpOAM_systemDateTimeQuery(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpOAM_systemDateTimeQuery_clientDateAndTime);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpOAM_systemDateTimeQuery_return);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpFaultManager/activityTestReq:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpFaultManager_activityTestReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_integrity_IpFaultManager_activityTestReq_activityTestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpFaultManager_activityTestReq_svcID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpFaultManager/appActivityTestRes:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpFaultManager_appActivityTestRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_integrity_IpFaultManager_appActivityTestRes_activityTestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpFaultManager_appActivityTestRes_activityTestResult);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpFaultManager/svcUnavailableInd:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpFaultManager_svcUnavailableInd(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpFaultManager_svcUnavailableInd_serviceID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpFaultManager/genFaultStatsRecordReq:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpFaultManager_genFaultStatsRecordReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_application_integrity_IpFaultManager_genFaultStatsRecordReq_serviceIDs;
    guint32   i_org_csapi_fw_fw_application_integrity_IpFaultManager_genFaultStatsRecordReq_serviceIDs;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_TpTimeInterval"  */
        decode_org_csapi_TpTimeInterval_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpTimeInterval"  */
        u_octet4_loop_org_csapi_fw_fw_application_integrity_IpFaultManager_genFaultStatsRecordReq_serviceIDs = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpFaultManager_genFaultStatsRecordReq_serviceIDs_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_integrity_IpFaultManager_genFaultStatsRecordReq_serviceIDs);

        for (i_org_csapi_fw_fw_application_integrity_IpFaultManager_genFaultStatsRecordReq_serviceIDs=0; i_org_csapi_fw_fw_application_integrity_IpFaultManager_genFaultStatsRecordReq_serviceIDs < u_octet4_loop_org_csapi_fw_fw_application_integrity_IpFaultManager_genFaultStatsRecordReq_serviceIDs; i_org_csapi_fw_fw_application_integrity_IpFaultManager_genFaultStatsRecordReq_serviceIDs++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpFaultManager_genFaultStatsRecordReq_serviceIDs);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpFaultManager/appActivityTestErr:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpFaultManager_appActivityTestErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_integrity_IpFaultManager_appActivityTestErr_activityTestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpFaultManager/appUnavailableInd:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpFaultManager_appUnavailableInd(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpFaultManager_appUnavailableInd_serviceID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpFaultManager/genFaultStatsRecordRes:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpFaultManager_genFaultStatsRecordRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_fw_TpFaultStatsRecord"  */
        decode_org_csapi_fw_TpFaultStatsRecord_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpFaultStatsRecord"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpFaultManager/genFaultStatsRecordErr:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpFaultManager_genFaultStatsRecordErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpFaultManager_genFaultStatsRecordErr_faultStatisticsError, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpFaultManager/appAvailStatusInd:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpFaultManager_appAvailStatusInd(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpFaultManager_appAvailStatusInd_reason, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpFaultManager/generateFaultStatisticsRecordReq:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordReq_serviceIDs;
    guint32   i_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordReq_serviceIDs;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordReq_faultStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_TpTimeInterval"  */
        decode_org_csapi_TpTimeInterval_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpTimeInterval"  */
        u_octet4_loop_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordReq_serviceIDs = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordReq_serviceIDs_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordReq_serviceIDs);

        for (i_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordReq_serviceIDs=0; i_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordReq_serviceIDs < u_octet4_loop_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordReq_serviceIDs; i_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordReq_serviceIDs++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordReq_serviceIDs);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpFaultManager/generateFaultStatisticsRecordRes:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordRes_faultStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_fw_TpFaultStatsRecord"  */
        decode_org_csapi_fw_TpFaultStatsRecord_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpFaultStatsRecord"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpFaultManager/generateFaultStatisticsRecordErr:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordErr_faultStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordErr_faultStatisticsError, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpHeartBeatMgmt/enableHeartBeat:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpHeartBeatMgmt_enableHeartBeat(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_integrity_IpHeartBeatMgmt_enableHeartBeat_interval, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpHeartBeatMgmt/disableHeartBeat:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpHeartBeatMgmt_disableHeartBeat(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpHeartBeatMgmt/changeInterval:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpHeartBeatMgmt_changeInterval(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_integrity_IpHeartBeatMgmt_changeInterval_interval, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppHeartBeat/pulse:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppHeartBeat_pulse(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpHeartBeat/pulse:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpHeartBeat_pulse(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppHeartBeatMgmt/enableAppHeartBeat:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppHeartBeatMgmt_enableAppHeartBeat(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_integrity_IpAppHeartBeatMgmt_enableAppHeartBeat_interval, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppHeartBeatMgmt/disableAppHeartBeat:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppHeartBeatMgmt_disableAppHeartBeat(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/integrity/IpAppHeartBeatMgmt/changeInterval:1.0
 */
static void
decode_org_csapi_fw_fw_application_integrity_IpAppHeartBeatMgmt_changeInterval(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_integrity_IpAppHeartBeatMgmt_changeInterval_interval, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/discovery/IpServiceDiscovery/listServiceTypes:1.0
 */
static void
decode_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listServiceTypes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listServiceTypes_return;
    guint32   i_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listServiceTypes_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listServiceTypes_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listServiceTypes_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listServiceTypes_return);

            for (i_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listServiceTypes_return=0; i_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listServiceTypes_return < u_octet4_loop_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listServiceTypes_return; i_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listServiceTypes_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listServiceTypes_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/discovery/IpServiceDiscovery/describeServiceType:1.0
 */
static void
decode_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_describeServiceType(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_describeServiceType_name);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_fw_TpServiceTypeDescription"  */
            decode_org_csapi_fw_TpServiceTypeDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpServiceTypeDescription"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/discovery/IpServiceDiscovery/discoverService:1.0
 */
static void
decode_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_desiredPropertyList;
    guint32   i_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_desiredPropertyList;
    guint32   u_octet4_loop_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_return;
    guint32   i_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_serviceTypeName);

        u_octet4_loop_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_desiredPropertyList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_desiredPropertyList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_desiredPropertyList);

        for (i_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_desiredPropertyList=0; i_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_desiredPropertyList < u_octet4_loop_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_desiredPropertyList; i_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_desiredPropertyList++) {

            /*  Begin struct "org_csapi_fw_TpServiceProperty"  */
            decode_org_csapi_fw_TpServiceProperty_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpServiceProperty"  */
        }

        proto_tree_add_int(tree, hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_max, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_return);

            for (i_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_return=0; i_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_return < u_octet4_loop_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_return; i_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_return++) {

                /*  Begin struct "org_csapi_fw_TpService"  */
                decode_org_csapi_fw_TpService_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_fw_TpService"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/discovery/IpServiceDiscovery/listSubscribedServices:1.0
 */
static void
decode_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listSubscribedServices(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listSubscribedServices_return;
    guint32   i_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listSubscribedServices_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listSubscribedServices_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listSubscribedServices_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listSubscribedServices_return);

            for (i_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listSubscribedServices_return=0; i_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listSubscribedServices_return < u_octet4_loop_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listSubscribedServices_return; i_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listSubscribedServices_return++) {

                /*  Begin struct "org_csapi_fw_TpService"  */
                decode_org_csapi_fw_TpService_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_fw_TpService"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/service_agreement/IpAppServiceAgreementManagement/signServiceAgreement:1.0
 */
static void
decode_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement_return;
    guint32   i_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement_serviceToken);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement_agreementText);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement_signingAlgorithm);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement_return);

            for (i_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement_return=0; i_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement_return < u_octet4_loop_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement_return; i_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement_return++) {

                proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement_return, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/service_agreement/IpAppServiceAgreementManagement/terminateServiceAgreement:1.0
 */
static void
decode_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_terminateServiceAgreement(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_terminateServiceAgreement_digitalSignature;
    guint32   i_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_terminateServiceAgreement_digitalSignature;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_terminateServiceAgreement_serviceToken);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_terminateServiceAgreement_terminationText);

        u_octet4_loop_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_terminateServiceAgreement_digitalSignature = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_terminateServiceAgreement_digitalSignature_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_terminateServiceAgreement_digitalSignature);

        for (i_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_terminateServiceAgreement_digitalSignature=0; i_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_terminateServiceAgreement_digitalSignature < u_octet4_loop_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_terminateServiceAgreement_digitalSignature; i_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_terminateServiceAgreement_digitalSignature++) {

            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_terminateServiceAgreement_digitalSignature, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/service_agreement/IpServiceAgreementManagement/signServiceAgreement:1.0
 */
static void
decode_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_signServiceAgreement(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_signServiceAgreement_serviceToken);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_signServiceAgreement_agreementText);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_signServiceAgreement_signingAlgorithm);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_fw_TpSignatureAndServiceMgr"  */
            decode_org_csapi_fw_TpSignatureAndServiceMgr_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpSignatureAndServiceMgr"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/service_agreement/IpServiceAgreementManagement/terminateServiceAgreement:1.0
 */
static void
decode_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_terminateServiceAgreement(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_terminateServiceAgreement_digitalSignature;
    guint32   i_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_terminateServiceAgreement_digitalSignature;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_terminateServiceAgreement_serviceToken);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_terminateServiceAgreement_terminationText);

        u_octet4_loop_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_terminateServiceAgreement_digitalSignature = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_terminateServiceAgreement_digitalSignature_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_terminateServiceAgreement_digitalSignature);

        for (i_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_terminateServiceAgreement_digitalSignature=0; i_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_terminateServiceAgreement_digitalSignature < u_octet4_loop_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_terminateServiceAgreement_digitalSignature; i_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_terminateServiceAgreement_digitalSignature++) {

            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_terminateServiceAgreement_digitalSignature, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/service_agreement/IpServiceAgreementManagement/selectService:1.0
 */
static void
decode_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_selectService(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_selectService_serviceID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_selectService_return);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_application/service_agreement/IpServiceAgreementManagement/initiateSignServiceAgreement:1.0
 */
static void
decode_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_initiateSignServiceAgreement(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_initiateSignServiceAgreement_serviceToken);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceProfileManagement/createServiceProfile:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_createServiceProfile(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_fw_TpServiceProfileDescription"  */
        decode_org_csapi_fw_TpServiceProfileDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpServiceProfileDescription"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_createServiceProfile_return);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceProfileManagement/modifyServiceProfile:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_modifyServiceProfile(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_fw_TpServiceProfile"  */
        decode_org_csapi_fw_TpServiceProfile_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpServiceProfile"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceProfileManagement/deleteServiceProfile:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_deleteServiceProfile(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_deleteServiceProfile_serviceProfileID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceProfileManagement/assign:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_assign(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_assign_sagID);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_assign_serviceProfileID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceProfileManagement/deassign:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_deassign(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_deassign_sagID);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_deassign_serviceProfileID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceProfileManagement/requestConflictInfo:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_requestConflictInfo(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_requestConflictInfo_return;
    guint32   i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_requestConflictInfo_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_requestConflictInfo_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_requestConflictInfo_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_requestConflictInfo_return);

            for (i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_requestConflictInfo_return=0; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_requestConflictInfo_return < u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_requestConflictInfo_return; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_requestConflictInfo_return++) {

                /*  Begin struct "org_csapi_fw_TpAssignSagToServiceProfileConflict"  */
                decode_org_csapi_fw_TpAssignSagToServiceProfileConflict_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_fw_TpAssignSagToServiceProfileConflict"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceProfileInfoQuery/listServiceProfiles:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listServiceProfiles(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listServiceProfiles_return;
    guint32   i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listServiceProfiles_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listServiceProfiles_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listServiceProfiles_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listServiceProfiles_return);

            for (i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listServiceProfiles_return=0; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listServiceProfiles_return < u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listServiceProfiles_return; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listServiceProfiles_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listServiceProfiles_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceProfileInfoQuery/describeServiceProfile:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_describeServiceProfile(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_describeServiceProfile_serviceProfileID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_fw_TpServiceProfileDescription"  */
            decode_org_csapi_fw_TpServiceProfileDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpServiceProfileDescription"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceProfileInfoQuery/listAssignedMembers:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listAssignedMembers(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listAssignedMembers_return;
    guint32   i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listAssignedMembers_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listAssignedMembers_serviceProfileID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listAssignedMembers_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listAssignedMembers_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listAssignedMembers_return);

            for (i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listAssignedMembers_return=0; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listAssignedMembers_return < u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listAssignedMembers_return; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listAssignedMembers_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listAssignedMembers_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceContractManagement/createServiceContract:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractManagement_createServiceContract(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_fw_TpServiceContractDescription"  */
        decode_org_csapi_fw_TpServiceContractDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpServiceContractDescription"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractManagement_createServiceContract_return);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceContractManagement/modifyServiceContract:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractManagement_modifyServiceContract(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_fw_TpServiceContract"  */
        decode_org_csapi_fw_TpServiceContract_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpServiceContract"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceContractManagement/deleteServiceContract:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractManagement_deleteServiceContract(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractManagement_deleteServiceContract_serviceContractID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceContractInfoQuery/describeServiceContract:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_describeServiceContract(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_describeServiceContract_serviceContractID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_fw_TpServiceContractDescription"  */
            decode_org_csapi_fw_TpServiceContractDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpServiceContractDescription"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceContractInfoQuery/listServiceContracts:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceContracts(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceContracts_return;
    guint32   i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceContracts_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceContracts_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceContracts_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceContracts_return);

            for (i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceContracts_return=0; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceContracts_return < u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceContracts_return; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceContracts_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceContracts_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceContractInfoQuery/listServiceProfiles:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceProfiles(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceProfiles_return;
    guint32   i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceProfiles_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceProfiles_serviceContractID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceProfiles_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceProfiles_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceProfiles_return);

            for (i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceProfiles_return=0; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceProfiles_return < u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceProfiles_return; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceProfiles_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceProfiles_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpEntOpAccountManagement/modifyEntOpAccount:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpEntOpAccountManagement_modifyEntOpAccount(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpEntOpAccountManagement_modifyEntOpAccount_enterpriseOperatorProperties;
    guint32   i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpEntOpAccountManagement_modifyEntOpAccount_enterpriseOperatorProperties;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpEntOpAccountManagement_modifyEntOpAccount_enterpriseOperatorProperties = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpEntOpAccountManagement_modifyEntOpAccount_enterpriseOperatorProperties_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpEntOpAccountManagement_modifyEntOpAccount_enterpriseOperatorProperties);

        for (i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpEntOpAccountManagement_modifyEntOpAccount_enterpriseOperatorProperties=0; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpEntOpAccountManagement_modifyEntOpAccount_enterpriseOperatorProperties < u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpEntOpAccountManagement_modifyEntOpAccount_enterpriseOperatorProperties; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpEntOpAccountManagement_modifyEntOpAccount_enterpriseOperatorProperties++) {

            /*  Begin struct "org_csapi_fw_TpProperty"  */
            decode_org_csapi_fw_TpProperty_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpProperty"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpEntOpAccountManagement/deleteEntOpAccount:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpEntOpAccountManagement_deleteEntOpAccount(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpEntOpAccountInfoQuery/describeEntOpAccount:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpEntOpAccountInfoQuery_describeEntOpAccount(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_fw_TpEntOp"  */
            decode_org_csapi_fw_TpEntOp_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpEntOp"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppManagement/createClientApp:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_createClientApp(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_fw_TpClientAppDescription"  */
        decode_org_csapi_fw_TpClientAppDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpClientAppDescription"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppManagement/modifyClientApp:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_modifyClientApp(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_fw_TpClientAppDescription"  */
        decode_org_csapi_fw_TpClientAppDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpClientAppDescription"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppManagement/deleteClientApp:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_deleteClientApp(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_deleteClientApp_clientAppID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppManagement/createSAG:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_createSAG(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_createSAG_clientAppIDs;
    guint32   i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_createSAG_clientAppIDs;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_fw_TpSag"  */
        decode_org_csapi_fw_TpSag_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpSag"  */
        u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_createSAG_clientAppIDs = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_createSAG_clientAppIDs_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_createSAG_clientAppIDs);

        for (i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_createSAG_clientAppIDs=0; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_createSAG_clientAppIDs < u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_createSAG_clientAppIDs; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_createSAG_clientAppIDs++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_createSAG_clientAppIDs);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppManagement/modifySAG:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_modifySAG(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_fw_TpSag"  */
        decode_org_csapi_fw_TpSag_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpSag"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppManagement/deleteSAG:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_deleteSAG(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_deleteSAG_sagID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppManagement/addSAGMembers:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_addSAGMembers(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_addSAGMembers_clientAppIDs;
    guint32   i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_addSAGMembers_clientAppIDs;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_addSAGMembers_sagID);

        u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_addSAGMembers_clientAppIDs = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_addSAGMembers_clientAppIDs_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_addSAGMembers_clientAppIDs);

        for (i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_addSAGMembers_clientAppIDs=0; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_addSAGMembers_clientAppIDs < u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_addSAGMembers_clientAppIDs; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_addSAGMembers_clientAppIDs++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_addSAGMembers_clientAppIDs);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppManagement/removeSAGMembers:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_removeSAGMembers(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_removeSAGMembers_clientAppIDList;
    guint32   i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_removeSAGMembers_clientAppIDList;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_removeSAGMembers_sagID);

        u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_removeSAGMembers_clientAppIDList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_removeSAGMembers_clientAppIDList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_removeSAGMembers_clientAppIDList);

        for (i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_removeSAGMembers_clientAppIDList=0; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_removeSAGMembers_clientAppIDList < u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_removeSAGMembers_clientAppIDList; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_removeSAGMembers_clientAppIDList++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_removeSAGMembers_clientAppIDList);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppManagement/requestConflictInfo:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_requestConflictInfo(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_requestConflictInfo_return;
    guint32   i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_requestConflictInfo_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_requestConflictInfo_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_requestConflictInfo_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_requestConflictInfo_return);

            for (i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_requestConflictInfo_return=0; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_requestConflictInfo_return < u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_requestConflictInfo_return; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_requestConflictInfo_return++) {

                /*  Begin struct "org_csapi_fw_TpAddSagMembersConflict"  */
                decode_org_csapi_fw_TpAddSagMembersConflict_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_fw_TpAddSagMembersConflict"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppInfoQuery/describeClientApp:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_describeClientApp(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_describeClientApp_clientAppID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_fw_TpClientAppDescription"  */
            decode_org_csapi_fw_TpClientAppDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpClientAppDescription"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppInfoQuery/listClientApps:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientApps(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientApps_return;
    guint32   i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientApps_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientApps_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientApps_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientApps_return);

            for (i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientApps_return=0; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientApps_return < u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientApps_return; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientApps_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientApps_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppInfoQuery/describeSAG:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_describeSAG(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_describeSAG_sagID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_describeSAG_return);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppInfoQuery/listSAGs:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGs(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGs_return;
    guint32   i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGs_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGs_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGs_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGs_return);

            for (i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGs_return=0; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGs_return < u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGs_return; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGs_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGs_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppInfoQuery/listSAGMembers:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGMembers(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGMembers_return;
    guint32   i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGMembers_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGMembers_sagID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGMembers_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGMembers_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGMembers_return);

            for (i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGMembers_return=0; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGMembers_return < u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGMembers_return; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGMembers_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGMembers_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppInfoQuery/listClientAppMembership:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientAppMembership(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientAppMembership_return;
    guint32   i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientAppMembership_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientAppMembership_clientAppID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientAppMembership_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientAppMembership_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientAppMembership_return);

            for (i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientAppMembership_return=0; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientAppMembership_return < u_octet4_loop_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientAppMembership_return; i_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientAppMembership_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientAppMembership_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/notification/IpClientEventNotification/reportNotification:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_notification_IpClientEventNotification_reportNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin union "org_csapi_fw_TpFwEventInfo"  */
        decode_org_csapi_fw_TpFwEventInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_fw_TpFwEventInfo"  */
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_enterprise_operator_notification_IpClientEventNotification_reportNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/notification/IpClientEventNotification/notificationTerminated:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_notification_IpClientEventNotification_notificationTerminated(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/notification/IpEventNotification/createNotification:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_notification_IpEventNotification_createNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin union "org_csapi_fw_TpFwEventCriteria"  */
        decode_org_csapi_fw_TpFwEventCriteria_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_fw_TpFwEventCriteria"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_fw_fw_enterprise_operator_notification_IpEventNotification_createNotification_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_enterprise_operator/notification/IpEventNotification/destroyNotification:1.0
 */
static void
decode_org_csapi_fw_fw_enterprise_operator_notification_IpEventNotification_destroyNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_enterprise_operator_notification_IpEventNotification_destroyNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/discovery/IpFwServiceDiscovery/listServiceTypes:1.0
 */
static void
decode_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listServiceTypes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listServiceTypes_return;
    guint32   i_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listServiceTypes_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listServiceTypes_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listServiceTypes_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listServiceTypes_return);

            for (i_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listServiceTypes_return=0; i_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listServiceTypes_return < u_octet4_loop_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listServiceTypes_return; i_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listServiceTypes_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listServiceTypes_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/discovery/IpFwServiceDiscovery/describeServiceType:1.0
 */
static void
decode_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_describeServiceType(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_describeServiceType_name);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_fw_TpServiceTypeDescription"  */
            decode_org_csapi_fw_TpServiceTypeDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpServiceTypeDescription"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/discovery/IpFwServiceDiscovery/discoverService:1.0
 */
static void
decode_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_desiredPropertyList;
    guint32   i_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_desiredPropertyList;
    guint32   u_octet4_loop_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_return;
    guint32   i_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_serviceTypeName);

        u_octet4_loop_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_desiredPropertyList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_desiredPropertyList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_desiredPropertyList);

        for (i_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_desiredPropertyList=0; i_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_desiredPropertyList < u_octet4_loop_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_desiredPropertyList; i_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_desiredPropertyList++) {

            /*  Begin struct "org_csapi_fw_TpServiceProperty"  */
            decode_org_csapi_fw_TpServiceProperty_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpServiceProperty"  */
        }

        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_max, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_return);

            for (i_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_return=0; i_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_return < u_octet4_loop_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_return; i_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_return++) {

                /*  Begin struct "org_csapi_fw_TpService"  */
                decode_org_csapi_fw_TpService_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_fw_TpService"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/discovery/IpFwServiceDiscovery/listRegisteredServices:1.0
 */
static void
decode_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listRegisteredServices(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listRegisteredServices_return;
    guint32   i_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listRegisteredServices_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listRegisteredServices_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listRegisteredServices_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listRegisteredServices_return);

            for (i_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listRegisteredServices_return=0; i_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listRegisteredServices_return < u_octet4_loop_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listRegisteredServices_return; i_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listRegisteredServices_return++) {

                /*  Begin struct "org_csapi_fw_TpService"  */
                decode_org_csapi_fw_TpService_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_fw_TpService"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/service_lifecycle/IpServiceInstanceLifecycleManager/createServiceManager:1.0
 */
static void
decode_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_createServiceManager(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_createServiceManager_serviceProperties;
    guint32   i_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_createServiceManager_serviceProperties;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_createServiceManager_application);

        u_octet4_loop_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_createServiceManager_serviceProperties = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_createServiceManager_serviceProperties_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_createServiceManager_serviceProperties);

        for (i_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_createServiceManager_serviceProperties=0; i_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_createServiceManager_serviceProperties < u_octet4_loop_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_createServiceManager_serviceProperties; i_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_createServiceManager_serviceProperties++) {

            /*  Begin struct "org_csapi_fw_TpServiceProperty"  */
            decode_org_csapi_fw_TpServiceProperty_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpServiceProperty"  */
        }

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_createServiceManager_serviceInstanceID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/service_lifecycle/IpServiceInstanceLifecycleManager/destroyServiceManager:1.0
 */
static void
decode_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_destroyServiceManager(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_destroyServiceManager_serviceInstance);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/service_registration/IpFwServiceRegistration/registerService:1.0
 */
static void
decode_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerService(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerService_servicePropertyList;
    guint32   i_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerService_servicePropertyList;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerService_serviceTypeName);

        u_octet4_loop_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerService_servicePropertyList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerService_servicePropertyList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerService_servicePropertyList);

        for (i_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerService_servicePropertyList=0; i_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerService_servicePropertyList < u_octet4_loop_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerService_servicePropertyList; i_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerService_servicePropertyList++) {

            /*  Begin struct "org_csapi_fw_TpServiceProperty"  */
            decode_org_csapi_fw_TpServiceProperty_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpServiceProperty"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerService_return);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/service_registration/IpFwServiceRegistration/announceServiceAvailability:1.0
 */
static void
decode_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_announceServiceAvailability(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_announceServiceAvailability_serviceID);

        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/service_registration/IpFwServiceRegistration/unregisterService:1.0
 */
static void
decode_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_unregisterService(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_unregisterService_serviceID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/service_registration/IpFwServiceRegistration/describeService:1.0
 */
static void
decode_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_describeService(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_describeService_serviceID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_fw_TpServiceDescription"  */
            decode_org_csapi_fw_TpServiceDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpServiceDescription"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/service_registration/IpFwServiceRegistration/unannounceService:1.0
 */
static void
decode_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_unannounceService(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_unannounceService_serviceID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/service_registration/IpFwServiceRegistration/registerServiceSubType:1.0
 */
static void
decode_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_servicePropertyList;
    guint32   i_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_servicePropertyList;
    guint32   u_octet4_loop_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_extendedServicePropertyList;
    guint32   i_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_extendedServicePropertyList;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_serviceTypeName);

        u_octet4_loop_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_servicePropertyList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_servicePropertyList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_servicePropertyList);

        for (i_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_servicePropertyList=0; i_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_servicePropertyList < u_octet4_loop_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_servicePropertyList; i_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_servicePropertyList++) {

            /*  Begin struct "org_csapi_fw_TpServiceProperty"  */
            decode_org_csapi_fw_TpServiceProperty_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpServiceProperty"  */
        }

        u_octet4_loop_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_extendedServicePropertyList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_extendedServicePropertyList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_extendedServicePropertyList);

        for (i_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_extendedServicePropertyList=0; i_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_extendedServicePropertyList < u_octet4_loop_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_extendedServicePropertyList; i_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_extendedServicePropertyList++) {

            /*  Begin struct "org_csapi_fw_TpServiceTypePropertyValue"  */
            decode_org_csapi_fw_TpServiceTypePropertyValue_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpServiceTypePropertyValue"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_return);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcLoadManager/querySvcLoadReq:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_querySvcLoadReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_TpTimeInterval"  */
        decode_org_csapi_TpTimeInterval_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpTimeInterval"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcLoadManager/queryLoadRes:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadRes_loadStatistics;
    guint32   i_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadRes_loadStatistics;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadRes_loadStatistics = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadRes_loadStatistics_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadRes_loadStatistics);

        for (i_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadRes_loadStatistics=0; i_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadRes_loadStatistics < u_octet4_loop_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadRes_loadStatistics; i_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadRes_loadStatistics++) {

            /*  Begin struct "org_csapi_fw_TpLoadStatistic"  */
            decode_org_csapi_fw_TpLoadStatistic_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpLoadStatistic"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcLoadManager/queryLoadErr:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadErr_loadStatisticsError, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcLoadManager/loadLevelNotification:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_loadLevelNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_loadLevelNotification_loadStatistics;
    guint32   i_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_loadLevelNotification_loadStatistics;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_loadLevelNotification_loadStatistics = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_loadLevelNotification_loadStatistics_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_loadLevelNotification_loadStatistics);

        for (i_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_loadLevelNotification_loadStatistics=0; i_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_loadLevelNotification_loadStatistics < u_octet4_loop_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_loadLevelNotification_loadStatistics; i_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_loadLevelNotification_loadStatistics++) {

            /*  Begin struct "org_csapi_fw_TpLoadStatistic"  */
            decode_org_csapi_fw_TpLoadStatistic_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpLoadStatistic"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcLoadManager/suspendNotification:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_suspendNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcLoadManager/resumeNotification:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_resumeNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcLoadManager/createLoadLevelNotification:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_createLoadLevelNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcLoadManager/destroyLoadLevelNotification:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_destroyLoadLevelNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcLoadManager/querySvcLoadStatsReq:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_querySvcLoadStatsReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_querySvcLoadStatsReq_loadStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_TpTimeInterval"  */
        decode_org_csapi_TpTimeInterval_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpTimeInterval"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcLoadManager/queryLoadStatsRes:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsRes_loadStatistics;
    guint32   i_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsRes_loadStatistics;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsRes_loadStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsRes_loadStatistics = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsRes_loadStatistics_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsRes_loadStatistics);

        for (i_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsRes_loadStatistics=0; i_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsRes_loadStatistics < u_octet4_loop_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsRes_loadStatistics; i_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsRes_loadStatistics++) {

            /*  Begin struct "org_csapi_fw_TpLoadStatistic"  */
            decode_org_csapi_fw_TpLoadStatistic_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpLoadStatistic"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcLoadManager/queryLoadStatsErr:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsErr_loadStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsErr_loadStatisticsError, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwLoadManager/reportLoad:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwLoadManager_reportLoad(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_reportLoad_loadLevel, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwLoadManager/queryLoadReq:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwLoadManager_queryLoadReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_queryLoadReq_querySubject, tvb, *offset-4, 4, u_octet4);

        /*  Begin struct "org_csapi_TpTimeInterval"  */
        decode_org_csapi_TpTimeInterval_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpTimeInterval"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwLoadManager/querySvcLoadRes:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadRes_loadStatistics;
    guint32   i_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadRes_loadStatistics;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadRes_loadStatistics = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadRes_loadStatistics_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadRes_loadStatistics);

        for (i_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadRes_loadStatistics=0; i_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadRes_loadStatistics < u_octet4_loop_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadRes_loadStatistics; i_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadRes_loadStatistics++) {

            /*  Begin struct "org_csapi_fw_TpLoadStatistic"  */
            decode_org_csapi_fw_TpLoadStatistic_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpLoadStatistic"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwLoadManager/querySvcLoadErr:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadErr_loadStatisticError, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwLoadManager/createLoadLevelNotification:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwLoadManager_createLoadLevelNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_createLoadLevelNotification_notificationSubject, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwLoadManager/destroyLoadLevelNotification:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwLoadManager_destroyLoadLevelNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_destroyLoadLevelNotification_notificationSubject, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwLoadManager/suspendNotification:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwLoadManager_suspendNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_suspendNotification_notificationSubject, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwLoadManager/resumeNotification:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwLoadManager_resumeNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_resumeNotification_notificationSubject, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwLoadManager/queryLoadStatsReq:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwLoadManager_queryLoadStatsReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_queryLoadStatsReq_loadStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_queryLoadStatsReq_querySubject, tvb, *offset-4, 4, u_octet4);

        /*  Begin struct "org_csapi_TpTimeInterval"  */
        decode_org_csapi_TpTimeInterval_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpTimeInterval"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwLoadManager/querySvcLoadStatsRes:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsRes_loadStatistics;
    guint32   i_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsRes_loadStatistics;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsRes_loadStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsRes_loadStatistics = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsRes_loadStatistics_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsRes_loadStatistics);

        for (i_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsRes_loadStatistics=0; i_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsRes_loadStatistics < u_octet4_loop_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsRes_loadStatistics; i_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsRes_loadStatistics++) {

            /*  Begin struct "org_csapi_fw_TpLoadStatistic"  */
            decode_org_csapi_fw_TpLoadStatistic_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_fw_TpLoadStatistic"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwLoadManager/querySvcLoadStatsErr:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsErr_loadStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsErr_loadStatisticError, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcFaultManager/activityTestRes:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_activityTestRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_activityTestRes_activityTestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_activityTestRes_activityTestResult);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcFaultManager/svcActivityTestReq:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_svcActivityTestReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_svcActivityTestReq_activityTestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcFaultManager/fwFaultReportInd:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_fwFaultReportInd(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_fwFaultReportInd_fault, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcFaultManager/fwFaultRecoveryInd:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_fwFaultRecoveryInd(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_fwFaultRecoveryInd_fault, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcFaultManager/fwUnavailableInd:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_fwUnavailableInd(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_fwUnavailableInd_reason, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcFaultManager/svcUnavailableInd:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_svcUnavailableInd(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcFaultManager/appUnavailableInd:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_appUnavailableInd(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcFaultManager/genFaultStatsRecordRes:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_fw_TpFaultStatsRecord"  */
        decode_org_csapi_fw_TpFaultStatsRecord_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpFaultStatsRecord"  */
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordRes_recordSubject, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcFaultManager/activityTestErr:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_activityTestErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_activityTestErr_activityTestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcFaultManager/genFaultStatsRecordErr:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordErr_faultStatisticsError, tvb, *offset-4, 4, u_octet4);

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordErr_recordSubject, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcFaultManager/genFaultStatsRecordReq:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordReq_serviceIDs;
    guint32   i_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordReq_serviceIDs;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_TpTimeInterval"  */
        decode_org_csapi_TpTimeInterval_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpTimeInterval"  */
        u_octet4_loop_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordReq_serviceIDs = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordReq_serviceIDs_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordReq_serviceIDs);

        for (i_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordReq_serviceIDs=0; i_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordReq_serviceIDs < u_octet4_loop_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordReq_serviceIDs; i_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordReq_serviceIDs++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordReq_serviceIDs);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcFaultManager/generateFaultStatsRecordReq:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatsRecordReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_TpTimeInterval"  */
        decode_org_csapi_TpTimeInterval_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpTimeInterval"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcFaultManager/appAvailStatusInd:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_appAvailStatusInd(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_appAvailStatusInd_reason, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcFaultManager/generateFaultStatisticsRecordRes:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordRes_faultStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_fw_TpFaultStatsRecord"  */
        decode_org_csapi_fw_TpFaultStatsRecord_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpFaultStatsRecord"  */
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordRes_recordSubject, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcFaultManager/generateFaultStatisticsRecordErr:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordErr_faultStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordErr_faultStatisticsError, tvb, *offset-4, 4, u_octet4);

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordErr_recordSubject, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcFaultManager/generateFaultStatisticsRecordReq:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordReq_faultStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_TpTimeInterval"  */
        decode_org_csapi_TpTimeInterval_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpTimeInterval"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcFaultManager/fwAvailStatusInd:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_fwAvailStatusInd(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_fwAvailStatusInd_reason, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwFaultManager/activityTestReq:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_activityTestReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_activityTestReq_activityTestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_activityTestReq_testSubject, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwFaultManager/svcActivityTestRes:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_svcActivityTestRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_svcActivityTestRes_activityTestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_svcActivityTestRes_activityTestResult);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwFaultManager/appUnavailableInd:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_appUnavailableInd(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwFaultManager/genFaultStatsRecordReq:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_TpTimeInterval"  */
        decode_org_csapi_TpTimeInterval_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpTimeInterval"  */
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordReq_recordSubject, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwFaultManager/svcUnavailableInd:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_svcUnavailableInd(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_svcUnavailableInd_reason, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwFaultManager/svcActivityTestErr:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_svcActivityTestErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_svcActivityTestErr_activityTestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwFaultManager/genFaultStatsRecordRes:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordRes_serviceIDs;
    guint32   i_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordRes_serviceIDs;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_fw_TpFaultStatsRecord"  */
        decode_org_csapi_fw_TpFaultStatsRecord_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpFaultStatsRecord"  */
        u_octet4_loop_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordRes_serviceIDs = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordRes_serviceIDs_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordRes_serviceIDs);

        for (i_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordRes_serviceIDs=0; i_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordRes_serviceIDs < u_octet4_loop_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordRes_serviceIDs; i_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordRes_serviceIDs++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordRes_serviceIDs);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwFaultManager/genFaultStatsRecordErr:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    guint32   u_octet4_loop_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordErr_serviceIDs;
    guint32   i_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordErr_serviceIDs;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordErr_faultStatisticsError, tvb, *offset-4, 4, u_octet4);

        u_octet4_loop_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordErr_serviceIDs = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordErr_serviceIDs_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordErr_serviceIDs);

        for (i_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordErr_serviceIDs=0; i_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordErr_serviceIDs < u_octet4_loop_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordErr_serviceIDs; i_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordErr_serviceIDs++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordErr_serviceIDs);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwFaultManager/generateFaultStatsRecordRes:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatsRecordRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_fw_TpFaultStatsRecord"  */
        decode_org_csapi_fw_TpFaultStatsRecord_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpFaultStatsRecord"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwFaultManager/generateFaultStatsRecordErr:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatsRecordErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatsRecordErr_faultStatisticsError, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwFaultManager/svcAvailStatusInd:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_svcAvailStatusInd(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_svcAvailStatusInd_reason, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwFaultManager/generateFaultStatisticsRecordReq:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatisticsRecordReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatisticsRecordReq_faultStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_TpTimeInterval"  */
        decode_org_csapi_TpTimeInterval_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpTimeInterval"  */
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatisticsRecordReq_recordSubject, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwFaultManager/generateFaultStatisticsRecordRes:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatisticsRecordRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatisticsRecordRes_faultStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_fw_TpFaultStatsRecord"  */
        decode_org_csapi_fw_TpFaultStatsRecord_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpFaultStatsRecord"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwFaultManager/generateFaultStatisticsRecordErr:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatisticsRecordErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatisticsRecordErr_faultStatsReqID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatisticsRecordErr_faultStatisticsError, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcOAM/systemDateTimeQuery:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcOAM_systemDateTimeQuery(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_service_integrity_IpSvcOAM_systemDateTimeQuery_systemDateAndTime);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_service_integrity_IpSvcOAM_systemDateTimeQuery_return);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwOAM/systemDateTimeQuery:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwOAM_systemDateTimeQuery(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_service_integrity_IpFwOAM_systemDateTimeQuery_clientDateAndTime);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_fw_service_integrity_IpFwOAM_systemDateTimeQuery_return);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwHeartBeatMgmt/enableHeartBeat:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwHeartBeatMgmt_enableHeartBeat(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_integrity_IpFwHeartBeatMgmt_enableHeartBeat_interval, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwHeartBeatMgmt/disableHeartBeat:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwHeartBeatMgmt_disableHeartBeat(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwHeartBeatMgmt/changeInterval:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwHeartBeatMgmt_changeInterval(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_integrity_IpFwHeartBeatMgmt_changeInterval_interval, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcHeartBeat/pulse:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcHeartBeat_pulse(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpFwHeartBeat/pulse:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpFwHeartBeat_pulse(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcHeartBeatMgmt/enableSvcHeartBeat:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcHeartBeatMgmt_enableSvcHeartBeat(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcHeartBeatMgmt_enableSvcHeartBeat_interval, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcHeartBeatMgmt/disableSvcHeartBeat:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcHeartBeatMgmt_disableSvcHeartBeat(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/integrity/IpSvcHeartBeatMgmt/changeInterval:1.0
 */
static void
decode_org_csapi_fw_fw_service_integrity_IpSvcHeartBeatMgmt_changeInterval(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_integrity_IpSvcHeartBeatMgmt_changeInterval_interval, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/notification/IpFwEventNotification/createNotification:1.0
 */
static void
decode_org_csapi_fw_fw_service_notification_IpFwEventNotification_createNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin union "org_csapi_fw_TpFwEventCriteria"  */
        decode_org_csapi_fw_TpFwEventCriteria_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_fw_TpFwEventCriteria"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_notification_IpFwEventNotification_createNotification_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/notification/IpFwEventNotification/destroyNotification:1.0
 */
static void
decode_org_csapi_fw_fw_service_notification_IpFwEventNotification_destroyNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_notification_IpFwEventNotification_destroyNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/notification/IpSvcEventNotification/reportNotification:1.0
 */
static void
decode_org_csapi_fw_fw_service_notification_IpSvcEventNotification_reportNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin union "org_csapi_fw_TpFwEventInfo"  */
        decode_org_csapi_fw_TpFwEventInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_fw_TpFwEventInfo"  */
        proto_tree_add_int(tree, hf_org_csapi_fw_fw_service_notification_IpSvcEventNotification_reportNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/fw/fw_service/notification/IpSvcEventNotification/notificationTerminated:1.0
 */
static void
decode_org_csapi_fw_fw_service_notification_IpSvcEventNotification_notificationTerminated(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpAppCall/routeRes:1.0
 */
static void
decode_org_csapi_cc_gccs_IpAppCall_routeRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpAppCall_routeRes_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_gccs_TpCallReport"  */
        decode_org_csapi_cc_gccs_TpCallReport_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_gccs_TpCallReport"  */
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpAppCall_routeRes_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpAppCall/routeErr:1.0
 */
static void
decode_org_csapi_cc_gccs_IpAppCall_routeErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpAppCall_routeErr_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_TpCallError"  */
        decode_org_csapi_cc_TpCallError_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallError"  */
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpAppCall_routeErr_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpAppCall/getCallInfoRes:1.0
 */
static void
decode_org_csapi_cc_gccs_IpAppCall_getCallInfoRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpAppCall_getCallInfoRes_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_gccs_TpCallInfoReport"  */
        decode_org_csapi_cc_gccs_TpCallInfoReport_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_gccs_TpCallInfoReport"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpAppCall/getCallInfoErr:1.0
 */
static void
decode_org_csapi_cc_gccs_IpAppCall_getCallInfoErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpAppCall_getCallInfoErr_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_TpCallError"  */
        decode_org_csapi_cc_TpCallError_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallError"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpAppCall/superviseCallRes:1.0
 */
static void
decode_org_csapi_cc_gccs_IpAppCall_superviseCallRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpAppCall_superviseCallRes_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpAppCall_superviseCallRes_report, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpAppCall_superviseCallRes_usedTime, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpAppCall/superviseCallErr:1.0
 */
static void
decode_org_csapi_cc_gccs_IpAppCall_superviseCallErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpAppCall_superviseCallErr_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_TpCallError"  */
        decode_org_csapi_cc_TpCallError_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallError"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpAppCall/callFaultDetected:1.0
 */
static void
decode_org_csapi_cc_gccs_IpAppCall_callFaultDetected(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpAppCall_callFaultDetected_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_gccs_IpAppCall_callFaultDetected_fault, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpAppCall/getMoreDialledDigitsRes:1.0
 */
static void
decode_org_csapi_cc_gccs_IpAppCall_getMoreDialledDigitsRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpAppCall_getMoreDialledDigitsRes_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_gccs_IpAppCall_getMoreDialledDigitsRes_digits);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpAppCall/getMoreDialledDigitsErr:1.0
 */
static void
decode_org_csapi_cc_gccs_IpAppCall_getMoreDialledDigitsErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpAppCall_getMoreDialledDigitsErr_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_TpCallError"  */
        decode_org_csapi_cc_TpCallError_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallError"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpAppCall/callEnded:1.0
 */
static void
decode_org_csapi_cc_gccs_IpAppCall_callEnded(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpAppCall_callEnded_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_gccs_TpCallEndedReport"  */
        decode_org_csapi_cc_gccs_TpCallEndedReport_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_gccs_TpCallEndedReport"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpCall/routeReq:1.0
 */
static void
decode_org_csapi_cc_gccs_IpCall_routeReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_gccs_IpCall_routeReq_responseRequested;
    guint32   i_org_csapi_cc_gccs_IpCall_routeReq_responseRequested;
    guint32   u_octet4_loop_org_csapi_cc_gccs_IpCall_routeReq_appInfo;
    guint32   i_org_csapi_cc_gccs_IpCall_routeReq_appInfo;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpCall_routeReq_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_cc_gccs_IpCall_routeReq_responseRequested = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_gccs_IpCall_routeReq_responseRequested_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_gccs_IpCall_routeReq_responseRequested);

        for (i_org_csapi_cc_gccs_IpCall_routeReq_responseRequested=0; i_org_csapi_cc_gccs_IpCall_routeReq_responseRequested < u_octet4_loop_org_csapi_cc_gccs_IpCall_routeReq_responseRequested; i_org_csapi_cc_gccs_IpCall_routeReq_responseRequested++) {

            /*  Begin struct "org_csapi_cc_gccs_TpCallReportRequest"  */
            decode_org_csapi_cc_gccs_TpCallReportRequest_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cc_gccs_TpCallReportRequest"  */
        }

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        u_octet4_loop_org_csapi_cc_gccs_IpCall_routeReq_appInfo = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_gccs_IpCall_routeReq_appInfo_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_gccs_IpCall_routeReq_appInfo);

        for (i_org_csapi_cc_gccs_IpCall_routeReq_appInfo=0; i_org_csapi_cc_gccs_IpCall_routeReq_appInfo < u_octet4_loop_org_csapi_cc_gccs_IpCall_routeReq_appInfo; i_org_csapi_cc_gccs_IpCall_routeReq_appInfo++) {

            /*  Begin union "org_csapi_cc_gccs_TpCallAppInfo"  */
            decode_org_csapi_cc_gccs_TpCallAppInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_cc_gccs_TpCallAppInfo"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpCall_routeReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpCall/release:1.0
 */
static void
decode_org_csapi_cc_gccs_IpCall_release(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpCall_release_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_gccs_TpCallReleaseCause"  */
        decode_org_csapi_cc_gccs_TpCallReleaseCause_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_gccs_TpCallReleaseCause"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpCall/deassignCall:1.0
 */
static void
decode_org_csapi_cc_gccs_IpCall_deassignCall(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpCall_deassignCall_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpCall/getCallInfoReq:1.0
 */
static void
decode_org_csapi_cc_gccs_IpCall_getCallInfoReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpCall_getCallInfoReq_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpCall_getCallInfoReq_callInfoRequested, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpCall/setCallChargePlan:1.0
 */
static void
decode_org_csapi_cc_gccs_IpCall_setCallChargePlan(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpCall_setCallChargePlan_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_TpCallChargePlan"  */
        decode_org_csapi_cc_TpCallChargePlan_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallChargePlan"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpCall/setAdviceOfCharge:1.0
 */
static void
decode_org_csapi_cc_gccs_IpCall_setAdviceOfCharge(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpCall_setAdviceOfCharge_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_TpAoCInfo"  */
        decode_org_csapi_TpAoCInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAoCInfo"  */
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpCall_setAdviceOfCharge_tariffSwitch, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpCall/getMoreDialledDigitsReq:1.0
 */
static void
decode_org_csapi_cc_gccs_IpCall_getMoreDialledDigitsReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpCall_getMoreDialledDigitsReq_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpCall_getMoreDialledDigitsReq_length, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpCall/superviseCallReq:1.0
 */
static void
decode_org_csapi_cc_gccs_IpCall_superviseCallReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpCall_superviseCallReq_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpCall_superviseCallReq_time, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpCall_superviseCallReq_treatment, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpCall/continueProcessing:1.0
 */
static void
decode_org_csapi_cc_gccs_IpCall_continueProcessing(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpCall_continueProcessing_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpAppCallControlManager/callAborted:1.0
 */
static void
decode_org_csapi_cc_gccs_IpAppCallControlManager_callAborted(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpAppCallControlManager_callAborted_callReference, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpAppCallControlManager/callEventNotify:1.0
 */
static void
decode_org_csapi_cc_gccs_IpAppCallControlManager_callEventNotify(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_cc_gccs_TpCallIdentifier"  */
        decode_org_csapi_cc_gccs_TpCallIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_gccs_TpCallIdentifier"  */
        /*  Begin struct "org_csapi_cc_gccs_TpCallEventInfo"  */
        decode_org_csapi_cc_gccs_TpCallEventInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_gccs_TpCallEventInfo"  */
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpAppCallControlManager_callEventNotify_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpAppCallControlManager/callNotificationInterrupted:1.0
 */
static void
decode_org_csapi_cc_gccs_IpAppCallControlManager_callNotificationInterrupted(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpAppCallControlManager/callNotificationContinued:1.0
 */
static void
decode_org_csapi_cc_gccs_IpAppCallControlManager_callNotificationContinued(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpAppCallControlManager/callOverloadEncountered:1.0
 */
static void
decode_org_csapi_cc_gccs_IpAppCallControlManager_callOverloadEncountered(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpAppCallControlManager_callOverloadEncountered_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpAppCallControlManager/callOverloadCeased:1.0
 */
static void
decode_org_csapi_cc_gccs_IpAppCallControlManager_callOverloadCeased(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpAppCallControlManager_callOverloadCeased_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpAppCallControlManager/abortMultipleCalls:1.0
 */
static void
decode_org_csapi_cc_gccs_IpAppCallControlManager_abortMultipleCalls(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_gccs_IpAppCallControlManager_abortMultipleCalls_callReferenceSet;
    guint32   i_org_csapi_cc_gccs_IpAppCallControlManager_abortMultipleCalls_callReferenceSet;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_cc_gccs_IpAppCallControlManager_abortMultipleCalls_callReferenceSet = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_gccs_IpAppCallControlManager_abortMultipleCalls_callReferenceSet_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_gccs_IpAppCallControlManager_abortMultipleCalls_callReferenceSet);

        for (i_org_csapi_cc_gccs_IpAppCallControlManager_abortMultipleCalls_callReferenceSet=0; i_org_csapi_cc_gccs_IpAppCallControlManager_abortMultipleCalls_callReferenceSet < u_octet4_loop_org_csapi_cc_gccs_IpAppCallControlManager_abortMultipleCalls_callReferenceSet; i_org_csapi_cc_gccs_IpAppCallControlManager_abortMultipleCalls_callReferenceSet++) {

            proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpAppCallControlManager_abortMultipleCalls_callReferenceSet, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpCallControlManager/createCall:1.0
 */
static void
decode_org_csapi_cc_gccs_IpCallControlManager_createCall(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cc_gccs_TpCallIdentifier"  */
            decode_org_csapi_cc_gccs_TpCallIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cc_gccs_TpCallIdentifier"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpCallControlManager/enableCallNotification:1.0
 */
static void
decode_org_csapi_cc_gccs_IpCallControlManager_enableCallNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        /*  Begin struct "org_csapi_cc_gccs_TpCallEventCriteria"  */
        decode_org_csapi_cc_gccs_TpCallEventCriteria_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_gccs_TpCallEventCriteria"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpCallControlManager_enableCallNotification_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpCallControlManager/disableCallNotification:1.0
 */
static void
decode_org_csapi_cc_gccs_IpCallControlManager_disableCallNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpCallControlManager_disableCallNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpCallControlManager/setCallLoadControl:1.0
 */
static void
decode_org_csapi_cc_gccs_IpCallControlManager_setCallLoadControl(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpCallControlManager_setCallLoadControl_duration, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin union "org_csapi_cc_TpCallLoadControlMechanism"  */
        decode_org_csapi_cc_TpCallLoadControlMechanism_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_cc_TpCallLoadControlMechanism"  */
        /*  Begin struct "org_csapi_cc_gccs_TpCallTreatment"  */
        decode_org_csapi_cc_gccs_TpCallTreatment_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_gccs_TpCallTreatment"  */
        /*  Begin struct "org_csapi_TpAddressRange"  */
        decode_org_csapi_TpAddressRange_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddressRange"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpCallControlManager_setCallLoadControl_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpCallControlManager/changeCallNotification:1.0
 */
static void
decode_org_csapi_cc_gccs_IpCallControlManager_changeCallNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_IpCallControlManager_changeCallNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_gccs_TpCallEventCriteria"  */
        decode_org_csapi_cc_gccs_TpCallEventCriteria_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_gccs_TpCallEventCriteria"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/gccs/IpCallControlManager/getCriteria:1.0
 */
static void
decode_org_csapi_cc_gccs_IpCallControlManager_getCriteria(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_gccs_IpCallControlManager_getCriteria_return;
    guint32   i_org_csapi_cc_gccs_IpCallControlManager_getCriteria_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_cc_gccs_IpCallControlManager_getCriteria_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_cc_gccs_IpCallControlManager_getCriteria_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_gccs_IpCallControlManager_getCriteria_return);

            for (i_org_csapi_cc_gccs_IpCallControlManager_getCriteria_return=0; i_org_csapi_cc_gccs_IpCallControlManager_getCriteria_return < u_octet4_loop_org_csapi_cc_gccs_IpCallControlManager_getCriteria_return; i_org_csapi_cc_gccs_IpCallControlManager_getCriteria_return++) {

                /*  Begin struct "org_csapi_cc_gccs_TpCallEventCriteriaResult"  */
                decode_org_csapi_cc_gccs_TpCallEventCriteriaResult_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_cc_gccs_TpCallEventCriteriaResult"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppCallLeg/eventReportRes:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppCallLeg_eventReportRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppCallLeg_eventReportRes_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_TpCallEventInfo"  */
        decode_org_csapi_cc_TpCallEventInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallEventInfo"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppCallLeg/eventReportErr:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppCallLeg_eventReportErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppCallLeg_eventReportErr_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_TpCallError"  */
        decode_org_csapi_cc_TpCallError_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallError"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppCallLeg/attachMediaRes:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppCallLeg_attachMediaRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppCallLeg_attachMediaRes_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppCallLeg/attachMediaErr:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppCallLeg_attachMediaErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppCallLeg_attachMediaErr_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_TpCallError"  */
        decode_org_csapi_cc_TpCallError_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallError"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppCallLeg/detachMediaRes:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppCallLeg_detachMediaRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppCallLeg_detachMediaRes_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppCallLeg/detachMediaErr:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppCallLeg_detachMediaErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppCallLeg_detachMediaErr_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_TpCallError"  */
        decode_org_csapi_cc_TpCallError_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallError"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppCallLeg/getInfoRes:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppCallLeg_getInfoRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppCallLeg_getInfoRes_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_TpCallLegInfoReport"  */
        decode_org_csapi_cc_TpCallLegInfoReport_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallLegInfoReport"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppCallLeg/getInfoErr:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppCallLeg_getInfoErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppCallLeg_getInfoErr_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_TpCallError"  */
        decode_org_csapi_cc_TpCallError_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallError"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppCallLeg/routeErr:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppCallLeg_routeErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppCallLeg_routeErr_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_TpCallError"  */
        decode_org_csapi_cc_TpCallError_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallError"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppCallLeg/superviseRes:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppCallLeg_superviseRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppCallLeg_superviseRes_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppCallLeg_superviseRes_report, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppCallLeg_superviseRes_usedTime, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppCallLeg/superviseErr:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppCallLeg_superviseErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppCallLeg_superviseErr_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_TpCallError"  */
        decode_org_csapi_cc_TpCallError_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallError"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppCallLeg/callLegEnded:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppCallLeg_callLegEnded(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppCallLeg_callLegEnded_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_mpccs_IpAppCallLeg_callLegEnded_cause, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpMultiPartyCallControlManager/createCall:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_createCall(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cc_mpccs_TpMultiPartyCallIdentifier"  */
            decode_org_csapi_cc_mpccs_TpMultiPartyCallIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cc_mpccs_TpMultiPartyCallIdentifier"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpMultiPartyCallControlManager/createNotification:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_createNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        /*  Begin struct "org_csapi_cc_TpCallNotificationRequest"  */
        decode_org_csapi_cc_TpCallNotificationRequest_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallNotificationRequest"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_createNotification_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpMultiPartyCallControlManager/destroyNotification:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_destroyNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_destroyNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpMultiPartyCallControlManager/changeNotification:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_changeNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_changeNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_TpCallNotificationRequest"  */
        decode_org_csapi_cc_TpCallNotificationRequest_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallNotificationRequest"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpMultiPartyCallControlManager/getNotification:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_getNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_getNotification_return;
    guint32   i_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_getNotification_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_getNotification_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_getNotification_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_getNotification_return);

            for (i_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_getNotification_return=0; i_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_getNotification_return < u_octet4_loop_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_getNotification_return; i_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_getNotification_return++) {

                /*  Begin struct "org_csapi_cc_TpNotificationRequested"  */
                decode_org_csapi_cc_TpNotificationRequested_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_cc_TpNotificationRequested"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpMultiPartyCallControlManager/setCallLoadControl:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_setCallLoadControl(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_setCallLoadControl_duration, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin union "org_csapi_cc_TpCallLoadControlMechanism"  */
        decode_org_csapi_cc_TpCallLoadControlMechanism_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_cc_TpCallLoadControlMechanism"  */
        /*  Begin struct "org_csapi_cc_TpCallTreatment"  */
        decode_org_csapi_cc_TpCallTreatment_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallTreatment"  */
        /*  Begin struct "org_csapi_TpAddressRange"  */
        decode_org_csapi_TpAddressRange_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddressRange"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_setCallLoadControl_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpMultiPartyCallControlManager/enableNotifications:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_enableNotifications(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_enableNotifications_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpMultiPartyCallControlManager/disableNotifications:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_disableNotifications(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpMultiPartyCallControlManager/getNextNotification:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_getNextNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_boolean(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_getNextNotification_reset, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cc_TpNotificationRequestedSetEntry"  */
            decode_org_csapi_cc_TpNotificationRequestedSetEntry_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cc_TpNotificationRequestedSetEntry"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpCallLeg/routeReq:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpCallLeg_routeReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_mpccs_IpCallLeg_routeReq_appInfo;
    guint32   i_org_csapi_cc_mpccs_IpCallLeg_routeReq_appInfo;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpCallLeg_routeReq_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        u_octet4_loop_org_csapi_cc_mpccs_IpCallLeg_routeReq_appInfo = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_mpccs_IpCallLeg_routeReq_appInfo_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_mpccs_IpCallLeg_routeReq_appInfo);

        for (i_org_csapi_cc_mpccs_IpCallLeg_routeReq_appInfo=0; i_org_csapi_cc_mpccs_IpCallLeg_routeReq_appInfo < u_octet4_loop_org_csapi_cc_mpccs_IpCallLeg_routeReq_appInfo; i_org_csapi_cc_mpccs_IpCallLeg_routeReq_appInfo++) {

            /*  Begin union "org_csapi_cc_TpCallAppInfo"  */
            decode_org_csapi_cc_TpCallAppInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_cc_TpCallAppInfo"  */
        }

        /*  Begin struct "org_csapi_cc_TpCallLegConnectionProperties"  */
        decode_org_csapi_cc_TpCallLegConnectionProperties_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallLegConnectionProperties"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpCallLeg/eventReportReq:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpCallLeg_eventReportReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_mpccs_IpCallLeg_eventReportReq_eventsRequested;
    guint32   i_org_csapi_cc_mpccs_IpCallLeg_eventReportReq_eventsRequested;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpCallLeg_eventReportReq_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_cc_mpccs_IpCallLeg_eventReportReq_eventsRequested = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_mpccs_IpCallLeg_eventReportReq_eventsRequested_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_mpccs_IpCallLeg_eventReportReq_eventsRequested);

        for (i_org_csapi_cc_mpccs_IpCallLeg_eventReportReq_eventsRequested=0; i_org_csapi_cc_mpccs_IpCallLeg_eventReportReq_eventsRequested < u_octet4_loop_org_csapi_cc_mpccs_IpCallLeg_eventReportReq_eventsRequested; i_org_csapi_cc_mpccs_IpCallLeg_eventReportReq_eventsRequested++) {

            /*  Begin struct "org_csapi_cc_TpCallEventRequest"  */
            decode_org_csapi_cc_TpCallEventRequest_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cc_TpCallEventRequest"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpCallLeg/release:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpCallLeg_release(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpCallLeg_release_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_mpccs_IpCallLeg_release_cause, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpCallLeg/getInfoReq:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpCallLeg_getInfoReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpCallLeg_getInfoReq_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpCallLeg_getInfoReq_callLegInfoRequested, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpCallLeg/getCall:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpCallLeg_getCall(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpCallLeg_getCall_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cc_mpccs_TpMultiPartyCallIdentifier"  */
            decode_org_csapi_cc_mpccs_TpMultiPartyCallIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cc_mpccs_TpMultiPartyCallIdentifier"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpCallLeg/attachMediaReq:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpCallLeg_attachMediaReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpCallLeg_attachMediaReq_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpCallLeg/detachMediaReq:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpCallLeg_detachMediaReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpCallLeg_detachMediaReq_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpCallLeg/getCurrentDestinationAddress:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpCallLeg_getCurrentDestinationAddress(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpCallLeg_getCurrentDestinationAddress_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpCallLeg/continueProcessing:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpCallLeg_continueProcessing(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpCallLeg_continueProcessing_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpCallLeg/setChargePlan:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpCallLeg_setChargePlan(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpCallLeg_setChargePlan_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_TpCallChargePlan"  */
        decode_org_csapi_cc_TpCallChargePlan_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallChargePlan"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpCallLeg/setAdviceOfCharge:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpCallLeg_setAdviceOfCharge(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpCallLeg_setAdviceOfCharge_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_TpAoCInfo"  */
        decode_org_csapi_TpAoCInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAoCInfo"  */
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpCallLeg_setAdviceOfCharge_tariffSwitch, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpCallLeg/superviseReq:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpCallLeg_superviseReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpCallLeg_superviseReq_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpCallLeg_superviseReq_time, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpCallLeg_superviseReq_treatment, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpCallLeg/deassign:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpCallLeg_deassign(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpCallLeg_deassign_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpCallLeg/getProperties:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpCallLeg_getProperties(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_mpccs_IpCallLeg_getProperties_propertyNames;
    guint32   i_org_csapi_cc_mpccs_IpCallLeg_getProperties_propertyNames;
    guint32   u_octet4_loop_org_csapi_cc_mpccs_IpCallLeg_getProperties_return;
    guint32   i_org_csapi_cc_mpccs_IpCallLeg_getProperties_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpCallLeg_getProperties_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_cc_mpccs_IpCallLeg_getProperties_propertyNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_mpccs_IpCallLeg_getProperties_propertyNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_mpccs_IpCallLeg_getProperties_propertyNames);

        for (i_org_csapi_cc_mpccs_IpCallLeg_getProperties_propertyNames=0; i_org_csapi_cc_mpccs_IpCallLeg_getProperties_propertyNames < u_octet4_loop_org_csapi_cc_mpccs_IpCallLeg_getProperties_propertyNames; i_org_csapi_cc_mpccs_IpCallLeg_getProperties_propertyNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_mpccs_IpCallLeg_getProperties_propertyNames);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_cc_mpccs_IpCallLeg_getProperties_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_cc_mpccs_IpCallLeg_getProperties_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_mpccs_IpCallLeg_getProperties_return);

            for (i_org_csapi_cc_mpccs_IpCallLeg_getProperties_return=0; i_org_csapi_cc_mpccs_IpCallLeg_getProperties_return < u_octet4_loop_org_csapi_cc_mpccs_IpCallLeg_getProperties_return; i_org_csapi_cc_mpccs_IpCallLeg_getProperties_return++) {

                /*  Begin struct "org_csapi_cc_TpCallLegProperty"  */
                decode_org_csapi_cc_TpCallLegProperty_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_cc_TpCallLegProperty"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpCallLeg/setProperties:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpCallLeg_setProperties(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_mpccs_IpCallLeg_setProperties_properties;
    guint32   i_org_csapi_cc_mpccs_IpCallLeg_setProperties_properties;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpCallLeg_setProperties_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_cc_mpccs_IpCallLeg_setProperties_properties = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_mpccs_IpCallLeg_setProperties_properties_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_mpccs_IpCallLeg_setProperties_properties);

        for (i_org_csapi_cc_mpccs_IpCallLeg_setProperties_properties=0; i_org_csapi_cc_mpccs_IpCallLeg_setProperties_properties < u_octet4_loop_org_csapi_cc_mpccs_IpCallLeg_setProperties_properties; i_org_csapi_cc_mpccs_IpCallLeg_setProperties_properties++) {

            /*  Begin struct "org_csapi_cc_TpCallLegProperty"  */
            decode_org_csapi_cc_TpCallLegProperty_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cc_TpCallLegProperty"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppMultiPartyCall/getInfoRes:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppMultiPartyCall_getInfoRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_getInfoRes_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_TpCallInfoReport"  */
        decode_org_csapi_cc_TpCallInfoReport_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallInfoReport"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppMultiPartyCall/getInfoErr:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppMultiPartyCall_getInfoErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_getInfoErr_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_TpCallError"  */
        decode_org_csapi_cc_TpCallError_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallError"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppMultiPartyCall/superviseRes:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppMultiPartyCall_superviseRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_superviseRes_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_superviseRes_report, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_superviseRes_usedTime, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppMultiPartyCall/superviseErr:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppMultiPartyCall_superviseErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_superviseErr_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_TpCallError"  */
        decode_org_csapi_cc_TpCallError_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallError"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppMultiPartyCall/callEnded:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppMultiPartyCall_callEnded(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_callEnded_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_TpCallEndedReport"  */
        decode_org_csapi_cc_TpCallEndedReport_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallEndedReport"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppMultiPartyCall/createAndRouteCallLegErr:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppMultiPartyCall_createAndRouteCallLegErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_createAndRouteCallLegErr_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_mpccs_TpCallLegIdentifier"  */
        decode_org_csapi_cc_mpccs_TpCallLegIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_mpccs_TpCallLegIdentifier"  */
        /*  Begin struct "org_csapi_cc_TpCallError"  */
        decode_org_csapi_cc_TpCallError_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallError"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpMultiPartyCall/getCallLegs:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpMultiPartyCall_getCallLegs(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_mpccs_IpMultiPartyCall_getCallLegs_return;
    guint32   i_org_csapi_cc_mpccs_IpMultiPartyCall_getCallLegs_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCall_getCallLegs_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_cc_mpccs_IpMultiPartyCall_getCallLegs_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCall_getCallLegs_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_mpccs_IpMultiPartyCall_getCallLegs_return);

            for (i_org_csapi_cc_mpccs_IpMultiPartyCall_getCallLegs_return=0; i_org_csapi_cc_mpccs_IpMultiPartyCall_getCallLegs_return < u_octet4_loop_org_csapi_cc_mpccs_IpMultiPartyCall_getCallLegs_return; i_org_csapi_cc_mpccs_IpMultiPartyCall_getCallLegs_return++) {

                /*  Begin struct "org_csapi_cc_mpccs_TpCallLegIdentifier"  */
                decode_org_csapi_cc_mpccs_TpCallLegIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_cc_mpccs_TpCallLegIdentifier"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpMultiPartyCall/createCallLeg:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpMultiPartyCall_createCallLeg(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCall_createCallLeg_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cc_mpccs_TpCallLegIdentifier"  */
            decode_org_csapi_cc_mpccs_TpCallLegIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cc_mpccs_TpCallLegIdentifier"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpMultiPartyCall/createAndRouteCallLegReq:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_eventsRequested;
    guint32   i_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_eventsRequested;
    guint32   u_octet4_loop_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_appInfo;
    guint32   i_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_appInfo;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_eventsRequested = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_eventsRequested_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_eventsRequested);

        for (i_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_eventsRequested=0; i_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_eventsRequested < u_octet4_loop_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_eventsRequested; i_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_eventsRequested++) {

            /*  Begin struct "org_csapi_cc_TpCallEventRequest"  */
            decode_org_csapi_cc_TpCallEventRequest_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cc_TpCallEventRequest"  */
        }

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        u_octet4_loop_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_appInfo = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_appInfo_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_appInfo);

        for (i_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_appInfo=0; i_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_appInfo < u_octet4_loop_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_appInfo; i_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_appInfo++) {

            /*  Begin union "org_csapi_cc_TpCallAppInfo"  */
            decode_org_csapi_cc_TpCallAppInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_cc_TpCallAppInfo"  */
        }

        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cc_mpccs_TpCallLegIdentifier"  */
            decode_org_csapi_cc_mpccs_TpCallLegIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cc_mpccs_TpCallLegIdentifier"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpMultiPartyCall/release:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpMultiPartyCall_release(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCall_release_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCall_release_cause, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpMultiPartyCall/deassignCall:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpMultiPartyCall_deassignCall(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCall_deassignCall_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpMultiPartyCall/getInfoReq:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpMultiPartyCall_getInfoReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCall_getInfoReq_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCall_getInfoReq_callInfoRequested, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpMultiPartyCall/setChargePlan:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpMultiPartyCall_setChargePlan(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCall_setChargePlan_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_TpCallChargePlan"  */
        decode_org_csapi_cc_TpCallChargePlan_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallChargePlan"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpMultiPartyCall/setAdviceOfCharge:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpMultiPartyCall_setAdviceOfCharge(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCall_setAdviceOfCharge_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_TpAoCInfo"  */
        decode_org_csapi_TpAoCInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAoCInfo"  */
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCall_setAdviceOfCharge_tariffSwitch, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpMultiPartyCall/superviseReq:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpMultiPartyCall_superviseReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCall_superviseReq_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCall_superviseReq_time, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpMultiPartyCall_superviseReq_treatment, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppMultiPartyCallControlManager/reportNotification:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_reportNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_reportNotification_callLegReferenceSet;
    guint32   i_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_reportNotification_callLegReferenceSet;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_cc_mpccs_TpMultiPartyCallIdentifier"  */
        decode_org_csapi_cc_mpccs_TpMultiPartyCallIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_mpccs_TpMultiPartyCallIdentifier"  */
        u_octet4_loop_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_reportNotification_callLegReferenceSet = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_reportNotification_callLegReferenceSet_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_reportNotification_callLegReferenceSet);

        for (i_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_reportNotification_callLegReferenceSet=0; i_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_reportNotification_callLegReferenceSet < u_octet4_loop_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_reportNotification_callLegReferenceSet; i_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_reportNotification_callLegReferenceSet++) {

            /*  Begin struct "org_csapi_cc_mpccs_TpCallLegIdentifier"  */
            decode_org_csapi_cc_mpccs_TpCallLegIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cc_mpccs_TpCallLegIdentifier"  */
        }

        /*  Begin struct "org_csapi_cc_TpCallNotificationInfo"  */
        decode_org_csapi_cc_TpCallNotificationInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallNotificationInfo"  */
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_reportNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin union "org_csapi_cc_mpccs_TpAppMultiPartyCallBack"  */
            decode_org_csapi_cc_mpccs_TpAppMultiPartyCallBack_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_cc_mpccs_TpAppMultiPartyCallBack"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppMultiPartyCallControlManager/callAborted:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_callAborted(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_callAborted_callReference, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppMultiPartyCallControlManager/managerInterrupted:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_managerInterrupted(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppMultiPartyCallControlManager/managerResumed:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_managerResumed(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppMultiPartyCallControlManager/callOverloadEncountered:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_callOverloadEncountered(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_callOverloadEncountered_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppMultiPartyCallControlManager/callOverloadCeased:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_callOverloadCeased(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_callOverloadCeased_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mpccs/IpAppMultiPartyCallControlManager/abortMultipleCalls:1.0
 */
static void
decode_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_abortMultipleCalls(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_abortMultipleCalls_callReferenceSet;
    guint32   i_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_abortMultipleCalls_callReferenceSet;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_abortMultipleCalls_callReferenceSet = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_abortMultipleCalls_callReferenceSet_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_abortMultipleCalls_callReferenceSet);

        for (i_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_abortMultipleCalls_callReferenceSet=0; i_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_abortMultipleCalls_callReferenceSet < u_octet4_loop_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_abortMultipleCalls_callReferenceSet; i_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_abortMultipleCalls_callReferenceSet++) {

            proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_abortMultipleCalls_callReferenceSet, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mmccs/IpAppMultiMediaCall/superviseVolumeRes:1.0
 */
static void
decode_org_csapi_cc_mmccs_IpAppMultiMediaCall_superviseVolumeRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_IpAppMultiMediaCall_superviseVolumeRes_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_IpAppMultiMediaCall_superviseVolumeRes_report, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_mmccs_TpCallSuperviseVolume"  */
        decode_org_csapi_cc_mmccs_TpCallSuperviseVolume_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_mmccs_TpCallSuperviseVolume"  */
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_mmccs_IpAppMultiMediaCall_superviseVolumeRes_qualityOfService, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mmccs/IpAppMultiMediaCall/superviseVolumeErr:1.0
 */
static void
decode_org_csapi_cc_mmccs_IpAppMultiMediaCall_superviseVolumeErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_IpAppMultiMediaCall_superviseVolumeErr_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_TpCallError"  */
        decode_org_csapi_cc_TpCallError_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallError"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mmccs/IpAppMultiMediaCallLeg/mediaStreamMonitorRes:1.0
 */
static void
decode_org_csapi_cc_mmccs_IpAppMultiMediaCallLeg_mediaStreamMonitorRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_mmccs_IpAppMultiMediaCallLeg_mediaStreamMonitorRes_streams;
    guint32   i_org_csapi_cc_mmccs_IpAppMultiMediaCallLeg_mediaStreamMonitorRes_streams;
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_IpAppMultiMediaCallLeg_mediaStreamMonitorRes_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_cc_mmccs_IpAppMultiMediaCallLeg_mediaStreamMonitorRes_streams = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_mmccs_IpAppMultiMediaCallLeg_mediaStreamMonitorRes_streams_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_mmccs_IpAppMultiMediaCallLeg_mediaStreamMonitorRes_streams);

        for (i_org_csapi_cc_mmccs_IpAppMultiMediaCallLeg_mediaStreamMonitorRes_streams=0; i_org_csapi_cc_mmccs_IpAppMultiMediaCallLeg_mediaStreamMonitorRes_streams < u_octet4_loop_org_csapi_cc_mmccs_IpAppMultiMediaCallLeg_mediaStreamMonitorRes_streams; i_org_csapi_cc_mmccs_IpAppMultiMediaCallLeg_mediaStreamMonitorRes_streams++) {

            /*  Begin struct "org_csapi_cc_mmccs_TpMediaStream"  */
            decode_org_csapi_cc_mmccs_TpMediaStream_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cc_mmccs_TpMediaStream"  */
        }

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_mmccs_IpAppMultiMediaCallLeg_mediaStreamMonitorRes_type, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mmccs/IpMultiMediaStream/subtract:1.0
 */
static void
decode_org_csapi_cc_mmccs_IpMultiMediaStream_subtract(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_IpMultiMediaStream_subtract_mediaStreamSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mmccs/IpMultiMediaCallLeg/mediaStreamAllow:1.0
 */
static void
decode_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamAllow(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamAllow_mediaStreamList;
    guint32   i_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamAllow_mediaStreamList;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamAllow_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamAllow_mediaStreamList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamAllow_mediaStreamList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamAllow_mediaStreamList);

        for (i_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamAllow_mediaStreamList=0; i_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamAllow_mediaStreamList < u_octet4_loop_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamAllow_mediaStreamList; i_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamAllow_mediaStreamList++) {

            proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamAllow_mediaStreamList, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mmccs/IpMultiMediaCallLeg/mediaStreamMonitorReq:1.0
 */
static void
decode_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamMonitorReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamMonitorReq_mediaStreamEventCriteria;
    guint32   i_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamMonitorReq_mediaStreamEventCriteria;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamMonitorReq_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamMonitorReq_mediaStreamEventCriteria = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamMonitorReq_mediaStreamEventCriteria_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamMonitorReq_mediaStreamEventCriteria);

        for (i_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamMonitorReq_mediaStreamEventCriteria=0; i_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamMonitorReq_mediaStreamEventCriteria < u_octet4_loop_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamMonitorReq_mediaStreamEventCriteria; i_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamMonitorReq_mediaStreamEventCriteria++) {

            /*  Begin struct "org_csapi_cc_mmccs_TpMediaStreamRequest"  */
            decode_org_csapi_cc_mmccs_TpMediaStreamRequest_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cc_mmccs_TpMediaStreamRequest"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mmccs/IpMultiMediaCallLeg/getMediaStreams:1.0
 */
static void
decode_org_csapi_cc_mmccs_IpMultiMediaCallLeg_getMediaStreams(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_mmccs_IpMultiMediaCallLeg_getMediaStreams_return;
    guint32   i_org_csapi_cc_mmccs_IpMultiMediaCallLeg_getMediaStreams_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_IpMultiMediaCallLeg_getMediaStreams_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_cc_mmccs_IpMultiMediaCallLeg_getMediaStreams_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_cc_mmccs_IpMultiMediaCallLeg_getMediaStreams_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_mmccs_IpMultiMediaCallLeg_getMediaStreams_return);

            for (i_org_csapi_cc_mmccs_IpMultiMediaCallLeg_getMediaStreams_return=0; i_org_csapi_cc_mmccs_IpMultiMediaCallLeg_getMediaStreams_return < u_octet4_loop_org_csapi_cc_mmccs_IpMultiMediaCallLeg_getMediaStreams_return; i_org_csapi_cc_mmccs_IpMultiMediaCallLeg_getMediaStreams_return++) {

                /*  Begin struct "org_csapi_cc_mmccs_TpMediaStream"  */
                decode_org_csapi_cc_mmccs_TpMediaStream_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_cc_mmccs_TpMediaStream"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mmccs/IpMultiMediaCall/superviseVolumeReq:1.0
 */
static void
decode_org_csapi_cc_mmccs_IpMultiMediaCall_superviseVolumeReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_IpMultiMediaCall_superviseVolumeReq_callSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_mmccs_TpCallSuperviseVolume"  */
        decode_org_csapi_cc_mmccs_TpCallSuperviseVolume_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_mmccs_TpCallSuperviseVolume"  */
        proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_IpMultiMediaCall_superviseVolumeReq_treatment, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mmccs/IpAppMultiMediaCallControlManager/reportMediaNotification:1.0
 */
static void
decode_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_callLegReferenceSet;
    guint32   i_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_callLegReferenceSet;
    guint32   u_octet4_loop_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_mediaStreams;
    guint32   i_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_mediaStreams;
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_cc_mmccs_TpMultiMediaCallIdentifier"  */
        decode_org_csapi_cc_mmccs_TpMultiMediaCallIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_mmccs_TpMultiMediaCallIdentifier"  */
        u_octet4_loop_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_callLegReferenceSet = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_callLegReferenceSet_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_callLegReferenceSet);

        for (i_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_callLegReferenceSet=0; i_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_callLegReferenceSet < u_octet4_loop_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_callLegReferenceSet; i_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_callLegReferenceSet++) {

            /*  Begin struct "org_csapi_cc_mmccs_TpMultiMediaCallLegIdentifier"  */
            decode_org_csapi_cc_mmccs_TpMultiMediaCallLegIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cc_mmccs_TpMultiMediaCallLegIdentifier"  */
        }

        u_octet4_loop_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_mediaStreams = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_mediaStreams_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_mediaStreams);

        for (i_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_mediaStreams=0; i_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_mediaStreams < u_octet4_loop_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_mediaStreams; i_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_mediaStreams++) {

            /*  Begin struct "org_csapi_cc_mmccs_TpMediaStream"  */
            decode_org_csapi_cc_mmccs_TpMediaStream_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cc_mmccs_TpMediaStream"  */
        }

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_type, tvb, *offset-4, 4, u_octet4);

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_qualityOfService, tvb, *offset-4, 4, u_octet4);

        proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin union "org_csapi_cc_mmccs_TpAppMultiMediaCallBack"  */
            decode_org_csapi_cc_mmccs_TpAppMultiMediaCallBack_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_cc_mmccs_TpAppMultiMediaCallBack"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mmccs/IpMultiMediaCallControlManager/createMediaNotification:1.0
 */
static void
decode_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_createMediaNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        /*  Begin struct "org_csapi_cc_mmccs_TpNotificationMediaRequest"  */
        decode_org_csapi_cc_mmccs_TpNotificationMediaRequest_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_mmccs_TpNotificationMediaRequest"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_createMediaNotification_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mmccs/IpMultiMediaCallControlManager/destroyMediaNotification:1.0
 */
static void
decode_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_destroyMediaNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_destroyMediaNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mmccs/IpMultiMediaCallControlManager/changeMediaNotification:1.0
 */
static void
decode_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_changeMediaNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_changeMediaNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_mmccs_TpNotificationMediaRequest"  */
        decode_org_csapi_cc_mmccs_TpNotificationMediaRequest_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_mmccs_TpNotificationMediaRequest"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/mmccs/IpMultiMediaCallControlManager/getMediaNotification:1.0
 */
static void
decode_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_getMediaNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_getMediaNotification_return;
    guint32   i_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_getMediaNotification_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_getMediaNotification_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_getMediaNotification_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_getMediaNotification_return);

            for (i_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_getMediaNotification_return=0; i_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_getMediaNotification_return < u_octet4_loop_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_getMediaNotification_return; i_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_getMediaNotification_return++) {

                /*  Begin struct "org_csapi_cc_mmccs_TpMediaNotificationRequested"  */
                decode_org_csapi_cc_mmccs_TpMediaNotificationRequested_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_cc_mmccs_TpMediaNotificationRequested"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/cccs/IpAppSubConfCall/chairSelection:1.0
 */
static void
decode_org_csapi_cc_cccs_IpAppSubConfCall_chairSelection(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpAppSubConfCall_chairSelection_subConferenceSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpAppSubConfCall_chairSelection_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/cccs/IpAppSubConfCall/floorRequest:1.0
 */
static void
decode_org_csapi_cc_cccs_IpAppSubConfCall_floorRequest(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpAppSubConfCall_floorRequest_subConferenceSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpAppSubConfCall_floorRequest_callLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/cccs/IpAppConfCall/partyJoined:1.0
 */
static void
decode_org_csapi_cc_cccs_IpAppConfCall_partyJoined(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpAppConfCall_partyJoined_conferenceSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cc_mpccs_TpCallLegIdentifier"  */
        decode_org_csapi_cc_mpccs_TpCallLegIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_mpccs_TpCallLegIdentifier"  */
        /*  Begin struct "org_csapi_cc_cccs_TpJoinEventInfo"  */
        decode_org_csapi_cc_cccs_TpJoinEventInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_cccs_TpJoinEventInfo"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/cccs/IpAppConfCall/leaveMonitorRes:1.0
 */
static void
decode_org_csapi_cc_cccs_IpAppConfCall_leaveMonitorRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpAppConfCall_leaveMonitorRes_conferenceSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpAppConfCall_leaveMonitorRes_callLeg, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/cccs/IpConfCallControlManager/createConference:1.0
 */
static void
decode_org_csapi_cc_cccs_IpConfCallControlManager_createConference(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpConfCallControlManager_createConference_numberOfSubConferences, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin union "org_csapi_cc_cccs_TpConfPolicy"  */
        decode_org_csapi_cc_cccs_TpConfPolicy_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_cc_cccs_TpConfPolicy"  */
        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpConfCallControlManager_createConference_numberOfParticipants, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpConfCallControlManager_createConference_duration, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cc_cccs_TpConfCallIdentifier"  */
            decode_org_csapi_cc_cccs_TpConfCallIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cc_cccs_TpConfCallIdentifier"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/cccs/IpConfCallControlManager/checkResources:1.0
 */
static void
decode_org_csapi_cc_cccs_IpConfCallControlManager_checkResources(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_cc_cccs_TpConfSearchCriteria"  */
        decode_org_csapi_cc_cccs_TpConfSearchCriteria_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_cccs_TpConfSearchCriteria"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cc_cccs_TpConfSearchResult"  */
            decode_org_csapi_cc_cccs_TpConfSearchResult_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cc_cccs_TpConfSearchResult"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/cccs/IpConfCallControlManager/reserveResources:1.0
 */
static void
decode_org_csapi_cc_cccs_IpConfCallControlManager_reserveResources(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_cccs_IpConfCallControlManager_reserveResources_startTime);

        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpConfCallControlManager_reserveResources_numberOfParticipants, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpConfCallControlManager_reserveResources_duration, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin union "org_csapi_cc_cccs_TpConfPolicy"  */
        decode_org_csapi_cc_cccs_TpConfPolicy_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_cc_cccs_TpConfPolicy"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cc_cccs_TpResourceReservation"  */
            decode_org_csapi_cc_cccs_TpResourceReservation_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cc_cccs_TpResourceReservation"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/cccs/IpConfCallControlManager/freeResources:1.0
 */
static void
decode_org_csapi_cc_cccs_IpConfCallControlManager_freeResources(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_cc_cccs_TpResourceReservation"  */
        decode_org_csapi_cc_cccs_TpResourceReservation_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_cccs_TpResourceReservation"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/cccs/IpAppConfCallControlManager/conferenceCreated:1.0
 */
static void
decode_org_csapi_cc_cccs_IpAppConfCallControlManager_conferenceCreated(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_cc_cccs_TpConfCallIdentifier"  */
        decode_org_csapi_cc_cccs_TpConfCallIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_cccs_TpConfCallIdentifier"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/cccs/IpSubConfCall/splitSubConference:1.0
 */
static void
decode_org_csapi_cc_cccs_IpSubConfCall_splitSubConference(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_cccs_IpSubConfCall_splitSubConference_callLegList;
    guint32   i_org_csapi_cc_cccs_IpSubConfCall_splitSubConference_callLegList;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpSubConfCall_splitSubConference_subConferenceSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_cc_cccs_IpSubConfCall_splitSubConference_callLegList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_cccs_IpSubConfCall_splitSubConference_callLegList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_cccs_IpSubConfCall_splitSubConference_callLegList);

        for (i_org_csapi_cc_cccs_IpSubConfCall_splitSubConference_callLegList=0; i_org_csapi_cc_cccs_IpSubConfCall_splitSubConference_callLegList < u_octet4_loop_org_csapi_cc_cccs_IpSubConfCall_splitSubConference_callLegList; i_org_csapi_cc_cccs_IpSubConfCall_splitSubConference_callLegList++) {

            proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpSubConfCall_splitSubConference_callLegList, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        }

        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cc_cccs_TpSubConfCallIdentifier"  */
            decode_org_csapi_cc_cccs_TpSubConfCallIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cc_cccs_TpSubConfCallIdentifier"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/cccs/IpSubConfCall/mergeSubConference:1.0
 */
static void
decode_org_csapi_cc_cccs_IpSubConfCall_mergeSubConference(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpSubConfCall_mergeSubConference_subConferenceCallSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpSubConfCall_mergeSubConference_targetSubConferenceCall, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/cccs/IpSubConfCall/moveCallLeg:1.0
 */
static void
decode_org_csapi_cc_cccs_IpSubConfCall_moveCallLeg(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpSubConfCall_moveCallLeg_subConferenceCallSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpSubConfCall_moveCallLeg_targetSubConferenceCall, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpSubConfCall_moveCallLeg_callLeg, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/cccs/IpSubConfCall/inspectVideo:1.0
 */
static void
decode_org_csapi_cc_cccs_IpSubConfCall_inspectVideo(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpSubConfCall_inspectVideo_subConferenceSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpSubConfCall_inspectVideo_inspectedCallLeg, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/cccs/IpSubConfCall/inspectVideoCancel:1.0
 */
static void
decode_org_csapi_cc_cccs_IpSubConfCall_inspectVideoCancel(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpSubConfCall_inspectVideoCancel_subConferenceSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/cccs/IpSubConfCall/appointSpeaker:1.0
 */
static void
decode_org_csapi_cc_cccs_IpSubConfCall_appointSpeaker(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpSubConfCall_appointSpeaker_subConferenceSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpSubConfCall_appointSpeaker_speakerCallLeg, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/cccs/IpSubConfCall/chairSelection:1.0
 */
static void
decode_org_csapi_cc_cccs_IpSubConfCall_chairSelection(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpSubConfCall_chairSelection_subConferenceSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpSubConfCall_chairSelection_chairCallLeg, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/cccs/IpSubConfCall/changeConferencePolicy:1.0
 */
static void
decode_org_csapi_cc_cccs_IpSubConfCall_changeConferencePolicy(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpSubConfCall_changeConferencePolicy_subConferenceSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin union "org_csapi_cc_cccs_TpConfPolicy"  */
        decode_org_csapi_cc_cccs_TpConfPolicy_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_cc_cccs_TpConfPolicy"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/cccs/IpConfCall/getSubConferences:1.0
 */
static void
decode_org_csapi_cc_cccs_IpConfCall_getSubConferences(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_cccs_IpConfCall_getSubConferences_return;
    guint32   i_org_csapi_cc_cccs_IpConfCall_getSubConferences_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpConfCall_getSubConferences_conferenceSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_cc_cccs_IpConfCall_getSubConferences_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_cc_cccs_IpConfCall_getSubConferences_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_cccs_IpConfCall_getSubConferences_return);

            for (i_org_csapi_cc_cccs_IpConfCall_getSubConferences_return=0; i_org_csapi_cc_cccs_IpConfCall_getSubConferences_return < u_octet4_loop_org_csapi_cc_cccs_IpConfCall_getSubConferences_return; i_org_csapi_cc_cccs_IpConfCall_getSubConferences_return++) {

                /*  Begin struct "org_csapi_cc_cccs_TpSubConfCallIdentifier"  */
                decode_org_csapi_cc_cccs_TpSubConfCallIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_cc_cccs_TpSubConfCallIdentifier"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/cccs/IpConfCall/createSubConference:1.0
 */
static void
decode_org_csapi_cc_cccs_IpConfCall_createSubConference(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpConfCall_createSubConference_conferenceSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        /*  Begin union "org_csapi_cc_cccs_TpConfPolicy"  */
        decode_org_csapi_cc_cccs_TpConfPolicy_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_cc_cccs_TpConfPolicy"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cc_cccs_TpSubConfCallIdentifier"  */
            decode_org_csapi_cc_cccs_TpSubConfCallIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cc_cccs_TpSubConfCallIdentifier"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/cccs/IpConfCall/leaveMonitorReq:1.0
 */
static void
decode_org_csapi_cc_cccs_IpConfCall_leaveMonitorReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpConfCall_leaveMonitorReq_conferenceSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cc/cccs/IpConfCall/getConferenceAddress:1.0
 */
static void
decode_org_csapi_cc_cccs_IpConfCall_getConferenceAddress(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cc_cccs_IpConfCall_getConferenceAddress_conferenceSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUI/sendInfoRes:1.0
 */
static void
decode_org_csapi_ui_IpAppUI_sendInfoRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUI_sendInfoRes_userInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUI_sendInfoRes_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_IpAppUI_sendInfoRes_response, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUI/sendInfoErr:1.0
 */
static void
decode_org_csapi_ui_IpAppUI_sendInfoErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUI_sendInfoErr_userInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUI_sendInfoErr_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_IpAppUI_sendInfoErr_error, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUI/sendInfoAndCollectRes:1.0
 */
static void
decode_org_csapi_ui_IpAppUI_sendInfoAndCollectRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUI_sendInfoAndCollectRes_userInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUI_sendInfoAndCollectRes_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_IpAppUI_sendInfoAndCollectRes_response, tvb, *offset-4, 4, u_octet4);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_IpAppUI_sendInfoAndCollectRes_collectedInfo);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUI/sendInfoAndCollectErr:1.0
 */
static void
decode_org_csapi_ui_IpAppUI_sendInfoAndCollectErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUI_sendInfoAndCollectErr_userInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUI_sendInfoAndCollectErr_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_IpAppUI_sendInfoAndCollectErr_error, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUI/userInteractionFaultDetected:1.0
 */
static void
decode_org_csapi_ui_IpAppUI_userInteractionFaultDetected(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUI_userInteractionFaultDetected_userInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_IpAppUI_userInteractionFaultDetected_fault, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUICall/recordMessageRes:1.0
 */
static void
decode_org_csapi_ui_IpAppUICall_recordMessageRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUICall_recordMessageRes_userInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUICall_recordMessageRes_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_IpAppUICall_recordMessageRes_response, tvb, *offset-4, 4, u_octet4);

        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUICall_recordMessageRes_messageID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUICall/recordMessageErr:1.0
 */
static void
decode_org_csapi_ui_IpAppUICall_recordMessageErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUICall_recordMessageErr_userInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUICall_recordMessageErr_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_IpAppUICall_recordMessageErr_error, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUICall/deleteMessageRes:1.0
 */
static void
decode_org_csapi_ui_IpAppUICall_deleteMessageRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUICall_deleteMessageRes_usrInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_IpAppUICall_deleteMessageRes_response, tvb, *offset-4, 4, u_octet4);

        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUICall_deleteMessageRes_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUICall/deleteMessageErr:1.0
 */
static void
decode_org_csapi_ui_IpAppUICall_deleteMessageErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUICall_deleteMessageErr_usrInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_IpAppUICall_deleteMessageErr_error, tvb, *offset-4, 4, u_octet4);

        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUICall_deleteMessageErr_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUICall/abortActionRes:1.0
 */
static void
decode_org_csapi_ui_IpAppUICall_abortActionRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUICall_abortActionRes_userInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUICall_abortActionRes_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUICall/abortActionErr:1.0
 */
static void
decode_org_csapi_ui_IpAppUICall_abortActionErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUICall_abortActionErr_userInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUICall_abortActionErr_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_IpAppUICall_abortActionErr_error, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUICall/getMessageRes:1.0
 */
static void
decode_org_csapi_ui_IpAppUICall_getMessageRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUICall_getMessageRes_userInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUICall_getMessageRes_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin union "org_csapi_ui_TpUIInfo"  */
        decode_org_csapi_ui_TpUIInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_ui_TpUIInfo"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUICall/getMessageErr:1.0
 */
static void
decode_org_csapi_ui_IpAppUICall_getMessageErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUICall_getMessageErr_userInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUICall_getMessageErr_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_IpAppUICall_getMessageErr_error, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUIManager/userInteractionAborted:1.0
 */
static void
decode_org_csapi_ui_IpAppUIManager_userInteractionAborted(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_ui_TpUIIdentifier"  */
        decode_org_csapi_ui_TpUIIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_ui_TpUIIdentifier"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUIManager/userInteractionNotificationInterrupted:1.0
 */
static void
decode_org_csapi_ui_IpAppUIManager_userInteractionNotificationInterrupted(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUIManager/userInteractionNotificationContinued:1.0
 */
static void
decode_org_csapi_ui_IpAppUIManager_userInteractionNotificationContinued(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUIManager/reportEventNotification:1.0
 */
static void
decode_org_csapi_ui_IpAppUIManager_reportEventNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_ui_TpUIIdentifier"  */
        decode_org_csapi_ui_TpUIIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_ui_TpUIIdentifier"  */
        /*  Begin struct "org_csapi_ui_TpUIEventNotificationInfo"  */
        decode_org_csapi_ui_TpUIEventNotificationInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_ui_TpUIEventNotificationInfo"  */
        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUIManager_reportEventNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUIManager/abortMultipleUserInteractions:1.0
 */
static void
decode_org_csapi_ui_IpAppUIManager_abortMultipleUserInteractions(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_ui_IpAppUIManager_abortMultipleUserInteractions_userInteractionSet;
    guint32   i_org_csapi_ui_IpAppUIManager_abortMultipleUserInteractions_userInteractionSet;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_ui_IpAppUIManager_abortMultipleUserInteractions_userInteractionSet = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_IpAppUIManager_abortMultipleUserInteractions_userInteractionSet_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_ui_IpAppUIManager_abortMultipleUserInteractions_userInteractionSet);

        for (i_org_csapi_ui_IpAppUIManager_abortMultipleUserInteractions_userInteractionSet=0; i_org_csapi_ui_IpAppUIManager_abortMultipleUserInteractions_userInteractionSet < u_octet4_loop_org_csapi_ui_IpAppUIManager_abortMultipleUserInteractions_userInteractionSet; i_org_csapi_ui_IpAppUIManager_abortMultipleUserInteractions_userInteractionSet++) {

            /*  Begin struct "org_csapi_ui_TpUIIdentifier"  */
            decode_org_csapi_ui_TpUIIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_ui_TpUIIdentifier"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpUI/sendInfoReq:1.0
 */
static void
decode_org_csapi_ui_IpUI_sendInfoReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_ui_IpUI_sendInfoReq_variableInfo;
    guint32   i_org_csapi_ui_IpUI_sendInfoReq_variableInfo;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpUI_sendInfoReq_userInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin union "org_csapi_ui_TpUIInfo"  */
        decode_org_csapi_ui_TpUIInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_ui_TpUIInfo"  */
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_IpUI_sendInfoReq_language);

        u_octet4_loop_org_csapi_ui_IpUI_sendInfoReq_variableInfo = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_IpUI_sendInfoReq_variableInfo_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_ui_IpUI_sendInfoReq_variableInfo);

        for (i_org_csapi_ui_IpUI_sendInfoReq_variableInfo=0; i_org_csapi_ui_IpUI_sendInfoReq_variableInfo < u_octet4_loop_org_csapi_ui_IpUI_sendInfoReq_variableInfo; i_org_csapi_ui_IpUI_sendInfoReq_variableInfo++) {

            /*  Begin union "org_csapi_ui_TpUIVariableInfo"  */
            decode_org_csapi_ui_TpUIVariableInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_ui_TpUIVariableInfo"  */
        }

        proto_tree_add_int(tree, hf_org_csapi_ui_IpUI_sendInfoReq_repeatIndicator, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_ui_IpUI_sendInfoReq_responseRequested, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_ui_IpUI_sendInfoReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpUI/sendInfoAndCollectReq:1.0
 */
static void
decode_org_csapi_ui_IpUI_sendInfoAndCollectReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_ui_IpUI_sendInfoAndCollectReq_variableInfo;
    guint32   i_org_csapi_ui_IpUI_sendInfoAndCollectReq_variableInfo;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpUI_sendInfoAndCollectReq_userInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin union "org_csapi_ui_TpUIInfo"  */
        decode_org_csapi_ui_TpUIInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_ui_TpUIInfo"  */
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_IpUI_sendInfoAndCollectReq_language);

        u_octet4_loop_org_csapi_ui_IpUI_sendInfoAndCollectReq_variableInfo = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_IpUI_sendInfoAndCollectReq_variableInfo_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_ui_IpUI_sendInfoAndCollectReq_variableInfo);

        for (i_org_csapi_ui_IpUI_sendInfoAndCollectReq_variableInfo=0; i_org_csapi_ui_IpUI_sendInfoAndCollectReq_variableInfo < u_octet4_loop_org_csapi_ui_IpUI_sendInfoAndCollectReq_variableInfo; i_org_csapi_ui_IpUI_sendInfoAndCollectReq_variableInfo++) {

            /*  Begin union "org_csapi_ui_TpUIVariableInfo"  */
            decode_org_csapi_ui_TpUIVariableInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_ui_TpUIVariableInfo"  */
        }

        /*  Begin struct "org_csapi_ui_TpUICollectCriteria"  */
        decode_org_csapi_ui_TpUICollectCriteria_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_ui_TpUICollectCriteria"  */
        proto_tree_add_int(tree, hf_org_csapi_ui_IpUI_sendInfoAndCollectReq_responseRequested, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_ui_IpUI_sendInfoAndCollectReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpUI/release:1.0
 */
static void
decode_org_csapi_ui_IpUI_release(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpUI_release_userInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpUI/setOriginatingAddress:1.0
 */
static void
decode_org_csapi_ui_IpUI_setOriginatingAddress(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpUI_setOriginatingAddress_userInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_IpUI_setOriginatingAddress_origin);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpUI/getOriginatingAddress:1.0
 */
static void
decode_org_csapi_ui_IpUI_getOriginatingAddress(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpUI_getOriginatingAddress_userInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_IpUI_getOriginatingAddress_return);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpUICall/recordMessageReq:1.0
 */
static void
decode_org_csapi_ui_IpUICall_recordMessageReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpUICall_recordMessageReq_userInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin union "org_csapi_ui_TpUIInfo"  */
        decode_org_csapi_ui_TpUIInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_ui_TpUIInfo"  */
        /*  Begin struct "org_csapi_ui_TpUIMessageCriteria"  */
        decode_org_csapi_ui_TpUIMessageCriteria_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_ui_TpUIMessageCriteria"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_ui_IpUICall_recordMessageReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpUICall/deleteMessageReq:1.0
 */
static void
decode_org_csapi_ui_IpUICall_deleteMessageReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpUICall_deleteMessageReq_usrInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_ui_IpUICall_deleteMessageReq_messageID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_ui_IpUICall_deleteMessageReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpUICall/abortActionReq:1.0
 */
static void
decode_org_csapi_ui_IpUICall_abortActionReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpUICall_abortActionReq_userInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_ui_IpUICall_abortActionReq_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpUICall/getMessageReq:1.0
 */
static void
decode_org_csapi_ui_IpUICall_getMessageReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpUICall_getMessageReq_userInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_ui_IpUICall_getMessageReq_messageID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_ui_IpUICall_getMessageReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpUIManager/createUI:1.0
 */
static void
decode_org_csapi_ui_IpUIManager_createUI(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_ui_TpUIIdentifier"  */
            decode_org_csapi_ui_TpUIIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_ui_TpUIIdentifier"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpUIManager/createUICall:1.0
 */
static void
decode_org_csapi_ui_IpUIManager_createUICall(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        /*  Begin union "org_csapi_ui_TpUITargetObject"  */
        decode_org_csapi_ui_TpUITargetObject_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_ui_TpUITargetObject"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_ui_TpUICallIdentifier"  */
            decode_org_csapi_ui_TpUICallIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_ui_TpUICallIdentifier"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpUIManager/createNotification:1.0
 */
static void
decode_org_csapi_ui_IpUIManager_createNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        /*  Begin struct "org_csapi_ui_TpUIEventCriteria"  */
        decode_org_csapi_ui_TpUIEventCriteria_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_ui_TpUIEventCriteria"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_ui_IpUIManager_createNotification_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpUIManager/destroyNotification:1.0
 */
static void
decode_org_csapi_ui_IpUIManager_destroyNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpUIManager_destroyNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpUIManager/changeNotification:1.0
 */
static void
decode_org_csapi_ui_IpUIManager_changeNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpUIManager_changeNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_ui_TpUIEventCriteria"  */
        decode_org_csapi_ui_TpUIEventCriteria_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_ui_TpUIEventCriteria"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpUIManager/getNotification:1.0
 */
static void
decode_org_csapi_ui_IpUIManager_getNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_ui_IpUIManager_getNotification_return;
    guint32   i_org_csapi_ui_IpUIManager_getNotification_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_ui_IpUIManager_getNotification_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_ui_IpUIManager_getNotification_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_ui_IpUIManager_getNotification_return);

            for (i_org_csapi_ui_IpUIManager_getNotification_return=0; i_org_csapi_ui_IpUIManager_getNotification_return < u_octet4_loop_org_csapi_ui_IpUIManager_getNotification_return; i_org_csapi_ui_IpUIManager_getNotification_return++) {

                /*  Begin struct "org_csapi_ui_TpUIEventCriteriaResult"  */
                decode_org_csapi_ui_TpUIEventCriteriaResult_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_ui_TpUIEventCriteriaResult"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpUIManager/enableNotifications:1.0
 */
static void
decode_org_csapi_ui_IpUIManager_enableNotifications(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_ui_IpUIManager_enableNotifications_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpUIManager/disableNotifications:1.0
 */
static void
decode_org_csapi_ui_IpUIManager_disableNotifications(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUIAdminManager/getMessageRes:1.0
 */
static void
decode_org_csapi_ui_IpAppUIAdminManager_getMessageRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUIAdminManager_getMessageRes_usrInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUIAdminManager_getMessageRes_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin union "org_csapi_ui_TpUIInfo"  */
        decode_org_csapi_ui_TpUIInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_ui_TpUIInfo"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUIAdminManager/getMessageErr:1.0
 */
static void
decode_org_csapi_ui_IpAppUIAdminManager_getMessageErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUIAdminManager_getMessageErr_usrInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUIAdminManager_getMessageErr_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_IpAppUIAdminManager_getMessageErr_error, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUIAdminManager/deleteMessageRes:1.0
 */
static void
decode_org_csapi_ui_IpAppUIAdminManager_deleteMessageRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUIAdminManager_deleteMessageRes_usrInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_IpAppUIAdminManager_deleteMessageRes_response, tvb, *offset-4, 4, u_octet4);

        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUIAdminManager_deleteMessageRes_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUIAdminManager/deleteMessageErr:1.0
 */
static void
decode_org_csapi_ui_IpAppUIAdminManager_deleteMessageErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUIAdminManager_deleteMessageErr_usrInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_IpAppUIAdminManager_deleteMessageErr_error, tvb, *offset-4, 4, u_octet4);

        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUIAdminManager_deleteMessageErr_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUIAdminManager/putMessageRes:1.0
 */
static void
decode_org_csapi_ui_IpAppUIAdminManager_putMessageRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUIAdminManager_putMessageRes_usrInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUIAdminManager_putMessageRes_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUIAdminManager_putMessageRes_messageID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUIAdminManager/putMessageErr:1.0
 */
static void
decode_org_csapi_ui_IpAppUIAdminManager_putMessageErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUIAdminManager_putMessageErr_usrInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUIAdminManager_putMessageErr_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_IpAppUIAdminManager_putMessageErr_error, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUIAdminManager/getMessageListRes:1.0
 */
static void
decode_org_csapi_ui_IpAppUIAdminManager_getMessageListRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_ui_IpAppUIAdminManager_getMessageListRes_messageIDList;
    guint32   i_org_csapi_ui_IpAppUIAdminManager_getMessageListRes_messageIDList;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUIAdminManager_getMessageListRes_usrInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUIAdminManager_getMessageListRes_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_ui_IpAppUIAdminManager_getMessageListRes_messageIDList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_IpAppUIAdminManager_getMessageListRes_messageIDList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_ui_IpAppUIAdminManager_getMessageListRes_messageIDList);

        for (i_org_csapi_ui_IpAppUIAdminManager_getMessageListRes_messageIDList=0; i_org_csapi_ui_IpAppUIAdminManager_getMessageListRes_messageIDList < u_octet4_loop_org_csapi_ui_IpAppUIAdminManager_getMessageListRes_messageIDList; i_org_csapi_ui_IpAppUIAdminManager_getMessageListRes_messageIDList++) {

            proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUIAdminManager_getMessageListRes_messageIDList, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        }

        proto_tree_add_boolean(tree, hf_org_csapi_ui_IpAppUIAdminManager_getMessageListRes_final, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpAppUIAdminManager/getMessageListErr:1.0
 */
static void
decode_org_csapi_ui_IpAppUIAdminManager_getMessageListErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUIAdminManager_getMessageListErr_usrInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_ui_IpAppUIAdminManager_getMessageListErr_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_IpAppUIAdminManager_getMessageListErr_error, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpUIAdminManager/getMessageReq:1.0
 */
static void
decode_org_csapi_ui_IpUIAdminManager_getMessageReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpUIAdminManager_getMessageReq_usrInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_ui_IpUIAdminManager_getMessageReq_messageID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_ui_IpUIAdminManager_getMessageReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpUIAdminManager/putMessageReq:1.0
 */
static void
decode_org_csapi_ui_IpUIAdminManager_putMessageReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpUIAdminManager_putMessageReq_usrInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin union "org_csapi_ui_TpUIInfo"  */
        decode_org_csapi_ui_TpUIInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_ui_TpUIInfo"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_ui_IpUIAdminManager_putMessageReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpUIAdminManager/deleteMessageReq:1.0
 */
static void
decode_org_csapi_ui_IpUIAdminManager_deleteMessageReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpUIAdminManager_deleteMessageReq_usrInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_ui_IpUIAdminManager_deleteMessageReq_messageID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_ui_IpUIAdminManager_deleteMessageReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/ui/IpUIAdminManager/getMessageListReq:1.0
 */
static void
decode_org_csapi_ui_IpUIAdminManager_getMessageListReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_ui_IpUIAdminManager_getMessageListReq_usrInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_boolean(tree, hf_org_csapi_ui_IpUIAdminManager_getMessageListReq_reset, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_ui_IpUIAdminManager_getMessageListReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ul/IpAppUserLocation/locationReportRes:1.0
 */
static void
decode_org_csapi_mm_ul_IpAppUserLocation_locationReportRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_ul_IpAppUserLocation_locationReportRes_locations;
    guint32   i_org_csapi_mm_ul_IpAppUserLocation_locationReportRes_locations;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_ul_IpAppUserLocation_locationReportRes_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_mm_ul_IpAppUserLocation_locationReportRes_locations = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ul_IpAppUserLocation_locationReportRes_locations_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_ul_IpAppUserLocation_locationReportRes_locations);

        for (i_org_csapi_mm_ul_IpAppUserLocation_locationReportRes_locations=0; i_org_csapi_mm_ul_IpAppUserLocation_locationReportRes_locations < u_octet4_loop_org_csapi_mm_ul_IpAppUserLocation_locationReportRes_locations; i_org_csapi_mm_ul_IpAppUserLocation_locationReportRes_locations++) {

            /*  Begin struct "org_csapi_mm_TpUserLocation"  */
            decode_org_csapi_mm_TpUserLocation_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mm_TpUserLocation"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ul/IpAppUserLocation/locationReportErr:1.0
 */
static void
decode_org_csapi_mm_ul_IpAppUserLocation_locationReportErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_ul_IpAppUserLocation_locationReportErr_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ul_IpAppUserLocation_locationReportErr_cause, tvb, *offset-4, 4, u_octet4);

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ul_IpAppUserLocation_locationReportErr_diagnostic, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ul/IpAppUserLocation/extendedLocationReportRes:1.0
 */
static void
decode_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportRes_locations;
    guint32   i_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportRes_locations;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportRes_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportRes_locations = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportRes_locations_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportRes_locations);

        for (i_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportRes_locations=0; i_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportRes_locations < u_octet4_loop_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportRes_locations; i_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportRes_locations++) {

            /*  Begin struct "org_csapi_mm_TpUserLocationExtended"  */
            decode_org_csapi_mm_TpUserLocationExtended_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mm_TpUserLocationExtended"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ul/IpAppUserLocation/extendedLocationReportErr:1.0
 */
static void
decode_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportErr_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportErr_cause, tvb, *offset-4, 4, u_octet4);

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportErr_diagnostic, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ul/IpAppUserLocation/periodicLocationReport:1.0
 */
static void
decode_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReport(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReport_locations;
    guint32   i_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReport_locations;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReport_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReport_locations = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReport_locations_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReport_locations);

        for (i_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReport_locations=0; i_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReport_locations < u_octet4_loop_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReport_locations; i_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReport_locations++) {

            /*  Begin struct "org_csapi_mm_TpUserLocationExtended"  */
            decode_org_csapi_mm_TpUserLocationExtended_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mm_TpUserLocationExtended"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ul/IpAppUserLocation/periodicLocationReportErr:1.0
 */
static void
decode_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReportErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReportErr_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReportErr_cause, tvb, *offset-4, 4, u_octet4);

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReportErr_diagnostic, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ul/IpUserLocation/locationReportReq:1.0
 */
static void
decode_org_csapi_mm_ul_IpUserLocation_locationReportReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_ul_IpUserLocation_locationReportReq_users;
    guint32   i_org_csapi_mm_ul_IpUserLocation_locationReportReq_users;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        u_octet4_loop_org_csapi_mm_ul_IpUserLocation_locationReportReq_users = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ul_IpUserLocation_locationReportReq_users_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_ul_IpUserLocation_locationReportReq_users);

        for (i_org_csapi_mm_ul_IpUserLocation_locationReportReq_users=0; i_org_csapi_mm_ul_IpUserLocation_locationReportReq_users < u_octet4_loop_org_csapi_mm_ul_IpUserLocation_locationReportReq_users; i_org_csapi_mm_ul_IpUserLocation_locationReportReq_users++) {

            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mm_ul_IpUserLocation_locationReportReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ul/IpUserLocation/extendedLocationReportReq:1.0
 */
static void
decode_org_csapi_mm_ul_IpUserLocation_extendedLocationReportReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_ul_IpUserLocation_extendedLocationReportReq_users;
    guint32   i_org_csapi_mm_ul_IpUserLocation_extendedLocationReportReq_users;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        u_octet4_loop_org_csapi_mm_ul_IpUserLocation_extendedLocationReportReq_users = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ul_IpUserLocation_extendedLocationReportReq_users_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_ul_IpUserLocation_extendedLocationReportReq_users);

        for (i_org_csapi_mm_ul_IpUserLocation_extendedLocationReportReq_users=0; i_org_csapi_mm_ul_IpUserLocation_extendedLocationReportReq_users < u_octet4_loop_org_csapi_mm_ul_IpUserLocation_extendedLocationReportReq_users; i_org_csapi_mm_ul_IpUserLocation_extendedLocationReportReq_users++) {

            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
        }

        /*  Begin struct "org_csapi_mm_TpLocationRequest"  */
        decode_org_csapi_mm_TpLocationRequest_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mm_TpLocationRequest"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mm_ul_IpUserLocation_extendedLocationReportReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ul/IpUserLocation/periodicLocationReportingStartReq:1.0
 */
static void
decode_org_csapi_mm_ul_IpUserLocation_periodicLocationReportingStartReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_ul_IpUserLocation_periodicLocationReportingStartReq_users;
    guint32   i_org_csapi_mm_ul_IpUserLocation_periodicLocationReportingStartReq_users;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        u_octet4_loop_org_csapi_mm_ul_IpUserLocation_periodicLocationReportingStartReq_users = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ul_IpUserLocation_periodicLocationReportingStartReq_users_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_ul_IpUserLocation_periodicLocationReportingStartReq_users);

        for (i_org_csapi_mm_ul_IpUserLocation_periodicLocationReportingStartReq_users=0; i_org_csapi_mm_ul_IpUserLocation_periodicLocationReportingStartReq_users < u_octet4_loop_org_csapi_mm_ul_IpUserLocation_periodicLocationReportingStartReq_users; i_org_csapi_mm_ul_IpUserLocation_periodicLocationReportingStartReq_users++) {

            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
        }

        /*  Begin struct "org_csapi_mm_TpLocationRequest"  */
        decode_org_csapi_mm_TpLocationRequest_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mm_TpLocationRequest"  */
        proto_tree_add_int(tree, hf_org_csapi_mm_ul_IpUserLocation_periodicLocationReportingStartReq_reportingInterval, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mm_ul_IpUserLocation_periodicLocationReportingStartReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ul/IpUserLocation/periodicLocationReportingStop:1.0
 */
static void
decode_org_csapi_mm_ul_IpUserLocation_periodicLocationReportingStop(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_mm_TpMobilityStopAssignmentData"  */
        decode_org_csapi_mm_TpMobilityStopAssignmentData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mm_TpMobilityStopAssignmentData"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ul/IpUserLocation/getNextPeriodicLocationRequest:1.0
 */
static void
decode_org_csapi_mm_ul_IpUserLocation_getNextPeriodicLocationRequest(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_boolean(tree, hf_org_csapi_mm_ul_IpUserLocation_getNextPeriodicLocationRequest_reset, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_mm_TpPeriodicLocationRequestSetEntry"  */
            decode_org_csapi_mm_TpPeriodicLocationRequestSetEntry_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mm_TpPeriodicLocationRequestSetEntry"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ul/IpAppTriggeredUserLocation/triggeredLocationReport:1.0
 */
static void
decode_org_csapi_mm_ul_IpAppTriggeredUserLocation_triggeredLocationReport(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_ul_IpAppTriggeredUserLocation_triggeredLocationReport_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_mm_TpUserLocationExtended"  */
        decode_org_csapi_mm_TpUserLocationExtended_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mm_TpUserLocationExtended"  */
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ul_IpAppTriggeredUserLocation_triggeredLocationReport_criterion, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ul/IpAppTriggeredUserLocation/triggeredLocationReportErr:1.0
 */
static void
decode_org_csapi_mm_ul_IpAppTriggeredUserLocation_triggeredLocationReportErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_ul_IpAppTriggeredUserLocation_triggeredLocationReportErr_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ul_IpAppTriggeredUserLocation_triggeredLocationReportErr_cause, tvb, *offset-4, 4, u_octet4);

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ul_IpAppTriggeredUserLocation_triggeredLocationReportErr_diagnostic, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ul/IpTriggeredUserLocation/triggeredLocationReportingStartReq:1.0
 */
static void
decode_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_users;
    guint32   i_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_users;
    guint32   u_octet4_loop_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_triggers;
    guint32   i_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_triggers;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        u_octet4_loop_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_users = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_users_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_users);

        for (i_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_users=0; i_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_users < u_octet4_loop_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_users; i_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_users++) {

            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
        }

        /*  Begin struct "org_csapi_mm_TpLocationRequest"  */
        decode_org_csapi_mm_TpLocationRequest_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mm_TpLocationRequest"  */
        u_octet4_loop_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_triggers = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_triggers_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_triggers);

        for (i_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_triggers=0; i_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_triggers < u_octet4_loop_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_triggers; i_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_triggers++) {

            /*  Begin struct "org_csapi_mm_TpLocationTrigger"  */
            decode_org_csapi_mm_TpLocationTrigger_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mm_TpLocationTrigger"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ul/IpTriggeredUserLocation/triggeredLocationReportingStop:1.0
 */
static void
decode_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStop(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_mm_TpMobilityStopAssignmentData"  */
        decode_org_csapi_mm_TpMobilityStopAssignmentData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mm_TpMobilityStopAssignmentData"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ul/IpTriggeredUserLocation/getNextTriggeredLocationRequest:1.0
 */
static void
decode_org_csapi_mm_ul_IpTriggeredUserLocation_getNextTriggeredLocationRequest(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_boolean(tree, hf_org_csapi_mm_ul_IpTriggeredUserLocation_getNextTriggeredLocationRequest_reset, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_mm_TpTriggeredLocationRequestSetEntry"  */
            decode_org_csapi_mm_TpTriggeredLocationRequestSetEntry_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mm_TpTriggeredLocationRequestSetEntry"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ulc/IpAppUserLocationCamel/locationReportRes:1.0
 */
static void
decode_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportRes_locations;
    guint32   i_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportRes_locations;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportRes_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportRes_locations = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportRes_locations_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportRes_locations);

        for (i_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportRes_locations=0; i_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportRes_locations < u_octet4_loop_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportRes_locations; i_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportRes_locations++) {

            /*  Begin struct "org_csapi_mm_TpUserLocationCamel"  */
            decode_org_csapi_mm_TpUserLocationCamel_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mm_TpUserLocationCamel"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ulc/IpAppUserLocationCamel/locationReportErr:1.0
 */
static void
decode_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportErr_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportErr_cause, tvb, *offset-4, 4, u_octet4);

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportErr_diagnostic, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ulc/IpAppUserLocationCamel/periodicLocationReport:1.0
 */
static void
decode_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReport(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReport_locations;
    guint32   i_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReport_locations;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReport_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReport_locations = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReport_locations_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReport_locations);

        for (i_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReport_locations=0; i_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReport_locations < u_octet4_loop_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReport_locations; i_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReport_locations++) {

            /*  Begin struct "org_csapi_mm_TpUserLocationCamel"  */
            decode_org_csapi_mm_TpUserLocationCamel_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mm_TpUserLocationCamel"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ulc/IpAppUserLocationCamel/periodicLocationReportErr:1.0
 */
static void
decode_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReportErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReportErr_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReportErr_cause, tvb, *offset-4, 4, u_octet4);

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReportErr_diagnostic, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ulc/IpAppUserLocationCamel/triggeredLocationReport:1.0
 */
static void
decode_org_csapi_mm_ulc_IpAppUserLocationCamel_triggeredLocationReport(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_ulc_IpAppUserLocationCamel_triggeredLocationReport_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_mm_TpUserLocationCamel"  */
        decode_org_csapi_mm_TpUserLocationCamel_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mm_TpUserLocationCamel"  */
        /*  Begin struct "org_csapi_mm_TpLocationTriggerCamel"  */
        decode_org_csapi_mm_TpLocationTriggerCamel_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mm_TpLocationTriggerCamel"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ulc/IpAppUserLocationCamel/triggeredLocationReportErr:1.0
 */
static void
decode_org_csapi_mm_ulc_IpAppUserLocationCamel_triggeredLocationReportErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_ulc_IpAppUserLocationCamel_triggeredLocationReportErr_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ulc_IpAppUserLocationCamel_triggeredLocationReportErr_cause, tvb, *offset-4, 4, u_octet4);

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ulc_IpAppUserLocationCamel_triggeredLocationReportErr_diagnostic, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ulc/IpUserLocationCamel/locationReportReq:1.0
 */
static void
decode_org_csapi_mm_ulc_IpUserLocationCamel_locationReportReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_ulc_IpUserLocationCamel_locationReportReq_users;
    guint32   i_org_csapi_mm_ulc_IpUserLocationCamel_locationReportReq_users;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        u_octet4_loop_org_csapi_mm_ulc_IpUserLocationCamel_locationReportReq_users = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ulc_IpUserLocationCamel_locationReportReq_users_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_ulc_IpUserLocationCamel_locationReportReq_users);

        for (i_org_csapi_mm_ulc_IpUserLocationCamel_locationReportReq_users=0; i_org_csapi_mm_ulc_IpUserLocationCamel_locationReportReq_users < u_octet4_loop_org_csapi_mm_ulc_IpUserLocationCamel_locationReportReq_users; i_org_csapi_mm_ulc_IpUserLocationCamel_locationReportReq_users++) {

            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mm_ulc_IpUserLocationCamel_locationReportReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ulc/IpUserLocationCamel/periodicLocationReportingStartReq:1.0
 */
static void
decode_org_csapi_mm_ulc_IpUserLocationCamel_periodicLocationReportingStartReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_ulc_IpUserLocationCamel_periodicLocationReportingStartReq_users;
    guint32   i_org_csapi_mm_ulc_IpUserLocationCamel_periodicLocationReportingStartReq_users;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        u_octet4_loop_org_csapi_mm_ulc_IpUserLocationCamel_periodicLocationReportingStartReq_users = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ulc_IpUserLocationCamel_periodicLocationReportingStartReq_users_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_ulc_IpUserLocationCamel_periodicLocationReportingStartReq_users);

        for (i_org_csapi_mm_ulc_IpUserLocationCamel_periodicLocationReportingStartReq_users=0; i_org_csapi_mm_ulc_IpUserLocationCamel_periodicLocationReportingStartReq_users < u_octet4_loop_org_csapi_mm_ulc_IpUserLocationCamel_periodicLocationReportingStartReq_users; i_org_csapi_mm_ulc_IpUserLocationCamel_periodicLocationReportingStartReq_users++) {

            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
        }

        proto_tree_add_int(tree, hf_org_csapi_mm_ulc_IpUserLocationCamel_periodicLocationReportingStartReq_reportingInterval, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mm_ulc_IpUserLocationCamel_periodicLocationReportingStartReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ulc/IpUserLocationCamel/periodicLocationReportingStop:1.0
 */
static void
decode_org_csapi_mm_ulc_IpUserLocationCamel_periodicLocationReportingStop(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_mm_TpMobilityStopAssignmentData"  */
        decode_org_csapi_mm_TpMobilityStopAssignmentData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mm_TpMobilityStopAssignmentData"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ulc/IpUserLocationCamel/triggeredLocationReportingStartReq:1.0
 */
static void
decode_org_csapi_mm_ulc_IpUserLocationCamel_triggeredLocationReportingStartReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_ulc_IpUserLocationCamel_triggeredLocationReportingStartReq_users;
    guint32   i_org_csapi_mm_ulc_IpUserLocationCamel_triggeredLocationReportingStartReq_users;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        u_octet4_loop_org_csapi_mm_ulc_IpUserLocationCamel_triggeredLocationReportingStartReq_users = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ulc_IpUserLocationCamel_triggeredLocationReportingStartReq_users_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_ulc_IpUserLocationCamel_triggeredLocationReportingStartReq_users);

        for (i_org_csapi_mm_ulc_IpUserLocationCamel_triggeredLocationReportingStartReq_users=0; i_org_csapi_mm_ulc_IpUserLocationCamel_triggeredLocationReportingStartReq_users < u_octet4_loop_org_csapi_mm_ulc_IpUserLocationCamel_triggeredLocationReportingStartReq_users; i_org_csapi_mm_ulc_IpUserLocationCamel_triggeredLocationReportingStartReq_users++) {

            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
        }

        /*  Begin struct "org_csapi_mm_TpLocationTriggerCamel"  */
        decode_org_csapi_mm_TpLocationTriggerCamel_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mm_TpLocationTriggerCamel"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mm_ulc_IpUserLocationCamel_triggeredLocationReportingStartReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ulc/IpUserLocationCamel/triggeredLocationReportingStop:1.0
 */
static void
decode_org_csapi_mm_ulc_IpUserLocationCamel_triggeredLocationReportingStop(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_mm_TpMobilityStopAssignmentData"  */
        decode_org_csapi_mm_TpMobilityStopAssignmentData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mm_TpMobilityStopAssignmentData"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ulc/IpUserLocationCamel/getNextPeriodicLocationRequest:1.0
 */
static void
decode_org_csapi_mm_ulc_IpUserLocationCamel_getNextPeriodicLocationRequest(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_boolean(tree, hf_org_csapi_mm_ulc_IpUserLocationCamel_getNextPeriodicLocationRequest_reset, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_mm_TpPeriodicLocationRequestSetEntry"  */
            decode_org_csapi_mm_TpPeriodicLocationRequestSetEntry_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mm_TpPeriodicLocationRequestSetEntry"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ulc/IpUserLocationCamel/getNextTriggeredLocationRequest:1.0
 */
static void
decode_org_csapi_mm_ulc_IpUserLocationCamel_getNextTriggeredLocationRequest(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_boolean(tree, hf_org_csapi_mm_ulc_IpUserLocationCamel_getNextTriggeredLocationRequest_reset, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_mm_TpTriggeredLocationRequestSetEntry"  */
            decode_org_csapi_mm_TpTriggeredLocationRequestSetEntry_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mm_TpTriggeredLocationRequestSetEntry"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ule/IpAppUserLocationEmergency/emergencyLocationReport:1.0
 */
static void
decode_org_csapi_mm_ule_IpAppUserLocationEmergency_emergencyLocationReport(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_ule_IpAppUserLocationEmergency_emergencyLocationReport_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_mm_TpUserLocationEmergency"  */
        decode_org_csapi_mm_TpUserLocationEmergency_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mm_TpUserLocationEmergency"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ule/IpAppUserLocationEmergency/emergencyLocationReportErr:1.0
 */
static void
decode_org_csapi_mm_ule_IpAppUserLocationEmergency_emergencyLocationReportErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_ule_IpAppUserLocationEmergency_emergencyLocationReportErr_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ule_IpAppUserLocationEmergency_emergencyLocationReportErr_cause, tvb, *offset-4, 4, u_octet4);

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ule_IpAppUserLocationEmergency_emergencyLocationReportErr_diagnostic, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ule/IpUserLocationEmergency/emergencyLocationReportReq:1.0
 */
static void
decode_org_csapi_mm_ule_IpUserLocationEmergency_emergencyLocationReportReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        /*  Begin struct "org_csapi_mm_TpUserLocationEmergencyRequest"  */
        decode_org_csapi_mm_TpUserLocationEmergencyRequest_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mm_TpUserLocationEmergencyRequest"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mm_ule_IpUserLocationEmergency_emergencyLocationReportReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ule/IpUserLocationEmergency/subscribeEmergencyLocationReports:1.0
 */
static void
decode_org_csapi_mm_ule_IpUserLocationEmergency_subscribeEmergencyLocationReports(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mm_ule_IpUserLocationEmergency_subscribeEmergencyLocationReports_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ule/IpUserLocationEmergency/unSubscribeEmergencyLocationReports:1.0
 */
static void
decode_org_csapi_mm_ule_IpUserLocationEmergency_unSubscribeEmergencyLocationReports(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_ule_IpUserLocationEmergency_unSubscribeEmergencyLocationReports_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/us/IpAppUserStatus/statusReportRes:1.0
 */
static void
decode_org_csapi_mm_us_IpAppUserStatus_statusReportRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_us_IpAppUserStatus_statusReportRes_status;
    guint32   i_org_csapi_mm_us_IpAppUserStatus_statusReportRes_status;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_us_IpAppUserStatus_statusReportRes_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_mm_us_IpAppUserStatus_statusReportRes_status = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_us_IpAppUserStatus_statusReportRes_status_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_us_IpAppUserStatus_statusReportRes_status);

        for (i_org_csapi_mm_us_IpAppUserStatus_statusReportRes_status=0; i_org_csapi_mm_us_IpAppUserStatus_statusReportRes_status < u_octet4_loop_org_csapi_mm_us_IpAppUserStatus_statusReportRes_status; i_org_csapi_mm_us_IpAppUserStatus_statusReportRes_status++) {

            /*  Begin struct "org_csapi_mm_TpUserStatus"  */
            decode_org_csapi_mm_TpUserStatus_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mm_TpUserStatus"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/us/IpAppUserStatus/statusReportErr:1.0
 */
static void
decode_org_csapi_mm_us_IpAppUserStatus_statusReportErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_us_IpAppUserStatus_statusReportErr_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_us_IpAppUserStatus_statusReportErr_cause, tvb, *offset-4, 4, u_octet4);

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_us_IpAppUserStatus_statusReportErr_diagnostic, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/us/IpAppUserStatus/triggeredStatusReport:1.0
 */
static void
decode_org_csapi_mm_us_IpAppUserStatus_triggeredStatusReport(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_us_IpAppUserStatus_triggeredStatusReport_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_mm_TpUserStatus"  */
        decode_org_csapi_mm_TpUserStatus_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mm_TpUserStatus"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/us/IpAppUserStatus/triggeredStatusReportErr:1.0
 */
static void
decode_org_csapi_mm_us_IpAppUserStatus_triggeredStatusReportErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_us_IpAppUserStatus_triggeredStatusReportErr_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_us_IpAppUserStatus_triggeredStatusReportErr_cause, tvb, *offset-4, 4, u_octet4);

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_us_IpAppUserStatus_triggeredStatusReportErr_diagnostic, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/us/IpAppUserStatus/extendedStatusReportRes:1.0
 */
static void
decode_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportRes_status;
    guint32   i_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportRes_status;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportRes_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportRes_status = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportRes_status_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportRes_status);

        for (i_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportRes_status=0; i_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportRes_status < u_octet4_loop_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportRes_status; i_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportRes_status++) {

            /*  Begin struct "org_csapi_mm_TpUserStatusExtended"  */
            decode_org_csapi_mm_TpUserStatusExtended_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mm_TpUserStatusExtended"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/us/IpAppUserStatus/extendedStatusReportErr:1.0
 */
static void
decode_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportErr_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportErr_cause, tvb, *offset-4, 4, u_octet4);

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportErr_diagnostic, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/us/IpAppUserStatus/extTriggeredStatusReport:1.0
 */
static void
decode_org_csapi_mm_us_IpAppUserStatus_extTriggeredStatusReport(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_us_IpAppUserStatus_extTriggeredStatusReport_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_mm_TpUserStatusExtended"  */
        decode_org_csapi_mm_TpUserStatusExtended_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mm_TpUserStatusExtended"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/us/IpAppUserStatus/extTriggeredStatusReportErr:1.0
 */
static void
decode_org_csapi_mm_us_IpAppUserStatus_extTriggeredStatusReportErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_us_IpAppUserStatus_extTriggeredStatusReportErr_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_us_IpAppUserStatus_extTriggeredStatusReportErr_cause, tvb, *offset-4, 4, u_octet4);

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_us_IpAppUserStatus_extTriggeredStatusReportErr_diagnostic, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/us/IpUserStatus/statusReportReq:1.0
 */
static void
decode_org_csapi_mm_us_IpUserStatus_statusReportReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_us_IpUserStatus_statusReportReq_users;
    guint32   i_org_csapi_mm_us_IpUserStatus_statusReportReq_users;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        u_octet4_loop_org_csapi_mm_us_IpUserStatus_statusReportReq_users = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_us_IpUserStatus_statusReportReq_users_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_us_IpUserStatus_statusReportReq_users);

        for (i_org_csapi_mm_us_IpUserStatus_statusReportReq_users=0; i_org_csapi_mm_us_IpUserStatus_statusReportReq_users < u_octet4_loop_org_csapi_mm_us_IpUserStatus_statusReportReq_users; i_org_csapi_mm_us_IpUserStatus_statusReportReq_users++) {

            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mm_us_IpUserStatus_statusReportReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/us/IpUserStatus/triggeredStatusReportingStartReq:1.0
 */
static void
decode_org_csapi_mm_us_IpUserStatus_triggeredStatusReportingStartReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_us_IpUserStatus_triggeredStatusReportingStartReq_users;
    guint32   i_org_csapi_mm_us_IpUserStatus_triggeredStatusReportingStartReq_users;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        u_octet4_loop_org_csapi_mm_us_IpUserStatus_triggeredStatusReportingStartReq_users = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_us_IpUserStatus_triggeredStatusReportingStartReq_users_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_us_IpUserStatus_triggeredStatusReportingStartReq_users);

        for (i_org_csapi_mm_us_IpUserStatus_triggeredStatusReportingStartReq_users=0; i_org_csapi_mm_us_IpUserStatus_triggeredStatusReportingStartReq_users < u_octet4_loop_org_csapi_mm_us_IpUserStatus_triggeredStatusReportingStartReq_users; i_org_csapi_mm_us_IpUserStatus_triggeredStatusReportingStartReq_users++) {

            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mm_us_IpUserStatus_triggeredStatusReportingStartReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/us/IpUserStatus/triggeredStatusReportingStop:1.0
 */
static void
decode_org_csapi_mm_us_IpUserStatus_triggeredStatusReportingStop(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_mm_TpMobilityStopAssignmentData"  */
        decode_org_csapi_mm_TpMobilityStopAssignmentData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mm_TpMobilityStopAssignmentData"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/us/IpUserStatus/getNextTriggeredStatusRequest:1.0
 */
static void
decode_org_csapi_mm_us_IpUserStatus_getNextTriggeredStatusRequest(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_boolean(tree, hf_org_csapi_mm_us_IpUserStatus_getNextTriggeredStatusRequest_reset, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_mm_TpTriggeredStatusRequestSetEntry"  */
            decode_org_csapi_mm_TpTriggeredStatusRequestSetEntry_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mm_TpTriggeredStatusRequestSetEntry"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/us/IpUserStatus/extendedStatusReportReq:1.0
 */
static void
decode_org_csapi_mm_us_IpUserStatus_extendedStatusReportReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_us_IpUserStatus_extendedStatusReportReq_users;
    guint32   i_org_csapi_mm_us_IpUserStatus_extendedStatusReportReq_users;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        u_octet4_loop_org_csapi_mm_us_IpUserStatus_extendedStatusReportReq_users = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_us_IpUserStatus_extendedStatusReportReq_users_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_us_IpUserStatus_extendedStatusReportReq_users);

        for (i_org_csapi_mm_us_IpUserStatus_extendedStatusReportReq_users=0; i_org_csapi_mm_us_IpUserStatus_extendedStatusReportReq_users < u_octet4_loop_org_csapi_mm_us_IpUserStatus_extendedStatusReportReq_users; i_org_csapi_mm_us_IpUserStatus_extendedStatusReportReq_users++) {

            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mm_us_IpUserStatus_extendedStatusReportReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/us/IpUserStatus/extTriggeredStatusReportingStartReq:1.0
 */
static void
decode_org_csapi_mm_us_IpUserStatus_extTriggeredStatusReportingStartReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_us_IpUserStatus_extTriggeredStatusReportingStartReq_users;
    guint32   i_org_csapi_mm_us_IpUserStatus_extTriggeredStatusReportingStartReq_users;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        u_octet4_loop_org_csapi_mm_us_IpUserStatus_extTriggeredStatusReportingStartReq_users = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_us_IpUserStatus_extTriggeredStatusReportingStartReq_users_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_us_IpUserStatus_extTriggeredStatusReportingStartReq_users);

        for (i_org_csapi_mm_us_IpUserStatus_extTriggeredStatusReportingStartReq_users=0; i_org_csapi_mm_us_IpUserStatus_extTriggeredStatusReportingStartReq_users < u_octet4_loop_org_csapi_mm_us_IpUserStatus_extTriggeredStatusReportingStartReq_users; i_org_csapi_mm_us_IpUserStatus_extTriggeredStatusReportingStartReq_users++) {

            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mm_us_IpUserStatus_extTriggeredStatusReportingStartReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/us/IpUserStatus/extTriggeredStatusReportingStop:1.0
 */
static void
decode_org_csapi_mm_us_IpUserStatus_extTriggeredStatusReportingStop(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_mm_TpMobilityStopAssignmentData"  */
        decode_org_csapi_mm_TpMobilityStopAssignmentData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mm_TpMobilityStopAssignmentData"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ub/IpAppUserBinding/triggeredBindingRequestNotification:1.0
 */
static void
decode_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_contactInformation;
    guint32   i_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_contactInformation;
    guint32   u_octet4_loop_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_return;
    guint32   i_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        u_octet4_loop_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_contactInformation = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_contactInformation_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_contactInformation);

        for (i_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_contactInformation=0; i_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_contactInformation < u_octet4_loop_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_contactInformation; i_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_contactInformation++) {

            u_octet4_loop_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_contactInformation = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_contactInformation_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_contactInformation);

            for (i_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_contactInformation=0; i_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_contactInformation < u_octet4_loop_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_contactInformation; i_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_contactInformation++) {

                /*  Begin union "org_csapi_mm_TpBindingEntry"  */
                decode_org_csapi_mm_TpBindingEntry_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

                /*  End union "org_csapi_mm_TpBindingEntry"  */
            }

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_return);

            for (i_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_return=0; i_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_return < u_octet4_loop_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_return; i_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_return++) {

                u_octet4_loop_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
                proto_tree_add_uint(tree, hf_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_return);

                for (i_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_return=0; i_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_return < u_octet4_loop_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_return; i_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_return++) {

                    /*  Begin union "org_csapi_mm_TpBindingEntry"  */
                    decode_org_csapi_mm_TpBindingEntry_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

                    /*  End union "org_csapi_mm_TpBindingEntry"  */
                }

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ub/IpAppUserBinding/triggeredBindingRequestNotificationStartErr:1.0
 */
static void
decode_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotificationStartErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotificationStartErr_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotificationStartErr_cause, tvb, *offset-4, 4, u_octet4);

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotificationStartErr_diagnostic, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ub/IpUserBinding/triggeredBindingRequestNotificationStartReq:1.0
 */
static void
decode_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_users;
    guint32   i_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_users;
    guint32   u_octet4_loop_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_criteria;
    guint32   i_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_criteria;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        u_octet4_loop_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_users = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_users_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_users);

        for (i_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_users=0; i_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_users < u_octet4_loop_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_users; i_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_users++) {

            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
        }

        u_octet4_loop_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_criteria = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_criteria_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_criteria);

        for (i_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_criteria=0; i_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_criteria < u_octet4_loop_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_criteria; i_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_criteria++) {

            /*  Begin struct "org_csapi_mm_TpBindingNotificationCriteria"  */
            decode_org_csapi_mm_TpBindingNotificationCriteria_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mm_TpBindingNotificationCriteria"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mm/ub/IpUserBinding/triggeredBindingRequestNotificationStop:1.0
 */
static void
decode_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStop(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStop_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/termcap/IpTerminalCapabilities/getTerminalCapabilities:1.0
 */
static void
decode_org_csapi_termcap_IpTerminalCapabilities_getTerminalCapabilities(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_termcap_IpTerminalCapabilities_getTerminalCapabilities_terminalIdentity);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_termcap_TpTerminalCapabilities"  */
            decode_org_csapi_termcap_TpTerminalCapabilities_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_termcap_TpTerminalCapabilities"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/termcap/IpAppExtendedTerminalCapabilities/triggeredTerminalCapabilityReport:1.0
 */
static void
decode_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReport(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReport_terminals;
    guint32   i_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReport_terminals;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReport_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReport_terminals = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReport_terminals_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReport_terminals);

        for (i_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReport_terminals=0; i_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReport_terminals < u_octet4_loop_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReport_terminals; i_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReport_terminals++) {

            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
        }

        proto_tree_add_int(tree, hf_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReport_criteria, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_termcap_TpTerminalCapabilities"  */
        decode_org_csapi_termcap_TpTerminalCapabilities_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_termcap_TpTerminalCapabilities"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/termcap/IpAppExtendedTerminalCapabilities/triggeredTerminalCapabilityReportErr:1.0
 */
static void
decode_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReportErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReportErr_terminals;
    guint32   i_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReportErr_terminals;
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReportErr_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReportErr_terminals = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReportErr_terminals_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReportErr_terminals);

        for (i_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReportErr_terminals=0; i_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReportErr_terminals < u_octet4_loop_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReportErr_terminals; i_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReportErr_terminals++) {

            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
        }

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReportErr_cause, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/termcap/IpExtendedTerminalCapabilities/triggeredTerminalCapabilityStartReq:1.0
 */
static void
decode_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStartReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStartReq_terminals;
    guint32   i_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStartReq_terminals;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        u_octet4_loop_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStartReq_terminals = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStartReq_terminals_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStartReq_terminals);

        for (i_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStartReq_terminals=0; i_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStartReq_terminals < u_octet4_loop_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStartReq_terminals; i_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStartReq_terminals++) {

            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
        }

        /*  Begin struct "org_csapi_termcap_TpTerminalCapabilityScope"  */
        decode_org_csapi_termcap_TpTerminalCapabilityScope_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_termcap_TpTerminalCapabilityScope"  */
        proto_tree_add_int(tree, hf_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStartReq_criteria, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStartReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/termcap/IpExtendedTerminalCapabilities/triggeredTerminalCapabilityStop:1.0
 */
static void
decode_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStop(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStop_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpAppDataSession/connectRes:1.0
 */
static void
decode_org_csapi_dsc_IpAppDataSession_connectRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_dsc_IpAppDataSession_connectRes_dataSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_dsc_TpDataSessionReport"  */
        decode_org_csapi_dsc_TpDataSessionReport_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_dsc_TpDataSessionReport"  */
        proto_tree_add_int(tree, hf_org_csapi_dsc_IpAppDataSession_connectRes_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpAppDataSession/connectErr:1.0
 */
static void
decode_org_csapi_dsc_IpAppDataSession_connectErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_dsc_IpAppDataSession_connectErr_dataSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_dsc_TpDataSessionError"  */
        decode_org_csapi_dsc_TpDataSessionError_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_dsc_TpDataSessionError"  */
        proto_tree_add_int(tree, hf_org_csapi_dsc_IpAppDataSession_connectErr_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpAppDataSession/superviseDataSessionRes:1.0
 */
static void
decode_org_csapi_dsc_IpAppDataSession_superviseDataSessionRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_dsc_IpAppDataSession_superviseDataSessionRes_dataSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_dsc_IpAppDataSession_superviseDataSessionRes_report, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_dsc_TpDataSessionSuperviseVolume"  */
        decode_org_csapi_dsc_TpDataSessionSuperviseVolume_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_dsc_TpDataSessionSuperviseVolume"  */
        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_dsc_IpAppDataSession_superviseDataSessionRes_qualityOfService, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpAppDataSession/superviseDataSessionErr:1.0
 */
static void
decode_org_csapi_dsc_IpAppDataSession_superviseDataSessionErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_dsc_IpAppDataSession_superviseDataSessionErr_dataSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_dsc_TpDataSessionError"  */
        decode_org_csapi_dsc_TpDataSessionError_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_dsc_TpDataSessionError"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpAppDataSession/dataSessionFaultDetected:1.0
 */
static void
decode_org_csapi_dsc_IpAppDataSession_dataSessionFaultDetected(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_dsc_IpAppDataSession_dataSessionFaultDetected_dataSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_dsc_IpAppDataSession_dataSessionFaultDetected_fault, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpAppDataSessionControlManager/dataSessionAborted:1.0
 */
static void
decode_org_csapi_dsc_IpAppDataSessionControlManager_dataSessionAborted(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_dsc_IpAppDataSessionControlManager_dataSessionAborted_dataSession, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpAppDataSessionControlManager/reportNotification:1.0
 */
static void
decode_org_csapi_dsc_IpAppDataSessionControlManager_reportNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_dsc_TpDataSessionIdentifier"  */
        decode_org_csapi_dsc_TpDataSessionIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_dsc_TpDataSessionIdentifier"  */
        /*  Begin struct "org_csapi_dsc_TpDataSessionEventInfo"  */
        decode_org_csapi_dsc_TpDataSessionEventInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_dsc_TpDataSessionEventInfo"  */
        proto_tree_add_int(tree, hf_org_csapi_dsc_IpAppDataSessionControlManager_reportNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpAppDataSessionControlManager/dataSessionNotificationContinued:1.0
 */
static void
decode_org_csapi_dsc_IpAppDataSessionControlManager_dataSessionNotificationContinued(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpAppDataSessionControlManager/dataSessionNotificationInterrupted:1.0
 */
static void
decode_org_csapi_dsc_IpAppDataSessionControlManager_dataSessionNotificationInterrupted(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpAppDataSessionControlManager/abortMultipleDataSessions:1.0
 */
static void
decode_org_csapi_dsc_IpAppDataSessionControlManager_abortMultipleDataSessions(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_dsc_IpAppDataSessionControlManager_abortMultipleDataSessions_dataSessionSet;
    guint32   i_org_csapi_dsc_IpAppDataSessionControlManager_abortMultipleDataSessions_dataSessionSet;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_dsc_IpAppDataSessionControlManager_abortMultipleDataSessions_dataSessionSet = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_dsc_IpAppDataSessionControlManager_abortMultipleDataSessions_dataSessionSet_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_dsc_IpAppDataSessionControlManager_abortMultipleDataSessions_dataSessionSet);

        for (i_org_csapi_dsc_IpAppDataSessionControlManager_abortMultipleDataSessions_dataSessionSet=0; i_org_csapi_dsc_IpAppDataSessionControlManager_abortMultipleDataSessions_dataSessionSet < u_octet4_loop_org_csapi_dsc_IpAppDataSessionControlManager_abortMultipleDataSessions_dataSessionSet; i_org_csapi_dsc_IpAppDataSessionControlManager_abortMultipleDataSessions_dataSessionSet++) {

            proto_tree_add_int(tree, hf_org_csapi_dsc_IpAppDataSessionControlManager_abortMultipleDataSessions_dataSessionSet, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpDataSession/connectReq:1.0
 */
static void
decode_org_csapi_dsc_IpDataSession_connectReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_dsc_IpDataSession_connectReq_responseRequested;
    guint32   i_org_csapi_dsc_IpDataSession_connectReq_responseRequested;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_dsc_IpDataSession_connectReq_dataSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_dsc_IpDataSession_connectReq_responseRequested = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_dsc_IpDataSession_connectReq_responseRequested_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_dsc_IpDataSession_connectReq_responseRequested);

        for (i_org_csapi_dsc_IpDataSession_connectReq_responseRequested=0; i_org_csapi_dsc_IpDataSession_connectReq_responseRequested < u_octet4_loop_org_csapi_dsc_IpDataSession_connectReq_responseRequested; i_org_csapi_dsc_IpDataSession_connectReq_responseRequested++) {

            /*  Begin struct "org_csapi_dsc_TpDataSessionReportRequest"  */
            decode_org_csapi_dsc_TpDataSessionReportRequest_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_dsc_TpDataSessionReportRequest"  */
        }

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_dsc_IpDataSession_connectReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpDataSession/release:1.0
 */
static void
decode_org_csapi_dsc_IpDataSession_release(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_dsc_IpDataSession_release_dataSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_dsc_TpDataSessionReleaseCause"  */
        decode_org_csapi_dsc_TpDataSessionReleaseCause_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_dsc_TpDataSessionReleaseCause"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpDataSession/superviseDataSessionReq:1.0
 */
static void
decode_org_csapi_dsc_IpDataSession_superviseDataSessionReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_dsc_IpDataSession_superviseDataSessionReq_dataSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_dsc_IpDataSession_superviseDataSessionReq_treatment, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_dsc_TpDataSessionSuperviseVolume"  */
        decode_org_csapi_dsc_TpDataSessionSuperviseVolume_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_dsc_TpDataSessionSuperviseVolume"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpDataSession/setDataSessionChargePlan:1.0
 */
static void
decode_org_csapi_dsc_IpDataSession_setDataSessionChargePlan(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_dsc_IpDataSession_setDataSessionChargePlan_dataSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_dsc_TpDataSessionChargePlan"  */
        decode_org_csapi_dsc_TpDataSessionChargePlan_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_dsc_TpDataSessionChargePlan"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpDataSession/setAdviceOfCharge:1.0
 */
static void
decode_org_csapi_dsc_IpDataSession_setAdviceOfCharge(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_dsc_IpDataSession_setAdviceOfCharge_dataSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_TpAoCInfo"  */
        decode_org_csapi_TpAoCInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAoCInfo"  */
        proto_tree_add_int(tree, hf_org_csapi_dsc_IpDataSession_setAdviceOfCharge_tariffSwitch, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpDataSession/deassignDataSession:1.0
 */
static void
decode_org_csapi_dsc_IpDataSession_deassignDataSession(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_dsc_IpDataSession_deassignDataSession_dataSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpDataSession/continueProcessing:1.0
 */
static void
decode_org_csapi_dsc_IpDataSession_continueProcessing(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_dsc_IpDataSession_continueProcessing_dataSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpDataSessionControlManager/createNotification:1.0
 */
static void
decode_org_csapi_dsc_IpDataSessionControlManager_createNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        /*  Begin struct "org_csapi_dsc_TpDataSessionEventCriteria"  */
        decode_org_csapi_dsc_TpDataSessionEventCriteria_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_dsc_TpDataSessionEventCriteria"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_dsc_IpDataSessionControlManager_createNotification_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpDataSessionControlManager/destroyNotification:1.0
 */
static void
decode_org_csapi_dsc_IpDataSessionControlManager_destroyNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_dsc_IpDataSessionControlManager_destroyNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpDataSessionControlManager/changeNotification:1.0
 */
static void
decode_org_csapi_dsc_IpDataSessionControlManager_changeNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_dsc_IpDataSessionControlManager_changeNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_dsc_TpDataSessionEventCriteria"  */
        decode_org_csapi_dsc_TpDataSessionEventCriteria_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_dsc_TpDataSessionEventCriteria"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpDataSessionControlManager/enableNotifications:1.0
 */
static void
decode_org_csapi_dsc_IpDataSessionControlManager_enableNotifications(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_dsc_IpDataSessionControlManager_enableNotifications_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpDataSessionControlManager/disableNotifications:1.0
 */
static void
decode_org_csapi_dsc_IpDataSessionControlManager_disableNotifications(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpDataSessionControlManager/getNotifications:1.0
 */
static void
decode_org_csapi_dsc_IpDataSessionControlManager_getNotifications(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_dsc_IpDataSessionControlManager_getNotifications_return;
    guint32   i_org_csapi_dsc_IpDataSessionControlManager_getNotifications_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_dsc_IpDataSessionControlManager_getNotifications_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_dsc_IpDataSessionControlManager_getNotifications_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_dsc_IpDataSessionControlManager_getNotifications_return);

            for (i_org_csapi_dsc_IpDataSessionControlManager_getNotifications_return=0; i_org_csapi_dsc_IpDataSessionControlManager_getNotifications_return < u_octet4_loop_org_csapi_dsc_IpDataSessionControlManager_getNotifications_return; i_org_csapi_dsc_IpDataSessionControlManager_getNotifications_return++) {

                /*  Begin struct "org_csapi_dsc_TpDataSessionEventCriteriaResult"  */
                decode_org_csapi_dsc_TpDataSessionEventCriteriaResult_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_dsc_TpDataSessionEventCriteriaResult"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/dsc/IpDataSessionControlManager/createNotifications:1.0
 */
static void
decode_org_csapi_dsc_IpDataSessionControlManager_createNotifications(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        /*  Begin struct "org_csapi_dsc_TpDataSessionEventCriteria"  */
        decode_org_csapi_dsc_TpDataSessionEventCriteria_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_dsc_TpDataSessionEventCriteria"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_dsc_IpDataSessionControlManager_createNotifications_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMessagingManager/openMailbox:1.0
 */
static void
decode_org_csapi_gms_IpMessagingManager_openMailbox(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_IpMessagingManager_openMailbox_authenticationInfo);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_gms_TpMailboxIdentifier"  */
            decode_org_csapi_gms_TpMailboxIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_gms_TpMailboxIdentifier"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMessagingManager/enableMessagingNotification:1.0
 */
static void
decode_org_csapi_gms_IpMessagingManager_enableMessagingNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        /*  Begin union "org_csapi_gms_TpMessagingEventCriteria"  */
        decode_org_csapi_gms_TpMessagingEventCriteria_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_gms_TpMessagingEventCriteria"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_gms_IpMessagingManager_enableMessagingNotification_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMessagingManager/disableMessagingNotification:1.0
 */
static void
decode_org_csapi_gms_IpMessagingManager_disableMessagingNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_gms_IpMessagingManager_disableMessagingNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpAppMessagingManager/mailboxTerminated:1.0
 */
static void
decode_org_csapi_gms_IpAppMessagingManager_mailboxTerminated(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        proto_tree_add_int(tree, hf_org_csapi_gms_IpAppMessagingManager_mailboxTerminated_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpAppMessagingManager/mailboxFaultDetected:1.0
 */
static void
decode_org_csapi_gms_IpAppMessagingManager_mailboxFaultDetected(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        proto_tree_add_int(tree, hf_org_csapi_gms_IpAppMessagingManager_mailboxFaultDetected_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_gms_IpAppMessagingManager_mailboxFaultDetected_fault, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpAppMessagingManager/messagingEventNotify:1.0
 */
static void
decode_org_csapi_gms_IpAppMessagingManager_messagingEventNotify(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        /*  Begin union "org_csapi_gms_TpMessagingEventInfo"  */
        decode_org_csapi_gms_TpMessagingEventInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_gms_TpMessagingEventInfo"  */
        proto_tree_add_int(tree, hf_org_csapi_gms_IpAppMessagingManager_messagingEventNotify_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpAppMessagingManager/messagingNotificationTerminated:1.0
 */
static void
decode_org_csapi_gms_IpAppMessagingManager_messagingNotificationTerminated(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMailbox/close:1.0
 */
static void
decode_org_csapi_gms_IpMailbox_close(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_gms_IpMailbox_close_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMailbox/lock:1.0
 */
static void
decode_org_csapi_gms_IpMailbox_lock(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_gms_IpMailbox_lock_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMailbox/unlock:1.0
 */
static void
decode_org_csapi_gms_IpMailbox_unlock(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_gms_IpMailbox_unlock_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMailbox/getInfoAmount:1.0
 */
static void
decode_org_csapi_gms_IpMailbox_getInfoAmount(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_gms_IpMailbox_getInfoAmount_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_gms_IpMailbox_getInfoAmount_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMailbox/getInfoProperties:1.0
 */
static void
decode_org_csapi_gms_IpMailbox_getInfoProperties(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_gms_IpMailbox_getInfoProperties_return;
    guint32   i_org_csapi_gms_IpMailbox_getInfoProperties_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_gms_IpMailbox_getInfoProperties_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_gms_IpMailbox_getInfoProperties_firstProperty, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_gms_IpMailbox_getInfoProperties_numberOfProperties, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_gms_IpMailbox_getInfoProperties_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_gms_IpMailbox_getInfoProperties_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_gms_IpMailbox_getInfoProperties_return);

            for (i_org_csapi_gms_IpMailbox_getInfoProperties_return=0; i_org_csapi_gms_IpMailbox_getInfoProperties_return < u_octet4_loop_org_csapi_gms_IpMailbox_getInfoProperties_return; i_org_csapi_gms_IpMailbox_getInfoProperties_return++) {

                /*  Begin union "org_csapi_gms_TpMailboxInfoProperty"  */
                decode_org_csapi_gms_TpMailboxInfoProperty_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

                /*  End union "org_csapi_gms_TpMailboxInfoProperty"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMailbox/setInfoProperties:1.0
 */
static void
decode_org_csapi_gms_IpMailbox_setInfoProperties(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_gms_IpMailbox_setInfoProperties_mailboxInfoProperties;
    guint32   i_org_csapi_gms_IpMailbox_setInfoProperties_mailboxInfoProperties;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_gms_IpMailbox_setInfoProperties_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_gms_IpMailbox_setInfoProperties_firstProperty, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_gms_IpMailbox_setInfoProperties_mailboxInfoProperties = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_gms_IpMailbox_setInfoProperties_mailboxInfoProperties_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_gms_IpMailbox_setInfoProperties_mailboxInfoProperties);

        for (i_org_csapi_gms_IpMailbox_setInfoProperties_mailboxInfoProperties=0; i_org_csapi_gms_IpMailbox_setInfoProperties_mailboxInfoProperties < u_octet4_loop_org_csapi_gms_IpMailbox_setInfoProperties_mailboxInfoProperties; i_org_csapi_gms_IpMailbox_setInfoProperties_mailboxInfoProperties++) {

            /*  Begin union "org_csapi_gms_TpMailboxInfoProperty"  */
            decode_org_csapi_gms_TpMailboxInfoProperty_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_gms_TpMailboxInfoProperty"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMailbox/openFolder:1.0
 */
static void
decode_org_csapi_gms_IpMailbox_openFolder(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_gms_IpMailbox_openFolder_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_IpMailbox_openFolder_folderID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_gms_TpMailboxFolderIdentifier"  */
            decode_org_csapi_gms_TpMailboxFolderIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_gms_TpMailboxFolderIdentifier"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMailbox/createFolder:1.0
 */
static void
decode_org_csapi_gms_IpMailbox_createFolder(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_gms_IpMailbox_createFolder_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_IpMailbox_createFolder_folderID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMailbox/remove:1.0
 */
static void
decode_org_csapi_gms_IpMailbox_remove(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_IpMailbox_remove_authenticationInfo);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMessage/getInfoAmount:1.0
 */
static void
decode_org_csapi_gms_IpMessage_getInfoAmount(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_gms_IpMessage_getInfoAmount_folderSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_IpMessage_getInfoAmount_messageID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_gms_IpMessage_getInfoAmount_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMessage/getInfoProperties:1.0
 */
static void
decode_org_csapi_gms_IpMessage_getInfoProperties(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_gms_IpMessage_getInfoProperties_return;
    guint32   i_org_csapi_gms_IpMessage_getInfoProperties_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_gms_IpMessage_getInfoProperties_folderSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_IpMessage_getInfoProperties_messageID);

        proto_tree_add_int(tree, hf_org_csapi_gms_IpMessage_getInfoProperties_firstProperty, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_gms_IpMessage_getInfoProperties_numberOfProperties, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_gms_IpMessage_getInfoProperties_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_gms_IpMessage_getInfoProperties_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_gms_IpMessage_getInfoProperties_return);

            for (i_org_csapi_gms_IpMessage_getInfoProperties_return=0; i_org_csapi_gms_IpMessage_getInfoProperties_return < u_octet4_loop_org_csapi_gms_IpMessage_getInfoProperties_return; i_org_csapi_gms_IpMessage_getInfoProperties_return++) {

                /*  Begin union "org_csapi_gms_TpMessageInfoProperty"  */
                decode_org_csapi_gms_TpMessageInfoProperty_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

                /*  End union "org_csapi_gms_TpMessageInfoProperty"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMessage/setInfoProperties:1.0
 */
static void
decode_org_csapi_gms_IpMessage_setInfoProperties(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_gms_IpMessage_setInfoProperties_messageInfoProperties;
    guint32   i_org_csapi_gms_IpMessage_setInfoProperties_messageInfoProperties;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_gms_IpMessage_setInfoProperties_folderSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_IpMessage_setInfoProperties_messageID);

        proto_tree_add_int(tree, hf_org_csapi_gms_IpMessage_setInfoProperties_firstProperty, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_gms_IpMessage_setInfoProperties_messageInfoProperties = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_gms_IpMessage_setInfoProperties_messageInfoProperties_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_gms_IpMessage_setInfoProperties_messageInfoProperties);

        for (i_org_csapi_gms_IpMessage_setInfoProperties_messageInfoProperties=0; i_org_csapi_gms_IpMessage_setInfoProperties_messageInfoProperties < u_octet4_loop_org_csapi_gms_IpMessage_setInfoProperties_messageInfoProperties; i_org_csapi_gms_IpMessage_setInfoProperties_messageInfoProperties++) {

            /*  Begin union "org_csapi_gms_TpMessageInfoProperty"  */
            decode_org_csapi_gms_TpMessageInfoProperty_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_gms_TpMessageInfoProperty"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMessage/remove:1.0
 */
static void
decode_org_csapi_gms_IpMessage_remove(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_gms_IpMessage_remove_folderSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_IpMessage_remove_messageID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMessage/getContent:1.0
 */
static void
decode_org_csapi_gms_IpMessage_getContent(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_gms_IpMessage_getContent_folderSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_IpMessage_getContent_messageID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_IpMessage_getContent_return);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMailboxFolder/getInfoAmount:1.0
 */
static void
decode_org_csapi_gms_IpMailboxFolder_getInfoAmount(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_gms_IpMailboxFolder_getInfoAmount_folderSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_gms_IpMailboxFolder_getInfoAmount_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMailboxFolder/getInfoProperties:1.0
 */
static void
decode_org_csapi_gms_IpMailboxFolder_getInfoProperties(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_gms_IpMailboxFolder_getInfoProperties_return;
    guint32   i_org_csapi_gms_IpMailboxFolder_getInfoProperties_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_gms_IpMailboxFolder_getInfoProperties_folderSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_gms_IpMailboxFolder_getInfoProperties_firstProperty, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_gms_IpMailboxFolder_getInfoProperties_numberOfProperties, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_gms_IpMailboxFolder_getInfoProperties_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_gms_IpMailboxFolder_getInfoProperties_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_gms_IpMailboxFolder_getInfoProperties_return);

            for (i_org_csapi_gms_IpMailboxFolder_getInfoProperties_return=0; i_org_csapi_gms_IpMailboxFolder_getInfoProperties_return < u_octet4_loop_org_csapi_gms_IpMailboxFolder_getInfoProperties_return; i_org_csapi_gms_IpMailboxFolder_getInfoProperties_return++) {

                /*  Begin union "org_csapi_gms_TpFolderInfoProperty"  */
                decode_org_csapi_gms_TpFolderInfoProperty_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

                /*  End union "org_csapi_gms_TpFolderInfoProperty"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMailboxFolder/setInfoProperties:1.0
 */
static void
decode_org_csapi_gms_IpMailboxFolder_setInfoProperties(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_gms_IpMailboxFolder_setInfoProperties_folderInfoProperties;
    guint32   i_org_csapi_gms_IpMailboxFolder_setInfoProperties_folderInfoProperties;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_gms_IpMailboxFolder_setInfoProperties_folderSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_gms_IpMailboxFolder_setInfoProperties_firstProperty, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_gms_IpMailboxFolder_setInfoProperties_folderInfoProperties = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_gms_IpMailboxFolder_setInfoProperties_folderInfoProperties_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_gms_IpMailboxFolder_setInfoProperties_folderInfoProperties);

        for (i_org_csapi_gms_IpMailboxFolder_setInfoProperties_folderInfoProperties=0; i_org_csapi_gms_IpMailboxFolder_setInfoProperties_folderInfoProperties < u_octet4_loop_org_csapi_gms_IpMailboxFolder_setInfoProperties_folderInfoProperties; i_org_csapi_gms_IpMailboxFolder_setInfoProperties_folderInfoProperties++) {

            /*  Begin union "org_csapi_gms_TpFolderInfoProperty"  */
            decode_org_csapi_gms_TpFolderInfoProperty_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_gms_TpFolderInfoProperty"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMailboxFolder/putMessage:1.0
 */
static void
decode_org_csapi_gms_IpMailboxFolder_putMessage(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_gms_IpMailboxFolder_putMessage_messageInfoProperties;
    guint32   i_org_csapi_gms_IpMailboxFolder_putMessage_messageInfoProperties;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_gms_IpMailboxFolder_putMessage_folderSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_IpMailboxFolder_putMessage_message);

        u_octet4_loop_org_csapi_gms_IpMailboxFolder_putMessage_messageInfoProperties = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_gms_IpMailboxFolder_putMessage_messageInfoProperties_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_gms_IpMailboxFolder_putMessage_messageInfoProperties);

        for (i_org_csapi_gms_IpMailboxFolder_putMessage_messageInfoProperties=0; i_org_csapi_gms_IpMailboxFolder_putMessage_messageInfoProperties < u_octet4_loop_org_csapi_gms_IpMailboxFolder_putMessage_messageInfoProperties; i_org_csapi_gms_IpMailboxFolder_putMessage_messageInfoProperties++) {

            /*  Begin union "org_csapi_gms_TpMessageInfoProperty"  */
            decode_org_csapi_gms_TpMessageInfoProperty_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_gms_TpMessageInfoProperty"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMailboxFolder/getMessage:1.0
 */
static void
decode_org_csapi_gms_IpMailboxFolder_getMessage(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_gms_IpMailboxFolder_getMessage_folderSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_IpMailboxFolder_getMessage_messageID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMailboxFolder/close:1.0
 */
static void
decode_org_csapi_gms_IpMailboxFolder_close(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_gms_IpMailboxFolder_close_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_gms_IpMailboxFolder_close_folderSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/gms/IpMailboxFolder/remove:1.0
 */
static void
decode_org_csapi_gms_IpMailboxFolder_remove(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_gms_IpMailboxFolder_remove_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_IpMailboxFolder_remove_folderID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpVPrP/getVPrPID:1.0
 */
static void
decode_org_csapi_cm_IpVPrP_getVPrPID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_IpVPrP_getVPrPID_return);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpVPrP/getSlaID:1.0
 */
static void
decode_org_csapi_cm_IpVPrP_getSlaID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_IpVPrP_getSlaID_return);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpVPrP/getStatus:1.0
 */
static void
decode_org_csapi_cm_IpVPrP_getStatus(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_cm_IpVPrP_getStatus_return, tvb, *offset-4, 4, u_octet4);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpVPrP/getProvisionedQoSInfo:1.0
 */
static void
decode_org_csapi_cm_IpVPrP_getProvisionedQoSInfo(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cm_TpProvisionedQoSInfo"  */
            decode_org_csapi_cm_TpProvisionedQoSInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cm_TpProvisionedQoSInfo"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpVPrP/getValidityInfo:1.0
 */
static void
decode_org_csapi_cm_IpVPrP_getValidityInfo(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cm_TpValidityInfo"  */
            decode_org_csapi_cm_TpValidityInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cm_TpValidityInfo"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpVPrP/getPipeQoSInfo:1.0
 */
static void
decode_org_csapi_cm_IpVPrP_getPipeQoSInfo(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cm_TpPipeQoSInfo"  */
            decode_org_csapi_cm_TpPipeQoSInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cm_TpPipeQoSInfo"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpVPrP/getDsCodepoint:1.0
 */
static void
decode_org_csapi_cm_IpVPrP_getDsCodepoint(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cm_TpDsCodepoint"  */
            decode_org_csapi_cm_TpDsCodepoint_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cm_TpDsCodepoint"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpVPrN/getVPrPList:1.0
 */
static void
decode_org_csapi_cm_IpVPrN_getVPrPList(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cm_IpVPrN_getVPrPList_return;
    guint32   i_org_csapi_cm_IpVPrN_getVPrPList_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_cm_IpVPrN_getVPrPList_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_cm_IpVPrN_getVPrPList_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cm_IpVPrN_getVPrPList_return);

            for (i_org_csapi_cm_IpVPrN_getVPrPList_return=0; i_org_csapi_cm_IpVPrN_getVPrPList_return < u_octet4_loop_org_csapi_cm_IpVPrN_getVPrPList_return; i_org_csapi_cm_IpVPrN_getVPrPList_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_IpVPrN_getVPrPList_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpVPrN/getVPrP:1.0
 */
static void
decode_org_csapi_cm_IpVPrN_getVPrP(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_IpVPrN_getVPrP_vPrPID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpVPrN/createVPrP:1.0
 */
static void
decode_org_csapi_cm_IpVPrN_createVPrP(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpVPrN/deleteVPrP:1.0
 */
static void
decode_org_csapi_cm_IpVPrN_deleteVPrP(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_IpVPrN_deleteVPrP_vPrPID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpQoSMenu/getTemplate:1.0
 */
static void
decode_org_csapi_cm_IpQoSMenu_getTemplate(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_IpQoSMenu_getTemplate_templateType);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpQoSMenu/getTemplateList:1.0
 */
static void
decode_org_csapi_cm_IpQoSMenu_getTemplateList(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cm_IpQoSMenu_getTemplateList_return;
    guint32   i_org_csapi_cm_IpQoSMenu_getTemplateList_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_cm_IpQoSMenu_getTemplateList_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_cm_IpQoSMenu_getTemplateList_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cm_IpQoSMenu_getTemplateList_return);

            for (i_org_csapi_cm_IpQoSMenu_getTemplateList_return=0; i_org_csapi_cm_IpQoSMenu_getTemplateList_return < u_octet4_loop_org_csapi_cm_IpQoSMenu_getTemplateList_return; i_org_csapi_cm_IpQoSMenu_getTemplateList_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_IpQoSMenu_getTemplateList_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpConnectivityManager/getQoSMenu:1.0
 */
static void
decode_org_csapi_cm_IpConnectivityManager_getQoSMenu(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpConnectivityManager/getEnterpriseNetwork:1.0
 */
static void
decode_org_csapi_cm_IpConnectivityManager_getEnterpriseNetwork(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpQoSTemplate/getTemplateType:1.0
 */
static void
decode_org_csapi_cm_IpQoSTemplate_getTemplateType(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_IpQoSTemplate_getTemplateType_return);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpQoSTemplate/getDescription:1.0
 */
static void
decode_org_csapi_cm_IpQoSTemplate_getDescription(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_IpQoSTemplate_getDescription_return);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpQoSTemplate/setSlaID:1.0
 */
static void
decode_org_csapi_cm_IpQoSTemplate_setSlaID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_IpQoSTemplate_setSlaID_slaID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpQoSTemplate/getPipeQoSInfo:1.0
 */
static void
decode_org_csapi_cm_IpQoSTemplate_getPipeQoSInfo(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cm_TpPipeQoSInfo"  */
            decode_org_csapi_cm_TpPipeQoSInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cm_TpPipeQoSInfo"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpQoSTemplate/setPipeQoSInfo:1.0
 */
static void
decode_org_csapi_cm_IpQoSTemplate_setPipeQoSInfo(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_cm_TpPipeQoSInfo"  */
        decode_org_csapi_cm_TpPipeQoSInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cm_TpPipeQoSInfo"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpQoSTemplate/getValidityInfo:1.0
 */
static void
decode_org_csapi_cm_IpQoSTemplate_getValidityInfo(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cm_TpValidityInfo"  */
            decode_org_csapi_cm_TpValidityInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cm_TpValidityInfo"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpQoSTemplate/setValidityInfo:1.0
 */
static void
decode_org_csapi_cm_IpQoSTemplate_setValidityInfo(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_cm_TpValidityInfo"  */
        decode_org_csapi_cm_TpValidityInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cm_TpValidityInfo"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpQoSTemplate/setProvisionedQoSInfo:1.0
 */
static void
decode_org_csapi_cm_IpQoSTemplate_setProvisionedQoSInfo(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_cm_TpProvisionedQoSInfo"  */
        decode_org_csapi_cm_TpProvisionedQoSInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cm_TpProvisionedQoSInfo"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpQoSTemplate/getProvisionedQoSInfo:1.0
 */
static void
decode_org_csapi_cm_IpQoSTemplate_getProvisionedQoSInfo(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cm_TpProvisionedQoSInfo"  */
            decode_org_csapi_cm_TpProvisionedQoSInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cm_TpProvisionedQoSInfo"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpQoSTemplate/getDsCodepoint:1.0
 */
static void
decode_org_csapi_cm_IpQoSTemplate_getDsCodepoint(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cm_TpDsCodepoint"  */
            decode_org_csapi_cm_TpDsCodepoint_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cm_TpDsCodepoint"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpEnterpriseNetwork/getSiteList:1.0
 */
static void
decode_org_csapi_cm_IpEnterpriseNetwork_getSiteList(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cm_IpEnterpriseNetwork_getSiteList_return;
    guint32   i_org_csapi_cm_IpEnterpriseNetwork_getSiteList_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_cm_IpEnterpriseNetwork_getSiteList_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_cm_IpEnterpriseNetwork_getSiteList_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cm_IpEnterpriseNetwork_getSiteList_return);

            for (i_org_csapi_cm_IpEnterpriseNetwork_getSiteList_return=0; i_org_csapi_cm_IpEnterpriseNetwork_getSiteList_return < u_octet4_loop_org_csapi_cm_IpEnterpriseNetwork_getSiteList_return; i_org_csapi_cm_IpEnterpriseNetwork_getSiteList_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_IpEnterpriseNetwork_getSiteList_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpEnterpriseNetwork/getVPrN:1.0
 */
static void
decode_org_csapi_cm_IpEnterpriseNetwork_getVPrN(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpEnterpriseNetwork/getSite:1.0
 */
static void
decode_org_csapi_cm_IpEnterpriseNetwork_getSite(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_IpEnterpriseNetwork_getSite_siteID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpEnterpriseNetworkSite/getSAPList:1.0
 */
static void
decode_org_csapi_cm_IpEnterpriseNetworkSite_getSAPList(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cm_IpEnterpriseNetworkSite_getSAPList_return;
    guint32   i_org_csapi_cm_IpEnterpriseNetworkSite_getSAPList_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_cm_IpEnterpriseNetworkSite_getSAPList_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_cm_IpEnterpriseNetworkSite_getSAPList_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cm_IpEnterpriseNetworkSite_getSAPList_return);

            for (i_org_csapi_cm_IpEnterpriseNetworkSite_getSAPList_return=0; i_org_csapi_cm_IpEnterpriseNetworkSite_getSAPList_return < u_octet4_loop_org_csapi_cm_IpEnterpriseNetworkSite_getSAPList_return; i_org_csapi_cm_IpEnterpriseNetworkSite_getSAPList_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_IpEnterpriseNetworkSite_getSAPList_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpEnterpriseNetworkSite/getSiteID:1.0
 */
static void
decode_org_csapi_cm_IpEnterpriseNetworkSite_getSiteID(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_IpEnterpriseNetworkSite_getSiteID_return);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpEnterpriseNetworkSite/getSiteLocation:1.0
 */
static void
decode_org_csapi_cm_IpEnterpriseNetworkSite_getSiteLocation(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_IpEnterpriseNetworkSite_getSiteLocation_return);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpEnterpriseNetworkSite/getSiteDescription:1.0
 */
static void
decode_org_csapi_cm_IpEnterpriseNetworkSite_getSiteDescription(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_IpEnterpriseNetworkSite_getSiteDescription_return);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpEnterpriseNetworkSite/getIPSubnet:1.0
 */
static void
decode_org_csapi_cm_IpEnterpriseNetworkSite_getIPSubnet(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cm_TpIPSubnet"  */
            decode_org_csapi_cm_TpIPSubnet_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cm_TpIPSubnet"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cm/IpEnterpriseNetworkSite/getSAPIPSubnet:1.0
 */
static void
decode_org_csapi_cm_IpEnterpriseNetworkSite_getSAPIPSubnet(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_IpEnterpriseNetworkSite_getSAPIPSubnet_sapID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cm_TpIPSubnet"  */
            decode_org_csapi_cm_TpIPSubnet_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cm_TpIPSubnet"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAppAccountManager/reportNotification:1.0
 */
static void
decode_org_csapi_am_IpAppAccountManager_reportNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_am_TpChargingEventInfo"  */
        decode_org_csapi_am_TpChargingEventInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_am_TpChargingEventInfo"  */
        proto_tree_add_int(tree, hf_org_csapi_am_IpAppAccountManager_reportNotification_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAppAccountManager/queryBalanceRes:1.0
 */
static void
decode_org_csapi_am_IpAppAccountManager_queryBalanceRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_am_IpAppAccountManager_queryBalanceRes_balances;
    guint32   i_org_csapi_am_IpAppAccountManager_queryBalanceRes_balances;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_am_IpAppAccountManager_queryBalanceRes_queryId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_am_IpAppAccountManager_queryBalanceRes_balances = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_am_IpAppAccountManager_queryBalanceRes_balances_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_am_IpAppAccountManager_queryBalanceRes_balances);

        for (i_org_csapi_am_IpAppAccountManager_queryBalanceRes_balances=0; i_org_csapi_am_IpAppAccountManager_queryBalanceRes_balances < u_octet4_loop_org_csapi_am_IpAppAccountManager_queryBalanceRes_balances; i_org_csapi_am_IpAppAccountManager_queryBalanceRes_balances++) {

            /*  Begin struct "org_csapi_am_TpBalance"  */
            decode_org_csapi_am_TpBalance_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_am_TpBalance"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAppAccountManager/queryBalanceErr:1.0
 */
static void
decode_org_csapi_am_IpAppAccountManager_queryBalanceErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_am_IpAppAccountManager_queryBalanceErr_queryId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_am_IpAppAccountManager_queryBalanceErr_cause, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAppAccountManager/retrieveTransactionHistoryRes:1.0
 */
static void
decode_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryRes_transactionHistory;
    guint32   i_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryRes_transactionHistory;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryRes_retrievalID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryRes_transactionHistory = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryRes_transactionHistory_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryRes_transactionHistory);

        for (i_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryRes_transactionHistory=0; i_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryRes_transactionHistory < u_octet4_loop_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryRes_transactionHistory; i_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryRes_transactionHistory++) {

            /*  Begin struct "org_csapi_am_TpTransactionHistory"  */
            decode_org_csapi_am_TpTransactionHistory_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_am_TpTransactionHistory"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAppAccountManager/retrieveTransactionHistoryErr:1.0
 */
static void
decode_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryErr_retrievalID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryErr_transactionHistoryError, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAppAccountManager/queryBalanceExpiryDateRes:1.0
 */
static void
decode_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateRes_dates;
    guint32   i_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateRes_dates;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateRes_queryId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateRes_dates = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateRes_dates_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateRes_dates);

        for (i_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateRes_dates=0; i_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateRes_dates < u_octet4_loop_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateRes_dates; i_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateRes_dates++) {

            /*  Begin struct "org_csapi_am_TpBalanceExpiryDate"  */
            decode_org_csapi_am_TpBalanceExpiryDate_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_am_TpBalanceExpiryDate"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAppAccountManager/queryBalanceExpiryDateErr:1.0
 */
static void
decode_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateErr_queryId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateErr_cause, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAppAccountManager/updateBalanceRes:1.0
 */
static void
decode_org_csapi_am_IpAppAccountManager_updateBalanceRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_am_IpAppAccountManager_updateBalanceRes_requestId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_am_TpBalance"  */
        decode_org_csapi_am_TpBalance_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_am_TpBalance"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAppAccountManager/updateBalanceErr:1.0
 */
static void
decode_org_csapi_am_IpAppAccountManager_updateBalanceErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_am_IpAppAccountManager_updateBalanceErr_requestId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_am_IpAppAccountManager_updateBalanceErr_cause, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAppAccountManager/createVoucherRes:1.0
 */
static void
decode_org_csapi_am_IpAppAccountManager_createVoucherRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_am_IpAppAccountManager_createVoucherRes_requestId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_am_IpAppAccountManager_createVoucherRes_voucherId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAppAccountManager/createVoucherErr:1.0
 */
static void
decode_org_csapi_am_IpAppAccountManager_createVoucherErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_am_IpAppAccountManager_createVoucherErr_requestId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_am_IpAppAccountManager_createVoucherErr_cause, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAppAccountManager/destroyVoucherRes:1.0
 */
static void
decode_org_csapi_am_IpAppAccountManager_destroyVoucherRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_am_IpAppAccountManager_destroyVoucherRes_requestId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_am_IpAppAccountManager_destroyVoucherRes_voucherId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAppAccountManager/destroyVoucherErr:1.0
 */
static void
decode_org_csapi_am_IpAppAccountManager_destroyVoucherErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_am_IpAppAccountManager_destroyVoucherErr_requestId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_am_IpAppAccountManager_destroyVoucherErr_voucherId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_am_IpAppAccountManager_destroyVoucherErr_cause, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAppAccountManager/queryVoucherRes:1.0
 */
static void
decode_org_csapi_am_IpAppAccountManager_queryVoucherRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_am_IpAppAccountManager_queryVoucherRes_queryId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_am_TpVoucher"  */
        decode_org_csapi_am_TpVoucher_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_am_TpVoucher"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAppAccountManager/queryVoucherErr:1.0
 */
static void
decode_org_csapi_am_IpAppAccountManager_queryVoucherErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_am_IpAppAccountManager_queryVoucherErr_queryId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_am_IpAppAccountManager_queryVoucherErr_voucherId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_am_IpAppAccountManager_queryVoucherErr_cause, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAppAccountManager/queryUserVouchersRes:1.0
 */
static void
decode_org_csapi_am_IpAppAccountManager_queryUserVouchersRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_am_IpAppAccountManager_queryUserVouchersRes_vouchers;
    guint32   i_org_csapi_am_IpAppAccountManager_queryUserVouchersRes_vouchers;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_am_IpAppAccountManager_queryUserVouchersRes_queryId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_am_IpAppAccountManager_queryUserVouchersRes_vouchers = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_am_IpAppAccountManager_queryUserVouchersRes_vouchers_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_am_IpAppAccountManager_queryUserVouchersRes_vouchers);

        for (i_org_csapi_am_IpAppAccountManager_queryUserVouchersRes_vouchers=0; i_org_csapi_am_IpAppAccountManager_queryUserVouchersRes_vouchers < u_octet4_loop_org_csapi_am_IpAppAccountManager_queryUserVouchersRes_vouchers; i_org_csapi_am_IpAppAccountManager_queryUserVouchersRes_vouchers++) {

            /*  Begin struct "org_csapi_am_TpVoucher"  */
            decode_org_csapi_am_TpVoucher_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_am_TpVoucher"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAppAccountManager/queryUserVouchersErr:1.0
 */
static void
decode_org_csapi_am_IpAppAccountManager_queryUserVouchersErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_am_IpAppAccountManager_queryUserVouchersErr_queryId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_am_IpAppAccountManager_queryUserVouchersErr_cause, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAccountManager/createNotification:1.0
 */
static void
decode_org_csapi_am_IpAccountManager_createNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        /*  Begin struct "org_csapi_am_TpChargingEventCriteria"  */
        decode_org_csapi_am_TpChargingEventCriteria_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_am_TpChargingEventCriteria"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_am_IpAccountManager_createNotification_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAccountManager/destroyNotification:1.0
 */
static void
decode_org_csapi_am_IpAccountManager_destroyNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_am_IpAccountManager_destroyNotification_assignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAccountManager/queryBalanceReq:1.0
 */
static void
decode_org_csapi_am_IpAccountManager_queryBalanceReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_am_IpAccountManager_queryBalanceReq_users;
    guint32   i_org_csapi_am_IpAccountManager_queryBalanceReq_users;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_am_IpAccountManager_queryBalanceReq_users = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_am_IpAccountManager_queryBalanceReq_users_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_am_IpAccountManager_queryBalanceReq_users);

        for (i_org_csapi_am_IpAccountManager_queryBalanceReq_users=0; i_org_csapi_am_IpAccountManager_queryBalanceReq_users < u_octet4_loop_org_csapi_am_IpAccountManager_queryBalanceReq_users; i_org_csapi_am_IpAccountManager_queryBalanceReq_users++) {

            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_am_IpAccountManager_queryBalanceReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAccountManager/changeNotification:1.0
 */
static void
decode_org_csapi_am_IpAccountManager_changeNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_am_IpAccountManager_changeNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_am_TpChargingEventCriteria"  */
        decode_org_csapi_am_TpChargingEventCriteria_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_am_TpChargingEventCriteria"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAccountManager/getNotification:1.0
 */
static void
decode_org_csapi_am_IpAccountManager_getNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_am_IpAccountManager_getNotification_return;
    guint32   i_org_csapi_am_IpAccountManager_getNotification_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_am_IpAccountManager_getNotification_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_am_IpAccountManager_getNotification_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_am_IpAccountManager_getNotification_return);

            for (i_org_csapi_am_IpAccountManager_getNotification_return=0; i_org_csapi_am_IpAccountManager_getNotification_return < u_octet4_loop_org_csapi_am_IpAccountManager_getNotification_return; i_org_csapi_am_IpAccountManager_getNotification_return++) {

                /*  Begin struct "org_csapi_am_TpChargingEventCriteriaResult"  */
                decode_org_csapi_am_TpChargingEventCriteriaResult_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_am_TpChargingEventCriteriaResult"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAccountManager/retrieveTransactionHistoryReq:1.0
 */
static void
decode_org_csapi_am_IpAccountManager_retrieveTransactionHistoryReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        /*  Begin struct "org_csapi_TpTimeInterval"  */
        decode_org_csapi_TpTimeInterval_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpTimeInterval"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_am_IpAccountManager_retrieveTransactionHistoryReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAccountManager/enableNotifications:1.0
 */
static void
decode_org_csapi_am_IpAccountManager_enableNotifications(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_am_IpAccountManager_enableNotifications_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAccountManager/disableNotifications:1.0
 */
static void
decode_org_csapi_am_IpAccountManager_disableNotifications(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAccountManager/queryBalanceExpiryDateReq:1.0
 */
static void
decode_org_csapi_am_IpAccountManager_queryBalanceExpiryDateReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_am_IpAccountManager_queryBalanceExpiryDateReq_users;
    guint32   i_org_csapi_am_IpAccountManager_queryBalanceExpiryDateReq_users;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_am_IpAccountManager_queryBalanceExpiryDateReq_users = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_am_IpAccountManager_queryBalanceExpiryDateReq_users_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_am_IpAccountManager_queryBalanceExpiryDateReq_users);

        for (i_org_csapi_am_IpAccountManager_queryBalanceExpiryDateReq_users=0; i_org_csapi_am_IpAccountManager_queryBalanceExpiryDateReq_users < u_octet4_loop_org_csapi_am_IpAccountManager_queryBalanceExpiryDateReq_users; i_org_csapi_am_IpAccountManager_queryBalanceExpiryDateReq_users++) {

            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_am_IpAccountManager_queryBalanceExpiryDateReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAccountManager/updateBalanceReq:1.0
 */
static void
decode_org_csapi_am_IpAccountManager_updateBalanceReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        proto_tree_add_boolean(tree, hf_org_csapi_am_IpAccountManager_updateBalanceReq_debit, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

        /*  Begin struct "org_csapi_am_TpBalanceInfo"  */
        decode_org_csapi_am_TpBalanceInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_am_TpBalanceInfo"  */
        proto_tree_add_int(tree, hf_org_csapi_am_IpAccountManager_updateBalanceReq_period, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_am_IpAccountManager_updateBalanceReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAccountManager/createVoucherReq:1.0
 */
static void
decode_org_csapi_am_IpAccountManager_createVoucherReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        /*  Begin struct "org_csapi_am_TpBalanceInfo"  */
        decode_org_csapi_am_TpBalanceInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_am_TpBalanceInfo"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_am_IpAccountManager_createVoucherReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAccountManager/destroyVoucherReq:1.0
 */
static void
decode_org_csapi_am_IpAccountManager_destroyVoucherReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_am_IpAccountManager_destroyVoucherReq_voucherId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_am_IpAccountManager_destroyVoucherReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAccountManager/queryVoucherReq:1.0
 */
static void
decode_org_csapi_am_IpAccountManager_queryVoucherReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_am_IpAccountManager_queryVoucherReq_voucherId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_am_IpAccountManager_queryVoucherReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/am/IpAccountManager/queryUserVouchersReq:1.0
 */
static void
decode_org_csapi_am_IpAccountManager_queryUserVouchersReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_am_IpAccountManager_queryUserVouchersReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/creditAmountErr:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_creditAmountErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_creditAmountErr_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_creditAmountErr_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpAppChargingSession_creditAmountErr_error, tvb, *offset-4, 4, u_octet4);

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_creditAmountErr_requestNumberNextRequest, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/creditAmountRes:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_creditAmountRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_creditAmountRes_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_creditAmountRes_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cs_TpChargingPrice"  */
        decode_org_csapi_cs_TpChargingPrice_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpChargingPrice"  */
        /*  Begin struct "org_csapi_cs_TpChargingPrice"  */
        decode_org_csapi_cs_TpChargingPrice_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpChargingPrice"  */
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_creditAmountRes_requestNumberNextRequest, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/creditUnitErr:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_creditUnitErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_creditUnitErr_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_creditUnitErr_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpAppChargingSession_creditUnitErr_error, tvb, *offset-4, 4, u_octet4);

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_creditUnitErr_requestNumberNextRequest, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/creditUnitRes:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_creditUnitRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cs_IpAppChargingSession_creditUnitRes_creditedVolumes;
    guint32   i_org_csapi_cs_IpAppChargingSession_creditUnitRes_creditedVolumes;
    guint32   u_octet4_loop_org_csapi_cs_IpAppChargingSession_creditUnitRes_reservedUnitsLeft;
    guint32   i_org_csapi_cs_IpAppChargingSession_creditUnitRes_reservedUnitsLeft;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_creditUnitRes_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_creditUnitRes_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_cs_IpAppChargingSession_creditUnitRes_creditedVolumes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpAppChargingSession_creditUnitRes_creditedVolumes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_IpAppChargingSession_creditUnitRes_creditedVolumes);

        for (i_org_csapi_cs_IpAppChargingSession_creditUnitRes_creditedVolumes=0; i_org_csapi_cs_IpAppChargingSession_creditUnitRes_creditedVolumes < u_octet4_loop_org_csapi_cs_IpAppChargingSession_creditUnitRes_creditedVolumes; i_org_csapi_cs_IpAppChargingSession_creditUnitRes_creditedVolumes++) {

            /*  Begin struct "org_csapi_cs_TpVolume"  */
            decode_org_csapi_cs_TpVolume_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cs_TpVolume"  */
        }

        u_octet4_loop_org_csapi_cs_IpAppChargingSession_creditUnitRes_reservedUnitsLeft = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpAppChargingSession_creditUnitRes_reservedUnitsLeft_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_IpAppChargingSession_creditUnitRes_reservedUnitsLeft);

        for (i_org_csapi_cs_IpAppChargingSession_creditUnitRes_reservedUnitsLeft=0; i_org_csapi_cs_IpAppChargingSession_creditUnitRes_reservedUnitsLeft < u_octet4_loop_org_csapi_cs_IpAppChargingSession_creditUnitRes_reservedUnitsLeft; i_org_csapi_cs_IpAppChargingSession_creditUnitRes_reservedUnitsLeft++) {

            /*  Begin struct "org_csapi_cs_TpVolume"  */
            decode_org_csapi_cs_TpVolume_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cs_TpVolume"  */
        }

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_creditUnitRes_requestNumberNextRequest, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/debitAmountErr:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_debitAmountErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_debitAmountErr_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_debitAmountErr_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpAppChargingSession_debitAmountErr_error, tvb, *offset-4, 4, u_octet4);

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_debitAmountErr_requestNumberNextRequest, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/debitAmountRes:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_debitAmountRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_debitAmountRes_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_debitAmountRes_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cs_TpChargingPrice"  */
        decode_org_csapi_cs_TpChargingPrice_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpChargingPrice"  */
        /*  Begin struct "org_csapi_cs_TpChargingPrice"  */
        decode_org_csapi_cs_TpChargingPrice_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpChargingPrice"  */
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_debitAmountRes_requestNumberNextRequest, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/debitUnitErr:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_debitUnitErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_debitUnitErr_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_debitUnitErr_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpAppChargingSession_debitUnitErr_error, tvb, *offset-4, 4, u_octet4);

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_debitUnitErr_requestNumberNextRequest, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/debitUnitRes:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_debitUnitRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cs_IpAppChargingSession_debitUnitRes_debitedVolumes;
    guint32   i_org_csapi_cs_IpAppChargingSession_debitUnitRes_debitedVolumes;
    guint32   u_octet4_loop_org_csapi_cs_IpAppChargingSession_debitUnitRes_reservedUnitsLeft;
    guint32   i_org_csapi_cs_IpAppChargingSession_debitUnitRes_reservedUnitsLeft;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_debitUnitRes_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_debitUnitRes_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_cs_IpAppChargingSession_debitUnitRes_debitedVolumes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpAppChargingSession_debitUnitRes_debitedVolumes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_IpAppChargingSession_debitUnitRes_debitedVolumes);

        for (i_org_csapi_cs_IpAppChargingSession_debitUnitRes_debitedVolumes=0; i_org_csapi_cs_IpAppChargingSession_debitUnitRes_debitedVolumes < u_octet4_loop_org_csapi_cs_IpAppChargingSession_debitUnitRes_debitedVolumes; i_org_csapi_cs_IpAppChargingSession_debitUnitRes_debitedVolumes++) {

            /*  Begin struct "org_csapi_cs_TpVolume"  */
            decode_org_csapi_cs_TpVolume_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cs_TpVolume"  */
        }

        u_octet4_loop_org_csapi_cs_IpAppChargingSession_debitUnitRes_reservedUnitsLeft = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpAppChargingSession_debitUnitRes_reservedUnitsLeft_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_IpAppChargingSession_debitUnitRes_reservedUnitsLeft);

        for (i_org_csapi_cs_IpAppChargingSession_debitUnitRes_reservedUnitsLeft=0; i_org_csapi_cs_IpAppChargingSession_debitUnitRes_reservedUnitsLeft < u_octet4_loop_org_csapi_cs_IpAppChargingSession_debitUnitRes_reservedUnitsLeft; i_org_csapi_cs_IpAppChargingSession_debitUnitRes_reservedUnitsLeft++) {

            /*  Begin struct "org_csapi_cs_TpVolume"  */
            decode_org_csapi_cs_TpVolume_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cs_TpVolume"  */
        }

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_debitUnitRes_requestNumberNextRequest, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/directCreditAmountErr:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_directCreditAmountErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directCreditAmountErr_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directCreditAmountErr_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpAppChargingSession_directCreditAmountErr_error, tvb, *offset-4, 4, u_octet4);

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directCreditAmountErr_requestNumberNextRequest, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/directCreditAmountRes:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_directCreditAmountRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directCreditAmountRes_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directCreditAmountRes_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cs_TpChargingPrice"  */
        decode_org_csapi_cs_TpChargingPrice_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpChargingPrice"  */
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directCreditAmountRes_requestNumberNextRequest, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/directCreditUnitErr:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_directCreditUnitErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directCreditUnitErr_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directCreditUnitErr_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpAppChargingSession_directCreditUnitErr_error, tvb, *offset-4, 4, u_octet4);

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directCreditUnitErr_requestNumberNextRequest, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/directCreditUnitRes:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_directCreditUnitRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cs_IpAppChargingSession_directCreditUnitRes_creditedVolumes;
    guint32   i_org_csapi_cs_IpAppChargingSession_directCreditUnitRes_creditedVolumes;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directCreditUnitRes_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directCreditUnitRes_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_cs_IpAppChargingSession_directCreditUnitRes_creditedVolumes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpAppChargingSession_directCreditUnitRes_creditedVolumes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_IpAppChargingSession_directCreditUnitRes_creditedVolumes);

        for (i_org_csapi_cs_IpAppChargingSession_directCreditUnitRes_creditedVolumes=0; i_org_csapi_cs_IpAppChargingSession_directCreditUnitRes_creditedVolumes < u_octet4_loop_org_csapi_cs_IpAppChargingSession_directCreditUnitRes_creditedVolumes; i_org_csapi_cs_IpAppChargingSession_directCreditUnitRes_creditedVolumes++) {

            /*  Begin struct "org_csapi_cs_TpVolume"  */
            decode_org_csapi_cs_TpVolume_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cs_TpVolume"  */
        }

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directCreditUnitRes_requestNumberNextRequest, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/directDebitAmountErr:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_directDebitAmountErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directDebitAmountErr_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directDebitAmountErr_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpAppChargingSession_directDebitAmountErr_error, tvb, *offset-4, 4, u_octet4);

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directDebitAmountErr_requestNumberNextRequest, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/directDebitAmountRes:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_directDebitAmountRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directDebitAmountRes_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directDebitAmountRes_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cs_TpChargingPrice"  */
        decode_org_csapi_cs_TpChargingPrice_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpChargingPrice"  */
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directDebitAmountRes_requestNumberNextRequest, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/directDebitUnitErr:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_directDebitUnitErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directDebitUnitErr_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directDebitUnitErr_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpAppChargingSession_directDebitUnitErr_error, tvb, *offset-4, 4, u_octet4);

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directDebitUnitErr_requestNumberNextRequest, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/directDebitUnitRes:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_directDebitUnitRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cs_IpAppChargingSession_directDebitUnitRes_debitedVolumes;
    guint32   i_org_csapi_cs_IpAppChargingSession_directDebitUnitRes_debitedVolumes;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directDebitUnitRes_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directDebitUnitRes_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_cs_IpAppChargingSession_directDebitUnitRes_debitedVolumes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpAppChargingSession_directDebitUnitRes_debitedVolumes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_IpAppChargingSession_directDebitUnitRes_debitedVolumes);

        for (i_org_csapi_cs_IpAppChargingSession_directDebitUnitRes_debitedVolumes=0; i_org_csapi_cs_IpAppChargingSession_directDebitUnitRes_debitedVolumes < u_octet4_loop_org_csapi_cs_IpAppChargingSession_directDebitUnitRes_debitedVolumes; i_org_csapi_cs_IpAppChargingSession_directDebitUnitRes_debitedVolumes++) {

            /*  Begin struct "org_csapi_cs_TpVolume"  */
            decode_org_csapi_cs_TpVolume_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cs_TpVolume"  */
        }

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_directDebitUnitRes_requestNumberNextRequest, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/extendLifeTimeErr:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_extendLifeTimeErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_extendLifeTimeErr_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpAppChargingSession_extendLifeTimeErr_error, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/extendLifeTimeRes:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_extendLifeTimeRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_extendLifeTimeRes_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_extendLifeTimeRes_sessionTimeLeft, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/rateErr:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_rateErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_rateErr_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpAppChargingSession_rateErr_error, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/rateRes:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_rateRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cs_IpAppChargingSession_rateRes_rates;
    guint32   i_org_csapi_cs_IpAppChargingSession_rateRes_rates;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_rateRes_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_cs_IpAppChargingSession_rateRes_rates = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpAppChargingSession_rateRes_rates_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_IpAppChargingSession_rateRes_rates);

        for (i_org_csapi_cs_IpAppChargingSession_rateRes_rates=0; i_org_csapi_cs_IpAppChargingSession_rateRes_rates < u_octet4_loop_org_csapi_cs_IpAppChargingSession_rateRes_rates; i_org_csapi_cs_IpAppChargingSession_rateRes_rates++) {

            /*  Begin struct "org_csapi_cs_TpPriceVolume"  */
            decode_org_csapi_cs_TpPriceVolume_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cs_TpPriceVolume"  */
        }

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_rateRes_validityTimeLeft, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/reserveAmountErr:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_reserveAmountErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_reserveAmountErr_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_reserveAmountErr_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpAppChargingSession_reserveAmountErr_error, tvb, *offset-4, 4, u_octet4);

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_reserveAmountErr_requestNumberNextRequest, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/reserveAmountRes:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_reserveAmountRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_reserveAmountRes_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_reserveAmountRes_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cs_TpChargingPrice"  */
        decode_org_csapi_cs_TpChargingPrice_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpChargingPrice"  */
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_reserveAmountRes_sessionTimeLeft, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_reserveAmountRes_requestNumberNextRequest, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/reserveUnitErr:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_reserveUnitErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_reserveUnitErr_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_reserveUnitErr_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpAppChargingSession_reserveUnitErr_error, tvb, *offset-4, 4, u_octet4);

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_reserveUnitErr_requestNumberNextRequest, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/reserveUnitRes:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_reserveUnitRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cs_IpAppChargingSession_reserveUnitRes_reservedUnits;
    guint32   i_org_csapi_cs_IpAppChargingSession_reserveUnitRes_reservedUnits;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_reserveUnitRes_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_reserveUnitRes_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_cs_IpAppChargingSession_reserveUnitRes_reservedUnits = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpAppChargingSession_reserveUnitRes_reservedUnits_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_IpAppChargingSession_reserveUnitRes_reservedUnits);

        for (i_org_csapi_cs_IpAppChargingSession_reserveUnitRes_reservedUnits=0; i_org_csapi_cs_IpAppChargingSession_reserveUnitRes_reservedUnits < u_octet4_loop_org_csapi_cs_IpAppChargingSession_reserveUnitRes_reservedUnits; i_org_csapi_cs_IpAppChargingSession_reserveUnitRes_reservedUnits++) {

            /*  Begin struct "org_csapi_cs_TpVolume"  */
            decode_org_csapi_cs_TpVolume_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cs_TpVolume"  */
        }

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_reserveUnitRes_sessionTimeLeft, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_reserveUnitRes_requestNumberNextRequest, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingSession/sessionEnded:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingSession_sessionEnded(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingSession_sessionEnded_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpAppChargingSession_sessionEnded_report, tvb, *offset-4, 4, u_octet4);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpChargingSession/creditAmountReq:1.0
 */
static void
decode_org_csapi_cs_IpChargingSession_creditAmountReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_creditAmountReq_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cs_TpApplicationDescription"  */
        decode_org_csapi_cs_TpApplicationDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpApplicationDescription"  */
        /*  Begin struct "org_csapi_cs_TpChargingPrice"  */
        decode_org_csapi_cs_TpChargingPrice_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpChargingPrice"  */
        proto_tree_add_boolean(tree, hf_org_csapi_cs_IpChargingSession_creditAmountReq_closeReservation, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_creditAmountReq_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpChargingSession/creditUnitReq:1.0
 */
static void
decode_org_csapi_cs_IpChargingSession_creditUnitReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cs_IpChargingSession_creditUnitReq_volumes;
    guint32   i_org_csapi_cs_IpChargingSession_creditUnitReq_volumes;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_creditUnitReq_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cs_TpApplicationDescription"  */
        decode_org_csapi_cs_TpApplicationDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpApplicationDescription"  */
        u_octet4_loop_org_csapi_cs_IpChargingSession_creditUnitReq_volumes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpChargingSession_creditUnitReq_volumes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_IpChargingSession_creditUnitReq_volumes);

        for (i_org_csapi_cs_IpChargingSession_creditUnitReq_volumes=0; i_org_csapi_cs_IpChargingSession_creditUnitReq_volumes < u_octet4_loop_org_csapi_cs_IpChargingSession_creditUnitReq_volumes; i_org_csapi_cs_IpChargingSession_creditUnitReq_volumes++) {

            /*  Begin struct "org_csapi_cs_TpVolume"  */
            decode_org_csapi_cs_TpVolume_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cs_TpVolume"  */
        }

        proto_tree_add_boolean(tree, hf_org_csapi_cs_IpChargingSession_creditUnitReq_closeReservation, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_creditUnitReq_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpChargingSession/debitAmountReq:1.0
 */
static void
decode_org_csapi_cs_IpChargingSession_debitAmountReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_debitAmountReq_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cs_TpApplicationDescription"  */
        decode_org_csapi_cs_TpApplicationDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpApplicationDescription"  */
        /*  Begin struct "org_csapi_cs_TpChargingPrice"  */
        decode_org_csapi_cs_TpChargingPrice_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpChargingPrice"  */
        proto_tree_add_boolean(tree, hf_org_csapi_cs_IpChargingSession_debitAmountReq_closeReservation, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_debitAmountReq_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpChargingSession/debitUnitReq:1.0
 */
static void
decode_org_csapi_cs_IpChargingSession_debitUnitReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cs_IpChargingSession_debitUnitReq_volumes;
    guint32   i_org_csapi_cs_IpChargingSession_debitUnitReq_volumes;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_debitUnitReq_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cs_TpApplicationDescription"  */
        decode_org_csapi_cs_TpApplicationDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpApplicationDescription"  */
        u_octet4_loop_org_csapi_cs_IpChargingSession_debitUnitReq_volumes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpChargingSession_debitUnitReq_volumes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_IpChargingSession_debitUnitReq_volumes);

        for (i_org_csapi_cs_IpChargingSession_debitUnitReq_volumes=0; i_org_csapi_cs_IpChargingSession_debitUnitReq_volumes < u_octet4_loop_org_csapi_cs_IpChargingSession_debitUnitReq_volumes; i_org_csapi_cs_IpChargingSession_debitUnitReq_volumes++) {

            /*  Begin struct "org_csapi_cs_TpVolume"  */
            decode_org_csapi_cs_TpVolume_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cs_TpVolume"  */
        }

        proto_tree_add_boolean(tree, hf_org_csapi_cs_IpChargingSession_debitUnitReq_closeReservation, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_debitUnitReq_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpChargingSession/directCreditAmountReq:1.0
 */
static void
decode_org_csapi_cs_IpChargingSession_directCreditAmountReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cs_IpChargingSession_directCreditAmountReq_chargingParameters;
    guint32   i_org_csapi_cs_IpChargingSession_directCreditAmountReq_chargingParameters;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_directCreditAmountReq_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cs_TpApplicationDescription"  */
        decode_org_csapi_cs_TpApplicationDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpApplicationDescription"  */
        u_octet4_loop_org_csapi_cs_IpChargingSession_directCreditAmountReq_chargingParameters = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpChargingSession_directCreditAmountReq_chargingParameters_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_IpChargingSession_directCreditAmountReq_chargingParameters);

        for (i_org_csapi_cs_IpChargingSession_directCreditAmountReq_chargingParameters=0; i_org_csapi_cs_IpChargingSession_directCreditAmountReq_chargingParameters < u_octet4_loop_org_csapi_cs_IpChargingSession_directCreditAmountReq_chargingParameters; i_org_csapi_cs_IpChargingSession_directCreditAmountReq_chargingParameters++) {

            /*  Begin struct "org_csapi_cs_TpChargingParameter"  */
            decode_org_csapi_cs_TpChargingParameter_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cs_TpChargingParameter"  */
        }

        /*  Begin struct "org_csapi_cs_TpChargingPrice"  */
        decode_org_csapi_cs_TpChargingPrice_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpChargingPrice"  */
        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_directCreditAmountReq_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpChargingSession/directCreditUnitReq:1.0
 */
static void
decode_org_csapi_cs_IpChargingSession_directCreditUnitReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cs_IpChargingSession_directCreditUnitReq_chargingParameters;
    guint32   i_org_csapi_cs_IpChargingSession_directCreditUnitReq_chargingParameters;
    guint32   u_octet4_loop_org_csapi_cs_IpChargingSession_directCreditUnitReq_volumes;
    guint32   i_org_csapi_cs_IpChargingSession_directCreditUnitReq_volumes;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_directCreditUnitReq_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cs_TpApplicationDescription"  */
        decode_org_csapi_cs_TpApplicationDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpApplicationDescription"  */
        u_octet4_loop_org_csapi_cs_IpChargingSession_directCreditUnitReq_chargingParameters = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpChargingSession_directCreditUnitReq_chargingParameters_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_IpChargingSession_directCreditUnitReq_chargingParameters);

        for (i_org_csapi_cs_IpChargingSession_directCreditUnitReq_chargingParameters=0; i_org_csapi_cs_IpChargingSession_directCreditUnitReq_chargingParameters < u_octet4_loop_org_csapi_cs_IpChargingSession_directCreditUnitReq_chargingParameters; i_org_csapi_cs_IpChargingSession_directCreditUnitReq_chargingParameters++) {

            /*  Begin struct "org_csapi_cs_TpChargingParameter"  */
            decode_org_csapi_cs_TpChargingParameter_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cs_TpChargingParameter"  */
        }

        u_octet4_loop_org_csapi_cs_IpChargingSession_directCreditUnitReq_volumes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpChargingSession_directCreditUnitReq_volumes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_IpChargingSession_directCreditUnitReq_volumes);

        for (i_org_csapi_cs_IpChargingSession_directCreditUnitReq_volumes=0; i_org_csapi_cs_IpChargingSession_directCreditUnitReq_volumes < u_octet4_loop_org_csapi_cs_IpChargingSession_directCreditUnitReq_volumes; i_org_csapi_cs_IpChargingSession_directCreditUnitReq_volumes++) {

            /*  Begin struct "org_csapi_cs_TpVolume"  */
            decode_org_csapi_cs_TpVolume_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cs_TpVolume"  */
        }

        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_directCreditUnitReq_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpChargingSession/directDebitAmountReq:1.0
 */
static void
decode_org_csapi_cs_IpChargingSession_directDebitAmountReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cs_IpChargingSession_directDebitAmountReq_chargingParameters;
    guint32   i_org_csapi_cs_IpChargingSession_directDebitAmountReq_chargingParameters;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_directDebitAmountReq_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cs_TpApplicationDescription"  */
        decode_org_csapi_cs_TpApplicationDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpApplicationDescription"  */
        u_octet4_loop_org_csapi_cs_IpChargingSession_directDebitAmountReq_chargingParameters = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpChargingSession_directDebitAmountReq_chargingParameters_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_IpChargingSession_directDebitAmountReq_chargingParameters);

        for (i_org_csapi_cs_IpChargingSession_directDebitAmountReq_chargingParameters=0; i_org_csapi_cs_IpChargingSession_directDebitAmountReq_chargingParameters < u_octet4_loop_org_csapi_cs_IpChargingSession_directDebitAmountReq_chargingParameters; i_org_csapi_cs_IpChargingSession_directDebitAmountReq_chargingParameters++) {

            /*  Begin struct "org_csapi_cs_TpChargingParameter"  */
            decode_org_csapi_cs_TpChargingParameter_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cs_TpChargingParameter"  */
        }

        /*  Begin struct "org_csapi_cs_TpChargingPrice"  */
        decode_org_csapi_cs_TpChargingPrice_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpChargingPrice"  */
        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_directDebitAmountReq_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpChargingSession/directDebitUnitReq:1.0
 */
static void
decode_org_csapi_cs_IpChargingSession_directDebitUnitReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cs_IpChargingSession_directDebitUnitReq_chargingParameters;
    guint32   i_org_csapi_cs_IpChargingSession_directDebitUnitReq_chargingParameters;
    guint32   u_octet4_loop_org_csapi_cs_IpChargingSession_directDebitUnitReq_volumes;
    guint32   i_org_csapi_cs_IpChargingSession_directDebitUnitReq_volumes;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_directDebitUnitReq_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cs_TpApplicationDescription"  */
        decode_org_csapi_cs_TpApplicationDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpApplicationDescription"  */
        u_octet4_loop_org_csapi_cs_IpChargingSession_directDebitUnitReq_chargingParameters = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpChargingSession_directDebitUnitReq_chargingParameters_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_IpChargingSession_directDebitUnitReq_chargingParameters);

        for (i_org_csapi_cs_IpChargingSession_directDebitUnitReq_chargingParameters=0; i_org_csapi_cs_IpChargingSession_directDebitUnitReq_chargingParameters < u_octet4_loop_org_csapi_cs_IpChargingSession_directDebitUnitReq_chargingParameters; i_org_csapi_cs_IpChargingSession_directDebitUnitReq_chargingParameters++) {

            /*  Begin struct "org_csapi_cs_TpChargingParameter"  */
            decode_org_csapi_cs_TpChargingParameter_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cs_TpChargingParameter"  */
        }

        u_octet4_loop_org_csapi_cs_IpChargingSession_directDebitUnitReq_volumes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpChargingSession_directDebitUnitReq_volumes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_IpChargingSession_directDebitUnitReq_volumes);

        for (i_org_csapi_cs_IpChargingSession_directDebitUnitReq_volumes=0; i_org_csapi_cs_IpChargingSession_directDebitUnitReq_volumes < u_octet4_loop_org_csapi_cs_IpChargingSession_directDebitUnitReq_volumes; i_org_csapi_cs_IpChargingSession_directDebitUnitReq_volumes++) {

            /*  Begin struct "org_csapi_cs_TpVolume"  */
            decode_org_csapi_cs_TpVolume_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cs_TpVolume"  */
        }

        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_directDebitUnitReq_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpChargingSession/extendLifeTimeReq:1.0
 */
static void
decode_org_csapi_cs_IpChargingSession_extendLifeTimeReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_extendLifeTimeReq_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpChargingSession/getAmountLeft:1.0
 */
static void
decode_org_csapi_cs_IpChargingSession_getAmountLeft(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_getAmountLeft_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cs_TpChargingPrice"  */
            decode_org_csapi_cs_TpChargingPrice_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cs_TpChargingPrice"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpChargingSession/getLifeTimeLeft:1.0
 */
static void
decode_org_csapi_cs_IpChargingSession_getLifeTimeLeft(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_getLifeTimeLeft_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_getLifeTimeLeft_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpChargingSession/getUnitLeft:1.0
 */
static void
decode_org_csapi_cs_IpChargingSession_getUnitLeft(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cs_IpChargingSession_getUnitLeft_return;
    guint32   i_org_csapi_cs_IpChargingSession_getUnitLeft_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_getUnitLeft_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_cs_IpChargingSession_getUnitLeft_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_cs_IpChargingSession_getUnitLeft_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_IpChargingSession_getUnitLeft_return);

            for (i_org_csapi_cs_IpChargingSession_getUnitLeft_return=0; i_org_csapi_cs_IpChargingSession_getUnitLeft_return < u_octet4_loop_org_csapi_cs_IpChargingSession_getUnitLeft_return; i_org_csapi_cs_IpChargingSession_getUnitLeft_return++) {

                /*  Begin struct "org_csapi_cs_TpVolume"  */
                decode_org_csapi_cs_TpVolume_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_cs_TpVolume"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpChargingSession/rateReq:1.0
 */
static void
decode_org_csapi_cs_IpChargingSession_rateReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cs_IpChargingSession_rateReq_chargingParameters;
    guint32   i_org_csapi_cs_IpChargingSession_rateReq_chargingParameters;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_rateReq_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_cs_IpChargingSession_rateReq_chargingParameters = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpChargingSession_rateReq_chargingParameters_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_IpChargingSession_rateReq_chargingParameters);

        for (i_org_csapi_cs_IpChargingSession_rateReq_chargingParameters=0; i_org_csapi_cs_IpChargingSession_rateReq_chargingParameters < u_octet4_loop_org_csapi_cs_IpChargingSession_rateReq_chargingParameters; i_org_csapi_cs_IpChargingSession_rateReq_chargingParameters++) {

            /*  Begin struct "org_csapi_cs_TpChargingParameter"  */
            decode_org_csapi_cs_TpChargingParameter_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cs_TpChargingParameter"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpChargingSession/release:1.0
 */
static void
decode_org_csapi_cs_IpChargingSession_release(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_release_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_release_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpChargingSession/reserveAmountReq:1.0
 */
static void
decode_org_csapi_cs_IpChargingSession_reserveAmountReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cs_IpChargingSession_reserveAmountReq_chargingParameters;
    guint32   i_org_csapi_cs_IpChargingSession_reserveAmountReq_chargingParameters;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_reserveAmountReq_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cs_TpApplicationDescription"  */
        decode_org_csapi_cs_TpApplicationDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpApplicationDescription"  */
        u_octet4_loop_org_csapi_cs_IpChargingSession_reserveAmountReq_chargingParameters = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpChargingSession_reserveAmountReq_chargingParameters_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_IpChargingSession_reserveAmountReq_chargingParameters);

        for (i_org_csapi_cs_IpChargingSession_reserveAmountReq_chargingParameters=0; i_org_csapi_cs_IpChargingSession_reserveAmountReq_chargingParameters < u_octet4_loop_org_csapi_cs_IpChargingSession_reserveAmountReq_chargingParameters; i_org_csapi_cs_IpChargingSession_reserveAmountReq_chargingParameters++) {

            /*  Begin struct "org_csapi_cs_TpChargingParameter"  */
            decode_org_csapi_cs_TpChargingParameter_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cs_TpChargingParameter"  */
        }

        /*  Begin struct "org_csapi_cs_TpChargingPrice"  */
        decode_org_csapi_cs_TpChargingPrice_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpChargingPrice"  */
        /*  Begin struct "org_csapi_cs_TpChargingPrice"  */
        decode_org_csapi_cs_TpChargingPrice_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpChargingPrice"  */
        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_reserveAmountReq_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpChargingSession/reserveUnitReq:1.0
 */
static void
decode_org_csapi_cs_IpChargingSession_reserveUnitReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cs_IpChargingSession_reserveUnitReq_chargingParameters;
    guint32   i_org_csapi_cs_IpChargingSession_reserveUnitReq_chargingParameters;
    guint32   u_octet4_loop_org_csapi_cs_IpChargingSession_reserveUnitReq_volumes;
    guint32   i_org_csapi_cs_IpChargingSession_reserveUnitReq_volumes;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_reserveUnitReq_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_cs_TpApplicationDescription"  */
        decode_org_csapi_cs_TpApplicationDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpApplicationDescription"  */
        u_octet4_loop_org_csapi_cs_IpChargingSession_reserveUnitReq_chargingParameters = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpChargingSession_reserveUnitReq_chargingParameters_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_IpChargingSession_reserveUnitReq_chargingParameters);

        for (i_org_csapi_cs_IpChargingSession_reserveUnitReq_chargingParameters=0; i_org_csapi_cs_IpChargingSession_reserveUnitReq_chargingParameters < u_octet4_loop_org_csapi_cs_IpChargingSession_reserveUnitReq_chargingParameters; i_org_csapi_cs_IpChargingSession_reserveUnitReq_chargingParameters++) {

            /*  Begin struct "org_csapi_cs_TpChargingParameter"  */
            decode_org_csapi_cs_TpChargingParameter_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cs_TpChargingParameter"  */
        }

        u_octet4_loop_org_csapi_cs_IpChargingSession_reserveUnitReq_volumes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpChargingSession_reserveUnitReq_volumes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_IpChargingSession_reserveUnitReq_volumes);

        for (i_org_csapi_cs_IpChargingSession_reserveUnitReq_volumes=0; i_org_csapi_cs_IpChargingSession_reserveUnitReq_volumes < u_octet4_loop_org_csapi_cs_IpChargingSession_reserveUnitReq_volumes; i_org_csapi_cs_IpChargingSession_reserveUnitReq_volumes++) {

            /*  Begin struct "org_csapi_cs_TpVolume"  */
            decode_org_csapi_cs_TpVolume_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cs_TpVolume"  */
        }

        proto_tree_add_int(tree, hf_org_csapi_cs_IpChargingSession_reserveUnitReq_requestNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingManager/sessionAborted:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingManager_sessionAborted(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingManager_sessionAborted_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpAppChargingManager/abortMultipleChargingSessions:1.0
 */
static void
decode_org_csapi_cs_IpAppChargingManager_abortMultipleChargingSessions(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cs_IpAppChargingManager_abortMultipleChargingSessions_chargingSessionSet;
    guint32   i_org_csapi_cs_IpAppChargingManager_abortMultipleChargingSessions_chargingSessionSet;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_cs_IpAppChargingManager_abortMultipleChargingSessions_chargingSessionSet = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpAppChargingManager_abortMultipleChargingSessions_chargingSessionSet_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_IpAppChargingManager_abortMultipleChargingSessions_chargingSessionSet);

        for (i_org_csapi_cs_IpAppChargingManager_abortMultipleChargingSessions_chargingSessionSet=0; i_org_csapi_cs_IpAppChargingManager_abortMultipleChargingSessions_chargingSessionSet < u_octet4_loop_org_csapi_cs_IpAppChargingManager_abortMultipleChargingSessions_chargingSessionSet; i_org_csapi_cs_IpAppChargingManager_abortMultipleChargingSessions_chargingSessionSet++) {

            proto_tree_add_int(tree, hf_org_csapi_cs_IpAppChargingManager_abortMultipleChargingSessions_chargingSessionSet, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpChargingManager/createChargingSession:1.0
 */
static void
decode_org_csapi_cs_IpChargingManager_createChargingSession(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cs_IpChargingManager_createChargingSession_sessionDescription);

        /*  Begin struct "org_csapi_cs_TpMerchantAccountID"  */
        decode_org_csapi_cs_TpMerchantAccountID_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpMerchantAccountID"  */
        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        /*  Begin struct "org_csapi_cs_TpCorrelationID"  */
        decode_org_csapi_cs_TpCorrelationID_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpCorrelationID"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cs_TpChargingSessionID"  */
            decode_org_csapi_cs_TpChargingSessionID_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cs_TpChargingSessionID"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/cs/IpChargingManager/createSplitChargingSession:1.0
 */
static void
decode_org_csapi_cs_IpChargingManager_createSplitChargingSession(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cs_IpChargingManager_createSplitChargingSession_users;
    guint32   i_org_csapi_cs_IpChargingManager_createSplitChargingSession_users;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cs_IpChargingManager_createSplitChargingSession_sessionDescription);

        /*  Begin struct "org_csapi_cs_TpMerchantAccountID"  */
        decode_org_csapi_cs_TpMerchantAccountID_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpMerchantAccountID"  */
        u_octet4_loop_org_csapi_cs_IpChargingManager_createSplitChargingSession_users = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_IpChargingManager_createSplitChargingSession_users_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_IpChargingManager_createSplitChargingSession_users);

        for (i_org_csapi_cs_IpChargingManager_createSplitChargingSession_users=0; i_org_csapi_cs_IpChargingManager_createSplitChargingSession_users < u_octet4_loop_org_csapi_cs_IpChargingManager_createSplitChargingSession_users; i_org_csapi_cs_IpChargingManager_createSplitChargingSession_users++) {

            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
        }

        /*  Begin struct "org_csapi_cs_TpCorrelationID"  */
        decode_org_csapi_cs_TpCorrelationID_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cs_TpCorrelationID"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_cs_TpChargingSessionID"  */
            decode_org_csapi_cs_TpChargingSessionID_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cs_TpChargingSessionID"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/evaluation/IpAppPolicyDomain/reportNotification:1.0
 */
static void
decode_org_csapi_policy_evaluation_IpAppPolicyDomain_reportNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_policy_evaluation_IpAppPolicyDomain_reportNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_policy_TpPolicyEvent"  */
        decode_org_csapi_policy_TpPolicyEvent_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_policy_TpPolicyEvent"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/evaluation/IpAppPolicyDomain/evalPolicyRes:1.0
 */
static void
decode_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyRes_outputVariables;
    guint32   i_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyRes_outputVariables;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyRes_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyRes_outputVariables = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyRes_outputVariables_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyRes_outputVariables);

        for (i_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyRes_outputVariables=0; i_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyRes_outputVariables < u_octet4_loop_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyRes_outputVariables; i_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyRes_outputVariables++) {

            /*  Begin struct "org_csapi_policy_TpPolicyNameValue"  */
            decode_org_csapi_policy_TpPolicyNameValue_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_policy_TpPolicyNameValue"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/evaluation/IpAppPolicyDomain/evalPolicyErr:1.0
 */
static void
decode_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyErr_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyErr_error, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/evaluation/IpPolicyEvalManager/evalPolicy:1.0
 */
static void
decode_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_inputVariables;
    guint32   i_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_inputVariables;
    guint32   u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_return;
    guint32   i_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_domainName);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_signatureName);

        u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_inputVariables = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_inputVariables_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_inputVariables);

        for (i_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_inputVariables=0; i_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_inputVariables < u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_inputVariables; i_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_inputVariables++) {

            /*  Begin struct "org_csapi_policy_TpPolicyNameValue"  */
            decode_org_csapi_policy_TpPolicyNameValue_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_policy_TpPolicyNameValue"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_return);

            for (i_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_return=0; i_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_return < u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_return; i_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_return++) {

                /*  Begin struct "org_csapi_policy_TpPolicyNameValue"  */
                decode_org_csapi_policy_TpPolicyNameValue_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_policy_TpPolicyNameValue"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/evaluation/IpPolicyEvalManager/evalPolicyReq:1.0
 */
static void
decode_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicyReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicyReq_inputVariables;
    guint32   i_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicyReq_inputVariables;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicyReq_domainName);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicyReq_signatureName);

        u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicyReq_inputVariables = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicyReq_inputVariables_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicyReq_inputVariables);

        for (i_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicyReq_inputVariables=0; i_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicyReq_inputVariables < u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicyReq_inputVariables; i_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicyReq_inputVariables++) {

            /*  Begin struct "org_csapi_policy_TpPolicyNameValue"  */
            decode_org_csapi_policy_TpPolicyNameValue_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_policy_TpPolicyNameValue"  */
        }

        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicyReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/evaluation/IpPolicyEvalManager/abortEvalPolicyReq:1.0
 */
static void
decode_org_csapi_policy_evaluation_IpPolicyEvalManager_abortEvalPolicyReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_evaluation_IpPolicyEvalManager_abortEvalPolicyReq_domainName);

        proto_tree_add_int(tree, hf_org_csapi_policy_evaluation_IpPolicyEvalManager_abortEvalPolicyReq_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/evaluation/IpPolicyEvalManager/generateEvent:1.0
 */
static void
decode_org_csapi_policy_evaluation_IpPolicyEvalManager_generateEvent(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_generateEvent_attributes;
    guint32   i_org_csapi_policy_evaluation_IpPolicyEvalManager_generateEvent_attributes;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_evaluation_IpPolicyEvalManager_generateEvent_domainName);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_evaluation_IpPolicyEvalManager_generateEvent_eventDefinitionName);

        u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_generateEvent_attributes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_evaluation_IpPolicyEvalManager_generateEvent_attributes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_generateEvent_attributes);

        for (i_org_csapi_policy_evaluation_IpPolicyEvalManager_generateEvent_attributes=0; i_org_csapi_policy_evaluation_IpPolicyEvalManager_generateEvent_attributes < u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_generateEvent_attributes; i_org_csapi_policy_evaluation_IpPolicyEvalManager_generateEvent_attributes++) {

            /*  Begin struct "org_csapi_TpAttribute"  */
            decode_org_csapi_TpAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAttribute"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/evaluation/IpPolicyEvalManager/createNotification:1.0
 */
static void
decode_org_csapi_policy_evaluation_IpPolicyEvalManager_createNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_createNotification_events;
    guint32   i_org_csapi_policy_evaluation_IpPolicyEvalManager_createNotification_events;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_evaluation_IpPolicyEvalManager_createNotification_domainName);

        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_createNotification_events = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_evaluation_IpPolicyEvalManager_createNotification_events_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_createNotification_events);

        for (i_org_csapi_policy_evaluation_IpPolicyEvalManager_createNotification_events=0; i_org_csapi_policy_evaluation_IpPolicyEvalManager_createNotification_events < u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_createNotification_events; i_org_csapi_policy_evaluation_IpPolicyEvalManager_createNotification_events++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_evaluation_IpPolicyEvalManager_createNotification_events);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_policy_evaluation_IpPolicyEvalManager_createNotification_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/evaluation/IpPolicyEvalManager/destroyNotification:1.0
 */
static void
decode_org_csapi_policy_evaluation_IpPolicyEvalManager_destroyNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_destroyNotification_events;
    guint32   i_org_csapi_policy_evaluation_IpPolicyEvalManager_destroyNotification_events;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_policy_evaluation_IpPolicyEvalManager_destroyNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_destroyNotification_events = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_evaluation_IpPolicyEvalManager_destroyNotification_events_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_destroyNotification_events);

        for (i_org_csapi_policy_evaluation_IpPolicyEvalManager_destroyNotification_events=0; i_org_csapi_policy_evaluation_IpPolicyEvalManager_destroyNotification_events < u_octet4_loop_org_csapi_policy_evaluation_IpPolicyEvalManager_destroyNotification_events; i_org_csapi_policy_evaluation_IpPolicyEvalManager_destroyNotification_events++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_evaluation_IpPolicyEvalManager_destroyNotification_events);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicy/getAttribute:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicy_getAttribute(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicy_getAttribute_attributeName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_TpAttribute"  */
            decode_org_csapi_TpAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAttribute"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicy/setAttribute:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicy_setAttribute(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_TpAttribute"  */
        decode_org_csapi_TpAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAttribute"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicy/getAttributes:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicy_getAttributes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicy_getAttributes_attributeNames;
    guint32   i_org_csapi_policy_provisioning_IpPolicy_getAttributes_attributeNames;
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicy_getAttributes_return;
    guint32   i_org_csapi_policy_provisioning_IpPolicy_getAttributes_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_policy_provisioning_IpPolicy_getAttributes_attributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicy_getAttributes_attributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicy_getAttributes_attributeNames);

        for (i_org_csapi_policy_provisioning_IpPolicy_getAttributes_attributeNames=0; i_org_csapi_policy_provisioning_IpPolicy_getAttributes_attributeNames < u_octet4_loop_org_csapi_policy_provisioning_IpPolicy_getAttributes_attributeNames; i_org_csapi_policy_provisioning_IpPolicy_getAttributes_attributeNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicy_getAttributes_attributeNames);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_policy_provisioning_IpPolicy_getAttributes_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicy_getAttributes_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicy_getAttributes_return);

            for (i_org_csapi_policy_provisioning_IpPolicy_getAttributes_return=0; i_org_csapi_policy_provisioning_IpPolicy_getAttributes_return < u_octet4_loop_org_csapi_policy_provisioning_IpPolicy_getAttributes_return; i_org_csapi_policy_provisioning_IpPolicy_getAttributes_return++) {

                /*  Begin struct "org_csapi_TpAttribute"  */
                decode_org_csapi_TpAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_TpAttribute"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicy/setAttributes:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicy_setAttributes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicy_setAttributes_targetAttributes;
    guint32   i_org_csapi_policy_provisioning_IpPolicy_setAttributes_targetAttributes;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_policy_provisioning_IpPolicy_setAttributes_targetAttributes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicy_setAttributes_targetAttributes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicy_setAttributes_targetAttributes);

        for (i_org_csapi_policy_provisioning_IpPolicy_setAttributes_targetAttributes=0; i_org_csapi_policy_provisioning_IpPolicy_setAttributes_targetAttributes < u_octet4_loop_org_csapi_policy_provisioning_IpPolicy_setAttributes_targetAttributes; i_org_csapi_policy_provisioning_IpPolicy_setAttributes_targetAttributes++) {

            /*  Begin struct "org_csapi_TpAttribute"  */
            decode_org_csapi_TpAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAttribute"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyIterator/getList:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyIterator_getList(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicyIterator_getList_return;
    guint32   i_org_csapi_policy_provisioning_IpPolicyIterator_getList_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_policy_provisioning_IpPolicyIterator_getList_startIndex, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_policy_provisioning_IpPolicyIterator_getList_numberRequested, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_policy_provisioning_IpPolicyIterator_getList_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicyIterator_getList_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicyIterator_getList_return);

            for (i_org_csapi_policy_provisioning_IpPolicyIterator_getList_return=0; i_org_csapi_policy_provisioning_IpPolicyIterator_getList_return < u_octet4_loop_org_csapi_policy_provisioning_IpPolicyIterator_getList_return; i_org_csapi_policy_provisioning_IpPolicyIterator_getList_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyIterator_getList_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRepository/getParentRepository:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRepository_getParentRepository(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRepository/createRepository:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRepository_createRepository(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyRepository_createRepository_repositoryName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRepository/getRepository:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRepository_getRepository(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyRepository_getRepository_repositoryName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRepository/removeRepository:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRepository_removeRepository(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyRepository_removeRepository_repositoryName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRepository/getRepositoryCount:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRepository_getRepositoryCount(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_policy_provisioning_IpPolicyRepository_getRepositoryCount_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRepository/getRepositoryIterator:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRepository_getRepositoryIterator(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRepository/createCondition:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRepository_createCondition(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRepository_createCondition_conditionAttributes;
    guint32   i_org_csapi_policy_provisioning_IpPolicyRepository_createCondition_conditionAttributes;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyRepository_createCondition_conditionName);

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicyRepository_createCondition_conditionType, tvb, *offset-4, 4, u_octet4);

        u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRepository_createCondition_conditionAttributes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicyRepository_createCondition_conditionAttributes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRepository_createCondition_conditionAttributes);

        for (i_org_csapi_policy_provisioning_IpPolicyRepository_createCondition_conditionAttributes=0; i_org_csapi_policy_provisioning_IpPolicyRepository_createCondition_conditionAttributes < u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRepository_createCondition_conditionAttributes; i_org_csapi_policy_provisioning_IpPolicyRepository_createCondition_conditionAttributes++) {

            /*  Begin struct "org_csapi_TpAttribute"  */
            decode_org_csapi_TpAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAttribute"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRepository/getCondition:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRepository_getCondition(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyRepository_getCondition_conditionName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRepository/removeCondition:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRepository_removeCondition(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyRepository_removeCondition_conditionName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRepository/getConditionCount:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRepository_getConditionCount(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_policy_provisioning_IpPolicyRepository_getConditionCount_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRepository/getConditionIterator:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRepository_getConditionIterator(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRepository/createAction:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRepository_createAction(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRepository_createAction_actionAttributes;
    guint32   i_org_csapi_policy_provisioning_IpPolicyRepository_createAction_actionAttributes;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyRepository_createAction_actionName);

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicyRepository_createAction_actionType, tvb, *offset-4, 4, u_octet4);

        u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRepository_createAction_actionAttributes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicyRepository_createAction_actionAttributes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRepository_createAction_actionAttributes);

        for (i_org_csapi_policy_provisioning_IpPolicyRepository_createAction_actionAttributes=0; i_org_csapi_policy_provisioning_IpPolicyRepository_createAction_actionAttributes < u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRepository_createAction_actionAttributes; i_org_csapi_policy_provisioning_IpPolicyRepository_createAction_actionAttributes++) {

            /*  Begin struct "org_csapi_TpAttribute"  */
            decode_org_csapi_TpAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAttribute"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRepository/getAction:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRepository_getAction(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyRepository_getAction_actionName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRepository/removeAction:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRepository_removeAction(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyRepository_removeAction_actionName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRepository/getActionCount:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRepository_getActionCount(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_policy_provisioning_IpPolicyRepository_getActionCount_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRepository/getActionIterator:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRepository_getActionIterator(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRule/getParentGroup:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRule_getParentGroup(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRule/getParentDomain:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRule_getParentDomain(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRule/createCondition:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRule_createCondition(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_createCondition_conditionAttributes;
    guint32   i_org_csapi_policy_provisioning_IpPolicyRule_createCondition_conditionAttributes;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyRule_createCondition_conditionName);

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicyRule_createCondition_conditionType, tvb, *offset-4, 4, u_octet4);

        u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_createCondition_conditionAttributes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicyRule_createCondition_conditionAttributes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_createCondition_conditionAttributes);

        for (i_org_csapi_policy_provisioning_IpPolicyRule_createCondition_conditionAttributes=0; i_org_csapi_policy_provisioning_IpPolicyRule_createCondition_conditionAttributes < u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_createCondition_conditionAttributes; i_org_csapi_policy_provisioning_IpPolicyRule_createCondition_conditionAttributes++) {

            /*  Begin struct "org_csapi_TpAttribute"  */
            decode_org_csapi_TpAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAttribute"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRule/getCondition:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRule_getCondition(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyRule_getCondition_conditionName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRule/removeCondition:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRule_removeCondition(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyRule_removeCondition_conditionName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRule/getConditionCount:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRule_getConditionCount(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_policy_provisioning_IpPolicyRule_getConditionCount_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRule/getConditionIterator:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRule_getConditionIterator(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRule/createAction:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRule_createAction(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_createAction_actionAttributes;
    guint32   i_org_csapi_policy_provisioning_IpPolicyRule_createAction_actionAttributes;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyRule_createAction_actionName);

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicyRule_createAction_actionType, tvb, *offset-4, 4, u_octet4);

        u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_createAction_actionAttributes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicyRule_createAction_actionAttributes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_createAction_actionAttributes);

        for (i_org_csapi_policy_provisioning_IpPolicyRule_createAction_actionAttributes=0; i_org_csapi_policy_provisioning_IpPolicyRule_createAction_actionAttributes < u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_createAction_actionAttributes; i_org_csapi_policy_provisioning_IpPolicyRule_createAction_actionAttributes++) {

            /*  Begin struct "org_csapi_TpAttribute"  */
            decode_org_csapi_TpAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAttribute"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRule/getAction:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRule_getAction(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyRule_getAction_actionName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRule/removeAction:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRule_removeAction(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyRule_removeAction_actionName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRule/getActionCount:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRule_getActionCount(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_policy_provisioning_IpPolicyRule_getActionCount_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRule/getActionIterator:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRule_getActionIterator(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRule/setValidityPeriodConditionByName:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRule_setValidityPeriodConditionByName(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyRule_setValidityPeriodConditionByName_conditionName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRule/setValidityPeriodCondition:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRule_setValidityPeriodCondition(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRule/getValidityPeriodCondition:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRule_getValidityPeriodCondition(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRule/unsetValidityPeriodCondition:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRule_unsetValidityPeriodCondition(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRule/setConditionList:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRule_setConditionList(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_setConditionList_conditionList;
    guint32   i_org_csapi_policy_provisioning_IpPolicyRule_setConditionList_conditionList;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_setConditionList_conditionList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicyRule_setConditionList_conditionList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_setConditionList_conditionList);

        for (i_org_csapi_policy_provisioning_IpPolicyRule_setConditionList_conditionList=0; i_org_csapi_policy_provisioning_IpPolicyRule_setConditionList_conditionList < u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_setConditionList_conditionList; i_org_csapi_policy_provisioning_IpPolicyRule_setConditionList_conditionList++) {

            /*  Begin struct "org_csapi_policy_TpPolicyConditionListElement"  */
            decode_org_csapi_policy_TpPolicyConditionListElement_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_policy_TpPolicyConditionListElement"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRule/getConditionList:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRule_getConditionList(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_getConditionList_return;
    guint32   i_org_csapi_policy_provisioning_IpPolicyRule_getConditionList_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_getConditionList_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicyRule_getConditionList_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_getConditionList_return);

            for (i_org_csapi_policy_provisioning_IpPolicyRule_getConditionList_return=0; i_org_csapi_policy_provisioning_IpPolicyRule_getConditionList_return < u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_getConditionList_return; i_org_csapi_policy_provisioning_IpPolicyRule_getConditionList_return++) {

                /*  Begin struct "org_csapi_policy_TpPolicyConditionListElement"  */
                decode_org_csapi_policy_TpPolicyConditionListElement_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_policy_TpPolicyConditionListElement"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRule/setActionList:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRule_setActionList(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_setActionList_actionList;
    guint32   i_org_csapi_policy_provisioning_IpPolicyRule_setActionList_actionList;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_setActionList_actionList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicyRule_setActionList_actionList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_setActionList_actionList);

        for (i_org_csapi_policy_provisioning_IpPolicyRule_setActionList_actionList=0; i_org_csapi_policy_provisioning_IpPolicyRule_setActionList_actionList < u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_setActionList_actionList; i_org_csapi_policy_provisioning_IpPolicyRule_setActionList_actionList++) {

            /*  Begin struct "org_csapi_policy_TpPolicyActionListElement"  */
            decode_org_csapi_policy_TpPolicyActionListElement_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_policy_TpPolicyActionListElement"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyRule/getActionList:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyRule_getActionList(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_getActionList_return;
    guint32   i_org_csapi_policy_provisioning_IpPolicyRule_getActionList_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_getActionList_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicyRule_getActionList_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_getActionList_return);

            for (i_org_csapi_policy_provisioning_IpPolicyRule_getActionList_return=0; i_org_csapi_policy_provisioning_IpPolicyRule_getActionList_return < u_octet4_loop_org_csapi_policy_provisioning_IpPolicyRule_getActionList_return; i_org_csapi_policy_provisioning_IpPolicyRule_getActionList_return++) {

                /*  Begin struct "org_csapi_policy_TpPolicyActionListElement"  */
                decode_org_csapi_policy_TpPolicyActionListElement_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_policy_TpPolicyActionListElement"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyManager/createDomain:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyManager_createDomain(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyManager_createDomain_domainName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyManager/getDomain:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyManager_getDomain(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyManager_getDomain_domainName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyManager/removeDomain:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyManager_removeDomain(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyManager_removeDomain_domainName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyManager/getDomainCount:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyManager_getDomainCount(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_policy_provisioning_IpPolicyManager_getDomainCount_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyManager/getDomainIterator:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyManager_getDomainIterator(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyManager/findMatchingDomains:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_matchingAttributes;
    guint32   i_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_matchingAttributes;
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_return;
    guint32   i_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_matchingAttributes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_matchingAttributes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_matchingAttributes);

        for (i_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_matchingAttributes=0; i_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_matchingAttributes < u_octet4_loop_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_matchingAttributes; i_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_matchingAttributes++) {

            /*  Begin struct "org_csapi_TpAttribute"  */
            decode_org_csapi_TpAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAttribute"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_return);

            for (i_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_return=0; i_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_return < u_octet4_loop_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_return; i_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyManager/createRepository:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyManager_createRepository(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyManager_createRepository_repositoryName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyManager/getRepository:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyManager_getRepository(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyManager_getRepository_repositoryName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyManager/removeRepository:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyManager_removeRepository(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyManager_removeRepository_repositoryName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyManager/getRepositoryCount:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyManager_getRepositoryCount(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_policy_provisioning_IpPolicyManager_getRepositoryCount_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyManager/getRepositoryIterator:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyManager_getRepositoryIterator(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyManager/startTransaction:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyManager_startTransaction(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyManager/commitTransaction:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyManager_commitTransaction(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_boolean(tree, hf_org_csapi_policy_provisioning_IpPolicyManager_commitTransaction_return, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyManager/abortTransaction:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyManager_abortTransaction(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyGroup/getParentDomain:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyGroup_getParentDomain(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyGroup/getParentGroup:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyGroup_getParentGroup(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyGroup/createGroup:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyGroup_createGroup(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyGroup_createGroup_groupName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyGroup/getGroup:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyGroup_getGroup(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyGroup_getGroup_groupName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyGroup/removeGroup:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyGroup_removeGroup(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyGroup_removeGroup_groupName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyGroup/getGroupCount:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyGroup_getGroupCount(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_policy_provisioning_IpPolicyGroup_getGroupCount_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyGroup/getGroupIterator:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyGroup_getGroupIterator(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyGroup/createRule:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyGroup_createRule(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyGroup_createRule_ruleName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyGroup/getRule:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyGroup_getRule(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyGroup_getRule_ruleName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyGroup/removeRule:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyGroup_removeRule(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyGroup_removeRule_ruleName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyGroup/getRuleCount:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyGroup_getRuleCount(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_policy_provisioning_IpPolicyGroup_getRuleCount_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyGroup/getRuleIterator:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyGroup_getRuleIterator(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyEventDefinition/setRequiredAttributes:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyEventDefinition_setRequiredAttributes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicyEventDefinition_setRequiredAttributes_requiredAttributes;
    guint32   i_org_csapi_policy_provisioning_IpPolicyEventDefinition_setRequiredAttributes_requiredAttributes;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_policy_provisioning_IpPolicyEventDefinition_setRequiredAttributes_requiredAttributes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicyEventDefinition_setRequiredAttributes_requiredAttributes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicyEventDefinition_setRequiredAttributes_requiredAttributes);

        for (i_org_csapi_policy_provisioning_IpPolicyEventDefinition_setRequiredAttributes_requiredAttributes=0; i_org_csapi_policy_provisioning_IpPolicyEventDefinition_setRequiredAttributes_requiredAttributes < u_octet4_loop_org_csapi_policy_provisioning_IpPolicyEventDefinition_setRequiredAttributes_requiredAttributes; i_org_csapi_policy_provisioning_IpPolicyEventDefinition_setRequiredAttributes_requiredAttributes++) {

            /*  Begin struct "org_csapi_TpAttribute"  */
            decode_org_csapi_TpAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAttribute"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyEventDefinition/setOptionalAttributes:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyEventDefinition_setOptionalAttributes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicyEventDefinition_setOptionalAttributes_optionalAttributes;
    guint32   i_org_csapi_policy_provisioning_IpPolicyEventDefinition_setOptionalAttributes_optionalAttributes;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_policy_provisioning_IpPolicyEventDefinition_setOptionalAttributes_optionalAttributes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicyEventDefinition_setOptionalAttributes_optionalAttributes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicyEventDefinition_setOptionalAttributes_optionalAttributes);

        for (i_org_csapi_policy_provisioning_IpPolicyEventDefinition_setOptionalAttributes_optionalAttributes=0; i_org_csapi_policy_provisioning_IpPolicyEventDefinition_setOptionalAttributes_optionalAttributes < u_octet4_loop_org_csapi_policy_provisioning_IpPolicyEventDefinition_setOptionalAttributes_optionalAttributes; i_org_csapi_policy_provisioning_IpPolicyEventDefinition_setOptionalAttributes_optionalAttributes++) {

            /*  Begin struct "org_csapi_TpAttribute"  */
            decode_org_csapi_TpAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAttribute"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyEventDefinition/getRequiredAttributes:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyEventDefinition_getRequiredAttributes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicyEventDefinition_getRequiredAttributes_return;
    guint32   i_org_csapi_policy_provisioning_IpPolicyEventDefinition_getRequiredAttributes_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_policy_provisioning_IpPolicyEventDefinition_getRequiredAttributes_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicyEventDefinition_getRequiredAttributes_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicyEventDefinition_getRequiredAttributes_return);

            for (i_org_csapi_policy_provisioning_IpPolicyEventDefinition_getRequiredAttributes_return=0; i_org_csapi_policy_provisioning_IpPolicyEventDefinition_getRequiredAttributes_return < u_octet4_loop_org_csapi_policy_provisioning_IpPolicyEventDefinition_getRequiredAttributes_return; i_org_csapi_policy_provisioning_IpPolicyEventDefinition_getRequiredAttributes_return++) {

                /*  Begin struct "org_csapi_TpAttribute"  */
                decode_org_csapi_TpAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_TpAttribute"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyEventDefinition/getOptionalAttributes:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyEventDefinition_getOptionalAttributes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicyEventDefinition_getOptionalAttributes_return;
    guint32   i_org_csapi_policy_provisioning_IpPolicyEventDefinition_getOptionalAttributes_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_policy_provisioning_IpPolicyEventDefinition_getOptionalAttributes_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicyEventDefinition_getOptionalAttributes_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicyEventDefinition_getOptionalAttributes_return);

            for (i_org_csapi_policy_provisioning_IpPolicyEventDefinition_getOptionalAttributes_return=0; i_org_csapi_policy_provisioning_IpPolicyEventDefinition_getOptionalAttributes_return < u_octet4_loop_org_csapi_policy_provisioning_IpPolicyEventDefinition_getOptionalAttributes_return; i_org_csapi_policy_provisioning_IpPolicyEventDefinition_getOptionalAttributes_return++) {

                /*  Begin struct "org_csapi_TpAttribute"  */
                decode_org_csapi_TpAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_TpAttribute"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyEventDefinition/getParentDomain:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyEventDefinition_getParentDomain(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyCondition/getParentRepository:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyCondition_getParentRepository(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyCondition/getParentRule:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyCondition_getParentRule(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyAction/getParentRepository:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyAction_getParentRepository(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyAction/getParentRule:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyAction_getParentRule(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicySignature/setInputVariables:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicySignature_setInputVariables(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_setInputVariables_inputVariables;
    guint32   i_org_csapi_policy_provisioning_IpPolicySignature_setInputVariables_inputVariables;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_setInputVariables_inputVariables = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicySignature_setInputVariables_inputVariables_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_setInputVariables_inputVariables);

        for (i_org_csapi_policy_provisioning_IpPolicySignature_setInputVariables_inputVariables=0; i_org_csapi_policy_provisioning_IpPolicySignature_setInputVariables_inputVariables < u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_setInputVariables_inputVariables; i_org_csapi_policy_provisioning_IpPolicySignature_setInputVariables_inputVariables++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicySignature_setInputVariables_inputVariables);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicySignature/setOutputVariables:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicySignature_setOutputVariables(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_setOutputVariables_outputVariables;
    guint32   i_org_csapi_policy_provisioning_IpPolicySignature_setOutputVariables_outputVariables;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_setOutputVariables_outputVariables = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicySignature_setOutputVariables_outputVariables_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_setOutputVariables_outputVariables);

        for (i_org_csapi_policy_provisioning_IpPolicySignature_setOutputVariables_outputVariables=0; i_org_csapi_policy_provisioning_IpPolicySignature_setOutputVariables_outputVariables < u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_setOutputVariables_outputVariables; i_org_csapi_policy_provisioning_IpPolicySignature_setOutputVariables_outputVariables++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicySignature_setOutputVariables_outputVariables);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicySignature/getInputVariables:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicySignature_getInputVariables(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_getInputVariables_return;
    guint32   i_org_csapi_policy_provisioning_IpPolicySignature_getInputVariables_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_getInputVariables_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicySignature_getInputVariables_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_getInputVariables_return);

            for (i_org_csapi_policy_provisioning_IpPolicySignature_getInputVariables_return=0; i_org_csapi_policy_provisioning_IpPolicySignature_getInputVariables_return < u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_getInputVariables_return; i_org_csapi_policy_provisioning_IpPolicySignature_getInputVariables_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicySignature_getInputVariables_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicySignature/getOutputVariables:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicySignature_getOutputVariables(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_getOutputVariables_return;
    guint32   i_org_csapi_policy_provisioning_IpPolicySignature_getOutputVariables_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_getOutputVariables_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicySignature_getOutputVariables_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_getOutputVariables_return);

            for (i_org_csapi_policy_provisioning_IpPolicySignature_getOutputVariables_return=0; i_org_csapi_policy_provisioning_IpPolicySignature_getOutputVariables_return < u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_getOutputVariables_return; i_org_csapi_policy_provisioning_IpPolicySignature_getOutputVariables_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicySignature_getOutputVariables_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicySignature/setGroupNames:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicySignature_setGroupNames(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_setGroupNames_groupNames;
    guint32   i_org_csapi_policy_provisioning_IpPolicySignature_setGroupNames_groupNames;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_setGroupNames_groupNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicySignature_setGroupNames_groupNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_setGroupNames_groupNames);

        for (i_org_csapi_policy_provisioning_IpPolicySignature_setGroupNames_groupNames=0; i_org_csapi_policy_provisioning_IpPolicySignature_setGroupNames_groupNames < u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_setGroupNames_groupNames; i_org_csapi_policy_provisioning_IpPolicySignature_setGroupNames_groupNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicySignature_setGroupNames_groupNames);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicySignature/setPolicyRoles:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicySignature_setPolicyRoles(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_setPolicyRoles_roleNames;
    guint32   i_org_csapi_policy_provisioning_IpPolicySignature_setPolicyRoles_roleNames;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_setPolicyRoles_roleNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicySignature_setPolicyRoles_roleNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_setPolicyRoles_roleNames);

        for (i_org_csapi_policy_provisioning_IpPolicySignature_setPolicyRoles_roleNames=0; i_org_csapi_policy_provisioning_IpPolicySignature_setPolicyRoles_roleNames < u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_setPolicyRoles_roleNames; i_org_csapi_policy_provisioning_IpPolicySignature_setPolicyRoles_roleNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicySignature_setPolicyRoles_roleNames);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicySignature/getGroupNames:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicySignature_getGroupNames(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_getGroupNames_return;
    guint32   i_org_csapi_policy_provisioning_IpPolicySignature_getGroupNames_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_getGroupNames_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicySignature_getGroupNames_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_getGroupNames_return);

            for (i_org_csapi_policy_provisioning_IpPolicySignature_getGroupNames_return=0; i_org_csapi_policy_provisioning_IpPolicySignature_getGroupNames_return < u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_getGroupNames_return; i_org_csapi_policy_provisioning_IpPolicySignature_getGroupNames_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicySignature_getGroupNames_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicySignature/getPolicyRoles:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicySignature_getPolicyRoles(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_getPolicyRoles_return;
    guint32   i_org_csapi_policy_provisioning_IpPolicySignature_getPolicyRoles_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_getPolicyRoles_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicySignature_getPolicyRoles_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_getPolicyRoles_return);

            for (i_org_csapi_policy_provisioning_IpPolicySignature_getPolicyRoles_return=0; i_org_csapi_policy_provisioning_IpPolicySignature_getPolicyRoles_return < u_octet4_loop_org_csapi_policy_provisioning_IpPolicySignature_getPolicyRoles_return; i_org_csapi_policy_provisioning_IpPolicySignature_getPolicyRoles_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicySignature_getPolicyRoles_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicySignature/getParentDomain:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicySignature_getParentDomain(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/getParentDomain:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_getParentDomain(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/createDomain:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_createDomain(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_createDomain_domainName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/getDomain:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_getDomain(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_getDomain_domainName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/removeDomain:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_removeDomain(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_removeDomain_domainName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/getDomainCount:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_getDomainCount(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_policy_provisioning_IpPolicyDomain_getDomainCount_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/getDomainIterator:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_getDomainIterator(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/createGroup:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_createGroup(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_createGroup_groupName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/getGroup:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_getGroup(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_getGroup_groupName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/removeGroup:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_removeGroup(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_removeGroup_groupName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/getGroupCount:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_getGroupCount(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_policy_provisioning_IpPolicyDomain_getGroupCount_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/getGroupIterator:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_getGroupIterator(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/createRule:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_createRule(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_createRule_ruleName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/getRule:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_getRule(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_getRule_ruleName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/removeRule:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_removeRule(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_removeRule_ruleName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/getRuleCount:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_getRuleCount(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_policy_provisioning_IpPolicyDomain_getRuleCount_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/getRuleIterator:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_getRuleIterator(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/createEventDefinition:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_requiredAttributes;
    guint32   i_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_requiredAttributes;
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_optionalAttributes;
    guint32   i_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_optionalAttributes;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_eventDefinitionName);

        u_octet4_loop_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_requiredAttributes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_requiredAttributes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_requiredAttributes);

        for (i_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_requiredAttributes=0; i_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_requiredAttributes < u_octet4_loop_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_requiredAttributes; i_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_requiredAttributes++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_requiredAttributes);

        }

        u_octet4_loop_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_optionalAttributes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_optionalAttributes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_optionalAttributes);

        for (i_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_optionalAttributes=0; i_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_optionalAttributes < u_octet4_loop_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_optionalAttributes; i_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_optionalAttributes++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_optionalAttributes);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/getEventDefinition:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_getEventDefinition(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_getEventDefinition_eventDefinitionName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/removeEventDefinition:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_removeEventDefinition(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_removeEventDefinition_eventDefinitionName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/getEventDefinitionCount:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_getEventDefinitionCount(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_policy_provisioning_IpPolicyDomain_getEventDefinitionCount_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/getEventDefinitionIterator:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_getEventDefinitionIterator(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/createVariableSet:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_createVariableSet(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_createVariableSet_variableSetName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/getVariableSet:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_getVariableSet(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_provisioning_IpPolicyDomain_getVariableSet_return;
    guint32   i_org_csapi_policy_provisioning_IpPolicyDomain_getVariableSet_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariableSet_variableSetName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_policy_provisioning_IpPolicyDomain_getVariableSet_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariableSet_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_provisioning_IpPolicyDomain_getVariableSet_return);

            for (i_org_csapi_policy_provisioning_IpPolicyDomain_getVariableSet_return=0; i_org_csapi_policy_provisioning_IpPolicyDomain_getVariableSet_return < u_octet4_loop_org_csapi_policy_provisioning_IpPolicyDomain_getVariableSet_return; i_org_csapi_policy_provisioning_IpPolicyDomain_getVariableSet_return++) {

                /*  Begin struct "org_csapi_policy_TpPolicyVar"  */
                decode_org_csapi_policy_TpPolicyVar_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_policy_TpPolicyVar"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/removeVariableSet:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_removeVariableSet(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_removeVariableSet_variableSetName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/getVariableSetCount:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_getVariableSetCount(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariableSetCount_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/getVariableSetIterator:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_getVariableSetIterator(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/createVariable:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_createVariable(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_createVariable_variableSetName);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_createVariable_variableName);

        /*  Begin union "org_csapi_policy_TpPolicyType"  */
        decode_org_csapi_policy_TpPolicyType_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_policy_TpPolicyType"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/setVariableValue:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_setVariableValue(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_setVariableValue_variableSetName);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_setVariableValue_variableName);

        get_CDR_any(tvb, pinfo, tree, item, offset, stream_is_big_endian, boundary, header);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/getVariableType:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_getVariableType(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariableType_variableSetName);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariableType_variableName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin union "org_csapi_policy_TpPolicyType"  */
            decode_org_csapi_policy_TpPolicyType_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_policy_TpPolicyType"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/getVariableValue:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_getVariableValue(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariableValue_variableSetName);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariableValue_variableName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_any(tvb, pinfo, tree, item, offset, stream_is_big_endian, boundary, header);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/getVariable:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_getVariable(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariable_variableSetName);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariable_variableName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_policy_TpPolicyVar"  */
            decode_org_csapi_policy_TpPolicyVar_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_policy_TpPolicyVar"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/removeVariable:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_removeVariable(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_removeVariable_variablSetName);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_removeVariable_variableName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/createSignature:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_createSignature(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_createSignature_signatureName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/getSignature:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_getSignature(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_getSignature_signatureName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/removeSignature:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_removeSignature(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_provisioning_IpPolicyDomain_removeSignature_signatureName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/getSignatureCount:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_getSignatureCount(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_policy_provisioning_IpPolicyDomain_getSignatureCount_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/policy/provisioning/IpPolicyDomain/getSignatureIterator:1.0
 */
static void
decode_org_csapi_policy_provisioning_IpPolicyDomain_getSignatureIterator(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityManagement/createIdentity:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_identityTypes;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_identityTypes;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_identity);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_identityTypes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_identityTypes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_identityTypes);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_identityTypes=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_identityTypes < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_identityTypes; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_identityTypes++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_identityTypes);

        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityManagement/deleteIdentity:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteIdentity(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteIdentity_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteIdentity_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteIdentity_identity);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteIdentity_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteIdentity_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteIdentity_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteIdentity_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteIdentity_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteIdentity_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteIdentity_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteIdentity_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityManagement/isIdentity:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_isIdentity(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_isIdentity_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_isIdentity_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isIdentity_identity);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_isIdentity_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isIdentity_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_isIdentity_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_isIdentity_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_isIdentity_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_isIdentity_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_isIdentity_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isIdentity_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_boolean(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isIdentity_return, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityManagement/createGroupIdentity:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_identityTypes;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_identityTypes;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_identity);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_identityTypes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_identityTypes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_identityTypes);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_identityTypes=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_identityTypes < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_identityTypes; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_identityTypes++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_identityTypes);

        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityManagement/deleteGroupIdentity:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteGroupIdentity(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteGroupIdentity_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteGroupIdentity_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteGroupIdentity_identity);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteGroupIdentity_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteGroupIdentity_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteGroupIdentity_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteGroupIdentity_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteGroupIdentity_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteGroupIdentity_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteGroupIdentity_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteGroupIdentity_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityManagement/addToGroup:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_addToGroup(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_addToGroup_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_addToGroup_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addToGroup_group);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addToGroup_member);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_addToGroup_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addToGroup_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_addToGroup_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_addToGroup_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_addToGroup_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_addToGroup_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_addToGroup_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addToGroup_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityManagement/removeFromGroup:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeFromGroup(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeFromGroup_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeFromGroup_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeFromGroup_group);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeFromGroup_identity);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeFromGroup_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeFromGroup_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeFromGroup_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeFromGroup_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeFromGroup_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeFromGroup_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeFromGroup_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeFromGroup_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityManagement/listMembers:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_authToken;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_identity);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_return);

            for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_return=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_return; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityManagement/isGroupIdentity:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_isGroupIdentity(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_isGroupIdentity_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_isGroupIdentity_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isGroupIdentity_identity);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_isGroupIdentity_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isGroupIdentity_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_isGroupIdentity_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_isGroupIdentity_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_isGroupIdentity_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_isGroupIdentity_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_isGroupIdentity_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isGroupIdentity_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_boolean(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isGroupIdentity_return, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityManagement/listGroupMembership:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_authToken;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_identity);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_return);

            for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_return=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_return; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityManagement/addAlias:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_addAlias(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_addAlias_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_addAlias_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addAlias_identity);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addAlias_alias);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_addAlias_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addAlias_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_addAlias_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_addAlias_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_addAlias_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_addAlias_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_addAlias_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addAlias_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityManagement/removeAliases:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeAliases(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeAliases_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeAliases_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeAliases_identity);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeAliases_alias);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeAliases_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeAliases_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeAliases_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeAliases_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeAliases_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeAliases_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeAliases_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeAliases_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityManagement/listAliases:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_authToken;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_identity);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_return);

            for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_return=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_return; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityManagement/lookupByAlias:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_lookupByAlias(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_lookupByAlias_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_lookupByAlias_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_lookupByAlias_alias);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_lookupByAlias_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_lookupByAlias_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_lookupByAlias_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_lookupByAlias_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_lookupByAlias_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_lookupByAlias_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_lookupByAlias_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_lookupByAlias_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_lookupByAlias_return);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityManagement/associateTypes:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_identityTypes;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_identityTypes;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_identity);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_identityTypes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_identityTypes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_identityTypes);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_identityTypes=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_identityTypes < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_identityTypes; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_identityTypes++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_identityTypes);

        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityManagement/disassociateTypes:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_identityTypes;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_identityTypes;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_identity);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_identityTypes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_identityTypes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_identityTypes);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_identityTypes=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_identityTypes < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_identityTypes; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_identityTypes++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_identityTypes);

        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityManagement/listTypesOfIdentity:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_authToken;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_identity);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_return);

            for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_return=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_return; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityManagement/hasType:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType_identity);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType_typeName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_boolean(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType_return, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityManagement/getIdentityAttributes:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_attributeNames;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_attributeNames;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_authToken;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_identity);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_identityType);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_attributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_attributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_attributeNames);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_attributeNames=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_attributeNames < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_attributeNames; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_attributeNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_attributeNames);

        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_return);

            for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_return=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_return; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_return++) {

                /*  Begin struct "org_csapi_pam_TpPAMAttribute"  */
                decode_org_csapi_pam_TpPAMAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_pam_TpPAMAttribute"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityManagement/setIdentityAttributes:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_attributes;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_attributes;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_identity);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_identityType);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_attributes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_attributes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_attributes);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_attributes=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_attributes < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_attributes; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_attributes++) {

            /*  Begin struct "org_csapi_pam_TpPAMAttribute"  */
            decode_org_csapi_pam_TpPAMAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_pam_TpPAMAttribute"  */
        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentManagement/createAgent:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_agentTypes;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_agentTypes;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_agentName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_agentTypes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_agentTypes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_agentTypes);

        for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_agentTypes=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_agentTypes < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_agentTypes; i_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_agentTypes++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_agentTypes);

        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_authToken; i_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentManagement/deleteAgent:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentManagement_deleteAgent(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_deleteAgent_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_deleteAgent_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_deleteAgent_agentName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_deleteAgent_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_deleteAgent_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_deleteAgent_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_deleteAgent_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_deleteAgent_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_deleteAgent_authToken; i_org_csapi_pam_provisioning_IpPAMAgentManagement_deleteAgent_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_deleteAgent_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentManagement/isAgent:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentManagement_isAgent(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_isAgent_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_isAgent_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isAgent_agentName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_isAgent_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isAgent_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_isAgent_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_isAgent_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_isAgent_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_isAgent_authToken; i_org_csapi_pam_provisioning_IpPAMAgentManagement_isAgent_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isAgent_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_boolean(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isAgent_return, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentManagement/enableCapabilities:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_capabilities;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_capabilities;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_agentName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_capabilities = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_capabilities_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_capabilities);

        for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_capabilities=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_capabilities < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_capabilities; i_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_capabilities++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_capabilities);

        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_authToken; i_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentManagement/disableCapabilities:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_capabilities;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_capabilities;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_agentName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_capabilities = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_capabilities_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_capabilities);

        for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_capabilities=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_capabilities < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_capabilities; i_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_capabilities++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_capabilities);

        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_authToken; i_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentManagement/listEnabledCapabilities:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_authToken;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_agentName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_authToken; i_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_return);

            for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_return=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_return; i_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentManagement/listAllCapabilities:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_authToken;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_agentName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_authToken; i_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_return);

            for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_return=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_return; i_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentManagement/isCapableOf:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf_agentName);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf_capability);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf_authToken; i_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_boolean(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf_return, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentManagement/associateTypes:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_agentTypes;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_agentTypes;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_agentName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_agentTypes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_agentTypes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_agentTypes);

        for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_agentTypes=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_agentTypes < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_agentTypes; i_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_agentTypes++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_agentTypes);

        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_authToken; i_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentManagement/disassociateTypes:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_agentTypes;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_agentTypes;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_agentName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_agentTypes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_agentTypes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_agentTypes);

        for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_agentTypes=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_agentTypes < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_agentTypes; i_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_agentTypes++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_agentTypes);

        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_authToken; i_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentManagement/listTypesOfAgent:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_authToken;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_agentName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_authToken; i_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_return);

            for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_return=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_return; i_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentManagement/hasType:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType_agentName);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType_typeName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType_authToken; i_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_boolean(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType_return, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentManagement/getAgentAttributes:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_attributeNames;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_attributeNames;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_authToken;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_agentName);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_agentType);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_attributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_attributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_attributeNames);

        for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_attributeNames=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_attributeNames < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_attributeNames; i_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_attributeNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_attributeNames);

        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_authToken; i_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_return);

            for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_return=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_return; i_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_return++) {

                /*  Begin struct "org_csapi_pam_TpPAMAttribute"  */
                decode_org_csapi_pam_TpPAMAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_pam_TpPAMAttribute"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentManagement/setAgentAttributes:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_attributes;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_attributes;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_agentName);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_agentType);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_attributes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_attributes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_attributes);

        for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_attributes=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_attributes < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_attributes; i_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_attributes++) {

            /*  Begin struct "org_csapi_pam_TpPAMAttribute"  */
            decode_org_csapi_pam_TpPAMAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_pam_TpPAMAttribute"  */
        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_authToken; i_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentAssignment/assignAgent:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentAssignment_assignAgent(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_assignAgent_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentAssignment_assignAgent_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_assignAgent_identity);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_assignAgent_agentName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_assignAgent_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_assignAgent_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_assignAgent_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentAssignment_assignAgent_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentAssignment_assignAgent_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_assignAgent_authToken; i_org_csapi_pam_provisioning_IpPAMAgentAssignment_assignAgent_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_assignAgent_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentAssignment/unassignAgent:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentAssignment_unassignAgent(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_unassignAgent_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentAssignment_unassignAgent_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_unassignAgent_identity);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_unassignAgent_agentName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_unassignAgent_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_unassignAgent_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_unassignAgent_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentAssignment_unassignAgent_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentAssignment_unassignAgent_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_unassignAgent_authToken; i_org_csapi_pam_provisioning_IpPAMAgentAssignment_unassignAgent_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_unassignAgent_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentAssignment/listAssignedAgents:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_authToken;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_identity);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_authToken; i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_return);

            for (i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_return=0; i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_return; i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentAssignment/listAssociatedIdentitiesOfAgent:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_authToken;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_agentName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_authToken; i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_return);

            for (i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_return=0; i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_return; i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentAssignment/listAssignedAgentsByCapability:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_authToken;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_identity);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_capability);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_authToken; i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_return);

            for (i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_return=0; i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_return; i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentAssignment/listCapabilitiesOfIdentity:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_authToken;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_identity);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_authToken; i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_return);

            for (i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_return=0; i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_return; i_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentAssignment/isIdentityCapableOf:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf_identity);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf_capability);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf_authToken; i_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_boolean(tree, hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf_return, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityTypeManagement/createIdentityAttribute:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityAttribute(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityAttribute_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityAttribute_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_pam_TpPAMAttributeDef"  */
        decode_org_csapi_pam_TpPAMAttributeDef_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMAttributeDef"  */
        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityAttribute_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityAttribute_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityAttribute_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityAttribute_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityAttribute_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityAttribute_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityAttribute_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityAttribute_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityTypeManagement/deleteIdentityAttribute:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityAttribute(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityAttribute_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityAttribute_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityAttribute_attributeName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityAttribute_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityAttribute_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityAttribute_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityAttribute_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityAttribute_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityAttribute_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityAttribute_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityAttribute_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityTypeManagement/getIdentityAttributeDefinition:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_getIdentityAttributeDefinition(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_getIdentityAttributeDefinition_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_getIdentityAttributeDefinition_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_getIdentityAttributeDefinition_attributeName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_getIdentityAttributeDefinition_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_getIdentityAttributeDefinition_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_getIdentityAttributeDefinition_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_getIdentityAttributeDefinition_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_getIdentityAttributeDefinition_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_getIdentityAttributeDefinition_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_getIdentityAttributeDefinition_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_getIdentityAttributeDefinition_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_pam_TpPAMAttributeDef"  */
            decode_org_csapi_pam_TpPAMAttributeDef_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_pam_TpPAMAttributeDef"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityTypeManagement/listAllIdentityAttributes:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_authToken;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_return);

            for (i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_return=0; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_return; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityTypeManagement/createIdentityType:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_attributeNames;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_attributeNames;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_typeName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_attributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_attributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_attributeNames);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_attributeNames=0; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_attributeNames < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_attributeNames; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_attributeNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_attributeNames);

        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityTypeManagement/deleteIdentityType:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityType(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityType_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityType_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityType_typeName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityType_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityType_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityType_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityType_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityType_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityType_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityType_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityType_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityTypeManagement/listIdentityTypes:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_authToken;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_return);

            for (i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_return=0; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_return; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityTypeManagement/addIdentityTypeAttributes:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_attributeNames;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_attributeNames;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_typeName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_attributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_attributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_attributeNames);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_attributeNames=0; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_attributeNames < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_attributeNames; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_attributeNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_attributeNames);

        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityTypeManagement/removeIdentityTypeAttributes:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_attributeNames;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_attributeNames;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_typeName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_attributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_attributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_attributeNames);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_attributeNames=0; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_attributeNames < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_attributeNames; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_attributeNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_attributeNames);

        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMIdentityTypeManagement/listIdentityTypeAttributes:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_authToken;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_typeName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_authToken=0; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_authToken; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_return);

            for (i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_return=0; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_return; i_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentTypeManagement/createAgentAttribute:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentAttribute(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentAttribute_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentAttribute_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_pam_TpPAMAttributeDef"  */
        decode_org_csapi_pam_TpPAMAttributeDef_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMAttributeDef"  */
        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentAttribute_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentAttribute_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentAttribute_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentAttribute_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentAttribute_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentAttribute_authToken; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentAttribute_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentAttribute_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentTypeManagement/deleteAgentAttribute:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentAttribute(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentAttribute_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentAttribute_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentAttribute_attributeName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentAttribute_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentAttribute_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentAttribute_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentAttribute_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentAttribute_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentAttribute_authToken; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentAttribute_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentAttribute_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentTypeManagement/getAgentAttributeDefinition:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_getAgentAttributeDefinition(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_getAgentAttributeDefinition_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_getAgentAttributeDefinition_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_getAgentAttributeDefinition_attributeName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_getAgentAttributeDefinition_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_getAgentAttributeDefinition_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_getAgentAttributeDefinition_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_getAgentAttributeDefinition_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_getAgentAttributeDefinition_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_getAgentAttributeDefinition_authToken; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_getAgentAttributeDefinition_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_getAgentAttributeDefinition_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_pam_TpPAMAttributeDef"  */
            decode_org_csapi_pam_TpPAMAttributeDef_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_pam_TpPAMAttributeDef"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentTypeManagement/listAllAgentAttributes:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_authToken;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_authToken; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_return);

            for (i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_return=0; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_return; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentTypeManagement/createAgentType:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_attributeNames;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_attributeNames;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_typeName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_attributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_attributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_attributeNames);

        for (i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_attributeNames=0; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_attributeNames < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_attributeNames; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_attributeNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_attributeNames);

        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_authToken; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentTypeManagement/deleteAgentType:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentType(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentType_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentType_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentType_typeName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentType_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentType_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentType_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentType_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentType_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentType_authToken; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentType_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentType_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentTypeManagement/listAgentTypes:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_authToken;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_authToken; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_return);

            for (i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_return=0; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_return; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentTypeManagement/addAgentTypeAttributes:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_attributeNames;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_attributeNames;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_typeName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_attributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_attributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_attributeNames);

        for (i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_attributeNames=0; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_attributeNames < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_attributeNames; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_attributeNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_attributeNames);

        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_authToken; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentTypeManagement/removeAgentTypeAttributes:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_attributeNames;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_attributeNames;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_typeName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_attributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_attributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_attributeNames);

        for (i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_attributeNames=0; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_attributeNames < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_attributeNames; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_attributeNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_attributeNames);

        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_authToken; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMAgentTypeManagement/listAgentTypeAttributes:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_authToken;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_typeName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_authToken=0; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_authToken; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_return);

            for (i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_return=0; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_return; i_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMCapabilityManagement/createCapabilityAttribute:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapabilityAttribute(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapabilityAttribute_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapabilityAttribute_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_pam_TpPAMAttributeDef"  */
        decode_org_csapi_pam_TpPAMAttributeDef_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMAttributeDef"  */
        u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapabilityAttribute_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapabilityAttribute_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapabilityAttribute_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapabilityAttribute_authToken=0; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapabilityAttribute_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapabilityAttribute_authToken; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapabilityAttribute_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapabilityAttribute_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMCapabilityManagement/deleteCapabilityAttribute:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapabilityAttribute(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapabilityAttribute_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapabilityAttribute_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapabilityAttribute_attributeName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapabilityAttribute_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapabilityAttribute_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapabilityAttribute_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapabilityAttribute_authToken=0; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapabilityAttribute_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapabilityAttribute_authToken; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapabilityAttribute_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapabilityAttribute_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMCapabilityManagement/getCapabilityAttributeDefinition:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_getCapabilityAttributeDefinition(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_getCapabilityAttributeDefinition_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_getCapabilityAttributeDefinition_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_getCapabilityAttributeDefinition_attributeName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_getCapabilityAttributeDefinition_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_getCapabilityAttributeDefinition_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_getCapabilityAttributeDefinition_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_getCapabilityAttributeDefinition_authToken=0; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_getCapabilityAttributeDefinition_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_getCapabilityAttributeDefinition_authToken; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_getCapabilityAttributeDefinition_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_getCapabilityAttributeDefinition_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_pam_TpPAMAttributeDef"  */
            decode_org_csapi_pam_TpPAMAttributeDef_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_pam_TpPAMAttributeDef"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMCapabilityManagement/listAllCapabilityAttributes:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_authToken;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_authToken=0; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_authToken; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_return);

            for (i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_return=0; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_return; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMCapabilityManagement/createCapability:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_attributeNames;
    guint32   i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_attributeNames;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_capabilityName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_attributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_attributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_attributeNames);

        for (i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_attributeNames=0; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_attributeNames < u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_attributeNames; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_attributeNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_attributeNames);

        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_authToken=0; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_authToken; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMCapabilityManagement/deleteCapability:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapability(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapability_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapability_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapability_capabilityName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapability_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapability_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapability_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapability_authToken=0; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapability_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapability_authToken; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapability_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapability_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMCapabilityManagement/listCapabilities:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_authToken;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_authToken=0; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_authToken; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_return);

            for (i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_return=0; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_return; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMCapabilityManagement/addCapabilityAttributes:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_attributeNames;
    guint32   i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_attributeNames;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_capabilityName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_attributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_attributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_attributeNames);

        for (i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_attributeNames=0; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_attributeNames < u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_attributeNames; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_attributeNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_attributeNames);

        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_authToken=0; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_authToken; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMCapabilityManagement/removeCapabilityAttributes:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_attributeNames;
    guint32   i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_attributeNames;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_capabilityName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_attributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_attributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_attributeNames);

        for (i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_attributeNames=0; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_attributeNames < u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_attributeNames; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_attributeNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_attributeNames);

        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_authToken=0; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_authToken; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMCapabilityManagement/listCapabilityAttributes:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_authToken;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_capabilityName);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_authToken=0; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_authToken; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_return);

            for (i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_return=0; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_return; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMCapabilityManagement/assignCapabilitiesToType:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_capabilities;
    guint32   i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_capabilities;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_agentType);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_capabilities = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_capabilities_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_capabilities);

        for (i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_capabilities=0; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_capabilities < u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_capabilities; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_capabilities++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_capabilities);

        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_authToken=0; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_authToken; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMCapabilityManagement/unassignCapabilitiesFromType:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_capabilities;
    guint32   i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_capabilities;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_agentType);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_capabilities = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_capabilities_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_capabilities);

        for (i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_capabilities=0; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_capabilities < u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_capabilities; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_capabilities++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_capabilities);

        }

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_authToken=0; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_authToken; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMCapabilityManagement/listCapabilitiesOfType:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_authToken;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_agentType);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_authToken=0; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_authToken; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_return);

            for (i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_return=0; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_return; i_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_return++) {

                giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_return);

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMProvisioningManager/getAuthToken:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_askerData;
    guint32   i_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_askerData;
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_return;
    guint32   i_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_askerData = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_askerData_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_askerData);

        for (i_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_askerData=0; i_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_askerData < u_octet4_loop_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_askerData; i_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_askerData++) {

            /*  Begin struct "org_csapi_TpAttribute"  */
            decode_org_csapi_TpAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAttribute"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_return);

            for (i_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_return=0; i_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_return < u_octet4_loop_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_return; i_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_return++) {

                proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_return, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMProvisioningManager/obtainInterface:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMProvisioningManager_obtainInterface(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_obtainInterface_interfaceName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMProvisioningManager/getAccessControl:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAccessControl(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAccessControl_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAccessControl_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAccessControl_identity);

        u_octet4_loop_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAccessControl_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAccessControl_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAccessControl_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAccessControl_authToken=0; i_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAccessControl_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAccessControl_authToken; i_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAccessControl_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAccessControl_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_pam_TpPAMAccessControlData"  */
            decode_org_csapi_pam_TpPAMAccessControlData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_pam_TpPAMAccessControlData"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/provisioning/IpPAMProvisioningManager/setAccessControl:1.0
 */
static void
decode_org_csapi_pam_provisioning_IpPAMProvisioningManager_setAccessControl(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_provisioning_IpPAMProvisioningManager_setAccessControl_authToken;
    guint32   i_org_csapi_pam_provisioning_IpPAMProvisioningManager_setAccessControl_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_setAccessControl_identity);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_setAccessControl_operation);

        /*  Begin struct "org_csapi_pam_TpPAMAccessControlData"  */
        decode_org_csapi_pam_TpPAMAccessControlData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMAccessControlData"  */
        u_octet4_loop_org_csapi_pam_provisioning_IpPAMProvisioningManager_setAccessControl_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_setAccessControl_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_provisioning_IpPAMProvisioningManager_setAccessControl_authToken);

        for (i_org_csapi_pam_provisioning_IpPAMProvisioningManager_setAccessControl_authToken=0; i_org_csapi_pam_provisioning_IpPAMProvisioningManager_setAccessControl_authToken < u_octet4_loop_org_csapi_pam_provisioning_IpPAMProvisioningManager_setAccessControl_authToken; i_org_csapi_pam_provisioning_IpPAMProvisioningManager_setAccessControl_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_setAccessControl_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/access/IpPAMIdentityPresence/setIdentityPresence:1.0
 */
static void
decode_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_attributes;
    guint32   i_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_attributes;
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_authToken;
    guint32   i_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_identity);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_identityType);

        u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_attributes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_attributes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_attributes);

        for (i_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_attributes=0; i_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_attributes < u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_attributes; i_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_attributes++) {

            /*  Begin struct "org_csapi_pam_TpPAMAttribute"  */
            decode_org_csapi_pam_TpPAMAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_pam_TpPAMAttribute"  */
        }

        u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_authToken);

        for (i_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_authToken=0; i_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_authToken < u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_authToken; i_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/access/IpPAMIdentityPresence/setIdentityPresenceExpiration:1.0
 */
static void
decode_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_attributeNames;
    guint32   i_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_attributeNames;
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_authToken;
    guint32   i_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_identity);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_identityType);

        u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_attributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_attributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_attributeNames);

        for (i_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_attributeNames=0; i_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_attributeNames < u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_attributeNames; i_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_attributeNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_attributeNames);

        }

        proto_tree_add_int64(tree, hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_expiresIn, tvb, *offset-8, 8, get_CDR_long_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_authToken);

        for (i_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_authToken=0; i_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_authToken < u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_authToken; i_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/access/IpPAMIdentityPresence/getIdentityPresence:1.0
 */
static void
decode_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_attributeNames;
    guint32   i_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_attributeNames;
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_authToken;
    guint32   i_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_authToken;
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_return;
    guint32   i_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_identity);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_identityType);

        u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_attributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_attributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_attributeNames);

        for (i_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_attributeNames=0; i_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_attributeNames < u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_attributeNames; i_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_attributeNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_attributeNames);

        }

        u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_authToken);

        for (i_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_authToken=0; i_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_authToken < u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_authToken; i_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_return);

            for (i_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_return=0; i_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_return < u_octet4_loop_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_return; i_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_return++) {

                /*  Begin struct "org_csapi_pam_TpPAMAttribute"  */
                decode_org_csapi_pam_TpPAMAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_pam_TpPAMAttribute"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/access/IpPAMAgentPresence/setAgentPresence:1.0
 */
static void
decode_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_attributes;
    guint32   i_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_attributes;
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_authToken;
    guint32   i_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_agent);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_agentType);

        u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_attributes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_attributes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_attributes);

        for (i_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_attributes=0; i_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_attributes < u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_attributes; i_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_attributes++) {

            /*  Begin struct "org_csapi_pam_TpPAMAttribute"  */
            decode_org_csapi_pam_TpPAMAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_pam_TpPAMAttribute"  */
        }

        u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_authToken);

        for (i_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_authToken=0; i_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_authToken < u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_authToken; i_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/access/IpPAMAgentPresence/setCapabilityPresence:1.0
 */
static void
decode_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_attributes;
    guint32   i_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_attributes;
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_authToken;
    guint32   i_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_agent);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_capability);

        u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_attributes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_attributes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_attributes);

        for (i_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_attributes=0; i_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_attributes < u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_attributes; i_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_attributes++) {

            /*  Begin struct "org_csapi_pam_TpPAMAttribute"  */
            decode_org_csapi_pam_TpPAMAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_pam_TpPAMAttribute"  */
        }

        u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_authToken);

        for (i_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_authToken=0; i_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_authToken < u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_authToken; i_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/access/IpPAMAgentPresence/setAgentPresenceExpiration:1.0
 */
static void
decode_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_attributeNames;
    guint32   i_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_attributeNames;
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_authToken;
    guint32   i_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_agent);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_agentType);

        u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_attributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_attributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_attributeNames);

        for (i_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_attributeNames=0; i_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_attributeNames < u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_attributeNames; i_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_attributeNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_attributeNames);

        }

        proto_tree_add_int64(tree, hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_expiresIn, tvb, *offset-8, 8, get_CDR_long_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_authToken);

        for (i_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_authToken=0; i_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_authToken < u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_authToken; i_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/access/IpPAMAgentPresence/setCapabilityPresenceExpiration:1.0
 */
static void
decode_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_attributeNames;
    guint32   i_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_attributeNames;
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_authToken;
    guint32   i_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_agent);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_capability);

        u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_attributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_attributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_attributeNames);

        for (i_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_attributeNames=0; i_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_attributeNames < u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_attributeNames; i_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_attributeNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_attributeNames);

        }

        proto_tree_add_int64(tree, hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_expiresIn, tvb, *offset-8, 8, get_CDR_long_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_authToken);

        for (i_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_authToken=0; i_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_authToken < u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_authToken; i_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/access/IpPAMAgentPresence/getAgentPresence:1.0
 */
static void
decode_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_attributeNames;
    guint32   i_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_attributeNames;
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_authToken;
    guint32   i_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_authToken;
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_return;
    guint32   i_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_agent);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_agentType);

        u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_attributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_attributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_attributeNames);

        for (i_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_attributeNames=0; i_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_attributeNames < u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_attributeNames; i_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_attributeNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_attributeNames);

        }

        u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_authToken);

        for (i_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_authToken=0; i_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_authToken < u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_authToken; i_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_return);

            for (i_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_return=0; i_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_return < u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_return; i_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_return++) {

                /*  Begin struct "org_csapi_pam_TpPAMAttribute"  */
                decode_org_csapi_pam_TpPAMAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_pam_TpPAMAttribute"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/access/IpPAMAgentPresence/getCapabilityPresence:1.0
 */
static void
decode_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_attributeNames;
    guint32   i_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_attributeNames;
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_authToken;
    guint32   i_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_authToken;
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_return;
    guint32   i_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_agent);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_capability);

        u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_attributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_attributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_attributeNames);

        for (i_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_attributeNames=0; i_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_attributeNames < u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_attributeNames; i_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_attributeNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_attributeNames);

        }

        u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_authToken);

        for (i_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_authToken=0; i_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_authToken < u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_authToken; i_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_return);

            for (i_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_return=0; i_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_return < u_octet4_loop_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_return; i_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_return++) {

                /*  Begin struct "org_csapi_pam_TpPAMAttribute"  */
                decode_org_csapi_pam_TpPAMAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_pam_TpPAMAttribute"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/access/IpAppPAMPreferenceCheck/computeAvailability:1.0
 */
static void
decode_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_attributeNames;
    guint32   i_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_attributeNames;
    guint32   u_octet4_loop_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_authToken;
    guint32   i_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_authToken;
    guint32   u_octet4_loop_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_return;
    guint32   i_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_identity);

        /*  Begin struct "org_csapi_pam_TpPAMContext"  */
        decode_org_csapi_pam_TpPAMContext_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMContext"  */
        u_octet4_loop_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_attributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_attributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_attributeNames);

        for (i_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_attributeNames=0; i_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_attributeNames < u_octet4_loop_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_attributeNames; i_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_attributeNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_attributeNames);

        }

        u_octet4_loop_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_authToken);

        for (i_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_authToken=0; i_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_authToken < u_octet4_loop_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_authToken; i_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_return);

            for (i_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_return=0; i_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_return < u_octet4_loop_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_return; i_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_return++) {

                /*  Begin struct "org_csapi_pam_TpPAMAvailabilityProfile"  */
                decode_org_csapi_pam_TpPAMAvailabilityProfile_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_pam_TpPAMAvailabilityProfile"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/access/IpPAMAvailability/getAvailability:1.0
 */
static void
decode_org_csapi_pam_access_IpPAMAvailability_getAvailability(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMAvailability_getAvailability_attributeNames;
    guint32   i_org_csapi_pam_access_IpPAMAvailability_getAvailability_attributeNames;
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMAvailability_getAvailability_authToken;
    guint32   i_org_csapi_pam_access_IpPAMAvailability_getAvailability_authToken;
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMAvailability_getAvailability_return;
    guint32   i_org_csapi_pam_access_IpPAMAvailability_getAvailability_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMAvailability_getAvailability_identity);

        /*  Begin struct "org_csapi_pam_TpPAMContext"  */
        decode_org_csapi_pam_TpPAMContext_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMContext"  */
        u_octet4_loop_org_csapi_pam_access_IpPAMAvailability_getAvailability_attributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAvailability_getAvailability_attributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMAvailability_getAvailability_attributeNames);

        for (i_org_csapi_pam_access_IpPAMAvailability_getAvailability_attributeNames=0; i_org_csapi_pam_access_IpPAMAvailability_getAvailability_attributeNames < u_octet4_loop_org_csapi_pam_access_IpPAMAvailability_getAvailability_attributeNames; i_org_csapi_pam_access_IpPAMAvailability_getAvailability_attributeNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMAvailability_getAvailability_attributeNames);

        }

        u_octet4_loop_org_csapi_pam_access_IpPAMAvailability_getAvailability_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAvailability_getAvailability_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMAvailability_getAvailability_authToken);

        for (i_org_csapi_pam_access_IpPAMAvailability_getAvailability_authToken=0; i_org_csapi_pam_access_IpPAMAvailability_getAvailability_authToken < u_octet4_loop_org_csapi_pam_access_IpPAMAvailability_getAvailability_authToken; i_org_csapi_pam_access_IpPAMAvailability_getAvailability_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAvailability_getAvailability_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_access_IpPAMAvailability_getAvailability_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAvailability_getAvailability_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMAvailability_getAvailability_return);

            for (i_org_csapi_pam_access_IpPAMAvailability_getAvailability_return=0; i_org_csapi_pam_access_IpPAMAvailability_getAvailability_return < u_octet4_loop_org_csapi_pam_access_IpPAMAvailability_getAvailability_return; i_org_csapi_pam_access_IpPAMAvailability_getAvailability_return++) {

                /*  Begin struct "org_csapi_pam_TpPAMAvailabilityProfile"  */
                decode_org_csapi_pam_TpPAMAvailabilityProfile_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
                /*  End struct "org_csapi_pam_TpPAMAvailabilityProfile"  */
            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/access/IpPAMAvailability/getPreference:1.0
 */
static void
decode_org_csapi_pam_access_IpPAMAvailability_getPreference(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMAvailability_getPreference_authToken;
    guint32   i_org_csapi_pam_access_IpPAMAvailability_getPreference_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMAvailability_getPreference_identity);

        /*  Begin struct "org_csapi_pam_TpPAMContext"  */
        decode_org_csapi_pam_TpPAMContext_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMContext"  */
        u_octet4_loop_org_csapi_pam_access_IpPAMAvailability_getPreference_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAvailability_getPreference_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMAvailability_getPreference_authToken);

        for (i_org_csapi_pam_access_IpPAMAvailability_getPreference_authToken=0; i_org_csapi_pam_access_IpPAMAvailability_getPreference_authToken < u_octet4_loop_org_csapi_pam_access_IpPAMAvailability_getPreference_authToken; i_org_csapi_pam_access_IpPAMAvailability_getPreference_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAvailability_getPreference_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin union "org_csapi_pam_TpPAMPreferenceData"  */
            decode_org_csapi_pam_TpPAMPreferenceData_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_pam_TpPAMPreferenceData"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/access/IpPAMAvailability/setPreference:1.0
 */
static void
decode_org_csapi_pam_access_IpPAMAvailability_setPreference(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMAvailability_setPreference_authToken;
    guint32   i_org_csapi_pam_access_IpPAMAvailability_setPreference_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMAvailability_setPreference_identity);

        /*  Begin struct "org_csapi_pam_TpPAMContext"  */
        decode_org_csapi_pam_TpPAMContext_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMContext"  */
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMAvailability_setPreference_operation);

        /*  Begin union "org_csapi_pam_TpPAMPreferenceData"  */
        decode_org_csapi_pam_TpPAMPreferenceData_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_pam_TpPAMPreferenceData"  */
        u_octet4_loop_org_csapi_pam_access_IpPAMAvailability_setPreference_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAvailability_setPreference_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMAvailability_setPreference_authToken);

        for (i_org_csapi_pam_access_IpPAMAvailability_setPreference_authToken=0; i_org_csapi_pam_access_IpPAMAvailability_setPreference_authToken < u_octet4_loop_org_csapi_pam_access_IpPAMAvailability_setPreference_authToken; i_org_csapi_pam_access_IpPAMAvailability_setPreference_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMAvailability_setPreference_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/access/IpPAMPresenceAvailabilityManager/getAuthToken:1.0
 */
static void
decode_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_askerData;
    guint32   i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_askerData;
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_return;
    guint32   i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_askerData = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_askerData_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_askerData);

        for (i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_askerData=0; i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_askerData < u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_askerData; i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_askerData++) {

            /*  Begin struct "org_csapi_TpAttribute"  */
            decode_org_csapi_TpAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAttribute"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_return);

            for (i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_return=0; i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_return < u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_return; i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_return++) {

                proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_return, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/access/IpPAMPresenceAvailabilityManager/obtainInterface:1.0
 */
static void
decode_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_obtainInterface(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_obtainInterface_interfaceName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/access/IpPAMPresenceAvailabilityManager/getAccessControl:1.0
 */
static void
decode_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAccessControl(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAccessControl_authToken;
    guint32   i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAccessControl_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAccessControl_identity);

        u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAccessControl_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAccessControl_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAccessControl_authToken);

        for (i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAccessControl_authToken=0; i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAccessControl_authToken < u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAccessControl_authToken; i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAccessControl_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAccessControl_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_pam_TpPAMAccessControlData"  */
            decode_org_csapi_pam_TpPAMAccessControlData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_pam_TpPAMAccessControlData"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/access/IpPAMPresenceAvailabilityManager/setAccessControl:1.0
 */
static void
decode_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_setAccessControl(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_setAccessControl_authToken;
    guint32   i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_setAccessControl_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_setAccessControl_identity);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_setAccessControl_operation);

        /*  Begin struct "org_csapi_pam_TpPAMAccessControlData"  */
        decode_org_csapi_pam_TpPAMAccessControlData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMAccessControlData"  */
        u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_setAccessControl_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_setAccessControl_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_setAccessControl_authToken);

        for (i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_setAccessControl_authToken=0; i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_setAccessControl_authToken < u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_setAccessControl_authToken; i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_setAccessControl_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_setAccessControl_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/access/IpPAMPresenceAvailabilityManager/activateService:1.0
 */
static void
decode_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_activateService(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_activateService_authToken;
    guint32   i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_activateService_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_activateService_identity);

        u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_activateService_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_activateService_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_activateService_authToken);

        for (i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_activateService_authToken=0; i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_activateService_authToken < u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_activateService_authToken; i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_activateService_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_activateService_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/access/IpPAMPresenceAvailabilityManager/deactivateService:1.0
 */
static void
decode_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_deactivateService(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_deactivateService_authToken;
    guint32   i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_deactivateService_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_deactivateService_identity);

        u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_deactivateService_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_deactivateService_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_deactivateService_authToken);

        for (i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_deactivateService_authToken=0; i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_deactivateService_authToken < u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_deactivateService_authToken; i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_deactivateService_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_deactivateService_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/access/IpPAMPresenceAvailabilityManager/isActiveIdentity:1.0
 */
static void
decode_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_isActiveIdentity(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_isActiveIdentity_authToken;
    guint32   i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_isActiveIdentity_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_isActiveIdentity_identity);

        u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_isActiveIdentity_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_isActiveIdentity_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_isActiveIdentity_authToken);

        for (i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_isActiveIdentity_authToken=0; i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_isActiveIdentity_authToken < u_octet4_loop_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_isActiveIdentity_authToken; i_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_isActiveIdentity_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_isActiveIdentity_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_boolean(tree, hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_isActiveIdentity_return, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/event/IpAppPAMEventHandler/eventNotify:1.0
 */
static void
decode_org_csapi_pam_event_IpAppPAMEventHandler_eventNotify(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_event_IpAppPAMEventHandler_eventNotify_eventInfo;
    guint32   i_org_csapi_pam_event_IpAppPAMEventHandler_eventNotify_eventInfo;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_pam_event_IpAppPAMEventHandler_eventNotify_eventID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_pam_event_IpAppPAMEventHandler_eventNotify_eventInfo = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpAppPAMEventHandler_eventNotify_eventInfo_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_event_IpAppPAMEventHandler_eventNotify_eventInfo);

        for (i_org_csapi_pam_event_IpAppPAMEventHandler_eventNotify_eventInfo=0; i_org_csapi_pam_event_IpAppPAMEventHandler_eventNotify_eventInfo < u_octet4_loop_org_csapi_pam_event_IpAppPAMEventHandler_eventNotify_eventInfo; i_org_csapi_pam_event_IpAppPAMEventHandler_eventNotify_eventInfo++) {

            /*  Begin union "org_csapi_pam_TpPAMNotificationInfo"  */
            decode_org_csapi_pam_TpPAMNotificationInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_pam_TpPAMNotificationInfo"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/event/IpAppPAMEventHandler/eventNotifyErr:1.0
 */
static void
decode_org_csapi_pam_event_IpAppPAMEventHandler_eventNotifyErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_pam_event_IpAppPAMEventHandler_eventNotifyErr_eventID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_pam_TpPAMErrorInfo"  */
        decode_org_csapi_pam_TpPAMErrorInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMErrorInfo"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/event/IpPAMEventHandler/isRegistered:1.0
 */
static void
decode_org_csapi_pam_event_IpPAMEventHandler_isRegistered(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_isRegistered_authToken;
    guint32   i_org_csapi_pam_event_IpPAMEventHandler_isRegistered_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_pam_event_IpPAMEventHandler_isRegistered_clientID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_isRegistered_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventHandler_isRegistered_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_isRegistered_authToken);

        for (i_org_csapi_pam_event_IpPAMEventHandler_isRegistered_authToken=0; i_org_csapi_pam_event_IpPAMEventHandler_isRegistered_authToken < u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_isRegistered_authToken; i_org_csapi_pam_event_IpPAMEventHandler_isRegistered_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventHandler_isRegistered_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_boolean(tree, hf_org_csapi_pam_event_IpPAMEventHandler_isRegistered_return, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/event/IpPAMEventHandler/registerAppInterface:1.0
 */
static void
decode_org_csapi_pam_event_IpPAMEventHandler_registerAppInterface(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_registerAppInterface_authToken;
    guint32   i_org_csapi_pam_event_IpPAMEventHandler_registerAppInterface_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_registerAppInterface_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventHandler_registerAppInterface_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_registerAppInterface_authToken);

        for (i_org_csapi_pam_event_IpPAMEventHandler_registerAppInterface_authToken=0; i_org_csapi_pam_event_IpPAMEventHandler_registerAppInterface_authToken < u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_registerAppInterface_authToken; i_org_csapi_pam_event_IpPAMEventHandler_registerAppInterface_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventHandler_registerAppInterface_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_pam_event_IpPAMEventHandler_registerAppInterface_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/event/IpPAMEventHandler/registerForEvent:1.0
 */
static void
decode_org_csapi_pam_event_IpPAMEventHandler_registerForEvent(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_eventList;
    guint32   i_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_eventList;
    guint32   u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_authToken;
    guint32   i_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_clientID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_eventList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_eventList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_eventList);

        for (i_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_eventList=0; i_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_eventList < u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_eventList; i_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_eventList++) {

            /*  Begin union "org_csapi_pam_TpPAMEventInfo"  */
            decode_org_csapi_pam_TpPAMEventInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_pam_TpPAMEventInfo"  */
        }

        proto_tree_add_int(tree, hf_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_validFor, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_authToken);

        for (i_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_authToken=0; i_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_authToken < u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_authToken; i_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/event/IpPAMEventHandler/deregisterAppInterface:1.0
 */
static void
decode_org_csapi_pam_event_IpPAMEventHandler_deregisterAppInterface(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_deregisterAppInterface_authToken;
    guint32   i_org_csapi_pam_event_IpPAMEventHandler_deregisterAppInterface_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_pam_event_IpPAMEventHandler_deregisterAppInterface_clientID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_deregisterAppInterface_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventHandler_deregisterAppInterface_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_deregisterAppInterface_authToken);

        for (i_org_csapi_pam_event_IpPAMEventHandler_deregisterAppInterface_authToken=0; i_org_csapi_pam_event_IpPAMEventHandler_deregisterAppInterface_authToken < u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_deregisterAppInterface_authToken; i_org_csapi_pam_event_IpPAMEventHandler_deregisterAppInterface_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventHandler_deregisterAppInterface_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/event/IpPAMEventHandler/deregisterFromEvent:1.0
 */
static void
decode_org_csapi_pam_event_IpPAMEventHandler_deregisterFromEvent(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_deregisterFromEvent_authToken;
    guint32   i_org_csapi_pam_event_IpPAMEventHandler_deregisterFromEvent_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_pam_event_IpPAMEventHandler_deregisterFromEvent_eventID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_deregisterFromEvent_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventHandler_deregisterFromEvent_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_deregisterFromEvent_authToken);

        for (i_org_csapi_pam_event_IpPAMEventHandler_deregisterFromEvent_authToken=0; i_org_csapi_pam_event_IpPAMEventHandler_deregisterFromEvent_authToken < u_octet4_loop_org_csapi_pam_event_IpPAMEventHandler_deregisterFromEvent_authToken; i_org_csapi_pam_event_IpPAMEventHandler_deregisterFromEvent_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventHandler_deregisterFromEvent_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/event/IpPAMEventManager/getAuthToken:1.0
 */
static void
decode_org_csapi_pam_event_IpPAMEventManager_getAuthToken(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_getAuthToken_askerData;
    guint32   i_org_csapi_pam_event_IpPAMEventManager_getAuthToken_askerData;
    guint32   u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_getAuthToken_return;
    guint32   i_org_csapi_pam_event_IpPAMEventManager_getAuthToken_return;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_getAuthToken_askerData = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventManager_getAuthToken_askerData_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_getAuthToken_askerData);

        for (i_org_csapi_pam_event_IpPAMEventManager_getAuthToken_askerData=0; i_org_csapi_pam_event_IpPAMEventManager_getAuthToken_askerData < u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_getAuthToken_askerData; i_org_csapi_pam_event_IpPAMEventManager_getAuthToken_askerData++) {

            /*  Begin struct "org_csapi_TpAttribute"  */
            decode_org_csapi_TpAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAttribute"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_getAuthToken_return = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventManager_getAuthToken_return_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_getAuthToken_return);

            for (i_org_csapi_pam_event_IpPAMEventManager_getAuthToken_return=0; i_org_csapi_pam_event_IpPAMEventManager_getAuthToken_return < u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_getAuthToken_return; i_org_csapi_pam_event_IpPAMEventManager_getAuthToken_return++) {

                proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventManager_getAuthToken_return, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

            }

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/event/IpPAMEventManager/obtainInterface:1.0
 */
static void
decode_org_csapi_pam_event_IpPAMEventManager_obtainInterface(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_event_IpPAMEventManager_obtainInterface_interfaceName);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/event/IpPAMEventManager/getAccessControl:1.0
 */
static void
decode_org_csapi_pam_event_IpPAMEventManager_getAccessControl(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_getAccessControl_authToken;
    guint32   i_org_csapi_pam_event_IpPAMEventManager_getAccessControl_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_event_IpPAMEventManager_getAccessControl_identity);

        u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_getAccessControl_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventManager_getAccessControl_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_getAccessControl_authToken);

        for (i_org_csapi_pam_event_IpPAMEventManager_getAccessControl_authToken=0; i_org_csapi_pam_event_IpPAMEventManager_getAccessControl_authToken < u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_getAccessControl_authToken; i_org_csapi_pam_event_IpPAMEventManager_getAccessControl_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventManager_getAccessControl_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_pam_TpPAMAccessControlData"  */
            decode_org_csapi_pam_TpPAMAccessControlData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_pam_TpPAMAccessControlData"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/event/IpPAMEventManager/setAccessControl:1.0
 */
static void
decode_org_csapi_pam_event_IpPAMEventManager_setAccessControl(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_setAccessControl_authToken;
    guint32   i_org_csapi_pam_event_IpPAMEventManager_setAccessControl_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_event_IpPAMEventManager_setAccessControl_identity);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_event_IpPAMEventManager_setAccessControl_operation);

        /*  Begin struct "org_csapi_pam_TpPAMAccessControlData"  */
        decode_org_csapi_pam_TpPAMAccessControlData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMAccessControlData"  */
        u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_setAccessControl_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventManager_setAccessControl_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_setAccessControl_authToken);

        for (i_org_csapi_pam_event_IpPAMEventManager_setAccessControl_authToken=0; i_org_csapi_pam_event_IpPAMEventManager_setAccessControl_authToken < u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_setAccessControl_authToken; i_org_csapi_pam_event_IpPAMEventManager_setAccessControl_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventManager_setAccessControl_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/event/IpPAMEventManager/activateService:1.0
 */
static void
decode_org_csapi_pam_event_IpPAMEventManager_activateService(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_activateService_authToken;
    guint32   i_org_csapi_pam_event_IpPAMEventManager_activateService_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_event_IpPAMEventManager_activateService_identity);

        u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_activateService_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventManager_activateService_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_activateService_authToken);

        for (i_org_csapi_pam_event_IpPAMEventManager_activateService_authToken=0; i_org_csapi_pam_event_IpPAMEventManager_activateService_authToken < u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_activateService_authToken; i_org_csapi_pam_event_IpPAMEventManager_activateService_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventManager_activateService_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/event/IpPAMEventManager/deactivateService:1.0
 */
static void
decode_org_csapi_pam_event_IpPAMEventManager_deactivateService(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_deactivateService_authToken;
    guint32   i_org_csapi_pam_event_IpPAMEventManager_deactivateService_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_event_IpPAMEventManager_deactivateService_identity);

        u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_deactivateService_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventManager_deactivateService_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_deactivateService_authToken);

        for (i_org_csapi_pam_event_IpPAMEventManager_deactivateService_authToken=0; i_org_csapi_pam_event_IpPAMEventManager_deactivateService_authToken < u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_deactivateService_authToken; i_org_csapi_pam_event_IpPAMEventManager_deactivateService_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventManager_deactivateService_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/pam/event/IpPAMEventManager/isActiveIdentity:1.0
 */
static void
decode_org_csapi_pam_event_IpPAMEventManager_isActiveIdentity(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_isActiveIdentity_authToken;
    guint32   i_org_csapi_pam_event_IpPAMEventManager_isActiveIdentity_authToken;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_event_IpPAMEventManager_isActiveIdentity_identity);

        u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_isActiveIdentity_authToken = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventManager_isActiveIdentity_authToken_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_isActiveIdentity_authToken);

        for (i_org_csapi_pam_event_IpPAMEventManager_isActiveIdentity_authToken=0; i_org_csapi_pam_event_IpPAMEventManager_isActiveIdentity_authToken < u_octet4_loop_org_csapi_pam_event_IpPAMEventManager_isActiveIdentity_authToken; i_org_csapi_pam_event_IpPAMEventManager_isActiveIdentity_authToken++) {

            proto_tree_add_uint(tree, hf_org_csapi_pam_event_IpPAMEventManager_isActiveIdentity_authToken, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_boolean(tree, hf_org_csapi_pam_event_IpPAMEventManager_isActiveIdentity_return, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMultiMediaMessagingManager/mailboxTerminated:1.0
 */
static void
decode_org_csapi_mmm_IpAppMultiMediaMessagingManager_mailboxTerminated(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_mmm_TpMailboxIdentifier"  */
        decode_org_csapi_mmm_TpMailboxIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mmm_TpMailboxIdentifier"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMultiMediaMessagingManager/reportNotification:1.0
 */
static void
decode_org_csapi_mmm_IpAppMultiMediaMessagingManager_reportNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessagingManager_reportNotification_eventInfo;
    guint32   i_org_csapi_mmm_IpAppMultiMediaMessagingManager_reportNotification_eventInfo;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMultiMediaMessagingManager_reportNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessagingManager_reportNotification_eventInfo = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMultiMediaMessagingManager_reportNotification_eventInfo_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessagingManager_reportNotification_eventInfo);

        for (i_org_csapi_mmm_IpAppMultiMediaMessagingManager_reportNotification_eventInfo=0; i_org_csapi_mmm_IpAppMultiMediaMessagingManager_reportNotification_eventInfo < u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessagingManager_reportNotification_eventInfo; i_org_csapi_mmm_IpAppMultiMediaMessagingManager_reportNotification_eventInfo++) {

            /*  Begin union "org_csapi_mmm_TpMessagingEventInfo"  */
            decode_org_csapi_mmm_TpMessagingEventInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_mmm_TpMessagingEventInfo"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMultiMediaMessagingManager/notificationsInterrupted:1.0
 */
static void
decode_org_csapi_mmm_IpAppMultiMediaMessagingManager_notificationsInterrupted(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMultiMediaMessagingManager/notificationsResumed:1.0
 */
static void
decode_org_csapi_mmm_IpAppMultiMediaMessagingManager_notificationsResumed(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMultiMediaMessagingManager/multiMediaMessagingTerminated:1.0
 */
static void
decode_org_csapi_mmm_IpAppMultiMediaMessagingManager_multiMediaMessagingTerminated(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_mmm_TpMultiMediaMessagingIdentifier"  */
        decode_org_csapi_mmm_TpMultiMediaMessagingIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mmm_TpMultiMediaMessagingIdentifier"  */
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMultiMediaMessagingManager/terminateMultipleMailboxes:1.0
 */
static void
decode_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMailboxes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMailboxes_mailboxSet;
    guint32   i_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMailboxes_mailboxSet;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMailboxes_mailboxSet = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMailboxes_mailboxSet_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMailboxes_mailboxSet);

        for (i_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMailboxes_mailboxSet=0; i_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMailboxes_mailboxSet < u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMailboxes_mailboxSet; i_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMailboxes_mailboxSet++) {

            /*  Begin struct "org_csapi_mmm_TpMailboxIdentifier"  */
            decode_org_csapi_mmm_TpMailboxIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mmm_TpMailboxIdentifier"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMultiMediaMessagingManager/terminateMultipleMultiMediaMessagingSessions:1.0
 */
static void
decode_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMultiMediaMessagingSessions(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMultiMediaMessagingSessions_multiMediaMessagingSet;
    guint32   i_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMultiMediaMessagingSessions_multiMediaMessagingSet;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMultiMediaMessagingSessions_multiMediaMessagingSet = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMultiMediaMessagingSessions_multiMediaMessagingSet_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMultiMediaMessagingSessions_multiMediaMessagingSet);

        for (i_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMultiMediaMessagingSessions_multiMediaMessagingSet=0; i_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMultiMediaMessagingSessions_multiMediaMessagingSet < u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMultiMediaMessagingSessions_multiMediaMessagingSet; i_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMultiMediaMessagingSessions_multiMediaMessagingSet++) {

            /*  Begin struct "org_csapi_mmm_TpMultiMediaMessagingIdentifier"  */
            decode_org_csapi_mmm_TpMultiMediaMessagingIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mmm_TpMultiMediaMessagingIdentifier"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMultiMediaMessagingManager/openMailbox:1.0
 */
static void
decode_org_csapi_mmm_IpMultiMediaMessagingManager_openMailbox(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMultiMediaMessagingManager_openMailbox_mailboxID);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMultiMediaMessagingManager_openMailbox_authenticationInfo);

        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_mmm_TpMailboxIdentifier"  */
            decode_org_csapi_mmm_TpMailboxIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mmm_TpMailboxIdentifier"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMultiMediaMessagingManager/openMultiMediaMessaging:1.0
 */
static void
decode_org_csapi_mmm_IpMultiMediaMessagingManager_openMultiMediaMessaging(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        /*  Begin struct "org_csapi_mmm_TpTerminatingAddressList"  */
        decode_org_csapi_mmm_TpTerminatingAddressList_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mmm_TpTerminatingAddressList"  */
        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_mmm_TpMultiMediaMessagingIdentifier"  */
            decode_org_csapi_mmm_TpMultiMediaMessagingIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mmm_TpMultiMediaMessagingIdentifier"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMultiMediaMessagingManager/createNotification:1.0
 */
static void
decode_org_csapi_mmm_IpMultiMediaMessagingManager_createNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_IpMultiMediaMessagingManager_createNotification_eventCriteria;
    guint32   i_org_csapi_mmm_IpMultiMediaMessagingManager_createNotification_eventCriteria;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        u_octet4_loop_org_csapi_mmm_IpMultiMediaMessagingManager_createNotification_eventCriteria = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpMultiMediaMessagingManager_createNotification_eventCriteria_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpMultiMediaMessagingManager_createNotification_eventCriteria);

        for (i_org_csapi_mmm_IpMultiMediaMessagingManager_createNotification_eventCriteria=0; i_org_csapi_mmm_IpMultiMediaMessagingManager_createNotification_eventCriteria < u_octet4_loop_org_csapi_mmm_IpMultiMediaMessagingManager_createNotification_eventCriteria; i_org_csapi_mmm_IpMultiMediaMessagingManager_createNotification_eventCriteria++) {

            /*  Begin union "org_csapi_mmm_TpMessagingEventCriteria"  */
            decode_org_csapi_mmm_TpMessagingEventCriteria_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_mmm_TpMessagingEventCriteria"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mmm_IpMultiMediaMessagingManager_createNotification_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMultiMediaMessagingManager/destroyNotification:1.0
 */
static void
decode_org_csapi_mmm_IpMultiMediaMessagingManager_destroyNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMultiMediaMessagingManager_destroyNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMultiMediaMessagingManager/changeNotification:1.0
 */
static void
decode_org_csapi_mmm_IpMultiMediaMessagingManager_changeNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_IpMultiMediaMessagingManager_changeNotification_eventCriteria;
    guint32   i_org_csapi_mmm_IpMultiMediaMessagingManager_changeNotification_eventCriteria;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMultiMediaMessagingManager_changeNotification_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_mmm_IpMultiMediaMessagingManager_changeNotification_eventCriteria = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpMultiMediaMessagingManager_changeNotification_eventCriteria_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpMultiMediaMessagingManager_changeNotification_eventCriteria);

        for (i_org_csapi_mmm_IpMultiMediaMessagingManager_changeNotification_eventCriteria=0; i_org_csapi_mmm_IpMultiMediaMessagingManager_changeNotification_eventCriteria < u_octet4_loop_org_csapi_mmm_IpMultiMediaMessagingManager_changeNotification_eventCriteria; i_org_csapi_mmm_IpMultiMediaMessagingManager_changeNotification_eventCriteria++) {

            /*  Begin union "org_csapi_mmm_TpMessagingEventCriteria"  */
            decode_org_csapi_mmm_TpMessagingEventCriteria_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_mmm_TpMessagingEventCriteria"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMultiMediaMessagingManager/getNextNotification:1.0
 */
static void
decode_org_csapi_mmm_IpMultiMediaMessagingManager_getNextNotification(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_boolean(tree, hf_org_csapi_mmm_IpMultiMediaMessagingManager_getNextNotification_reset, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /*  Begin struct "org_csapi_mmm_TpMessagingNotificationRequestedSetEntry"  */
            decode_org_csapi_mmm_TpMessagingNotificationRequestedSetEntry_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mmm_TpMessagingNotificationRequestedSetEntry"  */
            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMultiMediaMessagingManager/enableNotifications:1.0
 */
static void
decode_org_csapi_mmm_IpMultiMediaMessagingManager_enableNotifications(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mmm_IpMultiMediaMessagingManager_enableNotifications_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMultiMediaMessagingManager/disableNotifications:1.0
 */
static void
decode_org_csapi_mmm_IpMultiMediaMessagingManager_disableNotifications(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/createFolderRes:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_createFolderRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_createFolderRes_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_createFolderRes_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_createFolderRes_folderID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/createFolderErr:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_createFolderErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_createFolderErr_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_createFolderErr_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_createFolderErr_error, tvb, *offset-4, 4, u_octet4);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_createFolderErr_errorDetails);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/getFoldersRes:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_getFoldersRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_IpAppMailbox_getFoldersRes_folderNames;
    guint32   i_org_csapi_mmm_IpAppMailbox_getFoldersRes_folderNames;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getFoldersRes_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getFoldersRes_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_getFoldersRes_folderID);

        u_octet4_loop_org_csapi_mmm_IpAppMailbox_getFoldersRes_folderNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_getFoldersRes_folderNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpAppMailbox_getFoldersRes_folderNames);

        for (i_org_csapi_mmm_IpAppMailbox_getFoldersRes_folderNames=0; i_org_csapi_mmm_IpAppMailbox_getFoldersRes_folderNames < u_octet4_loop_org_csapi_mmm_IpAppMailbox_getFoldersRes_folderNames; i_org_csapi_mmm_IpAppMailbox_getFoldersRes_folderNames++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_getFoldersRes_folderNames);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/getFoldersErr:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_getFoldersErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getFoldersErr_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getFoldersErr_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_getFoldersErr_folderID);

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_getFoldersErr_error, tvb, *offset-4, 4, u_octet4);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_getFoldersErr_errorDetails);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/deleteFolderRes:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_deleteFolderRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_deleteFolderRes_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_deleteFolderRes_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/deleteFolderErr:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_deleteFolderErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_deleteFolderErr_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_deleteFolderErr_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_deleteFolderErr_error, tvb, *offset-4, 4, u_octet4);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_deleteFolderErr_errorDetails);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/copyFolderRes:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_copyFolderRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_copyFolderRes_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_copyFolderRes_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/copyFolderErr:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_copyFolderErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_copyFolderErr_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_copyFolderErr_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_copyFolderErr_error, tvb, *offset-4, 4, u_octet4);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_copyFolderErr_errorDetails);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/moveFolderRes:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_moveFolderRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_moveFolderRes_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_moveFolderRes_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/moveFolderErr:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_moveFolderErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_moveFolderErr_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_moveFolderErr_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_moveFolderErr_error, tvb, *offset-4, 4, u_octet4);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_moveFolderErr_errorDetails);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/putMessageRes:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_putMessageRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_putMessageRes_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_putMessageRes_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_putMessageRes_messageID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/putMessageErr:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_putMessageErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_putMessageErr_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_putMessageErr_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_putMessageErr_error, tvb, *offset-4, 4, u_octet4);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_putMessageErr_errorDetails);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/copyMessageRes:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_copyMessageRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_copyMessageRes_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_copyMessageRes_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/copyMessageErr:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_copyMessageErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_copyMessageErr_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_copyMessageErr_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_copyMessageErr_error, tvb, *offset-4, 4, u_octet4);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_copyMessageErr_errorDetails);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/moveMessageRes:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_moveMessageRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_moveMessageRes_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_moveMessageRes_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/moveMessageErr:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_moveMessageErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_moveMessageErr_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_moveMessageErr_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_moveMessageErr_error, tvb, *offset-4, 4, u_octet4);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_moveMessageErr_errorDetails);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/deleteMessageRes:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_deleteMessageRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_deleteMessageRes_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_deleteMessageRes_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/deleteMessageErr:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_deleteMessageErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_deleteMessageErr_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_deleteMessageErr_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_deleteMessageErr_error, tvb, *offset-4, 4, u_octet4);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_deleteMessageErr_errorDetails);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/listMessagesRes:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_listMessagesRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_IpAppMailbox_listMessagesRes_messageList;
    guint32   i_org_csapi_mmm_IpAppMailbox_listMessagesRes_messageList;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_listMessagesRes_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_listMessagesRes_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_mmm_IpAppMailbox_listMessagesRes_messageList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_listMessagesRes_messageList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpAppMailbox_listMessagesRes_messageList);

        for (i_org_csapi_mmm_IpAppMailbox_listMessagesRes_messageList=0; i_org_csapi_mmm_IpAppMailbox_listMessagesRes_messageList < u_octet4_loop_org_csapi_mmm_IpAppMailbox_listMessagesRes_messageList; i_org_csapi_mmm_IpAppMailbox_listMessagesRes_messageList++) {

            /*  Begin struct "org_csapi_mmm_TpMessageDescription"  */
            decode_org_csapi_mmm_TpMessageDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mmm_TpMessageDescription"  */
        }

        /*  Begin struct "org_csapi_mmm_TpMailboxFolderStatusInformation"  */
        decode_org_csapi_mmm_TpMailboxFolderStatusInformation_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mmm_TpMailboxFolderStatusInformation"  */
        proto_tree_add_boolean(tree, hf_org_csapi_mmm_IpAppMailbox_listMessagesRes_final, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/listMessagesErr:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_listMessagesErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_listMessagesErr_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_listMessagesErr_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_listMessagesErr_error, tvb, *offset-4, 4, u_octet4);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_listMessagesErr_errorDetails);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/listMessageBodyPartsRes:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsRes_partsList;
    guint32   i_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsRes_partsList;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsRes_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsRes_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsRes_partsList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsRes_partsList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsRes_partsList);

        for (i_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsRes_partsList=0; i_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsRes_partsList < u_octet4_loop_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsRes_partsList; i_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsRes_partsList++) {

            /*  Begin struct "org_csapi_mmm_TpBodyPartDescription"  */
            decode_org_csapi_mmm_TpBodyPartDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mmm_TpBodyPartDescription"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/listMessageBodyPartsErr:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsErr_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsErr_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsErr_error, tvb, *offset-4, 4, u_octet4);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsErr_errorDetails);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/getMessageBodyPartsRes:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsRes_bodyParts;
    guint32   i_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsRes_bodyParts;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsRes_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsRes_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsRes_bodyParts = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsRes_bodyParts_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsRes_bodyParts);

        for (i_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsRes_bodyParts=0; i_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsRes_bodyParts < u_octet4_loop_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsRes_bodyParts; i_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsRes_bodyParts++) {

            /*  Begin struct "org_csapi_mmm_TpBodyPart"  */
            decode_org_csapi_mmm_TpBodyPart_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mmm_TpBodyPart"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/getMessageBodyPartsErr:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsErr_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsErr_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsErr_error, tvb, *offset-4, 4, u_octet4);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsErr_errorDetails);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/getMessageHeadersRes:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_getMessageHeadersRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_IpAppMailbox_getMessageHeadersRes_headers;
    guint32   i_org_csapi_mmm_IpAppMailbox_getMessageHeadersRes_headers;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageHeadersRes_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageHeadersRes_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_mmm_IpAppMailbox_getMessageHeadersRes_headers = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageHeadersRes_headers_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpAppMailbox_getMessageHeadersRes_headers);

        for (i_org_csapi_mmm_IpAppMailbox_getMessageHeadersRes_headers=0; i_org_csapi_mmm_IpAppMailbox_getMessageHeadersRes_headers < u_octet4_loop_org_csapi_mmm_IpAppMailbox_getMessageHeadersRes_headers; i_org_csapi_mmm_IpAppMailbox_getMessageHeadersRes_headers++) {

            /*  Begin union "org_csapi_mmm_TpMessageHeaderField"  */
            decode_org_csapi_mmm_TpMessageHeaderField_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_mmm_TpMessageHeaderField"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/getMessageHeadersErr:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_getMessageHeadersErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageHeadersErr_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageHeadersErr_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageHeadersErr_error, tvb, *offset-4, 4, u_octet4);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_getMessageHeadersErr_errorDetails);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/getMessageContentRes:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_getMessageContentRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_IpAppMailbox_getMessageContentRes_content;
    guint32   i_org_csapi_mmm_IpAppMailbox_getMessageContentRes_content;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageContentRes_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageContentRes_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_getMessageContentRes_contentType);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_getMessageContentRes_contentTransferEncoding);

        u_octet4_loop_org_csapi_mmm_IpAppMailbox_getMessageContentRes_content = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageContentRes_content_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpAppMailbox_getMessageContentRes_content);

        for (i_org_csapi_mmm_IpAppMailbox_getMessageContentRes_content=0; i_org_csapi_mmm_IpAppMailbox_getMessageContentRes_content < u_octet4_loop_org_csapi_mmm_IpAppMailbox_getMessageContentRes_content; i_org_csapi_mmm_IpAppMailbox_getMessageContentRes_content++) {

            proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageContentRes_content, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/getMessageContentErr:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_getMessageContentErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageContentErr_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageContentErr_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageContentErr_error, tvb, *offset-4, 4, u_octet4);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_getMessageContentErr_errorDetails);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/getFullMessageRes:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_getFullMessageRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_IpAppMailbox_getFullMessageRes_message;
    guint32   i_org_csapi_mmm_IpAppMailbox_getFullMessageRes_message;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getFullMessageRes_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getFullMessageRes_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_mmm_IpAppMailbox_getFullMessageRes_message = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_getFullMessageRes_message_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpAppMailbox_getFullMessageRes_message);

        for (i_org_csapi_mmm_IpAppMailbox_getFullMessageRes_message=0; i_org_csapi_mmm_IpAppMailbox_getFullMessageRes_message < u_octet4_loop_org_csapi_mmm_IpAppMailbox_getFullMessageRes_message; i_org_csapi_mmm_IpAppMailbox_getFullMessageRes_message++) {

            proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_getFullMessageRes_message, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/getFullMessageErr:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_getFullMessageErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getFullMessageErr_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getFullMessageErr_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_getFullMessageErr_error, tvb, *offset-4, 4, u_octet4);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_getFullMessageErr_errorDetails);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/getMailboxInfoPropertiesRes:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesRes_returnedProperties;
    guint32   i_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesRes_returnedProperties;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesRes_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesRes_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesRes_returnedProperties = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesRes_returnedProperties_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesRes_returnedProperties);

        for (i_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesRes_returnedProperties=0; i_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesRes_returnedProperties < u_octet4_loop_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesRes_returnedProperties; i_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesRes_returnedProperties++) {

            /*  Begin union "org_csapi_mmm_TpMailboxInfoProperty"  */
            decode_org_csapi_mmm_TpMailboxInfoProperty_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_mmm_TpMailboxInfoProperty"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/getFolderInfoPropertiesRes:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesRes_returnedProperties;
    guint32   i_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesRes_returnedProperties;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesRes_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesRes_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesRes_folderID);

        u_octet4_loop_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesRes_returnedProperties = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesRes_returnedProperties_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesRes_returnedProperties);

        for (i_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesRes_returnedProperties=0; i_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesRes_returnedProperties < u_octet4_loop_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesRes_returnedProperties; i_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesRes_returnedProperties++) {

            /*  Begin union "org_csapi_mmm_TpFolderInfoProperty"  */
            decode_org_csapi_mmm_TpFolderInfoProperty_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_mmm_TpFolderInfoProperty"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/getMessageInfoPropertiesRes:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesRes_returnedProperties;
    guint32   i_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesRes_returnedProperties;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesRes_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesRes_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesRes_messageID);

        u_octet4_loop_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesRes_returnedProperties = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesRes_returnedProperties_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesRes_returnedProperties);

        for (i_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesRes_returnedProperties=0; i_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesRes_returnedProperties < u_octet4_loop_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesRes_returnedProperties; i_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesRes_returnedProperties++) {

            /*  Begin union "org_csapi_mmm_TpMessageInfoProperty"  */
            decode_org_csapi_mmm_TpMessageInfoProperty_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_mmm_TpMessageInfoProperty"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/setMessageInfoPropertiesRes:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesRes_propertiesUpdated;
    guint32   i_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesRes_propertiesUpdated;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesRes_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesRes_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesRes_messageID);

        u_octet4_loop_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesRes_propertiesUpdated = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesRes_propertiesUpdated_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesRes_propertiesUpdated);

        for (i_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesRes_propertiesUpdated=0; i_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesRes_propertiesUpdated < u_octet4_loop_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesRes_propertiesUpdated; i_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesRes_propertiesUpdated++) {

            /*  Begin union "org_csapi_mmm_TpMessageInfoProperty"  */
            decode_org_csapi_mmm_TpMessageInfoProperty_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_mmm_TpMessageInfoProperty"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/setMessageInfoPropertiesErr:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesErr_propertiesNotUpdated;
    guint32   i_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesErr_propertiesNotUpdated;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesErr_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesErr_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesErr_messageID);

        u_octet4_loop_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesErr_propertiesNotUpdated = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesErr_propertiesNotUpdated_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesErr_propertiesNotUpdated);

        for (i_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesErr_propertiesNotUpdated=0; i_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesErr_propertiesNotUpdated < u_octet4_loop_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesErr_propertiesNotUpdated; i_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesErr_propertiesNotUpdated++) {

            /*  Begin struct "org_csapi_mmm_TpMessageInfoPropertyError"  */
            decode_org_csapi_mmm_TpMessageInfoPropertyError_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mmm_TpMessageInfoPropertyError"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/getMailboxInfoPropertiesErr:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesErr_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesErr_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesErr_error, tvb, *offset-4, 4, u_octet4);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesErr_errorDetails);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/getFolderInfoPropertiesErr:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesErr_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesErr_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesErr_error, tvb, *offset-4, 4, u_octet4);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesErr_errorDetails);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMailbox/getMessageInfoPropertiesErr:1.0
 */
static void
decode_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesErr_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesErr_requestID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesErr_error, tvb, *offset-4, 4, u_octet4);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesErr_errorDetails);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMailbox/close:1.0
 */
static void
decode_org_csapi_mmm_IpMailbox_close(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_close_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMailbox/createFolderReq:1.0
 */
static void
decode_org_csapi_mmm_IpMailbox_createFolderReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_createFolderReq_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_createFolderReq_folderID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_createFolderReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMailbox/getFoldersReq:1.0
 */
static void
decode_org_csapi_mmm_IpMailbox_getFoldersReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_getFoldersReq_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_getFoldersReq_folderID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_getFoldersReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMailbox/deleteFolderReq:1.0
 */
static void
decode_org_csapi_mmm_IpMailbox_deleteFolderReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_deleteFolderReq_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_deleteFolderReq_folderID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_deleteFolderReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMailbox/copyFolderReq:1.0
 */
static void
decode_org_csapi_mmm_IpMailbox_copyFolderReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_copyFolderReq_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_copyFolderReq_sourceFolderID);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_copyFolderReq_destinationFolderID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_copyFolderReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMailbox/moveFolderReq:1.0
 */
static void
decode_org_csapi_mmm_IpMailbox_moveFolderReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_moveFolderReq_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_moveFolderReq_sourceFolderID);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_moveFolderReq_destinationFolderID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_moveFolderReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMailbox/putMessageReq:1.0
 */
static void
decode_org_csapi_mmm_IpMailbox_putMessageReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_IpMailbox_putMessageReq_message;
    guint32   i_org_csapi_mmm_IpMailbox_putMessageReq_message;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_putMessageReq_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_putMessageReq_folderID);

        u_octet4_loop_org_csapi_mmm_IpMailbox_putMessageReq_message = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpMailbox_putMessageReq_message_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpMailbox_putMessageReq_message);

        for (i_org_csapi_mmm_IpMailbox_putMessageReq_message=0; i_org_csapi_mmm_IpMailbox_putMessageReq_message < u_octet4_loop_org_csapi_mmm_IpMailbox_putMessageReq_message; i_org_csapi_mmm_IpMailbox_putMessageReq_message++) {

            proto_tree_add_uint(tree, hf_org_csapi_mmm_IpMailbox_putMessageReq_message, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_putMessageReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMailbox/copyMessageReq:1.0
 */
static void
decode_org_csapi_mmm_IpMailbox_copyMessageReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_copyMessageReq_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_copyMessageReq_fromFolderID);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_copyMessageReq_toFolderID);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_copyMessageReq_messageID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_copyMessageReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMailbox/moveMessageReq:1.0
 */
static void
decode_org_csapi_mmm_IpMailbox_moveMessageReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_moveMessageReq_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_moveMessageReq_fromFolderID);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_moveMessageReq_toFolderID);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_moveMessageReq_messageID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_moveMessageReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMailbox/deleteMessageReq:1.0
 */
static void
decode_org_csapi_mmm_IpMailbox_deleteMessageReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_deleteMessageReq_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_deleteMessageReq_fromFolderID);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_deleteMessageReq_messageID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_deleteMessageReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMailbox/listMessagesReq:1.0
 */
static void
decode_org_csapi_mmm_IpMailbox_listMessagesReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_listMessagesReq_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_listMessagesReq_folderID);

        /*  Begin struct "org_csapi_mmm_TpListMessagesCriteria"  */
        decode_org_csapi_mmm_TpListMessagesCriteria_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mmm_TpListMessagesCriteria"  */
        proto_tree_add_boolean(tree, hf_org_csapi_mmm_IpMailbox_listMessagesReq_reset, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_listMessagesReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMailbox/listMessageBodyPartsReq:1.0
 */
static void
decode_org_csapi_mmm_IpMailbox_listMessageBodyPartsReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_listMessageBodyPartsReq_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_listMessageBodyPartsReq_folderID);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_listMessageBodyPartsReq_messageID);

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_listMessageBodyPartsReq_maxNestingLevel, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_listMessageBodyPartsReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMailbox/getMessageBodyPartsReq:1.0
 */
static void
decode_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_partIDs;
    guint32   i_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_partIDs;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_folderID);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_messageID);

        u_octet4_loop_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_partIDs = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_partIDs_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_partIDs);

        for (i_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_partIDs=0; i_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_partIDs < u_octet4_loop_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_partIDs; i_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_partIDs++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_partIDs);

        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMailbox/getMessageHeadersReq:1.0
 */
static void
decode_org_csapi_mmm_IpMailbox_getMessageHeadersReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_getMessageHeadersReq_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_getMessageHeadersReq_folderID);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_getMessageHeadersReq_messageID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_getMessageHeadersReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMailbox/getMessageContentReq:1.0
 */
static void
decode_org_csapi_mmm_IpMailbox_getMessageContentReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_getMessageContentReq_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_getMessageContentReq_folderID);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_getMessageContentReq_messageID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_getMessageContentReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMailbox/getFullMessageReq:1.0
 */
static void
decode_org_csapi_mmm_IpMailbox_getFullMessageReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_getFullMessageReq_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_getFullMessageReq_folderID);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_getFullMessageReq_messageID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_getFullMessageReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMailbox/getMailboxInfoPropertiesReq:1.0
 */
static void
decode_org_csapi_mmm_IpMailbox_getMailboxInfoPropertiesReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_getMailboxInfoPropertiesReq_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_getMailboxInfoPropertiesReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMailbox/getFolderInfoPropertiesReq:1.0
 */
static void
decode_org_csapi_mmm_IpMailbox_getFolderInfoPropertiesReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_getFolderInfoPropertiesReq_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_getFolderInfoPropertiesReq_folderID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_getFolderInfoPropertiesReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMailbox/getMessageInfoPropertiesReq:1.0
 */
static void
decode_org_csapi_mmm_IpMailbox_getMessageInfoPropertiesReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_getMessageInfoPropertiesReq_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_getMessageInfoPropertiesReq_messageID);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_getMessageInfoPropertiesReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMailbox/setMessageInfoPropertiesReq:1.0
 */
static void
decode_org_csapi_mmm_IpMailbox_setMessageInfoPropertiesReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_IpMailbox_setMessageInfoPropertiesReq_properties;
    guint32   i_org_csapi_mmm_IpMailbox_setMessageInfoPropertiesReq_properties;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_setMessageInfoPropertiesReq_mailboxSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMailbox_setMessageInfoPropertiesReq_messageID);

        u_octet4_loop_org_csapi_mmm_IpMailbox_setMessageInfoPropertiesReq_properties = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpMailbox_setMessageInfoPropertiesReq_properties_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpMailbox_setMessageInfoPropertiesReq_properties);

        for (i_org_csapi_mmm_IpMailbox_setMessageInfoPropertiesReq_properties=0; i_org_csapi_mmm_IpMailbox_setMessageInfoPropertiesReq_properties < u_octet4_loop_org_csapi_mmm_IpMailbox_setMessageInfoPropertiesReq_properties; i_org_csapi_mmm_IpMailbox_setMessageInfoPropertiesReq_properties++) {

            /*  Begin union "org_csapi_mmm_TpMessageInfoProperty"  */
            decode_org_csapi_mmm_TpMessageInfoProperty_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_mmm_TpMessageInfoProperty"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mmm_IpMailbox_setMessageInfoPropertiesReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMultiMediaMessaging/sendMessageRes:1.0
 */
static void
decode_org_csapi_mmm_IpAppMultiMediaMessaging_sendMessageRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMultiMediaMessaging_sendMessageRes_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMultiMediaMessaging_sendMessageRes_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMultiMediaMessaging/sendMessageErr:1.0
 */
static void
decode_org_csapi_mmm_IpAppMultiMediaMessaging_sendMessageErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMultiMediaMessaging_sendMessageErr_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMultiMediaMessaging_sendMessageErr_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMultiMediaMessaging_sendMessageErr_error, tvb, *offset-4, 4, u_octet4);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMultiMediaMessaging_sendMessageErr_errorDetails);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMultiMediaMessaging/cancelMessageRes:1.0
 */
static void
decode_org_csapi_mmm_IpAppMultiMediaMessaging_cancelMessageRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMultiMediaMessaging_cancelMessageRes_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMultiMediaMessaging_cancelMessageRes_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMultiMediaMessaging/cancelMessageErr:1.0
 */
static void
decode_org_csapi_mmm_IpAppMultiMediaMessaging_cancelMessageErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMultiMediaMessaging_cancelMessageErr_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMultiMediaMessaging_cancelMessageErr_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMultiMediaMessaging_cancelMessageErr_error, tvb, *offset-4, 4, u_octet4);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMultiMediaMessaging_cancelMessageErr_errorDetails);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMultiMediaMessaging/queryStatusRes:1.0
 */
static void
decode_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusRes(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusRes_result;
    guint32   i_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusRes_result;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusRes_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusRes_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusRes_result = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusRes_result_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusRes_result);

        for (i_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusRes_result=0; i_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusRes_result < u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusRes_result; i_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusRes_result++) {

            /*  Begin struct "org_csapi_mmm_TpQueryStatusReport"  */
            decode_org_csapi_mmm_TpQueryStatusReport_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_mmm_TpQueryStatusReport"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMultiMediaMessaging/queryStatusErr:1.0
 */
static void
decode_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusErr(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusErr_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusErr_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusErr_error, tvb, *offset-4, 4, u_octet4);

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusErr_errorDetails);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMultiMediaMessaging/messageStatusReport:1.0
 */
static void
decode_org_csapi_mmm_IpAppMultiMediaMessaging_messageStatusReport(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageStatusReport_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageStatusReport_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageStatusReport_deliveryReportType, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageStatusReport_deliveryReportInfo);

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpAppMultiMediaMessaging/messageReceived:1.0
 */
static void
decode_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_message;
    guint32   i_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_message;
    guint32   u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_headers;
    guint32   i_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_headers;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_message = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_message_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_message);

        for (i_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_message=0; i_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_message < u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_message; i_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_message++) {

            proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_message, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_headers = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_headers_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_headers);

        for (i_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_headers=0; i_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_headers < u_octet4_loop_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_headers; i_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_headers++) {

            /*  Begin union "org_csapi_mmm_TpMessageHeaderField"  */
            decode_org_csapi_mmm_TpMessageHeaderField_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_mmm_TpMessageHeaderField"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMultiMediaMessaging/sendMessageReq:1.0
 */
static void
decode_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_messageTreatment;
    guint32   i_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_messageTreatment;
    guint32   u_octet4_loop_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_message;
    guint32   i_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_message;
    guint32   u_octet4_loop_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_additionalHeaders;
    guint32   i_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_additionalHeaders;
    /* Operation specific Variable declarations End */

    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        /*  Begin struct "org_csapi_mmm_TpTerminatingAddressList"  */
        decode_org_csapi_mmm_TpTerminatingAddressList_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mmm_TpTerminatingAddressList"  */
        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_deliveryType);

        u_octet4_loop_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_messageTreatment = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_messageTreatment_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_messageTreatment);

        for (i_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_messageTreatment=0; i_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_messageTreatment < u_octet4_loop_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_messageTreatment; i_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_messageTreatment++) {

            /*  Begin union "org_csapi_mmm_TpMessageTreatment"  */
            decode_org_csapi_mmm_TpMessageTreatment_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_mmm_TpMessageTreatment"  */
        }

        u_octet4_loop_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_message = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_message_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_message);

        for (i_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_message=0; i_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_message < u_octet4_loop_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_message; i_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_message++) {

            proto_tree_add_uint(tree, hf_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_message, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        u_octet4_loop_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_additionalHeaders = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_additionalHeaders_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_additionalHeaders);

        for (i_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_additionalHeaders=0; i_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_additionalHeaders < u_octet4_loop_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_additionalHeaders; i_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_additionalHeaders++) {

            /*  Begin union "org_csapi_mmm_TpMessageHeaderField"  */
            decode_org_csapi_mmm_TpMessageHeaderField_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

            /*  End union "org_csapi_mmm_TpMessageHeaderField"  */
        }

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            proto_tree_add_int(tree, hf_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_return, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMultiMediaMessaging/cancelMessageReq:1.0
 */
static void
decode_org_csapi_mmm_IpMultiMediaMessaging_cancelMessageReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMultiMediaMessaging_cancelMessageReq_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMultiMediaMessaging_cancelMessageReq_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMultiMediaMessaging/queryStatusReq:1.0
 */
static void
decode_org_csapi_mmm_IpMultiMediaMessaging_queryStatusReq(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMultiMediaMessaging_queryStatusReq_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMultiMediaMessaging_queryStatusReq_assignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/*
 * IDL:org/csapi/mmm/IpMultiMediaMessaging/close:1.0
 */
static void
decode_org_csapi_mmm_IpMultiMediaMessaging_close(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    switch(header->message_type) {
    case Request:
        proto_tree_add_int(tree, hf_org_csapi_mmm_IpMultiMediaMessaging_close_sessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        break;
    case Reply:
        switch(header->rep_status) {
        case NO_EXCEPTION:
            /* Function returns void */

            break;
        case USER_EXCEPTION:
            break;
        default:
            /* Unknown Exception */
            expert_add_info_format(pinfo, item, &ei_parlay_unknown_exception, "Unknown exception %d", header->rep_status);
            break;
        }   /* switch(header->rep_status) */

        break;
    default:
        /* Unknown GIOP Message */
        expert_add_info_format(pinfo, item, &ei_parlay_unknown_giop_msg, "Unknown GIOP message %d", header->message_type);
        break;
    } /* switch(header->message_type) */
}

/* Struct = IDL:org/csapi/TpAddress:1.0 */
static void
decode_org_csapi_TpAddress_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_TpAddress_Plan, tvb, *offset-4, 4, u_octet4);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_TpAddress_AddrString);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_TpAddress_Name);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_TpAddress_Presentation, tvb, *offset-4, 4, u_octet4);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_TpAddress_Screening, tvb, *offset-4, 4, u_octet4);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_TpAddress_SubAddressString);

}

/* Struct = IDL:org/csapi/TpAddressRange:1.0 */
static void
decode_org_csapi_TpAddressRange_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_TpAddressRange_Plan, tvb, *offset-4, 4, u_octet4);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_TpAddressRange_AddrString);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_TpAddressRange_Name);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_TpAddressRange_SubAddressString);

}

/* Struct = IDL:org/csapi/TpCAIElements:1.0 */
static void
decode_org_csapi_TpCAIElements_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    proto_tree_add_int(tree, hf_org_csapi_TpCAIElements_UnitsPerInterval, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_TpCAIElements_SecondsPerTimeInterval, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_TpCAIElements_ScalingFactor, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_TpCAIElements_UnitIncrement, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_TpCAIElements_UnitsPerDataInterval, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_TpCAIElements_SegmentsPerDataInterval, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_TpCAIElements_InitialSecsPerTimeInterval, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/TpChargeAdviceInfo:1.0 */
static void
decode_org_csapi_TpChargeAdviceInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_TpCAIElements"  */
    decode_org_csapi_TpCAIElements_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpCAIElements"  */
    /*  Begin struct "org_csapi_TpCAIElements"  */
    decode_org_csapi_TpCAIElements_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpCAIElements"  */
}

/* Struct = IDL:org/csapi/TpChargePerTime:1.0 */
static void
decode_org_csapi_TpChargePerTime_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    proto_tree_add_int(tree, hf_org_csapi_TpChargePerTime_InitialCharge, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_TpChargePerTime_CurrentChargePerMinute, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_TpChargePerTime_NextChargePerMinute, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/TpAoCInfo:1.0 */
static void
decode_org_csapi_TpAoCInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin union "org_csapi_TpAoCOrder"  */
    decode_org_csapi_TpAoCOrder_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

    /*  End union "org_csapi_TpAoCOrder"  */
    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_TpAoCInfo_Currency);

}

/* Struct = IDL:org/csapi/TpTimeInterval:1.0 */
static void
decode_org_csapi_TpTimeInterval_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_TpTimeInterval_StartTime);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_TpTimeInterval_StopTime);

}

/* Struct = IDL:org/csapi/TpStructuredAttributeValue:1.0 */
static void
decode_org_csapi_TpStructuredAttributeValue_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_TpStructuredAttributeValue_Type);

    get_CDR_any(tvb, pinfo, tree, item, offset, stream_is_big_endian, boundary, header);

}

/* Struct = IDL:org/csapi/TpAttribute:1.0 */
static void
decode_org_csapi_TpAttribute_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_TpAttribute_AttributeName);

    /*  Begin union "org_csapi_TpAttributeValue"  */
    decode_org_csapi_TpAttributeValue_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

    /*  End union "org_csapi_TpAttributeValue"  */
}

/* Struct = IDL:org/csapi/fw/TpFaultStats:1.0 */
static void
decode_org_csapi_fw_TpFaultStats_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpFaultStats_Fault, tvb, *offset-4, 4, u_octet4);

    proto_tree_add_int(tree, hf_org_csapi_fw_TpFaultStats_Occurrences, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_fw_TpFaultStats_MaxDuration, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_fw_TpFaultStats_TotalDuration, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_fw_TpFaultStats_NumberOfClientsAffected, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/fw/TpFaultStatsRecord:1.0 */
static void
decode_org_csapi_fw_TpFaultStatsRecord_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_TpFaultStatsRecord_FaultStatsSet;
    guint32   i_org_csapi_fw_TpFaultStatsRecord_FaultStatsSet;
    /* Operation specific Variable declarations End */

    /*  Begin struct "org_csapi_TpTimeInterval"  */
    decode_org_csapi_TpTimeInterval_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpTimeInterval"  */
    u_octet4_loop_org_csapi_fw_TpFaultStatsRecord_FaultStatsSet = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpFaultStatsRecord_FaultStatsSet_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_TpFaultStatsRecord_FaultStatsSet);

    for (i_org_csapi_fw_TpFaultStatsRecord_FaultStatsSet=0; i_org_csapi_fw_TpFaultStatsRecord_FaultStatsSet < u_octet4_loop_org_csapi_fw_TpFaultStatsRecord_FaultStatsSet; i_org_csapi_fw_TpFaultStatsRecord_FaultStatsSet++) {

        /*  Begin struct "org_csapi_fw_TpFaultStats"  */
        decode_org_csapi_fw_TpFaultStats_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpFaultStats"  */
    }

}

/* Struct = IDL:org/csapi/fw/TpLoadPolicy:1.0 */
static void
decode_org_csapi_fw_TpLoadPolicy_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpLoadPolicy_LoadPolicy);

}

/* Struct = IDL:org/csapi/fw/TpLoadStatisticData:1.0 */
static void
decode_org_csapi_fw_TpLoadStatisticData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    proto_tree_add_float(tree, hf_org_csapi_fw_TpLoadStatisticData_LoadValue, tvb, *offset-4, 4, get_CDR_float(tvb,offset,stream_is_big_endian, boundary));

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpLoadStatisticData_LoadLevel, tvb, *offset-4, 4, u_octet4);

}

/* Struct = IDL:org/csapi/fw/TpLoadThreshold:1.0 */
static void
decode_org_csapi_fw_TpLoadThreshold_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    proto_tree_add_float(tree, hf_org_csapi_fw_TpLoadThreshold_LoadThreshold, tvb, *offset-4, 4, get_CDR_float(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/fw/TpLoadInitVal:1.0 */
static void
decode_org_csapi_fw_TpLoadInitVal_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpLoadInitVal_LoadLevel, tvb, *offset-4, 4, u_octet4);

    /*  Begin struct "org_csapi_fw_TpLoadThreshold"  */
    decode_org_csapi_fw_TpLoadThreshold_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_fw_TpLoadThreshold"  */
}

/* Struct = IDL:org/csapi/fw/TpProperty:1.0 */
static void
decode_org_csapi_fw_TpProperty_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpProperty_PropertyName);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpProperty_PropertyValue);

}

/* Struct = IDL:org/csapi/fw/TpClientAppDescription:1.0 */
static void
decode_org_csapi_fw_TpClientAppDescription_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_TpClientAppDescription_ClientAppProperties;
    guint32   i_org_csapi_fw_TpClientAppDescription_ClientAppProperties;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpClientAppDescription_ClientAppID);

    u_octet4_loop_org_csapi_fw_TpClientAppDescription_ClientAppProperties = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpClientAppDescription_ClientAppProperties_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_TpClientAppDescription_ClientAppProperties);

    for (i_org_csapi_fw_TpClientAppDescription_ClientAppProperties=0; i_org_csapi_fw_TpClientAppDescription_ClientAppProperties < u_octet4_loop_org_csapi_fw_TpClientAppDescription_ClientAppProperties; i_org_csapi_fw_TpClientAppDescription_ClientAppProperties++) {

        /*  Begin struct "org_csapi_fw_TpProperty"  */
        decode_org_csapi_fw_TpProperty_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpProperty"  */
    }

    proto_tree_add_boolean(tree, hf_org_csapi_fw_TpClientAppDescription_HasAccessSession, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    proto_tree_add_boolean(tree, hf_org_csapi_fw_TpClientAppDescription_HasServiceInstances, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

}

/* Struct = IDL:org/csapi/fw/TpEntOp:1.0 */
static void
decode_org_csapi_fw_TpEntOp_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_TpEntOp_EntOpProperties;
    guint32   i_org_csapi_fw_TpEntOp_EntOpProperties;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpEntOp_EntOpID);

    u_octet4_loop_org_csapi_fw_TpEntOp_EntOpProperties = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpEntOp_EntOpProperties_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_TpEntOp_EntOpProperties);

    for (i_org_csapi_fw_TpEntOp_EntOpProperties=0; i_org_csapi_fw_TpEntOp_EntOpProperties < u_octet4_loop_org_csapi_fw_TpEntOp_EntOpProperties; i_org_csapi_fw_TpEntOp_EntOpProperties++) {

        /*  Begin struct "org_csapi_fw_TpProperty"  */
        decode_org_csapi_fw_TpProperty_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpProperty"  */
    }

}

/* Struct = IDL:org/csapi/fw/TpSag:1.0 */
static void
decode_org_csapi_fw_TpSag_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpSag_SagID);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpSag_SagDescription);

}

/* Struct = IDL:org/csapi/fw/TpServiceProperty:1.0 */
static void
decode_org_csapi_fw_TpServiceProperty_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_TpServiceProperty_ServicePropertyValueList;
    guint32   i_org_csapi_fw_TpServiceProperty_ServicePropertyValueList;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpServiceProperty_ServicePropertyName);

    u_octet4_loop_org_csapi_fw_TpServiceProperty_ServicePropertyValueList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpServiceProperty_ServicePropertyValueList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_TpServiceProperty_ServicePropertyValueList);

    for (i_org_csapi_fw_TpServiceProperty_ServicePropertyValueList=0; i_org_csapi_fw_TpServiceProperty_ServicePropertyValueList < u_octet4_loop_org_csapi_fw_TpServiceProperty_ServicePropertyValueList; i_org_csapi_fw_TpServiceProperty_ServicePropertyValueList++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpServiceProperty_ServicePropertyValueList);

    }

}

/* Struct = IDL:org/csapi/fw/TpServiceDescription:1.0 */
static void
decode_org_csapi_fw_TpServiceDescription_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_TpServiceDescription_ServicePropertyList;
    guint32   i_org_csapi_fw_TpServiceDescription_ServicePropertyList;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpServiceDescription_ServiceTypeName);

    u_octet4_loop_org_csapi_fw_TpServiceDescription_ServicePropertyList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpServiceDescription_ServicePropertyList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_TpServiceDescription_ServicePropertyList);

    for (i_org_csapi_fw_TpServiceDescription_ServicePropertyList=0; i_org_csapi_fw_TpServiceDescription_ServicePropertyList < u_octet4_loop_org_csapi_fw_TpServiceDescription_ServicePropertyList; i_org_csapi_fw_TpServiceDescription_ServicePropertyList++) {

        /*  Begin struct "org_csapi_fw_TpServiceProperty"  */
        decode_org_csapi_fw_TpServiceProperty_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpServiceProperty"  */
    }

}

/* Struct = IDL:org/csapi/fw/TpService:1.0 */
static void
decode_org_csapi_fw_TpService_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpService_ServiceID);

    /*  Begin struct "org_csapi_fw_TpServiceDescription"  */
    decode_org_csapi_fw_TpServiceDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_fw_TpServiceDescription"  */
}

/* Struct = IDL:org/csapi/fw/TpServiceProfileDescription:1.0 */
static void
decode_org_csapi_fw_TpServiceProfileDescription_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_TpServiceProfileDescription_ServiceSubscriptionProperties;
    guint32   i_org_csapi_fw_TpServiceProfileDescription_ServiceSubscriptionProperties;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpServiceProfileDescription_ServiceContractID);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpServiceProfileDescription_ServiceStartDate);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpServiceProfileDescription_ServiceEndDate);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpServiceProfileDescription_ServiceTypeName);

    u_octet4_loop_org_csapi_fw_TpServiceProfileDescription_ServiceSubscriptionProperties = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpServiceProfileDescription_ServiceSubscriptionProperties_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_TpServiceProfileDescription_ServiceSubscriptionProperties);

    for (i_org_csapi_fw_TpServiceProfileDescription_ServiceSubscriptionProperties=0; i_org_csapi_fw_TpServiceProfileDescription_ServiceSubscriptionProperties < u_octet4_loop_org_csapi_fw_TpServiceProfileDescription_ServiceSubscriptionProperties; i_org_csapi_fw_TpServiceProfileDescription_ServiceSubscriptionProperties++) {

        /*  Begin struct "org_csapi_fw_TpServiceProperty"  */
        decode_org_csapi_fw_TpServiceProperty_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpServiceProperty"  */
    }

    proto_tree_add_boolean(tree, hf_org_csapi_fw_TpServiceProfileDescription_InUse, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpServiceProfileDescription_ServiceID);

}

/* Struct = IDL:org/csapi/fw/TpServiceTypeProperty:1.0 */
static void
decode_org_csapi_fw_TpServiceTypeProperty_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpServiceTypeProperty_ServicePropertyName);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpServiceTypeProperty_ServiceTypePropertyMode, tvb, *offset-4, 4, u_octet4);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpServiceTypeProperty_ServicePropertyTypeName);

}

/* Struct = IDL:org/csapi/fw/TpServiceTypeDescription:1.0 */
static void
decode_org_csapi_fw_TpServiceTypeDescription_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_TpServiceTypeDescription_ServiceTypePropertyList;
    guint32   i_org_csapi_fw_TpServiceTypeDescription_ServiceTypePropertyList;
    guint32   u_octet4_loop_org_csapi_fw_TpServiceTypeDescription_ServiceTypeNameList;
    guint32   i_org_csapi_fw_TpServiceTypeDescription_ServiceTypeNameList;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_fw_TpServiceTypeDescription_ServiceTypePropertyList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpServiceTypeDescription_ServiceTypePropertyList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_TpServiceTypeDescription_ServiceTypePropertyList);

    for (i_org_csapi_fw_TpServiceTypeDescription_ServiceTypePropertyList=0; i_org_csapi_fw_TpServiceTypeDescription_ServiceTypePropertyList < u_octet4_loop_org_csapi_fw_TpServiceTypeDescription_ServiceTypePropertyList; i_org_csapi_fw_TpServiceTypeDescription_ServiceTypePropertyList++) {

        /*  Begin struct "org_csapi_fw_TpServiceTypeProperty"  */
        decode_org_csapi_fw_TpServiceTypeProperty_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpServiceTypeProperty"  */
    }

    u_octet4_loop_org_csapi_fw_TpServiceTypeDescription_ServiceTypeNameList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpServiceTypeDescription_ServiceTypeNameList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_TpServiceTypeDescription_ServiceTypeNameList);

    for (i_org_csapi_fw_TpServiceTypeDescription_ServiceTypeNameList=0; i_org_csapi_fw_TpServiceTypeDescription_ServiceTypeNameList < u_octet4_loop_org_csapi_fw_TpServiceTypeDescription_ServiceTypeNameList; i_org_csapi_fw_TpServiceTypeDescription_ServiceTypeNameList++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpServiceTypeDescription_ServiceTypeNameList);

    }

    proto_tree_add_boolean(tree, hf_org_csapi_fw_TpServiceTypeDescription_AvailableOrUnavailable, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

}

/* Struct = IDL:org/csapi/fw/TpSignatureAndServiceMgr:1.0 */
static void
decode_org_csapi_fw_TpSignatureAndServiceMgr_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_TpSignatureAndServiceMgr_DigitalSignature;
    guint32   i_org_csapi_fw_TpSignatureAndServiceMgr_DigitalSignature;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_fw_TpSignatureAndServiceMgr_DigitalSignature = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpSignatureAndServiceMgr_DigitalSignature_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_TpSignatureAndServiceMgr_DigitalSignature);

    for (i_org_csapi_fw_TpSignatureAndServiceMgr_DigitalSignature=0; i_org_csapi_fw_TpSignatureAndServiceMgr_DigitalSignature < u_octet4_loop_org_csapi_fw_TpSignatureAndServiceMgr_DigitalSignature; i_org_csapi_fw_TpSignatureAndServiceMgr_DigitalSignature++) {

        proto_tree_add_uint(tree, hf_org_csapi_fw_TpSignatureAndServiceMgr_DigitalSignature, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

    }

    get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

}

/* Struct = IDL:org/csapi/fw/TpPerson:1.0 */
static void
decode_org_csapi_fw_TpPerson_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_TpPerson_PersonProperties;
    guint32   i_org_csapi_fw_TpPerson_PersonProperties;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpPerson_PersonName);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpPerson_PostalAddress);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpPerson_TelephoneNumber);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpPerson_Email);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpPerson_HomePage);

    u_octet4_loop_org_csapi_fw_TpPerson_PersonProperties = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpPerson_PersonProperties_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_TpPerson_PersonProperties);

    for (i_org_csapi_fw_TpPerson_PersonProperties=0; i_org_csapi_fw_TpPerson_PersonProperties < u_octet4_loop_org_csapi_fw_TpPerson_PersonProperties; i_org_csapi_fw_TpPerson_PersonProperties++) {

        /*  Begin struct "org_csapi_fw_TpProperty"  */
        decode_org_csapi_fw_TpProperty_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpProperty"  */
    }

}

/* Struct = IDL:org/csapi/fw/TpServiceContractDescription:1.0 */
static void
decode_org_csapi_fw_TpServiceContractDescription_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_TpServiceContractDescription_ServiceSubscriptionProperties;
    guint32   i_org_csapi_fw_TpServiceContractDescription_ServiceSubscriptionProperties;
    /* Operation specific Variable declarations End */

    /*  Begin struct "org_csapi_fw_TpPerson"  */
    decode_org_csapi_fw_TpPerson_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_fw_TpPerson"  */
    /*  Begin struct "org_csapi_fw_TpPerson"  */
    decode_org_csapi_fw_TpPerson_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_fw_TpPerson"  */
    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpServiceContractDescription_ServiceStartDate);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpServiceContractDescription_ServiceEndDate);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpServiceContractDescription_ServiceTypeName);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpServiceContractDescription_ServiceID);

    u_octet4_loop_org_csapi_fw_TpServiceContractDescription_ServiceSubscriptionProperties = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpServiceContractDescription_ServiceSubscriptionProperties_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_TpServiceContractDescription_ServiceSubscriptionProperties);

    for (i_org_csapi_fw_TpServiceContractDescription_ServiceSubscriptionProperties=0; i_org_csapi_fw_TpServiceContractDescription_ServiceSubscriptionProperties < u_octet4_loop_org_csapi_fw_TpServiceContractDescription_ServiceSubscriptionProperties; i_org_csapi_fw_TpServiceContractDescription_ServiceSubscriptionProperties++) {

        /*  Begin struct "org_csapi_fw_TpServiceProperty"  */
        decode_org_csapi_fw_TpServiceProperty_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpServiceProperty"  */
    }

    proto_tree_add_boolean(tree, hf_org_csapi_fw_TpServiceContractDescription_InUse, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

}

/* Struct = IDL:org/csapi/fw/TpLoadStatistic:1.0 */
static void
decode_org_csapi_fw_TpLoadStatistic_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin union "org_csapi_fw_TpLoadStatisticEntityID"  */
    decode_org_csapi_fw_TpLoadStatisticEntityID_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

    /*  End union "org_csapi_fw_TpLoadStatisticEntityID"  */
    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpLoadStatistic_TimeStamp);

    /*  Begin union "org_csapi_fw_TpLoadStatisticInfo"  */
    decode_org_csapi_fw_TpLoadStatisticInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

    /*  End union "org_csapi_fw_TpLoadStatisticInfo"  */
}

/* Struct = IDL:org/csapi/fw/TpServiceContract:1.0 */
static void
decode_org_csapi_fw_TpServiceContract_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpServiceContract_ServiceContractID);

    /*  Begin struct "org_csapi_fw_TpServiceContractDescription"  */
    decode_org_csapi_fw_TpServiceContractDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_fw_TpServiceContractDescription"  */
}

/* Struct = IDL:org/csapi/fw/TpServiceProfile:1.0 */
static void
decode_org_csapi_fw_TpServiceProfile_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpServiceProfile_ServiceProfileID);

    /*  Begin struct "org_csapi_fw_TpServiceProfileDescription"  */
    decode_org_csapi_fw_TpServiceProfileDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_fw_TpServiceProfileDescription"  */
}

/* Struct = IDL:org/csapi/fw/TpAuthDomain:1.0 */
static void
decode_org_csapi_fw_TpAuthDomain_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin union "org_csapi_fw_TpDomainID"  */
    decode_org_csapi_fw_TpDomainID_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

    /*  End union "org_csapi_fw_TpDomainID"  */
    get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

}

/* Struct = IDL:org/csapi/fw/TpSagProfilePair:1.0 */
static void
decode_org_csapi_fw_TpSagProfilePair_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpSagProfilePair_Sag);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpSagProfilePair_ServiceProfile);

}

/* Struct = IDL:org/csapi/fw/TpAddSagMembersConflict:1.0 */
static void
decode_org_csapi_fw_TpAddSagMembersConflict_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpAddSagMembersConflict_ClientApplication);

    /*  Begin struct "org_csapi_fw_TpSagProfilePair"  */
    decode_org_csapi_fw_TpSagProfilePair_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_fw_TpSagProfilePair"  */
    /*  Begin struct "org_csapi_fw_TpSagProfilePair"  */
    decode_org_csapi_fw_TpSagProfilePair_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_fw_TpSagProfilePair"  */
    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpAddSagMembersConflict_Service);

}

/* Struct = IDL:org/csapi/fw/TpAssignSagToServiceProfileConflict:1.0 */
static void
decode_org_csapi_fw_TpAssignSagToServiceProfileConflict_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpAssignSagToServiceProfileConflict_ClientApplication);

    /*  Begin struct "org_csapi_fw_TpSagProfilePair"  */
    decode_org_csapi_fw_TpSagProfilePair_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_fw_TpSagProfilePair"  */
    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpAssignSagToServiceProfileConflict_Service);

}

/* Struct = IDL:org/csapi/fw/TpServiceTypePropertyValue:1.0 */
static void
decode_org_csapi_fw_TpServiceTypePropertyValue_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    guint32   u_octet4_loop_org_csapi_fw_TpServiceTypePropertyValue_ServicePropertyValueList;
    guint32   i_org_csapi_fw_TpServiceTypePropertyValue_ServicePropertyValueList;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpServiceTypePropertyValue_ServicePropertyName);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpServiceTypePropertyValue_ServiceTypePropertyMode, tvb, *offset-4, 4, u_octet4);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpServiceTypePropertyValue_ServicePropertyTypeName);

    u_octet4_loop_org_csapi_fw_TpServiceTypePropertyValue_ServicePropertyValueList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpServiceTypePropertyValue_ServicePropertyValueList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_TpServiceTypePropertyValue_ServicePropertyValueList);

    for (i_org_csapi_fw_TpServiceTypePropertyValue_ServicePropertyValueList=0; i_org_csapi_fw_TpServiceTypePropertyValue_ServicePropertyValueList < u_octet4_loop_org_csapi_fw_TpServiceTypePropertyValue_ServicePropertyValueList; i_org_csapi_fw_TpServiceTypePropertyValue_ServicePropertyValueList++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpServiceTypePropertyValue_ServicePropertyValueList);

    }

}

/* Struct = IDL:org/csapi/fw/TpFwMigrationServiceAvailableInfo:1.0 */
static void
decode_org_csapi_fw_TpFwMigrationServiceAvailableInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_fw_TpFwMigrationServiceAvailableInfo_MigrationAdditionalInfo;
    guint32   i_org_csapi_fw_TpFwMigrationServiceAvailableInfo_MigrationAdditionalInfo;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_ServiceType);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_ServiceID);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_CompatibleServiceID);

    proto_tree_add_boolean(tree, hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_BackwardCompatibilityLevel, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    proto_tree_add_boolean(tree, hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_MigrationRequired, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    proto_tree_add_boolean(tree, hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_DataMigrated, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_MigrationDataAndTime);

    u_octet4_loop_org_csapi_fw_TpFwMigrationServiceAvailableInfo_MigrationAdditionalInfo = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_MigrationAdditionalInfo_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_TpFwMigrationServiceAvailableInfo_MigrationAdditionalInfo);

    for (i_org_csapi_fw_TpFwMigrationServiceAvailableInfo_MigrationAdditionalInfo=0; i_org_csapi_fw_TpFwMigrationServiceAvailableInfo_MigrationAdditionalInfo < u_octet4_loop_org_csapi_fw_TpFwMigrationServiceAvailableInfo_MigrationAdditionalInfo; i_org_csapi_fw_TpFwMigrationServiceAvailableInfo_MigrationAdditionalInfo++) {

        /*  Begin union "org_csapi_fw_TpMigrationAdditionalInfo"  */
        decode_org_csapi_fw_TpMigrationAdditionalInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_fw_TpMigrationAdditionalInfo"  */
    }

}

/* Struct = IDL:org/csapi/fw/TpFwAgreementInfo:1.0 */
static void
decode_org_csapi_fw_TpFwAgreementInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpFwAgreementInfo_ClientApplicationID);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpFwAgreementInfo_ServiceID);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpFwAgreementInfo_ServiceContractID);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpFwAgreementInfo_ServiceProfileID);

}

/* Struct = IDL:org/csapi/ui/TpUIMessageCriteria:1.0 */
static void
decode_org_csapi_ui_TpUIMessageCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_TpUIMessageCriteria_EndSequence);

    proto_tree_add_int(tree, hf_org_csapi_ui_TpUIMessageCriteria_MaxMessageTime, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_ui_TpUIMessageCriteria_MaxMessageSize, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/ui/TpUIEventInfo:1.0 */
static void
decode_org_csapi_ui_TpUIEventInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_TpUIEventInfo_ServiceCode);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_ui_TpUIEventInfo_DataTypeIndication, tvb, *offset-4, 4, u_octet4);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_TpUIEventInfo_DataString);

}

/* Struct = IDL:org/csapi/ui/TpUIEventCriteria:1.0 */
static void
decode_org_csapi_ui_TpUIEventCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_TpAddressRange"  */
    decode_org_csapi_TpAddressRange_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddressRange"  */
    /*  Begin struct "org_csapi_TpAddressRange"  */
    decode_org_csapi_TpAddressRange_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddressRange"  */
    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_TpUIEventCriteria_ServiceCode);

}

/* Struct = IDL:org/csapi/ui/TpUIEventCriteriaResult:1.0 */
static void
decode_org_csapi_ui_TpUIEventCriteriaResult_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_ui_TpUIEventCriteria"  */
    decode_org_csapi_ui_TpUIEventCriteria_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_ui_TpUIEventCriteria"  */
    proto_tree_add_int(tree, hf_org_csapi_ui_TpUIEventCriteriaResult_AssignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/ui/TpUIEventNotificationInfo:1.0 */
static void
decode_org_csapi_ui_TpUIEventNotificationInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    guint32   u_octet4_loop_org_csapi_ui_TpUIEventNotificationInfo_UIEventData;
    guint32   i_org_csapi_ui_TpUIEventNotificationInfo_UIEventData;
    /* Operation specific Variable declarations End */

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_TpUIEventNotificationInfo_ServiceCode);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_ui_TpUIEventNotificationInfo_DataTypeIndication, tvb, *offset-4, 4, u_octet4);

    u_octet4_loop_org_csapi_ui_TpUIEventNotificationInfo_UIEventData = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_ui_TpUIEventNotificationInfo_UIEventData_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_ui_TpUIEventNotificationInfo_UIEventData);

    for (i_org_csapi_ui_TpUIEventNotificationInfo_UIEventData=0; i_org_csapi_ui_TpUIEventNotificationInfo_UIEventData < u_octet4_loop_org_csapi_ui_TpUIEventNotificationInfo_UIEventData; i_org_csapi_ui_TpUIEventNotificationInfo_UIEventData++) {

        proto_tree_add_uint(tree, hf_org_csapi_ui_TpUIEventNotificationInfo_UIEventData, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

    }

}

/* Struct = IDL:org/csapi/ui/TpUIRecognitionProperty:1.0 */
static void
decode_org_csapi_ui_TpUIRecognitionProperty_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_TpUIRecognitionProperty_PropertyName);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_TpUIRecognitionProperty_PropertyValue);

}

/* Struct = IDL:org/csapi/ui/TpUIRecognitionCriteria:1.0 */
static void
decode_org_csapi_ui_TpUIRecognitionCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_ui_TpUIRecognitionCriteria_Properties;
    guint32   i_org_csapi_ui_TpUIRecognitionCriteria_Properties;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_TpUIRecognitionCriteria_SpeakerID);

    u_octet4_loop_org_csapi_ui_TpUIRecognitionCriteria_Properties = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_ui_TpUIRecognitionCriteria_Properties_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_ui_TpUIRecognitionCriteria_Properties);

    for (i_org_csapi_ui_TpUIRecognitionCriteria_Properties=0; i_org_csapi_ui_TpUIRecognitionCriteria_Properties < u_octet4_loop_org_csapi_ui_TpUIRecognitionCriteria_Properties; i_org_csapi_ui_TpUIRecognitionCriteria_Properties++) {

        /*  Begin struct "org_csapi_ui_TpUIRecognitionProperty"  */
        decode_org_csapi_ui_TpUIRecognitionProperty_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_ui_TpUIRecognitionProperty"  */
    }

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_TpUIRecognitionCriteria_Grammar);

}

/* Struct = IDL:org/csapi/ui/TpUICollectCriteria:1.0 */
static void
decode_org_csapi_ui_TpUICollectCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    proto_tree_add_int(tree, hf_org_csapi_ui_TpUICollectCriteria_MinLength, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_ui_TpUICollectCriteria_MaxLength, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_TpUICollectCriteria_EndSequence);

    proto_tree_add_int(tree, hf_org_csapi_ui_TpUICollectCriteria_StartTimeout, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_ui_TpUICollectCriteria_InterCharTimeout, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_ui_TpUICollectCriteria_CollectMode, tvb, *offset-4, 4, u_octet4);

    /*  Begin struct "org_csapi_ui_TpUIRecognitionCriteria"  */
    decode_org_csapi_ui_TpUIRecognitionCriteria_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_ui_TpUIRecognitionCriteria"  */
}

/* Struct = IDL:org/csapi/ui/TpUIWordOverride:1.0 */
static void
decode_org_csapi_ui_TpUIWordOverride_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_TpUIWordOverride_Spelling);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_ui_TpUIWordOverride_PronounceType, tvb, *offset-4, 4, u_octet4);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_TpUIWordOverride_PronounceAs);

}

/* Struct = IDL:org/csapi/ui/TpUISynthesisInfoData:1.0 */
static void
decode_org_csapi_ui_TpUISynthesisInfoData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    guint32   u_octet4_loop_org_csapi_ui_TpUISynthesisInfoData_WordOverrideSet;
    guint32   i_org_csapi_ui_TpUISynthesisInfoData_WordOverrideSet;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_ui_TpUISynthesisInfoData_SpeakerGender, tvb, *offset-4, 4, u_octet4);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_ui_TpUISynthesisInfoData_SpeakerAge, tvb, *offset-4, 4, u_octet4);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_ui_TpUISynthesisInfoData_SpeakerRate, tvb, *offset-4, 4, u_octet4);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_ui_TpUISynthesisInfoData_SpeakerRange, tvb, *offset-4, 4, u_octet4);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_TpUISynthesisInfoData_TextData);

    u_octet4_loop_org_csapi_ui_TpUISynthesisInfoData_WordOverrideSet = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_ui_TpUISynthesisInfoData_WordOverrideSet_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_ui_TpUISynthesisInfoData_WordOverrideSet);

    for (i_org_csapi_ui_TpUISynthesisInfoData_WordOverrideSet=0; i_org_csapi_ui_TpUISynthesisInfoData_WordOverrideSet < u_octet4_loop_org_csapi_ui_TpUISynthesisInfoData_WordOverrideSet; i_org_csapi_ui_TpUISynthesisInfoData_WordOverrideSet++) {

        /*  Begin struct "org_csapi_ui_TpUIWordOverride"  */
        decode_org_csapi_ui_TpUIWordOverride_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_ui_TpUIWordOverride"  */
    }

}

/* Struct = IDL:org/csapi/cc/TpCallError:1.0 */
static void
decode_org_csapi_cc_TpCallError_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_TpCallError_ErrorTime);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallError_ErrorType, tvb, *offset-4, 4, u_octet4);

    /*  Begin union "org_csapi_cc_TpCallAdditionalErrorInfo"  */
    decode_org_csapi_cc_TpCallAdditionalErrorInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

    /*  End union "org_csapi_cc_TpCallAdditionalErrorInfo"  */
}

/* Struct = IDL:org/csapi/cc/TpCallEndedReport:1.0 */
static void
decode_org_csapi_cc_TpCallEndedReport_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    proto_tree_add_int(tree, hf_org_csapi_cc_TpCallEndedReport_CallLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallEndedReport_Cause, tvb, *offset-4, 4, u_octet4);

}

/* Struct = IDL:org/csapi/cc/TpCallInfoReport:1.0 */
static void
decode_org_csapi_cc_TpCallInfoReport_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    proto_tree_add_int(tree, hf_org_csapi_cc_TpCallInfoReport_CallInfoType, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_TpCallInfoReport_CallInitiationStartTime);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_TpCallInfoReport_CallConnectedToResourceTime);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_TpCallInfoReport_CallConnectedToDestinationTime);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_TpCallInfoReport_CallEndTime);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallInfoReport_Cause, tvb, *offset-4, 4, u_octet4);

}

/* Struct = IDL:org/csapi/cc/TpCallServiceCode:1.0 */
static void
decode_org_csapi_cc_TpCallServiceCode_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallServiceCode_CallServiceCodeType, tvb, *offset-4, 4, u_octet4);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_TpCallServiceCode_ServiceCodeValue);

}

/* Struct = IDL:org/csapi/cc/TpCallTreatment:1.0 */
static void
decode_org_csapi_cc_TpCallTreatment_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallTreatment_CallTreatmentType, tvb, *offset-4, 4, u_octet4);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallTreatment_ReleaseCause, tvb, *offset-4, 4, u_octet4);

    /*  Begin union "org_csapi_cc_TpCallAdditionalTreatmentInfo"  */
    decode_org_csapi_cc_TpCallAdditionalTreatmentInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

    /*  End union "org_csapi_cc_TpCallAdditionalTreatmentInfo"  */
}

/* Struct = IDL:org/csapi/cc/TpCallChargePlan:1.0 */
static void
decode_org_csapi_cc_TpCallChargePlan_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    guint32   u_octet4_loop_org_csapi_cc_TpCallChargePlan_TransparentCharge;
    guint32   i_org_csapi_cc_TpCallChargePlan_TransparentCharge;
    guint32   u_octet4_loop_org_csapi_cc_TpCallChargePlan_AdditionalInfo;
    guint32   i_org_csapi_cc_TpCallChargePlan_AdditionalInfo;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallChargePlan_ChargeOrderType, tvb, *offset-4, 4, u_octet4);

    u_octet4_loop_org_csapi_cc_TpCallChargePlan_TransparentCharge = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallChargePlan_TransparentCharge_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_TpCallChargePlan_TransparentCharge);

    for (i_org_csapi_cc_TpCallChargePlan_TransparentCharge=0; i_org_csapi_cc_TpCallChargePlan_TransparentCharge < u_octet4_loop_org_csapi_cc_TpCallChargePlan_TransparentCharge; i_org_csapi_cc_TpCallChargePlan_TransparentCharge++) {

        proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallChargePlan_TransparentCharge, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

    }

    proto_tree_add_int(tree, hf_org_csapi_cc_TpCallChargePlan_ChargePlan, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    u_octet4_loop_org_csapi_cc_TpCallChargePlan_AdditionalInfo = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallChargePlan_AdditionalInfo_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_TpCallChargePlan_AdditionalInfo);

    for (i_org_csapi_cc_TpCallChargePlan_AdditionalInfo=0; i_org_csapi_cc_TpCallChargePlan_AdditionalInfo < u_octet4_loop_org_csapi_cc_TpCallChargePlan_AdditionalInfo; i_org_csapi_cc_TpCallChargePlan_AdditionalInfo++) {

        proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallChargePlan_AdditionalInfo, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

    }

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallChargePlan_PartyToCharge, tvb, *offset-4, 4, u_octet4);

    /*  Begin union "org_csapi_cc_TpCallPartyToChargeAdditionalInfo"  */
    decode_org_csapi_cc_TpCallPartyToChargeAdditionalInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

    /*  End union "org_csapi_cc_TpCallPartyToChargeAdditionalInfo"  */
}

/* Struct = IDL:org/csapi/cc/gccs/TpCallEventCriteria:1.0 */
static void
decode_org_csapi_cc_gccs_TpCallEventCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    /*  Begin struct "org_csapi_TpAddressRange"  */
    decode_org_csapi_TpAddressRange_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddressRange"  */
    /*  Begin struct "org_csapi_TpAddressRange"  */
    decode_org_csapi_TpAddressRange_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddressRange"  */
    proto_tree_add_int(tree, hf_org_csapi_cc_gccs_TpCallEventCriteria_CallEventName, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_gccs_TpCallEventCriteria_CallNotificationType, tvb, *offset-4, 4, u_octet4);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_gccs_TpCallEventCriteria_MonitorMode, tvb, *offset-4, 4, u_octet4);

}

/* Struct = IDL:org/csapi/cc/gccs/TpCallEventCriteriaResult:1.0 */
static void
decode_org_csapi_cc_gccs_TpCallEventCriteriaResult_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_cc_gccs_TpCallEventCriteria"  */
    decode_org_csapi_cc_gccs_TpCallEventCriteria_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cc_gccs_TpCallEventCriteria"  */
    proto_tree_add_int(tree, hf_org_csapi_cc_gccs_TpCallEventCriteriaResult_AssignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cc/gccs/TpCallReleaseCause:1.0 */
static void
decode_org_csapi_cc_gccs_TpCallReleaseCause_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    proto_tree_add_int(tree, hf_org_csapi_cc_gccs_TpCallReleaseCause_Value, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_cc_gccs_TpCallReleaseCause_Location, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cc/gccs/TpCallReport:1.0 */
static void
decode_org_csapi_cc_gccs_TpCallReport_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_gccs_TpCallReport_MonitorMode, tvb, *offset-4, 4, u_octet4);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_gccs_TpCallReport_CallEventTime);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_gccs_TpCallReport_CallReportType, tvb, *offset-4, 4, u_octet4);

    /*  Begin union "org_csapi_cc_gccs_TpCallAdditionalReportInfo"  */
    decode_org_csapi_cc_gccs_TpCallAdditionalReportInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

    /*  End union "org_csapi_cc_gccs_TpCallAdditionalReportInfo"  */
}

/* Struct = IDL:org/csapi/cc/gccs/TpCallReportRequest:1.0 */
static void
decode_org_csapi_cc_gccs_TpCallReportRequest_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_gccs_TpCallReportRequest_MonitorMode, tvb, *offset-4, 4, u_octet4);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_gccs_TpCallReportRequest_CallReportType, tvb, *offset-4, 4, u_octet4);

    /*  Begin union "org_csapi_cc_gccs_TpCallAdditionalReportCriteria"  */
    decode_org_csapi_cc_gccs_TpCallAdditionalReportCriteria_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

    /*  End union "org_csapi_cc_gccs_TpCallAdditionalReportCriteria"  */
}

/* Struct = IDL:org/csapi/cc/gccs/TpCallEventInfo:1.0 */
static void
decode_org_csapi_cc_gccs_TpCallEventInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_gccs_TpCallEventInfo_CallAppInfo;
    guint32   i_org_csapi_cc_gccs_TpCallEventInfo_CallAppInfo;
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    u_octet4_loop_org_csapi_cc_gccs_TpCallEventInfo_CallAppInfo = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_gccs_TpCallEventInfo_CallAppInfo_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_gccs_TpCallEventInfo_CallAppInfo);

    for (i_org_csapi_cc_gccs_TpCallEventInfo_CallAppInfo=0; i_org_csapi_cc_gccs_TpCallEventInfo_CallAppInfo < u_octet4_loop_org_csapi_cc_gccs_TpCallEventInfo_CallAppInfo; i_org_csapi_cc_gccs_TpCallEventInfo_CallAppInfo++) {

        /*  Begin union "org_csapi_cc_gccs_TpCallAppInfo"  */
        decode_org_csapi_cc_gccs_TpCallAppInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_cc_gccs_TpCallAppInfo"  */
    }

    proto_tree_add_int(tree, hf_org_csapi_cc_gccs_TpCallEventInfo_CallEventName, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_gccs_TpCallEventInfo_CallNotificationType, tvb, *offset-4, 4, u_octet4);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_gccs_TpCallEventInfo_MonitorMode, tvb, *offset-4, 4, u_octet4);

}

/* Struct = IDL:org/csapi/cc/gccs/TpCallEndedReport:1.0 */
static void
decode_org_csapi_cc_gccs_TpCallEndedReport_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    proto_tree_add_int(tree, hf_org_csapi_cc_gccs_TpCallEndedReport_CallLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    /*  Begin struct "org_csapi_cc_gccs_TpCallReleaseCause"  */
    decode_org_csapi_cc_gccs_TpCallReleaseCause_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cc_gccs_TpCallReleaseCause"  */
}

/* Struct = IDL:org/csapi/cc/gccs/TpCallInfoReport:1.0 */
static void
decode_org_csapi_cc_gccs_TpCallInfoReport_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    proto_tree_add_int(tree, hf_org_csapi_cc_gccs_TpCallInfoReport_CallInfoType, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_gccs_TpCallInfoReport_CallInitiationStartTime);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_gccs_TpCallInfoReport_CallConnectedToResourceTime);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_gccs_TpCallInfoReport_CallConnectedToDestinationTime);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_gccs_TpCallInfoReport_CallEndTime);

    /*  Begin struct "org_csapi_cc_gccs_TpCallReleaseCause"  */
    decode_org_csapi_cc_gccs_TpCallReleaseCause_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cc_gccs_TpCallReleaseCause"  */
}

/* Struct = IDL:org/csapi/cc/gccs/TpCallTreatment:1.0 */
static void
decode_org_csapi_cc_gccs_TpCallTreatment_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_gccs_TpCallTreatment_CallTreatmentType, tvb, *offset-4, 4, u_octet4);

    /*  Begin struct "org_csapi_cc_gccs_TpCallReleaseCause"  */
    decode_org_csapi_cc_gccs_TpCallReleaseCause_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cc_gccs_TpCallReleaseCause"  */
    /*  Begin union "org_csapi_cc_TpCallAdditionalTreatmentInfo"  */
    decode_org_csapi_cc_TpCallAdditionalTreatmentInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

    /*  End union "org_csapi_cc_TpCallAdditionalTreatmentInfo"  */
}

/* Struct = IDL:org/csapi/cc/gccs/TpCallIdentifier:1.0 */
static void
decode_org_csapi_cc_gccs_TpCallIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

    proto_tree_add_int(tree, hf_org_csapi_cc_gccs_TpCallIdentifier_CallSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cc/TpCallNotificationScope:1.0 */
static void
decode_org_csapi_cc_TpCallNotificationScope_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_TpAddressRange"  */
    decode_org_csapi_TpAddressRange_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddressRange"  */
    /*  Begin struct "org_csapi_TpAddressRange"  */
    decode_org_csapi_TpAddressRange_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddressRange"  */
}

/* Struct = IDL:org/csapi/cc/TpCallNotificationReportScope:1.0 */
static void
decode_org_csapi_cc_TpCallNotificationReportScope_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
}

/* Struct = IDL:org/csapi/cc/TpCallEventRequest:1.0 */
static void
decode_org_csapi_cc_TpCallEventRequest_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallEventRequest_CallEventType, tvb, *offset-4, 4, u_octet4);

    /*  Begin union "org_csapi_cc_TpAdditionalCallEventCriteria"  */
    decode_org_csapi_cc_TpAdditionalCallEventCriteria_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

    /*  End union "org_csapi_cc_TpAdditionalCallEventCriteria"  */
    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallEventRequest_CallMonitorMode, tvb, *offset-4, 4, u_octet4);

}

/* Struct = IDL:org/csapi/cc/TpCallNotificationRequest:1.0 */
static void
decode_org_csapi_cc_TpCallNotificationRequest_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_TpCallNotificationRequest_CallEventsRequested;
    guint32   i_org_csapi_cc_TpCallNotificationRequest_CallEventsRequested;
    /* Operation specific Variable declarations End */

    /*  Begin struct "org_csapi_cc_TpCallNotificationScope"  */
    decode_org_csapi_cc_TpCallNotificationScope_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cc_TpCallNotificationScope"  */
    u_octet4_loop_org_csapi_cc_TpCallNotificationRequest_CallEventsRequested = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallNotificationRequest_CallEventsRequested_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_TpCallNotificationRequest_CallEventsRequested);

    for (i_org_csapi_cc_TpCallNotificationRequest_CallEventsRequested=0; i_org_csapi_cc_TpCallNotificationRequest_CallEventsRequested < u_octet4_loop_org_csapi_cc_TpCallNotificationRequest_CallEventsRequested; i_org_csapi_cc_TpCallNotificationRequest_CallEventsRequested++) {

        /*  Begin struct "org_csapi_cc_TpCallEventRequest"  */
        decode_org_csapi_cc_TpCallEventRequest_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallEventRequest"  */
    }

}

/* Struct = IDL:org/csapi/cc/TpNotificationRequested:1.0 */
static void
decode_org_csapi_cc_TpNotificationRequested_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_cc_TpCallNotificationRequest"  */
    decode_org_csapi_cc_TpCallNotificationRequest_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cc_TpCallNotificationRequest"  */
    proto_tree_add_int(tree, hf_org_csapi_cc_TpNotificationRequested_AssignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cc/TpCallLegConnectionProperties:1.0 */
static void
decode_org_csapi_cc_TpCallLegConnectionProperties_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallLegConnectionProperties_AttachMechanism, tvb, *offset-4, 4, u_octet4);

}

/* Struct = IDL:org/csapi/cc/TpCallEventInfo:1.0 */
static void
decode_org_csapi_cc_TpCallEventInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallEventInfo_CallEventType, tvb, *offset-4, 4, u_octet4);

    /*  Begin union "org_csapi_cc_TpCallAdditionalEventInfo"  */
    decode_org_csapi_cc_TpCallAdditionalEventInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

    /*  End union "org_csapi_cc_TpCallAdditionalEventInfo"  */
    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallEventInfo_CallMonitorMode, tvb, *offset-4, 4, u_octet4);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_TpCallEventInfo_CallEventTime);

}

/* Struct = IDL:org/csapi/cc/TpNotificationRequestedSetEntry:1.0 */
static void
decode_org_csapi_cc_TpNotificationRequestedSetEntry_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_TpNotificationRequestedSetEntry_NotificationRequestSet;
    guint32   i_org_csapi_cc_TpNotificationRequestedSetEntry_NotificationRequestSet;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_cc_TpNotificationRequestedSetEntry_NotificationRequestSet = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpNotificationRequestedSetEntry_NotificationRequestSet_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_TpNotificationRequestedSetEntry_NotificationRequestSet);

    for (i_org_csapi_cc_TpNotificationRequestedSetEntry_NotificationRequestSet=0; i_org_csapi_cc_TpNotificationRequestedSetEntry_NotificationRequestSet < u_octet4_loop_org_csapi_cc_TpNotificationRequestedSetEntry_NotificationRequestSet; i_org_csapi_cc_TpNotificationRequestedSetEntry_NotificationRequestSet++) {

        /*  Begin struct "org_csapi_cc_TpNotificationRequested"  */
        decode_org_csapi_cc_TpNotificationRequested_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpNotificationRequested"  */
    }

    proto_tree_add_boolean(tree, hf_org_csapi_cc_TpNotificationRequestedSetEntry_Final, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

}

/* Struct = IDL:org/csapi/cc/TpCarrier:1.0 */
static void
decode_org_csapi_cc_TpCarrier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_TpCarrier_CarrierID;
    guint32   i_org_csapi_cc_TpCarrier_CarrierID;
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_cc_TpCarrier_CarrierID = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCarrier_CarrierID_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_TpCarrier_CarrierID);

    for (i_org_csapi_cc_TpCarrier_CarrierID=0; i_org_csapi_cc_TpCarrier_CarrierID < u_octet4_loop_org_csapi_cc_TpCarrier_CarrierID; i_org_csapi_cc_TpCarrier_CarrierID++) {

        proto_tree_add_uint(tree, hf_org_csapi_cc_TpCarrier_CarrierID, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

    }

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCarrier_CarrierSelectionField, tvb, *offset-4, 4, u_octet4);

}

/* Struct = IDL:org/csapi/cc/TpCallNotificationInfo:1.0 */
static void
decode_org_csapi_cc_TpCallNotificationInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_TpCallNotificationInfo_CallAppInfo;
    guint32   i_org_csapi_cc_TpCallNotificationInfo_CallAppInfo;
    /* Operation specific Variable declarations End */

    /*  Begin struct "org_csapi_cc_TpCallNotificationReportScope"  */
    decode_org_csapi_cc_TpCallNotificationReportScope_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cc_TpCallNotificationReportScope"  */
    u_octet4_loop_org_csapi_cc_TpCallNotificationInfo_CallAppInfo = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallNotificationInfo_CallAppInfo_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_TpCallNotificationInfo_CallAppInfo);

    for (i_org_csapi_cc_TpCallNotificationInfo_CallAppInfo=0; i_org_csapi_cc_TpCallNotificationInfo_CallAppInfo < u_octet4_loop_org_csapi_cc_TpCallNotificationInfo_CallAppInfo; i_org_csapi_cc_TpCallNotificationInfo_CallAppInfo++) {

        /*  Begin union "org_csapi_cc_TpCallAppInfo"  */
        decode_org_csapi_cc_TpCallAppInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_cc_TpCallAppInfo"  */
    }

    /*  Begin struct "org_csapi_cc_TpCallEventInfo"  */
    decode_org_csapi_cc_TpCallEventInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cc_TpCallEventInfo"  */
}

/* Struct = IDL:org/csapi/cc/TpCallLegInfoReport:1.0 */
static void
decode_org_csapi_cc_TpCallLegInfoReport_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    guint32   u_octet4_loop_org_csapi_cc_TpCallLegInfoReport_CallAppInfo;
    guint32   i_org_csapi_cc_TpCallLegInfoReport_CallAppInfo;
    /* Operation specific Variable declarations End */

    proto_tree_add_int(tree, hf_org_csapi_cc_TpCallLegInfoReport_CallLegInfoType, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_TpCallLegInfoReport_CallLegStartTime);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_TpCallLegInfoReport_CallLegConnectedToResourceTime);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_TpCallLegInfoReport_CallLegConnectedToAddressTime);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_TpCallLegInfoReport_CallLegEndTime);

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallLegInfoReport_CallLegReleaseCause, tvb, *offset-4, 4, u_octet4);

    u_octet4_loop_org_csapi_cc_TpCallLegInfoReport_CallAppInfo = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallLegInfoReport_CallAppInfo_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_TpCallLegInfoReport_CallAppInfo);

    for (i_org_csapi_cc_TpCallLegInfoReport_CallAppInfo=0; i_org_csapi_cc_TpCallLegInfoReport_CallAppInfo < u_octet4_loop_org_csapi_cc_TpCallLegInfoReport_CallAppInfo; i_org_csapi_cc_TpCallLegInfoReport_CallAppInfo++) {

        /*  Begin union "org_csapi_cc_TpCallAppInfo"  */
        decode_org_csapi_cc_TpCallAppInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_cc_TpCallAppInfo"  */
    }

}

/* Struct = IDL:org/csapi/cc/TpCallLegProperty:1.0 */
static void
decode_org_csapi_cc_TpCallLegProperty_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_TpCallLegProperty_CallLegPropertyName);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_TpCallLegProperty_CallLegPropertyValue);

}

/* Struct = IDL:org/csapi/cc/mpccs/TpMultiPartyCallIdentifier:1.0 */
static void
decode_org_csapi_cc_mpccs_TpMultiPartyCallIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

    proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_TpMultiPartyCallIdentifier_CallSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cc/mpccs/TpCallLegIdentifier:1.0 */
static void
decode_org_csapi_cc_mpccs_TpCallLegIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

    proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_TpCallLegIdentifier_CallLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cc/mpccs/TpAppCallLegCallBack:1.0 */
static void
decode_org_csapi_cc_mpccs_TpAppCallLegCallBack_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_mpccs_TpAppCallLegCallBack_AppCallLegSet;
    guint32   i_org_csapi_cc_mpccs_TpAppCallLegCallBack_AppCallLegSet;
    /* Operation specific Variable declarations End */

    get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

    u_octet4_loop_org_csapi_cc_mpccs_TpAppCallLegCallBack_AppCallLegSet = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_mpccs_TpAppCallLegCallBack_AppCallLegSet_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_mpccs_TpAppCallLegCallBack_AppCallLegSet);

    for (i_org_csapi_cc_mpccs_TpAppCallLegCallBack_AppCallLegSet=0; i_org_csapi_cc_mpccs_TpAppCallLegCallBack_AppCallLegSet < u_octet4_loop_org_csapi_cc_mpccs_TpAppCallLegCallBack_AppCallLegSet; i_org_csapi_cc_mpccs_TpAppCallLegCallBack_AppCallLegSet++) {

        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

    }

}

/* Struct = IDL:org/csapi/cc/mmccs/TpCallSuperviseVolume:1.0 */
static void
decode_org_csapi_cc_mmccs_TpCallSuperviseVolume_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_TpCallSuperviseVolume_VolumeQuantity, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_TpCallSuperviseVolume_VolumeUnit, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cc/mmccs/TpMediaStreamRequest:1.0 */
static void
decode_org_csapi_cc_mmccs_TpMediaStreamRequest_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_mmccs_TpMediaStreamRequest_Direction, tvb, *offset-4, 4, u_octet4);

    /*  Begin union "org_csapi_cc_mmccs_TpMediaStreamDataTypeRequest"  */
    decode_org_csapi_cc_mmccs_TpMediaStreamDataTypeRequest_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

    /*  End union "org_csapi_cc_mmccs_TpMediaStreamDataTypeRequest"  */
    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_mmccs_TpMediaStreamRequest_MediaMonitorMode, tvb, *offset-4, 4, u_octet4);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_mmccs_TpMediaStreamRequest_EventType, tvb, *offset-4, 4, u_octet4);

}

/* Struct = IDL:org/csapi/cc/mmccs/TpMediaStream:1.0 */
static void
decode_org_csapi_cc_mmccs_TpMediaStream_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_mmccs_TpMediaStream_Direction, tvb, *offset-4, 4, u_octet4);

    /*  Begin union "org_csapi_cc_mmccs_TpMediaStreamDataTypeRequest"  */
    decode_org_csapi_cc_mmccs_TpMediaStreamDataTypeRequest_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

    /*  End union "org_csapi_cc_mmccs_TpMediaStreamDataTypeRequest"  */
    proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_TpMediaStream_ChannelSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

}

/* Struct = IDL:org/csapi/cc/mmccs/TpNotificationMediaRequest:1.0 */
static void
decode_org_csapi_cc_mmccs_TpNotificationMediaRequest_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_mmccs_TpNotificationMediaRequest_MediaStreamsRequested;
    guint32   i_org_csapi_cc_mmccs_TpNotificationMediaRequest_MediaStreamsRequested;
    /* Operation specific Variable declarations End */

    /*  Begin struct "org_csapi_cc_TpCallNotificationScope"  */
    decode_org_csapi_cc_TpCallNotificationScope_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cc_TpCallNotificationScope"  */
    u_octet4_loop_org_csapi_cc_mmccs_TpNotificationMediaRequest_MediaStreamsRequested = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_mmccs_TpNotificationMediaRequest_MediaStreamsRequested_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_mmccs_TpNotificationMediaRequest_MediaStreamsRequested);

    for (i_org_csapi_cc_mmccs_TpNotificationMediaRequest_MediaStreamsRequested=0; i_org_csapi_cc_mmccs_TpNotificationMediaRequest_MediaStreamsRequested < u_octet4_loop_org_csapi_cc_mmccs_TpNotificationMediaRequest_MediaStreamsRequested; i_org_csapi_cc_mmccs_TpNotificationMediaRequest_MediaStreamsRequested++) {

        /*  Begin struct "org_csapi_cc_mmccs_TpMediaStreamRequest"  */
        decode_org_csapi_cc_mmccs_TpMediaStreamRequest_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_mmccs_TpMediaStreamRequest"  */
    }

}

/* Struct = IDL:org/csapi/cc/mmccs/TpMediaNotificationRequested:1.0 */
static void
decode_org_csapi_cc_mmccs_TpMediaNotificationRequested_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_cc_mmccs_TpNotificationMediaRequest"  */
    decode_org_csapi_cc_mmccs_TpNotificationMediaRequest_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cc_mmccs_TpNotificationMediaRequest"  */
    proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_TpMediaNotificationRequested_AssignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cc/mmccs/TpMultiMediaCallIdentifier:1.0 */
static void
decode_org_csapi_cc_mmccs_TpMultiMediaCallIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

    proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_TpMultiMediaCallIdentifier_MMCallSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cc/mmccs/TpMultiMediaCallLegIdentifier:1.0 */
static void
decode_org_csapi_cc_mmccs_TpMultiMediaCallLegIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

    proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_TpMultiMediaCallLegIdentifier_MMCallLegSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cc/mmccs/TpAppMultiMediaCallLegCallBack:1.0 */
static void
decode_org_csapi_cc_mmccs_TpAppMultiMediaCallLegCallBack_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_mmccs_TpAppMultiMediaCallLegCallBack_AppCallLegSet;
    guint32   i_org_csapi_cc_mmccs_TpAppMultiMediaCallLegCallBack_AppCallLegSet;
    /* Operation specific Variable declarations End */

    get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

    u_octet4_loop_org_csapi_cc_mmccs_TpAppMultiMediaCallLegCallBack_AppCallLegSet = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_mmccs_TpAppMultiMediaCallLegCallBack_AppCallLegSet_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_mmccs_TpAppMultiMediaCallLegCallBack_AppCallLegSet);

    for (i_org_csapi_cc_mmccs_TpAppMultiMediaCallLegCallBack_AppCallLegSet=0; i_org_csapi_cc_mmccs_TpAppMultiMediaCallLegCallBack_AppCallLegSet < u_octet4_loop_org_csapi_cc_mmccs_TpAppMultiMediaCallLegCallBack_AppCallLegSet; i_org_csapi_cc_mmccs_TpAppMultiMediaCallLegCallBack_AppCallLegSet++) {

        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

    }

}

/* Struct = IDL:org/csapi/cc/cccs/TpMonoMediaConfPolicy:1.0 */
static void
decode_org_csapi_cc_cccs_TpMonoMediaConfPolicy_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    proto_tree_add_boolean(tree, hf_org_csapi_cc_cccs_TpMonoMediaConfPolicy_JoinAllowed, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

}

/* Struct = IDL:org/csapi/cc/cccs/TpJoinEventInfo:1.0 */
static void
decode_org_csapi_cc_cccs_TpJoinEventInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cc_cccs_TpJoinEventInfo_CallAppInfo;
    guint32   i_org_csapi_cc_cccs_TpJoinEventInfo_CallAppInfo;
    /* Operation specific Variable declarations End */

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    u_octet4_loop_org_csapi_cc_cccs_TpJoinEventInfo_CallAppInfo = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_cccs_TpJoinEventInfo_CallAppInfo_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_cccs_TpJoinEventInfo_CallAppInfo);

    for (i_org_csapi_cc_cccs_TpJoinEventInfo_CallAppInfo=0; i_org_csapi_cc_cccs_TpJoinEventInfo_CallAppInfo < u_octet4_loop_org_csapi_cc_cccs_TpJoinEventInfo_CallAppInfo; i_org_csapi_cc_cccs_TpJoinEventInfo_CallAppInfo++) {

        /*  Begin union "org_csapi_cc_TpCallAppInfo"  */
        decode_org_csapi_cc_TpCallAppInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_cc_TpCallAppInfo"  */
    }

}

/* Struct = IDL:org/csapi/cc/cccs/TpConfSearchCriteria:1.0 */
static void
decode_org_csapi_cc_cccs_TpConfSearchCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_cccs_TpConfSearchCriteria_StartSearch);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_cccs_TpConfSearchCriteria_StopSearch);

    proto_tree_add_int(tree, hf_org_csapi_cc_cccs_TpConfSearchCriteria_RequestedResources, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_cc_cccs_TpConfSearchCriteria_RequestedDuration, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cc/cccs/TpConfSearchResult:1.0 */
static void
decode_org_csapi_cc_cccs_TpConfSearchResult_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    proto_tree_add_boolean(tree, hf_org_csapi_cc_cccs_TpConfSearchResult_MatchFound, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_cccs_TpConfSearchResult_ActualStartTime);

    proto_tree_add_int(tree, hf_org_csapi_cc_cccs_TpConfSearchResult_ActualResources, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_cc_cccs_TpConfSearchResult_ActualDuration, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cc/cccs/TpMultiMediaConfPolicy:1.0 */
static void
decode_org_csapi_cc_cccs_TpMultiMediaConfPolicy_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    proto_tree_add_boolean(tree, hf_org_csapi_cc_cccs_TpMultiMediaConfPolicy_JoinAllowed, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    proto_tree_add_int(tree, hf_org_csapi_cc_cccs_TpMultiMediaConfPolicy_MediaAllowed, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_boolean(tree, hf_org_csapi_cc_cccs_TpMultiMediaConfPolicy_Chaired, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_cccs_TpMultiMediaConfPolicy_VideoHandling, tvb, *offset-4, 4, u_octet4);

}

/* Struct = IDL:org/csapi/cc/cccs/TpResourceReservation:1.0 */
static void
decode_org_csapi_cc_cccs_TpResourceReservation_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    proto_tree_add_int(tree, hf_org_csapi_cc_cccs_TpResourceReservation_ReservationID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cc/cccs/TpConfCallIdentifier:1.0 */
static void
decode_org_csapi_cc_cccs_TpConfCallIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

    proto_tree_add_int(tree, hf_org_csapi_cc_cccs_TpConfCallIdentifier_ConfCallSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cc/cccs/TpSubConfCallIdentifier:1.0 */
static void
decode_org_csapi_cc_cccs_TpSubConfCallIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

    proto_tree_add_int(tree, hf_org_csapi_cc_cccs_TpSubConfCallIdentifier_SubConfCallSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/ui/TpUIIdentifier:1.0 */
static void
decode_org_csapi_ui_TpUIIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

    proto_tree_add_int(tree, hf_org_csapi_ui_TpUIIdentifier_UserInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/ui/TpUICallIdentifier:1.0 */
static void
decode_org_csapi_ui_TpUICallIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

    proto_tree_add_int(tree, hf_org_csapi_ui_TpUICallIdentifier_UserInteractionSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/mm/TpLocationResponseTime:1.0 */
static void
decode_org_csapi_mm_TpLocationResponseTime_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpLocationResponseTime_ResponseTime, tvb, *offset-4, 4, u_octet4);

    proto_tree_add_int(tree, hf_org_csapi_mm_TpLocationResponseTime_TimerValue, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/mm/TpLocationTrigger:1.0 */
static void
decode_org_csapi_mm_TpLocationTrigger_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    proto_tree_add_float(tree, hf_org_csapi_mm_TpLocationTrigger_Longitude, tvb, *offset-4, 4, get_CDR_float(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_float(tree, hf_org_csapi_mm_TpLocationTrigger_Latitude, tvb, *offset-4, 4, get_CDR_float(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_float(tree, hf_org_csapi_mm_TpLocationTrigger_AreaSemiMajor, tvb, *offset-4, 4, get_CDR_float(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_float(tree, hf_org_csapi_mm_TpLocationTrigger_AreaSemiMinor, tvb, *offset-4, 4, get_CDR_float(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_mm_TpLocationTrigger_AngleOfSemiMajor, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpLocationTrigger_Criterion, tvb, *offset-4, 4, u_octet4);

    proto_tree_add_int(tree, hf_org_csapi_mm_TpLocationTrigger_ReportingInterval, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/mm/TpLocationRequest:1.0 */
static void
decode_org_csapi_mm_TpLocationRequest_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    proto_tree_add_float(tree, hf_org_csapi_mm_TpLocationRequest_RequestedAccuracy, tvb, *offset-4, 4, get_CDR_float(tvb,offset,stream_is_big_endian, boundary));

    /*  Begin struct "org_csapi_mm_TpLocationResponseTime"  */
    decode_org_csapi_mm_TpLocationResponseTime_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_mm_TpLocationResponseTime"  */
    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpLocationRequest_AltitudeRequested, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpLocationRequest_Type, tvb, *offset-4, 4, u_octet4);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpLocationRequest_Priority, tvb, *offset-4, 4, u_octet4);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_TpLocationRequest_RequestedLocationMethod);

}

/* Struct = IDL:org/csapi/mm/TpGeographicalPosition:1.0 */
static void
decode_org_csapi_mm_TpGeographicalPosition_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    proto_tree_add_float(tree, hf_org_csapi_mm_TpGeographicalPosition_Longitude, tvb, *offset-4, 4, get_CDR_float(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_float(tree, hf_org_csapi_mm_TpGeographicalPosition_Latitude, tvb, *offset-4, 4, get_CDR_float(tvb,offset,stream_is_big_endian, boundary));

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpGeographicalPosition_TypeOfUncertaintyShape, tvb, *offset-4, 4, u_octet4);

    proto_tree_add_float(tree, hf_org_csapi_mm_TpGeographicalPosition_UncertaintyInnerSemiMajor, tvb, *offset-4, 4, get_CDR_float(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_float(tree, hf_org_csapi_mm_TpGeographicalPosition_UncertaintyOuterSemiMajor, tvb, *offset-4, 4, get_CDR_float(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_float(tree, hf_org_csapi_mm_TpGeographicalPosition_UncertaintyInnerSemiMinor, tvb, *offset-4, 4, get_CDR_float(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_float(tree, hf_org_csapi_mm_TpGeographicalPosition_UncertaintyOuterSemiMinor, tvb, *offset-4, 4, get_CDR_float(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_mm_TpGeographicalPosition_AngleOfSemiMajor, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_mm_TpGeographicalPosition_SegmentStartAngle, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_mm_TpGeographicalPosition_SegmentEndAngle, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/mm/TpMobilityStopAssignmentData:1.0 */
static void
decode_org_csapi_mm_TpMobilityStopAssignmentData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    guint32   u_octet4_loop_org_csapi_mm_TpMobilityStopAssignmentData_Users;
    guint32   i_org_csapi_mm_TpMobilityStopAssignmentData_Users;
    /* Operation specific Variable declarations End */

    proto_tree_add_int(tree, hf_org_csapi_mm_TpMobilityStopAssignmentData_AssignmentId, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpMobilityStopAssignmentData_StopScope, tvb, *offset-4, 4, u_octet4);

    u_octet4_loop_org_csapi_mm_TpMobilityStopAssignmentData_Users = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpMobilityStopAssignmentData_Users_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_TpMobilityStopAssignmentData_Users);

    for (i_org_csapi_mm_TpMobilityStopAssignmentData_Users=0; i_org_csapi_mm_TpMobilityStopAssignmentData_Users < u_octet4_loop_org_csapi_mm_TpMobilityStopAssignmentData_Users; i_org_csapi_mm_TpMobilityStopAssignmentData_Users++) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
    }

}

/* Struct = IDL:org/csapi/mm/TpUlExtendedData:1.0 */
static void
decode_org_csapi_mm_TpUlExtendedData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    /*  Begin struct "org_csapi_mm_TpGeographicalPosition"  */
    decode_org_csapi_mm_TpGeographicalPosition_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_mm_TpGeographicalPosition"  */
    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpUlExtendedData_TerminalType, tvb, *offset-4, 4, u_octet4);

    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpUlExtendedData_AltitudePresent, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    proto_tree_add_float(tree, hf_org_csapi_mm_TpUlExtendedData_Altitude, tvb, *offset-4, 4, get_CDR_float(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_float(tree, hf_org_csapi_mm_TpUlExtendedData_UncertaintyAltitude, tvb, *offset-4, 4, get_CDR_float(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpUlExtendedData_TimestampPresent, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_TpUlExtendedData_Timestamp);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_TpUlExtendedData_UsedLocationMethod);

}

/* Struct = IDL:org/csapi/mm/TpUserLocation:1.0 */
static void
decode_org_csapi_mm_TpUserLocation_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpUserLocation_StatusCode, tvb, *offset-4, 4, u_octet4);

    /*  Begin struct "org_csapi_mm_TpGeographicalPosition"  */
    decode_org_csapi_mm_TpGeographicalPosition_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_mm_TpGeographicalPosition"  */
}

/* Struct = IDL:org/csapi/mm/TpUserLocationExtended:1.0 */
static void
decode_org_csapi_mm_TpUserLocationExtended_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    guint32   u_octet4_loop_org_csapi_mm_TpUserLocationExtended_Locations;
    guint32   i_org_csapi_mm_TpUserLocationExtended_Locations;
    /* Operation specific Variable declarations End */

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpUserLocationExtended_StatusCode, tvb, *offset-4, 4, u_octet4);

    u_octet4_loop_org_csapi_mm_TpUserLocationExtended_Locations = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpUserLocationExtended_Locations_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_TpUserLocationExtended_Locations);

    for (i_org_csapi_mm_TpUserLocationExtended_Locations=0; i_org_csapi_mm_TpUserLocationExtended_Locations < u_octet4_loop_org_csapi_mm_TpUserLocationExtended_Locations; i_org_csapi_mm_TpUserLocationExtended_Locations++) {

        /*  Begin struct "org_csapi_mm_TpUlExtendedData"  */
        decode_org_csapi_mm_TpUlExtendedData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mm_TpUlExtendedData"  */
    }

}

/* Struct = IDL:org/csapi/mm/TpLocationTriggerCamel:1.0 */
static void
decode_org_csapi_mm_TpLocationTriggerCamel_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpLocationTriggerCamel_UpdateInsideVlr, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpLocationTriggerCamel_UpdateOutsideVlr, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

}

/* Struct = IDL:org/csapi/mm/TpUserLocationCamel:1.0 */
static void
decode_org_csapi_mm_TpUserLocationCamel_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpUserLocationCamel_StatusCode, tvb, *offset-4, 4, u_octet4);

    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpUserLocationCamel_GeographicalPositionPresent, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    /*  Begin struct "org_csapi_mm_TpGeographicalPosition"  */
    decode_org_csapi_mm_TpGeographicalPosition_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_mm_TpGeographicalPosition"  */
    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpUserLocationCamel_TimestampPresent, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_TpUserLocationCamel_Timestamp);

    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpUserLocationCamel_VlrNumberPresent, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpUserLocationCamel_LocationNumberPresent, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpUserLocationCamel_CellIdOrLaiPresent, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_TpUserLocationCamel_CellIdOrLai);

}

/* Struct = IDL:org/csapi/mm/TpUserLocationEmergencyRequest:1.0 */
static void
decode_org_csapi_mm_TpUserLocationEmergencyRequest_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpUserLocationEmergencyRequest_UserAddressPresent, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpUserLocationEmergencyRequest_NaEsrdPresent, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_TpUserLocationEmergencyRequest_NaEsrd);

    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpUserLocationEmergencyRequest_NaEsrkPresent, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_TpUserLocationEmergencyRequest_NaEsrk);

    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpUserLocationEmergencyRequest_ImeiPresent, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_TpUserLocationEmergencyRequest_Imei);

    /*  Begin struct "org_csapi_mm_TpLocationRequest"  */
    decode_org_csapi_mm_TpLocationRequest_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_mm_TpLocationRequest"  */
}

/* Struct = IDL:org/csapi/mm/TpUserLocationEmergency:1.0 */
static void
decode_org_csapi_mm_TpUserLocationEmergency_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpUserLocationEmergency_StatusCode, tvb, *offset-4, 4, u_octet4);

    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpUserLocationEmergency_UserIdPresent, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpUserLocationEmergency_NaEsrdPresent, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_TpUserLocationEmergency_NaEsrd);

    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpUserLocationEmergency_NaEsrkPresent, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_TpUserLocationEmergency_NaEsrk);

    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpUserLocationEmergency_ImeiPresent, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_TpUserLocationEmergency_Imei);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpUserLocationEmergency_TriggeringEvent, tvb, *offset-4, 4, u_octet4);

    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpUserLocationEmergency_GeographicalPositionPresent, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    /*  Begin struct "org_csapi_mm_TpGeographicalPosition"  */
    decode_org_csapi_mm_TpGeographicalPosition_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_mm_TpGeographicalPosition"  */
    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpUserLocationEmergency_AltitudePresent, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    proto_tree_add_float(tree, hf_org_csapi_mm_TpUserLocationEmergency_Altitude, tvb, *offset-4, 4, get_CDR_float(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_float(tree, hf_org_csapi_mm_TpUserLocationEmergency_UncertaintyAltitude, tvb, *offset-4, 4, get_CDR_float(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpUserLocationEmergency_TimestampPresent, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_TpUserLocationEmergency_Timestamp);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_TpUserLocationEmergency_UsedLocationMethod);

}

/* Struct = IDL:org/csapi/mm/TpUserStatus:1.0 */
static void
decode_org_csapi_mm_TpUserStatus_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpUserStatus_StatusCode, tvb, *offset-4, 4, u_octet4);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpUserStatus_Status, tvb, *offset-4, 4, u_octet4);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpUserStatus_TerminalType, tvb, *offset-4, 4, u_octet4);

}

/* Struct = IDL:org/csapi/mm/TpTriggeredStatusRequest:1.0 */
static void
decode_org_csapi_mm_TpTriggeredStatusRequest_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_TpTriggeredStatusRequest_Users;
    guint32   i_org_csapi_mm_TpTriggeredStatusRequest_Users;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_mm_TpTriggeredStatusRequest_Users = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpTriggeredStatusRequest_Users_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_TpTriggeredStatusRequest_Users);

    for (i_org_csapi_mm_TpTriggeredStatusRequest_Users=0; i_org_csapi_mm_TpTriggeredStatusRequest_Users < u_octet4_loop_org_csapi_mm_TpTriggeredStatusRequest_Users; i_org_csapi_mm_TpTriggeredStatusRequest_Users++) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
    }

    proto_tree_add_int(tree, hf_org_csapi_mm_TpTriggeredStatusRequest_AssignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/mm/TpTriggeredStatusRequestSetEntry:1.0 */
static void
decode_org_csapi_mm_TpTriggeredStatusRequestSetEntry_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_TpTriggeredStatusRequestSetEntry_Requests;
    guint32   i_org_csapi_mm_TpTriggeredStatusRequestSetEntry_Requests;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_mm_TpTriggeredStatusRequestSetEntry_Requests = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpTriggeredStatusRequestSetEntry_Requests_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_TpTriggeredStatusRequestSetEntry_Requests);

    for (i_org_csapi_mm_TpTriggeredStatusRequestSetEntry_Requests=0; i_org_csapi_mm_TpTriggeredStatusRequestSetEntry_Requests < u_octet4_loop_org_csapi_mm_TpTriggeredStatusRequestSetEntry_Requests; i_org_csapi_mm_TpTriggeredStatusRequestSetEntry_Requests++) {

        /*  Begin struct "org_csapi_mm_TpTriggeredStatusRequest"  */
        decode_org_csapi_mm_TpTriggeredStatusRequest_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mm_TpTriggeredStatusRequest"  */
    }

    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpTriggeredStatusRequestSetEntry_Final, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

}

/* Struct = IDL:org/csapi/mm/TpPeriodicLocationRequest:1.0 */
static void
decode_org_csapi_mm_TpPeriodicLocationRequest_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_TpPeriodicLocationRequest_Users;
    guint32   i_org_csapi_mm_TpPeriodicLocationRequest_Users;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_mm_TpPeriodicLocationRequest_Users = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpPeriodicLocationRequest_Users_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_TpPeriodicLocationRequest_Users);

    for (i_org_csapi_mm_TpPeriodicLocationRequest_Users=0; i_org_csapi_mm_TpPeriodicLocationRequest_Users < u_octet4_loop_org_csapi_mm_TpPeriodicLocationRequest_Users; i_org_csapi_mm_TpPeriodicLocationRequest_Users++) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
    }

    /*  Begin struct "org_csapi_mm_TpLocationRequest"  */
    decode_org_csapi_mm_TpLocationRequest_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_mm_TpLocationRequest"  */
    proto_tree_add_int(tree, hf_org_csapi_mm_TpPeriodicLocationRequest_ReportingInterval, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_mm_TpPeriodicLocationRequest_AssignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/mm/TpPeriodicLocationRequestSetEntry:1.0 */
static void
decode_org_csapi_mm_TpPeriodicLocationRequestSetEntry_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_TpPeriodicLocationRequestSetEntry_Requests;
    guint32   i_org_csapi_mm_TpPeriodicLocationRequestSetEntry_Requests;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_mm_TpPeriodicLocationRequestSetEntry_Requests = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpPeriodicLocationRequestSetEntry_Requests_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_TpPeriodicLocationRequestSetEntry_Requests);

    for (i_org_csapi_mm_TpPeriodicLocationRequestSetEntry_Requests=0; i_org_csapi_mm_TpPeriodicLocationRequestSetEntry_Requests < u_octet4_loop_org_csapi_mm_TpPeriodicLocationRequestSetEntry_Requests; i_org_csapi_mm_TpPeriodicLocationRequestSetEntry_Requests++) {

        /*  Begin struct "org_csapi_mm_TpPeriodicLocationRequest"  */
        decode_org_csapi_mm_TpPeriodicLocationRequest_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mm_TpPeriodicLocationRequest"  */
    }

    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpPeriodicLocationRequestSetEntry_Final, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

}

/* Struct = IDL:org/csapi/mm/TpTriggeredLocationRequest:1.0 */
static void
decode_org_csapi_mm_TpTriggeredLocationRequest_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_TpTriggeredLocationRequest_Users;
    guint32   i_org_csapi_mm_TpTriggeredLocationRequest_Users;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_mm_TpTriggeredLocationRequest_Users = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpTriggeredLocationRequest_Users_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_TpTriggeredLocationRequest_Users);

    for (i_org_csapi_mm_TpTriggeredLocationRequest_Users=0; i_org_csapi_mm_TpTriggeredLocationRequest_Users < u_octet4_loop_org_csapi_mm_TpTriggeredLocationRequest_Users; i_org_csapi_mm_TpTriggeredLocationRequest_Users++) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
    }

    /*  Begin struct "org_csapi_mm_TpLocationRequest"  */
    decode_org_csapi_mm_TpLocationRequest_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_mm_TpLocationRequest"  */
    proto_tree_add_int(tree, hf_org_csapi_mm_TpTriggeredLocationRequest_AssignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/mm/TpTriggeredLocationRequestSetEntry:1.0 */
static void
decode_org_csapi_mm_TpTriggeredLocationRequestSetEntry_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_TpTriggeredLocationRequestSetEntry_Requests;
    guint32   i_org_csapi_mm_TpTriggeredLocationRequestSetEntry_Requests;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_mm_TpTriggeredLocationRequestSetEntry_Requests = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpTriggeredLocationRequestSetEntry_Requests_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_TpTriggeredLocationRequestSetEntry_Requests);

    for (i_org_csapi_mm_TpTriggeredLocationRequestSetEntry_Requests=0; i_org_csapi_mm_TpTriggeredLocationRequestSetEntry_Requests < u_octet4_loop_org_csapi_mm_TpTriggeredLocationRequestSetEntry_Requests; i_org_csapi_mm_TpTriggeredLocationRequestSetEntry_Requests++) {

        /*  Begin struct "org_csapi_mm_TpTriggeredLocationRequest"  */
        decode_org_csapi_mm_TpTriggeredLocationRequest_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mm_TpTriggeredLocationRequest"  */
    }

    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpTriggeredLocationRequestSetEntry_Final, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

}

/* Struct = IDL:org/csapi/mm/TpUserStatusIndicatorExtended:1.0 */
static void
decode_org_csapi_mm_TpUserStatusIndicatorExtended_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpUserStatusIndicatorExtended_UserStatusIndicator, tvb, *offset-4, 4, u_octet4);

    proto_tree_add_boolean(tree, hf_org_csapi_mm_TpUserStatusIndicatorExtended_PDPContextActive, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

}

/* Struct = IDL:org/csapi/mm/TpUserInfo:1.0 */
static void
decode_org_csapi_mm_TpUserInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_TpUserInfo_UserName);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_TpUserInfo_Password);

}

/* Struct = IDL:org/csapi/mm/TpNetworkStatusIndicator:1.0 */
static void
decode_org_csapi_mm_TpNetworkStatusIndicator_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_TpNetworkStatusIndicator_CountryCode);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_TpNetworkStatusIndicator_MobileNetworkCode);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpNetworkStatusIndicator_AccessTechnology, tvb, *offset-4, 4, u_octet4);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpNetworkStatusIndicator_RoamingStatus, tvb, *offset-4, 4, u_octet4);

}

/* Struct = IDL:org/csapi/mm/TpUserStatusExtended:1.0 */
static void
decode_org_csapi_mm_TpUserStatusExtended_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_TpUserStatusExtended_UserID);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpUserStatusExtended_StatusCode, tvb, *offset-4, 4, u_octet4);

    /*  Begin struct "org_csapi_mm_TpUserStatusIndicatorExtended"  */
    decode_org_csapi_mm_TpUserStatusIndicatorExtended_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_mm_TpUserStatusIndicatorExtended"  */
    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpUserStatusExtended_TerminalType, tvb, *offset-4, 4, u_octet4);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpUserStatusExtended_AuthenticationStatus, tvb, *offset-4, 4, u_octet4);

    /*  Begin struct "org_csapi_mm_TpNetworkStatusIndicator"  */
    decode_org_csapi_mm_TpNetworkStatusIndicator_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_mm_TpNetworkStatusIndicator"  */
    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_TpUserStatusExtended_UserIPAddress);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_TpUserStatusExtended_UserMSISDN);

    /*  Begin struct "org_csapi_mm_TpUserInfo"  */
    decode_org_csapi_mm_TpUserInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_mm_TpUserInfo"  */
    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_TpUserStatusExtended_UserConnectionID);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mm_TpUserStatusExtended_AccessPointName);

}

/* Struct = IDL:org/csapi/mm/TpBindingNotificationCriteria:1.0 */
static void
decode_org_csapi_mm_TpBindingNotificationCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mm_TpBindingNotificationCriteria_ContactAddressSet;
    guint32   i_org_csapi_mm_TpBindingNotificationCriteria_ContactAddressSet;
    /* Operation specific Variable declarations End */

    proto_tree_add_int(tree, hf_org_csapi_mm_TpBindingNotificationCriteria_NotificationCriteriaType, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    u_octet4_loop_org_csapi_mm_TpBindingNotificationCriteria_ContactAddressSet = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpBindingNotificationCriteria_ContactAddressSet_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mm_TpBindingNotificationCriteria_ContactAddressSet);

    for (i_org_csapi_mm_TpBindingNotificationCriteria_ContactAddressSet=0; i_org_csapi_mm_TpBindingNotificationCriteria_ContactAddressSet < u_octet4_loop_org_csapi_mm_TpBindingNotificationCriteria_ContactAddressSet; i_org_csapi_mm_TpBindingNotificationCriteria_ContactAddressSet++) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
    }

}

/* Struct = IDL:org/csapi/termcap/TpTerminalCapabilities:1.0 */
static void
decode_org_csapi_termcap_TpTerminalCapabilities_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_termcap_TpTerminalCapabilities_TerminalCapabilities);

    proto_tree_add_boolean(tree, hf_org_csapi_termcap_TpTerminalCapabilities_StatusCode, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

}

/* Struct = IDL:org/csapi/termcap/TpTerminalCapabilityScope:1.0 */
static void
decode_org_csapi_termcap_TpTerminalCapabilityScope_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_termcap_TpTerminalCapabilityScope_ScopeType, tvb, *offset-4, 4, u_octet4);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_termcap_TpTerminalCapabilityScope_Scope);

}

/* Struct = IDL:org/csapi/dsc/TpChargePerVolume:1.0 */
static void
decode_org_csapi_dsc_TpChargePerVolume_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    proto_tree_add_int(tree, hf_org_csapi_dsc_TpChargePerVolume_InitialCharge, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_dsc_TpChargePerVolume_CurrentChargePerKilobyte, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_dsc_TpChargePerVolume_NextChargePerKilobyte, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/dsc/TpDataSessionChargePlan:1.0 */
static void
decode_org_csapi_dsc_TpDataSessionChargePlan_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin union "org_csapi_dsc_TpDataSessionChargeOrder"  */
    decode_org_csapi_dsc_TpDataSessionChargeOrder_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

    /*  End union "org_csapi_dsc_TpDataSessionChargeOrder"  */
    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_dsc_TpDataSessionChargePlan_Currency);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_dsc_TpDataSessionChargePlan_AdditionalInfo);

}

/* Struct = IDL:org/csapi/dsc/TpDataSessionError:1.0 */
static void
decode_org_csapi_dsc_TpDataSessionError_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_dsc_TpDataSessionError_ErrorTime);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_dsc_TpDataSessionError_ErrorType, tvb, *offset-4, 4, u_octet4);

    /*  Begin union "org_csapi_dsc_TpDataSessionAdditionalErrorInfo"  */
    decode_org_csapi_dsc_TpDataSessionAdditionalErrorInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

    /*  End union "org_csapi_dsc_TpDataSessionAdditionalErrorInfo"  */
}

/* Struct = IDL:org/csapi/dsc/TpDataSessionEventCriteria:1.0 */
static void
decode_org_csapi_dsc_TpDataSessionEventCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    /*  Begin struct "org_csapi_TpAddressRange"  */
    decode_org_csapi_TpAddressRange_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddressRange"  */
    /*  Begin struct "org_csapi_TpAddressRange"  */
    decode_org_csapi_TpAddressRange_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddressRange"  */
    proto_tree_add_int(tree, hf_org_csapi_dsc_TpDataSessionEventCriteria_DataSessionEventName, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_dsc_TpDataSessionEventCriteria_MonitorMode, tvb, *offset-4, 4, u_octet4);

}

/* Struct = IDL:org/csapi/dsc/TpDataSessionEventInfo:1.0 */
static void
decode_org_csapi_dsc_TpDataSessionEventInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    proto_tree_add_int(tree, hf_org_csapi_dsc_TpDataSessionEventInfo_DataSessionEventName, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_dsc_TpDataSessionEventInfo_MonitorMode, tvb, *offset-4, 4, u_octet4);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_dsc_TpDataSessionEventInfo_QoSClass, tvb, *offset-4, 4, u_octet4);

}

/* Struct = IDL:org/csapi/dsc/TpDataSessionReleaseCause:1.0 */
static void
decode_org_csapi_dsc_TpDataSessionReleaseCause_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    proto_tree_add_int(tree, hf_org_csapi_dsc_TpDataSessionReleaseCause_Value, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_dsc_TpDataSessionReleaseCause_Location, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/dsc/TpDataSessionReport:1.0 */
static void
decode_org_csapi_dsc_TpDataSessionReport_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_dsc_TpDataSessionReport_MonitorMode, tvb, *offset-4, 4, u_octet4);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_dsc_TpDataSessionReport_DataSessionEventTime);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_dsc_TpDataSessionReport_DataSessionReportType, tvb, *offset-4, 4, u_octet4);

    /*  Begin union "org_csapi_dsc_TpDataSessionAdditionalReportInfo"  */
    decode_org_csapi_dsc_TpDataSessionAdditionalReportInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

    /*  End union "org_csapi_dsc_TpDataSessionAdditionalReportInfo"  */
}

/* Struct = IDL:org/csapi/dsc/TpDataSessionReportRequest:1.0 */
static void
decode_org_csapi_dsc_TpDataSessionReportRequest_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_dsc_TpDataSessionReportRequest_MonitorMode, tvb, *offset-4, 4, u_octet4);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_dsc_TpDataSessionReportRequest_DataSessionReportType, tvb, *offset-4, 4, u_octet4);

}

/* Struct = IDL:org/csapi/dsc/TpDataSessionSuperviseVolume:1.0 */
static void
decode_org_csapi_dsc_TpDataSessionSuperviseVolume_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    proto_tree_add_int(tree, hf_org_csapi_dsc_TpDataSessionSuperviseVolume_VolumeQuantity, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_dsc_TpDataSessionSuperviseVolume_VolumeUnit, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/dsc/TpDataSessionEventCriteriaResult:1.0 */
static void
decode_org_csapi_dsc_TpDataSessionEventCriteriaResult_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_dsc_TpDataSessionEventCriteria"  */
    decode_org_csapi_dsc_TpDataSessionEventCriteria_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_dsc_TpDataSessionEventCriteria"  */
    proto_tree_add_int(tree, hf_org_csapi_dsc_TpDataSessionEventCriteriaResult_AssignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/dsc/TpDataSessionIdentifier:1.0 */
static void
decode_org_csapi_dsc_TpDataSessionIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

    proto_tree_add_int(tree, hf_org_csapi_dsc_TpDataSessionIdentifier_DataSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/gms/TpGMSNewMessageArrivedInfo:1.0 */
static void
decode_org_csapi_gms_TpGMSNewMessageArrivedInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_TpGMSNewMessageArrivedInfo_FolderID);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_TpGMSNewMessageArrivedInfo_MessageID);

    proto_tree_add_int(tree, hf_org_csapi_gms_TpGMSNewMessageArrivedInfo_NumberOfProperties, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/gms/TpGMSNewMessageArrivedCriteria:1.0 */
static void
decode_org_csapi_gms_TpGMSNewMessageArrivedCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_TpGMSNewMessageArrivedCriteria_AuthenticationInfo);

}

/* Struct = IDL:org/csapi/gms/TpMailboxIdentifier:1.0 */
static void
decode_org_csapi_gms_TpMailboxIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

    proto_tree_add_int(tree, hf_org_csapi_gms_TpMailboxIdentifier_SessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/gms/TpMailboxFolderIdentifier:1.0 */
static void
decode_org_csapi_gms_TpMailboxFolderIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

    proto_tree_add_int(tree, hf_org_csapi_gms_TpMailboxFolderIdentifier_SessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cm/TpNameDescrpTagTimePeriod:1.0 */
static void
decode_org_csapi_cm_TpNameDescrpTagTimePeriod_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    proto_tree_add_int(tree, hf_org_csapi_cm_TpNameDescrpTagTimePeriod_duration, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cm/TpNameDescrpTagTimeOfDay:1.0 */
static void
decode_org_csapi_cm_TpNameDescrpTagTimeOfDay_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpNameDescrpTagTimeOfDay_name);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpNameDescrpTagTimeOfDay_description);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cm_TpNameDescrpTagTimeOfDay_tag, tvb, *offset-4, 4, u_octet4);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpNameDescrpTagTimeOfDay_value);

}

/* Struct = IDL:org/csapi/cm/TpNameDescrpTagString:1.0 */
static void
decode_org_csapi_cm_TpNameDescrpTagString_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpNameDescrpTagString_name);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpNameDescrpTagString_description);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cm_TpNameDescrpTagString_tag, tvb, *offset-4, 4, u_octet4);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpNameDescrpTagString_value);

}

/* Struct = IDL:org/csapi/cm/TpNameDescrpTagMonth:1.0 */
static void
decode_org_csapi_cm_TpNameDescrpTagMonth_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpNameDescrpTagMonth_name);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpNameDescrpTagMonth_description);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cm_TpNameDescrpTagMonth_tag, tvb, *offset-4, 4, u_octet4);

    proto_tree_add_int(tree, hf_org_csapi_cm_TpNameDescrpTagMonth_value, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cm/TpNameDescrpTagInt:1.0 */
static void
decode_org_csapi_cm_TpNameDescrpTagInt_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpNameDescrpTagInt_name);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpNameDescrpTagInt_description);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cm_TpNameDescrpTagInt_tag, tvb, *offset-4, 4, u_octet4);

    proto_tree_add_int(tree, hf_org_csapi_cm_TpNameDescrpTagInt_value, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cm/TpNameDescrpTagDir:1.0 */
static void
decode_org_csapi_cm_TpNameDescrpTagDir_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpNameDescrpTagDir_name);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpNameDescrpTagDir_description);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cm_TpNameDescrpTagDir_tag, tvb, *offset-4, 4, u_octet4);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cm_TpNameDescrpTagDir_value, tvb, *offset-4, 4, u_octet4);

}

/* Struct = IDL:org/csapi/cm/TpNameDescrpTagDayOfWeek:1.0 */
static void
decode_org_csapi_cm_TpNameDescrpTagDayOfWeek_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpNameDescrpTagDayOfWeek_name);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpNameDescrpTagDayOfWeek_description);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cm_TpNameDescrpTagDayOfWeek_tag, tvb, *offset-4, 4, u_octet4);

    proto_tree_add_int(tree, hf_org_csapi_cm_TpNameDescrpTagDayOfWeek_value, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cm/TpNameDescrpTagDateTime:1.0 */
static void
decode_org_csapi_cm_TpNameDescrpTagDateTime_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpNameDescrpTagDateTime_name);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpNameDescrpTagDateTime_description);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cm_TpNameDescrpTagDateTime_tag, tvb, *offset-4, 4, u_octet4);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpNameDescrpTagDateTime_value);

}

/* Struct = IDL:org/csapi/cm/TpLossDescriptor:1.0 */
static void
decode_org_csapi_cm_TpLossDescriptor_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_cm_TpNameDescrpTagInt"  */
    decode_org_csapi_cm_TpNameDescrpTagInt_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagInt"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagInt"  */
    decode_org_csapi_cm_TpNameDescrpTagInt_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagInt"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagInt"  */
    decode_org_csapi_cm_TpNameDescrpTagInt_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagInt"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagInt"  */
    decode_org_csapi_cm_TpNameDescrpTagInt_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagInt"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagInt"  */
    decode_org_csapi_cm_TpNameDescrpTagInt_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagInt"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagString"  */
    decode_org_csapi_cm_TpNameDescrpTagString_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagString"  */
}

/* Struct = IDL:org/csapi/cm/TpLoadDescriptor:1.0 */
static void
decode_org_csapi_cm_TpLoadDescriptor_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_cm_TpNameDescrpTagInt"  */
    decode_org_csapi_cm_TpNameDescrpTagInt_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagInt"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagInt"  */
    decode_org_csapi_cm_TpNameDescrpTagInt_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagInt"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagInt"  */
    decode_org_csapi_cm_TpNameDescrpTagInt_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagInt"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagInt"  */
    decode_org_csapi_cm_TpNameDescrpTagInt_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagInt"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagInt"  */
    decode_org_csapi_cm_TpNameDescrpTagInt_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagInt"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagInt"  */
    decode_org_csapi_cm_TpNameDescrpTagInt_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagInt"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagInt"  */
    decode_org_csapi_cm_TpNameDescrpTagInt_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagInt"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagString"  */
    decode_org_csapi_cm_TpNameDescrpTagString_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagString"  */
}

/* Struct = IDL:org/csapi/cm/TpJitterDescriptor:1.0 */
static void
decode_org_csapi_cm_TpJitterDescriptor_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_cm_TpNameDescrpTagInt"  */
    decode_org_csapi_cm_TpNameDescrpTagInt_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagInt"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagInt"  */
    decode_org_csapi_cm_TpNameDescrpTagInt_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagInt"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagInt"  */
    decode_org_csapi_cm_TpNameDescrpTagInt_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagInt"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagInt"  */
    decode_org_csapi_cm_TpNameDescrpTagInt_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagInt"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagInt"  */
    decode_org_csapi_cm_TpNameDescrpTagInt_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagInt"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagString"  */
    decode_org_csapi_cm_TpNameDescrpTagString_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagString"  */
}

/* Struct = IDL:org/csapi/cm/TpEndpoint:1.0 */
static void
decode_org_csapi_cm_TpEndpoint_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cm_TpEndpoint_type, tvb, *offset-4, 4, u_octet4);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpEndpoint_id);

}

/* Struct = IDL:org/csapi/cm/TpDelayDescriptor:1.0 */
static void
decode_org_csapi_cm_TpDelayDescriptor_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_cm_TpNameDescrpTagInt"  */
    decode_org_csapi_cm_TpNameDescrpTagInt_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagInt"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagInt"  */
    decode_org_csapi_cm_TpNameDescrpTagInt_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagInt"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagInt"  */
    decode_org_csapi_cm_TpNameDescrpTagInt_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagInt"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagInt"  */
    decode_org_csapi_cm_TpNameDescrpTagInt_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagInt"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagInt"  */
    decode_org_csapi_cm_TpNameDescrpTagInt_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagInt"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagString"  */
    decode_org_csapi_cm_TpNameDescrpTagString_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagString"  */
}

/* Struct = IDL:org/csapi/cm/TpNameDescrpTagExcessLoadAction:1.0 */
static void
decode_org_csapi_cm_TpNameDescrpTagExcessLoadAction_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpNameDescrpTagExcessLoadAction_name);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpNameDescrpTagExcessLoadAction_description);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cm_TpNameDescrpTagExcessLoadAction_tag, tvb, *offset-4, 4, u_octet4);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cm_TpNameDescrpTagExcessLoadAction_value, tvb, *offset-4, 4, u_octet4);

}

/* Struct = IDL:org/csapi/cm/TpValidityInfo:1.0 */
static void
decode_org_csapi_cm_TpValidityInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_cm_TpNameDescrpTagDateTime"  */
    decode_org_csapi_cm_TpNameDescrpTagDateTime_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagDateTime"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagTimePeriod"  */
    decode_org_csapi_cm_TpNameDescrpTagTimePeriod_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagTimePeriod"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagTimeOfDay"  */
    decode_org_csapi_cm_TpNameDescrpTagTimeOfDay_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagTimeOfDay"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagTimePeriod"  */
    decode_org_csapi_cm_TpNameDescrpTagTimePeriod_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagTimePeriod"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagDayOfWeek"  */
    decode_org_csapi_cm_TpNameDescrpTagDayOfWeek_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagDayOfWeek"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagMonth"  */
    decode_org_csapi_cm_TpNameDescrpTagMonth_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagMonth"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagString"  */
    decode_org_csapi_cm_TpNameDescrpTagString_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagString"  */
}

/* Struct = IDL:org/csapi/cm/TpProvisionedQoSInfo:1.0 */
static void
decode_org_csapi_cm_TpProvisionedQoSInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_cm_TpDelayDescriptor"  */
    decode_org_csapi_cm_TpDelayDescriptor_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpDelayDescriptor"  */
    /*  Begin struct "org_csapi_cm_TpLossDescriptor"  */
    decode_org_csapi_cm_TpLossDescriptor_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpLossDescriptor"  */
    /*  Begin struct "org_csapi_cm_TpJitterDescriptor"  */
    decode_org_csapi_cm_TpJitterDescriptor_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpJitterDescriptor"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagExcessLoadAction"  */
    decode_org_csapi_cm_TpNameDescrpTagExcessLoadAction_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagExcessLoadAction"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagString"  */
    decode_org_csapi_cm_TpNameDescrpTagString_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagString"  */
}

/* Struct = IDL:org/csapi/cm/TpPipeQoSInfo:1.0 */
static void
decode_org_csapi_cm_TpPipeQoSInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_cm_TpNameDescrpTagDir"  */
    decode_org_csapi_cm_TpNameDescrpTagDir_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagDir"  */
    /*  Begin struct "org_csapi_cm_TpEndpoint"  */
    decode_org_csapi_cm_TpEndpoint_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpEndpoint"  */
    /*  Begin struct "org_csapi_cm_TpEndpoint"  */
    decode_org_csapi_cm_TpEndpoint_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpEndpoint"  */
    /*  Begin struct "org_csapi_cm_TpLoadDescriptor"  */
    decode_org_csapi_cm_TpLoadDescriptor_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpLoadDescriptor"  */
    /*  Begin struct "org_csapi_cm_TpLoadDescriptor"  */
    decode_org_csapi_cm_TpLoadDescriptor_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpLoadDescriptor"  */
    /*  Begin struct "org_csapi_cm_TpNameDescrpTagString"  */
    decode_org_csapi_cm_TpNameDescrpTagString_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cm_TpNameDescrpTagString"  */
}

/* Struct = IDL:org/csapi/cm/TpDsCodepoint:1.0 */
static void
decode_org_csapi_cm_TpDsCodepoint_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpDsCodepoint_match);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpDsCodepoint_mask);

}

/* Struct = IDL:org/csapi/cm/TpIPSubnet:1.0 */
static void
decode_org_csapi_cm_TpIPSubnet_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpIPSubnet_subnetNumber);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cm_TpIPSubnet_subnetMask);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cm_TpIPSubnet_addressType, tvb, *offset-4, 4, u_octet4);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cm_TpIPSubnet_IPVersionSupport, tvb, *offset-4, 4, u_octet4);

}

/* Struct = IDL:org/csapi/am/TpBalanceInfo:1.0 */
static void
decode_org_csapi_am_TpBalanceInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_am_TpBalanceInfo_Currency);

    proto_tree_add_int(tree, hf_org_csapi_am_TpBalanceInfo_ValuePartA, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_am_TpBalanceInfo_ValuePartB, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_am_TpBalanceInfo_Exponent, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_am_TpBalanceInfo_AdditionalInfo);

}

/* Struct = IDL:org/csapi/am/TpChargingEventInfo:1.0 */
static void
decode_org_csapi_am_TpChargingEventInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_am_TpChargingEventInfo_ChargingEventName, tvb, *offset-4, 4, u_octet4);

    /*  Begin struct "org_csapi_am_TpBalanceInfo"  */
    decode_org_csapi_am_TpBalanceInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_am_TpBalanceInfo"  */
    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_am_TpChargingEventInfo_ChargingEventTime);

}

/* Struct = IDL:org/csapi/am/TpBalance:1.0 */
static void
decode_org_csapi_am_TpBalance_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_am_TpBalance_StatusCode, tvb, *offset-4, 4, u_octet4);

    /*  Begin struct "org_csapi_am_TpBalanceInfo"  */
    decode_org_csapi_am_TpBalanceInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_am_TpBalanceInfo"  */
}

/* Struct = IDL:org/csapi/am/TpTransactionHistory:1.0 */
static void
decode_org_csapi_am_TpTransactionHistory_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    proto_tree_add_int(tree, hf_org_csapi_am_TpTransactionHistory_TransactionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_am_TpTransactionHistory_TimeStamp);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_am_TpTransactionHistory_AdditionalInfo);

}

/* Struct = IDL:org/csapi/am/TpChargingEventCriteria:1.0 */
static void
decode_org_csapi_am_TpChargingEventCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_am_TpChargingEventCriteria_ChargingEvents;
    guint32   i_org_csapi_am_TpChargingEventCriteria_ChargingEvents;
    guint32   u_octet4;
    guint32   u_octet4_loop_org_csapi_am_TpChargingEventCriteria_Users;
    guint32   i_org_csapi_am_TpChargingEventCriteria_Users;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_am_TpChargingEventCriteria_ChargingEvents = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_am_TpChargingEventCriteria_ChargingEvents_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_am_TpChargingEventCriteria_ChargingEvents);

    for (i_org_csapi_am_TpChargingEventCriteria_ChargingEvents=0; i_org_csapi_am_TpChargingEventCriteria_ChargingEvents < u_octet4_loop_org_csapi_am_TpChargingEventCriteria_ChargingEvents; i_org_csapi_am_TpChargingEventCriteria_ChargingEvents++) {

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_am_TpChargingEventCriteria_ChargingEvents, tvb, *offset-4, 4, u_octet4);

    }

    u_octet4_loop_org_csapi_am_TpChargingEventCriteria_Users = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_am_TpChargingEventCriteria_Users_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_am_TpChargingEventCriteria_Users);

    for (i_org_csapi_am_TpChargingEventCriteria_Users=0; i_org_csapi_am_TpChargingEventCriteria_Users < u_octet4_loop_org_csapi_am_TpChargingEventCriteria_Users; i_org_csapi_am_TpChargingEventCriteria_Users++) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
    }

}

/* Struct = IDL:org/csapi/am/TpChargingEventCriteriaResult:1.0 */
static void
decode_org_csapi_am_TpChargingEventCriteriaResult_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_am_TpChargingEventCriteria"  */
    decode_org_csapi_am_TpChargingEventCriteria_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_am_TpChargingEventCriteria"  */
    proto_tree_add_int(tree, hf_org_csapi_am_TpChargingEventCriteriaResult_AssignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/am/TpBalanceExpiryDate:1.0 */
static void
decode_org_csapi_am_TpBalanceExpiryDate_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_am_TpBalanceExpiryDate_StatusCode, tvb, *offset-4, 4, u_octet4);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_am_TpBalanceExpiryDate_ExpiryDate);

}

/* Struct = IDL:org/csapi/am/TpVoucher:1.0 */
static void
decode_org_csapi_am_TpVoucher_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    proto_tree_add_int(tree, hf_org_csapi_am_TpVoucher_VoucherID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    /*  Begin struct "org_csapi_am_TpBalanceInfo"  */
    decode_org_csapi_am_TpBalanceInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_am_TpBalanceInfo"  */
}

/* Struct = IDL:org/csapi/cs/TpMerchantAccountID:1.0 */
static void
decode_org_csapi_cs_TpMerchantAccountID_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cs_TpMerchantAccountID_MerchantID);

    proto_tree_add_int(tree, hf_org_csapi_cs_TpMerchantAccountID_AccountID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cs/TpCorrelationID:1.0 */
static void
decode_org_csapi_cs_TpCorrelationID_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    proto_tree_add_int(tree, hf_org_csapi_cs_TpCorrelationID_CorrelationID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_cs_TpCorrelationID_CorrelationType, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cs/TpAmount:1.0 */
static void
decode_org_csapi_cs_TpAmount_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    proto_tree_add_int(tree, hf_org_csapi_cs_TpAmount_Number, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_cs_TpAmount_Exponent, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cs/TpChargingPrice:1.0 */
static void
decode_org_csapi_cs_TpChargingPrice_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cs_TpChargingPrice_Currency);

    /*  Begin struct "org_csapi_cs_TpAmount"  */
    decode_org_csapi_cs_TpAmount_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cs_TpAmount"  */
}

/* Struct = IDL:org/csapi/cs/TpVolume:1.0 */
static void
decode_org_csapi_cs_TpVolume_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_cs_TpAmount"  */
    decode_org_csapi_cs_TpAmount_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cs_TpAmount"  */
    proto_tree_add_int(tree, hf_org_csapi_cs_TpVolume_Unit, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cs/TpChargingSessionID:1.0 */
static void
decode_org_csapi_cs_TpChargingSessionID_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

    proto_tree_add_int(tree, hf_org_csapi_cs_TpChargingSessionID_ChargingSessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_int(tree, hf_org_csapi_cs_TpChargingSessionID_RequestNumberFirstRequest, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/cs/TpPriceVolume:1.0 */
static void
decode_org_csapi_cs_TpPriceVolume_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_cs_TpChargingPrice"  */
    decode_org_csapi_cs_TpChargingPrice_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cs_TpChargingPrice"  */
    /*  Begin struct "org_csapi_cs_TpVolume"  */
    decode_org_csapi_cs_TpVolume_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_cs_TpVolume"  */
}

/* Struct = IDL:org/csapi/cs/TpApplicationDescription:1.0 */
static void
decode_org_csapi_cs_TpApplicationDescription_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_cs_TpApplicationDescription_AppInformation;
    guint32   i_org_csapi_cs_TpApplicationDescription_AppInformation;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cs_TpApplicationDescription_Text);

    u_octet4_loop_org_csapi_cs_TpApplicationDescription_AppInformation = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cs_TpApplicationDescription_AppInformation_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_TpApplicationDescription_AppInformation);

    for (i_org_csapi_cs_TpApplicationDescription_AppInformation=0; i_org_csapi_cs_TpApplicationDescription_AppInformation < u_octet4_loop_org_csapi_cs_TpApplicationDescription_AppInformation; i_org_csapi_cs_TpApplicationDescription_AppInformation++) {

        /*  Begin union "org_csapi_cs_TpAppInformation"  */
        decode_org_csapi_cs_TpAppInformation_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_cs_TpAppInformation"  */
    }

}

/* Struct = IDL:org/csapi/cs/TpChargingParameter:1.0 */
static void
decode_org_csapi_cs_TpChargingParameter_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    proto_tree_add_int(tree, hf_org_csapi_cs_TpChargingParameter_ParameterID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    /*  Begin union "org_csapi_cs_TpChargingParameterValue"  */
    decode_org_csapi_cs_TpChargingParameterValue_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

    /*  End union "org_csapi_cs_TpChargingParameterValue"  */
}

/* Struct = IDL:org/csapi/policy/TpPolicyEvent:1.0 */
static void
decode_org_csapi_policy_TpPolicyEvent_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_TpPolicyEvent_Attributes;
    guint32   i_org_csapi_policy_TpPolicyEvent_Attributes;
    /* Operation specific Variable declarations End */

    proto_tree_add_int(tree, hf_org_csapi_policy_TpPolicyEvent_EventID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_TpPolicyEvent_TimeGenerated);

    u_octet4_loop_org_csapi_policy_TpPolicyEvent_Attributes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_policy_TpPolicyEvent_Attributes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_TpPolicyEvent_Attributes);

    for (i_org_csapi_policy_TpPolicyEvent_Attributes=0; i_org_csapi_policy_TpPolicyEvent_Attributes < u_octet4_loop_org_csapi_policy_TpPolicyEvent_Attributes; i_org_csapi_policy_TpPolicyEvent_Attributes++) {

        /*  Begin struct "org_csapi_TpAttribute"  */
        decode_org_csapi_TpAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAttribute"  */
    }

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_TpPolicyEvent_EventDefinitionName);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_TpPolicyEvent_EventDomainName);

}

/* Struct = IDL:org/csapi/policy/TpPolicyNameValue:1.0 */
static void
decode_org_csapi_policy_TpPolicyNameValue_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_TpPolicyNameValue_Name);

    get_CDR_any(tvb, pinfo, tree, item, offset, stream_is_big_endian, boundary, header);

}

/* Struct = IDL:org/csapi/policy/TpPolicyType/TpPolicyRecordType:1.0 */
static void
decode_org_csapi_policy_TpPolicyType_TpPolicyRecordType_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Names;
    guint32   i_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Names;
    guint32   u_octet4_loop_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Types;
    guint32   i_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Types;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Names = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Names_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Names);

    for (i_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Names=0; i_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Names < u_octet4_loop_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Names; i_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Names++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Names);

    }

    u_octet4_loop_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Types = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Types_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Types);

    for (i_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Types=0; i_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Types < u_octet4_loop_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Types; i_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Types++) {

        /*  Begin union "org_csapi_policy_TpPolicyType"  */
        decode_org_csapi_policy_TpPolicyType_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_policy_TpPolicyType"  */
    }

}

/* Struct = IDL:org/csapi/policy/TpPolicyType/TpPolicyListType:1.0 */
static void
decode_org_csapi_policy_TpPolicyType_TpPolicyListType_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_policy_TpPolicyType_TpPolicyListType_ElementType;
    guint32   i_org_csapi_policy_TpPolicyType_TpPolicyListType_ElementType;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_policy_TpPolicyType_TpPolicyListType_ElementType = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_policy_TpPolicyType_TpPolicyListType_ElementType_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_policy_TpPolicyType_TpPolicyListType_ElementType);

    for (i_org_csapi_policy_TpPolicyType_TpPolicyListType_ElementType=0; i_org_csapi_policy_TpPolicyType_TpPolicyListType_ElementType < u_octet4_loop_org_csapi_policy_TpPolicyType_TpPolicyListType_ElementType; i_org_csapi_policy_TpPolicyType_TpPolicyListType_ElementType++) {

        /*  Begin union "org_csapi_policy_TpPolicyType"  */
        decode_org_csapi_policy_TpPolicyType_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_policy_TpPolicyType"  */
    }

}

/* Struct = IDL:org/csapi/policy/TpPolicyVar:1.0 */
static void
decode_org_csapi_policy_TpPolicyVar_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_TpPolicyVar_VarName);

    /*  Begin union "org_csapi_policy_TpPolicyType"  */
    decode_org_csapi_policy_TpPolicyType_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

    /*  End union "org_csapi_policy_TpPolicyType"  */
    get_CDR_any(tvb, pinfo, tree, item, offset, stream_is_big_endian, boundary, header);

}

/* Struct = IDL:org/csapi/policy/TpPolicyConditionListElement:1.0 */
static void
decode_org_csapi_policy_TpPolicyConditionListElement_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

    proto_tree_add_int(tree, hf_org_csapi_policy_TpPolicyConditionListElement_GroupNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    proto_tree_add_boolean(tree, hf_org_csapi_policy_TpPolicyConditionListElement_Negated, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

}

/* Struct = IDL:org/csapi/policy/TpPolicyActionListElement:1.0 */
static void
decode_org_csapi_policy_TpPolicyActionListElement_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

    proto_tree_add_int(tree, hf_org_csapi_policy_TpPolicyActionListElement_SequenceNumber, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/pam/TpPAMAttributeDef:1.0 */
static void
decode_org_csapi_pam_TpPAMAttributeDef_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAttributeDef_Name);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAttributeDef_Type);

    proto_tree_add_boolean(tree, hf_org_csapi_pam_TpPAMAttributeDef_IsStatic, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    proto_tree_add_boolean(tree, hf_org_csapi_pam_TpPAMAttributeDef_IsRevertOnExpiration, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

    get_CDR_any(tvb, pinfo, tree, item, offset, stream_is_big_endian, boundary, header);

}

/* Struct = IDL:org/csapi/pam/TpPAMAttribute:1.0 */
static void
decode_org_csapi_pam_TpPAMAttribute_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAttribute_AttributeName);

    /*  Begin union "org_csapi_TpAttributeValue"  */
    decode_org_csapi_TpAttributeValue_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

    /*  End union "org_csapi_TpAttributeValue"  */
    proto_tree_add_int64(tree, hf_org_csapi_pam_TpPAMAttribute_ExpiresIn, tvb, *offset-8, 8, get_CDR_long_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/pam/TpPAMPresenceData:1.0 */
static void
decode_org_csapi_pam_TpPAMPresenceData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMPresenceData_Name);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMPresenceData_subscriberStatus);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMPresenceData_networkStatus);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMPresenceData_communicationMeans);

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMPresenceData_subscriberProvidedLocation);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMPresenceData_networkProvidedLocation);

    proto_tree_add_int(tree, hf_org_csapi_pam_TpPAMPresenceData_Priority, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMPresenceData_otherInfo);

}

/* Struct = IDL:org/csapi/pam/TpPAMAvailabilityProfile:1.0 */
static void
decode_org_csapi_pam_TpPAMAvailabilityProfile_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAvailabilityProfile_PrivacyCode);

    /*  Begin struct "org_csapi_pam_TpPAMPresenceData"  */
    decode_org_csapi_pam_TpPAMPresenceData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_pam_TpPAMPresenceData"  */
}

/* Struct = IDL:org/csapi/pam/TpPAMCommunicationContext:1.0 */
static void
decode_org_csapi_pam_TpPAMCommunicationContext_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMCommunicationContext_CommunicationCapability);

}

/* Struct = IDL:org/csapi/pam/TpPAMContext:1.0 */
static void
decode_org_csapi_pam_TpPAMContext_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMContext_AskerData;
    guint32   i_org_csapi_pam_TpPAMContext_AskerData;
    /* Operation specific Variable declarations End */

    /*  Begin union "org_csapi_pam_TpPAMContextData"  */
    decode_org_csapi_pam_TpPAMContextData_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

    /*  End union "org_csapi_pam_TpPAMContextData"  */
    u_octet4_loop_org_csapi_pam_TpPAMContext_AskerData = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMContext_AskerData_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMContext_AskerData);

    for (i_org_csapi_pam_TpPAMContext_AskerData=0; i_org_csapi_pam_TpPAMContext_AskerData < u_octet4_loop_org_csapi_pam_TpPAMContext_AskerData; i_org_csapi_pam_TpPAMContext_AskerData++) {

        /*  Begin struct "org_csapi_TpAttribute"  */
        decode_org_csapi_TpAttribute_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAttribute"  */
    }

}

/* Struct = IDL:org/csapi/pam/TpPAMAccessControlData:1.0 */
static void
decode_org_csapi_pam_TpPAMAccessControlData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMAccessControlData_AllowList;
    guint32   i_org_csapi_pam_TpPAMAccessControlData_AllowList;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMAccessControlData_DenyList;
    guint32   i_org_csapi_pam_TpPAMAccessControlData_DenyList;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMAccessControlData_DefaultPolicy, tvb, *offset-4, 4, u_octet4);

    u_octet4_loop_org_csapi_pam_TpPAMAccessControlData_AllowList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMAccessControlData_AllowList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMAccessControlData_AllowList);

    for (i_org_csapi_pam_TpPAMAccessControlData_AllowList=0; i_org_csapi_pam_TpPAMAccessControlData_AllowList < u_octet4_loop_org_csapi_pam_TpPAMAccessControlData_AllowList; i_org_csapi_pam_TpPAMAccessControlData_AllowList++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAccessControlData_AllowList);

    }

    u_octet4_loop_org_csapi_pam_TpPAMAccessControlData_DenyList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMAccessControlData_DenyList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMAccessControlData_DenyList);

    for (i_org_csapi_pam_TpPAMAccessControlData_DenyList=0; i_org_csapi_pam_TpPAMAccessControlData_DenyList < u_octet4_loop_org_csapi_pam_TpPAMAccessControlData_DenyList; i_org_csapi_pam_TpPAMAccessControlData_DenyList++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAccessControlData_DenyList);

    }

}

/* Struct = IDL:org/csapi/pam/TpPAMICEventData:1.0 */
static void
decode_org_csapi_pam_TpPAMICEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMICEventData_IdentityType;
    guint32   i_org_csapi_pam_TpPAMICEventData_IdentityType;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_pam_TpPAMICEventData_IdentityType = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMICEventData_IdentityType_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMICEventData_IdentityType);

    for (i_org_csapi_pam_TpPAMICEventData_IdentityType=0; i_org_csapi_pam_TpPAMICEventData_IdentityType < u_octet4_loop_org_csapi_pam_TpPAMICEventData_IdentityType; i_org_csapi_pam_TpPAMICEventData_IdentityType++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMICEventData_IdentityType);

    }

}

/* Struct = IDL:org/csapi/pam/TpPAMICNotificationData:1.0 */
static void
decode_org_csapi_pam_TpPAMICNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMICNotificationData_Identities;
    guint32   i_org_csapi_pam_TpPAMICNotificationData_Identities;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_pam_TpPAMICNotificationData_Identities = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMICNotificationData_Identities_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMICNotificationData_Identities);

    for (i_org_csapi_pam_TpPAMICNotificationData_Identities=0; i_org_csapi_pam_TpPAMICNotificationData_Identities < u_octet4_loop_org_csapi_pam_TpPAMICNotificationData_Identities; i_org_csapi_pam_TpPAMICNotificationData_Identities++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMICNotificationData_Identities);

    }

}

/* Struct = IDL:org/csapi/pam/TpPAMIDEventData:1.0 */
static void
decode_org_csapi_pam_TpPAMIDEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMIDEventData_IdentityName;
    guint32   i_org_csapi_pam_TpPAMIDEventData_IdentityName;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMIDEventData_IdentityType;
    guint32   i_org_csapi_pam_TpPAMIDEventData_IdentityType;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_pam_TpPAMIDEventData_IdentityName = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMIDEventData_IdentityName_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMIDEventData_IdentityName);

    for (i_org_csapi_pam_TpPAMIDEventData_IdentityName=0; i_org_csapi_pam_TpPAMIDEventData_IdentityName < u_octet4_loop_org_csapi_pam_TpPAMIDEventData_IdentityName; i_org_csapi_pam_TpPAMIDEventData_IdentityName++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMIDEventData_IdentityName);

    }

    u_octet4_loop_org_csapi_pam_TpPAMIDEventData_IdentityType = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMIDEventData_IdentityType_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMIDEventData_IdentityType);

    for (i_org_csapi_pam_TpPAMIDEventData_IdentityType=0; i_org_csapi_pam_TpPAMIDEventData_IdentityType < u_octet4_loop_org_csapi_pam_TpPAMIDEventData_IdentityType; i_org_csapi_pam_TpPAMIDEventData_IdentityType++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMIDEventData_IdentityType);

    }

}

/* Struct = IDL:org/csapi/pam/TpPAMIDNotificationData:1.0 */
static void
decode_org_csapi_pam_TpPAMIDNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMIDNotificationData_Identities;
    guint32   i_org_csapi_pam_TpPAMIDNotificationData_Identities;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_pam_TpPAMIDNotificationData_Identities = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMIDNotificationData_Identities_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMIDNotificationData_Identities);

    for (i_org_csapi_pam_TpPAMIDNotificationData_Identities=0; i_org_csapi_pam_TpPAMIDNotificationData_Identities < u_octet4_loop_org_csapi_pam_TpPAMIDNotificationData_Identities; i_org_csapi_pam_TpPAMIDNotificationData_Identities++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMIDNotificationData_Identities);

    }

}

/* Struct = IDL:org/csapi/pam/TpPAMGMCEventData:1.0 */
static void
decode_org_csapi_pam_TpPAMGMCEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMGMCEventData_GroupName;
    guint32   i_org_csapi_pam_TpPAMGMCEventData_GroupName;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMGMCEventData_GroupType;
    guint32   i_org_csapi_pam_TpPAMGMCEventData_GroupType;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_pam_TpPAMGMCEventData_GroupName = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMGMCEventData_GroupName_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMGMCEventData_GroupName);

    for (i_org_csapi_pam_TpPAMGMCEventData_GroupName=0; i_org_csapi_pam_TpPAMGMCEventData_GroupName < u_octet4_loop_org_csapi_pam_TpPAMGMCEventData_GroupName; i_org_csapi_pam_TpPAMGMCEventData_GroupName++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMGMCEventData_GroupName);

    }

    u_octet4_loop_org_csapi_pam_TpPAMGMCEventData_GroupType = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMGMCEventData_GroupType_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMGMCEventData_GroupType);

    for (i_org_csapi_pam_TpPAMGMCEventData_GroupType=0; i_org_csapi_pam_TpPAMGMCEventData_GroupType < u_octet4_loop_org_csapi_pam_TpPAMGMCEventData_GroupType; i_org_csapi_pam_TpPAMGMCEventData_GroupType++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMGMCEventData_GroupType);

    }

}

/* Struct = IDL:org/csapi/pam/TpPAMGMCNotificationData:1.0 */
static void
decode_org_csapi_pam_TpPAMGMCNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMGMCNotificationData_Groups;
    guint32   i_org_csapi_pam_TpPAMGMCNotificationData_Groups;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_pam_TpPAMGMCNotificationData_Groups = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMGMCNotificationData_Groups_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMGMCNotificationData_Groups);

    for (i_org_csapi_pam_TpPAMGMCNotificationData_Groups=0; i_org_csapi_pam_TpPAMGMCNotificationData_Groups < u_octet4_loop_org_csapi_pam_TpPAMGMCNotificationData_Groups; i_org_csapi_pam_TpPAMGMCNotificationData_Groups++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMGMCNotificationData_Groups);

    }

}

/* Struct = IDL:org/csapi/pam/TpPAMACEventData:1.0 */
static void
decode_org_csapi_pam_TpPAMACEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMACEventData_AgentType;
    guint32   i_org_csapi_pam_TpPAMACEventData_AgentType;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_pam_TpPAMACEventData_AgentType = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMACEventData_AgentType_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMACEventData_AgentType);

    for (i_org_csapi_pam_TpPAMACEventData_AgentType=0; i_org_csapi_pam_TpPAMACEventData_AgentType < u_octet4_loop_org_csapi_pam_TpPAMACEventData_AgentType; i_org_csapi_pam_TpPAMACEventData_AgentType++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMACEventData_AgentType);

    }

}

/* Struct = IDL:org/csapi/pam/TpPAMACNotificationData:1.0 */
static void
decode_org_csapi_pam_TpPAMACNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMACNotificationData_Agents;
    guint32   i_org_csapi_pam_TpPAMACNotificationData_Agents;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_pam_TpPAMACNotificationData_Agents = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMACNotificationData_Agents_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMACNotificationData_Agents);

    for (i_org_csapi_pam_TpPAMACNotificationData_Agents=0; i_org_csapi_pam_TpPAMACNotificationData_Agents < u_octet4_loop_org_csapi_pam_TpPAMACNotificationData_Agents; i_org_csapi_pam_TpPAMACNotificationData_Agents++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMACNotificationData_Agents);

    }

}

/* Struct = IDL:org/csapi/pam/TpPAMADEventData:1.0 */
static void
decode_org_csapi_pam_TpPAMADEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMADEventData_AgentName;
    guint32   i_org_csapi_pam_TpPAMADEventData_AgentName;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMADEventData_AgentType;
    guint32   i_org_csapi_pam_TpPAMADEventData_AgentType;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_pam_TpPAMADEventData_AgentName = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMADEventData_AgentName_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMADEventData_AgentName);

    for (i_org_csapi_pam_TpPAMADEventData_AgentName=0; i_org_csapi_pam_TpPAMADEventData_AgentName < u_octet4_loop_org_csapi_pam_TpPAMADEventData_AgentName; i_org_csapi_pam_TpPAMADEventData_AgentName++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMADEventData_AgentName);

    }

    u_octet4_loop_org_csapi_pam_TpPAMADEventData_AgentType = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMADEventData_AgentType_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMADEventData_AgentType);

    for (i_org_csapi_pam_TpPAMADEventData_AgentType=0; i_org_csapi_pam_TpPAMADEventData_AgentType < u_octet4_loop_org_csapi_pam_TpPAMADEventData_AgentType; i_org_csapi_pam_TpPAMADEventData_AgentType++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMADEventData_AgentType);

    }

}

/* Struct = IDL:org/csapi/pam/TpPAMADNotificationData:1.0 */
static void
decode_org_csapi_pam_TpPAMADNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMADNotificationData_Agents;
    guint32   i_org_csapi_pam_TpPAMADNotificationData_Agents;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_pam_TpPAMADNotificationData_Agents = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMADNotificationData_Agents_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMADNotificationData_Agents);

    for (i_org_csapi_pam_TpPAMADNotificationData_Agents=0; i_org_csapi_pam_TpPAMADNotificationData_Agents < u_octet4_loop_org_csapi_pam_TpPAMADNotificationData_Agents; i_org_csapi_pam_TpPAMADNotificationData_Agents++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMADNotificationData_Agents);

    }

}

/* Struct = IDL:org/csapi/pam/TpPAMAAEventData:1.0 */
static void
decode_org_csapi_pam_TpPAMAAEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMAAEventData_IdentityName;
    guint32   i_org_csapi_pam_TpPAMAAEventData_IdentityName;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMAAEventData_IdentityType;
    guint32   i_org_csapi_pam_TpPAMAAEventData_IdentityType;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMAAEventData_AgentName;
    guint32   i_org_csapi_pam_TpPAMAAEventData_AgentName;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMAAEventData_AgentType;
    guint32   i_org_csapi_pam_TpPAMAAEventData_AgentType;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_pam_TpPAMAAEventData_IdentityName = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMAAEventData_IdentityName_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMAAEventData_IdentityName);

    for (i_org_csapi_pam_TpPAMAAEventData_IdentityName=0; i_org_csapi_pam_TpPAMAAEventData_IdentityName < u_octet4_loop_org_csapi_pam_TpPAMAAEventData_IdentityName; i_org_csapi_pam_TpPAMAAEventData_IdentityName++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAAEventData_IdentityName);

    }

    u_octet4_loop_org_csapi_pam_TpPAMAAEventData_IdentityType = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMAAEventData_IdentityType_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMAAEventData_IdentityType);

    for (i_org_csapi_pam_TpPAMAAEventData_IdentityType=0; i_org_csapi_pam_TpPAMAAEventData_IdentityType < u_octet4_loop_org_csapi_pam_TpPAMAAEventData_IdentityType; i_org_csapi_pam_TpPAMAAEventData_IdentityType++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAAEventData_IdentityType);

    }

    u_octet4_loop_org_csapi_pam_TpPAMAAEventData_AgentName = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMAAEventData_AgentName_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMAAEventData_AgentName);

    for (i_org_csapi_pam_TpPAMAAEventData_AgentName=0; i_org_csapi_pam_TpPAMAAEventData_AgentName < u_octet4_loop_org_csapi_pam_TpPAMAAEventData_AgentName; i_org_csapi_pam_TpPAMAAEventData_AgentName++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAAEventData_AgentName);

    }

    u_octet4_loop_org_csapi_pam_TpPAMAAEventData_AgentType = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMAAEventData_AgentType_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMAAEventData_AgentType);

    for (i_org_csapi_pam_TpPAMAAEventData_AgentType=0; i_org_csapi_pam_TpPAMAAEventData_AgentType < u_octet4_loop_org_csapi_pam_TpPAMAAEventData_AgentType; i_org_csapi_pam_TpPAMAAEventData_AgentType++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAAEventData_AgentType);

    }

}

/* Struct = IDL:org/csapi/pam/TpPAMAANotificationData:1.0 */
static void
decode_org_csapi_pam_TpPAMAANotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAANotificationData_Identity);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAANotificationData_Agent);

}

/* Struct = IDL:org/csapi/pam/TpPAMAUEventData:1.0 */
static void
decode_org_csapi_pam_TpPAMAUEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMAUEventData_IdentityName;
    guint32   i_org_csapi_pam_TpPAMAUEventData_IdentityName;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMAUEventData_IdentityType;
    guint32   i_org_csapi_pam_TpPAMAUEventData_IdentityType;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMAUEventData_AgentName;
    guint32   i_org_csapi_pam_TpPAMAUEventData_AgentName;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMAUEventData_AgentType;
    guint32   i_org_csapi_pam_TpPAMAUEventData_AgentType;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_pam_TpPAMAUEventData_IdentityName = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMAUEventData_IdentityName_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMAUEventData_IdentityName);

    for (i_org_csapi_pam_TpPAMAUEventData_IdentityName=0; i_org_csapi_pam_TpPAMAUEventData_IdentityName < u_octet4_loop_org_csapi_pam_TpPAMAUEventData_IdentityName; i_org_csapi_pam_TpPAMAUEventData_IdentityName++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAUEventData_IdentityName);

    }

    u_octet4_loop_org_csapi_pam_TpPAMAUEventData_IdentityType = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMAUEventData_IdentityType_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMAUEventData_IdentityType);

    for (i_org_csapi_pam_TpPAMAUEventData_IdentityType=0; i_org_csapi_pam_TpPAMAUEventData_IdentityType < u_octet4_loop_org_csapi_pam_TpPAMAUEventData_IdentityType; i_org_csapi_pam_TpPAMAUEventData_IdentityType++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAUEventData_IdentityType);

    }

    u_octet4_loop_org_csapi_pam_TpPAMAUEventData_AgentName = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMAUEventData_AgentName_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMAUEventData_AgentName);

    for (i_org_csapi_pam_TpPAMAUEventData_AgentName=0; i_org_csapi_pam_TpPAMAUEventData_AgentName < u_octet4_loop_org_csapi_pam_TpPAMAUEventData_AgentName; i_org_csapi_pam_TpPAMAUEventData_AgentName++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAUEventData_AgentName);

    }

    u_octet4_loop_org_csapi_pam_TpPAMAUEventData_AgentType = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMAUEventData_AgentType_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMAUEventData_AgentType);

    for (i_org_csapi_pam_TpPAMAUEventData_AgentType=0; i_org_csapi_pam_TpPAMAUEventData_AgentType < u_octet4_loop_org_csapi_pam_TpPAMAUEventData_AgentType; i_org_csapi_pam_TpPAMAUEventData_AgentType++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAUEventData_AgentType);

    }

}

/* Struct = IDL:org/csapi/pam/TpPAMAUNotificationData:1.0 */
static void
decode_org_csapi_pam_TpPAMAUNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAUNotificationData_Identity);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAUNotificationData_Agent);

}

/* Struct = IDL:org/csapi/pam/TpPAMCCEventData:1.0 */
static void
decode_org_csapi_pam_TpPAMCCEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMCCEventData_IdentityName;
    guint32   i_org_csapi_pam_TpPAMCCEventData_IdentityName;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMCCEventData_IdentityType;
    guint32   i_org_csapi_pam_TpPAMCCEventData_IdentityType;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMCCEventData_Capabilities;
    guint32   i_org_csapi_pam_TpPAMCCEventData_Capabilities;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_pam_TpPAMCCEventData_IdentityName = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMCCEventData_IdentityName_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMCCEventData_IdentityName);

    for (i_org_csapi_pam_TpPAMCCEventData_IdentityName=0; i_org_csapi_pam_TpPAMCCEventData_IdentityName < u_octet4_loop_org_csapi_pam_TpPAMCCEventData_IdentityName; i_org_csapi_pam_TpPAMCCEventData_IdentityName++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMCCEventData_IdentityName);

    }

    u_octet4_loop_org_csapi_pam_TpPAMCCEventData_IdentityType = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMCCEventData_IdentityType_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMCCEventData_IdentityType);

    for (i_org_csapi_pam_TpPAMCCEventData_IdentityType=0; i_org_csapi_pam_TpPAMCCEventData_IdentityType < u_octet4_loop_org_csapi_pam_TpPAMCCEventData_IdentityType; i_org_csapi_pam_TpPAMCCEventData_IdentityType++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMCCEventData_IdentityType);

    }

    u_octet4_loop_org_csapi_pam_TpPAMCCEventData_Capabilities = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMCCEventData_Capabilities_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMCCEventData_Capabilities);

    for (i_org_csapi_pam_TpPAMCCEventData_Capabilities=0; i_org_csapi_pam_TpPAMCCEventData_Capabilities < u_octet4_loop_org_csapi_pam_TpPAMCCEventData_Capabilities; i_org_csapi_pam_TpPAMCCEventData_Capabilities++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMCCEventData_Capabilities);

    }

}

/* Struct = IDL:org/csapi/pam/TpPAMCCNotificationData:1.0 */
static void
decode_org_csapi_pam_TpPAMCCNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMCCNotificationData_Capabilities;
    guint32   i_org_csapi_pam_TpPAMCCNotificationData_Capabilities;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMCCNotificationData_Identity);

    u_octet4_loop_org_csapi_pam_TpPAMCCNotificationData_Capabilities = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMCCNotificationData_Capabilities_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMCCNotificationData_Capabilities);

    for (i_org_csapi_pam_TpPAMCCNotificationData_Capabilities=0; i_org_csapi_pam_TpPAMCCNotificationData_Capabilities < u_octet4_loop_org_csapi_pam_TpPAMCCNotificationData_Capabilities; i_org_csapi_pam_TpPAMCCNotificationData_Capabilities++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMCCNotificationData_Capabilities);

    }

}

/* Struct = IDL:org/csapi/pam/TpPAMACPSEventData:1.0 */
static void
decode_org_csapi_pam_TpPAMACPSEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMACPSEventData_AgentName;
    guint32   i_org_csapi_pam_TpPAMACPSEventData_AgentName;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMACPSEventData_AgentType;
    guint32   i_org_csapi_pam_TpPAMACPSEventData_AgentType;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMACPSEventData_Capabilities;
    guint32   i_org_csapi_pam_TpPAMACPSEventData_Capabilities;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMACPSEventData_AttributeNames;
    guint32   i_org_csapi_pam_TpPAMACPSEventData_AttributeNames;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_pam_TpPAMACPSEventData_AgentName = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMACPSEventData_AgentName_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMACPSEventData_AgentName);

    for (i_org_csapi_pam_TpPAMACPSEventData_AgentName=0; i_org_csapi_pam_TpPAMACPSEventData_AgentName < u_octet4_loop_org_csapi_pam_TpPAMACPSEventData_AgentName; i_org_csapi_pam_TpPAMACPSEventData_AgentName++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMACPSEventData_AgentName);

    }

    u_octet4_loop_org_csapi_pam_TpPAMACPSEventData_AgentType = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMACPSEventData_AgentType_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMACPSEventData_AgentType);

    for (i_org_csapi_pam_TpPAMACPSEventData_AgentType=0; i_org_csapi_pam_TpPAMACPSEventData_AgentType < u_octet4_loop_org_csapi_pam_TpPAMACPSEventData_AgentType; i_org_csapi_pam_TpPAMACPSEventData_AgentType++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMACPSEventData_AgentType);

    }

    u_octet4_loop_org_csapi_pam_TpPAMACPSEventData_Capabilities = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMACPSEventData_Capabilities_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMACPSEventData_Capabilities);

    for (i_org_csapi_pam_TpPAMACPSEventData_Capabilities=0; i_org_csapi_pam_TpPAMACPSEventData_Capabilities < u_octet4_loop_org_csapi_pam_TpPAMACPSEventData_Capabilities; i_org_csapi_pam_TpPAMACPSEventData_Capabilities++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMACPSEventData_Capabilities);

    }

    u_octet4_loop_org_csapi_pam_TpPAMACPSEventData_AttributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMACPSEventData_AttributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMACPSEventData_AttributeNames);

    for (i_org_csapi_pam_TpPAMACPSEventData_AttributeNames=0; i_org_csapi_pam_TpPAMACPSEventData_AttributeNames < u_octet4_loop_org_csapi_pam_TpPAMACPSEventData_AttributeNames; i_org_csapi_pam_TpPAMACPSEventData_AttributeNames++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMACPSEventData_AttributeNames);

    }

    proto_tree_add_int64(tree, hf_org_csapi_pam_TpPAMACPSEventData_ReportingPeriod, tvb, *offset-8, 8, get_CDR_long_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/pam/TpPAMACPSNotificationData:1.0 */
static void
decode_org_csapi_pam_TpPAMACPSNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMACPSNotificationData_AttributeNames;
    guint32   i_org_csapi_pam_TpPAMACPSNotificationData_AttributeNames;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMACPSNotificationData_Agent);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMACPSNotificationData_Capability);

    u_octet4_loop_org_csapi_pam_TpPAMACPSNotificationData_AttributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMACPSNotificationData_AttributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMACPSNotificationData_AttributeNames);

    for (i_org_csapi_pam_TpPAMACPSNotificationData_AttributeNames=0; i_org_csapi_pam_TpPAMACPSNotificationData_AttributeNames < u_octet4_loop_org_csapi_pam_TpPAMACPSNotificationData_AttributeNames; i_org_csapi_pam_TpPAMACPSNotificationData_AttributeNames++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMACPSNotificationData_AttributeNames);

    }

}

/* Struct = IDL:org/csapi/pam/TpPAMAPSEventData:1.0 */
static void
decode_org_csapi_pam_TpPAMAPSEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMAPSEventData_AgentName;
    guint32   i_org_csapi_pam_TpPAMAPSEventData_AgentName;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMAPSEventData_AgentType;
    guint32   i_org_csapi_pam_TpPAMAPSEventData_AgentType;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMAPSEventData_AttributeNames;
    guint32   i_org_csapi_pam_TpPAMAPSEventData_AttributeNames;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_pam_TpPAMAPSEventData_AgentName = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMAPSEventData_AgentName_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMAPSEventData_AgentName);

    for (i_org_csapi_pam_TpPAMAPSEventData_AgentName=0; i_org_csapi_pam_TpPAMAPSEventData_AgentName < u_octet4_loop_org_csapi_pam_TpPAMAPSEventData_AgentName; i_org_csapi_pam_TpPAMAPSEventData_AgentName++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAPSEventData_AgentName);

    }

    u_octet4_loop_org_csapi_pam_TpPAMAPSEventData_AgentType = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMAPSEventData_AgentType_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMAPSEventData_AgentType);

    for (i_org_csapi_pam_TpPAMAPSEventData_AgentType=0; i_org_csapi_pam_TpPAMAPSEventData_AgentType < u_octet4_loop_org_csapi_pam_TpPAMAPSEventData_AgentType; i_org_csapi_pam_TpPAMAPSEventData_AgentType++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAPSEventData_AgentType);

    }

    u_octet4_loop_org_csapi_pam_TpPAMAPSEventData_AttributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMAPSEventData_AttributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMAPSEventData_AttributeNames);

    for (i_org_csapi_pam_TpPAMAPSEventData_AttributeNames=0; i_org_csapi_pam_TpPAMAPSEventData_AttributeNames < u_octet4_loop_org_csapi_pam_TpPAMAPSEventData_AttributeNames; i_org_csapi_pam_TpPAMAPSEventData_AttributeNames++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAPSEventData_AttributeNames);

    }

    proto_tree_add_int64(tree, hf_org_csapi_pam_TpPAMAPSEventData_ReportingPeriod, tvb, *offset-8, 8, get_CDR_long_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/pam/TpPAMAPSNotificationData:1.0 */
static void
decode_org_csapi_pam_TpPAMAPSNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMAPSNotificationData_AttributeNames;
    guint32   i_org_csapi_pam_TpPAMAPSNotificationData_AttributeNames;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAPSNotificationData_Agent);

    u_octet4_loop_org_csapi_pam_TpPAMAPSNotificationData_AttributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMAPSNotificationData_AttributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMAPSNotificationData_AttributeNames);

    for (i_org_csapi_pam_TpPAMAPSNotificationData_AttributeNames=0; i_org_csapi_pam_TpPAMAPSNotificationData_AttributeNames < u_octet4_loop_org_csapi_pam_TpPAMAPSNotificationData_AttributeNames; i_org_csapi_pam_TpPAMAPSNotificationData_AttributeNames++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAPSNotificationData_AttributeNames);

    }

}

/* Struct = IDL:org/csapi/pam/TpPAMIPSEventData:1.0 */
static void
decode_org_csapi_pam_TpPAMIPSEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMIPSEventData_IdentityName;
    guint32   i_org_csapi_pam_TpPAMIPSEventData_IdentityName;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMIPSEventData_IdentityType;
    guint32   i_org_csapi_pam_TpPAMIPSEventData_IdentityType;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMIPSEventData_AttributeNames;
    guint32   i_org_csapi_pam_TpPAMIPSEventData_AttributeNames;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_pam_TpPAMIPSEventData_IdentityName = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMIPSEventData_IdentityName_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMIPSEventData_IdentityName);

    for (i_org_csapi_pam_TpPAMIPSEventData_IdentityName=0; i_org_csapi_pam_TpPAMIPSEventData_IdentityName < u_octet4_loop_org_csapi_pam_TpPAMIPSEventData_IdentityName; i_org_csapi_pam_TpPAMIPSEventData_IdentityName++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMIPSEventData_IdentityName);

    }

    u_octet4_loop_org_csapi_pam_TpPAMIPSEventData_IdentityType = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMIPSEventData_IdentityType_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMIPSEventData_IdentityType);

    for (i_org_csapi_pam_TpPAMIPSEventData_IdentityType=0; i_org_csapi_pam_TpPAMIPSEventData_IdentityType < u_octet4_loop_org_csapi_pam_TpPAMIPSEventData_IdentityType; i_org_csapi_pam_TpPAMIPSEventData_IdentityType++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMIPSEventData_IdentityType);

    }

    u_octet4_loop_org_csapi_pam_TpPAMIPSEventData_AttributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMIPSEventData_AttributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMIPSEventData_AttributeNames);

    for (i_org_csapi_pam_TpPAMIPSEventData_AttributeNames=0; i_org_csapi_pam_TpPAMIPSEventData_AttributeNames < u_octet4_loop_org_csapi_pam_TpPAMIPSEventData_AttributeNames; i_org_csapi_pam_TpPAMIPSEventData_AttributeNames++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMIPSEventData_AttributeNames);

    }

    proto_tree_add_int64(tree, hf_org_csapi_pam_TpPAMIPSEventData_ReportingPeriod, tvb, *offset-8, 8, get_CDR_long_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/pam/TpPAMIPSNotificationData:1.0 */
static void
decode_org_csapi_pam_TpPAMIPSNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMIPSNotificationData_Attributes;
    guint32   i_org_csapi_pam_TpPAMIPSNotificationData_Attributes;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMIPSNotificationData_Identity);

    u_octet4_loop_org_csapi_pam_TpPAMIPSNotificationData_Attributes = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMIPSNotificationData_Attributes_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMIPSNotificationData_Attributes);

    for (i_org_csapi_pam_TpPAMIPSNotificationData_Attributes=0; i_org_csapi_pam_TpPAMIPSNotificationData_Attributes < u_octet4_loop_org_csapi_pam_TpPAMIPSNotificationData_Attributes; i_org_csapi_pam_TpPAMIPSNotificationData_Attributes++) {

        /*  Begin struct "org_csapi_pam_TpPAMPresenceData"  */
        decode_org_csapi_pam_TpPAMPresenceData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMPresenceData"  */
    }

}

/* Struct = IDL:org/csapi/pam/TpPAMAVCEventData:1.0 */
static void
decode_org_csapi_pam_TpPAMAVCEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMAVCEventData_IdentityName;
    guint32   i_org_csapi_pam_TpPAMAVCEventData_IdentityName;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMAVCEventData_IdentityType;
    guint32   i_org_csapi_pam_TpPAMAVCEventData_IdentityType;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMAVCEventData_PAMContext;
    guint32   i_org_csapi_pam_TpPAMAVCEventData_PAMContext;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMAVCEventData_AttributeNames;
    guint32   i_org_csapi_pam_TpPAMAVCEventData_AttributeNames;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_pam_TpPAMAVCEventData_IdentityName = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMAVCEventData_IdentityName_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMAVCEventData_IdentityName);

    for (i_org_csapi_pam_TpPAMAVCEventData_IdentityName=0; i_org_csapi_pam_TpPAMAVCEventData_IdentityName < u_octet4_loop_org_csapi_pam_TpPAMAVCEventData_IdentityName; i_org_csapi_pam_TpPAMAVCEventData_IdentityName++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAVCEventData_IdentityName);

    }

    u_octet4_loop_org_csapi_pam_TpPAMAVCEventData_IdentityType = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMAVCEventData_IdentityType_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMAVCEventData_IdentityType);

    for (i_org_csapi_pam_TpPAMAVCEventData_IdentityType=0; i_org_csapi_pam_TpPAMAVCEventData_IdentityType < u_octet4_loop_org_csapi_pam_TpPAMAVCEventData_IdentityType; i_org_csapi_pam_TpPAMAVCEventData_IdentityType++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAVCEventData_IdentityType);

    }

    u_octet4_loop_org_csapi_pam_TpPAMAVCEventData_PAMContext = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMAVCEventData_PAMContext_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMAVCEventData_PAMContext);

    for (i_org_csapi_pam_TpPAMAVCEventData_PAMContext=0; i_org_csapi_pam_TpPAMAVCEventData_PAMContext < u_octet4_loop_org_csapi_pam_TpPAMAVCEventData_PAMContext; i_org_csapi_pam_TpPAMAVCEventData_PAMContext++) {

        /*  Begin struct "org_csapi_pam_TpPAMContext"  */
        decode_org_csapi_pam_TpPAMContext_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMContext"  */
    }

    u_octet4_loop_org_csapi_pam_TpPAMAVCEventData_AttributeNames = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMAVCEventData_AttributeNames_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMAVCEventData_AttributeNames);

    for (i_org_csapi_pam_TpPAMAVCEventData_AttributeNames=0; i_org_csapi_pam_TpPAMAVCEventData_AttributeNames < u_octet4_loop_org_csapi_pam_TpPAMAVCEventData_AttributeNames; i_org_csapi_pam_TpPAMAVCEventData_AttributeNames++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAVCEventData_AttributeNames);

    }

    proto_tree_add_int64(tree, hf_org_csapi_pam_TpPAMAVCEventData_ReportingPeriod, tvb, *offset-8, 8, get_CDR_long_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/pam/TpPAMAVCNotificationData:1.0 */
static void
decode_org_csapi_pam_TpPAMAVCNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMAVCNotificationData_Availability;
    guint32   i_org_csapi_pam_TpPAMAVCNotificationData_Availability;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMAVCNotificationData_Identity);

    u_octet4_loop_org_csapi_pam_TpPAMAVCNotificationData_Availability = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMAVCNotificationData_Availability_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMAVCNotificationData_Availability);

    for (i_org_csapi_pam_TpPAMAVCNotificationData_Availability=0; i_org_csapi_pam_TpPAMAVCNotificationData_Availability < u_octet4_loop_org_csapi_pam_TpPAMAVCNotificationData_Availability; i_org_csapi_pam_TpPAMAVCNotificationData_Availability++) {

        /*  Begin struct "org_csapi_pam_TpPAMAvailabilityProfile"  */
        decode_org_csapi_pam_TpPAMAvailabilityProfile_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMAvailabilityProfile"  */
    }

}

/* Struct = IDL:org/csapi/pam/TpPAMWCEventData:1.0 */
static void
decode_org_csapi_pam_TpPAMWCEventData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_pam_TpPAMWCEventData_Events;
    guint32   i_org_csapi_pam_TpPAMWCEventData_Events;
    guint32   u_octet4;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMWCEventData_IdentityName;
    guint32   i_org_csapi_pam_TpPAMWCEventData_IdentityName;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMWCEventData_IdentityType;
    guint32   i_org_csapi_pam_TpPAMWCEventData_IdentityType;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_pam_TpPAMWCEventData_Events = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMWCEventData_Events_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMWCEventData_Events);

    for (i_org_csapi_pam_TpPAMWCEventData_Events=0; i_org_csapi_pam_TpPAMWCEventData_Events < u_octet4_loop_org_csapi_pam_TpPAMWCEventData_Events; i_org_csapi_pam_TpPAMWCEventData_Events++) {

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMWCEventData_Events, tvb, *offset-4, 4, u_octet4);

    }

    u_octet4_loop_org_csapi_pam_TpPAMWCEventData_IdentityName = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMWCEventData_IdentityName_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMWCEventData_IdentityName);

    for (i_org_csapi_pam_TpPAMWCEventData_IdentityName=0; i_org_csapi_pam_TpPAMWCEventData_IdentityName < u_octet4_loop_org_csapi_pam_TpPAMWCEventData_IdentityName; i_org_csapi_pam_TpPAMWCEventData_IdentityName++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMWCEventData_IdentityName);

    }

    u_octet4_loop_org_csapi_pam_TpPAMWCEventData_IdentityType = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMWCEventData_IdentityType_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMWCEventData_IdentityType);

    for (i_org_csapi_pam_TpPAMWCEventData_IdentityType=0; i_org_csapi_pam_TpPAMWCEventData_IdentityType < u_octet4_loop_org_csapi_pam_TpPAMWCEventData_IdentityType; i_org_csapi_pam_TpPAMWCEventData_IdentityType++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMWCEventData_IdentityType);

    }

    proto_tree_add_int64(tree, hf_org_csapi_pam_TpPAMWCEventData_ReportingPeriod, tvb, *offset-8, 8, get_CDR_long_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/pam/TpPAMWCNotificationData:1.0 */
static void
decode_org_csapi_pam_TpPAMWCNotificationData_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    guint32   u_octet4_loop_org_csapi_pam_TpPAMWCNotificationData_Watchers;
    guint32   i_org_csapi_pam_TpPAMWCNotificationData_Watchers;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMWCNotificationData_Event, tvb, *offset-4, 4, u_octet4);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMWCNotificationData_ChangeType, tvb, *offset-4, 4, u_octet4);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMWCNotificationData_Identity);

    u_octet4_loop_org_csapi_pam_TpPAMWCNotificationData_Watchers = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMWCNotificationData_Watchers_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_pam_TpPAMWCNotificationData_Watchers);

    for (i_org_csapi_pam_TpPAMWCNotificationData_Watchers=0; i_org_csapi_pam_TpPAMWCNotificationData_Watchers < u_octet4_loop_org_csapi_pam_TpPAMWCNotificationData_Watchers; i_org_csapi_pam_TpPAMWCNotificationData_Watchers++) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_pam_TpPAMWCNotificationData_Watchers);

    }

}

/* Struct = IDL:org/csapi/pam/TpPAMErrorInfo:1.0 */
static void
decode_org_csapi_pam_TpPAMErrorInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMErrorInfo_Cause, tvb, *offset-4, 4, u_octet4);

    /*  Begin union "org_csapi_pam_TpPAMNotificationInfo"  */
    decode_org_csapi_pam_TpPAMNotificationInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

    /*  End union "org_csapi_pam_TpPAMNotificationInfo"  */
}

/* Struct = IDL:org/csapi/mmm/TpMailboxIdentifier:1.0 */
static void
decode_org_csapi_mmm_TpMailboxIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

    proto_tree_add_int(tree, hf_org_csapi_mmm_TpMailboxIdentifier_SessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/mmm/TpMultiMediaMessagingIdentifier:1.0 */
static void
decode_org_csapi_mmm_TpMultiMediaMessagingIdentifier_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

    proto_tree_add_int(tree, hf_org_csapi_mmm_TpMultiMediaMessagingIdentifier_SessionID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/mmm/TpListMessagesCriteria:1.0 */
static void
decode_org_csapi_mmm_TpListMessagesCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    proto_tree_add_boolean(tree, hf_org_csapi_mmm_TpListMessagesCriteria_OnlyUnreadMessages, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

}

/* Struct = IDL:org/csapi/mmm/TpMailboxFolderStatusInformation:1.0 */
static void
decode_org_csapi_mmm_TpMailboxFolderStatusInformation_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    proto_tree_add_int(tree, hf_org_csapi_mmm_TpMailboxFolderStatusInformation_TotalMessageCount, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/mmm/TpMessageDescription:1.0 */
static void
decode_org_csapi_mmm_TpMessageDescription_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_TpMessageDescription_To;
    guint32   i_org_csapi_mmm_TpMessageDescription_To;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMessageDescription_MessageID);

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    u_octet4_loop_org_csapi_mmm_TpMessageDescription_To = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mmm_TpMessageDescription_To_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_TpMessageDescription_To);

    for (i_org_csapi_mmm_TpMessageDescription_To=0; i_org_csapi_mmm_TpMessageDescription_To < u_octet4_loop_org_csapi_mmm_TpMessageDescription_To; i_org_csapi_mmm_TpMessageDescription_To++) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
    }

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMessageDescription_Subject);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMessageDescription_ReceivedDate);

    proto_tree_add_int(tree, hf_org_csapi_mmm_TpMessageDescription_Size, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/mmm/TpBodyPartDescription:1.0 */
static void
decode_org_csapi_mmm_TpBodyPartDescription_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpBodyPartDescription_ContentDescription);

    proto_tree_add_int(tree, hf_org_csapi_mmm_TpBodyPartDescription_ContentSize, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpBodyPartDescription_ContentType);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpBodyPartDescription_ContentTransferEncoding);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpBodyPartDescription_ContentID);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpBodyPartDescription_ContentDisposition);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpBodyPartDescription_PartID);

    proto_tree_add_int(tree, hf_org_csapi_mmm_TpBodyPartDescription_NestingLevel, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/mmm/TpBodyPart:1.0 */
static void
decode_org_csapi_mmm_TpBodyPart_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_TpBodyPart_BodyPartContent;
    guint32   i_org_csapi_mmm_TpBodyPart_BodyPartContent;
    /* Operation specific Variable declarations End */

    /*  Begin struct "org_csapi_mmm_TpBodyPartDescription"  */
    decode_org_csapi_mmm_TpBodyPartDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_mmm_TpBodyPartDescription"  */
    u_octet4_loop_org_csapi_mmm_TpBodyPart_BodyPartContent = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mmm_TpBodyPart_BodyPartContent_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_TpBodyPart_BodyPartContent);

    for (i_org_csapi_mmm_TpBodyPart_BodyPartContent=0; i_org_csapi_mmm_TpBodyPart_BodyPartContent < u_octet4_loop_org_csapi_mmm_TpBodyPart_BodyPartContent; i_org_csapi_mmm_TpBodyPart_BodyPartContent++) {

        proto_tree_add_uint(tree, hf_org_csapi_mmm_TpBodyPart_BodyPartContent, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

    }

}

/* Struct = IDL:org/csapi/mmm/TpGenericHeaderField:1.0 */
static void
decode_org_csapi_mmm_TpGenericHeaderField_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpGenericHeaderField_FieldName);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpGenericHeaderField_FieldValue);

}

/* Struct = IDL:org/csapi/mmm/TpMessageInfoPropertyError:1.0 */
static void
decode_org_csapi_mmm_TpMessageInfoPropertyError_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mmm_TpMessageInfoPropertyError_MessagePropertyName, tvb, *offset-4, 4, u_octet4);

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mmm_TpMessageInfoPropertyError_Error, tvb, *offset-4, 4, u_octet4);

}

/* Struct = IDL:org/csapi/mmm/TpNewMailboxMessageArrivedCriteria:1.0 */
static void
decode_org_csapi_mmm_TpNewMailboxMessageArrivedCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpNewMailboxMessageArrivedCriteria_MailboxID);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpNewMailboxMessageArrivedCriteria_AuthenticationInfo);

}

/* Struct = IDL:org/csapi/mmm/TpNewMailboxMessageArrivedInfo:1.0 */
static void
decode_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_MessageDescription;
    guint32   i_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_MessageDescription;
    guint32   u_octet4_loop_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_ExtendedHeaderInformation;
    guint32   i_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_ExtendedHeaderInformation;
    /* Operation specific Variable declarations End */

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_MailboxID);

    giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_FolderID);

    u_octet4_loop_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_MessageDescription = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_MessageDescription_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_MessageDescription);

    for (i_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_MessageDescription=0; i_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_MessageDescription < u_octet4_loop_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_MessageDescription; i_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_MessageDescription++) {

        /*  Begin struct "org_csapi_mmm_TpMessageDescription"  */
        decode_org_csapi_mmm_TpMessageDescription_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mmm_TpMessageDescription"  */
    }

    u_octet4_loop_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_ExtendedHeaderInformation = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_ExtendedHeaderInformation_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_ExtendedHeaderInformation);

    for (i_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_ExtendedHeaderInformation=0; i_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_ExtendedHeaderInformation < u_octet4_loop_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_ExtendedHeaderInformation; i_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_ExtendedHeaderInformation++) {

        /*  Begin union "org_csapi_mmm_TpMessageHeaderField"  */
        decode_org_csapi_mmm_TpMessageHeaderField_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_mmm_TpMessageHeaderField"  */
    }

}

/* Struct = IDL:org/csapi/mmm/TpQueryStatusReport:1.0 */
static void
decode_org_csapi_mmm_TpQueryStatusReport_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    proto_tree_add_int(tree, hf_org_csapi_mmm_TpQueryStatusReport_ReportedStatus, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/mmm/TpTerminatingAddressList:1.0 */
static void
decode_org_csapi_mmm_TpTerminatingAddressList_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_TpTerminatingAddressList_ToAddressList;
    guint32   i_org_csapi_mmm_TpTerminatingAddressList_ToAddressList;
    guint32   u_octet4_loop_org_csapi_mmm_TpTerminatingAddressList_CcAddressList;
    guint32   i_org_csapi_mmm_TpTerminatingAddressList_CcAddressList;
    guint32   u_octet4_loop_org_csapi_mmm_TpTerminatingAddressList_BccAddressList;
    guint32   i_org_csapi_mmm_TpTerminatingAddressList_BccAddressList;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_mmm_TpTerminatingAddressList_ToAddressList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mmm_TpTerminatingAddressList_ToAddressList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_TpTerminatingAddressList_ToAddressList);

    for (i_org_csapi_mmm_TpTerminatingAddressList_ToAddressList=0; i_org_csapi_mmm_TpTerminatingAddressList_ToAddressList < u_octet4_loop_org_csapi_mmm_TpTerminatingAddressList_ToAddressList; i_org_csapi_mmm_TpTerminatingAddressList_ToAddressList++) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
    }

    u_octet4_loop_org_csapi_mmm_TpTerminatingAddressList_CcAddressList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mmm_TpTerminatingAddressList_CcAddressList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_TpTerminatingAddressList_CcAddressList);

    for (i_org_csapi_mmm_TpTerminatingAddressList_CcAddressList=0; i_org_csapi_mmm_TpTerminatingAddressList_CcAddressList < u_octet4_loop_org_csapi_mmm_TpTerminatingAddressList_CcAddressList; i_org_csapi_mmm_TpTerminatingAddressList_CcAddressList++) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
    }

    u_octet4_loop_org_csapi_mmm_TpTerminatingAddressList_BccAddressList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mmm_TpTerminatingAddressList_BccAddressList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_TpTerminatingAddressList_BccAddressList);

    for (i_org_csapi_mmm_TpTerminatingAddressList_BccAddressList=0; i_org_csapi_mmm_TpTerminatingAddressList_BccAddressList < u_octet4_loop_org_csapi_mmm_TpTerminatingAddressList_BccAddressList; i_org_csapi_mmm_TpTerminatingAddressList_BccAddressList++) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
    }

}

/* Struct = IDL:org/csapi/mmm/TpNewMessageArrivedCriteria:1.0 */
static void
decode_org_csapi_mmm_TpNewMessageArrivedCriteria_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /*  Begin struct "org_csapi_TpAddressRange"  */
    decode_org_csapi_TpAddressRange_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddressRange"  */
    /*  Begin struct "org_csapi_TpAddressRange"  */
    decode_org_csapi_TpAddressRange_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddressRange"  */
    proto_tree_add_boolean(tree, hf_org_csapi_mmm_TpNewMessageArrivedCriteria_CreateMultiMediaMessagingSession, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

}

/* Struct = IDL:org/csapi/mmm/TpMessagingNotificationRequested:1.0 */
static void
decode_org_csapi_mmm_TpMessagingNotificationRequested_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_TpMessagingNotificationRequested_EventCriteria;
    guint32   i_org_csapi_mmm_TpMessagingNotificationRequested_EventCriteria;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_mmm_TpMessagingNotificationRequested_EventCriteria = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mmm_TpMessagingNotificationRequested_EventCriteria_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_TpMessagingNotificationRequested_EventCriteria);

    for (i_org_csapi_mmm_TpMessagingNotificationRequested_EventCriteria=0; i_org_csapi_mmm_TpMessagingNotificationRequested_EventCriteria < u_octet4_loop_org_csapi_mmm_TpMessagingNotificationRequested_EventCriteria; i_org_csapi_mmm_TpMessagingNotificationRequested_EventCriteria++) {

        /*  Begin union "org_csapi_mmm_TpMessagingEventCriteria"  */
        decode_org_csapi_mmm_TpMessagingEventCriteria_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_mmm_TpMessagingEventCriteria"  */
    }

    proto_tree_add_int(tree, hf_org_csapi_mmm_TpMessagingNotificationRequested_AssignmentID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

}

/* Struct = IDL:org/csapi/mmm/TpMessagingNotificationRequestedSetEntry:1.0 */
static void
decode_org_csapi_mmm_TpMessagingNotificationRequestedSetEntry_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_TpMessagingNotificationRequestedSetEntry_MessagingNotificationRequestedSet;
    guint32   i_org_csapi_mmm_TpMessagingNotificationRequestedSetEntry_MessagingNotificationRequestedSet;
    /* Operation specific Variable declarations End */

    u_octet4_loop_org_csapi_mmm_TpMessagingNotificationRequestedSetEntry_MessagingNotificationRequestedSet = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mmm_TpMessagingNotificationRequestedSetEntry_MessagingNotificationRequestedSet_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_TpMessagingNotificationRequestedSetEntry_MessagingNotificationRequestedSet);

    for (i_org_csapi_mmm_TpMessagingNotificationRequestedSetEntry_MessagingNotificationRequestedSet=0; i_org_csapi_mmm_TpMessagingNotificationRequestedSetEntry_MessagingNotificationRequestedSet < u_octet4_loop_org_csapi_mmm_TpMessagingNotificationRequestedSetEntry_MessagingNotificationRequestedSet; i_org_csapi_mmm_TpMessagingNotificationRequestedSetEntry_MessagingNotificationRequestedSet++) {

        /*  Begin struct "org_csapi_mmm_TpMessagingNotificationRequested"  */
        decode_org_csapi_mmm_TpMessagingNotificationRequested_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mmm_TpMessagingNotificationRequested"  */
    }

    proto_tree_add_boolean(tree, hf_org_csapi_mmm_TpMessagingNotificationRequestedSetEntry_Final, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

}

/* Struct = IDL:org/csapi/mmm/TpNewMessageArrivedInfo:1.0 */
static void
decode_org_csapi_mmm_TpNewMessageArrivedInfo_st(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, proto_item *item _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4_loop_org_csapi_mmm_TpNewMessageArrivedInfo_DestinationAddressSet;
    guint32   i_org_csapi_mmm_TpNewMessageArrivedInfo_DestinationAddressSet;
    guint32   u_octet4_loop_org_csapi_mmm_TpNewMessageArrivedInfo_Message;
    guint32   i_org_csapi_mmm_TpNewMessageArrivedInfo_Message;
    guint32   u_octet4_loop_org_csapi_mmm_TpNewMessageArrivedInfo_Headers;
    guint32   i_org_csapi_mmm_TpNewMessageArrivedInfo_Headers;
    /* Operation specific Variable declarations End */

    /*  Begin struct "org_csapi_TpAddress"  */
    decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_TpAddress"  */
    u_octet4_loop_org_csapi_mmm_TpNewMessageArrivedInfo_DestinationAddressSet = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mmm_TpNewMessageArrivedInfo_DestinationAddressSet_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_TpNewMessageArrivedInfo_DestinationAddressSet);

    for (i_org_csapi_mmm_TpNewMessageArrivedInfo_DestinationAddressSet=0; i_org_csapi_mmm_TpNewMessageArrivedInfo_DestinationAddressSet < u_octet4_loop_org_csapi_mmm_TpNewMessageArrivedInfo_DestinationAddressSet; i_org_csapi_mmm_TpNewMessageArrivedInfo_DestinationAddressSet++) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
    }

    u_octet4_loop_org_csapi_mmm_TpNewMessageArrivedInfo_Message = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mmm_TpNewMessageArrivedInfo_Message_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_TpNewMessageArrivedInfo_Message);

    for (i_org_csapi_mmm_TpNewMessageArrivedInfo_Message=0; i_org_csapi_mmm_TpNewMessageArrivedInfo_Message < u_octet4_loop_org_csapi_mmm_TpNewMessageArrivedInfo_Message; i_org_csapi_mmm_TpNewMessageArrivedInfo_Message++) {

        proto_tree_add_uint(tree, hf_org_csapi_mmm_TpNewMessageArrivedInfo_Message, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

    }

    u_octet4_loop_org_csapi_mmm_TpNewMessageArrivedInfo_Headers = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mmm_TpNewMessageArrivedInfo_Headers_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_TpNewMessageArrivedInfo_Headers);

    for (i_org_csapi_mmm_TpNewMessageArrivedInfo_Headers=0; i_org_csapi_mmm_TpNewMessageArrivedInfo_Headers < u_octet4_loop_org_csapi_mmm_TpNewMessageArrivedInfo_Headers; i_org_csapi_mmm_TpNewMessageArrivedInfo_Headers++) {

        /*  Begin union "org_csapi_mmm_TpMessageHeaderField"  */
        decode_org_csapi_mmm_TpMessageHeaderField_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_mmm_TpMessageHeaderField"  */
    }

    /*  Begin struct "org_csapi_mmm_TpMultiMediaMessagingIdentifier"  */
    decode_org_csapi_mmm_TpMultiMediaMessagingIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
    /*  End struct "org_csapi_mmm_TpMultiMediaMessagingIdentifier"  */
}

/* Union = IDL:org/csapi/TpAoCOrder:1.0 */
static void
decode_org_csapi_TpAoCOrder_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpAoCOrder;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/TpAoCOrder:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_TpAoCOrder_TpAoCOrder, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/TpCallAoCOrderCategory:1.0
     */

    disc_s_TpAoCOrder = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpAoCOrder == 0 ) {

        /*  Begin struct "org_csapi_TpChargeAdviceInfo"  */
        decode_org_csapi_TpChargeAdviceInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpChargeAdviceInfo"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpAoCOrder == 1 ) {

        /*  Begin struct "org_csapi_TpChargePerTime"  */
        decode_org_csapi_TpChargePerTime_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpChargePerTime"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpAoCOrder == 2 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_TpAoCOrder_NetworkCharge);

        return;     /* End Compare for this discriminant type */
    }

}

/* Union = IDL:org/csapi/TpSimpleAttributeValue:1.0 */
static void
decode_org_csapi_TpSimpleAttributeValue_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpSimpleAttributeValue;
    gint8     s_octet1;
    const gchar   *seq = NULL;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/TpSimpleAttributeValue:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_TpSimpleAttributeValue_TpSimpleAttributeValue, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/TpSimpleAttributeTypeInfo:1.0
     */

    disc_s_TpSimpleAttributeValue = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpSimpleAttributeValue == 0 ) {

        proto_tree_add_boolean(tree, hf_org_csapi_TpSimpleAttributeValue_BooleanValue, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpSimpleAttributeValue == 1 ) {

        proto_tree_add_uint(tree, hf_org_csapi_TpSimpleAttributeValue_OctetValue, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpSimpleAttributeValue == 2 ) {

        proto_tree_add_uint(tree, hf_org_csapi_TpSimpleAttributeValue_CharValue, tvb, *offset-1, 1, get_CDR_char(tvb,offset));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpSimpleAttributeValue == 3 ) {

        s_octet1 = get_CDR_wchar(tvb, &seq, offset, header);
        if (tree) {
            if (s_octet1 > 0)
                proto_tree_add_uint(tree, hf_org_csapi_TpSimpleAttributeValue_WCharValue_len, tvb, *offset-1-s_octet1, 1, s_octet1);

            if (s_octet1 < 0)
                s_octet1 = -s_octet1;

            if (s_octet1 > 0)
                proto_tree_add_string(tree, hf_org_csapi_TpSimpleAttributeValue_WCharValue, tvb, *offset-s_octet1, s_octet1, seq);
        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpSimpleAttributeValue == 4 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_TpSimpleAttributeValue_StringValue);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpSimpleAttributeValue == 5 ) {

        u_octet4 = get_CDR_wstring(tvb, &seq, offset, stream_is_big_endian, boundary, header);
        proto_tree_add_string(tree, hf_org_csapi_TpSimpleAttributeValue_WStringValue, tvb, *offset-u_octet4, u_octet4, (u_octet4 > 0) ? seq : "");

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpSimpleAttributeValue == 6 ) {

        proto_tree_add_int(tree, hf_org_csapi_TpSimpleAttributeValue_Int16Value, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpSimpleAttributeValue == 7 ) {

        proto_tree_add_uint(tree, hf_org_csapi_TpSimpleAttributeValue_UnsignedInt16Value, tvb, *offset-2, 2, get_CDR_ushort(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpSimpleAttributeValue == 8 ) {

        proto_tree_add_int(tree, hf_org_csapi_TpSimpleAttributeValue_Int32Value, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpSimpleAttributeValue == 9 ) {

        proto_tree_add_uint(tree, hf_org_csapi_TpSimpleAttributeValue_UnsignedInt32Value, tvb, *offset-4, 4, get_CDR_ulong(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpSimpleAttributeValue == 10 ) {

        proto_tree_add_int64(tree, hf_org_csapi_TpSimpleAttributeValue_Int64Value, tvb, *offset-8, 8, get_CDR_long_long(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpSimpleAttributeValue == 11 ) {

        proto_tree_add_uint64(tree, hf_org_csapi_TpSimpleAttributeValue_UnsignedInt64Value, tvb, *offset-8, 8, get_CDR_ulong_long(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpSimpleAttributeValue == 12 ) {

        proto_tree_add_float(tree, hf_org_csapi_TpSimpleAttributeValue_FloatValue, tvb, *offset-4, 4, get_CDR_float(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpSimpleAttributeValue == 13 ) {

        proto_tree_add_double(tree, hf_org_csapi_TpSimpleAttributeValue_DoubleValue, tvb, *offset-8, 8, get_CDR_double(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

}

/* Union = IDL:org/csapi/TpAttributeValue:1.0 */
static void
decode_org_csapi_TpAttributeValue_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpAttributeValue;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/TpAttributeValue:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_TpAttributeValue_TpAttributeValue, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/TpAttributeTagInfo:1.0
     */

    disc_s_TpAttributeValue = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpAttributeValue == 0 ) {

        /*  Begin union "org_csapi_TpSimpleAttributeValue"  */
        decode_org_csapi_TpSimpleAttributeValue_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_TpSimpleAttributeValue"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpAttributeValue == 1 ) {

        /*  Begin struct "org_csapi_TpStructuredAttributeValue"  */
        decode_org_csapi_TpStructuredAttributeValue_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpStructuredAttributeValue"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpAttributeValue == 2 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_TpAttributeValue_XMLValue);

        return;     /* End Compare for this discriminant type */
    }

}

/* Union = IDL:org/csapi/fw/TpFwEventCriteria:1.0 */
static void
decode_org_csapi_fw_TpFwEventCriteria_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpFwEventCriteria;
    guint32   u_octet4_loop_org_csapi_fw_TpFwEventCriteria_ServiceTypeNameList;
    guint32   i_org_csapi_fw_TpFwEventCriteria_ServiceTypeNameList;
    guint32   u_octet4_loop_org_csapi_fw_TpFwEventCriteria_UnavailableServiceTypeNameList;
    guint32   i_org_csapi_fw_TpFwEventCriteria_UnavailableServiceTypeNameList;
    guint32   u_octet4_loop_org_csapi_fw_TpFwEventCriteria_CompatibleServiceTypeNameList;
    guint32   i_org_csapi_fw_TpFwEventCriteria_CompatibleServiceTypeNameList;
    guint32   u_octet4_loop_org_csapi_fw_TpFwEventCriteria_SessionCreatedList;
    guint32   i_org_csapi_fw_TpFwEventCriteria_SessionCreatedList;
    guint32   u_octet4_loop_org_csapi_fw_TpFwEventCriteria_SessionTerminatedList;
    guint32   i_org_csapi_fw_TpFwEventCriteria_SessionTerminatedList;
    guint32   u_octet4_loop_org_csapi_fw_TpFwEventCriteria_AgreementSignedList;
    guint32   i_org_csapi_fw_TpFwEventCriteria_AgreementSignedList;
    guint32   u_octet4_loop_org_csapi_fw_TpFwEventCriteria_AgreementEndedList;
    guint32   i_org_csapi_fw_TpFwEventCriteria_AgreementEndedList;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/fw/TpFwEventCriteria:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpFwEventCriteria_TpFwEventCriteria, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/fw/TpFwEventName:1.0
     */

    disc_s_TpFwEventCriteria = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpFwEventCriteria == 0 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpFwEventCriteria_EventNameUndefined);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpFwEventCriteria == 1 ) {

        u_octet4_loop_org_csapi_fw_TpFwEventCriteria_ServiceTypeNameList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_TpFwEventCriteria_ServiceTypeNameList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_TpFwEventCriteria_ServiceTypeNameList);

        for (i_org_csapi_fw_TpFwEventCriteria_ServiceTypeNameList=0; i_org_csapi_fw_TpFwEventCriteria_ServiceTypeNameList < u_octet4_loop_org_csapi_fw_TpFwEventCriteria_ServiceTypeNameList; i_org_csapi_fw_TpFwEventCriteria_ServiceTypeNameList++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpFwEventCriteria_ServiceTypeNameList);

        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpFwEventCriteria == 2 ) {

        u_octet4_loop_org_csapi_fw_TpFwEventCriteria_UnavailableServiceTypeNameList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_TpFwEventCriteria_UnavailableServiceTypeNameList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_TpFwEventCriteria_UnavailableServiceTypeNameList);

        for (i_org_csapi_fw_TpFwEventCriteria_UnavailableServiceTypeNameList=0; i_org_csapi_fw_TpFwEventCriteria_UnavailableServiceTypeNameList < u_octet4_loop_org_csapi_fw_TpFwEventCriteria_UnavailableServiceTypeNameList; i_org_csapi_fw_TpFwEventCriteria_UnavailableServiceTypeNameList++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpFwEventCriteria_UnavailableServiceTypeNameList);

        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpFwEventCriteria == 3 ) {

        u_octet4_loop_org_csapi_fw_TpFwEventCriteria_CompatibleServiceTypeNameList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_TpFwEventCriteria_CompatibleServiceTypeNameList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_TpFwEventCriteria_CompatibleServiceTypeNameList);

        for (i_org_csapi_fw_TpFwEventCriteria_CompatibleServiceTypeNameList=0; i_org_csapi_fw_TpFwEventCriteria_CompatibleServiceTypeNameList < u_octet4_loop_org_csapi_fw_TpFwEventCriteria_CompatibleServiceTypeNameList; i_org_csapi_fw_TpFwEventCriteria_CompatibleServiceTypeNameList++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpFwEventCriteria_CompatibleServiceTypeNameList);

        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpFwEventCriteria == 4 ) {

        u_octet4_loop_org_csapi_fw_TpFwEventCriteria_SessionCreatedList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_TpFwEventCriteria_SessionCreatedList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_TpFwEventCriteria_SessionCreatedList);

        for (i_org_csapi_fw_TpFwEventCriteria_SessionCreatedList=0; i_org_csapi_fw_TpFwEventCriteria_SessionCreatedList < u_octet4_loop_org_csapi_fw_TpFwEventCriteria_SessionCreatedList; i_org_csapi_fw_TpFwEventCriteria_SessionCreatedList++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpFwEventCriteria_SessionCreatedList);

        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpFwEventCriteria == 5 ) {

        u_octet4_loop_org_csapi_fw_TpFwEventCriteria_SessionTerminatedList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_TpFwEventCriteria_SessionTerminatedList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_TpFwEventCriteria_SessionTerminatedList);

        for (i_org_csapi_fw_TpFwEventCriteria_SessionTerminatedList=0; i_org_csapi_fw_TpFwEventCriteria_SessionTerminatedList < u_octet4_loop_org_csapi_fw_TpFwEventCriteria_SessionTerminatedList; i_org_csapi_fw_TpFwEventCriteria_SessionTerminatedList++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpFwEventCriteria_SessionTerminatedList);

        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpFwEventCriteria == 6 ) {

        u_octet4_loop_org_csapi_fw_TpFwEventCriteria_AgreementSignedList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_TpFwEventCriteria_AgreementSignedList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_TpFwEventCriteria_AgreementSignedList);

        for (i_org_csapi_fw_TpFwEventCriteria_AgreementSignedList=0; i_org_csapi_fw_TpFwEventCriteria_AgreementSignedList < u_octet4_loop_org_csapi_fw_TpFwEventCriteria_AgreementSignedList; i_org_csapi_fw_TpFwEventCriteria_AgreementSignedList++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpFwEventCriteria_AgreementSignedList);

        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpFwEventCriteria == 7 ) {

        u_octet4_loop_org_csapi_fw_TpFwEventCriteria_AgreementEndedList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_TpFwEventCriteria_AgreementEndedList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_TpFwEventCriteria_AgreementEndedList);

        for (i_org_csapi_fw_TpFwEventCriteria_AgreementEndedList=0; i_org_csapi_fw_TpFwEventCriteria_AgreementEndedList < u_octet4_loop_org_csapi_fw_TpFwEventCriteria_AgreementEndedList; i_org_csapi_fw_TpFwEventCriteria_AgreementEndedList++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpFwEventCriteria_AgreementEndedList);

        }

        return;     /* End Compare for this discriminant type */
    }

}

/* Union = IDL:org/csapi/fw/TpLoadStatisticEntityID:1.0 */
static void
decode_org_csapi_fw_TpLoadStatisticEntityID_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpLoadStatisticEntityID;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/fw/TpLoadStatisticEntityID:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpLoadStatisticEntityID_TpLoadStatisticEntityID, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/fw/TpLoadStatisticEntityType:1.0
     */

    disc_s_TpLoadStatisticEntityID = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpLoadStatisticEntityID == 0 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpLoadStatisticEntityID_FrameworkID);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpLoadStatisticEntityID == 1 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpLoadStatisticEntityID_ServiceID);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpLoadStatisticEntityID == 2 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpLoadStatisticEntityID_ClientAppID);

        return;     /* End Compare for this discriminant type */
    }

}

/* Union = IDL:org/csapi/fw/TpLoadStatisticInfo:1.0 */
static void
decode_org_csapi_fw_TpLoadStatisticInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpLoadStatisticInfo;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/fw/TpLoadStatisticInfo:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpLoadStatisticInfo_TpLoadStatisticInfo, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/fw/TpLoadStatisticInfoType:1.0
     */

    disc_s_TpLoadStatisticInfo = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpLoadStatisticInfo == 0 ) {

        /*  Begin struct "org_csapi_fw_TpLoadStatisticData"  */
        decode_org_csapi_fw_TpLoadStatisticData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpLoadStatisticData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpLoadStatisticInfo == 1 ) {

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_TpLoadStatisticInfo_LoadStatisticError, tvb, *offset-4, 4, u_octet4);

        return;     /* End Compare for this discriminant type */
    }

}

/* Union = IDL:org/csapi/fw/TpDomainID:1.0 */
static void
decode_org_csapi_fw_TpDomainID_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpDomainID;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/fw/TpDomainID:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpDomainID_TpDomainID, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/fw/TpDomainIDType:1.0
     */

    disc_s_TpDomainID = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpDomainID == 0 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpDomainID_FwID);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpDomainID == 1 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpDomainID_ClientAppID);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpDomainID == 2 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpDomainID_EntOpID);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpDomainID == 3 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpDomainID_ServiceID);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpDomainID == 4 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpDomainID_ServiceSupplierID);

        return;     /* End Compare for this discriminant type */
    }

}

/* Union = IDL:org/csapi/fw/TpMigrationAdditionalInfo:1.0 */
static void
decode_org_csapi_fw_TpMigrationAdditionalInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/fw/TpMigrationAdditionalInfo:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpMigrationAdditionalInfo_TpMigrationAdditionalInfo, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/fw/TpMigrationAdditionalInfoType:1.0
     */

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_fw_TpMigrationAdditionalInfo_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/fw/TpFwEventInfo:1.0 */
static void
decode_org_csapi_fw_TpFwEventInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpFwEventInfo;
    guint32   u_octet4_loop_org_csapi_fw_TpFwEventInfo_ServiceIDList;
    guint32   i_org_csapi_fw_TpFwEventInfo_ServiceIDList;
    guint32   u_octet4_loop_org_csapi_fw_TpFwEventInfo_UnavailableServiceIDList;
    guint32   i_org_csapi_fw_TpFwEventInfo_UnavailableServiceIDList;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/fw/TpFwEventInfo:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_fw_TpFwEventInfo_TpFwEventInfo, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/fw/TpFwEventName:1.0
     */

    disc_s_TpFwEventInfo = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpFwEventInfo == 0 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpFwEventInfo_EventNameUndefined);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpFwEventInfo == 1 ) {

        u_octet4_loop_org_csapi_fw_TpFwEventInfo_ServiceIDList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_TpFwEventInfo_ServiceIDList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_TpFwEventInfo_ServiceIDList);

        for (i_org_csapi_fw_TpFwEventInfo_ServiceIDList=0; i_org_csapi_fw_TpFwEventInfo_ServiceIDList < u_octet4_loop_org_csapi_fw_TpFwEventInfo_ServiceIDList; i_org_csapi_fw_TpFwEventInfo_ServiceIDList++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpFwEventInfo_ServiceIDList);

        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpFwEventInfo == 2 ) {

        u_octet4_loop_org_csapi_fw_TpFwEventInfo_UnavailableServiceIDList = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_fw_TpFwEventInfo_UnavailableServiceIDList_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_fw_TpFwEventInfo_UnavailableServiceIDList);

        for (i_org_csapi_fw_TpFwEventInfo_UnavailableServiceIDList=0; i_org_csapi_fw_TpFwEventInfo_UnavailableServiceIDList < u_octet4_loop_org_csapi_fw_TpFwEventInfo_UnavailableServiceIDList; i_org_csapi_fw_TpFwEventInfo_UnavailableServiceIDList++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpFwEventInfo_UnavailableServiceIDList);

        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpFwEventInfo == 3 ) {

        /*  Begin struct "org_csapi_fw_TpFwMigrationServiceAvailableInfo"  */
        decode_org_csapi_fw_TpFwMigrationServiceAvailableInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpFwMigrationServiceAvailableInfo"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpFwEventInfo == 4 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpFwEventInfo_AppSessionCreated);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpFwEventInfo == 5 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_fw_TpFwEventInfo_AppSessionTerminated);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpFwEventInfo == 6 ) {

        /*  Begin struct "org_csapi_fw_TpFwAgreementInfo"  */
        decode_org_csapi_fw_TpFwAgreementInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpFwAgreementInfo"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpFwEventInfo == 7 ) {

        /*  Begin struct "org_csapi_fw_TpFwAgreementInfo"  */
        decode_org_csapi_fw_TpFwAgreementInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_fw_TpFwAgreementInfo"  */
        return;     /* End Compare for this discriminant type */
    }

}

/* Union = IDL:org/csapi/ui/TpUIVariableInfo:1.0 */
static void
decode_org_csapi_ui_TpUIVariableInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpUIVariableInfo;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/ui/TpUIVariableInfo:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_ui_TpUIVariableInfo_TpUIVariableInfo, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/ui/TpUIVariablePartType:1.0
     */

    disc_s_TpUIVariableInfo = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpUIVariableInfo == 0 ) {

        proto_tree_add_int(tree, hf_org_csapi_ui_TpUIVariableInfo_VariablePartInteger, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpUIVariableInfo == 1 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_TpUIVariableInfo_VariablePartAddress);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpUIVariableInfo == 2 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_TpUIVariableInfo_VariablePartTime);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpUIVariableInfo == 3 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_TpUIVariableInfo_VariablePartDate);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpUIVariableInfo == 4 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_TpUIVariableInfo_VariablePartPrice);

        return;     /* End Compare for this discriminant type */
    }

}

/* Union = IDL:org/csapi/ui/TpUIInfo:1.0 */
static void
decode_org_csapi_ui_TpUIInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpUIInfo;
    guint32   u_octet4_loop_org_csapi_ui_TpUIInfo_InfoBinData;
    guint32   i_org_csapi_ui_TpUIInfo_InfoBinData;
    guint32   u_octet4_loop_org_csapi_ui_TpUIInfo_InfoMimeData;
    guint32   i_org_csapi_ui_TpUIInfo_InfoMimeData;
    guint32   u_octet4_loop_org_csapi_ui_TpUIInfo_InfoWaveData;
    guint32   i_org_csapi_ui_TpUIInfo_InfoWaveData;
    guint32   u_octet4_loop_org_csapi_ui_TpUIInfo_InfoAuData;
    guint32   i_org_csapi_ui_TpUIInfo_InfoAuData;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/ui/TpUIInfo:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_ui_TpUIInfo_TpUIInfo, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/ui/TpUIInfoType:1.0
     */

    disc_s_TpUIInfo = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpUIInfo == 0 ) {

        proto_tree_add_int(tree, hf_org_csapi_ui_TpUIInfo_InfoID, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpUIInfo == 1 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_TpUIInfo_InfoData);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpUIInfo == 2 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_TpUIInfo_InfoAddress);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpUIInfo == 3 ) {

        u_octet4_loop_org_csapi_ui_TpUIInfo_InfoBinData = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_TpUIInfo_InfoBinData_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_ui_TpUIInfo_InfoBinData);

        for (i_org_csapi_ui_TpUIInfo_InfoBinData=0; i_org_csapi_ui_TpUIInfo_InfoBinData < u_octet4_loop_org_csapi_ui_TpUIInfo_InfoBinData; i_org_csapi_ui_TpUIInfo_InfoBinData++) {

            proto_tree_add_uint(tree, hf_org_csapi_ui_TpUIInfo_InfoBinData, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpUIInfo == 4 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_TpUIInfo_InfoUUEncData);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpUIInfo == 5 ) {

        u_octet4_loop_org_csapi_ui_TpUIInfo_InfoMimeData = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_TpUIInfo_InfoMimeData_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_ui_TpUIInfo_InfoMimeData);

        for (i_org_csapi_ui_TpUIInfo_InfoMimeData=0; i_org_csapi_ui_TpUIInfo_InfoMimeData < u_octet4_loop_org_csapi_ui_TpUIInfo_InfoMimeData; i_org_csapi_ui_TpUIInfo_InfoMimeData++) {

            proto_tree_add_uint(tree, hf_org_csapi_ui_TpUIInfo_InfoMimeData, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpUIInfo == 6 ) {

        u_octet4_loop_org_csapi_ui_TpUIInfo_InfoWaveData = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_TpUIInfo_InfoWaveData_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_ui_TpUIInfo_InfoWaveData);

        for (i_org_csapi_ui_TpUIInfo_InfoWaveData=0; i_org_csapi_ui_TpUIInfo_InfoWaveData < u_octet4_loop_org_csapi_ui_TpUIInfo_InfoWaveData; i_org_csapi_ui_TpUIInfo_InfoWaveData++) {

            proto_tree_add_uint(tree, hf_org_csapi_ui_TpUIInfo_InfoWaveData, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpUIInfo == 7 ) {

        u_octet4_loop_org_csapi_ui_TpUIInfo_InfoAuData = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_ui_TpUIInfo_InfoAuData_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_ui_TpUIInfo_InfoAuData);

        for (i_org_csapi_ui_TpUIInfo_InfoAuData=0; i_org_csapi_ui_TpUIInfo_InfoAuData < u_octet4_loop_org_csapi_ui_TpUIInfo_InfoAuData; i_org_csapi_ui_TpUIInfo_InfoAuData++) {

            proto_tree_add_uint(tree, hf_org_csapi_ui_TpUIInfo_InfoAuData, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpUIInfo == 8 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_ui_TpUIInfo_InfoVXMLData);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpUIInfo == 9 ) {

        /*  Begin struct "org_csapi_ui_TpUISynthesisInfoData"  */
        decode_org_csapi_ui_TpUISynthesisInfoData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_ui_TpUISynthesisInfoData"  */
        return;     /* End Compare for this discriminant type */
    }

}

/* Union = IDL:org/csapi/cc/TpCallAdditionalErrorInfo:1.0 */
static void
decode_org_csapi_cc_TpCallAdditionalErrorInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpCallAdditionalErrorInfo;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/cc/TpCallAdditionalErrorInfo:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallAdditionalErrorInfo_TpCallAdditionalErrorInfo, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/cc/TpCallErrorType:1.0
     */

    disc_s_TpCallAdditionalErrorInfo = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpCallAdditionalErrorInfo == 1 ) {

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallAdditionalErrorInfo_CallErrorInvalidAddress, tvb, *offset-4, 4, u_octet4);

        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_cc_TpCallAdditionalErrorInfo_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/cc/TpCallLoadControlMechanism:1.0 */
static void
decode_org_csapi_cc_TpCallLoadControlMechanism_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpCallLoadControlMechanism;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/cc/TpCallLoadControlMechanism:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallLoadControlMechanism_TpCallLoadControlMechanism, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/cc/TpCallLoadControlMechanismType:1.0
     */

    disc_s_TpCallLoadControlMechanism = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpCallLoadControlMechanism == 0 ) {

        proto_tree_add_int(tree, hf_org_csapi_cc_TpCallLoadControlMechanism_CallLoadControlPerInterval, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

}

/* Union = IDL:org/csapi/cc/TpCallAdditionalTreatmentInfo:1.0 */
static void
decode_org_csapi_cc_TpCallAdditionalTreatmentInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpCallAdditionalTreatmentInfo;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/cc/TpCallAdditionalTreatmentInfo:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallAdditionalTreatmentInfo_TpCallAdditionalTreatmentInfo, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/cc/TpCallTreatmentType:1.0
     */

    disc_s_TpCallAdditionalTreatmentInfo = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpCallAdditionalTreatmentInfo == 2 ) {

        /*  Begin union "org_csapi_ui_TpUIInfo"  */
        decode_org_csapi_ui_TpUIInfo_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_ui_TpUIInfo"  */
        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_cc_TpCallAdditionalTreatmentInfo_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/cc/TpCallPartyToChargeAdditionalInfo:1.0 */
static void
decode_org_csapi_cc_TpCallPartyToChargeAdditionalInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpCallPartyToChargeAdditionalInfo;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/cc/TpCallPartyToChargeAdditionalInfo:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallPartyToChargeAdditionalInfo_TpCallPartyToChargeAdditionalInfo, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/cc/TpCallPartyToChargeType:1.0
     */

    disc_s_TpCallPartyToChargeAdditionalInfo = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpCallPartyToChargeAdditionalInfo == 2 ) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_cc_TpCallPartyToChargeAdditionalInfo_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/cc/gccs/TpCallAppInfo:1.0 */
static void
decode_org_csapi_cc_gccs_TpCallAppInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpCallAppInfo;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/cc/gccs/TpCallAppInfo:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_gccs_TpCallAppInfo_TpCallAppInfo, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/cc/gccs/TpCallAppInfoType:1.0
     */

    disc_s_TpCallAppInfo = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpCallAppInfo == 1 ) {

        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_TpCallAppInfo_CallAppAlertingMechanism, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAppInfo == 2 ) {

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_gccs_TpCallAppInfo_CallAppNetworkAccessType, tvb, *offset-4, 4, u_octet4);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAppInfo == 3 ) {

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_gccs_TpCallAppInfo_CallAppTeleService, tvb, *offset-4, 4, u_octet4);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAppInfo == 4 ) {

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_gccs_TpCallAppInfo_CallAppBearerService, tvb, *offset-4, 4, u_octet4);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAppInfo == 5 ) {

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_gccs_TpCallAppInfo_CallAppPartyCategory, tvb, *offset-4, 4, u_octet4);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAppInfo == 6 ) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAppInfo == 7 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_gccs_TpCallAppInfo_CallAppGenericInfo);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAppInfo == 8 ) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_cc_gccs_TpCallAppInfo_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/cc/gccs/TpCallAdditionalReportInfo:1.0 */
static void
decode_org_csapi_cc_gccs_TpCallAdditionalReportInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpCallAdditionalReportInfo;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/cc/gccs/TpCallAdditionalReportInfo:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_gccs_TpCallAdditionalReportInfo_TpCallAdditionalReportInfo, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/cc/gccs/TpCallReportType:1.0
     */

    disc_s_TpCallAdditionalReportInfo = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpCallAdditionalReportInfo == 4 ) {

        /*  Begin struct "org_csapi_cc_gccs_TpCallReleaseCause"  */
        decode_org_csapi_cc_gccs_TpCallReleaseCause_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_gccs_TpCallReleaseCause"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAdditionalReportInfo == 6 ) {

        /*  Begin struct "org_csapi_cc_gccs_TpCallReleaseCause"  */
        decode_org_csapi_cc_gccs_TpCallReleaseCause_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_gccs_TpCallReleaseCause"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAdditionalReportInfo == 7 ) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAdditionalReportInfo == 8 ) {

        /*  Begin struct "org_csapi_cc_TpCallServiceCode"  */
        decode_org_csapi_cc_TpCallServiceCode_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallServiceCode"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAdditionalReportInfo == 9 ) {

        /*  Begin struct "org_csapi_cc_gccs_TpCallReleaseCause"  */
        decode_org_csapi_cc_gccs_TpCallReleaseCause_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_gccs_TpCallReleaseCause"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAdditionalReportInfo == 10 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_gccs_TpCallAdditionalReportInfo_QueueStatus);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAdditionalReportInfo == 11 ) {

        /*  Begin struct "org_csapi_cc_gccs_TpCallReleaseCause"  */
        decode_org_csapi_cc_gccs_TpCallReleaseCause_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_gccs_TpCallReleaseCause"  */
        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_cc_gccs_TpCallAdditionalReportInfo_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/cc/gccs/TpCallAdditionalReportCriteria:1.0 */
static void
decode_org_csapi_cc_gccs_TpCallAdditionalReportCriteria_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpCallAdditionalReportCriteria;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/cc/gccs/TpCallAdditionalReportCriteria:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_gccs_TpCallAdditionalReportCriteria_TpCallAdditionalReportCriteria, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/cc/gccs/TpCallReportType:1.0
     */

    disc_s_TpCallAdditionalReportCriteria = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpCallAdditionalReportCriteria == 5 ) {

        proto_tree_add_int(tree, hf_org_csapi_cc_gccs_TpCallAdditionalReportCriteria_NoAnswerDuration, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAdditionalReportCriteria == 8 ) {

        /*  Begin struct "org_csapi_cc_TpCallServiceCode"  */
        decode_org_csapi_cc_TpCallServiceCode_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallServiceCode"  */
        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_cc_gccs_TpCallAdditionalReportCriteria_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/cc/TpCallAdditionalEventInfo:1.0 */
static void
decode_org_csapi_cc_TpCallAdditionalEventInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpCallAdditionalEventInfo;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/cc/TpCallAdditionalEventInfo:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallAdditionalEventInfo_TpCallAdditionalEventInfo, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/cc/TpCallEventType:1.0
     */

    disc_s_TpCallAdditionalEventInfo = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpCallAdditionalEventInfo == 3 ) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAdditionalEventInfo == 4 ) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAdditionalEventInfo == 5 ) {

        /*  Begin struct "org_csapi_cc_TpCallServiceCode"  */
        decode_org_csapi_cc_TpCallServiceCode_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallServiceCode"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAdditionalEventInfo == 6 ) {

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallAdditionalEventInfo_OriginatingReleaseCause, tvb, *offset-4, 4, u_octet4);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAdditionalEventInfo == 11 ) {

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallAdditionalEventInfo_TerminatingReleaseCause, tvb, *offset-4, 4, u_octet4);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAdditionalEventInfo == 12 ) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAdditionalEventInfo == 13 ) {

        /*  Begin struct "org_csapi_cc_TpCallServiceCode"  */
        decode_org_csapi_cc_TpCallServiceCode_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_TpCallServiceCode"  */
        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_cc_TpCallAdditionalEventInfo_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/cc/TpAdditionalCallEventCriteria:1.0 */
static void
decode_org_csapi_cc_TpAdditionalCallEventCriteria_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpAdditionalCallEventCriteria;
    guint32   u_octet4_loop_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingServiceCode;
    guint32   i_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingServiceCode;
    guint32   u_octet4_loop_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingReleaseCauseSet;
    guint32   i_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingReleaseCauseSet;
    guint32   u_octet4_loop_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingReleaseCauseSet;
    guint32   i_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingReleaseCauseSet;
    guint32   u_octet4_loop_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingServiceCode;
    guint32   i_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingServiceCode;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/cc/TpAdditionalCallEventCriteria:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpAdditionalCallEventCriteria_TpAdditionalCallEventCriteria, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/cc/TpCallEventType:1.0
     */

    disc_s_TpAdditionalCallEventCriteria = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpAdditionalCallEventCriteria == 3 ) {

        proto_tree_add_int(tree, hf_org_csapi_cc_TpAdditionalCallEventCriteria_MinAddressLength, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpAdditionalCallEventCriteria == 5 ) {

        u_octet4_loop_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingServiceCode = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingServiceCode_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingServiceCode);

        for (i_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingServiceCode=0; i_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingServiceCode < u_octet4_loop_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingServiceCode; i_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingServiceCode++) {

            /*  Begin struct "org_csapi_cc_TpCallServiceCode"  */
            decode_org_csapi_cc_TpCallServiceCode_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cc_TpCallServiceCode"  */
        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpAdditionalCallEventCriteria == 6 ) {

        u_octet4_loop_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingReleaseCauseSet = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingReleaseCauseSet_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingReleaseCauseSet);

        for (i_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingReleaseCauseSet=0; i_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingReleaseCauseSet < u_octet4_loop_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingReleaseCauseSet; i_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingReleaseCauseSet++) {

            u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingReleaseCauseSet, tvb, *offset-4, 4, u_octet4);

        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpAdditionalCallEventCriteria == 11 ) {

        u_octet4_loop_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingReleaseCauseSet = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingReleaseCauseSet_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingReleaseCauseSet);

        for (i_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingReleaseCauseSet=0; i_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingReleaseCauseSet < u_octet4_loop_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingReleaseCauseSet; i_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingReleaseCauseSet++) {

            u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
            proto_tree_add_uint(tree, hf_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingReleaseCauseSet, tvb, *offset-4, 4, u_octet4);

        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpAdditionalCallEventCriteria == 13 ) {

        u_octet4_loop_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingServiceCode = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingServiceCode_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingServiceCode);

        for (i_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingServiceCode=0; i_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingServiceCode < u_octet4_loop_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingServiceCode; i_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingServiceCode++) {

            /*  Begin struct "org_csapi_cc_TpCallServiceCode"  */
            decode_org_csapi_cc_TpCallServiceCode_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cc_TpCallServiceCode"  */
        }

        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_cc_TpAdditionalCallEventCriteria_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/cc/TpCallAppInfo:1.0 */
static void
decode_org_csapi_cc_TpCallAppInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpCallAppInfo;
    guint32   u_octet4_loop_org_csapi_cc_TpCallAppInfo_CallAppCarrier;
    guint32   i_org_csapi_cc_TpCallAppInfo_CallAppCarrier;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/cc/TpCallAppInfo:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallAppInfo_TpCallAppInfo, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/cc/TpCallAppInfoType:1.0
     */

    disc_s_TpCallAppInfo = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpCallAppInfo == 1 ) {

        proto_tree_add_int(tree, hf_org_csapi_cc_TpCallAppInfo_CallAppAlertingMechanism, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAppInfo == 2 ) {

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallAppInfo_CallAppNetworkAccessType, tvb, *offset-4, 4, u_octet4);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAppInfo == 3 ) {

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallAppInfo_CallAppTeleService, tvb, *offset-4, 4, u_octet4);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAppInfo == 4 ) {

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallAppInfo_CallAppBearerService, tvb, *offset-4, 4, u_octet4);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAppInfo == 5 ) {

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallAppInfo_CallAppPartyCategory, tvb, *offset-4, 4, u_octet4);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAppInfo == 6 ) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAppInfo == 7 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cc_TpCallAppInfo_CallAppGenericInfo);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAppInfo == 8 ) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAppInfo == 9 ) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAppInfo == 10 ) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAppInfo == 11 ) {

        proto_tree_add_int(tree, hf_org_csapi_cc_TpCallAppInfo_CallHighProbabilityCompletion, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpCallAppInfo == 12 ) {

        u_octet4_loop_org_csapi_cc_TpCallAppInfo_CallAppCarrier = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cc_TpCallAppInfo_CallAppCarrier_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cc_TpCallAppInfo_CallAppCarrier);

        for (i_org_csapi_cc_TpCallAppInfo_CallAppCarrier=0; i_org_csapi_cc_TpCallAppInfo_CallAppCarrier < u_octet4_loop_org_csapi_cc_TpCallAppInfo_CallAppCarrier; i_org_csapi_cc_TpCallAppInfo_CallAppCarrier++) {

            /*  Begin struct "org_csapi_cc_TpCarrier"  */
            decode_org_csapi_cc_TpCarrier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_cc_TpCarrier"  */
        }

        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_cc_TpCallAppInfo_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/cc/mpccs/TpAppMultiPartyCallBack:1.0 */
static void
decode_org_csapi_cc_mpccs_TpAppMultiPartyCallBack_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpAppMultiPartyCallBack;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/cc/mpccs/TpAppMultiPartyCallBack:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_mpccs_TpAppMultiPartyCallBack_TpAppMultiPartyCallBack, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/cc/mpccs/TpAppMultiPartyCallBackRefType:1.0
     */

    disc_s_TpAppMultiPartyCallBack = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpAppMultiPartyCallBack == 1 ) {

        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpAppMultiPartyCallBack == 2 ) {

        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpAppMultiPartyCallBack == 3 ) {

        /*  Begin struct "org_csapi_cc_mpccs_TpAppCallLegCallBack"  */
        decode_org_csapi_cc_mpccs_TpAppCallLegCallBack_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_mpccs_TpAppCallLegCallBack"  */
        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_cc_mpccs_TpAppMultiPartyCallBack_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/cc/mmccs/TpMediaStreamDataTypeRequest:1.0 */
static void
decode_org_csapi_cc_mmccs_TpMediaStreamDataTypeRequest_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpMediaStreamDataTypeRequest;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/cc/mmccs/TpMediaStreamDataTypeRequest:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_mmccs_TpMediaStreamDataTypeRequest_TpMediaStreamDataTypeRequest, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/cc/mmccs/TpMediaStreamDataTypeRequestType:1.0
     */

    disc_s_TpMediaStreamDataTypeRequest = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpMediaStreamDataTypeRequest == 0 ) {

        proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_TpMediaStreamDataTypeRequest_Audio, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMediaStreamDataTypeRequest == 1 ) {

        proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_TpMediaStreamDataTypeRequest_Video, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMediaStreamDataTypeRequest == 2 ) {

        proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_TpMediaStreamDataTypeRequest_Data, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

}

/* Union = IDL:org/csapi/cc/mmccs/TpAppMultiMediaCallBack:1.0 */
static void
decode_org_csapi_cc_mmccs_TpAppMultiMediaCallBack_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpAppMultiMediaCallBack;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/cc/mmccs/TpAppMultiMediaCallBack:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_mmccs_TpAppMultiMediaCallBack_TpAppMultiMediaCallBack, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/cc/mmccs/TpAppMultiMediaCallBackRefType:1.0
     */

    disc_s_TpAppMultiMediaCallBack = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpAppMultiMediaCallBack == 1 ) {

        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpAppMultiMediaCallBack == 2 ) {

        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpAppMultiMediaCallBack == 3 ) {

        /*  Begin struct "org_csapi_cc_mmccs_TpAppMultiMediaCallLegCallBack"  */
        decode_org_csapi_cc_mmccs_TpAppMultiMediaCallLegCallBack_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_mmccs_TpAppMultiMediaCallLegCallBack"  */
        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_cc_mmccs_TpAppMultiMediaCallBack_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/cc/cccs/TpConfPolicy:1.0 */
static void
decode_org_csapi_cc_cccs_TpConfPolicy_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpConfPolicy;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/cc/cccs/TpConfPolicy:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cc_cccs_TpConfPolicy_TpConfPolicy, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/cc/cccs/TpConfPolicyType:1.0
     */

    disc_s_TpConfPolicy = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpConfPolicy == 1 ) {

        /*  Begin struct "org_csapi_cc_cccs_TpMonoMediaConfPolicy"  */
        decode_org_csapi_cc_cccs_TpMonoMediaConfPolicy_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_cccs_TpMonoMediaConfPolicy"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpConfPolicy == 2 ) {

        /*  Begin struct "org_csapi_cc_cccs_TpMultiMediaConfPolicy"  */
        decode_org_csapi_cc_cccs_TpMultiMediaConfPolicy_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_cccs_TpMultiMediaConfPolicy"  */
        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_cc_cccs_TpConfPolicy_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/ui/TpUITargetObject:1.0 */
static void
decode_org_csapi_ui_TpUITargetObject_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpUITargetObject;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/ui/TpUITargetObject:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_ui_TpUITargetObject_TpUITargetObject, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/ui/TpUITargetObjectType:1.0
     */

    disc_s_TpUITargetObject = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpUITargetObject == 0 ) {

        /*  Begin struct "org_csapi_cc_gccs_TpCallIdentifier"  */
        decode_org_csapi_cc_gccs_TpCallIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_gccs_TpCallIdentifier"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpUITargetObject == 1 ) {

        /*  Begin struct "org_csapi_cc_mpccs_TpMultiPartyCallIdentifier"  */
        decode_org_csapi_cc_mpccs_TpMultiPartyCallIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_mpccs_TpMultiPartyCallIdentifier"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpUITargetObject == 2 ) {

        /*  Begin struct "org_csapi_cc_mpccs_TpCallLegIdentifier"  */
        decode_org_csapi_cc_mpccs_TpCallLegIdentifier_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_cc_mpccs_TpCallLegIdentifier"  */
        return;     /* End Compare for this discriminant type */
    }

}

/* Union = IDL:org/csapi/mm/TpBindingEntry:1.0 */
static void
decode_org_csapi_mm_TpBindingEntry_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpBindingEntry;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/mm/TpBindingEntry:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mm_TpBindingEntry_TpBindingEntry, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/mm/TpBindingEntryType:1.0
     */

    disc_s_TpBindingEntry = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpBindingEntry == 1 ) {

        proto_tree_add_int(tree, hf_org_csapi_mm_TpBindingEntry_UBExpires, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpBindingEntry == 2 ) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpBindingEntry == 3 ) {

        proto_tree_add_float(tree, hf_org_csapi_mm_TpBindingEntry_UBPreference, tvb, *offset-4, 4, get_CDR_float(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_mm_TpBindingEntry_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/dsc/TpDataSessionChargeOrder:1.0 */
static void
decode_org_csapi_dsc_TpDataSessionChargeOrder_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpDataSessionChargeOrder;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/dsc/TpDataSessionChargeOrder:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_dsc_TpDataSessionChargeOrder_TpDataSessionChargeOrder, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/dsc/TpDataSessionChargeOrderCategory:1.0
     */

    disc_s_TpDataSessionChargeOrder = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpDataSessionChargeOrder == 0 ) {

        /*  Begin struct "org_csapi_dsc_TpChargePerVolume"  */
        decode_org_csapi_dsc_TpChargePerVolume_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_dsc_TpChargePerVolume"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpDataSessionChargeOrder == 1 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_dsc_TpDataSessionChargeOrder_NetworkCharge);

        return;     /* End Compare for this discriminant type */
    }

}

/* Union = IDL:org/csapi/dsc/TpDataSessionAdditionalErrorInfo:1.0 */
static void
decode_org_csapi_dsc_TpDataSessionAdditionalErrorInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpDataSessionAdditionalErrorInfo;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/dsc/TpDataSessionAdditionalErrorInfo:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_dsc_TpDataSessionAdditionalErrorInfo_TpDataSessionAdditionalErrorInfo, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/dsc/TpDataSessionErrorType:1.0
     */

    disc_s_TpDataSessionAdditionalErrorInfo = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpDataSessionAdditionalErrorInfo == 1 ) {

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_dsc_TpDataSessionAdditionalErrorInfo_DataSessionErrorInvalidAddress, tvb, *offset-4, 4, u_octet4);

        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_dsc_TpDataSessionAdditionalErrorInfo_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/dsc/TpDataSessionAdditionalReportInfo:1.0 */
static void
decode_org_csapi_dsc_TpDataSessionAdditionalReportInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpDataSessionAdditionalReportInfo;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/dsc/TpDataSessionAdditionalReportInfo:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_dsc_TpDataSessionAdditionalReportInfo_TpDataSessionAdditionalReportInfo, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/dsc/TpDataSessionReportType:1.0
     */

    disc_s_TpDataSessionAdditionalReportInfo = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpDataSessionAdditionalReportInfo == 2 ) {

        /*  Begin struct "org_csapi_dsc_TpDataSessionReleaseCause"  */
        decode_org_csapi_dsc_TpDataSessionReleaseCause_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_dsc_TpDataSessionReleaseCause"  */
        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_dsc_TpDataSessionAdditionalReportInfo_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/gms/TpMessageInfoProperty:1.0 */
static void
decode_org_csapi_gms_TpMessageInfoProperty_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpMessageInfoProperty;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/gms/TpMessageInfoProperty:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_gms_TpMessageInfoProperty_TpMessageInfoProperty, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/gms/TpMessageInfoPropertyName:1.0
     */

    disc_s_TpMessageInfoProperty = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpMessageInfoProperty == 1 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageID);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageInfoProperty == 2 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageSubject);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageInfoProperty == 3 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageDateSent);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageInfoProperty == 4 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageDateReceived);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageInfoProperty == 5 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageDateChanged);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageInfoProperty == 6 ) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageInfoProperty == 7 ) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageInfoProperty == 8 ) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageInfoProperty == 9 ) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageInfoProperty == 10 ) {

        proto_tree_add_int(tree, hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageSize, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageInfoProperty == 11 ) {

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessagePriority, tvb, *offset-4, 4, u_octet4);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageInfoProperty == 12 ) {

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageFormat, tvb, *offset-4, 4, u_octet4);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageInfoProperty == 13 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageFolder);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageInfoProperty == 14 ) {

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageStatus, tvb, *offset-4, 4, u_octet4);

        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_gms_TpMessageInfoProperty_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/gms/TpMailboxInfoProperty:1.0 */
static void
decode_org_csapi_gms_TpMailboxInfoProperty_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpMailboxInfoProperty;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/gms/TpMailboxInfoProperty:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_gms_TpMailboxInfoProperty_TpMailboxInfoProperty, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/gms/TpMailboxInfoPropertyName:1.0
     */

    disc_s_TpMailboxInfoProperty = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpMailboxInfoProperty == 1 ) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMailboxInfoProperty == 2 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_TpMailboxInfoProperty_MessagingMailboxOwner);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMailboxInfoProperty == 3 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_TpMailboxInfoProperty_MessagingMailboxFolder);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMailboxInfoProperty == 4 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_TpMailboxInfoProperty_MessagingMailboxDateCreated);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMailboxInfoProperty == 5 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_TpMailboxInfoProperty_MessagingMailboxDateChanged);

        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_gms_TpMailboxInfoProperty_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/gms/TpMessagingEventInfo:1.0 */
static void
decode_org_csapi_gms_TpMessagingEventInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpMessagingEventInfo;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/gms/TpMessagingEventInfo:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_gms_TpMessagingEventInfo_TpMessagingEventInfo, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/gms/TpMessagingEventName:1.0
     */

    disc_s_TpMessagingEventInfo = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpMessagingEventInfo == 0 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_TpMessagingEventInfo_EventNameUndefined);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessagingEventInfo == 1 ) {

        /*  Begin struct "org_csapi_gms_TpGMSNewMessageArrivedInfo"  */
        decode_org_csapi_gms_TpGMSNewMessageArrivedInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_gms_TpGMSNewMessageArrivedInfo"  */
        return;     /* End Compare for this discriminant type */
    }

}

/* Union = IDL:org/csapi/gms/TpMessagingEventCriteria:1.0 */
static void
decode_org_csapi_gms_TpMessagingEventCriteria_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpMessagingEventCriteria;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/gms/TpMessagingEventCriteria:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_gms_TpMessagingEventCriteria_TpMessagingEventCriteria, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/gms/TpMessagingEventName:1.0
     */

    disc_s_TpMessagingEventCriteria = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpMessagingEventCriteria == 1 ) {

        /*  Begin struct "org_csapi_gms_TpGMSNewMessageArrivedCriteria"  */
        decode_org_csapi_gms_TpGMSNewMessageArrivedCriteria_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_gms_TpGMSNewMessageArrivedCriteria"  */
        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_gms_TpMessagingEventCriteria_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/gms/TpFolderInfoProperty:1.0 */
static void
decode_org_csapi_gms_TpFolderInfoProperty_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpFolderInfoProperty;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/gms/TpFolderInfoProperty:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_gms_TpFolderInfoProperty_TpFolderInfoProperty, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/gms/TpFolderInfoPropertyName:1.0
     */

    disc_s_TpFolderInfoProperty = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpFolderInfoProperty == 1 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_TpFolderInfoProperty_MessagingFolderID);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpFolderInfoProperty == 2 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_TpFolderInfoProperty_MessagingFolderMessage);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpFolderInfoProperty == 3 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_TpFolderInfoProperty_MessagingFolderSubfolder);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpFolderInfoProperty == 4 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_TpFolderInfoProperty_MessagingFolderDateCreated);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpFolderInfoProperty == 5 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_gms_TpFolderInfoProperty_MessagingFolderDateChanged);

        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_gms_TpFolderInfoProperty_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/cs/TpChargingParameterValue:1.0 */
static void
decode_org_csapi_cs_TpChargingParameterValue_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpChargingParameterValue;
    guint32   u_octet4_loop_org_csapi_cs_TpChargingParameterValue_OctetValue;
    guint32   i_org_csapi_cs_TpChargingParameterValue_OctetValue;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/cs/TpChargingParameterValue:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cs_TpChargingParameterValue_TpChargingParameterValue, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/cs/TpChargingParameterValueType:1.0
     */

    disc_s_TpChargingParameterValue = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpChargingParameterValue == 0 ) {

        proto_tree_add_int(tree, hf_org_csapi_cs_TpChargingParameterValue_IntValue, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpChargingParameterValue == 1 ) {

        proto_tree_add_float(tree, hf_org_csapi_cs_TpChargingParameterValue_FloatValue, tvb, *offset-4, 4, get_CDR_float(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpChargingParameterValue == 2 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cs_TpChargingParameterValue_StringValue);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpChargingParameterValue == 3 ) {

        proto_tree_add_boolean(tree, hf_org_csapi_cs_TpChargingParameterValue_BooleanValue, tvb, *offset-1, 1, get_CDR_boolean(tvb,offset));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpChargingParameterValue == 4 ) {

        u_octet4_loop_org_csapi_cs_TpChargingParameterValue_OctetValue = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_cs_TpChargingParameterValue_OctetValue_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_cs_TpChargingParameterValue_OctetValue);

        for (i_org_csapi_cs_TpChargingParameterValue_OctetValue=0; i_org_csapi_cs_TpChargingParameterValue_OctetValue < u_octet4_loop_org_csapi_cs_TpChargingParameterValue_OctetValue; i_org_csapi_cs_TpChargingParameterValue_OctetValue++) {

            proto_tree_add_uint(tree, hf_org_csapi_cs_TpChargingParameterValue_OctetValue, tvb, *offset-1, 1, get_CDR_octet(tvb,offset));

        }

        return;     /* End Compare for this discriminant type */
    }

}

/* Union = IDL:org/csapi/cs/TpAppInformation:1.0 */
static void
decode_org_csapi_cs_TpAppInformation_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpAppInformation;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/cs/TpAppInformation:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_cs_TpAppInformation_TpAppInformation, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/cs/TpAppInformationType:1.0
     */

    disc_s_TpAppInformation = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpAppInformation == 0 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_cs_TpAppInformation_Timestamp);

        return;     /* End Compare for this discriminant type */
    }

}

/* Union = IDL:org/csapi/policy/TpPolicyType:1.0 */
// NOLINTBEGIN(misc-no-recursion)
static void
decode_org_csapi_policy_TpPolicyType_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpPolicyType;
    /* Operation specific Variable declarations End */

    // Cycle: this -> decode_org_csapi_policy_TpPolicyType_TpPolicyListType_st -> this
    unsigned recursion_depth = p_get_proto_depth(pinfo, proto_parlay);
    DISSECTOR_ASSERT(recursion_depth <= MAX_RECURSION_DEPTH);
    p_set_proto_depth(pinfo, proto_parlay, recursion_depth + 1);

    /*
     * IDL Union Start - IDL:org/csapi/policy/TpPolicyType:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_policy_TpPolicyType_TpPolicyType, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/policy/TpPolicyTypeInfo:1.0
     */

    disc_s_TpPolicyType = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpPolicyType == 0 ) {

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_policy_TpPolicyType_SimpleType, tvb, *offset-4, 4, u_octet4);

        p_set_proto_depth(pinfo, proto_parlay, recursion_depth);
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPolicyType == 1 ) {

        /*  Begin struct "org_csapi_policy_TpPolicyType_TpPolicyRecordType"  */
        decode_org_csapi_policy_TpPolicyType_TpPolicyRecordType_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_policy_TpPolicyType_TpPolicyRecordType"  */
        p_set_proto_depth(pinfo, proto_parlay, recursion_depth);
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPolicyType == 2 ) {

        /*  Begin struct "org_csapi_policy_TpPolicyType_TpPolicyListType"  */
        decode_org_csapi_policy_TpPolicyType_TpPolicyListType_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_policy_TpPolicyType_TpPolicyListType"  */
        p_set_proto_depth(pinfo, proto_parlay, recursion_depth);
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPolicyType == 3 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_TpPolicyType_StructuredType);

        p_set_proto_depth(pinfo, proto_parlay, recursion_depth);
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPolicyType == 4 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_policy_TpPolicyType_XMLString);

        p_set_proto_depth(pinfo, proto_parlay, recursion_depth);
        return;     /* End Compare for this discriminant type */
    }

    p_set_proto_depth(pinfo, proto_parlay, recursion_depth);
}
// NOLINTEND(misc-no-recursion)

/* Union = IDL:org/csapi/pam/TpPAMContextData:1.0 */
static void
decode_org_csapi_pam_TpPAMContextData_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpPAMContextData;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/pam/TpPAMContextData:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMContextData_TpPAMContextData, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/pam/TpPAMContextName:1.0
     */

    disc_s_TpPAMContextData = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpPAMContextData == 1 ) {

        /*  Begin struct "org_csapi_pam_TpPAMCommunicationContext"  */
        decode_org_csapi_pam_TpPAMCommunicationContext_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMCommunicationContext"  */
        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_pam_TpPAMContextData_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/pam/TpPAMEventInfo:1.0 */
static void
decode_org_csapi_pam_TpPAMEventInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpPAMEventInfo;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/pam/TpPAMEventInfo:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMEventInfo_TpPAMEventInfo, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/pam/TpPAMEventName:1.0
     */

    disc_s_TpPAMEventInfo = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpPAMEventInfo == 0 ) {

        /*  Begin struct "org_csapi_pam_TpPAMIPSEventData"  */
        decode_org_csapi_pam_TpPAMIPSEventData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMIPSEventData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMEventInfo == 1 ) {

        /*  Begin struct "org_csapi_pam_TpPAMAVCEventData"  */
        decode_org_csapi_pam_TpPAMAVCEventData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMAVCEventData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMEventInfo == 2 ) {

        /*  Begin struct "org_csapi_pam_TpPAMWCEventData"  */
        decode_org_csapi_pam_TpPAMWCEventData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMWCEventData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMEventInfo == 3 ) {

        /*  Begin struct "org_csapi_pam_TpPAMICEventData"  */
        decode_org_csapi_pam_TpPAMICEventData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMICEventData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMEventInfo == 4 ) {

        /*  Begin struct "org_csapi_pam_TpPAMIDEventData"  */
        decode_org_csapi_pam_TpPAMIDEventData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMIDEventData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMEventInfo == 5 ) {

        /*  Begin struct "org_csapi_pam_TpPAMGMCEventData"  */
        decode_org_csapi_pam_TpPAMGMCEventData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMGMCEventData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMEventInfo == 6 ) {

        /*  Begin struct "org_csapi_pam_TpPAMACEventData"  */
        decode_org_csapi_pam_TpPAMACEventData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMACEventData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMEventInfo == 7 ) {

        /*  Begin struct "org_csapi_pam_TpPAMADEventData"  */
        decode_org_csapi_pam_TpPAMADEventData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMADEventData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMEventInfo == 8 ) {

        /*  Begin struct "org_csapi_pam_TpPAMAAEventData"  */
        decode_org_csapi_pam_TpPAMAAEventData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMAAEventData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMEventInfo == 9 ) {

        /*  Begin struct "org_csapi_pam_TpPAMAUEventData"  */
        decode_org_csapi_pam_TpPAMAUEventData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMAUEventData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMEventInfo == 10 ) {

        /*  Begin struct "org_csapi_pam_TpPAMCCEventData"  */
        decode_org_csapi_pam_TpPAMCCEventData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMCCEventData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMEventInfo == 11 ) {

        /*  Begin struct "org_csapi_pam_TpPAMACPSEventData"  */
        decode_org_csapi_pam_TpPAMACPSEventData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMACPSEventData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMEventInfo == 12 ) {

        /*  Begin struct "org_csapi_pam_TpPAMAPSEventData"  */
        decode_org_csapi_pam_TpPAMAPSEventData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMAPSEventData"  */
        return;     /* End Compare for this discriminant type */
    }

}

/* Union = IDL:org/csapi/pam/TpPAMNotificationInfo:1.0 */
static void
decode_org_csapi_pam_TpPAMNotificationInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpPAMNotificationInfo;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/pam/TpPAMNotificationInfo:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMNotificationInfo_TpPAMNotificationInfo, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/pam/TpPAMEventName:1.0
     */

    disc_s_TpPAMNotificationInfo = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpPAMNotificationInfo == 0 ) {

        /*  Begin struct "org_csapi_pam_TpPAMIPSNotificationData"  */
        decode_org_csapi_pam_TpPAMIPSNotificationData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMIPSNotificationData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMNotificationInfo == 1 ) {

        /*  Begin struct "org_csapi_pam_TpPAMAVCNotificationData"  */
        decode_org_csapi_pam_TpPAMAVCNotificationData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMAVCNotificationData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMNotificationInfo == 2 ) {

        /*  Begin struct "org_csapi_pam_TpPAMWCNotificationData"  */
        decode_org_csapi_pam_TpPAMWCNotificationData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMWCNotificationData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMNotificationInfo == 3 ) {

        /*  Begin struct "org_csapi_pam_TpPAMICNotificationData"  */
        decode_org_csapi_pam_TpPAMICNotificationData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMICNotificationData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMNotificationInfo == 4 ) {

        /*  Begin struct "org_csapi_pam_TpPAMIDNotificationData"  */
        decode_org_csapi_pam_TpPAMIDNotificationData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMIDNotificationData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMNotificationInfo == 5 ) {

        /*  Begin struct "org_csapi_pam_TpPAMGMCNotificationData"  */
        decode_org_csapi_pam_TpPAMGMCNotificationData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMGMCNotificationData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMNotificationInfo == 6 ) {

        /*  Begin struct "org_csapi_pam_TpPAMACNotificationData"  */
        decode_org_csapi_pam_TpPAMACNotificationData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMACNotificationData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMNotificationInfo == 7 ) {

        /*  Begin struct "org_csapi_pam_TpPAMADNotificationData"  */
        decode_org_csapi_pam_TpPAMADNotificationData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMADNotificationData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMNotificationInfo == 8 ) {

        /*  Begin struct "org_csapi_pam_TpPAMAANotificationData"  */
        decode_org_csapi_pam_TpPAMAANotificationData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMAANotificationData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMNotificationInfo == 9 ) {

        /*  Begin struct "org_csapi_pam_TpPAMAUNotificationData"  */
        decode_org_csapi_pam_TpPAMAUNotificationData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMAUNotificationData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMNotificationInfo == 10 ) {

        /*  Begin struct "org_csapi_pam_TpPAMCCNotificationData"  */
        decode_org_csapi_pam_TpPAMCCNotificationData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMCCNotificationData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMNotificationInfo == 11 ) {

        /*  Begin struct "org_csapi_pam_TpPAMACPSNotificationData"  */
        decode_org_csapi_pam_TpPAMACPSNotificationData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMACPSNotificationData"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpPAMNotificationInfo == 12 ) {

        /*  Begin struct "org_csapi_pam_TpPAMAPSNotificationData"  */
        decode_org_csapi_pam_TpPAMAPSNotificationData_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_pam_TpPAMAPSNotificationData"  */
        return;     /* End Compare for this discriminant type */
    }

}

/* Union = IDL:org/csapi/pam/TpPAMPreferenceData:1.0 */
static void
decode_org_csapi_pam_TpPAMPreferenceData_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpPAMPreferenceData;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/pam/TpPAMPreferenceData:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_pam_TpPAMPreferenceData_TpPAMPreferenceData, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/pam/TpPAMPreferenceType:1.0
     */

    disc_s_TpPAMPreferenceData = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpPAMPreferenceData == 1 ) {

        get_CDR_object(tvb, pinfo, tree, offset, stream_is_big_endian, boundary);

        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_pam_TpPAMPreferenceData_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/mmm/TpFolderInfoProperty:1.0 */
static void
decode_org_csapi_mmm_TpFolderInfoProperty_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpFolderInfoProperty;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/mmm/TpFolderInfoProperty:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mmm_TpFolderInfoProperty_TpFolderInfoProperty, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/mmm/TpFolderInfoPropertyName:1.0
     */

    disc_s_TpFolderInfoProperty = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpFolderInfoProperty == 1 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpFolderInfoProperty_FolderDateCreated);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpFolderInfoProperty == 2 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpFolderInfoProperty_FolderDateChanged);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpFolderInfoProperty == 3 ) {

        proto_tree_add_int(tree, hf_org_csapi_mmm_TpFolderInfoProperty_FolderSize, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpFolderInfoProperty == 4 ) {

        proto_tree_add_int(tree, hf_org_csapi_mmm_TpFolderInfoProperty_FolderNumberOfMessages, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_mmm_TpFolderInfoProperty_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/mmm/TpMailboxInfoProperty:1.0 */
static void
decode_org_csapi_mmm_TpMailboxInfoProperty_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpMailboxInfoProperty;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/mmm/TpMailboxInfoProperty:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mmm_TpMailboxInfoProperty_TpMailboxInfoProperty, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/mmm/TpMailboxInfoPropertyName:1.0
     */

    disc_s_TpMailboxInfoProperty = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpMailboxInfoProperty == 1 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMailboxInfoProperty_MailboxOwner);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMailboxInfoProperty == 2 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMailboxInfoProperty_MailboxDateCreated);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMailboxInfoProperty == 3 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMailboxInfoProperty_MailboxDateChanged);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMailboxInfoProperty == 4 ) {

        proto_tree_add_int(tree, hf_org_csapi_mmm_TpMailboxInfoProperty_MailboxSize, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_mmm_TpMailboxInfoProperty_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/mmm/TpMessageInfoProperty:1.0 */
static void
decode_org_csapi_mmm_TpMessageInfoProperty_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpMessageInfoProperty;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/mmm/TpMessageInfoProperty:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mmm_TpMessageInfoProperty_TpMessageInfoProperty, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/mmm/TpMessageInfoPropertyName:1.0
     */

    disc_s_TpMessageInfoProperty = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpMessageInfoProperty == 1 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMessageInfoProperty_MessageDateCreated);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageInfoProperty == 2 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMessageInfoProperty_MessageDateReceived);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageInfoProperty == 3 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMessageInfoProperty_MessageDateChanged);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageInfoProperty == 4 ) {

        proto_tree_add_int(tree, hf_org_csapi_mmm_TpMessageInfoProperty_MessageSize, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageInfoProperty == 5 ) {

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_TpMessageInfoProperty_MessageStatus, tvb, *offset-4, 4, u_octet4);

        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_mmm_TpMessageInfoProperty_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/mmm/TpMessageHeaderField:1.0 */
static void
decode_org_csapi_mmm_TpMessageHeaderField_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpMessageHeaderField;
    guint32   u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_From;
    guint32   i_org_csapi_mmm_TpMessageHeaderField_From;
    guint32   u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_ReplyTo;
    guint32   i_org_csapi_mmm_TpMessageHeaderField_ReplyTo;
    guint32   u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_To;
    guint32   i_org_csapi_mmm_TpMessageHeaderField_To;
    guint32   u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_Cc;
    guint32   i_org_csapi_mmm_TpMessageHeaderField_Cc;
    guint32   u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_Bcc;
    guint32   i_org_csapi_mmm_TpMessageHeaderField_Bcc;
    guint32   u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_InReplyTo;
    guint32   i_org_csapi_mmm_TpMessageHeaderField_InReplyTo;
    guint32   u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_References;
    guint32   i_org_csapi_mmm_TpMessageHeaderField_References;
    guint32   u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_Keywords;
    guint32   i_org_csapi_mmm_TpMessageHeaderField_Keywords;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/mmm/TpMessageHeaderField:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mmm_TpMessageHeaderField_TpMessageHeaderField, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/mmm/TpMessageHeaderFieldType:1.0
     */

    disc_s_TpMessageHeaderField = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpMessageHeaderField == 0 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMessageHeaderField_DateSent);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageHeaderField == 1 ) {

        u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_From = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_TpMessageHeaderField_From_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_From);

        for (i_org_csapi_mmm_TpMessageHeaderField_From=0; i_org_csapi_mmm_TpMessageHeaderField_From < u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_From; i_org_csapi_mmm_TpMessageHeaderField_From++) {

            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageHeaderField == 2 ) {

        /*  Begin struct "org_csapi_TpAddress"  */
        decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_TpAddress"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageHeaderField == 3 ) {

        u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_ReplyTo = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_TpMessageHeaderField_ReplyTo_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_ReplyTo);

        for (i_org_csapi_mmm_TpMessageHeaderField_ReplyTo=0; i_org_csapi_mmm_TpMessageHeaderField_ReplyTo < u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_ReplyTo; i_org_csapi_mmm_TpMessageHeaderField_ReplyTo++) {

            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageHeaderField == 4 ) {

        u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_To = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_TpMessageHeaderField_To_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_To);

        for (i_org_csapi_mmm_TpMessageHeaderField_To=0; i_org_csapi_mmm_TpMessageHeaderField_To < u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_To; i_org_csapi_mmm_TpMessageHeaderField_To++) {

            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageHeaderField == 5 ) {

        u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_Cc = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_TpMessageHeaderField_Cc_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_Cc);

        for (i_org_csapi_mmm_TpMessageHeaderField_Cc=0; i_org_csapi_mmm_TpMessageHeaderField_Cc < u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_Cc; i_org_csapi_mmm_TpMessageHeaderField_Cc++) {

            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageHeaderField == 6 ) {

        u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_Bcc = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_TpMessageHeaderField_Bcc_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_Bcc);

        for (i_org_csapi_mmm_TpMessageHeaderField_Bcc=0; i_org_csapi_mmm_TpMessageHeaderField_Bcc < u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_Bcc; i_org_csapi_mmm_TpMessageHeaderField_Bcc++) {

            /*  Begin struct "org_csapi_TpAddress"  */
            decode_org_csapi_TpAddress_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            /*  End struct "org_csapi_TpAddress"  */
        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageHeaderField == 7 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMessageHeaderField_RFC822MessageID);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageHeaderField == 8 ) {

        u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_InReplyTo = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_TpMessageHeaderField_InReplyTo_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_InReplyTo);

        for (i_org_csapi_mmm_TpMessageHeaderField_InReplyTo=0; i_org_csapi_mmm_TpMessageHeaderField_InReplyTo < u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_InReplyTo; i_org_csapi_mmm_TpMessageHeaderField_InReplyTo++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMessageHeaderField_InReplyTo);

        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageHeaderField == 9 ) {

        u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_References = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_TpMessageHeaderField_References_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_References);

        for (i_org_csapi_mmm_TpMessageHeaderField_References=0; i_org_csapi_mmm_TpMessageHeaderField_References < u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_References; i_org_csapi_mmm_TpMessageHeaderField_References++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMessageHeaderField_References);

        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageHeaderField == 10 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMessageHeaderField_Subject);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageHeaderField == 11 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMessageHeaderField_Comments);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageHeaderField == 12 ) {

        u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_Keywords = get_CDR_ulong(tvb, offset, stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_TpMessageHeaderField_Keywords_loop, tvb,*offset-4, 4, u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_Keywords);

        for (i_org_csapi_mmm_TpMessageHeaderField_Keywords=0; i_org_csapi_mmm_TpMessageHeaderField_Keywords < u_octet4_loop_org_csapi_mmm_TpMessageHeaderField_Keywords; i_org_csapi_mmm_TpMessageHeaderField_Keywords++) {

            giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMessageHeaderField_Keywords);

        }

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageHeaderField == 13 ) {

        /*  Begin struct "org_csapi_mmm_TpGenericHeaderField"  */
        decode_org_csapi_mmm_TpGenericHeaderField_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mmm_TpGenericHeaderField"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageHeaderField == 14 ) {

        /*  Begin struct "org_csapi_mmm_TpGenericHeaderField"  */
        decode_org_csapi_mmm_TpGenericHeaderField_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mmm_TpGenericHeaderField"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageHeaderField == 15 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMessageHeaderField_MimeVersion);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageHeaderField == 16 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMessageHeaderField_MimeContent);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageHeaderField == 17 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMessageHeaderField_MimeEncoding);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageHeaderField == 18 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMessageHeaderField_MimeID);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageHeaderField == 19 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMessageHeaderField_MimeDescription);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageHeaderField == 20 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMessageHeaderField_MimeDisposition);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageHeaderField == 21 ) {

        /*  Begin struct "org_csapi_mmm_TpGenericHeaderField"  */
        decode_org_csapi_mmm_TpGenericHeaderField_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mmm_TpGenericHeaderField"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageHeaderField == 22 ) {

        /*  Begin struct "org_csapi_mmm_TpGenericHeaderField"  */
        decode_org_csapi_mmm_TpGenericHeaderField_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mmm_TpGenericHeaderField"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageHeaderField == 23 ) {

        u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
        proto_tree_add_uint(tree, hf_org_csapi_mmm_TpMessageHeaderField_Priority, tvb, *offset-4, 4, u_octet4);

        return;     /* End Compare for this discriminant type */
    }

}

/* Union = IDL:org/csapi/mmm/TpDeliveryTime:1.0 */
static void
decode_org_csapi_mmm_TpDeliveryTime_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpDeliveryTime;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/mmm/TpDeliveryTime:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mmm_TpDeliveryTime_TpDeliveryTime, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/mmm/TpDeliveryTimeType:1.0
     */

    disc_s_TpDeliveryTime = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpDeliveryTime == 1 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpDeliveryTime_DeliveryTime);

        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_mmm_TpDeliveryTime_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/mmm/TpMessageTreatment:1.0 */
static void
decode_org_csapi_mmm_TpMessageTreatment_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpMessageTreatment;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/mmm/TpMessageTreatment:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mmm_TpMessageTreatment_TpMessageTreatment, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/mmm/TpMessageTreatmentType:1.0
     */

    disc_s_TpMessageTreatment = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpMessageTreatment == 1 ) {

        proto_tree_add_int(tree, hf_org_csapi_mmm_TpMessageTreatment_DeliveryReport, tvb, *offset-4, 4, get_CDR_long(tvb,offset,stream_is_big_endian, boundary));

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageTreatment == 2 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMessageTreatment_BillingID);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageTreatment == 3 ) {

        /*  Begin union "org_csapi_mmm_TpDeliveryTime"  */
        decode_org_csapi_mmm_TpDeliveryTime_un(tvb, pinfo, tree, offset, header, operation, stream_is_big_endian);

        /*  End union "org_csapi_mmm_TpDeliveryTime"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessageTreatment == 4 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMessageTreatment_ValidityTime);

        return;     /* End Compare for this discriminant type */
    }

}

/* Union = IDL:org/csapi/mmm/TpMessagingEventCriteria:1.0 */
static void
decode_org_csapi_mmm_TpMessagingEventCriteria_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpMessagingEventCriteria;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/mmm/TpMessagingEventCriteria:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mmm_TpMessagingEventCriteria_TpMessagingEventCriteria, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/mmm/TpMessagingEventName:1.0
     */

    disc_s_TpMessagingEventCriteria = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpMessagingEventCriteria == 1 ) {

        /*  Begin struct "org_csapi_mmm_TpNewMailboxMessageArrivedCriteria"  */
        decode_org_csapi_mmm_TpNewMailboxMessageArrivedCriteria_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mmm_TpNewMailboxMessageArrivedCriteria"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessagingEventCriteria == 2 ) {

        /*  Begin struct "org_csapi_mmm_TpNewMessageArrivedCriteria"  */
        decode_org_csapi_mmm_TpNewMessageArrivedCriteria_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mmm_TpNewMessageArrivedCriteria"  */
        return;     /* End Compare for this discriminant type */
    }

    
    /* Default Union Case Start */

    proto_tree_add_int(tree, hf_org_csapi_mmm_TpMessagingEventCriteria_Dummy, tvb, *offset-2, 2, get_CDR_short(tvb,offset,stream_is_big_endian, boundary));

    /* Default Union Case End */

}

/* Union = IDL:org/csapi/mmm/TpMessagingEventInfo:1.0 */
static void
decode_org_csapi_mmm_TpMessagingEventInfo_un(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, int *offset _U_, MessageHeader *header _U_, const gchar *operation _U_, gboolean stream_is_big_endian _U_)
{
    proto_item* item = NULL;

    /* Operation specific Variable declarations Begin */
    guint32   u_octet4;
    gint32    disc_s_TpMessagingEventInfo;
    /* Operation specific Variable declarations End */

    /*
     * IDL Union Start - IDL:org/csapi/mmm/TpMessagingEventInfo:1.0
     */

    u_octet4 = get_CDR_enum(tvb,offset,stream_is_big_endian, boundary);
    proto_tree_add_uint(tree, hf_org_csapi_mmm_TpMessagingEventInfo_TpMessagingEventInfo, tvb, *offset-4, 4, u_octet4);

    /*
     * IDL Union - Discriminant - IDL:org/csapi/mmm/TpMessagingEventName:1.0
     */

    disc_s_TpMessagingEventInfo = (gint32) u_octet4;     /* save Enum Value  discriminant and cast to gint32 */

    if (disc_s_TpMessagingEventInfo == 0 ) {

        giop_add_CDR_string(tree, tvb, offset, stream_is_big_endian, boundary, hf_org_csapi_mmm_TpMessagingEventInfo_EventNameUndefined);

        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessagingEventInfo == 1 ) {

        /*  Begin struct "org_csapi_mmm_TpNewMailboxMessageArrivedInfo"  */
        decode_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mmm_TpNewMailboxMessageArrivedInfo"  */
        return;     /* End Compare for this discriminant type */
    }

    if (disc_s_TpMessagingEventInfo == 2 ) {

        /*  Begin struct "org_csapi_mmm_TpNewMessageArrivedInfo"  */
        decode_org_csapi_mmm_TpNewMessageArrivedInfo_st(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
        /*  End struct "org_csapi_mmm_TpNewMessageArrivedInfo"  */
        return;     /* End Compare for this discriminant type */
    }

}

/*
 * Called once we accept the packet as being for us; it sets the
 * Protocol and Info columns and creates the top-level protocol
 * tree item.
 */
static proto_tree *
start_dissecting(tvbuff_t *tvb, packet_info *pinfo, proto_tree *ptree, int *offset)
{

    proto_item *ti = NULL;
    proto_tree *tree = NULL;            /* init later, inside if(tree) */

    col_set_str(pinfo->cinfo, COL_PROTOCOL, "PARLAY");

    /*
     * Do not clear COL_INFO, as nothing is being written there by
     * this dissector yet. So leave it as is from the GIOP dissector.
     * TODO: add something useful to COL_INFO
     *     col_clear(pinfo->cinfo, COL_INFO);
     */

    if (ptree) {
        ti = proto_tree_add_item(ptree, proto_parlay, tvb, *offset, tvb_reported_length_remaining(tvb, *offset), ENC_NA);
        tree = proto_item_add_subtree(ti, ett_parlay);
    }
    return tree;
}

static proto_item*
process_RequestOperation(tvbuff_t *tvb, packet_info *pinfo, proto_tree *ptree, MessageHeader *header, const gchar *operation)
{
    proto_item *pi;
    if(header->message_type == Reply) {
        /* fill-up info column */
        col_append_fstr(pinfo->cinfo, COL_INFO, " op = %s",operation);
    }
    /* fill-up the field */
    pi=proto_tree_add_string(ptree, hf_operationrequest, tvb, 0, 0, operation);
    proto_item_set_generated(pi);
    return pi;
}

static gboolean
dissect_parlay(tvbuff_t *tvb, packet_info *pinfo, proto_tree *ptree, int *offset, MessageHeader *header, const gchar *operation, gchar *idlname)
{
    proto_item *item _U_;
    proto_tree *tree _U_;
    gboolean stream_is_big_endian = is_big_endian(header); /* get endianess */

    /* If we have a USER Exception, then decode it and return */
    if ((header->message_type == Reply) && (header->rep_status == USER_EXCEPTION)) {
       return decode_user_exception(tvb, pinfo, ptree, offset, header, operation, stream_is_big_endian);
    }

    switch(header->message_type) {

    case Request:
    case Reply:

        if (strcmp(operation, "setCallback") == 0
            && (!idlname || strcmp(idlname, "org/csapi/IpService") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_IpService_setCallback(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setCallbackWithSessionID") == 0
            && (!idlname || strcmp(idlname, "org/csapi/IpService") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_IpService_setCallbackWithSessionID(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "initiateAuthentication") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_access/trust_and_security/IpInitial") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_access_trust_and_security_IpInitial_initiateAuthentication(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "initiateAuthenticationWithVersion") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_access/trust_and_security/IpInitial") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_access_trust_and_security_IpInitial_initiateAuthenticationWithVersion(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "requestAccess") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_access/trust_and_security/IpAuthentication") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_access_trust_and_security_IpAuthentication_requestAccess(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "terminateAccess") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_access/trust_and_security/IpClientAccess") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_access_trust_and_security_IpClientAccess_terminateAccess(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "obtainInterface") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_access/trust_and_security/IpAccess") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_access_trust_and_security_IpAccess_obtainInterface(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "obtainInterfaceWithCallback") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_access/trust_and_security/IpAccess") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_access_trust_and_security_IpAccess_obtainInterfaceWithCallback(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "endAccess") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_access/trust_and_security/IpAccess") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_access_trust_and_security_IpAccess_endAccess(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listInterfaces") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_access/trust_and_security/IpAccess") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_access_trust_and_security_IpAccess_listInterfaces(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "releaseInterface") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_access/trust_and_security/IpAccess") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_access_trust_and_security_IpAccess_releaseInterface(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "selectSigningAlgorithm") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_access/trust_and_security/IpAccess") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_access_trust_and_security_IpAccess_selectSigningAlgorithm(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "terminateAccess") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_access/trust_and_security/IpAccess") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_access_trust_and_security_IpAccess_terminateAccess(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "relinquishInterface") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_access/trust_and_security/IpAccess") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_access_trust_and_security_IpAccess_relinquishInterface(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "authenticate") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_access/trust_and_security/IpClientAPILevelAuthentication") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "abortAuthentication") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_access/trust_and_security/IpClientAPILevelAuthentication") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_abortAuthentication(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "authenticationSucceeded") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_access/trust_and_security/IpClientAPILevelAuthentication") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticationSucceeded(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "challenge") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_access/trust_and_security/IpClientAPILevelAuthentication") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "selectEncryptionMethod") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_access/trust_and_security/IpAPILevelAuthentication") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_selectEncryptionMethod(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "authenticate") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_access/trust_and_security/IpAPILevelAuthentication") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "abortAuthentication") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_access/trust_and_security/IpAPILevelAuthentication") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_abortAuthentication(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "authenticationSucceeded") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_access/trust_and_security/IpAPILevelAuthentication") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticationSucceeded(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "selectAuthenticationMechanism") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_access/trust_and_security/IpAPILevelAuthentication") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_selectAuthenticationMechanism(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "challenge") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_access/trust_and_security/IpAPILevelAuthentication") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "reportNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/notification/IpAppEventNotification") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_notification_IpAppEventNotification_reportNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "notificationTerminated") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/notification/IpAppEventNotification") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_notification_IpAppEventNotification_notificationTerminated(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/notification/IpEventNotification") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_notification_IpEventNotification_createNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "destroyNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/notification/IpEventNotification") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_notification_IpEventNotification_destroyNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "activityTestRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_activityTestRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "appActivityTestReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_appActivityTestReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "fwFaultReportInd") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_fwFaultReportInd(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "fwFaultRecoveryInd") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_fwFaultRecoveryInd(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "svcUnavailableInd") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_svcUnavailableInd(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "genFaultStatsRecordRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "fwUnavailableInd") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_fwUnavailableInd(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "activityTestErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_activityTestErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "genFaultStatsRecordErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "appUnavailableInd") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_appUnavailableInd(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "genFaultStatsRecordReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "svcAvailStatusInd") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_svcAvailStatusInd(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "generateFaultStatisticsRecordRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "generateFaultStatisticsRecordErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "generateFaultStatisticsRecordReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "fwAvailStatusInd") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppFaultManager_fwAvailStatusInd(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryAppLoadReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryAppLoadReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryLoadRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryLoadErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "loadLevelNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppLoadManager_loadLevelNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "resumeNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppLoadManager_resumeNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "suspendNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppLoadManager_suspendNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createLoadLevelNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppLoadManager_createLoadLevelNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "destroyLoadLevelNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppLoadManager_destroyLoadLevelNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryAppLoadStatsReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryAppLoadStatsReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryLoadStatsRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryLoadStatsErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "reportLoad") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpLoadManager_reportLoad(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryLoadReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryAppLoadRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryAppLoadErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createLoadLevelNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpLoadManager_createLoadLevelNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "destroyLoadLevelNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpLoadManager_destroyLoadLevelNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "resumeNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpLoadManager_resumeNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "suspendNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpLoadManager_suspendNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryLoadStatsReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadStatsReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryAppLoadStatsRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryAppLoadStatsErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "systemDateTimeQuery") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppOAM") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppOAM_systemDateTimeQuery(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "systemDateTimeQuery") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpOAM") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpOAM_systemDateTimeQuery(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "activityTestReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpFaultManager_activityTestReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "appActivityTestRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpFaultManager_appActivityTestRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "svcUnavailableInd") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpFaultManager_svcUnavailableInd(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "genFaultStatsRecordReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpFaultManager_genFaultStatsRecordReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "appActivityTestErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpFaultManager_appActivityTestErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "appUnavailableInd") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpFaultManager_appUnavailableInd(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "genFaultStatsRecordRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpFaultManager_genFaultStatsRecordRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "genFaultStatsRecordErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpFaultManager_genFaultStatsRecordErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "appAvailStatusInd") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpFaultManager_appAvailStatusInd(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "generateFaultStatisticsRecordReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "generateFaultStatisticsRecordRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "generateFaultStatisticsRecordErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "enableHeartBeat") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpHeartBeatMgmt") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpHeartBeatMgmt_enableHeartBeat(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "disableHeartBeat") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpHeartBeatMgmt") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpHeartBeatMgmt_disableHeartBeat(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "changeInterval") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpHeartBeatMgmt") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpHeartBeatMgmt_changeInterval(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "pulse") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppHeartBeat") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppHeartBeat_pulse(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "pulse") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpHeartBeat") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpHeartBeat_pulse(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "enableAppHeartBeat") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppHeartBeatMgmt") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppHeartBeatMgmt_enableAppHeartBeat(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "disableAppHeartBeat") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppHeartBeatMgmt") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppHeartBeatMgmt_disableAppHeartBeat(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "changeInterval") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/integrity/IpAppHeartBeatMgmt") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_integrity_IpAppHeartBeatMgmt_changeInterval(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listServiceTypes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/discovery/IpServiceDiscovery") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listServiceTypes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "describeServiceType") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/discovery/IpServiceDiscovery") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_describeServiceType(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "discoverService") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/discovery/IpServiceDiscovery") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listSubscribedServices") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/discovery/IpServiceDiscovery") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listSubscribedServices(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "signServiceAgreement") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/service_agreement/IpAppServiceAgreementManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "terminateServiceAgreement") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/service_agreement/IpAppServiceAgreementManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_terminateServiceAgreement(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "signServiceAgreement") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/service_agreement/IpServiceAgreementManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_signServiceAgreement(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "terminateServiceAgreement") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/service_agreement/IpServiceAgreementManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_terminateServiceAgreement(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "selectService") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/service_agreement/IpServiceAgreementManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_selectService(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "initiateSignServiceAgreement") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_application/service_agreement/IpServiceAgreementManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_initiateSignServiceAgreement(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createServiceProfile") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceProfileManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_createServiceProfile(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "modifyServiceProfile") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceProfileManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_modifyServiceProfile(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteServiceProfile") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceProfileManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_deleteServiceProfile(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "assign") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceProfileManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_assign(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deassign") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceProfileManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_deassign(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "requestConflictInfo") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceProfileManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_requestConflictInfo(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listServiceProfiles") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceProfileInfoQuery") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listServiceProfiles(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "describeServiceProfile") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceProfileInfoQuery") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_describeServiceProfile(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listAssignedMembers") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceProfileInfoQuery") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listAssignedMembers(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createServiceContract") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceContractManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractManagement_createServiceContract(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "modifyServiceContract") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceContractManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractManagement_modifyServiceContract(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteServiceContract") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceContractManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractManagement_deleteServiceContract(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "describeServiceContract") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceContractInfoQuery") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_describeServiceContract(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listServiceContracts") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceContractInfoQuery") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceContracts(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listServiceProfiles") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceContractInfoQuery") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceProfiles(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "modifyEntOpAccount") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpEntOpAccountManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpEntOpAccountManagement_modifyEntOpAccount(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteEntOpAccount") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpEntOpAccountManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpEntOpAccountManagement_deleteEntOpAccount(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "describeEntOpAccount") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpEntOpAccountInfoQuery") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpEntOpAccountInfoQuery_describeEntOpAccount(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createClientApp") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_createClientApp(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "modifyClientApp") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_modifyClientApp(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteClientApp") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_deleteClientApp(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createSAG") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_createSAG(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "modifySAG") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_modifySAG(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteSAG") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_deleteSAG(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "addSAGMembers") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_addSAGMembers(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "removeSAGMembers") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_removeSAGMembers(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "requestConflictInfo") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_requestConflictInfo(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "describeClientApp") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppInfoQuery") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_describeClientApp(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listClientApps") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppInfoQuery") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientApps(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "describeSAG") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppInfoQuery") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_describeSAG(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listSAGs") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppInfoQuery") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGs(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listSAGMembers") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppInfoQuery") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGMembers(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listClientAppMembership") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppInfoQuery") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientAppMembership(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "reportNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/notification/IpClientEventNotification") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_notification_IpClientEventNotification_reportNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "notificationTerminated") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/notification/IpClientEventNotification") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_notification_IpClientEventNotification_notificationTerminated(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/notification/IpEventNotification") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_notification_IpEventNotification_createNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "destroyNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_enterprise_operator/notification/IpEventNotification") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_enterprise_operator_notification_IpEventNotification_destroyNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listServiceTypes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/discovery/IpFwServiceDiscovery") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listServiceTypes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "describeServiceType") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/discovery/IpFwServiceDiscovery") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_describeServiceType(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "discoverService") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/discovery/IpFwServiceDiscovery") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listRegisteredServices") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/discovery/IpFwServiceDiscovery") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listRegisteredServices(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createServiceManager") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/service_lifecycle/IpServiceInstanceLifecycleManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_createServiceManager(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "destroyServiceManager") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/service_lifecycle/IpServiceInstanceLifecycleManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_destroyServiceManager(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "registerService") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/service_registration/IpFwServiceRegistration") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerService(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "announceServiceAvailability") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/service_registration/IpFwServiceRegistration") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_announceServiceAvailability(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "unregisterService") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/service_registration/IpFwServiceRegistration") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_unregisterService(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "describeService") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/service_registration/IpFwServiceRegistration") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_describeService(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "unannounceService") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/service_registration/IpFwServiceRegistration") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_unannounceService(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "registerServiceSubType") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/service_registration/IpFwServiceRegistration") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "querySvcLoadReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_querySvcLoadReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryLoadRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryLoadErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "loadLevelNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_loadLevelNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "suspendNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_suspendNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "resumeNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_resumeNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createLoadLevelNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_createLoadLevelNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "destroyLoadLevelNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_destroyLoadLevelNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "querySvcLoadStatsReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_querySvcLoadStatsReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryLoadStatsRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryLoadStatsErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "reportLoad") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwLoadManager_reportLoad(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryLoadReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwLoadManager_queryLoadReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "querySvcLoadRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "querySvcLoadErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createLoadLevelNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwLoadManager_createLoadLevelNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "destroyLoadLevelNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwLoadManager_destroyLoadLevelNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "suspendNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwLoadManager_suspendNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "resumeNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwLoadManager_resumeNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryLoadStatsReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwLoadManager_queryLoadStatsReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "querySvcLoadStatsRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "querySvcLoadStatsErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwLoadManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "activityTestRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_activityTestRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "svcActivityTestReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_svcActivityTestReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "fwFaultReportInd") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_fwFaultReportInd(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "fwFaultRecoveryInd") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_fwFaultRecoveryInd(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "fwUnavailableInd") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_fwUnavailableInd(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "svcUnavailableInd") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_svcUnavailableInd(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "appUnavailableInd") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_appUnavailableInd(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "genFaultStatsRecordRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "activityTestErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_activityTestErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "genFaultStatsRecordErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "genFaultStatsRecordReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "generateFaultStatsRecordReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatsRecordReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "appAvailStatusInd") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_appAvailStatusInd(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "generateFaultStatisticsRecordRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "generateFaultStatisticsRecordErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "generateFaultStatisticsRecordReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "fwAvailStatusInd") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_fwAvailStatusInd(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "activityTestReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_activityTestReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "svcActivityTestRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_svcActivityTestRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "appUnavailableInd") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_appUnavailableInd(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "genFaultStatsRecordReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "svcUnavailableInd") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_svcUnavailableInd(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "svcActivityTestErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_svcActivityTestErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "genFaultStatsRecordRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "genFaultStatsRecordErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "generateFaultStatsRecordRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatsRecordRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "generateFaultStatsRecordErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatsRecordErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "svcAvailStatusInd") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_svcAvailStatusInd(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "generateFaultStatisticsRecordReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatisticsRecordReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "generateFaultStatisticsRecordRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatisticsRecordRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "generateFaultStatisticsRecordErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwFaultManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatisticsRecordErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "systemDateTimeQuery") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcOAM") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcOAM_systemDateTimeQuery(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "systemDateTimeQuery") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwOAM") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwOAM_systemDateTimeQuery(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "enableHeartBeat") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwHeartBeatMgmt") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwHeartBeatMgmt_enableHeartBeat(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "disableHeartBeat") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwHeartBeatMgmt") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwHeartBeatMgmt_disableHeartBeat(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "changeInterval") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwHeartBeatMgmt") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwHeartBeatMgmt_changeInterval(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "pulse") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcHeartBeat") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcHeartBeat_pulse(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "pulse") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpFwHeartBeat") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpFwHeartBeat_pulse(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "enableSvcHeartBeat") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcHeartBeatMgmt") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcHeartBeatMgmt_enableSvcHeartBeat(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "disableSvcHeartBeat") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcHeartBeatMgmt") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcHeartBeatMgmt_disableSvcHeartBeat(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "changeInterval") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/integrity/IpSvcHeartBeatMgmt") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_integrity_IpSvcHeartBeatMgmt_changeInterval(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/notification/IpFwEventNotification") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_notification_IpFwEventNotification_createNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "destroyNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/notification/IpFwEventNotification") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_notification_IpFwEventNotification_destroyNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "reportNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/notification/IpSvcEventNotification") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_notification_IpSvcEventNotification_reportNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "notificationTerminated") == 0
            && (!idlname || strcmp(idlname, "org/csapi/fw/fw_service/notification/IpSvcEventNotification") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_fw_fw_service_notification_IpSvcEventNotification_notificationTerminated(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "routeRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpAppCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpAppCall_routeRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "routeErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpAppCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpAppCall_routeErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getCallInfoRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpAppCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpAppCall_getCallInfoRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getCallInfoErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpAppCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpAppCall_getCallInfoErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "superviseCallRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpAppCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpAppCall_superviseCallRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "superviseCallErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpAppCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpAppCall_superviseCallErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "callFaultDetected") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpAppCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpAppCall_callFaultDetected(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMoreDialledDigitsRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpAppCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpAppCall_getMoreDialledDigitsRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMoreDialledDigitsErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpAppCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpAppCall_getMoreDialledDigitsErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "callEnded") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpAppCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpAppCall_callEnded(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "routeReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpCall_routeReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "release") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpCall_release(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deassignCall") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpCall_deassignCall(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getCallInfoReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpCall_getCallInfoReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setCallChargePlan") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpCall_setCallChargePlan(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setAdviceOfCharge") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpCall_setAdviceOfCharge(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMoreDialledDigitsReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpCall_getMoreDialledDigitsReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "superviseCallReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpCall_superviseCallReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "continueProcessing") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpCall_continueProcessing(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "callAborted") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpAppCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpAppCallControlManager_callAborted(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "callEventNotify") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpAppCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpAppCallControlManager_callEventNotify(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "callNotificationInterrupted") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpAppCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpAppCallControlManager_callNotificationInterrupted(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "callNotificationContinued") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpAppCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpAppCallControlManager_callNotificationContinued(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "callOverloadEncountered") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpAppCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpAppCallControlManager_callOverloadEncountered(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "callOverloadCeased") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpAppCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpAppCallControlManager_callOverloadCeased(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "abortMultipleCalls") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpAppCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpAppCallControlManager_abortMultipleCalls(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createCall") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpCallControlManager_createCall(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "enableCallNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpCallControlManager_enableCallNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "disableCallNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpCallControlManager_disableCallNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setCallLoadControl") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpCallControlManager_setCallLoadControl(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "changeCallNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpCallControlManager_changeCallNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getCriteria") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/gccs/IpCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_gccs_IpCallControlManager_getCriteria(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "eventReportRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppCallLeg_eventReportRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "eventReportErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppCallLeg_eventReportErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "attachMediaRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppCallLeg_attachMediaRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "attachMediaErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppCallLeg_attachMediaErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "detachMediaRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppCallLeg_detachMediaRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "detachMediaErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppCallLeg_detachMediaErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getInfoRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppCallLeg_getInfoRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getInfoErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppCallLeg_getInfoErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "routeErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppCallLeg_routeErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "superviseRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppCallLeg_superviseRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "superviseErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppCallLeg_superviseErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "callLegEnded") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppCallLeg_callLegEnded(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createCall") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpMultiPartyCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_createCall(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpMultiPartyCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_createNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "destroyNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpMultiPartyCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_destroyNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "changeNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpMultiPartyCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_changeNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpMultiPartyCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_getNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setCallLoadControl") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpMultiPartyCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_setCallLoadControl(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "enableNotifications") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpMultiPartyCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_enableNotifications(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "disableNotifications") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpMultiPartyCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_disableNotifications(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getNextNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpMultiPartyCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_getNextNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "routeReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpCallLeg_routeReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "eventReportReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpCallLeg_eventReportReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "release") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpCallLeg_release(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getInfoReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpCallLeg_getInfoReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getCall") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpCallLeg_getCall(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "attachMediaReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpCallLeg_attachMediaReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "detachMediaReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpCallLeg_detachMediaReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getCurrentDestinationAddress") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpCallLeg_getCurrentDestinationAddress(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "continueProcessing") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpCallLeg_continueProcessing(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setChargePlan") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpCallLeg_setChargePlan(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setAdviceOfCharge") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpCallLeg_setAdviceOfCharge(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "superviseReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpCallLeg_superviseReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deassign") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpCallLeg_deassign(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getProperties") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpCallLeg_getProperties(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setProperties") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpCallLeg_setProperties(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getInfoRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppMultiPartyCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppMultiPartyCall_getInfoRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getInfoErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppMultiPartyCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppMultiPartyCall_getInfoErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "superviseRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppMultiPartyCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppMultiPartyCall_superviseRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "superviseErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppMultiPartyCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppMultiPartyCall_superviseErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "callEnded") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppMultiPartyCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppMultiPartyCall_callEnded(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createAndRouteCallLegErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppMultiPartyCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppMultiPartyCall_createAndRouteCallLegErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getCallLegs") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpMultiPartyCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpMultiPartyCall_getCallLegs(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createCallLeg") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpMultiPartyCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpMultiPartyCall_createCallLeg(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createAndRouteCallLegReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpMultiPartyCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "release") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpMultiPartyCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpMultiPartyCall_release(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deassignCall") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpMultiPartyCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpMultiPartyCall_deassignCall(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getInfoReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpMultiPartyCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpMultiPartyCall_getInfoReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setChargePlan") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpMultiPartyCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpMultiPartyCall_setChargePlan(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setAdviceOfCharge") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpMultiPartyCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpMultiPartyCall_setAdviceOfCharge(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "superviseReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpMultiPartyCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpMultiPartyCall_superviseReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "reportNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppMultiPartyCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_reportNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "callAborted") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppMultiPartyCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_callAborted(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "managerInterrupted") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppMultiPartyCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_managerInterrupted(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "managerResumed") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppMultiPartyCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_managerResumed(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "callOverloadEncountered") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppMultiPartyCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_callOverloadEncountered(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "callOverloadCeased") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppMultiPartyCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_callOverloadCeased(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "abortMultipleCalls") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mpccs/IpAppMultiPartyCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_abortMultipleCalls(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "superviseVolumeRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mmccs/IpAppMultiMediaCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mmccs_IpAppMultiMediaCall_superviseVolumeRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "superviseVolumeErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mmccs/IpAppMultiMediaCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mmccs_IpAppMultiMediaCall_superviseVolumeErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "mediaStreamMonitorRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mmccs/IpAppMultiMediaCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mmccs_IpAppMultiMediaCallLeg_mediaStreamMonitorRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "subtract") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mmccs/IpMultiMediaStream") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mmccs_IpMultiMediaStream_subtract(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "mediaStreamAllow") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mmccs/IpMultiMediaCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamAllow(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "mediaStreamMonitorReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mmccs/IpMultiMediaCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamMonitorReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMediaStreams") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mmccs/IpMultiMediaCallLeg") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mmccs_IpMultiMediaCallLeg_getMediaStreams(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "superviseVolumeReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mmccs/IpMultiMediaCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mmccs_IpMultiMediaCall_superviseVolumeReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "reportMediaNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mmccs/IpAppMultiMediaCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createMediaNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mmccs/IpMultiMediaCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_createMediaNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "destroyMediaNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mmccs/IpMultiMediaCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_destroyMediaNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "changeMediaNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mmccs/IpMultiMediaCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_changeMediaNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMediaNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/mmccs/IpMultiMediaCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_getMediaNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "chairSelection") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/cccs/IpAppSubConfCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_cccs_IpAppSubConfCall_chairSelection(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "floorRequest") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/cccs/IpAppSubConfCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_cccs_IpAppSubConfCall_floorRequest(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "partyJoined") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/cccs/IpAppConfCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_cccs_IpAppConfCall_partyJoined(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "leaveMonitorRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/cccs/IpAppConfCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_cccs_IpAppConfCall_leaveMonitorRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createConference") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/cccs/IpConfCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_cccs_IpConfCallControlManager_createConference(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "checkResources") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/cccs/IpConfCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_cccs_IpConfCallControlManager_checkResources(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "reserveResources") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/cccs/IpConfCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_cccs_IpConfCallControlManager_reserveResources(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "freeResources") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/cccs/IpConfCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_cccs_IpConfCallControlManager_freeResources(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "conferenceCreated") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/cccs/IpAppConfCallControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_cccs_IpAppConfCallControlManager_conferenceCreated(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "splitSubConference") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/cccs/IpSubConfCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_cccs_IpSubConfCall_splitSubConference(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "mergeSubConference") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/cccs/IpSubConfCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_cccs_IpSubConfCall_mergeSubConference(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "moveCallLeg") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/cccs/IpSubConfCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_cccs_IpSubConfCall_moveCallLeg(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "inspectVideo") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/cccs/IpSubConfCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_cccs_IpSubConfCall_inspectVideo(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "inspectVideoCancel") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/cccs/IpSubConfCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_cccs_IpSubConfCall_inspectVideoCancel(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "appointSpeaker") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/cccs/IpSubConfCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_cccs_IpSubConfCall_appointSpeaker(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "chairSelection") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/cccs/IpSubConfCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_cccs_IpSubConfCall_chairSelection(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "changeConferencePolicy") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/cccs/IpSubConfCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_cccs_IpSubConfCall_changeConferencePolicy(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getSubConferences") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/cccs/IpConfCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_cccs_IpConfCall_getSubConferences(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createSubConference") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/cccs/IpConfCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_cccs_IpConfCall_createSubConference(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "leaveMonitorReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/cccs/IpConfCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_cccs_IpConfCall_leaveMonitorReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getConferenceAddress") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cc/cccs/IpConfCall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cc_cccs_IpConfCall_getConferenceAddress(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "sendInfoRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUI") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUI_sendInfoRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "sendInfoErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUI") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUI_sendInfoErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "sendInfoAndCollectRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUI") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUI_sendInfoAndCollectRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "sendInfoAndCollectErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUI") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUI_sendInfoAndCollectErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "userInteractionFaultDetected") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUI") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUI_userInteractionFaultDetected(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "recordMessageRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUICall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUICall_recordMessageRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "recordMessageErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUICall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUICall_recordMessageErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteMessageRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUICall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUICall_deleteMessageRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteMessageErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUICall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUICall_deleteMessageErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "abortActionRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUICall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUICall_abortActionRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "abortActionErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUICall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUICall_abortActionErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMessageRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUICall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUICall_getMessageRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMessageErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUICall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUICall_getMessageErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "userInteractionAborted") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUIManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUIManager_userInteractionAborted(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "userInteractionNotificationInterrupted") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUIManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUIManager_userInteractionNotificationInterrupted(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "userInteractionNotificationContinued") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUIManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUIManager_userInteractionNotificationContinued(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "reportEventNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUIManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUIManager_reportEventNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "abortMultipleUserInteractions") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUIManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUIManager_abortMultipleUserInteractions(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "sendInfoReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpUI") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpUI_sendInfoReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "sendInfoAndCollectReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpUI") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpUI_sendInfoAndCollectReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "release") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpUI") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpUI_release(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setOriginatingAddress") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpUI") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpUI_setOriginatingAddress(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getOriginatingAddress") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpUI") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpUI_getOriginatingAddress(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "recordMessageReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpUICall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpUICall_recordMessageReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteMessageReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpUICall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpUICall_deleteMessageReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "abortActionReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpUICall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpUICall_abortActionReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMessageReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpUICall") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpUICall_getMessageReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createUI") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpUIManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpUIManager_createUI(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createUICall") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpUIManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpUIManager_createUICall(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpUIManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpUIManager_createNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "destroyNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpUIManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpUIManager_destroyNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "changeNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpUIManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpUIManager_changeNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpUIManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpUIManager_getNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "enableNotifications") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpUIManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpUIManager_enableNotifications(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "disableNotifications") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpUIManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpUIManager_disableNotifications(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMessageRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUIAdminManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUIAdminManager_getMessageRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMessageErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUIAdminManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUIAdminManager_getMessageErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteMessageRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUIAdminManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUIAdminManager_deleteMessageRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteMessageErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUIAdminManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUIAdminManager_deleteMessageErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "putMessageRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUIAdminManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUIAdminManager_putMessageRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "putMessageErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUIAdminManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUIAdminManager_putMessageErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMessageListRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUIAdminManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUIAdminManager_getMessageListRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMessageListErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpAppUIAdminManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpAppUIAdminManager_getMessageListErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMessageReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpUIAdminManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpUIAdminManager_getMessageReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "putMessageReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpUIAdminManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpUIAdminManager_putMessageReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteMessageReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpUIAdminManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpUIAdminManager_deleteMessageReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMessageListReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/ui/IpUIAdminManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_ui_IpUIAdminManager_getMessageListReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "locationReportRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ul/IpAppUserLocation") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ul_IpAppUserLocation_locationReportRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "locationReportErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ul/IpAppUserLocation") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ul_IpAppUserLocation_locationReportErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "extendedLocationReportRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ul/IpAppUserLocation") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "extendedLocationReportErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ul/IpAppUserLocation") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "periodicLocationReport") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ul/IpAppUserLocation") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReport(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "periodicLocationReportErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ul/IpAppUserLocation") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReportErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "locationReportReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ul/IpUserLocation") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ul_IpUserLocation_locationReportReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "extendedLocationReportReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ul/IpUserLocation") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ul_IpUserLocation_extendedLocationReportReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "periodicLocationReportingStartReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ul/IpUserLocation") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ul_IpUserLocation_periodicLocationReportingStartReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "periodicLocationReportingStop") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ul/IpUserLocation") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ul_IpUserLocation_periodicLocationReportingStop(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getNextPeriodicLocationRequest") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ul/IpUserLocation") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ul_IpUserLocation_getNextPeriodicLocationRequest(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "triggeredLocationReport") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ul/IpAppTriggeredUserLocation") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ul_IpAppTriggeredUserLocation_triggeredLocationReport(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "triggeredLocationReportErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ul/IpAppTriggeredUserLocation") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ul_IpAppTriggeredUserLocation_triggeredLocationReportErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "triggeredLocationReportingStartReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ul/IpTriggeredUserLocation") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "triggeredLocationReportingStop") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ul/IpTriggeredUserLocation") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStop(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getNextTriggeredLocationRequest") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ul/IpTriggeredUserLocation") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ul_IpTriggeredUserLocation_getNextTriggeredLocationRequest(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "locationReportRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ulc/IpAppUserLocationCamel") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "locationReportErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ulc/IpAppUserLocationCamel") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "periodicLocationReport") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ulc/IpAppUserLocationCamel") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReport(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "periodicLocationReportErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ulc/IpAppUserLocationCamel") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReportErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "triggeredLocationReport") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ulc/IpAppUserLocationCamel") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ulc_IpAppUserLocationCamel_triggeredLocationReport(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "triggeredLocationReportErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ulc/IpAppUserLocationCamel") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ulc_IpAppUserLocationCamel_triggeredLocationReportErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "locationReportReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ulc/IpUserLocationCamel") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ulc_IpUserLocationCamel_locationReportReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "periodicLocationReportingStartReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ulc/IpUserLocationCamel") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ulc_IpUserLocationCamel_periodicLocationReportingStartReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "periodicLocationReportingStop") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ulc/IpUserLocationCamel") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ulc_IpUserLocationCamel_periodicLocationReportingStop(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "triggeredLocationReportingStartReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ulc/IpUserLocationCamel") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ulc_IpUserLocationCamel_triggeredLocationReportingStartReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "triggeredLocationReportingStop") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ulc/IpUserLocationCamel") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ulc_IpUserLocationCamel_triggeredLocationReportingStop(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getNextPeriodicLocationRequest") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ulc/IpUserLocationCamel") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ulc_IpUserLocationCamel_getNextPeriodicLocationRequest(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getNextTriggeredLocationRequest") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ulc/IpUserLocationCamel") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ulc_IpUserLocationCamel_getNextTriggeredLocationRequest(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "emergencyLocationReport") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ule/IpAppUserLocationEmergency") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ule_IpAppUserLocationEmergency_emergencyLocationReport(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "emergencyLocationReportErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ule/IpAppUserLocationEmergency") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ule_IpAppUserLocationEmergency_emergencyLocationReportErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "emergencyLocationReportReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ule/IpUserLocationEmergency") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ule_IpUserLocationEmergency_emergencyLocationReportReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "subscribeEmergencyLocationReports") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ule/IpUserLocationEmergency") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ule_IpUserLocationEmergency_subscribeEmergencyLocationReports(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "unSubscribeEmergencyLocationReports") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ule/IpUserLocationEmergency") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ule_IpUserLocationEmergency_unSubscribeEmergencyLocationReports(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "statusReportRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/us/IpAppUserStatus") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_us_IpAppUserStatus_statusReportRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "statusReportErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/us/IpAppUserStatus") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_us_IpAppUserStatus_statusReportErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "triggeredStatusReport") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/us/IpAppUserStatus") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_us_IpAppUserStatus_triggeredStatusReport(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "triggeredStatusReportErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/us/IpAppUserStatus") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_us_IpAppUserStatus_triggeredStatusReportErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "extendedStatusReportRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/us/IpAppUserStatus") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "extendedStatusReportErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/us/IpAppUserStatus") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "extTriggeredStatusReport") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/us/IpAppUserStatus") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_us_IpAppUserStatus_extTriggeredStatusReport(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "extTriggeredStatusReportErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/us/IpAppUserStatus") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_us_IpAppUserStatus_extTriggeredStatusReportErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "statusReportReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/us/IpUserStatus") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_us_IpUserStatus_statusReportReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "triggeredStatusReportingStartReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/us/IpUserStatus") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_us_IpUserStatus_triggeredStatusReportingStartReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "triggeredStatusReportingStop") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/us/IpUserStatus") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_us_IpUserStatus_triggeredStatusReportingStop(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getNextTriggeredStatusRequest") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/us/IpUserStatus") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_us_IpUserStatus_getNextTriggeredStatusRequest(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "extendedStatusReportReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/us/IpUserStatus") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_us_IpUserStatus_extendedStatusReportReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "extTriggeredStatusReportingStartReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/us/IpUserStatus") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_us_IpUserStatus_extTriggeredStatusReportingStartReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "extTriggeredStatusReportingStop") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/us/IpUserStatus") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_us_IpUserStatus_extTriggeredStatusReportingStop(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "triggeredBindingRequestNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ub/IpAppUserBinding") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "triggeredBindingRequestNotificationStartErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ub/IpAppUserBinding") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotificationStartErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "triggeredBindingRequestNotificationStartReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ub/IpUserBinding") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "triggeredBindingRequestNotificationStop") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mm/ub/IpUserBinding") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStop(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getTerminalCapabilities") == 0
            && (!idlname || strcmp(idlname, "org/csapi/termcap/IpTerminalCapabilities") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_termcap_IpTerminalCapabilities_getTerminalCapabilities(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "triggeredTerminalCapabilityReport") == 0
            && (!idlname || strcmp(idlname, "org/csapi/termcap/IpAppExtendedTerminalCapabilities") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReport(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "triggeredTerminalCapabilityReportErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/termcap/IpAppExtendedTerminalCapabilities") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReportErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "triggeredTerminalCapabilityStartReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/termcap/IpExtendedTerminalCapabilities") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStartReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "triggeredTerminalCapabilityStop") == 0
            && (!idlname || strcmp(idlname, "org/csapi/termcap/IpExtendedTerminalCapabilities") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStop(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "connectRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpAppDataSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpAppDataSession_connectRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "connectErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpAppDataSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpAppDataSession_connectErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "superviseDataSessionRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpAppDataSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpAppDataSession_superviseDataSessionRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "superviseDataSessionErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpAppDataSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpAppDataSession_superviseDataSessionErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "dataSessionFaultDetected") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpAppDataSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpAppDataSession_dataSessionFaultDetected(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "dataSessionAborted") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpAppDataSessionControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpAppDataSessionControlManager_dataSessionAborted(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "reportNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpAppDataSessionControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpAppDataSessionControlManager_reportNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "dataSessionNotificationContinued") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpAppDataSessionControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpAppDataSessionControlManager_dataSessionNotificationContinued(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "dataSessionNotificationInterrupted") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpAppDataSessionControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpAppDataSessionControlManager_dataSessionNotificationInterrupted(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "abortMultipleDataSessions") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpAppDataSessionControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpAppDataSessionControlManager_abortMultipleDataSessions(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "connectReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpDataSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpDataSession_connectReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "release") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpDataSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpDataSession_release(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "superviseDataSessionReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpDataSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpDataSession_superviseDataSessionReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setDataSessionChargePlan") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpDataSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpDataSession_setDataSessionChargePlan(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setAdviceOfCharge") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpDataSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpDataSession_setAdviceOfCharge(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deassignDataSession") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpDataSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpDataSession_deassignDataSession(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "continueProcessing") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpDataSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpDataSession_continueProcessing(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpDataSessionControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpDataSessionControlManager_createNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "destroyNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpDataSessionControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpDataSessionControlManager_destroyNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "changeNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpDataSessionControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpDataSessionControlManager_changeNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "enableNotifications") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpDataSessionControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpDataSessionControlManager_enableNotifications(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "disableNotifications") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpDataSessionControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpDataSessionControlManager_disableNotifications(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getNotifications") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpDataSessionControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpDataSessionControlManager_getNotifications(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createNotifications") == 0
            && (!idlname || strcmp(idlname, "org/csapi/dsc/IpDataSessionControlManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_dsc_IpDataSessionControlManager_createNotifications(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "openMailbox") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMessagingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMessagingManager_openMailbox(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "enableMessagingNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMessagingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMessagingManager_enableMessagingNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "disableMessagingNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMessagingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMessagingManager_disableMessagingNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "mailboxTerminated") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpAppMessagingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpAppMessagingManager_mailboxTerminated(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "mailboxFaultDetected") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpAppMessagingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpAppMessagingManager_mailboxFaultDetected(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "messagingEventNotify") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpAppMessagingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpAppMessagingManager_messagingEventNotify(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "messagingNotificationTerminated") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpAppMessagingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpAppMessagingManager_messagingNotificationTerminated(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "close") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMailbox_close(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "lock") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMailbox_lock(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "unlock") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMailbox_unlock(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getInfoAmount") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMailbox_getInfoAmount(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getInfoProperties") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMailbox_getInfoProperties(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setInfoProperties") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMailbox_setInfoProperties(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "openFolder") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMailbox_openFolder(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createFolder") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMailbox_createFolder(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "remove") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMailbox_remove(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getInfoAmount") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMessage") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMessage_getInfoAmount(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getInfoProperties") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMessage") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMessage_getInfoProperties(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setInfoProperties") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMessage") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMessage_setInfoProperties(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "remove") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMessage") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMessage_remove(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getContent") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMessage") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMessage_getContent(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getInfoAmount") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMailboxFolder") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMailboxFolder_getInfoAmount(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getInfoProperties") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMailboxFolder") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMailboxFolder_getInfoProperties(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setInfoProperties") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMailboxFolder") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMailboxFolder_setInfoProperties(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "putMessage") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMailboxFolder") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMailboxFolder_putMessage(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMessage") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMailboxFolder") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMailboxFolder_getMessage(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "close") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMailboxFolder") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMailboxFolder_close(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "remove") == 0
            && (!idlname || strcmp(idlname, "org/csapi/gms/IpMailboxFolder") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_gms_IpMailboxFolder_remove(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getVPrPID") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpVPrP") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpVPrP_getVPrPID(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getSlaID") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpVPrP") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpVPrP_getSlaID(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getStatus") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpVPrP") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpVPrP_getStatus(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getProvisionedQoSInfo") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpVPrP") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpVPrP_getProvisionedQoSInfo(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getValidityInfo") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpVPrP") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpVPrP_getValidityInfo(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getPipeQoSInfo") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpVPrP") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpVPrP_getPipeQoSInfo(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getDsCodepoint") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpVPrP") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpVPrP_getDsCodepoint(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getVPrPList") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpVPrN") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpVPrN_getVPrPList(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getVPrP") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpVPrN") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpVPrN_getVPrP(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createVPrP") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpVPrN") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpVPrN_createVPrP(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteVPrP") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpVPrN") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpVPrN_deleteVPrP(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getTemplate") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpQoSMenu") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpQoSMenu_getTemplate(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getTemplateList") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpQoSMenu") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpQoSMenu_getTemplateList(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getQoSMenu") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpConnectivityManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpConnectivityManager_getQoSMenu(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getEnterpriseNetwork") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpConnectivityManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpConnectivityManager_getEnterpriseNetwork(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getTemplateType") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpQoSTemplate") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpQoSTemplate_getTemplateType(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getDescription") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpQoSTemplate") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpQoSTemplate_getDescription(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setSlaID") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpQoSTemplate") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpQoSTemplate_setSlaID(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getPipeQoSInfo") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpQoSTemplate") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpQoSTemplate_getPipeQoSInfo(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setPipeQoSInfo") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpQoSTemplate") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpQoSTemplate_setPipeQoSInfo(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getValidityInfo") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpQoSTemplate") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpQoSTemplate_getValidityInfo(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setValidityInfo") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpQoSTemplate") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpQoSTemplate_setValidityInfo(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setProvisionedQoSInfo") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpQoSTemplate") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpQoSTemplate_setProvisionedQoSInfo(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getProvisionedQoSInfo") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpQoSTemplate") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpQoSTemplate_getProvisionedQoSInfo(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getDsCodepoint") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpQoSTemplate") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpQoSTemplate_getDsCodepoint(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getSiteList") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpEnterpriseNetwork") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpEnterpriseNetwork_getSiteList(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getVPrN") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpEnterpriseNetwork") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpEnterpriseNetwork_getVPrN(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getSite") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpEnterpriseNetwork") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpEnterpriseNetwork_getSite(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getSAPList") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpEnterpriseNetworkSite") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpEnterpriseNetworkSite_getSAPList(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getSiteID") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpEnterpriseNetworkSite") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpEnterpriseNetworkSite_getSiteID(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getSiteLocation") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpEnterpriseNetworkSite") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpEnterpriseNetworkSite_getSiteLocation(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getSiteDescription") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpEnterpriseNetworkSite") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpEnterpriseNetworkSite_getSiteDescription(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getIPSubnet") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpEnterpriseNetworkSite") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpEnterpriseNetworkSite_getIPSubnet(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getSAPIPSubnet") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cm/IpEnterpriseNetworkSite") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cm_IpEnterpriseNetworkSite_getSAPIPSubnet(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "reportNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAppAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAppAccountManager_reportNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryBalanceRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAppAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAppAccountManager_queryBalanceRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryBalanceErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAppAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAppAccountManager_queryBalanceErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "retrieveTransactionHistoryRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAppAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "retrieveTransactionHistoryErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAppAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryBalanceExpiryDateRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAppAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryBalanceExpiryDateErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAppAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "updateBalanceRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAppAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAppAccountManager_updateBalanceRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "updateBalanceErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAppAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAppAccountManager_updateBalanceErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createVoucherRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAppAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAppAccountManager_createVoucherRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createVoucherErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAppAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAppAccountManager_createVoucherErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "destroyVoucherRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAppAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAppAccountManager_destroyVoucherRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "destroyVoucherErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAppAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAppAccountManager_destroyVoucherErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryVoucherRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAppAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAppAccountManager_queryVoucherRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryVoucherErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAppAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAppAccountManager_queryVoucherErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryUserVouchersRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAppAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAppAccountManager_queryUserVouchersRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryUserVouchersErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAppAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAppAccountManager_queryUserVouchersErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAccountManager_createNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "destroyNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAccountManager_destroyNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryBalanceReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAccountManager_queryBalanceReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "changeNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAccountManager_changeNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAccountManager_getNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "retrieveTransactionHistoryReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAccountManager_retrieveTransactionHistoryReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "enableNotifications") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAccountManager_enableNotifications(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "disableNotifications") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAccountManager_disableNotifications(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryBalanceExpiryDateReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAccountManager_queryBalanceExpiryDateReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "updateBalanceReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAccountManager_updateBalanceReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createVoucherReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAccountManager_createVoucherReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "destroyVoucherReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAccountManager_destroyVoucherReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryVoucherReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAccountManager_queryVoucherReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryUserVouchersReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/am/IpAccountManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_am_IpAccountManager_queryUserVouchersReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "creditAmountErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_creditAmountErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "creditAmountRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_creditAmountRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "creditUnitErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_creditUnitErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "creditUnitRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_creditUnitRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "debitAmountErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_debitAmountErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "debitAmountRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_debitAmountRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "debitUnitErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_debitUnitErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "debitUnitRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_debitUnitRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "directCreditAmountErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_directCreditAmountErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "directCreditAmountRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_directCreditAmountRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "directCreditUnitErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_directCreditUnitErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "directCreditUnitRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_directCreditUnitRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "directDebitAmountErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_directDebitAmountErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "directDebitAmountRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_directDebitAmountRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "directDebitUnitErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_directDebitUnitErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "directDebitUnitRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_directDebitUnitRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "extendLifeTimeErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_extendLifeTimeErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "extendLifeTimeRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_extendLifeTimeRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "rateErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_rateErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "rateRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_rateRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "reserveAmountErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_reserveAmountErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "reserveAmountRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_reserveAmountRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "reserveUnitErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_reserveUnitErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "reserveUnitRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_reserveUnitRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "sessionEnded") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingSession_sessionEnded(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "creditAmountReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpChargingSession_creditAmountReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "creditUnitReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpChargingSession_creditUnitReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "debitAmountReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpChargingSession_debitAmountReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "debitUnitReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpChargingSession_debitUnitReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "directCreditAmountReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpChargingSession_directCreditAmountReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "directCreditUnitReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpChargingSession_directCreditUnitReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "directDebitAmountReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpChargingSession_directDebitAmountReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "directDebitUnitReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpChargingSession_directDebitUnitReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "extendLifeTimeReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpChargingSession_extendLifeTimeReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getAmountLeft") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpChargingSession_getAmountLeft(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getLifeTimeLeft") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpChargingSession_getLifeTimeLeft(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getUnitLeft") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpChargingSession_getUnitLeft(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "rateReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpChargingSession_rateReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "release") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpChargingSession_release(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "reserveAmountReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpChargingSession_reserveAmountReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "reserveUnitReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpChargingSession") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpChargingSession_reserveUnitReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "sessionAborted") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingManager_sessionAborted(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "abortMultipleChargingSessions") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpAppChargingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpAppChargingManager_abortMultipleChargingSessions(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createChargingSession") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpChargingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpChargingManager_createChargingSession(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createSplitChargingSession") == 0
            && (!idlname || strcmp(idlname, "org/csapi/cs/IpChargingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_cs_IpChargingManager_createSplitChargingSession(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "reportNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/evaluation/IpAppPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_evaluation_IpAppPolicyDomain_reportNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "evalPolicyRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/evaluation/IpAppPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "evalPolicyErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/evaluation/IpAppPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "evalPolicy") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/evaluation/IpPolicyEvalManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "evalPolicyReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/evaluation/IpPolicyEvalManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicyReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "abortEvalPolicyReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/evaluation/IpPolicyEvalManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_evaluation_IpPolicyEvalManager_abortEvalPolicyReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "generateEvent") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/evaluation/IpPolicyEvalManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_evaluation_IpPolicyEvalManager_generateEvent(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/evaluation/IpPolicyEvalManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_evaluation_IpPolicyEvalManager_createNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "destroyNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/evaluation/IpPolicyEvalManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_evaluation_IpPolicyEvalManager_destroyNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getAttribute") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicy") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicy_getAttribute(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setAttribute") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicy") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicy_setAttribute(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getAttributes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicy") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicy_getAttributes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setAttributes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicy") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicy_setAttributes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getList") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyIterator") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyIterator_getList(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getParentRepository") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRepository") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRepository_getParentRepository(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createRepository") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRepository") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRepository_createRepository(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getRepository") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRepository") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRepository_getRepository(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "removeRepository") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRepository") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRepository_removeRepository(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getRepositoryCount") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRepository") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRepository_getRepositoryCount(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getRepositoryIterator") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRepository") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRepository_getRepositoryIterator(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createCondition") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRepository") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRepository_createCondition(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getCondition") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRepository") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRepository_getCondition(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "removeCondition") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRepository") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRepository_removeCondition(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getConditionCount") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRepository") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRepository_getConditionCount(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getConditionIterator") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRepository") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRepository_getConditionIterator(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createAction") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRepository") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRepository_createAction(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getAction") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRepository") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRepository_getAction(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "removeAction") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRepository") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRepository_removeAction(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getActionCount") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRepository") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRepository_getActionCount(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getActionIterator") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRepository") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRepository_getActionIterator(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getParentGroup") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRule") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRule_getParentGroup(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getParentDomain") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRule") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRule_getParentDomain(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createCondition") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRule") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRule_createCondition(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getCondition") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRule") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRule_getCondition(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "removeCondition") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRule") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRule_removeCondition(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getConditionCount") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRule") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRule_getConditionCount(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getConditionIterator") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRule") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRule_getConditionIterator(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createAction") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRule") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRule_createAction(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getAction") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRule") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRule_getAction(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "removeAction") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRule") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRule_removeAction(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getActionCount") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRule") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRule_getActionCount(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getActionIterator") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRule") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRule_getActionIterator(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setValidityPeriodConditionByName") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRule") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRule_setValidityPeriodConditionByName(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setValidityPeriodCondition") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRule") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRule_setValidityPeriodCondition(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getValidityPeriodCondition") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRule") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRule_getValidityPeriodCondition(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "unsetValidityPeriodCondition") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRule") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRule_unsetValidityPeriodCondition(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setConditionList") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRule") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRule_setConditionList(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getConditionList") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRule") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRule_getConditionList(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setActionList") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRule") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRule_setActionList(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getActionList") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyRule") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyRule_getActionList(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createDomain") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyManager_createDomain(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getDomain") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyManager_getDomain(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "removeDomain") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyManager_removeDomain(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getDomainCount") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyManager_getDomainCount(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getDomainIterator") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyManager_getDomainIterator(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "findMatchingDomains") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createRepository") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyManager_createRepository(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getRepository") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyManager_getRepository(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "removeRepository") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyManager_removeRepository(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getRepositoryCount") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyManager_getRepositoryCount(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getRepositoryIterator") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyManager_getRepositoryIterator(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "startTransaction") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyManager_startTransaction(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "commitTransaction") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyManager_commitTransaction(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "abortTransaction") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyManager_abortTransaction(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getParentDomain") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyGroup") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyGroup_getParentDomain(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getParentGroup") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyGroup") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyGroup_getParentGroup(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createGroup") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyGroup") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyGroup_createGroup(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getGroup") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyGroup") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyGroup_getGroup(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "removeGroup") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyGroup") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyGroup_removeGroup(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getGroupCount") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyGroup") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyGroup_getGroupCount(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getGroupIterator") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyGroup") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyGroup_getGroupIterator(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createRule") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyGroup") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyGroup_createRule(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getRule") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyGroup") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyGroup_getRule(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "removeRule") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyGroup") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyGroup_removeRule(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getRuleCount") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyGroup") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyGroup_getRuleCount(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getRuleIterator") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyGroup") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyGroup_getRuleIterator(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setRequiredAttributes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyEventDefinition") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyEventDefinition_setRequiredAttributes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setOptionalAttributes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyEventDefinition") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyEventDefinition_setOptionalAttributes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getRequiredAttributes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyEventDefinition") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyEventDefinition_getRequiredAttributes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getOptionalAttributes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyEventDefinition") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyEventDefinition_getOptionalAttributes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getParentDomain") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyEventDefinition") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyEventDefinition_getParentDomain(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getParentRepository") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyCondition") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyCondition_getParentRepository(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getParentRule") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyCondition") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyCondition_getParentRule(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getParentRepository") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyAction") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyAction_getParentRepository(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getParentRule") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyAction") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyAction_getParentRule(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setInputVariables") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicySignature") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicySignature_setInputVariables(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setOutputVariables") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicySignature") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicySignature_setOutputVariables(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getInputVariables") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicySignature") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicySignature_getInputVariables(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getOutputVariables") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicySignature") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicySignature_getOutputVariables(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setGroupNames") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicySignature") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicySignature_setGroupNames(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setPolicyRoles") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicySignature") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicySignature_setPolicyRoles(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getGroupNames") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicySignature") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicySignature_getGroupNames(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getPolicyRoles") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicySignature") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicySignature_getPolicyRoles(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getParentDomain") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicySignature") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicySignature_getParentDomain(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getParentDomain") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_getParentDomain(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createDomain") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_createDomain(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getDomain") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_getDomain(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "removeDomain") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_removeDomain(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getDomainCount") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_getDomainCount(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getDomainIterator") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_getDomainIterator(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createGroup") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_createGroup(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getGroup") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_getGroup(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "removeGroup") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_removeGroup(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getGroupCount") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_getGroupCount(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getGroupIterator") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_getGroupIterator(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createRule") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_createRule(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getRule") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_getRule(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "removeRule") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_removeRule(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getRuleCount") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_getRuleCount(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getRuleIterator") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_getRuleIterator(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createEventDefinition") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getEventDefinition") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_getEventDefinition(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "removeEventDefinition") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_removeEventDefinition(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getEventDefinitionCount") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_getEventDefinitionCount(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getEventDefinitionIterator") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_getEventDefinitionIterator(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createVariableSet") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_createVariableSet(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getVariableSet") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_getVariableSet(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "removeVariableSet") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_removeVariableSet(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getVariableSetCount") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_getVariableSetCount(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getVariableSetIterator") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_getVariableSetIterator(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createVariable") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_createVariable(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setVariableValue") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_setVariableValue(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getVariableType") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_getVariableType(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getVariableValue") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_getVariableValue(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getVariable") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_getVariable(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "removeVariable") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_removeVariable(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createSignature") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_createSignature(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getSignature") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_getSignature(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "removeSignature") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_removeSignature(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getSignatureCount") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_getSignatureCount(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getSignatureIterator") == 0
            && (!idlname || strcmp(idlname, "org/csapi/policy/provisioning/IpPolicyDomain") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_policy_provisioning_IpPolicyDomain_getSignatureIterator(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createIdentity") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteIdentity") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteIdentity(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "isIdentity") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_isIdentity(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createGroupIdentity") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteGroupIdentity") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteGroupIdentity(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "addToGroup") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_addToGroup(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "removeFromGroup") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeFromGroup(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listMembers") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "isGroupIdentity") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_isGroupIdentity(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listGroupMembership") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "addAlias") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_addAlias(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "removeAliases") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeAliases(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listAliases") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "lookupByAlias") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_lookupByAlias(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "associateTypes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "disassociateTypes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listTypesOfIdentity") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "hasType") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getIdentityAttributes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setIdentityAttributes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createAgent") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteAgent") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentManagement_deleteAgent(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "isAgent") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentManagement_isAgent(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "enableCapabilities") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "disableCapabilities") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listEnabledCapabilities") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listAllCapabilities") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "isCapableOf") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "associateTypes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "disassociateTypes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listTypesOfAgent") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "hasType") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getAgentAttributes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setAgentAttributes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "assignAgent") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentAssignment") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentAssignment_assignAgent(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "unassignAgent") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentAssignment") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentAssignment_unassignAgent(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listAssignedAgents") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentAssignment") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listAssociatedIdentitiesOfAgent") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentAssignment") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listAssignedAgentsByCapability") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentAssignment") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listCapabilitiesOfIdentity") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentAssignment") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "isIdentityCapableOf") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentAssignment") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createIdentityAttribute") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityTypeManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityAttribute(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteIdentityAttribute") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityTypeManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityAttribute(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getIdentityAttributeDefinition") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityTypeManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_getIdentityAttributeDefinition(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listAllIdentityAttributes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityTypeManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createIdentityType") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityTypeManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteIdentityType") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityTypeManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityType(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listIdentityTypes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityTypeManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "addIdentityTypeAttributes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityTypeManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "removeIdentityTypeAttributes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityTypeManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listIdentityTypeAttributes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMIdentityTypeManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createAgentAttribute") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentTypeManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentAttribute(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteAgentAttribute") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentTypeManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentAttribute(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getAgentAttributeDefinition") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentTypeManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_getAgentAttributeDefinition(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listAllAgentAttributes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentTypeManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createAgentType") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentTypeManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteAgentType") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentTypeManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentType(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listAgentTypes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentTypeManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "addAgentTypeAttributes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentTypeManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "removeAgentTypeAttributes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentTypeManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listAgentTypeAttributes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMAgentTypeManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createCapabilityAttribute") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMCapabilityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapabilityAttribute(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteCapabilityAttribute") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMCapabilityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapabilityAttribute(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getCapabilityAttributeDefinition") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMCapabilityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_getCapabilityAttributeDefinition(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listAllCapabilityAttributes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMCapabilityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createCapability") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMCapabilityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteCapability") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMCapabilityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapability(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listCapabilities") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMCapabilityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "addCapabilityAttributes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMCapabilityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "removeCapabilityAttributes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMCapabilityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listCapabilityAttributes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMCapabilityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "assignCapabilitiesToType") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMCapabilityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "unassignCapabilitiesFromType") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMCapabilityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listCapabilitiesOfType") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMCapabilityManagement") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getAuthToken") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMProvisioningManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "obtainInterface") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMProvisioningManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMProvisioningManager_obtainInterface(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getAccessControl") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMProvisioningManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAccessControl(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setAccessControl") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/provisioning/IpPAMProvisioningManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_provisioning_IpPAMProvisioningManager_setAccessControl(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setIdentityPresence") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/access/IpPAMIdentityPresence") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setIdentityPresenceExpiration") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/access/IpPAMIdentityPresence") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getIdentityPresence") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/access/IpPAMIdentityPresence") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setAgentPresence") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/access/IpPAMAgentPresence") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setCapabilityPresence") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/access/IpPAMAgentPresence") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setAgentPresenceExpiration") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/access/IpPAMAgentPresence") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setCapabilityPresenceExpiration") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/access/IpPAMAgentPresence") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getAgentPresence") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/access/IpPAMAgentPresence") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getCapabilityPresence") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/access/IpPAMAgentPresence") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "computeAvailability") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/access/IpAppPAMPreferenceCheck") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getAvailability") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/access/IpPAMAvailability") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_access_IpPAMAvailability_getAvailability(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getPreference") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/access/IpPAMAvailability") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_access_IpPAMAvailability_getPreference(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setPreference") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/access/IpPAMAvailability") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_access_IpPAMAvailability_setPreference(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getAuthToken") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/access/IpPAMPresenceAvailabilityManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "obtainInterface") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/access/IpPAMPresenceAvailabilityManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_obtainInterface(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getAccessControl") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/access/IpPAMPresenceAvailabilityManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAccessControl(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setAccessControl") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/access/IpPAMPresenceAvailabilityManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_setAccessControl(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "activateService") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/access/IpPAMPresenceAvailabilityManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_activateService(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deactivateService") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/access/IpPAMPresenceAvailabilityManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_deactivateService(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "isActiveIdentity") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/access/IpPAMPresenceAvailabilityManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_isActiveIdentity(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "eventNotify") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/event/IpAppPAMEventHandler") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_event_IpAppPAMEventHandler_eventNotify(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "eventNotifyErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/event/IpAppPAMEventHandler") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_event_IpAppPAMEventHandler_eventNotifyErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "isRegistered") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/event/IpPAMEventHandler") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_event_IpPAMEventHandler_isRegistered(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "registerAppInterface") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/event/IpPAMEventHandler") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_event_IpPAMEventHandler_registerAppInterface(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "registerForEvent") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/event/IpPAMEventHandler") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_event_IpPAMEventHandler_registerForEvent(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deregisterAppInterface") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/event/IpPAMEventHandler") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_event_IpPAMEventHandler_deregisterAppInterface(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deregisterFromEvent") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/event/IpPAMEventHandler") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_event_IpPAMEventHandler_deregisterFromEvent(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getAuthToken") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/event/IpPAMEventManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_event_IpPAMEventManager_getAuthToken(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "obtainInterface") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/event/IpPAMEventManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_event_IpPAMEventManager_obtainInterface(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getAccessControl") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/event/IpPAMEventManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_event_IpPAMEventManager_getAccessControl(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setAccessControl") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/event/IpPAMEventManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_event_IpPAMEventManager_setAccessControl(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "activateService") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/event/IpPAMEventManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_event_IpPAMEventManager_activateService(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deactivateService") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/event/IpPAMEventManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_event_IpPAMEventManager_deactivateService(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "isActiveIdentity") == 0
            && (!idlname || strcmp(idlname, "org/csapi/pam/event/IpPAMEventManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_pam_event_IpPAMEventManager_isActiveIdentity(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "mailboxTerminated") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMultiMediaMessagingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMultiMediaMessagingManager_mailboxTerminated(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "reportNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMultiMediaMessagingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMultiMediaMessagingManager_reportNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "notificationsInterrupted") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMultiMediaMessagingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMultiMediaMessagingManager_notificationsInterrupted(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "notificationsResumed") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMultiMediaMessagingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMultiMediaMessagingManager_notificationsResumed(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "multiMediaMessagingTerminated") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMultiMediaMessagingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMultiMediaMessagingManager_multiMediaMessagingTerminated(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "terminateMultipleMailboxes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMultiMediaMessagingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMailboxes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "terminateMultipleMultiMediaMessagingSessions") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMultiMediaMessagingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMultiMediaMessagingSessions(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "openMailbox") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMultiMediaMessagingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMultiMediaMessagingManager_openMailbox(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "openMultiMediaMessaging") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMultiMediaMessagingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMultiMediaMessagingManager_openMultiMediaMessaging(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMultiMediaMessagingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMultiMediaMessagingManager_createNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "destroyNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMultiMediaMessagingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMultiMediaMessagingManager_destroyNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "changeNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMultiMediaMessagingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMultiMediaMessagingManager_changeNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getNextNotification") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMultiMediaMessagingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMultiMediaMessagingManager_getNextNotification(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "enableNotifications") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMultiMediaMessagingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMultiMediaMessagingManager_enableNotifications(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "disableNotifications") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMultiMediaMessagingManager") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMultiMediaMessagingManager_disableNotifications(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createFolderRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_createFolderRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createFolderErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_createFolderErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getFoldersRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_getFoldersRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getFoldersErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_getFoldersErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteFolderRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_deleteFolderRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteFolderErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_deleteFolderErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "copyFolderRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_copyFolderRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "copyFolderErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_copyFolderErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "moveFolderRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_moveFolderRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "moveFolderErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_moveFolderErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "putMessageRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_putMessageRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "putMessageErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_putMessageErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "copyMessageRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_copyMessageRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "copyMessageErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_copyMessageErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "moveMessageRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_moveMessageRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "moveMessageErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_moveMessageErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteMessageRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_deleteMessageRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteMessageErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_deleteMessageErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listMessagesRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_listMessagesRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listMessagesErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_listMessagesErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listMessageBodyPartsRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listMessageBodyPartsErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMessageBodyPartsRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMessageBodyPartsErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMessageHeadersRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_getMessageHeadersRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMessageHeadersErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_getMessageHeadersErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMessageContentRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_getMessageContentRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMessageContentErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_getMessageContentErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getFullMessageRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_getFullMessageRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getFullMessageErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_getFullMessageErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMailboxInfoPropertiesRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getFolderInfoPropertiesRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMessageInfoPropertiesRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setMessageInfoPropertiesRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setMessageInfoPropertiesErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMailboxInfoPropertiesErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getFolderInfoPropertiesErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMessageInfoPropertiesErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "close") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMailbox_close(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "createFolderReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMailbox_createFolderReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getFoldersReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMailbox_getFoldersReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteFolderReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMailbox_deleteFolderReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "copyFolderReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMailbox_copyFolderReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "moveFolderReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMailbox_moveFolderReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "putMessageReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMailbox_putMessageReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "copyMessageReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMailbox_copyMessageReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "moveMessageReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMailbox_moveMessageReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "deleteMessageReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMailbox_deleteMessageReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listMessagesReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMailbox_listMessagesReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "listMessageBodyPartsReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMailbox_listMessageBodyPartsReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMessageBodyPartsReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMessageHeadersReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMailbox_getMessageHeadersReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMessageContentReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMailbox_getMessageContentReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getFullMessageReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMailbox_getFullMessageReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMailboxInfoPropertiesReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMailbox_getMailboxInfoPropertiesReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getFolderInfoPropertiesReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMailbox_getFolderInfoPropertiesReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "getMessageInfoPropertiesReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMailbox_getMessageInfoPropertiesReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "setMessageInfoPropertiesReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMailbox") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMailbox_setMessageInfoPropertiesReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "sendMessageRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMultiMediaMessaging") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMultiMediaMessaging_sendMessageRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "sendMessageErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMultiMediaMessaging") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMultiMediaMessaging_sendMessageErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "cancelMessageRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMultiMediaMessaging") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMultiMediaMessaging_cancelMessageRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "cancelMessageErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMultiMediaMessaging") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMultiMediaMessaging_cancelMessageErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryStatusRes") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMultiMediaMessaging") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusRes(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryStatusErr") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMultiMediaMessaging") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusErr(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "messageStatusReport") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMultiMediaMessaging") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMultiMediaMessaging_messageStatusReport(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "messageReceived") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpAppMultiMediaMessaging") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "sendMessageReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMultiMediaMessaging") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "cancelMessageReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMultiMediaMessaging") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMultiMediaMessaging_cancelMessageReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "queryStatusReq") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMultiMediaMessaging") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMultiMediaMessaging_queryStatusReq(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        if (strcmp(operation, "close") == 0
            && (!idlname || strcmp(idlname, "org/csapi/mmm/IpMultiMediaMessaging") == 0)) {
            item = process_RequestOperation(tvb, pinfo, ptree, header, operation);  /* fill-up Request_Operation field & info column */
            tree = start_dissecting(tvb, pinfo, ptree, offset);
            decode_org_csapi_mmm_IpMultiMediaMessaging_close(tvb, pinfo, tree, item, offset, header, operation, stream_is_big_endian);
            return TRUE;
        }

        break;

    case CancelRequest:
    case LocateRequest:
    case LocateReply:
    case CloseConnection:
    case MessageError:
    case Fragment:
       return FALSE;      /* not handled yet */

    default:
       return FALSE;      /* not handled yet */

    }   /* switch */


    return FALSE;

}  /* End of main dissector  */


/* Register the protocol with Wireshark */
void proto_register_giop_parlay(void)
{
    /* setup list of header fields */
    static hf_register_info hf[] = {
        /* field that indicates the currently ongoing request/reply exchange */
            {&hf_operationrequest, {"Request_Operation","giop-parlay.Request_Operation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        /* Operation filters */
        {&hf_org_csapi_IpService_setCallbackWithSessionID_sessionID, {"sessionID","giop-parlay.csapi.setCallbackWithSessionID.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpInitial_initiateAuthentication_authType, {"authType","giop-parlay.csapi.initiateAuthentication.authType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpInitial_initiateAuthenticationWithVersion_authType, {"authType","giop-parlay.csapi.initiateAuthenticationWithVersion.authType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpInitial_initiateAuthenticationWithVersion_frameworkVersion, {"frameworkVersion","giop-parlay.csapi.initiateAuthenticationWithVersion.frameworkVersion",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAuthentication_requestAccess_accessType, {"accessType","giop-parlay.csapi.requestAccess.accessType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpClientAccess_terminateAccess_terminationText, {"terminationText","giop-parlay.csapi.terminateAccess.terminationText",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpClientAccess_terminateAccess_signingAlgorithm, {"signingAlgorithm","giop-parlay.csapi.terminateAccess.signingAlgorithm",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpClientAccess_terminateAccess_digitalSignature_loop, {"Seq length of digitalSignature","giop-parlay.csapi.terminateAccess.digitalSignature.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpClientAccess_terminateAccess_digitalSignature, {"digitalSignature","giop-parlay.csapi.terminateAccess.digitalSignature",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_obtainInterface_interfaceName, {"interfaceName","giop-parlay.csapi.obtainInterface.interfaceName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_obtainInterfaceWithCallback_interfaceName, {"interfaceName","giop-parlay.csapi.obtainInterfaceWithCallback.interfaceName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_endAccess_endAccessProperties_loop, {"Seq length of endAccessProperties","giop-parlay.csapi.endAccess.endAccessProperties.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_listInterfaces_return_loop, {"Seq length of TpInterfaceNameList","giop-parlay.csapi.listInterfaces.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_listInterfaces_return, {"TpInterfaceNameList","giop-parlay.csapi.listInterfaces.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_releaseInterface_interfaceName, {"interfaceName","giop-parlay.csapi.releaseInterface.interfaceName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_selectSigningAlgorithm_return, {"TpSigningAlgorithm","giop-parlay.csapi.selectSigningAlgorithm.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_selectSigningAlgorithm_signingAlgorithmCaps, {"signingAlgorithmCaps","giop-parlay.csapi.selectSigningAlgorithm.signingAlgorithmCaps",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_terminateAccess_terminationText, {"terminationText","giop-parlay.csapi.terminateAccess.terminationText",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_terminateAccess_digitalSignature_loop, {"Seq length of digitalSignature","giop-parlay.csapi.terminateAccess.digitalSignature.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_terminateAccess_digitalSignature, {"digitalSignature","giop-parlay.csapi.terminateAccess.digitalSignature",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_relinquishInterface_interfaceName, {"interfaceName","giop-parlay.csapi.relinquishInterface.interfaceName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_relinquishInterface_terminationText, {"terminationText","giop-parlay.csapi.relinquishInterface.terminationText",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_relinquishInterface_digitalSignature_loop, {"Seq length of digitalSignature","giop-parlay.csapi.relinquishInterface.digitalSignature.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAccess_relinquishInterface_digitalSignature, {"digitalSignature","giop-parlay.csapi.relinquishInterface.digitalSignature",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_return_loop, {"Seq length of TpOctetSet","giop-parlay.csapi.authenticate.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_return, {"TpOctetSet","giop-parlay.csapi.authenticate.return",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_challenge_loop, {"Seq length of challenge","giop-parlay.csapi.authenticate.challenge.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_authenticate_challenge, {"challenge","giop-parlay.csapi.authenticate.challenge",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_return_loop, {"Seq length of TpOctetSet","giop-parlay.csapi.challenge.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_return, {"TpOctetSet","giop-parlay.csapi.challenge.return",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_challenge_loop, {"Seq length of challenge","giop-parlay.csapi.challenge.challenge.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpClientAPILevelAuthentication_challenge_challenge, {"challenge","giop-parlay.csapi.challenge.challenge",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_selectEncryptionMethod_return, {"TpEncryptionCapability","giop-parlay.csapi.selectEncryptionMethod.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_selectEncryptionMethod_encryptionCaps, {"encryptionCaps","giop-parlay.csapi.selectEncryptionMethod.encryptionCaps",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_return_loop, {"Seq length of TpOctetSet","giop-parlay.csapi.authenticate.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_return, {"TpOctetSet","giop-parlay.csapi.authenticate.return",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_challenge_loop, {"Seq length of challenge","giop-parlay.csapi.authenticate.challenge.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_authenticate_challenge, {"challenge","giop-parlay.csapi.authenticate.challenge",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_selectAuthenticationMechanism_return, {"TpAuthMechanism","giop-parlay.csapi.selectAuthenticationMechanism.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_selectAuthenticationMechanism_authMechanismList, {"authMechanismList","giop-parlay.csapi.selectAuthenticationMechanism.authMechanismList",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_return_loop, {"Seq length of TpOctetSet","giop-parlay.csapi.challenge.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_return, {"TpOctetSet","giop-parlay.csapi.challenge.return",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_challenge_loop, {"Seq length of challenge","giop-parlay.csapi.challenge.challenge.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_access_trust_and_security_IpAPILevelAuthentication_challenge_challenge, {"challenge","giop-parlay.csapi.challenge.challenge",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_notification_IpAppEventNotification_reportNotification_assignmentID, {"assignmentID","giop-parlay.csapi.reportNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_notification_IpEventNotification_createNotification_return, {"TpAssignmentID","giop-parlay.csapi.createNotification.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_notification_IpEventNotification_destroyNotification_assignmentID, {"assignmentID","giop-parlay.csapi.destroyNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_activityTestRes_activityTestID, {"activityTestID","giop-parlay.csapi.activityTestRes.activityTestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_activityTestRes_activityTestResult, {"activityTestResult","giop-parlay.csapi.activityTestRes.activityTestResult",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_appActivityTestReq_activityTestID, {"activityTestID","giop-parlay.csapi.appActivityTestReq.activityTestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_fwFaultReportInd_fault, {"fault","giop-parlay.csapi.fwFaultReportInd.fault",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpInterfaceFault),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_fwFaultRecoveryInd_fault, {"fault","giop-parlay.csapi.fwFaultRecoveryInd.fault",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpInterfaceFault),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_svcUnavailableInd_serviceID, {"serviceID","giop-parlay.csapi.svcUnavailableInd.serviceID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_svcUnavailableInd_reason, {"reason","giop-parlay.csapi.svcUnavailableInd.reason",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpSvcUnavailReason),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordRes_serviceIDs_loop, {"Seq length of serviceIDs","giop-parlay.csapi.genFaultStatsRecordRes.serviceIDs.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordRes_serviceIDs, {"serviceIDs","giop-parlay.csapi.genFaultStatsRecordRes.serviceIDs",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_fwUnavailableInd_reason, {"reason","giop-parlay.csapi.fwUnavailableInd.reason",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpFwUnavailReason),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_activityTestErr_activityTestID, {"activityTestID","giop-parlay.csapi.activityTestErr.activityTestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordErr_faultStatisticsError, {"faultStatisticsError","giop-parlay.csapi.genFaultStatsRecordErr.faultStatisticsError",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpFaultStatisticsError),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordErr_serviceIDs_loop, {"Seq length of serviceIDs","giop-parlay.csapi.genFaultStatsRecordErr.serviceIDs.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_genFaultStatsRecordErr_serviceIDs, {"serviceIDs","giop-parlay.csapi.genFaultStatsRecordErr.serviceIDs",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_appUnavailableInd_serviceID, {"serviceID","giop-parlay.csapi.appUnavailableInd.serviceID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_svcAvailStatusInd_serviceID, {"serviceID","giop-parlay.csapi.svcAvailStatusInd.serviceID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_svcAvailStatusInd_reason, {"reason","giop-parlay.csapi.svcAvailStatusInd.reason",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpSvcAvailStatusReason),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordRes_faultStatsReqID, {"faultStatsReqID","giop-parlay.csapi.generateFaultStatisticsRecordRes.faultStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordRes_serviceIDs_loop, {"Seq length of serviceIDs","giop-parlay.csapi.generateFaultStatisticsRecordRes.serviceIDs.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordRes_serviceIDs, {"serviceIDs","giop-parlay.csapi.generateFaultStatisticsRecordRes.serviceIDs",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_faultStatsReqID, {"faultStatsReqID","giop-parlay.csapi.generateFaultStatisticsRecordErr.faultStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_faultStatistics_loop, {"Seq length of faultStatistics","giop-parlay.csapi.generateFaultStatisticsRecordErr.faultStatistics.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_faultStatistics, {"faultStatistics","giop-parlay.csapi.generateFaultStatisticsRecordErr.faultStatistics",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpFaultStatisticsError),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_serviceIDs_loop, {"Seq length of serviceIDs","giop-parlay.csapi.generateFaultStatisticsRecordErr.serviceIDs.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordErr_serviceIDs, {"serviceIDs","giop-parlay.csapi.generateFaultStatisticsRecordErr.serviceIDs",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_generateFaultStatisticsRecordReq_faultStatsReqID, {"faultStatsReqID","giop-parlay.csapi.generateFaultStatisticsRecordReq.faultStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppFaultManager_fwAvailStatusInd_reason, {"reason","giop-parlay.csapi.fwAvailStatusInd.reason",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpFwAvailStatusReason),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadRes_loadStatistics_loop, {"Seq length of loadStatistics","giop-parlay.csapi.queryLoadRes.loadStatistics.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadErr_loadStatisticsError, {"loadStatisticsError","giop-parlay.csapi.queryLoadErr.loadStatisticsError",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpLoadStatisticError),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_loadLevelNotification_loadStatistics_loop, {"Seq length of loadStatistics","giop-parlay.csapi.loadLevelNotification.loadStatistics.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryAppLoadStatsReq_loadStatsReqID, {"loadStatsReqID","giop-parlay.csapi.queryAppLoadStatsReq.loadStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsRes_loadStatsReqID, {"loadStatsReqID","giop-parlay.csapi.queryLoadStatsRes.loadStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsRes_loadStatistics_loop, {"Seq length of loadStatistics","giop-parlay.csapi.queryLoadStatsRes.loadStatistics.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsErr_loadStatsReqID, {"loadStatsReqID","giop-parlay.csapi.queryLoadStatsErr.loadStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppLoadManager_queryLoadStatsErr_loadStatisticsError, {"loadStatisticsError","giop-parlay.csapi.queryLoadStatsErr.loadStatisticsError",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpLoadStatisticError),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpLoadManager_reportLoad_loadLevel, {"loadLevel","giop-parlay.csapi.reportLoad.loadLevel",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpLoadLevel),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadReq_serviceIDs_loop, {"Seq length of serviceIDs","giop-parlay.csapi.queryLoadReq.serviceIDs.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadReq_serviceIDs, {"serviceIDs","giop-parlay.csapi.queryLoadReq.serviceIDs",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadRes_loadStatistics_loop, {"Seq length of loadStatistics","giop-parlay.csapi.queryAppLoadRes.loadStatistics.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadErr_loadStatisticsError, {"loadStatisticsError","giop-parlay.csapi.queryAppLoadErr.loadStatisticsError",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpLoadStatisticError),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpLoadManager_createLoadLevelNotification_serviceIDs_loop, {"Seq length of serviceIDs","giop-parlay.csapi.createLoadLevelNotification.serviceIDs.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpLoadManager_createLoadLevelNotification_serviceIDs, {"serviceIDs","giop-parlay.csapi.createLoadLevelNotification.serviceIDs",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpLoadManager_destroyLoadLevelNotification_serviceIDs_loop, {"Seq length of serviceIDs","giop-parlay.csapi.destroyLoadLevelNotification.serviceIDs.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpLoadManager_destroyLoadLevelNotification_serviceIDs, {"serviceIDs","giop-parlay.csapi.destroyLoadLevelNotification.serviceIDs",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpLoadManager_resumeNotification_serviceIDs_loop, {"Seq length of serviceIDs","giop-parlay.csapi.resumeNotification.serviceIDs.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpLoadManager_resumeNotification_serviceIDs, {"serviceIDs","giop-parlay.csapi.resumeNotification.serviceIDs",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpLoadManager_suspendNotification_serviceIDs_loop, {"Seq length of serviceIDs","giop-parlay.csapi.suspendNotification.serviceIDs.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpLoadManager_suspendNotification_serviceIDs, {"serviceIDs","giop-parlay.csapi.suspendNotification.serviceIDs",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadStatsReq_loadStatsReqID, {"loadStatsReqID","giop-parlay.csapi.queryLoadStatsReq.loadStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadStatsReq_serviceIDs_loop, {"Seq length of serviceIDs","giop-parlay.csapi.queryLoadStatsReq.serviceIDs.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryLoadStatsReq_serviceIDs, {"serviceIDs","giop-parlay.csapi.queryLoadStatsReq.serviceIDs",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsRes_loadStatsReqID, {"loadStatsReqID","giop-parlay.csapi.queryAppLoadStatsRes.loadStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsRes_loadStatistics_loop, {"Seq length of loadStatistics","giop-parlay.csapi.queryAppLoadStatsRes.loadStatistics.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsErr_loadStatsReqID, {"loadStatsReqID","giop-parlay.csapi.queryAppLoadStatsErr.loadStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpLoadManager_queryAppLoadStatsErr_loadStatisticsError, {"loadStatisticsError","giop-parlay.csapi.queryAppLoadStatsErr.loadStatisticsError",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpLoadStatisticError),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppOAM_systemDateTimeQuery_return, {"TpDateAndTime","giop-parlay.csapi.systemDateTimeQuery.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppOAM_systemDateTimeQuery_systemDateAndTime, {"systemDateAndTime","giop-parlay.csapi.systemDateTimeQuery.systemDateAndTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpOAM_systemDateTimeQuery_return, {"TpDateAndTime","giop-parlay.csapi.systemDateTimeQuery.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpOAM_systemDateTimeQuery_clientDateAndTime, {"clientDateAndTime","giop-parlay.csapi.systemDateTimeQuery.clientDateAndTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpFaultManager_activityTestReq_activityTestID, {"activityTestID","giop-parlay.csapi.activityTestReq.activityTestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpFaultManager_activityTestReq_svcID, {"svcID","giop-parlay.csapi.activityTestReq.svcID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpFaultManager_appActivityTestRes_activityTestID, {"activityTestID","giop-parlay.csapi.appActivityTestRes.activityTestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpFaultManager_appActivityTestRes_activityTestResult, {"activityTestResult","giop-parlay.csapi.appActivityTestRes.activityTestResult",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpFaultManager_svcUnavailableInd_serviceID, {"serviceID","giop-parlay.csapi.svcUnavailableInd.serviceID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpFaultManager_genFaultStatsRecordReq_serviceIDs_loop, {"Seq length of serviceIDs","giop-parlay.csapi.genFaultStatsRecordReq.serviceIDs.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpFaultManager_genFaultStatsRecordReq_serviceIDs, {"serviceIDs","giop-parlay.csapi.genFaultStatsRecordReq.serviceIDs",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpFaultManager_appActivityTestErr_activityTestID, {"activityTestID","giop-parlay.csapi.appActivityTestErr.activityTestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpFaultManager_appUnavailableInd_serviceID, {"serviceID","giop-parlay.csapi.appUnavailableInd.serviceID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpFaultManager_genFaultStatsRecordErr_faultStatisticsError, {"faultStatisticsError","giop-parlay.csapi.genFaultStatsRecordErr.faultStatisticsError",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpFaultStatisticsError),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpFaultManager_appAvailStatusInd_reason, {"reason","giop-parlay.csapi.appAvailStatusInd.reason",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpAppAvailStatusReason),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordReq_faultStatsReqID, {"faultStatsReqID","giop-parlay.csapi.generateFaultStatisticsRecordReq.faultStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordReq_serviceIDs_loop, {"Seq length of serviceIDs","giop-parlay.csapi.generateFaultStatisticsRecordReq.serviceIDs.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordReq_serviceIDs, {"serviceIDs","giop-parlay.csapi.generateFaultStatisticsRecordReq.serviceIDs",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordRes_faultStatsReqID, {"faultStatsReqID","giop-parlay.csapi.generateFaultStatisticsRecordRes.faultStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordErr_faultStatsReqID, {"faultStatsReqID","giop-parlay.csapi.generateFaultStatisticsRecordErr.faultStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpFaultManager_generateFaultStatisticsRecordErr_faultStatisticsError, {"faultStatisticsError","giop-parlay.csapi.generateFaultStatisticsRecordErr.faultStatisticsError",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpFaultStatisticsError),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpHeartBeatMgmt_enableHeartBeat_interval, {"interval","giop-parlay.csapi.enableHeartBeat.interval",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpHeartBeatMgmt_changeInterval_interval, {"interval","giop-parlay.csapi.changeInterval.interval",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppHeartBeatMgmt_enableAppHeartBeat_interval, {"interval","giop-parlay.csapi.enableAppHeartBeat.interval",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_integrity_IpAppHeartBeatMgmt_changeInterval_interval, {"interval","giop-parlay.csapi.changeInterval.interval",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listServiceTypes_return_loop, {"Seq length of TpServiceTypeNameList","giop-parlay.csapi.listServiceTypes.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listServiceTypes_return, {"TpServiceTypeNameList","giop-parlay.csapi.listServiceTypes.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_describeServiceType_name, {"name","giop-parlay.csapi.describeServiceType.name",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_return_loop, {"Seq length of TpServiceList","giop-parlay.csapi.discoverService.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_serviceTypeName, {"serviceTypeName","giop-parlay.csapi.discoverService.serviceTypeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_desiredPropertyList_loop, {"Seq length of desiredPropertyList","giop-parlay.csapi.discoverService.desiredPropertyList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_discoverService_max, {"max","giop-parlay.csapi.discoverService.max",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_discovery_IpServiceDiscovery_listSubscribedServices_return_loop, {"Seq length of TpServiceList","giop-parlay.csapi.listSubscribedServices.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement_return_loop, {"Seq length of TpOctetSet","giop-parlay.csapi.signServiceAgreement.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement_return, {"TpOctetSet","giop-parlay.csapi.signServiceAgreement.return",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement_serviceToken, {"serviceToken","giop-parlay.csapi.signServiceAgreement.serviceToken",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement_agreementText, {"agreementText","giop-parlay.csapi.signServiceAgreement.agreementText",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_signServiceAgreement_signingAlgorithm, {"signingAlgorithm","giop-parlay.csapi.signServiceAgreement.signingAlgorithm",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_terminateServiceAgreement_serviceToken, {"serviceToken","giop-parlay.csapi.terminateServiceAgreement.serviceToken",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_terminateServiceAgreement_terminationText, {"terminationText","giop-parlay.csapi.terminateServiceAgreement.terminationText",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_terminateServiceAgreement_digitalSignature_loop, {"Seq length of digitalSignature","giop-parlay.csapi.terminateServiceAgreement.digitalSignature.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_service_agreement_IpAppServiceAgreementManagement_terminateServiceAgreement_digitalSignature, {"digitalSignature","giop-parlay.csapi.terminateServiceAgreement.digitalSignature",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_signServiceAgreement_serviceToken, {"serviceToken","giop-parlay.csapi.signServiceAgreement.serviceToken",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_signServiceAgreement_agreementText, {"agreementText","giop-parlay.csapi.signServiceAgreement.agreementText",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_signServiceAgreement_signingAlgorithm, {"signingAlgorithm","giop-parlay.csapi.signServiceAgreement.signingAlgorithm",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_terminateServiceAgreement_serviceToken, {"serviceToken","giop-parlay.csapi.terminateServiceAgreement.serviceToken",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_terminateServiceAgreement_terminationText, {"terminationText","giop-parlay.csapi.terminateServiceAgreement.terminationText",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_terminateServiceAgreement_digitalSignature_loop, {"Seq length of digitalSignature","giop-parlay.csapi.terminateServiceAgreement.digitalSignature.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_terminateServiceAgreement_digitalSignature, {"digitalSignature","giop-parlay.csapi.terminateServiceAgreement.digitalSignature",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_selectService_return, {"TpServiceToken","giop-parlay.csapi.selectService.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_selectService_serviceID, {"serviceID","giop-parlay.csapi.selectService.serviceID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_application_service_agreement_IpServiceAgreementManagement_initiateSignServiceAgreement_serviceToken, {"serviceToken","giop-parlay.csapi.initiateSignServiceAgreement.serviceToken",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_createServiceProfile_return, {"TpServiceProfileID","giop-parlay.csapi.createServiceProfile.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_deleteServiceProfile_serviceProfileID, {"serviceProfileID","giop-parlay.csapi.deleteServiceProfile.serviceProfileID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_assign_sagID, {"sagID","giop-parlay.csapi.assign.sagID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_assign_serviceProfileID, {"serviceProfileID","giop-parlay.csapi.assign.serviceProfileID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_deassign_sagID, {"sagID","giop-parlay.csapi.deassign.sagID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_deassign_serviceProfileID, {"serviceProfileID","giop-parlay.csapi.deassign.serviceProfileID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileManagement_requestConflictInfo_return_loop, {"Seq length of TpAssignSagToServiceProfileConflictList","giop-parlay.csapi.requestConflictInfo.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listServiceProfiles_return_loop, {"Seq length of TpServiceProfileIDList","giop-parlay.csapi.listServiceProfiles.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listServiceProfiles_return, {"TpServiceProfileIDList","giop-parlay.csapi.listServiceProfiles.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_describeServiceProfile_serviceProfileID, {"serviceProfileID","giop-parlay.csapi.describeServiceProfile.serviceProfileID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listAssignedMembers_return_loop, {"Seq length of TpSagIDList","giop-parlay.csapi.listAssignedMembers.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listAssignedMembers_return, {"TpSagIDList","giop-parlay.csapi.listAssignedMembers.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceProfileInfoQuery_listAssignedMembers_serviceProfileID, {"serviceProfileID","giop-parlay.csapi.listAssignedMembers.serviceProfileID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractManagement_createServiceContract_return, {"TpServiceContractID","giop-parlay.csapi.createServiceContract.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractManagement_deleteServiceContract_serviceContractID, {"serviceContractID","giop-parlay.csapi.deleteServiceContract.serviceContractID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_describeServiceContract_serviceContractID, {"serviceContractID","giop-parlay.csapi.describeServiceContract.serviceContractID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceContracts_return_loop, {"Seq length of TpServiceContractIDList","giop-parlay.csapi.listServiceContracts.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceContracts_return, {"TpServiceContractIDList","giop-parlay.csapi.listServiceContracts.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceProfiles_return_loop, {"Seq length of TpServiceProfileIDList","giop-parlay.csapi.listServiceProfiles.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceProfiles_return, {"TpServiceProfileIDList","giop-parlay.csapi.listServiceProfiles.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpServiceContractInfoQuery_listServiceProfiles_serviceContractID, {"serviceContractID","giop-parlay.csapi.listServiceProfiles.serviceContractID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpEntOpAccountManagement_modifyEntOpAccount_enterpriseOperatorProperties_loop, {"Seq length of enterpriseOperatorProperties","giop-parlay.csapi.modifyEntOpAccount.enterpriseOperatorProperties.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_deleteClientApp_clientAppID, {"clientAppID","giop-parlay.csapi.deleteClientApp.clientAppID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_createSAG_clientAppIDs_loop, {"Seq length of clientAppIDs","giop-parlay.csapi.createSAG.clientAppIDs.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_createSAG_clientAppIDs, {"clientAppIDs","giop-parlay.csapi.createSAG.clientAppIDs",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_deleteSAG_sagID, {"sagID","giop-parlay.csapi.deleteSAG.sagID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_addSAGMembers_sagID, {"sagID","giop-parlay.csapi.addSAGMembers.sagID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_addSAGMembers_clientAppIDs_loop, {"Seq length of clientAppIDs","giop-parlay.csapi.addSAGMembers.clientAppIDs.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_addSAGMembers_clientAppIDs, {"clientAppIDs","giop-parlay.csapi.addSAGMembers.clientAppIDs",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_removeSAGMembers_sagID, {"sagID","giop-parlay.csapi.removeSAGMembers.sagID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_removeSAGMembers_clientAppIDList_loop, {"Seq length of clientAppIDList","giop-parlay.csapi.removeSAGMembers.clientAppIDList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_removeSAGMembers_clientAppIDList, {"clientAppIDList","giop-parlay.csapi.removeSAGMembers.clientAppIDList",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppManagement_requestConflictInfo_return_loop, {"Seq length of TpAddSagMembersConflictList","giop-parlay.csapi.requestConflictInfo.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_describeClientApp_clientAppID, {"clientAppID","giop-parlay.csapi.describeClientApp.clientAppID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientApps_return_loop, {"Seq length of TpClientAppIDList","giop-parlay.csapi.listClientApps.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientApps_return, {"TpClientAppIDList","giop-parlay.csapi.listClientApps.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_describeSAG_return, {"TpSagDescription","giop-parlay.csapi.describeSAG.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_describeSAG_sagID, {"sagID","giop-parlay.csapi.describeSAG.sagID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGs_return_loop, {"Seq length of TpSagIDList","giop-parlay.csapi.listSAGs.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGs_return, {"TpSagIDList","giop-parlay.csapi.listSAGs.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGMembers_return_loop, {"Seq length of TpClientAppIDList","giop-parlay.csapi.listSAGMembers.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGMembers_return, {"TpClientAppIDList","giop-parlay.csapi.listSAGMembers.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listSAGMembers_sagID, {"sagID","giop-parlay.csapi.listSAGMembers.sagID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientAppMembership_return_loop, {"Seq length of TpSagIDList","giop-parlay.csapi.listClientAppMembership.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientAppMembership_return, {"TpSagIDList","giop-parlay.csapi.listClientAppMembership.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_service_subscription_IpClientAppInfoQuery_listClientAppMembership_clientAppID, {"clientAppID","giop-parlay.csapi.listClientAppMembership.clientAppID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_notification_IpClientEventNotification_reportNotification_assignmentID, {"assignmentID","giop-parlay.csapi.reportNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_notification_IpEventNotification_createNotification_return, {"TpAssignmentID","giop-parlay.csapi.createNotification.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_enterprise_operator_notification_IpEventNotification_destroyNotification_assignmentID, {"assignmentID","giop-parlay.csapi.destroyNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listServiceTypes_return_loop, {"Seq length of TpServiceTypeNameList","giop-parlay.csapi.listServiceTypes.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listServiceTypes_return, {"TpServiceTypeNameList","giop-parlay.csapi.listServiceTypes.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_describeServiceType_name, {"name","giop-parlay.csapi.describeServiceType.name",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_return_loop, {"Seq length of TpServiceList","giop-parlay.csapi.discoverService.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_serviceTypeName, {"serviceTypeName","giop-parlay.csapi.discoverService.serviceTypeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_desiredPropertyList_loop, {"Seq length of desiredPropertyList","giop-parlay.csapi.discoverService.desiredPropertyList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_discoverService_max, {"max","giop-parlay.csapi.discoverService.max",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_discovery_IpFwServiceDiscovery_listRegisteredServices_return_loop, {"Seq length of TpServiceList","giop-parlay.csapi.listRegisteredServices.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_createServiceManager_application, {"application","giop-parlay.csapi.createServiceManager.application",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_createServiceManager_serviceProperties_loop, {"Seq length of serviceProperties","giop-parlay.csapi.createServiceManager.serviceProperties.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_createServiceManager_serviceInstanceID, {"serviceInstanceID","giop-parlay.csapi.createServiceManager.serviceInstanceID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_service_lifecycle_IpServiceInstanceLifecycleManager_destroyServiceManager_serviceInstance, {"serviceInstance","giop-parlay.csapi.destroyServiceManager.serviceInstance",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerService_return, {"TpServiceID","giop-parlay.csapi.registerService.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerService_serviceTypeName, {"serviceTypeName","giop-parlay.csapi.registerService.serviceTypeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerService_servicePropertyList_loop, {"Seq length of servicePropertyList","giop-parlay.csapi.registerService.servicePropertyList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_announceServiceAvailability_serviceID, {"serviceID","giop-parlay.csapi.announceServiceAvailability.serviceID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_unregisterService_serviceID, {"serviceID","giop-parlay.csapi.unregisterService.serviceID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_describeService_serviceID, {"serviceID","giop-parlay.csapi.describeService.serviceID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_unannounceService_serviceID, {"serviceID","giop-parlay.csapi.unannounceService.serviceID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_return, {"TpServiceID","giop-parlay.csapi.registerServiceSubType.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_serviceTypeName, {"serviceTypeName","giop-parlay.csapi.registerServiceSubType.serviceTypeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_servicePropertyList_loop, {"Seq length of servicePropertyList","giop-parlay.csapi.registerServiceSubType.servicePropertyList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_service_registration_IpFwServiceRegistration_registerServiceSubType_extendedServicePropertyList_loop, {"Seq length of extendedServicePropertyList","giop-parlay.csapi.registerServiceSubType.extendedServicePropertyList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadRes_loadStatistics_loop, {"Seq length of loadStatistics","giop-parlay.csapi.queryLoadRes.loadStatistics.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadErr_loadStatisticsError, {"loadStatisticsError","giop-parlay.csapi.queryLoadErr.loadStatisticsError",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpLoadStatisticError),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_loadLevelNotification_loadStatistics_loop, {"Seq length of loadStatistics","giop-parlay.csapi.loadLevelNotification.loadStatistics.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_querySvcLoadStatsReq_loadStatsReqID, {"loadStatsReqID","giop-parlay.csapi.querySvcLoadStatsReq.loadStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsRes_loadStatsReqID, {"loadStatsReqID","giop-parlay.csapi.queryLoadStatsRes.loadStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsRes_loadStatistics_loop, {"Seq length of loadStatistics","giop-parlay.csapi.queryLoadStatsRes.loadStatistics.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsErr_loadStatsReqID, {"loadStatsReqID","giop-parlay.csapi.queryLoadStatsErr.loadStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcLoadManager_queryLoadStatsErr_loadStatisticsError, {"loadStatisticsError","giop-parlay.csapi.queryLoadStatsErr.loadStatisticsError",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpLoadStatisticError),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_reportLoad_loadLevel, {"loadLevel","giop-parlay.csapi.reportLoad.loadLevel",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpLoadLevel),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_queryLoadReq_querySubject, {"querySubject","giop-parlay.csapi.queryLoadReq.querySubject",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpSubjectType),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadRes_loadStatistics_loop, {"Seq length of loadStatistics","giop-parlay.csapi.querySvcLoadRes.loadStatistics.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadErr_loadStatisticError, {"loadStatisticError","giop-parlay.csapi.querySvcLoadErr.loadStatisticError",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpLoadStatisticError),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_createLoadLevelNotification_notificationSubject, {"notificationSubject","giop-parlay.csapi.createLoadLevelNotification.notificationSubject",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpSubjectType),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_destroyLoadLevelNotification_notificationSubject, {"notificationSubject","giop-parlay.csapi.destroyLoadLevelNotification.notificationSubject",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpSubjectType),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_suspendNotification_notificationSubject, {"notificationSubject","giop-parlay.csapi.suspendNotification.notificationSubject",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpSubjectType),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_resumeNotification_notificationSubject, {"notificationSubject","giop-parlay.csapi.resumeNotification.notificationSubject",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpSubjectType),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_queryLoadStatsReq_loadStatsReqID, {"loadStatsReqID","giop-parlay.csapi.queryLoadStatsReq.loadStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_queryLoadStatsReq_querySubject, {"querySubject","giop-parlay.csapi.queryLoadStatsReq.querySubject",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpSubjectType),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsRes_loadStatsReqID, {"loadStatsReqID","giop-parlay.csapi.querySvcLoadStatsRes.loadStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsRes_loadStatistics_loop, {"Seq length of loadStatistics","giop-parlay.csapi.querySvcLoadStatsRes.loadStatistics.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsErr_loadStatsReqID, {"loadStatsReqID","giop-parlay.csapi.querySvcLoadStatsErr.loadStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwLoadManager_querySvcLoadStatsErr_loadStatisticError, {"loadStatisticError","giop-parlay.csapi.querySvcLoadStatsErr.loadStatisticError",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpLoadStatisticError),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_activityTestRes_activityTestID, {"activityTestID","giop-parlay.csapi.activityTestRes.activityTestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_activityTestRes_activityTestResult, {"activityTestResult","giop-parlay.csapi.activityTestRes.activityTestResult",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_svcActivityTestReq_activityTestID, {"activityTestID","giop-parlay.csapi.svcActivityTestReq.activityTestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_fwFaultReportInd_fault, {"fault","giop-parlay.csapi.fwFaultReportInd.fault",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpInterfaceFault),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_fwFaultRecoveryInd_fault, {"fault","giop-parlay.csapi.fwFaultRecoveryInd.fault",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpInterfaceFault),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_fwUnavailableInd_reason, {"reason","giop-parlay.csapi.fwUnavailableInd.reason",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpFwUnavailReason),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordRes_recordSubject, {"recordSubject","giop-parlay.csapi.genFaultStatsRecordRes.recordSubject",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpSubjectType),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_activityTestErr_activityTestID, {"activityTestID","giop-parlay.csapi.activityTestErr.activityTestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordErr_faultStatisticsError, {"faultStatisticsError","giop-parlay.csapi.genFaultStatsRecordErr.faultStatisticsError",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpFaultStatisticsError),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordErr_recordSubject, {"recordSubject","giop-parlay.csapi.genFaultStatsRecordErr.recordSubject",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpSubjectType),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordReq_serviceIDs_loop, {"Seq length of serviceIDs","giop-parlay.csapi.genFaultStatsRecordReq.serviceIDs.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_genFaultStatsRecordReq_serviceIDs, {"serviceIDs","giop-parlay.csapi.genFaultStatsRecordReq.serviceIDs",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_appAvailStatusInd_reason, {"reason","giop-parlay.csapi.appAvailStatusInd.reason",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpAppAvailStatusReason),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordRes_faultStatsReqID, {"faultStatsReqID","giop-parlay.csapi.generateFaultStatisticsRecordRes.faultStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordRes_recordSubject, {"recordSubject","giop-parlay.csapi.generateFaultStatisticsRecordRes.recordSubject",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpSubjectType),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordErr_faultStatsReqID, {"faultStatsReqID","giop-parlay.csapi.generateFaultStatisticsRecordErr.faultStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordErr_faultStatisticsError, {"faultStatisticsError","giop-parlay.csapi.generateFaultStatisticsRecordErr.faultStatisticsError",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpFaultStatisticsError),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordErr_recordSubject, {"recordSubject","giop-parlay.csapi.generateFaultStatisticsRecordErr.recordSubject",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpSubjectType),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_generateFaultStatisticsRecordReq_faultStatsReqID, {"faultStatsReqID","giop-parlay.csapi.generateFaultStatisticsRecordReq.faultStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcFaultManager_fwAvailStatusInd_reason, {"reason","giop-parlay.csapi.fwAvailStatusInd.reason",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpFwAvailStatusReason),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_activityTestReq_activityTestID, {"activityTestID","giop-parlay.csapi.activityTestReq.activityTestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_activityTestReq_testSubject, {"testSubject","giop-parlay.csapi.activityTestReq.testSubject",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpSubjectType),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_svcActivityTestRes_activityTestID, {"activityTestID","giop-parlay.csapi.svcActivityTestRes.activityTestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_svcActivityTestRes_activityTestResult, {"activityTestResult","giop-parlay.csapi.svcActivityTestRes.activityTestResult",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordReq_recordSubject, {"recordSubject","giop-parlay.csapi.genFaultStatsRecordReq.recordSubject",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpSubjectType),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_svcUnavailableInd_reason, {"reason","giop-parlay.csapi.svcUnavailableInd.reason",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpSvcUnavailReason),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_svcActivityTestErr_activityTestID, {"activityTestID","giop-parlay.csapi.svcActivityTestErr.activityTestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordRes_serviceIDs_loop, {"Seq length of serviceIDs","giop-parlay.csapi.genFaultStatsRecordRes.serviceIDs.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordRes_serviceIDs, {"serviceIDs","giop-parlay.csapi.genFaultStatsRecordRes.serviceIDs",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordErr_faultStatisticsError, {"faultStatisticsError","giop-parlay.csapi.genFaultStatsRecordErr.faultStatisticsError",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpFaultStatisticsError),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordErr_serviceIDs_loop, {"Seq length of serviceIDs","giop-parlay.csapi.genFaultStatsRecordErr.serviceIDs.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_genFaultStatsRecordErr_serviceIDs, {"serviceIDs","giop-parlay.csapi.genFaultStatsRecordErr.serviceIDs",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatsRecordErr_faultStatisticsError, {"faultStatisticsError","giop-parlay.csapi.generateFaultStatsRecordErr.faultStatisticsError",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpFaultStatisticsError),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_svcAvailStatusInd_reason, {"reason","giop-parlay.csapi.svcAvailStatusInd.reason",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpSvcAvailStatusReason),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatisticsRecordReq_faultStatsReqID, {"faultStatsReqID","giop-parlay.csapi.generateFaultStatisticsRecordReq.faultStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatisticsRecordReq_recordSubject, {"recordSubject","giop-parlay.csapi.generateFaultStatisticsRecordReq.recordSubject",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpSubjectType),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatisticsRecordRes_faultStatsReqID, {"faultStatsReqID","giop-parlay.csapi.generateFaultStatisticsRecordRes.faultStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatisticsRecordErr_faultStatsReqID, {"faultStatsReqID","giop-parlay.csapi.generateFaultStatisticsRecordErr.faultStatsReqID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwFaultManager_generateFaultStatisticsRecordErr_faultStatisticsError, {"faultStatisticsError","giop-parlay.csapi.generateFaultStatisticsRecordErr.faultStatisticsError",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpFaultStatisticsError),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcOAM_systemDateTimeQuery_return, {"TpDateAndTime","giop-parlay.csapi.systemDateTimeQuery.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcOAM_systemDateTimeQuery_systemDateAndTime, {"systemDateAndTime","giop-parlay.csapi.systemDateTimeQuery.systemDateAndTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwOAM_systemDateTimeQuery_return, {"TpDateAndTime","giop-parlay.csapi.systemDateTimeQuery.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwOAM_systemDateTimeQuery_clientDateAndTime, {"clientDateAndTime","giop-parlay.csapi.systemDateTimeQuery.clientDateAndTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwHeartBeatMgmt_enableHeartBeat_interval, {"interval","giop-parlay.csapi.enableHeartBeat.interval",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpFwHeartBeatMgmt_changeInterval_interval, {"interval","giop-parlay.csapi.changeInterval.interval",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcHeartBeatMgmt_enableSvcHeartBeat_interval, {"interval","giop-parlay.csapi.enableSvcHeartBeat.interval",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_integrity_IpSvcHeartBeatMgmt_changeInterval_interval, {"interval","giop-parlay.csapi.changeInterval.interval",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_notification_IpFwEventNotification_createNotification_return, {"TpAssignmentID","giop-parlay.csapi.createNotification.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_notification_IpFwEventNotification_destroyNotification_assignmentID, {"assignmentID","giop-parlay.csapi.destroyNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_fw_service_notification_IpSvcEventNotification_reportNotification_assignmentID, {"assignmentID","giop-parlay.csapi.reportNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpAppCall_routeRes_callSessionID, {"callSessionID","giop-parlay.csapi.routeRes.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpAppCall_routeRes_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.routeRes.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpAppCall_routeErr_callSessionID, {"callSessionID","giop-parlay.csapi.routeErr.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpAppCall_routeErr_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.routeErr.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpAppCall_getCallInfoRes_callSessionID, {"callSessionID","giop-parlay.csapi.getCallInfoRes.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpAppCall_getCallInfoErr_callSessionID, {"callSessionID","giop-parlay.csapi.getCallInfoErr.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpAppCall_superviseCallRes_callSessionID, {"callSessionID","giop-parlay.csapi.superviseCallRes.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpAppCall_superviseCallRes_report, {"report","giop-parlay.csapi.superviseCallRes.report",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpAppCall_superviseCallRes_usedTime, {"usedTime","giop-parlay.csapi.superviseCallRes.usedTime",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpAppCall_superviseCallErr_callSessionID, {"callSessionID","giop-parlay.csapi.superviseCallErr.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpAppCall_callFaultDetected_callSessionID, {"callSessionID","giop-parlay.csapi.callFaultDetected.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpAppCall_callFaultDetected_fault, {"fault","giop-parlay.csapi.callFaultDetected.fault",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_gccs_TpCallFault),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpAppCall_getMoreDialledDigitsRes_callSessionID, {"callSessionID","giop-parlay.csapi.getMoreDialledDigitsRes.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpAppCall_getMoreDialledDigitsRes_digits, {"digits","giop-parlay.csapi.getMoreDialledDigitsRes.digits",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpAppCall_getMoreDialledDigitsErr_callSessionID, {"callSessionID","giop-parlay.csapi.getMoreDialledDigitsErr.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpAppCall_callEnded_callSessionID, {"callSessionID","giop-parlay.csapi.callEnded.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpCall_routeReq_return, {"TpSessionID","giop-parlay.csapi.routeReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpCall_routeReq_callSessionID, {"callSessionID","giop-parlay.csapi.routeReq.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpCall_routeReq_responseRequested_loop, {"Seq length of responseRequested","giop-parlay.csapi.routeReq.responseRequested.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpCall_routeReq_appInfo_loop, {"Seq length of appInfo","giop-parlay.csapi.routeReq.appInfo.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpCall_release_callSessionID, {"callSessionID","giop-parlay.csapi.release.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpCall_deassignCall_callSessionID, {"callSessionID","giop-parlay.csapi.deassignCall.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpCall_getCallInfoReq_callSessionID, {"callSessionID","giop-parlay.csapi.getCallInfoReq.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpCall_getCallInfoReq_callInfoRequested, {"callInfoRequested","giop-parlay.csapi.getCallInfoReq.callInfoRequested",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpCall_setCallChargePlan_callSessionID, {"callSessionID","giop-parlay.csapi.setCallChargePlan.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpCall_setAdviceOfCharge_callSessionID, {"callSessionID","giop-parlay.csapi.setAdviceOfCharge.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpCall_setAdviceOfCharge_tariffSwitch, {"tariffSwitch","giop-parlay.csapi.setAdviceOfCharge.tariffSwitch",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpCall_getMoreDialledDigitsReq_callSessionID, {"callSessionID","giop-parlay.csapi.getMoreDialledDigitsReq.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpCall_getMoreDialledDigitsReq_length, {"length","giop-parlay.csapi.getMoreDialledDigitsReq.length",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpCall_superviseCallReq_callSessionID, {"callSessionID","giop-parlay.csapi.superviseCallReq.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpCall_superviseCallReq_time, {"time","giop-parlay.csapi.superviseCallReq.time",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpCall_superviseCallReq_treatment, {"treatment","giop-parlay.csapi.superviseCallReq.treatment",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpCall_continueProcessing_callSessionID, {"callSessionID","giop-parlay.csapi.continueProcessing.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpAppCallControlManager_callAborted_callReference, {"callReference","giop-parlay.csapi.callAborted.callReference",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpAppCallControlManager_callEventNotify_assignmentID, {"assignmentID","giop-parlay.csapi.callEventNotify.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpAppCallControlManager_callOverloadEncountered_assignmentID, {"assignmentID","giop-parlay.csapi.callOverloadEncountered.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpAppCallControlManager_callOverloadCeased_assignmentID, {"assignmentID","giop-parlay.csapi.callOverloadCeased.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpAppCallControlManager_abortMultipleCalls_callReferenceSet_loop, {"Seq length of callReferenceSet","giop-parlay.csapi.abortMultipleCalls.callReferenceSet.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpAppCallControlManager_abortMultipleCalls_callReferenceSet, {"callReferenceSet","giop-parlay.csapi.abortMultipleCalls.callReferenceSet",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpCallControlManager_enableCallNotification_return, {"TpAssignmentID","giop-parlay.csapi.enableCallNotification.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpCallControlManager_disableCallNotification_assignmentID, {"assignmentID","giop-parlay.csapi.disableCallNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpCallControlManager_setCallLoadControl_return, {"TpAssignmentID","giop-parlay.csapi.setCallLoadControl.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpCallControlManager_setCallLoadControl_duration, {"duration","giop-parlay.csapi.setCallLoadControl.duration",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpCallControlManager_changeCallNotification_assignmentID, {"assignmentID","giop-parlay.csapi.changeCallNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_IpCallControlManager_getCriteria_return_loop, {"Seq length of TpCallEventCriteriaResultSet","giop-parlay.csapi.getCriteria.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppCallLeg_eventReportRes_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.eventReportRes.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppCallLeg_eventReportErr_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.eventReportErr.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppCallLeg_attachMediaRes_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.attachMediaRes.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppCallLeg_attachMediaErr_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.attachMediaErr.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppCallLeg_detachMediaRes_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.detachMediaRes.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppCallLeg_detachMediaErr_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.detachMediaErr.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppCallLeg_getInfoRes_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.getInfoRes.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppCallLeg_getInfoErr_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.getInfoErr.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppCallLeg_routeErr_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.routeErr.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppCallLeg_superviseRes_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.superviseRes.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppCallLeg_superviseRes_report, {"report","giop-parlay.csapi.superviseRes.report",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppCallLeg_superviseRes_usedTime, {"usedTime","giop-parlay.csapi.superviseRes.usedTime",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppCallLeg_superviseErr_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.superviseErr.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppCallLeg_callLegEnded_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.callLegEnded.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppCallLeg_callLegEnded_cause, {"cause","giop-parlay.csapi.callLegEnded.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpReleaseCause),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_createNotification_return, {"TpAssignmentID","giop-parlay.csapi.createNotification.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_destroyNotification_assignmentID, {"assignmentID","giop-parlay.csapi.destroyNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_changeNotification_assignmentID, {"assignmentID","giop-parlay.csapi.changeNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_getNotification_return_loop, {"Seq length of TpNotificationRequestedSet","giop-parlay.csapi.getNotification.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_setCallLoadControl_return, {"TpAssignmentID","giop-parlay.csapi.setCallLoadControl.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_setCallLoadControl_duration, {"duration","giop-parlay.csapi.setCallLoadControl.duration",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_enableNotifications_return, {"TpAssignmentID","giop-parlay.csapi.enableNotifications.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCallControlManager_getNextNotification_reset, {"reset","giop-parlay.csapi.getNextNotification.reset",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_routeReq_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.routeReq.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_routeReq_appInfo_loop, {"Seq length of appInfo","giop-parlay.csapi.routeReq.appInfo.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_eventReportReq_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.eventReportReq.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_eventReportReq_eventsRequested_loop, {"Seq length of eventsRequested","giop-parlay.csapi.eventReportReq.eventsRequested.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_release_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.release.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_release_cause, {"cause","giop-parlay.csapi.release.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpReleaseCause),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_getInfoReq_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.getInfoReq.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_getInfoReq_callLegInfoRequested, {"callLegInfoRequested","giop-parlay.csapi.getInfoReq.callLegInfoRequested",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_getCall_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.getCall.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_attachMediaReq_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.attachMediaReq.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_detachMediaReq_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.detachMediaReq.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_getCurrentDestinationAddress_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.getCurrentDestinationAddress.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_continueProcessing_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.continueProcessing.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_setChargePlan_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.setChargePlan.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_setAdviceOfCharge_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.setAdviceOfCharge.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_setAdviceOfCharge_tariffSwitch, {"tariffSwitch","giop-parlay.csapi.setAdviceOfCharge.tariffSwitch",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_superviseReq_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.superviseReq.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_superviseReq_time, {"time","giop-parlay.csapi.superviseReq.time",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_superviseReq_treatment, {"treatment","giop-parlay.csapi.superviseReq.treatment",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_deassign_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.deassign.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_getProperties_return_loop, {"Seq length of TpCallLegPropertyList","giop-parlay.csapi.getProperties.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_getProperties_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.getProperties.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_getProperties_propertyNames_loop, {"Seq length of propertyNames","giop-parlay.csapi.getProperties.propertyNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_getProperties_propertyNames, {"propertyNames","giop-parlay.csapi.getProperties.propertyNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_setProperties_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.setProperties.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpCallLeg_setProperties_properties_loop, {"Seq length of properties","giop-parlay.csapi.setProperties.properties.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_getInfoRes_callSessionID, {"callSessionID","giop-parlay.csapi.getInfoRes.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_getInfoErr_callSessionID, {"callSessionID","giop-parlay.csapi.getInfoErr.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_superviseRes_callSessionID, {"callSessionID","giop-parlay.csapi.superviseRes.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_superviseRes_report, {"report","giop-parlay.csapi.superviseRes.report",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_superviseRes_usedTime, {"usedTime","giop-parlay.csapi.superviseRes.usedTime",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_superviseErr_callSessionID, {"callSessionID","giop-parlay.csapi.superviseErr.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_callEnded_callSessionID, {"callSessionID","giop-parlay.csapi.callEnded.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppMultiPartyCall_createAndRouteCallLegErr_callSessionID, {"callSessionID","giop-parlay.csapi.createAndRouteCallLegErr.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCall_getCallLegs_return_loop, {"Seq length of TpCallLegIdentifierSet","giop-parlay.csapi.getCallLegs.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCall_getCallLegs_callSessionID, {"callSessionID","giop-parlay.csapi.getCallLegs.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCall_createCallLeg_callSessionID, {"callSessionID","giop-parlay.csapi.createCallLeg.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_callSessionID, {"callSessionID","giop-parlay.csapi.createAndRouteCallLegReq.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_eventsRequested_loop, {"Seq length of eventsRequested","giop-parlay.csapi.createAndRouteCallLegReq.eventsRequested.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCall_createAndRouteCallLegReq_appInfo_loop, {"Seq length of appInfo","giop-parlay.csapi.createAndRouteCallLegReq.appInfo.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCall_release_callSessionID, {"callSessionID","giop-parlay.csapi.release.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCall_release_cause, {"cause","giop-parlay.csapi.release.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpReleaseCause),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCall_deassignCall_callSessionID, {"callSessionID","giop-parlay.csapi.deassignCall.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCall_getInfoReq_callSessionID, {"callSessionID","giop-parlay.csapi.getInfoReq.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCall_getInfoReq_callInfoRequested, {"callInfoRequested","giop-parlay.csapi.getInfoReq.callInfoRequested",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCall_setChargePlan_callSessionID, {"callSessionID","giop-parlay.csapi.setChargePlan.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCall_setAdviceOfCharge_callSessionID, {"callSessionID","giop-parlay.csapi.setAdviceOfCharge.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCall_setAdviceOfCharge_tariffSwitch, {"tariffSwitch","giop-parlay.csapi.setAdviceOfCharge.tariffSwitch",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCall_superviseReq_callSessionID, {"callSessionID","giop-parlay.csapi.superviseReq.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCall_superviseReq_time, {"time","giop-parlay.csapi.superviseReq.time",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpMultiPartyCall_superviseReq_treatment, {"treatment","giop-parlay.csapi.superviseReq.treatment",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_reportNotification_callLegReferenceSet_loop, {"Seq length of callLegReferenceSet","giop-parlay.csapi.reportNotification.callLegReferenceSet.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_reportNotification_assignmentID, {"assignmentID","giop-parlay.csapi.reportNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_callAborted_callReference, {"callReference","giop-parlay.csapi.callAborted.callReference",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_callOverloadEncountered_assignmentID, {"assignmentID","giop-parlay.csapi.callOverloadEncountered.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_callOverloadCeased_assignmentID, {"assignmentID","giop-parlay.csapi.callOverloadCeased.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_abortMultipleCalls_callReferenceSet_loop, {"Seq length of callReferenceSet","giop-parlay.csapi.abortMultipleCalls.callReferenceSet.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_IpAppMultiPartyCallControlManager_abortMultipleCalls_callReferenceSet, {"callReferenceSet","giop-parlay.csapi.abortMultipleCalls.callReferenceSet",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpAppMultiMediaCall_superviseVolumeRes_callSessionID, {"callSessionID","giop-parlay.csapi.superviseVolumeRes.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpAppMultiMediaCall_superviseVolumeRes_report, {"report","giop-parlay.csapi.superviseVolumeRes.report",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpAppMultiMediaCall_superviseVolumeRes_qualityOfService, {"qualityOfService","giop-parlay.csapi.superviseVolumeRes.qualityOfService",FT_UINT32,BASE_DEC,VALS(org_csapi_TpDataSessionQosClass),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpAppMultiMediaCall_superviseVolumeErr_callSessionID, {"callSessionID","giop-parlay.csapi.superviseVolumeErr.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpAppMultiMediaCallLeg_mediaStreamMonitorRes_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.mediaStreamMonitorRes.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpAppMultiMediaCallLeg_mediaStreamMonitorRes_streams_loop, {"Seq length of streams","giop-parlay.csapi.mediaStreamMonitorRes.streams.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpAppMultiMediaCallLeg_mediaStreamMonitorRes_type, {"type","giop-parlay.csapi.mediaStreamMonitorRes.type",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_mmccs_TpMediaStreamEventType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpMultiMediaStream_subtract_mediaStreamSessionID, {"mediaStreamSessionID","giop-parlay.csapi.subtract.mediaStreamSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamAllow_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.mediaStreamAllow.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamAllow_mediaStreamList_loop, {"Seq length of mediaStreamList","giop-parlay.csapi.mediaStreamAllow.mediaStreamList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamAllow_mediaStreamList, {"mediaStreamList","giop-parlay.csapi.mediaStreamAllow.mediaStreamList",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamMonitorReq_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.mediaStreamMonitorReq.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpMultiMediaCallLeg_mediaStreamMonitorReq_mediaStreamEventCriteria_loop, {"Seq length of mediaStreamEventCriteria","giop-parlay.csapi.mediaStreamMonitorReq.mediaStreamEventCriteria.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpMultiMediaCallLeg_getMediaStreams_return_loop, {"Seq length of TpMediaStreamSet","giop-parlay.csapi.getMediaStreams.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpMultiMediaCallLeg_getMediaStreams_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.getMediaStreams.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpMultiMediaCall_superviseVolumeReq_callSessionID, {"callSessionID","giop-parlay.csapi.superviseVolumeReq.callSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpMultiMediaCall_superviseVolumeReq_treatment, {"treatment","giop-parlay.csapi.superviseVolumeReq.treatment",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_callLegReferenceSet_loop, {"Seq length of callLegReferenceSet","giop-parlay.csapi.reportMediaNotification.callLegReferenceSet.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_mediaStreams_loop, {"Seq length of mediaStreams","giop-parlay.csapi.reportMediaNotification.mediaStreams.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_type, {"type","giop-parlay.csapi.reportMediaNotification.type",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_mmccs_TpMediaStreamEventType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_qualityOfService, {"qualityOfService","giop-parlay.csapi.reportMediaNotification.qualityOfService",FT_UINT32,BASE_DEC,VALS(org_csapi_TpDataSessionQosClass),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpAppMultiMediaCallControlManager_reportMediaNotification_assignmentID, {"assignmentID","giop-parlay.csapi.reportMediaNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_createMediaNotification_return, {"TpAssignmentID","giop-parlay.csapi.createMediaNotification.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_destroyMediaNotification_assignmentID, {"assignmentID","giop-parlay.csapi.destroyMediaNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_changeMediaNotification_assignmentID, {"assignmentID","giop-parlay.csapi.changeMediaNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_IpMultiMediaCallControlManager_getMediaNotification_return_loop, {"Seq length of TpMediaNotificationRequestedSet","giop-parlay.csapi.getMediaNotification.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpAppSubConfCall_chairSelection_subConferenceSessionID, {"subConferenceSessionID","giop-parlay.csapi.chairSelection.subConferenceSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpAppSubConfCall_chairSelection_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.chairSelection.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpAppSubConfCall_floorRequest_subConferenceSessionID, {"subConferenceSessionID","giop-parlay.csapi.floorRequest.subConferenceSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpAppSubConfCall_floorRequest_callLegSessionID, {"callLegSessionID","giop-parlay.csapi.floorRequest.callLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpAppConfCall_partyJoined_conferenceSessionID, {"conferenceSessionID","giop-parlay.csapi.partyJoined.conferenceSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpAppConfCall_leaveMonitorRes_conferenceSessionID, {"conferenceSessionID","giop-parlay.csapi.leaveMonitorRes.conferenceSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpAppConfCall_leaveMonitorRes_callLeg, {"callLeg","giop-parlay.csapi.leaveMonitorRes.callLeg",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpConfCallControlManager_createConference_numberOfSubConferences, {"numberOfSubConferences","giop-parlay.csapi.createConference.numberOfSubConferences",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpConfCallControlManager_createConference_numberOfParticipants, {"numberOfParticipants","giop-parlay.csapi.createConference.numberOfParticipants",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpConfCallControlManager_createConference_duration, {"duration","giop-parlay.csapi.createConference.duration",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpConfCallControlManager_reserveResources_startTime, {"startTime","giop-parlay.csapi.reserveResources.startTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpConfCallControlManager_reserveResources_numberOfParticipants, {"numberOfParticipants","giop-parlay.csapi.reserveResources.numberOfParticipants",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpConfCallControlManager_reserveResources_duration, {"duration","giop-parlay.csapi.reserveResources.duration",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpSubConfCall_splitSubConference_subConferenceSessionID, {"subConferenceSessionID","giop-parlay.csapi.splitSubConference.subConferenceSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpSubConfCall_splitSubConference_callLegList_loop, {"Seq length of callLegList","giop-parlay.csapi.splitSubConference.callLegList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpSubConfCall_splitSubConference_callLegList, {"callLegList","giop-parlay.csapi.splitSubConference.callLegList",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpSubConfCall_mergeSubConference_subConferenceCallSessionID, {"subConferenceCallSessionID","giop-parlay.csapi.mergeSubConference.subConferenceCallSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpSubConfCall_mergeSubConference_targetSubConferenceCall, {"targetSubConferenceCall","giop-parlay.csapi.mergeSubConference.targetSubConferenceCall",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpSubConfCall_moveCallLeg_subConferenceCallSessionID, {"subConferenceCallSessionID","giop-parlay.csapi.moveCallLeg.subConferenceCallSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpSubConfCall_moveCallLeg_targetSubConferenceCall, {"targetSubConferenceCall","giop-parlay.csapi.moveCallLeg.targetSubConferenceCall",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpSubConfCall_moveCallLeg_callLeg, {"callLeg","giop-parlay.csapi.moveCallLeg.callLeg",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpSubConfCall_inspectVideo_subConferenceSessionID, {"subConferenceSessionID","giop-parlay.csapi.inspectVideo.subConferenceSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpSubConfCall_inspectVideo_inspectedCallLeg, {"inspectedCallLeg","giop-parlay.csapi.inspectVideo.inspectedCallLeg",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpSubConfCall_inspectVideoCancel_subConferenceSessionID, {"subConferenceSessionID","giop-parlay.csapi.inspectVideoCancel.subConferenceSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpSubConfCall_appointSpeaker_subConferenceSessionID, {"subConferenceSessionID","giop-parlay.csapi.appointSpeaker.subConferenceSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpSubConfCall_appointSpeaker_speakerCallLeg, {"speakerCallLeg","giop-parlay.csapi.appointSpeaker.speakerCallLeg",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpSubConfCall_chairSelection_subConferenceSessionID, {"subConferenceSessionID","giop-parlay.csapi.chairSelection.subConferenceSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpSubConfCall_chairSelection_chairCallLeg, {"chairCallLeg","giop-parlay.csapi.chairSelection.chairCallLeg",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpSubConfCall_changeConferencePolicy_subConferenceSessionID, {"subConferenceSessionID","giop-parlay.csapi.changeConferencePolicy.subConferenceSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpConfCall_getSubConferences_return_loop, {"Seq length of TpSubConfCallIdentifierSet","giop-parlay.csapi.getSubConferences.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpConfCall_getSubConferences_conferenceSessionID, {"conferenceSessionID","giop-parlay.csapi.getSubConferences.conferenceSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpConfCall_createSubConference_conferenceSessionID, {"conferenceSessionID","giop-parlay.csapi.createSubConference.conferenceSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpConfCall_leaveMonitorReq_conferenceSessionID, {"conferenceSessionID","giop-parlay.csapi.leaveMonitorReq.conferenceSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_IpConfCall_getConferenceAddress_conferenceSessionID, {"conferenceSessionID","giop-parlay.csapi.getConferenceAddress.conferenceSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUI_sendInfoRes_userInteractionSessionID, {"userInteractionSessionID","giop-parlay.csapi.sendInfoRes.userInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUI_sendInfoRes_assignmentID, {"assignmentID","giop-parlay.csapi.sendInfoRes.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUI_sendInfoRes_response, {"response","giop-parlay.csapi.sendInfoRes.response",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUIReport),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUI_sendInfoErr_userInteractionSessionID, {"userInteractionSessionID","giop-parlay.csapi.sendInfoErr.userInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUI_sendInfoErr_assignmentID, {"assignmentID","giop-parlay.csapi.sendInfoErr.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUI_sendInfoErr_error, {"error","giop-parlay.csapi.sendInfoErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUIError),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUI_sendInfoAndCollectRes_userInteractionSessionID, {"userInteractionSessionID","giop-parlay.csapi.sendInfoAndCollectRes.userInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUI_sendInfoAndCollectRes_assignmentID, {"assignmentID","giop-parlay.csapi.sendInfoAndCollectRes.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUI_sendInfoAndCollectRes_response, {"response","giop-parlay.csapi.sendInfoAndCollectRes.response",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUIReport),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUI_sendInfoAndCollectRes_collectedInfo, {"collectedInfo","giop-parlay.csapi.sendInfoAndCollectRes.collectedInfo",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUI_sendInfoAndCollectErr_userInteractionSessionID, {"userInteractionSessionID","giop-parlay.csapi.sendInfoAndCollectErr.userInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUI_sendInfoAndCollectErr_assignmentID, {"assignmentID","giop-parlay.csapi.sendInfoAndCollectErr.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUI_sendInfoAndCollectErr_error, {"error","giop-parlay.csapi.sendInfoAndCollectErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUIError),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUI_userInteractionFaultDetected_userInteractionSessionID, {"userInteractionSessionID","giop-parlay.csapi.userInteractionFaultDetected.userInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUI_userInteractionFaultDetected_fault, {"fault","giop-parlay.csapi.userInteractionFaultDetected.fault",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUIFault),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUICall_recordMessageRes_userInteractionSessionID, {"userInteractionSessionID","giop-parlay.csapi.recordMessageRes.userInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUICall_recordMessageRes_assignmentID, {"assignmentID","giop-parlay.csapi.recordMessageRes.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUICall_recordMessageRes_response, {"response","giop-parlay.csapi.recordMessageRes.response",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUIReport),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUICall_recordMessageRes_messageID, {"messageID","giop-parlay.csapi.recordMessageRes.messageID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUICall_recordMessageErr_userInteractionSessionID, {"userInteractionSessionID","giop-parlay.csapi.recordMessageErr.userInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUICall_recordMessageErr_assignmentID, {"assignmentID","giop-parlay.csapi.recordMessageErr.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUICall_recordMessageErr_error, {"error","giop-parlay.csapi.recordMessageErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUIError),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUICall_deleteMessageRes_usrInteractionSessionID, {"usrInteractionSessionID","giop-parlay.csapi.deleteMessageRes.usrInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUICall_deleteMessageRes_response, {"response","giop-parlay.csapi.deleteMessageRes.response",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUIReport),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUICall_deleteMessageRes_assignmentID, {"assignmentID","giop-parlay.csapi.deleteMessageRes.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUICall_deleteMessageErr_usrInteractionSessionID, {"usrInteractionSessionID","giop-parlay.csapi.deleteMessageErr.usrInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUICall_deleteMessageErr_error, {"error","giop-parlay.csapi.deleteMessageErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUIError),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUICall_deleteMessageErr_assignmentID, {"assignmentID","giop-parlay.csapi.deleteMessageErr.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUICall_abortActionRes_userInteractionSessionID, {"userInteractionSessionID","giop-parlay.csapi.abortActionRes.userInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUICall_abortActionRes_assignmentID, {"assignmentID","giop-parlay.csapi.abortActionRes.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUICall_abortActionErr_userInteractionSessionID, {"userInteractionSessionID","giop-parlay.csapi.abortActionErr.userInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUICall_abortActionErr_assignmentID, {"assignmentID","giop-parlay.csapi.abortActionErr.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUICall_abortActionErr_error, {"error","giop-parlay.csapi.abortActionErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUIError),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUICall_getMessageRes_userInteractionSessionID, {"userInteractionSessionID","giop-parlay.csapi.getMessageRes.userInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUICall_getMessageRes_assignmentID, {"assignmentID","giop-parlay.csapi.getMessageRes.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUICall_getMessageErr_userInteractionSessionID, {"userInteractionSessionID","giop-parlay.csapi.getMessageErr.userInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUICall_getMessageErr_assignmentID, {"assignmentID","giop-parlay.csapi.getMessageErr.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUICall_getMessageErr_error, {"error","giop-parlay.csapi.getMessageErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUIError),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIManager_reportEventNotification_assignmentID, {"assignmentID","giop-parlay.csapi.reportEventNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIManager_abortMultipleUserInteractions_userInteractionSet_loop, {"Seq length of userInteractionSet","giop-parlay.csapi.abortMultipleUserInteractions.userInteractionSet.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUI_sendInfoReq_return, {"TpAssignmentID","giop-parlay.csapi.sendInfoReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUI_sendInfoReq_userInteractionSessionID, {"userInteractionSessionID","giop-parlay.csapi.sendInfoReq.userInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUI_sendInfoReq_language, {"language","giop-parlay.csapi.sendInfoReq.language",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUI_sendInfoReq_variableInfo_loop, {"Seq length of variableInfo","giop-parlay.csapi.sendInfoReq.variableInfo.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUI_sendInfoReq_repeatIndicator, {"repeatIndicator","giop-parlay.csapi.sendInfoReq.repeatIndicator",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUI_sendInfoReq_responseRequested, {"responseRequested","giop-parlay.csapi.sendInfoReq.responseRequested",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUI_sendInfoAndCollectReq_return, {"TpAssignmentID","giop-parlay.csapi.sendInfoAndCollectReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUI_sendInfoAndCollectReq_userInteractionSessionID, {"userInteractionSessionID","giop-parlay.csapi.sendInfoAndCollectReq.userInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUI_sendInfoAndCollectReq_language, {"language","giop-parlay.csapi.sendInfoAndCollectReq.language",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUI_sendInfoAndCollectReq_variableInfo_loop, {"Seq length of variableInfo","giop-parlay.csapi.sendInfoAndCollectReq.variableInfo.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUI_sendInfoAndCollectReq_responseRequested, {"responseRequested","giop-parlay.csapi.sendInfoAndCollectReq.responseRequested",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUI_release_userInteractionSessionID, {"userInteractionSessionID","giop-parlay.csapi.release.userInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUI_setOriginatingAddress_userInteractionSessionID, {"userInteractionSessionID","giop-parlay.csapi.setOriginatingAddress.userInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUI_setOriginatingAddress_origin, {"origin","giop-parlay.csapi.setOriginatingAddress.origin",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUI_getOriginatingAddress_return, {"TpString","giop-parlay.csapi.getOriginatingAddress.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUI_getOriginatingAddress_userInteractionSessionID, {"userInteractionSessionID","giop-parlay.csapi.getOriginatingAddress.userInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUICall_recordMessageReq_return, {"TpAssignmentID","giop-parlay.csapi.recordMessageReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUICall_recordMessageReq_userInteractionSessionID, {"userInteractionSessionID","giop-parlay.csapi.recordMessageReq.userInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUICall_deleteMessageReq_return, {"TpAssignmentID","giop-parlay.csapi.deleteMessageReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUICall_deleteMessageReq_usrInteractionSessionID, {"usrInteractionSessionID","giop-parlay.csapi.deleteMessageReq.usrInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUICall_deleteMessageReq_messageID, {"messageID","giop-parlay.csapi.deleteMessageReq.messageID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUICall_abortActionReq_userInteractionSessionID, {"userInteractionSessionID","giop-parlay.csapi.abortActionReq.userInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUICall_abortActionReq_assignmentID, {"assignmentID","giop-parlay.csapi.abortActionReq.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUICall_getMessageReq_return, {"TpAssignmentID","giop-parlay.csapi.getMessageReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUICall_getMessageReq_userInteractionSessionID, {"userInteractionSessionID","giop-parlay.csapi.getMessageReq.userInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUICall_getMessageReq_messageID, {"messageID","giop-parlay.csapi.getMessageReq.messageID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUIManager_createNotification_return, {"TpAssignmentID","giop-parlay.csapi.createNotification.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUIManager_destroyNotification_assignmentID, {"assignmentID","giop-parlay.csapi.destroyNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUIManager_changeNotification_assignmentID, {"assignmentID","giop-parlay.csapi.changeNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUIManager_getNotification_return_loop, {"Seq length of TpUIEventCriteriaResultSet","giop-parlay.csapi.getNotification.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUIManager_enableNotifications_return, {"TpAssignmentID","giop-parlay.csapi.enableNotifications.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_getMessageRes_usrInteractionSessionID, {"usrInteractionSessionID","giop-parlay.csapi.getMessageRes.usrInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_getMessageRes_assignmentID, {"assignmentID","giop-parlay.csapi.getMessageRes.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_getMessageErr_usrInteractionSessionID, {"usrInteractionSessionID","giop-parlay.csapi.getMessageErr.usrInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_getMessageErr_assignmentID, {"assignmentID","giop-parlay.csapi.getMessageErr.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_getMessageErr_error, {"error","giop-parlay.csapi.getMessageErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUIError),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_deleteMessageRes_usrInteractionSessionID, {"usrInteractionSessionID","giop-parlay.csapi.deleteMessageRes.usrInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_deleteMessageRes_response, {"response","giop-parlay.csapi.deleteMessageRes.response",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUIReport),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_deleteMessageRes_assignmentID, {"assignmentID","giop-parlay.csapi.deleteMessageRes.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_deleteMessageErr_usrInteractionSessionID, {"usrInteractionSessionID","giop-parlay.csapi.deleteMessageErr.usrInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_deleteMessageErr_error, {"error","giop-parlay.csapi.deleteMessageErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUIError),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_deleteMessageErr_assignmentID, {"assignmentID","giop-parlay.csapi.deleteMessageErr.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_putMessageRes_usrInteractionSessionID, {"usrInteractionSessionID","giop-parlay.csapi.putMessageRes.usrInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_putMessageRes_assignmentID, {"assignmentID","giop-parlay.csapi.putMessageRes.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_putMessageRes_messageID, {"messageID","giop-parlay.csapi.putMessageRes.messageID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_putMessageErr_usrInteractionSessionID, {"usrInteractionSessionID","giop-parlay.csapi.putMessageErr.usrInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_putMessageErr_assignmentID, {"assignmentID","giop-parlay.csapi.putMessageErr.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_putMessageErr_error, {"error","giop-parlay.csapi.putMessageErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUIError),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_getMessageListRes_usrInteractionSessionID, {"usrInteractionSessionID","giop-parlay.csapi.getMessageListRes.usrInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_getMessageListRes_assignmentID, {"assignmentID","giop-parlay.csapi.getMessageListRes.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_getMessageListRes_messageIDList_loop, {"Seq length of messageIDList","giop-parlay.csapi.getMessageListRes.messageIDList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_getMessageListRes_messageIDList, {"messageIDList","giop-parlay.csapi.getMessageListRes.messageIDList",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_getMessageListRes_final, {"final","giop-parlay.csapi.getMessageListRes.final",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_getMessageListErr_usrInteractionSessionID, {"usrInteractionSessionID","giop-parlay.csapi.getMessageListErr.usrInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_getMessageListErr_assignmentID, {"assignmentID","giop-parlay.csapi.getMessageListErr.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpAppUIAdminManager_getMessageListErr_error, {"error","giop-parlay.csapi.getMessageListErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUIError),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUIAdminManager_getMessageReq_return, {"TpAssignmentID","giop-parlay.csapi.getMessageReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUIAdminManager_getMessageReq_usrInteractionSessionID, {"usrInteractionSessionID","giop-parlay.csapi.getMessageReq.usrInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUIAdminManager_getMessageReq_messageID, {"messageID","giop-parlay.csapi.getMessageReq.messageID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUIAdminManager_putMessageReq_return, {"TpAssignmentID","giop-parlay.csapi.putMessageReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUIAdminManager_putMessageReq_usrInteractionSessionID, {"usrInteractionSessionID","giop-parlay.csapi.putMessageReq.usrInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUIAdminManager_deleteMessageReq_return, {"TpAssignmentID","giop-parlay.csapi.deleteMessageReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUIAdminManager_deleteMessageReq_usrInteractionSessionID, {"usrInteractionSessionID","giop-parlay.csapi.deleteMessageReq.usrInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUIAdminManager_deleteMessageReq_messageID, {"messageID","giop-parlay.csapi.deleteMessageReq.messageID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUIAdminManager_getMessageListReq_return, {"TpAssignmentID","giop-parlay.csapi.getMessageListReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUIAdminManager_getMessageListReq_usrInteractionSessionID, {"usrInteractionSessionID","giop-parlay.csapi.getMessageListReq.usrInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_IpUIAdminManager_getMessageListReq_reset, {"reset","giop-parlay.csapi.getMessageListReq.reset",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpAppUserLocation_locationReportRes_assignmentId, {"assignmentId","giop-parlay.csapi.locationReportRes.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpAppUserLocation_locationReportRes_locations_loop, {"Seq length of locations","giop-parlay.csapi.locationReportRes.locations.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpAppUserLocation_locationReportErr_assignmentId, {"assignmentId","giop-parlay.csapi.locationReportErr.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpAppUserLocation_locationReportErr_cause, {"cause","giop-parlay.csapi.locationReportErr.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpAppUserLocation_locationReportErr_diagnostic, {"diagnostic","giop-parlay.csapi.locationReportErr.diagnostic",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityDiagnostic),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportRes_assignmentId, {"assignmentId","giop-parlay.csapi.extendedLocationReportRes.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportRes_locations_loop, {"Seq length of locations","giop-parlay.csapi.extendedLocationReportRes.locations.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportErr_assignmentId, {"assignmentId","giop-parlay.csapi.extendedLocationReportErr.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportErr_cause, {"cause","giop-parlay.csapi.extendedLocationReportErr.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpAppUserLocation_extendedLocationReportErr_diagnostic, {"diagnostic","giop-parlay.csapi.extendedLocationReportErr.diagnostic",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityDiagnostic),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReport_assignmentId, {"assignmentId","giop-parlay.csapi.periodicLocationReport.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReport_locations_loop, {"Seq length of locations","giop-parlay.csapi.periodicLocationReport.locations.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReportErr_assignmentId, {"assignmentId","giop-parlay.csapi.periodicLocationReportErr.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReportErr_cause, {"cause","giop-parlay.csapi.periodicLocationReportErr.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpAppUserLocation_periodicLocationReportErr_diagnostic, {"diagnostic","giop-parlay.csapi.periodicLocationReportErr.diagnostic",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityDiagnostic),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpUserLocation_locationReportReq_return, {"TpAssignmentID","giop-parlay.csapi.locationReportReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpUserLocation_locationReportReq_users_loop, {"Seq length of users","giop-parlay.csapi.locationReportReq.users.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpUserLocation_extendedLocationReportReq_return, {"TpAssignmentID","giop-parlay.csapi.extendedLocationReportReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpUserLocation_extendedLocationReportReq_users_loop, {"Seq length of users","giop-parlay.csapi.extendedLocationReportReq.users.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpUserLocation_periodicLocationReportingStartReq_return, {"TpAssignmentID","giop-parlay.csapi.periodicLocationReportingStartReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpUserLocation_periodicLocationReportingStartReq_users_loop, {"Seq length of users","giop-parlay.csapi.periodicLocationReportingStartReq.users.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpUserLocation_periodicLocationReportingStartReq_reportingInterval, {"reportingInterval","giop-parlay.csapi.periodicLocationReportingStartReq.reportingInterval",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpUserLocation_getNextPeriodicLocationRequest_reset, {"reset","giop-parlay.csapi.getNextPeriodicLocationRequest.reset",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpAppTriggeredUserLocation_triggeredLocationReport_assignmentId, {"assignmentId","giop-parlay.csapi.triggeredLocationReport.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpAppTriggeredUserLocation_triggeredLocationReport_criterion, {"criterion","giop-parlay.csapi.triggeredLocationReport.criterion",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpLocationTriggerCriteria),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpAppTriggeredUserLocation_triggeredLocationReportErr_assignmentId, {"assignmentId","giop-parlay.csapi.triggeredLocationReportErr.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpAppTriggeredUserLocation_triggeredLocationReportErr_cause, {"cause","giop-parlay.csapi.triggeredLocationReportErr.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpAppTriggeredUserLocation_triggeredLocationReportErr_diagnostic, {"diagnostic","giop-parlay.csapi.triggeredLocationReportErr.diagnostic",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityDiagnostic),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_return, {"TpAssignmentID","giop-parlay.csapi.triggeredLocationReportingStartReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_users_loop, {"Seq length of users","giop-parlay.csapi.triggeredLocationReportingStartReq.users.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpTriggeredUserLocation_triggeredLocationReportingStartReq_triggers_loop, {"Seq length of triggers","giop-parlay.csapi.triggeredLocationReportingStartReq.triggers.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ul_IpTriggeredUserLocation_getNextTriggeredLocationRequest_reset, {"reset","giop-parlay.csapi.getNextTriggeredLocationRequest.reset",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportRes_assignmentId, {"assignmentId","giop-parlay.csapi.locationReportRes.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportRes_locations_loop, {"Seq length of locations","giop-parlay.csapi.locationReportRes.locations.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportErr_assignmentId, {"assignmentId","giop-parlay.csapi.locationReportErr.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportErr_cause, {"cause","giop-parlay.csapi.locationReportErr.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ulc_IpAppUserLocationCamel_locationReportErr_diagnostic, {"diagnostic","giop-parlay.csapi.locationReportErr.diagnostic",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityDiagnostic),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReport_assignmentId, {"assignmentId","giop-parlay.csapi.periodicLocationReport.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReport_locations_loop, {"Seq length of locations","giop-parlay.csapi.periodicLocationReport.locations.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReportErr_assignmentId, {"assignmentId","giop-parlay.csapi.periodicLocationReportErr.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReportErr_cause, {"cause","giop-parlay.csapi.periodicLocationReportErr.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ulc_IpAppUserLocationCamel_periodicLocationReportErr_diagnostic, {"diagnostic","giop-parlay.csapi.periodicLocationReportErr.diagnostic",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityDiagnostic),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ulc_IpAppUserLocationCamel_triggeredLocationReport_assignmentId, {"assignmentId","giop-parlay.csapi.triggeredLocationReport.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ulc_IpAppUserLocationCamel_triggeredLocationReportErr_assignmentId, {"assignmentId","giop-parlay.csapi.triggeredLocationReportErr.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ulc_IpAppUserLocationCamel_triggeredLocationReportErr_cause, {"cause","giop-parlay.csapi.triggeredLocationReportErr.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ulc_IpAppUserLocationCamel_triggeredLocationReportErr_diagnostic, {"diagnostic","giop-parlay.csapi.triggeredLocationReportErr.diagnostic",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityDiagnostic),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ulc_IpUserLocationCamel_locationReportReq_return, {"TpAssignmentID","giop-parlay.csapi.locationReportReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ulc_IpUserLocationCamel_locationReportReq_users_loop, {"Seq length of users","giop-parlay.csapi.locationReportReq.users.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ulc_IpUserLocationCamel_periodicLocationReportingStartReq_return, {"TpAssignmentID","giop-parlay.csapi.periodicLocationReportingStartReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ulc_IpUserLocationCamel_periodicLocationReportingStartReq_users_loop, {"Seq length of users","giop-parlay.csapi.periodicLocationReportingStartReq.users.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ulc_IpUserLocationCamel_periodicLocationReportingStartReq_reportingInterval, {"reportingInterval","giop-parlay.csapi.periodicLocationReportingStartReq.reportingInterval",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ulc_IpUserLocationCamel_triggeredLocationReportingStartReq_return, {"TpAssignmentID","giop-parlay.csapi.triggeredLocationReportingStartReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ulc_IpUserLocationCamel_triggeredLocationReportingStartReq_users_loop, {"Seq length of users","giop-parlay.csapi.triggeredLocationReportingStartReq.users.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ulc_IpUserLocationCamel_getNextPeriodicLocationRequest_reset, {"reset","giop-parlay.csapi.getNextPeriodicLocationRequest.reset",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_ulc_IpUserLocationCamel_getNextTriggeredLocationRequest_reset, {"reset","giop-parlay.csapi.getNextTriggeredLocationRequest.reset",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_ule_IpAppUserLocationEmergency_emergencyLocationReport_assignmentId, {"assignmentId","giop-parlay.csapi.emergencyLocationReport.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ule_IpAppUserLocationEmergency_emergencyLocationReportErr_assignmentId, {"assignmentId","giop-parlay.csapi.emergencyLocationReportErr.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ule_IpAppUserLocationEmergency_emergencyLocationReportErr_cause, {"cause","giop-parlay.csapi.emergencyLocationReportErr.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ule_IpAppUserLocationEmergency_emergencyLocationReportErr_diagnostic, {"diagnostic","giop-parlay.csapi.emergencyLocationReportErr.diagnostic",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityDiagnostic),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ule_IpUserLocationEmergency_emergencyLocationReportReq_return, {"TpAssignmentID","giop-parlay.csapi.emergencyLocationReportReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ule_IpUserLocationEmergency_subscribeEmergencyLocationReports_return, {"TpAssignmentID","giop-parlay.csapi.subscribeEmergencyLocationReports.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ule_IpUserLocationEmergency_unSubscribeEmergencyLocationReports_assignmentId, {"assignmentId","giop-parlay.csapi.unSubscribeEmergencyLocationReports.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpAppUserStatus_statusReportRes_assignmentId, {"assignmentId","giop-parlay.csapi.statusReportRes.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpAppUserStatus_statusReportRes_status_loop, {"Seq length of status","giop-parlay.csapi.statusReportRes.status.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpAppUserStatus_statusReportErr_assignmentId, {"assignmentId","giop-parlay.csapi.statusReportErr.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpAppUserStatus_statusReportErr_cause, {"cause","giop-parlay.csapi.statusReportErr.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpAppUserStatus_statusReportErr_diagnostic, {"diagnostic","giop-parlay.csapi.statusReportErr.diagnostic",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityDiagnostic),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpAppUserStatus_triggeredStatusReport_assignmentId, {"assignmentId","giop-parlay.csapi.triggeredStatusReport.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpAppUserStatus_triggeredStatusReportErr_assignmentId, {"assignmentId","giop-parlay.csapi.triggeredStatusReportErr.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpAppUserStatus_triggeredStatusReportErr_cause, {"cause","giop-parlay.csapi.triggeredStatusReportErr.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpAppUserStatus_triggeredStatusReportErr_diagnostic, {"diagnostic","giop-parlay.csapi.triggeredStatusReportErr.diagnostic",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityDiagnostic),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportRes_assignmentId, {"assignmentId","giop-parlay.csapi.extendedStatusReportRes.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportRes_status_loop, {"Seq length of status","giop-parlay.csapi.extendedStatusReportRes.status.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportErr_assignmentId, {"assignmentId","giop-parlay.csapi.extendedStatusReportErr.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportErr_cause, {"cause","giop-parlay.csapi.extendedStatusReportErr.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpAppUserStatus_extendedStatusReportErr_diagnostic, {"diagnostic","giop-parlay.csapi.extendedStatusReportErr.diagnostic",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityDiagnostic),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpAppUserStatus_extTriggeredStatusReport_assignmentId, {"assignmentId","giop-parlay.csapi.extTriggeredStatusReport.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpAppUserStatus_extTriggeredStatusReportErr_assignmentId, {"assignmentId","giop-parlay.csapi.extTriggeredStatusReportErr.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpAppUserStatus_extTriggeredStatusReportErr_cause, {"cause","giop-parlay.csapi.extTriggeredStatusReportErr.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpAppUserStatus_extTriggeredStatusReportErr_diagnostic, {"diagnostic","giop-parlay.csapi.extTriggeredStatusReportErr.diagnostic",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityDiagnostic),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpUserStatus_statusReportReq_return, {"TpAssignmentID","giop-parlay.csapi.statusReportReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpUserStatus_statusReportReq_users_loop, {"Seq length of users","giop-parlay.csapi.statusReportReq.users.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpUserStatus_triggeredStatusReportingStartReq_return, {"TpAssignmentID","giop-parlay.csapi.triggeredStatusReportingStartReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpUserStatus_triggeredStatusReportingStartReq_users_loop, {"Seq length of users","giop-parlay.csapi.triggeredStatusReportingStartReq.users.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpUserStatus_getNextTriggeredStatusRequest_reset, {"reset","giop-parlay.csapi.getNextTriggeredStatusRequest.reset",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpUserStatus_extendedStatusReportReq_return, {"TpAssignmentID","giop-parlay.csapi.extendedStatusReportReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpUserStatus_extendedStatusReportReq_users_loop, {"Seq length of users","giop-parlay.csapi.extendedStatusReportReq.users.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpUserStatus_extTriggeredStatusReportingStartReq_return, {"TpAssignmentID","giop-parlay.csapi.extTriggeredStatusReportingStartReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_us_IpUserStatus_extTriggeredStatusReportingStartReq_users_loop, {"Seq length of users","giop-parlay.csapi.extTriggeredStatusReportingStartReq.users.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_return_loop, {"Seq length of TpBindingSet","giop-parlay.csapi.triggeredBindingRequestNotification.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_assignmentId, {"assignmentId","giop-parlay.csapi.triggeredBindingRequestNotification.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotification_contactInformation_loop, {"Seq length of contactInformation","giop-parlay.csapi.triggeredBindingRequestNotification.contactInformation.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotificationStartErr_assignmentId, {"assignmentId","giop-parlay.csapi.triggeredBindingRequestNotificationStartErr.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotificationStartErr_cause, {"cause","giop-parlay.csapi.triggeredBindingRequestNotificationStartErr.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ub_IpAppUserBinding_triggeredBindingRequestNotificationStartErr_diagnostic, {"diagnostic","giop-parlay.csapi.triggeredBindingRequestNotificationStartErr.diagnostic",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityDiagnostic),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_return, {"TpAssignmentID","giop-parlay.csapi.triggeredBindingRequestNotificationStartReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_users_loop, {"Seq length of users","giop-parlay.csapi.triggeredBindingRequestNotificationStartReq.users.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStartReq_criteria_loop, {"Seq length of criteria","giop-parlay.csapi.triggeredBindingRequestNotificationStartReq.criteria.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_ub_IpUserBinding_triggeredBindingRequestNotificationStop_assignmentId, {"assignmentId","giop-parlay.csapi.triggeredBindingRequestNotificationStop.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_termcap_IpTerminalCapabilities_getTerminalCapabilities_terminalIdentity, {"terminalIdentity","giop-parlay.csapi.getTerminalCapabilities.terminalIdentity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReport_assignmentID, {"assignmentID","giop-parlay.csapi.triggeredTerminalCapabilityReport.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReport_terminals_loop, {"Seq length of terminals","giop-parlay.csapi.triggeredTerminalCapabilityReport.terminals.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReport_criteria, {"criteria","giop-parlay.csapi.triggeredTerminalCapabilityReport.criteria",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReportErr_assignmentId, {"assignmentId","giop-parlay.csapi.triggeredTerminalCapabilityReportErr.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReportErr_terminals_loop, {"Seq length of terminals","giop-parlay.csapi.triggeredTerminalCapabilityReportErr.terminals.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_termcap_IpAppExtendedTerminalCapabilities_triggeredTerminalCapabilityReportErr_cause, {"cause","giop-parlay.csapi.triggeredTerminalCapabilityReportErr.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_termcap_TpTerminalCapabilitiesError),0x0,NULL,HFILL}},
        {&hf_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStartReq_return, {"TpAssignmentID","giop-parlay.csapi.triggeredTerminalCapabilityStartReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStartReq_terminals_loop, {"Seq length of terminals","giop-parlay.csapi.triggeredTerminalCapabilityStartReq.terminals.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStartReq_criteria, {"criteria","giop-parlay.csapi.triggeredTerminalCapabilityStartReq.criteria",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_termcap_IpExtendedTerminalCapabilities_triggeredTerminalCapabilityStop_assignmentID, {"assignmentID","giop-parlay.csapi.triggeredTerminalCapabilityStop.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpAppDataSession_connectRes_dataSessionID, {"dataSessionID","giop-parlay.csapi.connectRes.dataSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpAppDataSession_connectRes_assignmentID, {"assignmentID","giop-parlay.csapi.connectRes.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpAppDataSession_connectErr_dataSessionID, {"dataSessionID","giop-parlay.csapi.connectErr.dataSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpAppDataSession_connectErr_assignmentID, {"assignmentID","giop-parlay.csapi.connectErr.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpAppDataSession_superviseDataSessionRes_dataSessionID, {"dataSessionID","giop-parlay.csapi.superviseDataSessionRes.dataSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpAppDataSession_superviseDataSessionRes_report, {"report","giop-parlay.csapi.superviseDataSessionRes.report",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpAppDataSession_superviseDataSessionRes_qualityOfService, {"qualityOfService","giop-parlay.csapi.superviseDataSessionRes.qualityOfService",FT_UINT32,BASE_DEC,VALS(org_csapi_TpDataSessionQosClass),0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpAppDataSession_superviseDataSessionErr_dataSessionID, {"dataSessionID","giop-parlay.csapi.superviseDataSessionErr.dataSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpAppDataSession_dataSessionFaultDetected_dataSessionID, {"dataSessionID","giop-parlay.csapi.dataSessionFaultDetected.dataSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpAppDataSession_dataSessionFaultDetected_fault, {"fault","giop-parlay.csapi.dataSessionFaultDetected.fault",FT_UINT32,BASE_DEC,VALS(org_csapi_dsc_TpDataSessionFault),0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpAppDataSessionControlManager_dataSessionAborted_dataSession, {"dataSession","giop-parlay.csapi.dataSessionAborted.dataSession",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpAppDataSessionControlManager_reportNotification_assignmentID, {"assignmentID","giop-parlay.csapi.reportNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpAppDataSessionControlManager_abortMultipleDataSessions_dataSessionSet_loop, {"Seq length of dataSessionSet","giop-parlay.csapi.abortMultipleDataSessions.dataSessionSet.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpAppDataSessionControlManager_abortMultipleDataSessions_dataSessionSet, {"dataSessionSet","giop-parlay.csapi.abortMultipleDataSessions.dataSessionSet",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpDataSession_connectReq_return, {"TpAssignmentID","giop-parlay.csapi.connectReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpDataSession_connectReq_dataSessionID, {"dataSessionID","giop-parlay.csapi.connectReq.dataSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpDataSession_connectReq_responseRequested_loop, {"Seq length of responseRequested","giop-parlay.csapi.connectReq.responseRequested.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpDataSession_release_dataSessionID, {"dataSessionID","giop-parlay.csapi.release.dataSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpDataSession_superviseDataSessionReq_dataSessionID, {"dataSessionID","giop-parlay.csapi.superviseDataSessionReq.dataSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpDataSession_superviseDataSessionReq_treatment, {"treatment","giop-parlay.csapi.superviseDataSessionReq.treatment",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpDataSession_setDataSessionChargePlan_dataSessionID, {"dataSessionID","giop-parlay.csapi.setDataSessionChargePlan.dataSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpDataSession_setAdviceOfCharge_dataSessionID, {"dataSessionID","giop-parlay.csapi.setAdviceOfCharge.dataSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpDataSession_setAdviceOfCharge_tariffSwitch, {"tariffSwitch","giop-parlay.csapi.setAdviceOfCharge.tariffSwitch",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpDataSession_deassignDataSession_dataSessionID, {"dataSessionID","giop-parlay.csapi.deassignDataSession.dataSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpDataSession_continueProcessing_dataSessionID, {"dataSessionID","giop-parlay.csapi.continueProcessing.dataSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpDataSessionControlManager_createNotification_return, {"TpAssignmentID","giop-parlay.csapi.createNotification.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpDataSessionControlManager_destroyNotification_assignmentID, {"assignmentID","giop-parlay.csapi.destroyNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpDataSessionControlManager_changeNotification_assignmentID, {"assignmentID","giop-parlay.csapi.changeNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpDataSessionControlManager_enableNotifications_return, {"TpAssignmentID","giop-parlay.csapi.enableNotifications.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpDataSessionControlManager_getNotifications_return_loop, {"Seq length of TpDataSessionEventCriteriaResultSet","giop-parlay.csapi.getNotifications.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_IpDataSessionControlManager_createNotifications_return, {"TpAssignmentID","giop-parlay.csapi.createNotifications.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMessagingManager_openMailbox_authenticationInfo, {"authenticationInfo","giop-parlay.csapi.openMailbox.authenticationInfo",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMessagingManager_enableMessagingNotification_return, {"TpAssignmentID","giop-parlay.csapi.enableMessagingNotification.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMessagingManager_disableMessagingNotification_assignmentID, {"assignmentID","giop-parlay.csapi.disableMessagingNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpAppMessagingManager_mailboxTerminated_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.mailboxTerminated.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpAppMessagingManager_mailboxFaultDetected_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.mailboxFaultDetected.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpAppMessagingManager_mailboxFaultDetected_fault, {"fault","giop-parlay.csapi.mailboxFaultDetected.fault",FT_UINT32,BASE_DEC,VALS(org_csapi_gms_TpMessagingFault),0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpAppMessagingManager_messagingEventNotify_assignmentID, {"assignmentID","giop-parlay.csapi.messagingEventNotify.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailbox_close_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.close.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailbox_lock_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.lock.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailbox_unlock_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.unlock.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailbox_getInfoAmount_return, {"TpInt32","giop-parlay.csapi.getInfoAmount.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailbox_getInfoAmount_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getInfoAmount.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailbox_getInfoProperties_return_loop, {"Seq length of TpMailboxInfoPropertySet","giop-parlay.csapi.getInfoProperties.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailbox_getInfoProperties_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getInfoProperties.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailbox_getInfoProperties_firstProperty, {"firstProperty","giop-parlay.csapi.getInfoProperties.firstProperty",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailbox_getInfoProperties_numberOfProperties, {"numberOfProperties","giop-parlay.csapi.getInfoProperties.numberOfProperties",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailbox_setInfoProperties_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.setInfoProperties.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailbox_setInfoProperties_firstProperty, {"firstProperty","giop-parlay.csapi.setInfoProperties.firstProperty",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailbox_setInfoProperties_mailboxInfoProperties_loop, {"Seq length of mailboxInfoProperties","giop-parlay.csapi.setInfoProperties.mailboxInfoProperties.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailbox_openFolder_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.openFolder.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailbox_openFolder_folderID, {"folderID","giop-parlay.csapi.openFolder.folderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailbox_createFolder_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.createFolder.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailbox_createFolder_folderID, {"folderID","giop-parlay.csapi.createFolder.folderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailbox_remove_authenticationInfo, {"authenticationInfo","giop-parlay.csapi.remove.authenticationInfo",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMessage_getInfoAmount_return, {"TpInt32","giop-parlay.csapi.getInfoAmount.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMessage_getInfoAmount_folderSessionID, {"folderSessionID","giop-parlay.csapi.getInfoAmount.folderSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMessage_getInfoAmount_messageID, {"messageID","giop-parlay.csapi.getInfoAmount.messageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMessage_getInfoProperties_return_loop, {"Seq length of TpMessageInfoPropertySet","giop-parlay.csapi.getInfoProperties.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMessage_getInfoProperties_folderSessionID, {"folderSessionID","giop-parlay.csapi.getInfoProperties.folderSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMessage_getInfoProperties_messageID, {"messageID","giop-parlay.csapi.getInfoProperties.messageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMessage_getInfoProperties_firstProperty, {"firstProperty","giop-parlay.csapi.getInfoProperties.firstProperty",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMessage_getInfoProperties_numberOfProperties, {"numberOfProperties","giop-parlay.csapi.getInfoProperties.numberOfProperties",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMessage_setInfoProperties_folderSessionID, {"folderSessionID","giop-parlay.csapi.setInfoProperties.folderSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMessage_setInfoProperties_messageID, {"messageID","giop-parlay.csapi.setInfoProperties.messageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMessage_setInfoProperties_firstProperty, {"firstProperty","giop-parlay.csapi.setInfoProperties.firstProperty",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMessage_setInfoProperties_messageInfoProperties_loop, {"Seq length of messageInfoProperties","giop-parlay.csapi.setInfoProperties.messageInfoProperties.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMessage_remove_folderSessionID, {"folderSessionID","giop-parlay.csapi.remove.folderSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMessage_remove_messageID, {"messageID","giop-parlay.csapi.remove.messageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMessage_getContent_return, {"TpMessage","giop-parlay.csapi.getContent.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMessage_getContent_folderSessionID, {"folderSessionID","giop-parlay.csapi.getContent.folderSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMessage_getContent_messageID, {"messageID","giop-parlay.csapi.getContent.messageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailboxFolder_getInfoAmount_return, {"TpInt32","giop-parlay.csapi.getInfoAmount.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailboxFolder_getInfoAmount_folderSessionID, {"folderSessionID","giop-parlay.csapi.getInfoAmount.folderSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailboxFolder_getInfoProperties_return_loop, {"Seq length of TpFolderInfoPropertySet","giop-parlay.csapi.getInfoProperties.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailboxFolder_getInfoProperties_folderSessionID, {"folderSessionID","giop-parlay.csapi.getInfoProperties.folderSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailboxFolder_getInfoProperties_firstProperty, {"firstProperty","giop-parlay.csapi.getInfoProperties.firstProperty",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailboxFolder_getInfoProperties_numberOfProperties, {"numberOfProperties","giop-parlay.csapi.getInfoProperties.numberOfProperties",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailboxFolder_setInfoProperties_folderSessionID, {"folderSessionID","giop-parlay.csapi.setInfoProperties.folderSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailboxFolder_setInfoProperties_firstProperty, {"firstProperty","giop-parlay.csapi.setInfoProperties.firstProperty",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailboxFolder_setInfoProperties_folderInfoProperties_loop, {"Seq length of folderInfoProperties","giop-parlay.csapi.setInfoProperties.folderInfoProperties.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailboxFolder_putMessage_folderSessionID, {"folderSessionID","giop-parlay.csapi.putMessage.folderSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailboxFolder_putMessage_message, {"message","giop-parlay.csapi.putMessage.message",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailboxFolder_putMessage_messageInfoProperties_loop, {"Seq length of messageInfoProperties","giop-parlay.csapi.putMessage.messageInfoProperties.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailboxFolder_getMessage_folderSessionID, {"folderSessionID","giop-parlay.csapi.getMessage.folderSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailboxFolder_getMessage_messageID, {"messageID","giop-parlay.csapi.getMessage.messageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailboxFolder_close_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.close.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailboxFolder_close_folderSessionID, {"folderSessionID","giop-parlay.csapi.close.folderSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailboxFolder_remove_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.remove.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_IpMailboxFolder_remove_folderID, {"folderID","giop-parlay.csapi.remove.folderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_IpVPrP_getVPrPID_return, {"TpString","giop-parlay.csapi.getVPrPID.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_IpVPrP_getSlaID_return, {"TpString","giop-parlay.csapi.getSlaID.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_IpVPrP_getStatus_return, {"Return value","giop-parlay.csapi.getStatus.return",FT_UINT32,BASE_DEC,VALS(org_csapi_cm_TpVprpStatus),0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_IpVPrN_getVPrPList_return_loop, {"Seq length of TpStringList","giop-parlay.csapi.getVPrPList.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_IpVPrN_getVPrPList_return, {"TpStringList","giop-parlay.csapi.getVPrPList.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_IpVPrN_getVPrP_vPrPID, {"vPrPID","giop-parlay.csapi.getVPrP.vPrPID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_IpVPrN_deleteVPrP_vPrPID, {"vPrPID","giop-parlay.csapi.deleteVPrP.vPrPID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_IpQoSMenu_getTemplate_templateType, {"templateType","giop-parlay.csapi.getTemplate.templateType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_IpQoSMenu_getTemplateList_return_loop, {"Seq length of TpStringList","giop-parlay.csapi.getTemplateList.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_IpQoSMenu_getTemplateList_return, {"TpStringList","giop-parlay.csapi.getTemplateList.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_IpQoSTemplate_getTemplateType_return, {"TpString","giop-parlay.csapi.getTemplateType.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_IpQoSTemplate_getDescription_return, {"TpString","giop-parlay.csapi.getDescription.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_IpQoSTemplate_setSlaID_slaID, {"slaID","giop-parlay.csapi.setSlaID.slaID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_IpEnterpriseNetwork_getSiteList_return_loop, {"Seq length of TpStringList","giop-parlay.csapi.getSiteList.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_IpEnterpriseNetwork_getSiteList_return, {"TpStringList","giop-parlay.csapi.getSiteList.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_IpEnterpriseNetwork_getSite_siteID, {"siteID","giop-parlay.csapi.getSite.siteID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_IpEnterpriseNetworkSite_getSAPList_return_loop, {"Seq length of TpStringList","giop-parlay.csapi.getSAPList.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_IpEnterpriseNetworkSite_getSAPList_return, {"TpStringList","giop-parlay.csapi.getSAPList.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_IpEnterpriseNetworkSite_getSiteID_return, {"TpString","giop-parlay.csapi.getSiteID.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_IpEnterpriseNetworkSite_getSiteLocation_return, {"TpString","giop-parlay.csapi.getSiteLocation.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_IpEnterpriseNetworkSite_getSiteDescription_return, {"TpString","giop-parlay.csapi.getSiteDescription.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_IpEnterpriseNetworkSite_getSAPIPSubnet_sapID, {"sapID","giop-parlay.csapi.getSAPIPSubnet.sapID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_reportNotification_assignmentId, {"assignmentId","giop-parlay.csapi.reportNotification.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_queryBalanceRes_queryId, {"queryId","giop-parlay.csapi.queryBalanceRes.queryId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_queryBalanceRes_balances_loop, {"Seq length of balances","giop-parlay.csapi.queryBalanceRes.balances.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_queryBalanceErr_queryId, {"queryId","giop-parlay.csapi.queryBalanceErr.queryId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_queryBalanceErr_cause, {"cause","giop-parlay.csapi.queryBalanceErr.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_am_TpBalanceQueryError),0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryRes_retrievalID, {"retrievalID","giop-parlay.csapi.retrieveTransactionHistoryRes.retrievalID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryRes_transactionHistory_loop, {"Seq length of transactionHistory","giop-parlay.csapi.retrieveTransactionHistoryRes.transactionHistory.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryErr_retrievalID, {"retrievalID","giop-parlay.csapi.retrieveTransactionHistoryErr.retrievalID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_retrieveTransactionHistoryErr_transactionHistoryError, {"transactionHistoryError","giop-parlay.csapi.retrieveTransactionHistoryErr.transactionHistoryError",FT_UINT32,BASE_DEC,VALS(org_csapi_am_TpTransactionHistoryStatus),0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateRes_queryId, {"queryId","giop-parlay.csapi.queryBalanceExpiryDateRes.queryId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateRes_dates_loop, {"Seq length of dates","giop-parlay.csapi.queryBalanceExpiryDateRes.dates.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateErr_queryId, {"queryId","giop-parlay.csapi.queryBalanceExpiryDateErr.queryId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_queryBalanceExpiryDateErr_cause, {"cause","giop-parlay.csapi.queryBalanceExpiryDateErr.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_am_TpBalanceQueryError),0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_updateBalanceRes_requestId, {"requestId","giop-parlay.csapi.updateBalanceRes.requestId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_updateBalanceErr_requestId, {"requestId","giop-parlay.csapi.updateBalanceErr.requestId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_updateBalanceErr_cause, {"cause","giop-parlay.csapi.updateBalanceErr.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_am_TpBalanceQueryError),0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_createVoucherRes_requestId, {"requestId","giop-parlay.csapi.createVoucherRes.requestId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_createVoucherRes_voucherId, {"voucherId","giop-parlay.csapi.createVoucherRes.voucherId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_createVoucherErr_requestId, {"requestId","giop-parlay.csapi.createVoucherErr.requestId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_createVoucherErr_cause, {"cause","giop-parlay.csapi.createVoucherErr.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_am_TpVoucherError),0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_destroyVoucherRes_requestId, {"requestId","giop-parlay.csapi.destroyVoucherRes.requestId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_destroyVoucherRes_voucherId, {"voucherId","giop-parlay.csapi.destroyVoucherRes.voucherId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_destroyVoucherErr_requestId, {"requestId","giop-parlay.csapi.destroyVoucherErr.requestId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_destroyVoucherErr_voucherId, {"voucherId","giop-parlay.csapi.destroyVoucherErr.voucherId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_destroyVoucherErr_cause, {"cause","giop-parlay.csapi.destroyVoucherErr.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_am_TpVoucherError),0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_queryVoucherRes_queryId, {"queryId","giop-parlay.csapi.queryVoucherRes.queryId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_queryVoucherErr_queryId, {"queryId","giop-parlay.csapi.queryVoucherErr.queryId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_queryVoucherErr_voucherId, {"voucherId","giop-parlay.csapi.queryVoucherErr.voucherId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_queryVoucherErr_cause, {"cause","giop-parlay.csapi.queryVoucherErr.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_am_TpVoucherError),0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_queryUserVouchersRes_queryId, {"queryId","giop-parlay.csapi.queryUserVouchersRes.queryId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_queryUserVouchersRes_vouchers_loop, {"Seq length of vouchers","giop-parlay.csapi.queryUserVouchersRes.vouchers.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_queryUserVouchersErr_queryId, {"queryId","giop-parlay.csapi.queryUserVouchersErr.queryId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAppAccountManager_queryUserVouchersErr_cause, {"cause","giop-parlay.csapi.queryUserVouchersErr.cause",FT_UINT32,BASE_DEC,VALS(org_csapi_am_TpVoucherError),0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAccountManager_createNotification_return, {"TpAssignmentID","giop-parlay.csapi.createNotification.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAccountManager_destroyNotification_assignmentId, {"assignmentId","giop-parlay.csapi.destroyNotification.assignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAccountManager_queryBalanceReq_return, {"TpAssignmentID","giop-parlay.csapi.queryBalanceReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAccountManager_queryBalanceReq_users_loop, {"Seq length of users","giop-parlay.csapi.queryBalanceReq.users.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAccountManager_changeNotification_assignmentID, {"assignmentID","giop-parlay.csapi.changeNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAccountManager_getNotification_return_loop, {"Seq length of TpChargingEventCriteriaResultSet","giop-parlay.csapi.getNotification.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAccountManager_retrieveTransactionHistoryReq_return, {"TpAssignmentID","giop-parlay.csapi.retrieveTransactionHistoryReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAccountManager_enableNotifications_return, {"TpAssignmentID","giop-parlay.csapi.enableNotifications.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAccountManager_queryBalanceExpiryDateReq_return, {"TpAssignmentID","giop-parlay.csapi.queryBalanceExpiryDateReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAccountManager_queryBalanceExpiryDateReq_users_loop, {"Seq length of users","giop-parlay.csapi.queryBalanceExpiryDateReq.users.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAccountManager_updateBalanceReq_return, {"TpAssignmentID","giop-parlay.csapi.updateBalanceReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAccountManager_updateBalanceReq_debit, {"debit","giop-parlay.csapi.updateBalanceReq.debit",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_am_IpAccountManager_updateBalanceReq_period, {"period","giop-parlay.csapi.updateBalanceReq.period",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAccountManager_createVoucherReq_return, {"TpAssignmentID","giop-parlay.csapi.createVoucherReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAccountManager_destroyVoucherReq_return, {"TpAssignmentID","giop-parlay.csapi.destroyVoucherReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAccountManager_destroyVoucherReq_voucherId, {"voucherId","giop-parlay.csapi.destroyVoucherReq.voucherId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAccountManager_queryVoucherReq_return, {"TpAssignmentID","giop-parlay.csapi.queryVoucherReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAccountManager_queryVoucherReq_voucherId, {"voucherId","giop-parlay.csapi.queryVoucherReq.voucherId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_IpAccountManager_queryUserVouchersReq_return, {"TpAssignmentID","giop-parlay.csapi.queryUserVouchersReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_creditAmountErr_sessionID, {"sessionID","giop-parlay.csapi.creditAmountErr.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_creditAmountErr_requestNumber, {"requestNumber","giop-parlay.csapi.creditAmountErr.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_creditAmountErr_error, {"error","giop-parlay.csapi.creditAmountErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_cs_TpChargingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_creditAmountErr_requestNumberNextRequest, {"requestNumberNextRequest","giop-parlay.csapi.creditAmountErr.requestNumberNextRequest",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_creditAmountRes_sessionID, {"sessionID","giop-parlay.csapi.creditAmountRes.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_creditAmountRes_requestNumber, {"requestNumber","giop-parlay.csapi.creditAmountRes.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_creditAmountRes_requestNumberNextRequest, {"requestNumberNextRequest","giop-parlay.csapi.creditAmountRes.requestNumberNextRequest",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_creditUnitErr_sessionID, {"sessionID","giop-parlay.csapi.creditUnitErr.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_creditUnitErr_requestNumber, {"requestNumber","giop-parlay.csapi.creditUnitErr.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_creditUnitErr_error, {"error","giop-parlay.csapi.creditUnitErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_cs_TpChargingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_creditUnitErr_requestNumberNextRequest, {"requestNumberNextRequest","giop-parlay.csapi.creditUnitErr.requestNumberNextRequest",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_creditUnitRes_sessionID, {"sessionID","giop-parlay.csapi.creditUnitRes.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_creditUnitRes_requestNumber, {"requestNumber","giop-parlay.csapi.creditUnitRes.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_creditUnitRes_creditedVolumes_loop, {"Seq length of creditedVolumes","giop-parlay.csapi.creditUnitRes.creditedVolumes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_creditUnitRes_reservedUnitsLeft_loop, {"Seq length of reservedUnitsLeft","giop-parlay.csapi.creditUnitRes.reservedUnitsLeft.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_creditUnitRes_requestNumberNextRequest, {"requestNumberNextRequest","giop-parlay.csapi.creditUnitRes.requestNumberNextRequest",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_debitAmountErr_sessionID, {"sessionID","giop-parlay.csapi.debitAmountErr.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_debitAmountErr_requestNumber, {"requestNumber","giop-parlay.csapi.debitAmountErr.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_debitAmountErr_error, {"error","giop-parlay.csapi.debitAmountErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_cs_TpChargingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_debitAmountErr_requestNumberNextRequest, {"requestNumberNextRequest","giop-parlay.csapi.debitAmountErr.requestNumberNextRequest",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_debitAmountRes_sessionID, {"sessionID","giop-parlay.csapi.debitAmountRes.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_debitAmountRes_requestNumber, {"requestNumber","giop-parlay.csapi.debitAmountRes.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_debitAmountRes_requestNumberNextRequest, {"requestNumberNextRequest","giop-parlay.csapi.debitAmountRes.requestNumberNextRequest",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_debitUnitErr_sessionID, {"sessionID","giop-parlay.csapi.debitUnitErr.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_debitUnitErr_requestNumber, {"requestNumber","giop-parlay.csapi.debitUnitErr.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_debitUnitErr_error, {"error","giop-parlay.csapi.debitUnitErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_cs_TpChargingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_debitUnitErr_requestNumberNextRequest, {"requestNumberNextRequest","giop-parlay.csapi.debitUnitErr.requestNumberNextRequest",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_debitUnitRes_sessionID, {"sessionID","giop-parlay.csapi.debitUnitRes.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_debitUnitRes_requestNumber, {"requestNumber","giop-parlay.csapi.debitUnitRes.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_debitUnitRes_debitedVolumes_loop, {"Seq length of debitedVolumes","giop-parlay.csapi.debitUnitRes.debitedVolumes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_debitUnitRes_reservedUnitsLeft_loop, {"Seq length of reservedUnitsLeft","giop-parlay.csapi.debitUnitRes.reservedUnitsLeft.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_debitUnitRes_requestNumberNextRequest, {"requestNumberNextRequest","giop-parlay.csapi.debitUnitRes.requestNumberNextRequest",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directCreditAmountErr_sessionID, {"sessionID","giop-parlay.csapi.directCreditAmountErr.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directCreditAmountErr_requestNumber, {"requestNumber","giop-parlay.csapi.directCreditAmountErr.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directCreditAmountErr_error, {"error","giop-parlay.csapi.directCreditAmountErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_cs_TpChargingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directCreditAmountErr_requestNumberNextRequest, {"requestNumberNextRequest","giop-parlay.csapi.directCreditAmountErr.requestNumberNextRequest",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directCreditAmountRes_sessionID, {"sessionID","giop-parlay.csapi.directCreditAmountRes.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directCreditAmountRes_requestNumber, {"requestNumber","giop-parlay.csapi.directCreditAmountRes.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directCreditAmountRes_requestNumberNextRequest, {"requestNumberNextRequest","giop-parlay.csapi.directCreditAmountRes.requestNumberNextRequest",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directCreditUnitErr_sessionID, {"sessionID","giop-parlay.csapi.directCreditUnitErr.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directCreditUnitErr_requestNumber, {"requestNumber","giop-parlay.csapi.directCreditUnitErr.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directCreditUnitErr_error, {"error","giop-parlay.csapi.directCreditUnitErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_cs_TpChargingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directCreditUnitErr_requestNumberNextRequest, {"requestNumberNextRequest","giop-parlay.csapi.directCreditUnitErr.requestNumberNextRequest",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directCreditUnitRes_sessionID, {"sessionID","giop-parlay.csapi.directCreditUnitRes.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directCreditUnitRes_requestNumber, {"requestNumber","giop-parlay.csapi.directCreditUnitRes.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directCreditUnitRes_creditedVolumes_loop, {"Seq length of creditedVolumes","giop-parlay.csapi.directCreditUnitRes.creditedVolumes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directCreditUnitRes_requestNumberNextRequest, {"requestNumberNextRequest","giop-parlay.csapi.directCreditUnitRes.requestNumberNextRequest",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directDebitAmountErr_sessionID, {"sessionID","giop-parlay.csapi.directDebitAmountErr.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directDebitAmountErr_requestNumber, {"requestNumber","giop-parlay.csapi.directDebitAmountErr.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directDebitAmountErr_error, {"error","giop-parlay.csapi.directDebitAmountErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_cs_TpChargingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directDebitAmountErr_requestNumberNextRequest, {"requestNumberNextRequest","giop-parlay.csapi.directDebitAmountErr.requestNumberNextRequest",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directDebitAmountRes_sessionID, {"sessionID","giop-parlay.csapi.directDebitAmountRes.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directDebitAmountRes_requestNumber, {"requestNumber","giop-parlay.csapi.directDebitAmountRes.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directDebitAmountRes_requestNumberNextRequest, {"requestNumberNextRequest","giop-parlay.csapi.directDebitAmountRes.requestNumberNextRequest",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directDebitUnitErr_sessionID, {"sessionID","giop-parlay.csapi.directDebitUnitErr.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directDebitUnitErr_requestNumber, {"requestNumber","giop-parlay.csapi.directDebitUnitErr.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directDebitUnitErr_error, {"error","giop-parlay.csapi.directDebitUnitErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_cs_TpChargingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directDebitUnitErr_requestNumberNextRequest, {"requestNumberNextRequest","giop-parlay.csapi.directDebitUnitErr.requestNumberNextRequest",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directDebitUnitRes_sessionID, {"sessionID","giop-parlay.csapi.directDebitUnitRes.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directDebitUnitRes_requestNumber, {"requestNumber","giop-parlay.csapi.directDebitUnitRes.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directDebitUnitRes_debitedVolumes_loop, {"Seq length of debitedVolumes","giop-parlay.csapi.directDebitUnitRes.debitedVolumes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_directDebitUnitRes_requestNumberNextRequest, {"requestNumberNextRequest","giop-parlay.csapi.directDebitUnitRes.requestNumberNextRequest",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_extendLifeTimeErr_sessionID, {"sessionID","giop-parlay.csapi.extendLifeTimeErr.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_extendLifeTimeErr_error, {"error","giop-parlay.csapi.extendLifeTimeErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_cs_TpChargingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_extendLifeTimeRes_sessionID, {"sessionID","giop-parlay.csapi.extendLifeTimeRes.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_extendLifeTimeRes_sessionTimeLeft, {"sessionTimeLeft","giop-parlay.csapi.extendLifeTimeRes.sessionTimeLeft",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_rateErr_sessionID, {"sessionID","giop-parlay.csapi.rateErr.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_rateErr_error, {"error","giop-parlay.csapi.rateErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_cs_TpChargingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_rateRes_sessionID, {"sessionID","giop-parlay.csapi.rateRes.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_rateRes_rates_loop, {"Seq length of rates","giop-parlay.csapi.rateRes.rates.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_rateRes_validityTimeLeft, {"validityTimeLeft","giop-parlay.csapi.rateRes.validityTimeLeft",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_reserveAmountErr_sessionID, {"sessionID","giop-parlay.csapi.reserveAmountErr.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_reserveAmountErr_requestNumber, {"requestNumber","giop-parlay.csapi.reserveAmountErr.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_reserveAmountErr_error, {"error","giop-parlay.csapi.reserveAmountErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_cs_TpChargingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_reserveAmountErr_requestNumberNextRequest, {"requestNumberNextRequest","giop-parlay.csapi.reserveAmountErr.requestNumberNextRequest",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_reserveAmountRes_sessionID, {"sessionID","giop-parlay.csapi.reserveAmountRes.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_reserveAmountRes_requestNumber, {"requestNumber","giop-parlay.csapi.reserveAmountRes.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_reserveAmountRes_sessionTimeLeft, {"sessionTimeLeft","giop-parlay.csapi.reserveAmountRes.sessionTimeLeft",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_reserveAmountRes_requestNumberNextRequest, {"requestNumberNextRequest","giop-parlay.csapi.reserveAmountRes.requestNumberNextRequest",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_reserveUnitErr_sessionID, {"sessionID","giop-parlay.csapi.reserveUnitErr.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_reserveUnitErr_requestNumber, {"requestNumber","giop-parlay.csapi.reserveUnitErr.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_reserveUnitErr_error, {"error","giop-parlay.csapi.reserveUnitErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_cs_TpChargingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_reserveUnitErr_requestNumberNextRequest, {"requestNumberNextRequest","giop-parlay.csapi.reserveUnitErr.requestNumberNextRequest",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_reserveUnitRes_sessionID, {"sessionID","giop-parlay.csapi.reserveUnitRes.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_reserveUnitRes_requestNumber, {"requestNumber","giop-parlay.csapi.reserveUnitRes.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_reserveUnitRes_reservedUnits_loop, {"Seq length of reservedUnits","giop-parlay.csapi.reserveUnitRes.reservedUnits.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_reserveUnitRes_sessionTimeLeft, {"sessionTimeLeft","giop-parlay.csapi.reserveUnitRes.sessionTimeLeft",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_reserveUnitRes_requestNumberNextRequest, {"requestNumberNextRequest","giop-parlay.csapi.reserveUnitRes.requestNumberNextRequest",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_sessionEnded_sessionID, {"sessionID","giop-parlay.csapi.sessionEnded.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingSession_sessionEnded_report, {"report","giop-parlay.csapi.sessionEnded.report",FT_UINT32,BASE_DEC,VALS(org_csapi_cs_TpSessionEndedCause),0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_creditAmountReq_sessionID, {"sessionID","giop-parlay.csapi.creditAmountReq.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_creditAmountReq_closeReservation, {"closeReservation","giop-parlay.csapi.creditAmountReq.closeReservation",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_creditAmountReq_requestNumber, {"requestNumber","giop-parlay.csapi.creditAmountReq.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_creditUnitReq_sessionID, {"sessionID","giop-parlay.csapi.creditUnitReq.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_creditUnitReq_volumes_loop, {"Seq length of volumes","giop-parlay.csapi.creditUnitReq.volumes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_creditUnitReq_closeReservation, {"closeReservation","giop-parlay.csapi.creditUnitReq.closeReservation",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_creditUnitReq_requestNumber, {"requestNumber","giop-parlay.csapi.creditUnitReq.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_debitAmountReq_sessionID, {"sessionID","giop-parlay.csapi.debitAmountReq.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_debitAmountReq_closeReservation, {"closeReservation","giop-parlay.csapi.debitAmountReq.closeReservation",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_debitAmountReq_requestNumber, {"requestNumber","giop-parlay.csapi.debitAmountReq.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_debitUnitReq_sessionID, {"sessionID","giop-parlay.csapi.debitUnitReq.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_debitUnitReq_volumes_loop, {"Seq length of volumes","giop-parlay.csapi.debitUnitReq.volumes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_debitUnitReq_closeReservation, {"closeReservation","giop-parlay.csapi.debitUnitReq.closeReservation",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_debitUnitReq_requestNumber, {"requestNumber","giop-parlay.csapi.debitUnitReq.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_directCreditAmountReq_sessionID, {"sessionID","giop-parlay.csapi.directCreditAmountReq.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_directCreditAmountReq_chargingParameters_loop, {"Seq length of chargingParameters","giop-parlay.csapi.directCreditAmountReq.chargingParameters.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_directCreditAmountReq_requestNumber, {"requestNumber","giop-parlay.csapi.directCreditAmountReq.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_directCreditUnitReq_sessionID, {"sessionID","giop-parlay.csapi.directCreditUnitReq.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_directCreditUnitReq_chargingParameters_loop, {"Seq length of chargingParameters","giop-parlay.csapi.directCreditUnitReq.chargingParameters.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_directCreditUnitReq_volumes_loop, {"Seq length of volumes","giop-parlay.csapi.directCreditUnitReq.volumes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_directCreditUnitReq_requestNumber, {"requestNumber","giop-parlay.csapi.directCreditUnitReq.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_directDebitAmountReq_sessionID, {"sessionID","giop-parlay.csapi.directDebitAmountReq.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_directDebitAmountReq_chargingParameters_loop, {"Seq length of chargingParameters","giop-parlay.csapi.directDebitAmountReq.chargingParameters.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_directDebitAmountReq_requestNumber, {"requestNumber","giop-parlay.csapi.directDebitAmountReq.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_directDebitUnitReq_sessionID, {"sessionID","giop-parlay.csapi.directDebitUnitReq.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_directDebitUnitReq_chargingParameters_loop, {"Seq length of chargingParameters","giop-parlay.csapi.directDebitUnitReq.chargingParameters.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_directDebitUnitReq_volumes_loop, {"Seq length of volumes","giop-parlay.csapi.directDebitUnitReq.volumes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_directDebitUnitReq_requestNumber, {"requestNumber","giop-parlay.csapi.directDebitUnitReq.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_extendLifeTimeReq_sessionID, {"sessionID","giop-parlay.csapi.extendLifeTimeReq.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_getAmountLeft_sessionID, {"sessionID","giop-parlay.csapi.getAmountLeft.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_getLifeTimeLeft_return, {"TpInt32","giop-parlay.csapi.getLifeTimeLeft.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_getLifeTimeLeft_sessionID, {"sessionID","giop-parlay.csapi.getLifeTimeLeft.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_getUnitLeft_return_loop, {"Seq length of TpVolumeSet","giop-parlay.csapi.getUnitLeft.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_getUnitLeft_sessionID, {"sessionID","giop-parlay.csapi.getUnitLeft.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_rateReq_sessionID, {"sessionID","giop-parlay.csapi.rateReq.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_rateReq_chargingParameters_loop, {"Seq length of chargingParameters","giop-parlay.csapi.rateReq.chargingParameters.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_release_sessionID, {"sessionID","giop-parlay.csapi.release.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_release_requestNumber, {"requestNumber","giop-parlay.csapi.release.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_reserveAmountReq_sessionID, {"sessionID","giop-parlay.csapi.reserveAmountReq.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_reserveAmountReq_chargingParameters_loop, {"Seq length of chargingParameters","giop-parlay.csapi.reserveAmountReq.chargingParameters.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_reserveAmountReq_requestNumber, {"requestNumber","giop-parlay.csapi.reserveAmountReq.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_reserveUnitReq_sessionID, {"sessionID","giop-parlay.csapi.reserveUnitReq.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_reserveUnitReq_chargingParameters_loop, {"Seq length of chargingParameters","giop-parlay.csapi.reserveUnitReq.chargingParameters.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_reserveUnitReq_volumes_loop, {"Seq length of volumes","giop-parlay.csapi.reserveUnitReq.volumes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingSession_reserveUnitReq_requestNumber, {"requestNumber","giop-parlay.csapi.reserveUnitReq.requestNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingManager_sessionAborted_sessionID, {"sessionID","giop-parlay.csapi.sessionAborted.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingManager_abortMultipleChargingSessions_chargingSessionSet_loop, {"Seq length of chargingSessionSet","giop-parlay.csapi.abortMultipleChargingSessions.chargingSessionSet.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpAppChargingManager_abortMultipleChargingSessions_chargingSessionSet, {"chargingSessionSet","giop-parlay.csapi.abortMultipleChargingSessions.chargingSessionSet",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingManager_createChargingSession_sessionDescription, {"sessionDescription","giop-parlay.csapi.createChargingSession.sessionDescription",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingManager_createSplitChargingSession_sessionDescription, {"sessionDescription","giop-parlay.csapi.createSplitChargingSession.sessionDescription",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_IpChargingManager_createSplitChargingSession_users_loop, {"Seq length of users","giop-parlay.csapi.createSplitChargingSession.users.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpAppPolicyDomain_reportNotification_assignmentID, {"assignmentID","giop-parlay.csapi.reportNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyRes_assignmentID, {"assignmentID","giop-parlay.csapi.evalPolicyRes.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyRes_outputVariables_loop, {"Seq length of outputVariables","giop-parlay.csapi.evalPolicyRes.outputVariables.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyErr_assignmentID, {"assignmentID","giop-parlay.csapi.evalPolicyErr.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpAppPolicyDomain_evalPolicyErr_error, {"error","giop-parlay.csapi.evalPolicyErr.error",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_return_loop, {"Seq length of TpPolicyNameValueList","giop-parlay.csapi.evalPolicy.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_domainName, {"domainName","giop-parlay.csapi.evalPolicy.domainName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_signatureName, {"signatureName","giop-parlay.csapi.evalPolicy.signatureName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicy_inputVariables_loop, {"Seq length of inputVariables","giop-parlay.csapi.evalPolicy.inputVariables.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicyReq_return, {"TpAssignmentID","giop-parlay.csapi.evalPolicyReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicyReq_domainName, {"domainName","giop-parlay.csapi.evalPolicyReq.domainName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicyReq_signatureName, {"signatureName","giop-parlay.csapi.evalPolicyReq.signatureName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpPolicyEvalManager_evalPolicyReq_inputVariables_loop, {"Seq length of inputVariables","giop-parlay.csapi.evalPolicyReq.inputVariables.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpPolicyEvalManager_abortEvalPolicyReq_domainName, {"domainName","giop-parlay.csapi.abortEvalPolicyReq.domainName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpPolicyEvalManager_abortEvalPolicyReq_assignmentID, {"assignmentID","giop-parlay.csapi.abortEvalPolicyReq.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpPolicyEvalManager_generateEvent_domainName, {"domainName","giop-parlay.csapi.generateEvent.domainName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpPolicyEvalManager_generateEvent_eventDefinitionName, {"eventDefinitionName","giop-parlay.csapi.generateEvent.eventDefinitionName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpPolicyEvalManager_generateEvent_attributes_loop, {"Seq length of attributes","giop-parlay.csapi.generateEvent.attributes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpPolicyEvalManager_createNotification_return, {"TpAssignmentID","giop-parlay.csapi.createNotification.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpPolicyEvalManager_createNotification_domainName, {"domainName","giop-parlay.csapi.createNotification.domainName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpPolicyEvalManager_createNotification_events_loop, {"Seq length of events","giop-parlay.csapi.createNotification.events.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpPolicyEvalManager_createNotification_events, {"events","giop-parlay.csapi.createNotification.events",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpPolicyEvalManager_destroyNotification_assignmentID, {"assignmentID","giop-parlay.csapi.destroyNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpPolicyEvalManager_destroyNotification_events_loop, {"Seq length of events","giop-parlay.csapi.destroyNotification.events.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_evaluation_IpPolicyEvalManager_destroyNotification_events, {"events","giop-parlay.csapi.destroyNotification.events",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicy_getAttribute_attributeName, {"attributeName","giop-parlay.csapi.getAttribute.attributeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicy_getAttributes_return_loop, {"Seq length of TpAttributeSet","giop-parlay.csapi.getAttributes.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicy_getAttributes_attributeNames_loop, {"Seq length of attributeNames","giop-parlay.csapi.getAttributes.attributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicy_getAttributes_attributeNames, {"attributeNames","giop-parlay.csapi.getAttributes.attributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicy_setAttributes_targetAttributes_loop, {"Seq length of targetAttributes","giop-parlay.csapi.setAttributes.targetAttributes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyIterator_getList_return_loop, {"Seq length of TpStringSet","giop-parlay.csapi.getList.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyIterator_getList_return, {"TpStringSet","giop-parlay.csapi.getList.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyIterator_getList_startIndex, {"startIndex","giop-parlay.csapi.getList.startIndex",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyIterator_getList_numberRequested, {"numberRequested","giop-parlay.csapi.getList.numberRequested",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRepository_createRepository_repositoryName, {"repositoryName","giop-parlay.csapi.createRepository.repositoryName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRepository_getRepository_repositoryName, {"repositoryName","giop-parlay.csapi.getRepository.repositoryName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRepository_removeRepository_repositoryName, {"repositoryName","giop-parlay.csapi.removeRepository.repositoryName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRepository_getRepositoryCount_return, {"TpInt32","giop-parlay.csapi.getRepositoryCount.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRepository_createCondition_conditionName, {"conditionName","giop-parlay.csapi.createCondition.conditionName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRepository_createCondition_conditionType, {"conditionType","giop-parlay.csapi.createCondition.conditionType",FT_UINT32,BASE_DEC,VALS(org_csapi_policy_TpPolicyConditionType),0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRepository_createCondition_conditionAttributes_loop, {"Seq length of conditionAttributes","giop-parlay.csapi.createCondition.conditionAttributes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRepository_getCondition_conditionName, {"conditionName","giop-parlay.csapi.getCondition.conditionName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRepository_removeCondition_conditionName, {"conditionName","giop-parlay.csapi.removeCondition.conditionName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRepository_getConditionCount_return, {"TpInt32","giop-parlay.csapi.getConditionCount.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRepository_createAction_actionName, {"actionName","giop-parlay.csapi.createAction.actionName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRepository_createAction_actionType, {"actionType","giop-parlay.csapi.createAction.actionType",FT_UINT32,BASE_DEC,VALS(org_csapi_policy_TpPolicyActionType),0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRepository_createAction_actionAttributes_loop, {"Seq length of actionAttributes","giop-parlay.csapi.createAction.actionAttributes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRepository_getAction_actionName, {"actionName","giop-parlay.csapi.getAction.actionName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRepository_removeAction_actionName, {"actionName","giop-parlay.csapi.removeAction.actionName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRepository_getActionCount_return, {"TpInt32","giop-parlay.csapi.getActionCount.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRule_createCondition_conditionName, {"conditionName","giop-parlay.csapi.createCondition.conditionName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRule_createCondition_conditionType, {"conditionType","giop-parlay.csapi.createCondition.conditionType",FT_UINT32,BASE_DEC,VALS(org_csapi_policy_TpPolicyConditionType),0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRule_createCondition_conditionAttributes_loop, {"Seq length of conditionAttributes","giop-parlay.csapi.createCondition.conditionAttributes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRule_getCondition_conditionName, {"conditionName","giop-parlay.csapi.getCondition.conditionName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRule_removeCondition_conditionName, {"conditionName","giop-parlay.csapi.removeCondition.conditionName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRule_getConditionCount_return, {"TpInt32","giop-parlay.csapi.getConditionCount.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRule_createAction_actionName, {"actionName","giop-parlay.csapi.createAction.actionName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRule_createAction_actionType, {"actionType","giop-parlay.csapi.createAction.actionType",FT_UINT32,BASE_DEC,VALS(org_csapi_policy_TpPolicyActionType),0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRule_createAction_actionAttributes_loop, {"Seq length of actionAttributes","giop-parlay.csapi.createAction.actionAttributes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRule_getAction_actionName, {"actionName","giop-parlay.csapi.getAction.actionName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRule_removeAction_actionName, {"actionName","giop-parlay.csapi.removeAction.actionName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRule_getActionCount_return, {"TpInt32","giop-parlay.csapi.getActionCount.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRule_setValidityPeriodConditionByName_conditionName, {"conditionName","giop-parlay.csapi.setValidityPeriodConditionByName.conditionName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRule_setConditionList_conditionList_loop, {"Seq length of conditionList","giop-parlay.csapi.setConditionList.conditionList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRule_getConditionList_return_loop, {"Seq length of TpPolicyConditionList","giop-parlay.csapi.getConditionList.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRule_setActionList_actionList_loop, {"Seq length of actionList","giop-parlay.csapi.setActionList.actionList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyRule_getActionList_return_loop, {"Seq length of TpPolicyActionList","giop-parlay.csapi.getActionList.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyManager_createDomain_domainName, {"domainName","giop-parlay.csapi.createDomain.domainName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyManager_getDomain_domainName, {"domainName","giop-parlay.csapi.getDomain.domainName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyManager_removeDomain_domainName, {"domainName","giop-parlay.csapi.removeDomain.domainName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyManager_getDomainCount_return, {"TpInt32","giop-parlay.csapi.getDomainCount.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_return_loop, {"Seq length of TpStringSet","giop-parlay.csapi.findMatchingDomains.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_return, {"TpStringSet","giop-parlay.csapi.findMatchingDomains.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyManager_findMatchingDomains_matchingAttributes_loop, {"Seq length of matchingAttributes","giop-parlay.csapi.findMatchingDomains.matchingAttributes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyManager_createRepository_repositoryName, {"repositoryName","giop-parlay.csapi.createRepository.repositoryName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyManager_getRepository_repositoryName, {"repositoryName","giop-parlay.csapi.getRepository.repositoryName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyManager_removeRepository_repositoryName, {"repositoryName","giop-parlay.csapi.removeRepository.repositoryName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyManager_getRepositoryCount_return, {"TpInt32","giop-parlay.csapi.getRepositoryCount.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyManager_commitTransaction_return, {"TpBoolean","giop-parlay.csapi.commitTransaction.return",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyGroup_createGroup_groupName, {"groupName","giop-parlay.csapi.createGroup.groupName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyGroup_getGroup_groupName, {"groupName","giop-parlay.csapi.getGroup.groupName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyGroup_removeGroup_groupName, {"groupName","giop-parlay.csapi.removeGroup.groupName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyGroup_getGroupCount_return, {"TpInt32","giop-parlay.csapi.getGroupCount.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyGroup_createRule_ruleName, {"ruleName","giop-parlay.csapi.createRule.ruleName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyGroup_getRule_ruleName, {"ruleName","giop-parlay.csapi.getRule.ruleName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyGroup_removeRule_ruleName, {"ruleName","giop-parlay.csapi.removeRule.ruleName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyGroup_getRuleCount_return, {"TpInt32","giop-parlay.csapi.getRuleCount.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyEventDefinition_setRequiredAttributes_requiredAttributes_loop, {"Seq length of requiredAttributes","giop-parlay.csapi.setRequiredAttributes.requiredAttributes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyEventDefinition_setOptionalAttributes_optionalAttributes_loop, {"Seq length of optionalAttributes","giop-parlay.csapi.setOptionalAttributes.optionalAttributes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyEventDefinition_getRequiredAttributes_return_loop, {"Seq length of TpAttributeSet","giop-parlay.csapi.getRequiredAttributes.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyEventDefinition_getOptionalAttributes_return_loop, {"Seq length of TpAttributeSet","giop-parlay.csapi.getOptionalAttributes.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicySignature_setInputVariables_inputVariables_loop, {"Seq length of inputVariables","giop-parlay.csapi.setInputVariables.inputVariables.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicySignature_setInputVariables_inputVariables, {"inputVariables","giop-parlay.csapi.setInputVariables.inputVariables",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicySignature_setOutputVariables_outputVariables_loop, {"Seq length of outputVariables","giop-parlay.csapi.setOutputVariables.outputVariables.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicySignature_setOutputVariables_outputVariables, {"outputVariables","giop-parlay.csapi.setOutputVariables.outputVariables",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicySignature_getInputVariables_return_loop, {"Seq length of TpStringSet","giop-parlay.csapi.getInputVariables.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicySignature_getInputVariables_return, {"TpStringSet","giop-parlay.csapi.getInputVariables.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicySignature_getOutputVariables_return_loop, {"Seq length of TpStringSet","giop-parlay.csapi.getOutputVariables.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicySignature_getOutputVariables_return, {"TpStringSet","giop-parlay.csapi.getOutputVariables.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicySignature_setGroupNames_groupNames_loop, {"Seq length of groupNames","giop-parlay.csapi.setGroupNames.groupNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicySignature_setGroupNames_groupNames, {"groupNames","giop-parlay.csapi.setGroupNames.groupNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicySignature_setPolicyRoles_roleNames_loop, {"Seq length of roleNames","giop-parlay.csapi.setPolicyRoles.roleNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicySignature_setPolicyRoles_roleNames, {"roleNames","giop-parlay.csapi.setPolicyRoles.roleNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicySignature_getGroupNames_return_loop, {"Seq length of TpStringSet","giop-parlay.csapi.getGroupNames.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicySignature_getGroupNames_return, {"TpStringSet","giop-parlay.csapi.getGroupNames.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicySignature_getPolicyRoles_return_loop, {"Seq length of TpStringSet","giop-parlay.csapi.getPolicyRoles.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicySignature_getPolicyRoles_return, {"TpStringSet","giop-parlay.csapi.getPolicyRoles.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_createDomain_domainName, {"domainName","giop-parlay.csapi.createDomain.domainName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_getDomain_domainName, {"domainName","giop-parlay.csapi.getDomain.domainName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_removeDomain_domainName, {"domainName","giop-parlay.csapi.removeDomain.domainName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_getDomainCount_return, {"TpInt32","giop-parlay.csapi.getDomainCount.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_createGroup_groupName, {"groupName","giop-parlay.csapi.createGroup.groupName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_getGroup_groupName, {"groupName","giop-parlay.csapi.getGroup.groupName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_removeGroup_groupName, {"groupName","giop-parlay.csapi.removeGroup.groupName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_getGroupCount_return, {"TpInt32","giop-parlay.csapi.getGroupCount.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_createRule_ruleName, {"ruleName","giop-parlay.csapi.createRule.ruleName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_getRule_ruleName, {"ruleName","giop-parlay.csapi.getRule.ruleName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_removeRule_ruleName, {"ruleName","giop-parlay.csapi.removeRule.ruleName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_getRuleCount_return, {"TpInt32","giop-parlay.csapi.getRuleCount.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_eventDefinitionName, {"eventDefinitionName","giop-parlay.csapi.createEventDefinition.eventDefinitionName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_requiredAttributes_loop, {"Seq length of requiredAttributes","giop-parlay.csapi.createEventDefinition.requiredAttributes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_requiredAttributes, {"requiredAttributes","giop-parlay.csapi.createEventDefinition.requiredAttributes",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_optionalAttributes_loop, {"Seq length of optionalAttributes","giop-parlay.csapi.createEventDefinition.optionalAttributes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_createEventDefinition_optionalAttributes, {"optionalAttributes","giop-parlay.csapi.createEventDefinition.optionalAttributes",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_getEventDefinition_eventDefinitionName, {"eventDefinitionName","giop-parlay.csapi.getEventDefinition.eventDefinitionName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_removeEventDefinition_eventDefinitionName, {"eventDefinitionName","giop-parlay.csapi.removeEventDefinition.eventDefinitionName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_getEventDefinitionCount_return, {"TpInt32","giop-parlay.csapi.getEventDefinitionCount.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_createVariableSet_variableSetName, {"variableSetName","giop-parlay.csapi.createVariableSet.variableSetName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariableSet_return_loop, {"Seq length of TpPolicyVarSet","giop-parlay.csapi.getVariableSet.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariableSet_variableSetName, {"variableSetName","giop-parlay.csapi.getVariableSet.variableSetName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_removeVariableSet_variableSetName, {"variableSetName","giop-parlay.csapi.removeVariableSet.variableSetName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariableSetCount_return, {"TpInt32","giop-parlay.csapi.getVariableSetCount.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_createVariable_variableSetName, {"variableSetName","giop-parlay.csapi.createVariable.variableSetName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_createVariable_variableName, {"variableName","giop-parlay.csapi.createVariable.variableName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_setVariableValue_variableSetName, {"variableSetName","giop-parlay.csapi.setVariableValue.variableSetName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_setVariableValue_variableName, {"variableName","giop-parlay.csapi.setVariableValue.variableName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariableType_variableSetName, {"variableSetName","giop-parlay.csapi.getVariableType.variableSetName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariableType_variableName, {"variableName","giop-parlay.csapi.getVariableType.variableName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariableValue_variableSetName, {"variableSetName","giop-parlay.csapi.getVariableValue.variableSetName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariableValue_variableName, {"variableName","giop-parlay.csapi.getVariableValue.variableName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariable_variableSetName, {"variableSetName","giop-parlay.csapi.getVariable.variableSetName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_getVariable_variableName, {"variableName","giop-parlay.csapi.getVariable.variableName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_removeVariable_variablSetName, {"variablSetName","giop-parlay.csapi.removeVariable.variablSetName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_removeVariable_variableName, {"variableName","giop-parlay.csapi.removeVariable.variableName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_createSignature_signatureName, {"signatureName","giop-parlay.csapi.createSignature.signatureName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_getSignature_signatureName, {"signatureName","giop-parlay.csapi.getSignature.signatureName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_removeSignature_signatureName, {"signatureName","giop-parlay.csapi.removeSignature.signatureName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_provisioning_IpPolicyDomain_getSignatureCount_return, {"TpInt32","giop-parlay.csapi.getSignatureCount.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_identity, {"identity","giop-parlay.csapi.createIdentity.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_identityTypes_loop, {"Seq length of identityTypes","giop-parlay.csapi.createIdentity.identityTypes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_identityTypes, {"identityTypes","giop-parlay.csapi.createIdentity.identityTypes",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.createIdentity.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createIdentity_authToken, {"authToken","giop-parlay.csapi.createIdentity.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteIdentity_identity, {"identity","giop-parlay.csapi.deleteIdentity.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteIdentity_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.deleteIdentity.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteIdentity_authToken, {"authToken","giop-parlay.csapi.deleteIdentity.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isIdentity_return, {"TpBoolean","giop-parlay.csapi.isIdentity.return",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isIdentity_identity, {"identity","giop-parlay.csapi.isIdentity.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isIdentity_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.isIdentity.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isIdentity_authToken, {"authToken","giop-parlay.csapi.isIdentity.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_identity, {"identity","giop-parlay.csapi.createGroupIdentity.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_identityTypes_loop, {"Seq length of identityTypes","giop-parlay.csapi.createGroupIdentity.identityTypes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_identityTypes, {"identityTypes","giop-parlay.csapi.createGroupIdentity.identityTypes",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.createGroupIdentity.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_createGroupIdentity_authToken, {"authToken","giop-parlay.csapi.createGroupIdentity.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteGroupIdentity_identity, {"identity","giop-parlay.csapi.deleteGroupIdentity.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteGroupIdentity_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.deleteGroupIdentity.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_deleteGroupIdentity_authToken, {"authToken","giop-parlay.csapi.deleteGroupIdentity.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addToGroup_group, {"group","giop-parlay.csapi.addToGroup.group",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addToGroup_member, {"member","giop-parlay.csapi.addToGroup.member",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addToGroup_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.addToGroup.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addToGroup_authToken, {"authToken","giop-parlay.csapi.addToGroup.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeFromGroup_group, {"group","giop-parlay.csapi.removeFromGroup.group",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeFromGroup_identity, {"identity","giop-parlay.csapi.removeFromGroup.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeFromGroup_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.removeFromGroup.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeFromGroup_authToken, {"authToken","giop-parlay.csapi.removeFromGroup.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_return_loop, {"Seq length of TpPAMFQNameList","giop-parlay.csapi.listMembers.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_return, {"TpPAMFQNameList","giop-parlay.csapi.listMembers.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_identity, {"identity","giop-parlay.csapi.listMembers.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.listMembers.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listMembers_authToken, {"authToken","giop-parlay.csapi.listMembers.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isGroupIdentity_return, {"TpBoolean","giop-parlay.csapi.isGroupIdentity.return",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isGroupIdentity_identity, {"identity","giop-parlay.csapi.isGroupIdentity.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isGroupIdentity_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.isGroupIdentity.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_isGroupIdentity_authToken, {"authToken","giop-parlay.csapi.isGroupIdentity.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_return_loop, {"Seq length of TpPAMFQNameList","giop-parlay.csapi.listGroupMembership.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_return, {"TpPAMFQNameList","giop-parlay.csapi.listGroupMembership.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_identity, {"identity","giop-parlay.csapi.listGroupMembership.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.listGroupMembership.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listGroupMembership_authToken, {"authToken","giop-parlay.csapi.listGroupMembership.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addAlias_identity, {"identity","giop-parlay.csapi.addAlias.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addAlias_alias, {"alias","giop-parlay.csapi.addAlias.alias",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addAlias_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.addAlias.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_addAlias_authToken, {"authToken","giop-parlay.csapi.addAlias.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeAliases_identity, {"identity","giop-parlay.csapi.removeAliases.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeAliases_alias, {"alias","giop-parlay.csapi.removeAliases.alias",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeAliases_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.removeAliases.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_removeAliases_authToken, {"authToken","giop-parlay.csapi.removeAliases.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_return_loop, {"Seq length of TpPAMFQNameList","giop-parlay.csapi.listAliases.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_return, {"TpPAMFQNameList","giop-parlay.csapi.listAliases.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_identity, {"identity","giop-parlay.csapi.listAliases.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.listAliases.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listAliases_authToken, {"authToken","giop-parlay.csapi.listAliases.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_lookupByAlias_return, {"TpPAMFQName","giop-parlay.csapi.lookupByAlias.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_lookupByAlias_alias, {"alias","giop-parlay.csapi.lookupByAlias.alias",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_lookupByAlias_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.lookupByAlias.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_lookupByAlias_authToken, {"authToken","giop-parlay.csapi.lookupByAlias.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_identity, {"identity","giop-parlay.csapi.associateTypes.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_identityTypes_loop, {"Seq length of identityTypes","giop-parlay.csapi.associateTypes.identityTypes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_identityTypes, {"identityTypes","giop-parlay.csapi.associateTypes.identityTypes",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.associateTypes.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_associateTypes_authToken, {"authToken","giop-parlay.csapi.associateTypes.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_identity, {"identity","giop-parlay.csapi.disassociateTypes.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_identityTypes_loop, {"Seq length of identityTypes","giop-parlay.csapi.disassociateTypes.identityTypes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_identityTypes, {"identityTypes","giop-parlay.csapi.disassociateTypes.identityTypes",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.disassociateTypes.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_disassociateTypes_authToken, {"authToken","giop-parlay.csapi.disassociateTypes.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_return_loop, {"Seq length of TpPAMFQNameList","giop-parlay.csapi.listTypesOfIdentity.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_return, {"TpPAMFQNameList","giop-parlay.csapi.listTypesOfIdentity.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_identity, {"identity","giop-parlay.csapi.listTypesOfIdentity.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.listTypesOfIdentity.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_listTypesOfIdentity_authToken, {"authToken","giop-parlay.csapi.listTypesOfIdentity.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType_return, {"TpBoolean","giop-parlay.csapi.hasType.return",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType_identity, {"identity","giop-parlay.csapi.hasType.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType_typeName, {"typeName","giop-parlay.csapi.hasType.typeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.hasType.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_hasType_authToken, {"authToken","giop-parlay.csapi.hasType.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_return_loop, {"Seq length of TpPAMAttributeList","giop-parlay.csapi.getIdentityAttributes.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_identity, {"identity","giop-parlay.csapi.getIdentityAttributes.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_identityType, {"identityType","giop-parlay.csapi.getIdentityAttributes.identityType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_attributeNames_loop, {"Seq length of attributeNames","giop-parlay.csapi.getIdentityAttributes.attributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_attributeNames, {"attributeNames","giop-parlay.csapi.getIdentityAttributes.attributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.getIdentityAttributes.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_getIdentityAttributes_authToken, {"authToken","giop-parlay.csapi.getIdentityAttributes.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_identity, {"identity","giop-parlay.csapi.setIdentityAttributes.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_identityType, {"identityType","giop-parlay.csapi.setIdentityAttributes.identityType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_attributes_loop, {"Seq length of attributes","giop-parlay.csapi.setIdentityAttributes.attributes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.setIdentityAttributes.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityManagement_setIdentityAttributes_authToken, {"authToken","giop-parlay.csapi.setIdentityAttributes.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_agentName, {"agentName","giop-parlay.csapi.createAgent.agentName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_agentTypes_loop, {"Seq length of agentTypes","giop-parlay.csapi.createAgent.agentTypes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_agentTypes, {"agentTypes","giop-parlay.csapi.createAgent.agentTypes",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.createAgent.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_createAgent_authToken, {"authToken","giop-parlay.csapi.createAgent.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_deleteAgent_agentName, {"agentName","giop-parlay.csapi.deleteAgent.agentName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_deleteAgent_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.deleteAgent.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_deleteAgent_authToken, {"authToken","giop-parlay.csapi.deleteAgent.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isAgent_return, {"TpBoolean","giop-parlay.csapi.isAgent.return",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isAgent_agentName, {"agentName","giop-parlay.csapi.isAgent.agentName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isAgent_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.isAgent.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isAgent_authToken, {"authToken","giop-parlay.csapi.isAgent.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_agentName, {"agentName","giop-parlay.csapi.enableCapabilities.agentName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_capabilities_loop, {"Seq length of capabilities","giop-parlay.csapi.enableCapabilities.capabilities.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_capabilities, {"capabilities","giop-parlay.csapi.enableCapabilities.capabilities",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.enableCapabilities.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_enableCapabilities_authToken, {"authToken","giop-parlay.csapi.enableCapabilities.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_agentName, {"agentName","giop-parlay.csapi.disableCapabilities.agentName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_capabilities_loop, {"Seq length of capabilities","giop-parlay.csapi.disableCapabilities.capabilities.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_capabilities, {"capabilities","giop-parlay.csapi.disableCapabilities.capabilities",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.disableCapabilities.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disableCapabilities_authToken, {"authToken","giop-parlay.csapi.disableCapabilities.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_return_loop, {"Seq length of TpPAMCapabilityList","giop-parlay.csapi.listEnabledCapabilities.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_return, {"TpPAMCapabilityList","giop-parlay.csapi.listEnabledCapabilities.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_agentName, {"agentName","giop-parlay.csapi.listEnabledCapabilities.agentName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.listEnabledCapabilities.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listEnabledCapabilities_authToken, {"authToken","giop-parlay.csapi.listEnabledCapabilities.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_return_loop, {"Seq length of TpPAMCapabilityList","giop-parlay.csapi.listAllCapabilities.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_return, {"TpPAMCapabilityList","giop-parlay.csapi.listAllCapabilities.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_agentName, {"agentName","giop-parlay.csapi.listAllCapabilities.agentName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.listAllCapabilities.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listAllCapabilities_authToken, {"authToken","giop-parlay.csapi.listAllCapabilities.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf_return, {"TpBoolean","giop-parlay.csapi.isCapableOf.return",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf_agentName, {"agentName","giop-parlay.csapi.isCapableOf.agentName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf_capability, {"capability","giop-parlay.csapi.isCapableOf.capability",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.isCapableOf.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_isCapableOf_authToken, {"authToken","giop-parlay.csapi.isCapableOf.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_agentName, {"agentName","giop-parlay.csapi.associateTypes.agentName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_agentTypes_loop, {"Seq length of agentTypes","giop-parlay.csapi.associateTypes.agentTypes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_agentTypes, {"agentTypes","giop-parlay.csapi.associateTypes.agentTypes",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.associateTypes.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_associateTypes_authToken, {"authToken","giop-parlay.csapi.associateTypes.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_agentName, {"agentName","giop-parlay.csapi.disassociateTypes.agentName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_agentTypes_loop, {"Seq length of agentTypes","giop-parlay.csapi.disassociateTypes.agentTypes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_agentTypes, {"agentTypes","giop-parlay.csapi.disassociateTypes.agentTypes",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.disassociateTypes.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_disassociateTypes_authToken, {"authToken","giop-parlay.csapi.disassociateTypes.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_return_loop, {"Seq length of TpStringList","giop-parlay.csapi.listTypesOfAgent.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_return, {"TpStringList","giop-parlay.csapi.listTypesOfAgent.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_agentName, {"agentName","giop-parlay.csapi.listTypesOfAgent.agentName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.listTypesOfAgent.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_listTypesOfAgent_authToken, {"authToken","giop-parlay.csapi.listTypesOfAgent.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType_return, {"TpBoolean","giop-parlay.csapi.hasType.return",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType_agentName, {"agentName","giop-parlay.csapi.hasType.agentName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType_typeName, {"typeName","giop-parlay.csapi.hasType.typeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.hasType.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_hasType_authToken, {"authToken","giop-parlay.csapi.hasType.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_return_loop, {"Seq length of TpPAMAttributeList","giop-parlay.csapi.getAgentAttributes.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_agentName, {"agentName","giop-parlay.csapi.getAgentAttributes.agentName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_agentType, {"agentType","giop-parlay.csapi.getAgentAttributes.agentType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_attributeNames_loop, {"Seq length of attributeNames","giop-parlay.csapi.getAgentAttributes.attributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_attributeNames, {"attributeNames","giop-parlay.csapi.getAgentAttributes.attributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.getAgentAttributes.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_getAgentAttributes_authToken, {"authToken","giop-parlay.csapi.getAgentAttributes.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_agentName, {"agentName","giop-parlay.csapi.setAgentAttributes.agentName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_agentType, {"agentType","giop-parlay.csapi.setAgentAttributes.agentType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_attributes_loop, {"Seq length of attributes","giop-parlay.csapi.setAgentAttributes.attributes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.setAgentAttributes.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentManagement_setAgentAttributes_authToken, {"authToken","giop-parlay.csapi.setAgentAttributes.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_assignAgent_identity, {"identity","giop-parlay.csapi.assignAgent.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_assignAgent_agentName, {"agentName","giop-parlay.csapi.assignAgent.agentName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_assignAgent_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.assignAgent.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_assignAgent_authToken, {"authToken","giop-parlay.csapi.assignAgent.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_unassignAgent_identity, {"identity","giop-parlay.csapi.unassignAgent.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_unassignAgent_agentName, {"agentName","giop-parlay.csapi.unassignAgent.agentName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_unassignAgent_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.unassignAgent.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_unassignAgent_authToken, {"authToken","giop-parlay.csapi.unassignAgent.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_return_loop, {"Seq length of TpPAMFQNameList","giop-parlay.csapi.listAssignedAgents.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_return, {"TpPAMFQNameList","giop-parlay.csapi.listAssignedAgents.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_identity, {"identity","giop-parlay.csapi.listAssignedAgents.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.listAssignedAgents.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgents_authToken, {"authToken","giop-parlay.csapi.listAssignedAgents.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_return_loop, {"Seq length of TpPAMFQNameList","giop-parlay.csapi.listAssociatedIdentitiesOfAgent.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_return, {"TpPAMFQNameList","giop-parlay.csapi.listAssociatedIdentitiesOfAgent.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_agentName, {"agentName","giop-parlay.csapi.listAssociatedIdentitiesOfAgent.agentName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.listAssociatedIdentitiesOfAgent.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssociatedIdentitiesOfAgent_authToken, {"authToken","giop-parlay.csapi.listAssociatedIdentitiesOfAgent.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_return_loop, {"Seq length of TpPAMFQNameList","giop-parlay.csapi.listAssignedAgentsByCapability.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_return, {"TpPAMFQNameList","giop-parlay.csapi.listAssignedAgentsByCapability.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_identity, {"identity","giop-parlay.csapi.listAssignedAgentsByCapability.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_capability, {"capability","giop-parlay.csapi.listAssignedAgentsByCapability.capability",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.listAssignedAgentsByCapability.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listAssignedAgentsByCapability_authToken, {"authToken","giop-parlay.csapi.listAssignedAgentsByCapability.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_return_loop, {"Seq length of TpPAMCapabilityList","giop-parlay.csapi.listCapabilitiesOfIdentity.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_return, {"TpPAMCapabilityList","giop-parlay.csapi.listCapabilitiesOfIdentity.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_identity, {"identity","giop-parlay.csapi.listCapabilitiesOfIdentity.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.listCapabilitiesOfIdentity.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_listCapabilitiesOfIdentity_authToken, {"authToken","giop-parlay.csapi.listCapabilitiesOfIdentity.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf_return, {"TpBoolean","giop-parlay.csapi.isIdentityCapableOf.return",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf_identity, {"identity","giop-parlay.csapi.isIdentityCapableOf.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf_capability, {"capability","giop-parlay.csapi.isIdentityCapableOf.capability",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.isIdentityCapableOf.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentAssignment_isIdentityCapableOf_authToken, {"authToken","giop-parlay.csapi.isIdentityCapableOf.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityAttribute_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.createIdentityAttribute.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityAttribute_authToken, {"authToken","giop-parlay.csapi.createIdentityAttribute.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityAttribute_attributeName, {"attributeName","giop-parlay.csapi.deleteIdentityAttribute.attributeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityAttribute_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.deleteIdentityAttribute.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityAttribute_authToken, {"authToken","giop-parlay.csapi.deleteIdentityAttribute.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_getIdentityAttributeDefinition_attributeName, {"attributeName","giop-parlay.csapi.getIdentityAttributeDefinition.attributeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_getIdentityAttributeDefinition_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.getIdentityAttributeDefinition.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_getIdentityAttributeDefinition_authToken, {"authToken","giop-parlay.csapi.getIdentityAttributeDefinition.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_return_loop, {"Seq length of TpStringList","giop-parlay.csapi.listAllIdentityAttributes.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_return, {"TpStringList","giop-parlay.csapi.listAllIdentityAttributes.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.listAllIdentityAttributes.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listAllIdentityAttributes_authToken, {"authToken","giop-parlay.csapi.listAllIdentityAttributes.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_typeName, {"typeName","giop-parlay.csapi.createIdentityType.typeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_attributeNames_loop, {"Seq length of attributeNames","giop-parlay.csapi.createIdentityType.attributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_attributeNames, {"attributeNames","giop-parlay.csapi.createIdentityType.attributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.createIdentityType.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_createIdentityType_authToken, {"authToken","giop-parlay.csapi.createIdentityType.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityType_typeName, {"typeName","giop-parlay.csapi.deleteIdentityType.typeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityType_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.deleteIdentityType.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_deleteIdentityType_authToken, {"authToken","giop-parlay.csapi.deleteIdentityType.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_return_loop, {"Seq length of TpStringList","giop-parlay.csapi.listIdentityTypes.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_return, {"TpStringList","giop-parlay.csapi.listIdentityTypes.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.listIdentityTypes.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypes_authToken, {"authToken","giop-parlay.csapi.listIdentityTypes.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_typeName, {"typeName","giop-parlay.csapi.addIdentityTypeAttributes.typeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_attributeNames_loop, {"Seq length of attributeNames","giop-parlay.csapi.addIdentityTypeAttributes.attributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_attributeNames, {"attributeNames","giop-parlay.csapi.addIdentityTypeAttributes.attributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.addIdentityTypeAttributes.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_addIdentityTypeAttributes_authToken, {"authToken","giop-parlay.csapi.addIdentityTypeAttributes.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_typeName, {"typeName","giop-parlay.csapi.removeIdentityTypeAttributes.typeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_attributeNames_loop, {"Seq length of attributeNames","giop-parlay.csapi.removeIdentityTypeAttributes.attributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_attributeNames, {"attributeNames","giop-parlay.csapi.removeIdentityTypeAttributes.attributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.removeIdentityTypeAttributes.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_removeIdentityTypeAttributes_authToken, {"authToken","giop-parlay.csapi.removeIdentityTypeAttributes.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_return_loop, {"Seq length of TpStringList","giop-parlay.csapi.listIdentityTypeAttributes.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_return, {"TpStringList","giop-parlay.csapi.listIdentityTypeAttributes.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_typeName, {"typeName","giop-parlay.csapi.listIdentityTypeAttributes.typeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.listIdentityTypeAttributes.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMIdentityTypeManagement_listIdentityTypeAttributes_authToken, {"authToken","giop-parlay.csapi.listIdentityTypeAttributes.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentAttribute_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.createAgentAttribute.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentAttribute_authToken, {"authToken","giop-parlay.csapi.createAgentAttribute.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentAttribute_attributeName, {"attributeName","giop-parlay.csapi.deleteAgentAttribute.attributeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentAttribute_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.deleteAgentAttribute.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentAttribute_authToken, {"authToken","giop-parlay.csapi.deleteAgentAttribute.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_getAgentAttributeDefinition_attributeName, {"attributeName","giop-parlay.csapi.getAgentAttributeDefinition.attributeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_getAgentAttributeDefinition_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.getAgentAttributeDefinition.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_getAgentAttributeDefinition_authToken, {"authToken","giop-parlay.csapi.getAgentAttributeDefinition.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_return_loop, {"Seq length of TpStringList","giop-parlay.csapi.listAllAgentAttributes.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_return, {"TpStringList","giop-parlay.csapi.listAllAgentAttributes.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.listAllAgentAttributes.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAllAgentAttributes_authToken, {"authToken","giop-parlay.csapi.listAllAgentAttributes.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_typeName, {"typeName","giop-parlay.csapi.createAgentType.typeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_attributeNames_loop, {"Seq length of attributeNames","giop-parlay.csapi.createAgentType.attributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_attributeNames, {"attributeNames","giop-parlay.csapi.createAgentType.attributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.createAgentType.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_createAgentType_authToken, {"authToken","giop-parlay.csapi.createAgentType.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentType_typeName, {"typeName","giop-parlay.csapi.deleteAgentType.typeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentType_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.deleteAgentType.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_deleteAgentType_authToken, {"authToken","giop-parlay.csapi.deleteAgentType.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_return_loop, {"Seq length of TpStringList","giop-parlay.csapi.listAgentTypes.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_return, {"TpStringList","giop-parlay.csapi.listAgentTypes.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.listAgentTypes.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypes_authToken, {"authToken","giop-parlay.csapi.listAgentTypes.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_typeName, {"typeName","giop-parlay.csapi.addAgentTypeAttributes.typeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_attributeNames_loop, {"Seq length of attributeNames","giop-parlay.csapi.addAgentTypeAttributes.attributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_attributeNames, {"attributeNames","giop-parlay.csapi.addAgentTypeAttributes.attributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.addAgentTypeAttributes.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_addAgentTypeAttributes_authToken, {"authToken","giop-parlay.csapi.addAgentTypeAttributes.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_typeName, {"typeName","giop-parlay.csapi.removeAgentTypeAttributes.typeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_attributeNames_loop, {"Seq length of attributeNames","giop-parlay.csapi.removeAgentTypeAttributes.attributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_attributeNames, {"attributeNames","giop-parlay.csapi.removeAgentTypeAttributes.attributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.removeAgentTypeAttributes.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_removeAgentTypeAttributes_authToken, {"authToken","giop-parlay.csapi.removeAgentTypeAttributes.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_return_loop, {"Seq length of TpStringList","giop-parlay.csapi.listAgentTypeAttributes.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_return, {"TpStringList","giop-parlay.csapi.listAgentTypeAttributes.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_typeName, {"typeName","giop-parlay.csapi.listAgentTypeAttributes.typeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.listAgentTypeAttributes.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMAgentTypeManagement_listAgentTypeAttributes_authToken, {"authToken","giop-parlay.csapi.listAgentTypeAttributes.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapabilityAttribute_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.createCapabilityAttribute.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapabilityAttribute_authToken, {"authToken","giop-parlay.csapi.createCapabilityAttribute.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapabilityAttribute_attributeName, {"attributeName","giop-parlay.csapi.deleteCapabilityAttribute.attributeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapabilityAttribute_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.deleteCapabilityAttribute.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapabilityAttribute_authToken, {"authToken","giop-parlay.csapi.deleteCapabilityAttribute.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_getCapabilityAttributeDefinition_attributeName, {"attributeName","giop-parlay.csapi.getCapabilityAttributeDefinition.attributeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_getCapabilityAttributeDefinition_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.getCapabilityAttributeDefinition.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_getCapabilityAttributeDefinition_authToken, {"authToken","giop-parlay.csapi.getCapabilityAttributeDefinition.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_return_loop, {"Seq length of TpStringList","giop-parlay.csapi.listAllCapabilityAttributes.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_return, {"TpStringList","giop-parlay.csapi.listAllCapabilityAttributes.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.listAllCapabilityAttributes.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listAllCapabilityAttributes_authToken, {"authToken","giop-parlay.csapi.listAllCapabilityAttributes.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_capabilityName, {"capabilityName","giop-parlay.csapi.createCapability.capabilityName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_attributeNames_loop, {"Seq length of attributeNames","giop-parlay.csapi.createCapability.attributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_attributeNames, {"attributeNames","giop-parlay.csapi.createCapability.attributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.createCapability.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_createCapability_authToken, {"authToken","giop-parlay.csapi.createCapability.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapability_capabilityName, {"capabilityName","giop-parlay.csapi.deleteCapability.capabilityName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapability_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.deleteCapability.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_deleteCapability_authToken, {"authToken","giop-parlay.csapi.deleteCapability.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_return_loop, {"Seq length of TpPAMCapabilityList","giop-parlay.csapi.listCapabilities.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_return, {"TpPAMCapabilityList","giop-parlay.csapi.listCapabilities.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.listCapabilities.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilities_authToken, {"authToken","giop-parlay.csapi.listCapabilities.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_capabilityName, {"capabilityName","giop-parlay.csapi.addCapabilityAttributes.capabilityName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_attributeNames_loop, {"Seq length of attributeNames","giop-parlay.csapi.addCapabilityAttributes.attributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_attributeNames, {"attributeNames","giop-parlay.csapi.addCapabilityAttributes.attributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.addCapabilityAttributes.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_addCapabilityAttributes_authToken, {"authToken","giop-parlay.csapi.addCapabilityAttributes.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_capabilityName, {"capabilityName","giop-parlay.csapi.removeCapabilityAttributes.capabilityName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_attributeNames_loop, {"Seq length of attributeNames","giop-parlay.csapi.removeCapabilityAttributes.attributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_attributeNames, {"attributeNames","giop-parlay.csapi.removeCapabilityAttributes.attributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.removeCapabilityAttributes.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_removeCapabilityAttributes_authToken, {"authToken","giop-parlay.csapi.removeCapabilityAttributes.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_return_loop, {"Seq length of TpStringList","giop-parlay.csapi.listCapabilityAttributes.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_return, {"TpStringList","giop-parlay.csapi.listCapabilityAttributes.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_capabilityName, {"capabilityName","giop-parlay.csapi.listCapabilityAttributes.capabilityName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.listCapabilityAttributes.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilityAttributes_authToken, {"authToken","giop-parlay.csapi.listCapabilityAttributes.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_agentType, {"agentType","giop-parlay.csapi.assignCapabilitiesToType.agentType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_capabilities_loop, {"Seq length of capabilities","giop-parlay.csapi.assignCapabilitiesToType.capabilities.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_capabilities, {"capabilities","giop-parlay.csapi.assignCapabilitiesToType.capabilities",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.assignCapabilitiesToType.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_assignCapabilitiesToType_authToken, {"authToken","giop-parlay.csapi.assignCapabilitiesToType.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_agentType, {"agentType","giop-parlay.csapi.unassignCapabilitiesFromType.agentType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_capabilities_loop, {"Seq length of capabilities","giop-parlay.csapi.unassignCapabilitiesFromType.capabilities.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_capabilities, {"capabilities","giop-parlay.csapi.unassignCapabilitiesFromType.capabilities",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.unassignCapabilitiesFromType.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_unassignCapabilitiesFromType_authToken, {"authToken","giop-parlay.csapi.unassignCapabilitiesFromType.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_return_loop, {"Seq length of TpPAMCapabilityList","giop-parlay.csapi.listCapabilitiesOfType.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_return, {"TpPAMCapabilityList","giop-parlay.csapi.listCapabilitiesOfType.return",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_agentType, {"agentType","giop-parlay.csapi.listCapabilitiesOfType.agentType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.listCapabilitiesOfType.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMCapabilityManagement_listCapabilitiesOfType_authToken, {"authToken","giop-parlay.csapi.listCapabilitiesOfType.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_return_loop, {"Seq length of TpPAMCredential","giop-parlay.csapi.getAuthToken.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_return, {"TpPAMCredential","giop-parlay.csapi.getAuthToken.return",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAuthToken_askerData_loop, {"Seq length of askerData","giop-parlay.csapi.getAuthToken.askerData.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_obtainInterface_interfaceName, {"interfaceName","giop-parlay.csapi.obtainInterface.interfaceName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAccessControl_identity, {"identity","giop-parlay.csapi.getAccessControl.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAccessControl_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.getAccessControl.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_getAccessControl_authToken, {"authToken","giop-parlay.csapi.getAccessControl.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_setAccessControl_identity, {"identity","giop-parlay.csapi.setAccessControl.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_setAccessControl_operation, {"operation","giop-parlay.csapi.setAccessControl.operation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_setAccessControl_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.setAccessControl.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_provisioning_IpPAMProvisioningManager_setAccessControl_authToken, {"authToken","giop-parlay.csapi.setAccessControl.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_identity, {"identity","giop-parlay.csapi.setIdentityPresence.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_identityType, {"identityType","giop-parlay.csapi.setIdentityPresence.identityType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_attributes_loop, {"Seq length of attributes","giop-parlay.csapi.setIdentityPresence.attributes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.setIdentityPresence.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresence_authToken, {"authToken","giop-parlay.csapi.setIdentityPresence.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_identity, {"identity","giop-parlay.csapi.setIdentityPresenceExpiration.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_identityType, {"identityType","giop-parlay.csapi.setIdentityPresenceExpiration.identityType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_attributeNames_loop, {"Seq length of attributeNames","giop-parlay.csapi.setIdentityPresenceExpiration.attributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_attributeNames, {"attributeNames","giop-parlay.csapi.setIdentityPresenceExpiration.attributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_expiresIn, {"expiresIn","giop-parlay.csapi.setIdentityPresenceExpiration.expiresIn",FT_INT64,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.setIdentityPresenceExpiration.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMIdentityPresence_setIdentityPresenceExpiration_authToken, {"authToken","giop-parlay.csapi.setIdentityPresenceExpiration.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_return_loop, {"Seq length of TpPAMAttributeList","giop-parlay.csapi.getIdentityPresence.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_identity, {"identity","giop-parlay.csapi.getIdentityPresence.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_identityType, {"identityType","giop-parlay.csapi.getIdentityPresence.identityType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_attributeNames_loop, {"Seq length of attributeNames","giop-parlay.csapi.getIdentityPresence.attributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_attributeNames, {"attributeNames","giop-parlay.csapi.getIdentityPresence.attributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.getIdentityPresence.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMIdentityPresence_getIdentityPresence_authToken, {"authToken","giop-parlay.csapi.getIdentityPresence.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_agent, {"agent","giop-parlay.csapi.setAgentPresence.agent",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_agentType, {"agentType","giop-parlay.csapi.setAgentPresence.agentType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_attributes_loop, {"Seq length of attributes","giop-parlay.csapi.setAgentPresence.attributes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.setAgentPresence.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresence_authToken, {"authToken","giop-parlay.csapi.setAgentPresence.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_agent, {"agent","giop-parlay.csapi.setCapabilityPresence.agent",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_capability, {"capability","giop-parlay.csapi.setCapabilityPresence.capability",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_attributes_loop, {"Seq length of attributes","giop-parlay.csapi.setCapabilityPresence.attributes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.setCapabilityPresence.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresence_authToken, {"authToken","giop-parlay.csapi.setCapabilityPresence.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_agent, {"agent","giop-parlay.csapi.setAgentPresenceExpiration.agent",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_agentType, {"agentType","giop-parlay.csapi.setAgentPresenceExpiration.agentType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_attributeNames_loop, {"Seq length of attributeNames","giop-parlay.csapi.setAgentPresenceExpiration.attributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_attributeNames, {"attributeNames","giop-parlay.csapi.setAgentPresenceExpiration.attributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_expiresIn, {"expiresIn","giop-parlay.csapi.setAgentPresenceExpiration.expiresIn",FT_INT64,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.setAgentPresenceExpiration.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setAgentPresenceExpiration_authToken, {"authToken","giop-parlay.csapi.setAgentPresenceExpiration.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_agent, {"agent","giop-parlay.csapi.setCapabilityPresenceExpiration.agent",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_capability, {"capability","giop-parlay.csapi.setCapabilityPresenceExpiration.capability",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_attributeNames_loop, {"Seq length of attributeNames","giop-parlay.csapi.setCapabilityPresenceExpiration.attributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_attributeNames, {"attributeNames","giop-parlay.csapi.setCapabilityPresenceExpiration.attributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_expiresIn, {"expiresIn","giop-parlay.csapi.setCapabilityPresenceExpiration.expiresIn",FT_INT64,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.setCapabilityPresenceExpiration.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_setCapabilityPresenceExpiration_authToken, {"authToken","giop-parlay.csapi.setCapabilityPresenceExpiration.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_return_loop, {"Seq length of TpPAMAttributeList","giop-parlay.csapi.getAgentPresence.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_agent, {"agent","giop-parlay.csapi.getAgentPresence.agent",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_agentType, {"agentType","giop-parlay.csapi.getAgentPresence.agentType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_attributeNames_loop, {"Seq length of attributeNames","giop-parlay.csapi.getAgentPresence.attributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_attributeNames, {"attributeNames","giop-parlay.csapi.getAgentPresence.attributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.getAgentPresence.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_getAgentPresence_authToken, {"authToken","giop-parlay.csapi.getAgentPresence.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_return_loop, {"Seq length of TpPAMAttributeList","giop-parlay.csapi.getCapabilityPresence.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_agent, {"agent","giop-parlay.csapi.getCapabilityPresence.agent",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_capability, {"capability","giop-parlay.csapi.getCapabilityPresence.capability",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_attributeNames_loop, {"Seq length of attributeNames","giop-parlay.csapi.getCapabilityPresence.attributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_attributeNames, {"attributeNames","giop-parlay.csapi.getCapabilityPresence.attributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.getCapabilityPresence.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAgentPresence_getCapabilityPresence_authToken, {"authToken","giop-parlay.csapi.getCapabilityPresence.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_return_loop, {"Seq length of TpPAMAvailabilityProfileList","giop-parlay.csapi.computeAvailability.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_identity, {"identity","giop-parlay.csapi.computeAvailability.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_attributeNames_loop, {"Seq length of attributeNames","giop-parlay.csapi.computeAvailability.attributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_attributeNames, {"attributeNames","giop-parlay.csapi.computeAvailability.attributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.computeAvailability.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpAppPAMPreferenceCheck_computeAvailability_authToken, {"authToken","giop-parlay.csapi.computeAvailability.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAvailability_getAvailability_return_loop, {"Seq length of TpPAMAvailabilityProfileList","giop-parlay.csapi.getAvailability.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAvailability_getAvailability_identity, {"identity","giop-parlay.csapi.getAvailability.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAvailability_getAvailability_attributeNames_loop, {"Seq length of attributeNames","giop-parlay.csapi.getAvailability.attributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAvailability_getAvailability_attributeNames, {"attributeNames","giop-parlay.csapi.getAvailability.attributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAvailability_getAvailability_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.getAvailability.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAvailability_getAvailability_authToken, {"authToken","giop-parlay.csapi.getAvailability.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAvailability_getPreference_identity, {"identity","giop-parlay.csapi.getPreference.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAvailability_getPreference_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.getPreference.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAvailability_getPreference_authToken, {"authToken","giop-parlay.csapi.getPreference.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAvailability_setPreference_identity, {"identity","giop-parlay.csapi.setPreference.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAvailability_setPreference_operation, {"operation","giop-parlay.csapi.setPreference.operation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAvailability_setPreference_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.setPreference.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMAvailability_setPreference_authToken, {"authToken","giop-parlay.csapi.setPreference.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_return_loop, {"Seq length of TpPAMCredential","giop-parlay.csapi.getAuthToken.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_return, {"TpPAMCredential","giop-parlay.csapi.getAuthToken.return",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAuthToken_askerData_loop, {"Seq length of askerData","giop-parlay.csapi.getAuthToken.askerData.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_obtainInterface_interfaceName, {"interfaceName","giop-parlay.csapi.obtainInterface.interfaceName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAccessControl_identity, {"identity","giop-parlay.csapi.getAccessControl.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAccessControl_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.getAccessControl.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_getAccessControl_authToken, {"authToken","giop-parlay.csapi.getAccessControl.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_setAccessControl_identity, {"identity","giop-parlay.csapi.setAccessControl.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_setAccessControl_operation, {"operation","giop-parlay.csapi.setAccessControl.operation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_setAccessControl_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.setAccessControl.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_setAccessControl_authToken, {"authToken","giop-parlay.csapi.setAccessControl.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_activateService_identity, {"identity","giop-parlay.csapi.activateService.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_activateService_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.activateService.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_activateService_authToken, {"authToken","giop-parlay.csapi.activateService.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_deactivateService_identity, {"identity","giop-parlay.csapi.deactivateService.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_deactivateService_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.deactivateService.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_deactivateService_authToken, {"authToken","giop-parlay.csapi.deactivateService.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_isActiveIdentity_return, {"TpBoolean","giop-parlay.csapi.isActiveIdentity.return",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_isActiveIdentity_identity, {"identity","giop-parlay.csapi.isActiveIdentity.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_isActiveIdentity_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.isActiveIdentity.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_access_IpPAMPresenceAvailabilityManager_isActiveIdentity_authToken, {"authToken","giop-parlay.csapi.isActiveIdentity.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpAppPAMEventHandler_eventNotify_eventID, {"eventID","giop-parlay.csapi.eventNotify.eventID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpAppPAMEventHandler_eventNotify_eventInfo_loop, {"Seq length of eventInfo","giop-parlay.csapi.eventNotify.eventInfo.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpAppPAMEventHandler_eventNotifyErr_eventID, {"eventID","giop-parlay.csapi.eventNotifyErr.eventID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventHandler_isRegistered_return, {"TpBoolean","giop-parlay.csapi.isRegistered.return",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventHandler_isRegistered_clientID, {"clientID","giop-parlay.csapi.isRegistered.clientID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventHandler_isRegistered_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.isRegistered.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventHandler_isRegistered_authToken, {"authToken","giop-parlay.csapi.isRegistered.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventHandler_registerAppInterface_return, {"TpPAMClientID","giop-parlay.csapi.registerAppInterface.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventHandler_registerAppInterface_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.registerAppInterface.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventHandler_registerAppInterface_authToken, {"authToken","giop-parlay.csapi.registerAppInterface.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_return, {"TpPAMEventID","giop-parlay.csapi.registerForEvent.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_clientID, {"clientID","giop-parlay.csapi.registerForEvent.clientID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_eventList_loop, {"Seq length of eventList","giop-parlay.csapi.registerForEvent.eventList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_validFor, {"validFor","giop-parlay.csapi.registerForEvent.validFor",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.registerForEvent.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventHandler_registerForEvent_authToken, {"authToken","giop-parlay.csapi.registerForEvent.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventHandler_deregisterAppInterface_clientID, {"clientID","giop-parlay.csapi.deregisterAppInterface.clientID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventHandler_deregisterAppInterface_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.deregisterAppInterface.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventHandler_deregisterAppInterface_authToken, {"authToken","giop-parlay.csapi.deregisterAppInterface.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventHandler_deregisterFromEvent_eventID, {"eventID","giop-parlay.csapi.deregisterFromEvent.eventID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventHandler_deregisterFromEvent_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.deregisterFromEvent.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventHandler_deregisterFromEvent_authToken, {"authToken","giop-parlay.csapi.deregisterFromEvent.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventManager_getAuthToken_return_loop, {"Seq length of TpPAMCredential","giop-parlay.csapi.getAuthToken.return.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventManager_getAuthToken_return, {"TpPAMCredential","giop-parlay.csapi.getAuthToken.return",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventManager_getAuthToken_askerData_loop, {"Seq length of askerData","giop-parlay.csapi.getAuthToken.askerData.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventManager_obtainInterface_interfaceName, {"interfaceName","giop-parlay.csapi.obtainInterface.interfaceName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventManager_getAccessControl_identity, {"identity","giop-parlay.csapi.getAccessControl.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventManager_getAccessControl_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.getAccessControl.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventManager_getAccessControl_authToken, {"authToken","giop-parlay.csapi.getAccessControl.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventManager_setAccessControl_identity, {"identity","giop-parlay.csapi.setAccessControl.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventManager_setAccessControl_operation, {"operation","giop-parlay.csapi.setAccessControl.operation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventManager_setAccessControl_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.setAccessControl.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventManager_setAccessControl_authToken, {"authToken","giop-parlay.csapi.setAccessControl.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventManager_activateService_identity, {"identity","giop-parlay.csapi.activateService.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventManager_activateService_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.activateService.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventManager_activateService_authToken, {"authToken","giop-parlay.csapi.activateService.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventManager_deactivateService_identity, {"identity","giop-parlay.csapi.deactivateService.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventManager_deactivateService_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.deactivateService.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventManager_deactivateService_authToken, {"authToken","giop-parlay.csapi.deactivateService.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventManager_isActiveIdentity_return, {"TpBoolean","giop-parlay.csapi.isActiveIdentity.return",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventManager_isActiveIdentity_identity, {"identity","giop-parlay.csapi.isActiveIdentity.identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventManager_isActiveIdentity_authToken_loop, {"Seq length of authToken","giop-parlay.csapi.isActiveIdentity.authToken.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_event_IpPAMEventManager_isActiveIdentity_authToken, {"authToken","giop-parlay.csapi.isActiveIdentity.authToken",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessagingManager_reportNotification_assignmentID, {"assignmentID","giop-parlay.csapi.reportNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessagingManager_reportNotification_eventInfo_loop, {"Seq length of eventInfo","giop-parlay.csapi.reportNotification.eventInfo.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMailboxes_mailboxSet_loop, {"Seq length of mailboxSet","giop-parlay.csapi.terminateMultipleMailboxes.mailboxSet.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessagingManager_terminateMultipleMultiMediaMessagingSessions_multiMediaMessagingSet_loop, {"Seq length of multiMediaMessagingSet","giop-parlay.csapi.terminateMultipleMultiMediaMessagingSessions.multiMediaMessagingSet.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMultiMediaMessagingManager_openMailbox_mailboxID, {"mailboxID","giop-parlay.csapi.openMailbox.mailboxID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMultiMediaMessagingManager_openMailbox_authenticationInfo, {"authenticationInfo","giop-parlay.csapi.openMailbox.authenticationInfo",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMultiMediaMessagingManager_createNotification_return, {"TpAssignmentID","giop-parlay.csapi.createNotification.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMultiMediaMessagingManager_createNotification_eventCriteria_loop, {"Seq length of eventCriteria","giop-parlay.csapi.createNotification.eventCriteria.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMultiMediaMessagingManager_destroyNotification_assignmentID, {"assignmentID","giop-parlay.csapi.destroyNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMultiMediaMessagingManager_changeNotification_assignmentID, {"assignmentID","giop-parlay.csapi.changeNotification.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMultiMediaMessagingManager_changeNotification_eventCriteria_loop, {"Seq length of eventCriteria","giop-parlay.csapi.changeNotification.eventCriteria.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMultiMediaMessagingManager_getNextNotification_reset, {"reset","giop-parlay.csapi.getNextNotification.reset",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMultiMediaMessagingManager_enableNotifications_return, {"TpAssignmentID","giop-parlay.csapi.enableNotifications.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_createFolderRes_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.createFolderRes.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_createFolderRes_requestID, {"requestID","giop-parlay.csapi.createFolderRes.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_createFolderRes_folderID, {"folderID","giop-parlay.csapi.createFolderRes.folderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_createFolderErr_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.createFolderErr.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_createFolderErr_requestID, {"requestID","giop-parlay.csapi.createFolderErr.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_createFolderErr_error, {"error","giop-parlay.csapi.createFolderErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_createFolderErr_errorDetails, {"errorDetails","giop-parlay.csapi.createFolderErr.errorDetails",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFoldersRes_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getFoldersRes.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFoldersRes_requestID, {"requestID","giop-parlay.csapi.getFoldersRes.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFoldersRes_folderID, {"folderID","giop-parlay.csapi.getFoldersRes.folderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFoldersRes_folderNames_loop, {"Seq length of folderNames","giop-parlay.csapi.getFoldersRes.folderNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFoldersRes_folderNames, {"folderNames","giop-parlay.csapi.getFoldersRes.folderNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFoldersErr_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getFoldersErr.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFoldersErr_requestID, {"requestID","giop-parlay.csapi.getFoldersErr.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFoldersErr_folderID, {"folderID","giop-parlay.csapi.getFoldersErr.folderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFoldersErr_error, {"error","giop-parlay.csapi.getFoldersErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFoldersErr_errorDetails, {"errorDetails","giop-parlay.csapi.getFoldersErr.errorDetails",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_deleteFolderRes_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.deleteFolderRes.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_deleteFolderRes_requestID, {"requestID","giop-parlay.csapi.deleteFolderRes.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_deleteFolderErr_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.deleteFolderErr.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_deleteFolderErr_requestID, {"requestID","giop-parlay.csapi.deleteFolderErr.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_deleteFolderErr_error, {"error","giop-parlay.csapi.deleteFolderErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_deleteFolderErr_errorDetails, {"errorDetails","giop-parlay.csapi.deleteFolderErr.errorDetails",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_copyFolderRes_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.copyFolderRes.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_copyFolderRes_requestID, {"requestID","giop-parlay.csapi.copyFolderRes.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_copyFolderErr_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.copyFolderErr.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_copyFolderErr_requestID, {"requestID","giop-parlay.csapi.copyFolderErr.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_copyFolderErr_error, {"error","giop-parlay.csapi.copyFolderErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_copyFolderErr_errorDetails, {"errorDetails","giop-parlay.csapi.copyFolderErr.errorDetails",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_moveFolderRes_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.moveFolderRes.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_moveFolderRes_requestID, {"requestID","giop-parlay.csapi.moveFolderRes.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_moveFolderErr_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.moveFolderErr.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_moveFolderErr_requestID, {"requestID","giop-parlay.csapi.moveFolderErr.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_moveFolderErr_error, {"error","giop-parlay.csapi.moveFolderErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_moveFolderErr_errorDetails, {"errorDetails","giop-parlay.csapi.moveFolderErr.errorDetails",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_putMessageRes_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.putMessageRes.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_putMessageRes_requestID, {"requestID","giop-parlay.csapi.putMessageRes.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_putMessageRes_messageID, {"messageID","giop-parlay.csapi.putMessageRes.messageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_putMessageErr_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.putMessageErr.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_putMessageErr_requestID, {"requestID","giop-parlay.csapi.putMessageErr.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_putMessageErr_error, {"error","giop-parlay.csapi.putMessageErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_putMessageErr_errorDetails, {"errorDetails","giop-parlay.csapi.putMessageErr.errorDetails",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_copyMessageRes_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.copyMessageRes.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_copyMessageRes_requestID, {"requestID","giop-parlay.csapi.copyMessageRes.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_copyMessageErr_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.copyMessageErr.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_copyMessageErr_requestID, {"requestID","giop-parlay.csapi.copyMessageErr.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_copyMessageErr_error, {"error","giop-parlay.csapi.copyMessageErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_copyMessageErr_errorDetails, {"errorDetails","giop-parlay.csapi.copyMessageErr.errorDetails",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_moveMessageRes_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.moveMessageRes.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_moveMessageRes_requestID, {"requestID","giop-parlay.csapi.moveMessageRes.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_moveMessageErr_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.moveMessageErr.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_moveMessageErr_requestID, {"requestID","giop-parlay.csapi.moveMessageErr.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_moveMessageErr_error, {"error","giop-parlay.csapi.moveMessageErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_moveMessageErr_errorDetails, {"errorDetails","giop-parlay.csapi.moveMessageErr.errorDetails",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_deleteMessageRes_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.deleteMessageRes.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_deleteMessageRes_requestID, {"requestID","giop-parlay.csapi.deleteMessageRes.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_deleteMessageErr_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.deleteMessageErr.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_deleteMessageErr_requestID, {"requestID","giop-parlay.csapi.deleteMessageErr.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_deleteMessageErr_error, {"error","giop-parlay.csapi.deleteMessageErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_deleteMessageErr_errorDetails, {"errorDetails","giop-parlay.csapi.deleteMessageErr.errorDetails",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_listMessagesRes_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.listMessagesRes.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_listMessagesRes_requestID, {"requestID","giop-parlay.csapi.listMessagesRes.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_listMessagesRes_messageList_loop, {"Seq length of messageList","giop-parlay.csapi.listMessagesRes.messageList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_listMessagesRes_final, {"final","giop-parlay.csapi.listMessagesRes.final",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_listMessagesErr_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.listMessagesErr.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_listMessagesErr_requestID, {"requestID","giop-parlay.csapi.listMessagesErr.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_listMessagesErr_error, {"error","giop-parlay.csapi.listMessagesErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_listMessagesErr_errorDetails, {"errorDetails","giop-parlay.csapi.listMessagesErr.errorDetails",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsRes_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.listMessageBodyPartsRes.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsRes_requestID, {"requestID","giop-parlay.csapi.listMessageBodyPartsRes.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsRes_partsList_loop, {"Seq length of partsList","giop-parlay.csapi.listMessageBodyPartsRes.partsList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsErr_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.listMessageBodyPartsErr.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsErr_requestID, {"requestID","giop-parlay.csapi.listMessageBodyPartsErr.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsErr_error, {"error","giop-parlay.csapi.listMessageBodyPartsErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_listMessageBodyPartsErr_errorDetails, {"errorDetails","giop-parlay.csapi.listMessageBodyPartsErr.errorDetails",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsRes_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getMessageBodyPartsRes.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsRes_requestID, {"requestID","giop-parlay.csapi.getMessageBodyPartsRes.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsRes_bodyParts_loop, {"Seq length of bodyParts","giop-parlay.csapi.getMessageBodyPartsRes.bodyParts.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsErr_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getMessageBodyPartsErr.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsErr_requestID, {"requestID","giop-parlay.csapi.getMessageBodyPartsErr.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsErr_error, {"error","giop-parlay.csapi.getMessageBodyPartsErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageBodyPartsErr_errorDetails, {"errorDetails","giop-parlay.csapi.getMessageBodyPartsErr.errorDetails",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageHeadersRes_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getMessageHeadersRes.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageHeadersRes_requestID, {"requestID","giop-parlay.csapi.getMessageHeadersRes.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageHeadersRes_headers_loop, {"Seq length of headers","giop-parlay.csapi.getMessageHeadersRes.headers.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageHeadersErr_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getMessageHeadersErr.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageHeadersErr_requestID, {"requestID","giop-parlay.csapi.getMessageHeadersErr.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageHeadersErr_error, {"error","giop-parlay.csapi.getMessageHeadersErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageHeadersErr_errorDetails, {"errorDetails","giop-parlay.csapi.getMessageHeadersErr.errorDetails",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageContentRes_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getMessageContentRes.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageContentRes_requestID, {"requestID","giop-parlay.csapi.getMessageContentRes.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageContentRes_contentType, {"contentType","giop-parlay.csapi.getMessageContentRes.contentType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageContentRes_contentTransferEncoding, {"contentTransferEncoding","giop-parlay.csapi.getMessageContentRes.contentTransferEncoding",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageContentRes_content_loop, {"Seq length of content","giop-parlay.csapi.getMessageContentRes.content.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageContentRes_content, {"content","giop-parlay.csapi.getMessageContentRes.content",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageContentErr_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getMessageContentErr.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageContentErr_requestID, {"requestID","giop-parlay.csapi.getMessageContentErr.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageContentErr_error, {"error","giop-parlay.csapi.getMessageContentErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageContentErr_errorDetails, {"errorDetails","giop-parlay.csapi.getMessageContentErr.errorDetails",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFullMessageRes_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getFullMessageRes.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFullMessageRes_requestID, {"requestID","giop-parlay.csapi.getFullMessageRes.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFullMessageRes_message_loop, {"Seq length of message","giop-parlay.csapi.getFullMessageRes.message.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFullMessageRes_message, {"message","giop-parlay.csapi.getFullMessageRes.message",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFullMessageErr_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getFullMessageErr.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFullMessageErr_requestID, {"requestID","giop-parlay.csapi.getFullMessageErr.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFullMessageErr_error, {"error","giop-parlay.csapi.getFullMessageErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFullMessageErr_errorDetails, {"errorDetails","giop-parlay.csapi.getFullMessageErr.errorDetails",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesRes_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getMailboxInfoPropertiesRes.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesRes_requestID, {"requestID","giop-parlay.csapi.getMailboxInfoPropertiesRes.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesRes_returnedProperties_loop, {"Seq length of returnedProperties","giop-parlay.csapi.getMailboxInfoPropertiesRes.returnedProperties.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesRes_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getFolderInfoPropertiesRes.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesRes_requestID, {"requestID","giop-parlay.csapi.getFolderInfoPropertiesRes.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesRes_folderID, {"folderID","giop-parlay.csapi.getFolderInfoPropertiesRes.folderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesRes_returnedProperties_loop, {"Seq length of returnedProperties","giop-parlay.csapi.getFolderInfoPropertiesRes.returnedProperties.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesRes_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getMessageInfoPropertiesRes.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesRes_requestID, {"requestID","giop-parlay.csapi.getMessageInfoPropertiesRes.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesRes_messageID, {"messageID","giop-parlay.csapi.getMessageInfoPropertiesRes.messageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesRes_returnedProperties_loop, {"Seq length of returnedProperties","giop-parlay.csapi.getMessageInfoPropertiesRes.returnedProperties.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesRes_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.setMessageInfoPropertiesRes.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesRes_requestID, {"requestID","giop-parlay.csapi.setMessageInfoPropertiesRes.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesRes_messageID, {"messageID","giop-parlay.csapi.setMessageInfoPropertiesRes.messageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesRes_propertiesUpdated_loop, {"Seq length of propertiesUpdated","giop-parlay.csapi.setMessageInfoPropertiesRes.propertiesUpdated.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesErr_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.setMessageInfoPropertiesErr.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesErr_requestID, {"requestID","giop-parlay.csapi.setMessageInfoPropertiesErr.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesErr_messageID, {"messageID","giop-parlay.csapi.setMessageInfoPropertiesErr.messageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_setMessageInfoPropertiesErr_propertiesNotUpdated_loop, {"Seq length of propertiesNotUpdated","giop-parlay.csapi.setMessageInfoPropertiesErr.propertiesNotUpdated.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesErr_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getMailboxInfoPropertiesErr.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesErr_requestID, {"requestID","giop-parlay.csapi.getMailboxInfoPropertiesErr.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesErr_error, {"error","giop-parlay.csapi.getMailboxInfoPropertiesErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMailboxInfoPropertiesErr_errorDetails, {"errorDetails","giop-parlay.csapi.getMailboxInfoPropertiesErr.errorDetails",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesErr_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getFolderInfoPropertiesErr.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesErr_requestID, {"requestID","giop-parlay.csapi.getFolderInfoPropertiesErr.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesErr_error, {"error","giop-parlay.csapi.getFolderInfoPropertiesErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getFolderInfoPropertiesErr_errorDetails, {"errorDetails","giop-parlay.csapi.getFolderInfoPropertiesErr.errorDetails",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesErr_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getMessageInfoPropertiesErr.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesErr_requestID, {"requestID","giop-parlay.csapi.getMessageInfoPropertiesErr.requestID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesErr_error, {"error","giop-parlay.csapi.getMessageInfoPropertiesErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMailbox_getMessageInfoPropertiesErr_errorDetails, {"errorDetails","giop-parlay.csapi.getMessageInfoPropertiesErr.errorDetails",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_close_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.close.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_createFolderReq_return, {"TpAssignmentID","giop-parlay.csapi.createFolderReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_createFolderReq_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.createFolderReq.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_createFolderReq_folderID, {"folderID","giop-parlay.csapi.createFolderReq.folderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getFoldersReq_return, {"TpAssignmentID","giop-parlay.csapi.getFoldersReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getFoldersReq_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getFoldersReq.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getFoldersReq_folderID, {"folderID","giop-parlay.csapi.getFoldersReq.folderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_deleteFolderReq_return, {"TpAssignmentID","giop-parlay.csapi.deleteFolderReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_deleteFolderReq_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.deleteFolderReq.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_deleteFolderReq_folderID, {"folderID","giop-parlay.csapi.deleteFolderReq.folderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_copyFolderReq_return, {"TpAssignmentID","giop-parlay.csapi.copyFolderReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_copyFolderReq_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.copyFolderReq.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_copyFolderReq_sourceFolderID, {"sourceFolderID","giop-parlay.csapi.copyFolderReq.sourceFolderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_copyFolderReq_destinationFolderID, {"destinationFolderID","giop-parlay.csapi.copyFolderReq.destinationFolderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_moveFolderReq_return, {"TpAssignmentID","giop-parlay.csapi.moveFolderReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_moveFolderReq_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.moveFolderReq.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_moveFolderReq_sourceFolderID, {"sourceFolderID","giop-parlay.csapi.moveFolderReq.sourceFolderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_moveFolderReq_destinationFolderID, {"destinationFolderID","giop-parlay.csapi.moveFolderReq.destinationFolderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_putMessageReq_return, {"TpAssignmentID","giop-parlay.csapi.putMessageReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_putMessageReq_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.putMessageReq.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_putMessageReq_folderID, {"folderID","giop-parlay.csapi.putMessageReq.folderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_putMessageReq_message_loop, {"Seq length of message","giop-parlay.csapi.putMessageReq.message.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_putMessageReq_message, {"message","giop-parlay.csapi.putMessageReq.message",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_copyMessageReq_return, {"TpAssignmentID","giop-parlay.csapi.copyMessageReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_copyMessageReq_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.copyMessageReq.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_copyMessageReq_fromFolderID, {"fromFolderID","giop-parlay.csapi.copyMessageReq.fromFolderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_copyMessageReq_toFolderID, {"toFolderID","giop-parlay.csapi.copyMessageReq.toFolderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_copyMessageReq_messageID, {"messageID","giop-parlay.csapi.copyMessageReq.messageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_moveMessageReq_return, {"TpAssignmentID","giop-parlay.csapi.moveMessageReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_moveMessageReq_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.moveMessageReq.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_moveMessageReq_fromFolderID, {"fromFolderID","giop-parlay.csapi.moveMessageReq.fromFolderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_moveMessageReq_toFolderID, {"toFolderID","giop-parlay.csapi.moveMessageReq.toFolderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_moveMessageReq_messageID, {"messageID","giop-parlay.csapi.moveMessageReq.messageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_deleteMessageReq_return, {"TpAssignmentID","giop-parlay.csapi.deleteMessageReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_deleteMessageReq_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.deleteMessageReq.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_deleteMessageReq_fromFolderID, {"fromFolderID","giop-parlay.csapi.deleteMessageReq.fromFolderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_deleteMessageReq_messageID, {"messageID","giop-parlay.csapi.deleteMessageReq.messageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_listMessagesReq_return, {"TpAssignmentID","giop-parlay.csapi.listMessagesReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_listMessagesReq_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.listMessagesReq.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_listMessagesReq_folderID, {"folderID","giop-parlay.csapi.listMessagesReq.folderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_listMessagesReq_reset, {"reset","giop-parlay.csapi.listMessagesReq.reset",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_listMessageBodyPartsReq_return, {"TpAssignmentID","giop-parlay.csapi.listMessageBodyPartsReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_listMessageBodyPartsReq_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.listMessageBodyPartsReq.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_listMessageBodyPartsReq_folderID, {"folderID","giop-parlay.csapi.listMessageBodyPartsReq.folderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_listMessageBodyPartsReq_messageID, {"messageID","giop-parlay.csapi.listMessageBodyPartsReq.messageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_listMessageBodyPartsReq_maxNestingLevel, {"maxNestingLevel","giop-parlay.csapi.listMessageBodyPartsReq.maxNestingLevel",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_return, {"TpAssignmentID","giop-parlay.csapi.getMessageBodyPartsReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getMessageBodyPartsReq.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_folderID, {"folderID","giop-parlay.csapi.getMessageBodyPartsReq.folderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_messageID, {"messageID","giop-parlay.csapi.getMessageBodyPartsReq.messageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_partIDs_loop, {"Seq length of partIDs","giop-parlay.csapi.getMessageBodyPartsReq.partIDs.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getMessageBodyPartsReq_partIDs, {"partIDs","giop-parlay.csapi.getMessageBodyPartsReq.partIDs",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getMessageHeadersReq_return, {"TpAssignmentID","giop-parlay.csapi.getMessageHeadersReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getMessageHeadersReq_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getMessageHeadersReq.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getMessageHeadersReq_folderID, {"folderID","giop-parlay.csapi.getMessageHeadersReq.folderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getMessageHeadersReq_messageID, {"messageID","giop-parlay.csapi.getMessageHeadersReq.messageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getMessageContentReq_return, {"TpAssignmentID","giop-parlay.csapi.getMessageContentReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getMessageContentReq_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getMessageContentReq.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getMessageContentReq_folderID, {"folderID","giop-parlay.csapi.getMessageContentReq.folderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getMessageContentReq_messageID, {"messageID","giop-parlay.csapi.getMessageContentReq.messageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getFullMessageReq_return, {"TpAssignmentID","giop-parlay.csapi.getFullMessageReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getFullMessageReq_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getFullMessageReq.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getFullMessageReq_folderID, {"folderID","giop-parlay.csapi.getFullMessageReq.folderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getFullMessageReq_messageID, {"messageID","giop-parlay.csapi.getFullMessageReq.messageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getMailboxInfoPropertiesReq_return, {"TpAssignmentID","giop-parlay.csapi.getMailboxInfoPropertiesReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getMailboxInfoPropertiesReq_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getMailboxInfoPropertiesReq.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getFolderInfoPropertiesReq_return, {"TpAssignmentID","giop-parlay.csapi.getFolderInfoPropertiesReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getFolderInfoPropertiesReq_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getFolderInfoPropertiesReq.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getFolderInfoPropertiesReq_folderID, {"folderID","giop-parlay.csapi.getFolderInfoPropertiesReq.folderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getMessageInfoPropertiesReq_return, {"TpAssignmentID","giop-parlay.csapi.getMessageInfoPropertiesReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getMessageInfoPropertiesReq_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.getMessageInfoPropertiesReq.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_getMessageInfoPropertiesReq_messageID, {"messageID","giop-parlay.csapi.getMessageInfoPropertiesReq.messageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_setMessageInfoPropertiesReq_return, {"TpAssignmentID","giop-parlay.csapi.setMessageInfoPropertiesReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_setMessageInfoPropertiesReq_mailboxSessionID, {"mailboxSessionID","giop-parlay.csapi.setMessageInfoPropertiesReq.mailboxSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_setMessageInfoPropertiesReq_messageID, {"messageID","giop-parlay.csapi.setMessageInfoPropertiesReq.messageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMailbox_setMessageInfoPropertiesReq_properties_loop, {"Seq length of properties","giop-parlay.csapi.setMessageInfoPropertiesReq.properties.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_sendMessageRes_sessionID, {"sessionID","giop-parlay.csapi.sendMessageRes.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_sendMessageRes_assignmentID, {"assignmentID","giop-parlay.csapi.sendMessageRes.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_sendMessageErr_sessionID, {"sessionID","giop-parlay.csapi.sendMessageErr.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_sendMessageErr_assignmentID, {"assignmentID","giop-parlay.csapi.sendMessageErr.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_sendMessageErr_error, {"error","giop-parlay.csapi.sendMessageErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_sendMessageErr_errorDetails, {"errorDetails","giop-parlay.csapi.sendMessageErr.errorDetails",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_cancelMessageRes_sessionID, {"sessionID","giop-parlay.csapi.cancelMessageRes.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_cancelMessageRes_assignmentID, {"assignmentID","giop-parlay.csapi.cancelMessageRes.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_cancelMessageErr_sessionID, {"sessionID","giop-parlay.csapi.cancelMessageErr.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_cancelMessageErr_assignmentID, {"assignmentID","giop-parlay.csapi.cancelMessageErr.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_cancelMessageErr_error, {"error","giop-parlay.csapi.cancelMessageErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_cancelMessageErr_errorDetails, {"errorDetails","giop-parlay.csapi.cancelMessageErr.errorDetails",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusRes_sessionID, {"sessionID","giop-parlay.csapi.queryStatusRes.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusRes_assignmentID, {"assignmentID","giop-parlay.csapi.queryStatusRes.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusRes_result_loop, {"Seq length of result","giop-parlay.csapi.queryStatusRes.result.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusErr_sessionID, {"sessionID","giop-parlay.csapi.queryStatusErr.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusErr_assignmentID, {"assignmentID","giop-parlay.csapi.queryStatusErr.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusErr_error, {"error","giop-parlay.csapi.queryStatusErr.error",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagingError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_queryStatusErr_errorDetails, {"errorDetails","giop-parlay.csapi.queryStatusErr.errorDetails",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageStatusReport_sessionID, {"sessionID","giop-parlay.csapi.messageStatusReport.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageStatusReport_assignmentID, {"assignmentID","giop-parlay.csapi.messageStatusReport.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageStatusReport_deliveryReportType, {"deliveryReportType","giop-parlay.csapi.messageStatusReport.deliveryReportType",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageStatusReport_deliveryReportInfo, {"deliveryReportInfo","giop-parlay.csapi.messageStatusReport.deliveryReportInfo",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_sessionID, {"sessionID","giop-parlay.csapi.messageReceived.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_message_loop, {"Seq length of message","giop-parlay.csapi.messageReceived.message.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_message, {"message","giop-parlay.csapi.messageReceived.message",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpAppMultiMediaMessaging_messageReceived_headers_loop, {"Seq length of headers","giop-parlay.csapi.messageReceived.headers.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_return, {"TpAssignmentID","giop-parlay.csapi.sendMessageReq.return",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_sessionID, {"sessionID","giop-parlay.csapi.sendMessageReq.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_deliveryType, {"deliveryType","giop-parlay.csapi.sendMessageReq.deliveryType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_messageTreatment_loop, {"Seq length of messageTreatment","giop-parlay.csapi.sendMessageReq.messageTreatment.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_message_loop, {"Seq length of message","giop-parlay.csapi.sendMessageReq.message.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_message, {"message","giop-parlay.csapi.sendMessageReq.message",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMultiMediaMessaging_sendMessageReq_additionalHeaders_loop, {"Seq length of additionalHeaders","giop-parlay.csapi.sendMessageReq.additionalHeaders.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMultiMediaMessaging_cancelMessageReq_sessionID, {"sessionID","giop-parlay.csapi.cancelMessageReq.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMultiMediaMessaging_cancelMessageReq_assignmentID, {"assignmentID","giop-parlay.csapi.cancelMessageReq.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMultiMediaMessaging_queryStatusReq_sessionID, {"sessionID","giop-parlay.csapi.queryStatusReq.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMultiMediaMessaging_queryStatusReq_assignmentID, {"assignmentID","giop-parlay.csapi.queryStatusReq.assignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_IpMultiMediaMessaging_close_sessionID, {"sessionID","giop-parlay.csapi.close.sessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        /* Attribute filters */
        /* Struct filters */
        {&hf_org_csapi_TpAddress_Plan, {"TpAddress_Plan","giop-parlay.TpAddress.Plan",FT_UINT32,BASE_DEC,VALS(org_csapi_TpAddressPlan),0x0,NULL,HFILL}},
        {&hf_org_csapi_TpAddress_AddrString, {"TpAddress_AddrString","giop-parlay.TpAddress.AddrString",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpAddress_Name, {"TpAddress_Name","giop-parlay.TpAddress.Name",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpAddress_Presentation, {"TpAddress_Presentation","giop-parlay.TpAddress.Presentation",FT_UINT32,BASE_DEC,VALS(org_csapi_TpAddressPresentation),0x0,NULL,HFILL}},
        {&hf_org_csapi_TpAddress_Screening, {"TpAddress_Screening","giop-parlay.TpAddress.Screening",FT_UINT32,BASE_DEC,VALS(org_csapi_TpAddressScreening),0x0,NULL,HFILL}},
        {&hf_org_csapi_TpAddress_SubAddressString, {"TpAddress_SubAddressString","giop-parlay.TpAddress.SubAddressString",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpAddressRange_Plan, {"TpAddressRange_Plan","giop-parlay.TpAddressRange.Plan",FT_UINT32,BASE_DEC,VALS(org_csapi_TpAddressPlan),0x0,NULL,HFILL}},
        {&hf_org_csapi_TpAddressRange_AddrString, {"TpAddressRange_AddrString","giop-parlay.TpAddressRange.AddrString",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpAddressRange_Name, {"TpAddressRange_Name","giop-parlay.TpAddressRange.Name",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpAddressRange_SubAddressString, {"TpAddressRange_SubAddressString","giop-parlay.TpAddressRange.SubAddressString",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpCAIElements_UnitsPerInterval, {"TpCAIElements_UnitsPerInterval","giop-parlay.TpCAIElements.UnitsPerInterval",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpCAIElements_SecondsPerTimeInterval, {"TpCAIElements_SecondsPerTimeInterval","giop-parlay.TpCAIElements.SecondsPerTimeInterval",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpCAIElements_ScalingFactor, {"TpCAIElements_ScalingFactor","giop-parlay.TpCAIElements.ScalingFactor",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpCAIElements_UnitIncrement, {"TpCAIElements_UnitIncrement","giop-parlay.TpCAIElements.UnitIncrement",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpCAIElements_UnitsPerDataInterval, {"TpCAIElements_UnitsPerDataInterval","giop-parlay.TpCAIElements.UnitsPerDataInterval",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpCAIElements_SegmentsPerDataInterval, {"TpCAIElements_SegmentsPerDataInterval","giop-parlay.TpCAIElements.SegmentsPerDataInterval",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpCAIElements_InitialSecsPerTimeInterval, {"TpCAIElements_InitialSecsPerTimeInterval","giop-parlay.TpCAIElements.InitialSecsPerTimeInterval",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpChargePerTime_InitialCharge, {"TpChargePerTime_InitialCharge","giop-parlay.TpChargePerTime.InitialCharge",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpChargePerTime_CurrentChargePerMinute, {"TpChargePerTime_CurrentChargePerMinute","giop-parlay.TpChargePerTime.CurrentChargePerMinute",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpChargePerTime_NextChargePerMinute, {"TpChargePerTime_NextChargePerMinute","giop-parlay.TpChargePerTime.NextChargePerMinute",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpAoCInfo_Currency, {"TpAoCInfo_Currency","giop-parlay.TpAoCInfo.Currency",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpTimeInterval_StartTime, {"TpTimeInterval_StartTime","giop-parlay.TpTimeInterval.StartTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpTimeInterval_StopTime, {"TpTimeInterval_StopTime","giop-parlay.TpTimeInterval.StopTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpStructuredAttributeValue_Type, {"TpStructuredAttributeValue_Type","giop-parlay.TpStructuredAttributeValue.Type",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpAttribute_AttributeName, {"TpAttribute_AttributeName","giop-parlay.TpAttribute.AttributeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFaultStats_Fault, {"TpFaultStats_Fault","giop-parlay.TpFaultStats.Fault",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpInterfaceFault),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFaultStats_Occurrences, {"TpFaultStats_Occurrences","giop-parlay.TpFaultStats.Occurrences",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFaultStats_MaxDuration, {"TpFaultStats_MaxDuration","giop-parlay.TpFaultStats.MaxDuration",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFaultStats_TotalDuration, {"TpFaultStats_TotalDuration","giop-parlay.TpFaultStats.TotalDuration",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFaultStats_NumberOfClientsAffected, {"TpFaultStats_NumberOfClientsAffected","giop-parlay.TpFaultStats.NumberOfClientsAffected",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFaultStatsRecord_FaultStatsSet_loop, {"Seq length of TpFaultStatsRecord_FaultStatsSet","giop-parlay.TpFaultStatsRecord.FaultStatsSet.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpLoadPolicy_LoadPolicy, {"TpLoadPolicy_LoadPolicy","giop-parlay.TpLoadPolicy.LoadPolicy",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpLoadStatisticData_LoadValue, {"TpLoadStatisticData_LoadValue","giop-parlay.TpLoadStatisticData.LoadValue",FT_FLOAT,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpLoadStatisticData_LoadLevel, {"TpLoadStatisticData_LoadLevel","giop-parlay.TpLoadStatisticData.LoadLevel",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpLoadLevel),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpLoadThreshold_LoadThreshold, {"TpLoadThreshold_LoadThreshold","giop-parlay.TpLoadThreshold.LoadThreshold",FT_FLOAT,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpLoadInitVal_LoadLevel, {"TpLoadInitVal_LoadLevel","giop-parlay.TpLoadInitVal.LoadLevel",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpLoadLevel),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpProperty_PropertyName, {"TpProperty_PropertyName","giop-parlay.TpProperty.PropertyName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpProperty_PropertyValue, {"TpProperty_PropertyValue","giop-parlay.TpProperty.PropertyValue",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpClientAppDescription_ClientAppID, {"TpClientAppDescription_ClientAppID","giop-parlay.TpClientAppDescription.ClientAppID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpClientAppDescription_ClientAppProperties_loop, {"Seq length of TpClientAppDescription_ClientAppProperties","giop-parlay.TpClientAppDescription.ClientAppProperties.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpClientAppDescription_HasAccessSession, {"TpClientAppDescription_HasAccessSession","giop-parlay.TpClientAppDescription.HasAccessSession",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_fw_TpClientAppDescription_HasServiceInstances, {"TpClientAppDescription_HasServiceInstances","giop-parlay.TpClientAppDescription.HasServiceInstances",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_fw_TpEntOp_EntOpID, {"TpEntOp_EntOpID","giop-parlay.TpEntOp.EntOpID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpEntOp_EntOpProperties_loop, {"Seq length of TpEntOp_EntOpProperties","giop-parlay.TpEntOp.EntOpProperties.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpSag_SagID, {"TpSag_SagID","giop-parlay.TpSag.SagID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpSag_SagDescription, {"TpSag_SagDescription","giop-parlay.TpSag.SagDescription",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceProperty_ServicePropertyName, {"TpServiceProperty_ServicePropertyName","giop-parlay.TpServiceProperty.ServicePropertyName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceProperty_ServicePropertyValueList_loop, {"Seq length of TpServiceProperty_ServicePropertyValueList","giop-parlay.TpServiceProperty.ServicePropertyValueList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceProperty_ServicePropertyValueList, {"TpServiceProperty_ServicePropertyValueList","giop-parlay.TpServiceProperty.ServicePropertyValueList",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceDescription_ServiceTypeName, {"TpServiceDescription_ServiceTypeName","giop-parlay.TpServiceDescription.ServiceTypeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceDescription_ServicePropertyList_loop, {"Seq length of TpServiceDescription_ServicePropertyList","giop-parlay.TpServiceDescription.ServicePropertyList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpService_ServiceID, {"TpService_ServiceID","giop-parlay.TpService.ServiceID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceProfileDescription_ServiceContractID, {"TpServiceProfileDescription_ServiceContractID","giop-parlay.TpServiceProfileDescription.ServiceContractID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceProfileDescription_ServiceStartDate, {"TpServiceProfileDescription_ServiceStartDate","giop-parlay.TpServiceProfileDescription.ServiceStartDate",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceProfileDescription_ServiceEndDate, {"TpServiceProfileDescription_ServiceEndDate","giop-parlay.TpServiceProfileDescription.ServiceEndDate",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceProfileDescription_ServiceTypeName, {"TpServiceProfileDescription_ServiceTypeName","giop-parlay.TpServiceProfileDescription.ServiceTypeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceProfileDescription_ServiceSubscriptionProperties_loop, {"Seq length of TpServiceProfileDescription_ServiceSubscriptionProperties","giop-parlay.TpServiceProfileDescription.ServiceSubscriptionProperties.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceProfileDescription_InUse, {"TpServiceProfileDescription_InUse","giop-parlay.TpServiceProfileDescription.InUse",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceProfileDescription_ServiceID, {"TpServiceProfileDescription_ServiceID","giop-parlay.TpServiceProfileDescription.ServiceID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceTypeProperty_ServicePropertyName, {"TpServiceTypeProperty_ServicePropertyName","giop-parlay.TpServiceTypeProperty.ServicePropertyName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceTypeProperty_ServiceTypePropertyMode, {"TpServiceTypeProperty_ServiceTypePropertyMode","giop-parlay.TpServiceTypeProperty.ServiceTypePropertyMode",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpServiceTypePropertyMode),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceTypeProperty_ServicePropertyTypeName, {"TpServiceTypeProperty_ServicePropertyTypeName","giop-parlay.TpServiceTypeProperty.ServicePropertyTypeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceTypeDescription_ServiceTypePropertyList_loop, {"Seq length of TpServiceTypeDescription_ServiceTypePropertyList","giop-parlay.TpServiceTypeDescription.ServiceTypePropertyList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceTypeDescription_ServiceTypeNameList_loop, {"Seq length of TpServiceTypeDescription_ServiceTypeNameList","giop-parlay.TpServiceTypeDescription.ServiceTypeNameList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceTypeDescription_ServiceTypeNameList, {"TpServiceTypeDescription_ServiceTypeNameList","giop-parlay.TpServiceTypeDescription.ServiceTypeNameList",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceTypeDescription_AvailableOrUnavailable, {"TpServiceTypeDescription_AvailableOrUnavailable","giop-parlay.TpServiceTypeDescription.AvailableOrUnavailable",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_fw_TpSignatureAndServiceMgr_DigitalSignature_loop, {"Seq length of TpSignatureAndServiceMgr_DigitalSignature","giop-parlay.TpSignatureAndServiceMgr.DigitalSignature.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpSignatureAndServiceMgr_DigitalSignature, {"TpSignatureAndServiceMgr_DigitalSignature","giop-parlay.TpSignatureAndServiceMgr.DigitalSignature",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpPerson_PersonName, {"TpPerson_PersonName","giop-parlay.TpPerson.PersonName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpPerson_PostalAddress, {"TpPerson_PostalAddress","giop-parlay.TpPerson.PostalAddress",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpPerson_TelephoneNumber, {"TpPerson_TelephoneNumber","giop-parlay.TpPerson.TelephoneNumber",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpPerson_Email, {"TpPerson_Email","giop-parlay.TpPerson.Email",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpPerson_HomePage, {"TpPerson_HomePage","giop-parlay.TpPerson.HomePage",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpPerson_PersonProperties_loop, {"Seq length of TpPerson_PersonProperties","giop-parlay.TpPerson.PersonProperties.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceContractDescription_ServiceStartDate, {"TpServiceContractDescription_ServiceStartDate","giop-parlay.TpServiceContractDescription.ServiceStartDate",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceContractDescription_ServiceEndDate, {"TpServiceContractDescription_ServiceEndDate","giop-parlay.TpServiceContractDescription.ServiceEndDate",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceContractDescription_ServiceTypeName, {"TpServiceContractDescription_ServiceTypeName","giop-parlay.TpServiceContractDescription.ServiceTypeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceContractDescription_ServiceID, {"TpServiceContractDescription_ServiceID","giop-parlay.TpServiceContractDescription.ServiceID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceContractDescription_ServiceSubscriptionProperties_loop, {"Seq length of TpServiceContractDescription_ServiceSubscriptionProperties","giop-parlay.TpServiceContractDescription.ServiceSubscriptionProperties.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceContractDescription_InUse, {"TpServiceContractDescription_InUse","giop-parlay.TpServiceContractDescription.InUse",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_fw_TpLoadStatistic_TimeStamp, {"TpLoadStatistic_TimeStamp","giop-parlay.TpLoadStatistic.TimeStamp",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceContract_ServiceContractID, {"TpServiceContract_ServiceContractID","giop-parlay.TpServiceContract.ServiceContractID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceProfile_ServiceProfileID, {"TpServiceProfile_ServiceProfileID","giop-parlay.TpServiceProfile.ServiceProfileID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpSagProfilePair_Sag, {"TpSagProfilePair_Sag","giop-parlay.TpSagProfilePair.Sag",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpSagProfilePair_ServiceProfile, {"TpSagProfilePair_ServiceProfile","giop-parlay.TpSagProfilePair.ServiceProfile",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpAddSagMembersConflict_ClientApplication, {"TpAddSagMembersConflict_ClientApplication","giop-parlay.TpAddSagMembersConflict.ClientApplication",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpAddSagMembersConflict_Service, {"TpAddSagMembersConflict_Service","giop-parlay.TpAddSagMembersConflict.Service",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpAssignSagToServiceProfileConflict_ClientApplication, {"TpAssignSagToServiceProfileConflict_ClientApplication","giop-parlay.TpAssignSagToServiceProfileConflict.ClientApplication",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpAssignSagToServiceProfileConflict_Service, {"TpAssignSagToServiceProfileConflict_Service","giop-parlay.TpAssignSagToServiceProfileConflict.Service",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceTypePropertyValue_ServicePropertyName, {"TpServiceTypePropertyValue_ServicePropertyName","giop-parlay.TpServiceTypePropertyValue.ServicePropertyName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceTypePropertyValue_ServiceTypePropertyMode, {"TpServiceTypePropertyValue_ServiceTypePropertyMode","giop-parlay.TpServiceTypePropertyValue.ServiceTypePropertyMode",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpServiceTypePropertyMode),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceTypePropertyValue_ServicePropertyTypeName, {"TpServiceTypePropertyValue_ServicePropertyTypeName","giop-parlay.TpServiceTypePropertyValue.ServicePropertyTypeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceTypePropertyValue_ServicePropertyValueList_loop, {"Seq length of TpServiceTypePropertyValue_ServicePropertyValueList","giop-parlay.TpServiceTypePropertyValue.ServicePropertyValueList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpServiceTypePropertyValue_ServicePropertyValueList, {"TpServiceTypePropertyValue_ServicePropertyValueList","giop-parlay.TpServiceTypePropertyValue.ServicePropertyValueList",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_ServiceType, {"TpFwMigrationServiceAvailableInfo_ServiceType","giop-parlay.TpFwMigrationServiceAvailableInfo.ServiceType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_ServiceID, {"TpFwMigrationServiceAvailableInfo_ServiceID","giop-parlay.TpFwMigrationServiceAvailableInfo.ServiceID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_CompatibleServiceID, {"TpFwMigrationServiceAvailableInfo_CompatibleServiceID","giop-parlay.TpFwMigrationServiceAvailableInfo.CompatibleServiceID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_BackwardCompatibilityLevel, {"TpFwMigrationServiceAvailableInfo_BackwardCompatibilityLevel","giop-parlay.TpFwMigrationServiceAvailableInfo.BackwardCompatibilityLevel",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_MigrationRequired, {"TpFwMigrationServiceAvailableInfo_MigrationRequired","giop-parlay.TpFwMigrationServiceAvailableInfo.MigrationRequired",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_DataMigrated, {"TpFwMigrationServiceAvailableInfo_DataMigrated","giop-parlay.TpFwMigrationServiceAvailableInfo.DataMigrated",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_MigrationDataAndTime, {"TpFwMigrationServiceAvailableInfo_MigrationDataAndTime","giop-parlay.TpFwMigrationServiceAvailableInfo.MigrationDataAndTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwMigrationServiceAvailableInfo_MigrationAdditionalInfo_loop, {"Seq length of TpFwMigrationServiceAvailableInfo_MigrationAdditionalInfo","giop-parlay.TpFwMigrationServiceAvailableInfo.MigrationAdditionalInfo.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwAgreementInfo_ClientApplicationID, {"TpFwAgreementInfo_ClientApplicationID","giop-parlay.TpFwAgreementInfo.ClientApplicationID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwAgreementInfo_ServiceID, {"TpFwAgreementInfo_ServiceID","giop-parlay.TpFwAgreementInfo.ServiceID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwAgreementInfo_ServiceContractID, {"TpFwAgreementInfo_ServiceContractID","giop-parlay.TpFwAgreementInfo.ServiceContractID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwAgreementInfo_ServiceProfileID, {"TpFwAgreementInfo_ServiceProfileID","giop-parlay.TpFwAgreementInfo.ServiceProfileID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIMessageCriteria_EndSequence, {"TpUIMessageCriteria_EndSequence","giop-parlay.TpUIMessageCriteria.EndSequence",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIMessageCriteria_MaxMessageTime, {"TpUIMessageCriteria_MaxMessageTime","giop-parlay.TpUIMessageCriteria.MaxMessageTime",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIMessageCriteria_MaxMessageSize, {"TpUIMessageCriteria_MaxMessageSize","giop-parlay.TpUIMessageCriteria.MaxMessageSize",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIEventInfo_ServiceCode, {"TpUIEventInfo_ServiceCode","giop-parlay.TpUIEventInfo.ServiceCode",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIEventInfo_DataTypeIndication, {"TpUIEventInfo_DataTypeIndication","giop-parlay.TpUIEventInfo.DataTypeIndication",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUIEventInfoDataType),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIEventInfo_DataString, {"TpUIEventInfo_DataString","giop-parlay.TpUIEventInfo.DataString",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIEventCriteria_ServiceCode, {"TpUIEventCriteria_ServiceCode","giop-parlay.TpUIEventCriteria.ServiceCode",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIEventCriteriaResult_AssignmentID, {"TpUIEventCriteriaResult_AssignmentID","giop-parlay.TpUIEventCriteriaResult.AssignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIEventNotificationInfo_ServiceCode, {"TpUIEventNotificationInfo_ServiceCode","giop-parlay.TpUIEventNotificationInfo.ServiceCode",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIEventNotificationInfo_DataTypeIndication, {"TpUIEventNotificationInfo_DataTypeIndication","giop-parlay.TpUIEventNotificationInfo.DataTypeIndication",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUIEventInfoDataType),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIEventNotificationInfo_UIEventData_loop, {"Seq length of TpUIEventNotificationInfo_UIEventData","giop-parlay.TpUIEventNotificationInfo.UIEventData.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIEventNotificationInfo_UIEventData, {"TpUIEventNotificationInfo_UIEventData","giop-parlay.TpUIEventNotificationInfo.UIEventData",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIRecognitionProperty_PropertyName, {"TpUIRecognitionProperty_PropertyName","giop-parlay.TpUIRecognitionProperty.PropertyName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIRecognitionProperty_PropertyValue, {"TpUIRecognitionProperty_PropertyValue","giop-parlay.TpUIRecognitionProperty.PropertyValue",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIRecognitionCriteria_SpeakerID, {"TpUIRecognitionCriteria_SpeakerID","giop-parlay.TpUIRecognitionCriteria.SpeakerID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIRecognitionCriteria_Properties_loop, {"Seq length of TpUIRecognitionCriteria_Properties","giop-parlay.TpUIRecognitionCriteria.Properties.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIRecognitionCriteria_Grammar, {"TpUIRecognitionCriteria_Grammar","giop-parlay.TpUIRecognitionCriteria.Grammar",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUICollectCriteria_MinLength, {"TpUICollectCriteria_MinLength","giop-parlay.TpUICollectCriteria.MinLength",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUICollectCriteria_MaxLength, {"TpUICollectCriteria_MaxLength","giop-parlay.TpUICollectCriteria.MaxLength",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUICollectCriteria_EndSequence, {"TpUICollectCriteria_EndSequence","giop-parlay.TpUICollectCriteria.EndSequence",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUICollectCriteria_StartTimeout, {"TpUICollectCriteria_StartTimeout","giop-parlay.TpUICollectCriteria.StartTimeout",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUICollectCriteria_InterCharTimeout, {"TpUICollectCriteria_InterCharTimeout","giop-parlay.TpUICollectCriteria.InterCharTimeout",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUICollectCriteria_CollectMode, {"TpUICollectCriteria_CollectMode","giop-parlay.TpUICollectCriteria.CollectMode",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUICollectMode),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIWordOverride_Spelling, {"TpUIWordOverride_Spelling","giop-parlay.TpUIWordOverride.Spelling",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIWordOverride_PronounceType, {"TpUIWordOverride_PronounceType","giop-parlay.TpUIWordOverride.PronounceType",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUIPronounceType),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIWordOverride_PronounceAs, {"TpUIWordOverride_PronounceAs","giop-parlay.TpUIWordOverride.PronounceAs",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUISynthesisInfoData_SpeakerGender, {"TpUISynthesisInfoData_SpeakerGender","giop-parlay.TpUISynthesisInfoData.SpeakerGender",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUISynthesisGender),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUISynthesisInfoData_SpeakerAge, {"TpUISynthesisInfoData_SpeakerAge","giop-parlay.TpUISynthesisInfoData.SpeakerAge",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUISynthesisAge),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUISynthesisInfoData_SpeakerRate, {"TpUISynthesisInfoData_SpeakerRate","giop-parlay.TpUISynthesisInfoData.SpeakerRate",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUISynthesisRate),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUISynthesisInfoData_SpeakerRange, {"TpUISynthesisInfoData_SpeakerRange","giop-parlay.TpUISynthesisInfoData.SpeakerRange",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUISynthesisRange),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUISynthesisInfoData_TextData, {"TpUISynthesisInfoData_TextData","giop-parlay.TpUISynthesisInfoData.TextData",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUISynthesisInfoData_WordOverrideSet_loop, {"Seq length of TpUISynthesisInfoData_WordOverrideSet","giop-parlay.TpUISynthesisInfoData.WordOverrideSet.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallError_ErrorTime, {"TpCallError_ErrorTime","giop-parlay.TpCallError.ErrorTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallError_ErrorType, {"TpCallError_ErrorType","giop-parlay.TpCallError.ErrorType",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallErrorType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallEndedReport_CallLegSessionID, {"TpCallEndedReport_CallLegSessionID","giop-parlay.TpCallEndedReport.CallLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallEndedReport_Cause, {"TpCallEndedReport_Cause","giop-parlay.TpCallEndedReport.Cause",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpReleaseCause),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallInfoReport_CallInfoType, {"TpCallInfoReport_CallInfoType","giop-parlay.TpCallInfoReport.CallInfoType",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallInfoReport_CallInitiationStartTime, {"TpCallInfoReport_CallInitiationStartTime","giop-parlay.TpCallInfoReport.CallInitiationStartTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallInfoReport_CallConnectedToResourceTime, {"TpCallInfoReport_CallConnectedToResourceTime","giop-parlay.TpCallInfoReport.CallConnectedToResourceTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallInfoReport_CallConnectedToDestinationTime, {"TpCallInfoReport_CallConnectedToDestinationTime","giop-parlay.TpCallInfoReport.CallConnectedToDestinationTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallInfoReport_CallEndTime, {"TpCallInfoReport_CallEndTime","giop-parlay.TpCallInfoReport.CallEndTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallInfoReport_Cause, {"TpCallInfoReport_Cause","giop-parlay.TpCallInfoReport.Cause",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpReleaseCause),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallServiceCode_CallServiceCodeType, {"TpCallServiceCode_CallServiceCodeType","giop-parlay.TpCallServiceCode.CallServiceCodeType",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallServiceCodeType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallServiceCode_ServiceCodeValue, {"TpCallServiceCode_ServiceCodeValue","giop-parlay.TpCallServiceCode.ServiceCodeValue",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallTreatment_CallTreatmentType, {"TpCallTreatment_CallTreatmentType","giop-parlay.TpCallTreatment.CallTreatmentType",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallTreatmentType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallTreatment_ReleaseCause, {"TpCallTreatment_ReleaseCause","giop-parlay.TpCallTreatment.ReleaseCause",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpReleaseCause),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallChargePlan_ChargeOrderType, {"TpCallChargePlan_ChargeOrderType","giop-parlay.TpCallChargePlan.ChargeOrderType",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallChargeOrderCategory),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallChargePlan_TransparentCharge_loop, {"Seq length of TpCallChargePlan_TransparentCharge","giop-parlay.TpCallChargePlan.TransparentCharge.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallChargePlan_TransparentCharge, {"TpCallChargePlan_TransparentCharge","giop-parlay.TpCallChargePlan.TransparentCharge",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallChargePlan_ChargePlan, {"TpCallChargePlan_ChargePlan","giop-parlay.TpCallChargePlan.ChargePlan",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallChargePlan_AdditionalInfo_loop, {"Seq length of TpCallChargePlan_AdditionalInfo","giop-parlay.TpCallChargePlan.AdditionalInfo.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallChargePlan_AdditionalInfo, {"TpCallChargePlan_AdditionalInfo","giop-parlay.TpCallChargePlan.AdditionalInfo",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallChargePlan_PartyToCharge, {"TpCallChargePlan_PartyToCharge","giop-parlay.TpCallChargePlan.PartyToCharge",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallPartyToChargeType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallEventCriteria_CallEventName, {"TpCallEventCriteria_CallEventName","giop-parlay.TpCallEventCriteria.CallEventName",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallEventCriteria_CallNotificationType, {"TpCallEventCriteria_CallNotificationType","giop-parlay.TpCallEventCriteria.CallNotificationType",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_gccs_TpCallNotificationType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallEventCriteria_MonitorMode, {"TpCallEventCriteria_MonitorMode","giop-parlay.TpCallEventCriteria.MonitorMode",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallMonitorMode),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallEventCriteriaResult_AssignmentID, {"TpCallEventCriteriaResult_AssignmentID","giop-parlay.TpCallEventCriteriaResult.AssignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallReleaseCause_Value, {"TpCallReleaseCause_Value","giop-parlay.TpCallReleaseCause.Value",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallReleaseCause_Location, {"TpCallReleaseCause_Location","giop-parlay.TpCallReleaseCause.Location",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallReport_MonitorMode, {"TpCallReport_MonitorMode","giop-parlay.TpCallReport.MonitorMode",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallMonitorMode),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallReport_CallEventTime, {"TpCallReport_CallEventTime","giop-parlay.TpCallReport.CallEventTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallReport_CallReportType, {"TpCallReport_CallReportType","giop-parlay.TpCallReport.CallReportType",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_gccs_TpCallReportType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallReportRequest_MonitorMode, {"TpCallReportRequest_MonitorMode","giop-parlay.TpCallReportRequest.MonitorMode",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallMonitorMode),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallReportRequest_CallReportType, {"TpCallReportRequest_CallReportType","giop-parlay.TpCallReportRequest.CallReportType",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_gccs_TpCallReportType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallEventInfo_CallAppInfo_loop, {"Seq length of TpCallEventInfo_CallAppInfo","giop-parlay.TpCallEventInfo.CallAppInfo.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallEventInfo_CallEventName, {"TpCallEventInfo_CallEventName","giop-parlay.TpCallEventInfo.CallEventName",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallEventInfo_CallNotificationType, {"TpCallEventInfo_CallNotificationType","giop-parlay.TpCallEventInfo.CallNotificationType",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_gccs_TpCallNotificationType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallEventInfo_MonitorMode, {"TpCallEventInfo_MonitorMode","giop-parlay.TpCallEventInfo.MonitorMode",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallMonitorMode),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallEndedReport_CallLegSessionID, {"TpCallEndedReport_CallLegSessionID","giop-parlay.TpCallEndedReport.CallLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallInfoReport_CallInfoType, {"TpCallInfoReport_CallInfoType","giop-parlay.TpCallInfoReport.CallInfoType",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallInfoReport_CallInitiationStartTime, {"TpCallInfoReport_CallInitiationStartTime","giop-parlay.TpCallInfoReport.CallInitiationStartTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallInfoReport_CallConnectedToResourceTime, {"TpCallInfoReport_CallConnectedToResourceTime","giop-parlay.TpCallInfoReport.CallConnectedToResourceTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallInfoReport_CallConnectedToDestinationTime, {"TpCallInfoReport_CallConnectedToDestinationTime","giop-parlay.TpCallInfoReport.CallConnectedToDestinationTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallInfoReport_CallEndTime, {"TpCallInfoReport_CallEndTime","giop-parlay.TpCallInfoReport.CallEndTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallTreatment_CallTreatmentType, {"TpCallTreatment_CallTreatmentType","giop-parlay.TpCallTreatment.CallTreatmentType",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallTreatmentType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallIdentifier_CallSessionID, {"TpCallIdentifier_CallSessionID","giop-parlay.TpCallIdentifier.CallSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallEventRequest_CallEventType, {"TpCallEventRequest_CallEventType","giop-parlay.TpCallEventRequest.CallEventType",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallEventType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallEventRequest_CallMonitorMode, {"TpCallEventRequest_CallMonitorMode","giop-parlay.TpCallEventRequest.CallMonitorMode",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallMonitorMode),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallNotificationRequest_CallEventsRequested_loop, {"Seq length of TpCallNotificationRequest_CallEventsRequested","giop-parlay.TpCallNotificationRequest.CallEventsRequested.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpNotificationRequested_AssignmentID, {"TpNotificationRequested_AssignmentID","giop-parlay.TpNotificationRequested.AssignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallLegConnectionProperties_AttachMechanism, {"TpCallLegConnectionProperties_AttachMechanism","giop-parlay.TpCallLegConnectionProperties.AttachMechanism",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallLegAttachMechanism),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallEventInfo_CallEventType, {"TpCallEventInfo_CallEventType","giop-parlay.TpCallEventInfo.CallEventType",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallEventType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallEventInfo_CallMonitorMode, {"TpCallEventInfo_CallMonitorMode","giop-parlay.TpCallEventInfo.CallMonitorMode",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallMonitorMode),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallEventInfo_CallEventTime, {"TpCallEventInfo_CallEventTime","giop-parlay.TpCallEventInfo.CallEventTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpNotificationRequestedSetEntry_NotificationRequestSet_loop, {"Seq length of TpNotificationRequestedSetEntry_NotificationRequestSet","giop-parlay.TpNotificationRequestedSetEntry.NotificationRequestSet.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpNotificationRequestedSetEntry_Final, {"TpNotificationRequestedSetEntry_Final","giop-parlay.TpNotificationRequestedSetEntry.Final",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCarrier_CarrierID_loop, {"Seq length of TpCarrier_CarrierID","giop-parlay.TpCarrier.CarrierID.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCarrier_CarrierID, {"TpCarrier_CarrierID","giop-parlay.TpCarrier.CarrierID",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCarrier_CarrierSelectionField, {"TpCarrier_CarrierSelectionField","giop-parlay.TpCarrier.CarrierSelectionField",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCarrierSelectionField),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallNotificationInfo_CallAppInfo_loop, {"Seq length of TpCallNotificationInfo_CallAppInfo","giop-parlay.TpCallNotificationInfo.CallAppInfo.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallLegInfoReport_CallLegInfoType, {"TpCallLegInfoReport_CallLegInfoType","giop-parlay.TpCallLegInfoReport.CallLegInfoType",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallLegInfoReport_CallLegStartTime, {"TpCallLegInfoReport_CallLegStartTime","giop-parlay.TpCallLegInfoReport.CallLegStartTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallLegInfoReport_CallLegConnectedToResourceTime, {"TpCallLegInfoReport_CallLegConnectedToResourceTime","giop-parlay.TpCallLegInfoReport.CallLegConnectedToResourceTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallLegInfoReport_CallLegConnectedToAddressTime, {"TpCallLegInfoReport_CallLegConnectedToAddressTime","giop-parlay.TpCallLegInfoReport.CallLegConnectedToAddressTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallLegInfoReport_CallLegEndTime, {"TpCallLegInfoReport_CallLegEndTime","giop-parlay.TpCallLegInfoReport.CallLegEndTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallLegInfoReport_CallLegReleaseCause, {"TpCallLegInfoReport_CallLegReleaseCause","giop-parlay.TpCallLegInfoReport.CallLegReleaseCause",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpReleaseCause),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallLegInfoReport_CallAppInfo_loop, {"Seq length of TpCallLegInfoReport_CallAppInfo","giop-parlay.TpCallLegInfoReport.CallAppInfo.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallLegProperty_CallLegPropertyName, {"TpCallLegProperty_CallLegPropertyName","giop-parlay.TpCallLegProperty.CallLegPropertyName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallLegProperty_CallLegPropertyValue, {"TpCallLegProperty_CallLegPropertyValue","giop-parlay.TpCallLegProperty.CallLegPropertyValue",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_TpMultiPartyCallIdentifier_CallSessionID, {"TpMultiPartyCallIdentifier_CallSessionID","giop-parlay.TpMultiPartyCallIdentifier.CallSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_TpCallLegIdentifier_CallLegSessionID, {"TpCallLegIdentifier_CallLegSessionID","giop-parlay.TpCallLegIdentifier.CallLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_TpAppCallLegCallBack_AppCallLegSet_loop, {"Seq length of TpAppCallLegCallBack_AppCallLegSet","giop-parlay.TpAppCallLegCallBack.AppCallLegSet.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_TpCallSuperviseVolume_VolumeQuantity, {"TpCallSuperviseVolume_VolumeQuantity","giop-parlay.TpCallSuperviseVolume.VolumeQuantity",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_TpCallSuperviseVolume_VolumeUnit, {"TpCallSuperviseVolume_VolumeUnit","giop-parlay.TpCallSuperviseVolume.VolumeUnit",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_TpMediaStreamRequest_Direction, {"TpMediaStreamRequest_Direction","giop-parlay.TpMediaStreamRequest.Direction",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_mmccs_TpMediaStreamDirection),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_TpMediaStreamRequest_MediaMonitorMode, {"TpMediaStreamRequest_MediaMonitorMode","giop-parlay.TpMediaStreamRequest.MediaMonitorMode",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallMonitorMode),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_TpMediaStreamRequest_EventType, {"TpMediaStreamRequest_EventType","giop-parlay.TpMediaStreamRequest.EventType",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_mmccs_TpMediaStreamEventType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_TpMediaStream_Direction, {"TpMediaStream_Direction","giop-parlay.TpMediaStream.Direction",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_mmccs_TpMediaStreamDirection),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_TpMediaStream_ChannelSessionID, {"TpMediaStream_ChannelSessionID","giop-parlay.TpMediaStream.ChannelSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_TpNotificationMediaRequest_MediaStreamsRequested_loop, {"Seq length of TpNotificationMediaRequest_MediaStreamsRequested","giop-parlay.TpNotificationMediaRequest.MediaStreamsRequested.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_TpMediaNotificationRequested_AssignmentID, {"TpMediaNotificationRequested_AssignmentID","giop-parlay.TpMediaNotificationRequested.AssignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_TpMultiMediaCallIdentifier_MMCallSessionID, {"TpMultiMediaCallIdentifier_MMCallSessionID","giop-parlay.TpMultiMediaCallIdentifier.MMCallSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_TpMultiMediaCallLegIdentifier_MMCallLegSessionID, {"TpMultiMediaCallLegIdentifier_MMCallLegSessionID","giop-parlay.TpMultiMediaCallLegIdentifier.MMCallLegSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_TpAppMultiMediaCallLegCallBack_AppCallLegSet_loop, {"Seq length of TpAppMultiMediaCallLegCallBack_AppCallLegSet","giop-parlay.TpAppMultiMediaCallLegCallBack.AppCallLegSet.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_TpMonoMediaConfPolicy_JoinAllowed, {"TpMonoMediaConfPolicy_JoinAllowed","giop-parlay.TpMonoMediaConfPolicy.JoinAllowed",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_TpJoinEventInfo_CallAppInfo_loop, {"Seq length of TpJoinEventInfo_CallAppInfo","giop-parlay.TpJoinEventInfo.CallAppInfo.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_TpConfSearchCriteria_StartSearch, {"TpConfSearchCriteria_StartSearch","giop-parlay.TpConfSearchCriteria.StartSearch",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_TpConfSearchCriteria_StopSearch, {"TpConfSearchCriteria_StopSearch","giop-parlay.TpConfSearchCriteria.StopSearch",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_TpConfSearchCriteria_RequestedResources, {"TpConfSearchCriteria_RequestedResources","giop-parlay.TpConfSearchCriteria.RequestedResources",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_TpConfSearchCriteria_RequestedDuration, {"TpConfSearchCriteria_RequestedDuration","giop-parlay.TpConfSearchCriteria.RequestedDuration",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_TpConfSearchResult_MatchFound, {"TpConfSearchResult_MatchFound","giop-parlay.TpConfSearchResult.MatchFound",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_TpConfSearchResult_ActualStartTime, {"TpConfSearchResult_ActualStartTime","giop-parlay.TpConfSearchResult.ActualStartTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_TpConfSearchResult_ActualResources, {"TpConfSearchResult_ActualResources","giop-parlay.TpConfSearchResult.ActualResources",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_TpConfSearchResult_ActualDuration, {"TpConfSearchResult_ActualDuration","giop-parlay.TpConfSearchResult.ActualDuration",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_TpMultiMediaConfPolicy_JoinAllowed, {"TpMultiMediaConfPolicy_JoinAllowed","giop-parlay.TpMultiMediaConfPolicy.JoinAllowed",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_TpMultiMediaConfPolicy_MediaAllowed, {"TpMultiMediaConfPolicy_MediaAllowed","giop-parlay.TpMultiMediaConfPolicy.MediaAllowed",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_TpMultiMediaConfPolicy_Chaired, {"TpMultiMediaConfPolicy_Chaired","giop-parlay.TpMultiMediaConfPolicy.Chaired",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_TpMultiMediaConfPolicy_VideoHandling, {"TpMultiMediaConfPolicy_VideoHandling","giop-parlay.TpMultiMediaConfPolicy.VideoHandling",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_cccs_TpVideoHandlingType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_TpResourceReservation_ReservationID, {"TpResourceReservation_ReservationID","giop-parlay.TpResourceReservation.ReservationID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_TpConfCallIdentifier_ConfCallSessionID, {"TpConfCallIdentifier_ConfCallSessionID","giop-parlay.TpConfCallIdentifier.ConfCallSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_TpSubConfCallIdentifier_SubConfCallSessionID, {"TpSubConfCallIdentifier_SubConfCallSessionID","giop-parlay.TpSubConfCallIdentifier.SubConfCallSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIIdentifier_UserInteractionSessionID, {"TpUIIdentifier_UserInteractionSessionID","giop-parlay.TpUIIdentifier.UserInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUICallIdentifier_UserInteractionSessionID, {"TpUICallIdentifier_UserInteractionSessionID","giop-parlay.TpUICallIdentifier.UserInteractionSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpLocationResponseTime_ResponseTime, {"TpLocationResponseTime_ResponseTime","giop-parlay.TpLocationResponseTime.ResponseTime",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpLocationResponseIndicator),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpLocationResponseTime_TimerValue, {"TpLocationResponseTime_TimerValue","giop-parlay.TpLocationResponseTime.TimerValue",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpLocationTrigger_Longitude, {"TpLocationTrigger_Longitude","giop-parlay.TpLocationTrigger.Longitude",FT_FLOAT,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpLocationTrigger_Latitude, {"TpLocationTrigger_Latitude","giop-parlay.TpLocationTrigger.Latitude",FT_FLOAT,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpLocationTrigger_AreaSemiMajor, {"TpLocationTrigger_AreaSemiMajor","giop-parlay.TpLocationTrigger.AreaSemiMajor",FT_FLOAT,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpLocationTrigger_AreaSemiMinor, {"TpLocationTrigger_AreaSemiMinor","giop-parlay.TpLocationTrigger.AreaSemiMinor",FT_FLOAT,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpLocationTrigger_AngleOfSemiMajor, {"TpLocationTrigger_AngleOfSemiMajor","giop-parlay.TpLocationTrigger.AngleOfSemiMajor",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpLocationTrigger_Criterion, {"TpLocationTrigger_Criterion","giop-parlay.TpLocationTrigger.Criterion",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpLocationTriggerCriteria),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpLocationTrigger_ReportingInterval, {"TpLocationTrigger_ReportingInterval","giop-parlay.TpLocationTrigger.ReportingInterval",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpLocationRequest_RequestedAccuracy, {"TpLocationRequest_RequestedAccuracy","giop-parlay.TpLocationRequest.RequestedAccuracy",FT_FLOAT,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpLocationRequest_AltitudeRequested, {"TpLocationRequest_AltitudeRequested","giop-parlay.TpLocationRequest.AltitudeRequested",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpLocationRequest_Type, {"TpLocationRequest_Type","giop-parlay.TpLocationRequest.Type",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpLocationType),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpLocationRequest_Priority, {"TpLocationRequest_Priority","giop-parlay.TpLocationRequest.Priority",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpLocationPriority),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpLocationRequest_RequestedLocationMethod, {"TpLocationRequest_RequestedLocationMethod","giop-parlay.TpLocationRequest.RequestedLocationMethod",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpGeographicalPosition_Longitude, {"TpGeographicalPosition_Longitude","giop-parlay.TpGeographicalPosition.Longitude",FT_FLOAT,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpGeographicalPosition_Latitude, {"TpGeographicalPosition_Latitude","giop-parlay.TpGeographicalPosition.Latitude",FT_FLOAT,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpGeographicalPosition_TypeOfUncertaintyShape, {"TpGeographicalPosition_TypeOfUncertaintyShape","giop-parlay.TpGeographicalPosition.TypeOfUncertaintyShape",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpLocationUncertaintyShape),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpGeographicalPosition_UncertaintyInnerSemiMajor, {"TpGeographicalPosition_UncertaintyInnerSemiMajor","giop-parlay.TpGeographicalPosition.UncertaintyInnerSemiMajor",FT_FLOAT,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpGeographicalPosition_UncertaintyOuterSemiMajor, {"TpGeographicalPosition_UncertaintyOuterSemiMajor","giop-parlay.TpGeographicalPosition.UncertaintyOuterSemiMajor",FT_FLOAT,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpGeographicalPosition_UncertaintyInnerSemiMinor, {"TpGeographicalPosition_UncertaintyInnerSemiMinor","giop-parlay.TpGeographicalPosition.UncertaintyInnerSemiMinor",FT_FLOAT,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpGeographicalPosition_UncertaintyOuterSemiMinor, {"TpGeographicalPosition_UncertaintyOuterSemiMinor","giop-parlay.TpGeographicalPosition.UncertaintyOuterSemiMinor",FT_FLOAT,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpGeographicalPosition_AngleOfSemiMajor, {"TpGeographicalPosition_AngleOfSemiMajor","giop-parlay.TpGeographicalPosition.AngleOfSemiMajor",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpGeographicalPosition_SegmentStartAngle, {"TpGeographicalPosition_SegmentStartAngle","giop-parlay.TpGeographicalPosition.SegmentStartAngle",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpGeographicalPosition_SegmentEndAngle, {"TpGeographicalPosition_SegmentEndAngle","giop-parlay.TpGeographicalPosition.SegmentEndAngle",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpMobilityStopAssignmentData_AssignmentId, {"TpMobilityStopAssignmentData_AssignmentId","giop-parlay.TpMobilityStopAssignmentData.AssignmentId",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpMobilityStopAssignmentData_StopScope, {"TpMobilityStopAssignmentData_StopScope","giop-parlay.TpMobilityStopAssignmentData.StopScope",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityStopScope),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpMobilityStopAssignmentData_Users_loop, {"Seq length of TpMobilityStopAssignmentData_Users","giop-parlay.TpMobilityStopAssignmentData.Users.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUlExtendedData_TerminalType, {"TpUlExtendedData_TerminalType","giop-parlay.TpUlExtendedData.TerminalType",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpTerminalType),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUlExtendedData_AltitudePresent, {"TpUlExtendedData_AltitudePresent","giop-parlay.TpUlExtendedData.AltitudePresent",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUlExtendedData_Altitude, {"TpUlExtendedData_Altitude","giop-parlay.TpUlExtendedData.Altitude",FT_FLOAT,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUlExtendedData_UncertaintyAltitude, {"TpUlExtendedData_UncertaintyAltitude","giop-parlay.TpUlExtendedData.UncertaintyAltitude",FT_FLOAT,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUlExtendedData_TimestampPresent, {"TpUlExtendedData_TimestampPresent","giop-parlay.TpUlExtendedData.TimestampPresent",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUlExtendedData_Timestamp, {"TpUlExtendedData_Timestamp","giop-parlay.TpUlExtendedData.Timestamp",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUlExtendedData_UsedLocationMethod, {"TpUlExtendedData_UsedLocationMethod","giop-parlay.TpUlExtendedData.UsedLocationMethod",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocation_StatusCode, {"TpUserLocation_StatusCode","giop-parlay.TpUserLocation.StatusCode",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationExtended_StatusCode, {"TpUserLocationExtended_StatusCode","giop-parlay.TpUserLocationExtended.StatusCode",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationExtended_Locations_loop, {"Seq length of TpUserLocationExtended_Locations","giop-parlay.TpUserLocationExtended.Locations.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpLocationTriggerCamel_UpdateInsideVlr, {"TpLocationTriggerCamel_UpdateInsideVlr","giop-parlay.TpLocationTriggerCamel.UpdateInsideVlr",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpLocationTriggerCamel_UpdateOutsideVlr, {"TpLocationTriggerCamel_UpdateOutsideVlr","giop-parlay.TpLocationTriggerCamel.UpdateOutsideVlr",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationCamel_StatusCode, {"TpUserLocationCamel_StatusCode","giop-parlay.TpUserLocationCamel.StatusCode",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationCamel_GeographicalPositionPresent, {"TpUserLocationCamel_GeographicalPositionPresent","giop-parlay.TpUserLocationCamel.GeographicalPositionPresent",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationCamel_TimestampPresent, {"TpUserLocationCamel_TimestampPresent","giop-parlay.TpUserLocationCamel.TimestampPresent",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationCamel_Timestamp, {"TpUserLocationCamel_Timestamp","giop-parlay.TpUserLocationCamel.Timestamp",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationCamel_VlrNumberPresent, {"TpUserLocationCamel_VlrNumberPresent","giop-parlay.TpUserLocationCamel.VlrNumberPresent",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationCamel_LocationNumberPresent, {"TpUserLocationCamel_LocationNumberPresent","giop-parlay.TpUserLocationCamel.LocationNumberPresent",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationCamel_CellIdOrLaiPresent, {"TpUserLocationCamel_CellIdOrLaiPresent","giop-parlay.TpUserLocationCamel.CellIdOrLaiPresent",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationCamel_CellIdOrLai, {"TpUserLocationCamel_CellIdOrLai","giop-parlay.TpUserLocationCamel.CellIdOrLai",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationEmergencyRequest_UserAddressPresent, {"TpUserLocationEmergencyRequest_UserAddressPresent","giop-parlay.TpUserLocationEmergencyRequest.UserAddressPresent",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationEmergencyRequest_NaEsrdPresent, {"TpUserLocationEmergencyRequest_NaEsrdPresent","giop-parlay.TpUserLocationEmergencyRequest.NaEsrdPresent",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationEmergencyRequest_NaEsrd, {"TpUserLocationEmergencyRequest_NaEsrd","giop-parlay.TpUserLocationEmergencyRequest.NaEsrd",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationEmergencyRequest_NaEsrkPresent, {"TpUserLocationEmergencyRequest_NaEsrkPresent","giop-parlay.TpUserLocationEmergencyRequest.NaEsrkPresent",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationEmergencyRequest_NaEsrk, {"TpUserLocationEmergencyRequest_NaEsrk","giop-parlay.TpUserLocationEmergencyRequest.NaEsrk",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationEmergencyRequest_ImeiPresent, {"TpUserLocationEmergencyRequest_ImeiPresent","giop-parlay.TpUserLocationEmergencyRequest.ImeiPresent",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationEmergencyRequest_Imei, {"TpUserLocationEmergencyRequest_Imei","giop-parlay.TpUserLocationEmergencyRequest.Imei",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationEmergency_StatusCode, {"TpUserLocationEmergency_StatusCode","giop-parlay.TpUserLocationEmergency.StatusCode",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationEmergency_UserIdPresent, {"TpUserLocationEmergency_UserIdPresent","giop-parlay.TpUserLocationEmergency.UserIdPresent",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationEmergency_NaEsrdPresent, {"TpUserLocationEmergency_NaEsrdPresent","giop-parlay.TpUserLocationEmergency.NaEsrdPresent",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationEmergency_NaEsrd, {"TpUserLocationEmergency_NaEsrd","giop-parlay.TpUserLocationEmergency.NaEsrd",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationEmergency_NaEsrkPresent, {"TpUserLocationEmergency_NaEsrkPresent","giop-parlay.TpUserLocationEmergency.NaEsrkPresent",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationEmergency_NaEsrk, {"TpUserLocationEmergency_NaEsrk","giop-parlay.TpUserLocationEmergency.NaEsrk",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationEmergency_ImeiPresent, {"TpUserLocationEmergency_ImeiPresent","giop-parlay.TpUserLocationEmergency.ImeiPresent",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationEmergency_Imei, {"TpUserLocationEmergency_Imei","giop-parlay.TpUserLocationEmergency.Imei",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationEmergency_TriggeringEvent, {"TpUserLocationEmergency_TriggeringEvent","giop-parlay.TpUserLocationEmergency.TriggeringEvent",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpUserLocationEmergencyTrigger),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationEmergency_GeographicalPositionPresent, {"TpUserLocationEmergency_GeographicalPositionPresent","giop-parlay.TpUserLocationEmergency.GeographicalPositionPresent",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationEmergency_AltitudePresent, {"TpUserLocationEmergency_AltitudePresent","giop-parlay.TpUserLocationEmergency.AltitudePresent",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationEmergency_Altitude, {"TpUserLocationEmergency_Altitude","giop-parlay.TpUserLocationEmergency.Altitude",FT_FLOAT,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationEmergency_UncertaintyAltitude, {"TpUserLocationEmergency_UncertaintyAltitude","giop-parlay.TpUserLocationEmergency.UncertaintyAltitude",FT_FLOAT,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationEmergency_TimestampPresent, {"TpUserLocationEmergency_TimestampPresent","giop-parlay.TpUserLocationEmergency.TimestampPresent",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationEmergency_Timestamp, {"TpUserLocationEmergency_Timestamp","giop-parlay.TpUserLocationEmergency.Timestamp",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserLocationEmergency_UsedLocationMethod, {"TpUserLocationEmergency_UsedLocationMethod","giop-parlay.TpUserLocationEmergency.UsedLocationMethod",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserStatus_StatusCode, {"TpUserStatus_StatusCode","giop-parlay.TpUserStatus.StatusCode",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserStatus_Status, {"TpUserStatus_Status","giop-parlay.TpUserStatus.Status",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpUserStatusIndicator),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserStatus_TerminalType, {"TpUserStatus_TerminalType","giop-parlay.TpUserStatus.TerminalType",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpTerminalType),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpTriggeredStatusRequest_Users_loop, {"Seq length of TpTriggeredStatusRequest_Users","giop-parlay.TpTriggeredStatusRequest.Users.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpTriggeredStatusRequest_AssignmentID, {"TpTriggeredStatusRequest_AssignmentID","giop-parlay.TpTriggeredStatusRequest.AssignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpTriggeredStatusRequestSetEntry_Requests_loop, {"Seq length of TpTriggeredStatusRequestSetEntry_Requests","giop-parlay.TpTriggeredStatusRequestSetEntry.Requests.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpTriggeredStatusRequestSetEntry_Final, {"TpTriggeredStatusRequestSetEntry_Final","giop-parlay.TpTriggeredStatusRequestSetEntry.Final",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpPeriodicLocationRequest_Users_loop, {"Seq length of TpPeriodicLocationRequest_Users","giop-parlay.TpPeriodicLocationRequest.Users.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpPeriodicLocationRequest_ReportingInterval, {"TpPeriodicLocationRequest_ReportingInterval","giop-parlay.TpPeriodicLocationRequest.ReportingInterval",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpPeriodicLocationRequest_AssignmentID, {"TpPeriodicLocationRequest_AssignmentID","giop-parlay.TpPeriodicLocationRequest.AssignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpPeriodicLocationRequestSetEntry_Requests_loop, {"Seq length of TpPeriodicLocationRequestSetEntry_Requests","giop-parlay.TpPeriodicLocationRequestSetEntry.Requests.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpPeriodicLocationRequestSetEntry_Final, {"TpPeriodicLocationRequestSetEntry_Final","giop-parlay.TpPeriodicLocationRequestSetEntry.Final",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpTriggeredLocationRequest_Users_loop, {"Seq length of TpTriggeredLocationRequest_Users","giop-parlay.TpTriggeredLocationRequest.Users.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpTriggeredLocationRequest_AssignmentID, {"TpTriggeredLocationRequest_AssignmentID","giop-parlay.TpTriggeredLocationRequest.AssignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpTriggeredLocationRequestSetEntry_Requests_loop, {"Seq length of TpTriggeredLocationRequestSetEntry_Requests","giop-parlay.TpTriggeredLocationRequestSetEntry.Requests.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpTriggeredLocationRequestSetEntry_Final, {"TpTriggeredLocationRequestSetEntry_Final","giop-parlay.TpTriggeredLocationRequestSetEntry.Final",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserStatusIndicatorExtended_UserStatusIndicator, {"TpUserStatusIndicatorExtended_UserStatusIndicator","giop-parlay.TpUserStatusIndicatorExtended.UserStatusIndicator",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpUserStatusIndicator),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserStatusIndicatorExtended_PDPContextActive, {"TpUserStatusIndicatorExtended_PDPContextActive","giop-parlay.TpUserStatusIndicatorExtended.PDPContextActive",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserInfo_UserName, {"TpUserInfo_UserName","giop-parlay.TpUserInfo.UserName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserInfo_Password, {"TpUserInfo_Password","giop-parlay.TpUserInfo.Password",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpNetworkStatusIndicator_CountryCode, {"TpNetworkStatusIndicator_CountryCode","giop-parlay.TpNetworkStatusIndicator.CountryCode",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpNetworkStatusIndicator_MobileNetworkCode, {"TpNetworkStatusIndicator_MobileNetworkCode","giop-parlay.TpNetworkStatusIndicator.MobileNetworkCode",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpNetworkStatusIndicator_AccessTechnology, {"TpNetworkStatusIndicator_AccessTechnology","giop-parlay.TpNetworkStatusIndicator.AccessTechnology",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpAccessTechnology),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpNetworkStatusIndicator_RoamingStatus, {"TpNetworkStatusIndicator_RoamingStatus","giop-parlay.TpNetworkStatusIndicator.RoamingStatus",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpRoamingStatus),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserStatusExtended_UserID, {"TpUserStatusExtended_UserID","giop-parlay.TpUserStatusExtended.UserID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserStatusExtended_StatusCode, {"TpUserStatusExtended_StatusCode","giop-parlay.TpUserStatusExtended.StatusCode",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpMobilityError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserStatusExtended_TerminalType, {"TpUserStatusExtended_TerminalType","giop-parlay.TpUserStatusExtended.TerminalType",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpTerminalType),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserStatusExtended_AuthenticationStatus, {"TpUserStatusExtended_AuthenticationStatus","giop-parlay.TpUserStatusExtended.AuthenticationStatus",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpAuthStatusIndicator),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserStatusExtended_UserIPAddress, {"TpUserStatusExtended_UserIPAddress","giop-parlay.TpUserStatusExtended.UserIPAddress",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserStatusExtended_UserMSISDN, {"TpUserStatusExtended_UserMSISDN","giop-parlay.TpUserStatusExtended.UserMSISDN",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserStatusExtended_UserConnectionID, {"TpUserStatusExtended_UserConnectionID","giop-parlay.TpUserStatusExtended.UserConnectionID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpUserStatusExtended_AccessPointName, {"TpUserStatusExtended_AccessPointName","giop-parlay.TpUserStatusExtended.AccessPointName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpBindingNotificationCriteria_NotificationCriteriaType, {"TpBindingNotificationCriteria_NotificationCriteriaType","giop-parlay.TpBindingNotificationCriteria.NotificationCriteriaType",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpBindingNotificationCriteria_ContactAddressSet_loop, {"Seq length of TpBindingNotificationCriteria_ContactAddressSet","giop-parlay.TpBindingNotificationCriteria.ContactAddressSet.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_termcap_TpTerminalCapabilities_TerminalCapabilities, {"TpTerminalCapabilities_TerminalCapabilities","giop-parlay.TpTerminalCapabilities.TerminalCapabilities",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_termcap_TpTerminalCapabilities_StatusCode, {"TpTerminalCapabilities_StatusCode","giop-parlay.TpTerminalCapabilities.StatusCode",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_termcap_TpTerminalCapabilityScope_ScopeType, {"TpTerminalCapabilityScope_ScopeType","giop-parlay.TpTerminalCapabilityScope.ScopeType",FT_UINT32,BASE_DEC,VALS(org_csapi_termcap_TpTerminalCapabilityScopeType),0x0,NULL,HFILL}},
        {&hf_org_csapi_termcap_TpTerminalCapabilityScope_Scope, {"TpTerminalCapabilityScope_Scope","giop-parlay.TpTerminalCapabilityScope.Scope",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpChargePerVolume_InitialCharge, {"TpChargePerVolume_InitialCharge","giop-parlay.TpChargePerVolume.InitialCharge",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpChargePerVolume_CurrentChargePerKilobyte, {"TpChargePerVolume_CurrentChargePerKilobyte","giop-parlay.TpChargePerVolume.CurrentChargePerKilobyte",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpChargePerVolume_NextChargePerKilobyte, {"TpChargePerVolume_NextChargePerKilobyte","giop-parlay.TpChargePerVolume.NextChargePerKilobyte",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionChargePlan_Currency, {"TpDataSessionChargePlan_Currency","giop-parlay.TpDataSessionChargePlan.Currency",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionChargePlan_AdditionalInfo, {"TpDataSessionChargePlan_AdditionalInfo","giop-parlay.TpDataSessionChargePlan.AdditionalInfo",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionError_ErrorTime, {"TpDataSessionError_ErrorTime","giop-parlay.TpDataSessionError.ErrorTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionError_ErrorType, {"TpDataSessionError_ErrorType","giop-parlay.TpDataSessionError.ErrorType",FT_UINT32,BASE_DEC,VALS(org_csapi_dsc_TpDataSessionErrorType),0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionEventCriteria_DataSessionEventName, {"TpDataSessionEventCriteria_DataSessionEventName","giop-parlay.TpDataSessionEventCriteria.DataSessionEventName",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionEventCriteria_MonitorMode, {"TpDataSessionEventCriteria_MonitorMode","giop-parlay.TpDataSessionEventCriteria.MonitorMode",FT_UINT32,BASE_DEC,VALS(org_csapi_dsc_TpDataSessionMonitorMode),0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionEventInfo_DataSessionEventName, {"TpDataSessionEventInfo_DataSessionEventName","giop-parlay.TpDataSessionEventInfo.DataSessionEventName",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionEventInfo_MonitorMode, {"TpDataSessionEventInfo_MonitorMode","giop-parlay.TpDataSessionEventInfo.MonitorMode",FT_UINT32,BASE_DEC,VALS(org_csapi_dsc_TpDataSessionMonitorMode),0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionEventInfo_QoSClass, {"TpDataSessionEventInfo_QoSClass","giop-parlay.TpDataSessionEventInfo.QoSClass",FT_UINT32,BASE_DEC,VALS(org_csapi_TpDataSessionQosClass),0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionReleaseCause_Value, {"TpDataSessionReleaseCause_Value","giop-parlay.TpDataSessionReleaseCause.Value",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionReleaseCause_Location, {"TpDataSessionReleaseCause_Location","giop-parlay.TpDataSessionReleaseCause.Location",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionReport_MonitorMode, {"TpDataSessionReport_MonitorMode","giop-parlay.TpDataSessionReport.MonitorMode",FT_UINT32,BASE_DEC,VALS(org_csapi_dsc_TpDataSessionMonitorMode),0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionReport_DataSessionEventTime, {"TpDataSessionReport_DataSessionEventTime","giop-parlay.TpDataSessionReport.DataSessionEventTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionReport_DataSessionReportType, {"TpDataSessionReport_DataSessionReportType","giop-parlay.TpDataSessionReport.DataSessionReportType",FT_UINT32,BASE_DEC,VALS(org_csapi_dsc_TpDataSessionReportType),0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionReportRequest_MonitorMode, {"TpDataSessionReportRequest_MonitorMode","giop-parlay.TpDataSessionReportRequest.MonitorMode",FT_UINT32,BASE_DEC,VALS(org_csapi_dsc_TpDataSessionMonitorMode),0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionReportRequest_DataSessionReportType, {"TpDataSessionReportRequest_DataSessionReportType","giop-parlay.TpDataSessionReportRequest.DataSessionReportType",FT_UINT32,BASE_DEC,VALS(org_csapi_dsc_TpDataSessionReportType),0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionSuperviseVolume_VolumeQuantity, {"TpDataSessionSuperviseVolume_VolumeQuantity","giop-parlay.TpDataSessionSuperviseVolume.VolumeQuantity",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionSuperviseVolume_VolumeUnit, {"TpDataSessionSuperviseVolume_VolumeUnit","giop-parlay.TpDataSessionSuperviseVolume.VolumeUnit",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionEventCriteriaResult_AssignmentID, {"TpDataSessionEventCriteriaResult_AssignmentID","giop-parlay.TpDataSessionEventCriteriaResult.AssignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionIdentifier_DataSessionID, {"TpDataSessionIdentifier_DataSessionID","giop-parlay.TpDataSessionIdentifier.DataSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpGMSNewMessageArrivedInfo_FolderID, {"TpGMSNewMessageArrivedInfo_FolderID","giop-parlay.TpGMSNewMessageArrivedInfo.FolderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpGMSNewMessageArrivedInfo_MessageID, {"TpGMSNewMessageArrivedInfo_MessageID","giop-parlay.TpGMSNewMessageArrivedInfo.MessageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpGMSNewMessageArrivedInfo_NumberOfProperties, {"TpGMSNewMessageArrivedInfo_NumberOfProperties","giop-parlay.TpGMSNewMessageArrivedInfo.NumberOfProperties",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpGMSNewMessageArrivedCriteria_AuthenticationInfo, {"TpGMSNewMessageArrivedCriteria_AuthenticationInfo","giop-parlay.TpGMSNewMessageArrivedCriteria.AuthenticationInfo",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMailboxIdentifier_SessionID, {"TpMailboxIdentifier_SessionID","giop-parlay.TpMailboxIdentifier.SessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMailboxFolderIdentifier_SessionID, {"TpMailboxFolderIdentifier_SessionID","giop-parlay.TpMailboxFolderIdentifier.SessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagTimePeriod_duration, {"TpNameDescrpTagTimePeriod_duration","giop-parlay.TpNameDescrpTagTimePeriod.duration",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagTimeOfDay_name, {"TpNameDescrpTagTimeOfDay_name","giop-parlay.TpNameDescrpTagTimeOfDay.name",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagTimeOfDay_description, {"TpNameDescrpTagTimeOfDay_description","giop-parlay.TpNameDescrpTagTimeOfDay.description",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagTimeOfDay_tag, {"TpNameDescrpTagTimeOfDay_tag","giop-parlay.TpNameDescrpTagTimeOfDay.tag",FT_UINT32,BASE_DEC,VALS(org_csapi_cm_TpTagValue),0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagTimeOfDay_value, {"TpNameDescrpTagTimeOfDay_value","giop-parlay.TpNameDescrpTagTimeOfDay.value",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagString_name, {"TpNameDescrpTagString_name","giop-parlay.TpNameDescrpTagString.name",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagString_description, {"TpNameDescrpTagString_description","giop-parlay.TpNameDescrpTagString.description",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagString_tag, {"TpNameDescrpTagString_tag","giop-parlay.TpNameDescrpTagString.tag",FT_UINT32,BASE_DEC,VALS(org_csapi_cm_TpTagValue),0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagString_value, {"TpNameDescrpTagString_value","giop-parlay.TpNameDescrpTagString.value",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagMonth_name, {"TpNameDescrpTagMonth_name","giop-parlay.TpNameDescrpTagMonth.name",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagMonth_description, {"TpNameDescrpTagMonth_description","giop-parlay.TpNameDescrpTagMonth.description",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagMonth_tag, {"TpNameDescrpTagMonth_tag","giop-parlay.TpNameDescrpTagMonth.tag",FT_UINT32,BASE_DEC,VALS(org_csapi_cm_TpTagValue),0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagMonth_value, {"TpNameDescrpTagMonth_value","giop-parlay.TpNameDescrpTagMonth.value",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagInt_name, {"TpNameDescrpTagInt_name","giop-parlay.TpNameDescrpTagInt.name",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagInt_description, {"TpNameDescrpTagInt_description","giop-parlay.TpNameDescrpTagInt.description",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagInt_tag, {"TpNameDescrpTagInt_tag","giop-parlay.TpNameDescrpTagInt.tag",FT_UINT32,BASE_DEC,VALS(org_csapi_cm_TpTagValue),0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagInt_value, {"TpNameDescrpTagInt_value","giop-parlay.TpNameDescrpTagInt.value",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagDir_name, {"TpNameDescrpTagDir_name","giop-parlay.TpNameDescrpTagDir.name",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagDir_description, {"TpNameDescrpTagDir_description","giop-parlay.TpNameDescrpTagDir.description",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagDir_tag, {"TpNameDescrpTagDir_tag","giop-parlay.TpNameDescrpTagDir.tag",FT_UINT32,BASE_DEC,VALS(org_csapi_cm_TpTagValue),0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagDir_value, {"TpNameDescrpTagDir_value","giop-parlay.TpNameDescrpTagDir.value",FT_UINT32,BASE_DEC,VALS(org_csapi_cm_TpTrafficDirection),0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagDayOfWeek_name, {"TpNameDescrpTagDayOfWeek_name","giop-parlay.TpNameDescrpTagDayOfWeek.name",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagDayOfWeek_description, {"TpNameDescrpTagDayOfWeek_description","giop-parlay.TpNameDescrpTagDayOfWeek.description",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagDayOfWeek_tag, {"TpNameDescrpTagDayOfWeek_tag","giop-parlay.TpNameDescrpTagDayOfWeek.tag",FT_UINT32,BASE_DEC,VALS(org_csapi_cm_TpTagValue),0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagDayOfWeek_value, {"TpNameDescrpTagDayOfWeek_value","giop-parlay.TpNameDescrpTagDayOfWeek.value",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagDateTime_name, {"TpNameDescrpTagDateTime_name","giop-parlay.TpNameDescrpTagDateTime.name",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagDateTime_description, {"TpNameDescrpTagDateTime_description","giop-parlay.TpNameDescrpTagDateTime.description",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagDateTime_tag, {"TpNameDescrpTagDateTime_tag","giop-parlay.TpNameDescrpTagDateTime.tag",FT_UINT32,BASE_DEC,VALS(org_csapi_cm_TpTagValue),0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagDateTime_value, {"TpNameDescrpTagDateTime_value","giop-parlay.TpNameDescrpTagDateTime.value",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpEndpoint_type, {"TpEndpoint_type","giop-parlay.TpEndpoint.type",FT_UINT32,BASE_DEC,VALS(org_csapi_cm_TpSiteOrSap),0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpEndpoint_id, {"TpEndpoint_id","giop-parlay.TpEndpoint.id",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagExcessLoadAction_name, {"TpNameDescrpTagExcessLoadAction_name","giop-parlay.TpNameDescrpTagExcessLoadAction.name",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagExcessLoadAction_description, {"TpNameDescrpTagExcessLoadAction_description","giop-parlay.TpNameDescrpTagExcessLoadAction.description",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagExcessLoadAction_tag, {"TpNameDescrpTagExcessLoadAction_tag","giop-parlay.TpNameDescrpTagExcessLoadAction.tag",FT_UINT32,BASE_DEC,VALS(org_csapi_cm_TpTagValue),0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpNameDescrpTagExcessLoadAction_value, {"TpNameDescrpTagExcessLoadAction_value","giop-parlay.TpNameDescrpTagExcessLoadAction.value",FT_UINT32,BASE_DEC,VALS(org_csapi_cm_TpAction),0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpDsCodepoint_match, {"TpDsCodepoint_match","giop-parlay.TpDsCodepoint.match",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpDsCodepoint_mask, {"TpDsCodepoint_mask","giop-parlay.TpDsCodepoint.mask",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpIPSubnet_subnetNumber, {"TpIPSubnet_subnetNumber","giop-parlay.TpIPSubnet.subnetNumber",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpIPSubnet_subnetMask, {"TpIPSubnet_subnetMask","giop-parlay.TpIPSubnet.subnetMask",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpIPSubnet_addressType, {"TpIPSubnet_addressType","giop-parlay.TpIPSubnet.addressType",FT_UINT32,BASE_DEC,VALS(org_csapi_cm_TpIPv4AddType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_TpIPSubnet_IPVersionSupport, {"TpIPSubnet_IPVersionSupport","giop-parlay.TpIPSubnet.IPVersionSupport",FT_UINT32,BASE_DEC,VALS(org_csapi_cm_TpIPVersion),0x0,NULL,HFILL}},
        {&hf_org_csapi_am_TpBalanceInfo_Currency, {"TpBalanceInfo_Currency","giop-parlay.TpBalanceInfo.Currency",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_TpBalanceInfo_ValuePartA, {"TpBalanceInfo_ValuePartA","giop-parlay.TpBalanceInfo.ValuePartA",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_TpBalanceInfo_ValuePartB, {"TpBalanceInfo_ValuePartB","giop-parlay.TpBalanceInfo.ValuePartB",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_TpBalanceInfo_Exponent, {"TpBalanceInfo_Exponent","giop-parlay.TpBalanceInfo.Exponent",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_TpBalanceInfo_AdditionalInfo, {"TpBalanceInfo_AdditionalInfo","giop-parlay.TpBalanceInfo.AdditionalInfo",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_TpChargingEventInfo_ChargingEventName, {"TpChargingEventInfo_ChargingEventName","giop-parlay.TpChargingEventInfo.ChargingEventName",FT_UINT32,BASE_DEC,VALS(org_csapi_am_TpChargingEventName),0x0,NULL,HFILL}},
        {&hf_org_csapi_am_TpChargingEventInfo_ChargingEventTime, {"TpChargingEventInfo_ChargingEventTime","giop-parlay.TpChargingEventInfo.ChargingEventTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_TpBalance_StatusCode, {"TpBalance_StatusCode","giop-parlay.TpBalance.StatusCode",FT_UINT32,BASE_DEC,VALS(org_csapi_am_TpBalanceQueryError),0x0,NULL,HFILL}},
        {&hf_org_csapi_am_TpTransactionHistory_TransactionID, {"TpTransactionHistory_TransactionID","giop-parlay.TpTransactionHistory.TransactionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_TpTransactionHistory_TimeStamp, {"TpTransactionHistory_TimeStamp","giop-parlay.TpTransactionHistory.TimeStamp",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_TpTransactionHistory_AdditionalInfo, {"TpTransactionHistory_AdditionalInfo","giop-parlay.TpTransactionHistory.AdditionalInfo",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_TpChargingEventCriteria_ChargingEvents_loop, {"Seq length of TpChargingEventCriteria_ChargingEvents","giop-parlay.TpChargingEventCriteria.ChargingEvents.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_TpChargingEventCriteria_ChargingEvents, {"TpChargingEventCriteria_ChargingEvents","giop-parlay.TpChargingEventCriteria.ChargingEvents",FT_UINT32,BASE_DEC,VALS(org_csapi_am_TpChargingEventName),0x0,NULL,HFILL}},
        {&hf_org_csapi_am_TpChargingEventCriteria_Users_loop, {"Seq length of TpChargingEventCriteria_Users","giop-parlay.TpChargingEventCriteria.Users.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_TpChargingEventCriteriaResult_AssignmentID, {"TpChargingEventCriteriaResult_AssignmentID","giop-parlay.TpChargingEventCriteriaResult.AssignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_TpBalanceExpiryDate_StatusCode, {"TpBalanceExpiryDate_StatusCode","giop-parlay.TpBalanceExpiryDate.StatusCode",FT_UINT32,BASE_DEC,VALS(org_csapi_am_TpBalanceQueryError),0x0,NULL,HFILL}},
        {&hf_org_csapi_am_TpBalanceExpiryDate_ExpiryDate, {"TpBalanceExpiryDate_ExpiryDate","giop-parlay.TpBalanceExpiryDate.ExpiryDate",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_TpVoucher_VoucherID, {"TpVoucher_VoucherID","giop-parlay.TpVoucher.VoucherID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_TpMerchantAccountID_MerchantID, {"TpMerchantAccountID_MerchantID","giop-parlay.TpMerchantAccountID.MerchantID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_TpMerchantAccountID_AccountID, {"TpMerchantAccountID_AccountID","giop-parlay.TpMerchantAccountID.AccountID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_TpCorrelationID_CorrelationID, {"TpCorrelationID_CorrelationID","giop-parlay.TpCorrelationID.CorrelationID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_TpCorrelationID_CorrelationType, {"TpCorrelationID_CorrelationType","giop-parlay.TpCorrelationID.CorrelationType",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_TpAmount_Number, {"TpAmount_Number","giop-parlay.TpAmount.Number",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_TpAmount_Exponent, {"TpAmount_Exponent","giop-parlay.TpAmount.Exponent",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_TpChargingPrice_Currency, {"TpChargingPrice_Currency","giop-parlay.TpChargingPrice.Currency",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_TpVolume_Unit, {"TpVolume_Unit","giop-parlay.TpVolume.Unit",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_TpChargingSessionID_ChargingSessionID, {"TpChargingSessionID_ChargingSessionID","giop-parlay.TpChargingSessionID.ChargingSessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_TpChargingSessionID_RequestNumberFirstRequest, {"TpChargingSessionID_RequestNumberFirstRequest","giop-parlay.TpChargingSessionID.RequestNumberFirstRequest",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_TpApplicationDescription_Text, {"TpApplicationDescription_Text","giop-parlay.TpApplicationDescription.Text",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_TpApplicationDescription_AppInformation_loop, {"Seq length of TpApplicationDescription_AppInformation","giop-parlay.TpApplicationDescription.AppInformation.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_TpChargingParameter_ParameterID, {"TpChargingParameter_ParameterID","giop-parlay.TpChargingParameter.ParameterID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_TpPolicyEvent_EventID, {"TpPolicyEvent_EventID","giop-parlay.TpPolicyEvent.EventID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_TpPolicyEvent_TimeGenerated, {"TpPolicyEvent_TimeGenerated","giop-parlay.TpPolicyEvent.TimeGenerated",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_TpPolicyEvent_Attributes_loop, {"Seq length of TpPolicyEvent_Attributes","giop-parlay.TpPolicyEvent.Attributes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_TpPolicyEvent_EventDefinitionName, {"TpPolicyEvent_EventDefinitionName","giop-parlay.TpPolicyEvent.EventDefinitionName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_TpPolicyEvent_EventDomainName, {"TpPolicyEvent_EventDomainName","giop-parlay.TpPolicyEvent.EventDomainName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_TpPolicyNameValue_Name, {"TpPolicyNameValue_Name","giop-parlay.TpPolicyNameValue.Name",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Names_loop, {"Seq length of TpPolicyRecordType_Names","giop-parlay.TpPolicyRecordType.Names.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Names, {"TpPolicyRecordType_Names","giop-parlay.TpPolicyRecordType.Names",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_TpPolicyType_TpPolicyRecordType_Types_loop, {"Seq length of TpPolicyRecordType_Types","giop-parlay.TpPolicyRecordType.Types.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_TpPolicyType_TpPolicyListType_ElementType_loop, {"Seq length of TpPolicyListType_ElementType","giop-parlay.TpPolicyListType.ElementType.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_TpPolicyVar_VarName, {"TpPolicyVar_VarName","giop-parlay.TpPolicyVar.VarName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_TpPolicyConditionListElement_GroupNumber, {"TpPolicyConditionListElement_GroupNumber","giop-parlay.TpPolicyConditionListElement.GroupNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_TpPolicyConditionListElement_Negated, {"TpPolicyConditionListElement_Negated","giop-parlay.TpPolicyConditionListElement.Negated",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_policy_TpPolicyActionListElement_SequenceNumber, {"TpPolicyActionListElement_SequenceNumber","giop-parlay.TpPolicyActionListElement.SequenceNumber",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAttributeDef_Name, {"TpPAMAttributeDef_Name","giop-parlay.TpPAMAttributeDef.Name",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAttributeDef_Type, {"TpPAMAttributeDef_Type","giop-parlay.TpPAMAttributeDef.Type",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAttributeDef_IsStatic, {"TpPAMAttributeDef_IsStatic","giop-parlay.TpPAMAttributeDef.IsStatic",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAttributeDef_IsRevertOnExpiration, {"TpPAMAttributeDef_IsRevertOnExpiration","giop-parlay.TpPAMAttributeDef.IsRevertOnExpiration",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAttribute_AttributeName, {"TpPAMAttribute_AttributeName","giop-parlay.TpPAMAttribute.AttributeName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAttribute_ExpiresIn, {"TpPAMAttribute_ExpiresIn","giop-parlay.TpPAMAttribute.ExpiresIn",FT_INT64,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMPresenceData_Name, {"TpPAMPresenceData_Name","giop-parlay.TpPAMPresenceData.Name",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMPresenceData_subscriberStatus, {"TpPAMPresenceData_subscriberStatus","giop-parlay.TpPAMPresenceData.subscriberStatus",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMPresenceData_networkStatus, {"TpPAMPresenceData_networkStatus","giop-parlay.TpPAMPresenceData.networkStatus",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMPresenceData_communicationMeans, {"TpPAMPresenceData_communicationMeans","giop-parlay.TpPAMPresenceData.communicationMeans",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMPresenceData_subscriberProvidedLocation, {"TpPAMPresenceData_subscriberProvidedLocation","giop-parlay.TpPAMPresenceData.subscriberProvidedLocation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMPresenceData_networkProvidedLocation, {"TpPAMPresenceData_networkProvidedLocation","giop-parlay.TpPAMPresenceData.networkProvidedLocation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMPresenceData_Priority, {"TpPAMPresenceData_Priority","giop-parlay.TpPAMPresenceData.Priority",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMPresenceData_otherInfo, {"TpPAMPresenceData_otherInfo","giop-parlay.TpPAMPresenceData.otherInfo",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAvailabilityProfile_PrivacyCode, {"TpPAMAvailabilityProfile_PrivacyCode","giop-parlay.TpPAMAvailabilityProfile.PrivacyCode",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMCommunicationContext_CommunicationCapability, {"TpPAMCommunicationContext_CommunicationCapability","giop-parlay.TpPAMCommunicationContext.CommunicationCapability",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMContext_AskerData_loop, {"Seq length of TpPAMContext_AskerData","giop-parlay.TpPAMContext.AskerData.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAccessControlData_DefaultPolicy, {"TpPAMAccessControlData_DefaultPolicy","giop-parlay.TpPAMAccessControlData.DefaultPolicy",FT_UINT32,BASE_DEC,VALS(org_csapi_pam_TpPAMACLDefault),0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAccessControlData_AllowList_loop, {"Seq length of TpPAMAccessControlData_AllowList","giop-parlay.TpPAMAccessControlData.AllowList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAccessControlData_AllowList, {"TpPAMAccessControlData_AllowList","giop-parlay.TpPAMAccessControlData.AllowList",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAccessControlData_DenyList_loop, {"Seq length of TpPAMAccessControlData_DenyList","giop-parlay.TpPAMAccessControlData.DenyList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAccessControlData_DenyList, {"TpPAMAccessControlData_DenyList","giop-parlay.TpPAMAccessControlData.DenyList",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMICEventData_IdentityType_loop, {"Seq length of TpPAMICEventData_IdentityType","giop-parlay.TpPAMICEventData.IdentityType.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMICEventData_IdentityType, {"TpPAMICEventData_IdentityType","giop-parlay.TpPAMICEventData.IdentityType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMICNotificationData_Identities_loop, {"Seq length of TpPAMICNotificationData_Identities","giop-parlay.TpPAMICNotificationData.Identities.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMICNotificationData_Identities, {"TpPAMICNotificationData_Identities","giop-parlay.TpPAMICNotificationData.Identities",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMIDEventData_IdentityName_loop, {"Seq length of TpPAMIDEventData_IdentityName","giop-parlay.TpPAMIDEventData.IdentityName.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMIDEventData_IdentityName, {"TpPAMIDEventData_IdentityName","giop-parlay.TpPAMIDEventData.IdentityName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMIDEventData_IdentityType_loop, {"Seq length of TpPAMIDEventData_IdentityType","giop-parlay.TpPAMIDEventData.IdentityType.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMIDEventData_IdentityType, {"TpPAMIDEventData_IdentityType","giop-parlay.TpPAMIDEventData.IdentityType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMIDNotificationData_Identities_loop, {"Seq length of TpPAMIDNotificationData_Identities","giop-parlay.TpPAMIDNotificationData.Identities.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMIDNotificationData_Identities, {"TpPAMIDNotificationData_Identities","giop-parlay.TpPAMIDNotificationData.Identities",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMGMCEventData_GroupName_loop, {"Seq length of TpPAMGMCEventData_GroupName","giop-parlay.TpPAMGMCEventData.GroupName.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMGMCEventData_GroupName, {"TpPAMGMCEventData_GroupName","giop-parlay.TpPAMGMCEventData.GroupName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMGMCEventData_GroupType_loop, {"Seq length of TpPAMGMCEventData_GroupType","giop-parlay.TpPAMGMCEventData.GroupType.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMGMCEventData_GroupType, {"TpPAMGMCEventData_GroupType","giop-parlay.TpPAMGMCEventData.GroupType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMGMCNotificationData_Groups_loop, {"Seq length of TpPAMGMCNotificationData_Groups","giop-parlay.TpPAMGMCNotificationData.Groups.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMGMCNotificationData_Groups, {"TpPAMGMCNotificationData_Groups","giop-parlay.TpPAMGMCNotificationData.Groups",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMACEventData_AgentType_loop, {"Seq length of TpPAMACEventData_AgentType","giop-parlay.TpPAMACEventData.AgentType.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMACEventData_AgentType, {"TpPAMACEventData_AgentType","giop-parlay.TpPAMACEventData.AgentType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMACNotificationData_Agents_loop, {"Seq length of TpPAMACNotificationData_Agents","giop-parlay.TpPAMACNotificationData.Agents.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMACNotificationData_Agents, {"TpPAMACNotificationData_Agents","giop-parlay.TpPAMACNotificationData.Agents",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMADEventData_AgentName_loop, {"Seq length of TpPAMADEventData_AgentName","giop-parlay.TpPAMADEventData.AgentName.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMADEventData_AgentName, {"TpPAMADEventData_AgentName","giop-parlay.TpPAMADEventData.AgentName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMADEventData_AgentType_loop, {"Seq length of TpPAMADEventData_AgentType","giop-parlay.TpPAMADEventData.AgentType.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMADEventData_AgentType, {"TpPAMADEventData_AgentType","giop-parlay.TpPAMADEventData.AgentType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMADNotificationData_Agents_loop, {"Seq length of TpPAMADNotificationData_Agents","giop-parlay.TpPAMADNotificationData.Agents.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMADNotificationData_Agents, {"TpPAMADNotificationData_Agents","giop-parlay.TpPAMADNotificationData.Agents",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAAEventData_IdentityName_loop, {"Seq length of TpPAMAAEventData_IdentityName","giop-parlay.TpPAMAAEventData.IdentityName.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAAEventData_IdentityName, {"TpPAMAAEventData_IdentityName","giop-parlay.TpPAMAAEventData.IdentityName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAAEventData_IdentityType_loop, {"Seq length of TpPAMAAEventData_IdentityType","giop-parlay.TpPAMAAEventData.IdentityType.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAAEventData_IdentityType, {"TpPAMAAEventData_IdentityType","giop-parlay.TpPAMAAEventData.IdentityType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAAEventData_AgentName_loop, {"Seq length of TpPAMAAEventData_AgentName","giop-parlay.TpPAMAAEventData.AgentName.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAAEventData_AgentName, {"TpPAMAAEventData_AgentName","giop-parlay.TpPAMAAEventData.AgentName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAAEventData_AgentType_loop, {"Seq length of TpPAMAAEventData_AgentType","giop-parlay.TpPAMAAEventData.AgentType.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAAEventData_AgentType, {"TpPAMAAEventData_AgentType","giop-parlay.TpPAMAAEventData.AgentType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAANotificationData_Identity, {"TpPAMAANotificationData_Identity","giop-parlay.TpPAMAANotificationData.Identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAANotificationData_Agent, {"TpPAMAANotificationData_Agent","giop-parlay.TpPAMAANotificationData.Agent",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAUEventData_IdentityName_loop, {"Seq length of TpPAMAUEventData_IdentityName","giop-parlay.TpPAMAUEventData.IdentityName.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAUEventData_IdentityName, {"TpPAMAUEventData_IdentityName","giop-parlay.TpPAMAUEventData.IdentityName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAUEventData_IdentityType_loop, {"Seq length of TpPAMAUEventData_IdentityType","giop-parlay.TpPAMAUEventData.IdentityType.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAUEventData_IdentityType, {"TpPAMAUEventData_IdentityType","giop-parlay.TpPAMAUEventData.IdentityType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAUEventData_AgentName_loop, {"Seq length of TpPAMAUEventData_AgentName","giop-parlay.TpPAMAUEventData.AgentName.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAUEventData_AgentName, {"TpPAMAUEventData_AgentName","giop-parlay.TpPAMAUEventData.AgentName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAUEventData_AgentType_loop, {"Seq length of TpPAMAUEventData_AgentType","giop-parlay.TpPAMAUEventData.AgentType.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAUEventData_AgentType, {"TpPAMAUEventData_AgentType","giop-parlay.TpPAMAUEventData.AgentType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAUNotificationData_Identity, {"TpPAMAUNotificationData_Identity","giop-parlay.TpPAMAUNotificationData.Identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAUNotificationData_Agent, {"TpPAMAUNotificationData_Agent","giop-parlay.TpPAMAUNotificationData.Agent",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMCCEventData_IdentityName_loop, {"Seq length of TpPAMCCEventData_IdentityName","giop-parlay.TpPAMCCEventData.IdentityName.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMCCEventData_IdentityName, {"TpPAMCCEventData_IdentityName","giop-parlay.TpPAMCCEventData.IdentityName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMCCEventData_IdentityType_loop, {"Seq length of TpPAMCCEventData_IdentityType","giop-parlay.TpPAMCCEventData.IdentityType.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMCCEventData_IdentityType, {"TpPAMCCEventData_IdentityType","giop-parlay.TpPAMCCEventData.IdentityType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMCCEventData_Capabilities_loop, {"Seq length of TpPAMCCEventData_Capabilities","giop-parlay.TpPAMCCEventData.Capabilities.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMCCEventData_Capabilities, {"TpPAMCCEventData_Capabilities","giop-parlay.TpPAMCCEventData.Capabilities",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMCCNotificationData_Identity, {"TpPAMCCNotificationData_Identity","giop-parlay.TpPAMCCNotificationData.Identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMCCNotificationData_Capabilities_loop, {"Seq length of TpPAMCCNotificationData_Capabilities","giop-parlay.TpPAMCCNotificationData.Capabilities.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMCCNotificationData_Capabilities, {"TpPAMCCNotificationData_Capabilities","giop-parlay.TpPAMCCNotificationData.Capabilities",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMACPSEventData_AgentName_loop, {"Seq length of TpPAMACPSEventData_AgentName","giop-parlay.TpPAMACPSEventData.AgentName.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMACPSEventData_AgentName, {"TpPAMACPSEventData_AgentName","giop-parlay.TpPAMACPSEventData.AgentName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMACPSEventData_AgentType_loop, {"Seq length of TpPAMACPSEventData_AgentType","giop-parlay.TpPAMACPSEventData.AgentType.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMACPSEventData_AgentType, {"TpPAMACPSEventData_AgentType","giop-parlay.TpPAMACPSEventData.AgentType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMACPSEventData_Capabilities_loop, {"Seq length of TpPAMACPSEventData_Capabilities","giop-parlay.TpPAMACPSEventData.Capabilities.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMACPSEventData_Capabilities, {"TpPAMACPSEventData_Capabilities","giop-parlay.TpPAMACPSEventData.Capabilities",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMACPSEventData_AttributeNames_loop, {"Seq length of TpPAMACPSEventData_AttributeNames","giop-parlay.TpPAMACPSEventData.AttributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMACPSEventData_AttributeNames, {"TpPAMACPSEventData_AttributeNames","giop-parlay.TpPAMACPSEventData.AttributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMACPSEventData_ReportingPeriod, {"TpPAMACPSEventData_ReportingPeriod","giop-parlay.TpPAMACPSEventData.ReportingPeriod",FT_INT64,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMACPSNotificationData_Agent, {"TpPAMACPSNotificationData_Agent","giop-parlay.TpPAMACPSNotificationData.Agent",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMACPSNotificationData_Capability, {"TpPAMACPSNotificationData_Capability","giop-parlay.TpPAMACPSNotificationData.Capability",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMACPSNotificationData_AttributeNames_loop, {"Seq length of TpPAMACPSNotificationData_AttributeNames","giop-parlay.TpPAMACPSNotificationData.AttributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMACPSNotificationData_AttributeNames, {"TpPAMACPSNotificationData_AttributeNames","giop-parlay.TpPAMACPSNotificationData.AttributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAPSEventData_AgentName_loop, {"Seq length of TpPAMAPSEventData_AgentName","giop-parlay.TpPAMAPSEventData.AgentName.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAPSEventData_AgentName, {"TpPAMAPSEventData_AgentName","giop-parlay.TpPAMAPSEventData.AgentName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAPSEventData_AgentType_loop, {"Seq length of TpPAMAPSEventData_AgentType","giop-parlay.TpPAMAPSEventData.AgentType.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAPSEventData_AgentType, {"TpPAMAPSEventData_AgentType","giop-parlay.TpPAMAPSEventData.AgentType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAPSEventData_AttributeNames_loop, {"Seq length of TpPAMAPSEventData_AttributeNames","giop-parlay.TpPAMAPSEventData.AttributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAPSEventData_AttributeNames, {"TpPAMAPSEventData_AttributeNames","giop-parlay.TpPAMAPSEventData.AttributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAPSEventData_ReportingPeriod, {"TpPAMAPSEventData_ReportingPeriod","giop-parlay.TpPAMAPSEventData.ReportingPeriod",FT_INT64,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAPSNotificationData_Agent, {"TpPAMAPSNotificationData_Agent","giop-parlay.TpPAMAPSNotificationData.Agent",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAPSNotificationData_AttributeNames_loop, {"Seq length of TpPAMAPSNotificationData_AttributeNames","giop-parlay.TpPAMAPSNotificationData.AttributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAPSNotificationData_AttributeNames, {"TpPAMAPSNotificationData_AttributeNames","giop-parlay.TpPAMAPSNotificationData.AttributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMIPSEventData_IdentityName_loop, {"Seq length of TpPAMIPSEventData_IdentityName","giop-parlay.TpPAMIPSEventData.IdentityName.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMIPSEventData_IdentityName, {"TpPAMIPSEventData_IdentityName","giop-parlay.TpPAMIPSEventData.IdentityName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMIPSEventData_IdentityType_loop, {"Seq length of TpPAMIPSEventData_IdentityType","giop-parlay.TpPAMIPSEventData.IdentityType.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMIPSEventData_IdentityType, {"TpPAMIPSEventData_IdentityType","giop-parlay.TpPAMIPSEventData.IdentityType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMIPSEventData_AttributeNames_loop, {"Seq length of TpPAMIPSEventData_AttributeNames","giop-parlay.TpPAMIPSEventData.AttributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMIPSEventData_AttributeNames, {"TpPAMIPSEventData_AttributeNames","giop-parlay.TpPAMIPSEventData.AttributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMIPSEventData_ReportingPeriod, {"TpPAMIPSEventData_ReportingPeriod","giop-parlay.TpPAMIPSEventData.ReportingPeriod",FT_INT64,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMIPSNotificationData_Identity, {"TpPAMIPSNotificationData_Identity","giop-parlay.TpPAMIPSNotificationData.Identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMIPSNotificationData_Attributes_loop, {"Seq length of TpPAMIPSNotificationData_Attributes","giop-parlay.TpPAMIPSNotificationData.Attributes.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAVCEventData_IdentityName_loop, {"Seq length of TpPAMAVCEventData_IdentityName","giop-parlay.TpPAMAVCEventData.IdentityName.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAVCEventData_IdentityName, {"TpPAMAVCEventData_IdentityName","giop-parlay.TpPAMAVCEventData.IdentityName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAVCEventData_IdentityType_loop, {"Seq length of TpPAMAVCEventData_IdentityType","giop-parlay.TpPAMAVCEventData.IdentityType.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAVCEventData_IdentityType, {"TpPAMAVCEventData_IdentityType","giop-parlay.TpPAMAVCEventData.IdentityType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAVCEventData_PAMContext_loop, {"Seq length of TpPAMAVCEventData_PAMContext","giop-parlay.TpPAMAVCEventData.PAMContext.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAVCEventData_AttributeNames_loop, {"Seq length of TpPAMAVCEventData_AttributeNames","giop-parlay.TpPAMAVCEventData.AttributeNames.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAVCEventData_AttributeNames, {"TpPAMAVCEventData_AttributeNames","giop-parlay.TpPAMAVCEventData.AttributeNames",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAVCEventData_ReportingPeriod, {"TpPAMAVCEventData_ReportingPeriod","giop-parlay.TpPAMAVCEventData.ReportingPeriod",FT_INT64,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAVCNotificationData_Identity, {"TpPAMAVCNotificationData_Identity","giop-parlay.TpPAMAVCNotificationData.Identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMAVCNotificationData_Availability_loop, {"Seq length of TpPAMAVCNotificationData_Availability","giop-parlay.TpPAMAVCNotificationData.Availability.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMWCEventData_Events_loop, {"Seq length of TpPAMWCEventData_Events","giop-parlay.TpPAMWCEventData.Events.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMWCEventData_Events, {"TpPAMWCEventData_Events","giop-parlay.TpPAMWCEventData.Events",FT_UINT32,BASE_DEC,VALS(org_csapi_pam_TpPAMEventName),0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMWCEventData_IdentityName_loop, {"Seq length of TpPAMWCEventData_IdentityName","giop-parlay.TpPAMWCEventData.IdentityName.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMWCEventData_IdentityName, {"TpPAMWCEventData_IdentityName","giop-parlay.TpPAMWCEventData.IdentityName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMWCEventData_IdentityType_loop, {"Seq length of TpPAMWCEventData_IdentityType","giop-parlay.TpPAMWCEventData.IdentityType.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMWCEventData_IdentityType, {"TpPAMWCEventData_IdentityType","giop-parlay.TpPAMWCEventData.IdentityType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMWCEventData_ReportingPeriod, {"TpPAMWCEventData_ReportingPeriod","giop-parlay.TpPAMWCEventData.ReportingPeriod",FT_INT64,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMWCNotificationData_Event, {"TpPAMWCNotificationData_Event","giop-parlay.TpPAMWCNotificationData.Event",FT_UINT32,BASE_DEC,VALS(org_csapi_pam_TpPAMEventName),0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMWCNotificationData_ChangeType, {"TpPAMWCNotificationData_ChangeType","giop-parlay.TpPAMWCNotificationData.ChangeType",FT_UINT32,BASE_DEC,VALS(org_csapi_pam_TpPAMwatcherChangeType),0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMWCNotificationData_Identity, {"TpPAMWCNotificationData_Identity","giop-parlay.TpPAMWCNotificationData.Identity",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMWCNotificationData_Watchers_loop, {"Seq length of TpPAMWCNotificationData_Watchers","giop-parlay.TpPAMWCNotificationData.Watchers.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMWCNotificationData_Watchers, {"TpPAMWCNotificationData_Watchers","giop-parlay.TpPAMWCNotificationData.Watchers",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMErrorInfo_Cause, {"TpPAMErrorInfo_Cause","giop-parlay.TpPAMErrorInfo.Cause",FT_UINT32,BASE_DEC,VALS(org_csapi_pam_TpPAMErrorCause),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMailboxIdentifier_SessionID, {"TpMailboxIdentifier_SessionID","giop-parlay.TpMailboxIdentifier.SessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMultiMediaMessagingIdentifier_SessionID, {"TpMultiMediaMessagingIdentifier_SessionID","giop-parlay.TpMultiMediaMessagingIdentifier.SessionID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpListMessagesCriteria_OnlyUnreadMessages, {"TpListMessagesCriteria_OnlyUnreadMessages","giop-parlay.TpListMessagesCriteria.OnlyUnreadMessages",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMailboxFolderStatusInformation_TotalMessageCount, {"TpMailboxFolderStatusInformation_TotalMessageCount","giop-parlay.TpMailboxFolderStatusInformation.TotalMessageCount",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageDescription_MessageID, {"TpMessageDescription_MessageID","giop-parlay.TpMessageDescription.MessageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageDescription_To_loop, {"Seq length of TpMessageDescription_To","giop-parlay.TpMessageDescription.To.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageDescription_Subject, {"TpMessageDescription_Subject","giop-parlay.TpMessageDescription.Subject",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageDescription_ReceivedDate, {"TpMessageDescription_ReceivedDate","giop-parlay.TpMessageDescription.ReceivedDate",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageDescription_Size, {"TpMessageDescription_Size","giop-parlay.TpMessageDescription.Size",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpBodyPartDescription_ContentDescription, {"TpBodyPartDescription_ContentDescription","giop-parlay.TpBodyPartDescription.ContentDescription",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpBodyPartDescription_ContentSize, {"TpBodyPartDescription_ContentSize","giop-parlay.TpBodyPartDescription.ContentSize",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpBodyPartDescription_ContentType, {"TpBodyPartDescription_ContentType","giop-parlay.TpBodyPartDescription.ContentType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpBodyPartDescription_ContentTransferEncoding, {"TpBodyPartDescription_ContentTransferEncoding","giop-parlay.TpBodyPartDescription.ContentTransferEncoding",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpBodyPartDescription_ContentID, {"TpBodyPartDescription_ContentID","giop-parlay.TpBodyPartDescription.ContentID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpBodyPartDescription_ContentDisposition, {"TpBodyPartDescription_ContentDisposition","giop-parlay.TpBodyPartDescription.ContentDisposition",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpBodyPartDescription_PartID, {"TpBodyPartDescription_PartID","giop-parlay.TpBodyPartDescription.PartID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpBodyPartDescription_NestingLevel, {"TpBodyPartDescription_NestingLevel","giop-parlay.TpBodyPartDescription.NestingLevel",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpBodyPart_BodyPartContent_loop, {"Seq length of TpBodyPart_BodyPartContent","giop-parlay.TpBodyPart.BodyPartContent.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpBodyPart_BodyPartContent, {"TpBodyPart_BodyPartContent","giop-parlay.TpBodyPart.BodyPartContent",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpGenericHeaderField_FieldName, {"TpGenericHeaderField_FieldName","giop-parlay.TpGenericHeaderField.FieldName",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpGenericHeaderField_FieldValue, {"TpGenericHeaderField_FieldValue","giop-parlay.TpGenericHeaderField.FieldValue",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageInfoPropertyError_MessagePropertyName, {"TpMessageInfoPropertyError_MessagePropertyName","giop-parlay.TpMessageInfoPropertyError.MessagePropertyName",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessageInfoPropertyName),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageInfoPropertyError_Error, {"TpMessageInfoPropertyError_Error","giop-parlay.TpMessageInfoPropertyError.Error",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpSetPropertyError),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpNewMailboxMessageArrivedCriteria_MailboxID, {"TpNewMailboxMessageArrivedCriteria_MailboxID","giop-parlay.TpNewMailboxMessageArrivedCriteria.MailboxID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpNewMailboxMessageArrivedCriteria_AuthenticationInfo, {"TpNewMailboxMessageArrivedCriteria_AuthenticationInfo","giop-parlay.TpNewMailboxMessageArrivedCriteria.AuthenticationInfo",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_MailboxID, {"TpNewMailboxMessageArrivedInfo_MailboxID","giop-parlay.TpNewMailboxMessageArrivedInfo.MailboxID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_FolderID, {"TpNewMailboxMessageArrivedInfo_FolderID","giop-parlay.TpNewMailboxMessageArrivedInfo.FolderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_MessageDescription_loop, {"Seq length of TpNewMailboxMessageArrivedInfo_MessageDescription","giop-parlay.TpNewMailboxMessageArrivedInfo.MessageDescription.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpNewMailboxMessageArrivedInfo_ExtendedHeaderInformation_loop, {"Seq length of TpNewMailboxMessageArrivedInfo_ExtendedHeaderInformation","giop-parlay.TpNewMailboxMessageArrivedInfo.ExtendedHeaderInformation.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpQueryStatusReport_ReportedStatus, {"TpQueryStatusReport_ReportedStatus","giop-parlay.TpQueryStatusReport.ReportedStatus",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpTerminatingAddressList_ToAddressList_loop, {"Seq length of TpTerminatingAddressList_ToAddressList","giop-parlay.TpTerminatingAddressList.ToAddressList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpTerminatingAddressList_CcAddressList_loop, {"Seq length of TpTerminatingAddressList_CcAddressList","giop-parlay.TpTerminatingAddressList.CcAddressList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpTerminatingAddressList_BccAddressList_loop, {"Seq length of TpTerminatingAddressList_BccAddressList","giop-parlay.TpTerminatingAddressList.BccAddressList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpNewMessageArrivedCriteria_CreateMultiMediaMessagingSession, {"TpNewMessageArrivedCriteria_CreateMultiMediaMessagingSession","giop-parlay.TpNewMessageArrivedCriteria.CreateMultiMediaMessagingSession",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessagingNotificationRequested_EventCriteria_loop, {"Seq length of TpMessagingNotificationRequested_EventCriteria","giop-parlay.TpMessagingNotificationRequested.EventCriteria.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessagingNotificationRequested_AssignmentID, {"TpMessagingNotificationRequested_AssignmentID","giop-parlay.TpMessagingNotificationRequested.AssignmentID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessagingNotificationRequestedSetEntry_MessagingNotificationRequestedSet_loop, {"Seq length of TpMessagingNotificationRequestedSetEntry_MessagingNotificationRequestedSet","giop-parlay.TpMessagingNotificationRequestedSetEntry.MessagingNotificationRequestedSet.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessagingNotificationRequestedSetEntry_Final, {"TpMessagingNotificationRequestedSetEntry_Final","giop-parlay.TpMessagingNotificationRequestedSetEntry.Final",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpNewMessageArrivedInfo_DestinationAddressSet_loop, {"Seq length of TpNewMessageArrivedInfo_DestinationAddressSet","giop-parlay.TpNewMessageArrivedInfo.DestinationAddressSet.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpNewMessageArrivedInfo_Message_loop, {"Seq length of TpNewMessageArrivedInfo_Message","giop-parlay.TpNewMessageArrivedInfo.Message.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpNewMessageArrivedInfo_Message, {"TpNewMessageArrivedInfo_Message","giop-parlay.TpNewMessageArrivedInfo.Message",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpNewMessageArrivedInfo_Headers_loop, {"Seq length of TpNewMessageArrivedInfo_Headers","giop-parlay.TpNewMessageArrivedInfo.Headers.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        /* User exception filters */
        {&hf_org_csapi_TpCommonExceptions_ExceptionType, {"TpCommonExceptions_ExceptionType","giop-parlay.csapi.TpCommonExceptions_ExceptionType",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpCommonExceptions_ExtraInformation, {"TpCommonExceptions_ExtraInformation","giop-parlay.csapi.TpCommonExceptions_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_P_INVALID_INTERFACE_TYPE_ExtraInformation, {"P_INVALID_INTERFACE_TYPE_ExtraInformation","giop-parlay.csapi.P_INVALID_INTERFACE_TYPE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_P_INVALID_SESSION_ID_ExtraInformation, {"P_INVALID_SESSION_ID_ExtraInformation","giop-parlay.csapi.P_INVALID_SESSION_ID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_INVALID_DOMAIN_ID_ExtraInformation, {"P_INVALID_DOMAIN_ID_ExtraInformation","giop-parlay.csapi.P_INVALID_DOMAIN_ID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_INVALID_AUTH_TYPE_ExtraInformation, {"P_INVALID_AUTH_TYPE_ExtraInformation","giop-parlay.csapi.P_INVALID_AUTH_TYPE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_P_INVALID_VERSION_ExtraInformation, {"P_INVALID_VERSION_ExtraInformation","giop-parlay.csapi.P_INVALID_VERSION_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_ACCESS_DENIED_ExtraInformation, {"P_ACCESS_DENIED_ExtraInformation","giop-parlay.csapi.P_ACCESS_DENIED_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_INVALID_ACCESS_TYPE_ExtraInformation, {"P_INVALID_ACCESS_TYPE_ExtraInformation","giop-parlay.csapi.P_INVALID_ACCESS_TYPE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_INVALID_SIGNING_ALGORITHM_ExtraInformation, {"P_INVALID_SIGNING_ALGORITHM_ExtraInformation","giop-parlay.csapi.P_INVALID_SIGNING_ALGORITHM_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_INVALID_SIGNATURE_ExtraInformation, {"P_INVALID_SIGNATURE_ExtraInformation","giop-parlay.csapi.P_INVALID_SIGNATURE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_P_INVALID_INTERFACE_NAME_ExtraInformation, {"P_INVALID_INTERFACE_NAME_ExtraInformation","giop-parlay.csapi.P_INVALID_INTERFACE_NAME_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_INVALID_PROPERTY_ExtraInformation, {"P_INVALID_PROPERTY_ExtraInformation","giop-parlay.csapi.P_INVALID_PROPERTY_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_NO_ACCEPTABLE_SIGNING_ALGORITHM_ExtraInformation, {"P_NO_ACCEPTABLE_SIGNING_ALGORITHM_ExtraInformation","giop-parlay.csapi.P_NO_ACCEPTABLE_SIGNING_ALGORITHM_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_NO_ACCEPTABLE_ENCRYPTION_CAPABILITY_ExtraInformation, {"P_NO_ACCEPTABLE_ENCRYPTION_CAPABILITY_ExtraInformation","giop-parlay.csapi.P_NO_ACCEPTABLE_ENCRYPTION_CAPABILITY_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_NO_ACCEPTABLE_AUTHENTICATION_MECHANISM_ExtraInformation, {"P_NO_ACCEPTABLE_AUTHENTICATION_MECHANISM_ExtraInformation","giop-parlay.csapi.P_NO_ACCEPTABLE_AUTHENTICATION_MECHANISM_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_P_INVALID_CRITERIA_ExtraInformation, {"P_INVALID_CRITERIA_ExtraInformation","giop-parlay.csapi.P_INVALID_CRITERIA_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_P_INVALID_EVENT_TYPE_ExtraInformation, {"P_INVALID_EVENT_TYPE_ExtraInformation","giop-parlay.csapi.P_INVALID_EVENT_TYPE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_P_INVALID_ASSIGNMENT_ID_ExtraInformation, {"P_INVALID_ASSIGNMENT_ID_ExtraInformation","giop-parlay.csapi.P_INVALID_ASSIGNMENT_ID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_INVALID_SERVICE_ID_ExtraInformation, {"P_INVALID_SERVICE_ID_ExtraInformation","giop-parlay.csapi.P_INVALID_SERVICE_ID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_SERVICE_NOT_ENABLED_ExtraInformation, {"P_SERVICE_NOT_ENABLED_ExtraInformation","giop-parlay.csapi.P_SERVICE_NOT_ENABLED_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_P_UNAUTHORISED_PARAMETER_VALUE_ExtraInformation, {"P_UNAUTHORISED_PARAMETER_VALUE_ExtraInformation","giop-parlay.csapi.P_UNAUTHORISED_PARAMETER_VALUE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_P_INVALID_TIME_AND_DATE_FORMAT_ExtraInformation, {"P_INVALID_TIME_AND_DATE_FORMAT_ExtraInformation","giop-parlay.csapi.P_INVALID_TIME_AND_DATE_FORMAT_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_INVALID_ACTIVITY_TEST_ID_ExtraInformation, {"P_INVALID_ACTIVITY_TEST_ID_ExtraInformation","giop-parlay.csapi.P_INVALID_ACTIVITY_TEST_ID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_ILLEGAL_SERVICE_TYPE_ExtraInformation, {"P_ILLEGAL_SERVICE_TYPE_ExtraInformation","giop-parlay.csapi.P_ILLEGAL_SERVICE_TYPE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_UNKNOWN_SERVICE_TYPE_ExtraInformation, {"P_UNKNOWN_SERVICE_TYPE_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_SERVICE_TYPE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_INVALID_AGREEMENT_TEXT_ExtraInformation, {"P_INVALID_AGREEMENT_TEXT_ExtraInformation","giop-parlay.csapi.P_INVALID_AGREEMENT_TEXT_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_INVALID_SERVICE_TOKEN_ExtraInformation, {"P_INVALID_SERVICE_TOKEN_ExtraInformation","giop-parlay.csapi.P_INVALID_SERVICE_TOKEN_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_SERVICE_ACCESS_DENIED_ExtraInformation, {"P_SERVICE_ACCESS_DENIED_ExtraInformation","giop-parlay.csapi.P_SERVICE_ACCESS_DENIED_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_INVALID_SERVICE_PROFILE_ID_ExtraInformation, {"P_INVALID_SERVICE_PROFILE_ID_ExtraInformation","giop-parlay.csapi.P_INVALID_SERVICE_PROFILE_ID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_INVALID_SAG_ID_ExtraInformation, {"P_INVALID_SAG_ID_ExtraInformation","giop-parlay.csapi.P_INVALID_SAG_ID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_INVALID_SAG_TO_SERVICE_PROFILE_ASSIGNMENT_ExtraInformation, {"P_INVALID_SAG_TO_SERVICE_PROFILE_ASSIGNMENT_ExtraInformation","giop-parlay.csapi.P_INVALID_SAG_TO_SERVICE_PROFILE_ASSIGNMENT_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_INVALID_SERVICE_CONTRACT_ID_ExtraInformation, {"P_INVALID_SERVICE_CONTRACT_ID_ExtraInformation","giop-parlay.csapi.P_INVALID_SERVICE_CONTRACT_ID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_INVALID_CLIENT_APP_ID_ExtraInformation, {"P_INVALID_CLIENT_APP_ID_ExtraInformation","giop-parlay.csapi.P_INVALID_CLIENT_APP_ID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_INVALID_ADDITION_TO_SAG_ExtraInformation, {"P_INVALID_ADDITION_TO_SAG_ExtraInformation","giop-parlay.csapi.P_INVALID_ADDITION_TO_SAG_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_PROPERTY_TYPE_MISMATCH_ExtraInformation, {"P_PROPERTY_TYPE_MISMATCH_ExtraInformation","giop-parlay.csapi.P_PROPERTY_TYPE_MISMATCH_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_DUPLICATE_PROPERTY_NAME_ExtraInformation, {"P_DUPLICATE_PROPERTY_NAME_ExtraInformation","giop-parlay.csapi.P_DUPLICATE_PROPERTY_NAME_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_MISSING_MANDATORY_PROPERTY_ExtraInformation, {"P_MISSING_MANDATORY_PROPERTY_ExtraInformation","giop-parlay.csapi.P_MISSING_MANDATORY_PROPERTY_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_SERVICE_TYPE_UNAVAILABLE_ExtraInformation, {"P_SERVICE_TYPE_UNAVAILABLE_ExtraInformation","giop-parlay.csapi.P_SERVICE_TYPE_UNAVAILABLE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_ILLEGAL_SERVICE_ID_ExtraInformation, {"P_ILLEGAL_SERVICE_ID_ExtraInformation","giop-parlay.csapi.P_ILLEGAL_SERVICE_ID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_P_UNKNOWN_SERVICE_ID_ExtraInformation, {"P_UNKNOWN_SERVICE_ID_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_SERVICE_ID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_P_INVALID_ADDRESS_ExtraInformation, {"P_INVALID_ADDRESS_ExtraInformation","giop-parlay.csapi.P_INVALID_ADDRESS_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_P_UNSUPPORTED_ADDRESS_PLAN_ExtraInformation, {"P_UNSUPPORTED_ADDRESS_PLAN_ExtraInformation","giop-parlay.csapi.P_UNSUPPORTED_ADDRESS_PLAN_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_P_INVALID_NETWORK_STATE_ExtraInformation, {"P_INVALID_NETWORK_STATE_ExtraInformation","giop-parlay.csapi.P_INVALID_NETWORK_STATE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_P_INVALID_CURRENCY_ExtraInformation, {"P_INVALID_CURRENCY_ExtraInformation","giop-parlay.csapi.P_INVALID_CURRENCY_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_P_INVALID_AMOUNT_ExtraInformation, {"P_INVALID_AMOUNT_ExtraInformation","giop-parlay.csapi.P_INVALID_AMOUNT_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_P_INFORMATION_NOT_AVAILABLE_ExtraInformation, {"P_INFORMATION_NOT_AVAILABLE_ExtraInformation","giop-parlay.csapi.P_INFORMATION_NOT_AVAILABLE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_P_ILLEGAL_ID_ExtraInformation, {"P_ILLEGAL_ID_ExtraInformation","giop-parlay.csapi.P_ILLEGAL_ID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_P_ID_NOT_FOUND_ExtraInformation, {"P_ID_NOT_FOUND_ExtraInformation","giop-parlay.csapi.P_ID_NOT_FOUND_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_P_ILLEGAL_RANGE_ExtraInformation, {"P_ILLEGAL_RANGE_ExtraInformation","giop-parlay.csapi.P_ILLEGAL_RANGE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_P_INVALID_COLLECTION_CRITERIA_ExtraInformation, {"P_INVALID_COLLECTION_CRITERIA_ExtraInformation","giop-parlay.csapi.P_INVALID_COLLECTION_CRITERIA_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_P_APPLICATION_NOT_ACTIVATED_ExtraInformation, {"P_APPLICATION_NOT_ACTIVATED_ExtraInformation","giop-parlay.csapi.P_APPLICATION_NOT_ACTIVATED_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_P_REQUESTED_ACCURACY_CANNOT_BE_DELIVERED_ExtraInformation, {"P_REQUESTED_ACCURACY_CANNOT_BE_DELIVERED_ExtraInformation","giop-parlay.csapi.P_REQUESTED_ACCURACY_CANNOT_BE_DELIVERED_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_P_REQUESTED_RESPONSE_TIME_CANNOT_BE_DELIVERED_ExtraInformation, {"P_REQUESTED_RESPONSE_TIME_CANNOT_BE_DELIVERED_ExtraInformation","giop-parlay.csapi.P_REQUESTED_RESPONSE_TIME_CANNOT_BE_DELIVERED_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_P_INVALID_REPORTING_INTERVAL_ExtraInformation, {"P_INVALID_REPORTING_INTERVAL_ExtraInformation","giop-parlay.csapi.P_INVALID_REPORTING_INTERVAL_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_P_TRIGGER_CONDITIONS_NOT_SUBSCRIBED_ExtraInformation, {"P_TRIGGER_CONDITIONS_NOT_SUBSCRIBED_ExtraInformation","giop-parlay.csapi.P_TRIGGER_CONDITIONS_NOT_SUBSCRIBED_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_P_UNKNOWN_SUBSCRIBER_ExtraInformation, {"P_UNKNOWN_SUBSCRIBER_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_SUBSCRIBER_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_termcap_P_INVALID_TERMINAL_ID_ExtraInformation, {"P_INVALID_TERMINAL_ID_ExtraInformation","giop-parlay.csapi.P_INVALID_TERMINAL_ID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_P_GMS_INVALID_MAILBOX_ExtraInformation, {"P_GMS_INVALID_MAILBOX_ExtraInformation","giop-parlay.csapi.P_GMS_INVALID_MAILBOX_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_P_GMS_INVALID_AUTHENTICATION_INFORMATION_ExtraInformation, {"P_GMS_INVALID_AUTHENTICATION_INFORMATION_ExtraInformation","giop-parlay.csapi.P_GMS_INVALID_AUTHENTICATION_INFORMATION_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_P_GMS_LOCKING_LOCKED_MAILBOX_ExtraInformation, {"P_GMS_LOCKING_LOCKED_MAILBOX_ExtraInformation","giop-parlay.csapi.P_GMS_LOCKING_LOCKED_MAILBOX_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_P_GMS_UNLOCKING_UNLOCKED_MAILBOX_ExtraInformation, {"P_GMS_UNLOCKING_UNLOCKED_MAILBOX_ExtraInformation","giop-parlay.csapi.P_GMS_UNLOCKING_UNLOCKED_MAILBOX_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_P_GMS_CANNOT_UNLOCK_MAILBOX_ExtraInformation, {"P_GMS_CANNOT_UNLOCK_MAILBOX_ExtraInformation","giop-parlay.csapi.P_GMS_CANNOT_UNLOCK_MAILBOX_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_P_GMS_NUMBER_NOT_POSITIVE_ExtraInformation, {"P_GMS_NUMBER_NOT_POSITIVE_ExtraInformation","giop-parlay.csapi.P_GMS_NUMBER_NOT_POSITIVE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_P_GMS_PROPERTY_NOT_SET_ExtraInformation, {"P_GMS_PROPERTY_NOT_SET_ExtraInformation","giop-parlay.csapi.P_GMS_PROPERTY_NOT_SET_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_P_GMS_MAILBOX_LOCKED_ExtraInformation, {"P_GMS_MAILBOX_LOCKED_ExtraInformation","giop-parlay.csapi.P_GMS_MAILBOX_LOCKED_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_P_GMS_FOLDER_IS_OPEN_ExtraInformation, {"P_GMS_FOLDER_IS_OPEN_ExtraInformation","giop-parlay.csapi.P_GMS_FOLDER_IS_OPEN_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_P_GMS_INVALID_FOLDER_ID_ExtraInformation, {"P_GMS_INVALID_FOLDER_ID_ExtraInformation","giop-parlay.csapi.P_GMS_INVALID_FOLDER_ID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_P_GMS_INSUFFICIENT_PRIVILEGE_ExtraInformation, {"P_GMS_INSUFFICIENT_PRIVILEGE_ExtraInformation","giop-parlay.csapi.P_GMS_INSUFFICIENT_PRIVILEGE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_P_GMS_MAILBOX_OPEN_ExtraInformation, {"P_GMS_MAILBOX_OPEN_ExtraInformation","giop-parlay.csapi.P_GMS_MAILBOX_OPEN_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_P_GMS_INVALID_MESSAGE_ID_ExtraInformation, {"P_GMS_INVALID_MESSAGE_ID_ExtraInformation","giop-parlay.csapi.P_GMS_INVALID_MESSAGE_ID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_P_GMS_MESSAGE_NOT_REMOVED_ExtraInformation, {"P_GMS_MESSAGE_NOT_REMOVED_ExtraInformation","giop-parlay.csapi.P_GMS_MESSAGE_NOT_REMOVED_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_UNKNOWN_VPRP_ID_ExtraInformation, {"P_UNKNOWN_VPRP_ID_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_VPRP_ID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_UNKNOWN_SLA_ID_ExtraInformation, {"P_UNKNOWN_SLA_ID_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_SLA_ID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_UNKNOWN_STATUS_ExtraInformation, {"P_UNKNOWN_STATUS_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_STATUS_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_UNKNOWN_QOS_INFO_ExtraInformation, {"P_UNKNOWN_QOS_INFO_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_QOS_INFO_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_UNKNOWN_VALIDITY_INFO_ExtraInformation, {"P_UNKNOWN_VALIDITY_INFO_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_VALIDITY_INFO_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_UNKNOWN_PIPEQOSINFO_ExtraInformation, {"P_UNKNOWN_PIPEQOSINFO_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_PIPEQOSINFO_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_UNKNOWN_DSCODEPOINT_ExtraInformation, {"P_UNKNOWN_DSCODEPOINT_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_DSCODEPOINT_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_UNKNOWN_VPRP_ExtraInformation, {"P_UNKNOWN_VPRP_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_VPRP_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_ILLEGAL_VPRPID_ExtraInformation, {"P_ILLEGAL_VPRPID_ExtraInformation","giop-parlay.csapi.P_ILLEGAL_VPRPID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_ILLEGAL_REF_VALUE_ExtraInformation, {"P_ILLEGAL_REF_VALUE_ExtraInformation","giop-parlay.csapi.P_ILLEGAL_REF_VALUE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_UNKNOWN_INTERFACE_ExtraInformation, {"P_UNKNOWN_INTERFACE_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_INTERFACE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_CANT_DELETE_VPRP_ExtraInformation, {"P_CANT_DELETE_VPRP_ExtraInformation","giop-parlay.csapi.P_CANT_DELETE_VPRP_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_UNKNOWN_TEMPLATES_ExtraInformation, {"P_UNKNOWN_TEMPLATES_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_TEMPLATES_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_UNKNOWN_MENU_ExtraInformation, {"P_UNKNOWN_MENU_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_MENU_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_UNKNOWN_ENTERPRISE_NETWORK_ExtraInformation, {"P_UNKNOWN_ENTERPRISE_NETWORK_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_ENTERPRISE_NETWORK_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_UNKNOWN_TEMPLATE_TYPE_ExtraInformation, {"P_UNKNOWN_TEMPLATE_TYPE_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_TEMPLATE_TYPE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_UNKNOWN_DESCRIPTION_ExtraInformation, {"P_UNKNOWN_DESCRIPTION_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_DESCRIPTION_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_ILLEGAL_SLA_ID_ExtraInformation, {"P_ILLEGAL_SLA_ID_ExtraInformation","giop-parlay.csapi.P_ILLEGAL_SLA_ID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_ILLEGAL_TAG_ExtraInformation, {"P_ILLEGAL_TAG_ExtraInformation","giop-parlay.csapi.P_ILLEGAL_TAG_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_ILLEGAL_VALUE_ExtraInformation, {"P_ILLEGAL_VALUE_ExtraInformation","giop-parlay.csapi.P_ILLEGAL_VALUE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_ILLEGAL_COMBINATION_ExtraInformation, {"P_ILLEGAL_COMBINATION_ExtraInformation","giop-parlay.csapi.P_ILLEGAL_COMBINATION_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_UNKNOWN_SITES_ExtraInformation, {"P_UNKNOWN_SITES_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_SITES_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_UNKNOWN_VPRN_ExtraInformation, {"P_UNKNOWN_VPRN_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_VPRN_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_ILLEGAL_SITE_ID_ExtraInformation, {"P_ILLEGAL_SITE_ID_ExtraInformation","giop-parlay.csapi.P_ILLEGAL_SITE_ID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_UNKNOWN_SITE_ID_ExtraInformation, {"P_UNKNOWN_SITE_ID_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_SITE_ID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_UNKNOWN_SAPS_ExtraInformation, {"P_UNKNOWN_SAPS_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_SAPS_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_UNKNOWN_SITE_LOCATION_ExtraInformation, {"P_UNKNOWN_SITE_LOCATION_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_SITE_LOCATION_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_UNKNOWN_SITE_DESCRIPTION_ExtraInformation, {"P_UNKNOWN_SITE_DESCRIPTION_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_SITE_DESCRIPTION_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_UNKNOWN_IPSUBNET_ExtraInformation, {"P_UNKNOWN_IPSUBNET_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_IPSUBNET_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cm_P_UNKNOWN_SAP_ExtraInformation, {"P_UNKNOWN_SAP_ExtraInformation","giop-parlay.csapi.P_UNKNOWN_SAP_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_am_P_UNAUTHORIZED_APPLICATION_ExtraInformation, {"P_UNAUTHORIZED_APPLICATION_ExtraInformation","giop-parlay.csapi.P_UNAUTHORIZED_APPLICATION_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_P_INVALID_REQUEST_NUMBER_ExtraInformation, {"P_INVALID_REQUEST_NUMBER_ExtraInformation","giop-parlay.csapi.P_INVALID_REQUEST_NUMBER_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_P_INVALID_VOLUME_ExtraInformation, {"P_INVALID_VOLUME_ExtraInformation","giop-parlay.csapi.P_INVALID_VOLUME_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_P_INVALID_USER_ExtraInformation, {"P_INVALID_USER_ExtraInformation","giop-parlay.csapi.P_INVALID_USER_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_P_INVALID_ACCOUNT_ExtraInformation, {"P_INVALID_ACCOUNT_ExtraInformation","giop-parlay.csapi.P_INVALID_ACCOUNT_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_P_ACCESS_VIOLATION_ExtraInformation, {"P_ACCESS_VIOLATION_ExtraInformation","giop-parlay.csapi.P_ACCESS_VIOLATION_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_P_SYNTAX_ERROR_ExtraInformation, {"P_SYNTAX_ERROR_ExtraInformation","giop-parlay.csapi.P_SYNTAX_ERROR_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_P_NAME_SPACE_ERROR_ExtraInformation, {"P_NAME_SPACE_ERROR_ExtraInformation","giop-parlay.csapi.P_NAME_SPACE_ERROR_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_P_NO_TRANSACTION_IN_PROCESS_ExtraInformation, {"P_NO_TRANSACTION_IN_PROCESS_ExtraInformation","giop-parlay.csapi.P_NO_TRANSACTION_IN_PROCESS_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_P_TRANSACTION_IN_PROCESS_ExtraInformation, {"P_TRANSACTION_IN_PROCESS_ExtraInformation","giop-parlay.csapi.P_TRANSACTION_IN_PROCESS_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_IDENTITY_EXISTS_ExtraInformation, {"P_PAM_IDENTITY_EXISTS_ExtraInformation","giop-parlay.csapi.P_PAM_IDENTITY_EXISTS_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_UNKNOWN_TYPE_ExtraInformation, {"P_PAM_UNKNOWN_TYPE_ExtraInformation","giop-parlay.csapi.P_PAM_UNKNOWN_TYPE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_INVALID_CREDENTIAL_ExtraInformation, {"P_PAM_INVALID_CREDENTIAL_ExtraInformation","giop-parlay.csapi.P_PAM_INVALID_CREDENTIAL_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_UNKNOWN_IDENTITY_ExtraInformation, {"P_PAM_UNKNOWN_IDENTITY_ExtraInformation","giop-parlay.csapi.P_PAM_UNKNOWN_IDENTITY_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_MEMBER_EXISTS_ExtraInformation, {"P_PAM_MEMBER_EXISTS_ExtraInformation","giop-parlay.csapi.P_PAM_MEMBER_EXISTS_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_UNKNOWN_GROUP_ExtraInformation, {"P_PAM_UNKNOWN_GROUP_ExtraInformation","giop-parlay.csapi.P_PAM_UNKNOWN_GROUP_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_UNKNOWN_MEMBER_ExtraInformation, {"P_PAM_UNKNOWN_MEMBER_ExtraInformation","giop-parlay.csapi.P_PAM_UNKNOWN_MEMBER_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_IS_CYCLIC_ExtraInformation, {"P_PAM_IS_CYCLIC_ExtraInformation","giop-parlay.csapi.P_PAM_IS_CYCLIC_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_NOT_MEMBER_ExtraInformation, {"P_PAM_NOT_MEMBER_ExtraInformation","giop-parlay.csapi.P_PAM_NOT_MEMBER_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_ALIAS_EXISTS_ExtraInformation, {"P_PAM_ALIAS_EXISTS_ExtraInformation","giop-parlay.csapi.P_PAM_ALIAS_EXISTS_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_ALIAS_NOT_UNIQUE_ExtraInformation, {"P_PAM_ALIAS_NOT_UNIQUE_ExtraInformation","giop-parlay.csapi.P_PAM_ALIAS_NOT_UNIQUE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_UNASSIGNED_ALIAS_ExtraInformation, {"P_PAM_UNASSIGNED_ALIAS_ExtraInformation","giop-parlay.csapi.P_PAM_UNASSIGNED_ALIAS_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_UNKNOWN_ALIAS_ExtraInformation, {"P_PAM_UNKNOWN_ALIAS_ExtraInformation","giop-parlay.csapi.P_PAM_UNKNOWN_ALIAS_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_TYPE_ASSOCIATED_ExtraInformation, {"P_PAM_TYPE_ASSOCIATED_ExtraInformation","giop-parlay.csapi.P_PAM_TYPE_ASSOCIATED_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_DISASSOCIATED_TYPE_ExtraInformation, {"P_PAM_DISASSOCIATED_TYPE_ExtraInformation","giop-parlay.csapi.P_PAM_DISASSOCIATED_TYPE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_UNKNOWN_ATTRIBUTE_ExtraInformation, {"P_PAM_UNKNOWN_ATTRIBUTE_ExtraInformation","giop-parlay.csapi.P_PAM_UNKNOWN_ATTRIBUTE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_UNKNOWN_ATTRIBUTES_ExtraInformation, {"P_PAM_UNKNOWN_ATTRIBUTES_ExtraInformation","giop-parlay.csapi.P_PAM_UNKNOWN_ATTRIBUTES_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_AGENT_EXISTS_ExtraInformation, {"P_PAM_AGENT_EXISTS_ExtraInformation","giop-parlay.csapi.P_PAM_AGENT_EXISTS_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_UNKNOWN_AGENT_ExtraInformation, {"P_PAM_UNKNOWN_AGENT_ExtraInformation","giop-parlay.csapi.P_PAM_UNKNOWN_AGENT_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_NO_CAPABILITY_ExtraInformation, {"P_PAM_NO_CAPABILITY_ExtraInformation","giop-parlay.csapi.P_PAM_NO_CAPABILITY_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_UNKNOWN_ASSIGNMENT_ExtraInformation, {"P_PAM_UNKNOWN_ASSIGNMENT_ExtraInformation","giop-parlay.csapi.P_PAM_UNKNOWN_ASSIGNMENT_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_ATTRIBUTE_EXISTS_ExtraInformation, {"P_PAM_ATTRIBUTE_EXISTS_ExtraInformation","giop-parlay.csapi.P_PAM_ATTRIBUTE_EXISTS_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_TYPE_EXISTS_ExtraInformation, {"P_PAM_TYPE_EXISTS_ExtraInformation","giop-parlay.csapi.P_PAM_TYPE_EXISTS_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_UNKNOWN_CAPABILITY_ExtraInformation, {"P_PAM_UNKNOWN_CAPABILITY_ExtraInformation","giop-parlay.csapi.P_PAM_UNKNOWN_CAPABILITY_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_UNAVAILABLE_INTERFACE_ExtraInformation, {"P_PAM_UNAVAILABLE_INTERFACE_ExtraInformation","giop-parlay.csapi.P_PAM_UNAVAILABLE_INTERFACE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_P_PAM_NOT_REGISTERED_ExtraInformation, {"P_PAM_NOT_REGISTERED_ExtraInformation","giop-parlay.csapi.P_PAM_NOT_REGISTERED_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_P_MMM_INVALID_MAILBOX_ExtraInformation, {"P_MMM_INVALID_MAILBOX_ExtraInformation","giop-parlay.csapi.P_MMM_INVALID_MAILBOX_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_P_MMM_INVALID_AUTHENTICATION_INFORMATION_ExtraInformation, {"P_MMM_INVALID_AUTHENTICATION_INFORMATION_ExtraInformation","giop-parlay.csapi.P_MMM_INVALID_AUTHENTICATION_INFORMATION_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_P_MMM_INVALID_FOLDER_ID_ExtraInformation, {"P_MMM_INVALID_FOLDER_ID_ExtraInformation","giop-parlay.csapi.P_MMM_INVALID_FOLDER_ID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_P_MMM_MAX_MESSAGE_SIZE_EXCEEDED_ExtraInformation, {"P_MMM_MAX_MESSAGE_SIZE_EXCEEDED_ExtraInformation","giop-parlay.csapi.P_MMM_MAX_MESSAGE_SIZE_EXCEEDED_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_P_MMM_INVALID_MESSAGE_ID_ExtraInformation, {"P_MMM_INVALID_MESSAGE_ID_ExtraInformation","giop-parlay.csapi.P_MMM_INVALID_MESSAGE_ID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_P_MMM_INVALID_PART_ID_ExtraInformation, {"P_MMM_INVALID_PART_ID_ExtraInformation","giop-parlay.csapi.P_MMM_INVALID_PART_ID_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_P_MMM_INVALID_PROPERTY_ExtraInformation, {"P_MMM_INVALID_PROPERTY_ExtraInformation","giop-parlay.csapi.P_MMM_INVALID_PROPERTY_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_P_MMM_INVALID_DELIVERY_TYPE_ExtraInformation, {"P_MMM_INVALID_DELIVERY_TYPE_ExtraInformation","giop-parlay.csapi.P_MMM_INVALID_DELIVERY_TYPE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_P_MMM_DELIVERY_TYPE_ADDRESS_TYPE_MISMATCH_ExtraInformation, {"P_MMM_DELIVERY_TYPE_ADDRESS_TYPE_MISMATCH_ExtraInformation","giop-parlay.csapi.P_MMM_DELIVERY_TYPE_ADDRESS_TYPE_MISMATCH_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_P_MMM_DELIVERY_TYPE_MESSAGE_TYPE_MISMATCH_ExtraInformation, {"P_MMM_DELIVERY_TYPE_MESSAGE_TYPE_MISMATCH_ExtraInformation","giop-parlay.csapi.P_MMM_DELIVERY_TYPE_MESSAGE_TYPE_MISMATCH_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_P_MMM_INVALID_DELIVERY_TIME_ExtraInformation, {"P_MMM_INVALID_DELIVERY_TIME_ExtraInformation","giop-parlay.csapi.P_MMM_INVALID_DELIVERY_TIME_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_P_MMM_INVALID_VALIDITY_TIME_ExtraInformation, {"P_MMM_INVALID_VALIDITY_TIME_ExtraInformation","giop-parlay.csapi.P_MMM_INVALID_VALIDITY_TIME_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_P_MMM_MAX_SUBJECT_SIZE_EXCEEDED_ExtraInformation, {"P_MMM_MAX_SUBJECT_SIZE_EXCEEDED_ExtraInformation","giop-parlay.csapi.P_MMM_MAX_SUBJECT_SIZE_EXCEEDED_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_P_MMM_INVALID_HEADER_ExtraInformation, {"P_MMM_INVALID_HEADER_ExtraInformation","giop-parlay.csapi.P_MMM_INVALID_HEADER_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_P_MMM_CANNOT_CANCEL_ExtraInformation, {"P_MMM_CANNOT_CANCEL_ExtraInformation","giop-parlay.csapi.P_MMM_CANNOT_CANCEL_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_P_MMM_INFORMATION_NOT_AVAILABLE_ExtraInformation, {"P_MMM_INFORMATION_NOT_AVAILABLE_ExtraInformation","giop-parlay.csapi.P_MMM_INFORMATION_NOT_AVAILABLE_ExtraInformation",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        /* Union filters */
        {&hf_org_csapi_TpAoCOrder_TpAoCOrder, {"TpAoCOrder","giop-parlay.org.csapi.TpAoCOrder",FT_UINT32,BASE_DEC,VALS(org_csapi_TpCallAoCOrderCategory),0x0,NULL,HFILL}},
        {&hf_org_csapi_TpAoCOrder_NetworkCharge, {"TpAoCOrder_NetworkCharge","giop-parlay.org.csapi.TpAoCOrder.NetworkCharge",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpSimpleAttributeValue_TpSimpleAttributeValue, {"TpSimpleAttributeValue","giop-parlay.org.csapi.TpSimpleAttributeValue",FT_UINT32,BASE_DEC,VALS(org_csapi_TpSimpleAttributeTypeInfo),0x0,NULL,HFILL}},
        {&hf_org_csapi_TpSimpleAttributeValue_BooleanValue, {"TpSimpleAttributeValue_BooleanValue","giop-parlay.org.csapi.TpSimpleAttributeValue.BooleanValue",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_TpSimpleAttributeValue_OctetValue, {"TpSimpleAttributeValue_OctetValue","giop-parlay.org.csapi.TpSimpleAttributeValue.OctetValue",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpSimpleAttributeValue_CharValue, {"TpSimpleAttributeValue_CharValue","giop-parlay.org.csapi.TpSimpleAttributeValue.CharValue",FT_UINT8,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpSimpleAttributeValue_WCharValue_len, {"TpSimpleAttributeValue_WCharValue Length","giop-parlay.org.csapi.TpSimpleAttributeValue.WCharValue.len",FT_UINT8,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpSimpleAttributeValue_WCharValue, {"TpSimpleAttributeValue_WCharValue","giop-parlay.org.csapi.TpSimpleAttributeValue.WCharValue",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpSimpleAttributeValue_StringValue, {"TpSimpleAttributeValue_StringValue","giop-parlay.org.csapi.TpSimpleAttributeValue.StringValue",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpSimpleAttributeValue_WStringValue, {"TpSimpleAttributeValue_WStringValue","giop-parlay.org.csapi.TpSimpleAttributeValue.WStringValue",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpSimpleAttributeValue_Int16Value, {"TpSimpleAttributeValue_Int16Value","giop-parlay.org.csapi.TpSimpleAttributeValue.Int16Value",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpSimpleAttributeValue_UnsignedInt16Value, {"TpSimpleAttributeValue_UnsignedInt16Value","giop-parlay.org.csapi.TpSimpleAttributeValue.UnsignedInt16Value",FT_UINT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpSimpleAttributeValue_Int32Value, {"TpSimpleAttributeValue_Int32Value","giop-parlay.org.csapi.TpSimpleAttributeValue.Int32Value",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpSimpleAttributeValue_UnsignedInt32Value, {"TpSimpleAttributeValue_UnsignedInt32Value","giop-parlay.org.csapi.TpSimpleAttributeValue.UnsignedInt32Value",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpSimpleAttributeValue_Int64Value, {"TpSimpleAttributeValue_Int64Value","giop-parlay.org.csapi.TpSimpleAttributeValue.Int64Value",FT_INT64,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpSimpleAttributeValue_UnsignedInt64Value, {"TpSimpleAttributeValue_UnsignedInt64Value","giop-parlay.org.csapi.TpSimpleAttributeValue.UnsignedInt64Value",FT_UINT64,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpSimpleAttributeValue_FloatValue, {"TpSimpleAttributeValue_FloatValue","giop-parlay.org.csapi.TpSimpleAttributeValue.FloatValue",FT_FLOAT,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpSimpleAttributeValue_DoubleValue, {"TpSimpleAttributeValue_DoubleValue","giop-parlay.org.csapi.TpSimpleAttributeValue.DoubleValue",FT_DOUBLE,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_TpAttributeValue_TpAttributeValue, {"TpAttributeValue","giop-parlay.org.csapi.TpAttributeValue",FT_UINT32,BASE_DEC,VALS(org_csapi_TpAttributeTagInfo),0x0,NULL,HFILL}},
        {&hf_org_csapi_TpAttributeValue_XMLValue, {"TpAttributeValue_XMLValue","giop-parlay.org.csapi.TpAttributeValue.XMLValue",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventCriteria_TpFwEventCriteria, {"TpFwEventCriteria","giop-parlay.org.csapi.fw.TpFwEventCriteria",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpFwEventName),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventCriteria_EventNameUndefined, {"TpFwEventCriteria_EventNameUndefined","giop-parlay.org.csapi.fw.TpFwEventCriteria.EventNameUndefined",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventCriteria_ServiceTypeNameList_loop, {"Seq length of TpFwEventCriteria_ServiceTypeNameList","giop-parlay.org.csapi.fw.TpFwEventCriteria.ServiceTypeNameList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventCriteria_ServiceTypeNameList, {"TpFwEventCriteria_ServiceTypeNameList","giop-parlay.org.csapi.fw.TpFwEventCriteria.ServiceTypeNameList",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventCriteria_UnavailableServiceTypeNameList_loop, {"Seq length of TpFwEventCriteria_UnavailableServiceTypeNameList","giop-parlay.org.csapi.fw.TpFwEventCriteria.UnavailableServiceTypeNameList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventCriteria_UnavailableServiceTypeNameList, {"TpFwEventCriteria_UnavailableServiceTypeNameList","giop-parlay.org.csapi.fw.TpFwEventCriteria.UnavailableServiceTypeNameList",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventCriteria_CompatibleServiceTypeNameList_loop, {"Seq length of TpFwEventCriteria_CompatibleServiceTypeNameList","giop-parlay.org.csapi.fw.TpFwEventCriteria.CompatibleServiceTypeNameList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventCriteria_CompatibleServiceTypeNameList, {"TpFwEventCriteria_CompatibleServiceTypeNameList","giop-parlay.org.csapi.fw.TpFwEventCriteria.CompatibleServiceTypeNameList",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventCriteria_SessionCreatedList_loop, {"Seq length of TpFwEventCriteria_SessionCreatedList","giop-parlay.org.csapi.fw.TpFwEventCriteria.SessionCreatedList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventCriteria_SessionCreatedList, {"TpFwEventCriteria_SessionCreatedList","giop-parlay.org.csapi.fw.TpFwEventCriteria.SessionCreatedList",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventCriteria_SessionTerminatedList_loop, {"Seq length of TpFwEventCriteria_SessionTerminatedList","giop-parlay.org.csapi.fw.TpFwEventCriteria.SessionTerminatedList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventCriteria_SessionTerminatedList, {"TpFwEventCriteria_SessionTerminatedList","giop-parlay.org.csapi.fw.TpFwEventCriteria.SessionTerminatedList",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventCriteria_AgreementSignedList_loop, {"Seq length of TpFwEventCriteria_AgreementSignedList","giop-parlay.org.csapi.fw.TpFwEventCriteria.AgreementSignedList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventCriteria_AgreementSignedList, {"TpFwEventCriteria_AgreementSignedList","giop-parlay.org.csapi.fw.TpFwEventCriteria.AgreementSignedList",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventCriteria_AgreementEndedList_loop, {"Seq length of TpFwEventCriteria_AgreementEndedList","giop-parlay.org.csapi.fw.TpFwEventCriteria.AgreementEndedList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventCriteria_AgreementEndedList, {"TpFwEventCriteria_AgreementEndedList","giop-parlay.org.csapi.fw.TpFwEventCriteria.AgreementEndedList",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpLoadStatisticEntityID_TpLoadStatisticEntityID, {"TpLoadStatisticEntityID","giop-parlay.org.csapi.fw.TpLoadStatisticEntityID",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpLoadStatisticEntityType),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpLoadStatisticEntityID_FrameworkID, {"TpLoadStatisticEntityID_FrameworkID","giop-parlay.org.csapi.fw.TpLoadStatisticEntityID.FrameworkID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpLoadStatisticEntityID_ServiceID, {"TpLoadStatisticEntityID_ServiceID","giop-parlay.org.csapi.fw.TpLoadStatisticEntityID.ServiceID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpLoadStatisticEntityID_ClientAppID, {"TpLoadStatisticEntityID_ClientAppID","giop-parlay.org.csapi.fw.TpLoadStatisticEntityID.ClientAppID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpLoadStatisticInfo_TpLoadStatisticInfo, {"TpLoadStatisticInfo","giop-parlay.org.csapi.fw.TpLoadStatisticInfo",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpLoadStatisticInfoType),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpLoadStatisticInfo_LoadStatisticError, {"TpLoadStatisticInfo_LoadStatisticError","giop-parlay.org.csapi.fw.TpLoadStatisticInfo.LoadStatisticError",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpLoadStatisticError),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpDomainID_TpDomainID, {"TpDomainID","giop-parlay.org.csapi.fw.TpDomainID",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpDomainIDType),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpDomainID_FwID, {"TpDomainID_FwID","giop-parlay.org.csapi.fw.TpDomainID.FwID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpDomainID_ClientAppID, {"TpDomainID_ClientAppID","giop-parlay.org.csapi.fw.TpDomainID.ClientAppID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpDomainID_EntOpID, {"TpDomainID_EntOpID","giop-parlay.org.csapi.fw.TpDomainID.EntOpID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpDomainID_ServiceID, {"TpDomainID_ServiceID","giop-parlay.org.csapi.fw.TpDomainID.ServiceID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpDomainID_ServiceSupplierID, {"TpDomainID_ServiceSupplierID","giop-parlay.org.csapi.fw.TpDomainID.ServiceSupplierID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpMigrationAdditionalInfo_TpMigrationAdditionalInfo, {"TpMigrationAdditionalInfo","giop-parlay.org.csapi.fw.TpMigrationAdditionalInfo",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpMigrationAdditionalInfoType),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpMigrationAdditionalInfo_Dummy, {"TpMigrationAdditionalInfo_Dummy","giop-parlay.org.csapi.fw.TpMigrationAdditionalInfo.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventInfo_TpFwEventInfo, {"TpFwEventInfo","giop-parlay.org.csapi.fw.TpFwEventInfo",FT_UINT32,BASE_DEC,VALS(org_csapi_fw_TpFwEventName),0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventInfo_EventNameUndefined, {"TpFwEventInfo_EventNameUndefined","giop-parlay.org.csapi.fw.TpFwEventInfo.EventNameUndefined",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventInfo_ServiceIDList_loop, {"Seq length of TpFwEventInfo_ServiceIDList","giop-parlay.org.csapi.fw.TpFwEventInfo.ServiceIDList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventInfo_ServiceIDList, {"TpFwEventInfo_ServiceIDList","giop-parlay.org.csapi.fw.TpFwEventInfo.ServiceIDList",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventInfo_UnavailableServiceIDList_loop, {"Seq length of TpFwEventInfo_UnavailableServiceIDList","giop-parlay.org.csapi.fw.TpFwEventInfo.UnavailableServiceIDList.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventInfo_UnavailableServiceIDList, {"TpFwEventInfo_UnavailableServiceIDList","giop-parlay.org.csapi.fw.TpFwEventInfo.UnavailableServiceIDList",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventInfo_AppSessionCreated, {"TpFwEventInfo_AppSessionCreated","giop-parlay.org.csapi.fw.TpFwEventInfo.AppSessionCreated",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_fw_TpFwEventInfo_AppSessionTerminated, {"TpFwEventInfo_AppSessionTerminated","giop-parlay.org.csapi.fw.TpFwEventInfo.AppSessionTerminated",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIVariableInfo_TpUIVariableInfo, {"TpUIVariableInfo","giop-parlay.org.csapi.ui.TpUIVariableInfo",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUIVariablePartType),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIVariableInfo_VariablePartInteger, {"TpUIVariableInfo_VariablePartInteger","giop-parlay.org.csapi.ui.TpUIVariableInfo.VariablePartInteger",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIVariableInfo_VariablePartAddress, {"TpUIVariableInfo_VariablePartAddress","giop-parlay.org.csapi.ui.TpUIVariableInfo.VariablePartAddress",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIVariableInfo_VariablePartTime, {"TpUIVariableInfo_VariablePartTime","giop-parlay.org.csapi.ui.TpUIVariableInfo.VariablePartTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIVariableInfo_VariablePartDate, {"TpUIVariableInfo_VariablePartDate","giop-parlay.org.csapi.ui.TpUIVariableInfo.VariablePartDate",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIVariableInfo_VariablePartPrice, {"TpUIVariableInfo_VariablePartPrice","giop-parlay.org.csapi.ui.TpUIVariableInfo.VariablePartPrice",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIInfo_TpUIInfo, {"TpUIInfo","giop-parlay.org.csapi.ui.TpUIInfo",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUIInfoType),0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIInfo_InfoID, {"TpUIInfo_InfoID","giop-parlay.org.csapi.ui.TpUIInfo.InfoID",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIInfo_InfoData, {"TpUIInfo_InfoData","giop-parlay.org.csapi.ui.TpUIInfo.InfoData",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIInfo_InfoAddress, {"TpUIInfo_InfoAddress","giop-parlay.org.csapi.ui.TpUIInfo.InfoAddress",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIInfo_InfoBinData_loop, {"Seq length of TpUIInfo_InfoBinData","giop-parlay.org.csapi.ui.TpUIInfo.InfoBinData.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIInfo_InfoBinData, {"TpUIInfo_InfoBinData","giop-parlay.org.csapi.ui.TpUIInfo.InfoBinData",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIInfo_InfoUUEncData, {"TpUIInfo_InfoUUEncData","giop-parlay.org.csapi.ui.TpUIInfo.InfoUUEncData",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIInfo_InfoMimeData_loop, {"Seq length of TpUIInfo_InfoMimeData","giop-parlay.org.csapi.ui.TpUIInfo.InfoMimeData.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIInfo_InfoMimeData, {"TpUIInfo_InfoMimeData","giop-parlay.org.csapi.ui.TpUIInfo.InfoMimeData",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIInfo_InfoWaveData_loop, {"Seq length of TpUIInfo_InfoWaveData","giop-parlay.org.csapi.ui.TpUIInfo.InfoWaveData.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIInfo_InfoWaveData, {"TpUIInfo_InfoWaveData","giop-parlay.org.csapi.ui.TpUIInfo.InfoWaveData",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIInfo_InfoAuData_loop, {"Seq length of TpUIInfo_InfoAuData","giop-parlay.org.csapi.ui.TpUIInfo.InfoAuData.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIInfo_InfoAuData, {"TpUIInfo_InfoAuData","giop-parlay.org.csapi.ui.TpUIInfo.InfoAuData",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUIInfo_InfoVXMLData, {"TpUIInfo_InfoVXMLData","giop-parlay.org.csapi.ui.TpUIInfo.InfoVXMLData",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallAdditionalErrorInfo_TpCallAdditionalErrorInfo, {"TpCallAdditionalErrorInfo","giop-parlay.org.csapi.cc.TpCallAdditionalErrorInfo",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallErrorType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallAdditionalErrorInfo_CallErrorInvalidAddress, {"TpCallAdditionalErrorInfo_CallErrorInvalidAddress","giop-parlay.org.csapi.cc.TpCallAdditionalErrorInfo.CallErrorInvalidAddress",FT_UINT32,BASE_DEC,VALS(org_csapi_TpAddressError),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallAdditionalErrorInfo_Dummy, {"TpCallAdditionalErrorInfo_Dummy","giop-parlay.org.csapi.cc.TpCallAdditionalErrorInfo.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallLoadControlMechanism_TpCallLoadControlMechanism, {"TpCallLoadControlMechanism","giop-parlay.org.csapi.cc.TpCallLoadControlMechanism",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallLoadControlMechanismType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallLoadControlMechanism_CallLoadControlPerInterval, {"TpCallLoadControlMechanism_CallLoadControlPerInterval","giop-parlay.org.csapi.cc.TpCallLoadControlMechanism.CallLoadControlPerInterval",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallAdditionalTreatmentInfo_TpCallAdditionalTreatmentInfo, {"TpCallAdditionalTreatmentInfo","giop-parlay.org.csapi.cc.TpCallAdditionalTreatmentInfo",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallTreatmentType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallAdditionalTreatmentInfo_Dummy, {"TpCallAdditionalTreatmentInfo_Dummy","giop-parlay.org.csapi.cc.TpCallAdditionalTreatmentInfo.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallPartyToChargeAdditionalInfo_TpCallPartyToChargeAdditionalInfo, {"TpCallPartyToChargeAdditionalInfo","giop-parlay.org.csapi.cc.TpCallPartyToChargeAdditionalInfo",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallPartyToChargeType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallPartyToChargeAdditionalInfo_Dummy, {"TpCallPartyToChargeAdditionalInfo_Dummy","giop-parlay.org.csapi.cc.TpCallPartyToChargeAdditionalInfo.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallAppInfo_TpCallAppInfo, {"TpCallAppInfo","giop-parlay.org.csapi.cc.gccs.TpCallAppInfo",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_gccs_TpCallAppInfoType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallAppInfo_CallAppAlertingMechanism, {"TpCallAppInfo_CallAppAlertingMechanism","giop-parlay.org.csapi.cc.gccs.TpCallAppInfo.CallAppAlertingMechanism",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallAppInfo_CallAppNetworkAccessType, {"TpCallAppInfo_CallAppNetworkAccessType","giop-parlay.org.csapi.cc.gccs.TpCallAppInfo.CallAppNetworkAccessType",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallNetworkAccessType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallAppInfo_CallAppTeleService, {"TpCallAppInfo_CallAppTeleService","giop-parlay.org.csapi.cc.gccs.TpCallAppInfo.CallAppTeleService",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallTeleService),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallAppInfo_CallAppBearerService, {"TpCallAppInfo_CallAppBearerService","giop-parlay.org.csapi.cc.gccs.TpCallAppInfo.CallAppBearerService",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallBearerService),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallAppInfo_CallAppPartyCategory, {"TpCallAppInfo_CallAppPartyCategory","giop-parlay.org.csapi.cc.gccs.TpCallAppInfo.CallAppPartyCategory",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallPartyCategory),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallAppInfo_CallAppGenericInfo, {"TpCallAppInfo_CallAppGenericInfo","giop-parlay.org.csapi.cc.gccs.TpCallAppInfo.CallAppGenericInfo",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallAppInfo_Dummy, {"TpCallAppInfo_Dummy","giop-parlay.org.csapi.cc.gccs.TpCallAppInfo.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallAdditionalReportInfo_TpCallAdditionalReportInfo, {"TpCallAdditionalReportInfo","giop-parlay.org.csapi.cc.gccs.TpCallAdditionalReportInfo",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_gccs_TpCallReportType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallAdditionalReportInfo_QueueStatus, {"TpCallAdditionalReportInfo_QueueStatus","giop-parlay.org.csapi.cc.gccs.TpCallAdditionalReportInfo.QueueStatus",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallAdditionalReportInfo_Dummy, {"TpCallAdditionalReportInfo_Dummy","giop-parlay.org.csapi.cc.gccs.TpCallAdditionalReportInfo.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallAdditionalReportCriteria_TpCallAdditionalReportCriteria, {"TpCallAdditionalReportCriteria","giop-parlay.org.csapi.cc.gccs.TpCallAdditionalReportCriteria",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_gccs_TpCallReportType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallAdditionalReportCriteria_NoAnswerDuration, {"TpCallAdditionalReportCriteria_NoAnswerDuration","giop-parlay.org.csapi.cc.gccs.TpCallAdditionalReportCriteria.NoAnswerDuration",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_gccs_TpCallAdditionalReportCriteria_Dummy, {"TpCallAdditionalReportCriteria_Dummy","giop-parlay.org.csapi.cc.gccs.TpCallAdditionalReportCriteria.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallAdditionalEventInfo_TpCallAdditionalEventInfo, {"TpCallAdditionalEventInfo","giop-parlay.org.csapi.cc.TpCallAdditionalEventInfo",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallEventType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallAdditionalEventInfo_OriginatingReleaseCause, {"TpCallAdditionalEventInfo_OriginatingReleaseCause","giop-parlay.org.csapi.cc.TpCallAdditionalEventInfo.OriginatingReleaseCause",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpReleaseCause),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallAdditionalEventInfo_TerminatingReleaseCause, {"TpCallAdditionalEventInfo_TerminatingReleaseCause","giop-parlay.org.csapi.cc.TpCallAdditionalEventInfo.TerminatingReleaseCause",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpReleaseCause),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallAdditionalEventInfo_Dummy, {"TpCallAdditionalEventInfo_Dummy","giop-parlay.org.csapi.cc.TpCallAdditionalEventInfo.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpAdditionalCallEventCriteria_TpAdditionalCallEventCriteria, {"TpAdditionalCallEventCriteria","giop-parlay.org.csapi.cc.TpAdditionalCallEventCriteria",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallEventType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpAdditionalCallEventCriteria_MinAddressLength, {"TpAdditionalCallEventCriteria_MinAddressLength","giop-parlay.org.csapi.cc.TpAdditionalCallEventCriteria.MinAddressLength",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingServiceCode_loop, {"Seq length of TpAdditionalCallEventCriteria_OriginatingServiceCode","giop-parlay.org.csapi.cc.TpAdditionalCallEventCriteria.OriginatingServiceCode.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingReleaseCauseSet_loop, {"Seq length of TpAdditionalCallEventCriteria_OriginatingReleaseCauseSet","giop-parlay.org.csapi.cc.TpAdditionalCallEventCriteria.OriginatingReleaseCauseSet.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpAdditionalCallEventCriteria_OriginatingReleaseCauseSet, {"TpAdditionalCallEventCriteria_OriginatingReleaseCauseSet","giop-parlay.org.csapi.cc.TpAdditionalCallEventCriteria.OriginatingReleaseCauseSet",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpReleaseCause),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingReleaseCauseSet_loop, {"Seq length of TpAdditionalCallEventCriteria_TerminatingReleaseCauseSet","giop-parlay.org.csapi.cc.TpAdditionalCallEventCriteria.TerminatingReleaseCauseSet.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingReleaseCauseSet, {"TpAdditionalCallEventCriteria_TerminatingReleaseCauseSet","giop-parlay.org.csapi.cc.TpAdditionalCallEventCriteria.TerminatingReleaseCauseSet",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpReleaseCause),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpAdditionalCallEventCriteria_TerminatingServiceCode_loop, {"Seq length of TpAdditionalCallEventCriteria_TerminatingServiceCode","giop-parlay.org.csapi.cc.TpAdditionalCallEventCriteria.TerminatingServiceCode.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpAdditionalCallEventCriteria_Dummy, {"TpAdditionalCallEventCriteria_Dummy","giop-parlay.org.csapi.cc.TpAdditionalCallEventCriteria.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallAppInfo_TpCallAppInfo, {"TpCallAppInfo","giop-parlay.org.csapi.cc.TpCallAppInfo",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallAppInfoType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallAppInfo_CallAppAlertingMechanism, {"TpCallAppInfo_CallAppAlertingMechanism","giop-parlay.org.csapi.cc.TpCallAppInfo.CallAppAlertingMechanism",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallAppInfo_CallAppNetworkAccessType, {"TpCallAppInfo_CallAppNetworkAccessType","giop-parlay.org.csapi.cc.TpCallAppInfo.CallAppNetworkAccessType",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallNetworkAccessType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallAppInfo_CallAppTeleService, {"TpCallAppInfo_CallAppTeleService","giop-parlay.org.csapi.cc.TpCallAppInfo.CallAppTeleService",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallTeleService),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallAppInfo_CallAppBearerService, {"TpCallAppInfo_CallAppBearerService","giop-parlay.org.csapi.cc.TpCallAppInfo.CallAppBearerService",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallBearerService),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallAppInfo_CallAppPartyCategory, {"TpCallAppInfo_CallAppPartyCategory","giop-parlay.org.csapi.cc.TpCallAppInfo.CallAppPartyCategory",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_TpCallPartyCategory),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallAppInfo_CallAppGenericInfo, {"TpCallAppInfo_CallAppGenericInfo","giop-parlay.org.csapi.cc.TpCallAppInfo.CallAppGenericInfo",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallAppInfo_CallHighProbabilityCompletion, {"TpCallAppInfo_CallHighProbabilityCompletion","giop-parlay.org.csapi.cc.TpCallAppInfo.CallHighProbabilityCompletion",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallAppInfo_CallAppCarrier_loop, {"Seq length of TpCallAppInfo_CallAppCarrier","giop-parlay.org.csapi.cc.TpCallAppInfo.CallAppCarrier.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_TpCallAppInfo_Dummy, {"TpCallAppInfo_Dummy","giop-parlay.org.csapi.cc.TpCallAppInfo.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_TpAppMultiPartyCallBack_TpAppMultiPartyCallBack, {"TpAppMultiPartyCallBack","giop-parlay.org.csapi.cc.mpccs.TpAppMultiPartyCallBack",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_mpccs_TpAppMultiPartyCallBackRefType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mpccs_TpAppMultiPartyCallBack_Dummy, {"TpAppMultiPartyCallBack_Dummy","giop-parlay.org.csapi.cc.mpccs.TpAppMultiPartyCallBack.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_TpMediaStreamDataTypeRequest_TpMediaStreamDataTypeRequest, {"TpMediaStreamDataTypeRequest","giop-parlay.org.csapi.cc.mmccs.TpMediaStreamDataTypeRequest",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_mmccs_TpMediaStreamDataTypeRequestType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_TpMediaStreamDataTypeRequest_Audio, {"TpMediaStreamDataTypeRequest_Audio","giop-parlay.org.csapi.cc.mmccs.TpMediaStreamDataTypeRequest.Audio",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_TpMediaStreamDataTypeRequest_Video, {"TpMediaStreamDataTypeRequest_Video","giop-parlay.org.csapi.cc.mmccs.TpMediaStreamDataTypeRequest.Video",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_TpMediaStreamDataTypeRequest_Data, {"TpMediaStreamDataTypeRequest_Data","giop-parlay.org.csapi.cc.mmccs.TpMediaStreamDataTypeRequest.Data",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_TpAppMultiMediaCallBack_TpAppMultiMediaCallBack, {"TpAppMultiMediaCallBack","giop-parlay.org.csapi.cc.mmccs.TpAppMultiMediaCallBack",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_mmccs_TpAppMultiMediaCallBackRefType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_mmccs_TpAppMultiMediaCallBack_Dummy, {"TpAppMultiMediaCallBack_Dummy","giop-parlay.org.csapi.cc.mmccs.TpAppMultiMediaCallBack.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_TpConfPolicy_TpConfPolicy, {"TpConfPolicy","giop-parlay.org.csapi.cc.cccs.TpConfPolicy",FT_UINT32,BASE_DEC,VALS(org_csapi_cc_cccs_TpConfPolicyType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cc_cccs_TpConfPolicy_Dummy, {"TpConfPolicy_Dummy","giop-parlay.org.csapi.cc.cccs.TpConfPolicy.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_ui_TpUITargetObject_TpUITargetObject, {"TpUITargetObject","giop-parlay.org.csapi.ui.TpUITargetObject",FT_UINT32,BASE_DEC,VALS(org_csapi_ui_TpUITargetObjectType),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpBindingEntry_TpBindingEntry, {"TpBindingEntry","giop-parlay.org.csapi.mm.TpBindingEntry",FT_UINT32,BASE_DEC,VALS(org_csapi_mm_TpBindingEntryType),0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpBindingEntry_UBExpires, {"TpBindingEntry_UBExpires","giop-parlay.org.csapi.mm.TpBindingEntry.UBExpires",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpBindingEntry_UBPreference, {"TpBindingEntry_UBPreference","giop-parlay.org.csapi.mm.TpBindingEntry.UBPreference",FT_FLOAT,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mm_TpBindingEntry_Dummy, {"TpBindingEntry_Dummy","giop-parlay.org.csapi.mm.TpBindingEntry.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionChargeOrder_TpDataSessionChargeOrder, {"TpDataSessionChargeOrder","giop-parlay.org.csapi.dsc.TpDataSessionChargeOrder",FT_UINT32,BASE_DEC,VALS(org_csapi_dsc_TpDataSessionChargeOrderCategory),0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionChargeOrder_NetworkCharge, {"TpDataSessionChargeOrder_NetworkCharge","giop-parlay.org.csapi.dsc.TpDataSessionChargeOrder.NetworkCharge",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionAdditionalErrorInfo_TpDataSessionAdditionalErrorInfo, {"TpDataSessionAdditionalErrorInfo","giop-parlay.org.csapi.dsc.TpDataSessionAdditionalErrorInfo",FT_UINT32,BASE_DEC,VALS(org_csapi_dsc_TpDataSessionErrorType),0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionAdditionalErrorInfo_DataSessionErrorInvalidAddress, {"TpDataSessionAdditionalErrorInfo_DataSessionErrorInvalidAddress","giop-parlay.org.csapi.dsc.TpDataSessionAdditionalErrorInfo.DataSessionErrorInvalidAddress",FT_UINT32,BASE_DEC,VALS(org_csapi_TpAddressError),0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionAdditionalErrorInfo_Dummy, {"TpDataSessionAdditionalErrorInfo_Dummy","giop-parlay.org.csapi.dsc.TpDataSessionAdditionalErrorInfo.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionAdditionalReportInfo_TpDataSessionAdditionalReportInfo, {"TpDataSessionAdditionalReportInfo","giop-parlay.org.csapi.dsc.TpDataSessionAdditionalReportInfo",FT_UINT32,BASE_DEC,VALS(org_csapi_dsc_TpDataSessionReportType),0x0,NULL,HFILL}},
        {&hf_org_csapi_dsc_TpDataSessionAdditionalReportInfo_Dummy, {"TpDataSessionAdditionalReportInfo_Dummy","giop-parlay.org.csapi.dsc.TpDataSessionAdditionalReportInfo.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMessageInfoProperty_TpMessageInfoProperty, {"TpMessageInfoProperty","giop-parlay.org.csapi.gms.TpMessageInfoProperty",FT_UINT32,BASE_DEC,VALS(org_csapi_gms_TpMessageInfoPropertyName),0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageID, {"TpMessageInfoProperty_MessagingMessageID","giop-parlay.org.csapi.gms.TpMessageInfoProperty.MessagingMessageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageSubject, {"TpMessageInfoProperty_MessagingMessageSubject","giop-parlay.org.csapi.gms.TpMessageInfoProperty.MessagingMessageSubject",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageDateSent, {"TpMessageInfoProperty_MessagingMessageDateSent","giop-parlay.org.csapi.gms.TpMessageInfoProperty.MessagingMessageDateSent",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageDateReceived, {"TpMessageInfoProperty_MessagingMessageDateReceived","giop-parlay.org.csapi.gms.TpMessageInfoProperty.MessagingMessageDateReceived",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageDateChanged, {"TpMessageInfoProperty_MessagingMessageDateChanged","giop-parlay.org.csapi.gms.TpMessageInfoProperty.MessagingMessageDateChanged",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageSize, {"TpMessageInfoProperty_MessagingMessageSize","giop-parlay.org.csapi.gms.TpMessageInfoProperty.MessagingMessageSize",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessagePriority, {"TpMessageInfoProperty_MessagingMessagePriority","giop-parlay.org.csapi.gms.TpMessageInfoProperty.MessagingMessagePriority",FT_UINT32,BASE_DEC,VALS(org_csapi_gms_TpMessagePriority),0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageFormat, {"TpMessageInfoProperty_MessagingMessageFormat","giop-parlay.org.csapi.gms.TpMessageInfoProperty.MessagingMessageFormat",FT_UINT32,BASE_DEC,VALS(org_csapi_gms_TpMessageFormat),0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageFolder, {"TpMessageInfoProperty_MessagingMessageFolder","giop-parlay.org.csapi.gms.TpMessageInfoProperty.MessagingMessageFolder",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMessageInfoProperty_MessagingMessageStatus, {"TpMessageInfoProperty_MessagingMessageStatus","giop-parlay.org.csapi.gms.TpMessageInfoProperty.MessagingMessageStatus",FT_UINT32,BASE_DEC,VALS(org_csapi_gms_TpMessageStatus),0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMessageInfoProperty_Dummy, {"TpMessageInfoProperty_Dummy","giop-parlay.org.csapi.gms.TpMessageInfoProperty.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMailboxInfoProperty_TpMailboxInfoProperty, {"TpMailboxInfoProperty","giop-parlay.org.csapi.gms.TpMailboxInfoProperty",FT_UINT32,BASE_DEC,VALS(org_csapi_gms_TpMailboxInfoPropertyName),0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMailboxInfoProperty_MessagingMailboxOwner, {"TpMailboxInfoProperty_MessagingMailboxOwner","giop-parlay.org.csapi.gms.TpMailboxInfoProperty.MessagingMailboxOwner",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMailboxInfoProperty_MessagingMailboxFolder, {"TpMailboxInfoProperty_MessagingMailboxFolder","giop-parlay.org.csapi.gms.TpMailboxInfoProperty.MessagingMailboxFolder",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMailboxInfoProperty_MessagingMailboxDateCreated, {"TpMailboxInfoProperty_MessagingMailboxDateCreated","giop-parlay.org.csapi.gms.TpMailboxInfoProperty.MessagingMailboxDateCreated",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMailboxInfoProperty_MessagingMailboxDateChanged, {"TpMailboxInfoProperty_MessagingMailboxDateChanged","giop-parlay.org.csapi.gms.TpMailboxInfoProperty.MessagingMailboxDateChanged",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMailboxInfoProperty_Dummy, {"TpMailboxInfoProperty_Dummy","giop-parlay.org.csapi.gms.TpMailboxInfoProperty.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMessagingEventInfo_TpMessagingEventInfo, {"TpMessagingEventInfo","giop-parlay.org.csapi.gms.TpMessagingEventInfo",FT_UINT32,BASE_DEC,VALS(org_csapi_gms_TpMessagingEventName),0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMessagingEventInfo_EventNameUndefined, {"TpMessagingEventInfo_EventNameUndefined","giop-parlay.org.csapi.gms.TpMessagingEventInfo.EventNameUndefined",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMessagingEventCriteria_TpMessagingEventCriteria, {"TpMessagingEventCriteria","giop-parlay.org.csapi.gms.TpMessagingEventCriteria",FT_UINT32,BASE_DEC,VALS(org_csapi_gms_TpMessagingEventName),0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpMessagingEventCriteria_Dummy, {"TpMessagingEventCriteria_Dummy","giop-parlay.org.csapi.gms.TpMessagingEventCriteria.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpFolderInfoProperty_TpFolderInfoProperty, {"TpFolderInfoProperty","giop-parlay.org.csapi.gms.TpFolderInfoProperty",FT_UINT32,BASE_DEC,VALS(org_csapi_gms_TpFolderInfoPropertyName),0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpFolderInfoProperty_MessagingFolderID, {"TpFolderInfoProperty_MessagingFolderID","giop-parlay.org.csapi.gms.TpFolderInfoProperty.MessagingFolderID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpFolderInfoProperty_MessagingFolderMessage, {"TpFolderInfoProperty_MessagingFolderMessage","giop-parlay.org.csapi.gms.TpFolderInfoProperty.MessagingFolderMessage",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpFolderInfoProperty_MessagingFolderSubfolder, {"TpFolderInfoProperty_MessagingFolderSubfolder","giop-parlay.org.csapi.gms.TpFolderInfoProperty.MessagingFolderSubfolder",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpFolderInfoProperty_MessagingFolderDateCreated, {"TpFolderInfoProperty_MessagingFolderDateCreated","giop-parlay.org.csapi.gms.TpFolderInfoProperty.MessagingFolderDateCreated",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpFolderInfoProperty_MessagingFolderDateChanged, {"TpFolderInfoProperty_MessagingFolderDateChanged","giop-parlay.org.csapi.gms.TpFolderInfoProperty.MessagingFolderDateChanged",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_gms_TpFolderInfoProperty_Dummy, {"TpFolderInfoProperty_Dummy","giop-parlay.org.csapi.gms.TpFolderInfoProperty.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_TpChargingParameterValue_TpChargingParameterValue, {"TpChargingParameterValue","giop-parlay.org.csapi.cs.TpChargingParameterValue",FT_UINT32,BASE_DEC,VALS(org_csapi_cs_TpChargingParameterValueType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_TpChargingParameterValue_IntValue, {"TpChargingParameterValue_IntValue","giop-parlay.org.csapi.cs.TpChargingParameterValue.IntValue",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_TpChargingParameterValue_FloatValue, {"TpChargingParameterValue_FloatValue","giop-parlay.org.csapi.cs.TpChargingParameterValue.FloatValue",FT_FLOAT,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_TpChargingParameterValue_StringValue, {"TpChargingParameterValue_StringValue","giop-parlay.org.csapi.cs.TpChargingParameterValue.StringValue",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_TpChargingParameterValue_BooleanValue, {"TpChargingParameterValue_BooleanValue","giop-parlay.org.csapi.cs.TpChargingParameterValue.BooleanValue",FT_BOOLEAN,8,NULL,0x01,NULL,HFILL}},
        {&hf_org_csapi_cs_TpChargingParameterValue_OctetValue_loop, {"Seq length of TpChargingParameterValue_OctetValue","giop-parlay.org.csapi.cs.TpChargingParameterValue.OctetValue.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_TpChargingParameterValue_OctetValue, {"TpChargingParameterValue_OctetValue","giop-parlay.org.csapi.cs.TpChargingParameterValue.OctetValue",FT_UINT8,BASE_HEX,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_TpAppInformation_TpAppInformation, {"TpAppInformation","giop-parlay.org.csapi.cs.TpAppInformation",FT_UINT32,BASE_DEC,VALS(org_csapi_cs_TpAppInformationType),0x0,NULL,HFILL}},
        {&hf_org_csapi_cs_TpAppInformation_Timestamp, {"TpAppInformation_Timestamp","giop-parlay.org.csapi.cs.TpAppInformation.Timestamp",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_TpPolicyType_TpPolicyType, {"TpPolicyType","giop-parlay.org.csapi.policy.TpPolicyType",FT_UINT32,BASE_DEC,VALS(org_csapi_policy_TpPolicyTypeInfo),0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_TpPolicyType_SimpleType, {"TpPolicyType_SimpleType","giop-parlay.org.csapi.policy.TpPolicyType.SimpleType",FT_UINT32,BASE_DEC,VALS(org_csapi_TpSimpleAttributeTypeInfo),0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_TpPolicyType_StructuredType, {"TpPolicyType_StructuredType","giop-parlay.org.csapi.policy.TpPolicyType.StructuredType",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_policy_TpPolicyType_XMLString, {"TpPolicyType_XMLString","giop-parlay.org.csapi.policy.TpPolicyType.XMLString",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMContextData_TpPAMContextData, {"TpPAMContextData","giop-parlay.org.csapi.pam.TpPAMContextData",FT_UINT32,BASE_DEC,VALS(org_csapi_pam_TpPAMContextName),0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMContextData_Dummy, {"TpPAMContextData_Dummy","giop-parlay.org.csapi.pam.TpPAMContextData.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMEventInfo_TpPAMEventInfo, {"TpPAMEventInfo","giop-parlay.org.csapi.pam.TpPAMEventInfo",FT_UINT32,BASE_DEC,VALS(org_csapi_pam_TpPAMEventName),0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMNotificationInfo_TpPAMNotificationInfo, {"TpPAMNotificationInfo","giop-parlay.org.csapi.pam.TpPAMNotificationInfo",FT_UINT32,BASE_DEC,VALS(org_csapi_pam_TpPAMEventName),0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMPreferenceData_TpPAMPreferenceData, {"TpPAMPreferenceData","giop-parlay.org.csapi.pam.TpPAMPreferenceData",FT_UINT32,BASE_DEC,VALS(org_csapi_pam_TpPAMPreferenceType),0x0,NULL,HFILL}},
        {&hf_org_csapi_pam_TpPAMPreferenceData_Dummy, {"TpPAMPreferenceData_Dummy","giop-parlay.org.csapi.pam.TpPAMPreferenceData.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpFolderInfoProperty_TpFolderInfoProperty, {"TpFolderInfoProperty","giop-parlay.org.csapi.mmm.TpFolderInfoProperty",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpFolderInfoPropertyName),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpFolderInfoProperty_FolderDateCreated, {"TpFolderInfoProperty_FolderDateCreated","giop-parlay.org.csapi.mmm.TpFolderInfoProperty.FolderDateCreated",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpFolderInfoProperty_FolderDateChanged, {"TpFolderInfoProperty_FolderDateChanged","giop-parlay.org.csapi.mmm.TpFolderInfoProperty.FolderDateChanged",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpFolderInfoProperty_FolderSize, {"TpFolderInfoProperty_FolderSize","giop-parlay.org.csapi.mmm.TpFolderInfoProperty.FolderSize",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpFolderInfoProperty_FolderNumberOfMessages, {"TpFolderInfoProperty_FolderNumberOfMessages","giop-parlay.org.csapi.mmm.TpFolderInfoProperty.FolderNumberOfMessages",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpFolderInfoProperty_Dummy, {"TpFolderInfoProperty_Dummy","giop-parlay.org.csapi.mmm.TpFolderInfoProperty.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMailboxInfoProperty_TpMailboxInfoProperty, {"TpMailboxInfoProperty","giop-parlay.org.csapi.mmm.TpMailboxInfoProperty",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMailboxInfoPropertyName),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMailboxInfoProperty_MailboxOwner, {"TpMailboxInfoProperty_MailboxOwner","giop-parlay.org.csapi.mmm.TpMailboxInfoProperty.MailboxOwner",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMailboxInfoProperty_MailboxDateCreated, {"TpMailboxInfoProperty_MailboxDateCreated","giop-parlay.org.csapi.mmm.TpMailboxInfoProperty.MailboxDateCreated",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMailboxInfoProperty_MailboxDateChanged, {"TpMailboxInfoProperty_MailboxDateChanged","giop-parlay.org.csapi.mmm.TpMailboxInfoProperty.MailboxDateChanged",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMailboxInfoProperty_MailboxSize, {"TpMailboxInfoProperty_MailboxSize","giop-parlay.org.csapi.mmm.TpMailboxInfoProperty.MailboxSize",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMailboxInfoProperty_Dummy, {"TpMailboxInfoProperty_Dummy","giop-parlay.org.csapi.mmm.TpMailboxInfoProperty.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageInfoProperty_TpMessageInfoProperty, {"TpMessageInfoProperty","giop-parlay.org.csapi.mmm.TpMessageInfoProperty",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessageInfoPropertyName),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageInfoProperty_MessageDateCreated, {"TpMessageInfoProperty_MessageDateCreated","giop-parlay.org.csapi.mmm.TpMessageInfoProperty.MessageDateCreated",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageInfoProperty_MessageDateReceived, {"TpMessageInfoProperty_MessageDateReceived","giop-parlay.org.csapi.mmm.TpMessageInfoProperty.MessageDateReceived",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageInfoProperty_MessageDateChanged, {"TpMessageInfoProperty_MessageDateChanged","giop-parlay.org.csapi.mmm.TpMessageInfoProperty.MessageDateChanged",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageInfoProperty_MessageSize, {"TpMessageInfoProperty_MessageSize","giop-parlay.org.csapi.mmm.TpMessageInfoProperty.MessageSize",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageInfoProperty_MessageStatus, {"TpMessageInfoProperty_MessageStatus","giop-parlay.org.csapi.mmm.TpMessageInfoProperty.MessageStatus",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMailboxMessageStatus),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageInfoProperty_Dummy, {"TpMessageInfoProperty_Dummy","giop-parlay.org.csapi.mmm.TpMessageInfoProperty.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageHeaderField_TpMessageHeaderField, {"TpMessageHeaderField","giop-parlay.org.csapi.mmm.TpMessageHeaderField",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessageHeaderFieldType),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageHeaderField_DateSent, {"TpMessageHeaderField_DateSent","giop-parlay.org.csapi.mmm.TpMessageHeaderField.DateSent",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageHeaderField_From_loop, {"Seq length of TpMessageHeaderField_From","giop-parlay.org.csapi.mmm.TpMessageHeaderField.From.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageHeaderField_ReplyTo_loop, {"Seq length of TpMessageHeaderField_ReplyTo","giop-parlay.org.csapi.mmm.TpMessageHeaderField.ReplyTo.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageHeaderField_To_loop, {"Seq length of TpMessageHeaderField_To","giop-parlay.org.csapi.mmm.TpMessageHeaderField.To.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageHeaderField_Cc_loop, {"Seq length of TpMessageHeaderField_Cc","giop-parlay.org.csapi.mmm.TpMessageHeaderField.Cc.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageHeaderField_Bcc_loop, {"Seq length of TpMessageHeaderField_Bcc","giop-parlay.org.csapi.mmm.TpMessageHeaderField.Bcc.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageHeaderField_RFC822MessageID, {"TpMessageHeaderField_RFC822MessageID","giop-parlay.org.csapi.mmm.TpMessageHeaderField.RFC822MessageID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageHeaderField_InReplyTo_loop, {"Seq length of TpMessageHeaderField_InReplyTo","giop-parlay.org.csapi.mmm.TpMessageHeaderField.InReplyTo.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageHeaderField_InReplyTo, {"TpMessageHeaderField_InReplyTo","giop-parlay.org.csapi.mmm.TpMessageHeaderField.InReplyTo",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageHeaderField_References_loop, {"Seq length of TpMessageHeaderField_References","giop-parlay.org.csapi.mmm.TpMessageHeaderField.References.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageHeaderField_References, {"TpMessageHeaderField_References","giop-parlay.org.csapi.mmm.TpMessageHeaderField.References",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageHeaderField_Subject, {"TpMessageHeaderField_Subject","giop-parlay.org.csapi.mmm.TpMessageHeaderField.Subject",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageHeaderField_Comments, {"TpMessageHeaderField_Comments","giop-parlay.org.csapi.mmm.TpMessageHeaderField.Comments",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageHeaderField_Keywords_loop, {"Seq length of TpMessageHeaderField_Keywords","giop-parlay.org.csapi.mmm.TpMessageHeaderField.Keywords.size",FT_UINT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageHeaderField_Keywords, {"TpMessageHeaderField_Keywords","giop-parlay.org.csapi.mmm.TpMessageHeaderField.Keywords",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageHeaderField_MimeVersion, {"TpMessageHeaderField_MimeVersion","giop-parlay.org.csapi.mmm.TpMessageHeaderField.MimeVersion",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageHeaderField_MimeContent, {"TpMessageHeaderField_MimeContent","giop-parlay.org.csapi.mmm.TpMessageHeaderField.MimeContent",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageHeaderField_MimeEncoding, {"TpMessageHeaderField_MimeEncoding","giop-parlay.org.csapi.mmm.TpMessageHeaderField.MimeEncoding",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageHeaderField_MimeID, {"TpMessageHeaderField_MimeID","giop-parlay.org.csapi.mmm.TpMessageHeaderField.MimeID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageHeaderField_MimeDescription, {"TpMessageHeaderField_MimeDescription","giop-parlay.org.csapi.mmm.TpMessageHeaderField.MimeDescription",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageHeaderField_MimeDisposition, {"TpMessageHeaderField_MimeDisposition","giop-parlay.org.csapi.mmm.TpMessageHeaderField.MimeDisposition",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageHeaderField_Priority, {"TpMessageHeaderField_Priority","giop-parlay.org.csapi.mmm.TpMessageHeaderField.Priority",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagePriority),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpDeliveryTime_TpDeliveryTime, {"TpDeliveryTime","giop-parlay.org.csapi.mmm.TpDeliveryTime",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpDeliveryTimeType),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpDeliveryTime_DeliveryTime, {"TpDeliveryTime_DeliveryTime","giop-parlay.org.csapi.mmm.TpDeliveryTime.DeliveryTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpDeliveryTime_Dummy, {"TpDeliveryTime_Dummy","giop-parlay.org.csapi.mmm.TpDeliveryTime.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageTreatment_TpMessageTreatment, {"TpMessageTreatment","giop-parlay.org.csapi.mmm.TpMessageTreatment",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessageTreatmentType),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageTreatment_DeliveryReport, {"TpMessageTreatment_DeliveryReport","giop-parlay.org.csapi.mmm.TpMessageTreatment.DeliveryReport",FT_INT32,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageTreatment_BillingID, {"TpMessageTreatment_BillingID","giop-parlay.org.csapi.mmm.TpMessageTreatment.BillingID",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessageTreatment_ValidityTime, {"TpMessageTreatment_ValidityTime","giop-parlay.org.csapi.mmm.TpMessageTreatment.ValidityTime",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessagingEventCriteria_TpMessagingEventCriteria, {"TpMessagingEventCriteria","giop-parlay.org.csapi.mmm.TpMessagingEventCriteria",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagingEventName),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessagingEventCriteria_Dummy, {"TpMessagingEventCriteria_Dummy","giop-parlay.org.csapi.mmm.TpMessagingEventCriteria.Dummy",FT_INT16,BASE_DEC,NULL,0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessagingEventInfo_TpMessagingEventInfo, {"TpMessagingEventInfo","giop-parlay.org.csapi.mmm.TpMessagingEventInfo",FT_UINT32,BASE_DEC,VALS(org_csapi_mmm_TpMessagingEventName),0x0,NULL,HFILL}},
        {&hf_org_csapi_mmm_TpMessagingEventInfo_EventNameUndefined, {"TpMessagingEventInfo_EventNameUndefined","giop-parlay.org.csapi.mmm.TpMessagingEventInfo.EventNameUndefined",FT_STRING,BASE_NONE,NULL,0x0,NULL,HFILL}},

    };

    static ei_register_info ei[] = {
        { &ei_parlay_unknown_giop_msg, { "giop-parlay.unknown_giop_msg", PI_PROTOCOL, PI_WARN, "Unknown GIOP message", EXPFILL }},
        { &ei_parlay_unknown_exception, { "giop-parlay.unknown_exception", PI_PROTOCOL, PI_WARN, "Unknown exception", EXPFILL }},
        { &ei_parlay_unknown_reply_status, { "giop-parlay.unknown_reply_status", PI_PROTOCOL, PI_WARN, "Unknown reply status", EXPFILL }},
    };

    /* setup protocol subtree array */

    static gint *ett[] = {
        &ett_parlay,
    };

    expert_module_t* expert_parlay;


    /* Register the protocol name and description */
    proto_parlay = proto_register_protocol("Parlay Dissector Using GIOP API" , "PARLAY", "giop-parlay" );
    proto_register_field_array(proto_parlay, hf, array_length(hf));
    proto_register_subtree_array(ett, array_length(ett));

    expert_parlay = expert_register_protocol(proto_parlay);
    expert_register_field_array(expert_parlay, ei, array_length(ei));
}

/* register me as handler for these interfaces */
void proto_reg_handoff_giop_parlay(void)
{
    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/IpService", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/am/IpAccountManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/am/IpAppAccountManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cc/cccs/IpAppConfCall", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cc/cccs/IpAppConfCallControlManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cc/cccs/IpAppSubConfCall", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cc/cccs/IpConfCall", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cc/cccs/IpConfCallControlManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cc/cccs/IpSubConfCall", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cc/gccs/IpAppCall", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cc/gccs/IpAppCallControlManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cc/gccs/IpCall", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cc/gccs/IpCallControlManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cc/mmccs/IpAppMultiMediaCall", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cc/mmccs/IpAppMultiMediaCallControlManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cc/mmccs/IpAppMultiMediaCallLeg", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cc/mmccs/IpMultiMediaCall", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cc/mmccs/IpMultiMediaCallControlManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cc/mmccs/IpMultiMediaCallLeg", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cc/mmccs/IpMultiMediaStream", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cc/mpccs/IpAppCallLeg", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cc/mpccs/IpAppMultiPartyCall", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cc/mpccs/IpAppMultiPartyCallControlManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cc/mpccs/IpCallLeg", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cc/mpccs/IpMultiPartyCall", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cc/mpccs/IpMultiPartyCallControlManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cm/IpConnectivityManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cm/IpEnterpriseNetwork", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cm/IpEnterpriseNetworkSite", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cm/IpQoSMenu", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cm/IpQoSTemplate", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cm/IpVPrN", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cm/IpVPrP", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cs/IpAppChargingManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cs/IpAppChargingSession", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cs/IpChargingManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/cs/IpChargingSession", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/dsc/IpAppDataSession", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/dsc/IpAppDataSessionControlManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/dsc/IpDataSession", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/dsc/IpDataSessionControlManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_access/trust_and_security/IpAPILevelAuthentication", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_access/trust_and_security/IpAccess", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_access/trust_and_security/IpAuthentication", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_access/trust_and_security/IpClientAPILevelAuthentication", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_access/trust_and_security/IpClientAccess", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_access/trust_and_security/IpInitial", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_application/discovery/IpServiceDiscovery", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_application/integrity/IpAppFaultManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_application/integrity/IpAppHeartBeat", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_application/integrity/IpAppHeartBeatMgmt", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_application/integrity/IpAppLoadManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_application/integrity/IpAppOAM", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_application/integrity/IpFaultManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_application/integrity/IpHeartBeat", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_application/integrity/IpHeartBeatMgmt", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_application/integrity/IpLoadManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_application/integrity/IpOAM", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_application/notification/IpAppEventNotification", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_application/notification/IpEventNotification", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_application/service_agreement/IpAppServiceAgreementManagement", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_application/service_agreement/IpServiceAgreementManagement", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_enterprise_operator/notification/IpClientEventNotification", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_enterprise_operator/notification/IpEventNotification", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppInfoQuery", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_enterprise_operator/service_subscription/IpClientAppManagement", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_enterprise_operator/service_subscription/IpEntOpAccountInfoQuery", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_enterprise_operator/service_subscription/IpEntOpAccountManagement", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceContractInfoQuery", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceContractManagement", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceProfileInfoQuery", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_enterprise_operator/service_subscription/IpServiceProfileManagement", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_service/discovery/IpFwServiceDiscovery", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_service/integrity/IpFwFaultManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_service/integrity/IpFwHeartBeat", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_service/integrity/IpFwHeartBeatMgmt", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_service/integrity/IpFwLoadManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_service/integrity/IpFwOAM", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_service/integrity/IpSvcFaultManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_service/integrity/IpSvcHeartBeat", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_service/integrity/IpSvcHeartBeatMgmt", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_service/integrity/IpSvcLoadManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_service/integrity/IpSvcOAM", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_service/notification/IpFwEventNotification", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_service/notification/IpSvcEventNotification", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_service/service_lifecycle/IpServiceInstanceLifecycleManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/fw/fw_service/service_registration/IpFwServiceRegistration", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/gms/IpAppMessagingManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/gms/IpMailbox", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/gms/IpMailboxFolder", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/gms/IpMessage", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/gms/IpMessagingManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/mm/ub/IpAppUserBinding", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/mm/ub/IpUserBinding", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/mm/ul/IpAppTriggeredUserLocation", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/mm/ul/IpAppUserLocation", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/mm/ul/IpTriggeredUserLocation", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/mm/ul/IpUserLocation", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/mm/ulc/IpAppUserLocationCamel", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/mm/ulc/IpUserLocationCamel", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/mm/ule/IpAppUserLocationEmergency", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/mm/ule/IpUserLocationEmergency", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/mm/us/IpAppUserStatus", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/mm/us/IpUserStatus", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/mmm/IpAppMailbox", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/mmm/IpAppMultiMediaMessaging", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/mmm/IpAppMultiMediaMessagingManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/mmm/IpMailbox", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/mmm/IpMultiMediaMessaging", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/mmm/IpMultiMediaMessagingManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/pam/access/IpAppPAMPreferenceCheck", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/pam/access/IpPAMAgentPresence", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/pam/access/IpPAMAvailability", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/pam/access/IpPAMIdentityPresence", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/pam/access/IpPAMPresenceAvailabilityManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/pam/event/IpAppPAMEventHandler", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/pam/event/IpPAMEventHandler", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/pam/event/IpPAMEventManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/pam/provisioning/IpPAMAgentAssignment", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/pam/provisioning/IpPAMAgentManagement", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/pam/provisioning/IpPAMAgentTypeManagement", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/pam/provisioning/IpPAMCapabilityManagement", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/pam/provisioning/IpPAMIdentityManagement", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/pam/provisioning/IpPAMIdentityTypeManagement", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/pam/provisioning/IpPAMProvisioningManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/policy/evaluation/IpAppPolicyDomain", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/policy/evaluation/IpPolicyEvalManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/policy/provisioning/IpPolicy", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/policy/provisioning/IpPolicyAction", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/policy/provisioning/IpPolicyCondition", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/policy/provisioning/IpPolicyDomain", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/policy/provisioning/IpPolicyEventDefinition", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/policy/provisioning/IpPolicyGroup", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/policy/provisioning/IpPolicyIterator", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/policy/provisioning/IpPolicyManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/policy/provisioning/IpPolicyRepository", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/policy/provisioning/IpPolicyRule", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/policy/provisioning/IpPolicySignature", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/termcap/IpAppExtendedTerminalCapabilities", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/termcap/IpExtendedTerminalCapabilities", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/termcap/IpTerminalCapabilities", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/ui/IpAppUI", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/ui/IpAppUIAdminManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/ui/IpAppUICall", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/ui/IpAppUIManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/ui/IpUI", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/ui/IpUIAdminManager", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/ui/IpUICall", proto_parlay );     /* explicit dissector */

    /* Register for Explicit Dissection */
    register_giop_user_module(dissect_parlay, "PARLAY", "org/csapi/ui/IpUIManager", proto_parlay );     /* explicit dissector */

    /* Register for Heuristic Dissection */
    register_giop_user(dissect_parlay, "PARLAY" ,proto_parlay);     /* heuristic dissector */

}

/*
 * Editor modelines  -  https://www.wireshark.org/tools/modelines.html
 *
 * Local Variables:
 * c-basic-offset: 4
 * tab-width: 8
 * indent-tabs-mode: nil
 * End:
 *
 * ex: set shiftwidth=4 tabstop=8 expandtab:
 * :indentSize=4:tabSize=8:noTabs=true:
 */
